/* Automatically @generated by tree-sitter */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#ifdef _MSC_VER
#pragma optimize("", off)
#elif defined(__clang__)
#pragma clang optimize off
#elif defined(__GNUC__)
#pragma GCC optimize ("O0")
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 7350
#define LARGE_STATE_COUNT 1386
#define SYMBOL_COUNT 456
#define ALIAS_COUNT 13
#define TOKEN_COUNT 152
#define EXTERNAL_TOKEN_COUNT 7
#define FIELD_COUNT 39
#define MAX_ALIAS_SEQUENCE_LENGTH 11
#define MAX_RESERVED_WORD_SET_SIZE 55
#define PRODUCTION_ID_COUNT 231
#define SUPERTYPE_COUNT 32

enum ts_symbol_identifiers {
  sym__lowercase_identifier = 1,
  anon_sym_SEMI_SEMI = 2,
  anon_sym_let = 3,
  anon_sym_rec = 4,
  anon_sym_and = 5,
  anon_sym_EQ = 6,
  anon_sym_TILDE = 7,
  anon_sym_QMARK = 8,
  anon_sym_COLON = 9,
  anon_sym_LPAREN = 10,
  anon_sym_RPAREN = 11,
  anon_sym_external = 12,
  anon_sym_type = 13,
  anon_sym_nonrec = 14,
  anon_sym_COLON_EQ = 15,
  anon_sym_private = 16,
  anon_sym_DOT_DOT = 17,
  anon_sym_PLUS_EQ = 18,
  anon_sym_COMMA = 19,
  anon_sym_PLUS = 20,
  anon_sym_DASH = 21,
  anon_sym_BANG = 22,
  anon_sym__ = 23,
  anon_sym_PIPE = 24,
  anon_sym_of = 25,
  anon_sym_COLON2 = 26,
  anon_sym_DOT = 27,
  anon_sym_DASH_GT = 28,
  anon_sym_LBRACK = 29,
  anon_sym_RBRACK = 30,
  anon_sym_true = 31,
  anon_sym_false = 32,
  anon_sym_STAR = 33,
  anon_sym_LBRACE = 34,
  anon_sym_SEMI = 35,
  anon_sym_RBRACE = 36,
  anon_sym_mutable = 37,
  anon_sym_constraint = 38,
  anon_sym_exception = 39,
  anon_sym_module = 40,
  anon_sym_open = 41,
  anon_sym_include = 42,
  anon_sym_class = 43,
  anon_sym_virtual = 44,
  anon_sym_val = 45,
  anon_sym_sig = 46,
  anon_sym_end = 47,
  anon_sym_with = 48,
  anon_sym_functor = 49,
  anon_sym_struct = 50,
  anon_sym_COLON_GT = 51,
  anon_sym_object = 52,
  anon_sym_inherit = 53,
  anon_sym_method = 54,
  anon_sym_in = 55,
  anon_sym_fun = 56,
  anon_sym_as = 57,
  anon_sym_initializer = 58,
  anon_sym_LBRACK_GT = 59,
  anon_sym_LBRACK_LT = 60,
  anon_sym_GT = 61,
  anon_sym_AMP = 62,
  anon_sym_LT = 63,
  anon_sym_POUND = 64,
  anon_sym_COLON_COLON = 65,
  anon_sym_LBRACK_PIPE = 66,
  anon_sym_PIPE_RBRACK = 67,
  anon_sym_LT_DASH = 68,
  anon_sym_if = 69,
  anon_sym_then = 70,
  anon_sym_else = 71,
  anon_sym_while = 72,
  anon_sym_do = 73,
  anon_sym_done = 74,
  anon_sym_for = 75,
  anon_sym_to = 76,
  anon_sym_downto = 77,
  anon_sym_match = 78,
  anon_sym_when = 79,
  anon_sym_function = 80,
  anon_sym_try = 81,
  anon_sym_assert = 82,
  anon_sym_lazy = 83,
  anon_sym_new = 84,
  anon_sym_LBRACE_LT = 85,
  anon_sym_GT_RBRACE = 86,
  anon_sym_begin = 87,
  sym_ocamlyacc_value = 88,
  anon_sym_effect = 89,
  aux_sym_attribute_token1 = 90,
  anon_sym_LBRACK_AT_AT = 91,
  anon_sym_LBRACK_AT_AT_AT = 92,
  anon_sym_LBRACK_PERCENT = 93,
  anon_sym_LBRACE_PERCENT = 94,
  aux_sym_quoted_extension_token1 = 95,
  anon_sym_LBRACK_PERCENT_PERCENT = 96,
  anon_sym_LBRACE_PERCENT_PERCENT = 97,
  anon_sym_PERCENT = 98,
  aux_sym_number_token1 = 99,
  aux_sym_signed_number_token1 = 100,
  anon_sym_SQUOTE = 101,
  anon_sym_SQUOTE2 = 102,
  aux_sym_character_content_token1 = 103,
  aux_sym_character_content_token2 = 104,
  anon_sym_DQUOTE = 105,
  aux_sym_string_content_token1 = 106,
  aux_sym_string_content_token2 = 107,
  aux_sym_string_content_token3 = 108,
  aux_sym_string_content_token4 = 109,
  aux_sym_string_content_token5 = 110,
  aux_sym_quoted_string_content_token1 = 111,
  aux_sym_escape_sequence_token1 = 112,
  aux_sym_escape_sequence_token2 = 113,
  aux_sym_escape_sequence_token3 = 114,
  aux_sym_escape_sequence_token4 = 115,
  sym_conversion_specification = 116,
  sym_pretty_printing_indication = 117,
  sym_prefix_operator = 118,
  aux_sym_sign_operator_token1 = 119,
  sym_hash_operator = 120,
  anon_sym_lsl = 121,
  anon_sym_lsr = 122,
  anon_sym_asr = 123,
  aux_sym_pow_operator_token1 = 124,
  anon_sym_mod = 125,
  anon_sym_land = 126,
  anon_sym_lor = 127,
  anon_sym_lxor = 128,
  aux_sym_mult_operator_token1 = 129,
  aux_sym_add_operator_token1 = 130,
  sym_concat_operator = 131,
  sym_rel_operator = 132,
  anon_sym_AMP_AMP = 133,
  anon_sym_or = 134,
  anon_sym_PIPE_PIPE = 135,
  sym_assign_operator = 136,
  sym_indexing_operator = 137,
  sym_let_operator = 138,
  sym_let_and_operator = 139,
  sym_match_operator = 140,
  sym__uppercase_identifier = 141,
  aux_sym_directive_token1 = 142,
  aux_sym_type_variable_token1 = 143,
  aux_sym_tag_token1 = 144,
  aux_sym_attribute_id_token1 = 145,
  sym_comment = 146,
  sym__left_quoted_string_delimiter = 147,
  sym__right_quoted_string_delimiter = 148,
  sym_line_number_directive = 149,
  sym__null = 150,
  sym__error_sentinel = 151,
  sym_type = 152,
  sym__structure = 153,
  sym_expression_item = 154,
  sym__signature = 155,
  sym_toplevel_directive = 156,
  sym__local_structure_item = 157,
  sym__structure_item = 158,
  sym_value_definition = 159,
  sym_let_binding = 160,
  sym_parameter = 161,
  sym_external = 162,
  sym_type_definition = 163,
  sym_type_binding = 164,
  sym__type_params = 165,
  sym__type_param = 166,
  sym_variant_declaration = 167,
  sym_constructor_declaration = 168,
  sym__constructor_declaration_name = 169,
  sym__constructor_argument = 170,
  sym_record_declaration = 171,
  sym_field_declaration = 172,
  sym_external_declaration = 173,
  sym_type_constraint = 174,
  sym_exception_definition = 175,
  sym_module_definition = 176,
  sym_module_binding = 177,
  sym_module_parameter = 178,
  sym_module_type_definition = 179,
  sym_open_module = 180,
  sym_include_module = 181,
  sym_class_definition = 182,
  sym_class_binding = 183,
  sym_class_type_definition = 184,
  sym_class_type_binding = 185,
  sym__signature_item = 186,
  sym_value_specification = 187,
  sym_include_module_type = 188,
  sym__module_typed = 189,
  sym__module_type = 190,
  sym_signature = 191,
  sym_module_type_constraint = 192,
  sym_constrain_type = 193,
  sym_constrain_module = 194,
  sym_constrain_module_type = 195,
  sym_module_type_of = 196,
  sym_functor_type = 197,
  sym_parenthesized_module_type = 198,
  sym__simple_module_expression = 199,
  sym__module_expression = 200,
  sym_structure = 201,
  sym_functor = 202,
  sym_module_application = 203,
  sym_typed_module_expression = 204,
  sym_packed_module = 205,
  sym_parenthesized_module_expression = 206,
  sym__class_typed = 207,
  sym__simple_class_type = 208,
  sym__class_type = 209,
  sym_instantiated_class_type = 210,
  sym_class_body_type = 211,
  sym__class_field_specification = 212,
  sym_inheritance_specification = 213,
  sym_instance_variable_specification = 214,
  sym_method_specification = 215,
  sym_type_parameter_constraint = 216,
  sym_let_open_class_type = 217,
  sym_class_function_type = 218,
  sym__simple_class_expression = 219,
  sym__class_expression = 220,
  sym_instantiated_class = 221,
  sym_typed_class_expression = 222,
  sym_class_function = 223,
  sym_class_application = 224,
  sym_let_class_expression = 225,
  sym__class_field = 226,
  sym_inheritance_definition = 227,
  sym_instance_variable_definition = 228,
  sym_method_definition = 229,
  sym_class_initializer = 230,
  sym_let_open_class_expression = 231,
  sym_parenthesized_class_expression = 232,
  sym__typed = 233,
  sym__simple_typed = 234,
  sym__coerced = 235,
  sym__type_constrained = 236,
  sym__maybe_polymorphic_typed = 237,
  sym__polymorphic_typed = 238,
  sym__polymorphic_type = 239,
  sym_polymorphic_type = 240,
  sym__abstract_type = 241,
  sym__parenthesized_abstract_type = 242,
  sym__delimited_type = 243,
  sym__simple_type = 244,
  sym__type = 245,
  sym_function_type = 246,
  sym_labeled_argument_type = 247,
  sym__proper_tuple_type = 248,
  sym__labeled_tuple_type = 249,
  sym_labeled_tuple_element_type = 250,
  sym__tuple_type_rhs = 251,
  sym_constructed_type = 252,
  sym_aliased_type = 253,
  sym_local_open_type = 254,
  sym_polymorphic_variant_type = 255,
  sym__tag_spec = 256,
  sym_tag_specification = 257,
  sym_package_type = 258,
  sym_object_type = 259,
  sym_method_type = 260,
  sym_hash_type = 261,
  sym_parenthesized_type = 262,
  sym__delimited_expression = 263,
  sym__simple_expression = 264,
  sym__expression = 265,
  sym__sequence_expression = 266,
  sym_typed_expression = 267,
  sym_labeled_tuple_element = 268,
  sym__tuple_expression = 269,
  sym_cons_expression = 270,
  sym_list_expression = 271,
  sym_array_expression = 272,
  sym__sequence_expression_content = 273,
  sym_record_expression = 274,
  sym_field_expression = 275,
  sym_application_expression = 276,
  sym_labeled_argument = 277,
  sym_prefix_expression = 278,
  sym_sign_expression = 279,
  sym_hash_expression = 280,
  sym_infix_expression = 281,
  sym_field_get_expression = 282,
  sym_array_get_expression = 283,
  sym_string_get_expression = 284,
  sym_bigarray_get_expression = 285,
  sym_set_expression = 286,
  sym_if_expression = 287,
  sym_then_clause = 288,
  sym_else_clause = 289,
  sym_while_expression = 290,
  sym_do_clause = 291,
  sym_for_expression = 292,
  sym__sequence_expression_anonymous = 293,
  sym_match_expression = 294,
  sym__match_cases = 295,
  sym_match_case = 296,
  sym_guard = 297,
  sym_refutation_case = 298,
  sym_function_expression = 299,
  sym_fun_expression = 300,
  sym_try_expression = 301,
  sym_let_expression = 302,
  sym_assert_expression = 303,
  sym_lazy_expression = 304,
  sym_local_open_expression = 305,
  sym_package_expression = 306,
  sym_new_expression = 307,
  sym_object_copy_expression = 308,
  sym_instance_variable_expression = 309,
  sym_method_invocation = 310,
  sym_object_expression = 311,
  sym_parenthesized_expression = 312,
  sym__delimited_pattern = 313,
  sym__simple_pattern = 314,
  sym__effect_pattern = 315,
  sym__pattern = 316,
  sym__delimited_binding_pattern = 317,
  sym__simple_binding_pattern = 318,
  sym__effect_binding_pattern = 319,
  sym__binding_pattern_no_exn = 320,
  sym__binding_pattern = 321,
  sym_alias_pattern = 322,
  sym_alias_binding_pattern_no_exn = 323,
  sym_alias_binding_pattern = 324,
  sym_typed_pattern = 325,
  sym_typed_binding_pattern = 326,
  sym__or_pattern_anonymous = 327,
  sym__or_binding_pattern_no_exn_anonymous = 328,
  sym__or_binding_pattern_anonymous = 329,
  sym_constructor_pattern = 330,
  sym_constructor_binding_pattern = 331,
  sym_tag_pattern = 332,
  sym_tag_binding_pattern = 333,
  sym_polymorphic_variant_pattern = 334,
  sym_labeled_tuple_element_pattern = 335,
  sym__tuple_pattern = 336,
  sym_labeled_tuple_element_binding_pattern = 337,
  sym__tuple_binding_pattern_no_exn = 338,
  sym__tuple_binding_pattern = 339,
  sym_record_pattern = 340,
  sym_field_pattern = 341,
  sym_record_binding_pattern = 342,
  sym_field_binding_pattern = 343,
  sym_list_pattern = 344,
  sym_list_binding_pattern = 345,
  sym_cons_pattern = 346,
  sym_cons_binding_pattern_no_exn = 347,
  sym_cons_binding_pattern = 348,
  sym_array_pattern = 349,
  sym_array_binding_pattern = 350,
  sym__sequence_pattern_content = 351,
  sym__sequence_binding_pattern_content = 352,
  sym_range_pattern = 353,
  sym_lazy_pattern = 354,
  sym_lazy_binding_pattern = 355,
  sym_local_open_pattern = 356,
  sym_local_open_binding_pattern = 357,
  sym_package_pattern = 358,
  sym_parenthesized_pattern = 359,
  sym_parenthesized_binding_pattern = 360,
  sym_exception_pattern = 361,
  sym_exception_binding_pattern = 362,
  sym_effect_pattern = 363,
  sym_effect_binding_pattern = 364,
  sym_attribute = 365,
  sym_item_attribute = 366,
  sym_floating_attribute = 367,
  sym_attribute_payload = 368,
  sym_extension = 369,
  sym_quoted_extension = 370,
  sym_item_extension = 371,
  sym_quoted_item_extension = 372,
  sym__attribute = 373,
  sym__constant = 374,
  sym__signed_constant = 375,
  sym_number = 376,
  sym_signed_number = 377,
  sym_character = 378,
  sym_character_content = 379,
  sym_string = 380,
  sym_string_content = 381,
  sym_quoted_string = 382,
  sym__quoted_string = 383,
  sym_quoted_string_content = 384,
  sym_escape_sequence = 385,
  sym_boolean = 386,
  sym_unit = 387,
  sym_sign_operator = 388,
  sym__infix_operator = 389,
  sym_pow_operator = 390,
  sym_mult_operator = 391,
  sym_add_operator = 392,
  sym_and_operator = 393,
  sym_or_operator = 394,
  sym_indexing_operator_path = 395,
  sym__value_name = 396,
  sym__simple_value_pattern = 397,
  sym_parenthesized_operator = 398,
  sym_value_path = 399,
  sym_module_path = 400,
  sym_extended_module_path = 401,
  sym_module_type_path = 402,
  sym_field_path = 403,
  sym_constructor_path = 404,
  sym_type_constructor_path = 405,
  sym_class_path = 406,
  sym_class_type_path = 407,
  sym__instance_variable_name = 408,
  sym__constructor_name = 409,
  sym_directive = 410,
  sym_type_variable = 411,
  sym_tag = 412,
  sym_attribute_id = 413,
  aux_sym__structure_repeat1 = 414,
  aux_sym__structure_repeat2 = 415,
  aux_sym_expression_item_repeat1 = 416,
  aux_sym__signature_repeat1 = 417,
  aux_sym_value_definition_repeat1 = 418,
  aux_sym_let_binding_repeat1 = 419,
  aux_sym_external_repeat1 = 420,
  aux_sym_type_definition_repeat1 = 421,
  aux_sym_type_binding_repeat1 = 422,
  aux_sym__type_params_repeat1 = 423,
  aux_sym__type_param_repeat1 = 424,
  aux_sym_variant_declaration_repeat1 = 425,
  aux_sym_constructor_declaration_repeat1 = 426,
  aux_sym__constructor_argument_repeat1 = 427,
  aux_sym_record_declaration_repeat1 = 428,
  aux_sym_module_definition_repeat1 = 429,
  aux_sym_module_binding_repeat1 = 430,
  aux_sym_class_definition_repeat1 = 431,
  aux_sym_class_type_definition_repeat1 = 432,
  aux_sym_module_type_constraint_repeat1 = 433,
  aux_sym_instantiated_class_type_repeat1 = 434,
  aux_sym_class_body_type_repeat1 = 435,
  aux_sym_instance_variable_specification_repeat1 = 436,
  aux_sym_method_specification_repeat1 = 437,
  aux_sym_class_application_repeat1 = 438,
  aux_sym__abstract_type_repeat1 = 439,
  aux_sym_polymorphic_variant_type_repeat1 = 440,
  aux_sym_polymorphic_variant_type_repeat2 = 441,
  aux_sym_tag_specification_repeat1 = 442,
  aux_sym_object_type_repeat1 = 443,
  aux_sym__sequence_expression_content_repeat1 = 444,
  aux_sym_record_expression_repeat1 = 445,
  aux_sym__match_cases_repeat1 = 446,
  aux_sym_object_copy_expression_repeat1 = 447,
  aux_sym_object_expression_repeat1 = 448,
  aux_sym_record_pattern_repeat1 = 449,
  aux_sym_record_binding_pattern_repeat1 = 450,
  aux_sym__sequence_pattern_content_repeat1 = 451,
  aux_sym__sequence_binding_pattern_content_repeat1 = 452,
  aux_sym_string_content_repeat1 = 453,
  aux_sym_quoted_string_content_repeat1 = 454,
  aux_sym_attribute_id_repeat1 = 455,
  alias_sym_class_name = 456,
  alias_sym_class_type_name = 457,
  alias_sym_field_name = 458,
  alias_sym_instance_variable_name = 459,
  alias_sym_label_name = 460,
  alias_sym_method_name = 461,
  alias_sym_module_name = 462,
  alias_sym_module_type_name = 463,
  alias_sym_sequence_expression = 464,
  alias_sym_tuple_expression = 465,
  alias_sym_type_constructor = 466,
  alias_sym_value_name = 467,
  alias_sym_value_pattern = 468,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym__lowercase_identifier] = "_lowercase_identifier",
  [anon_sym_SEMI_SEMI] = ";;",
  [anon_sym_let] = "let",
  [anon_sym_rec] = "rec",
  [anon_sym_and] = "and",
  [anon_sym_EQ] = "=",
  [anon_sym_TILDE] = "~",
  [anon_sym_QMARK] = "\?",
  [anon_sym_COLON] = ":",
  [anon_sym_LPAREN] = "(",
  [anon_sym_RPAREN] = ")",
  [anon_sym_external] = "external",
  [anon_sym_type] = "type",
  [anon_sym_nonrec] = "nonrec",
  [anon_sym_COLON_EQ] = ":=",
  [anon_sym_private] = "private",
  [anon_sym_DOT_DOT] = "..",
  [anon_sym_PLUS_EQ] = "+=",
  [anon_sym_COMMA] = ",",
  [anon_sym_PLUS] = "+",
  [anon_sym_DASH] = "-",
  [anon_sym_BANG] = "!",
  [anon_sym__] = "_",
  [anon_sym_PIPE] = "|",
  [anon_sym_of] = "of",
  [anon_sym_COLON2] = ":",
  [anon_sym_DOT] = ".",
  [anon_sym_DASH_GT] = "->",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [anon_sym_true] = "true",
  [anon_sym_false] = "false",
  [anon_sym_STAR] = "*",
  [anon_sym_LBRACE] = "{",
  [anon_sym_SEMI] = ";",
  [anon_sym_RBRACE] = "}",
  [anon_sym_mutable] = "mutable",
  [anon_sym_constraint] = "constraint",
  [anon_sym_exception] = "exception",
  [anon_sym_module] = "module",
  [anon_sym_open] = "open",
  [anon_sym_include] = "include",
  [anon_sym_class] = "class",
  [anon_sym_virtual] = "virtual",
  [anon_sym_val] = "val",
  [anon_sym_sig] = "sig",
  [anon_sym_end] = "end",
  [anon_sym_with] = "with",
  [anon_sym_functor] = "functor",
  [anon_sym_struct] = "struct",
  [anon_sym_COLON_GT] = ":>",
  [anon_sym_object] = "object",
  [anon_sym_inherit] = "inherit",
  [anon_sym_method] = "method",
  [anon_sym_in] = "in",
  [anon_sym_fun] = "fun",
  [anon_sym_as] = "as",
  [anon_sym_initializer] = "initializer",
  [anon_sym_LBRACK_GT] = "[>",
  [anon_sym_LBRACK_LT] = "[<",
  [anon_sym_GT] = ">",
  [anon_sym_AMP] = "&",
  [anon_sym_LT] = "<",
  [anon_sym_POUND] = "#",
  [anon_sym_COLON_COLON] = "::",
  [anon_sym_LBRACK_PIPE] = "[|",
  [anon_sym_PIPE_RBRACK] = "|]",
  [anon_sym_LT_DASH] = "<-",
  [anon_sym_if] = "if",
  [anon_sym_then] = "then",
  [anon_sym_else] = "else",
  [anon_sym_while] = "while",
  [anon_sym_do] = "do",
  [anon_sym_done] = "done",
  [anon_sym_for] = "for",
  [anon_sym_to] = "to",
  [anon_sym_downto] = "downto",
  [anon_sym_match] = "match",
  [anon_sym_when] = "when",
  [anon_sym_function] = "function",
  [anon_sym_try] = "try",
  [anon_sym_assert] = "assert",
  [anon_sym_lazy] = "lazy",
  [anon_sym_new] = "new",
  [anon_sym_LBRACE_LT] = "{<",
  [anon_sym_GT_RBRACE] = ">}",
  [anon_sym_begin] = "begin",
  [sym_ocamlyacc_value] = "ocamlyacc_value",
  [anon_sym_effect] = "effect",
  [aux_sym_attribute_token1] = "[@",
  [anon_sym_LBRACK_AT_AT] = "[@@",
  [anon_sym_LBRACK_AT_AT_AT] = "[@@@",
  [anon_sym_LBRACK_PERCENT] = "[%",
  [anon_sym_LBRACE_PERCENT] = "{%",
  [aux_sym_quoted_extension_token1] = "quoted_extension_token1",
  [anon_sym_LBRACK_PERCENT_PERCENT] = "[%%",
  [anon_sym_LBRACE_PERCENT_PERCENT] = "{%%",
  [anon_sym_PERCENT] = "%",
  [aux_sym_number_token1] = "number_token1",
  [aux_sym_signed_number_token1] = "signed_number_token1",
  [anon_sym_SQUOTE] = "'",
  [anon_sym_SQUOTE2] = "'",
  [aux_sym_character_content_token1] = "character_content_token1",
  [aux_sym_character_content_token2] = "character_content_token2",
  [anon_sym_DQUOTE] = "\"",
  [aux_sym_string_content_token1] = "string_content_token1",
  [aux_sym_string_content_token2] = "string_content_token2",
  [aux_sym_string_content_token3] = "string_content_token3",
  [aux_sym_string_content_token4] = "escape_sequence",
  [aux_sym_string_content_token5] = "escape_sequence",
  [aux_sym_quoted_string_content_token1] = "quoted_string_content_token1",
  [aux_sym_escape_sequence_token1] = "escape_sequence_token1",
  [aux_sym_escape_sequence_token2] = "escape_sequence_token2",
  [aux_sym_escape_sequence_token3] = "escape_sequence_token3",
  [aux_sym_escape_sequence_token4] = "escape_sequence_token4",
  [sym_conversion_specification] = "conversion_specification",
  [sym_pretty_printing_indication] = "pretty_printing_indication",
  [sym_prefix_operator] = "prefix_operator",
  [aux_sym_sign_operator_token1] = "sign_operator_token1",
  [sym_hash_operator] = "hash_operator",
  [anon_sym_lsl] = "lsl",
  [anon_sym_lsr] = "lsr",
  [anon_sym_asr] = "asr",
  [aux_sym_pow_operator_token1] = "pow_operator_token1",
  [anon_sym_mod] = "mod",
  [anon_sym_land] = "land",
  [anon_sym_lor] = "lor",
  [anon_sym_lxor] = "lxor",
  [aux_sym_mult_operator_token1] = "mult_operator_token1",
  [aux_sym_add_operator_token1] = "add_operator_token1",
  [sym_concat_operator] = "concat_operator",
  [sym_rel_operator] = "rel_operator",
  [anon_sym_AMP_AMP] = "&&",
  [anon_sym_or] = "or",
  [anon_sym_PIPE_PIPE] = "||",
  [sym_assign_operator] = "assign_operator",
  [sym_indexing_operator] = "indexing_operator",
  [sym_let_operator] = "let_operator",
  [sym_let_and_operator] = "let_and_operator",
  [sym_match_operator] = "match_operator",
  [sym__uppercase_identifier] = "_uppercase_identifier",
  [aux_sym_directive_token1] = "directive_token1",
  [aux_sym_type_variable_token1] = "type_variable_token1",
  [aux_sym_tag_token1] = "tag_token1",
  [aux_sym_attribute_id_token1] = "attribute_id_token1",
  [sym_comment] = "comment",
  [sym__left_quoted_string_delimiter] = "_left_quoted_string_delimiter",
  [sym__right_quoted_string_delimiter] = "_right_quoted_string_delimiter",
  [sym_line_number_directive] = "line_number_directive",
  [sym__null] = "_null",
  [sym__error_sentinel] = "_error_sentinel",
  [sym_type] = "type",
  [sym__structure] = "_structure",
  [sym_expression_item] = "expression_item",
  [sym__signature] = "_signature",
  [sym_toplevel_directive] = "toplevel_directive",
  [sym__local_structure_item] = "_local_structure_item",
  [sym__structure_item] = "_structure_item",
  [sym_value_definition] = "value_definition",
  [sym_let_binding] = "let_binding",
  [sym_parameter] = "parameter",
  [sym_external] = "external",
  [sym_type_definition] = "type_definition",
  [sym_type_binding] = "type_binding",
  [sym__type_params] = "_type_params",
  [sym__type_param] = "_type_param",
  [sym_variant_declaration] = "variant_declaration",
  [sym_constructor_declaration] = "constructor_declaration",
  [sym__constructor_declaration_name] = "constructor_name",
  [sym__constructor_argument] = "_constructor_argument",
  [sym_record_declaration] = "record_declaration",
  [sym_field_declaration] = "field_declaration",
  [sym_external_declaration] = "external_declaration",
  [sym_type_constraint] = "type_constraint",
  [sym_exception_definition] = "exception_definition",
  [sym_module_definition] = "module_definition",
  [sym_module_binding] = "module_binding",
  [sym_module_parameter] = "module_parameter",
  [sym_module_type_definition] = "module_type_definition",
  [sym_open_module] = "open_module",
  [sym_include_module] = "include_module",
  [sym_class_definition] = "class_definition",
  [sym_class_binding] = "class_binding",
  [sym_class_type_definition] = "class_type_definition",
  [sym_class_type_binding] = "class_type_binding",
  [sym__signature_item] = "_signature_item",
  [sym_value_specification] = "value_specification",
  [sym_include_module_type] = "include_module_type",
  [sym__module_typed] = "_module_typed",
  [sym__module_type] = "_module_type",
  [sym_signature] = "signature",
  [sym_module_type_constraint] = "module_type_constraint",
  [sym_constrain_type] = "constrain_type",
  [sym_constrain_module] = "constrain_module",
  [sym_constrain_module_type] = "constrain_module_type",
  [sym_module_type_of] = "module_type_of",
  [sym_functor_type] = "functor_type",
  [sym_parenthesized_module_type] = "parenthesized_module_type",
  [sym__simple_module_expression] = "_simple_module_expression",
  [sym__module_expression] = "_module_expression",
  [sym_structure] = "structure",
  [sym_functor] = "functor",
  [sym_module_application] = "module_application",
  [sym_typed_module_expression] = "typed_module_expression",
  [sym_packed_module] = "packed_module",
  [sym_parenthesized_module_expression] = "parenthesized_module_expression",
  [sym__class_typed] = "_class_typed",
  [sym__simple_class_type] = "_simple_class_type",
  [sym__class_type] = "_class_type",
  [sym_instantiated_class_type] = "instantiated_class_type",
  [sym_class_body_type] = "class_body_type",
  [sym__class_field_specification] = "_class_field_specification",
  [sym_inheritance_specification] = "inheritance_specification",
  [sym_instance_variable_specification] = "instance_variable_specification",
  [sym_method_specification] = "method_specification",
  [sym_type_parameter_constraint] = "type_parameter_constraint",
  [sym_let_open_class_type] = "let_open_class_type",
  [sym_class_function_type] = "class_function_type",
  [sym__simple_class_expression] = "_simple_class_expression",
  [sym__class_expression] = "_class_expression",
  [sym_instantiated_class] = "instantiated_class",
  [sym_typed_class_expression] = "typed_class_expression",
  [sym_class_function] = "class_function",
  [sym_class_application] = "class_application",
  [sym_let_class_expression] = "let_class_expression",
  [sym__class_field] = "_class_field",
  [sym_inheritance_definition] = "inheritance_definition",
  [sym_instance_variable_definition] = "instance_variable_definition",
  [sym_method_definition] = "method_definition",
  [sym_class_initializer] = "class_initializer",
  [sym_let_open_class_expression] = "let_open_class_expression",
  [sym_parenthesized_class_expression] = "parenthesized_class_expression",
  [sym__typed] = "_typed",
  [sym__simple_typed] = "_simple_typed",
  [sym__coerced] = "_coerced",
  [sym__type_constrained] = "_type_constrained",
  [sym__maybe_polymorphic_typed] = "_maybe_polymorphic_typed",
  [sym__polymorphic_typed] = "_polymorphic_typed",
  [sym__polymorphic_type] = "_polymorphic_type",
  [sym_polymorphic_type] = "polymorphic_type",
  [sym__abstract_type] = "_abstract_type",
  [sym__parenthesized_abstract_type] = "abstract_type",
  [sym__delimited_type] = "_delimited_type",
  [sym__simple_type] = "_simple_type",
  [sym__type] = "_type",
  [sym_function_type] = "function_type",
  [sym_labeled_argument_type] = "labeled_argument_type",
  [sym__proper_tuple_type] = "tuple_type",
  [sym__labeled_tuple_type] = "tuple_type",
  [sym_labeled_tuple_element_type] = "labeled_tuple_element_type",
  [sym__tuple_type_rhs] = "_tuple_type_rhs",
  [sym_constructed_type] = "constructed_type",
  [sym_aliased_type] = "aliased_type",
  [sym_local_open_type] = "local_open_type",
  [sym_polymorphic_variant_type] = "polymorphic_variant_type",
  [sym__tag_spec] = "_tag_spec",
  [sym_tag_specification] = "tag_specification",
  [sym_package_type] = "package_type",
  [sym_object_type] = "object_type",
  [sym_method_type] = "method_type",
  [sym_hash_type] = "hash_type",
  [sym_parenthesized_type] = "parenthesized_type",
  [sym__delimited_expression] = "_delimited_expression",
  [sym__simple_expression] = "_simple_expression",
  [sym__expression] = "_expression",
  [sym__sequence_expression] = "_sequence_expression",
  [sym_typed_expression] = "typed_expression",
  [sym_labeled_tuple_element] = "labeled_tuple_element",
  [sym__tuple_expression] = "_tuple_expression",
  [sym_cons_expression] = "cons_expression",
  [sym_list_expression] = "list_expression",
  [sym_array_expression] = "array_expression",
  [sym__sequence_expression_content] = "_sequence_expression_content",
  [sym_record_expression] = "record_expression",
  [sym_field_expression] = "field_expression",
  [sym_application_expression] = "application_expression",
  [sym_labeled_argument] = "labeled_argument",
  [sym_prefix_expression] = "prefix_expression",
  [sym_sign_expression] = "sign_expression",
  [sym_hash_expression] = "hash_expression",
  [sym_infix_expression] = "infix_expression",
  [sym_field_get_expression] = "field_get_expression",
  [sym_array_get_expression] = "array_get_expression",
  [sym_string_get_expression] = "string_get_expression",
  [sym_bigarray_get_expression] = "bigarray_get_expression",
  [sym_set_expression] = "set_expression",
  [sym_if_expression] = "if_expression",
  [sym_then_clause] = "then_clause",
  [sym_else_clause] = "else_clause",
  [sym_while_expression] = "while_expression",
  [sym_do_clause] = "do_clause",
  [sym_for_expression] = "for_expression",
  [sym__sequence_expression_anonymous] = "_sequence_expression_anonymous",
  [sym_match_expression] = "match_expression",
  [sym__match_cases] = "_match_cases",
  [sym_match_case] = "match_case",
  [sym_guard] = "guard",
  [sym_refutation_case] = "refutation_case",
  [sym_function_expression] = "function_expression",
  [sym_fun_expression] = "fun_expression",
  [sym_try_expression] = "try_expression",
  [sym_let_expression] = "let_expression",
  [sym_assert_expression] = "assert_expression",
  [sym_lazy_expression] = "lazy_expression",
  [sym_local_open_expression] = "local_open_expression",
  [sym_package_expression] = "package_expression",
  [sym_new_expression] = "new_expression",
  [sym_object_copy_expression] = "object_copy_expression",
  [sym_instance_variable_expression] = "instance_variable_expression",
  [sym_method_invocation] = "method_invocation",
  [sym_object_expression] = "object_expression",
  [sym_parenthesized_expression] = "parenthesized_expression",
  [sym__delimited_pattern] = "_delimited_pattern",
  [sym__simple_pattern] = "_simple_pattern",
  [sym__effect_pattern] = "_effect_pattern",
  [sym__pattern] = "_pattern",
  [sym__delimited_binding_pattern] = "_delimited_binding_pattern",
  [sym__simple_binding_pattern] = "_simple_binding_pattern",
  [sym__effect_binding_pattern] = "_effect_binding_pattern",
  [sym__binding_pattern_no_exn] = "_binding_pattern_no_exn",
  [sym__binding_pattern] = "_binding_pattern",
  [sym_alias_pattern] = "alias_pattern",
  [sym_alias_binding_pattern_no_exn] = "alias_pattern",
  [sym_alias_binding_pattern] = "alias_pattern",
  [sym_typed_pattern] = "typed_pattern",
  [sym_typed_binding_pattern] = "typed_pattern",
  [sym__or_pattern_anonymous] = "_or_pattern_anonymous",
  [sym__or_binding_pattern_no_exn_anonymous] = "or_pattern",
  [sym__or_binding_pattern_anonymous] = "_or_binding_pattern_anonymous",
  [sym_constructor_pattern] = "constructor_pattern",
  [sym_constructor_binding_pattern] = "constructor_pattern",
  [sym_tag_pattern] = "tag_pattern",
  [sym_tag_binding_pattern] = "tag_pattern",
  [sym_polymorphic_variant_pattern] = "polymorphic_variant_pattern",
  [sym_labeled_tuple_element_pattern] = "labeled_tuple_element_pattern",
  [sym__tuple_pattern] = "_tuple_pattern",
  [sym_labeled_tuple_element_binding_pattern] = "labeled_tuple_element_binding_pattern",
  [sym__tuple_binding_pattern_no_exn] = "tuple_pattern",
  [sym__tuple_binding_pattern] = "_tuple_binding_pattern",
  [sym_record_pattern] = "record_pattern",
  [sym_field_pattern] = "field_pattern",
  [sym_record_binding_pattern] = "record_pattern",
  [sym_field_binding_pattern] = "field_pattern",
  [sym_list_pattern] = "list_pattern",
  [sym_list_binding_pattern] = "list_pattern",
  [sym_cons_pattern] = "cons_pattern",
  [sym_cons_binding_pattern_no_exn] = "cons_pattern",
  [sym_cons_binding_pattern] = "cons_pattern",
  [sym_array_pattern] = "array_pattern",
  [sym_array_binding_pattern] = "array_pattern",
  [sym__sequence_pattern_content] = "_sequence_pattern_content",
  [sym__sequence_binding_pattern_content] = "_sequence_binding_pattern_content",
  [sym_range_pattern] = "range_pattern",
  [sym_lazy_pattern] = "lazy_pattern",
  [sym_lazy_binding_pattern] = "lazy_pattern",
  [sym_local_open_pattern] = "local_open_pattern",
  [sym_local_open_binding_pattern] = "local_open_pattern",
  [sym_package_pattern] = "package_pattern",
  [sym_parenthesized_pattern] = "parenthesized_pattern",
  [sym_parenthesized_binding_pattern] = "parenthesized_pattern",
  [sym_exception_pattern] = "exception_pattern",
  [sym_exception_binding_pattern] = "exception_pattern",
  [sym_effect_pattern] = "effect_pattern",
  [sym_effect_binding_pattern] = "effect_pattern",
  [sym_attribute] = "attribute",
  [sym_item_attribute] = "item_attribute",
  [sym_floating_attribute] = "floating_attribute",
  [sym_attribute_payload] = "attribute_payload",
  [sym_extension] = "extension",
  [sym_quoted_extension] = "quoted_extension",
  [sym_item_extension] = "item_extension",
  [sym_quoted_item_extension] = "quoted_item_extension",
  [sym__attribute] = "_attribute",
  [sym__constant] = "_constant",
  [sym__signed_constant] = "_signed_constant",
  [sym_number] = "number",
  [sym_signed_number] = "signed_number",
  [sym_character] = "character",
  [sym_character_content] = "character_content",
  [sym_string] = "string",
  [sym_string_content] = "string_content",
  [sym_quoted_string] = "quoted_string",
  [sym__quoted_string] = "_quoted_string",
  [sym_quoted_string_content] = "quoted_string_content",
  [sym_escape_sequence] = "escape_sequence",
  [sym_boolean] = "boolean",
  [sym_unit] = "unit",
  [sym_sign_operator] = "sign_operator",
  [sym__infix_operator] = "_infix_operator",
  [sym_pow_operator] = "pow_operator",
  [sym_mult_operator] = "mult_operator",
  [sym_add_operator] = "add_operator",
  [sym_and_operator] = "and_operator",
  [sym_or_operator] = "or_operator",
  [sym_indexing_operator_path] = "indexing_operator_path",
  [sym__value_name] = "_value_name",
  [sym__simple_value_pattern] = "_simple_value_pattern",
  [sym_parenthesized_operator] = "parenthesized_operator",
  [sym_value_path] = "value_path",
  [sym_module_path] = "module_path",
  [sym_extended_module_path] = "extended_module_path",
  [sym_module_type_path] = "module_type_path",
  [sym_field_path] = "field_path",
  [sym_constructor_path] = "constructor_path",
  [sym_type_constructor_path] = "type_constructor_path",
  [sym_class_path] = "class_path",
  [sym_class_type_path] = "class_type_path",
  [sym__instance_variable_name] = "_instance_variable_name",
  [sym__constructor_name] = "_constructor_name",
  [sym_directive] = "directive",
  [sym_type_variable] = "type_variable",
  [sym_tag] = "tag",
  [sym_attribute_id] = "attribute_id",
  [aux_sym__structure_repeat1] = "_structure_repeat1",
  [aux_sym__structure_repeat2] = "_structure_repeat2",
  [aux_sym_expression_item_repeat1] = "expression_item_repeat1",
  [aux_sym__signature_repeat1] = "_signature_repeat1",
  [aux_sym_value_definition_repeat1] = "value_definition_repeat1",
  [aux_sym_let_binding_repeat1] = "let_binding_repeat1",
  [aux_sym_external_repeat1] = "external_repeat1",
  [aux_sym_type_definition_repeat1] = "type_definition_repeat1",
  [aux_sym_type_binding_repeat1] = "type_binding_repeat1",
  [aux_sym__type_params_repeat1] = "_type_params_repeat1",
  [aux_sym__type_param_repeat1] = "_type_param_repeat1",
  [aux_sym_variant_declaration_repeat1] = "variant_declaration_repeat1",
  [aux_sym_constructor_declaration_repeat1] = "constructor_declaration_repeat1",
  [aux_sym__constructor_argument_repeat1] = "_constructor_argument_repeat1",
  [aux_sym_record_declaration_repeat1] = "record_declaration_repeat1",
  [aux_sym_module_definition_repeat1] = "module_definition_repeat1",
  [aux_sym_module_binding_repeat1] = "module_binding_repeat1",
  [aux_sym_class_definition_repeat1] = "class_definition_repeat1",
  [aux_sym_class_type_definition_repeat1] = "class_type_definition_repeat1",
  [aux_sym_module_type_constraint_repeat1] = "module_type_constraint_repeat1",
  [aux_sym_instantiated_class_type_repeat1] = "instantiated_class_type_repeat1",
  [aux_sym_class_body_type_repeat1] = "class_body_type_repeat1",
  [aux_sym_instance_variable_specification_repeat1] = "instance_variable_specification_repeat1",
  [aux_sym_method_specification_repeat1] = "method_specification_repeat1",
  [aux_sym_class_application_repeat1] = "class_application_repeat1",
  [aux_sym__abstract_type_repeat1] = "_abstract_type_repeat1",
  [aux_sym_polymorphic_variant_type_repeat1] = "polymorphic_variant_type_repeat1",
  [aux_sym_polymorphic_variant_type_repeat2] = "polymorphic_variant_type_repeat2",
  [aux_sym_tag_specification_repeat1] = "tag_specification_repeat1",
  [aux_sym_object_type_repeat1] = "object_type_repeat1",
  [aux_sym__sequence_expression_content_repeat1] = "_sequence_expression_content_repeat1",
  [aux_sym_record_expression_repeat1] = "record_expression_repeat1",
  [aux_sym__match_cases_repeat1] = "_match_cases_repeat1",
  [aux_sym_object_copy_expression_repeat1] = "object_copy_expression_repeat1",
  [aux_sym_object_expression_repeat1] = "object_expression_repeat1",
  [aux_sym_record_pattern_repeat1] = "record_pattern_repeat1",
  [aux_sym_record_binding_pattern_repeat1] = "record_binding_pattern_repeat1",
  [aux_sym__sequence_pattern_content_repeat1] = "_sequence_pattern_content_repeat1",
  [aux_sym__sequence_binding_pattern_content_repeat1] = "_sequence_binding_pattern_content_repeat1",
  [aux_sym_string_content_repeat1] = "string_content_repeat1",
  [aux_sym_quoted_string_content_repeat1] = "quoted_string_content_repeat1",
  [aux_sym_attribute_id_repeat1] = "attribute_id_repeat1",
  [alias_sym_class_name] = "class_name",
  [alias_sym_class_type_name] = "class_type_name",
  [alias_sym_field_name] = "field_name",
  [alias_sym_instance_variable_name] = "instance_variable_name",
  [alias_sym_label_name] = "label_name",
  [alias_sym_method_name] = "method_name",
  [alias_sym_module_name] = "module_name",
  [alias_sym_module_type_name] = "module_type_name",
  [alias_sym_sequence_expression] = "sequence_expression",
  [alias_sym_tuple_expression] = "tuple_expression",
  [alias_sym_type_constructor] = "type_constructor",
  [alias_sym_value_name] = "value_name",
  [alias_sym_value_pattern] = "value_pattern",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym__lowercase_identifier] = sym__lowercase_identifier,
  [anon_sym_SEMI_SEMI] = anon_sym_SEMI_SEMI,
  [anon_sym_let] = anon_sym_let,
  [anon_sym_rec] = anon_sym_rec,
  [anon_sym_and] = anon_sym_and,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_TILDE] = anon_sym_TILDE,
  [anon_sym_QMARK] = anon_sym_QMARK,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [anon_sym_external] = anon_sym_external,
  [anon_sym_type] = anon_sym_type,
  [anon_sym_nonrec] = anon_sym_nonrec,
  [anon_sym_COLON_EQ] = anon_sym_COLON_EQ,
  [anon_sym_private] = anon_sym_private,
  [anon_sym_DOT_DOT] = anon_sym_DOT_DOT,
  [anon_sym_PLUS_EQ] = anon_sym_PLUS_EQ,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_BANG] = anon_sym_BANG,
  [anon_sym__] = anon_sym__,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_of] = anon_sym_of,
  [anon_sym_COLON2] = anon_sym_COLON,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_DASH_GT] = anon_sym_DASH_GT,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym_true] = anon_sym_true,
  [anon_sym_false] = anon_sym_false,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_SEMI] = anon_sym_SEMI,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [anon_sym_mutable] = anon_sym_mutable,
  [anon_sym_constraint] = anon_sym_constraint,
  [anon_sym_exception] = anon_sym_exception,
  [anon_sym_module] = anon_sym_module,
  [anon_sym_open] = anon_sym_open,
  [anon_sym_include] = anon_sym_include,
  [anon_sym_class] = anon_sym_class,
  [anon_sym_virtual] = anon_sym_virtual,
  [anon_sym_val] = anon_sym_val,
  [anon_sym_sig] = anon_sym_sig,
  [anon_sym_end] = anon_sym_end,
  [anon_sym_with] = anon_sym_with,
  [anon_sym_functor] = anon_sym_functor,
  [anon_sym_struct] = anon_sym_struct,
  [anon_sym_COLON_GT] = anon_sym_COLON_GT,
  [anon_sym_object] = anon_sym_object,
  [anon_sym_inherit] = anon_sym_inherit,
  [anon_sym_method] = anon_sym_method,
  [anon_sym_in] = anon_sym_in,
  [anon_sym_fun] = anon_sym_fun,
  [anon_sym_as] = anon_sym_as,
  [anon_sym_initializer] = anon_sym_initializer,
  [anon_sym_LBRACK_GT] = anon_sym_LBRACK_GT,
  [anon_sym_LBRACK_LT] = anon_sym_LBRACK_LT,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_AMP] = anon_sym_AMP,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_POUND] = anon_sym_POUND,
  [anon_sym_COLON_COLON] = anon_sym_COLON_COLON,
  [anon_sym_LBRACK_PIPE] = anon_sym_LBRACK_PIPE,
  [anon_sym_PIPE_RBRACK] = anon_sym_PIPE_RBRACK,
  [anon_sym_LT_DASH] = anon_sym_LT_DASH,
  [anon_sym_if] = anon_sym_if,
  [anon_sym_then] = anon_sym_then,
  [anon_sym_else] = anon_sym_else,
  [anon_sym_while] = anon_sym_while,
  [anon_sym_do] = anon_sym_do,
  [anon_sym_done] = anon_sym_done,
  [anon_sym_for] = anon_sym_for,
  [anon_sym_to] = anon_sym_to,
  [anon_sym_downto] = anon_sym_downto,
  [anon_sym_match] = anon_sym_match,
  [anon_sym_when] = anon_sym_when,
  [anon_sym_function] = anon_sym_function,
  [anon_sym_try] = anon_sym_try,
  [anon_sym_assert] = anon_sym_assert,
  [anon_sym_lazy] = anon_sym_lazy,
  [anon_sym_new] = anon_sym_new,
  [anon_sym_LBRACE_LT] = anon_sym_LBRACE_LT,
  [anon_sym_GT_RBRACE] = anon_sym_GT_RBRACE,
  [anon_sym_begin] = anon_sym_begin,
  [sym_ocamlyacc_value] = sym_ocamlyacc_value,
  [anon_sym_effect] = anon_sym_effect,
  [aux_sym_attribute_token1] = aux_sym_attribute_token1,
  [anon_sym_LBRACK_AT_AT] = anon_sym_LBRACK_AT_AT,
  [anon_sym_LBRACK_AT_AT_AT] = anon_sym_LBRACK_AT_AT_AT,
  [anon_sym_LBRACK_PERCENT] = anon_sym_LBRACK_PERCENT,
  [anon_sym_LBRACE_PERCENT] = anon_sym_LBRACE_PERCENT,
  [aux_sym_quoted_extension_token1] = aux_sym_quoted_extension_token1,
  [anon_sym_LBRACK_PERCENT_PERCENT] = anon_sym_LBRACK_PERCENT_PERCENT,
  [anon_sym_LBRACE_PERCENT_PERCENT] = anon_sym_LBRACE_PERCENT_PERCENT,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [aux_sym_number_token1] = aux_sym_number_token1,
  [aux_sym_signed_number_token1] = aux_sym_signed_number_token1,
  [anon_sym_SQUOTE] = anon_sym_SQUOTE,
  [anon_sym_SQUOTE2] = anon_sym_SQUOTE,
  [aux_sym_character_content_token1] = aux_sym_character_content_token1,
  [aux_sym_character_content_token2] = aux_sym_character_content_token2,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [aux_sym_string_content_token1] = aux_sym_string_content_token1,
  [aux_sym_string_content_token2] = aux_sym_string_content_token2,
  [aux_sym_string_content_token3] = aux_sym_string_content_token3,
  [aux_sym_string_content_token4] = sym_escape_sequence,
  [aux_sym_string_content_token5] = sym_escape_sequence,
  [aux_sym_quoted_string_content_token1] = aux_sym_quoted_string_content_token1,
  [aux_sym_escape_sequence_token1] = aux_sym_escape_sequence_token1,
  [aux_sym_escape_sequence_token2] = aux_sym_escape_sequence_token2,
  [aux_sym_escape_sequence_token3] = aux_sym_escape_sequence_token3,
  [aux_sym_escape_sequence_token4] = aux_sym_escape_sequence_token4,
  [sym_conversion_specification] = sym_conversion_specification,
  [sym_pretty_printing_indication] = sym_pretty_printing_indication,
  [sym_prefix_operator] = sym_prefix_operator,
  [aux_sym_sign_operator_token1] = aux_sym_sign_operator_token1,
  [sym_hash_operator] = sym_hash_operator,
  [anon_sym_lsl] = anon_sym_lsl,
  [anon_sym_lsr] = anon_sym_lsr,
  [anon_sym_asr] = anon_sym_asr,
  [aux_sym_pow_operator_token1] = aux_sym_pow_operator_token1,
  [anon_sym_mod] = anon_sym_mod,
  [anon_sym_land] = anon_sym_land,
  [anon_sym_lor] = anon_sym_lor,
  [anon_sym_lxor] = anon_sym_lxor,
  [aux_sym_mult_operator_token1] = aux_sym_mult_operator_token1,
  [aux_sym_add_operator_token1] = aux_sym_add_operator_token1,
  [sym_concat_operator] = sym_concat_operator,
  [sym_rel_operator] = sym_rel_operator,
  [anon_sym_AMP_AMP] = anon_sym_AMP_AMP,
  [anon_sym_or] = anon_sym_or,
  [anon_sym_PIPE_PIPE] = anon_sym_PIPE_PIPE,
  [sym_assign_operator] = sym_assign_operator,
  [sym_indexing_operator] = sym_indexing_operator,
  [sym_let_operator] = sym_let_operator,
  [sym_let_and_operator] = sym_let_and_operator,
  [sym_match_operator] = sym_match_operator,
  [sym__uppercase_identifier] = sym__uppercase_identifier,
  [aux_sym_directive_token1] = aux_sym_directive_token1,
  [aux_sym_type_variable_token1] = aux_sym_type_variable_token1,
  [aux_sym_tag_token1] = aux_sym_tag_token1,
  [aux_sym_attribute_id_token1] = aux_sym_attribute_id_token1,
  [sym_comment] = sym_comment,
  [sym__left_quoted_string_delimiter] = sym__left_quoted_string_delimiter,
  [sym__right_quoted_string_delimiter] = sym__right_quoted_string_delimiter,
  [sym_line_number_directive] = sym_line_number_directive,
  [sym__null] = sym__null,
  [sym__error_sentinel] = sym__error_sentinel,
  [sym_type] = sym_type,
  [sym__structure] = sym__structure,
  [sym_expression_item] = sym_expression_item,
  [sym__signature] = sym__signature,
  [sym_toplevel_directive] = sym_toplevel_directive,
  [sym__local_structure_item] = sym__local_structure_item,
  [sym__structure_item] = sym__structure_item,
  [sym_value_definition] = sym_value_definition,
  [sym_let_binding] = sym_let_binding,
  [sym_parameter] = sym_parameter,
  [sym_external] = sym_external,
  [sym_type_definition] = sym_type_definition,
  [sym_type_binding] = sym_type_binding,
  [sym__type_params] = sym__type_params,
  [sym__type_param] = sym__type_param,
  [sym_variant_declaration] = sym_variant_declaration,
  [sym_constructor_declaration] = sym_constructor_declaration,
  [sym__constructor_declaration_name] = sym__constructor_declaration_name,
  [sym__constructor_argument] = sym__constructor_argument,
  [sym_record_declaration] = sym_record_declaration,
  [sym_field_declaration] = sym_field_declaration,
  [sym_external_declaration] = sym_external_declaration,
  [sym_type_constraint] = sym_type_constraint,
  [sym_exception_definition] = sym_exception_definition,
  [sym_module_definition] = sym_module_definition,
  [sym_module_binding] = sym_module_binding,
  [sym_module_parameter] = sym_module_parameter,
  [sym_module_type_definition] = sym_module_type_definition,
  [sym_open_module] = sym_open_module,
  [sym_include_module] = sym_include_module,
  [sym_class_definition] = sym_class_definition,
  [sym_class_binding] = sym_class_binding,
  [sym_class_type_definition] = sym_class_type_definition,
  [sym_class_type_binding] = sym_class_type_binding,
  [sym__signature_item] = sym__signature_item,
  [sym_value_specification] = sym_value_specification,
  [sym_include_module_type] = sym_include_module_type,
  [sym__module_typed] = sym__module_typed,
  [sym__module_type] = sym__module_type,
  [sym_signature] = sym_signature,
  [sym_module_type_constraint] = sym_module_type_constraint,
  [sym_constrain_type] = sym_constrain_type,
  [sym_constrain_module] = sym_constrain_module,
  [sym_constrain_module_type] = sym_constrain_module_type,
  [sym_module_type_of] = sym_module_type_of,
  [sym_functor_type] = sym_functor_type,
  [sym_parenthesized_module_type] = sym_parenthesized_module_type,
  [sym__simple_module_expression] = sym__simple_module_expression,
  [sym__module_expression] = sym__module_expression,
  [sym_structure] = sym_structure,
  [sym_functor] = sym_functor,
  [sym_module_application] = sym_module_application,
  [sym_typed_module_expression] = sym_typed_module_expression,
  [sym_packed_module] = sym_packed_module,
  [sym_parenthesized_module_expression] = sym_parenthesized_module_expression,
  [sym__class_typed] = sym__class_typed,
  [sym__simple_class_type] = sym__simple_class_type,
  [sym__class_type] = sym__class_type,
  [sym_instantiated_class_type] = sym_instantiated_class_type,
  [sym_class_body_type] = sym_class_body_type,
  [sym__class_field_specification] = sym__class_field_specification,
  [sym_inheritance_specification] = sym_inheritance_specification,
  [sym_instance_variable_specification] = sym_instance_variable_specification,
  [sym_method_specification] = sym_method_specification,
  [sym_type_parameter_constraint] = sym_type_parameter_constraint,
  [sym_let_open_class_type] = sym_let_open_class_type,
  [sym_class_function_type] = sym_class_function_type,
  [sym__simple_class_expression] = sym__simple_class_expression,
  [sym__class_expression] = sym__class_expression,
  [sym_instantiated_class] = sym_instantiated_class,
  [sym_typed_class_expression] = sym_typed_class_expression,
  [sym_class_function] = sym_class_function,
  [sym_class_application] = sym_class_application,
  [sym_let_class_expression] = sym_let_class_expression,
  [sym__class_field] = sym__class_field,
  [sym_inheritance_definition] = sym_inheritance_definition,
  [sym_instance_variable_definition] = sym_instance_variable_definition,
  [sym_method_definition] = sym_method_definition,
  [sym_class_initializer] = sym_class_initializer,
  [sym_let_open_class_expression] = sym_let_open_class_expression,
  [sym_parenthesized_class_expression] = sym_parenthesized_class_expression,
  [sym__typed] = sym__typed,
  [sym__simple_typed] = sym__simple_typed,
  [sym__coerced] = sym__coerced,
  [sym__type_constrained] = sym__type_constrained,
  [sym__maybe_polymorphic_typed] = sym__maybe_polymorphic_typed,
  [sym__polymorphic_typed] = sym__polymorphic_typed,
  [sym__polymorphic_type] = sym__polymorphic_type,
  [sym_polymorphic_type] = sym_polymorphic_type,
  [sym__abstract_type] = sym__abstract_type,
  [sym__parenthesized_abstract_type] = sym__parenthesized_abstract_type,
  [sym__delimited_type] = sym__delimited_type,
  [sym__simple_type] = sym__simple_type,
  [sym__type] = sym__type,
  [sym_function_type] = sym_function_type,
  [sym_labeled_argument_type] = sym_labeled_argument_type,
  [sym__proper_tuple_type] = sym__proper_tuple_type,
  [sym__labeled_tuple_type] = sym__proper_tuple_type,
  [sym_labeled_tuple_element_type] = sym_labeled_tuple_element_type,
  [sym__tuple_type_rhs] = sym__tuple_type_rhs,
  [sym_constructed_type] = sym_constructed_type,
  [sym_aliased_type] = sym_aliased_type,
  [sym_local_open_type] = sym_local_open_type,
  [sym_polymorphic_variant_type] = sym_polymorphic_variant_type,
  [sym__tag_spec] = sym__tag_spec,
  [sym_tag_specification] = sym_tag_specification,
  [sym_package_type] = sym_package_type,
  [sym_object_type] = sym_object_type,
  [sym_method_type] = sym_method_type,
  [sym_hash_type] = sym_hash_type,
  [sym_parenthesized_type] = sym_parenthesized_type,
  [sym__delimited_expression] = sym__delimited_expression,
  [sym__simple_expression] = sym__simple_expression,
  [sym__expression] = sym__expression,
  [sym__sequence_expression] = sym__sequence_expression,
  [sym_typed_expression] = sym_typed_expression,
  [sym_labeled_tuple_element] = sym_labeled_tuple_element,
  [sym__tuple_expression] = sym__tuple_expression,
  [sym_cons_expression] = sym_cons_expression,
  [sym_list_expression] = sym_list_expression,
  [sym_array_expression] = sym_array_expression,
  [sym__sequence_expression_content] = sym__sequence_expression_content,
  [sym_record_expression] = sym_record_expression,
  [sym_field_expression] = sym_field_expression,
  [sym_application_expression] = sym_application_expression,
  [sym_labeled_argument] = sym_labeled_argument,
  [sym_prefix_expression] = sym_prefix_expression,
  [sym_sign_expression] = sym_sign_expression,
  [sym_hash_expression] = sym_hash_expression,
  [sym_infix_expression] = sym_infix_expression,
  [sym_field_get_expression] = sym_field_get_expression,
  [sym_array_get_expression] = sym_array_get_expression,
  [sym_string_get_expression] = sym_string_get_expression,
  [sym_bigarray_get_expression] = sym_bigarray_get_expression,
  [sym_set_expression] = sym_set_expression,
  [sym_if_expression] = sym_if_expression,
  [sym_then_clause] = sym_then_clause,
  [sym_else_clause] = sym_else_clause,
  [sym_while_expression] = sym_while_expression,
  [sym_do_clause] = sym_do_clause,
  [sym_for_expression] = sym_for_expression,
  [sym__sequence_expression_anonymous] = sym__sequence_expression_anonymous,
  [sym_match_expression] = sym_match_expression,
  [sym__match_cases] = sym__match_cases,
  [sym_match_case] = sym_match_case,
  [sym_guard] = sym_guard,
  [sym_refutation_case] = sym_refutation_case,
  [sym_function_expression] = sym_function_expression,
  [sym_fun_expression] = sym_fun_expression,
  [sym_try_expression] = sym_try_expression,
  [sym_let_expression] = sym_let_expression,
  [sym_assert_expression] = sym_assert_expression,
  [sym_lazy_expression] = sym_lazy_expression,
  [sym_local_open_expression] = sym_local_open_expression,
  [sym_package_expression] = sym_package_expression,
  [sym_new_expression] = sym_new_expression,
  [sym_object_copy_expression] = sym_object_copy_expression,
  [sym_instance_variable_expression] = sym_instance_variable_expression,
  [sym_method_invocation] = sym_method_invocation,
  [sym_object_expression] = sym_object_expression,
  [sym_parenthesized_expression] = sym_parenthesized_expression,
  [sym__delimited_pattern] = sym__delimited_pattern,
  [sym__simple_pattern] = sym__simple_pattern,
  [sym__effect_pattern] = sym__effect_pattern,
  [sym__pattern] = sym__pattern,
  [sym__delimited_binding_pattern] = sym__delimited_binding_pattern,
  [sym__simple_binding_pattern] = sym__simple_binding_pattern,
  [sym__effect_binding_pattern] = sym__effect_binding_pattern,
  [sym__binding_pattern_no_exn] = sym__binding_pattern_no_exn,
  [sym__binding_pattern] = sym__binding_pattern,
  [sym_alias_pattern] = sym_alias_pattern,
  [sym_alias_binding_pattern_no_exn] = sym_alias_pattern,
  [sym_alias_binding_pattern] = sym_alias_pattern,
  [sym_typed_pattern] = sym_typed_pattern,
  [sym_typed_binding_pattern] = sym_typed_pattern,
  [sym__or_pattern_anonymous] = sym__or_pattern_anonymous,
  [sym__or_binding_pattern_no_exn_anonymous] = sym__or_binding_pattern_no_exn_anonymous,
  [sym__or_binding_pattern_anonymous] = sym__or_binding_pattern_anonymous,
  [sym_constructor_pattern] = sym_constructor_pattern,
  [sym_constructor_binding_pattern] = sym_constructor_pattern,
  [sym_tag_pattern] = sym_tag_pattern,
  [sym_tag_binding_pattern] = sym_tag_pattern,
  [sym_polymorphic_variant_pattern] = sym_polymorphic_variant_pattern,
  [sym_labeled_tuple_element_pattern] = sym_labeled_tuple_element_pattern,
  [sym__tuple_pattern] = sym__tuple_pattern,
  [sym_labeled_tuple_element_binding_pattern] = sym_labeled_tuple_element_binding_pattern,
  [sym__tuple_binding_pattern_no_exn] = sym__tuple_binding_pattern_no_exn,
  [sym__tuple_binding_pattern] = sym__tuple_binding_pattern,
  [sym_record_pattern] = sym_record_pattern,
  [sym_field_pattern] = sym_field_pattern,
  [sym_record_binding_pattern] = sym_record_pattern,
  [sym_field_binding_pattern] = sym_field_pattern,
  [sym_list_pattern] = sym_list_pattern,
  [sym_list_binding_pattern] = sym_list_pattern,
  [sym_cons_pattern] = sym_cons_pattern,
  [sym_cons_binding_pattern_no_exn] = sym_cons_pattern,
  [sym_cons_binding_pattern] = sym_cons_pattern,
  [sym_array_pattern] = sym_array_pattern,
  [sym_array_binding_pattern] = sym_array_pattern,
  [sym__sequence_pattern_content] = sym__sequence_pattern_content,
  [sym__sequence_binding_pattern_content] = sym__sequence_binding_pattern_content,
  [sym_range_pattern] = sym_range_pattern,
  [sym_lazy_pattern] = sym_lazy_pattern,
  [sym_lazy_binding_pattern] = sym_lazy_pattern,
  [sym_local_open_pattern] = sym_local_open_pattern,
  [sym_local_open_binding_pattern] = sym_local_open_pattern,
  [sym_package_pattern] = sym_package_pattern,
  [sym_parenthesized_pattern] = sym_parenthesized_pattern,
  [sym_parenthesized_binding_pattern] = sym_parenthesized_pattern,
  [sym_exception_pattern] = sym_exception_pattern,
  [sym_exception_binding_pattern] = sym_exception_pattern,
  [sym_effect_pattern] = sym_effect_pattern,
  [sym_effect_binding_pattern] = sym_effect_pattern,
  [sym_attribute] = sym_attribute,
  [sym_item_attribute] = sym_item_attribute,
  [sym_floating_attribute] = sym_floating_attribute,
  [sym_attribute_payload] = sym_attribute_payload,
  [sym_extension] = sym_extension,
  [sym_quoted_extension] = sym_quoted_extension,
  [sym_item_extension] = sym_item_extension,
  [sym_quoted_item_extension] = sym_quoted_item_extension,
  [sym__attribute] = sym__attribute,
  [sym__constant] = sym__constant,
  [sym__signed_constant] = sym__signed_constant,
  [sym_number] = sym_number,
  [sym_signed_number] = sym_signed_number,
  [sym_character] = sym_character,
  [sym_character_content] = sym_character_content,
  [sym_string] = sym_string,
  [sym_string_content] = sym_string_content,
  [sym_quoted_string] = sym_quoted_string,
  [sym__quoted_string] = sym__quoted_string,
  [sym_quoted_string_content] = sym_quoted_string_content,
  [sym_escape_sequence] = sym_escape_sequence,
  [sym_boolean] = sym_boolean,
  [sym_unit] = sym_unit,
  [sym_sign_operator] = sym_sign_operator,
  [sym__infix_operator] = sym__infix_operator,
  [sym_pow_operator] = sym_pow_operator,
  [sym_mult_operator] = sym_mult_operator,
  [sym_add_operator] = sym_add_operator,
  [sym_and_operator] = sym_and_operator,
  [sym_or_operator] = sym_or_operator,
  [sym_indexing_operator_path] = sym_indexing_operator_path,
  [sym__value_name] = sym__value_name,
  [sym__simple_value_pattern] = sym__simple_value_pattern,
  [sym_parenthesized_operator] = sym_parenthesized_operator,
  [sym_value_path] = sym_value_path,
  [sym_module_path] = sym_module_path,
  [sym_extended_module_path] = sym_extended_module_path,
  [sym_module_type_path] = sym_module_type_path,
  [sym_field_path] = sym_field_path,
  [sym_constructor_path] = sym_constructor_path,
  [sym_type_constructor_path] = sym_type_constructor_path,
  [sym_class_path] = sym_class_path,
  [sym_class_type_path] = sym_class_type_path,
  [sym__instance_variable_name] = sym__instance_variable_name,
  [sym__constructor_name] = sym__constructor_name,
  [sym_directive] = sym_directive,
  [sym_type_variable] = sym_type_variable,
  [sym_tag] = sym_tag,
  [sym_attribute_id] = sym_attribute_id,
  [aux_sym__structure_repeat1] = aux_sym__structure_repeat1,
  [aux_sym__structure_repeat2] = aux_sym__structure_repeat2,
  [aux_sym_expression_item_repeat1] = aux_sym_expression_item_repeat1,
  [aux_sym__signature_repeat1] = aux_sym__signature_repeat1,
  [aux_sym_value_definition_repeat1] = aux_sym_value_definition_repeat1,
  [aux_sym_let_binding_repeat1] = aux_sym_let_binding_repeat1,
  [aux_sym_external_repeat1] = aux_sym_external_repeat1,
  [aux_sym_type_definition_repeat1] = aux_sym_type_definition_repeat1,
  [aux_sym_type_binding_repeat1] = aux_sym_type_binding_repeat1,
  [aux_sym__type_params_repeat1] = aux_sym__type_params_repeat1,
  [aux_sym__type_param_repeat1] = aux_sym__type_param_repeat1,
  [aux_sym_variant_declaration_repeat1] = aux_sym_variant_declaration_repeat1,
  [aux_sym_constructor_declaration_repeat1] = aux_sym_constructor_declaration_repeat1,
  [aux_sym__constructor_argument_repeat1] = aux_sym__constructor_argument_repeat1,
  [aux_sym_record_declaration_repeat1] = aux_sym_record_declaration_repeat1,
  [aux_sym_module_definition_repeat1] = aux_sym_module_definition_repeat1,
  [aux_sym_module_binding_repeat1] = aux_sym_module_binding_repeat1,
  [aux_sym_class_definition_repeat1] = aux_sym_class_definition_repeat1,
  [aux_sym_class_type_definition_repeat1] = aux_sym_class_type_definition_repeat1,
  [aux_sym_module_type_constraint_repeat1] = aux_sym_module_type_constraint_repeat1,
  [aux_sym_instantiated_class_type_repeat1] = aux_sym_instantiated_class_type_repeat1,
  [aux_sym_class_body_type_repeat1] = aux_sym_class_body_type_repeat1,
  [aux_sym_instance_variable_specification_repeat1] = aux_sym_instance_variable_specification_repeat1,
  [aux_sym_method_specification_repeat1] = aux_sym_method_specification_repeat1,
  [aux_sym_class_application_repeat1] = aux_sym_class_application_repeat1,
  [aux_sym__abstract_type_repeat1] = aux_sym__abstract_type_repeat1,
  [aux_sym_polymorphic_variant_type_repeat1] = aux_sym_polymorphic_variant_type_repeat1,
  [aux_sym_polymorphic_variant_type_repeat2] = aux_sym_polymorphic_variant_type_repeat2,
  [aux_sym_tag_specification_repeat1] = aux_sym_tag_specification_repeat1,
  [aux_sym_object_type_repeat1] = aux_sym_object_type_repeat1,
  [aux_sym__sequence_expression_content_repeat1] = aux_sym__sequence_expression_content_repeat1,
  [aux_sym_record_expression_repeat1] = aux_sym_record_expression_repeat1,
  [aux_sym__match_cases_repeat1] = aux_sym__match_cases_repeat1,
  [aux_sym_object_copy_expression_repeat1] = aux_sym_object_copy_expression_repeat1,
  [aux_sym_object_expression_repeat1] = aux_sym_object_expression_repeat1,
  [aux_sym_record_pattern_repeat1] = aux_sym_record_pattern_repeat1,
  [aux_sym_record_binding_pattern_repeat1] = aux_sym_record_binding_pattern_repeat1,
  [aux_sym__sequence_pattern_content_repeat1] = aux_sym__sequence_pattern_content_repeat1,
  [aux_sym__sequence_binding_pattern_content_repeat1] = aux_sym__sequence_binding_pattern_content_repeat1,
  [aux_sym_string_content_repeat1] = aux_sym_string_content_repeat1,
  [aux_sym_quoted_string_content_repeat1] = aux_sym_quoted_string_content_repeat1,
  [aux_sym_attribute_id_repeat1] = aux_sym_attribute_id_repeat1,
  [alias_sym_class_name] = alias_sym_class_name,
  [alias_sym_class_type_name] = alias_sym_class_type_name,
  [alias_sym_field_name] = alias_sym_field_name,
  [alias_sym_instance_variable_name] = alias_sym_instance_variable_name,
  [alias_sym_label_name] = alias_sym_label_name,
  [alias_sym_method_name] = alias_sym_method_name,
  [alias_sym_module_name] = alias_sym_module_name,
  [alias_sym_module_type_name] = alias_sym_module_type_name,
  [alias_sym_sequence_expression] = alias_sym_sequence_expression,
  [alias_sym_tuple_expression] = alias_sym_tuple_expression,
  [alias_sym_type_constructor] = alias_sym_type_constructor,
  [alias_sym_value_name] = alias_sym_value_name,
  [alias_sym_value_pattern] = alias_sym_value_pattern,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym__lowercase_identifier] = {
    .visible = false,
    .named = true,
  },
  [anon_sym_SEMI_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_let] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_rec] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_and] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_external] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_type] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_nonrec] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_private] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_of] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_true] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_false] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_mutable] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_constraint] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_exception] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_module] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_open] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_include] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_class] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_virtual] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_val] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_sig] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_end] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_with] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_functor] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_struct] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_object] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_inherit] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_method] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_in] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_fun] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_as] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_initializer] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUND] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_if] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_then] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_else] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_while] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_do] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_done] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_for] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_to] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_downto] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_match] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_when] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_function] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_try] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_assert] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_lazy] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_new] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_begin] = {
    .visible = true,
    .named = false,
  },
  [sym_ocamlyacc_value] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_effect] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_attribute_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_AT_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_AT_AT_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_quoted_extension_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_LBRACK_PERCENT_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE_PERCENT_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_number_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_signed_number_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SQUOTE2] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_character_content_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_character_content_token2] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_string_content_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_content_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_content_token3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_content_token4] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_string_content_token5] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_quoted_string_content_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_escape_sequence_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_escape_sequence_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_escape_sequence_token3] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_escape_sequence_token4] = {
    .visible = false,
    .named = false,
  },
  [sym_conversion_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_pretty_printing_indication] = {
    .visible = true,
    .named = true,
  },
  [sym_prefix_operator] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_sign_operator_token1] = {
    .visible = false,
    .named = false,
  },
  [sym_hash_operator] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_lsl] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_lsr] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_asr] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_pow_operator_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_mod] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_land] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_lor] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_lxor] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_mult_operator_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_add_operator_token1] = {
    .visible = false,
    .named = false,
  },
  [sym_concat_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_rel_operator] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_AMP_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_or] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE] = {
    .visible = true,
    .named = false,
  },
  [sym_assign_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_indexing_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_let_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_let_and_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_match_operator] = {
    .visible = true,
    .named = true,
  },
  [sym__uppercase_identifier] = {
    .visible = false,
    .named = true,
  },
  [aux_sym_directive_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_variable_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_tag_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attribute_id_token1] = {
    .visible = false,
    .named = false,
  },
  [sym_comment] = {
    .visible = true,
    .named = true,
  },
  [sym__left_quoted_string_delimiter] = {
    .visible = false,
    .named = true,
  },
  [sym__right_quoted_string_delimiter] = {
    .visible = false,
    .named = true,
  },
  [sym_line_number_directive] = {
    .visible = true,
    .named = true,
  },
  [sym__null] = {
    .visible = false,
    .named = true,
  },
  [sym__error_sentinel] = {
    .visible = false,
    .named = true,
  },
  [sym_type] = {
    .visible = true,
    .named = true,
  },
  [sym__structure] = {
    .visible = false,
    .named = true,
  },
  [sym_expression_item] = {
    .visible = true,
    .named = true,
  },
  [sym__signature] = {
    .visible = false,
    .named = true,
  },
  [sym_toplevel_directive] = {
    .visible = true,
    .named = true,
  },
  [sym__local_structure_item] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__structure_item] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_value_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_let_binding] = {
    .visible = true,
    .named = true,
  },
  [sym_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_external] = {
    .visible = true,
    .named = true,
  },
  [sym_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_type_binding] = {
    .visible = true,
    .named = true,
  },
  [sym__type_params] = {
    .visible = false,
    .named = true,
  },
  [sym__type_param] = {
    .visible = false,
    .named = true,
  },
  [sym_variant_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_constructor_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym__constructor_declaration_name] = {
    .visible = true,
    .named = true,
  },
  [sym__constructor_argument] = {
    .visible = false,
    .named = true,
  },
  [sym_record_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_field_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_external_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_type_constraint] = {
    .visible = true,
    .named = true,
  },
  [sym_exception_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_module_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_module_binding] = {
    .visible = true,
    .named = true,
  },
  [sym_module_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_module_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_open_module] = {
    .visible = true,
    .named = true,
  },
  [sym_include_module] = {
    .visible = true,
    .named = true,
  },
  [sym_class_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_class_binding] = {
    .visible = true,
    .named = true,
  },
  [sym_class_type_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_class_type_binding] = {
    .visible = true,
    .named = true,
  },
  [sym__signature_item] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_value_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_include_module_type] = {
    .visible = true,
    .named = true,
  },
  [sym__module_typed] = {
    .visible = false,
    .named = true,
  },
  [sym__module_type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_signature] = {
    .visible = true,
    .named = true,
  },
  [sym_module_type_constraint] = {
    .visible = true,
    .named = true,
  },
  [sym_constrain_type] = {
    .visible = true,
    .named = true,
  },
  [sym_constrain_module] = {
    .visible = true,
    .named = true,
  },
  [sym_constrain_module_type] = {
    .visible = true,
    .named = true,
  },
  [sym_module_type_of] = {
    .visible = true,
    .named = true,
  },
  [sym_functor_type] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_module_type] = {
    .visible = true,
    .named = true,
  },
  [sym__simple_module_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__module_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_structure] = {
    .visible = true,
    .named = true,
  },
  [sym_functor] = {
    .visible = true,
    .named = true,
  },
  [sym_module_application] = {
    .visible = true,
    .named = true,
  },
  [sym_typed_module_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_packed_module] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_module_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__class_typed] = {
    .visible = false,
    .named = true,
  },
  [sym__simple_class_type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__class_type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_instantiated_class_type] = {
    .visible = true,
    .named = true,
  },
  [sym_class_body_type] = {
    .visible = true,
    .named = true,
  },
  [sym__class_field_specification] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_inheritance_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_instance_variable_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_method_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_type_parameter_constraint] = {
    .visible = true,
    .named = true,
  },
  [sym_let_open_class_type] = {
    .visible = true,
    .named = true,
  },
  [sym_class_function_type] = {
    .visible = true,
    .named = true,
  },
  [sym__simple_class_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__class_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_instantiated_class] = {
    .visible = true,
    .named = true,
  },
  [sym_typed_class_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_class_function] = {
    .visible = true,
    .named = true,
  },
  [sym_class_application] = {
    .visible = true,
    .named = true,
  },
  [sym_let_class_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__class_field] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_inheritance_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_instance_variable_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_method_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_class_initializer] = {
    .visible = true,
    .named = true,
  },
  [sym_let_open_class_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_class_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__typed] = {
    .visible = false,
    .named = true,
  },
  [sym__simple_typed] = {
    .visible = false,
    .named = true,
  },
  [sym__coerced] = {
    .visible = false,
    .named = true,
  },
  [sym__type_constrained] = {
    .visible = false,
    .named = true,
  },
  [sym__maybe_polymorphic_typed] = {
    .visible = false,
    .named = true,
  },
  [sym__polymorphic_typed] = {
    .visible = false,
    .named = true,
  },
  [sym__polymorphic_type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_polymorphic_type] = {
    .visible = true,
    .named = true,
  },
  [sym__abstract_type] = {
    .visible = false,
    .named = true,
  },
  [sym__parenthesized_abstract_type] = {
    .visible = true,
    .named = true,
  },
  [sym__delimited_type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__simple_type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_function_type] = {
    .visible = true,
    .named = true,
  },
  [sym_labeled_argument_type] = {
    .visible = true,
    .named = true,
  },
  [sym__proper_tuple_type] = {
    .visible = true,
    .named = true,
  },
  [sym__labeled_tuple_type] = {
    .visible = true,
    .named = true,
  },
  [sym_labeled_tuple_element_type] = {
    .visible = true,
    .named = true,
  },
  [sym__tuple_type_rhs] = {
    .visible = false,
    .named = true,
  },
  [sym_constructed_type] = {
    .visible = true,
    .named = true,
  },
  [sym_aliased_type] = {
    .visible = true,
    .named = true,
  },
  [sym_local_open_type] = {
    .visible = true,
    .named = true,
  },
  [sym_polymorphic_variant_type] = {
    .visible = true,
    .named = true,
  },
  [sym__tag_spec] = {
    .visible = false,
    .named = true,
  },
  [sym_tag_specification] = {
    .visible = true,
    .named = true,
  },
  [sym_package_type] = {
    .visible = true,
    .named = true,
  },
  [sym_object_type] = {
    .visible = true,
    .named = true,
  },
  [sym_method_type] = {
    .visible = true,
    .named = true,
  },
  [sym_hash_type] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_type] = {
    .visible = true,
    .named = true,
  },
  [sym__delimited_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__simple_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__sequence_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_typed_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_labeled_tuple_element] = {
    .visible = true,
    .named = true,
  },
  [sym__tuple_expression] = {
    .visible = false,
    .named = true,
  },
  [sym_cons_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_list_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_array_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__sequence_expression_content] = {
    .visible = false,
    .named = true,
  },
  [sym_record_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_field_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_application_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_labeled_argument] = {
    .visible = true,
    .named = true,
  },
  [sym_prefix_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_sign_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_hash_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_infix_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_field_get_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_array_get_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_string_get_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_bigarray_get_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_set_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_if_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_then_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_else_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_while_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_do_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_for_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__sequence_expression_anonymous] = {
    .visible = false,
    .named = true,
  },
  [sym_match_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__match_cases] = {
    .visible = false,
    .named = true,
  },
  [sym_match_case] = {
    .visible = true,
    .named = true,
  },
  [sym_guard] = {
    .visible = true,
    .named = true,
  },
  [sym_refutation_case] = {
    .visible = true,
    .named = true,
  },
  [sym_function_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_fun_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_try_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_let_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_assert_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_lazy_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_local_open_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_package_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_new_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_object_copy_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_instance_variable_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_method_invocation] = {
    .visible = true,
    .named = true,
  },
  [sym_object_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__delimited_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__simple_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__effect_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__delimited_binding_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__simple_binding_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__effect_binding_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__binding_pattern_no_exn] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__binding_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_alias_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_alias_binding_pattern_no_exn] = {
    .visible = true,
    .named = true,
  },
  [sym_alias_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_typed_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_typed_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym__or_pattern_anonymous] = {
    .visible = false,
    .named = true,
  },
  [sym__or_binding_pattern_no_exn_anonymous] = {
    .visible = true,
    .named = true,
  },
  [sym__or_binding_pattern_anonymous] = {
    .visible = false,
    .named = true,
  },
  [sym_constructor_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_constructor_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_tag_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_tag_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_polymorphic_variant_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_labeled_tuple_element_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym__tuple_pattern] = {
    .visible = false,
    .named = true,
  },
  [sym_labeled_tuple_element_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym__tuple_binding_pattern_no_exn] = {
    .visible = true,
    .named = true,
  },
  [sym__tuple_binding_pattern] = {
    .visible = false,
    .named = true,
  },
  [sym_record_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_field_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_record_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_field_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_list_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_list_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_cons_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_cons_binding_pattern_no_exn] = {
    .visible = true,
    .named = true,
  },
  [sym_cons_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_array_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_array_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym__sequence_pattern_content] = {
    .visible = false,
    .named = true,
  },
  [sym__sequence_binding_pattern_content] = {
    .visible = false,
    .named = true,
  },
  [sym_range_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_lazy_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_lazy_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_local_open_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_local_open_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_package_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_exception_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_exception_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_effect_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_effect_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym_item_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym_floating_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_payload] = {
    .visible = true,
    .named = true,
  },
  [sym_extension] = {
    .visible = true,
    .named = true,
  },
  [sym_quoted_extension] = {
    .visible = true,
    .named = true,
  },
  [sym_item_extension] = {
    .visible = true,
    .named = true,
  },
  [sym_quoted_item_extension] = {
    .visible = true,
    .named = true,
  },
  [sym__attribute] = {
    .visible = false,
    .named = true,
  },
  [sym__constant] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__signed_constant] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_number] = {
    .visible = true,
    .named = true,
  },
  [sym_signed_number] = {
    .visible = true,
    .named = true,
  },
  [sym_character] = {
    .visible = true,
    .named = true,
  },
  [sym_character_content] = {
    .visible = true,
    .named = true,
  },
  [sym_string] = {
    .visible = true,
    .named = true,
  },
  [sym_string_content] = {
    .visible = true,
    .named = true,
  },
  [sym_quoted_string] = {
    .visible = true,
    .named = true,
  },
  [sym__quoted_string] = {
    .visible = false,
    .named = true,
  },
  [sym_quoted_string_content] = {
    .visible = true,
    .named = true,
  },
  [sym_escape_sequence] = {
    .visible = true,
    .named = true,
  },
  [sym_boolean] = {
    .visible = true,
    .named = true,
  },
  [sym_unit] = {
    .visible = true,
    .named = true,
  },
  [sym_sign_operator] = {
    .visible = true,
    .named = true,
  },
  [sym__infix_operator] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_pow_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_mult_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_add_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_and_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_or_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_indexing_operator_path] = {
    .visible = true,
    .named = true,
  },
  [sym__value_name] = {
    .visible = false,
    .named = true,
  },
  [sym__simple_value_pattern] = {
    .visible = false,
    .named = true,
  },
  [sym_parenthesized_operator] = {
    .visible = true,
    .named = true,
  },
  [sym_value_path] = {
    .visible = true,
    .named = true,
  },
  [sym_module_path] = {
    .visible = true,
    .named = true,
  },
  [sym_extended_module_path] = {
    .visible = true,
    .named = true,
  },
  [sym_module_type_path] = {
    .visible = true,
    .named = true,
  },
  [sym_field_path] = {
    .visible = true,
    .named = true,
  },
  [sym_constructor_path] = {
    .visible = true,
    .named = true,
  },
  [sym_type_constructor_path] = {
    .visible = true,
    .named = true,
  },
  [sym_class_path] = {
    .visible = true,
    .named = true,
  },
  [sym_class_type_path] = {
    .visible = true,
    .named = true,
  },
  [sym__instance_variable_name] = {
    .visible = false,
    .named = true,
  },
  [sym__constructor_name] = {
    .visible = false,
    .named = true,
  },
  [sym_directive] = {
    .visible = true,
    .named = true,
  },
  [sym_type_variable] = {
    .visible = true,
    .named = true,
  },
  [sym_tag] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_id] = {
    .visible = true,
    .named = true,
  },
  [aux_sym__structure_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__structure_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_expression_item_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__signature_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_value_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_let_binding_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_external_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_binding_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__type_params_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__type_param_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_variant_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_constructor_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__constructor_argument_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_record_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_module_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_module_binding_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_type_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_module_type_constraint_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_instantiated_class_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_body_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_instance_variable_specification_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_method_specification_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_application_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__abstract_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_polymorphic_variant_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_polymorphic_variant_type_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_tag_specification_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_object_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__sequence_expression_content_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_record_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__match_cases_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_object_copy_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_object_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_record_pattern_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_record_binding_pattern_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__sequence_pattern_content_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__sequence_binding_pattern_content_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_content_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_quoted_string_content_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attribute_id_repeat1] = {
    .visible = false,
    .named = false,
  },
  [alias_sym_class_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_class_type_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_field_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_instance_variable_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_label_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_method_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_module_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_module_type_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_sequence_expression] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_tuple_expression] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_type_constructor] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_value_name] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_value_pattern] = {
    .visible = true,
    .named = true,
  },
};

enum ts_field_identifiers {
  field_alias = 1,
  field_argument = 2,
  field_array = 3,
  field_body = 4,
  field_class = 5,
  field_class_type = 6,
  field_codomain = 7,
  field_coercion = 8,
  field_condition = 9,
  field_constraint = 10,
  field_continuation = 11,
  field_default = 12,
  field_domain = 13,
  field_effect = 14,
  field_equation = 15,
  field_expression = 16,
  field_field = 17,
  field_from = 18,
  field_function = 19,
  field_functor = 20,
  field_index = 21,
  field_initializer = 22,
  field_left = 23,
  field_method = 24,
  field_module = 25,
  field_module_type = 26,
  field_name = 27,
  field_object = 28,
  field_operator = 29,
  field_pattern = 30,
  field_record = 31,
  field_right = 32,
  field_self = 33,
  field_self_type = 34,
  field_string = 35,
  field_synonym = 36,
  field_to = 37,
  field_type = 38,
  field_value = 39,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_alias] = "alias",
  [field_argument] = "argument",
  [field_array] = "array",
  [field_body] = "body",
  [field_class] = "class",
  [field_class_type] = "class_type",
  [field_codomain] = "codomain",
  [field_coercion] = "coercion",
  [field_condition] = "condition",
  [field_constraint] = "constraint",
  [field_continuation] = "continuation",
  [field_default] = "default",
  [field_domain] = "domain",
  [field_effect] = "effect",
  [field_equation] = "equation",
  [field_expression] = "expression",
  [field_field] = "field",
  [field_from] = "from",
  [field_function] = "function",
  [field_functor] = "functor",
  [field_index] = "index",
  [field_initializer] = "initializer",
  [field_left] = "left",
  [field_method] = "method",
  [field_module] = "module",
  [field_module_type] = "module_type",
  [field_name] = "name",
  [field_object] = "object",
  [field_operator] = "operator",
  [field_pattern] = "pattern",
  [field_record] = "record",
  [field_right] = "right",
  [field_self] = "self",
  [field_self_type] = "self_type",
  [field_string] = "string",
  [field_synonym] = "synonym",
  [field_to] = "to",
  [field_type] = "type",
  [field_value] = "value",
};

static const TSMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [10] = {.index = 0, .length = 1},
  [11] = {.index = 1, .length = 1},
  [12] = {.index = 2, .length = 2},
  [13] = {.index = 4, .length = 2},
  [16] = {.index = 1, .length = 1},
  [17] = {.index = 6, .length = 1},
  [23] = {.index = 7, .length = 1},
  [25] = {.index = 8, .length = 1},
  [27] = {.index = 9, .length = 1},
  [28] = {.index = 10, .length = 2},
  [29] = {.index = 12, .length = 1},
  [30] = {.index = 13, .length = 2},
  [31] = {.index = 15, .length = 1},
  [32] = {.index = 16, .length = 1},
  [33] = {.index = 17, .length = 1},
  [34] = {.index = 1, .length = 1},
  [35] = {.index = 18, .length = 1},
  [37] = {.index = 19, .length = 1},
  [39] = {.index = 20, .length = 1},
  [40] = {.index = 21, .length = 1},
  [41] = {.index = 22, .length = 1},
  [43] = {.index = 23, .length = 1},
  [44] = {.index = 24, .length = 2},
  [46] = {.index = 26, .length = 1},
  [47] = {.index = 27, .length = 1},
  [48] = {.index = 28, .length = 2},
  [50] = {.index = 30, .length = 1},
  [51] = {.index = 31, .length = 1},
  [52] = {.index = 32, .length = 1},
  [54] = {.index = 33, .length = 1},
  [55] = {.index = 34, .length = 1},
  [56] = {.index = 35, .length = 1},
  [57] = {.index = 36, .length = 1},
  [58] = {.index = 37, .length = 2},
  [59] = {.index = 39, .length = 2},
  [60] = {.index = 41, .length = 3},
  [61] = {.index = 44, .length = 2},
  [62] = {.index = 46, .length = 2},
  [63] = {.index = 48, .length = 1},
  [64] = {.index = 49, .length = 1},
  [65] = {.index = 50, .length = 1},
  [66] = {.index = 51, .length = 4},
  [67] = {.index = 55, .length = 3},
  [68] = {.index = 0, .length = 1},
  [69] = {.index = 58, .length = 1},
  [70] = {.index = 59, .length = 1},
  [71] = {.index = 60, .length = 2},
  [72] = {.index = 62, .length = 1},
  [73] = {.index = 63, .length = 2},
  [74] = {.index = 65, .length = 1},
  [75] = {.index = 66, .length = 2},
  [76] = {.index = 68, .length = 3},
  [77] = {.index = 71, .length = 2},
  [78] = {.index = 71, .length = 2},
  [81] = {.index = 36, .length = 1},
  [82] = {.index = 73, .length = 1},
  [83] = {.index = 74, .length = 1},
  [84] = {.index = 75, .length = 1},
  [85] = {.index = 76, .length = 1},
  [86] = {.index = 36, .length = 1},
  [87] = {.index = 77, .length = 1},
  [88] = {.index = 76, .length = 1},
  [89] = {.index = 78, .length = 1},
  [90] = {.index = 79, .length = 2},
  [91] = {.index = 81, .length = 1},
  [92] = {.index = 82, .length = 2},
  [94] = {.index = 31, .length = 1},
  [95] = {.index = 84, .length = 1},
  [97] = {.index = 85, .length = 3},
  [98] = {.index = 73, .length = 1},
  [99] = {.index = 88, .length = 1},
  [100] = {.index = 89, .length = 2},
  [101] = {.index = 91, .length = 1},
  [102] = {.index = 92, .length = 2},
  [103] = {.index = 94, .length = 3},
  [104] = {.index = 97, .length = 3},
  [105] = {.index = 100, .length = 2},
  [106] = {.index = 102, .length = 1},
  [107] = {.index = 103, .length = 2},
  [108] = {.index = 105, .length = 2},
  [109] = {.index = 107, .length = 2},
  [110] = {.index = 109, .length = 2},
  [111] = {.index = 111, .length = 2},
  [112] = {.index = 111, .length = 2},
  [113] = {.index = 109, .length = 2},
  [114] = {.index = 113, .length = 1},
  [115] = {.index = 114, .length = 3},
  [116] = {.index = 102, .length = 1},
  [117] = {.index = 117, .length = 2},
  [118] = {.index = 59, .length = 1},
  [119] = {.index = 119, .length = 1},
  [120] = {.index = 120, .length = 2},
  [121] = {.index = 36, .length = 1},
  [123] = {.index = 59, .length = 1},
  [124] = {.index = 122, .length = 1},
  [125] = {.index = 123, .length = 2},
  [126] = {.index = 59, .length = 1},
  [127] = {.index = 125, .length = 1},
  [128] = {.index = 126, .length = 2},
  [129] = {.index = 103, .length = 2},
  [130] = {.index = 128, .length = 2},
  [131] = {.index = 78, .length = 1},
  [133] = {.index = 59, .length = 1},
  [134] = {.index = 78, .length = 1},
  [135] = {.index = 130, .length = 2},
  [136] = {.index = 132, .length = 2},
  [137] = {.index = 134, .length = 2},
  [138] = {.index = 136, .length = 2},
  [139] = {.index = 78, .length = 1},
  [140] = {.index = 138, .length = 1},
  [142] = {.index = 139, .length = 2},
  [143] = {.index = 141, .length = 4},
  [144] = {.index = 145, .length = 3},
  [145] = {.index = 148, .length = 3},
  [146] = {.index = 151, .length = 2},
  [147] = {.index = 0, .length = 1},
  [148] = {.index = 153, .length = 3},
  [149] = {.index = 156, .length = 2},
  [150] = {.index = 156, .length = 2},
  [151] = {.index = 158, .length = 1},
  [152] = {.index = 159, .length = 2},
  [153] = {.index = 161, .length = 2},
  [154] = {.index = 59, .length = 1},
  [155] = {.index = 163, .length = 1},
  [157] = {.index = 164, .length = 2},
  [158] = {.index = 102, .length = 1},
  [159] = {.index = 164, .length = 2},
  [160] = {.index = 166, .length = 2},
  [161] = {.index = 168, .length = 2},
  [162] = {.index = 170, .length = 3},
  [163] = {.index = 173, .length = 2},
  [164] = {.index = 175, .length = 2},
  [165] = {.index = 102, .length = 1},
  [166] = {.index = 177, .length = 1},
  [167] = {.index = 177, .length = 1},
  [168] = {.index = 130, .length = 2},
  [169] = {.index = 102, .length = 1},
  [170] = {.index = 178, .length = 1},
  [171] = {.index = 179, .length = 2},
  [172] = {.index = 181, .length = 2},
  [173] = {.index = 183, .length = 3},
  [174] = {.index = 186, .length = 3},
  [175] = {.index = 189, .length = 1},
  [176] = {.index = 190, .length = 1},
  [177] = {.index = 191, .length = 1},
  [178] = {.index = 192, .length = 4},
  [179] = {.index = 31, .length = 1},
  [180] = {.index = 196, .length = 3},
  [181] = {.index = 199, .length = 3},
  [182] = {.index = 202, .length = 2},
  [183] = {.index = 158, .length = 1},
  [184] = {.index = 204, .length = 1},
  [185] = {.index = 204, .length = 1},
  [187] = {.index = 205, .length = 2},
  [188] = {.index = 207, .length = 3},
  [189] = {.index = 158, .length = 1},
  [190] = {.index = 181, .length = 2},
  [191] = {.index = 158, .length = 1},
  [192] = {.index = 158, .length = 1},
  [193] = {.index = 210, .length = 1},
  [194] = {.index = 181, .length = 2},
  [195] = {.index = 211, .length = 2},
  [196] = {.index = 213, .length = 2},
  [197] = {.index = 215, .length = 3},
  [198] = {.index = 218, .length = 1},
  [199] = {.index = 219, .length = 1},
  [200] = {.index = 220, .length = 3},
  [201] = {.index = 223, .length = 3},
  [202] = {.index = 158, .length = 1},
  [203] = {.index = 226, .length = 2},
  [204] = {.index = 228, .length = 1},
  [205] = {.index = 228, .length = 1},
  [206] = {.index = 229, .length = 1},
  [207] = {.index = 229, .length = 1},
  [208] = {.index = 230, .length = 1},
  [209] = {.index = 231, .length = 3},
  [210] = {.index = 234, .length = 2},
  [211] = {.index = 234, .length = 2},
  [212] = {.index = 228, .length = 1},
  [213] = {.index = 236, .length = 3},
  [214] = {.index = 239, .length = 2},
  [215] = {.index = 241, .length = 3},
  [216] = {.index = 244, .length = 1},
  [217] = {.index = 228, .length = 1},
  [218] = {.index = 245, .length = 2},
  [219] = {.index = 245, .length = 2},
  [220] = {.index = 247, .length = 1},
  [221] = {.index = 247, .length = 1},
  [222] = {.index = 248, .length = 1},
  [223] = {.index = 249, .length = 2},
  [224] = {.index = 251, .length = 2},
  [225] = {.index = 247, .length = 1},
  [226] = {.index = 253, .length = 2},
  [227] = {.index = 253, .length = 2},
  [228] = {.index = 255, .length = 1},
  [229] = {.index = 256, .length = 3},
  [230] = {.index = 259, .length = 2},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_type, 1, .inherited = true},
  [1] =
    {field_type, 2},
  [2] =
    {field_codomain, 2},
    {field_domain, 0},
  [4] =
    {field_alias, 2},
    {field_type, 0},
  [6] =
    {field_pattern, 0},
  [7] =
    {field_name, 0},
  [8] =
    {field_module, 1},
  [9] =
    {field_expression, 1},
  [10] =
    {field_expression, 1},
    {field_operator, 0},
  [12] =
    {field_argument, 0},
  [13] =
    {field_argument, 1, .inherited = true},
    {field_function, 0},
  [15] =
    {field_type, 3},
  [16] =
    {field_codomain, 1},
  [17] =
    {field_module_type, 2},
  [18] =
    {field_type, 1},
  [19] =
    {field_pattern, 1},
  [20] =
    {field_type, 0, .inherited = true},
  [21] =
    {field_coercion, 0, .inherited = true},
  [22] =
    {field_name, 1},
  [23] =
    {field_module_type, 1},
  [24] =
    {field_coercion, 1, .inherited = true},
    {field_type, 1, .inherited = true},
  [26] =
    {field_module_type, 1, .inherited = true},
  [27] =
    {field_module, 2},
  [28] =
    {field_argument, 1},
    {field_functor, 0},
  [30] =
    {field_class_type, 1, .inherited = true},
  [31] =
    {field_type, 2, .inherited = true},
  [32] =
    {field_class, 1},
  [33] =
    {field_initializer, 1},
  [34] =
    {field_condition, 1},
  [35] =
    {field_expression, 2},
  [36] =
    {field_body, 2},
  [37] =
    {field_field, 2},
    {field_record, 0},
  [39] =
    {field_method, 2},
    {field_object, 0},
  [41] =
    {field_left, 0},
    {field_operator, 1},
    {field_right, 2},
  [44] =
    {field_argument, 0, .inherited = true},
    {field_argument, 1, .inherited = true},
  [46] =
    {field_left, 0},
    {field_right, 2},
  [48] =
    {field_codomain, 2},
  [49] =
    {field_module_type, 0},
  [50] =
    {field_module_type, 3},
  [51] =
    {field_codomain, 4},
    {field_domain, 0},
    {field_domain, 1},
    {field_domain, 2},
  [55] =
    {field_type, 2},
    {field_type, 3},
    {field_type, 4},
  [58] =
    {field_pattern, 2},
  [59] =
    {field_body, 3},
  [60] =
    {field_body, 2},
    {field_pattern, 0},
  [62] =
    {field_coercion, 1},
  [63] =
    {field_alias, 2},
    {field_pattern, 0},
  [65] =
    {field_expression, 3},
  [66] =
    {field_coercion, 1, .inherited = true},
    {field_type, 0, .inherited = true},
  [68] =
    {field_coercion, 2, .inherited = true},
    {field_expression, 1},
    {field_type, 2, .inherited = true},
  [71] =
    {field_body, 2},
    {field_name, 0},
  [73] =
    {field_module_type, 2, .inherited = true},
  [74] =
    {field_module, 3},
  [75] =
    {field_functor, 0},
  [76] =
    {field_class_type, 2, .inherited = true},
  [77] =
    {field_class_type, 1},
  [78] =
    {field_type, 3, .inherited = true},
  [79] =
    {field_coercion, 2, .inherited = true},
    {field_type, 2, .inherited = true},
  [81] =
    {field_class, 2},
  [82] =
    {field_argument, 1, .inherited = true},
    {field_class, 0},
  [84] =
    {field_condition, 2},
  [85] =
    {field_type, 3},
    {field_type, 4},
    {field_type, 5},
  [88] =
    {field_codomain, 3},
  [89] =
    {field_module, 2},
    {field_module_type, 4},
  [91] =
    {field_pattern, 3},
  [92] =
    {field_pattern, 1},
    {field_type, 2, .inherited = true},
  [94] =
    {field_body, 3},
    {field_coercion, 1, .inherited = true},
    {field_pattern, 0},
  [97] =
    {field_body, 3},
    {field_pattern, 0},
    {field_type, 1, .inherited = true},
  [100] =
    {field_body, 3},
    {field_pattern, 0},
  [102] =
    {field_body, 4},
  [103] =
    {field_coercion, 3, .inherited = true},
    {field_type, 3, .inherited = true},
  [105] =
    {field_continuation, 3},
    {field_effect, 1},
  [107] =
    {field_module, 2},
    {field_module_type, 3, .inherited = true},
  [109] =
    {field_body, 3},
    {field_name, 0},
  [111] =
    {field_body, 3},
    {field_name, 1},
  [113] =
    {field_record, 1},
  [114] =
    {field_body, 3},
    {field_coercion, 1, .inherited = true},
    {field_type, 1, .inherited = true},
  [117] =
    {field_body, 3},
    {field_module_type, 1, .inherited = true},
  [119] =
    {field_value, 2},
  [120] =
    {field_module, 1},
    {field_module_type, 2, .inherited = true},
  [122] =
    {field_class_type, 3, .inherited = true},
  [123] =
    {field_body, 3},
    {field_class_type, 1, .inherited = true},
  [125] =
    {field_self, 2},
  [126] =
    {field_constraint, 3},
    {field_type, 1},
  [128] =
    {field_alias, 3},
    {field_class, 1},
  [130] =
    {field_body, 4},
    {field_type, 2, .inherited = true},
  [132] =
    {field_array, 0},
    {field_index, 3},
  [134] =
    {field_index, 3},
    {field_string, 0},
  [136] =
    {field_module, 3},
    {field_module_type, 5},
  [138] =
    {field_module_type, 3, .inherited = true},
  [139] =
    {field_pattern, 3},
    {field_type, 1, .inherited = true},
  [141] =
    {field_body, 4},
    {field_coercion, 2, .inherited = true},
    {field_pattern, 0},
    {field_type, 1, .inherited = true},
  [145] =
    {field_body, 4},
    {field_coercion, 2, .inherited = true},
    {field_pattern, 0},
  [148] =
    {field_body, 4},
    {field_pattern, 0},
    {field_type, 2, .inherited = true},
  [151] =
    {field_module, 3},
    {field_module_type, 4, .inherited = true},
  [153] =
    {field_body, 4},
    {field_name, 0},
    {field_synonym, 2},
  [156] =
    {field_body, 4},
    {field_name, 1},
  [158] =
    {field_body, 5},
  [159] =
    {field_body, 4},
    {field_module_type, 2, .inherited = true},
  [161] =
    {field_module_type, 3, .inherited = true},
    {field_value, 2},
  [163] =
    {field_class_type, 4, .inherited = true},
  [164] =
    {field_body, 4},
    {field_class_type, 2, .inherited = true},
  [166] =
    {field_self, 2},
    {field_type, 3, .inherited = true},
  [168] =
    {field_coercion, 4, .inherited = true},
    {field_type, 4, .inherited = true},
  [170] =
    {field_body, 4},
    {field_coercion, 2, .inherited = true},
    {field_type, 2, .inherited = true},
  [173] =
    {field_class, 1},
    {field_class_type, 2, .inherited = true},
  [175] =
    {field_alias, 4},
    {field_class, 2},
  [177] =
    {field_type, 4, .inherited = true},
  [178] =
    {field_self, 3},
  [179] =
    {field_pattern, 2},
    {field_type, 3, .inherited = true},
  [181] =
    {field_body, 5},
    {field_type, 3, .inherited = true},
  [183] =
    {field_array, 0},
    {field_index, 4},
    {field_operator, 2},
  [186] =
    {field_index, 4},
    {field_operator, 2},
    {field_string, 0},
  [189] =
    {field_equation, 3},
  [190] =
    {field_constraint, 3},
  [191] =
    {field_module_type, 4, .inherited = true},
  [192] =
    {field_body, 5},
    {field_coercion, 3, .inherited = true},
    {field_pattern, 0},
    {field_type, 2, .inherited = true},
  [196] =
    {field_body, 5},
    {field_name, 0},
    {field_synonym, 2},
  [199] =
    {field_body, 5},
    {field_name, 1},
    {field_synonym, 3},
  [202] =
    {field_coercion, 4},
    {field_value, 2},
  [204] =
    {field_class_type, 5, .inherited = true},
  [205] =
    {field_body, 5},
    {field_class_type, 3, .inherited = true},
  [207] =
    {field_body, 5},
    {field_coercion, 3, .inherited = true},
    {field_type, 3, .inherited = true},
  [210] =
    {field_type, 5, .inherited = true},
  [211] =
    {field_self, 3},
    {field_type, 4, .inherited = true},
  [213] =
    {field_default, 4},
    {field_pattern, 2},
  [215] =
    {field_from, 3},
    {field_name, 1},
    {field_to, 5},
  [218] =
    {field_equation, 4},
  [219] =
    {field_constraint, 4},
  [220] =
    {field_body, 6},
    {field_name, 1},
    {field_synonym, 3},
  [223] =
    {field_coercion, 5},
    {field_module_type, 3, .inherited = true},
    {field_value, 2},
  [226] =
    {field_body, 6},
    {field_class_type, 4, .inherited = true},
  [228] =
    {field_body, 6},
  [229] =
    {field_class_type, 6, .inherited = true},
  [230] =
    {field_self_type, 2},
  [231] =
    {field_body, 6},
    {field_coercion, 4, .inherited = true},
    {field_type, 4, .inherited = true},
  [234] =
    {field_body, 6},
    {field_type, 4, .inherited = true},
  [236] =
    {field_default, 5},
    {field_pattern, 2},
    {field_type, 3, .inherited = true},
  [239] =
    {field_pattern, 4},
    {field_type, 5, .inherited = true},
  [241] =
    {field_from, 4},
    {field_name, 2},
    {field_to, 6},
  [244] =
    {field_equation, 5},
  [245] =
    {field_body, 7},
    {field_class_type, 5, .inherited = true},
  [247] =
    {field_body, 7},
  [248] =
    {field_class_type, 7, .inherited = true},
  [249] =
    {field_body, 7},
    {field_type, 5, .inherited = true},
  [251] =
    {field_default, 6},
    {field_pattern, 4},
  [253] =
    {field_body, 8},
    {field_class_type, 6, .inherited = true},
  [255] =
    {field_body, 8},
  [256] =
    {field_default, 7},
    {field_pattern, 4},
    {field_type, 5, .inherited = true},
  [259] =
    {field_body, 9},
    {field_class_type, 7, .inherited = true},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [1] = {
    [0] = alias_sym_type_constructor,
  },
  [2] = {
    [0] = alias_sym_module_name,
  },
  [3] = {
    [0] = alias_sym_class_type_name,
  },
  [4] = {
    [0] = alias_sym_value_name,
  },
  [5] = {
    [0] = alias_sym_instance_variable_name,
  },
  [6] = {
    [0] = sym__constructor_declaration_name,
  },
  [7] = {
    [0] = alias_sym_tuple_expression,
  },
  [8] = {
    [0] = alias_sym_sequence_expression,
  },
  [9] = {
    [0] = alias_sym_module_type_name,
  },
  [10] = {
    [0] = alias_sym_method_name,
  },
  [11] = {
    [0] = alias_sym_label_name,
  },
  [14] = {
    [2] = alias_sym_type_constructor,
  },
  [15] = {
    [2] = alias_sym_module_name,
  },
  [18] = {
    [1] = alias_sym_label_name,
  },
  [19] = {
    [0] = alias_sym_value_pattern,
  },
  [20] = {
    [0] = sym__or_binding_pattern_no_exn_anonymous,
  },
  [21] = {
    [0] = sym__tuple_binding_pattern_no_exn,
  },
  [22] = {
    [0] = sym_type_variable,
  },
  [23] = {
    [0] = alias_sym_type_constructor,
  },
  [24] = {
    [0] = alias_sym_field_name,
  },
  [26] = {
    [0] = alias_sym_class_name,
  },
  [31] = {
    [1] = alias_sym_label_name,
  },
  [34] = {
    [0] = sym__parenthesized_abstract_type,
  },
  [36] = {
    [2] = alias_sym_class_type_name,
  },
  [38] = {
    [1] = sym__constructor_declaration_name,
  },
  [41] = {
    [1] = alias_sym_type_constructor,
  },
  [42] = {
    [1] = sym_type_variable,
  },
  [45] = {
    [2] = alias_sym_module_type_name,
  },
  [46] = {
    [0] = alias_sym_module_name,
  },
  [49] = {
    [1] = alias_sym_class_name,
  },
  [50] = {
    [0] = alias_sym_class_name,
  },
  [53] = {
    [1] = alias_sym_method_name,
  },
  [59] = {
    [2] = alias_sym_method_name,
  },
  [67] = {
    [0] = alias_sym_label_name,
  },
  [74] = {
    [1] = alias_sym_label_name,
  },
  [77] = {
    [0] = alias_sym_type_constructor,
  },
  [79] = {
    [2] = alias_sym_field_name,
  },
  [80] = {
    [3] = alias_sym_module_type_name,
  },
  [81] = {
    [0] = alias_sym_module_name,
  },
  [82] = {
    [0] = alias_sym_module_name,
  },
  [85] = {
    [1] = alias_sym_class_name,
  },
  [86] = {
    [0] = alias_sym_class_name,
  },
  [88] = {
    [0] = alias_sym_class_name,
  },
  [93] = {
    [2] = alias_sym_method_name,
  },
  [94] = {
    [1] = alias_sym_method_name,
  },
  [96] = {
    [2] = alias_sym_class_name,
  },
  [97] = {
    [1] = alias_sym_label_name,
  },
  [98] = {
    [1] = alias_sym_module_name,
  },
  [100] = {
    [2] = alias_sym_module_name,
  },
  [101] = {
    [1] = alias_sym_label_name,
  },
  [107] = {
    [2] = alias_sym_label_name,
  },
  [110] = {
    [0] = alias_sym_type_constructor,
  },
  [111] = {
    [1] = alias_sym_type_constructor,
  },
  [116] = {
    [2] = alias_sym_module_type_name,
  },
  [117] = {
    [0] = alias_sym_module_name,
  },
  [118] = {
    [0] = alias_sym_module_name,
  },
  [121] = {
    [0] = alias_sym_class_type_name,
  },
  [122] = {
    [3] = alias_sym_class_name,
  },
  [123] = {
    [1] = alias_sym_class_name,
  },
  [124] = {
    [1] = alias_sym_class_name,
  },
  [125] = {
    [0] = alias_sym_class_name,
  },
  [126] = {
    [0] = alias_sym_class_name,
  },
  [131] = {
    [2] = alias_sym_method_name,
  },
  [132] = {
    [3] = alias_sym_method_name,
  },
  [133] = {
    [1] = alias_sym_method_name,
  },
  [134] = {
    [1] = alias_sym_method_name,
  },
  [138] = {
    [3] = alias_sym_module_name,
  },
  [139] = {
    [2] = alias_sym_label_name,
  },
  [140] = {
    [2] = alias_sym_module_name,
  },
  [141] = {
    [3] = alias_sym_module_name,
  },
  [147] = {
    [0] = alias_sym_field_name,
  },
  [148] = {
    [0] = alias_sym_type_constructor,
  },
  [149] = {
    [1] = alias_sym_type_constructor,
  },
  [151] = {
    [3] = alias_sym_module_type_name,
  },
  [152] = {
    [0] = alias_sym_module_name,
  },
  [154] = {
    [1] = alias_sym_class_type_name,
  },
  [155] = {
    [3] = alias_sym_class_name,
  },
  [156] = {
    [4] = alias_sym_class_name,
  },
  [157] = {
    [1] = alias_sym_class_name,
  },
  [158] = {
    [1] = alias_sym_class_name,
  },
  [159] = {
    [0] = alias_sym_class_name,
  },
  [165] = {
    [2] = alias_sym_method_name,
  },
  [166] = {
    [2] = alias_sym_method_name,
  },
  [167] = {
    [3] = alias_sym_method_name,
  },
  [168] = {
    [1] = alias_sym_method_name,
  },
  [169] = {
    [1] = alias_sym_method_name,
  },
  [177] = {
    [3] = alias_sym_module_name,
  },
  [179] = {
    [1] = alias_sym_field_name,
  },
  [180] = {
    [0] = alias_sym_type_constructor,
  },
  [181] = {
    [1] = alias_sym_type_constructor,
  },
  [183] = {
    [3] = alias_sym_class_name,
  },
  [184] = {
    [3] = alias_sym_class_name,
  },
  [185] = {
    [4] = alias_sym_class_name,
  },
  [186] = {
    [5] = alias_sym_class_name,
  },
  [187] = {
    [1] = alias_sym_class_name,
  },
  [190] = {
    [2] = alias_sym_method_name,
  },
  [191] = {
    [2] = alias_sym_method_name,
  },
  [192] = {
    [3] = alias_sym_method_name,
  },
  [193] = {
    [3] = alias_sym_method_name,
  },
  [194] = {
    [1] = alias_sym_method_name,
  },
  [200] = {
    [1] = alias_sym_type_constructor,
  },
  [202] = {
    [3] = alias_sym_class_type_name,
  },
  [203] = {
    [3] = alias_sym_class_name,
  },
  [204] = {
    [3] = alias_sym_class_name,
  },
  [205] = {
    [4] = alias_sym_class_name,
  },
  [206] = {
    [4] = alias_sym_class_name,
  },
  [207] = {
    [5] = alias_sym_class_name,
  },
  [210] = {
    [2] = alias_sym_method_name,
  },
  [211] = {
    [3] = alias_sym_method_name,
  },
  [212] = {
    [3] = alias_sym_method_name,
  },
  [214] = {
    [1] = alias_sym_label_name,
  },
  [217] = {
    [4] = alias_sym_class_type_name,
  },
  [218] = {
    [3] = alias_sym_class_name,
  },
  [219] = {
    [4] = alias_sym_class_name,
  },
  [220] = {
    [4] = alias_sym_class_name,
  },
  [221] = {
    [5] = alias_sym_class_name,
  },
  [222] = {
    [5] = alias_sym_class_name,
  },
  [223] = {
    [3] = alias_sym_method_name,
  },
  [224] = {
    [1] = alias_sym_label_name,
  },
  [225] = {
    [5] = alias_sym_class_type_name,
  },
  [226] = {
    [4] = alias_sym_class_name,
  },
  [227] = {
    [5] = alias_sym_class_name,
  },
  [228] = {
    [5] = alias_sym_class_name,
  },
  [229] = {
    [1] = alias_sym_label_name,
  },
  [230] = {
    [5] = alias_sym_class_name,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  sym__abstract_type, 2,
    sym__abstract_type,
    sym__parenthesized_abstract_type,
  sym__tuple_expression, 2,
    sym__tuple_expression,
    alias_sym_tuple_expression,
  sym__sequence_expression_anonymous, 2,
    sym__sequence_expression_anonymous,
    alias_sym_sequence_expression,
  sym__or_pattern_anonymous, 2,
    sym__or_pattern_anonymous,
    sym__or_binding_pattern_no_exn_anonymous,
  sym__or_binding_pattern_anonymous, 2,
    sym__or_binding_pattern_anonymous,
    sym__or_binding_pattern_no_exn_anonymous,
  sym__tuple_pattern, 2,
    sym__tuple_pattern,
    sym__tuple_binding_pattern_no_exn,
  sym__tuple_binding_pattern, 2,
    sym__tuple_binding_pattern,
    sym__tuple_binding_pattern_no_exn,
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 2,
  [4] = 2,
  [5] = 2,
  [6] = 6,
  [7] = 7,
  [8] = 8,
  [9] = 6,
  [10] = 7,
  [11] = 8,
  [12] = 12,
  [13] = 6,
  [14] = 6,
  [15] = 12,
  [16] = 6,
  [17] = 12,
  [18] = 6,
  [19] = 6,
  [20] = 6,
  [21] = 6,
  [22] = 6,
  [23] = 6,
  [24] = 6,
  [25] = 6,
  [26] = 6,
  [27] = 6,
  [28] = 12,
  [29] = 6,
  [30] = 6,
  [31] = 6,
  [32] = 6,
  [33] = 6,
  [34] = 6,
  [35] = 8,
  [36] = 36,
  [37] = 37,
  [38] = 38,
  [39] = 39,
  [40] = 40,
  [41] = 37,
  [42] = 42,
  [43] = 2,
  [44] = 44,
  [45] = 2,
  [46] = 46,
  [47] = 47,
  [48] = 47,
  [49] = 46,
  [50] = 46,
  [51] = 47,
  [52] = 46,
  [53] = 47,
  [54] = 46,
  [55] = 47,
  [56] = 46,
  [57] = 47,
  [58] = 46,
  [59] = 47,
  [60] = 46,
  [61] = 47,
  [62] = 46,
  [63] = 47,
  [64] = 46,
  [65] = 47,
  [66] = 47,
  [67] = 46,
  [68] = 2,
  [69] = 2,
  [70] = 2,
  [71] = 2,
  [72] = 2,
  [73] = 2,
  [74] = 2,
  [75] = 2,
  [76] = 2,
  [77] = 77,
  [78] = 78,
  [79] = 2,
  [80] = 77,
  [81] = 77,
  [82] = 82,
  [83] = 78,
  [84] = 82,
  [85] = 78,
  [86] = 82,
  [87] = 77,
  [88] = 2,
  [89] = 78,
  [90] = 2,
  [91] = 82,
  [92] = 2,
  [93] = 2,
  [94] = 94,
  [95] = 95,
  [96] = 95,
  [97] = 94,
  [98] = 95,
  [99] = 99,
  [100] = 95,
  [101] = 95,
  [102] = 95,
  [103] = 103,
  [104] = 95,
  [105] = 94,
  [106] = 99,
  [107] = 95,
  [108] = 95,
  [109] = 95,
  [110] = 103,
  [111] = 103,
  [112] = 95,
  [113] = 94,
  [114] = 114,
  [115] = 114,
  [116] = 116,
  [117] = 117,
  [118] = 116,
  [119] = 117,
  [120] = 120,
  [121] = 121,
  [122] = 117,
  [123] = 123,
  [124] = 117,
  [125] = 117,
  [126] = 116,
  [127] = 117,
  [128] = 116,
  [129] = 116,
  [130] = 117,
  [131] = 116,
  [132] = 114,
  [133] = 116,
  [134] = 120,
  [135] = 121,
  [136] = 117,
  [137] = 116,
  [138] = 120,
  [139] = 121,
  [140] = 120,
  [141] = 121,
  [142] = 120,
  [143] = 120,
  [144] = 120,
  [145] = 120,
  [146] = 120,
  [147] = 120,
  [148] = 120,
  [149] = 120,
  [150] = 123,
  [151] = 114,
  [152] = 2,
  [153] = 120,
  [154] = 123,
  [155] = 114,
  [156] = 123,
  [157] = 114,
  [158] = 123,
  [159] = 123,
  [160] = 114,
  [161] = 123,
  [162] = 114,
  [163] = 123,
  [164] = 114,
  [165] = 123,
  [166] = 114,
  [167] = 123,
  [168] = 2,
  [169] = 169,
  [170] = 170,
  [171] = 171,
  [172] = 170,
  [173] = 173,
  [174] = 174,
  [175] = 171,
  [176] = 170,
  [177] = 169,
  [178] = 178,
  [179] = 171,
  [180] = 170,
  [181] = 169,
  [182] = 169,
  [183] = 173,
  [184] = 174,
  [185] = 170,
  [186] = 169,
  [187] = 170,
  [188] = 178,
  [189] = 173,
  [190] = 173,
  [191] = 171,
  [192] = 169,
  [193] = 169,
  [194] = 173,
  [195] = 170,
  [196] = 173,
  [197] = 173,
  [198] = 169,
  [199] = 173,
  [200] = 174,
  [201] = 173,
  [202] = 169,
  [203] = 173,
  [204] = 174,
  [205] = 171,
  [206] = 171,
  [207] = 169,
  [208] = 178,
  [209] = 170,
  [210] = 169,
  [211] = 173,
  [212] = 212,
  [213] = 213,
  [214] = 214,
  [215] = 215,
  [216] = 216,
  [217] = 217,
  [218] = 218,
  [219] = 219,
  [220] = 220,
  [221] = 221,
  [222] = 222,
  [223] = 223,
  [224] = 224,
  [225] = 225,
  [226] = 226,
  [227] = 227,
  [228] = 228,
  [229] = 229,
  [230] = 230,
  [231] = 231,
  [232] = 232,
  [233] = 233,
  [234] = 234,
  [235] = 235,
  [236] = 236,
  [237] = 237,
  [238] = 238,
  [239] = 239,
  [240] = 240,
  [241] = 233,
  [242] = 214,
  [243] = 213,
  [244] = 214,
  [245] = 213,
  [246] = 214,
  [247] = 215,
  [248] = 216,
  [249] = 217,
  [250] = 218,
  [251] = 233,
  [252] = 224,
  [253] = 213,
  [254] = 214,
  [255] = 215,
  [256] = 216,
  [257] = 217,
  [258] = 218,
  [259] = 259,
  [260] = 224,
  [261] = 225,
  [262] = 212,
  [263] = 215,
  [264] = 216,
  [265] = 229,
  [266] = 217,
  [267] = 267,
  [268] = 218,
  [269] = 269,
  [270] = 233,
  [271] = 271,
  [272] = 272,
  [273] = 213,
  [274] = 224,
  [275] = 225,
  [276] = 214,
  [277] = 225,
  [278] = 229,
  [279] = 215,
  [280] = 216,
  [281] = 224,
  [282] = 225,
  [283] = 217,
  [284] = 233,
  [285] = 229,
  [286] = 218,
  [287] = 233,
  [288] = 229,
  [289] = 213,
  [290] = 224,
  [291] = 225,
  [292] = 214,
  [293] = 293,
  [294] = 229,
  [295] = 215,
  [296] = 216,
  [297] = 217,
  [298] = 224,
  [299] = 225,
  [300] = 234,
  [301] = 218,
  [302] = 235,
  [303] = 233,
  [304] = 236,
  [305] = 215,
  [306] = 224,
  [307] = 225,
  [308] = 213,
  [309] = 214,
  [310] = 215,
  [311] = 216,
  [312] = 224,
  [313] = 225,
  [314] = 217,
  [315] = 315,
  [316] = 218,
  [317] = 317,
  [318] = 216,
  [319] = 217,
  [320] = 320,
  [321] = 213,
  [322] = 231,
  [323] = 323,
  [324] = 224,
  [325] = 225,
  [326] = 326,
  [327] = 327,
  [328] = 328,
  [329] = 237,
  [330] = 238,
  [331] = 239,
  [332] = 229,
  [333] = 333,
  [334] = 334,
  [335] = 335,
  [336] = 231,
  [337] = 293,
  [338] = 338,
  [339] = 269,
  [340] = 271,
  [341] = 272,
  [342] = 323,
  [343] = 326,
  [344] = 327,
  [345] = 345,
  [346] = 346,
  [347] = 347,
  [348] = 219,
  [349] = 220,
  [350] = 227,
  [351] = 230,
  [352] = 233,
  [353] = 353,
  [354] = 213,
  [355] = 214,
  [356] = 215,
  [357] = 216,
  [358] = 217,
  [359] = 218,
  [360] = 345,
  [361] = 346,
  [362] = 234,
  [363] = 363,
  [364] = 235,
  [365] = 229,
  [366] = 229,
  [367] = 229,
  [368] = 236,
  [369] = 237,
  [370] = 238,
  [371] = 293,
  [372] = 269,
  [373] = 271,
  [374] = 272,
  [375] = 323,
  [376] = 326,
  [377] = 327,
  [378] = 347,
  [379] = 220,
  [380] = 239,
  [381] = 229,
  [382] = 218,
  [383] = 240,
  [384] = 233,
  [385] = 293,
  [386] = 269,
  [387] = 271,
  [388] = 272,
  [389] = 323,
  [390] = 326,
  [391] = 327,
  [392] = 347,
  [393] = 220,
  [394] = 347,
  [395] = 213,
  [396] = 269,
  [397] = 271,
  [398] = 272,
  [399] = 323,
  [400] = 326,
  [401] = 327,
  [402] = 214,
  [403] = 269,
  [404] = 271,
  [405] = 272,
  [406] = 323,
  [407] = 326,
  [408] = 327,
  [409] = 240,
  [410] = 269,
  [411] = 271,
  [412] = 272,
  [413] = 323,
  [414] = 326,
  [415] = 327,
  [416] = 269,
  [417] = 271,
  [418] = 272,
  [419] = 323,
  [420] = 326,
  [421] = 327,
  [422] = 422,
  [423] = 269,
  [424] = 271,
  [425] = 272,
  [426] = 323,
  [427] = 326,
  [428] = 327,
  [429] = 215,
  [430] = 269,
  [431] = 271,
  [432] = 272,
  [433] = 323,
  [434] = 326,
  [435] = 327,
  [436] = 216,
  [437] = 217,
  [438] = 438,
  [439] = 218,
  [440] = 212,
  [441] = 328,
  [442] = 232,
  [443] = 233,
  [444] = 233,
  [445] = 213,
  [446] = 214,
  [447] = 215,
  [448] = 216,
  [449] = 217,
  [450] = 218,
  [451] = 213,
  [452] = 229,
  [453] = 214,
  [454] = 215,
  [455] = 212,
  [456] = 328,
  [457] = 232,
  [458] = 216,
  [459] = 217,
  [460] = 212,
  [461] = 328,
  [462] = 232,
  [463] = 218,
  [464] = 233,
  [465] = 465,
  [466] = 328,
  [467] = 232,
  [468] = 213,
  [469] = 214,
  [470] = 212,
  [471] = 328,
  [472] = 232,
  [473] = 473,
  [474] = 212,
  [475] = 328,
  [476] = 232,
  [477] = 215,
  [478] = 216,
  [479] = 212,
  [480] = 328,
  [481] = 232,
  [482] = 217,
  [483] = 483,
  [484] = 212,
  [485] = 328,
  [486] = 232,
  [487] = 218,
  [488] = 233,
  [489] = 212,
  [490] = 328,
  [491] = 232,
  [492] = 259,
  [493] = 320,
  [494] = 259,
  [495] = 320,
  [496] = 259,
  [497] = 320,
  [498] = 229,
  [499] = 499,
  [500] = 499,
  [501] = 499,
  [502] = 499,
  [503] = 499,
  [504] = 499,
  [505] = 499,
  [506] = 499,
  [507] = 499,
  [508] = 499,
  [509] = 499,
  [510] = 499,
  [511] = 511,
  [512] = 499,
  [513] = 499,
  [514] = 499,
  [515] = 511,
  [516] = 499,
  [517] = 499,
  [518] = 518,
  [519] = 499,
  [520] = 518,
  [521] = 499,
  [522] = 499,
  [523] = 499,
  [524] = 499,
  [525] = 525,
  [526] = 526,
  [527] = 527,
  [528] = 528,
  [529] = 529,
  [530] = 530,
  [531] = 531,
  [532] = 532,
  [533] = 533,
  [534] = 534,
  [535] = 535,
  [536] = 536,
  [537] = 525,
  [538] = 538,
  [539] = 539,
  [540] = 540,
  [541] = 526,
  [542] = 542,
  [543] = 528,
  [544] = 529,
  [545] = 530,
  [546] = 531,
  [547] = 532,
  [548] = 533,
  [549] = 534,
  [550] = 535,
  [551] = 536,
  [552] = 538,
  [553] = 539,
  [554] = 526,
  [555] = 540,
  [556] = 526,
  [557] = 528,
  [558] = 558,
  [559] = 528,
  [560] = 529,
  [561] = 530,
  [562] = 531,
  [563] = 532,
  [564] = 533,
  [565] = 534,
  [566] = 535,
  [567] = 536,
  [568] = 525,
  [569] = 538,
  [570] = 539,
  [571] = 540,
  [572] = 526,
  [573] = 529,
  [574] = 530,
  [575] = 531,
  [576] = 532,
  [577] = 533,
  [578] = 534,
  [579] = 535,
  [580] = 536,
  [581] = 525,
  [582] = 538,
  [583] = 539,
  [584] = 528,
  [585] = 529,
  [586] = 530,
  [587] = 531,
  [588] = 532,
  [589] = 533,
  [590] = 534,
  [591] = 535,
  [592] = 536,
  [593] = 525,
  [594] = 538,
  [595] = 539,
  [596] = 540,
  [597] = 526,
  [598] = 598,
  [599] = 558,
  [600] = 528,
  [601] = 529,
  [602] = 530,
  [603] = 531,
  [604] = 532,
  [605] = 533,
  [606] = 534,
  [607] = 535,
  [608] = 536,
  [609] = 525,
  [610] = 538,
  [611] = 539,
  [612] = 540,
  [613] = 526,
  [614] = 528,
  [615] = 529,
  [616] = 530,
  [617] = 531,
  [618] = 532,
  [619] = 533,
  [620] = 534,
  [621] = 536,
  [622] = 525,
  [623] = 538,
  [624] = 539,
  [625] = 540,
  [626] = 526,
  [627] = 528,
  [628] = 529,
  [629] = 530,
  [630] = 531,
  [631] = 532,
  [632] = 533,
  [633] = 534,
  [634] = 535,
  [635] = 536,
  [636] = 525,
  [637] = 538,
  [638] = 539,
  [639] = 540,
  [640] = 526,
  [641] = 528,
  [642] = 529,
  [643] = 530,
  [644] = 531,
  [645] = 532,
  [646] = 533,
  [647] = 534,
  [648] = 535,
  [649] = 536,
  [650] = 525,
  [651] = 538,
  [652] = 539,
  [653] = 540,
  [654] = 526,
  [655] = 528,
  [656] = 528,
  [657] = 529,
  [658] = 530,
  [659] = 531,
  [660] = 532,
  [661] = 533,
  [662] = 534,
  [663] = 535,
  [664] = 536,
  [665] = 525,
  [666] = 538,
  [667] = 539,
  [668] = 540,
  [669] = 526,
  [670] = 529,
  [671] = 530,
  [672] = 531,
  [673] = 532,
  [674] = 533,
  [675] = 534,
  [676] = 535,
  [677] = 536,
  [678] = 525,
  [679] = 538,
  [680] = 539,
  [681] = 681,
  [682] = 598,
  [683] = 540,
  [684] = 526,
  [685] = 528,
  [686] = 528,
  [687] = 529,
  [688] = 530,
  [689] = 531,
  [690] = 532,
  [691] = 533,
  [692] = 534,
  [693] = 535,
  [694] = 536,
  [695] = 525,
  [696] = 538,
  [697] = 539,
  [698] = 540,
  [699] = 529,
  [700] = 530,
  [701] = 531,
  [702] = 526,
  [703] = 532,
  [704] = 533,
  [705] = 534,
  [706] = 535,
  [707] = 536,
  [708] = 525,
  [709] = 538,
  [710] = 539,
  [711] = 540,
  [712] = 540,
  [713] = 535,
  [714] = 714,
  [715] = 77,
  [716] = 716,
  [717] = 82,
  [718] = 78,
  [719] = 719,
  [720] = 719,
  [721] = 719,
  [722] = 719,
  [723] = 719,
  [724] = 719,
  [725] = 719,
  [726] = 719,
  [727] = 719,
  [728] = 719,
  [729] = 719,
  [730] = 719,
  [731] = 719,
  [732] = 719,
  [733] = 733,
  [734] = 734,
  [735] = 734,
  [736] = 734,
  [737] = 734,
  [738] = 734,
  [739] = 734,
  [740] = 734,
  [741] = 734,
  [742] = 733,
  [743] = 734,
  [744] = 734,
  [745] = 734,
  [746] = 734,
  [747] = 734,
  [748] = 734,
  [749] = 749,
  [750] = 750,
  [751] = 751,
  [752] = 752,
  [753] = 753,
  [754] = 754,
  [755] = 755,
  [756] = 756,
  [757] = 757,
  [758] = 758,
  [759] = 759,
  [760] = 760,
  [761] = 761,
  [762] = 762,
  [763] = 763,
  [764] = 764,
  [765] = 765,
  [766] = 766,
  [767] = 767,
  [768] = 768,
  [769] = 769,
  [770] = 770,
  [771] = 771,
  [772] = 772,
  [773] = 773,
  [774] = 774,
  [775] = 775,
  [776] = 776,
  [777] = 777,
  [778] = 778,
  [779] = 779,
  [780] = 780,
  [781] = 781,
  [782] = 782,
  [783] = 783,
  [784] = 784,
  [785] = 785,
  [786] = 786,
  [787] = 787,
  [788] = 788,
  [789] = 789,
  [790] = 790,
  [791] = 791,
  [792] = 792,
  [793] = 793,
  [794] = 794,
  [795] = 795,
  [796] = 796,
  [797] = 797,
  [798] = 798,
  [799] = 799,
  [800] = 800,
  [801] = 801,
  [802] = 802,
  [803] = 803,
  [804] = 804,
  [805] = 805,
  [806] = 806,
  [807] = 807,
  [808] = 808,
  [809] = 809,
  [810] = 810,
  [811] = 811,
  [812] = 812,
  [813] = 813,
  [814] = 814,
  [815] = 815,
  [816] = 816,
  [817] = 817,
  [818] = 818,
  [819] = 819,
  [820] = 820,
  [821] = 821,
  [822] = 822,
  [823] = 823,
  [824] = 824,
  [825] = 825,
  [826] = 826,
  [827] = 827,
  [828] = 828,
  [829] = 829,
  [830] = 830,
  [831] = 831,
  [832] = 749,
  [833] = 750,
  [834] = 78,
  [835] = 835,
  [836] = 750,
  [837] = 749,
  [838] = 714,
  [839] = 839,
  [840] = 753,
  [841] = 756,
  [842] = 842,
  [843] = 757,
  [844] = 754,
  [845] = 845,
  [846] = 835,
  [847] = 847,
  [848] = 848,
  [849] = 849,
  [850] = 850,
  [851] = 755,
  [852] = 752,
  [853] = 853,
  [854] = 751,
  [855] = 855,
  [856] = 856,
  [857] = 857,
  [858] = 758,
  [859] = 759,
  [860] = 860,
  [861] = 861,
  [862] = 716,
  [863] = 800,
  [864] = 809,
  [865] = 811,
  [866] = 772,
  [867] = 803,
  [868] = 791,
  [869] = 802,
  [870] = 761,
  [871] = 792,
  [872] = 766,
  [873] = 773,
  [874] = 775,
  [875] = 782,
  [876] = 812,
  [877] = 813,
  [878] = 760,
  [879] = 805,
  [880] = 806,
  [881] = 808,
  [882] = 816,
  [883] = 830,
  [884] = 804,
  [885] = 793,
  [886] = 796,
  [887] = 798,
  [888] = 820,
  [889] = 824,
  [890] = 828,
  [891] = 829,
  [892] = 789,
  [893] = 790,
  [894] = 795,
  [895] = 821,
  [896] = 825,
  [897] = 814,
  [898] = 817,
  [899] = 819,
  [900] = 807,
  [901] = 774,
  [902] = 801,
  [903] = 822,
  [904] = 826,
  [905] = 810,
  [906] = 815,
  [907] = 818,
  [908] = 908,
  [909] = 784,
  [910] = 794,
  [911] = 785,
  [912] = 823,
  [913] = 762,
  [914] = 767,
  [915] = 771,
  [916] = 786,
  [917] = 787,
  [918] = 788,
  [919] = 831,
  [920] = 797,
  [921] = 827,
  [922] = 763,
  [923] = 764,
  [924] = 765,
  [925] = 768,
  [926] = 769,
  [927] = 770,
  [928] = 776,
  [929] = 777,
  [930] = 778,
  [931] = 779,
  [932] = 780,
  [933] = 781,
  [934] = 783,
  [935] = 799,
  [936] = 936,
  [937] = 835,
  [938] = 938,
  [939] = 754,
  [940] = 755,
  [941] = 752,
  [942] = 751,
  [943] = 759,
  [944] = 757,
  [945] = 753,
  [946] = 758,
  [947] = 756,
  [948] = 772,
  [949] = 799,
  [950] = 810,
  [951] = 815,
  [952] = 818,
  [953] = 803,
  [954] = 805,
  [955] = 806,
  [956] = 808,
  [957] = 793,
  [958] = 816,
  [959] = 783,
  [960] = 809,
  [961] = 855,
  [962] = 811,
  [963] = 791,
  [964] = 766,
  [965] = 812,
  [966] = 813,
  [967] = 848,
  [968] = 849,
  [969] = 775,
  [970] = 857,
  [971] = 796,
  [972] = 842,
  [973] = 860,
  [974] = 861,
  [975] = 850,
  [976] = 853,
  [977] = 856,
  [978] = 830,
  [979] = 804,
  [980] = 839,
  [981] = 760,
  [982] = 982,
  [983] = 782,
  [984] = 828,
  [985] = 829,
  [986] = 784,
  [987] = 794,
  [988] = 785,
  [989] = 807,
  [990] = 800,
  [991] = 802,
  [992] = 823,
  [993] = 762,
  [994] = 767,
  [995] = 789,
  [996] = 790,
  [997] = 795,
  [998] = 771,
  [999] = 774,
  [1000] = 798,
  [1001] = 801,
  [1002] = 821,
  [1003] = 822,
  [1004] = 824,
  [1005] = 938,
  [1006] = 825,
  [1007] = 826,
  [1008] = 814,
  [1009] = 786,
  [1010] = 787,
  [1011] = 788,
  [1012] = 792,
  [1013] = 817,
  [1014] = 831,
  [1015] = 797,
  [1016] = 827,
  [1017] = 761,
  [1018] = 819,
  [1019] = 763,
  [1020] = 764,
  [1021] = 765,
  [1022] = 768,
  [1023] = 769,
  [1024] = 770,
  [1025] = 773,
  [1026] = 781,
  [1027] = 776,
  [1028] = 777,
  [1029] = 778,
  [1030] = 779,
  [1031] = 780,
  [1032] = 820,
  [1033] = 938,
  [1034] = 758,
  [1035] = 759,
  [1036] = 753,
  [1037] = 845,
  [1038] = 714,
  [1039] = 751,
  [1040] = 752,
  [1041] = 847,
  [1042] = 908,
  [1043] = 757,
  [1044] = 756,
  [1045] = 754,
  [1046] = 755,
  [1047] = 771,
  [1048] = 766,
  [1049] = 787,
  [1050] = 827,
  [1051] = 829,
  [1052] = 794,
  [1053] = 761,
  [1054] = 762,
  [1055] = 767,
  [1056] = 938,
  [1057] = 795,
  [1058] = 778,
  [1059] = 779,
  [1060] = 780,
  [1061] = 781,
  [1062] = 786,
  [1063] = 782,
  [1064] = 788,
  [1065] = 801,
  [1066] = 822,
  [1067] = 798,
  [1068] = 802,
  [1069] = 790,
  [1070] = 768,
  [1071] = 765,
  [1072] = 826,
  [1073] = 810,
  [1074] = 815,
  [1075] = 769,
  [1076] = 791,
  [1077] = 830,
  [1078] = 772,
  [1079] = 796,
  [1080] = 839,
  [1081] = 818,
  [1082] = 800,
  [1083] = 783,
  [1084] = 784,
  [1085] = 792,
  [1086] = 789,
  [1087] = 770,
  [1088] = 908,
  [1089] = 809,
  [1090] = 811,
  [1091] = 823,
  [1092] = 812,
  [1093] = 813,
  [1094] = 714,
  [1095] = 835,
  [1096] = 773,
  [1097] = 936,
  [1098] = 825,
  [1099] = 793,
  [1100] = 760,
  [1101] = 817,
  [1102] = 803,
  [1103] = 831,
  [1104] = 805,
  [1105] = 763,
  [1106] = 806,
  [1107] = 808,
  [1108] = 819,
  [1109] = 816,
  [1110] = 807,
  [1111] = 820,
  [1112] = 764,
  [1113] = 821,
  [1114] = 775,
  [1115] = 776,
  [1116] = 774,
  [1117] = 777,
  [1118] = 785,
  [1119] = 828,
  [1120] = 716,
  [1121] = 797,
  [1122] = 824,
  [1123] = 714,
  [1124] = 799,
  [1125] = 804,
  [1126] = 814,
  [1127] = 936,
  [1128] = 716,
  [1129] = 716,
  [1130] = 982,
  [1131] = 845,
  [1132] = 847,
  [1133] = 1133,
  [1134] = 1133,
  [1135] = 1133,
  [1136] = 1133,
  [1137] = 1133,
  [1138] = 1133,
  [1139] = 1133,
  [1140] = 1133,
  [1141] = 1133,
  [1142] = 839,
  [1143] = 1133,
  [1144] = 845,
  [1145] = 982,
  [1146] = 847,
  [1147] = 1147,
  [1148] = 1148,
  [1149] = 1149,
  [1150] = 1150,
  [1151] = 1151,
  [1152] = 1149,
  [1153] = 1153,
  [1154] = 1147,
  [1155] = 1155,
  [1156] = 1149,
  [1157] = 1157,
  [1158] = 1155,
  [1159] = 1151,
  [1160] = 1155,
  [1161] = 1150,
  [1162] = 1147,
  [1163] = 1147,
  [1164] = 1150,
  [1165] = 1147,
  [1166] = 1149,
  [1167] = 1147,
  [1168] = 1149,
  [1169] = 1151,
  [1170] = 1170,
  [1171] = 1150,
  [1172] = 1147,
  [1173] = 1150,
  [1174] = 1149,
  [1175] = 1155,
  [1176] = 1176,
  [1177] = 1151,
  [1178] = 1149,
  [1179] = 1151,
  [1180] = 1150,
  [1181] = 1150,
  [1182] = 1149,
  [1183] = 1155,
  [1184] = 1151,
  [1185] = 733,
  [1186] = 1147,
  [1187] = 1150,
  [1188] = 1151,
  [1189] = 1151,
  [1190] = 1147,
  [1191] = 1155,
  [1192] = 1155,
  [1193] = 1155,
  [1194] = 1194,
  [1195] = 1155,
  [1196] = 1150,
  [1197] = 1149,
  [1198] = 1147,
  [1199] = 1150,
  [1200] = 1151,
  [1201] = 1149,
  [1202] = 1151,
  [1203] = 1155,
  [1204] = 1204,
  [1205] = 1205,
  [1206] = 1206,
  [1207] = 1207,
  [1208] = 1208,
  [1209] = 1209,
  [1210] = 714,
  [1211] = 1211,
  [1212] = 1208,
  [1213] = 1206,
  [1214] = 1214,
  [1215] = 1215,
  [1216] = 1205,
  [1217] = 1208,
  [1218] = 1206,
  [1219] = 1209,
  [1220] = 1220,
  [1221] = 1209,
  [1222] = 1209,
  [1223] = 1215,
  [1224] = 1209,
  [1225] = 1225,
  [1226] = 1226,
  [1227] = 1211,
  [1228] = 1226,
  [1229] = 1226,
  [1230] = 1211,
  [1231] = 1207,
  [1232] = 1208,
  [1233] = 1206,
  [1234] = 1204,
  [1235] = 1205,
  [1236] = 1207,
  [1237] = 1208,
  [1238] = 1206,
  [1239] = 1215,
  [1240] = 1204,
  [1241] = 1215,
  [1242] = 1207,
  [1243] = 1208,
  [1244] = 1206,
  [1245] = 1245,
  [1246] = 1214,
  [1247] = 1206,
  [1248] = 1220,
  [1249] = 1207,
  [1250] = 1205,
  [1251] = 1220,
  [1252] = 908,
  [1253] = 1208,
  [1254] = 1206,
  [1255] = 1214,
  [1256] = 1220,
  [1257] = 1226,
  [1258] = 1206,
  [1259] = 1211,
  [1260] = 1208,
  [1261] = 1207,
  [1262] = 1225,
  [1263] = 1225,
  [1264] = 1225,
  [1265] = 1204,
  [1266] = 1266,
  [1267] = 1267,
  [1268] = 1268,
  [1269] = 1269,
  [1270] = 1268,
  [1271] = 1271,
  [1272] = 1272,
  [1273] = 1273,
  [1274] = 1274,
  [1275] = 1267,
  [1276] = 1267,
  [1277] = 1272,
  [1278] = 716,
  [1279] = 1268,
  [1280] = 1272,
  [1281] = 1268,
  [1282] = 1273,
  [1283] = 1272,
  [1284] = 1268,
  [1285] = 1269,
  [1286] = 1272,
  [1287] = 1269,
  [1288] = 1268,
  [1289] = 1289,
  [1290] = 1289,
  [1291] = 1268,
  [1292] = 1272,
  [1293] = 1269,
  [1294] = 1268,
  [1295] = 1266,
  [1296] = 1272,
  [1297] = 936,
  [1298] = 1274,
  [1299] = 1269,
  [1300] = 1271,
  [1301] = 1272,
  [1302] = 1302,
  [1303] = 1303,
  [1304] = 1303,
  [1305] = 1305,
  [1306] = 1306,
  [1307] = 1307,
  [1308] = 1308,
  [1309] = 1309,
  [1310] = 1310,
  [1311] = 1311,
  [1312] = 1312,
  [1313] = 1313,
  [1314] = 1314,
  [1315] = 1307,
  [1316] = 1316,
  [1317] = 1311,
  [1318] = 1318,
  [1319] = 1314,
  [1320] = 1310,
  [1321] = 1310,
  [1322] = 1305,
  [1323] = 1307,
  [1324] = 1324,
  [1325] = 1305,
  [1326] = 1306,
  [1327] = 1314,
  [1328] = 1306,
  [1329] = 1302,
  [1330] = 1308,
  [1331] = 1307,
  [1332] = 1311,
  [1333] = 1310,
  [1334] = 1314,
  [1335] = 1302,
  [1336] = 1311,
  [1337] = 1310,
  [1338] = 1307,
  [1339] = 1339,
  [1340] = 1316,
  [1341] = 1314,
  [1342] = 1342,
  [1343] = 1339,
  [1344] = 1316,
  [1345] = 1307,
  [1346] = 1311,
  [1347] = 1347,
  [1348] = 1314,
  [1349] = 1314,
  [1350] = 1307,
  [1351] = 1351,
  [1352] = 1307,
  [1353] = 1353,
  [1354] = 1347,
  [1355] = 1314,
  [1356] = 1312,
  [1357] = 1353,
  [1358] = 1353,
  [1359] = 1307,
  [1360] = 1313,
  [1361] = 1308,
  [1362] = 1314,
  [1363] = 1310,
  [1364] = 1353,
  [1365] = 1347,
  [1366] = 1307,
  [1367] = 1318,
  [1368] = 1302,
  [1369] = 1314,
  [1370] = 1307,
  [1371] = 1314,
  [1372] = 1372,
  [1373] = 1373,
  [1374] = 1308,
  [1375] = 1339,
  [1376] = 1376,
  [1377] = 1342,
  [1378] = 1311,
  [1379] = 1302,
  [1380] = 1316,
  [1381] = 1342,
  [1382] = 1339,
  [1383] = 1342,
  [1384] = 1316,
  [1385] = 1339,
  [1386] = 1386,
  [1387] = 938,
  [1388] = 1388,
  [1389] = 1388,
  [1390] = 938,
  [1391] = 1388,
  [1392] = 1388,
  [1393] = 1388,
  [1394] = 1388,
  [1395] = 1388,
  [1396] = 1388,
  [1397] = 1388,
  [1398] = 1388,
  [1399] = 1386,
  [1400] = 1400,
  [1401] = 1388,
  [1402] = 1386,
  [1403] = 1386,
  [1404] = 1386,
  [1405] = 1386,
  [1406] = 1386,
  [1407] = 1386,
  [1408] = 1386,
  [1409] = 1386,
  [1410] = 1386,
  [1411] = 1400,
  [1412] = 1412,
  [1413] = 1412,
  [1414] = 1414,
  [1415] = 1415,
  [1416] = 1414,
  [1417] = 1417,
  [1418] = 1412,
  [1419] = 1415,
  [1420] = 1417,
  [1421] = 1415,
  [1422] = 1417,
  [1423] = 1414,
  [1424] = 842,
  [1425] = 1425,
  [1426] = 1426,
  [1427] = 850,
  [1428] = 853,
  [1429] = 1429,
  [1430] = 856,
  [1431] = 855,
  [1432] = 1429,
  [1433] = 1429,
  [1434] = 1425,
  [1435] = 1429,
  [1436] = 1426,
  [1437] = 1429,
  [1438] = 1429,
  [1439] = 1425,
  [1440] = 1429,
  [1441] = 848,
  [1442] = 849,
  [1443] = 1426,
  [1444] = 860,
  [1445] = 1425,
  [1446] = 1429,
  [1447] = 861,
  [1448] = 1429,
  [1449] = 1426,
  [1450] = 857,
  [1451] = 1429,
  [1452] = 1452,
  [1453] = 1452,
  [1454] = 1452,
  [1455] = 1452,
  [1456] = 1452,
  [1457] = 1452,
  [1458] = 1452,
  [1459] = 1452,
  [1460] = 1452,
  [1461] = 1452,
  [1462] = 1452,
  [1463] = 1463,
  [1464] = 1464,
  [1465] = 1465,
  [1466] = 1463,
  [1467] = 1463,
  [1468] = 1464,
  [1469] = 1464,
  [1470] = 1465,
  [1471] = 1463,
  [1472] = 1472,
  [1473] = 1473,
  [1474] = 1474,
  [1475] = 1464,
  [1476] = 1476,
  [1477] = 1465,
  [1478] = 1463,
  [1479] = 1463,
  [1480] = 1480,
  [1481] = 1463,
  [1482] = 1464,
  [1483] = 1465,
  [1484] = 1464,
  [1485] = 1464,
  [1486] = 1486,
  [1487] = 1463,
  [1488] = 1465,
  [1489] = 1472,
  [1490] = 1473,
  [1491] = 1463,
  [1492] = 1476,
  [1493] = 1464,
  [1494] = 1464,
  [1495] = 1464,
  [1496] = 1472,
  [1497] = 1473,
  [1498] = 1472,
  [1499] = 1473,
  [1500] = 1476,
  [1501] = 1463,
  [1502] = 1464,
  [1503] = 1476,
  [1504] = 1476,
  [1505] = 1465,
  [1506] = 1463,
  [1507] = 758,
  [1508] = 759,
  [1509] = 1509,
  [1510] = 1510,
  [1511] = 1509,
  [1512] = 753,
  [1513] = 757,
  [1514] = 756,
  [1515] = 1515,
  [1516] = 754,
  [1517] = 835,
  [1518] = 755,
  [1519] = 1509,
  [1520] = 938,
  [1521] = 1521,
  [1522] = 752,
  [1523] = 751,
  [1524] = 816,
  [1525] = 798,
  [1526] = 801,
  [1527] = 938,
  [1528] = 821,
  [1529] = 822,
  [1530] = 772,
  [1531] = 824,
  [1532] = 825,
  [1533] = 826,
  [1534] = 814,
  [1535] = 817,
  [1536] = 819,
  [1537] = 807,
  [1538] = 1538,
  [1539] = 796,
  [1540] = 820,
  [1541] = 784,
  [1542] = 799,
  [1543] = 794,
  [1544] = 810,
  [1545] = 785,
  [1546] = 800,
  [1547] = 823,
  [1548] = 762,
  [1549] = 767,
  [1550] = 771,
  [1551] = 786,
  [1552] = 787,
  [1553] = 788,
  [1554] = 831,
  [1555] = 797,
  [1556] = 827,
  [1557] = 763,
  [1558] = 764,
  [1559] = 938,
  [1560] = 765,
  [1561] = 815,
  [1562] = 768,
  [1563] = 769,
  [1564] = 774,
  [1565] = 818,
  [1566] = 776,
  [1567] = 777,
  [1568] = 778,
  [1569] = 779,
  [1570] = 780,
  [1571] = 781,
  [1572] = 828,
  [1573] = 783,
  [1574] = 829,
  [1575] = 809,
  [1576] = 811,
  [1577] = 793,
  [1578] = 789,
  [1579] = 803,
  [1580] = 791,
  [1581] = 790,
  [1582] = 802,
  [1583] = 792,
  [1584] = 766,
  [1585] = 773,
  [1586] = 782,
  [1587] = 812,
  [1588] = 813,
  [1589] = 760,
  [1590] = 805,
  [1591] = 806,
  [1592] = 808,
  [1593] = 830,
  [1594] = 804,
  [1595] = 761,
  [1596] = 795,
  [1597] = 775,
  [1598] = 770,
  [1599] = 1599,
  [1600] = 1599,
  [1601] = 1601,
  [1602] = 1599,
  [1603] = 1601,
  [1604] = 1601,
  [1605] = 1599,
  [1606] = 1601,
  [1607] = 1601,
  [1608] = 1599,
  [1609] = 1599,
  [1610] = 1601,
  [1611] = 1599,
  [1612] = 1601,
  [1613] = 1601,
  [1614] = 1599,
  [1615] = 1601,
  [1616] = 1599,
  [1617] = 1599,
  [1618] = 1599,
  [1619] = 1601,
  [1620] = 938,
  [1621] = 1599,
  [1622] = 1601,
  [1623] = 1599,
  [1624] = 1601,
  [1625] = 1601,
  [1626] = 1626,
  [1627] = 839,
  [1628] = 1628,
  [1629] = 1628,
  [1630] = 1628,
  [1631] = 1628,
  [1632] = 1628,
  [1633] = 1628,
  [1634] = 1628,
  [1635] = 1628,
  [1636] = 1628,
  [1637] = 1628,
  [1638] = 938,
  [1639] = 1628,
  [1640] = 938,
  [1641] = 1628,
  [1642] = 1628,
  [1643] = 804,
  [1644] = 845,
  [1645] = 828,
  [1646] = 1646,
  [1647] = 1647,
  [1648] = 1648,
  [1649] = 1649,
  [1650] = 793,
  [1651] = 1651,
  [1652] = 1652,
  [1653] = 1653,
  [1654] = 847,
  [1655] = 1649,
  [1656] = 1649,
  [1657] = 1657,
  [1658] = 938,
  [1659] = 1659,
  [1660] = 1660,
  [1661] = 1661,
  [1662] = 830,
  [1663] = 1649,
  [1664] = 1664,
  [1665] = 1665,
  [1666] = 1666,
  [1667] = 1667,
  [1668] = 1668,
  [1669] = 1669,
  [1670] = 1670,
  [1671] = 1671,
  [1672] = 1672,
  [1673] = 1673,
  [1674] = 1665,
  [1675] = 1651,
  [1676] = 1653,
  [1677] = 794,
  [1678] = 1672,
  [1679] = 827,
  [1680] = 768,
  [1681] = 1681,
  [1682] = 769,
  [1683] = 770,
  [1684] = 773,
  [1685] = 793,
  [1686] = 751,
  [1687] = 1687,
  [1688] = 1672,
  [1689] = 763,
  [1690] = 761,
  [1691] = 1681,
  [1692] = 778,
  [1693] = 779,
  [1694] = 796,
  [1695] = 1687,
  [1696] = 1659,
  [1697] = 791,
  [1698] = 796,
  [1699] = 780,
  [1700] = 762,
  [1701] = 781,
  [1702] = 774,
  [1703] = 767,
  [1704] = 802,
  [1705] = 1672,
  [1706] = 1681,
  [1707] = 1669,
  [1708] = 775,
  [1709] = 776,
  [1710] = 782,
  [1711] = 820,
  [1712] = 1672,
  [1713] = 1681,
  [1714] = 816,
  [1715] = 786,
  [1716] = 787,
  [1717] = 1672,
  [1718] = 785,
  [1719] = 1681,
  [1720] = 1672,
  [1721] = 1660,
  [1722] = 1661,
  [1723] = 783,
  [1724] = 808,
  [1725] = 1672,
  [1726] = 1666,
  [1727] = 758,
  [1728] = 1670,
  [1729] = 1672,
  [1730] = 759,
  [1731] = 1668,
  [1732] = 1671,
  [1733] = 1672,
  [1734] = 1647,
  [1735] = 1652,
  [1736] = 753,
  [1737] = 820,
  [1738] = 1672,
  [1739] = 1667,
  [1740] = 788,
  [1741] = 1672,
  [1742] = 777,
  [1743] = 1672,
  [1744] = 809,
  [1745] = 811,
  [1746] = 771,
  [1747] = 812,
  [1748] = 1673,
  [1749] = 813,
  [1750] = 823,
  [1751] = 828,
  [1752] = 792,
  [1753] = 829,
  [1754] = 760,
  [1755] = 831,
  [1756] = 757,
  [1757] = 756,
  [1758] = 765,
  [1759] = 766,
  [1760] = 797,
  [1761] = 800,
  [1762] = 829,
  [1763] = 1673,
  [1764] = 1648,
  [1765] = 754,
  [1766] = 803,
  [1767] = 1673,
  [1768] = 805,
  [1769] = 1687,
  [1770] = 806,
  [1771] = 1657,
  [1772] = 1664,
  [1773] = 781,
  [1774] = 782,
  [1775] = 1474,
  [1776] = 814,
  [1777] = 812,
  [1778] = 1665,
  [1779] = 1651,
  [1780] = 1653,
  [1781] = 813,
  [1782] = 760,
  [1783] = 828,
  [1784] = 817,
  [1785] = 805,
  [1786] = 806,
  [1787] = 808,
  [1788] = 819,
  [1789] = 816,
  [1790] = 807,
  [1791] = 799,
  [1792] = 1486,
  [1793] = 1648,
  [1794] = 776,
  [1795] = 789,
  [1796] = 810,
  [1797] = 815,
  [1798] = 818,
  [1799] = 790,
  [1800] = 778,
  [1801] = 779,
  [1802] = 780,
  [1803] = 795,
  [1804] = 777,
  [1805] = 783,
  [1806] = 758,
  [1807] = 759,
  [1808] = 753,
  [1809] = 809,
  [1810] = 811,
  [1811] = 804,
  [1812] = 1812,
  [1813] = 1813,
  [1814] = 757,
  [1815] = 756,
  [1816] = 754,
  [1817] = 803,
  [1818] = 755,
  [1819] = 1812,
  [1820] = 1813,
  [1821] = 801,
  [1822] = 830,
  [1823] = 1823,
  [1824] = 1659,
  [1825] = 1825,
  [1826] = 1826,
  [1827] = 789,
  [1828] = 1669,
  [1829] = 1829,
  [1830] = 752,
  [1831] = 790,
  [1832] = 793,
  [1833] = 796,
  [1834] = 822,
  [1835] = 828,
  [1836] = 829,
  [1837] = 795,
  [1838] = 791,
  [1839] = 820,
  [1840] = 1840,
  [1841] = 1841,
  [1842] = 825,
  [1843] = 826,
  [1844] = 755,
  [1845] = 764,
  [1846] = 1660,
  [1847] = 1661,
  [1848] = 1648,
  [1849] = 1664,
  [1850] = 1666,
  [1851] = 1670,
  [1852] = 1668,
  [1853] = 770,
  [1854] = 1647,
  [1855] = 774,
  [1856] = 793,
  [1857] = 798,
  [1858] = 1652,
  [1859] = 802,
  [1860] = 761,
  [1861] = 1480,
  [1862] = 752,
  [1863] = 801,
  [1864] = 1667,
  [1865] = 821,
  [1866] = 789,
  [1867] = 790,
  [1868] = 795,
  [1869] = 822,
  [1870] = 774,
  [1871] = 792,
  [1872] = 1872,
  [1873] = 798,
  [1874] = 801,
  [1875] = 824,
  [1876] = 825,
  [1877] = 821,
  [1878] = 822,
  [1879] = 826,
  [1880] = 824,
  [1881] = 825,
  [1882] = 826,
  [1883] = 766,
  [1884] = 1872,
  [1885] = 814,
  [1886] = 817,
  [1887] = 773,
  [1888] = 1657,
  [1889] = 819,
  [1890] = 775,
  [1891] = 807,
  [1892] = 1840,
  [1893] = 1841,
  [1894] = 1659,
  [1895] = 1669,
  [1896] = 794,
  [1897] = 785,
  [1898] = 799,
  [1899] = 800,
  [1900] = 823,
  [1901] = 762,
  [1902] = 767,
  [1903] = 771,
  [1904] = 810,
  [1905] = 815,
  [1906] = 786,
  [1907] = 787,
  [1908] = 788,
  [1909] = 831,
  [1910] = 797,
  [1911] = 827,
  [1912] = 751,
  [1913] = 763,
  [1914] = 764,
  [1915] = 765,
  [1916] = 818,
  [1917] = 768,
  [1918] = 769,
  [1919] = 1671,
  [1920] = 756,
  [1921] = 818,
  [1922] = 1922,
  [1923] = 1652,
  [1924] = 830,
  [1925] = 1925,
  [1926] = 1926,
  [1927] = 1927,
  [1928] = 804,
  [1929] = 1929,
  [1930] = 1930,
  [1931] = 1931,
  [1932] = 1932,
  [1933] = 1933,
  [1934] = 1934,
  [1935] = 817,
  [1936] = 1660,
  [1937] = 794,
  [1938] = 821,
  [1939] = 757,
  [1940] = 761,
  [1941] = 1659,
  [1942] = 798,
  [1943] = 1943,
  [1944] = 824,
  [1945] = 1945,
  [1946] = 755,
  [1947] = 792,
  [1948] = 819,
  [1949] = 789,
  [1950] = 790,
  [1951] = 795,
  [1952] = 1952,
  [1953] = 785,
  [1954] = 1954,
  [1955] = 766,
  [1956] = 754,
  [1957] = 821,
  [1958] = 773,
  [1959] = 1841,
  [1960] = 825,
  [1961] = 814,
  [1962] = 817,
  [1963] = 819,
  [1964] = 824,
  [1965] = 807,
  [1966] = 775,
  [1967] = 752,
  [1968] = 774,
  [1969] = 1969,
  [1970] = 1970,
  [1971] = 782,
  [1972] = 801,
  [1973] = 822,
  [1974] = 826,
  [1975] = 1975,
  [1976] = 793,
  [1977] = 796,
  [1978] = 820,
  [1979] = 828,
  [1980] = 829,
  [1981] = 807,
  [1982] = 799,
  [1983] = 810,
  [1984] = 815,
  [1985] = 818,
  [1986] = 812,
  [1987] = 813,
  [1988] = 760,
  [1989] = 805,
  [1990] = 806,
  [1991] = 808,
  [1992] = 816,
  [1993] = 1659,
  [1994] = 1994,
  [1995] = 1995,
  [1996] = 1996,
  [1997] = 1997,
  [1998] = 829,
  [1999] = 800,
  [2000] = 2000,
  [2001] = 2001,
  [2002] = 1648,
  [2003] = 823,
  [2004] = 1667,
  [2005] = 762,
  [2006] = 767,
  [2007] = 1661,
  [2008] = 796,
  [2009] = 2009,
  [2010] = 1664,
  [2011] = 771,
  [2012] = 814,
  [2013] = 2013,
  [2014] = 1666,
  [2015] = 1670,
  [2016] = 1665,
  [2017] = 1651,
  [2018] = 1653,
  [2019] = 1668,
  [2020] = 786,
  [2021] = 787,
  [2022] = 1648,
  [2023] = 788,
  [2024] = 2024,
  [2025] = 774,
  [2026] = 799,
  [2027] = 2027,
  [2028] = 1669,
  [2029] = 1812,
  [2030] = 1813,
  [2031] = 831,
  [2032] = 797,
  [2033] = 2033,
  [2034] = 827,
  [2035] = 830,
  [2036] = 2036,
  [2037] = 2037,
  [2038] = 804,
  [2039] = 751,
  [2040] = 763,
  [2041] = 803,
  [2042] = 2042,
  [2043] = 2043,
  [2044] = 764,
  [2045] = 1671,
  [2046] = 765,
  [2047] = 1647,
  [2048] = 2048,
  [2049] = 2049,
  [2050] = 768,
  [2051] = 2051,
  [2052] = 1872,
  [2053] = 769,
  [2054] = 770,
  [2055] = 783,
  [2056] = 2056,
  [2057] = 2057,
  [2058] = 809,
  [2059] = 2059,
  [2060] = 2060,
  [2061] = 2061,
  [2062] = 2062,
  [2063] = 2063,
  [2064] = 2064,
  [2065] = 798,
  [2066] = 811,
  [2067] = 776,
  [2068] = 2068,
  [2069] = 2069,
  [2070] = 777,
  [2071] = 2071,
  [2072] = 2072,
  [2073] = 820,
  [2074] = 1657,
  [2075] = 778,
  [2076] = 2076,
  [2077] = 779,
  [2078] = 810,
  [2079] = 2079,
  [2080] = 815,
  [2081] = 758,
  [2082] = 780,
  [2083] = 791,
  [2084] = 759,
  [2085] = 2085,
  [2086] = 2086,
  [2087] = 2087,
  [2088] = 1669,
  [2089] = 2089,
  [2090] = 802,
  [2091] = 781,
  [2092] = 753,
  [2093] = 1840,
  [2094] = 2094,
  [2095] = 2095,
  [2096] = 2096,
  [2097] = 2097,
  [2098] = 2098,
  [2099] = 2099,
  [2100] = 2100,
  [2101] = 2101,
  [2102] = 2101,
  [2103] = 2103,
  [2104] = 2104,
  [2105] = 2105,
  [2106] = 2096,
  [2107] = 2107,
  [2108] = 2108,
  [2109] = 2109,
  [2110] = 2101,
  [2111] = 2096,
  [2112] = 2100,
  [2113] = 2113,
  [2114] = 2114,
  [2115] = 2115,
  [2116] = 2116,
  [2117] = 2117,
  [2118] = 2105,
  [2119] = 2117,
  [2120] = 1669,
  [2121] = 2121,
  [2122] = 2122,
  [2123] = 2123,
  [2124] = 2124,
  [2125] = 2125,
  [2126] = 2126,
  [2127] = 2125,
  [2128] = 2128,
  [2129] = 2129,
  [2130] = 2130,
  [2131] = 2131,
  [2132] = 2132,
  [2133] = 2133,
  [2134] = 2130,
  [2135] = 2129,
  [2136] = 2126,
  [2137] = 2137,
  [2138] = 2138,
  [2139] = 2139,
  [2140] = 2140,
  [2141] = 2141,
  [2142] = 2142,
  [2143] = 2143,
  [2144] = 2144,
  [2145] = 2122,
  [2146] = 1659,
  [2147] = 2147,
  [2148] = 2148,
  [2149] = 2149,
  [2150] = 2150,
  [2151] = 2133,
  [2152] = 2115,
  [2153] = 2121,
  [2154] = 2140,
  [2155] = 2155,
  [2156] = 2138,
  [2157] = 2148,
  [2158] = 2109,
  [2159] = 2139,
  [2160] = 2132,
  [2161] = 2108,
  [2162] = 2131,
  [2163] = 2124,
  [2164] = 2113,
  [2165] = 2147,
  [2166] = 2095,
  [2167] = 2167,
  [2168] = 2144,
  [2169] = 2143,
  [2170] = 2142,
  [2171] = 2171,
  [2172] = 2141,
  [2173] = 2098,
  [2174] = 2099,
  [2175] = 2171,
  [2176] = 2100,
  [2177] = 2149,
  [2178] = 1648,
  [2179] = 2128,
  [2180] = 2123,
  [2181] = 2181,
  [2182] = 822,
  [2183] = 2137,
  [2184] = 2122,
  [2185] = 2095,
  [2186] = 824,
  [2187] = 825,
  [2188] = 826,
  [2189] = 2132,
  [2190] = 2190,
  [2191] = 2191,
  [2192] = 2192,
  [2193] = 2056,
  [2194] = 2194,
  [2195] = 2194,
  [2196] = 2181,
  [2197] = 2141,
  [2198] = 2194,
  [2199] = 2199,
  [2200] = 2200,
  [2201] = 2109,
  [2202] = 1474,
  [2203] = 2203,
  [2204] = 2133,
  [2205] = 2205,
  [2206] = 2130,
  [2207] = 2207,
  [2208] = 2194,
  [2209] = 2125,
  [2210] = 1969,
  [2211] = 2143,
  [2212] = 1823,
  [2213] = 1994,
  [2214] = 2129,
  [2215] = 814,
  [2216] = 2190,
  [2217] = 2181,
  [2218] = 817,
  [2219] = 1486,
  [2220] = 819,
  [2221] = 807,
  [2222] = 2121,
  [2223] = 1480,
  [2224] = 2224,
  [2225] = 2225,
  [2226] = 2128,
  [2227] = 2148,
  [2228] = 2123,
  [2229] = 2171,
  [2230] = 2230,
  [2231] = 2200,
  [2232] = 2232,
  [2233] = 2140,
  [2234] = 2234,
  [2235] = 2235,
  [2236] = 2013,
  [2237] = 2108,
  [2238] = 2238,
  [2239] = 2239,
  [2240] = 2240,
  [2241] = 2194,
  [2242] = 2242,
  [2243] = 2243,
  [2244] = 821,
  [2245] = 2116,
  [2246] = 2200,
  [2247] = 2192,
  [2248] = 2248,
  [2249] = 2205,
  [2250] = 2250,
  [2251] = 2251,
  [2252] = 2194,
  [2253] = 2253,
  [2254] = 799,
  [2255] = 1995,
  [2256] = 2190,
  [2257] = 810,
  [2258] = 815,
  [2259] = 1829,
  [2260] = 789,
  [2261] = 2124,
  [2262] = 2126,
  [2263] = 790,
  [2264] = 2264,
  [2265] = 2149,
  [2266] = 2147,
  [2267] = 2200,
  [2268] = 818,
  [2269] = 2115,
  [2270] = 2117,
  [2271] = 2131,
  [2272] = 2098,
  [2273] = 2150,
  [2274] = 2144,
  [2275] = 2142,
  [2276] = 2099,
  [2277] = 2155,
  [2278] = 2167,
  [2279] = 2192,
  [2280] = 795,
  [2281] = 2105,
  [2282] = 2113,
  [2283] = 2181,
  [2284] = 2203,
  [2285] = 2285,
  [2286] = 2286,
  [2287] = 2287,
  [2288] = 798,
  [2289] = 2289,
  [2290] = 2290,
  [2291] = 2205,
  [2292] = 2292,
  [2293] = 801,
  [2294] = 2138,
  [2295] = 2139,
  [2296] = 2296,
  [2297] = 2297,
  [2298] = 1943,
  [2299] = 1945,
  [2300] = 2181,
  [2301] = 2203,
  [2302] = 2098,
  [2303] = 2303,
  [2304] = 2099,
  [2305] = 2108,
  [2306] = 2105,
  [2307] = 2303,
  [2308] = 2113,
  [2309] = 2303,
  [2310] = 2095,
  [2311] = 2137,
  [2312] = 2312,
  [2313] = 2312,
  [2314] = 2109,
  [2315] = 2312,
  [2316] = 2137,
  [2317] = 784,
  [2318] = 781,
  [2319] = 774,
  [2320] = 783,
  [2321] = 794,
  [2322] = 2013,
  [2323] = 799,
  [2324] = 758,
  [2325] = 759,
  [2326] = 753,
  [2327] = 789,
  [2328] = 809,
  [2329] = 811,
  [2330] = 785,
  [2331] = 782,
  [2332] = 814,
  [2333] = 2104,
  [2334] = 817,
  [2335] = 757,
  [2336] = 756,
  [2337] = 754,
  [2338] = 819,
  [2339] = 803,
  [2340] = 790,
  [2341] = 2103,
  [2342] = 2097,
  [2343] = 801,
  [2344] = 2344,
  [2345] = 791,
  [2346] = 818,
  [2347] = 802,
  [2348] = 800,
  [2349] = 823,
  [2350] = 761,
  [2351] = 761,
  [2352] = 762,
  [2353] = 810,
  [2354] = 767,
  [2355] = 807,
  [2356] = 792,
  [2357] = 771,
  [2358] = 828,
  [2359] = 772,
  [2360] = 824,
  [2361] = 795,
  [2362] = 810,
  [2363] = 825,
  [2364] = 775,
  [2365] = 793,
  [2366] = 766,
  [2367] = 815,
  [2368] = 812,
  [2369] = 796,
  [2370] = 786,
  [2371] = 773,
  [2372] = 787,
  [2373] = 780,
  [2374] = 788,
  [2375] = 822,
  [2376] = 813,
  [2377] = 831,
  [2378] = 797,
  [2379] = 760,
  [2380] = 826,
  [2381] = 827,
  [2382] = 782,
  [2383] = 792,
  [2384] = 798,
  [2385] = 773,
  [2386] = 812,
  [2387] = 813,
  [2388] = 760,
  [2389] = 829,
  [2390] = 751,
  [2391] = 763,
  [2392] = 805,
  [2393] = 806,
  [2394] = 808,
  [2395] = 816,
  [2396] = 764,
  [2397] = 793,
  [2398] = 805,
  [2399] = 752,
  [2400] = 830,
  [2401] = 804,
  [2402] = 765,
  [2403] = 806,
  [2404] = 808,
  [2405] = 802,
  [2406] = 768,
  [2407] = 769,
  [2408] = 770,
  [2409] = 755,
  [2410] = 820,
  [2411] = 815,
  [2412] = 775,
  [2413] = 776,
  [2414] = 2107,
  [2415] = 816,
  [2416] = 777,
  [2417] = 818,
  [2418] = 828,
  [2419] = 821,
  [2420] = 799,
  [2421] = 778,
  [2422] = 779,
  [2423] = 2114,
  [2424] = 2248,
  [2425] = 1480,
  [2426] = 1823,
  [2427] = 2056,
  [2428] = 2287,
  [2429] = 822,
  [2430] = 824,
  [2431] = 2235,
  [2432] = 2289,
  [2433] = 2292,
  [2434] = 795,
  [2435] = 1994,
  [2436] = 825,
  [2437] = 2232,
  [2438] = 2264,
  [2439] = 2239,
  [2440] = 2013,
  [2441] = 2441,
  [2442] = 2240,
  [2443] = 2253,
  [2444] = 2242,
  [2445] = 774,
  [2446] = 2243,
  [2447] = 1995,
  [2448] = 826,
  [2449] = 2250,
  [2450] = 2094,
  [2451] = 814,
  [2452] = 1829,
  [2453] = 2297,
  [2454] = 2230,
  [2455] = 819,
  [2456] = 2234,
  [2457] = 799,
  [2458] = 820,
  [2459] = 798,
  [2460] = 810,
  [2461] = 807,
  [2462] = 2290,
  [2463] = 2224,
  [2464] = 821,
  [2465] = 815,
  [2466] = 2251,
  [2467] = 2296,
  [2468] = 1486,
  [2469] = 2238,
  [2470] = 817,
  [2471] = 2199,
  [2472] = 2285,
  [2473] = 2207,
  [2474] = 2286,
  [2475] = 1969,
  [2476] = 796,
  [2477] = 2225,
  [2478] = 818,
  [2479] = 2191,
  [2480] = 790,
  [2481] = 1943,
  [2482] = 1945,
  [2483] = 789,
  [2484] = 2441,
  [2485] = 2441,
  [2486] = 829,
  [2487] = 1474,
  [2488] = 801,
  [2489] = 1994,
  [2490] = 839,
  [2491] = 2491,
  [2492] = 2492,
  [2493] = 2491,
  [2494] = 2494,
  [2495] = 2495,
  [2496] = 2491,
  [2497] = 2492,
  [2498] = 2498,
  [2499] = 2492,
  [2500] = 1648,
  [2501] = 847,
  [2502] = 2502,
  [2503] = 2503,
  [2504] = 2504,
  [2505] = 2504,
  [2506] = 2506,
  [2507] = 1659,
  [2508] = 1669,
  [2509] = 2509,
  [2510] = 1648,
  [2511] = 1651,
  [2512] = 2512,
  [2513] = 2513,
  [2514] = 1647,
  [2515] = 1657,
  [2516] = 2504,
  [2517] = 1669,
  [2518] = 1652,
  [2519] = 1666,
  [2520] = 2502,
  [2521] = 2504,
  [2522] = 2513,
  [2523] = 1670,
  [2524] = 1659,
  [2525] = 2525,
  [2526] = 2526,
  [2527] = 1657,
  [2528] = 1653,
  [2529] = 1660,
  [2530] = 2506,
  [2531] = 1667,
  [2532] = 1661,
  [2533] = 1664,
  [2534] = 1666,
  [2535] = 1670,
  [2536] = 2526,
  [2537] = 2512,
  [2538] = 1665,
  [2539] = 1651,
  [2540] = 1653,
  [2541] = 2541,
  [2542] = 2504,
  [2543] = 1668,
  [2544] = 1671,
  [2545] = 1647,
  [2546] = 2526,
  [2547] = 2547,
  [2548] = 2502,
  [2549] = 1652,
  [2550] = 2506,
  [2551] = 2513,
  [2552] = 845,
  [2553] = 2513,
  [2554] = 1667,
  [2555] = 2344,
  [2556] = 2506,
  [2557] = 793,
  [2558] = 1671,
  [2559] = 1665,
  [2560] = 1661,
  [2561] = 2513,
  [2562] = 2513,
  [2563] = 2525,
  [2564] = 1664,
  [2565] = 828,
  [2566] = 1660,
  [2567] = 1668,
  [2568] = 2512,
  [2569] = 2525,
  [2570] = 2570,
  [2571] = 2571,
  [2572] = 2572,
  [2573] = 2573,
  [2574] = 2574,
  [2575] = 2575,
  [2576] = 2576,
  [2577] = 2577,
  [2578] = 2578,
  [2579] = 2579,
  [2580] = 2580,
  [2581] = 2581,
  [2582] = 2582,
  [2583] = 2583,
  [2584] = 2584,
  [2585] = 2585,
  [2586] = 2586,
  [2587] = 2587,
  [2588] = 2588,
  [2589] = 2589,
  [2590] = 2590,
  [2591] = 2591,
  [2592] = 2592,
  [2593] = 2593,
  [2594] = 2594,
  [2595] = 2595,
  [2596] = 2596,
  [2597] = 796,
  [2598] = 820,
  [2599] = 829,
  [2600] = 2600,
  [2601] = 774,
  [2602] = 2602,
  [2603] = 2603,
  [2604] = 2604,
  [2605] = 2605,
  [2606] = 2587,
  [2607] = 2596,
  [2608] = 2608,
  [2609] = 2609,
  [2610] = 2602,
  [2611] = 2611,
  [2612] = 2612,
  [2613] = 2613,
  [2614] = 2614,
  [2615] = 2588,
  [2616] = 2594,
  [2617] = 2612,
  [2618] = 2618,
  [2619] = 1659,
  [2620] = 1669,
  [2621] = 2587,
  [2622] = 2596,
  [2623] = 2602,
  [2624] = 2588,
  [2625] = 2594,
  [2626] = 2612,
  [2627] = 2587,
  [2628] = 2602,
  [2629] = 2629,
  [2630] = 1648,
  [2631] = 2596,
  [2632] = 2632,
  [2633] = 2633,
  [2634] = 2634,
  [2635] = 2344,
  [2636] = 2636,
  [2637] = 2637,
  [2638] = 2638,
  [2639] = 2639,
  [2640] = 1648,
  [2641] = 2641,
  [2642] = 2642,
  [2643] = 2643,
  [2644] = 2644,
  [2645] = 2645,
  [2646] = 2646,
  [2647] = 1659,
  [2648] = 2648,
  [2649] = 1669,
  [2650] = 2650,
  [2651] = 2651,
  [2652] = 2652,
  [2653] = 815,
  [2654] = 802,
  [2655] = 761,
  [2656] = 755,
  [2657] = 792,
  [2658] = 766,
  [2659] = 773,
  [2660] = 775,
  [2661] = 752,
  [2662] = 782,
  [2663] = 812,
  [2664] = 813,
  [2665] = 760,
  [2666] = 805,
  [2667] = 806,
  [2668] = 808,
  [2669] = 816,
  [2670] = 1515,
  [2671] = 1510,
  [2672] = 2672,
  [2673] = 1812,
  [2674] = 793,
  [2675] = 830,
  [2676] = 828,
  [2677] = 781,
  [2678] = 804,
  [2679] = 798,
  [2680] = 824,
  [2681] = 789,
  [2682] = 790,
  [2683] = 795,
  [2684] = 821,
  [2685] = 825,
  [2686] = 814,
  [2687] = 817,
  [2688] = 819,
  [2689] = 807,
  [2690] = 774,
  [2691] = 801,
  [2692] = 822,
  [2693] = 826,
  [2694] = 793,
  [2695] = 796,
  [2696] = 820,
  [2697] = 828,
  [2698] = 829,
  [2699] = 799,
  [2700] = 810,
  [2701] = 2701,
  [2702] = 818,
  [2703] = 2703,
  [2704] = 1813,
  [2705] = 2705,
  [2706] = 2706,
  [2707] = 2705,
  [2708] = 783,
  [2709] = 780,
  [2710] = 1841,
  [2711] = 1872,
  [2712] = 758,
  [2713] = 794,
  [2714] = 759,
  [2715] = 753,
  [2716] = 809,
  [2717] = 811,
  [2718] = 2703,
  [2719] = 785,
  [2720] = 800,
  [2721] = 823,
  [2722] = 2108,
  [2723] = 762,
  [2724] = 767,
  [2725] = 1521,
  [2726] = 2726,
  [2727] = 771,
  [2728] = 757,
  [2729] = 786,
  [2730] = 2703,
  [2731] = 787,
  [2732] = 2701,
  [2733] = 788,
  [2734] = 2734,
  [2735] = 2705,
  [2736] = 2706,
  [2737] = 831,
  [2738] = 797,
  [2739] = 2109,
  [2740] = 756,
  [2741] = 827,
  [2742] = 751,
  [2743] = 763,
  [2744] = 764,
  [2745] = 765,
  [2746] = 754,
  [2747] = 2701,
  [2748] = 2109,
  [2749] = 768,
  [2750] = 769,
  [2751] = 2705,
  [2752] = 770,
  [2753] = 776,
  [2754] = 2706,
  [2755] = 777,
  [2756] = 778,
  [2757] = 2108,
  [2758] = 803,
  [2759] = 2706,
  [2760] = 779,
  [2761] = 791,
  [2762] = 1840,
  [2763] = 2763,
  [2764] = 2764,
  [2765] = 2765,
  [2766] = 2763,
  [2767] = 2767,
  [2768] = 796,
  [2769] = 829,
  [2770] = 774,
  [2771] = 2771,
  [2772] = 820,
  [2773] = 2773,
  [2774] = 2771,
  [2775] = 2775,
  [2776] = 2776,
  [2777] = 1538,
  [2778] = 2778,
  [2779] = 2779,
  [2780] = 2771,
  [2781] = 2775,
  [2782] = 2776,
  [2783] = 2765,
  [2784] = 2778,
  [2785] = 2779,
  [2786] = 2771,
  [2787] = 2765,
  [2788] = 2775,
  [2789] = 2764,
  [2790] = 2776,
  [2791] = 2791,
  [2792] = 2792,
  [2793] = 1659,
  [2794] = 1669,
  [2795] = 2765,
  [2796] = 1657,
  [2797] = 1665,
  [2798] = 1651,
  [2799] = 1653,
  [2800] = 1648,
  [2801] = 2801,
  [2802] = 2801,
  [2803] = 2803,
  [2804] = 2775,
  [2805] = 2805,
  [2806] = 2776,
  [2807] = 2779,
  [2808] = 2803,
  [2809] = 2809,
  [2810] = 2771,
  [2811] = 2775,
  [2812] = 2776,
  [2813] = 2765,
  [2814] = 2801,
  [2815] = 2803,
  [2816] = 2809,
  [2817] = 2763,
  [2818] = 2778,
  [2819] = 2779,
  [2820] = 1660,
  [2821] = 1661,
  [2822] = 1664,
  [2823] = 1666,
  [2824] = 1670,
  [2825] = 1668,
  [2826] = 1671,
  [2827] = 1647,
  [2828] = 1652,
  [2829] = 1667,
  [2830] = 2764,
  [2831] = 2771,
  [2832] = 2832,
  [2833] = 2765,
  [2834] = 2809,
  [2835] = 2764,
  [2836] = 2137,
  [2837] = 2809,
  [2838] = 2838,
  [2839] = 2838,
  [2840] = 2767,
  [2841] = 2778,
  [2842] = 2838,
  [2843] = 2137,
  [2844] = 1667,
  [2845] = 2494,
  [2846] = 2100,
  [2847] = 1657,
  [2848] = 1659,
  [2849] = 1669,
  [2850] = 1660,
  [2851] = 1661,
  [2852] = 1664,
  [2853] = 1666,
  [2854] = 1670,
  [2855] = 1665,
  [2856] = 1651,
  [2857] = 1653,
  [2858] = 1668,
  [2859] = 1671,
  [2860] = 1647,
  [2861] = 1652,
  [2862] = 1648,
  [2863] = 2636,
  [2864] = 1521,
  [2865] = 1661,
  [2866] = 1664,
  [2867] = 1666,
  [2868] = 1670,
  [2869] = 1668,
  [2870] = 1671,
  [2871] = 1647,
  [2872] = 1652,
  [2873] = 1667,
  [2874] = 2498,
  [2875] = 1510,
  [2876] = 1648,
  [2877] = 1657,
  [2878] = 1665,
  [2879] = 1651,
  [2880] = 1653,
  [2881] = 1659,
  [2882] = 1669,
  [2883] = 1515,
  [2884] = 1660,
  [2885] = 2140,
  [2886] = 2115,
  [2887] = 1652,
  [2888] = 2138,
  [2889] = 2139,
  [2890] = 2150,
  [2891] = 1669,
  [2892] = 1486,
  [2893] = 2155,
  [2894] = 2167,
  [2895] = 2122,
  [2896] = 1480,
  [2897] = 1667,
  [2898] = 1657,
  [2899] = 1648,
  [2900] = 1665,
  [2901] = 1651,
  [2902] = 1653,
  [2903] = 2123,
  [2904] = 1661,
  [2905] = 2129,
  [2906] = 2124,
  [2907] = 2149,
  [2908] = 2121,
  [2909] = 2909,
  [2910] = 1670,
  [2911] = 2171,
  [2912] = 1648,
  [2913] = 2125,
  [2914] = 2141,
  [2915] = 1647,
  [2916] = 2133,
  [2917] = 2917,
  [2918] = 2117,
  [2919] = 1668,
  [2920] = 2131,
  [2921] = 1664,
  [2922] = 2116,
  [2923] = 2142,
  [2924] = 1825,
  [2925] = 1826,
  [2926] = 2132,
  [2927] = 789,
  [2928] = 790,
  [2929] = 795,
  [2930] = 2130,
  [2931] = 2909,
  [2932] = 1660,
  [2933] = 2495,
  [2934] = 801,
  [2935] = 2148,
  [2936] = 2143,
  [2937] = 1659,
  [2938] = 2938,
  [2939] = 1474,
  [2940] = 1823,
  [2941] = 822,
  [2942] = 825,
  [2943] = 826,
  [2944] = 1671,
  [2945] = 1666,
  [2946] = 1659,
  [2947] = 2147,
  [2948] = 2909,
  [2949] = 1829,
  [2950] = 2909,
  [2951] = 1538,
  [2952] = 2909,
  [2953] = 2128,
  [2954] = 1669,
  [2955] = 2126,
  [2956] = 2144,
  [2957] = 1994,
  [2958] = 2636,
  [2959] = 2959,
  [2960] = 1659,
  [2961] = 2060,
  [2962] = 2061,
  [2963] = 828,
  [2964] = 2062,
  [2965] = 1925,
  [2966] = 1952,
  [2967] = 2036,
  [2968] = 2013,
  [2969] = 1926,
  [2970] = 1931,
  [2971] = 2584,
  [2972] = 1922,
  [2973] = 2063,
  [2974] = 2064,
  [2975] = 2975,
  [2976] = 2086,
  [2977] = 1934,
  [2978] = 1970,
  [2979] = 1997,
  [2980] = 2980,
  [2981] = 2604,
  [2982] = 2605,
  [2983] = 1648,
  [2984] = 2585,
  [2985] = 2603,
  [2986] = 820,
  [2987] = 2076,
  [2988] = 2589,
  [2989] = 2571,
  [2990] = 799,
  [2991] = 2991,
  [2992] = 810,
  [2993] = 815,
  [2994] = 2608,
  [2995] = 2590,
  [2996] = 818,
  [2997] = 2997,
  [2998] = 2001,
  [2999] = 2609,
  [3000] = 2570,
  [3001] = 1669,
  [3002] = 1975,
  [3003] = 2613,
  [3004] = 2577,
  [3005] = 2049,
  [3006] = 2009,
  [3007] = 2027,
  [3008] = 2033,
  [3009] = 2042,
  [3010] = 2079,
  [3011] = 1954,
  [3012] = 2000,
  [3013] = 1996,
  [3014] = 2578,
  [3015] = 2057,
  [3016] = 2579,
  [3017] = 2503,
  [3018] = 1648,
  [3019] = 2089,
  [3020] = 2618,
  [3021] = 804,
  [3022] = 798,
  [3023] = 2051,
  [3024] = 2085,
  [3025] = 3025,
  [3026] = 2024,
  [3027] = 2059,
  [3028] = 2068,
  [3029] = 1994,
  [3030] = 2069,
  [3031] = 2586,
  [3032] = 2071,
  [3033] = 1657,
  [3034] = 1660,
  [3035] = 830,
  [3036] = 1661,
  [3037] = 1664,
  [3038] = 1666,
  [3039] = 1670,
  [3040] = 1665,
  [3041] = 2980,
  [3042] = 1653,
  [3043] = 1668,
  [3044] = 1671,
  [3045] = 1647,
  [3046] = 1652,
  [3047] = 2013,
  [3048] = 1667,
  [3049] = 2072,
  [3050] = 821,
  [3051] = 2087,
  [3052] = 824,
  [3053] = 2037,
  [3054] = 2043,
  [3055] = 2573,
  [3056] = 814,
  [3057] = 817,
  [3058] = 819,
  [3059] = 807,
  [3060] = 3060,
  [3061] = 1660,
  [3062] = 1659,
  [3063] = 1661,
  [3064] = 1664,
  [3065] = 1666,
  [3066] = 1670,
  [3067] = 1668,
  [3068] = 1671,
  [3069] = 1647,
  [3070] = 1652,
  [3071] = 1669,
  [3072] = 1667,
  [3073] = 1659,
  [3074] = 1669,
  [3075] = 1648,
  [3076] = 799,
  [3077] = 2048,
  [3078] = 2580,
  [3079] = 3079,
  [3080] = 1995,
  [3081] = 3081,
  [3082] = 2576,
  [3083] = 3083,
  [3084] = 810,
  [3085] = 815,
  [3086] = 3086,
  [3087] = 2593,
  [3088] = 2582,
  [3089] = 3089,
  [3090] = 2595,
  [3091] = 2991,
  [3092] = 2611,
  [3093] = 818,
  [3094] = 2629,
  [3095] = 3095,
  [3096] = 2574,
  [3097] = 3097,
  [3098] = 2632,
  [3099] = 1927,
  [3100] = 1969,
  [3101] = 1657,
  [3102] = 1665,
  [3103] = 1651,
  [3104] = 1653,
  [3105] = 793,
  [3106] = 1929,
  [3107] = 1930,
  [3108] = 1932,
  [3109] = 2583,
  [3110] = 1933,
  [3111] = 1943,
  [3112] = 1945,
  [3113] = 2056,
  [3114] = 1651,
  [3115] = 1671,
  [3116] = 2109,
  [3117] = 2098,
  [3118] = 2099,
  [3119] = 829,
  [3120] = 3120,
  [3121] = 3121,
  [3122] = 3122,
  [3123] = 3123,
  [3124] = 3124,
  [3125] = 3125,
  [3126] = 3126,
  [3127] = 3127,
  [3128] = 799,
  [3129] = 3123,
  [3130] = 2095,
  [3131] = 3131,
  [3132] = 3132,
  [3133] = 2105,
  [3134] = 2113,
  [3135] = 810,
  [3136] = 3136,
  [3137] = 815,
  [3138] = 3138,
  [3139] = 3139,
  [3140] = 3140,
  [3141] = 1657,
  [3142] = 1660,
  [3143] = 3143,
  [3144] = 1664,
  [3145] = 1666,
  [3146] = 1670,
  [3147] = 1665,
  [3148] = 1651,
  [3149] = 1653,
  [3150] = 1668,
  [3151] = 796,
  [3152] = 1647,
  [3153] = 1652,
  [3154] = 1667,
  [3155] = 3123,
  [3156] = 3156,
  [3157] = 3157,
  [3158] = 3158,
  [3159] = 1659,
  [3160] = 1669,
  [3161] = 1648,
  [3162] = 3162,
  [3163] = 3163,
  [3164] = 3164,
  [3165] = 3132,
  [3166] = 818,
  [3167] = 3167,
  [3168] = 2108,
  [3169] = 3169,
  [3170] = 3170,
  [3171] = 1661,
  [3172] = 3172,
  [3173] = 3173,
  [3174] = 3174,
  [3175] = 3175,
  [3176] = 3176,
  [3177] = 3177,
  [3178] = 3178,
  [3179] = 3179,
  [3180] = 3180,
  [3181] = 3125,
  [3182] = 3182,
  [3183] = 3183,
  [3184] = 2109,
  [3185] = 2108,
  [3186] = 2109,
  [3187] = 2098,
  [3188] = 2099,
  [3189] = 2105,
  [3190] = 2113,
  [3191] = 3173,
  [3192] = 1648,
  [3193] = 793,
  [3194] = 3194,
  [3195] = 3174,
  [3196] = 3178,
  [3197] = 3197,
  [3198] = 2137,
  [3199] = 799,
  [3200] = 3173,
  [3201] = 810,
  [3202] = 815,
  [3203] = 2108,
  [3204] = 3174,
  [3205] = 3178,
  [3206] = 3206,
  [3207] = 3194,
  [3208] = 3194,
  [3209] = 2917,
  [3210] = 2938,
  [3211] = 3211,
  [3212] = 1659,
  [3213] = 3182,
  [3214] = 1659,
  [3215] = 1669,
  [3216] = 828,
  [3217] = 1669,
  [3218] = 2095,
  [3219] = 3219,
  [3220] = 818,
  [3221] = 3211,
  [3222] = 1648,
  [3223] = 3182,
  [3224] = 799,
  [3225] = 3225,
  [3226] = 810,
  [3227] = 815,
  [3228] = 3228,
  [3229] = 796,
  [3230] = 774,
  [3231] = 818,
  [3232] = 820,
  [3233] = 2498,
  [3234] = 3234,
  [3235] = 3235,
  [3236] = 2108,
  [3237] = 3234,
  [3238] = 3235,
  [3239] = 3239,
  [3240] = 3240,
  [3241] = 3225,
  [3242] = 3225,
  [3243] = 3234,
  [3244] = 3225,
  [3245] = 3225,
  [3246] = 3225,
  [3247] = 3247,
  [3248] = 3248,
  [3249] = 3170,
  [3250] = 3234,
  [3251] = 3086,
  [3252] = 3252,
  [3253] = 3235,
  [3254] = 3254,
  [3255] = 2137,
  [3256] = 3256,
  [3257] = 3235,
  [3258] = 3258,
  [3259] = 2137,
  [3260] = 3260,
  [3261] = 3156,
  [3262] = 3262,
  [3263] = 3263,
  [3264] = 3264,
  [3265] = 829,
  [3266] = 3234,
  [3267] = 2109,
  [3268] = 3268,
  [3269] = 3234,
  [3270] = 3235,
  [3271] = 3235,
  [3272] = 2098,
  [3273] = 2105,
  [3274] = 3274,
  [3275] = 3275,
  [3276] = 3276,
  [3277] = 2095,
  [3278] = 2105,
  [3279] = 2113,
  [3280] = 3280,
  [3281] = 3281,
  [3282] = 2109,
  [3283] = 2099,
  [3284] = 2113,
  [3285] = 3285,
  [3286] = 3286,
  [3287] = 3287,
  [3288] = 3288,
  [3289] = 3289,
  [3290] = 3290,
  [3291] = 3291,
  [3292] = 3292,
  [3293] = 2109,
  [3294] = 3294,
  [3295] = 3295,
  [3296] = 3125,
  [3297] = 2095,
  [3298] = 814,
  [3299] = 817,
  [3300] = 3300,
  [3301] = 3264,
  [3302] = 819,
  [3303] = 3303,
  [3304] = 807,
  [3305] = 2137,
  [3306] = 2098,
  [3307] = 2108,
  [3308] = 2099,
  [3309] = 3309,
  [3310] = 3310,
  [3311] = 2108,
  [3312] = 821,
  [3313] = 3313,
  [3314] = 3314,
  [3315] = 3176,
  [3316] = 3316,
  [3317] = 3317,
  [3318] = 3318,
  [3319] = 3319,
  [3320] = 3320,
  [3321] = 3321,
  [3322] = 3322,
  [3323] = 3323,
  [3324] = 3324,
  [3325] = 3325,
  [3326] = 3326,
  [3327] = 3327,
  [3328] = 3328,
  [3329] = 3300,
  [3330] = 3314,
  [3331] = 810,
  [3332] = 3319,
  [3333] = 815,
  [3334] = 3314,
  [3335] = 3335,
  [3336] = 3336,
  [3337] = 3337,
  [3338] = 3338,
  [3339] = 3339,
  [3340] = 3321,
  [3341] = 3341,
  [3342] = 3322,
  [3343] = 3343,
  [3344] = 3344,
  [3345] = 3323,
  [3346] = 3346,
  [3347] = 3347,
  [3348] = 2137,
  [3349] = 3349,
  [3350] = 3350,
  [3351] = 2116,
  [3352] = 3352,
  [3353] = 3353,
  [3354] = 3354,
  [3355] = 3320,
  [3356] = 3356,
  [3357] = 3357,
  [3358] = 3358,
  [3359] = 3359,
  [3360] = 3337,
  [3361] = 3338,
  [3362] = 3356,
  [3363] = 3339,
  [3364] = 3364,
  [3365] = 3364,
  [3366] = 3366,
  [3367] = 3367,
  [3368] = 3172,
  [3369] = 3324,
  [3370] = 3337,
  [3371] = 3338,
  [3372] = 3341,
  [3373] = 3343,
  [3374] = 3353,
  [3375] = 3375,
  [3376] = 3324,
  [3377] = 3325,
  [3378] = 3366,
  [3379] = 3325,
  [3380] = 3316,
  [3381] = 3344,
  [3382] = 3382,
  [3383] = 3339,
  [3384] = 3354,
  [3385] = 3320,
  [3386] = 3343,
  [3387] = 3344,
  [3388] = 3327,
  [3389] = 3389,
  [3390] = 3341,
  [3391] = 3391,
  [3392] = 3375,
  [3393] = 3393,
  [3394] = 3353,
  [3395] = 3382,
  [3396] = 3396,
  [3397] = 3391,
  [3398] = 3398,
  [3399] = 3317,
  [3400] = 3398,
  [3401] = 3319,
  [3402] = 3321,
  [3403] = 3322,
  [3404] = 3323,
  [3405] = 3356,
  [3406] = 3314,
  [3407] = 3364,
  [3408] = 3366,
  [3409] = 3375,
  [3410] = 818,
  [3411] = 2150,
  [3412] = 2137,
  [3413] = 2155,
  [3414] = 2167,
  [3415] = 3316,
  [3416] = 2109,
  [3417] = 3417,
  [3418] = 3336,
  [3419] = 3179,
  [3420] = 3382,
  [3421] = 3354,
  [3422] = 3328,
  [3423] = 3328,
  [3424] = 799,
  [3425] = 3391,
  [3426] = 3353,
  [3427] = 3314,
  [3428] = 3428,
  [3429] = 3398,
  [3430] = 3430,
  [3431] = 3354,
  [3432] = 3317,
  [3433] = 3197,
  [3434] = 3320,
  [3435] = 2108,
  [3436] = 3336,
  [3437] = 3437,
  [3438] = 3438,
  [3439] = 3439,
  [3440] = 3440,
  [3441] = 3441,
  [3442] = 2137,
  [3443] = 3443,
  [3444] = 3444,
  [3445] = 3228,
  [3446] = 3446,
  [3447] = 3447,
  [3448] = 3448,
  [3449] = 3449,
  [3450] = 3450,
  [3451] = 3451,
  [3452] = 3452,
  [3453] = 3453,
  [3454] = 3454,
  [3455] = 3455,
  [3456] = 3456,
  [3457] = 3240,
  [3458] = 3286,
  [3459] = 3459,
  [3460] = 3460,
  [3461] = 3347,
  [3462] = 3462,
  [3463] = 3463,
  [3464] = 3464,
  [3465] = 3465,
  [3466] = 3466,
  [3467] = 3467,
  [3468] = 3468,
  [3469] = 3469,
  [3470] = 3470,
  [3471] = 3471,
  [3472] = 3472,
  [3473] = 3326,
  [3474] = 3474,
  [3475] = 3280,
  [3476] = 3254,
  [3477] = 3477,
  [3478] = 3256,
  [3479] = 3258,
  [3480] = 3480,
  [3481] = 3481,
  [3482] = 3482,
  [3483] = 3275,
  [3484] = 3484,
  [3485] = 3485,
  [3486] = 3486,
  [3487] = 3487,
  [3488] = 3488,
  [3489] = 3489,
  [3490] = 3490,
  [3491] = 3491,
  [3492] = 3492,
  [3493] = 3493,
  [3494] = 3494,
  [3495] = 3495,
  [3496] = 804,
  [3497] = 3497,
  [3498] = 3498,
  [3499] = 3499,
  [3500] = 3500,
  [3501] = 830,
  [3502] = 3497,
  [3503] = 3503,
  [3504] = 3294,
  [3505] = 3290,
  [3506] = 3506,
  [3507] = 3507,
  [3508] = 3508,
  [3509] = 3509,
  [3510] = 3510,
  [3511] = 3511,
  [3512] = 3512,
  [3513] = 3513,
  [3514] = 3514,
  [3515] = 3515,
  [3516] = 3499,
  [3517] = 3499,
  [3518] = 3518,
  [3519] = 3295,
  [3520] = 3274,
  [3521] = 3276,
  [3522] = 3430,
  [3523] = 3523,
  [3524] = 3524,
  [3525] = 3525,
  [3526] = 3526,
  [3527] = 3527,
  [3528] = 3289,
  [3529] = 3335,
  [3530] = 3530,
  [3531] = 3530,
  [3532] = 3532,
  [3533] = 3533,
  [3534] = 3534,
  [3535] = 3535,
  [3536] = 3536,
  [3537] = 3318,
  [3538] = 3538,
  [3539] = 1521,
  [3540] = 3540,
  [3541] = 3530,
  [3542] = 1515,
  [3543] = 3543,
  [3544] = 3544,
  [3545] = 3530,
  [3546] = 3546,
  [3547] = 3547,
  [3548] = 3548,
  [3549] = 3549,
  [3550] = 3550,
  [3551] = 3551,
  [3552] = 3552,
  [3553] = 3553,
  [3554] = 3530,
  [3555] = 1510,
  [3556] = 3556,
  [3557] = 3350,
  [3558] = 3530,
  [3559] = 3530,
  [3560] = 3530,
  [3561] = 3561,
  [3562] = 3562,
  [3563] = 3530,
  [3564] = 3564,
  [3565] = 3530,
  [3566] = 3530,
  [3567] = 3396,
  [3568] = 3568,
  [3569] = 3346,
  [3570] = 2285,
  [3571] = 3086,
  [3572] = 3572,
  [3573] = 3573,
  [3574] = 3574,
  [3575] = 3575,
  [3576] = 3576,
  [3577] = 3577,
  [3578] = 3578,
  [3579] = 3579,
  [3580] = 3580,
  [3581] = 3581,
  [3582] = 3582,
  [3583] = 3583,
  [3584] = 3584,
  [3585] = 3585,
  [3586] = 3586,
  [3587] = 3587,
  [3588] = 3588,
  [3589] = 3589,
  [3590] = 3486,
  [3591] = 3574,
  [3592] = 822,
  [3593] = 3593,
  [3594] = 3594,
  [3595] = 3595,
  [3596] = 3596,
  [3597] = 3597,
  [3598] = 3598,
  [3599] = 3599,
  [3600] = 3440,
  [3601] = 3601,
  [3602] = 3602,
  [3603] = 3603,
  [3604] = 3604,
  [3605] = 3605,
  [3606] = 3606,
  [3607] = 3607,
  [3608] = 3608,
  [3609] = 3609,
  [3610] = 3610,
  [3611] = 3611,
  [3612] = 3612,
  [3613] = 3613,
  [3614] = 3614,
  [3615] = 3615,
  [3616] = 3616,
  [3617] = 3617,
  [3618] = 3618,
  [3619] = 3589,
  [3620] = 3620,
  [3621] = 3621,
  [3622] = 3622,
  [3623] = 3623,
  [3624] = 3624,
  [3625] = 3625,
  [3626] = 3626,
  [3627] = 3627,
  [3628] = 3628,
  [3629] = 3629,
  [3630] = 3630,
  [3631] = 3631,
  [3632] = 3632,
  [3633] = 3633,
  [3634] = 1538,
  [3635] = 3635,
  [3636] = 3636,
  [3637] = 3637,
  [3638] = 3638,
  [3639] = 3639,
  [3640] = 3640,
  [3641] = 3641,
  [3642] = 3628,
  [3643] = 3574,
  [3644] = 3644,
  [3645] = 3645,
  [3646] = 3646,
  [3647] = 3647,
  [3648] = 3648,
  [3649] = 3649,
  [3650] = 3628,
  [3651] = 3635,
  [3652] = 3652,
  [3653] = 3653,
  [3654] = 3617,
  [3655] = 3588,
  [3656] = 3589,
  [3657] = 3453,
  [3658] = 3637,
  [3659] = 3659,
  [3660] = 3574,
  [3661] = 3606,
  [3662] = 3460,
  [3663] = 3482,
  [3664] = 3469,
  [3665] = 3617,
  [3666] = 3444,
  [3667] = 3474,
  [3668] = 3668,
  [3669] = 3582,
  [3670] = 3484,
  [3671] = 3671,
  [3672] = 3451,
  [3673] = 3448,
  [3674] = 3456,
  [3675] = 3437,
  [3676] = 3620,
  [3677] = 3677,
  [3678] = 3455,
  [3679] = 3470,
  [3680] = 3443,
  [3681] = 3464,
  [3682] = 3452,
  [3683] = 3459,
  [3684] = 3466,
  [3685] = 3438,
  [3686] = 3636,
  [3687] = 3617,
  [3688] = 3589,
  [3689] = 3689,
  [3690] = 3690,
  [3691] = 3691,
  [3692] = 3574,
  [3693] = 3693,
  [3694] = 3694,
  [3695] = 3695,
  [3696] = 3696,
  [3697] = 3697,
  [3698] = 3637,
  [3699] = 3699,
  [3700] = 3700,
  [3701] = 3494,
  [3702] = 3702,
  [3703] = 3703,
  [3704] = 3704,
  [3705] = 3705,
  [3706] = 826,
  [3707] = 3707,
  [3708] = 3708,
  [3709] = 3709,
  [3710] = 3710,
  [3711] = 3711,
  [3712] = 3712,
  [3713] = 3713,
  [3714] = 3628,
  [3715] = 3715,
  [3716] = 3716,
  [3717] = 3717,
  [3718] = 3718,
  [3719] = 3719,
  [3720] = 3720,
  [3721] = 3721,
  [3722] = 801,
  [3723] = 3723,
  [3724] = 3724,
  [3725] = 3725,
  [3726] = 3726,
  [3727] = 3727,
  [3728] = 3635,
  [3729] = 3729,
  [3730] = 3588,
  [3731] = 3606,
  [3732] = 3732,
  [3733] = 3733,
  [3734] = 3734,
  [3735] = 3735,
  [3736] = 3736,
  [3737] = 3737,
  [3738] = 3695,
  [3739] = 3739,
  [3740] = 3695,
  [3741] = 3741,
  [3742] = 3732,
  [3743] = 3743,
  [3744] = 3744,
  [3745] = 3745,
  [3746] = 3746,
  [3747] = 3747,
  [3748] = 3748,
  [3749] = 3749,
  [3750] = 3750,
  [3751] = 3751,
  [3752] = 3752,
  [3753] = 3753,
  [3754] = 3754,
  [3755] = 3755,
  [3756] = 3756,
  [3757] = 3757,
  [3758] = 3758,
  [3759] = 3759,
  [3760] = 3637,
  [3761] = 3636,
  [3762] = 3467,
  [3763] = 3732,
  [3764] = 3463,
  [3765] = 3765,
  [3766] = 3766,
  [3767] = 3628,
  [3768] = 3617,
  [3769] = 3589,
  [3770] = 3574,
  [3771] = 3471,
  [3772] = 3695,
  [3773] = 3628,
  [3774] = 3617,
  [3775] = 3589,
  [3776] = 3776,
  [3777] = 3777,
  [3778] = 3778,
  [3779] = 3777,
  [3780] = 3780,
  [3781] = 3781,
  [3782] = 3777,
  [3783] = 3783,
  [3784] = 2636,
  [3785] = 3785,
  [3786] = 2609,
  [3787] = 3787,
  [3788] = 3788,
  [3789] = 3789,
  [3790] = 3790,
  [3791] = 3631,
  [3792] = 3524,
  [3793] = 3525,
  [3794] = 3758,
  [3795] = 3586,
  [3796] = 3623,
  [3797] = 3797,
  [3798] = 3777,
  [3799] = 3526,
  [3800] = 3694,
  [3801] = 3801,
  [3802] = 3802,
  [3803] = 2917,
  [3804] = 3804,
  [3805] = 3805,
  [3806] = 3806,
  [3807] = 3518,
  [3808] = 3777,
  [3809] = 3809,
  [3810] = 3810,
  [3811] = 3493,
  [3812] = 3812,
  [3813] = 3813,
  [3814] = 3814,
  [3815] = 3815,
  [3816] = 3777,
  [3817] = 3777,
  [3818] = 3818,
  [3819] = 3819,
  [3820] = 3820,
  [3821] = 3495,
  [3822] = 3822,
  [3823] = 3823,
  [3824] = 3824,
  [3825] = 3777,
  [3826] = 3826,
  [3827] = 3827,
  [3828] = 801,
  [3829] = 822,
  [3830] = 826,
  [3831] = 3831,
  [3832] = 3646,
  [3833] = 3527,
  [3834] = 3605,
  [3835] = 3500,
  [3836] = 3777,
  [3837] = 3837,
  [3838] = 801,
  [3839] = 822,
  [3840] = 826,
  [3841] = 2611,
  [3842] = 3842,
  [3843] = 2917,
  [3844] = 2938,
  [3845] = 3788,
  [3846] = 2938,
  [3847] = 3777,
  [3848] = 3777,
  [3849] = 3849,
  [3850] = 3546,
  [3851] = 3851,
  [3852] = 3852,
  [3853] = 3853,
  [3854] = 3854,
  [3855] = 3855,
  [3856] = 3851,
  [3857] = 3857,
  [3858] = 3858,
  [3859] = 3859,
  [3860] = 3860,
  [3861] = 3861,
  [3862] = 3556,
  [3863] = 3863,
  [3864] = 3864,
  [3865] = 3864,
  [3866] = 3866,
  [3867] = 3867,
  [3868] = 3866,
  [3869] = 3869,
  [3870] = 3549,
  [3871] = 3859,
  [3872] = 3863,
  [3873] = 3852,
  [3874] = 3534,
  [3875] = 3568,
  [3876] = 3861,
  [3877] = 3864,
  [3878] = 3853,
  [3879] = 3854,
  [3880] = 3869,
  [3881] = 3550,
  [3882] = 3859,
  [3883] = 3855,
  [3884] = 3861,
  [3885] = 3538,
  [3886] = 3853,
  [3887] = 3852,
  [3888] = 3853,
  [3889] = 3854,
  [3890] = 3890,
  [3891] = 3855,
  [3892] = 3551,
  [3893] = 3544,
  [3894] = 3854,
  [3895] = 3867,
  [3896] = 3896,
  [3897] = 3897,
  [3898] = 3552,
  [3899] = 3553,
  [3900] = 3900,
  [3901] = 3901,
  [3902] = 3902,
  [3903] = 3572,
  [3904] = 3532,
  [3905] = 3863,
  [3906] = 3547,
  [3907] = 3907,
  [3908] = 3908,
  [3909] = 3909,
  [3910] = 3864,
  [3911] = 3911,
  [3912] = 3866,
  [3913] = 3867,
  [3914] = 3863,
  [3915] = 3915,
  [3916] = 3869,
  [3917] = 3866,
  [3918] = 3918,
  [3919] = 3859,
  [3920] = 3861,
  [3921] = 3859,
  [3922] = 3864,
  [3923] = 3923,
  [3924] = 3533,
  [3925] = 3852,
  [3926] = 3536,
  [3927] = 3853,
  [3928] = 3928,
  [3929] = 3854,
  [3930] = 3863,
  [3931] = 3866,
  [3932] = 3855,
  [3933] = 3867,
  [3934] = 3934,
  [3935] = 3869,
  [3936] = 3936,
  [3937] = 3860,
  [3938] = 3859,
  [3939] = 3861,
  [3940] = 3869,
  [3941] = 3852,
  [3942] = 3859,
  [3943] = 3863,
  [3944] = 3866,
  [3945] = 3861,
  [3946] = 3867,
  [3947] = 3864,
  [3948] = 3852,
  [3949] = 3853,
  [3950] = 3854,
  [3951] = 3855,
  [3952] = 3952,
  [3953] = 3866,
  [3954] = 3952,
  [3955] = 3869,
  [3956] = 3859,
  [3957] = 3863,
  [3958] = 3866,
  [3959] = 3867,
  [3960] = 3960,
  [3961] = 3869,
  [3962] = 3851,
  [3963] = 3963,
  [3964] = 3861,
  [3965] = 3965,
  [3966] = 3859,
  [3967] = 3967,
  [3968] = 3864,
  [3969] = 3861,
  [3970] = 3852,
  [3971] = 3867,
  [3972] = 3851,
  [3973] = 3973,
  [3974] = 3864,
  [3975] = 3852,
  [3976] = 3853,
  [3977] = 3977,
  [3978] = 3978,
  [3979] = 3853,
  [3980] = 3851,
  [3981] = 3851,
  [3982] = 3866,
  [3983] = 3851,
  [3984] = 3851,
  [3985] = 3851,
  [3986] = 3851,
  [3987] = 3987,
  [3988] = 3854,
  [3989] = 3854,
  [3990] = 3855,
  [3991] = 3991,
  [3992] = 3867,
  [3993] = 3993,
  [3994] = 3561,
  [3995] = 3543,
  [3996] = 3855,
  [3997] = 3863,
  [3998] = 3866,
  [3999] = 3855,
  [4000] = 3867,
  [4001] = 3855,
  [4002] = 4002,
  [4003] = 3853,
  [4004] = 3869,
  [4005] = 4005,
  [4006] = 4006,
  [4007] = 3859,
  [4008] = 3861,
  [4009] = 3864,
  [4010] = 3086,
  [4011] = 3867,
  [4012] = 3852,
  [4013] = 3853,
  [4014] = 3854,
  [4015] = 3855,
  [4016] = 3854,
  [4017] = 3861,
  [4018] = 3863,
  [4019] = 3866,
  [4020] = 3867,
  [4021] = 4021,
  [4022] = 3869,
  [4023] = 3859,
  [4024] = 3861,
  [4025] = 3864,
  [4026] = 3852,
  [4027] = 3853,
  [4028] = 3854,
  [4029] = 3855,
  [4030] = 774,
  [4031] = 3869,
  [4032] = 3864,
  [4033] = 3562,
  [4034] = 3863,
  [4035] = 3852,
  [4036] = 3869,
  [4037] = 3863,
  [4038] = 4038,
  [4039] = 4039,
  [4040] = 4040,
  [4041] = 4041,
  [4042] = 4042,
  [4043] = 4043,
  [4044] = 4044,
  [4045] = 4045,
  [4046] = 4043,
  [4047] = 4043,
  [4048] = 4048,
  [4049] = 4044,
  [4050] = 3440,
  [4051] = 4051,
  [4052] = 4052,
  [4053] = 4053,
  [4054] = 4054,
  [4055] = 4055,
  [4056] = 4052,
  [4057] = 4057,
  [4058] = 4058,
  [4059] = 4059,
  [4060] = 814,
  [4061] = 4061,
  [4062] = 817,
  [4063] = 819,
  [4064] = 4064,
  [4065] = 807,
  [4066] = 4066,
  [4067] = 4061,
  [4068] = 4068,
  [4069] = 4069,
  [4070] = 4070,
  [4071] = 4071,
  [4072] = 790,
  [4073] = 4061,
  [4074] = 4068,
  [4075] = 4075,
  [4076] = 4076,
  [4077] = 4077,
  [4078] = 4078,
  [4079] = 4079,
  [4080] = 821,
  [4081] = 4081,
  [4082] = 4082,
  [4083] = 4069,
  [4084] = 4084,
  [4085] = 4077,
  [4086] = 4061,
  [4087] = 4061,
  [4088] = 789,
  [4089] = 825,
  [4090] = 4061,
  [4091] = 825,
  [4092] = 4077,
  [4093] = 4093,
  [4094] = 4094,
  [4095] = 4093,
  [4096] = 791,
  [4097] = 4058,
  [4098] = 4081,
  [4099] = 4084,
  [4100] = 795,
  [4101] = 4066,
  [4102] = 4094,
  [4103] = 4064,
  [4104] = 4061,
  [4105] = 789,
  [4106] = 4084,
  [4107] = 4107,
  [4108] = 790,
  [4109] = 4084,
  [4110] = 4110,
  [4111] = 4111,
  [4112] = 4061,
  [4113] = 766,
  [4114] = 4061,
  [4115] = 4057,
  [4116] = 4075,
  [4117] = 795,
  [4118] = 4070,
  [4119] = 4061,
  [4120] = 4077,
  [4121] = 4121,
  [4122] = 4121,
  [4123] = 4121,
  [4124] = 798,
  [4125] = 4121,
  [4126] = 4121,
  [4127] = 4121,
  [4128] = 3960,
  [4129] = 4021,
  [4130] = 3991,
  [4131] = 3857,
  [4132] = 3934,
  [4133] = 4121,
  [4134] = 4134,
  [4135] = 4121,
  [4136] = 4121,
  [4137] = 4134,
  [4138] = 2672,
  [4139] = 3896,
  [4140] = 3897,
  [4141] = 3902,
  [4142] = 3858,
  [4143] = 824,
  [4144] = 4121,
  [4145] = 4121,
  [4146] = 3977,
  [4147] = 3978,
  [4148] = 4121,
  [4149] = 4006,
  [4150] = 4053,
  [4151] = 4121,
  [4152] = 4055,
  [4153] = 4121,
  [4154] = 3890,
  [4155] = 3936,
  [4156] = 4134,
  [4157] = 4121,
  [4158] = 4121,
  [4159] = 799,
  [4160] = 810,
  [4161] = 815,
  [4162] = 4121,
  [4163] = 818,
  [4164] = 3911,
  [4165] = 4121,
  [4166] = 3928,
  [4167] = 3907,
  [4168] = 3900,
  [4169] = 3901,
  [4170] = 3987,
  [4171] = 3908,
  [4172] = 3923,
  [4173] = 2734,
  [4174] = 4174,
  [4175] = 4175,
  [4176] = 4176,
  [4177] = 1994,
  [4178] = 4178,
  [4179] = 2938,
  [4180] = 4180,
  [4181] = 4181,
  [4182] = 4182,
  [4183] = 4174,
  [4184] = 4176,
  [4185] = 2917,
  [4186] = 4186,
  [4187] = 4175,
  [4188] = 4188,
  [4189] = 4176,
  [4190] = 4180,
  [4191] = 4180,
  [4192] = 4188,
  [4193] = 4182,
  [4194] = 4194,
  [4195] = 4186,
  [4196] = 4180,
  [4197] = 4197,
  [4198] = 4198,
  [4199] = 4178,
  [4200] = 4200,
  [4201] = 3295,
  [4202] = 4175,
  [4203] = 4178,
  [4204] = 4178,
  [4205] = 3295,
  [4206] = 4200,
  [4207] = 1626,
  [4208] = 4197,
  [4209] = 4182,
  [4210] = 4210,
  [4211] = 2013,
  [4212] = 4212,
  [4213] = 4213,
  [4214] = 4214,
  [4215] = 4215,
  [4216] = 4216,
  [4217] = 4217,
  [4218] = 3089,
  [4219] = 3095,
  [4220] = 3097,
  [4221] = 4213,
  [4222] = 4222,
  [4223] = 2975,
  [4224] = 4224,
  [4225] = 1646,
  [4226] = 4217,
  [4227] = 3081,
  [4228] = 3083,
  [4229] = 3350,
  [4230] = 3896,
  [4231] = 4217,
  [4232] = 4213,
  [4233] = 2672,
  [4234] = 3897,
  [4235] = 820,
  [4236] = 3350,
  [4237] = 4216,
  [4238] = 4213,
  [4239] = 3902,
  [4240] = 4240,
  [4241] = 4216,
  [4242] = 4242,
  [4243] = 4243,
  [4244] = 2734,
  [4245] = 2651,
  [4246] = 4246,
  [4247] = 4247,
  [4248] = 4246,
  [4249] = 4249,
  [4250] = 2938,
  [4251] = 2652,
  [4252] = 4243,
  [4253] = 4253,
  [4254] = 1823,
  [4255] = 4243,
  [4256] = 3158,
  [4257] = 4243,
  [4258] = 2638,
  [4259] = 4259,
  [4260] = 2637,
  [4261] = 4261,
  [4262] = 3120,
  [4263] = 2498,
  [4264] = 4264,
  [4265] = 2648,
  [4266] = 1829,
  [4267] = 4267,
  [4268] = 4268,
  [4269] = 4269,
  [4270] = 4243,
  [4271] = 4271,
  [4272] = 2495,
  [4273] = 2734,
  [4274] = 2641,
  [4275] = 4275,
  [4276] = 4276,
  [4277] = 4243,
  [4278] = 2672,
  [4279] = 4269,
  [4280] = 4269,
  [4281] = 4243,
  [4282] = 4269,
  [4283] = 4243,
  [4284] = 3126,
  [4285] = 4285,
  [4286] = 4269,
  [4287] = 4287,
  [4288] = 4246,
  [4289] = 4269,
  [4290] = 4290,
  [4291] = 4275,
  [4292] = 4243,
  [4293] = 2494,
  [4294] = 4294,
  [4295] = 4269,
  [4296] = 4296,
  [4297] = 2917,
  [4298] = 4275,
  [4299] = 4269,
  [4300] = 2650,
  [4301] = 4301,
  [4302] = 4302,
  [4303] = 3172,
  [4304] = 4304,
  [4305] = 4305,
  [4306] = 4302,
  [4307] = 2235,
  [4308] = 2292,
  [4309] = 4309,
  [4310] = 4310,
  [4311] = 4311,
  [4312] = 3197,
  [4313] = 4313,
  [4314] = 4314,
  [4315] = 4315,
  [4316] = 4316,
  [4317] = 4317,
  [4318] = 4313,
  [4319] = 3809,
  [4320] = 4320,
  [4321] = 4304,
  [4322] = 4322,
  [4323] = 4323,
  [4324] = 4324,
  [4325] = 4325,
  [4326] = 4302,
  [4327] = 4327,
  [4328] = 4328,
  [4329] = 2250,
  [4330] = 4330,
  [4331] = 4331,
  [4332] = 4332,
  [4333] = 1969,
  [4334] = 3086,
  [4335] = 4335,
  [4336] = 2230,
  [4337] = 1995,
  [4338] = 4338,
  [4339] = 4302,
  [4340] = 2234,
  [4341] = 4341,
  [4342] = 4304,
  [4343] = 4302,
  [4344] = 4344,
  [4345] = 2232,
  [4346] = 4302,
  [4347] = 4347,
  [4348] = 4348,
  [4349] = 4349,
  [4350] = 4350,
  [4351] = 4351,
  [4352] = 4352,
  [4353] = 4353,
  [4354] = 2503,
  [4355] = 4355,
  [4356] = 4304,
  [4357] = 4357,
  [4358] = 4304,
  [4359] = 4359,
  [4360] = 3804,
  [4361] = 3805,
  [4362] = 4304,
  [4363] = 4363,
  [4364] = 4364,
  [4365] = 4365,
  [4366] = 4302,
  [4367] = 1943,
  [4368] = 1945,
  [4369] = 4302,
  [4370] = 4370,
  [4371] = 4371,
  [4372] = 4372,
  [4373] = 4373,
  [4374] = 4374,
  [4375] = 2224,
  [4376] = 3806,
  [4377] = 3176,
  [4378] = 4378,
  [4379] = 4379,
  [4380] = 4380,
  [4381] = 4381,
  [4382] = 4382,
  [4383] = 4304,
  [4384] = 3781,
  [4385] = 4385,
  [4386] = 4386,
  [4387] = 2239,
  [4388] = 4388,
  [4389] = 4389,
  [4390] = 4390,
  [4391] = 4391,
  [4392] = 4392,
  [4393] = 4304,
  [4394] = 3179,
  [4395] = 4395,
  [4396] = 4396,
  [4397] = 4302,
  [4398] = 4398,
  [4399] = 4399,
  [4400] = 4400,
  [4401] = 4401,
  [4402] = 4402,
  [4403] = 2056,
  [4404] = 4404,
  [4405] = 4405,
  [4406] = 4406,
  [4407] = 2240,
  [4408] = 4408,
  [4409] = 793,
  [4410] = 4410,
  [4411] = 4411,
  [4412] = 2264,
  [4413] = 4413,
  [4414] = 2287,
  [4415] = 828,
  [4416] = 4304,
  [4417] = 4417,
  [4418] = 4418,
  [4419] = 4419,
  [4420] = 4302,
  [4421] = 4421,
  [4422] = 2242,
  [4423] = 3785,
  [4424] = 4424,
  [4425] = 2037,
  [4426] = 3789,
  [4427] = 4304,
  [4428] = 2286,
  [4429] = 3790,
  [4430] = 4430,
  [4431] = 4431,
  [4432] = 4432,
  [4433] = 4433,
  [4434] = 4434,
  [4435] = 4435,
  [4436] = 2289,
  [4437] = 4437,
  [4438] = 4438,
  [4439] = 2604,
  [4440] = 2613,
  [4441] = 4441,
  [4442] = 2570,
  [4443] = 2577,
  [4444] = 2618,
  [4445] = 4445,
  [4446] = 2578,
  [4447] = 2579,
  [4448] = 4448,
  [4449] = 2585,
  [4450] = 4450,
  [4451] = 4451,
  [4452] = 2574,
  [4453] = 2571,
  [4454] = 796,
  [4455] = 2582,
  [4456] = 4456,
  [4457] = 2580,
  [4458] = 2586,
  [4459] = 2609,
  [4460] = 4460,
  [4461] = 3228,
  [4462] = 2997,
  [4463] = 3254,
  [4464] = 2583,
  [4465] = 4465,
  [4466] = 3060,
  [4467] = 4467,
  [4468] = 4456,
  [4469] = 3240,
  [4470] = 4470,
  [4471] = 2589,
  [4472] = 2590,
  [4473] = 2608,
  [4474] = 4474,
  [4475] = 4465,
  [4476] = 2576,
  [4477] = 3256,
  [4478] = 4478,
  [4479] = 3079,
  [4480] = 2573,
  [4481] = 2593,
  [4482] = 3258,
  [4483] = 4456,
  [4484] = 2595,
  [4485] = 2611,
  [4486] = 2632,
  [4487] = 4487,
  [4488] = 2629,
  [4489] = 4465,
  [4490] = 2603,
  [4491] = 2584,
  [4492] = 2605,
  [4493] = 829,
  [4494] = 4494,
  [4495] = 3157,
  [4496] = 4496,
  [4497] = 3131,
  [4498] = 3127,
  [4499] = 3294,
  [4500] = 3162,
  [4501] = 4501,
  [4502] = 3121,
  [4503] = 4503,
  [4504] = 4496,
  [4505] = 4505,
  [4506] = 4506,
  [4507] = 4507,
  [4508] = 3288,
  [4509] = 4505,
  [4510] = 752,
  [4511] = 4511,
  [4512] = 3140,
  [4513] = 3164,
  [4514] = 4514,
  [4515] = 4494,
  [4516] = 4516,
  [4517] = 3290,
  [4518] = 4496,
  [4519] = 4507,
  [4520] = 4505,
  [4521] = 3122,
  [4522] = 3274,
  [4523] = 4511,
  [4524] = 1825,
  [4525] = 3136,
  [4526] = 3275,
  [4527] = 3124,
  [4528] = 4528,
  [4529] = 3138,
  [4530] = 4511,
  [4531] = 3295,
  [4532] = 4528,
  [4533] = 4528,
  [4534] = 3276,
  [4535] = 3289,
  [4536] = 4536,
  [4537] = 755,
  [4538] = 3167,
  [4539] = 4539,
  [4540] = 4540,
  [4541] = 4539,
  [4542] = 2001,
  [4543] = 3248,
  [4544] = 3335,
  [4545] = 3256,
  [4546] = 2048,
  [4547] = 3263,
  [4548] = 2051,
  [4549] = 4549,
  [4550] = 3346,
  [4551] = 4539,
  [4552] = 4552,
  [4553] = 2085,
  [4554] = 4552,
  [4555] = 4539,
  [4556] = 3262,
  [4557] = 4552,
  [4558] = 4539,
  [4559] = 3239,
  [4560] = 2063,
  [4561] = 4552,
  [4562] = 2059,
  [4563] = 3318,
  [4564] = 2087,
  [4565] = 2064,
  [4566] = 4552,
  [4567] = 4552,
  [4568] = 4552,
  [4569] = 4539,
  [4570] = 2036,
  [4571] = 4571,
  [4572] = 1952,
  [4573] = 2498,
  [4574] = 3260,
  [4575] = 2043,
  [4576] = 4552,
  [4577] = 4539,
  [4578] = 4539,
  [4579] = 1931,
  [4580] = 2076,
  [4581] = 1926,
  [4582] = 3247,
  [4583] = 4539,
  [4584] = 1970,
  [4585] = 2068,
  [4586] = 4552,
  [4587] = 2057,
  [4588] = 2069,
  [4589] = 3252,
  [4590] = 4539,
  [4591] = 4539,
  [4592] = 2071,
  [4593] = 3258,
  [4594] = 2072,
  [4595] = 4539,
  [4596] = 4539,
  [4597] = 4539,
  [4598] = 4552,
  [4599] = 2060,
  [4600] = 2061,
  [4601] = 4552,
  [4602] = 2062,
  [4603] = 3396,
  [4604] = 4539,
  [4605] = 1997,
  [4606] = 4539,
  [4607] = 3310,
  [4608] = 3254,
  [4609] = 1922,
  [4610] = 3350,
  [4611] = 1925,
  [4612] = 1975,
  [4613] = 4539,
  [4614] = 1934,
  [4615] = 4615,
  [4616] = 1996,
  [4617] = 4617,
  [4618] = 2089,
  [4619] = 4619,
  [4620] = 4620,
  [4621] = 3295,
  [4622] = 4619,
  [4623] = 4623,
  [4624] = 4624,
  [4625] = 4625,
  [4626] = 3477,
  [4627] = 3439,
  [4628] = 3440,
  [4629] = 4620,
  [4630] = 4625,
  [4631] = 4631,
  [4632] = 4632,
  [4633] = 3460,
  [4634] = 4620,
  [4635] = 3482,
  [4636] = 3275,
  [4637] = 4637,
  [4638] = 1826,
  [4639] = 4451,
  [4640] = 4640,
  [4641] = 3276,
  [4642] = 3467,
  [4643] = 3309,
  [4644] = 3453,
  [4645] = 4645,
  [4646] = 3469,
  [4647] = 4645,
  [4648] = 4648,
  [4649] = 3441,
  [4650] = 3474,
  [4651] = 3518,
  [4652] = 4652,
  [4653] = 4625,
  [4654] = 3463,
  [4655] = 4648,
  [4656] = 3451,
  [4657] = 3448,
  [4658] = 3456,
  [4659] = 3437,
  [4660] = 3495,
  [4661] = 4661,
  [4662] = 4648,
  [4663] = 4631,
  [4664] = 4652,
  [4665] = 3455,
  [4666] = 3470,
  [4667] = 3274,
  [4668] = 3443,
  [4669] = 3464,
  [4670] = 3294,
  [4671] = 4631,
  [4672] = 4640,
  [4673] = 4652,
  [4674] = 4674,
  [4675] = 3452,
  [4676] = 3459,
  [4677] = 3466,
  [4678] = 4678,
  [4679] = 3438,
  [4680] = 3471,
  [4681] = 3444,
  [4682] = 3357,
  [4683] = 3506,
  [4684] = 3527,
  [4685] = 3358,
  [4686] = 3352,
  [4687] = 4687,
  [4688] = 4688,
  [4689] = 4689,
  [4690] = 4690,
  [4691] = 4501,
  [4692] = 4692,
  [4693] = 795,
  [4694] = 4694,
  [4695] = 4688,
  [4696] = 4451,
  [4697] = 4687,
  [4698] = 4698,
  [4699] = 3472,
  [4700] = 3526,
  [4701] = 4701,
  [4702] = 3349,
  [4703] = 2024,
  [4704] = 4694,
  [4705] = 4705,
  [4706] = 4694,
  [4707] = 4707,
  [4708] = 1927,
  [4709] = 4707,
  [4710] = 4710,
  [4711] = 1929,
  [4712] = 1930,
  [4713] = 1932,
  [4714] = 1933,
  [4715] = 4705,
  [4716] = 4716,
  [4717] = 1823,
  [4718] = 4718,
  [4719] = 4719,
  [4720] = 4516,
  [4721] = 4719,
  [4722] = 4722,
  [4723] = 4722,
  [4724] = 4716,
  [4725] = 4285,
  [4726] = 4688,
  [4727] = 4718,
  [4728] = 4536,
  [4729] = 4729,
  [4730] = 4514,
  [4731] = 4701,
  [4732] = 4732,
  [4733] = 4716,
  [4734] = 4506,
  [4735] = 3515,
  [4736] = 4719,
  [4737] = 1826,
  [4738] = 4722,
  [4739] = 1825,
  [4740] = 4729,
  [4741] = 4732,
  [4742] = 4710,
  [4743] = 4729,
  [4744] = 4503,
  [4745] = 4707,
  [4746] = 4729,
  [4747] = 3367,
  [4748] = 4732,
  [4749] = 4749,
  [4750] = 4729,
  [4751] = 4732,
  [4752] = 4688,
  [4753] = 4710,
  [4754] = 4754,
  [4755] = 4729,
  [4756] = 3493,
  [4757] = 4757,
  [4758] = 4732,
  [4759] = 4729,
  [4760] = 3449,
  [4761] = 4729,
  [4762] = 4688,
  [4763] = 4732,
  [4764] = 801,
  [4765] = 4729,
  [4766] = 4732,
  [4767] = 4722,
  [4768] = 4690,
  [4769] = 4729,
  [4770] = 4732,
  [4771] = 4698,
  [4772] = 4729,
  [4773] = 4732,
  [4774] = 4687,
  [4775] = 4688,
  [4776] = 4732,
  [4777] = 822,
  [4778] = 4694,
  [4779] = 4779,
  [4780] = 1826,
  [4781] = 4687,
  [4782] = 4705,
  [4783] = 4783,
  [4784] = 4688,
  [4785] = 4716,
  [4786] = 789,
  [4787] = 1829,
  [4788] = 4788,
  [4789] = 4719,
  [4790] = 4790,
  [4791] = 790,
  [4792] = 3350,
  [4793] = 4692,
  [4794] = 4707,
  [4795] = 825,
  [4796] = 3417,
  [4797] = 4705,
  [4798] = 793,
  [4799] = 3465,
  [4800] = 3393,
  [4801] = 3524,
  [4802] = 4710,
  [4803] = 826,
  [4804] = 804,
  [4805] = 4718,
  [4806] = 4688,
  [4807] = 3525,
  [4808] = 830,
  [4809] = 3359,
  [4810] = 3500,
  [4811] = 4732,
  [4812] = 828,
  [4813] = 4701,
  [4814] = 4814,
  [4815] = 4374,
  [4816] = 4816,
  [4817] = 2071,
  [4818] = 4818,
  [4819] = 824,
  [4820] = 4820,
  [4821] = 4821,
  [4822] = 2024,
  [4823] = 2072,
  [4824] = 4824,
  [4825] = 4825,
  [4826] = 4430,
  [4827] = 804,
  [4828] = 2037,
  [4829] = 3562,
  [4830] = 4405,
  [4831] = 2043,
  [4832] = 4322,
  [4833] = 4418,
  [4834] = 3467,
  [4835] = 2048,
  [4836] = 2235,
  [4837] = 2232,
  [4838] = 2234,
  [4839] = 2239,
  [4840] = 2240,
  [4841] = 1927,
  [4842] = 4516,
  [4843] = 2242,
  [4844] = 1929,
  [4845] = 4818,
  [4846] = 2060,
  [4847] = 814,
  [4848] = 2061,
  [4849] = 3534,
  [4850] = 2062,
  [4851] = 3568,
  [4852] = 1930,
  [4853] = 817,
  [4854] = 4854,
  [4855] = 1932,
  [4856] = 1996,
  [4857] = 819,
  [4858] = 807,
  [4859] = 1994,
  [4860] = 4536,
  [4861] = 4861,
  [4862] = 1933,
  [4863] = 830,
  [4864] = 3498,
  [4865] = 2013,
  [4866] = 4866,
  [4867] = 4514,
  [4868] = 4501,
  [4869] = 4816,
  [4870] = 4818,
  [4871] = 4871,
  [4872] = 2224,
  [4873] = 4873,
  [4874] = 4341,
  [4875] = 4506,
  [4876] = 4816,
  [4877] = 3463,
  [4878] = 4351,
  [4879] = 4355,
  [4880] = 4880,
  [4881] = 2089,
  [4882] = 4821,
  [4883] = 1925,
  [4884] = 1926,
  [4885] = 1931,
  [4886] = 1934,
  [4887] = 4821,
  [4888] = 4866,
  [4889] = 4503,
  [4890] = 3507,
  [4891] = 3508,
  [4892] = 4892,
  [4893] = 3509,
  [4894] = 2250,
  [4895] = 3510,
  [4896] = 4824,
  [4897] = 4435,
  [4898] = 4898,
  [4899] = 4309,
  [4900] = 4820,
  [4901] = 2036,
  [4902] = 4854,
  [4903] = 4824,
  [4904] = 4904,
  [4905] = 4424,
  [4906] = 3430,
  [4907] = 4861,
  [4908] = 4908,
  [4909] = 3550,
  [4910] = 4816,
  [4911] = 2264,
  [4912] = 4861,
  [4913] = 3551,
  [4914] = 2051,
  [4915] = 774,
  [4916] = 2059,
  [4917] = 4431,
  [4918] = 1922,
  [4919] = 4818,
  [4920] = 4316,
  [4921] = 2085,
  [4922] = 2087,
  [4923] = 3561,
  [4924] = 3546,
  [4925] = 4816,
  [4926] = 2024,
  [4927] = 4821,
  [4928] = 3538,
  [4929] = 4929,
  [4930] = 4323,
  [4931] = 4854,
  [4932] = 4898,
  [4933] = 4824,
  [4934] = 3471,
  [4935] = 4854,
  [4936] = 4861,
  [4937] = 798,
  [4938] = 4324,
  [4939] = 4818,
  [4940] = 2063,
  [4941] = 4941,
  [4942] = 2064,
  [4943] = 3556,
  [4944] = 1975,
  [4945] = 4861,
  [4946] = 3487,
  [4947] = 4330,
  [4948] = 3488,
  [4949] = 3489,
  [4950] = 3490,
  [4951] = 3491,
  [4952] = 3492,
  [4953] = 4818,
  [4954] = 4854,
  [4955] = 4861,
  [4956] = 3533,
  [4957] = 3536,
  [4958] = 4821,
  [4959] = 4816,
  [4960] = 2076,
  [4961] = 4821,
  [4962] = 2285,
  [4963] = 1969,
  [4964] = 3543,
  [4965] = 4904,
  [4966] = 4824,
  [4967] = 3544,
  [4968] = 4968,
  [4969] = 4816,
  [4970] = 2286,
  [4971] = 2068,
  [4972] = 1927,
  [4973] = 4854,
  [4974] = 4861,
  [4975] = 1929,
  [4976] = 1930,
  [4977] = 4873,
  [4978] = 2287,
  [4979] = 4824,
  [4980] = 1932,
  [4981] = 4981,
  [4982] = 4818,
  [4983] = 4983,
  [4984] = 3485,
  [4985] = 1933,
  [4986] = 4816,
  [4987] = 4821,
  [4988] = 4824,
  [4989] = 1970,
  [4990] = 4824,
  [4991] = 1997,
  [4992] = 3511,
  [4993] = 3512,
  [4994] = 3513,
  [4995] = 3514,
  [4996] = 4854,
  [4997] = 4821,
  [4998] = 821,
  [4999] = 4861,
  [5000] = 2289,
  [5001] = 1995,
  [5002] = 2292,
  [5003] = 4818,
  [5004] = 3549,
  [5005] = 4818,
  [5006] = 2069,
  [5007] = 2001,
  [5008] = 4816,
  [5009] = 1943,
  [5010] = 1945,
  [5011] = 4821,
  [5012] = 3552,
  [5013] = 4824,
  [5014] = 2230,
  [5015] = 3553,
  [5016] = 4873,
  [5017] = 5017,
  [5018] = 4981,
  [5019] = 4854,
  [5020] = 4861,
  [5021] = 4861,
  [5022] = 1952,
  [5023] = 3523,
  [5024] = 5024,
  [5025] = 796,
  [5026] = 3572,
  [5027] = 3532,
  [5028] = 4818,
  [5029] = 4968,
  [5030] = 2056,
  [5031] = 4816,
  [5032] = 820,
  [5033] = 4821,
  [5034] = 4968,
  [5035] = 3547,
  [5036] = 4824,
  [5037] = 4981,
  [5038] = 2057,
  [5039] = 829,
  [5040] = 4854,
  [5041] = 4854,
  [5042] = 3602,
  [5043] = 5043,
  [5044] = 5044,
  [5045] = 5045,
  [5046] = 3708,
  [5047] = 3576,
  [5048] = 5048,
  [5049] = 5049,
  [5050] = 5050,
  [5051] = 5051,
  [5052] = 3765,
  [5053] = 3577,
  [5054] = 3717,
  [5055] = 3750,
  [5056] = 5056,
  [5057] = 3720,
  [5058] = 3727,
  [5059] = 3729,
  [5060] = 3754,
  [5061] = 3743,
  [5062] = 3584,
  [5063] = 5063,
  [5064] = 3601,
  [5065] = 3587,
  [5066] = 3597,
  [5067] = 3611,
  [5068] = 3613,
  [5069] = 5069,
  [5070] = 3626,
  [5071] = 3627,
  [5072] = 5072,
  [5073] = 3573,
  [5074] = 3648,
  [5075] = 3649,
  [5076] = 5045,
  [5077] = 3593,
  [5078] = 5043,
  [5079] = 3540,
  [5080] = 5043,
  [5081] = 5049,
  [5082] = 5082,
  [5083] = 5051,
  [5084] = 5043,
  [5085] = 5085,
  [5086] = 3700,
  [5087] = 5056,
  [5088] = 5088,
  [5089] = 5089,
  [5090] = 5043,
  [5091] = 5091,
  [5092] = 5051,
  [5093] = 5056,
  [5094] = 3735,
  [5095] = 3776,
  [5096] = 5096,
  [5097] = 5051,
  [5098] = 5098,
  [5099] = 5049,
  [5100] = 5056,
  [5101] = 3748,
  [5102] = 5102,
  [5103] = 5051,
  [5104] = 5104,
  [5105] = 3598,
  [5106] = 5106,
  [5107] = 3659,
  [5108] = 3599,
  [5109] = 3535,
  [5110] = 3603,
  [5111] = 5102,
  [5112] = 5049,
  [5113] = 5051,
  [5114] = 5114,
  [5115] = 5051,
  [5116] = 5116,
  [5117] = 3607,
  [5118] = 3608,
  [5119] = 3609,
  [5120] = 3610,
  [5121] = 5121,
  [5122] = 5051,
  [5123] = 3615,
  [5124] = 3616,
  [5125] = 3618,
  [5126] = 3745,
  [5127] = 3621,
  [5128] = 3622,
  [5129] = 3625,
  [5130] = 5051,
  [5131] = 5056,
  [5132] = 3629,
  [5133] = 3630,
  [5134] = 3632,
  [5135] = 3633,
  [5136] = 3677,
  [5137] = 3639,
  [5138] = 3640,
  [5139] = 3641,
  [5140] = 5043,
  [5141] = 5141,
  [5142] = 5088,
  [5143] = 3751,
  [5144] = 5144,
  [5145] = 5063,
  [5146] = 5146,
  [5147] = 5106,
  [5148] = 5146,
  [5149] = 5051,
  [5150] = 5150,
  [5151] = 5104,
  [5152] = 5096,
  [5153] = 3581,
  [5154] = 5154,
  [5155] = 5043,
  [5156] = 5156,
  [5157] = 5157,
  [5158] = 3578,
  [5159] = 5063,
  [5160] = 3749,
  [5161] = 5102,
  [5162] = 5085,
  [5163] = 3548,
  [5164] = 5043,
  [5165] = 3705,
  [5166] = 3756,
  [5167] = 5167,
  [5168] = 3707,
  [5169] = 5096,
  [5170] = 3710,
  [5171] = 5045,
  [5172] = 3711,
  [5173] = 5173,
  [5174] = 3579,
  [5175] = 5085,
  [5176] = 3712,
  [5177] = 3713,
  [5178] = 3715,
  [5179] = 3716,
  [5180] = 3580,
  [5181] = 3718,
  [5182] = 3721,
  [5183] = 5183,
  [5184] = 5184,
  [5185] = 3723,
  [5186] = 5146,
  [5187] = 3430,
  [5188] = 5056,
  [5189] = 3595,
  [5190] = 3724,
  [5191] = 3725,
  [5192] = 3726,
  [5193] = 5193,
  [5194] = 5043,
  [5195] = 3733,
  [5196] = 5088,
  [5197] = 5063,
  [5198] = 3564,
  [5199] = 3747,
  [5200] = 3752,
  [5201] = 3753,
  [5202] = 3755,
  [5203] = 3759,
  [5204] = 3766,
  [5205] = 3647,
  [5206] = 3644,
  [5207] = 3624,
  [5208] = 3583,
  [5209] = 3594,
  [5210] = 5104,
  [5211] = 3596,
  [5212] = 3575,
  [5213] = 3604,
  [5214] = 4214,
  [5215] = 764,
  [5216] = 5216,
  [5217] = 3668,
  [5218] = 3671,
  [5219] = 5121,
  [5220] = 3689,
  [5221] = 3690,
  [5222] = 3691,
  [5223] = 3693,
  [5224] = 5048,
  [5225] = 5048,
  [5226] = 3696,
  [5227] = 5227,
  [5228] = 3697,
  [5229] = 5088,
  [5230] = 4268,
  [5231] = 5146,
  [5232] = 3699,
  [5233] = 3703,
  [5234] = 5106,
  [5235] = 3704,
  [5236] = 3709,
  [5237] = 3719,
  [5238] = 3734,
  [5239] = 3736,
  [5240] = 5050,
  [5241] = 5106,
  [5242] = 3645,
  [5243] = 5216,
  [5244] = 5104,
  [5245] = 3746,
  [5246] = 3524,
  [5247] = 3525,
  [5248] = 3526,
  [5249] = 3527,
  [5250] = 5050,
  [5251] = 5216,
  [5252] = 5050,
  [5253] = 5216,
  [5254] = 5121,
  [5255] = 5043,
  [5256] = 5096,
  [5257] = 3737,
  [5258] = 3739,
  [5259] = 3741,
  [5260] = 3744,
  [5261] = 3702,
  [5262] = 5262,
  [5263] = 4355,
  [5264] = 5264,
  [5265] = 5265,
  [5266] = 5266,
  [5267] = 5267,
  [5268] = 5268,
  [5269] = 5269,
  [5270] = 5270,
  [5271] = 5271,
  [5272] = 5272,
  [5273] = 3804,
  [5274] = 5274,
  [5275] = 5275,
  [5276] = 5276,
  [5277] = 5277,
  [5278] = 5270,
  [5279] = 5279,
  [5280] = 3805,
  [5281] = 5264,
  [5282] = 5282,
  [5283] = 5283,
  [5284] = 5284,
  [5285] = 3806,
  [5286] = 5286,
  [5287] = 5271,
  [5288] = 5288,
  [5289] = 5289,
  [5290] = 3787,
  [5291] = 5291,
  [5292] = 5292,
  [5293] = 5293,
  [5294] = 5294,
  [5295] = 3809,
  [5296] = 5296,
  [5297] = 3814,
  [5298] = 5288,
  [5299] = 4249,
  [5300] = 3781,
  [5301] = 5271,
  [5302] = 5302,
  [5303] = 5303,
  [5304] = 3826,
  [5305] = 5305,
  [5306] = 5306,
  [5307] = 5276,
  [5308] = 5292,
  [5309] = 3837,
  [5310] = 3631,
  [5311] = 5311,
  [5312] = 5279,
  [5313] = 5313,
  [5314] = 5292,
  [5315] = 5315,
  [5316] = 5316,
  [5317] = 5283,
  [5318] = 5294,
  [5319] = 5319,
  [5320] = 5293,
  [5321] = 5321,
  [5322] = 2064,
  [5323] = 3842,
  [5324] = 5291,
  [5325] = 5292,
  [5326] = 5275,
  [5327] = 5277,
  [5328] = 5277,
  [5329] = 3849,
  [5330] = 5282,
  [5331] = 5302,
  [5332] = 5266,
  [5333] = 3527,
  [5334] = 2609,
  [5335] = 5305,
  [5336] = 5291,
  [5337] = 2611,
  [5338] = 5262,
  [5339] = 5271,
  [5340] = 5311,
  [5341] = 5341,
  [5342] = 5342,
  [5343] = 5343,
  [5344] = 5311,
  [5345] = 5303,
  [5346] = 5346,
  [5347] = 5284,
  [5348] = 5348,
  [5349] = 5289,
  [5350] = 5288,
  [5351] = 3500,
  [5352] = 5293,
  [5353] = 5353,
  [5354] = 1872,
  [5355] = 3525,
  [5356] = 5272,
  [5357] = 5357,
  [5358] = 5358,
  [5359] = 5289,
  [5360] = 5275,
  [5361] = 4341,
  [5362] = 5362,
  [5363] = 5271,
  [5364] = 5313,
  [5365] = 5365,
  [5366] = 3623,
  [5367] = 3823,
  [5368] = 5288,
  [5369] = 3524,
  [5370] = 5292,
  [5371] = 5357,
  [5372] = 5302,
  [5373] = 5271,
  [5374] = 5272,
  [5375] = 5375,
  [5376] = 5313,
  [5377] = 5271,
  [5378] = 5291,
  [5379] = 5379,
  [5380] = 5316,
  [5381] = 3527,
  [5382] = 3525,
  [5383] = 5291,
  [5384] = 5264,
  [5385] = 5319,
  [5386] = 4351,
  [5387] = 5296,
  [5388] = 5362,
  [5389] = 1521,
  [5390] = 2072,
  [5391] = 5283,
  [5392] = 3824,
  [5393] = 5393,
  [5394] = 3815,
  [5395] = 5319,
  [5396] = 3810,
  [5397] = 4330,
  [5398] = 5398,
  [5399] = 5292,
  [5400] = 5400,
  [5401] = 3468,
  [5402] = 5293,
  [5403] = 3447,
  [5404] = 5348,
  [5405] = 5365,
  [5406] = 3524,
  [5407] = 5315,
  [5408] = 5408,
  [5409] = 5283,
  [5410] = 5269,
  [5411] = 3694,
  [5412] = 5266,
  [5413] = 5288,
  [5414] = 3526,
  [5415] = 5305,
  [5416] = 5265,
  [5417] = 5398,
  [5418] = 5418,
  [5419] = 5283,
  [5420] = 5271,
  [5421] = 5421,
  [5422] = 3758,
  [5423] = 3813,
  [5424] = 5268,
  [5425] = 2048,
  [5426] = 3820,
  [5427] = 5283,
  [5428] = 5428,
  [5429] = 5270,
  [5430] = 5270,
  [5431] = 5431,
  [5432] = 5431,
  [5433] = 3484,
  [5434] = 5271,
  [5435] = 5283,
  [5436] = 5379,
  [5437] = 3480,
  [5438] = 5438,
  [5439] = 2063,
  [5440] = 5275,
  [5441] = 5319,
  [5442] = 5418,
  [5443] = 5283,
  [5444] = 5321,
  [5445] = 5445,
  [5446] = 5277,
  [5447] = 3614,
  [5448] = 3446,
  [5449] = 3450,
  [5450] = 5283,
  [5451] = 3646,
  [5452] = 5431,
  [5453] = 1515,
  [5454] = 5454,
  [5455] = 5362,
  [5456] = 5283,
  [5457] = 4424,
  [5458] = 3785,
  [5459] = 5267,
  [5460] = 5292,
  [5461] = 3789,
  [5462] = 3790,
  [5463] = 5343,
  [5464] = 3896,
  [5465] = 5393,
  [5466] = 3902,
  [5467] = 5379,
  [5468] = 5319,
  [5469] = 5271,
  [5470] = 3802,
  [5471] = 5289,
  [5472] = 5305,
  [5473] = 1922,
  [5474] = 5319,
  [5475] = 5428,
  [5476] = 5303,
  [5477] = 5288,
  [5478] = 5478,
  [5479] = 3897,
  [5480] = 5480,
  [5481] = 1510,
  [5482] = 5292,
  [5483] = 5319,
  [5484] = 5428,
  [5485] = 5485,
  [5486] = 5275,
  [5487] = 5277,
  [5488] = 5289,
  [5489] = 5272,
  [5490] = 5393,
  [5491] = 5348,
  [5492] = 3818,
  [5493] = 5292,
  [5494] = 5272,
  [5495] = 5279,
  [5496] = 5496,
  [5497] = 5268,
  [5498] = 5294,
  [5499] = 3801,
  [5500] = 5500,
  [5501] = 5292,
  [5502] = 3494,
  [5503] = 5291,
  [5504] = 5504,
  [5505] = 3605,
  [5506] = 5500,
  [5507] = 5375,
  [5508] = 5353,
  [5509] = 3757,
  [5510] = 5267,
  [5511] = 5291,
  [5512] = 5319,
  [5513] = 3585,
  [5514] = 5315,
  [5515] = 3526,
  [5516] = 3454,
  [5517] = 5319,
  [5518] = 5353,
  [5519] = 5343,
  [5520] = 5316,
  [5521] = 3481,
  [5522] = 5500,
  [5523] = 5319,
  [5524] = 5357,
  [5525] = 5478,
  [5526] = 3638,
  [5527] = 5321,
  [5528] = 3486,
  [5529] = 5529,
  [5530] = 5275,
  [5531] = 5277,
  [5532] = 5289,
  [5533] = 5398,
  [5534] = 3462,
  [5535] = 3586,
  [5536] = 5270,
  [5537] = 5265,
  [5538] = 5269,
  [5539] = 5265,
  [5540] = 5540,
  [5541] = 5541,
  [5542] = 5542,
  [5543] = 1538,
  [5544] = 5544,
  [5545] = 5545,
  [5546] = 5546,
  [5547] = 5547,
  [5548] = 5548,
  [5549] = 5549,
  [5550] = 5550,
  [5551] = 5551,
  [5552] = 5552,
  [5553] = 5553,
  [5554] = 5554,
  [5555] = 5555,
  [5556] = 5556,
  [5557] = 5557,
  [5558] = 5558,
  [5559] = 5559,
  [5560] = 5560,
  [5561] = 3991,
  [5562] = 5562,
  [5563] = 5563,
  [5564] = 5546,
  [5565] = 5565,
  [5566] = 5550,
  [5567] = 5553,
  [5568] = 5555,
  [5569] = 5569,
  [5570] = 3908,
  [5571] = 5571,
  [5572] = 5557,
  [5573] = 5558,
  [5574] = 5574,
  [5575] = 5562,
  [5576] = 5576,
  [5577] = 5546,
  [5578] = 5578,
  [5579] = 3778,
  [5580] = 5580,
  [5581] = 5581,
  [5582] = 3858,
  [5583] = 5557,
  [5584] = 5542,
  [5585] = 5544,
  [5586] = 5547,
  [5587] = 5548,
  [5588] = 5549,
  [5589] = 5589,
  [5590] = 5554,
  [5591] = 5560,
  [5592] = 5541,
  [5593] = 5593,
  [5594] = 5594,
  [5595] = 5562,
  [5596] = 5596,
  [5597] = 5597,
  [5598] = 5546,
  [5599] = 5599,
  [5600] = 5546,
  [5601] = 5601,
  [5602] = 5602,
  [5603] = 5603,
  [5604] = 5604,
  [5605] = 5605,
  [5606] = 5606,
  [5607] = 5569,
  [5608] = 5556,
  [5609] = 5609,
  [5610] = 5610,
  [5611] = 5611,
  [5612] = 5612,
  [5613] = 5606,
  [5614] = 5601,
  [5615] = 5550,
  [5616] = 5616,
  [5617] = 4006,
  [5618] = 5618,
  [5619] = 5602,
  [5620] = 5604,
  [5621] = 5621,
  [5622] = 5622,
  [5623] = 5553,
  [5624] = 5624,
  [5625] = 5625,
  [5626] = 5626,
  [5627] = 5627,
  [5628] = 5628,
  [5629] = 5629,
  [5630] = 5630,
  [5631] = 5631,
  [5632] = 5632,
  [5633] = 5633,
  [5634] = 5626,
  [5635] = 5635,
  [5636] = 5622,
  [5637] = 5631,
  [5638] = 5638,
  [5639] = 5609,
  [5640] = 5640,
  [5641] = 5555,
  [5642] = 3890,
  [5643] = 5643,
  [5644] = 5576,
  [5645] = 5645,
  [5646] = 3494,
  [5647] = 5647,
  [5648] = 5557,
  [5649] = 5649,
  [5650] = 5650,
  [5651] = 5651,
  [5652] = 5612,
  [5653] = 3936,
  [5654] = 5558,
  [5655] = 5655,
  [5656] = 5656,
  [5657] = 5657,
  [5658] = 5550,
  [5659] = 3907,
  [5660] = 5550,
  [5661] = 5638,
  [5662] = 3900,
  [5663] = 5663,
  [5664] = 5630,
  [5665] = 5603,
  [5666] = 5565,
  [5667] = 3901,
  [5668] = 5553,
  [5669] = 5669,
  [5670] = 5629,
  [5671] = 3965,
  [5672] = 3973,
  [5673] = 5630,
  [5674] = 5633,
  [5675] = 5609,
  [5676] = 3911,
  [5677] = 4021,
  [5678] = 5678,
  [5679] = 5558,
  [5680] = 5635,
  [5681] = 5555,
  [5682] = 5655,
  [5683] = 5683,
  [5684] = 5684,
  [5685] = 5647,
  [5686] = 5638,
  [5687] = 5687,
  [5688] = 5625,
  [5689] = 5560,
  [5690] = 5554,
  [5691] = 3484,
  [5692] = 5692,
  [5693] = 5562,
  [5694] = 5694,
  [5695] = 5542,
  [5696] = 5635,
  [5697] = 5546,
  [5698] = 5555,
  [5699] = 5609,
  [5700] = 5700,
  [5701] = 5612,
  [5702] = 5593,
  [5703] = 5550,
  [5704] = 5544,
  [5705] = 5553,
  [5706] = 5706,
  [5707] = 5633,
  [5708] = 5541,
  [5709] = 5635,
  [5710] = 5557,
  [5711] = 5593,
  [5712] = 5638,
  [5713] = 5555,
  [5714] = 5599,
  [5715] = 5603,
  [5716] = 5557,
  [5717] = 5569,
  [5718] = 5556,
  [5719] = 5651,
  [5720] = 5611,
  [5721] = 5558,
  [5722] = 5616,
  [5723] = 5557,
  [5724] = 5625,
  [5725] = 5628,
  [5726] = 5622,
  [5727] = 5631,
  [5728] = 3960,
  [5729] = 5643,
  [5730] = 5645,
  [5731] = 5731,
  [5732] = 5574,
  [5733] = 5649,
  [5734] = 5611,
  [5735] = 5651,
  [5736] = 3987,
  [5737] = 3857,
  [5738] = 5738,
  [5739] = 3963,
  [5740] = 4002,
  [5741] = 5741,
  [5742] = 5541,
  [5743] = 5593,
  [5744] = 3934,
  [5745] = 5616,
  [5746] = 3285,
  [5747] = 5569,
  [5748] = 5560,
  [5749] = 5749,
  [5750] = 5656,
  [5751] = 5562,
  [5752] = 5752,
  [5753] = 5541,
  [5754] = 5593,
  [5755] = 5755,
  [5756] = 5756,
  [5757] = 5558,
  [5758] = 5546,
  [5759] = 5655,
  [5760] = 3486,
  [5761] = 5541,
  [5762] = 5593,
  [5763] = 5550,
  [5764] = 3923,
  [5765] = 5562,
  [5766] = 5633,
  [5767] = 5541,
  [5768] = 5593,
  [5769] = 5555,
  [5770] = 5643,
  [5771] = 5557,
  [5772] = 5651,
  [5773] = 5773,
  [5774] = 5541,
  [5775] = 5593,
  [5776] = 5558,
  [5777] = 5541,
  [5778] = 5593,
  [5779] = 5683,
  [5780] = 3928,
  [5781] = 5687,
  [5782] = 5782,
  [5783] = 5650,
  [5784] = 3967,
  [5785] = 5593,
  [5786] = 5731,
  [5787] = 5560,
  [5788] = 5692,
  [5789] = 5562,
  [5790] = 5541,
  [5791] = 5593,
  [5792] = 5738,
  [5793] = 5541,
  [5794] = 5593,
  [5795] = 3977,
  [5796] = 5541,
  [5797] = 5593,
  [5798] = 5546,
  [5799] = 5541,
  [5800] = 5593,
  [5801] = 5541,
  [5802] = 5593,
  [5803] = 3978,
  [5804] = 5541,
  [5805] = 5593,
  [5806] = 5541,
  [5807] = 5593,
  [5808] = 5541,
  [5809] = 5593,
  [5810] = 5541,
  [5811] = 5593,
  [5812] = 5541,
  [5813] = 5593,
  [5814] = 5541,
  [5815] = 5593,
  [5816] = 5816,
  [5817] = 5649,
  [5818] = 3291,
  [5819] = 5597,
  [5820] = 5551,
  [5821] = 5821,
  [5822] = 5541,
  [5823] = 5560,
  [5824] = 3313,
  [5825] = 5550,
  [5826] = 5650,
  [5827] = 5589,
  [5828] = 5606,
  [5829] = 3303,
  [5830] = 5553,
  [5831] = 5831,
  [5832] = 5618,
  [5833] = 5633,
  [5834] = 5555,
  [5835] = 5545,
  [5836] = 5553,
  [5837] = 5669,
  [5838] = 5557,
  [5839] = 5651,
  [5840] = 5558,
  [5841] = 5657,
  [5842] = 5651,
  [5843] = 5633,
  [5844] = 5612,
  [5845] = 5562,
  [5846] = 5547,
  [5847] = 5847,
  [5848] = 5546,
  [5849] = 5629,
  [5850] = 5816,
  [5851] = 5640,
  [5852] = 5550,
  [5853] = 5553,
  [5854] = 5831,
  [5855] = 5555,
  [5856] = 5856,
  [5857] = 5706,
  [5858] = 5545,
  [5859] = 5816,
  [5860] = 5610,
  [5861] = 5557,
  [5862] = 5831,
  [5863] = 5558,
  [5864] = 5545,
  [5865] = 5816,
  [5866] = 5628,
  [5867] = 5816,
  [5868] = 5831,
  [5869] = 5624,
  [5870] = 5545,
  [5871] = 5816,
  [5872] = 5645,
  [5873] = 5548,
  [5874] = 5831,
  [5875] = 5545,
  [5876] = 5816,
  [5877] = 5632,
  [5878] = 5878,
  [5879] = 5831,
  [5880] = 5601,
  [5881] = 5545,
  [5882] = 5816,
  [5883] = 5831,
  [5884] = 5545,
  [5885] = 5816,
  [5886] = 5562,
  [5887] = 5571,
  [5888] = 5831,
  [5889] = 5571,
  [5890] = 5545,
  [5891] = 5816,
  [5892] = 5562,
  [5893] = 5831,
  [5894] = 5831,
  [5895] = 5574,
  [5896] = 5545,
  [5897] = 5816,
  [5898] = 5831,
  [5899] = 5816,
  [5900] = 5831,
  [5901] = 5816,
  [5902] = 5831,
  [5903] = 5816,
  [5904] = 5831,
  [5905] = 5816,
  [5906] = 5831,
  [5907] = 5816,
  [5908] = 5831,
  [5909] = 5816,
  [5910] = 5831,
  [5911] = 5816,
  [5912] = 5816,
  [5913] = 5816,
  [5914] = 5816,
  [5915] = 5546,
  [5916] = 5551,
  [5917] = 5603,
  [5918] = 5878,
  [5919] = 5589,
  [5920] = 5920,
  [5921] = 5752,
  [5922] = 5756,
  [5923] = 5749,
  [5924] = 5550,
  [5925] = 5542,
  [5926] = 5549,
  [5927] = 5553,
  [5928] = 4268,
  [5929] = 5555,
  [5930] = 5930,
  [5931] = 5683,
  [5932] = 5687,
  [5933] = 5557,
  [5934] = 5558,
  [5935] = 5602,
  [5936] = 5596,
  [5937] = 5604,
  [5938] = 5741,
  [5939] = 5559,
  [5940] = 5576,
  [5941] = 5624,
  [5942] = 5942,
  [5943] = 5597,
  [5944] = 5599,
  [5945] = 5624,
  [5946] = 5946,
  [5947] = 5553,
  [5948] = 5948,
  [5949] = 5949,
  [5950] = 5948,
  [5951] = 5951,
  [5952] = 5952,
  [5953] = 5952,
  [5954] = 5954,
  [5955] = 5955,
  [5956] = 5956,
  [5957] = 5955,
  [5958] = 5958,
  [5959] = 5959,
  [5960] = 5960,
  [5961] = 5961,
  [5962] = 5962,
  [5963] = 5963,
  [5964] = 5964,
  [5965] = 5965,
  [5966] = 5966,
  [5967] = 5967,
  [5968] = 5968,
  [5969] = 5969,
  [5970] = 5970,
  [5971] = 5952,
  [5972] = 5972,
  [5973] = 5969,
  [5974] = 5974,
  [5975] = 5974,
  [5976] = 5961,
  [5977] = 5949,
  [5978] = 5978,
  [5979] = 5963,
  [5980] = 5980,
  [5981] = 5954,
  [5982] = 5982,
  [5983] = 5959,
  [5984] = 5984,
  [5985] = 5985,
  [5986] = 5986,
  [5987] = 5987,
  [5988] = 5958,
  [5989] = 5965,
  [5990] = 5966,
  [5991] = 5967,
  [5992] = 5968,
  [5993] = 5993,
  [5994] = 5994,
  [5995] = 5995,
  [5996] = 5949,
  [5997] = 5954,
  [5998] = 5998,
  [5999] = 5999,
  [6000] = 5959,
  [6001] = 5966,
  [6002] = 5967,
  [6003] = 5948,
  [6004] = 6004,
  [6005] = 5965,
  [6006] = 5966,
  [6007] = 5967,
  [6008] = 5968,
  [6009] = 5982,
  [6010] = 5968,
  [6011] = 5952,
  [6012] = 5961,
  [6013] = 6013,
  [6014] = 5974,
  [6015] = 5949,
  [6016] = 6016,
  [6017] = 5954,
  [6018] = 6018,
  [6019] = 5952,
  [6020] = 5966,
  [6021] = 5967,
  [6022] = 5968,
  [6023] = 6023,
  [6024] = 6024,
  [6025] = 5952,
  [6026] = 5949,
  [6027] = 6027,
  [6028] = 6028,
  [6029] = 5949,
  [6030] = 6030,
  [6031] = 6031,
  [6032] = 5954,
  [6033] = 5966,
  [6034] = 5967,
  [6035] = 5968,
  [6036] = 5978,
  [6037] = 5961,
  [6038] = 6016,
  [6039] = 6039,
  [6040] = 6040,
  [6041] = 5956,
  [6042] = 5954,
  [6043] = 6043,
  [6044] = 6028,
  [6045] = 5963,
  [6046] = 6046,
  [6047] = 6028,
  [6048] = 4051,
  [6049] = 5966,
  [6050] = 5967,
  [6051] = 5968,
  [6052] = 6039,
  [6053] = 6016,
  [6054] = 5985,
  [6055] = 6055,
  [6056] = 5961,
  [6057] = 6057,
  [6058] = 5960,
  [6059] = 5954,
  [6060] = 6060,
  [6061] = 5993,
  [6062] = 5954,
  [6063] = 6063,
  [6064] = 5966,
  [6065] = 5967,
  [6066] = 5968,
  [6067] = 5982,
  [6068] = 6068,
  [6069] = 5963,
  [6070] = 6070,
  [6071] = 5952,
  [6072] = 5954,
  [6073] = 5969,
  [6074] = 5966,
  [6075] = 5967,
  [6076] = 5968,
  [6077] = 5969,
  [6078] = 6078,
  [6079] = 6079,
  [6080] = 5984,
  [6081] = 5954,
  [6082] = 5966,
  [6083] = 5967,
  [6084] = 5968,
  [6085] = 6085,
  [6086] = 5982,
  [6087] = 6087,
  [6088] = 5954,
  [6089] = 5966,
  [6090] = 5967,
  [6091] = 5968,
  [6092] = 6092,
  [6093] = 6039,
  [6094] = 5954,
  [6095] = 5966,
  [6096] = 5967,
  [6097] = 5968,
  [6098] = 5969,
  [6099] = 6099,
  [6100] = 6100,
  [6101] = 5954,
  [6102] = 5966,
  [6103] = 5967,
  [6104] = 5968,
  [6105] = 5954,
  [6106] = 6106,
  [6107] = 5954,
  [6108] = 5966,
  [6109] = 5967,
  [6110] = 5968,
  [6111] = 6023,
  [6112] = 6112,
  [6113] = 5954,
  [6114] = 5966,
  [6115] = 5967,
  [6116] = 5968,
  [6117] = 6016,
  [6118] = 5961,
  [6119] = 5954,
  [6120] = 6030,
  [6121] = 4048,
  [6122] = 5954,
  [6123] = 5986,
  [6124] = 5986,
  [6125] = 5954,
  [6126] = 6024,
  [6127] = 6127,
  [6128] = 5954,
  [6129] = 5963,
  [6130] = 5986,
  [6131] = 5962,
  [6132] = 6078,
  [6133] = 6079,
  [6134] = 5984,
  [6135] = 5962,
  [6136] = 6078,
  [6137] = 5963,
  [6138] = 6078,
  [6139] = 6139,
  [6140] = 6023,
  [6141] = 6030,
  [6142] = 6142,
  [6143] = 6043,
  [6144] = 6144,
  [6145] = 5970,
  [6146] = 6146,
  [6147] = 5964,
  [6148] = 2100,
  [6149] = 6149,
  [6150] = 5982,
  [6151] = 6151,
  [6152] = 6016,
  [6153] = 6153,
  [6154] = 6154,
  [6155] = 6155,
  [6156] = 5955,
  [6157] = 5955,
  [6158] = 6158,
  [6159] = 5960,
  [6160] = 6023,
  [6161] = 6046,
  [6162] = 6162,
  [6163] = 6092,
  [6164] = 6164,
  [6165] = 6031,
  [6166] = 5960,
  [6167] = 6167,
  [6168] = 5978,
  [6169] = 5959,
  [6170] = 5952,
  [6171] = 6171,
  [6172] = 5961,
  [6173] = 6004,
  [6174] = 5986,
  [6175] = 6175,
  [6176] = 5961,
  [6177] = 4451,
  [6178] = 6178,
  [6179] = 5963,
  [6180] = 5959,
  [6181] = 6016,
  [6182] = 6142,
  [6183] = 6016,
  [6184] = 5993,
  [6185] = 5960,
  [6186] = 6030,
  [6187] = 6016,
  [6188] = 6018,
  [6189] = 6189,
  [6190] = 6057,
  [6191] = 5969,
  [6192] = 6016,
  [6193] = 6023,
  [6194] = 5969,
  [6195] = 6195,
  [6196] = 5949,
  [6197] = 5969,
  [6198] = 6178,
  [6199] = 5984,
  [6200] = 5995,
  [6201] = 6142,
  [6202] = 6202,
  [6203] = 5982,
  [6204] = 6204,
  [6205] = 6205,
  [6206] = 6027,
  [6207] = 6153,
  [6208] = 6164,
  [6209] = 6057,
  [6210] = 5959,
  [6211] = 5982,
  [6212] = 6070,
  [6213] = 6167,
  [6214] = 6214,
  [6215] = 5986,
  [6216] = 6167,
  [6217] = 5952,
  [6218] = 5969,
  [6219] = 5952,
  [6220] = 6220,
  [6221] = 5965,
  [6222] = 6070,
  [6223] = 6167,
  [6224] = 6024,
  [6225] = 1826,
  [6226] = 6175,
  [6227] = 6085,
  [6228] = 5969,
  [6229] = 5982,
  [6230] = 6164,
  [6231] = 5982,
  [6232] = 5961,
  [6233] = 6024,
  [6234] = 5952,
  [6235] = 6158,
  [6236] = 6085,
  [6237] = 5969,
  [6238] = 5963,
  [6239] = 5969,
  [6240] = 5951,
  [6241] = 5963,
  [6242] = 6099,
  [6243] = 5961,
  [6244] = 5999,
  [6245] = 5961,
  [6246] = 5963,
  [6247] = 6220,
  [6248] = 5960,
  [6249] = 6016,
  [6250] = 5974,
  [6251] = 5982,
  [6252] = 5949,
  [6253] = 5963,
  [6254] = 6254,
  [6255] = 5960,
  [6256] = 6085,
  [6257] = 5963,
  [6258] = 5952,
  [6259] = 5954,
  [6260] = 6260,
  [6261] = 5987,
  [6262] = 5961,
  [6263] = 6263,
  [6264] = 5964,
  [6265] = 5959,
  [6266] = 5965,
  [6267] = 6016,
  [6268] = 5963,
  [6269] = 5995,
  [6270] = 5948,
  [6271] = 6099,
  [6272] = 5952,
  [6273] = 5952,
  [6274] = 5961,
  [6275] = 5980,
  [6276] = 6155,
  [6277] = 6031,
  [6278] = 5963,
  [6279] = 5964,
  [6280] = 5965,
  [6281] = 5966,
  [6282] = 5966,
  [6283] = 5967,
  [6284] = 5968,
  [6285] = 5967,
  [6286] = 5968,
  [6287] = 6092,
  [6288] = 6028,
  [6289] = 6079,
  [6290] = 5955,
  [6291] = 5952,
  [6292] = 5961,
  [6293] = 6039,
  [6294] = 5963,
  [6295] = 5962,
  [6296] = 5960,
  [6297] = 6078,
  [6298] = 5952,
  [6299] = 6046,
  [6300] = 5982,
  [6301] = 5999,
  [6302] = 5961,
  [6303] = 5993,
  [6304] = 5961,
  [6305] = 6016,
  [6306] = 6164,
  [6307] = 6092,
  [6308] = 6254,
  [6309] = 5995,
  [6310] = 5952,
  [6311] = 6158,
  [6312] = 6178,
  [6313] = 6016,
  [6314] = 5978,
  [6315] = 4039,
  [6316] = 5963,
  [6317] = 6079,
  [6318] = 6100,
  [6319] = 5961,
  [6320] = 4045,
  [6321] = 6024,
  [6322] = 5963,
  [6323] = 6153,
  [6324] = 5963,
  [6325] = 6087,
  [6326] = 6043,
  [6327] = 5956,
  [6328] = 5962,
  [6329] = 5969,
  [6330] = 6046,
  [6331] = 6331,
  [6332] = 6332,
  [6333] = 6031,
  [6334] = 6023,
  [6335] = 5956,
  [6336] = 5974,
  [6337] = 6070,
  [6338] = 6338,
  [6339] = 6339,
  [6340] = 6340,
  [6341] = 6341,
  [6342] = 6342,
  [6343] = 6343,
  [6344] = 6344,
  [6345] = 6345,
  [6346] = 6346,
  [6347] = 6347,
  [6348] = 6348,
  [6349] = 6349,
  [6350] = 6350,
  [6351] = 6351,
  [6352] = 6347,
  [6353] = 6353,
  [6354] = 6354,
  [6355] = 6340,
  [6356] = 6356,
  [6357] = 6357,
  [6358] = 6358,
  [6359] = 6359,
  [6360] = 6360,
  [6361] = 6349,
  [6362] = 6362,
  [6363] = 6350,
  [6364] = 6364,
  [6365] = 6365,
  [6366] = 6366,
  [6367] = 6367,
  [6368] = 6368,
  [6369] = 6353,
  [6370] = 6370,
  [6371] = 6358,
  [6372] = 6359,
  [6373] = 6373,
  [6374] = 6370,
  [6375] = 6375,
  [6376] = 6376,
  [6377] = 6377,
  [6378] = 6378,
  [6379] = 6362,
  [6380] = 6380,
  [6381] = 6381,
  [6382] = 6348,
  [6383] = 6383,
  [6384] = 6384,
  [6385] = 6385,
  [6386] = 6367,
  [6387] = 6387,
  [6388] = 6384,
  [6389] = 6375,
  [6390] = 6366,
  [6391] = 6378,
  [6392] = 6392,
  [6393] = 6353,
  [6394] = 6394,
  [6395] = 6368,
  [6396] = 6358,
  [6397] = 6397,
  [6398] = 6398,
  [6399] = 6359,
  [6400] = 6400,
  [6401] = 6394,
  [6402] = 6402,
  [6403] = 6375,
  [6404] = 6404,
  [6405] = 6364,
  [6406] = 6406,
  [6407] = 6383,
  [6408] = 6406,
  [6409] = 6409,
  [6410] = 6410,
  [6411] = 6411,
  [6412] = 6412,
  [6413] = 6413,
  [6414] = 6402,
  [6415] = 6415,
  [6416] = 6416,
  [6417] = 6417,
  [6418] = 6418,
  [6419] = 6373,
  [6420] = 6420,
  [6421] = 6411,
  [6422] = 6422,
  [6423] = 6423,
  [6424] = 6406,
  [6425] = 6411,
  [6426] = 6413,
  [6427] = 6427,
  [6428] = 6428,
  [6429] = 6376,
  [6430] = 6377,
  [6431] = 6431,
  [6432] = 6432,
  [6433] = 6433,
  [6434] = 6434,
  [6435] = 6435,
  [6436] = 6436,
  [6437] = 6437,
  [6438] = 6438,
  [6439] = 6439,
  [6440] = 6413,
  [6441] = 6441,
  [6442] = 6415,
  [6443] = 6415,
  [6444] = 6444,
  [6445] = 6445,
  [6446] = 6383,
  [6447] = 6447,
  [6448] = 6448,
  [6449] = 6449,
  [6450] = 6450,
  [6451] = 6380,
  [6452] = 6452,
  [6453] = 6447,
  [6454] = 6387,
  [6455] = 6455,
  [6456] = 6344,
  [6457] = 6381,
  [6458] = 6458,
  [6459] = 6459,
  [6460] = 2644,
  [6461] = 6461,
  [6462] = 6365,
  [6463] = 6463,
  [6464] = 6448,
  [6465] = 6465,
  [6466] = 6348,
  [6467] = 6387,
  [6468] = 6404,
  [6469] = 6469,
  [6470] = 6470,
  [6471] = 6471,
  [6472] = 6365,
  [6473] = 2645,
  [6474] = 6474,
  [6475] = 6475,
  [6476] = 6476,
  [6477] = 6477,
  [6478] = 6478,
  [6479] = 6479,
  [6480] = 6480,
  [6481] = 6481,
  [6482] = 6482,
  [6483] = 6392,
  [6484] = 6343,
  [6485] = 6485,
  [6486] = 6486,
  [6487] = 6487,
  [6488] = 6488,
  [6489] = 6489,
  [6490] = 6354,
  [6491] = 6406,
  [6492] = 6378,
  [6493] = 6362,
  [6494] = 6474,
  [6495] = 6475,
  [6496] = 6366,
  [6497] = 6411,
  [6498] = 6485,
  [6499] = 6413,
  [6500] = 6441,
  [6501] = 6373,
  [6502] = 6376,
  [6503] = 6377,
  [6504] = 6504,
  [6505] = 6505,
  [6506] = 6487,
  [6507] = 6471,
  [6508] = 6508,
  [6509] = 6447,
  [6510] = 6510,
  [6511] = 6489,
  [6512] = 6504,
  [6513] = 6397,
  [6514] = 6398,
  [6515] = 6400,
  [6516] = 6397,
  [6517] = 6517,
  [6518] = 6517,
  [6519] = 6404,
  [6520] = 6348,
  [6521] = 6398,
  [6522] = 6463,
  [6523] = 6420,
  [6524] = 6400,
  [6525] = 6525,
  [6526] = 6422,
  [6527] = 6423,
  [6528] = 6368,
  [6529] = 6529,
  [6530] = 6364,
  [6531] = 6434,
  [6532] = 6434,
  [6533] = 6533,
  [6534] = 6433,
  [6535] = 6435,
  [6536] = 6536,
  [6537] = 6437,
  [6538] = 6438,
  [6539] = 6439,
  [6540] = 6540,
  [6541] = 6541,
  [6542] = 6542,
  [6543] = 6342,
  [6544] = 6445,
  [6545] = 6404,
  [6546] = 6546,
  [6547] = 6378,
  [6548] = 6367,
  [6549] = 6549,
  [6550] = 6384,
  [6551] = 6533,
  [6552] = 6510,
  [6553] = 6354,
  [6554] = 6554,
  [6555] = 6362,
  [6556] = 6366,
  [6557] = 6370,
  [6558] = 6432,
  [6559] = 6348,
  [6560] = 6373,
  [6561] = 6376,
  [6562] = 6377,
  [6563] = 6463,
  [6564] = 6346,
  [6565] = 6449,
  [6566] = 6367,
  [6567] = 6384,
  [6568] = 6471,
  [6569] = 6398,
  [6570] = 6400,
  [6571] = 6489,
  [6572] = 6409,
  [6573] = 6402,
  [6574] = 6404,
  [6575] = 6410,
  [6576] = 6474,
  [6577] = 6349,
  [6578] = 6367,
  [6579] = 6422,
  [6580] = 6580,
  [6581] = 6432,
  [6582] = 6450,
  [6583] = 6459,
  [6584] = 6463,
  [6585] = 6435,
  [6586] = 6437,
  [6587] = 6438,
  [6588] = 6439,
  [6589] = 6485,
  [6590] = 6590,
  [6591] = 6475,
  [6592] = 6384,
  [6593] = 6377,
  [6594] = 6594,
  [6595] = 6595,
  [6596] = 6596,
  [6597] = 6343,
  [6598] = 6354,
  [6599] = 6416,
  [6600] = 6362,
  [6601] = 6366,
  [6602] = 6415,
  [6603] = 6346,
  [6604] = 6373,
  [6605] = 6376,
  [6606] = 6377,
  [6607] = 6607,
  [6608] = 6485,
  [6609] = 6609,
  [6610] = 6471,
  [6611] = 6611,
  [6612] = 6398,
  [6613] = 6400,
  [6614] = 6420,
  [6615] = 6546,
  [6616] = 6455,
  [6617] = 6404,
  [6618] = 6417,
  [6619] = 6418,
  [6620] = 6487,
  [6621] = 6489,
  [6622] = 6422,
  [6623] = 6623,
  [6624] = 6504,
  [6625] = 6447,
  [6626] = 6554,
  [6627] = 2646,
  [6628] = 6435,
  [6629] = 6437,
  [6630] = 6438,
  [6631] = 6439,
  [6632] = 6432,
  [6633] = 6609,
  [6634] = 6554,
  [6635] = 2642,
  [6636] = 6517,
  [6637] = 6351,
  [6638] = 6638,
  [6639] = 6450,
  [6640] = 6640,
  [6641] = 6354,
  [6642] = 6422,
  [6643] = 6362,
  [6644] = 6366,
  [6645] = 6590,
  [6646] = 6349,
  [6647] = 6373,
  [6648] = 6376,
  [6649] = 6377,
  [6650] = 6447,
  [6651] = 6423,
  [6652] = 6580,
  [6653] = 6398,
  [6654] = 6400,
  [6655] = 6353,
  [6656] = 6358,
  [6657] = 6359,
  [6658] = 6404,
  [6659] = 6485,
  [6660] = 6344,
  [6661] = 6385,
  [6662] = 6422,
  [6663] = 6345,
  [6664] = 6381,
  [6665] = 6349,
  [6666] = 6487,
  [6667] = 6435,
  [6668] = 6437,
  [6669] = 6438,
  [6670] = 6439,
  [6671] = 6378,
  [6672] = 6504,
  [6673] = 6450,
  [6674] = 6452,
  [6675] = 6351,
  [6676] = 6517,
  [6677] = 6434,
  [6678] = 6678,
  [6679] = 6412,
  [6680] = 6354,
  [6681] = 6458,
  [6682] = 6362,
  [6683] = 6366,
  [6684] = 6412,
  [6685] = 6406,
  [6686] = 6373,
  [6687] = 6376,
  [6688] = 6377,
  [6689] = 6411,
  [6690] = 6413,
  [6691] = 6427,
  [6692] = 6398,
  [6693] = 6400,
  [6694] = 6415,
  [6695] = 6406,
  [6696] = 6411,
  [6697] = 6404,
  [6698] = 6367,
  [6699] = 6384,
  [6700] = 6431,
  [6701] = 6422,
  [6702] = 6413,
  [6703] = 6432,
  [6704] = 6459,
  [6705] = 6482,
  [6706] = 6435,
  [6707] = 6437,
  [6708] = 6438,
  [6709] = 6439,
  [6710] = 6447,
  [6711] = 6397,
  [6712] = 6465,
  [6713] = 6463,
  [6714] = 6714,
  [6715] = 6427,
  [6716] = 6716,
  [6717] = 6365,
  [6718] = 6354,
  [6719] = 6415,
  [6720] = 6366,
  [6721] = 6435,
  [6722] = 6345,
  [6723] = 6373,
  [6724] = 6376,
  [6725] = 6377,
  [6726] = 6554,
  [6727] = 6447,
  [6728] = 6428,
  [6729] = 6398,
  [6730] = 6400,
  [6731] = 6731,
  [6732] = 6412,
  [6733] = 6436,
  [6734] = 6404,
  [6735] = 6596,
  [6736] = 6433,
  [6737] = 6349,
  [6738] = 6422,
  [6739] = 6364,
  [6740] = 6415,
  [6741] = 6435,
  [6742] = 6350,
  [6743] = 6435,
  [6744] = 6437,
  [6745] = 6438,
  [6746] = 6439,
  [6747] = 6485,
  [6748] = 6487,
  [6749] = 6474,
  [6750] = 6504,
  [6751] = 6475,
  [6752] = 6485,
  [6753] = 6487,
  [6754] = 6452,
  [6755] = 6354,
  [6756] = 6517,
  [6757] = 6437,
  [6758] = 6438,
  [6759] = 6373,
  [6760] = 6376,
  [6761] = 6377,
  [6762] = 6510,
  [6763] = 6489,
  [6764] = 6504,
  [6765] = 6398,
  [6766] = 6400,
  [6767] = 6439,
  [6768] = 6353,
  [6769] = 6517,
  [6770] = 6404,
  [6771] = 6358,
  [6772] = 6359,
  [6773] = 6434,
  [6774] = 6422,
  [6775] = 6375,
  [6776] = 6351,
  [6777] = 6368,
  [6778] = 6367,
  [6779] = 6435,
  [6780] = 6437,
  [6781] = 6438,
  [6782] = 6439,
  [6783] = 6364,
  [6784] = 6476,
  [6785] = 6470,
  [6786] = 6380,
  [6787] = 6434,
  [6788] = 6596,
  [6789] = 6341,
  [6790] = 6346,
  [6791] = 6354,
  [6792] = 6384,
  [6793] = 6342,
  [6794] = 6377,
  [6795] = 6432,
  [6796] = 6404,
  [6797] = 6381,
  [6798] = 6351,
  [6799] = 6415,
  [6800] = 6422,
  [6801] = 6434,
  [6802] = 6435,
  [6803] = 6437,
  [6804] = 6438,
  [6805] = 6439,
  [6806] = 6806,
  [6807] = 6477,
  [6808] = 6433,
  [6809] = 6809,
  [6810] = 6370,
  [6811] = 6447,
  [6812] = 6487,
  [6813] = 6479,
  [6814] = 6814,
  [6815] = 6815,
  [6816] = 6437,
  [6817] = 6438,
  [6818] = 6439,
  [6819] = 6378,
  [6820] = 6365,
  [6821] = 4276,
  [6822] = 6485,
  [6823] = 6485,
  [6824] = 6487,
  [6825] = 6590,
  [6826] = 6437,
  [6827] = 6438,
  [6828] = 6439,
  [6829] = 6504,
  [6830] = 6830,
  [6831] = 6348,
  [6832] = 6427,
  [6833] = 6517,
  [6834] = 6343,
  [6835] = 6367,
  [6836] = 6437,
  [6837] = 6438,
  [6838] = 6439,
  [6839] = 6434,
  [6840] = 6428,
  [6841] = 6488,
  [6842] = 6367,
  [6843] = 6412,
  [6844] = 6406,
  [6845] = 6384,
  [6846] = 6437,
  [6847] = 6438,
  [6848] = 6439,
  [6849] = 6411,
  [6850] = 6413,
  [6851] = 6367,
  [6852] = 6427,
  [6853] = 6428,
  [6854] = 6432,
  [6855] = 6362,
  [6856] = 6437,
  [6857] = 6438,
  [6858] = 6439,
  [6859] = 6402,
  [6860] = 6436,
  [6861] = 6384,
  [6862] = 6415,
  [6863] = 6415,
  [6864] = 6814,
  [6865] = 6342,
  [6866] = 6437,
  [6867] = 6438,
  [6868] = 6439,
  [6869] = 6611,
  [6870] = 6870,
  [6871] = 6716,
  [6872] = 6447,
  [6873] = 6504,
  [6874] = 1480,
  [6875] = 6875,
  [6876] = 6487,
  [6877] = 6448,
  [6878] = 6370,
  [6879] = 6485,
  [6880] = 6449,
  [6881] = 6487,
  [6882] = 6342,
  [6883] = 6883,
  [6884] = 6884,
  [6885] = 6481,
  [6886] = 6339,
  [6887] = 6640,
  [6888] = 6402,
  [6889] = 6889,
  [6890] = 6504,
  [6891] = 6611,
  [6892] = 6517,
  [6893] = 6350,
  [6894] = 6434,
  [6895] = 6447,
  [6896] = 6432,
  [6897] = 6459,
  [6898] = 6463,
  [6899] = 6367,
  [6900] = 6346,
  [6901] = 6384,
  [6902] = 6716,
  [6903] = 6343,
  [6904] = 6432,
  [6905] = 6905,
  [6906] = 6368,
  [6907] = 6907,
  [6908] = 6434,
  [6909] = 6368,
  [6910] = 6415,
  [6911] = 6471,
  [6912] = 6912,
  [6913] = 6348,
  [6914] = 6471,
  [6915] = 6475,
  [6916] = 6447,
  [6917] = 6830,
  [6918] = 6918,
  [6919] = 6546,
  [6920] = 6465,
  [6921] = 6367,
  [6922] = 6510,
  [6923] = 6923,
  [6924] = 6505,
  [6925] = 6925,
  [6926] = 6907,
  [6927] = 6546,
  [6928] = 6384,
  [6929] = 6554,
  [6930] = 6384,
  [6931] = 6368,
  [6932] = 6436,
  [6933] = 6351,
  [6934] = 6338,
  [6935] = 6485,
  [6936] = 6554,
  [6937] = 6937,
  [6938] = 6938,
  [6939] = 6487,
  [6940] = 6525,
  [6941] = 6349,
  [6942] = 6533,
  [6943] = 6380,
  [6944] = 6504,
  [6945] = 6925,
  [6946] = 1474,
  [6947] = 6364,
  [6948] = 6353,
  [6949] = 6517,
  [6950] = 6480,
  [6951] = 6541,
  [6952] = 6918,
  [6953] = 6358,
  [6954] = 6536,
  [6955] = 6359,
  [6956] = 6434,
  [6957] = 6415,
  [6958] = 6958,
  [6959] = 6381,
  [6960] = 6486,
  [6961] = 6357,
  [6962] = 6870,
  [6963] = 6542,
  [6964] = 6367,
  [6965] = 6378,
  [6966] = 6416,
  [6967] = 6384,
  [6968] = 6448,
  [6969] = 6449,
  [6970] = 6432,
  [6971] = 6412,
  [6972] = 6406,
  [6973] = 6411,
  [6974] = 6413,
  [6975] = 6427,
  [6976] = 6415,
  [6977] = 6422,
  [6978] = 6481,
  [6979] = 6365,
  [6980] = 6517,
  [6981] = 6415,
  [6982] = 6437,
  [6983] = 6381,
  [6984] = 6447,
  [6985] = 6607,
  [6986] = 6438,
  [6987] = 6366,
  [6988] = 6365,
  [6989] = 6439,
  [6990] = 6461,
  [6991] = 6409,
  [6992] = 6402,
  [6993] = 6447,
  [6994] = 6508,
  [6995] = 6469,
  [6996] = 6549,
  [6997] = 6997,
  [6998] = 6470,
  [6999] = 6410,
  [7000] = 6474,
  [7001] = 6475,
  [7002] = 6485,
  [7003] = 6487,
  [7004] = 6609,
  [7005] = 6489,
  [7006] = 6504,
  [7007] = 6342,
  [7008] = 6517,
  [7009] = 6447,
  [7010] = 6402,
  [7011] = 6368,
  [7012] = 6432,
  [7013] = 6364,
  [7014] = 6485,
  [7015] = 6434,
  [7016] = 7016,
  [7017] = 6938,
  [7018] = 6339,
  [7019] = 6487,
  [7020] = 6342,
  [7021] = 6504,
  [7022] = 6365,
  [7023] = 6517,
  [7024] = 6477,
  [7025] = 6370,
  [7026] = 6455,
  [7027] = 6348,
  [7028] = 6474,
  [7029] = 6475,
  [7030] = 6489,
  [7031] = 6731,
  [7032] = 6434,
  [7033] = 6367,
  [7034] = 6463,
  [7035] = 6384,
  [7036] = 6471,
  [7037] = 6349,
  [7038] = 6367,
  [7039] = 6402,
  [7040] = 6384,
  [7041] = 6432,
  [7042] = 6463,
  [7043] = 6937,
  [7044] = 7044,
  [7045] = 6958,
  [7046] = 6432,
  [7047] = 6343,
  [7048] = 6883,
  [7049] = 6463,
  [7050] = 6481,
  [7051] = 6889,
  [7052] = 6415,
  [7053] = 6485,
  [7054] = 6905,
  [7055] = 7055,
  [7056] = 6346,
  [7057] = 6471,
  [7058] = 6479,
  [7059] = 3612,
  [7060] = 6907,
  [7061] = 6447,
  [7062] = 6546,
  [7063] = 6343,
  [7064] = 6815,
  [7065] = 6423,
  [7066] = 6368,
  [7067] = 6554,
  [7068] = 7068,
  [7069] = 6505,
  [7070] = 6485,
  [7071] = 6344,
  [7072] = 6351,
  [7073] = 6373,
  [7074] = 6487,
  [7075] = 6504,
  [7076] = 6434,
  [7077] = 6937,
  [7078] = 6384,
  [7079] = 6349,
  [7080] = 6345,
  [7081] = 6480,
  [7082] = 6541,
  [7083] = 6918,
  [7084] = 6432,
  [7085] = 6489,
  [7086] = 6638,
  [7087] = 6474,
  [7088] = 6475,
  [7089] = 6353,
  [7090] = 6358,
  [7091] = 6907,
  [7092] = 6359,
  [7093] = 6349,
  [7094] = 6471,
  [7095] = 6485,
  [7096] = 6505,
  [7097] = 6487,
  [7098] = 6378,
  [7099] = 6400,
  [7100] = 6485,
  [7101] = 6937,
  [7102] = 6487,
  [7103] = 6504,
  [7104] = 6432,
  [7105] = 6480,
  [7106] = 6541,
  [7107] = 6918,
  [7108] = 6488,
  [7109] = 6485,
  [7110] = 6458,
  [7111] = 6487,
  [7112] = 6406,
  [7113] = 6411,
  [7114] = 6413,
  [7115] = 6415,
  [7116] = 6474,
  [7117] = 6432,
  [7118] = 6510,
  [7119] = 6937,
  [7120] = 6510,
  [7121] = 6480,
  [7122] = 6541,
  [7123] = 6918,
  [7124] = 3652,
  [7125] = 1486,
  [7126] = 6489,
  [7127] = 6340,
  [7128] = 6351,
  [7129] = 6447,
  [7130] = 6504,
  [7131] = 6883,
  [7132] = 6485,
  [7133] = 6365,
  [7134] = 6937,
  [7135] = 6517,
  [7136] = 6480,
  [7137] = 6541,
  [7138] = 6918,
  [7139] = 6487,
  [7140] = 6365,
  [7141] = 7141,
  [7142] = 6640,
  [7143] = 6432,
  [7144] = 6459,
  [7145] = 6368,
  [7146] = 6504,
  [7147] = 6364,
  [7148] = 6489,
  [7149] = 6370,
  [7150] = 6480,
  [7151] = 6541,
  [7152] = 6918,
  [7153] = 6434,
  [7154] = 4107,
  [7155] = 6417,
  [7156] = 6485,
  [7157] = 6487,
  [7158] = 7158,
  [7159] = 6489,
  [7160] = 6418,
  [7161] = 6342,
  [7162] = 6504,
  [7163] = 6487,
  [7164] = 6480,
  [7165] = 6541,
  [7166] = 6918,
  [7167] = 6815,
  [7168] = 6883,
  [7169] = 6517,
  [7170] = 6481,
  [7171] = 6609,
  [7172] = 6471,
  [7173] = 6368,
  [7174] = 6504,
  [7175] = 6339,
  [7176] = 6640,
  [7177] = 6370,
  [7178] = 6480,
  [7179] = 6541,
  [7180] = 6918,
  [7181] = 6432,
  [7182] = 6364,
  [7183] = 6348,
  [7184] = 6434,
  [7185] = 6394,
  [7186] = 6434,
  [7187] = 6517,
  [7188] = 6342,
  [7189] = 6367,
  [7190] = 6384,
  [7191] = 6432,
  [7192] = 6480,
  [7193] = 6541,
  [7194] = 6918,
  [7195] = 6546,
  [7196] = 6474,
  [7197] = 6475,
  [7198] = 6370,
  [7199] = 6480,
  [7200] = 6541,
  [7201] = 6918,
  [7202] = 6402,
  [7203] = 6889,
  [7204] = 6480,
  [7205] = 6541,
  [7206] = 6918,
  [7207] = 6432,
  [7208] = 7208,
  [7209] = 6480,
  [7210] = 6541,
  [7211] = 6918,
  [7212] = 6489,
  [7213] = 6450,
  [7214] = 6480,
  [7215] = 6541,
  [7216] = 6918,
  [7217] = 6452,
  [7218] = 7218,
  [7219] = 6480,
  [7220] = 6541,
  [7221] = 6918,
  [7222] = 6353,
  [7223] = 6458,
  [7224] = 6480,
  [7225] = 6541,
  [7226] = 6918,
  [7227] = 6459,
  [7228] = 6465,
  [7229] = 6480,
  [7230] = 6541,
  [7231] = 6918,
  [7232] = 6463,
  [7233] = 6463,
  [7234] = 7208,
  [7235] = 6385,
  [7236] = 6471,
  [7237] = 6339,
  [7238] = 6348,
  [7239] = 6368,
  [7240] = 6340,
  [7241] = 7241,
  [7242] = 6368,
  [7243] = 7044,
  [7244] = 6367,
  [7245] = 6384,
  [7246] = 6883,
  [7247] = 6402,
  [7248] = 7248,
  [7249] = 6470,
  [7250] = 6432,
  [7251] = 6463,
  [7252] = 6883,
  [7253] = 6477,
  [7254] = 6607,
  [7255] = 6875,
  [7256] = 7208,
  [7257] = 6343,
  [7258] = 6358,
  [7259] = 6349,
  [7260] = 6343,
  [7261] = 6431,
  [7262] = 6482,
  [7263] = 6471,
  [7264] = 6596,
  [7265] = 6341,
  [7266] = 6546,
  [7267] = 6525,
  [7268] = 6364,
  [7269] = 7269,
  [7270] = 6353,
  [7271] = 6339,
  [7272] = 6554,
  [7273] = 7241,
  [7274] = 6471,
  [7275] = 6351,
  [7276] = 6623,
  [7277] = 7241,
  [7278] = 6358,
  [7279] = 7279,
  [7280] = 7241,
  [7281] = 7241,
  [7282] = 7241,
  [7283] = 7241,
  [7284] = 7241,
  [7285] = 7241,
  [7286] = 7241,
  [7287] = 7241,
  [7288] = 7241,
  [7289] = 7241,
  [7290] = 7241,
  [7291] = 7241,
  [7292] = 7241,
  [7293] = 6478,
  [7294] = 6349,
  [7295] = 6474,
  [7296] = 6343,
  [7297] = 6475,
  [7298] = 6353,
  [7299] = 6554,
  [7300] = 6358,
  [7301] = 6640,
  [7302] = 6359,
  [7303] = 6714,
  [7304] = 6359,
  [7305] = 6378,
  [7306] = 6546,
  [7307] = 6434,
  [7308] = 6351,
  [7309] = 6354,
  [7310] = 7241,
  [7311] = 6406,
  [7312] = 6411,
  [7313] = 6413,
  [7314] = 6415,
  [7315] = 6714,
  [7316] = 6716,
  [7317] = 6714,
  [7318] = 6359,
  [7319] = 6554,
  [7320] = 6517,
  [7321] = 6447,
  [7322] = 6370,
  [7323] = 7323,
  [7324] = 6474,
  [7325] = 6475,
  [7326] = 6463,
  [7327] = 6392,
  [7328] = 6609,
  [7329] = 6474,
  [7330] = 6475,
  [7331] = 6376,
  [7332] = 2643,
  [7333] = 6884,
  [7334] = 6455,
  [7335] = 6474,
  [7336] = 6475,
  [7337] = 6485,
  [7338] = 6487,
  [7339] = 6489,
  [7340] = 6504,
  [7341] = 6342,
  [7342] = 6517,
  [7343] = 6354,
  [7344] = 6481,
  [7345] = 6349,
  [7346] = 6378,
  [7347] = 6398,
  [7348] = 7348,
  [7349] = 7349,
};

static const TSSymbol ts_supertype_symbols[SUPERTYPE_COUNT] = {
  sym__binding_pattern,
  sym__binding_pattern_no_exn,
  sym__class_expression,
  sym__class_field,
  sym__class_field_specification,
  sym__class_type,
  sym__constant,
  sym__delimited_binding_pattern,
  sym__delimited_expression,
  sym__delimited_pattern,
  sym__delimited_type,
  sym__effect_binding_pattern,
  sym__effect_pattern,
  sym__expression,
  sym__infix_operator,
  sym__local_structure_item,
  sym__module_expression,
  sym__module_type,
  sym__pattern,
  sym__polymorphic_type,
  sym__sequence_expression,
  sym__signature_item,
  sym__signed_constant,
  sym__simple_binding_pattern,
  sym__simple_class_expression,
  sym__simple_class_type,
  sym__simple_expression,
  sym__simple_module_expression,
  sym__simple_pattern,
  sym__simple_type,
  sym__structure_item,
  sym__type,
};

static const TSMapSlice ts_supertype_map_slices[] = {
  [sym__binding_pattern] = {.index = 0, .length = 14},
  [sym__binding_pattern_no_exn] = {.index = 14, .length = 10},
  [sym__class_expression] = {.index = 24, .length = 5},
  [sym__class_field] = {.index = 29, .length = 7},
  [sym__class_field_specification] = {.index = 36, .length = 6},
  [sym__class_type] = {.index = 42, .length = 2},
  [sym__constant] = {.index = 44, .length = 5},
  [sym__delimited_binding_pattern] = {.index = 49, .length = 9},
  [sym__delimited_expression] = {.index = 58, .length = 7},
  [sym__delimited_pattern] = {.index = 65, .length = 5},
  [sym__delimited_type] = {.index = 70, .length = 3},
  [sym__effect_binding_pattern] = {.index = 73, .length = 7},
  [sym__effect_pattern] = {.index = 80, .length = 4},
  [sym__expression] = {.index = 84, .length = 17},
  [sym__infix_operator] = {.index = 101, .length = 8},
  [sym__local_structure_item] = {.index = 109, .length = 11},
  [sym__module_expression] = {.index = 120, .length = 5},
  [sym__module_type] = {.index = 125, .length = 8},
  [sym__pattern] = {.index = 133, .length = 8},
  [sym__polymorphic_type] = {.index = 141, .length = 2},
  [sym__sequence_expression] = {.index = 143, .length = 2},
  [sym__signature_item] = {.index = 145, .length = 13},
  [sym__signed_constant] = {.index = 158, .length = 2},
  [sym__simple_binding_pattern] = {.index = 160, .length = 14},
  [sym__simple_class_expression] = {.index = 174, .length = 7},
  [sym__simple_class_type] = {.index = 181, .length = 6},
  [sym__simple_expression] = {.index = 187, .length = 19},
  [sym__simple_module_expression] = {.index = 206, .length = 5},
  [sym__simple_pattern] = {.index = 211, .length = 12},
  [sym__simple_type] = {.index = 223, .length = 9},
  [sym__structure_item] = {.index = 232, .length = 4},
  [sym__type] = {.index = 236, .length = 5},
};

static const TSSymbol ts_supertype_map_entries[] = {
  [0] =
    sym__effect_binding_pattern,
    sym__or_binding_pattern_no_exn_anonymous,
    sym__tuple_binding_pattern_no_exn,
    sym_alias_binding_pattern,
    sym_alias_binding_pattern_no_exn,
    sym_alias_pattern,
    sym_cons_binding_pattern,
    sym_cons_binding_pattern_no_exn,
    sym_cons_pattern,
    sym_effect_binding_pattern,
    sym_effect_pattern,
    sym_exception_binding_pattern,
    sym_exception_pattern,
    sym_range_pattern,
  [14] =
    sym__effect_binding_pattern,
    sym__or_binding_pattern_no_exn_anonymous,
    sym__tuple_binding_pattern_no_exn,
    sym_alias_binding_pattern,
    sym_alias_binding_pattern_no_exn,
    sym_alias_pattern,
    sym_cons_binding_pattern,
    sym_cons_binding_pattern_no_exn,
    sym_cons_pattern,
    sym_range_pattern,
  [24] =
    sym__simple_class_expression,
    sym_class_application,
    sym_class_function,
    sym_let_class_expression,
    sym_let_open_class_expression,
  [29] =
    sym_class_initializer,
    sym_inheritance_definition,
    sym_instance_variable_definition,
    sym_item_extension,
    sym_method_definition,
    sym_quoted_item_extension,
    sym_type_parameter_constraint,
  [36] =
    sym_inheritance_specification,
    sym_instance_variable_specification,
    sym_item_extension,
    sym_method_specification,
    sym_quoted_item_extension,
    sym_type_parameter_constraint,
  [42] =
    sym__simple_class_type,
    sym_class_function_type,
  [44] =
    sym_boolean,
    sym_character,
    sym_number,
    sym_quoted_string,
    sym_string,
  [49] =
    sym_array_binding_pattern,
    sym_array_pattern,
    sym_list_binding_pattern,
    sym_list_pattern,
    sym_parenthesized_binding_pattern,
    sym_parenthesized_pattern,
    sym_record_binding_pattern,
    sym_record_pattern,
    sym_unit,
  [58] =
    sym_array_expression,
    sym_list_expression,
    sym_object_copy_expression,
    sym_package_expression,
    sym_parenthesized_expression,
    sym_record_expression,
    sym_unit,
  [65] =
    sym_array_pattern,
    sym_list_pattern,
    sym_parenthesized_pattern,
    sym_record_pattern,
    sym_unit,
  [70] =
    sym_package_type,
    sym_parenthesized_type,
    sym_polymorphic_variant_type,
  [73] =
    sym__simple_binding_pattern,
    sym_constructor_binding_pattern,
    sym_constructor_pattern,
    sym_lazy_binding_pattern,
    sym_lazy_pattern,
    sym_tag_binding_pattern,
    sym_tag_pattern,
  [80] =
    sym__simple_pattern,
    sym_constructor_pattern,
    sym_lazy_pattern,
    sym_tag_pattern,
  [84] =
    alias_sym_tuple_expression,
    sym__simple_expression,
    sym_application_expression,
    sym_assert_expression,
    sym_cons_expression,
    sym_for_expression,
    sym_fun_expression,
    sym_function_expression,
    sym_if_expression,
    sym_infix_expression,
    sym_lazy_expression,
    sym_let_expression,
    sym_match_expression,
    sym_set_expression,
    sym_sign_expression,
    sym_try_expression,
    sym_while_expression,
  [101] =
    sym_add_operator,
    sym_and_operator,
    sym_assign_operator,
    sym_concat_operator,
    sym_mult_operator,
    sym_or_operator,
    sym_pow_operator,
    sym_rel_operator,
  [109] =
    sym_class_definition,
    sym_class_type_definition,
    sym_exception_definition,
    sym_external,
    sym_floating_attribute,
    sym_item_extension,
    sym_module_definition,
    sym_module_type_definition,
    sym_open_module,
    sym_quoted_item_extension,
    sym_type_definition,
  [120] =
    sym__simple_module_expression,
    sym_functor,
    sym_module_application,
    sym_module_path,
    sym_structure,
  [125] =
    sym_extension,
    sym_functor_type,
    sym_module_type_constraint,
    sym_module_type_of,
    sym_module_type_path,
    sym_parenthesized_module_type,
    sym_quoted_extension,
    sym_signature,
  [133] =
    sym__effect_pattern,
    sym__or_binding_pattern_no_exn_anonymous,
    sym__tuple_binding_pattern_no_exn,
    sym_alias_pattern,
    sym_cons_pattern,
    sym_effect_pattern,
    sym_exception_pattern,
    sym_range_pattern,
  [141] =
    sym__type,
    sym_polymorphic_type,
  [143] =
    alias_sym_sequence_expression,
    sym__expression,
  [145] =
    sym_class_definition,
    sym_class_type_definition,
    sym_exception_definition,
    sym_external,
    sym_floating_attribute,
    sym_include_module_type,
    sym_item_extension,
    sym_module_definition,
    sym_module_type_definition,
    sym_open_module,
    sym_quoted_item_extension,
    sym_type_definition,
    sym_value_specification,
  [158] =
    sym__constant,
    sym_signed_number,
  [160] =
    alias_sym_value_name,
    sym__delimited_binding_pattern,
    sym__signed_constant,
    sym_constructor_path,
    sym_extension,
    sym_local_open_binding_pattern,
    sym_local_open_pattern,
    sym_package_pattern,
    sym_parenthesized_operator,
    sym_polymorphic_variant_pattern,
    sym_quoted_extension,
    sym_tag,
    sym_typed_binding_pattern,
    sym_typed_pattern,
  [174] =
    sym_class_path,
    sym_extension,
    sym_instantiated_class,
    sym_object_expression,
    sym_parenthesized_class_expression,
    sym_quoted_extension,
    sym_typed_class_expression,
  [181] =
    sym_class_body_type,
    sym_class_type_path,
    sym_extension,
    sym_instantiated_class_type,
    sym_let_open_class_type,
    sym_quoted_extension,
  [187] =
    sym__constant,
    sym__delimited_expression,
    sym_array_get_expression,
    sym_bigarray_get_expression,
    sym_constructor_path,
    sym_extension,
    sym_field_get_expression,
    sym_hash_expression,
    sym_local_open_expression,
    sym_method_invocation,
    sym_new_expression,
    sym_object_expression,
    sym_ocamlyacc_value,
    sym_prefix_expression,
    sym_quoted_extension,
    sym_string_get_expression,
    sym_tag,
    sym_typed_expression,
    sym_value_path,
  [206] =
    sym_extension,
    sym_packed_module,
    sym_parenthesized_module_expression,
    sym_quoted_extension,
    sym_typed_module_expression,
  [211] =
    alias_sym_value_pattern,
    sym__delimited_pattern,
    sym__signed_constant,
    sym_constructor_path,
    sym_extension,
    sym_local_open_pattern,
    sym_package_pattern,
    sym_parenthesized_operator,
    sym_polymorphic_variant_pattern,
    sym_quoted_extension,
    sym_tag,
    sym_typed_pattern,
  [223] =
    sym__delimited_type,
    sym_constructed_type,
    sym_extension,
    sym_hash_type,
    sym_local_open_type,
    sym_object_type,
    sym_quoted_extension,
    sym_type_constructor_path,
    sym_type_variable,
  [232] =
    sym__local_structure_item,
    sym_include_module,
    sym_value_definition,
    sym_value_specification,
  [236] =
    sym__labeled_tuple_type,
    sym__proper_tuple_type,
    sym__simple_type,
    sym_aliased_type,
    sym_function_type,
};

static const TSCharacterRange sym_conversion_specification_character_set_6[] = {
  {'!', '!'}, {'%', '%'}, {',', ','}, {'@', '@'}, {'B', 'C'}, {'E', 'H'}, {'L', 'L'}, {'N', 'N'},
  {'S', 'S'}, {'X', 'X'}, {'a', 'i'}, {'l', 'l'}, {'n', 'o'}, {'s', 'u'}, {'x', 'x'},
};

static const TSCharacterRange sym_conversion_specification_character_set_7[] = {
  {'!', '!'}, {'%', '%'}, {',', ','}, {'0', '9'}, {'@', '@'}, {'B', 'C'}, {'E', 'H'}, {'L', 'L'},
  {'N', 'N'}, {'S', 'S'}, {'X', 'X'}, {'a', 'i'}, {'l', 'l'}, {'n', 'o'}, {'s', 'u'}, {'x', 'x'},
};

static const TSCharacterRange sym_prefix_operator_character_set_1[] = {
  {'!', '!'}, {'#', '&'}, {'*', '+'}, {'-', '/'}, {':', ':'}, {'<', '@'}, {'^', '^'}, {'|', '|'},
  {'~', '~'},
};

static const TSCharacterRange aux_sym_pow_operator_token1_character_set_1[] = {
  {'!', '!'}, {'$', '&'}, {'*', '+'}, {'-', '/'}, {':', ':'}, {'<', '@'}, {'^', '^'}, {'|', '|'},
  {'~', '~'},
};

static const TSCharacterRange sym_indexing_operator_character_set_1[] = {
  {'!', '!'}, {'$', '&'}, {'*', '+'}, {'-', '-'}, {'/', '/'}, {':', ':'}, {'=', '@'}, {'^', '^'},
  {'|', '|'},
};

static const TSCharacterRange sym_let_operator_character_set_1[] = {
  {'$', '$'}, {'&', '&'}, {'*', '+'}, {'-', '-'}, {'/', '/'}, {'<', '>'}, {'@', '@'}, {'^', '^'},
  {'|', '|'},
};

static const TSCharacterRange sym__lowercase_identifier_character_set_2[] = {
  {'\'', '\''}, {'0', '9'}, {'A', 'Z'}, {'_', '_'}, {'a', 'z'}, {0xaa, 0xaa}, {0xb5, 0xb5}, {0xb7, 0xb7},
  {0xba, 0xba}, {0xc0, 0xd6}, {0xd8, 0xf6}, {0xf8, 0x2c1}, {0x2c6, 0x2d1}, {0x2e0, 0x2e4}, {0x2ec, 0x2ec}, {0x2ee, 0x2ee},
  {0x300, 0x374}, {0x376, 0x377}, {0x37b, 0x37d}, {0x37f, 0x37f}, {0x386, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x3a1}, {0x3a3, 0x3f5},
  {0x3f7, 0x481}, {0x483, 0x487}, {0x48a, 0x52f}, {0x531, 0x556}, {0x559, 0x559}, {0x560, 0x588}, {0x591, 0x5bd}, {0x5bf, 0x5bf},
  {0x5c1, 0x5c2}, {0x5c4, 0x5c5}, {0x5c7, 0x5c7}, {0x5d0, 0x5ea}, {0x5ef, 0x5f2}, {0x610, 0x61a}, {0x620, 0x669}, {0x66e, 0x6d3},
  {0x6d5, 0x6dc}, {0x6df, 0x6e8}, {0x6ea, 0x6fc}, {0x6ff, 0x6ff}, {0x710, 0x74a}, {0x74d, 0x7b1}, {0x7c0, 0x7f5}, {0x7fa, 0x7fa},
  {0x7fd, 0x7fd}, {0x800, 0x82d}, {0x840, 0x85b}, {0x860, 0x86a}, {0x870, 0x887}, {0x889, 0x88e}, {0x897, 0x8e1}, {0x8e3, 0x963},
  {0x966, 0x96f}, {0x971, 0x983}, {0x985, 0x98c}, {0x98f, 0x990}, {0x993, 0x9a8}, {0x9aa, 0x9b0}, {0x9b2, 0x9b2}, {0x9b6, 0x9b9},
  {0x9bc, 0x9c4}, {0x9c7, 0x9c8}, {0x9cb, 0x9ce}, {0x9d7, 0x9d7}, {0x9dc, 0x9dd}, {0x9df, 0x9e3}, {0x9e6, 0x9f1}, {0x9fc, 0x9fc},
  {0x9fe, 0x9fe}, {0xa01, 0xa03}, {0xa05, 0xa0a}, {0xa0f, 0xa10}, {0xa13, 0xa28}, {0xa2a, 0xa30}, {0xa32, 0xa33}, {0xa35, 0xa36},
  {0xa38, 0xa39}, {0xa3c, 0xa3c}, {0xa3e, 0xa42}, {0xa47, 0xa48}, {0xa4b, 0xa4d}, {0xa51, 0xa51}, {0xa59, 0xa5c}, {0xa5e, 0xa5e},
  {0xa66, 0xa75}, {0xa81, 0xa83}, {0xa85, 0xa8d}, {0xa8f, 0xa91}, {0xa93, 0xaa8}, {0xaaa, 0xab0}, {0xab2, 0xab3}, {0xab5, 0xab9},
  {0xabc, 0xac5}, {0xac7, 0xac9}, {0xacb, 0xacd}, {0xad0, 0xad0}, {0xae0, 0xae3}, {0xae6, 0xaef}, {0xaf9, 0xaff}, {0xb01, 0xb03},
  {0xb05, 0xb0c}, {0xb0f, 0xb10}, {0xb13, 0xb28}, {0xb2a, 0xb30}, {0xb32, 0xb33}, {0xb35, 0xb39}, {0xb3c, 0xb44}, {0xb47, 0xb48},
  {0xb4b, 0xb4d}, {0xb55, 0xb57}, {0xb5c, 0xb5d}, {0xb5f, 0xb63}, {0xb66, 0xb6f}, {0xb71, 0xb71}, {0xb82, 0xb83}, {0xb85, 0xb8a},
  {0xb8e, 0xb90}, {0xb92, 0xb95}, {0xb99, 0xb9a}, {0xb9c, 0xb9c}, {0xb9e, 0xb9f}, {0xba3, 0xba4}, {0xba8, 0xbaa}, {0xbae, 0xbb9},
  {0xbbe, 0xbc2}, {0xbc6, 0xbc8}, {0xbca, 0xbcd}, {0xbd0, 0xbd0}, {0xbd7, 0xbd7}, {0xbe6, 0xbef}, {0xc00, 0xc0c}, {0xc0e, 0xc10},
  {0xc12, 0xc28}, {0xc2a, 0xc39}, {0xc3c, 0xc44}, {0xc46, 0xc48}, {0xc4a, 0xc4d}, {0xc55, 0xc56}, {0xc58, 0xc5a}, {0xc5d, 0xc5d},
  {0xc60, 0xc63}, {0xc66, 0xc6f}, {0xc80, 0xc83}, {0xc85, 0xc8c}, {0xc8e, 0xc90}, {0xc92, 0xca8}, {0xcaa, 0xcb3}, {0xcb5, 0xcb9},
  {0xcbc, 0xcc4}, {0xcc6, 0xcc8}, {0xcca, 0xccd}, {0xcd5, 0xcd6}, {0xcdd, 0xcde}, {0xce0, 0xce3}, {0xce6, 0xcef}, {0xcf1, 0xcf3},
  {0xd00, 0xd0c}, {0xd0e, 0xd10}, {0xd12, 0xd44}, {0xd46, 0xd48}, {0xd4a, 0xd4e}, {0xd54, 0xd57}, {0xd5f, 0xd63}, {0xd66, 0xd6f},
  {0xd7a, 0xd7f}, {0xd81, 0xd83}, {0xd85, 0xd96}, {0xd9a, 0xdb1}, {0xdb3, 0xdbb}, {0xdbd, 0xdbd}, {0xdc0, 0xdc6}, {0xdca, 0xdca},
  {0xdcf, 0xdd4}, {0xdd6, 0xdd6}, {0xdd8, 0xddf}, {0xde6, 0xdef}, {0xdf2, 0xdf3}, {0xe01, 0xe3a}, {0xe40, 0xe4e}, {0xe50, 0xe59},
  {0xe81, 0xe82}, {0xe84, 0xe84}, {0xe86, 0xe8a}, {0xe8c, 0xea3}, {0xea5, 0xea5}, {0xea7, 0xebd}, {0xec0, 0xec4}, {0xec6, 0xec6},
  {0xec8, 0xece}, {0xed0, 0xed9}, {0xedc, 0xedf}, {0xf00, 0xf00}, {0xf18, 0xf19}, {0xf20, 0xf29}, {0xf35, 0xf35}, {0xf37, 0xf37},
  {0xf39, 0xf39}, {0xf3e, 0xf47}, {0xf49, 0xf6c}, {0xf71, 0xf84}, {0xf86, 0xf97}, {0xf99, 0xfbc}, {0xfc6, 0xfc6}, {0x1000, 0x1049},
  {0x1050, 0x109d}, {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x10d0, 0x10fa}, {0x10fc, 0x1248}, {0x124a, 0x124d}, {0x1250, 0x1256},
  {0x1258, 0x1258}, {0x125a, 0x125d}, {0x1260, 0x1288}, {0x128a, 0x128d}, {0x1290, 0x12b0}, {0x12b2, 0x12b5}, {0x12b8, 0x12be}, {0x12c0, 0x12c0},
  {0x12c2, 0x12c5}, {0x12c8, 0x12d6}, {0x12d8, 0x1310}, {0x1312, 0x1315}, {0x1318, 0x135a}, {0x135d, 0x135f}, {0x1369, 0x1371}, {0x1380, 0x138f},
  {0x13a0, 0x13f5}, {0x13f8, 0x13fd}, {0x1401, 0x166c}, {0x166f, 0x167f}, {0x1681, 0x169a}, {0x16a0, 0x16ea}, {0x16ee, 0x16f8}, {0x1700, 0x1715},
  {0x171f, 0x1734}, {0x1740, 0x1753}, {0x1760, 0x176c}, {0x176e, 0x1770}, {0x1772, 0x1773}, {0x1780, 0x17d3}, {0x17d7, 0x17d7}, {0x17dc, 0x17dd},
  {0x17e0, 0x17e9}, {0x180b, 0x180d}, {0x180f, 0x1819}, {0x1820, 0x1878}, {0x1880, 0x18aa}, {0x18b0, 0x18f5}, {0x1900, 0x191e}, {0x1920, 0x192b},
  {0x1930, 0x193b}, {0x1946, 0x196d}, {0x1970, 0x1974}, {0x1980, 0x19ab}, {0x19b0, 0x19c9}, {0x19d0, 0x19da}, {0x1a00, 0x1a1b}, {0x1a20, 0x1a5e},
  {0x1a60, 0x1a7c}, {0x1a7f, 0x1a89}, {0x1a90, 0x1a99}, {0x1aa7, 0x1aa7}, {0x1ab0, 0x1abd}, {0x1abf, 0x1ace}, {0x1b00, 0x1b4c}, {0x1b50, 0x1b59},
  {0x1b6b, 0x1b73}, {0x1b80, 0x1bf3}, {0x1c00, 0x1c37}, {0x1c40, 0x1c49}, {0x1c4d, 0x1c7d}, {0x1c80, 0x1c8a}, {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf},
  {0x1cd0, 0x1cd2}, {0x1cd4, 0x1cfa}, {0x1d00, 0x1f15}, {0x1f18, 0x1f1d}, {0x1f20, 0x1f45}, {0x1f48, 0x1f4d}, {0x1f50, 0x1f57}, {0x1f59, 0x1f59},
  {0x1f5b, 0x1f5b}, {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f7d}, {0x1f80, 0x1fb4}, {0x1fb6, 0x1fbc}, {0x1fbe, 0x1fbe}, {0x1fc2, 0x1fc4}, {0x1fc6, 0x1fcc},
  {0x1fd0, 0x1fd3}, {0x1fd6, 0x1fdb}, {0x1fe0, 0x1fec}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ffc}, {0x200c, 0x200d}, {0x203f, 0x2040}, {0x2054, 0x2054},
  {0x2071, 0x2071}, {0x207f, 0x207f}, {0x2090, 0x209c}, {0x20d0, 0x20dc}, {0x20e1, 0x20e1}, {0x20e5, 0x20f0}, {0x2102, 0x2102}, {0x2107, 0x2107},
  {0x210a, 0x2113}, {0x2115, 0x2115}, {0x2118, 0x211d}, {0x2124, 0x2124}, {0x2126, 0x2126}, {0x2128, 0x2128}, {0x212a, 0x2139}, {0x213c, 0x213f},
  {0x2145, 0x2149}, {0x214e, 0x214e}, {0x2160, 0x2188}, {0x2c00, 0x2ce4}, {0x2ceb, 0x2cf3}, {0x2d00, 0x2d25}, {0x2d27, 0x2d27}, {0x2d2d, 0x2d2d},
  {0x2d30, 0x2d67}, {0x2d6f, 0x2d6f}, {0x2d7f, 0x2d96}, {0x2da0, 0x2da6}, {0x2da8, 0x2dae}, {0x2db0, 0x2db6}, {0x2db8, 0x2dbe}, {0x2dc0, 0x2dc6},
  {0x2dc8, 0x2dce}, {0x2dd0, 0x2dd6}, {0x2dd8, 0x2dde}, {0x2de0, 0x2dff}, {0x3005, 0x3007}, {0x3021, 0x302f}, {0x3031, 0x3035}, {0x3038, 0x303c},
  {0x3041, 0x3096}, {0x3099, 0x309a}, {0x309d, 0x309f}, {0x30a1, 0x30ff}, {0x3105, 0x312f}, {0x3131, 0x318e}, {0x31a0, 0x31bf}, {0x31f0, 0x31ff},
  {0x3400, 0x4dbf}, {0x4e00, 0xa48c}, {0xa4d0, 0xa4fd}, {0xa500, 0xa60c}, {0xa610, 0xa62b}, {0xa640, 0xa66f}, {0xa674, 0xa67d}, {0xa67f, 0xa6f1},
  {0xa717, 0xa71f}, {0xa722, 0xa788}, {0xa78b, 0xa7cd}, {0xa7d0, 0xa7d1}, {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7dc}, {0xa7f2, 0xa827}, {0xa82c, 0xa82c},
  {0xa840, 0xa873}, {0xa880, 0xa8c5}, {0xa8d0, 0xa8d9}, {0xa8e0, 0xa8f7}, {0xa8fb, 0xa8fb}, {0xa8fd, 0xa92d}, {0xa930, 0xa953}, {0xa960, 0xa97c},
  {0xa980, 0xa9c0}, {0xa9cf, 0xa9d9}, {0xa9e0, 0xa9fe}, {0xaa00, 0xaa36}, {0xaa40, 0xaa4d}, {0xaa50, 0xaa59}, {0xaa60, 0xaa76}, {0xaa7a, 0xaac2},
  {0xaadb, 0xaadd}, {0xaae0, 0xaaef}, {0xaaf2, 0xaaf6}, {0xab01, 0xab06}, {0xab09, 0xab0e}, {0xab11, 0xab16}, {0xab20, 0xab26}, {0xab28, 0xab2e},
  {0xab30, 0xab5a}, {0xab5c, 0xab69}, {0xab70, 0xabea}, {0xabec, 0xabed}, {0xabf0, 0xabf9}, {0xac00, 0xd7a3}, {0xd7b0, 0xd7c6}, {0xd7cb, 0xd7fb},
  {0xf900, 0xfa6d}, {0xfa70, 0xfad9}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xfb1d, 0xfb28}, {0xfb2a, 0xfb36}, {0xfb38, 0xfb3c}, {0xfb3e, 0xfb3e},
  {0xfb40, 0xfb41}, {0xfb43, 0xfb44}, {0xfb46, 0xfbb1}, {0xfbd3, 0xfc5d}, {0xfc64, 0xfd3d}, {0xfd50, 0xfd8f}, {0xfd92, 0xfdc7}, {0xfdf0, 0xfdf9},
  {0xfe00, 0xfe0f}, {0xfe20, 0xfe2f}, {0xfe33, 0xfe34}, {0xfe4d, 0xfe4f}, {0xfe71, 0xfe71}, {0xfe73, 0xfe73}, {0xfe77, 0xfe77}, {0xfe79, 0xfe79},
  {0xfe7b, 0xfe7b}, {0xfe7d, 0xfe7d}, {0xfe7f, 0xfefc}, {0xff10, 0xff19}, {0xff21, 0xff3a}, {0xff3f, 0xff3f}, {0xff41, 0xff5a}, {0xff65, 0xffbe},
  {0xffc2, 0xffc7}, {0xffca, 0xffcf}, {0xffd2, 0xffd7}, {0xffda, 0xffdc}, {0x10000, 0x1000b}, {0x1000d, 0x10026}, {0x10028, 0x1003a}, {0x1003c, 0x1003d},
  {0x1003f, 0x1004d}, {0x10050, 0x1005d}, {0x10080, 0x100fa}, {0x10140, 0x10174}, {0x101fd, 0x101fd}, {0x10280, 0x1029c}, {0x102a0, 0x102d0}, {0x102e0, 0x102e0},
  {0x10300, 0x1031f}, {0x1032d, 0x1034a}, {0x10350, 0x1037a}, {0x10380, 0x1039d}, {0x103a0, 0x103c3}, {0x103c8, 0x103cf}, {0x103d1, 0x103d5}, {0x10400, 0x1049d},
  {0x104a0, 0x104a9}, {0x104b0, 0x104d3}, {0x104d8, 0x104fb}, {0x10500, 0x10527}, {0x10530, 0x10563}, {0x10570, 0x1057a}, {0x1057c, 0x1058a}, {0x1058c, 0x10592},
  {0x10594, 0x10595}, {0x10597, 0x105a1}, {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc}, {0x105c0, 0x105f3}, {0x10600, 0x10736}, {0x10740, 0x10755},
  {0x10760, 0x10767}, {0x10780, 0x10785}, {0x10787, 0x107b0}, {0x107b2, 0x107ba}, {0x10800, 0x10805}, {0x10808, 0x10808}, {0x1080a, 0x10835}, {0x10837, 0x10838},
  {0x1083c, 0x1083c}, {0x1083f, 0x10855}, {0x10860, 0x10876}, {0x10880, 0x1089e}, {0x108e0, 0x108f2}, {0x108f4, 0x108f5}, {0x10900, 0x10915}, {0x10920, 0x10939},
  {0x10980, 0x109b7}, {0x109be, 0x109bf}, {0x10a00, 0x10a03}, {0x10a05, 0x10a06}, {0x10a0c, 0x10a13}, {0x10a15, 0x10a17}, {0x10a19, 0x10a35}, {0x10a38, 0x10a3a},
  {0x10a3f, 0x10a3f}, {0x10a60, 0x10a7c}, {0x10a80, 0x10a9c}, {0x10ac0, 0x10ac7}, {0x10ac9, 0x10ae6}, {0x10b00, 0x10b35}, {0x10b40, 0x10b55}, {0x10b60, 0x10b72},
  {0x10b80, 0x10b91}, {0x10c00, 0x10c48}, {0x10c80, 0x10cb2}, {0x10cc0, 0x10cf2}, {0x10d00, 0x10d27}, {0x10d30, 0x10d39}, {0x10d40, 0x10d65}, {0x10d69, 0x10d6d},
  {0x10d6f, 0x10d85}, {0x10e80, 0x10ea9}, {0x10eab, 0x10eac}, {0x10eb0, 0x10eb1}, {0x10ec2, 0x10ec4}, {0x10efc, 0x10f1c}, {0x10f27, 0x10f27}, {0x10f30, 0x10f50},
  {0x10f70, 0x10f85}, {0x10fb0, 0x10fc4}, {0x10fe0, 0x10ff6}, {0x11000, 0x11046}, {0x11066, 0x11075}, {0x1107f, 0x110ba}, {0x110c2, 0x110c2}, {0x110d0, 0x110e8},
  {0x110f0, 0x110f9}, {0x11100, 0x11134}, {0x11136, 0x1113f}, {0x11144, 0x11147}, {0x11150, 0x11173}, {0x11176, 0x11176}, {0x11180, 0x111c4}, {0x111c9, 0x111cc},
  {0x111ce, 0x111da}, {0x111dc, 0x111dc}, {0x11200, 0x11211}, {0x11213, 0x11237}, {0x1123e, 0x11241}, {0x11280, 0x11286}, {0x11288, 0x11288}, {0x1128a, 0x1128d},
  {0x1128f, 0x1129d}, {0x1129f, 0x112a8}, {0x112b0, 0x112ea}, {0x112f0, 0x112f9}, {0x11300, 0x11303}, {0x11305, 0x1130c}, {0x1130f, 0x11310}, {0x11313, 0x11328},
  {0x1132a, 0x11330}, {0x11332, 0x11333}, {0x11335, 0x11339}, {0x1133b, 0x11344}, {0x11347, 0x11348}, {0x1134b, 0x1134d}, {0x11350, 0x11350}, {0x11357, 0x11357},
  {0x1135d, 0x11363}, {0x11366, 0x1136c}, {0x11370, 0x11374}, {0x11380, 0x11389}, {0x1138b, 0x1138b}, {0x1138e, 0x1138e}, {0x11390, 0x113b5}, {0x113b7, 0x113c0},
  {0x113c2, 0x113c2}, {0x113c5, 0x113c5}, {0x113c7, 0x113ca}, {0x113cc, 0x113d3}, {0x113e1, 0x113e2}, {0x11400, 0x1144a}, {0x11450, 0x11459}, {0x1145e, 0x11461},
  {0x11480, 0x114c5}, {0x114c7, 0x114c7}, {0x114d0, 0x114d9}, {0x11580, 0x115b5}, {0x115b8, 0x115c0}, {0x115d8, 0x115dd}, {0x11600, 0x11640}, {0x11644, 0x11644},
  {0x11650, 0x11659}, {0x11680, 0x116b8}, {0x116c0, 0x116c9}, {0x116d0, 0x116e3}, {0x11700, 0x1171a}, {0x1171d, 0x1172b}, {0x11730, 0x11739}, {0x11740, 0x11746},
  {0x11800, 0x1183a}, {0x118a0, 0x118e9}, {0x118ff, 0x11906}, {0x11909, 0x11909}, {0x1190c, 0x11913}, {0x11915, 0x11916}, {0x11918, 0x11935}, {0x11937, 0x11938},
  {0x1193b, 0x11943}, {0x11950, 0x11959}, {0x119a0, 0x119a7}, {0x119aa, 0x119d7}, {0x119da, 0x119e1}, {0x119e3, 0x119e4}, {0x11a00, 0x11a3e}, {0x11a47, 0x11a47},
  {0x11a50, 0x11a99}, {0x11a9d, 0x11a9d}, {0x11ab0, 0x11af8}, {0x11bc0, 0x11be0}, {0x11bf0, 0x11bf9}, {0x11c00, 0x11c08}, {0x11c0a, 0x11c36}, {0x11c38, 0x11c40},
  {0x11c50, 0x11c59}, {0x11c72, 0x11c8f}, {0x11c92, 0x11ca7}, {0x11ca9, 0x11cb6}, {0x11d00, 0x11d06}, {0x11d08, 0x11d09}, {0x11d0b, 0x11d36}, {0x11d3a, 0x11d3a},
  {0x11d3c, 0x11d3d}, {0x11d3f, 0x11d47}, {0x11d50, 0x11d59}, {0x11d60, 0x11d65}, {0x11d67, 0x11d68}, {0x11d6a, 0x11d8e}, {0x11d90, 0x11d91}, {0x11d93, 0x11d98},
  {0x11da0, 0x11da9}, {0x11ee0, 0x11ef6}, {0x11f00, 0x11f10}, {0x11f12, 0x11f3a}, {0x11f3e, 0x11f42}, {0x11f50, 0x11f5a}, {0x11fb0, 0x11fb0}, {0x12000, 0x12399},
  {0x12400, 0x1246e}, {0x12480, 0x12543}, {0x12f90, 0x12ff0}, {0x13000, 0x1342f}, {0x13440, 0x13455}, {0x13460, 0x143fa}, {0x14400, 0x14646}, {0x16100, 0x16139},
  {0x16800, 0x16a38}, {0x16a40, 0x16a5e}, {0x16a60, 0x16a69}, {0x16a70, 0x16abe}, {0x16ac0, 0x16ac9}, {0x16ad0, 0x16aed}, {0x16af0, 0x16af4}, {0x16b00, 0x16b36},
  {0x16b40, 0x16b43}, {0x16b50, 0x16b59}, {0x16b63, 0x16b77}, {0x16b7d, 0x16b8f}, {0x16d40, 0x16d6c}, {0x16d70, 0x16d79}, {0x16e40, 0x16e7f}, {0x16f00, 0x16f4a},
  {0x16f4f, 0x16f87}, {0x16f8f, 0x16f9f}, {0x16fe0, 0x16fe1}, {0x16fe3, 0x16fe4}, {0x16ff0, 0x16ff1}, {0x17000, 0x187f7}, {0x18800, 0x18cd5}, {0x18cff, 0x18d08},
  {0x1aff0, 0x1aff3}, {0x1aff5, 0x1affb}, {0x1affd, 0x1affe}, {0x1b000, 0x1b122}, {0x1b132, 0x1b132}, {0x1b150, 0x1b152}, {0x1b155, 0x1b155}, {0x1b164, 0x1b167},
  {0x1b170, 0x1b2fb}, {0x1bc00, 0x1bc6a}, {0x1bc70, 0x1bc7c}, {0x1bc80, 0x1bc88}, {0x1bc90, 0x1bc99}, {0x1bc9d, 0x1bc9e}, {0x1ccf0, 0x1ccf9}, {0x1cf00, 0x1cf2d},
  {0x1cf30, 0x1cf46}, {0x1d165, 0x1d169}, {0x1d16d, 0x1d172}, {0x1d17b, 0x1d182}, {0x1d185, 0x1d18b}, {0x1d1aa, 0x1d1ad}, {0x1d242, 0x1d244}, {0x1d400, 0x1d454},
  {0x1d456, 0x1d49c}, {0x1d49e, 0x1d49f}, {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b9}, {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3},
  {0x1d4c5, 0x1d505}, {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514}, {0x1d516, 0x1d51c}, {0x1d51e, 0x1d539}, {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544}, {0x1d546, 0x1d546},
  {0x1d54a, 0x1d550}, {0x1d552, 0x1d6a5}, {0x1d6a8, 0x1d6c0}, {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6fa}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d734}, {0x1d736, 0x1d74e},
  {0x1d750, 0x1d76e}, {0x1d770, 0x1d788}, {0x1d78a, 0x1d7a8}, {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7cb}, {0x1d7ce, 0x1d7ff}, {0x1da00, 0x1da36}, {0x1da3b, 0x1da6c},
  {0x1da75, 0x1da75}, {0x1da84, 0x1da84}, {0x1da9b, 0x1da9f}, {0x1daa1, 0x1daaf}, {0x1df00, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e000, 0x1e006}, {0x1e008, 0x1e018},
  {0x1e01b, 0x1e021}, {0x1e023, 0x1e024}, {0x1e026, 0x1e02a}, {0x1e030, 0x1e06d}, {0x1e08f, 0x1e08f}, {0x1e100, 0x1e12c}, {0x1e130, 0x1e13d}, {0x1e140, 0x1e149},
  {0x1e14e, 0x1e14e}, {0x1e290, 0x1e2ae}, {0x1e2c0, 0x1e2f9}, {0x1e4d0, 0x1e4f9}, {0x1e5d0, 0x1e5fa}, {0x1e7e0, 0x1e7e6}, {0x1e7e8, 0x1e7eb}, {0x1e7ed, 0x1e7ee},
  {0x1e7f0, 0x1e7fe}, {0x1e800, 0x1e8c4}, {0x1e8d0, 0x1e8d6}, {0x1e900, 0x1e94b}, {0x1e950, 0x1e959}, {0x1ee00, 0x1ee03}, {0x1ee05, 0x1ee1f}, {0x1ee21, 0x1ee22},
  {0x1ee24, 0x1ee24}, {0x1ee27, 0x1ee27}, {0x1ee29, 0x1ee32}, {0x1ee34, 0x1ee37}, {0x1ee39, 0x1ee39}, {0x1ee3b, 0x1ee3b}, {0x1ee42, 0x1ee42}, {0x1ee47, 0x1ee47},
  {0x1ee49, 0x1ee49}, {0x1ee4b, 0x1ee4b}, {0x1ee4d, 0x1ee4f}, {0x1ee51, 0x1ee52}, {0x1ee54, 0x1ee54}, {0x1ee57, 0x1ee57}, {0x1ee59, 0x1ee59}, {0x1ee5b, 0x1ee5b},
  {0x1ee5d, 0x1ee5d}, {0x1ee5f, 0x1ee5f}, {0x1ee61, 0x1ee62}, {0x1ee64, 0x1ee64}, {0x1ee67, 0x1ee6a}, {0x1ee6c, 0x1ee72}, {0x1ee74, 0x1ee77}, {0x1ee79, 0x1ee7c},
  {0x1ee7e, 0x1ee7e}, {0x1ee80, 0x1ee89}, {0x1ee8b, 0x1ee9b}, {0x1eea1, 0x1eea3}, {0x1eea5, 0x1eea9}, {0x1eeab, 0x1eebb}, {0x1fbf0, 0x1fbf9}, {0x20000, 0x2a6df},
  {0x2a700, 0x2b739}, {0x2b740, 0x2b81d}, {0x2b820, 0x2cea1}, {0x2ceb0, 0x2ebe0}, {0x2ebf0, 0x2ee5d}, {0x2f800, 0x2fa1d}, {0x30000, 0x3134a}, {0x31350, 0x323af},
  {0xe0100, 0xe01ef},
};

static const TSCharacterRange sym__lowercase_identifier_character_set_3[] = {
  {'_', '_'}, {'a', 'z'}, {0xb5, 0xb5}, {0xdf, 0xf6}, {0xf8, 0xff}, {0x101, 0x101}, {0x103, 0x103}, {0x105, 0x105},
  {0x107, 0x107}, {0x109, 0x109}, {0x10b, 0x10b}, {0x10d, 0x10d}, {0x10f, 0x10f}, {0x111, 0x111}, {0x113, 0x113}, {0x115, 0x115},
  {0x117, 0x117}, {0x119, 0x119}, {0x11b, 0x11b}, {0x11d, 0x11d}, {0x11f, 0x11f}, {0x121, 0x121}, {0x123, 0x123}, {0x125, 0x125},
  {0x127, 0x127}, {0x129, 0x129}, {0x12b, 0x12b}, {0x12d, 0x12d}, {0x12f, 0x12f}, {0x131, 0x131}, {0x133, 0x133}, {0x135, 0x135},
  {0x137, 0x138}, {0x13a, 0x13a}, {0x13c, 0x13c}, {0x13e, 0x13e}, {0x140, 0x140}, {0x142, 0x142}, {0x144, 0x144}, {0x146, 0x146},
  {0x148, 0x149}, {0x14b, 0x14b}, {0x14d, 0x14d}, {0x14f, 0x14f}, {0x151, 0x151}, {0x153, 0x153}, {0x155, 0x155}, {0x157, 0x157},
  {0x159, 0x159}, {0x15b, 0x15b}, {0x15d, 0x15d}, {0x15f, 0x15f}, {0x161, 0x161}, {0x163, 0x163}, {0x165, 0x165}, {0x167, 0x167},
  {0x169, 0x169}, {0x16b, 0x16b}, {0x16d, 0x16d}, {0x16f, 0x16f}, {0x171, 0x171}, {0x173, 0x173}, {0x175, 0x175}, {0x177, 0x177},
  {0x17a, 0x17a}, {0x17c, 0x17c}, {0x17e, 0x180}, {0x183, 0x183}, {0x185, 0x185}, {0x188, 0x188}, {0x18c, 0x18d}, {0x192, 0x192},
  {0x195, 0x195}, {0x199, 0x19b}, {0x19e, 0x19e}, {0x1a1, 0x1a1}, {0x1a3, 0x1a3}, {0x1a5, 0x1a5}, {0x1a8, 0x1a8}, {0x1aa, 0x1ab},
  {0x1ad, 0x1ad}, {0x1b0, 0x1b0}, {0x1b4, 0x1b4}, {0x1b6, 0x1b6}, {0x1b9, 0x1ba}, {0x1bd, 0x1bf}, {0x1c6, 0x1c6}, {0x1c9, 0x1c9},
  {0x1cc, 0x1cc}, {0x1ce, 0x1ce}, {0x1d0, 0x1d0}, {0x1d2, 0x1d2}, {0x1d4, 0x1d4}, {0x1d6, 0x1d6}, {0x1d8, 0x1d8}, {0x1da, 0x1da},
  {0x1dc, 0x1dd}, {0x1df, 0x1df}, {0x1e1, 0x1e1}, {0x1e3, 0x1e3}, {0x1e5, 0x1e5}, {0x1e7, 0x1e7}, {0x1e9, 0x1e9}, {0x1eb, 0x1eb},
  {0x1ed, 0x1ed}, {0x1ef, 0x1f0}, {0x1f3, 0x1f3}, {0x1f5, 0x1f5}, {0x1f9, 0x1f9}, {0x1fb, 0x1fb}, {0x1fd, 0x1fd}, {0x1ff, 0x1ff},
  {0x201, 0x201}, {0x203, 0x203}, {0x205, 0x205}, {0x207, 0x207}, {0x209, 0x209}, {0x20b, 0x20b}, {0x20d, 0x20d}, {0x20f, 0x20f},
  {0x211, 0x211}, {0x213, 0x213}, {0x215, 0x215}, {0x217, 0x217}, {0x219, 0x219}, {0x21b, 0x21b}, {0x21d, 0x21d}, {0x21f, 0x21f},
  {0x221, 0x221}, {0x223, 0x223}, {0x225, 0x225}, {0x227, 0x227}, {0x229, 0x229}, {0x22b, 0x22b}, {0x22d, 0x22d}, {0x22f, 0x22f},
  {0x231, 0x231}, {0x233, 0x239}, {0x23c, 0x23c}, {0x23f, 0x240}, {0x242, 0x242}, {0x247, 0x247}, {0x249, 0x249}, {0x24b, 0x24b},
  {0x24d, 0x24d}, {0x24f, 0x293}, {0x295, 0x2af}, {0x371, 0x371}, {0x373, 0x373}, {0x377, 0x377}, {0x37b, 0x37d}, {0x390, 0x390},
  {0x3ac, 0x3ce}, {0x3d0, 0x3d1}, {0x3d5, 0x3d7}, {0x3d9, 0x3d9}, {0x3db, 0x3db}, {0x3dd, 0x3dd}, {0x3df, 0x3df}, {0x3e1, 0x3e1},
  {0x3e3, 0x3e3}, {0x3e5, 0x3e5}, {0x3e7, 0x3e7}, {0x3e9, 0x3e9}, {0x3eb, 0x3eb}, {0x3ed, 0x3ed}, {0x3ef, 0x3f3}, {0x3f5, 0x3f5},
  {0x3f8, 0x3f8}, {0x3fb, 0x3fc}, {0x430, 0x45f}, {0x461, 0x461}, {0x463, 0x463}, {0x465, 0x465}, {0x467, 0x467}, {0x469, 0x469},
  {0x46b, 0x46b}, {0x46d, 0x46d}, {0x46f, 0x46f}, {0x471, 0x471}, {0x473, 0x473}, {0x475, 0x475}, {0x477, 0x477}, {0x479, 0x479},
  {0x47b, 0x47b}, {0x47d, 0x47d}, {0x47f, 0x47f}, {0x481, 0x481}, {0x48b, 0x48b}, {0x48d, 0x48d}, {0x48f, 0x48f}, {0x491, 0x491},
  {0x493, 0x493}, {0x495, 0x495}, {0x497, 0x497}, {0x499, 0x499}, {0x49b, 0x49b}, {0x49d, 0x49d}, {0x49f, 0x49f}, {0x4a1, 0x4a1},
  {0x4a3, 0x4a3}, {0x4a5, 0x4a5}, {0x4a7, 0x4a7}, {0x4a9, 0x4a9}, {0x4ab, 0x4ab}, {0x4ad, 0x4ad}, {0x4af, 0x4af}, {0x4b1, 0x4b1},
  {0x4b3, 0x4b3}, {0x4b5, 0x4b5}, {0x4b7, 0x4b7}, {0x4b9, 0x4b9}, {0x4bb, 0x4bb}, {0x4bd, 0x4bd}, {0x4bf, 0x4bf}, {0x4c2, 0x4c2},
  {0x4c4, 0x4c4}, {0x4c6, 0x4c6}, {0x4c8, 0x4c8}, {0x4ca, 0x4ca}, {0x4cc, 0x4cc}, {0x4ce, 0x4cf}, {0x4d1, 0x4d1}, {0x4d3, 0x4d3},
  {0x4d5, 0x4d5}, {0x4d7, 0x4d7}, {0x4d9, 0x4d9}, {0x4db, 0x4db}, {0x4dd, 0x4dd}, {0x4df, 0x4df}, {0x4e1, 0x4e1}, {0x4e3, 0x4e3},
  {0x4e5, 0x4e5}, {0x4e7, 0x4e7}, {0x4e9, 0x4e9}, {0x4eb, 0x4eb}, {0x4ed, 0x4ed}, {0x4ef, 0x4ef}, {0x4f1, 0x4f1}, {0x4f3, 0x4f3},
  {0x4f5, 0x4f5}, {0x4f7, 0x4f7}, {0x4f9, 0x4f9}, {0x4fb, 0x4fb}, {0x4fd, 0x4fd}, {0x4ff, 0x4ff}, {0x501, 0x501}, {0x503, 0x503},
  {0x505, 0x505}, {0x507, 0x507}, {0x509, 0x509}, {0x50b, 0x50b}, {0x50d, 0x50d}, {0x50f, 0x50f}, {0x511, 0x511}, {0x513, 0x513},
  {0x515, 0x515}, {0x517, 0x517}, {0x519, 0x519}, {0x51b, 0x51b}, {0x51d, 0x51d}, {0x51f, 0x51f}, {0x521, 0x521}, {0x523, 0x523},
  {0x525, 0x525}, {0x527, 0x527}, {0x529, 0x529}, {0x52b, 0x52b}, {0x52d, 0x52d}, {0x52f, 0x52f}, {0x560, 0x588}, {0x10d0, 0x10fa},
  {0x10fd, 0x10ff}, {0x13f8, 0x13fd}, {0x1c80, 0x1c88}, {0x1c8a, 0x1c8a}, {0x1d00, 0x1d2b}, {0x1d6b, 0x1d77}, {0x1d79, 0x1d9a}, {0x1e01, 0x1e01},
  {0x1e03, 0x1e03}, {0x1e05, 0x1e05}, {0x1e07, 0x1e07}, {0x1e09, 0x1e09}, {0x1e0b, 0x1e0b}, {0x1e0d, 0x1e0d}, {0x1e0f, 0x1e0f}, {0x1e11, 0x1e11},
  {0x1e13, 0x1e13}, {0x1e15, 0x1e15}, {0x1e17, 0x1e17}, {0x1e19, 0x1e19}, {0x1e1b, 0x1e1b}, {0x1e1d, 0x1e1d}, {0x1e1f, 0x1e1f}, {0x1e21, 0x1e21},
  {0x1e23, 0x1e23}, {0x1e25, 0x1e25}, {0x1e27, 0x1e27}, {0x1e29, 0x1e29}, {0x1e2b, 0x1e2b}, {0x1e2d, 0x1e2d}, {0x1e2f, 0x1e2f}, {0x1e31, 0x1e31},
  {0x1e33, 0x1e33}, {0x1e35, 0x1e35}, {0x1e37, 0x1e37}, {0x1e39, 0x1e39}, {0x1e3b, 0x1e3b}, {0x1e3d, 0x1e3d}, {0x1e3f, 0x1e3f}, {0x1e41, 0x1e41},
  {0x1e43, 0x1e43}, {0x1e45, 0x1e45}, {0x1e47, 0x1e47}, {0x1e49, 0x1e49}, {0x1e4b, 0x1e4b}, {0x1e4d, 0x1e4d}, {0x1e4f, 0x1e4f}, {0x1e51, 0x1e51},
  {0x1e53, 0x1e53}, {0x1e55, 0x1e55}, {0x1e57, 0x1e57}, {0x1e59, 0x1e59}, {0x1e5b, 0x1e5b}, {0x1e5d, 0x1e5d}, {0x1e5f, 0x1e5f}, {0x1e61, 0x1e61},
  {0x1e63, 0x1e63}, {0x1e65, 0x1e65}, {0x1e67, 0x1e67}, {0x1e69, 0x1e69}, {0x1e6b, 0x1e6b}, {0x1e6d, 0x1e6d}, {0x1e6f, 0x1e6f}, {0x1e71, 0x1e71},
  {0x1e73, 0x1e73}, {0x1e75, 0x1e75}, {0x1e77, 0x1e77}, {0x1e79, 0x1e79}, {0x1e7b, 0x1e7b}, {0x1e7d, 0x1e7d}, {0x1e7f, 0x1e7f}, {0x1e81, 0x1e81},
  {0x1e83, 0x1e83}, {0x1e85, 0x1e85}, {0x1e87, 0x1e87}, {0x1e89, 0x1e89}, {0x1e8b, 0x1e8b}, {0x1e8d, 0x1e8d}, {0x1e8f, 0x1e8f}, {0x1e91, 0x1e91},
  {0x1e93, 0x1e93}, {0x1e95, 0x1e9d}, {0x1e9f, 0x1e9f}, {0x1ea1, 0x1ea1}, {0x1ea3, 0x1ea3}, {0x1ea5, 0x1ea5}, {0x1ea7, 0x1ea7}, {0x1ea9, 0x1ea9},
  {0x1eab, 0x1eab}, {0x1ead, 0x1ead}, {0x1eaf, 0x1eaf}, {0x1eb1, 0x1eb1}, {0x1eb3, 0x1eb3}, {0x1eb5, 0x1eb5}, {0x1eb7, 0x1eb7}, {0x1eb9, 0x1eb9},
  {0x1ebb, 0x1ebb}, {0x1ebd, 0x1ebd}, {0x1ebf, 0x1ebf}, {0x1ec1, 0x1ec1}, {0x1ec3, 0x1ec3}, {0x1ec5, 0x1ec5}, {0x1ec7, 0x1ec7}, {0x1ec9, 0x1ec9},
  {0x1ecb, 0x1ecb}, {0x1ecd, 0x1ecd}, {0x1ecf, 0x1ecf}, {0x1ed1, 0x1ed1}, {0x1ed3, 0x1ed3}, {0x1ed5, 0x1ed5}, {0x1ed7, 0x1ed7}, {0x1ed9, 0x1ed9},
  {0x1edb, 0x1edb}, {0x1edd, 0x1edd}, {0x1edf, 0x1edf}, {0x1ee1, 0x1ee1}, {0x1ee3, 0x1ee3}, {0x1ee5, 0x1ee5}, {0x1ee7, 0x1ee7}, {0x1ee9, 0x1ee9},
  {0x1eeb, 0x1eeb}, {0x1eed, 0x1eed}, {0x1eef, 0x1eef}, {0x1ef1, 0x1ef1}, {0x1ef3, 0x1ef3}, {0x1ef5, 0x1ef5}, {0x1ef7, 0x1ef7}, {0x1ef9, 0x1ef9},
  {0x1efb, 0x1efb}, {0x1efd, 0x1efd}, {0x1eff, 0x1f07}, {0x1f10, 0x1f15}, {0x1f20, 0x1f27}, {0x1f30, 0x1f37}, {0x1f40, 0x1f45}, {0x1f50, 0x1f57},
  {0x1f60, 0x1f67}, {0x1f70, 0x1f7d}, {0x1f80, 0x1f87}, {0x1f90, 0x1f97}, {0x1fa0, 0x1fa7}, {0x1fb0, 0x1fb4}, {0x1fb6, 0x1fb7}, {0x1fbe, 0x1fbe},
  {0x1fc2, 0x1fc4}, {0x1fc6, 0x1fc7}, {0x1fd0, 0x1fd3}, {0x1fd6, 0x1fd7}, {0x1fe0, 0x1fe7}, {0x1ff2, 0x1ff4}, {0x1ff6, 0x1ff7}, {0x210a, 0x210a},
  {0x210e, 0x210f}, {0x2113, 0x2113}, {0x212f, 0x212f}, {0x2134, 0x2134}, {0x2139, 0x2139}, {0x213c, 0x213d}, {0x2146, 0x2149}, {0x214e, 0x214e},
  {0x2184, 0x2184}, {0x2c30, 0x2c5f}, {0x2c61, 0x2c61}, {0x2c65, 0x2c66}, {0x2c68, 0x2c68}, {0x2c6a, 0x2c6a}, {0x2c6c, 0x2c6c}, {0x2c71, 0x2c71},
  {0x2c73, 0x2c74}, {0x2c76, 0x2c7b}, {0x2c81, 0x2c81}, {0x2c83, 0x2c83}, {0x2c85, 0x2c85}, {0x2c87, 0x2c87}, {0x2c89, 0x2c89}, {0x2c8b, 0x2c8b},
  {0x2c8d, 0x2c8d}, {0x2c8f, 0x2c8f}, {0x2c91, 0x2c91}, {0x2c93, 0x2c93}, {0x2c95, 0x2c95}, {0x2c97, 0x2c97}, {0x2c99, 0x2c99}, {0x2c9b, 0x2c9b},
  {0x2c9d, 0x2c9d}, {0x2c9f, 0x2c9f}, {0x2ca1, 0x2ca1}, {0x2ca3, 0x2ca3}, {0x2ca5, 0x2ca5}, {0x2ca7, 0x2ca7}, {0x2ca9, 0x2ca9}, {0x2cab, 0x2cab},
  {0x2cad, 0x2cad}, {0x2caf, 0x2caf}, {0x2cb1, 0x2cb1}, {0x2cb3, 0x2cb3}, {0x2cb5, 0x2cb5}, {0x2cb7, 0x2cb7}, {0x2cb9, 0x2cb9}, {0x2cbb, 0x2cbb},
  {0x2cbd, 0x2cbd}, {0x2cbf, 0x2cbf}, {0x2cc1, 0x2cc1}, {0x2cc3, 0x2cc3}, {0x2cc5, 0x2cc5}, {0x2cc7, 0x2cc7}, {0x2cc9, 0x2cc9}, {0x2ccb, 0x2ccb},
  {0x2ccd, 0x2ccd}, {0x2ccf, 0x2ccf}, {0x2cd1, 0x2cd1}, {0x2cd3, 0x2cd3}, {0x2cd5, 0x2cd5}, {0x2cd7, 0x2cd7}, {0x2cd9, 0x2cd9}, {0x2cdb, 0x2cdb},
  {0x2cdd, 0x2cdd}, {0x2cdf, 0x2cdf}, {0x2ce1, 0x2ce1}, {0x2ce3, 0x2ce4}, {0x2cec, 0x2cec}, {0x2cee, 0x2cee}, {0x2cf3, 0x2cf3}, {0x2d00, 0x2d25},
  {0x2d27, 0x2d27}, {0x2d2d, 0x2d2d}, {0xa641, 0xa641}, {0xa643, 0xa643}, {0xa645, 0xa645}, {0xa647, 0xa647}, {0xa649, 0xa649}, {0xa64b, 0xa64b},
  {0xa64d, 0xa64d}, {0xa64f, 0xa64f}, {0xa651, 0xa651}, {0xa653, 0xa653}, {0xa655, 0xa655}, {0xa657, 0xa657}, {0xa659, 0xa659}, {0xa65b, 0xa65b},
  {0xa65d, 0xa65d}, {0xa65f, 0xa65f}, {0xa661, 0xa661}, {0xa663, 0xa663}, {0xa665, 0xa665}, {0xa667, 0xa667}, {0xa669, 0xa669}, {0xa66b, 0xa66b},
  {0xa66d, 0xa66d}, {0xa681, 0xa681}, {0xa683, 0xa683}, {0xa685, 0xa685}, {0xa687, 0xa687}, {0xa689, 0xa689}, {0xa68b, 0xa68b}, {0xa68d, 0xa68d},
  {0xa68f, 0xa68f}, {0xa691, 0xa691}, {0xa693, 0xa693}, {0xa695, 0xa695}, {0xa697, 0xa697}, {0xa699, 0xa699}, {0xa69b, 0xa69b}, {0xa723, 0xa723},
  {0xa725, 0xa725}, {0xa727, 0xa727}, {0xa729, 0xa729}, {0xa72b, 0xa72b}, {0xa72d, 0xa72d}, {0xa72f, 0xa731}, {0xa733, 0xa733}, {0xa735, 0xa735},
  {0xa737, 0xa737}, {0xa739, 0xa739}, {0xa73b, 0xa73b}, {0xa73d, 0xa73d}, {0xa73f, 0xa73f}, {0xa741, 0xa741}, {0xa743, 0xa743}, {0xa745, 0xa745},
  {0xa747, 0xa747}, {0xa749, 0xa749}, {0xa74b, 0xa74b}, {0xa74d, 0xa74d}, {0xa74f, 0xa74f}, {0xa751, 0xa751}, {0xa753, 0xa753}, {0xa755, 0xa755},
  {0xa757, 0xa757}, {0xa759, 0xa759}, {0xa75b, 0xa75b}, {0xa75d, 0xa75d}, {0xa75f, 0xa75f}, {0xa761, 0xa761}, {0xa763, 0xa763}, {0xa765, 0xa765},
  {0xa767, 0xa767}, {0xa769, 0xa769}, {0xa76b, 0xa76b}, {0xa76d, 0xa76d}, {0xa76f, 0xa76f}, {0xa771, 0xa778}, {0xa77a, 0xa77a}, {0xa77c, 0xa77c},
  {0xa77f, 0xa77f}, {0xa781, 0xa781}, {0xa783, 0xa783}, {0xa785, 0xa785}, {0xa787, 0xa787}, {0xa78c, 0xa78c}, {0xa78e, 0xa78e}, {0xa791, 0xa791},
  {0xa793, 0xa795}, {0xa797, 0xa797}, {0xa799, 0xa799}, {0xa79b, 0xa79b}, {0xa79d, 0xa79d}, {0xa79f, 0xa79f}, {0xa7a1, 0xa7a1}, {0xa7a3, 0xa7a3},
  {0xa7a5, 0xa7a5}, {0xa7a7, 0xa7a7}, {0xa7a9, 0xa7a9}, {0xa7af, 0xa7af}, {0xa7b5, 0xa7b5}, {0xa7b7, 0xa7b7}, {0xa7b9, 0xa7b9}, {0xa7bb, 0xa7bb},
  {0xa7bd, 0xa7bd}, {0xa7bf, 0xa7bf}, {0xa7c1, 0xa7c1}, {0xa7c3, 0xa7c3}, {0xa7c8, 0xa7c8}, {0xa7ca, 0xa7ca}, {0xa7cd, 0xa7cd}, {0xa7d1, 0xa7d1},
  {0xa7d3, 0xa7d3}, {0xa7d5, 0xa7d5}, {0xa7d7, 0xa7d7}, {0xa7d9, 0xa7d9}, {0xa7db, 0xa7db}, {0xa7f6, 0xa7f6}, {0xa7fa, 0xa7fa}, {0xab30, 0xab5a},
  {0xab60, 0xab68}, {0xab70, 0xabbf}, {0xfb00, 0xfb06}, {0xfb13, 0xfb17}, {0xff41, 0xff5a}, {0x10428, 0x1044f}, {0x104d8, 0x104fb}, {0x10597, 0x105a1},
  {0x105a3, 0x105b1}, {0x105b3, 0x105b9}, {0x105bb, 0x105bc}, {0x10cc0, 0x10cf2}, {0x10d70, 0x10d85}, {0x118c0, 0x118df}, {0x16e60, 0x16e7f}, {0x1d41a, 0x1d433},
  {0x1d44e, 0x1d454}, {0x1d456, 0x1d467}, {0x1d482, 0x1d49b}, {0x1d4b6, 0x1d4b9}, {0x1d4bb, 0x1d4bb}, {0x1d4bd, 0x1d4c3}, {0x1d4c5, 0x1d4cf}, {0x1d4ea, 0x1d503},
  {0x1d51e, 0x1d537}, {0x1d552, 0x1d56b}, {0x1d586, 0x1d59f}, {0x1d5ba, 0x1d5d3}, {0x1d5ee, 0x1d607}, {0x1d622, 0x1d63b}, {0x1d656, 0x1d66f}, {0x1d68a, 0x1d6a5},
  {0x1d6c2, 0x1d6da}, {0x1d6dc, 0x1d6e1}, {0x1d6fc, 0x1d714}, {0x1d716, 0x1d71b}, {0x1d736, 0x1d74e}, {0x1d750, 0x1d755}, {0x1d770, 0x1d788}, {0x1d78a, 0x1d78f},
  {0x1d7aa, 0x1d7c2}, {0x1d7c4, 0x1d7c9}, {0x1d7cb, 0x1d7cb}, {0x1df00, 0x1df09}, {0x1df0b, 0x1df1e}, {0x1df25, 0x1df2a}, {0x1e922, 0x1e943},
};

static const TSCharacterRange sym__uppercase_identifier_character_set_1[] = {
  {'A', 'Z'}, {0xc0, 0xd6}, {0xd8, 0xde}, {0x100, 0x100}, {0x102, 0x102}, {0x104, 0x104}, {0x106, 0x106}, {0x108, 0x108},
  {0x10a, 0x10a}, {0x10c, 0x10c}, {0x10e, 0x10e}, {0x110, 0x110}, {0x112, 0x112}, {0x114, 0x114}, {0x116, 0x116}, {0x118, 0x118},
  {0x11a, 0x11a}, {0x11c, 0x11c}, {0x11e, 0x11e}, {0x120, 0x120}, {0x122, 0x122}, {0x124, 0x124}, {0x126, 0x126}, {0x128, 0x128},
  {0x12a, 0x12a}, {0x12c, 0x12c}, {0x12e, 0x12e}, {0x130, 0x130}, {0x132, 0x132}, {0x134, 0x134}, {0x136, 0x136}, {0x139, 0x139},
  {0x13b, 0x13b}, {0x13d, 0x13d}, {0x13f, 0x13f}, {0x141, 0x141}, {0x143, 0x143}, {0x145, 0x145}, {0x147, 0x147}, {0x14a, 0x14a},
  {0x14c, 0x14c}, {0x14e, 0x14e}, {0x150, 0x150}, {0x152, 0x152}, {0x154, 0x154}, {0x156, 0x156}, {0x158, 0x158}, {0x15a, 0x15a},
  {0x15c, 0x15c}, {0x15e, 0x15e}, {0x160, 0x160}, {0x162, 0x162}, {0x164, 0x164}, {0x166, 0x166}, {0x168, 0x168}, {0x16a, 0x16a},
  {0x16c, 0x16c}, {0x16e, 0x16e}, {0x170, 0x170}, {0x172, 0x172}, {0x174, 0x174}, {0x176, 0x176}, {0x178, 0x179}, {0x17b, 0x17b},
  {0x17d, 0x17d}, {0x181, 0x182}, {0x184, 0x184}, {0x186, 0x187}, {0x189, 0x18b}, {0x18e, 0x191}, {0x193, 0x194}, {0x196, 0x198},
  {0x19c, 0x19d}, {0x19f, 0x1a0}, {0x1a2, 0x1a2}, {0x1a4, 0x1a4}, {0x1a6, 0x1a7}, {0x1a9, 0x1a9}, {0x1ac, 0x1ac}, {0x1ae, 0x1af},
  {0x1b1, 0x1b3}, {0x1b5, 0x1b5}, {0x1b7, 0x1b8}, {0x1bc, 0x1bc}, {0x1c4, 0x1c4}, {0x1c7, 0x1c7}, {0x1ca, 0x1ca}, {0x1cd, 0x1cd},
  {0x1cf, 0x1cf}, {0x1d1, 0x1d1}, {0x1d3, 0x1d3}, {0x1d5, 0x1d5}, {0x1d7, 0x1d7}, {0x1d9, 0x1d9}, {0x1db, 0x1db}, {0x1de, 0x1de},
  {0x1e0, 0x1e0}, {0x1e2, 0x1e2}, {0x1e4, 0x1e4}, {0x1e6, 0x1e6}, {0x1e8, 0x1e8}, {0x1ea, 0x1ea}, {0x1ec, 0x1ec}, {0x1ee, 0x1ee},
  {0x1f1, 0x1f1}, {0x1f4, 0x1f4}, {0x1f6, 0x1f8}, {0x1fa, 0x1fa}, {0x1fc, 0x1fc}, {0x1fe, 0x1fe}, {0x200, 0x200}, {0x202, 0x202},
  {0x204, 0x204}, {0x206, 0x206}, {0x208, 0x208}, {0x20a, 0x20a}, {0x20c, 0x20c}, {0x20e, 0x20e}, {0x210, 0x210}, {0x212, 0x212},
  {0x214, 0x214}, {0x216, 0x216}, {0x218, 0x218}, {0x21a, 0x21a}, {0x21c, 0x21c}, {0x21e, 0x21e}, {0x220, 0x220}, {0x222, 0x222},
  {0x224, 0x224}, {0x226, 0x226}, {0x228, 0x228}, {0x22a, 0x22a}, {0x22c, 0x22c}, {0x22e, 0x22e}, {0x230, 0x230}, {0x232, 0x232},
  {0x23a, 0x23b}, {0x23d, 0x23e}, {0x241, 0x241}, {0x243, 0x246}, {0x248, 0x248}, {0x24a, 0x24a}, {0x24c, 0x24c}, {0x24e, 0x24e},
  {0x370, 0x370}, {0x372, 0x372}, {0x376, 0x376}, {0x37f, 0x37f}, {0x386, 0x386}, {0x388, 0x38a}, {0x38c, 0x38c}, {0x38e, 0x38f},
  {0x391, 0x3a1}, {0x3a3, 0x3ab}, {0x3cf, 0x3cf}, {0x3d2, 0x3d4}, {0x3d8, 0x3d8}, {0x3da, 0x3da}, {0x3dc, 0x3dc}, {0x3de, 0x3de},
  {0x3e0, 0x3e0}, {0x3e2, 0x3e2}, {0x3e4, 0x3e4}, {0x3e6, 0x3e6}, {0x3e8, 0x3e8}, {0x3ea, 0x3ea}, {0x3ec, 0x3ec}, {0x3ee, 0x3ee},
  {0x3f4, 0x3f4}, {0x3f7, 0x3f7}, {0x3f9, 0x3fa}, {0x3fd, 0x42f}, {0x460, 0x460}, {0x462, 0x462}, {0x464, 0x464}, {0x466, 0x466},
  {0x468, 0x468}, {0x46a, 0x46a}, {0x46c, 0x46c}, {0x46e, 0x46e}, {0x470, 0x470}, {0x472, 0x472}, {0x474, 0x474}, {0x476, 0x476},
  {0x478, 0x478}, {0x47a, 0x47a}, {0x47c, 0x47c}, {0x47e, 0x47e}, {0x480, 0x480}, {0x48a, 0x48a}, {0x48c, 0x48c}, {0x48e, 0x48e},
  {0x490, 0x490}, {0x492, 0x492}, {0x494, 0x494}, {0x496, 0x496}, {0x498, 0x498}, {0x49a, 0x49a}, {0x49c, 0x49c}, {0x49e, 0x49e},
  {0x4a0, 0x4a0}, {0x4a2, 0x4a2}, {0x4a4, 0x4a4}, {0x4a6, 0x4a6}, {0x4a8, 0x4a8}, {0x4aa, 0x4aa}, {0x4ac, 0x4ac}, {0x4ae, 0x4ae},
  {0x4b0, 0x4b0}, {0x4b2, 0x4b2}, {0x4b4, 0x4b4}, {0x4b6, 0x4b6}, {0x4b8, 0x4b8}, {0x4ba, 0x4ba}, {0x4bc, 0x4bc}, {0x4be, 0x4be},
  {0x4c0, 0x4c1}, {0x4c3, 0x4c3}, {0x4c5, 0x4c5}, {0x4c7, 0x4c7}, {0x4c9, 0x4c9}, {0x4cb, 0x4cb}, {0x4cd, 0x4cd}, {0x4d0, 0x4d0},
  {0x4d2, 0x4d2}, {0x4d4, 0x4d4}, {0x4d6, 0x4d6}, {0x4d8, 0x4d8}, {0x4da, 0x4da}, {0x4dc, 0x4dc}, {0x4de, 0x4de}, {0x4e0, 0x4e0},
  {0x4e2, 0x4e2}, {0x4e4, 0x4e4}, {0x4e6, 0x4e6}, {0x4e8, 0x4e8}, {0x4ea, 0x4ea}, {0x4ec, 0x4ec}, {0x4ee, 0x4ee}, {0x4f0, 0x4f0},
  {0x4f2, 0x4f2}, {0x4f4, 0x4f4}, {0x4f6, 0x4f6}, {0x4f8, 0x4f8}, {0x4fa, 0x4fa}, {0x4fc, 0x4fc}, {0x4fe, 0x4fe}, {0x500, 0x500},
  {0x502, 0x502}, {0x504, 0x504}, {0x506, 0x506}, {0x508, 0x508}, {0x50a, 0x50a}, {0x50c, 0x50c}, {0x50e, 0x50e}, {0x510, 0x510},
  {0x512, 0x512}, {0x514, 0x514}, {0x516, 0x516}, {0x518, 0x518}, {0x51a, 0x51a}, {0x51c, 0x51c}, {0x51e, 0x51e}, {0x520, 0x520},
  {0x522, 0x522}, {0x524, 0x524}, {0x526, 0x526}, {0x528, 0x528}, {0x52a, 0x52a}, {0x52c, 0x52c}, {0x52e, 0x52e}, {0x531, 0x556},
  {0x10a0, 0x10c5}, {0x10c7, 0x10c7}, {0x10cd, 0x10cd}, {0x13a0, 0x13f5}, {0x1c89, 0x1c89}, {0x1c90, 0x1cba}, {0x1cbd, 0x1cbf}, {0x1e00, 0x1e00},
  {0x1e02, 0x1e02}, {0x1e04, 0x1e04}, {0x1e06, 0x1e06}, {0x1e08, 0x1e08}, {0x1e0a, 0x1e0a}, {0x1e0c, 0x1e0c}, {0x1e0e, 0x1e0e}, {0x1e10, 0x1e10},
  {0x1e12, 0x1e12}, {0x1e14, 0x1e14}, {0x1e16, 0x1e16}, {0x1e18, 0x1e18}, {0x1e1a, 0x1e1a}, {0x1e1c, 0x1e1c}, {0x1e1e, 0x1e1e}, {0x1e20, 0x1e20},
  {0x1e22, 0x1e22}, {0x1e24, 0x1e24}, {0x1e26, 0x1e26}, {0x1e28, 0x1e28}, {0x1e2a, 0x1e2a}, {0x1e2c, 0x1e2c}, {0x1e2e, 0x1e2e}, {0x1e30, 0x1e30},
  {0x1e32, 0x1e32}, {0x1e34, 0x1e34}, {0x1e36, 0x1e36}, {0x1e38, 0x1e38}, {0x1e3a, 0x1e3a}, {0x1e3c, 0x1e3c}, {0x1e3e, 0x1e3e}, {0x1e40, 0x1e40},
  {0x1e42, 0x1e42}, {0x1e44, 0x1e44}, {0x1e46, 0x1e46}, {0x1e48, 0x1e48}, {0x1e4a, 0x1e4a}, {0x1e4c, 0x1e4c}, {0x1e4e, 0x1e4e}, {0x1e50, 0x1e50},
  {0x1e52, 0x1e52}, {0x1e54, 0x1e54}, {0x1e56, 0x1e56}, {0x1e58, 0x1e58}, {0x1e5a, 0x1e5a}, {0x1e5c, 0x1e5c}, {0x1e5e, 0x1e5e}, {0x1e60, 0x1e60},
  {0x1e62, 0x1e62}, {0x1e64, 0x1e64}, {0x1e66, 0x1e66}, {0x1e68, 0x1e68}, {0x1e6a, 0x1e6a}, {0x1e6c, 0x1e6c}, {0x1e6e, 0x1e6e}, {0x1e70, 0x1e70},
  {0x1e72, 0x1e72}, {0x1e74, 0x1e74}, {0x1e76, 0x1e76}, {0x1e78, 0x1e78}, {0x1e7a, 0x1e7a}, {0x1e7c, 0x1e7c}, {0x1e7e, 0x1e7e}, {0x1e80, 0x1e80},
  {0x1e82, 0x1e82}, {0x1e84, 0x1e84}, {0x1e86, 0x1e86}, {0x1e88, 0x1e88}, {0x1e8a, 0x1e8a}, {0x1e8c, 0x1e8c}, {0x1e8e, 0x1e8e}, {0x1e90, 0x1e90},
  {0x1e92, 0x1e92}, {0x1e94, 0x1e94}, {0x1e9e, 0x1e9e}, {0x1ea0, 0x1ea0}, {0x1ea2, 0x1ea2}, {0x1ea4, 0x1ea4}, {0x1ea6, 0x1ea6}, {0x1ea8, 0x1ea8},
  {0x1eaa, 0x1eaa}, {0x1eac, 0x1eac}, {0x1eae, 0x1eae}, {0x1eb0, 0x1eb0}, {0x1eb2, 0x1eb2}, {0x1eb4, 0x1eb4}, {0x1eb6, 0x1eb6}, {0x1eb8, 0x1eb8},
  {0x1eba, 0x1eba}, {0x1ebc, 0x1ebc}, {0x1ebe, 0x1ebe}, {0x1ec0, 0x1ec0}, {0x1ec2, 0x1ec2}, {0x1ec4, 0x1ec4}, {0x1ec6, 0x1ec6}, {0x1ec8, 0x1ec8},
  {0x1eca, 0x1eca}, {0x1ecc, 0x1ecc}, {0x1ece, 0x1ece}, {0x1ed0, 0x1ed0}, {0x1ed2, 0x1ed2}, {0x1ed4, 0x1ed4}, {0x1ed6, 0x1ed6}, {0x1ed8, 0x1ed8},
  {0x1eda, 0x1eda}, {0x1edc, 0x1edc}, {0x1ede, 0x1ede}, {0x1ee0, 0x1ee0}, {0x1ee2, 0x1ee2}, {0x1ee4, 0x1ee4}, {0x1ee6, 0x1ee6}, {0x1ee8, 0x1ee8},
  {0x1eea, 0x1eea}, {0x1eec, 0x1eec}, {0x1eee, 0x1eee}, {0x1ef0, 0x1ef0}, {0x1ef2, 0x1ef2}, {0x1ef4, 0x1ef4}, {0x1ef6, 0x1ef6}, {0x1ef8, 0x1ef8},
  {0x1efa, 0x1efa}, {0x1efc, 0x1efc}, {0x1efe, 0x1efe}, {0x1f08, 0x1f0f}, {0x1f18, 0x1f1d}, {0x1f28, 0x1f2f}, {0x1f38, 0x1f3f}, {0x1f48, 0x1f4d},
  {0x1f59, 0x1f59}, {0x1f5b, 0x1f5b}, {0x1f5d, 0x1f5d}, {0x1f5f, 0x1f5f}, {0x1f68, 0x1f6f}, {0x1fb8, 0x1fbb}, {0x1fc8, 0x1fcb}, {0x1fd8, 0x1fdb},
  {0x1fe8, 0x1fec}, {0x1ff8, 0x1ffb}, {0x2102, 0x2102}, {0x2107, 0x2107}, {0x210b, 0x210d}, {0x2110, 0x2112}, {0x2115, 0x2115}, {0x2119, 0x211d},
  {0x2124, 0x2124}, {0x2126, 0x2126}, {0x2128, 0x2128}, {0x212a, 0x212d}, {0x2130, 0x2133}, {0x213e, 0x213f}, {0x2145, 0x2145}, {0x2183, 0x2183},
  {0x2c00, 0x2c2f}, {0x2c60, 0x2c60}, {0x2c62, 0x2c64}, {0x2c67, 0x2c67}, {0x2c69, 0x2c69}, {0x2c6b, 0x2c6b}, {0x2c6d, 0x2c70}, {0x2c72, 0x2c72},
  {0x2c75, 0x2c75}, {0x2c7e, 0x2c80}, {0x2c82, 0x2c82}, {0x2c84, 0x2c84}, {0x2c86, 0x2c86}, {0x2c88, 0x2c88}, {0x2c8a, 0x2c8a}, {0x2c8c, 0x2c8c},
  {0x2c8e, 0x2c8e}, {0x2c90, 0x2c90}, {0x2c92, 0x2c92}, {0x2c94, 0x2c94}, {0x2c96, 0x2c96}, {0x2c98, 0x2c98}, {0x2c9a, 0x2c9a}, {0x2c9c, 0x2c9c},
  {0x2c9e, 0x2c9e}, {0x2ca0, 0x2ca0}, {0x2ca2, 0x2ca2}, {0x2ca4, 0x2ca4}, {0x2ca6, 0x2ca6}, {0x2ca8, 0x2ca8}, {0x2caa, 0x2caa}, {0x2cac, 0x2cac},
  {0x2cae, 0x2cae}, {0x2cb0, 0x2cb0}, {0x2cb2, 0x2cb2}, {0x2cb4, 0x2cb4}, {0x2cb6, 0x2cb6}, {0x2cb8, 0x2cb8}, {0x2cba, 0x2cba}, {0x2cbc, 0x2cbc},
  {0x2cbe, 0x2cbe}, {0x2cc0, 0x2cc0}, {0x2cc2, 0x2cc2}, {0x2cc4, 0x2cc4}, {0x2cc6, 0x2cc6}, {0x2cc8, 0x2cc8}, {0x2cca, 0x2cca}, {0x2ccc, 0x2ccc},
  {0x2cce, 0x2cce}, {0x2cd0, 0x2cd0}, {0x2cd2, 0x2cd2}, {0x2cd4, 0x2cd4}, {0x2cd6, 0x2cd6}, {0x2cd8, 0x2cd8}, {0x2cda, 0x2cda}, {0x2cdc, 0x2cdc},
  {0x2cde, 0x2cde}, {0x2ce0, 0x2ce0}, {0x2ce2, 0x2ce2}, {0x2ceb, 0x2ceb}, {0x2ced, 0x2ced}, {0x2cf2, 0x2cf2}, {0xa640, 0xa640}, {0xa642, 0xa642},
  {0xa644, 0xa644}, {0xa646, 0xa646}, {0xa648, 0xa648}, {0xa64a, 0xa64a}, {0xa64c, 0xa64c}, {0xa64e, 0xa64e}, {0xa650, 0xa650}, {0xa652, 0xa652},
  {0xa654, 0xa654}, {0xa656, 0xa656}, {0xa658, 0xa658}, {0xa65a, 0xa65a}, {0xa65c, 0xa65c}, {0xa65e, 0xa65e}, {0xa660, 0xa660}, {0xa662, 0xa662},
  {0xa664, 0xa664}, {0xa666, 0xa666}, {0xa668, 0xa668}, {0xa66a, 0xa66a}, {0xa66c, 0xa66c}, {0xa680, 0xa680}, {0xa682, 0xa682}, {0xa684, 0xa684},
  {0xa686, 0xa686}, {0xa688, 0xa688}, {0xa68a, 0xa68a}, {0xa68c, 0xa68c}, {0xa68e, 0xa68e}, {0xa690, 0xa690}, {0xa692, 0xa692}, {0xa694, 0xa694},
  {0xa696, 0xa696}, {0xa698, 0xa698}, {0xa69a, 0xa69a}, {0xa722, 0xa722}, {0xa724, 0xa724}, {0xa726, 0xa726}, {0xa728, 0xa728}, {0xa72a, 0xa72a},
  {0xa72c, 0xa72c}, {0xa72e, 0xa72e}, {0xa732, 0xa732}, {0xa734, 0xa734}, {0xa736, 0xa736}, {0xa738, 0xa738}, {0xa73a, 0xa73a}, {0xa73c, 0xa73c},
  {0xa73e, 0xa73e}, {0xa740, 0xa740}, {0xa742, 0xa742}, {0xa744, 0xa744}, {0xa746, 0xa746}, {0xa748, 0xa748}, {0xa74a, 0xa74a}, {0xa74c, 0xa74c},
  {0xa74e, 0xa74e}, {0xa750, 0xa750}, {0xa752, 0xa752}, {0xa754, 0xa754}, {0xa756, 0xa756}, {0xa758, 0xa758}, {0xa75a, 0xa75a}, {0xa75c, 0xa75c},
  {0xa75e, 0xa75e}, {0xa760, 0xa760}, {0xa762, 0xa762}, {0xa764, 0xa764}, {0xa766, 0xa766}, {0xa768, 0xa768}, {0xa76a, 0xa76a}, {0xa76c, 0xa76c},
  {0xa76e, 0xa76e}, {0xa779, 0xa779}, {0xa77b, 0xa77b}, {0xa77d, 0xa77e}, {0xa780, 0xa780}, {0xa782, 0xa782}, {0xa784, 0xa784}, {0xa786, 0xa786},
  {0xa78b, 0xa78b}, {0xa78d, 0xa78d}, {0xa790, 0xa790}, {0xa792, 0xa792}, {0xa796, 0xa796}, {0xa798, 0xa798}, {0xa79a, 0xa79a}, {0xa79c, 0xa79c},
  {0xa79e, 0xa79e}, {0xa7a0, 0xa7a0}, {0xa7a2, 0xa7a2}, {0xa7a4, 0xa7a4}, {0xa7a6, 0xa7a6}, {0xa7a8, 0xa7a8}, {0xa7aa, 0xa7ae}, {0xa7b0, 0xa7b4},
  {0xa7b6, 0xa7b6}, {0xa7b8, 0xa7b8}, {0xa7ba, 0xa7ba}, {0xa7bc, 0xa7bc}, {0xa7be, 0xa7be}, {0xa7c0, 0xa7c0}, {0xa7c2, 0xa7c2}, {0xa7c4, 0xa7c7},
  {0xa7c9, 0xa7c9}, {0xa7cb, 0xa7cc}, {0xa7d0, 0xa7d0}, {0xa7d6, 0xa7d6}, {0xa7d8, 0xa7d8}, {0xa7da, 0xa7da}, {0xa7dc, 0xa7dc}, {0xa7f5, 0xa7f5},
  {0xff21, 0xff3a}, {0x10400, 0x10427}, {0x104b0, 0x104d3}, {0x10570, 0x1057a}, {0x1057c, 0x1058a}, {0x1058c, 0x10592}, {0x10594, 0x10595}, {0x10c80, 0x10cb2},
  {0x10d50, 0x10d65}, {0x118a0, 0x118bf}, {0x16e40, 0x16e5f}, {0x1d400, 0x1d419}, {0x1d434, 0x1d44d}, {0x1d468, 0x1d481}, {0x1d49c, 0x1d49c}, {0x1d49e, 0x1d49f},
  {0x1d4a2, 0x1d4a2}, {0x1d4a5, 0x1d4a6}, {0x1d4a9, 0x1d4ac}, {0x1d4ae, 0x1d4b5}, {0x1d4d0, 0x1d4e9}, {0x1d504, 0x1d505}, {0x1d507, 0x1d50a}, {0x1d50d, 0x1d514},
  {0x1d516, 0x1d51c}, {0x1d538, 0x1d539}, {0x1d53b, 0x1d53e}, {0x1d540, 0x1d544}, {0x1d546, 0x1d546}, {0x1d54a, 0x1d550}, {0x1d56c, 0x1d585}, {0x1d5a0, 0x1d5b9},
  {0x1d5d4, 0x1d5ed}, {0x1d608, 0x1d621}, {0x1d63c, 0x1d655}, {0x1d670, 0x1d689}, {0x1d6a8, 0x1d6c0}, {0x1d6e2, 0x1d6fa}, {0x1d71c, 0x1d734}, {0x1d756, 0x1d76e},
  {0x1d790, 0x1d7a8}, {0x1d7ca, 0x1d7ca}, {0x1e900, 0x1e921},
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(152);
      ADVANCE_MAP(
        '!', 180,
        '"', 282,
        '#', 233,
        '$', 132,
        '%', 256,
        '&', 228,
        '\'', 278,
        '(', 169,
        ')', 170,
        '*', 213,
        '+', 176,
        ',', 174,
        '-', 178,
        '.', 197,
        '0', 259,
        ':', 165,
        ';', 221,
        '<', 231,
        '=', 158,
        '>', 226,
        '?', 161,
        '@', 316,
        '[', 200,
        '\\', 2,
        ']', 211,
        '^', 318,
        '`', 358,
        'a', 344,
        'l', 340,
        'm', 335,
        '{', 216,
        '|', 183,
        '}', 222,
        '~', 159,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(145);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 1:
      if (lookahead == '\n') ADVANCE(290);
      if (lookahead == '\r') ADVANCE(1);
      END_STATE();
    case 2:
      ADVANCE_MAP(
        '\n', 290,
        '\r', 1,
        '#', 144,
        'o', 129,
        'u', 121,
        'x', 143,
        ' ', 296,
        '"', 296,
        '\'', 296,
        '\\', 296,
        'b', 296,
        'n', 296,
        'r', 296,
        't', 296,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(136);
      END_STATE();
    case 3:
      ADVANCE_MAP(
        '\n', 290,
        '\r', 1,
        'o', 129,
        'u', 121,
        'x', 143,
        ' ', 296,
        '"', 296,
        '\'', 296,
        '\\', 296,
        'b', 296,
        'n', 296,
        'r', 296,
        't', 296,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(136);
      END_STATE();
    case 4:
      if (lookahead == '\n') ADVANCE(279);
      if (lookahead == '\r') ADVANCE(5);
      if (lookahead == '[') ADVANCE(281);
      if (lookahead == '\\') ADVANCE(120);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') ADVANCE(280);
      if (lookahead != 0 &&
          lookahead != '\'') ADVANCE(280);
      END_STATE();
    case 5:
      if (lookahead == '\n') ADVANCE(279);
      if (lookahead == '\r') ADVANCE(5);
      if (lookahead == '[') ADVANCE(113);
      if (lookahead == '\\') ADVANCE(120);
      if (('\t' <= lookahead && lookahead <= '\f') ||
          lookahead == ' ') SKIP(116);
      END_STATE();
    case 6:
      ADVANCE_MAP(
        '!', 180,
        '"', 282,
        '%', 256,
        '&', 227,
        '\'', 277,
        '(', 169,
        ')', 170,
        '*', 212,
        '+', 270,
        ',', 174,
        '-', 276,
        '.', 97,
        '0', 259,
        ':', 189,
        ';', 104,
        '=', 158,
        '>', 226,
        '[', 203,
        '\\', 74,
        ']', 211,
        '{', 215,
        '|', 181,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(6);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 7:
      ADVANCE_MAP(
        '!', 180,
        '%', 256,
        '\'', 357,
        '(', 169,
        '+', 175,
        '-', 177,
        '[', 113,
        '\\', 74,
        'a', 345,
        'l', 341,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(7);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 8:
      ADVANCE_MAP(
        '!', 180,
        '\'', 357,
        '(', 169,
        '+', 175,
        '-', 179,
        '.', 359,
        '[', 205,
        '\\', 74,
        'a', 345,
        'l', 341,
        '{', 79,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(8);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 9:
      ADVANCE_MAP(
        '!', 180,
        '\'', 357,
        '+', 175,
        ',', 174,
        '-', 177,
        '.', 97,
        ':', 101,
        ';', 220,
        '[', 113,
        '\\', 74,
        '|', 182,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(9);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 10:
      ADVANCE_MAP(
        '!', 180,
        '\'', 357,
        '+', 175,
        ',', 174,
        '-', 177,
        '.', 97,
        ':', 164,
        ';', 220,
        '[', 113,
        '\\', 74,
        '|', 182,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(9);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 11:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '#', 233,
        '$', 323,
        '%', 257,
        '&', 229,
        '\'', 277,
        '(', 169,
        ')', 170,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 274,
        '.', 196,
        '/', 313,
        '0', 259,
        ':', 191,
        ';', 220,
        '<', 320,
        '=', 324,
        '>', 322,
        '?', 140,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 340,
        'm', 335,
        '{', 216,
        '|', 185,
        '}', 222,
        '~', 160,
        '@', 318,
        '^', 318,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(11);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 12:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '#', 233,
        '$', 323,
        '&', 229,
        '\'', 277,
        '(', 169,
        ')', 170,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 274,
        '.', 196,
        '0', 259,
        ':', 191,
        ';', 220,
        '<', 321,
        '=', 324,
        '>', 322,
        '?', 162,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 341,
        '{', 216,
        '|', 185,
        '}', 222,
        '~', 160,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(12);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 13:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '#', 233,
        '$', 323,
        '&', 229,
        '\'', 277,
        '(', 169,
        ')', 170,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 274,
        '.', 196,
        '0', 259,
        ':', 191,
        ';', 220,
        '<', 321,
        '=', 324,
        '>', 322,
        '?', 162,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 341,
        '{', 216,
        '|', 184,
        '}', 222,
        '~', 160,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(13);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 14:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '#', 233,
        '$', 323,
        '&', 229,
        '\'', 277,
        '(', 169,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 275,
        '.', 196,
        '0', 259,
        ':', 166,
        ';', 220,
        '<', 321,
        '?', 162,
        '[', 208,
        '\\', 74,
        '`', 358,
        'a', 344,
        'l', 341,
        '{', 218,
        '|', 185,
        '~', 160,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(17);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 15:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '#', 233,
        '$', 323,
        '&', 229,
        '\'', 277,
        '(', 169,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 275,
        '.', 196,
        '0', 259,
        ':', 103,
        ';', 221,
        '<', 321,
        '?', 162,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 344,
        'l', 340,
        '{', 216,
        '|', 185,
        '~', 160,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(15);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 16:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '#', 233,
        '$', 323,
        '&', 229,
        '\'', 277,
        '(', 169,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 275,
        '.', 196,
        '0', 259,
        ':', 103,
        ';', 221,
        '<', 321,
        '?', 162,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 340,
        '{', 216,
        '|', 185,
        '~', 160,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(16);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 17:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '#', 233,
        '$', 323,
        '&', 229,
        '\'', 277,
        '(', 169,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 275,
        '.', 196,
        '0', 259,
        ':', 103,
        ';', 220,
        '<', 321,
        '?', 162,
        '[', 208,
        '\\', 74,
        '`', 358,
        'a', 344,
        'l', 341,
        '{', 218,
        '|', 185,
        '~', 160,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(17);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 18:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '#', 233,
        '&', 229,
        '\'', 277,
        '(', 169,
        ')', 170,
        '*', 312,
        '+', 273,
        '-', 275,
        '.', 196,
        '0', 259,
        ':', 103,
        '<', 320,
        '?', 140,
        '[', 207,
        '\\', 74,
        '`', 358,
        'a', 344,
        'l', 340,
        'm', 335,
        '{', 217,
        '|', 122,
        '~', 160,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
        '$', 324,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(18);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 19:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '#', 356,
        '$', 323,
        '%', 257,
        '&', 229,
        '\'', 277,
        '(', 169,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 275,
        '/', 313,
        '0', 259,
        ':', 166,
        ';', 221,
        '<', 320,
        '?', 140,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 344,
        'l', 340,
        'm', 335,
        '{', 216,
        '|', 185,
        '~', 160,
        '@', 318,
        '^', 318,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(20);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 20:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '#', 356,
        '$', 323,
        '%', 257,
        '&', 229,
        '\'', 277,
        '(', 169,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 275,
        '/', 313,
        '0', 259,
        ':', 103,
        ';', 221,
        '<', 320,
        '?', 140,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 344,
        'l', 340,
        'm', 335,
        '{', 216,
        '|', 185,
        '~', 160,
        '@', 318,
        '^', 318,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(20);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 21:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '#', 356,
        '$', 323,
        '%', 257,
        '&', 229,
        '\'', 277,
        '(', 169,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 275,
        '/', 313,
        '0', 259,
        ':', 103,
        ';', 221,
        '<', 320,
        '?', 140,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 340,
        'm', 335,
        '{', 216,
        '|', 185,
        '~', 160,
        '@', 318,
        '^', 318,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(21);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 22:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '#', 356,
        '$', 323,
        '&', 229,
        '\'', 277,
        '(', 169,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 275,
        '0', 259,
        ':', 166,
        ';', 221,
        '<', 320,
        '?', 162,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 344,
        'l', 340,
        '{', 216,
        '|', 185,
        '~', 160,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(24);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 23:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '#', 356,
        '$', 323,
        '&', 229,
        '\'', 277,
        '(', 169,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 275,
        '0', 259,
        ':', 166,
        ';', 221,
        '<', 320,
        '?', 162,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 340,
        '{', 216,
        '|', 185,
        '~', 160,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(25);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 24:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '#', 356,
        '$', 323,
        '&', 229,
        '\'', 277,
        '(', 169,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 275,
        '0', 259,
        ':', 103,
        ';', 221,
        '<', 320,
        '?', 162,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 344,
        'l', 340,
        '{', 216,
        '|', 185,
        '~', 160,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(24);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 25:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '#', 356,
        '$', 323,
        '&', 229,
        '\'', 277,
        '(', 169,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 275,
        '0', 259,
        ':', 103,
        ';', 221,
        '<', 320,
        '?', 162,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 340,
        '{', 216,
        '|', 185,
        '~', 160,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(25);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 26:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '#', 139,
        '$', 323,
        '&', 229,
        '\'', 277,
        '(', 169,
        ')', 170,
        '*', 312,
        '+', 273,
        '-', 275,
        '.', 196,
        '0', 259,
        ':', 103,
        ';', 220,
        '<', 320,
        '?', 140,
        '[', 207,
        '\\', 74,
        '`', 358,
        'a', 344,
        'l', 340,
        'm', 335,
        '{', 218,
        '|', 118,
        '~', 160,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(26);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 27:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '$', 323,
        '%', 257,
        '&', 229,
        '\'', 277,
        '(', 169,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 275,
        '/', 313,
        '0', 259,
        ':', 103,
        ';', 220,
        '<', 320,
        '?', 140,
        '[', 207,
        '\\', 74,
        '`', 358,
        'a', 345,
        'l', 340,
        'm', 335,
        '{', 218,
        '|', 184,
        '~', 160,
        '@', 318,
        '^', 318,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(27);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 28:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '$', 323,
        '%', 257,
        '&', 229,
        '\'', 277,
        '(', 169,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 275,
        '/', 313,
        '0', 259,
        ':', 103,
        ';', 220,
        '<', 320,
        '?', 140,
        '[', 208,
        '\\', 74,
        '`', 358,
        'a', 344,
        'l', 340,
        'm', 335,
        '{', 218,
        '|', 185,
        '~', 160,
        '@', 318,
        '^', 318,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(28);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 29:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '$', 323,
        '&', 229,
        '\'', 277,
        '(', 169,
        ')', 170,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 274,
        '.', 359,
        '0', 259,
        ':', 167,
        ';', 220,
        '<', 320,
        '=', 324,
        '>', 322,
        '?', 162,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 341,
        '{', 216,
        '|', 184,
        '}', 222,
        '~', 160,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(30);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 30:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '$', 323,
        '&', 229,
        '\'', 277,
        '(', 169,
        ')', 170,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 274,
        '.', 359,
        '0', 259,
        ':', 191,
        ';', 220,
        '<', 320,
        '=', 324,
        '>', 322,
        '?', 162,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 341,
        '{', 216,
        '|', 184,
        '}', 222,
        '~', 160,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(30);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 31:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '$', 323,
        '&', 229,
        '\'', 277,
        '(', 169,
        ')', 170,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 274,
        '0', 259,
        ':', 167,
        ';', 220,
        '<', 320,
        '=', 324,
        '>', 322,
        '?', 162,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 341,
        '{', 216,
        '|', 185,
        '}', 222,
        '~', 160,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(32);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 32:
      ADVANCE_MAP(
        '!', 305,
        '"', 282,
        '$', 323,
        '&', 229,
        '\'', 277,
        '(', 169,
        ')', 170,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 274,
        '0', 259,
        ':', 191,
        ';', 220,
        '<', 320,
        '=', 324,
        '>', 322,
        '?', 162,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 341,
        '{', 216,
        '|', 185,
        '}', 222,
        '~', 160,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(32);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 33:
      ADVANCE_MAP(
        '!', 305,
        '#', 139,
        '&', 229,
        '*', 312,
        '+', 273,
        '-', 275,
        '.', 196,
        ':', 108,
        '<', 320,
        '[', 113,
        '\\', 74,
        'a', 346,
        'l', 342,
        'm', 335,
        '|', 122,
        '%', 313,
        '/', 313,
        '?', 140,
        '~', 140,
        '@', 318,
        '^', 318,
        '$', 324,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(33);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 34:
      ADVANCE_MAP(
        '!', 109,
        '"', 282,
        '#', 356,
        '&', 229,
        '(', 169,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 275,
        '.', 196,
        ':', 103,
        ';', 221,
        '<', 320,
        '[', 78,
        '\\', 74,
        ']', 211,
        'l', 340,
        '{', 215,
        '|', 185,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
        '$', 324,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(34);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 35:
      ADVANCE_MAP(
        '!', 109,
        '#', 233,
        '&', 229,
        '\'', 278,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 275,
        '.', 196,
        ':', 190,
        ';', 221,
        '<', 320,
        '[', 84,
        '\\', 74,
        ']', 211,
        'a', 344,
        'l', 340,
        '{', 86,
        '|', 185,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
        '$', 324,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(39);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 36:
      ADVANCE_MAP(
        '!', 109,
        '#', 233,
        '&', 229,
        '(', 169,
        ')', 170,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 274,
        '.', 196,
        ':', 191,
        ';', 220,
        '<', 320,
        '>', 322,
        '[', 209,
        '\\', 74,
        ']', 211,
        '{', 86,
        '|', 185,
        '}', 222,
        '$', 324,
        '=', 324,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(36);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 37:
      ADVANCE_MAP(
        '!', 109,
        '#', 233,
        '&', 229,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 275,
        '.', 196,
        ':', 166,
        ';', 220,
        '<', 320,
        '[', 115,
        '\\', 74,
        'a', 344,
        '|', 185,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
        '$', 324,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(38);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 38:
      ADVANCE_MAP(
        '!', 109,
        '#', 233,
        '&', 229,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 275,
        '.', 196,
        ':', 103,
        ';', 220,
        '<', 320,
        '[', 115,
        '\\', 74,
        'a', 344,
        '|', 185,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
        '$', 324,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(38);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 39:
      ADVANCE_MAP(
        '!', 109,
        '#', 233,
        '&', 229,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 275,
        '.', 196,
        ':', 190,
        ';', 221,
        '<', 320,
        '[', 84,
        '\\', 74,
        ']', 211,
        'a', 344,
        'l', 340,
        '{', 86,
        '|', 185,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
        '$', 324,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(39);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 40:
      ADVANCE_MAP(
        '!', 109,
        '#', 356,
        '&', 229,
        '\'', 357,
        '(', 169,
        ')', 170,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 275,
        '.', 196,
        ':', 191,
        ';', 221,
        '<', 320,
        '[', 209,
        '\\', 74,
        ']', 211,
        'a', 344,
        'l', 340,
        '{', 219,
        '|', 185,
        '}', 222,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
        '$', 324,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(40);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 41:
      ADVANCE_MAP(
        '!', 109,
        '#', 356,
        '&', 229,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 275,
        ':', 166,
        ';', 221,
        '<', 320,
        '[', 84,
        '\\', 74,
        ']', 211,
        'a', 344,
        'l', 340,
        '{', 86,
        '|', 185,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
        '$', 324,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(42);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 42:
      ADVANCE_MAP(
        '!', 109,
        '#', 356,
        '&', 229,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 275,
        ':', 103,
        ';', 221,
        '<', 320,
        '[', 84,
        '\\', 74,
        ']', 211,
        'a', 344,
        'l', 340,
        '{', 86,
        '|', 185,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
        '$', 324,
        '=', 324,
        '>', 324,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(42);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 43:
      ADVANCE_MAP(
        '!', 109,
        '&', 229,
        ')', 170,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 274,
        ':', 167,
        ';', 220,
        '<', 320,
        '>', 322,
        '[', 84,
        '\\', 74,
        ']', 211,
        '{', 86,
        '|', 185,
        '}', 222,
        '$', 324,
        '=', 324,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(45);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 44:
      ADVANCE_MAP(
        '!', 109,
        '&', 229,
        ')', 170,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 274,
        ':', 167,
        ';', 220,
        '<', 320,
        '>', 322,
        '[', 84,
        '\\', 74,
        ']', 211,
        '{', 86,
        '|', 184,
        '}', 222,
        '$', 324,
        '=', 324,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(46);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 45:
      ADVANCE_MAP(
        '!', 109,
        '&', 229,
        ')', 170,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 274,
        ':', 191,
        ';', 220,
        '<', 320,
        '>', 322,
        '[', 84,
        '\\', 74,
        ']', 211,
        '{', 86,
        '|', 185,
        '}', 222,
        '$', 324,
        '=', 324,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(45);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 46:
      ADVANCE_MAP(
        '!', 109,
        '&', 229,
        ')', 170,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 274,
        ':', 191,
        ';', 220,
        '<', 320,
        '>', 322,
        '[', 84,
        '\\', 74,
        ']', 211,
        '{', 86,
        '|', 184,
        '}', 222,
        '$', 324,
        '=', 324,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(46);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 47:
      ADVANCE_MAP(
        '!', 109,
        '&', 229,
        ')', 170,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 274,
        ':', 191,
        ';', 220,
        '<', 320,
        '>', 322,
        '[', 115,
        '\\', 74,
        ']', 211,
        'a', 344,
        '|', 184,
        '}', 222,
        '$', 324,
        '=', 324,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(47);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 48:
      ADVANCE_MAP(
        '!', 304,
        '"', 282,
        '#', 233,
        '$', 132,
        '\'', 277,
        '(', 169,
        '.', 196,
        '0', 259,
        ':', 163,
        ';', 104,
        '?', 162,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 341,
        '{', 216,
        '~', 160,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(49);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 49:
      ADVANCE_MAP(
        '!', 304,
        '"', 282,
        '#', 233,
        '$', 132,
        '\'', 277,
        '(', 169,
        '.', 196,
        '0', 259,
        ';', 104,
        '?', 162,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 341,
        '{', 216,
        '~', 160,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(49);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 50:
      ADVANCE_MAP(
        '!', 304,
        '"', 282,
        '#', 232,
        '$', 132,
        '%', 256,
        '\'', 277,
        '(', 169,
        ')', 170,
        '+', 270,
        ',', 174,
        '-', 276,
        '.', 97,
        '0', 259,
        ':', 187,
        ';', 104,
        '=', 158,
        '?', 162,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 341,
        '{', 216,
        '|', 181,
        '~', 160,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(50);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 51:
      ADVANCE_MAP(
        '!', 304,
        '"', 282,
        '#', 356,
        '$', 132,
        '%', 256,
        '\'', 277,
        '(', 169,
        ')', 170,
        '+', 271,
        '-', 272,
        '.', 196,
        '0', 259,
        ':', 194,
        ';', 104,
        '<', 96,
        '=', 158,
        '?', 162,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 340,
        'm', 335,
        '{', 216,
        '|', 117,
        '}', 222,
        '~', 160,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(51);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 52:
      ADVANCE_MAP(
        '!', 304,
        '"', 282,
        '#', 356,
        '$', 132,
        '%', 256,
        '\'', 277,
        '(', 169,
        '0', 259,
        ';', 104,
        '?', 140,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 344,
        'l', 340,
        'm', 335,
        '{', 216,
        '~', 160,
        '+', 271,
        '-', 271,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(52);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 53:
      ADVANCE_MAP(
        '!', 304,
        '"', 282,
        '#', 356,
        '$', 132,
        '\'', 277,
        '(', 169,
        '.', 359,
        '0', 259,
        ':', 186,
        ';', 104,
        '?', 162,
        '[', 206,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 340,
        'm', 335,
        '{', 216,
        '~', 160,
        '+', 271,
        '-', 271,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(53);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 54:
      ADVANCE_MAP(
        '"', 282,
        '#', 232,
        '%', 256,
        '\'', 277,
        '(', 169,
        '+', 270,
        '-', 276,
        '.', 196,
        '0', 259,
        ':', 195,
        ';', 104,
        '=', 158,
        '?', 161,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 341,
        '{', 215,
        '|', 117,
        '~', 159,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(54);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 55:
      ADVANCE_MAP(
        '"', 282,
        '#', 232,
        '\'', 277,
        '(', 169,
        '*', 212,
        '+', 270,
        '-', 276,
        '.', 196,
        '0', 259,
        ':', 193,
        ';', 104,
        '=', 158,
        '?', 161,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 340,
        '{', 215,
        '|', 181,
        '~', 159,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(55);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 56:
      ADVANCE_MAP(
        '"', 282,
        '#', 232,
        '\'', 277,
        '(', 169,
        '+', 270,
        ',', 174,
        '-', 276,
        '.', 196,
        '0', 259,
        ':', 192,
        ';', 220,
        '=', 158,
        '>', 124,
        '?', 161,
        '[', 204,
        '\\', 74,
        '`', 358,
        'a', 345,
        'l', 341,
        '{', 215,
        '|', 182,
        '~', 159,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(56);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 57:
      ADVANCE_MAP(
        '"', 282,
        '#', 232,
        '\'', 277,
        '(', 169,
        '+', 270,
        ',', 174,
        '-', 276,
        '.', 196,
        '0', 259,
        ':', 168,
        ';', 220,
        '=', 158,
        '>', 124,
        '?', 161,
        '[', 204,
        '\\', 74,
        '`', 358,
        'a', 345,
        'l', 341,
        '{', 215,
        '|', 182,
        '~', 159,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(56);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 58:
      ADVANCE_MAP(
        '"', 282,
        '#', 232,
        '\'', 277,
        '(', 169,
        ',', 174,
        '.', 97,
        '0', 259,
        ':', 192,
        ';', 104,
        '=', 158,
        '?', 161,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 344,
        'l', 340,
        '{', 215,
        '|', 181,
        '~', 159,
        '+', 270,
        '-', 270,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(58);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 59:
      ADVANCE_MAP(
        '"', 282,
        '#', 232,
        '\'', 277,
        '(', 169,
        ',', 174,
        '.', 97,
        '0', 259,
        ':', 192,
        '=', 158,
        '?', 161,
        '[', 208,
        '\\', 74,
        '`', 358,
        'a', 344,
        'l', 341,
        '{', 217,
        '|', 181,
        '~', 159,
        '+', 270,
        '-', 270,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(59);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 60:
      ADVANCE_MAP(
        '"', 282,
        '#', 232,
        '\'', 277,
        '(', 169,
        ',', 174,
        '.', 196,
        '0', 259,
        ':', 192,
        ';', 104,
        '=', 158,
        '?', 161,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 344,
        'l', 340,
        '{', 215,
        '|', 181,
        '~', 159,
        '+', 270,
        '-', 270,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(60);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 61:
      ADVANCE_MAP(
        '"', 282,
        '#', 232,
        '\'', 277,
        '(', 169,
        ',', 174,
        '.', 196,
        '0', 259,
        ':', 192,
        ';', 220,
        '=', 158,
        '>', 226,
        '?', 161,
        '[', 208,
        '\\', 74,
        '`', 358,
        'a', 344,
        'l', 341,
        '{', 217,
        '|', 181,
        '}', 222,
        '~', 159,
        '+', 270,
        '-', 270,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(61);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 62:
      ADVANCE_MAP(
        '"', 282,
        '#', 232,
        '\'', 277,
        '(', 169,
        ',', 174,
        '0', 259,
        ':', 192,
        ';', 104,
        '=', 158,
        '?', 161,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 344,
        'l', 340,
        '{', 215,
        '|', 181,
        '~', 159,
        '+', 270,
        '-', 270,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(62);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 63:
      ADVANCE_MAP(
        '"', 282,
        '#', 232,
        '\'', 277,
        '(', 169,
        ',', 174,
        '0', 259,
        ':', 192,
        '=', 158,
        '?', 161,
        '[', 208,
        '\\', 74,
        '`', 358,
        'a', 344,
        'l', 341,
        '{', 217,
        '|', 181,
        '~', 159,
        '+', 270,
        '-', 270,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(63);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 64:
      ADVANCE_MAP(
        '"', 282,
        '#', 232,
        '\'', 277,
        '(', 169,
        ',', 174,
        '0', 259,
        ':', 168,
        ';', 104,
        '=', 158,
        '?', 161,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 344,
        'l', 340,
        '{', 215,
        '|', 181,
        '~', 159,
        '+', 270,
        '-', 270,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(62);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 65:
      ADVANCE_MAP(
        '"', 282,
        '#', 232,
        '\'', 277,
        '(', 169,
        ',', 174,
        '0', 259,
        ':', 168,
        '=', 158,
        '?', 161,
        '[', 208,
        '\\', 74,
        '`', 358,
        'a', 344,
        'l', 341,
        '{', 217,
        '|', 181,
        '~', 159,
        '+', 270,
        '-', 270,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(63);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 66:
      ADVANCE_MAP(
        '"', 282,
        '#', 232,
        '\'', 277,
        '(', 169,
        '.', 359,
        '0', 259,
        ':', 186,
        ';', 104,
        '=', 158,
        '?', 161,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 341,
        '{', 215,
        '|', 181,
        '~', 159,
        '+', 270,
        '-', 270,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(66);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 67:
      ADVANCE_MAP(
        '"', 282,
        '#', 232,
        '\'', 277,
        '(', 169,
        '.', 359,
        '0', 259,
        ':', 163,
        ';', 104,
        '=', 158,
        '?', 161,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 341,
        '{', 215,
        '|', 181,
        '~', 159,
        '+', 270,
        '-', 270,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(66);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 68:
      ADVANCE_MAP(
        '"', 282,
        '#', 232,
        '\'', 277,
        '(', 169,
        '0', 259,
        ':', 186,
        ';', 104,
        '=', 158,
        '?', 161,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 340,
        '{', 215,
        '~', 159,
        '+', 270,
        '-', 270,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(68);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 69:
      ADVANCE_MAP(
        '"', 282,
        '#', 232,
        '\'', 277,
        '(', 169,
        '0', 259,
        ':', 163,
        ';', 104,
        '=', 158,
        '?', 161,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 340,
        '{', 215,
        '~', 159,
        '+', 270,
        '-', 270,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(68);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 70:
      if (lookahead == '"') ADVANCE(282);
      if (lookahead == '%') ADVANCE(285);
      if (lookahead == '@') ADVANCE(286);
      if (lookahead == '[') ADVANCE(287);
      if (lookahead == '\\') ADVANCE(3);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(283);
      if (lookahead != 0) ADVANCE(288);
      END_STATE();
    case 71:
      ADVANCE_MAP(
        '#', 233,
        '(', 169,
        ')', 170,
        '*', 212,
        ',', 174,
        '-', 110,
        '.', 196,
        ':', 192,
        ';', 220,
        '=', 158,
        '>', 226,
        '[', 84,
        '\\', 74,
        ']', 211,
        '`', 358,
        '{', 86,
        '|', 181,
        '}', 222,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(71);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 72:
      ADVANCE_MAP(
        '#', 233,
        ')', 170,
        ',', 174,
        '-', 110,
        '.', 197,
        ':', 189,
        ';', 220,
        '=', 158,
        '>', 226,
        '[', 115,
        '\\', 74,
        ']', 211,
        'a', 345,
        '|', 181,
        '}', 222,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(72);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 73:
      ADVANCE_MAP(
        '#', 233,
        ')', 170,
        ',', 174,
        '-', 110,
        '.', 197,
        ':', 192,
        ';', 220,
        '=', 158,
        '>', 124,
        '[', 84,
        '\\', 74,
        ']', 211,
        '{', 86,
        '|', 181,
        '}', 222,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(73);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 74:
      if (lookahead == '#') ADVANCE(144);
      END_STATE();
    case 75:
      ADVANCE_MAP(
        '#', 232,
        '\'', 357,
        ')', 170,
        '*', 212,
        ',', 174,
        '-', 110,
        '.', 196,
        ':', 111,
        ';', 220,
        '=', 158,
        '>', 226,
        '[', 84,
        '\\', 74,
        'a', 345,
        'l', 341,
        '{', 86,
        '|', 181,
        '}', 222,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(75);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 76:
      ADVANCE_MAP(
        '#', 232,
        '\'', 357,
        '*', 212,
        '-', 110,
        '.', 196,
        ':', 106,
        ';', 104,
        '=', 158,
        '[', 84,
        '\\', 74,
        ']', 211,
        'a', 345,
        'l', 340,
        '{', 86,
        '|', 181,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(76);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 77:
      ADVANCE_MAP(
        '#', 356,
        ')', 170,
        ',', 174,
        '-', 110,
        ':', 188,
        ';', 104,
        '=', 158,
        '[', 84,
        '\\', 74,
        ']', 211,
        'l', 340,
        '{', 86,
        '|', 181,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(77);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 78:
      if (lookahead == '%') ADVANCE(250);
      if (lookahead == '@') ADVANCE(243);
      END_STATE();
    case 79:
      if (lookahead == '%') ADVANCE(252);
      END_STATE();
    case 80:
      if (lookahead == '%') ADVANCE(254);
      END_STATE();
    case 81:
      if (lookahead == '%') ADVANCE(255);
      END_STATE();
    case 82:
      if (lookahead == '%') ADVANCE(249);
      if (lookahead == '@') ADVANCE(244);
      END_STATE();
    case 83:
      if (lookahead == '%') ADVANCE(251);
      END_STATE();
    case 84:
      if (lookahead == '%') ADVANCE(80);
      if (lookahead == '@') ADVANCE(243);
      END_STATE();
    case 85:
      if (lookahead == '%') ADVANCE(292);
      if (lookahead == '@') ADVANCE(293);
      if (lookahead == '[') ADVANCE(294);
      if (lookahead == '|') ADVANCE(291);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(283);
      if (lookahead != 0) ADVANCE(295);
      END_STATE();
    case 86:
      if (lookahead == '%') ADVANCE(81);
      END_STATE();
    case 87:
      ADVANCE_MAP(
        '(', 169,
        ',', 174,
        '-', 110,
        ':', 102,
        ';', 220,
        '=', 158,
        '[', 82,
        '\\', 74,
        'a', 345,
        '{', 83,
        '|', 182,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(87);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 88:
      if (lookahead == '(') ADVANCE(169);
      if (lookahead == '[') ADVANCE(210);
      if (lookahead == '\\') ADVANCE(74);
      if (lookahead == 'a') ADVANCE(345);
      if (lookahead == 'l') ADVANCE(341);
      if (lookahead == '{') ADVANCE(214);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(88);
      if (set_contains(sym_indexing_operator_character_set_1, 9, lookahead)) ADVANCE(331);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 89:
      ADVANCE_MAP(
        ')', 170,
        ',', 174,
        '-', 110,
        '.', 97,
        ':', 187,
        ';', 220,
        '=', 158,
        '[', 113,
        '\\', 74,
        ']', 211,
        '|', 181,
        '}', 222,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(89);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 90:
      ADVANCE_MAP(
        ')', 170,
        ',', 174,
        '-', 110,
        ':', 187,
        ';', 220,
        '=', 158,
        '[', 113,
        '\\', 74,
        ']', 211,
        '|', 181,
        '}', 222,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(90);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 91:
      ADVANCE_MAP(
        ')', 170,
        ',', 174,
        '-', 110,
        ':', 164,
        ';', 220,
        '=', 158,
        '[', 113,
        '\\', 74,
        ']', 211,
        '|', 181,
        '}', 222,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(90);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 92:
      if (lookahead == '*') ADVANCE(100);
      if (lookahead == '.') ADVANCE(93);
      if (lookahead == 'L' ||
          lookahead == 'l' ||
          lookahead == 'n') ADVANCE(301);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(99);
      if (set_contains(sym_conversion_specification_character_set_6, 15, lookahead)) ADVANCE(300);
      END_STATE();
    case 93:
      if (lookahead == '*') ADVANCE(128);
      if (lookahead == 'L' ||
          lookahead == 'l' ||
          lookahead == 'n') ADVANCE(301);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(127);
      if (set_contains(sym_conversion_specification_character_set_7, 16, lookahead)) ADVANCE(300);
      END_STATE();
    case 94:
      ADVANCE_MAP(
        ',', 174,
        '.', 196,
        ':', 101,
        ';', 220,
        '[', 115,
        '\\', 74,
        ']', 211,
        'a', 345,
        '|', 181,
        '}', 222,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(94);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 95:
      if (lookahead == ',') ADVANCE(174);
      if (lookahead == '.') ADVANCE(196);
      if (lookahead == ':') ADVANCE(101);
      if (lookahead == ';') ADVANCE(220);
      if (lookahead == '[') ADVANCE(113);
      if (lookahead == '\\') ADVANCE(74);
      if (lookahead == '|') ADVANCE(182);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(95);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 96:
      if (lookahead == '-') ADVANCE(237);
      END_STATE();
    case 97:
      if (lookahead == '.') ADVANCE(172);
      END_STATE();
    case 98:
      if (lookahead == '.') ADVANCE(359);
      if (lookahead == '[') ADVANCE(113);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(253);
      END_STATE();
    case 99:
      if (lookahead == '.') ADVANCE(93);
      if (lookahead == 'L' ||
          lookahead == 'l' ||
          lookahead == 'n') ADVANCE(301);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(99);
      if (set_contains(sym_conversion_specification_character_set_7, 16, lookahead)) ADVANCE(300);
      END_STATE();
    case 100:
      if (lookahead == '.') ADVANCE(93);
      if (lookahead == 'L' ||
          lookahead == 'l' ||
          lookahead == 'n') ADVANCE(301);
      if (set_contains(sym_conversion_specification_character_set_6, 15, lookahead)) ADVANCE(300);
      END_STATE();
    case 101:
      if (lookahead == ':') ADVANCE(234);
      END_STATE();
    case 102:
      if (lookahead == ':') ADVANCE(234);
      if (lookahead == '=') ADVANCE(171);
      END_STATE();
    case 103:
      if (lookahead == ':') ADVANCE(234);
      if (lookahead == '=') ADVANCE(330);
      END_STATE();
    case 104:
      if (lookahead == ';') ADVANCE(153);
      END_STATE();
    case 105:
      if (lookahead == '=') ADVANCE(173);
      END_STATE();
    case 106:
      if (lookahead == '=') ADVANCE(171);
      END_STATE();
    case 107:
      if (lookahead == '=') ADVANCE(171);
      if (lookahead == '>') ADVANCE(223);
      END_STATE();
    case 108:
      if (lookahead == '=') ADVANCE(330);
      END_STATE();
    case 109:
      if (lookahead == '=') ADVANCE(319);
      END_STATE();
    case 110:
      if (lookahead == '>') ADVANCE(198);
      END_STATE();
    case 111:
      if (lookahead == '>') ADVANCE(223);
      END_STATE();
    case 112:
      if (lookahead == '>') ADVANCE(302);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(112);
      END_STATE();
    case 113:
      if (lookahead == '@') ADVANCE(242);
      END_STATE();
    case 114:
      if (lookahead == '@') ADVANCE(248);
      END_STATE();
    case 115:
      if (lookahead == '@') ADVANCE(244);
      END_STATE();
    case 116:
      if (lookahead == '[') ADVANCE(113);
      if (lookahead == '\\') ADVANCE(120);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(116);
      END_STATE();
    case 117:
      if (lookahead == ']') ADVANCE(236);
      END_STATE();
    case 118:
      if (lookahead == ']') ADVANCE(236);
      if (lookahead == '|') ADVANCE(329);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(324);
      END_STATE();
    case 119:
      if (lookahead == 'n') ADVANCE(302);
      END_STATE();
    case 120:
      ADVANCE_MAP(
        'o', 129,
        'x', 143,
        ' ', 296,
        '"', 296,
        '\'', 296,
        '\\', 296,
        'b', 296,
        'n', 296,
        'r', 296,
        't', 296,
      );
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(136);
      END_STATE();
    case 121:
      if (lookahead == '{') ADVANCE(141);
      END_STATE();
    case 122:
      if (lookahead == '|') ADVANCE(329);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(324);
      END_STATE();
    case 123:
      if (lookahead == '}') ADVANCE(289);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(123);
      END_STATE();
    case 124:
      if (lookahead == '}') ADVANCE(240);
      END_STATE();
    case 125:
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(133);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      END_STATE();
    case 126:
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(265);
      END_STATE();
    case 127:
      if (lookahead == 'L' ||
          lookahead == 'l' ||
          lookahead == 'n') ADVANCE(301);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(127);
      if (set_contains(sym_conversion_specification_character_set_7, 16, lookahead)) ADVANCE(300);
      END_STATE();
    case 128:
      if (lookahead == 'L' ||
          lookahead == 'l' ||
          lookahead == 'n') ADVANCE(301);
      if (set_contains(sym_conversion_specification_character_set_6, 15, lookahead)) ADVANCE(300);
      END_STATE();
    case 129:
      if (('0' <= lookahead && lookahead <= '3')) ADVANCE(131);
      END_STATE();
    case 130:
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(299);
      END_STATE();
    case 131:
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(130);
      END_STATE();
    case 132:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(241);
      END_STATE();
    case 133:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      END_STATE();
    case 134:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(112);
      END_STATE();
    case 135:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(297);
      END_STATE();
    case 136:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(135);
      END_STATE();
    case 137:
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(324);
      END_STATE();
    case 138:
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(315);
      END_STATE();
    case 139:
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(310);
      END_STATE();
    case 140:
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(306);
      END_STATE();
    case 141:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(123);
      END_STATE();
    case 142:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(298);
      END_STATE();
    case 143:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(142);
      END_STATE();
    case 144:
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 145:
      if (eof) ADVANCE(152);
      ADVANCE_MAP(
        '!', 180,
        '"', 282,
        '#', 233,
        '$', 132,
        '%', 256,
        '&', 228,
        '\'', 277,
        '(', 169,
        ')', 170,
        '*', 213,
        '+', 176,
        ',', 174,
        '-', 178,
        '.', 197,
        '0', 259,
        ':', 189,
        ';', 221,
        '<', 231,
        '=', 158,
        '>', 226,
        '?', 161,
        '@', 316,
        '[', 201,
        '\\', 2,
        ']', 211,
        '^', 318,
        '`', 358,
        'a', 344,
        'l', 340,
        'm', 335,
        '{', 216,
        '|', 183,
        '}', 222,
        '~', 159,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(145);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 146:
      if (eof) ADVANCE(152);
      ADVANCE_MAP(
        '!', 180,
        '"', 282,
        '#', 232,
        '%', 256,
        '&', 227,
        '\'', 357,
        '(', 169,
        ')', 170,
        '*', 212,
        '+', 105,
        ',', 174,
        '-', 110,
        '.', 97,
        '0', 259,
        ':', 189,
        ';', 104,
        '<', 230,
        '=', 158,
        '>', 226,
        '?', 161,
        '[', 202,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 341,
        '{', 215,
        '|', 181,
        '}', 222,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(146);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 147:
      if (eof) ADVANCE(152);
      ADVANCE_MAP(
        '!', 180,
        '#', 356,
        '&', 227,
        '(', 169,
        ')', 170,
        '*', 212,
        '+', 105,
        ',', 174,
        '-', 110,
        '.', 196,
        ':', 189,
        ';', 104,
        '=', 158,
        '>', 226,
        '[', 203,
        '\\', 74,
        ']', 211,
        'a', 345,
        'l', 340,
        '{', 79,
        '|', 181,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(147);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 148:
      if (eof) ADVANCE(152);
      ADVANCE_MAP(
        '!', 109,
        '"', 282,
        '#', 233,
        '&', 229,
        '(', 169,
        ')', 170,
        '*', 312,
        '+', 273,
        ',', 174,
        '-', 274,
        '.', 196,
        ':', 191,
        ';', 220,
        '<', 320,
        '>', 322,
        '[', 204,
        '\\', 74,
        ']', 211,
        '{', 215,
        '|', 184,
        '}', 222,
        '$', 324,
        '=', 324,
        '%', 313,
        '/', 313,
        '@', 318,
        '^', 318,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(148);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 149:
      if (eof) ADVANCE(152);
      ADVANCE_MAP(
        '"', 282,
        '#', 233,
        '&', 227,
        '\'', 277,
        '(', 169,
        ')', 170,
        '*', 212,
        '+', 270,
        ',', 174,
        '-', 276,
        '.', 196,
        '0', 259,
        ':', 189,
        ';', 220,
        '=', 158,
        '>', 226,
        '?', 161,
        '[', 204,
        '\\', 74,
        ']', 211,
        '`', 358,
        'a', 345,
        'l', 341,
        '{', 215,
        '|', 181,
        '}', 222,
        '~', 159,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(149);
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(260);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 150:
      if (eof) ADVANCE(152);
      ADVANCE_MAP(
        '#', 232,
        '%', 256,
        '&', 227,
        '\'', 357,
        ')', 170,
        '*', 212,
        ',', 174,
        '-', 110,
        '.', 196,
        ':', 107,
        ';', 104,
        '=', 158,
        '>', 226,
        '[', 203,
        '\\', 74,
        ']', 211,
        'a', 345,
        'l', 341,
        '{', 79,
        '|', 181,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(150);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 151:
      if (eof) ADVANCE(152);
      ADVANCE_MAP(
        '&', 227,
        '(', 169,
        ')', 170,
        '*', 212,
        '+', 105,
        ',', 174,
        '-', 110,
        '.', 196,
        ':', 194,
        ';', 104,
        '=', 158,
        '>', 226,
        '[', 203,
        '\\', 74,
        ']', 211,
        'a', 345,
        '{', 79,
        '|', 181,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(151);
      if (set_contains(sym__uppercase_identifier_character_set_1, 651, lookahead)) ADVANCE(355);
      if (set_contains(sym__lowercase_identifier_character_set_3, 663, lookahead)) ADVANCE(354);
      END_STATE();
    case 152:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 153:
      ACCEPT_TOKEN(anon_sym_SEMI_SEMI);
      END_STATE();
    case 154:
      ACCEPT_TOKEN(anon_sym_let);
      if (set_contains(sym_let_operator_character_set_1, 9, lookahead)) ADVANCE(332);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 155:
      ACCEPT_TOKEN(anon_sym_let);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 156:
      ACCEPT_TOKEN(anon_sym_and);
      if (set_contains(sym_let_operator_character_set_1, 9, lookahead)) ADVANCE(333);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 157:
      ACCEPT_TOKEN(anon_sym_and);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 158:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 159:
      ACCEPT_TOKEN(anon_sym_TILDE);
      END_STATE();
    case 160:
      ACCEPT_TOKEN(anon_sym_TILDE);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(306);
      END_STATE();
    case 161:
      ACCEPT_TOKEN(anon_sym_QMARK);
      END_STATE();
    case 162:
      ACCEPT_TOKEN(anon_sym_QMARK);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(306);
      END_STATE();
    case 163:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 164:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(234);
      END_STATE();
    case 165:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(234);
      if (lookahead == '=') ADVANCE(171);
      if (lookahead == '>') ADVANCE(223);
      END_STATE();
    case 166:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(234);
      if (lookahead == '=') ADVANCE(330);
      END_STATE();
    case 167:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(234);
      if (lookahead == '=') ADVANCE(330);
      if (lookahead == '>') ADVANCE(223);
      END_STATE();
    case 168:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(234);
      if (lookahead == '>') ADVANCE(223);
      END_STATE();
    case 169:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 170:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 171:
      ACCEPT_TOKEN(anon_sym_COLON_EQ);
      END_STATE();
    case 172:
      ACCEPT_TOKEN(anon_sym_DOT_DOT);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(anon_sym_PLUS_EQ);
      END_STATE();
    case 174:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 175:
      ACCEPT_TOKEN(anon_sym_PLUS);
      END_STATE();
    case 176:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '.') ADVANCE(307);
      if (lookahead == '=') ADVANCE(173);
      END_STATE();
    case 177:
      ACCEPT_TOKEN(anon_sym_DASH);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '.') ADVANCE(307);
      if (lookahead == '>') ADVANCE(198);
      END_STATE();
    case 179:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '>') ADVANCE(198);
      END_STATE();
    case 180:
      ACCEPT_TOKEN(anon_sym_BANG);
      END_STATE();
    case 181:
      ACCEPT_TOKEN(anon_sym_PIPE);
      END_STATE();
    case 182:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == ']') ADVANCE(236);
      END_STATE();
    case 183:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == ']') ADVANCE(236);
      if (lookahead == '|') ADVANCE(328);
      END_STATE();
    case 184:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == ']') ADVANCE(236);
      if (lookahead == '|') ADVANCE(329);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(324);
      END_STATE();
    case 185:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '|') ADVANCE(329);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(324);
      END_STATE();
    case 186:
      ACCEPT_TOKEN(anon_sym_COLON2);
      END_STATE();
    case 187:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == ':') ADVANCE(234);
      END_STATE();
    case 188:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == ':') ADVANCE(234);
      if (lookahead == '=') ADVANCE(171);
      END_STATE();
    case 189:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == ':') ADVANCE(234);
      if (lookahead == '=') ADVANCE(171);
      if (lookahead == '>') ADVANCE(223);
      END_STATE();
    case 190:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == ':') ADVANCE(234);
      if (lookahead == '=') ADVANCE(330);
      END_STATE();
    case 191:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == ':') ADVANCE(234);
      if (lookahead == '=') ADVANCE(330);
      if (lookahead == '>') ADVANCE(223);
      END_STATE();
    case 192:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == ':') ADVANCE(234);
      if (lookahead == '>') ADVANCE(223);
      END_STATE();
    case 193:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == '=') ADVANCE(171);
      END_STATE();
    case 194:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == '=') ADVANCE(171);
      if (lookahead == '>') ADVANCE(223);
      END_STATE();
    case 195:
      ACCEPT_TOKEN(anon_sym_COLON2);
      if (lookahead == '>') ADVANCE(223);
      END_STATE();
    case 196:
      ACCEPT_TOKEN(anon_sym_DOT);
      END_STATE();
    case 197:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(172);
      END_STATE();
    case 198:
      ACCEPT_TOKEN(anon_sym_DASH_GT);
      END_STATE();
    case 199:
      ACCEPT_TOKEN(anon_sym_DASH_GT);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(315);
      END_STATE();
    case 200:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(250);
      if (lookahead == '<') ADVANCE(225);
      if (lookahead == '>') ADVANCE(224);
      if (lookahead == '@') ADVANCE(284);
      if (lookahead == '|') ADVANCE(235);
      END_STATE();
    case 201:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(250);
      if (lookahead == '<') ADVANCE(225);
      if (lookahead == '>') ADVANCE(224);
      if (lookahead == '@') ADVANCE(242);
      if (lookahead == '|') ADVANCE(235);
      END_STATE();
    case 202:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(250);
      if (lookahead == '<') ADVANCE(225);
      if (lookahead == '>') ADVANCE(224);
      if (lookahead == '@') ADVANCE(243);
      END_STATE();
    case 203:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(250);
      if (lookahead == '@') ADVANCE(243);
      END_STATE();
    case 204:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(250);
      if (lookahead == '@') ADVANCE(243);
      if (lookahead == '|') ADVANCE(235);
      END_STATE();
    case 205:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(250);
      if (lookahead == '@') ADVANCE(245);
      END_STATE();
    case 206:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(250);
      if (lookahead == '@') ADVANCE(245);
      if (lookahead == '|') ADVANCE(235);
      END_STATE();
    case 207:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(249);
      if (lookahead == '@') ADVANCE(242);
      if (lookahead == '|') ADVANCE(235);
      END_STATE();
    case 208:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(249);
      if (lookahead == '@') ADVANCE(244);
      if (lookahead == '|') ADVANCE(235);
      END_STATE();
    case 209:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '%') ADVANCE(80);
      if (lookahead == '@') ADVANCE(243);
      END_STATE();
    case 210:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      if (lookahead == '@') ADVANCE(242);
      END_STATE();
    case 211:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 212:
      ACCEPT_TOKEN(anon_sym_STAR);
      END_STATE();
    case 213:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '*') ADVANCE(311);
      END_STATE();
    case 214:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 215:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      if (lookahead == '%') ADVANCE(252);
      END_STATE();
    case 216:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      if (lookahead == '%') ADVANCE(252);
      if (lookahead == '<') ADVANCE(239);
      END_STATE();
    case 217:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      if (lookahead == '%') ADVANCE(251);
      END_STATE();
    case 218:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      if (lookahead == '%') ADVANCE(251);
      if (lookahead == '<') ADVANCE(239);
      END_STATE();
    case 219:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      if (lookahead == '%') ADVANCE(81);
      END_STATE();
    case 220:
      ACCEPT_TOKEN(anon_sym_SEMI);
      END_STATE();
    case 221:
      ACCEPT_TOKEN(anon_sym_SEMI);
      if (lookahead == ';') ADVANCE(153);
      END_STATE();
    case 222:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 223:
      ACCEPT_TOKEN(anon_sym_COLON_GT);
      END_STATE();
    case 224:
      ACCEPT_TOKEN(anon_sym_LBRACK_GT);
      END_STATE();
    case 225:
      ACCEPT_TOKEN(anon_sym_LBRACK_LT);
      END_STATE();
    case 226:
      ACCEPT_TOKEN(anon_sym_GT);
      END_STATE();
    case 227:
      ACCEPT_TOKEN(anon_sym_AMP);
      END_STATE();
    case 228:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(326);
      END_STATE();
    case 229:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(327);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(324);
      END_STATE();
    case 230:
      ACCEPT_TOKEN(anon_sym_LT);
      END_STATE();
    case 231:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '-') ADVANCE(237);
      END_STATE();
    case 232:
      ACCEPT_TOKEN(anon_sym_POUND);
      END_STATE();
    case 233:
      ACCEPT_TOKEN(anon_sym_POUND);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(310);
      END_STATE();
    case 234:
      ACCEPT_TOKEN(anon_sym_COLON_COLON);
      END_STATE();
    case 235:
      ACCEPT_TOKEN(anon_sym_LBRACK_PIPE);
      END_STATE();
    case 236:
      ACCEPT_TOKEN(anon_sym_PIPE_RBRACK);
      END_STATE();
    case 237:
      ACCEPT_TOKEN(anon_sym_LT_DASH);
      END_STATE();
    case 238:
      ACCEPT_TOKEN(anon_sym_LT_DASH);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(324);
      END_STATE();
    case 239:
      ACCEPT_TOKEN(anon_sym_LBRACE_LT);
      END_STATE();
    case 240:
      ACCEPT_TOKEN(anon_sym_GT_RBRACE);
      END_STATE();
    case 241:
      ACCEPT_TOKEN(sym_ocamlyacc_value);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(241);
      END_STATE();
    case 242:
      ACCEPT_TOKEN(aux_sym_attribute_token1);
      END_STATE();
    case 243:
      ACCEPT_TOKEN(aux_sym_attribute_token1);
      if (lookahead == '@') ADVANCE(247);
      END_STATE();
    case 244:
      ACCEPT_TOKEN(aux_sym_attribute_token1);
      if (lookahead == '@') ADVANCE(246);
      END_STATE();
    case 245:
      ACCEPT_TOKEN(aux_sym_attribute_token1);
      if (lookahead == '@') ADVANCE(114);
      END_STATE();
    case 246:
      ACCEPT_TOKEN(anon_sym_LBRACK_AT_AT);
      END_STATE();
    case 247:
      ACCEPT_TOKEN(anon_sym_LBRACK_AT_AT);
      if (lookahead == '@') ADVANCE(248);
      END_STATE();
    case 248:
      ACCEPT_TOKEN(anon_sym_LBRACK_AT_AT_AT);
      END_STATE();
    case 249:
      ACCEPT_TOKEN(anon_sym_LBRACK_PERCENT);
      END_STATE();
    case 250:
      ACCEPT_TOKEN(anon_sym_LBRACK_PERCENT);
      if (lookahead == '%') ADVANCE(254);
      END_STATE();
    case 251:
      ACCEPT_TOKEN(anon_sym_LBRACE_PERCENT);
      END_STATE();
    case 252:
      ACCEPT_TOKEN(anon_sym_LBRACE_PERCENT);
      if (lookahead == '%') ADVANCE(255);
      END_STATE();
    case 253:
      ACCEPT_TOKEN(aux_sym_quoted_extension_token1);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(253);
      END_STATE();
    case 254:
      ACCEPT_TOKEN(anon_sym_LBRACK_PERCENT_PERCENT);
      END_STATE();
    case 255:
      ACCEPT_TOKEN(anon_sym_LBRACE_PERCENT_PERCENT);
      END_STATE();
    case 256:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      END_STATE();
    case 257:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(313);
      END_STATE();
    case 258:
      ACCEPT_TOKEN(aux_sym_number_token1);
      END_STATE();
    case 259:
      ACCEPT_TOKEN(aux_sym_number_token1);
      ADVANCE_MAP(
        '.', 263,
        'B', 126,
        'b', 126,
        'E', 125,
        'e', 125,
        'O', 266,
        'o', 266,
        'X', 269,
        'x', 269,
      );
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(260);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(258);
      END_STATE();
    case 260:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '.') ADVANCE(263);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(125);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(260);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(258);
      END_STATE();
    case 261:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '.') ADVANCE(264);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(262);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(261);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(258);
      END_STATE();
    case 262:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(133);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(268);
      END_STATE();
    case 263:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(125);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(263);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(258);
      END_STATE();
    case 264:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == 'P' ||
          lookahead == 'p') ADVANCE(262);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(264);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(258);
      END_STATE();
    case 265:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (lookahead == '0' ||
          lookahead == '1' ||
          lookahead == '_') ADVANCE(265);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(258);
      END_STATE();
    case 266:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(267);
      END_STATE();
    case 267:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (('0' <= lookahead && lookahead <= '7') ||
          lookahead == '_') ADVANCE(267);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(258);
      END_STATE();
    case 268:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(268);
      if (('G' <= lookahead && lookahead <= 'Z') ||
          ('g' <= lookahead && lookahead <= 'z')) ADVANCE(258);
      END_STATE();
    case 269:
      ACCEPT_TOKEN(aux_sym_number_token1);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(261);
      END_STATE();
    case 270:
      ACCEPT_TOKEN(aux_sym_signed_number_token1);
      END_STATE();
    case 271:
      ACCEPT_TOKEN(aux_sym_signed_number_token1);
      if (lookahead == '.') ADVANCE(307);
      END_STATE();
    case 272:
      ACCEPT_TOKEN(aux_sym_signed_number_token1);
      if (lookahead == '.') ADVANCE(307);
      if (lookahead == '>') ADVANCE(198);
      END_STATE();
    case 273:
      ACCEPT_TOKEN(aux_sym_signed_number_token1);
      if (lookahead == '.') ADVANCE(309);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(315);
      END_STATE();
    case 274:
      ACCEPT_TOKEN(aux_sym_signed_number_token1);
      if (lookahead == '.') ADVANCE(308);
      if (lookahead == '>') ADVANCE(199);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(314);
      END_STATE();
    case 275:
      ACCEPT_TOKEN(aux_sym_signed_number_token1);
      if (lookahead == '.') ADVANCE(308);
      if (lookahead == '>') ADVANCE(138);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(314);
      END_STATE();
    case 276:
      ACCEPT_TOKEN(aux_sym_signed_number_token1);
      if (lookahead == '>') ADVANCE(198);
      END_STATE();
    case 277:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      END_STATE();
    case 278:
      ACCEPT_TOKEN(anon_sym_SQUOTE2);
      END_STATE();
    case 279:
      ACCEPT_TOKEN(aux_sym_character_content_token1);
      END_STATE();
    case 280:
      ACCEPT_TOKEN(aux_sym_character_content_token2);
      END_STATE();
    case 281:
      ACCEPT_TOKEN(aux_sym_character_content_token2);
      if (lookahead == '@') ADVANCE(242);
      END_STATE();
    case 282:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 283:
      ACCEPT_TOKEN(aux_sym_string_content_token1);
      END_STATE();
    case 284:
      ACCEPT_TOKEN(aux_sym_string_content_token2);
      END_STATE();
    case 285:
      ACCEPT_TOKEN(aux_sym_string_content_token3);
      ADVANCE_MAP(
        '*', 100,
        '.', 93,
        'L', 301,
        'l', 301,
        'n', 301,
        ' ', 92,
        '#', 92,
        '+', 92,
        '-', 92,
        '0', 92,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(99);
      if (set_contains(sym_conversion_specification_character_set_7, 16, lookahead)) ADVANCE(300);
      END_STATE();
    case 286:
      ACCEPT_TOKEN(aux_sym_string_content_token3);
      ADVANCE_MAP(
        '<', 134,
        '\\', 119,
        ' ', 302,
        ',', 302,
        '.', 302,
        ';', 302,
        '?', 302,
        '[', 302,
        ']', 302,
        '{', 302,
        '}', 302,
      );
      END_STATE();
    case 287:
      ACCEPT_TOKEN(aux_sym_string_content_token3);
      if (lookahead == '@') ADVANCE(284);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '%' &&
          lookahead != '\\') ADVANCE(288);
      END_STATE();
    case 288:
      ACCEPT_TOKEN(aux_sym_string_content_token3);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '%' &&
          lookahead != '@' &&
          lookahead != '\\') ADVANCE(288);
      END_STATE();
    case 289:
      ACCEPT_TOKEN(aux_sym_string_content_token4);
      END_STATE();
    case 290:
      ACCEPT_TOKEN(aux_sym_string_content_token5);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(290);
      END_STATE();
    case 291:
      ACCEPT_TOKEN(aux_sym_quoted_string_content_token1);
      END_STATE();
    case 292:
      ACCEPT_TOKEN(aux_sym_quoted_string_content_token1);
      ADVANCE_MAP(
        '*', 100,
        '.', 93,
        'L', 301,
        'l', 301,
        'n', 301,
        ' ', 92,
        '#', 92,
        '+', 92,
        '-', 92,
        '0', 92,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(99);
      if (set_contains(sym_conversion_specification_character_set_7, 16, lookahead)) ADVANCE(300);
      END_STATE();
    case 293:
      ACCEPT_TOKEN(aux_sym_quoted_string_content_token1);
      ADVANCE_MAP(
        '<', 134,
        '\\', 119,
        ' ', 302,
        ',', 302,
        '.', 302,
        ';', 302,
        '?', 302,
        '[', 302,
        ']', 302,
        '{', 302,
        '}', 302,
      );
      END_STATE();
    case 294:
      ACCEPT_TOKEN(aux_sym_quoted_string_content_token1);
      if (lookahead == '@') ADVANCE(284);
      if (lookahead != 0 &&
          lookahead != '%' &&
          lookahead != '|') ADVANCE(295);
      END_STATE();
    case 295:
      ACCEPT_TOKEN(aux_sym_quoted_string_content_token1);
      if (lookahead != 0 &&
          lookahead != '%' &&
          lookahead != '@' &&
          lookahead != '|') ADVANCE(295);
      END_STATE();
    case 296:
      ACCEPT_TOKEN(aux_sym_escape_sequence_token1);
      END_STATE();
    case 297:
      ACCEPT_TOKEN(aux_sym_escape_sequence_token2);
      END_STATE();
    case 298:
      ACCEPT_TOKEN(aux_sym_escape_sequence_token3);
      END_STATE();
    case 299:
      ACCEPT_TOKEN(aux_sym_escape_sequence_token4);
      END_STATE();
    case 300:
      ACCEPT_TOKEN(sym_conversion_specification);
      END_STATE();
    case 301:
      ACCEPT_TOKEN(sym_conversion_specification);
      if (lookahead == 'X' ||
          lookahead == 'd' ||
          lookahead == 'i' ||
          lookahead == 'o' ||
          lookahead == 'u' ||
          lookahead == 'x') ADVANCE(300);
      END_STATE();
    case 302:
      ACCEPT_TOKEN(sym_pretty_printing_indication);
      END_STATE();
    case 303:
      ACCEPT_TOKEN(sym_pretty_printing_indication);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(318);
      END_STATE();
    case 304:
      ACCEPT_TOKEN(sym_prefix_operator);
      if (lookahead == '=') ADVANCE(140);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(306);
      END_STATE();
    case 305:
      ACCEPT_TOKEN(sym_prefix_operator);
      if (lookahead == '=') ADVANCE(325);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(306);
      END_STATE();
    case 306:
      ACCEPT_TOKEN(sym_prefix_operator);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(306);
      END_STATE();
    case 307:
      ACCEPT_TOKEN(aux_sym_sign_operator_token1);
      END_STATE();
    case 308:
      ACCEPT_TOKEN(aux_sym_sign_operator_token1);
      if (lookahead == '>') ADVANCE(315);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(314);
      END_STATE();
    case 309:
      ACCEPT_TOKEN(aux_sym_sign_operator_token1);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(315);
      END_STATE();
    case 310:
      ACCEPT_TOKEN(sym_hash_operator);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(310);
      END_STATE();
    case 311:
      ACCEPT_TOKEN(aux_sym_pow_operator_token1);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(311);
      END_STATE();
    case 312:
      ACCEPT_TOKEN(aux_sym_mult_operator_token1);
      if (lookahead == '*') ADVANCE(311);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(313);
      END_STATE();
    case 313:
      ACCEPT_TOKEN(aux_sym_mult_operator_token1);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(313);
      END_STATE();
    case 314:
      ACCEPT_TOKEN(aux_sym_add_operator_token1);
      if (lookahead == '>') ADVANCE(315);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(314);
      END_STATE();
    case 315:
      ACCEPT_TOKEN(aux_sym_add_operator_token1);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(315);
      END_STATE();
    case 316:
      ACCEPT_TOKEN(sym_concat_operator);
      ADVANCE_MAP(
        '<', 317,
        '\\', 119,
        '.', 303,
        '?', 303,
        ' ', 302,
        ',', 302,
        ';', 302,
        '[', 302,
        ']', 302,
        '{', 302,
        '}', 302,
      );
      if (lookahead == '!' ||
          ('$' <= lookahead && lookahead <= '&') ||
          ('*' <= lookahead && lookahead <= '/') ||
          (':' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          ('|' <= lookahead && lookahead <= '~')) ADVANCE(318);
      END_STATE();
    case 317:
      ACCEPT_TOKEN(sym_concat_operator);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(112);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(318);
      END_STATE();
    case 318:
      ACCEPT_TOKEN(sym_concat_operator);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(318);
      END_STATE();
    case 319:
      ACCEPT_TOKEN(sym_rel_operator);
      END_STATE();
    case 320:
      ACCEPT_TOKEN(sym_rel_operator);
      if (lookahead == '-') ADVANCE(137);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(324);
      END_STATE();
    case 321:
      ACCEPT_TOKEN(sym_rel_operator);
      if (lookahead == '-') ADVANCE(238);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(324);
      END_STATE();
    case 322:
      ACCEPT_TOKEN(sym_rel_operator);
      if (lookahead == '}') ADVANCE(240);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(324);
      END_STATE();
    case 323:
      ACCEPT_TOKEN(sym_rel_operator);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(241);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(324);
      END_STATE();
    case 324:
      ACCEPT_TOKEN(sym_rel_operator);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(324);
      END_STATE();
    case 325:
      ACCEPT_TOKEN(sym_rel_operator);
      if (set_contains(sym_prefix_operator_character_set_1, 9, lookahead)) ADVANCE(306);
      END_STATE();
    case 326:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      END_STATE();
    case 327:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(324);
      END_STATE();
    case 328:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      END_STATE();
    case 329:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(324);
      END_STATE();
    case 330:
      ACCEPT_TOKEN(sym_assign_operator);
      END_STATE();
    case 331:
      ACCEPT_TOKEN(sym_indexing_operator);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(331);
      END_STATE();
    case 332:
      ACCEPT_TOKEN(sym_let_operator);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(332);
      END_STATE();
    case 333:
      ACCEPT_TOKEN(sym_let_and_operator);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(333);
      END_STATE();
    case 334:
      ACCEPT_TOKEN(sym_match_operator);
      if (set_contains(aux_sym_pow_operator_token1_character_set_1, 9, lookahead)) ADVANCE(334);
      END_STATE();
    case 335:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'a') ADVANCE(348);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 336:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'c') ADVANCE(343);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 337:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'd') ADVANCE(156);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 338:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'd') ADVANCE(157);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 339:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'd') ADVANCE(351);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 340:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'e') ADVANCE(347);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 341:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'e') ADVANCE(349);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 342:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'e') ADVANCE(350);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 343:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'h') ADVANCE(353);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 344:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'n') ADVANCE(337);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 345:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'n') ADVANCE(338);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 346:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 'n') ADVANCE(339);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 347:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 't') ADVANCE(154);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 348:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 't') ADVANCE(336);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 349:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 't') ADVANCE(155);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 350:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (lookahead == 't') ADVANCE(352);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 351:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (set_contains(sym_let_operator_character_set_1, 9, lookahead)) ADVANCE(333);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 352:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (set_contains(sym_let_operator_character_set_1, 9, lookahead)) ADVANCE(332);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 353:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (set_contains(sym_let_operator_character_set_1, 9, lookahead)) ADVANCE(334);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 354:
      ACCEPT_TOKEN(sym__lowercase_identifier);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(354);
      END_STATE();
    case 355:
      ACCEPT_TOKEN(sym__uppercase_identifier);
      if (set_contains(sym__lowercase_identifier_character_set_2, 801, lookahead)) ADVANCE(355);
      END_STATE();
    case 356:
      ACCEPT_TOKEN(aux_sym_directive_token1);
      END_STATE();
    case 357:
      ACCEPT_TOKEN(aux_sym_type_variable_token1);
      END_STATE();
    case 358:
      ACCEPT_TOKEN(aux_sym_tag_token1);
      END_STATE();
    case 359:
      ACCEPT_TOKEN(aux_sym_attribute_id_token1);
      END_STATE();
    default:
      return false;
  }
}

static bool ts_lex_keywords(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      ADVANCE_MAP(
        '_', 1,
        'a', 2,
        'b', 3,
        'c', 4,
        'd', 5,
        'e', 6,
        'f', 7,
        'i', 8,
        'l', 9,
        'm', 10,
        'n', 11,
        'o', 12,
        'p', 13,
        'r', 14,
        's', 15,
        't', 16,
        'v', 17,
        'w', 18,
      );
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(0);
      END_STATE();
    case 1:
      ACCEPT_TOKEN(anon_sym__);
      END_STATE();
    case 2:
      if (lookahead == 's') ADVANCE(19);
      END_STATE();
    case 3:
      if (lookahead == 'e') ADVANCE(20);
      END_STATE();
    case 4:
      if (lookahead == 'l') ADVANCE(21);
      if (lookahead == 'o') ADVANCE(22);
      END_STATE();
    case 5:
      if (lookahead == 'o') ADVANCE(23);
      END_STATE();
    case 6:
      if (lookahead == 'f') ADVANCE(24);
      if (lookahead == 'l') ADVANCE(25);
      if (lookahead == 'n') ADVANCE(26);
      if (lookahead == 'x') ADVANCE(27);
      END_STATE();
    case 7:
      if (lookahead == 'a') ADVANCE(28);
      if (lookahead == 'o') ADVANCE(29);
      if (lookahead == 'u') ADVANCE(30);
      END_STATE();
    case 8:
      if (lookahead == 'f') ADVANCE(31);
      if (lookahead == 'n') ADVANCE(32);
      END_STATE();
    case 9:
      if (lookahead == 'a') ADVANCE(33);
      if (lookahead == 'o') ADVANCE(34);
      if (lookahead == 's') ADVANCE(35);
      if (lookahead == 'x') ADVANCE(36);
      END_STATE();
    case 10:
      if (lookahead == 'a') ADVANCE(37);
      if (lookahead == 'e') ADVANCE(38);
      if (lookahead == 'o') ADVANCE(39);
      if (lookahead == 'u') ADVANCE(40);
      END_STATE();
    case 11:
      if (lookahead == 'e') ADVANCE(41);
      if (lookahead == 'o') ADVANCE(42);
      END_STATE();
    case 12:
      if (lookahead == 'b') ADVANCE(43);
      if (lookahead == 'f') ADVANCE(44);
      if (lookahead == 'p') ADVANCE(45);
      if (lookahead == 'r') ADVANCE(46);
      END_STATE();
    case 13:
      if (lookahead == 'r') ADVANCE(47);
      END_STATE();
    case 14:
      if (lookahead == 'e') ADVANCE(48);
      END_STATE();
    case 15:
      if (lookahead == 'i') ADVANCE(49);
      if (lookahead == 't') ADVANCE(50);
      END_STATE();
    case 16:
      if (lookahead == 'h') ADVANCE(51);
      if (lookahead == 'o') ADVANCE(52);
      if (lookahead == 'r') ADVANCE(53);
      if (lookahead == 'y') ADVANCE(54);
      END_STATE();
    case 17:
      if (lookahead == 'a') ADVANCE(55);
      if (lookahead == 'i') ADVANCE(56);
      END_STATE();
    case 18:
      if (lookahead == 'h') ADVANCE(57);
      if (lookahead == 'i') ADVANCE(58);
      END_STATE();
    case 19:
      ACCEPT_TOKEN(anon_sym_as);
      if (lookahead == 'r') ADVANCE(59);
      if (lookahead == 's') ADVANCE(60);
      END_STATE();
    case 20:
      if (lookahead == 'g') ADVANCE(61);
      END_STATE();
    case 21:
      if (lookahead == 'a') ADVANCE(62);
      END_STATE();
    case 22:
      if (lookahead == 'n') ADVANCE(63);
      END_STATE();
    case 23:
      ACCEPT_TOKEN(anon_sym_do);
      if (lookahead == 'n') ADVANCE(64);
      if (lookahead == 'w') ADVANCE(65);
      END_STATE();
    case 24:
      if (lookahead == 'f') ADVANCE(66);
      END_STATE();
    case 25:
      if (lookahead == 's') ADVANCE(67);
      END_STATE();
    case 26:
      if (lookahead == 'd') ADVANCE(68);
      END_STATE();
    case 27:
      if (lookahead == 'c') ADVANCE(69);
      if (lookahead == 't') ADVANCE(70);
      END_STATE();
    case 28:
      if (lookahead == 'l') ADVANCE(71);
      END_STATE();
    case 29:
      if (lookahead == 'r') ADVANCE(72);
      END_STATE();
    case 30:
      if (lookahead == 'n') ADVANCE(73);
      END_STATE();
    case 31:
      ACCEPT_TOKEN(anon_sym_if);
      END_STATE();
    case 32:
      ACCEPT_TOKEN(anon_sym_in);
      if (lookahead == 'c') ADVANCE(74);
      if (lookahead == 'h') ADVANCE(75);
      if (lookahead == 'i') ADVANCE(76);
      END_STATE();
    case 33:
      if (lookahead == 'n') ADVANCE(77);
      if (lookahead == 'z') ADVANCE(78);
      END_STATE();
    case 34:
      if (lookahead == 'r') ADVANCE(79);
      END_STATE();
    case 35:
      if (lookahead == 'l') ADVANCE(80);
      if (lookahead == 'r') ADVANCE(81);
      END_STATE();
    case 36:
      if (lookahead == 'o') ADVANCE(82);
      END_STATE();
    case 37:
      if (lookahead == 't') ADVANCE(83);
      END_STATE();
    case 38:
      if (lookahead == 't') ADVANCE(84);
      END_STATE();
    case 39:
      if (lookahead == 'd') ADVANCE(85);
      END_STATE();
    case 40:
      if (lookahead == 't') ADVANCE(86);
      END_STATE();
    case 41:
      if (lookahead == 'w') ADVANCE(87);
      END_STATE();
    case 42:
      if (lookahead == 'n') ADVANCE(88);
      END_STATE();
    case 43:
      if (lookahead == 'j') ADVANCE(89);
      END_STATE();
    case 44:
      ACCEPT_TOKEN(anon_sym_of);
      END_STATE();
    case 45:
      if (lookahead == 'e') ADVANCE(90);
      END_STATE();
    case 46:
      ACCEPT_TOKEN(anon_sym_or);
      END_STATE();
    case 47:
      if (lookahead == 'i') ADVANCE(91);
      END_STATE();
    case 48:
      if (lookahead == 'c') ADVANCE(92);
      END_STATE();
    case 49:
      if (lookahead == 'g') ADVANCE(93);
      END_STATE();
    case 50:
      if (lookahead == 'r') ADVANCE(94);
      END_STATE();
    case 51:
      if (lookahead == 'e') ADVANCE(95);
      END_STATE();
    case 52:
      ACCEPT_TOKEN(anon_sym_to);
      END_STATE();
    case 53:
      if (lookahead == 'u') ADVANCE(96);
      if (lookahead == 'y') ADVANCE(97);
      END_STATE();
    case 54:
      if (lookahead == 'p') ADVANCE(98);
      END_STATE();
    case 55:
      if (lookahead == 'l') ADVANCE(99);
      END_STATE();
    case 56:
      if (lookahead == 'r') ADVANCE(100);
      END_STATE();
    case 57:
      if (lookahead == 'e') ADVANCE(101);
      if (lookahead == 'i') ADVANCE(102);
      END_STATE();
    case 58:
      if (lookahead == 't') ADVANCE(103);
      END_STATE();
    case 59:
      ACCEPT_TOKEN(anon_sym_asr);
      END_STATE();
    case 60:
      if (lookahead == 'e') ADVANCE(104);
      END_STATE();
    case 61:
      if (lookahead == 'i') ADVANCE(105);
      END_STATE();
    case 62:
      if (lookahead == 's') ADVANCE(106);
      END_STATE();
    case 63:
      if (lookahead == 's') ADVANCE(107);
      END_STATE();
    case 64:
      if (lookahead == 'e') ADVANCE(108);
      END_STATE();
    case 65:
      if (lookahead == 'n') ADVANCE(109);
      END_STATE();
    case 66:
      if (lookahead == 'e') ADVANCE(110);
      END_STATE();
    case 67:
      if (lookahead == 'e') ADVANCE(111);
      END_STATE();
    case 68:
      ACCEPT_TOKEN(anon_sym_end);
      END_STATE();
    case 69:
      if (lookahead == 'e') ADVANCE(112);
      END_STATE();
    case 70:
      if (lookahead == 'e') ADVANCE(113);
      END_STATE();
    case 71:
      if (lookahead == 's') ADVANCE(114);
      END_STATE();
    case 72:
      ACCEPT_TOKEN(anon_sym_for);
      END_STATE();
    case 73:
      ACCEPT_TOKEN(anon_sym_fun);
      if (lookahead == 'c') ADVANCE(115);
      END_STATE();
    case 74:
      if (lookahead == 'l') ADVANCE(116);
      END_STATE();
    case 75:
      if (lookahead == 'e') ADVANCE(117);
      END_STATE();
    case 76:
      if (lookahead == 't') ADVANCE(118);
      END_STATE();
    case 77:
      if (lookahead == 'd') ADVANCE(119);
      END_STATE();
    case 78:
      if (lookahead == 'y') ADVANCE(120);
      END_STATE();
    case 79:
      ACCEPT_TOKEN(anon_sym_lor);
      END_STATE();
    case 80:
      ACCEPT_TOKEN(anon_sym_lsl);
      END_STATE();
    case 81:
      ACCEPT_TOKEN(anon_sym_lsr);
      END_STATE();
    case 82:
      if (lookahead == 'r') ADVANCE(121);
      END_STATE();
    case 83:
      if (lookahead == 'c') ADVANCE(122);
      END_STATE();
    case 84:
      if (lookahead == 'h') ADVANCE(123);
      END_STATE();
    case 85:
      ACCEPT_TOKEN(anon_sym_mod);
      if (lookahead == 'u') ADVANCE(124);
      END_STATE();
    case 86:
      if (lookahead == 'a') ADVANCE(125);
      END_STATE();
    case 87:
      ACCEPT_TOKEN(anon_sym_new);
      END_STATE();
    case 88:
      if (lookahead == 'r') ADVANCE(126);
      END_STATE();
    case 89:
      if (lookahead == 'e') ADVANCE(127);
      END_STATE();
    case 90:
      if (lookahead == 'n') ADVANCE(128);
      END_STATE();
    case 91:
      if (lookahead == 'v') ADVANCE(129);
      END_STATE();
    case 92:
      ACCEPT_TOKEN(anon_sym_rec);
      END_STATE();
    case 93:
      ACCEPT_TOKEN(anon_sym_sig);
      END_STATE();
    case 94:
      if (lookahead == 'u') ADVANCE(130);
      END_STATE();
    case 95:
      if (lookahead == 'n') ADVANCE(131);
      END_STATE();
    case 96:
      if (lookahead == 'e') ADVANCE(132);
      END_STATE();
    case 97:
      ACCEPT_TOKEN(anon_sym_try);
      END_STATE();
    case 98:
      if (lookahead == 'e') ADVANCE(133);
      END_STATE();
    case 99:
      ACCEPT_TOKEN(anon_sym_val);
      END_STATE();
    case 100:
      if (lookahead == 't') ADVANCE(134);
      END_STATE();
    case 101:
      if (lookahead == 'n') ADVANCE(135);
      END_STATE();
    case 102:
      if (lookahead == 'l') ADVANCE(136);
      END_STATE();
    case 103:
      if (lookahead == 'h') ADVANCE(137);
      END_STATE();
    case 104:
      if (lookahead == 'r') ADVANCE(138);
      END_STATE();
    case 105:
      if (lookahead == 'n') ADVANCE(139);
      END_STATE();
    case 106:
      if (lookahead == 's') ADVANCE(140);
      END_STATE();
    case 107:
      if (lookahead == 't') ADVANCE(141);
      END_STATE();
    case 108:
      ACCEPT_TOKEN(anon_sym_done);
      END_STATE();
    case 109:
      if (lookahead == 't') ADVANCE(142);
      END_STATE();
    case 110:
      if (lookahead == 'c') ADVANCE(143);
      END_STATE();
    case 111:
      ACCEPT_TOKEN(anon_sym_else);
      END_STATE();
    case 112:
      if (lookahead == 'p') ADVANCE(144);
      END_STATE();
    case 113:
      if (lookahead == 'r') ADVANCE(145);
      END_STATE();
    case 114:
      if (lookahead == 'e') ADVANCE(146);
      END_STATE();
    case 115:
      if (lookahead == 't') ADVANCE(147);
      END_STATE();
    case 116:
      if (lookahead == 'u') ADVANCE(148);
      END_STATE();
    case 117:
      if (lookahead == 'r') ADVANCE(149);
      END_STATE();
    case 118:
      if (lookahead == 'i') ADVANCE(150);
      END_STATE();
    case 119:
      ACCEPT_TOKEN(anon_sym_land);
      END_STATE();
    case 120:
      ACCEPT_TOKEN(anon_sym_lazy);
      END_STATE();
    case 121:
      ACCEPT_TOKEN(anon_sym_lxor);
      END_STATE();
    case 122:
      if (lookahead == 'h') ADVANCE(151);
      END_STATE();
    case 123:
      if (lookahead == 'o') ADVANCE(152);
      END_STATE();
    case 124:
      if (lookahead == 'l') ADVANCE(153);
      END_STATE();
    case 125:
      if (lookahead == 'b') ADVANCE(154);
      END_STATE();
    case 126:
      if (lookahead == 'e') ADVANCE(155);
      END_STATE();
    case 127:
      if (lookahead == 'c') ADVANCE(156);
      END_STATE();
    case 128:
      ACCEPT_TOKEN(anon_sym_open);
      END_STATE();
    case 129:
      if (lookahead == 'a') ADVANCE(157);
      END_STATE();
    case 130:
      if (lookahead == 'c') ADVANCE(158);
      END_STATE();
    case 131:
      ACCEPT_TOKEN(anon_sym_then);
      END_STATE();
    case 132:
      ACCEPT_TOKEN(anon_sym_true);
      END_STATE();
    case 133:
      ACCEPT_TOKEN(anon_sym_type);
      END_STATE();
    case 134:
      if (lookahead == 'u') ADVANCE(159);
      END_STATE();
    case 135:
      ACCEPT_TOKEN(anon_sym_when);
      END_STATE();
    case 136:
      if (lookahead == 'e') ADVANCE(160);
      END_STATE();
    case 137:
      ACCEPT_TOKEN(anon_sym_with);
      END_STATE();
    case 138:
      if (lookahead == 't') ADVANCE(161);
      END_STATE();
    case 139:
      ACCEPT_TOKEN(anon_sym_begin);
      END_STATE();
    case 140:
      ACCEPT_TOKEN(anon_sym_class);
      END_STATE();
    case 141:
      if (lookahead == 'r') ADVANCE(162);
      END_STATE();
    case 142:
      if (lookahead == 'o') ADVANCE(163);
      END_STATE();
    case 143:
      if (lookahead == 't') ADVANCE(164);
      END_STATE();
    case 144:
      if (lookahead == 't') ADVANCE(165);
      END_STATE();
    case 145:
      if (lookahead == 'n') ADVANCE(166);
      END_STATE();
    case 146:
      ACCEPT_TOKEN(anon_sym_false);
      END_STATE();
    case 147:
      if (lookahead == 'i') ADVANCE(167);
      if (lookahead == 'o') ADVANCE(168);
      END_STATE();
    case 148:
      if (lookahead == 'd') ADVANCE(169);
      END_STATE();
    case 149:
      if (lookahead == 'i') ADVANCE(170);
      END_STATE();
    case 150:
      if (lookahead == 'a') ADVANCE(171);
      END_STATE();
    case 151:
      ACCEPT_TOKEN(anon_sym_match);
      END_STATE();
    case 152:
      if (lookahead == 'd') ADVANCE(172);
      END_STATE();
    case 153:
      if (lookahead == 'e') ADVANCE(173);
      END_STATE();
    case 154:
      if (lookahead == 'l') ADVANCE(174);
      END_STATE();
    case 155:
      if (lookahead == 'c') ADVANCE(175);
      END_STATE();
    case 156:
      if (lookahead == 't') ADVANCE(176);
      END_STATE();
    case 157:
      if (lookahead == 't') ADVANCE(177);
      END_STATE();
    case 158:
      if (lookahead == 't') ADVANCE(178);
      END_STATE();
    case 159:
      if (lookahead == 'a') ADVANCE(179);
      END_STATE();
    case 160:
      ACCEPT_TOKEN(anon_sym_while);
      END_STATE();
    case 161:
      ACCEPT_TOKEN(anon_sym_assert);
      END_STATE();
    case 162:
      if (lookahead == 'a') ADVANCE(180);
      END_STATE();
    case 163:
      ACCEPT_TOKEN(anon_sym_downto);
      END_STATE();
    case 164:
      ACCEPT_TOKEN(anon_sym_effect);
      END_STATE();
    case 165:
      if (lookahead == 'i') ADVANCE(181);
      END_STATE();
    case 166:
      if (lookahead == 'a') ADVANCE(182);
      END_STATE();
    case 167:
      if (lookahead == 'o') ADVANCE(183);
      END_STATE();
    case 168:
      if (lookahead == 'r') ADVANCE(184);
      END_STATE();
    case 169:
      if (lookahead == 'e') ADVANCE(185);
      END_STATE();
    case 170:
      if (lookahead == 't') ADVANCE(186);
      END_STATE();
    case 171:
      if (lookahead == 'l') ADVANCE(187);
      END_STATE();
    case 172:
      ACCEPT_TOKEN(anon_sym_method);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(anon_sym_module);
      END_STATE();
    case 174:
      if (lookahead == 'e') ADVANCE(188);
      END_STATE();
    case 175:
      ACCEPT_TOKEN(anon_sym_nonrec);
      END_STATE();
    case 176:
      ACCEPT_TOKEN(anon_sym_object);
      END_STATE();
    case 177:
      if (lookahead == 'e') ADVANCE(189);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(anon_sym_struct);
      END_STATE();
    case 179:
      if (lookahead == 'l') ADVANCE(190);
      END_STATE();
    case 180:
      if (lookahead == 'i') ADVANCE(191);
      END_STATE();
    case 181:
      if (lookahead == 'o') ADVANCE(192);
      END_STATE();
    case 182:
      if (lookahead == 'l') ADVANCE(193);
      END_STATE();
    case 183:
      if (lookahead == 'n') ADVANCE(194);
      END_STATE();
    case 184:
      ACCEPT_TOKEN(anon_sym_functor);
      END_STATE();
    case 185:
      ACCEPT_TOKEN(anon_sym_include);
      END_STATE();
    case 186:
      ACCEPT_TOKEN(anon_sym_inherit);
      END_STATE();
    case 187:
      if (lookahead == 'i') ADVANCE(195);
      END_STATE();
    case 188:
      ACCEPT_TOKEN(anon_sym_mutable);
      END_STATE();
    case 189:
      ACCEPT_TOKEN(anon_sym_private);
      END_STATE();
    case 190:
      ACCEPT_TOKEN(anon_sym_virtual);
      END_STATE();
    case 191:
      if (lookahead == 'n') ADVANCE(196);
      END_STATE();
    case 192:
      if (lookahead == 'n') ADVANCE(197);
      END_STATE();
    case 193:
      ACCEPT_TOKEN(anon_sym_external);
      END_STATE();
    case 194:
      ACCEPT_TOKEN(anon_sym_function);
      END_STATE();
    case 195:
      if (lookahead == 'z') ADVANCE(198);
      END_STATE();
    case 196:
      if (lookahead == 't') ADVANCE(199);
      END_STATE();
    case 197:
      ACCEPT_TOKEN(anon_sym_exception);
      END_STATE();
    case 198:
      if (lookahead == 'e') ADVANCE(200);
      END_STATE();
    case 199:
      ACCEPT_TOKEN(anon_sym_constraint);
      END_STATE();
    case 200:
      if (lookahead == 'r') ADVANCE(201);
      END_STATE();
    case 201:
      ACCEPT_TOKEN(anon_sym_initializer);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2] = {.lex_state = 11, .external_lex_state = 3, .reserved_word_set_id = 2},
  [3] = {.lex_state = 19, .external_lex_state = 3, .reserved_word_set_id = 3},
  [4] = {.lex_state = 19, .external_lex_state = 3, .reserved_word_set_id = 4},
  [5] = {.lex_state = 21, .external_lex_state = 3, .reserved_word_set_id = 5},
  [6] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [7] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [8] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [9] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [10] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [11] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [12] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [13] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [14] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [15] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [16] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [17] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [18] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [19] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [20] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [21] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [22] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [23] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [24] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [25] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [26] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [27] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [28] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [29] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [30] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [31] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [32] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [33] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [34] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [35] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [36] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [37] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 7},
  [38] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 7},
  [39] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 7},
  [40] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 7},
  [41] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 7},
  [42] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 7},
  [43] = {.lex_state = 11, .external_lex_state = 3, .reserved_word_set_id = 8},
  [44] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 6},
  [45] = {.lex_state = 11, .external_lex_state = 3, .reserved_word_set_id = 9},
  [46] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [47] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [48] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [49] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [50] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [51] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [52] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [53] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [54] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [55] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [56] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [57] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [58] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [59] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [60] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [61] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [62] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [63] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [64] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [65] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [66] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [67] = {.lex_state = 26, .external_lex_state = 3, .reserved_word_set_id = 10},
  [68] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 11},
  [69] = {.lex_state = 11, .external_lex_state = 3, .reserved_word_set_id = 12},
  [70] = {.lex_state = 28, .external_lex_state = 3, .reserved_word_set_id = 11},
  [71] = {.lex_state = 11, .external_lex_state = 3, .reserved_word_set_id = 12},
  [72] = {.lex_state = 27, .external_lex_state = 3, .reserved_word_set_id = 13},
  [73] = {.lex_state = 11, .external_lex_state = 3, .reserved_word_set_id = 14},
  [74] = {.lex_state = 11, .external_lex_state = 3, .reserved_word_set_id = 14},
  [75] = {.lex_state = 52, .external_lex_state = 3, .reserved_word_set_id = 15},
  [76] = {.lex_state = 52, .external_lex_state = 3, .reserved_word_set_id = 16},
  [77] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 17},
  [78] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [79] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 7},
  [80] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [81] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [82] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 17},
  [83] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [84] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [85] = {.lex_state = 22, .external_lex_state = 3, .reserved_word_set_id = 19},
  [86] = {.lex_state = 22, .external_lex_state = 3, .reserved_word_set_id = 19},
  [87] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 20},
  [88] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [89] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 21},
  [90] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 22},
  [91] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 20},
  [92] = {.lex_state = 52, .external_lex_state = 3, .reserved_word_set_id = 1},
  [93] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 23},
  [94] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 24},
  [95] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [96] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [97] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 1},
  [98] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [99] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 24},
  [100] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [101] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [102] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [103] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 24},
  [104] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [105] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 24},
  [106] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 1},
  [107] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [108] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [109] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [110] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 24},
  [111] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 24},
  [112] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [113] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 1},
  [114] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [115] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [116] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [117] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [118] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [119] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [120] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [121] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [122] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [123] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [124] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [125] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [126] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [127] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [128] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [129] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [130] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [131] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [132] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [133] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [134] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [135] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [136] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [137] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [138] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [139] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [140] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [141] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [142] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [143] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [144] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [145] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [146] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [147] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [148] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [149] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [150] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [151] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [152] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [153] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [154] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [155] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [156] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [157] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [158] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [159] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [160] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [161] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [162] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [163] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [164] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [165] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [166] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [167] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [168] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [169] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [170] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 1},
  [171] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 1},
  [172] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 1},
  [173] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [174] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 27},
  [175] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 28},
  [176] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 28},
  [177] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [178] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 28},
  [179] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 28},
  [180] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 28},
  [181] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [182] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [183] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [184] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 27},
  [185] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 28},
  [186] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [187] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 28},
  [188] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 28},
  [189] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [190] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [191] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 28},
  [192] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [193] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [194] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [195] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 28},
  [196] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [197] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [198] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [199] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [200] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 27},
  [201] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [202] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [203] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [204] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 27},
  [205] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 1},
  [206] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 1},
  [207] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [208] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 28},
  [209] = {.lex_state = 18, .external_lex_state = 3, .reserved_word_set_id = 1},
  [210] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [211] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [212] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [213] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [214] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [215] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [216] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [217] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [218] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [219] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [220] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [221] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [222] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [223] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [224] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [225] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [226] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [227] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [228] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [229] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [230] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [231] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [232] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [233] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [234] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [235] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [236] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [237] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [238] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [239] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [240] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [241] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [242] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [243] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [244] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [245] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [246] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [247] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [248] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [249] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [250] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [251] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [252] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [253] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [254] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [255] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [256] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [257] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [258] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [259] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [260] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [261] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [262] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [263] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [264] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [265] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [266] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [267] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [268] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [269] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [270] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [271] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [272] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [273] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [274] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [275] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [276] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [277] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [278] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [279] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [280] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [281] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [282] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [283] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [284] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [285] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [286] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [287] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [288] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [289] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [290] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [291] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [292] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [293] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [294] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [295] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [296] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [297] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [298] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [299] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [300] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [301] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [302] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [303] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [304] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [305] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [306] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [307] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [308] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [309] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [310] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [311] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [312] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [313] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [314] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [315] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [316] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [317] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [318] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [319] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [320] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [321] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [322] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [323] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [324] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [325] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [326] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [327] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [328] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [329] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [330] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [331] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [332] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [333] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [334] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [335] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [336] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [337] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [338] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [339] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [340] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [341] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [342] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [343] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [344] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [345] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [346] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [347] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [348] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [349] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [350] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [351] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [352] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [353] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [354] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [355] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [356] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [357] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [358] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [359] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [360] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [361] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [362] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [363] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [364] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [365] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [366] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [367] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [368] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [369] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [370] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [371] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [372] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [373] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [374] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [375] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [376] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [377] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [378] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [379] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [380] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [381] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [382] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [383] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [384] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [385] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [386] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [387] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [388] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [389] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [390] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [391] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [392] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [393] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [394] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [395] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [396] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [397] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [398] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [399] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [400] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [401] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [402] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [403] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [404] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [405] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [406] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [407] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [408] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [409] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [410] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [411] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [412] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [413] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [414] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [415] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [416] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [417] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [418] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [419] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [420] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [421] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [422] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [423] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [424] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [425] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [426] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [427] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [428] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [429] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [430] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [431] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [432] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [433] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [434] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [435] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [436] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [437] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [438] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [439] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [440] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [441] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [442] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [443] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [444] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [445] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [446] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [447] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [448] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [449] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [450] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [451] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [452] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [453] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [454] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [455] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [456] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [457] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [458] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [459] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [460] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [461] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [462] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [463] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [464] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [465] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [466] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [467] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [468] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [469] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [470] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [471] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [472] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [473] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [474] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [475] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [476] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [477] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [478] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [479] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [480] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [481] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [482] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [483] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [484] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [485] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [486] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [487] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [488] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [489] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [490] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [491] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [492] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [493] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [494] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [495] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [496] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [497] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [498] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [499] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [500] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [501] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [502] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [503] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [504] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [505] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [506] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [507] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [508] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [509] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [510] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [511] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [512] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [513] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [514] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [515] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [516] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [517] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [518] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [519] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [520] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [521] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [522] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [523] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [524] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [525] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [526] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [527] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [528] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [529] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [530] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [531] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [532] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [533] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [534] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [535] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [536] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [537] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [538] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [539] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [540] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [541] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [542] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [543] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [544] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [545] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [546] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [547] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [548] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [549] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [550] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [551] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [552] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [553] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [554] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [555] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [556] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [557] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [558] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [559] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [560] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [561] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [562] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [563] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [564] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [565] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [566] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [567] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [568] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [569] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [570] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [571] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [572] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [573] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [574] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [575] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [576] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [577] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [578] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [579] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [580] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [581] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [582] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [583] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [584] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [585] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [586] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [587] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [588] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [589] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [590] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [591] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [592] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [593] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [594] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [595] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [596] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [597] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [598] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [599] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [600] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [601] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [602] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [603] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [604] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [605] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [606] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [607] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [608] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [609] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [610] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [611] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [612] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [613] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [614] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [615] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [616] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [617] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [618] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [619] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [620] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [621] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [622] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [623] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [624] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [625] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [626] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [627] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [628] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [629] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [630] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [631] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [632] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [633] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [634] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [635] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [636] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [637] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [638] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [639] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [640] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [641] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [642] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [643] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [644] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [645] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [646] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [647] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [648] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [649] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [650] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [651] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [652] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [653] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [654] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [655] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [656] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [657] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [658] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [659] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [660] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [661] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [662] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [663] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [664] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [665] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [666] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [667] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [668] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [669] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [670] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [671] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [672] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [673] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [674] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [675] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [676] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [677] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [678] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [679] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [680] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [681] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [682] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [683] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [684] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [685] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [686] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [687] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [688] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [689] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [690] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [691] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [692] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [693] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [694] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [695] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [696] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [697] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [698] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [699] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [700] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [701] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [702] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [703] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [704] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [705] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [706] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [707] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [708] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [709] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [710] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [711] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [712] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [713] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 26},
  [714] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 29},
  [715] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [716] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 29},
  [717] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [718] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [719] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [720] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [721] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [722] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [723] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [724] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [725] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [726] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [727] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [728] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [729] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [730] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [731] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [732] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [733] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [734] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [735] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [736] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [737] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [738] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [739] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [740] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [741] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [742] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 33},
  [743] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [744] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [745] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [746] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [747] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [748] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 31},
  [749] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [750] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [751] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [752] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [753] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [754] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [755] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [756] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [757] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [758] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [759] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [760] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [761] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [762] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [763] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [764] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [765] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [766] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [767] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [768] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [769] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [770] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [771] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [772] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [773] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [774] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [775] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [776] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [777] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [778] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [779] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [780] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [781] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [782] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [783] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [784] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [785] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [786] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [787] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [788] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [789] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [790] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [791] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [792] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [793] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [794] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [795] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [796] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [797] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [798] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [799] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [800] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [801] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [802] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [803] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [804] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [805] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [806] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [807] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [808] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [809] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [810] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [811] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [812] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [813] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [814] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [815] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [816] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [817] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [818] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [819] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [820] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [821] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [822] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [823] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [824] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [825] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [826] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [827] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [828] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [829] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [830] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [831] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [832] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 35},
  [833] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 35},
  [834] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 35},
  [835] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 17},
  [836] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 36},
  [837] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 36},
  [838] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [839] = {.lex_state = 29, .external_lex_state = 3, .reserved_word_set_id = 18},
  [840] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [841] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [842] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [843] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [844] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [845] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [846] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [847] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 18},
  [848] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [849] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [850] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [851] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [852] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [853] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [854] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [855] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [856] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [857] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [858] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [859] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [860] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [861] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [862] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [863] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [864] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [865] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [866] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [867] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [868] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [869] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [870] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [871] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [872] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [873] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [874] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [875] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [876] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [877] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [878] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [879] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [880] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [881] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [882] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [883] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [884] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [885] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [886] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [887] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [888] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [889] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [890] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [891] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [892] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [893] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [894] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [895] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [896] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [897] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [898] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [899] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [900] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [901] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [902] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [903] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [904] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [905] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [906] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [907] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [908] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 38},
  [909] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [910] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [911] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [912] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [913] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [914] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [915] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [916] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [917] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [918] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [919] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [920] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [921] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [922] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [923] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [924] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [925] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [926] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [927] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [928] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [929] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [930] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [931] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [932] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [933] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [934] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [935] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [936] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 38},
  [937] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [938] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [939] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [940] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [941] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [942] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [943] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [944] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [945] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [946] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [947] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [948] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [949] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [950] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [951] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [952] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [953] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [954] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [955] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [956] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [957] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [958] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [959] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [960] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [961] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 39},
  [962] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [963] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [964] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [965] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [966] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [967] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 39},
  [968] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 39},
  [969] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [970] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 39},
  [971] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [972] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 39},
  [973] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 39},
  [974] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 39},
  [975] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 39},
  [976] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 39},
  [977] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 39},
  [978] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [979] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [980] = {.lex_state = 31, .external_lex_state = 3, .reserved_word_set_id = 17},
  [981] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [982] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [983] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [984] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [985] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [986] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [987] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [988] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [989] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [990] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [991] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [992] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [993] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [994] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [995] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [996] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [997] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [998] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [999] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1000] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1001] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1002] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1003] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1004] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1005] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1006] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1007] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1008] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1009] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1010] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1011] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1012] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1013] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1014] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1015] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1016] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1017] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1018] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1019] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1020] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1021] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1022] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1023] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1024] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1025] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1026] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1027] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1028] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1029] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1030] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1031] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1032] = {.lex_state = 15, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1033] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 40},
  [1034] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1035] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1036] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1037] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [1038] = {.lex_state = 149, .external_lex_state = 3, .reserved_word_set_id = 41},
  [1039] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1040] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1041] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 17},
  [1042] = {.lex_state = 149, .external_lex_state = 3, .reserved_word_set_id = 41},
  [1043] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1044] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1045] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1046] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1047] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1048] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1049] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1050] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1051] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1052] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1053] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1054] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1055] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1056] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 20},
  [1057] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1058] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1059] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1060] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1061] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1062] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1063] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1064] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1065] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1066] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1067] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1068] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1069] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1070] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1071] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1072] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1073] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1074] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1075] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1076] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1077] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1078] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1079] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1080] = {.lex_state = 22, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1081] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1082] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1083] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1084] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1085] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1086] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1087] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1088] = {.lex_state = 57, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1089] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1090] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1091] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1092] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1093] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1094] = {.lex_state = 57, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1095] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 20},
  [1096] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1097] = {.lex_state = 149, .external_lex_state = 3, .reserved_word_set_id = 41},
  [1098] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1099] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1100] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1101] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1102] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1103] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1104] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1105] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1106] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1107] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1108] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1109] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1110] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1111] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1112] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1113] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1114] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1115] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1116] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1117] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1118] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1119] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1120] = {.lex_state = 149, .external_lex_state = 3, .reserved_word_set_id = 41},
  [1121] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1122] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1123] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 41},
  [1124] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1125] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1126] = {.lex_state = 16, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1127] = {.lex_state = 57, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1128] = {.lex_state = 57, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1129] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 41},
  [1130] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 39},
  [1131] = {.lex_state = 22, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1132] = {.lex_state = 22, .external_lex_state = 3, .reserved_word_set_id = 19},
  [1133] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1134] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1135] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1136] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1137] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1138] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1139] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1140] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1141] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1142] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1143] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1144] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1145] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 43},
  [1146] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 21},
  [1147] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1148] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1149] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1150] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1151] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1152] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1153] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1154] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1155] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1156] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1157] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1158] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1159] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1160] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1161] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1162] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1163] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1164] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1165] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1166] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1167] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1168] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1169] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1170] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1171] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1172] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1173] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1174] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1175] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1176] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1177] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1178] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1179] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1180] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1181] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1182] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1183] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1184] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1185] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1186] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1187] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1188] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1189] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1190] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1191] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1192] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1193] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1194] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1195] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1196] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1197] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1198] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1199] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1200] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1201] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1202] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1203] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1204] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1205] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1206] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1207] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1208] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1209] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1210] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1211] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1212] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1213] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1214] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1215] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 44},
  [1216] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1217] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1218] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1219] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1220] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1221] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1222] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1223] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 44},
  [1224] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1225] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1226] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1227] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1228] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1229] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1230] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1231] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1232] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1233] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1234] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1235] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1236] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1237] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1238] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1239] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1240] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1241] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 44},
  [1242] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1243] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1244] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1245] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1246] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1247] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1248] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1249] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1250] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1251] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1252] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1253] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1254] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1255] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1256] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1257] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1258] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1259] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1260] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1261] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1262] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1263] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1264] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1265] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1266] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1267] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1268] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1269] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1270] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1271] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1272] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1273] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1274] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1275] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1276] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1277] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1278] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1279] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1280] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1281] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1282] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1283] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1284] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1285] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1286] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1287] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1288] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1289] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1290] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1291] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1292] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1293] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1294] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1295] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1296] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1297] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1298] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1299] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1300] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1301] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1302] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1303] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1304] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1305] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1306] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1307] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1308] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1309] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1310] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1311] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1312] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1313] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1314] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1315] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1316] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1317] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1318] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1319] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1320] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1321] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1322] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1323] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1324] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1325] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1326] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1327] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1328] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1329] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1330] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1331] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1332] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1333] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1334] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1335] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1336] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1337] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1338] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1339] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1340] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1341] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1342] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1343] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1344] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1345] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1346] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1347] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1348] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1349] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1350] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1351] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1352] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1353] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1354] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1355] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1356] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1357] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1358] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1359] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1360] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1361] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1362] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1363] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1364] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1365] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1366] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1367] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1368] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1369] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1370] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1371] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1372] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1373] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 45},
  [1374] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1375] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1376] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1377] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1378] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1379] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1380] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1381] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1382] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1383] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1384] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1385] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 42},
  [1386] = {.lex_state = 13, .external_lex_state = 4, .reserved_word_set_id = 46},
  [1387] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 47},
  [1388] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1389] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1390] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 49},
  [1391] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1392] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1393] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1394] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1395] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1396] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1397] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1398] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1399] = {.lex_state = 13, .external_lex_state = 4, .reserved_word_set_id = 46},
  [1400] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1401] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1402] = {.lex_state = 13, .external_lex_state = 4, .reserved_word_set_id = 46},
  [1403] = {.lex_state = 13, .external_lex_state = 4, .reserved_word_set_id = 46},
  [1404] = {.lex_state = 13, .external_lex_state = 4, .reserved_word_set_id = 46},
  [1405] = {.lex_state = 13, .external_lex_state = 4, .reserved_word_set_id = 46},
  [1406] = {.lex_state = 13, .external_lex_state = 4, .reserved_word_set_id = 46},
  [1407] = {.lex_state = 13, .external_lex_state = 4, .reserved_word_set_id = 46},
  [1408] = {.lex_state = 13, .external_lex_state = 4, .reserved_word_set_id = 46},
  [1409] = {.lex_state = 13, .external_lex_state = 4, .reserved_word_set_id = 46},
  [1410] = {.lex_state = 13, .external_lex_state = 4, .reserved_word_set_id = 46},
  [1411] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 45},
  [1412] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1413] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1414] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1415] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1416] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1417] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1418] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1419] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1420] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1421] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1422] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1423] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1424] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1425] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1426] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1427] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1428] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1429] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1430] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1431] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1432] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1433] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1434] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1435] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1436] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1437] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1438] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1439] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1440] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1441] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1442] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1443] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1444] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1445] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1446] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1447] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1448] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1449] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1450] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1451] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1452] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1453] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1454] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1455] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1456] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1457] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1458] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1459] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1460] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1461] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1462] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1463] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1464] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1465] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1466] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1467] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1468] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1469] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1470] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1471] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1472] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1473] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1474] = {.lex_state = 40, .external_lex_state = 2},
  [1475] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1476] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1477] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1478] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1479] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1480] = {.lex_state = 40, .external_lex_state = 2},
  [1481] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1482] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1483] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1484] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1485] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1486] = {.lex_state = 40, .external_lex_state = 2},
  [1487] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1488] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1489] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1490] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1491] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1492] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1493] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1494] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1495] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1496] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1497] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1498] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1499] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1500] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1501] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1502] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1503] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1504] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1505] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1506] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 46},
  [1507] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1508] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1509] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1510] = {.lex_state = 53, .external_lex_state = 3, .reserved_word_set_id = 7},
  [1511] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1512] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1513] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1514] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1515] = {.lex_state = 53, .external_lex_state = 3, .reserved_word_set_id = 7},
  [1516] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1517] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1518] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1519] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1520] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1521] = {.lex_state = 53, .external_lex_state = 3, .reserved_word_set_id = 7},
  [1522] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1523] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1524] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1525] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1526] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1527] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1528] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1529] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1530] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1531] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1532] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1533] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1534] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1535] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1536] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1537] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1538] = {.lex_state = 53, .external_lex_state = 3, .reserved_word_set_id = 7},
  [1539] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1540] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1541] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1542] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1543] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1544] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1545] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1546] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1547] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1548] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1549] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1550] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1551] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1552] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1553] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1554] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1555] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1556] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1557] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1558] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1559] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 51},
  [1560] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1561] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1562] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1563] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1564] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1565] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1566] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1567] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1568] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1569] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1570] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1571] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1572] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1573] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1574] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1575] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1576] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1577] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1578] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1579] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1580] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1581] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1582] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1583] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1584] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1585] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1586] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1587] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1588] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1589] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1590] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1591] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1592] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1593] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1594] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1595] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1596] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1597] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1598] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1599] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1600] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1601] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1602] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1603] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1604] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1605] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1606] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1607] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1608] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1609] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1610] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1611] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1612] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1613] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1614] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1615] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1616] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1617] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1618] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1619] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1620] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 51},
  [1621] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1622] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1623] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1624] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1625] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1626] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 7},
  [1627] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1628] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1629] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1630] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1631] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1632] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1633] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1634] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1635] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1636] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1637] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1638] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 52},
  [1639] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1640] = {.lex_state = 12, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1641] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1642] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1643] = {.lex_state = 35, .external_lex_state = 2},
  [1644] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1645] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1646] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 7},
  [1647] = {.lex_state = 40, .external_lex_state = 2},
  [1648] = {.lex_state = 40, .external_lex_state = 2},
  [1649] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1650] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1651] = {.lex_state = 40, .external_lex_state = 2},
  [1652] = {.lex_state = 40, .external_lex_state = 2},
  [1653] = {.lex_state = 40, .external_lex_state = 2},
  [1654] = {.lex_state = 14, .external_lex_state = 3, .reserved_word_set_id = 30},
  [1655] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1656] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1657] = {.lex_state = 40, .external_lex_state = 2},
  [1658] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 53},
  [1659] = {.lex_state = 40, .external_lex_state = 2},
  [1660] = {.lex_state = 40, .external_lex_state = 2},
  [1661] = {.lex_state = 40, .external_lex_state = 2},
  [1662] = {.lex_state = 35, .external_lex_state = 2},
  [1663] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1664] = {.lex_state = 40, .external_lex_state = 2},
  [1665] = {.lex_state = 40, .external_lex_state = 2},
  [1666] = {.lex_state = 40, .external_lex_state = 2},
  [1667] = {.lex_state = 40, .external_lex_state = 2},
  [1668] = {.lex_state = 40, .external_lex_state = 2},
  [1669] = {.lex_state = 40, .external_lex_state = 2},
  [1670] = {.lex_state = 40, .external_lex_state = 2},
  [1671] = {.lex_state = 40, .external_lex_state = 2},
  [1672] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1673] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1674] = {.lex_state = 40, .external_lex_state = 2},
  [1675] = {.lex_state = 40, .external_lex_state = 2},
  [1676] = {.lex_state = 40, .external_lex_state = 2},
  [1677] = {.lex_state = 35, .external_lex_state = 2},
  [1678] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1679] = {.lex_state = 35, .external_lex_state = 2},
  [1680] = {.lex_state = 35, .external_lex_state = 2},
  [1681] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1682] = {.lex_state = 35, .external_lex_state = 2},
  [1683] = {.lex_state = 35, .external_lex_state = 2},
  [1684] = {.lex_state = 35, .external_lex_state = 2},
  [1685] = {.lex_state = 35, .external_lex_state = 2},
  [1686] = {.lex_state = 35, .external_lex_state = 2},
  [1687] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1688] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1689] = {.lex_state = 35, .external_lex_state = 2},
  [1690] = {.lex_state = 35, .external_lex_state = 2},
  [1691] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1692] = {.lex_state = 35, .external_lex_state = 2},
  [1693] = {.lex_state = 35, .external_lex_state = 2},
  [1694] = {.lex_state = 35, .external_lex_state = 2},
  [1695] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1696] = {.lex_state = 40, .external_lex_state = 2},
  [1697] = {.lex_state = 35, .external_lex_state = 2},
  [1698] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1699] = {.lex_state = 35, .external_lex_state = 2},
  [1700] = {.lex_state = 35, .external_lex_state = 2},
  [1701] = {.lex_state = 35, .external_lex_state = 2},
  [1702] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1703] = {.lex_state = 35, .external_lex_state = 2},
  [1704] = {.lex_state = 35, .external_lex_state = 2},
  [1705] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1706] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1707] = {.lex_state = 40, .external_lex_state = 2},
  [1708] = {.lex_state = 35, .external_lex_state = 2},
  [1709] = {.lex_state = 35, .external_lex_state = 2},
  [1710] = {.lex_state = 35, .external_lex_state = 2},
  [1711] = {.lex_state = 35, .external_lex_state = 2},
  [1712] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1713] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1714] = {.lex_state = 35, .external_lex_state = 2},
  [1715] = {.lex_state = 35, .external_lex_state = 2},
  [1716] = {.lex_state = 35, .external_lex_state = 2},
  [1717] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1718] = {.lex_state = 35, .external_lex_state = 2},
  [1719] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1720] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1721] = {.lex_state = 40, .external_lex_state = 2},
  [1722] = {.lex_state = 40, .external_lex_state = 2},
  [1723] = {.lex_state = 35, .external_lex_state = 2},
  [1724] = {.lex_state = 35, .external_lex_state = 2},
  [1725] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1726] = {.lex_state = 40, .external_lex_state = 2},
  [1727] = {.lex_state = 35, .external_lex_state = 2},
  [1728] = {.lex_state = 40, .external_lex_state = 2},
  [1729] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1730] = {.lex_state = 35, .external_lex_state = 2},
  [1731] = {.lex_state = 40, .external_lex_state = 2},
  [1732] = {.lex_state = 40, .external_lex_state = 2},
  [1733] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1734] = {.lex_state = 40, .external_lex_state = 2},
  [1735] = {.lex_state = 40, .external_lex_state = 2},
  [1736] = {.lex_state = 35, .external_lex_state = 2},
  [1737] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1738] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1739] = {.lex_state = 40, .external_lex_state = 2},
  [1740] = {.lex_state = 35, .external_lex_state = 2},
  [1741] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1742] = {.lex_state = 35, .external_lex_state = 2},
  [1743] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1744] = {.lex_state = 35, .external_lex_state = 2},
  [1745] = {.lex_state = 35, .external_lex_state = 2},
  [1746] = {.lex_state = 35, .external_lex_state = 2},
  [1747] = {.lex_state = 35, .external_lex_state = 2},
  [1748] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1749] = {.lex_state = 35, .external_lex_state = 2},
  [1750] = {.lex_state = 35, .external_lex_state = 2},
  [1751] = {.lex_state = 35, .external_lex_state = 2},
  [1752] = {.lex_state = 35, .external_lex_state = 2},
  [1753] = {.lex_state = 35, .external_lex_state = 2},
  [1754] = {.lex_state = 35, .external_lex_state = 2},
  [1755] = {.lex_state = 35, .external_lex_state = 2},
  [1756] = {.lex_state = 35, .external_lex_state = 2},
  [1757] = {.lex_state = 35, .external_lex_state = 2},
  [1758] = {.lex_state = 35, .external_lex_state = 2},
  [1759] = {.lex_state = 35, .external_lex_state = 2},
  [1760] = {.lex_state = 35, .external_lex_state = 2},
  [1761] = {.lex_state = 35, .external_lex_state = 2},
  [1762] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 29},
  [1763] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 1},
  [1764] = {.lex_state = 40, .external_lex_state = 2},
  [1765] = {.lex_state = 35, .external_lex_state = 2},
  [1766] = {.lex_state = 35, .external_lex_state = 2},
  [1767] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 50},
  [1768] = {.lex_state = 35, .external_lex_state = 2},
  [1769] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 48},
  [1770] = {.lex_state = 35, .external_lex_state = 2},
  [1771] = {.lex_state = 40, .external_lex_state = 2},
  [1772] = {.lex_state = 40, .external_lex_state = 2},
  [1773] = {.lex_state = 148, .external_lex_state = 2},
  [1774] = {.lex_state = 148, .external_lex_state = 2},
  [1775] = {.lex_state = 40, .external_lex_state = 2},
  [1776] = {.lex_state = 148, .external_lex_state = 2},
  [1777] = {.lex_state = 148, .external_lex_state = 2},
  [1778] = {.lex_state = 36, .external_lex_state = 2},
  [1779] = {.lex_state = 36, .external_lex_state = 2},
  [1780] = {.lex_state = 36, .external_lex_state = 2},
  [1781] = {.lex_state = 148, .external_lex_state = 2},
  [1782] = {.lex_state = 148, .external_lex_state = 2},
  [1783] = {.lex_state = 149, .external_lex_state = 3, .reserved_word_set_id = 54},
  [1784] = {.lex_state = 148, .external_lex_state = 2},
  [1785] = {.lex_state = 148, .external_lex_state = 2},
  [1786] = {.lex_state = 148, .external_lex_state = 2},
  [1787] = {.lex_state = 148, .external_lex_state = 2},
  [1788] = {.lex_state = 148, .external_lex_state = 2},
  [1789] = {.lex_state = 148, .external_lex_state = 2},
  [1790] = {.lex_state = 148, .external_lex_state = 2},
  [1791] = {.lex_state = 148, .external_lex_state = 2},
  [1792] = {.lex_state = 40, .external_lex_state = 2},
  [1793] = {.lex_state = 40, .external_lex_state = 2},
  [1794] = {.lex_state = 148, .external_lex_state = 2},
  [1795] = {.lex_state = 58, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1796] = {.lex_state = 148, .external_lex_state = 2},
  [1797] = {.lex_state = 148, .external_lex_state = 2},
  [1798] = {.lex_state = 148, .external_lex_state = 2},
  [1799] = {.lex_state = 58, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1800] = {.lex_state = 148, .external_lex_state = 2},
  [1801] = {.lex_state = 148, .external_lex_state = 2},
  [1802] = {.lex_state = 148, .external_lex_state = 2},
  [1803] = {.lex_state = 58, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1804] = {.lex_state = 148, .external_lex_state = 2},
  [1805] = {.lex_state = 148, .external_lex_state = 2},
  [1806] = {.lex_state = 148, .external_lex_state = 2},
  [1807] = {.lex_state = 148, .external_lex_state = 2},
  [1808] = {.lex_state = 148, .external_lex_state = 2},
  [1809] = {.lex_state = 148, .external_lex_state = 2},
  [1810] = {.lex_state = 148, .external_lex_state = 2},
  [1811] = {.lex_state = 148, .external_lex_state = 2},
  [1812] = {.lex_state = 35, .external_lex_state = 2},
  [1813] = {.lex_state = 35, .external_lex_state = 2},
  [1814] = {.lex_state = 148, .external_lex_state = 2},
  [1815] = {.lex_state = 148, .external_lex_state = 2},
  [1816] = {.lex_state = 148, .external_lex_state = 2},
  [1817] = {.lex_state = 148, .external_lex_state = 2},
  [1818] = {.lex_state = 35, .external_lex_state = 2},
  [1819] = {.lex_state = 148, .external_lex_state = 2},
  [1820] = {.lex_state = 148, .external_lex_state = 2},
  [1821] = {.lex_state = 58, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1822] = {.lex_state = 148, .external_lex_state = 2},
  [1823] = {.lex_state = 58, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1824] = {.lex_state = 40, .external_lex_state = 2},
  [1825] = {.lex_state = 58, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1826] = {.lex_state = 64, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1827] = {.lex_state = 148, .external_lex_state = 2},
  [1828] = {.lex_state = 40, .external_lex_state = 2},
  [1829] = {.lex_state = 58, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1830] = {.lex_state = 35, .external_lex_state = 2},
  [1831] = {.lex_state = 148, .external_lex_state = 2},
  [1832] = {.lex_state = 148, .external_lex_state = 2},
  [1833] = {.lex_state = 148, .external_lex_state = 2},
  [1834] = {.lex_state = 58, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1835] = {.lex_state = 148, .external_lex_state = 2},
  [1836] = {.lex_state = 148, .external_lex_state = 2},
  [1837] = {.lex_state = 148, .external_lex_state = 2},
  [1838] = {.lex_state = 148, .external_lex_state = 2},
  [1839] = {.lex_state = 148, .external_lex_state = 2},
  [1840] = {.lex_state = 148, .external_lex_state = 2},
  [1841] = {.lex_state = 148, .external_lex_state = 2},
  [1842] = {.lex_state = 58, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1843] = {.lex_state = 58, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1844] = {.lex_state = 148, .external_lex_state = 2},
  [1845] = {.lex_state = 35, .external_lex_state = 2},
  [1846] = {.lex_state = 36, .external_lex_state = 2},
  [1847] = {.lex_state = 36, .external_lex_state = 2},
  [1848] = {.lex_state = 36, .external_lex_state = 2},
  [1849] = {.lex_state = 36, .external_lex_state = 2},
  [1850] = {.lex_state = 36, .external_lex_state = 2},
  [1851] = {.lex_state = 36, .external_lex_state = 2},
  [1852] = {.lex_state = 36, .external_lex_state = 2},
  [1853] = {.lex_state = 148, .external_lex_state = 2},
  [1854] = {.lex_state = 36, .external_lex_state = 2},
  [1855] = {.lex_state = 148, .external_lex_state = 2},
  [1856] = {.lex_state = 149, .external_lex_state = 3, .reserved_word_set_id = 54},
  [1857] = {.lex_state = 148, .external_lex_state = 2},
  [1858] = {.lex_state = 36, .external_lex_state = 2},
  [1859] = {.lex_state = 148, .external_lex_state = 2},
  [1860] = {.lex_state = 148, .external_lex_state = 2},
  [1861] = {.lex_state = 40, .external_lex_state = 2},
  [1862] = {.lex_state = 148, .external_lex_state = 2},
  [1863] = {.lex_state = 148, .external_lex_state = 2},
  [1864] = {.lex_state = 36, .external_lex_state = 2},
  [1865] = {.lex_state = 148, .external_lex_state = 2},
  [1866] = {.lex_state = 35, .external_lex_state = 2},
  [1867] = {.lex_state = 35, .external_lex_state = 2},
  [1868] = {.lex_state = 35, .external_lex_state = 2},
  [1869] = {.lex_state = 148, .external_lex_state = 2},
  [1870] = {.lex_state = 35, .external_lex_state = 2},
  [1871] = {.lex_state = 148, .external_lex_state = 2},
  [1872] = {.lex_state = 148, .external_lex_state = 2},
  [1873] = {.lex_state = 35, .external_lex_state = 2},
  [1874] = {.lex_state = 35, .external_lex_state = 2},
  [1875] = {.lex_state = 148, .external_lex_state = 2},
  [1876] = {.lex_state = 148, .external_lex_state = 2},
  [1877] = {.lex_state = 35, .external_lex_state = 2},
  [1878] = {.lex_state = 35, .external_lex_state = 2},
  [1879] = {.lex_state = 148, .external_lex_state = 2},
  [1880] = {.lex_state = 35, .external_lex_state = 2},
  [1881] = {.lex_state = 35, .external_lex_state = 2},
  [1882] = {.lex_state = 35, .external_lex_state = 2},
  [1883] = {.lex_state = 148, .external_lex_state = 2},
  [1884] = {.lex_state = 35, .external_lex_state = 2},
  [1885] = {.lex_state = 35, .external_lex_state = 2},
  [1886] = {.lex_state = 35, .external_lex_state = 2},
  [1887] = {.lex_state = 148, .external_lex_state = 2},
  [1888] = {.lex_state = 36, .external_lex_state = 2},
  [1889] = {.lex_state = 35, .external_lex_state = 2},
  [1890] = {.lex_state = 148, .external_lex_state = 2},
  [1891] = {.lex_state = 35, .external_lex_state = 2},
  [1892] = {.lex_state = 35, .external_lex_state = 2},
  [1893] = {.lex_state = 35, .external_lex_state = 2},
  [1894] = {.lex_state = 36, .external_lex_state = 2},
  [1895] = {.lex_state = 36, .external_lex_state = 2},
  [1896] = {.lex_state = 148, .external_lex_state = 2},
  [1897] = {.lex_state = 148, .external_lex_state = 2},
  [1898] = {.lex_state = 35, .external_lex_state = 2},
  [1899] = {.lex_state = 148, .external_lex_state = 2},
  [1900] = {.lex_state = 148, .external_lex_state = 2},
  [1901] = {.lex_state = 148, .external_lex_state = 2},
  [1902] = {.lex_state = 148, .external_lex_state = 2},
  [1903] = {.lex_state = 148, .external_lex_state = 2},
  [1904] = {.lex_state = 35, .external_lex_state = 2},
  [1905] = {.lex_state = 35, .external_lex_state = 2},
  [1906] = {.lex_state = 148, .external_lex_state = 2},
  [1907] = {.lex_state = 148, .external_lex_state = 2},
  [1908] = {.lex_state = 148, .external_lex_state = 2},
  [1909] = {.lex_state = 148, .external_lex_state = 2},
  [1910] = {.lex_state = 148, .external_lex_state = 2},
  [1911] = {.lex_state = 148, .external_lex_state = 2},
  [1912] = {.lex_state = 148, .external_lex_state = 2},
  [1913] = {.lex_state = 148, .external_lex_state = 2},
  [1914] = {.lex_state = 148, .external_lex_state = 2},
  [1915] = {.lex_state = 148, .external_lex_state = 2},
  [1916] = {.lex_state = 35, .external_lex_state = 2},
  [1917] = {.lex_state = 148, .external_lex_state = 2},
  [1918] = {.lex_state = 148, .external_lex_state = 2},
  [1919] = {.lex_state = 36, .external_lex_state = 2},
  [1920] = {.lex_state = 36, .external_lex_state = 2},
  [1921] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1922] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1923] = {.lex_state = 34, .external_lex_state = 2},
  [1924] = {.lex_state = 36, .external_lex_state = 2},
  [1925] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1926] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1927] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1928] = {.lex_state = 36, .external_lex_state = 2},
  [1929] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1930] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1931] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1932] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1933] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1934] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1935] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1936] = {.lex_state = 34, .external_lex_state = 2},
  [1937] = {.lex_state = 36, .external_lex_state = 2},
  [1938] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1939] = {.lex_state = 36, .external_lex_state = 2},
  [1940] = {.lex_state = 36, .external_lex_state = 2},
  [1941] = {.lex_state = 34, .external_lex_state = 2},
  [1942] = {.lex_state = 36, .external_lex_state = 2},
  [1943] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1944] = {.lex_state = 36, .external_lex_state = 2},
  [1945] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1946] = {.lex_state = 36, .external_lex_state = 2},
  [1947] = {.lex_state = 36, .external_lex_state = 2},
  [1948] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1949] = {.lex_state = 36, .external_lex_state = 2},
  [1950] = {.lex_state = 36, .external_lex_state = 2},
  [1951] = {.lex_state = 36, .external_lex_state = 2},
  [1952] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1953] = {.lex_state = 36, .external_lex_state = 2},
  [1954] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1955] = {.lex_state = 36, .external_lex_state = 2},
  [1956] = {.lex_state = 36, .external_lex_state = 2},
  [1957] = {.lex_state = 36, .external_lex_state = 2},
  [1958] = {.lex_state = 36, .external_lex_state = 2},
  [1959] = {.lex_state = 36, .external_lex_state = 2},
  [1960] = {.lex_state = 36, .external_lex_state = 2},
  [1961] = {.lex_state = 36, .external_lex_state = 2},
  [1962] = {.lex_state = 36, .external_lex_state = 2},
  [1963] = {.lex_state = 36, .external_lex_state = 2},
  [1964] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1965] = {.lex_state = 36, .external_lex_state = 2},
  [1966] = {.lex_state = 36, .external_lex_state = 2},
  [1967] = {.lex_state = 36, .external_lex_state = 2},
  [1968] = {.lex_state = 36, .external_lex_state = 2},
  [1969] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1970] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1971] = {.lex_state = 36, .external_lex_state = 2},
  [1972] = {.lex_state = 36, .external_lex_state = 2},
  [1973] = {.lex_state = 36, .external_lex_state = 2},
  [1974] = {.lex_state = 36, .external_lex_state = 2},
  [1975] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1976] = {.lex_state = 36, .external_lex_state = 2},
  [1977] = {.lex_state = 36, .external_lex_state = 2},
  [1978] = {.lex_state = 36, .external_lex_state = 2},
  [1979] = {.lex_state = 36, .external_lex_state = 2},
  [1980] = {.lex_state = 36, .external_lex_state = 2},
  [1981] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1982] = {.lex_state = 36, .external_lex_state = 2},
  [1983] = {.lex_state = 36, .external_lex_state = 2},
  [1984] = {.lex_state = 36, .external_lex_state = 2},
  [1985] = {.lex_state = 36, .external_lex_state = 2},
  [1986] = {.lex_state = 36, .external_lex_state = 2},
  [1987] = {.lex_state = 36, .external_lex_state = 2},
  [1988] = {.lex_state = 36, .external_lex_state = 2},
  [1989] = {.lex_state = 36, .external_lex_state = 2},
  [1990] = {.lex_state = 36, .external_lex_state = 2},
  [1991] = {.lex_state = 36, .external_lex_state = 2},
  [1992] = {.lex_state = 36, .external_lex_state = 2},
  [1993] = {.lex_state = 40, .external_lex_state = 2},
  [1994] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1995] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1996] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1997] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [1998] = {.lex_state = 149, .external_lex_state = 3, .reserved_word_set_id = 54},
  [1999] = {.lex_state = 36, .external_lex_state = 2},
  [2000] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2001] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2002] = {.lex_state = 34, .external_lex_state = 2},
  [2003] = {.lex_state = 36, .external_lex_state = 2},
  [2004] = {.lex_state = 34, .external_lex_state = 2},
  [2005] = {.lex_state = 36, .external_lex_state = 2},
  [2006] = {.lex_state = 36, .external_lex_state = 2},
  [2007] = {.lex_state = 34, .external_lex_state = 2},
  [2008] = {.lex_state = 149, .external_lex_state = 3, .reserved_word_set_id = 54},
  [2009] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2010] = {.lex_state = 34, .external_lex_state = 2},
  [2011] = {.lex_state = 36, .external_lex_state = 2},
  [2012] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2013] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2014] = {.lex_state = 34, .external_lex_state = 2},
  [2015] = {.lex_state = 34, .external_lex_state = 2},
  [2016] = {.lex_state = 34, .external_lex_state = 2},
  [2017] = {.lex_state = 34, .external_lex_state = 2},
  [2018] = {.lex_state = 34, .external_lex_state = 2},
  [2019] = {.lex_state = 34, .external_lex_state = 2},
  [2020] = {.lex_state = 36, .external_lex_state = 2},
  [2021] = {.lex_state = 36, .external_lex_state = 2},
  [2022] = {.lex_state = 40, .external_lex_state = 2},
  [2023] = {.lex_state = 36, .external_lex_state = 2},
  [2024] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2025] = {.lex_state = 149, .external_lex_state = 3, .reserved_word_set_id = 54},
  [2026] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2027] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2028] = {.lex_state = 40, .external_lex_state = 2},
  [2029] = {.lex_state = 36, .external_lex_state = 2},
  [2030] = {.lex_state = 36, .external_lex_state = 2},
  [2031] = {.lex_state = 36, .external_lex_state = 2},
  [2032] = {.lex_state = 36, .external_lex_state = 2},
  [2033] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2034] = {.lex_state = 36, .external_lex_state = 2},
  [2035] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2036] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2037] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2038] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2039] = {.lex_state = 36, .external_lex_state = 2},
  [2040] = {.lex_state = 36, .external_lex_state = 2},
  [2041] = {.lex_state = 36, .external_lex_state = 2},
  [2042] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2043] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2044] = {.lex_state = 36, .external_lex_state = 2},
  [2045] = {.lex_state = 34, .external_lex_state = 2},
  [2046] = {.lex_state = 36, .external_lex_state = 2},
  [2047] = {.lex_state = 34, .external_lex_state = 2},
  [2048] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2049] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2050] = {.lex_state = 36, .external_lex_state = 2},
  [2051] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2052] = {.lex_state = 36, .external_lex_state = 2},
  [2053] = {.lex_state = 36, .external_lex_state = 2},
  [2054] = {.lex_state = 36, .external_lex_state = 2},
  [2055] = {.lex_state = 36, .external_lex_state = 2},
  [2056] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2057] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2058] = {.lex_state = 36, .external_lex_state = 2},
  [2059] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2060] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2061] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2062] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2063] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2064] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2065] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2066] = {.lex_state = 36, .external_lex_state = 2},
  [2067] = {.lex_state = 36, .external_lex_state = 2},
  [2068] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2069] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2070] = {.lex_state = 36, .external_lex_state = 2},
  [2071] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2072] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2073] = {.lex_state = 149, .external_lex_state = 3, .reserved_word_set_id = 54},
  [2074] = {.lex_state = 34, .external_lex_state = 2},
  [2075] = {.lex_state = 36, .external_lex_state = 2},
  [2076] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2077] = {.lex_state = 36, .external_lex_state = 2},
  [2078] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2079] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2080] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2081] = {.lex_state = 36, .external_lex_state = 2},
  [2082] = {.lex_state = 36, .external_lex_state = 2},
  [2083] = {.lex_state = 36, .external_lex_state = 2},
  [2084] = {.lex_state = 36, .external_lex_state = 2},
  [2085] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2086] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2087] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2088] = {.lex_state = 34, .external_lex_state = 2},
  [2089] = {.lex_state = 60, .external_lex_state = 3, .reserved_word_set_id = 32},
  [2090] = {.lex_state = 36, .external_lex_state = 2},
  [2091] = {.lex_state = 36, .external_lex_state = 2},
  [2092] = {.lex_state = 36, .external_lex_state = 2},
  [2093] = {.lex_state = 36, .external_lex_state = 2},
  [2094] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 55},
  [2095] = {.lex_state = 40, .external_lex_state = 2},
  [2096] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2097] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 56},
  [2098] = {.lex_state = 40, .external_lex_state = 2},
  [2099] = {.lex_state = 40, .external_lex_state = 2},
  [2100] = {.lex_state = 44, .external_lex_state = 2},
  [2101] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 57},
  [2102] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2103] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 56},
  [2104] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 56},
  [2105] = {.lex_state = 40, .external_lex_state = 2},
  [2106] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 57},
  [2107] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 56},
  [2108] = {.lex_state = 40, .external_lex_state = 2},
  [2109] = {.lex_state = 40, .external_lex_state = 2},
  [2110] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 57},
  [2111] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 57},
  [2112] = {.lex_state = 41, .external_lex_state = 2},
  [2113] = {.lex_state = 40, .external_lex_state = 2},
  [2114] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 56},
  [2115] = {.lex_state = 148, .external_lex_state = 2},
  [2116] = {.lex_state = 40, .external_lex_state = 2},
  [2117] = {.lex_state = 148, .external_lex_state = 2},
  [2118] = {.lex_state = 40, .external_lex_state = 2},
  [2119] = {.lex_state = 40, .external_lex_state = 2},
  [2120] = {.lex_state = 34, .external_lex_state = 2},
  [2121] = {.lex_state = 40, .external_lex_state = 2},
  [2122] = {.lex_state = 40, .external_lex_state = 2},
  [2123] = {.lex_state = 40, .external_lex_state = 2},
  [2124] = {.lex_state = 40, .external_lex_state = 2},
  [2125] = {.lex_state = 148, .external_lex_state = 2},
  [2126] = {.lex_state = 40, .external_lex_state = 2},
  [2127] = {.lex_state = 40, .external_lex_state = 2},
  [2128] = {.lex_state = 148, .external_lex_state = 2},
  [2129] = {.lex_state = 40, .external_lex_state = 2},
  [2130] = {.lex_state = 148, .external_lex_state = 2},
  [2131] = {.lex_state = 40, .external_lex_state = 2},
  [2132] = {.lex_state = 40, .external_lex_state = 2},
  [2133] = {.lex_state = 40, .external_lex_state = 2},
  [2134] = {.lex_state = 40, .external_lex_state = 2},
  [2135] = {.lex_state = 148, .external_lex_state = 2},
  [2136] = {.lex_state = 148, .external_lex_state = 2},
  [2137] = {.lex_state = 40, .external_lex_state = 2},
  [2138] = {.lex_state = 148, .external_lex_state = 2},
  [2139] = {.lex_state = 148, .external_lex_state = 2},
  [2140] = {.lex_state = 40, .external_lex_state = 2},
  [2141] = {.lex_state = 40, .external_lex_state = 2},
  [2142] = {.lex_state = 148, .external_lex_state = 2},
  [2143] = {.lex_state = 40, .external_lex_state = 2},
  [2144] = {.lex_state = 148, .external_lex_state = 2},
  [2145] = {.lex_state = 148, .external_lex_state = 2},
  [2146] = {.lex_state = 34, .external_lex_state = 2},
  [2147] = {.lex_state = 148, .external_lex_state = 2},
  [2148] = {.lex_state = 40, .external_lex_state = 2},
  [2149] = {.lex_state = 40, .external_lex_state = 2},
  [2150] = {.lex_state = 40, .external_lex_state = 2},
  [2151] = {.lex_state = 148, .external_lex_state = 2},
  [2152] = {.lex_state = 40, .external_lex_state = 2},
  [2153] = {.lex_state = 148, .external_lex_state = 2},
  [2154] = {.lex_state = 148, .external_lex_state = 2},
  [2155] = {.lex_state = 40, .external_lex_state = 2},
  [2156] = {.lex_state = 40, .external_lex_state = 2},
  [2157] = {.lex_state = 148, .external_lex_state = 2},
  [2158] = {.lex_state = 40, .external_lex_state = 2},
  [2159] = {.lex_state = 40, .external_lex_state = 2},
  [2160] = {.lex_state = 148, .external_lex_state = 2},
  [2161] = {.lex_state = 40, .external_lex_state = 2},
  [2162] = {.lex_state = 148, .external_lex_state = 2},
  [2163] = {.lex_state = 148, .external_lex_state = 2},
  [2164] = {.lex_state = 40, .external_lex_state = 2},
  [2165] = {.lex_state = 40, .external_lex_state = 2},
  [2166] = {.lex_state = 40, .external_lex_state = 2},
  [2167] = {.lex_state = 40, .external_lex_state = 2},
  [2168] = {.lex_state = 40, .external_lex_state = 2},
  [2169] = {.lex_state = 148, .external_lex_state = 2},
  [2170] = {.lex_state = 40, .external_lex_state = 2},
  [2171] = {.lex_state = 148, .external_lex_state = 2},
  [2172] = {.lex_state = 148, .external_lex_state = 2},
  [2173] = {.lex_state = 40, .external_lex_state = 2},
  [2174] = {.lex_state = 40, .external_lex_state = 2},
  [2175] = {.lex_state = 40, .external_lex_state = 2},
  [2176] = {.lex_state = 43, .external_lex_state = 2},
  [2177] = {.lex_state = 148, .external_lex_state = 2},
  [2178] = {.lex_state = 34, .external_lex_state = 2},
  [2179] = {.lex_state = 40, .external_lex_state = 2},
  [2180] = {.lex_state = 148, .external_lex_state = 2},
  [2181] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 50},
  [2182] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2183] = {.lex_state = 40, .external_lex_state = 2},
  [2184] = {.lex_state = 36, .external_lex_state = 2},
  [2185] = {.lex_state = 36, .external_lex_state = 2},
  [2186] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2187] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2188] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2189] = {.lex_state = 36, .external_lex_state = 2},
  [2190] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 59},
  [2191] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2192] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 59},
  [2193] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2194] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2195] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 50},
  [2196] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2197] = {.lex_state = 36, .external_lex_state = 2},
  [2198] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 50},
  [2199] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2200] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2201] = {.lex_state = 36, .external_lex_state = 2},
  [2202] = {.lex_state = 36, .external_lex_state = 2},
  [2203] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2204] = {.lex_state = 36, .external_lex_state = 2},
  [2205] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 50},
  [2206] = {.lex_state = 36, .external_lex_state = 2},
  [2207] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2208] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2209] = {.lex_state = 36, .external_lex_state = 2},
  [2210] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2211] = {.lex_state = 36, .external_lex_state = 2},
  [2212] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2213] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2214] = {.lex_state = 36, .external_lex_state = 2},
  [2215] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2216] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 59},
  [2217] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2218] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2219] = {.lex_state = 36, .external_lex_state = 2},
  [2220] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2221] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2222] = {.lex_state = 36, .external_lex_state = 2},
  [2223] = {.lex_state = 36, .external_lex_state = 2},
  [2224] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2225] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2226] = {.lex_state = 36, .external_lex_state = 2},
  [2227] = {.lex_state = 36, .external_lex_state = 2},
  [2228] = {.lex_state = 36, .external_lex_state = 2},
  [2229] = {.lex_state = 36, .external_lex_state = 2},
  [2230] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2231] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 50},
  [2232] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2233] = {.lex_state = 36, .external_lex_state = 2},
  [2234] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2235] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2236] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2237] = {.lex_state = 36, .external_lex_state = 2},
  [2238] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2239] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2240] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2241] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 50},
  [2242] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2243] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2244] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2245] = {.lex_state = 36, .external_lex_state = 2},
  [2246] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 50},
  [2247] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2248] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2249] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 50},
  [2250] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2251] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2252] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2253] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2254] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2255] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2256] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2257] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2258] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2259] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2260] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2261] = {.lex_state = 36, .external_lex_state = 2},
  [2262] = {.lex_state = 36, .external_lex_state = 2},
  [2263] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2264] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2265] = {.lex_state = 36, .external_lex_state = 2},
  [2266] = {.lex_state = 36, .external_lex_state = 2},
  [2267] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2268] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2269] = {.lex_state = 36, .external_lex_state = 2},
  [2270] = {.lex_state = 36, .external_lex_state = 2},
  [2271] = {.lex_state = 36, .external_lex_state = 2},
  [2272] = {.lex_state = 36, .external_lex_state = 2},
  [2273] = {.lex_state = 36, .external_lex_state = 2},
  [2274] = {.lex_state = 36, .external_lex_state = 2},
  [2275] = {.lex_state = 36, .external_lex_state = 2},
  [2276] = {.lex_state = 36, .external_lex_state = 2},
  [2277] = {.lex_state = 36, .external_lex_state = 2},
  [2278] = {.lex_state = 36, .external_lex_state = 2},
  [2279] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 59},
  [2280] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2281] = {.lex_state = 36, .external_lex_state = 2},
  [2282] = {.lex_state = 36, .external_lex_state = 2},
  [2283] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 50},
  [2284] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 50},
  [2285] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2286] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2287] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2288] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2289] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2290] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2291] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 50},
  [2292] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2293] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2294] = {.lex_state = 36, .external_lex_state = 2},
  [2295] = {.lex_state = 36, .external_lex_state = 2},
  [2296] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2297] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2298] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2299] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 58},
  [2300] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 50},
  [2301] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 50},
  [2302] = {.lex_state = 34, .external_lex_state = 2},
  [2303] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 60},
  [2304] = {.lex_state = 34, .external_lex_state = 2},
  [2305] = {.lex_state = 34, .external_lex_state = 2},
  [2306] = {.lex_state = 34, .external_lex_state = 2},
  [2307] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2308] = {.lex_state = 34, .external_lex_state = 2},
  [2309] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 60},
  [2310] = {.lex_state = 34, .external_lex_state = 2},
  [2311] = {.lex_state = 36, .external_lex_state = 2},
  [2312] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 60},
  [2313] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 60},
  [2314] = {.lex_state = 34, .external_lex_state = 2},
  [2315] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2316] = {.lex_state = 34, .external_lex_state = 2},
  [2317] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2318] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2319] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2320] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2321] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2322] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 61},
  [2323] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2324] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2325] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2326] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2327] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2328] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2329] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2330] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2331] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2332] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2333] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2334] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2335] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2336] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2337] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2338] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2339] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2340] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2341] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2342] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2343] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2344] = {.lex_state = 69, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2345] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2346] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2347] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2348] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2349] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2350] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2351] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2352] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2353] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2354] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2355] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2356] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2357] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2358] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2359] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2360] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2361] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2362] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2363] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2364] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2365] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2366] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2367] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2368] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2369] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2370] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2371] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2372] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2373] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2374] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2375] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2376] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2377] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2378] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2379] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2380] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2381] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2382] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2383] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2384] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2385] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2386] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2387] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2388] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2389] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2390] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2391] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2392] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2393] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2394] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2395] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2396] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2397] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2398] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2399] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2400] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2401] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2402] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2403] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2404] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2405] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2406] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2407] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2408] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2409] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2410] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2411] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2412] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2413] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2414] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2415] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2416] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2417] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2418] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2419] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2420] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2421] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2422] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2423] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 34},
  [2424] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2425] = {.lex_state = 47, .external_lex_state = 2},
  [2426] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2427] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2428] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2429] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2430] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2431] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2432] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2433] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2434] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2435] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2436] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2437] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2438] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2439] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2440] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2441] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 62},
  [2442] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2443] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2444] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2445] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2446] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2447] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2448] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2449] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2450] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2451] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2452] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2453] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2454] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2455] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2456] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2457] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2458] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2459] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2460] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2461] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2462] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2463] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2464] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2465] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2466] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2467] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2468] = {.lex_state = 47, .external_lex_state = 2},
  [2469] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2470] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2471] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2472] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2473] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2474] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2475] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2476] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2477] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2478] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2479] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2480] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2481] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2482] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2483] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2484] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 62},
  [2485] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 62},
  [2486] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2487] = {.lex_state = 47, .external_lex_state = 2},
  [2488] = {.lex_state = 55, .external_lex_state = 3, .reserved_word_set_id = 37},
  [2489] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 61},
  [2490] = {.lex_state = 48, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2491] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2492] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2493] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2494] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2495] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2496] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2497] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2498] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2499] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2500] = {.lex_state = 148, .external_lex_state = 2},
  [2501] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2502] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2503] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2504] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2505] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2506] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2507] = {.lex_state = 148, .external_lex_state = 2},
  [2508] = {.lex_state = 148, .external_lex_state = 2},
  [2509] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2510] = {.lex_state = 148, .external_lex_state = 2},
  [2511] = {.lex_state = 148, .external_lex_state = 2},
  [2512] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2513] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 62},
  [2514] = {.lex_state = 148, .external_lex_state = 2},
  [2515] = {.lex_state = 148, .external_lex_state = 2},
  [2516] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2517] = {.lex_state = 148, .external_lex_state = 2},
  [2518] = {.lex_state = 148, .external_lex_state = 2},
  [2519] = {.lex_state = 148, .external_lex_state = 2},
  [2520] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2521] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2522] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 62},
  [2523] = {.lex_state = 148, .external_lex_state = 2},
  [2524] = {.lex_state = 148, .external_lex_state = 2},
  [2525] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2526] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2527] = {.lex_state = 148, .external_lex_state = 2},
  [2528] = {.lex_state = 148, .external_lex_state = 2},
  [2529] = {.lex_state = 148, .external_lex_state = 2},
  [2530] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 62},
  [2531] = {.lex_state = 148, .external_lex_state = 2},
  [2532] = {.lex_state = 148, .external_lex_state = 2},
  [2533] = {.lex_state = 148, .external_lex_state = 2},
  [2534] = {.lex_state = 148, .external_lex_state = 2},
  [2535] = {.lex_state = 148, .external_lex_state = 2},
  [2536] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2537] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2538] = {.lex_state = 148, .external_lex_state = 2},
  [2539] = {.lex_state = 148, .external_lex_state = 2},
  [2540] = {.lex_state = 148, .external_lex_state = 2},
  [2541] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2542] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 64},
  [2543] = {.lex_state = 148, .external_lex_state = 2},
  [2544] = {.lex_state = 148, .external_lex_state = 2},
  [2545] = {.lex_state = 148, .external_lex_state = 2},
  [2546] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2547] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2548] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2549] = {.lex_state = 148, .external_lex_state = 2},
  [2550] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2551] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 62},
  [2552] = {.lex_state = 50, .external_lex_state = 3, .reserved_word_set_id = 35},
  [2553] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 62},
  [2554] = {.lex_state = 148, .external_lex_state = 2},
  [2555] = {.lex_state = 67, .external_lex_state = 3, .reserved_word_set_id = 39},
  [2556] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 62},
  [2557] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 39},
  [2558] = {.lex_state = 148, .external_lex_state = 2},
  [2559] = {.lex_state = 148, .external_lex_state = 2},
  [2560] = {.lex_state = 148, .external_lex_state = 2},
  [2561] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 62},
  [2562] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 62},
  [2563] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2564] = {.lex_state = 148, .external_lex_state = 2},
  [2565] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 39},
  [2566] = {.lex_state = 148, .external_lex_state = 2},
  [2567] = {.lex_state = 148, .external_lex_state = 2},
  [2568] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2569] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2570] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2571] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2572] = {.lex_state = 34, .external_lex_state = 2},
  [2573] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2574] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2575] = {.lex_state = 34, .external_lex_state = 2},
  [2576] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2577] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2578] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2579] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2580] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2581] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2582] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2583] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2584] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2585] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2586] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2587] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 65},
  [2588] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2589] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2590] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2591] = {.lex_state = 34, .external_lex_state = 2},
  [2592] = {.lex_state = 34, .external_lex_state = 2},
  [2593] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2594] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2595] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2596] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 65},
  [2597] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 39},
  [2598] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 39},
  [2599] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 39},
  [2600] = {.lex_state = 34, .external_lex_state = 2},
  [2601] = {.lex_state = 54, .external_lex_state = 3, .reserved_word_set_id = 39},
  [2602] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2603] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2604] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2605] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2606] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 65},
  [2607] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 65},
  [2608] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2609] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2610] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 65},
  [2611] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2612] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2613] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2614] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2615] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2616] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2617] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2618] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2619] = {.lex_state = 148, .external_lex_state = 2},
  [2620] = {.lex_state = 148, .external_lex_state = 2},
  [2621] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2622] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2623] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2624] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2625] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2626] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2627] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2628] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 65},
  [2629] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2630] = {.lex_state = 148, .external_lex_state = 2},
  [2631] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2632] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 63},
  [2633] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2634] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 66},
  [2635] = {.lex_state = 57, .external_lex_state = 3, .reserved_word_set_id = 43},
  [2636] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 25},
  [2637] = {.lex_state = 147, .external_lex_state = 2},
  [2638] = {.lex_state = 147, .external_lex_state = 2},
  [2639] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2640] = {.lex_state = 148, .external_lex_state = 2},
  [2641] = {.lex_state = 147, .external_lex_state = 2},
  [2642] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2643] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2644] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2645] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2646] = {.lex_state = 51, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2647] = {.lex_state = 148, .external_lex_state = 2},
  [2648] = {.lex_state = 147, .external_lex_state = 2},
  [2649] = {.lex_state = 148, .external_lex_state = 2},
  [2650] = {.lex_state = 147, .external_lex_state = 2},
  [2651] = {.lex_state = 147, .external_lex_state = 2},
  [2652] = {.lex_state = 147, .external_lex_state = 2},
  [2653] = {.lex_state = 37, .external_lex_state = 2},
  [2654] = {.lex_state = 37, .external_lex_state = 2},
  [2655] = {.lex_state = 37, .external_lex_state = 2},
  [2656] = {.lex_state = 37, .external_lex_state = 2},
  [2657] = {.lex_state = 37, .external_lex_state = 2},
  [2658] = {.lex_state = 37, .external_lex_state = 2},
  [2659] = {.lex_state = 37, .external_lex_state = 2},
  [2660] = {.lex_state = 37, .external_lex_state = 2},
  [2661] = {.lex_state = 37, .external_lex_state = 2},
  [2662] = {.lex_state = 37, .external_lex_state = 2},
  [2663] = {.lex_state = 37, .external_lex_state = 2},
  [2664] = {.lex_state = 37, .external_lex_state = 2},
  [2665] = {.lex_state = 37, .external_lex_state = 2},
  [2666] = {.lex_state = 37, .external_lex_state = 2},
  [2667] = {.lex_state = 37, .external_lex_state = 2},
  [2668] = {.lex_state = 37, .external_lex_state = 2},
  [2669] = {.lex_state = 37, .external_lex_state = 2},
  [2670] = {.lex_state = 67, .external_lex_state = 3, .reserved_word_set_id = 31},
  [2671] = {.lex_state = 67, .external_lex_state = 3, .reserved_word_set_id = 31},
  [2672] = {.lex_state = 147, .external_lex_state = 2},
  [2673] = {.lex_state = 37, .external_lex_state = 2},
  [2674] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 67},
  [2675] = {.lex_state = 37, .external_lex_state = 2},
  [2676] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 67},
  [2677] = {.lex_state = 37, .external_lex_state = 2},
  [2678] = {.lex_state = 37, .external_lex_state = 2},
  [2679] = {.lex_state = 37, .external_lex_state = 2},
  [2680] = {.lex_state = 37, .external_lex_state = 2},
  [2681] = {.lex_state = 37, .external_lex_state = 2},
  [2682] = {.lex_state = 37, .external_lex_state = 2},
  [2683] = {.lex_state = 37, .external_lex_state = 2},
  [2684] = {.lex_state = 37, .external_lex_state = 2},
  [2685] = {.lex_state = 37, .external_lex_state = 2},
  [2686] = {.lex_state = 37, .external_lex_state = 2},
  [2687] = {.lex_state = 37, .external_lex_state = 2},
  [2688] = {.lex_state = 37, .external_lex_state = 2},
  [2689] = {.lex_state = 37, .external_lex_state = 2},
  [2690] = {.lex_state = 37, .external_lex_state = 2},
  [2691] = {.lex_state = 37, .external_lex_state = 2},
  [2692] = {.lex_state = 37, .external_lex_state = 2},
  [2693] = {.lex_state = 37, .external_lex_state = 2},
  [2694] = {.lex_state = 37, .external_lex_state = 2},
  [2695] = {.lex_state = 37, .external_lex_state = 2},
  [2696] = {.lex_state = 37, .external_lex_state = 2},
  [2697] = {.lex_state = 37, .external_lex_state = 2},
  [2698] = {.lex_state = 37, .external_lex_state = 2},
  [2699] = {.lex_state = 37, .external_lex_state = 2},
  [2700] = {.lex_state = 37, .external_lex_state = 2},
  [2701] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 68},
  [2702] = {.lex_state = 37, .external_lex_state = 2},
  [2703] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 68},
  [2704] = {.lex_state = 37, .external_lex_state = 2},
  [2705] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 69},
  [2706] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 69},
  [2707] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 69},
  [2708] = {.lex_state = 37, .external_lex_state = 2},
  [2709] = {.lex_state = 37, .external_lex_state = 2},
  [2710] = {.lex_state = 37, .external_lex_state = 2},
  [2711] = {.lex_state = 37, .external_lex_state = 2},
  [2712] = {.lex_state = 37, .external_lex_state = 2},
  [2713] = {.lex_state = 37, .external_lex_state = 2},
  [2714] = {.lex_state = 37, .external_lex_state = 2},
  [2715] = {.lex_state = 37, .external_lex_state = 2},
  [2716] = {.lex_state = 37, .external_lex_state = 2},
  [2717] = {.lex_state = 37, .external_lex_state = 2},
  [2718] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 68},
  [2719] = {.lex_state = 37, .external_lex_state = 2},
  [2720] = {.lex_state = 37, .external_lex_state = 2},
  [2721] = {.lex_state = 37, .external_lex_state = 2},
  [2722] = {.lex_state = 148, .external_lex_state = 2},
  [2723] = {.lex_state = 37, .external_lex_state = 2},
  [2724] = {.lex_state = 37, .external_lex_state = 2},
  [2725] = {.lex_state = 67, .external_lex_state = 3, .reserved_word_set_id = 31},
  [2726] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2727] = {.lex_state = 37, .external_lex_state = 2},
  [2728] = {.lex_state = 37, .external_lex_state = 2},
  [2729] = {.lex_state = 37, .external_lex_state = 2},
  [2730] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 68},
  [2731] = {.lex_state = 37, .external_lex_state = 2},
  [2732] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 68},
  [2733] = {.lex_state = 37, .external_lex_state = 2},
  [2734] = {.lex_state = 147, .external_lex_state = 2},
  [2735] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 69},
  [2736] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 69},
  [2737] = {.lex_state = 37, .external_lex_state = 2},
  [2738] = {.lex_state = 37, .external_lex_state = 2},
  [2739] = {.lex_state = 148, .external_lex_state = 2},
  [2740] = {.lex_state = 37, .external_lex_state = 2},
  [2741] = {.lex_state = 37, .external_lex_state = 2},
  [2742] = {.lex_state = 37, .external_lex_state = 2},
  [2743] = {.lex_state = 37, .external_lex_state = 2},
  [2744] = {.lex_state = 37, .external_lex_state = 2},
  [2745] = {.lex_state = 37, .external_lex_state = 2},
  [2746] = {.lex_state = 37, .external_lex_state = 2},
  [2747] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 68},
  [2748] = {.lex_state = 148, .external_lex_state = 2},
  [2749] = {.lex_state = 37, .external_lex_state = 2},
  [2750] = {.lex_state = 37, .external_lex_state = 2},
  [2751] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 69},
  [2752] = {.lex_state = 37, .external_lex_state = 2},
  [2753] = {.lex_state = 37, .external_lex_state = 2},
  [2754] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 69},
  [2755] = {.lex_state = 37, .external_lex_state = 2},
  [2756] = {.lex_state = 37, .external_lex_state = 2},
  [2757] = {.lex_state = 148, .external_lex_state = 2},
  [2758] = {.lex_state = 37, .external_lex_state = 2},
  [2759] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 69},
  [2760] = {.lex_state = 37, .external_lex_state = 2},
  [2761] = {.lex_state = 37, .external_lex_state = 2},
  [2762] = {.lex_state = 37, .external_lex_state = 2},
  [2763] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2764] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2765] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2766] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2767] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2768] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 67},
  [2769] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 67},
  [2770] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 67},
  [2771] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2772] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 67},
  [2773] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2774] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2775] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2776] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2777] = {.lex_state = 67, .external_lex_state = 3, .reserved_word_set_id = 31},
  [2778] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2779] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2780] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2781] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2782] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2783] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2784] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2785] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2786] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2787] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2788] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2789] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2790] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2791] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2792] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2793] = {.lex_state = 37, .external_lex_state = 2},
  [2794] = {.lex_state = 37, .external_lex_state = 2},
  [2795] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2796] = {.lex_state = 37, .external_lex_state = 2},
  [2797] = {.lex_state = 37, .external_lex_state = 2},
  [2798] = {.lex_state = 37, .external_lex_state = 2},
  [2799] = {.lex_state = 37, .external_lex_state = 2},
  [2800] = {.lex_state = 37, .external_lex_state = 2},
  [2801] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2802] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2803] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2804] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2805] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2806] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2807] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2808] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2809] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2810] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2811] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2812] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2813] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2814] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2815] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2816] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2817] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2818] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2819] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2820] = {.lex_state = 37, .external_lex_state = 2},
  [2821] = {.lex_state = 37, .external_lex_state = 2},
  [2822] = {.lex_state = 37, .external_lex_state = 2},
  [2823] = {.lex_state = 37, .external_lex_state = 2},
  [2824] = {.lex_state = 37, .external_lex_state = 2},
  [2825] = {.lex_state = 37, .external_lex_state = 2},
  [2826] = {.lex_state = 37, .external_lex_state = 2},
  [2827] = {.lex_state = 37, .external_lex_state = 2},
  [2828] = {.lex_state = 37, .external_lex_state = 2},
  [2829] = {.lex_state = 37, .external_lex_state = 2},
  [2830] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2831] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2832] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2833] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2834] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2835] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2836] = {.lex_state = 148, .external_lex_state = 2},
  [2837] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2838] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2839] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2840] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2841] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2842] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2843] = {.lex_state = 148, .external_lex_state = 2},
  [2844] = {.lex_state = 47, .external_lex_state = 2},
  [2845] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2846] = {.lex_state = 37, .external_lex_state = 2},
  [2847] = {.lex_state = 36, .external_lex_state = 2},
  [2848] = {.lex_state = 36, .external_lex_state = 2},
  [2849] = {.lex_state = 36, .external_lex_state = 2},
  [2850] = {.lex_state = 47, .external_lex_state = 2},
  [2851] = {.lex_state = 47, .external_lex_state = 2},
  [2852] = {.lex_state = 47, .external_lex_state = 2},
  [2853] = {.lex_state = 47, .external_lex_state = 2},
  [2854] = {.lex_state = 47, .external_lex_state = 2},
  [2855] = {.lex_state = 36, .external_lex_state = 2},
  [2856] = {.lex_state = 36, .external_lex_state = 2},
  [2857] = {.lex_state = 36, .external_lex_state = 2},
  [2858] = {.lex_state = 47, .external_lex_state = 2},
  [2859] = {.lex_state = 47, .external_lex_state = 2},
  [2860] = {.lex_state = 47, .external_lex_state = 2},
  [2861] = {.lex_state = 47, .external_lex_state = 2},
  [2862] = {.lex_state = 47, .external_lex_state = 2},
  [2863] = {.lex_state = 149, .external_lex_state = 3, .reserved_word_set_id = 31},
  [2864] = {.lex_state = 8, .external_lex_state = 2, .reserved_word_set_id = 71},
  [2865] = {.lex_state = 36, .external_lex_state = 2},
  [2866] = {.lex_state = 36, .external_lex_state = 2},
  [2867] = {.lex_state = 36, .external_lex_state = 2},
  [2868] = {.lex_state = 36, .external_lex_state = 2},
  [2869] = {.lex_state = 36, .external_lex_state = 2},
  [2870] = {.lex_state = 36, .external_lex_state = 2},
  [2871] = {.lex_state = 36, .external_lex_state = 2},
  [2872] = {.lex_state = 36, .external_lex_state = 2},
  [2873] = {.lex_state = 36, .external_lex_state = 2},
  [2874] = {.lex_state = 76, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2875] = {.lex_state = 8, .external_lex_state = 2, .reserved_word_set_id = 71},
  [2876] = {.lex_state = 36, .external_lex_state = 2},
  [2877] = {.lex_state = 47, .external_lex_state = 2},
  [2878] = {.lex_state = 47, .external_lex_state = 2},
  [2879] = {.lex_state = 47, .external_lex_state = 2},
  [2880] = {.lex_state = 47, .external_lex_state = 2},
  [2881] = {.lex_state = 47, .external_lex_state = 2},
  [2882] = {.lex_state = 47, .external_lex_state = 2},
  [2883] = {.lex_state = 8, .external_lex_state = 2, .reserved_word_set_id = 71},
  [2884] = {.lex_state = 36, .external_lex_state = 2},
  [2885] = {.lex_state = 37, .external_lex_state = 2},
  [2886] = {.lex_state = 37, .external_lex_state = 2},
  [2887] = {.lex_state = 148, .external_lex_state = 2},
  [2888] = {.lex_state = 37, .external_lex_state = 2},
  [2889] = {.lex_state = 37, .external_lex_state = 2},
  [2890] = {.lex_state = 37, .external_lex_state = 2},
  [2891] = {.lex_state = 47, .external_lex_state = 2},
  [2892] = {.lex_state = 37, .external_lex_state = 2},
  [2893] = {.lex_state = 37, .external_lex_state = 2},
  [2894] = {.lex_state = 37, .external_lex_state = 2},
  [2895] = {.lex_state = 37, .external_lex_state = 2},
  [2896] = {.lex_state = 37, .external_lex_state = 2},
  [2897] = {.lex_state = 148, .external_lex_state = 2},
  [2898] = {.lex_state = 148, .external_lex_state = 2},
  [2899] = {.lex_state = 47, .external_lex_state = 2},
  [2900] = {.lex_state = 148, .external_lex_state = 2},
  [2901] = {.lex_state = 148, .external_lex_state = 2},
  [2902] = {.lex_state = 148, .external_lex_state = 2},
  [2903] = {.lex_state = 37, .external_lex_state = 2},
  [2904] = {.lex_state = 148, .external_lex_state = 2},
  [2905] = {.lex_state = 37, .external_lex_state = 2},
  [2906] = {.lex_state = 37, .external_lex_state = 2},
  [2907] = {.lex_state = 37, .external_lex_state = 2},
  [2908] = {.lex_state = 37, .external_lex_state = 2},
  [2909] = {.lex_state = 33, .external_lex_state = 2},
  [2910] = {.lex_state = 148, .external_lex_state = 2},
  [2911] = {.lex_state = 37, .external_lex_state = 2},
  [2912] = {.lex_state = 148, .external_lex_state = 2},
  [2913] = {.lex_state = 37, .external_lex_state = 2},
  [2914] = {.lex_state = 37, .external_lex_state = 2},
  [2915] = {.lex_state = 148, .external_lex_state = 2},
  [2916] = {.lex_state = 37, .external_lex_state = 2},
  [2917] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2918] = {.lex_state = 37, .external_lex_state = 2},
  [2919] = {.lex_state = 148, .external_lex_state = 2},
  [2920] = {.lex_state = 37, .external_lex_state = 2},
  [2921] = {.lex_state = 148, .external_lex_state = 2},
  [2922] = {.lex_state = 37, .external_lex_state = 2},
  [2923] = {.lex_state = 37, .external_lex_state = 2},
  [2924] = {.lex_state = 59, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2925] = {.lex_state = 65, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2926] = {.lex_state = 37, .external_lex_state = 2},
  [2927] = {.lex_state = 59, .external_lex_state = 3, .reserved_word_set_id = 72},
  [2928] = {.lex_state = 59, .external_lex_state = 3, .reserved_word_set_id = 72},
  [2929] = {.lex_state = 59, .external_lex_state = 3, .reserved_word_set_id = 72},
  [2930] = {.lex_state = 37, .external_lex_state = 2},
  [2931] = {.lex_state = 33, .external_lex_state = 2},
  [2932] = {.lex_state = 148, .external_lex_state = 2},
  [2933] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2934] = {.lex_state = 59, .external_lex_state = 3, .reserved_word_set_id = 72},
  [2935] = {.lex_state = 37, .external_lex_state = 2},
  [2936] = {.lex_state = 37, .external_lex_state = 2},
  [2937] = {.lex_state = 47, .external_lex_state = 2},
  [2938] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2939] = {.lex_state = 37, .external_lex_state = 2},
  [2940] = {.lex_state = 59, .external_lex_state = 3, .reserved_word_set_id = 72},
  [2941] = {.lex_state = 59, .external_lex_state = 3, .reserved_word_set_id = 72},
  [2942] = {.lex_state = 59, .external_lex_state = 3, .reserved_word_set_id = 72},
  [2943] = {.lex_state = 59, .external_lex_state = 3, .reserved_word_set_id = 72},
  [2944] = {.lex_state = 148, .external_lex_state = 2},
  [2945] = {.lex_state = 148, .external_lex_state = 2},
  [2946] = {.lex_state = 148, .external_lex_state = 2},
  [2947] = {.lex_state = 37, .external_lex_state = 2},
  [2948] = {.lex_state = 33, .external_lex_state = 2},
  [2949] = {.lex_state = 59, .external_lex_state = 3, .reserved_word_set_id = 72},
  [2950] = {.lex_state = 33, .external_lex_state = 2},
  [2951] = {.lex_state = 8, .external_lex_state = 2, .reserved_word_set_id = 71},
  [2952] = {.lex_state = 33, .external_lex_state = 2},
  [2953] = {.lex_state = 37, .external_lex_state = 2},
  [2954] = {.lex_state = 148, .external_lex_state = 2},
  [2955] = {.lex_state = 37, .external_lex_state = 2},
  [2956] = {.lex_state = 37, .external_lex_state = 2},
  [2957] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2958] = {.lex_state = 8, .external_lex_state = 2, .reserved_word_set_id = 71},
  [2959] = {.lex_state = 6, .external_lex_state = 2},
  [2960] = {.lex_state = 148, .external_lex_state = 2},
  [2961] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2962] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2963] = {.lex_state = 147, .external_lex_state = 2},
  [2964] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2965] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2966] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2967] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2968] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2969] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2970] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2971] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2972] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2973] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2974] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2975] = {.lex_state = 23, .external_lex_state = 2},
  [2976] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2977] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2978] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2979] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2980] = {.lex_state = 148, .external_lex_state = 2},
  [2981] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2982] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2983] = {.lex_state = 36, .external_lex_state = 2},
  [2984] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2985] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2986] = {.lex_state = 147, .external_lex_state = 2},
  [2987] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2988] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2989] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2990] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2991] = {.lex_state = 6, .external_lex_state = 2},
  [2992] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2993] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2994] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2995] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2996] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [2997] = {.lex_state = 147, .external_lex_state = 2},
  [2998] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [2999] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3000] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3001] = {.lex_state = 148, .external_lex_state = 2},
  [3002] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3003] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3004] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3005] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3006] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3007] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3008] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3009] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3010] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3011] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3012] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3013] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3014] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3015] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3016] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3017] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3018] = {.lex_state = 148, .external_lex_state = 2},
  [3019] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3020] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3021] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3022] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3023] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3024] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3025] = {.lex_state = 6, .external_lex_state = 2},
  [3026] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3027] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3028] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3029] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3030] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3031] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3032] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3033] = {.lex_state = 148, .external_lex_state = 2},
  [3034] = {.lex_state = 148, .external_lex_state = 2},
  [3035] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3036] = {.lex_state = 148, .external_lex_state = 2},
  [3037] = {.lex_state = 148, .external_lex_state = 2},
  [3038] = {.lex_state = 148, .external_lex_state = 2},
  [3039] = {.lex_state = 148, .external_lex_state = 2},
  [3040] = {.lex_state = 148, .external_lex_state = 2},
  [3041] = {.lex_state = 148, .external_lex_state = 2},
  [3042] = {.lex_state = 148, .external_lex_state = 2},
  [3043] = {.lex_state = 148, .external_lex_state = 2},
  [3044] = {.lex_state = 148, .external_lex_state = 2},
  [3045] = {.lex_state = 148, .external_lex_state = 2},
  [3046] = {.lex_state = 148, .external_lex_state = 2},
  [3047] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3048] = {.lex_state = 148, .external_lex_state = 2},
  [3049] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3050] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3051] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3052] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3053] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3054] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3055] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3056] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3057] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3058] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3059] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3060] = {.lex_state = 147, .external_lex_state = 2},
  [3061] = {.lex_state = 36, .external_lex_state = 2},
  [3062] = {.lex_state = 148, .external_lex_state = 2},
  [3063] = {.lex_state = 36, .external_lex_state = 2},
  [3064] = {.lex_state = 36, .external_lex_state = 2},
  [3065] = {.lex_state = 36, .external_lex_state = 2},
  [3066] = {.lex_state = 36, .external_lex_state = 2},
  [3067] = {.lex_state = 36, .external_lex_state = 2},
  [3068] = {.lex_state = 36, .external_lex_state = 2},
  [3069] = {.lex_state = 36, .external_lex_state = 2},
  [3070] = {.lex_state = 36, .external_lex_state = 2},
  [3071] = {.lex_state = 148, .external_lex_state = 2},
  [3072] = {.lex_state = 36, .external_lex_state = 2},
  [3073] = {.lex_state = 36, .external_lex_state = 2},
  [3074] = {.lex_state = 36, .external_lex_state = 2},
  [3075] = {.lex_state = 148, .external_lex_state = 2},
  [3076] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3077] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3078] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3079] = {.lex_state = 147, .external_lex_state = 2},
  [3080] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3081] = {.lex_state = 23, .external_lex_state = 2},
  [3082] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3083] = {.lex_state = 23, .external_lex_state = 2},
  [3084] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3085] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3086] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3087] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3088] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3089] = {.lex_state = 23, .external_lex_state = 2},
  [3090] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3091] = {.lex_state = 6, .external_lex_state = 2},
  [3092] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3093] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3094] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3095] = {.lex_state = 23, .external_lex_state = 2},
  [3096] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3097] = {.lex_state = 23, .external_lex_state = 2},
  [3098] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3099] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3100] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3101] = {.lex_state = 36, .external_lex_state = 2},
  [3102] = {.lex_state = 36, .external_lex_state = 2},
  [3103] = {.lex_state = 36, .external_lex_state = 2},
  [3104] = {.lex_state = 36, .external_lex_state = 2},
  [3105] = {.lex_state = 147, .external_lex_state = 2},
  [3106] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3107] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3108] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3109] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 70},
  [3110] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3111] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3112] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3113] = {.lex_state = 61, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3114] = {.lex_state = 148, .external_lex_state = 2},
  [3115] = {.lex_state = 148, .external_lex_state = 2},
  [3116] = {.lex_state = 37, .external_lex_state = 2},
  [3117] = {.lex_state = 37, .external_lex_state = 2},
  [3118] = {.lex_state = 37, .external_lex_state = 2},
  [3119] = {.lex_state = 147, .external_lex_state = 2},
  [3120] = {.lex_state = 34, .external_lex_state = 2},
  [3121] = {.lex_state = 147, .external_lex_state = 2},
  [3122] = {.lex_state = 147, .external_lex_state = 2},
  [3123] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3124] = {.lex_state = 147, .external_lex_state = 2},
  [3125] = {.lex_state = 76, .external_lex_state = 2, .reserved_word_set_id = 73},
  [3126] = {.lex_state = 34, .external_lex_state = 2},
  [3127] = {.lex_state = 147, .external_lex_state = 2},
  [3128] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 74},
  [3129] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3130] = {.lex_state = 37, .external_lex_state = 2},
  [3131] = {.lex_state = 147, .external_lex_state = 2},
  [3132] = {.lex_state = 148, .external_lex_state = 2},
  [3133] = {.lex_state = 37, .external_lex_state = 2},
  [3134] = {.lex_state = 37, .external_lex_state = 2},
  [3135] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 74},
  [3136] = {.lex_state = 147, .external_lex_state = 2},
  [3137] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 74},
  [3138] = {.lex_state = 147, .external_lex_state = 2},
  [3139] = {.lex_state = 6, .external_lex_state = 2},
  [3140] = {.lex_state = 147, .external_lex_state = 2},
  [3141] = {.lex_state = 148, .external_lex_state = 2},
  [3142] = {.lex_state = 148, .external_lex_state = 2},
  [3143] = {.lex_state = 34, .external_lex_state = 2},
  [3144] = {.lex_state = 148, .external_lex_state = 2},
  [3145] = {.lex_state = 148, .external_lex_state = 2},
  [3146] = {.lex_state = 148, .external_lex_state = 2},
  [3147] = {.lex_state = 148, .external_lex_state = 2},
  [3148] = {.lex_state = 148, .external_lex_state = 2},
  [3149] = {.lex_state = 148, .external_lex_state = 2},
  [3150] = {.lex_state = 148, .external_lex_state = 2},
  [3151] = {.lex_state = 147, .external_lex_state = 2},
  [3152] = {.lex_state = 148, .external_lex_state = 2},
  [3153] = {.lex_state = 148, .external_lex_state = 2},
  [3154] = {.lex_state = 148, .external_lex_state = 2},
  [3155] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3156] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 75},
  [3157] = {.lex_state = 147, .external_lex_state = 2},
  [3158] = {.lex_state = 34, .external_lex_state = 2},
  [3159] = {.lex_state = 148, .external_lex_state = 2},
  [3160] = {.lex_state = 148, .external_lex_state = 2},
  [3161] = {.lex_state = 148, .external_lex_state = 2},
  [3162] = {.lex_state = 147, .external_lex_state = 2},
  [3163] = {.lex_state = 6, .external_lex_state = 2},
  [3164] = {.lex_state = 147, .external_lex_state = 2},
  [3165] = {.lex_state = 148, .external_lex_state = 2},
  [3166] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 74},
  [3167] = {.lex_state = 147, .external_lex_state = 2},
  [3168] = {.lex_state = 37, .external_lex_state = 2},
  [3169] = {.lex_state = 34, .external_lex_state = 2},
  [3170] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 75},
  [3171] = {.lex_state = 148, .external_lex_state = 2},
  [3172] = {.lex_state = 147, .external_lex_state = 2},
  [3173] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3174] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3175] = {.lex_state = 148, .external_lex_state = 2},
  [3176] = {.lex_state = 147, .external_lex_state = 2},
  [3177] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3178] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3179] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 75},
  [3180] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3181] = {.lex_state = 75, .external_lex_state = 2, .reserved_word_set_id = 76},
  [3182] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3183] = {.lex_state = 148, .external_lex_state = 2},
  [3184] = {.lex_state = 36, .external_lex_state = 2},
  [3185] = {.lex_state = 36, .external_lex_state = 2},
  [3186] = {.lex_state = 47, .external_lex_state = 2},
  [3187] = {.lex_state = 36, .external_lex_state = 2},
  [3188] = {.lex_state = 36, .external_lex_state = 2},
  [3189] = {.lex_state = 36, .external_lex_state = 2},
  [3190] = {.lex_state = 36, .external_lex_state = 2},
  [3191] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3192] = {.lex_state = 148, .external_lex_state = 2},
  [3193] = {.lex_state = 57, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3194] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3195] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3196] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3197] = {.lex_state = 147, .external_lex_state = 2},
  [3198] = {.lex_state = 37, .external_lex_state = 2},
  [3199] = {.lex_state = 147, .external_lex_state = 2},
  [3200] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3201] = {.lex_state = 147, .external_lex_state = 2},
  [3202] = {.lex_state = 147, .external_lex_state = 2},
  [3203] = {.lex_state = 47, .external_lex_state = 2},
  [3204] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3205] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3206] = {.lex_state = 148, .external_lex_state = 2},
  [3207] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3208] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3209] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 77},
  [3210] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 77},
  [3211] = {.lex_state = 148, .external_lex_state = 2},
  [3212] = {.lex_state = 148, .external_lex_state = 2},
  [3213] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3214] = {.lex_state = 148, .external_lex_state = 2},
  [3215] = {.lex_state = 148, .external_lex_state = 2},
  [3216] = {.lex_state = 57, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3217] = {.lex_state = 148, .external_lex_state = 2},
  [3218] = {.lex_state = 36, .external_lex_state = 2},
  [3219] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3220] = {.lex_state = 147, .external_lex_state = 2},
  [3221] = {.lex_state = 148, .external_lex_state = 2},
  [3222] = {.lex_state = 148, .external_lex_state = 2},
  [3223] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3224] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 77},
  [3225] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3226] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 77},
  [3227] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 77},
  [3228] = {.lex_state = 147, .external_lex_state = 2},
  [3229] = {.lex_state = 57, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3230] = {.lex_state = 57, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3231] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 77},
  [3232] = {.lex_state = 57, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3233] = {.lex_state = 147, .external_lex_state = 2},
  [3234] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3235] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3236] = {.lex_state = 148, .external_lex_state = 2},
  [3237] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3238] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3239] = {.lex_state = 147, .external_lex_state = 2},
  [3240] = {.lex_state = 147, .external_lex_state = 2},
  [3241] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3242] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3243] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3244] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3245] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3246] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3247] = {.lex_state = 147, .external_lex_state = 2},
  [3248] = {.lex_state = 147, .external_lex_state = 2},
  [3249] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 78},
  [3250] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3251] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 77},
  [3252] = {.lex_state = 147, .external_lex_state = 2},
  [3253] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3254] = {.lex_state = 147, .external_lex_state = 2},
  [3255] = {.lex_state = 47, .external_lex_state = 2},
  [3256] = {.lex_state = 147, .external_lex_state = 2},
  [3257] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3258] = {.lex_state = 147, .external_lex_state = 2},
  [3259] = {.lex_state = 36, .external_lex_state = 2},
  [3260] = {.lex_state = 147, .external_lex_state = 2},
  [3261] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 78},
  [3262] = {.lex_state = 147, .external_lex_state = 2},
  [3263] = {.lex_state = 147, .external_lex_state = 2},
  [3264] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 75},
  [3265] = {.lex_state = 57, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3266] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3267] = {.lex_state = 148, .external_lex_state = 2},
  [3268] = {.lex_state = 6, .external_lex_state = 2},
  [3269] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3270] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3271] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3272] = {.lex_state = 36, .external_lex_state = 2},
  [3273] = {.lex_state = 148, .external_lex_state = 2},
  [3274] = {.lex_state = 147, .external_lex_state = 2},
  [3275] = {.lex_state = 147, .external_lex_state = 2},
  [3276] = {.lex_state = 147, .external_lex_state = 2},
  [3277] = {.lex_state = 36, .external_lex_state = 2},
  [3278] = {.lex_state = 36, .external_lex_state = 2},
  [3279] = {.lex_state = 36, .external_lex_state = 2},
  [3280] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 75},
  [3281] = {.lex_state = 23, .external_lex_state = 3, .reserved_word_set_id = 66},
  [3282] = {.lex_state = 148, .external_lex_state = 2},
  [3283] = {.lex_state = 36, .external_lex_state = 2},
  [3284] = {.lex_state = 148, .external_lex_state = 2},
  [3285] = {.lex_state = 147, .external_lex_state = 2},
  [3286] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 75},
  [3287] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3288] = {.lex_state = 147, .external_lex_state = 2},
  [3289] = {.lex_state = 147, .external_lex_state = 2},
  [3290] = {.lex_state = 147, .external_lex_state = 2},
  [3291] = {.lex_state = 147, .external_lex_state = 2},
  [3292] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3293] = {.lex_state = 36, .external_lex_state = 2},
  [3294] = {.lex_state = 147, .external_lex_state = 2},
  [3295] = {.lex_state = 147, .external_lex_state = 2},
  [3296] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 79},
  [3297] = {.lex_state = 148, .external_lex_state = 2},
  [3298] = {.lex_state = 71, .external_lex_state = 2},
  [3299] = {.lex_state = 71, .external_lex_state = 2},
  [3300] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 80},
  [3301] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 78},
  [3302] = {.lex_state = 71, .external_lex_state = 2},
  [3303] = {.lex_state = 147, .external_lex_state = 2},
  [3304] = {.lex_state = 71, .external_lex_state = 2},
  [3305] = {.lex_state = 148, .external_lex_state = 2},
  [3306] = {.lex_state = 148, .external_lex_state = 2},
  [3307] = {.lex_state = 148, .external_lex_state = 2},
  [3308] = {.lex_state = 148, .external_lex_state = 2},
  [3309] = {.lex_state = 147, .external_lex_state = 2},
  [3310] = {.lex_state = 147, .external_lex_state = 2},
  [3311] = {.lex_state = 36, .external_lex_state = 2},
  [3312] = {.lex_state = 71, .external_lex_state = 2},
  [3313] = {.lex_state = 147, .external_lex_state = 2},
  [3314] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3315] = {.lex_state = 151, .external_lex_state = 2},
  [3316] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3317] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3318] = {.lex_state = 23, .external_lex_state = 2},
  [3319] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3320] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3321] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3322] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3323] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3324] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3325] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3326] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 80},
  [3327] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3328] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3329] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3330] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3331] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3332] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3333] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3334] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3335] = {.lex_state = 23, .external_lex_state = 2},
  [3336] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3337] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3338] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3339] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3340] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3341] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3342] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3343] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3344] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3345] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3346] = {.lex_state = 34, .external_lex_state = 3},
  [3347] = {.lex_state = 55, .external_lex_state = 2, .reserved_word_set_id = 80},
  [3348] = {.lex_state = 36, .external_lex_state = 2},
  [3349] = {.lex_state = 147, .external_lex_state = 2},
  [3350] = {.lex_state = 147, .external_lex_state = 2},
  [3351] = {.lex_state = 148, .external_lex_state = 2},
  [3352] = {.lex_state = 147, .external_lex_state = 2},
  [3353] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3354] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3355] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3356] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3357] = {.lex_state = 147, .external_lex_state = 2},
  [3358] = {.lex_state = 147, .external_lex_state = 2},
  [3359] = {.lex_state = 147, .external_lex_state = 2},
  [3360] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3361] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3362] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3363] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3364] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3365] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3366] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3367] = {.lex_state = 147, .external_lex_state = 2},
  [3368] = {.lex_state = 151, .external_lex_state = 2},
  [3369] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3370] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3371] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3372] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3373] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3374] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3375] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3376] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3377] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3378] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3379] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3380] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3381] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3382] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3383] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3384] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3385] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3386] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3387] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3388] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3389] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3390] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3391] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3392] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3393] = {.lex_state = 147, .external_lex_state = 2},
  [3394] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3395] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3396] = {.lex_state = 34, .external_lex_state = 3},
  [3397] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3398] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3399] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3400] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3401] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3402] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3403] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3404] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3405] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3406] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3407] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3408] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3409] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3410] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3411] = {.lex_state = 148, .external_lex_state = 2},
  [3412] = {.lex_state = 148, .external_lex_state = 2},
  [3413] = {.lex_state = 148, .external_lex_state = 2},
  [3414] = {.lex_state = 148, .external_lex_state = 2},
  [3415] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3416] = {.lex_state = 148, .external_lex_state = 2},
  [3417] = {.lex_state = 147, .external_lex_state = 2},
  [3418] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3419] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 84},
  [3420] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3421] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3422] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3423] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3424] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3425] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3426] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3427] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3428] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3429] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3430] = {.lex_state = 23, .external_lex_state = 2},
  [3431] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3432] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 81},
  [3433] = {.lex_state = 151, .external_lex_state = 2},
  [3434] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3435] = {.lex_state = 148, .external_lex_state = 2},
  [3436] = {.lex_state = 147, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3437] = {.lex_state = 23, .external_lex_state = 2},
  [3438] = {.lex_state = 23, .external_lex_state = 2},
  [3439] = {.lex_state = 147, .external_lex_state = 2},
  [3440] = {.lex_state = 147, .external_lex_state = 2},
  [3441] = {.lex_state = 147, .external_lex_state = 2},
  [3442] = {.lex_state = 148, .external_lex_state = 2},
  [3443] = {.lex_state = 23, .external_lex_state = 2},
  [3444] = {.lex_state = 23, .external_lex_state = 2},
  [3445] = {.lex_state = 151, .external_lex_state = 2},
  [3446] = {.lex_state = 147, .external_lex_state = 2},
  [3447] = {.lex_state = 147, .external_lex_state = 2},
  [3448] = {.lex_state = 23, .external_lex_state = 2},
  [3449] = {.lex_state = 147, .external_lex_state = 2},
  [3450] = {.lex_state = 147, .external_lex_state = 2},
  [3451] = {.lex_state = 23, .external_lex_state = 2},
  [3452] = {.lex_state = 23, .external_lex_state = 2},
  [3453] = {.lex_state = 23, .external_lex_state = 2},
  [3454] = {.lex_state = 147, .external_lex_state = 2},
  [3455] = {.lex_state = 23, .external_lex_state = 2},
  [3456] = {.lex_state = 23, .external_lex_state = 2},
  [3457] = {.lex_state = 151, .external_lex_state = 2},
  [3458] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 78},
  [3459] = {.lex_state = 23, .external_lex_state = 2},
  [3460] = {.lex_state = 23, .external_lex_state = 2},
  [3461] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3462] = {.lex_state = 147, .external_lex_state = 2},
  [3463] = {.lex_state = 147, .external_lex_state = 2},
  [3464] = {.lex_state = 23, .external_lex_state = 2},
  [3465] = {.lex_state = 147, .external_lex_state = 2},
  [3466] = {.lex_state = 23, .external_lex_state = 2},
  [3467] = {.lex_state = 147, .external_lex_state = 2},
  [3468] = {.lex_state = 147, .external_lex_state = 2},
  [3469] = {.lex_state = 23, .external_lex_state = 2},
  [3470] = {.lex_state = 23, .external_lex_state = 2},
  [3471] = {.lex_state = 147, .external_lex_state = 2},
  [3472] = {.lex_state = 147, .external_lex_state = 2},
  [3473] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 82},
  [3474] = {.lex_state = 23, .external_lex_state = 2},
  [3475] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 78},
  [3476] = {.lex_state = 151, .external_lex_state = 2},
  [3477] = {.lex_state = 147, .external_lex_state = 2},
  [3478] = {.lex_state = 151, .external_lex_state = 2},
  [3479] = {.lex_state = 151, .external_lex_state = 2},
  [3480] = {.lex_state = 147, .external_lex_state = 2},
  [3481] = {.lex_state = 147, .external_lex_state = 2},
  [3482] = {.lex_state = 23, .external_lex_state = 2},
  [3483] = {.lex_state = 151, .external_lex_state = 2},
  [3484] = {.lex_state = 23, .external_lex_state = 2},
  [3485] = {.lex_state = 22, .external_lex_state = 2},
  [3486] = {.lex_state = 23, .external_lex_state = 2},
  [3487] = {.lex_state = 22, .external_lex_state = 2},
  [3488] = {.lex_state = 22, .external_lex_state = 2},
  [3489] = {.lex_state = 22, .external_lex_state = 2},
  [3490] = {.lex_state = 22, .external_lex_state = 2},
  [3491] = {.lex_state = 22, .external_lex_state = 2},
  [3492] = {.lex_state = 22, .external_lex_state = 2},
  [3493] = {.lex_state = 34, .external_lex_state = 3},
  [3494] = {.lex_state = 23, .external_lex_state = 2},
  [3495] = {.lex_state = 147, .external_lex_state = 2},
  [3496] = {.lex_state = 77, .external_lex_state = 2},
  [3497] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3498] = {.lex_state = 22, .external_lex_state = 2},
  [3499] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3500] = {.lex_state = 147, .external_lex_state = 2},
  [3501] = {.lex_state = 77, .external_lex_state = 2},
  [3502] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3503] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3504] = {.lex_state = 151, .external_lex_state = 2},
  [3505] = {.lex_state = 151, .external_lex_state = 2},
  [3506] = {.lex_state = 147, .external_lex_state = 2},
  [3507] = {.lex_state = 22, .external_lex_state = 2},
  [3508] = {.lex_state = 22, .external_lex_state = 2},
  [3509] = {.lex_state = 22, .external_lex_state = 2},
  [3510] = {.lex_state = 22, .external_lex_state = 2},
  [3511] = {.lex_state = 22, .external_lex_state = 2},
  [3512] = {.lex_state = 22, .external_lex_state = 2},
  [3513] = {.lex_state = 22, .external_lex_state = 2},
  [3514] = {.lex_state = 22, .external_lex_state = 2},
  [3515] = {.lex_state = 147, .external_lex_state = 2},
  [3516] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3517] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3518] = {.lex_state = 147, .external_lex_state = 2},
  [3519] = {.lex_state = 151, .external_lex_state = 2},
  [3520] = {.lex_state = 151, .external_lex_state = 2},
  [3521] = {.lex_state = 151, .external_lex_state = 2},
  [3522] = {.lex_state = 22, .external_lex_state = 2},
  [3523] = {.lex_state = 22, .external_lex_state = 2},
  [3524] = {.lex_state = 147, .external_lex_state = 2},
  [3525] = {.lex_state = 147, .external_lex_state = 2},
  [3526] = {.lex_state = 147, .external_lex_state = 2},
  [3527] = {.lex_state = 147, .external_lex_state = 2},
  [3528] = {.lex_state = 151, .external_lex_state = 2},
  [3529] = {.lex_state = 151, .external_lex_state = 2},
  [3530] = {.lex_state = 6, .external_lex_state = 2},
  [3531] = {.lex_state = 6, .external_lex_state = 2},
  [3532] = {.lex_state = 147, .external_lex_state = 2},
  [3533] = {.lex_state = 147, .external_lex_state = 2},
  [3534] = {.lex_state = 147, .external_lex_state = 2},
  [3535] = {.lex_state = 147, .external_lex_state = 2},
  [3536] = {.lex_state = 147, .external_lex_state = 2},
  [3537] = {.lex_state = 151, .external_lex_state = 2},
  [3538] = {.lex_state = 147, .external_lex_state = 2},
  [3539] = {.lex_state = 29, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3540] = {.lex_state = 147, .external_lex_state = 2},
  [3541] = {.lex_state = 6, .external_lex_state = 2},
  [3542] = {.lex_state = 29, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3543] = {.lex_state = 77, .external_lex_state = 2},
  [3544] = {.lex_state = 147, .external_lex_state = 2},
  [3545] = {.lex_state = 6, .external_lex_state = 2},
  [3546] = {.lex_state = 77, .external_lex_state = 2},
  [3547] = {.lex_state = 147, .external_lex_state = 2},
  [3548] = {.lex_state = 147, .external_lex_state = 2},
  [3549] = {.lex_state = 147, .external_lex_state = 2},
  [3550] = {.lex_state = 147, .external_lex_state = 2},
  [3551] = {.lex_state = 147, .external_lex_state = 2},
  [3552] = {.lex_state = 147, .external_lex_state = 2},
  [3553] = {.lex_state = 147, .external_lex_state = 2},
  [3554] = {.lex_state = 6, .external_lex_state = 2},
  [3555] = {.lex_state = 29, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3556] = {.lex_state = 147, .external_lex_state = 2},
  [3557] = {.lex_state = 151, .external_lex_state = 2},
  [3558] = {.lex_state = 6, .external_lex_state = 2},
  [3559] = {.lex_state = 6, .external_lex_state = 2},
  [3560] = {.lex_state = 6, .external_lex_state = 2},
  [3561] = {.lex_state = 77, .external_lex_state = 2},
  [3562] = {.lex_state = 77, .external_lex_state = 2},
  [3563] = {.lex_state = 6, .external_lex_state = 2},
  [3564] = {.lex_state = 147, .external_lex_state = 2},
  [3565] = {.lex_state = 6, .external_lex_state = 2},
  [3566] = {.lex_state = 6, .external_lex_state = 2},
  [3567] = {.lex_state = 6, .external_lex_state = 3},
  [3568] = {.lex_state = 147, .external_lex_state = 2},
  [3569] = {.lex_state = 6, .external_lex_state = 3},
  [3570] = {.lex_state = 71, .external_lex_state = 2},
  [3571] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 86},
  [3572] = {.lex_state = 147, .external_lex_state = 2},
  [3573] = {.lex_state = 23, .external_lex_state = 2},
  [3574] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3575] = {.lex_state = 23, .external_lex_state = 2},
  [3576] = {.lex_state = 23, .external_lex_state = 2},
  [3577] = {.lex_state = 23, .external_lex_state = 2},
  [3578] = {.lex_state = 23, .external_lex_state = 2},
  [3579] = {.lex_state = 23, .external_lex_state = 2},
  [3580] = {.lex_state = 23, .external_lex_state = 2},
  [3581] = {.lex_state = 23, .external_lex_state = 2},
  [3582] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3583] = {.lex_state = 23, .external_lex_state = 2},
  [3584] = {.lex_state = 23, .external_lex_state = 2},
  [3585] = {.lex_state = 147, .external_lex_state = 2},
  [3586] = {.lex_state = 22, .external_lex_state = 2},
  [3587] = {.lex_state = 23, .external_lex_state = 2},
  [3588] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3589] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3590] = {.lex_state = 22, .external_lex_state = 2},
  [3591] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3592] = {.lex_state = 72, .external_lex_state = 2},
  [3593] = {.lex_state = 147, .external_lex_state = 2},
  [3594] = {.lex_state = 23, .external_lex_state = 2},
  [3595] = {.lex_state = 23, .external_lex_state = 2},
  [3596] = {.lex_state = 23, .external_lex_state = 2},
  [3597] = {.lex_state = 23, .external_lex_state = 2},
  [3598] = {.lex_state = 23, .external_lex_state = 2},
  [3599] = {.lex_state = 23, .external_lex_state = 2},
  [3600] = {.lex_state = 151, .external_lex_state = 2},
  [3601] = {.lex_state = 23, .external_lex_state = 2},
  [3602] = {.lex_state = 23, .external_lex_state = 2},
  [3603] = {.lex_state = 147, .external_lex_state = 2},
  [3604] = {.lex_state = 23, .external_lex_state = 2},
  [3605] = {.lex_state = 22, .external_lex_state = 2},
  [3606] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3607] = {.lex_state = 23, .external_lex_state = 2},
  [3608] = {.lex_state = 23, .external_lex_state = 2},
  [3609] = {.lex_state = 23, .external_lex_state = 2},
  [3610] = {.lex_state = 23, .external_lex_state = 2},
  [3611] = {.lex_state = 23, .external_lex_state = 2},
  [3612] = {.lex_state = 34, .external_lex_state = 2},
  [3613] = {.lex_state = 23, .external_lex_state = 2},
  [3614] = {.lex_state = 147, .external_lex_state = 2},
  [3615] = {.lex_state = 23, .external_lex_state = 2},
  [3616] = {.lex_state = 23, .external_lex_state = 2},
  [3617] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3618] = {.lex_state = 23, .external_lex_state = 2},
  [3619] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3620] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3621] = {.lex_state = 23, .external_lex_state = 2},
  [3622] = {.lex_state = 23, .external_lex_state = 2},
  [3623] = {.lex_state = 22, .external_lex_state = 2},
  [3624] = {.lex_state = 23, .external_lex_state = 2},
  [3625] = {.lex_state = 23, .external_lex_state = 2},
  [3626] = {.lex_state = 23, .external_lex_state = 2},
  [3627] = {.lex_state = 23, .external_lex_state = 2},
  [3628] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3629] = {.lex_state = 23, .external_lex_state = 2},
  [3630] = {.lex_state = 23, .external_lex_state = 2},
  [3631] = {.lex_state = 22, .external_lex_state = 2},
  [3632] = {.lex_state = 23, .external_lex_state = 2},
  [3633] = {.lex_state = 23, .external_lex_state = 2},
  [3634] = {.lex_state = 29, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3635] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3636] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3637] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3638] = {.lex_state = 147, .external_lex_state = 2},
  [3639] = {.lex_state = 23, .external_lex_state = 2},
  [3640] = {.lex_state = 23, .external_lex_state = 2},
  [3641] = {.lex_state = 23, .external_lex_state = 2},
  [3642] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3643] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3644] = {.lex_state = 23, .external_lex_state = 2},
  [3645] = {.lex_state = 23, .external_lex_state = 2},
  [3646] = {.lex_state = 22, .external_lex_state = 2},
  [3647] = {.lex_state = 23, .external_lex_state = 2},
  [3648] = {.lex_state = 23, .external_lex_state = 2},
  [3649] = {.lex_state = 23, .external_lex_state = 2},
  [3650] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3651] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3652] = {.lex_state = 34, .external_lex_state = 2},
  [3653] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3654] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3655] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3656] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3657] = {.lex_state = 151, .external_lex_state = 2},
  [3658] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3659] = {.lex_state = 23, .external_lex_state = 2},
  [3660] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3661] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3662] = {.lex_state = 151, .external_lex_state = 2},
  [3663] = {.lex_state = 151, .external_lex_state = 2},
  [3664] = {.lex_state = 151, .external_lex_state = 2},
  [3665] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3666] = {.lex_state = 151, .external_lex_state = 2},
  [3667] = {.lex_state = 151, .external_lex_state = 2},
  [3668] = {.lex_state = 23, .external_lex_state = 2},
  [3669] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3670] = {.lex_state = 22, .external_lex_state = 2},
  [3671] = {.lex_state = 23, .external_lex_state = 2},
  [3672] = {.lex_state = 151, .external_lex_state = 2},
  [3673] = {.lex_state = 151, .external_lex_state = 2},
  [3674] = {.lex_state = 151, .external_lex_state = 2},
  [3675] = {.lex_state = 151, .external_lex_state = 2},
  [3676] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3677] = {.lex_state = 23, .external_lex_state = 2},
  [3678] = {.lex_state = 151, .external_lex_state = 2},
  [3679] = {.lex_state = 151, .external_lex_state = 2},
  [3680] = {.lex_state = 151, .external_lex_state = 2},
  [3681] = {.lex_state = 151, .external_lex_state = 2},
  [3682] = {.lex_state = 151, .external_lex_state = 2},
  [3683] = {.lex_state = 151, .external_lex_state = 2},
  [3684] = {.lex_state = 151, .external_lex_state = 2},
  [3685] = {.lex_state = 151, .external_lex_state = 2},
  [3686] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3687] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3688] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3689] = {.lex_state = 23, .external_lex_state = 2},
  [3690] = {.lex_state = 23, .external_lex_state = 2},
  [3691] = {.lex_state = 23, .external_lex_state = 2},
  [3692] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3693] = {.lex_state = 23, .external_lex_state = 2},
  [3694] = {.lex_state = 22, .external_lex_state = 2},
  [3695] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3696] = {.lex_state = 23, .external_lex_state = 2},
  [3697] = {.lex_state = 23, .external_lex_state = 2},
  [3698] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3699] = {.lex_state = 23, .external_lex_state = 2},
  [3700] = {.lex_state = 23, .external_lex_state = 2},
  [3701] = {.lex_state = 22, .external_lex_state = 2},
  [3702] = {.lex_state = 23, .external_lex_state = 2},
  [3703] = {.lex_state = 23, .external_lex_state = 2},
  [3704] = {.lex_state = 23, .external_lex_state = 2},
  [3705] = {.lex_state = 23, .external_lex_state = 2},
  [3706] = {.lex_state = 72, .external_lex_state = 2},
  [3707] = {.lex_state = 23, .external_lex_state = 2},
  [3708] = {.lex_state = 23, .external_lex_state = 2},
  [3709] = {.lex_state = 23, .external_lex_state = 2},
  [3710] = {.lex_state = 23, .external_lex_state = 2},
  [3711] = {.lex_state = 23, .external_lex_state = 2},
  [3712] = {.lex_state = 23, .external_lex_state = 2},
  [3713] = {.lex_state = 23, .external_lex_state = 2},
  [3714] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3715] = {.lex_state = 23, .external_lex_state = 2},
  [3716] = {.lex_state = 23, .external_lex_state = 2},
  [3717] = {.lex_state = 23, .external_lex_state = 2},
  [3718] = {.lex_state = 23, .external_lex_state = 2},
  [3719] = {.lex_state = 23, .external_lex_state = 2},
  [3720] = {.lex_state = 23, .external_lex_state = 2},
  [3721] = {.lex_state = 23, .external_lex_state = 2},
  [3722] = {.lex_state = 72, .external_lex_state = 2},
  [3723] = {.lex_state = 23, .external_lex_state = 2},
  [3724] = {.lex_state = 23, .external_lex_state = 2},
  [3725] = {.lex_state = 23, .external_lex_state = 2},
  [3726] = {.lex_state = 23, .external_lex_state = 2},
  [3727] = {.lex_state = 23, .external_lex_state = 2},
  [3728] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3729] = {.lex_state = 23, .external_lex_state = 2},
  [3730] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3731] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3732] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3733] = {.lex_state = 23, .external_lex_state = 2},
  [3734] = {.lex_state = 23, .external_lex_state = 2},
  [3735] = {.lex_state = 23, .external_lex_state = 2},
  [3736] = {.lex_state = 23, .external_lex_state = 2},
  [3737] = {.lex_state = 23, .external_lex_state = 2},
  [3738] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3739] = {.lex_state = 23, .external_lex_state = 2},
  [3740] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3741] = {.lex_state = 23, .external_lex_state = 2},
  [3742] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3743] = {.lex_state = 23, .external_lex_state = 2},
  [3744] = {.lex_state = 23, .external_lex_state = 2},
  [3745] = {.lex_state = 23, .external_lex_state = 2},
  [3746] = {.lex_state = 23, .external_lex_state = 2},
  [3747] = {.lex_state = 23, .external_lex_state = 2},
  [3748] = {.lex_state = 23, .external_lex_state = 2},
  [3749] = {.lex_state = 23, .external_lex_state = 2},
  [3750] = {.lex_state = 23, .external_lex_state = 2},
  [3751] = {.lex_state = 23, .external_lex_state = 2},
  [3752] = {.lex_state = 23, .external_lex_state = 2},
  [3753] = {.lex_state = 23, .external_lex_state = 2},
  [3754] = {.lex_state = 23, .external_lex_state = 2},
  [3755] = {.lex_state = 23, .external_lex_state = 2},
  [3756] = {.lex_state = 23, .external_lex_state = 2},
  [3757] = {.lex_state = 147, .external_lex_state = 2},
  [3758] = {.lex_state = 22, .external_lex_state = 2},
  [3759] = {.lex_state = 23, .external_lex_state = 2},
  [3760] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3761] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3762] = {.lex_state = 151, .external_lex_state = 2},
  [3763] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3764] = {.lex_state = 151, .external_lex_state = 2},
  [3765] = {.lex_state = 23, .external_lex_state = 2},
  [3766] = {.lex_state = 23, .external_lex_state = 2},
  [3767] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3768] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3769] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3770] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3771] = {.lex_state = 151, .external_lex_state = 2},
  [3772] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3773] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3774] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3775] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 85},
  [3776] = {.lex_state = 23, .external_lex_state = 2},
  [3777] = {.lex_state = 148, .external_lex_state = 2},
  [3778] = {.lex_state = 22, .external_lex_state = 2},
  [3779] = {.lex_state = 148, .external_lex_state = 2},
  [3780] = {.lex_state = 34, .external_lex_state = 2},
  [3781] = {.lex_state = 34, .external_lex_state = 2},
  [3782] = {.lex_state = 148, .external_lex_state = 2},
  [3783] = {.lex_state = 34, .external_lex_state = 2},
  [3784] = {.lex_state = 13, .external_lex_state = 3, .reserved_word_set_id = 1},
  [3785] = {.lex_state = 34, .external_lex_state = 2},
  [3786] = {.lex_state = 147, .external_lex_state = 2},
  [3787] = {.lex_state = 34, .external_lex_state = 2},
  [3788] = {.lex_state = 6, .external_lex_state = 2},
  [3789] = {.lex_state = 34, .external_lex_state = 2},
  [3790] = {.lex_state = 34, .external_lex_state = 2},
  [3791] = {.lex_state = 22, .external_lex_state = 2},
  [3792] = {.lex_state = 151, .external_lex_state = 2},
  [3793] = {.lex_state = 151, .external_lex_state = 2},
  [3794] = {.lex_state = 22, .external_lex_state = 2},
  [3795] = {.lex_state = 22, .external_lex_state = 2},
  [3796] = {.lex_state = 22, .external_lex_state = 2},
  [3797] = {.lex_state = 34, .external_lex_state = 2},
  [3798] = {.lex_state = 148, .external_lex_state = 2},
  [3799] = {.lex_state = 151, .external_lex_state = 2},
  [3800] = {.lex_state = 22, .external_lex_state = 2},
  [3801] = {.lex_state = 34, .external_lex_state = 2},
  [3802] = {.lex_state = 34, .external_lex_state = 2},
  [3803] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 87},
  [3804] = {.lex_state = 34, .external_lex_state = 2},
  [3805] = {.lex_state = 34, .external_lex_state = 2},
  [3806] = {.lex_state = 34, .external_lex_state = 2},
  [3807] = {.lex_state = 151, .external_lex_state = 2},
  [3808] = {.lex_state = 148, .external_lex_state = 2},
  [3809] = {.lex_state = 34, .external_lex_state = 2},
  [3810] = {.lex_state = 34, .external_lex_state = 2},
  [3811] = {.lex_state = 6, .external_lex_state = 3},
  [3812] = {.lex_state = 34, .external_lex_state = 2},
  [3813] = {.lex_state = 34, .external_lex_state = 2},
  [3814] = {.lex_state = 34, .external_lex_state = 2},
  [3815] = {.lex_state = 34, .external_lex_state = 2},
  [3816] = {.lex_state = 148, .external_lex_state = 2},
  [3817] = {.lex_state = 148, .external_lex_state = 2},
  [3818] = {.lex_state = 34, .external_lex_state = 3},
  [3819] = {.lex_state = 34, .external_lex_state = 2},
  [3820] = {.lex_state = 23, .external_lex_state = 2},
  [3821] = {.lex_state = 151, .external_lex_state = 2},
  [3822] = {.lex_state = 34, .external_lex_state = 2},
  [3823] = {.lex_state = 34, .external_lex_state = 2},
  [3824] = {.lex_state = 34, .external_lex_state = 2},
  [3825] = {.lex_state = 148, .external_lex_state = 2},
  [3826] = {.lex_state = 34, .external_lex_state = 2},
  [3827] = {.lex_state = 34, .external_lex_state = 2},
  [3828] = {.lex_state = 34, .external_lex_state = 3},
  [3829] = {.lex_state = 34, .external_lex_state = 3},
  [3830] = {.lex_state = 34, .external_lex_state = 3},
  [3831] = {.lex_state = 34, .external_lex_state = 2},
  [3832] = {.lex_state = 22, .external_lex_state = 2},
  [3833] = {.lex_state = 151, .external_lex_state = 2},
  [3834] = {.lex_state = 22, .external_lex_state = 2},
  [3835] = {.lex_state = 151, .external_lex_state = 2},
  [3836] = {.lex_state = 148, .external_lex_state = 2},
  [3837] = {.lex_state = 34, .external_lex_state = 2},
  [3838] = {.lex_state = 23, .external_lex_state = 2},
  [3839] = {.lex_state = 23, .external_lex_state = 2},
  [3840] = {.lex_state = 23, .external_lex_state = 2},
  [3841] = {.lex_state = 147, .external_lex_state = 2},
  [3842] = {.lex_state = 34, .external_lex_state = 2},
  [3843] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3844] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 83},
  [3845] = {.lex_state = 6, .external_lex_state = 2},
  [3846] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 87},
  [3847] = {.lex_state = 148, .external_lex_state = 2},
  [3848] = {.lex_state = 148, .external_lex_state = 2},
  [3849] = {.lex_state = 34, .external_lex_state = 2},
  [3850] = {.lex_state = 6, .external_lex_state = 2},
  [3851] = {.lex_state = 6, .external_lex_state = 2},
  [3852] = {.lex_state = 148, .external_lex_state = 2},
  [3853] = {.lex_state = 148, .external_lex_state = 2},
  [3854] = {.lex_state = 148, .external_lex_state = 2},
  [3855] = {.lex_state = 148, .external_lex_state = 2},
  [3856] = {.lex_state = 6, .external_lex_state = 2},
  [3857] = {.lex_state = 23, .external_lex_state = 2},
  [3858] = {.lex_state = 23, .external_lex_state = 2},
  [3859] = {.lex_state = 148, .external_lex_state = 2},
  [3860] = {.lex_state = 6, .external_lex_state = 2},
  [3861] = {.lex_state = 148, .external_lex_state = 2},
  [3862] = {.lex_state = 151, .external_lex_state = 2},
  [3863] = {.lex_state = 148, .external_lex_state = 2},
  [3864] = {.lex_state = 148, .external_lex_state = 2},
  [3865] = {.lex_state = 148, .external_lex_state = 2},
  [3866] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 88},
  [3867] = {.lex_state = 148, .external_lex_state = 2},
  [3868] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 88},
  [3869] = {.lex_state = 148, .external_lex_state = 2},
  [3870] = {.lex_state = 151, .external_lex_state = 2},
  [3871] = {.lex_state = 148, .external_lex_state = 2},
  [3872] = {.lex_state = 148, .external_lex_state = 2},
  [3873] = {.lex_state = 148, .external_lex_state = 2},
  [3874] = {.lex_state = 151, .external_lex_state = 2},
  [3875] = {.lex_state = 151, .external_lex_state = 2},
  [3876] = {.lex_state = 148, .external_lex_state = 2},
  [3877] = {.lex_state = 148, .external_lex_state = 2},
  [3878] = {.lex_state = 148, .external_lex_state = 2},
  [3879] = {.lex_state = 148, .external_lex_state = 2},
  [3880] = {.lex_state = 148, .external_lex_state = 2},
  [3881] = {.lex_state = 151, .external_lex_state = 2},
  [3882] = {.lex_state = 148, .external_lex_state = 2},
  [3883] = {.lex_state = 148, .external_lex_state = 2},
  [3884] = {.lex_state = 148, .external_lex_state = 2},
  [3885] = {.lex_state = 151, .external_lex_state = 2},
  [3886] = {.lex_state = 148, .external_lex_state = 2},
  [3887] = {.lex_state = 148, .external_lex_state = 2},
  [3888] = {.lex_state = 148, .external_lex_state = 2},
  [3889] = {.lex_state = 148, .external_lex_state = 2},
  [3890] = {.lex_state = 23, .external_lex_state = 2},
  [3891] = {.lex_state = 148, .external_lex_state = 2},
  [3892] = {.lex_state = 151, .external_lex_state = 2},
  [3893] = {.lex_state = 151, .external_lex_state = 2},
  [3894] = {.lex_state = 148, .external_lex_state = 2},
  [3895] = {.lex_state = 148, .external_lex_state = 2},
  [3896] = {.lex_state = 34, .external_lex_state = 2},
  [3897] = {.lex_state = 34, .external_lex_state = 2},
  [3898] = {.lex_state = 151, .external_lex_state = 2},
  [3899] = {.lex_state = 151, .external_lex_state = 2},
  [3900] = {.lex_state = 23, .external_lex_state = 2},
  [3901] = {.lex_state = 23, .external_lex_state = 2},
  [3902] = {.lex_state = 34, .external_lex_state = 2},
  [3903] = {.lex_state = 151, .external_lex_state = 2},
  [3904] = {.lex_state = 151, .external_lex_state = 2},
  [3905] = {.lex_state = 148, .external_lex_state = 2},
  [3906] = {.lex_state = 151, .external_lex_state = 2},
  [3907] = {.lex_state = 23, .external_lex_state = 2},
  [3908] = {.lex_state = 23, .external_lex_state = 2},
  [3909] = {.lex_state = 6, .external_lex_state = 2},
  [3910] = {.lex_state = 148, .external_lex_state = 2},
  [3911] = {.lex_state = 23, .external_lex_state = 2},
  [3912] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 88},
  [3913] = {.lex_state = 148, .external_lex_state = 2},
  [3914] = {.lex_state = 148, .external_lex_state = 2},
  [3915] = {.lex_state = 6, .external_lex_state = 2},
  [3916] = {.lex_state = 148, .external_lex_state = 2},
  [3917] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 88},
  [3918] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [3919] = {.lex_state = 148, .external_lex_state = 2},
  [3920] = {.lex_state = 148, .external_lex_state = 2},
  [3921] = {.lex_state = 148, .external_lex_state = 2},
  [3922] = {.lex_state = 148, .external_lex_state = 2},
  [3923] = {.lex_state = 23, .external_lex_state = 2},
  [3924] = {.lex_state = 151, .external_lex_state = 2},
  [3925] = {.lex_state = 148, .external_lex_state = 2},
  [3926] = {.lex_state = 151, .external_lex_state = 2},
  [3927] = {.lex_state = 148, .external_lex_state = 2},
  [3928] = {.lex_state = 23, .external_lex_state = 2},
  [3929] = {.lex_state = 148, .external_lex_state = 2},
  [3930] = {.lex_state = 148, .external_lex_state = 2},
  [3931] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 88},
  [3932] = {.lex_state = 148, .external_lex_state = 2},
  [3933] = {.lex_state = 148, .external_lex_state = 2},
  [3934] = {.lex_state = 23, .external_lex_state = 2},
  [3935] = {.lex_state = 148, .external_lex_state = 2},
  [3936] = {.lex_state = 23, .external_lex_state = 2},
  [3937] = {.lex_state = 6, .external_lex_state = 2},
  [3938] = {.lex_state = 148, .external_lex_state = 2},
  [3939] = {.lex_state = 148, .external_lex_state = 2},
  [3940] = {.lex_state = 148, .external_lex_state = 2},
  [3941] = {.lex_state = 148, .external_lex_state = 2},
  [3942] = {.lex_state = 148, .external_lex_state = 2},
  [3943] = {.lex_state = 148, .external_lex_state = 2},
  [3944] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 88},
  [3945] = {.lex_state = 148, .external_lex_state = 2},
  [3946] = {.lex_state = 148, .external_lex_state = 2},
  [3947] = {.lex_state = 148, .external_lex_state = 2},
  [3948] = {.lex_state = 148, .external_lex_state = 2},
  [3949] = {.lex_state = 148, .external_lex_state = 2},
  [3950] = {.lex_state = 148, .external_lex_state = 2},
  [3951] = {.lex_state = 148, .external_lex_state = 2},
  [3952] = {.lex_state = 148, .external_lex_state = 2},
  [3953] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 88},
  [3954] = {.lex_state = 148, .external_lex_state = 2},
  [3955] = {.lex_state = 148, .external_lex_state = 2},
  [3956] = {.lex_state = 148, .external_lex_state = 2},
  [3957] = {.lex_state = 148, .external_lex_state = 2},
  [3958] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 88},
  [3959] = {.lex_state = 148, .external_lex_state = 2},
  [3960] = {.lex_state = 23, .external_lex_state = 2},
  [3961] = {.lex_state = 148, .external_lex_state = 2},
  [3962] = {.lex_state = 6, .external_lex_state = 2},
  [3963] = {.lex_state = 23, .external_lex_state = 2},
  [3964] = {.lex_state = 148, .external_lex_state = 2},
  [3965] = {.lex_state = 23, .external_lex_state = 2},
  [3966] = {.lex_state = 148, .external_lex_state = 2},
  [3967] = {.lex_state = 23, .external_lex_state = 2},
  [3968] = {.lex_state = 148, .external_lex_state = 2},
  [3969] = {.lex_state = 148, .external_lex_state = 2},
  [3970] = {.lex_state = 148, .external_lex_state = 2},
  [3971] = {.lex_state = 148, .external_lex_state = 2},
  [3972] = {.lex_state = 6, .external_lex_state = 2},
  [3973] = {.lex_state = 23, .external_lex_state = 2},
  [3974] = {.lex_state = 148, .external_lex_state = 2},
  [3975] = {.lex_state = 148, .external_lex_state = 2},
  [3976] = {.lex_state = 148, .external_lex_state = 2},
  [3977] = {.lex_state = 23, .external_lex_state = 2},
  [3978] = {.lex_state = 23, .external_lex_state = 2},
  [3979] = {.lex_state = 148, .external_lex_state = 2},
  [3980] = {.lex_state = 6, .external_lex_state = 2},
  [3981] = {.lex_state = 6, .external_lex_state = 2},
  [3982] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 88},
  [3983] = {.lex_state = 6, .external_lex_state = 2},
  [3984] = {.lex_state = 6, .external_lex_state = 2},
  [3985] = {.lex_state = 6, .external_lex_state = 2},
  [3986] = {.lex_state = 6, .external_lex_state = 2},
  [3987] = {.lex_state = 23, .external_lex_state = 2},
  [3988] = {.lex_state = 148, .external_lex_state = 2},
  [3989] = {.lex_state = 148, .external_lex_state = 2},
  [3990] = {.lex_state = 148, .external_lex_state = 2},
  [3991] = {.lex_state = 23, .external_lex_state = 2},
  [3992] = {.lex_state = 148, .external_lex_state = 2},
  [3993] = {.lex_state = 148, .external_lex_state = 2},
  [3994] = {.lex_state = 6, .external_lex_state = 2},
  [3995] = {.lex_state = 6, .external_lex_state = 2},
  [3996] = {.lex_state = 148, .external_lex_state = 2},
  [3997] = {.lex_state = 148, .external_lex_state = 2},
  [3998] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 88},
  [3999] = {.lex_state = 148, .external_lex_state = 2},
  [4000] = {.lex_state = 148, .external_lex_state = 2},
  [4001] = {.lex_state = 148, .external_lex_state = 2},
  [4002] = {.lex_state = 23, .external_lex_state = 2},
  [4003] = {.lex_state = 148, .external_lex_state = 2},
  [4004] = {.lex_state = 148, .external_lex_state = 2},
  [4005] = {.lex_state = 6, .external_lex_state = 2},
  [4006] = {.lex_state = 23, .external_lex_state = 2},
  [4007] = {.lex_state = 148, .external_lex_state = 2},
  [4008] = {.lex_state = 148, .external_lex_state = 2},
  [4009] = {.lex_state = 148, .external_lex_state = 2},
  [4010] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 83},
  [4011] = {.lex_state = 148, .external_lex_state = 2},
  [4012] = {.lex_state = 148, .external_lex_state = 2},
  [4013] = {.lex_state = 148, .external_lex_state = 2},
  [4014] = {.lex_state = 148, .external_lex_state = 2},
  [4015] = {.lex_state = 148, .external_lex_state = 2},
  [4016] = {.lex_state = 148, .external_lex_state = 2},
  [4017] = {.lex_state = 148, .external_lex_state = 2},
  [4018] = {.lex_state = 148, .external_lex_state = 2},
  [4019] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 88},
  [4020] = {.lex_state = 148, .external_lex_state = 2},
  [4021] = {.lex_state = 23, .external_lex_state = 2},
  [4022] = {.lex_state = 148, .external_lex_state = 2},
  [4023] = {.lex_state = 148, .external_lex_state = 2},
  [4024] = {.lex_state = 148, .external_lex_state = 2},
  [4025] = {.lex_state = 148, .external_lex_state = 2},
  [4026] = {.lex_state = 148, .external_lex_state = 2},
  [4027] = {.lex_state = 148, .external_lex_state = 2},
  [4028] = {.lex_state = 148, .external_lex_state = 2},
  [4029] = {.lex_state = 148, .external_lex_state = 2},
  [4030] = {.lex_state = 71, .external_lex_state = 2},
  [4031] = {.lex_state = 148, .external_lex_state = 2},
  [4032] = {.lex_state = 148, .external_lex_state = 2},
  [4033] = {.lex_state = 6, .external_lex_state = 2},
  [4034] = {.lex_state = 148, .external_lex_state = 2},
  [4035] = {.lex_state = 148, .external_lex_state = 2},
  [4036] = {.lex_state = 148, .external_lex_state = 2},
  [4037] = {.lex_state = 148, .external_lex_state = 2},
  [4038] = {.lex_state = 34, .external_lex_state = 2},
  [4039] = {.lex_state = 23, .external_lex_state = 2},
  [4040] = {.lex_state = 6, .external_lex_state = 2},
  [4041] = {.lex_state = 34, .external_lex_state = 2},
  [4042] = {.lex_state = 6, .external_lex_state = 2},
  [4043] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4044] = {.lex_state = 148, .external_lex_state = 2},
  [4045] = {.lex_state = 23, .external_lex_state = 2},
  [4046] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4047] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4048] = {.lex_state = 23, .external_lex_state = 2},
  [4049] = {.lex_state = 148, .external_lex_state = 2},
  [4050] = {.lex_state = 6, .external_lex_state = 2},
  [4051] = {.lex_state = 23, .external_lex_state = 2},
  [4052] = {.lex_state = 148, .external_lex_state = 2},
  [4053] = {.lex_state = 34, .external_lex_state = 2},
  [4054] = {.lex_state = 6, .external_lex_state = 2},
  [4055] = {.lex_state = 34, .external_lex_state = 2},
  [4056] = {.lex_state = 148, .external_lex_state = 2},
  [4057] = {.lex_state = 148, .external_lex_state = 2},
  [4058] = {.lex_state = 148, .external_lex_state = 2},
  [4059] = {.lex_state = 34, .external_lex_state = 2},
  [4060] = {.lex_state = 34, .external_lex_state = 2},
  [4061] = {.lex_state = 148, .external_lex_state = 2},
  [4062] = {.lex_state = 34, .external_lex_state = 2},
  [4063] = {.lex_state = 34, .external_lex_state = 2},
  [4064] = {.lex_state = 148, .external_lex_state = 2},
  [4065] = {.lex_state = 34, .external_lex_state = 2},
  [4066] = {.lex_state = 148, .external_lex_state = 2},
  [4067] = {.lex_state = 148, .external_lex_state = 2},
  [4068] = {.lex_state = 148, .external_lex_state = 2},
  [4069] = {.lex_state = 148, .external_lex_state = 2},
  [4070] = {.lex_state = 148, .external_lex_state = 2},
  [4071] = {.lex_state = 6, .external_lex_state = 2},
  [4072] = {.lex_state = 73, .external_lex_state = 2},
  [4073] = {.lex_state = 148, .external_lex_state = 2},
  [4074] = {.lex_state = 148, .external_lex_state = 2},
  [4075] = {.lex_state = 148, .external_lex_state = 2},
  [4076] = {.lex_state = 148, .external_lex_state = 2},
  [4077] = {.lex_state = 148, .external_lex_state = 2},
  [4078] = {.lex_state = 34, .external_lex_state = 2},
  [4079] = {.lex_state = 148, .external_lex_state = 2},
  [4080] = {.lex_state = 34, .external_lex_state = 2},
  [4081] = {.lex_state = 148, .external_lex_state = 2},
  [4082] = {.lex_state = 6, .external_lex_state = 2},
  [4083] = {.lex_state = 148, .external_lex_state = 2},
  [4084] = {.lex_state = 148, .external_lex_state = 2},
  [4085] = {.lex_state = 148, .external_lex_state = 2},
  [4086] = {.lex_state = 148, .external_lex_state = 2},
  [4087] = {.lex_state = 148, .external_lex_state = 2},
  [4088] = {.lex_state = 73, .external_lex_state = 2},
  [4089] = {.lex_state = 73, .external_lex_state = 2},
  [4090] = {.lex_state = 148, .external_lex_state = 2},
  [4091] = {.lex_state = 34, .external_lex_state = 2},
  [4092] = {.lex_state = 148, .external_lex_state = 2},
  [4093] = {.lex_state = 148, .external_lex_state = 2},
  [4094] = {.lex_state = 148, .external_lex_state = 2},
  [4095] = {.lex_state = 148, .external_lex_state = 2},
  [4096] = {.lex_state = 34, .external_lex_state = 2},
  [4097] = {.lex_state = 148, .external_lex_state = 2},
  [4098] = {.lex_state = 148, .external_lex_state = 2},
  [4099] = {.lex_state = 148, .external_lex_state = 2},
  [4100] = {.lex_state = 73, .external_lex_state = 2},
  [4101] = {.lex_state = 148, .external_lex_state = 2},
  [4102] = {.lex_state = 148, .external_lex_state = 2},
  [4103] = {.lex_state = 148, .external_lex_state = 2},
  [4104] = {.lex_state = 148, .external_lex_state = 2},
  [4105] = {.lex_state = 34, .external_lex_state = 2},
  [4106] = {.lex_state = 148, .external_lex_state = 2},
  [4107] = {.lex_state = 34, .external_lex_state = 2},
  [4108] = {.lex_state = 34, .external_lex_state = 2},
  [4109] = {.lex_state = 148, .external_lex_state = 2},
  [4110] = {.lex_state = 34, .external_lex_state = 2},
  [4111] = {.lex_state = 34, .external_lex_state = 2},
  [4112] = {.lex_state = 148, .external_lex_state = 2},
  [4113] = {.lex_state = 34, .external_lex_state = 2},
  [4114] = {.lex_state = 148, .external_lex_state = 2},
  [4115] = {.lex_state = 148, .external_lex_state = 2},
  [4116] = {.lex_state = 148, .external_lex_state = 2},
  [4117] = {.lex_state = 34, .external_lex_state = 2},
  [4118] = {.lex_state = 148, .external_lex_state = 2},
  [4119] = {.lex_state = 148, .external_lex_state = 2},
  [4120] = {.lex_state = 148, .external_lex_state = 2},
  [4121] = {.lex_state = 70, .external_lex_state = 5},
  [4122] = {.lex_state = 70, .external_lex_state = 5},
  [4123] = {.lex_state = 70, .external_lex_state = 5},
  [4124] = {.lex_state = 71, .external_lex_state = 2},
  [4125] = {.lex_state = 70, .external_lex_state = 5},
  [4126] = {.lex_state = 70, .external_lex_state = 5},
  [4127] = {.lex_state = 70, .external_lex_state = 5},
  [4128] = {.lex_state = 151, .external_lex_state = 2},
  [4129] = {.lex_state = 151, .external_lex_state = 2},
  [4130] = {.lex_state = 151, .external_lex_state = 2},
  [4131] = {.lex_state = 151, .external_lex_state = 2},
  [4132] = {.lex_state = 151, .external_lex_state = 2},
  [4133] = {.lex_state = 70, .external_lex_state = 5},
  [4134] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4135] = {.lex_state = 70, .external_lex_state = 5},
  [4136] = {.lex_state = 70, .external_lex_state = 5},
  [4137] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4138] = {.lex_state = 151, .external_lex_state = 2},
  [4139] = {.lex_state = 6, .external_lex_state = 2},
  [4140] = {.lex_state = 6, .external_lex_state = 2},
  [4141] = {.lex_state = 6, .external_lex_state = 2},
  [4142] = {.lex_state = 151, .external_lex_state = 2},
  [4143] = {.lex_state = 71, .external_lex_state = 2},
  [4144] = {.lex_state = 70, .external_lex_state = 5},
  [4145] = {.lex_state = 70, .external_lex_state = 5},
  [4146] = {.lex_state = 151, .external_lex_state = 2},
  [4147] = {.lex_state = 151, .external_lex_state = 2},
  [4148] = {.lex_state = 70, .external_lex_state = 5},
  [4149] = {.lex_state = 151, .external_lex_state = 2},
  [4150] = {.lex_state = 71, .external_lex_state = 2},
  [4151] = {.lex_state = 70, .external_lex_state = 5},
  [4152] = {.lex_state = 71, .external_lex_state = 2},
  [4153] = {.lex_state = 70, .external_lex_state = 5},
  [4154] = {.lex_state = 151, .external_lex_state = 2},
  [4155] = {.lex_state = 151, .external_lex_state = 2},
  [4156] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4157] = {.lex_state = 70, .external_lex_state = 5},
  [4158] = {.lex_state = 70, .external_lex_state = 5},
  [4159] = {.lex_state = 87, .external_lex_state = 2},
  [4160] = {.lex_state = 87, .external_lex_state = 2},
  [4161] = {.lex_state = 87, .external_lex_state = 2},
  [4162] = {.lex_state = 70, .external_lex_state = 5},
  [4163] = {.lex_state = 87, .external_lex_state = 2},
  [4164] = {.lex_state = 151, .external_lex_state = 2},
  [4165] = {.lex_state = 70, .external_lex_state = 5},
  [4166] = {.lex_state = 151, .external_lex_state = 2},
  [4167] = {.lex_state = 151, .external_lex_state = 2},
  [4168] = {.lex_state = 151, .external_lex_state = 2},
  [4169] = {.lex_state = 151, .external_lex_state = 2},
  [4170] = {.lex_state = 151, .external_lex_state = 2},
  [4171] = {.lex_state = 151, .external_lex_state = 2},
  [4172] = {.lex_state = 151, .external_lex_state = 2},
  [4173] = {.lex_state = 151, .external_lex_state = 2},
  [4174] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4175] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4176] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4177] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4178] = {.lex_state = 6, .external_lex_state = 3},
  [4179] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 86},
  [4180] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4181] = {.lex_state = 73, .external_lex_state = 2},
  [4182] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4183] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 65},
  [4184] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4185] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 86},
  [4186] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 65},
  [4187] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4188] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 65},
  [4189] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4190] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 65},
  [4191] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 65},
  [4192] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4193] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4194] = {.lex_state = 70, .external_lex_state = 5},
  [4195] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4196] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4197] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 65},
  [4198] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4199] = {.lex_state = 6, .external_lex_state = 3},
  [4200] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4201] = {.lex_state = 6, .external_lex_state = 2},
  [4202] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4203] = {.lex_state = 6, .external_lex_state = 3},
  [4204] = {.lex_state = 6, .external_lex_state = 3},
  [4205] = {.lex_state = 151, .external_lex_state = 2},
  [4206] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 65},
  [4207] = {.lex_state = 6, .external_lex_state = 2},
  [4208] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4209] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4210] = {.lex_state = 70, .external_lex_state = 5},
  [4211] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4212] = {.lex_state = 6, .external_lex_state = 2},
  [4213] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4214] = {.lex_state = 6, .external_lex_state = 2},
  [4215] = {.lex_state = 71, .external_lex_state = 2},
  [4216] = {.lex_state = 6, .external_lex_state = 2},
  [4217] = {.lex_state = 6, .external_lex_state = 2},
  [4218] = {.lex_state = 151, .external_lex_state = 2},
  [4219] = {.lex_state = 151, .external_lex_state = 2},
  [4220] = {.lex_state = 151, .external_lex_state = 2},
  [4221] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4222] = {.lex_state = 6, .external_lex_state = 2},
  [4223] = {.lex_state = 151, .external_lex_state = 2},
  [4224] = {.lex_state = 71, .external_lex_state = 2},
  [4225] = {.lex_state = 6, .external_lex_state = 2},
  [4226] = {.lex_state = 6, .external_lex_state = 2},
  [4227] = {.lex_state = 151, .external_lex_state = 2},
  [4228] = {.lex_state = 151, .external_lex_state = 2},
  [4229] = {.lex_state = 151, .external_lex_state = 2},
  [4230] = {.lex_state = 6, .external_lex_state = 2},
  [4231] = {.lex_state = 6, .external_lex_state = 2},
  [4232] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4233] = {.lex_state = 6, .external_lex_state = 2},
  [4234] = {.lex_state = 6, .external_lex_state = 2},
  [4235] = {.lex_state = 72, .external_lex_state = 2},
  [4236] = {.lex_state = 6, .external_lex_state = 2},
  [4237] = {.lex_state = 6, .external_lex_state = 2},
  [4238] = {.lex_state = 7, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4239] = {.lex_state = 6, .external_lex_state = 2},
  [4240] = {.lex_state = 6, .external_lex_state = 2},
  [4241] = {.lex_state = 6, .external_lex_state = 2},
  [4242] = {.lex_state = 6, .external_lex_state = 2},
  [4243] = {.lex_state = 148, .external_lex_state = 2},
  [4244] = {.lex_state = 6, .external_lex_state = 2},
  [4245] = {.lex_state = 72, .external_lex_state = 2},
  [4246] = {.lex_state = 6, .external_lex_state = 2},
  [4247] = {.lex_state = 70, .external_lex_state = 5},
  [4248] = {.lex_state = 6, .external_lex_state = 2},
  [4249] = {.lex_state = 6, .external_lex_state = 2},
  [4250] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4251] = {.lex_state = 72, .external_lex_state = 2},
  [4252] = {.lex_state = 148, .external_lex_state = 2},
  [4253] = {.lex_state = 148, .external_lex_state = 2},
  [4254] = {.lex_state = 89, .external_lex_state = 2},
  [4255] = {.lex_state = 148, .external_lex_state = 2},
  [4256] = {.lex_state = 148, .external_lex_state = 2},
  [4257] = {.lex_state = 148, .external_lex_state = 2},
  [4258] = {.lex_state = 72, .external_lex_state = 2},
  [4259] = {.lex_state = 6, .external_lex_state = 2},
  [4260] = {.lex_state = 72, .external_lex_state = 2},
  [4261] = {.lex_state = 6, .external_lex_state = 2},
  [4262] = {.lex_state = 148, .external_lex_state = 2},
  [4263] = {.lex_state = 75, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4264] = {.lex_state = 6, .external_lex_state = 2},
  [4265] = {.lex_state = 72, .external_lex_state = 2},
  [4266] = {.lex_state = 89, .external_lex_state = 2},
  [4267] = {.lex_state = 6, .external_lex_state = 2},
  [4268] = {.lex_state = 6, .external_lex_state = 2},
  [4269] = {.lex_state = 148, .external_lex_state = 2},
  [4270] = {.lex_state = 148, .external_lex_state = 2},
  [4271] = {.lex_state = 6, .external_lex_state = 2},
  [4272] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4273] = {.lex_state = 6, .external_lex_state = 2},
  [4274] = {.lex_state = 72, .external_lex_state = 2},
  [4275] = {.lex_state = 6, .external_lex_state = 2},
  [4276] = {.lex_state = 70, .external_lex_state = 5},
  [4277] = {.lex_state = 148, .external_lex_state = 2},
  [4278] = {.lex_state = 6, .external_lex_state = 2},
  [4279] = {.lex_state = 148, .external_lex_state = 2},
  [4280] = {.lex_state = 148, .external_lex_state = 2},
  [4281] = {.lex_state = 148, .external_lex_state = 2},
  [4282] = {.lex_state = 148, .external_lex_state = 2},
  [4283] = {.lex_state = 148, .external_lex_state = 2},
  [4284] = {.lex_state = 148, .external_lex_state = 2},
  [4285] = {.lex_state = 89, .external_lex_state = 2},
  [4286] = {.lex_state = 148, .external_lex_state = 2},
  [4287] = {.lex_state = 148, .external_lex_state = 2},
  [4288] = {.lex_state = 6, .external_lex_state = 2},
  [4289] = {.lex_state = 148, .external_lex_state = 2},
  [4290] = {.lex_state = 6, .external_lex_state = 2},
  [4291] = {.lex_state = 6, .external_lex_state = 2},
  [4292] = {.lex_state = 148, .external_lex_state = 2},
  [4293] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4294] = {.lex_state = 148, .external_lex_state = 2},
  [4295] = {.lex_state = 148, .external_lex_state = 2},
  [4296] = {.lex_state = 6, .external_lex_state = 2},
  [4297] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4298] = {.lex_state = 6, .external_lex_state = 2},
  [4299] = {.lex_state = 148, .external_lex_state = 2},
  [4300] = {.lex_state = 72, .external_lex_state = 2},
  [4301] = {.lex_state = 6, .external_lex_state = 2},
  [4302] = {.lex_state = 146, .external_lex_state = 2},
  [4303] = {.lex_state = 151, .external_lex_state = 2},
  [4304] = {.lex_state = 146, .external_lex_state = 2},
  [4305] = {.lex_state = 148, .external_lex_state = 2},
  [4306] = {.lex_state = 146, .external_lex_state = 2},
  [4307] = {.lex_state = 71, .external_lex_state = 2},
  [4308] = {.lex_state = 71, .external_lex_state = 2},
  [4309] = {.lex_state = 71, .external_lex_state = 2},
  [4310] = {.lex_state = 148, .external_lex_state = 2},
  [4311] = {.lex_state = 148, .external_lex_state = 2},
  [4312] = {.lex_state = 151, .external_lex_state = 2},
  [4313] = {.lex_state = 146, .external_lex_state = 2},
  [4314] = {.lex_state = 148, .external_lex_state = 2},
  [4315] = {.lex_state = 148, .external_lex_state = 2},
  [4316] = {.lex_state = 71, .external_lex_state = 2},
  [4317] = {.lex_state = 148, .external_lex_state = 2},
  [4318] = {.lex_state = 146, .external_lex_state = 2},
  [4319] = {.lex_state = 148, .external_lex_state = 2},
  [4320] = {.lex_state = 148, .external_lex_state = 2},
  [4321] = {.lex_state = 146, .external_lex_state = 2},
  [4322] = {.lex_state = 71, .external_lex_state = 2},
  [4323] = {.lex_state = 71, .external_lex_state = 2},
  [4324] = {.lex_state = 71, .external_lex_state = 2},
  [4325] = {.lex_state = 148, .external_lex_state = 2},
  [4326] = {.lex_state = 146, .external_lex_state = 2},
  [4327] = {.lex_state = 148, .external_lex_state = 2},
  [4328] = {.lex_state = 148, .external_lex_state = 2},
  [4329] = {.lex_state = 71, .external_lex_state = 2},
  [4330] = {.lex_state = 71, .external_lex_state = 2},
  [4331] = {.lex_state = 148, .external_lex_state = 2},
  [4332] = {.lex_state = 148, .external_lex_state = 2},
  [4333] = {.lex_state = 71, .external_lex_state = 2},
  [4334] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4335] = {.lex_state = 148, .external_lex_state = 2},
  [4336] = {.lex_state = 71, .external_lex_state = 2},
  [4337] = {.lex_state = 71, .external_lex_state = 2},
  [4338] = {.lex_state = 148, .external_lex_state = 2},
  [4339] = {.lex_state = 146, .external_lex_state = 2},
  [4340] = {.lex_state = 71, .external_lex_state = 2},
  [4341] = {.lex_state = 71, .external_lex_state = 2},
  [4342] = {.lex_state = 146, .external_lex_state = 2},
  [4343] = {.lex_state = 146, .external_lex_state = 2},
  [4344] = {.lex_state = 148, .external_lex_state = 2},
  [4345] = {.lex_state = 71, .external_lex_state = 2},
  [4346] = {.lex_state = 146, .external_lex_state = 2},
  [4347] = {.lex_state = 148, .external_lex_state = 2},
  [4348] = {.lex_state = 148, .external_lex_state = 2},
  [4349] = {.lex_state = 148, .external_lex_state = 2},
  [4350] = {.lex_state = 148, .external_lex_state = 2},
  [4351] = {.lex_state = 71, .external_lex_state = 2},
  [4352] = {.lex_state = 148, .external_lex_state = 2},
  [4353] = {.lex_state = 148, .external_lex_state = 2},
  [4354] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4355] = {.lex_state = 71, .external_lex_state = 2},
  [4356] = {.lex_state = 146, .external_lex_state = 2},
  [4357] = {.lex_state = 148, .external_lex_state = 2},
  [4358] = {.lex_state = 146, .external_lex_state = 2},
  [4359] = {.lex_state = 148, .external_lex_state = 2},
  [4360] = {.lex_state = 148, .external_lex_state = 2},
  [4361] = {.lex_state = 148, .external_lex_state = 2},
  [4362] = {.lex_state = 146, .external_lex_state = 2},
  [4363] = {.lex_state = 148, .external_lex_state = 2},
  [4364] = {.lex_state = 148, .external_lex_state = 2},
  [4365] = {.lex_state = 148, .external_lex_state = 2},
  [4366] = {.lex_state = 146, .external_lex_state = 2},
  [4367] = {.lex_state = 71, .external_lex_state = 2},
  [4368] = {.lex_state = 71, .external_lex_state = 2},
  [4369] = {.lex_state = 146, .external_lex_state = 2},
  [4370] = {.lex_state = 148, .external_lex_state = 2},
  [4371] = {.lex_state = 148, .external_lex_state = 2},
  [4372] = {.lex_state = 148, .external_lex_state = 2},
  [4373] = {.lex_state = 148, .external_lex_state = 2},
  [4374] = {.lex_state = 71, .external_lex_state = 2},
  [4375] = {.lex_state = 71, .external_lex_state = 2},
  [4376] = {.lex_state = 148, .external_lex_state = 2},
  [4377] = {.lex_state = 151, .external_lex_state = 2},
  [4378] = {.lex_state = 148, .external_lex_state = 2},
  [4379] = {.lex_state = 148, .external_lex_state = 2},
  [4380] = {.lex_state = 148, .external_lex_state = 2},
  [4381] = {.lex_state = 148, .external_lex_state = 2},
  [4382] = {.lex_state = 148, .external_lex_state = 2},
  [4383] = {.lex_state = 146, .external_lex_state = 2},
  [4384] = {.lex_state = 148, .external_lex_state = 2},
  [4385] = {.lex_state = 148, .external_lex_state = 2},
  [4386] = {.lex_state = 148, .external_lex_state = 2},
  [4387] = {.lex_state = 71, .external_lex_state = 2},
  [4388] = {.lex_state = 148, .external_lex_state = 2},
  [4389] = {.lex_state = 148, .external_lex_state = 2},
  [4390] = {.lex_state = 148, .external_lex_state = 2},
  [4391] = {.lex_state = 148, .external_lex_state = 2},
  [4392] = {.lex_state = 148, .external_lex_state = 2},
  [4393] = {.lex_state = 146, .external_lex_state = 2},
  [4394] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 90},
  [4395] = {.lex_state = 148, .external_lex_state = 2},
  [4396] = {.lex_state = 148, .external_lex_state = 2},
  [4397] = {.lex_state = 146, .external_lex_state = 2},
  [4398] = {.lex_state = 148, .external_lex_state = 2},
  [4399] = {.lex_state = 148, .external_lex_state = 2},
  [4400] = {.lex_state = 148, .external_lex_state = 2},
  [4401] = {.lex_state = 148, .external_lex_state = 2},
  [4402] = {.lex_state = 148, .external_lex_state = 2},
  [4403] = {.lex_state = 71, .external_lex_state = 2},
  [4404] = {.lex_state = 148, .external_lex_state = 2},
  [4405] = {.lex_state = 71, .external_lex_state = 2},
  [4406] = {.lex_state = 148, .external_lex_state = 2},
  [4407] = {.lex_state = 71, .external_lex_state = 2},
  [4408] = {.lex_state = 148, .external_lex_state = 2},
  [4409] = {.lex_state = 94, .external_lex_state = 2},
  [4410] = {.lex_state = 148, .external_lex_state = 2},
  [4411] = {.lex_state = 148, .external_lex_state = 2},
  [4412] = {.lex_state = 71, .external_lex_state = 2},
  [4413] = {.lex_state = 148, .external_lex_state = 2},
  [4414] = {.lex_state = 71, .external_lex_state = 2},
  [4415] = {.lex_state = 94, .external_lex_state = 2},
  [4416] = {.lex_state = 146, .external_lex_state = 2},
  [4417] = {.lex_state = 148, .external_lex_state = 2},
  [4418] = {.lex_state = 71, .external_lex_state = 2},
  [4419] = {.lex_state = 148, .external_lex_state = 2},
  [4420] = {.lex_state = 146, .external_lex_state = 2},
  [4421] = {.lex_state = 148, .external_lex_state = 2},
  [4422] = {.lex_state = 71, .external_lex_state = 2},
  [4423] = {.lex_state = 148, .external_lex_state = 2},
  [4424] = {.lex_state = 71, .external_lex_state = 2},
  [4425] = {.lex_state = 71, .external_lex_state = 2},
  [4426] = {.lex_state = 148, .external_lex_state = 2},
  [4427] = {.lex_state = 146, .external_lex_state = 2},
  [4428] = {.lex_state = 71, .external_lex_state = 2},
  [4429] = {.lex_state = 148, .external_lex_state = 2},
  [4430] = {.lex_state = 71, .external_lex_state = 2},
  [4431] = {.lex_state = 71, .external_lex_state = 2},
  [4432] = {.lex_state = 148, .external_lex_state = 2},
  [4433] = {.lex_state = 148, .external_lex_state = 2},
  [4434] = {.lex_state = 148, .external_lex_state = 2},
  [4435] = {.lex_state = 71, .external_lex_state = 2},
  [4436] = {.lex_state = 71, .external_lex_state = 2},
  [4437] = {.lex_state = 148, .external_lex_state = 2},
  [4438] = {.lex_state = 148, .external_lex_state = 2},
  [4439] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4440] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4441] = {.lex_state = 148, .external_lex_state = 2},
  [4442] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4443] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4444] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4445] = {.lex_state = 148, .external_lex_state = 2},
  [4446] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4447] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4448] = {.lex_state = 148, .external_lex_state = 2},
  [4449] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4450] = {.lex_state = 148, .external_lex_state = 2},
  [4451] = {.lex_state = 91, .external_lex_state = 2},
  [4452] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4453] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4454] = {.lex_state = 72, .external_lex_state = 2},
  [4455] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4456] = {.lex_state = 6, .external_lex_state = 2},
  [4457] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4458] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4459] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4460] = {.lex_state = 148, .external_lex_state = 2},
  [4461] = {.lex_state = 151, .external_lex_state = 2},
  [4462] = {.lex_state = 151, .external_lex_state = 2},
  [4463] = {.lex_state = 151, .external_lex_state = 2},
  [4464] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4465] = {.lex_state = 6, .external_lex_state = 2},
  [4466] = {.lex_state = 151, .external_lex_state = 2},
  [4467] = {.lex_state = 148, .external_lex_state = 2},
  [4468] = {.lex_state = 6, .external_lex_state = 2},
  [4469] = {.lex_state = 151, .external_lex_state = 2},
  [4470] = {.lex_state = 148, .external_lex_state = 2},
  [4471] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4472] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4473] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4474] = {.lex_state = 148, .external_lex_state = 2},
  [4475] = {.lex_state = 6, .external_lex_state = 2},
  [4476] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4477] = {.lex_state = 151, .external_lex_state = 2},
  [4478] = {.lex_state = 148, .external_lex_state = 2},
  [4479] = {.lex_state = 151, .external_lex_state = 2},
  [4480] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4481] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4482] = {.lex_state = 151, .external_lex_state = 2},
  [4483] = {.lex_state = 6, .external_lex_state = 2},
  [4484] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4485] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4486] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4487] = {.lex_state = 148, .external_lex_state = 2},
  [4488] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4489] = {.lex_state = 6, .external_lex_state = 2},
  [4490] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4491] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4492] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 89},
  [4493] = {.lex_state = 72, .external_lex_state = 2},
  [4494] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4495] = {.lex_state = 151, .external_lex_state = 2},
  [4496] = {.lex_state = 6, .external_lex_state = 2},
  [4497] = {.lex_state = 151, .external_lex_state = 2},
  [4498] = {.lex_state = 151, .external_lex_state = 2},
  [4499] = {.lex_state = 151, .external_lex_state = 2},
  [4500] = {.lex_state = 151, .external_lex_state = 2},
  [4501] = {.lex_state = 6, .external_lex_state = 2},
  [4502] = {.lex_state = 151, .external_lex_state = 2},
  [4503] = {.lex_state = 6, .external_lex_state = 2},
  [4504] = {.lex_state = 6, .external_lex_state = 2},
  [4505] = {.lex_state = 6, .external_lex_state = 2},
  [4506] = {.lex_state = 6, .external_lex_state = 2},
  [4507] = {.lex_state = 6, .external_lex_state = 2},
  [4508] = {.lex_state = 151, .external_lex_state = 2},
  [4509] = {.lex_state = 6, .external_lex_state = 2},
  [4510] = {.lex_state = 71, .external_lex_state = 2},
  [4511] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4512] = {.lex_state = 151, .external_lex_state = 2},
  [4513] = {.lex_state = 151, .external_lex_state = 2},
  [4514] = {.lex_state = 6, .external_lex_state = 2},
  [4515] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4516] = {.lex_state = 6, .external_lex_state = 2},
  [4517] = {.lex_state = 151, .external_lex_state = 2},
  [4518] = {.lex_state = 6, .external_lex_state = 2},
  [4519] = {.lex_state = 6, .external_lex_state = 2},
  [4520] = {.lex_state = 6, .external_lex_state = 2},
  [4521] = {.lex_state = 151, .external_lex_state = 2},
  [4522] = {.lex_state = 151, .external_lex_state = 2},
  [4523] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4524] = {.lex_state = 89, .external_lex_state = 2},
  [4525] = {.lex_state = 151, .external_lex_state = 2},
  [4526] = {.lex_state = 151, .external_lex_state = 2},
  [4527] = {.lex_state = 151, .external_lex_state = 2},
  [4528] = {.lex_state = 6, .external_lex_state = 2},
  [4529] = {.lex_state = 151, .external_lex_state = 2},
  [4530] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4531] = {.lex_state = 151, .external_lex_state = 2},
  [4532] = {.lex_state = 6, .external_lex_state = 2},
  [4533] = {.lex_state = 6, .external_lex_state = 2},
  [4534] = {.lex_state = 151, .external_lex_state = 2},
  [4535] = {.lex_state = 151, .external_lex_state = 2},
  [4536] = {.lex_state = 6, .external_lex_state = 2},
  [4537] = {.lex_state = 71, .external_lex_state = 2},
  [4538] = {.lex_state = 151, .external_lex_state = 2},
  [4539] = {.lex_state = 4, .external_lex_state = 6},
  [4540] = {.lex_state = 148, .external_lex_state = 2},
  [4541] = {.lex_state = 4, .external_lex_state = 6},
  [4542] = {.lex_state = 71, .external_lex_state = 2},
  [4543] = {.lex_state = 151, .external_lex_state = 2},
  [4544] = {.lex_state = 151, .external_lex_state = 2},
  [4545] = {.lex_state = 151, .external_lex_state = 2},
  [4546] = {.lex_state = 71, .external_lex_state = 2},
  [4547] = {.lex_state = 151, .external_lex_state = 2},
  [4548] = {.lex_state = 71, .external_lex_state = 2},
  [4549] = {.lex_state = 85, .external_lex_state = 7},
  [4550] = {.lex_state = 148, .external_lex_state = 3},
  [4551] = {.lex_state = 4, .external_lex_state = 6},
  [4552] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4553] = {.lex_state = 71, .external_lex_state = 2},
  [4554] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4555] = {.lex_state = 4, .external_lex_state = 6},
  [4556] = {.lex_state = 151, .external_lex_state = 2},
  [4557] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4558] = {.lex_state = 4, .external_lex_state = 6},
  [4559] = {.lex_state = 151, .external_lex_state = 2},
  [4560] = {.lex_state = 71, .external_lex_state = 2},
  [4561] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4562] = {.lex_state = 71, .external_lex_state = 2},
  [4563] = {.lex_state = 151, .external_lex_state = 2},
  [4564] = {.lex_state = 71, .external_lex_state = 2},
  [4565] = {.lex_state = 71, .external_lex_state = 2},
  [4566] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4567] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4568] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4569] = {.lex_state = 4, .external_lex_state = 6},
  [4570] = {.lex_state = 71, .external_lex_state = 2},
  [4571] = {.lex_state = 71, .external_lex_state = 2},
  [4572] = {.lex_state = 71, .external_lex_state = 2},
  [4573] = {.lex_state = 151, .external_lex_state = 2},
  [4574] = {.lex_state = 151, .external_lex_state = 2},
  [4575] = {.lex_state = 71, .external_lex_state = 2},
  [4576] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4577] = {.lex_state = 4, .external_lex_state = 6},
  [4578] = {.lex_state = 4, .external_lex_state = 6},
  [4579] = {.lex_state = 71, .external_lex_state = 2},
  [4580] = {.lex_state = 71, .external_lex_state = 2},
  [4581] = {.lex_state = 71, .external_lex_state = 2},
  [4582] = {.lex_state = 151, .external_lex_state = 2},
  [4583] = {.lex_state = 4, .external_lex_state = 6},
  [4584] = {.lex_state = 71, .external_lex_state = 2},
  [4585] = {.lex_state = 71, .external_lex_state = 2},
  [4586] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4587] = {.lex_state = 71, .external_lex_state = 2},
  [4588] = {.lex_state = 71, .external_lex_state = 2},
  [4589] = {.lex_state = 151, .external_lex_state = 2},
  [4590] = {.lex_state = 4, .external_lex_state = 6},
  [4591] = {.lex_state = 4, .external_lex_state = 6},
  [4592] = {.lex_state = 71, .external_lex_state = 2},
  [4593] = {.lex_state = 151, .external_lex_state = 2},
  [4594] = {.lex_state = 71, .external_lex_state = 2},
  [4595] = {.lex_state = 4, .external_lex_state = 6},
  [4596] = {.lex_state = 4, .external_lex_state = 6},
  [4597] = {.lex_state = 4, .external_lex_state = 6},
  [4598] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4599] = {.lex_state = 71, .external_lex_state = 2},
  [4600] = {.lex_state = 71, .external_lex_state = 2},
  [4601] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4602] = {.lex_state = 71, .external_lex_state = 2},
  [4603] = {.lex_state = 148, .external_lex_state = 3},
  [4604] = {.lex_state = 4, .external_lex_state = 6},
  [4605] = {.lex_state = 71, .external_lex_state = 2},
  [4606] = {.lex_state = 4, .external_lex_state = 6},
  [4607] = {.lex_state = 151, .external_lex_state = 2},
  [4608] = {.lex_state = 151, .external_lex_state = 2},
  [4609] = {.lex_state = 71, .external_lex_state = 2},
  [4610] = {.lex_state = 151, .external_lex_state = 2},
  [4611] = {.lex_state = 71, .external_lex_state = 2},
  [4612] = {.lex_state = 71, .external_lex_state = 2},
  [4613] = {.lex_state = 4, .external_lex_state = 6},
  [4614] = {.lex_state = 71, .external_lex_state = 2},
  [4615] = {.lex_state = 71, .external_lex_state = 2},
  [4616] = {.lex_state = 71, .external_lex_state = 2},
  [4617] = {.lex_state = 148, .external_lex_state = 2},
  [4618] = {.lex_state = 71, .external_lex_state = 2},
  [4619] = {.lex_state = 10, .external_lex_state = 2},
  [4620] = {.lex_state = 36, .external_lex_state = 2},
  [4621] = {.lex_state = 71, .external_lex_state = 2},
  [4622] = {.lex_state = 10, .external_lex_state = 2},
  [4623] = {.lex_state = 148, .external_lex_state = 2},
  [4624] = {.lex_state = 148, .external_lex_state = 2},
  [4625] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4626] = {.lex_state = 151, .external_lex_state = 2},
  [4627] = {.lex_state = 151, .external_lex_state = 2},
  [4628] = {.lex_state = 151, .external_lex_state = 2},
  [4629] = {.lex_state = 36, .external_lex_state = 2},
  [4630] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4631] = {.lex_state = 10, .external_lex_state = 2},
  [4632] = {.lex_state = 85, .external_lex_state = 7},
  [4633] = {.lex_state = 151, .external_lex_state = 2},
  [4634] = {.lex_state = 36, .external_lex_state = 2},
  [4635] = {.lex_state = 151, .external_lex_state = 2},
  [4636] = {.lex_state = 151, .external_lex_state = 2},
  [4637] = {.lex_state = 10, .external_lex_state = 2},
  [4638] = {.lex_state = 91, .external_lex_state = 2},
  [4639] = {.lex_state = 91, .external_lex_state = 2},
  [4640] = {.lex_state = 6, .external_lex_state = 2},
  [4641] = {.lex_state = 151, .external_lex_state = 2},
  [4642] = {.lex_state = 151, .external_lex_state = 2},
  [4643] = {.lex_state = 151, .external_lex_state = 2},
  [4644] = {.lex_state = 151, .external_lex_state = 2},
  [4645] = {.lex_state = 10, .external_lex_state = 2},
  [4646] = {.lex_state = 151, .external_lex_state = 2},
  [4647] = {.lex_state = 10, .external_lex_state = 2},
  [4648] = {.lex_state = 36, .external_lex_state = 2},
  [4649] = {.lex_state = 151, .external_lex_state = 2},
  [4650] = {.lex_state = 151, .external_lex_state = 2},
  [4651] = {.lex_state = 151, .external_lex_state = 2},
  [4652] = {.lex_state = 10, .external_lex_state = 2},
  [4653] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4654] = {.lex_state = 151, .external_lex_state = 2},
  [4655] = {.lex_state = 36, .external_lex_state = 2},
  [4656] = {.lex_state = 151, .external_lex_state = 2},
  [4657] = {.lex_state = 151, .external_lex_state = 2},
  [4658] = {.lex_state = 151, .external_lex_state = 2},
  [4659] = {.lex_state = 151, .external_lex_state = 2},
  [4660] = {.lex_state = 151, .external_lex_state = 2},
  [4661] = {.lex_state = 10, .external_lex_state = 2},
  [4662] = {.lex_state = 36, .external_lex_state = 2},
  [4663] = {.lex_state = 10, .external_lex_state = 2},
  [4664] = {.lex_state = 10, .external_lex_state = 2},
  [4665] = {.lex_state = 151, .external_lex_state = 2},
  [4666] = {.lex_state = 151, .external_lex_state = 2},
  [4667] = {.lex_state = 151, .external_lex_state = 2},
  [4668] = {.lex_state = 151, .external_lex_state = 2},
  [4669] = {.lex_state = 151, .external_lex_state = 2},
  [4670] = {.lex_state = 151, .external_lex_state = 2},
  [4671] = {.lex_state = 10, .external_lex_state = 2},
  [4672] = {.lex_state = 6, .external_lex_state = 2},
  [4673] = {.lex_state = 10, .external_lex_state = 2},
  [4674] = {.lex_state = 85, .external_lex_state = 7},
  [4675] = {.lex_state = 151, .external_lex_state = 2},
  [4676] = {.lex_state = 151, .external_lex_state = 2},
  [4677] = {.lex_state = 151, .external_lex_state = 2},
  [4678] = {.lex_state = 61, .external_lex_state = 2},
  [4679] = {.lex_state = 151, .external_lex_state = 2},
  [4680] = {.lex_state = 151, .external_lex_state = 2},
  [4681] = {.lex_state = 151, .external_lex_state = 2},
  [4682] = {.lex_state = 151, .external_lex_state = 2},
  [4683] = {.lex_state = 151, .external_lex_state = 2},
  [4684] = {.lex_state = 151, .external_lex_state = 2},
  [4685] = {.lex_state = 151, .external_lex_state = 2},
  [4686] = {.lex_state = 151, .external_lex_state = 2},
  [4687] = {.lex_state = 6, .external_lex_state = 2},
  [4688] = {.lex_state = 6, .external_lex_state = 2},
  [4689] = {.lex_state = 6, .external_lex_state = 2},
  [4690] = {.lex_state = 6, .external_lex_state = 2},
  [4691] = {.lex_state = 71, .external_lex_state = 2},
  [4692] = {.lex_state = 10, .external_lex_state = 2},
  [4693] = {.lex_state = 10, .external_lex_state = 2},
  [4694] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4695] = {.lex_state = 6, .external_lex_state = 2},
  [4696] = {.lex_state = 10, .external_lex_state = 2},
  [4697] = {.lex_state = 6, .external_lex_state = 2},
  [4698] = {.lex_state = 10, .external_lex_state = 2},
  [4699] = {.lex_state = 151, .external_lex_state = 2},
  [4700] = {.lex_state = 151, .external_lex_state = 2},
  [4701] = {.lex_state = 6, .external_lex_state = 2},
  [4702] = {.lex_state = 151, .external_lex_state = 2},
  [4703] = {.lex_state = 71, .external_lex_state = 2},
  [4704] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4705] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4706] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4707] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4708] = {.lex_state = 71, .external_lex_state = 2},
  [4709] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4710] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4711] = {.lex_state = 71, .external_lex_state = 2},
  [4712] = {.lex_state = 71, .external_lex_state = 2},
  [4713] = {.lex_state = 71, .external_lex_state = 2},
  [4714] = {.lex_state = 71, .external_lex_state = 2},
  [4715] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4716] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4717] = {.lex_state = 10, .external_lex_state = 2},
  [4718] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4719] = {.lex_state = 146, .external_lex_state = 8, .reserved_word_set_id = 1},
  [4720] = {.lex_state = 71, .external_lex_state = 2},
  [4721] = {.lex_state = 146, .external_lex_state = 8, .reserved_word_set_id = 1},
  [4722] = {.lex_state = 146, .external_lex_state = 8, .reserved_word_set_id = 1},
  [4723] = {.lex_state = 146, .external_lex_state = 8, .reserved_word_set_id = 1},
  [4724] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4725] = {.lex_state = 10, .external_lex_state = 2},
  [4726] = {.lex_state = 6, .external_lex_state = 2},
  [4727] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 91},
  [4728] = {.lex_state = 71, .external_lex_state = 2},
  [4729] = {.lex_state = 6, .external_lex_state = 2},
  [4730] = {.lex_state = 71, .external_lex_state = 2},
  [4731] = {.lex_state = 6, .external_lex_state = 2},
  [4732] = {.lex_state = 6, .external_lex_state = 2},
  [4733] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4734] = {.lex_state = 71, .external_lex_state = 2},
  [4735] = {.lex_state = 151, .external_lex_state = 2},
  [4736] = {.lex_state = 146, .external_lex_state = 8, .reserved_word_set_id = 1},
  [4737] = {.lex_state = 91, .external_lex_state = 2},
  [4738] = {.lex_state = 146, .external_lex_state = 8, .reserved_word_set_id = 1},
  [4739] = {.lex_state = 10, .external_lex_state = 2},
  [4740] = {.lex_state = 6, .external_lex_state = 2},
  [4741] = {.lex_state = 6, .external_lex_state = 2},
  [4742] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4743] = {.lex_state = 6, .external_lex_state = 2},
  [4744] = {.lex_state = 71, .external_lex_state = 2},
  [4745] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4746] = {.lex_state = 6, .external_lex_state = 2},
  [4747] = {.lex_state = 151, .external_lex_state = 2},
  [4748] = {.lex_state = 6, .external_lex_state = 2},
  [4749] = {.lex_state = 10, .external_lex_state = 2},
  [4750] = {.lex_state = 6, .external_lex_state = 2},
  [4751] = {.lex_state = 6, .external_lex_state = 2},
  [4752] = {.lex_state = 6, .external_lex_state = 2},
  [4753] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4754] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4755] = {.lex_state = 6, .external_lex_state = 2},
  [4756] = {.lex_state = 148, .external_lex_state = 3},
  [4757] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4758] = {.lex_state = 6, .external_lex_state = 2},
  [4759] = {.lex_state = 6, .external_lex_state = 2},
  [4760] = {.lex_state = 151, .external_lex_state = 2},
  [4761] = {.lex_state = 6, .external_lex_state = 2},
  [4762] = {.lex_state = 6, .external_lex_state = 2},
  [4763] = {.lex_state = 6, .external_lex_state = 2},
  [4764] = {.lex_state = 10, .external_lex_state = 2},
  [4765] = {.lex_state = 6, .external_lex_state = 2},
  [4766] = {.lex_state = 6, .external_lex_state = 2},
  [4767] = {.lex_state = 146, .external_lex_state = 8, .reserved_word_set_id = 1},
  [4768] = {.lex_state = 6, .external_lex_state = 2},
  [4769] = {.lex_state = 6, .external_lex_state = 2},
  [4770] = {.lex_state = 6, .external_lex_state = 2},
  [4771] = {.lex_state = 71, .external_lex_state = 2},
  [4772] = {.lex_state = 6, .external_lex_state = 2},
  [4773] = {.lex_state = 6, .external_lex_state = 2},
  [4774] = {.lex_state = 6, .external_lex_state = 2},
  [4775] = {.lex_state = 6, .external_lex_state = 2},
  [4776] = {.lex_state = 6, .external_lex_state = 2},
  [4777] = {.lex_state = 10, .external_lex_state = 2},
  [4778] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4779] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4780] = {.lex_state = 10, .external_lex_state = 2},
  [4781] = {.lex_state = 6, .external_lex_state = 2},
  [4782] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4783] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4784] = {.lex_state = 6, .external_lex_state = 2},
  [4785] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4786] = {.lex_state = 10, .external_lex_state = 2},
  [4787] = {.lex_state = 10, .external_lex_state = 2},
  [4788] = {.lex_state = 149, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4789] = {.lex_state = 146, .external_lex_state = 8, .reserved_word_set_id = 1},
  [4790] = {.lex_state = 71, .external_lex_state = 2},
  [4791] = {.lex_state = 10, .external_lex_state = 2},
  [4792] = {.lex_state = 71, .external_lex_state = 2},
  [4793] = {.lex_state = 71, .external_lex_state = 2},
  [4794] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4795] = {.lex_state = 10, .external_lex_state = 2},
  [4796] = {.lex_state = 151, .external_lex_state = 2},
  [4797] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4798] = {.lex_state = 95, .external_lex_state = 2},
  [4799] = {.lex_state = 151, .external_lex_state = 2},
  [4800] = {.lex_state = 151, .external_lex_state = 2},
  [4801] = {.lex_state = 151, .external_lex_state = 2},
  [4802] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4803] = {.lex_state = 10, .external_lex_state = 2},
  [4804] = {.lex_state = 71, .external_lex_state = 2},
  [4805] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 91},
  [4806] = {.lex_state = 6, .external_lex_state = 2},
  [4807] = {.lex_state = 151, .external_lex_state = 2},
  [4808] = {.lex_state = 71, .external_lex_state = 2},
  [4809] = {.lex_state = 151, .external_lex_state = 2},
  [4810] = {.lex_state = 151, .external_lex_state = 2},
  [4811] = {.lex_state = 6, .external_lex_state = 2},
  [4812] = {.lex_state = 95, .external_lex_state = 2},
  [4813] = {.lex_state = 6, .external_lex_state = 2},
  [4814] = {.lex_state = 85, .external_lex_state = 7},
  [4815] = {.lex_state = 10, .external_lex_state = 2},
  [4816] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4817] = {.lex_state = 10, .external_lex_state = 2},
  [4818] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4819] = {.lex_state = 10, .external_lex_state = 2},
  [4820] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4821] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4822] = {.lex_state = 10, .external_lex_state = 2},
  [4823] = {.lex_state = 10, .external_lex_state = 2},
  [4824] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4825] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4826] = {.lex_state = 10, .external_lex_state = 2},
  [4827] = {.lex_state = 10, .external_lex_state = 2},
  [4828] = {.lex_state = 10, .external_lex_state = 2},
  [4829] = {.lex_state = 6, .external_lex_state = 2},
  [4830] = {.lex_state = 10, .external_lex_state = 2},
  [4831] = {.lex_state = 10, .external_lex_state = 2},
  [4832] = {.lex_state = 10, .external_lex_state = 2},
  [4833] = {.lex_state = 10, .external_lex_state = 2},
  [4834] = {.lex_state = 151, .external_lex_state = 2},
  [4835] = {.lex_state = 10, .external_lex_state = 2},
  [4836] = {.lex_state = 10, .external_lex_state = 2},
  [4837] = {.lex_state = 10, .external_lex_state = 2},
  [4838] = {.lex_state = 10, .external_lex_state = 2},
  [4839] = {.lex_state = 10, .external_lex_state = 2},
  [4840] = {.lex_state = 10, .external_lex_state = 2},
  [4841] = {.lex_state = 6, .external_lex_state = 2},
  [4842] = {.lex_state = 10, .external_lex_state = 2},
  [4843] = {.lex_state = 10, .external_lex_state = 2},
  [4844] = {.lex_state = 6, .external_lex_state = 2},
  [4845] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4846] = {.lex_state = 10, .external_lex_state = 2},
  [4847] = {.lex_state = 10, .external_lex_state = 2},
  [4848] = {.lex_state = 10, .external_lex_state = 2},
  [4849] = {.lex_state = 151, .external_lex_state = 2},
  [4850] = {.lex_state = 10, .external_lex_state = 2},
  [4851] = {.lex_state = 151, .external_lex_state = 2},
  [4852] = {.lex_state = 6, .external_lex_state = 2},
  [4853] = {.lex_state = 10, .external_lex_state = 2},
  [4854] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4855] = {.lex_state = 6, .external_lex_state = 2},
  [4856] = {.lex_state = 10, .external_lex_state = 2},
  [4857] = {.lex_state = 10, .external_lex_state = 2},
  [4858] = {.lex_state = 10, .external_lex_state = 2},
  [4859] = {.lex_state = 10, .external_lex_state = 2},
  [4860] = {.lex_state = 10, .external_lex_state = 2},
  [4861] = {.lex_state = 146, .external_lex_state = 2},
  [4862] = {.lex_state = 6, .external_lex_state = 2},
  [4863] = {.lex_state = 10, .external_lex_state = 2},
  [4864] = {.lex_state = 47, .external_lex_state = 2},
  [4865] = {.lex_state = 10, .external_lex_state = 2},
  [4866] = {.lex_state = 71, .external_lex_state = 2},
  [4867] = {.lex_state = 10, .external_lex_state = 2},
  [4868] = {.lex_state = 10, .external_lex_state = 2},
  [4869] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4870] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4871] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4872] = {.lex_state = 10, .external_lex_state = 2},
  [4873] = {.lex_state = 71, .external_lex_state = 2},
  [4874] = {.lex_state = 10, .external_lex_state = 2},
  [4875] = {.lex_state = 10, .external_lex_state = 2},
  [4876] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4877] = {.lex_state = 151, .external_lex_state = 2},
  [4878] = {.lex_state = 10, .external_lex_state = 2},
  [4879] = {.lex_state = 10, .external_lex_state = 2},
  [4880] = {.lex_state = 10, .external_lex_state = 2},
  [4881] = {.lex_state = 10, .external_lex_state = 2},
  [4882] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4883] = {.lex_state = 10, .external_lex_state = 2},
  [4884] = {.lex_state = 10, .external_lex_state = 2},
  [4885] = {.lex_state = 10, .external_lex_state = 2},
  [4886] = {.lex_state = 10, .external_lex_state = 2},
  [4887] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4888] = {.lex_state = 10, .external_lex_state = 2},
  [4889] = {.lex_state = 10, .external_lex_state = 2},
  [4890] = {.lex_state = 47, .external_lex_state = 2},
  [4891] = {.lex_state = 47, .external_lex_state = 2},
  [4892] = {.lex_state = 71, .external_lex_state = 2},
  [4893] = {.lex_state = 47, .external_lex_state = 2},
  [4894] = {.lex_state = 10, .external_lex_state = 2},
  [4895] = {.lex_state = 47, .external_lex_state = 2},
  [4896] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4897] = {.lex_state = 10, .external_lex_state = 2},
  [4898] = {.lex_state = 10, .external_lex_state = 2},
  [4899] = {.lex_state = 10, .external_lex_state = 2},
  [4900] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 92},
  [4901] = {.lex_state = 10, .external_lex_state = 2},
  [4902] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4903] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4904] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4905] = {.lex_state = 10, .external_lex_state = 2},
  [4906] = {.lex_state = 47, .external_lex_state = 2},
  [4907] = {.lex_state = 146, .external_lex_state = 2},
  [4908] = {.lex_state = 71, .external_lex_state = 2},
  [4909] = {.lex_state = 151, .external_lex_state = 2},
  [4910] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4911] = {.lex_state = 10, .external_lex_state = 2},
  [4912] = {.lex_state = 146, .external_lex_state = 2},
  [4913] = {.lex_state = 151, .external_lex_state = 2},
  [4914] = {.lex_state = 10, .external_lex_state = 2},
  [4915] = {.lex_state = 10, .external_lex_state = 2},
  [4916] = {.lex_state = 10, .external_lex_state = 2},
  [4917] = {.lex_state = 10, .external_lex_state = 2},
  [4918] = {.lex_state = 10, .external_lex_state = 2},
  [4919] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4920] = {.lex_state = 10, .external_lex_state = 2},
  [4921] = {.lex_state = 10, .external_lex_state = 2},
  [4922] = {.lex_state = 10, .external_lex_state = 2},
  [4923] = {.lex_state = 6, .external_lex_state = 2},
  [4924] = {.lex_state = 6, .external_lex_state = 2},
  [4925] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4926] = {.lex_state = 6, .external_lex_state = 2},
  [4927] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4928] = {.lex_state = 151, .external_lex_state = 2},
  [4929] = {.lex_state = 71, .external_lex_state = 2},
  [4930] = {.lex_state = 10, .external_lex_state = 2},
  [4931] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4932] = {.lex_state = 71, .external_lex_state = 2},
  [4933] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4934] = {.lex_state = 151, .external_lex_state = 2},
  [4935] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4936] = {.lex_state = 146, .external_lex_state = 2},
  [4937] = {.lex_state = 10, .external_lex_state = 2},
  [4938] = {.lex_state = 10, .external_lex_state = 2},
  [4939] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4940] = {.lex_state = 10, .external_lex_state = 2},
  [4941] = {.lex_state = 71, .external_lex_state = 2},
  [4942] = {.lex_state = 10, .external_lex_state = 2},
  [4943] = {.lex_state = 151, .external_lex_state = 2},
  [4944] = {.lex_state = 10, .external_lex_state = 2},
  [4945] = {.lex_state = 146, .external_lex_state = 2},
  [4946] = {.lex_state = 47, .external_lex_state = 2},
  [4947] = {.lex_state = 10, .external_lex_state = 2},
  [4948] = {.lex_state = 47, .external_lex_state = 2},
  [4949] = {.lex_state = 47, .external_lex_state = 2},
  [4950] = {.lex_state = 47, .external_lex_state = 2},
  [4951] = {.lex_state = 47, .external_lex_state = 2},
  [4952] = {.lex_state = 47, .external_lex_state = 2},
  [4953] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4954] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4955] = {.lex_state = 146, .external_lex_state = 2},
  [4956] = {.lex_state = 151, .external_lex_state = 2},
  [4957] = {.lex_state = 151, .external_lex_state = 2},
  [4958] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4959] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4960] = {.lex_state = 10, .external_lex_state = 2},
  [4961] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4962] = {.lex_state = 10, .external_lex_state = 2},
  [4963] = {.lex_state = 10, .external_lex_state = 2},
  [4964] = {.lex_state = 6, .external_lex_state = 2},
  [4965] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4966] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4967] = {.lex_state = 151, .external_lex_state = 2},
  [4968] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4969] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4970] = {.lex_state = 10, .external_lex_state = 2},
  [4971] = {.lex_state = 10, .external_lex_state = 2},
  [4972] = {.lex_state = 10, .external_lex_state = 2},
  [4973] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4974] = {.lex_state = 146, .external_lex_state = 2},
  [4975] = {.lex_state = 10, .external_lex_state = 2},
  [4976] = {.lex_state = 10, .external_lex_state = 2},
  [4977] = {.lex_state = 71, .external_lex_state = 2},
  [4978] = {.lex_state = 10, .external_lex_state = 2},
  [4979] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4980] = {.lex_state = 10, .external_lex_state = 2},
  [4981] = {.lex_state = 71, .external_lex_state = 2},
  [4982] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4983] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4984] = {.lex_state = 47, .external_lex_state = 2},
  [4985] = {.lex_state = 10, .external_lex_state = 2},
  [4986] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4987] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4988] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4989] = {.lex_state = 10, .external_lex_state = 2},
  [4990] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4991] = {.lex_state = 10, .external_lex_state = 2},
  [4992] = {.lex_state = 47, .external_lex_state = 2},
  [4993] = {.lex_state = 47, .external_lex_state = 2},
  [4994] = {.lex_state = 47, .external_lex_state = 2},
  [4995] = {.lex_state = 47, .external_lex_state = 2},
  [4996] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4997] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [4998] = {.lex_state = 10, .external_lex_state = 2},
  [4999] = {.lex_state = 146, .external_lex_state = 2},
  [5000] = {.lex_state = 10, .external_lex_state = 2},
  [5001] = {.lex_state = 10, .external_lex_state = 2},
  [5002] = {.lex_state = 10, .external_lex_state = 2},
  [5003] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5004] = {.lex_state = 151, .external_lex_state = 2},
  [5005] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5006] = {.lex_state = 10, .external_lex_state = 2},
  [5007] = {.lex_state = 10, .external_lex_state = 2},
  [5008] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5009] = {.lex_state = 10, .external_lex_state = 2},
  [5010] = {.lex_state = 10, .external_lex_state = 2},
  [5011] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5012] = {.lex_state = 151, .external_lex_state = 2},
  [5013] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5014] = {.lex_state = 10, .external_lex_state = 2},
  [5015] = {.lex_state = 151, .external_lex_state = 2},
  [5016] = {.lex_state = 71, .external_lex_state = 2},
  [5017] = {.lex_state = 6, .external_lex_state = 2},
  [5018] = {.lex_state = 71, .external_lex_state = 2},
  [5019] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5020] = {.lex_state = 146, .external_lex_state = 2},
  [5021] = {.lex_state = 146, .external_lex_state = 2},
  [5022] = {.lex_state = 10, .external_lex_state = 2},
  [5023] = {.lex_state = 47, .external_lex_state = 2},
  [5024] = {.lex_state = 6, .external_lex_state = 2},
  [5025] = {.lex_state = 10, .external_lex_state = 2},
  [5026] = {.lex_state = 151, .external_lex_state = 2},
  [5027] = {.lex_state = 151, .external_lex_state = 2},
  [5028] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5029] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5030] = {.lex_state = 10, .external_lex_state = 2},
  [5031] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5032] = {.lex_state = 10, .external_lex_state = 2},
  [5033] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5034] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5035] = {.lex_state = 151, .external_lex_state = 2},
  [5036] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5037] = {.lex_state = 71, .external_lex_state = 2},
  [5038] = {.lex_state = 10, .external_lex_state = 2},
  [5039] = {.lex_state = 10, .external_lex_state = 2},
  [5040] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5041] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5042] = {.lex_state = 151, .external_lex_state = 2},
  [5043] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5044] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5045] = {.lex_state = 146, .external_lex_state = 2},
  [5046] = {.lex_state = 151, .external_lex_state = 2},
  [5047] = {.lex_state = 151, .external_lex_state = 2},
  [5048] = {.lex_state = 146, .external_lex_state = 3},
  [5049] = {.lex_state = 146, .external_lex_state = 2},
  [5050] = {.lex_state = 6, .external_lex_state = 2},
  [5051] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5052] = {.lex_state = 151, .external_lex_state = 2},
  [5053] = {.lex_state = 151, .external_lex_state = 2},
  [5054] = {.lex_state = 151, .external_lex_state = 2},
  [5055] = {.lex_state = 151, .external_lex_state = 2},
  [5056] = {.lex_state = 146, .external_lex_state = 2},
  [5057] = {.lex_state = 151, .external_lex_state = 2},
  [5058] = {.lex_state = 151, .external_lex_state = 2},
  [5059] = {.lex_state = 151, .external_lex_state = 2},
  [5060] = {.lex_state = 151, .external_lex_state = 2},
  [5061] = {.lex_state = 151, .external_lex_state = 2},
  [5062] = {.lex_state = 151, .external_lex_state = 2},
  [5063] = {.lex_state = 6, .external_lex_state = 2},
  [5064] = {.lex_state = 151, .external_lex_state = 2},
  [5065] = {.lex_state = 151, .external_lex_state = 2},
  [5066] = {.lex_state = 151, .external_lex_state = 2},
  [5067] = {.lex_state = 151, .external_lex_state = 2},
  [5068] = {.lex_state = 151, .external_lex_state = 2},
  [5069] = {.lex_state = 61, .external_lex_state = 2},
  [5070] = {.lex_state = 151, .external_lex_state = 2},
  [5071] = {.lex_state = 151, .external_lex_state = 2},
  [5072] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5073] = {.lex_state = 151, .external_lex_state = 2},
  [5074] = {.lex_state = 151, .external_lex_state = 2},
  [5075] = {.lex_state = 151, .external_lex_state = 2},
  [5076] = {.lex_state = 146, .external_lex_state = 2},
  [5077] = {.lex_state = 151, .external_lex_state = 2},
  [5078] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5079] = {.lex_state = 151, .external_lex_state = 2},
  [5080] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5081] = {.lex_state = 146, .external_lex_state = 2},
  [5082] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5083] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5084] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5085] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5086] = {.lex_state = 151, .external_lex_state = 2},
  [5087] = {.lex_state = 146, .external_lex_state = 2},
  [5088] = {.lex_state = 148, .external_lex_state = 2},
  [5089] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5090] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5091] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5092] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5093] = {.lex_state = 146, .external_lex_state = 2},
  [5094] = {.lex_state = 151, .external_lex_state = 2},
  [5095] = {.lex_state = 151, .external_lex_state = 2},
  [5096] = {.lex_state = 6, .external_lex_state = 2},
  [5097] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5098] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5099] = {.lex_state = 146, .external_lex_state = 2},
  [5100] = {.lex_state = 146, .external_lex_state = 2},
  [5101] = {.lex_state = 151, .external_lex_state = 2},
  [5102] = {.lex_state = 146, .external_lex_state = 3},
  [5103] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5104] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5105] = {.lex_state = 151, .external_lex_state = 2},
  [5106] = {.lex_state = 148, .external_lex_state = 2},
  [5107] = {.lex_state = 151, .external_lex_state = 2},
  [5108] = {.lex_state = 151, .external_lex_state = 2},
  [5109] = {.lex_state = 151, .external_lex_state = 2},
  [5110] = {.lex_state = 151, .external_lex_state = 2},
  [5111] = {.lex_state = 146, .external_lex_state = 3},
  [5112] = {.lex_state = 146, .external_lex_state = 2},
  [5113] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5114] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5115] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5116] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5117] = {.lex_state = 151, .external_lex_state = 2},
  [5118] = {.lex_state = 151, .external_lex_state = 2},
  [5119] = {.lex_state = 151, .external_lex_state = 2},
  [5120] = {.lex_state = 151, .external_lex_state = 2},
  [5121] = {.lex_state = 6, .external_lex_state = 2},
  [5122] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5123] = {.lex_state = 151, .external_lex_state = 2},
  [5124] = {.lex_state = 151, .external_lex_state = 2},
  [5125] = {.lex_state = 151, .external_lex_state = 2},
  [5126] = {.lex_state = 151, .external_lex_state = 2},
  [5127] = {.lex_state = 151, .external_lex_state = 2},
  [5128] = {.lex_state = 151, .external_lex_state = 2},
  [5129] = {.lex_state = 151, .external_lex_state = 2},
  [5130] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5131] = {.lex_state = 146, .external_lex_state = 2},
  [5132] = {.lex_state = 151, .external_lex_state = 2},
  [5133] = {.lex_state = 151, .external_lex_state = 2},
  [5134] = {.lex_state = 151, .external_lex_state = 2},
  [5135] = {.lex_state = 151, .external_lex_state = 2},
  [5136] = {.lex_state = 151, .external_lex_state = 2},
  [5137] = {.lex_state = 151, .external_lex_state = 2},
  [5138] = {.lex_state = 151, .external_lex_state = 2},
  [5139] = {.lex_state = 151, .external_lex_state = 2},
  [5140] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5141] = {.lex_state = 61, .external_lex_state = 2},
  [5142] = {.lex_state = 148, .external_lex_state = 2},
  [5143] = {.lex_state = 151, .external_lex_state = 2},
  [5144] = {.lex_state = 10, .external_lex_state = 2},
  [5145] = {.lex_state = 6, .external_lex_state = 2},
  [5146] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5147] = {.lex_state = 148, .external_lex_state = 2},
  [5148] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5149] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5150] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5151] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5152] = {.lex_state = 6, .external_lex_state = 2},
  [5153] = {.lex_state = 151, .external_lex_state = 2},
  [5154] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5155] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5156] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5157] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5158] = {.lex_state = 151, .external_lex_state = 2},
  [5159] = {.lex_state = 6, .external_lex_state = 2},
  [5160] = {.lex_state = 151, .external_lex_state = 2},
  [5161] = {.lex_state = 146, .external_lex_state = 3},
  [5162] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5163] = {.lex_state = 151, .external_lex_state = 2},
  [5164] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5165] = {.lex_state = 151, .external_lex_state = 2},
  [5166] = {.lex_state = 151, .external_lex_state = 2},
  [5167] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5168] = {.lex_state = 151, .external_lex_state = 2},
  [5169] = {.lex_state = 6, .external_lex_state = 2},
  [5170] = {.lex_state = 151, .external_lex_state = 2},
  [5171] = {.lex_state = 146, .external_lex_state = 2},
  [5172] = {.lex_state = 151, .external_lex_state = 2},
  [5173] = {.lex_state = 146, .external_lex_state = 2},
  [5174] = {.lex_state = 151, .external_lex_state = 2},
  [5175] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5176] = {.lex_state = 151, .external_lex_state = 2},
  [5177] = {.lex_state = 151, .external_lex_state = 2},
  [5178] = {.lex_state = 151, .external_lex_state = 2},
  [5179] = {.lex_state = 151, .external_lex_state = 2},
  [5180] = {.lex_state = 151, .external_lex_state = 2},
  [5181] = {.lex_state = 151, .external_lex_state = 2},
  [5182] = {.lex_state = 151, .external_lex_state = 2},
  [5183] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5184] = {.lex_state = 6, .external_lex_state = 2},
  [5185] = {.lex_state = 151, .external_lex_state = 2},
  [5186] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5187] = {.lex_state = 151, .external_lex_state = 2},
  [5188] = {.lex_state = 146, .external_lex_state = 2},
  [5189] = {.lex_state = 151, .external_lex_state = 2},
  [5190] = {.lex_state = 151, .external_lex_state = 2},
  [5191] = {.lex_state = 151, .external_lex_state = 2},
  [5192] = {.lex_state = 151, .external_lex_state = 2},
  [5193] = {.lex_state = 146, .external_lex_state = 2},
  [5194] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5195] = {.lex_state = 151, .external_lex_state = 2},
  [5196] = {.lex_state = 148, .external_lex_state = 2},
  [5197] = {.lex_state = 6, .external_lex_state = 2},
  [5198] = {.lex_state = 151, .external_lex_state = 2},
  [5199] = {.lex_state = 151, .external_lex_state = 2},
  [5200] = {.lex_state = 151, .external_lex_state = 2},
  [5201] = {.lex_state = 151, .external_lex_state = 2},
  [5202] = {.lex_state = 151, .external_lex_state = 2},
  [5203] = {.lex_state = 151, .external_lex_state = 2},
  [5204] = {.lex_state = 151, .external_lex_state = 2},
  [5205] = {.lex_state = 151, .external_lex_state = 2},
  [5206] = {.lex_state = 151, .external_lex_state = 2},
  [5207] = {.lex_state = 151, .external_lex_state = 2},
  [5208] = {.lex_state = 151, .external_lex_state = 2},
  [5209] = {.lex_state = 151, .external_lex_state = 2},
  [5210] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5211] = {.lex_state = 151, .external_lex_state = 2},
  [5212] = {.lex_state = 151, .external_lex_state = 2},
  [5213] = {.lex_state = 151, .external_lex_state = 2},
  [5214] = {.lex_state = 71, .external_lex_state = 2},
  [5215] = {.lex_state = 148, .external_lex_state = 2},
  [5216] = {.lex_state = 6, .external_lex_state = 2},
  [5217] = {.lex_state = 151, .external_lex_state = 2},
  [5218] = {.lex_state = 151, .external_lex_state = 2},
  [5219] = {.lex_state = 6, .external_lex_state = 2},
  [5220] = {.lex_state = 151, .external_lex_state = 2},
  [5221] = {.lex_state = 151, .external_lex_state = 2},
  [5222] = {.lex_state = 151, .external_lex_state = 2},
  [5223] = {.lex_state = 151, .external_lex_state = 2},
  [5224] = {.lex_state = 146, .external_lex_state = 3},
  [5225] = {.lex_state = 146, .external_lex_state = 3},
  [5226] = {.lex_state = 151, .external_lex_state = 2},
  [5227] = {.lex_state = 146, .external_lex_state = 2},
  [5228] = {.lex_state = 151, .external_lex_state = 2},
  [5229] = {.lex_state = 148, .external_lex_state = 2},
  [5230] = {.lex_state = 61, .external_lex_state = 2},
  [5231] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5232] = {.lex_state = 151, .external_lex_state = 2},
  [5233] = {.lex_state = 151, .external_lex_state = 2},
  [5234] = {.lex_state = 148, .external_lex_state = 2},
  [5235] = {.lex_state = 151, .external_lex_state = 2},
  [5236] = {.lex_state = 151, .external_lex_state = 2},
  [5237] = {.lex_state = 151, .external_lex_state = 2},
  [5238] = {.lex_state = 151, .external_lex_state = 2},
  [5239] = {.lex_state = 151, .external_lex_state = 2},
  [5240] = {.lex_state = 6, .external_lex_state = 2},
  [5241] = {.lex_state = 148, .external_lex_state = 2},
  [5242] = {.lex_state = 151, .external_lex_state = 2},
  [5243] = {.lex_state = 6, .external_lex_state = 2},
  [5244] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5245] = {.lex_state = 151, .external_lex_state = 2},
  [5246] = {.lex_state = 151, .external_lex_state = 2},
  [5247] = {.lex_state = 151, .external_lex_state = 2},
  [5248] = {.lex_state = 151, .external_lex_state = 2},
  [5249] = {.lex_state = 151, .external_lex_state = 2},
  [5250] = {.lex_state = 6, .external_lex_state = 2},
  [5251] = {.lex_state = 6, .external_lex_state = 2},
  [5252] = {.lex_state = 6, .external_lex_state = 2},
  [5253] = {.lex_state = 6, .external_lex_state = 2},
  [5254] = {.lex_state = 6, .external_lex_state = 2},
  [5255] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5256] = {.lex_state = 6, .external_lex_state = 2},
  [5257] = {.lex_state = 151, .external_lex_state = 2},
  [5258] = {.lex_state = 151, .external_lex_state = 2},
  [5259] = {.lex_state = 151, .external_lex_state = 2},
  [5260] = {.lex_state = 151, .external_lex_state = 2},
  [5261] = {.lex_state = 151, .external_lex_state = 2},
  [5262] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 93},
  [5263] = {.lex_state = 6, .external_lex_state = 2},
  [5264] = {.lex_state = 146, .external_lex_state = 2},
  [5265] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5266] = {.lex_state = 146, .external_lex_state = 2},
  [5267] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5268] = {.lex_state = 148, .external_lex_state = 2},
  [5269] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5270] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5271] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5272] = {.lex_state = 40, .external_lex_state = 2},
  [5273] = {.lex_state = 148, .external_lex_state = 2},
  [5274] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5275] = {.lex_state = 146, .external_lex_state = 2},
  [5276] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5277] = {.lex_state = 146, .external_lex_state = 2},
  [5278] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5279] = {.lex_state = 146, .external_lex_state = 2},
  [5280] = {.lex_state = 148, .external_lex_state = 2},
  [5281] = {.lex_state = 146, .external_lex_state = 2},
  [5282] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5283] = {.lex_state = 148, .external_lex_state = 2},
  [5284] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5285] = {.lex_state = 148, .external_lex_state = 2},
  [5286] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5287] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5288] = {.lex_state = 6, .external_lex_state = 2},
  [5289] = {.lex_state = 146, .external_lex_state = 2},
  [5290] = {.lex_state = 148, .external_lex_state = 2},
  [5291] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5292] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5293] = {.lex_state = 146, .external_lex_state = 2},
  [5294] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5295] = {.lex_state = 148, .external_lex_state = 2},
  [5296] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 93},
  [5297] = {.lex_state = 148, .external_lex_state = 2},
  [5298] = {.lex_state = 6, .external_lex_state = 2},
  [5299] = {.lex_state = 71, .external_lex_state = 2},
  [5300] = {.lex_state = 148, .external_lex_state = 2},
  [5301] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5302] = {.lex_state = 40, .external_lex_state = 2},
  [5303] = {.lex_state = 148, .external_lex_state = 2},
  [5304] = {.lex_state = 148, .external_lex_state = 2},
  [5305] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5306] = {.lex_state = 40, .external_lex_state = 2},
  [5307] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5308] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5309] = {.lex_state = 148, .external_lex_state = 2},
  [5310] = {.lex_state = 47, .external_lex_state = 2},
  [5311] = {.lex_state = 148, .external_lex_state = 2},
  [5312] = {.lex_state = 146, .external_lex_state = 2},
  [5313] = {.lex_state = 146, .external_lex_state = 2},
  [5314] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5315] = {.lex_state = 146, .external_lex_state = 2},
  [5316] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5317] = {.lex_state = 148, .external_lex_state = 2},
  [5318] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5319] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5320] = {.lex_state = 146, .external_lex_state = 2},
  [5321] = {.lex_state = 146, .external_lex_state = 2},
  [5322] = {.lex_state = 6, .external_lex_state = 2},
  [5323] = {.lex_state = 148, .external_lex_state = 2},
  [5324] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5325] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5326] = {.lex_state = 146, .external_lex_state = 2},
  [5327] = {.lex_state = 146, .external_lex_state = 2},
  [5328] = {.lex_state = 146, .external_lex_state = 2},
  [5329] = {.lex_state = 148, .external_lex_state = 2},
  [5330] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5331] = {.lex_state = 40, .external_lex_state = 2},
  [5332] = {.lex_state = 146, .external_lex_state = 2},
  [5333] = {.lex_state = 6, .external_lex_state = 2},
  [5334] = {.lex_state = 151, .external_lex_state = 2},
  [5335] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5336] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5337] = {.lex_state = 151, .external_lex_state = 2},
  [5338] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 93},
  [5339] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5340] = {.lex_state = 148, .external_lex_state = 2},
  [5341] = {.lex_state = 146, .external_lex_state = 2},
  [5342] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5343] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5344] = {.lex_state = 148, .external_lex_state = 2},
  [5345] = {.lex_state = 148, .external_lex_state = 2},
  [5346] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5347] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5348] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5349] = {.lex_state = 146, .external_lex_state = 2},
  [5350] = {.lex_state = 6, .external_lex_state = 2},
  [5351] = {.lex_state = 6, .external_lex_state = 2},
  [5352] = {.lex_state = 146, .external_lex_state = 2},
  [5353] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5354] = {.lex_state = 11, .external_lex_state = 2},
  [5355] = {.lex_state = 6, .external_lex_state = 2},
  [5356] = {.lex_state = 40, .external_lex_state = 2},
  [5357] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5358] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5359] = {.lex_state = 146, .external_lex_state = 2},
  [5360] = {.lex_state = 146, .external_lex_state = 2},
  [5361] = {.lex_state = 6, .external_lex_state = 2},
  [5362] = {.lex_state = 148, .external_lex_state = 2},
  [5363] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5364] = {.lex_state = 146, .external_lex_state = 2},
  [5365] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5366] = {.lex_state = 47, .external_lex_state = 2},
  [5367] = {.lex_state = 148, .external_lex_state = 2},
  [5368] = {.lex_state = 6, .external_lex_state = 2},
  [5369] = {.lex_state = 6, .external_lex_state = 2},
  [5370] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5371] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5372] = {.lex_state = 40, .external_lex_state = 2},
  [5373] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5374] = {.lex_state = 40, .external_lex_state = 2},
  [5375] = {.lex_state = 146, .external_lex_state = 2},
  [5376] = {.lex_state = 146, .external_lex_state = 2},
  [5377] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5378] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5379] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5380] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5381] = {.lex_state = 6, .external_lex_state = 2},
  [5382] = {.lex_state = 6, .external_lex_state = 2},
  [5383] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5384] = {.lex_state = 146, .external_lex_state = 2},
  [5385] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5386] = {.lex_state = 6, .external_lex_state = 2},
  [5387] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 93},
  [5388] = {.lex_state = 148, .external_lex_state = 2},
  [5389] = {.lex_state = 98, .external_lex_state = 8},
  [5390] = {.lex_state = 6, .external_lex_state = 2},
  [5391] = {.lex_state = 148, .external_lex_state = 2},
  [5392] = {.lex_state = 148, .external_lex_state = 2},
  [5393] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5394] = {.lex_state = 148, .external_lex_state = 2},
  [5395] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5396] = {.lex_state = 148, .external_lex_state = 2},
  [5397] = {.lex_state = 6, .external_lex_state = 2},
  [5398] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5399] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5400] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5401] = {.lex_state = 151, .external_lex_state = 2},
  [5402] = {.lex_state = 146, .external_lex_state = 2},
  [5403] = {.lex_state = 151, .external_lex_state = 2},
  [5404] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5405] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5406] = {.lex_state = 6, .external_lex_state = 2},
  [5407] = {.lex_state = 146, .external_lex_state = 2},
  [5408] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5409] = {.lex_state = 148, .external_lex_state = 2},
  [5410] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5411] = {.lex_state = 47, .external_lex_state = 2},
  [5412] = {.lex_state = 146, .external_lex_state = 2},
  [5413] = {.lex_state = 6, .external_lex_state = 2},
  [5414] = {.lex_state = 6, .external_lex_state = 2},
  [5415] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5416] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5417] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5418] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5419] = {.lex_state = 148, .external_lex_state = 2},
  [5420] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5421] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5422] = {.lex_state = 47, .external_lex_state = 2},
  [5423] = {.lex_state = 148, .external_lex_state = 2},
  [5424] = {.lex_state = 148, .external_lex_state = 2},
  [5425] = {.lex_state = 6, .external_lex_state = 2},
  [5426] = {.lex_state = 151, .external_lex_state = 2},
  [5427] = {.lex_state = 148, .external_lex_state = 2},
  [5428] = {.lex_state = 146, .external_lex_state = 2},
  [5429] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5430] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5431] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5432] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5433] = {.lex_state = 47, .external_lex_state = 2},
  [5434] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5435] = {.lex_state = 148, .external_lex_state = 2},
  [5436] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5437] = {.lex_state = 151, .external_lex_state = 2},
  [5438] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5439] = {.lex_state = 6, .external_lex_state = 2},
  [5440] = {.lex_state = 146, .external_lex_state = 2},
  [5441] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5442] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5443] = {.lex_state = 148, .external_lex_state = 2},
  [5444] = {.lex_state = 146, .external_lex_state = 2},
  [5445] = {.lex_state = 146, .external_lex_state = 2},
  [5446] = {.lex_state = 146, .external_lex_state = 2},
  [5447] = {.lex_state = 151, .external_lex_state = 2},
  [5448] = {.lex_state = 151, .external_lex_state = 2},
  [5449] = {.lex_state = 151, .external_lex_state = 2},
  [5450] = {.lex_state = 148, .external_lex_state = 2},
  [5451] = {.lex_state = 47, .external_lex_state = 2},
  [5452] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5453] = {.lex_state = 98, .external_lex_state = 8},
  [5454] = {.lex_state = 6, .external_lex_state = 2},
  [5455] = {.lex_state = 148, .external_lex_state = 2},
  [5456] = {.lex_state = 148, .external_lex_state = 2},
  [5457] = {.lex_state = 6, .external_lex_state = 2},
  [5458] = {.lex_state = 148, .external_lex_state = 2},
  [5459] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5460] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5461] = {.lex_state = 148, .external_lex_state = 2},
  [5462] = {.lex_state = 148, .external_lex_state = 2},
  [5463] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5464] = {.lex_state = 71, .external_lex_state = 2},
  [5465] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5466] = {.lex_state = 71, .external_lex_state = 2},
  [5467] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5468] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5469] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5470] = {.lex_state = 148, .external_lex_state = 2},
  [5471] = {.lex_state = 146, .external_lex_state = 2},
  [5472] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5473] = {.lex_state = 6, .external_lex_state = 2},
  [5474] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5475] = {.lex_state = 146, .external_lex_state = 2},
  [5476] = {.lex_state = 148, .external_lex_state = 2},
  [5477] = {.lex_state = 6, .external_lex_state = 2},
  [5478] = {.lex_state = 146, .external_lex_state = 3},
  [5479] = {.lex_state = 71, .external_lex_state = 2},
  [5480] = {.lex_state = 146, .external_lex_state = 2},
  [5481] = {.lex_state = 98, .external_lex_state = 8},
  [5482] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5483] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5484] = {.lex_state = 146, .external_lex_state = 2},
  [5485] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5486] = {.lex_state = 146, .external_lex_state = 2},
  [5487] = {.lex_state = 146, .external_lex_state = 2},
  [5488] = {.lex_state = 146, .external_lex_state = 2},
  [5489] = {.lex_state = 40, .external_lex_state = 2},
  [5490] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5491] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5492] = {.lex_state = 148, .external_lex_state = 3},
  [5493] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5494] = {.lex_state = 40, .external_lex_state = 2},
  [5495] = {.lex_state = 146, .external_lex_state = 2},
  [5496] = {.lex_state = 146, .external_lex_state = 2},
  [5497] = {.lex_state = 148, .external_lex_state = 2},
  [5498] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5499] = {.lex_state = 148, .external_lex_state = 2},
  [5500] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5501] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5502] = {.lex_state = 47, .external_lex_state = 2},
  [5503] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5504] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5505] = {.lex_state = 47, .external_lex_state = 2},
  [5506] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5507] = {.lex_state = 146, .external_lex_state = 2},
  [5508] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5509] = {.lex_state = 151, .external_lex_state = 2},
  [5510] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5511] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5512] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5513] = {.lex_state = 151, .external_lex_state = 2},
  [5514] = {.lex_state = 146, .external_lex_state = 2},
  [5515] = {.lex_state = 6, .external_lex_state = 2},
  [5516] = {.lex_state = 151, .external_lex_state = 2},
  [5517] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5518] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 92},
  [5519] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5520] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5521] = {.lex_state = 151, .external_lex_state = 2},
  [5522] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5523] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5524] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5525] = {.lex_state = 146, .external_lex_state = 3},
  [5526] = {.lex_state = 151, .external_lex_state = 2},
  [5527] = {.lex_state = 146, .external_lex_state = 2},
  [5528] = {.lex_state = 47, .external_lex_state = 2},
  [5529] = {.lex_state = 148, .external_lex_state = 2},
  [5530] = {.lex_state = 146, .external_lex_state = 2},
  [5531] = {.lex_state = 146, .external_lex_state = 2},
  [5532] = {.lex_state = 146, .external_lex_state = 2},
  [5533] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5534] = {.lex_state = 151, .external_lex_state = 2},
  [5535] = {.lex_state = 47, .external_lex_state = 2},
  [5536] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5537] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5538] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5539] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5540] = {.lex_state = 6, .external_lex_state = 2},
  [5541] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5542] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5543] = {.lex_state = 98, .external_lex_state = 8},
  [5544] = {.lex_state = 6, .external_lex_state = 2},
  [5545] = {.lex_state = 40, .external_lex_state = 2},
  [5546] = {.lex_state = 11, .external_lex_state = 2},
  [5547] = {.lex_state = 146, .external_lex_state = 2},
  [5548] = {.lex_state = 146, .external_lex_state = 2},
  [5549] = {.lex_state = 146, .external_lex_state = 2},
  [5550] = {.lex_state = 11, .external_lex_state = 2},
  [5551] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5552] = {.lex_state = 146, .external_lex_state = 2},
  [5553] = {.lex_state = 11, .external_lex_state = 2},
  [5554] = {.lex_state = 61, .external_lex_state = 2},
  [5555] = {.lex_state = 11, .external_lex_state = 2},
  [5556] = {.lex_state = 146, .external_lex_state = 2},
  [5557] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5558] = {.lex_state = 11, .external_lex_state = 2},
  [5559] = {.lex_state = 11, .external_lex_state = 2},
  [5560] = {.lex_state = 6, .external_lex_state = 2},
  [5561] = {.lex_state = 151, .external_lex_state = 2},
  [5562] = {.lex_state = 11, .external_lex_state = 2},
  [5563] = {.lex_state = 6, .external_lex_state = 2},
  [5564] = {.lex_state = 11, .external_lex_state = 2},
  [5565] = {.lex_state = 148, .external_lex_state = 2},
  [5566] = {.lex_state = 11, .external_lex_state = 2},
  [5567] = {.lex_state = 11, .external_lex_state = 2},
  [5568] = {.lex_state = 11, .external_lex_state = 2},
  [5569] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5570] = {.lex_state = 151, .external_lex_state = 2},
  [5571] = {.lex_state = 146, .external_lex_state = 2},
  [5572] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5573] = {.lex_state = 11, .external_lex_state = 2},
  [5574] = {.lex_state = 61, .external_lex_state = 2},
  [5575] = {.lex_state = 11, .external_lex_state = 2},
  [5576] = {.lex_state = 148, .external_lex_state = 2},
  [5577] = {.lex_state = 11, .external_lex_state = 2},
  [5578] = {.lex_state = 57, .external_lex_state = 2},
  [5579] = {.lex_state = 47, .external_lex_state = 2},
  [5580] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5581] = {.lex_state = 61, .external_lex_state = 2},
  [5582] = {.lex_state = 151, .external_lex_state = 2},
  [5583] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5584] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5585] = {.lex_state = 6, .external_lex_state = 2},
  [5586] = {.lex_state = 146, .external_lex_state = 2},
  [5587] = {.lex_state = 146, .external_lex_state = 2},
  [5588] = {.lex_state = 146, .external_lex_state = 2},
  [5589] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5590] = {.lex_state = 61, .external_lex_state = 2},
  [5591] = {.lex_state = 6, .external_lex_state = 2},
  [5592] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5593] = {.lex_state = 98, .external_lex_state = 8},
  [5594] = {.lex_state = 146, .external_lex_state = 2},
  [5595] = {.lex_state = 11, .external_lex_state = 2},
  [5596] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5597] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5598] = {.lex_state = 11, .external_lex_state = 2},
  [5599] = {.lex_state = 146, .external_lex_state = 2},
  [5600] = {.lex_state = 11, .external_lex_state = 2},
  [5601] = {.lex_state = 6, .external_lex_state = 2},
  [5602] = {.lex_state = 146, .external_lex_state = 2},
  [5603] = {.lex_state = 146, .external_lex_state = 2},
  [5604] = {.lex_state = 146, .external_lex_state = 2},
  [5605] = {.lex_state = 11, .external_lex_state = 2},
  [5606] = {.lex_state = 98, .external_lex_state = 8},
  [5607] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5608] = {.lex_state = 146, .external_lex_state = 2},
  [5609] = {.lex_state = 11, .external_lex_state = 2},
  [5610] = {.lex_state = 6, .external_lex_state = 2},
  [5611] = {.lex_state = 146, .external_lex_state = 2},
  [5612] = {.lex_state = 11, .external_lex_state = 2},
  [5613] = {.lex_state = 98, .external_lex_state = 8},
  [5614] = {.lex_state = 6, .external_lex_state = 2},
  [5615] = {.lex_state = 11, .external_lex_state = 2},
  [5616] = {.lex_state = 146, .external_lex_state = 2},
  [5617] = {.lex_state = 151, .external_lex_state = 2},
  [5618] = {.lex_state = 11, .external_lex_state = 2},
  [5619] = {.lex_state = 146, .external_lex_state = 2},
  [5620] = {.lex_state = 146, .external_lex_state = 2},
  [5621] = {.lex_state = 13, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5622] = {.lex_state = 146, .external_lex_state = 2},
  [5623] = {.lex_state = 11, .external_lex_state = 2},
  [5624] = {.lex_state = 148, .external_lex_state = 2},
  [5625] = {.lex_state = 146, .external_lex_state = 2},
  [5626] = {.lex_state = 26, .external_lex_state = 2},
  [5627] = {.lex_state = 40, .external_lex_state = 2},
  [5628] = {.lex_state = 146, .external_lex_state = 2},
  [5629] = {.lex_state = 6, .external_lex_state = 2},
  [5630] = {.lex_state = 146, .external_lex_state = 2},
  [5631] = {.lex_state = 146, .external_lex_state = 2},
  [5632] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 93},
  [5633] = {.lex_state = 146, .external_lex_state = 2},
  [5634] = {.lex_state = 11, .external_lex_state = 2},
  [5635] = {.lex_state = 11, .external_lex_state = 2},
  [5636] = {.lex_state = 146, .external_lex_state = 2},
  [5637] = {.lex_state = 146, .external_lex_state = 2},
  [5638] = {.lex_state = 11, .external_lex_state = 2},
  [5639] = {.lex_state = 11, .external_lex_state = 2},
  [5640] = {.lex_state = 26, .external_lex_state = 2},
  [5641] = {.lex_state = 11, .external_lex_state = 2},
  [5642] = {.lex_state = 151, .external_lex_state = 2},
  [5643] = {.lex_state = 146, .external_lex_state = 2},
  [5644] = {.lex_state = 148, .external_lex_state = 2},
  [5645] = {.lex_state = 146, .external_lex_state = 2},
  [5646] = {.lex_state = 151, .external_lex_state = 2},
  [5647] = {.lex_state = 11, .external_lex_state = 2},
  [5648] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5649] = {.lex_state = 6, .external_lex_state = 2},
  [5650] = {.lex_state = 146, .external_lex_state = 2},
  [5651] = {.lex_state = 146, .external_lex_state = 2},
  [5652] = {.lex_state = 11, .external_lex_state = 2},
  [5653] = {.lex_state = 151, .external_lex_state = 2},
  [5654] = {.lex_state = 11, .external_lex_state = 2},
  [5655] = {.lex_state = 146, .external_lex_state = 2},
  [5656] = {.lex_state = 11, .external_lex_state = 2},
  [5657] = {.lex_state = 6, .external_lex_state = 2},
  [5658] = {.lex_state = 11, .external_lex_state = 2},
  [5659] = {.lex_state = 151, .external_lex_state = 2},
  [5660] = {.lex_state = 11, .external_lex_state = 2},
  [5661] = {.lex_state = 11, .external_lex_state = 2},
  [5662] = {.lex_state = 151, .external_lex_state = 2},
  [5663] = {.lex_state = 61, .external_lex_state = 2},
  [5664] = {.lex_state = 146, .external_lex_state = 2},
  [5665] = {.lex_state = 146, .external_lex_state = 2},
  [5666] = {.lex_state = 148, .external_lex_state = 2},
  [5667] = {.lex_state = 151, .external_lex_state = 2},
  [5668] = {.lex_state = 11, .external_lex_state = 2},
  [5669] = {.lex_state = 11, .external_lex_state = 2},
  [5670] = {.lex_state = 6, .external_lex_state = 2},
  [5671] = {.lex_state = 151, .external_lex_state = 2},
  [5672] = {.lex_state = 151, .external_lex_state = 2},
  [5673] = {.lex_state = 146, .external_lex_state = 2},
  [5674] = {.lex_state = 146, .external_lex_state = 2},
  [5675] = {.lex_state = 11, .external_lex_state = 2},
  [5676] = {.lex_state = 151, .external_lex_state = 2},
  [5677] = {.lex_state = 151, .external_lex_state = 2},
  [5678] = {.lex_state = 61, .external_lex_state = 2},
  [5679] = {.lex_state = 11, .external_lex_state = 2},
  [5680] = {.lex_state = 11, .external_lex_state = 2},
  [5681] = {.lex_state = 11, .external_lex_state = 2},
  [5682] = {.lex_state = 146, .external_lex_state = 2},
  [5683] = {.lex_state = 148, .external_lex_state = 2},
  [5684] = {.lex_state = 11, .external_lex_state = 2},
  [5685] = {.lex_state = 26, .external_lex_state = 2},
  [5686] = {.lex_state = 11, .external_lex_state = 2},
  [5687] = {.lex_state = 148, .external_lex_state = 2},
  [5688] = {.lex_state = 146, .external_lex_state = 2},
  [5689] = {.lex_state = 6, .external_lex_state = 2},
  [5690] = {.lex_state = 61, .external_lex_state = 2},
  [5691] = {.lex_state = 151, .external_lex_state = 2},
  [5692] = {.lex_state = 11, .external_lex_state = 2},
  [5693] = {.lex_state = 11, .external_lex_state = 2},
  [5694] = {.lex_state = 11, .external_lex_state = 2},
  [5695] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5696] = {.lex_state = 11, .external_lex_state = 2},
  [5697] = {.lex_state = 11, .external_lex_state = 2},
  [5698] = {.lex_state = 11, .external_lex_state = 2},
  [5699] = {.lex_state = 11, .external_lex_state = 2},
  [5700] = {.lex_state = 61, .external_lex_state = 2},
  [5701] = {.lex_state = 11, .external_lex_state = 2},
  [5702] = {.lex_state = 98, .external_lex_state = 8},
  [5703] = {.lex_state = 11, .external_lex_state = 2},
  [5704] = {.lex_state = 6, .external_lex_state = 2},
  [5705] = {.lex_state = 11, .external_lex_state = 2},
  [5706] = {.lex_state = 146, .external_lex_state = 2},
  [5707] = {.lex_state = 146, .external_lex_state = 2},
  [5708] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5709] = {.lex_state = 11, .external_lex_state = 2},
  [5710] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5711] = {.lex_state = 98, .external_lex_state = 8},
  [5712] = {.lex_state = 11, .external_lex_state = 2},
  [5713] = {.lex_state = 11, .external_lex_state = 2},
  [5714] = {.lex_state = 146, .external_lex_state = 2},
  [5715] = {.lex_state = 146, .external_lex_state = 2},
  [5716] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5717] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5718] = {.lex_state = 146, .external_lex_state = 2},
  [5719] = {.lex_state = 146, .external_lex_state = 2},
  [5720] = {.lex_state = 146, .external_lex_state = 2},
  [5721] = {.lex_state = 11, .external_lex_state = 2},
  [5722] = {.lex_state = 146, .external_lex_state = 2},
  [5723] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5724] = {.lex_state = 146, .external_lex_state = 2},
  [5725] = {.lex_state = 146, .external_lex_state = 2},
  [5726] = {.lex_state = 146, .external_lex_state = 2},
  [5727] = {.lex_state = 146, .external_lex_state = 2},
  [5728] = {.lex_state = 151, .external_lex_state = 2},
  [5729] = {.lex_state = 146, .external_lex_state = 2},
  [5730] = {.lex_state = 146, .external_lex_state = 2},
  [5731] = {.lex_state = 51, .external_lex_state = 2},
  [5732] = {.lex_state = 61, .external_lex_state = 2},
  [5733] = {.lex_state = 6, .external_lex_state = 2},
  [5734] = {.lex_state = 146, .external_lex_state = 2},
  [5735] = {.lex_state = 146, .external_lex_state = 2},
  [5736] = {.lex_state = 151, .external_lex_state = 2},
  [5737] = {.lex_state = 151, .external_lex_state = 2},
  [5738] = {.lex_state = 6, .external_lex_state = 2},
  [5739] = {.lex_state = 151, .external_lex_state = 2},
  [5740] = {.lex_state = 151, .external_lex_state = 2},
  [5741] = {.lex_state = 26, .external_lex_state = 2},
  [5742] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5743] = {.lex_state = 98, .external_lex_state = 8},
  [5744] = {.lex_state = 151, .external_lex_state = 2},
  [5745] = {.lex_state = 146, .external_lex_state = 2},
  [5746] = {.lex_state = 151, .external_lex_state = 2},
  [5747] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5748] = {.lex_state = 6, .external_lex_state = 2},
  [5749] = {.lex_state = 146, .external_lex_state = 2},
  [5750] = {.lex_state = 11, .external_lex_state = 2},
  [5751] = {.lex_state = 11, .external_lex_state = 2},
  [5752] = {.lex_state = 146, .external_lex_state = 2},
  [5753] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5754] = {.lex_state = 98, .external_lex_state = 8},
  [5755] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5756] = {.lex_state = 146, .external_lex_state = 2},
  [5757] = {.lex_state = 11, .external_lex_state = 2},
  [5758] = {.lex_state = 11, .external_lex_state = 2},
  [5759] = {.lex_state = 146, .external_lex_state = 2},
  [5760] = {.lex_state = 151, .external_lex_state = 2},
  [5761] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5762] = {.lex_state = 98, .external_lex_state = 8},
  [5763] = {.lex_state = 11, .external_lex_state = 2},
  [5764] = {.lex_state = 151, .external_lex_state = 2},
  [5765] = {.lex_state = 11, .external_lex_state = 2},
  [5766] = {.lex_state = 146, .external_lex_state = 2},
  [5767] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5768] = {.lex_state = 98, .external_lex_state = 8},
  [5769] = {.lex_state = 11, .external_lex_state = 2},
  [5770] = {.lex_state = 146, .external_lex_state = 2},
  [5771] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5772] = {.lex_state = 146, .external_lex_state = 2},
  [5773] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5774] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5775] = {.lex_state = 98, .external_lex_state = 8},
  [5776] = {.lex_state = 11, .external_lex_state = 2},
  [5777] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5778] = {.lex_state = 98, .external_lex_state = 8},
  [5779] = {.lex_state = 148, .external_lex_state = 2},
  [5780] = {.lex_state = 151, .external_lex_state = 2},
  [5781] = {.lex_state = 148, .external_lex_state = 2},
  [5782] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5783] = {.lex_state = 146, .external_lex_state = 2},
  [5784] = {.lex_state = 151, .external_lex_state = 2},
  [5785] = {.lex_state = 98, .external_lex_state = 8},
  [5786] = {.lex_state = 51, .external_lex_state = 2},
  [5787] = {.lex_state = 6, .external_lex_state = 2},
  [5788] = {.lex_state = 11, .external_lex_state = 2},
  [5789] = {.lex_state = 11, .external_lex_state = 2},
  [5790] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5791] = {.lex_state = 98, .external_lex_state = 8},
  [5792] = {.lex_state = 6, .external_lex_state = 2},
  [5793] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5794] = {.lex_state = 98, .external_lex_state = 8},
  [5795] = {.lex_state = 151, .external_lex_state = 2},
  [5796] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5797] = {.lex_state = 98, .external_lex_state = 8},
  [5798] = {.lex_state = 11, .external_lex_state = 2},
  [5799] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5800] = {.lex_state = 98, .external_lex_state = 8},
  [5801] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5802] = {.lex_state = 98, .external_lex_state = 8},
  [5803] = {.lex_state = 151, .external_lex_state = 2},
  [5804] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5805] = {.lex_state = 98, .external_lex_state = 8},
  [5806] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5807] = {.lex_state = 98, .external_lex_state = 8},
  [5808] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5809] = {.lex_state = 98, .external_lex_state = 8},
  [5810] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5811] = {.lex_state = 98, .external_lex_state = 8},
  [5812] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5813] = {.lex_state = 98, .external_lex_state = 8},
  [5814] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5815] = {.lex_state = 98, .external_lex_state = 8},
  [5816] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5817] = {.lex_state = 6, .external_lex_state = 2},
  [5818] = {.lex_state = 151, .external_lex_state = 2},
  [5819] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5820] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5821] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5822] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5823] = {.lex_state = 6, .external_lex_state = 2},
  [5824] = {.lex_state = 151, .external_lex_state = 2},
  [5825] = {.lex_state = 11, .external_lex_state = 2},
  [5826] = {.lex_state = 146, .external_lex_state = 2},
  [5827] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5828] = {.lex_state = 98, .external_lex_state = 8},
  [5829] = {.lex_state = 151, .external_lex_state = 2},
  [5830] = {.lex_state = 11, .external_lex_state = 2},
  [5831] = {.lex_state = 40, .external_lex_state = 2},
  [5832] = {.lex_state = 26, .external_lex_state = 2},
  [5833] = {.lex_state = 146, .external_lex_state = 2},
  [5834] = {.lex_state = 11, .external_lex_state = 2},
  [5835] = {.lex_state = 40, .external_lex_state = 2},
  [5836] = {.lex_state = 11, .external_lex_state = 2},
  [5837] = {.lex_state = 26, .external_lex_state = 2},
  [5838] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5839] = {.lex_state = 146, .external_lex_state = 2},
  [5840] = {.lex_state = 11, .external_lex_state = 2},
  [5841] = {.lex_state = 6, .external_lex_state = 2},
  [5842] = {.lex_state = 146, .external_lex_state = 2},
  [5843] = {.lex_state = 146, .external_lex_state = 2},
  [5844] = {.lex_state = 11, .external_lex_state = 2},
  [5845] = {.lex_state = 11, .external_lex_state = 2},
  [5846] = {.lex_state = 146, .external_lex_state = 2},
  [5847] = {.lex_state = 146, .external_lex_state = 2},
  [5848] = {.lex_state = 11, .external_lex_state = 2},
  [5849] = {.lex_state = 6, .external_lex_state = 2},
  [5850] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5851] = {.lex_state = 11, .external_lex_state = 2},
  [5852] = {.lex_state = 11, .external_lex_state = 2},
  [5853] = {.lex_state = 11, .external_lex_state = 2},
  [5854] = {.lex_state = 40, .external_lex_state = 2},
  [5855] = {.lex_state = 11, .external_lex_state = 2},
  [5856] = {.lex_state = 11, .external_lex_state = 2},
  [5857] = {.lex_state = 146, .external_lex_state = 2},
  [5858] = {.lex_state = 40, .external_lex_state = 2},
  [5859] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5860] = {.lex_state = 6, .external_lex_state = 2},
  [5861] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5862] = {.lex_state = 40, .external_lex_state = 2},
  [5863] = {.lex_state = 11, .external_lex_state = 2},
  [5864] = {.lex_state = 40, .external_lex_state = 2},
  [5865] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5866] = {.lex_state = 146, .external_lex_state = 2},
  [5867] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5868] = {.lex_state = 40, .external_lex_state = 2},
  [5869] = {.lex_state = 148, .external_lex_state = 2},
  [5870] = {.lex_state = 40, .external_lex_state = 2},
  [5871] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5872] = {.lex_state = 146, .external_lex_state = 2},
  [5873] = {.lex_state = 146, .external_lex_state = 2},
  [5874] = {.lex_state = 40, .external_lex_state = 2},
  [5875] = {.lex_state = 40, .external_lex_state = 2},
  [5876] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5877] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 93},
  [5878] = {.lex_state = 146, .external_lex_state = 2},
  [5879] = {.lex_state = 40, .external_lex_state = 2},
  [5880] = {.lex_state = 6, .external_lex_state = 2},
  [5881] = {.lex_state = 40, .external_lex_state = 2},
  [5882] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5883] = {.lex_state = 40, .external_lex_state = 2},
  [5884] = {.lex_state = 40, .external_lex_state = 2},
  [5885] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5886] = {.lex_state = 11, .external_lex_state = 2},
  [5887] = {.lex_state = 146, .external_lex_state = 2},
  [5888] = {.lex_state = 40, .external_lex_state = 2},
  [5889] = {.lex_state = 146, .external_lex_state = 2},
  [5890] = {.lex_state = 40, .external_lex_state = 2},
  [5891] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5892] = {.lex_state = 11, .external_lex_state = 2},
  [5893] = {.lex_state = 40, .external_lex_state = 2},
  [5894] = {.lex_state = 40, .external_lex_state = 2},
  [5895] = {.lex_state = 61, .external_lex_state = 2},
  [5896] = {.lex_state = 40, .external_lex_state = 2},
  [5897] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5898] = {.lex_state = 40, .external_lex_state = 2},
  [5899] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5900] = {.lex_state = 40, .external_lex_state = 2},
  [5901] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5902] = {.lex_state = 40, .external_lex_state = 2},
  [5903] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5904] = {.lex_state = 40, .external_lex_state = 2},
  [5905] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5906] = {.lex_state = 40, .external_lex_state = 2},
  [5907] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5908] = {.lex_state = 40, .external_lex_state = 2},
  [5909] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5910] = {.lex_state = 40, .external_lex_state = 2},
  [5911] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5912] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5913] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5914] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5915] = {.lex_state = 11, .external_lex_state = 2},
  [5916] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5917] = {.lex_state = 146, .external_lex_state = 2},
  [5918] = {.lex_state = 146, .external_lex_state = 2},
  [5919] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5920] = {.lex_state = 11, .external_lex_state = 2},
  [5921] = {.lex_state = 146, .external_lex_state = 2},
  [5922] = {.lex_state = 146, .external_lex_state = 2},
  [5923] = {.lex_state = 146, .external_lex_state = 2},
  [5924] = {.lex_state = 11, .external_lex_state = 2},
  [5925] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5926] = {.lex_state = 146, .external_lex_state = 2},
  [5927] = {.lex_state = 11, .external_lex_state = 2},
  [5928] = {.lex_state = 146, .external_lex_state = 2},
  [5929] = {.lex_state = 11, .external_lex_state = 2},
  [5930] = {.lex_state = 146, .external_lex_state = 2},
  [5931] = {.lex_state = 148, .external_lex_state = 2},
  [5932] = {.lex_state = 148, .external_lex_state = 2},
  [5933] = {.lex_state = 88, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5934] = {.lex_state = 11, .external_lex_state = 2},
  [5935] = {.lex_state = 146, .external_lex_state = 2},
  [5936] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5937] = {.lex_state = 146, .external_lex_state = 2},
  [5938] = {.lex_state = 11, .external_lex_state = 2},
  [5939] = {.lex_state = 11, .external_lex_state = 2},
  [5940] = {.lex_state = 148, .external_lex_state = 2},
  [5941] = {.lex_state = 148, .external_lex_state = 2},
  [5942] = {.lex_state = 146, .external_lex_state = 2},
  [5943] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5944] = {.lex_state = 146, .external_lex_state = 2},
  [5945] = {.lex_state = 148, .external_lex_state = 2},
  [5946] = {.lex_state = 40, .external_lex_state = 2},
  [5947] = {.lex_state = 11, .external_lex_state = 2},
  [5948] = {.lex_state = 11, .external_lex_state = 2},
  [5949] = {.lex_state = 11, .external_lex_state = 2},
  [5950] = {.lex_state = 11, .external_lex_state = 2},
  [5951] = {.lex_state = 146, .external_lex_state = 2},
  [5952] = {.lex_state = 51, .external_lex_state = 2},
  [5953] = {.lex_state = 51, .external_lex_state = 2},
  [5954] = {.lex_state = 146, .external_lex_state = 8},
  [5955] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5956] = {.lex_state = 146, .external_lex_state = 2},
  [5957] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5958] = {.lex_state = 146, .external_lex_state = 2},
  [5959] = {.lex_state = 146, .external_lex_state = 8},
  [5960] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5961] = {.lex_state = 51, .external_lex_state = 2},
  [5962] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [5963] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5964] = {.lex_state = 146, .external_lex_state = 2},
  [5965] = {.lex_state = 148, .external_lex_state = 2},
  [5966] = {.lex_state = 11, .external_lex_state = 2},
  [5967] = {.lex_state = 11, .external_lex_state = 2},
  [5968] = {.lex_state = 11, .external_lex_state = 2},
  [5969] = {.lex_state = 148, .external_lex_state = 2},
  [5970] = {.lex_state = 146, .external_lex_state = 2},
  [5971] = {.lex_state = 51, .external_lex_state = 2},
  [5972] = {.lex_state = 6, .external_lex_state = 2},
  [5973] = {.lex_state = 148, .external_lex_state = 2},
  [5974] = {.lex_state = 11, .external_lex_state = 2},
  [5975] = {.lex_state = 11, .external_lex_state = 2},
  [5976] = {.lex_state = 51, .external_lex_state = 2},
  [5977] = {.lex_state = 11, .external_lex_state = 2},
  [5978] = {.lex_state = 146, .external_lex_state = 2},
  [5979] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5980] = {.lex_state = 146, .external_lex_state = 2},
  [5981] = {.lex_state = 146, .external_lex_state = 8},
  [5982] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5983] = {.lex_state = 146, .external_lex_state = 8},
  [5984] = {.lex_state = 148, .external_lex_state = 2},
  [5985] = {.lex_state = 148, .external_lex_state = 2},
  [5986] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [5987] = {.lex_state = 146, .external_lex_state = 2},
  [5988] = {.lex_state = 146, .external_lex_state = 2},
  [5989] = {.lex_state = 148, .external_lex_state = 2},
  [5990] = {.lex_state = 11, .external_lex_state = 2},
  [5991] = {.lex_state = 11, .external_lex_state = 2},
  [5992] = {.lex_state = 11, .external_lex_state = 2},
  [5993] = {.lex_state = 148, .external_lex_state = 2},
  [5994] = {.lex_state = 61, .external_lex_state = 2},
  [5995] = {.lex_state = 11, .external_lex_state = 2},
  [5996] = {.lex_state = 11, .external_lex_state = 2},
  [5997] = {.lex_state = 146, .external_lex_state = 8},
  [5998] = {.lex_state = 146, .external_lex_state = 2},
  [5999] = {.lex_state = 11, .external_lex_state = 2},
  [6000] = {.lex_state = 146, .external_lex_state = 8},
  [6001] = {.lex_state = 11, .external_lex_state = 2},
  [6002] = {.lex_state = 11, .external_lex_state = 2},
  [6003] = {.lex_state = 11, .external_lex_state = 2},
  [6004] = {.lex_state = 146, .external_lex_state = 2},
  [6005] = {.lex_state = 148, .external_lex_state = 2},
  [6006] = {.lex_state = 11, .external_lex_state = 2},
  [6007] = {.lex_state = 11, .external_lex_state = 2},
  [6008] = {.lex_state = 11, .external_lex_state = 2},
  [6009] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6010] = {.lex_state = 11, .external_lex_state = 2},
  [6011] = {.lex_state = 51, .external_lex_state = 2},
  [6012] = {.lex_state = 51, .external_lex_state = 2},
  [6013] = {.lex_state = 146, .external_lex_state = 2},
  [6014] = {.lex_state = 11, .external_lex_state = 2},
  [6015] = {.lex_state = 11, .external_lex_state = 2},
  [6016] = {.lex_state = 148, .external_lex_state = 2},
  [6017] = {.lex_state = 146, .external_lex_state = 8},
  [6018] = {.lex_state = 146, .external_lex_state = 2},
  [6019] = {.lex_state = 51, .external_lex_state = 2},
  [6020] = {.lex_state = 11, .external_lex_state = 2},
  [6021] = {.lex_state = 11, .external_lex_state = 2},
  [6022] = {.lex_state = 11, .external_lex_state = 2},
  [6023] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6024] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6025] = {.lex_state = 51, .external_lex_state = 2},
  [6026] = {.lex_state = 11, .external_lex_state = 2},
  [6027] = {.lex_state = 146, .external_lex_state = 2},
  [6028] = {.lex_state = 11, .external_lex_state = 2},
  [6029] = {.lex_state = 11, .external_lex_state = 2},
  [6030] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6031] = {.lex_state = 146, .external_lex_state = 2},
  [6032] = {.lex_state = 146, .external_lex_state = 8},
  [6033] = {.lex_state = 11, .external_lex_state = 2},
  [6034] = {.lex_state = 11, .external_lex_state = 2},
  [6035] = {.lex_state = 11, .external_lex_state = 2},
  [6036] = {.lex_state = 146, .external_lex_state = 2},
  [6037] = {.lex_state = 51, .external_lex_state = 2},
  [6038] = {.lex_state = 148, .external_lex_state = 2},
  [6039] = {.lex_state = 11, .external_lex_state = 2},
  [6040] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6041] = {.lex_state = 146, .external_lex_state = 2},
  [6042] = {.lex_state = 146, .external_lex_state = 8},
  [6043] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6044] = {.lex_state = 11, .external_lex_state = 2},
  [6045] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6046] = {.lex_state = 148, .external_lex_state = 2},
  [6047] = {.lex_state = 11, .external_lex_state = 2},
  [6048] = {.lex_state = 151, .external_lex_state = 2},
  [6049] = {.lex_state = 11, .external_lex_state = 2},
  [6050] = {.lex_state = 11, .external_lex_state = 2},
  [6051] = {.lex_state = 11, .external_lex_state = 2},
  [6052] = {.lex_state = 11, .external_lex_state = 2},
  [6053] = {.lex_state = 148, .external_lex_state = 2},
  [6054] = {.lex_state = 148, .external_lex_state = 2},
  [6055] = {.lex_state = 146, .external_lex_state = 2},
  [6056] = {.lex_state = 51, .external_lex_state = 2},
  [6057] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6058] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6059] = {.lex_state = 146, .external_lex_state = 8},
  [6060] = {.lex_state = 11, .external_lex_state = 2},
  [6061] = {.lex_state = 148, .external_lex_state = 2},
  [6062] = {.lex_state = 146, .external_lex_state = 8},
  [6063] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6064] = {.lex_state = 11, .external_lex_state = 2},
  [6065] = {.lex_state = 11, .external_lex_state = 2},
  [6066] = {.lex_state = 11, .external_lex_state = 2},
  [6067] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6068] = {.lex_state = 146, .external_lex_state = 2},
  [6069] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6070] = {.lex_state = 148, .external_lex_state = 2},
  [6071] = {.lex_state = 51, .external_lex_state = 2},
  [6072] = {.lex_state = 146, .external_lex_state = 8},
  [6073] = {.lex_state = 148, .external_lex_state = 2},
  [6074] = {.lex_state = 11, .external_lex_state = 2},
  [6075] = {.lex_state = 11, .external_lex_state = 2},
  [6076] = {.lex_state = 11, .external_lex_state = 2},
  [6077] = {.lex_state = 148, .external_lex_state = 2},
  [6078] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6079] = {.lex_state = 146, .external_lex_state = 2},
  [6080] = {.lex_state = 148, .external_lex_state = 2},
  [6081] = {.lex_state = 146, .external_lex_state = 8},
  [6082] = {.lex_state = 11, .external_lex_state = 2},
  [6083] = {.lex_state = 11, .external_lex_state = 2},
  [6084] = {.lex_state = 11, .external_lex_state = 2},
  [6085] = {.lex_state = 148, .external_lex_state = 2},
  [6086] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6087] = {.lex_state = 11, .external_lex_state = 2},
  [6088] = {.lex_state = 146, .external_lex_state = 8},
  [6089] = {.lex_state = 11, .external_lex_state = 2},
  [6090] = {.lex_state = 11, .external_lex_state = 2},
  [6091] = {.lex_state = 11, .external_lex_state = 2},
  [6092] = {.lex_state = 148, .external_lex_state = 2},
  [6093] = {.lex_state = 11, .external_lex_state = 2},
  [6094] = {.lex_state = 146, .external_lex_state = 8},
  [6095] = {.lex_state = 11, .external_lex_state = 2},
  [6096] = {.lex_state = 11, .external_lex_state = 2},
  [6097] = {.lex_state = 11, .external_lex_state = 2},
  [6098] = {.lex_state = 148, .external_lex_state = 2},
  [6099] = {.lex_state = 146, .external_lex_state = 2},
  [6100] = {.lex_state = 146, .external_lex_state = 2},
  [6101] = {.lex_state = 146, .external_lex_state = 8},
  [6102] = {.lex_state = 11, .external_lex_state = 2},
  [6103] = {.lex_state = 11, .external_lex_state = 2},
  [6104] = {.lex_state = 11, .external_lex_state = 2},
  [6105] = {.lex_state = 146, .external_lex_state = 8},
  [6106] = {.lex_state = 11, .external_lex_state = 2},
  [6107] = {.lex_state = 146, .external_lex_state = 8},
  [6108] = {.lex_state = 11, .external_lex_state = 2},
  [6109] = {.lex_state = 11, .external_lex_state = 2},
  [6110] = {.lex_state = 11, .external_lex_state = 2},
  [6111] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6112] = {.lex_state = 61, .external_lex_state = 2},
  [6113] = {.lex_state = 146, .external_lex_state = 8},
  [6114] = {.lex_state = 11, .external_lex_state = 2},
  [6115] = {.lex_state = 11, .external_lex_state = 2},
  [6116] = {.lex_state = 11, .external_lex_state = 2},
  [6117] = {.lex_state = 148, .external_lex_state = 2},
  [6118] = {.lex_state = 51, .external_lex_state = 2},
  [6119] = {.lex_state = 146, .external_lex_state = 8},
  [6120] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6121] = {.lex_state = 151, .external_lex_state = 2},
  [6122] = {.lex_state = 146, .external_lex_state = 8},
  [6123] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6124] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6125] = {.lex_state = 146, .external_lex_state = 8},
  [6126] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6127] = {.lex_state = 148, .external_lex_state = 2},
  [6128] = {.lex_state = 146, .external_lex_state = 8},
  [6129] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6130] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6131] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [6132] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6133] = {.lex_state = 146, .external_lex_state = 2},
  [6134] = {.lex_state = 148, .external_lex_state = 2},
  [6135] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [6136] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6137] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6138] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6139] = {.lex_state = 146, .external_lex_state = 2},
  [6140] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6141] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6142] = {.lex_state = 146, .external_lex_state = 2},
  [6143] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6144] = {.lex_state = 11, .external_lex_state = 2},
  [6145] = {.lex_state = 146, .external_lex_state = 2},
  [6146] = {.lex_state = 146, .external_lex_state = 2},
  [6147] = {.lex_state = 146, .external_lex_state = 2},
  [6148] = {.lex_state = 19, .external_lex_state = 2},
  [6149] = {.lex_state = 146, .external_lex_state = 2},
  [6150] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6151] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6152] = {.lex_state = 148, .external_lex_state = 2},
  [6153] = {.lex_state = 146, .external_lex_state = 2},
  [6154] = {.lex_state = 146, .external_lex_state = 2},
  [6155] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6156] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6157] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6158] = {.lex_state = 146, .external_lex_state = 8},
  [6159] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6160] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6161] = {.lex_state = 148, .external_lex_state = 2},
  [6162] = {.lex_state = 146, .external_lex_state = 2},
  [6163] = {.lex_state = 148, .external_lex_state = 2},
  [6164] = {.lex_state = 6, .external_lex_state = 2},
  [6165] = {.lex_state = 146, .external_lex_state = 2},
  [6166] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6167] = {.lex_state = 148, .external_lex_state = 2},
  [6168] = {.lex_state = 146, .external_lex_state = 2},
  [6169] = {.lex_state = 146, .external_lex_state = 8},
  [6170] = {.lex_state = 51, .external_lex_state = 2},
  [6171] = {.lex_state = 11, .external_lex_state = 2},
  [6172] = {.lex_state = 51, .external_lex_state = 2},
  [6173] = {.lex_state = 146, .external_lex_state = 2},
  [6174] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6175] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6176] = {.lex_state = 51, .external_lex_state = 2},
  [6177] = {.lex_state = 19, .external_lex_state = 2},
  [6178] = {.lex_state = 148, .external_lex_state = 2},
  [6179] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6180] = {.lex_state = 146, .external_lex_state = 8},
  [6181] = {.lex_state = 148, .external_lex_state = 2},
  [6182] = {.lex_state = 146, .external_lex_state = 2},
  [6183] = {.lex_state = 148, .external_lex_state = 2},
  [6184] = {.lex_state = 148, .external_lex_state = 2},
  [6185] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6186] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6187] = {.lex_state = 148, .external_lex_state = 2},
  [6188] = {.lex_state = 146, .external_lex_state = 2},
  [6189] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6190] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6191] = {.lex_state = 148, .external_lex_state = 2},
  [6192] = {.lex_state = 148, .external_lex_state = 2},
  [6193] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6194] = {.lex_state = 148, .external_lex_state = 2},
  [6195] = {.lex_state = 11, .external_lex_state = 2},
  [6196] = {.lex_state = 11, .external_lex_state = 2},
  [6197] = {.lex_state = 148, .external_lex_state = 2},
  [6198] = {.lex_state = 148, .external_lex_state = 2},
  [6199] = {.lex_state = 148, .external_lex_state = 2},
  [6200] = {.lex_state = 11, .external_lex_state = 2},
  [6201] = {.lex_state = 146, .external_lex_state = 2},
  [6202] = {.lex_state = 11, .external_lex_state = 2},
  [6203] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6204] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6205] = {.lex_state = 40, .external_lex_state = 2},
  [6206] = {.lex_state = 146, .external_lex_state = 2},
  [6207] = {.lex_state = 146, .external_lex_state = 2},
  [6208] = {.lex_state = 6, .external_lex_state = 2},
  [6209] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6210] = {.lex_state = 146, .external_lex_state = 8},
  [6211] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6212] = {.lex_state = 148, .external_lex_state = 2},
  [6213] = {.lex_state = 148, .external_lex_state = 2},
  [6214] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6215] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6216] = {.lex_state = 148, .external_lex_state = 2},
  [6217] = {.lex_state = 51, .external_lex_state = 2},
  [6218] = {.lex_state = 148, .external_lex_state = 2},
  [6219] = {.lex_state = 51, .external_lex_state = 2},
  [6220] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6221] = {.lex_state = 148, .external_lex_state = 2},
  [6222] = {.lex_state = 148, .external_lex_state = 2},
  [6223] = {.lex_state = 148, .external_lex_state = 2},
  [6224] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6225] = {.lex_state = 19, .external_lex_state = 2},
  [6226] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6227] = {.lex_state = 148, .external_lex_state = 2},
  [6228] = {.lex_state = 148, .external_lex_state = 2},
  [6229] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6230] = {.lex_state = 6, .external_lex_state = 2},
  [6231] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6232] = {.lex_state = 51, .external_lex_state = 2},
  [6233] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6234] = {.lex_state = 51, .external_lex_state = 2},
  [6235] = {.lex_state = 146, .external_lex_state = 8},
  [6236] = {.lex_state = 148, .external_lex_state = 2},
  [6237] = {.lex_state = 148, .external_lex_state = 2},
  [6238] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6239] = {.lex_state = 148, .external_lex_state = 2},
  [6240] = {.lex_state = 146, .external_lex_state = 2},
  [6241] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6242] = {.lex_state = 146, .external_lex_state = 2},
  [6243] = {.lex_state = 51, .external_lex_state = 2},
  [6244] = {.lex_state = 11, .external_lex_state = 2},
  [6245] = {.lex_state = 51, .external_lex_state = 2},
  [6246] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6247] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6248] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6249] = {.lex_state = 148, .external_lex_state = 2},
  [6250] = {.lex_state = 11, .external_lex_state = 2},
  [6251] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6252] = {.lex_state = 11, .external_lex_state = 2},
  [6253] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6254] = {.lex_state = 11, .external_lex_state = 2},
  [6255] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6256] = {.lex_state = 148, .external_lex_state = 2},
  [6257] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6258] = {.lex_state = 51, .external_lex_state = 2},
  [6259] = {.lex_state = 146, .external_lex_state = 8},
  [6260] = {.lex_state = 146, .external_lex_state = 2},
  [6261] = {.lex_state = 146, .external_lex_state = 2},
  [6262] = {.lex_state = 51, .external_lex_state = 2},
  [6263] = {.lex_state = 146, .external_lex_state = 2},
  [6264] = {.lex_state = 146, .external_lex_state = 2},
  [6265] = {.lex_state = 146, .external_lex_state = 8},
  [6266] = {.lex_state = 148, .external_lex_state = 2},
  [6267] = {.lex_state = 148, .external_lex_state = 2},
  [6268] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6269] = {.lex_state = 11, .external_lex_state = 2},
  [6270] = {.lex_state = 11, .external_lex_state = 2},
  [6271] = {.lex_state = 146, .external_lex_state = 2},
  [6272] = {.lex_state = 51, .external_lex_state = 2},
  [6273] = {.lex_state = 51, .external_lex_state = 2},
  [6274] = {.lex_state = 51, .external_lex_state = 2},
  [6275] = {.lex_state = 146, .external_lex_state = 2},
  [6276] = {.lex_state = 150, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6277] = {.lex_state = 146, .external_lex_state = 2},
  [6278] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6279] = {.lex_state = 146, .external_lex_state = 2},
  [6280] = {.lex_state = 148, .external_lex_state = 2},
  [6281] = {.lex_state = 11, .external_lex_state = 2},
  [6282] = {.lex_state = 11, .external_lex_state = 2},
  [6283] = {.lex_state = 11, .external_lex_state = 2},
  [6284] = {.lex_state = 11, .external_lex_state = 2},
  [6285] = {.lex_state = 11, .external_lex_state = 2},
  [6286] = {.lex_state = 11, .external_lex_state = 2},
  [6287] = {.lex_state = 148, .external_lex_state = 2},
  [6288] = {.lex_state = 11, .external_lex_state = 2},
  [6289] = {.lex_state = 146, .external_lex_state = 2},
  [6290] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6291] = {.lex_state = 51, .external_lex_state = 2},
  [6292] = {.lex_state = 51, .external_lex_state = 2},
  [6293] = {.lex_state = 11, .external_lex_state = 2},
  [6294] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6295] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [6296] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6297] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6298] = {.lex_state = 51, .external_lex_state = 2},
  [6299] = {.lex_state = 148, .external_lex_state = 2},
  [6300] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6301] = {.lex_state = 11, .external_lex_state = 2},
  [6302] = {.lex_state = 51, .external_lex_state = 2},
  [6303] = {.lex_state = 148, .external_lex_state = 2},
  [6304] = {.lex_state = 51, .external_lex_state = 2},
  [6305] = {.lex_state = 148, .external_lex_state = 2},
  [6306] = {.lex_state = 6, .external_lex_state = 2},
  [6307] = {.lex_state = 148, .external_lex_state = 2},
  [6308] = {.lex_state = 11, .external_lex_state = 2},
  [6309] = {.lex_state = 11, .external_lex_state = 2},
  [6310] = {.lex_state = 51, .external_lex_state = 2},
  [6311] = {.lex_state = 146, .external_lex_state = 8},
  [6312] = {.lex_state = 148, .external_lex_state = 2},
  [6313] = {.lex_state = 148, .external_lex_state = 2},
  [6314] = {.lex_state = 146, .external_lex_state = 2},
  [6315] = {.lex_state = 151, .external_lex_state = 2},
  [6316] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6317] = {.lex_state = 146, .external_lex_state = 2},
  [6318] = {.lex_state = 146, .external_lex_state = 2},
  [6319] = {.lex_state = 51, .external_lex_state = 2},
  [6320] = {.lex_state = 151, .external_lex_state = 2},
  [6321] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6322] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6323] = {.lex_state = 146, .external_lex_state = 2},
  [6324] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6325] = {.lex_state = 11, .external_lex_state = 2},
  [6326] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6327] = {.lex_state = 146, .external_lex_state = 2},
  [6328] = {.lex_state = 148, .external_lex_state = 2, .reserved_word_set_id = 94},
  [6329] = {.lex_state = 148, .external_lex_state = 2},
  [6330] = {.lex_state = 148, .external_lex_state = 2},
  [6331] = {.lex_state = 148, .external_lex_state = 2},
  [6332] = {.lex_state = 11, .external_lex_state = 2},
  [6333] = {.lex_state = 146, .external_lex_state = 2},
  [6334] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6335] = {.lex_state = 146, .external_lex_state = 2},
  [6336] = {.lex_state = 11, .external_lex_state = 2},
  [6337] = {.lex_state = 148, .external_lex_state = 2},
  [6338] = {.lex_state = 146, .external_lex_state = 2},
  [6339] = {.lex_state = 146, .external_lex_state = 2},
  [6340] = {.lex_state = 148, .external_lex_state = 2},
  [6341] = {.lex_state = 146, .external_lex_state = 2},
  [6342] = {.lex_state = 51, .external_lex_state = 2},
  [6343] = {.lex_state = 146, .external_lex_state = 2},
  [6344] = {.lex_state = 146, .external_lex_state = 2},
  [6345] = {.lex_state = 146, .external_lex_state = 2},
  [6346] = {.lex_state = 146, .external_lex_state = 2},
  [6347] = {.lex_state = 146, .external_lex_state = 2},
  [6348] = {.lex_state = 148, .external_lex_state = 2},
  [6349] = {.lex_state = 146, .external_lex_state = 2},
  [6350] = {.lex_state = 148, .external_lex_state = 2},
  [6351] = {.lex_state = 146, .external_lex_state = 2},
  [6352] = {.lex_state = 146, .external_lex_state = 2},
  [6353] = {.lex_state = 146, .external_lex_state = 2},
  [6354] = {.lex_state = 11, .external_lex_state = 2},
  [6355] = {.lex_state = 148, .external_lex_state = 2},
  [6356] = {.lex_state = 146, .external_lex_state = 2},
  [6357] = {.lex_state = 146, .external_lex_state = 2},
  [6358] = {.lex_state = 146, .external_lex_state = 2},
  [6359] = {.lex_state = 146, .external_lex_state = 2},
  [6360] = {.lex_state = 146, .external_lex_state = 2},
  [6361] = {.lex_state = 146, .external_lex_state = 2},
  [6362] = {.lex_state = 11, .external_lex_state = 2},
  [6363] = {.lex_state = 148, .external_lex_state = 2},
  [6364] = {.lex_state = 146, .external_lex_state = 2},
  [6365] = {.lex_state = 146, .external_lex_state = 2},
  [6366] = {.lex_state = 11, .external_lex_state = 2},
  [6367] = {.lex_state = 35, .external_lex_state = 2},
  [6368] = {.lex_state = 148, .external_lex_state = 2},
  [6369] = {.lex_state = 146, .external_lex_state = 2},
  [6370] = {.lex_state = 11, .external_lex_state = 2},
  [6371] = {.lex_state = 146, .external_lex_state = 2},
  [6372] = {.lex_state = 146, .external_lex_state = 2},
  [6373] = {.lex_state = 148, .external_lex_state = 2},
  [6374] = {.lex_state = 11, .external_lex_state = 2},
  [6375] = {.lex_state = 146, .external_lex_state = 2},
  [6376] = {.lex_state = 148, .external_lex_state = 2},
  [6377] = {.lex_state = 11, .external_lex_state = 2},
  [6378] = {.lex_state = 146, .external_lex_state = 2},
  [6379] = {.lex_state = 11, .external_lex_state = 2},
  [6380] = {.lex_state = 146, .external_lex_state = 2},
  [6381] = {.lex_state = 146, .external_lex_state = 2},
  [6382] = {.lex_state = 148, .external_lex_state = 2},
  [6383] = {.lex_state = 146, .external_lex_state = 2},
  [6384] = {.lex_state = 146, .external_lex_state = 3},
  [6385] = {.lex_state = 146, .external_lex_state = 2},
  [6386] = {.lex_state = 35, .external_lex_state = 2},
  [6387] = {.lex_state = 146, .external_lex_state = 2},
  [6388] = {.lex_state = 146, .external_lex_state = 3},
  [6389] = {.lex_state = 146, .external_lex_state = 2},
  [6390] = {.lex_state = 11, .external_lex_state = 2},
  [6391] = {.lex_state = 146, .external_lex_state = 2},
  [6392] = {.lex_state = 146, .external_lex_state = 2},
  [6393] = {.lex_state = 146, .external_lex_state = 2},
  [6394] = {.lex_state = 146, .external_lex_state = 2},
  [6395] = {.lex_state = 148, .external_lex_state = 2},
  [6396] = {.lex_state = 146, .external_lex_state = 2},
  [6397] = {.lex_state = 148, .external_lex_state = 2},
  [6398] = {.lex_state = 148, .external_lex_state = 2},
  [6399] = {.lex_state = 146, .external_lex_state = 2},
  [6400] = {.lex_state = 148, .external_lex_state = 2},
  [6401] = {.lex_state = 146, .external_lex_state = 2},
  [6402] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6403] = {.lex_state = 146, .external_lex_state = 2},
  [6404] = {.lex_state = 11, .external_lex_state = 2},
  [6405] = {.lex_state = 146, .external_lex_state = 2},
  [6406] = {.lex_state = 146, .external_lex_state = 2},
  [6407] = {.lex_state = 146, .external_lex_state = 2},
  [6408] = {.lex_state = 146, .external_lex_state = 2},
  [6409] = {.lex_state = 146, .external_lex_state = 2},
  [6410] = {.lex_state = 146, .external_lex_state = 2},
  [6411] = {.lex_state = 146, .external_lex_state = 2},
  [6412] = {.lex_state = 146, .external_lex_state = 2},
  [6413] = {.lex_state = 146, .external_lex_state = 2},
  [6414] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6415] = {.lex_state = 146, .external_lex_state = 2},
  [6416] = {.lex_state = 146, .external_lex_state = 2},
  [6417] = {.lex_state = 146, .external_lex_state = 2},
  [6418] = {.lex_state = 11, .external_lex_state = 2},
  [6419] = {.lex_state = 148, .external_lex_state = 2},
  [6420] = {.lex_state = 146, .external_lex_state = 2},
  [6421] = {.lex_state = 146, .external_lex_state = 2},
  [6422] = {.lex_state = 146, .external_lex_state = 2},
  [6423] = {.lex_state = 148, .external_lex_state = 2},
  [6424] = {.lex_state = 146, .external_lex_state = 2},
  [6425] = {.lex_state = 146, .external_lex_state = 2},
  [6426] = {.lex_state = 146, .external_lex_state = 2},
  [6427] = {.lex_state = 146, .external_lex_state = 2},
  [6428] = {.lex_state = 146, .external_lex_state = 2},
  [6429] = {.lex_state = 148, .external_lex_state = 2},
  [6430] = {.lex_state = 11, .external_lex_state = 2},
  [6431] = {.lex_state = 146, .external_lex_state = 2},
  [6432] = {.lex_state = 146, .external_lex_state = 2},
  [6433] = {.lex_state = 146, .external_lex_state = 2},
  [6434] = {.lex_state = 146, .external_lex_state = 2},
  [6435] = {.lex_state = 146, .external_lex_state = 2},
  [6436] = {.lex_state = 146, .external_lex_state = 2},
  [6437] = {.lex_state = 146, .external_lex_state = 2},
  [6438] = {.lex_state = 146, .external_lex_state = 2},
  [6439] = {.lex_state = 146, .external_lex_state = 2},
  [6440] = {.lex_state = 146, .external_lex_state = 2},
  [6441] = {.lex_state = 146, .external_lex_state = 2},
  [6442] = {.lex_state = 146, .external_lex_state = 2},
  [6443] = {.lex_state = 146, .external_lex_state = 2},
  [6444] = {.lex_state = 146, .external_lex_state = 2},
  [6445] = {.lex_state = 146, .external_lex_state = 2},
  [6446] = {.lex_state = 146, .external_lex_state = 2},
  [6447] = {.lex_state = 146, .external_lex_state = 2},
  [6448] = {.lex_state = 146, .external_lex_state = 2},
  [6449] = {.lex_state = 146, .external_lex_state = 2},
  [6450] = {.lex_state = 146, .external_lex_state = 2},
  [6451] = {.lex_state = 146, .external_lex_state = 2},
  [6452] = {.lex_state = 146, .external_lex_state = 2},
  [6453] = {.lex_state = 146, .external_lex_state = 2},
  [6454] = {.lex_state = 146, .external_lex_state = 2},
  [6455] = {.lex_state = 146, .external_lex_state = 2},
  [6456] = {.lex_state = 146, .external_lex_state = 2},
  [6457] = {.lex_state = 146, .external_lex_state = 2},
  [6458] = {.lex_state = 51, .external_lex_state = 2},
  [6459] = {.lex_state = 148, .external_lex_state = 2},
  [6460] = {.lex_state = 146, .external_lex_state = 2},
  [6461] = {.lex_state = 146, .external_lex_state = 2},
  [6462] = {.lex_state = 146, .external_lex_state = 2},
  [6463] = {.lex_state = 148, .external_lex_state = 2},
  [6464] = {.lex_state = 146, .external_lex_state = 2},
  [6465] = {.lex_state = 148, .external_lex_state = 2},
  [6466] = {.lex_state = 148, .external_lex_state = 2},
  [6467] = {.lex_state = 146, .external_lex_state = 2},
  [6468] = {.lex_state = 11, .external_lex_state = 2},
  [6469] = {.lex_state = 148, .external_lex_state = 2},
  [6470] = {.lex_state = 146, .external_lex_state = 2},
  [6471] = {.lex_state = 146, .external_lex_state = 2},
  [6472] = {.lex_state = 146, .external_lex_state = 2},
  [6473] = {.lex_state = 146, .external_lex_state = 2},
  [6474] = {.lex_state = 146, .external_lex_state = 2},
  [6475] = {.lex_state = 51, .external_lex_state = 2},
  [6476] = {.lex_state = 146, .external_lex_state = 2},
  [6477] = {.lex_state = 51, .external_lex_state = 2},
  [6478] = {.lex_state = 146, .external_lex_state = 2},
  [6479] = {.lex_state = 146, .external_lex_state = 2},
  [6480] = {.lex_state = 146, .external_lex_state = 2},
  [6481] = {.lex_state = 146, .external_lex_state = 2},
  [6482] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6483] = {.lex_state = 146, .external_lex_state = 2},
  [6484] = {.lex_state = 146, .external_lex_state = 2},
  [6485] = {.lex_state = 146, .external_lex_state = 2},
  [6486] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6487] = {.lex_state = 146, .external_lex_state = 2},
  [6488] = {.lex_state = 146, .external_lex_state = 2},
  [6489] = {.lex_state = 148, .external_lex_state = 2},
  [6490] = {.lex_state = 11, .external_lex_state = 2},
  [6491] = {.lex_state = 146, .external_lex_state = 2},
  [6492] = {.lex_state = 146, .external_lex_state = 2},
  [6493] = {.lex_state = 11, .external_lex_state = 2},
  [6494] = {.lex_state = 146, .external_lex_state = 2},
  [6495] = {.lex_state = 51, .external_lex_state = 2},
  [6496] = {.lex_state = 11, .external_lex_state = 2},
  [6497] = {.lex_state = 146, .external_lex_state = 2},
  [6498] = {.lex_state = 146, .external_lex_state = 2},
  [6499] = {.lex_state = 146, .external_lex_state = 2},
  [6500] = {.lex_state = 146, .external_lex_state = 2},
  [6501] = {.lex_state = 148, .external_lex_state = 2},
  [6502] = {.lex_state = 148, .external_lex_state = 2},
  [6503] = {.lex_state = 11, .external_lex_state = 2},
  [6504] = {.lex_state = 146, .external_lex_state = 2},
  [6505] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6506] = {.lex_state = 146, .external_lex_state = 2},
  [6507] = {.lex_state = 146, .external_lex_state = 2},
  [6508] = {.lex_state = 146, .external_lex_state = 2},
  [6509] = {.lex_state = 146, .external_lex_state = 2},
  [6510] = {.lex_state = 146, .external_lex_state = 2},
  [6511] = {.lex_state = 148, .external_lex_state = 2},
  [6512] = {.lex_state = 146, .external_lex_state = 2},
  [6513] = {.lex_state = 148, .external_lex_state = 2},
  [6514] = {.lex_state = 148, .external_lex_state = 2},
  [6515] = {.lex_state = 148, .external_lex_state = 2},
  [6516] = {.lex_state = 148, .external_lex_state = 2},
  [6517] = {.lex_state = 146, .external_lex_state = 2},
  [6518] = {.lex_state = 146, .external_lex_state = 2},
  [6519] = {.lex_state = 11, .external_lex_state = 2},
  [6520] = {.lex_state = 148, .external_lex_state = 2},
  [6521] = {.lex_state = 148, .external_lex_state = 2},
  [6522] = {.lex_state = 148, .external_lex_state = 2},
  [6523] = {.lex_state = 146, .external_lex_state = 2},
  [6524] = {.lex_state = 148, .external_lex_state = 2},
  [6525] = {.lex_state = 146, .external_lex_state = 2},
  [6526] = {.lex_state = 146, .external_lex_state = 2},
  [6527] = {.lex_state = 148, .external_lex_state = 2},
  [6528] = {.lex_state = 148, .external_lex_state = 2},
  [6529] = {.lex_state = 146, .external_lex_state = 2},
  [6530] = {.lex_state = 146, .external_lex_state = 2},
  [6531] = {.lex_state = 146, .external_lex_state = 2},
  [6532] = {.lex_state = 146, .external_lex_state = 2},
  [6533] = {.lex_state = 146, .external_lex_state = 2},
  [6534] = {.lex_state = 146, .external_lex_state = 2},
  [6535] = {.lex_state = 146, .external_lex_state = 2},
  [6536] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6537] = {.lex_state = 146, .external_lex_state = 2},
  [6538] = {.lex_state = 146, .external_lex_state = 2},
  [6539] = {.lex_state = 146, .external_lex_state = 2},
  [6540] = {.lex_state = 146, .external_lex_state = 2},
  [6541] = {.lex_state = 146, .external_lex_state = 2},
  [6542] = {.lex_state = 146, .external_lex_state = 2},
  [6543] = {.lex_state = 51, .external_lex_state = 2},
  [6544] = {.lex_state = 146, .external_lex_state = 2},
  [6545] = {.lex_state = 11, .external_lex_state = 2},
  [6546] = {.lex_state = 146, .external_lex_state = 2},
  [6547] = {.lex_state = 146, .external_lex_state = 2},
  [6548] = {.lex_state = 35, .external_lex_state = 2},
  [6549] = {.lex_state = 146, .external_lex_state = 2},
  [6550] = {.lex_state = 146, .external_lex_state = 3},
  [6551] = {.lex_state = 146, .external_lex_state = 2},
  [6552] = {.lex_state = 146, .external_lex_state = 2},
  [6553] = {.lex_state = 11, .external_lex_state = 2},
  [6554] = {.lex_state = 148, .external_lex_state = 2},
  [6555] = {.lex_state = 11, .external_lex_state = 2},
  [6556] = {.lex_state = 11, .external_lex_state = 2},
  [6557] = {.lex_state = 11, .external_lex_state = 2},
  [6558] = {.lex_state = 146, .external_lex_state = 2},
  [6559] = {.lex_state = 148, .external_lex_state = 2},
  [6560] = {.lex_state = 148, .external_lex_state = 2},
  [6561] = {.lex_state = 148, .external_lex_state = 2},
  [6562] = {.lex_state = 11, .external_lex_state = 2},
  [6563] = {.lex_state = 148, .external_lex_state = 2},
  [6564] = {.lex_state = 146, .external_lex_state = 2},
  [6565] = {.lex_state = 146, .external_lex_state = 2},
  [6566] = {.lex_state = 35, .external_lex_state = 2},
  [6567] = {.lex_state = 146, .external_lex_state = 3},
  [6568] = {.lex_state = 146, .external_lex_state = 2},
  [6569] = {.lex_state = 148, .external_lex_state = 2},
  [6570] = {.lex_state = 148, .external_lex_state = 2},
  [6571] = {.lex_state = 148, .external_lex_state = 2},
  [6572] = {.lex_state = 146, .external_lex_state = 2},
  [6573] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6574] = {.lex_state = 11, .external_lex_state = 2},
  [6575] = {.lex_state = 146, .external_lex_state = 2},
  [6576] = {.lex_state = 146, .external_lex_state = 2},
  [6577] = {.lex_state = 146, .external_lex_state = 2},
  [6578] = {.lex_state = 35, .external_lex_state = 2},
  [6579] = {.lex_state = 146, .external_lex_state = 2},
  [6580] = {.lex_state = 146, .external_lex_state = 2},
  [6581] = {.lex_state = 146, .external_lex_state = 2},
  [6582] = {.lex_state = 146, .external_lex_state = 2},
  [6583] = {.lex_state = 148, .external_lex_state = 2},
  [6584] = {.lex_state = 148, .external_lex_state = 2},
  [6585] = {.lex_state = 146, .external_lex_state = 2},
  [6586] = {.lex_state = 146, .external_lex_state = 2},
  [6587] = {.lex_state = 146, .external_lex_state = 2},
  [6588] = {.lex_state = 146, .external_lex_state = 2},
  [6589] = {.lex_state = 146, .external_lex_state = 2},
  [6590] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6591] = {.lex_state = 51, .external_lex_state = 2},
  [6592] = {.lex_state = 146, .external_lex_state = 3},
  [6593] = {.lex_state = 11, .external_lex_state = 2},
  [6594] = {.lex_state = 146, .external_lex_state = 2},
  [6595] = {.lex_state = 146, .external_lex_state = 2},
  [6596] = {.lex_state = 148, .external_lex_state = 2},
  [6597] = {.lex_state = 146, .external_lex_state = 2},
  [6598] = {.lex_state = 11, .external_lex_state = 2},
  [6599] = {.lex_state = 146, .external_lex_state = 2},
  [6600] = {.lex_state = 11, .external_lex_state = 2},
  [6601] = {.lex_state = 11, .external_lex_state = 2},
  [6602] = {.lex_state = 146, .external_lex_state = 2},
  [6603] = {.lex_state = 146, .external_lex_state = 2},
  [6604] = {.lex_state = 148, .external_lex_state = 2},
  [6605] = {.lex_state = 148, .external_lex_state = 2},
  [6606] = {.lex_state = 11, .external_lex_state = 2},
  [6607] = {.lex_state = 146, .external_lex_state = 2},
  [6608] = {.lex_state = 146, .external_lex_state = 2},
  [6609] = {.lex_state = 146, .external_lex_state = 2},
  [6610] = {.lex_state = 146, .external_lex_state = 2},
  [6611] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6612] = {.lex_state = 148, .external_lex_state = 2},
  [6613] = {.lex_state = 148, .external_lex_state = 2},
  [6614] = {.lex_state = 146, .external_lex_state = 2},
  [6615] = {.lex_state = 146, .external_lex_state = 2},
  [6616] = {.lex_state = 146, .external_lex_state = 2},
  [6617] = {.lex_state = 11, .external_lex_state = 2},
  [6618] = {.lex_state = 146, .external_lex_state = 2},
  [6619] = {.lex_state = 11, .external_lex_state = 2},
  [6620] = {.lex_state = 146, .external_lex_state = 2},
  [6621] = {.lex_state = 148, .external_lex_state = 2},
  [6622] = {.lex_state = 146, .external_lex_state = 2},
  [6623] = {.lex_state = 146, .external_lex_state = 2},
  [6624] = {.lex_state = 146, .external_lex_state = 2},
  [6625] = {.lex_state = 146, .external_lex_state = 2},
  [6626] = {.lex_state = 148, .external_lex_state = 2},
  [6627] = {.lex_state = 146, .external_lex_state = 2},
  [6628] = {.lex_state = 146, .external_lex_state = 2},
  [6629] = {.lex_state = 146, .external_lex_state = 2},
  [6630] = {.lex_state = 146, .external_lex_state = 2},
  [6631] = {.lex_state = 146, .external_lex_state = 2},
  [6632] = {.lex_state = 146, .external_lex_state = 2},
  [6633] = {.lex_state = 146, .external_lex_state = 2},
  [6634] = {.lex_state = 148, .external_lex_state = 2},
  [6635] = {.lex_state = 146, .external_lex_state = 2},
  [6636] = {.lex_state = 146, .external_lex_state = 2},
  [6637] = {.lex_state = 146, .external_lex_state = 2},
  [6638] = {.lex_state = 146, .external_lex_state = 2},
  [6639] = {.lex_state = 146, .external_lex_state = 2},
  [6640] = {.lex_state = 11, .external_lex_state = 2},
  [6641] = {.lex_state = 11, .external_lex_state = 2},
  [6642] = {.lex_state = 146, .external_lex_state = 2},
  [6643] = {.lex_state = 11, .external_lex_state = 2},
  [6644] = {.lex_state = 11, .external_lex_state = 2},
  [6645] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6646] = {.lex_state = 146, .external_lex_state = 2},
  [6647] = {.lex_state = 148, .external_lex_state = 2},
  [6648] = {.lex_state = 148, .external_lex_state = 2},
  [6649] = {.lex_state = 11, .external_lex_state = 2},
  [6650] = {.lex_state = 146, .external_lex_state = 2},
  [6651] = {.lex_state = 148, .external_lex_state = 2},
  [6652] = {.lex_state = 146, .external_lex_state = 2},
  [6653] = {.lex_state = 148, .external_lex_state = 2},
  [6654] = {.lex_state = 148, .external_lex_state = 2},
  [6655] = {.lex_state = 146, .external_lex_state = 2},
  [6656] = {.lex_state = 146, .external_lex_state = 2},
  [6657] = {.lex_state = 146, .external_lex_state = 2},
  [6658] = {.lex_state = 11, .external_lex_state = 2},
  [6659] = {.lex_state = 146, .external_lex_state = 2},
  [6660] = {.lex_state = 146, .external_lex_state = 2},
  [6661] = {.lex_state = 146, .external_lex_state = 2},
  [6662] = {.lex_state = 146, .external_lex_state = 2},
  [6663] = {.lex_state = 146, .external_lex_state = 2},
  [6664] = {.lex_state = 146, .external_lex_state = 2},
  [6665] = {.lex_state = 146, .external_lex_state = 2},
  [6666] = {.lex_state = 146, .external_lex_state = 2},
  [6667] = {.lex_state = 146, .external_lex_state = 2},
  [6668] = {.lex_state = 146, .external_lex_state = 2},
  [6669] = {.lex_state = 146, .external_lex_state = 2},
  [6670] = {.lex_state = 146, .external_lex_state = 2},
  [6671] = {.lex_state = 146, .external_lex_state = 2},
  [6672] = {.lex_state = 146, .external_lex_state = 2},
  [6673] = {.lex_state = 146, .external_lex_state = 2},
  [6674] = {.lex_state = 146, .external_lex_state = 2},
  [6675] = {.lex_state = 146, .external_lex_state = 2},
  [6676] = {.lex_state = 146, .external_lex_state = 2},
  [6677] = {.lex_state = 146, .external_lex_state = 2},
  [6678] = {.lex_state = 146, .external_lex_state = 2},
  [6679] = {.lex_state = 146, .external_lex_state = 2},
  [6680] = {.lex_state = 11, .external_lex_state = 2},
  [6681] = {.lex_state = 51, .external_lex_state = 2},
  [6682] = {.lex_state = 11, .external_lex_state = 2},
  [6683] = {.lex_state = 11, .external_lex_state = 2},
  [6684] = {.lex_state = 146, .external_lex_state = 2},
  [6685] = {.lex_state = 146, .external_lex_state = 2},
  [6686] = {.lex_state = 148, .external_lex_state = 2},
  [6687] = {.lex_state = 148, .external_lex_state = 2},
  [6688] = {.lex_state = 11, .external_lex_state = 2},
  [6689] = {.lex_state = 146, .external_lex_state = 2},
  [6690] = {.lex_state = 146, .external_lex_state = 2},
  [6691] = {.lex_state = 146, .external_lex_state = 2},
  [6692] = {.lex_state = 148, .external_lex_state = 2},
  [6693] = {.lex_state = 148, .external_lex_state = 2},
  [6694] = {.lex_state = 146, .external_lex_state = 2},
  [6695] = {.lex_state = 146, .external_lex_state = 2},
  [6696] = {.lex_state = 146, .external_lex_state = 2},
  [6697] = {.lex_state = 11, .external_lex_state = 2},
  [6698] = {.lex_state = 35, .external_lex_state = 2},
  [6699] = {.lex_state = 146, .external_lex_state = 3},
  [6700] = {.lex_state = 146, .external_lex_state = 2},
  [6701] = {.lex_state = 146, .external_lex_state = 2},
  [6702] = {.lex_state = 146, .external_lex_state = 2},
  [6703] = {.lex_state = 146, .external_lex_state = 2},
  [6704] = {.lex_state = 148, .external_lex_state = 2},
  [6705] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6706] = {.lex_state = 146, .external_lex_state = 2},
  [6707] = {.lex_state = 146, .external_lex_state = 2},
  [6708] = {.lex_state = 146, .external_lex_state = 2},
  [6709] = {.lex_state = 146, .external_lex_state = 2},
  [6710] = {.lex_state = 146, .external_lex_state = 2},
  [6711] = {.lex_state = 148, .external_lex_state = 2},
  [6712] = {.lex_state = 148, .external_lex_state = 2},
  [6713] = {.lex_state = 148, .external_lex_state = 2},
  [6714] = {.lex_state = 146, .external_lex_state = 2},
  [6715] = {.lex_state = 146, .external_lex_state = 2},
  [6716] = {.lex_state = 146, .external_lex_state = 2},
  [6717] = {.lex_state = 146, .external_lex_state = 2},
  [6718] = {.lex_state = 11, .external_lex_state = 2},
  [6719] = {.lex_state = 146, .external_lex_state = 2},
  [6720] = {.lex_state = 11, .external_lex_state = 2},
  [6721] = {.lex_state = 146, .external_lex_state = 2},
  [6722] = {.lex_state = 146, .external_lex_state = 2},
  [6723] = {.lex_state = 148, .external_lex_state = 2},
  [6724] = {.lex_state = 148, .external_lex_state = 2},
  [6725] = {.lex_state = 11, .external_lex_state = 2},
  [6726] = {.lex_state = 148, .external_lex_state = 2},
  [6727] = {.lex_state = 146, .external_lex_state = 2},
  [6728] = {.lex_state = 146, .external_lex_state = 2},
  [6729] = {.lex_state = 148, .external_lex_state = 2},
  [6730] = {.lex_state = 148, .external_lex_state = 2},
  [6731] = {.lex_state = 146, .external_lex_state = 2},
  [6732] = {.lex_state = 146, .external_lex_state = 2},
  [6733] = {.lex_state = 146, .external_lex_state = 2},
  [6734] = {.lex_state = 11, .external_lex_state = 2},
  [6735] = {.lex_state = 148, .external_lex_state = 2},
  [6736] = {.lex_state = 146, .external_lex_state = 2},
  [6737] = {.lex_state = 146, .external_lex_state = 2},
  [6738] = {.lex_state = 146, .external_lex_state = 2},
  [6739] = {.lex_state = 146, .external_lex_state = 2},
  [6740] = {.lex_state = 146, .external_lex_state = 2},
  [6741] = {.lex_state = 146, .external_lex_state = 2},
  [6742] = {.lex_state = 148, .external_lex_state = 2},
  [6743] = {.lex_state = 146, .external_lex_state = 2},
  [6744] = {.lex_state = 146, .external_lex_state = 2},
  [6745] = {.lex_state = 146, .external_lex_state = 2},
  [6746] = {.lex_state = 146, .external_lex_state = 2},
  [6747] = {.lex_state = 146, .external_lex_state = 2},
  [6748] = {.lex_state = 146, .external_lex_state = 2},
  [6749] = {.lex_state = 146, .external_lex_state = 2},
  [6750] = {.lex_state = 146, .external_lex_state = 2},
  [6751] = {.lex_state = 51, .external_lex_state = 2},
  [6752] = {.lex_state = 146, .external_lex_state = 2},
  [6753] = {.lex_state = 146, .external_lex_state = 2},
  [6754] = {.lex_state = 146, .external_lex_state = 2},
  [6755] = {.lex_state = 11, .external_lex_state = 2},
  [6756] = {.lex_state = 146, .external_lex_state = 2},
  [6757] = {.lex_state = 146, .external_lex_state = 2},
  [6758] = {.lex_state = 146, .external_lex_state = 2},
  [6759] = {.lex_state = 148, .external_lex_state = 2},
  [6760] = {.lex_state = 148, .external_lex_state = 2},
  [6761] = {.lex_state = 11, .external_lex_state = 2},
  [6762] = {.lex_state = 146, .external_lex_state = 2},
  [6763] = {.lex_state = 148, .external_lex_state = 2},
  [6764] = {.lex_state = 146, .external_lex_state = 2},
  [6765] = {.lex_state = 148, .external_lex_state = 2},
  [6766] = {.lex_state = 148, .external_lex_state = 2},
  [6767] = {.lex_state = 146, .external_lex_state = 2},
  [6768] = {.lex_state = 146, .external_lex_state = 2},
  [6769] = {.lex_state = 146, .external_lex_state = 2},
  [6770] = {.lex_state = 11, .external_lex_state = 2},
  [6771] = {.lex_state = 146, .external_lex_state = 2},
  [6772] = {.lex_state = 146, .external_lex_state = 2},
  [6773] = {.lex_state = 146, .external_lex_state = 2},
  [6774] = {.lex_state = 146, .external_lex_state = 2},
  [6775] = {.lex_state = 146, .external_lex_state = 2},
  [6776] = {.lex_state = 146, .external_lex_state = 2},
  [6777] = {.lex_state = 148, .external_lex_state = 2},
  [6778] = {.lex_state = 35, .external_lex_state = 2},
  [6779] = {.lex_state = 146, .external_lex_state = 2},
  [6780] = {.lex_state = 146, .external_lex_state = 2},
  [6781] = {.lex_state = 146, .external_lex_state = 2},
  [6782] = {.lex_state = 146, .external_lex_state = 2},
  [6783] = {.lex_state = 146, .external_lex_state = 2},
  [6784] = {.lex_state = 146, .external_lex_state = 2},
  [6785] = {.lex_state = 146, .external_lex_state = 2},
  [6786] = {.lex_state = 146, .external_lex_state = 2},
  [6787] = {.lex_state = 146, .external_lex_state = 2},
  [6788] = {.lex_state = 148, .external_lex_state = 2},
  [6789] = {.lex_state = 146, .external_lex_state = 2},
  [6790] = {.lex_state = 146, .external_lex_state = 2},
  [6791] = {.lex_state = 11, .external_lex_state = 2},
  [6792] = {.lex_state = 146, .external_lex_state = 3},
  [6793] = {.lex_state = 51, .external_lex_state = 2},
  [6794] = {.lex_state = 11, .external_lex_state = 2},
  [6795] = {.lex_state = 146, .external_lex_state = 2},
  [6796] = {.lex_state = 11, .external_lex_state = 2},
  [6797] = {.lex_state = 146, .external_lex_state = 2},
  [6798] = {.lex_state = 146, .external_lex_state = 2},
  [6799] = {.lex_state = 146, .external_lex_state = 2},
  [6800] = {.lex_state = 146, .external_lex_state = 2},
  [6801] = {.lex_state = 146, .external_lex_state = 2},
  [6802] = {.lex_state = 146, .external_lex_state = 2},
  [6803] = {.lex_state = 146, .external_lex_state = 2},
  [6804] = {.lex_state = 146, .external_lex_state = 2},
  [6805] = {.lex_state = 146, .external_lex_state = 2},
  [6806] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6807] = {.lex_state = 51, .external_lex_state = 2},
  [6808] = {.lex_state = 146, .external_lex_state = 2},
  [6809] = {.lex_state = 35, .external_lex_state = 2},
  [6810] = {.lex_state = 11, .external_lex_state = 2},
  [6811] = {.lex_state = 146, .external_lex_state = 2},
  [6812] = {.lex_state = 146, .external_lex_state = 2},
  [6813] = {.lex_state = 146, .external_lex_state = 2},
  [6814] = {.lex_state = 146, .external_lex_state = 2},
  [6815] = {.lex_state = 146, .external_lex_state = 2},
  [6816] = {.lex_state = 146, .external_lex_state = 2},
  [6817] = {.lex_state = 146, .external_lex_state = 2},
  [6818] = {.lex_state = 146, .external_lex_state = 2},
  [6819] = {.lex_state = 146, .external_lex_state = 2},
  [6820] = {.lex_state = 146, .external_lex_state = 2},
  [6821] = {.lex_state = 35, .external_lex_state = 2},
  [6822] = {.lex_state = 146, .external_lex_state = 2},
  [6823] = {.lex_state = 146, .external_lex_state = 2},
  [6824] = {.lex_state = 146, .external_lex_state = 2},
  [6825] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6826] = {.lex_state = 146, .external_lex_state = 2},
  [6827] = {.lex_state = 146, .external_lex_state = 2},
  [6828] = {.lex_state = 146, .external_lex_state = 2},
  [6829] = {.lex_state = 146, .external_lex_state = 2},
  [6830] = {.lex_state = 146, .external_lex_state = 2},
  [6831] = {.lex_state = 148, .external_lex_state = 2},
  [6832] = {.lex_state = 146, .external_lex_state = 2},
  [6833] = {.lex_state = 146, .external_lex_state = 2},
  [6834] = {.lex_state = 146, .external_lex_state = 2},
  [6835] = {.lex_state = 35, .external_lex_state = 2},
  [6836] = {.lex_state = 146, .external_lex_state = 2},
  [6837] = {.lex_state = 146, .external_lex_state = 2},
  [6838] = {.lex_state = 146, .external_lex_state = 2},
  [6839] = {.lex_state = 146, .external_lex_state = 2},
  [6840] = {.lex_state = 146, .external_lex_state = 2},
  [6841] = {.lex_state = 146, .external_lex_state = 2},
  [6842] = {.lex_state = 35, .external_lex_state = 2},
  [6843] = {.lex_state = 146, .external_lex_state = 2},
  [6844] = {.lex_state = 146, .external_lex_state = 2},
  [6845] = {.lex_state = 146, .external_lex_state = 3},
  [6846] = {.lex_state = 146, .external_lex_state = 2},
  [6847] = {.lex_state = 146, .external_lex_state = 2},
  [6848] = {.lex_state = 146, .external_lex_state = 2},
  [6849] = {.lex_state = 146, .external_lex_state = 2},
  [6850] = {.lex_state = 146, .external_lex_state = 2},
  [6851] = {.lex_state = 35, .external_lex_state = 2},
  [6852] = {.lex_state = 146, .external_lex_state = 2},
  [6853] = {.lex_state = 146, .external_lex_state = 2},
  [6854] = {.lex_state = 146, .external_lex_state = 2},
  [6855] = {.lex_state = 11, .external_lex_state = 2},
  [6856] = {.lex_state = 146, .external_lex_state = 2},
  [6857] = {.lex_state = 146, .external_lex_state = 2},
  [6858] = {.lex_state = 146, .external_lex_state = 2},
  [6859] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6860] = {.lex_state = 146, .external_lex_state = 2},
  [6861] = {.lex_state = 146, .external_lex_state = 3},
  [6862] = {.lex_state = 146, .external_lex_state = 2},
  [6863] = {.lex_state = 146, .external_lex_state = 2},
  [6864] = {.lex_state = 146, .external_lex_state = 2},
  [6865] = {.lex_state = 51, .external_lex_state = 2},
  [6866] = {.lex_state = 146, .external_lex_state = 2},
  [6867] = {.lex_state = 146, .external_lex_state = 2},
  [6868] = {.lex_state = 146, .external_lex_state = 2},
  [6869] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6870] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6871] = {.lex_state = 146, .external_lex_state = 2},
  [6872] = {.lex_state = 146, .external_lex_state = 2},
  [6873] = {.lex_state = 146, .external_lex_state = 2},
  [6874] = {.lex_state = 146, .external_lex_state = 2},
  [6875] = {.lex_state = 148, .external_lex_state = 2},
  [6876] = {.lex_state = 146, .external_lex_state = 2},
  [6877] = {.lex_state = 146, .external_lex_state = 2},
  [6878] = {.lex_state = 11, .external_lex_state = 2},
  [6879] = {.lex_state = 146, .external_lex_state = 2},
  [6880] = {.lex_state = 146, .external_lex_state = 2},
  [6881] = {.lex_state = 146, .external_lex_state = 2},
  [6882] = {.lex_state = 51, .external_lex_state = 2},
  [6883] = {.lex_state = 146, .external_lex_state = 2},
  [6884] = {.lex_state = 146, .external_lex_state = 2},
  [6885] = {.lex_state = 146, .external_lex_state = 2},
  [6886] = {.lex_state = 146, .external_lex_state = 2},
  [6887] = {.lex_state = 11, .external_lex_state = 2},
  [6888] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6889] = {.lex_state = 146, .external_lex_state = 2},
  [6890] = {.lex_state = 146, .external_lex_state = 2},
  [6891] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6892] = {.lex_state = 146, .external_lex_state = 2},
  [6893] = {.lex_state = 148, .external_lex_state = 2},
  [6894] = {.lex_state = 146, .external_lex_state = 2},
  [6895] = {.lex_state = 146, .external_lex_state = 2},
  [6896] = {.lex_state = 146, .external_lex_state = 2},
  [6897] = {.lex_state = 148, .external_lex_state = 2},
  [6898] = {.lex_state = 148, .external_lex_state = 2},
  [6899] = {.lex_state = 35, .external_lex_state = 2},
  [6900] = {.lex_state = 146, .external_lex_state = 2},
  [6901] = {.lex_state = 146, .external_lex_state = 3},
  [6902] = {.lex_state = 146, .external_lex_state = 2},
  [6903] = {.lex_state = 146, .external_lex_state = 2},
  [6904] = {.lex_state = 146, .external_lex_state = 2},
  [6905] = {.lex_state = 146, .external_lex_state = 2},
  [6906] = {.lex_state = 148, .external_lex_state = 2},
  [6907] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6908] = {.lex_state = 146, .external_lex_state = 2},
  [6909] = {.lex_state = 148, .external_lex_state = 2},
  [6910] = {.lex_state = 146, .external_lex_state = 2},
  [6911] = {.lex_state = 146, .external_lex_state = 2},
  [6912] = {.lex_state = 146, .external_lex_state = 2},
  [6913] = {.lex_state = 148, .external_lex_state = 2},
  [6914] = {.lex_state = 146, .external_lex_state = 2},
  [6915] = {.lex_state = 51, .external_lex_state = 2},
  [6916] = {.lex_state = 146, .external_lex_state = 2},
  [6917] = {.lex_state = 146, .external_lex_state = 2},
  [6918] = {.lex_state = 146, .external_lex_state = 2},
  [6919] = {.lex_state = 146, .external_lex_state = 2},
  [6920] = {.lex_state = 148, .external_lex_state = 2},
  [6921] = {.lex_state = 35, .external_lex_state = 2},
  [6922] = {.lex_state = 146, .external_lex_state = 2},
  [6923] = {.lex_state = 146, .external_lex_state = 2},
  [6924] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6925] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6926] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6927] = {.lex_state = 146, .external_lex_state = 2},
  [6928] = {.lex_state = 146, .external_lex_state = 3},
  [6929] = {.lex_state = 148, .external_lex_state = 2},
  [6930] = {.lex_state = 146, .external_lex_state = 3},
  [6931] = {.lex_state = 148, .external_lex_state = 2},
  [6932] = {.lex_state = 146, .external_lex_state = 2},
  [6933] = {.lex_state = 146, .external_lex_state = 2},
  [6934] = {.lex_state = 146, .external_lex_state = 2},
  [6935] = {.lex_state = 146, .external_lex_state = 2},
  [6936] = {.lex_state = 148, .external_lex_state = 2},
  [6937] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6938] = {.lex_state = 146, .external_lex_state = 2},
  [6939] = {.lex_state = 146, .external_lex_state = 2},
  [6940] = {.lex_state = 146, .external_lex_state = 2},
  [6941] = {.lex_state = 146, .external_lex_state = 2},
  [6942] = {.lex_state = 146, .external_lex_state = 2},
  [6943] = {.lex_state = 146, .external_lex_state = 2},
  [6944] = {.lex_state = 146, .external_lex_state = 2},
  [6945] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6946] = {.lex_state = 146, .external_lex_state = 2},
  [6947] = {.lex_state = 146, .external_lex_state = 2},
  [6948] = {.lex_state = 146, .external_lex_state = 2},
  [6949] = {.lex_state = 146, .external_lex_state = 2},
  [6950] = {.lex_state = 146, .external_lex_state = 2},
  [6951] = {.lex_state = 146, .external_lex_state = 2},
  [6952] = {.lex_state = 146, .external_lex_state = 2},
  [6953] = {.lex_state = 146, .external_lex_state = 2},
  [6954] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6955] = {.lex_state = 146, .external_lex_state = 2},
  [6956] = {.lex_state = 146, .external_lex_state = 2},
  [6957] = {.lex_state = 146, .external_lex_state = 2},
  [6958] = {.lex_state = 146, .external_lex_state = 2},
  [6959] = {.lex_state = 146, .external_lex_state = 2},
  [6960] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6961] = {.lex_state = 146, .external_lex_state = 2},
  [6962] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6963] = {.lex_state = 146, .external_lex_state = 2},
  [6964] = {.lex_state = 35, .external_lex_state = 2},
  [6965] = {.lex_state = 146, .external_lex_state = 2},
  [6966] = {.lex_state = 146, .external_lex_state = 2},
  [6967] = {.lex_state = 146, .external_lex_state = 3},
  [6968] = {.lex_state = 146, .external_lex_state = 2},
  [6969] = {.lex_state = 146, .external_lex_state = 2},
  [6970] = {.lex_state = 146, .external_lex_state = 2},
  [6971] = {.lex_state = 146, .external_lex_state = 2},
  [6972] = {.lex_state = 146, .external_lex_state = 2},
  [6973] = {.lex_state = 146, .external_lex_state = 2},
  [6974] = {.lex_state = 146, .external_lex_state = 2},
  [6975] = {.lex_state = 146, .external_lex_state = 2},
  [6976] = {.lex_state = 146, .external_lex_state = 2},
  [6977] = {.lex_state = 146, .external_lex_state = 2},
  [6978] = {.lex_state = 146, .external_lex_state = 2},
  [6979] = {.lex_state = 146, .external_lex_state = 2},
  [6980] = {.lex_state = 146, .external_lex_state = 2},
  [6981] = {.lex_state = 146, .external_lex_state = 2},
  [6982] = {.lex_state = 146, .external_lex_state = 2},
  [6983] = {.lex_state = 146, .external_lex_state = 2},
  [6984] = {.lex_state = 146, .external_lex_state = 2},
  [6985] = {.lex_state = 146, .external_lex_state = 2},
  [6986] = {.lex_state = 146, .external_lex_state = 2},
  [6987] = {.lex_state = 11, .external_lex_state = 2},
  [6988] = {.lex_state = 146, .external_lex_state = 2},
  [6989] = {.lex_state = 146, .external_lex_state = 2},
  [6990] = {.lex_state = 146, .external_lex_state = 2},
  [6991] = {.lex_state = 146, .external_lex_state = 2},
  [6992] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [6993] = {.lex_state = 146, .external_lex_state = 2},
  [6994] = {.lex_state = 146, .external_lex_state = 2},
  [6995] = {.lex_state = 148, .external_lex_state = 2},
  [6996] = {.lex_state = 146, .external_lex_state = 2},
  [6997] = {.lex_state = 11, .external_lex_state = 2},
  [6998] = {.lex_state = 146, .external_lex_state = 2},
  [6999] = {.lex_state = 146, .external_lex_state = 2},
  [7000] = {.lex_state = 146, .external_lex_state = 2},
  [7001] = {.lex_state = 51, .external_lex_state = 2},
  [7002] = {.lex_state = 146, .external_lex_state = 2},
  [7003] = {.lex_state = 146, .external_lex_state = 2},
  [7004] = {.lex_state = 146, .external_lex_state = 2},
  [7005] = {.lex_state = 148, .external_lex_state = 2},
  [7006] = {.lex_state = 146, .external_lex_state = 2},
  [7007] = {.lex_state = 51, .external_lex_state = 2},
  [7008] = {.lex_state = 146, .external_lex_state = 2},
  [7009] = {.lex_state = 146, .external_lex_state = 2},
  [7010] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [7011] = {.lex_state = 148, .external_lex_state = 2},
  [7012] = {.lex_state = 146, .external_lex_state = 2},
  [7013] = {.lex_state = 146, .external_lex_state = 2},
  [7014] = {.lex_state = 146, .external_lex_state = 2},
  [7015] = {.lex_state = 146, .external_lex_state = 2},
  [7016] = {.lex_state = 146, .external_lex_state = 2},
  [7017] = {.lex_state = 146, .external_lex_state = 2},
  [7018] = {.lex_state = 146, .external_lex_state = 2},
  [7019] = {.lex_state = 146, .external_lex_state = 2},
  [7020] = {.lex_state = 51, .external_lex_state = 2},
  [7021] = {.lex_state = 146, .external_lex_state = 2},
  [7022] = {.lex_state = 146, .external_lex_state = 2},
  [7023] = {.lex_state = 146, .external_lex_state = 2},
  [7024] = {.lex_state = 51, .external_lex_state = 2},
  [7025] = {.lex_state = 11, .external_lex_state = 2},
  [7026] = {.lex_state = 146, .external_lex_state = 2},
  [7027] = {.lex_state = 148, .external_lex_state = 2},
  [7028] = {.lex_state = 146, .external_lex_state = 2},
  [7029] = {.lex_state = 51, .external_lex_state = 2},
  [7030] = {.lex_state = 148, .external_lex_state = 2},
  [7031] = {.lex_state = 146, .external_lex_state = 2},
  [7032] = {.lex_state = 146, .external_lex_state = 2},
  [7033] = {.lex_state = 35, .external_lex_state = 2},
  [7034] = {.lex_state = 148, .external_lex_state = 2},
  [7035] = {.lex_state = 146, .external_lex_state = 3},
  [7036] = {.lex_state = 146, .external_lex_state = 2},
  [7037] = {.lex_state = 146, .external_lex_state = 2},
  [7038] = {.lex_state = 35, .external_lex_state = 2},
  [7039] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [7040] = {.lex_state = 146, .external_lex_state = 3},
  [7041] = {.lex_state = 146, .external_lex_state = 2},
  [7042] = {.lex_state = 148, .external_lex_state = 2},
  [7043] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [7044] = {.lex_state = 146, .external_lex_state = 2},
  [7045] = {.lex_state = 146, .external_lex_state = 2},
  [7046] = {.lex_state = 146, .external_lex_state = 2},
  [7047] = {.lex_state = 146, .external_lex_state = 2},
  [7048] = {.lex_state = 146, .external_lex_state = 2},
  [7049] = {.lex_state = 148, .external_lex_state = 2},
  [7050] = {.lex_state = 146, .external_lex_state = 2},
  [7051] = {.lex_state = 146, .external_lex_state = 2},
  [7052] = {.lex_state = 146, .external_lex_state = 2},
  [7053] = {.lex_state = 146, .external_lex_state = 2},
  [7054] = {.lex_state = 146, .external_lex_state = 2},
  [7055] = {.lex_state = 146, .external_lex_state = 2},
  [7056] = {.lex_state = 146, .external_lex_state = 2},
  [7057] = {.lex_state = 146, .external_lex_state = 2},
  [7058] = {.lex_state = 146, .external_lex_state = 2},
  [7059] = {.lex_state = 148, .external_lex_state = 2},
  [7060] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [7061] = {.lex_state = 146, .external_lex_state = 2},
  [7062] = {.lex_state = 146, .external_lex_state = 2},
  [7063] = {.lex_state = 146, .external_lex_state = 2},
  [7064] = {.lex_state = 146, .external_lex_state = 2},
  [7065] = {.lex_state = 148, .external_lex_state = 2},
  [7066] = {.lex_state = 148, .external_lex_state = 2},
  [7067] = {.lex_state = 148, .external_lex_state = 2},
  [7068] = {.lex_state = 146, .external_lex_state = 2},
  [7069] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [7070] = {.lex_state = 146, .external_lex_state = 2},
  [7071] = {.lex_state = 146, .external_lex_state = 2},
  [7072] = {.lex_state = 146, .external_lex_state = 2},
  [7073] = {.lex_state = 148, .external_lex_state = 2},
  [7074] = {.lex_state = 146, .external_lex_state = 2},
  [7075] = {.lex_state = 146, .external_lex_state = 2},
  [7076] = {.lex_state = 146, .external_lex_state = 2},
  [7077] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [7078] = {.lex_state = 146, .external_lex_state = 3},
  [7079] = {.lex_state = 146, .external_lex_state = 2},
  [7080] = {.lex_state = 146, .external_lex_state = 2},
  [7081] = {.lex_state = 146, .external_lex_state = 2},
  [7082] = {.lex_state = 146, .external_lex_state = 2},
  [7083] = {.lex_state = 146, .external_lex_state = 2},
  [7084] = {.lex_state = 146, .external_lex_state = 2},
  [7085] = {.lex_state = 148, .external_lex_state = 2},
  [7086] = {.lex_state = 146, .external_lex_state = 2},
  [7087] = {.lex_state = 146, .external_lex_state = 2},
  [7088] = {.lex_state = 51, .external_lex_state = 2},
  [7089] = {.lex_state = 146, .external_lex_state = 2},
  [7090] = {.lex_state = 146, .external_lex_state = 2},
  [7091] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [7092] = {.lex_state = 146, .external_lex_state = 2},
  [7093] = {.lex_state = 146, .external_lex_state = 2},
  [7094] = {.lex_state = 146, .external_lex_state = 2},
  [7095] = {.lex_state = 146, .external_lex_state = 2},
  [7096] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [7097] = {.lex_state = 146, .external_lex_state = 2},
  [7098] = {.lex_state = 146, .external_lex_state = 2},
  [7099] = {.lex_state = 148, .external_lex_state = 2},
  [7100] = {.lex_state = 146, .external_lex_state = 2},
  [7101] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [7102] = {.lex_state = 146, .external_lex_state = 2},
  [7103] = {.lex_state = 146, .external_lex_state = 2},
  [7104] = {.lex_state = 146, .external_lex_state = 2},
  [7105] = {.lex_state = 146, .external_lex_state = 2},
  [7106] = {.lex_state = 146, .external_lex_state = 2},
  [7107] = {.lex_state = 146, .external_lex_state = 2},
  [7108] = {.lex_state = 146, .external_lex_state = 2},
  [7109] = {.lex_state = 146, .external_lex_state = 2},
  [7110] = {.lex_state = 51, .external_lex_state = 2},
  [7111] = {.lex_state = 146, .external_lex_state = 2},
  [7112] = {.lex_state = 146, .external_lex_state = 2},
  [7113] = {.lex_state = 146, .external_lex_state = 2},
  [7114] = {.lex_state = 146, .external_lex_state = 2},
  [7115] = {.lex_state = 146, .external_lex_state = 2},
  [7116] = {.lex_state = 146, .external_lex_state = 2},
  [7117] = {.lex_state = 146, .external_lex_state = 2},
  [7118] = {.lex_state = 146, .external_lex_state = 2},
  [7119] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [7120] = {.lex_state = 146, .external_lex_state = 2},
  [7121] = {.lex_state = 146, .external_lex_state = 2},
  [7122] = {.lex_state = 146, .external_lex_state = 2},
  [7123] = {.lex_state = 146, .external_lex_state = 2},
  [7124] = {.lex_state = 148, .external_lex_state = 2},
  [7125] = {.lex_state = 146, .external_lex_state = 2},
  [7126] = {.lex_state = 148, .external_lex_state = 2},
  [7127] = {.lex_state = 148, .external_lex_state = 2},
  [7128] = {.lex_state = 146, .external_lex_state = 2},
  [7129] = {.lex_state = 146, .external_lex_state = 2},
  [7130] = {.lex_state = 146, .external_lex_state = 2},
  [7131] = {.lex_state = 146, .external_lex_state = 2},
  [7132] = {.lex_state = 146, .external_lex_state = 2},
  [7133] = {.lex_state = 146, .external_lex_state = 2},
  [7134] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [7135] = {.lex_state = 146, .external_lex_state = 2},
  [7136] = {.lex_state = 146, .external_lex_state = 2},
  [7137] = {.lex_state = 146, .external_lex_state = 2},
  [7138] = {.lex_state = 146, .external_lex_state = 2},
  [7139] = {.lex_state = 146, .external_lex_state = 2},
  [7140] = {.lex_state = 146, .external_lex_state = 2},
  [7141] = {.lex_state = 146, .external_lex_state = 2},
  [7142] = {.lex_state = 11, .external_lex_state = 2},
  [7143] = {.lex_state = 146, .external_lex_state = 2},
  [7144] = {.lex_state = 148, .external_lex_state = 2},
  [7145] = {.lex_state = 148, .external_lex_state = 2},
  [7146] = {.lex_state = 146, .external_lex_state = 2},
  [7147] = {.lex_state = 146, .external_lex_state = 2},
  [7148] = {.lex_state = 148, .external_lex_state = 2},
  [7149] = {.lex_state = 11, .external_lex_state = 2},
  [7150] = {.lex_state = 146, .external_lex_state = 2},
  [7151] = {.lex_state = 146, .external_lex_state = 2},
  [7152] = {.lex_state = 146, .external_lex_state = 2},
  [7153] = {.lex_state = 146, .external_lex_state = 2},
  [7154] = {.lex_state = 148, .external_lex_state = 2},
  [7155] = {.lex_state = 146, .external_lex_state = 2},
  [7156] = {.lex_state = 146, .external_lex_state = 2},
  [7157] = {.lex_state = 146, .external_lex_state = 2},
  [7158] = {.lex_state = 146, .external_lex_state = 9},
  [7159] = {.lex_state = 148, .external_lex_state = 2},
  [7160] = {.lex_state = 11, .external_lex_state = 2},
  [7161] = {.lex_state = 51, .external_lex_state = 2},
  [7162] = {.lex_state = 146, .external_lex_state = 2},
  [7163] = {.lex_state = 146, .external_lex_state = 2},
  [7164] = {.lex_state = 146, .external_lex_state = 2},
  [7165] = {.lex_state = 146, .external_lex_state = 2},
  [7166] = {.lex_state = 146, .external_lex_state = 2},
  [7167] = {.lex_state = 146, .external_lex_state = 2},
  [7168] = {.lex_state = 146, .external_lex_state = 2},
  [7169] = {.lex_state = 146, .external_lex_state = 2},
  [7170] = {.lex_state = 146, .external_lex_state = 2},
  [7171] = {.lex_state = 146, .external_lex_state = 2},
  [7172] = {.lex_state = 146, .external_lex_state = 2},
  [7173] = {.lex_state = 148, .external_lex_state = 2},
  [7174] = {.lex_state = 146, .external_lex_state = 2},
  [7175] = {.lex_state = 146, .external_lex_state = 2},
  [7176] = {.lex_state = 11, .external_lex_state = 2},
  [7177] = {.lex_state = 11, .external_lex_state = 2},
  [7178] = {.lex_state = 146, .external_lex_state = 2},
  [7179] = {.lex_state = 146, .external_lex_state = 2},
  [7180] = {.lex_state = 146, .external_lex_state = 2},
  [7181] = {.lex_state = 146, .external_lex_state = 2},
  [7182] = {.lex_state = 146, .external_lex_state = 2},
  [7183] = {.lex_state = 148, .external_lex_state = 2},
  [7184] = {.lex_state = 146, .external_lex_state = 2},
  [7185] = {.lex_state = 146, .external_lex_state = 2},
  [7186] = {.lex_state = 146, .external_lex_state = 2},
  [7187] = {.lex_state = 146, .external_lex_state = 2},
  [7188] = {.lex_state = 51, .external_lex_state = 2},
  [7189] = {.lex_state = 35, .external_lex_state = 2},
  [7190] = {.lex_state = 146, .external_lex_state = 3},
  [7191] = {.lex_state = 146, .external_lex_state = 2},
  [7192] = {.lex_state = 146, .external_lex_state = 2},
  [7193] = {.lex_state = 146, .external_lex_state = 2},
  [7194] = {.lex_state = 146, .external_lex_state = 2},
  [7195] = {.lex_state = 146, .external_lex_state = 2},
  [7196] = {.lex_state = 146, .external_lex_state = 2},
  [7197] = {.lex_state = 51, .external_lex_state = 2},
  [7198] = {.lex_state = 11, .external_lex_state = 2},
  [7199] = {.lex_state = 146, .external_lex_state = 2},
  [7200] = {.lex_state = 146, .external_lex_state = 2},
  [7201] = {.lex_state = 146, .external_lex_state = 2},
  [7202] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [7203] = {.lex_state = 146, .external_lex_state = 2},
  [7204] = {.lex_state = 146, .external_lex_state = 2},
  [7205] = {.lex_state = 146, .external_lex_state = 2},
  [7206] = {.lex_state = 146, .external_lex_state = 2},
  [7207] = {.lex_state = 146, .external_lex_state = 2},
  [7208] = {.lex_state = 146, .external_lex_state = 2},
  [7209] = {.lex_state = 146, .external_lex_state = 2},
  [7210] = {.lex_state = 146, .external_lex_state = 2},
  [7211] = {.lex_state = 146, .external_lex_state = 2},
  [7212] = {.lex_state = 148, .external_lex_state = 2},
  [7213] = {.lex_state = 146, .external_lex_state = 2},
  [7214] = {.lex_state = 146, .external_lex_state = 2},
  [7215] = {.lex_state = 146, .external_lex_state = 2},
  [7216] = {.lex_state = 146, .external_lex_state = 2},
  [7217] = {.lex_state = 146, .external_lex_state = 2},
  [7218] = {.lex_state = 146, .external_lex_state = 2},
  [7219] = {.lex_state = 146, .external_lex_state = 2},
  [7220] = {.lex_state = 146, .external_lex_state = 2},
  [7221] = {.lex_state = 146, .external_lex_state = 2},
  [7222] = {.lex_state = 146, .external_lex_state = 2},
  [7223] = {.lex_state = 51, .external_lex_state = 2},
  [7224] = {.lex_state = 146, .external_lex_state = 2},
  [7225] = {.lex_state = 146, .external_lex_state = 2},
  [7226] = {.lex_state = 146, .external_lex_state = 2},
  [7227] = {.lex_state = 148, .external_lex_state = 2},
  [7228] = {.lex_state = 148, .external_lex_state = 2},
  [7229] = {.lex_state = 146, .external_lex_state = 2},
  [7230] = {.lex_state = 146, .external_lex_state = 2},
  [7231] = {.lex_state = 146, .external_lex_state = 2},
  [7232] = {.lex_state = 148, .external_lex_state = 2},
  [7233] = {.lex_state = 148, .external_lex_state = 2},
  [7234] = {.lex_state = 146, .external_lex_state = 2},
  [7235] = {.lex_state = 146, .external_lex_state = 2},
  [7236] = {.lex_state = 146, .external_lex_state = 2},
  [7237] = {.lex_state = 146, .external_lex_state = 2},
  [7238] = {.lex_state = 148, .external_lex_state = 2},
  [7239] = {.lex_state = 148, .external_lex_state = 2},
  [7240] = {.lex_state = 148, .external_lex_state = 2},
  [7241] = {.lex_state = 88, .external_lex_state = 2},
  [7242] = {.lex_state = 148, .external_lex_state = 2},
  [7243] = {.lex_state = 146, .external_lex_state = 2},
  [7244] = {.lex_state = 35, .external_lex_state = 2},
  [7245] = {.lex_state = 146, .external_lex_state = 3},
  [7246] = {.lex_state = 146, .external_lex_state = 2},
  [7247] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [7248] = {.lex_state = 11, .external_lex_state = 2},
  [7249] = {.lex_state = 146, .external_lex_state = 2},
  [7250] = {.lex_state = 146, .external_lex_state = 2},
  [7251] = {.lex_state = 148, .external_lex_state = 2},
  [7252] = {.lex_state = 146, .external_lex_state = 2},
  [7253] = {.lex_state = 51, .external_lex_state = 2},
  [7254] = {.lex_state = 146, .external_lex_state = 2},
  [7255] = {.lex_state = 148, .external_lex_state = 2},
  [7256] = {.lex_state = 146, .external_lex_state = 2},
  [7257] = {.lex_state = 146, .external_lex_state = 2},
  [7258] = {.lex_state = 146, .external_lex_state = 2},
  [7259] = {.lex_state = 146, .external_lex_state = 2},
  [7260] = {.lex_state = 146, .external_lex_state = 2},
  [7261] = {.lex_state = 146, .external_lex_state = 2},
  [7262] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [7263] = {.lex_state = 146, .external_lex_state = 2},
  [7264] = {.lex_state = 148, .external_lex_state = 2},
  [7265] = {.lex_state = 146, .external_lex_state = 2},
  [7266] = {.lex_state = 146, .external_lex_state = 2},
  [7267] = {.lex_state = 146, .external_lex_state = 2},
  [7268] = {.lex_state = 146, .external_lex_state = 2},
  [7269] = {.lex_state = 146, .external_lex_state = 2},
  [7270] = {.lex_state = 146, .external_lex_state = 2},
  [7271] = {.lex_state = 146, .external_lex_state = 2},
  [7272] = {.lex_state = 148, .external_lex_state = 2},
  [7273] = {.lex_state = 88, .external_lex_state = 2},
  [7274] = {.lex_state = 146, .external_lex_state = 2},
  [7275] = {.lex_state = 146, .external_lex_state = 2},
  [7276] = {.lex_state = 146, .external_lex_state = 2},
  [7277] = {.lex_state = 88, .external_lex_state = 2},
  [7278] = {.lex_state = 146, .external_lex_state = 2},
  [7279] = {.lex_state = 146, .external_lex_state = 2, .reserved_word_set_id = 1},
  [7280] = {.lex_state = 88, .external_lex_state = 2},
  [7281] = {.lex_state = 88, .external_lex_state = 2},
  [7282] = {.lex_state = 88, .external_lex_state = 2},
  [7283] = {.lex_state = 88, .external_lex_state = 2},
  [7284] = {.lex_state = 88, .external_lex_state = 2},
  [7285] = {.lex_state = 88, .external_lex_state = 2},
  [7286] = {.lex_state = 88, .external_lex_state = 2},
  [7287] = {.lex_state = 88, .external_lex_state = 2},
  [7288] = {.lex_state = 88, .external_lex_state = 2},
  [7289] = {.lex_state = 88, .external_lex_state = 2},
  [7290] = {.lex_state = 88, .external_lex_state = 2},
  [7291] = {.lex_state = 88, .external_lex_state = 2},
  [7292] = {.lex_state = 88, .external_lex_state = 2},
  [7293] = {.lex_state = 146, .external_lex_state = 2},
  [7294] = {.lex_state = 146, .external_lex_state = 2},
  [7295] = {.lex_state = 146, .external_lex_state = 2},
  [7296] = {.lex_state = 146, .external_lex_state = 2},
  [7297] = {.lex_state = 51, .external_lex_state = 2},
  [7298] = {.lex_state = 146, .external_lex_state = 2},
  [7299] = {.lex_state = 148, .external_lex_state = 2},
  [7300] = {.lex_state = 146, .external_lex_state = 2},
  [7301] = {.lex_state = 11, .external_lex_state = 2},
  [7302] = {.lex_state = 146, .external_lex_state = 2},
  [7303] = {.lex_state = 146, .external_lex_state = 2},
  [7304] = {.lex_state = 146, .external_lex_state = 2},
  [7305] = {.lex_state = 146, .external_lex_state = 2},
  [7306] = {.lex_state = 146, .external_lex_state = 2},
  [7307] = {.lex_state = 146, .external_lex_state = 2},
  [7308] = {.lex_state = 146, .external_lex_state = 2},
  [7309] = {.lex_state = 11, .external_lex_state = 2},
  [7310] = {.lex_state = 88, .external_lex_state = 2},
  [7311] = {.lex_state = 146, .external_lex_state = 2},
  [7312] = {.lex_state = 146, .external_lex_state = 2},
  [7313] = {.lex_state = 146, .external_lex_state = 2},
  [7314] = {.lex_state = 146, .external_lex_state = 2},
  [7315] = {.lex_state = 146, .external_lex_state = 2},
  [7316] = {.lex_state = 146, .external_lex_state = 2},
  [7317] = {.lex_state = 146, .external_lex_state = 2},
  [7318] = {.lex_state = 146, .external_lex_state = 2},
  [7319] = {.lex_state = 148, .external_lex_state = 2},
  [7320] = {.lex_state = 146, .external_lex_state = 2},
  [7321] = {.lex_state = 146, .external_lex_state = 2},
  [7322] = {.lex_state = 11, .external_lex_state = 2},
  [7323] = {.lex_state = 146, .external_lex_state = 2},
  [7324] = {.lex_state = 146, .external_lex_state = 2},
  [7325] = {.lex_state = 51, .external_lex_state = 2},
  [7326] = {.lex_state = 148, .external_lex_state = 2},
  [7327] = {.lex_state = 146, .external_lex_state = 2},
  [7328] = {.lex_state = 146, .external_lex_state = 2},
  [7329] = {.lex_state = 146, .external_lex_state = 2},
  [7330] = {.lex_state = 51, .external_lex_state = 2},
  [7331] = {.lex_state = 148, .external_lex_state = 2},
  [7332] = {.lex_state = 146, .external_lex_state = 2},
  [7333] = {.lex_state = 146, .external_lex_state = 2},
  [7334] = {.lex_state = 146, .external_lex_state = 2},
  [7335] = {.lex_state = 146, .external_lex_state = 2},
  [7336] = {.lex_state = 51, .external_lex_state = 2},
  [7337] = {.lex_state = 146, .external_lex_state = 2},
  [7338] = {.lex_state = 146, .external_lex_state = 2},
  [7339] = {.lex_state = 148, .external_lex_state = 2},
  [7340] = {.lex_state = 146, .external_lex_state = 2},
  [7341] = {.lex_state = 51, .external_lex_state = 2},
  [7342] = {.lex_state = 146, .external_lex_state = 2},
  [7343] = {.lex_state = 11, .external_lex_state = 2},
  [7344] = {.lex_state = 146, .external_lex_state = 2},
  [7345] = {.lex_state = 146, .external_lex_state = 2},
  [7346] = {.lex_state = 146, .external_lex_state = 2},
  [7347] = {.lex_state = 148, .external_lex_state = 2},
  [7348] = {(TSStateId)(-1),},
  [7349] = {(TSStateId)(-1),},
};

static const TSSymbol ts_reserved_words[95][MAX_RESERVED_WORD_SET_SIZE] = {
  [1] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [2] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_in,
    anon_sym_as,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [3] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [4] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [5] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [6] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [7] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [8] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_in,
    anon_sym_as,
    anon_sym_then,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [9] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [10] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [11] = {
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [12] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_do,
    anon_sym_when,
  },
  [13] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [14] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
  },
  [15] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [16] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [17] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_if,
    anon_sym_while,
    anon_sym_for,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [18] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_while,
    anon_sym_for,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [19] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [20] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [21] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [22] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [23] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [24] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
  },
  [25] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [26] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_done,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [27] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_then,
    anon_sym_else,
    anon_sym_do,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_when,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [28] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
  },
  [29] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [30] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [31] = {
    anon_sym_let,
    anon_sym_and,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_include,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [32] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [33] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [34] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [35] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [36] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [37] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [38] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [39] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [40] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_if,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [41] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [42] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [43] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [44] = {
    anon_sym_let,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [45] = {
    anon_sym_let,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [46] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [47] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [48] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [49] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [50] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [51] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_for,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [52] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [53] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_while,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
  },
  [54] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [55] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [56] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [57] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_type,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [58] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [59] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [60] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [61] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [62] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [63] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [64] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [65] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [66] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [67] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [68] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [69] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [70] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [71] = {
    anon_sym_let,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_with,
    anon_sym_object,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [72] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [73] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [74] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [75] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [76] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [77] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [78] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [79] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [80] = {
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [81] = {
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [82] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [83] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [84] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_virtual,
    anon_sym_sig,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [85] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_end,
    anon_sym_with,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [86] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [87] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [88] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [89] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [90] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [91] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [92] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_mutable,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [93] = {
    anon_sym_let,
    anon_sym_rec,
    anon_sym_and,
    anon_sym_external,
    anon_sym_type,
    anon_sym_private,
    anon_sym_of,
    anon_sym_true,
    anon_sym_false,
    anon_sym_constraint,
    anon_sym_exception,
    anon_sym_module,
    anon_sym_open,
    anon_sym_include,
    anon_sym_class,
    anon_sym_virtual,
    anon_sym_val,
    anon_sym_sig,
    anon_sym_end,
    anon_sym_with,
    anon_sym_functor,
    anon_sym_struct,
    anon_sym_object,
    anon_sym_inherit,
    anon_sym_method,
    anon_sym_in,
    anon_sym_fun,
    anon_sym_as,
    anon_sym_initializer,
    anon_sym_if,
    anon_sym_then,
    anon_sym_else,
    anon_sym_while,
    anon_sym_do,
    anon_sym_done,
    anon_sym_for,
    anon_sym_to,
    anon_sym_downto,
    anon_sym_match,
    anon_sym_when,
    anon_sym_function,
    anon_sym_try,
    anon_sym_assert,
    anon_sym_lazy,
    anon_sym_new,
    anon_sym_begin,
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
    anon_sym_or,
  },
  [94] = {
    anon_sym_lsl,
    anon_sym_lsr,
    anon_sym_asr,
    anon_sym_mod,
    anon_sym_land,
    anon_sym_lor,
    anon_sym_lxor,
  },
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [sym_attribute] = STATE(0),
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym__lowercase_identifier] = ACTIONS(1),
    [anon_sym_SEMI_SEMI] = ACTIONS(1),
    [anon_sym_let] = ACTIONS(1),
    [anon_sym_rec] = ACTIONS(1),
    [anon_sym_and] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_TILDE] = ACTIONS(1),
    [anon_sym_QMARK] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [anon_sym_external] = ACTIONS(1),
    [anon_sym_type] = ACTIONS(1),
    [anon_sym_nonrec] = ACTIONS(1),
    [anon_sym_COLON_EQ] = ACTIONS(1),
    [anon_sym_private] = ACTIONS(1),
    [anon_sym_DOT_DOT] = ACTIONS(1),
    [anon_sym_PLUS_EQ] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_BANG] = ACTIONS(1),
    [anon_sym__] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_of] = ACTIONS(1),
    [anon_sym_COLON2] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_DASH_GT] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_true] = ACTIONS(1),
    [anon_sym_false] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_SEMI] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [anon_sym_mutable] = ACTIONS(1),
    [anon_sym_constraint] = ACTIONS(1),
    [anon_sym_exception] = ACTIONS(1),
    [anon_sym_module] = ACTIONS(1),
    [anon_sym_open] = ACTIONS(1),
    [anon_sym_include] = ACTIONS(1),
    [anon_sym_class] = ACTIONS(1),
    [anon_sym_virtual] = ACTIONS(1),
    [anon_sym_val] = ACTIONS(1),
    [anon_sym_sig] = ACTIONS(1),
    [anon_sym_end] = ACTIONS(1),
    [anon_sym_with] = ACTIONS(1),
    [anon_sym_functor] = ACTIONS(1),
    [anon_sym_struct] = ACTIONS(1),
    [anon_sym_COLON_GT] = ACTIONS(1),
    [anon_sym_object] = ACTIONS(1),
    [anon_sym_inherit] = ACTIONS(1),
    [anon_sym_method] = ACTIONS(1),
    [anon_sym_in] = ACTIONS(1),
    [anon_sym_fun] = ACTIONS(1),
    [anon_sym_as] = ACTIONS(1),
    [anon_sym_initializer] = ACTIONS(1),
    [anon_sym_LBRACK_GT] = ACTIONS(1),
    [anon_sym_LBRACK_LT] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_AMP] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_POUND] = ACTIONS(1),
    [anon_sym_COLON_COLON] = ACTIONS(1),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1),
    [anon_sym_LT_DASH] = ACTIONS(1),
    [anon_sym_if] = ACTIONS(1),
    [anon_sym_then] = ACTIONS(1),
    [anon_sym_else] = ACTIONS(1),
    [anon_sym_while] = ACTIONS(1),
    [anon_sym_do] = ACTIONS(1),
    [anon_sym_done] = ACTIONS(1),
    [anon_sym_for] = ACTIONS(1),
    [anon_sym_to] = ACTIONS(1),
    [anon_sym_downto] = ACTIONS(1),
    [anon_sym_match] = ACTIONS(1),
    [anon_sym_when] = ACTIONS(1),
    [anon_sym_function] = ACTIONS(1),
    [anon_sym_try] = ACTIONS(1),
    [anon_sym_assert] = ACTIONS(1),
    [anon_sym_lazy] = ACTIONS(1),
    [anon_sym_new] = ACTIONS(1),
    [anon_sym_LBRACE_LT] = ACTIONS(1),
    [anon_sym_begin] = ACTIONS(1),
    [sym_ocamlyacc_value] = ACTIONS(1),
    [anon_sym_effect] = ACTIONS(1),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [aux_sym_number_token1] = ACTIONS(1),
    [aux_sym_signed_number_token1] = ACTIONS(1),
    [anon_sym_SQUOTE] = ACTIONS(1),
    [anon_sym_SQUOTE2] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [aux_sym_string_content_token2] = ACTIONS(1),
    [aux_sym_string_content_token4] = ACTIONS(1),
    [aux_sym_string_content_token5] = ACTIONS(1),
    [aux_sym_escape_sequence_token1] = ACTIONS(1),
    [aux_sym_escape_sequence_token2] = ACTIONS(1),
    [aux_sym_escape_sequence_token3] = ACTIONS(1),
    [aux_sym_escape_sequence_token4] = ACTIONS(1),
    [sym_pretty_printing_indication] = ACTIONS(1),
    [aux_sym_sign_operator_token1] = ACTIONS(1),
    [sym_hash_operator] = ACTIONS(1),
    [anon_sym_lsl] = ACTIONS(1),
    [anon_sym_lsr] = ACTIONS(1),
    [anon_sym_asr] = ACTIONS(1),
    [aux_sym_pow_operator_token1] = ACTIONS(1),
    [anon_sym_mod] = ACTIONS(1),
    [anon_sym_land] = ACTIONS(1),
    [anon_sym_lor] = ACTIONS(1),
    [anon_sym_lxor] = ACTIONS(1),
    [sym_concat_operator] = ACTIONS(1),
    [anon_sym_AMP_AMP] = ACTIONS(1),
    [anon_sym_or] = ACTIONS(1),
    [anon_sym_PIPE_PIPE] = ACTIONS(1),
    [sym_assign_operator] = ACTIONS(1),
    [sym_let_operator] = ACTIONS(1),
    [sym_let_and_operator] = ACTIONS(1),
    [sym_match_operator] = ACTIONS(1),
    [sym__uppercase_identifier] = ACTIONS(1),
    [aux_sym_directive_token1] = ACTIONS(1),
    [aux_sym_type_variable_token1] = ACTIONS(1),
    [aux_sym_tag_token1] = ACTIONS(1),
    [aux_sym_attribute_id_token1] = ACTIONS(1),
    [sym_comment] = ACTIONS(5),
    [sym__left_quoted_string_delimiter] = ACTIONS(1),
    [sym__right_quoted_string_delimiter] = ACTIONS(1),
    [sym_line_number_directive] = ACTIONS(5),
    [sym__null] = ACTIONS(1),
    [sym__error_sentinel] = ACTIONS(1),
  },
  [STATE(1)] = {
    [sym_type] = STATE(6912),
    [sym__delimited_type] = STATE(2574),
    [sym__simple_type] = STATE(3803),
    [sym__type] = STATE(6331),
    [sym_function_type] = STATE(2650),
    [sym_labeled_argument_type] = STATE(7246),
    [sym__proper_tuple_type] = STATE(4229),
    [sym__labeled_tuple_type] = STATE(2650),
    [sym_labeled_tuple_element_type] = STATE(6978),
    [sym_constructed_type] = STATE(2574),
    [sym_aliased_type] = STATE(2650),
    [sym_local_open_type] = STATE(2574),
    [sym_polymorphic_variant_type] = STATE(2586),
    [sym_package_type] = STATE(2586),
    [sym_object_type] = STATE(2574),
    [sym_hash_type] = STATE(2574),
    [sym_parenthesized_type] = STATE(2586),
    [sym_attribute] = STATE(1),
    [sym_extension] = STATE(2574),
    [sym_quoted_extension] = STATE(2574),
    [sym_extended_module_path] = STATE(6301),
    [sym_type_constructor_path] = STATE(2574),
    [sym_type_variable] = STATE(2574),
    [sym__lowercase_identifier] = ACTIONS(7),
    [anon_sym_QMARK] = ACTIONS(9),
    [anon_sym_LPAREN] = ACTIONS(11),
    [anon_sym_LBRACK] = ACTIONS(13),
    [anon_sym_LBRACK_GT] = ACTIONS(15),
    [anon_sym_LBRACK_LT] = ACTIONS(17),
    [anon_sym_LT] = ACTIONS(19),
    [anon_sym_POUND] = ACTIONS(21),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [sym__uppercase_identifier] = ACTIONS(29),
    [aux_sym_type_variable_token1] = ACTIONS(31),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(2)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1894),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2202),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(2),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__attribute] = STATE(519),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_RPAREN] = ACTIONS(41),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_PIPE] = ACTIONS(43),
    [anon_sym_COLON2] = ACTIONS(43),
    [anon_sym_DASH_GT] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_RBRACK] = ACTIONS(41),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_SEMI] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(41),
    [anon_sym_constraint] = ACTIONS(43),
    [anon_sym_val] = ACTIONS(43),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_with] = ACTIONS(43),
    [anon_sym_COLON_GT] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_inherit] = ACTIONS(43),
    [anon_sym_method] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_initializer] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_then] = ACTIONS(43),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_GT_RBRACE] = ACTIONS(41),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(43),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(81),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(83),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(3)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1659),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1775),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(3),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__attribute] = STATE(517),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_SEMI_SEMI] = ACTIONS(41),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_and] = ACTIONS(43),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_external] = ACTIONS(43),
    [anon_sym_type] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_PIPE] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_RBRACK] = ACTIONS(41),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_SEMI] = ACTIONS(43),
    [anon_sym_exception] = ACTIONS(43),
    [anon_sym_module] = ACTIONS(43),
    [anon_sym_open] = ACTIONS(43),
    [anon_sym_include] = ACTIONS(43),
    [anon_sym_class] = ACTIONS(43),
    [anon_sym_val] = ACTIONS(43),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_in] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(43),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(147),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_let_and_operator] = ACTIONS(41),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_directive_token1] = ACTIONS(41),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(4)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1696),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1474),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(4),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__attribute] = STATE(507),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_SEMI_SEMI] = ACTIONS(41),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_and] = ACTIONS(43),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_external] = ACTIONS(43),
    [anon_sym_type] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_PIPE] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_RBRACK] = ACTIONS(41),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_SEMI] = ACTIONS(43),
    [anon_sym_exception] = ACTIONS(43),
    [anon_sym_module] = ACTIONS(43),
    [anon_sym_open] = ACTIONS(43),
    [anon_sym_include] = ACTIONS(43),
    [anon_sym_class] = ACTIONS(43),
    [anon_sym_val] = ACTIONS(43),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(43),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(147),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_let_and_operator] = ACTIONS(41),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_directive_token1] = ACTIONS(41),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(5)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(1941),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1474),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(5),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__attribute] = STATE(524),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(41),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(43),
    [anon_sym_type] = ACTIONS(43),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_PIPE] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(41),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_SEMI] = ACTIONS(43),
    [anon_sym_exception] = ACTIONS(43),
    [anon_sym_module] = ACTIONS(43),
    [anon_sym_open] = ACTIONS(43),
    [anon_sym_include] = ACTIONS(43),
    [anon_sym_class] = ACTIONS(43),
    [anon_sym_val] = ACTIONS(43),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(43),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(41),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(6)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(6),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(6879),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(243),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(7)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(7),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(7017),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(269),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(8)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(8),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(7267),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(271),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(9)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(9),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(7053),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(273),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(10)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(10),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(6938),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(275),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(11)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(11),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(6940),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(277),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(12)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(12),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(6403),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(279),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(13)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(13),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(6589),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(281),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(14)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(14),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(6822),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(283),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(15)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(15),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(6775),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(285),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(16)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(16),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(7095),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(287),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(17)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(17),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(6375),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(289),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(18)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(18),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(6498),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(291),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(19)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(19),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(6752),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(293),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(20)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(20),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(7002),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(295),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(21)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(21),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(7156),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(297),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(22)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(22),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(7337),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(299),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(23)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(23),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(6608),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(301),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(24)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(24),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(6485),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(303),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(25)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(25),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(6659),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(305),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(26)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(26),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(6747),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(307),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(27)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(27),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(6823),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(309),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(28)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(28),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(6389),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(311),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(29)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(29),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(6935),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(313),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(30)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(30),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(7014),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(315),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(31)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(31),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(7070),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(317),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(32)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(32),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(7100),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(319),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(33)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(33),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(7109),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(321),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(34)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(34),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(7132),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(323),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(35)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(35),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(6525),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(325),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(36)] = {
    [sym__structure] = STATE(6923),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(36),
    [sym_floating_attribute] = STATE(4107),
    [sym_attribute_payload] = STATE(6444),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_QMARK] = ACTIONS(235),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_COLON2] = ACTIONS(241),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(327),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(37)] = {
    [sym__structure] = STATE(7255),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(37),
    [sym_floating_attribute] = STATE(4107),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_end] = ACTIONS(329),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(38)] = {
    [sym_expression_item] = STATE(4111),
    [sym_toplevel_directive] = STATE(4111),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(4111),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(38),
    [sym_floating_attribute] = STATE(4107),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(331),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_end] = ACTIONS(333),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(39)] = {
    [sym_expression_item] = STATE(2600),
    [sym_toplevel_directive] = STATE(2600),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2600),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(39),
    [sym_floating_attribute] = STATE(4107),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(335),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_end] = ACTIONS(337),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(40)] = {
    [sym_expression_item] = STATE(4111),
    [sym_toplevel_directive] = STATE(4111),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(4111),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(40),
    [sym_floating_attribute] = STATE(4107),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(339),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_end] = ACTIONS(341),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(41)] = {
    [sym__structure] = STATE(6875),
    [sym_expression_item] = STATE(2591),
    [sym_toplevel_directive] = STATE(2591),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(2591),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(41),
    [sym_floating_attribute] = STATE(4107),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(39),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_end] = ACTIONS(343),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(42)] = {
    [sym_expression_item] = STATE(4111),
    [sym_toplevel_directive] = STATE(4111),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(4111),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(42),
    [sym_floating_attribute] = STATE(4107),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(345),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_end] = ACTIONS(347),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(43)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2524),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2202),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(43),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(501),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_SEMI] = ACTIONS(41),
    [anon_sym_constraint] = ACTIONS(43),
    [anon_sym_val] = ACTIONS(43),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_inherit] = ACTIONS(43),
    [anon_sym_method] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_initializer] = ACTIONS(43),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(43),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(393),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(44)] = {
    [sym_expression_item] = STATE(4111),
    [sym_toplevel_directive] = STATE(4111),
    [sym__local_structure_item] = STATE(4110),
    [sym__structure_item] = STATE(4111),
    [sym_value_definition] = STATE(4038),
    [sym_external] = STATE(4107),
    [sym_type_definition] = STATE(4107),
    [sym_exception_definition] = STATE(4107),
    [sym_module_definition] = STATE(4107),
    [sym_module_type_definition] = STATE(4107),
    [sym_open_module] = STATE(4107),
    [sym_include_module] = STATE(4110),
    [sym_class_definition] = STATE(4107),
    [sym_class_type_definition] = STATE(4107),
    [sym_value_specification] = STATE(4110),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2178),
    [sym__sequence_expression] = STATE(3783),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(44),
    [sym_floating_attribute] = STATE(4107),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym_item_extension] = STATE(4107),
    [sym_quoted_item_extension] = STATE(4107),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_directive] = STATE(2634),
    [sym_tag] = STATE(1052),
    [aux_sym__structure_repeat1] = STATE(1626),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(231),
    [anon_sym_let] = ACTIONS(233),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(237),
    [anon_sym_type] = ACTIONS(239),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(245),
    [anon_sym_module] = ACTIONS(247),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(251),
    [anon_sym_class] = ACTIONS(253),
    [anon_sym_val] = ACTIONS(255),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(265),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(267),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(45)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2507),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2487),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(45),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(513),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(41),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_COLON2] = ACTIONS(43),
    [anon_sym_DASH_GT] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_RBRACK] = ACTIONS(41),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_SEMI] = ACTIONS(41),
    [anon_sym_RBRACE] = ACTIONS(41),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_with] = ACTIONS(43),
    [anon_sym_COLON_GT] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_then] = ACTIONS(43),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_GT_RBRACE] = ACTIONS(41),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(46)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6979),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(46),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(6833),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(419),
    [anon_sym_DOT] = ACTIONS(421),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(423),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(431),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(433),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(449),
    [sym_let_and_operator] = ACTIONS(433),
    [sym_match_operator] = ACTIONS(451),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(47)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(5020),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(47),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(7135),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(453),
    [anon_sym_DOT] = ACTIONS(455),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(457),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(459),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(461),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(463),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(465),
    [sym_let_and_operator] = ACTIONS(463),
    [sym_match_operator] = ACTIONS(467),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(48)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(4955),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(48),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(7187),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(469),
    [anon_sym_DOT] = ACTIONS(471),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(473),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(475),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(477),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(479),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(481),
    [sym_let_and_operator] = ACTIONS(479),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(49)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7022),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(49),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(6769),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(485),
    [anon_sym_DOT] = ACTIONS(487),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(473),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(475),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(489),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(491),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(493),
    [sym_let_and_operator] = ACTIONS(491),
    [sym_match_operator] = ACTIONS(495),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(50)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7022),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(50),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(7187),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(469),
    [anon_sym_DOT] = ACTIONS(471),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(473),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(475),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(477),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(479),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(481),
    [sym_let_and_operator] = ACTIONS(479),
    [sym_match_operator] = ACTIONS(483),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(51)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(4955),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(51),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(6769),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(485),
    [anon_sym_DOT] = ACTIONS(487),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(473),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(475),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(489),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(491),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(493),
    [sym_let_and_operator] = ACTIONS(491),
    [sym_match_operator] = ACTIONS(495),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(52)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6365),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(52),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(6636),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(497),
    [anon_sym_DOT] = ACTIONS(499),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(501),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(503),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(505),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(507),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(509),
    [sym_let_and_operator] = ACTIONS(507),
    [sym_match_operator] = ACTIONS(511),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(53)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(4861),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(53),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(6636),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(497),
    [anon_sym_DOT] = ACTIONS(499),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(501),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(503),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(505),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(507),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(509),
    [sym_let_and_operator] = ACTIONS(507),
    [sym_match_operator] = ACTIONS(511),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(54)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7133),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(54),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(6676),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(513),
    [anon_sym_DOT] = ACTIONS(515),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(517),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(519),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(521),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(523),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(525),
    [sym_let_and_operator] = ACTIONS(523),
    [sym_match_operator] = ACTIONS(527),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(55)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(4907),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(55),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(6676),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(513),
    [anon_sym_DOT] = ACTIONS(515),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(517),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(519),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(521),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(523),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(525),
    [sym_let_and_operator] = ACTIONS(523),
    [sym_match_operator] = ACTIONS(527),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(56)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6820),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(56),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(6756),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(529),
    [anon_sym_DOT] = ACTIONS(531),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(533),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(535),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(539),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(541),
    [sym_let_and_operator] = ACTIONS(539),
    [sym_match_operator] = ACTIONS(543),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(57)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(4945),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(57),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(6756),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(529),
    [anon_sym_DOT] = ACTIONS(531),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(533),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(535),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(537),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(539),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(541),
    [sym_let_and_operator] = ACTIONS(539),
    [sym_match_operator] = ACTIONS(543),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(58)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7140),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(58),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(7135),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(453),
    [anon_sym_DOT] = ACTIONS(455),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(457),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(459),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(461),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(463),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(465),
    [sym_let_and_operator] = ACTIONS(463),
    [sym_match_operator] = ACTIONS(467),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(59)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(4974),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(59),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(6833),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(419),
    [anon_sym_DOT] = ACTIONS(421),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(423),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(427),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(431),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(433),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(449),
    [sym_let_and_operator] = ACTIONS(433),
    [sym_match_operator] = ACTIONS(451),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(60)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6462),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(60),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(6892),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(545),
    [anon_sym_DOT] = ACTIONS(547),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(549),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(551),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(553),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(555),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(557),
    [sym_let_and_operator] = ACTIONS(555),
    [sym_match_operator] = ACTIONS(559),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(61)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(4999),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(61),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(6892),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(545),
    [anon_sym_DOT] = ACTIONS(547),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(549),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(551),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(553),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(555),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(557),
    [sym_let_and_operator] = ACTIONS(555),
    [sym_match_operator] = ACTIONS(559),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(62)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6717),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(62),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(6949),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(561),
    [anon_sym_DOT] = ACTIONS(563),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(565),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(567),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(569),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(571),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(573),
    [sym_let_and_operator] = ACTIONS(571),
    [sym_match_operator] = ACTIONS(575),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(63)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(5021),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(63),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(6949),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(561),
    [anon_sym_DOT] = ACTIONS(563),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(565),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(567),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(569),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(571),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(573),
    [sym_let_and_operator] = ACTIONS(571),
    [sym_match_operator] = ACTIONS(575),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(64)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6988),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(64),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(7023),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(577),
    [anon_sym_DOT] = ACTIONS(579),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(581),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(583),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(585),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(587),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(589),
    [sym_let_and_operator] = ACTIONS(587),
    [sym_match_operator] = ACTIONS(591),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(65)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(4912),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(65),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(7023),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(577),
    [anon_sym_DOT] = ACTIONS(579),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(581),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(583),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(585),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(587),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(589),
    [sym_let_and_operator] = ACTIONS(587),
    [sym_match_operator] = ACTIONS(591),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(66)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(4936),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(66),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(7320),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(593),
    [anon_sym_DOT] = ACTIONS(595),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(597),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(599),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(601),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(603),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(605),
    [sym_let_and_operator] = ACTIONS(603),
    [sym_match_operator] = ACTIONS(607),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(67)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6472),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(67),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__infix_operator] = STATE(7320),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(593),
    [anon_sym_DOT] = ACTIONS(595),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_module] = ACTIONS(597),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_COLON_COLON] = ACTIONS(599),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(429),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(601),
    [aux_sym_sign_operator_token1] = ACTIONS(429),
    [sym_hash_operator] = ACTIONS(603),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(605),
    [sym_let_and_operator] = ACTIONS(603),
    [sym_match_operator] = ACTIONS(607),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(68)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2793),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2939),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(68),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__attribute] = STATE(508),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_and] = ACTIONS(43),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_PIPE] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_SEMI] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_in] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_let_and_operator] = ACTIONS(41),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(69)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2848),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2939),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(69),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__attribute] = STATE(510),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_PIPE] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_SEMI] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_done] = ACTIONS(43),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_to] = ACTIONS(43),
    [anon_sym_downto] = ACTIONS(43),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(70)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2881),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2487),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(70),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__attribute] = STATE(500),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_and] = ACTIONS(43),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_SEMI] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_in] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_let_and_operator] = ACTIONS(41),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(71)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2946),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2939),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(71),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(514),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_SEMI] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_done] = ACTIONS(43),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_to] = ACTIONS(43),
    [anon_sym_downto] = ACTIONS(43),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(72)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3062),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2487),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(72),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(509),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_PIPE] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_SEMI] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_PIPE_RBRACK] = ACTIONS(41),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(73)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(3073),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2202),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(73),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__attribute] = STATE(523),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_PIPE] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_SEMI] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(74)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3159),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2487),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(74),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(503),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_COMMA] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_SEMI] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_AMP] = ACTIONS(43),
    [anon_sym_COLON_COLON] = ACTIONS(41),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_else] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(85),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(89),
    [anon_sym_lsl] = ACTIONS(43),
    [anon_sym_lsr] = ACTIONS(43),
    [anon_sym_asr] = ACTIONS(43),
    [aux_sym_pow_operator_token1] = ACTIONS(41),
    [anon_sym_mod] = ACTIONS(43),
    [anon_sym_land] = ACTIONS(43),
    [anon_sym_lor] = ACTIONS(43),
    [anon_sym_lxor] = ACTIONS(43),
    [aux_sym_mult_operator_token1] = ACTIONS(43),
    [aux_sym_add_operator_token1] = ACTIONS(43),
    [sym_concat_operator] = ACTIONS(41),
    [sym_rel_operator] = ACTIONS(43),
    [anon_sym_AMP_AMP] = ACTIONS(43),
    [anon_sym_or] = ACTIONS(43),
    [anon_sym_PIPE_PIPE] = ACTIONS(43),
    [sym_assign_operator] = ACTIONS(41),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(75)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1824),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1775),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(75),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__attribute] = STATE(506),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_SEMI_SEMI] = ACTIONS(41),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_and] = ACTIONS(43),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_external] = ACTIONS(43),
    [anon_sym_type] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_RBRACK] = ACTIONS(41),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_exception] = ACTIONS(43),
    [anon_sym_module] = ACTIONS(43),
    [anon_sym_open] = ACTIONS(43),
    [anon_sym_include] = ACTIONS(43),
    [anon_sym_class] = ACTIONS(43),
    [anon_sym_val] = ACTIONS(43),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_in] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(43),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(147),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_let_and_operator] = ACTIONS(41),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_directive_token1] = ACTIONS(41),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(76)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1993),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1474),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(76),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__attribute] = STATE(505),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_SEMI_SEMI] = ACTIONS(41),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_and] = ACTIONS(43),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_external] = ACTIONS(43),
    [anon_sym_type] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_RBRACK] = ACTIONS(41),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_exception] = ACTIONS(43),
    [anon_sym_module] = ACTIONS(43),
    [anon_sym_open] = ACTIONS(43),
    [anon_sym_include] = ACTIONS(43),
    [anon_sym_class] = ACTIONS(43),
    [anon_sym_val] = ACTIONS(43),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(43),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(147),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_let_and_operator] = ACTIONS(41),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_directive_token1] = ACTIONS(41),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(77)] = {
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(772),
    [sym_typed_expression] = STATE(794),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_labeled_argument] = STATE(845),
    [sym_prefix_expression] = STATE(794),
    [sym_hash_expression] = STATE(794),
    [sym_field_get_expression] = STATE(794),
    [sym_array_get_expression] = STATE(794),
    [sym_string_get_expression] = STATE(794),
    [sym_bigarray_get_expression] = STATE(794),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(77),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [aux_sym_class_application_repeat1] = STATE(82),
    [sym__lowercase_identifier] = ACTIONS(743),
    [anon_sym_TILDE] = ACTIONS(745),
    [anon_sym_QMARK] = ACTIONS(745),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(747),
    [anon_sym_COMMA] = ACTIONS(747),
    [anon_sym_PIPE] = ACTIONS(749),
    [anon_sym_COLON2] = ACTIONS(749),
    [anon_sym_DOT] = ACTIONS(751),
    [anon_sym_DASH_GT] = ACTIONS(749),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_RBRACK] = ACTIONS(747),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_SEMI] = ACTIONS(747),
    [anon_sym_RBRACE] = ACTIONS(747),
    [anon_sym_constraint] = ACTIONS(749),
    [anon_sym_val] = ACTIONS(749),
    [anon_sym_end] = ACTIONS(749),
    [anon_sym_with] = ACTIONS(749),
    [anon_sym_COLON_GT] = ACTIONS(747),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_inherit] = ACTIONS(749),
    [anon_sym_method] = ACTIONS(749),
    [anon_sym_initializer] = ACTIONS(749),
    [anon_sym_AMP] = ACTIONS(749),
    [anon_sym_POUND] = ACTIONS(753),
    [anon_sym_COLON_COLON] = ACTIONS(747),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_PIPE_RBRACK] = ACTIONS(747),
    [anon_sym_then] = ACTIONS(749),
    [anon_sym_else] = ACTIONS(749),
    [anon_sym_do] = ACTIONS(749),
    [anon_sym_done] = ACTIONS(749),
    [anon_sym_to] = ACTIONS(749),
    [anon_sym_downto] = ACTIONS(749),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_GT_RBRACE] = ACTIONS(747),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(749),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(747),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(393),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(747),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(747),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(749),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(749),
    [sym_hash_operator] = ACTIONS(755),
    [anon_sym_lsl] = ACTIONS(749),
    [anon_sym_lsr] = ACTIONS(749),
    [anon_sym_asr] = ACTIONS(749),
    [aux_sym_pow_operator_token1] = ACTIONS(747),
    [anon_sym_mod] = ACTIONS(749),
    [anon_sym_land] = ACTIONS(749),
    [anon_sym_lor] = ACTIONS(749),
    [anon_sym_lxor] = ACTIONS(749),
    [aux_sym_mult_operator_token1] = ACTIONS(749),
    [aux_sym_add_operator_token1] = ACTIONS(749),
    [sym_concat_operator] = ACTIONS(747),
    [sym_rel_operator] = ACTIONS(749),
    [anon_sym_AMP_AMP] = ACTIONS(749),
    [anon_sym_or] = ACTIONS(749),
    [anon_sym_PIPE_PIPE] = ACTIONS(749),
    [sym_assign_operator] = ACTIONS(747),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(78)] = {
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(772),
    [sym_typed_expression] = STATE(794),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_labeled_argument] = STATE(845),
    [sym_prefix_expression] = STATE(794),
    [sym_hash_expression] = STATE(794),
    [sym_field_get_expression] = STATE(794),
    [sym_array_get_expression] = STATE(794),
    [sym_string_get_expression] = STATE(794),
    [sym_bigarray_get_expression] = STATE(794),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(78),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [aux_sym_class_application_repeat1] = STATE(78),
    [sym__lowercase_identifier] = ACTIONS(757),
    [anon_sym_and] = ACTIONS(760),
    [anon_sym_TILDE] = ACTIONS(762),
    [anon_sym_QMARK] = ACTIONS(762),
    [anon_sym_LPAREN] = ACTIONS(765),
    [anon_sym_RPAREN] = ACTIONS(768),
    [anon_sym_COMMA] = ACTIONS(768),
    [anon_sym_PIPE] = ACTIONS(760),
    [anon_sym_COLON2] = ACTIONS(760),
    [anon_sym_DASH_GT] = ACTIONS(760),
    [anon_sym_LBRACK] = ACTIONS(770),
    [anon_sym_RBRACK] = ACTIONS(768),
    [anon_sym_true] = ACTIONS(773),
    [anon_sym_false] = ACTIONS(773),
    [anon_sym_LBRACE] = ACTIONS(776),
    [anon_sym_SEMI] = ACTIONS(768),
    [anon_sym_RBRACE] = ACTIONS(768),
    [anon_sym_constraint] = ACTIONS(760),
    [anon_sym_val] = ACTIONS(760),
    [anon_sym_end] = ACTIONS(760),
    [anon_sym_with] = ACTIONS(760),
    [anon_sym_COLON_GT] = ACTIONS(768),
    [anon_sym_object] = ACTIONS(779),
    [anon_sym_inherit] = ACTIONS(760),
    [anon_sym_method] = ACTIONS(760),
    [anon_sym_in] = ACTIONS(760),
    [anon_sym_as] = ACTIONS(760),
    [anon_sym_initializer] = ACTIONS(760),
    [anon_sym_AMP] = ACTIONS(760),
    [anon_sym_COLON_COLON] = ACTIONS(768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(782),
    [anon_sym_PIPE_RBRACK] = ACTIONS(768),
    [anon_sym_then] = ACTIONS(760),
    [anon_sym_else] = ACTIONS(760),
    [anon_sym_do] = ACTIONS(760),
    [anon_sym_done] = ACTIONS(760),
    [anon_sym_to] = ACTIONS(760),
    [anon_sym_downto] = ACTIONS(760),
    [anon_sym_new] = ACTIONS(785),
    [anon_sym_LBRACE_LT] = ACTIONS(788),
    [anon_sym_GT_RBRACE] = ACTIONS(768),
    [anon_sym_begin] = ACTIONS(791),
    [sym_ocamlyacc_value] = ACTIONS(794),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(760),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(768),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(797),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(800),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(768),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(768),
    [aux_sym_number_token1] = ACTIONS(803),
    [aux_sym_signed_number_token1] = ACTIONS(760),
    [anon_sym_SQUOTE] = ACTIONS(806),
    [anon_sym_DQUOTE] = ACTIONS(809),
    [sym_prefix_operator] = ACTIONS(812),
    [aux_sym_sign_operator_token1] = ACTIONS(760),
    [anon_sym_lsl] = ACTIONS(760),
    [anon_sym_lsr] = ACTIONS(760),
    [anon_sym_asr] = ACTIONS(760),
    [aux_sym_pow_operator_token1] = ACTIONS(768),
    [anon_sym_mod] = ACTIONS(760),
    [anon_sym_land] = ACTIONS(760),
    [anon_sym_lor] = ACTIONS(760),
    [anon_sym_lxor] = ACTIONS(760),
    [aux_sym_mult_operator_token1] = ACTIONS(760),
    [aux_sym_add_operator_token1] = ACTIONS(760),
    [sym_concat_operator] = ACTIONS(768),
    [sym_rel_operator] = ACTIONS(760),
    [anon_sym_AMP_AMP] = ACTIONS(760),
    [anon_sym_or] = ACTIONS(760),
    [anon_sym_PIPE_PIPE] = ACTIONS(760),
    [sym_assign_operator] = ACTIONS(768),
    [sym__uppercase_identifier] = ACTIONS(815),
    [aux_sym_tag_token1] = ACTIONS(818),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(79)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2146),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1474),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(79),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__attribute] = STATE(502),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_SEMI_SEMI] = ACTIONS(41),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(43),
    [anon_sym_type] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(41),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(43),
    [anon_sym_module] = ACTIONS(43),
    [anon_sym_open] = ACTIONS(43),
    [anon_sym_include] = ACTIONS(43),
    [anon_sym_class] = ACTIONS(43),
    [anon_sym_val] = ACTIONS(43),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(43),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(41),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(80)] = {
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(866),
    [sym_typed_expression] = STATE(910),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_labeled_argument] = STATE(1037),
    [sym_prefix_expression] = STATE(910),
    [sym_hash_expression] = STATE(910),
    [sym_field_get_expression] = STATE(910),
    [sym_array_get_expression] = STATE(910),
    [sym_string_get_expression] = STATE(910),
    [sym_bigarray_get_expression] = STATE(910),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(80),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [aux_sym_class_application_repeat1] = STATE(84),
    [sym__lowercase_identifier] = ACTIONS(821),
    [anon_sym_TILDE] = ACTIONS(823),
    [anon_sym_QMARK] = ACTIONS(823),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_RPAREN] = ACTIONS(747),
    [anon_sym_COMMA] = ACTIONS(747),
    [anon_sym_PIPE] = ACTIONS(749),
    [anon_sym_COLON2] = ACTIONS(749),
    [anon_sym_DOT] = ACTIONS(825),
    [anon_sym_DASH_GT] = ACTIONS(749),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_RBRACK] = ACTIONS(747),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_SEMI] = ACTIONS(747),
    [anon_sym_RBRACE] = ACTIONS(747),
    [anon_sym_constraint] = ACTIONS(749),
    [anon_sym_val] = ACTIONS(749),
    [anon_sym_end] = ACTIONS(749),
    [anon_sym_with] = ACTIONS(749),
    [anon_sym_COLON_GT] = ACTIONS(747),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_inherit] = ACTIONS(749),
    [anon_sym_method] = ACTIONS(749),
    [anon_sym_initializer] = ACTIONS(749),
    [anon_sym_AMP] = ACTIONS(749),
    [anon_sym_POUND] = ACTIONS(827),
    [anon_sym_COLON_COLON] = ACTIONS(747),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_then] = ACTIONS(749),
    [anon_sym_else] = ACTIONS(749),
    [anon_sym_do] = ACTIONS(749),
    [anon_sym_done] = ACTIONS(749),
    [anon_sym_to] = ACTIONS(749),
    [anon_sym_downto] = ACTIONS(749),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_GT_RBRACE] = ACTIONS(747),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(749),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(747),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(81),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(83),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(747),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(747),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(749),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(749),
    [sym_hash_operator] = ACTIONS(829),
    [anon_sym_lsl] = ACTIONS(749),
    [anon_sym_lsr] = ACTIONS(749),
    [anon_sym_asr] = ACTIONS(749),
    [aux_sym_pow_operator_token1] = ACTIONS(747),
    [anon_sym_mod] = ACTIONS(749),
    [anon_sym_land] = ACTIONS(749),
    [anon_sym_lor] = ACTIONS(749),
    [anon_sym_lxor] = ACTIONS(749),
    [aux_sym_mult_operator_token1] = ACTIONS(749),
    [aux_sym_add_operator_token1] = ACTIONS(749),
    [sym_concat_operator] = ACTIONS(747),
    [sym_rel_operator] = ACTIONS(749),
    [anon_sym_AMP_AMP] = ACTIONS(749),
    [anon_sym_or] = ACTIONS(749),
    [anon_sym_PIPE_PIPE] = ACTIONS(749),
    [sym_assign_operator] = ACTIONS(747),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(81)] = {
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(948),
    [sym_typed_expression] = STATE(987),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_labeled_argument] = STATE(1131),
    [sym_prefix_expression] = STATE(987),
    [sym_hash_expression] = STATE(987),
    [sym_field_get_expression] = STATE(987),
    [sym_array_get_expression] = STATE(987),
    [sym_string_get_expression] = STATE(987),
    [sym_bigarray_get_expression] = STATE(987),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(81),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [aux_sym_class_application_repeat1] = STATE(86),
    [sym__lowercase_identifier] = ACTIONS(831),
    [anon_sym_SEMI_SEMI] = ACTIONS(747),
    [anon_sym_let] = ACTIONS(749),
    [anon_sym_and] = ACTIONS(749),
    [anon_sym_TILDE] = ACTIONS(833),
    [anon_sym_QMARK] = ACTIONS(833),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_external] = ACTIONS(749),
    [anon_sym_type] = ACTIONS(749),
    [anon_sym_COMMA] = ACTIONS(747),
    [anon_sym_PIPE] = ACTIONS(749),
    [anon_sym_DOT] = ACTIONS(835),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_RBRACK] = ACTIONS(747),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_SEMI] = ACTIONS(749),
    [anon_sym_exception] = ACTIONS(749),
    [anon_sym_module] = ACTIONS(749),
    [anon_sym_open] = ACTIONS(749),
    [anon_sym_include] = ACTIONS(749),
    [anon_sym_class] = ACTIONS(749),
    [anon_sym_val] = ACTIONS(749),
    [anon_sym_end] = ACTIONS(749),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_in] = ACTIONS(749),
    [anon_sym_AMP] = ACTIONS(749),
    [anon_sym_POUND] = ACTIONS(837),
    [anon_sym_COLON_COLON] = ACTIONS(747),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_else] = ACTIONS(749),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(749),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(747),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(147),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(747),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(747),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(749),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(749),
    [sym_hash_operator] = ACTIONS(839),
    [anon_sym_lsl] = ACTIONS(749),
    [anon_sym_lsr] = ACTIONS(749),
    [anon_sym_asr] = ACTIONS(749),
    [aux_sym_pow_operator_token1] = ACTIONS(747),
    [anon_sym_mod] = ACTIONS(749),
    [anon_sym_land] = ACTIONS(749),
    [anon_sym_lor] = ACTIONS(749),
    [anon_sym_lxor] = ACTIONS(749),
    [aux_sym_mult_operator_token1] = ACTIONS(749),
    [aux_sym_add_operator_token1] = ACTIONS(749),
    [sym_concat_operator] = ACTIONS(747),
    [sym_rel_operator] = ACTIONS(749),
    [anon_sym_AMP_AMP] = ACTIONS(749),
    [anon_sym_or] = ACTIONS(749),
    [anon_sym_PIPE_PIPE] = ACTIONS(749),
    [sym_assign_operator] = ACTIONS(747),
    [sym_let_operator] = ACTIONS(747),
    [sym_let_and_operator] = ACTIONS(747),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_directive_token1] = ACTIONS(749),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(82)] = {
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(772),
    [sym_typed_expression] = STATE(794),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_labeled_argument] = STATE(845),
    [sym_prefix_expression] = STATE(794),
    [sym_hash_expression] = STATE(794),
    [sym_field_get_expression] = STATE(794),
    [sym_array_get_expression] = STATE(794),
    [sym_string_get_expression] = STATE(794),
    [sym_bigarray_get_expression] = STATE(794),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(82),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [aux_sym_class_application_repeat1] = STATE(78),
    [sym__lowercase_identifier] = ACTIONS(743),
    [anon_sym_TILDE] = ACTIONS(745),
    [anon_sym_QMARK] = ACTIONS(745),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(841),
    [anon_sym_COMMA] = ACTIONS(841),
    [anon_sym_PIPE] = ACTIONS(843),
    [anon_sym_COLON2] = ACTIONS(843),
    [anon_sym_DASH_GT] = ACTIONS(843),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_RBRACK] = ACTIONS(841),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_SEMI] = ACTIONS(841),
    [anon_sym_RBRACE] = ACTIONS(841),
    [anon_sym_constraint] = ACTIONS(843),
    [anon_sym_val] = ACTIONS(843),
    [anon_sym_end] = ACTIONS(843),
    [anon_sym_with] = ACTIONS(843),
    [anon_sym_COLON_GT] = ACTIONS(841),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_inherit] = ACTIONS(843),
    [anon_sym_method] = ACTIONS(843),
    [anon_sym_initializer] = ACTIONS(843),
    [anon_sym_AMP] = ACTIONS(843),
    [anon_sym_COLON_COLON] = ACTIONS(841),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_PIPE_RBRACK] = ACTIONS(841),
    [anon_sym_then] = ACTIONS(843),
    [anon_sym_else] = ACTIONS(843),
    [anon_sym_do] = ACTIONS(843),
    [anon_sym_done] = ACTIONS(843),
    [anon_sym_to] = ACTIONS(843),
    [anon_sym_downto] = ACTIONS(843),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_GT_RBRACE] = ACTIONS(841),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(843),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(841),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(393),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(841),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(841),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(843),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(843),
    [anon_sym_lsl] = ACTIONS(843),
    [anon_sym_lsr] = ACTIONS(843),
    [anon_sym_asr] = ACTIONS(843),
    [aux_sym_pow_operator_token1] = ACTIONS(841),
    [anon_sym_mod] = ACTIONS(843),
    [anon_sym_land] = ACTIONS(843),
    [anon_sym_lor] = ACTIONS(843),
    [anon_sym_lxor] = ACTIONS(843),
    [aux_sym_mult_operator_token1] = ACTIONS(843),
    [aux_sym_add_operator_token1] = ACTIONS(843),
    [sym_concat_operator] = ACTIONS(841),
    [sym_rel_operator] = ACTIONS(843),
    [anon_sym_AMP_AMP] = ACTIONS(843),
    [anon_sym_or] = ACTIONS(843),
    [anon_sym_PIPE_PIPE] = ACTIONS(843),
    [sym_assign_operator] = ACTIONS(841),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(83)] = {
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(866),
    [sym_typed_expression] = STATE(910),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_labeled_argument] = STATE(1037),
    [sym_prefix_expression] = STATE(910),
    [sym_hash_expression] = STATE(910),
    [sym_field_get_expression] = STATE(910),
    [sym_array_get_expression] = STATE(910),
    [sym_string_get_expression] = STATE(910),
    [sym_bigarray_get_expression] = STATE(910),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(83),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [aux_sym_class_application_repeat1] = STATE(83),
    [sym__lowercase_identifier] = ACTIONS(845),
    [anon_sym_TILDE] = ACTIONS(848),
    [anon_sym_QMARK] = ACTIONS(848),
    [anon_sym_LPAREN] = ACTIONS(851),
    [anon_sym_RPAREN] = ACTIONS(768),
    [anon_sym_COMMA] = ACTIONS(768),
    [anon_sym_PIPE] = ACTIONS(760),
    [anon_sym_COLON2] = ACTIONS(760),
    [anon_sym_DASH_GT] = ACTIONS(760),
    [anon_sym_LBRACK] = ACTIONS(854),
    [anon_sym_RBRACK] = ACTIONS(768),
    [anon_sym_true] = ACTIONS(857),
    [anon_sym_false] = ACTIONS(857),
    [anon_sym_LBRACE] = ACTIONS(860),
    [anon_sym_SEMI] = ACTIONS(768),
    [anon_sym_RBRACE] = ACTIONS(768),
    [anon_sym_constraint] = ACTIONS(760),
    [anon_sym_val] = ACTIONS(760),
    [anon_sym_end] = ACTIONS(760),
    [anon_sym_with] = ACTIONS(760),
    [anon_sym_COLON_GT] = ACTIONS(768),
    [anon_sym_object] = ACTIONS(863),
    [anon_sym_inherit] = ACTIONS(760),
    [anon_sym_method] = ACTIONS(760),
    [anon_sym_initializer] = ACTIONS(760),
    [anon_sym_AMP] = ACTIONS(760),
    [anon_sym_COLON_COLON] = ACTIONS(768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(866),
    [anon_sym_then] = ACTIONS(760),
    [anon_sym_else] = ACTIONS(760),
    [anon_sym_do] = ACTIONS(760),
    [anon_sym_done] = ACTIONS(760),
    [anon_sym_to] = ACTIONS(760),
    [anon_sym_downto] = ACTIONS(760),
    [anon_sym_new] = ACTIONS(869),
    [anon_sym_LBRACE_LT] = ACTIONS(872),
    [anon_sym_GT_RBRACE] = ACTIONS(768),
    [anon_sym_begin] = ACTIONS(875),
    [sym_ocamlyacc_value] = ACTIONS(878),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(760),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(768),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(881),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(884),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(768),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(768),
    [aux_sym_number_token1] = ACTIONS(887),
    [aux_sym_signed_number_token1] = ACTIONS(760),
    [anon_sym_SQUOTE] = ACTIONS(890),
    [anon_sym_DQUOTE] = ACTIONS(893),
    [sym_prefix_operator] = ACTIONS(896),
    [aux_sym_sign_operator_token1] = ACTIONS(760),
    [anon_sym_lsl] = ACTIONS(760),
    [anon_sym_lsr] = ACTIONS(760),
    [anon_sym_asr] = ACTIONS(760),
    [aux_sym_pow_operator_token1] = ACTIONS(768),
    [anon_sym_mod] = ACTIONS(760),
    [anon_sym_land] = ACTIONS(760),
    [anon_sym_lor] = ACTIONS(760),
    [anon_sym_lxor] = ACTIONS(760),
    [aux_sym_mult_operator_token1] = ACTIONS(760),
    [aux_sym_add_operator_token1] = ACTIONS(760),
    [sym_concat_operator] = ACTIONS(768),
    [sym_rel_operator] = ACTIONS(760),
    [anon_sym_AMP_AMP] = ACTIONS(760),
    [anon_sym_or] = ACTIONS(760),
    [anon_sym_PIPE_PIPE] = ACTIONS(760),
    [sym_assign_operator] = ACTIONS(768),
    [sym__uppercase_identifier] = ACTIONS(899),
    [aux_sym_tag_token1] = ACTIONS(902),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(84)] = {
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(866),
    [sym_typed_expression] = STATE(910),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_labeled_argument] = STATE(1037),
    [sym_prefix_expression] = STATE(910),
    [sym_hash_expression] = STATE(910),
    [sym_field_get_expression] = STATE(910),
    [sym_array_get_expression] = STATE(910),
    [sym_string_get_expression] = STATE(910),
    [sym_bigarray_get_expression] = STATE(910),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(84),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [aux_sym_class_application_repeat1] = STATE(83),
    [sym__lowercase_identifier] = ACTIONS(821),
    [anon_sym_TILDE] = ACTIONS(823),
    [anon_sym_QMARK] = ACTIONS(823),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_RPAREN] = ACTIONS(841),
    [anon_sym_COMMA] = ACTIONS(841),
    [anon_sym_PIPE] = ACTIONS(843),
    [anon_sym_COLON2] = ACTIONS(843),
    [anon_sym_DASH_GT] = ACTIONS(843),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_RBRACK] = ACTIONS(841),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_SEMI] = ACTIONS(841),
    [anon_sym_RBRACE] = ACTIONS(841),
    [anon_sym_constraint] = ACTIONS(843),
    [anon_sym_val] = ACTIONS(843),
    [anon_sym_end] = ACTIONS(843),
    [anon_sym_with] = ACTIONS(843),
    [anon_sym_COLON_GT] = ACTIONS(841),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_inherit] = ACTIONS(843),
    [anon_sym_method] = ACTIONS(843),
    [anon_sym_initializer] = ACTIONS(843),
    [anon_sym_AMP] = ACTIONS(843),
    [anon_sym_COLON_COLON] = ACTIONS(841),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_then] = ACTIONS(843),
    [anon_sym_else] = ACTIONS(843),
    [anon_sym_do] = ACTIONS(843),
    [anon_sym_done] = ACTIONS(843),
    [anon_sym_to] = ACTIONS(843),
    [anon_sym_downto] = ACTIONS(843),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_GT_RBRACE] = ACTIONS(841),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(843),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(841),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(81),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(83),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(841),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(841),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(843),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(843),
    [anon_sym_lsl] = ACTIONS(843),
    [anon_sym_lsr] = ACTIONS(843),
    [anon_sym_asr] = ACTIONS(843),
    [aux_sym_pow_operator_token1] = ACTIONS(841),
    [anon_sym_mod] = ACTIONS(843),
    [anon_sym_land] = ACTIONS(843),
    [anon_sym_lor] = ACTIONS(843),
    [anon_sym_lxor] = ACTIONS(843),
    [aux_sym_mult_operator_token1] = ACTIONS(843),
    [aux_sym_add_operator_token1] = ACTIONS(843),
    [sym_concat_operator] = ACTIONS(841),
    [sym_rel_operator] = ACTIONS(843),
    [anon_sym_AMP_AMP] = ACTIONS(843),
    [anon_sym_or] = ACTIONS(843),
    [anon_sym_PIPE_PIPE] = ACTIONS(843),
    [sym_assign_operator] = ACTIONS(841),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(85)] = {
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(948),
    [sym_typed_expression] = STATE(987),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_labeled_argument] = STATE(1131),
    [sym_prefix_expression] = STATE(987),
    [sym_hash_expression] = STATE(987),
    [sym_field_get_expression] = STATE(987),
    [sym_array_get_expression] = STATE(987),
    [sym_string_get_expression] = STATE(987),
    [sym_bigarray_get_expression] = STATE(987),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(85),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [aux_sym_class_application_repeat1] = STATE(85),
    [sym__lowercase_identifier] = ACTIONS(905),
    [anon_sym_SEMI_SEMI] = ACTIONS(768),
    [anon_sym_let] = ACTIONS(760),
    [anon_sym_and] = ACTIONS(760),
    [anon_sym_TILDE] = ACTIONS(908),
    [anon_sym_QMARK] = ACTIONS(908),
    [anon_sym_LPAREN] = ACTIONS(911),
    [anon_sym_external] = ACTIONS(760),
    [anon_sym_type] = ACTIONS(760),
    [anon_sym_COMMA] = ACTIONS(768),
    [anon_sym_PIPE] = ACTIONS(760),
    [anon_sym_LBRACK] = ACTIONS(914),
    [anon_sym_RBRACK] = ACTIONS(768),
    [anon_sym_true] = ACTIONS(917),
    [anon_sym_false] = ACTIONS(917),
    [anon_sym_LBRACE] = ACTIONS(920),
    [anon_sym_SEMI] = ACTIONS(760),
    [anon_sym_exception] = ACTIONS(760),
    [anon_sym_module] = ACTIONS(760),
    [anon_sym_open] = ACTIONS(760),
    [anon_sym_include] = ACTIONS(760),
    [anon_sym_class] = ACTIONS(760),
    [anon_sym_val] = ACTIONS(760),
    [anon_sym_end] = ACTIONS(760),
    [anon_sym_object] = ACTIONS(923),
    [anon_sym_in] = ACTIONS(760),
    [anon_sym_AMP] = ACTIONS(760),
    [anon_sym_COLON_COLON] = ACTIONS(768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(926),
    [anon_sym_else] = ACTIONS(760),
    [anon_sym_new] = ACTIONS(929),
    [anon_sym_LBRACE_LT] = ACTIONS(932),
    [anon_sym_begin] = ACTIONS(935),
    [sym_ocamlyacc_value] = ACTIONS(938),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(760),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(768),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(941),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(944),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(768),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(768),
    [aux_sym_number_token1] = ACTIONS(947),
    [aux_sym_signed_number_token1] = ACTIONS(760),
    [anon_sym_SQUOTE] = ACTIONS(950),
    [anon_sym_DQUOTE] = ACTIONS(953),
    [sym_prefix_operator] = ACTIONS(956),
    [aux_sym_sign_operator_token1] = ACTIONS(760),
    [anon_sym_lsl] = ACTIONS(760),
    [anon_sym_lsr] = ACTIONS(760),
    [anon_sym_asr] = ACTIONS(760),
    [aux_sym_pow_operator_token1] = ACTIONS(768),
    [anon_sym_mod] = ACTIONS(760),
    [anon_sym_land] = ACTIONS(760),
    [anon_sym_lor] = ACTIONS(760),
    [anon_sym_lxor] = ACTIONS(760),
    [aux_sym_mult_operator_token1] = ACTIONS(760),
    [aux_sym_add_operator_token1] = ACTIONS(760),
    [sym_concat_operator] = ACTIONS(768),
    [sym_rel_operator] = ACTIONS(760),
    [anon_sym_AMP_AMP] = ACTIONS(760),
    [anon_sym_or] = ACTIONS(760),
    [anon_sym_PIPE_PIPE] = ACTIONS(760),
    [sym_assign_operator] = ACTIONS(768),
    [sym_let_operator] = ACTIONS(768),
    [sym_let_and_operator] = ACTIONS(768),
    [sym__uppercase_identifier] = ACTIONS(959),
    [aux_sym_directive_token1] = ACTIONS(768),
    [aux_sym_tag_token1] = ACTIONS(962),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(86)] = {
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(948),
    [sym_typed_expression] = STATE(987),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_labeled_argument] = STATE(1131),
    [sym_prefix_expression] = STATE(987),
    [sym_hash_expression] = STATE(987),
    [sym_field_get_expression] = STATE(987),
    [sym_array_get_expression] = STATE(987),
    [sym_string_get_expression] = STATE(987),
    [sym_bigarray_get_expression] = STATE(987),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(86),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [aux_sym_class_application_repeat1] = STATE(85),
    [sym__lowercase_identifier] = ACTIONS(831),
    [anon_sym_SEMI_SEMI] = ACTIONS(841),
    [anon_sym_let] = ACTIONS(843),
    [anon_sym_and] = ACTIONS(843),
    [anon_sym_TILDE] = ACTIONS(833),
    [anon_sym_QMARK] = ACTIONS(833),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_external] = ACTIONS(843),
    [anon_sym_type] = ACTIONS(843),
    [anon_sym_COMMA] = ACTIONS(841),
    [anon_sym_PIPE] = ACTIONS(843),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_RBRACK] = ACTIONS(841),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_SEMI] = ACTIONS(843),
    [anon_sym_exception] = ACTIONS(843),
    [anon_sym_module] = ACTIONS(843),
    [anon_sym_open] = ACTIONS(843),
    [anon_sym_include] = ACTIONS(843),
    [anon_sym_class] = ACTIONS(843),
    [anon_sym_val] = ACTIONS(843),
    [anon_sym_end] = ACTIONS(843),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_in] = ACTIONS(843),
    [anon_sym_AMP] = ACTIONS(843),
    [anon_sym_COLON_COLON] = ACTIONS(841),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_else] = ACTIONS(843),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(843),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(841),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(147),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(149),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(841),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(841),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(843),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(843),
    [anon_sym_lsl] = ACTIONS(843),
    [anon_sym_lsr] = ACTIONS(843),
    [anon_sym_asr] = ACTIONS(843),
    [aux_sym_pow_operator_token1] = ACTIONS(841),
    [anon_sym_mod] = ACTIONS(843),
    [anon_sym_land] = ACTIONS(843),
    [anon_sym_lor] = ACTIONS(843),
    [anon_sym_lxor] = ACTIONS(843),
    [aux_sym_mult_operator_token1] = ACTIONS(843),
    [aux_sym_add_operator_token1] = ACTIONS(843),
    [sym_concat_operator] = ACTIONS(841),
    [sym_rel_operator] = ACTIONS(843),
    [anon_sym_AMP_AMP] = ACTIONS(843),
    [anon_sym_or] = ACTIONS(843),
    [anon_sym_PIPE_PIPE] = ACTIONS(843),
    [sym_assign_operator] = ACTIONS(841),
    [sym_let_operator] = ACTIONS(841),
    [sym_let_and_operator] = ACTIONS(841),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_directive_token1] = ACTIONS(841),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(87)] = {
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(1078),
    [sym_typed_expression] = STATE(1052),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_labeled_argument] = STATE(1144),
    [sym_prefix_expression] = STATE(1052),
    [sym_hash_expression] = STATE(1052),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(87),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [aux_sym_class_application_repeat1] = STATE(91),
    [sym__lowercase_identifier] = ACTIONS(965),
    [anon_sym_SEMI_SEMI] = ACTIONS(747),
    [anon_sym_let] = ACTIONS(749),
    [anon_sym_TILDE] = ACTIONS(967),
    [anon_sym_QMARK] = ACTIONS(967),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(749),
    [anon_sym_type] = ACTIONS(749),
    [anon_sym_COMMA] = ACTIONS(747),
    [anon_sym_PIPE] = ACTIONS(749),
    [anon_sym_DOT] = ACTIONS(969),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(747),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_SEMI] = ACTIONS(749),
    [anon_sym_exception] = ACTIONS(749),
    [anon_sym_module] = ACTIONS(749),
    [anon_sym_open] = ACTIONS(749),
    [anon_sym_include] = ACTIONS(749),
    [anon_sym_class] = ACTIONS(749),
    [anon_sym_val] = ACTIONS(749),
    [anon_sym_end] = ACTIONS(749),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_AMP] = ACTIONS(749),
    [anon_sym_POUND] = ACTIONS(971),
    [anon_sym_COLON_COLON] = ACTIONS(747),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_else] = ACTIONS(749),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(749),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(747),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(747),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(747),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(749),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(749),
    [sym_hash_operator] = ACTIONS(973),
    [anon_sym_lsl] = ACTIONS(749),
    [anon_sym_lsr] = ACTIONS(749),
    [anon_sym_asr] = ACTIONS(749),
    [aux_sym_pow_operator_token1] = ACTIONS(747),
    [anon_sym_mod] = ACTIONS(749),
    [anon_sym_land] = ACTIONS(749),
    [anon_sym_lor] = ACTIONS(749),
    [anon_sym_lxor] = ACTIONS(749),
    [aux_sym_mult_operator_token1] = ACTIONS(749),
    [aux_sym_add_operator_token1] = ACTIONS(749),
    [sym_concat_operator] = ACTIONS(747),
    [sym_rel_operator] = ACTIONS(749),
    [anon_sym_AMP_AMP] = ACTIONS(749),
    [anon_sym_or] = ACTIONS(749),
    [anon_sym_PIPE_PIPE] = ACTIONS(749),
    [sym_assign_operator] = ACTIONS(747),
    [sym_let_operator] = ACTIONS(747),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(749),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(88)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2619),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1474),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(88),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(516),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_constraint] = ACTIONS(43),
    [anon_sym_val] = ACTIONS(43),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_inherit] = ACTIONS(43),
    [anon_sym_method] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_initializer] = ACTIONS(43),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(43),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(393),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(41),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(89)] = {
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(1078),
    [sym_typed_expression] = STATE(1052),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_labeled_argument] = STATE(1144),
    [sym_prefix_expression] = STATE(1052),
    [sym_hash_expression] = STATE(1052),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(89),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [aux_sym_class_application_repeat1] = STATE(89),
    [sym__lowercase_identifier] = ACTIONS(975),
    [anon_sym_SEMI_SEMI] = ACTIONS(768),
    [anon_sym_let] = ACTIONS(760),
    [anon_sym_and] = ACTIONS(760),
    [anon_sym_TILDE] = ACTIONS(978),
    [anon_sym_QMARK] = ACTIONS(978),
    [anon_sym_LPAREN] = ACTIONS(981),
    [anon_sym_external] = ACTIONS(760),
    [anon_sym_type] = ACTIONS(760),
    [anon_sym_COMMA] = ACTIONS(768),
    [anon_sym_PIPE] = ACTIONS(760),
    [anon_sym_LBRACK] = ACTIONS(984),
    [anon_sym_RBRACK] = ACTIONS(768),
    [anon_sym_true] = ACTIONS(987),
    [anon_sym_false] = ACTIONS(987),
    [anon_sym_LBRACE] = ACTIONS(990),
    [anon_sym_SEMI] = ACTIONS(760),
    [anon_sym_exception] = ACTIONS(760),
    [anon_sym_module] = ACTIONS(760),
    [anon_sym_open] = ACTIONS(760),
    [anon_sym_include] = ACTIONS(760),
    [anon_sym_class] = ACTIONS(760),
    [anon_sym_val] = ACTIONS(760),
    [anon_sym_end] = ACTIONS(760),
    [anon_sym_object] = ACTIONS(993),
    [anon_sym_AMP] = ACTIONS(760),
    [anon_sym_COLON_COLON] = ACTIONS(768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(996),
    [anon_sym_else] = ACTIONS(760),
    [anon_sym_new] = ACTIONS(999),
    [anon_sym_LBRACE_LT] = ACTIONS(1002),
    [anon_sym_begin] = ACTIONS(1005),
    [sym_ocamlyacc_value] = ACTIONS(1008),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(760),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(768),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1011),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1014),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(768),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(768),
    [aux_sym_number_token1] = ACTIONS(1017),
    [aux_sym_signed_number_token1] = ACTIONS(760),
    [anon_sym_SQUOTE] = ACTIONS(1020),
    [anon_sym_DQUOTE] = ACTIONS(1023),
    [sym_prefix_operator] = ACTIONS(1026),
    [aux_sym_sign_operator_token1] = ACTIONS(760),
    [anon_sym_lsl] = ACTIONS(760),
    [anon_sym_lsr] = ACTIONS(760),
    [anon_sym_asr] = ACTIONS(760),
    [aux_sym_pow_operator_token1] = ACTIONS(768),
    [anon_sym_mod] = ACTIONS(760),
    [anon_sym_land] = ACTIONS(760),
    [anon_sym_lor] = ACTIONS(760),
    [anon_sym_lxor] = ACTIONS(760),
    [aux_sym_mult_operator_token1] = ACTIONS(760),
    [aux_sym_add_operator_token1] = ACTIONS(760),
    [sym_concat_operator] = ACTIONS(768),
    [sym_rel_operator] = ACTIONS(760),
    [anon_sym_AMP_AMP] = ACTIONS(760),
    [anon_sym_or] = ACTIONS(760),
    [anon_sym_PIPE_PIPE] = ACTIONS(760),
    [sym_assign_operator] = ACTIONS(768),
    [sym_let_operator] = ACTIONS(768),
    [sym__uppercase_identifier] = ACTIONS(1029),
    [aux_sym_directive_token1] = ACTIONS(768),
    [aux_sym_tag_token1] = ACTIONS(1032),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(90)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2647),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1474),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(90),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(499),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(41),
    [anon_sym_COLON2] = ACTIONS(43),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_RBRACK] = ACTIONS(41),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_RBRACE] = ACTIONS(41),
    [anon_sym_end] = ACTIONS(43),
    [anon_sym_with] = ACTIONS(43),
    [anon_sym_COLON_GT] = ACTIONS(41),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_then] = ACTIONS(43),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(91)] = {
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(1078),
    [sym_typed_expression] = STATE(1052),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_labeled_argument] = STATE(1144),
    [sym_prefix_expression] = STATE(1052),
    [sym_hash_expression] = STATE(1052),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(91),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [aux_sym_class_application_repeat1] = STATE(89),
    [sym__lowercase_identifier] = ACTIONS(965),
    [anon_sym_SEMI_SEMI] = ACTIONS(841),
    [anon_sym_let] = ACTIONS(843),
    [anon_sym_TILDE] = ACTIONS(967),
    [anon_sym_QMARK] = ACTIONS(967),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(843),
    [anon_sym_type] = ACTIONS(843),
    [anon_sym_COMMA] = ACTIONS(841),
    [anon_sym_PIPE] = ACTIONS(843),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(841),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_SEMI] = ACTIONS(843),
    [anon_sym_exception] = ACTIONS(843),
    [anon_sym_module] = ACTIONS(843),
    [anon_sym_open] = ACTIONS(843),
    [anon_sym_include] = ACTIONS(843),
    [anon_sym_class] = ACTIONS(843),
    [anon_sym_val] = ACTIONS(843),
    [anon_sym_end] = ACTIONS(843),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_AMP] = ACTIONS(843),
    [anon_sym_COLON_COLON] = ACTIONS(841),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_else] = ACTIONS(843),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(843),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(841),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(841),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(841),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(843),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(843),
    [anon_sym_lsl] = ACTIONS(843),
    [anon_sym_lsr] = ACTIONS(843),
    [anon_sym_asr] = ACTIONS(843),
    [aux_sym_pow_operator_token1] = ACTIONS(841),
    [anon_sym_mod] = ACTIONS(843),
    [anon_sym_land] = ACTIONS(843),
    [anon_sym_lor] = ACTIONS(843),
    [anon_sym_lxor] = ACTIONS(843),
    [aux_sym_mult_operator_token1] = ACTIONS(843),
    [aux_sym_add_operator_token1] = ACTIONS(843),
    [sym_concat_operator] = ACTIONS(841),
    [sym_rel_operator] = ACTIONS(843),
    [anon_sym_AMP_AMP] = ACTIONS(843),
    [anon_sym_or] = ACTIONS(843),
    [anon_sym_PIPE_PIPE] = ACTIONS(843),
    [sym_assign_operator] = ACTIONS(841),
    [sym_let_operator] = ACTIONS(841),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(841),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(92)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2937),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2487),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(92),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__attribute] = STATE(512),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_and] = ACTIONS(43),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_in] = ACTIONS(43),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(41),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_let_and_operator] = ACTIONS(41),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(93)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2960),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(1775),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(93),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(521),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_done] = ACTIONS(43),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_to] = ACTIONS(43),
    [anon_sym_downto] = ACTIONS(43),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(94)] = {
    [sym__abstract_type] = STATE(7054),
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4781),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(94),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__infix_operator] = STATE(6980),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_RPAREN] = ACTIONS(1041),
    [anon_sym_type] = ACTIONS(1043),
    [anon_sym_DOT] = ACTIONS(1045),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_module] = ACTIONS(1055),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1059),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(1077),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(1077),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(1077),
    [sym_let_and_operator] = ACTIONS(1077),
    [sym_match_operator] = ACTIONS(1077),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(95)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6520),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(95),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(196),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1083),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(96)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6382),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(96),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(197),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1085),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(97)] = {
    [sym__abstract_type] = STATE(7054),
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4697),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(97),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__infix_operator] = STATE(7342),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_RPAREN] = ACTIONS(1087),
    [anon_sym_type] = ACTIONS(1043),
    [anon_sym_DOT] = ACTIONS(1089),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_module] = ACTIONS(1091),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(1095),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(1095),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(1095),
    [sym_let_and_operator] = ACTIONS(1095),
    [sym_match_operator] = ACTIONS(1095),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(98)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7183),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(98),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(199),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1097),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(99)] = {
    [sym__abstract_type] = STATE(7054),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4768),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(99),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__infix_operator] = STATE(7187),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(1101),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(469),
    [anon_sym_type] = ACTIONS(1043),
    [anon_sym_DOT] = ACTIONS(471),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_module] = ACTIONS(1111),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(479),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(479),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(479),
    [sym_let_and_operator] = ACTIONS(479),
    [sym_match_operator] = ACTIONS(479),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(100)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6348),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(100),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(194),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1121),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(101)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6559),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(101),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(203),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1123),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(102)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6831),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(102),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(190),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1125),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(103)] = {
    [sym__abstract_type] = STATE(6905),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4640),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(103),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__infix_operator] = STATE(6518),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(1101),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(1127),
    [anon_sym_type] = ACTIONS(1043),
    [anon_sym_DOT] = ACTIONS(1129),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_module] = ACTIONS(1131),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(1135),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(1135),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(1135),
    [sym_let_and_operator] = ACTIONS(1135),
    [sym_match_operator] = ACTIONS(1135),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(104)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7027),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(104),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(201),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1137),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(105)] = {
    [sym__abstract_type] = STATE(7054),
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4687),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(105),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__infix_operator] = STATE(7187),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_RPAREN] = ACTIONS(469),
    [anon_sym_type] = ACTIONS(1043),
    [anon_sym_DOT] = ACTIONS(471),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_module] = ACTIONS(1111),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(479),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(479),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(479),
    [sym_let_and_operator] = ACTIONS(479),
    [sym_match_operator] = ACTIONS(479),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(106)] = {
    [sym__abstract_type] = STATE(7054),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4690),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(106),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__infix_operator] = STATE(7169),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(1101),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(1139),
    [anon_sym_type] = ACTIONS(1043),
    [anon_sym_DOT] = ACTIONS(1141),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_module] = ACTIONS(1143),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(1147),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(1147),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(1147),
    [sym_let_and_operator] = ACTIONS(1147),
    [sym_match_operator] = ACTIONS(1147),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(107)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7238),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(107),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(173),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1149),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(108)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6466),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(108),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(183),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1151),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(109)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6913),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(109),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(211),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1153),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(110)] = {
    [sym__abstract_type] = STATE(7054),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4672),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(110),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__infix_operator] = STATE(6517),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(1101),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(1155),
    [anon_sym_type] = ACTIONS(1043),
    [anon_sym_DOT] = ACTIONS(1157),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_module] = ACTIONS(1159),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(1161),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(1161),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(1161),
    [sym_let_and_operator] = ACTIONS(1161),
    [sym_match_operator] = ACTIONS(1161),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(111)] = {
    [sym__abstract_type] = STATE(7054),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4672),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(111),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__infix_operator] = STATE(6517),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(1101),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(1155),
    [anon_sym_type] = ACTIONS(1043),
    [anon_sym_DOT] = ACTIONS(1157),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_module] = ACTIONS(1159),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1163),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(1161),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(1161),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(1161),
    [sym_let_and_operator] = ACTIONS(1161),
    [sym_match_operator] = ACTIONS(1161),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(112)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6382),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(112),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(189),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1165),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(113)] = {
    [sym__abstract_type] = STATE(7054),
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4774),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(113),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__infix_operator] = STATE(7169),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_RPAREN] = ACTIONS(1139),
    [anon_sym_type] = ACTIONS(1043),
    [anon_sym_DOT] = ACTIONS(1141),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_module] = ACTIONS(1143),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(1147),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(1147),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(1147),
    [sym_let_and_operator] = ACTIONS(1147),
    [sym_match_operator] = ACTIONS(1147),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(114)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6561),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(114),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(467),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(115)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6760),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(115),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(491),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(116)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2800),
    [sym__sequence_expression] = STATE(2922),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_refutation_case] = STATE(2922),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(116),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_DOT] = ACTIONS(1167),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(117)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2876),
    [sym__sequence_expression] = STATE(2894),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_refutation_case] = STATE(2894),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(117),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_DOT] = ACTIONS(1167),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(118)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1848),
    [sym__sequence_expression] = STATE(2245),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2204),
    [sym_refutation_case] = STATE(2245),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(118),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_DOT] = ACTIONS(1169),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(119)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2002),
    [sym__sequence_expression] = STATE(2167),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_refutation_case] = STATE(2167),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(119),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_DOT] = ACTIONS(1171),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(120)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6098),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(120),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(288),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(121)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3222),
    [sym__sequence_expression] = STATE(6134),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(121),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(293),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(122)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2800),
    [sym__sequence_expression] = STATE(2894),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_refutation_case] = STATE(2894),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(122),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_DOT] = ACTIONS(1167),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(123)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7073),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(123),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(328),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(124)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2983),
    [sym__sequence_expression] = STATE(2278),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2204),
    [sym_refutation_case] = STATE(2278),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(124),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_DOT] = ACTIONS(1169),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(125)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3075),
    [sym__sequence_expression] = STATE(3414),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_refutation_case] = STATE(3414),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(125),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_DOT] = ACTIONS(1177),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(126)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2983),
    [sym__sequence_expression] = STATE(2245),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2204),
    [sym_refutation_case] = STATE(2245),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(126),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_DOT] = ACTIONS(1169),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(127)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1764),
    [sym__sequence_expression] = STATE(2167),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_refutation_case] = STATE(2167),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(127),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_DOT] = ACTIONS(1171),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(128)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1764),
    [sym__sequence_expression] = STATE(2116),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_refutation_case] = STATE(2116),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(128),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_DOT] = ACTIONS(1171),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(129)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3075),
    [sym__sequence_expression] = STATE(3351),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_refutation_case] = STATE(3351),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(129),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_DOT] = ACTIONS(1177),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(130)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1648),
    [sym__sequence_expression] = STATE(2167),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2133),
    [sym_refutation_case] = STATE(2167),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(130),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_DOT] = ACTIONS(1171),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(131)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1648),
    [sym__sequence_expression] = STATE(2116),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2133),
    [sym_refutation_case] = STATE(2116),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(131),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_DOT] = ACTIONS(1171),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(132)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7331),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(132),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(232),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(133)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2002),
    [sym__sequence_expression] = STATE(2116),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_refutation_case] = STATE(2116),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(133),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_DOT] = ACTIONS(1171),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(134)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(5969),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(134),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(332),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(135)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3222),
    [sym__sequence_expression] = STATE(5984),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(135),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(337),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(136)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1848),
    [sym__sequence_expression] = STATE(2278),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2204),
    [sym_refutation_case] = STATE(2278),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(136),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_DOT] = ACTIONS(1169),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(137)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2876),
    [sym__sequence_expression] = STATE(2922),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_refutation_case] = STATE(2922),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(137),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_DOT] = ACTIONS(1167),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(138)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6218),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(138),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(498),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(139)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3222),
    [sym__sequence_expression] = STATE(6080),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(139),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(371),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(140)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6329),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(140),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(381),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(141)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3222),
    [sym__sequence_expression] = STATE(6199),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(141),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(385),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(142)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6077),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(142),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(229),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(143)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6197),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(143),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(265),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(144)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(5973),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(144),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(278),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(145)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6073),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(145),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(285),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(146)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6239),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(146),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(367),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(147)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6228),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(147),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(365),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(148)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6191),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(148),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(294),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(149)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6237),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(149),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(366),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(150)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6419),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(150),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(441),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(151)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6429),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(151),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(442),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(152)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3212),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1474),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(152),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(522),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(153)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6194),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(153),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(452),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(154)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6373),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(154),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(456),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(155)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6376),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(155),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(457),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(156)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6501),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(156),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(461),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(157)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6502),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(157),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(462),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(158)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6560),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(158),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(466),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(159)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6604),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(159),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(471),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(160)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6605),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(160),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(472),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(161)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6647),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(161),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(475),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(162)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6648),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(162),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(476),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(163)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6686),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(163),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(480),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(164)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6687),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(164),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(481),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(165)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6723),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(165),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(485),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(166)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6724),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(166),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(486),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(167)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6759),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(167),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(490),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(168)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3214),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(6946),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(168),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__attribute] = STATE(504),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_DASH_GT] = ACTIONS(41),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [anon_sym_PERCENT] = ACTIONS(741),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(169)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6419),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(169),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(1183),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(170)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4687),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(170),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__infix_operator] = STATE(7187),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_RPAREN] = ACTIONS(469),
    [anon_sym_DOT] = ACTIONS(471),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_module] = ACTIONS(1111),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(475),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(479),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(479),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(479),
    [sym_let_and_operator] = ACTIONS(479),
    [sym_match_operator] = ACTIONS(479),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(171)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4690),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(171),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__infix_operator] = STATE(7169),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(1101),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(1139),
    [anon_sym_DOT] = ACTIONS(1141),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_module] = ACTIONS(1143),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1185),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(1147),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(1147),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(1147),
    [sym_let_and_operator] = ACTIONS(1147),
    [sym_match_operator] = ACTIONS(1147),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(172)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4774),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(172),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__infix_operator] = STATE(7169),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_RPAREN] = ACTIONS(1139),
    [anon_sym_DOT] = ACTIONS(1141),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_module] = ACTIONS(1143),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1185),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(1147),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(1147),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(1147),
    [sym_let_and_operator] = ACTIONS(1147),
    [sym_match_operator] = ACTIONS(1147),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(173)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7272),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(173),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1187),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(174)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3018),
    [sym__sequence_expression] = STATE(6350),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(174),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_done] = ACTIONS(1189),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(175)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4768),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(175),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__infix_operator] = STATE(7187),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(1101),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(469),
    [anon_sym_DOT] = ACTIONS(471),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_module] = ACTIONS(1111),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1191),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(479),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(479),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(479),
    [sym_let_and_operator] = ACTIONS(479),
    [sym_match_operator] = ACTIONS(479),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(176)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4687),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(176),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__infix_operator] = STATE(7187),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_RPAREN] = ACTIONS(469),
    [anon_sym_DOT] = ACTIONS(471),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_module] = ACTIONS(1111),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1191),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(479),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(479),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(479),
    [sym_let_and_operator] = ACTIONS(479),
    [sym_match_operator] = ACTIONS(479),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(177)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6419),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(177),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(1193),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(178)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4519),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(178),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__infix_operator] = STATE(6518),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(1101),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(1127),
    [anon_sym_DOT] = ACTIONS(1129),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_module] = ACTIONS(1131),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(1135),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(1135),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(1135),
    [sym_let_and_operator] = ACTIONS(1135),
    [sym_match_operator] = ACTIONS(1135),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(179)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4768),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(179),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__infix_operator] = STATE(7187),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(1101),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(469),
    [anon_sym_DOT] = ACTIONS(471),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_module] = ACTIONS(1111),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1195),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(479),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(479),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(479),
    [sym_let_and_operator] = ACTIONS(479),
    [sym_match_operator] = ACTIONS(479),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(180)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4687),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(180),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__infix_operator] = STATE(7187),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_RPAREN] = ACTIONS(469),
    [anon_sym_DOT] = ACTIONS(471),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_module] = ACTIONS(1111),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1195),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(479),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(479),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(479),
    [sym_let_and_operator] = ACTIONS(479),
    [sym_match_operator] = ACTIONS(479),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(181)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6419),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(181),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(1197),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(182)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6419),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(182),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(1199),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(183)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6634),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(183),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1201),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(184)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3018),
    [sym__sequence_expression] = STATE(6893),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(184),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_done] = ACTIONS(1203),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(185)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4697),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(185),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__infix_operator] = STATE(7342),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_RPAREN] = ACTIONS(1087),
    [anon_sym_DOT] = ACTIONS(1089),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_module] = ACTIONS(1091),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(1095),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(1095),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(1095),
    [sym_let_and_operator] = ACTIONS(1095),
    [sym_match_operator] = ACTIONS(1095),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(186)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6419),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(186),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(1205),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(187)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4781),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(187),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__infix_operator] = STATE(6980),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_RPAREN] = ACTIONS(1041),
    [anon_sym_DOT] = ACTIONS(1045),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_module] = ACTIONS(1055),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1059),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(1077),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(1077),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(1077),
    [sym_let_and_operator] = ACTIONS(1077),
    [sym_match_operator] = ACTIONS(1077),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(188)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4507),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(188),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__infix_operator] = STATE(6517),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(1101),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(1155),
    [anon_sym_DOT] = ACTIONS(1157),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_module] = ACTIONS(1159),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(1161),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(1161),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(1161),
    [sym_let_and_operator] = ACTIONS(1161),
    [sym_match_operator] = ACTIONS(1161),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(189)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6554),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(189),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1207),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(190)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6929),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(190),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1209),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(191)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4768),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(191),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__infix_operator] = STATE(7187),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(1101),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(469),
    [anon_sym_DOT] = ACTIONS(471),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_module] = ACTIONS(1111),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(479),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(479),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(479),
    [sym_let_and_operator] = ACTIONS(479),
    [sym_match_operator] = ACTIONS(479),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(192)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6419),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(192),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(1211),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(193)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6419),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(193),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(1213),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(194)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6936),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(194),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1215),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(195)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4687),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(195),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__infix_operator] = STATE(7187),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_RPAREN] = ACTIONS(469),
    [anon_sym_DOT] = ACTIONS(471),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_module] = ACTIONS(1111),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(479),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(479),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(479),
    [sym_let_and_operator] = ACTIONS(479),
    [sym_match_operator] = ACTIONS(479),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(196)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6726),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(196),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1217),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(197)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6554),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(197),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1219),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(198)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6419),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(198),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(1221),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(199)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7319),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(199),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1223),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(200)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3018),
    [sym__sequence_expression] = STATE(6742),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(200),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_done] = ACTIONS(1225),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(201)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7067),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(201),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1227),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(202)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6419),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(202),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(1229),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(203)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6626),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(203),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1231),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(204)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3018),
    [sym__sequence_expression] = STATE(6363),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(204),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_done] = ACTIONS(1233),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(205)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4768),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(205),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__infix_operator] = STATE(7187),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(1101),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(469),
    [anon_sym_DOT] = ACTIONS(471),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_module] = ACTIONS(1111),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(475),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(479),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(479),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(479),
    [sym_let_and_operator] = ACTIONS(479),
    [sym_match_operator] = ACTIONS(479),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(206)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4690),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(206),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__infix_operator] = STATE(7169),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(1101),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(1139),
    [anon_sym_DOT] = ACTIONS(1141),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_module] = ACTIONS(1143),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(1147),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(1147),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(1147),
    [sym_let_and_operator] = ACTIONS(1147),
    [sym_match_operator] = ACTIONS(1147),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(207)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6419),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(207),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(1235),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(208)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4507),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(208),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__infix_operator] = STATE(6517),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(1101),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(1155),
    [anon_sym_DOT] = ACTIONS(1157),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_module] = ACTIONS(1159),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1163),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(1161),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(1161),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(1161),
    [sym_let_and_operator] = ACTIONS(1161),
    [sym_match_operator] = ACTIONS(1161),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(209)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4774),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(209),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__infix_operator] = STATE(7169),
    [sym_pow_operator] = STATE(6595),
    [sym_mult_operator] = STATE(6595),
    [sym_add_operator] = STATE(6595),
    [sym_and_operator] = STATE(6595),
    [sym_or_operator] = STATE(6595),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_RPAREN] = ACTIONS(1139),
    [anon_sym_DOT] = ACTIONS(1141),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_module] = ACTIONS(1143),
    [anon_sym_AMP] = ACTIONS(425),
    [anon_sym_POUND] = ACTIONS(1057),
    [anon_sym_COLON_COLON] = ACTIONS(1145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(1071),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym_prefix_operator] = ACTIONS(1147),
    [aux_sym_sign_operator_token1] = ACTIONS(441),
    [sym_hash_operator] = ACTIONS(1147),
    [anon_sym_lsl] = ACTIONS(435),
    [anon_sym_lsr] = ACTIONS(435),
    [anon_sym_asr] = ACTIONS(435),
    [aux_sym_pow_operator_token1] = ACTIONS(437),
    [anon_sym_mod] = ACTIONS(439),
    [anon_sym_land] = ACTIONS(439),
    [anon_sym_lor] = ACTIONS(439),
    [anon_sym_lxor] = ACTIONS(439),
    [aux_sym_mult_operator_token1] = ACTIONS(439),
    [aux_sym_add_operator_token1] = ACTIONS(441),
    [sym_concat_operator] = ACTIONS(443),
    [sym_rel_operator] = ACTIONS(445),
    [anon_sym_AMP_AMP] = ACTIONS(425),
    [anon_sym_or] = ACTIONS(447),
    [anon_sym_PIPE_PIPE] = ACTIONS(447),
    [sym_assign_operator] = ACTIONS(443),
    [sym_let_operator] = ACTIONS(1147),
    [sym_let_and_operator] = ACTIONS(1147),
    [sym_match_operator] = ACTIONS(1147),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(210)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6419),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(210),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(1237),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(211)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7299),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(211),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_end] = ACTIONS(1239),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(212)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6560),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(212),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(213)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2983),
    [sym__sequence_expression] = STATE(2266),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(213),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(214)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2983),
    [sym__sequence_expression] = STATE(2226),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(214),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(215)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2983),
    [sym__sequence_expression] = STATE(2228),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(215),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(216)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2983),
    [sym__sequence_expression] = STATE(2261),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(216),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(217)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2983),
    [sym__sequence_expression] = STATE(2265),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(217),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(218)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2983),
    [sym__sequence_expression] = STATE(2189),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(218),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(219)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6508),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(219),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(220)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3222),
    [sym__sequence_expression] = STATE(5993),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(220),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(221)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4348),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(221),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(222)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4438),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(222),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(223)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4327),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(223),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(224)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3165),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym__sequence_expression_content] = STATE(7268),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(224),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_RBRACK] = ACTIONS(1241),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(225)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3132),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym__sequence_expression_content] = STATE(7341),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(225),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1243),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(226)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4328),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(226),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(227)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6731),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(227),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(228)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4310),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(228),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(229)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6313),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(229),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(230)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6958),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(230),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(231)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(2022),
    [sym__sequence_expression] = STATE(3485),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(231),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(232)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7099),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(232),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(233)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2510),
    [sym__sequence_expression] = STATE(2171),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(233),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(234)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(2022),
    [sym__sequence_expression] = STATE(3508),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(234),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(235)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(2022),
    [sym__sequence_expression] = STATE(3509),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(235),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(236)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(2022),
    [sym__sequence_expression] = STATE(3510),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(236),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(237)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(2022),
    [sym__sequence_expression] = STATE(3489),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(237),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(238)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(2022),
    [sym__sequence_expression] = STATE(3491),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(238),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(239)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(2022),
    [sym__sequence_expression] = STATE(3492),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(239),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(240)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(2022),
    [sym__sequence_expression] = STATE(3514),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(240),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(241)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2912),
    [sym__sequence_expression] = STATE(2911),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(241),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(242)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2002),
    [sym__sequence_expression] = STATE(2179),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(242),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(243)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2912),
    [sym__sequence_expression] = STATE(2947),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(243),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(244)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2912),
    [sym__sequence_expression] = STATE(2953),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(244),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(245)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2510),
    [sym__sequence_expression] = STATE(2147),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(245),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(246)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2510),
    [sym__sequence_expression] = STATE(2128),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(246),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(247)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2912),
    [sym__sequence_expression] = STATE(2903),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(247),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(248)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2912),
    [sym__sequence_expression] = STATE(2906),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(248),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(249)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2912),
    [sym__sequence_expression] = STATE(2907),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(249),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(250)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2912),
    [sym__sequence_expression] = STATE(2926),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(250),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(251)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2876),
    [sym__sequence_expression] = STATE(2911),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(251),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(252)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3165),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym__sequence_expression_content] = STATE(6947),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(252),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_RBRACK] = ACTIONS(1245),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(253)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2876),
    [sym__sequence_expression] = STATE(2947),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(253),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(254)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2876),
    [sym__sequence_expression] = STATE(2953),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(254),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(255)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2876),
    [sym__sequence_expression] = STATE(2903),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(255),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(256)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2876),
    [sym__sequence_expression] = STATE(2906),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(256),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(257)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2876),
    [sym__sequence_expression] = STATE(2907),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(257),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(258)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2876),
    [sym__sequence_expression] = STATE(2926),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(258),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(259)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3018),
    [sym__sequence_expression] = STATE(6070),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(259),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(260)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3165),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym__sequence_expression_content] = STATE(7147),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(260),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_RBRACK] = ACTIONS(1247),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(261)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3132),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym__sequence_expression_content] = STATE(7161),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(261),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1249),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(262)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7073),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(262),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(263)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2510),
    [sym__sequence_expression] = STATE(2180),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(263),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(264)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2510),
    [sym__sequence_expression] = STATE(2163),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(264),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(265)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6267),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(265),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(266)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2510),
    [sym__sequence_expression] = STATE(2177),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(266),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(267)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4391),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(267),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(268)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2510),
    [sym__sequence_expression] = STATE(2160),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(268),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(269)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7270),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(269),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(270)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1648),
    [sym__sequence_expression] = STATE(2175),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(270),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(271)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7278),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(271),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(272)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7304),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(272),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(273)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1648),
    [sym__sequence_expression] = STATE(2165),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(273),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(274)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3165),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym__sequence_expression_content] = STATE(6530),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(274),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_RBRACK] = ACTIONS(1251),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(275)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3132),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym__sequence_expression_content] = STATE(6543),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(275),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1253),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(276)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1648),
    [sym__sequence_expression] = STATE(2179),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(276),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(277)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3132),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym__sequence_expression_content] = STATE(6882),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(277),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1255),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(278)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6187),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(278),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(279)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1648),
    [sym__sequence_expression] = STATE(2123),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(279),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(280)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1648),
    [sym__sequence_expression] = STATE(2124),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(280),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(281)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3165),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym__sequence_expression_content] = STATE(6783),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(281),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_RBRACK] = ACTIONS(1257),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(282)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3132),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym__sequence_expression_content] = STATE(6793),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(282),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1259),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(283)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1648),
    [sym__sequence_expression] = STATE(2149),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(283),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(284)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2002),
    [sym__sequence_expression] = STATE(2175),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(284),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(285)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6038),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(285),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(286)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1648),
    [sym__sequence_expression] = STATE(2132),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(286),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(287)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2862),
    [sym__sequence_expression] = STATE(2911),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(287),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(288)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6249),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(288),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(289)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2862),
    [sym__sequence_expression] = STATE(2947),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(289),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(290)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3165),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym__sequence_expression_content] = STATE(7013),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(290),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_RBRACK] = ACTIONS(1261),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(291)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3132),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym__sequence_expression_content] = STATE(7020),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(291),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1263),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(292)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2862),
    [sym__sequence_expression] = STATE(2953),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(292),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(293)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3222),
    [sym__sequence_expression] = STATE(6256),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(293),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(294)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6016),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(294),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(295)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2862),
    [sym__sequence_expression] = STATE(2903),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(295),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(296)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2862),
    [sym__sequence_expression] = STATE(2906),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(296),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(297)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2862),
    [sym__sequence_expression] = STATE(2907),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(297),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(298)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3165),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym__sequence_expression_content] = STATE(7182),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(298),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_RBRACK] = ACTIONS(1265),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(299)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3132),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym__sequence_expression_content] = STATE(7188),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(299),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1267),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(300)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1793),
    [sym__sequence_expression] = STATE(3508),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(300),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(301)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2862),
    [sym__sequence_expression] = STATE(2926),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(301),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(302)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1793),
    [sym__sequence_expression] = STATE(3509),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(302),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(303)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1848),
    [sym__sequence_expression] = STATE(2229),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(303),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(304)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1793),
    [sym__sequence_expression] = STATE(3510),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(304),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(305)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2002),
    [sym__sequence_expression] = STATE(2123),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(305),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(306)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3165),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym__sequence_expression_content] = STATE(6364),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(306),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_RBRACK] = ACTIONS(1269),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(307)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3132),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym__sequence_expression_content] = STATE(6342),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(307),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1271),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(308)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1848),
    [sym__sequence_expression] = STATE(2266),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(308),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(309)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1848),
    [sym__sequence_expression] = STATE(2226),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(309),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(310)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1848),
    [sym__sequence_expression] = STATE(2228),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(310),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(311)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1848),
    [sym__sequence_expression] = STATE(2261),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(311),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(312)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3165),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym__sequence_expression_content] = STATE(6739),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(312),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_RBRACK] = ACTIONS(1273),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(313)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3132),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym__sequence_expression_content] = STATE(6865),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(313),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1275),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(314)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1848),
    [sym__sequence_expression] = STATE(2265),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(314),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(315)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4344),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(315),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(316)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1848),
    [sym__sequence_expression] = STATE(2189),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(316),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(317)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4335),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(317),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(318)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2002),
    [sym__sequence_expression] = STATE(2124),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(318),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(319)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2002),
    [sym__sequence_expression] = STATE(2149),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(319),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(320)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3018),
    [sym__sequence_expression] = STATE(6216),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(320),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(321)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2002),
    [sym__sequence_expression] = STATE(2165),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(321),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(322)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1793),
    [sym__sequence_expression] = STATE(3485),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(322),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(323)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6491),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(323),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(324)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3165),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym__sequence_expression_content] = STATE(6405),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(324),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_RBRACK] = ACTIONS(1277),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(325)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3132),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym__sequence_expression_content] = STATE(7007),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(325),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1279),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(326)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6497),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(326),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(327)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6499),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(327),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(328)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7347),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(328),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(329)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1793),
    [sym__sequence_expression] = STATE(3489),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(329),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(330)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1793),
    [sym__sequence_expression] = STATE(3491),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(330),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(331)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1793),
    [sym__sequence_expression] = STATE(3492),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(331),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(332)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6183),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(332),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(333)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4364),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(333),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(334)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4433),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(334),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(335)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4331),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(335),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(336)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2899),
    [sym__sequence_expression] = STATE(4984),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(336),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(337)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3222),
    [sym__sequence_expression] = STATE(6085),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(337),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(338)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4406),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(338),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(339)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6393),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(339),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(340)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6396),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(340),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(341)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6399),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(341),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(342)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6695),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(342),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(343)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6696),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(343),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(344)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6702),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(344),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(345)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6934),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(345),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(346)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3192),
    [sym__sequence_expression] = STATE(6260),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(6874),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(346),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(347)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3222),
    [sym__sequence_expression] = STATE(6161),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(347),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(348)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6994),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(348),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(349)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3222),
    [sym__sequence_expression] = STATE(6184),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(349),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(350)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7031),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(350),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(351)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7045),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(351),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(352)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3075),
    [sym__sequence_expression] = STATE(2171),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(352),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(353)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4421),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(353),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(354)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3075),
    [sym__sequence_expression] = STATE(2147),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(354),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(355)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3075),
    [sym__sequence_expression] = STATE(2128),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(355),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(356)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3075),
    [sym__sequence_expression] = STATE(2180),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(356),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(357)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3075),
    [sym__sequence_expression] = STATE(2163),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(357),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(358)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3075),
    [sym__sequence_expression] = STATE(2177),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(358),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(359)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3075),
    [sym__sequence_expression] = STATE(2160),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(359),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(360)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6338),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(360),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(361)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6260),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(361),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(362)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2899),
    [sym__sequence_expression] = STATE(4891),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(362),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(363)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4404),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(363),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(364)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2899),
    [sym__sequence_expression] = STATE(4893),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(364),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(365)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6117),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(365),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(366)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6152),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(366),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(367)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6192),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(367),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(368)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2899),
    [sym__sequence_expression] = STATE(4895),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(368),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(369)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2899),
    [sym__sequence_expression] = STATE(4949),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(369),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(370)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2899),
    [sym__sequence_expression] = STATE(4951),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(370),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(371)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3222),
    [sym__sequence_expression] = STATE(6236),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(371),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(372)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6768),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(372),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(373)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6771),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(373),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(374)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6772),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(374),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(375)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6844),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(375),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(376)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6849),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(376),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(377)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6850),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(377),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(378)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3222),
    [sym__sequence_expression] = STATE(6046),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(378),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(379)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3222),
    [sym__sequence_expression] = STATE(6061),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(379),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(380)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2899),
    [sym__sequence_expression] = STATE(4952),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(380),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(381)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6305),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(381),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(382)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2002),
    [sym__sequence_expression] = STATE(2132),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(382),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(383)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2899),
    [sym__sequence_expression] = STATE(4995),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(383),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(384)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2500),
    [sym__sequence_expression] = STATE(2171),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(384),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(385)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3222),
    [sym__sequence_expression] = STATE(6227),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(385),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(386)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6369),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(386),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(387)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6371),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(387),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(388)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6372),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(388),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(389)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6424),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(389),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(390)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6425),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(390),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(391)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6426),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(391),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(392)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3222),
    [sym__sequence_expression] = STATE(6299),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(392),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(393)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3222),
    [sym__sequence_expression] = STATE(6303),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(393),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(394)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3222),
    [sym__sequence_expression] = STATE(6330),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(394),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(395)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2500),
    [sym__sequence_expression] = STATE(2147),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(395),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(396)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6655),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(396),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(397)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6656),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(397),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(398)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6657),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(398),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(399)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6685),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(399),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(400)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6689),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(400),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(401)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6690),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(401),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(402)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2500),
    [sym__sequence_expression] = STATE(2128),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(402),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(403)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6948),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(403),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(404)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6953),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(404),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(405)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6955),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(405),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(406)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6972),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(406),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(407)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6973),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(407),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(408)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6974),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(408),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(409)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1793),
    [sym__sequence_expression] = STATE(3514),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(409),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(410)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7089),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(410),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(411)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7090),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(411),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(412)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7092),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(412),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(413)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7112),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(413),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(414)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7113),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(414),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(415)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7114),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(415),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(416)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7298),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(416),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(417)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7300),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(417),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(418)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7302),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(418),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(419)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7311),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(419),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(420)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7312),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(420),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(421)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7313),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(421),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(422)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4365),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(422),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(423)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7222),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(423),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(424)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7258),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(424),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(425)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(7318),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(425),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(426)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6406),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(426),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(427)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6421),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(427),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(428)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6440),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(428),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(429)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2500),
    [sym__sequence_expression] = STATE(2180),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(429),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(430)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6353),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(430),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(431)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6358),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(431),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(432)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6359),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(432),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(433)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6408),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(433),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(434)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6411),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(434),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(435)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6413),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(435),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(436)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2500),
    [sym__sequence_expression] = STATE(2163),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(436),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(437)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2500),
    [sym__sequence_expression] = STATE(2177),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(437),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(438)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4371),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(438),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(439)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2500),
    [sym__sequence_expression] = STATE(2160),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(439),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(440)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6419),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(440),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(441)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6521),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(441),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(442)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6524),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(442),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(443)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3161),
    [sym__sequence_expression] = STATE(2171),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(443),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(444)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1764),
    [sym__sequence_expression] = STATE(2175),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(444),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(445)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3161),
    [sym__sequence_expression] = STATE(2147),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(445),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(446)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3161),
    [sym__sequence_expression] = STATE(2128),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(446),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(447)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3161),
    [sym__sequence_expression] = STATE(2180),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(447),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(448)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3161),
    [sym__sequence_expression] = STATE(2163),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(448),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(449)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3161),
    [sym__sequence_expression] = STATE(2177),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(449),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(450)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3161),
    [sym__sequence_expression] = STATE(2160),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2425),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(450),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(451)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1764),
    [sym__sequence_expression] = STATE(2165),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(451),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(452)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6181),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(452),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(453)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1764),
    [sym__sequence_expression] = STATE(2179),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(453),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(454)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1764),
    [sym__sequence_expression] = STATE(2123),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(454),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(455)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6373),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(455),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(456)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6398),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(456),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(457)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6400),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(457),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(458)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1764),
    [sym__sequence_expression] = STATE(2124),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(458),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(459)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1764),
    [sym__sequence_expression] = STATE(2149),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(459),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(460)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6501),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(460),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(461)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6514),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(461),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(462)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6515),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(462),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(463)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1764),
    [sym__sequence_expression] = STATE(2132),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(463),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(464)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2800),
    [sym__sequence_expression] = STATE(2911),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(464),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(465)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4385),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(465),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(466)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6569),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(466),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(467)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6570),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(467),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(468)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2800),
    [sym__sequence_expression] = STATE(2947),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(468),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(469)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2800),
    [sym__sequence_expression] = STATE(2953),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(469),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(470)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6604),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(470),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(471)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6612),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(471),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(472)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6613),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(472),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(473)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4386),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(473),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(474)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6647),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(474),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(475)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6653),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(475),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(476)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6654),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(476),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(477)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2800),
    [sym__sequence_expression] = STATE(2903),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(477),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(478)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2800),
    [sym__sequence_expression] = STATE(2906),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(478),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(479)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6686),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(479),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(480)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6692),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(480),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(481)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6693),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(481),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(482)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2800),
    [sym__sequence_expression] = STATE(2907),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(482),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(483)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2630),
    [sym__sequence_expression] = STATE(4390),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(483),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(484)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6723),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(484),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(485)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6729),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(485),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(486)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6730),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(486),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(487)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2800),
    [sym__sequence_expression] = STATE(2926),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2896),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(487),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(488)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2983),
    [sym__sequence_expression] = STATE(2229),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2223),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(488),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(489)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6759),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(489),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(490)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6765),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(490),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(491)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6766),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(491),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(492)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3018),
    [sym__sequence_expression] = STATE(6337),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(492),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(493)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3018),
    [sym__sequence_expression] = STATE(6167),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(493),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(494)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3018),
    [sym__sequence_expression] = STATE(6212),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(494),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(495)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3018),
    [sym__sequence_expression] = STATE(6213),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(495),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(496)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3018),
    [sym__sequence_expression] = STATE(6222),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(496),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(497)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3018),
    [sym__sequence_expression] = STATE(6223),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(1861),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(497),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(498)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2640),
    [sym__sequence_expression] = STATE(6053),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1480),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(498),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(499)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2649),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1486),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(499),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(500)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2882),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(500),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(501)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2517),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2219),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(501),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(502)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2120),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1486),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(502),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(503)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3160),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(503),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(504)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3217),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(7125),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(504),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(505)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(2028),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1486),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(505),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(506)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1828),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1792),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(506),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(507)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1707),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1486),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(507),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(508)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2794),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2892),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(508),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(509)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3071),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(509),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(510)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2849),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2892),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(510),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(511)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3221),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(511),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_RBRACK] = ACTIONS(1281),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(512)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2891),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(512),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(513)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2508),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(2468),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(513),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(514)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2954),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(2892),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(514),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(515)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3211),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(515),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1281),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(516)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2620),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1486),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(516),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(517)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1669),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1792),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(517),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(518)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3221),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(518),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_RBRACK] = ACTIONS(1283),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(519)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1895),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2219),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(519),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(520)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3211),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(520),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1283),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(521)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3001),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym__sequence_expression_anonymous] = STATE(1792),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(521),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(522)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3215),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym__sequence_expression_anonymous] = STATE(1486),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(522),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(523)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(3074),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym__sequence_expression_anonymous] = STATE(2219),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(523),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(524)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2088),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym__sequence_expression_anonymous] = STATE(1486),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(524),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(525)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2919),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(525),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(526)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2844),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(526),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(527)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3183),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(527),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(528)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1888),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(528),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(529)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1846),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(2311),
    [sym__tuple_expression] = STATE(2209),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(529),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(1285),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(530)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1847),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(530),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(531)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1849),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(531),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(532)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1850),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(532),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(533)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1851),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(533),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(534)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1778),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(534),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(535)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1779),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(535),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(536)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1780),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(536),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(537)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1852),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(537),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(538)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1919),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(538),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(539)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1854),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(539),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(540)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1858),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(540),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(541)] = {
    [sym_value_definition] = STATE(6528),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(1864),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(6749),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1033),
    [sym_array_get_expression] = STATE(1033),
    [sym_string_get_expression] = STATE(1033),
    [sym_bigarray_get_expression] = STATE(1033),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(541),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(528),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(6751),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(35),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(53),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(57),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(63),
    [anon_sym_function] = ACTIONS(65),
    [anon_sym_try] = ACTIONS(67),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(99),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(542)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3206),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(542),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(543)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2898),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(543),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(544)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2932),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(3305),
    [sym__tuple_expression] = STATE(2913),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(544),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(1287),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(545)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2904),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(545),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(546)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2921),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(546),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(547)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2945),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(547),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(548)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2910),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(548),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(549)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2900),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(549),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(550)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2901),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(550),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(551)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2902),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(551),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(552)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2944),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(552),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(553)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2915),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(553),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(554)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2004),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(554),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(555)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2887),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(555),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(556)] = {
    [sym_value_definition] = STATE(6368),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2897),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7087),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1620),
    [sym_array_get_expression] = STATE(1620),
    [sym_string_get_expression] = STATE(1620),
    [sym_bigarray_get_expression] = STATE(1620),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(556),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(543),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7088),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(693),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(695),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(697),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(557)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2074),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(557),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(558)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3221),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(558),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(559)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2847),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(559),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(560)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2884),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(3259),
    [sym__tuple_expression] = STATE(2913),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(560),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(1287),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(561)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2865),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(561),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(562)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2866),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(562),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(563)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2867),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(563),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(564)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2868),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(564),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(565)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2855),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(565),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(566)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2856),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(566),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(567)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2857),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(567),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(568)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2869),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(568),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(569)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2870),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(569),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(570)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2871),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(570),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(571)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2872),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(571),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(572)] = {
    [sym_value_definition] = STATE(7173),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(2873),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7000),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1559),
    [sym_array_get_expression] = STATE(1559),
    [sym_string_get_expression] = STATE(1559),
    [sym_bigarray_get_expression] = STATE(1559),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(572),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(559),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7001),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(669),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(671),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(673),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(675),
    [anon_sym_function] = ACTIONS(677),
    [anon_sym_try] = ACTIONS(679),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(685),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(573)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(1936),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(2316),
    [sym__tuple_expression] = STATE(2127),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(573),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(1289),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(574)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2007),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(574),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(575)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2010),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(575),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(576)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2014),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(576),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(577)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2015),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(577),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(578)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2016),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(578),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(579)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2017),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(579),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(580)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2018),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(580),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(581)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2019),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(581),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(582)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2045),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(582),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(583)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(2047),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(583),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(584)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2515),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(584),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(585)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2566),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(2843),
    [sym__tuple_expression] = STATE(2125),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(585),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(1291),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(586)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2560),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(586),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(587)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2564),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(587),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(588)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2519),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(588),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(589)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2523),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(589),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(590)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2559),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(590),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(591)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2511),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(591),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(592)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2528),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(592),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(593)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2567),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(593),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(594)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2558),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(594),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(595)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2514),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(595),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(596)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2518),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(596),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(597)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2531),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(597),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(598)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3041),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(598),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(599)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3211),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(599),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(600)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3033),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(600),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(601)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3034),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(3412),
    [sym__tuple_expression] = STATE(2125),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(601),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(1291),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(602)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3036),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(602),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(603)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3037),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(603),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(604)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3038),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(604),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(605)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3039),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(605),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(606)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3040),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(606),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(607)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3114),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(607),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(608)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3042),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(608),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(609)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3043),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(609),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(610)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3044),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(610),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(611)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3045),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(611),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(612)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3046),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(612),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(613)] = {
    [sym_value_definition] = STATE(6909),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3048),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7196),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1638),
    [sym_array_get_expression] = STATE(1638),
    [sym_string_get_expression] = STATE(1638),
    [sym_bigarray_get_expression] = STATE(1638),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(613),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(600),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7197),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(699),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(703),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(705),
    [anon_sym_function] = ACTIONS(707),
    [anon_sym_try] = ACTIONS(709),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(711),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(614)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1771),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(614),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(615)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1721),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(2183),
    [sym__tuple_expression] = STATE(2127),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(615),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(1289),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(616)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1722),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(616),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(617)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1772),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(617),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(618)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1726),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(618),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(619)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1728),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(619),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(620)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1674),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(620),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(621)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1676),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(621),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(622)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1731),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(622),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(623)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1732),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(623),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(624)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1734),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(624),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(625)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1735),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(625),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(626)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1739),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(626),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(627)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2796),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(627),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(628)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2820),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(3198),
    [sym__tuple_expression] = STATE(2913),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(628),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(1287),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(629)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2821),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(629),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(630)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2822),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(630),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(631)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2823),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(631),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(632)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2824),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(632),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(633)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2797),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(633),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(634)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2798),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(634),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(635)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2799),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(635),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(636)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2825),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(636),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(637)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2826),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(637),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(638)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2827),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(638),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(639)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2828),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(639),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(640)] = {
    [sym_value_definition] = STATE(6906),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2829),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(7324),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1520),
    [sym_array_get_expression] = STATE(1520),
    [sym_string_get_expression] = STATE(1520),
    [sym_bigarray_get_expression] = STATE(1520),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(640),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(627),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(7325),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(611),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(623),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(627),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(641)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(3101),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(641),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(642)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(3061),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(3348),
    [sym__tuple_expression] = STATE(2209),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(642),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(1285),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(643)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(3063),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(643),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(644)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(3064),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(644),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(645)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(3065),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(645),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(646)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(3066),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(646),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(647)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(3102),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(647),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(648)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(3103),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(648),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(649)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(3104),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(649),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(650)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(3067),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(650),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(651)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(3068),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(651),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(652)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(3069),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(652),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(653)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(3070),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(653),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(654)] = {
    [sym_value_definition] = STATE(6777),
    [sym__delimited_expression] = STATE(910),
    [sym__simple_expression] = STATE(80),
    [sym__expression] = STATE(3072),
    [sym_typed_expression] = STATE(910),
    [sym_labeled_tuple_element] = STATE(7329),
    [sym__tuple_expression] = STATE(2197),
    [sym_cons_expression] = STATE(2204),
    [sym_list_expression] = STATE(911),
    [sym_array_expression] = STATE(911),
    [sym_record_expression] = STATE(911),
    [sym_application_expression] = STATE(2204),
    [sym_prefix_expression] = STATE(910),
    [sym_sign_expression] = STATE(2204),
    [sym_hash_expression] = STATE(910),
    [sym_infix_expression] = STATE(2204),
    [sym_field_get_expression] = STATE(1640),
    [sym_array_get_expression] = STATE(1640),
    [sym_string_get_expression] = STATE(1640),
    [sym_bigarray_get_expression] = STATE(1640),
    [sym_set_expression] = STATE(2204),
    [sym_if_expression] = STATE(2204),
    [sym_while_expression] = STATE(2204),
    [sym_for_expression] = STATE(2204),
    [sym_match_expression] = STATE(2204),
    [sym_function_expression] = STATE(2204),
    [sym_fun_expression] = STATE(2204),
    [sym_try_expression] = STATE(2204),
    [sym_let_expression] = STATE(2204),
    [sym_assert_expression] = STATE(2204),
    [sym_lazy_expression] = STATE(2204),
    [sym_local_open_expression] = STATE(910),
    [sym_package_expression] = STATE(911),
    [sym_new_expression] = STATE(910),
    [sym_object_copy_expression] = STATE(911),
    [sym_method_invocation] = STATE(910),
    [sym_object_expression] = STATE(910),
    [sym_parenthesized_expression] = STATE(911),
    [sym_attribute] = STATE(654),
    [sym_extension] = STATE(910),
    [sym_quoted_extension] = STATE(910),
    [sym__constant] = STATE(910),
    [sym_number] = STATE(894),
    [sym_character] = STATE(894),
    [sym_string] = STATE(894),
    [sym_quoted_string] = STATE(894),
    [sym_boolean] = STATE(894),
    [sym_unit] = STATE(911),
    [sym_sign_operator] = STATE(641),
    [sym__value_name] = STATE(868),
    [sym_parenthesized_operator] = STATE(883),
    [sym_value_path] = STATE(910),
    [sym_module_path] = STATE(6680),
    [sym_constructor_path] = STATE(910),
    [sym__instance_variable_name] = STATE(7330),
    [sym__constructor_name] = STATE(886),
    [sym_tag] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(33),
    [anon_sym_let] = ACTIONS(713),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(39),
    [anon_sym_LBRACK] = ACTIONS(45),
    [anon_sym_true] = ACTIONS(47),
    [anon_sym_false] = ACTIONS(47),
    [anon_sym_LBRACE] = ACTIONS(49),
    [anon_sym_object] = ACTIONS(51),
    [anon_sym_fun] = ACTIONS(715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(55),
    [anon_sym_if] = ACTIONS(717),
    [anon_sym_while] = ACTIONS(59),
    [anon_sym_for] = ACTIONS(61),
    [anon_sym_match] = ACTIONS(719),
    [anon_sym_function] = ACTIONS(721),
    [anon_sym_try] = ACTIONS(723),
    [anon_sym_assert] = ACTIONS(69),
    [anon_sym_lazy] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_LBRACE_LT] = ACTIONS(75),
    [anon_sym_begin] = ACTIONS(77),
    [sym_ocamlyacc_value] = ACTIONS(79),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(681),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(683),
    [aux_sym_number_token1] = ACTIONS(87),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(91),
    [anon_sym_DQUOTE] = ACTIONS(93),
    [sym_prefix_operator] = ACTIONS(95),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(725),
    [sym__uppercase_identifier] = ACTIONS(101),
    [aux_sym_tag_token1] = ACTIONS(103),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(655)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1657),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(655),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(656)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2527),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(656),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(657)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2529),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(2836),
    [sym__tuple_expression] = STATE(2125),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(657),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(1291),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(658)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2532),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(658),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(659)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2533),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(659),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(660)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2534),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(660),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(661)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2535),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(661),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(662)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2538),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(662),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(663)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2539),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(663),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(664)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2540),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(664),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(665)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2543),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(665),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(666)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2544),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(666),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(667)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2545),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(667),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(668)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2549),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(668),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(669)] = {
    [sym_value_definition] = STATE(6395),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2554),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(6474),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1387),
    [sym_array_get_expression] = STATE(1387),
    [sym_string_get_expression] = STATE(1387),
    [sym_bigarray_get_expression] = STATE(1387),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(669),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(656),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(6475),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(351),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(363),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(367),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(670)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1660),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(2137),
    [sym__tuple_expression] = STATE(2127),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(670),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(1289),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(671)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1661),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(671),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(672)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1664),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(672),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(673)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1666),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(673),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(674)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1670),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(674),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(675)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1665),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(675),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(676)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1651),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(676),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(677)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1653),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(677),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(678)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1668),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(678),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(679)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1671),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(679),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(680)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1647),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(680),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(681)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3175),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(681),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(682)] = {
    [sym_value_definition] = STATE(6931),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(2980),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7335),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1390),
    [sym_array_get_expression] = STATE(1390),
    [sym_string_get_expression] = STATE(1390),
    [sym_bigarray_get_expression] = STATE(1390),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(682),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(584),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7336),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(409),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(411),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(413),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(373),
    [anon_sym_function] = ACTIONS(375),
    [anon_sym_try] = ACTIONS(377),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(403),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(683)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1652),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(683),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(684)] = {
    [sym_value_definition] = STATE(7239),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1667),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(6576),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(938),
    [sym_array_get_expression] = STATE(938),
    [sym_string_get_expression] = STATE(938),
    [sym_bigarray_get_expression] = STATE(938),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(684),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(655),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(6591),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(119),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(123),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(129),
    [anon_sym_function] = ACTIONS(131),
    [anon_sym_try] = ACTIONS(133),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(159),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(685)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2877),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(685),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(686)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3141),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(686),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(687)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3142),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(3442),
    [sym__tuple_expression] = STATE(2125),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(687),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(1291),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(688)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3171),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(688),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(689)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3144),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(689),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(690)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3145),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(690),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(691)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3146),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(691),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(692)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3147),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(692),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(693)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3148),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(693),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(694)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3149),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(694),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(695)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3150),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(695),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(696)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3115),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(696),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(697)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3152),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(697),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(698)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3153),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(698),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(699)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2850),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(3255),
    [sym__tuple_expression] = STATE(2913),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(699),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(1287),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(700)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2851),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(700),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(701)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2852),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(701),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(702)] = {
    [sym_value_definition] = STATE(7242),
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(77),
    [sym__expression] = STATE(3154),
    [sym_typed_expression] = STATE(794),
    [sym_labeled_tuple_element] = STATE(7028),
    [sym__tuple_expression] = STATE(2172),
    [sym_cons_expression] = STATE(2151),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_application_expression] = STATE(2151),
    [sym_prefix_expression] = STATE(794),
    [sym_sign_expression] = STATE(2151),
    [sym_hash_expression] = STATE(794),
    [sym_infix_expression] = STATE(2151),
    [sym_field_get_expression] = STATE(1658),
    [sym_array_get_expression] = STATE(1658),
    [sym_string_get_expression] = STATE(1658),
    [sym_bigarray_get_expression] = STATE(1658),
    [sym_set_expression] = STATE(2151),
    [sym_if_expression] = STATE(2151),
    [sym_while_expression] = STATE(2151),
    [sym_for_expression] = STATE(2151),
    [sym_match_expression] = STATE(2151),
    [sym_function_expression] = STATE(2151),
    [sym_fun_expression] = STATE(2151),
    [sym_try_expression] = STATE(2151),
    [sym_let_expression] = STATE(2151),
    [sym_assert_expression] = STATE(2151),
    [sym_lazy_expression] = STATE(2151),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(702),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym_sign_operator] = STATE(686),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__instance_variable_name] = STATE(7029),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(349),
    [anon_sym_let] = ACTIONS(727),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_fun] = ACTIONS(729),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_if] = ACTIONS(731),
    [anon_sym_while] = ACTIONS(369),
    [anon_sym_for] = ACTIONS(371),
    [anon_sym_match] = ACTIONS(733),
    [anon_sym_function] = ACTIONS(735),
    [anon_sym_try] = ACTIONS(737),
    [anon_sym_assert] = ACTIONS(379),
    [anon_sym_lazy] = ACTIONS(381),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(415),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(417),
    [aux_sym_number_token1] = ACTIONS(395),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(739),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(703)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2853),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(703),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(704)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2854),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(704),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(705)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2878),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(705),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(706)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2879),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(706),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(707)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2880),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(707),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(708)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2858),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(708),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(709)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2859),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(709),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(710)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2860),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(710),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(711)] = {
    [sym_value_definition] = STATE(7066),
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(87),
    [sym__expression] = STATE(1923),
    [sym_typed_expression] = STATE(1052),
    [sym_labeled_tuple_element] = STATE(7116),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(1052),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(1052),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1056),
    [sym_array_get_expression] = STATE(1056),
    [sym_string_get_expression] = STATE(1056),
    [sym_bigarray_get_expression] = STATE(1056),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(711),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym_sign_operator] = STATE(557),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__instance_variable_name] = STATE(6915),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(179),
    [anon_sym_let] = ACTIONS(181),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_fun] = ACTIONS(193),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_if] = ACTIONS(197),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(199),
    [anon_sym_function] = ACTIONS(201),
    [anon_sym_try] = ACTIONS(203),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1173),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1175),
    [aux_sym_number_token1] = ACTIONS(217),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(225),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(712)] = {
    [sym_value_definition] = STATE(7145),
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(715),
    [sym__expression] = STATE(2861),
    [sym_typed_expression] = STATE(1543),
    [sym_labeled_tuple_element] = STATE(6494),
    [sym__tuple_expression] = STATE(2914),
    [sym_cons_expression] = STATE(2916),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_application_expression] = STATE(2916),
    [sym_prefix_expression] = STATE(1543),
    [sym_sign_expression] = STATE(2916),
    [sym_hash_expression] = STATE(1543),
    [sym_infix_expression] = STATE(2916),
    [sym_field_get_expression] = STATE(1527),
    [sym_array_get_expression] = STATE(1527),
    [sym_string_get_expression] = STATE(1527),
    [sym_bigarray_get_expression] = STATE(1527),
    [sym_set_expression] = STATE(2916),
    [sym_if_expression] = STATE(2916),
    [sym_while_expression] = STATE(2916),
    [sym_for_expression] = STATE(2916),
    [sym_match_expression] = STATE(2916),
    [sym_function_expression] = STATE(2916),
    [sym_fun_expression] = STATE(2916),
    [sym_try_expression] = STATE(2916),
    [sym_let_expression] = STATE(2916),
    [sym_assert_expression] = STATE(2916),
    [sym_lazy_expression] = STATE(2916),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(712),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym_sign_operator] = STATE(685),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__instance_variable_name] = STATE(6495),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [sym__lowercase_identifier] = ACTIONS(609),
    [anon_sym_let] = ACTIONS(687),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_fun] = ACTIONS(689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_if] = ACTIONS(691),
    [anon_sym_while] = ACTIONS(629),
    [anon_sym_for] = ACTIONS(631),
    [anon_sym_match] = ACTIONS(633),
    [anon_sym_function] = ACTIONS(635),
    [anon_sym_try] = ACTIONS(637),
    [anon_sym_assert] = ACTIONS(639),
    [anon_sym_lazy] = ACTIONS(641),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(663),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(713)] = {
    [sym_value_definition] = STATE(7011),
    [sym__delimited_expression] = STATE(987),
    [sym__simple_expression] = STATE(81),
    [sym__expression] = STATE(1675),
    [sym_typed_expression] = STATE(987),
    [sym_labeled_tuple_element] = STATE(7295),
    [sym__tuple_expression] = STATE(2141),
    [sym_cons_expression] = STATE(2133),
    [sym_list_expression] = STATE(988),
    [sym_array_expression] = STATE(988),
    [sym_record_expression] = STATE(988),
    [sym_application_expression] = STATE(2133),
    [sym_prefix_expression] = STATE(987),
    [sym_sign_expression] = STATE(2133),
    [sym_hash_expression] = STATE(987),
    [sym_infix_expression] = STATE(2133),
    [sym_field_get_expression] = STATE(1005),
    [sym_array_get_expression] = STATE(1005),
    [sym_string_get_expression] = STATE(1005),
    [sym_bigarray_get_expression] = STATE(1005),
    [sym_set_expression] = STATE(2133),
    [sym_if_expression] = STATE(2133),
    [sym_while_expression] = STATE(2133),
    [sym_for_expression] = STATE(2133),
    [sym_match_expression] = STATE(2133),
    [sym_function_expression] = STATE(2133),
    [sym_fun_expression] = STATE(2133),
    [sym_try_expression] = STATE(2133),
    [sym_let_expression] = STATE(2133),
    [sym_assert_expression] = STATE(2133),
    [sym_lazy_expression] = STATE(2133),
    [sym_local_open_expression] = STATE(987),
    [sym_package_expression] = STATE(988),
    [sym_new_expression] = STATE(987),
    [sym_object_copy_expression] = STATE(988),
    [sym_method_invocation] = STATE(987),
    [sym_object_expression] = STATE(987),
    [sym_parenthesized_expression] = STATE(988),
    [sym_attribute] = STATE(713),
    [sym_extension] = STATE(987),
    [sym_quoted_extension] = STATE(987),
    [sym__constant] = STATE(987),
    [sym_number] = STATE(997),
    [sym_character] = STATE(997),
    [sym_string] = STATE(997),
    [sym_quoted_string] = STATE(997),
    [sym_boolean] = STATE(997),
    [sym_unit] = STATE(988),
    [sym_sign_operator] = STATE(614),
    [sym__value_name] = STATE(963),
    [sym_parenthesized_operator] = STATE(978),
    [sym_value_path] = STATE(987),
    [sym_module_path] = STATE(6598),
    [sym_constructor_path] = STATE(987),
    [sym__instance_variable_name] = STATE(7297),
    [sym__constructor_name] = STATE(971),
    [sym_tag] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(105),
    [anon_sym_let] = ACTIONS(165),
    [anon_sym_TILDE] = ACTIONS(37),
    [anon_sym_LPAREN] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_true] = ACTIONS(113),
    [anon_sym_false] = ACTIONS(113),
    [anon_sym_LBRACE] = ACTIONS(115),
    [anon_sym_object] = ACTIONS(117),
    [anon_sym_fun] = ACTIONS(167),
    [anon_sym_LBRACK_PIPE] = ACTIONS(121),
    [anon_sym_if] = ACTIONS(169),
    [anon_sym_while] = ACTIONS(125),
    [anon_sym_for] = ACTIONS(127),
    [anon_sym_match] = ACTIONS(171),
    [anon_sym_function] = ACTIONS(173),
    [anon_sym_try] = ACTIONS(175),
    [anon_sym_assert] = ACTIONS(135),
    [anon_sym_lazy] = ACTIONS(137),
    [anon_sym_new] = ACTIONS(139),
    [anon_sym_LBRACE_LT] = ACTIONS(141),
    [anon_sym_begin] = ACTIONS(143),
    [sym_ocamlyacc_value] = ACTIONS(145),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1179),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1181),
    [aux_sym_number_token1] = ACTIONS(151),
    [aux_sym_signed_number_token1] = ACTIONS(89),
    [anon_sym_SQUOTE] = ACTIONS(153),
    [anon_sym_DQUOTE] = ACTIONS(155),
    [sym_prefix_operator] = ACTIONS(157),
    [aux_sym_sign_operator_token1] = ACTIONS(263),
    [sym_let_operator] = ACTIONS(97),
    [sym_match_operator] = ACTIONS(177),
    [sym__uppercase_identifier] = ACTIONS(161),
    [aux_sym_tag_token1] = ACTIONS(163),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(714)] = {
    [sym__parenthesized_abstract_type] = STATE(1310),
    [sym__delimited_binding_pattern] = STATE(2051),
    [sym__simple_binding_pattern] = STATE(2085),
    [sym__effect_binding_pattern] = STATE(2060),
    [sym__binding_pattern] = STATE(2063),
    [sym_alias_binding_pattern] = STATE(2060),
    [sym_typed_binding_pattern] = STATE(2051),
    [sym__or_binding_pattern_anonymous] = STATE(2061),
    [sym_constructor_binding_pattern] = STATE(2085),
    [sym_tag_binding_pattern] = STATE(2085),
    [sym_polymorphic_variant_pattern] = STATE(2051),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7213),
    [sym__tuple_binding_pattern] = STATE(2062),
    [sym_record_binding_pattern] = STATE(2087),
    [sym_list_binding_pattern] = STATE(2087),
    [sym_cons_binding_pattern] = STATE(2060),
    [sym_array_binding_pattern] = STATE(2087),
    [sym_range_pattern] = STATE(2060),
    [sym_lazy_binding_pattern] = STATE(2085),
    [sym_local_open_binding_pattern] = STATE(2051),
    [sym_package_pattern] = STATE(2051),
    [sym_parenthesized_binding_pattern] = STATE(2087),
    [sym_exception_binding_pattern] = STATE(2060),
    [sym_effect_binding_pattern] = STATE(2060),
    [sym_attribute] = STATE(714),
    [sym_extension] = STATE(2051),
    [sym_quoted_extension] = STATE(2051),
    [sym__constant] = STATE(1823),
    [sym__signed_constant] = STATE(1825),
    [sym_number] = STATE(1803),
    [sym_signed_number] = STATE(1823),
    [sym_character] = STATE(1803),
    [sym_string] = STATE(1803),
    [sym_quoted_string] = STATE(1803),
    [sym_boolean] = STATE(1803),
    [sym_unit] = STATE(2087),
    [sym__value_name] = STATE(2051),
    [sym_parenthesized_operator] = STATE(2035),
    [sym_module_path] = STATE(6855),
    [sym_constructor_path] = STATE(714),
    [sym__constructor_name] = STATE(1698),
    [sym_tag] = STATE(716),
    [sym__lowercase_identifier] = ACTIONS(1293),
    [anon_sym_SEMI_SEMI] = ACTIONS(1295),
    [anon_sym_let] = ACTIONS(1297),
    [anon_sym_and] = ACTIONS(1297),
    [anon_sym_EQ] = ACTIONS(1295),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_QMARK] = ACTIONS(1295),
    [anon_sym_LPAREN] = ACTIONS(1301),
    [anon_sym_external] = ACTIONS(1297),
    [anon_sym_type] = ACTIONS(1297),
    [anon_sym_COMMA] = ACTIONS(1295),
    [anon_sym_PIPE] = ACTIONS(1295),
    [anon_sym_COLON2] = ACTIONS(1297),
    [anon_sym_LBRACK] = ACTIONS(1303),
    [anon_sym_RBRACK] = ACTIONS(1295),
    [anon_sym_true] = ACTIONS(1305),
    [anon_sym_false] = ACTIONS(1305),
    [anon_sym_LBRACE] = ACTIONS(1307),
    [anon_sym_exception] = ACTIONS(1309),
    [anon_sym_module] = ACTIONS(1297),
    [anon_sym_open] = ACTIONS(1297),
    [anon_sym_include] = ACTIONS(1297),
    [anon_sym_class] = ACTIONS(1297),
    [anon_sym_val] = ACTIONS(1297),
    [anon_sym_end] = ACTIONS(1297),
    [anon_sym_COLON_GT] = ACTIONS(1295),
    [anon_sym_in] = ACTIONS(1297),
    [anon_sym_as] = ACTIONS(1297),
    [anon_sym_POUND] = ACTIONS(1311),
    [anon_sym_COLON_COLON] = ACTIONS(1295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1313),
    [anon_sym_lazy] = ACTIONS(1315),
    [anon_sym_begin] = ACTIONS(1317),
    [anon_sym_effect] = ACTIONS(1319),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1297),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1295),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1321),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1323),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1295),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1295),
    [aux_sym_number_token1] = ACTIONS(1325),
    [aux_sym_signed_number_token1] = ACTIONS(1327),
    [anon_sym_SQUOTE] = ACTIONS(1329),
    [anon_sym_DQUOTE] = ACTIONS(1331),
    [sym_let_operator] = ACTIONS(1295),
    [sym_let_and_operator] = ACTIONS(1295),
    [sym__uppercase_identifier] = ACTIONS(1333),
    [aux_sym_directive_token1] = ACTIONS(1297),
    [aux_sym_tag_token1] = ACTIONS(1335),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(715)] = {
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(1530),
    [sym_typed_expression] = STATE(1543),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_labeled_argument] = STATE(1644),
    [sym_prefix_expression] = STATE(1543),
    [sym_hash_expression] = STATE(1543),
    [sym_field_get_expression] = STATE(1543),
    [sym_array_get_expression] = STATE(1543),
    [sym_string_get_expression] = STATE(1543),
    [sym_bigarray_get_expression] = STATE(1543),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(715),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [aux_sym_class_application_repeat1] = STATE(717),
    [sym__lowercase_identifier] = ACTIONS(1337),
    [anon_sym_and] = ACTIONS(749),
    [anon_sym_TILDE] = ACTIONS(1339),
    [anon_sym_QMARK] = ACTIONS(1339),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_COMMA] = ACTIONS(747),
    [anon_sym_PIPE] = ACTIONS(749),
    [anon_sym_DOT] = ACTIONS(1341),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_SEMI] = ACTIONS(747),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_in] = ACTIONS(749),
    [anon_sym_AMP] = ACTIONS(749),
    [anon_sym_POUND] = ACTIONS(1343),
    [anon_sym_COLON_COLON] = ACTIONS(747),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_else] = ACTIONS(749),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(747),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(749),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(749),
    [sym_hash_operator] = ACTIONS(1345),
    [anon_sym_lsl] = ACTIONS(749),
    [anon_sym_lsr] = ACTIONS(749),
    [anon_sym_asr] = ACTIONS(749),
    [aux_sym_pow_operator_token1] = ACTIONS(747),
    [anon_sym_mod] = ACTIONS(749),
    [anon_sym_land] = ACTIONS(749),
    [anon_sym_lor] = ACTIONS(749),
    [anon_sym_lxor] = ACTIONS(749),
    [aux_sym_mult_operator_token1] = ACTIONS(749),
    [aux_sym_add_operator_token1] = ACTIONS(749),
    [sym_concat_operator] = ACTIONS(747),
    [sym_rel_operator] = ACTIONS(749),
    [anon_sym_AMP_AMP] = ACTIONS(749),
    [anon_sym_or] = ACTIONS(749),
    [anon_sym_PIPE_PIPE] = ACTIONS(749),
    [sym_assign_operator] = ACTIONS(747),
    [sym_let_and_operator] = ACTIONS(747),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(716)] = {
    [sym__delimited_binding_pattern] = STATE(2051),
    [sym__simple_binding_pattern] = STATE(2085),
    [sym__effect_binding_pattern] = STATE(2060),
    [sym__binding_pattern] = STATE(2064),
    [sym_alias_binding_pattern] = STATE(2060),
    [sym_typed_binding_pattern] = STATE(2051),
    [sym__or_binding_pattern_anonymous] = STATE(2061),
    [sym_constructor_binding_pattern] = STATE(2085),
    [sym_tag_binding_pattern] = STATE(2085),
    [sym_polymorphic_variant_pattern] = STATE(2051),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7213),
    [sym__tuple_binding_pattern] = STATE(2062),
    [sym_record_binding_pattern] = STATE(2087),
    [sym_list_binding_pattern] = STATE(2087),
    [sym_cons_binding_pattern] = STATE(2060),
    [sym_array_binding_pattern] = STATE(2087),
    [sym_range_pattern] = STATE(2060),
    [sym_lazy_binding_pattern] = STATE(2085),
    [sym_local_open_binding_pattern] = STATE(2051),
    [sym_package_pattern] = STATE(2051),
    [sym_parenthesized_binding_pattern] = STATE(2087),
    [sym_exception_binding_pattern] = STATE(2060),
    [sym_effect_binding_pattern] = STATE(2060),
    [sym_attribute] = STATE(716),
    [sym_extension] = STATE(2051),
    [sym_quoted_extension] = STATE(2051),
    [sym__constant] = STATE(1823),
    [sym__signed_constant] = STATE(1825),
    [sym_number] = STATE(1803),
    [sym_signed_number] = STATE(1823),
    [sym_character] = STATE(1803),
    [sym_string] = STATE(1803),
    [sym_quoted_string] = STATE(1803),
    [sym_boolean] = STATE(1803),
    [sym_unit] = STATE(2087),
    [sym__value_name] = STATE(2051),
    [sym_parenthesized_operator] = STATE(2035),
    [sym_module_path] = STATE(6855),
    [sym_constructor_path] = STATE(714),
    [sym__constructor_name] = STATE(1698),
    [sym_tag] = STATE(716),
    [sym__lowercase_identifier] = ACTIONS(1293),
    [anon_sym_SEMI_SEMI] = ACTIONS(1295),
    [anon_sym_let] = ACTIONS(1297),
    [anon_sym_and] = ACTIONS(1297),
    [anon_sym_EQ] = ACTIONS(1295),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_QMARK] = ACTIONS(1295),
    [anon_sym_LPAREN] = ACTIONS(1347),
    [anon_sym_external] = ACTIONS(1297),
    [anon_sym_type] = ACTIONS(1297),
    [anon_sym_COMMA] = ACTIONS(1295),
    [anon_sym_PIPE] = ACTIONS(1295),
    [anon_sym_COLON2] = ACTIONS(1297),
    [anon_sym_LBRACK] = ACTIONS(1303),
    [anon_sym_RBRACK] = ACTIONS(1295),
    [anon_sym_true] = ACTIONS(1305),
    [anon_sym_false] = ACTIONS(1305),
    [anon_sym_LBRACE] = ACTIONS(1307),
    [anon_sym_exception] = ACTIONS(1309),
    [anon_sym_module] = ACTIONS(1297),
    [anon_sym_open] = ACTIONS(1297),
    [anon_sym_include] = ACTIONS(1297),
    [anon_sym_class] = ACTIONS(1297),
    [anon_sym_val] = ACTIONS(1297),
    [anon_sym_end] = ACTIONS(1297),
    [anon_sym_COLON_GT] = ACTIONS(1295),
    [anon_sym_in] = ACTIONS(1297),
    [anon_sym_as] = ACTIONS(1297),
    [anon_sym_POUND] = ACTIONS(1311),
    [anon_sym_COLON_COLON] = ACTIONS(1295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1313),
    [anon_sym_lazy] = ACTIONS(1315),
    [anon_sym_begin] = ACTIONS(1317),
    [anon_sym_effect] = ACTIONS(1319),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1297),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1295),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1321),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1323),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1295),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1295),
    [aux_sym_number_token1] = ACTIONS(1325),
    [aux_sym_signed_number_token1] = ACTIONS(1327),
    [anon_sym_SQUOTE] = ACTIONS(1329),
    [anon_sym_DQUOTE] = ACTIONS(1331),
    [sym_let_operator] = ACTIONS(1295),
    [sym_let_and_operator] = ACTIONS(1295),
    [sym__uppercase_identifier] = ACTIONS(1333),
    [aux_sym_directive_token1] = ACTIONS(1297),
    [aux_sym_tag_token1] = ACTIONS(1335),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(717)] = {
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(1530),
    [sym_typed_expression] = STATE(1543),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_labeled_argument] = STATE(1644),
    [sym_prefix_expression] = STATE(1543),
    [sym_hash_expression] = STATE(1543),
    [sym_field_get_expression] = STATE(1543),
    [sym_array_get_expression] = STATE(1543),
    [sym_string_get_expression] = STATE(1543),
    [sym_bigarray_get_expression] = STATE(1543),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(717),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [aux_sym_class_application_repeat1] = STATE(718),
    [sym__lowercase_identifier] = ACTIONS(1337),
    [anon_sym_and] = ACTIONS(843),
    [anon_sym_TILDE] = ACTIONS(1339),
    [anon_sym_QMARK] = ACTIONS(1339),
    [anon_sym_LPAREN] = ACTIONS(613),
    [anon_sym_COMMA] = ACTIONS(841),
    [anon_sym_PIPE] = ACTIONS(843),
    [anon_sym_LBRACK] = ACTIONS(615),
    [anon_sym_true] = ACTIONS(617),
    [anon_sym_false] = ACTIONS(617),
    [anon_sym_LBRACE] = ACTIONS(619),
    [anon_sym_SEMI] = ACTIONS(841),
    [anon_sym_object] = ACTIONS(621),
    [anon_sym_in] = ACTIONS(843),
    [anon_sym_AMP] = ACTIONS(843),
    [anon_sym_COLON_COLON] = ACTIONS(841),
    [anon_sym_LBRACK_PIPE] = ACTIONS(625),
    [anon_sym_else] = ACTIONS(843),
    [anon_sym_new] = ACTIONS(643),
    [anon_sym_LBRACE_LT] = ACTIONS(645),
    [anon_sym_begin] = ACTIONS(647),
    [sym_ocamlyacc_value] = ACTIONS(649),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(841),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(653),
    [aux_sym_number_token1] = ACTIONS(655),
    [aux_sym_signed_number_token1] = ACTIONS(843),
    [anon_sym_SQUOTE] = ACTIONS(657),
    [anon_sym_DQUOTE] = ACTIONS(659),
    [sym_prefix_operator] = ACTIONS(661),
    [aux_sym_sign_operator_token1] = ACTIONS(843),
    [anon_sym_lsl] = ACTIONS(843),
    [anon_sym_lsr] = ACTIONS(843),
    [anon_sym_asr] = ACTIONS(843),
    [aux_sym_pow_operator_token1] = ACTIONS(841),
    [anon_sym_mod] = ACTIONS(843),
    [anon_sym_land] = ACTIONS(843),
    [anon_sym_lor] = ACTIONS(843),
    [anon_sym_lxor] = ACTIONS(843),
    [aux_sym_mult_operator_token1] = ACTIONS(843),
    [aux_sym_add_operator_token1] = ACTIONS(843),
    [sym_concat_operator] = ACTIONS(841),
    [sym_rel_operator] = ACTIONS(843),
    [anon_sym_AMP_AMP] = ACTIONS(843),
    [anon_sym_or] = ACTIONS(843),
    [anon_sym_PIPE_PIPE] = ACTIONS(843),
    [sym_assign_operator] = ACTIONS(841),
    [sym_let_and_operator] = ACTIONS(841),
    [sym__uppercase_identifier] = ACTIONS(665),
    [aux_sym_tag_token1] = ACTIONS(667),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(718)] = {
    [sym__delimited_expression] = STATE(1543),
    [sym__simple_expression] = STATE(1530),
    [sym_typed_expression] = STATE(1543),
    [sym_list_expression] = STATE(1545),
    [sym_array_expression] = STATE(1545),
    [sym_record_expression] = STATE(1545),
    [sym_labeled_argument] = STATE(1644),
    [sym_prefix_expression] = STATE(1543),
    [sym_hash_expression] = STATE(1543),
    [sym_field_get_expression] = STATE(1543),
    [sym_array_get_expression] = STATE(1543),
    [sym_string_get_expression] = STATE(1543),
    [sym_bigarray_get_expression] = STATE(1543),
    [sym_local_open_expression] = STATE(1543),
    [sym_package_expression] = STATE(1545),
    [sym_new_expression] = STATE(1543),
    [sym_object_copy_expression] = STATE(1545),
    [sym_method_invocation] = STATE(1543),
    [sym_object_expression] = STATE(1543),
    [sym_parenthesized_expression] = STATE(1545),
    [sym_attribute] = STATE(718),
    [sym_extension] = STATE(1543),
    [sym_quoted_extension] = STATE(1543),
    [sym__constant] = STATE(1543),
    [sym_number] = STATE(1596),
    [sym_character] = STATE(1596),
    [sym_string] = STATE(1596),
    [sym_quoted_string] = STATE(1596),
    [sym_boolean] = STATE(1596),
    [sym_unit] = STATE(1545),
    [sym__value_name] = STATE(1580),
    [sym_parenthesized_operator] = STATE(1593),
    [sym_value_path] = STATE(1543),
    [sym_module_path] = STATE(6641),
    [sym_constructor_path] = STATE(1543),
    [sym__constructor_name] = STATE(1539),
    [sym_tag] = STATE(1543),
    [aux_sym_class_application_repeat1] = STATE(718),
    [sym__lowercase_identifier] = ACTIONS(1349),
    [anon_sym_and] = ACTIONS(760),
    [anon_sym_TILDE] = ACTIONS(1352),
    [anon_sym_QMARK] = ACTIONS(1352),
    [anon_sym_LPAREN] = ACTIONS(1355),
    [anon_sym_COMMA] = ACTIONS(768),
    [anon_sym_PIPE] = ACTIONS(760),
    [anon_sym_LBRACK] = ACTIONS(1358),
    [anon_sym_true] = ACTIONS(1361),
    [anon_sym_false] = ACTIONS(1361),
    [anon_sym_LBRACE] = ACTIONS(1364),
    [anon_sym_SEMI] = ACTIONS(768),
    [anon_sym_object] = ACTIONS(1367),
    [anon_sym_in] = ACTIONS(760),
    [anon_sym_AMP] = ACTIONS(760),
    [anon_sym_COLON_COLON] = ACTIONS(768),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1370),
    [anon_sym_else] = ACTIONS(760),
    [anon_sym_new] = ACTIONS(1373),
    [anon_sym_LBRACE_LT] = ACTIONS(1376),
    [anon_sym_begin] = ACTIONS(1379),
    [sym_ocamlyacc_value] = ACTIONS(1382),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(768),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1385),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1388),
    [aux_sym_number_token1] = ACTIONS(1391),
    [aux_sym_signed_number_token1] = ACTIONS(760),
    [anon_sym_SQUOTE] = ACTIONS(1394),
    [anon_sym_DQUOTE] = ACTIONS(1397),
    [sym_prefix_operator] = ACTIONS(1400),
    [aux_sym_sign_operator_token1] = ACTIONS(760),
    [anon_sym_lsl] = ACTIONS(760),
    [anon_sym_lsr] = ACTIONS(760),
    [anon_sym_asr] = ACTIONS(760),
    [aux_sym_pow_operator_token1] = ACTIONS(768),
    [anon_sym_mod] = ACTIONS(760),
    [anon_sym_land] = ACTIONS(760),
    [anon_sym_lor] = ACTIONS(760),
    [anon_sym_lxor] = ACTIONS(760),
    [aux_sym_mult_operator_token1] = ACTIONS(760),
    [aux_sym_add_operator_token1] = ACTIONS(760),
    [sym_concat_operator] = ACTIONS(768),
    [sym_rel_operator] = ACTIONS(760),
    [anon_sym_AMP_AMP] = ACTIONS(760),
    [anon_sym_or] = ACTIONS(760),
    [anon_sym_PIPE_PIPE] = ACTIONS(760),
    [sym_assign_operator] = ACTIONS(768),
    [sym_let_and_operator] = ACTIONS(768),
    [sym__uppercase_identifier] = ACTIONS(1403),
    [aux_sym_tag_token1] = ACTIONS(1406),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(719)] = {
    [sym__local_structure_item] = STATE(7030),
    [sym_let_binding] = STATE(5451),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(719),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__attribute] = STATE(734),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1411),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(720)] = {
    [sym__local_structure_item] = STATE(6571),
    [sym_let_binding] = STATE(5451),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(720),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__attribute] = STATE(735),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1411),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(721)] = {
    [sym__local_structure_item] = STATE(7148),
    [sym_let_binding] = STATE(5451),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(721),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__attribute] = STATE(739),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1411),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(722)] = {
    [sym__local_structure_item] = STATE(7159),
    [sym_let_binding] = STATE(5451),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(722),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__attribute] = STATE(737),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1411),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(723)] = {
    [sym__local_structure_item] = STATE(7126),
    [sym_let_binding] = STATE(5451),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(723),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__attribute] = STATE(738),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1411),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(724)] = {
    [sym__local_structure_item] = STATE(6511),
    [sym_let_binding] = STATE(5451),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(724),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__attribute] = STATE(747),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1411),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(725)] = {
    [sym__local_structure_item] = STATE(6763),
    [sym_let_binding] = STATE(5451),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(725),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__attribute] = STATE(745),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1411),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(726)] = {
    [sym__local_structure_item] = STATE(7005),
    [sym_let_binding] = STATE(5451),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(726),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__attribute] = STATE(743),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1411),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(727)] = {
    [sym__local_structure_item] = STATE(7339),
    [sym_let_binding] = STATE(5451),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(727),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__attribute] = STATE(740),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1411),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(728)] = {
    [sym__local_structure_item] = STATE(6621),
    [sym_let_binding] = STATE(5451),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(728),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__attribute] = STATE(741),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1411),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(729)] = {
    [sym__local_structure_item] = STATE(6489),
    [sym_let_binding] = STATE(5451),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(729),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__attribute] = STATE(744),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1411),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(730)] = {
    [sym__local_structure_item] = STATE(7085),
    [sym_let_binding] = STATE(3646),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(2051),
    [sym__simple_binding_pattern] = STATE(2085),
    [sym__effect_binding_pattern] = STATE(2086),
    [sym__binding_pattern_no_exn] = STATE(733),
    [sym_alias_binding_pattern_no_exn] = STATE(2086),
    [sym_typed_binding_pattern] = STATE(2051),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2086),
    [sym_constructor_binding_pattern] = STATE(2085),
    [sym_tag_binding_pattern] = STATE(2085),
    [sym_polymorphic_variant_pattern] = STATE(2051),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6652),
    [sym__tuple_binding_pattern_no_exn] = STATE(2086),
    [sym_record_binding_pattern] = STATE(2087),
    [sym_list_binding_pattern] = STATE(2087),
    [sym_cons_binding_pattern_no_exn] = STATE(2086),
    [sym_array_binding_pattern] = STATE(2087),
    [sym_range_pattern] = STATE(2086),
    [sym_lazy_binding_pattern] = STATE(2085),
    [sym_local_open_binding_pattern] = STATE(2051),
    [sym_package_pattern] = STATE(2051),
    [sym_parenthesized_binding_pattern] = STATE(2087),
    [sym_attribute] = STATE(730),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(2051),
    [sym_quoted_extension] = STATE(2051),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__attribute] = STATE(736),
    [sym__constant] = STATE(1823),
    [sym__signed_constant] = STATE(1825),
    [sym_number] = STATE(1803),
    [sym_signed_number] = STATE(1823),
    [sym_character] = STATE(1803),
    [sym_string] = STATE(1803),
    [sym_quoted_string] = STATE(1803),
    [sym_boolean] = STATE(1803),
    [sym_unit] = STATE(2087),
    [sym__value_name] = STATE(2051),
    [sym_parenthesized_operator] = STATE(2035),
    [sym_module_path] = STATE(6855),
    [sym_constructor_path] = STATE(714),
    [sym__constructor_name] = STATE(1698),
    [sym_tag] = STATE(716),
    [sym__lowercase_identifier] = ACTIONS(1293),
    [anon_sym_rec] = ACTIONS(1465),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1347),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1303),
    [anon_sym_true] = ACTIONS(1305),
    [anon_sym_false] = ACTIONS(1305),
    [anon_sym_LBRACE] = ACTIONS(1307),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1311),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1313),
    [anon_sym_lazy] = ACTIONS(1315),
    [anon_sym_begin] = ACTIONS(1317),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1321),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1323),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1325),
    [aux_sym_signed_number_token1] = ACTIONS(1327),
    [anon_sym_SQUOTE] = ACTIONS(1329),
    [anon_sym_DQUOTE] = ACTIONS(1331),
    [sym__uppercase_identifier] = ACTIONS(1333),
    [aux_sym_tag_token1] = ACTIONS(1335),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(731)] = {
    [sym__local_structure_item] = STATE(7212),
    [sym_let_binding] = STATE(5451),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(731),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__attribute] = STATE(746),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1411),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(732)] = {
    [sym__local_structure_item] = STATE(7085),
    [sym_let_binding] = STATE(5451),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(732),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__attribute] = STATE(748),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1411),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(733)] = {
    [sym_parameter] = STATE(2207),
    [sym__coerced] = STATE(6454),
    [sym__maybe_polymorphic_typed] = STATE(5734),
    [sym__parenthesized_abstract_type] = STATE(2207),
    [sym__delimited_pattern] = STATE(2286),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2286),
    [sym_polymorphic_variant_pattern] = STATE(2286),
    [sym_record_pattern] = STATE(2287),
    [sym_list_pattern] = STATE(2287),
    [sym_array_pattern] = STATE(2287),
    [sym_local_open_pattern] = STATE(2286),
    [sym_package_pattern] = STATE(2286),
    [sym_parenthesized_pattern] = STATE(2287),
    [sym_attribute] = STATE(733),
    [sym_item_attribute] = STATE(3701),
    [sym_extension] = STATE(2286),
    [sym_quoted_extension] = STATE(2286),
    [sym__constant] = STATE(2212),
    [sym__signed_constant] = STATE(2286),
    [sym_number] = STATE(2280),
    [sym_signed_number] = STATE(2212),
    [sym_character] = STATE(2280),
    [sym_string] = STATE(2280),
    [sym_quoted_string] = STATE(2280),
    [sym_boolean] = STATE(2280),
    [sym_unit] = STATE(2287),
    [sym__simple_value_pattern] = STATE(2286),
    [sym_parenthesized_operator] = STATE(2286),
    [sym_module_path] = STATE(6390),
    [sym_constructor_path] = STATE(2286),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(2286),
    [aux_sym_expression_item_repeat1] = STATE(3498),
    [aux_sym_let_binding_repeat1] = STATE(1511),
    [sym__lowercase_identifier] = ACTIONS(1467),
    [anon_sym_SEMI_SEMI] = ACTIONS(1469),
    [anon_sym_let] = ACTIONS(1471),
    [anon_sym_and] = ACTIONS(1471),
    [anon_sym_EQ] = ACTIONS(1473),
    [anon_sym_TILDE] = ACTIONS(1475),
    [anon_sym_QMARK] = ACTIONS(1475),
    [anon_sym_LPAREN] = ACTIONS(1477),
    [anon_sym_external] = ACTIONS(1471),
    [anon_sym_type] = ACTIONS(1471),
    [anon_sym_COMMA] = ACTIONS(1479),
    [anon_sym_PIPE] = ACTIONS(1481),
    [anon_sym_COLON2] = ACTIONS(1483),
    [anon_sym_LBRACK] = ACTIONS(1485),
    [anon_sym_RBRACK] = ACTIONS(1469),
    [anon_sym_true] = ACTIONS(1487),
    [anon_sym_false] = ACTIONS(1487),
    [anon_sym_LBRACE] = ACTIONS(1489),
    [anon_sym_exception] = ACTIONS(1471),
    [anon_sym_module] = ACTIONS(1471),
    [anon_sym_open] = ACTIONS(1471),
    [anon_sym_include] = ACTIONS(1471),
    [anon_sym_class] = ACTIONS(1471),
    [anon_sym_val] = ACTIONS(1471),
    [anon_sym_end] = ACTIONS(1471),
    [anon_sym_COLON_GT] = ACTIONS(1491),
    [anon_sym_in] = ACTIONS(1471),
    [anon_sym_as] = ACTIONS(1493),
    [anon_sym_POUND] = ACTIONS(1495),
    [anon_sym_COLON_COLON] = ACTIONS(1497),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1499),
    [anon_sym_begin] = ACTIONS(1501),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1503),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1469),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1469),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1469),
    [aux_sym_number_token1] = ACTIONS(1509),
    [aux_sym_signed_number_token1] = ACTIONS(1511),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE] = ACTIONS(1515),
    [sym_let_operator] = ACTIONS(1469),
    [sym_let_and_operator] = ACTIONS(1469),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_directive_token1] = ACTIONS(1471),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(734)] = {
    [sym__local_structure_item] = STATE(7034),
    [sym_let_binding] = STATE(5505),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(734),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(735)] = {
    [sym__local_structure_item] = STATE(6713),
    [sym_let_binding] = STATE(5505),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(735),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(736)] = {
    [sym__local_structure_item] = STATE(6522),
    [sym_let_binding] = STATE(3605),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(2051),
    [sym__simple_binding_pattern] = STATE(2085),
    [sym__effect_binding_pattern] = STATE(2086),
    [sym__binding_pattern_no_exn] = STATE(733),
    [sym_alias_binding_pattern_no_exn] = STATE(2086),
    [sym_typed_binding_pattern] = STATE(2051),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2086),
    [sym_constructor_binding_pattern] = STATE(2085),
    [sym_tag_binding_pattern] = STATE(2085),
    [sym_polymorphic_variant_pattern] = STATE(2051),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6652),
    [sym__tuple_binding_pattern_no_exn] = STATE(2086),
    [sym_record_binding_pattern] = STATE(2087),
    [sym_list_binding_pattern] = STATE(2087),
    [sym_cons_binding_pattern_no_exn] = STATE(2086),
    [sym_array_binding_pattern] = STATE(2087),
    [sym_range_pattern] = STATE(2086),
    [sym_lazy_binding_pattern] = STATE(2085),
    [sym_local_open_binding_pattern] = STATE(2051),
    [sym_package_pattern] = STATE(2051),
    [sym_parenthesized_binding_pattern] = STATE(2087),
    [sym_attribute] = STATE(736),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(2051),
    [sym_quoted_extension] = STATE(2051),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__constant] = STATE(1823),
    [sym__signed_constant] = STATE(1825),
    [sym_number] = STATE(1803),
    [sym_signed_number] = STATE(1823),
    [sym_character] = STATE(1803),
    [sym_string] = STATE(1803),
    [sym_quoted_string] = STATE(1803),
    [sym_boolean] = STATE(1803),
    [sym_unit] = STATE(2087),
    [sym__value_name] = STATE(2051),
    [sym_parenthesized_operator] = STATE(2035),
    [sym_module_path] = STATE(6855),
    [sym_constructor_path] = STATE(714),
    [sym__constructor_name] = STATE(1698),
    [sym_tag] = STATE(716),
    [sym__lowercase_identifier] = ACTIONS(1293),
    [anon_sym_rec] = ACTIONS(1519),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1347),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1303),
    [anon_sym_true] = ACTIONS(1305),
    [anon_sym_false] = ACTIONS(1305),
    [anon_sym_LBRACE] = ACTIONS(1307),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1311),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1313),
    [anon_sym_lazy] = ACTIONS(1315),
    [anon_sym_begin] = ACTIONS(1317),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1321),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1323),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [aux_sym_number_token1] = ACTIONS(1325),
    [aux_sym_signed_number_token1] = ACTIONS(1327),
    [anon_sym_SQUOTE] = ACTIONS(1329),
    [anon_sym_DQUOTE] = ACTIONS(1331),
    [sym__uppercase_identifier] = ACTIONS(1333),
    [aux_sym_tag_token1] = ACTIONS(1335),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(737)] = {
    [sym__local_structure_item] = STATE(7251),
    [sym_let_binding] = STATE(5505),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(737),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(738)] = {
    [sym__local_structure_item] = STATE(7232),
    [sym_let_binding] = STATE(5505),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(738),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(739)] = {
    [sym__local_structure_item] = STATE(6463),
    [sym_let_binding] = STATE(5505),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(739),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(740)] = {
    [sym__local_structure_item] = STATE(7326),
    [sym_let_binding] = STATE(5505),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(740),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(741)] = {
    [sym__local_structure_item] = STATE(7049),
    [sym_let_binding] = STATE(5505),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(741),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(742)] = {
    [sym_parameter] = STATE(2207),
    [sym__coerced] = STATE(6387),
    [sym__maybe_polymorphic_typed] = STATE(5720),
    [sym__parenthesized_abstract_type] = STATE(2207),
    [sym__delimited_pattern] = STATE(2286),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2286),
    [sym_polymorphic_variant_pattern] = STATE(2286),
    [sym_record_pattern] = STATE(2287),
    [sym_list_pattern] = STATE(2287),
    [sym_array_pattern] = STATE(2287),
    [sym_local_open_pattern] = STATE(2286),
    [sym_package_pattern] = STATE(2286),
    [sym_parenthesized_pattern] = STATE(2287),
    [sym_attribute] = STATE(742),
    [sym_item_attribute] = STATE(3701),
    [sym_extension] = STATE(2286),
    [sym_quoted_extension] = STATE(2286),
    [sym__constant] = STATE(2212),
    [sym__signed_constant] = STATE(2286),
    [sym_number] = STATE(2280),
    [sym_signed_number] = STATE(2212),
    [sym_character] = STATE(2280),
    [sym_string] = STATE(2280),
    [sym_quoted_string] = STATE(2280),
    [sym_boolean] = STATE(2280),
    [sym_unit] = STATE(2287),
    [sym__simple_value_pattern] = STATE(2286),
    [sym_parenthesized_operator] = STATE(2286),
    [sym_module_path] = STATE(6390),
    [sym_constructor_path] = STATE(2286),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(2286),
    [aux_sym_expression_item_repeat1] = STATE(3498),
    [aux_sym_let_binding_repeat1] = STATE(1509),
    [sym__lowercase_identifier] = ACTIONS(1467),
    [anon_sym_SEMI_SEMI] = ACTIONS(1469),
    [anon_sym_let] = ACTIONS(1471),
    [anon_sym_and] = ACTIONS(1471),
    [anon_sym_EQ] = ACTIONS(1521),
    [anon_sym_TILDE] = ACTIONS(1475),
    [anon_sym_QMARK] = ACTIONS(1475),
    [anon_sym_LPAREN] = ACTIONS(1477),
    [anon_sym_external] = ACTIONS(1471),
    [anon_sym_type] = ACTIONS(1471),
    [anon_sym_COMMA] = ACTIONS(1479),
    [anon_sym_PIPE] = ACTIONS(1481),
    [anon_sym_COLON2] = ACTIONS(1483),
    [anon_sym_LBRACK] = ACTIONS(1485),
    [anon_sym_RBRACK] = ACTIONS(1469),
    [anon_sym_true] = ACTIONS(1487),
    [anon_sym_false] = ACTIONS(1487),
    [anon_sym_LBRACE] = ACTIONS(1489),
    [anon_sym_exception] = ACTIONS(1471),
    [anon_sym_module] = ACTIONS(1471),
    [anon_sym_open] = ACTIONS(1471),
    [anon_sym_include] = ACTIONS(1471),
    [anon_sym_class] = ACTIONS(1471),
    [anon_sym_val] = ACTIONS(1471),
    [anon_sym_end] = ACTIONS(1471),
    [anon_sym_COLON_GT] = ACTIONS(1491),
    [anon_sym_as] = ACTIONS(1493),
    [anon_sym_POUND] = ACTIONS(1495),
    [anon_sym_COLON_COLON] = ACTIONS(1497),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1499),
    [anon_sym_begin] = ACTIONS(1501),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1503),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1469),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1469),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1469),
    [aux_sym_number_token1] = ACTIONS(1509),
    [aux_sym_signed_number_token1] = ACTIONS(1511),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE] = ACTIONS(1515),
    [sym_let_operator] = ACTIONS(1469),
    [sym_let_and_operator] = ACTIONS(1469),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_directive_token1] = ACTIONS(1471),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(743)] = {
    [sym__local_structure_item] = STATE(7042),
    [sym_let_binding] = STATE(5505),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(743),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(744)] = {
    [sym__local_structure_item] = STATE(6563),
    [sym_let_binding] = STATE(5505),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(744),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(745)] = {
    [sym__local_structure_item] = STATE(6898),
    [sym_let_binding] = STATE(5505),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(745),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(746)] = {
    [sym__local_structure_item] = STATE(7233),
    [sym_let_binding] = STATE(5505),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(746),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(747)] = {
    [sym__local_structure_item] = STATE(6584),
    [sym_let_binding] = STATE(5505),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(747),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(748)] = {
    [sym__local_structure_item] = STATE(6522),
    [sym_let_binding] = STATE(5505),
    [sym_external] = STATE(7154),
    [sym_type_definition] = STATE(7154),
    [sym_exception_definition] = STATE(7154),
    [sym_module_definition] = STATE(7154),
    [sym_module_type_definition] = STATE(7154),
    [sym_open_module] = STATE(7154),
    [sym_class_definition] = STATE(7154),
    [sym_class_type_definition] = STATE(7154),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(748),
    [sym_floating_attribute] = STATE(7154),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym_item_extension] = STATE(7154),
    [sym_quoted_item_extension] = STATE(7154),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1517),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_external] = ACTIONS(1415),
    [anon_sym_type] = ACTIONS(1417),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1425),
    [anon_sym_module] = ACTIONS(1427),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_class] = ACTIONS(1431),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1441),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1443),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1445),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1447),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1449),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(749)] = {
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(1078),
    [sym_typed_expression] = STATE(1052),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_labeled_argument] = STATE(1144),
    [sym_prefix_expression] = STATE(1052),
    [sym_hash_expression] = STATE(1052),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(749),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [aux_sym_class_application_repeat1] = STATE(750),
    [sym__lowercase_identifier] = ACTIONS(965),
    [anon_sym_SEMI_SEMI] = ACTIONS(1523),
    [anon_sym_let] = ACTIONS(1525),
    [anon_sym_and] = ACTIONS(1525),
    [anon_sym_TILDE] = ACTIONS(967),
    [anon_sym_QMARK] = ACTIONS(967),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(1525),
    [anon_sym_type] = ACTIONS(1525),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(1523),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(1525),
    [anon_sym_module] = ACTIONS(1525),
    [anon_sym_open] = ACTIONS(1525),
    [anon_sym_include] = ACTIONS(1525),
    [anon_sym_class] = ACTIONS(1525),
    [anon_sym_val] = ACTIONS(1525),
    [anon_sym_end] = ACTIONS(1525),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1525),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1523),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1523),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1523),
    [aux_sym_number_token1] = ACTIONS(217),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [sym_let_operator] = ACTIONS(1523),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(1523),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(750)] = {
    [sym__delimited_expression] = STATE(1052),
    [sym__simple_expression] = STATE(1078),
    [sym_typed_expression] = STATE(1052),
    [sym_list_expression] = STATE(1118),
    [sym_array_expression] = STATE(1118),
    [sym_record_expression] = STATE(1118),
    [sym_labeled_argument] = STATE(1144),
    [sym_prefix_expression] = STATE(1052),
    [sym_hash_expression] = STATE(1052),
    [sym_field_get_expression] = STATE(1052),
    [sym_array_get_expression] = STATE(1052),
    [sym_string_get_expression] = STATE(1052),
    [sym_bigarray_get_expression] = STATE(1052),
    [sym_local_open_expression] = STATE(1052),
    [sym_package_expression] = STATE(1118),
    [sym_new_expression] = STATE(1052),
    [sym_object_copy_expression] = STATE(1118),
    [sym_method_invocation] = STATE(1052),
    [sym_object_expression] = STATE(1052),
    [sym_parenthesized_expression] = STATE(1118),
    [sym_attribute] = STATE(750),
    [sym_extension] = STATE(1052),
    [sym_quoted_extension] = STATE(1052),
    [sym__constant] = STATE(1052),
    [sym_number] = STATE(1057),
    [sym_character] = STATE(1057),
    [sym_string] = STATE(1057),
    [sym_quoted_string] = STATE(1057),
    [sym_boolean] = STATE(1057),
    [sym_unit] = STATE(1118),
    [sym__value_name] = STATE(1076),
    [sym_parenthesized_operator] = STATE(1077),
    [sym_value_path] = STATE(1052),
    [sym_module_path] = STATE(7309),
    [sym_constructor_path] = STATE(1052),
    [sym__constructor_name] = STATE(1079),
    [sym_tag] = STATE(1052),
    [aux_sym_class_application_repeat1] = STATE(89),
    [sym__lowercase_identifier] = ACTIONS(965),
    [anon_sym_SEMI_SEMI] = ACTIONS(1527),
    [anon_sym_let] = ACTIONS(1529),
    [anon_sym_and] = ACTIONS(1529),
    [anon_sym_TILDE] = ACTIONS(967),
    [anon_sym_QMARK] = ACTIONS(967),
    [anon_sym_LPAREN] = ACTIONS(183),
    [anon_sym_external] = ACTIONS(1529),
    [anon_sym_type] = ACTIONS(1529),
    [anon_sym_LBRACK] = ACTIONS(185),
    [anon_sym_RBRACK] = ACTIONS(1527),
    [anon_sym_true] = ACTIONS(187),
    [anon_sym_false] = ACTIONS(187),
    [anon_sym_LBRACE] = ACTIONS(189),
    [anon_sym_exception] = ACTIONS(1529),
    [anon_sym_module] = ACTIONS(1529),
    [anon_sym_open] = ACTIONS(1529),
    [anon_sym_include] = ACTIONS(1529),
    [anon_sym_class] = ACTIONS(1529),
    [anon_sym_val] = ACTIONS(1529),
    [anon_sym_end] = ACTIONS(1529),
    [anon_sym_object] = ACTIONS(191),
    [anon_sym_LBRACK_PIPE] = ACTIONS(195),
    [anon_sym_new] = ACTIONS(205),
    [anon_sym_LBRACE_LT] = ACTIONS(207),
    [anon_sym_begin] = ACTIONS(209),
    [sym_ocamlyacc_value] = ACTIONS(211),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1529),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1527),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(213),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(215),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1527),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1527),
    [aux_sym_number_token1] = ACTIONS(217),
    [anon_sym_SQUOTE] = ACTIONS(219),
    [anon_sym_DQUOTE] = ACTIONS(221),
    [sym_prefix_operator] = ACTIONS(223),
    [sym_let_operator] = ACTIONS(1527),
    [sym__uppercase_identifier] = ACTIONS(227),
    [aux_sym_directive_token1] = ACTIONS(1527),
    [aux_sym_tag_token1] = ACTIONS(229),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(751)] = {
    [sym_attribute] = STATE(751),
    [sym__lowercase_identifier] = ACTIONS(1531),
    [anon_sym_and] = ACTIONS(1531),
    [anon_sym_TILDE] = ACTIONS(1531),
    [anon_sym_QMARK] = ACTIONS(1531),
    [anon_sym_LPAREN] = ACTIONS(1533),
    [anon_sym_RPAREN] = ACTIONS(1533),
    [anon_sym_COMMA] = ACTIONS(1533),
    [anon_sym_PIPE] = ACTIONS(1531),
    [anon_sym_COLON2] = ACTIONS(1531),
    [anon_sym_DOT] = ACTIONS(1533),
    [anon_sym_DASH_GT] = ACTIONS(1531),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_RBRACK] = ACTIONS(1533),
    [anon_sym_true] = ACTIONS(1531),
    [anon_sym_false] = ACTIONS(1531),
    [anon_sym_LBRACE] = ACTIONS(1531),
    [anon_sym_SEMI] = ACTIONS(1533),
    [anon_sym_RBRACE] = ACTIONS(1533),
    [anon_sym_constraint] = ACTIONS(1531),
    [anon_sym_val] = ACTIONS(1531),
    [anon_sym_end] = ACTIONS(1531),
    [anon_sym_with] = ACTIONS(1531),
    [anon_sym_COLON_GT] = ACTIONS(1533),
    [anon_sym_object] = ACTIONS(1531),
    [anon_sym_inherit] = ACTIONS(1531),
    [anon_sym_method] = ACTIONS(1531),
    [anon_sym_in] = ACTIONS(1531),
    [anon_sym_as] = ACTIONS(1531),
    [anon_sym_initializer] = ACTIONS(1531),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_POUND] = ACTIONS(1531),
    [anon_sym_COLON_COLON] = ACTIONS(1533),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1533),
    [anon_sym_LT_DASH] = ACTIONS(1531),
    [anon_sym_then] = ACTIONS(1531),
    [anon_sym_else] = ACTIONS(1531),
    [anon_sym_do] = ACTIONS(1531),
    [anon_sym_done] = ACTIONS(1531),
    [anon_sym_to] = ACTIONS(1531),
    [anon_sym_downto] = ACTIONS(1531),
    [anon_sym_new] = ACTIONS(1531),
    [anon_sym_LBRACE_LT] = ACTIONS(1533),
    [anon_sym_GT_RBRACE] = ACTIONS(1533),
    [anon_sym_begin] = ACTIONS(1531),
    [sym_ocamlyacc_value] = ACTIONS(1533),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1531),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1533),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1531),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1531),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1533),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1533),
    [aux_sym_number_token1] = ACTIONS(1533),
    [aux_sym_signed_number_token1] = ACTIONS(1531),
    [anon_sym_SQUOTE] = ACTIONS(1533),
    [anon_sym_DQUOTE] = ACTIONS(1533),
    [sym_prefix_operator] = ACTIONS(1533),
    [aux_sym_sign_operator_token1] = ACTIONS(1531),
    [sym_hash_operator] = ACTIONS(1533),
    [anon_sym_lsl] = ACTIONS(1531),
    [anon_sym_lsr] = ACTIONS(1531),
    [anon_sym_asr] = ACTIONS(1531),
    [aux_sym_pow_operator_token1] = ACTIONS(1533),
    [anon_sym_mod] = ACTIONS(1531),
    [anon_sym_land] = ACTIONS(1531),
    [anon_sym_lor] = ACTIONS(1531),
    [anon_sym_lxor] = ACTIONS(1531),
    [aux_sym_mult_operator_token1] = ACTIONS(1531),
    [aux_sym_add_operator_token1] = ACTIONS(1531),
    [sym_concat_operator] = ACTIONS(1533),
    [sym_rel_operator] = ACTIONS(1531),
    [anon_sym_AMP_AMP] = ACTIONS(1531),
    [anon_sym_or] = ACTIONS(1531),
    [anon_sym_PIPE_PIPE] = ACTIONS(1531),
    [sym_assign_operator] = ACTIONS(1533),
    [sym__uppercase_identifier] = ACTIONS(1533),
    [aux_sym_tag_token1] = ACTIONS(1533),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(752)] = {
    [sym_attribute] = STATE(752),
    [sym__lowercase_identifier] = ACTIONS(1535),
    [anon_sym_and] = ACTIONS(1535),
    [anon_sym_TILDE] = ACTIONS(1535),
    [anon_sym_QMARK] = ACTIONS(1535),
    [anon_sym_LPAREN] = ACTIONS(1537),
    [anon_sym_RPAREN] = ACTIONS(1537),
    [anon_sym_COMMA] = ACTIONS(1537),
    [anon_sym_PIPE] = ACTIONS(1535),
    [anon_sym_COLON2] = ACTIONS(1535),
    [anon_sym_DOT] = ACTIONS(1537),
    [anon_sym_DASH_GT] = ACTIONS(1535),
    [anon_sym_LBRACK] = ACTIONS(1535),
    [anon_sym_RBRACK] = ACTIONS(1537),
    [anon_sym_true] = ACTIONS(1535),
    [anon_sym_false] = ACTIONS(1535),
    [anon_sym_LBRACE] = ACTIONS(1535),
    [anon_sym_SEMI] = ACTIONS(1537),
    [anon_sym_RBRACE] = ACTIONS(1537),
    [anon_sym_constraint] = ACTIONS(1535),
    [anon_sym_val] = ACTIONS(1535),
    [anon_sym_end] = ACTIONS(1535),
    [anon_sym_with] = ACTIONS(1535),
    [anon_sym_COLON_GT] = ACTIONS(1537),
    [anon_sym_object] = ACTIONS(1535),
    [anon_sym_inherit] = ACTIONS(1535),
    [anon_sym_method] = ACTIONS(1535),
    [anon_sym_in] = ACTIONS(1535),
    [anon_sym_as] = ACTIONS(1535),
    [anon_sym_initializer] = ACTIONS(1535),
    [anon_sym_AMP] = ACTIONS(1535),
    [anon_sym_POUND] = ACTIONS(1535),
    [anon_sym_COLON_COLON] = ACTIONS(1537),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1537),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1537),
    [anon_sym_LT_DASH] = ACTIONS(1535),
    [anon_sym_then] = ACTIONS(1535),
    [anon_sym_else] = ACTIONS(1535),
    [anon_sym_do] = ACTIONS(1535),
    [anon_sym_done] = ACTIONS(1535),
    [anon_sym_to] = ACTIONS(1535),
    [anon_sym_downto] = ACTIONS(1535),
    [anon_sym_new] = ACTIONS(1535),
    [anon_sym_LBRACE_LT] = ACTIONS(1537),
    [anon_sym_GT_RBRACE] = ACTIONS(1537),
    [anon_sym_begin] = ACTIONS(1535),
    [sym_ocamlyacc_value] = ACTIONS(1537),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1535),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1537),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1535),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1535),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1537),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1537),
    [aux_sym_number_token1] = ACTIONS(1537),
    [aux_sym_signed_number_token1] = ACTIONS(1535),
    [anon_sym_SQUOTE] = ACTIONS(1537),
    [anon_sym_DQUOTE] = ACTIONS(1537),
    [sym_prefix_operator] = ACTIONS(1537),
    [aux_sym_sign_operator_token1] = ACTIONS(1535),
    [sym_hash_operator] = ACTIONS(1537),
    [anon_sym_lsl] = ACTIONS(1535),
    [anon_sym_lsr] = ACTIONS(1535),
    [anon_sym_asr] = ACTIONS(1535),
    [aux_sym_pow_operator_token1] = ACTIONS(1537),
    [anon_sym_mod] = ACTIONS(1535),
    [anon_sym_land] = ACTIONS(1535),
    [anon_sym_lor] = ACTIONS(1535),
    [anon_sym_lxor] = ACTIONS(1535),
    [aux_sym_mult_operator_token1] = ACTIONS(1535),
    [aux_sym_add_operator_token1] = ACTIONS(1535),
    [sym_concat_operator] = ACTIONS(1537),
    [sym_rel_operator] = ACTIONS(1535),
    [anon_sym_AMP_AMP] = ACTIONS(1535),
    [anon_sym_or] = ACTIONS(1535),
    [anon_sym_PIPE_PIPE] = ACTIONS(1535),
    [sym_assign_operator] = ACTIONS(1537),
    [sym__uppercase_identifier] = ACTIONS(1537),
    [aux_sym_tag_token1] = ACTIONS(1537),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(753)] = {
    [sym_attribute] = STATE(753),
    [sym__lowercase_identifier] = ACTIONS(1539),
    [anon_sym_and] = ACTIONS(1539),
    [anon_sym_TILDE] = ACTIONS(1539),
    [anon_sym_QMARK] = ACTIONS(1539),
    [anon_sym_LPAREN] = ACTIONS(1541),
    [anon_sym_RPAREN] = ACTIONS(1541),
    [anon_sym_COMMA] = ACTIONS(1541),
    [anon_sym_PIPE] = ACTIONS(1539),
    [anon_sym_COLON2] = ACTIONS(1539),
    [anon_sym_DOT] = ACTIONS(1541),
    [anon_sym_DASH_GT] = ACTIONS(1539),
    [anon_sym_LBRACK] = ACTIONS(1539),
    [anon_sym_RBRACK] = ACTIONS(1541),
    [anon_sym_true] = ACTIONS(1539),
    [anon_sym_false] = ACTIONS(1539),
    [anon_sym_LBRACE] = ACTIONS(1539),
    [anon_sym_SEMI] = ACTIONS(1541),
    [anon_sym_RBRACE] = ACTIONS(1541),
    [anon_sym_constraint] = ACTIONS(1539),
    [anon_sym_val] = ACTIONS(1539),
    [anon_sym_end] = ACTIONS(1539),
    [anon_sym_with] = ACTIONS(1539),
    [anon_sym_COLON_GT] = ACTIONS(1541),
    [anon_sym_object] = ACTIONS(1539),
    [anon_sym_inherit] = ACTIONS(1539),
    [anon_sym_method] = ACTIONS(1539),
    [anon_sym_in] = ACTIONS(1539),
    [anon_sym_as] = ACTIONS(1539),
    [anon_sym_initializer] = ACTIONS(1539),
    [anon_sym_AMP] = ACTIONS(1539),
    [anon_sym_POUND] = ACTIONS(1539),
    [anon_sym_COLON_COLON] = ACTIONS(1541),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1541),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1541),
    [anon_sym_LT_DASH] = ACTIONS(1539),
    [anon_sym_then] = ACTIONS(1539),
    [anon_sym_else] = ACTIONS(1539),
    [anon_sym_do] = ACTIONS(1539),
    [anon_sym_done] = ACTIONS(1539),
    [anon_sym_to] = ACTIONS(1539),
    [anon_sym_downto] = ACTIONS(1539),
    [anon_sym_new] = ACTIONS(1539),
    [anon_sym_LBRACE_LT] = ACTIONS(1541),
    [anon_sym_GT_RBRACE] = ACTIONS(1541),
    [anon_sym_begin] = ACTIONS(1539),
    [sym_ocamlyacc_value] = ACTIONS(1541),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1539),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1541),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1539),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1539),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1541),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1541),
    [aux_sym_number_token1] = ACTIONS(1541),
    [aux_sym_signed_number_token1] = ACTIONS(1539),
    [anon_sym_SQUOTE] = ACTIONS(1541),
    [anon_sym_DQUOTE] = ACTIONS(1541),
    [sym_prefix_operator] = ACTIONS(1541),
    [aux_sym_sign_operator_token1] = ACTIONS(1539),
    [sym_hash_operator] = ACTIONS(1541),
    [anon_sym_lsl] = ACTIONS(1539),
    [anon_sym_lsr] = ACTIONS(1539),
    [anon_sym_asr] = ACTIONS(1539),
    [aux_sym_pow_operator_token1] = ACTIONS(1541),
    [anon_sym_mod] = ACTIONS(1539),
    [anon_sym_land] = ACTIONS(1539),
    [anon_sym_lor] = ACTIONS(1539),
    [anon_sym_lxor] = ACTIONS(1539),
    [aux_sym_mult_operator_token1] = ACTIONS(1539),
    [aux_sym_add_operator_token1] = ACTIONS(1539),
    [sym_concat_operator] = ACTIONS(1541),
    [sym_rel_operator] = ACTIONS(1539),
    [anon_sym_AMP_AMP] = ACTIONS(1539),
    [anon_sym_or] = ACTIONS(1539),
    [anon_sym_PIPE_PIPE] = ACTIONS(1539),
    [sym_assign_operator] = ACTIONS(1541),
    [sym__uppercase_identifier] = ACTIONS(1541),
    [aux_sym_tag_token1] = ACTIONS(1541),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(754)] = {
    [sym_attribute] = STATE(754),
    [sym__lowercase_identifier] = ACTIONS(1543),
    [anon_sym_and] = ACTIONS(1543),
    [anon_sym_TILDE] = ACTIONS(1543),
    [anon_sym_QMARK] = ACTIONS(1543),
    [anon_sym_LPAREN] = ACTIONS(1545),
    [anon_sym_RPAREN] = ACTIONS(1545),
    [anon_sym_COMMA] = ACTIONS(1545),
    [anon_sym_PIPE] = ACTIONS(1543),
    [anon_sym_COLON2] = ACTIONS(1543),
    [anon_sym_DOT] = ACTIONS(1545),
    [anon_sym_DASH_GT] = ACTIONS(1543),
    [anon_sym_LBRACK] = ACTIONS(1543),
    [anon_sym_RBRACK] = ACTIONS(1545),
    [anon_sym_true] = ACTIONS(1543),
    [anon_sym_false] = ACTIONS(1543),
    [anon_sym_LBRACE] = ACTIONS(1543),
    [anon_sym_SEMI] = ACTIONS(1545),
    [anon_sym_RBRACE] = ACTIONS(1545),
    [anon_sym_constraint] = ACTIONS(1543),
    [anon_sym_val] = ACTIONS(1543),
    [anon_sym_end] = ACTIONS(1543),
    [anon_sym_with] = ACTIONS(1543),
    [anon_sym_COLON_GT] = ACTIONS(1545),
    [anon_sym_object] = ACTIONS(1543),
    [anon_sym_inherit] = ACTIONS(1543),
    [anon_sym_method] = ACTIONS(1543),
    [anon_sym_in] = ACTIONS(1543),
    [anon_sym_as] = ACTIONS(1543),
    [anon_sym_initializer] = ACTIONS(1543),
    [anon_sym_AMP] = ACTIONS(1543),
    [anon_sym_POUND] = ACTIONS(1543),
    [anon_sym_COLON_COLON] = ACTIONS(1545),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1545),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1545),
    [anon_sym_LT_DASH] = ACTIONS(1543),
    [anon_sym_then] = ACTIONS(1543),
    [anon_sym_else] = ACTIONS(1543),
    [anon_sym_do] = ACTIONS(1543),
    [anon_sym_done] = ACTIONS(1543),
    [anon_sym_to] = ACTIONS(1543),
    [anon_sym_downto] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1543),
    [anon_sym_LBRACE_LT] = ACTIONS(1545),
    [anon_sym_GT_RBRACE] = ACTIONS(1545),
    [anon_sym_begin] = ACTIONS(1543),
    [sym_ocamlyacc_value] = ACTIONS(1545),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1543),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1545),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1543),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1543),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1545),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1545),
    [aux_sym_number_token1] = ACTIONS(1545),
    [aux_sym_signed_number_token1] = ACTIONS(1543),
    [anon_sym_SQUOTE] = ACTIONS(1545),
    [anon_sym_DQUOTE] = ACTIONS(1545),
    [sym_prefix_operator] = ACTIONS(1545),
    [aux_sym_sign_operator_token1] = ACTIONS(1543),
    [sym_hash_operator] = ACTIONS(1545),
    [anon_sym_lsl] = ACTIONS(1543),
    [anon_sym_lsr] = ACTIONS(1543),
    [anon_sym_asr] = ACTIONS(1543),
    [aux_sym_pow_operator_token1] = ACTIONS(1545),
    [anon_sym_mod] = ACTIONS(1543),
    [anon_sym_land] = ACTIONS(1543),
    [anon_sym_lor] = ACTIONS(1543),
    [anon_sym_lxor] = ACTIONS(1543),
    [aux_sym_mult_operator_token1] = ACTIONS(1543),
    [aux_sym_add_operator_token1] = ACTIONS(1543),
    [sym_concat_operator] = ACTIONS(1545),
    [sym_rel_operator] = ACTIONS(1543),
    [anon_sym_AMP_AMP] = ACTIONS(1543),
    [anon_sym_or] = ACTIONS(1543),
    [anon_sym_PIPE_PIPE] = ACTIONS(1543),
    [sym_assign_operator] = ACTIONS(1545),
    [sym__uppercase_identifier] = ACTIONS(1545),
    [aux_sym_tag_token1] = ACTIONS(1545),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(755)] = {
    [sym_attribute] = STATE(755),
    [sym__lowercase_identifier] = ACTIONS(1547),
    [anon_sym_and] = ACTIONS(1547),
    [anon_sym_TILDE] = ACTIONS(1547),
    [anon_sym_QMARK] = ACTIONS(1547),
    [anon_sym_LPAREN] = ACTIONS(1549),
    [anon_sym_RPAREN] = ACTIONS(1549),
    [anon_sym_COMMA] = ACTIONS(1549),
    [anon_sym_PIPE] = ACTIONS(1547),
    [anon_sym_COLON2] = ACTIONS(1547),
    [anon_sym_DOT] = ACTIONS(1549),
    [anon_sym_DASH_GT] = ACTIONS(1547),
    [anon_sym_LBRACK] = ACTIONS(1547),
    [anon_sym_RBRACK] = ACTIONS(1549),
    [anon_sym_true] = ACTIONS(1547),
    [anon_sym_false] = ACTIONS(1547),
    [anon_sym_LBRACE] = ACTIONS(1547),
    [anon_sym_SEMI] = ACTIONS(1549),
    [anon_sym_RBRACE] = ACTIONS(1549),
    [anon_sym_constraint] = ACTIONS(1547),
    [anon_sym_val] = ACTIONS(1547),
    [anon_sym_end] = ACTIONS(1547),
    [anon_sym_with] = ACTIONS(1547),
    [anon_sym_COLON_GT] = ACTIONS(1549),
    [anon_sym_object] = ACTIONS(1547),
    [anon_sym_inherit] = ACTIONS(1547),
    [anon_sym_method] = ACTIONS(1547),
    [anon_sym_in] = ACTIONS(1547),
    [anon_sym_as] = ACTIONS(1547),
    [anon_sym_initializer] = ACTIONS(1547),
    [anon_sym_AMP] = ACTIONS(1547),
    [anon_sym_POUND] = ACTIONS(1547),
    [anon_sym_COLON_COLON] = ACTIONS(1549),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1549),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1549),
    [anon_sym_LT_DASH] = ACTIONS(1547),
    [anon_sym_then] = ACTIONS(1547),
    [anon_sym_else] = ACTIONS(1547),
    [anon_sym_do] = ACTIONS(1547),
    [anon_sym_done] = ACTIONS(1547),
    [anon_sym_to] = ACTIONS(1547),
    [anon_sym_downto] = ACTIONS(1547),
    [anon_sym_new] = ACTIONS(1547),
    [anon_sym_LBRACE_LT] = ACTIONS(1549),
    [anon_sym_GT_RBRACE] = ACTIONS(1549),
    [anon_sym_begin] = ACTIONS(1547),
    [sym_ocamlyacc_value] = ACTIONS(1549),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1547),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1549),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1547),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1547),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1549),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1549),
    [aux_sym_number_token1] = ACTIONS(1549),
    [aux_sym_signed_number_token1] = ACTIONS(1547),
    [anon_sym_SQUOTE] = ACTIONS(1549),
    [anon_sym_DQUOTE] = ACTIONS(1549),
    [sym_prefix_operator] = ACTIONS(1549),
    [aux_sym_sign_operator_token1] = ACTIONS(1547),
    [sym_hash_operator] = ACTIONS(1549),
    [anon_sym_lsl] = ACTIONS(1547),
    [anon_sym_lsr] = ACTIONS(1547),
    [anon_sym_asr] = ACTIONS(1547),
    [aux_sym_pow_operator_token1] = ACTIONS(1549),
    [anon_sym_mod] = ACTIONS(1547),
    [anon_sym_land] = ACTIONS(1547),
    [anon_sym_lor] = ACTIONS(1547),
    [anon_sym_lxor] = ACTIONS(1547),
    [aux_sym_mult_operator_token1] = ACTIONS(1547),
    [aux_sym_add_operator_token1] = ACTIONS(1547),
    [sym_concat_operator] = ACTIONS(1549),
    [sym_rel_operator] = ACTIONS(1547),
    [anon_sym_AMP_AMP] = ACTIONS(1547),
    [anon_sym_or] = ACTIONS(1547),
    [anon_sym_PIPE_PIPE] = ACTIONS(1547),
    [sym_assign_operator] = ACTIONS(1549),
    [sym__uppercase_identifier] = ACTIONS(1549),
    [aux_sym_tag_token1] = ACTIONS(1549),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(756)] = {
    [sym_attribute] = STATE(756),
    [sym__lowercase_identifier] = ACTIONS(1551),
    [anon_sym_and] = ACTIONS(1551),
    [anon_sym_TILDE] = ACTIONS(1551),
    [anon_sym_QMARK] = ACTIONS(1551),
    [anon_sym_LPAREN] = ACTIONS(1553),
    [anon_sym_RPAREN] = ACTIONS(1553),
    [anon_sym_COMMA] = ACTIONS(1553),
    [anon_sym_PIPE] = ACTIONS(1551),
    [anon_sym_COLON2] = ACTIONS(1551),
    [anon_sym_DOT] = ACTIONS(1553),
    [anon_sym_DASH_GT] = ACTIONS(1551),
    [anon_sym_LBRACK] = ACTIONS(1551),
    [anon_sym_RBRACK] = ACTIONS(1553),
    [anon_sym_true] = ACTIONS(1551),
    [anon_sym_false] = ACTIONS(1551),
    [anon_sym_LBRACE] = ACTIONS(1551),
    [anon_sym_SEMI] = ACTIONS(1553),
    [anon_sym_RBRACE] = ACTIONS(1553),
    [anon_sym_constraint] = ACTIONS(1551),
    [anon_sym_val] = ACTIONS(1551),
    [anon_sym_end] = ACTIONS(1551),
    [anon_sym_with] = ACTIONS(1551),
    [anon_sym_COLON_GT] = ACTIONS(1553),
    [anon_sym_object] = ACTIONS(1551),
    [anon_sym_inherit] = ACTIONS(1551),
    [anon_sym_method] = ACTIONS(1551),
    [anon_sym_in] = ACTIONS(1551),
    [anon_sym_as] = ACTIONS(1551),
    [anon_sym_initializer] = ACTIONS(1551),
    [anon_sym_AMP] = ACTIONS(1551),
    [anon_sym_POUND] = ACTIONS(1551),
    [anon_sym_COLON_COLON] = ACTIONS(1553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1553),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1553),
    [anon_sym_LT_DASH] = ACTIONS(1551),
    [anon_sym_then] = ACTIONS(1551),
    [anon_sym_else] = ACTIONS(1551),
    [anon_sym_do] = ACTIONS(1551),
    [anon_sym_done] = ACTIONS(1551),
    [anon_sym_to] = ACTIONS(1551),
    [anon_sym_downto] = ACTIONS(1551),
    [anon_sym_new] = ACTIONS(1551),
    [anon_sym_LBRACE_LT] = ACTIONS(1553),
    [anon_sym_GT_RBRACE] = ACTIONS(1553),
    [anon_sym_begin] = ACTIONS(1551),
    [sym_ocamlyacc_value] = ACTIONS(1553),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1551),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1553),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1551),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1551),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1553),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1553),
    [aux_sym_number_token1] = ACTIONS(1553),
    [aux_sym_signed_number_token1] = ACTIONS(1551),
    [anon_sym_SQUOTE] = ACTIONS(1553),
    [anon_sym_DQUOTE] = ACTIONS(1553),
    [sym_prefix_operator] = ACTIONS(1553),
    [aux_sym_sign_operator_token1] = ACTIONS(1551),
    [sym_hash_operator] = ACTIONS(1553),
    [anon_sym_lsl] = ACTIONS(1551),
    [anon_sym_lsr] = ACTIONS(1551),
    [anon_sym_asr] = ACTIONS(1551),
    [aux_sym_pow_operator_token1] = ACTIONS(1553),
    [anon_sym_mod] = ACTIONS(1551),
    [anon_sym_land] = ACTIONS(1551),
    [anon_sym_lor] = ACTIONS(1551),
    [anon_sym_lxor] = ACTIONS(1551),
    [aux_sym_mult_operator_token1] = ACTIONS(1551),
    [aux_sym_add_operator_token1] = ACTIONS(1551),
    [sym_concat_operator] = ACTIONS(1553),
    [sym_rel_operator] = ACTIONS(1551),
    [anon_sym_AMP_AMP] = ACTIONS(1551),
    [anon_sym_or] = ACTIONS(1551),
    [anon_sym_PIPE_PIPE] = ACTIONS(1551),
    [sym_assign_operator] = ACTIONS(1553),
    [sym__uppercase_identifier] = ACTIONS(1553),
    [aux_sym_tag_token1] = ACTIONS(1553),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(757)] = {
    [sym_attribute] = STATE(757),
    [sym__lowercase_identifier] = ACTIONS(1555),
    [anon_sym_and] = ACTIONS(1555),
    [anon_sym_TILDE] = ACTIONS(1555),
    [anon_sym_QMARK] = ACTIONS(1555),
    [anon_sym_LPAREN] = ACTIONS(1557),
    [anon_sym_RPAREN] = ACTIONS(1557),
    [anon_sym_COMMA] = ACTIONS(1557),
    [anon_sym_PIPE] = ACTIONS(1555),
    [anon_sym_COLON2] = ACTIONS(1555),
    [anon_sym_DOT] = ACTIONS(1557),
    [anon_sym_DASH_GT] = ACTIONS(1555),
    [anon_sym_LBRACK] = ACTIONS(1555),
    [anon_sym_RBRACK] = ACTIONS(1557),
    [anon_sym_true] = ACTIONS(1555),
    [anon_sym_false] = ACTIONS(1555),
    [anon_sym_LBRACE] = ACTIONS(1555),
    [anon_sym_SEMI] = ACTIONS(1557),
    [anon_sym_RBRACE] = ACTIONS(1557),
    [anon_sym_constraint] = ACTIONS(1555),
    [anon_sym_val] = ACTIONS(1555),
    [anon_sym_end] = ACTIONS(1555),
    [anon_sym_with] = ACTIONS(1555),
    [anon_sym_COLON_GT] = ACTIONS(1557),
    [anon_sym_object] = ACTIONS(1555),
    [anon_sym_inherit] = ACTIONS(1555),
    [anon_sym_method] = ACTIONS(1555),
    [anon_sym_in] = ACTIONS(1555),
    [anon_sym_as] = ACTIONS(1555),
    [anon_sym_initializer] = ACTIONS(1555),
    [anon_sym_AMP] = ACTIONS(1555),
    [anon_sym_POUND] = ACTIONS(1555),
    [anon_sym_COLON_COLON] = ACTIONS(1557),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1557),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1557),
    [anon_sym_LT_DASH] = ACTIONS(1555),
    [anon_sym_then] = ACTIONS(1555),
    [anon_sym_else] = ACTIONS(1555),
    [anon_sym_do] = ACTIONS(1555),
    [anon_sym_done] = ACTIONS(1555),
    [anon_sym_to] = ACTIONS(1555),
    [anon_sym_downto] = ACTIONS(1555),
    [anon_sym_new] = ACTIONS(1555),
    [anon_sym_LBRACE_LT] = ACTIONS(1557),
    [anon_sym_GT_RBRACE] = ACTIONS(1557),
    [anon_sym_begin] = ACTIONS(1555),
    [sym_ocamlyacc_value] = ACTIONS(1557),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1555),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1557),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1555),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1555),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1557),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1557),
    [aux_sym_number_token1] = ACTIONS(1557),
    [aux_sym_signed_number_token1] = ACTIONS(1555),
    [anon_sym_SQUOTE] = ACTIONS(1557),
    [anon_sym_DQUOTE] = ACTIONS(1557),
    [sym_prefix_operator] = ACTIONS(1557),
    [aux_sym_sign_operator_token1] = ACTIONS(1555),
    [sym_hash_operator] = ACTIONS(1557),
    [anon_sym_lsl] = ACTIONS(1555),
    [anon_sym_lsr] = ACTIONS(1555),
    [anon_sym_asr] = ACTIONS(1555),
    [aux_sym_pow_operator_token1] = ACTIONS(1557),
    [anon_sym_mod] = ACTIONS(1555),
    [anon_sym_land] = ACTIONS(1555),
    [anon_sym_lor] = ACTIONS(1555),
    [anon_sym_lxor] = ACTIONS(1555),
    [aux_sym_mult_operator_token1] = ACTIONS(1555),
    [aux_sym_add_operator_token1] = ACTIONS(1555),
    [sym_concat_operator] = ACTIONS(1557),
    [sym_rel_operator] = ACTIONS(1555),
    [anon_sym_AMP_AMP] = ACTIONS(1555),
    [anon_sym_or] = ACTIONS(1555),
    [anon_sym_PIPE_PIPE] = ACTIONS(1555),
    [sym_assign_operator] = ACTIONS(1557),
    [sym__uppercase_identifier] = ACTIONS(1557),
    [aux_sym_tag_token1] = ACTIONS(1557),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(758)] = {
    [sym_attribute] = STATE(758),
    [sym__lowercase_identifier] = ACTIONS(1559),
    [anon_sym_and] = ACTIONS(1559),
    [anon_sym_TILDE] = ACTIONS(1559),
    [anon_sym_QMARK] = ACTIONS(1559),
    [anon_sym_LPAREN] = ACTIONS(1561),
    [anon_sym_RPAREN] = ACTIONS(1561),
    [anon_sym_COMMA] = ACTIONS(1561),
    [anon_sym_PIPE] = ACTIONS(1559),
    [anon_sym_COLON2] = ACTIONS(1559),
    [anon_sym_DOT] = ACTIONS(1561),
    [anon_sym_DASH_GT] = ACTIONS(1559),
    [anon_sym_LBRACK] = ACTIONS(1559),
    [anon_sym_RBRACK] = ACTIONS(1561),
    [anon_sym_true] = ACTIONS(1559),
    [anon_sym_false] = ACTIONS(1559),
    [anon_sym_LBRACE] = ACTIONS(1559),
    [anon_sym_SEMI] = ACTIONS(1561),
    [anon_sym_RBRACE] = ACTIONS(1561),
    [anon_sym_constraint] = ACTIONS(1559),
    [anon_sym_val] = ACTIONS(1559),
    [anon_sym_end] = ACTIONS(1559),
    [anon_sym_with] = ACTIONS(1559),
    [anon_sym_COLON_GT] = ACTIONS(1561),
    [anon_sym_object] = ACTIONS(1559),
    [anon_sym_inherit] = ACTIONS(1559),
    [anon_sym_method] = ACTIONS(1559),
    [anon_sym_in] = ACTIONS(1559),
    [anon_sym_as] = ACTIONS(1559),
    [anon_sym_initializer] = ACTIONS(1559),
    [anon_sym_AMP] = ACTIONS(1559),
    [anon_sym_POUND] = ACTIONS(1559),
    [anon_sym_COLON_COLON] = ACTIONS(1561),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1561),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1561),
    [anon_sym_LT_DASH] = ACTIONS(1559),
    [anon_sym_then] = ACTIONS(1559),
    [anon_sym_else] = ACTIONS(1559),
    [anon_sym_do] = ACTIONS(1559),
    [anon_sym_done] = ACTIONS(1559),
    [anon_sym_to] = ACTIONS(1559),
    [anon_sym_downto] = ACTIONS(1559),
    [anon_sym_new] = ACTIONS(1559),
    [anon_sym_LBRACE_LT] = ACTIONS(1561),
    [anon_sym_GT_RBRACE] = ACTIONS(1561),
    [anon_sym_begin] = ACTIONS(1559),
    [sym_ocamlyacc_value] = ACTIONS(1561),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1559),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1561),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1559),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1559),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1561),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1561),
    [aux_sym_number_token1] = ACTIONS(1561),
    [aux_sym_signed_number_token1] = ACTIONS(1559),
    [anon_sym_SQUOTE] = ACTIONS(1561),
    [anon_sym_DQUOTE] = ACTIONS(1561),
    [sym_prefix_operator] = ACTIONS(1561),
    [aux_sym_sign_operator_token1] = ACTIONS(1559),
    [sym_hash_operator] = ACTIONS(1561),
    [anon_sym_lsl] = ACTIONS(1559),
    [anon_sym_lsr] = ACTIONS(1559),
    [anon_sym_asr] = ACTIONS(1559),
    [aux_sym_pow_operator_token1] = ACTIONS(1561),
    [anon_sym_mod] = ACTIONS(1559),
    [anon_sym_land] = ACTIONS(1559),
    [anon_sym_lor] = ACTIONS(1559),
    [anon_sym_lxor] = ACTIONS(1559),
    [aux_sym_mult_operator_token1] = ACTIONS(1559),
    [aux_sym_add_operator_token1] = ACTIONS(1559),
    [sym_concat_operator] = ACTIONS(1561),
    [sym_rel_operator] = ACTIONS(1559),
    [anon_sym_AMP_AMP] = ACTIONS(1559),
    [anon_sym_or] = ACTIONS(1559),
    [anon_sym_PIPE_PIPE] = ACTIONS(1559),
    [sym_assign_operator] = ACTIONS(1561),
    [sym__uppercase_identifier] = ACTIONS(1561),
    [aux_sym_tag_token1] = ACTIONS(1561),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(759)] = {
    [sym_attribute] = STATE(759),
    [sym__lowercase_identifier] = ACTIONS(1563),
    [anon_sym_and] = ACTIONS(1563),
    [anon_sym_TILDE] = ACTIONS(1563),
    [anon_sym_QMARK] = ACTIONS(1563),
    [anon_sym_LPAREN] = ACTIONS(1565),
    [anon_sym_RPAREN] = ACTIONS(1565),
    [anon_sym_COMMA] = ACTIONS(1565),
    [anon_sym_PIPE] = ACTIONS(1563),
    [anon_sym_COLON2] = ACTIONS(1563),
    [anon_sym_DOT] = ACTIONS(1565),
    [anon_sym_DASH_GT] = ACTIONS(1563),
    [anon_sym_LBRACK] = ACTIONS(1563),
    [anon_sym_RBRACK] = ACTIONS(1565),
    [anon_sym_true] = ACTIONS(1563),
    [anon_sym_false] = ACTIONS(1563),
    [anon_sym_LBRACE] = ACTIONS(1563),
    [anon_sym_SEMI] = ACTIONS(1565),
    [anon_sym_RBRACE] = ACTIONS(1565),
    [anon_sym_constraint] = ACTIONS(1563),
    [anon_sym_val] = ACTIONS(1563),
    [anon_sym_end] = ACTIONS(1563),
    [anon_sym_with] = ACTIONS(1563),
    [anon_sym_COLON_GT] = ACTIONS(1565),
    [anon_sym_object] = ACTIONS(1563),
    [anon_sym_inherit] = ACTIONS(1563),
    [anon_sym_method] = ACTIONS(1563),
    [anon_sym_in] = ACTIONS(1563),
    [anon_sym_as] = ACTIONS(1563),
    [anon_sym_initializer] = ACTIONS(1563),
    [anon_sym_AMP] = ACTIONS(1563),
    [anon_sym_POUND] = ACTIONS(1563),
    [anon_sym_COLON_COLON] = ACTIONS(1565),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1565),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1565),
    [anon_sym_LT_DASH] = ACTIONS(1563),
    [anon_sym_then] = ACTIONS(1563),
    [anon_sym_else] = ACTIONS(1563),
    [anon_sym_do] = ACTIONS(1563),
    [anon_sym_done] = ACTIONS(1563),
    [anon_sym_to] = ACTIONS(1563),
    [anon_sym_downto] = ACTIONS(1563),
    [anon_sym_new] = ACTIONS(1563),
    [anon_sym_LBRACE_LT] = ACTIONS(1565),
    [anon_sym_GT_RBRACE] = ACTIONS(1565),
    [anon_sym_begin] = ACTIONS(1563),
    [sym_ocamlyacc_value] = ACTIONS(1565),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1563),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1565),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1563),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1563),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1565),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1565),
    [aux_sym_number_token1] = ACTIONS(1565),
    [aux_sym_signed_number_token1] = ACTIONS(1563),
    [anon_sym_SQUOTE] = ACTIONS(1565),
    [anon_sym_DQUOTE] = ACTIONS(1565),
    [sym_prefix_operator] = ACTIONS(1565),
    [aux_sym_sign_operator_token1] = ACTIONS(1563),
    [sym_hash_operator] = ACTIONS(1565),
    [anon_sym_lsl] = ACTIONS(1563),
    [anon_sym_lsr] = ACTIONS(1563),
    [anon_sym_asr] = ACTIONS(1563),
    [aux_sym_pow_operator_token1] = ACTIONS(1565),
    [anon_sym_mod] = ACTIONS(1563),
    [anon_sym_land] = ACTIONS(1563),
    [anon_sym_lor] = ACTIONS(1563),
    [anon_sym_lxor] = ACTIONS(1563),
    [aux_sym_mult_operator_token1] = ACTIONS(1563),
    [aux_sym_add_operator_token1] = ACTIONS(1563),
    [sym_concat_operator] = ACTIONS(1565),
    [sym_rel_operator] = ACTIONS(1563),
    [anon_sym_AMP_AMP] = ACTIONS(1563),
    [anon_sym_or] = ACTIONS(1563),
    [anon_sym_PIPE_PIPE] = ACTIONS(1563),
    [sym_assign_operator] = ACTIONS(1565),
    [sym__uppercase_identifier] = ACTIONS(1565),
    [aux_sym_tag_token1] = ACTIONS(1565),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(760)] = {
    [sym_attribute] = STATE(760),
    [sym__lowercase_identifier] = ACTIONS(1567),
    [anon_sym_and] = ACTIONS(1567),
    [anon_sym_TILDE] = ACTIONS(1567),
    [anon_sym_QMARK] = ACTIONS(1567),
    [anon_sym_LPAREN] = ACTIONS(1569),
    [anon_sym_RPAREN] = ACTIONS(1569),
    [anon_sym_COMMA] = ACTIONS(1569),
    [anon_sym_PIPE] = ACTIONS(1567),
    [anon_sym_COLON2] = ACTIONS(1567),
    [anon_sym_DOT] = ACTIONS(1569),
    [anon_sym_DASH_GT] = ACTIONS(1567),
    [anon_sym_LBRACK] = ACTIONS(1567),
    [anon_sym_RBRACK] = ACTIONS(1569),
    [anon_sym_true] = ACTIONS(1567),
    [anon_sym_false] = ACTIONS(1567),
    [anon_sym_LBRACE] = ACTIONS(1567),
    [anon_sym_SEMI] = ACTIONS(1569),
    [anon_sym_RBRACE] = ACTIONS(1569),
    [anon_sym_constraint] = ACTIONS(1567),
    [anon_sym_val] = ACTIONS(1567),
    [anon_sym_end] = ACTIONS(1567),
    [anon_sym_with] = ACTIONS(1567),
    [anon_sym_COLON_GT] = ACTIONS(1569),
    [anon_sym_object] = ACTIONS(1567),
    [anon_sym_inherit] = ACTIONS(1567),
    [anon_sym_method] = ACTIONS(1567),
    [anon_sym_in] = ACTIONS(1567),
    [anon_sym_as] = ACTIONS(1567),
    [anon_sym_initializer] = ACTIONS(1567),
    [anon_sym_AMP] = ACTIONS(1567),
    [anon_sym_POUND] = ACTIONS(1567),
    [anon_sym_COLON_COLON] = ACTIONS(1569),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1569),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1569),
    [anon_sym_then] = ACTIONS(1567),
    [anon_sym_else] = ACTIONS(1567),
    [anon_sym_do] = ACTIONS(1567),
    [anon_sym_done] = ACTIONS(1567),
    [anon_sym_to] = ACTIONS(1567),
    [anon_sym_downto] = ACTIONS(1567),
    [anon_sym_new] = ACTIONS(1567),
    [anon_sym_LBRACE_LT] = ACTIONS(1569),
    [anon_sym_GT_RBRACE] = ACTIONS(1569),
    [anon_sym_begin] = ACTIONS(1567),
    [sym_ocamlyacc_value] = ACTIONS(1569),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1567),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1569),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1567),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1567),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1569),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1569),
    [aux_sym_number_token1] = ACTIONS(1569),
    [aux_sym_signed_number_token1] = ACTIONS(1567),
    [anon_sym_SQUOTE] = ACTIONS(1569),
    [anon_sym_DQUOTE] = ACTIONS(1569),
    [sym_prefix_operator] = ACTIONS(1569),
    [aux_sym_sign_operator_token1] = ACTIONS(1567),
    [sym_hash_operator] = ACTIONS(1569),
    [anon_sym_lsl] = ACTIONS(1567),
    [anon_sym_lsr] = ACTIONS(1567),
    [anon_sym_asr] = ACTIONS(1567),
    [aux_sym_pow_operator_token1] = ACTIONS(1569),
    [anon_sym_mod] = ACTIONS(1567),
    [anon_sym_land] = ACTIONS(1567),
    [anon_sym_lor] = ACTIONS(1567),
    [anon_sym_lxor] = ACTIONS(1567),
    [aux_sym_mult_operator_token1] = ACTIONS(1567),
    [aux_sym_add_operator_token1] = ACTIONS(1567),
    [sym_concat_operator] = ACTIONS(1569),
    [sym_rel_operator] = ACTIONS(1567),
    [anon_sym_AMP_AMP] = ACTIONS(1567),
    [anon_sym_or] = ACTIONS(1567),
    [anon_sym_PIPE_PIPE] = ACTIONS(1567),
    [sym_assign_operator] = ACTIONS(1569),
    [sym__uppercase_identifier] = ACTIONS(1569),
    [aux_sym_tag_token1] = ACTIONS(1569),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(761)] = {
    [sym_attribute] = STATE(761),
    [sym__lowercase_identifier] = ACTIONS(1571),
    [anon_sym_and] = ACTIONS(1571),
    [anon_sym_TILDE] = ACTIONS(1571),
    [anon_sym_QMARK] = ACTIONS(1571),
    [anon_sym_LPAREN] = ACTIONS(1573),
    [anon_sym_RPAREN] = ACTIONS(1573),
    [anon_sym_COMMA] = ACTIONS(1573),
    [anon_sym_PIPE] = ACTIONS(1571),
    [anon_sym_COLON2] = ACTIONS(1571),
    [anon_sym_DOT] = ACTIONS(1573),
    [anon_sym_DASH_GT] = ACTIONS(1571),
    [anon_sym_LBRACK] = ACTIONS(1571),
    [anon_sym_RBRACK] = ACTIONS(1573),
    [anon_sym_true] = ACTIONS(1571),
    [anon_sym_false] = ACTIONS(1571),
    [anon_sym_LBRACE] = ACTIONS(1571),
    [anon_sym_SEMI] = ACTIONS(1573),
    [anon_sym_RBRACE] = ACTIONS(1573),
    [anon_sym_constraint] = ACTIONS(1571),
    [anon_sym_val] = ACTIONS(1571),
    [anon_sym_end] = ACTIONS(1571),
    [anon_sym_with] = ACTIONS(1571),
    [anon_sym_COLON_GT] = ACTIONS(1573),
    [anon_sym_object] = ACTIONS(1571),
    [anon_sym_inherit] = ACTIONS(1571),
    [anon_sym_method] = ACTIONS(1571),
    [anon_sym_in] = ACTIONS(1571),
    [anon_sym_as] = ACTIONS(1571),
    [anon_sym_initializer] = ACTIONS(1571),
    [anon_sym_AMP] = ACTIONS(1571),
    [anon_sym_POUND] = ACTIONS(1571),
    [anon_sym_COLON_COLON] = ACTIONS(1573),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1573),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1573),
    [anon_sym_then] = ACTIONS(1571),
    [anon_sym_else] = ACTIONS(1571),
    [anon_sym_do] = ACTIONS(1571),
    [anon_sym_done] = ACTIONS(1571),
    [anon_sym_to] = ACTIONS(1571),
    [anon_sym_downto] = ACTIONS(1571),
    [anon_sym_new] = ACTIONS(1571),
    [anon_sym_LBRACE_LT] = ACTIONS(1573),
    [anon_sym_GT_RBRACE] = ACTIONS(1573),
    [anon_sym_begin] = ACTIONS(1571),
    [sym_ocamlyacc_value] = ACTIONS(1573),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1571),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1573),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1571),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1571),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1573),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1573),
    [aux_sym_number_token1] = ACTIONS(1573),
    [aux_sym_signed_number_token1] = ACTIONS(1571),
    [anon_sym_SQUOTE] = ACTIONS(1573),
    [anon_sym_DQUOTE] = ACTIONS(1573),
    [sym_prefix_operator] = ACTIONS(1573),
    [aux_sym_sign_operator_token1] = ACTIONS(1571),
    [sym_hash_operator] = ACTIONS(1573),
    [anon_sym_lsl] = ACTIONS(1571),
    [anon_sym_lsr] = ACTIONS(1571),
    [anon_sym_asr] = ACTIONS(1571),
    [aux_sym_pow_operator_token1] = ACTIONS(1573),
    [anon_sym_mod] = ACTIONS(1571),
    [anon_sym_land] = ACTIONS(1571),
    [anon_sym_lor] = ACTIONS(1571),
    [anon_sym_lxor] = ACTIONS(1571),
    [aux_sym_mult_operator_token1] = ACTIONS(1571),
    [aux_sym_add_operator_token1] = ACTIONS(1571),
    [sym_concat_operator] = ACTIONS(1573),
    [sym_rel_operator] = ACTIONS(1571),
    [anon_sym_AMP_AMP] = ACTIONS(1571),
    [anon_sym_or] = ACTIONS(1571),
    [anon_sym_PIPE_PIPE] = ACTIONS(1571),
    [sym_assign_operator] = ACTIONS(1573),
    [sym__uppercase_identifier] = ACTIONS(1573),
    [aux_sym_tag_token1] = ACTIONS(1573),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(762)] = {
    [sym_attribute] = STATE(762),
    [sym__lowercase_identifier] = ACTIONS(1575),
    [anon_sym_and] = ACTIONS(1575),
    [anon_sym_TILDE] = ACTIONS(1575),
    [anon_sym_QMARK] = ACTIONS(1575),
    [anon_sym_LPAREN] = ACTIONS(1577),
    [anon_sym_RPAREN] = ACTIONS(1577),
    [anon_sym_COMMA] = ACTIONS(1577),
    [anon_sym_PIPE] = ACTIONS(1575),
    [anon_sym_COLON2] = ACTIONS(1575),
    [anon_sym_DOT] = ACTIONS(1577),
    [anon_sym_DASH_GT] = ACTIONS(1575),
    [anon_sym_LBRACK] = ACTIONS(1575),
    [anon_sym_RBRACK] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(1575),
    [anon_sym_false] = ACTIONS(1575),
    [anon_sym_LBRACE] = ACTIONS(1575),
    [anon_sym_SEMI] = ACTIONS(1577),
    [anon_sym_RBRACE] = ACTIONS(1577),
    [anon_sym_constraint] = ACTIONS(1575),
    [anon_sym_val] = ACTIONS(1575),
    [anon_sym_end] = ACTIONS(1575),
    [anon_sym_with] = ACTIONS(1575),
    [anon_sym_COLON_GT] = ACTIONS(1577),
    [anon_sym_object] = ACTIONS(1575),
    [anon_sym_inherit] = ACTIONS(1575),
    [anon_sym_method] = ACTIONS(1575),
    [anon_sym_in] = ACTIONS(1575),
    [anon_sym_as] = ACTIONS(1575),
    [anon_sym_initializer] = ACTIONS(1575),
    [anon_sym_AMP] = ACTIONS(1575),
    [anon_sym_POUND] = ACTIONS(1575),
    [anon_sym_COLON_COLON] = ACTIONS(1577),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1577),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1577),
    [anon_sym_then] = ACTIONS(1575),
    [anon_sym_else] = ACTIONS(1575),
    [anon_sym_do] = ACTIONS(1575),
    [anon_sym_done] = ACTIONS(1575),
    [anon_sym_to] = ACTIONS(1575),
    [anon_sym_downto] = ACTIONS(1575),
    [anon_sym_new] = ACTIONS(1575),
    [anon_sym_LBRACE_LT] = ACTIONS(1577),
    [anon_sym_GT_RBRACE] = ACTIONS(1577),
    [anon_sym_begin] = ACTIONS(1575),
    [sym_ocamlyacc_value] = ACTIONS(1577),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1575),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1577),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1575),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1575),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1577),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1577),
    [aux_sym_number_token1] = ACTIONS(1577),
    [aux_sym_signed_number_token1] = ACTIONS(1575),
    [anon_sym_SQUOTE] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(1577),
    [sym_prefix_operator] = ACTIONS(1577),
    [aux_sym_sign_operator_token1] = ACTIONS(1575),
    [sym_hash_operator] = ACTIONS(1577),
    [anon_sym_lsl] = ACTIONS(1575),
    [anon_sym_lsr] = ACTIONS(1575),
    [anon_sym_asr] = ACTIONS(1575),
    [aux_sym_pow_operator_token1] = ACTIONS(1577),
    [anon_sym_mod] = ACTIONS(1575),
    [anon_sym_land] = ACTIONS(1575),
    [anon_sym_lor] = ACTIONS(1575),
    [anon_sym_lxor] = ACTIONS(1575),
    [aux_sym_mult_operator_token1] = ACTIONS(1575),
    [aux_sym_add_operator_token1] = ACTIONS(1575),
    [sym_concat_operator] = ACTIONS(1577),
    [sym_rel_operator] = ACTIONS(1575),
    [anon_sym_AMP_AMP] = ACTIONS(1575),
    [anon_sym_or] = ACTIONS(1575),
    [anon_sym_PIPE_PIPE] = ACTIONS(1575),
    [sym_assign_operator] = ACTIONS(1577),
    [sym__uppercase_identifier] = ACTIONS(1577),
    [aux_sym_tag_token1] = ACTIONS(1577),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(763)] = {
    [sym_attribute] = STATE(763),
    [sym__lowercase_identifier] = ACTIONS(1579),
    [anon_sym_and] = ACTIONS(1579),
    [anon_sym_TILDE] = ACTIONS(1579),
    [anon_sym_QMARK] = ACTIONS(1579),
    [anon_sym_LPAREN] = ACTIONS(1581),
    [anon_sym_RPAREN] = ACTIONS(1581),
    [anon_sym_COMMA] = ACTIONS(1581),
    [anon_sym_PIPE] = ACTIONS(1579),
    [anon_sym_COLON2] = ACTIONS(1579),
    [anon_sym_DOT] = ACTIONS(1581),
    [anon_sym_DASH_GT] = ACTIONS(1579),
    [anon_sym_LBRACK] = ACTIONS(1579),
    [anon_sym_RBRACK] = ACTIONS(1581),
    [anon_sym_true] = ACTIONS(1579),
    [anon_sym_false] = ACTIONS(1579),
    [anon_sym_LBRACE] = ACTIONS(1579),
    [anon_sym_SEMI] = ACTIONS(1581),
    [anon_sym_RBRACE] = ACTIONS(1581),
    [anon_sym_constraint] = ACTIONS(1579),
    [anon_sym_val] = ACTIONS(1579),
    [anon_sym_end] = ACTIONS(1579),
    [anon_sym_with] = ACTIONS(1579),
    [anon_sym_COLON_GT] = ACTIONS(1581),
    [anon_sym_object] = ACTIONS(1579),
    [anon_sym_inherit] = ACTIONS(1579),
    [anon_sym_method] = ACTIONS(1579),
    [anon_sym_in] = ACTIONS(1579),
    [anon_sym_as] = ACTIONS(1579),
    [anon_sym_initializer] = ACTIONS(1579),
    [anon_sym_AMP] = ACTIONS(1579),
    [anon_sym_POUND] = ACTIONS(1579),
    [anon_sym_COLON_COLON] = ACTIONS(1581),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1581),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1581),
    [anon_sym_then] = ACTIONS(1579),
    [anon_sym_else] = ACTIONS(1579),
    [anon_sym_do] = ACTIONS(1579),
    [anon_sym_done] = ACTIONS(1579),
    [anon_sym_to] = ACTIONS(1579),
    [anon_sym_downto] = ACTIONS(1579),
    [anon_sym_new] = ACTIONS(1579),
    [anon_sym_LBRACE_LT] = ACTIONS(1581),
    [anon_sym_GT_RBRACE] = ACTIONS(1581),
    [anon_sym_begin] = ACTIONS(1579),
    [sym_ocamlyacc_value] = ACTIONS(1581),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1579),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1581),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1579),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1579),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1581),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1581),
    [aux_sym_number_token1] = ACTIONS(1581),
    [aux_sym_signed_number_token1] = ACTIONS(1579),
    [anon_sym_SQUOTE] = ACTIONS(1581),
    [anon_sym_DQUOTE] = ACTIONS(1581),
    [sym_prefix_operator] = ACTIONS(1581),
    [aux_sym_sign_operator_token1] = ACTIONS(1579),
    [sym_hash_operator] = ACTIONS(1581),
    [anon_sym_lsl] = ACTIONS(1579),
    [anon_sym_lsr] = ACTIONS(1579),
    [anon_sym_asr] = ACTIONS(1579),
    [aux_sym_pow_operator_token1] = ACTIONS(1581),
    [anon_sym_mod] = ACTIONS(1579),
    [anon_sym_land] = ACTIONS(1579),
    [anon_sym_lor] = ACTIONS(1579),
    [anon_sym_lxor] = ACTIONS(1579),
    [aux_sym_mult_operator_token1] = ACTIONS(1579),
    [aux_sym_add_operator_token1] = ACTIONS(1579),
    [sym_concat_operator] = ACTIONS(1581),
    [sym_rel_operator] = ACTIONS(1579),
    [anon_sym_AMP_AMP] = ACTIONS(1579),
    [anon_sym_or] = ACTIONS(1579),
    [anon_sym_PIPE_PIPE] = ACTIONS(1579),
    [sym_assign_operator] = ACTIONS(1581),
    [sym__uppercase_identifier] = ACTIONS(1581),
    [aux_sym_tag_token1] = ACTIONS(1581),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(764)] = {
    [sym_attribute] = STATE(764),
    [sym__lowercase_identifier] = ACTIONS(1583),
    [anon_sym_and] = ACTIONS(1583),
    [anon_sym_TILDE] = ACTIONS(1583),
    [anon_sym_QMARK] = ACTIONS(1583),
    [anon_sym_LPAREN] = ACTIONS(1585),
    [anon_sym_RPAREN] = ACTIONS(1585),
    [anon_sym_COMMA] = ACTIONS(1585),
    [anon_sym_PIPE] = ACTIONS(1583),
    [anon_sym_COLON2] = ACTIONS(1583),
    [anon_sym_DOT] = ACTIONS(751),
    [anon_sym_DASH_GT] = ACTIONS(1583),
    [anon_sym_LBRACK] = ACTIONS(1583),
    [anon_sym_RBRACK] = ACTIONS(1585),
    [anon_sym_true] = ACTIONS(1583),
    [anon_sym_false] = ACTIONS(1583),
    [anon_sym_LBRACE] = ACTIONS(1583),
    [anon_sym_SEMI] = ACTIONS(1585),
    [anon_sym_RBRACE] = ACTIONS(1585),
    [anon_sym_constraint] = ACTIONS(1583),
    [anon_sym_val] = ACTIONS(1583),
    [anon_sym_end] = ACTIONS(1583),
    [anon_sym_with] = ACTIONS(1583),
    [anon_sym_COLON_GT] = ACTIONS(1585),
    [anon_sym_object] = ACTIONS(1583),
    [anon_sym_inherit] = ACTIONS(1583),
    [anon_sym_method] = ACTIONS(1583),
    [anon_sym_in] = ACTIONS(1583),
    [anon_sym_as] = ACTIONS(1583),
    [anon_sym_initializer] = ACTIONS(1583),
    [anon_sym_AMP] = ACTIONS(1583),
    [anon_sym_POUND] = ACTIONS(1583),
    [anon_sym_COLON_COLON] = ACTIONS(1585),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1585),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1585),
    [anon_sym_then] = ACTIONS(1583),
    [anon_sym_else] = ACTIONS(1583),
    [anon_sym_do] = ACTIONS(1583),
    [anon_sym_done] = ACTIONS(1583),
    [anon_sym_to] = ACTIONS(1583),
    [anon_sym_downto] = ACTIONS(1583),
    [anon_sym_new] = ACTIONS(1583),
    [anon_sym_LBRACE_LT] = ACTIONS(1585),
    [anon_sym_GT_RBRACE] = ACTIONS(1585),
    [anon_sym_begin] = ACTIONS(1583),
    [sym_ocamlyacc_value] = ACTIONS(1585),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1583),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1585),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1583),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1583),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1585),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1585),
    [aux_sym_number_token1] = ACTIONS(1585),
    [aux_sym_signed_number_token1] = ACTIONS(1583),
    [anon_sym_SQUOTE] = ACTIONS(1585),
    [anon_sym_DQUOTE] = ACTIONS(1585),
    [sym_prefix_operator] = ACTIONS(1585),
    [aux_sym_sign_operator_token1] = ACTIONS(1583),
    [sym_hash_operator] = ACTIONS(1585),
    [anon_sym_lsl] = ACTIONS(1583),
    [anon_sym_lsr] = ACTIONS(1583),
    [anon_sym_asr] = ACTIONS(1583),
    [aux_sym_pow_operator_token1] = ACTIONS(1585),
    [anon_sym_mod] = ACTIONS(1583),
    [anon_sym_land] = ACTIONS(1583),
    [anon_sym_lor] = ACTIONS(1583),
    [anon_sym_lxor] = ACTIONS(1583),
    [aux_sym_mult_operator_token1] = ACTIONS(1583),
    [aux_sym_add_operator_token1] = ACTIONS(1583),
    [sym_concat_operator] = ACTIONS(1585),
    [sym_rel_operator] = ACTIONS(1583),
    [anon_sym_AMP_AMP] = ACTIONS(1583),
    [anon_sym_or] = ACTIONS(1583),
    [anon_sym_PIPE_PIPE] = ACTIONS(1583),
    [sym_assign_operator] = ACTIONS(1585),
    [sym__uppercase_identifier] = ACTIONS(1585),
    [aux_sym_tag_token1] = ACTIONS(1585),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(765)] = {
    [sym_attribute] = STATE(765),
    [sym__lowercase_identifier] = ACTIONS(1587),
    [anon_sym_and] = ACTIONS(1587),
    [anon_sym_TILDE] = ACTIONS(1587),
    [anon_sym_QMARK] = ACTIONS(1587),
    [anon_sym_LPAREN] = ACTIONS(1589),
    [anon_sym_RPAREN] = ACTIONS(1589),
    [anon_sym_COMMA] = ACTIONS(1589),
    [anon_sym_PIPE] = ACTIONS(1587),
    [anon_sym_COLON2] = ACTIONS(1587),
    [anon_sym_DOT] = ACTIONS(1589),
    [anon_sym_DASH_GT] = ACTIONS(1587),
    [anon_sym_LBRACK] = ACTIONS(1587),
    [anon_sym_RBRACK] = ACTIONS(1589),
    [anon_sym_true] = ACTIONS(1587),
    [anon_sym_false] = ACTIONS(1587),
    [anon_sym_LBRACE] = ACTIONS(1587),
    [anon_sym_SEMI] = ACTIONS(1589),
    [anon_sym_RBRACE] = ACTIONS(1589),
    [anon_sym_constraint] = ACTIONS(1587),
    [anon_sym_val] = ACTIONS(1587),
    [anon_sym_end] = ACTIONS(1587),
    [anon_sym_with] = ACTIONS(1587),
    [anon_sym_COLON_GT] = ACTIONS(1589),
    [anon_sym_object] = ACTIONS(1587),
    [anon_sym_inherit] = ACTIONS(1587),
    [anon_sym_method] = ACTIONS(1587),
    [anon_sym_in] = ACTIONS(1587),
    [anon_sym_as] = ACTIONS(1587),
    [anon_sym_initializer] = ACTIONS(1587),
    [anon_sym_AMP] = ACTIONS(1587),
    [anon_sym_POUND] = ACTIONS(1587),
    [anon_sym_COLON_COLON] = ACTIONS(1589),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1589),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1589),
    [anon_sym_then] = ACTIONS(1587),
    [anon_sym_else] = ACTIONS(1587),
    [anon_sym_do] = ACTIONS(1587),
    [anon_sym_done] = ACTIONS(1587),
    [anon_sym_to] = ACTIONS(1587),
    [anon_sym_downto] = ACTIONS(1587),
    [anon_sym_new] = ACTIONS(1587),
    [anon_sym_LBRACE_LT] = ACTIONS(1589),
    [anon_sym_GT_RBRACE] = ACTIONS(1589),
    [anon_sym_begin] = ACTIONS(1587),
    [sym_ocamlyacc_value] = ACTIONS(1589),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1587),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1589),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1587),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1587),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1589),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1589),
    [aux_sym_number_token1] = ACTIONS(1589),
    [aux_sym_signed_number_token1] = ACTIONS(1587),
    [anon_sym_SQUOTE] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(1589),
    [sym_prefix_operator] = ACTIONS(1589),
    [aux_sym_sign_operator_token1] = ACTIONS(1587),
    [sym_hash_operator] = ACTIONS(1589),
    [anon_sym_lsl] = ACTIONS(1587),
    [anon_sym_lsr] = ACTIONS(1587),
    [anon_sym_asr] = ACTIONS(1587),
    [aux_sym_pow_operator_token1] = ACTIONS(1589),
    [anon_sym_mod] = ACTIONS(1587),
    [anon_sym_land] = ACTIONS(1587),
    [anon_sym_lor] = ACTIONS(1587),
    [anon_sym_lxor] = ACTIONS(1587),
    [aux_sym_mult_operator_token1] = ACTIONS(1587),
    [aux_sym_add_operator_token1] = ACTIONS(1587),
    [sym_concat_operator] = ACTIONS(1589),
    [sym_rel_operator] = ACTIONS(1587),
    [anon_sym_AMP_AMP] = ACTIONS(1587),
    [anon_sym_or] = ACTIONS(1587),
    [anon_sym_PIPE_PIPE] = ACTIONS(1587),
    [sym_assign_operator] = ACTIONS(1589),
    [sym__uppercase_identifier] = ACTIONS(1589),
    [aux_sym_tag_token1] = ACTIONS(1589),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(766)] = {
    [sym_attribute] = STATE(766),
    [sym__lowercase_identifier] = ACTIONS(1591),
    [anon_sym_and] = ACTIONS(1591),
    [anon_sym_TILDE] = ACTIONS(1591),
    [anon_sym_QMARK] = ACTIONS(1591),
    [anon_sym_LPAREN] = ACTIONS(1593),
    [anon_sym_RPAREN] = ACTIONS(1593),
    [anon_sym_COMMA] = ACTIONS(1593),
    [anon_sym_PIPE] = ACTIONS(1591),
    [anon_sym_COLON2] = ACTIONS(1591),
    [anon_sym_DOT] = ACTIONS(1593),
    [anon_sym_DASH_GT] = ACTIONS(1591),
    [anon_sym_LBRACK] = ACTIONS(1591),
    [anon_sym_RBRACK] = ACTIONS(1593),
    [anon_sym_true] = ACTIONS(1591),
    [anon_sym_false] = ACTIONS(1591),
    [anon_sym_LBRACE] = ACTIONS(1591),
    [anon_sym_SEMI] = ACTIONS(1593),
    [anon_sym_RBRACE] = ACTIONS(1593),
    [anon_sym_constraint] = ACTIONS(1591),
    [anon_sym_val] = ACTIONS(1591),
    [anon_sym_end] = ACTIONS(1591),
    [anon_sym_with] = ACTIONS(1591),
    [anon_sym_COLON_GT] = ACTIONS(1593),
    [anon_sym_object] = ACTIONS(1591),
    [anon_sym_inherit] = ACTIONS(1591),
    [anon_sym_method] = ACTIONS(1591),
    [anon_sym_in] = ACTIONS(1591),
    [anon_sym_as] = ACTIONS(1591),
    [anon_sym_initializer] = ACTIONS(1591),
    [anon_sym_AMP] = ACTIONS(1591),
    [anon_sym_POUND] = ACTIONS(1591),
    [anon_sym_COLON_COLON] = ACTIONS(1593),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1593),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1593),
    [anon_sym_then] = ACTIONS(1591),
    [anon_sym_else] = ACTIONS(1591),
    [anon_sym_do] = ACTIONS(1591),
    [anon_sym_done] = ACTIONS(1591),
    [anon_sym_to] = ACTIONS(1591),
    [anon_sym_downto] = ACTIONS(1591),
    [anon_sym_new] = ACTIONS(1591),
    [anon_sym_LBRACE_LT] = ACTIONS(1593),
    [anon_sym_GT_RBRACE] = ACTIONS(1593),
    [anon_sym_begin] = ACTIONS(1591),
    [sym_ocamlyacc_value] = ACTIONS(1593),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1591),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1593),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1591),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1591),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1593),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1593),
    [aux_sym_number_token1] = ACTIONS(1593),
    [aux_sym_signed_number_token1] = ACTIONS(1591),
    [anon_sym_SQUOTE] = ACTIONS(1593),
    [anon_sym_DQUOTE] = ACTIONS(1593),
    [sym_prefix_operator] = ACTIONS(1593),
    [aux_sym_sign_operator_token1] = ACTIONS(1591),
    [sym_hash_operator] = ACTIONS(1593),
    [anon_sym_lsl] = ACTIONS(1591),
    [anon_sym_lsr] = ACTIONS(1591),
    [anon_sym_asr] = ACTIONS(1591),
    [aux_sym_pow_operator_token1] = ACTIONS(1593),
    [anon_sym_mod] = ACTIONS(1591),
    [anon_sym_land] = ACTIONS(1591),
    [anon_sym_lor] = ACTIONS(1591),
    [anon_sym_lxor] = ACTIONS(1591),
    [aux_sym_mult_operator_token1] = ACTIONS(1591),
    [aux_sym_add_operator_token1] = ACTIONS(1591),
    [sym_concat_operator] = ACTIONS(1593),
    [sym_rel_operator] = ACTIONS(1591),
    [anon_sym_AMP_AMP] = ACTIONS(1591),
    [anon_sym_or] = ACTIONS(1591),
    [anon_sym_PIPE_PIPE] = ACTIONS(1591),
    [sym_assign_operator] = ACTIONS(1593),
    [sym__uppercase_identifier] = ACTIONS(1593),
    [aux_sym_tag_token1] = ACTIONS(1593),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(767)] = {
    [sym_attribute] = STATE(767),
    [sym__lowercase_identifier] = ACTIONS(1595),
    [anon_sym_and] = ACTIONS(1595),
    [anon_sym_TILDE] = ACTIONS(1595),
    [anon_sym_QMARK] = ACTIONS(1595),
    [anon_sym_LPAREN] = ACTIONS(1597),
    [anon_sym_RPAREN] = ACTIONS(1597),
    [anon_sym_COMMA] = ACTIONS(1597),
    [anon_sym_PIPE] = ACTIONS(1595),
    [anon_sym_COLON2] = ACTIONS(1595),
    [anon_sym_DOT] = ACTIONS(1597),
    [anon_sym_DASH_GT] = ACTIONS(1595),
    [anon_sym_LBRACK] = ACTIONS(1595),
    [anon_sym_RBRACK] = ACTIONS(1597),
    [anon_sym_true] = ACTIONS(1595),
    [anon_sym_false] = ACTIONS(1595),
    [anon_sym_LBRACE] = ACTIONS(1595),
    [anon_sym_SEMI] = ACTIONS(1597),
    [anon_sym_RBRACE] = ACTIONS(1597),
    [anon_sym_constraint] = ACTIONS(1595),
    [anon_sym_val] = ACTIONS(1595),
    [anon_sym_end] = ACTIONS(1595),
    [anon_sym_with] = ACTIONS(1595),
    [anon_sym_COLON_GT] = ACTIONS(1597),
    [anon_sym_object] = ACTIONS(1595),
    [anon_sym_inherit] = ACTIONS(1595),
    [anon_sym_method] = ACTIONS(1595),
    [anon_sym_in] = ACTIONS(1595),
    [anon_sym_as] = ACTIONS(1595),
    [anon_sym_initializer] = ACTIONS(1595),
    [anon_sym_AMP] = ACTIONS(1595),
    [anon_sym_POUND] = ACTIONS(1595),
    [anon_sym_COLON_COLON] = ACTIONS(1597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1597),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1597),
    [anon_sym_then] = ACTIONS(1595),
    [anon_sym_else] = ACTIONS(1595),
    [anon_sym_do] = ACTIONS(1595),
    [anon_sym_done] = ACTIONS(1595),
    [anon_sym_to] = ACTIONS(1595),
    [anon_sym_downto] = ACTIONS(1595),
    [anon_sym_new] = ACTIONS(1595),
    [anon_sym_LBRACE_LT] = ACTIONS(1597),
    [anon_sym_GT_RBRACE] = ACTIONS(1597),
    [anon_sym_begin] = ACTIONS(1595),
    [sym_ocamlyacc_value] = ACTIONS(1597),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1595),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1597),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1595),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1595),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1597),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1597),
    [aux_sym_number_token1] = ACTIONS(1597),
    [aux_sym_signed_number_token1] = ACTIONS(1595),
    [anon_sym_SQUOTE] = ACTIONS(1597),
    [anon_sym_DQUOTE] = ACTIONS(1597),
    [sym_prefix_operator] = ACTIONS(1597),
    [aux_sym_sign_operator_token1] = ACTIONS(1595),
    [sym_hash_operator] = ACTIONS(1597),
    [anon_sym_lsl] = ACTIONS(1595),
    [anon_sym_lsr] = ACTIONS(1595),
    [anon_sym_asr] = ACTIONS(1595),
    [aux_sym_pow_operator_token1] = ACTIONS(1597),
    [anon_sym_mod] = ACTIONS(1595),
    [anon_sym_land] = ACTIONS(1595),
    [anon_sym_lor] = ACTIONS(1595),
    [anon_sym_lxor] = ACTIONS(1595),
    [aux_sym_mult_operator_token1] = ACTIONS(1595),
    [aux_sym_add_operator_token1] = ACTIONS(1595),
    [sym_concat_operator] = ACTIONS(1597),
    [sym_rel_operator] = ACTIONS(1595),
    [anon_sym_AMP_AMP] = ACTIONS(1595),
    [anon_sym_or] = ACTIONS(1595),
    [anon_sym_PIPE_PIPE] = ACTIONS(1595),
    [sym_assign_operator] = ACTIONS(1597),
    [sym__uppercase_identifier] = ACTIONS(1597),
    [aux_sym_tag_token1] = ACTIONS(1597),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(768)] = {
    [sym_attribute] = STATE(768),
    [sym__lowercase_identifier] = ACTIONS(1599),
    [anon_sym_and] = ACTIONS(1599),
    [anon_sym_TILDE] = ACTIONS(1599),
    [anon_sym_QMARK] = ACTIONS(1599),
    [anon_sym_LPAREN] = ACTIONS(1601),
    [anon_sym_RPAREN] = ACTIONS(1601),
    [anon_sym_COMMA] = ACTIONS(1601),
    [anon_sym_PIPE] = ACTIONS(1599),
    [anon_sym_COLON2] = ACTIONS(1599),
    [anon_sym_DOT] = ACTIONS(1601),
    [anon_sym_DASH_GT] = ACTIONS(1599),
    [anon_sym_LBRACK] = ACTIONS(1599),
    [anon_sym_RBRACK] = ACTIONS(1601),
    [anon_sym_true] = ACTIONS(1599),
    [anon_sym_false] = ACTIONS(1599),
    [anon_sym_LBRACE] = ACTIONS(1599),
    [anon_sym_SEMI] = ACTIONS(1601),
    [anon_sym_RBRACE] = ACTIONS(1601),
    [anon_sym_constraint] = ACTIONS(1599),
    [anon_sym_val] = ACTIONS(1599),
    [anon_sym_end] = ACTIONS(1599),
    [anon_sym_with] = ACTIONS(1599),
    [anon_sym_COLON_GT] = ACTIONS(1601),
    [anon_sym_object] = ACTIONS(1599),
    [anon_sym_inherit] = ACTIONS(1599),
    [anon_sym_method] = ACTIONS(1599),
    [anon_sym_in] = ACTIONS(1599),
    [anon_sym_as] = ACTIONS(1599),
    [anon_sym_initializer] = ACTIONS(1599),
    [anon_sym_AMP] = ACTIONS(1599),
    [anon_sym_POUND] = ACTIONS(1599),
    [anon_sym_COLON_COLON] = ACTIONS(1601),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1601),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1601),
    [anon_sym_then] = ACTIONS(1599),
    [anon_sym_else] = ACTIONS(1599),
    [anon_sym_do] = ACTIONS(1599),
    [anon_sym_done] = ACTIONS(1599),
    [anon_sym_to] = ACTIONS(1599),
    [anon_sym_downto] = ACTIONS(1599),
    [anon_sym_new] = ACTIONS(1599),
    [anon_sym_LBRACE_LT] = ACTIONS(1601),
    [anon_sym_GT_RBRACE] = ACTIONS(1601),
    [anon_sym_begin] = ACTIONS(1599),
    [sym_ocamlyacc_value] = ACTIONS(1601),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1599),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1601),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1599),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1599),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1601),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1601),
    [aux_sym_number_token1] = ACTIONS(1601),
    [aux_sym_signed_number_token1] = ACTIONS(1599),
    [anon_sym_SQUOTE] = ACTIONS(1601),
    [anon_sym_DQUOTE] = ACTIONS(1601),
    [sym_prefix_operator] = ACTIONS(1601),
    [aux_sym_sign_operator_token1] = ACTIONS(1599),
    [sym_hash_operator] = ACTIONS(1601),
    [anon_sym_lsl] = ACTIONS(1599),
    [anon_sym_lsr] = ACTIONS(1599),
    [anon_sym_asr] = ACTIONS(1599),
    [aux_sym_pow_operator_token1] = ACTIONS(1601),
    [anon_sym_mod] = ACTIONS(1599),
    [anon_sym_land] = ACTIONS(1599),
    [anon_sym_lor] = ACTIONS(1599),
    [anon_sym_lxor] = ACTIONS(1599),
    [aux_sym_mult_operator_token1] = ACTIONS(1599),
    [aux_sym_add_operator_token1] = ACTIONS(1599),
    [sym_concat_operator] = ACTIONS(1601),
    [sym_rel_operator] = ACTIONS(1599),
    [anon_sym_AMP_AMP] = ACTIONS(1599),
    [anon_sym_or] = ACTIONS(1599),
    [anon_sym_PIPE_PIPE] = ACTIONS(1599),
    [sym_assign_operator] = ACTIONS(1601),
    [sym__uppercase_identifier] = ACTIONS(1601),
    [aux_sym_tag_token1] = ACTIONS(1601),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(769)] = {
    [sym_attribute] = STATE(769),
    [sym__lowercase_identifier] = ACTIONS(1603),
    [anon_sym_and] = ACTIONS(1603),
    [anon_sym_TILDE] = ACTIONS(1603),
    [anon_sym_QMARK] = ACTIONS(1603),
    [anon_sym_LPAREN] = ACTIONS(1605),
    [anon_sym_RPAREN] = ACTIONS(1605),
    [anon_sym_COMMA] = ACTIONS(1605),
    [anon_sym_PIPE] = ACTIONS(1603),
    [anon_sym_COLON2] = ACTIONS(1603),
    [anon_sym_DOT] = ACTIONS(1605),
    [anon_sym_DASH_GT] = ACTIONS(1603),
    [anon_sym_LBRACK] = ACTIONS(1603),
    [anon_sym_RBRACK] = ACTIONS(1605),
    [anon_sym_true] = ACTIONS(1603),
    [anon_sym_false] = ACTIONS(1603),
    [anon_sym_LBRACE] = ACTIONS(1603),
    [anon_sym_SEMI] = ACTIONS(1605),
    [anon_sym_RBRACE] = ACTIONS(1605),
    [anon_sym_constraint] = ACTIONS(1603),
    [anon_sym_val] = ACTIONS(1603),
    [anon_sym_end] = ACTIONS(1603),
    [anon_sym_with] = ACTIONS(1603),
    [anon_sym_COLON_GT] = ACTIONS(1605),
    [anon_sym_object] = ACTIONS(1603),
    [anon_sym_inherit] = ACTIONS(1603),
    [anon_sym_method] = ACTIONS(1603),
    [anon_sym_in] = ACTIONS(1603),
    [anon_sym_as] = ACTIONS(1603),
    [anon_sym_initializer] = ACTIONS(1603),
    [anon_sym_AMP] = ACTIONS(1603),
    [anon_sym_POUND] = ACTIONS(1603),
    [anon_sym_COLON_COLON] = ACTIONS(1605),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1605),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1605),
    [anon_sym_then] = ACTIONS(1603),
    [anon_sym_else] = ACTIONS(1603),
    [anon_sym_do] = ACTIONS(1603),
    [anon_sym_done] = ACTIONS(1603),
    [anon_sym_to] = ACTIONS(1603),
    [anon_sym_downto] = ACTIONS(1603),
    [anon_sym_new] = ACTIONS(1603),
    [anon_sym_LBRACE_LT] = ACTIONS(1605),
    [anon_sym_GT_RBRACE] = ACTIONS(1605),
    [anon_sym_begin] = ACTIONS(1603),
    [sym_ocamlyacc_value] = ACTIONS(1605),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1603),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1605),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1603),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1603),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1605),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1605),
    [aux_sym_number_token1] = ACTIONS(1605),
    [aux_sym_signed_number_token1] = ACTIONS(1603),
    [anon_sym_SQUOTE] = ACTIONS(1605),
    [anon_sym_DQUOTE] = ACTIONS(1605),
    [sym_prefix_operator] = ACTIONS(1605),
    [aux_sym_sign_operator_token1] = ACTIONS(1603),
    [sym_hash_operator] = ACTIONS(1605),
    [anon_sym_lsl] = ACTIONS(1603),
    [anon_sym_lsr] = ACTIONS(1603),
    [anon_sym_asr] = ACTIONS(1603),
    [aux_sym_pow_operator_token1] = ACTIONS(1605),
    [anon_sym_mod] = ACTIONS(1603),
    [anon_sym_land] = ACTIONS(1603),
    [anon_sym_lor] = ACTIONS(1603),
    [anon_sym_lxor] = ACTIONS(1603),
    [aux_sym_mult_operator_token1] = ACTIONS(1603),
    [aux_sym_add_operator_token1] = ACTIONS(1603),
    [sym_concat_operator] = ACTIONS(1605),
    [sym_rel_operator] = ACTIONS(1603),
    [anon_sym_AMP_AMP] = ACTIONS(1603),
    [anon_sym_or] = ACTIONS(1603),
    [anon_sym_PIPE_PIPE] = ACTIONS(1603),
    [sym_assign_operator] = ACTIONS(1605),
    [sym__uppercase_identifier] = ACTIONS(1605),
    [aux_sym_tag_token1] = ACTIONS(1605),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(770)] = {
    [sym_attribute] = STATE(770),
    [sym__lowercase_identifier] = ACTIONS(1607),
    [anon_sym_and] = ACTIONS(1607),
    [anon_sym_TILDE] = ACTIONS(1607),
    [anon_sym_QMARK] = ACTIONS(1607),
    [anon_sym_LPAREN] = ACTIONS(1609),
    [anon_sym_RPAREN] = ACTIONS(1609),
    [anon_sym_COMMA] = ACTIONS(1609),
    [anon_sym_PIPE] = ACTIONS(1607),
    [anon_sym_COLON2] = ACTIONS(1607),
    [anon_sym_DOT] = ACTIONS(1609),
    [anon_sym_DASH_GT] = ACTIONS(1607),
    [anon_sym_LBRACK] = ACTIONS(1607),
    [anon_sym_RBRACK] = ACTIONS(1609),
    [anon_sym_true] = ACTIONS(1607),
    [anon_sym_false] = ACTIONS(1607),
    [anon_sym_LBRACE] = ACTIONS(1607),
    [anon_sym_SEMI] = ACTIONS(1609),
    [anon_sym_RBRACE] = ACTIONS(1609),
    [anon_sym_constraint] = ACTIONS(1607),
    [anon_sym_val] = ACTIONS(1607),
    [anon_sym_end] = ACTIONS(1607),
    [anon_sym_with] = ACTIONS(1607),
    [anon_sym_COLON_GT] = ACTIONS(1609),
    [anon_sym_object] = ACTIONS(1607),
    [anon_sym_inherit] = ACTIONS(1607),
    [anon_sym_method] = ACTIONS(1607),
    [anon_sym_in] = ACTIONS(1607),
    [anon_sym_as] = ACTIONS(1607),
    [anon_sym_initializer] = ACTIONS(1607),
    [anon_sym_AMP] = ACTIONS(1607),
    [anon_sym_POUND] = ACTIONS(1607),
    [anon_sym_COLON_COLON] = ACTIONS(1609),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1609),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1609),
    [anon_sym_then] = ACTIONS(1607),
    [anon_sym_else] = ACTIONS(1607),
    [anon_sym_do] = ACTIONS(1607),
    [anon_sym_done] = ACTIONS(1607),
    [anon_sym_to] = ACTIONS(1607),
    [anon_sym_downto] = ACTIONS(1607),
    [anon_sym_new] = ACTIONS(1607),
    [anon_sym_LBRACE_LT] = ACTIONS(1609),
    [anon_sym_GT_RBRACE] = ACTIONS(1609),
    [anon_sym_begin] = ACTIONS(1607),
    [sym_ocamlyacc_value] = ACTIONS(1609),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1607),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1609),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1607),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1607),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1609),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1609),
    [aux_sym_number_token1] = ACTIONS(1609),
    [aux_sym_signed_number_token1] = ACTIONS(1607),
    [anon_sym_SQUOTE] = ACTIONS(1609),
    [anon_sym_DQUOTE] = ACTIONS(1609),
    [sym_prefix_operator] = ACTIONS(1609),
    [aux_sym_sign_operator_token1] = ACTIONS(1607),
    [sym_hash_operator] = ACTIONS(1609),
    [anon_sym_lsl] = ACTIONS(1607),
    [anon_sym_lsr] = ACTIONS(1607),
    [anon_sym_asr] = ACTIONS(1607),
    [aux_sym_pow_operator_token1] = ACTIONS(1609),
    [anon_sym_mod] = ACTIONS(1607),
    [anon_sym_land] = ACTIONS(1607),
    [anon_sym_lor] = ACTIONS(1607),
    [anon_sym_lxor] = ACTIONS(1607),
    [aux_sym_mult_operator_token1] = ACTIONS(1607),
    [aux_sym_add_operator_token1] = ACTIONS(1607),
    [sym_concat_operator] = ACTIONS(1609),
    [sym_rel_operator] = ACTIONS(1607),
    [anon_sym_AMP_AMP] = ACTIONS(1607),
    [anon_sym_or] = ACTIONS(1607),
    [anon_sym_PIPE_PIPE] = ACTIONS(1607),
    [sym_assign_operator] = ACTIONS(1609),
    [sym__uppercase_identifier] = ACTIONS(1609),
    [aux_sym_tag_token1] = ACTIONS(1609),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(771)] = {
    [sym_attribute] = STATE(771),
    [sym__lowercase_identifier] = ACTIONS(1611),
    [anon_sym_and] = ACTIONS(1611),
    [anon_sym_TILDE] = ACTIONS(1611),
    [anon_sym_QMARK] = ACTIONS(1611),
    [anon_sym_LPAREN] = ACTIONS(1613),
    [anon_sym_RPAREN] = ACTIONS(1613),
    [anon_sym_COMMA] = ACTIONS(1613),
    [anon_sym_PIPE] = ACTIONS(1611),
    [anon_sym_COLON2] = ACTIONS(1611),
    [anon_sym_DOT] = ACTIONS(1613),
    [anon_sym_DASH_GT] = ACTIONS(1611),
    [anon_sym_LBRACK] = ACTIONS(1611),
    [anon_sym_RBRACK] = ACTIONS(1613),
    [anon_sym_true] = ACTIONS(1611),
    [anon_sym_false] = ACTIONS(1611),
    [anon_sym_LBRACE] = ACTIONS(1611),
    [anon_sym_SEMI] = ACTIONS(1613),
    [anon_sym_RBRACE] = ACTIONS(1613),
    [anon_sym_constraint] = ACTIONS(1611),
    [anon_sym_val] = ACTIONS(1611),
    [anon_sym_end] = ACTIONS(1611),
    [anon_sym_with] = ACTIONS(1611),
    [anon_sym_COLON_GT] = ACTIONS(1613),
    [anon_sym_object] = ACTIONS(1611),
    [anon_sym_inherit] = ACTIONS(1611),
    [anon_sym_method] = ACTIONS(1611),
    [anon_sym_in] = ACTIONS(1611),
    [anon_sym_as] = ACTIONS(1611),
    [anon_sym_initializer] = ACTIONS(1611),
    [anon_sym_AMP] = ACTIONS(1611),
    [anon_sym_POUND] = ACTIONS(1611),
    [anon_sym_COLON_COLON] = ACTIONS(1613),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1613),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1613),
    [anon_sym_then] = ACTIONS(1611),
    [anon_sym_else] = ACTIONS(1611),
    [anon_sym_do] = ACTIONS(1611),
    [anon_sym_done] = ACTIONS(1611),
    [anon_sym_to] = ACTIONS(1611),
    [anon_sym_downto] = ACTIONS(1611),
    [anon_sym_new] = ACTIONS(1611),
    [anon_sym_LBRACE_LT] = ACTIONS(1613),
    [anon_sym_GT_RBRACE] = ACTIONS(1613),
    [anon_sym_begin] = ACTIONS(1611),
    [sym_ocamlyacc_value] = ACTIONS(1613),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1611),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1613),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1611),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1611),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1613),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1613),
    [aux_sym_number_token1] = ACTIONS(1613),
    [aux_sym_signed_number_token1] = ACTIONS(1611),
    [anon_sym_SQUOTE] = ACTIONS(1613),
    [anon_sym_DQUOTE] = ACTIONS(1613),
    [sym_prefix_operator] = ACTIONS(1613),
    [aux_sym_sign_operator_token1] = ACTIONS(1611),
    [sym_hash_operator] = ACTIONS(1613),
    [anon_sym_lsl] = ACTIONS(1611),
    [anon_sym_lsr] = ACTIONS(1611),
    [anon_sym_asr] = ACTIONS(1611),
    [aux_sym_pow_operator_token1] = ACTIONS(1613),
    [anon_sym_mod] = ACTIONS(1611),
    [anon_sym_land] = ACTIONS(1611),
    [anon_sym_lor] = ACTIONS(1611),
    [anon_sym_lxor] = ACTIONS(1611),
    [aux_sym_mult_operator_token1] = ACTIONS(1611),
    [aux_sym_add_operator_token1] = ACTIONS(1611),
    [sym_concat_operator] = ACTIONS(1613),
    [sym_rel_operator] = ACTIONS(1611),
    [anon_sym_AMP_AMP] = ACTIONS(1611),
    [anon_sym_or] = ACTIONS(1611),
    [anon_sym_PIPE_PIPE] = ACTIONS(1611),
    [sym_assign_operator] = ACTIONS(1613),
    [sym__uppercase_identifier] = ACTIONS(1613),
    [aux_sym_tag_token1] = ACTIONS(1613),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(772)] = {
    [sym_attribute] = STATE(772),
    [sym__lowercase_identifier] = ACTIONS(1615),
    [anon_sym_and] = ACTIONS(1615),
    [anon_sym_TILDE] = ACTIONS(1615),
    [anon_sym_QMARK] = ACTIONS(1615),
    [anon_sym_LPAREN] = ACTIONS(1617),
    [anon_sym_RPAREN] = ACTIONS(1617),
    [anon_sym_COMMA] = ACTIONS(1617),
    [anon_sym_PIPE] = ACTIONS(1615),
    [anon_sym_COLON2] = ACTIONS(1615),
    [anon_sym_DOT] = ACTIONS(751),
    [anon_sym_DASH_GT] = ACTIONS(1615),
    [anon_sym_LBRACK] = ACTIONS(1615),
    [anon_sym_RBRACK] = ACTIONS(1617),
    [anon_sym_true] = ACTIONS(1615),
    [anon_sym_false] = ACTIONS(1615),
    [anon_sym_LBRACE] = ACTIONS(1615),
    [anon_sym_SEMI] = ACTIONS(1617),
    [anon_sym_RBRACE] = ACTIONS(1617),
    [anon_sym_constraint] = ACTIONS(1615),
    [anon_sym_val] = ACTIONS(1615),
    [anon_sym_end] = ACTIONS(1615),
    [anon_sym_with] = ACTIONS(1615),
    [anon_sym_COLON_GT] = ACTIONS(1617),
    [anon_sym_object] = ACTIONS(1615),
    [anon_sym_inherit] = ACTIONS(1615),
    [anon_sym_method] = ACTIONS(1615),
    [anon_sym_in] = ACTIONS(1615),
    [anon_sym_as] = ACTIONS(1615),
    [anon_sym_initializer] = ACTIONS(1615),
    [anon_sym_AMP] = ACTIONS(1615),
    [anon_sym_POUND] = ACTIONS(753),
    [anon_sym_COLON_COLON] = ACTIONS(1617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1617),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1617),
    [anon_sym_then] = ACTIONS(1615),
    [anon_sym_else] = ACTIONS(1615),
    [anon_sym_do] = ACTIONS(1615),
    [anon_sym_done] = ACTIONS(1615),
    [anon_sym_to] = ACTIONS(1615),
    [anon_sym_downto] = ACTIONS(1615),
    [anon_sym_new] = ACTIONS(1615),
    [anon_sym_LBRACE_LT] = ACTIONS(1617),
    [anon_sym_GT_RBRACE] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1615),
    [sym_ocamlyacc_value] = ACTIONS(1617),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1615),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1617),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1615),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1615),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1617),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1617),
    [aux_sym_number_token1] = ACTIONS(1617),
    [aux_sym_signed_number_token1] = ACTIONS(1615),
    [anon_sym_SQUOTE] = ACTIONS(1617),
    [anon_sym_DQUOTE] = ACTIONS(1617),
    [sym_prefix_operator] = ACTIONS(1617),
    [aux_sym_sign_operator_token1] = ACTIONS(1615),
    [sym_hash_operator] = ACTIONS(755),
    [anon_sym_lsl] = ACTIONS(1615),
    [anon_sym_lsr] = ACTIONS(1615),
    [anon_sym_asr] = ACTIONS(1615),
    [aux_sym_pow_operator_token1] = ACTIONS(1617),
    [anon_sym_mod] = ACTIONS(1615),
    [anon_sym_land] = ACTIONS(1615),
    [anon_sym_lor] = ACTIONS(1615),
    [anon_sym_lxor] = ACTIONS(1615),
    [aux_sym_mult_operator_token1] = ACTIONS(1615),
    [aux_sym_add_operator_token1] = ACTIONS(1615),
    [sym_concat_operator] = ACTIONS(1617),
    [sym_rel_operator] = ACTIONS(1615),
    [anon_sym_AMP_AMP] = ACTIONS(1615),
    [anon_sym_or] = ACTIONS(1615),
    [anon_sym_PIPE_PIPE] = ACTIONS(1615),
    [sym_assign_operator] = ACTIONS(1617),
    [sym__uppercase_identifier] = ACTIONS(1617),
    [aux_sym_tag_token1] = ACTIONS(1617),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(773)] = {
    [sym_attribute] = STATE(773),
    [sym__lowercase_identifier] = ACTIONS(1619),
    [anon_sym_and] = ACTIONS(1619),
    [anon_sym_TILDE] = ACTIONS(1619),
    [anon_sym_QMARK] = ACTIONS(1619),
    [anon_sym_LPAREN] = ACTIONS(1621),
    [anon_sym_RPAREN] = ACTIONS(1621),
    [anon_sym_COMMA] = ACTIONS(1621),
    [anon_sym_PIPE] = ACTIONS(1619),
    [anon_sym_COLON2] = ACTIONS(1619),
    [anon_sym_DOT] = ACTIONS(1621),
    [anon_sym_DASH_GT] = ACTIONS(1619),
    [anon_sym_LBRACK] = ACTIONS(1619),
    [anon_sym_RBRACK] = ACTIONS(1621),
    [anon_sym_true] = ACTIONS(1619),
    [anon_sym_false] = ACTIONS(1619),
    [anon_sym_LBRACE] = ACTIONS(1619),
    [anon_sym_SEMI] = ACTIONS(1621),
    [anon_sym_RBRACE] = ACTIONS(1621),
    [anon_sym_constraint] = ACTIONS(1619),
    [anon_sym_val] = ACTIONS(1619),
    [anon_sym_end] = ACTIONS(1619),
    [anon_sym_with] = ACTIONS(1619),
    [anon_sym_COLON_GT] = ACTIONS(1621),
    [anon_sym_object] = ACTIONS(1619),
    [anon_sym_inherit] = ACTIONS(1619),
    [anon_sym_method] = ACTIONS(1619),
    [anon_sym_in] = ACTIONS(1619),
    [anon_sym_as] = ACTIONS(1619),
    [anon_sym_initializer] = ACTIONS(1619),
    [anon_sym_AMP] = ACTIONS(1619),
    [anon_sym_POUND] = ACTIONS(1619),
    [anon_sym_COLON_COLON] = ACTIONS(1621),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1621),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1621),
    [anon_sym_then] = ACTIONS(1619),
    [anon_sym_else] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1619),
    [anon_sym_done] = ACTIONS(1619),
    [anon_sym_to] = ACTIONS(1619),
    [anon_sym_downto] = ACTIONS(1619),
    [anon_sym_new] = ACTIONS(1619),
    [anon_sym_LBRACE_LT] = ACTIONS(1621),
    [anon_sym_GT_RBRACE] = ACTIONS(1621),
    [anon_sym_begin] = ACTIONS(1619),
    [sym_ocamlyacc_value] = ACTIONS(1621),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1619),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1621),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1619),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1619),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1621),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1621),
    [aux_sym_number_token1] = ACTIONS(1621),
    [aux_sym_signed_number_token1] = ACTIONS(1619),
    [anon_sym_SQUOTE] = ACTIONS(1621),
    [anon_sym_DQUOTE] = ACTIONS(1621),
    [sym_prefix_operator] = ACTIONS(1621),
    [aux_sym_sign_operator_token1] = ACTIONS(1619),
    [sym_hash_operator] = ACTIONS(1621),
    [anon_sym_lsl] = ACTIONS(1619),
    [anon_sym_lsr] = ACTIONS(1619),
    [anon_sym_asr] = ACTIONS(1619),
    [aux_sym_pow_operator_token1] = ACTIONS(1621),
    [anon_sym_mod] = ACTIONS(1619),
    [anon_sym_land] = ACTIONS(1619),
    [anon_sym_lor] = ACTIONS(1619),
    [anon_sym_lxor] = ACTIONS(1619),
    [aux_sym_mult_operator_token1] = ACTIONS(1619),
    [aux_sym_add_operator_token1] = ACTIONS(1619),
    [sym_concat_operator] = ACTIONS(1621),
    [sym_rel_operator] = ACTIONS(1619),
    [anon_sym_AMP_AMP] = ACTIONS(1619),
    [anon_sym_or] = ACTIONS(1619),
    [anon_sym_PIPE_PIPE] = ACTIONS(1619),
    [sym_assign_operator] = ACTIONS(1621),
    [sym__uppercase_identifier] = ACTIONS(1621),
    [aux_sym_tag_token1] = ACTIONS(1621),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(774)] = {
    [sym_attribute] = STATE(774),
    [sym__lowercase_identifier] = ACTIONS(1623),
    [anon_sym_and] = ACTIONS(1623),
    [anon_sym_TILDE] = ACTIONS(1623),
    [anon_sym_QMARK] = ACTIONS(1623),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_RPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(1625),
    [anon_sym_PIPE] = ACTIONS(1623),
    [anon_sym_COLON2] = ACTIONS(1623),
    [anon_sym_DOT] = ACTIONS(1625),
    [anon_sym_DASH_GT] = ACTIONS(1623),
    [anon_sym_LBRACK] = ACTIONS(1623),
    [anon_sym_RBRACK] = ACTIONS(1625),
    [anon_sym_true] = ACTIONS(1623),
    [anon_sym_false] = ACTIONS(1623),
    [anon_sym_LBRACE] = ACTIONS(1623),
    [anon_sym_SEMI] = ACTIONS(1625),
    [anon_sym_RBRACE] = ACTIONS(1625),
    [anon_sym_constraint] = ACTIONS(1623),
    [anon_sym_val] = ACTIONS(1623),
    [anon_sym_end] = ACTIONS(1623),
    [anon_sym_with] = ACTIONS(1623),
    [anon_sym_COLON_GT] = ACTIONS(1625),
    [anon_sym_object] = ACTIONS(1623),
    [anon_sym_inherit] = ACTIONS(1623),
    [anon_sym_method] = ACTIONS(1623),
    [anon_sym_in] = ACTIONS(1623),
    [anon_sym_as] = ACTIONS(1623),
    [anon_sym_initializer] = ACTIONS(1623),
    [anon_sym_AMP] = ACTIONS(1623),
    [anon_sym_POUND] = ACTIONS(1623),
    [anon_sym_COLON_COLON] = ACTIONS(1625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1625),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1625),
    [anon_sym_then] = ACTIONS(1623),
    [anon_sym_else] = ACTIONS(1623),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_done] = ACTIONS(1623),
    [anon_sym_to] = ACTIONS(1623),
    [anon_sym_downto] = ACTIONS(1623),
    [anon_sym_new] = ACTIONS(1623),
    [anon_sym_LBRACE_LT] = ACTIONS(1625),
    [anon_sym_GT_RBRACE] = ACTIONS(1625),
    [anon_sym_begin] = ACTIONS(1623),
    [sym_ocamlyacc_value] = ACTIONS(1625),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1623),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1625),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1623),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1623),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1625),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1625),
    [aux_sym_number_token1] = ACTIONS(1625),
    [aux_sym_signed_number_token1] = ACTIONS(1623),
    [anon_sym_SQUOTE] = ACTIONS(1625),
    [anon_sym_DQUOTE] = ACTIONS(1625),
    [sym_prefix_operator] = ACTIONS(1625),
    [aux_sym_sign_operator_token1] = ACTIONS(1623),
    [sym_hash_operator] = ACTIONS(1625),
    [anon_sym_lsl] = ACTIONS(1623),
    [anon_sym_lsr] = ACTIONS(1623),
    [anon_sym_asr] = ACTIONS(1623),
    [aux_sym_pow_operator_token1] = ACTIONS(1625),
    [anon_sym_mod] = ACTIONS(1623),
    [anon_sym_land] = ACTIONS(1623),
    [anon_sym_lor] = ACTIONS(1623),
    [anon_sym_lxor] = ACTIONS(1623),
    [aux_sym_mult_operator_token1] = ACTIONS(1623),
    [aux_sym_add_operator_token1] = ACTIONS(1623),
    [sym_concat_operator] = ACTIONS(1625),
    [sym_rel_operator] = ACTIONS(1623),
    [anon_sym_AMP_AMP] = ACTIONS(1623),
    [anon_sym_or] = ACTIONS(1623),
    [anon_sym_PIPE_PIPE] = ACTIONS(1623),
    [sym_assign_operator] = ACTIONS(1625),
    [sym__uppercase_identifier] = ACTIONS(1625),
    [aux_sym_tag_token1] = ACTIONS(1625),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(775)] = {
    [sym_attribute] = STATE(775),
    [sym__lowercase_identifier] = ACTIONS(1627),
    [anon_sym_and] = ACTIONS(1627),
    [anon_sym_TILDE] = ACTIONS(1627),
    [anon_sym_QMARK] = ACTIONS(1627),
    [anon_sym_LPAREN] = ACTIONS(1629),
    [anon_sym_RPAREN] = ACTIONS(1629),
    [anon_sym_COMMA] = ACTIONS(1629),
    [anon_sym_PIPE] = ACTIONS(1627),
    [anon_sym_COLON2] = ACTIONS(1627),
    [anon_sym_DOT] = ACTIONS(1629),
    [anon_sym_DASH_GT] = ACTIONS(1627),
    [anon_sym_LBRACK] = ACTIONS(1627),
    [anon_sym_RBRACK] = ACTIONS(1629),
    [anon_sym_true] = ACTIONS(1627),
    [anon_sym_false] = ACTIONS(1627),
    [anon_sym_LBRACE] = ACTIONS(1627),
    [anon_sym_SEMI] = ACTIONS(1629),
    [anon_sym_RBRACE] = ACTIONS(1629),
    [anon_sym_constraint] = ACTIONS(1627),
    [anon_sym_val] = ACTIONS(1627),
    [anon_sym_end] = ACTIONS(1627),
    [anon_sym_with] = ACTIONS(1627),
    [anon_sym_COLON_GT] = ACTIONS(1629),
    [anon_sym_object] = ACTIONS(1627),
    [anon_sym_inherit] = ACTIONS(1627),
    [anon_sym_method] = ACTIONS(1627),
    [anon_sym_in] = ACTIONS(1627),
    [anon_sym_as] = ACTIONS(1627),
    [anon_sym_initializer] = ACTIONS(1627),
    [anon_sym_AMP] = ACTIONS(1627),
    [anon_sym_POUND] = ACTIONS(1627),
    [anon_sym_COLON_COLON] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1629),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1629),
    [anon_sym_then] = ACTIONS(1627),
    [anon_sym_else] = ACTIONS(1627),
    [anon_sym_do] = ACTIONS(1627),
    [anon_sym_done] = ACTIONS(1627),
    [anon_sym_to] = ACTIONS(1627),
    [anon_sym_downto] = ACTIONS(1627),
    [anon_sym_new] = ACTIONS(1627),
    [anon_sym_LBRACE_LT] = ACTIONS(1629),
    [anon_sym_GT_RBRACE] = ACTIONS(1629),
    [anon_sym_begin] = ACTIONS(1627),
    [sym_ocamlyacc_value] = ACTIONS(1629),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1627),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1629),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1627),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1627),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1629),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1629),
    [aux_sym_number_token1] = ACTIONS(1629),
    [aux_sym_signed_number_token1] = ACTIONS(1627),
    [anon_sym_SQUOTE] = ACTIONS(1629),
    [anon_sym_DQUOTE] = ACTIONS(1629),
    [sym_prefix_operator] = ACTIONS(1629),
    [aux_sym_sign_operator_token1] = ACTIONS(1627),
    [sym_hash_operator] = ACTIONS(1629),
    [anon_sym_lsl] = ACTIONS(1627),
    [anon_sym_lsr] = ACTIONS(1627),
    [anon_sym_asr] = ACTIONS(1627),
    [aux_sym_pow_operator_token1] = ACTIONS(1629),
    [anon_sym_mod] = ACTIONS(1627),
    [anon_sym_land] = ACTIONS(1627),
    [anon_sym_lor] = ACTIONS(1627),
    [anon_sym_lxor] = ACTIONS(1627),
    [aux_sym_mult_operator_token1] = ACTIONS(1627),
    [aux_sym_add_operator_token1] = ACTIONS(1627),
    [sym_concat_operator] = ACTIONS(1629),
    [sym_rel_operator] = ACTIONS(1627),
    [anon_sym_AMP_AMP] = ACTIONS(1627),
    [anon_sym_or] = ACTIONS(1627),
    [anon_sym_PIPE_PIPE] = ACTIONS(1627),
    [sym_assign_operator] = ACTIONS(1629),
    [sym__uppercase_identifier] = ACTIONS(1629),
    [aux_sym_tag_token1] = ACTIONS(1629),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(776)] = {
    [sym_attribute] = STATE(776),
    [sym__lowercase_identifier] = ACTIONS(1631),
    [anon_sym_and] = ACTIONS(1631),
    [anon_sym_TILDE] = ACTIONS(1631),
    [anon_sym_QMARK] = ACTIONS(1631),
    [anon_sym_LPAREN] = ACTIONS(1633),
    [anon_sym_RPAREN] = ACTIONS(1633),
    [anon_sym_COMMA] = ACTIONS(1633),
    [anon_sym_PIPE] = ACTIONS(1631),
    [anon_sym_COLON2] = ACTIONS(1631),
    [anon_sym_DOT] = ACTIONS(1633),
    [anon_sym_DASH_GT] = ACTIONS(1631),
    [anon_sym_LBRACK] = ACTIONS(1631),
    [anon_sym_RBRACK] = ACTIONS(1633),
    [anon_sym_true] = ACTIONS(1631),
    [anon_sym_false] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1631),
    [anon_sym_SEMI] = ACTIONS(1633),
    [anon_sym_RBRACE] = ACTIONS(1633),
    [anon_sym_constraint] = ACTIONS(1631),
    [anon_sym_val] = ACTIONS(1631),
    [anon_sym_end] = ACTIONS(1631),
    [anon_sym_with] = ACTIONS(1631),
    [anon_sym_COLON_GT] = ACTIONS(1633),
    [anon_sym_object] = ACTIONS(1631),
    [anon_sym_inherit] = ACTIONS(1631),
    [anon_sym_method] = ACTIONS(1631),
    [anon_sym_in] = ACTIONS(1631),
    [anon_sym_as] = ACTIONS(1631),
    [anon_sym_initializer] = ACTIONS(1631),
    [anon_sym_AMP] = ACTIONS(1631),
    [anon_sym_POUND] = ACTIONS(1631),
    [anon_sym_COLON_COLON] = ACTIONS(1633),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1633),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1633),
    [anon_sym_then] = ACTIONS(1631),
    [anon_sym_else] = ACTIONS(1631),
    [anon_sym_do] = ACTIONS(1631),
    [anon_sym_done] = ACTIONS(1631),
    [anon_sym_to] = ACTIONS(1631),
    [anon_sym_downto] = ACTIONS(1631),
    [anon_sym_new] = ACTIONS(1631),
    [anon_sym_LBRACE_LT] = ACTIONS(1633),
    [anon_sym_GT_RBRACE] = ACTIONS(1633),
    [anon_sym_begin] = ACTIONS(1631),
    [sym_ocamlyacc_value] = ACTIONS(1633),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1631),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1633),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1631),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1631),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1633),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1633),
    [aux_sym_number_token1] = ACTIONS(1633),
    [aux_sym_signed_number_token1] = ACTIONS(1631),
    [anon_sym_SQUOTE] = ACTIONS(1633),
    [anon_sym_DQUOTE] = ACTIONS(1633),
    [sym_prefix_operator] = ACTIONS(1633),
    [aux_sym_sign_operator_token1] = ACTIONS(1631),
    [sym_hash_operator] = ACTIONS(1633),
    [anon_sym_lsl] = ACTIONS(1631),
    [anon_sym_lsr] = ACTIONS(1631),
    [anon_sym_asr] = ACTIONS(1631),
    [aux_sym_pow_operator_token1] = ACTIONS(1633),
    [anon_sym_mod] = ACTIONS(1631),
    [anon_sym_land] = ACTIONS(1631),
    [anon_sym_lor] = ACTIONS(1631),
    [anon_sym_lxor] = ACTIONS(1631),
    [aux_sym_mult_operator_token1] = ACTIONS(1631),
    [aux_sym_add_operator_token1] = ACTIONS(1631),
    [sym_concat_operator] = ACTIONS(1633),
    [sym_rel_operator] = ACTIONS(1631),
    [anon_sym_AMP_AMP] = ACTIONS(1631),
    [anon_sym_or] = ACTIONS(1631),
    [anon_sym_PIPE_PIPE] = ACTIONS(1631),
    [sym_assign_operator] = ACTIONS(1633),
    [sym__uppercase_identifier] = ACTIONS(1633),
    [aux_sym_tag_token1] = ACTIONS(1633),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(777)] = {
    [sym_attribute] = STATE(777),
    [sym__lowercase_identifier] = ACTIONS(1635),
    [anon_sym_and] = ACTIONS(1635),
    [anon_sym_TILDE] = ACTIONS(1635),
    [anon_sym_QMARK] = ACTIONS(1635),
    [anon_sym_LPAREN] = ACTIONS(1637),
    [anon_sym_RPAREN] = ACTIONS(1637),
    [anon_sym_COMMA] = ACTIONS(1637),
    [anon_sym_PIPE] = ACTIONS(1635),
    [anon_sym_COLON2] = ACTIONS(1635),
    [anon_sym_DOT] = ACTIONS(1637),
    [anon_sym_DASH_GT] = ACTIONS(1635),
    [anon_sym_LBRACK] = ACTIONS(1635),
    [anon_sym_RBRACK] = ACTIONS(1637),
    [anon_sym_true] = ACTIONS(1635),
    [anon_sym_false] = ACTIONS(1635),
    [anon_sym_LBRACE] = ACTIONS(1635),
    [anon_sym_SEMI] = ACTIONS(1637),
    [anon_sym_RBRACE] = ACTIONS(1637),
    [anon_sym_constraint] = ACTIONS(1635),
    [anon_sym_val] = ACTIONS(1635),
    [anon_sym_end] = ACTIONS(1635),
    [anon_sym_with] = ACTIONS(1635),
    [anon_sym_COLON_GT] = ACTIONS(1637),
    [anon_sym_object] = ACTIONS(1635),
    [anon_sym_inherit] = ACTIONS(1635),
    [anon_sym_method] = ACTIONS(1635),
    [anon_sym_in] = ACTIONS(1635),
    [anon_sym_as] = ACTIONS(1635),
    [anon_sym_initializer] = ACTIONS(1635),
    [anon_sym_AMP] = ACTIONS(1635),
    [anon_sym_POUND] = ACTIONS(1635),
    [anon_sym_COLON_COLON] = ACTIONS(1637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1637),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1637),
    [anon_sym_then] = ACTIONS(1635),
    [anon_sym_else] = ACTIONS(1635),
    [anon_sym_do] = ACTIONS(1635),
    [anon_sym_done] = ACTIONS(1635),
    [anon_sym_to] = ACTIONS(1635),
    [anon_sym_downto] = ACTIONS(1635),
    [anon_sym_new] = ACTIONS(1635),
    [anon_sym_LBRACE_LT] = ACTIONS(1637),
    [anon_sym_GT_RBRACE] = ACTIONS(1637),
    [anon_sym_begin] = ACTIONS(1635),
    [sym_ocamlyacc_value] = ACTIONS(1637),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1635),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1637),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1635),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1635),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1637),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1637),
    [aux_sym_number_token1] = ACTIONS(1637),
    [aux_sym_signed_number_token1] = ACTIONS(1635),
    [anon_sym_SQUOTE] = ACTIONS(1637),
    [anon_sym_DQUOTE] = ACTIONS(1637),
    [sym_prefix_operator] = ACTIONS(1637),
    [aux_sym_sign_operator_token1] = ACTIONS(1635),
    [sym_hash_operator] = ACTIONS(1637),
    [anon_sym_lsl] = ACTIONS(1635),
    [anon_sym_lsr] = ACTIONS(1635),
    [anon_sym_asr] = ACTIONS(1635),
    [aux_sym_pow_operator_token1] = ACTIONS(1637),
    [anon_sym_mod] = ACTIONS(1635),
    [anon_sym_land] = ACTIONS(1635),
    [anon_sym_lor] = ACTIONS(1635),
    [anon_sym_lxor] = ACTIONS(1635),
    [aux_sym_mult_operator_token1] = ACTIONS(1635),
    [aux_sym_add_operator_token1] = ACTIONS(1635),
    [sym_concat_operator] = ACTIONS(1637),
    [sym_rel_operator] = ACTIONS(1635),
    [anon_sym_AMP_AMP] = ACTIONS(1635),
    [anon_sym_or] = ACTIONS(1635),
    [anon_sym_PIPE_PIPE] = ACTIONS(1635),
    [sym_assign_operator] = ACTIONS(1637),
    [sym__uppercase_identifier] = ACTIONS(1637),
    [aux_sym_tag_token1] = ACTIONS(1637),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(778)] = {
    [sym_attribute] = STATE(778),
    [sym__lowercase_identifier] = ACTIONS(1639),
    [anon_sym_and] = ACTIONS(1639),
    [anon_sym_TILDE] = ACTIONS(1639),
    [anon_sym_QMARK] = ACTIONS(1639),
    [anon_sym_LPAREN] = ACTIONS(1641),
    [anon_sym_RPAREN] = ACTIONS(1641),
    [anon_sym_COMMA] = ACTIONS(1641),
    [anon_sym_PIPE] = ACTIONS(1639),
    [anon_sym_COLON2] = ACTIONS(1639),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_DASH_GT] = ACTIONS(1639),
    [anon_sym_LBRACK] = ACTIONS(1639),
    [anon_sym_RBRACK] = ACTIONS(1641),
    [anon_sym_true] = ACTIONS(1639),
    [anon_sym_false] = ACTIONS(1639),
    [anon_sym_LBRACE] = ACTIONS(1639),
    [anon_sym_SEMI] = ACTIONS(1641),
    [anon_sym_RBRACE] = ACTIONS(1641),
    [anon_sym_constraint] = ACTIONS(1639),
    [anon_sym_val] = ACTIONS(1639),
    [anon_sym_end] = ACTIONS(1639),
    [anon_sym_with] = ACTIONS(1639),
    [anon_sym_COLON_GT] = ACTIONS(1641),
    [anon_sym_object] = ACTIONS(1639),
    [anon_sym_inherit] = ACTIONS(1639),
    [anon_sym_method] = ACTIONS(1639),
    [anon_sym_in] = ACTIONS(1639),
    [anon_sym_as] = ACTIONS(1639),
    [anon_sym_initializer] = ACTIONS(1639),
    [anon_sym_AMP] = ACTIONS(1639),
    [anon_sym_POUND] = ACTIONS(1639),
    [anon_sym_COLON_COLON] = ACTIONS(1641),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1641),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1641),
    [anon_sym_then] = ACTIONS(1639),
    [anon_sym_else] = ACTIONS(1639),
    [anon_sym_do] = ACTIONS(1639),
    [anon_sym_done] = ACTIONS(1639),
    [anon_sym_to] = ACTIONS(1639),
    [anon_sym_downto] = ACTIONS(1639),
    [anon_sym_new] = ACTIONS(1639),
    [anon_sym_LBRACE_LT] = ACTIONS(1641),
    [anon_sym_GT_RBRACE] = ACTIONS(1641),
    [anon_sym_begin] = ACTIONS(1639),
    [sym_ocamlyacc_value] = ACTIONS(1641),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1639),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1641),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1639),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1639),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1641),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1641),
    [aux_sym_number_token1] = ACTIONS(1641),
    [aux_sym_signed_number_token1] = ACTIONS(1639),
    [anon_sym_SQUOTE] = ACTIONS(1641),
    [anon_sym_DQUOTE] = ACTIONS(1641),
    [sym_prefix_operator] = ACTIONS(1641),
    [aux_sym_sign_operator_token1] = ACTIONS(1639),
    [sym_hash_operator] = ACTIONS(1641),
    [anon_sym_lsl] = ACTIONS(1639),
    [anon_sym_lsr] = ACTIONS(1639),
    [anon_sym_asr] = ACTIONS(1639),
    [aux_sym_pow_operator_token1] = ACTIONS(1641),
    [anon_sym_mod] = ACTIONS(1639),
    [anon_sym_land] = ACTIONS(1639),
    [anon_sym_lor] = ACTIONS(1639),
    [anon_sym_lxor] = ACTIONS(1639),
    [aux_sym_mult_operator_token1] = ACTIONS(1639),
    [aux_sym_add_operator_token1] = ACTIONS(1639),
    [sym_concat_operator] = ACTIONS(1641),
    [sym_rel_operator] = ACTIONS(1639),
    [anon_sym_AMP_AMP] = ACTIONS(1639),
    [anon_sym_or] = ACTIONS(1639),
    [anon_sym_PIPE_PIPE] = ACTIONS(1639),
    [sym_assign_operator] = ACTIONS(1641),
    [sym__uppercase_identifier] = ACTIONS(1641),
    [aux_sym_tag_token1] = ACTIONS(1641),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(779)] = {
    [sym_attribute] = STATE(779),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_and] = ACTIONS(1643),
    [anon_sym_TILDE] = ACTIONS(1643),
    [anon_sym_QMARK] = ACTIONS(1643),
    [anon_sym_LPAREN] = ACTIONS(1645),
    [anon_sym_RPAREN] = ACTIONS(1645),
    [anon_sym_COMMA] = ACTIONS(1645),
    [anon_sym_PIPE] = ACTIONS(1643),
    [anon_sym_COLON2] = ACTIONS(1643),
    [anon_sym_DOT] = ACTIONS(1645),
    [anon_sym_DASH_GT] = ACTIONS(1643),
    [anon_sym_LBRACK] = ACTIONS(1643),
    [anon_sym_RBRACK] = ACTIONS(1645),
    [anon_sym_true] = ACTIONS(1643),
    [anon_sym_false] = ACTIONS(1643),
    [anon_sym_LBRACE] = ACTIONS(1643),
    [anon_sym_SEMI] = ACTIONS(1645),
    [anon_sym_RBRACE] = ACTIONS(1645),
    [anon_sym_constraint] = ACTIONS(1643),
    [anon_sym_val] = ACTIONS(1643),
    [anon_sym_end] = ACTIONS(1643),
    [anon_sym_with] = ACTIONS(1643),
    [anon_sym_COLON_GT] = ACTIONS(1645),
    [anon_sym_object] = ACTIONS(1643),
    [anon_sym_inherit] = ACTIONS(1643),
    [anon_sym_method] = ACTIONS(1643),
    [anon_sym_in] = ACTIONS(1643),
    [anon_sym_as] = ACTIONS(1643),
    [anon_sym_initializer] = ACTIONS(1643),
    [anon_sym_AMP] = ACTIONS(1643),
    [anon_sym_POUND] = ACTIONS(1643),
    [anon_sym_COLON_COLON] = ACTIONS(1645),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1645),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1645),
    [anon_sym_then] = ACTIONS(1643),
    [anon_sym_else] = ACTIONS(1643),
    [anon_sym_do] = ACTIONS(1643),
    [anon_sym_done] = ACTIONS(1643),
    [anon_sym_to] = ACTIONS(1643),
    [anon_sym_downto] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1643),
    [anon_sym_LBRACE_LT] = ACTIONS(1645),
    [anon_sym_GT_RBRACE] = ACTIONS(1645),
    [anon_sym_begin] = ACTIONS(1643),
    [sym_ocamlyacc_value] = ACTIONS(1645),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1643),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1645),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1643),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1643),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1645),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1645),
    [aux_sym_number_token1] = ACTIONS(1645),
    [aux_sym_signed_number_token1] = ACTIONS(1643),
    [anon_sym_SQUOTE] = ACTIONS(1645),
    [anon_sym_DQUOTE] = ACTIONS(1645),
    [sym_prefix_operator] = ACTIONS(1645),
    [aux_sym_sign_operator_token1] = ACTIONS(1643),
    [sym_hash_operator] = ACTIONS(1645),
    [anon_sym_lsl] = ACTIONS(1643),
    [anon_sym_lsr] = ACTIONS(1643),
    [anon_sym_asr] = ACTIONS(1643),
    [aux_sym_pow_operator_token1] = ACTIONS(1645),
    [anon_sym_mod] = ACTIONS(1643),
    [anon_sym_land] = ACTIONS(1643),
    [anon_sym_lor] = ACTIONS(1643),
    [anon_sym_lxor] = ACTIONS(1643),
    [aux_sym_mult_operator_token1] = ACTIONS(1643),
    [aux_sym_add_operator_token1] = ACTIONS(1643),
    [sym_concat_operator] = ACTIONS(1645),
    [sym_rel_operator] = ACTIONS(1643),
    [anon_sym_AMP_AMP] = ACTIONS(1643),
    [anon_sym_or] = ACTIONS(1643),
    [anon_sym_PIPE_PIPE] = ACTIONS(1643),
    [sym_assign_operator] = ACTIONS(1645),
    [sym__uppercase_identifier] = ACTIONS(1645),
    [aux_sym_tag_token1] = ACTIONS(1645),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(780)] = {
    [sym_attribute] = STATE(780),
    [sym__lowercase_identifier] = ACTIONS(1647),
    [anon_sym_and] = ACTIONS(1647),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_QMARK] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1649),
    [anon_sym_RPAREN] = ACTIONS(1649),
    [anon_sym_COMMA] = ACTIONS(1649),
    [anon_sym_PIPE] = ACTIONS(1647),
    [anon_sym_COLON2] = ACTIONS(1647),
    [anon_sym_DOT] = ACTIONS(1649),
    [anon_sym_DASH_GT] = ACTIONS(1647),
    [anon_sym_LBRACK] = ACTIONS(1647),
    [anon_sym_RBRACK] = ACTIONS(1649),
    [anon_sym_true] = ACTIONS(1647),
    [anon_sym_false] = ACTIONS(1647),
    [anon_sym_LBRACE] = ACTIONS(1647),
    [anon_sym_SEMI] = ACTIONS(1649),
    [anon_sym_RBRACE] = ACTIONS(1649),
    [anon_sym_constraint] = ACTIONS(1647),
    [anon_sym_val] = ACTIONS(1647),
    [anon_sym_end] = ACTIONS(1647),
    [anon_sym_with] = ACTIONS(1647),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_object] = ACTIONS(1647),
    [anon_sym_inherit] = ACTIONS(1647),
    [anon_sym_method] = ACTIONS(1647),
    [anon_sym_in] = ACTIONS(1647),
    [anon_sym_as] = ACTIONS(1647),
    [anon_sym_initializer] = ACTIONS(1647),
    [anon_sym_AMP] = ACTIONS(1647),
    [anon_sym_POUND] = ACTIONS(1647),
    [anon_sym_COLON_COLON] = ACTIONS(1649),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1649),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1649),
    [anon_sym_then] = ACTIONS(1647),
    [anon_sym_else] = ACTIONS(1647),
    [anon_sym_do] = ACTIONS(1647),
    [anon_sym_done] = ACTIONS(1647),
    [anon_sym_to] = ACTIONS(1647),
    [anon_sym_downto] = ACTIONS(1647),
    [anon_sym_new] = ACTIONS(1647),
    [anon_sym_LBRACE_LT] = ACTIONS(1649),
    [anon_sym_GT_RBRACE] = ACTIONS(1649),
    [anon_sym_begin] = ACTIONS(1647),
    [sym_ocamlyacc_value] = ACTIONS(1649),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1647),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1649),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1647),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1647),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1649),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1649),
    [aux_sym_number_token1] = ACTIONS(1649),
    [aux_sym_signed_number_token1] = ACTIONS(1647),
    [anon_sym_SQUOTE] = ACTIONS(1649),
    [anon_sym_DQUOTE] = ACTIONS(1649),
    [sym_prefix_operator] = ACTIONS(1649),
    [aux_sym_sign_operator_token1] = ACTIONS(1647),
    [sym_hash_operator] = ACTIONS(1649),
    [anon_sym_lsl] = ACTIONS(1647),
    [anon_sym_lsr] = ACTIONS(1647),
    [anon_sym_asr] = ACTIONS(1647),
    [aux_sym_pow_operator_token1] = ACTIONS(1649),
    [anon_sym_mod] = ACTIONS(1647),
    [anon_sym_land] = ACTIONS(1647),
    [anon_sym_lor] = ACTIONS(1647),
    [anon_sym_lxor] = ACTIONS(1647),
    [aux_sym_mult_operator_token1] = ACTIONS(1647),
    [aux_sym_add_operator_token1] = ACTIONS(1647),
    [sym_concat_operator] = ACTIONS(1649),
    [sym_rel_operator] = ACTIONS(1647),
    [anon_sym_AMP_AMP] = ACTIONS(1647),
    [anon_sym_or] = ACTIONS(1647),
    [anon_sym_PIPE_PIPE] = ACTIONS(1647),
    [sym_assign_operator] = ACTIONS(1649),
    [sym__uppercase_identifier] = ACTIONS(1649),
    [aux_sym_tag_token1] = ACTIONS(1649),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(781)] = {
    [sym_attribute] = STATE(781),
    [sym__lowercase_identifier] = ACTIONS(1651),
    [anon_sym_and] = ACTIONS(1651),
    [anon_sym_TILDE] = ACTIONS(1651),
    [anon_sym_QMARK] = ACTIONS(1651),
    [anon_sym_LPAREN] = ACTIONS(1653),
    [anon_sym_RPAREN] = ACTIONS(1653),
    [anon_sym_COMMA] = ACTIONS(1653),
    [anon_sym_PIPE] = ACTIONS(1651),
    [anon_sym_COLON2] = ACTIONS(1651),
    [anon_sym_DOT] = ACTIONS(1653),
    [anon_sym_DASH_GT] = ACTIONS(1651),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_RBRACK] = ACTIONS(1653),
    [anon_sym_true] = ACTIONS(1651),
    [anon_sym_false] = ACTIONS(1651),
    [anon_sym_LBRACE] = ACTIONS(1651),
    [anon_sym_SEMI] = ACTIONS(1653),
    [anon_sym_RBRACE] = ACTIONS(1653),
    [anon_sym_constraint] = ACTIONS(1651),
    [anon_sym_val] = ACTIONS(1651),
    [anon_sym_end] = ACTIONS(1651),
    [anon_sym_with] = ACTIONS(1651),
    [anon_sym_COLON_GT] = ACTIONS(1653),
    [anon_sym_object] = ACTIONS(1651),
    [anon_sym_inherit] = ACTIONS(1651),
    [anon_sym_method] = ACTIONS(1651),
    [anon_sym_in] = ACTIONS(1651),
    [anon_sym_as] = ACTIONS(1651),
    [anon_sym_initializer] = ACTIONS(1651),
    [anon_sym_AMP] = ACTIONS(1651),
    [anon_sym_POUND] = ACTIONS(1651),
    [anon_sym_COLON_COLON] = ACTIONS(1653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1653),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1653),
    [anon_sym_then] = ACTIONS(1651),
    [anon_sym_else] = ACTIONS(1651),
    [anon_sym_do] = ACTIONS(1651),
    [anon_sym_done] = ACTIONS(1651),
    [anon_sym_to] = ACTIONS(1651),
    [anon_sym_downto] = ACTIONS(1651),
    [anon_sym_new] = ACTIONS(1651),
    [anon_sym_LBRACE_LT] = ACTIONS(1653),
    [anon_sym_GT_RBRACE] = ACTIONS(1653),
    [anon_sym_begin] = ACTIONS(1651),
    [sym_ocamlyacc_value] = ACTIONS(1653),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1651),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1653),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1651),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1653),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1653),
    [aux_sym_number_token1] = ACTIONS(1653),
    [aux_sym_signed_number_token1] = ACTIONS(1651),
    [anon_sym_SQUOTE] = ACTIONS(1653),
    [anon_sym_DQUOTE] = ACTIONS(1653),
    [sym_prefix_operator] = ACTIONS(1653),
    [aux_sym_sign_operator_token1] = ACTIONS(1651),
    [sym_hash_operator] = ACTIONS(1653),
    [anon_sym_lsl] = ACTIONS(1651),
    [anon_sym_lsr] = ACTIONS(1651),
    [anon_sym_asr] = ACTIONS(1651),
    [aux_sym_pow_operator_token1] = ACTIONS(1653),
    [anon_sym_mod] = ACTIONS(1651),
    [anon_sym_land] = ACTIONS(1651),
    [anon_sym_lor] = ACTIONS(1651),
    [anon_sym_lxor] = ACTIONS(1651),
    [aux_sym_mult_operator_token1] = ACTIONS(1651),
    [aux_sym_add_operator_token1] = ACTIONS(1651),
    [sym_concat_operator] = ACTIONS(1653),
    [sym_rel_operator] = ACTIONS(1651),
    [anon_sym_AMP_AMP] = ACTIONS(1651),
    [anon_sym_or] = ACTIONS(1651),
    [anon_sym_PIPE_PIPE] = ACTIONS(1651),
    [sym_assign_operator] = ACTIONS(1653),
    [sym__uppercase_identifier] = ACTIONS(1653),
    [aux_sym_tag_token1] = ACTIONS(1653),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(782)] = {
    [sym_attribute] = STATE(782),
    [sym__lowercase_identifier] = ACTIONS(1655),
    [anon_sym_and] = ACTIONS(1655),
    [anon_sym_TILDE] = ACTIONS(1655),
    [anon_sym_QMARK] = ACTIONS(1655),
    [anon_sym_LPAREN] = ACTIONS(1657),
    [anon_sym_RPAREN] = ACTIONS(1657),
    [anon_sym_COMMA] = ACTIONS(1657),
    [anon_sym_PIPE] = ACTIONS(1655),
    [anon_sym_COLON2] = ACTIONS(1655),
    [anon_sym_DOT] = ACTIONS(1657),
    [anon_sym_DASH_GT] = ACTIONS(1655),
    [anon_sym_LBRACK] = ACTIONS(1655),
    [anon_sym_RBRACK] = ACTIONS(1657),
    [anon_sym_true] = ACTIONS(1655),
    [anon_sym_false] = ACTIONS(1655),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_SEMI] = ACTIONS(1657),
    [anon_sym_RBRACE] = ACTIONS(1657),
    [anon_sym_constraint] = ACTIONS(1655),
    [anon_sym_val] = ACTIONS(1655),
    [anon_sym_end] = ACTIONS(1655),
    [anon_sym_with] = ACTIONS(1655),
    [anon_sym_COLON_GT] = ACTIONS(1657),
    [anon_sym_object] = ACTIONS(1655),
    [anon_sym_inherit] = ACTIONS(1655),
    [anon_sym_method] = ACTIONS(1655),
    [anon_sym_in] = ACTIONS(1655),
    [anon_sym_as] = ACTIONS(1655),
    [anon_sym_initializer] = ACTIONS(1655),
    [anon_sym_AMP] = ACTIONS(1655),
    [anon_sym_POUND] = ACTIONS(1655),
    [anon_sym_COLON_COLON] = ACTIONS(1657),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1657),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1657),
    [anon_sym_then] = ACTIONS(1655),
    [anon_sym_else] = ACTIONS(1655),
    [anon_sym_do] = ACTIONS(1655),
    [anon_sym_done] = ACTIONS(1655),
    [anon_sym_to] = ACTIONS(1655),
    [anon_sym_downto] = ACTIONS(1655),
    [anon_sym_new] = ACTIONS(1655),
    [anon_sym_LBRACE_LT] = ACTIONS(1657),
    [anon_sym_GT_RBRACE] = ACTIONS(1657),
    [anon_sym_begin] = ACTIONS(1655),
    [sym_ocamlyacc_value] = ACTIONS(1657),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1655),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1657),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1655),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1655),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1657),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1657),
    [aux_sym_number_token1] = ACTIONS(1657),
    [aux_sym_signed_number_token1] = ACTIONS(1655),
    [anon_sym_SQUOTE] = ACTIONS(1657),
    [anon_sym_DQUOTE] = ACTIONS(1657),
    [sym_prefix_operator] = ACTIONS(1657),
    [aux_sym_sign_operator_token1] = ACTIONS(1655),
    [sym_hash_operator] = ACTIONS(1657),
    [anon_sym_lsl] = ACTIONS(1655),
    [anon_sym_lsr] = ACTIONS(1655),
    [anon_sym_asr] = ACTIONS(1655),
    [aux_sym_pow_operator_token1] = ACTIONS(1657),
    [anon_sym_mod] = ACTIONS(1655),
    [anon_sym_land] = ACTIONS(1655),
    [anon_sym_lor] = ACTIONS(1655),
    [anon_sym_lxor] = ACTIONS(1655),
    [aux_sym_mult_operator_token1] = ACTIONS(1655),
    [aux_sym_add_operator_token1] = ACTIONS(1655),
    [sym_concat_operator] = ACTIONS(1657),
    [sym_rel_operator] = ACTIONS(1655),
    [anon_sym_AMP_AMP] = ACTIONS(1655),
    [anon_sym_or] = ACTIONS(1655),
    [anon_sym_PIPE_PIPE] = ACTIONS(1655),
    [sym_assign_operator] = ACTIONS(1657),
    [sym__uppercase_identifier] = ACTIONS(1657),
    [aux_sym_tag_token1] = ACTIONS(1657),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(783)] = {
    [sym_attribute] = STATE(783),
    [sym__lowercase_identifier] = ACTIONS(1659),
    [anon_sym_and] = ACTIONS(1659),
    [anon_sym_TILDE] = ACTIONS(1659),
    [anon_sym_QMARK] = ACTIONS(1659),
    [anon_sym_LPAREN] = ACTIONS(1661),
    [anon_sym_RPAREN] = ACTIONS(1661),
    [anon_sym_COMMA] = ACTIONS(1661),
    [anon_sym_PIPE] = ACTIONS(1659),
    [anon_sym_COLON2] = ACTIONS(1659),
    [anon_sym_DOT] = ACTIONS(1661),
    [anon_sym_DASH_GT] = ACTIONS(1659),
    [anon_sym_LBRACK] = ACTIONS(1659),
    [anon_sym_RBRACK] = ACTIONS(1661),
    [anon_sym_true] = ACTIONS(1659),
    [anon_sym_false] = ACTIONS(1659),
    [anon_sym_LBRACE] = ACTIONS(1659),
    [anon_sym_SEMI] = ACTIONS(1661),
    [anon_sym_RBRACE] = ACTIONS(1661),
    [anon_sym_constraint] = ACTIONS(1659),
    [anon_sym_val] = ACTIONS(1659),
    [anon_sym_end] = ACTIONS(1659),
    [anon_sym_with] = ACTIONS(1659),
    [anon_sym_COLON_GT] = ACTIONS(1661),
    [anon_sym_object] = ACTIONS(1659),
    [anon_sym_inherit] = ACTIONS(1659),
    [anon_sym_method] = ACTIONS(1659),
    [anon_sym_in] = ACTIONS(1659),
    [anon_sym_as] = ACTIONS(1659),
    [anon_sym_initializer] = ACTIONS(1659),
    [anon_sym_AMP] = ACTIONS(1659),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_COLON_COLON] = ACTIONS(1661),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1661),
    [anon_sym_then] = ACTIONS(1659),
    [anon_sym_else] = ACTIONS(1659),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_done] = ACTIONS(1659),
    [anon_sym_to] = ACTIONS(1659),
    [anon_sym_downto] = ACTIONS(1659),
    [anon_sym_new] = ACTIONS(1659),
    [anon_sym_LBRACE_LT] = ACTIONS(1661),
    [anon_sym_GT_RBRACE] = ACTIONS(1661),
    [anon_sym_begin] = ACTIONS(1659),
    [sym_ocamlyacc_value] = ACTIONS(1661),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1659),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1661),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1659),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1659),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1661),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1661),
    [aux_sym_number_token1] = ACTIONS(1661),
    [aux_sym_signed_number_token1] = ACTIONS(1659),
    [anon_sym_SQUOTE] = ACTIONS(1661),
    [anon_sym_DQUOTE] = ACTIONS(1661),
    [sym_prefix_operator] = ACTIONS(1661),
    [aux_sym_sign_operator_token1] = ACTIONS(1659),
    [sym_hash_operator] = ACTIONS(1661),
    [anon_sym_lsl] = ACTIONS(1659),
    [anon_sym_lsr] = ACTIONS(1659),
    [anon_sym_asr] = ACTIONS(1659),
    [aux_sym_pow_operator_token1] = ACTIONS(1661),
    [anon_sym_mod] = ACTIONS(1659),
    [anon_sym_land] = ACTIONS(1659),
    [anon_sym_lor] = ACTIONS(1659),
    [anon_sym_lxor] = ACTIONS(1659),
    [aux_sym_mult_operator_token1] = ACTIONS(1659),
    [aux_sym_add_operator_token1] = ACTIONS(1659),
    [sym_concat_operator] = ACTIONS(1661),
    [sym_rel_operator] = ACTIONS(1659),
    [anon_sym_AMP_AMP] = ACTIONS(1659),
    [anon_sym_or] = ACTIONS(1659),
    [anon_sym_PIPE_PIPE] = ACTIONS(1659),
    [sym_assign_operator] = ACTIONS(1661),
    [sym__uppercase_identifier] = ACTIONS(1661),
    [aux_sym_tag_token1] = ACTIONS(1661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(784)] = {
    [sym_attribute] = STATE(784),
    [sym__lowercase_identifier] = ACTIONS(1663),
    [anon_sym_and] = ACTIONS(1663),
    [anon_sym_TILDE] = ACTIONS(1663),
    [anon_sym_QMARK] = ACTIONS(1663),
    [anon_sym_LPAREN] = ACTIONS(1665),
    [anon_sym_RPAREN] = ACTIONS(1665),
    [anon_sym_COMMA] = ACTIONS(1665),
    [anon_sym_PIPE] = ACTIONS(1663),
    [anon_sym_COLON2] = ACTIONS(1663),
    [anon_sym_DOT] = ACTIONS(751),
    [anon_sym_DASH_GT] = ACTIONS(1663),
    [anon_sym_LBRACK] = ACTIONS(1663),
    [anon_sym_RBRACK] = ACTIONS(1665),
    [anon_sym_true] = ACTIONS(1663),
    [anon_sym_false] = ACTIONS(1663),
    [anon_sym_LBRACE] = ACTIONS(1663),
    [anon_sym_SEMI] = ACTIONS(1665),
    [anon_sym_RBRACE] = ACTIONS(1665),
    [anon_sym_constraint] = ACTIONS(1663),
    [anon_sym_val] = ACTIONS(1663),
    [anon_sym_end] = ACTIONS(1663),
    [anon_sym_with] = ACTIONS(1663),
    [anon_sym_COLON_GT] = ACTIONS(1665),
    [anon_sym_object] = ACTIONS(1663),
    [anon_sym_inherit] = ACTIONS(1663),
    [anon_sym_method] = ACTIONS(1663),
    [anon_sym_in] = ACTIONS(1663),
    [anon_sym_as] = ACTIONS(1663),
    [anon_sym_initializer] = ACTIONS(1663),
    [anon_sym_AMP] = ACTIONS(1663),
    [anon_sym_POUND] = ACTIONS(753),
    [anon_sym_COLON_COLON] = ACTIONS(1665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1665),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1665),
    [anon_sym_then] = ACTIONS(1663),
    [anon_sym_else] = ACTIONS(1663),
    [anon_sym_do] = ACTIONS(1663),
    [anon_sym_done] = ACTIONS(1663),
    [anon_sym_to] = ACTIONS(1663),
    [anon_sym_downto] = ACTIONS(1663),
    [anon_sym_new] = ACTIONS(1663),
    [anon_sym_LBRACE_LT] = ACTIONS(1665),
    [anon_sym_GT_RBRACE] = ACTIONS(1665),
    [anon_sym_begin] = ACTIONS(1663),
    [sym_ocamlyacc_value] = ACTIONS(1665),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1663),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1665),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1663),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1663),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1665),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1665),
    [aux_sym_number_token1] = ACTIONS(1665),
    [aux_sym_signed_number_token1] = ACTIONS(1663),
    [anon_sym_SQUOTE] = ACTIONS(1665),
    [anon_sym_DQUOTE] = ACTIONS(1665),
    [sym_prefix_operator] = ACTIONS(1665),
    [aux_sym_sign_operator_token1] = ACTIONS(1663),
    [sym_hash_operator] = ACTIONS(755),
    [anon_sym_lsl] = ACTIONS(1663),
    [anon_sym_lsr] = ACTIONS(1663),
    [anon_sym_asr] = ACTIONS(1663),
    [aux_sym_pow_operator_token1] = ACTIONS(1665),
    [anon_sym_mod] = ACTIONS(1663),
    [anon_sym_land] = ACTIONS(1663),
    [anon_sym_lor] = ACTIONS(1663),
    [anon_sym_lxor] = ACTIONS(1663),
    [aux_sym_mult_operator_token1] = ACTIONS(1663),
    [aux_sym_add_operator_token1] = ACTIONS(1663),
    [sym_concat_operator] = ACTIONS(1665),
    [sym_rel_operator] = ACTIONS(1663),
    [anon_sym_AMP_AMP] = ACTIONS(1663),
    [anon_sym_or] = ACTIONS(1663),
    [anon_sym_PIPE_PIPE] = ACTIONS(1663),
    [sym_assign_operator] = ACTIONS(1665),
    [sym__uppercase_identifier] = ACTIONS(1665),
    [aux_sym_tag_token1] = ACTIONS(1665),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(785)] = {
    [sym_attribute] = STATE(785),
    [sym__lowercase_identifier] = ACTIONS(1667),
    [anon_sym_and] = ACTIONS(1667),
    [anon_sym_TILDE] = ACTIONS(1667),
    [anon_sym_QMARK] = ACTIONS(1667),
    [anon_sym_LPAREN] = ACTIONS(1669),
    [anon_sym_RPAREN] = ACTIONS(1669),
    [anon_sym_COMMA] = ACTIONS(1669),
    [anon_sym_PIPE] = ACTIONS(1667),
    [anon_sym_COLON2] = ACTIONS(1667),
    [anon_sym_DOT] = ACTIONS(1669),
    [anon_sym_DASH_GT] = ACTIONS(1667),
    [anon_sym_LBRACK] = ACTIONS(1667),
    [anon_sym_RBRACK] = ACTIONS(1669),
    [anon_sym_true] = ACTIONS(1667),
    [anon_sym_false] = ACTIONS(1667),
    [anon_sym_LBRACE] = ACTIONS(1667),
    [anon_sym_SEMI] = ACTIONS(1669),
    [anon_sym_RBRACE] = ACTIONS(1669),
    [anon_sym_constraint] = ACTIONS(1667),
    [anon_sym_val] = ACTIONS(1667),
    [anon_sym_end] = ACTIONS(1667),
    [anon_sym_with] = ACTIONS(1667),
    [anon_sym_COLON_GT] = ACTIONS(1669),
    [anon_sym_object] = ACTIONS(1667),
    [anon_sym_inherit] = ACTIONS(1667),
    [anon_sym_method] = ACTIONS(1667),
    [anon_sym_in] = ACTIONS(1667),
    [anon_sym_as] = ACTIONS(1667),
    [anon_sym_initializer] = ACTIONS(1667),
    [anon_sym_AMP] = ACTIONS(1667),
    [anon_sym_POUND] = ACTIONS(1667),
    [anon_sym_COLON_COLON] = ACTIONS(1669),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1669),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1669),
    [anon_sym_then] = ACTIONS(1667),
    [anon_sym_else] = ACTIONS(1667),
    [anon_sym_do] = ACTIONS(1667),
    [anon_sym_done] = ACTIONS(1667),
    [anon_sym_to] = ACTIONS(1667),
    [anon_sym_downto] = ACTIONS(1667),
    [anon_sym_new] = ACTIONS(1667),
    [anon_sym_LBRACE_LT] = ACTIONS(1669),
    [anon_sym_GT_RBRACE] = ACTIONS(1669),
    [anon_sym_begin] = ACTIONS(1667),
    [sym_ocamlyacc_value] = ACTIONS(1669),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1667),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1669),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1667),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1667),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1669),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1669),
    [aux_sym_number_token1] = ACTIONS(1669),
    [aux_sym_signed_number_token1] = ACTIONS(1667),
    [anon_sym_SQUOTE] = ACTIONS(1669),
    [anon_sym_DQUOTE] = ACTIONS(1669),
    [sym_prefix_operator] = ACTIONS(1669),
    [aux_sym_sign_operator_token1] = ACTIONS(1667),
    [sym_hash_operator] = ACTIONS(1669),
    [anon_sym_lsl] = ACTIONS(1667),
    [anon_sym_lsr] = ACTIONS(1667),
    [anon_sym_asr] = ACTIONS(1667),
    [aux_sym_pow_operator_token1] = ACTIONS(1669),
    [anon_sym_mod] = ACTIONS(1667),
    [anon_sym_land] = ACTIONS(1667),
    [anon_sym_lor] = ACTIONS(1667),
    [anon_sym_lxor] = ACTIONS(1667),
    [aux_sym_mult_operator_token1] = ACTIONS(1667),
    [aux_sym_add_operator_token1] = ACTIONS(1667),
    [sym_concat_operator] = ACTIONS(1669),
    [sym_rel_operator] = ACTIONS(1667),
    [anon_sym_AMP_AMP] = ACTIONS(1667),
    [anon_sym_or] = ACTIONS(1667),
    [anon_sym_PIPE_PIPE] = ACTIONS(1667),
    [sym_assign_operator] = ACTIONS(1669),
    [sym__uppercase_identifier] = ACTIONS(1669),
    [aux_sym_tag_token1] = ACTIONS(1669),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(786)] = {
    [sym_attribute] = STATE(786),
    [sym__lowercase_identifier] = ACTIONS(1671),
    [anon_sym_and] = ACTIONS(1671),
    [anon_sym_TILDE] = ACTIONS(1671),
    [anon_sym_QMARK] = ACTIONS(1671),
    [anon_sym_LPAREN] = ACTIONS(1673),
    [anon_sym_RPAREN] = ACTIONS(1673),
    [anon_sym_COMMA] = ACTIONS(1673),
    [anon_sym_PIPE] = ACTIONS(1671),
    [anon_sym_COLON2] = ACTIONS(1671),
    [anon_sym_DOT] = ACTIONS(1673),
    [anon_sym_DASH_GT] = ACTIONS(1671),
    [anon_sym_LBRACK] = ACTIONS(1671),
    [anon_sym_RBRACK] = ACTIONS(1673),
    [anon_sym_true] = ACTIONS(1671),
    [anon_sym_false] = ACTIONS(1671),
    [anon_sym_LBRACE] = ACTIONS(1671),
    [anon_sym_SEMI] = ACTIONS(1673),
    [anon_sym_RBRACE] = ACTIONS(1673),
    [anon_sym_constraint] = ACTIONS(1671),
    [anon_sym_val] = ACTIONS(1671),
    [anon_sym_end] = ACTIONS(1671),
    [anon_sym_with] = ACTIONS(1671),
    [anon_sym_COLON_GT] = ACTIONS(1673),
    [anon_sym_object] = ACTIONS(1671),
    [anon_sym_inherit] = ACTIONS(1671),
    [anon_sym_method] = ACTIONS(1671),
    [anon_sym_in] = ACTIONS(1671),
    [anon_sym_as] = ACTIONS(1671),
    [anon_sym_initializer] = ACTIONS(1671),
    [anon_sym_AMP] = ACTIONS(1671),
    [anon_sym_POUND] = ACTIONS(1671),
    [anon_sym_COLON_COLON] = ACTIONS(1673),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1673),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1673),
    [anon_sym_then] = ACTIONS(1671),
    [anon_sym_else] = ACTIONS(1671),
    [anon_sym_do] = ACTIONS(1671),
    [anon_sym_done] = ACTIONS(1671),
    [anon_sym_to] = ACTIONS(1671),
    [anon_sym_downto] = ACTIONS(1671),
    [anon_sym_new] = ACTIONS(1671),
    [anon_sym_LBRACE_LT] = ACTIONS(1673),
    [anon_sym_GT_RBRACE] = ACTIONS(1673),
    [anon_sym_begin] = ACTIONS(1671),
    [sym_ocamlyacc_value] = ACTIONS(1673),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1671),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1673),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1671),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1671),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1673),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1673),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1671),
    [anon_sym_SQUOTE] = ACTIONS(1673),
    [anon_sym_DQUOTE] = ACTIONS(1673),
    [sym_prefix_operator] = ACTIONS(1673),
    [aux_sym_sign_operator_token1] = ACTIONS(1671),
    [sym_hash_operator] = ACTIONS(1673),
    [anon_sym_lsl] = ACTIONS(1671),
    [anon_sym_lsr] = ACTIONS(1671),
    [anon_sym_asr] = ACTIONS(1671),
    [aux_sym_pow_operator_token1] = ACTIONS(1673),
    [anon_sym_mod] = ACTIONS(1671),
    [anon_sym_land] = ACTIONS(1671),
    [anon_sym_lor] = ACTIONS(1671),
    [anon_sym_lxor] = ACTIONS(1671),
    [aux_sym_mult_operator_token1] = ACTIONS(1671),
    [aux_sym_add_operator_token1] = ACTIONS(1671),
    [sym_concat_operator] = ACTIONS(1673),
    [sym_rel_operator] = ACTIONS(1671),
    [anon_sym_AMP_AMP] = ACTIONS(1671),
    [anon_sym_or] = ACTIONS(1671),
    [anon_sym_PIPE_PIPE] = ACTIONS(1671),
    [sym_assign_operator] = ACTIONS(1673),
    [sym__uppercase_identifier] = ACTIONS(1673),
    [aux_sym_tag_token1] = ACTIONS(1673),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(787)] = {
    [sym_attribute] = STATE(787),
    [sym__lowercase_identifier] = ACTIONS(1675),
    [anon_sym_and] = ACTIONS(1675),
    [anon_sym_TILDE] = ACTIONS(1675),
    [anon_sym_QMARK] = ACTIONS(1675),
    [anon_sym_LPAREN] = ACTIONS(1677),
    [anon_sym_RPAREN] = ACTIONS(1677),
    [anon_sym_COMMA] = ACTIONS(1677),
    [anon_sym_PIPE] = ACTIONS(1675),
    [anon_sym_COLON2] = ACTIONS(1675),
    [anon_sym_DOT] = ACTIONS(1677),
    [anon_sym_DASH_GT] = ACTIONS(1675),
    [anon_sym_LBRACK] = ACTIONS(1675),
    [anon_sym_RBRACK] = ACTIONS(1677),
    [anon_sym_true] = ACTIONS(1675),
    [anon_sym_false] = ACTIONS(1675),
    [anon_sym_LBRACE] = ACTIONS(1675),
    [anon_sym_SEMI] = ACTIONS(1677),
    [anon_sym_RBRACE] = ACTIONS(1677),
    [anon_sym_constraint] = ACTIONS(1675),
    [anon_sym_val] = ACTIONS(1675),
    [anon_sym_end] = ACTIONS(1675),
    [anon_sym_with] = ACTIONS(1675),
    [anon_sym_COLON_GT] = ACTIONS(1677),
    [anon_sym_object] = ACTIONS(1675),
    [anon_sym_inherit] = ACTIONS(1675),
    [anon_sym_method] = ACTIONS(1675),
    [anon_sym_in] = ACTIONS(1675),
    [anon_sym_as] = ACTIONS(1675),
    [anon_sym_initializer] = ACTIONS(1675),
    [anon_sym_AMP] = ACTIONS(1675),
    [anon_sym_POUND] = ACTIONS(1675),
    [anon_sym_COLON_COLON] = ACTIONS(1677),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1677),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1677),
    [anon_sym_then] = ACTIONS(1675),
    [anon_sym_else] = ACTIONS(1675),
    [anon_sym_do] = ACTIONS(1675),
    [anon_sym_done] = ACTIONS(1675),
    [anon_sym_to] = ACTIONS(1675),
    [anon_sym_downto] = ACTIONS(1675),
    [anon_sym_new] = ACTIONS(1675),
    [anon_sym_LBRACE_LT] = ACTIONS(1677),
    [anon_sym_GT_RBRACE] = ACTIONS(1677),
    [anon_sym_begin] = ACTIONS(1675),
    [sym_ocamlyacc_value] = ACTIONS(1677),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1675),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1677),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1675),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1675),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1677),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1677),
    [aux_sym_number_token1] = ACTIONS(1677),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1677),
    [sym_prefix_operator] = ACTIONS(1677),
    [aux_sym_sign_operator_token1] = ACTIONS(1675),
    [sym_hash_operator] = ACTIONS(1677),
    [anon_sym_lsl] = ACTIONS(1675),
    [anon_sym_lsr] = ACTIONS(1675),
    [anon_sym_asr] = ACTIONS(1675),
    [aux_sym_pow_operator_token1] = ACTIONS(1677),
    [anon_sym_mod] = ACTIONS(1675),
    [anon_sym_land] = ACTIONS(1675),
    [anon_sym_lor] = ACTIONS(1675),
    [anon_sym_lxor] = ACTIONS(1675),
    [aux_sym_mult_operator_token1] = ACTIONS(1675),
    [aux_sym_add_operator_token1] = ACTIONS(1675),
    [sym_concat_operator] = ACTIONS(1677),
    [sym_rel_operator] = ACTIONS(1675),
    [anon_sym_AMP_AMP] = ACTIONS(1675),
    [anon_sym_or] = ACTIONS(1675),
    [anon_sym_PIPE_PIPE] = ACTIONS(1675),
    [sym_assign_operator] = ACTIONS(1677),
    [sym__uppercase_identifier] = ACTIONS(1677),
    [aux_sym_tag_token1] = ACTIONS(1677),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(788)] = {
    [sym_attribute] = STATE(788),
    [sym__lowercase_identifier] = ACTIONS(1679),
    [anon_sym_and] = ACTIONS(1679),
    [anon_sym_TILDE] = ACTIONS(1679),
    [anon_sym_QMARK] = ACTIONS(1679),
    [anon_sym_LPAREN] = ACTIONS(1681),
    [anon_sym_RPAREN] = ACTIONS(1681),
    [anon_sym_COMMA] = ACTIONS(1681),
    [anon_sym_PIPE] = ACTIONS(1679),
    [anon_sym_COLON2] = ACTIONS(1679),
    [anon_sym_DOT] = ACTIONS(1681),
    [anon_sym_DASH_GT] = ACTIONS(1679),
    [anon_sym_LBRACK] = ACTIONS(1679),
    [anon_sym_RBRACK] = ACTIONS(1681),
    [anon_sym_true] = ACTIONS(1679),
    [anon_sym_false] = ACTIONS(1679),
    [anon_sym_LBRACE] = ACTIONS(1679),
    [anon_sym_SEMI] = ACTIONS(1681),
    [anon_sym_RBRACE] = ACTIONS(1681),
    [anon_sym_constraint] = ACTIONS(1679),
    [anon_sym_val] = ACTIONS(1679),
    [anon_sym_end] = ACTIONS(1679),
    [anon_sym_with] = ACTIONS(1679),
    [anon_sym_COLON_GT] = ACTIONS(1681),
    [anon_sym_object] = ACTIONS(1679),
    [anon_sym_inherit] = ACTIONS(1679),
    [anon_sym_method] = ACTIONS(1679),
    [anon_sym_in] = ACTIONS(1679),
    [anon_sym_as] = ACTIONS(1679),
    [anon_sym_initializer] = ACTIONS(1679),
    [anon_sym_AMP] = ACTIONS(1679),
    [anon_sym_POUND] = ACTIONS(1679),
    [anon_sym_COLON_COLON] = ACTIONS(1681),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1681),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1681),
    [anon_sym_then] = ACTIONS(1679),
    [anon_sym_else] = ACTIONS(1679),
    [anon_sym_do] = ACTIONS(1679),
    [anon_sym_done] = ACTIONS(1679),
    [anon_sym_to] = ACTIONS(1679),
    [anon_sym_downto] = ACTIONS(1679),
    [anon_sym_new] = ACTIONS(1679),
    [anon_sym_LBRACE_LT] = ACTIONS(1681),
    [anon_sym_GT_RBRACE] = ACTIONS(1681),
    [anon_sym_begin] = ACTIONS(1679),
    [sym_ocamlyacc_value] = ACTIONS(1681),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1679),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1681),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1679),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1679),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1681),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1681),
    [aux_sym_number_token1] = ACTIONS(1681),
    [aux_sym_signed_number_token1] = ACTIONS(1679),
    [anon_sym_SQUOTE] = ACTIONS(1681),
    [anon_sym_DQUOTE] = ACTIONS(1681),
    [sym_prefix_operator] = ACTIONS(1681),
    [aux_sym_sign_operator_token1] = ACTIONS(1679),
    [sym_hash_operator] = ACTIONS(1681),
    [anon_sym_lsl] = ACTIONS(1679),
    [anon_sym_lsr] = ACTIONS(1679),
    [anon_sym_asr] = ACTIONS(1679),
    [aux_sym_pow_operator_token1] = ACTIONS(1681),
    [anon_sym_mod] = ACTIONS(1679),
    [anon_sym_land] = ACTIONS(1679),
    [anon_sym_lor] = ACTIONS(1679),
    [anon_sym_lxor] = ACTIONS(1679),
    [aux_sym_mult_operator_token1] = ACTIONS(1679),
    [aux_sym_add_operator_token1] = ACTIONS(1679),
    [sym_concat_operator] = ACTIONS(1681),
    [sym_rel_operator] = ACTIONS(1679),
    [anon_sym_AMP_AMP] = ACTIONS(1679),
    [anon_sym_or] = ACTIONS(1679),
    [anon_sym_PIPE_PIPE] = ACTIONS(1679),
    [sym_assign_operator] = ACTIONS(1681),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_tag_token1] = ACTIONS(1681),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(789)] = {
    [sym_attribute] = STATE(789),
    [sym__lowercase_identifier] = ACTIONS(1683),
    [anon_sym_and] = ACTIONS(1683),
    [anon_sym_TILDE] = ACTIONS(1683),
    [anon_sym_QMARK] = ACTIONS(1683),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_RPAREN] = ACTIONS(1685),
    [anon_sym_COMMA] = ACTIONS(1685),
    [anon_sym_PIPE] = ACTIONS(1683),
    [anon_sym_COLON2] = ACTIONS(1683),
    [anon_sym_DOT] = ACTIONS(1685),
    [anon_sym_DASH_GT] = ACTIONS(1683),
    [anon_sym_LBRACK] = ACTIONS(1683),
    [anon_sym_RBRACK] = ACTIONS(1685),
    [anon_sym_true] = ACTIONS(1683),
    [anon_sym_false] = ACTIONS(1683),
    [anon_sym_LBRACE] = ACTIONS(1683),
    [anon_sym_SEMI] = ACTIONS(1685),
    [anon_sym_RBRACE] = ACTIONS(1685),
    [anon_sym_constraint] = ACTIONS(1683),
    [anon_sym_val] = ACTIONS(1683),
    [anon_sym_end] = ACTIONS(1683),
    [anon_sym_with] = ACTIONS(1683),
    [anon_sym_COLON_GT] = ACTIONS(1685),
    [anon_sym_object] = ACTIONS(1683),
    [anon_sym_inherit] = ACTIONS(1683),
    [anon_sym_method] = ACTIONS(1683),
    [anon_sym_in] = ACTIONS(1683),
    [anon_sym_as] = ACTIONS(1683),
    [anon_sym_initializer] = ACTIONS(1683),
    [anon_sym_AMP] = ACTIONS(1683),
    [anon_sym_POUND] = ACTIONS(1683),
    [anon_sym_COLON_COLON] = ACTIONS(1685),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1685),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1685),
    [anon_sym_then] = ACTIONS(1683),
    [anon_sym_else] = ACTIONS(1683),
    [anon_sym_do] = ACTIONS(1683),
    [anon_sym_done] = ACTIONS(1683),
    [anon_sym_to] = ACTIONS(1683),
    [anon_sym_downto] = ACTIONS(1683),
    [anon_sym_new] = ACTIONS(1683),
    [anon_sym_LBRACE_LT] = ACTIONS(1685),
    [anon_sym_GT_RBRACE] = ACTIONS(1685),
    [anon_sym_begin] = ACTIONS(1683),
    [sym_ocamlyacc_value] = ACTIONS(1685),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1683),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1685),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1683),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1683),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1685),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1685),
    [aux_sym_number_token1] = ACTIONS(1685),
    [aux_sym_signed_number_token1] = ACTIONS(1683),
    [anon_sym_SQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1685),
    [sym_prefix_operator] = ACTIONS(1685),
    [aux_sym_sign_operator_token1] = ACTIONS(1683),
    [sym_hash_operator] = ACTIONS(1685),
    [anon_sym_lsl] = ACTIONS(1683),
    [anon_sym_lsr] = ACTIONS(1683),
    [anon_sym_asr] = ACTIONS(1683),
    [aux_sym_pow_operator_token1] = ACTIONS(1685),
    [anon_sym_mod] = ACTIONS(1683),
    [anon_sym_land] = ACTIONS(1683),
    [anon_sym_lor] = ACTIONS(1683),
    [anon_sym_lxor] = ACTIONS(1683),
    [aux_sym_mult_operator_token1] = ACTIONS(1683),
    [aux_sym_add_operator_token1] = ACTIONS(1683),
    [sym_concat_operator] = ACTIONS(1685),
    [sym_rel_operator] = ACTIONS(1683),
    [anon_sym_AMP_AMP] = ACTIONS(1683),
    [anon_sym_or] = ACTIONS(1683),
    [anon_sym_PIPE_PIPE] = ACTIONS(1683),
    [sym_assign_operator] = ACTIONS(1685),
    [sym__uppercase_identifier] = ACTIONS(1685),
    [aux_sym_tag_token1] = ACTIONS(1685),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(790)] = {
    [sym_attribute] = STATE(790),
    [sym__lowercase_identifier] = ACTIONS(1687),
    [anon_sym_and] = ACTIONS(1687),
    [anon_sym_TILDE] = ACTIONS(1687),
    [anon_sym_QMARK] = ACTIONS(1687),
    [anon_sym_LPAREN] = ACTIONS(1689),
    [anon_sym_RPAREN] = ACTIONS(1689),
    [anon_sym_COMMA] = ACTIONS(1689),
    [anon_sym_PIPE] = ACTIONS(1687),
    [anon_sym_COLON2] = ACTIONS(1687),
    [anon_sym_DOT] = ACTIONS(1689),
    [anon_sym_DASH_GT] = ACTIONS(1687),
    [anon_sym_LBRACK] = ACTIONS(1687),
    [anon_sym_RBRACK] = ACTIONS(1689),
    [anon_sym_true] = ACTIONS(1687),
    [anon_sym_false] = ACTIONS(1687),
    [anon_sym_LBRACE] = ACTIONS(1687),
    [anon_sym_SEMI] = ACTIONS(1689),
    [anon_sym_RBRACE] = ACTIONS(1689),
    [anon_sym_constraint] = ACTIONS(1687),
    [anon_sym_val] = ACTIONS(1687),
    [anon_sym_end] = ACTIONS(1687),
    [anon_sym_with] = ACTIONS(1687),
    [anon_sym_COLON_GT] = ACTIONS(1689),
    [anon_sym_object] = ACTIONS(1687),
    [anon_sym_inherit] = ACTIONS(1687),
    [anon_sym_method] = ACTIONS(1687),
    [anon_sym_in] = ACTIONS(1687),
    [anon_sym_as] = ACTIONS(1687),
    [anon_sym_initializer] = ACTIONS(1687),
    [anon_sym_AMP] = ACTIONS(1687),
    [anon_sym_POUND] = ACTIONS(1687),
    [anon_sym_COLON_COLON] = ACTIONS(1689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1689),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1689),
    [anon_sym_then] = ACTIONS(1687),
    [anon_sym_else] = ACTIONS(1687),
    [anon_sym_do] = ACTIONS(1687),
    [anon_sym_done] = ACTIONS(1687),
    [anon_sym_to] = ACTIONS(1687),
    [anon_sym_downto] = ACTIONS(1687),
    [anon_sym_new] = ACTIONS(1687),
    [anon_sym_LBRACE_LT] = ACTIONS(1689),
    [anon_sym_GT_RBRACE] = ACTIONS(1689),
    [anon_sym_begin] = ACTIONS(1687),
    [sym_ocamlyacc_value] = ACTIONS(1689),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1687),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1689),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1687),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1687),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1689),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1689),
    [aux_sym_number_token1] = ACTIONS(1689),
    [aux_sym_signed_number_token1] = ACTIONS(1687),
    [anon_sym_SQUOTE] = ACTIONS(1689),
    [anon_sym_DQUOTE] = ACTIONS(1689),
    [sym_prefix_operator] = ACTIONS(1689),
    [aux_sym_sign_operator_token1] = ACTIONS(1687),
    [sym_hash_operator] = ACTIONS(1689),
    [anon_sym_lsl] = ACTIONS(1687),
    [anon_sym_lsr] = ACTIONS(1687),
    [anon_sym_asr] = ACTIONS(1687),
    [aux_sym_pow_operator_token1] = ACTIONS(1689),
    [anon_sym_mod] = ACTIONS(1687),
    [anon_sym_land] = ACTIONS(1687),
    [anon_sym_lor] = ACTIONS(1687),
    [anon_sym_lxor] = ACTIONS(1687),
    [aux_sym_mult_operator_token1] = ACTIONS(1687),
    [aux_sym_add_operator_token1] = ACTIONS(1687),
    [sym_concat_operator] = ACTIONS(1689),
    [sym_rel_operator] = ACTIONS(1687),
    [anon_sym_AMP_AMP] = ACTIONS(1687),
    [anon_sym_or] = ACTIONS(1687),
    [anon_sym_PIPE_PIPE] = ACTIONS(1687),
    [sym_assign_operator] = ACTIONS(1689),
    [sym__uppercase_identifier] = ACTIONS(1689),
    [aux_sym_tag_token1] = ACTIONS(1689),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(791)] = {
    [sym_attribute] = STATE(791),
    [sym__lowercase_identifier] = ACTIONS(1691),
    [anon_sym_and] = ACTIONS(1691),
    [anon_sym_TILDE] = ACTIONS(1691),
    [anon_sym_QMARK] = ACTIONS(1691),
    [anon_sym_LPAREN] = ACTIONS(1693),
    [anon_sym_RPAREN] = ACTIONS(1693),
    [anon_sym_COMMA] = ACTIONS(1693),
    [anon_sym_PIPE] = ACTIONS(1691),
    [anon_sym_COLON2] = ACTIONS(1691),
    [anon_sym_DOT] = ACTIONS(1693),
    [anon_sym_DASH_GT] = ACTIONS(1691),
    [anon_sym_LBRACK] = ACTIONS(1691),
    [anon_sym_RBRACK] = ACTIONS(1693),
    [anon_sym_true] = ACTIONS(1691),
    [anon_sym_false] = ACTIONS(1691),
    [anon_sym_LBRACE] = ACTIONS(1691),
    [anon_sym_SEMI] = ACTIONS(1693),
    [anon_sym_RBRACE] = ACTIONS(1693),
    [anon_sym_constraint] = ACTIONS(1691),
    [anon_sym_val] = ACTIONS(1691),
    [anon_sym_end] = ACTIONS(1691),
    [anon_sym_with] = ACTIONS(1691),
    [anon_sym_COLON_GT] = ACTIONS(1693),
    [anon_sym_object] = ACTIONS(1691),
    [anon_sym_inherit] = ACTIONS(1691),
    [anon_sym_method] = ACTIONS(1691),
    [anon_sym_in] = ACTIONS(1691),
    [anon_sym_as] = ACTIONS(1691),
    [anon_sym_initializer] = ACTIONS(1691),
    [anon_sym_AMP] = ACTIONS(1691),
    [anon_sym_POUND] = ACTIONS(1691),
    [anon_sym_COLON_COLON] = ACTIONS(1693),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1693),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1693),
    [anon_sym_then] = ACTIONS(1691),
    [anon_sym_else] = ACTIONS(1691),
    [anon_sym_do] = ACTIONS(1691),
    [anon_sym_done] = ACTIONS(1691),
    [anon_sym_to] = ACTIONS(1691),
    [anon_sym_downto] = ACTIONS(1691),
    [anon_sym_new] = ACTIONS(1691),
    [anon_sym_LBRACE_LT] = ACTIONS(1693),
    [anon_sym_GT_RBRACE] = ACTIONS(1693),
    [anon_sym_begin] = ACTIONS(1691),
    [sym_ocamlyacc_value] = ACTIONS(1693),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1691),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1693),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1691),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1691),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1693),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1693),
    [aux_sym_number_token1] = ACTIONS(1693),
    [aux_sym_signed_number_token1] = ACTIONS(1691),
    [anon_sym_SQUOTE] = ACTIONS(1693),
    [anon_sym_DQUOTE] = ACTIONS(1693),
    [sym_prefix_operator] = ACTIONS(1693),
    [aux_sym_sign_operator_token1] = ACTIONS(1691),
    [sym_hash_operator] = ACTIONS(1693),
    [anon_sym_lsl] = ACTIONS(1691),
    [anon_sym_lsr] = ACTIONS(1691),
    [anon_sym_asr] = ACTIONS(1691),
    [aux_sym_pow_operator_token1] = ACTIONS(1693),
    [anon_sym_mod] = ACTIONS(1691),
    [anon_sym_land] = ACTIONS(1691),
    [anon_sym_lor] = ACTIONS(1691),
    [anon_sym_lxor] = ACTIONS(1691),
    [aux_sym_mult_operator_token1] = ACTIONS(1691),
    [aux_sym_add_operator_token1] = ACTIONS(1691),
    [sym_concat_operator] = ACTIONS(1693),
    [sym_rel_operator] = ACTIONS(1691),
    [anon_sym_AMP_AMP] = ACTIONS(1691),
    [anon_sym_or] = ACTIONS(1691),
    [anon_sym_PIPE_PIPE] = ACTIONS(1691),
    [sym_assign_operator] = ACTIONS(1693),
    [sym__uppercase_identifier] = ACTIONS(1693),
    [aux_sym_tag_token1] = ACTIONS(1693),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(792)] = {
    [sym_attribute] = STATE(792),
    [sym__lowercase_identifier] = ACTIONS(1695),
    [anon_sym_and] = ACTIONS(1695),
    [anon_sym_TILDE] = ACTIONS(1695),
    [anon_sym_QMARK] = ACTIONS(1695),
    [anon_sym_LPAREN] = ACTIONS(1697),
    [anon_sym_RPAREN] = ACTIONS(1697),
    [anon_sym_COMMA] = ACTIONS(1697),
    [anon_sym_PIPE] = ACTIONS(1695),
    [anon_sym_COLON2] = ACTIONS(1695),
    [anon_sym_DOT] = ACTIONS(1697),
    [anon_sym_DASH_GT] = ACTIONS(1695),
    [anon_sym_LBRACK] = ACTIONS(1695),
    [anon_sym_RBRACK] = ACTIONS(1697),
    [anon_sym_true] = ACTIONS(1695),
    [anon_sym_false] = ACTIONS(1695),
    [anon_sym_LBRACE] = ACTIONS(1695),
    [anon_sym_SEMI] = ACTIONS(1697),
    [anon_sym_RBRACE] = ACTIONS(1697),
    [anon_sym_constraint] = ACTIONS(1695),
    [anon_sym_val] = ACTIONS(1695),
    [anon_sym_end] = ACTIONS(1695),
    [anon_sym_with] = ACTIONS(1695),
    [anon_sym_COLON_GT] = ACTIONS(1697),
    [anon_sym_object] = ACTIONS(1695),
    [anon_sym_inherit] = ACTIONS(1695),
    [anon_sym_method] = ACTIONS(1695),
    [anon_sym_in] = ACTIONS(1695),
    [anon_sym_as] = ACTIONS(1695),
    [anon_sym_initializer] = ACTIONS(1695),
    [anon_sym_AMP] = ACTIONS(1695),
    [anon_sym_POUND] = ACTIONS(1695),
    [anon_sym_COLON_COLON] = ACTIONS(1697),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1697),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1697),
    [anon_sym_then] = ACTIONS(1695),
    [anon_sym_else] = ACTIONS(1695),
    [anon_sym_do] = ACTIONS(1695),
    [anon_sym_done] = ACTIONS(1695),
    [anon_sym_to] = ACTIONS(1695),
    [anon_sym_downto] = ACTIONS(1695),
    [anon_sym_new] = ACTIONS(1695),
    [anon_sym_LBRACE_LT] = ACTIONS(1697),
    [anon_sym_GT_RBRACE] = ACTIONS(1697),
    [anon_sym_begin] = ACTIONS(1695),
    [sym_ocamlyacc_value] = ACTIONS(1697),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1695),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1697),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1695),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1695),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1697),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1697),
    [aux_sym_number_token1] = ACTIONS(1697),
    [aux_sym_signed_number_token1] = ACTIONS(1695),
    [anon_sym_SQUOTE] = ACTIONS(1697),
    [anon_sym_DQUOTE] = ACTIONS(1697),
    [sym_prefix_operator] = ACTIONS(1697),
    [aux_sym_sign_operator_token1] = ACTIONS(1695),
    [sym_hash_operator] = ACTIONS(1697),
    [anon_sym_lsl] = ACTIONS(1695),
    [anon_sym_lsr] = ACTIONS(1695),
    [anon_sym_asr] = ACTIONS(1695),
    [aux_sym_pow_operator_token1] = ACTIONS(1697),
    [anon_sym_mod] = ACTIONS(1695),
    [anon_sym_land] = ACTIONS(1695),
    [anon_sym_lor] = ACTIONS(1695),
    [anon_sym_lxor] = ACTIONS(1695),
    [aux_sym_mult_operator_token1] = ACTIONS(1695),
    [aux_sym_add_operator_token1] = ACTIONS(1695),
    [sym_concat_operator] = ACTIONS(1697),
    [sym_rel_operator] = ACTIONS(1695),
    [anon_sym_AMP_AMP] = ACTIONS(1695),
    [anon_sym_or] = ACTIONS(1695),
    [anon_sym_PIPE_PIPE] = ACTIONS(1695),
    [sym_assign_operator] = ACTIONS(1697),
    [sym__uppercase_identifier] = ACTIONS(1697),
    [aux_sym_tag_token1] = ACTIONS(1697),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(793)] = {
    [sym_attribute] = STATE(793),
    [sym__lowercase_identifier] = ACTIONS(1699),
    [anon_sym_and] = ACTIONS(1699),
    [anon_sym_TILDE] = ACTIONS(1699),
    [anon_sym_QMARK] = ACTIONS(1699),
    [anon_sym_LPAREN] = ACTIONS(1701),
    [anon_sym_RPAREN] = ACTIONS(1701),
    [anon_sym_COMMA] = ACTIONS(1701),
    [anon_sym_PIPE] = ACTIONS(1699),
    [anon_sym_COLON2] = ACTIONS(1699),
    [anon_sym_DOT] = ACTIONS(1703),
    [anon_sym_DASH_GT] = ACTIONS(1699),
    [anon_sym_LBRACK] = ACTIONS(1699),
    [anon_sym_RBRACK] = ACTIONS(1701),
    [anon_sym_true] = ACTIONS(1699),
    [anon_sym_false] = ACTIONS(1699),
    [anon_sym_LBRACE] = ACTIONS(1699),
    [anon_sym_SEMI] = ACTIONS(1701),
    [anon_sym_RBRACE] = ACTIONS(1701),
    [anon_sym_constraint] = ACTIONS(1699),
    [anon_sym_val] = ACTIONS(1699),
    [anon_sym_end] = ACTIONS(1699),
    [anon_sym_with] = ACTIONS(1699),
    [anon_sym_COLON_GT] = ACTIONS(1701),
    [anon_sym_object] = ACTIONS(1699),
    [anon_sym_inherit] = ACTIONS(1699),
    [anon_sym_method] = ACTIONS(1699),
    [anon_sym_in] = ACTIONS(1699),
    [anon_sym_as] = ACTIONS(1699),
    [anon_sym_initializer] = ACTIONS(1699),
    [anon_sym_AMP] = ACTIONS(1699),
    [anon_sym_POUND] = ACTIONS(1699),
    [anon_sym_COLON_COLON] = ACTIONS(1701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1701),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1701),
    [anon_sym_then] = ACTIONS(1699),
    [anon_sym_else] = ACTIONS(1699),
    [anon_sym_do] = ACTIONS(1699),
    [anon_sym_done] = ACTIONS(1699),
    [anon_sym_to] = ACTIONS(1699),
    [anon_sym_downto] = ACTIONS(1699),
    [anon_sym_new] = ACTIONS(1699),
    [anon_sym_LBRACE_LT] = ACTIONS(1701),
    [anon_sym_GT_RBRACE] = ACTIONS(1701),
    [anon_sym_begin] = ACTIONS(1699),
    [sym_ocamlyacc_value] = ACTIONS(1701),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1699),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1701),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1699),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1699),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1701),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1701),
    [aux_sym_number_token1] = ACTIONS(1701),
    [aux_sym_signed_number_token1] = ACTIONS(1699),
    [anon_sym_SQUOTE] = ACTIONS(1701),
    [anon_sym_DQUOTE] = ACTIONS(1701),
    [sym_prefix_operator] = ACTIONS(1701),
    [aux_sym_sign_operator_token1] = ACTIONS(1699),
    [sym_hash_operator] = ACTIONS(1701),
    [anon_sym_lsl] = ACTIONS(1699),
    [anon_sym_lsr] = ACTIONS(1699),
    [anon_sym_asr] = ACTIONS(1699),
    [aux_sym_pow_operator_token1] = ACTIONS(1701),
    [anon_sym_mod] = ACTIONS(1699),
    [anon_sym_land] = ACTIONS(1699),
    [anon_sym_lor] = ACTIONS(1699),
    [anon_sym_lxor] = ACTIONS(1699),
    [aux_sym_mult_operator_token1] = ACTIONS(1699),
    [aux_sym_add_operator_token1] = ACTIONS(1699),
    [sym_concat_operator] = ACTIONS(1701),
    [sym_rel_operator] = ACTIONS(1699),
    [anon_sym_AMP_AMP] = ACTIONS(1699),
    [anon_sym_or] = ACTIONS(1699),
    [anon_sym_PIPE_PIPE] = ACTIONS(1699),
    [sym_assign_operator] = ACTIONS(1701),
    [sym__uppercase_identifier] = ACTIONS(1701),
    [aux_sym_tag_token1] = ACTIONS(1701),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(794)] = {
    [sym_attribute] = STATE(794),
    [sym__lowercase_identifier] = ACTIONS(1705),
    [anon_sym_and] = ACTIONS(1705),
    [anon_sym_TILDE] = ACTIONS(1705),
    [anon_sym_QMARK] = ACTIONS(1705),
    [anon_sym_LPAREN] = ACTIONS(1707),
    [anon_sym_RPAREN] = ACTIONS(1707),
    [anon_sym_COMMA] = ACTIONS(1707),
    [anon_sym_PIPE] = ACTIONS(1705),
    [anon_sym_COLON2] = ACTIONS(1705),
    [anon_sym_DOT] = ACTIONS(1707),
    [anon_sym_DASH_GT] = ACTIONS(1705),
    [anon_sym_LBRACK] = ACTIONS(1705),
    [anon_sym_RBRACK] = ACTIONS(1707),
    [anon_sym_true] = ACTIONS(1705),
    [anon_sym_false] = ACTIONS(1705),
    [anon_sym_LBRACE] = ACTIONS(1705),
    [anon_sym_SEMI] = ACTIONS(1707),
    [anon_sym_RBRACE] = ACTIONS(1707),
    [anon_sym_constraint] = ACTIONS(1705),
    [anon_sym_val] = ACTIONS(1705),
    [anon_sym_end] = ACTIONS(1705),
    [anon_sym_with] = ACTIONS(1705),
    [anon_sym_COLON_GT] = ACTIONS(1707),
    [anon_sym_object] = ACTIONS(1705),
    [anon_sym_inherit] = ACTIONS(1705),
    [anon_sym_method] = ACTIONS(1705),
    [anon_sym_in] = ACTIONS(1705),
    [anon_sym_as] = ACTIONS(1705),
    [anon_sym_initializer] = ACTIONS(1705),
    [anon_sym_AMP] = ACTIONS(1705),
    [anon_sym_POUND] = ACTIONS(1705),
    [anon_sym_COLON_COLON] = ACTIONS(1707),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1707),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1707),
    [anon_sym_then] = ACTIONS(1705),
    [anon_sym_else] = ACTIONS(1705),
    [anon_sym_do] = ACTIONS(1705),
    [anon_sym_done] = ACTIONS(1705),
    [anon_sym_to] = ACTIONS(1705),
    [anon_sym_downto] = ACTIONS(1705),
    [anon_sym_new] = ACTIONS(1705),
    [anon_sym_LBRACE_LT] = ACTIONS(1707),
    [anon_sym_GT_RBRACE] = ACTIONS(1707),
    [anon_sym_begin] = ACTIONS(1705),
    [sym_ocamlyacc_value] = ACTIONS(1707),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1705),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1707),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1705),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1705),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1707),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1707),
    [aux_sym_number_token1] = ACTIONS(1707),
    [aux_sym_signed_number_token1] = ACTIONS(1705),
    [anon_sym_SQUOTE] = ACTIONS(1707),
    [anon_sym_DQUOTE] = ACTIONS(1707),
    [sym_prefix_operator] = ACTIONS(1707),
    [aux_sym_sign_operator_token1] = ACTIONS(1705),
    [sym_hash_operator] = ACTIONS(1707),
    [anon_sym_lsl] = ACTIONS(1705),
    [anon_sym_lsr] = ACTIONS(1705),
    [anon_sym_asr] = ACTIONS(1705),
    [aux_sym_pow_operator_token1] = ACTIONS(1707),
    [anon_sym_mod] = ACTIONS(1705),
    [anon_sym_land] = ACTIONS(1705),
    [anon_sym_lor] = ACTIONS(1705),
    [anon_sym_lxor] = ACTIONS(1705),
    [aux_sym_mult_operator_token1] = ACTIONS(1705),
    [aux_sym_add_operator_token1] = ACTIONS(1705),
    [sym_concat_operator] = ACTIONS(1707),
    [sym_rel_operator] = ACTIONS(1705),
    [anon_sym_AMP_AMP] = ACTIONS(1705),
    [anon_sym_or] = ACTIONS(1705),
    [anon_sym_PIPE_PIPE] = ACTIONS(1705),
    [sym_assign_operator] = ACTIONS(1707),
    [sym__uppercase_identifier] = ACTIONS(1707),
    [aux_sym_tag_token1] = ACTIONS(1707),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(795)] = {
    [sym_attribute] = STATE(795),
    [sym__lowercase_identifier] = ACTIONS(1709),
    [anon_sym_and] = ACTIONS(1709),
    [anon_sym_TILDE] = ACTIONS(1709),
    [anon_sym_QMARK] = ACTIONS(1709),
    [anon_sym_LPAREN] = ACTIONS(1711),
    [anon_sym_RPAREN] = ACTIONS(1711),
    [anon_sym_COMMA] = ACTIONS(1711),
    [anon_sym_PIPE] = ACTIONS(1709),
    [anon_sym_COLON2] = ACTIONS(1709),
    [anon_sym_DOT] = ACTIONS(1711),
    [anon_sym_DASH_GT] = ACTIONS(1709),
    [anon_sym_LBRACK] = ACTIONS(1709),
    [anon_sym_RBRACK] = ACTIONS(1711),
    [anon_sym_true] = ACTIONS(1709),
    [anon_sym_false] = ACTIONS(1709),
    [anon_sym_LBRACE] = ACTIONS(1709),
    [anon_sym_SEMI] = ACTIONS(1711),
    [anon_sym_RBRACE] = ACTIONS(1711),
    [anon_sym_constraint] = ACTIONS(1709),
    [anon_sym_val] = ACTIONS(1709),
    [anon_sym_end] = ACTIONS(1709),
    [anon_sym_with] = ACTIONS(1709),
    [anon_sym_COLON_GT] = ACTIONS(1711),
    [anon_sym_object] = ACTIONS(1709),
    [anon_sym_inherit] = ACTIONS(1709),
    [anon_sym_method] = ACTIONS(1709),
    [anon_sym_in] = ACTIONS(1709),
    [anon_sym_as] = ACTIONS(1709),
    [anon_sym_initializer] = ACTIONS(1709),
    [anon_sym_AMP] = ACTIONS(1709),
    [anon_sym_POUND] = ACTIONS(1709),
    [anon_sym_COLON_COLON] = ACTIONS(1711),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1711),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1711),
    [anon_sym_then] = ACTIONS(1709),
    [anon_sym_else] = ACTIONS(1709),
    [anon_sym_do] = ACTIONS(1709),
    [anon_sym_done] = ACTIONS(1709),
    [anon_sym_to] = ACTIONS(1709),
    [anon_sym_downto] = ACTIONS(1709),
    [anon_sym_new] = ACTIONS(1709),
    [anon_sym_LBRACE_LT] = ACTIONS(1711),
    [anon_sym_GT_RBRACE] = ACTIONS(1711),
    [anon_sym_begin] = ACTIONS(1709),
    [sym_ocamlyacc_value] = ACTIONS(1711),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1709),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1711),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1709),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1709),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1711),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1711),
    [aux_sym_number_token1] = ACTIONS(1711),
    [aux_sym_signed_number_token1] = ACTIONS(1709),
    [anon_sym_SQUOTE] = ACTIONS(1711),
    [anon_sym_DQUOTE] = ACTIONS(1711),
    [sym_prefix_operator] = ACTIONS(1711),
    [aux_sym_sign_operator_token1] = ACTIONS(1709),
    [sym_hash_operator] = ACTIONS(1711),
    [anon_sym_lsl] = ACTIONS(1709),
    [anon_sym_lsr] = ACTIONS(1709),
    [anon_sym_asr] = ACTIONS(1709),
    [aux_sym_pow_operator_token1] = ACTIONS(1711),
    [anon_sym_mod] = ACTIONS(1709),
    [anon_sym_land] = ACTIONS(1709),
    [anon_sym_lor] = ACTIONS(1709),
    [anon_sym_lxor] = ACTIONS(1709),
    [aux_sym_mult_operator_token1] = ACTIONS(1709),
    [aux_sym_add_operator_token1] = ACTIONS(1709),
    [sym_concat_operator] = ACTIONS(1711),
    [sym_rel_operator] = ACTIONS(1709),
    [anon_sym_AMP_AMP] = ACTIONS(1709),
    [anon_sym_or] = ACTIONS(1709),
    [anon_sym_PIPE_PIPE] = ACTIONS(1709),
    [sym_assign_operator] = ACTIONS(1711),
    [sym__uppercase_identifier] = ACTIONS(1711),
    [aux_sym_tag_token1] = ACTIONS(1711),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(796)] = {
    [sym_attribute] = STATE(796),
    [sym__lowercase_identifier] = ACTIONS(1713),
    [anon_sym_and] = ACTIONS(1713),
    [anon_sym_TILDE] = ACTIONS(1713),
    [anon_sym_QMARK] = ACTIONS(1713),
    [anon_sym_LPAREN] = ACTIONS(1715),
    [anon_sym_RPAREN] = ACTIONS(1715),
    [anon_sym_COMMA] = ACTIONS(1715),
    [anon_sym_PIPE] = ACTIONS(1713),
    [anon_sym_COLON2] = ACTIONS(1713),
    [anon_sym_DOT] = ACTIONS(1715),
    [anon_sym_DASH_GT] = ACTIONS(1713),
    [anon_sym_LBRACK] = ACTIONS(1713),
    [anon_sym_RBRACK] = ACTIONS(1715),
    [anon_sym_true] = ACTIONS(1713),
    [anon_sym_false] = ACTIONS(1713),
    [anon_sym_LBRACE] = ACTIONS(1713),
    [anon_sym_SEMI] = ACTIONS(1715),
    [anon_sym_RBRACE] = ACTIONS(1715),
    [anon_sym_constraint] = ACTIONS(1713),
    [anon_sym_val] = ACTIONS(1713),
    [anon_sym_end] = ACTIONS(1713),
    [anon_sym_with] = ACTIONS(1713),
    [anon_sym_COLON_GT] = ACTIONS(1715),
    [anon_sym_object] = ACTIONS(1713),
    [anon_sym_inherit] = ACTIONS(1713),
    [anon_sym_method] = ACTIONS(1713),
    [anon_sym_in] = ACTIONS(1713),
    [anon_sym_as] = ACTIONS(1713),
    [anon_sym_initializer] = ACTIONS(1713),
    [anon_sym_AMP] = ACTIONS(1713),
    [anon_sym_POUND] = ACTIONS(1713),
    [anon_sym_COLON_COLON] = ACTIONS(1715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1715),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1715),
    [anon_sym_then] = ACTIONS(1713),
    [anon_sym_else] = ACTIONS(1713),
    [anon_sym_do] = ACTIONS(1713),
    [anon_sym_done] = ACTIONS(1713),
    [anon_sym_to] = ACTIONS(1713),
    [anon_sym_downto] = ACTIONS(1713),
    [anon_sym_new] = ACTIONS(1713),
    [anon_sym_LBRACE_LT] = ACTIONS(1715),
    [anon_sym_GT_RBRACE] = ACTIONS(1715),
    [anon_sym_begin] = ACTIONS(1713),
    [sym_ocamlyacc_value] = ACTIONS(1715),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1713),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1715),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1713),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1713),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1715),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1715),
    [aux_sym_number_token1] = ACTIONS(1715),
    [aux_sym_signed_number_token1] = ACTIONS(1713),
    [anon_sym_SQUOTE] = ACTIONS(1715),
    [anon_sym_DQUOTE] = ACTIONS(1715),
    [sym_prefix_operator] = ACTIONS(1715),
    [aux_sym_sign_operator_token1] = ACTIONS(1713),
    [sym_hash_operator] = ACTIONS(1715),
    [anon_sym_lsl] = ACTIONS(1713),
    [anon_sym_lsr] = ACTIONS(1713),
    [anon_sym_asr] = ACTIONS(1713),
    [aux_sym_pow_operator_token1] = ACTIONS(1715),
    [anon_sym_mod] = ACTIONS(1713),
    [anon_sym_land] = ACTIONS(1713),
    [anon_sym_lor] = ACTIONS(1713),
    [anon_sym_lxor] = ACTIONS(1713),
    [aux_sym_mult_operator_token1] = ACTIONS(1713),
    [aux_sym_add_operator_token1] = ACTIONS(1713),
    [sym_concat_operator] = ACTIONS(1715),
    [sym_rel_operator] = ACTIONS(1713),
    [anon_sym_AMP_AMP] = ACTIONS(1713),
    [anon_sym_or] = ACTIONS(1713),
    [anon_sym_PIPE_PIPE] = ACTIONS(1713),
    [sym_assign_operator] = ACTIONS(1715),
    [sym__uppercase_identifier] = ACTIONS(1715),
    [aux_sym_tag_token1] = ACTIONS(1715),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(797)] = {
    [sym_attribute] = STATE(797),
    [sym__lowercase_identifier] = ACTIONS(1717),
    [anon_sym_and] = ACTIONS(1717),
    [anon_sym_TILDE] = ACTIONS(1717),
    [anon_sym_QMARK] = ACTIONS(1717),
    [anon_sym_LPAREN] = ACTIONS(1719),
    [anon_sym_RPAREN] = ACTIONS(1719),
    [anon_sym_COMMA] = ACTIONS(1719),
    [anon_sym_PIPE] = ACTIONS(1717),
    [anon_sym_COLON2] = ACTIONS(1717),
    [anon_sym_DOT] = ACTIONS(1719),
    [anon_sym_DASH_GT] = ACTIONS(1717),
    [anon_sym_LBRACK] = ACTIONS(1717),
    [anon_sym_RBRACK] = ACTIONS(1719),
    [anon_sym_true] = ACTIONS(1717),
    [anon_sym_false] = ACTIONS(1717),
    [anon_sym_LBRACE] = ACTIONS(1717),
    [anon_sym_SEMI] = ACTIONS(1719),
    [anon_sym_RBRACE] = ACTIONS(1719),
    [anon_sym_constraint] = ACTIONS(1717),
    [anon_sym_val] = ACTIONS(1717),
    [anon_sym_end] = ACTIONS(1717),
    [anon_sym_with] = ACTIONS(1717),
    [anon_sym_COLON_GT] = ACTIONS(1719),
    [anon_sym_object] = ACTIONS(1717),
    [anon_sym_inherit] = ACTIONS(1717),
    [anon_sym_method] = ACTIONS(1717),
    [anon_sym_in] = ACTIONS(1717),
    [anon_sym_as] = ACTIONS(1717),
    [anon_sym_initializer] = ACTIONS(1717),
    [anon_sym_AMP] = ACTIONS(1717),
    [anon_sym_POUND] = ACTIONS(1717),
    [anon_sym_COLON_COLON] = ACTIONS(1719),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1719),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1719),
    [anon_sym_then] = ACTIONS(1717),
    [anon_sym_else] = ACTIONS(1717),
    [anon_sym_do] = ACTIONS(1717),
    [anon_sym_done] = ACTIONS(1717),
    [anon_sym_to] = ACTIONS(1717),
    [anon_sym_downto] = ACTIONS(1717),
    [anon_sym_new] = ACTIONS(1717),
    [anon_sym_LBRACE_LT] = ACTIONS(1719),
    [anon_sym_GT_RBRACE] = ACTIONS(1719),
    [anon_sym_begin] = ACTIONS(1717),
    [sym_ocamlyacc_value] = ACTIONS(1719),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1717),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1719),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1717),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1717),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1719),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1719),
    [aux_sym_number_token1] = ACTIONS(1719),
    [aux_sym_signed_number_token1] = ACTIONS(1717),
    [anon_sym_SQUOTE] = ACTIONS(1719),
    [anon_sym_DQUOTE] = ACTIONS(1719),
    [sym_prefix_operator] = ACTIONS(1719),
    [aux_sym_sign_operator_token1] = ACTIONS(1717),
    [sym_hash_operator] = ACTIONS(1719),
    [anon_sym_lsl] = ACTIONS(1717),
    [anon_sym_lsr] = ACTIONS(1717),
    [anon_sym_asr] = ACTIONS(1717),
    [aux_sym_pow_operator_token1] = ACTIONS(1719),
    [anon_sym_mod] = ACTIONS(1717),
    [anon_sym_land] = ACTIONS(1717),
    [anon_sym_lor] = ACTIONS(1717),
    [anon_sym_lxor] = ACTIONS(1717),
    [aux_sym_mult_operator_token1] = ACTIONS(1717),
    [aux_sym_add_operator_token1] = ACTIONS(1717),
    [sym_concat_operator] = ACTIONS(1719),
    [sym_rel_operator] = ACTIONS(1717),
    [anon_sym_AMP_AMP] = ACTIONS(1717),
    [anon_sym_or] = ACTIONS(1717),
    [anon_sym_PIPE_PIPE] = ACTIONS(1717),
    [sym_assign_operator] = ACTIONS(1719),
    [sym__uppercase_identifier] = ACTIONS(1719),
    [aux_sym_tag_token1] = ACTIONS(1719),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(798)] = {
    [sym_attribute] = STATE(798),
    [sym__lowercase_identifier] = ACTIONS(1721),
    [anon_sym_and] = ACTIONS(1721),
    [anon_sym_TILDE] = ACTIONS(1721),
    [anon_sym_QMARK] = ACTIONS(1721),
    [anon_sym_LPAREN] = ACTIONS(1723),
    [anon_sym_RPAREN] = ACTIONS(1723),
    [anon_sym_COMMA] = ACTIONS(1723),
    [anon_sym_PIPE] = ACTIONS(1721),
    [anon_sym_COLON2] = ACTIONS(1721),
    [anon_sym_DOT] = ACTIONS(1723),
    [anon_sym_DASH_GT] = ACTIONS(1721),
    [anon_sym_LBRACK] = ACTIONS(1721),
    [anon_sym_RBRACK] = ACTIONS(1723),
    [anon_sym_true] = ACTIONS(1721),
    [anon_sym_false] = ACTIONS(1721),
    [anon_sym_LBRACE] = ACTIONS(1721),
    [anon_sym_SEMI] = ACTIONS(1723),
    [anon_sym_RBRACE] = ACTIONS(1723),
    [anon_sym_constraint] = ACTIONS(1721),
    [anon_sym_val] = ACTIONS(1721),
    [anon_sym_end] = ACTIONS(1721),
    [anon_sym_with] = ACTIONS(1721),
    [anon_sym_COLON_GT] = ACTIONS(1723),
    [anon_sym_object] = ACTIONS(1721),
    [anon_sym_inherit] = ACTIONS(1721),
    [anon_sym_method] = ACTIONS(1721),
    [anon_sym_in] = ACTIONS(1721),
    [anon_sym_as] = ACTIONS(1721),
    [anon_sym_initializer] = ACTIONS(1721),
    [anon_sym_AMP] = ACTIONS(1721),
    [anon_sym_POUND] = ACTIONS(1721),
    [anon_sym_COLON_COLON] = ACTIONS(1723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1723),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1723),
    [anon_sym_then] = ACTIONS(1721),
    [anon_sym_else] = ACTIONS(1721),
    [anon_sym_do] = ACTIONS(1721),
    [anon_sym_done] = ACTIONS(1721),
    [anon_sym_to] = ACTIONS(1721),
    [anon_sym_downto] = ACTIONS(1721),
    [anon_sym_new] = ACTIONS(1721),
    [anon_sym_LBRACE_LT] = ACTIONS(1723),
    [anon_sym_GT_RBRACE] = ACTIONS(1723),
    [anon_sym_begin] = ACTIONS(1721),
    [sym_ocamlyacc_value] = ACTIONS(1723),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1721),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1723),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1721),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1721),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1723),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1723),
    [aux_sym_number_token1] = ACTIONS(1723),
    [aux_sym_signed_number_token1] = ACTIONS(1721),
    [anon_sym_SQUOTE] = ACTIONS(1723),
    [anon_sym_DQUOTE] = ACTIONS(1723),
    [sym_prefix_operator] = ACTIONS(1723),
    [aux_sym_sign_operator_token1] = ACTIONS(1721),
    [sym_hash_operator] = ACTIONS(1723),
    [anon_sym_lsl] = ACTIONS(1721),
    [anon_sym_lsr] = ACTIONS(1721),
    [anon_sym_asr] = ACTIONS(1721),
    [aux_sym_pow_operator_token1] = ACTIONS(1723),
    [anon_sym_mod] = ACTIONS(1721),
    [anon_sym_land] = ACTIONS(1721),
    [anon_sym_lor] = ACTIONS(1721),
    [anon_sym_lxor] = ACTIONS(1721),
    [aux_sym_mult_operator_token1] = ACTIONS(1721),
    [aux_sym_add_operator_token1] = ACTIONS(1721),
    [sym_concat_operator] = ACTIONS(1723),
    [sym_rel_operator] = ACTIONS(1721),
    [anon_sym_AMP_AMP] = ACTIONS(1721),
    [anon_sym_or] = ACTIONS(1721),
    [anon_sym_PIPE_PIPE] = ACTIONS(1721),
    [sym_assign_operator] = ACTIONS(1723),
    [sym__uppercase_identifier] = ACTIONS(1723),
    [aux_sym_tag_token1] = ACTIONS(1723),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(799)] = {
    [sym_attribute] = STATE(799),
    [sym__lowercase_identifier] = ACTIONS(1725),
    [anon_sym_and] = ACTIONS(1725),
    [anon_sym_TILDE] = ACTIONS(1725),
    [anon_sym_QMARK] = ACTIONS(1725),
    [anon_sym_LPAREN] = ACTIONS(1727),
    [anon_sym_RPAREN] = ACTIONS(1727),
    [anon_sym_COMMA] = ACTIONS(1727),
    [anon_sym_PIPE] = ACTIONS(1725),
    [anon_sym_COLON2] = ACTIONS(1725),
    [anon_sym_DOT] = ACTIONS(1727),
    [anon_sym_DASH_GT] = ACTIONS(1725),
    [anon_sym_LBRACK] = ACTIONS(1725),
    [anon_sym_RBRACK] = ACTIONS(1727),
    [anon_sym_true] = ACTIONS(1725),
    [anon_sym_false] = ACTIONS(1725),
    [anon_sym_LBRACE] = ACTIONS(1725),
    [anon_sym_SEMI] = ACTIONS(1727),
    [anon_sym_RBRACE] = ACTIONS(1727),
    [anon_sym_constraint] = ACTIONS(1725),
    [anon_sym_val] = ACTIONS(1725),
    [anon_sym_end] = ACTIONS(1725),
    [anon_sym_with] = ACTIONS(1725),
    [anon_sym_COLON_GT] = ACTIONS(1727),
    [anon_sym_object] = ACTIONS(1725),
    [anon_sym_inherit] = ACTIONS(1725),
    [anon_sym_method] = ACTIONS(1725),
    [anon_sym_in] = ACTIONS(1725),
    [anon_sym_as] = ACTIONS(1725),
    [anon_sym_initializer] = ACTIONS(1725),
    [anon_sym_AMP] = ACTIONS(1725),
    [anon_sym_POUND] = ACTIONS(1725),
    [anon_sym_COLON_COLON] = ACTIONS(1727),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1727),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1727),
    [anon_sym_then] = ACTIONS(1725),
    [anon_sym_else] = ACTIONS(1725),
    [anon_sym_do] = ACTIONS(1725),
    [anon_sym_done] = ACTIONS(1725),
    [anon_sym_to] = ACTIONS(1725),
    [anon_sym_downto] = ACTIONS(1725),
    [anon_sym_new] = ACTIONS(1725),
    [anon_sym_LBRACE_LT] = ACTIONS(1727),
    [anon_sym_GT_RBRACE] = ACTIONS(1727),
    [anon_sym_begin] = ACTIONS(1725),
    [sym_ocamlyacc_value] = ACTIONS(1727),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1725),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1727),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1725),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1725),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1727),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1727),
    [aux_sym_number_token1] = ACTIONS(1727),
    [aux_sym_signed_number_token1] = ACTIONS(1725),
    [anon_sym_SQUOTE] = ACTIONS(1727),
    [anon_sym_DQUOTE] = ACTIONS(1727),
    [sym_prefix_operator] = ACTIONS(1727),
    [aux_sym_sign_operator_token1] = ACTIONS(1725),
    [sym_hash_operator] = ACTIONS(1727),
    [anon_sym_lsl] = ACTIONS(1725),
    [anon_sym_lsr] = ACTIONS(1725),
    [anon_sym_asr] = ACTIONS(1725),
    [aux_sym_pow_operator_token1] = ACTIONS(1727),
    [anon_sym_mod] = ACTIONS(1725),
    [anon_sym_land] = ACTIONS(1725),
    [anon_sym_lor] = ACTIONS(1725),
    [anon_sym_lxor] = ACTIONS(1725),
    [aux_sym_mult_operator_token1] = ACTIONS(1725),
    [aux_sym_add_operator_token1] = ACTIONS(1725),
    [sym_concat_operator] = ACTIONS(1727),
    [sym_rel_operator] = ACTIONS(1725),
    [anon_sym_AMP_AMP] = ACTIONS(1725),
    [anon_sym_or] = ACTIONS(1725),
    [anon_sym_PIPE_PIPE] = ACTIONS(1725),
    [sym_assign_operator] = ACTIONS(1727),
    [sym__uppercase_identifier] = ACTIONS(1727),
    [aux_sym_tag_token1] = ACTIONS(1727),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(800)] = {
    [sym_attribute] = STATE(800),
    [sym__lowercase_identifier] = ACTIONS(1729),
    [anon_sym_and] = ACTIONS(1729),
    [anon_sym_TILDE] = ACTIONS(1729),
    [anon_sym_QMARK] = ACTIONS(1729),
    [anon_sym_LPAREN] = ACTIONS(1731),
    [anon_sym_RPAREN] = ACTIONS(1731),
    [anon_sym_COMMA] = ACTIONS(1731),
    [anon_sym_PIPE] = ACTIONS(1729),
    [anon_sym_COLON2] = ACTIONS(1729),
    [anon_sym_DOT] = ACTIONS(1731),
    [anon_sym_DASH_GT] = ACTIONS(1729),
    [anon_sym_LBRACK] = ACTIONS(1729),
    [anon_sym_RBRACK] = ACTIONS(1731),
    [anon_sym_true] = ACTIONS(1729),
    [anon_sym_false] = ACTIONS(1729),
    [anon_sym_LBRACE] = ACTIONS(1729),
    [anon_sym_SEMI] = ACTIONS(1731),
    [anon_sym_RBRACE] = ACTIONS(1731),
    [anon_sym_constraint] = ACTIONS(1729),
    [anon_sym_val] = ACTIONS(1729),
    [anon_sym_end] = ACTIONS(1729),
    [anon_sym_with] = ACTIONS(1729),
    [anon_sym_COLON_GT] = ACTIONS(1731),
    [anon_sym_object] = ACTIONS(1729),
    [anon_sym_inherit] = ACTIONS(1729),
    [anon_sym_method] = ACTIONS(1729),
    [anon_sym_in] = ACTIONS(1729),
    [anon_sym_as] = ACTIONS(1729),
    [anon_sym_initializer] = ACTIONS(1729),
    [anon_sym_AMP] = ACTIONS(1729),
    [anon_sym_POUND] = ACTIONS(1729),
    [anon_sym_COLON_COLON] = ACTIONS(1731),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1731),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1731),
    [anon_sym_then] = ACTIONS(1729),
    [anon_sym_else] = ACTIONS(1729),
    [anon_sym_do] = ACTIONS(1729),
    [anon_sym_done] = ACTIONS(1729),
    [anon_sym_to] = ACTIONS(1729),
    [anon_sym_downto] = ACTIONS(1729),
    [anon_sym_new] = ACTIONS(1729),
    [anon_sym_LBRACE_LT] = ACTIONS(1731),
    [anon_sym_GT_RBRACE] = ACTIONS(1731),
    [anon_sym_begin] = ACTIONS(1729),
    [sym_ocamlyacc_value] = ACTIONS(1731),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1729),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1731),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1729),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1729),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1731),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1731),
    [aux_sym_number_token1] = ACTIONS(1731),
    [aux_sym_signed_number_token1] = ACTIONS(1729),
    [anon_sym_SQUOTE] = ACTIONS(1731),
    [anon_sym_DQUOTE] = ACTIONS(1731),
    [sym_prefix_operator] = ACTIONS(1731),
    [aux_sym_sign_operator_token1] = ACTIONS(1729),
    [sym_hash_operator] = ACTIONS(1731),
    [anon_sym_lsl] = ACTIONS(1729),
    [anon_sym_lsr] = ACTIONS(1729),
    [anon_sym_asr] = ACTIONS(1729),
    [aux_sym_pow_operator_token1] = ACTIONS(1731),
    [anon_sym_mod] = ACTIONS(1729),
    [anon_sym_land] = ACTIONS(1729),
    [anon_sym_lor] = ACTIONS(1729),
    [anon_sym_lxor] = ACTIONS(1729),
    [aux_sym_mult_operator_token1] = ACTIONS(1729),
    [aux_sym_add_operator_token1] = ACTIONS(1729),
    [sym_concat_operator] = ACTIONS(1731),
    [sym_rel_operator] = ACTIONS(1729),
    [anon_sym_AMP_AMP] = ACTIONS(1729),
    [anon_sym_or] = ACTIONS(1729),
    [anon_sym_PIPE_PIPE] = ACTIONS(1729),
    [sym_assign_operator] = ACTIONS(1731),
    [sym__uppercase_identifier] = ACTIONS(1731),
    [aux_sym_tag_token1] = ACTIONS(1731),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(801)] = {
    [sym_attribute] = STATE(801),
    [sym__lowercase_identifier] = ACTIONS(1733),
    [anon_sym_and] = ACTIONS(1733),
    [anon_sym_TILDE] = ACTIONS(1733),
    [anon_sym_QMARK] = ACTIONS(1733),
    [anon_sym_LPAREN] = ACTIONS(1735),
    [anon_sym_RPAREN] = ACTIONS(1735),
    [anon_sym_COMMA] = ACTIONS(1735),
    [anon_sym_PIPE] = ACTIONS(1733),
    [anon_sym_COLON2] = ACTIONS(1733),
    [anon_sym_DOT] = ACTIONS(1735),
    [anon_sym_DASH_GT] = ACTIONS(1733),
    [anon_sym_LBRACK] = ACTIONS(1733),
    [anon_sym_RBRACK] = ACTIONS(1735),
    [anon_sym_true] = ACTIONS(1733),
    [anon_sym_false] = ACTIONS(1733),
    [anon_sym_LBRACE] = ACTIONS(1733),
    [anon_sym_SEMI] = ACTIONS(1735),
    [anon_sym_RBRACE] = ACTIONS(1735),
    [anon_sym_constraint] = ACTIONS(1733),
    [anon_sym_val] = ACTIONS(1733),
    [anon_sym_end] = ACTIONS(1733),
    [anon_sym_with] = ACTIONS(1733),
    [anon_sym_COLON_GT] = ACTIONS(1735),
    [anon_sym_object] = ACTIONS(1733),
    [anon_sym_inherit] = ACTIONS(1733),
    [anon_sym_method] = ACTIONS(1733),
    [anon_sym_in] = ACTIONS(1733),
    [anon_sym_as] = ACTIONS(1733),
    [anon_sym_initializer] = ACTIONS(1733),
    [anon_sym_AMP] = ACTIONS(1733),
    [anon_sym_POUND] = ACTIONS(1733),
    [anon_sym_COLON_COLON] = ACTIONS(1735),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1735),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1735),
    [anon_sym_then] = ACTIONS(1733),
    [anon_sym_else] = ACTIONS(1733),
    [anon_sym_do] = ACTIONS(1733),
    [anon_sym_done] = ACTIONS(1733),
    [anon_sym_to] = ACTIONS(1733),
    [anon_sym_downto] = ACTIONS(1733),
    [anon_sym_new] = ACTIONS(1733),
    [anon_sym_LBRACE_LT] = ACTIONS(1735),
    [anon_sym_GT_RBRACE] = ACTIONS(1735),
    [anon_sym_begin] = ACTIONS(1733),
    [sym_ocamlyacc_value] = ACTIONS(1735),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1733),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1735),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1733),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1733),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1735),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1735),
    [aux_sym_number_token1] = ACTIONS(1735),
    [aux_sym_signed_number_token1] = ACTIONS(1733),
    [anon_sym_SQUOTE] = ACTIONS(1735),
    [anon_sym_DQUOTE] = ACTIONS(1735),
    [sym_prefix_operator] = ACTIONS(1735),
    [aux_sym_sign_operator_token1] = ACTIONS(1733),
    [sym_hash_operator] = ACTIONS(1735),
    [anon_sym_lsl] = ACTIONS(1733),
    [anon_sym_lsr] = ACTIONS(1733),
    [anon_sym_asr] = ACTIONS(1733),
    [aux_sym_pow_operator_token1] = ACTIONS(1735),
    [anon_sym_mod] = ACTIONS(1733),
    [anon_sym_land] = ACTIONS(1733),
    [anon_sym_lor] = ACTIONS(1733),
    [anon_sym_lxor] = ACTIONS(1733),
    [aux_sym_mult_operator_token1] = ACTIONS(1733),
    [aux_sym_add_operator_token1] = ACTIONS(1733),
    [sym_concat_operator] = ACTIONS(1735),
    [sym_rel_operator] = ACTIONS(1733),
    [anon_sym_AMP_AMP] = ACTIONS(1733),
    [anon_sym_or] = ACTIONS(1733),
    [anon_sym_PIPE_PIPE] = ACTIONS(1733),
    [sym_assign_operator] = ACTIONS(1735),
    [sym__uppercase_identifier] = ACTIONS(1735),
    [aux_sym_tag_token1] = ACTIONS(1735),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(802)] = {
    [sym_attribute] = STATE(802),
    [sym__lowercase_identifier] = ACTIONS(1737),
    [anon_sym_and] = ACTIONS(1737),
    [anon_sym_TILDE] = ACTIONS(1737),
    [anon_sym_QMARK] = ACTIONS(1737),
    [anon_sym_LPAREN] = ACTIONS(1739),
    [anon_sym_RPAREN] = ACTIONS(1739),
    [anon_sym_COMMA] = ACTIONS(1739),
    [anon_sym_PIPE] = ACTIONS(1737),
    [anon_sym_COLON2] = ACTIONS(1737),
    [anon_sym_DOT] = ACTIONS(1739),
    [anon_sym_DASH_GT] = ACTIONS(1737),
    [anon_sym_LBRACK] = ACTIONS(1737),
    [anon_sym_RBRACK] = ACTIONS(1739),
    [anon_sym_true] = ACTIONS(1737),
    [anon_sym_false] = ACTIONS(1737),
    [anon_sym_LBRACE] = ACTIONS(1737),
    [anon_sym_SEMI] = ACTIONS(1739),
    [anon_sym_RBRACE] = ACTIONS(1739),
    [anon_sym_constraint] = ACTIONS(1737),
    [anon_sym_val] = ACTIONS(1737),
    [anon_sym_end] = ACTIONS(1737),
    [anon_sym_with] = ACTIONS(1737),
    [anon_sym_COLON_GT] = ACTIONS(1739),
    [anon_sym_object] = ACTIONS(1737),
    [anon_sym_inherit] = ACTIONS(1737),
    [anon_sym_method] = ACTIONS(1737),
    [anon_sym_in] = ACTIONS(1737),
    [anon_sym_as] = ACTIONS(1737),
    [anon_sym_initializer] = ACTIONS(1737),
    [anon_sym_AMP] = ACTIONS(1737),
    [anon_sym_POUND] = ACTIONS(1737),
    [anon_sym_COLON_COLON] = ACTIONS(1739),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1739),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1739),
    [anon_sym_then] = ACTIONS(1737),
    [anon_sym_else] = ACTIONS(1737),
    [anon_sym_do] = ACTIONS(1737),
    [anon_sym_done] = ACTIONS(1737),
    [anon_sym_to] = ACTIONS(1737),
    [anon_sym_downto] = ACTIONS(1737),
    [anon_sym_new] = ACTIONS(1737),
    [anon_sym_LBRACE_LT] = ACTIONS(1739),
    [anon_sym_GT_RBRACE] = ACTIONS(1739),
    [anon_sym_begin] = ACTIONS(1737),
    [sym_ocamlyacc_value] = ACTIONS(1739),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1737),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1739),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1737),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1737),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1739),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1739),
    [aux_sym_number_token1] = ACTIONS(1739),
    [aux_sym_signed_number_token1] = ACTIONS(1737),
    [anon_sym_SQUOTE] = ACTIONS(1739),
    [anon_sym_DQUOTE] = ACTIONS(1739),
    [sym_prefix_operator] = ACTIONS(1739),
    [aux_sym_sign_operator_token1] = ACTIONS(1737),
    [sym_hash_operator] = ACTIONS(1739),
    [anon_sym_lsl] = ACTIONS(1737),
    [anon_sym_lsr] = ACTIONS(1737),
    [anon_sym_asr] = ACTIONS(1737),
    [aux_sym_pow_operator_token1] = ACTIONS(1739),
    [anon_sym_mod] = ACTIONS(1737),
    [anon_sym_land] = ACTIONS(1737),
    [anon_sym_lor] = ACTIONS(1737),
    [anon_sym_lxor] = ACTIONS(1737),
    [aux_sym_mult_operator_token1] = ACTIONS(1737),
    [aux_sym_add_operator_token1] = ACTIONS(1737),
    [sym_concat_operator] = ACTIONS(1739),
    [sym_rel_operator] = ACTIONS(1737),
    [anon_sym_AMP_AMP] = ACTIONS(1737),
    [anon_sym_or] = ACTIONS(1737),
    [anon_sym_PIPE_PIPE] = ACTIONS(1737),
    [sym_assign_operator] = ACTIONS(1739),
    [sym__uppercase_identifier] = ACTIONS(1739),
    [aux_sym_tag_token1] = ACTIONS(1739),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(803)] = {
    [sym_attribute] = STATE(803),
    [sym__lowercase_identifier] = ACTIONS(1741),
    [anon_sym_and] = ACTIONS(1741),
    [anon_sym_TILDE] = ACTIONS(1741),
    [anon_sym_QMARK] = ACTIONS(1741),
    [anon_sym_LPAREN] = ACTIONS(1743),
    [anon_sym_RPAREN] = ACTIONS(1743),
    [anon_sym_COMMA] = ACTIONS(1743),
    [anon_sym_PIPE] = ACTIONS(1741),
    [anon_sym_COLON2] = ACTIONS(1741),
    [anon_sym_DOT] = ACTIONS(1743),
    [anon_sym_DASH_GT] = ACTIONS(1741),
    [anon_sym_LBRACK] = ACTIONS(1741),
    [anon_sym_RBRACK] = ACTIONS(1743),
    [anon_sym_true] = ACTIONS(1741),
    [anon_sym_false] = ACTIONS(1741),
    [anon_sym_LBRACE] = ACTIONS(1741),
    [anon_sym_SEMI] = ACTIONS(1743),
    [anon_sym_RBRACE] = ACTIONS(1743),
    [anon_sym_constraint] = ACTIONS(1741),
    [anon_sym_val] = ACTIONS(1741),
    [anon_sym_end] = ACTIONS(1741),
    [anon_sym_with] = ACTIONS(1741),
    [anon_sym_COLON_GT] = ACTIONS(1743),
    [anon_sym_object] = ACTIONS(1741),
    [anon_sym_inherit] = ACTIONS(1741),
    [anon_sym_method] = ACTIONS(1741),
    [anon_sym_in] = ACTIONS(1741),
    [anon_sym_as] = ACTIONS(1741),
    [anon_sym_initializer] = ACTIONS(1741),
    [anon_sym_AMP] = ACTIONS(1741),
    [anon_sym_POUND] = ACTIONS(1741),
    [anon_sym_COLON_COLON] = ACTIONS(1743),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1743),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1743),
    [anon_sym_then] = ACTIONS(1741),
    [anon_sym_else] = ACTIONS(1741),
    [anon_sym_do] = ACTIONS(1741),
    [anon_sym_done] = ACTIONS(1741),
    [anon_sym_to] = ACTIONS(1741),
    [anon_sym_downto] = ACTIONS(1741),
    [anon_sym_new] = ACTIONS(1741),
    [anon_sym_LBRACE_LT] = ACTIONS(1743),
    [anon_sym_GT_RBRACE] = ACTIONS(1743),
    [anon_sym_begin] = ACTIONS(1741),
    [sym_ocamlyacc_value] = ACTIONS(1743),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1741),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1743),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1741),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1741),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1743),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1743),
    [aux_sym_number_token1] = ACTIONS(1743),
    [aux_sym_signed_number_token1] = ACTIONS(1741),
    [anon_sym_SQUOTE] = ACTIONS(1743),
    [anon_sym_DQUOTE] = ACTIONS(1743),
    [sym_prefix_operator] = ACTIONS(1743),
    [aux_sym_sign_operator_token1] = ACTIONS(1741),
    [sym_hash_operator] = ACTIONS(1743),
    [anon_sym_lsl] = ACTIONS(1741),
    [anon_sym_lsr] = ACTIONS(1741),
    [anon_sym_asr] = ACTIONS(1741),
    [aux_sym_pow_operator_token1] = ACTIONS(1743),
    [anon_sym_mod] = ACTIONS(1741),
    [anon_sym_land] = ACTIONS(1741),
    [anon_sym_lor] = ACTIONS(1741),
    [anon_sym_lxor] = ACTIONS(1741),
    [aux_sym_mult_operator_token1] = ACTIONS(1741),
    [aux_sym_add_operator_token1] = ACTIONS(1741),
    [sym_concat_operator] = ACTIONS(1743),
    [sym_rel_operator] = ACTIONS(1741),
    [anon_sym_AMP_AMP] = ACTIONS(1741),
    [anon_sym_or] = ACTIONS(1741),
    [anon_sym_PIPE_PIPE] = ACTIONS(1741),
    [sym_assign_operator] = ACTIONS(1743),
    [sym__uppercase_identifier] = ACTIONS(1743),
    [aux_sym_tag_token1] = ACTIONS(1743),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(804)] = {
    [sym_attribute] = STATE(804),
    [sym__lowercase_identifier] = ACTIONS(1745),
    [anon_sym_and] = ACTIONS(1745),
    [anon_sym_TILDE] = ACTIONS(1745),
    [anon_sym_QMARK] = ACTIONS(1745),
    [anon_sym_LPAREN] = ACTIONS(1747),
    [anon_sym_RPAREN] = ACTIONS(1747),
    [anon_sym_COMMA] = ACTIONS(1747),
    [anon_sym_PIPE] = ACTIONS(1745),
    [anon_sym_COLON2] = ACTIONS(1745),
    [anon_sym_DOT] = ACTIONS(1747),
    [anon_sym_DASH_GT] = ACTIONS(1745),
    [anon_sym_LBRACK] = ACTIONS(1745),
    [anon_sym_RBRACK] = ACTIONS(1747),
    [anon_sym_true] = ACTIONS(1745),
    [anon_sym_false] = ACTIONS(1745),
    [anon_sym_LBRACE] = ACTIONS(1745),
    [anon_sym_SEMI] = ACTIONS(1747),
    [anon_sym_RBRACE] = ACTIONS(1747),
    [anon_sym_constraint] = ACTIONS(1745),
    [anon_sym_val] = ACTIONS(1745),
    [anon_sym_end] = ACTIONS(1745),
    [anon_sym_with] = ACTIONS(1745),
    [anon_sym_COLON_GT] = ACTIONS(1747),
    [anon_sym_object] = ACTIONS(1745),
    [anon_sym_inherit] = ACTIONS(1745),
    [anon_sym_method] = ACTIONS(1745),
    [anon_sym_in] = ACTIONS(1745),
    [anon_sym_as] = ACTIONS(1745),
    [anon_sym_initializer] = ACTIONS(1745),
    [anon_sym_AMP] = ACTIONS(1745),
    [anon_sym_POUND] = ACTIONS(1745),
    [anon_sym_COLON_COLON] = ACTIONS(1747),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1747),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1747),
    [anon_sym_then] = ACTIONS(1745),
    [anon_sym_else] = ACTIONS(1745),
    [anon_sym_do] = ACTIONS(1745),
    [anon_sym_done] = ACTIONS(1745),
    [anon_sym_to] = ACTIONS(1745),
    [anon_sym_downto] = ACTIONS(1745),
    [anon_sym_new] = ACTIONS(1745),
    [anon_sym_LBRACE_LT] = ACTIONS(1747),
    [anon_sym_GT_RBRACE] = ACTIONS(1747),
    [anon_sym_begin] = ACTIONS(1745),
    [sym_ocamlyacc_value] = ACTIONS(1747),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1745),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1747),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1745),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1745),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1747),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1747),
    [aux_sym_number_token1] = ACTIONS(1747),
    [aux_sym_signed_number_token1] = ACTIONS(1745),
    [anon_sym_SQUOTE] = ACTIONS(1747),
    [anon_sym_DQUOTE] = ACTIONS(1747),
    [sym_prefix_operator] = ACTIONS(1747),
    [aux_sym_sign_operator_token1] = ACTIONS(1745),
    [sym_hash_operator] = ACTIONS(1747),
    [anon_sym_lsl] = ACTIONS(1745),
    [anon_sym_lsr] = ACTIONS(1745),
    [anon_sym_asr] = ACTIONS(1745),
    [aux_sym_pow_operator_token1] = ACTIONS(1747),
    [anon_sym_mod] = ACTIONS(1745),
    [anon_sym_land] = ACTIONS(1745),
    [anon_sym_lor] = ACTIONS(1745),
    [anon_sym_lxor] = ACTIONS(1745),
    [aux_sym_mult_operator_token1] = ACTIONS(1745),
    [aux_sym_add_operator_token1] = ACTIONS(1745),
    [sym_concat_operator] = ACTIONS(1747),
    [sym_rel_operator] = ACTIONS(1745),
    [anon_sym_AMP_AMP] = ACTIONS(1745),
    [anon_sym_or] = ACTIONS(1745),
    [anon_sym_PIPE_PIPE] = ACTIONS(1745),
    [sym_assign_operator] = ACTIONS(1747),
    [sym__uppercase_identifier] = ACTIONS(1747),
    [aux_sym_tag_token1] = ACTIONS(1747),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(805)] = {
    [sym_attribute] = STATE(805),
    [sym__lowercase_identifier] = ACTIONS(1749),
    [anon_sym_and] = ACTIONS(1749),
    [anon_sym_TILDE] = ACTIONS(1749),
    [anon_sym_QMARK] = ACTIONS(1749),
    [anon_sym_LPAREN] = ACTIONS(1751),
    [anon_sym_RPAREN] = ACTIONS(1751),
    [anon_sym_COMMA] = ACTIONS(1751),
    [anon_sym_PIPE] = ACTIONS(1749),
    [anon_sym_COLON2] = ACTIONS(1749),
    [anon_sym_DOT] = ACTIONS(1751),
    [anon_sym_DASH_GT] = ACTIONS(1749),
    [anon_sym_LBRACK] = ACTIONS(1749),
    [anon_sym_RBRACK] = ACTIONS(1751),
    [anon_sym_true] = ACTIONS(1749),
    [anon_sym_false] = ACTIONS(1749),
    [anon_sym_LBRACE] = ACTIONS(1749),
    [anon_sym_SEMI] = ACTIONS(1751),
    [anon_sym_RBRACE] = ACTIONS(1751),
    [anon_sym_constraint] = ACTIONS(1749),
    [anon_sym_val] = ACTIONS(1749),
    [anon_sym_end] = ACTIONS(1749),
    [anon_sym_with] = ACTIONS(1749),
    [anon_sym_COLON_GT] = ACTIONS(1751),
    [anon_sym_object] = ACTIONS(1749),
    [anon_sym_inherit] = ACTIONS(1749),
    [anon_sym_method] = ACTIONS(1749),
    [anon_sym_in] = ACTIONS(1749),
    [anon_sym_as] = ACTIONS(1749),
    [anon_sym_initializer] = ACTIONS(1749),
    [anon_sym_AMP] = ACTIONS(1749),
    [anon_sym_POUND] = ACTIONS(1749),
    [anon_sym_COLON_COLON] = ACTIONS(1751),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1751),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1751),
    [anon_sym_then] = ACTIONS(1749),
    [anon_sym_else] = ACTIONS(1749),
    [anon_sym_do] = ACTIONS(1749),
    [anon_sym_done] = ACTIONS(1749),
    [anon_sym_to] = ACTIONS(1749),
    [anon_sym_downto] = ACTIONS(1749),
    [anon_sym_new] = ACTIONS(1749),
    [anon_sym_LBRACE_LT] = ACTIONS(1751),
    [anon_sym_GT_RBRACE] = ACTIONS(1751),
    [anon_sym_begin] = ACTIONS(1749),
    [sym_ocamlyacc_value] = ACTIONS(1751),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1749),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1751),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1749),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1749),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1751),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1751),
    [aux_sym_number_token1] = ACTIONS(1751),
    [aux_sym_signed_number_token1] = ACTIONS(1749),
    [anon_sym_SQUOTE] = ACTIONS(1751),
    [anon_sym_DQUOTE] = ACTIONS(1751),
    [sym_prefix_operator] = ACTIONS(1751),
    [aux_sym_sign_operator_token1] = ACTIONS(1749),
    [sym_hash_operator] = ACTIONS(1751),
    [anon_sym_lsl] = ACTIONS(1749),
    [anon_sym_lsr] = ACTIONS(1749),
    [anon_sym_asr] = ACTIONS(1749),
    [aux_sym_pow_operator_token1] = ACTIONS(1751),
    [anon_sym_mod] = ACTIONS(1749),
    [anon_sym_land] = ACTIONS(1749),
    [anon_sym_lor] = ACTIONS(1749),
    [anon_sym_lxor] = ACTIONS(1749),
    [aux_sym_mult_operator_token1] = ACTIONS(1749),
    [aux_sym_add_operator_token1] = ACTIONS(1749),
    [sym_concat_operator] = ACTIONS(1751),
    [sym_rel_operator] = ACTIONS(1749),
    [anon_sym_AMP_AMP] = ACTIONS(1749),
    [anon_sym_or] = ACTIONS(1749),
    [anon_sym_PIPE_PIPE] = ACTIONS(1749),
    [sym_assign_operator] = ACTIONS(1751),
    [sym__uppercase_identifier] = ACTIONS(1751),
    [aux_sym_tag_token1] = ACTIONS(1751),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(806)] = {
    [sym_attribute] = STATE(806),
    [sym__lowercase_identifier] = ACTIONS(1753),
    [anon_sym_and] = ACTIONS(1753),
    [anon_sym_TILDE] = ACTIONS(1753),
    [anon_sym_QMARK] = ACTIONS(1753),
    [anon_sym_LPAREN] = ACTIONS(1755),
    [anon_sym_RPAREN] = ACTIONS(1755),
    [anon_sym_COMMA] = ACTIONS(1755),
    [anon_sym_PIPE] = ACTIONS(1753),
    [anon_sym_COLON2] = ACTIONS(1753),
    [anon_sym_DOT] = ACTIONS(1755),
    [anon_sym_DASH_GT] = ACTIONS(1753),
    [anon_sym_LBRACK] = ACTIONS(1753),
    [anon_sym_RBRACK] = ACTIONS(1755),
    [anon_sym_true] = ACTIONS(1753),
    [anon_sym_false] = ACTIONS(1753),
    [anon_sym_LBRACE] = ACTIONS(1753),
    [anon_sym_SEMI] = ACTIONS(1755),
    [anon_sym_RBRACE] = ACTIONS(1755),
    [anon_sym_constraint] = ACTIONS(1753),
    [anon_sym_val] = ACTIONS(1753),
    [anon_sym_end] = ACTIONS(1753),
    [anon_sym_with] = ACTIONS(1753),
    [anon_sym_COLON_GT] = ACTIONS(1755),
    [anon_sym_object] = ACTIONS(1753),
    [anon_sym_inherit] = ACTIONS(1753),
    [anon_sym_method] = ACTIONS(1753),
    [anon_sym_in] = ACTIONS(1753),
    [anon_sym_as] = ACTIONS(1753),
    [anon_sym_initializer] = ACTIONS(1753),
    [anon_sym_AMP] = ACTIONS(1753),
    [anon_sym_POUND] = ACTIONS(1753),
    [anon_sym_COLON_COLON] = ACTIONS(1755),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1755),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1755),
    [anon_sym_then] = ACTIONS(1753),
    [anon_sym_else] = ACTIONS(1753),
    [anon_sym_do] = ACTIONS(1753),
    [anon_sym_done] = ACTIONS(1753),
    [anon_sym_to] = ACTIONS(1753),
    [anon_sym_downto] = ACTIONS(1753),
    [anon_sym_new] = ACTIONS(1753),
    [anon_sym_LBRACE_LT] = ACTIONS(1755),
    [anon_sym_GT_RBRACE] = ACTIONS(1755),
    [anon_sym_begin] = ACTIONS(1753),
    [sym_ocamlyacc_value] = ACTIONS(1755),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1753),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1755),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1753),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1753),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1755),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1755),
    [aux_sym_number_token1] = ACTIONS(1755),
    [aux_sym_signed_number_token1] = ACTIONS(1753),
    [anon_sym_SQUOTE] = ACTIONS(1755),
    [anon_sym_DQUOTE] = ACTIONS(1755),
    [sym_prefix_operator] = ACTIONS(1755),
    [aux_sym_sign_operator_token1] = ACTIONS(1753),
    [sym_hash_operator] = ACTIONS(1755),
    [anon_sym_lsl] = ACTIONS(1753),
    [anon_sym_lsr] = ACTIONS(1753),
    [anon_sym_asr] = ACTIONS(1753),
    [aux_sym_pow_operator_token1] = ACTIONS(1755),
    [anon_sym_mod] = ACTIONS(1753),
    [anon_sym_land] = ACTIONS(1753),
    [anon_sym_lor] = ACTIONS(1753),
    [anon_sym_lxor] = ACTIONS(1753),
    [aux_sym_mult_operator_token1] = ACTIONS(1753),
    [aux_sym_add_operator_token1] = ACTIONS(1753),
    [sym_concat_operator] = ACTIONS(1755),
    [sym_rel_operator] = ACTIONS(1753),
    [anon_sym_AMP_AMP] = ACTIONS(1753),
    [anon_sym_or] = ACTIONS(1753),
    [anon_sym_PIPE_PIPE] = ACTIONS(1753),
    [sym_assign_operator] = ACTIONS(1755),
    [sym__uppercase_identifier] = ACTIONS(1755),
    [aux_sym_tag_token1] = ACTIONS(1755),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(807)] = {
    [sym_attribute] = STATE(807),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_and] = ACTIONS(1757),
    [anon_sym_TILDE] = ACTIONS(1757),
    [anon_sym_QMARK] = ACTIONS(1757),
    [anon_sym_LPAREN] = ACTIONS(1759),
    [anon_sym_RPAREN] = ACTIONS(1759),
    [anon_sym_COMMA] = ACTIONS(1759),
    [anon_sym_PIPE] = ACTIONS(1757),
    [anon_sym_COLON2] = ACTIONS(1757),
    [anon_sym_DOT] = ACTIONS(1759),
    [anon_sym_DASH_GT] = ACTIONS(1757),
    [anon_sym_LBRACK] = ACTIONS(1757),
    [anon_sym_RBRACK] = ACTIONS(1759),
    [anon_sym_true] = ACTIONS(1757),
    [anon_sym_false] = ACTIONS(1757),
    [anon_sym_LBRACE] = ACTIONS(1757),
    [anon_sym_SEMI] = ACTIONS(1759),
    [anon_sym_RBRACE] = ACTIONS(1759),
    [anon_sym_constraint] = ACTIONS(1757),
    [anon_sym_val] = ACTIONS(1757),
    [anon_sym_end] = ACTIONS(1757),
    [anon_sym_with] = ACTIONS(1757),
    [anon_sym_COLON_GT] = ACTIONS(1759),
    [anon_sym_object] = ACTIONS(1757),
    [anon_sym_inherit] = ACTIONS(1757),
    [anon_sym_method] = ACTIONS(1757),
    [anon_sym_in] = ACTIONS(1757),
    [anon_sym_as] = ACTIONS(1757),
    [anon_sym_initializer] = ACTIONS(1757),
    [anon_sym_AMP] = ACTIONS(1757),
    [anon_sym_POUND] = ACTIONS(1757),
    [anon_sym_COLON_COLON] = ACTIONS(1759),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1759),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1759),
    [anon_sym_then] = ACTIONS(1757),
    [anon_sym_else] = ACTIONS(1757),
    [anon_sym_do] = ACTIONS(1757),
    [anon_sym_done] = ACTIONS(1757),
    [anon_sym_to] = ACTIONS(1757),
    [anon_sym_downto] = ACTIONS(1757),
    [anon_sym_new] = ACTIONS(1757),
    [anon_sym_LBRACE_LT] = ACTIONS(1759),
    [anon_sym_GT_RBRACE] = ACTIONS(1759),
    [anon_sym_begin] = ACTIONS(1757),
    [sym_ocamlyacc_value] = ACTIONS(1759),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1757),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1759),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1757),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1757),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1759),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1759),
    [aux_sym_number_token1] = ACTIONS(1759),
    [aux_sym_signed_number_token1] = ACTIONS(1757),
    [anon_sym_SQUOTE] = ACTIONS(1759),
    [anon_sym_DQUOTE] = ACTIONS(1759),
    [sym_prefix_operator] = ACTIONS(1759),
    [aux_sym_sign_operator_token1] = ACTIONS(1757),
    [sym_hash_operator] = ACTIONS(1759),
    [anon_sym_lsl] = ACTIONS(1757),
    [anon_sym_lsr] = ACTIONS(1757),
    [anon_sym_asr] = ACTIONS(1757),
    [aux_sym_pow_operator_token1] = ACTIONS(1759),
    [anon_sym_mod] = ACTIONS(1757),
    [anon_sym_land] = ACTIONS(1757),
    [anon_sym_lor] = ACTIONS(1757),
    [anon_sym_lxor] = ACTIONS(1757),
    [aux_sym_mult_operator_token1] = ACTIONS(1757),
    [aux_sym_add_operator_token1] = ACTIONS(1757),
    [sym_concat_operator] = ACTIONS(1759),
    [sym_rel_operator] = ACTIONS(1757),
    [anon_sym_AMP_AMP] = ACTIONS(1757),
    [anon_sym_or] = ACTIONS(1757),
    [anon_sym_PIPE_PIPE] = ACTIONS(1757),
    [sym_assign_operator] = ACTIONS(1759),
    [sym__uppercase_identifier] = ACTIONS(1759),
    [aux_sym_tag_token1] = ACTIONS(1759),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(808)] = {
    [sym_attribute] = STATE(808),
    [sym__lowercase_identifier] = ACTIONS(1761),
    [anon_sym_and] = ACTIONS(1761),
    [anon_sym_TILDE] = ACTIONS(1761),
    [anon_sym_QMARK] = ACTIONS(1761),
    [anon_sym_LPAREN] = ACTIONS(1763),
    [anon_sym_RPAREN] = ACTIONS(1763),
    [anon_sym_COMMA] = ACTIONS(1763),
    [anon_sym_PIPE] = ACTIONS(1761),
    [anon_sym_COLON2] = ACTIONS(1761),
    [anon_sym_DOT] = ACTIONS(1763),
    [anon_sym_DASH_GT] = ACTIONS(1761),
    [anon_sym_LBRACK] = ACTIONS(1761),
    [anon_sym_RBRACK] = ACTIONS(1763),
    [anon_sym_true] = ACTIONS(1761),
    [anon_sym_false] = ACTIONS(1761),
    [anon_sym_LBRACE] = ACTIONS(1761),
    [anon_sym_SEMI] = ACTIONS(1763),
    [anon_sym_RBRACE] = ACTIONS(1763),
    [anon_sym_constraint] = ACTIONS(1761),
    [anon_sym_val] = ACTIONS(1761),
    [anon_sym_end] = ACTIONS(1761),
    [anon_sym_with] = ACTIONS(1761),
    [anon_sym_COLON_GT] = ACTIONS(1763),
    [anon_sym_object] = ACTIONS(1761),
    [anon_sym_inherit] = ACTIONS(1761),
    [anon_sym_method] = ACTIONS(1761),
    [anon_sym_in] = ACTIONS(1761),
    [anon_sym_as] = ACTIONS(1761),
    [anon_sym_initializer] = ACTIONS(1761),
    [anon_sym_AMP] = ACTIONS(1761),
    [anon_sym_POUND] = ACTIONS(1761),
    [anon_sym_COLON_COLON] = ACTIONS(1763),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1763),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1763),
    [anon_sym_then] = ACTIONS(1761),
    [anon_sym_else] = ACTIONS(1761),
    [anon_sym_do] = ACTIONS(1761),
    [anon_sym_done] = ACTIONS(1761),
    [anon_sym_to] = ACTIONS(1761),
    [anon_sym_downto] = ACTIONS(1761),
    [anon_sym_new] = ACTIONS(1761),
    [anon_sym_LBRACE_LT] = ACTIONS(1763),
    [anon_sym_GT_RBRACE] = ACTIONS(1763),
    [anon_sym_begin] = ACTIONS(1761),
    [sym_ocamlyacc_value] = ACTIONS(1763),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1761),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1763),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1761),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1761),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1763),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1763),
    [aux_sym_number_token1] = ACTIONS(1763),
    [aux_sym_signed_number_token1] = ACTIONS(1761),
    [anon_sym_SQUOTE] = ACTIONS(1763),
    [anon_sym_DQUOTE] = ACTIONS(1763),
    [sym_prefix_operator] = ACTIONS(1763),
    [aux_sym_sign_operator_token1] = ACTIONS(1761),
    [sym_hash_operator] = ACTIONS(1763),
    [anon_sym_lsl] = ACTIONS(1761),
    [anon_sym_lsr] = ACTIONS(1761),
    [anon_sym_asr] = ACTIONS(1761),
    [aux_sym_pow_operator_token1] = ACTIONS(1763),
    [anon_sym_mod] = ACTIONS(1761),
    [anon_sym_land] = ACTIONS(1761),
    [anon_sym_lor] = ACTIONS(1761),
    [anon_sym_lxor] = ACTIONS(1761),
    [aux_sym_mult_operator_token1] = ACTIONS(1761),
    [aux_sym_add_operator_token1] = ACTIONS(1761),
    [sym_concat_operator] = ACTIONS(1763),
    [sym_rel_operator] = ACTIONS(1761),
    [anon_sym_AMP_AMP] = ACTIONS(1761),
    [anon_sym_or] = ACTIONS(1761),
    [anon_sym_PIPE_PIPE] = ACTIONS(1761),
    [sym_assign_operator] = ACTIONS(1763),
    [sym__uppercase_identifier] = ACTIONS(1763),
    [aux_sym_tag_token1] = ACTIONS(1763),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(809)] = {
    [sym_attribute] = STATE(809),
    [sym__lowercase_identifier] = ACTIONS(1765),
    [anon_sym_and] = ACTIONS(1765),
    [anon_sym_TILDE] = ACTIONS(1765),
    [anon_sym_QMARK] = ACTIONS(1765),
    [anon_sym_LPAREN] = ACTIONS(1767),
    [anon_sym_RPAREN] = ACTIONS(1767),
    [anon_sym_COMMA] = ACTIONS(1767),
    [anon_sym_PIPE] = ACTIONS(1765),
    [anon_sym_COLON2] = ACTIONS(1765),
    [anon_sym_DOT] = ACTIONS(1767),
    [anon_sym_DASH_GT] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_RBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1765),
    [anon_sym_false] = ACTIONS(1765),
    [anon_sym_LBRACE] = ACTIONS(1765),
    [anon_sym_SEMI] = ACTIONS(1767),
    [anon_sym_RBRACE] = ACTIONS(1767),
    [anon_sym_constraint] = ACTIONS(1765),
    [anon_sym_val] = ACTIONS(1765),
    [anon_sym_end] = ACTIONS(1765),
    [anon_sym_with] = ACTIONS(1765),
    [anon_sym_COLON_GT] = ACTIONS(1767),
    [anon_sym_object] = ACTIONS(1765),
    [anon_sym_inherit] = ACTIONS(1765),
    [anon_sym_method] = ACTIONS(1765),
    [anon_sym_in] = ACTIONS(1765),
    [anon_sym_as] = ACTIONS(1765),
    [anon_sym_initializer] = ACTIONS(1765),
    [anon_sym_AMP] = ACTIONS(1765),
    [anon_sym_POUND] = ACTIONS(1765),
    [anon_sym_COLON_COLON] = ACTIONS(1767),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1767),
    [anon_sym_then] = ACTIONS(1765),
    [anon_sym_else] = ACTIONS(1765),
    [anon_sym_do] = ACTIONS(1765),
    [anon_sym_done] = ACTIONS(1765),
    [anon_sym_to] = ACTIONS(1765),
    [anon_sym_downto] = ACTIONS(1765),
    [anon_sym_new] = ACTIONS(1765),
    [anon_sym_LBRACE_LT] = ACTIONS(1767),
    [anon_sym_GT_RBRACE] = ACTIONS(1767),
    [anon_sym_begin] = ACTIONS(1765),
    [sym_ocamlyacc_value] = ACTIONS(1767),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1765),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1767),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1765),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1765),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1767),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1767),
    [aux_sym_number_token1] = ACTIONS(1767),
    [aux_sym_signed_number_token1] = ACTIONS(1765),
    [anon_sym_SQUOTE] = ACTIONS(1767),
    [anon_sym_DQUOTE] = ACTIONS(1767),
    [sym_prefix_operator] = ACTIONS(1767),
    [aux_sym_sign_operator_token1] = ACTIONS(1765),
    [sym_hash_operator] = ACTIONS(1767),
    [anon_sym_lsl] = ACTIONS(1765),
    [anon_sym_lsr] = ACTIONS(1765),
    [anon_sym_asr] = ACTIONS(1765),
    [aux_sym_pow_operator_token1] = ACTIONS(1767),
    [anon_sym_mod] = ACTIONS(1765),
    [anon_sym_land] = ACTIONS(1765),
    [anon_sym_lor] = ACTIONS(1765),
    [anon_sym_lxor] = ACTIONS(1765),
    [aux_sym_mult_operator_token1] = ACTIONS(1765),
    [aux_sym_add_operator_token1] = ACTIONS(1765),
    [sym_concat_operator] = ACTIONS(1767),
    [sym_rel_operator] = ACTIONS(1765),
    [anon_sym_AMP_AMP] = ACTIONS(1765),
    [anon_sym_or] = ACTIONS(1765),
    [anon_sym_PIPE_PIPE] = ACTIONS(1765),
    [sym_assign_operator] = ACTIONS(1767),
    [sym__uppercase_identifier] = ACTIONS(1767),
    [aux_sym_tag_token1] = ACTIONS(1767),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(810)] = {
    [sym_attribute] = STATE(810),
    [sym__lowercase_identifier] = ACTIONS(1769),
    [anon_sym_and] = ACTIONS(1769),
    [anon_sym_TILDE] = ACTIONS(1769),
    [anon_sym_QMARK] = ACTIONS(1769),
    [anon_sym_LPAREN] = ACTIONS(1771),
    [anon_sym_RPAREN] = ACTIONS(1771),
    [anon_sym_COMMA] = ACTIONS(1771),
    [anon_sym_PIPE] = ACTIONS(1769),
    [anon_sym_COLON2] = ACTIONS(1769),
    [anon_sym_DOT] = ACTIONS(1771),
    [anon_sym_DASH_GT] = ACTIONS(1769),
    [anon_sym_LBRACK] = ACTIONS(1769),
    [anon_sym_RBRACK] = ACTIONS(1771),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1769),
    [anon_sym_SEMI] = ACTIONS(1771),
    [anon_sym_RBRACE] = ACTIONS(1771),
    [anon_sym_constraint] = ACTIONS(1769),
    [anon_sym_val] = ACTIONS(1769),
    [anon_sym_end] = ACTIONS(1769),
    [anon_sym_with] = ACTIONS(1769),
    [anon_sym_COLON_GT] = ACTIONS(1771),
    [anon_sym_object] = ACTIONS(1769),
    [anon_sym_inherit] = ACTIONS(1769),
    [anon_sym_method] = ACTIONS(1769),
    [anon_sym_in] = ACTIONS(1769),
    [anon_sym_as] = ACTIONS(1769),
    [anon_sym_initializer] = ACTIONS(1769),
    [anon_sym_AMP] = ACTIONS(1769),
    [anon_sym_POUND] = ACTIONS(1769),
    [anon_sym_COLON_COLON] = ACTIONS(1771),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1771),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1771),
    [anon_sym_then] = ACTIONS(1769),
    [anon_sym_else] = ACTIONS(1769),
    [anon_sym_do] = ACTIONS(1769),
    [anon_sym_done] = ACTIONS(1769),
    [anon_sym_to] = ACTIONS(1769),
    [anon_sym_downto] = ACTIONS(1769),
    [anon_sym_new] = ACTIONS(1769),
    [anon_sym_LBRACE_LT] = ACTIONS(1771),
    [anon_sym_GT_RBRACE] = ACTIONS(1771),
    [anon_sym_begin] = ACTIONS(1769),
    [sym_ocamlyacc_value] = ACTIONS(1771),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1769),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1771),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1769),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1769),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1771),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1771),
    [aux_sym_number_token1] = ACTIONS(1771),
    [aux_sym_signed_number_token1] = ACTIONS(1769),
    [anon_sym_SQUOTE] = ACTIONS(1771),
    [anon_sym_DQUOTE] = ACTIONS(1771),
    [sym_prefix_operator] = ACTIONS(1771),
    [aux_sym_sign_operator_token1] = ACTIONS(1769),
    [sym_hash_operator] = ACTIONS(1771),
    [anon_sym_lsl] = ACTIONS(1769),
    [anon_sym_lsr] = ACTIONS(1769),
    [anon_sym_asr] = ACTIONS(1769),
    [aux_sym_pow_operator_token1] = ACTIONS(1771),
    [anon_sym_mod] = ACTIONS(1769),
    [anon_sym_land] = ACTIONS(1769),
    [anon_sym_lor] = ACTIONS(1769),
    [anon_sym_lxor] = ACTIONS(1769),
    [aux_sym_mult_operator_token1] = ACTIONS(1769),
    [aux_sym_add_operator_token1] = ACTIONS(1769),
    [sym_concat_operator] = ACTIONS(1771),
    [sym_rel_operator] = ACTIONS(1769),
    [anon_sym_AMP_AMP] = ACTIONS(1769),
    [anon_sym_or] = ACTIONS(1769),
    [anon_sym_PIPE_PIPE] = ACTIONS(1769),
    [sym_assign_operator] = ACTIONS(1771),
    [sym__uppercase_identifier] = ACTIONS(1771),
    [aux_sym_tag_token1] = ACTIONS(1771),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(811)] = {
    [sym_attribute] = STATE(811),
    [sym__lowercase_identifier] = ACTIONS(1773),
    [anon_sym_and] = ACTIONS(1773),
    [anon_sym_TILDE] = ACTIONS(1773),
    [anon_sym_QMARK] = ACTIONS(1773),
    [anon_sym_LPAREN] = ACTIONS(1775),
    [anon_sym_RPAREN] = ACTIONS(1775),
    [anon_sym_COMMA] = ACTIONS(1775),
    [anon_sym_PIPE] = ACTIONS(1773),
    [anon_sym_COLON2] = ACTIONS(1773),
    [anon_sym_DOT] = ACTIONS(1775),
    [anon_sym_DASH_GT] = ACTIONS(1773),
    [anon_sym_LBRACK] = ACTIONS(1773),
    [anon_sym_RBRACK] = ACTIONS(1775),
    [anon_sym_true] = ACTIONS(1773),
    [anon_sym_false] = ACTIONS(1773),
    [anon_sym_LBRACE] = ACTIONS(1773),
    [anon_sym_SEMI] = ACTIONS(1775),
    [anon_sym_RBRACE] = ACTIONS(1775),
    [anon_sym_constraint] = ACTIONS(1773),
    [anon_sym_val] = ACTIONS(1773),
    [anon_sym_end] = ACTIONS(1773),
    [anon_sym_with] = ACTIONS(1773),
    [anon_sym_COLON_GT] = ACTIONS(1775),
    [anon_sym_object] = ACTIONS(1773),
    [anon_sym_inherit] = ACTIONS(1773),
    [anon_sym_method] = ACTIONS(1773),
    [anon_sym_in] = ACTIONS(1773),
    [anon_sym_as] = ACTIONS(1773),
    [anon_sym_initializer] = ACTIONS(1773),
    [anon_sym_AMP] = ACTIONS(1773),
    [anon_sym_POUND] = ACTIONS(1773),
    [anon_sym_COLON_COLON] = ACTIONS(1775),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1775),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1775),
    [anon_sym_then] = ACTIONS(1773),
    [anon_sym_else] = ACTIONS(1773),
    [anon_sym_do] = ACTIONS(1773),
    [anon_sym_done] = ACTIONS(1773),
    [anon_sym_to] = ACTIONS(1773),
    [anon_sym_downto] = ACTIONS(1773),
    [anon_sym_new] = ACTIONS(1773),
    [anon_sym_LBRACE_LT] = ACTIONS(1775),
    [anon_sym_GT_RBRACE] = ACTIONS(1775),
    [anon_sym_begin] = ACTIONS(1773),
    [sym_ocamlyacc_value] = ACTIONS(1775),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1773),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1775),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1773),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1773),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1775),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1775),
    [aux_sym_number_token1] = ACTIONS(1775),
    [aux_sym_signed_number_token1] = ACTIONS(1773),
    [anon_sym_SQUOTE] = ACTIONS(1775),
    [anon_sym_DQUOTE] = ACTIONS(1775),
    [sym_prefix_operator] = ACTIONS(1775),
    [aux_sym_sign_operator_token1] = ACTIONS(1773),
    [sym_hash_operator] = ACTIONS(1775),
    [anon_sym_lsl] = ACTIONS(1773),
    [anon_sym_lsr] = ACTIONS(1773),
    [anon_sym_asr] = ACTIONS(1773),
    [aux_sym_pow_operator_token1] = ACTIONS(1775),
    [anon_sym_mod] = ACTIONS(1773),
    [anon_sym_land] = ACTIONS(1773),
    [anon_sym_lor] = ACTIONS(1773),
    [anon_sym_lxor] = ACTIONS(1773),
    [aux_sym_mult_operator_token1] = ACTIONS(1773),
    [aux_sym_add_operator_token1] = ACTIONS(1773),
    [sym_concat_operator] = ACTIONS(1775),
    [sym_rel_operator] = ACTIONS(1773),
    [anon_sym_AMP_AMP] = ACTIONS(1773),
    [anon_sym_or] = ACTIONS(1773),
    [anon_sym_PIPE_PIPE] = ACTIONS(1773),
    [sym_assign_operator] = ACTIONS(1775),
    [sym__uppercase_identifier] = ACTIONS(1775),
    [aux_sym_tag_token1] = ACTIONS(1775),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(812)] = {
    [sym_attribute] = STATE(812),
    [sym__lowercase_identifier] = ACTIONS(1777),
    [anon_sym_and] = ACTIONS(1777),
    [anon_sym_TILDE] = ACTIONS(1777),
    [anon_sym_QMARK] = ACTIONS(1777),
    [anon_sym_LPAREN] = ACTIONS(1779),
    [anon_sym_RPAREN] = ACTIONS(1779),
    [anon_sym_COMMA] = ACTIONS(1779),
    [anon_sym_PIPE] = ACTIONS(1777),
    [anon_sym_COLON2] = ACTIONS(1777),
    [anon_sym_DOT] = ACTIONS(1779),
    [anon_sym_DASH_GT] = ACTIONS(1777),
    [anon_sym_LBRACK] = ACTIONS(1777),
    [anon_sym_RBRACK] = ACTIONS(1779),
    [anon_sym_true] = ACTIONS(1777),
    [anon_sym_false] = ACTIONS(1777),
    [anon_sym_LBRACE] = ACTIONS(1777),
    [anon_sym_SEMI] = ACTIONS(1779),
    [anon_sym_RBRACE] = ACTIONS(1779),
    [anon_sym_constraint] = ACTIONS(1777),
    [anon_sym_val] = ACTIONS(1777),
    [anon_sym_end] = ACTIONS(1777),
    [anon_sym_with] = ACTIONS(1777),
    [anon_sym_COLON_GT] = ACTIONS(1779),
    [anon_sym_object] = ACTIONS(1777),
    [anon_sym_inherit] = ACTIONS(1777),
    [anon_sym_method] = ACTIONS(1777),
    [anon_sym_in] = ACTIONS(1777),
    [anon_sym_as] = ACTIONS(1777),
    [anon_sym_initializer] = ACTIONS(1777),
    [anon_sym_AMP] = ACTIONS(1777),
    [anon_sym_POUND] = ACTIONS(1777),
    [anon_sym_COLON_COLON] = ACTIONS(1779),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1779),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1779),
    [anon_sym_then] = ACTIONS(1777),
    [anon_sym_else] = ACTIONS(1777),
    [anon_sym_do] = ACTIONS(1777),
    [anon_sym_done] = ACTIONS(1777),
    [anon_sym_to] = ACTIONS(1777),
    [anon_sym_downto] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1777),
    [anon_sym_LBRACE_LT] = ACTIONS(1779),
    [anon_sym_GT_RBRACE] = ACTIONS(1779),
    [anon_sym_begin] = ACTIONS(1777),
    [sym_ocamlyacc_value] = ACTIONS(1779),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1777),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1779),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1777),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1777),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1779),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1779),
    [aux_sym_number_token1] = ACTIONS(1779),
    [aux_sym_signed_number_token1] = ACTIONS(1777),
    [anon_sym_SQUOTE] = ACTIONS(1779),
    [anon_sym_DQUOTE] = ACTIONS(1779),
    [sym_prefix_operator] = ACTIONS(1779),
    [aux_sym_sign_operator_token1] = ACTIONS(1777),
    [sym_hash_operator] = ACTIONS(1779),
    [anon_sym_lsl] = ACTIONS(1777),
    [anon_sym_lsr] = ACTIONS(1777),
    [anon_sym_asr] = ACTIONS(1777),
    [aux_sym_pow_operator_token1] = ACTIONS(1779),
    [anon_sym_mod] = ACTIONS(1777),
    [anon_sym_land] = ACTIONS(1777),
    [anon_sym_lor] = ACTIONS(1777),
    [anon_sym_lxor] = ACTIONS(1777),
    [aux_sym_mult_operator_token1] = ACTIONS(1777),
    [aux_sym_add_operator_token1] = ACTIONS(1777),
    [sym_concat_operator] = ACTIONS(1779),
    [sym_rel_operator] = ACTIONS(1777),
    [anon_sym_AMP_AMP] = ACTIONS(1777),
    [anon_sym_or] = ACTIONS(1777),
    [anon_sym_PIPE_PIPE] = ACTIONS(1777),
    [sym_assign_operator] = ACTIONS(1779),
    [sym__uppercase_identifier] = ACTIONS(1779),
    [aux_sym_tag_token1] = ACTIONS(1779),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(813)] = {
    [sym_attribute] = STATE(813),
    [sym__lowercase_identifier] = ACTIONS(1781),
    [anon_sym_and] = ACTIONS(1781),
    [anon_sym_TILDE] = ACTIONS(1781),
    [anon_sym_QMARK] = ACTIONS(1781),
    [anon_sym_LPAREN] = ACTIONS(1783),
    [anon_sym_RPAREN] = ACTIONS(1783),
    [anon_sym_COMMA] = ACTIONS(1783),
    [anon_sym_PIPE] = ACTIONS(1781),
    [anon_sym_COLON2] = ACTIONS(1781),
    [anon_sym_DOT] = ACTIONS(1783),
    [anon_sym_DASH_GT] = ACTIONS(1781),
    [anon_sym_LBRACK] = ACTIONS(1781),
    [anon_sym_RBRACK] = ACTIONS(1783),
    [anon_sym_true] = ACTIONS(1781),
    [anon_sym_false] = ACTIONS(1781),
    [anon_sym_LBRACE] = ACTIONS(1781),
    [anon_sym_SEMI] = ACTIONS(1783),
    [anon_sym_RBRACE] = ACTIONS(1783),
    [anon_sym_constraint] = ACTIONS(1781),
    [anon_sym_val] = ACTIONS(1781),
    [anon_sym_end] = ACTIONS(1781),
    [anon_sym_with] = ACTIONS(1781),
    [anon_sym_COLON_GT] = ACTIONS(1783),
    [anon_sym_object] = ACTIONS(1781),
    [anon_sym_inherit] = ACTIONS(1781),
    [anon_sym_method] = ACTIONS(1781),
    [anon_sym_in] = ACTIONS(1781),
    [anon_sym_as] = ACTIONS(1781),
    [anon_sym_initializer] = ACTIONS(1781),
    [anon_sym_AMP] = ACTIONS(1781),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_COLON_COLON] = ACTIONS(1783),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1783),
    [anon_sym_then] = ACTIONS(1781),
    [anon_sym_else] = ACTIONS(1781),
    [anon_sym_do] = ACTIONS(1781),
    [anon_sym_done] = ACTIONS(1781),
    [anon_sym_to] = ACTIONS(1781),
    [anon_sym_downto] = ACTIONS(1781),
    [anon_sym_new] = ACTIONS(1781),
    [anon_sym_LBRACE_LT] = ACTIONS(1783),
    [anon_sym_GT_RBRACE] = ACTIONS(1783),
    [anon_sym_begin] = ACTIONS(1781),
    [sym_ocamlyacc_value] = ACTIONS(1783),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1781),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1783),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1781),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1781),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1783),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1783),
    [aux_sym_number_token1] = ACTIONS(1783),
    [aux_sym_signed_number_token1] = ACTIONS(1781),
    [anon_sym_SQUOTE] = ACTIONS(1783),
    [anon_sym_DQUOTE] = ACTIONS(1783),
    [sym_prefix_operator] = ACTIONS(1783),
    [aux_sym_sign_operator_token1] = ACTIONS(1781),
    [sym_hash_operator] = ACTIONS(1783),
    [anon_sym_lsl] = ACTIONS(1781),
    [anon_sym_lsr] = ACTIONS(1781),
    [anon_sym_asr] = ACTIONS(1781),
    [aux_sym_pow_operator_token1] = ACTIONS(1783),
    [anon_sym_mod] = ACTIONS(1781),
    [anon_sym_land] = ACTIONS(1781),
    [anon_sym_lor] = ACTIONS(1781),
    [anon_sym_lxor] = ACTIONS(1781),
    [aux_sym_mult_operator_token1] = ACTIONS(1781),
    [aux_sym_add_operator_token1] = ACTIONS(1781),
    [sym_concat_operator] = ACTIONS(1783),
    [sym_rel_operator] = ACTIONS(1781),
    [anon_sym_AMP_AMP] = ACTIONS(1781),
    [anon_sym_or] = ACTIONS(1781),
    [anon_sym_PIPE_PIPE] = ACTIONS(1781),
    [sym_assign_operator] = ACTIONS(1783),
    [sym__uppercase_identifier] = ACTIONS(1783),
    [aux_sym_tag_token1] = ACTIONS(1783),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(814)] = {
    [sym_attribute] = STATE(814),
    [sym__lowercase_identifier] = ACTIONS(1785),
    [anon_sym_and] = ACTIONS(1785),
    [anon_sym_TILDE] = ACTIONS(1785),
    [anon_sym_QMARK] = ACTIONS(1785),
    [anon_sym_LPAREN] = ACTIONS(1787),
    [anon_sym_RPAREN] = ACTIONS(1787),
    [anon_sym_COMMA] = ACTIONS(1787),
    [anon_sym_PIPE] = ACTIONS(1785),
    [anon_sym_COLON2] = ACTIONS(1785),
    [anon_sym_DOT] = ACTIONS(1787),
    [anon_sym_DASH_GT] = ACTIONS(1785),
    [anon_sym_LBRACK] = ACTIONS(1785),
    [anon_sym_RBRACK] = ACTIONS(1787),
    [anon_sym_true] = ACTIONS(1785),
    [anon_sym_false] = ACTIONS(1785),
    [anon_sym_LBRACE] = ACTIONS(1785),
    [anon_sym_SEMI] = ACTIONS(1787),
    [anon_sym_RBRACE] = ACTIONS(1787),
    [anon_sym_constraint] = ACTIONS(1785),
    [anon_sym_val] = ACTIONS(1785),
    [anon_sym_end] = ACTIONS(1785),
    [anon_sym_with] = ACTIONS(1785),
    [anon_sym_COLON_GT] = ACTIONS(1787),
    [anon_sym_object] = ACTIONS(1785),
    [anon_sym_inherit] = ACTIONS(1785),
    [anon_sym_method] = ACTIONS(1785),
    [anon_sym_in] = ACTIONS(1785),
    [anon_sym_as] = ACTIONS(1785),
    [anon_sym_initializer] = ACTIONS(1785),
    [anon_sym_AMP] = ACTIONS(1785),
    [anon_sym_POUND] = ACTIONS(1785),
    [anon_sym_COLON_COLON] = ACTIONS(1787),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1787),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1787),
    [anon_sym_then] = ACTIONS(1785),
    [anon_sym_else] = ACTIONS(1785),
    [anon_sym_do] = ACTIONS(1785),
    [anon_sym_done] = ACTIONS(1785),
    [anon_sym_to] = ACTIONS(1785),
    [anon_sym_downto] = ACTIONS(1785),
    [anon_sym_new] = ACTIONS(1785),
    [anon_sym_LBRACE_LT] = ACTIONS(1787),
    [anon_sym_GT_RBRACE] = ACTIONS(1787),
    [anon_sym_begin] = ACTIONS(1785),
    [sym_ocamlyacc_value] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1785),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1787),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1785),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1785),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1787),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1787),
    [aux_sym_number_token1] = ACTIONS(1787),
    [aux_sym_signed_number_token1] = ACTIONS(1785),
    [anon_sym_SQUOTE] = ACTIONS(1787),
    [anon_sym_DQUOTE] = ACTIONS(1787),
    [sym_prefix_operator] = ACTIONS(1787),
    [aux_sym_sign_operator_token1] = ACTIONS(1785),
    [sym_hash_operator] = ACTIONS(1787),
    [anon_sym_lsl] = ACTIONS(1785),
    [anon_sym_lsr] = ACTIONS(1785),
    [anon_sym_asr] = ACTIONS(1785),
    [aux_sym_pow_operator_token1] = ACTIONS(1787),
    [anon_sym_mod] = ACTIONS(1785),
    [anon_sym_land] = ACTIONS(1785),
    [anon_sym_lor] = ACTIONS(1785),
    [anon_sym_lxor] = ACTIONS(1785),
    [aux_sym_mult_operator_token1] = ACTIONS(1785),
    [aux_sym_add_operator_token1] = ACTIONS(1785),
    [sym_concat_operator] = ACTIONS(1787),
    [sym_rel_operator] = ACTIONS(1785),
    [anon_sym_AMP_AMP] = ACTIONS(1785),
    [anon_sym_or] = ACTIONS(1785),
    [anon_sym_PIPE_PIPE] = ACTIONS(1785),
    [sym_assign_operator] = ACTIONS(1787),
    [sym__uppercase_identifier] = ACTIONS(1787),
    [aux_sym_tag_token1] = ACTIONS(1787),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(815)] = {
    [sym_attribute] = STATE(815),
    [sym__lowercase_identifier] = ACTIONS(1789),
    [anon_sym_and] = ACTIONS(1789),
    [anon_sym_TILDE] = ACTIONS(1789),
    [anon_sym_QMARK] = ACTIONS(1789),
    [anon_sym_LPAREN] = ACTIONS(1791),
    [anon_sym_RPAREN] = ACTIONS(1791),
    [anon_sym_COMMA] = ACTIONS(1791),
    [anon_sym_PIPE] = ACTIONS(1789),
    [anon_sym_COLON2] = ACTIONS(1789),
    [anon_sym_DOT] = ACTIONS(1791),
    [anon_sym_DASH_GT] = ACTIONS(1789),
    [anon_sym_LBRACK] = ACTIONS(1789),
    [anon_sym_RBRACK] = ACTIONS(1791),
    [anon_sym_true] = ACTIONS(1789),
    [anon_sym_false] = ACTIONS(1789),
    [anon_sym_LBRACE] = ACTIONS(1789),
    [anon_sym_SEMI] = ACTIONS(1791),
    [anon_sym_RBRACE] = ACTIONS(1791),
    [anon_sym_constraint] = ACTIONS(1789),
    [anon_sym_val] = ACTIONS(1789),
    [anon_sym_end] = ACTIONS(1789),
    [anon_sym_with] = ACTIONS(1789),
    [anon_sym_COLON_GT] = ACTIONS(1791),
    [anon_sym_object] = ACTIONS(1789),
    [anon_sym_inherit] = ACTIONS(1789),
    [anon_sym_method] = ACTIONS(1789),
    [anon_sym_in] = ACTIONS(1789),
    [anon_sym_as] = ACTIONS(1789),
    [anon_sym_initializer] = ACTIONS(1789),
    [anon_sym_AMP] = ACTIONS(1789),
    [anon_sym_POUND] = ACTIONS(1789),
    [anon_sym_COLON_COLON] = ACTIONS(1791),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1791),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1791),
    [anon_sym_then] = ACTIONS(1789),
    [anon_sym_else] = ACTIONS(1789),
    [anon_sym_do] = ACTIONS(1789),
    [anon_sym_done] = ACTIONS(1789),
    [anon_sym_to] = ACTIONS(1789),
    [anon_sym_downto] = ACTIONS(1789),
    [anon_sym_new] = ACTIONS(1789),
    [anon_sym_LBRACE_LT] = ACTIONS(1791),
    [anon_sym_GT_RBRACE] = ACTIONS(1791),
    [anon_sym_begin] = ACTIONS(1789),
    [sym_ocamlyacc_value] = ACTIONS(1791),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1791),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1789),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1791),
    [aux_sym_number_token1] = ACTIONS(1791),
    [aux_sym_signed_number_token1] = ACTIONS(1789),
    [anon_sym_SQUOTE] = ACTIONS(1791),
    [anon_sym_DQUOTE] = ACTIONS(1791),
    [sym_prefix_operator] = ACTIONS(1791),
    [aux_sym_sign_operator_token1] = ACTIONS(1789),
    [sym_hash_operator] = ACTIONS(1791),
    [anon_sym_lsl] = ACTIONS(1789),
    [anon_sym_lsr] = ACTIONS(1789),
    [anon_sym_asr] = ACTIONS(1789),
    [aux_sym_pow_operator_token1] = ACTIONS(1791),
    [anon_sym_mod] = ACTIONS(1789),
    [anon_sym_land] = ACTIONS(1789),
    [anon_sym_lor] = ACTIONS(1789),
    [anon_sym_lxor] = ACTIONS(1789),
    [aux_sym_mult_operator_token1] = ACTIONS(1789),
    [aux_sym_add_operator_token1] = ACTIONS(1789),
    [sym_concat_operator] = ACTIONS(1791),
    [sym_rel_operator] = ACTIONS(1789),
    [anon_sym_AMP_AMP] = ACTIONS(1789),
    [anon_sym_or] = ACTIONS(1789),
    [anon_sym_PIPE_PIPE] = ACTIONS(1789),
    [sym_assign_operator] = ACTIONS(1791),
    [sym__uppercase_identifier] = ACTIONS(1791),
    [aux_sym_tag_token1] = ACTIONS(1791),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(816)] = {
    [sym_attribute] = STATE(816),
    [sym__lowercase_identifier] = ACTIONS(1793),
    [anon_sym_and] = ACTIONS(1793),
    [anon_sym_TILDE] = ACTIONS(1793),
    [anon_sym_QMARK] = ACTIONS(1793),
    [anon_sym_LPAREN] = ACTIONS(1795),
    [anon_sym_RPAREN] = ACTIONS(1795),
    [anon_sym_COMMA] = ACTIONS(1795),
    [anon_sym_PIPE] = ACTIONS(1793),
    [anon_sym_COLON2] = ACTIONS(1793),
    [anon_sym_DOT] = ACTIONS(1795),
    [anon_sym_DASH_GT] = ACTIONS(1793),
    [anon_sym_LBRACK] = ACTIONS(1793),
    [anon_sym_RBRACK] = ACTIONS(1795),
    [anon_sym_true] = ACTIONS(1793),
    [anon_sym_false] = ACTIONS(1793),
    [anon_sym_LBRACE] = ACTIONS(1793),
    [anon_sym_SEMI] = ACTIONS(1795),
    [anon_sym_RBRACE] = ACTIONS(1795),
    [anon_sym_constraint] = ACTIONS(1793),
    [anon_sym_val] = ACTIONS(1793),
    [anon_sym_end] = ACTIONS(1793),
    [anon_sym_with] = ACTIONS(1793),
    [anon_sym_COLON_GT] = ACTIONS(1795),
    [anon_sym_object] = ACTIONS(1793),
    [anon_sym_inherit] = ACTIONS(1793),
    [anon_sym_method] = ACTIONS(1793),
    [anon_sym_in] = ACTIONS(1793),
    [anon_sym_as] = ACTIONS(1793),
    [anon_sym_initializer] = ACTIONS(1793),
    [anon_sym_AMP] = ACTIONS(1793),
    [anon_sym_POUND] = ACTIONS(1793),
    [anon_sym_COLON_COLON] = ACTIONS(1795),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1795),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1795),
    [anon_sym_then] = ACTIONS(1793),
    [anon_sym_else] = ACTIONS(1793),
    [anon_sym_do] = ACTIONS(1793),
    [anon_sym_done] = ACTIONS(1793),
    [anon_sym_to] = ACTIONS(1793),
    [anon_sym_downto] = ACTIONS(1793),
    [anon_sym_new] = ACTIONS(1793),
    [anon_sym_LBRACE_LT] = ACTIONS(1795),
    [anon_sym_GT_RBRACE] = ACTIONS(1795),
    [anon_sym_begin] = ACTIONS(1793),
    [sym_ocamlyacc_value] = ACTIONS(1795),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1793),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1795),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1795),
    [aux_sym_number_token1] = ACTIONS(1795),
    [aux_sym_signed_number_token1] = ACTIONS(1793),
    [anon_sym_SQUOTE] = ACTIONS(1795),
    [anon_sym_DQUOTE] = ACTIONS(1795),
    [sym_prefix_operator] = ACTIONS(1795),
    [aux_sym_sign_operator_token1] = ACTIONS(1793),
    [sym_hash_operator] = ACTIONS(1795),
    [anon_sym_lsl] = ACTIONS(1793),
    [anon_sym_lsr] = ACTIONS(1793),
    [anon_sym_asr] = ACTIONS(1793),
    [aux_sym_pow_operator_token1] = ACTIONS(1795),
    [anon_sym_mod] = ACTIONS(1793),
    [anon_sym_land] = ACTIONS(1793),
    [anon_sym_lor] = ACTIONS(1793),
    [anon_sym_lxor] = ACTIONS(1793),
    [aux_sym_mult_operator_token1] = ACTIONS(1793),
    [aux_sym_add_operator_token1] = ACTIONS(1793),
    [sym_concat_operator] = ACTIONS(1795),
    [sym_rel_operator] = ACTIONS(1793),
    [anon_sym_AMP_AMP] = ACTIONS(1793),
    [anon_sym_or] = ACTIONS(1793),
    [anon_sym_PIPE_PIPE] = ACTIONS(1793),
    [sym_assign_operator] = ACTIONS(1795),
    [sym__uppercase_identifier] = ACTIONS(1795),
    [aux_sym_tag_token1] = ACTIONS(1795),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(817)] = {
    [sym_attribute] = STATE(817),
    [sym__lowercase_identifier] = ACTIONS(1797),
    [anon_sym_and] = ACTIONS(1797),
    [anon_sym_TILDE] = ACTIONS(1797),
    [anon_sym_QMARK] = ACTIONS(1797),
    [anon_sym_LPAREN] = ACTIONS(1799),
    [anon_sym_RPAREN] = ACTIONS(1799),
    [anon_sym_COMMA] = ACTIONS(1799),
    [anon_sym_PIPE] = ACTIONS(1797),
    [anon_sym_COLON2] = ACTIONS(1797),
    [anon_sym_DOT] = ACTIONS(1799),
    [anon_sym_DASH_GT] = ACTIONS(1797),
    [anon_sym_LBRACK] = ACTIONS(1797),
    [anon_sym_RBRACK] = ACTIONS(1799),
    [anon_sym_true] = ACTIONS(1797),
    [anon_sym_false] = ACTIONS(1797),
    [anon_sym_LBRACE] = ACTIONS(1797),
    [anon_sym_SEMI] = ACTIONS(1799),
    [anon_sym_RBRACE] = ACTIONS(1799),
    [anon_sym_constraint] = ACTIONS(1797),
    [anon_sym_val] = ACTIONS(1797),
    [anon_sym_end] = ACTIONS(1797),
    [anon_sym_with] = ACTIONS(1797),
    [anon_sym_COLON_GT] = ACTIONS(1799),
    [anon_sym_object] = ACTIONS(1797),
    [anon_sym_inherit] = ACTIONS(1797),
    [anon_sym_method] = ACTIONS(1797),
    [anon_sym_in] = ACTIONS(1797),
    [anon_sym_as] = ACTIONS(1797),
    [anon_sym_initializer] = ACTIONS(1797),
    [anon_sym_AMP] = ACTIONS(1797),
    [anon_sym_POUND] = ACTIONS(1797),
    [anon_sym_COLON_COLON] = ACTIONS(1799),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1799),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1799),
    [anon_sym_then] = ACTIONS(1797),
    [anon_sym_else] = ACTIONS(1797),
    [anon_sym_do] = ACTIONS(1797),
    [anon_sym_done] = ACTIONS(1797),
    [anon_sym_to] = ACTIONS(1797),
    [anon_sym_downto] = ACTIONS(1797),
    [anon_sym_new] = ACTIONS(1797),
    [anon_sym_LBRACE_LT] = ACTIONS(1799),
    [anon_sym_GT_RBRACE] = ACTIONS(1799),
    [anon_sym_begin] = ACTIONS(1797),
    [sym_ocamlyacc_value] = ACTIONS(1799),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1797),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1799),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1797),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1797),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1799),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1799),
    [aux_sym_signed_number_token1] = ACTIONS(1797),
    [anon_sym_SQUOTE] = ACTIONS(1799),
    [anon_sym_DQUOTE] = ACTIONS(1799),
    [sym_prefix_operator] = ACTIONS(1799),
    [aux_sym_sign_operator_token1] = ACTIONS(1797),
    [sym_hash_operator] = ACTIONS(1799),
    [anon_sym_lsl] = ACTIONS(1797),
    [anon_sym_lsr] = ACTIONS(1797),
    [anon_sym_asr] = ACTIONS(1797),
    [aux_sym_pow_operator_token1] = ACTIONS(1799),
    [anon_sym_mod] = ACTIONS(1797),
    [anon_sym_land] = ACTIONS(1797),
    [anon_sym_lor] = ACTIONS(1797),
    [anon_sym_lxor] = ACTIONS(1797),
    [aux_sym_mult_operator_token1] = ACTIONS(1797),
    [aux_sym_add_operator_token1] = ACTIONS(1797),
    [sym_concat_operator] = ACTIONS(1799),
    [sym_rel_operator] = ACTIONS(1797),
    [anon_sym_AMP_AMP] = ACTIONS(1797),
    [anon_sym_or] = ACTIONS(1797),
    [anon_sym_PIPE_PIPE] = ACTIONS(1797),
    [sym_assign_operator] = ACTIONS(1799),
    [sym__uppercase_identifier] = ACTIONS(1799),
    [aux_sym_tag_token1] = ACTIONS(1799),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(818)] = {
    [sym_attribute] = STATE(818),
    [sym__lowercase_identifier] = ACTIONS(1801),
    [anon_sym_and] = ACTIONS(1801),
    [anon_sym_TILDE] = ACTIONS(1801),
    [anon_sym_QMARK] = ACTIONS(1801),
    [anon_sym_LPAREN] = ACTIONS(1803),
    [anon_sym_RPAREN] = ACTIONS(1803),
    [anon_sym_COMMA] = ACTIONS(1803),
    [anon_sym_PIPE] = ACTIONS(1801),
    [anon_sym_COLON2] = ACTIONS(1801),
    [anon_sym_DOT] = ACTIONS(1803),
    [anon_sym_DASH_GT] = ACTIONS(1801),
    [anon_sym_LBRACK] = ACTIONS(1801),
    [anon_sym_RBRACK] = ACTIONS(1803),
    [anon_sym_true] = ACTIONS(1801),
    [anon_sym_false] = ACTIONS(1801),
    [anon_sym_LBRACE] = ACTIONS(1801),
    [anon_sym_SEMI] = ACTIONS(1803),
    [anon_sym_RBRACE] = ACTIONS(1803),
    [anon_sym_constraint] = ACTIONS(1801),
    [anon_sym_val] = ACTIONS(1801),
    [anon_sym_end] = ACTIONS(1801),
    [anon_sym_with] = ACTIONS(1801),
    [anon_sym_COLON_GT] = ACTIONS(1803),
    [anon_sym_object] = ACTIONS(1801),
    [anon_sym_inherit] = ACTIONS(1801),
    [anon_sym_method] = ACTIONS(1801),
    [anon_sym_in] = ACTIONS(1801),
    [anon_sym_as] = ACTIONS(1801),
    [anon_sym_initializer] = ACTIONS(1801),
    [anon_sym_AMP] = ACTIONS(1801),
    [anon_sym_POUND] = ACTIONS(1801),
    [anon_sym_COLON_COLON] = ACTIONS(1803),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1803),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1803),
    [anon_sym_then] = ACTIONS(1801),
    [anon_sym_else] = ACTIONS(1801),
    [anon_sym_do] = ACTIONS(1801),
    [anon_sym_done] = ACTIONS(1801),
    [anon_sym_to] = ACTIONS(1801),
    [anon_sym_downto] = ACTIONS(1801),
    [anon_sym_new] = ACTIONS(1801),
    [anon_sym_LBRACE_LT] = ACTIONS(1803),
    [anon_sym_GT_RBRACE] = ACTIONS(1803),
    [anon_sym_begin] = ACTIONS(1801),
    [sym_ocamlyacc_value] = ACTIONS(1803),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1801),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1803),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1801),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1801),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1803),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1803),
    [aux_sym_number_token1] = ACTIONS(1803),
    [aux_sym_signed_number_token1] = ACTIONS(1801),
    [anon_sym_SQUOTE] = ACTIONS(1803),
    [anon_sym_DQUOTE] = ACTIONS(1803),
    [sym_prefix_operator] = ACTIONS(1803),
    [aux_sym_sign_operator_token1] = ACTIONS(1801),
    [sym_hash_operator] = ACTIONS(1803),
    [anon_sym_lsl] = ACTIONS(1801),
    [anon_sym_lsr] = ACTIONS(1801),
    [anon_sym_asr] = ACTIONS(1801),
    [aux_sym_pow_operator_token1] = ACTIONS(1803),
    [anon_sym_mod] = ACTIONS(1801),
    [anon_sym_land] = ACTIONS(1801),
    [anon_sym_lor] = ACTIONS(1801),
    [anon_sym_lxor] = ACTIONS(1801),
    [aux_sym_mult_operator_token1] = ACTIONS(1801),
    [aux_sym_add_operator_token1] = ACTIONS(1801),
    [sym_concat_operator] = ACTIONS(1803),
    [sym_rel_operator] = ACTIONS(1801),
    [anon_sym_AMP_AMP] = ACTIONS(1801),
    [anon_sym_or] = ACTIONS(1801),
    [anon_sym_PIPE_PIPE] = ACTIONS(1801),
    [sym_assign_operator] = ACTIONS(1803),
    [sym__uppercase_identifier] = ACTIONS(1803),
    [aux_sym_tag_token1] = ACTIONS(1803),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(819)] = {
    [sym_attribute] = STATE(819),
    [sym__lowercase_identifier] = ACTIONS(1805),
    [anon_sym_and] = ACTIONS(1805),
    [anon_sym_TILDE] = ACTIONS(1805),
    [anon_sym_QMARK] = ACTIONS(1805),
    [anon_sym_LPAREN] = ACTIONS(1807),
    [anon_sym_RPAREN] = ACTIONS(1807),
    [anon_sym_COMMA] = ACTIONS(1807),
    [anon_sym_PIPE] = ACTIONS(1805),
    [anon_sym_COLON2] = ACTIONS(1805),
    [anon_sym_DOT] = ACTIONS(1807),
    [anon_sym_DASH_GT] = ACTIONS(1805),
    [anon_sym_LBRACK] = ACTIONS(1805),
    [anon_sym_RBRACK] = ACTIONS(1807),
    [anon_sym_true] = ACTIONS(1805),
    [anon_sym_false] = ACTIONS(1805),
    [anon_sym_LBRACE] = ACTIONS(1805),
    [anon_sym_SEMI] = ACTIONS(1807),
    [anon_sym_RBRACE] = ACTIONS(1807),
    [anon_sym_constraint] = ACTIONS(1805),
    [anon_sym_val] = ACTIONS(1805),
    [anon_sym_end] = ACTIONS(1805),
    [anon_sym_with] = ACTIONS(1805),
    [anon_sym_COLON_GT] = ACTIONS(1807),
    [anon_sym_object] = ACTIONS(1805),
    [anon_sym_inherit] = ACTIONS(1805),
    [anon_sym_method] = ACTIONS(1805),
    [anon_sym_in] = ACTIONS(1805),
    [anon_sym_as] = ACTIONS(1805),
    [anon_sym_initializer] = ACTIONS(1805),
    [anon_sym_AMP] = ACTIONS(1805),
    [anon_sym_POUND] = ACTIONS(1805),
    [anon_sym_COLON_COLON] = ACTIONS(1807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1807),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1807),
    [anon_sym_then] = ACTIONS(1805),
    [anon_sym_else] = ACTIONS(1805),
    [anon_sym_do] = ACTIONS(1805),
    [anon_sym_done] = ACTIONS(1805),
    [anon_sym_to] = ACTIONS(1805),
    [anon_sym_downto] = ACTIONS(1805),
    [anon_sym_new] = ACTIONS(1805),
    [anon_sym_LBRACE_LT] = ACTIONS(1807),
    [anon_sym_GT_RBRACE] = ACTIONS(1807),
    [anon_sym_begin] = ACTIONS(1805),
    [sym_ocamlyacc_value] = ACTIONS(1807),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1805),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1807),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1805),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1805),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1807),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1807),
    [aux_sym_number_token1] = ACTIONS(1807),
    [aux_sym_signed_number_token1] = ACTIONS(1805),
    [anon_sym_SQUOTE] = ACTIONS(1807),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym_prefix_operator] = ACTIONS(1807),
    [aux_sym_sign_operator_token1] = ACTIONS(1805),
    [sym_hash_operator] = ACTIONS(1807),
    [anon_sym_lsl] = ACTIONS(1805),
    [anon_sym_lsr] = ACTIONS(1805),
    [anon_sym_asr] = ACTIONS(1805),
    [aux_sym_pow_operator_token1] = ACTIONS(1807),
    [anon_sym_mod] = ACTIONS(1805),
    [anon_sym_land] = ACTIONS(1805),
    [anon_sym_lor] = ACTIONS(1805),
    [anon_sym_lxor] = ACTIONS(1805),
    [aux_sym_mult_operator_token1] = ACTIONS(1805),
    [aux_sym_add_operator_token1] = ACTIONS(1805),
    [sym_concat_operator] = ACTIONS(1807),
    [sym_rel_operator] = ACTIONS(1805),
    [anon_sym_AMP_AMP] = ACTIONS(1805),
    [anon_sym_or] = ACTIONS(1805),
    [anon_sym_PIPE_PIPE] = ACTIONS(1805),
    [sym_assign_operator] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1807),
    [aux_sym_tag_token1] = ACTIONS(1807),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(820)] = {
    [sym_attribute] = STATE(820),
    [sym__lowercase_identifier] = ACTIONS(1809),
    [anon_sym_and] = ACTIONS(1809),
    [anon_sym_TILDE] = ACTIONS(1809),
    [anon_sym_QMARK] = ACTIONS(1809),
    [anon_sym_LPAREN] = ACTIONS(1811),
    [anon_sym_RPAREN] = ACTIONS(1811),
    [anon_sym_COMMA] = ACTIONS(1811),
    [anon_sym_PIPE] = ACTIONS(1809),
    [anon_sym_COLON2] = ACTIONS(1809),
    [anon_sym_DOT] = ACTIONS(1811),
    [anon_sym_DASH_GT] = ACTIONS(1809),
    [anon_sym_LBRACK] = ACTIONS(1809),
    [anon_sym_RBRACK] = ACTIONS(1811),
    [anon_sym_true] = ACTIONS(1809),
    [anon_sym_false] = ACTIONS(1809),
    [anon_sym_LBRACE] = ACTIONS(1809),
    [anon_sym_SEMI] = ACTIONS(1811),
    [anon_sym_RBRACE] = ACTIONS(1811),
    [anon_sym_constraint] = ACTIONS(1809),
    [anon_sym_val] = ACTIONS(1809),
    [anon_sym_end] = ACTIONS(1809),
    [anon_sym_with] = ACTIONS(1809),
    [anon_sym_COLON_GT] = ACTIONS(1811),
    [anon_sym_object] = ACTIONS(1809),
    [anon_sym_inherit] = ACTIONS(1809),
    [anon_sym_method] = ACTIONS(1809),
    [anon_sym_in] = ACTIONS(1809),
    [anon_sym_as] = ACTIONS(1809),
    [anon_sym_initializer] = ACTIONS(1809),
    [anon_sym_AMP] = ACTIONS(1809),
    [anon_sym_POUND] = ACTIONS(1809),
    [anon_sym_COLON_COLON] = ACTIONS(1811),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1811),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1811),
    [anon_sym_then] = ACTIONS(1809),
    [anon_sym_else] = ACTIONS(1809),
    [anon_sym_do] = ACTIONS(1809),
    [anon_sym_done] = ACTIONS(1809),
    [anon_sym_to] = ACTIONS(1809),
    [anon_sym_downto] = ACTIONS(1809),
    [anon_sym_new] = ACTIONS(1809),
    [anon_sym_LBRACE_LT] = ACTIONS(1811),
    [anon_sym_GT_RBRACE] = ACTIONS(1811),
    [anon_sym_begin] = ACTIONS(1809),
    [sym_ocamlyacc_value] = ACTIONS(1811),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1809),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1811),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1809),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1809),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1811),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1811),
    [aux_sym_number_token1] = ACTIONS(1811),
    [aux_sym_signed_number_token1] = ACTIONS(1809),
    [anon_sym_SQUOTE] = ACTIONS(1811),
    [anon_sym_DQUOTE] = ACTIONS(1811),
    [sym_prefix_operator] = ACTIONS(1811),
    [aux_sym_sign_operator_token1] = ACTIONS(1809),
    [sym_hash_operator] = ACTIONS(1811),
    [anon_sym_lsl] = ACTIONS(1809),
    [anon_sym_lsr] = ACTIONS(1809),
    [anon_sym_asr] = ACTIONS(1809),
    [aux_sym_pow_operator_token1] = ACTIONS(1811),
    [anon_sym_mod] = ACTIONS(1809),
    [anon_sym_land] = ACTIONS(1809),
    [anon_sym_lor] = ACTIONS(1809),
    [anon_sym_lxor] = ACTIONS(1809),
    [aux_sym_mult_operator_token1] = ACTIONS(1809),
    [aux_sym_add_operator_token1] = ACTIONS(1809),
    [sym_concat_operator] = ACTIONS(1811),
    [sym_rel_operator] = ACTIONS(1809),
    [anon_sym_AMP_AMP] = ACTIONS(1809),
    [anon_sym_or] = ACTIONS(1809),
    [anon_sym_PIPE_PIPE] = ACTIONS(1809),
    [sym_assign_operator] = ACTIONS(1811),
    [sym__uppercase_identifier] = ACTIONS(1811),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(821)] = {
    [sym_attribute] = STATE(821),
    [sym__lowercase_identifier] = ACTIONS(1813),
    [anon_sym_and] = ACTIONS(1813),
    [anon_sym_TILDE] = ACTIONS(1813),
    [anon_sym_QMARK] = ACTIONS(1813),
    [anon_sym_LPAREN] = ACTIONS(1815),
    [anon_sym_RPAREN] = ACTIONS(1815),
    [anon_sym_COMMA] = ACTIONS(1815),
    [anon_sym_PIPE] = ACTIONS(1813),
    [anon_sym_COLON2] = ACTIONS(1813),
    [anon_sym_DOT] = ACTIONS(1815),
    [anon_sym_DASH_GT] = ACTIONS(1813),
    [anon_sym_LBRACK] = ACTIONS(1813),
    [anon_sym_RBRACK] = ACTIONS(1815),
    [anon_sym_true] = ACTIONS(1813),
    [anon_sym_false] = ACTIONS(1813),
    [anon_sym_LBRACE] = ACTIONS(1813),
    [anon_sym_SEMI] = ACTIONS(1815),
    [anon_sym_RBRACE] = ACTIONS(1815),
    [anon_sym_constraint] = ACTIONS(1813),
    [anon_sym_val] = ACTIONS(1813),
    [anon_sym_end] = ACTIONS(1813),
    [anon_sym_with] = ACTIONS(1813),
    [anon_sym_COLON_GT] = ACTIONS(1815),
    [anon_sym_object] = ACTIONS(1813),
    [anon_sym_inherit] = ACTIONS(1813),
    [anon_sym_method] = ACTIONS(1813),
    [anon_sym_in] = ACTIONS(1813),
    [anon_sym_as] = ACTIONS(1813),
    [anon_sym_initializer] = ACTIONS(1813),
    [anon_sym_AMP] = ACTIONS(1813),
    [anon_sym_POUND] = ACTIONS(1813),
    [anon_sym_COLON_COLON] = ACTIONS(1815),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1815),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1815),
    [anon_sym_then] = ACTIONS(1813),
    [anon_sym_else] = ACTIONS(1813),
    [anon_sym_do] = ACTIONS(1813),
    [anon_sym_done] = ACTIONS(1813),
    [anon_sym_to] = ACTIONS(1813),
    [anon_sym_downto] = ACTIONS(1813),
    [anon_sym_new] = ACTIONS(1813),
    [anon_sym_LBRACE_LT] = ACTIONS(1815),
    [anon_sym_GT_RBRACE] = ACTIONS(1815),
    [anon_sym_begin] = ACTIONS(1813),
    [sym_ocamlyacc_value] = ACTIONS(1815),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1813),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1815),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1813),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1813),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1815),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1815),
    [aux_sym_number_token1] = ACTIONS(1815),
    [aux_sym_signed_number_token1] = ACTIONS(1813),
    [anon_sym_SQUOTE] = ACTIONS(1815),
    [anon_sym_DQUOTE] = ACTIONS(1815),
    [sym_prefix_operator] = ACTIONS(1815),
    [aux_sym_sign_operator_token1] = ACTIONS(1813),
    [sym_hash_operator] = ACTIONS(1815),
    [anon_sym_lsl] = ACTIONS(1813),
    [anon_sym_lsr] = ACTIONS(1813),
    [anon_sym_asr] = ACTIONS(1813),
    [aux_sym_pow_operator_token1] = ACTIONS(1815),
    [anon_sym_mod] = ACTIONS(1813),
    [anon_sym_land] = ACTIONS(1813),
    [anon_sym_lor] = ACTIONS(1813),
    [anon_sym_lxor] = ACTIONS(1813),
    [aux_sym_mult_operator_token1] = ACTIONS(1813),
    [aux_sym_add_operator_token1] = ACTIONS(1813),
    [sym_concat_operator] = ACTIONS(1815),
    [sym_rel_operator] = ACTIONS(1813),
    [anon_sym_AMP_AMP] = ACTIONS(1813),
    [anon_sym_or] = ACTIONS(1813),
    [anon_sym_PIPE_PIPE] = ACTIONS(1813),
    [sym_assign_operator] = ACTIONS(1815),
    [sym__uppercase_identifier] = ACTIONS(1815),
    [aux_sym_tag_token1] = ACTIONS(1815),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(822)] = {
    [sym_attribute] = STATE(822),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_and] = ACTIONS(1817),
    [anon_sym_TILDE] = ACTIONS(1817),
    [anon_sym_QMARK] = ACTIONS(1817),
    [anon_sym_LPAREN] = ACTIONS(1819),
    [anon_sym_RPAREN] = ACTIONS(1819),
    [anon_sym_COMMA] = ACTIONS(1819),
    [anon_sym_PIPE] = ACTIONS(1817),
    [anon_sym_COLON2] = ACTIONS(1817),
    [anon_sym_DOT] = ACTIONS(1819),
    [anon_sym_DASH_GT] = ACTIONS(1817),
    [anon_sym_LBRACK] = ACTIONS(1817),
    [anon_sym_RBRACK] = ACTIONS(1819),
    [anon_sym_true] = ACTIONS(1817),
    [anon_sym_false] = ACTIONS(1817),
    [anon_sym_LBRACE] = ACTIONS(1817),
    [anon_sym_SEMI] = ACTIONS(1819),
    [anon_sym_RBRACE] = ACTIONS(1819),
    [anon_sym_constraint] = ACTIONS(1817),
    [anon_sym_val] = ACTIONS(1817),
    [anon_sym_end] = ACTIONS(1817),
    [anon_sym_with] = ACTIONS(1817),
    [anon_sym_COLON_GT] = ACTIONS(1819),
    [anon_sym_object] = ACTIONS(1817),
    [anon_sym_inherit] = ACTIONS(1817),
    [anon_sym_method] = ACTIONS(1817),
    [anon_sym_in] = ACTIONS(1817),
    [anon_sym_as] = ACTIONS(1817),
    [anon_sym_initializer] = ACTIONS(1817),
    [anon_sym_AMP] = ACTIONS(1817),
    [anon_sym_POUND] = ACTIONS(1817),
    [anon_sym_COLON_COLON] = ACTIONS(1819),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1819),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1819),
    [anon_sym_then] = ACTIONS(1817),
    [anon_sym_else] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1817),
    [anon_sym_done] = ACTIONS(1817),
    [anon_sym_to] = ACTIONS(1817),
    [anon_sym_downto] = ACTIONS(1817),
    [anon_sym_new] = ACTIONS(1817),
    [anon_sym_LBRACE_LT] = ACTIONS(1819),
    [anon_sym_GT_RBRACE] = ACTIONS(1819),
    [anon_sym_begin] = ACTIONS(1817),
    [sym_ocamlyacc_value] = ACTIONS(1819),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1817),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1819),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1817),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1817),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1819),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1819),
    [aux_sym_number_token1] = ACTIONS(1819),
    [aux_sym_signed_number_token1] = ACTIONS(1817),
    [anon_sym_SQUOTE] = ACTIONS(1819),
    [anon_sym_DQUOTE] = ACTIONS(1819),
    [sym_prefix_operator] = ACTIONS(1819),
    [aux_sym_sign_operator_token1] = ACTIONS(1817),
    [sym_hash_operator] = ACTIONS(1819),
    [anon_sym_lsl] = ACTIONS(1817),
    [anon_sym_lsr] = ACTIONS(1817),
    [anon_sym_asr] = ACTIONS(1817),
    [aux_sym_pow_operator_token1] = ACTIONS(1819),
    [anon_sym_mod] = ACTIONS(1817),
    [anon_sym_land] = ACTIONS(1817),
    [anon_sym_lor] = ACTIONS(1817),
    [anon_sym_lxor] = ACTIONS(1817),
    [aux_sym_mult_operator_token1] = ACTIONS(1817),
    [aux_sym_add_operator_token1] = ACTIONS(1817),
    [sym_concat_operator] = ACTIONS(1819),
    [sym_rel_operator] = ACTIONS(1817),
    [anon_sym_AMP_AMP] = ACTIONS(1817),
    [anon_sym_or] = ACTIONS(1817),
    [anon_sym_PIPE_PIPE] = ACTIONS(1817),
    [sym_assign_operator] = ACTIONS(1819),
    [sym__uppercase_identifier] = ACTIONS(1819),
    [aux_sym_tag_token1] = ACTIONS(1819),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(823)] = {
    [sym_attribute] = STATE(823),
    [sym__lowercase_identifier] = ACTIONS(1821),
    [anon_sym_and] = ACTIONS(1821),
    [anon_sym_TILDE] = ACTIONS(1821),
    [anon_sym_QMARK] = ACTIONS(1821),
    [anon_sym_LPAREN] = ACTIONS(1823),
    [anon_sym_RPAREN] = ACTIONS(1823),
    [anon_sym_COMMA] = ACTIONS(1823),
    [anon_sym_PIPE] = ACTIONS(1821),
    [anon_sym_COLON2] = ACTIONS(1821),
    [anon_sym_DOT] = ACTIONS(1823),
    [anon_sym_DASH_GT] = ACTIONS(1821),
    [anon_sym_LBRACK] = ACTIONS(1821),
    [anon_sym_RBRACK] = ACTIONS(1823),
    [anon_sym_true] = ACTIONS(1821),
    [anon_sym_false] = ACTIONS(1821),
    [anon_sym_LBRACE] = ACTIONS(1821),
    [anon_sym_SEMI] = ACTIONS(1823),
    [anon_sym_RBRACE] = ACTIONS(1823),
    [anon_sym_constraint] = ACTIONS(1821),
    [anon_sym_val] = ACTIONS(1821),
    [anon_sym_end] = ACTIONS(1821),
    [anon_sym_with] = ACTIONS(1821),
    [anon_sym_COLON_GT] = ACTIONS(1823),
    [anon_sym_object] = ACTIONS(1821),
    [anon_sym_inherit] = ACTIONS(1821),
    [anon_sym_method] = ACTIONS(1821),
    [anon_sym_in] = ACTIONS(1821),
    [anon_sym_as] = ACTIONS(1821),
    [anon_sym_initializer] = ACTIONS(1821),
    [anon_sym_AMP] = ACTIONS(1821),
    [anon_sym_POUND] = ACTIONS(1821),
    [anon_sym_COLON_COLON] = ACTIONS(1823),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1823),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1823),
    [anon_sym_then] = ACTIONS(1821),
    [anon_sym_else] = ACTIONS(1821),
    [anon_sym_do] = ACTIONS(1821),
    [anon_sym_done] = ACTIONS(1821),
    [anon_sym_to] = ACTIONS(1821),
    [anon_sym_downto] = ACTIONS(1821),
    [anon_sym_new] = ACTIONS(1821),
    [anon_sym_LBRACE_LT] = ACTIONS(1823),
    [anon_sym_GT_RBRACE] = ACTIONS(1823),
    [anon_sym_begin] = ACTIONS(1821),
    [sym_ocamlyacc_value] = ACTIONS(1823),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1821),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1823),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1821),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1821),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1823),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1823),
    [aux_sym_number_token1] = ACTIONS(1823),
    [aux_sym_signed_number_token1] = ACTIONS(1821),
    [anon_sym_SQUOTE] = ACTIONS(1823),
    [anon_sym_DQUOTE] = ACTIONS(1823),
    [sym_prefix_operator] = ACTIONS(1823),
    [aux_sym_sign_operator_token1] = ACTIONS(1821),
    [sym_hash_operator] = ACTIONS(1823),
    [anon_sym_lsl] = ACTIONS(1821),
    [anon_sym_lsr] = ACTIONS(1821),
    [anon_sym_asr] = ACTIONS(1821),
    [aux_sym_pow_operator_token1] = ACTIONS(1823),
    [anon_sym_mod] = ACTIONS(1821),
    [anon_sym_land] = ACTIONS(1821),
    [anon_sym_lor] = ACTIONS(1821),
    [anon_sym_lxor] = ACTIONS(1821),
    [aux_sym_mult_operator_token1] = ACTIONS(1821),
    [aux_sym_add_operator_token1] = ACTIONS(1821),
    [sym_concat_operator] = ACTIONS(1823),
    [sym_rel_operator] = ACTIONS(1821),
    [anon_sym_AMP_AMP] = ACTIONS(1821),
    [anon_sym_or] = ACTIONS(1821),
    [anon_sym_PIPE_PIPE] = ACTIONS(1821),
    [sym_assign_operator] = ACTIONS(1823),
    [sym__uppercase_identifier] = ACTIONS(1823),
    [aux_sym_tag_token1] = ACTIONS(1823),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(824)] = {
    [sym_attribute] = STATE(824),
    [sym__lowercase_identifier] = ACTIONS(1825),
    [anon_sym_and] = ACTIONS(1825),
    [anon_sym_TILDE] = ACTIONS(1825),
    [anon_sym_QMARK] = ACTIONS(1825),
    [anon_sym_LPAREN] = ACTIONS(1827),
    [anon_sym_RPAREN] = ACTIONS(1827),
    [anon_sym_COMMA] = ACTIONS(1827),
    [anon_sym_PIPE] = ACTIONS(1825),
    [anon_sym_COLON2] = ACTIONS(1825),
    [anon_sym_DOT] = ACTIONS(1827),
    [anon_sym_DASH_GT] = ACTIONS(1825),
    [anon_sym_LBRACK] = ACTIONS(1825),
    [anon_sym_RBRACK] = ACTIONS(1827),
    [anon_sym_true] = ACTIONS(1825),
    [anon_sym_false] = ACTIONS(1825),
    [anon_sym_LBRACE] = ACTIONS(1825),
    [anon_sym_SEMI] = ACTIONS(1827),
    [anon_sym_RBRACE] = ACTIONS(1827),
    [anon_sym_constraint] = ACTIONS(1825),
    [anon_sym_val] = ACTIONS(1825),
    [anon_sym_end] = ACTIONS(1825),
    [anon_sym_with] = ACTIONS(1825),
    [anon_sym_COLON_GT] = ACTIONS(1827),
    [anon_sym_object] = ACTIONS(1825),
    [anon_sym_inherit] = ACTIONS(1825),
    [anon_sym_method] = ACTIONS(1825),
    [anon_sym_in] = ACTIONS(1825),
    [anon_sym_as] = ACTIONS(1825),
    [anon_sym_initializer] = ACTIONS(1825),
    [anon_sym_AMP] = ACTIONS(1825),
    [anon_sym_POUND] = ACTIONS(1825),
    [anon_sym_COLON_COLON] = ACTIONS(1827),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1827),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1827),
    [anon_sym_then] = ACTIONS(1825),
    [anon_sym_else] = ACTIONS(1825),
    [anon_sym_do] = ACTIONS(1825),
    [anon_sym_done] = ACTIONS(1825),
    [anon_sym_to] = ACTIONS(1825),
    [anon_sym_downto] = ACTIONS(1825),
    [anon_sym_new] = ACTIONS(1825),
    [anon_sym_LBRACE_LT] = ACTIONS(1827),
    [anon_sym_GT_RBRACE] = ACTIONS(1827),
    [anon_sym_begin] = ACTIONS(1825),
    [sym_ocamlyacc_value] = ACTIONS(1827),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1825),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1827),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1825),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1825),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1827),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1827),
    [aux_sym_number_token1] = ACTIONS(1827),
    [aux_sym_signed_number_token1] = ACTIONS(1825),
    [anon_sym_SQUOTE] = ACTIONS(1827),
    [anon_sym_DQUOTE] = ACTIONS(1827),
    [sym_prefix_operator] = ACTIONS(1827),
    [aux_sym_sign_operator_token1] = ACTIONS(1825),
    [sym_hash_operator] = ACTIONS(1827),
    [anon_sym_lsl] = ACTIONS(1825),
    [anon_sym_lsr] = ACTIONS(1825),
    [anon_sym_asr] = ACTIONS(1825),
    [aux_sym_pow_operator_token1] = ACTIONS(1827),
    [anon_sym_mod] = ACTIONS(1825),
    [anon_sym_land] = ACTIONS(1825),
    [anon_sym_lor] = ACTIONS(1825),
    [anon_sym_lxor] = ACTIONS(1825),
    [aux_sym_mult_operator_token1] = ACTIONS(1825),
    [aux_sym_add_operator_token1] = ACTIONS(1825),
    [sym_concat_operator] = ACTIONS(1827),
    [sym_rel_operator] = ACTIONS(1825),
    [anon_sym_AMP_AMP] = ACTIONS(1825),
    [anon_sym_or] = ACTIONS(1825),
    [anon_sym_PIPE_PIPE] = ACTIONS(1825),
    [sym_assign_operator] = ACTIONS(1827),
    [sym__uppercase_identifier] = ACTIONS(1827),
    [aux_sym_tag_token1] = ACTIONS(1827),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(825)] = {
    [sym_attribute] = STATE(825),
    [sym__lowercase_identifier] = ACTIONS(1829),
    [anon_sym_and] = ACTIONS(1829),
    [anon_sym_TILDE] = ACTIONS(1829),
    [anon_sym_QMARK] = ACTIONS(1829),
    [anon_sym_LPAREN] = ACTIONS(1831),
    [anon_sym_RPAREN] = ACTIONS(1831),
    [anon_sym_COMMA] = ACTIONS(1831),
    [anon_sym_PIPE] = ACTIONS(1829),
    [anon_sym_COLON2] = ACTIONS(1829),
    [anon_sym_DOT] = ACTIONS(1831),
    [anon_sym_DASH_GT] = ACTIONS(1829),
    [anon_sym_LBRACK] = ACTIONS(1829),
    [anon_sym_RBRACK] = ACTIONS(1831),
    [anon_sym_true] = ACTIONS(1829),
    [anon_sym_false] = ACTIONS(1829),
    [anon_sym_LBRACE] = ACTIONS(1829),
    [anon_sym_SEMI] = ACTIONS(1831),
    [anon_sym_RBRACE] = ACTIONS(1831),
    [anon_sym_constraint] = ACTIONS(1829),
    [anon_sym_val] = ACTIONS(1829),
    [anon_sym_end] = ACTIONS(1829),
    [anon_sym_with] = ACTIONS(1829),
    [anon_sym_COLON_GT] = ACTIONS(1831),
    [anon_sym_object] = ACTIONS(1829),
    [anon_sym_inherit] = ACTIONS(1829),
    [anon_sym_method] = ACTIONS(1829),
    [anon_sym_in] = ACTIONS(1829),
    [anon_sym_as] = ACTIONS(1829),
    [anon_sym_initializer] = ACTIONS(1829),
    [anon_sym_AMP] = ACTIONS(1829),
    [anon_sym_POUND] = ACTIONS(1829),
    [anon_sym_COLON_COLON] = ACTIONS(1831),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1831),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1831),
    [anon_sym_then] = ACTIONS(1829),
    [anon_sym_else] = ACTIONS(1829),
    [anon_sym_do] = ACTIONS(1829),
    [anon_sym_done] = ACTIONS(1829),
    [anon_sym_to] = ACTIONS(1829),
    [anon_sym_downto] = ACTIONS(1829),
    [anon_sym_new] = ACTIONS(1829),
    [anon_sym_LBRACE_LT] = ACTIONS(1831),
    [anon_sym_GT_RBRACE] = ACTIONS(1831),
    [anon_sym_begin] = ACTIONS(1829),
    [sym_ocamlyacc_value] = ACTIONS(1831),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1829),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1831),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1829),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1829),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1831),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1831),
    [aux_sym_number_token1] = ACTIONS(1831),
    [aux_sym_signed_number_token1] = ACTIONS(1829),
    [anon_sym_SQUOTE] = ACTIONS(1831),
    [anon_sym_DQUOTE] = ACTIONS(1831),
    [sym_prefix_operator] = ACTIONS(1831),
    [aux_sym_sign_operator_token1] = ACTIONS(1829),
    [sym_hash_operator] = ACTIONS(1831),
    [anon_sym_lsl] = ACTIONS(1829),
    [anon_sym_lsr] = ACTIONS(1829),
    [anon_sym_asr] = ACTIONS(1829),
    [aux_sym_pow_operator_token1] = ACTIONS(1831),
    [anon_sym_mod] = ACTIONS(1829),
    [anon_sym_land] = ACTIONS(1829),
    [anon_sym_lor] = ACTIONS(1829),
    [anon_sym_lxor] = ACTIONS(1829),
    [aux_sym_mult_operator_token1] = ACTIONS(1829),
    [aux_sym_add_operator_token1] = ACTIONS(1829),
    [sym_concat_operator] = ACTIONS(1831),
    [sym_rel_operator] = ACTIONS(1829),
    [anon_sym_AMP_AMP] = ACTIONS(1829),
    [anon_sym_or] = ACTIONS(1829),
    [anon_sym_PIPE_PIPE] = ACTIONS(1829),
    [sym_assign_operator] = ACTIONS(1831),
    [sym__uppercase_identifier] = ACTIONS(1831),
    [aux_sym_tag_token1] = ACTIONS(1831),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(826)] = {
    [sym_attribute] = STATE(826),
    [sym__lowercase_identifier] = ACTIONS(1833),
    [anon_sym_and] = ACTIONS(1833),
    [anon_sym_TILDE] = ACTIONS(1833),
    [anon_sym_QMARK] = ACTIONS(1833),
    [anon_sym_LPAREN] = ACTIONS(1835),
    [anon_sym_RPAREN] = ACTIONS(1835),
    [anon_sym_COMMA] = ACTIONS(1835),
    [anon_sym_PIPE] = ACTIONS(1833),
    [anon_sym_COLON2] = ACTIONS(1833),
    [anon_sym_DOT] = ACTIONS(1835),
    [anon_sym_DASH_GT] = ACTIONS(1833),
    [anon_sym_LBRACK] = ACTIONS(1833),
    [anon_sym_RBRACK] = ACTIONS(1835),
    [anon_sym_true] = ACTIONS(1833),
    [anon_sym_false] = ACTIONS(1833),
    [anon_sym_LBRACE] = ACTIONS(1833),
    [anon_sym_SEMI] = ACTIONS(1835),
    [anon_sym_RBRACE] = ACTIONS(1835),
    [anon_sym_constraint] = ACTIONS(1833),
    [anon_sym_val] = ACTIONS(1833),
    [anon_sym_end] = ACTIONS(1833),
    [anon_sym_with] = ACTIONS(1833),
    [anon_sym_COLON_GT] = ACTIONS(1835),
    [anon_sym_object] = ACTIONS(1833),
    [anon_sym_inherit] = ACTIONS(1833),
    [anon_sym_method] = ACTIONS(1833),
    [anon_sym_in] = ACTIONS(1833),
    [anon_sym_as] = ACTIONS(1833),
    [anon_sym_initializer] = ACTIONS(1833),
    [anon_sym_AMP] = ACTIONS(1833),
    [anon_sym_POUND] = ACTIONS(1833),
    [anon_sym_COLON_COLON] = ACTIONS(1835),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1835),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1835),
    [anon_sym_then] = ACTIONS(1833),
    [anon_sym_else] = ACTIONS(1833),
    [anon_sym_do] = ACTIONS(1833),
    [anon_sym_done] = ACTIONS(1833),
    [anon_sym_to] = ACTIONS(1833),
    [anon_sym_downto] = ACTIONS(1833),
    [anon_sym_new] = ACTIONS(1833),
    [anon_sym_LBRACE_LT] = ACTIONS(1835),
    [anon_sym_GT_RBRACE] = ACTIONS(1835),
    [anon_sym_begin] = ACTIONS(1833),
    [sym_ocamlyacc_value] = ACTIONS(1835),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1833),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1835),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1833),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1833),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1835),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1835),
    [aux_sym_number_token1] = ACTIONS(1835),
    [aux_sym_signed_number_token1] = ACTIONS(1833),
    [anon_sym_SQUOTE] = ACTIONS(1835),
    [anon_sym_DQUOTE] = ACTIONS(1835),
    [sym_prefix_operator] = ACTIONS(1835),
    [aux_sym_sign_operator_token1] = ACTIONS(1833),
    [sym_hash_operator] = ACTIONS(1835),
    [anon_sym_lsl] = ACTIONS(1833),
    [anon_sym_lsr] = ACTIONS(1833),
    [anon_sym_asr] = ACTIONS(1833),
    [aux_sym_pow_operator_token1] = ACTIONS(1835),
    [anon_sym_mod] = ACTIONS(1833),
    [anon_sym_land] = ACTIONS(1833),
    [anon_sym_lor] = ACTIONS(1833),
    [anon_sym_lxor] = ACTIONS(1833),
    [aux_sym_mult_operator_token1] = ACTIONS(1833),
    [aux_sym_add_operator_token1] = ACTIONS(1833),
    [sym_concat_operator] = ACTIONS(1835),
    [sym_rel_operator] = ACTIONS(1833),
    [anon_sym_AMP_AMP] = ACTIONS(1833),
    [anon_sym_or] = ACTIONS(1833),
    [anon_sym_PIPE_PIPE] = ACTIONS(1833),
    [sym_assign_operator] = ACTIONS(1835),
    [sym__uppercase_identifier] = ACTIONS(1835),
    [aux_sym_tag_token1] = ACTIONS(1835),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(827)] = {
    [sym_attribute] = STATE(827),
    [sym__lowercase_identifier] = ACTIONS(1837),
    [anon_sym_and] = ACTIONS(1837),
    [anon_sym_TILDE] = ACTIONS(1837),
    [anon_sym_QMARK] = ACTIONS(1837),
    [anon_sym_LPAREN] = ACTIONS(1839),
    [anon_sym_RPAREN] = ACTIONS(1839),
    [anon_sym_COMMA] = ACTIONS(1839),
    [anon_sym_PIPE] = ACTIONS(1837),
    [anon_sym_COLON2] = ACTIONS(1837),
    [anon_sym_DOT] = ACTIONS(1839),
    [anon_sym_DASH_GT] = ACTIONS(1837),
    [anon_sym_LBRACK] = ACTIONS(1837),
    [anon_sym_RBRACK] = ACTIONS(1839),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1837),
    [anon_sym_SEMI] = ACTIONS(1839),
    [anon_sym_RBRACE] = ACTIONS(1839),
    [anon_sym_constraint] = ACTIONS(1837),
    [anon_sym_val] = ACTIONS(1837),
    [anon_sym_end] = ACTIONS(1837),
    [anon_sym_with] = ACTIONS(1837),
    [anon_sym_COLON_GT] = ACTIONS(1839),
    [anon_sym_object] = ACTIONS(1837),
    [anon_sym_inherit] = ACTIONS(1837),
    [anon_sym_method] = ACTIONS(1837),
    [anon_sym_in] = ACTIONS(1837),
    [anon_sym_as] = ACTIONS(1837),
    [anon_sym_initializer] = ACTIONS(1837),
    [anon_sym_AMP] = ACTIONS(1837),
    [anon_sym_POUND] = ACTIONS(1837),
    [anon_sym_COLON_COLON] = ACTIONS(1839),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1839),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1839),
    [anon_sym_then] = ACTIONS(1837),
    [anon_sym_else] = ACTIONS(1837),
    [anon_sym_do] = ACTIONS(1837),
    [anon_sym_done] = ACTIONS(1837),
    [anon_sym_to] = ACTIONS(1837),
    [anon_sym_downto] = ACTIONS(1837),
    [anon_sym_new] = ACTIONS(1837),
    [anon_sym_LBRACE_LT] = ACTIONS(1839),
    [anon_sym_GT_RBRACE] = ACTIONS(1839),
    [anon_sym_begin] = ACTIONS(1837),
    [sym_ocamlyacc_value] = ACTIONS(1839),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1837),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1839),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1837),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1837),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1839),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1839),
    [aux_sym_number_token1] = ACTIONS(1839),
    [aux_sym_signed_number_token1] = ACTIONS(1837),
    [anon_sym_SQUOTE] = ACTIONS(1839),
    [anon_sym_DQUOTE] = ACTIONS(1839),
    [sym_prefix_operator] = ACTIONS(1839),
    [aux_sym_sign_operator_token1] = ACTIONS(1837),
    [sym_hash_operator] = ACTIONS(1839),
    [anon_sym_lsl] = ACTIONS(1837),
    [anon_sym_lsr] = ACTIONS(1837),
    [anon_sym_asr] = ACTIONS(1837),
    [aux_sym_pow_operator_token1] = ACTIONS(1839),
    [anon_sym_mod] = ACTIONS(1837),
    [anon_sym_land] = ACTIONS(1837),
    [anon_sym_lor] = ACTIONS(1837),
    [anon_sym_lxor] = ACTIONS(1837),
    [aux_sym_mult_operator_token1] = ACTIONS(1837),
    [aux_sym_add_operator_token1] = ACTIONS(1837),
    [sym_concat_operator] = ACTIONS(1839),
    [sym_rel_operator] = ACTIONS(1837),
    [anon_sym_AMP_AMP] = ACTIONS(1837),
    [anon_sym_or] = ACTIONS(1837),
    [anon_sym_PIPE_PIPE] = ACTIONS(1837),
    [sym_assign_operator] = ACTIONS(1839),
    [sym__uppercase_identifier] = ACTIONS(1839),
    [aux_sym_tag_token1] = ACTIONS(1839),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(828)] = {
    [sym_attribute] = STATE(828),
    [sym__lowercase_identifier] = ACTIONS(1699),
    [anon_sym_and] = ACTIONS(1699),
    [anon_sym_TILDE] = ACTIONS(1699),
    [anon_sym_QMARK] = ACTIONS(1699),
    [anon_sym_LPAREN] = ACTIONS(1701),
    [anon_sym_RPAREN] = ACTIONS(1701),
    [anon_sym_COMMA] = ACTIONS(1701),
    [anon_sym_PIPE] = ACTIONS(1699),
    [anon_sym_COLON2] = ACTIONS(1699),
    [anon_sym_DOT] = ACTIONS(1841),
    [anon_sym_DASH_GT] = ACTIONS(1699),
    [anon_sym_LBRACK] = ACTIONS(1699),
    [anon_sym_RBRACK] = ACTIONS(1701),
    [anon_sym_true] = ACTIONS(1699),
    [anon_sym_false] = ACTIONS(1699),
    [anon_sym_LBRACE] = ACTIONS(1699),
    [anon_sym_SEMI] = ACTIONS(1701),
    [anon_sym_RBRACE] = ACTIONS(1701),
    [anon_sym_constraint] = ACTIONS(1699),
    [anon_sym_val] = ACTIONS(1699),
    [anon_sym_end] = ACTIONS(1699),
    [anon_sym_with] = ACTIONS(1699),
    [anon_sym_COLON_GT] = ACTIONS(1701),
    [anon_sym_object] = ACTIONS(1699),
    [anon_sym_inherit] = ACTIONS(1699),
    [anon_sym_method] = ACTIONS(1699),
    [anon_sym_in] = ACTIONS(1699),
    [anon_sym_as] = ACTIONS(1699),
    [anon_sym_initializer] = ACTIONS(1699),
    [anon_sym_AMP] = ACTIONS(1699),
    [anon_sym_POUND] = ACTIONS(1699),
    [anon_sym_COLON_COLON] = ACTIONS(1701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1701),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1701),
    [anon_sym_then] = ACTIONS(1699),
    [anon_sym_else] = ACTIONS(1699),
    [anon_sym_do] = ACTIONS(1699),
    [anon_sym_done] = ACTIONS(1699),
    [anon_sym_to] = ACTIONS(1699),
    [anon_sym_downto] = ACTIONS(1699),
    [anon_sym_new] = ACTIONS(1699),
    [anon_sym_LBRACE_LT] = ACTIONS(1701),
    [anon_sym_GT_RBRACE] = ACTIONS(1701),
    [anon_sym_begin] = ACTIONS(1699),
    [sym_ocamlyacc_value] = ACTIONS(1701),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1699),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1701),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1699),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1699),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1701),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1701),
    [aux_sym_number_token1] = ACTIONS(1701),
    [aux_sym_signed_number_token1] = ACTIONS(1699),
    [anon_sym_SQUOTE] = ACTIONS(1701),
    [anon_sym_DQUOTE] = ACTIONS(1701),
    [sym_prefix_operator] = ACTIONS(1701),
    [aux_sym_sign_operator_token1] = ACTIONS(1699),
    [sym_hash_operator] = ACTIONS(1701),
    [anon_sym_lsl] = ACTIONS(1699),
    [anon_sym_lsr] = ACTIONS(1699),
    [anon_sym_asr] = ACTIONS(1699),
    [aux_sym_pow_operator_token1] = ACTIONS(1701),
    [anon_sym_mod] = ACTIONS(1699),
    [anon_sym_land] = ACTIONS(1699),
    [anon_sym_lor] = ACTIONS(1699),
    [anon_sym_lxor] = ACTIONS(1699),
    [aux_sym_mult_operator_token1] = ACTIONS(1699),
    [aux_sym_add_operator_token1] = ACTIONS(1699),
    [sym_concat_operator] = ACTIONS(1701),
    [sym_rel_operator] = ACTIONS(1699),
    [anon_sym_AMP_AMP] = ACTIONS(1699),
    [anon_sym_or] = ACTIONS(1699),
    [anon_sym_PIPE_PIPE] = ACTIONS(1699),
    [sym_assign_operator] = ACTIONS(1701),
    [sym__uppercase_identifier] = ACTIONS(1701),
    [aux_sym_tag_token1] = ACTIONS(1701),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(829)] = {
    [sym_attribute] = STATE(829),
    [sym__lowercase_identifier] = ACTIONS(1843),
    [anon_sym_and] = ACTIONS(1843),
    [anon_sym_TILDE] = ACTIONS(1843),
    [anon_sym_QMARK] = ACTIONS(1843),
    [anon_sym_LPAREN] = ACTIONS(1845),
    [anon_sym_RPAREN] = ACTIONS(1845),
    [anon_sym_COMMA] = ACTIONS(1845),
    [anon_sym_PIPE] = ACTIONS(1843),
    [anon_sym_COLON2] = ACTIONS(1843),
    [anon_sym_DOT] = ACTIONS(1845),
    [anon_sym_DASH_GT] = ACTIONS(1843),
    [anon_sym_LBRACK] = ACTIONS(1843),
    [anon_sym_RBRACK] = ACTIONS(1845),
    [anon_sym_true] = ACTIONS(1843),
    [anon_sym_false] = ACTIONS(1843),
    [anon_sym_LBRACE] = ACTIONS(1843),
    [anon_sym_SEMI] = ACTIONS(1845),
    [anon_sym_RBRACE] = ACTIONS(1845),
    [anon_sym_constraint] = ACTIONS(1843),
    [anon_sym_val] = ACTIONS(1843),
    [anon_sym_end] = ACTIONS(1843),
    [anon_sym_with] = ACTIONS(1843),
    [anon_sym_COLON_GT] = ACTIONS(1845),
    [anon_sym_object] = ACTIONS(1843),
    [anon_sym_inherit] = ACTIONS(1843),
    [anon_sym_method] = ACTIONS(1843),
    [anon_sym_in] = ACTIONS(1843),
    [anon_sym_as] = ACTIONS(1843),
    [anon_sym_initializer] = ACTIONS(1843),
    [anon_sym_AMP] = ACTIONS(1843),
    [anon_sym_POUND] = ACTIONS(1843),
    [anon_sym_COLON_COLON] = ACTIONS(1845),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1845),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1845),
    [anon_sym_then] = ACTIONS(1843),
    [anon_sym_else] = ACTIONS(1843),
    [anon_sym_do] = ACTIONS(1843),
    [anon_sym_done] = ACTIONS(1843),
    [anon_sym_to] = ACTIONS(1843),
    [anon_sym_downto] = ACTIONS(1843),
    [anon_sym_new] = ACTIONS(1843),
    [anon_sym_LBRACE_LT] = ACTIONS(1845),
    [anon_sym_GT_RBRACE] = ACTIONS(1845),
    [anon_sym_begin] = ACTIONS(1843),
    [sym_ocamlyacc_value] = ACTIONS(1845),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1843),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1845),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1843),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1843),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1845),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1845),
    [aux_sym_number_token1] = ACTIONS(1845),
    [aux_sym_signed_number_token1] = ACTIONS(1843),
    [anon_sym_SQUOTE] = ACTIONS(1845),
    [anon_sym_DQUOTE] = ACTIONS(1845),
    [sym_prefix_operator] = ACTIONS(1845),
    [aux_sym_sign_operator_token1] = ACTIONS(1843),
    [sym_hash_operator] = ACTIONS(1845),
    [anon_sym_lsl] = ACTIONS(1843),
    [anon_sym_lsr] = ACTIONS(1843),
    [anon_sym_asr] = ACTIONS(1843),
    [aux_sym_pow_operator_token1] = ACTIONS(1845),
    [anon_sym_mod] = ACTIONS(1843),
    [anon_sym_land] = ACTIONS(1843),
    [anon_sym_lor] = ACTIONS(1843),
    [anon_sym_lxor] = ACTIONS(1843),
    [aux_sym_mult_operator_token1] = ACTIONS(1843),
    [aux_sym_add_operator_token1] = ACTIONS(1843),
    [sym_concat_operator] = ACTIONS(1845),
    [sym_rel_operator] = ACTIONS(1843),
    [anon_sym_AMP_AMP] = ACTIONS(1843),
    [anon_sym_or] = ACTIONS(1843),
    [anon_sym_PIPE_PIPE] = ACTIONS(1843),
    [sym_assign_operator] = ACTIONS(1845),
    [sym__uppercase_identifier] = ACTIONS(1845),
    [aux_sym_tag_token1] = ACTIONS(1845),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(830)] = {
    [sym_attribute] = STATE(830),
    [sym__lowercase_identifier] = ACTIONS(1847),
    [anon_sym_and] = ACTIONS(1847),
    [anon_sym_TILDE] = ACTIONS(1847),
    [anon_sym_QMARK] = ACTIONS(1847),
    [anon_sym_LPAREN] = ACTIONS(1849),
    [anon_sym_RPAREN] = ACTIONS(1849),
    [anon_sym_COMMA] = ACTIONS(1849),
    [anon_sym_PIPE] = ACTIONS(1847),
    [anon_sym_COLON2] = ACTIONS(1847),
    [anon_sym_DOT] = ACTIONS(1849),
    [anon_sym_DASH_GT] = ACTIONS(1847),
    [anon_sym_LBRACK] = ACTIONS(1847),
    [anon_sym_RBRACK] = ACTIONS(1849),
    [anon_sym_true] = ACTIONS(1847),
    [anon_sym_false] = ACTIONS(1847),
    [anon_sym_LBRACE] = ACTIONS(1847),
    [anon_sym_SEMI] = ACTIONS(1849),
    [anon_sym_RBRACE] = ACTIONS(1849),
    [anon_sym_constraint] = ACTIONS(1847),
    [anon_sym_val] = ACTIONS(1847),
    [anon_sym_end] = ACTIONS(1847),
    [anon_sym_with] = ACTIONS(1847),
    [anon_sym_COLON_GT] = ACTIONS(1849),
    [anon_sym_object] = ACTIONS(1847),
    [anon_sym_inherit] = ACTIONS(1847),
    [anon_sym_method] = ACTIONS(1847),
    [anon_sym_in] = ACTIONS(1847),
    [anon_sym_as] = ACTIONS(1847),
    [anon_sym_initializer] = ACTIONS(1847),
    [anon_sym_AMP] = ACTIONS(1847),
    [anon_sym_POUND] = ACTIONS(1847),
    [anon_sym_COLON_COLON] = ACTIONS(1849),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1849),
    [anon_sym_then] = ACTIONS(1847),
    [anon_sym_else] = ACTIONS(1847),
    [anon_sym_do] = ACTIONS(1847),
    [anon_sym_done] = ACTIONS(1847),
    [anon_sym_to] = ACTIONS(1847),
    [anon_sym_downto] = ACTIONS(1847),
    [anon_sym_new] = ACTIONS(1847),
    [anon_sym_LBRACE_LT] = ACTIONS(1849),
    [anon_sym_GT_RBRACE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1847),
    [sym_ocamlyacc_value] = ACTIONS(1849),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1847),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1849),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1847),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1847),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1849),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1849),
    [aux_sym_number_token1] = ACTIONS(1849),
    [aux_sym_signed_number_token1] = ACTIONS(1847),
    [anon_sym_SQUOTE] = ACTIONS(1849),
    [anon_sym_DQUOTE] = ACTIONS(1849),
    [sym_prefix_operator] = ACTIONS(1849),
    [aux_sym_sign_operator_token1] = ACTIONS(1847),
    [sym_hash_operator] = ACTIONS(1849),
    [anon_sym_lsl] = ACTIONS(1847),
    [anon_sym_lsr] = ACTIONS(1847),
    [anon_sym_asr] = ACTIONS(1847),
    [aux_sym_pow_operator_token1] = ACTIONS(1849),
    [anon_sym_mod] = ACTIONS(1847),
    [anon_sym_land] = ACTIONS(1847),
    [anon_sym_lor] = ACTIONS(1847),
    [anon_sym_lxor] = ACTIONS(1847),
    [aux_sym_mult_operator_token1] = ACTIONS(1847),
    [aux_sym_add_operator_token1] = ACTIONS(1847),
    [sym_concat_operator] = ACTIONS(1849),
    [sym_rel_operator] = ACTIONS(1847),
    [anon_sym_AMP_AMP] = ACTIONS(1847),
    [anon_sym_or] = ACTIONS(1847),
    [anon_sym_PIPE_PIPE] = ACTIONS(1847),
    [sym_assign_operator] = ACTIONS(1849),
    [sym__uppercase_identifier] = ACTIONS(1849),
    [aux_sym_tag_token1] = ACTIONS(1849),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(831)] = {
    [sym_attribute] = STATE(831),
    [sym__lowercase_identifier] = ACTIONS(1851),
    [anon_sym_and] = ACTIONS(1851),
    [anon_sym_TILDE] = ACTIONS(1851),
    [anon_sym_QMARK] = ACTIONS(1851),
    [anon_sym_LPAREN] = ACTIONS(1853),
    [anon_sym_RPAREN] = ACTIONS(1853),
    [anon_sym_COMMA] = ACTIONS(1853),
    [anon_sym_PIPE] = ACTIONS(1851),
    [anon_sym_COLON2] = ACTIONS(1851),
    [anon_sym_DOT] = ACTIONS(1853),
    [anon_sym_DASH_GT] = ACTIONS(1851),
    [anon_sym_LBRACK] = ACTIONS(1851),
    [anon_sym_RBRACK] = ACTIONS(1853),
    [anon_sym_true] = ACTIONS(1851),
    [anon_sym_false] = ACTIONS(1851),
    [anon_sym_LBRACE] = ACTIONS(1851),
    [anon_sym_SEMI] = ACTIONS(1853),
    [anon_sym_RBRACE] = ACTIONS(1853),
    [anon_sym_constraint] = ACTIONS(1851),
    [anon_sym_val] = ACTIONS(1851),
    [anon_sym_end] = ACTIONS(1851),
    [anon_sym_with] = ACTIONS(1851),
    [anon_sym_COLON_GT] = ACTIONS(1853),
    [anon_sym_object] = ACTIONS(1851),
    [anon_sym_inherit] = ACTIONS(1851),
    [anon_sym_method] = ACTIONS(1851),
    [anon_sym_in] = ACTIONS(1851),
    [anon_sym_as] = ACTIONS(1851),
    [anon_sym_initializer] = ACTIONS(1851),
    [anon_sym_AMP] = ACTIONS(1851),
    [anon_sym_POUND] = ACTIONS(1851),
    [anon_sym_COLON_COLON] = ACTIONS(1853),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1853),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1853),
    [anon_sym_then] = ACTIONS(1851),
    [anon_sym_else] = ACTIONS(1851),
    [anon_sym_do] = ACTIONS(1851),
    [anon_sym_done] = ACTIONS(1851),
    [anon_sym_to] = ACTIONS(1851),
    [anon_sym_downto] = ACTIONS(1851),
    [anon_sym_new] = ACTIONS(1851),
    [anon_sym_LBRACE_LT] = ACTIONS(1853),
    [anon_sym_GT_RBRACE] = ACTIONS(1853),
    [anon_sym_begin] = ACTIONS(1851),
    [sym_ocamlyacc_value] = ACTIONS(1853),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1851),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1853),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1851),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1851),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1853),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1853),
    [aux_sym_number_token1] = ACTIONS(1853),
    [aux_sym_signed_number_token1] = ACTIONS(1851),
    [anon_sym_SQUOTE] = ACTIONS(1853),
    [anon_sym_DQUOTE] = ACTIONS(1853),
    [sym_prefix_operator] = ACTIONS(1853),
    [aux_sym_sign_operator_token1] = ACTIONS(1851),
    [sym_hash_operator] = ACTIONS(1853),
    [anon_sym_lsl] = ACTIONS(1851),
    [anon_sym_lsr] = ACTIONS(1851),
    [anon_sym_asr] = ACTIONS(1851),
    [aux_sym_pow_operator_token1] = ACTIONS(1853),
    [anon_sym_mod] = ACTIONS(1851),
    [anon_sym_land] = ACTIONS(1851),
    [anon_sym_lor] = ACTIONS(1851),
    [anon_sym_lxor] = ACTIONS(1851),
    [aux_sym_mult_operator_token1] = ACTIONS(1851),
    [aux_sym_add_operator_token1] = ACTIONS(1851),
    [sym_concat_operator] = ACTIONS(1853),
    [sym_rel_operator] = ACTIONS(1851),
    [anon_sym_AMP_AMP] = ACTIONS(1851),
    [anon_sym_or] = ACTIONS(1851),
    [anon_sym_PIPE_PIPE] = ACTIONS(1851),
    [sym_assign_operator] = ACTIONS(1853),
    [sym__uppercase_identifier] = ACTIONS(1853),
    [aux_sym_tag_token1] = ACTIONS(1853),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(832)] = {
    [sym__delimited_expression] = STATE(2321),
    [sym__simple_expression] = STATE(2359),
    [sym_typed_expression] = STATE(2321),
    [sym_list_expression] = STATE(2330),
    [sym_array_expression] = STATE(2330),
    [sym_record_expression] = STATE(2330),
    [sym_labeled_argument] = STATE(2552),
    [sym_prefix_expression] = STATE(2321),
    [sym_hash_expression] = STATE(2321),
    [sym_field_get_expression] = STATE(2321),
    [sym_array_get_expression] = STATE(2321),
    [sym_string_get_expression] = STATE(2321),
    [sym_bigarray_get_expression] = STATE(2321),
    [sym_local_open_expression] = STATE(2321),
    [sym_package_expression] = STATE(2330),
    [sym_new_expression] = STATE(2321),
    [sym_object_copy_expression] = STATE(2330),
    [sym_method_invocation] = STATE(2321),
    [sym_object_expression] = STATE(2321),
    [sym_parenthesized_expression] = STATE(2330),
    [sym_attribute] = STATE(832),
    [sym_extension] = STATE(2321),
    [sym_quoted_extension] = STATE(2321),
    [sym__constant] = STATE(2321),
    [sym_number] = STATE(2361),
    [sym_character] = STATE(2361),
    [sym_string] = STATE(2361),
    [sym_quoted_string] = STATE(2361),
    [sym_boolean] = STATE(2361),
    [sym_unit] = STATE(2330),
    [sym__value_name] = STATE(2345),
    [sym_parenthesized_operator] = STATE(2400),
    [sym_value_path] = STATE(2321),
    [sym_module_path] = STATE(6791),
    [sym_constructor_path] = STATE(2321),
    [sym__constructor_name] = STATE(2369),
    [sym_tag] = STATE(2321),
    [aux_sym_class_application_repeat1] = STATE(833),
    [sym__lowercase_identifier] = ACTIONS(1855),
    [anon_sym_SEMI_SEMI] = ACTIONS(1523),
    [anon_sym_and] = ACTIONS(1525),
    [anon_sym_TILDE] = ACTIONS(1857),
    [anon_sym_QMARK] = ACTIONS(1857),
    [anon_sym_LPAREN] = ACTIONS(1859),
    [anon_sym_external] = ACTIONS(1525),
    [anon_sym_type] = ACTIONS(1525),
    [anon_sym_LBRACK] = ACTIONS(1861),
    [anon_sym_RBRACK] = ACTIONS(1523),
    [anon_sym_true] = ACTIONS(1863),
    [anon_sym_false] = ACTIONS(1863),
    [anon_sym_LBRACE] = ACTIONS(1865),
    [anon_sym_exception] = ACTIONS(1525),
    [anon_sym_module] = ACTIONS(1525),
    [anon_sym_open] = ACTIONS(1525),
    [anon_sym_include] = ACTIONS(1525),
    [anon_sym_class] = ACTIONS(1525),
    [anon_sym_val] = ACTIONS(1525),
    [anon_sym_end] = ACTIONS(1525),
    [anon_sym_object] = ACTIONS(1867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1869),
    [anon_sym_new] = ACTIONS(1871),
    [anon_sym_LBRACE_LT] = ACTIONS(1873),
    [anon_sym_begin] = ACTIONS(1875),
    [sym_ocamlyacc_value] = ACTIONS(1877),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1525),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1523),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1879),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1881),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1523),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1523),
    [aux_sym_number_token1] = ACTIONS(1883),
    [anon_sym_SQUOTE] = ACTIONS(1885),
    [anon_sym_DQUOTE] = ACTIONS(1887),
    [sym_prefix_operator] = ACTIONS(1889),
    [sym__uppercase_identifier] = ACTIONS(1891),
    [aux_sym_tag_token1] = ACTIONS(1893),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(833)] = {
    [sym__delimited_expression] = STATE(2321),
    [sym__simple_expression] = STATE(2359),
    [sym_typed_expression] = STATE(2321),
    [sym_list_expression] = STATE(2330),
    [sym_array_expression] = STATE(2330),
    [sym_record_expression] = STATE(2330),
    [sym_labeled_argument] = STATE(2552),
    [sym_prefix_expression] = STATE(2321),
    [sym_hash_expression] = STATE(2321),
    [sym_field_get_expression] = STATE(2321),
    [sym_array_get_expression] = STATE(2321),
    [sym_string_get_expression] = STATE(2321),
    [sym_bigarray_get_expression] = STATE(2321),
    [sym_local_open_expression] = STATE(2321),
    [sym_package_expression] = STATE(2330),
    [sym_new_expression] = STATE(2321),
    [sym_object_copy_expression] = STATE(2330),
    [sym_method_invocation] = STATE(2321),
    [sym_object_expression] = STATE(2321),
    [sym_parenthesized_expression] = STATE(2330),
    [sym_attribute] = STATE(833),
    [sym_extension] = STATE(2321),
    [sym_quoted_extension] = STATE(2321),
    [sym__constant] = STATE(2321),
    [sym_number] = STATE(2361),
    [sym_character] = STATE(2361),
    [sym_string] = STATE(2361),
    [sym_quoted_string] = STATE(2361),
    [sym_boolean] = STATE(2361),
    [sym_unit] = STATE(2330),
    [sym__value_name] = STATE(2345),
    [sym_parenthesized_operator] = STATE(2400),
    [sym_value_path] = STATE(2321),
    [sym_module_path] = STATE(6791),
    [sym_constructor_path] = STATE(2321),
    [sym__constructor_name] = STATE(2369),
    [sym_tag] = STATE(2321),
    [aux_sym_class_application_repeat1] = STATE(834),
    [sym__lowercase_identifier] = ACTIONS(1855),
    [anon_sym_SEMI_SEMI] = ACTIONS(1527),
    [anon_sym_and] = ACTIONS(1529),
    [anon_sym_TILDE] = ACTIONS(1857),
    [anon_sym_QMARK] = ACTIONS(1857),
    [anon_sym_LPAREN] = ACTIONS(1859),
    [anon_sym_external] = ACTIONS(1529),
    [anon_sym_type] = ACTIONS(1529),
    [anon_sym_LBRACK] = ACTIONS(1861),
    [anon_sym_RBRACK] = ACTIONS(1527),
    [anon_sym_true] = ACTIONS(1863),
    [anon_sym_false] = ACTIONS(1863),
    [anon_sym_LBRACE] = ACTIONS(1865),
    [anon_sym_exception] = ACTIONS(1529),
    [anon_sym_module] = ACTIONS(1529),
    [anon_sym_open] = ACTIONS(1529),
    [anon_sym_include] = ACTIONS(1529),
    [anon_sym_class] = ACTIONS(1529),
    [anon_sym_val] = ACTIONS(1529),
    [anon_sym_end] = ACTIONS(1529),
    [anon_sym_object] = ACTIONS(1867),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1869),
    [anon_sym_new] = ACTIONS(1871),
    [anon_sym_LBRACE_LT] = ACTIONS(1873),
    [anon_sym_begin] = ACTIONS(1875),
    [sym_ocamlyacc_value] = ACTIONS(1877),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1529),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1527),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1879),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1881),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1527),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1527),
    [aux_sym_number_token1] = ACTIONS(1883),
    [anon_sym_SQUOTE] = ACTIONS(1885),
    [anon_sym_DQUOTE] = ACTIONS(1887),
    [sym_prefix_operator] = ACTIONS(1889),
    [sym__uppercase_identifier] = ACTIONS(1891),
    [aux_sym_tag_token1] = ACTIONS(1893),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(834)] = {
    [sym__delimited_expression] = STATE(2321),
    [sym__simple_expression] = STATE(2359),
    [sym_typed_expression] = STATE(2321),
    [sym_list_expression] = STATE(2330),
    [sym_array_expression] = STATE(2330),
    [sym_record_expression] = STATE(2330),
    [sym_labeled_argument] = STATE(2552),
    [sym_prefix_expression] = STATE(2321),
    [sym_hash_expression] = STATE(2321),
    [sym_field_get_expression] = STATE(2321),
    [sym_array_get_expression] = STATE(2321),
    [sym_string_get_expression] = STATE(2321),
    [sym_bigarray_get_expression] = STATE(2321),
    [sym_local_open_expression] = STATE(2321),
    [sym_package_expression] = STATE(2330),
    [sym_new_expression] = STATE(2321),
    [sym_object_copy_expression] = STATE(2330),
    [sym_method_invocation] = STATE(2321),
    [sym_object_expression] = STATE(2321),
    [sym_parenthesized_expression] = STATE(2330),
    [sym_attribute] = STATE(834),
    [sym_extension] = STATE(2321),
    [sym_quoted_extension] = STATE(2321),
    [sym__constant] = STATE(2321),
    [sym_number] = STATE(2361),
    [sym_character] = STATE(2361),
    [sym_string] = STATE(2361),
    [sym_quoted_string] = STATE(2361),
    [sym_boolean] = STATE(2361),
    [sym_unit] = STATE(2330),
    [sym__value_name] = STATE(2345),
    [sym_parenthesized_operator] = STATE(2400),
    [sym_value_path] = STATE(2321),
    [sym_module_path] = STATE(6791),
    [sym_constructor_path] = STATE(2321),
    [sym__constructor_name] = STATE(2369),
    [sym_tag] = STATE(2321),
    [aux_sym_class_application_repeat1] = STATE(834),
    [sym__lowercase_identifier] = ACTIONS(1895),
    [anon_sym_SEMI_SEMI] = ACTIONS(768),
    [anon_sym_and] = ACTIONS(760),
    [anon_sym_TILDE] = ACTIONS(1898),
    [anon_sym_QMARK] = ACTIONS(1898),
    [anon_sym_LPAREN] = ACTIONS(1901),
    [anon_sym_external] = ACTIONS(760),
    [anon_sym_type] = ACTIONS(760),
    [anon_sym_LBRACK] = ACTIONS(1904),
    [anon_sym_RBRACK] = ACTIONS(768),
    [anon_sym_true] = ACTIONS(1907),
    [anon_sym_false] = ACTIONS(1907),
    [anon_sym_LBRACE] = ACTIONS(1910),
    [anon_sym_exception] = ACTIONS(760),
    [anon_sym_module] = ACTIONS(760),
    [anon_sym_open] = ACTIONS(760),
    [anon_sym_include] = ACTIONS(760),
    [anon_sym_class] = ACTIONS(760),
    [anon_sym_val] = ACTIONS(760),
    [anon_sym_end] = ACTIONS(760),
    [anon_sym_object] = ACTIONS(1913),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1916),
    [anon_sym_new] = ACTIONS(1919),
    [anon_sym_LBRACE_LT] = ACTIONS(1922),
    [anon_sym_begin] = ACTIONS(1925),
    [sym_ocamlyacc_value] = ACTIONS(1928),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(760),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(768),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1931),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1934),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(768),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(768),
    [aux_sym_number_token1] = ACTIONS(1937),
    [anon_sym_SQUOTE] = ACTIONS(1940),
    [anon_sym_DQUOTE] = ACTIONS(1943),
    [sym_prefix_operator] = ACTIONS(1946),
    [sym__uppercase_identifier] = ACTIONS(1949),
    [aux_sym_tag_token1] = ACTIONS(1952),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(835)] = {
    [sym_attribute] = STATE(835),
    [sym__lowercase_identifier] = ACTIONS(1745),
    [anon_sym_TILDE] = ACTIONS(1745),
    [anon_sym_QMARK] = ACTIONS(1745),
    [anon_sym_LPAREN] = ACTIONS(1747),
    [anon_sym_RPAREN] = ACTIONS(1747),
    [anon_sym_COMMA] = ACTIONS(1747),
    [anon_sym_PIPE] = ACTIONS(1745),
    [anon_sym_COLON2] = ACTIONS(1745),
    [anon_sym_DOT] = ACTIONS(1747),
    [anon_sym_DASH_GT] = ACTIONS(1745),
    [anon_sym_LBRACK] = ACTIONS(1745),
    [anon_sym_RBRACK] = ACTIONS(1747),
    [anon_sym_true] = ACTIONS(1745),
    [anon_sym_false] = ACTIONS(1745),
    [anon_sym_LBRACE] = ACTIONS(1745),
    [anon_sym_SEMI] = ACTIONS(1747),
    [anon_sym_RBRACE] = ACTIONS(1747),
    [anon_sym_constraint] = ACTIONS(1745),
    [anon_sym_val] = ACTIONS(1745),
    [anon_sym_end] = ACTIONS(1745),
    [anon_sym_with] = ACTIONS(1745),
    [anon_sym_COLON_GT] = ACTIONS(1747),
    [anon_sym_object] = ACTIONS(1745),
    [anon_sym_inherit] = ACTIONS(1745),
    [anon_sym_method] = ACTIONS(1745),
    [anon_sym_initializer] = ACTIONS(1745),
    [anon_sym_AMP] = ACTIONS(1745),
    [anon_sym_POUND] = ACTIONS(1745),
    [anon_sym_COLON_COLON] = ACTIONS(1747),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1747),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1747),
    [anon_sym_LT_DASH] = ACTIONS(1955),
    [anon_sym_then] = ACTIONS(1745),
    [anon_sym_else] = ACTIONS(1745),
    [anon_sym_do] = ACTIONS(1745),
    [anon_sym_done] = ACTIONS(1745),
    [anon_sym_to] = ACTIONS(1745),
    [anon_sym_downto] = ACTIONS(1745),
    [anon_sym_new] = ACTIONS(1745),
    [anon_sym_LBRACE_LT] = ACTIONS(1747),
    [anon_sym_GT_RBRACE] = ACTIONS(1747),
    [anon_sym_begin] = ACTIONS(1745),
    [sym_ocamlyacc_value] = ACTIONS(1747),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1745),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1747),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1745),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1745),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1747),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1747),
    [aux_sym_number_token1] = ACTIONS(1747),
    [aux_sym_signed_number_token1] = ACTIONS(1745),
    [anon_sym_SQUOTE] = ACTIONS(1747),
    [anon_sym_DQUOTE] = ACTIONS(1747),
    [sym_prefix_operator] = ACTIONS(1747),
    [aux_sym_sign_operator_token1] = ACTIONS(1745),
    [sym_hash_operator] = ACTIONS(1747),
    [anon_sym_lsl] = ACTIONS(1745),
    [anon_sym_lsr] = ACTIONS(1745),
    [anon_sym_asr] = ACTIONS(1745),
    [aux_sym_pow_operator_token1] = ACTIONS(1747),
    [anon_sym_mod] = ACTIONS(1745),
    [anon_sym_land] = ACTIONS(1745),
    [anon_sym_lor] = ACTIONS(1745),
    [anon_sym_lxor] = ACTIONS(1745),
    [aux_sym_mult_operator_token1] = ACTIONS(1745),
    [aux_sym_add_operator_token1] = ACTIONS(1745),
    [sym_concat_operator] = ACTIONS(1747),
    [sym_rel_operator] = ACTIONS(1745),
    [anon_sym_AMP_AMP] = ACTIONS(1745),
    [anon_sym_or] = ACTIONS(1745),
    [anon_sym_PIPE_PIPE] = ACTIONS(1745),
    [sym_assign_operator] = ACTIONS(1747),
    [sym__uppercase_identifier] = ACTIONS(1747),
    [aux_sym_tag_token1] = ACTIONS(1747),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(836)] = {
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(772),
    [sym_typed_expression] = STATE(794),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_labeled_argument] = STATE(845),
    [sym_prefix_expression] = STATE(794),
    [sym_hash_expression] = STATE(794),
    [sym_field_get_expression] = STATE(794),
    [sym_array_get_expression] = STATE(794),
    [sym_string_get_expression] = STATE(794),
    [sym_bigarray_get_expression] = STATE(794),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(836),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [aux_sym_class_application_repeat1] = STATE(78),
    [sym__lowercase_identifier] = ACTIONS(743),
    [anon_sym_and] = ACTIONS(1529),
    [anon_sym_TILDE] = ACTIONS(745),
    [anon_sym_QMARK] = ACTIONS(745),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(1527),
    [anon_sym_COLON2] = ACTIONS(1527),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_constraint] = ACTIONS(1529),
    [anon_sym_val] = ACTIONS(1529),
    [anon_sym_end] = ACTIONS(1529),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_inherit] = ACTIONS(1529),
    [anon_sym_method] = ACTIONS(1529),
    [anon_sym_in] = ACTIONS(1529),
    [anon_sym_as] = ACTIONS(1529),
    [anon_sym_initializer] = ACTIONS(1529),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1529),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1527),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(393),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1527),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1527),
    [aux_sym_number_token1] = ACTIONS(395),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(837)] = {
    [sym__delimited_expression] = STATE(794),
    [sym__simple_expression] = STATE(772),
    [sym_typed_expression] = STATE(794),
    [sym_list_expression] = STATE(785),
    [sym_array_expression] = STATE(785),
    [sym_record_expression] = STATE(785),
    [sym_labeled_argument] = STATE(845),
    [sym_prefix_expression] = STATE(794),
    [sym_hash_expression] = STATE(794),
    [sym_field_get_expression] = STATE(794),
    [sym_array_get_expression] = STATE(794),
    [sym_string_get_expression] = STATE(794),
    [sym_bigarray_get_expression] = STATE(794),
    [sym_local_open_expression] = STATE(794),
    [sym_package_expression] = STATE(785),
    [sym_new_expression] = STATE(794),
    [sym_object_copy_expression] = STATE(785),
    [sym_method_invocation] = STATE(794),
    [sym_object_expression] = STATE(794),
    [sym_parenthesized_expression] = STATE(785),
    [sym_attribute] = STATE(837),
    [sym_extension] = STATE(794),
    [sym_quoted_extension] = STATE(794),
    [sym__constant] = STATE(794),
    [sym_number] = STATE(795),
    [sym_character] = STATE(795),
    [sym_string] = STATE(795),
    [sym_quoted_string] = STATE(795),
    [sym_boolean] = STATE(795),
    [sym_unit] = STATE(785),
    [sym__value_name] = STATE(791),
    [sym_parenthesized_operator] = STATE(830),
    [sym_value_path] = STATE(794),
    [sym_module_path] = STATE(7343),
    [sym_constructor_path] = STATE(794),
    [sym__constructor_name] = STATE(796),
    [sym_tag] = STATE(794),
    [aux_sym_class_application_repeat1] = STATE(836),
    [sym__lowercase_identifier] = ACTIONS(743),
    [anon_sym_and] = ACTIONS(1525),
    [anon_sym_TILDE] = ACTIONS(745),
    [anon_sym_QMARK] = ACTIONS(745),
    [anon_sym_LPAREN] = ACTIONS(353),
    [anon_sym_RPAREN] = ACTIONS(1523),
    [anon_sym_COLON2] = ACTIONS(1523),
    [anon_sym_LBRACK] = ACTIONS(355),
    [anon_sym_true] = ACTIONS(357),
    [anon_sym_false] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(359),
    [anon_sym_constraint] = ACTIONS(1525),
    [anon_sym_val] = ACTIONS(1525),
    [anon_sym_end] = ACTIONS(1525),
    [anon_sym_object] = ACTIONS(361),
    [anon_sym_inherit] = ACTIONS(1525),
    [anon_sym_method] = ACTIONS(1525),
    [anon_sym_in] = ACTIONS(1525),
    [anon_sym_as] = ACTIONS(1525),
    [anon_sym_initializer] = ACTIONS(1525),
    [anon_sym_LBRACK_PIPE] = ACTIONS(365),
    [anon_sym_new] = ACTIONS(383),
    [anon_sym_LBRACE_LT] = ACTIONS(385),
    [anon_sym_begin] = ACTIONS(387),
    [sym_ocamlyacc_value] = ACTIONS(389),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1525),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1523),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(391),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(393),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1523),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1523),
    [aux_sym_number_token1] = ACTIONS(395),
    [anon_sym_SQUOTE] = ACTIONS(397),
    [anon_sym_DQUOTE] = ACTIONS(399),
    [sym_prefix_operator] = ACTIONS(401),
    [sym__uppercase_identifier] = ACTIONS(405),
    [aux_sym_tag_token1] = ACTIONS(407),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(838)] = {
    [sym__parenthesized_abstract_type] = STATE(1363),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2961),
    [sym__binding_pattern] = STATE(2973),
    [sym_alias_binding_pattern] = STATE(2961),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_anonymous] = STATE(2962),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6582),
    [sym__tuple_binding_pattern] = STATE(2964),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern] = STATE(2961),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2961),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_exception_binding_pattern] = STATE(2961),
    [sym_effect_binding_pattern] = STATE(2961),
    [sym_attribute] = STATE(838),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_and] = ACTIONS(1297),
    [anon_sym_EQ] = ACTIONS(1295),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_QMARK] = ACTIONS(1295),
    [anon_sym_LPAREN] = ACTIONS(1957),
    [anon_sym_COMMA] = ACTIONS(1295),
    [anon_sym_PIPE] = ACTIONS(1295),
    [anon_sym_COLON2] = ACTIONS(1297),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1959),
    [anon_sym_COLON_GT] = ACTIONS(1295),
    [anon_sym_in] = ACTIONS(1297),
    [anon_sym_as] = ACTIONS(1297),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_COLON_COLON] = ACTIONS(1295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [anon_sym_effect] = ACTIONS(1961),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1295),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1965),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym_let_and_operator] = ACTIONS(1295),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(839)] = {
    [sym_attribute] = STATE(839),
    [sym__lowercase_identifier] = ACTIONS(1967),
    [anon_sym_and] = ACTIONS(1967),
    [anon_sym_TILDE] = ACTIONS(1967),
    [anon_sym_QMARK] = ACTIONS(1967),
    [anon_sym_COLON] = ACTIONS(1969),
    [anon_sym_LPAREN] = ACTIONS(1971),
    [anon_sym_RPAREN] = ACTIONS(1971),
    [anon_sym_COMMA] = ACTIONS(1971),
    [anon_sym_PIPE] = ACTIONS(1967),
    [anon_sym_COLON2] = ACTIONS(1967),
    [anon_sym_DASH_GT] = ACTIONS(1967),
    [anon_sym_LBRACK] = ACTIONS(1967),
    [anon_sym_RBRACK] = ACTIONS(1971),
    [anon_sym_true] = ACTIONS(1967),
    [anon_sym_false] = ACTIONS(1967),
    [anon_sym_LBRACE] = ACTIONS(1967),
    [anon_sym_SEMI] = ACTIONS(1971),
    [anon_sym_RBRACE] = ACTIONS(1971),
    [anon_sym_constraint] = ACTIONS(1967),
    [anon_sym_val] = ACTIONS(1967),
    [anon_sym_end] = ACTIONS(1967),
    [anon_sym_with] = ACTIONS(1967),
    [anon_sym_COLON_GT] = ACTIONS(1971),
    [anon_sym_object] = ACTIONS(1967),
    [anon_sym_inherit] = ACTIONS(1967),
    [anon_sym_method] = ACTIONS(1967),
    [anon_sym_in] = ACTIONS(1967),
    [anon_sym_as] = ACTIONS(1967),
    [anon_sym_initializer] = ACTIONS(1967),
    [anon_sym_AMP] = ACTIONS(1967),
    [anon_sym_COLON_COLON] = ACTIONS(1971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1971),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1971),
    [anon_sym_then] = ACTIONS(1967),
    [anon_sym_else] = ACTIONS(1967),
    [anon_sym_do] = ACTIONS(1967),
    [anon_sym_done] = ACTIONS(1967),
    [anon_sym_to] = ACTIONS(1967),
    [anon_sym_downto] = ACTIONS(1967),
    [anon_sym_new] = ACTIONS(1967),
    [anon_sym_LBRACE_LT] = ACTIONS(1971),
    [anon_sym_GT_RBRACE] = ACTIONS(1971),
    [anon_sym_begin] = ACTIONS(1967),
    [sym_ocamlyacc_value] = ACTIONS(1971),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1967),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1971),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1967),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1967),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1971),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1971),
    [aux_sym_number_token1] = ACTIONS(1971),
    [aux_sym_signed_number_token1] = ACTIONS(1967),
    [anon_sym_SQUOTE] = ACTIONS(1971),
    [anon_sym_DQUOTE] = ACTIONS(1971),
    [sym_prefix_operator] = ACTIONS(1971),
    [aux_sym_sign_operator_token1] = ACTIONS(1967),
    [anon_sym_lsl] = ACTIONS(1967),
    [anon_sym_lsr] = ACTIONS(1967),
    [anon_sym_asr] = ACTIONS(1967),
    [aux_sym_pow_operator_token1] = ACTIONS(1971),
    [anon_sym_mod] = ACTIONS(1967),
    [anon_sym_land] = ACTIONS(1967),
    [anon_sym_lor] = ACTIONS(1967),
    [anon_sym_lxor] = ACTIONS(1967),
    [aux_sym_mult_operator_token1] = ACTIONS(1967),
    [aux_sym_add_operator_token1] = ACTIONS(1967),
    [sym_concat_operator] = ACTIONS(1971),
    [sym_rel_operator] = ACTIONS(1967),
    [anon_sym_AMP_AMP] = ACTIONS(1967),
    [anon_sym_or] = ACTIONS(1967),
    [anon_sym_PIPE_PIPE] = ACTIONS(1967),
    [sym_assign_operator] = ACTIONS(1971),
    [sym__uppercase_identifier] = ACTIONS(1971),
    [aux_sym_tag_token1] = ACTIONS(1971),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(840)] = {
    [sym_attribute] = STATE(840),
    [sym__lowercase_identifier] = ACTIONS(1539),
    [anon_sym_TILDE] = ACTIONS(1539),
    [anon_sym_QMARK] = ACTIONS(1539),
    [anon_sym_LPAREN] = ACTIONS(1541),
    [anon_sym_RPAREN] = ACTIONS(1541),
    [anon_sym_COMMA] = ACTIONS(1541),
    [anon_sym_PIPE] = ACTIONS(1539),
    [anon_sym_COLON2] = ACTIONS(1539),
    [anon_sym_DOT] = ACTIONS(1541),
    [anon_sym_DASH_GT] = ACTIONS(1539),
    [anon_sym_LBRACK] = ACTIONS(1539),
    [anon_sym_RBRACK] = ACTIONS(1541),
    [anon_sym_true] = ACTIONS(1539),
    [anon_sym_false] = ACTIONS(1539),
    [anon_sym_LBRACE] = ACTIONS(1539),
    [anon_sym_SEMI] = ACTIONS(1541),
    [anon_sym_RBRACE] = ACTIONS(1541),
    [anon_sym_constraint] = ACTIONS(1539),
    [anon_sym_val] = ACTIONS(1539),
    [anon_sym_end] = ACTIONS(1539),
    [anon_sym_with] = ACTIONS(1539),
    [anon_sym_COLON_GT] = ACTIONS(1541),
    [anon_sym_object] = ACTIONS(1539),
    [anon_sym_inherit] = ACTIONS(1539),
    [anon_sym_method] = ACTIONS(1539),
    [anon_sym_initializer] = ACTIONS(1539),
    [anon_sym_AMP] = ACTIONS(1539),
    [anon_sym_POUND] = ACTIONS(1539),
    [anon_sym_COLON_COLON] = ACTIONS(1541),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1541),
    [anon_sym_LT_DASH] = ACTIONS(1539),
    [anon_sym_then] = ACTIONS(1539),
    [anon_sym_else] = ACTIONS(1539),
    [anon_sym_do] = ACTIONS(1539),
    [anon_sym_done] = ACTIONS(1539),
    [anon_sym_to] = ACTIONS(1539),
    [anon_sym_downto] = ACTIONS(1539),
    [anon_sym_new] = ACTIONS(1539),
    [anon_sym_LBRACE_LT] = ACTIONS(1541),
    [anon_sym_GT_RBRACE] = ACTIONS(1541),
    [anon_sym_begin] = ACTIONS(1539),
    [sym_ocamlyacc_value] = ACTIONS(1541),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1539),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1541),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1539),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1539),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1541),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1541),
    [aux_sym_number_token1] = ACTIONS(1541),
    [aux_sym_signed_number_token1] = ACTIONS(1539),
    [anon_sym_SQUOTE] = ACTIONS(1541),
    [anon_sym_DQUOTE] = ACTIONS(1541),
    [sym_prefix_operator] = ACTIONS(1541),
    [aux_sym_sign_operator_token1] = ACTIONS(1539),
    [sym_hash_operator] = ACTIONS(1541),
    [anon_sym_lsl] = ACTIONS(1539),
    [anon_sym_lsr] = ACTIONS(1539),
    [anon_sym_asr] = ACTIONS(1539),
    [aux_sym_pow_operator_token1] = ACTIONS(1541),
    [anon_sym_mod] = ACTIONS(1539),
    [anon_sym_land] = ACTIONS(1539),
    [anon_sym_lor] = ACTIONS(1539),
    [anon_sym_lxor] = ACTIONS(1539),
    [aux_sym_mult_operator_token1] = ACTIONS(1539),
    [aux_sym_add_operator_token1] = ACTIONS(1539),
    [sym_concat_operator] = ACTIONS(1541),
    [sym_rel_operator] = ACTIONS(1539),
    [anon_sym_AMP_AMP] = ACTIONS(1539),
    [anon_sym_or] = ACTIONS(1539),
    [anon_sym_PIPE_PIPE] = ACTIONS(1539),
    [sym_assign_operator] = ACTIONS(1541),
    [sym__uppercase_identifier] = ACTIONS(1541),
    [aux_sym_tag_token1] = ACTIONS(1541),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(841)] = {
    [sym_attribute] = STATE(841),
    [sym__lowercase_identifier] = ACTIONS(1551),
    [anon_sym_TILDE] = ACTIONS(1551),
    [anon_sym_QMARK] = ACTIONS(1551),
    [anon_sym_LPAREN] = ACTIONS(1553),
    [anon_sym_RPAREN] = ACTIONS(1553),
    [anon_sym_COMMA] = ACTIONS(1553),
    [anon_sym_PIPE] = ACTIONS(1551),
    [anon_sym_COLON2] = ACTIONS(1551),
    [anon_sym_DOT] = ACTIONS(1553),
    [anon_sym_DASH_GT] = ACTIONS(1551),
    [anon_sym_LBRACK] = ACTIONS(1551),
    [anon_sym_RBRACK] = ACTIONS(1553),
    [anon_sym_true] = ACTIONS(1551),
    [anon_sym_false] = ACTIONS(1551),
    [anon_sym_LBRACE] = ACTIONS(1551),
    [anon_sym_SEMI] = ACTIONS(1553),
    [anon_sym_RBRACE] = ACTIONS(1553),
    [anon_sym_constraint] = ACTIONS(1551),
    [anon_sym_val] = ACTIONS(1551),
    [anon_sym_end] = ACTIONS(1551),
    [anon_sym_with] = ACTIONS(1551),
    [anon_sym_COLON_GT] = ACTIONS(1553),
    [anon_sym_object] = ACTIONS(1551),
    [anon_sym_inherit] = ACTIONS(1551),
    [anon_sym_method] = ACTIONS(1551),
    [anon_sym_initializer] = ACTIONS(1551),
    [anon_sym_AMP] = ACTIONS(1551),
    [anon_sym_POUND] = ACTIONS(1551),
    [anon_sym_COLON_COLON] = ACTIONS(1553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1553),
    [anon_sym_LT_DASH] = ACTIONS(1551),
    [anon_sym_then] = ACTIONS(1551),
    [anon_sym_else] = ACTIONS(1551),
    [anon_sym_do] = ACTIONS(1551),
    [anon_sym_done] = ACTIONS(1551),
    [anon_sym_to] = ACTIONS(1551),
    [anon_sym_downto] = ACTIONS(1551),
    [anon_sym_new] = ACTIONS(1551),
    [anon_sym_LBRACE_LT] = ACTIONS(1553),
    [anon_sym_GT_RBRACE] = ACTIONS(1553),
    [anon_sym_begin] = ACTIONS(1551),
    [sym_ocamlyacc_value] = ACTIONS(1553),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1551),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1553),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1551),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1551),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1553),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1553),
    [aux_sym_number_token1] = ACTIONS(1553),
    [aux_sym_signed_number_token1] = ACTIONS(1551),
    [anon_sym_SQUOTE] = ACTIONS(1553),
    [anon_sym_DQUOTE] = ACTIONS(1553),
    [sym_prefix_operator] = ACTIONS(1553),
    [aux_sym_sign_operator_token1] = ACTIONS(1551),
    [sym_hash_operator] = ACTIONS(1553),
    [anon_sym_lsl] = ACTIONS(1551),
    [anon_sym_lsr] = ACTIONS(1551),
    [anon_sym_asr] = ACTIONS(1551),
    [aux_sym_pow_operator_token1] = ACTIONS(1553),
    [anon_sym_mod] = ACTIONS(1551),
    [anon_sym_land] = ACTIONS(1551),
    [anon_sym_lor] = ACTIONS(1551),
    [anon_sym_lxor] = ACTIONS(1551),
    [aux_sym_mult_operator_token1] = ACTIONS(1551),
    [aux_sym_add_operator_token1] = ACTIONS(1551),
    [sym_concat_operator] = ACTIONS(1553),
    [sym_rel_operator] = ACTIONS(1551),
    [anon_sym_AMP_AMP] = ACTIONS(1551),
    [anon_sym_or] = ACTIONS(1551),
    [anon_sym_PIPE_PIPE] = ACTIONS(1551),
    [sym_assign_operator] = ACTIONS(1553),
    [sym__uppercase_identifier] = ACTIONS(1553),
    [aux_sym_tag_token1] = ACTIONS(1553),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(842)] = {
    [sym_parameter] = STATE(2473),
    [sym__class_typed] = STATE(3544),
    [sym__parenthesized_abstract_type] = STATE(2473),
    [sym__delimited_pattern] = STATE(2474),
    [sym__simple_pattern] = STATE(2446),
    [sym_typed_pattern] = STATE(2474),
    [sym_polymorphic_variant_pattern] = STATE(2474),
    [sym_record_pattern] = STATE(2428),
    [sym_list_pattern] = STATE(2428),
    [sym_array_pattern] = STATE(2428),
    [sym_local_open_pattern] = STATE(2474),
    [sym_package_pattern] = STATE(2474),
    [sym_parenthesized_pattern] = STATE(2428),
    [sym_attribute] = STATE(842),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2474),
    [sym_quoted_extension] = STATE(2474),
    [sym__constant] = STATE(2426),
    [sym__signed_constant] = STATE(2474),
    [sym_number] = STATE(2434),
    [sym_signed_number] = STATE(2426),
    [sym_character] = STATE(2434),
    [sym_string] = STATE(2434),
    [sym_quoted_string] = STATE(2434),
    [sym_boolean] = STATE(2434),
    [sym_unit] = STATE(2428),
    [sym__simple_value_pattern] = STATE(2474),
    [sym_parenthesized_operator] = STATE(2474),
    [sym_module_path] = STATE(6496),
    [sym_constructor_path] = STATE(2474),
    [sym__constructor_name] = STATE(2476),
    [sym_tag] = STATE(2474),
    [aux_sym_expression_item_repeat1] = STATE(3625),
    [aux_sym_let_binding_repeat1] = STATE(860),
    [sym__lowercase_identifier] = ACTIONS(1973),
    [anon_sym_SEMI_SEMI] = ACTIONS(1975),
    [anon_sym_let] = ACTIONS(1977),
    [anon_sym_and] = ACTIONS(1977),
    [anon_sym_EQ] = ACTIONS(1979),
    [anon_sym_TILDE] = ACTIONS(1981),
    [anon_sym_QMARK] = ACTIONS(1981),
    [anon_sym_LPAREN] = ACTIONS(1983),
    [anon_sym_external] = ACTIONS(1977),
    [anon_sym_type] = ACTIONS(1977),
    [anon_sym_COLON2] = ACTIONS(1985),
    [anon_sym_LBRACK] = ACTIONS(1987),
    [anon_sym_RBRACK] = ACTIONS(1975),
    [anon_sym_true] = ACTIONS(1989),
    [anon_sym_false] = ACTIONS(1989),
    [anon_sym_LBRACE] = ACTIONS(1991),
    [anon_sym_exception] = ACTIONS(1977),
    [anon_sym_module] = ACTIONS(1977),
    [anon_sym_open] = ACTIONS(1977),
    [anon_sym_include] = ACTIONS(1977),
    [anon_sym_class] = ACTIONS(1977),
    [anon_sym_val] = ACTIONS(1977),
    [anon_sym_end] = ACTIONS(1977),
    [anon_sym_POUND] = ACTIONS(1993),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1995),
    [anon_sym_begin] = ACTIONS(1997),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1975),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2001),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1975),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1975),
    [aux_sym_number_token1] = ACTIONS(2005),
    [aux_sym_signed_number_token1] = ACTIONS(2007),
    [anon_sym_SQUOTE] = ACTIONS(2009),
    [anon_sym_DQUOTE] = ACTIONS(2011),
    [sym_let_operator] = ACTIONS(1975),
    [sym__uppercase_identifier] = ACTIONS(2013),
    [aux_sym_directive_token1] = ACTIONS(1977),
    [aux_sym_tag_token1] = ACTIONS(2015),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(843)] = {
    [sym_attribute] = STATE(843),
    [sym__lowercase_identifier] = ACTIONS(1555),
    [anon_sym_TILDE] = ACTIONS(1555),
    [anon_sym_QMARK] = ACTIONS(1555),
    [anon_sym_LPAREN] = ACTIONS(1557),
    [anon_sym_RPAREN] = ACTIONS(1557),
    [anon_sym_COMMA] = ACTIONS(1557),
    [anon_sym_PIPE] = ACTIONS(1555),
    [anon_sym_COLON2] = ACTIONS(1555),
    [anon_sym_DOT] = ACTIONS(1557),
    [anon_sym_DASH_GT] = ACTIONS(1555),
    [anon_sym_LBRACK] = ACTIONS(1555),
    [anon_sym_RBRACK] = ACTIONS(1557),
    [anon_sym_true] = ACTIONS(1555),
    [anon_sym_false] = ACTIONS(1555),
    [anon_sym_LBRACE] = ACTIONS(1555),
    [anon_sym_SEMI] = ACTIONS(1557),
    [anon_sym_RBRACE] = ACTIONS(1557),
    [anon_sym_constraint] = ACTIONS(1555),
    [anon_sym_val] = ACTIONS(1555),
    [anon_sym_end] = ACTIONS(1555),
    [anon_sym_with] = ACTIONS(1555),
    [anon_sym_COLON_GT] = ACTIONS(1557),
    [anon_sym_object] = ACTIONS(1555),
    [anon_sym_inherit] = ACTIONS(1555),
    [anon_sym_method] = ACTIONS(1555),
    [anon_sym_initializer] = ACTIONS(1555),
    [anon_sym_AMP] = ACTIONS(1555),
    [anon_sym_POUND] = ACTIONS(1555),
    [anon_sym_COLON_COLON] = ACTIONS(1557),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1557),
    [anon_sym_LT_DASH] = ACTIONS(1555),
    [anon_sym_then] = ACTIONS(1555),
    [anon_sym_else] = ACTIONS(1555),
    [anon_sym_do] = ACTIONS(1555),
    [anon_sym_done] = ACTIONS(1555),
    [anon_sym_to] = ACTIONS(1555),
    [anon_sym_downto] = ACTIONS(1555),
    [anon_sym_new] = ACTIONS(1555),
    [anon_sym_LBRACE_LT] = ACTIONS(1557),
    [anon_sym_GT_RBRACE] = ACTIONS(1557),
    [anon_sym_begin] = ACTIONS(1555),
    [sym_ocamlyacc_value] = ACTIONS(1557),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1555),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1557),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1555),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1555),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1557),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1557),
    [aux_sym_number_token1] = ACTIONS(1557),
    [aux_sym_signed_number_token1] = ACTIONS(1555),
    [anon_sym_SQUOTE] = ACTIONS(1557),
    [anon_sym_DQUOTE] = ACTIONS(1557),
    [sym_prefix_operator] = ACTIONS(1557),
    [aux_sym_sign_operator_token1] = ACTIONS(1555),
    [sym_hash_operator] = ACTIONS(1557),
    [anon_sym_lsl] = ACTIONS(1555),
    [anon_sym_lsr] = ACTIONS(1555),
    [anon_sym_asr] = ACTIONS(1555),
    [aux_sym_pow_operator_token1] = ACTIONS(1557),
    [anon_sym_mod] = ACTIONS(1555),
    [anon_sym_land] = ACTIONS(1555),
    [anon_sym_lor] = ACTIONS(1555),
    [anon_sym_lxor] = ACTIONS(1555),
    [aux_sym_mult_operator_token1] = ACTIONS(1555),
    [aux_sym_add_operator_token1] = ACTIONS(1555),
    [sym_concat_operator] = ACTIONS(1557),
    [sym_rel_operator] = ACTIONS(1555),
    [anon_sym_AMP_AMP] = ACTIONS(1555),
    [anon_sym_or] = ACTIONS(1555),
    [anon_sym_PIPE_PIPE] = ACTIONS(1555),
    [sym_assign_operator] = ACTIONS(1557),
    [sym__uppercase_identifier] = ACTIONS(1557),
    [aux_sym_tag_token1] = ACTIONS(1557),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(844)] = {
    [sym_attribute] = STATE(844),
    [sym__lowercase_identifier] = ACTIONS(1543),
    [anon_sym_TILDE] = ACTIONS(1543),
    [anon_sym_QMARK] = ACTIONS(1543),
    [anon_sym_LPAREN] = ACTIONS(1545),
    [anon_sym_RPAREN] = ACTIONS(1545),
    [anon_sym_COMMA] = ACTIONS(1545),
    [anon_sym_PIPE] = ACTIONS(1543),
    [anon_sym_COLON2] = ACTIONS(1543),
    [anon_sym_DOT] = ACTIONS(1545),
    [anon_sym_DASH_GT] = ACTIONS(1543),
    [anon_sym_LBRACK] = ACTIONS(1543),
    [anon_sym_RBRACK] = ACTIONS(1545),
    [anon_sym_true] = ACTIONS(1543),
    [anon_sym_false] = ACTIONS(1543),
    [anon_sym_LBRACE] = ACTIONS(1543),
    [anon_sym_SEMI] = ACTIONS(1545),
    [anon_sym_RBRACE] = ACTIONS(1545),
    [anon_sym_constraint] = ACTIONS(1543),
    [anon_sym_val] = ACTIONS(1543),
    [anon_sym_end] = ACTIONS(1543),
    [anon_sym_with] = ACTIONS(1543),
    [anon_sym_COLON_GT] = ACTIONS(1545),
    [anon_sym_object] = ACTIONS(1543),
    [anon_sym_inherit] = ACTIONS(1543),
    [anon_sym_method] = ACTIONS(1543),
    [anon_sym_initializer] = ACTIONS(1543),
    [anon_sym_AMP] = ACTIONS(1543),
    [anon_sym_POUND] = ACTIONS(1543),
    [anon_sym_COLON_COLON] = ACTIONS(1545),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1545),
    [anon_sym_LT_DASH] = ACTIONS(1543),
    [anon_sym_then] = ACTIONS(1543),
    [anon_sym_else] = ACTIONS(1543),
    [anon_sym_do] = ACTIONS(1543),
    [anon_sym_done] = ACTIONS(1543),
    [anon_sym_to] = ACTIONS(1543),
    [anon_sym_downto] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1543),
    [anon_sym_LBRACE_LT] = ACTIONS(1545),
    [anon_sym_GT_RBRACE] = ACTIONS(1545),
    [anon_sym_begin] = ACTIONS(1543),
    [sym_ocamlyacc_value] = ACTIONS(1545),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1543),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1545),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1543),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1543),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1545),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1545),
    [aux_sym_number_token1] = ACTIONS(1545),
    [aux_sym_signed_number_token1] = ACTIONS(1543),
    [anon_sym_SQUOTE] = ACTIONS(1545),
    [anon_sym_DQUOTE] = ACTIONS(1545),
    [sym_prefix_operator] = ACTIONS(1545),
    [aux_sym_sign_operator_token1] = ACTIONS(1543),
    [sym_hash_operator] = ACTIONS(1545),
    [anon_sym_lsl] = ACTIONS(1543),
    [anon_sym_lsr] = ACTIONS(1543),
    [anon_sym_asr] = ACTIONS(1543),
    [aux_sym_pow_operator_token1] = ACTIONS(1545),
    [anon_sym_mod] = ACTIONS(1543),
    [anon_sym_land] = ACTIONS(1543),
    [anon_sym_lor] = ACTIONS(1543),
    [anon_sym_lxor] = ACTIONS(1543),
    [aux_sym_mult_operator_token1] = ACTIONS(1543),
    [aux_sym_add_operator_token1] = ACTIONS(1543),
    [sym_concat_operator] = ACTIONS(1545),
    [sym_rel_operator] = ACTIONS(1543),
    [anon_sym_AMP_AMP] = ACTIONS(1543),
    [anon_sym_or] = ACTIONS(1543),
    [anon_sym_PIPE_PIPE] = ACTIONS(1543),
    [sym_assign_operator] = ACTIONS(1545),
    [sym__uppercase_identifier] = ACTIONS(1545),
    [aux_sym_tag_token1] = ACTIONS(1545),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(845)] = {
    [sym_attribute] = STATE(845),
    [sym__lowercase_identifier] = ACTIONS(1615),
    [anon_sym_and] = ACTIONS(1615),
    [anon_sym_TILDE] = ACTIONS(1615),
    [anon_sym_QMARK] = ACTIONS(1615),
    [anon_sym_LPAREN] = ACTIONS(1617),
    [anon_sym_RPAREN] = ACTIONS(1617),
    [anon_sym_COMMA] = ACTIONS(1617),
    [anon_sym_PIPE] = ACTIONS(1615),
    [anon_sym_COLON2] = ACTIONS(1615),
    [anon_sym_DASH_GT] = ACTIONS(1615),
    [anon_sym_LBRACK] = ACTIONS(1615),
    [anon_sym_RBRACK] = ACTIONS(1617),
    [anon_sym_true] = ACTIONS(1615),
    [anon_sym_false] = ACTIONS(1615),
    [anon_sym_LBRACE] = ACTIONS(1615),
    [anon_sym_SEMI] = ACTIONS(1617),
    [anon_sym_RBRACE] = ACTIONS(1617),
    [anon_sym_constraint] = ACTIONS(1615),
    [anon_sym_val] = ACTIONS(1615),
    [anon_sym_end] = ACTIONS(1615),
    [anon_sym_with] = ACTIONS(1615),
    [anon_sym_COLON_GT] = ACTIONS(1617),
    [anon_sym_object] = ACTIONS(1615),
    [anon_sym_inherit] = ACTIONS(1615),
    [anon_sym_method] = ACTIONS(1615),
    [anon_sym_in] = ACTIONS(1615),
    [anon_sym_as] = ACTIONS(1615),
    [anon_sym_initializer] = ACTIONS(1615),
    [anon_sym_AMP] = ACTIONS(1615),
    [anon_sym_COLON_COLON] = ACTIONS(1617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1617),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1617),
    [anon_sym_then] = ACTIONS(1615),
    [anon_sym_else] = ACTIONS(1615),
    [anon_sym_do] = ACTIONS(1615),
    [anon_sym_done] = ACTIONS(1615),
    [anon_sym_to] = ACTIONS(1615),
    [anon_sym_downto] = ACTIONS(1615),
    [anon_sym_new] = ACTIONS(1615),
    [anon_sym_LBRACE_LT] = ACTIONS(1617),
    [anon_sym_GT_RBRACE] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1615),
    [sym_ocamlyacc_value] = ACTIONS(1617),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1615),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1617),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1615),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1615),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1617),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1617),
    [aux_sym_number_token1] = ACTIONS(1617),
    [aux_sym_signed_number_token1] = ACTIONS(1615),
    [anon_sym_SQUOTE] = ACTIONS(1617),
    [anon_sym_DQUOTE] = ACTIONS(1617),
    [sym_prefix_operator] = ACTIONS(1617),
    [aux_sym_sign_operator_token1] = ACTIONS(1615),
    [anon_sym_lsl] = ACTIONS(1615),
    [anon_sym_lsr] = ACTIONS(1615),
    [anon_sym_asr] = ACTIONS(1615),
    [aux_sym_pow_operator_token1] = ACTIONS(1617),
    [anon_sym_mod] = ACTIONS(1615),
    [anon_sym_land] = ACTIONS(1615),
    [anon_sym_lor] = ACTIONS(1615),
    [anon_sym_lxor] = ACTIONS(1615),
    [aux_sym_mult_operator_token1] = ACTIONS(1615),
    [aux_sym_add_operator_token1] = ACTIONS(1615),
    [sym_concat_operator] = ACTIONS(1617),
    [sym_rel_operator] = ACTIONS(1615),
    [anon_sym_AMP_AMP] = ACTIONS(1615),
    [anon_sym_or] = ACTIONS(1615),
    [anon_sym_PIPE_PIPE] = ACTIONS(1615),
    [sym_assign_operator] = ACTIONS(1617),
    [sym__uppercase_identifier] = ACTIONS(1617),
    [aux_sym_tag_token1] = ACTIONS(1617),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(846)] = {
    [sym_attribute] = STATE(846),
    [sym__lowercase_identifier] = ACTIONS(1745),
    [anon_sym_TILDE] = ACTIONS(1745),
    [anon_sym_QMARK] = ACTIONS(1745),
    [anon_sym_LPAREN] = ACTIONS(1747),
    [anon_sym_RPAREN] = ACTIONS(1747),
    [anon_sym_COMMA] = ACTIONS(1747),
    [anon_sym_PIPE] = ACTIONS(1745),
    [anon_sym_COLON2] = ACTIONS(1745),
    [anon_sym_DOT] = ACTIONS(1747),
    [anon_sym_DASH_GT] = ACTIONS(1745),
    [anon_sym_LBRACK] = ACTIONS(1745),
    [anon_sym_RBRACK] = ACTIONS(1747),
    [anon_sym_true] = ACTIONS(1745),
    [anon_sym_false] = ACTIONS(1745),
    [anon_sym_LBRACE] = ACTIONS(1745),
    [anon_sym_SEMI] = ACTIONS(1747),
    [anon_sym_RBRACE] = ACTIONS(1747),
    [anon_sym_constraint] = ACTIONS(1745),
    [anon_sym_val] = ACTIONS(1745),
    [anon_sym_end] = ACTIONS(1745),
    [anon_sym_with] = ACTIONS(1745),
    [anon_sym_COLON_GT] = ACTIONS(1747),
    [anon_sym_object] = ACTIONS(1745),
    [anon_sym_inherit] = ACTIONS(1745),
    [anon_sym_method] = ACTIONS(1745),
    [anon_sym_initializer] = ACTIONS(1745),
    [anon_sym_AMP] = ACTIONS(1745),
    [anon_sym_POUND] = ACTIONS(1745),
    [anon_sym_COLON_COLON] = ACTIONS(1747),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1747),
    [anon_sym_LT_DASH] = ACTIONS(1955),
    [anon_sym_then] = ACTIONS(1745),
    [anon_sym_else] = ACTIONS(1745),
    [anon_sym_do] = ACTIONS(1745),
    [anon_sym_done] = ACTIONS(1745),
    [anon_sym_to] = ACTIONS(1745),
    [anon_sym_downto] = ACTIONS(1745),
    [anon_sym_new] = ACTIONS(1745),
    [anon_sym_LBRACE_LT] = ACTIONS(1747),
    [anon_sym_GT_RBRACE] = ACTIONS(1747),
    [anon_sym_begin] = ACTIONS(1745),
    [sym_ocamlyacc_value] = ACTIONS(1747),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1745),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1747),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1745),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1745),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1747),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1747),
    [aux_sym_number_token1] = ACTIONS(1747),
    [aux_sym_signed_number_token1] = ACTIONS(1745),
    [anon_sym_SQUOTE] = ACTIONS(1747),
    [anon_sym_DQUOTE] = ACTIONS(1747),
    [sym_prefix_operator] = ACTIONS(1747),
    [aux_sym_sign_operator_token1] = ACTIONS(1745),
    [sym_hash_operator] = ACTIONS(1747),
    [anon_sym_lsl] = ACTIONS(1745),
    [anon_sym_lsr] = ACTIONS(1745),
    [anon_sym_asr] = ACTIONS(1745),
    [aux_sym_pow_operator_token1] = ACTIONS(1747),
    [anon_sym_mod] = ACTIONS(1745),
    [anon_sym_land] = ACTIONS(1745),
    [anon_sym_lor] = ACTIONS(1745),
    [anon_sym_lxor] = ACTIONS(1745),
    [aux_sym_mult_operator_token1] = ACTIONS(1745),
    [aux_sym_add_operator_token1] = ACTIONS(1745),
    [sym_concat_operator] = ACTIONS(1747),
    [sym_rel_operator] = ACTIONS(1745),
    [anon_sym_AMP_AMP] = ACTIONS(1745),
    [anon_sym_or] = ACTIONS(1745),
    [anon_sym_PIPE_PIPE] = ACTIONS(1745),
    [sym_assign_operator] = ACTIONS(1747),
    [sym__uppercase_identifier] = ACTIONS(1747),
    [aux_sym_tag_token1] = ACTIONS(1747),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(847)] = {
    [sym_attribute] = STATE(847),
    [sym__lowercase_identifier] = ACTIONS(2017),
    [anon_sym_and] = ACTIONS(2017),
    [anon_sym_TILDE] = ACTIONS(2017),
    [anon_sym_QMARK] = ACTIONS(2017),
    [anon_sym_LPAREN] = ACTIONS(2019),
    [anon_sym_RPAREN] = ACTIONS(2019),
    [anon_sym_COMMA] = ACTIONS(2019),
    [anon_sym_PIPE] = ACTIONS(2017),
    [anon_sym_COLON2] = ACTIONS(2017),
    [anon_sym_DASH_GT] = ACTIONS(2017),
    [anon_sym_LBRACK] = ACTIONS(2017),
    [anon_sym_RBRACK] = ACTIONS(2019),
    [anon_sym_true] = ACTIONS(2017),
    [anon_sym_false] = ACTIONS(2017),
    [anon_sym_LBRACE] = ACTIONS(2017),
    [anon_sym_SEMI] = ACTIONS(2019),
    [anon_sym_RBRACE] = ACTIONS(2019),
    [anon_sym_constraint] = ACTIONS(2017),
    [anon_sym_val] = ACTIONS(2017),
    [anon_sym_end] = ACTIONS(2017),
    [anon_sym_with] = ACTIONS(2017),
    [anon_sym_COLON_GT] = ACTIONS(2019),
    [anon_sym_object] = ACTIONS(2017),
    [anon_sym_inherit] = ACTIONS(2017),
    [anon_sym_method] = ACTIONS(2017),
    [anon_sym_in] = ACTIONS(2017),
    [anon_sym_as] = ACTIONS(2017),
    [anon_sym_initializer] = ACTIONS(2017),
    [anon_sym_AMP] = ACTIONS(2017),
    [anon_sym_COLON_COLON] = ACTIONS(2019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2019),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2019),
    [anon_sym_then] = ACTIONS(2017),
    [anon_sym_else] = ACTIONS(2017),
    [anon_sym_do] = ACTIONS(2017),
    [anon_sym_done] = ACTIONS(2017),
    [anon_sym_to] = ACTIONS(2017),
    [anon_sym_downto] = ACTIONS(2017),
    [anon_sym_new] = ACTIONS(2017),
    [anon_sym_LBRACE_LT] = ACTIONS(2019),
    [anon_sym_GT_RBRACE] = ACTIONS(2019),
    [anon_sym_begin] = ACTIONS(2017),
    [sym_ocamlyacc_value] = ACTIONS(2019),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2017),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2019),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2019),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2019),
    [aux_sym_number_token1] = ACTIONS(2019),
    [aux_sym_signed_number_token1] = ACTIONS(2017),
    [anon_sym_SQUOTE] = ACTIONS(2019),
    [anon_sym_DQUOTE] = ACTIONS(2019),
    [sym_prefix_operator] = ACTIONS(2019),
    [aux_sym_sign_operator_token1] = ACTIONS(2017),
    [anon_sym_lsl] = ACTIONS(2017),
    [anon_sym_lsr] = ACTIONS(2017),
    [anon_sym_asr] = ACTIONS(2017),
    [aux_sym_pow_operator_token1] = ACTIONS(2019),
    [anon_sym_mod] = ACTIONS(2017),
    [anon_sym_land] = ACTIONS(2017),
    [anon_sym_lor] = ACTIONS(2017),
    [anon_sym_lxor] = ACTIONS(2017),
    [aux_sym_mult_operator_token1] = ACTIONS(2017),
    [aux_sym_add_operator_token1] = ACTIONS(2017),
    [sym_concat_operator] = ACTIONS(2019),
    [sym_rel_operator] = ACTIONS(2017),
    [anon_sym_AMP_AMP] = ACTIONS(2017),
    [anon_sym_or] = ACTIONS(2017),
    [anon_sym_PIPE_PIPE] = ACTIONS(2017),
    [sym_assign_operator] = ACTIONS(2019),
    [sym__uppercase_identifier] = ACTIONS(2019),
    [aux_sym_tag_token1] = ACTIONS(2019),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(848)] = {
    [sym_parameter] = STATE(2473),
    [sym__class_typed] = STATE(3534),
    [sym__parenthesized_abstract_type] = STATE(2473),
    [sym__delimited_pattern] = STATE(2474),
    [sym__simple_pattern] = STATE(2446),
    [sym_typed_pattern] = STATE(2474),
    [sym_polymorphic_variant_pattern] = STATE(2474),
    [sym_record_pattern] = STATE(2428),
    [sym_list_pattern] = STATE(2428),
    [sym_array_pattern] = STATE(2428),
    [sym_local_open_pattern] = STATE(2474),
    [sym_package_pattern] = STATE(2474),
    [sym_parenthesized_pattern] = STATE(2428),
    [sym_attribute] = STATE(848),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2474),
    [sym_quoted_extension] = STATE(2474),
    [sym__constant] = STATE(2426),
    [sym__signed_constant] = STATE(2474),
    [sym_number] = STATE(2434),
    [sym_signed_number] = STATE(2426),
    [sym_character] = STATE(2434),
    [sym_string] = STATE(2434),
    [sym_quoted_string] = STATE(2434),
    [sym_boolean] = STATE(2434),
    [sym_unit] = STATE(2428),
    [sym__simple_value_pattern] = STATE(2474),
    [sym_parenthesized_operator] = STATE(2474),
    [sym_module_path] = STATE(6496),
    [sym_constructor_path] = STATE(2474),
    [sym__constructor_name] = STATE(2476),
    [sym_tag] = STATE(2474),
    [aux_sym_expression_item_repeat1] = STATE(3765),
    [aux_sym_let_binding_repeat1] = STATE(857),
    [sym__lowercase_identifier] = ACTIONS(1973),
    [anon_sym_SEMI_SEMI] = ACTIONS(2021),
    [anon_sym_let] = ACTIONS(2023),
    [anon_sym_and] = ACTIONS(2023),
    [anon_sym_EQ] = ACTIONS(2025),
    [anon_sym_TILDE] = ACTIONS(1981),
    [anon_sym_QMARK] = ACTIONS(1981),
    [anon_sym_LPAREN] = ACTIONS(1983),
    [anon_sym_external] = ACTIONS(2023),
    [anon_sym_type] = ACTIONS(2023),
    [anon_sym_COLON2] = ACTIONS(1985),
    [anon_sym_LBRACK] = ACTIONS(1987),
    [anon_sym_RBRACK] = ACTIONS(2021),
    [anon_sym_true] = ACTIONS(1989),
    [anon_sym_false] = ACTIONS(1989),
    [anon_sym_LBRACE] = ACTIONS(1991),
    [anon_sym_exception] = ACTIONS(2023),
    [anon_sym_module] = ACTIONS(2023),
    [anon_sym_open] = ACTIONS(2023),
    [anon_sym_include] = ACTIONS(2023),
    [anon_sym_class] = ACTIONS(2023),
    [anon_sym_val] = ACTIONS(2023),
    [anon_sym_end] = ACTIONS(2023),
    [anon_sym_POUND] = ACTIONS(1993),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1995),
    [anon_sym_begin] = ACTIONS(1997),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2021),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2001),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2021),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2021),
    [aux_sym_number_token1] = ACTIONS(2005),
    [aux_sym_signed_number_token1] = ACTIONS(2007),
    [anon_sym_SQUOTE] = ACTIONS(2009),
    [anon_sym_DQUOTE] = ACTIONS(2011),
    [sym_let_operator] = ACTIONS(2021),
    [sym__uppercase_identifier] = ACTIONS(2013),
    [aux_sym_directive_token1] = ACTIONS(2023),
    [aux_sym_tag_token1] = ACTIONS(2015),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(849)] = {
    [sym_parameter] = STATE(2473),
    [sym__class_typed] = STATE(3568),
    [sym__parenthesized_abstract_type] = STATE(2473),
    [sym__delimited_pattern] = STATE(2474),
    [sym__simple_pattern] = STATE(2446),
    [sym_typed_pattern] = STATE(2474),
    [sym_polymorphic_variant_pattern] = STATE(2474),
    [sym_record_pattern] = STATE(2428),
    [sym_list_pattern] = STATE(2428),
    [sym_array_pattern] = STATE(2428),
    [sym_local_open_pattern] = STATE(2474),
    [sym_package_pattern] = STATE(2474),
    [sym_parenthesized_pattern] = STATE(2428),
    [sym_attribute] = STATE(849),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2474),
    [sym_quoted_extension] = STATE(2474),
    [sym__constant] = STATE(2426),
    [sym__signed_constant] = STATE(2474),
    [sym_number] = STATE(2434),
    [sym_signed_number] = STATE(2426),
    [sym_character] = STATE(2434),
    [sym_string] = STATE(2434),
    [sym_quoted_string] = STATE(2434),
    [sym_boolean] = STATE(2434),
    [sym_unit] = STATE(2428),
    [sym__simple_value_pattern] = STATE(2474),
    [sym_parenthesized_operator] = STATE(2474),
    [sym_module_path] = STATE(6496),
    [sym_constructor_path] = STATE(2474),
    [sym__constructor_name] = STATE(2476),
    [sym_tag] = STATE(2474),
    [aux_sym_expression_item_repeat1] = STATE(3584),
    [aux_sym_let_binding_repeat1] = STATE(982),
    [sym__lowercase_identifier] = ACTIONS(1973),
    [anon_sym_SEMI_SEMI] = ACTIONS(2027),
    [anon_sym_let] = ACTIONS(2029),
    [anon_sym_and] = ACTIONS(2029),
    [anon_sym_EQ] = ACTIONS(2031),
    [anon_sym_TILDE] = ACTIONS(1981),
    [anon_sym_QMARK] = ACTIONS(1981),
    [anon_sym_LPAREN] = ACTIONS(1983),
    [anon_sym_external] = ACTIONS(2029),
    [anon_sym_type] = ACTIONS(2029),
    [anon_sym_COLON2] = ACTIONS(1985),
    [anon_sym_LBRACK] = ACTIONS(1987),
    [anon_sym_RBRACK] = ACTIONS(2027),
    [anon_sym_true] = ACTIONS(1989),
    [anon_sym_false] = ACTIONS(1989),
    [anon_sym_LBRACE] = ACTIONS(1991),
    [anon_sym_exception] = ACTIONS(2029),
    [anon_sym_module] = ACTIONS(2029),
    [anon_sym_open] = ACTIONS(2029),
    [anon_sym_include] = ACTIONS(2029),
    [anon_sym_class] = ACTIONS(2029),
    [anon_sym_val] = ACTIONS(2029),
    [anon_sym_end] = ACTIONS(2029),
    [anon_sym_POUND] = ACTIONS(1993),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1995),
    [anon_sym_begin] = ACTIONS(1997),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2027),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2001),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2027),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2027),
    [aux_sym_number_token1] = ACTIONS(2005),
    [aux_sym_signed_number_token1] = ACTIONS(2007),
    [anon_sym_SQUOTE] = ACTIONS(2009),
    [anon_sym_DQUOTE] = ACTIONS(2011),
    [sym_let_operator] = ACTIONS(2027),
    [sym__uppercase_identifier] = ACTIONS(2013),
    [aux_sym_directive_token1] = ACTIONS(2029),
    [aux_sym_tag_token1] = ACTIONS(2015),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(850)] = {
    [sym_parameter] = STATE(2473),
    [sym__class_typed] = STATE(3572),
    [sym__parenthesized_abstract_type] = STATE(2473),
    [sym__delimited_pattern] = STATE(2474),
    [sym__simple_pattern] = STATE(2446),
    [sym_typed_pattern] = STATE(2474),
    [sym_polymorphic_variant_pattern] = STATE(2474),
    [sym_record_pattern] = STATE(2428),
    [sym_list_pattern] = STATE(2428),
    [sym_array_pattern] = STATE(2428),
    [sym_local_open_pattern] = STATE(2474),
    [sym_package_pattern] = STATE(2474),
    [sym_parenthesized_pattern] = STATE(2428),
    [sym_attribute] = STATE(850),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2474),
    [sym_quoted_extension] = STATE(2474),
    [sym__constant] = STATE(2426),
    [sym__signed_constant] = STATE(2474),
    [sym_number] = STATE(2434),
    [sym_signed_number] = STATE(2426),
    [sym_character] = STATE(2434),
    [sym_string] = STATE(2434),
    [sym_quoted_string] = STATE(2434),
    [sym_boolean] = STATE(2434),
    [sym_unit] = STATE(2428),
    [sym__simple_value_pattern] = STATE(2474),
    [sym_parenthesized_operator] = STATE(2474),
    [sym_module_path] = STATE(6496),
    [sym_constructor_path] = STATE(2474),
    [sym__constructor_name] = STATE(2476),
    [sym_tag] = STATE(2474),
    [aux_sym_expression_item_repeat1] = STATE(3596),
    [aux_sym_let_binding_repeat1] = STATE(982),
    [sym__lowercase_identifier] = ACTIONS(1973),
    [anon_sym_SEMI_SEMI] = ACTIONS(2033),
    [anon_sym_let] = ACTIONS(2035),
    [anon_sym_and] = ACTIONS(2035),
    [anon_sym_EQ] = ACTIONS(2037),
    [anon_sym_TILDE] = ACTIONS(1981),
    [anon_sym_QMARK] = ACTIONS(1981),
    [anon_sym_LPAREN] = ACTIONS(1983),
    [anon_sym_external] = ACTIONS(2035),
    [anon_sym_type] = ACTIONS(2035),
    [anon_sym_COLON2] = ACTIONS(1985),
    [anon_sym_LBRACK] = ACTIONS(1987),
    [anon_sym_RBRACK] = ACTIONS(2033),
    [anon_sym_true] = ACTIONS(1989),
    [anon_sym_false] = ACTIONS(1989),
    [anon_sym_LBRACE] = ACTIONS(1991),
    [anon_sym_exception] = ACTIONS(2035),
    [anon_sym_module] = ACTIONS(2035),
    [anon_sym_open] = ACTIONS(2035),
    [anon_sym_include] = ACTIONS(2035),
    [anon_sym_class] = ACTIONS(2035),
    [anon_sym_val] = ACTIONS(2035),
    [anon_sym_end] = ACTIONS(2035),
    [anon_sym_POUND] = ACTIONS(1993),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1995),
    [anon_sym_begin] = ACTIONS(1997),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2033),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2001),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2033),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2033),
    [aux_sym_number_token1] = ACTIONS(2005),
    [aux_sym_signed_number_token1] = ACTIONS(2007),
    [anon_sym_SQUOTE] = ACTIONS(2009),
    [anon_sym_DQUOTE] = ACTIONS(2011),
    [sym_let_operator] = ACTIONS(2033),
    [sym__uppercase_identifier] = ACTIONS(2013),
    [aux_sym_directive_token1] = ACTIONS(2035),
    [aux_sym_tag_token1] = ACTIONS(2015),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(851)] = {
    [sym_attribute] = STATE(851),
    [sym__lowercase_identifier] = ACTIONS(1547),
    [anon_sym_TILDE] = ACTIONS(1547),
    [anon_sym_QMARK] = ACTIONS(1547),
    [anon_sym_LPAREN] = ACTIONS(1549),
    [anon_sym_RPAREN] = ACTIONS(1549),
    [anon_sym_COMMA] = ACTIONS(1549),
    [anon_sym_PIPE] = ACTIONS(1547),
    [anon_sym_COLON2] = ACTIONS(1547),
    [anon_sym_DOT] = ACTIONS(1549),
    [anon_sym_DASH_GT] = ACTIONS(1547),
    [anon_sym_LBRACK] = ACTIONS(1547),
    [anon_sym_RBRACK] = ACTIONS(1549),
    [anon_sym_true] = ACTIONS(1547),
    [anon_sym_false] = ACTIONS(1547),
    [anon_sym_LBRACE] = ACTIONS(1547),
    [anon_sym_SEMI] = ACTIONS(1549),
    [anon_sym_RBRACE] = ACTIONS(1549),
    [anon_sym_constraint] = ACTIONS(1547),
    [anon_sym_val] = ACTIONS(1547),
    [anon_sym_end] = ACTIONS(1547),
    [anon_sym_with] = ACTIONS(1547),
    [anon_sym_COLON_GT] = ACTIONS(1549),
    [anon_sym_object] = ACTIONS(1547),
    [anon_sym_inherit] = ACTIONS(1547),
    [anon_sym_method] = ACTIONS(1547),
    [anon_sym_initializer] = ACTIONS(1547),
    [anon_sym_AMP] = ACTIONS(1547),
    [anon_sym_POUND] = ACTIONS(1547),
    [anon_sym_COLON_COLON] = ACTIONS(1549),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1549),
    [anon_sym_LT_DASH] = ACTIONS(1547),
    [anon_sym_then] = ACTIONS(1547),
    [anon_sym_else] = ACTIONS(1547),
    [anon_sym_do] = ACTIONS(1547),
    [anon_sym_done] = ACTIONS(1547),
    [anon_sym_to] = ACTIONS(1547),
    [anon_sym_downto] = ACTIONS(1547),
    [anon_sym_new] = ACTIONS(1547),
    [anon_sym_LBRACE_LT] = ACTIONS(1549),
    [anon_sym_GT_RBRACE] = ACTIONS(1549),
    [anon_sym_begin] = ACTIONS(1547),
    [sym_ocamlyacc_value] = ACTIONS(1549),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1547),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1549),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1547),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1547),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1549),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1549),
    [aux_sym_number_token1] = ACTIONS(1549),
    [aux_sym_signed_number_token1] = ACTIONS(1547),
    [anon_sym_SQUOTE] = ACTIONS(1549),
    [anon_sym_DQUOTE] = ACTIONS(1549),
    [sym_prefix_operator] = ACTIONS(1549),
    [aux_sym_sign_operator_token1] = ACTIONS(1547),
    [sym_hash_operator] = ACTIONS(1549),
    [anon_sym_lsl] = ACTIONS(1547),
    [anon_sym_lsr] = ACTIONS(1547),
    [anon_sym_asr] = ACTIONS(1547),
    [aux_sym_pow_operator_token1] = ACTIONS(1549),
    [anon_sym_mod] = ACTIONS(1547),
    [anon_sym_land] = ACTIONS(1547),
    [anon_sym_lor] = ACTIONS(1547),
    [anon_sym_lxor] = ACTIONS(1547),
    [aux_sym_mult_operator_token1] = ACTIONS(1547),
    [aux_sym_add_operator_token1] = ACTIONS(1547),
    [sym_concat_operator] = ACTIONS(1549),
    [sym_rel_operator] = ACTIONS(1547),
    [anon_sym_AMP_AMP] = ACTIONS(1547),
    [anon_sym_or] = ACTIONS(1547),
    [anon_sym_PIPE_PIPE] = ACTIONS(1547),
    [sym_assign_operator] = ACTIONS(1549),
    [sym__uppercase_identifier] = ACTIONS(1549),
    [aux_sym_tag_token1] = ACTIONS(1549),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(852)] = {
    [sym_attribute] = STATE(852),
    [sym__lowercase_identifier] = ACTIONS(1535),
    [anon_sym_TILDE] = ACTIONS(1535),
    [anon_sym_QMARK] = ACTIONS(1535),
    [anon_sym_LPAREN] = ACTIONS(1537),
    [anon_sym_RPAREN] = ACTIONS(1537),
    [anon_sym_COMMA] = ACTIONS(1537),
    [anon_sym_PIPE] = ACTIONS(1535),
    [anon_sym_COLON2] = ACTIONS(1535),
    [anon_sym_DOT] = ACTIONS(1537),
    [anon_sym_DASH_GT] = ACTIONS(1535),
    [anon_sym_LBRACK] = ACTIONS(1535),
    [anon_sym_RBRACK] = ACTIONS(1537),
    [anon_sym_true] = ACTIONS(1535),
    [anon_sym_false] = ACTIONS(1535),
    [anon_sym_LBRACE] = ACTIONS(1535),
    [anon_sym_SEMI] = ACTIONS(1537),
    [anon_sym_RBRACE] = ACTIONS(1537),
    [anon_sym_constraint] = ACTIONS(1535),
    [anon_sym_val] = ACTIONS(1535),
    [anon_sym_end] = ACTIONS(1535),
    [anon_sym_with] = ACTIONS(1535),
    [anon_sym_COLON_GT] = ACTIONS(1537),
    [anon_sym_object] = ACTIONS(1535),
    [anon_sym_inherit] = ACTIONS(1535),
    [anon_sym_method] = ACTIONS(1535),
    [anon_sym_initializer] = ACTIONS(1535),
    [anon_sym_AMP] = ACTIONS(1535),
    [anon_sym_POUND] = ACTIONS(1535),
    [anon_sym_COLON_COLON] = ACTIONS(1537),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1537),
    [anon_sym_LT_DASH] = ACTIONS(1535),
    [anon_sym_then] = ACTIONS(1535),
    [anon_sym_else] = ACTIONS(1535),
    [anon_sym_do] = ACTIONS(1535),
    [anon_sym_done] = ACTIONS(1535),
    [anon_sym_to] = ACTIONS(1535),
    [anon_sym_downto] = ACTIONS(1535),
    [anon_sym_new] = ACTIONS(1535),
    [anon_sym_LBRACE_LT] = ACTIONS(1537),
    [anon_sym_GT_RBRACE] = ACTIONS(1537),
    [anon_sym_begin] = ACTIONS(1535),
    [sym_ocamlyacc_value] = ACTIONS(1537),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1535),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1537),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1535),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1535),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1537),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1537),
    [aux_sym_number_token1] = ACTIONS(1537),
    [aux_sym_signed_number_token1] = ACTIONS(1535),
    [anon_sym_SQUOTE] = ACTIONS(1537),
    [anon_sym_DQUOTE] = ACTIONS(1537),
    [sym_prefix_operator] = ACTIONS(1537),
    [aux_sym_sign_operator_token1] = ACTIONS(1535),
    [sym_hash_operator] = ACTIONS(1537),
    [anon_sym_lsl] = ACTIONS(1535),
    [anon_sym_lsr] = ACTIONS(1535),
    [anon_sym_asr] = ACTIONS(1535),
    [aux_sym_pow_operator_token1] = ACTIONS(1537),
    [anon_sym_mod] = ACTIONS(1535),
    [anon_sym_land] = ACTIONS(1535),
    [anon_sym_lor] = ACTIONS(1535),
    [anon_sym_lxor] = ACTIONS(1535),
    [aux_sym_mult_operator_token1] = ACTIONS(1535),
    [aux_sym_add_operator_token1] = ACTIONS(1535),
    [sym_concat_operator] = ACTIONS(1537),
    [sym_rel_operator] = ACTIONS(1535),
    [anon_sym_AMP_AMP] = ACTIONS(1535),
    [anon_sym_or] = ACTIONS(1535),
    [anon_sym_PIPE_PIPE] = ACTIONS(1535),
    [sym_assign_operator] = ACTIONS(1537),
    [sym__uppercase_identifier] = ACTIONS(1537),
    [aux_sym_tag_token1] = ACTIONS(1537),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(853)] = {
    [sym_parameter] = STATE(2473),
    [sym__class_typed] = STATE(3532),
    [sym__parenthesized_abstract_type] = STATE(2473),
    [sym__delimited_pattern] = STATE(2474),
    [sym__simple_pattern] = STATE(2446),
    [sym_typed_pattern] = STATE(2474),
    [sym_polymorphic_variant_pattern] = STATE(2474),
    [sym_record_pattern] = STATE(2428),
    [sym_list_pattern] = STATE(2428),
    [sym_array_pattern] = STATE(2428),
    [sym_local_open_pattern] = STATE(2474),
    [sym_package_pattern] = STATE(2474),
    [sym_parenthesized_pattern] = STATE(2428),
    [sym_attribute] = STATE(853),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2474),
    [sym_quoted_extension] = STATE(2474),
    [sym__constant] = STATE(2426),
    [sym__signed_constant] = STATE(2474),
    [sym_number] = STATE(2434),
    [sym_signed_number] = STATE(2426),
    [sym_character] = STATE(2434),
    [sym_string] = STATE(2434),
    [sym_quoted_string] = STATE(2434),
    [sym_boolean] = STATE(2434),
    [sym_unit] = STATE(2428),
    [sym__simple_value_pattern] = STATE(2474),
    [sym_parenthesized_operator] = STATE(2474),
    [sym_module_path] = STATE(6496),
    [sym_constructor_path] = STATE(2474),
    [sym__constructor_name] = STATE(2476),
    [sym_tag] = STATE(2474),
    [aux_sym_expression_item_repeat1] = STATE(3602),
    [aux_sym_let_binding_repeat1] = STATE(856),
    [sym__lowercase_identifier] = ACTIONS(1973),
    [anon_sym_SEMI_SEMI] = ACTIONS(2039),
    [anon_sym_let] = ACTIONS(2041),
    [anon_sym_and] = ACTIONS(2041),
    [anon_sym_EQ] = ACTIONS(2043),
    [anon_sym_TILDE] = ACTIONS(1981),
    [anon_sym_QMARK] = ACTIONS(1981),
    [anon_sym_LPAREN] = ACTIONS(1983),
    [anon_sym_external] = ACTIONS(2041),
    [anon_sym_type] = ACTIONS(2041),
    [anon_sym_COLON2] = ACTIONS(1985),
    [anon_sym_LBRACK] = ACTIONS(1987),
    [anon_sym_RBRACK] = ACTIONS(2039),
    [anon_sym_true] = ACTIONS(1989),
    [anon_sym_false] = ACTIONS(1989),
    [anon_sym_LBRACE] = ACTIONS(1991),
    [anon_sym_exception] = ACTIONS(2041),
    [anon_sym_module] = ACTIONS(2041),
    [anon_sym_open] = ACTIONS(2041),
    [anon_sym_include] = ACTIONS(2041),
    [anon_sym_class] = ACTIONS(2041),
    [anon_sym_val] = ACTIONS(2041),
    [anon_sym_end] = ACTIONS(2041),
    [anon_sym_POUND] = ACTIONS(1993),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1995),
    [anon_sym_begin] = ACTIONS(1997),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2039),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2001),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2039),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2039),
    [aux_sym_number_token1] = ACTIONS(2005),
    [aux_sym_signed_number_token1] = ACTIONS(2007),
    [anon_sym_SQUOTE] = ACTIONS(2009),
    [anon_sym_DQUOTE] = ACTIONS(2011),
    [sym_let_operator] = ACTIONS(2039),
    [sym__uppercase_identifier] = ACTIONS(2013),
    [aux_sym_directive_token1] = ACTIONS(2041),
    [aux_sym_tag_token1] = ACTIONS(2015),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(854)] = {
    [sym_attribute] = STATE(854),
    [sym__lowercase_identifier] = ACTIONS(1531),
    [anon_sym_TILDE] = ACTIONS(1531),
    [anon_sym_QMARK] = ACTIONS(1531),
    [anon_sym_LPAREN] = ACTIONS(1533),
    [anon_sym_RPAREN] = ACTIONS(1533),
    [anon_sym_COMMA] = ACTIONS(1533),
    [anon_sym_PIPE] = ACTIONS(1531),
    [anon_sym_COLON2] = ACTIONS(1531),
    [anon_sym_DOT] = ACTIONS(1533),
    [anon_sym_DASH_GT] = ACTIONS(1531),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_RBRACK] = ACTIONS(1533),
    [anon_sym_true] = ACTIONS(1531),
    [anon_sym_false] = ACTIONS(1531),
    [anon_sym_LBRACE] = ACTIONS(1531),
    [anon_sym_SEMI] = ACTIONS(1533),
    [anon_sym_RBRACE] = ACTIONS(1533),
    [anon_sym_constraint] = ACTIONS(1531),
    [anon_sym_val] = ACTIONS(1531),
    [anon_sym_end] = ACTIONS(1531),
    [anon_sym_with] = ACTIONS(1531),
    [anon_sym_COLON_GT] = ACTIONS(1533),
    [anon_sym_object] = ACTIONS(1531),
    [anon_sym_inherit] = ACTIONS(1531),
    [anon_sym_method] = ACTIONS(1531),
    [anon_sym_initializer] = ACTIONS(1531),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_POUND] = ACTIONS(1531),
    [anon_sym_COLON_COLON] = ACTIONS(1533),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LT_DASH] = ACTIONS(1531),
    [anon_sym_then] = ACTIONS(1531),
    [anon_sym_else] = ACTIONS(1531),
    [anon_sym_do] = ACTIONS(1531),
    [anon_sym_done] = ACTIONS(1531),
    [anon_sym_to] = ACTIONS(1531),
    [anon_sym_downto] = ACTIONS(1531),
    [anon_sym_new] = ACTIONS(1531),
    [anon_sym_LBRACE_LT] = ACTIONS(1533),
    [anon_sym_GT_RBRACE] = ACTIONS(1533),
    [anon_sym_begin] = ACTIONS(1531),
    [sym_ocamlyacc_value] = ACTIONS(1533),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1531),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1533),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1531),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1531),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1533),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1533),
    [aux_sym_number_token1] = ACTIONS(1533),
    [aux_sym_signed_number_token1] = ACTIONS(1531),
    [anon_sym_SQUOTE] = ACTIONS(1533),
    [anon_sym_DQUOTE] = ACTIONS(1533),
    [sym_prefix_operator] = ACTIONS(1533),
    [aux_sym_sign_operator_token1] = ACTIONS(1531),
    [sym_hash_operator] = ACTIONS(1533),
    [anon_sym_lsl] = ACTIONS(1531),
    [anon_sym_lsr] = ACTIONS(1531),
    [anon_sym_asr] = ACTIONS(1531),
    [aux_sym_pow_operator_token1] = ACTIONS(1533),
    [anon_sym_mod] = ACTIONS(1531),
    [anon_sym_land] = ACTIONS(1531),
    [anon_sym_lor] = ACTIONS(1531),
    [anon_sym_lxor] = ACTIONS(1531),
    [aux_sym_mult_operator_token1] = ACTIONS(1531),
    [aux_sym_add_operator_token1] = ACTIONS(1531),
    [sym_concat_operator] = ACTIONS(1533),
    [sym_rel_operator] = ACTIONS(1531),
    [anon_sym_AMP_AMP] = ACTIONS(1531),
    [anon_sym_or] = ACTIONS(1531),
    [anon_sym_PIPE_PIPE] = ACTIONS(1531),
    [sym_assign_operator] = ACTIONS(1533),
    [sym__uppercase_identifier] = ACTIONS(1533),
    [aux_sym_tag_token1] = ACTIONS(1533),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(855)] = {
    [sym_parameter] = STATE(2473),
    [sym__class_typed] = STATE(3556),
    [sym__parenthesized_abstract_type] = STATE(2473),
    [sym__delimited_pattern] = STATE(2474),
    [sym__simple_pattern] = STATE(2446),
    [sym_typed_pattern] = STATE(2474),
    [sym_polymorphic_variant_pattern] = STATE(2474),
    [sym_record_pattern] = STATE(2428),
    [sym_list_pattern] = STATE(2428),
    [sym_array_pattern] = STATE(2428),
    [sym_local_open_pattern] = STATE(2474),
    [sym_package_pattern] = STATE(2474),
    [sym_parenthesized_pattern] = STATE(2428),
    [sym_attribute] = STATE(855),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2474),
    [sym_quoted_extension] = STATE(2474),
    [sym__constant] = STATE(2426),
    [sym__signed_constant] = STATE(2474),
    [sym_number] = STATE(2434),
    [sym_signed_number] = STATE(2426),
    [sym_character] = STATE(2434),
    [sym_string] = STATE(2434),
    [sym_quoted_string] = STATE(2434),
    [sym_boolean] = STATE(2434),
    [sym_unit] = STATE(2428),
    [sym__simple_value_pattern] = STATE(2474),
    [sym_parenthesized_operator] = STATE(2474),
    [sym_module_path] = STATE(6496),
    [sym_constructor_path] = STATE(2474),
    [sym__constructor_name] = STATE(2476),
    [sym_tag] = STATE(2474),
    [aux_sym_expression_item_repeat1] = STATE(3735),
    [aux_sym_let_binding_repeat1] = STATE(849),
    [sym__lowercase_identifier] = ACTIONS(1973),
    [anon_sym_SEMI_SEMI] = ACTIONS(2045),
    [anon_sym_let] = ACTIONS(2047),
    [anon_sym_and] = ACTIONS(2047),
    [anon_sym_EQ] = ACTIONS(2049),
    [anon_sym_TILDE] = ACTIONS(1981),
    [anon_sym_QMARK] = ACTIONS(1981),
    [anon_sym_LPAREN] = ACTIONS(1983),
    [anon_sym_external] = ACTIONS(2047),
    [anon_sym_type] = ACTIONS(2047),
    [anon_sym_COLON2] = ACTIONS(1985),
    [anon_sym_LBRACK] = ACTIONS(1987),
    [anon_sym_RBRACK] = ACTIONS(2045),
    [anon_sym_true] = ACTIONS(1989),
    [anon_sym_false] = ACTIONS(1989),
    [anon_sym_LBRACE] = ACTIONS(1991),
    [anon_sym_exception] = ACTIONS(2047),
    [anon_sym_module] = ACTIONS(2047),
    [anon_sym_open] = ACTIONS(2047),
    [anon_sym_include] = ACTIONS(2047),
    [anon_sym_class] = ACTIONS(2047),
    [anon_sym_val] = ACTIONS(2047),
    [anon_sym_end] = ACTIONS(2047),
    [anon_sym_POUND] = ACTIONS(1993),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1995),
    [anon_sym_begin] = ACTIONS(1997),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2045),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2001),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2045),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2045),
    [aux_sym_number_token1] = ACTIONS(2005),
    [aux_sym_signed_number_token1] = ACTIONS(2007),
    [anon_sym_SQUOTE] = ACTIONS(2009),
    [anon_sym_DQUOTE] = ACTIONS(2011),
    [sym_let_operator] = ACTIONS(2045),
    [sym__uppercase_identifier] = ACTIONS(2013),
    [aux_sym_directive_token1] = ACTIONS(2047),
    [aux_sym_tag_token1] = ACTIONS(2015),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(856)] = {
    [sym_parameter] = STATE(2473),
    [sym__class_typed] = STATE(3547),
    [sym__parenthesized_abstract_type] = STATE(2473),
    [sym__delimited_pattern] = STATE(2474),
    [sym__simple_pattern] = STATE(2446),
    [sym_typed_pattern] = STATE(2474),
    [sym_polymorphic_variant_pattern] = STATE(2474),
    [sym_record_pattern] = STATE(2428),
    [sym_list_pattern] = STATE(2428),
    [sym_array_pattern] = STATE(2428),
    [sym_local_open_pattern] = STATE(2474),
    [sym_package_pattern] = STATE(2474),
    [sym_parenthesized_pattern] = STATE(2428),
    [sym_attribute] = STATE(856),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2474),
    [sym_quoted_extension] = STATE(2474),
    [sym__constant] = STATE(2426),
    [sym__signed_constant] = STATE(2474),
    [sym_number] = STATE(2434),
    [sym_signed_number] = STATE(2426),
    [sym_character] = STATE(2434),
    [sym_string] = STATE(2434),
    [sym_quoted_string] = STATE(2434),
    [sym_boolean] = STATE(2434),
    [sym_unit] = STATE(2428),
    [sym__simple_value_pattern] = STATE(2474),
    [sym_parenthesized_operator] = STATE(2474),
    [sym_module_path] = STATE(6496),
    [sym_constructor_path] = STATE(2474),
    [sym__constructor_name] = STATE(2476),
    [sym_tag] = STATE(2474),
    [aux_sym_expression_item_repeat1] = STATE(3736),
    [aux_sym_let_binding_repeat1] = STATE(982),
    [sym__lowercase_identifier] = ACTIONS(1973),
    [anon_sym_SEMI_SEMI] = ACTIONS(2051),
    [anon_sym_let] = ACTIONS(2053),
    [anon_sym_and] = ACTIONS(2053),
    [anon_sym_EQ] = ACTIONS(2055),
    [anon_sym_TILDE] = ACTIONS(1981),
    [anon_sym_QMARK] = ACTIONS(1981),
    [anon_sym_LPAREN] = ACTIONS(1983),
    [anon_sym_external] = ACTIONS(2053),
    [anon_sym_type] = ACTIONS(2053),
    [anon_sym_COLON2] = ACTIONS(1985),
    [anon_sym_LBRACK] = ACTIONS(1987),
    [anon_sym_RBRACK] = ACTIONS(2051),
    [anon_sym_true] = ACTIONS(1989),
    [anon_sym_false] = ACTIONS(1989),
    [anon_sym_LBRACE] = ACTIONS(1991),
    [anon_sym_exception] = ACTIONS(2053),
    [anon_sym_module] = ACTIONS(2053),
    [anon_sym_open] = ACTIONS(2053),
    [anon_sym_include] = ACTIONS(2053),
    [anon_sym_class] = ACTIONS(2053),
    [anon_sym_val] = ACTIONS(2053),
    [anon_sym_end] = ACTIONS(2053),
    [anon_sym_POUND] = ACTIONS(1993),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1995),
    [anon_sym_begin] = ACTIONS(1997),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2051),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2001),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2051),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2051),
    [aux_sym_number_token1] = ACTIONS(2005),
    [aux_sym_signed_number_token1] = ACTIONS(2007),
    [anon_sym_SQUOTE] = ACTIONS(2009),
    [anon_sym_DQUOTE] = ACTIONS(2011),
    [sym_let_operator] = ACTIONS(2051),
    [sym__uppercase_identifier] = ACTIONS(2013),
    [aux_sym_directive_token1] = ACTIONS(2053),
    [aux_sym_tag_token1] = ACTIONS(2015),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(857)] = {
    [sym_parameter] = STATE(2473),
    [sym__class_typed] = STATE(3538),
    [sym__parenthesized_abstract_type] = STATE(2473),
    [sym__delimited_pattern] = STATE(2474),
    [sym__simple_pattern] = STATE(2446),
    [sym_typed_pattern] = STATE(2474),
    [sym_polymorphic_variant_pattern] = STATE(2474),
    [sym_record_pattern] = STATE(2428),
    [sym_list_pattern] = STATE(2428),
    [sym_array_pattern] = STATE(2428),
    [sym_local_open_pattern] = STATE(2474),
    [sym_package_pattern] = STATE(2474),
    [sym_parenthesized_pattern] = STATE(2428),
    [sym_attribute] = STATE(857),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2474),
    [sym_quoted_extension] = STATE(2474),
    [sym__constant] = STATE(2426),
    [sym__signed_constant] = STATE(2474),
    [sym_number] = STATE(2434),
    [sym_signed_number] = STATE(2426),
    [sym_character] = STATE(2434),
    [sym_string] = STATE(2434),
    [sym_quoted_string] = STATE(2434),
    [sym_boolean] = STATE(2434),
    [sym_unit] = STATE(2428),
    [sym__simple_value_pattern] = STATE(2474),
    [sym_parenthesized_operator] = STATE(2474),
    [sym_module_path] = STATE(6496),
    [sym_constructor_path] = STATE(2474),
    [sym__constructor_name] = STATE(2476),
    [sym_tag] = STATE(2474),
    [aux_sym_expression_item_repeat1] = STATE(3626),
    [aux_sym_let_binding_repeat1] = STATE(982),
    [sym__lowercase_identifier] = ACTIONS(1973),
    [anon_sym_SEMI_SEMI] = ACTIONS(2057),
    [anon_sym_let] = ACTIONS(2059),
    [anon_sym_and] = ACTIONS(2059),
    [anon_sym_EQ] = ACTIONS(2061),
    [anon_sym_TILDE] = ACTIONS(1981),
    [anon_sym_QMARK] = ACTIONS(1981),
    [anon_sym_LPAREN] = ACTIONS(1983),
    [anon_sym_external] = ACTIONS(2059),
    [anon_sym_type] = ACTIONS(2059),
    [anon_sym_COLON2] = ACTIONS(1985),
    [anon_sym_LBRACK] = ACTIONS(1987),
    [anon_sym_RBRACK] = ACTIONS(2057),
    [anon_sym_true] = ACTIONS(1989),
    [anon_sym_false] = ACTIONS(1989),
    [anon_sym_LBRACE] = ACTIONS(1991),
    [anon_sym_exception] = ACTIONS(2059),
    [anon_sym_module] = ACTIONS(2059),
    [anon_sym_open] = ACTIONS(2059),
    [anon_sym_include] = ACTIONS(2059),
    [anon_sym_class] = ACTIONS(2059),
    [anon_sym_val] = ACTIONS(2059),
    [anon_sym_end] = ACTIONS(2059),
    [anon_sym_POUND] = ACTIONS(1993),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1995),
    [anon_sym_begin] = ACTIONS(1997),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2057),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2001),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2057),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2057),
    [aux_sym_number_token1] = ACTIONS(2005),
    [aux_sym_signed_number_token1] = ACTIONS(2007),
    [anon_sym_SQUOTE] = ACTIONS(2009),
    [anon_sym_DQUOTE] = ACTIONS(2011),
    [sym_let_operator] = ACTIONS(2057),
    [sym__uppercase_identifier] = ACTIONS(2013),
    [aux_sym_directive_token1] = ACTIONS(2059),
    [aux_sym_tag_token1] = ACTIONS(2015),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(858)] = {
    [sym_attribute] = STATE(858),
    [sym__lowercase_identifier] = ACTIONS(1559),
    [anon_sym_TILDE] = ACTIONS(1559),
    [anon_sym_QMARK] = ACTIONS(1559),
    [anon_sym_LPAREN] = ACTIONS(1561),
    [anon_sym_RPAREN] = ACTIONS(1561),
    [anon_sym_COMMA] = ACTIONS(1561),
    [anon_sym_PIPE] = ACTIONS(1559),
    [anon_sym_COLON2] = ACTIONS(1559),
    [anon_sym_DOT] = ACTIONS(1561),
    [anon_sym_DASH_GT] = ACTIONS(1559),
    [anon_sym_LBRACK] = ACTIONS(1559),
    [anon_sym_RBRACK] = ACTIONS(1561),
    [anon_sym_true] = ACTIONS(1559),
    [anon_sym_false] = ACTIONS(1559),
    [anon_sym_LBRACE] = ACTIONS(1559),
    [anon_sym_SEMI] = ACTIONS(1561),
    [anon_sym_RBRACE] = ACTIONS(1561),
    [anon_sym_constraint] = ACTIONS(1559),
    [anon_sym_val] = ACTIONS(1559),
    [anon_sym_end] = ACTIONS(1559),
    [anon_sym_with] = ACTIONS(1559),
    [anon_sym_COLON_GT] = ACTIONS(1561),
    [anon_sym_object] = ACTIONS(1559),
    [anon_sym_inherit] = ACTIONS(1559),
    [anon_sym_method] = ACTIONS(1559),
    [anon_sym_initializer] = ACTIONS(1559),
    [anon_sym_AMP] = ACTIONS(1559),
    [anon_sym_POUND] = ACTIONS(1559),
    [anon_sym_COLON_COLON] = ACTIONS(1561),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1561),
    [anon_sym_LT_DASH] = ACTIONS(1559),
    [anon_sym_then] = ACTIONS(1559),
    [anon_sym_else] = ACTIONS(1559),
    [anon_sym_do] = ACTIONS(1559),
    [anon_sym_done] = ACTIONS(1559),
    [anon_sym_to] = ACTIONS(1559),
    [anon_sym_downto] = ACTIONS(1559),
    [anon_sym_new] = ACTIONS(1559),
    [anon_sym_LBRACE_LT] = ACTIONS(1561),
    [anon_sym_GT_RBRACE] = ACTIONS(1561),
    [anon_sym_begin] = ACTIONS(1559),
    [sym_ocamlyacc_value] = ACTIONS(1561),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1559),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1561),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1559),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1559),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1561),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1561),
    [aux_sym_number_token1] = ACTIONS(1561),
    [aux_sym_signed_number_token1] = ACTIONS(1559),
    [anon_sym_SQUOTE] = ACTIONS(1561),
    [anon_sym_DQUOTE] = ACTIONS(1561),
    [sym_prefix_operator] = ACTIONS(1561),
    [aux_sym_sign_operator_token1] = ACTIONS(1559),
    [sym_hash_operator] = ACTIONS(1561),
    [anon_sym_lsl] = ACTIONS(1559),
    [anon_sym_lsr] = ACTIONS(1559),
    [anon_sym_asr] = ACTIONS(1559),
    [aux_sym_pow_operator_token1] = ACTIONS(1561),
    [anon_sym_mod] = ACTIONS(1559),
    [anon_sym_land] = ACTIONS(1559),
    [anon_sym_lor] = ACTIONS(1559),
    [anon_sym_lxor] = ACTIONS(1559),
    [aux_sym_mult_operator_token1] = ACTIONS(1559),
    [aux_sym_add_operator_token1] = ACTIONS(1559),
    [sym_concat_operator] = ACTIONS(1561),
    [sym_rel_operator] = ACTIONS(1559),
    [anon_sym_AMP_AMP] = ACTIONS(1559),
    [anon_sym_or] = ACTIONS(1559),
    [anon_sym_PIPE_PIPE] = ACTIONS(1559),
    [sym_assign_operator] = ACTIONS(1561),
    [sym__uppercase_identifier] = ACTIONS(1561),
    [aux_sym_tag_token1] = ACTIONS(1561),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(859)] = {
    [sym_attribute] = STATE(859),
    [sym__lowercase_identifier] = ACTIONS(1563),
    [anon_sym_TILDE] = ACTIONS(1563),
    [anon_sym_QMARK] = ACTIONS(1563),
    [anon_sym_LPAREN] = ACTIONS(1565),
    [anon_sym_RPAREN] = ACTIONS(1565),
    [anon_sym_COMMA] = ACTIONS(1565),
    [anon_sym_PIPE] = ACTIONS(1563),
    [anon_sym_COLON2] = ACTIONS(1563),
    [anon_sym_DOT] = ACTIONS(1565),
    [anon_sym_DASH_GT] = ACTIONS(1563),
    [anon_sym_LBRACK] = ACTIONS(1563),
    [anon_sym_RBRACK] = ACTIONS(1565),
    [anon_sym_true] = ACTIONS(1563),
    [anon_sym_false] = ACTIONS(1563),
    [anon_sym_LBRACE] = ACTIONS(1563),
    [anon_sym_SEMI] = ACTIONS(1565),
    [anon_sym_RBRACE] = ACTIONS(1565),
    [anon_sym_constraint] = ACTIONS(1563),
    [anon_sym_val] = ACTIONS(1563),
    [anon_sym_end] = ACTIONS(1563),
    [anon_sym_with] = ACTIONS(1563),
    [anon_sym_COLON_GT] = ACTIONS(1565),
    [anon_sym_object] = ACTIONS(1563),
    [anon_sym_inherit] = ACTIONS(1563),
    [anon_sym_method] = ACTIONS(1563),
    [anon_sym_initializer] = ACTIONS(1563),
    [anon_sym_AMP] = ACTIONS(1563),
    [anon_sym_POUND] = ACTIONS(1563),
    [anon_sym_COLON_COLON] = ACTIONS(1565),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1565),
    [anon_sym_LT_DASH] = ACTIONS(1563),
    [anon_sym_then] = ACTIONS(1563),
    [anon_sym_else] = ACTIONS(1563),
    [anon_sym_do] = ACTIONS(1563),
    [anon_sym_done] = ACTIONS(1563),
    [anon_sym_to] = ACTIONS(1563),
    [anon_sym_downto] = ACTIONS(1563),
    [anon_sym_new] = ACTIONS(1563),
    [anon_sym_LBRACE_LT] = ACTIONS(1565),
    [anon_sym_GT_RBRACE] = ACTIONS(1565),
    [anon_sym_begin] = ACTIONS(1563),
    [sym_ocamlyacc_value] = ACTIONS(1565),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1563),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1565),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1563),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1563),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1565),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1565),
    [aux_sym_number_token1] = ACTIONS(1565),
    [aux_sym_signed_number_token1] = ACTIONS(1563),
    [anon_sym_SQUOTE] = ACTIONS(1565),
    [anon_sym_DQUOTE] = ACTIONS(1565),
    [sym_prefix_operator] = ACTIONS(1565),
    [aux_sym_sign_operator_token1] = ACTIONS(1563),
    [sym_hash_operator] = ACTIONS(1565),
    [anon_sym_lsl] = ACTIONS(1563),
    [anon_sym_lsr] = ACTIONS(1563),
    [anon_sym_asr] = ACTIONS(1563),
    [aux_sym_pow_operator_token1] = ACTIONS(1565),
    [anon_sym_mod] = ACTIONS(1563),
    [anon_sym_land] = ACTIONS(1563),
    [anon_sym_lor] = ACTIONS(1563),
    [anon_sym_lxor] = ACTIONS(1563),
    [aux_sym_mult_operator_token1] = ACTIONS(1563),
    [aux_sym_add_operator_token1] = ACTIONS(1563),
    [sym_concat_operator] = ACTIONS(1565),
    [sym_rel_operator] = ACTIONS(1563),
    [anon_sym_AMP_AMP] = ACTIONS(1563),
    [anon_sym_or] = ACTIONS(1563),
    [anon_sym_PIPE_PIPE] = ACTIONS(1563),
    [sym_assign_operator] = ACTIONS(1565),
    [sym__uppercase_identifier] = ACTIONS(1565),
    [aux_sym_tag_token1] = ACTIONS(1565),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(860)] = {
    [sym_parameter] = STATE(2473),
    [sym__class_typed] = STATE(3552),
    [sym__parenthesized_abstract_type] = STATE(2473),
    [sym__delimited_pattern] = STATE(2474),
    [sym__simple_pattern] = STATE(2446),
    [sym_typed_pattern] = STATE(2474),
    [sym_polymorphic_variant_pattern] = STATE(2474),
    [sym_record_pattern] = STATE(2428),
    [sym_list_pattern] = STATE(2428),
    [sym_array_pattern] = STATE(2428),
    [sym_local_open_pattern] = STATE(2474),
    [sym_package_pattern] = STATE(2474),
    [sym_parenthesized_pattern] = STATE(2428),
    [sym_attribute] = STATE(860),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2474),
    [sym_quoted_extension] = STATE(2474),
    [sym__constant] = STATE(2426),
    [sym__signed_constant] = STATE(2474),
    [sym_number] = STATE(2434),
    [sym_signed_number] = STATE(2426),
    [sym_character] = STATE(2434),
    [sym_string] = STATE(2434),
    [sym_quoted_string] = STATE(2434),
    [sym_boolean] = STATE(2434),
    [sym_unit] = STATE(2428),
    [sym__simple_value_pattern] = STATE(2474),
    [sym_parenthesized_operator] = STATE(2474),
    [sym_module_path] = STATE(6496),
    [sym_constructor_path] = STATE(2474),
    [sym__constructor_name] = STATE(2476),
    [sym_tag] = STATE(2474),
    [aux_sym_expression_item_repeat1] = STATE(3721),
    [aux_sym_let_binding_repeat1] = STATE(982),
    [sym__lowercase_identifier] = ACTIONS(1973),
    [anon_sym_SEMI_SEMI] = ACTIONS(2063),
    [anon_sym_let] = ACTIONS(2065),
    [anon_sym_and] = ACTIONS(2065),
    [anon_sym_EQ] = ACTIONS(2067),
    [anon_sym_TILDE] = ACTIONS(1981),
    [anon_sym_QMARK] = ACTIONS(1981),
    [anon_sym_LPAREN] = ACTIONS(1983),
    [anon_sym_external] = ACTIONS(2065),
    [anon_sym_type] = ACTIONS(2065),
    [anon_sym_COLON2] = ACTIONS(1985),
    [anon_sym_LBRACK] = ACTIONS(1987),
    [anon_sym_RBRACK] = ACTIONS(2063),
    [anon_sym_true] = ACTIONS(1989),
    [anon_sym_false] = ACTIONS(1989),
    [anon_sym_LBRACE] = ACTIONS(1991),
    [anon_sym_exception] = ACTIONS(2065),
    [anon_sym_module] = ACTIONS(2065),
    [anon_sym_open] = ACTIONS(2065),
    [anon_sym_include] = ACTIONS(2065),
    [anon_sym_class] = ACTIONS(2065),
    [anon_sym_val] = ACTIONS(2065),
    [anon_sym_end] = ACTIONS(2065),
    [anon_sym_POUND] = ACTIONS(1993),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1995),
    [anon_sym_begin] = ACTIONS(1997),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2063),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2001),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2063),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2063),
    [aux_sym_number_token1] = ACTIONS(2005),
    [aux_sym_signed_number_token1] = ACTIONS(2007),
    [anon_sym_SQUOTE] = ACTIONS(2009),
    [anon_sym_DQUOTE] = ACTIONS(2011),
    [sym_let_operator] = ACTIONS(2063),
    [sym__uppercase_identifier] = ACTIONS(2013),
    [aux_sym_directive_token1] = ACTIONS(2065),
    [aux_sym_tag_token1] = ACTIONS(2015),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(861)] = {
    [sym_parameter] = STATE(2473),
    [sym__class_typed] = STATE(3553),
    [sym__parenthesized_abstract_type] = STATE(2473),
    [sym__delimited_pattern] = STATE(2474),
    [sym__simple_pattern] = STATE(2446),
    [sym_typed_pattern] = STATE(2474),
    [sym_polymorphic_variant_pattern] = STATE(2474),
    [sym_record_pattern] = STATE(2428),
    [sym_list_pattern] = STATE(2428),
    [sym_array_pattern] = STATE(2428),
    [sym_local_open_pattern] = STATE(2474),
    [sym_package_pattern] = STATE(2474),
    [sym_parenthesized_pattern] = STATE(2428),
    [sym_attribute] = STATE(861),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2474),
    [sym_quoted_extension] = STATE(2474),
    [sym__constant] = STATE(2426),
    [sym__signed_constant] = STATE(2474),
    [sym_number] = STATE(2434),
    [sym_signed_number] = STATE(2426),
    [sym_character] = STATE(2434),
    [sym_string] = STATE(2434),
    [sym_quoted_string] = STATE(2434),
    [sym_boolean] = STATE(2434),
    [sym_unit] = STATE(2428),
    [sym__simple_value_pattern] = STATE(2474),
    [sym_parenthesized_operator] = STATE(2474),
    [sym_module_path] = STATE(6496),
    [sym_constructor_path] = STATE(2474),
    [sym__constructor_name] = STATE(2476),
    [sym_tag] = STATE(2474),
    [aux_sym_expression_item_repeat1] = STATE(3723),
    [aux_sym_let_binding_repeat1] = STATE(850),
    [sym__lowercase_identifier] = ACTIONS(1973),
    [anon_sym_SEMI_SEMI] = ACTIONS(2069),
    [anon_sym_let] = ACTIONS(2071),
    [anon_sym_and] = ACTIONS(2071),
    [anon_sym_EQ] = ACTIONS(2073),
    [anon_sym_TILDE] = ACTIONS(1981),
    [anon_sym_QMARK] = ACTIONS(1981),
    [anon_sym_LPAREN] = ACTIONS(1983),
    [anon_sym_external] = ACTIONS(2071),
    [anon_sym_type] = ACTIONS(2071),
    [anon_sym_COLON2] = ACTIONS(1985),
    [anon_sym_LBRACK] = ACTIONS(1987),
    [anon_sym_RBRACK] = ACTIONS(2069),
    [anon_sym_true] = ACTIONS(1989),
    [anon_sym_false] = ACTIONS(1989),
    [anon_sym_LBRACE] = ACTIONS(1991),
    [anon_sym_exception] = ACTIONS(2071),
    [anon_sym_module] = ACTIONS(2071),
    [anon_sym_open] = ACTIONS(2071),
    [anon_sym_include] = ACTIONS(2071),
    [anon_sym_class] = ACTIONS(2071),
    [anon_sym_val] = ACTIONS(2071),
    [anon_sym_end] = ACTIONS(2071),
    [anon_sym_POUND] = ACTIONS(1993),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1995),
    [anon_sym_begin] = ACTIONS(1997),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2069),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2001),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2003),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2069),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2069),
    [aux_sym_number_token1] = ACTIONS(2005),
    [aux_sym_signed_number_token1] = ACTIONS(2007),
    [anon_sym_SQUOTE] = ACTIONS(2009),
    [anon_sym_DQUOTE] = ACTIONS(2011),
    [sym_let_operator] = ACTIONS(2069),
    [sym__uppercase_identifier] = ACTIONS(2013),
    [aux_sym_directive_token1] = ACTIONS(2071),
    [aux_sym_tag_token1] = ACTIONS(2015),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(862)] = {
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2961),
    [sym__binding_pattern] = STATE(2974),
    [sym_alias_binding_pattern] = STATE(2961),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_anonymous] = STATE(2962),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6582),
    [sym__tuple_binding_pattern] = STATE(2964),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern] = STATE(2961),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2961),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_exception_binding_pattern] = STATE(2961),
    [sym_effect_binding_pattern] = STATE(2961),
    [sym_attribute] = STATE(862),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_and] = ACTIONS(1297),
    [anon_sym_EQ] = ACTIONS(1295),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_QMARK] = ACTIONS(1295),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_COMMA] = ACTIONS(1295),
    [anon_sym_PIPE] = ACTIONS(1295),
    [anon_sym_COLON2] = ACTIONS(1297),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1959),
    [anon_sym_COLON_GT] = ACTIONS(1295),
    [anon_sym_in] = ACTIONS(1297),
    [anon_sym_as] = ACTIONS(1297),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_COLON_COLON] = ACTIONS(1295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [anon_sym_effect] = ACTIONS(1961),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1295),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1965),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym_let_and_operator] = ACTIONS(1295),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(863)] = {
    [sym_attribute] = STATE(863),
    [sym__lowercase_identifier] = ACTIONS(1729),
    [anon_sym_TILDE] = ACTIONS(1729),
    [anon_sym_QMARK] = ACTIONS(1729),
    [anon_sym_LPAREN] = ACTIONS(1731),
    [anon_sym_RPAREN] = ACTIONS(1731),
    [anon_sym_COMMA] = ACTIONS(1731),
    [anon_sym_PIPE] = ACTIONS(1729),
    [anon_sym_COLON2] = ACTIONS(1729),
    [anon_sym_DOT] = ACTIONS(1731),
    [anon_sym_DASH_GT] = ACTIONS(1729),
    [anon_sym_LBRACK] = ACTIONS(1729),
    [anon_sym_RBRACK] = ACTIONS(1731),
    [anon_sym_true] = ACTIONS(1729),
    [anon_sym_false] = ACTIONS(1729),
    [anon_sym_LBRACE] = ACTIONS(1729),
    [anon_sym_SEMI] = ACTIONS(1731),
    [anon_sym_RBRACE] = ACTIONS(1731),
    [anon_sym_constraint] = ACTIONS(1729),
    [anon_sym_val] = ACTIONS(1729),
    [anon_sym_end] = ACTIONS(1729),
    [anon_sym_with] = ACTIONS(1729),
    [anon_sym_COLON_GT] = ACTIONS(1731),
    [anon_sym_object] = ACTIONS(1729),
    [anon_sym_inherit] = ACTIONS(1729),
    [anon_sym_method] = ACTIONS(1729),
    [anon_sym_initializer] = ACTIONS(1729),
    [anon_sym_AMP] = ACTIONS(1729),
    [anon_sym_POUND] = ACTIONS(1729),
    [anon_sym_COLON_COLON] = ACTIONS(1731),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1731),
    [anon_sym_then] = ACTIONS(1729),
    [anon_sym_else] = ACTIONS(1729),
    [anon_sym_do] = ACTIONS(1729),
    [anon_sym_done] = ACTIONS(1729),
    [anon_sym_to] = ACTIONS(1729),
    [anon_sym_downto] = ACTIONS(1729),
    [anon_sym_new] = ACTIONS(1729),
    [anon_sym_LBRACE_LT] = ACTIONS(1731),
    [anon_sym_GT_RBRACE] = ACTIONS(1731),
    [anon_sym_begin] = ACTIONS(1729),
    [sym_ocamlyacc_value] = ACTIONS(1731),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1729),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1731),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1729),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1729),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1731),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1731),
    [aux_sym_number_token1] = ACTIONS(1731),
    [aux_sym_signed_number_token1] = ACTIONS(1729),
    [anon_sym_SQUOTE] = ACTIONS(1731),
    [anon_sym_DQUOTE] = ACTIONS(1731),
    [sym_prefix_operator] = ACTIONS(1731),
    [aux_sym_sign_operator_token1] = ACTIONS(1729),
    [sym_hash_operator] = ACTIONS(1731),
    [anon_sym_lsl] = ACTIONS(1729),
    [anon_sym_lsr] = ACTIONS(1729),
    [anon_sym_asr] = ACTIONS(1729),
    [aux_sym_pow_operator_token1] = ACTIONS(1731),
    [anon_sym_mod] = ACTIONS(1729),
    [anon_sym_land] = ACTIONS(1729),
    [anon_sym_lor] = ACTIONS(1729),
    [anon_sym_lxor] = ACTIONS(1729),
    [aux_sym_mult_operator_token1] = ACTIONS(1729),
    [aux_sym_add_operator_token1] = ACTIONS(1729),
    [sym_concat_operator] = ACTIONS(1731),
    [sym_rel_operator] = ACTIONS(1729),
    [anon_sym_AMP_AMP] = ACTIONS(1729),
    [anon_sym_or] = ACTIONS(1729),
    [anon_sym_PIPE_PIPE] = ACTIONS(1729),
    [sym_assign_operator] = ACTIONS(1731),
    [sym__uppercase_identifier] = ACTIONS(1731),
    [aux_sym_tag_token1] = ACTIONS(1731),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(864)] = {
    [sym_attribute] = STATE(864),
    [sym__lowercase_identifier] = ACTIONS(1765),
    [anon_sym_TILDE] = ACTIONS(1765),
    [anon_sym_QMARK] = ACTIONS(1765),
    [anon_sym_LPAREN] = ACTIONS(1767),
    [anon_sym_RPAREN] = ACTIONS(1767),
    [anon_sym_COMMA] = ACTIONS(1767),
    [anon_sym_PIPE] = ACTIONS(1765),
    [anon_sym_COLON2] = ACTIONS(1765),
    [anon_sym_DOT] = ACTIONS(1767),
    [anon_sym_DASH_GT] = ACTIONS(1765),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_RBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1765),
    [anon_sym_false] = ACTIONS(1765),
    [anon_sym_LBRACE] = ACTIONS(1765),
    [anon_sym_SEMI] = ACTIONS(1767),
    [anon_sym_RBRACE] = ACTIONS(1767),
    [anon_sym_constraint] = ACTIONS(1765),
    [anon_sym_val] = ACTIONS(1765),
    [anon_sym_end] = ACTIONS(1765),
    [anon_sym_with] = ACTIONS(1765),
    [anon_sym_COLON_GT] = ACTIONS(1767),
    [anon_sym_object] = ACTIONS(1765),
    [anon_sym_inherit] = ACTIONS(1765),
    [anon_sym_method] = ACTIONS(1765),
    [anon_sym_initializer] = ACTIONS(1765),
    [anon_sym_AMP] = ACTIONS(1765),
    [anon_sym_POUND] = ACTIONS(1765),
    [anon_sym_COLON_COLON] = ACTIONS(1767),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_then] = ACTIONS(1765),
    [anon_sym_else] = ACTIONS(1765),
    [anon_sym_do] = ACTIONS(1765),
    [anon_sym_done] = ACTIONS(1765),
    [anon_sym_to] = ACTIONS(1765),
    [anon_sym_downto] = ACTIONS(1765),
    [anon_sym_new] = ACTIONS(1765),
    [anon_sym_LBRACE_LT] = ACTIONS(1767),
    [anon_sym_GT_RBRACE] = ACTIONS(1767),
    [anon_sym_begin] = ACTIONS(1765),
    [sym_ocamlyacc_value] = ACTIONS(1767),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1765),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1767),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1765),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1765),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1767),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1767),
    [aux_sym_number_token1] = ACTIONS(1767),
    [aux_sym_signed_number_token1] = ACTIONS(1765),
    [anon_sym_SQUOTE] = ACTIONS(1767),
    [anon_sym_DQUOTE] = ACTIONS(1767),
    [sym_prefix_operator] = ACTIONS(1767),
    [aux_sym_sign_operator_token1] = ACTIONS(1765),
    [sym_hash_operator] = ACTIONS(1767),
    [anon_sym_lsl] = ACTIONS(1765),
    [anon_sym_lsr] = ACTIONS(1765),
    [anon_sym_asr] = ACTIONS(1765),
    [aux_sym_pow_operator_token1] = ACTIONS(1767),
    [anon_sym_mod] = ACTIONS(1765),
    [anon_sym_land] = ACTIONS(1765),
    [anon_sym_lor] = ACTIONS(1765),
    [anon_sym_lxor] = ACTIONS(1765),
    [aux_sym_mult_operator_token1] = ACTIONS(1765),
    [aux_sym_add_operator_token1] = ACTIONS(1765),
    [sym_concat_operator] = ACTIONS(1767),
    [sym_rel_operator] = ACTIONS(1765),
    [anon_sym_AMP_AMP] = ACTIONS(1765),
    [anon_sym_or] = ACTIONS(1765),
    [anon_sym_PIPE_PIPE] = ACTIONS(1765),
    [sym_assign_operator] = ACTIONS(1767),
    [sym__uppercase_identifier] = ACTIONS(1767),
    [aux_sym_tag_token1] = ACTIONS(1767),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(865)] = {
    [sym_attribute] = STATE(865),
    [sym__lowercase_identifier] = ACTIONS(1773),
    [anon_sym_TILDE] = ACTIONS(1773),
    [anon_sym_QMARK] = ACTIONS(1773),
    [anon_sym_LPAREN] = ACTIONS(1775),
    [anon_sym_RPAREN] = ACTIONS(1775),
    [anon_sym_COMMA] = ACTIONS(1775),
    [anon_sym_PIPE] = ACTIONS(1773),
    [anon_sym_COLON2] = ACTIONS(1773),
    [anon_sym_DOT] = ACTIONS(1775),
    [anon_sym_DASH_GT] = ACTIONS(1773),
    [anon_sym_LBRACK] = ACTIONS(1773),
    [anon_sym_RBRACK] = ACTIONS(1775),
    [anon_sym_true] = ACTIONS(1773),
    [anon_sym_false] = ACTIONS(1773),
    [anon_sym_LBRACE] = ACTIONS(1773),
    [anon_sym_SEMI] = ACTIONS(1775),
    [anon_sym_RBRACE] = ACTIONS(1775),
    [anon_sym_constraint] = ACTIONS(1773),
    [anon_sym_val] = ACTIONS(1773),
    [anon_sym_end] = ACTIONS(1773),
    [anon_sym_with] = ACTIONS(1773),
    [anon_sym_COLON_GT] = ACTIONS(1775),
    [anon_sym_object] = ACTIONS(1773),
    [anon_sym_inherit] = ACTIONS(1773),
    [anon_sym_method] = ACTIONS(1773),
    [anon_sym_initializer] = ACTIONS(1773),
    [anon_sym_AMP] = ACTIONS(1773),
    [anon_sym_POUND] = ACTIONS(1773),
    [anon_sym_COLON_COLON] = ACTIONS(1775),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1775),
    [anon_sym_then] = ACTIONS(1773),
    [anon_sym_else] = ACTIONS(1773),
    [anon_sym_do] = ACTIONS(1773),
    [anon_sym_done] = ACTIONS(1773),
    [anon_sym_to] = ACTIONS(1773),
    [anon_sym_downto] = ACTIONS(1773),
    [anon_sym_new] = ACTIONS(1773),
    [anon_sym_LBRACE_LT] = ACTIONS(1775),
    [anon_sym_GT_RBRACE] = ACTIONS(1775),
    [anon_sym_begin] = ACTIONS(1773),
    [sym_ocamlyacc_value] = ACTIONS(1775),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1773),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1775),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1773),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1773),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1775),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1775),
    [aux_sym_number_token1] = ACTIONS(1775),
    [aux_sym_signed_number_token1] = ACTIONS(1773),
    [anon_sym_SQUOTE] = ACTIONS(1775),
    [anon_sym_DQUOTE] = ACTIONS(1775),
    [sym_prefix_operator] = ACTIONS(1775),
    [aux_sym_sign_operator_token1] = ACTIONS(1773),
    [sym_hash_operator] = ACTIONS(1775),
    [anon_sym_lsl] = ACTIONS(1773),
    [anon_sym_lsr] = ACTIONS(1773),
    [anon_sym_asr] = ACTIONS(1773),
    [aux_sym_pow_operator_token1] = ACTIONS(1775),
    [anon_sym_mod] = ACTIONS(1773),
    [anon_sym_land] = ACTIONS(1773),
    [anon_sym_lor] = ACTIONS(1773),
    [anon_sym_lxor] = ACTIONS(1773),
    [aux_sym_mult_operator_token1] = ACTIONS(1773),
    [aux_sym_add_operator_token1] = ACTIONS(1773),
    [sym_concat_operator] = ACTIONS(1775),
    [sym_rel_operator] = ACTIONS(1773),
    [anon_sym_AMP_AMP] = ACTIONS(1773),
    [anon_sym_or] = ACTIONS(1773),
    [anon_sym_PIPE_PIPE] = ACTIONS(1773),
    [sym_assign_operator] = ACTIONS(1775),
    [sym__uppercase_identifier] = ACTIONS(1775),
    [aux_sym_tag_token1] = ACTIONS(1775),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(866)] = {
    [sym_attribute] = STATE(866),
    [sym__lowercase_identifier] = ACTIONS(1615),
    [anon_sym_TILDE] = ACTIONS(1615),
    [anon_sym_QMARK] = ACTIONS(1615),
    [anon_sym_LPAREN] = ACTIONS(1617),
    [anon_sym_RPAREN] = ACTIONS(1617),
    [anon_sym_COMMA] = ACTIONS(1617),
    [anon_sym_PIPE] = ACTIONS(1615),
    [anon_sym_COLON2] = ACTIONS(1615),
    [anon_sym_DOT] = ACTIONS(825),
    [anon_sym_DASH_GT] = ACTIONS(1615),
    [anon_sym_LBRACK] = ACTIONS(1615),
    [anon_sym_RBRACK] = ACTIONS(1617),
    [anon_sym_true] = ACTIONS(1615),
    [anon_sym_false] = ACTIONS(1615),
    [anon_sym_LBRACE] = ACTIONS(1615),
    [anon_sym_SEMI] = ACTIONS(1617),
    [anon_sym_RBRACE] = ACTIONS(1617),
    [anon_sym_constraint] = ACTIONS(1615),
    [anon_sym_val] = ACTIONS(1615),
    [anon_sym_end] = ACTIONS(1615),
    [anon_sym_with] = ACTIONS(1615),
    [anon_sym_COLON_GT] = ACTIONS(1617),
    [anon_sym_object] = ACTIONS(1615),
    [anon_sym_inherit] = ACTIONS(1615),
    [anon_sym_method] = ACTIONS(1615),
    [anon_sym_initializer] = ACTIONS(1615),
    [anon_sym_AMP] = ACTIONS(1615),
    [anon_sym_POUND] = ACTIONS(827),
    [anon_sym_COLON_COLON] = ACTIONS(1617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1617),
    [anon_sym_then] = ACTIONS(1615),
    [anon_sym_else] = ACTIONS(1615),
    [anon_sym_do] = ACTIONS(1615),
    [anon_sym_done] = ACTIONS(1615),
    [anon_sym_to] = ACTIONS(1615),
    [anon_sym_downto] = ACTIONS(1615),
    [anon_sym_new] = ACTIONS(1615),
    [anon_sym_LBRACE_LT] = ACTIONS(1617),
    [anon_sym_GT_RBRACE] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1615),
    [sym_ocamlyacc_value] = ACTIONS(1617),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1615),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1617),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1615),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1615),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1617),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1617),
    [aux_sym_number_token1] = ACTIONS(1617),
    [aux_sym_signed_number_token1] = ACTIONS(1615),
    [anon_sym_SQUOTE] = ACTIONS(1617),
    [anon_sym_DQUOTE] = ACTIONS(1617),
    [sym_prefix_operator] = ACTIONS(1617),
    [aux_sym_sign_operator_token1] = ACTIONS(1615),
    [sym_hash_operator] = ACTIONS(829),
    [anon_sym_lsl] = ACTIONS(1615),
    [anon_sym_lsr] = ACTIONS(1615),
    [anon_sym_asr] = ACTIONS(1615),
    [aux_sym_pow_operator_token1] = ACTIONS(1617),
    [anon_sym_mod] = ACTIONS(1615),
    [anon_sym_land] = ACTIONS(1615),
    [anon_sym_lor] = ACTIONS(1615),
    [anon_sym_lxor] = ACTIONS(1615),
    [aux_sym_mult_operator_token1] = ACTIONS(1615),
    [aux_sym_add_operator_token1] = ACTIONS(1615),
    [sym_concat_operator] = ACTIONS(1617),
    [sym_rel_operator] = ACTIONS(1615),
    [anon_sym_AMP_AMP] = ACTIONS(1615),
    [anon_sym_or] = ACTIONS(1615),
    [anon_sym_PIPE_PIPE] = ACTIONS(1615),
    [sym_assign_operator] = ACTIONS(1617),
    [sym__uppercase_identifier] = ACTIONS(1617),
    [aux_sym_tag_token1] = ACTIONS(1617),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(867)] = {
    [sym_attribute] = STATE(867),
    [sym__lowercase_identifier] = ACTIONS(1741),
    [anon_sym_TILDE] = ACTIONS(1741),
    [anon_sym_QMARK] = ACTIONS(1741),
    [anon_sym_LPAREN] = ACTIONS(1743),
    [anon_sym_RPAREN] = ACTIONS(1743),
    [anon_sym_COMMA] = ACTIONS(1743),
    [anon_sym_PIPE] = ACTIONS(1741),
    [anon_sym_COLON2] = ACTIONS(1741),
    [anon_sym_DOT] = ACTIONS(1743),
    [anon_sym_DASH_GT] = ACTIONS(1741),
    [anon_sym_LBRACK] = ACTIONS(1741),
    [anon_sym_RBRACK] = ACTIONS(1743),
    [anon_sym_true] = ACTIONS(1741),
    [anon_sym_false] = ACTIONS(1741),
    [anon_sym_LBRACE] = ACTIONS(1741),
    [anon_sym_SEMI] = ACTIONS(1743),
    [anon_sym_RBRACE] = ACTIONS(1743),
    [anon_sym_constraint] = ACTIONS(1741),
    [anon_sym_val] = ACTIONS(1741),
    [anon_sym_end] = ACTIONS(1741),
    [anon_sym_with] = ACTIONS(1741),
    [anon_sym_COLON_GT] = ACTIONS(1743),
    [anon_sym_object] = ACTIONS(1741),
    [anon_sym_inherit] = ACTIONS(1741),
    [anon_sym_method] = ACTIONS(1741),
    [anon_sym_initializer] = ACTIONS(1741),
    [anon_sym_AMP] = ACTIONS(1741),
    [anon_sym_POUND] = ACTIONS(1741),
    [anon_sym_COLON_COLON] = ACTIONS(1743),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1743),
    [anon_sym_then] = ACTIONS(1741),
    [anon_sym_else] = ACTIONS(1741),
    [anon_sym_do] = ACTIONS(1741),
    [anon_sym_done] = ACTIONS(1741),
    [anon_sym_to] = ACTIONS(1741),
    [anon_sym_downto] = ACTIONS(1741),
    [anon_sym_new] = ACTIONS(1741),
    [anon_sym_LBRACE_LT] = ACTIONS(1743),
    [anon_sym_GT_RBRACE] = ACTIONS(1743),
    [anon_sym_begin] = ACTIONS(1741),
    [sym_ocamlyacc_value] = ACTIONS(1743),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1741),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1743),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1741),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1741),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1743),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1743),
    [aux_sym_number_token1] = ACTIONS(1743),
    [aux_sym_signed_number_token1] = ACTIONS(1741),
    [anon_sym_SQUOTE] = ACTIONS(1743),
    [anon_sym_DQUOTE] = ACTIONS(1743),
    [sym_prefix_operator] = ACTIONS(1743),
    [aux_sym_sign_operator_token1] = ACTIONS(1741),
    [sym_hash_operator] = ACTIONS(1743),
    [anon_sym_lsl] = ACTIONS(1741),
    [anon_sym_lsr] = ACTIONS(1741),
    [anon_sym_asr] = ACTIONS(1741),
    [aux_sym_pow_operator_token1] = ACTIONS(1743),
    [anon_sym_mod] = ACTIONS(1741),
    [anon_sym_land] = ACTIONS(1741),
    [anon_sym_lor] = ACTIONS(1741),
    [anon_sym_lxor] = ACTIONS(1741),
    [aux_sym_mult_operator_token1] = ACTIONS(1741),
    [aux_sym_add_operator_token1] = ACTIONS(1741),
    [sym_concat_operator] = ACTIONS(1743),
    [sym_rel_operator] = ACTIONS(1741),
    [anon_sym_AMP_AMP] = ACTIONS(1741),
    [anon_sym_or] = ACTIONS(1741),
    [anon_sym_PIPE_PIPE] = ACTIONS(1741),
    [sym_assign_operator] = ACTIONS(1743),
    [sym__uppercase_identifier] = ACTIONS(1743),
    [aux_sym_tag_token1] = ACTIONS(1743),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(868)] = {
    [sym_attribute] = STATE(868),
    [sym__lowercase_identifier] = ACTIONS(1691),
    [anon_sym_TILDE] = ACTIONS(1691),
    [anon_sym_QMARK] = ACTIONS(1691),
    [anon_sym_LPAREN] = ACTIONS(1693),
    [anon_sym_RPAREN] = ACTIONS(1693),
    [anon_sym_COMMA] = ACTIONS(1693),
    [anon_sym_PIPE] = ACTIONS(1691),
    [anon_sym_COLON2] = ACTIONS(1691),
    [anon_sym_DOT] = ACTIONS(1693),
    [anon_sym_DASH_GT] = ACTIONS(1691),
    [anon_sym_LBRACK] = ACTIONS(1691),
    [anon_sym_RBRACK] = ACTIONS(1693),
    [anon_sym_true] = ACTIONS(1691),
    [anon_sym_false] = ACTIONS(1691),
    [anon_sym_LBRACE] = ACTIONS(1691),
    [anon_sym_SEMI] = ACTIONS(1693),
    [anon_sym_RBRACE] = ACTIONS(1693),
    [anon_sym_constraint] = ACTIONS(1691),
    [anon_sym_val] = ACTIONS(1691),
    [anon_sym_end] = ACTIONS(1691),
    [anon_sym_with] = ACTIONS(1691),
    [anon_sym_COLON_GT] = ACTIONS(1693),
    [anon_sym_object] = ACTIONS(1691),
    [anon_sym_inherit] = ACTIONS(1691),
    [anon_sym_method] = ACTIONS(1691),
    [anon_sym_initializer] = ACTIONS(1691),
    [anon_sym_AMP] = ACTIONS(1691),
    [anon_sym_POUND] = ACTIONS(1691),
    [anon_sym_COLON_COLON] = ACTIONS(1693),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1693),
    [anon_sym_then] = ACTIONS(1691),
    [anon_sym_else] = ACTIONS(1691),
    [anon_sym_do] = ACTIONS(1691),
    [anon_sym_done] = ACTIONS(1691),
    [anon_sym_to] = ACTIONS(1691),
    [anon_sym_downto] = ACTIONS(1691),
    [anon_sym_new] = ACTIONS(1691),
    [anon_sym_LBRACE_LT] = ACTIONS(1693),
    [anon_sym_GT_RBRACE] = ACTIONS(1693),
    [anon_sym_begin] = ACTIONS(1691),
    [sym_ocamlyacc_value] = ACTIONS(1693),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1691),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1693),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1691),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1691),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1693),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1693),
    [aux_sym_number_token1] = ACTIONS(1693),
    [aux_sym_signed_number_token1] = ACTIONS(1691),
    [anon_sym_SQUOTE] = ACTIONS(1693),
    [anon_sym_DQUOTE] = ACTIONS(1693),
    [sym_prefix_operator] = ACTIONS(1693),
    [aux_sym_sign_operator_token1] = ACTIONS(1691),
    [sym_hash_operator] = ACTIONS(1693),
    [anon_sym_lsl] = ACTIONS(1691),
    [anon_sym_lsr] = ACTIONS(1691),
    [anon_sym_asr] = ACTIONS(1691),
    [aux_sym_pow_operator_token1] = ACTIONS(1693),
    [anon_sym_mod] = ACTIONS(1691),
    [anon_sym_land] = ACTIONS(1691),
    [anon_sym_lor] = ACTIONS(1691),
    [anon_sym_lxor] = ACTIONS(1691),
    [aux_sym_mult_operator_token1] = ACTIONS(1691),
    [aux_sym_add_operator_token1] = ACTIONS(1691),
    [sym_concat_operator] = ACTIONS(1693),
    [sym_rel_operator] = ACTIONS(1691),
    [anon_sym_AMP_AMP] = ACTIONS(1691),
    [anon_sym_or] = ACTIONS(1691),
    [anon_sym_PIPE_PIPE] = ACTIONS(1691),
    [sym_assign_operator] = ACTIONS(1693),
    [sym__uppercase_identifier] = ACTIONS(1693),
    [aux_sym_tag_token1] = ACTIONS(1693),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(869)] = {
    [sym_attribute] = STATE(869),
    [sym__lowercase_identifier] = ACTIONS(1737),
    [anon_sym_TILDE] = ACTIONS(1737),
    [anon_sym_QMARK] = ACTIONS(1737),
    [anon_sym_LPAREN] = ACTIONS(1739),
    [anon_sym_RPAREN] = ACTIONS(1739),
    [anon_sym_COMMA] = ACTIONS(1739),
    [anon_sym_PIPE] = ACTIONS(1737),
    [anon_sym_COLON2] = ACTIONS(1737),
    [anon_sym_DOT] = ACTIONS(1739),
    [anon_sym_DASH_GT] = ACTIONS(1737),
    [anon_sym_LBRACK] = ACTIONS(1737),
    [anon_sym_RBRACK] = ACTIONS(1739),
    [anon_sym_true] = ACTIONS(1737),
    [anon_sym_false] = ACTIONS(1737),
    [anon_sym_LBRACE] = ACTIONS(1737),
    [anon_sym_SEMI] = ACTIONS(1739),
    [anon_sym_RBRACE] = ACTIONS(1739),
    [anon_sym_constraint] = ACTIONS(1737),
    [anon_sym_val] = ACTIONS(1737),
    [anon_sym_end] = ACTIONS(1737),
    [anon_sym_with] = ACTIONS(1737),
    [anon_sym_COLON_GT] = ACTIONS(1739),
    [anon_sym_object] = ACTIONS(1737),
    [anon_sym_inherit] = ACTIONS(1737),
    [anon_sym_method] = ACTIONS(1737),
    [anon_sym_initializer] = ACTIONS(1737),
    [anon_sym_AMP] = ACTIONS(1737),
    [anon_sym_POUND] = ACTIONS(1737),
    [anon_sym_COLON_COLON] = ACTIONS(1739),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1739),
    [anon_sym_then] = ACTIONS(1737),
    [anon_sym_else] = ACTIONS(1737),
    [anon_sym_do] = ACTIONS(1737),
    [anon_sym_done] = ACTIONS(1737),
    [anon_sym_to] = ACTIONS(1737),
    [anon_sym_downto] = ACTIONS(1737),
    [anon_sym_new] = ACTIONS(1737),
    [anon_sym_LBRACE_LT] = ACTIONS(1739),
    [anon_sym_GT_RBRACE] = ACTIONS(1739),
    [anon_sym_begin] = ACTIONS(1737),
    [sym_ocamlyacc_value] = ACTIONS(1739),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1737),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1739),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1737),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1737),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1739),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1739),
    [aux_sym_number_token1] = ACTIONS(1739),
    [aux_sym_signed_number_token1] = ACTIONS(1737),
    [anon_sym_SQUOTE] = ACTIONS(1739),
    [anon_sym_DQUOTE] = ACTIONS(1739),
    [sym_prefix_operator] = ACTIONS(1739),
    [aux_sym_sign_operator_token1] = ACTIONS(1737),
    [sym_hash_operator] = ACTIONS(1739),
    [anon_sym_lsl] = ACTIONS(1737),
    [anon_sym_lsr] = ACTIONS(1737),
    [anon_sym_asr] = ACTIONS(1737),
    [aux_sym_pow_operator_token1] = ACTIONS(1739),
    [anon_sym_mod] = ACTIONS(1737),
    [anon_sym_land] = ACTIONS(1737),
    [anon_sym_lor] = ACTIONS(1737),
    [anon_sym_lxor] = ACTIONS(1737),
    [aux_sym_mult_operator_token1] = ACTIONS(1737),
    [aux_sym_add_operator_token1] = ACTIONS(1737),
    [sym_concat_operator] = ACTIONS(1739),
    [sym_rel_operator] = ACTIONS(1737),
    [anon_sym_AMP_AMP] = ACTIONS(1737),
    [anon_sym_or] = ACTIONS(1737),
    [anon_sym_PIPE_PIPE] = ACTIONS(1737),
    [sym_assign_operator] = ACTIONS(1739),
    [sym__uppercase_identifier] = ACTIONS(1739),
    [aux_sym_tag_token1] = ACTIONS(1739),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(870)] = {
    [sym_attribute] = STATE(870),
    [sym__lowercase_identifier] = ACTIONS(1571),
    [anon_sym_TILDE] = ACTIONS(1571),
    [anon_sym_QMARK] = ACTIONS(1571),
    [anon_sym_LPAREN] = ACTIONS(1573),
    [anon_sym_RPAREN] = ACTIONS(1573),
    [anon_sym_COMMA] = ACTIONS(1573),
    [anon_sym_PIPE] = ACTIONS(1571),
    [anon_sym_COLON2] = ACTIONS(1571),
    [anon_sym_DOT] = ACTIONS(1573),
    [anon_sym_DASH_GT] = ACTIONS(1571),
    [anon_sym_LBRACK] = ACTIONS(1571),
    [anon_sym_RBRACK] = ACTIONS(1573),
    [anon_sym_true] = ACTIONS(1571),
    [anon_sym_false] = ACTIONS(1571),
    [anon_sym_LBRACE] = ACTIONS(1571),
    [anon_sym_SEMI] = ACTIONS(1573),
    [anon_sym_RBRACE] = ACTIONS(1573),
    [anon_sym_constraint] = ACTIONS(1571),
    [anon_sym_val] = ACTIONS(1571),
    [anon_sym_end] = ACTIONS(1571),
    [anon_sym_with] = ACTIONS(1571),
    [anon_sym_COLON_GT] = ACTIONS(1573),
    [anon_sym_object] = ACTIONS(1571),
    [anon_sym_inherit] = ACTIONS(1571),
    [anon_sym_method] = ACTIONS(1571),
    [anon_sym_initializer] = ACTIONS(1571),
    [anon_sym_AMP] = ACTIONS(1571),
    [anon_sym_POUND] = ACTIONS(1571),
    [anon_sym_COLON_COLON] = ACTIONS(1573),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1573),
    [anon_sym_then] = ACTIONS(1571),
    [anon_sym_else] = ACTIONS(1571),
    [anon_sym_do] = ACTIONS(1571),
    [anon_sym_done] = ACTIONS(1571),
    [anon_sym_to] = ACTIONS(1571),
    [anon_sym_downto] = ACTIONS(1571),
    [anon_sym_new] = ACTIONS(1571),
    [anon_sym_LBRACE_LT] = ACTIONS(1573),
    [anon_sym_GT_RBRACE] = ACTIONS(1573),
    [anon_sym_begin] = ACTIONS(1571),
    [sym_ocamlyacc_value] = ACTIONS(1573),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1571),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1573),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1571),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1571),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1573),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1573),
    [aux_sym_number_token1] = ACTIONS(1573),
    [aux_sym_signed_number_token1] = ACTIONS(1571),
    [anon_sym_SQUOTE] = ACTIONS(1573),
    [anon_sym_DQUOTE] = ACTIONS(1573),
    [sym_prefix_operator] = ACTIONS(1573),
    [aux_sym_sign_operator_token1] = ACTIONS(1571),
    [sym_hash_operator] = ACTIONS(1573),
    [anon_sym_lsl] = ACTIONS(1571),
    [anon_sym_lsr] = ACTIONS(1571),
    [anon_sym_asr] = ACTIONS(1571),
    [aux_sym_pow_operator_token1] = ACTIONS(1573),
    [anon_sym_mod] = ACTIONS(1571),
    [anon_sym_land] = ACTIONS(1571),
    [anon_sym_lor] = ACTIONS(1571),
    [anon_sym_lxor] = ACTIONS(1571),
    [aux_sym_mult_operator_token1] = ACTIONS(1571),
    [aux_sym_add_operator_token1] = ACTIONS(1571),
    [sym_concat_operator] = ACTIONS(1573),
    [sym_rel_operator] = ACTIONS(1571),
    [anon_sym_AMP_AMP] = ACTIONS(1571),
    [anon_sym_or] = ACTIONS(1571),
    [anon_sym_PIPE_PIPE] = ACTIONS(1571),
    [sym_assign_operator] = ACTIONS(1573),
    [sym__uppercase_identifier] = ACTIONS(1573),
    [aux_sym_tag_token1] = ACTIONS(1573),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(871)] = {
    [sym_attribute] = STATE(871),
    [sym__lowercase_identifier] = ACTIONS(1695),
    [anon_sym_TILDE] = ACTIONS(1695),
    [anon_sym_QMARK] = ACTIONS(1695),
    [anon_sym_LPAREN] = ACTIONS(1697),
    [anon_sym_RPAREN] = ACTIONS(1697),
    [anon_sym_COMMA] = ACTIONS(1697),
    [anon_sym_PIPE] = ACTIONS(1695),
    [anon_sym_COLON2] = ACTIONS(1695),
    [anon_sym_DOT] = ACTIONS(1697),
    [anon_sym_DASH_GT] = ACTIONS(1695),
    [anon_sym_LBRACK] = ACTIONS(1695),
    [anon_sym_RBRACK] = ACTIONS(1697),
    [anon_sym_true] = ACTIONS(1695),
    [anon_sym_false] = ACTIONS(1695),
    [anon_sym_LBRACE] = ACTIONS(1695),
    [anon_sym_SEMI] = ACTIONS(1697),
    [anon_sym_RBRACE] = ACTIONS(1697),
    [anon_sym_constraint] = ACTIONS(1695),
    [anon_sym_val] = ACTIONS(1695),
    [anon_sym_end] = ACTIONS(1695),
    [anon_sym_with] = ACTIONS(1695),
    [anon_sym_COLON_GT] = ACTIONS(1697),
    [anon_sym_object] = ACTIONS(1695),
    [anon_sym_inherit] = ACTIONS(1695),
    [anon_sym_method] = ACTIONS(1695),
    [anon_sym_initializer] = ACTIONS(1695),
    [anon_sym_AMP] = ACTIONS(1695),
    [anon_sym_POUND] = ACTIONS(1695),
    [anon_sym_COLON_COLON] = ACTIONS(1697),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1697),
    [anon_sym_then] = ACTIONS(1695),
    [anon_sym_else] = ACTIONS(1695),
    [anon_sym_do] = ACTIONS(1695),
    [anon_sym_done] = ACTIONS(1695),
    [anon_sym_to] = ACTIONS(1695),
    [anon_sym_downto] = ACTIONS(1695),
    [anon_sym_new] = ACTIONS(1695),
    [anon_sym_LBRACE_LT] = ACTIONS(1697),
    [anon_sym_GT_RBRACE] = ACTIONS(1697),
    [anon_sym_begin] = ACTIONS(1695),
    [sym_ocamlyacc_value] = ACTIONS(1697),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1695),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1697),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1695),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1695),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1697),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1697),
    [aux_sym_number_token1] = ACTIONS(1697),
    [aux_sym_signed_number_token1] = ACTIONS(1695),
    [anon_sym_SQUOTE] = ACTIONS(1697),
    [anon_sym_DQUOTE] = ACTIONS(1697),
    [sym_prefix_operator] = ACTIONS(1697),
    [aux_sym_sign_operator_token1] = ACTIONS(1695),
    [sym_hash_operator] = ACTIONS(1697),
    [anon_sym_lsl] = ACTIONS(1695),
    [anon_sym_lsr] = ACTIONS(1695),
    [anon_sym_asr] = ACTIONS(1695),
    [aux_sym_pow_operator_token1] = ACTIONS(1697),
    [anon_sym_mod] = ACTIONS(1695),
    [anon_sym_land] = ACTIONS(1695),
    [anon_sym_lor] = ACTIONS(1695),
    [anon_sym_lxor] = ACTIONS(1695),
    [aux_sym_mult_operator_token1] = ACTIONS(1695),
    [aux_sym_add_operator_token1] = ACTIONS(1695),
    [sym_concat_operator] = ACTIONS(1697),
    [sym_rel_operator] = ACTIONS(1695),
    [anon_sym_AMP_AMP] = ACTIONS(1695),
    [anon_sym_or] = ACTIONS(1695),
    [anon_sym_PIPE_PIPE] = ACTIONS(1695),
    [sym_assign_operator] = ACTIONS(1697),
    [sym__uppercase_identifier] = ACTIONS(1697),
    [aux_sym_tag_token1] = ACTIONS(1697),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(872)] = {
    [sym_attribute] = STATE(872),
    [sym__lowercase_identifier] = ACTIONS(1591),
    [anon_sym_TILDE] = ACTIONS(1591),
    [anon_sym_QMARK] = ACTIONS(1591),
    [anon_sym_LPAREN] = ACTIONS(1593),
    [anon_sym_RPAREN] = ACTIONS(1593),
    [anon_sym_COMMA] = ACTIONS(1593),
    [anon_sym_PIPE] = ACTIONS(1591),
    [anon_sym_COLON2] = ACTIONS(1591),
    [anon_sym_DOT] = ACTIONS(1593),
    [anon_sym_DASH_GT] = ACTIONS(1591),
    [anon_sym_LBRACK] = ACTIONS(1591),
    [anon_sym_RBRACK] = ACTIONS(1593),
    [anon_sym_true] = ACTIONS(1591),
    [anon_sym_false] = ACTIONS(1591),
    [anon_sym_LBRACE] = ACTIONS(1591),
    [anon_sym_SEMI] = ACTIONS(1593),
    [anon_sym_RBRACE] = ACTIONS(1593),
    [anon_sym_constraint] = ACTIONS(1591),
    [anon_sym_val] = ACTIONS(1591),
    [anon_sym_end] = ACTIONS(1591),
    [anon_sym_with] = ACTIONS(1591),
    [anon_sym_COLON_GT] = ACTIONS(1593),
    [anon_sym_object] = ACTIONS(1591),
    [anon_sym_inherit] = ACTIONS(1591),
    [anon_sym_method] = ACTIONS(1591),
    [anon_sym_initializer] = ACTIONS(1591),
    [anon_sym_AMP] = ACTIONS(1591),
    [anon_sym_POUND] = ACTIONS(1591),
    [anon_sym_COLON_COLON] = ACTIONS(1593),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1593),
    [anon_sym_then] = ACTIONS(1591),
    [anon_sym_else] = ACTIONS(1591),
    [anon_sym_do] = ACTIONS(1591),
    [anon_sym_done] = ACTIONS(1591),
    [anon_sym_to] = ACTIONS(1591),
    [anon_sym_downto] = ACTIONS(1591),
    [anon_sym_new] = ACTIONS(1591),
    [anon_sym_LBRACE_LT] = ACTIONS(1593),
    [anon_sym_GT_RBRACE] = ACTIONS(1593),
    [anon_sym_begin] = ACTIONS(1591),
    [sym_ocamlyacc_value] = ACTIONS(1593),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1591),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1593),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1591),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1591),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1593),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1593),
    [aux_sym_number_token1] = ACTIONS(1593),
    [aux_sym_signed_number_token1] = ACTIONS(1591),
    [anon_sym_SQUOTE] = ACTIONS(1593),
    [anon_sym_DQUOTE] = ACTIONS(1593),
    [sym_prefix_operator] = ACTIONS(1593),
    [aux_sym_sign_operator_token1] = ACTIONS(1591),
    [sym_hash_operator] = ACTIONS(1593),
    [anon_sym_lsl] = ACTIONS(1591),
    [anon_sym_lsr] = ACTIONS(1591),
    [anon_sym_asr] = ACTIONS(1591),
    [aux_sym_pow_operator_token1] = ACTIONS(1593),
    [anon_sym_mod] = ACTIONS(1591),
    [anon_sym_land] = ACTIONS(1591),
    [anon_sym_lor] = ACTIONS(1591),
    [anon_sym_lxor] = ACTIONS(1591),
    [aux_sym_mult_operator_token1] = ACTIONS(1591),
    [aux_sym_add_operator_token1] = ACTIONS(1591),
    [sym_concat_operator] = ACTIONS(1593),
    [sym_rel_operator] = ACTIONS(1591),
    [anon_sym_AMP_AMP] = ACTIONS(1591),
    [anon_sym_or] = ACTIONS(1591),
    [anon_sym_PIPE_PIPE] = ACTIONS(1591),
    [sym_assign_operator] = ACTIONS(1593),
    [sym__uppercase_identifier] = ACTIONS(1593),
    [aux_sym_tag_token1] = ACTIONS(1593),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(873)] = {
    [sym_attribute] = STATE(873),
    [sym__lowercase_identifier] = ACTIONS(1619),
    [anon_sym_TILDE] = ACTIONS(1619),
    [anon_sym_QMARK] = ACTIONS(1619),
    [anon_sym_LPAREN] = ACTIONS(1621),
    [anon_sym_RPAREN] = ACTIONS(1621),
    [anon_sym_COMMA] = ACTIONS(1621),
    [anon_sym_PIPE] = ACTIONS(1619),
    [anon_sym_COLON2] = ACTIONS(1619),
    [anon_sym_DOT] = ACTIONS(1621),
    [anon_sym_DASH_GT] = ACTIONS(1619),
    [anon_sym_LBRACK] = ACTIONS(1619),
    [anon_sym_RBRACK] = ACTIONS(1621),
    [anon_sym_true] = ACTIONS(1619),
    [anon_sym_false] = ACTIONS(1619),
    [anon_sym_LBRACE] = ACTIONS(1619),
    [anon_sym_SEMI] = ACTIONS(1621),
    [anon_sym_RBRACE] = ACTIONS(1621),
    [anon_sym_constraint] = ACTIONS(1619),
    [anon_sym_val] = ACTIONS(1619),
    [anon_sym_end] = ACTIONS(1619),
    [anon_sym_with] = ACTIONS(1619),
    [anon_sym_COLON_GT] = ACTIONS(1621),
    [anon_sym_object] = ACTIONS(1619),
    [anon_sym_inherit] = ACTIONS(1619),
    [anon_sym_method] = ACTIONS(1619),
    [anon_sym_initializer] = ACTIONS(1619),
    [anon_sym_AMP] = ACTIONS(1619),
    [anon_sym_POUND] = ACTIONS(1619),
    [anon_sym_COLON_COLON] = ACTIONS(1621),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1621),
    [anon_sym_then] = ACTIONS(1619),
    [anon_sym_else] = ACTIONS(1619),
    [anon_sym_do] = ACTIONS(1619),
    [anon_sym_done] = ACTIONS(1619),
    [anon_sym_to] = ACTIONS(1619),
    [anon_sym_downto] = ACTIONS(1619),
    [anon_sym_new] = ACTIONS(1619),
    [anon_sym_LBRACE_LT] = ACTIONS(1621),
    [anon_sym_GT_RBRACE] = ACTIONS(1621),
    [anon_sym_begin] = ACTIONS(1619),
    [sym_ocamlyacc_value] = ACTIONS(1621),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1619),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1621),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1619),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1619),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1621),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1621),
    [aux_sym_number_token1] = ACTIONS(1621),
    [aux_sym_signed_number_token1] = ACTIONS(1619),
    [anon_sym_SQUOTE] = ACTIONS(1621),
    [anon_sym_DQUOTE] = ACTIONS(1621),
    [sym_prefix_operator] = ACTIONS(1621),
    [aux_sym_sign_operator_token1] = ACTIONS(1619),
    [sym_hash_operator] = ACTIONS(1621),
    [anon_sym_lsl] = ACTIONS(1619),
    [anon_sym_lsr] = ACTIONS(1619),
    [anon_sym_asr] = ACTIONS(1619),
    [aux_sym_pow_operator_token1] = ACTIONS(1621),
    [anon_sym_mod] = ACTIONS(1619),
    [anon_sym_land] = ACTIONS(1619),
    [anon_sym_lor] = ACTIONS(1619),
    [anon_sym_lxor] = ACTIONS(1619),
    [aux_sym_mult_operator_token1] = ACTIONS(1619),
    [aux_sym_add_operator_token1] = ACTIONS(1619),
    [sym_concat_operator] = ACTIONS(1621),
    [sym_rel_operator] = ACTIONS(1619),
    [anon_sym_AMP_AMP] = ACTIONS(1619),
    [anon_sym_or] = ACTIONS(1619),
    [anon_sym_PIPE_PIPE] = ACTIONS(1619),
    [sym_assign_operator] = ACTIONS(1621),
    [sym__uppercase_identifier] = ACTIONS(1621),
    [aux_sym_tag_token1] = ACTIONS(1621),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(874)] = {
    [sym_attribute] = STATE(874),
    [sym__lowercase_identifier] = ACTIONS(1627),
    [anon_sym_TILDE] = ACTIONS(1627),
    [anon_sym_QMARK] = ACTIONS(1627),
    [anon_sym_LPAREN] = ACTIONS(1629),
    [anon_sym_RPAREN] = ACTIONS(1629),
    [anon_sym_COMMA] = ACTIONS(1629),
    [anon_sym_PIPE] = ACTIONS(1627),
    [anon_sym_COLON2] = ACTIONS(1627),
    [anon_sym_DOT] = ACTIONS(1629),
    [anon_sym_DASH_GT] = ACTIONS(1627),
    [anon_sym_LBRACK] = ACTIONS(1627),
    [anon_sym_RBRACK] = ACTIONS(1629),
    [anon_sym_true] = ACTIONS(1627),
    [anon_sym_false] = ACTIONS(1627),
    [anon_sym_LBRACE] = ACTIONS(1627),
    [anon_sym_SEMI] = ACTIONS(1629),
    [anon_sym_RBRACE] = ACTIONS(1629),
    [anon_sym_constraint] = ACTIONS(1627),
    [anon_sym_val] = ACTIONS(1627),
    [anon_sym_end] = ACTIONS(1627),
    [anon_sym_with] = ACTIONS(1627),
    [anon_sym_COLON_GT] = ACTIONS(1629),
    [anon_sym_object] = ACTIONS(1627),
    [anon_sym_inherit] = ACTIONS(1627),
    [anon_sym_method] = ACTIONS(1627),
    [anon_sym_initializer] = ACTIONS(1627),
    [anon_sym_AMP] = ACTIONS(1627),
    [anon_sym_POUND] = ACTIONS(1627),
    [anon_sym_COLON_COLON] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1629),
    [anon_sym_then] = ACTIONS(1627),
    [anon_sym_else] = ACTIONS(1627),
    [anon_sym_do] = ACTIONS(1627),
    [anon_sym_done] = ACTIONS(1627),
    [anon_sym_to] = ACTIONS(1627),
    [anon_sym_downto] = ACTIONS(1627),
    [anon_sym_new] = ACTIONS(1627),
    [anon_sym_LBRACE_LT] = ACTIONS(1629),
    [anon_sym_GT_RBRACE] = ACTIONS(1629),
    [anon_sym_begin] = ACTIONS(1627),
    [sym_ocamlyacc_value] = ACTIONS(1629),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1627),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1629),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1627),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1627),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1629),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1629),
    [aux_sym_number_token1] = ACTIONS(1629),
    [aux_sym_signed_number_token1] = ACTIONS(1627),
    [anon_sym_SQUOTE] = ACTIONS(1629),
    [anon_sym_DQUOTE] = ACTIONS(1629),
    [sym_prefix_operator] = ACTIONS(1629),
    [aux_sym_sign_operator_token1] = ACTIONS(1627),
    [sym_hash_operator] = ACTIONS(1629),
    [anon_sym_lsl] = ACTIONS(1627),
    [anon_sym_lsr] = ACTIONS(1627),
    [anon_sym_asr] = ACTIONS(1627),
    [aux_sym_pow_operator_token1] = ACTIONS(1629),
    [anon_sym_mod] = ACTIONS(1627),
    [anon_sym_land] = ACTIONS(1627),
    [anon_sym_lor] = ACTIONS(1627),
    [anon_sym_lxor] = ACTIONS(1627),
    [aux_sym_mult_operator_token1] = ACTIONS(1627),
    [aux_sym_add_operator_token1] = ACTIONS(1627),
    [sym_concat_operator] = ACTIONS(1629),
    [sym_rel_operator] = ACTIONS(1627),
    [anon_sym_AMP_AMP] = ACTIONS(1627),
    [anon_sym_or] = ACTIONS(1627),
    [anon_sym_PIPE_PIPE] = ACTIONS(1627),
    [sym_assign_operator] = ACTIONS(1629),
    [sym__uppercase_identifier] = ACTIONS(1629),
    [aux_sym_tag_token1] = ACTIONS(1629),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(875)] = {
    [sym_attribute] = STATE(875),
    [sym__lowercase_identifier] = ACTIONS(1655),
    [anon_sym_TILDE] = ACTIONS(1655),
    [anon_sym_QMARK] = ACTIONS(1655),
    [anon_sym_LPAREN] = ACTIONS(1657),
    [anon_sym_RPAREN] = ACTIONS(1657),
    [anon_sym_COMMA] = ACTIONS(1657),
    [anon_sym_PIPE] = ACTIONS(1655),
    [anon_sym_COLON2] = ACTIONS(1655),
    [anon_sym_DOT] = ACTIONS(1657),
    [anon_sym_DASH_GT] = ACTIONS(1655),
    [anon_sym_LBRACK] = ACTIONS(1655),
    [anon_sym_RBRACK] = ACTIONS(1657),
    [anon_sym_true] = ACTIONS(1655),
    [anon_sym_false] = ACTIONS(1655),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_SEMI] = ACTIONS(1657),
    [anon_sym_RBRACE] = ACTIONS(1657),
    [anon_sym_constraint] = ACTIONS(1655),
    [anon_sym_val] = ACTIONS(1655),
    [anon_sym_end] = ACTIONS(1655),
    [anon_sym_with] = ACTIONS(1655),
    [anon_sym_COLON_GT] = ACTIONS(1657),
    [anon_sym_object] = ACTIONS(1655),
    [anon_sym_inherit] = ACTIONS(1655),
    [anon_sym_method] = ACTIONS(1655),
    [anon_sym_initializer] = ACTIONS(1655),
    [anon_sym_AMP] = ACTIONS(1655),
    [anon_sym_POUND] = ACTIONS(1655),
    [anon_sym_COLON_COLON] = ACTIONS(1657),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1657),
    [anon_sym_then] = ACTIONS(1655),
    [anon_sym_else] = ACTIONS(1655),
    [anon_sym_do] = ACTIONS(1655),
    [anon_sym_done] = ACTIONS(1655),
    [anon_sym_to] = ACTIONS(1655),
    [anon_sym_downto] = ACTIONS(1655),
    [anon_sym_new] = ACTIONS(1655),
    [anon_sym_LBRACE_LT] = ACTIONS(1657),
    [anon_sym_GT_RBRACE] = ACTIONS(1657),
    [anon_sym_begin] = ACTIONS(1655),
    [sym_ocamlyacc_value] = ACTIONS(1657),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1655),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1657),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1655),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1655),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1657),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1657),
    [aux_sym_number_token1] = ACTIONS(1657),
    [aux_sym_signed_number_token1] = ACTIONS(1655),
    [anon_sym_SQUOTE] = ACTIONS(1657),
    [anon_sym_DQUOTE] = ACTIONS(1657),
    [sym_prefix_operator] = ACTIONS(1657),
    [aux_sym_sign_operator_token1] = ACTIONS(1655),
    [sym_hash_operator] = ACTIONS(1657),
    [anon_sym_lsl] = ACTIONS(1655),
    [anon_sym_lsr] = ACTIONS(1655),
    [anon_sym_asr] = ACTIONS(1655),
    [aux_sym_pow_operator_token1] = ACTIONS(1657),
    [anon_sym_mod] = ACTIONS(1655),
    [anon_sym_land] = ACTIONS(1655),
    [anon_sym_lor] = ACTIONS(1655),
    [anon_sym_lxor] = ACTIONS(1655),
    [aux_sym_mult_operator_token1] = ACTIONS(1655),
    [aux_sym_add_operator_token1] = ACTIONS(1655),
    [sym_concat_operator] = ACTIONS(1657),
    [sym_rel_operator] = ACTIONS(1655),
    [anon_sym_AMP_AMP] = ACTIONS(1655),
    [anon_sym_or] = ACTIONS(1655),
    [anon_sym_PIPE_PIPE] = ACTIONS(1655),
    [sym_assign_operator] = ACTIONS(1657),
    [sym__uppercase_identifier] = ACTIONS(1657),
    [aux_sym_tag_token1] = ACTIONS(1657),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(876)] = {
    [sym_attribute] = STATE(876),
    [sym__lowercase_identifier] = ACTIONS(1777),
    [anon_sym_TILDE] = ACTIONS(1777),
    [anon_sym_QMARK] = ACTIONS(1777),
    [anon_sym_LPAREN] = ACTIONS(1779),
    [anon_sym_RPAREN] = ACTIONS(1779),
    [anon_sym_COMMA] = ACTIONS(1779),
    [anon_sym_PIPE] = ACTIONS(1777),
    [anon_sym_COLON2] = ACTIONS(1777),
    [anon_sym_DOT] = ACTIONS(1779),
    [anon_sym_DASH_GT] = ACTIONS(1777),
    [anon_sym_LBRACK] = ACTIONS(1777),
    [anon_sym_RBRACK] = ACTIONS(1779),
    [anon_sym_true] = ACTIONS(1777),
    [anon_sym_false] = ACTIONS(1777),
    [anon_sym_LBRACE] = ACTIONS(1777),
    [anon_sym_SEMI] = ACTIONS(1779),
    [anon_sym_RBRACE] = ACTIONS(1779),
    [anon_sym_constraint] = ACTIONS(1777),
    [anon_sym_val] = ACTIONS(1777),
    [anon_sym_end] = ACTIONS(1777),
    [anon_sym_with] = ACTIONS(1777),
    [anon_sym_COLON_GT] = ACTIONS(1779),
    [anon_sym_object] = ACTIONS(1777),
    [anon_sym_inherit] = ACTIONS(1777),
    [anon_sym_method] = ACTIONS(1777),
    [anon_sym_initializer] = ACTIONS(1777),
    [anon_sym_AMP] = ACTIONS(1777),
    [anon_sym_POUND] = ACTIONS(1777),
    [anon_sym_COLON_COLON] = ACTIONS(1779),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1779),
    [anon_sym_then] = ACTIONS(1777),
    [anon_sym_else] = ACTIONS(1777),
    [anon_sym_do] = ACTIONS(1777),
    [anon_sym_done] = ACTIONS(1777),
    [anon_sym_to] = ACTIONS(1777),
    [anon_sym_downto] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1777),
    [anon_sym_LBRACE_LT] = ACTIONS(1779),
    [anon_sym_GT_RBRACE] = ACTIONS(1779),
    [anon_sym_begin] = ACTIONS(1777),
    [sym_ocamlyacc_value] = ACTIONS(1779),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1777),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1779),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1777),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1777),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1779),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1779),
    [aux_sym_number_token1] = ACTIONS(1779),
    [aux_sym_signed_number_token1] = ACTIONS(1777),
    [anon_sym_SQUOTE] = ACTIONS(1779),
    [anon_sym_DQUOTE] = ACTIONS(1779),
    [sym_prefix_operator] = ACTIONS(1779),
    [aux_sym_sign_operator_token1] = ACTIONS(1777),
    [sym_hash_operator] = ACTIONS(1779),
    [anon_sym_lsl] = ACTIONS(1777),
    [anon_sym_lsr] = ACTIONS(1777),
    [anon_sym_asr] = ACTIONS(1777),
    [aux_sym_pow_operator_token1] = ACTIONS(1779),
    [anon_sym_mod] = ACTIONS(1777),
    [anon_sym_land] = ACTIONS(1777),
    [anon_sym_lor] = ACTIONS(1777),
    [anon_sym_lxor] = ACTIONS(1777),
    [aux_sym_mult_operator_token1] = ACTIONS(1777),
    [aux_sym_add_operator_token1] = ACTIONS(1777),
    [sym_concat_operator] = ACTIONS(1779),
    [sym_rel_operator] = ACTIONS(1777),
    [anon_sym_AMP_AMP] = ACTIONS(1777),
    [anon_sym_or] = ACTIONS(1777),
    [anon_sym_PIPE_PIPE] = ACTIONS(1777),
    [sym_assign_operator] = ACTIONS(1779),
    [sym__uppercase_identifier] = ACTIONS(1779),
    [aux_sym_tag_token1] = ACTIONS(1779),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(877)] = {
    [sym_attribute] = STATE(877),
    [sym__lowercase_identifier] = ACTIONS(1781),
    [anon_sym_TILDE] = ACTIONS(1781),
    [anon_sym_QMARK] = ACTIONS(1781),
    [anon_sym_LPAREN] = ACTIONS(1783),
    [anon_sym_RPAREN] = ACTIONS(1783),
    [anon_sym_COMMA] = ACTIONS(1783),
    [anon_sym_PIPE] = ACTIONS(1781),
    [anon_sym_COLON2] = ACTIONS(1781),
    [anon_sym_DOT] = ACTIONS(1783),
    [anon_sym_DASH_GT] = ACTIONS(1781),
    [anon_sym_LBRACK] = ACTIONS(1781),
    [anon_sym_RBRACK] = ACTIONS(1783),
    [anon_sym_true] = ACTIONS(1781),
    [anon_sym_false] = ACTIONS(1781),
    [anon_sym_LBRACE] = ACTIONS(1781),
    [anon_sym_SEMI] = ACTIONS(1783),
    [anon_sym_RBRACE] = ACTIONS(1783),
    [anon_sym_constraint] = ACTIONS(1781),
    [anon_sym_val] = ACTIONS(1781),
    [anon_sym_end] = ACTIONS(1781),
    [anon_sym_with] = ACTIONS(1781),
    [anon_sym_COLON_GT] = ACTIONS(1783),
    [anon_sym_object] = ACTIONS(1781),
    [anon_sym_inherit] = ACTIONS(1781),
    [anon_sym_method] = ACTIONS(1781),
    [anon_sym_initializer] = ACTIONS(1781),
    [anon_sym_AMP] = ACTIONS(1781),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_COLON_COLON] = ACTIONS(1783),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_then] = ACTIONS(1781),
    [anon_sym_else] = ACTIONS(1781),
    [anon_sym_do] = ACTIONS(1781),
    [anon_sym_done] = ACTIONS(1781),
    [anon_sym_to] = ACTIONS(1781),
    [anon_sym_downto] = ACTIONS(1781),
    [anon_sym_new] = ACTIONS(1781),
    [anon_sym_LBRACE_LT] = ACTIONS(1783),
    [anon_sym_GT_RBRACE] = ACTIONS(1783),
    [anon_sym_begin] = ACTIONS(1781),
    [sym_ocamlyacc_value] = ACTIONS(1783),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1781),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1783),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1781),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1781),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1783),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1783),
    [aux_sym_number_token1] = ACTIONS(1783),
    [aux_sym_signed_number_token1] = ACTIONS(1781),
    [anon_sym_SQUOTE] = ACTIONS(1783),
    [anon_sym_DQUOTE] = ACTIONS(1783),
    [sym_prefix_operator] = ACTIONS(1783),
    [aux_sym_sign_operator_token1] = ACTIONS(1781),
    [sym_hash_operator] = ACTIONS(1783),
    [anon_sym_lsl] = ACTIONS(1781),
    [anon_sym_lsr] = ACTIONS(1781),
    [anon_sym_asr] = ACTIONS(1781),
    [aux_sym_pow_operator_token1] = ACTIONS(1783),
    [anon_sym_mod] = ACTIONS(1781),
    [anon_sym_land] = ACTIONS(1781),
    [anon_sym_lor] = ACTIONS(1781),
    [anon_sym_lxor] = ACTIONS(1781),
    [aux_sym_mult_operator_token1] = ACTIONS(1781),
    [aux_sym_add_operator_token1] = ACTIONS(1781),
    [sym_concat_operator] = ACTIONS(1783),
    [sym_rel_operator] = ACTIONS(1781),
    [anon_sym_AMP_AMP] = ACTIONS(1781),
    [anon_sym_or] = ACTIONS(1781),
    [anon_sym_PIPE_PIPE] = ACTIONS(1781),
    [sym_assign_operator] = ACTIONS(1783),
    [sym__uppercase_identifier] = ACTIONS(1783),
    [aux_sym_tag_token1] = ACTIONS(1783),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(878)] = {
    [sym_attribute] = STATE(878),
    [sym__lowercase_identifier] = ACTIONS(1567),
    [anon_sym_TILDE] = ACTIONS(1567),
    [anon_sym_QMARK] = ACTIONS(1567),
    [anon_sym_LPAREN] = ACTIONS(1569),
    [anon_sym_RPAREN] = ACTIONS(1569),
    [anon_sym_COMMA] = ACTIONS(1569),
    [anon_sym_PIPE] = ACTIONS(1567),
    [anon_sym_COLON2] = ACTIONS(1567),
    [anon_sym_DOT] = ACTIONS(1569),
    [anon_sym_DASH_GT] = ACTIONS(1567),
    [anon_sym_LBRACK] = ACTIONS(1567),
    [anon_sym_RBRACK] = ACTIONS(1569),
    [anon_sym_true] = ACTIONS(1567),
    [anon_sym_false] = ACTIONS(1567),
    [anon_sym_LBRACE] = ACTIONS(1567),
    [anon_sym_SEMI] = ACTIONS(1569),
    [anon_sym_RBRACE] = ACTIONS(1569),
    [anon_sym_constraint] = ACTIONS(1567),
    [anon_sym_val] = ACTIONS(1567),
    [anon_sym_end] = ACTIONS(1567),
    [anon_sym_with] = ACTIONS(1567),
    [anon_sym_COLON_GT] = ACTIONS(1569),
    [anon_sym_object] = ACTIONS(1567),
    [anon_sym_inherit] = ACTIONS(1567),
    [anon_sym_method] = ACTIONS(1567),
    [anon_sym_initializer] = ACTIONS(1567),
    [anon_sym_AMP] = ACTIONS(1567),
    [anon_sym_POUND] = ACTIONS(1567),
    [anon_sym_COLON_COLON] = ACTIONS(1569),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1569),
    [anon_sym_then] = ACTIONS(1567),
    [anon_sym_else] = ACTIONS(1567),
    [anon_sym_do] = ACTIONS(1567),
    [anon_sym_done] = ACTIONS(1567),
    [anon_sym_to] = ACTIONS(1567),
    [anon_sym_downto] = ACTIONS(1567),
    [anon_sym_new] = ACTIONS(1567),
    [anon_sym_LBRACE_LT] = ACTIONS(1569),
    [anon_sym_GT_RBRACE] = ACTIONS(1569),
    [anon_sym_begin] = ACTIONS(1567),
    [sym_ocamlyacc_value] = ACTIONS(1569),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1567),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1569),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1567),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1567),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1569),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1569),
    [aux_sym_number_token1] = ACTIONS(1569),
    [aux_sym_signed_number_token1] = ACTIONS(1567),
    [anon_sym_SQUOTE] = ACTIONS(1569),
    [anon_sym_DQUOTE] = ACTIONS(1569),
    [sym_prefix_operator] = ACTIONS(1569),
    [aux_sym_sign_operator_token1] = ACTIONS(1567),
    [sym_hash_operator] = ACTIONS(1569),
    [anon_sym_lsl] = ACTIONS(1567),
    [anon_sym_lsr] = ACTIONS(1567),
    [anon_sym_asr] = ACTIONS(1567),
    [aux_sym_pow_operator_token1] = ACTIONS(1569),
    [anon_sym_mod] = ACTIONS(1567),
    [anon_sym_land] = ACTIONS(1567),
    [anon_sym_lor] = ACTIONS(1567),
    [anon_sym_lxor] = ACTIONS(1567),
    [aux_sym_mult_operator_token1] = ACTIONS(1567),
    [aux_sym_add_operator_token1] = ACTIONS(1567),
    [sym_concat_operator] = ACTIONS(1569),
    [sym_rel_operator] = ACTIONS(1567),
    [anon_sym_AMP_AMP] = ACTIONS(1567),
    [anon_sym_or] = ACTIONS(1567),
    [anon_sym_PIPE_PIPE] = ACTIONS(1567),
    [sym_assign_operator] = ACTIONS(1569),
    [sym__uppercase_identifier] = ACTIONS(1569),
    [aux_sym_tag_token1] = ACTIONS(1569),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(879)] = {
    [sym_attribute] = STATE(879),
    [sym__lowercase_identifier] = ACTIONS(1749),
    [anon_sym_TILDE] = ACTIONS(1749),
    [anon_sym_QMARK] = ACTIONS(1749),
    [anon_sym_LPAREN] = ACTIONS(1751),
    [anon_sym_RPAREN] = ACTIONS(1751),
    [anon_sym_COMMA] = ACTIONS(1751),
    [anon_sym_PIPE] = ACTIONS(1749),
    [anon_sym_COLON2] = ACTIONS(1749),
    [anon_sym_DOT] = ACTIONS(1751),
    [anon_sym_DASH_GT] = ACTIONS(1749),
    [anon_sym_LBRACK] = ACTIONS(1749),
    [anon_sym_RBRACK] = ACTIONS(1751),
    [anon_sym_true] = ACTIONS(1749),
    [anon_sym_false] = ACTIONS(1749),
    [anon_sym_LBRACE] = ACTIONS(1749),
    [anon_sym_SEMI] = ACTIONS(1751),
    [anon_sym_RBRACE] = ACTIONS(1751),
    [anon_sym_constraint] = ACTIONS(1749),
    [anon_sym_val] = ACTIONS(1749),
    [anon_sym_end] = ACTIONS(1749),
    [anon_sym_with] = ACTIONS(1749),
    [anon_sym_COLON_GT] = ACTIONS(1751),
    [anon_sym_object] = ACTIONS(1749),
    [anon_sym_inherit] = ACTIONS(1749),
    [anon_sym_method] = ACTIONS(1749),
    [anon_sym_initializer] = ACTIONS(1749),
    [anon_sym_AMP] = ACTIONS(1749),
    [anon_sym_POUND] = ACTIONS(1749),
    [anon_sym_COLON_COLON] = ACTIONS(1751),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1751),
    [anon_sym_then] = ACTIONS(1749),
    [anon_sym_else] = ACTIONS(1749),
    [anon_sym_do] = ACTIONS(1749),
    [anon_sym_done] = ACTIONS(1749),
    [anon_sym_to] = ACTIONS(1749),
    [anon_sym_downto] = ACTIONS(1749),
    [anon_sym_new] = ACTIONS(1749),
    [anon_sym_LBRACE_LT] = ACTIONS(1751),
    [anon_sym_GT_RBRACE] = ACTIONS(1751),
    [anon_sym_begin] = ACTIONS(1749),
    [sym_ocamlyacc_value] = ACTIONS(1751),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1749),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1751),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1749),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1749),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1751),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1751),
    [aux_sym_number_token1] = ACTIONS(1751),
    [aux_sym_signed_number_token1] = ACTIONS(1749),
    [anon_sym_SQUOTE] = ACTIONS(1751),
    [anon_sym_DQUOTE] = ACTIONS(1751),
    [sym_prefix_operator] = ACTIONS(1751),
    [aux_sym_sign_operator_token1] = ACTIONS(1749),
    [sym_hash_operator] = ACTIONS(1751),
    [anon_sym_lsl] = ACTIONS(1749),
    [anon_sym_lsr] = ACTIONS(1749),
    [anon_sym_asr] = ACTIONS(1749),
    [aux_sym_pow_operator_token1] = ACTIONS(1751),
    [anon_sym_mod] = ACTIONS(1749),
    [anon_sym_land] = ACTIONS(1749),
    [anon_sym_lor] = ACTIONS(1749),
    [anon_sym_lxor] = ACTIONS(1749),
    [aux_sym_mult_operator_token1] = ACTIONS(1749),
    [aux_sym_add_operator_token1] = ACTIONS(1749),
    [sym_concat_operator] = ACTIONS(1751),
    [sym_rel_operator] = ACTIONS(1749),
    [anon_sym_AMP_AMP] = ACTIONS(1749),
    [anon_sym_or] = ACTIONS(1749),
    [anon_sym_PIPE_PIPE] = ACTIONS(1749),
    [sym_assign_operator] = ACTIONS(1751),
    [sym__uppercase_identifier] = ACTIONS(1751),
    [aux_sym_tag_token1] = ACTIONS(1751),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(880)] = {
    [sym_attribute] = STATE(880),
    [sym__lowercase_identifier] = ACTIONS(1753),
    [anon_sym_TILDE] = ACTIONS(1753),
    [anon_sym_QMARK] = ACTIONS(1753),
    [anon_sym_LPAREN] = ACTIONS(1755),
    [anon_sym_RPAREN] = ACTIONS(1755),
    [anon_sym_COMMA] = ACTIONS(1755),
    [anon_sym_PIPE] = ACTIONS(1753),
    [anon_sym_COLON2] = ACTIONS(1753),
    [anon_sym_DOT] = ACTIONS(1755),
    [anon_sym_DASH_GT] = ACTIONS(1753),
    [anon_sym_LBRACK] = ACTIONS(1753),
    [anon_sym_RBRACK] = ACTIONS(1755),
    [anon_sym_true] = ACTIONS(1753),
    [anon_sym_false] = ACTIONS(1753),
    [anon_sym_LBRACE] = ACTIONS(1753),
    [anon_sym_SEMI] = ACTIONS(1755),
    [anon_sym_RBRACE] = ACTIONS(1755),
    [anon_sym_constraint] = ACTIONS(1753),
    [anon_sym_val] = ACTIONS(1753),
    [anon_sym_end] = ACTIONS(1753),
    [anon_sym_with] = ACTIONS(1753),
    [anon_sym_COLON_GT] = ACTIONS(1755),
    [anon_sym_object] = ACTIONS(1753),
    [anon_sym_inherit] = ACTIONS(1753),
    [anon_sym_method] = ACTIONS(1753),
    [anon_sym_initializer] = ACTIONS(1753),
    [anon_sym_AMP] = ACTIONS(1753),
    [anon_sym_POUND] = ACTIONS(1753),
    [anon_sym_COLON_COLON] = ACTIONS(1755),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1755),
    [anon_sym_then] = ACTIONS(1753),
    [anon_sym_else] = ACTIONS(1753),
    [anon_sym_do] = ACTIONS(1753),
    [anon_sym_done] = ACTIONS(1753),
    [anon_sym_to] = ACTIONS(1753),
    [anon_sym_downto] = ACTIONS(1753),
    [anon_sym_new] = ACTIONS(1753),
    [anon_sym_LBRACE_LT] = ACTIONS(1755),
    [anon_sym_GT_RBRACE] = ACTIONS(1755),
    [anon_sym_begin] = ACTIONS(1753),
    [sym_ocamlyacc_value] = ACTIONS(1755),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1753),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1755),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1753),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1753),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1755),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1755),
    [aux_sym_number_token1] = ACTIONS(1755),
    [aux_sym_signed_number_token1] = ACTIONS(1753),
    [anon_sym_SQUOTE] = ACTIONS(1755),
    [anon_sym_DQUOTE] = ACTIONS(1755),
    [sym_prefix_operator] = ACTIONS(1755),
    [aux_sym_sign_operator_token1] = ACTIONS(1753),
    [sym_hash_operator] = ACTIONS(1755),
    [anon_sym_lsl] = ACTIONS(1753),
    [anon_sym_lsr] = ACTIONS(1753),
    [anon_sym_asr] = ACTIONS(1753),
    [aux_sym_pow_operator_token1] = ACTIONS(1755),
    [anon_sym_mod] = ACTIONS(1753),
    [anon_sym_land] = ACTIONS(1753),
    [anon_sym_lor] = ACTIONS(1753),
    [anon_sym_lxor] = ACTIONS(1753),
    [aux_sym_mult_operator_token1] = ACTIONS(1753),
    [aux_sym_add_operator_token1] = ACTIONS(1753),
    [sym_concat_operator] = ACTIONS(1755),
    [sym_rel_operator] = ACTIONS(1753),
    [anon_sym_AMP_AMP] = ACTIONS(1753),
    [anon_sym_or] = ACTIONS(1753),
    [anon_sym_PIPE_PIPE] = ACTIONS(1753),
    [sym_assign_operator] = ACTIONS(1755),
    [sym__uppercase_identifier] = ACTIONS(1755),
    [aux_sym_tag_token1] = ACTIONS(1755),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(881)] = {
    [sym_attribute] = STATE(881),
    [sym__lowercase_identifier] = ACTIONS(1761),
    [anon_sym_TILDE] = ACTIONS(1761),
    [anon_sym_QMARK] = ACTIONS(1761),
    [anon_sym_LPAREN] = ACTIONS(1763),
    [anon_sym_RPAREN] = ACTIONS(1763),
    [anon_sym_COMMA] = ACTIONS(1763),
    [anon_sym_PIPE] = ACTIONS(1761),
    [anon_sym_COLON2] = ACTIONS(1761),
    [anon_sym_DOT] = ACTIONS(1763),
    [anon_sym_DASH_GT] = ACTIONS(1761),
    [anon_sym_LBRACK] = ACTIONS(1761),
    [anon_sym_RBRACK] = ACTIONS(1763),
    [anon_sym_true] = ACTIONS(1761),
    [anon_sym_false] = ACTIONS(1761),
    [anon_sym_LBRACE] = ACTIONS(1761),
    [anon_sym_SEMI] = ACTIONS(1763),
    [anon_sym_RBRACE] = ACTIONS(1763),
    [anon_sym_constraint] = ACTIONS(1761),
    [anon_sym_val] = ACTIONS(1761),
    [anon_sym_end] = ACTIONS(1761),
    [anon_sym_with] = ACTIONS(1761),
    [anon_sym_COLON_GT] = ACTIONS(1763),
    [anon_sym_object] = ACTIONS(1761),
    [anon_sym_inherit] = ACTIONS(1761),
    [anon_sym_method] = ACTIONS(1761),
    [anon_sym_initializer] = ACTIONS(1761),
    [anon_sym_AMP] = ACTIONS(1761),
    [anon_sym_POUND] = ACTIONS(1761),
    [anon_sym_COLON_COLON] = ACTIONS(1763),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1763),
    [anon_sym_then] = ACTIONS(1761),
    [anon_sym_else] = ACTIONS(1761),
    [anon_sym_do] = ACTIONS(1761),
    [anon_sym_done] = ACTIONS(1761),
    [anon_sym_to] = ACTIONS(1761),
    [anon_sym_downto] = ACTIONS(1761),
    [anon_sym_new] = ACTIONS(1761),
    [anon_sym_LBRACE_LT] = ACTIONS(1763),
    [anon_sym_GT_RBRACE] = ACTIONS(1763),
    [anon_sym_begin] = ACTIONS(1761),
    [sym_ocamlyacc_value] = ACTIONS(1763),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1761),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1763),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1761),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1761),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1763),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1763),
    [aux_sym_number_token1] = ACTIONS(1763),
    [aux_sym_signed_number_token1] = ACTIONS(1761),
    [anon_sym_SQUOTE] = ACTIONS(1763),
    [anon_sym_DQUOTE] = ACTIONS(1763),
    [sym_prefix_operator] = ACTIONS(1763),
    [aux_sym_sign_operator_token1] = ACTIONS(1761),
    [sym_hash_operator] = ACTIONS(1763),
    [anon_sym_lsl] = ACTIONS(1761),
    [anon_sym_lsr] = ACTIONS(1761),
    [anon_sym_asr] = ACTIONS(1761),
    [aux_sym_pow_operator_token1] = ACTIONS(1763),
    [anon_sym_mod] = ACTIONS(1761),
    [anon_sym_land] = ACTIONS(1761),
    [anon_sym_lor] = ACTIONS(1761),
    [anon_sym_lxor] = ACTIONS(1761),
    [aux_sym_mult_operator_token1] = ACTIONS(1761),
    [aux_sym_add_operator_token1] = ACTIONS(1761),
    [sym_concat_operator] = ACTIONS(1763),
    [sym_rel_operator] = ACTIONS(1761),
    [anon_sym_AMP_AMP] = ACTIONS(1761),
    [anon_sym_or] = ACTIONS(1761),
    [anon_sym_PIPE_PIPE] = ACTIONS(1761),
    [sym_assign_operator] = ACTIONS(1763),
    [sym__uppercase_identifier] = ACTIONS(1763),
    [aux_sym_tag_token1] = ACTIONS(1763),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(882)] = {
    [sym_attribute] = STATE(882),
    [sym__lowercase_identifier] = ACTIONS(1793),
    [anon_sym_TILDE] = ACTIONS(1793),
    [anon_sym_QMARK] = ACTIONS(1793),
    [anon_sym_LPAREN] = ACTIONS(1795),
    [anon_sym_RPAREN] = ACTIONS(1795),
    [anon_sym_COMMA] = ACTIONS(1795),
    [anon_sym_PIPE] = ACTIONS(1793),
    [anon_sym_COLON2] = ACTIONS(1793),
    [anon_sym_DOT] = ACTIONS(1795),
    [anon_sym_DASH_GT] = ACTIONS(1793),
    [anon_sym_LBRACK] = ACTIONS(1793),
    [anon_sym_RBRACK] = ACTIONS(1795),
    [anon_sym_true] = ACTIONS(1793),
    [anon_sym_false] = ACTIONS(1793),
    [anon_sym_LBRACE] = ACTIONS(1793),
    [anon_sym_SEMI] = ACTIONS(1795),
    [anon_sym_RBRACE] = ACTIONS(1795),
    [anon_sym_constraint] = ACTIONS(1793),
    [anon_sym_val] = ACTIONS(1793),
    [anon_sym_end] = ACTIONS(1793),
    [anon_sym_with] = ACTIONS(1793),
    [anon_sym_COLON_GT] = ACTIONS(1795),
    [anon_sym_object] = ACTIONS(1793),
    [anon_sym_inherit] = ACTIONS(1793),
    [anon_sym_method] = ACTIONS(1793),
    [anon_sym_initializer] = ACTIONS(1793),
    [anon_sym_AMP] = ACTIONS(1793),
    [anon_sym_POUND] = ACTIONS(1793),
    [anon_sym_COLON_COLON] = ACTIONS(1795),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1795),
    [anon_sym_then] = ACTIONS(1793),
    [anon_sym_else] = ACTIONS(1793),
    [anon_sym_do] = ACTIONS(1793),
    [anon_sym_done] = ACTIONS(1793),
    [anon_sym_to] = ACTIONS(1793),
    [anon_sym_downto] = ACTIONS(1793),
    [anon_sym_new] = ACTIONS(1793),
    [anon_sym_LBRACE_LT] = ACTIONS(1795),
    [anon_sym_GT_RBRACE] = ACTIONS(1795),
    [anon_sym_begin] = ACTIONS(1793),
    [sym_ocamlyacc_value] = ACTIONS(1795),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1793),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1795),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1795),
    [aux_sym_number_token1] = ACTIONS(1795),
    [aux_sym_signed_number_token1] = ACTIONS(1793),
    [anon_sym_SQUOTE] = ACTIONS(1795),
    [anon_sym_DQUOTE] = ACTIONS(1795),
    [sym_prefix_operator] = ACTIONS(1795),
    [aux_sym_sign_operator_token1] = ACTIONS(1793),
    [sym_hash_operator] = ACTIONS(1795),
    [anon_sym_lsl] = ACTIONS(1793),
    [anon_sym_lsr] = ACTIONS(1793),
    [anon_sym_asr] = ACTIONS(1793),
    [aux_sym_pow_operator_token1] = ACTIONS(1795),
    [anon_sym_mod] = ACTIONS(1793),
    [anon_sym_land] = ACTIONS(1793),
    [anon_sym_lor] = ACTIONS(1793),
    [anon_sym_lxor] = ACTIONS(1793),
    [aux_sym_mult_operator_token1] = ACTIONS(1793),
    [aux_sym_add_operator_token1] = ACTIONS(1793),
    [sym_concat_operator] = ACTIONS(1795),
    [sym_rel_operator] = ACTIONS(1793),
    [anon_sym_AMP_AMP] = ACTIONS(1793),
    [anon_sym_or] = ACTIONS(1793),
    [anon_sym_PIPE_PIPE] = ACTIONS(1793),
    [sym_assign_operator] = ACTIONS(1795),
    [sym__uppercase_identifier] = ACTIONS(1795),
    [aux_sym_tag_token1] = ACTIONS(1795),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(883)] = {
    [sym_attribute] = STATE(883),
    [sym__lowercase_identifier] = ACTIONS(1847),
    [anon_sym_TILDE] = ACTIONS(1847),
    [anon_sym_QMARK] = ACTIONS(1847),
    [anon_sym_LPAREN] = ACTIONS(1849),
    [anon_sym_RPAREN] = ACTIONS(1849),
    [anon_sym_COMMA] = ACTIONS(1849),
    [anon_sym_PIPE] = ACTIONS(1847),
    [anon_sym_COLON2] = ACTIONS(1847),
    [anon_sym_DOT] = ACTIONS(1849),
    [anon_sym_DASH_GT] = ACTIONS(1847),
    [anon_sym_LBRACK] = ACTIONS(1847),
    [anon_sym_RBRACK] = ACTIONS(1849),
    [anon_sym_true] = ACTIONS(1847),
    [anon_sym_false] = ACTIONS(1847),
    [anon_sym_LBRACE] = ACTIONS(1847),
    [anon_sym_SEMI] = ACTIONS(1849),
    [anon_sym_RBRACE] = ACTIONS(1849),
    [anon_sym_constraint] = ACTIONS(1847),
    [anon_sym_val] = ACTIONS(1847),
    [anon_sym_end] = ACTIONS(1847),
    [anon_sym_with] = ACTIONS(1847),
    [anon_sym_COLON_GT] = ACTIONS(1849),
    [anon_sym_object] = ACTIONS(1847),
    [anon_sym_inherit] = ACTIONS(1847),
    [anon_sym_method] = ACTIONS(1847),
    [anon_sym_initializer] = ACTIONS(1847),
    [anon_sym_AMP] = ACTIONS(1847),
    [anon_sym_POUND] = ACTIONS(1847),
    [anon_sym_COLON_COLON] = ACTIONS(1849),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_then] = ACTIONS(1847),
    [anon_sym_else] = ACTIONS(1847),
    [anon_sym_do] = ACTIONS(1847),
    [anon_sym_done] = ACTIONS(1847),
    [anon_sym_to] = ACTIONS(1847),
    [anon_sym_downto] = ACTIONS(1847),
    [anon_sym_new] = ACTIONS(1847),
    [anon_sym_LBRACE_LT] = ACTIONS(1849),
    [anon_sym_GT_RBRACE] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1847),
    [sym_ocamlyacc_value] = ACTIONS(1849),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1847),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1849),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1847),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1847),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1849),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1849),
    [aux_sym_number_token1] = ACTIONS(1849),
    [aux_sym_signed_number_token1] = ACTIONS(1847),
    [anon_sym_SQUOTE] = ACTIONS(1849),
    [anon_sym_DQUOTE] = ACTIONS(1849),
    [sym_prefix_operator] = ACTIONS(1849),
    [aux_sym_sign_operator_token1] = ACTIONS(1847),
    [sym_hash_operator] = ACTIONS(1849),
    [anon_sym_lsl] = ACTIONS(1847),
    [anon_sym_lsr] = ACTIONS(1847),
    [anon_sym_asr] = ACTIONS(1847),
    [aux_sym_pow_operator_token1] = ACTIONS(1849),
    [anon_sym_mod] = ACTIONS(1847),
    [anon_sym_land] = ACTIONS(1847),
    [anon_sym_lor] = ACTIONS(1847),
    [anon_sym_lxor] = ACTIONS(1847),
    [aux_sym_mult_operator_token1] = ACTIONS(1847),
    [aux_sym_add_operator_token1] = ACTIONS(1847),
    [sym_concat_operator] = ACTIONS(1849),
    [sym_rel_operator] = ACTIONS(1847),
    [anon_sym_AMP_AMP] = ACTIONS(1847),
    [anon_sym_or] = ACTIONS(1847),
    [anon_sym_PIPE_PIPE] = ACTIONS(1847),
    [sym_assign_operator] = ACTIONS(1849),
    [sym__uppercase_identifier] = ACTIONS(1849),
    [aux_sym_tag_token1] = ACTIONS(1849),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(884)] = {
    [sym_attribute] = STATE(884),
    [sym__lowercase_identifier] = ACTIONS(1745),
    [anon_sym_TILDE] = ACTIONS(1745),
    [anon_sym_QMARK] = ACTIONS(1745),
    [anon_sym_LPAREN] = ACTIONS(1747),
    [anon_sym_RPAREN] = ACTIONS(1747),
    [anon_sym_COMMA] = ACTIONS(1747),
    [anon_sym_PIPE] = ACTIONS(1745),
    [anon_sym_COLON2] = ACTIONS(1745),
    [anon_sym_DOT] = ACTIONS(1747),
    [anon_sym_DASH_GT] = ACTIONS(1745),
    [anon_sym_LBRACK] = ACTIONS(1745),
    [anon_sym_RBRACK] = ACTIONS(1747),
    [anon_sym_true] = ACTIONS(1745),
    [anon_sym_false] = ACTIONS(1745),
    [anon_sym_LBRACE] = ACTIONS(1745),
    [anon_sym_SEMI] = ACTIONS(1747),
    [anon_sym_RBRACE] = ACTIONS(1747),
    [anon_sym_constraint] = ACTIONS(1745),
    [anon_sym_val] = ACTIONS(1745),
    [anon_sym_end] = ACTIONS(1745),
    [anon_sym_with] = ACTIONS(1745),
    [anon_sym_COLON_GT] = ACTIONS(1747),
    [anon_sym_object] = ACTIONS(1745),
    [anon_sym_inherit] = ACTIONS(1745),
    [anon_sym_method] = ACTIONS(1745),
    [anon_sym_initializer] = ACTIONS(1745),
    [anon_sym_AMP] = ACTIONS(1745),
    [anon_sym_POUND] = ACTIONS(1745),
    [anon_sym_COLON_COLON] = ACTIONS(1747),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1747),
    [anon_sym_then] = ACTIONS(1745),
    [anon_sym_else] = ACTIONS(1745),
    [anon_sym_do] = ACTIONS(1745),
    [anon_sym_done] = ACTIONS(1745),
    [anon_sym_to] = ACTIONS(1745),
    [anon_sym_downto] = ACTIONS(1745),
    [anon_sym_new] = ACTIONS(1745),
    [anon_sym_LBRACE_LT] = ACTIONS(1747),
    [anon_sym_GT_RBRACE] = ACTIONS(1747),
    [anon_sym_begin] = ACTIONS(1745),
    [sym_ocamlyacc_value] = ACTIONS(1747),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1745),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1747),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1745),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1745),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1747),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1747),
    [aux_sym_number_token1] = ACTIONS(1747),
    [aux_sym_signed_number_token1] = ACTIONS(1745),
    [anon_sym_SQUOTE] = ACTIONS(1747),
    [anon_sym_DQUOTE] = ACTIONS(1747),
    [sym_prefix_operator] = ACTIONS(1747),
    [aux_sym_sign_operator_token1] = ACTIONS(1745),
    [sym_hash_operator] = ACTIONS(1747),
    [anon_sym_lsl] = ACTIONS(1745),
    [anon_sym_lsr] = ACTIONS(1745),
    [anon_sym_asr] = ACTIONS(1745),
    [aux_sym_pow_operator_token1] = ACTIONS(1747),
    [anon_sym_mod] = ACTIONS(1745),
    [anon_sym_land] = ACTIONS(1745),
    [anon_sym_lor] = ACTIONS(1745),
    [anon_sym_lxor] = ACTIONS(1745),
    [aux_sym_mult_operator_token1] = ACTIONS(1745),
    [aux_sym_add_operator_token1] = ACTIONS(1745),
    [sym_concat_operator] = ACTIONS(1747),
    [sym_rel_operator] = ACTIONS(1745),
    [anon_sym_AMP_AMP] = ACTIONS(1745),
    [anon_sym_or] = ACTIONS(1745),
    [anon_sym_PIPE_PIPE] = ACTIONS(1745),
    [sym_assign_operator] = ACTIONS(1747),
    [sym__uppercase_identifier] = ACTIONS(1747),
    [aux_sym_tag_token1] = ACTIONS(1747),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(885)] = {
    [sym_attribute] = STATE(885),
    [sym__lowercase_identifier] = ACTIONS(1699),
    [anon_sym_TILDE] = ACTIONS(1699),
    [anon_sym_QMARK] = ACTIONS(1699),
    [anon_sym_LPAREN] = ACTIONS(1701),
    [anon_sym_RPAREN] = ACTIONS(1701),
    [anon_sym_COMMA] = ACTIONS(1701),
    [anon_sym_PIPE] = ACTIONS(1699),
    [anon_sym_COLON2] = ACTIONS(1699),
    [anon_sym_DOT] = ACTIONS(1703),
    [anon_sym_DASH_GT] = ACTIONS(1699),
    [anon_sym_LBRACK] = ACTIONS(1699),
    [anon_sym_RBRACK] = ACTIONS(1701),
    [anon_sym_true] = ACTIONS(1699),
    [anon_sym_false] = ACTIONS(1699),
    [anon_sym_LBRACE] = ACTIONS(1699),
    [anon_sym_SEMI] = ACTIONS(1701),
    [anon_sym_RBRACE] = ACTIONS(1701),
    [anon_sym_constraint] = ACTIONS(1699),
    [anon_sym_val] = ACTIONS(1699),
    [anon_sym_end] = ACTIONS(1699),
    [anon_sym_with] = ACTIONS(1699),
    [anon_sym_COLON_GT] = ACTIONS(1701),
    [anon_sym_object] = ACTIONS(1699),
    [anon_sym_inherit] = ACTIONS(1699),
    [anon_sym_method] = ACTIONS(1699),
    [anon_sym_initializer] = ACTIONS(1699),
    [anon_sym_AMP] = ACTIONS(1699),
    [anon_sym_POUND] = ACTIONS(1699),
    [anon_sym_COLON_COLON] = ACTIONS(1701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1701),
    [anon_sym_then] = ACTIONS(1699),
    [anon_sym_else] = ACTIONS(1699),
    [anon_sym_do] = ACTIONS(1699),
    [anon_sym_done] = ACTIONS(1699),
    [anon_sym_to] = ACTIONS(1699),
    [anon_sym_downto] = ACTIONS(1699),
    [anon_sym_new] = ACTIONS(1699),
    [anon_sym_LBRACE_LT] = ACTIONS(1701),
    [anon_sym_GT_RBRACE] = ACTIONS(1701),
    [anon_sym_begin] = ACTIONS(1699),
    [sym_ocamlyacc_value] = ACTIONS(1701),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1699),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1701),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1699),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1699),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1701),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1701),
    [aux_sym_number_token1] = ACTIONS(1701),
    [aux_sym_signed_number_token1] = ACTIONS(1699),
    [anon_sym_SQUOTE] = ACTIONS(1701),
    [anon_sym_DQUOTE] = ACTIONS(1701),
    [sym_prefix_operator] = ACTIONS(1701),
    [aux_sym_sign_operator_token1] = ACTIONS(1699),
    [sym_hash_operator] = ACTIONS(1701),
    [anon_sym_lsl] = ACTIONS(1699),
    [anon_sym_lsr] = ACTIONS(1699),
    [anon_sym_asr] = ACTIONS(1699),
    [aux_sym_pow_operator_token1] = ACTIONS(1701),
    [anon_sym_mod] = ACTIONS(1699),
    [anon_sym_land] = ACTIONS(1699),
    [anon_sym_lor] = ACTIONS(1699),
    [anon_sym_lxor] = ACTIONS(1699),
    [aux_sym_mult_operator_token1] = ACTIONS(1699),
    [aux_sym_add_operator_token1] = ACTIONS(1699),
    [sym_concat_operator] = ACTIONS(1701),
    [sym_rel_operator] = ACTIONS(1699),
    [anon_sym_AMP_AMP] = ACTIONS(1699),
    [anon_sym_or] = ACTIONS(1699),
    [anon_sym_PIPE_PIPE] = ACTIONS(1699),
    [sym_assign_operator] = ACTIONS(1701),
    [sym__uppercase_identifier] = ACTIONS(1701),
    [aux_sym_tag_token1] = ACTIONS(1701),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(886)] = {
    [sym_attribute] = STATE(886),
    [sym__lowercase_identifier] = ACTIONS(1713),
    [anon_sym_TILDE] = ACTIONS(1713),
    [anon_sym_QMARK] = ACTIONS(1713),
    [anon_sym_LPAREN] = ACTIONS(1715),
    [anon_sym_RPAREN] = ACTIONS(1715),
    [anon_sym_COMMA] = ACTIONS(1715),
    [anon_sym_PIPE] = ACTIONS(1713),
    [anon_sym_COLON2] = ACTIONS(1713),
    [anon_sym_DOT] = ACTIONS(1715),
    [anon_sym_DASH_GT] = ACTIONS(1713),
    [anon_sym_LBRACK] = ACTIONS(1713),
    [anon_sym_RBRACK] = ACTIONS(1715),
    [anon_sym_true] = ACTIONS(1713),
    [anon_sym_false] = ACTIONS(1713),
    [anon_sym_LBRACE] = ACTIONS(1713),
    [anon_sym_SEMI] = ACTIONS(1715),
    [anon_sym_RBRACE] = ACTIONS(1715),
    [anon_sym_constraint] = ACTIONS(1713),
    [anon_sym_val] = ACTIONS(1713),
    [anon_sym_end] = ACTIONS(1713),
    [anon_sym_with] = ACTIONS(1713),
    [anon_sym_COLON_GT] = ACTIONS(1715),
    [anon_sym_object] = ACTIONS(1713),
    [anon_sym_inherit] = ACTIONS(1713),
    [anon_sym_method] = ACTIONS(1713),
    [anon_sym_initializer] = ACTIONS(1713),
    [anon_sym_AMP] = ACTIONS(1713),
    [anon_sym_POUND] = ACTIONS(1713),
    [anon_sym_COLON_COLON] = ACTIONS(1715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1715),
    [anon_sym_then] = ACTIONS(1713),
    [anon_sym_else] = ACTIONS(1713),
    [anon_sym_do] = ACTIONS(1713),
    [anon_sym_done] = ACTIONS(1713),
    [anon_sym_to] = ACTIONS(1713),
    [anon_sym_downto] = ACTIONS(1713),
    [anon_sym_new] = ACTIONS(1713),
    [anon_sym_LBRACE_LT] = ACTIONS(1715),
    [anon_sym_GT_RBRACE] = ACTIONS(1715),
    [anon_sym_begin] = ACTIONS(1713),
    [sym_ocamlyacc_value] = ACTIONS(1715),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1713),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1715),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1713),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1713),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1715),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1715),
    [aux_sym_number_token1] = ACTIONS(1715),
    [aux_sym_signed_number_token1] = ACTIONS(1713),
    [anon_sym_SQUOTE] = ACTIONS(1715),
    [anon_sym_DQUOTE] = ACTIONS(1715),
    [sym_prefix_operator] = ACTIONS(1715),
    [aux_sym_sign_operator_token1] = ACTIONS(1713),
    [sym_hash_operator] = ACTIONS(1715),
    [anon_sym_lsl] = ACTIONS(1713),
    [anon_sym_lsr] = ACTIONS(1713),
    [anon_sym_asr] = ACTIONS(1713),
    [aux_sym_pow_operator_token1] = ACTIONS(1715),
    [anon_sym_mod] = ACTIONS(1713),
    [anon_sym_land] = ACTIONS(1713),
    [anon_sym_lor] = ACTIONS(1713),
    [anon_sym_lxor] = ACTIONS(1713),
    [aux_sym_mult_operator_token1] = ACTIONS(1713),
    [aux_sym_add_operator_token1] = ACTIONS(1713),
    [sym_concat_operator] = ACTIONS(1715),
    [sym_rel_operator] = ACTIONS(1713),
    [anon_sym_AMP_AMP] = ACTIONS(1713),
    [anon_sym_or] = ACTIONS(1713),
    [anon_sym_PIPE_PIPE] = ACTIONS(1713),
    [sym_assign_operator] = ACTIONS(1715),
    [sym__uppercase_identifier] = ACTIONS(1715),
    [aux_sym_tag_token1] = ACTIONS(1715),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(887)] = {
    [sym_attribute] = STATE(887),
    [sym__lowercase_identifier] = ACTIONS(1721),
    [anon_sym_TILDE] = ACTIONS(1721),
    [anon_sym_QMARK] = ACTIONS(1721),
    [anon_sym_LPAREN] = ACTIONS(1723),
    [anon_sym_RPAREN] = ACTIONS(1723),
    [anon_sym_COMMA] = ACTIONS(1723),
    [anon_sym_PIPE] = ACTIONS(1721),
    [anon_sym_COLON2] = ACTIONS(1721),
    [anon_sym_DOT] = ACTIONS(1723),
    [anon_sym_DASH_GT] = ACTIONS(1721),
    [anon_sym_LBRACK] = ACTIONS(1721),
    [anon_sym_RBRACK] = ACTIONS(1723),
    [anon_sym_true] = ACTIONS(1721),
    [anon_sym_false] = ACTIONS(1721),
    [anon_sym_LBRACE] = ACTIONS(1721),
    [anon_sym_SEMI] = ACTIONS(1723),
    [anon_sym_RBRACE] = ACTIONS(1723),
    [anon_sym_constraint] = ACTIONS(1721),
    [anon_sym_val] = ACTIONS(1721),
    [anon_sym_end] = ACTIONS(1721),
    [anon_sym_with] = ACTIONS(1721),
    [anon_sym_COLON_GT] = ACTIONS(1723),
    [anon_sym_object] = ACTIONS(1721),
    [anon_sym_inherit] = ACTIONS(1721),
    [anon_sym_method] = ACTIONS(1721),
    [anon_sym_initializer] = ACTIONS(1721),
    [anon_sym_AMP] = ACTIONS(1721),
    [anon_sym_POUND] = ACTIONS(1721),
    [anon_sym_COLON_COLON] = ACTIONS(1723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1723),
    [anon_sym_then] = ACTIONS(1721),
    [anon_sym_else] = ACTIONS(1721),
    [anon_sym_do] = ACTIONS(1721),
    [anon_sym_done] = ACTIONS(1721),
    [anon_sym_to] = ACTIONS(1721),
    [anon_sym_downto] = ACTIONS(1721),
    [anon_sym_new] = ACTIONS(1721),
    [anon_sym_LBRACE_LT] = ACTIONS(1723),
    [anon_sym_GT_RBRACE] = ACTIONS(1723),
    [anon_sym_begin] = ACTIONS(1721),
    [sym_ocamlyacc_value] = ACTIONS(1723),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1721),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1723),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1721),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1721),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1723),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1723),
    [aux_sym_number_token1] = ACTIONS(1723),
    [aux_sym_signed_number_token1] = ACTIONS(1721),
    [anon_sym_SQUOTE] = ACTIONS(1723),
    [anon_sym_DQUOTE] = ACTIONS(1723),
    [sym_prefix_operator] = ACTIONS(1723),
    [aux_sym_sign_operator_token1] = ACTIONS(1721),
    [sym_hash_operator] = ACTIONS(1723),
    [anon_sym_lsl] = ACTIONS(1721),
    [anon_sym_lsr] = ACTIONS(1721),
    [anon_sym_asr] = ACTIONS(1721),
    [aux_sym_pow_operator_token1] = ACTIONS(1723),
    [anon_sym_mod] = ACTIONS(1721),
    [anon_sym_land] = ACTIONS(1721),
    [anon_sym_lor] = ACTIONS(1721),
    [anon_sym_lxor] = ACTIONS(1721),
    [aux_sym_mult_operator_token1] = ACTIONS(1721),
    [aux_sym_add_operator_token1] = ACTIONS(1721),
    [sym_concat_operator] = ACTIONS(1723),
    [sym_rel_operator] = ACTIONS(1721),
    [anon_sym_AMP_AMP] = ACTIONS(1721),
    [anon_sym_or] = ACTIONS(1721),
    [anon_sym_PIPE_PIPE] = ACTIONS(1721),
    [sym_assign_operator] = ACTIONS(1723),
    [sym__uppercase_identifier] = ACTIONS(1723),
    [aux_sym_tag_token1] = ACTIONS(1723),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(888)] = {
    [sym_attribute] = STATE(888),
    [sym__lowercase_identifier] = ACTIONS(1809),
    [anon_sym_TILDE] = ACTIONS(1809),
    [anon_sym_QMARK] = ACTIONS(1809),
    [anon_sym_LPAREN] = ACTIONS(1811),
    [anon_sym_RPAREN] = ACTIONS(1811),
    [anon_sym_COMMA] = ACTIONS(1811),
    [anon_sym_PIPE] = ACTIONS(1809),
    [anon_sym_COLON2] = ACTIONS(1809),
    [anon_sym_DOT] = ACTIONS(1811),
    [anon_sym_DASH_GT] = ACTIONS(1809),
    [anon_sym_LBRACK] = ACTIONS(1809),
    [anon_sym_RBRACK] = ACTIONS(1811),
    [anon_sym_true] = ACTIONS(1809),
    [anon_sym_false] = ACTIONS(1809),
    [anon_sym_LBRACE] = ACTIONS(1809),
    [anon_sym_SEMI] = ACTIONS(1811),
    [anon_sym_RBRACE] = ACTIONS(1811),
    [anon_sym_constraint] = ACTIONS(1809),
    [anon_sym_val] = ACTIONS(1809),
    [anon_sym_end] = ACTIONS(1809),
    [anon_sym_with] = ACTIONS(1809),
    [anon_sym_COLON_GT] = ACTIONS(1811),
    [anon_sym_object] = ACTIONS(1809),
    [anon_sym_inherit] = ACTIONS(1809),
    [anon_sym_method] = ACTIONS(1809),
    [anon_sym_initializer] = ACTIONS(1809),
    [anon_sym_AMP] = ACTIONS(1809),
    [anon_sym_POUND] = ACTIONS(1809),
    [anon_sym_COLON_COLON] = ACTIONS(1811),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1811),
    [anon_sym_then] = ACTIONS(1809),
    [anon_sym_else] = ACTIONS(1809),
    [anon_sym_do] = ACTIONS(1809),
    [anon_sym_done] = ACTIONS(1809),
    [anon_sym_to] = ACTIONS(1809),
    [anon_sym_downto] = ACTIONS(1809),
    [anon_sym_new] = ACTIONS(1809),
    [anon_sym_LBRACE_LT] = ACTIONS(1811),
    [anon_sym_GT_RBRACE] = ACTIONS(1811),
    [anon_sym_begin] = ACTIONS(1809),
    [sym_ocamlyacc_value] = ACTIONS(1811),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1809),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1811),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1809),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1809),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1811),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1811),
    [aux_sym_number_token1] = ACTIONS(1811),
    [aux_sym_signed_number_token1] = ACTIONS(1809),
    [anon_sym_SQUOTE] = ACTIONS(1811),
    [anon_sym_DQUOTE] = ACTIONS(1811),
    [sym_prefix_operator] = ACTIONS(1811),
    [aux_sym_sign_operator_token1] = ACTIONS(1809),
    [sym_hash_operator] = ACTIONS(1811),
    [anon_sym_lsl] = ACTIONS(1809),
    [anon_sym_lsr] = ACTIONS(1809),
    [anon_sym_asr] = ACTIONS(1809),
    [aux_sym_pow_operator_token1] = ACTIONS(1811),
    [anon_sym_mod] = ACTIONS(1809),
    [anon_sym_land] = ACTIONS(1809),
    [anon_sym_lor] = ACTIONS(1809),
    [anon_sym_lxor] = ACTIONS(1809),
    [aux_sym_mult_operator_token1] = ACTIONS(1809),
    [aux_sym_add_operator_token1] = ACTIONS(1809),
    [sym_concat_operator] = ACTIONS(1811),
    [sym_rel_operator] = ACTIONS(1809),
    [anon_sym_AMP_AMP] = ACTIONS(1809),
    [anon_sym_or] = ACTIONS(1809),
    [anon_sym_PIPE_PIPE] = ACTIONS(1809),
    [sym_assign_operator] = ACTIONS(1811),
    [sym__uppercase_identifier] = ACTIONS(1811),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(889)] = {
    [sym_attribute] = STATE(889),
    [sym__lowercase_identifier] = ACTIONS(1825),
    [anon_sym_TILDE] = ACTIONS(1825),
    [anon_sym_QMARK] = ACTIONS(1825),
    [anon_sym_LPAREN] = ACTIONS(1827),
    [anon_sym_RPAREN] = ACTIONS(1827),
    [anon_sym_COMMA] = ACTIONS(1827),
    [anon_sym_PIPE] = ACTIONS(1825),
    [anon_sym_COLON2] = ACTIONS(1825),
    [anon_sym_DOT] = ACTIONS(1827),
    [anon_sym_DASH_GT] = ACTIONS(1825),
    [anon_sym_LBRACK] = ACTIONS(1825),
    [anon_sym_RBRACK] = ACTIONS(1827),
    [anon_sym_true] = ACTIONS(1825),
    [anon_sym_false] = ACTIONS(1825),
    [anon_sym_LBRACE] = ACTIONS(1825),
    [anon_sym_SEMI] = ACTIONS(1827),
    [anon_sym_RBRACE] = ACTIONS(1827),
    [anon_sym_constraint] = ACTIONS(1825),
    [anon_sym_val] = ACTIONS(1825),
    [anon_sym_end] = ACTIONS(1825),
    [anon_sym_with] = ACTIONS(1825),
    [anon_sym_COLON_GT] = ACTIONS(1827),
    [anon_sym_object] = ACTIONS(1825),
    [anon_sym_inherit] = ACTIONS(1825),
    [anon_sym_method] = ACTIONS(1825),
    [anon_sym_initializer] = ACTIONS(1825),
    [anon_sym_AMP] = ACTIONS(1825),
    [anon_sym_POUND] = ACTIONS(1825),
    [anon_sym_COLON_COLON] = ACTIONS(1827),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1827),
    [anon_sym_then] = ACTIONS(1825),
    [anon_sym_else] = ACTIONS(1825),
    [anon_sym_do] = ACTIONS(1825),
    [anon_sym_done] = ACTIONS(1825),
    [anon_sym_to] = ACTIONS(1825),
    [anon_sym_downto] = ACTIONS(1825),
    [anon_sym_new] = ACTIONS(1825),
    [anon_sym_LBRACE_LT] = ACTIONS(1827),
    [anon_sym_GT_RBRACE] = ACTIONS(1827),
    [anon_sym_begin] = ACTIONS(1825),
    [sym_ocamlyacc_value] = ACTIONS(1827),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1825),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1827),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1825),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1825),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1827),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1827),
    [aux_sym_number_token1] = ACTIONS(1827),
    [aux_sym_signed_number_token1] = ACTIONS(1825),
    [anon_sym_SQUOTE] = ACTIONS(1827),
    [anon_sym_DQUOTE] = ACTIONS(1827),
    [sym_prefix_operator] = ACTIONS(1827),
    [aux_sym_sign_operator_token1] = ACTIONS(1825),
    [sym_hash_operator] = ACTIONS(1827),
    [anon_sym_lsl] = ACTIONS(1825),
    [anon_sym_lsr] = ACTIONS(1825),
    [anon_sym_asr] = ACTIONS(1825),
    [aux_sym_pow_operator_token1] = ACTIONS(1827),
    [anon_sym_mod] = ACTIONS(1825),
    [anon_sym_land] = ACTIONS(1825),
    [anon_sym_lor] = ACTIONS(1825),
    [anon_sym_lxor] = ACTIONS(1825),
    [aux_sym_mult_operator_token1] = ACTIONS(1825),
    [aux_sym_add_operator_token1] = ACTIONS(1825),
    [sym_concat_operator] = ACTIONS(1827),
    [sym_rel_operator] = ACTIONS(1825),
    [anon_sym_AMP_AMP] = ACTIONS(1825),
    [anon_sym_or] = ACTIONS(1825),
    [anon_sym_PIPE_PIPE] = ACTIONS(1825),
    [sym_assign_operator] = ACTIONS(1827),
    [sym__uppercase_identifier] = ACTIONS(1827),
    [aux_sym_tag_token1] = ACTIONS(1827),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(890)] = {
    [sym_attribute] = STATE(890),
    [sym__lowercase_identifier] = ACTIONS(1699),
    [anon_sym_TILDE] = ACTIONS(1699),
    [anon_sym_QMARK] = ACTIONS(1699),
    [anon_sym_LPAREN] = ACTIONS(1701),
    [anon_sym_RPAREN] = ACTIONS(1701),
    [anon_sym_COMMA] = ACTIONS(1701),
    [anon_sym_PIPE] = ACTIONS(1699),
    [anon_sym_COLON2] = ACTIONS(1699),
    [anon_sym_DOT] = ACTIONS(1841),
    [anon_sym_DASH_GT] = ACTIONS(1699),
    [anon_sym_LBRACK] = ACTIONS(1699),
    [anon_sym_RBRACK] = ACTIONS(1701),
    [anon_sym_true] = ACTIONS(1699),
    [anon_sym_false] = ACTIONS(1699),
    [anon_sym_LBRACE] = ACTIONS(1699),
    [anon_sym_SEMI] = ACTIONS(1701),
    [anon_sym_RBRACE] = ACTIONS(1701),
    [anon_sym_constraint] = ACTIONS(1699),
    [anon_sym_val] = ACTIONS(1699),
    [anon_sym_end] = ACTIONS(1699),
    [anon_sym_with] = ACTIONS(1699),
    [anon_sym_COLON_GT] = ACTIONS(1701),
    [anon_sym_object] = ACTIONS(1699),
    [anon_sym_inherit] = ACTIONS(1699),
    [anon_sym_method] = ACTIONS(1699),
    [anon_sym_initializer] = ACTIONS(1699),
    [anon_sym_AMP] = ACTIONS(1699),
    [anon_sym_POUND] = ACTIONS(1699),
    [anon_sym_COLON_COLON] = ACTIONS(1701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1701),
    [anon_sym_then] = ACTIONS(1699),
    [anon_sym_else] = ACTIONS(1699),
    [anon_sym_do] = ACTIONS(1699),
    [anon_sym_done] = ACTIONS(1699),
    [anon_sym_to] = ACTIONS(1699),
    [anon_sym_downto] = ACTIONS(1699),
    [anon_sym_new] = ACTIONS(1699),
    [anon_sym_LBRACE_LT] = ACTIONS(1701),
    [anon_sym_GT_RBRACE] = ACTIONS(1701),
    [anon_sym_begin] = ACTIONS(1699),
    [sym_ocamlyacc_value] = ACTIONS(1701),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1699),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1701),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1699),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1699),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1701),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1701),
    [aux_sym_number_token1] = ACTIONS(1701),
    [aux_sym_signed_number_token1] = ACTIONS(1699),
    [anon_sym_SQUOTE] = ACTIONS(1701),
    [anon_sym_DQUOTE] = ACTIONS(1701),
    [sym_prefix_operator] = ACTIONS(1701),
    [aux_sym_sign_operator_token1] = ACTIONS(1699),
    [sym_hash_operator] = ACTIONS(1701),
    [anon_sym_lsl] = ACTIONS(1699),
    [anon_sym_lsr] = ACTIONS(1699),
    [anon_sym_asr] = ACTIONS(1699),
    [aux_sym_pow_operator_token1] = ACTIONS(1701),
    [anon_sym_mod] = ACTIONS(1699),
    [anon_sym_land] = ACTIONS(1699),
    [anon_sym_lor] = ACTIONS(1699),
    [anon_sym_lxor] = ACTIONS(1699),
    [aux_sym_mult_operator_token1] = ACTIONS(1699),
    [aux_sym_add_operator_token1] = ACTIONS(1699),
    [sym_concat_operator] = ACTIONS(1701),
    [sym_rel_operator] = ACTIONS(1699),
    [anon_sym_AMP_AMP] = ACTIONS(1699),
    [anon_sym_or] = ACTIONS(1699),
    [anon_sym_PIPE_PIPE] = ACTIONS(1699),
    [sym_assign_operator] = ACTIONS(1701),
    [sym__uppercase_identifier] = ACTIONS(1701),
    [aux_sym_tag_token1] = ACTIONS(1701),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(891)] = {
    [sym_attribute] = STATE(891),
    [sym__lowercase_identifier] = ACTIONS(1843),
    [anon_sym_TILDE] = ACTIONS(1843),
    [anon_sym_QMARK] = ACTIONS(1843),
    [anon_sym_LPAREN] = ACTIONS(1845),
    [anon_sym_RPAREN] = ACTIONS(1845),
    [anon_sym_COMMA] = ACTIONS(1845),
    [anon_sym_PIPE] = ACTIONS(1843),
    [anon_sym_COLON2] = ACTIONS(1843),
    [anon_sym_DOT] = ACTIONS(1845),
    [anon_sym_DASH_GT] = ACTIONS(1843),
    [anon_sym_LBRACK] = ACTIONS(1843),
    [anon_sym_RBRACK] = ACTIONS(1845),
    [anon_sym_true] = ACTIONS(1843),
    [anon_sym_false] = ACTIONS(1843),
    [anon_sym_LBRACE] = ACTIONS(1843),
    [anon_sym_SEMI] = ACTIONS(1845),
    [anon_sym_RBRACE] = ACTIONS(1845),
    [anon_sym_constraint] = ACTIONS(1843),
    [anon_sym_val] = ACTIONS(1843),
    [anon_sym_end] = ACTIONS(1843),
    [anon_sym_with] = ACTIONS(1843),
    [anon_sym_COLON_GT] = ACTIONS(1845),
    [anon_sym_object] = ACTIONS(1843),
    [anon_sym_inherit] = ACTIONS(1843),
    [anon_sym_method] = ACTIONS(1843),
    [anon_sym_initializer] = ACTIONS(1843),
    [anon_sym_AMP] = ACTIONS(1843),
    [anon_sym_POUND] = ACTIONS(1843),
    [anon_sym_COLON_COLON] = ACTIONS(1845),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1845),
    [anon_sym_then] = ACTIONS(1843),
    [anon_sym_else] = ACTIONS(1843),
    [anon_sym_do] = ACTIONS(1843),
    [anon_sym_done] = ACTIONS(1843),
    [anon_sym_to] = ACTIONS(1843),
    [anon_sym_downto] = ACTIONS(1843),
    [anon_sym_new] = ACTIONS(1843),
    [anon_sym_LBRACE_LT] = ACTIONS(1845),
    [anon_sym_GT_RBRACE] = ACTIONS(1845),
    [anon_sym_begin] = ACTIONS(1843),
    [sym_ocamlyacc_value] = ACTIONS(1845),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1843),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1845),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1843),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1843),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1845),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1845),
    [aux_sym_number_token1] = ACTIONS(1845),
    [aux_sym_signed_number_token1] = ACTIONS(1843),
    [anon_sym_SQUOTE] = ACTIONS(1845),
    [anon_sym_DQUOTE] = ACTIONS(1845),
    [sym_prefix_operator] = ACTIONS(1845),
    [aux_sym_sign_operator_token1] = ACTIONS(1843),
    [sym_hash_operator] = ACTIONS(1845),
    [anon_sym_lsl] = ACTIONS(1843),
    [anon_sym_lsr] = ACTIONS(1843),
    [anon_sym_asr] = ACTIONS(1843),
    [aux_sym_pow_operator_token1] = ACTIONS(1845),
    [anon_sym_mod] = ACTIONS(1843),
    [anon_sym_land] = ACTIONS(1843),
    [anon_sym_lor] = ACTIONS(1843),
    [anon_sym_lxor] = ACTIONS(1843),
    [aux_sym_mult_operator_token1] = ACTIONS(1843),
    [aux_sym_add_operator_token1] = ACTIONS(1843),
    [sym_concat_operator] = ACTIONS(1845),
    [sym_rel_operator] = ACTIONS(1843),
    [anon_sym_AMP_AMP] = ACTIONS(1843),
    [anon_sym_or] = ACTIONS(1843),
    [anon_sym_PIPE_PIPE] = ACTIONS(1843),
    [sym_assign_operator] = ACTIONS(1845),
    [sym__uppercase_identifier] = ACTIONS(1845),
    [aux_sym_tag_token1] = ACTIONS(1845),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(892)] = {
    [sym_attribute] = STATE(892),
    [sym__lowercase_identifier] = ACTIONS(1683),
    [anon_sym_TILDE] = ACTIONS(1683),
    [anon_sym_QMARK] = ACTIONS(1683),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_RPAREN] = ACTIONS(1685),
    [anon_sym_COMMA] = ACTIONS(1685),
    [anon_sym_PIPE] = ACTIONS(1683),
    [anon_sym_COLON2] = ACTIONS(1683),
    [anon_sym_DOT] = ACTIONS(1685),
    [anon_sym_DASH_GT] = ACTIONS(1683),
    [anon_sym_LBRACK] = ACTIONS(1683),
    [anon_sym_RBRACK] = ACTIONS(1685),
    [anon_sym_true] = ACTIONS(1683),
    [anon_sym_false] = ACTIONS(1683),
    [anon_sym_LBRACE] = ACTIONS(1683),
    [anon_sym_SEMI] = ACTIONS(1685),
    [anon_sym_RBRACE] = ACTIONS(1685),
    [anon_sym_constraint] = ACTIONS(1683),
    [anon_sym_val] = ACTIONS(1683),
    [anon_sym_end] = ACTIONS(1683),
    [anon_sym_with] = ACTIONS(1683),
    [anon_sym_COLON_GT] = ACTIONS(1685),
    [anon_sym_object] = ACTIONS(1683),
    [anon_sym_inherit] = ACTIONS(1683),
    [anon_sym_method] = ACTIONS(1683),
    [anon_sym_initializer] = ACTIONS(1683),
    [anon_sym_AMP] = ACTIONS(1683),
    [anon_sym_POUND] = ACTIONS(1683),
    [anon_sym_COLON_COLON] = ACTIONS(1685),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1685),
    [anon_sym_then] = ACTIONS(1683),
    [anon_sym_else] = ACTIONS(1683),
    [anon_sym_do] = ACTIONS(1683),
    [anon_sym_done] = ACTIONS(1683),
    [anon_sym_to] = ACTIONS(1683),
    [anon_sym_downto] = ACTIONS(1683),
    [anon_sym_new] = ACTIONS(1683),
    [anon_sym_LBRACE_LT] = ACTIONS(1685),
    [anon_sym_GT_RBRACE] = ACTIONS(1685),
    [anon_sym_begin] = ACTIONS(1683),
    [sym_ocamlyacc_value] = ACTIONS(1685),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1683),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1685),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1683),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1683),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1685),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1685),
    [aux_sym_number_token1] = ACTIONS(1685),
    [aux_sym_signed_number_token1] = ACTIONS(1683),
    [anon_sym_SQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1685),
    [sym_prefix_operator] = ACTIONS(1685),
    [aux_sym_sign_operator_token1] = ACTIONS(1683),
    [sym_hash_operator] = ACTIONS(1685),
    [anon_sym_lsl] = ACTIONS(1683),
    [anon_sym_lsr] = ACTIONS(1683),
    [anon_sym_asr] = ACTIONS(1683),
    [aux_sym_pow_operator_token1] = ACTIONS(1685),
    [anon_sym_mod] = ACTIONS(1683),
    [anon_sym_land] = ACTIONS(1683),
    [anon_sym_lor] = ACTIONS(1683),
    [anon_sym_lxor] = ACTIONS(1683),
    [aux_sym_mult_operator_token1] = ACTIONS(1683),
    [aux_sym_add_operator_token1] = ACTIONS(1683),
    [sym_concat_operator] = ACTIONS(1685),
    [sym_rel_operator] = ACTIONS(1683),
    [anon_sym_AMP_AMP] = ACTIONS(1683),
    [anon_sym_or] = ACTIONS(1683),
    [anon_sym_PIPE_PIPE] = ACTIONS(1683),
    [sym_assign_operator] = ACTIONS(1685),
    [sym__uppercase_identifier] = ACTIONS(1685),
    [aux_sym_tag_token1] = ACTIONS(1685),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(893)] = {
    [sym_attribute] = STATE(893),
    [sym__lowercase_identifier] = ACTIONS(1687),
    [anon_sym_TILDE] = ACTIONS(1687),
    [anon_sym_QMARK] = ACTIONS(1687),
    [anon_sym_LPAREN] = ACTIONS(1689),
    [anon_sym_RPAREN] = ACTIONS(1689),
    [anon_sym_COMMA] = ACTIONS(1689),
    [anon_sym_PIPE] = ACTIONS(1687),
    [anon_sym_COLON2] = ACTIONS(1687),
    [anon_sym_DOT] = ACTIONS(1689),
    [anon_sym_DASH_GT] = ACTIONS(1687),
    [anon_sym_LBRACK] = ACTIONS(1687),
    [anon_sym_RBRACK] = ACTIONS(1689),
    [anon_sym_true] = ACTIONS(1687),
    [anon_sym_false] = ACTIONS(1687),
    [anon_sym_LBRACE] = ACTIONS(1687),
    [anon_sym_SEMI] = ACTIONS(1689),
    [anon_sym_RBRACE] = ACTIONS(1689),
    [anon_sym_constraint] = ACTIONS(1687),
    [anon_sym_val] = ACTIONS(1687),
    [anon_sym_end] = ACTIONS(1687),
    [anon_sym_with] = ACTIONS(1687),
    [anon_sym_COLON_GT] = ACTIONS(1689),
    [anon_sym_object] = ACTIONS(1687),
    [anon_sym_inherit] = ACTIONS(1687),
    [anon_sym_method] = ACTIONS(1687),
    [anon_sym_initializer] = ACTIONS(1687),
    [anon_sym_AMP] = ACTIONS(1687),
    [anon_sym_POUND] = ACTIONS(1687),
    [anon_sym_COLON_COLON] = ACTIONS(1689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1689),
    [anon_sym_then] = ACTIONS(1687),
    [anon_sym_else] = ACTIONS(1687),
    [anon_sym_do] = ACTIONS(1687),
    [anon_sym_done] = ACTIONS(1687),
    [anon_sym_to] = ACTIONS(1687),
    [anon_sym_downto] = ACTIONS(1687),
    [anon_sym_new] = ACTIONS(1687),
    [anon_sym_LBRACE_LT] = ACTIONS(1689),
    [anon_sym_GT_RBRACE] = ACTIONS(1689),
    [anon_sym_begin] = ACTIONS(1687),
    [sym_ocamlyacc_value] = ACTIONS(1689),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1687),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1689),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1687),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1687),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1689),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1689),
    [aux_sym_number_token1] = ACTIONS(1689),
    [aux_sym_signed_number_token1] = ACTIONS(1687),
    [anon_sym_SQUOTE] = ACTIONS(1689),
    [anon_sym_DQUOTE] = ACTIONS(1689),
    [sym_prefix_operator] = ACTIONS(1689),
    [aux_sym_sign_operator_token1] = ACTIONS(1687),
    [sym_hash_operator] = ACTIONS(1689),
    [anon_sym_lsl] = ACTIONS(1687),
    [anon_sym_lsr] = ACTIONS(1687),
    [anon_sym_asr] = ACTIONS(1687),
    [aux_sym_pow_operator_token1] = ACTIONS(1689),
    [anon_sym_mod] = ACTIONS(1687),
    [anon_sym_land] = ACTIONS(1687),
    [anon_sym_lor] = ACTIONS(1687),
    [anon_sym_lxor] = ACTIONS(1687),
    [aux_sym_mult_operator_token1] = ACTIONS(1687),
    [aux_sym_add_operator_token1] = ACTIONS(1687),
    [sym_concat_operator] = ACTIONS(1689),
    [sym_rel_operator] = ACTIONS(1687),
    [anon_sym_AMP_AMP] = ACTIONS(1687),
    [anon_sym_or] = ACTIONS(1687),
    [anon_sym_PIPE_PIPE] = ACTIONS(1687),
    [sym_assign_operator] = ACTIONS(1689),
    [sym__uppercase_identifier] = ACTIONS(1689),
    [aux_sym_tag_token1] = ACTIONS(1689),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(894)] = {
    [sym_attribute] = STATE(894),
    [sym__lowercase_identifier] = ACTIONS(1709),
    [anon_sym_TILDE] = ACTIONS(1709),
    [anon_sym_QMARK] = ACTIONS(1709),
    [anon_sym_LPAREN] = ACTIONS(1711),
    [anon_sym_RPAREN] = ACTIONS(1711),
    [anon_sym_COMMA] = ACTIONS(1711),
    [anon_sym_PIPE] = ACTIONS(1709),
    [anon_sym_COLON2] = ACTIONS(1709),
    [anon_sym_DOT] = ACTIONS(1711),
    [anon_sym_DASH_GT] = ACTIONS(1709),
    [anon_sym_LBRACK] = ACTIONS(1709),
    [anon_sym_RBRACK] = ACTIONS(1711),
    [anon_sym_true] = ACTIONS(1709),
    [anon_sym_false] = ACTIONS(1709),
    [anon_sym_LBRACE] = ACTIONS(1709),
    [anon_sym_SEMI] = ACTIONS(1711),
    [anon_sym_RBRACE] = ACTIONS(1711),
    [anon_sym_constraint] = ACTIONS(1709),
    [anon_sym_val] = ACTIONS(1709),
    [anon_sym_end] = ACTIONS(1709),
    [anon_sym_with] = ACTIONS(1709),
    [anon_sym_COLON_GT] = ACTIONS(1711),
    [anon_sym_object] = ACTIONS(1709),
    [anon_sym_inherit] = ACTIONS(1709),
    [anon_sym_method] = ACTIONS(1709),
    [anon_sym_initializer] = ACTIONS(1709),
    [anon_sym_AMP] = ACTIONS(1709),
    [anon_sym_POUND] = ACTIONS(1709),
    [anon_sym_COLON_COLON] = ACTIONS(1711),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1711),
    [anon_sym_then] = ACTIONS(1709),
    [anon_sym_else] = ACTIONS(1709),
    [anon_sym_do] = ACTIONS(1709),
    [anon_sym_done] = ACTIONS(1709),
    [anon_sym_to] = ACTIONS(1709),
    [anon_sym_downto] = ACTIONS(1709),
    [anon_sym_new] = ACTIONS(1709),
    [anon_sym_LBRACE_LT] = ACTIONS(1711),
    [anon_sym_GT_RBRACE] = ACTIONS(1711),
    [anon_sym_begin] = ACTIONS(1709),
    [sym_ocamlyacc_value] = ACTIONS(1711),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1709),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1711),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1709),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1709),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1711),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1711),
    [aux_sym_number_token1] = ACTIONS(1711),
    [aux_sym_signed_number_token1] = ACTIONS(1709),
    [anon_sym_SQUOTE] = ACTIONS(1711),
    [anon_sym_DQUOTE] = ACTIONS(1711),
    [sym_prefix_operator] = ACTIONS(1711),
    [aux_sym_sign_operator_token1] = ACTIONS(1709),
    [sym_hash_operator] = ACTIONS(1711),
    [anon_sym_lsl] = ACTIONS(1709),
    [anon_sym_lsr] = ACTIONS(1709),
    [anon_sym_asr] = ACTIONS(1709),
    [aux_sym_pow_operator_token1] = ACTIONS(1711),
    [anon_sym_mod] = ACTIONS(1709),
    [anon_sym_land] = ACTIONS(1709),
    [anon_sym_lor] = ACTIONS(1709),
    [anon_sym_lxor] = ACTIONS(1709),
    [aux_sym_mult_operator_token1] = ACTIONS(1709),
    [aux_sym_add_operator_token1] = ACTIONS(1709),
    [sym_concat_operator] = ACTIONS(1711),
    [sym_rel_operator] = ACTIONS(1709),
    [anon_sym_AMP_AMP] = ACTIONS(1709),
    [anon_sym_or] = ACTIONS(1709),
    [anon_sym_PIPE_PIPE] = ACTIONS(1709),
    [sym_assign_operator] = ACTIONS(1711),
    [sym__uppercase_identifier] = ACTIONS(1711),
    [aux_sym_tag_token1] = ACTIONS(1711),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(895)] = {
    [sym_attribute] = STATE(895),
    [sym__lowercase_identifier] = ACTIONS(1813),
    [anon_sym_TILDE] = ACTIONS(1813),
    [anon_sym_QMARK] = ACTIONS(1813),
    [anon_sym_LPAREN] = ACTIONS(1815),
    [anon_sym_RPAREN] = ACTIONS(1815),
    [anon_sym_COMMA] = ACTIONS(1815),
    [anon_sym_PIPE] = ACTIONS(1813),
    [anon_sym_COLON2] = ACTIONS(1813),
    [anon_sym_DOT] = ACTIONS(1815),
    [anon_sym_DASH_GT] = ACTIONS(1813),
    [anon_sym_LBRACK] = ACTIONS(1813),
    [anon_sym_RBRACK] = ACTIONS(1815),
    [anon_sym_true] = ACTIONS(1813),
    [anon_sym_false] = ACTIONS(1813),
    [anon_sym_LBRACE] = ACTIONS(1813),
    [anon_sym_SEMI] = ACTIONS(1815),
    [anon_sym_RBRACE] = ACTIONS(1815),
    [anon_sym_constraint] = ACTIONS(1813),
    [anon_sym_val] = ACTIONS(1813),
    [anon_sym_end] = ACTIONS(1813),
    [anon_sym_with] = ACTIONS(1813),
    [anon_sym_COLON_GT] = ACTIONS(1815),
    [anon_sym_object] = ACTIONS(1813),
    [anon_sym_inherit] = ACTIONS(1813),
    [anon_sym_method] = ACTIONS(1813),
    [anon_sym_initializer] = ACTIONS(1813),
    [anon_sym_AMP] = ACTIONS(1813),
    [anon_sym_POUND] = ACTIONS(1813),
    [anon_sym_COLON_COLON] = ACTIONS(1815),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1815),
    [anon_sym_then] = ACTIONS(1813),
    [anon_sym_else] = ACTIONS(1813),
    [anon_sym_do] = ACTIONS(1813),
    [anon_sym_done] = ACTIONS(1813),
    [anon_sym_to] = ACTIONS(1813),
    [anon_sym_downto] = ACTIONS(1813),
    [anon_sym_new] = ACTIONS(1813),
    [anon_sym_LBRACE_LT] = ACTIONS(1815),
    [anon_sym_GT_RBRACE] = ACTIONS(1815),
    [anon_sym_begin] = ACTIONS(1813),
    [sym_ocamlyacc_value] = ACTIONS(1815),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1813),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1815),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1813),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1813),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1815),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1815),
    [aux_sym_number_token1] = ACTIONS(1815),
    [aux_sym_signed_number_token1] = ACTIONS(1813),
    [anon_sym_SQUOTE] = ACTIONS(1815),
    [anon_sym_DQUOTE] = ACTIONS(1815),
    [sym_prefix_operator] = ACTIONS(1815),
    [aux_sym_sign_operator_token1] = ACTIONS(1813),
    [sym_hash_operator] = ACTIONS(1815),
    [anon_sym_lsl] = ACTIONS(1813),
    [anon_sym_lsr] = ACTIONS(1813),
    [anon_sym_asr] = ACTIONS(1813),
    [aux_sym_pow_operator_token1] = ACTIONS(1815),
    [anon_sym_mod] = ACTIONS(1813),
    [anon_sym_land] = ACTIONS(1813),
    [anon_sym_lor] = ACTIONS(1813),
    [anon_sym_lxor] = ACTIONS(1813),
    [aux_sym_mult_operator_token1] = ACTIONS(1813),
    [aux_sym_add_operator_token1] = ACTIONS(1813),
    [sym_concat_operator] = ACTIONS(1815),
    [sym_rel_operator] = ACTIONS(1813),
    [anon_sym_AMP_AMP] = ACTIONS(1813),
    [anon_sym_or] = ACTIONS(1813),
    [anon_sym_PIPE_PIPE] = ACTIONS(1813),
    [sym_assign_operator] = ACTIONS(1815),
    [sym__uppercase_identifier] = ACTIONS(1815),
    [aux_sym_tag_token1] = ACTIONS(1815),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(896)] = {
    [sym_attribute] = STATE(896),
    [sym__lowercase_identifier] = ACTIONS(1829),
    [anon_sym_TILDE] = ACTIONS(1829),
    [anon_sym_QMARK] = ACTIONS(1829),
    [anon_sym_LPAREN] = ACTIONS(1831),
    [anon_sym_RPAREN] = ACTIONS(1831),
    [anon_sym_COMMA] = ACTIONS(1831),
    [anon_sym_PIPE] = ACTIONS(1829),
    [anon_sym_COLON2] = ACTIONS(1829),
    [anon_sym_DOT] = ACTIONS(1831),
    [anon_sym_DASH_GT] = ACTIONS(1829),
    [anon_sym_LBRACK] = ACTIONS(1829),
    [anon_sym_RBRACK] = ACTIONS(1831),
    [anon_sym_true] = ACTIONS(1829),
    [anon_sym_false] = ACTIONS(1829),
    [anon_sym_LBRACE] = ACTIONS(1829),
    [anon_sym_SEMI] = ACTIONS(1831),
    [anon_sym_RBRACE] = ACTIONS(1831),
    [anon_sym_constraint] = ACTIONS(1829),
    [anon_sym_val] = ACTIONS(1829),
    [anon_sym_end] = ACTIONS(1829),
    [anon_sym_with] = ACTIONS(1829),
    [anon_sym_COLON_GT] = ACTIONS(1831),
    [anon_sym_object] = ACTIONS(1829),
    [anon_sym_inherit] = ACTIONS(1829),
    [anon_sym_method] = ACTIONS(1829),
    [anon_sym_initializer] = ACTIONS(1829),
    [anon_sym_AMP] = ACTIONS(1829),
    [anon_sym_POUND] = ACTIONS(1829),
    [anon_sym_COLON_COLON] = ACTIONS(1831),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1831),
    [anon_sym_then] = ACTIONS(1829),
    [anon_sym_else] = ACTIONS(1829),
    [anon_sym_do] = ACTIONS(1829),
    [anon_sym_done] = ACTIONS(1829),
    [anon_sym_to] = ACTIONS(1829),
    [anon_sym_downto] = ACTIONS(1829),
    [anon_sym_new] = ACTIONS(1829),
    [anon_sym_LBRACE_LT] = ACTIONS(1831),
    [anon_sym_GT_RBRACE] = ACTIONS(1831),
    [anon_sym_begin] = ACTIONS(1829),
    [sym_ocamlyacc_value] = ACTIONS(1831),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1829),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1831),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1829),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1829),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1831),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1831),
    [aux_sym_number_token1] = ACTIONS(1831),
    [aux_sym_signed_number_token1] = ACTIONS(1829),
    [anon_sym_SQUOTE] = ACTIONS(1831),
    [anon_sym_DQUOTE] = ACTIONS(1831),
    [sym_prefix_operator] = ACTIONS(1831),
    [aux_sym_sign_operator_token1] = ACTIONS(1829),
    [sym_hash_operator] = ACTIONS(1831),
    [anon_sym_lsl] = ACTIONS(1829),
    [anon_sym_lsr] = ACTIONS(1829),
    [anon_sym_asr] = ACTIONS(1829),
    [aux_sym_pow_operator_token1] = ACTIONS(1831),
    [anon_sym_mod] = ACTIONS(1829),
    [anon_sym_land] = ACTIONS(1829),
    [anon_sym_lor] = ACTIONS(1829),
    [anon_sym_lxor] = ACTIONS(1829),
    [aux_sym_mult_operator_token1] = ACTIONS(1829),
    [aux_sym_add_operator_token1] = ACTIONS(1829),
    [sym_concat_operator] = ACTIONS(1831),
    [sym_rel_operator] = ACTIONS(1829),
    [anon_sym_AMP_AMP] = ACTIONS(1829),
    [anon_sym_or] = ACTIONS(1829),
    [anon_sym_PIPE_PIPE] = ACTIONS(1829),
    [sym_assign_operator] = ACTIONS(1831),
    [sym__uppercase_identifier] = ACTIONS(1831),
    [aux_sym_tag_token1] = ACTIONS(1831),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(897)] = {
    [sym_attribute] = STATE(897),
    [sym__lowercase_identifier] = ACTIONS(1785),
    [anon_sym_TILDE] = ACTIONS(1785),
    [anon_sym_QMARK] = ACTIONS(1785),
    [anon_sym_LPAREN] = ACTIONS(1787),
    [anon_sym_RPAREN] = ACTIONS(1787),
    [anon_sym_COMMA] = ACTIONS(1787),
    [anon_sym_PIPE] = ACTIONS(1785),
    [anon_sym_COLON2] = ACTIONS(1785),
    [anon_sym_DOT] = ACTIONS(1787),
    [anon_sym_DASH_GT] = ACTIONS(1785),
    [anon_sym_LBRACK] = ACTIONS(1785),
    [anon_sym_RBRACK] = ACTIONS(1787),
    [anon_sym_true] = ACTIONS(1785),
    [anon_sym_false] = ACTIONS(1785),
    [anon_sym_LBRACE] = ACTIONS(1785),
    [anon_sym_SEMI] = ACTIONS(1787),
    [anon_sym_RBRACE] = ACTIONS(1787),
    [anon_sym_constraint] = ACTIONS(1785),
    [anon_sym_val] = ACTIONS(1785),
    [anon_sym_end] = ACTIONS(1785),
    [anon_sym_with] = ACTIONS(1785),
    [anon_sym_COLON_GT] = ACTIONS(1787),
    [anon_sym_object] = ACTIONS(1785),
    [anon_sym_inherit] = ACTIONS(1785),
    [anon_sym_method] = ACTIONS(1785),
    [anon_sym_initializer] = ACTIONS(1785),
    [anon_sym_AMP] = ACTIONS(1785),
    [anon_sym_POUND] = ACTIONS(1785),
    [anon_sym_COLON_COLON] = ACTIONS(1787),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1787),
    [anon_sym_then] = ACTIONS(1785),
    [anon_sym_else] = ACTIONS(1785),
    [anon_sym_do] = ACTIONS(1785),
    [anon_sym_done] = ACTIONS(1785),
    [anon_sym_to] = ACTIONS(1785),
    [anon_sym_downto] = ACTIONS(1785),
    [anon_sym_new] = ACTIONS(1785),
    [anon_sym_LBRACE_LT] = ACTIONS(1787),
    [anon_sym_GT_RBRACE] = ACTIONS(1787),
    [anon_sym_begin] = ACTIONS(1785),
    [sym_ocamlyacc_value] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1785),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1787),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1785),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1785),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1787),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1787),
    [aux_sym_number_token1] = ACTIONS(1787),
    [aux_sym_signed_number_token1] = ACTIONS(1785),
    [anon_sym_SQUOTE] = ACTIONS(1787),
    [anon_sym_DQUOTE] = ACTIONS(1787),
    [sym_prefix_operator] = ACTIONS(1787),
    [aux_sym_sign_operator_token1] = ACTIONS(1785),
    [sym_hash_operator] = ACTIONS(1787),
    [anon_sym_lsl] = ACTIONS(1785),
    [anon_sym_lsr] = ACTIONS(1785),
    [anon_sym_asr] = ACTIONS(1785),
    [aux_sym_pow_operator_token1] = ACTIONS(1787),
    [anon_sym_mod] = ACTIONS(1785),
    [anon_sym_land] = ACTIONS(1785),
    [anon_sym_lor] = ACTIONS(1785),
    [anon_sym_lxor] = ACTIONS(1785),
    [aux_sym_mult_operator_token1] = ACTIONS(1785),
    [aux_sym_add_operator_token1] = ACTIONS(1785),
    [sym_concat_operator] = ACTIONS(1787),
    [sym_rel_operator] = ACTIONS(1785),
    [anon_sym_AMP_AMP] = ACTIONS(1785),
    [anon_sym_or] = ACTIONS(1785),
    [anon_sym_PIPE_PIPE] = ACTIONS(1785),
    [sym_assign_operator] = ACTIONS(1787),
    [sym__uppercase_identifier] = ACTIONS(1787),
    [aux_sym_tag_token1] = ACTIONS(1787),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(898)] = {
    [sym_attribute] = STATE(898),
    [sym__lowercase_identifier] = ACTIONS(1797),
    [anon_sym_TILDE] = ACTIONS(1797),
    [anon_sym_QMARK] = ACTIONS(1797),
    [anon_sym_LPAREN] = ACTIONS(1799),
    [anon_sym_RPAREN] = ACTIONS(1799),
    [anon_sym_COMMA] = ACTIONS(1799),
    [anon_sym_PIPE] = ACTIONS(1797),
    [anon_sym_COLON2] = ACTIONS(1797),
    [anon_sym_DOT] = ACTIONS(1799),
    [anon_sym_DASH_GT] = ACTIONS(1797),
    [anon_sym_LBRACK] = ACTIONS(1797),
    [anon_sym_RBRACK] = ACTIONS(1799),
    [anon_sym_true] = ACTIONS(1797),
    [anon_sym_false] = ACTIONS(1797),
    [anon_sym_LBRACE] = ACTIONS(1797),
    [anon_sym_SEMI] = ACTIONS(1799),
    [anon_sym_RBRACE] = ACTIONS(1799),
    [anon_sym_constraint] = ACTIONS(1797),
    [anon_sym_val] = ACTIONS(1797),
    [anon_sym_end] = ACTIONS(1797),
    [anon_sym_with] = ACTIONS(1797),
    [anon_sym_COLON_GT] = ACTIONS(1799),
    [anon_sym_object] = ACTIONS(1797),
    [anon_sym_inherit] = ACTIONS(1797),
    [anon_sym_method] = ACTIONS(1797),
    [anon_sym_initializer] = ACTIONS(1797),
    [anon_sym_AMP] = ACTIONS(1797),
    [anon_sym_POUND] = ACTIONS(1797),
    [anon_sym_COLON_COLON] = ACTIONS(1799),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1799),
    [anon_sym_then] = ACTIONS(1797),
    [anon_sym_else] = ACTIONS(1797),
    [anon_sym_do] = ACTIONS(1797),
    [anon_sym_done] = ACTIONS(1797),
    [anon_sym_to] = ACTIONS(1797),
    [anon_sym_downto] = ACTIONS(1797),
    [anon_sym_new] = ACTIONS(1797),
    [anon_sym_LBRACE_LT] = ACTIONS(1799),
    [anon_sym_GT_RBRACE] = ACTIONS(1799),
    [anon_sym_begin] = ACTIONS(1797),
    [sym_ocamlyacc_value] = ACTIONS(1799),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1797),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1799),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1797),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1797),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1799),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1799),
    [aux_sym_signed_number_token1] = ACTIONS(1797),
    [anon_sym_SQUOTE] = ACTIONS(1799),
    [anon_sym_DQUOTE] = ACTIONS(1799),
    [sym_prefix_operator] = ACTIONS(1799),
    [aux_sym_sign_operator_token1] = ACTIONS(1797),
    [sym_hash_operator] = ACTIONS(1799),
    [anon_sym_lsl] = ACTIONS(1797),
    [anon_sym_lsr] = ACTIONS(1797),
    [anon_sym_asr] = ACTIONS(1797),
    [aux_sym_pow_operator_token1] = ACTIONS(1799),
    [anon_sym_mod] = ACTIONS(1797),
    [anon_sym_land] = ACTIONS(1797),
    [anon_sym_lor] = ACTIONS(1797),
    [anon_sym_lxor] = ACTIONS(1797),
    [aux_sym_mult_operator_token1] = ACTIONS(1797),
    [aux_sym_add_operator_token1] = ACTIONS(1797),
    [sym_concat_operator] = ACTIONS(1799),
    [sym_rel_operator] = ACTIONS(1797),
    [anon_sym_AMP_AMP] = ACTIONS(1797),
    [anon_sym_or] = ACTIONS(1797),
    [anon_sym_PIPE_PIPE] = ACTIONS(1797),
    [sym_assign_operator] = ACTIONS(1799),
    [sym__uppercase_identifier] = ACTIONS(1799),
    [aux_sym_tag_token1] = ACTIONS(1799),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(899)] = {
    [sym_attribute] = STATE(899),
    [sym__lowercase_identifier] = ACTIONS(1805),
    [anon_sym_TILDE] = ACTIONS(1805),
    [anon_sym_QMARK] = ACTIONS(1805),
    [anon_sym_LPAREN] = ACTIONS(1807),
    [anon_sym_RPAREN] = ACTIONS(1807),
    [anon_sym_COMMA] = ACTIONS(1807),
    [anon_sym_PIPE] = ACTIONS(1805),
    [anon_sym_COLON2] = ACTIONS(1805),
    [anon_sym_DOT] = ACTIONS(1807),
    [anon_sym_DASH_GT] = ACTIONS(1805),
    [anon_sym_LBRACK] = ACTIONS(1805),
    [anon_sym_RBRACK] = ACTIONS(1807),
    [anon_sym_true] = ACTIONS(1805),
    [anon_sym_false] = ACTIONS(1805),
    [anon_sym_LBRACE] = ACTIONS(1805),
    [anon_sym_SEMI] = ACTIONS(1807),
    [anon_sym_RBRACE] = ACTIONS(1807),
    [anon_sym_constraint] = ACTIONS(1805),
    [anon_sym_val] = ACTIONS(1805),
    [anon_sym_end] = ACTIONS(1805),
    [anon_sym_with] = ACTIONS(1805),
    [anon_sym_COLON_GT] = ACTIONS(1807),
    [anon_sym_object] = ACTIONS(1805),
    [anon_sym_inherit] = ACTIONS(1805),
    [anon_sym_method] = ACTIONS(1805),
    [anon_sym_initializer] = ACTIONS(1805),
    [anon_sym_AMP] = ACTIONS(1805),
    [anon_sym_POUND] = ACTIONS(1805),
    [anon_sym_COLON_COLON] = ACTIONS(1807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1807),
    [anon_sym_then] = ACTIONS(1805),
    [anon_sym_else] = ACTIONS(1805),
    [anon_sym_do] = ACTIONS(1805),
    [anon_sym_done] = ACTIONS(1805),
    [anon_sym_to] = ACTIONS(1805),
    [anon_sym_downto] = ACTIONS(1805),
    [anon_sym_new] = ACTIONS(1805),
    [anon_sym_LBRACE_LT] = ACTIONS(1807),
    [anon_sym_GT_RBRACE] = ACTIONS(1807),
    [anon_sym_begin] = ACTIONS(1805),
    [sym_ocamlyacc_value] = ACTIONS(1807),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1805),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1807),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1805),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1805),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1807),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1807),
    [aux_sym_number_token1] = ACTIONS(1807),
    [aux_sym_signed_number_token1] = ACTIONS(1805),
    [anon_sym_SQUOTE] = ACTIONS(1807),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym_prefix_operator] = ACTIONS(1807),
    [aux_sym_sign_operator_token1] = ACTIONS(1805),
    [sym_hash_operator] = ACTIONS(1807),
    [anon_sym_lsl] = ACTIONS(1805),
    [anon_sym_lsr] = ACTIONS(1805),
    [anon_sym_asr] = ACTIONS(1805),
    [aux_sym_pow_operator_token1] = ACTIONS(1807),
    [anon_sym_mod] = ACTIONS(1805),
    [anon_sym_land] = ACTIONS(1805),
    [anon_sym_lor] = ACTIONS(1805),
    [anon_sym_lxor] = ACTIONS(1805),
    [aux_sym_mult_operator_token1] = ACTIONS(1805),
    [aux_sym_add_operator_token1] = ACTIONS(1805),
    [sym_concat_operator] = ACTIONS(1807),
    [sym_rel_operator] = ACTIONS(1805),
    [anon_sym_AMP_AMP] = ACTIONS(1805),
    [anon_sym_or] = ACTIONS(1805),
    [anon_sym_PIPE_PIPE] = ACTIONS(1805),
    [sym_assign_operator] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1807),
    [aux_sym_tag_token1] = ACTIONS(1807),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(900)] = {
    [sym_attribute] = STATE(900),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_TILDE] = ACTIONS(1757),
    [anon_sym_QMARK] = ACTIONS(1757),
    [anon_sym_LPAREN] = ACTIONS(1759),
    [anon_sym_RPAREN] = ACTIONS(1759),
    [anon_sym_COMMA] = ACTIONS(1759),
    [anon_sym_PIPE] = ACTIONS(1757),
    [anon_sym_COLON2] = ACTIONS(1757),
    [anon_sym_DOT] = ACTIONS(1759),
    [anon_sym_DASH_GT] = ACTIONS(1757),
    [anon_sym_LBRACK] = ACTIONS(1757),
    [anon_sym_RBRACK] = ACTIONS(1759),
    [anon_sym_true] = ACTIONS(1757),
    [anon_sym_false] = ACTIONS(1757),
    [anon_sym_LBRACE] = ACTIONS(1757),
    [anon_sym_SEMI] = ACTIONS(1759),
    [anon_sym_RBRACE] = ACTIONS(1759),
    [anon_sym_constraint] = ACTIONS(1757),
    [anon_sym_val] = ACTIONS(1757),
    [anon_sym_end] = ACTIONS(1757),
    [anon_sym_with] = ACTIONS(1757),
    [anon_sym_COLON_GT] = ACTIONS(1759),
    [anon_sym_object] = ACTIONS(1757),
    [anon_sym_inherit] = ACTIONS(1757),
    [anon_sym_method] = ACTIONS(1757),
    [anon_sym_initializer] = ACTIONS(1757),
    [anon_sym_AMP] = ACTIONS(1757),
    [anon_sym_POUND] = ACTIONS(1757),
    [anon_sym_COLON_COLON] = ACTIONS(1759),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1759),
    [anon_sym_then] = ACTIONS(1757),
    [anon_sym_else] = ACTIONS(1757),
    [anon_sym_do] = ACTIONS(1757),
    [anon_sym_done] = ACTIONS(1757),
    [anon_sym_to] = ACTIONS(1757),
    [anon_sym_downto] = ACTIONS(1757),
    [anon_sym_new] = ACTIONS(1757),
    [anon_sym_LBRACE_LT] = ACTIONS(1759),
    [anon_sym_GT_RBRACE] = ACTIONS(1759),
    [anon_sym_begin] = ACTIONS(1757),
    [sym_ocamlyacc_value] = ACTIONS(1759),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1757),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1759),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1757),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1757),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1759),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1759),
    [aux_sym_number_token1] = ACTIONS(1759),
    [aux_sym_signed_number_token1] = ACTIONS(1757),
    [anon_sym_SQUOTE] = ACTIONS(1759),
    [anon_sym_DQUOTE] = ACTIONS(1759),
    [sym_prefix_operator] = ACTIONS(1759),
    [aux_sym_sign_operator_token1] = ACTIONS(1757),
    [sym_hash_operator] = ACTIONS(1759),
    [anon_sym_lsl] = ACTIONS(1757),
    [anon_sym_lsr] = ACTIONS(1757),
    [anon_sym_asr] = ACTIONS(1757),
    [aux_sym_pow_operator_token1] = ACTIONS(1759),
    [anon_sym_mod] = ACTIONS(1757),
    [anon_sym_land] = ACTIONS(1757),
    [anon_sym_lor] = ACTIONS(1757),
    [anon_sym_lxor] = ACTIONS(1757),
    [aux_sym_mult_operator_token1] = ACTIONS(1757),
    [aux_sym_add_operator_token1] = ACTIONS(1757),
    [sym_concat_operator] = ACTIONS(1759),
    [sym_rel_operator] = ACTIONS(1757),
    [anon_sym_AMP_AMP] = ACTIONS(1757),
    [anon_sym_or] = ACTIONS(1757),
    [anon_sym_PIPE_PIPE] = ACTIONS(1757),
    [sym_assign_operator] = ACTIONS(1759),
    [sym__uppercase_identifier] = ACTIONS(1759),
    [aux_sym_tag_token1] = ACTIONS(1759),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(901)] = {
    [sym_attribute] = STATE(901),
    [sym__lowercase_identifier] = ACTIONS(1623),
    [anon_sym_TILDE] = ACTIONS(1623),
    [anon_sym_QMARK] = ACTIONS(1623),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_RPAREN] = ACTIONS(1625),
    [anon_sym_COMMA] = ACTIONS(1625),
    [anon_sym_PIPE] = ACTIONS(1623),
    [anon_sym_COLON2] = ACTIONS(1623),
    [anon_sym_DOT] = ACTIONS(1625),
    [anon_sym_DASH_GT] = ACTIONS(1623),
    [anon_sym_LBRACK] = ACTIONS(1623),
    [anon_sym_RBRACK] = ACTIONS(1625),
    [anon_sym_true] = ACTIONS(1623),
    [anon_sym_false] = ACTIONS(1623),
    [anon_sym_LBRACE] = ACTIONS(1623),
    [anon_sym_SEMI] = ACTIONS(1625),
    [anon_sym_RBRACE] = ACTIONS(1625),
    [anon_sym_constraint] = ACTIONS(1623),
    [anon_sym_val] = ACTIONS(1623),
    [anon_sym_end] = ACTIONS(1623),
    [anon_sym_with] = ACTIONS(1623),
    [anon_sym_COLON_GT] = ACTIONS(1625),
    [anon_sym_object] = ACTIONS(1623),
    [anon_sym_inherit] = ACTIONS(1623),
    [anon_sym_method] = ACTIONS(1623),
    [anon_sym_initializer] = ACTIONS(1623),
    [anon_sym_AMP] = ACTIONS(1623),
    [anon_sym_POUND] = ACTIONS(1623),
    [anon_sym_COLON_COLON] = ACTIONS(1625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1625),
    [anon_sym_then] = ACTIONS(1623),
    [anon_sym_else] = ACTIONS(1623),
    [anon_sym_do] = ACTIONS(1623),
    [anon_sym_done] = ACTIONS(1623),
    [anon_sym_to] = ACTIONS(1623),
    [anon_sym_downto] = ACTIONS(1623),
    [anon_sym_new] = ACTIONS(1623),
    [anon_sym_LBRACE_LT] = ACTIONS(1625),
    [anon_sym_GT_RBRACE] = ACTIONS(1625),
    [anon_sym_begin] = ACTIONS(1623),
    [sym_ocamlyacc_value] = ACTIONS(1625),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1623),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1625),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1623),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1623),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1625),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1625),
    [aux_sym_number_token1] = ACTIONS(1625),
    [aux_sym_signed_number_token1] = ACTIONS(1623),
    [anon_sym_SQUOTE] = ACTIONS(1625),
    [anon_sym_DQUOTE] = ACTIONS(1625),
    [sym_prefix_operator] = ACTIONS(1625),
    [aux_sym_sign_operator_token1] = ACTIONS(1623),
    [sym_hash_operator] = ACTIONS(1625),
    [anon_sym_lsl] = ACTIONS(1623),
    [anon_sym_lsr] = ACTIONS(1623),
    [anon_sym_asr] = ACTIONS(1623),
    [aux_sym_pow_operator_token1] = ACTIONS(1625),
    [anon_sym_mod] = ACTIONS(1623),
    [anon_sym_land] = ACTIONS(1623),
    [anon_sym_lor] = ACTIONS(1623),
    [anon_sym_lxor] = ACTIONS(1623),
    [aux_sym_mult_operator_token1] = ACTIONS(1623),
    [aux_sym_add_operator_token1] = ACTIONS(1623),
    [sym_concat_operator] = ACTIONS(1625),
    [sym_rel_operator] = ACTIONS(1623),
    [anon_sym_AMP_AMP] = ACTIONS(1623),
    [anon_sym_or] = ACTIONS(1623),
    [anon_sym_PIPE_PIPE] = ACTIONS(1623),
    [sym_assign_operator] = ACTIONS(1625),
    [sym__uppercase_identifier] = ACTIONS(1625),
    [aux_sym_tag_token1] = ACTIONS(1625),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(902)] = {
    [sym_attribute] = STATE(902),
    [sym__lowercase_identifier] = ACTIONS(1733),
    [anon_sym_TILDE] = ACTIONS(1733),
    [anon_sym_QMARK] = ACTIONS(1733),
    [anon_sym_LPAREN] = ACTIONS(1735),
    [anon_sym_RPAREN] = ACTIONS(1735),
    [anon_sym_COMMA] = ACTIONS(1735),
    [anon_sym_PIPE] = ACTIONS(1733),
    [anon_sym_COLON2] = ACTIONS(1733),
    [anon_sym_DOT] = ACTIONS(1735),
    [anon_sym_DASH_GT] = ACTIONS(1733),
    [anon_sym_LBRACK] = ACTIONS(1733),
    [anon_sym_RBRACK] = ACTIONS(1735),
    [anon_sym_true] = ACTIONS(1733),
    [anon_sym_false] = ACTIONS(1733),
    [anon_sym_LBRACE] = ACTIONS(1733),
    [anon_sym_SEMI] = ACTIONS(1735),
    [anon_sym_RBRACE] = ACTIONS(1735),
    [anon_sym_constraint] = ACTIONS(1733),
    [anon_sym_val] = ACTIONS(1733),
    [anon_sym_end] = ACTIONS(1733),
    [anon_sym_with] = ACTIONS(1733),
    [anon_sym_COLON_GT] = ACTIONS(1735),
    [anon_sym_object] = ACTIONS(1733),
    [anon_sym_inherit] = ACTIONS(1733),
    [anon_sym_method] = ACTIONS(1733),
    [anon_sym_initializer] = ACTIONS(1733),
    [anon_sym_AMP] = ACTIONS(1733),
    [anon_sym_POUND] = ACTIONS(1733),
    [anon_sym_COLON_COLON] = ACTIONS(1735),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1735),
    [anon_sym_then] = ACTIONS(1733),
    [anon_sym_else] = ACTIONS(1733),
    [anon_sym_do] = ACTIONS(1733),
    [anon_sym_done] = ACTIONS(1733),
    [anon_sym_to] = ACTIONS(1733),
    [anon_sym_downto] = ACTIONS(1733),
    [anon_sym_new] = ACTIONS(1733),
    [anon_sym_LBRACE_LT] = ACTIONS(1735),
    [anon_sym_GT_RBRACE] = ACTIONS(1735),
    [anon_sym_begin] = ACTIONS(1733),
    [sym_ocamlyacc_value] = ACTIONS(1735),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1733),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1735),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1733),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1733),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1735),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1735),
    [aux_sym_number_token1] = ACTIONS(1735),
    [aux_sym_signed_number_token1] = ACTIONS(1733),
    [anon_sym_SQUOTE] = ACTIONS(1735),
    [anon_sym_DQUOTE] = ACTIONS(1735),
    [sym_prefix_operator] = ACTIONS(1735),
    [aux_sym_sign_operator_token1] = ACTIONS(1733),
    [sym_hash_operator] = ACTIONS(1735),
    [anon_sym_lsl] = ACTIONS(1733),
    [anon_sym_lsr] = ACTIONS(1733),
    [anon_sym_asr] = ACTIONS(1733),
    [aux_sym_pow_operator_token1] = ACTIONS(1735),
    [anon_sym_mod] = ACTIONS(1733),
    [anon_sym_land] = ACTIONS(1733),
    [anon_sym_lor] = ACTIONS(1733),
    [anon_sym_lxor] = ACTIONS(1733),
    [aux_sym_mult_operator_token1] = ACTIONS(1733),
    [aux_sym_add_operator_token1] = ACTIONS(1733),
    [sym_concat_operator] = ACTIONS(1735),
    [sym_rel_operator] = ACTIONS(1733),
    [anon_sym_AMP_AMP] = ACTIONS(1733),
    [anon_sym_or] = ACTIONS(1733),
    [anon_sym_PIPE_PIPE] = ACTIONS(1733),
    [sym_assign_operator] = ACTIONS(1735),
    [sym__uppercase_identifier] = ACTIONS(1735),
    [aux_sym_tag_token1] = ACTIONS(1735),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(903)] = {
    [sym_attribute] = STATE(903),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_TILDE] = ACTIONS(1817),
    [anon_sym_QMARK] = ACTIONS(1817),
    [anon_sym_LPAREN] = ACTIONS(1819),
    [anon_sym_RPAREN] = ACTIONS(1819),
    [anon_sym_COMMA] = ACTIONS(1819),
    [anon_sym_PIPE] = ACTIONS(1817),
    [anon_sym_COLON2] = ACTIONS(1817),
    [anon_sym_DOT] = ACTIONS(1819),
    [anon_sym_DASH_GT] = ACTIONS(1817),
    [anon_sym_LBRACK] = ACTIONS(1817),
    [anon_sym_RBRACK] = ACTIONS(1819),
    [anon_sym_true] = ACTIONS(1817),
    [anon_sym_false] = ACTIONS(1817),
    [anon_sym_LBRACE] = ACTIONS(1817),
    [anon_sym_SEMI] = ACTIONS(1819),
    [anon_sym_RBRACE] = ACTIONS(1819),
    [anon_sym_constraint] = ACTIONS(1817),
    [anon_sym_val] = ACTIONS(1817),
    [anon_sym_end] = ACTIONS(1817),
    [anon_sym_with] = ACTIONS(1817),
    [anon_sym_COLON_GT] = ACTIONS(1819),
    [anon_sym_object] = ACTIONS(1817),
    [anon_sym_inherit] = ACTIONS(1817),
    [anon_sym_method] = ACTIONS(1817),
    [anon_sym_initializer] = ACTIONS(1817),
    [anon_sym_AMP] = ACTIONS(1817),
    [anon_sym_POUND] = ACTIONS(1817),
    [anon_sym_COLON_COLON] = ACTIONS(1819),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1819),
    [anon_sym_then] = ACTIONS(1817),
    [anon_sym_else] = ACTIONS(1817),
    [anon_sym_do] = ACTIONS(1817),
    [anon_sym_done] = ACTIONS(1817),
    [anon_sym_to] = ACTIONS(1817),
    [anon_sym_downto] = ACTIONS(1817),
    [anon_sym_new] = ACTIONS(1817),
    [anon_sym_LBRACE_LT] = ACTIONS(1819),
    [anon_sym_GT_RBRACE] = ACTIONS(1819),
    [anon_sym_begin] = ACTIONS(1817),
    [sym_ocamlyacc_value] = ACTIONS(1819),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1817),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1819),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1817),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1817),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1819),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1819),
    [aux_sym_number_token1] = ACTIONS(1819),
    [aux_sym_signed_number_token1] = ACTIONS(1817),
    [anon_sym_SQUOTE] = ACTIONS(1819),
    [anon_sym_DQUOTE] = ACTIONS(1819),
    [sym_prefix_operator] = ACTIONS(1819),
    [aux_sym_sign_operator_token1] = ACTIONS(1817),
    [sym_hash_operator] = ACTIONS(1819),
    [anon_sym_lsl] = ACTIONS(1817),
    [anon_sym_lsr] = ACTIONS(1817),
    [anon_sym_asr] = ACTIONS(1817),
    [aux_sym_pow_operator_token1] = ACTIONS(1819),
    [anon_sym_mod] = ACTIONS(1817),
    [anon_sym_land] = ACTIONS(1817),
    [anon_sym_lor] = ACTIONS(1817),
    [anon_sym_lxor] = ACTIONS(1817),
    [aux_sym_mult_operator_token1] = ACTIONS(1817),
    [aux_sym_add_operator_token1] = ACTIONS(1817),
    [sym_concat_operator] = ACTIONS(1819),
    [sym_rel_operator] = ACTIONS(1817),
    [anon_sym_AMP_AMP] = ACTIONS(1817),
    [anon_sym_or] = ACTIONS(1817),
    [anon_sym_PIPE_PIPE] = ACTIONS(1817),
    [sym_assign_operator] = ACTIONS(1819),
    [sym__uppercase_identifier] = ACTIONS(1819),
    [aux_sym_tag_token1] = ACTIONS(1819),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(904)] = {
    [sym_attribute] = STATE(904),
    [sym__lowercase_identifier] = ACTIONS(1833),
    [anon_sym_TILDE] = ACTIONS(1833),
    [anon_sym_QMARK] = ACTIONS(1833),
    [anon_sym_LPAREN] = ACTIONS(1835),
    [anon_sym_RPAREN] = ACTIONS(1835),
    [anon_sym_COMMA] = ACTIONS(1835),
    [anon_sym_PIPE] = ACTIONS(1833),
    [anon_sym_COLON2] = ACTIONS(1833),
    [anon_sym_DOT] = ACTIONS(1835),
    [anon_sym_DASH_GT] = ACTIONS(1833),
    [anon_sym_LBRACK] = ACTIONS(1833),
    [anon_sym_RBRACK] = ACTIONS(1835),
    [anon_sym_true] = ACTIONS(1833),
    [anon_sym_false] = ACTIONS(1833),
    [anon_sym_LBRACE] = ACTIONS(1833),
    [anon_sym_SEMI] = ACTIONS(1835),
    [anon_sym_RBRACE] = ACTIONS(1835),
    [anon_sym_constraint] = ACTIONS(1833),
    [anon_sym_val] = ACTIONS(1833),
    [anon_sym_end] = ACTIONS(1833),
    [anon_sym_with] = ACTIONS(1833),
    [anon_sym_COLON_GT] = ACTIONS(1835),
    [anon_sym_object] = ACTIONS(1833),
    [anon_sym_inherit] = ACTIONS(1833),
    [anon_sym_method] = ACTIONS(1833),
    [anon_sym_initializer] = ACTIONS(1833),
    [anon_sym_AMP] = ACTIONS(1833),
    [anon_sym_POUND] = ACTIONS(1833),
    [anon_sym_COLON_COLON] = ACTIONS(1835),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1835),
    [anon_sym_then] = ACTIONS(1833),
    [anon_sym_else] = ACTIONS(1833),
    [anon_sym_do] = ACTIONS(1833),
    [anon_sym_done] = ACTIONS(1833),
    [anon_sym_to] = ACTIONS(1833),
    [anon_sym_downto] = ACTIONS(1833),
    [anon_sym_new] = ACTIONS(1833),
    [anon_sym_LBRACE_LT] = ACTIONS(1835),
    [anon_sym_GT_RBRACE] = ACTIONS(1835),
    [anon_sym_begin] = ACTIONS(1833),
    [sym_ocamlyacc_value] = ACTIONS(1835),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1833),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1835),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1833),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1833),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1835),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1835),
    [aux_sym_number_token1] = ACTIONS(1835),
    [aux_sym_signed_number_token1] = ACTIONS(1833),
    [anon_sym_SQUOTE] = ACTIONS(1835),
    [anon_sym_DQUOTE] = ACTIONS(1835),
    [sym_prefix_operator] = ACTIONS(1835),
    [aux_sym_sign_operator_token1] = ACTIONS(1833),
    [sym_hash_operator] = ACTIONS(1835),
    [anon_sym_lsl] = ACTIONS(1833),
    [anon_sym_lsr] = ACTIONS(1833),
    [anon_sym_asr] = ACTIONS(1833),
    [aux_sym_pow_operator_token1] = ACTIONS(1835),
    [anon_sym_mod] = ACTIONS(1833),
    [anon_sym_land] = ACTIONS(1833),
    [anon_sym_lor] = ACTIONS(1833),
    [anon_sym_lxor] = ACTIONS(1833),
    [aux_sym_mult_operator_token1] = ACTIONS(1833),
    [aux_sym_add_operator_token1] = ACTIONS(1833),
    [sym_concat_operator] = ACTIONS(1835),
    [sym_rel_operator] = ACTIONS(1833),
    [anon_sym_AMP_AMP] = ACTIONS(1833),
    [anon_sym_or] = ACTIONS(1833),
    [anon_sym_PIPE_PIPE] = ACTIONS(1833),
    [sym_assign_operator] = ACTIONS(1835),
    [sym__uppercase_identifier] = ACTIONS(1835),
    [aux_sym_tag_token1] = ACTIONS(1835),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(905)] = {
    [sym_attribute] = STATE(905),
    [sym__lowercase_identifier] = ACTIONS(1769),
    [anon_sym_TILDE] = ACTIONS(1769),
    [anon_sym_QMARK] = ACTIONS(1769),
    [anon_sym_LPAREN] = ACTIONS(1771),
    [anon_sym_RPAREN] = ACTIONS(1771),
    [anon_sym_COMMA] = ACTIONS(1771),
    [anon_sym_PIPE] = ACTIONS(1769),
    [anon_sym_COLON2] = ACTIONS(1769),
    [anon_sym_DOT] = ACTIONS(1771),
    [anon_sym_DASH_GT] = ACTIONS(1769),
    [anon_sym_LBRACK] = ACTIONS(1769),
    [anon_sym_RBRACK] = ACTIONS(1771),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1769),
    [anon_sym_SEMI] = ACTIONS(1771),
    [anon_sym_RBRACE] = ACTIONS(1771),
    [anon_sym_constraint] = ACTIONS(1769),
    [anon_sym_val] = ACTIONS(1769),
    [anon_sym_end] = ACTIONS(1769),
    [anon_sym_with] = ACTIONS(1769),
    [anon_sym_COLON_GT] = ACTIONS(1771),
    [anon_sym_object] = ACTIONS(1769),
    [anon_sym_inherit] = ACTIONS(1769),
    [anon_sym_method] = ACTIONS(1769),
    [anon_sym_initializer] = ACTIONS(1769),
    [anon_sym_AMP] = ACTIONS(1769),
    [anon_sym_POUND] = ACTIONS(1769),
    [anon_sym_COLON_COLON] = ACTIONS(1771),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1771),
    [anon_sym_then] = ACTIONS(1769),
    [anon_sym_else] = ACTIONS(1769),
    [anon_sym_do] = ACTIONS(1769),
    [anon_sym_done] = ACTIONS(1769),
    [anon_sym_to] = ACTIONS(1769),
    [anon_sym_downto] = ACTIONS(1769),
    [anon_sym_new] = ACTIONS(1769),
    [anon_sym_LBRACE_LT] = ACTIONS(1771),
    [anon_sym_GT_RBRACE] = ACTIONS(1771),
    [anon_sym_begin] = ACTIONS(1769),
    [sym_ocamlyacc_value] = ACTIONS(1771),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1769),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1771),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1769),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1769),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1771),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1771),
    [aux_sym_number_token1] = ACTIONS(1771),
    [aux_sym_signed_number_token1] = ACTIONS(1769),
    [anon_sym_SQUOTE] = ACTIONS(1771),
    [anon_sym_DQUOTE] = ACTIONS(1771),
    [sym_prefix_operator] = ACTIONS(1771),
    [aux_sym_sign_operator_token1] = ACTIONS(1769),
    [sym_hash_operator] = ACTIONS(1771),
    [anon_sym_lsl] = ACTIONS(1769),
    [anon_sym_lsr] = ACTIONS(1769),
    [anon_sym_asr] = ACTIONS(1769),
    [aux_sym_pow_operator_token1] = ACTIONS(1771),
    [anon_sym_mod] = ACTIONS(1769),
    [anon_sym_land] = ACTIONS(1769),
    [anon_sym_lor] = ACTIONS(1769),
    [anon_sym_lxor] = ACTIONS(1769),
    [aux_sym_mult_operator_token1] = ACTIONS(1769),
    [aux_sym_add_operator_token1] = ACTIONS(1769),
    [sym_concat_operator] = ACTIONS(1771),
    [sym_rel_operator] = ACTIONS(1769),
    [anon_sym_AMP_AMP] = ACTIONS(1769),
    [anon_sym_or] = ACTIONS(1769),
    [anon_sym_PIPE_PIPE] = ACTIONS(1769),
    [sym_assign_operator] = ACTIONS(1771),
    [sym__uppercase_identifier] = ACTIONS(1771),
    [aux_sym_tag_token1] = ACTIONS(1771),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(906)] = {
    [sym_attribute] = STATE(906),
    [sym__lowercase_identifier] = ACTIONS(1789),
    [anon_sym_TILDE] = ACTIONS(1789),
    [anon_sym_QMARK] = ACTIONS(1789),
    [anon_sym_LPAREN] = ACTIONS(1791),
    [anon_sym_RPAREN] = ACTIONS(1791),
    [anon_sym_COMMA] = ACTIONS(1791),
    [anon_sym_PIPE] = ACTIONS(1789),
    [anon_sym_COLON2] = ACTIONS(1789),
    [anon_sym_DOT] = ACTIONS(1791),
    [anon_sym_DASH_GT] = ACTIONS(1789),
    [anon_sym_LBRACK] = ACTIONS(1789),
    [anon_sym_RBRACK] = ACTIONS(1791),
    [anon_sym_true] = ACTIONS(1789),
    [anon_sym_false] = ACTIONS(1789),
    [anon_sym_LBRACE] = ACTIONS(1789),
    [anon_sym_SEMI] = ACTIONS(1791),
    [anon_sym_RBRACE] = ACTIONS(1791),
    [anon_sym_constraint] = ACTIONS(1789),
    [anon_sym_val] = ACTIONS(1789),
    [anon_sym_end] = ACTIONS(1789),
    [anon_sym_with] = ACTIONS(1789),
    [anon_sym_COLON_GT] = ACTIONS(1791),
    [anon_sym_object] = ACTIONS(1789),
    [anon_sym_inherit] = ACTIONS(1789),
    [anon_sym_method] = ACTIONS(1789),
    [anon_sym_initializer] = ACTIONS(1789),
    [anon_sym_AMP] = ACTIONS(1789),
    [anon_sym_POUND] = ACTIONS(1789),
    [anon_sym_COLON_COLON] = ACTIONS(1791),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1791),
    [anon_sym_then] = ACTIONS(1789),
    [anon_sym_else] = ACTIONS(1789),
    [anon_sym_do] = ACTIONS(1789),
    [anon_sym_done] = ACTIONS(1789),
    [anon_sym_to] = ACTIONS(1789),
    [anon_sym_downto] = ACTIONS(1789),
    [anon_sym_new] = ACTIONS(1789),
    [anon_sym_LBRACE_LT] = ACTIONS(1791),
    [anon_sym_GT_RBRACE] = ACTIONS(1791),
    [anon_sym_begin] = ACTIONS(1789),
    [sym_ocamlyacc_value] = ACTIONS(1791),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1791),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1789),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1791),
    [aux_sym_number_token1] = ACTIONS(1791),
    [aux_sym_signed_number_token1] = ACTIONS(1789),
    [anon_sym_SQUOTE] = ACTIONS(1791),
    [anon_sym_DQUOTE] = ACTIONS(1791),
    [sym_prefix_operator] = ACTIONS(1791),
    [aux_sym_sign_operator_token1] = ACTIONS(1789),
    [sym_hash_operator] = ACTIONS(1791),
    [anon_sym_lsl] = ACTIONS(1789),
    [anon_sym_lsr] = ACTIONS(1789),
    [anon_sym_asr] = ACTIONS(1789),
    [aux_sym_pow_operator_token1] = ACTIONS(1791),
    [anon_sym_mod] = ACTIONS(1789),
    [anon_sym_land] = ACTIONS(1789),
    [anon_sym_lor] = ACTIONS(1789),
    [anon_sym_lxor] = ACTIONS(1789),
    [aux_sym_mult_operator_token1] = ACTIONS(1789),
    [aux_sym_add_operator_token1] = ACTIONS(1789),
    [sym_concat_operator] = ACTIONS(1791),
    [sym_rel_operator] = ACTIONS(1789),
    [anon_sym_AMP_AMP] = ACTIONS(1789),
    [anon_sym_or] = ACTIONS(1789),
    [anon_sym_PIPE_PIPE] = ACTIONS(1789),
    [sym_assign_operator] = ACTIONS(1791),
    [sym__uppercase_identifier] = ACTIONS(1791),
    [aux_sym_tag_token1] = ACTIONS(1791),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(907)] = {
    [sym_attribute] = STATE(907),
    [sym__lowercase_identifier] = ACTIONS(1801),
    [anon_sym_TILDE] = ACTIONS(1801),
    [anon_sym_QMARK] = ACTIONS(1801),
    [anon_sym_LPAREN] = ACTIONS(1803),
    [anon_sym_RPAREN] = ACTIONS(1803),
    [anon_sym_COMMA] = ACTIONS(1803),
    [anon_sym_PIPE] = ACTIONS(1801),
    [anon_sym_COLON2] = ACTIONS(1801),
    [anon_sym_DOT] = ACTIONS(1803),
    [anon_sym_DASH_GT] = ACTIONS(1801),
    [anon_sym_LBRACK] = ACTIONS(1801),
    [anon_sym_RBRACK] = ACTIONS(1803),
    [anon_sym_true] = ACTIONS(1801),
    [anon_sym_false] = ACTIONS(1801),
    [anon_sym_LBRACE] = ACTIONS(1801),
    [anon_sym_SEMI] = ACTIONS(1803),
    [anon_sym_RBRACE] = ACTIONS(1803),
    [anon_sym_constraint] = ACTIONS(1801),
    [anon_sym_val] = ACTIONS(1801),
    [anon_sym_end] = ACTIONS(1801),
    [anon_sym_with] = ACTIONS(1801),
    [anon_sym_COLON_GT] = ACTIONS(1803),
    [anon_sym_object] = ACTIONS(1801),
    [anon_sym_inherit] = ACTIONS(1801),
    [anon_sym_method] = ACTIONS(1801),
    [anon_sym_initializer] = ACTIONS(1801),
    [anon_sym_AMP] = ACTIONS(1801),
    [anon_sym_POUND] = ACTIONS(1801),
    [anon_sym_COLON_COLON] = ACTIONS(1803),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1803),
    [anon_sym_then] = ACTIONS(1801),
    [anon_sym_else] = ACTIONS(1801),
    [anon_sym_do] = ACTIONS(1801),
    [anon_sym_done] = ACTIONS(1801),
    [anon_sym_to] = ACTIONS(1801),
    [anon_sym_downto] = ACTIONS(1801),
    [anon_sym_new] = ACTIONS(1801),
    [anon_sym_LBRACE_LT] = ACTIONS(1803),
    [anon_sym_GT_RBRACE] = ACTIONS(1803),
    [anon_sym_begin] = ACTIONS(1801),
    [sym_ocamlyacc_value] = ACTIONS(1803),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1801),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1803),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1801),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1801),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1803),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1803),
    [aux_sym_number_token1] = ACTIONS(1803),
    [aux_sym_signed_number_token1] = ACTIONS(1801),
    [anon_sym_SQUOTE] = ACTIONS(1803),
    [anon_sym_DQUOTE] = ACTIONS(1803),
    [sym_prefix_operator] = ACTIONS(1803),
    [aux_sym_sign_operator_token1] = ACTIONS(1801),
    [sym_hash_operator] = ACTIONS(1803),
    [anon_sym_lsl] = ACTIONS(1801),
    [anon_sym_lsr] = ACTIONS(1801),
    [anon_sym_asr] = ACTIONS(1801),
    [aux_sym_pow_operator_token1] = ACTIONS(1803),
    [anon_sym_mod] = ACTIONS(1801),
    [anon_sym_land] = ACTIONS(1801),
    [anon_sym_lor] = ACTIONS(1801),
    [anon_sym_lxor] = ACTIONS(1801),
    [aux_sym_mult_operator_token1] = ACTIONS(1801),
    [aux_sym_add_operator_token1] = ACTIONS(1801),
    [sym_concat_operator] = ACTIONS(1803),
    [sym_rel_operator] = ACTIONS(1801),
    [anon_sym_AMP_AMP] = ACTIONS(1801),
    [anon_sym_or] = ACTIONS(1801),
    [anon_sym_PIPE_PIPE] = ACTIONS(1801),
    [sym_assign_operator] = ACTIONS(1803),
    [sym__uppercase_identifier] = ACTIONS(1803),
    [aux_sym_tag_token1] = ACTIONS(1803),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(908)] = {
    [sym__parenthesized_abstract_type] = STATE(1361),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4351),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(908),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_EQ] = ACTIONS(2075),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(2079),
    [anon_sym_RPAREN] = ACTIONS(2075),
    [anon_sym_COMMA] = ACTIONS(2075),
    [anon_sym_PIPE] = ACTIONS(2075),
    [anon_sym_COLON2] = ACTIONS(2081),
    [anon_sym_DASH_GT] = ACTIONS(2075),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_RBRACK] = ACTIONS(2075),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_as] = ACTIONS(2081),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(2075),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_when] = ACTIONS(2081),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2085),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(909)] = {
    [sym_attribute] = STATE(909),
    [sym__lowercase_identifier] = ACTIONS(1663),
    [anon_sym_TILDE] = ACTIONS(1663),
    [anon_sym_QMARK] = ACTIONS(1663),
    [anon_sym_LPAREN] = ACTIONS(1665),
    [anon_sym_RPAREN] = ACTIONS(1665),
    [anon_sym_COMMA] = ACTIONS(1665),
    [anon_sym_PIPE] = ACTIONS(1663),
    [anon_sym_COLON2] = ACTIONS(1663),
    [anon_sym_DOT] = ACTIONS(825),
    [anon_sym_DASH_GT] = ACTIONS(1663),
    [anon_sym_LBRACK] = ACTIONS(1663),
    [anon_sym_RBRACK] = ACTIONS(1665),
    [anon_sym_true] = ACTIONS(1663),
    [anon_sym_false] = ACTIONS(1663),
    [anon_sym_LBRACE] = ACTIONS(1663),
    [anon_sym_SEMI] = ACTIONS(1665),
    [anon_sym_RBRACE] = ACTIONS(1665),
    [anon_sym_constraint] = ACTIONS(1663),
    [anon_sym_val] = ACTIONS(1663),
    [anon_sym_end] = ACTIONS(1663),
    [anon_sym_with] = ACTIONS(1663),
    [anon_sym_COLON_GT] = ACTIONS(1665),
    [anon_sym_object] = ACTIONS(1663),
    [anon_sym_inherit] = ACTIONS(1663),
    [anon_sym_method] = ACTIONS(1663),
    [anon_sym_initializer] = ACTIONS(1663),
    [anon_sym_AMP] = ACTIONS(1663),
    [anon_sym_POUND] = ACTIONS(827),
    [anon_sym_COLON_COLON] = ACTIONS(1665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1665),
    [anon_sym_then] = ACTIONS(1663),
    [anon_sym_else] = ACTIONS(1663),
    [anon_sym_do] = ACTIONS(1663),
    [anon_sym_done] = ACTIONS(1663),
    [anon_sym_to] = ACTIONS(1663),
    [anon_sym_downto] = ACTIONS(1663),
    [anon_sym_new] = ACTIONS(1663),
    [anon_sym_LBRACE_LT] = ACTIONS(1665),
    [anon_sym_GT_RBRACE] = ACTIONS(1665),
    [anon_sym_begin] = ACTIONS(1663),
    [sym_ocamlyacc_value] = ACTIONS(1665),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1663),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1665),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1663),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1663),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1665),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1665),
    [aux_sym_number_token1] = ACTIONS(1665),
    [aux_sym_signed_number_token1] = ACTIONS(1663),
    [anon_sym_SQUOTE] = ACTIONS(1665),
    [anon_sym_DQUOTE] = ACTIONS(1665),
    [sym_prefix_operator] = ACTIONS(1665),
    [aux_sym_sign_operator_token1] = ACTIONS(1663),
    [sym_hash_operator] = ACTIONS(829),
    [anon_sym_lsl] = ACTIONS(1663),
    [anon_sym_lsr] = ACTIONS(1663),
    [anon_sym_asr] = ACTIONS(1663),
    [aux_sym_pow_operator_token1] = ACTIONS(1665),
    [anon_sym_mod] = ACTIONS(1663),
    [anon_sym_land] = ACTIONS(1663),
    [anon_sym_lor] = ACTIONS(1663),
    [anon_sym_lxor] = ACTIONS(1663),
    [aux_sym_mult_operator_token1] = ACTIONS(1663),
    [aux_sym_add_operator_token1] = ACTIONS(1663),
    [sym_concat_operator] = ACTIONS(1665),
    [sym_rel_operator] = ACTIONS(1663),
    [anon_sym_AMP_AMP] = ACTIONS(1663),
    [anon_sym_or] = ACTIONS(1663),
    [anon_sym_PIPE_PIPE] = ACTIONS(1663),
    [sym_assign_operator] = ACTIONS(1665),
    [sym__uppercase_identifier] = ACTIONS(1665),
    [aux_sym_tag_token1] = ACTIONS(1665),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(910)] = {
    [sym_attribute] = STATE(910),
    [sym__lowercase_identifier] = ACTIONS(1705),
    [anon_sym_TILDE] = ACTIONS(1705),
    [anon_sym_QMARK] = ACTIONS(1705),
    [anon_sym_LPAREN] = ACTIONS(1707),
    [anon_sym_RPAREN] = ACTIONS(1707),
    [anon_sym_COMMA] = ACTIONS(1707),
    [anon_sym_PIPE] = ACTIONS(1705),
    [anon_sym_COLON2] = ACTIONS(1705),
    [anon_sym_DOT] = ACTIONS(1707),
    [anon_sym_DASH_GT] = ACTIONS(1705),
    [anon_sym_LBRACK] = ACTIONS(1705),
    [anon_sym_RBRACK] = ACTIONS(1707),
    [anon_sym_true] = ACTIONS(1705),
    [anon_sym_false] = ACTIONS(1705),
    [anon_sym_LBRACE] = ACTIONS(1705),
    [anon_sym_SEMI] = ACTIONS(1707),
    [anon_sym_RBRACE] = ACTIONS(1707),
    [anon_sym_constraint] = ACTIONS(1705),
    [anon_sym_val] = ACTIONS(1705),
    [anon_sym_end] = ACTIONS(1705),
    [anon_sym_with] = ACTIONS(1705),
    [anon_sym_COLON_GT] = ACTIONS(1707),
    [anon_sym_object] = ACTIONS(1705),
    [anon_sym_inherit] = ACTIONS(1705),
    [anon_sym_method] = ACTIONS(1705),
    [anon_sym_initializer] = ACTIONS(1705),
    [anon_sym_AMP] = ACTIONS(1705),
    [anon_sym_POUND] = ACTIONS(1705),
    [anon_sym_COLON_COLON] = ACTIONS(1707),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1707),
    [anon_sym_then] = ACTIONS(1705),
    [anon_sym_else] = ACTIONS(1705),
    [anon_sym_do] = ACTIONS(1705),
    [anon_sym_done] = ACTIONS(1705),
    [anon_sym_to] = ACTIONS(1705),
    [anon_sym_downto] = ACTIONS(1705),
    [anon_sym_new] = ACTIONS(1705),
    [anon_sym_LBRACE_LT] = ACTIONS(1707),
    [anon_sym_GT_RBRACE] = ACTIONS(1707),
    [anon_sym_begin] = ACTIONS(1705),
    [sym_ocamlyacc_value] = ACTIONS(1707),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1705),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1707),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1705),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1705),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1707),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1707),
    [aux_sym_number_token1] = ACTIONS(1707),
    [aux_sym_signed_number_token1] = ACTIONS(1705),
    [anon_sym_SQUOTE] = ACTIONS(1707),
    [anon_sym_DQUOTE] = ACTIONS(1707),
    [sym_prefix_operator] = ACTIONS(1707),
    [aux_sym_sign_operator_token1] = ACTIONS(1705),
    [sym_hash_operator] = ACTIONS(1707),
    [anon_sym_lsl] = ACTIONS(1705),
    [anon_sym_lsr] = ACTIONS(1705),
    [anon_sym_asr] = ACTIONS(1705),
    [aux_sym_pow_operator_token1] = ACTIONS(1707),
    [anon_sym_mod] = ACTIONS(1705),
    [anon_sym_land] = ACTIONS(1705),
    [anon_sym_lor] = ACTIONS(1705),
    [anon_sym_lxor] = ACTIONS(1705),
    [aux_sym_mult_operator_token1] = ACTIONS(1705),
    [aux_sym_add_operator_token1] = ACTIONS(1705),
    [sym_concat_operator] = ACTIONS(1707),
    [sym_rel_operator] = ACTIONS(1705),
    [anon_sym_AMP_AMP] = ACTIONS(1705),
    [anon_sym_or] = ACTIONS(1705),
    [anon_sym_PIPE_PIPE] = ACTIONS(1705),
    [sym_assign_operator] = ACTIONS(1707),
    [sym__uppercase_identifier] = ACTIONS(1707),
    [aux_sym_tag_token1] = ACTIONS(1707),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(911)] = {
    [sym_attribute] = STATE(911),
    [sym__lowercase_identifier] = ACTIONS(1667),
    [anon_sym_TILDE] = ACTIONS(1667),
    [anon_sym_QMARK] = ACTIONS(1667),
    [anon_sym_LPAREN] = ACTIONS(1669),
    [anon_sym_RPAREN] = ACTIONS(1669),
    [anon_sym_COMMA] = ACTIONS(1669),
    [anon_sym_PIPE] = ACTIONS(1667),
    [anon_sym_COLON2] = ACTIONS(1667),
    [anon_sym_DOT] = ACTIONS(1669),
    [anon_sym_DASH_GT] = ACTIONS(1667),
    [anon_sym_LBRACK] = ACTIONS(1667),
    [anon_sym_RBRACK] = ACTIONS(1669),
    [anon_sym_true] = ACTIONS(1667),
    [anon_sym_false] = ACTIONS(1667),
    [anon_sym_LBRACE] = ACTIONS(1667),
    [anon_sym_SEMI] = ACTIONS(1669),
    [anon_sym_RBRACE] = ACTIONS(1669),
    [anon_sym_constraint] = ACTIONS(1667),
    [anon_sym_val] = ACTIONS(1667),
    [anon_sym_end] = ACTIONS(1667),
    [anon_sym_with] = ACTIONS(1667),
    [anon_sym_COLON_GT] = ACTIONS(1669),
    [anon_sym_object] = ACTIONS(1667),
    [anon_sym_inherit] = ACTIONS(1667),
    [anon_sym_method] = ACTIONS(1667),
    [anon_sym_initializer] = ACTIONS(1667),
    [anon_sym_AMP] = ACTIONS(1667),
    [anon_sym_POUND] = ACTIONS(1667),
    [anon_sym_COLON_COLON] = ACTIONS(1669),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1669),
    [anon_sym_then] = ACTIONS(1667),
    [anon_sym_else] = ACTIONS(1667),
    [anon_sym_do] = ACTIONS(1667),
    [anon_sym_done] = ACTIONS(1667),
    [anon_sym_to] = ACTIONS(1667),
    [anon_sym_downto] = ACTIONS(1667),
    [anon_sym_new] = ACTIONS(1667),
    [anon_sym_LBRACE_LT] = ACTIONS(1669),
    [anon_sym_GT_RBRACE] = ACTIONS(1669),
    [anon_sym_begin] = ACTIONS(1667),
    [sym_ocamlyacc_value] = ACTIONS(1669),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1667),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1669),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1667),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1667),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1669),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1669),
    [aux_sym_number_token1] = ACTIONS(1669),
    [aux_sym_signed_number_token1] = ACTIONS(1667),
    [anon_sym_SQUOTE] = ACTIONS(1669),
    [anon_sym_DQUOTE] = ACTIONS(1669),
    [sym_prefix_operator] = ACTIONS(1669),
    [aux_sym_sign_operator_token1] = ACTIONS(1667),
    [sym_hash_operator] = ACTIONS(1669),
    [anon_sym_lsl] = ACTIONS(1667),
    [anon_sym_lsr] = ACTIONS(1667),
    [anon_sym_asr] = ACTIONS(1667),
    [aux_sym_pow_operator_token1] = ACTIONS(1669),
    [anon_sym_mod] = ACTIONS(1667),
    [anon_sym_land] = ACTIONS(1667),
    [anon_sym_lor] = ACTIONS(1667),
    [anon_sym_lxor] = ACTIONS(1667),
    [aux_sym_mult_operator_token1] = ACTIONS(1667),
    [aux_sym_add_operator_token1] = ACTIONS(1667),
    [sym_concat_operator] = ACTIONS(1669),
    [sym_rel_operator] = ACTIONS(1667),
    [anon_sym_AMP_AMP] = ACTIONS(1667),
    [anon_sym_or] = ACTIONS(1667),
    [anon_sym_PIPE_PIPE] = ACTIONS(1667),
    [sym_assign_operator] = ACTIONS(1669),
    [sym__uppercase_identifier] = ACTIONS(1669),
    [aux_sym_tag_token1] = ACTIONS(1669),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(912)] = {
    [sym_attribute] = STATE(912),
    [sym__lowercase_identifier] = ACTIONS(1821),
    [anon_sym_TILDE] = ACTIONS(1821),
    [anon_sym_QMARK] = ACTIONS(1821),
    [anon_sym_LPAREN] = ACTIONS(1823),
    [anon_sym_RPAREN] = ACTIONS(1823),
    [anon_sym_COMMA] = ACTIONS(1823),
    [anon_sym_PIPE] = ACTIONS(1821),
    [anon_sym_COLON2] = ACTIONS(1821),
    [anon_sym_DOT] = ACTIONS(1823),
    [anon_sym_DASH_GT] = ACTIONS(1821),
    [anon_sym_LBRACK] = ACTIONS(1821),
    [anon_sym_RBRACK] = ACTIONS(1823),
    [anon_sym_true] = ACTIONS(1821),
    [anon_sym_false] = ACTIONS(1821),
    [anon_sym_LBRACE] = ACTIONS(1821),
    [anon_sym_SEMI] = ACTIONS(1823),
    [anon_sym_RBRACE] = ACTIONS(1823),
    [anon_sym_constraint] = ACTIONS(1821),
    [anon_sym_val] = ACTIONS(1821),
    [anon_sym_end] = ACTIONS(1821),
    [anon_sym_with] = ACTIONS(1821),
    [anon_sym_COLON_GT] = ACTIONS(1823),
    [anon_sym_object] = ACTIONS(1821),
    [anon_sym_inherit] = ACTIONS(1821),
    [anon_sym_method] = ACTIONS(1821),
    [anon_sym_initializer] = ACTIONS(1821),
    [anon_sym_AMP] = ACTIONS(1821),
    [anon_sym_POUND] = ACTIONS(1821),
    [anon_sym_COLON_COLON] = ACTIONS(1823),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1823),
    [anon_sym_then] = ACTIONS(1821),
    [anon_sym_else] = ACTIONS(1821),
    [anon_sym_do] = ACTIONS(1821),
    [anon_sym_done] = ACTIONS(1821),
    [anon_sym_to] = ACTIONS(1821),
    [anon_sym_downto] = ACTIONS(1821),
    [anon_sym_new] = ACTIONS(1821),
    [anon_sym_LBRACE_LT] = ACTIONS(1823),
    [anon_sym_GT_RBRACE] = ACTIONS(1823),
    [anon_sym_begin] = ACTIONS(1821),
    [sym_ocamlyacc_value] = ACTIONS(1823),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1821),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1823),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1821),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1821),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1823),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1823),
    [aux_sym_number_token1] = ACTIONS(1823),
    [aux_sym_signed_number_token1] = ACTIONS(1821),
    [anon_sym_SQUOTE] = ACTIONS(1823),
    [anon_sym_DQUOTE] = ACTIONS(1823),
    [sym_prefix_operator] = ACTIONS(1823),
    [aux_sym_sign_operator_token1] = ACTIONS(1821),
    [sym_hash_operator] = ACTIONS(1823),
    [anon_sym_lsl] = ACTIONS(1821),
    [anon_sym_lsr] = ACTIONS(1821),
    [anon_sym_asr] = ACTIONS(1821),
    [aux_sym_pow_operator_token1] = ACTIONS(1823),
    [anon_sym_mod] = ACTIONS(1821),
    [anon_sym_land] = ACTIONS(1821),
    [anon_sym_lor] = ACTIONS(1821),
    [anon_sym_lxor] = ACTIONS(1821),
    [aux_sym_mult_operator_token1] = ACTIONS(1821),
    [aux_sym_add_operator_token1] = ACTIONS(1821),
    [sym_concat_operator] = ACTIONS(1823),
    [sym_rel_operator] = ACTIONS(1821),
    [anon_sym_AMP_AMP] = ACTIONS(1821),
    [anon_sym_or] = ACTIONS(1821),
    [anon_sym_PIPE_PIPE] = ACTIONS(1821),
    [sym_assign_operator] = ACTIONS(1823),
    [sym__uppercase_identifier] = ACTIONS(1823),
    [aux_sym_tag_token1] = ACTIONS(1823),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(913)] = {
    [sym_attribute] = STATE(913),
    [sym__lowercase_identifier] = ACTIONS(1575),
    [anon_sym_TILDE] = ACTIONS(1575),
    [anon_sym_QMARK] = ACTIONS(1575),
    [anon_sym_LPAREN] = ACTIONS(1577),
    [anon_sym_RPAREN] = ACTIONS(1577),
    [anon_sym_COMMA] = ACTIONS(1577),
    [anon_sym_PIPE] = ACTIONS(1575),
    [anon_sym_COLON2] = ACTIONS(1575),
    [anon_sym_DOT] = ACTIONS(1577),
    [anon_sym_DASH_GT] = ACTIONS(1575),
    [anon_sym_LBRACK] = ACTIONS(1575),
    [anon_sym_RBRACK] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(1575),
    [anon_sym_false] = ACTIONS(1575),
    [anon_sym_LBRACE] = ACTIONS(1575),
    [anon_sym_SEMI] = ACTIONS(1577),
    [anon_sym_RBRACE] = ACTIONS(1577),
    [anon_sym_constraint] = ACTIONS(1575),
    [anon_sym_val] = ACTIONS(1575),
    [anon_sym_end] = ACTIONS(1575),
    [anon_sym_with] = ACTIONS(1575),
    [anon_sym_COLON_GT] = ACTIONS(1577),
    [anon_sym_object] = ACTIONS(1575),
    [anon_sym_inherit] = ACTIONS(1575),
    [anon_sym_method] = ACTIONS(1575),
    [anon_sym_initializer] = ACTIONS(1575),
    [anon_sym_AMP] = ACTIONS(1575),
    [anon_sym_POUND] = ACTIONS(1575),
    [anon_sym_COLON_COLON] = ACTIONS(1577),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1577),
    [anon_sym_then] = ACTIONS(1575),
    [anon_sym_else] = ACTIONS(1575),
    [anon_sym_do] = ACTIONS(1575),
    [anon_sym_done] = ACTIONS(1575),
    [anon_sym_to] = ACTIONS(1575),
    [anon_sym_downto] = ACTIONS(1575),
    [anon_sym_new] = ACTIONS(1575),
    [anon_sym_LBRACE_LT] = ACTIONS(1577),
    [anon_sym_GT_RBRACE] = ACTIONS(1577),
    [anon_sym_begin] = ACTIONS(1575),
    [sym_ocamlyacc_value] = ACTIONS(1577),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1575),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1577),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1575),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1575),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1577),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1577),
    [aux_sym_number_token1] = ACTIONS(1577),
    [aux_sym_signed_number_token1] = ACTIONS(1575),
    [anon_sym_SQUOTE] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(1577),
    [sym_prefix_operator] = ACTIONS(1577),
    [aux_sym_sign_operator_token1] = ACTIONS(1575),
    [sym_hash_operator] = ACTIONS(1577),
    [anon_sym_lsl] = ACTIONS(1575),
    [anon_sym_lsr] = ACTIONS(1575),
    [anon_sym_asr] = ACTIONS(1575),
    [aux_sym_pow_operator_token1] = ACTIONS(1577),
    [anon_sym_mod] = ACTIONS(1575),
    [anon_sym_land] = ACTIONS(1575),
    [anon_sym_lor] = ACTIONS(1575),
    [anon_sym_lxor] = ACTIONS(1575),
    [aux_sym_mult_operator_token1] = ACTIONS(1575),
    [aux_sym_add_operator_token1] = ACTIONS(1575),
    [sym_concat_operator] = ACTIONS(1577),
    [sym_rel_operator] = ACTIONS(1575),
    [anon_sym_AMP_AMP] = ACTIONS(1575),
    [anon_sym_or] = ACTIONS(1575),
    [anon_sym_PIPE_PIPE] = ACTIONS(1575),
    [sym_assign_operator] = ACTIONS(1577),
    [sym__uppercase_identifier] = ACTIONS(1577),
    [aux_sym_tag_token1] = ACTIONS(1577),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(914)] = {
    [sym_attribute] = STATE(914),
    [sym__lowercase_identifier] = ACTIONS(1595),
    [anon_sym_TILDE] = ACTIONS(1595),
    [anon_sym_QMARK] = ACTIONS(1595),
    [anon_sym_LPAREN] = ACTIONS(1597),
    [anon_sym_RPAREN] = ACTIONS(1597),
    [anon_sym_COMMA] = ACTIONS(1597),
    [anon_sym_PIPE] = ACTIONS(1595),
    [anon_sym_COLON2] = ACTIONS(1595),
    [anon_sym_DOT] = ACTIONS(1597),
    [anon_sym_DASH_GT] = ACTIONS(1595),
    [anon_sym_LBRACK] = ACTIONS(1595),
    [anon_sym_RBRACK] = ACTIONS(1597),
    [anon_sym_true] = ACTIONS(1595),
    [anon_sym_false] = ACTIONS(1595),
    [anon_sym_LBRACE] = ACTIONS(1595),
    [anon_sym_SEMI] = ACTIONS(1597),
    [anon_sym_RBRACE] = ACTIONS(1597),
    [anon_sym_constraint] = ACTIONS(1595),
    [anon_sym_val] = ACTIONS(1595),
    [anon_sym_end] = ACTIONS(1595),
    [anon_sym_with] = ACTIONS(1595),
    [anon_sym_COLON_GT] = ACTIONS(1597),
    [anon_sym_object] = ACTIONS(1595),
    [anon_sym_inherit] = ACTIONS(1595),
    [anon_sym_method] = ACTIONS(1595),
    [anon_sym_initializer] = ACTIONS(1595),
    [anon_sym_AMP] = ACTIONS(1595),
    [anon_sym_POUND] = ACTIONS(1595),
    [anon_sym_COLON_COLON] = ACTIONS(1597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1597),
    [anon_sym_then] = ACTIONS(1595),
    [anon_sym_else] = ACTIONS(1595),
    [anon_sym_do] = ACTIONS(1595),
    [anon_sym_done] = ACTIONS(1595),
    [anon_sym_to] = ACTIONS(1595),
    [anon_sym_downto] = ACTIONS(1595),
    [anon_sym_new] = ACTIONS(1595),
    [anon_sym_LBRACE_LT] = ACTIONS(1597),
    [anon_sym_GT_RBRACE] = ACTIONS(1597),
    [anon_sym_begin] = ACTIONS(1595),
    [sym_ocamlyacc_value] = ACTIONS(1597),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1595),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1597),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1595),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1595),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1597),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1597),
    [aux_sym_number_token1] = ACTIONS(1597),
    [aux_sym_signed_number_token1] = ACTIONS(1595),
    [anon_sym_SQUOTE] = ACTIONS(1597),
    [anon_sym_DQUOTE] = ACTIONS(1597),
    [sym_prefix_operator] = ACTIONS(1597),
    [aux_sym_sign_operator_token1] = ACTIONS(1595),
    [sym_hash_operator] = ACTIONS(1597),
    [anon_sym_lsl] = ACTIONS(1595),
    [anon_sym_lsr] = ACTIONS(1595),
    [anon_sym_asr] = ACTIONS(1595),
    [aux_sym_pow_operator_token1] = ACTIONS(1597),
    [anon_sym_mod] = ACTIONS(1595),
    [anon_sym_land] = ACTIONS(1595),
    [anon_sym_lor] = ACTIONS(1595),
    [anon_sym_lxor] = ACTIONS(1595),
    [aux_sym_mult_operator_token1] = ACTIONS(1595),
    [aux_sym_add_operator_token1] = ACTIONS(1595),
    [sym_concat_operator] = ACTIONS(1597),
    [sym_rel_operator] = ACTIONS(1595),
    [anon_sym_AMP_AMP] = ACTIONS(1595),
    [anon_sym_or] = ACTIONS(1595),
    [anon_sym_PIPE_PIPE] = ACTIONS(1595),
    [sym_assign_operator] = ACTIONS(1597),
    [sym__uppercase_identifier] = ACTIONS(1597),
    [aux_sym_tag_token1] = ACTIONS(1597),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(915)] = {
    [sym_attribute] = STATE(915),
    [sym__lowercase_identifier] = ACTIONS(1611),
    [anon_sym_TILDE] = ACTIONS(1611),
    [anon_sym_QMARK] = ACTIONS(1611),
    [anon_sym_LPAREN] = ACTIONS(1613),
    [anon_sym_RPAREN] = ACTIONS(1613),
    [anon_sym_COMMA] = ACTIONS(1613),
    [anon_sym_PIPE] = ACTIONS(1611),
    [anon_sym_COLON2] = ACTIONS(1611),
    [anon_sym_DOT] = ACTIONS(1613),
    [anon_sym_DASH_GT] = ACTIONS(1611),
    [anon_sym_LBRACK] = ACTIONS(1611),
    [anon_sym_RBRACK] = ACTIONS(1613),
    [anon_sym_true] = ACTIONS(1611),
    [anon_sym_false] = ACTIONS(1611),
    [anon_sym_LBRACE] = ACTIONS(1611),
    [anon_sym_SEMI] = ACTIONS(1613),
    [anon_sym_RBRACE] = ACTIONS(1613),
    [anon_sym_constraint] = ACTIONS(1611),
    [anon_sym_val] = ACTIONS(1611),
    [anon_sym_end] = ACTIONS(1611),
    [anon_sym_with] = ACTIONS(1611),
    [anon_sym_COLON_GT] = ACTIONS(1613),
    [anon_sym_object] = ACTIONS(1611),
    [anon_sym_inherit] = ACTIONS(1611),
    [anon_sym_method] = ACTIONS(1611),
    [anon_sym_initializer] = ACTIONS(1611),
    [anon_sym_AMP] = ACTIONS(1611),
    [anon_sym_POUND] = ACTIONS(1611),
    [anon_sym_COLON_COLON] = ACTIONS(1613),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1613),
    [anon_sym_then] = ACTIONS(1611),
    [anon_sym_else] = ACTIONS(1611),
    [anon_sym_do] = ACTIONS(1611),
    [anon_sym_done] = ACTIONS(1611),
    [anon_sym_to] = ACTIONS(1611),
    [anon_sym_downto] = ACTIONS(1611),
    [anon_sym_new] = ACTIONS(1611),
    [anon_sym_LBRACE_LT] = ACTIONS(1613),
    [anon_sym_GT_RBRACE] = ACTIONS(1613),
    [anon_sym_begin] = ACTIONS(1611),
    [sym_ocamlyacc_value] = ACTIONS(1613),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1611),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1613),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1611),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1611),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1613),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1613),
    [aux_sym_number_token1] = ACTIONS(1613),
    [aux_sym_signed_number_token1] = ACTIONS(1611),
    [anon_sym_SQUOTE] = ACTIONS(1613),
    [anon_sym_DQUOTE] = ACTIONS(1613),
    [sym_prefix_operator] = ACTIONS(1613),
    [aux_sym_sign_operator_token1] = ACTIONS(1611),
    [sym_hash_operator] = ACTIONS(1613),
    [anon_sym_lsl] = ACTIONS(1611),
    [anon_sym_lsr] = ACTIONS(1611),
    [anon_sym_asr] = ACTIONS(1611),
    [aux_sym_pow_operator_token1] = ACTIONS(1613),
    [anon_sym_mod] = ACTIONS(1611),
    [anon_sym_land] = ACTIONS(1611),
    [anon_sym_lor] = ACTIONS(1611),
    [anon_sym_lxor] = ACTIONS(1611),
    [aux_sym_mult_operator_token1] = ACTIONS(1611),
    [aux_sym_add_operator_token1] = ACTIONS(1611),
    [sym_concat_operator] = ACTIONS(1613),
    [sym_rel_operator] = ACTIONS(1611),
    [anon_sym_AMP_AMP] = ACTIONS(1611),
    [anon_sym_or] = ACTIONS(1611),
    [anon_sym_PIPE_PIPE] = ACTIONS(1611),
    [sym_assign_operator] = ACTIONS(1613),
    [sym__uppercase_identifier] = ACTIONS(1613),
    [aux_sym_tag_token1] = ACTIONS(1613),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(916)] = {
    [sym_attribute] = STATE(916),
    [sym__lowercase_identifier] = ACTIONS(1671),
    [anon_sym_TILDE] = ACTIONS(1671),
    [anon_sym_QMARK] = ACTIONS(1671),
    [anon_sym_LPAREN] = ACTIONS(1673),
    [anon_sym_RPAREN] = ACTIONS(1673),
    [anon_sym_COMMA] = ACTIONS(1673),
    [anon_sym_PIPE] = ACTIONS(1671),
    [anon_sym_COLON2] = ACTIONS(1671),
    [anon_sym_DOT] = ACTIONS(1673),
    [anon_sym_DASH_GT] = ACTIONS(1671),
    [anon_sym_LBRACK] = ACTIONS(1671),
    [anon_sym_RBRACK] = ACTIONS(1673),
    [anon_sym_true] = ACTIONS(1671),
    [anon_sym_false] = ACTIONS(1671),
    [anon_sym_LBRACE] = ACTIONS(1671),
    [anon_sym_SEMI] = ACTIONS(1673),
    [anon_sym_RBRACE] = ACTIONS(1673),
    [anon_sym_constraint] = ACTIONS(1671),
    [anon_sym_val] = ACTIONS(1671),
    [anon_sym_end] = ACTIONS(1671),
    [anon_sym_with] = ACTIONS(1671),
    [anon_sym_COLON_GT] = ACTIONS(1673),
    [anon_sym_object] = ACTIONS(1671),
    [anon_sym_inherit] = ACTIONS(1671),
    [anon_sym_method] = ACTIONS(1671),
    [anon_sym_initializer] = ACTIONS(1671),
    [anon_sym_AMP] = ACTIONS(1671),
    [anon_sym_POUND] = ACTIONS(1671),
    [anon_sym_COLON_COLON] = ACTIONS(1673),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1673),
    [anon_sym_then] = ACTIONS(1671),
    [anon_sym_else] = ACTIONS(1671),
    [anon_sym_do] = ACTIONS(1671),
    [anon_sym_done] = ACTIONS(1671),
    [anon_sym_to] = ACTIONS(1671),
    [anon_sym_downto] = ACTIONS(1671),
    [anon_sym_new] = ACTIONS(1671),
    [anon_sym_LBRACE_LT] = ACTIONS(1673),
    [anon_sym_GT_RBRACE] = ACTIONS(1673),
    [anon_sym_begin] = ACTIONS(1671),
    [sym_ocamlyacc_value] = ACTIONS(1673),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1671),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1673),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1671),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1671),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1673),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1673),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1671),
    [anon_sym_SQUOTE] = ACTIONS(1673),
    [anon_sym_DQUOTE] = ACTIONS(1673),
    [sym_prefix_operator] = ACTIONS(1673),
    [aux_sym_sign_operator_token1] = ACTIONS(1671),
    [sym_hash_operator] = ACTIONS(1673),
    [anon_sym_lsl] = ACTIONS(1671),
    [anon_sym_lsr] = ACTIONS(1671),
    [anon_sym_asr] = ACTIONS(1671),
    [aux_sym_pow_operator_token1] = ACTIONS(1673),
    [anon_sym_mod] = ACTIONS(1671),
    [anon_sym_land] = ACTIONS(1671),
    [anon_sym_lor] = ACTIONS(1671),
    [anon_sym_lxor] = ACTIONS(1671),
    [aux_sym_mult_operator_token1] = ACTIONS(1671),
    [aux_sym_add_operator_token1] = ACTIONS(1671),
    [sym_concat_operator] = ACTIONS(1673),
    [sym_rel_operator] = ACTIONS(1671),
    [anon_sym_AMP_AMP] = ACTIONS(1671),
    [anon_sym_or] = ACTIONS(1671),
    [anon_sym_PIPE_PIPE] = ACTIONS(1671),
    [sym_assign_operator] = ACTIONS(1673),
    [sym__uppercase_identifier] = ACTIONS(1673),
    [aux_sym_tag_token1] = ACTIONS(1673),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(917)] = {
    [sym_attribute] = STATE(917),
    [sym__lowercase_identifier] = ACTIONS(1675),
    [anon_sym_TILDE] = ACTIONS(1675),
    [anon_sym_QMARK] = ACTIONS(1675),
    [anon_sym_LPAREN] = ACTIONS(1677),
    [anon_sym_RPAREN] = ACTIONS(1677),
    [anon_sym_COMMA] = ACTIONS(1677),
    [anon_sym_PIPE] = ACTIONS(1675),
    [anon_sym_COLON2] = ACTIONS(1675),
    [anon_sym_DOT] = ACTIONS(1677),
    [anon_sym_DASH_GT] = ACTIONS(1675),
    [anon_sym_LBRACK] = ACTIONS(1675),
    [anon_sym_RBRACK] = ACTIONS(1677),
    [anon_sym_true] = ACTIONS(1675),
    [anon_sym_false] = ACTIONS(1675),
    [anon_sym_LBRACE] = ACTIONS(1675),
    [anon_sym_SEMI] = ACTIONS(1677),
    [anon_sym_RBRACE] = ACTIONS(1677),
    [anon_sym_constraint] = ACTIONS(1675),
    [anon_sym_val] = ACTIONS(1675),
    [anon_sym_end] = ACTIONS(1675),
    [anon_sym_with] = ACTIONS(1675),
    [anon_sym_COLON_GT] = ACTIONS(1677),
    [anon_sym_object] = ACTIONS(1675),
    [anon_sym_inherit] = ACTIONS(1675),
    [anon_sym_method] = ACTIONS(1675),
    [anon_sym_initializer] = ACTIONS(1675),
    [anon_sym_AMP] = ACTIONS(1675),
    [anon_sym_POUND] = ACTIONS(1675),
    [anon_sym_COLON_COLON] = ACTIONS(1677),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1677),
    [anon_sym_then] = ACTIONS(1675),
    [anon_sym_else] = ACTIONS(1675),
    [anon_sym_do] = ACTIONS(1675),
    [anon_sym_done] = ACTIONS(1675),
    [anon_sym_to] = ACTIONS(1675),
    [anon_sym_downto] = ACTIONS(1675),
    [anon_sym_new] = ACTIONS(1675),
    [anon_sym_LBRACE_LT] = ACTIONS(1677),
    [anon_sym_GT_RBRACE] = ACTIONS(1677),
    [anon_sym_begin] = ACTIONS(1675),
    [sym_ocamlyacc_value] = ACTIONS(1677),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1675),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1677),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1675),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1675),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1677),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1677),
    [aux_sym_number_token1] = ACTIONS(1677),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1677),
    [sym_prefix_operator] = ACTIONS(1677),
    [aux_sym_sign_operator_token1] = ACTIONS(1675),
    [sym_hash_operator] = ACTIONS(1677),
    [anon_sym_lsl] = ACTIONS(1675),
    [anon_sym_lsr] = ACTIONS(1675),
    [anon_sym_asr] = ACTIONS(1675),
    [aux_sym_pow_operator_token1] = ACTIONS(1677),
    [anon_sym_mod] = ACTIONS(1675),
    [anon_sym_land] = ACTIONS(1675),
    [anon_sym_lor] = ACTIONS(1675),
    [anon_sym_lxor] = ACTIONS(1675),
    [aux_sym_mult_operator_token1] = ACTIONS(1675),
    [aux_sym_add_operator_token1] = ACTIONS(1675),
    [sym_concat_operator] = ACTIONS(1677),
    [sym_rel_operator] = ACTIONS(1675),
    [anon_sym_AMP_AMP] = ACTIONS(1675),
    [anon_sym_or] = ACTIONS(1675),
    [anon_sym_PIPE_PIPE] = ACTIONS(1675),
    [sym_assign_operator] = ACTIONS(1677),
    [sym__uppercase_identifier] = ACTIONS(1677),
    [aux_sym_tag_token1] = ACTIONS(1677),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(918)] = {
    [sym_attribute] = STATE(918),
    [sym__lowercase_identifier] = ACTIONS(1679),
    [anon_sym_TILDE] = ACTIONS(1679),
    [anon_sym_QMARK] = ACTIONS(1679),
    [anon_sym_LPAREN] = ACTIONS(1681),
    [anon_sym_RPAREN] = ACTIONS(1681),
    [anon_sym_COMMA] = ACTIONS(1681),
    [anon_sym_PIPE] = ACTIONS(1679),
    [anon_sym_COLON2] = ACTIONS(1679),
    [anon_sym_DOT] = ACTIONS(1681),
    [anon_sym_DASH_GT] = ACTIONS(1679),
    [anon_sym_LBRACK] = ACTIONS(1679),
    [anon_sym_RBRACK] = ACTIONS(1681),
    [anon_sym_true] = ACTIONS(1679),
    [anon_sym_false] = ACTIONS(1679),
    [anon_sym_LBRACE] = ACTIONS(1679),
    [anon_sym_SEMI] = ACTIONS(1681),
    [anon_sym_RBRACE] = ACTIONS(1681),
    [anon_sym_constraint] = ACTIONS(1679),
    [anon_sym_val] = ACTIONS(1679),
    [anon_sym_end] = ACTIONS(1679),
    [anon_sym_with] = ACTIONS(1679),
    [anon_sym_COLON_GT] = ACTIONS(1681),
    [anon_sym_object] = ACTIONS(1679),
    [anon_sym_inherit] = ACTIONS(1679),
    [anon_sym_method] = ACTIONS(1679),
    [anon_sym_initializer] = ACTIONS(1679),
    [anon_sym_AMP] = ACTIONS(1679),
    [anon_sym_POUND] = ACTIONS(1679),
    [anon_sym_COLON_COLON] = ACTIONS(1681),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1681),
    [anon_sym_then] = ACTIONS(1679),
    [anon_sym_else] = ACTIONS(1679),
    [anon_sym_do] = ACTIONS(1679),
    [anon_sym_done] = ACTIONS(1679),
    [anon_sym_to] = ACTIONS(1679),
    [anon_sym_downto] = ACTIONS(1679),
    [anon_sym_new] = ACTIONS(1679),
    [anon_sym_LBRACE_LT] = ACTIONS(1681),
    [anon_sym_GT_RBRACE] = ACTIONS(1681),
    [anon_sym_begin] = ACTIONS(1679),
    [sym_ocamlyacc_value] = ACTIONS(1681),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1679),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1681),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1679),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1679),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1681),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1681),
    [aux_sym_number_token1] = ACTIONS(1681),
    [aux_sym_signed_number_token1] = ACTIONS(1679),
    [anon_sym_SQUOTE] = ACTIONS(1681),
    [anon_sym_DQUOTE] = ACTIONS(1681),
    [sym_prefix_operator] = ACTIONS(1681),
    [aux_sym_sign_operator_token1] = ACTIONS(1679),
    [sym_hash_operator] = ACTIONS(1681),
    [anon_sym_lsl] = ACTIONS(1679),
    [anon_sym_lsr] = ACTIONS(1679),
    [anon_sym_asr] = ACTIONS(1679),
    [aux_sym_pow_operator_token1] = ACTIONS(1681),
    [anon_sym_mod] = ACTIONS(1679),
    [anon_sym_land] = ACTIONS(1679),
    [anon_sym_lor] = ACTIONS(1679),
    [anon_sym_lxor] = ACTIONS(1679),
    [aux_sym_mult_operator_token1] = ACTIONS(1679),
    [aux_sym_add_operator_token1] = ACTIONS(1679),
    [sym_concat_operator] = ACTIONS(1681),
    [sym_rel_operator] = ACTIONS(1679),
    [anon_sym_AMP_AMP] = ACTIONS(1679),
    [anon_sym_or] = ACTIONS(1679),
    [anon_sym_PIPE_PIPE] = ACTIONS(1679),
    [sym_assign_operator] = ACTIONS(1681),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_tag_token1] = ACTIONS(1681),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(919)] = {
    [sym_attribute] = STATE(919),
    [sym__lowercase_identifier] = ACTIONS(1851),
    [anon_sym_TILDE] = ACTIONS(1851),
    [anon_sym_QMARK] = ACTIONS(1851),
    [anon_sym_LPAREN] = ACTIONS(1853),
    [anon_sym_RPAREN] = ACTIONS(1853),
    [anon_sym_COMMA] = ACTIONS(1853),
    [anon_sym_PIPE] = ACTIONS(1851),
    [anon_sym_COLON2] = ACTIONS(1851),
    [anon_sym_DOT] = ACTIONS(1853),
    [anon_sym_DASH_GT] = ACTIONS(1851),
    [anon_sym_LBRACK] = ACTIONS(1851),
    [anon_sym_RBRACK] = ACTIONS(1853),
    [anon_sym_true] = ACTIONS(1851),
    [anon_sym_false] = ACTIONS(1851),
    [anon_sym_LBRACE] = ACTIONS(1851),
    [anon_sym_SEMI] = ACTIONS(1853),
    [anon_sym_RBRACE] = ACTIONS(1853),
    [anon_sym_constraint] = ACTIONS(1851),
    [anon_sym_val] = ACTIONS(1851),
    [anon_sym_end] = ACTIONS(1851),
    [anon_sym_with] = ACTIONS(1851),
    [anon_sym_COLON_GT] = ACTIONS(1853),
    [anon_sym_object] = ACTIONS(1851),
    [anon_sym_inherit] = ACTIONS(1851),
    [anon_sym_method] = ACTIONS(1851),
    [anon_sym_initializer] = ACTIONS(1851),
    [anon_sym_AMP] = ACTIONS(1851),
    [anon_sym_POUND] = ACTIONS(1851),
    [anon_sym_COLON_COLON] = ACTIONS(1853),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1853),
    [anon_sym_then] = ACTIONS(1851),
    [anon_sym_else] = ACTIONS(1851),
    [anon_sym_do] = ACTIONS(1851),
    [anon_sym_done] = ACTIONS(1851),
    [anon_sym_to] = ACTIONS(1851),
    [anon_sym_downto] = ACTIONS(1851),
    [anon_sym_new] = ACTIONS(1851),
    [anon_sym_LBRACE_LT] = ACTIONS(1853),
    [anon_sym_GT_RBRACE] = ACTIONS(1853),
    [anon_sym_begin] = ACTIONS(1851),
    [sym_ocamlyacc_value] = ACTIONS(1853),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1851),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1853),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1851),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1851),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1853),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1853),
    [aux_sym_number_token1] = ACTIONS(1853),
    [aux_sym_signed_number_token1] = ACTIONS(1851),
    [anon_sym_SQUOTE] = ACTIONS(1853),
    [anon_sym_DQUOTE] = ACTIONS(1853),
    [sym_prefix_operator] = ACTIONS(1853),
    [aux_sym_sign_operator_token1] = ACTIONS(1851),
    [sym_hash_operator] = ACTIONS(1853),
    [anon_sym_lsl] = ACTIONS(1851),
    [anon_sym_lsr] = ACTIONS(1851),
    [anon_sym_asr] = ACTIONS(1851),
    [aux_sym_pow_operator_token1] = ACTIONS(1853),
    [anon_sym_mod] = ACTIONS(1851),
    [anon_sym_land] = ACTIONS(1851),
    [anon_sym_lor] = ACTIONS(1851),
    [anon_sym_lxor] = ACTIONS(1851),
    [aux_sym_mult_operator_token1] = ACTIONS(1851),
    [aux_sym_add_operator_token1] = ACTIONS(1851),
    [sym_concat_operator] = ACTIONS(1853),
    [sym_rel_operator] = ACTIONS(1851),
    [anon_sym_AMP_AMP] = ACTIONS(1851),
    [anon_sym_or] = ACTIONS(1851),
    [anon_sym_PIPE_PIPE] = ACTIONS(1851),
    [sym_assign_operator] = ACTIONS(1853),
    [sym__uppercase_identifier] = ACTIONS(1853),
    [aux_sym_tag_token1] = ACTIONS(1853),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(920)] = {
    [sym_attribute] = STATE(920),
    [sym__lowercase_identifier] = ACTIONS(1717),
    [anon_sym_TILDE] = ACTIONS(1717),
    [anon_sym_QMARK] = ACTIONS(1717),
    [anon_sym_LPAREN] = ACTIONS(1719),
    [anon_sym_RPAREN] = ACTIONS(1719),
    [anon_sym_COMMA] = ACTIONS(1719),
    [anon_sym_PIPE] = ACTIONS(1717),
    [anon_sym_COLON2] = ACTIONS(1717),
    [anon_sym_DOT] = ACTIONS(1719),
    [anon_sym_DASH_GT] = ACTIONS(1717),
    [anon_sym_LBRACK] = ACTIONS(1717),
    [anon_sym_RBRACK] = ACTIONS(1719),
    [anon_sym_true] = ACTIONS(1717),
    [anon_sym_false] = ACTIONS(1717),
    [anon_sym_LBRACE] = ACTIONS(1717),
    [anon_sym_SEMI] = ACTIONS(1719),
    [anon_sym_RBRACE] = ACTIONS(1719),
    [anon_sym_constraint] = ACTIONS(1717),
    [anon_sym_val] = ACTIONS(1717),
    [anon_sym_end] = ACTIONS(1717),
    [anon_sym_with] = ACTIONS(1717),
    [anon_sym_COLON_GT] = ACTIONS(1719),
    [anon_sym_object] = ACTIONS(1717),
    [anon_sym_inherit] = ACTIONS(1717),
    [anon_sym_method] = ACTIONS(1717),
    [anon_sym_initializer] = ACTIONS(1717),
    [anon_sym_AMP] = ACTIONS(1717),
    [anon_sym_POUND] = ACTIONS(1717),
    [anon_sym_COLON_COLON] = ACTIONS(1719),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1719),
    [anon_sym_then] = ACTIONS(1717),
    [anon_sym_else] = ACTIONS(1717),
    [anon_sym_do] = ACTIONS(1717),
    [anon_sym_done] = ACTIONS(1717),
    [anon_sym_to] = ACTIONS(1717),
    [anon_sym_downto] = ACTIONS(1717),
    [anon_sym_new] = ACTIONS(1717),
    [anon_sym_LBRACE_LT] = ACTIONS(1719),
    [anon_sym_GT_RBRACE] = ACTIONS(1719),
    [anon_sym_begin] = ACTIONS(1717),
    [sym_ocamlyacc_value] = ACTIONS(1719),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1717),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1719),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1717),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1717),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1719),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1719),
    [aux_sym_number_token1] = ACTIONS(1719),
    [aux_sym_signed_number_token1] = ACTIONS(1717),
    [anon_sym_SQUOTE] = ACTIONS(1719),
    [anon_sym_DQUOTE] = ACTIONS(1719),
    [sym_prefix_operator] = ACTIONS(1719),
    [aux_sym_sign_operator_token1] = ACTIONS(1717),
    [sym_hash_operator] = ACTIONS(1719),
    [anon_sym_lsl] = ACTIONS(1717),
    [anon_sym_lsr] = ACTIONS(1717),
    [anon_sym_asr] = ACTIONS(1717),
    [aux_sym_pow_operator_token1] = ACTIONS(1719),
    [anon_sym_mod] = ACTIONS(1717),
    [anon_sym_land] = ACTIONS(1717),
    [anon_sym_lor] = ACTIONS(1717),
    [anon_sym_lxor] = ACTIONS(1717),
    [aux_sym_mult_operator_token1] = ACTIONS(1717),
    [aux_sym_add_operator_token1] = ACTIONS(1717),
    [sym_concat_operator] = ACTIONS(1719),
    [sym_rel_operator] = ACTIONS(1717),
    [anon_sym_AMP_AMP] = ACTIONS(1717),
    [anon_sym_or] = ACTIONS(1717),
    [anon_sym_PIPE_PIPE] = ACTIONS(1717),
    [sym_assign_operator] = ACTIONS(1719),
    [sym__uppercase_identifier] = ACTIONS(1719),
    [aux_sym_tag_token1] = ACTIONS(1719),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(921)] = {
    [sym_attribute] = STATE(921),
    [sym__lowercase_identifier] = ACTIONS(1837),
    [anon_sym_TILDE] = ACTIONS(1837),
    [anon_sym_QMARK] = ACTIONS(1837),
    [anon_sym_LPAREN] = ACTIONS(1839),
    [anon_sym_RPAREN] = ACTIONS(1839),
    [anon_sym_COMMA] = ACTIONS(1839),
    [anon_sym_PIPE] = ACTIONS(1837),
    [anon_sym_COLON2] = ACTIONS(1837),
    [anon_sym_DOT] = ACTIONS(1839),
    [anon_sym_DASH_GT] = ACTIONS(1837),
    [anon_sym_LBRACK] = ACTIONS(1837),
    [anon_sym_RBRACK] = ACTIONS(1839),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1837),
    [anon_sym_SEMI] = ACTIONS(1839),
    [anon_sym_RBRACE] = ACTIONS(1839),
    [anon_sym_constraint] = ACTIONS(1837),
    [anon_sym_val] = ACTIONS(1837),
    [anon_sym_end] = ACTIONS(1837),
    [anon_sym_with] = ACTIONS(1837),
    [anon_sym_COLON_GT] = ACTIONS(1839),
    [anon_sym_object] = ACTIONS(1837),
    [anon_sym_inherit] = ACTIONS(1837),
    [anon_sym_method] = ACTIONS(1837),
    [anon_sym_initializer] = ACTIONS(1837),
    [anon_sym_AMP] = ACTIONS(1837),
    [anon_sym_POUND] = ACTIONS(1837),
    [anon_sym_COLON_COLON] = ACTIONS(1839),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1839),
    [anon_sym_then] = ACTIONS(1837),
    [anon_sym_else] = ACTIONS(1837),
    [anon_sym_do] = ACTIONS(1837),
    [anon_sym_done] = ACTIONS(1837),
    [anon_sym_to] = ACTIONS(1837),
    [anon_sym_downto] = ACTIONS(1837),
    [anon_sym_new] = ACTIONS(1837),
    [anon_sym_LBRACE_LT] = ACTIONS(1839),
    [anon_sym_GT_RBRACE] = ACTIONS(1839),
    [anon_sym_begin] = ACTIONS(1837),
    [sym_ocamlyacc_value] = ACTIONS(1839),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1837),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1839),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1837),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1837),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1839),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1839),
    [aux_sym_number_token1] = ACTIONS(1839),
    [aux_sym_signed_number_token1] = ACTIONS(1837),
    [anon_sym_SQUOTE] = ACTIONS(1839),
    [anon_sym_DQUOTE] = ACTIONS(1839),
    [sym_prefix_operator] = ACTIONS(1839),
    [aux_sym_sign_operator_token1] = ACTIONS(1837),
    [sym_hash_operator] = ACTIONS(1839),
    [anon_sym_lsl] = ACTIONS(1837),
    [anon_sym_lsr] = ACTIONS(1837),
    [anon_sym_asr] = ACTIONS(1837),
    [aux_sym_pow_operator_token1] = ACTIONS(1839),
    [anon_sym_mod] = ACTIONS(1837),
    [anon_sym_land] = ACTIONS(1837),
    [anon_sym_lor] = ACTIONS(1837),
    [anon_sym_lxor] = ACTIONS(1837),
    [aux_sym_mult_operator_token1] = ACTIONS(1837),
    [aux_sym_add_operator_token1] = ACTIONS(1837),
    [sym_concat_operator] = ACTIONS(1839),
    [sym_rel_operator] = ACTIONS(1837),
    [anon_sym_AMP_AMP] = ACTIONS(1837),
    [anon_sym_or] = ACTIONS(1837),
    [anon_sym_PIPE_PIPE] = ACTIONS(1837),
    [sym_assign_operator] = ACTIONS(1839),
    [sym__uppercase_identifier] = ACTIONS(1839),
    [aux_sym_tag_token1] = ACTIONS(1839),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(922)] = {
    [sym_attribute] = STATE(922),
    [sym__lowercase_identifier] = ACTIONS(1579),
    [anon_sym_TILDE] = ACTIONS(1579),
    [anon_sym_QMARK] = ACTIONS(1579),
    [anon_sym_LPAREN] = ACTIONS(1581),
    [anon_sym_RPAREN] = ACTIONS(1581),
    [anon_sym_COMMA] = ACTIONS(1581),
    [anon_sym_PIPE] = ACTIONS(1579),
    [anon_sym_COLON2] = ACTIONS(1579),
    [anon_sym_DOT] = ACTIONS(1581),
    [anon_sym_DASH_GT] = ACTIONS(1579),
    [anon_sym_LBRACK] = ACTIONS(1579),
    [anon_sym_RBRACK] = ACTIONS(1581),
    [anon_sym_true] = ACTIONS(1579),
    [anon_sym_false] = ACTIONS(1579),
    [anon_sym_LBRACE] = ACTIONS(1579),
    [anon_sym_SEMI] = ACTIONS(1581),
    [anon_sym_RBRACE] = ACTIONS(1581),
    [anon_sym_constraint] = ACTIONS(1579),
    [anon_sym_val] = ACTIONS(1579),
    [anon_sym_end] = ACTIONS(1579),
    [anon_sym_with] = ACTIONS(1579),
    [anon_sym_COLON_GT] = ACTIONS(1581),
    [anon_sym_object] = ACTIONS(1579),
    [anon_sym_inherit] = ACTIONS(1579),
    [anon_sym_method] = ACTIONS(1579),
    [anon_sym_initializer] = ACTIONS(1579),
    [anon_sym_AMP] = ACTIONS(1579),
    [anon_sym_POUND] = ACTIONS(1579),
    [anon_sym_COLON_COLON] = ACTIONS(1581),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1581),
    [anon_sym_then] = ACTIONS(1579),
    [anon_sym_else] = ACTIONS(1579),
    [anon_sym_do] = ACTIONS(1579),
    [anon_sym_done] = ACTIONS(1579),
    [anon_sym_to] = ACTIONS(1579),
    [anon_sym_downto] = ACTIONS(1579),
    [anon_sym_new] = ACTIONS(1579),
    [anon_sym_LBRACE_LT] = ACTIONS(1581),
    [anon_sym_GT_RBRACE] = ACTIONS(1581),
    [anon_sym_begin] = ACTIONS(1579),
    [sym_ocamlyacc_value] = ACTIONS(1581),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1579),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1581),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1579),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1579),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1581),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1581),
    [aux_sym_number_token1] = ACTIONS(1581),
    [aux_sym_signed_number_token1] = ACTIONS(1579),
    [anon_sym_SQUOTE] = ACTIONS(1581),
    [anon_sym_DQUOTE] = ACTIONS(1581),
    [sym_prefix_operator] = ACTIONS(1581),
    [aux_sym_sign_operator_token1] = ACTIONS(1579),
    [sym_hash_operator] = ACTIONS(1581),
    [anon_sym_lsl] = ACTIONS(1579),
    [anon_sym_lsr] = ACTIONS(1579),
    [anon_sym_asr] = ACTIONS(1579),
    [aux_sym_pow_operator_token1] = ACTIONS(1581),
    [anon_sym_mod] = ACTIONS(1579),
    [anon_sym_land] = ACTIONS(1579),
    [anon_sym_lor] = ACTIONS(1579),
    [anon_sym_lxor] = ACTIONS(1579),
    [aux_sym_mult_operator_token1] = ACTIONS(1579),
    [aux_sym_add_operator_token1] = ACTIONS(1579),
    [sym_concat_operator] = ACTIONS(1581),
    [sym_rel_operator] = ACTIONS(1579),
    [anon_sym_AMP_AMP] = ACTIONS(1579),
    [anon_sym_or] = ACTIONS(1579),
    [anon_sym_PIPE_PIPE] = ACTIONS(1579),
    [sym_assign_operator] = ACTIONS(1581),
    [sym__uppercase_identifier] = ACTIONS(1581),
    [aux_sym_tag_token1] = ACTIONS(1581),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(923)] = {
    [sym_attribute] = STATE(923),
    [sym__lowercase_identifier] = ACTIONS(1583),
    [anon_sym_TILDE] = ACTIONS(1583),
    [anon_sym_QMARK] = ACTIONS(1583),
    [anon_sym_LPAREN] = ACTIONS(1585),
    [anon_sym_RPAREN] = ACTIONS(1585),
    [anon_sym_COMMA] = ACTIONS(1585),
    [anon_sym_PIPE] = ACTIONS(1583),
    [anon_sym_COLON2] = ACTIONS(1583),
    [anon_sym_DOT] = ACTIONS(825),
    [anon_sym_DASH_GT] = ACTIONS(1583),
    [anon_sym_LBRACK] = ACTIONS(1583),
    [anon_sym_RBRACK] = ACTIONS(1585),
    [anon_sym_true] = ACTIONS(1583),
    [anon_sym_false] = ACTIONS(1583),
    [anon_sym_LBRACE] = ACTIONS(1583),
    [anon_sym_SEMI] = ACTIONS(1585),
    [anon_sym_RBRACE] = ACTIONS(1585),
    [anon_sym_constraint] = ACTIONS(1583),
    [anon_sym_val] = ACTIONS(1583),
    [anon_sym_end] = ACTIONS(1583),
    [anon_sym_with] = ACTIONS(1583),
    [anon_sym_COLON_GT] = ACTIONS(1585),
    [anon_sym_object] = ACTIONS(1583),
    [anon_sym_inherit] = ACTIONS(1583),
    [anon_sym_method] = ACTIONS(1583),
    [anon_sym_initializer] = ACTIONS(1583),
    [anon_sym_AMP] = ACTIONS(1583),
    [anon_sym_POUND] = ACTIONS(1583),
    [anon_sym_COLON_COLON] = ACTIONS(1585),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1585),
    [anon_sym_then] = ACTIONS(1583),
    [anon_sym_else] = ACTIONS(1583),
    [anon_sym_do] = ACTIONS(1583),
    [anon_sym_done] = ACTIONS(1583),
    [anon_sym_to] = ACTIONS(1583),
    [anon_sym_downto] = ACTIONS(1583),
    [anon_sym_new] = ACTIONS(1583),
    [anon_sym_LBRACE_LT] = ACTIONS(1585),
    [anon_sym_GT_RBRACE] = ACTIONS(1585),
    [anon_sym_begin] = ACTIONS(1583),
    [sym_ocamlyacc_value] = ACTIONS(1585),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1583),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1585),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1583),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1583),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1585),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1585),
    [aux_sym_number_token1] = ACTIONS(1585),
    [aux_sym_signed_number_token1] = ACTIONS(1583),
    [anon_sym_SQUOTE] = ACTIONS(1585),
    [anon_sym_DQUOTE] = ACTIONS(1585),
    [sym_prefix_operator] = ACTIONS(1585),
    [aux_sym_sign_operator_token1] = ACTIONS(1583),
    [sym_hash_operator] = ACTIONS(1585),
    [anon_sym_lsl] = ACTIONS(1583),
    [anon_sym_lsr] = ACTIONS(1583),
    [anon_sym_asr] = ACTIONS(1583),
    [aux_sym_pow_operator_token1] = ACTIONS(1585),
    [anon_sym_mod] = ACTIONS(1583),
    [anon_sym_land] = ACTIONS(1583),
    [anon_sym_lor] = ACTIONS(1583),
    [anon_sym_lxor] = ACTIONS(1583),
    [aux_sym_mult_operator_token1] = ACTIONS(1583),
    [aux_sym_add_operator_token1] = ACTIONS(1583),
    [sym_concat_operator] = ACTIONS(1585),
    [sym_rel_operator] = ACTIONS(1583),
    [anon_sym_AMP_AMP] = ACTIONS(1583),
    [anon_sym_or] = ACTIONS(1583),
    [anon_sym_PIPE_PIPE] = ACTIONS(1583),
    [sym_assign_operator] = ACTIONS(1585),
    [sym__uppercase_identifier] = ACTIONS(1585),
    [aux_sym_tag_token1] = ACTIONS(1585),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(924)] = {
    [sym_attribute] = STATE(924),
    [sym__lowercase_identifier] = ACTIONS(1587),
    [anon_sym_TILDE] = ACTIONS(1587),
    [anon_sym_QMARK] = ACTIONS(1587),
    [anon_sym_LPAREN] = ACTIONS(1589),
    [anon_sym_RPAREN] = ACTIONS(1589),
    [anon_sym_COMMA] = ACTIONS(1589),
    [anon_sym_PIPE] = ACTIONS(1587),
    [anon_sym_COLON2] = ACTIONS(1587),
    [anon_sym_DOT] = ACTIONS(1589),
    [anon_sym_DASH_GT] = ACTIONS(1587),
    [anon_sym_LBRACK] = ACTIONS(1587),
    [anon_sym_RBRACK] = ACTIONS(1589),
    [anon_sym_true] = ACTIONS(1587),
    [anon_sym_false] = ACTIONS(1587),
    [anon_sym_LBRACE] = ACTIONS(1587),
    [anon_sym_SEMI] = ACTIONS(1589),
    [anon_sym_RBRACE] = ACTIONS(1589),
    [anon_sym_constraint] = ACTIONS(1587),
    [anon_sym_val] = ACTIONS(1587),
    [anon_sym_end] = ACTIONS(1587),
    [anon_sym_with] = ACTIONS(1587),
    [anon_sym_COLON_GT] = ACTIONS(1589),
    [anon_sym_object] = ACTIONS(1587),
    [anon_sym_inherit] = ACTIONS(1587),
    [anon_sym_method] = ACTIONS(1587),
    [anon_sym_initializer] = ACTIONS(1587),
    [anon_sym_AMP] = ACTIONS(1587),
    [anon_sym_POUND] = ACTIONS(1587),
    [anon_sym_COLON_COLON] = ACTIONS(1589),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1589),
    [anon_sym_then] = ACTIONS(1587),
    [anon_sym_else] = ACTIONS(1587),
    [anon_sym_do] = ACTIONS(1587),
    [anon_sym_done] = ACTIONS(1587),
    [anon_sym_to] = ACTIONS(1587),
    [anon_sym_downto] = ACTIONS(1587),
    [anon_sym_new] = ACTIONS(1587),
    [anon_sym_LBRACE_LT] = ACTIONS(1589),
    [anon_sym_GT_RBRACE] = ACTIONS(1589),
    [anon_sym_begin] = ACTIONS(1587),
    [sym_ocamlyacc_value] = ACTIONS(1589),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1587),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1589),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1587),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1587),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1589),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1589),
    [aux_sym_number_token1] = ACTIONS(1589),
    [aux_sym_signed_number_token1] = ACTIONS(1587),
    [anon_sym_SQUOTE] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(1589),
    [sym_prefix_operator] = ACTIONS(1589),
    [aux_sym_sign_operator_token1] = ACTIONS(1587),
    [sym_hash_operator] = ACTIONS(1589),
    [anon_sym_lsl] = ACTIONS(1587),
    [anon_sym_lsr] = ACTIONS(1587),
    [anon_sym_asr] = ACTIONS(1587),
    [aux_sym_pow_operator_token1] = ACTIONS(1589),
    [anon_sym_mod] = ACTIONS(1587),
    [anon_sym_land] = ACTIONS(1587),
    [anon_sym_lor] = ACTIONS(1587),
    [anon_sym_lxor] = ACTIONS(1587),
    [aux_sym_mult_operator_token1] = ACTIONS(1587),
    [aux_sym_add_operator_token1] = ACTIONS(1587),
    [sym_concat_operator] = ACTIONS(1589),
    [sym_rel_operator] = ACTIONS(1587),
    [anon_sym_AMP_AMP] = ACTIONS(1587),
    [anon_sym_or] = ACTIONS(1587),
    [anon_sym_PIPE_PIPE] = ACTIONS(1587),
    [sym_assign_operator] = ACTIONS(1589),
    [sym__uppercase_identifier] = ACTIONS(1589),
    [aux_sym_tag_token1] = ACTIONS(1589),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(925)] = {
    [sym_attribute] = STATE(925),
    [sym__lowercase_identifier] = ACTIONS(1599),
    [anon_sym_TILDE] = ACTIONS(1599),
    [anon_sym_QMARK] = ACTIONS(1599),
    [anon_sym_LPAREN] = ACTIONS(1601),
    [anon_sym_RPAREN] = ACTIONS(1601),
    [anon_sym_COMMA] = ACTIONS(1601),
    [anon_sym_PIPE] = ACTIONS(1599),
    [anon_sym_COLON2] = ACTIONS(1599),
    [anon_sym_DOT] = ACTIONS(1601),
    [anon_sym_DASH_GT] = ACTIONS(1599),
    [anon_sym_LBRACK] = ACTIONS(1599),
    [anon_sym_RBRACK] = ACTIONS(1601),
    [anon_sym_true] = ACTIONS(1599),
    [anon_sym_false] = ACTIONS(1599),
    [anon_sym_LBRACE] = ACTIONS(1599),
    [anon_sym_SEMI] = ACTIONS(1601),
    [anon_sym_RBRACE] = ACTIONS(1601),
    [anon_sym_constraint] = ACTIONS(1599),
    [anon_sym_val] = ACTIONS(1599),
    [anon_sym_end] = ACTIONS(1599),
    [anon_sym_with] = ACTIONS(1599),
    [anon_sym_COLON_GT] = ACTIONS(1601),
    [anon_sym_object] = ACTIONS(1599),
    [anon_sym_inherit] = ACTIONS(1599),
    [anon_sym_method] = ACTIONS(1599),
    [anon_sym_initializer] = ACTIONS(1599),
    [anon_sym_AMP] = ACTIONS(1599),
    [anon_sym_POUND] = ACTIONS(1599),
    [anon_sym_COLON_COLON] = ACTIONS(1601),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1601),
    [anon_sym_then] = ACTIONS(1599),
    [anon_sym_else] = ACTIONS(1599),
    [anon_sym_do] = ACTIONS(1599),
    [anon_sym_done] = ACTIONS(1599),
    [anon_sym_to] = ACTIONS(1599),
    [anon_sym_downto] = ACTIONS(1599),
    [anon_sym_new] = ACTIONS(1599),
    [anon_sym_LBRACE_LT] = ACTIONS(1601),
    [anon_sym_GT_RBRACE] = ACTIONS(1601),
    [anon_sym_begin] = ACTIONS(1599),
    [sym_ocamlyacc_value] = ACTIONS(1601),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1599),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1601),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1599),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1599),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1601),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1601),
    [aux_sym_number_token1] = ACTIONS(1601),
    [aux_sym_signed_number_token1] = ACTIONS(1599),
    [anon_sym_SQUOTE] = ACTIONS(1601),
    [anon_sym_DQUOTE] = ACTIONS(1601),
    [sym_prefix_operator] = ACTIONS(1601),
    [aux_sym_sign_operator_token1] = ACTIONS(1599),
    [sym_hash_operator] = ACTIONS(1601),
    [anon_sym_lsl] = ACTIONS(1599),
    [anon_sym_lsr] = ACTIONS(1599),
    [anon_sym_asr] = ACTIONS(1599),
    [aux_sym_pow_operator_token1] = ACTIONS(1601),
    [anon_sym_mod] = ACTIONS(1599),
    [anon_sym_land] = ACTIONS(1599),
    [anon_sym_lor] = ACTIONS(1599),
    [anon_sym_lxor] = ACTIONS(1599),
    [aux_sym_mult_operator_token1] = ACTIONS(1599),
    [aux_sym_add_operator_token1] = ACTIONS(1599),
    [sym_concat_operator] = ACTIONS(1601),
    [sym_rel_operator] = ACTIONS(1599),
    [anon_sym_AMP_AMP] = ACTIONS(1599),
    [anon_sym_or] = ACTIONS(1599),
    [anon_sym_PIPE_PIPE] = ACTIONS(1599),
    [sym_assign_operator] = ACTIONS(1601),
    [sym__uppercase_identifier] = ACTIONS(1601),
    [aux_sym_tag_token1] = ACTIONS(1601),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(926)] = {
    [sym_attribute] = STATE(926),
    [sym__lowercase_identifier] = ACTIONS(1603),
    [anon_sym_TILDE] = ACTIONS(1603),
    [anon_sym_QMARK] = ACTIONS(1603),
    [anon_sym_LPAREN] = ACTIONS(1605),
    [anon_sym_RPAREN] = ACTIONS(1605),
    [anon_sym_COMMA] = ACTIONS(1605),
    [anon_sym_PIPE] = ACTIONS(1603),
    [anon_sym_COLON2] = ACTIONS(1603),
    [anon_sym_DOT] = ACTIONS(1605),
    [anon_sym_DASH_GT] = ACTIONS(1603),
    [anon_sym_LBRACK] = ACTIONS(1603),
    [anon_sym_RBRACK] = ACTIONS(1605),
    [anon_sym_true] = ACTIONS(1603),
    [anon_sym_false] = ACTIONS(1603),
    [anon_sym_LBRACE] = ACTIONS(1603),
    [anon_sym_SEMI] = ACTIONS(1605),
    [anon_sym_RBRACE] = ACTIONS(1605),
    [anon_sym_constraint] = ACTIONS(1603),
    [anon_sym_val] = ACTIONS(1603),
    [anon_sym_end] = ACTIONS(1603),
    [anon_sym_with] = ACTIONS(1603),
    [anon_sym_COLON_GT] = ACTIONS(1605),
    [anon_sym_object] = ACTIONS(1603),
    [anon_sym_inherit] = ACTIONS(1603),
    [anon_sym_method] = ACTIONS(1603),
    [anon_sym_initializer] = ACTIONS(1603),
    [anon_sym_AMP] = ACTIONS(1603),
    [anon_sym_POUND] = ACTIONS(1603),
    [anon_sym_COLON_COLON] = ACTIONS(1605),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1605),
    [anon_sym_then] = ACTIONS(1603),
    [anon_sym_else] = ACTIONS(1603),
    [anon_sym_do] = ACTIONS(1603),
    [anon_sym_done] = ACTIONS(1603),
    [anon_sym_to] = ACTIONS(1603),
    [anon_sym_downto] = ACTIONS(1603),
    [anon_sym_new] = ACTIONS(1603),
    [anon_sym_LBRACE_LT] = ACTIONS(1605),
    [anon_sym_GT_RBRACE] = ACTIONS(1605),
    [anon_sym_begin] = ACTIONS(1603),
    [sym_ocamlyacc_value] = ACTIONS(1605),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1603),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1605),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1603),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1603),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1605),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1605),
    [aux_sym_number_token1] = ACTIONS(1605),
    [aux_sym_signed_number_token1] = ACTIONS(1603),
    [anon_sym_SQUOTE] = ACTIONS(1605),
    [anon_sym_DQUOTE] = ACTIONS(1605),
    [sym_prefix_operator] = ACTIONS(1605),
    [aux_sym_sign_operator_token1] = ACTIONS(1603),
    [sym_hash_operator] = ACTIONS(1605),
    [anon_sym_lsl] = ACTIONS(1603),
    [anon_sym_lsr] = ACTIONS(1603),
    [anon_sym_asr] = ACTIONS(1603),
    [aux_sym_pow_operator_token1] = ACTIONS(1605),
    [anon_sym_mod] = ACTIONS(1603),
    [anon_sym_land] = ACTIONS(1603),
    [anon_sym_lor] = ACTIONS(1603),
    [anon_sym_lxor] = ACTIONS(1603),
    [aux_sym_mult_operator_token1] = ACTIONS(1603),
    [aux_sym_add_operator_token1] = ACTIONS(1603),
    [sym_concat_operator] = ACTIONS(1605),
    [sym_rel_operator] = ACTIONS(1603),
    [anon_sym_AMP_AMP] = ACTIONS(1603),
    [anon_sym_or] = ACTIONS(1603),
    [anon_sym_PIPE_PIPE] = ACTIONS(1603),
    [sym_assign_operator] = ACTIONS(1605),
    [sym__uppercase_identifier] = ACTIONS(1605),
    [aux_sym_tag_token1] = ACTIONS(1605),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(927)] = {
    [sym_attribute] = STATE(927),
    [sym__lowercase_identifier] = ACTIONS(1607),
    [anon_sym_TILDE] = ACTIONS(1607),
    [anon_sym_QMARK] = ACTIONS(1607),
    [anon_sym_LPAREN] = ACTIONS(1609),
    [anon_sym_RPAREN] = ACTIONS(1609),
    [anon_sym_COMMA] = ACTIONS(1609),
    [anon_sym_PIPE] = ACTIONS(1607),
    [anon_sym_COLON2] = ACTIONS(1607),
    [anon_sym_DOT] = ACTIONS(1609),
    [anon_sym_DASH_GT] = ACTIONS(1607),
    [anon_sym_LBRACK] = ACTIONS(1607),
    [anon_sym_RBRACK] = ACTIONS(1609),
    [anon_sym_true] = ACTIONS(1607),
    [anon_sym_false] = ACTIONS(1607),
    [anon_sym_LBRACE] = ACTIONS(1607),
    [anon_sym_SEMI] = ACTIONS(1609),
    [anon_sym_RBRACE] = ACTIONS(1609),
    [anon_sym_constraint] = ACTIONS(1607),
    [anon_sym_val] = ACTIONS(1607),
    [anon_sym_end] = ACTIONS(1607),
    [anon_sym_with] = ACTIONS(1607),
    [anon_sym_COLON_GT] = ACTIONS(1609),
    [anon_sym_object] = ACTIONS(1607),
    [anon_sym_inherit] = ACTIONS(1607),
    [anon_sym_method] = ACTIONS(1607),
    [anon_sym_initializer] = ACTIONS(1607),
    [anon_sym_AMP] = ACTIONS(1607),
    [anon_sym_POUND] = ACTIONS(1607),
    [anon_sym_COLON_COLON] = ACTIONS(1609),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1609),
    [anon_sym_then] = ACTIONS(1607),
    [anon_sym_else] = ACTIONS(1607),
    [anon_sym_do] = ACTIONS(1607),
    [anon_sym_done] = ACTIONS(1607),
    [anon_sym_to] = ACTIONS(1607),
    [anon_sym_downto] = ACTIONS(1607),
    [anon_sym_new] = ACTIONS(1607),
    [anon_sym_LBRACE_LT] = ACTIONS(1609),
    [anon_sym_GT_RBRACE] = ACTIONS(1609),
    [anon_sym_begin] = ACTIONS(1607),
    [sym_ocamlyacc_value] = ACTIONS(1609),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1607),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1609),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1607),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1607),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1609),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1609),
    [aux_sym_number_token1] = ACTIONS(1609),
    [aux_sym_signed_number_token1] = ACTIONS(1607),
    [anon_sym_SQUOTE] = ACTIONS(1609),
    [anon_sym_DQUOTE] = ACTIONS(1609),
    [sym_prefix_operator] = ACTIONS(1609),
    [aux_sym_sign_operator_token1] = ACTIONS(1607),
    [sym_hash_operator] = ACTIONS(1609),
    [anon_sym_lsl] = ACTIONS(1607),
    [anon_sym_lsr] = ACTIONS(1607),
    [anon_sym_asr] = ACTIONS(1607),
    [aux_sym_pow_operator_token1] = ACTIONS(1609),
    [anon_sym_mod] = ACTIONS(1607),
    [anon_sym_land] = ACTIONS(1607),
    [anon_sym_lor] = ACTIONS(1607),
    [anon_sym_lxor] = ACTIONS(1607),
    [aux_sym_mult_operator_token1] = ACTIONS(1607),
    [aux_sym_add_operator_token1] = ACTIONS(1607),
    [sym_concat_operator] = ACTIONS(1609),
    [sym_rel_operator] = ACTIONS(1607),
    [anon_sym_AMP_AMP] = ACTIONS(1607),
    [anon_sym_or] = ACTIONS(1607),
    [anon_sym_PIPE_PIPE] = ACTIONS(1607),
    [sym_assign_operator] = ACTIONS(1609),
    [sym__uppercase_identifier] = ACTIONS(1609),
    [aux_sym_tag_token1] = ACTIONS(1609),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(928)] = {
    [sym_attribute] = STATE(928),
    [sym__lowercase_identifier] = ACTIONS(1631),
    [anon_sym_TILDE] = ACTIONS(1631),
    [anon_sym_QMARK] = ACTIONS(1631),
    [anon_sym_LPAREN] = ACTIONS(1633),
    [anon_sym_RPAREN] = ACTIONS(1633),
    [anon_sym_COMMA] = ACTIONS(1633),
    [anon_sym_PIPE] = ACTIONS(1631),
    [anon_sym_COLON2] = ACTIONS(1631),
    [anon_sym_DOT] = ACTIONS(1633),
    [anon_sym_DASH_GT] = ACTIONS(1631),
    [anon_sym_LBRACK] = ACTIONS(1631),
    [anon_sym_RBRACK] = ACTIONS(1633),
    [anon_sym_true] = ACTIONS(1631),
    [anon_sym_false] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1631),
    [anon_sym_SEMI] = ACTIONS(1633),
    [anon_sym_RBRACE] = ACTIONS(1633),
    [anon_sym_constraint] = ACTIONS(1631),
    [anon_sym_val] = ACTIONS(1631),
    [anon_sym_end] = ACTIONS(1631),
    [anon_sym_with] = ACTIONS(1631),
    [anon_sym_COLON_GT] = ACTIONS(1633),
    [anon_sym_object] = ACTIONS(1631),
    [anon_sym_inherit] = ACTIONS(1631),
    [anon_sym_method] = ACTIONS(1631),
    [anon_sym_initializer] = ACTIONS(1631),
    [anon_sym_AMP] = ACTIONS(1631),
    [anon_sym_POUND] = ACTIONS(1631),
    [anon_sym_COLON_COLON] = ACTIONS(1633),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1633),
    [anon_sym_then] = ACTIONS(1631),
    [anon_sym_else] = ACTIONS(1631),
    [anon_sym_do] = ACTIONS(1631),
    [anon_sym_done] = ACTIONS(1631),
    [anon_sym_to] = ACTIONS(1631),
    [anon_sym_downto] = ACTIONS(1631),
    [anon_sym_new] = ACTIONS(1631),
    [anon_sym_LBRACE_LT] = ACTIONS(1633),
    [anon_sym_GT_RBRACE] = ACTIONS(1633),
    [anon_sym_begin] = ACTIONS(1631),
    [sym_ocamlyacc_value] = ACTIONS(1633),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1631),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1633),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1631),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1631),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1633),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1633),
    [aux_sym_number_token1] = ACTIONS(1633),
    [aux_sym_signed_number_token1] = ACTIONS(1631),
    [anon_sym_SQUOTE] = ACTIONS(1633),
    [anon_sym_DQUOTE] = ACTIONS(1633),
    [sym_prefix_operator] = ACTIONS(1633),
    [aux_sym_sign_operator_token1] = ACTIONS(1631),
    [sym_hash_operator] = ACTIONS(1633),
    [anon_sym_lsl] = ACTIONS(1631),
    [anon_sym_lsr] = ACTIONS(1631),
    [anon_sym_asr] = ACTIONS(1631),
    [aux_sym_pow_operator_token1] = ACTIONS(1633),
    [anon_sym_mod] = ACTIONS(1631),
    [anon_sym_land] = ACTIONS(1631),
    [anon_sym_lor] = ACTIONS(1631),
    [anon_sym_lxor] = ACTIONS(1631),
    [aux_sym_mult_operator_token1] = ACTIONS(1631),
    [aux_sym_add_operator_token1] = ACTIONS(1631),
    [sym_concat_operator] = ACTIONS(1633),
    [sym_rel_operator] = ACTIONS(1631),
    [anon_sym_AMP_AMP] = ACTIONS(1631),
    [anon_sym_or] = ACTIONS(1631),
    [anon_sym_PIPE_PIPE] = ACTIONS(1631),
    [sym_assign_operator] = ACTIONS(1633),
    [sym__uppercase_identifier] = ACTIONS(1633),
    [aux_sym_tag_token1] = ACTIONS(1633),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(929)] = {
    [sym_attribute] = STATE(929),
    [sym__lowercase_identifier] = ACTIONS(1635),
    [anon_sym_TILDE] = ACTIONS(1635),
    [anon_sym_QMARK] = ACTIONS(1635),
    [anon_sym_LPAREN] = ACTIONS(1637),
    [anon_sym_RPAREN] = ACTIONS(1637),
    [anon_sym_COMMA] = ACTIONS(1637),
    [anon_sym_PIPE] = ACTIONS(1635),
    [anon_sym_COLON2] = ACTIONS(1635),
    [anon_sym_DOT] = ACTIONS(1637),
    [anon_sym_DASH_GT] = ACTIONS(1635),
    [anon_sym_LBRACK] = ACTIONS(1635),
    [anon_sym_RBRACK] = ACTIONS(1637),
    [anon_sym_true] = ACTIONS(1635),
    [anon_sym_false] = ACTIONS(1635),
    [anon_sym_LBRACE] = ACTIONS(1635),
    [anon_sym_SEMI] = ACTIONS(1637),
    [anon_sym_RBRACE] = ACTIONS(1637),
    [anon_sym_constraint] = ACTIONS(1635),
    [anon_sym_val] = ACTIONS(1635),
    [anon_sym_end] = ACTIONS(1635),
    [anon_sym_with] = ACTIONS(1635),
    [anon_sym_COLON_GT] = ACTIONS(1637),
    [anon_sym_object] = ACTIONS(1635),
    [anon_sym_inherit] = ACTIONS(1635),
    [anon_sym_method] = ACTIONS(1635),
    [anon_sym_initializer] = ACTIONS(1635),
    [anon_sym_AMP] = ACTIONS(1635),
    [anon_sym_POUND] = ACTIONS(1635),
    [anon_sym_COLON_COLON] = ACTIONS(1637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1637),
    [anon_sym_then] = ACTIONS(1635),
    [anon_sym_else] = ACTIONS(1635),
    [anon_sym_do] = ACTIONS(1635),
    [anon_sym_done] = ACTIONS(1635),
    [anon_sym_to] = ACTIONS(1635),
    [anon_sym_downto] = ACTIONS(1635),
    [anon_sym_new] = ACTIONS(1635),
    [anon_sym_LBRACE_LT] = ACTIONS(1637),
    [anon_sym_GT_RBRACE] = ACTIONS(1637),
    [anon_sym_begin] = ACTIONS(1635),
    [sym_ocamlyacc_value] = ACTIONS(1637),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1635),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1637),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1635),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1635),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1637),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1637),
    [aux_sym_number_token1] = ACTIONS(1637),
    [aux_sym_signed_number_token1] = ACTIONS(1635),
    [anon_sym_SQUOTE] = ACTIONS(1637),
    [anon_sym_DQUOTE] = ACTIONS(1637),
    [sym_prefix_operator] = ACTIONS(1637),
    [aux_sym_sign_operator_token1] = ACTIONS(1635),
    [sym_hash_operator] = ACTIONS(1637),
    [anon_sym_lsl] = ACTIONS(1635),
    [anon_sym_lsr] = ACTIONS(1635),
    [anon_sym_asr] = ACTIONS(1635),
    [aux_sym_pow_operator_token1] = ACTIONS(1637),
    [anon_sym_mod] = ACTIONS(1635),
    [anon_sym_land] = ACTIONS(1635),
    [anon_sym_lor] = ACTIONS(1635),
    [anon_sym_lxor] = ACTIONS(1635),
    [aux_sym_mult_operator_token1] = ACTIONS(1635),
    [aux_sym_add_operator_token1] = ACTIONS(1635),
    [sym_concat_operator] = ACTIONS(1637),
    [sym_rel_operator] = ACTIONS(1635),
    [anon_sym_AMP_AMP] = ACTIONS(1635),
    [anon_sym_or] = ACTIONS(1635),
    [anon_sym_PIPE_PIPE] = ACTIONS(1635),
    [sym_assign_operator] = ACTIONS(1637),
    [sym__uppercase_identifier] = ACTIONS(1637),
    [aux_sym_tag_token1] = ACTIONS(1637),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(930)] = {
    [sym_attribute] = STATE(930),
    [sym__lowercase_identifier] = ACTIONS(1639),
    [anon_sym_TILDE] = ACTIONS(1639),
    [anon_sym_QMARK] = ACTIONS(1639),
    [anon_sym_LPAREN] = ACTIONS(1641),
    [anon_sym_RPAREN] = ACTIONS(1641),
    [anon_sym_COMMA] = ACTIONS(1641),
    [anon_sym_PIPE] = ACTIONS(1639),
    [anon_sym_COLON2] = ACTIONS(1639),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_DASH_GT] = ACTIONS(1639),
    [anon_sym_LBRACK] = ACTIONS(1639),
    [anon_sym_RBRACK] = ACTIONS(1641),
    [anon_sym_true] = ACTIONS(1639),
    [anon_sym_false] = ACTIONS(1639),
    [anon_sym_LBRACE] = ACTIONS(1639),
    [anon_sym_SEMI] = ACTIONS(1641),
    [anon_sym_RBRACE] = ACTIONS(1641),
    [anon_sym_constraint] = ACTIONS(1639),
    [anon_sym_val] = ACTIONS(1639),
    [anon_sym_end] = ACTIONS(1639),
    [anon_sym_with] = ACTIONS(1639),
    [anon_sym_COLON_GT] = ACTIONS(1641),
    [anon_sym_object] = ACTIONS(1639),
    [anon_sym_inherit] = ACTIONS(1639),
    [anon_sym_method] = ACTIONS(1639),
    [anon_sym_initializer] = ACTIONS(1639),
    [anon_sym_AMP] = ACTIONS(1639),
    [anon_sym_POUND] = ACTIONS(1639),
    [anon_sym_COLON_COLON] = ACTIONS(1641),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1641),
    [anon_sym_then] = ACTIONS(1639),
    [anon_sym_else] = ACTIONS(1639),
    [anon_sym_do] = ACTIONS(1639),
    [anon_sym_done] = ACTIONS(1639),
    [anon_sym_to] = ACTIONS(1639),
    [anon_sym_downto] = ACTIONS(1639),
    [anon_sym_new] = ACTIONS(1639),
    [anon_sym_LBRACE_LT] = ACTIONS(1641),
    [anon_sym_GT_RBRACE] = ACTIONS(1641),
    [anon_sym_begin] = ACTIONS(1639),
    [sym_ocamlyacc_value] = ACTIONS(1641),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1639),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1641),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1639),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1639),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1641),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1641),
    [aux_sym_number_token1] = ACTIONS(1641),
    [aux_sym_signed_number_token1] = ACTIONS(1639),
    [anon_sym_SQUOTE] = ACTIONS(1641),
    [anon_sym_DQUOTE] = ACTIONS(1641),
    [sym_prefix_operator] = ACTIONS(1641),
    [aux_sym_sign_operator_token1] = ACTIONS(1639),
    [sym_hash_operator] = ACTIONS(1641),
    [anon_sym_lsl] = ACTIONS(1639),
    [anon_sym_lsr] = ACTIONS(1639),
    [anon_sym_asr] = ACTIONS(1639),
    [aux_sym_pow_operator_token1] = ACTIONS(1641),
    [anon_sym_mod] = ACTIONS(1639),
    [anon_sym_land] = ACTIONS(1639),
    [anon_sym_lor] = ACTIONS(1639),
    [anon_sym_lxor] = ACTIONS(1639),
    [aux_sym_mult_operator_token1] = ACTIONS(1639),
    [aux_sym_add_operator_token1] = ACTIONS(1639),
    [sym_concat_operator] = ACTIONS(1641),
    [sym_rel_operator] = ACTIONS(1639),
    [anon_sym_AMP_AMP] = ACTIONS(1639),
    [anon_sym_or] = ACTIONS(1639),
    [anon_sym_PIPE_PIPE] = ACTIONS(1639),
    [sym_assign_operator] = ACTIONS(1641),
    [sym__uppercase_identifier] = ACTIONS(1641),
    [aux_sym_tag_token1] = ACTIONS(1641),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(931)] = {
    [sym_attribute] = STATE(931),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_TILDE] = ACTIONS(1643),
    [anon_sym_QMARK] = ACTIONS(1643),
    [anon_sym_LPAREN] = ACTIONS(1645),
    [anon_sym_RPAREN] = ACTIONS(1645),
    [anon_sym_COMMA] = ACTIONS(1645),
    [anon_sym_PIPE] = ACTIONS(1643),
    [anon_sym_COLON2] = ACTIONS(1643),
    [anon_sym_DOT] = ACTIONS(1645),
    [anon_sym_DASH_GT] = ACTIONS(1643),
    [anon_sym_LBRACK] = ACTIONS(1643),
    [anon_sym_RBRACK] = ACTIONS(1645),
    [anon_sym_true] = ACTIONS(1643),
    [anon_sym_false] = ACTIONS(1643),
    [anon_sym_LBRACE] = ACTIONS(1643),
    [anon_sym_SEMI] = ACTIONS(1645),
    [anon_sym_RBRACE] = ACTIONS(1645),
    [anon_sym_constraint] = ACTIONS(1643),
    [anon_sym_val] = ACTIONS(1643),
    [anon_sym_end] = ACTIONS(1643),
    [anon_sym_with] = ACTIONS(1643),
    [anon_sym_COLON_GT] = ACTIONS(1645),
    [anon_sym_object] = ACTIONS(1643),
    [anon_sym_inherit] = ACTIONS(1643),
    [anon_sym_method] = ACTIONS(1643),
    [anon_sym_initializer] = ACTIONS(1643),
    [anon_sym_AMP] = ACTIONS(1643),
    [anon_sym_POUND] = ACTIONS(1643),
    [anon_sym_COLON_COLON] = ACTIONS(1645),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1645),
    [anon_sym_then] = ACTIONS(1643),
    [anon_sym_else] = ACTIONS(1643),
    [anon_sym_do] = ACTIONS(1643),
    [anon_sym_done] = ACTIONS(1643),
    [anon_sym_to] = ACTIONS(1643),
    [anon_sym_downto] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1643),
    [anon_sym_LBRACE_LT] = ACTIONS(1645),
    [anon_sym_GT_RBRACE] = ACTIONS(1645),
    [anon_sym_begin] = ACTIONS(1643),
    [sym_ocamlyacc_value] = ACTIONS(1645),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1643),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1645),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1643),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1643),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1645),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1645),
    [aux_sym_number_token1] = ACTIONS(1645),
    [aux_sym_signed_number_token1] = ACTIONS(1643),
    [anon_sym_SQUOTE] = ACTIONS(1645),
    [anon_sym_DQUOTE] = ACTIONS(1645),
    [sym_prefix_operator] = ACTIONS(1645),
    [aux_sym_sign_operator_token1] = ACTIONS(1643),
    [sym_hash_operator] = ACTIONS(1645),
    [anon_sym_lsl] = ACTIONS(1643),
    [anon_sym_lsr] = ACTIONS(1643),
    [anon_sym_asr] = ACTIONS(1643),
    [aux_sym_pow_operator_token1] = ACTIONS(1645),
    [anon_sym_mod] = ACTIONS(1643),
    [anon_sym_land] = ACTIONS(1643),
    [anon_sym_lor] = ACTIONS(1643),
    [anon_sym_lxor] = ACTIONS(1643),
    [aux_sym_mult_operator_token1] = ACTIONS(1643),
    [aux_sym_add_operator_token1] = ACTIONS(1643),
    [sym_concat_operator] = ACTIONS(1645),
    [sym_rel_operator] = ACTIONS(1643),
    [anon_sym_AMP_AMP] = ACTIONS(1643),
    [anon_sym_or] = ACTIONS(1643),
    [anon_sym_PIPE_PIPE] = ACTIONS(1643),
    [sym_assign_operator] = ACTIONS(1645),
    [sym__uppercase_identifier] = ACTIONS(1645),
    [aux_sym_tag_token1] = ACTIONS(1645),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(932)] = {
    [sym_attribute] = STATE(932),
    [sym__lowercase_identifier] = ACTIONS(1647),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_QMARK] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1649),
    [anon_sym_RPAREN] = ACTIONS(1649),
    [anon_sym_COMMA] = ACTIONS(1649),
    [anon_sym_PIPE] = ACTIONS(1647),
    [anon_sym_COLON2] = ACTIONS(1647),
    [anon_sym_DOT] = ACTIONS(1649),
    [anon_sym_DASH_GT] = ACTIONS(1647),
    [anon_sym_LBRACK] = ACTIONS(1647),
    [anon_sym_RBRACK] = ACTIONS(1649),
    [anon_sym_true] = ACTIONS(1647),
    [anon_sym_false] = ACTIONS(1647),
    [anon_sym_LBRACE] = ACTIONS(1647),
    [anon_sym_SEMI] = ACTIONS(1649),
    [anon_sym_RBRACE] = ACTIONS(1649),
    [anon_sym_constraint] = ACTIONS(1647),
    [anon_sym_val] = ACTIONS(1647),
    [anon_sym_end] = ACTIONS(1647),
    [anon_sym_with] = ACTIONS(1647),
    [anon_sym_COLON_GT] = ACTIONS(1649),
    [anon_sym_object] = ACTIONS(1647),
    [anon_sym_inherit] = ACTIONS(1647),
    [anon_sym_method] = ACTIONS(1647),
    [anon_sym_initializer] = ACTIONS(1647),
    [anon_sym_AMP] = ACTIONS(1647),
    [anon_sym_POUND] = ACTIONS(1647),
    [anon_sym_COLON_COLON] = ACTIONS(1649),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1649),
    [anon_sym_then] = ACTIONS(1647),
    [anon_sym_else] = ACTIONS(1647),
    [anon_sym_do] = ACTIONS(1647),
    [anon_sym_done] = ACTIONS(1647),
    [anon_sym_to] = ACTIONS(1647),
    [anon_sym_downto] = ACTIONS(1647),
    [anon_sym_new] = ACTIONS(1647),
    [anon_sym_LBRACE_LT] = ACTIONS(1649),
    [anon_sym_GT_RBRACE] = ACTIONS(1649),
    [anon_sym_begin] = ACTIONS(1647),
    [sym_ocamlyacc_value] = ACTIONS(1649),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1647),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1649),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1647),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1647),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1649),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1649),
    [aux_sym_number_token1] = ACTIONS(1649),
    [aux_sym_signed_number_token1] = ACTIONS(1647),
    [anon_sym_SQUOTE] = ACTIONS(1649),
    [anon_sym_DQUOTE] = ACTIONS(1649),
    [sym_prefix_operator] = ACTIONS(1649),
    [aux_sym_sign_operator_token1] = ACTIONS(1647),
    [sym_hash_operator] = ACTIONS(1649),
    [anon_sym_lsl] = ACTIONS(1647),
    [anon_sym_lsr] = ACTIONS(1647),
    [anon_sym_asr] = ACTIONS(1647),
    [aux_sym_pow_operator_token1] = ACTIONS(1649),
    [anon_sym_mod] = ACTIONS(1647),
    [anon_sym_land] = ACTIONS(1647),
    [anon_sym_lor] = ACTIONS(1647),
    [anon_sym_lxor] = ACTIONS(1647),
    [aux_sym_mult_operator_token1] = ACTIONS(1647),
    [aux_sym_add_operator_token1] = ACTIONS(1647),
    [sym_concat_operator] = ACTIONS(1649),
    [sym_rel_operator] = ACTIONS(1647),
    [anon_sym_AMP_AMP] = ACTIONS(1647),
    [anon_sym_or] = ACTIONS(1647),
    [anon_sym_PIPE_PIPE] = ACTIONS(1647),
    [sym_assign_operator] = ACTIONS(1649),
    [sym__uppercase_identifier] = ACTIONS(1649),
    [aux_sym_tag_token1] = ACTIONS(1649),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(933)] = {
    [sym_attribute] = STATE(933),
    [sym__lowercase_identifier] = ACTIONS(1651),
    [anon_sym_TILDE] = ACTIONS(1651),
    [anon_sym_QMARK] = ACTIONS(1651),
    [anon_sym_LPAREN] = ACTIONS(1653),
    [anon_sym_RPAREN] = ACTIONS(1653),
    [anon_sym_COMMA] = ACTIONS(1653),
    [anon_sym_PIPE] = ACTIONS(1651),
    [anon_sym_COLON2] = ACTIONS(1651),
    [anon_sym_DOT] = ACTIONS(1653),
    [anon_sym_DASH_GT] = ACTIONS(1651),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_RBRACK] = ACTIONS(1653),
    [anon_sym_true] = ACTIONS(1651),
    [anon_sym_false] = ACTIONS(1651),
    [anon_sym_LBRACE] = ACTIONS(1651),
    [anon_sym_SEMI] = ACTIONS(1653),
    [anon_sym_RBRACE] = ACTIONS(1653),
    [anon_sym_constraint] = ACTIONS(1651),
    [anon_sym_val] = ACTIONS(1651),
    [anon_sym_end] = ACTIONS(1651),
    [anon_sym_with] = ACTIONS(1651),
    [anon_sym_COLON_GT] = ACTIONS(1653),
    [anon_sym_object] = ACTIONS(1651),
    [anon_sym_inherit] = ACTIONS(1651),
    [anon_sym_method] = ACTIONS(1651),
    [anon_sym_initializer] = ACTIONS(1651),
    [anon_sym_AMP] = ACTIONS(1651),
    [anon_sym_POUND] = ACTIONS(1651),
    [anon_sym_COLON_COLON] = ACTIONS(1653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1653),
    [anon_sym_then] = ACTIONS(1651),
    [anon_sym_else] = ACTIONS(1651),
    [anon_sym_do] = ACTIONS(1651),
    [anon_sym_done] = ACTIONS(1651),
    [anon_sym_to] = ACTIONS(1651),
    [anon_sym_downto] = ACTIONS(1651),
    [anon_sym_new] = ACTIONS(1651),
    [anon_sym_LBRACE_LT] = ACTIONS(1653),
    [anon_sym_GT_RBRACE] = ACTIONS(1653),
    [anon_sym_begin] = ACTIONS(1651),
    [sym_ocamlyacc_value] = ACTIONS(1653),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1651),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1653),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1651),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1653),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1653),
    [aux_sym_number_token1] = ACTIONS(1653),
    [aux_sym_signed_number_token1] = ACTIONS(1651),
    [anon_sym_SQUOTE] = ACTIONS(1653),
    [anon_sym_DQUOTE] = ACTIONS(1653),
    [sym_prefix_operator] = ACTIONS(1653),
    [aux_sym_sign_operator_token1] = ACTIONS(1651),
    [sym_hash_operator] = ACTIONS(1653),
    [anon_sym_lsl] = ACTIONS(1651),
    [anon_sym_lsr] = ACTIONS(1651),
    [anon_sym_asr] = ACTIONS(1651),
    [aux_sym_pow_operator_token1] = ACTIONS(1653),
    [anon_sym_mod] = ACTIONS(1651),
    [anon_sym_land] = ACTIONS(1651),
    [anon_sym_lor] = ACTIONS(1651),
    [anon_sym_lxor] = ACTIONS(1651),
    [aux_sym_mult_operator_token1] = ACTIONS(1651),
    [aux_sym_add_operator_token1] = ACTIONS(1651),
    [sym_concat_operator] = ACTIONS(1653),
    [sym_rel_operator] = ACTIONS(1651),
    [anon_sym_AMP_AMP] = ACTIONS(1651),
    [anon_sym_or] = ACTIONS(1651),
    [anon_sym_PIPE_PIPE] = ACTIONS(1651),
    [sym_assign_operator] = ACTIONS(1653),
    [sym__uppercase_identifier] = ACTIONS(1653),
    [aux_sym_tag_token1] = ACTIONS(1653),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(934)] = {
    [sym_attribute] = STATE(934),
    [sym__lowercase_identifier] = ACTIONS(1659),
    [anon_sym_TILDE] = ACTIONS(1659),
    [anon_sym_QMARK] = ACTIONS(1659),
    [anon_sym_LPAREN] = ACTIONS(1661),
    [anon_sym_RPAREN] = ACTIONS(1661),
    [anon_sym_COMMA] = ACTIONS(1661),
    [anon_sym_PIPE] = ACTIONS(1659),
    [anon_sym_COLON2] = ACTIONS(1659),
    [anon_sym_DOT] = ACTIONS(1661),
    [anon_sym_DASH_GT] = ACTIONS(1659),
    [anon_sym_LBRACK] = ACTIONS(1659),
    [anon_sym_RBRACK] = ACTIONS(1661),
    [anon_sym_true] = ACTIONS(1659),
    [anon_sym_false] = ACTIONS(1659),
    [anon_sym_LBRACE] = ACTIONS(1659),
    [anon_sym_SEMI] = ACTIONS(1661),
    [anon_sym_RBRACE] = ACTIONS(1661),
    [anon_sym_constraint] = ACTIONS(1659),
    [anon_sym_val] = ACTIONS(1659),
    [anon_sym_end] = ACTIONS(1659),
    [anon_sym_with] = ACTIONS(1659),
    [anon_sym_COLON_GT] = ACTIONS(1661),
    [anon_sym_object] = ACTIONS(1659),
    [anon_sym_inherit] = ACTIONS(1659),
    [anon_sym_method] = ACTIONS(1659),
    [anon_sym_initializer] = ACTIONS(1659),
    [anon_sym_AMP] = ACTIONS(1659),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_COLON_COLON] = ACTIONS(1661),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_then] = ACTIONS(1659),
    [anon_sym_else] = ACTIONS(1659),
    [anon_sym_do] = ACTIONS(1659),
    [anon_sym_done] = ACTIONS(1659),
    [anon_sym_to] = ACTIONS(1659),
    [anon_sym_downto] = ACTIONS(1659),
    [anon_sym_new] = ACTIONS(1659),
    [anon_sym_LBRACE_LT] = ACTIONS(1661),
    [anon_sym_GT_RBRACE] = ACTIONS(1661),
    [anon_sym_begin] = ACTIONS(1659),
    [sym_ocamlyacc_value] = ACTIONS(1661),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1659),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1661),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1659),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1659),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1661),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1661),
    [aux_sym_number_token1] = ACTIONS(1661),
    [aux_sym_signed_number_token1] = ACTIONS(1659),
    [anon_sym_SQUOTE] = ACTIONS(1661),
    [anon_sym_DQUOTE] = ACTIONS(1661),
    [sym_prefix_operator] = ACTIONS(1661),
    [aux_sym_sign_operator_token1] = ACTIONS(1659),
    [sym_hash_operator] = ACTIONS(1661),
    [anon_sym_lsl] = ACTIONS(1659),
    [anon_sym_lsr] = ACTIONS(1659),
    [anon_sym_asr] = ACTIONS(1659),
    [aux_sym_pow_operator_token1] = ACTIONS(1661),
    [anon_sym_mod] = ACTIONS(1659),
    [anon_sym_land] = ACTIONS(1659),
    [anon_sym_lor] = ACTIONS(1659),
    [anon_sym_lxor] = ACTIONS(1659),
    [aux_sym_mult_operator_token1] = ACTIONS(1659),
    [aux_sym_add_operator_token1] = ACTIONS(1659),
    [sym_concat_operator] = ACTIONS(1661),
    [sym_rel_operator] = ACTIONS(1659),
    [anon_sym_AMP_AMP] = ACTIONS(1659),
    [anon_sym_or] = ACTIONS(1659),
    [anon_sym_PIPE_PIPE] = ACTIONS(1659),
    [sym_assign_operator] = ACTIONS(1661),
    [sym__uppercase_identifier] = ACTIONS(1661),
    [aux_sym_tag_token1] = ACTIONS(1661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(935)] = {
    [sym_attribute] = STATE(935),
    [sym__lowercase_identifier] = ACTIONS(1725),
    [anon_sym_TILDE] = ACTIONS(1725),
    [anon_sym_QMARK] = ACTIONS(1725),
    [anon_sym_LPAREN] = ACTIONS(1727),
    [anon_sym_RPAREN] = ACTIONS(1727),
    [anon_sym_COMMA] = ACTIONS(1727),
    [anon_sym_PIPE] = ACTIONS(1725),
    [anon_sym_COLON2] = ACTIONS(1725),
    [anon_sym_DOT] = ACTIONS(1727),
    [anon_sym_DASH_GT] = ACTIONS(1725),
    [anon_sym_LBRACK] = ACTIONS(1725),
    [anon_sym_RBRACK] = ACTIONS(1727),
    [anon_sym_true] = ACTIONS(1725),
    [anon_sym_false] = ACTIONS(1725),
    [anon_sym_LBRACE] = ACTIONS(1725),
    [anon_sym_SEMI] = ACTIONS(1727),
    [anon_sym_RBRACE] = ACTIONS(1727),
    [anon_sym_constraint] = ACTIONS(1725),
    [anon_sym_val] = ACTIONS(1725),
    [anon_sym_end] = ACTIONS(1725),
    [anon_sym_with] = ACTIONS(1725),
    [anon_sym_COLON_GT] = ACTIONS(1727),
    [anon_sym_object] = ACTIONS(1725),
    [anon_sym_inherit] = ACTIONS(1725),
    [anon_sym_method] = ACTIONS(1725),
    [anon_sym_initializer] = ACTIONS(1725),
    [anon_sym_AMP] = ACTIONS(1725),
    [anon_sym_POUND] = ACTIONS(1725),
    [anon_sym_COLON_COLON] = ACTIONS(1727),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1727),
    [anon_sym_then] = ACTIONS(1725),
    [anon_sym_else] = ACTIONS(1725),
    [anon_sym_do] = ACTIONS(1725),
    [anon_sym_done] = ACTIONS(1725),
    [anon_sym_to] = ACTIONS(1725),
    [anon_sym_downto] = ACTIONS(1725),
    [anon_sym_new] = ACTIONS(1725),
    [anon_sym_LBRACE_LT] = ACTIONS(1727),
    [anon_sym_GT_RBRACE] = ACTIONS(1727),
    [anon_sym_begin] = ACTIONS(1725),
    [sym_ocamlyacc_value] = ACTIONS(1727),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1725),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1727),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1725),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1725),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1727),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1727),
    [aux_sym_number_token1] = ACTIONS(1727),
    [aux_sym_signed_number_token1] = ACTIONS(1725),
    [anon_sym_SQUOTE] = ACTIONS(1727),
    [anon_sym_DQUOTE] = ACTIONS(1727),
    [sym_prefix_operator] = ACTIONS(1727),
    [aux_sym_sign_operator_token1] = ACTIONS(1725),
    [sym_hash_operator] = ACTIONS(1727),
    [anon_sym_lsl] = ACTIONS(1725),
    [anon_sym_lsr] = ACTIONS(1725),
    [anon_sym_asr] = ACTIONS(1725),
    [aux_sym_pow_operator_token1] = ACTIONS(1727),
    [anon_sym_mod] = ACTIONS(1725),
    [anon_sym_land] = ACTIONS(1725),
    [anon_sym_lor] = ACTIONS(1725),
    [anon_sym_lxor] = ACTIONS(1725),
    [aux_sym_mult_operator_token1] = ACTIONS(1725),
    [aux_sym_add_operator_token1] = ACTIONS(1725),
    [sym_concat_operator] = ACTIONS(1727),
    [sym_rel_operator] = ACTIONS(1725),
    [anon_sym_AMP_AMP] = ACTIONS(1725),
    [anon_sym_or] = ACTIONS(1725),
    [anon_sym_PIPE_PIPE] = ACTIONS(1725),
    [sym_assign_operator] = ACTIONS(1727),
    [sym__uppercase_identifier] = ACTIONS(1727),
    [aux_sym_tag_token1] = ACTIONS(1727),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(936)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4355),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(936),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_EQ] = ACTIONS(2075),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(2075),
    [anon_sym_COMMA] = ACTIONS(2075),
    [anon_sym_PIPE] = ACTIONS(2075),
    [anon_sym_COLON2] = ACTIONS(2081),
    [anon_sym_DASH_GT] = ACTIONS(2075),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_RBRACK] = ACTIONS(2075),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_as] = ACTIONS(2081),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(2075),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_when] = ACTIONS(2081),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2085),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(937)] = {
    [sym_attribute] = STATE(937),
    [sym__lowercase_identifier] = ACTIONS(1745),
    [anon_sym_SEMI_SEMI] = ACTIONS(1747),
    [anon_sym_let] = ACTIONS(1745),
    [anon_sym_and] = ACTIONS(1745),
    [anon_sym_TILDE] = ACTIONS(1745),
    [anon_sym_QMARK] = ACTIONS(1745),
    [anon_sym_LPAREN] = ACTIONS(1747),
    [anon_sym_external] = ACTIONS(1745),
    [anon_sym_type] = ACTIONS(1745),
    [anon_sym_COMMA] = ACTIONS(1747),
    [anon_sym_PIPE] = ACTIONS(1745),
    [anon_sym_DOT] = ACTIONS(1747),
    [anon_sym_LBRACK] = ACTIONS(1745),
    [anon_sym_RBRACK] = ACTIONS(1747),
    [anon_sym_true] = ACTIONS(1745),
    [anon_sym_false] = ACTIONS(1745),
    [anon_sym_LBRACE] = ACTIONS(1745),
    [anon_sym_SEMI] = ACTIONS(1745),
    [anon_sym_exception] = ACTIONS(1745),
    [anon_sym_module] = ACTIONS(1745),
    [anon_sym_open] = ACTIONS(1745),
    [anon_sym_include] = ACTIONS(1745),
    [anon_sym_class] = ACTIONS(1745),
    [anon_sym_val] = ACTIONS(1745),
    [anon_sym_end] = ACTIONS(1745),
    [anon_sym_object] = ACTIONS(1745),
    [anon_sym_in] = ACTIONS(1745),
    [anon_sym_AMP] = ACTIONS(1745),
    [anon_sym_POUND] = ACTIONS(1745),
    [anon_sym_COLON_COLON] = ACTIONS(1747),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1747),
    [anon_sym_LT_DASH] = ACTIONS(1955),
    [anon_sym_else] = ACTIONS(1745),
    [anon_sym_new] = ACTIONS(1745),
    [anon_sym_LBRACE_LT] = ACTIONS(1747),
    [anon_sym_begin] = ACTIONS(1745),
    [sym_ocamlyacc_value] = ACTIONS(1747),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1745),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1747),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1745),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1745),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1747),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1747),
    [aux_sym_number_token1] = ACTIONS(1747),
    [aux_sym_signed_number_token1] = ACTIONS(1745),
    [anon_sym_SQUOTE] = ACTIONS(1747),
    [anon_sym_DQUOTE] = ACTIONS(1747),
    [sym_prefix_operator] = ACTIONS(1747),
    [aux_sym_sign_operator_token1] = ACTIONS(1745),
    [sym_hash_operator] = ACTIONS(1747),
    [anon_sym_lsl] = ACTIONS(1745),
    [anon_sym_lsr] = ACTIONS(1745),
    [anon_sym_asr] = ACTIONS(1745),
    [aux_sym_pow_operator_token1] = ACTIONS(1747),
    [anon_sym_mod] = ACTIONS(1745),
    [anon_sym_land] = ACTIONS(1745),
    [anon_sym_lor] = ACTIONS(1745),
    [anon_sym_lxor] = ACTIONS(1745),
    [aux_sym_mult_operator_token1] = ACTIONS(1745),
    [aux_sym_add_operator_token1] = ACTIONS(1745),
    [sym_concat_operator] = ACTIONS(1747),
    [sym_rel_operator] = ACTIONS(1745),
    [anon_sym_AMP_AMP] = ACTIONS(1745),
    [anon_sym_or] = ACTIONS(1745),
    [anon_sym_PIPE_PIPE] = ACTIONS(1745),
    [sym_assign_operator] = ACTIONS(1747),
    [sym_let_operator] = ACTIONS(1747),
    [sym_let_and_operator] = ACTIONS(1747),
    [sym__uppercase_identifier] = ACTIONS(1747),
    [aux_sym_directive_token1] = ACTIONS(1745),
    [aux_sym_tag_token1] = ACTIONS(1747),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(938)] = {
    [sym_attribute] = STATE(938),
    [sym__lowercase_identifier] = ACTIONS(1705),
    [anon_sym_SEMI_SEMI] = ACTIONS(1707),
    [anon_sym_let] = ACTIONS(1705),
    [anon_sym_and] = ACTIONS(1705),
    [anon_sym_TILDE] = ACTIONS(1705),
    [anon_sym_QMARK] = ACTIONS(1705),
    [anon_sym_LPAREN] = ACTIONS(1707),
    [anon_sym_external] = ACTIONS(1705),
    [anon_sym_type] = ACTIONS(1705),
    [anon_sym_COMMA] = ACTIONS(1707),
    [anon_sym_PIPE] = ACTIONS(1705),
    [anon_sym_DOT] = ACTIONS(1707),
    [anon_sym_LBRACK] = ACTIONS(1705),
    [anon_sym_RBRACK] = ACTIONS(1707),
    [anon_sym_true] = ACTIONS(1705),
    [anon_sym_false] = ACTIONS(1705),
    [anon_sym_LBRACE] = ACTIONS(1705),
    [anon_sym_SEMI] = ACTIONS(1705),
    [anon_sym_exception] = ACTIONS(1705),
    [anon_sym_module] = ACTIONS(1705),
    [anon_sym_open] = ACTIONS(1705),
    [anon_sym_include] = ACTIONS(1705),
    [anon_sym_class] = ACTIONS(1705),
    [anon_sym_val] = ACTIONS(1705),
    [anon_sym_end] = ACTIONS(1705),
    [anon_sym_object] = ACTIONS(1705),
    [anon_sym_in] = ACTIONS(1705),
    [anon_sym_AMP] = ACTIONS(1705),
    [anon_sym_POUND] = ACTIONS(1705),
    [anon_sym_COLON_COLON] = ACTIONS(1707),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1707),
    [anon_sym_LT_DASH] = ACTIONS(2087),
    [anon_sym_else] = ACTIONS(1705),
    [anon_sym_new] = ACTIONS(1705),
    [anon_sym_LBRACE_LT] = ACTIONS(1707),
    [anon_sym_begin] = ACTIONS(1705),
    [sym_ocamlyacc_value] = ACTIONS(1707),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1705),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1707),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1705),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1705),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1707),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1707),
    [aux_sym_number_token1] = ACTIONS(1707),
    [aux_sym_signed_number_token1] = ACTIONS(1705),
    [anon_sym_SQUOTE] = ACTIONS(1707),
    [anon_sym_DQUOTE] = ACTIONS(1707),
    [sym_prefix_operator] = ACTIONS(1707),
    [aux_sym_sign_operator_token1] = ACTIONS(1705),
    [sym_hash_operator] = ACTIONS(1707),
    [anon_sym_lsl] = ACTIONS(1705),
    [anon_sym_lsr] = ACTIONS(1705),
    [anon_sym_asr] = ACTIONS(1705),
    [aux_sym_pow_operator_token1] = ACTIONS(1707),
    [anon_sym_mod] = ACTIONS(1705),
    [anon_sym_land] = ACTIONS(1705),
    [anon_sym_lor] = ACTIONS(1705),
    [anon_sym_lxor] = ACTIONS(1705),
    [aux_sym_mult_operator_token1] = ACTIONS(1705),
    [aux_sym_add_operator_token1] = ACTIONS(1705),
    [sym_concat_operator] = ACTIONS(1707),
    [sym_rel_operator] = ACTIONS(1705),
    [anon_sym_AMP_AMP] = ACTIONS(1705),
    [anon_sym_or] = ACTIONS(1705),
    [anon_sym_PIPE_PIPE] = ACTIONS(1705),
    [sym_assign_operator] = ACTIONS(1707),
    [sym_let_operator] = ACTIONS(1707),
    [sym_let_and_operator] = ACTIONS(1707),
    [sym__uppercase_identifier] = ACTIONS(1707),
    [aux_sym_directive_token1] = ACTIONS(1705),
    [aux_sym_tag_token1] = ACTIONS(1707),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(939)] = {
    [sym_attribute] = STATE(939),
    [sym__lowercase_identifier] = ACTIONS(1543),
    [anon_sym_SEMI_SEMI] = ACTIONS(1545),
    [anon_sym_let] = ACTIONS(1543),
    [anon_sym_and] = ACTIONS(1543),
    [anon_sym_TILDE] = ACTIONS(1543),
    [anon_sym_QMARK] = ACTIONS(1543),
    [anon_sym_LPAREN] = ACTIONS(1545),
    [anon_sym_external] = ACTIONS(1543),
    [anon_sym_type] = ACTIONS(1543),
    [anon_sym_COMMA] = ACTIONS(1545),
    [anon_sym_PIPE] = ACTIONS(1543),
    [anon_sym_DOT] = ACTIONS(1545),
    [anon_sym_LBRACK] = ACTIONS(1543),
    [anon_sym_RBRACK] = ACTIONS(1545),
    [anon_sym_true] = ACTIONS(1543),
    [anon_sym_false] = ACTIONS(1543),
    [anon_sym_LBRACE] = ACTIONS(1543),
    [anon_sym_SEMI] = ACTIONS(1543),
    [anon_sym_exception] = ACTIONS(1543),
    [anon_sym_module] = ACTIONS(1543),
    [anon_sym_open] = ACTIONS(1543),
    [anon_sym_include] = ACTIONS(1543),
    [anon_sym_class] = ACTIONS(1543),
    [anon_sym_val] = ACTIONS(1543),
    [anon_sym_end] = ACTIONS(1543),
    [anon_sym_object] = ACTIONS(1543),
    [anon_sym_in] = ACTIONS(1543),
    [anon_sym_AMP] = ACTIONS(1543),
    [anon_sym_POUND] = ACTIONS(1543),
    [anon_sym_COLON_COLON] = ACTIONS(1545),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1545),
    [anon_sym_LT_DASH] = ACTIONS(1543),
    [anon_sym_else] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1543),
    [anon_sym_LBRACE_LT] = ACTIONS(1545),
    [anon_sym_begin] = ACTIONS(1543),
    [sym_ocamlyacc_value] = ACTIONS(1545),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1543),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1545),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1543),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1543),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1545),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1545),
    [aux_sym_number_token1] = ACTIONS(1545),
    [aux_sym_signed_number_token1] = ACTIONS(1543),
    [anon_sym_SQUOTE] = ACTIONS(1545),
    [anon_sym_DQUOTE] = ACTIONS(1545),
    [sym_prefix_operator] = ACTIONS(1545),
    [aux_sym_sign_operator_token1] = ACTIONS(1543),
    [sym_hash_operator] = ACTIONS(1545),
    [anon_sym_lsl] = ACTIONS(1543),
    [anon_sym_lsr] = ACTIONS(1543),
    [anon_sym_asr] = ACTIONS(1543),
    [aux_sym_pow_operator_token1] = ACTIONS(1545),
    [anon_sym_mod] = ACTIONS(1543),
    [anon_sym_land] = ACTIONS(1543),
    [anon_sym_lor] = ACTIONS(1543),
    [anon_sym_lxor] = ACTIONS(1543),
    [aux_sym_mult_operator_token1] = ACTIONS(1543),
    [aux_sym_add_operator_token1] = ACTIONS(1543),
    [sym_concat_operator] = ACTIONS(1545),
    [sym_rel_operator] = ACTIONS(1543),
    [anon_sym_AMP_AMP] = ACTIONS(1543),
    [anon_sym_or] = ACTIONS(1543),
    [anon_sym_PIPE_PIPE] = ACTIONS(1543),
    [sym_assign_operator] = ACTIONS(1545),
    [sym_let_operator] = ACTIONS(1545),
    [sym_let_and_operator] = ACTIONS(1545),
    [sym__uppercase_identifier] = ACTIONS(1545),
    [aux_sym_directive_token1] = ACTIONS(1543),
    [aux_sym_tag_token1] = ACTIONS(1545),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(940)] = {
    [sym_attribute] = STATE(940),
    [sym__lowercase_identifier] = ACTIONS(1547),
    [anon_sym_SEMI_SEMI] = ACTIONS(1549),
    [anon_sym_let] = ACTIONS(1547),
    [anon_sym_and] = ACTIONS(1547),
    [anon_sym_TILDE] = ACTIONS(1547),
    [anon_sym_QMARK] = ACTIONS(1547),
    [anon_sym_LPAREN] = ACTIONS(1549),
    [anon_sym_external] = ACTIONS(1547),
    [anon_sym_type] = ACTIONS(1547),
    [anon_sym_COMMA] = ACTIONS(1549),
    [anon_sym_PIPE] = ACTIONS(1547),
    [anon_sym_DOT] = ACTIONS(1549),
    [anon_sym_LBRACK] = ACTIONS(1547),
    [anon_sym_RBRACK] = ACTIONS(1549),
    [anon_sym_true] = ACTIONS(1547),
    [anon_sym_false] = ACTIONS(1547),
    [anon_sym_LBRACE] = ACTIONS(1547),
    [anon_sym_SEMI] = ACTIONS(1547),
    [anon_sym_exception] = ACTIONS(1547),
    [anon_sym_module] = ACTIONS(1547),
    [anon_sym_open] = ACTIONS(1547),
    [anon_sym_include] = ACTIONS(1547),
    [anon_sym_class] = ACTIONS(1547),
    [anon_sym_val] = ACTIONS(1547),
    [anon_sym_end] = ACTIONS(1547),
    [anon_sym_object] = ACTIONS(1547),
    [anon_sym_in] = ACTIONS(1547),
    [anon_sym_AMP] = ACTIONS(1547),
    [anon_sym_POUND] = ACTIONS(1547),
    [anon_sym_COLON_COLON] = ACTIONS(1549),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1549),
    [anon_sym_LT_DASH] = ACTIONS(1547),
    [anon_sym_else] = ACTIONS(1547),
    [anon_sym_new] = ACTIONS(1547),
    [anon_sym_LBRACE_LT] = ACTIONS(1549),
    [anon_sym_begin] = ACTIONS(1547),
    [sym_ocamlyacc_value] = ACTIONS(1549),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1547),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1549),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1547),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1547),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1549),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1549),
    [aux_sym_number_token1] = ACTIONS(1549),
    [aux_sym_signed_number_token1] = ACTIONS(1547),
    [anon_sym_SQUOTE] = ACTIONS(1549),
    [anon_sym_DQUOTE] = ACTIONS(1549),
    [sym_prefix_operator] = ACTIONS(1549),
    [aux_sym_sign_operator_token1] = ACTIONS(1547),
    [sym_hash_operator] = ACTIONS(1549),
    [anon_sym_lsl] = ACTIONS(1547),
    [anon_sym_lsr] = ACTIONS(1547),
    [anon_sym_asr] = ACTIONS(1547),
    [aux_sym_pow_operator_token1] = ACTIONS(1549),
    [anon_sym_mod] = ACTIONS(1547),
    [anon_sym_land] = ACTIONS(1547),
    [anon_sym_lor] = ACTIONS(1547),
    [anon_sym_lxor] = ACTIONS(1547),
    [aux_sym_mult_operator_token1] = ACTIONS(1547),
    [aux_sym_add_operator_token1] = ACTIONS(1547),
    [sym_concat_operator] = ACTIONS(1549),
    [sym_rel_operator] = ACTIONS(1547),
    [anon_sym_AMP_AMP] = ACTIONS(1547),
    [anon_sym_or] = ACTIONS(1547),
    [anon_sym_PIPE_PIPE] = ACTIONS(1547),
    [sym_assign_operator] = ACTIONS(1549),
    [sym_let_operator] = ACTIONS(1549),
    [sym_let_and_operator] = ACTIONS(1549),
    [sym__uppercase_identifier] = ACTIONS(1549),
    [aux_sym_directive_token1] = ACTIONS(1547),
    [aux_sym_tag_token1] = ACTIONS(1549),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(941)] = {
    [sym_attribute] = STATE(941),
    [sym__lowercase_identifier] = ACTIONS(1535),
    [anon_sym_SEMI_SEMI] = ACTIONS(1537),
    [anon_sym_let] = ACTIONS(1535),
    [anon_sym_and] = ACTIONS(1535),
    [anon_sym_TILDE] = ACTIONS(1535),
    [anon_sym_QMARK] = ACTIONS(1535),
    [anon_sym_LPAREN] = ACTIONS(1537),
    [anon_sym_external] = ACTIONS(1535),
    [anon_sym_type] = ACTIONS(1535),
    [anon_sym_COMMA] = ACTIONS(1537),
    [anon_sym_PIPE] = ACTIONS(1535),
    [anon_sym_DOT] = ACTIONS(1537),
    [anon_sym_LBRACK] = ACTIONS(1535),
    [anon_sym_RBRACK] = ACTIONS(1537),
    [anon_sym_true] = ACTIONS(1535),
    [anon_sym_false] = ACTIONS(1535),
    [anon_sym_LBRACE] = ACTIONS(1535),
    [anon_sym_SEMI] = ACTIONS(1535),
    [anon_sym_exception] = ACTIONS(1535),
    [anon_sym_module] = ACTIONS(1535),
    [anon_sym_open] = ACTIONS(1535),
    [anon_sym_include] = ACTIONS(1535),
    [anon_sym_class] = ACTIONS(1535),
    [anon_sym_val] = ACTIONS(1535),
    [anon_sym_end] = ACTIONS(1535),
    [anon_sym_object] = ACTIONS(1535),
    [anon_sym_in] = ACTIONS(1535),
    [anon_sym_AMP] = ACTIONS(1535),
    [anon_sym_POUND] = ACTIONS(1535),
    [anon_sym_COLON_COLON] = ACTIONS(1537),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1537),
    [anon_sym_LT_DASH] = ACTIONS(1535),
    [anon_sym_else] = ACTIONS(1535),
    [anon_sym_new] = ACTIONS(1535),
    [anon_sym_LBRACE_LT] = ACTIONS(1537),
    [anon_sym_begin] = ACTIONS(1535),
    [sym_ocamlyacc_value] = ACTIONS(1537),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1535),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1537),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1535),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1535),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1537),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1537),
    [aux_sym_number_token1] = ACTIONS(1537),
    [aux_sym_signed_number_token1] = ACTIONS(1535),
    [anon_sym_SQUOTE] = ACTIONS(1537),
    [anon_sym_DQUOTE] = ACTIONS(1537),
    [sym_prefix_operator] = ACTIONS(1537),
    [aux_sym_sign_operator_token1] = ACTIONS(1535),
    [sym_hash_operator] = ACTIONS(1537),
    [anon_sym_lsl] = ACTIONS(1535),
    [anon_sym_lsr] = ACTIONS(1535),
    [anon_sym_asr] = ACTIONS(1535),
    [aux_sym_pow_operator_token1] = ACTIONS(1537),
    [anon_sym_mod] = ACTIONS(1535),
    [anon_sym_land] = ACTIONS(1535),
    [anon_sym_lor] = ACTIONS(1535),
    [anon_sym_lxor] = ACTIONS(1535),
    [aux_sym_mult_operator_token1] = ACTIONS(1535),
    [aux_sym_add_operator_token1] = ACTIONS(1535),
    [sym_concat_operator] = ACTIONS(1537),
    [sym_rel_operator] = ACTIONS(1535),
    [anon_sym_AMP_AMP] = ACTIONS(1535),
    [anon_sym_or] = ACTIONS(1535),
    [anon_sym_PIPE_PIPE] = ACTIONS(1535),
    [sym_assign_operator] = ACTIONS(1537),
    [sym_let_operator] = ACTIONS(1537),
    [sym_let_and_operator] = ACTIONS(1537),
    [sym__uppercase_identifier] = ACTIONS(1537),
    [aux_sym_directive_token1] = ACTIONS(1535),
    [aux_sym_tag_token1] = ACTIONS(1537),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(942)] = {
    [sym_attribute] = STATE(942),
    [sym__lowercase_identifier] = ACTIONS(1531),
    [anon_sym_SEMI_SEMI] = ACTIONS(1533),
    [anon_sym_let] = ACTIONS(1531),
    [anon_sym_and] = ACTIONS(1531),
    [anon_sym_TILDE] = ACTIONS(1531),
    [anon_sym_QMARK] = ACTIONS(1531),
    [anon_sym_LPAREN] = ACTIONS(1533),
    [anon_sym_external] = ACTIONS(1531),
    [anon_sym_type] = ACTIONS(1531),
    [anon_sym_COMMA] = ACTIONS(1533),
    [anon_sym_PIPE] = ACTIONS(1531),
    [anon_sym_DOT] = ACTIONS(1533),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_RBRACK] = ACTIONS(1533),
    [anon_sym_true] = ACTIONS(1531),
    [anon_sym_false] = ACTIONS(1531),
    [anon_sym_LBRACE] = ACTIONS(1531),
    [anon_sym_SEMI] = ACTIONS(1531),
    [anon_sym_exception] = ACTIONS(1531),
    [anon_sym_module] = ACTIONS(1531),
    [anon_sym_open] = ACTIONS(1531),
    [anon_sym_include] = ACTIONS(1531),
    [anon_sym_class] = ACTIONS(1531),
    [anon_sym_val] = ACTIONS(1531),
    [anon_sym_end] = ACTIONS(1531),
    [anon_sym_object] = ACTIONS(1531),
    [anon_sym_in] = ACTIONS(1531),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_POUND] = ACTIONS(1531),
    [anon_sym_COLON_COLON] = ACTIONS(1533),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LT_DASH] = ACTIONS(1531),
    [anon_sym_else] = ACTIONS(1531),
    [anon_sym_new] = ACTIONS(1531),
    [anon_sym_LBRACE_LT] = ACTIONS(1533),
    [anon_sym_begin] = ACTIONS(1531),
    [sym_ocamlyacc_value] = ACTIONS(1533),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1531),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1533),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1531),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1531),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1533),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1533),
    [aux_sym_number_token1] = ACTIONS(1533),
    [aux_sym_signed_number_token1] = ACTIONS(1531),
    [anon_sym_SQUOTE] = ACTIONS(1533),
    [anon_sym_DQUOTE] = ACTIONS(1533),
    [sym_prefix_operator] = ACTIONS(1533),
    [aux_sym_sign_operator_token1] = ACTIONS(1531),
    [sym_hash_operator] = ACTIONS(1533),
    [anon_sym_lsl] = ACTIONS(1531),
    [anon_sym_lsr] = ACTIONS(1531),
    [anon_sym_asr] = ACTIONS(1531),
    [aux_sym_pow_operator_token1] = ACTIONS(1533),
    [anon_sym_mod] = ACTIONS(1531),
    [anon_sym_land] = ACTIONS(1531),
    [anon_sym_lor] = ACTIONS(1531),
    [anon_sym_lxor] = ACTIONS(1531),
    [aux_sym_mult_operator_token1] = ACTIONS(1531),
    [aux_sym_add_operator_token1] = ACTIONS(1531),
    [sym_concat_operator] = ACTIONS(1533),
    [sym_rel_operator] = ACTIONS(1531),
    [anon_sym_AMP_AMP] = ACTIONS(1531),
    [anon_sym_or] = ACTIONS(1531),
    [anon_sym_PIPE_PIPE] = ACTIONS(1531),
    [sym_assign_operator] = ACTIONS(1533),
    [sym_let_operator] = ACTIONS(1533),
    [sym_let_and_operator] = ACTIONS(1533),
    [sym__uppercase_identifier] = ACTIONS(1533),
    [aux_sym_directive_token1] = ACTIONS(1531),
    [aux_sym_tag_token1] = ACTIONS(1533),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(943)] = {
    [sym_attribute] = STATE(943),
    [sym__lowercase_identifier] = ACTIONS(1563),
    [anon_sym_SEMI_SEMI] = ACTIONS(1565),
    [anon_sym_let] = ACTIONS(1563),
    [anon_sym_and] = ACTIONS(1563),
    [anon_sym_TILDE] = ACTIONS(1563),
    [anon_sym_QMARK] = ACTIONS(1563),
    [anon_sym_LPAREN] = ACTIONS(1565),
    [anon_sym_external] = ACTIONS(1563),
    [anon_sym_type] = ACTIONS(1563),
    [anon_sym_COMMA] = ACTIONS(1565),
    [anon_sym_PIPE] = ACTIONS(1563),
    [anon_sym_DOT] = ACTIONS(1565),
    [anon_sym_LBRACK] = ACTIONS(1563),
    [anon_sym_RBRACK] = ACTIONS(1565),
    [anon_sym_true] = ACTIONS(1563),
    [anon_sym_false] = ACTIONS(1563),
    [anon_sym_LBRACE] = ACTIONS(1563),
    [anon_sym_SEMI] = ACTIONS(1563),
    [anon_sym_exception] = ACTIONS(1563),
    [anon_sym_module] = ACTIONS(1563),
    [anon_sym_open] = ACTIONS(1563),
    [anon_sym_include] = ACTIONS(1563),
    [anon_sym_class] = ACTIONS(1563),
    [anon_sym_val] = ACTIONS(1563),
    [anon_sym_end] = ACTIONS(1563),
    [anon_sym_object] = ACTIONS(1563),
    [anon_sym_in] = ACTIONS(1563),
    [anon_sym_AMP] = ACTIONS(1563),
    [anon_sym_POUND] = ACTIONS(1563),
    [anon_sym_COLON_COLON] = ACTIONS(1565),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1565),
    [anon_sym_LT_DASH] = ACTIONS(1563),
    [anon_sym_else] = ACTIONS(1563),
    [anon_sym_new] = ACTIONS(1563),
    [anon_sym_LBRACE_LT] = ACTIONS(1565),
    [anon_sym_begin] = ACTIONS(1563),
    [sym_ocamlyacc_value] = ACTIONS(1565),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1563),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1565),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1563),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1563),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1565),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1565),
    [aux_sym_number_token1] = ACTIONS(1565),
    [aux_sym_signed_number_token1] = ACTIONS(1563),
    [anon_sym_SQUOTE] = ACTIONS(1565),
    [anon_sym_DQUOTE] = ACTIONS(1565),
    [sym_prefix_operator] = ACTIONS(1565),
    [aux_sym_sign_operator_token1] = ACTIONS(1563),
    [sym_hash_operator] = ACTIONS(1565),
    [anon_sym_lsl] = ACTIONS(1563),
    [anon_sym_lsr] = ACTIONS(1563),
    [anon_sym_asr] = ACTIONS(1563),
    [aux_sym_pow_operator_token1] = ACTIONS(1565),
    [anon_sym_mod] = ACTIONS(1563),
    [anon_sym_land] = ACTIONS(1563),
    [anon_sym_lor] = ACTIONS(1563),
    [anon_sym_lxor] = ACTIONS(1563),
    [aux_sym_mult_operator_token1] = ACTIONS(1563),
    [aux_sym_add_operator_token1] = ACTIONS(1563),
    [sym_concat_operator] = ACTIONS(1565),
    [sym_rel_operator] = ACTIONS(1563),
    [anon_sym_AMP_AMP] = ACTIONS(1563),
    [anon_sym_or] = ACTIONS(1563),
    [anon_sym_PIPE_PIPE] = ACTIONS(1563),
    [sym_assign_operator] = ACTIONS(1565),
    [sym_let_operator] = ACTIONS(1565),
    [sym_let_and_operator] = ACTIONS(1565),
    [sym__uppercase_identifier] = ACTIONS(1565),
    [aux_sym_directive_token1] = ACTIONS(1563),
    [aux_sym_tag_token1] = ACTIONS(1565),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(944)] = {
    [sym_attribute] = STATE(944),
    [sym__lowercase_identifier] = ACTIONS(1555),
    [anon_sym_SEMI_SEMI] = ACTIONS(1557),
    [anon_sym_let] = ACTIONS(1555),
    [anon_sym_and] = ACTIONS(1555),
    [anon_sym_TILDE] = ACTIONS(1555),
    [anon_sym_QMARK] = ACTIONS(1555),
    [anon_sym_LPAREN] = ACTIONS(1557),
    [anon_sym_external] = ACTIONS(1555),
    [anon_sym_type] = ACTIONS(1555),
    [anon_sym_COMMA] = ACTIONS(1557),
    [anon_sym_PIPE] = ACTIONS(1555),
    [anon_sym_DOT] = ACTIONS(1557),
    [anon_sym_LBRACK] = ACTIONS(1555),
    [anon_sym_RBRACK] = ACTIONS(1557),
    [anon_sym_true] = ACTIONS(1555),
    [anon_sym_false] = ACTIONS(1555),
    [anon_sym_LBRACE] = ACTIONS(1555),
    [anon_sym_SEMI] = ACTIONS(1555),
    [anon_sym_exception] = ACTIONS(1555),
    [anon_sym_module] = ACTIONS(1555),
    [anon_sym_open] = ACTIONS(1555),
    [anon_sym_include] = ACTIONS(1555),
    [anon_sym_class] = ACTIONS(1555),
    [anon_sym_val] = ACTIONS(1555),
    [anon_sym_end] = ACTIONS(1555),
    [anon_sym_object] = ACTIONS(1555),
    [anon_sym_in] = ACTIONS(1555),
    [anon_sym_AMP] = ACTIONS(1555),
    [anon_sym_POUND] = ACTIONS(1555),
    [anon_sym_COLON_COLON] = ACTIONS(1557),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1557),
    [anon_sym_LT_DASH] = ACTIONS(1555),
    [anon_sym_else] = ACTIONS(1555),
    [anon_sym_new] = ACTIONS(1555),
    [anon_sym_LBRACE_LT] = ACTIONS(1557),
    [anon_sym_begin] = ACTIONS(1555),
    [sym_ocamlyacc_value] = ACTIONS(1557),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1555),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1557),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1555),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1555),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1557),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1557),
    [aux_sym_number_token1] = ACTIONS(1557),
    [aux_sym_signed_number_token1] = ACTIONS(1555),
    [anon_sym_SQUOTE] = ACTIONS(1557),
    [anon_sym_DQUOTE] = ACTIONS(1557),
    [sym_prefix_operator] = ACTIONS(1557),
    [aux_sym_sign_operator_token1] = ACTIONS(1555),
    [sym_hash_operator] = ACTIONS(1557),
    [anon_sym_lsl] = ACTIONS(1555),
    [anon_sym_lsr] = ACTIONS(1555),
    [anon_sym_asr] = ACTIONS(1555),
    [aux_sym_pow_operator_token1] = ACTIONS(1557),
    [anon_sym_mod] = ACTIONS(1555),
    [anon_sym_land] = ACTIONS(1555),
    [anon_sym_lor] = ACTIONS(1555),
    [anon_sym_lxor] = ACTIONS(1555),
    [aux_sym_mult_operator_token1] = ACTIONS(1555),
    [aux_sym_add_operator_token1] = ACTIONS(1555),
    [sym_concat_operator] = ACTIONS(1557),
    [sym_rel_operator] = ACTIONS(1555),
    [anon_sym_AMP_AMP] = ACTIONS(1555),
    [anon_sym_or] = ACTIONS(1555),
    [anon_sym_PIPE_PIPE] = ACTIONS(1555),
    [sym_assign_operator] = ACTIONS(1557),
    [sym_let_operator] = ACTIONS(1557),
    [sym_let_and_operator] = ACTIONS(1557),
    [sym__uppercase_identifier] = ACTIONS(1557),
    [aux_sym_directive_token1] = ACTIONS(1555),
    [aux_sym_tag_token1] = ACTIONS(1557),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(945)] = {
    [sym_attribute] = STATE(945),
    [sym__lowercase_identifier] = ACTIONS(1539),
    [anon_sym_SEMI_SEMI] = ACTIONS(1541),
    [anon_sym_let] = ACTIONS(1539),
    [anon_sym_and] = ACTIONS(1539),
    [anon_sym_TILDE] = ACTIONS(1539),
    [anon_sym_QMARK] = ACTIONS(1539),
    [anon_sym_LPAREN] = ACTIONS(1541),
    [anon_sym_external] = ACTIONS(1539),
    [anon_sym_type] = ACTIONS(1539),
    [anon_sym_COMMA] = ACTIONS(1541),
    [anon_sym_PIPE] = ACTIONS(1539),
    [anon_sym_DOT] = ACTIONS(1541),
    [anon_sym_LBRACK] = ACTIONS(1539),
    [anon_sym_RBRACK] = ACTIONS(1541),
    [anon_sym_true] = ACTIONS(1539),
    [anon_sym_false] = ACTIONS(1539),
    [anon_sym_LBRACE] = ACTIONS(1539),
    [anon_sym_SEMI] = ACTIONS(1539),
    [anon_sym_exception] = ACTIONS(1539),
    [anon_sym_module] = ACTIONS(1539),
    [anon_sym_open] = ACTIONS(1539),
    [anon_sym_include] = ACTIONS(1539),
    [anon_sym_class] = ACTIONS(1539),
    [anon_sym_val] = ACTIONS(1539),
    [anon_sym_end] = ACTIONS(1539),
    [anon_sym_object] = ACTIONS(1539),
    [anon_sym_in] = ACTIONS(1539),
    [anon_sym_AMP] = ACTIONS(1539),
    [anon_sym_POUND] = ACTIONS(1539),
    [anon_sym_COLON_COLON] = ACTIONS(1541),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1541),
    [anon_sym_LT_DASH] = ACTIONS(1539),
    [anon_sym_else] = ACTIONS(1539),
    [anon_sym_new] = ACTIONS(1539),
    [anon_sym_LBRACE_LT] = ACTIONS(1541),
    [anon_sym_begin] = ACTIONS(1539),
    [sym_ocamlyacc_value] = ACTIONS(1541),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1539),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1541),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1539),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1539),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1541),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1541),
    [aux_sym_number_token1] = ACTIONS(1541),
    [aux_sym_signed_number_token1] = ACTIONS(1539),
    [anon_sym_SQUOTE] = ACTIONS(1541),
    [anon_sym_DQUOTE] = ACTIONS(1541),
    [sym_prefix_operator] = ACTIONS(1541),
    [aux_sym_sign_operator_token1] = ACTIONS(1539),
    [sym_hash_operator] = ACTIONS(1541),
    [anon_sym_lsl] = ACTIONS(1539),
    [anon_sym_lsr] = ACTIONS(1539),
    [anon_sym_asr] = ACTIONS(1539),
    [aux_sym_pow_operator_token1] = ACTIONS(1541),
    [anon_sym_mod] = ACTIONS(1539),
    [anon_sym_land] = ACTIONS(1539),
    [anon_sym_lor] = ACTIONS(1539),
    [anon_sym_lxor] = ACTIONS(1539),
    [aux_sym_mult_operator_token1] = ACTIONS(1539),
    [aux_sym_add_operator_token1] = ACTIONS(1539),
    [sym_concat_operator] = ACTIONS(1541),
    [sym_rel_operator] = ACTIONS(1539),
    [anon_sym_AMP_AMP] = ACTIONS(1539),
    [anon_sym_or] = ACTIONS(1539),
    [anon_sym_PIPE_PIPE] = ACTIONS(1539),
    [sym_assign_operator] = ACTIONS(1541),
    [sym_let_operator] = ACTIONS(1541),
    [sym_let_and_operator] = ACTIONS(1541),
    [sym__uppercase_identifier] = ACTIONS(1541),
    [aux_sym_directive_token1] = ACTIONS(1539),
    [aux_sym_tag_token1] = ACTIONS(1541),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(946)] = {
    [sym_attribute] = STATE(946),
    [sym__lowercase_identifier] = ACTIONS(1559),
    [anon_sym_SEMI_SEMI] = ACTIONS(1561),
    [anon_sym_let] = ACTIONS(1559),
    [anon_sym_and] = ACTIONS(1559),
    [anon_sym_TILDE] = ACTIONS(1559),
    [anon_sym_QMARK] = ACTIONS(1559),
    [anon_sym_LPAREN] = ACTIONS(1561),
    [anon_sym_external] = ACTIONS(1559),
    [anon_sym_type] = ACTIONS(1559),
    [anon_sym_COMMA] = ACTIONS(1561),
    [anon_sym_PIPE] = ACTIONS(1559),
    [anon_sym_DOT] = ACTIONS(1561),
    [anon_sym_LBRACK] = ACTIONS(1559),
    [anon_sym_RBRACK] = ACTIONS(1561),
    [anon_sym_true] = ACTIONS(1559),
    [anon_sym_false] = ACTIONS(1559),
    [anon_sym_LBRACE] = ACTIONS(1559),
    [anon_sym_SEMI] = ACTIONS(1559),
    [anon_sym_exception] = ACTIONS(1559),
    [anon_sym_module] = ACTIONS(1559),
    [anon_sym_open] = ACTIONS(1559),
    [anon_sym_include] = ACTIONS(1559),
    [anon_sym_class] = ACTIONS(1559),
    [anon_sym_val] = ACTIONS(1559),
    [anon_sym_end] = ACTIONS(1559),
    [anon_sym_object] = ACTIONS(1559),
    [anon_sym_in] = ACTIONS(1559),
    [anon_sym_AMP] = ACTIONS(1559),
    [anon_sym_POUND] = ACTIONS(1559),
    [anon_sym_COLON_COLON] = ACTIONS(1561),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1561),
    [anon_sym_LT_DASH] = ACTIONS(1559),
    [anon_sym_else] = ACTIONS(1559),
    [anon_sym_new] = ACTIONS(1559),
    [anon_sym_LBRACE_LT] = ACTIONS(1561),
    [anon_sym_begin] = ACTIONS(1559),
    [sym_ocamlyacc_value] = ACTIONS(1561),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1559),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1561),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1559),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1559),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1561),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1561),
    [aux_sym_number_token1] = ACTIONS(1561),
    [aux_sym_signed_number_token1] = ACTIONS(1559),
    [anon_sym_SQUOTE] = ACTIONS(1561),
    [anon_sym_DQUOTE] = ACTIONS(1561),
    [sym_prefix_operator] = ACTIONS(1561),
    [aux_sym_sign_operator_token1] = ACTIONS(1559),
    [sym_hash_operator] = ACTIONS(1561),
    [anon_sym_lsl] = ACTIONS(1559),
    [anon_sym_lsr] = ACTIONS(1559),
    [anon_sym_asr] = ACTIONS(1559),
    [aux_sym_pow_operator_token1] = ACTIONS(1561),
    [anon_sym_mod] = ACTIONS(1559),
    [anon_sym_land] = ACTIONS(1559),
    [anon_sym_lor] = ACTIONS(1559),
    [anon_sym_lxor] = ACTIONS(1559),
    [aux_sym_mult_operator_token1] = ACTIONS(1559),
    [aux_sym_add_operator_token1] = ACTIONS(1559),
    [sym_concat_operator] = ACTIONS(1561),
    [sym_rel_operator] = ACTIONS(1559),
    [anon_sym_AMP_AMP] = ACTIONS(1559),
    [anon_sym_or] = ACTIONS(1559),
    [anon_sym_PIPE_PIPE] = ACTIONS(1559),
    [sym_assign_operator] = ACTIONS(1561),
    [sym_let_operator] = ACTIONS(1561),
    [sym_let_and_operator] = ACTIONS(1561),
    [sym__uppercase_identifier] = ACTIONS(1561),
    [aux_sym_directive_token1] = ACTIONS(1559),
    [aux_sym_tag_token1] = ACTIONS(1561),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(947)] = {
    [sym_attribute] = STATE(947),
    [sym__lowercase_identifier] = ACTIONS(1551),
    [anon_sym_SEMI_SEMI] = ACTIONS(1553),
    [anon_sym_let] = ACTIONS(1551),
    [anon_sym_and] = ACTIONS(1551),
    [anon_sym_TILDE] = ACTIONS(1551),
    [anon_sym_QMARK] = ACTIONS(1551),
    [anon_sym_LPAREN] = ACTIONS(1553),
    [anon_sym_external] = ACTIONS(1551),
    [anon_sym_type] = ACTIONS(1551),
    [anon_sym_COMMA] = ACTIONS(1553),
    [anon_sym_PIPE] = ACTIONS(1551),
    [anon_sym_DOT] = ACTIONS(1553),
    [anon_sym_LBRACK] = ACTIONS(1551),
    [anon_sym_RBRACK] = ACTIONS(1553),
    [anon_sym_true] = ACTIONS(1551),
    [anon_sym_false] = ACTIONS(1551),
    [anon_sym_LBRACE] = ACTIONS(1551),
    [anon_sym_SEMI] = ACTIONS(1551),
    [anon_sym_exception] = ACTIONS(1551),
    [anon_sym_module] = ACTIONS(1551),
    [anon_sym_open] = ACTIONS(1551),
    [anon_sym_include] = ACTIONS(1551),
    [anon_sym_class] = ACTIONS(1551),
    [anon_sym_val] = ACTIONS(1551),
    [anon_sym_end] = ACTIONS(1551),
    [anon_sym_object] = ACTIONS(1551),
    [anon_sym_in] = ACTIONS(1551),
    [anon_sym_AMP] = ACTIONS(1551),
    [anon_sym_POUND] = ACTIONS(1551),
    [anon_sym_COLON_COLON] = ACTIONS(1553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1553),
    [anon_sym_LT_DASH] = ACTIONS(1551),
    [anon_sym_else] = ACTIONS(1551),
    [anon_sym_new] = ACTIONS(1551),
    [anon_sym_LBRACE_LT] = ACTIONS(1553),
    [anon_sym_begin] = ACTIONS(1551),
    [sym_ocamlyacc_value] = ACTIONS(1553),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1551),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1553),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1551),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1551),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1553),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1553),
    [aux_sym_number_token1] = ACTIONS(1553),
    [aux_sym_signed_number_token1] = ACTIONS(1551),
    [anon_sym_SQUOTE] = ACTIONS(1553),
    [anon_sym_DQUOTE] = ACTIONS(1553),
    [sym_prefix_operator] = ACTIONS(1553),
    [aux_sym_sign_operator_token1] = ACTIONS(1551),
    [sym_hash_operator] = ACTIONS(1553),
    [anon_sym_lsl] = ACTIONS(1551),
    [anon_sym_lsr] = ACTIONS(1551),
    [anon_sym_asr] = ACTIONS(1551),
    [aux_sym_pow_operator_token1] = ACTIONS(1553),
    [anon_sym_mod] = ACTIONS(1551),
    [anon_sym_land] = ACTIONS(1551),
    [anon_sym_lor] = ACTIONS(1551),
    [anon_sym_lxor] = ACTIONS(1551),
    [aux_sym_mult_operator_token1] = ACTIONS(1551),
    [aux_sym_add_operator_token1] = ACTIONS(1551),
    [sym_concat_operator] = ACTIONS(1553),
    [sym_rel_operator] = ACTIONS(1551),
    [anon_sym_AMP_AMP] = ACTIONS(1551),
    [anon_sym_or] = ACTIONS(1551),
    [anon_sym_PIPE_PIPE] = ACTIONS(1551),
    [sym_assign_operator] = ACTIONS(1553),
    [sym_let_operator] = ACTIONS(1553),
    [sym_let_and_operator] = ACTIONS(1553),
    [sym__uppercase_identifier] = ACTIONS(1553),
    [aux_sym_directive_token1] = ACTIONS(1551),
    [aux_sym_tag_token1] = ACTIONS(1553),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(948)] = {
    [sym_attribute] = STATE(948),
    [sym__lowercase_identifier] = ACTIONS(1615),
    [anon_sym_SEMI_SEMI] = ACTIONS(1617),
    [anon_sym_let] = ACTIONS(1615),
    [anon_sym_and] = ACTIONS(1615),
    [anon_sym_TILDE] = ACTIONS(1615),
    [anon_sym_QMARK] = ACTIONS(1615),
    [anon_sym_LPAREN] = ACTIONS(1617),
    [anon_sym_external] = ACTIONS(1615),
    [anon_sym_type] = ACTIONS(1615),
    [anon_sym_COMMA] = ACTIONS(1617),
    [anon_sym_PIPE] = ACTIONS(1615),
    [anon_sym_DOT] = ACTIONS(835),
    [anon_sym_LBRACK] = ACTIONS(1615),
    [anon_sym_RBRACK] = ACTIONS(1617),
    [anon_sym_true] = ACTIONS(1615),
    [anon_sym_false] = ACTIONS(1615),
    [anon_sym_LBRACE] = ACTIONS(1615),
    [anon_sym_SEMI] = ACTIONS(1615),
    [anon_sym_exception] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(1615),
    [anon_sym_open] = ACTIONS(1615),
    [anon_sym_include] = ACTIONS(1615),
    [anon_sym_class] = ACTIONS(1615),
    [anon_sym_val] = ACTIONS(1615),
    [anon_sym_end] = ACTIONS(1615),
    [anon_sym_object] = ACTIONS(1615),
    [anon_sym_in] = ACTIONS(1615),
    [anon_sym_AMP] = ACTIONS(1615),
    [anon_sym_POUND] = ACTIONS(837),
    [anon_sym_COLON_COLON] = ACTIONS(1617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1617),
    [anon_sym_else] = ACTIONS(1615),
    [anon_sym_new] = ACTIONS(1615),
    [anon_sym_LBRACE_LT] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1615),
    [sym_ocamlyacc_value] = ACTIONS(1617),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1615),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1617),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1615),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1615),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1617),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1617),
    [aux_sym_number_token1] = ACTIONS(1617),
    [aux_sym_signed_number_token1] = ACTIONS(1615),
    [anon_sym_SQUOTE] = ACTIONS(1617),
    [anon_sym_DQUOTE] = ACTIONS(1617),
    [sym_prefix_operator] = ACTIONS(1617),
    [aux_sym_sign_operator_token1] = ACTIONS(1615),
    [sym_hash_operator] = ACTIONS(839),
    [anon_sym_lsl] = ACTIONS(1615),
    [anon_sym_lsr] = ACTIONS(1615),
    [anon_sym_asr] = ACTIONS(1615),
    [aux_sym_pow_operator_token1] = ACTIONS(1617),
    [anon_sym_mod] = ACTIONS(1615),
    [anon_sym_land] = ACTIONS(1615),
    [anon_sym_lor] = ACTIONS(1615),
    [anon_sym_lxor] = ACTIONS(1615),
    [aux_sym_mult_operator_token1] = ACTIONS(1615),
    [aux_sym_add_operator_token1] = ACTIONS(1615),
    [sym_concat_operator] = ACTIONS(1617),
    [sym_rel_operator] = ACTIONS(1615),
    [anon_sym_AMP_AMP] = ACTIONS(1615),
    [anon_sym_or] = ACTIONS(1615),
    [anon_sym_PIPE_PIPE] = ACTIONS(1615),
    [sym_assign_operator] = ACTIONS(1617),
    [sym_let_operator] = ACTIONS(1617),
    [sym_let_and_operator] = ACTIONS(1617),
    [sym__uppercase_identifier] = ACTIONS(1617),
    [aux_sym_directive_token1] = ACTIONS(1615),
    [aux_sym_tag_token1] = ACTIONS(1617),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(949)] = {
    [sym_attribute] = STATE(949),
    [sym__lowercase_identifier] = ACTIONS(1725),
    [anon_sym_SEMI_SEMI] = ACTIONS(1727),
    [anon_sym_let] = ACTIONS(1725),
    [anon_sym_and] = ACTIONS(1725),
    [anon_sym_TILDE] = ACTIONS(1725),
    [anon_sym_QMARK] = ACTIONS(1725),
    [anon_sym_LPAREN] = ACTIONS(1727),
    [anon_sym_external] = ACTIONS(1725),
    [anon_sym_type] = ACTIONS(1725),
    [anon_sym_COMMA] = ACTIONS(1727),
    [anon_sym_PIPE] = ACTIONS(1725),
    [anon_sym_DOT] = ACTIONS(1727),
    [anon_sym_LBRACK] = ACTIONS(1725),
    [anon_sym_RBRACK] = ACTIONS(1727),
    [anon_sym_true] = ACTIONS(1725),
    [anon_sym_false] = ACTIONS(1725),
    [anon_sym_LBRACE] = ACTIONS(1725),
    [anon_sym_SEMI] = ACTIONS(1725),
    [anon_sym_exception] = ACTIONS(1725),
    [anon_sym_module] = ACTIONS(1725),
    [anon_sym_open] = ACTIONS(1725),
    [anon_sym_include] = ACTIONS(1725),
    [anon_sym_class] = ACTIONS(1725),
    [anon_sym_val] = ACTIONS(1725),
    [anon_sym_end] = ACTIONS(1725),
    [anon_sym_object] = ACTIONS(1725),
    [anon_sym_in] = ACTIONS(1725),
    [anon_sym_AMP] = ACTIONS(1725),
    [anon_sym_POUND] = ACTIONS(1725),
    [anon_sym_COLON_COLON] = ACTIONS(1727),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1727),
    [anon_sym_else] = ACTIONS(1725),
    [anon_sym_new] = ACTIONS(1725),
    [anon_sym_LBRACE_LT] = ACTIONS(1727),
    [anon_sym_begin] = ACTIONS(1725),
    [sym_ocamlyacc_value] = ACTIONS(1727),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1725),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1727),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1725),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1725),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1727),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1727),
    [aux_sym_number_token1] = ACTIONS(1727),
    [aux_sym_signed_number_token1] = ACTIONS(1725),
    [anon_sym_SQUOTE] = ACTIONS(1727),
    [anon_sym_DQUOTE] = ACTIONS(1727),
    [sym_prefix_operator] = ACTIONS(1727),
    [aux_sym_sign_operator_token1] = ACTIONS(1725),
    [sym_hash_operator] = ACTIONS(1727),
    [anon_sym_lsl] = ACTIONS(1725),
    [anon_sym_lsr] = ACTIONS(1725),
    [anon_sym_asr] = ACTIONS(1725),
    [aux_sym_pow_operator_token1] = ACTIONS(1727),
    [anon_sym_mod] = ACTIONS(1725),
    [anon_sym_land] = ACTIONS(1725),
    [anon_sym_lor] = ACTIONS(1725),
    [anon_sym_lxor] = ACTIONS(1725),
    [aux_sym_mult_operator_token1] = ACTIONS(1725),
    [aux_sym_add_operator_token1] = ACTIONS(1725),
    [sym_concat_operator] = ACTIONS(1727),
    [sym_rel_operator] = ACTIONS(1725),
    [anon_sym_AMP_AMP] = ACTIONS(1725),
    [anon_sym_or] = ACTIONS(1725),
    [anon_sym_PIPE_PIPE] = ACTIONS(1725),
    [sym_assign_operator] = ACTIONS(1727),
    [sym_let_operator] = ACTIONS(1727),
    [sym_let_and_operator] = ACTIONS(1727),
    [sym__uppercase_identifier] = ACTIONS(1727),
    [aux_sym_directive_token1] = ACTIONS(1725),
    [aux_sym_tag_token1] = ACTIONS(1727),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(950)] = {
    [sym_attribute] = STATE(950),
    [sym__lowercase_identifier] = ACTIONS(1769),
    [anon_sym_SEMI_SEMI] = ACTIONS(1771),
    [anon_sym_let] = ACTIONS(1769),
    [anon_sym_and] = ACTIONS(1769),
    [anon_sym_TILDE] = ACTIONS(1769),
    [anon_sym_QMARK] = ACTIONS(1769),
    [anon_sym_LPAREN] = ACTIONS(1771),
    [anon_sym_external] = ACTIONS(1769),
    [anon_sym_type] = ACTIONS(1769),
    [anon_sym_COMMA] = ACTIONS(1771),
    [anon_sym_PIPE] = ACTIONS(1769),
    [anon_sym_DOT] = ACTIONS(1771),
    [anon_sym_LBRACK] = ACTIONS(1769),
    [anon_sym_RBRACK] = ACTIONS(1771),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1769),
    [anon_sym_SEMI] = ACTIONS(1769),
    [anon_sym_exception] = ACTIONS(1769),
    [anon_sym_module] = ACTIONS(1769),
    [anon_sym_open] = ACTIONS(1769),
    [anon_sym_include] = ACTIONS(1769),
    [anon_sym_class] = ACTIONS(1769),
    [anon_sym_val] = ACTIONS(1769),
    [anon_sym_end] = ACTIONS(1769),
    [anon_sym_object] = ACTIONS(1769),
    [anon_sym_in] = ACTIONS(1769),
    [anon_sym_AMP] = ACTIONS(1769),
    [anon_sym_POUND] = ACTIONS(1769),
    [anon_sym_COLON_COLON] = ACTIONS(1771),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1771),
    [anon_sym_else] = ACTIONS(1769),
    [anon_sym_new] = ACTIONS(1769),
    [anon_sym_LBRACE_LT] = ACTIONS(1771),
    [anon_sym_begin] = ACTIONS(1769),
    [sym_ocamlyacc_value] = ACTIONS(1771),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1769),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1771),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1769),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1769),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1771),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1771),
    [aux_sym_number_token1] = ACTIONS(1771),
    [aux_sym_signed_number_token1] = ACTIONS(1769),
    [anon_sym_SQUOTE] = ACTIONS(1771),
    [anon_sym_DQUOTE] = ACTIONS(1771),
    [sym_prefix_operator] = ACTIONS(1771),
    [aux_sym_sign_operator_token1] = ACTIONS(1769),
    [sym_hash_operator] = ACTIONS(1771),
    [anon_sym_lsl] = ACTIONS(1769),
    [anon_sym_lsr] = ACTIONS(1769),
    [anon_sym_asr] = ACTIONS(1769),
    [aux_sym_pow_operator_token1] = ACTIONS(1771),
    [anon_sym_mod] = ACTIONS(1769),
    [anon_sym_land] = ACTIONS(1769),
    [anon_sym_lor] = ACTIONS(1769),
    [anon_sym_lxor] = ACTIONS(1769),
    [aux_sym_mult_operator_token1] = ACTIONS(1769),
    [aux_sym_add_operator_token1] = ACTIONS(1769),
    [sym_concat_operator] = ACTIONS(1771),
    [sym_rel_operator] = ACTIONS(1769),
    [anon_sym_AMP_AMP] = ACTIONS(1769),
    [anon_sym_or] = ACTIONS(1769),
    [anon_sym_PIPE_PIPE] = ACTIONS(1769),
    [sym_assign_operator] = ACTIONS(1771),
    [sym_let_operator] = ACTIONS(1771),
    [sym_let_and_operator] = ACTIONS(1771),
    [sym__uppercase_identifier] = ACTIONS(1771),
    [aux_sym_directive_token1] = ACTIONS(1769),
    [aux_sym_tag_token1] = ACTIONS(1771),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(951)] = {
    [sym_attribute] = STATE(951),
    [sym__lowercase_identifier] = ACTIONS(1789),
    [anon_sym_SEMI_SEMI] = ACTIONS(1791),
    [anon_sym_let] = ACTIONS(1789),
    [anon_sym_and] = ACTIONS(1789),
    [anon_sym_TILDE] = ACTIONS(1789),
    [anon_sym_QMARK] = ACTIONS(1789),
    [anon_sym_LPAREN] = ACTIONS(1791),
    [anon_sym_external] = ACTIONS(1789),
    [anon_sym_type] = ACTIONS(1789),
    [anon_sym_COMMA] = ACTIONS(1791),
    [anon_sym_PIPE] = ACTIONS(1789),
    [anon_sym_DOT] = ACTIONS(1791),
    [anon_sym_LBRACK] = ACTIONS(1789),
    [anon_sym_RBRACK] = ACTIONS(1791),
    [anon_sym_true] = ACTIONS(1789),
    [anon_sym_false] = ACTIONS(1789),
    [anon_sym_LBRACE] = ACTIONS(1789),
    [anon_sym_SEMI] = ACTIONS(1789),
    [anon_sym_exception] = ACTIONS(1789),
    [anon_sym_module] = ACTIONS(1789),
    [anon_sym_open] = ACTIONS(1789),
    [anon_sym_include] = ACTIONS(1789),
    [anon_sym_class] = ACTIONS(1789),
    [anon_sym_val] = ACTIONS(1789),
    [anon_sym_end] = ACTIONS(1789),
    [anon_sym_object] = ACTIONS(1789),
    [anon_sym_in] = ACTIONS(1789),
    [anon_sym_AMP] = ACTIONS(1789),
    [anon_sym_POUND] = ACTIONS(1789),
    [anon_sym_COLON_COLON] = ACTIONS(1791),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1791),
    [anon_sym_else] = ACTIONS(1789),
    [anon_sym_new] = ACTIONS(1789),
    [anon_sym_LBRACE_LT] = ACTIONS(1791),
    [anon_sym_begin] = ACTIONS(1789),
    [sym_ocamlyacc_value] = ACTIONS(1791),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1791),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1789),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1791),
    [aux_sym_number_token1] = ACTIONS(1791),
    [aux_sym_signed_number_token1] = ACTIONS(1789),
    [anon_sym_SQUOTE] = ACTIONS(1791),
    [anon_sym_DQUOTE] = ACTIONS(1791),
    [sym_prefix_operator] = ACTIONS(1791),
    [aux_sym_sign_operator_token1] = ACTIONS(1789),
    [sym_hash_operator] = ACTIONS(1791),
    [anon_sym_lsl] = ACTIONS(1789),
    [anon_sym_lsr] = ACTIONS(1789),
    [anon_sym_asr] = ACTIONS(1789),
    [aux_sym_pow_operator_token1] = ACTIONS(1791),
    [anon_sym_mod] = ACTIONS(1789),
    [anon_sym_land] = ACTIONS(1789),
    [anon_sym_lor] = ACTIONS(1789),
    [anon_sym_lxor] = ACTIONS(1789),
    [aux_sym_mult_operator_token1] = ACTIONS(1789),
    [aux_sym_add_operator_token1] = ACTIONS(1789),
    [sym_concat_operator] = ACTIONS(1791),
    [sym_rel_operator] = ACTIONS(1789),
    [anon_sym_AMP_AMP] = ACTIONS(1789),
    [anon_sym_or] = ACTIONS(1789),
    [anon_sym_PIPE_PIPE] = ACTIONS(1789),
    [sym_assign_operator] = ACTIONS(1791),
    [sym_let_operator] = ACTIONS(1791),
    [sym_let_and_operator] = ACTIONS(1791),
    [sym__uppercase_identifier] = ACTIONS(1791),
    [aux_sym_directive_token1] = ACTIONS(1789),
    [aux_sym_tag_token1] = ACTIONS(1791),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(952)] = {
    [sym_attribute] = STATE(952),
    [sym__lowercase_identifier] = ACTIONS(1801),
    [anon_sym_SEMI_SEMI] = ACTIONS(1803),
    [anon_sym_let] = ACTIONS(1801),
    [anon_sym_and] = ACTIONS(1801),
    [anon_sym_TILDE] = ACTIONS(1801),
    [anon_sym_QMARK] = ACTIONS(1801),
    [anon_sym_LPAREN] = ACTIONS(1803),
    [anon_sym_external] = ACTIONS(1801),
    [anon_sym_type] = ACTIONS(1801),
    [anon_sym_COMMA] = ACTIONS(1803),
    [anon_sym_PIPE] = ACTIONS(1801),
    [anon_sym_DOT] = ACTIONS(1803),
    [anon_sym_LBRACK] = ACTIONS(1801),
    [anon_sym_RBRACK] = ACTIONS(1803),
    [anon_sym_true] = ACTIONS(1801),
    [anon_sym_false] = ACTIONS(1801),
    [anon_sym_LBRACE] = ACTIONS(1801),
    [anon_sym_SEMI] = ACTIONS(1801),
    [anon_sym_exception] = ACTIONS(1801),
    [anon_sym_module] = ACTIONS(1801),
    [anon_sym_open] = ACTIONS(1801),
    [anon_sym_include] = ACTIONS(1801),
    [anon_sym_class] = ACTIONS(1801),
    [anon_sym_val] = ACTIONS(1801),
    [anon_sym_end] = ACTIONS(1801),
    [anon_sym_object] = ACTIONS(1801),
    [anon_sym_in] = ACTIONS(1801),
    [anon_sym_AMP] = ACTIONS(1801),
    [anon_sym_POUND] = ACTIONS(1801),
    [anon_sym_COLON_COLON] = ACTIONS(1803),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1803),
    [anon_sym_else] = ACTIONS(1801),
    [anon_sym_new] = ACTIONS(1801),
    [anon_sym_LBRACE_LT] = ACTIONS(1803),
    [anon_sym_begin] = ACTIONS(1801),
    [sym_ocamlyacc_value] = ACTIONS(1803),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1801),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1803),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1801),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1801),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1803),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1803),
    [aux_sym_number_token1] = ACTIONS(1803),
    [aux_sym_signed_number_token1] = ACTIONS(1801),
    [anon_sym_SQUOTE] = ACTIONS(1803),
    [anon_sym_DQUOTE] = ACTIONS(1803),
    [sym_prefix_operator] = ACTIONS(1803),
    [aux_sym_sign_operator_token1] = ACTIONS(1801),
    [sym_hash_operator] = ACTIONS(1803),
    [anon_sym_lsl] = ACTIONS(1801),
    [anon_sym_lsr] = ACTIONS(1801),
    [anon_sym_asr] = ACTIONS(1801),
    [aux_sym_pow_operator_token1] = ACTIONS(1803),
    [anon_sym_mod] = ACTIONS(1801),
    [anon_sym_land] = ACTIONS(1801),
    [anon_sym_lor] = ACTIONS(1801),
    [anon_sym_lxor] = ACTIONS(1801),
    [aux_sym_mult_operator_token1] = ACTIONS(1801),
    [aux_sym_add_operator_token1] = ACTIONS(1801),
    [sym_concat_operator] = ACTIONS(1803),
    [sym_rel_operator] = ACTIONS(1801),
    [anon_sym_AMP_AMP] = ACTIONS(1801),
    [anon_sym_or] = ACTIONS(1801),
    [anon_sym_PIPE_PIPE] = ACTIONS(1801),
    [sym_assign_operator] = ACTIONS(1803),
    [sym_let_operator] = ACTIONS(1803),
    [sym_let_and_operator] = ACTIONS(1803),
    [sym__uppercase_identifier] = ACTIONS(1803),
    [aux_sym_directive_token1] = ACTIONS(1801),
    [aux_sym_tag_token1] = ACTIONS(1803),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(953)] = {
    [sym_attribute] = STATE(953),
    [sym__lowercase_identifier] = ACTIONS(1741),
    [anon_sym_SEMI_SEMI] = ACTIONS(1743),
    [anon_sym_let] = ACTIONS(1741),
    [anon_sym_and] = ACTIONS(1741),
    [anon_sym_TILDE] = ACTIONS(1741),
    [anon_sym_QMARK] = ACTIONS(1741),
    [anon_sym_LPAREN] = ACTIONS(1743),
    [anon_sym_external] = ACTIONS(1741),
    [anon_sym_type] = ACTIONS(1741),
    [anon_sym_COMMA] = ACTIONS(1743),
    [anon_sym_PIPE] = ACTIONS(1741),
    [anon_sym_DOT] = ACTIONS(1743),
    [anon_sym_LBRACK] = ACTIONS(1741),
    [anon_sym_RBRACK] = ACTIONS(1743),
    [anon_sym_true] = ACTIONS(1741),
    [anon_sym_false] = ACTIONS(1741),
    [anon_sym_LBRACE] = ACTIONS(1741),
    [anon_sym_SEMI] = ACTIONS(1741),
    [anon_sym_exception] = ACTIONS(1741),
    [anon_sym_module] = ACTIONS(1741),
    [anon_sym_open] = ACTIONS(1741),
    [anon_sym_include] = ACTIONS(1741),
    [anon_sym_class] = ACTIONS(1741),
    [anon_sym_val] = ACTIONS(1741),
    [anon_sym_end] = ACTIONS(1741),
    [anon_sym_object] = ACTIONS(1741),
    [anon_sym_in] = ACTIONS(1741),
    [anon_sym_AMP] = ACTIONS(1741),
    [anon_sym_POUND] = ACTIONS(1741),
    [anon_sym_COLON_COLON] = ACTIONS(1743),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1743),
    [anon_sym_else] = ACTIONS(1741),
    [anon_sym_new] = ACTIONS(1741),
    [anon_sym_LBRACE_LT] = ACTIONS(1743),
    [anon_sym_begin] = ACTIONS(1741),
    [sym_ocamlyacc_value] = ACTIONS(1743),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1741),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1743),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1741),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1741),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1743),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1743),
    [aux_sym_number_token1] = ACTIONS(1743),
    [aux_sym_signed_number_token1] = ACTIONS(1741),
    [anon_sym_SQUOTE] = ACTIONS(1743),
    [anon_sym_DQUOTE] = ACTIONS(1743),
    [sym_prefix_operator] = ACTIONS(1743),
    [aux_sym_sign_operator_token1] = ACTIONS(1741),
    [sym_hash_operator] = ACTIONS(1743),
    [anon_sym_lsl] = ACTIONS(1741),
    [anon_sym_lsr] = ACTIONS(1741),
    [anon_sym_asr] = ACTIONS(1741),
    [aux_sym_pow_operator_token1] = ACTIONS(1743),
    [anon_sym_mod] = ACTIONS(1741),
    [anon_sym_land] = ACTIONS(1741),
    [anon_sym_lor] = ACTIONS(1741),
    [anon_sym_lxor] = ACTIONS(1741),
    [aux_sym_mult_operator_token1] = ACTIONS(1741),
    [aux_sym_add_operator_token1] = ACTIONS(1741),
    [sym_concat_operator] = ACTIONS(1743),
    [sym_rel_operator] = ACTIONS(1741),
    [anon_sym_AMP_AMP] = ACTIONS(1741),
    [anon_sym_or] = ACTIONS(1741),
    [anon_sym_PIPE_PIPE] = ACTIONS(1741),
    [sym_assign_operator] = ACTIONS(1743),
    [sym_let_operator] = ACTIONS(1743),
    [sym_let_and_operator] = ACTIONS(1743),
    [sym__uppercase_identifier] = ACTIONS(1743),
    [aux_sym_directive_token1] = ACTIONS(1741),
    [aux_sym_tag_token1] = ACTIONS(1743),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(954)] = {
    [sym_attribute] = STATE(954),
    [sym__lowercase_identifier] = ACTIONS(1749),
    [anon_sym_SEMI_SEMI] = ACTIONS(1751),
    [anon_sym_let] = ACTIONS(1749),
    [anon_sym_and] = ACTIONS(1749),
    [anon_sym_TILDE] = ACTIONS(1749),
    [anon_sym_QMARK] = ACTIONS(1749),
    [anon_sym_LPAREN] = ACTIONS(1751),
    [anon_sym_external] = ACTIONS(1749),
    [anon_sym_type] = ACTIONS(1749),
    [anon_sym_COMMA] = ACTIONS(1751),
    [anon_sym_PIPE] = ACTIONS(1749),
    [anon_sym_DOT] = ACTIONS(1751),
    [anon_sym_LBRACK] = ACTIONS(1749),
    [anon_sym_RBRACK] = ACTIONS(1751),
    [anon_sym_true] = ACTIONS(1749),
    [anon_sym_false] = ACTIONS(1749),
    [anon_sym_LBRACE] = ACTIONS(1749),
    [anon_sym_SEMI] = ACTIONS(1749),
    [anon_sym_exception] = ACTIONS(1749),
    [anon_sym_module] = ACTIONS(1749),
    [anon_sym_open] = ACTIONS(1749),
    [anon_sym_include] = ACTIONS(1749),
    [anon_sym_class] = ACTIONS(1749),
    [anon_sym_val] = ACTIONS(1749),
    [anon_sym_end] = ACTIONS(1749),
    [anon_sym_object] = ACTIONS(1749),
    [anon_sym_in] = ACTIONS(1749),
    [anon_sym_AMP] = ACTIONS(1749),
    [anon_sym_POUND] = ACTIONS(1749),
    [anon_sym_COLON_COLON] = ACTIONS(1751),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1751),
    [anon_sym_else] = ACTIONS(1749),
    [anon_sym_new] = ACTIONS(1749),
    [anon_sym_LBRACE_LT] = ACTIONS(1751),
    [anon_sym_begin] = ACTIONS(1749),
    [sym_ocamlyacc_value] = ACTIONS(1751),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1749),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1751),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1749),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1749),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1751),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1751),
    [aux_sym_number_token1] = ACTIONS(1751),
    [aux_sym_signed_number_token1] = ACTIONS(1749),
    [anon_sym_SQUOTE] = ACTIONS(1751),
    [anon_sym_DQUOTE] = ACTIONS(1751),
    [sym_prefix_operator] = ACTIONS(1751),
    [aux_sym_sign_operator_token1] = ACTIONS(1749),
    [sym_hash_operator] = ACTIONS(1751),
    [anon_sym_lsl] = ACTIONS(1749),
    [anon_sym_lsr] = ACTIONS(1749),
    [anon_sym_asr] = ACTIONS(1749),
    [aux_sym_pow_operator_token1] = ACTIONS(1751),
    [anon_sym_mod] = ACTIONS(1749),
    [anon_sym_land] = ACTIONS(1749),
    [anon_sym_lor] = ACTIONS(1749),
    [anon_sym_lxor] = ACTIONS(1749),
    [aux_sym_mult_operator_token1] = ACTIONS(1749),
    [aux_sym_add_operator_token1] = ACTIONS(1749),
    [sym_concat_operator] = ACTIONS(1751),
    [sym_rel_operator] = ACTIONS(1749),
    [anon_sym_AMP_AMP] = ACTIONS(1749),
    [anon_sym_or] = ACTIONS(1749),
    [anon_sym_PIPE_PIPE] = ACTIONS(1749),
    [sym_assign_operator] = ACTIONS(1751),
    [sym_let_operator] = ACTIONS(1751),
    [sym_let_and_operator] = ACTIONS(1751),
    [sym__uppercase_identifier] = ACTIONS(1751),
    [aux_sym_directive_token1] = ACTIONS(1749),
    [aux_sym_tag_token1] = ACTIONS(1751),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(955)] = {
    [sym_attribute] = STATE(955),
    [sym__lowercase_identifier] = ACTIONS(1753),
    [anon_sym_SEMI_SEMI] = ACTIONS(1755),
    [anon_sym_let] = ACTIONS(1753),
    [anon_sym_and] = ACTIONS(1753),
    [anon_sym_TILDE] = ACTIONS(1753),
    [anon_sym_QMARK] = ACTIONS(1753),
    [anon_sym_LPAREN] = ACTIONS(1755),
    [anon_sym_external] = ACTIONS(1753),
    [anon_sym_type] = ACTIONS(1753),
    [anon_sym_COMMA] = ACTIONS(1755),
    [anon_sym_PIPE] = ACTIONS(1753),
    [anon_sym_DOT] = ACTIONS(1755),
    [anon_sym_LBRACK] = ACTIONS(1753),
    [anon_sym_RBRACK] = ACTIONS(1755),
    [anon_sym_true] = ACTIONS(1753),
    [anon_sym_false] = ACTIONS(1753),
    [anon_sym_LBRACE] = ACTIONS(1753),
    [anon_sym_SEMI] = ACTIONS(1753),
    [anon_sym_exception] = ACTIONS(1753),
    [anon_sym_module] = ACTIONS(1753),
    [anon_sym_open] = ACTIONS(1753),
    [anon_sym_include] = ACTIONS(1753),
    [anon_sym_class] = ACTIONS(1753),
    [anon_sym_val] = ACTIONS(1753),
    [anon_sym_end] = ACTIONS(1753),
    [anon_sym_object] = ACTIONS(1753),
    [anon_sym_in] = ACTIONS(1753),
    [anon_sym_AMP] = ACTIONS(1753),
    [anon_sym_POUND] = ACTIONS(1753),
    [anon_sym_COLON_COLON] = ACTIONS(1755),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1755),
    [anon_sym_else] = ACTIONS(1753),
    [anon_sym_new] = ACTIONS(1753),
    [anon_sym_LBRACE_LT] = ACTIONS(1755),
    [anon_sym_begin] = ACTIONS(1753),
    [sym_ocamlyacc_value] = ACTIONS(1755),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1753),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1755),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1753),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1753),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1755),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1755),
    [aux_sym_number_token1] = ACTIONS(1755),
    [aux_sym_signed_number_token1] = ACTIONS(1753),
    [anon_sym_SQUOTE] = ACTIONS(1755),
    [anon_sym_DQUOTE] = ACTIONS(1755),
    [sym_prefix_operator] = ACTIONS(1755),
    [aux_sym_sign_operator_token1] = ACTIONS(1753),
    [sym_hash_operator] = ACTIONS(1755),
    [anon_sym_lsl] = ACTIONS(1753),
    [anon_sym_lsr] = ACTIONS(1753),
    [anon_sym_asr] = ACTIONS(1753),
    [aux_sym_pow_operator_token1] = ACTIONS(1755),
    [anon_sym_mod] = ACTIONS(1753),
    [anon_sym_land] = ACTIONS(1753),
    [anon_sym_lor] = ACTIONS(1753),
    [anon_sym_lxor] = ACTIONS(1753),
    [aux_sym_mult_operator_token1] = ACTIONS(1753),
    [aux_sym_add_operator_token1] = ACTIONS(1753),
    [sym_concat_operator] = ACTIONS(1755),
    [sym_rel_operator] = ACTIONS(1753),
    [anon_sym_AMP_AMP] = ACTIONS(1753),
    [anon_sym_or] = ACTIONS(1753),
    [anon_sym_PIPE_PIPE] = ACTIONS(1753),
    [sym_assign_operator] = ACTIONS(1755),
    [sym_let_operator] = ACTIONS(1755),
    [sym_let_and_operator] = ACTIONS(1755),
    [sym__uppercase_identifier] = ACTIONS(1755),
    [aux_sym_directive_token1] = ACTIONS(1753),
    [aux_sym_tag_token1] = ACTIONS(1755),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(956)] = {
    [sym_attribute] = STATE(956),
    [sym__lowercase_identifier] = ACTIONS(1761),
    [anon_sym_SEMI_SEMI] = ACTIONS(1763),
    [anon_sym_let] = ACTIONS(1761),
    [anon_sym_and] = ACTIONS(1761),
    [anon_sym_TILDE] = ACTIONS(1761),
    [anon_sym_QMARK] = ACTIONS(1761),
    [anon_sym_LPAREN] = ACTIONS(1763),
    [anon_sym_external] = ACTIONS(1761),
    [anon_sym_type] = ACTIONS(1761),
    [anon_sym_COMMA] = ACTIONS(1763),
    [anon_sym_PIPE] = ACTIONS(1761),
    [anon_sym_DOT] = ACTIONS(1763),
    [anon_sym_LBRACK] = ACTIONS(1761),
    [anon_sym_RBRACK] = ACTIONS(1763),
    [anon_sym_true] = ACTIONS(1761),
    [anon_sym_false] = ACTIONS(1761),
    [anon_sym_LBRACE] = ACTIONS(1761),
    [anon_sym_SEMI] = ACTIONS(1761),
    [anon_sym_exception] = ACTIONS(1761),
    [anon_sym_module] = ACTIONS(1761),
    [anon_sym_open] = ACTIONS(1761),
    [anon_sym_include] = ACTIONS(1761),
    [anon_sym_class] = ACTIONS(1761),
    [anon_sym_val] = ACTIONS(1761),
    [anon_sym_end] = ACTIONS(1761),
    [anon_sym_object] = ACTIONS(1761),
    [anon_sym_in] = ACTIONS(1761),
    [anon_sym_AMP] = ACTIONS(1761),
    [anon_sym_POUND] = ACTIONS(1761),
    [anon_sym_COLON_COLON] = ACTIONS(1763),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1763),
    [anon_sym_else] = ACTIONS(1761),
    [anon_sym_new] = ACTIONS(1761),
    [anon_sym_LBRACE_LT] = ACTIONS(1763),
    [anon_sym_begin] = ACTIONS(1761),
    [sym_ocamlyacc_value] = ACTIONS(1763),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1761),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1763),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1761),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1761),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1763),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1763),
    [aux_sym_number_token1] = ACTIONS(1763),
    [aux_sym_signed_number_token1] = ACTIONS(1761),
    [anon_sym_SQUOTE] = ACTIONS(1763),
    [anon_sym_DQUOTE] = ACTIONS(1763),
    [sym_prefix_operator] = ACTIONS(1763),
    [aux_sym_sign_operator_token1] = ACTIONS(1761),
    [sym_hash_operator] = ACTIONS(1763),
    [anon_sym_lsl] = ACTIONS(1761),
    [anon_sym_lsr] = ACTIONS(1761),
    [anon_sym_asr] = ACTIONS(1761),
    [aux_sym_pow_operator_token1] = ACTIONS(1763),
    [anon_sym_mod] = ACTIONS(1761),
    [anon_sym_land] = ACTIONS(1761),
    [anon_sym_lor] = ACTIONS(1761),
    [anon_sym_lxor] = ACTIONS(1761),
    [aux_sym_mult_operator_token1] = ACTIONS(1761),
    [aux_sym_add_operator_token1] = ACTIONS(1761),
    [sym_concat_operator] = ACTIONS(1763),
    [sym_rel_operator] = ACTIONS(1761),
    [anon_sym_AMP_AMP] = ACTIONS(1761),
    [anon_sym_or] = ACTIONS(1761),
    [anon_sym_PIPE_PIPE] = ACTIONS(1761),
    [sym_assign_operator] = ACTIONS(1763),
    [sym_let_operator] = ACTIONS(1763),
    [sym_let_and_operator] = ACTIONS(1763),
    [sym__uppercase_identifier] = ACTIONS(1763),
    [aux_sym_directive_token1] = ACTIONS(1761),
    [aux_sym_tag_token1] = ACTIONS(1763),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(957)] = {
    [sym_attribute] = STATE(957),
    [sym__lowercase_identifier] = ACTIONS(1699),
    [anon_sym_SEMI_SEMI] = ACTIONS(1701),
    [anon_sym_let] = ACTIONS(1699),
    [anon_sym_and] = ACTIONS(1699),
    [anon_sym_TILDE] = ACTIONS(1699),
    [anon_sym_QMARK] = ACTIONS(1699),
    [anon_sym_LPAREN] = ACTIONS(1701),
    [anon_sym_external] = ACTIONS(1699),
    [anon_sym_type] = ACTIONS(1699),
    [anon_sym_COMMA] = ACTIONS(1701),
    [anon_sym_PIPE] = ACTIONS(1699),
    [anon_sym_DOT] = ACTIONS(1703),
    [anon_sym_LBRACK] = ACTIONS(1699),
    [anon_sym_RBRACK] = ACTIONS(1701),
    [anon_sym_true] = ACTIONS(1699),
    [anon_sym_false] = ACTIONS(1699),
    [anon_sym_LBRACE] = ACTIONS(1699),
    [anon_sym_SEMI] = ACTIONS(1699),
    [anon_sym_exception] = ACTIONS(1699),
    [anon_sym_module] = ACTIONS(1699),
    [anon_sym_open] = ACTIONS(1699),
    [anon_sym_include] = ACTIONS(1699),
    [anon_sym_class] = ACTIONS(1699),
    [anon_sym_val] = ACTIONS(1699),
    [anon_sym_end] = ACTIONS(1699),
    [anon_sym_object] = ACTIONS(1699),
    [anon_sym_in] = ACTIONS(1699),
    [anon_sym_AMP] = ACTIONS(1699),
    [anon_sym_POUND] = ACTIONS(1699),
    [anon_sym_COLON_COLON] = ACTIONS(1701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1701),
    [anon_sym_else] = ACTIONS(1699),
    [anon_sym_new] = ACTIONS(1699),
    [anon_sym_LBRACE_LT] = ACTIONS(1701),
    [anon_sym_begin] = ACTIONS(1699),
    [sym_ocamlyacc_value] = ACTIONS(1701),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1699),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1701),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1699),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1699),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1701),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1701),
    [aux_sym_number_token1] = ACTIONS(1701),
    [aux_sym_signed_number_token1] = ACTIONS(1699),
    [anon_sym_SQUOTE] = ACTIONS(1701),
    [anon_sym_DQUOTE] = ACTIONS(1701),
    [sym_prefix_operator] = ACTIONS(1701),
    [aux_sym_sign_operator_token1] = ACTIONS(1699),
    [sym_hash_operator] = ACTIONS(1701),
    [anon_sym_lsl] = ACTIONS(1699),
    [anon_sym_lsr] = ACTIONS(1699),
    [anon_sym_asr] = ACTIONS(1699),
    [aux_sym_pow_operator_token1] = ACTIONS(1701),
    [anon_sym_mod] = ACTIONS(1699),
    [anon_sym_land] = ACTIONS(1699),
    [anon_sym_lor] = ACTIONS(1699),
    [anon_sym_lxor] = ACTIONS(1699),
    [aux_sym_mult_operator_token1] = ACTIONS(1699),
    [aux_sym_add_operator_token1] = ACTIONS(1699),
    [sym_concat_operator] = ACTIONS(1701),
    [sym_rel_operator] = ACTIONS(1699),
    [anon_sym_AMP_AMP] = ACTIONS(1699),
    [anon_sym_or] = ACTIONS(1699),
    [anon_sym_PIPE_PIPE] = ACTIONS(1699),
    [sym_assign_operator] = ACTIONS(1701),
    [sym_let_operator] = ACTIONS(1701),
    [sym_let_and_operator] = ACTIONS(1701),
    [sym__uppercase_identifier] = ACTIONS(1701),
    [aux_sym_directive_token1] = ACTIONS(1699),
    [aux_sym_tag_token1] = ACTIONS(1701),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(958)] = {
    [sym_attribute] = STATE(958),
    [sym__lowercase_identifier] = ACTIONS(1793),
    [anon_sym_SEMI_SEMI] = ACTIONS(1795),
    [anon_sym_let] = ACTIONS(1793),
    [anon_sym_and] = ACTIONS(1793),
    [anon_sym_TILDE] = ACTIONS(1793),
    [anon_sym_QMARK] = ACTIONS(1793),
    [anon_sym_LPAREN] = ACTIONS(1795),
    [anon_sym_external] = ACTIONS(1793),
    [anon_sym_type] = ACTIONS(1793),
    [anon_sym_COMMA] = ACTIONS(1795),
    [anon_sym_PIPE] = ACTIONS(1793),
    [anon_sym_DOT] = ACTIONS(1795),
    [anon_sym_LBRACK] = ACTIONS(1793),
    [anon_sym_RBRACK] = ACTIONS(1795),
    [anon_sym_true] = ACTIONS(1793),
    [anon_sym_false] = ACTIONS(1793),
    [anon_sym_LBRACE] = ACTIONS(1793),
    [anon_sym_SEMI] = ACTIONS(1793),
    [anon_sym_exception] = ACTIONS(1793),
    [anon_sym_module] = ACTIONS(1793),
    [anon_sym_open] = ACTIONS(1793),
    [anon_sym_include] = ACTIONS(1793),
    [anon_sym_class] = ACTIONS(1793),
    [anon_sym_val] = ACTIONS(1793),
    [anon_sym_end] = ACTIONS(1793),
    [anon_sym_object] = ACTIONS(1793),
    [anon_sym_in] = ACTIONS(1793),
    [anon_sym_AMP] = ACTIONS(1793),
    [anon_sym_POUND] = ACTIONS(1793),
    [anon_sym_COLON_COLON] = ACTIONS(1795),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1795),
    [anon_sym_else] = ACTIONS(1793),
    [anon_sym_new] = ACTIONS(1793),
    [anon_sym_LBRACE_LT] = ACTIONS(1795),
    [anon_sym_begin] = ACTIONS(1793),
    [sym_ocamlyacc_value] = ACTIONS(1795),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1793),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1795),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1795),
    [aux_sym_number_token1] = ACTIONS(1795),
    [aux_sym_signed_number_token1] = ACTIONS(1793),
    [anon_sym_SQUOTE] = ACTIONS(1795),
    [anon_sym_DQUOTE] = ACTIONS(1795),
    [sym_prefix_operator] = ACTIONS(1795),
    [aux_sym_sign_operator_token1] = ACTIONS(1793),
    [sym_hash_operator] = ACTIONS(1795),
    [anon_sym_lsl] = ACTIONS(1793),
    [anon_sym_lsr] = ACTIONS(1793),
    [anon_sym_asr] = ACTIONS(1793),
    [aux_sym_pow_operator_token1] = ACTIONS(1795),
    [anon_sym_mod] = ACTIONS(1793),
    [anon_sym_land] = ACTIONS(1793),
    [anon_sym_lor] = ACTIONS(1793),
    [anon_sym_lxor] = ACTIONS(1793),
    [aux_sym_mult_operator_token1] = ACTIONS(1793),
    [aux_sym_add_operator_token1] = ACTIONS(1793),
    [sym_concat_operator] = ACTIONS(1795),
    [sym_rel_operator] = ACTIONS(1793),
    [anon_sym_AMP_AMP] = ACTIONS(1793),
    [anon_sym_or] = ACTIONS(1793),
    [anon_sym_PIPE_PIPE] = ACTIONS(1793),
    [sym_assign_operator] = ACTIONS(1795),
    [sym_let_operator] = ACTIONS(1795),
    [sym_let_and_operator] = ACTIONS(1795),
    [sym__uppercase_identifier] = ACTIONS(1795),
    [aux_sym_directive_token1] = ACTIONS(1793),
    [aux_sym_tag_token1] = ACTIONS(1795),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(959)] = {
    [sym_attribute] = STATE(959),
    [sym__lowercase_identifier] = ACTIONS(1659),
    [anon_sym_SEMI_SEMI] = ACTIONS(1661),
    [anon_sym_let] = ACTIONS(1659),
    [anon_sym_and] = ACTIONS(1659),
    [anon_sym_TILDE] = ACTIONS(1659),
    [anon_sym_QMARK] = ACTIONS(1659),
    [anon_sym_LPAREN] = ACTIONS(1661),
    [anon_sym_external] = ACTIONS(1659),
    [anon_sym_type] = ACTIONS(1659),
    [anon_sym_COMMA] = ACTIONS(1661),
    [anon_sym_PIPE] = ACTIONS(1659),
    [anon_sym_DOT] = ACTIONS(1661),
    [anon_sym_LBRACK] = ACTIONS(1659),
    [anon_sym_RBRACK] = ACTIONS(1661),
    [anon_sym_true] = ACTIONS(1659),
    [anon_sym_false] = ACTIONS(1659),
    [anon_sym_LBRACE] = ACTIONS(1659),
    [anon_sym_SEMI] = ACTIONS(1659),
    [anon_sym_exception] = ACTIONS(1659),
    [anon_sym_module] = ACTIONS(1659),
    [anon_sym_open] = ACTIONS(1659),
    [anon_sym_include] = ACTIONS(1659),
    [anon_sym_class] = ACTIONS(1659),
    [anon_sym_val] = ACTIONS(1659),
    [anon_sym_end] = ACTIONS(1659),
    [anon_sym_object] = ACTIONS(1659),
    [anon_sym_in] = ACTIONS(1659),
    [anon_sym_AMP] = ACTIONS(1659),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_COLON_COLON] = ACTIONS(1661),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_else] = ACTIONS(1659),
    [anon_sym_new] = ACTIONS(1659),
    [anon_sym_LBRACE_LT] = ACTIONS(1661),
    [anon_sym_begin] = ACTIONS(1659),
    [sym_ocamlyacc_value] = ACTIONS(1661),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1659),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1661),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1659),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1659),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1661),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1661),
    [aux_sym_number_token1] = ACTIONS(1661),
    [aux_sym_signed_number_token1] = ACTIONS(1659),
    [anon_sym_SQUOTE] = ACTIONS(1661),
    [anon_sym_DQUOTE] = ACTIONS(1661),
    [sym_prefix_operator] = ACTIONS(1661),
    [aux_sym_sign_operator_token1] = ACTIONS(1659),
    [sym_hash_operator] = ACTIONS(1661),
    [anon_sym_lsl] = ACTIONS(1659),
    [anon_sym_lsr] = ACTIONS(1659),
    [anon_sym_asr] = ACTIONS(1659),
    [aux_sym_pow_operator_token1] = ACTIONS(1661),
    [anon_sym_mod] = ACTIONS(1659),
    [anon_sym_land] = ACTIONS(1659),
    [anon_sym_lor] = ACTIONS(1659),
    [anon_sym_lxor] = ACTIONS(1659),
    [aux_sym_mult_operator_token1] = ACTIONS(1659),
    [aux_sym_add_operator_token1] = ACTIONS(1659),
    [sym_concat_operator] = ACTIONS(1661),
    [sym_rel_operator] = ACTIONS(1659),
    [anon_sym_AMP_AMP] = ACTIONS(1659),
    [anon_sym_or] = ACTIONS(1659),
    [anon_sym_PIPE_PIPE] = ACTIONS(1659),
    [sym_assign_operator] = ACTIONS(1661),
    [sym_let_operator] = ACTIONS(1661),
    [sym_let_and_operator] = ACTIONS(1661),
    [sym__uppercase_identifier] = ACTIONS(1661),
    [aux_sym_directive_token1] = ACTIONS(1659),
    [aux_sym_tag_token1] = ACTIONS(1661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(960)] = {
    [sym_attribute] = STATE(960),
    [sym__lowercase_identifier] = ACTIONS(1765),
    [anon_sym_SEMI_SEMI] = ACTIONS(1767),
    [anon_sym_let] = ACTIONS(1765),
    [anon_sym_and] = ACTIONS(1765),
    [anon_sym_TILDE] = ACTIONS(1765),
    [anon_sym_QMARK] = ACTIONS(1765),
    [anon_sym_LPAREN] = ACTIONS(1767),
    [anon_sym_external] = ACTIONS(1765),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_COMMA] = ACTIONS(1767),
    [anon_sym_PIPE] = ACTIONS(1765),
    [anon_sym_DOT] = ACTIONS(1767),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_RBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1765),
    [anon_sym_false] = ACTIONS(1765),
    [anon_sym_LBRACE] = ACTIONS(1765),
    [anon_sym_SEMI] = ACTIONS(1765),
    [anon_sym_exception] = ACTIONS(1765),
    [anon_sym_module] = ACTIONS(1765),
    [anon_sym_open] = ACTIONS(1765),
    [anon_sym_include] = ACTIONS(1765),
    [anon_sym_class] = ACTIONS(1765),
    [anon_sym_val] = ACTIONS(1765),
    [anon_sym_end] = ACTIONS(1765),
    [anon_sym_object] = ACTIONS(1765),
    [anon_sym_in] = ACTIONS(1765),
    [anon_sym_AMP] = ACTIONS(1765),
    [anon_sym_POUND] = ACTIONS(1765),
    [anon_sym_COLON_COLON] = ACTIONS(1767),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_else] = ACTIONS(1765),
    [anon_sym_new] = ACTIONS(1765),
    [anon_sym_LBRACE_LT] = ACTIONS(1767),
    [anon_sym_begin] = ACTIONS(1765),
    [sym_ocamlyacc_value] = ACTIONS(1767),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1765),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1767),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1765),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1765),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1767),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1767),
    [aux_sym_number_token1] = ACTIONS(1767),
    [aux_sym_signed_number_token1] = ACTIONS(1765),
    [anon_sym_SQUOTE] = ACTIONS(1767),
    [anon_sym_DQUOTE] = ACTIONS(1767),
    [sym_prefix_operator] = ACTIONS(1767),
    [aux_sym_sign_operator_token1] = ACTIONS(1765),
    [sym_hash_operator] = ACTIONS(1767),
    [anon_sym_lsl] = ACTIONS(1765),
    [anon_sym_lsr] = ACTIONS(1765),
    [anon_sym_asr] = ACTIONS(1765),
    [aux_sym_pow_operator_token1] = ACTIONS(1767),
    [anon_sym_mod] = ACTIONS(1765),
    [anon_sym_land] = ACTIONS(1765),
    [anon_sym_lor] = ACTIONS(1765),
    [anon_sym_lxor] = ACTIONS(1765),
    [aux_sym_mult_operator_token1] = ACTIONS(1765),
    [aux_sym_add_operator_token1] = ACTIONS(1765),
    [sym_concat_operator] = ACTIONS(1767),
    [sym_rel_operator] = ACTIONS(1765),
    [anon_sym_AMP_AMP] = ACTIONS(1765),
    [anon_sym_or] = ACTIONS(1765),
    [anon_sym_PIPE_PIPE] = ACTIONS(1765),
    [sym_assign_operator] = ACTIONS(1767),
    [sym_let_operator] = ACTIONS(1767),
    [sym_let_and_operator] = ACTIONS(1767),
    [sym__uppercase_identifier] = ACTIONS(1767),
    [aux_sym_directive_token1] = ACTIONS(1765),
    [aux_sym_tag_token1] = ACTIONS(1767),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(961)] = {
    [sym_parameter] = STATE(2207),
    [sym__class_typed] = STATE(3862),
    [sym__parenthesized_abstract_type] = STATE(2207),
    [sym__delimited_pattern] = STATE(2286),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2286),
    [sym_polymorphic_variant_pattern] = STATE(2286),
    [sym_record_pattern] = STATE(2287),
    [sym_list_pattern] = STATE(2287),
    [sym_array_pattern] = STATE(2287),
    [sym_local_open_pattern] = STATE(2286),
    [sym_package_pattern] = STATE(2286),
    [sym_parenthesized_pattern] = STATE(2287),
    [sym_attribute] = STATE(961),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2286),
    [sym_quoted_extension] = STATE(2286),
    [sym__constant] = STATE(2212),
    [sym__signed_constant] = STATE(2286),
    [sym_number] = STATE(2280),
    [sym_signed_number] = STATE(2212),
    [sym_character] = STATE(2280),
    [sym_string] = STATE(2280),
    [sym_quoted_string] = STATE(2280),
    [sym_boolean] = STATE(2280),
    [sym_unit] = STATE(2287),
    [sym__simple_value_pattern] = STATE(2286),
    [sym_parenthesized_operator] = STATE(2286),
    [sym_module_path] = STATE(6556),
    [sym_constructor_path] = STATE(2286),
    [sym__constructor_name] = STATE(2597),
    [sym_tag] = STATE(2286),
    [aux_sym_expression_item_repeat1] = STATE(3735),
    [aux_sym_let_binding_repeat1] = STATE(968),
    [sym__lowercase_identifier] = ACTIONS(1467),
    [anon_sym_SEMI_SEMI] = ACTIONS(2045),
    [anon_sym_and] = ACTIONS(2047),
    [anon_sym_EQ] = ACTIONS(2089),
    [anon_sym_TILDE] = ACTIONS(2091),
    [anon_sym_QMARK] = ACTIONS(2091),
    [anon_sym_LPAREN] = ACTIONS(2093),
    [anon_sym_external] = ACTIONS(2047),
    [anon_sym_type] = ACTIONS(2047),
    [anon_sym_COLON2] = ACTIONS(2095),
    [anon_sym_LBRACK] = ACTIONS(1485),
    [anon_sym_RBRACK] = ACTIONS(2045),
    [anon_sym_true] = ACTIONS(1487),
    [anon_sym_false] = ACTIONS(1487),
    [anon_sym_LBRACE] = ACTIONS(1489),
    [anon_sym_exception] = ACTIONS(2047),
    [anon_sym_module] = ACTIONS(2047),
    [anon_sym_open] = ACTIONS(2047),
    [anon_sym_include] = ACTIONS(2047),
    [anon_sym_class] = ACTIONS(2047),
    [anon_sym_val] = ACTIONS(2047),
    [anon_sym_end] = ACTIONS(2047),
    [anon_sym_POUND] = ACTIONS(1495),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1499),
    [anon_sym_begin] = ACTIONS(1501),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2045),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2045),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2045),
    [aux_sym_number_token1] = ACTIONS(1509),
    [aux_sym_signed_number_token1] = ACTIONS(1511),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE] = ACTIONS(1515),
    [sym__uppercase_identifier] = ACTIONS(2097),
    [aux_sym_tag_token1] = ACTIONS(2099),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(962)] = {
    [sym_attribute] = STATE(962),
    [sym__lowercase_identifier] = ACTIONS(1773),
    [anon_sym_SEMI_SEMI] = ACTIONS(1775),
    [anon_sym_let] = ACTIONS(1773),
    [anon_sym_and] = ACTIONS(1773),
    [anon_sym_TILDE] = ACTIONS(1773),
    [anon_sym_QMARK] = ACTIONS(1773),
    [anon_sym_LPAREN] = ACTIONS(1775),
    [anon_sym_external] = ACTIONS(1773),
    [anon_sym_type] = ACTIONS(1773),
    [anon_sym_COMMA] = ACTIONS(1775),
    [anon_sym_PIPE] = ACTIONS(1773),
    [anon_sym_DOT] = ACTIONS(1775),
    [anon_sym_LBRACK] = ACTIONS(1773),
    [anon_sym_RBRACK] = ACTIONS(1775),
    [anon_sym_true] = ACTIONS(1773),
    [anon_sym_false] = ACTIONS(1773),
    [anon_sym_LBRACE] = ACTIONS(1773),
    [anon_sym_SEMI] = ACTIONS(1773),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1773),
    [anon_sym_open] = ACTIONS(1773),
    [anon_sym_include] = ACTIONS(1773),
    [anon_sym_class] = ACTIONS(1773),
    [anon_sym_val] = ACTIONS(1773),
    [anon_sym_end] = ACTIONS(1773),
    [anon_sym_object] = ACTIONS(1773),
    [anon_sym_in] = ACTIONS(1773),
    [anon_sym_AMP] = ACTIONS(1773),
    [anon_sym_POUND] = ACTIONS(1773),
    [anon_sym_COLON_COLON] = ACTIONS(1775),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1775),
    [anon_sym_else] = ACTIONS(1773),
    [anon_sym_new] = ACTIONS(1773),
    [anon_sym_LBRACE_LT] = ACTIONS(1775),
    [anon_sym_begin] = ACTIONS(1773),
    [sym_ocamlyacc_value] = ACTIONS(1775),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1773),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1775),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1773),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1773),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1775),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1775),
    [aux_sym_number_token1] = ACTIONS(1775),
    [aux_sym_signed_number_token1] = ACTIONS(1773),
    [anon_sym_SQUOTE] = ACTIONS(1775),
    [anon_sym_DQUOTE] = ACTIONS(1775),
    [sym_prefix_operator] = ACTIONS(1775),
    [aux_sym_sign_operator_token1] = ACTIONS(1773),
    [sym_hash_operator] = ACTIONS(1775),
    [anon_sym_lsl] = ACTIONS(1773),
    [anon_sym_lsr] = ACTIONS(1773),
    [anon_sym_asr] = ACTIONS(1773),
    [aux_sym_pow_operator_token1] = ACTIONS(1775),
    [anon_sym_mod] = ACTIONS(1773),
    [anon_sym_land] = ACTIONS(1773),
    [anon_sym_lor] = ACTIONS(1773),
    [anon_sym_lxor] = ACTIONS(1773),
    [aux_sym_mult_operator_token1] = ACTIONS(1773),
    [aux_sym_add_operator_token1] = ACTIONS(1773),
    [sym_concat_operator] = ACTIONS(1775),
    [sym_rel_operator] = ACTIONS(1773),
    [anon_sym_AMP_AMP] = ACTIONS(1773),
    [anon_sym_or] = ACTIONS(1773),
    [anon_sym_PIPE_PIPE] = ACTIONS(1773),
    [sym_assign_operator] = ACTIONS(1775),
    [sym_let_operator] = ACTIONS(1775),
    [sym_let_and_operator] = ACTIONS(1775),
    [sym__uppercase_identifier] = ACTIONS(1775),
    [aux_sym_directive_token1] = ACTIONS(1773),
    [aux_sym_tag_token1] = ACTIONS(1775),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(963)] = {
    [sym_attribute] = STATE(963),
    [sym__lowercase_identifier] = ACTIONS(1691),
    [anon_sym_SEMI_SEMI] = ACTIONS(1693),
    [anon_sym_let] = ACTIONS(1691),
    [anon_sym_and] = ACTIONS(1691),
    [anon_sym_TILDE] = ACTIONS(1691),
    [anon_sym_QMARK] = ACTIONS(1691),
    [anon_sym_LPAREN] = ACTIONS(1693),
    [anon_sym_external] = ACTIONS(1691),
    [anon_sym_type] = ACTIONS(1691),
    [anon_sym_COMMA] = ACTIONS(1693),
    [anon_sym_PIPE] = ACTIONS(1691),
    [anon_sym_DOT] = ACTIONS(1693),
    [anon_sym_LBRACK] = ACTIONS(1691),
    [anon_sym_RBRACK] = ACTIONS(1693),
    [anon_sym_true] = ACTIONS(1691),
    [anon_sym_false] = ACTIONS(1691),
    [anon_sym_LBRACE] = ACTIONS(1691),
    [anon_sym_SEMI] = ACTIONS(1691),
    [anon_sym_exception] = ACTIONS(1691),
    [anon_sym_module] = ACTIONS(1691),
    [anon_sym_open] = ACTIONS(1691),
    [anon_sym_include] = ACTIONS(1691),
    [anon_sym_class] = ACTIONS(1691),
    [anon_sym_val] = ACTIONS(1691),
    [anon_sym_end] = ACTIONS(1691),
    [anon_sym_object] = ACTIONS(1691),
    [anon_sym_in] = ACTIONS(1691),
    [anon_sym_AMP] = ACTIONS(1691),
    [anon_sym_POUND] = ACTIONS(1691),
    [anon_sym_COLON_COLON] = ACTIONS(1693),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1693),
    [anon_sym_else] = ACTIONS(1691),
    [anon_sym_new] = ACTIONS(1691),
    [anon_sym_LBRACE_LT] = ACTIONS(1693),
    [anon_sym_begin] = ACTIONS(1691),
    [sym_ocamlyacc_value] = ACTIONS(1693),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1691),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1693),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1691),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1691),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1693),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1693),
    [aux_sym_number_token1] = ACTIONS(1693),
    [aux_sym_signed_number_token1] = ACTIONS(1691),
    [anon_sym_SQUOTE] = ACTIONS(1693),
    [anon_sym_DQUOTE] = ACTIONS(1693),
    [sym_prefix_operator] = ACTIONS(1693),
    [aux_sym_sign_operator_token1] = ACTIONS(1691),
    [sym_hash_operator] = ACTIONS(1693),
    [anon_sym_lsl] = ACTIONS(1691),
    [anon_sym_lsr] = ACTIONS(1691),
    [anon_sym_asr] = ACTIONS(1691),
    [aux_sym_pow_operator_token1] = ACTIONS(1693),
    [anon_sym_mod] = ACTIONS(1691),
    [anon_sym_land] = ACTIONS(1691),
    [anon_sym_lor] = ACTIONS(1691),
    [anon_sym_lxor] = ACTIONS(1691),
    [aux_sym_mult_operator_token1] = ACTIONS(1691),
    [aux_sym_add_operator_token1] = ACTIONS(1691),
    [sym_concat_operator] = ACTIONS(1693),
    [sym_rel_operator] = ACTIONS(1691),
    [anon_sym_AMP_AMP] = ACTIONS(1691),
    [anon_sym_or] = ACTIONS(1691),
    [anon_sym_PIPE_PIPE] = ACTIONS(1691),
    [sym_assign_operator] = ACTIONS(1693),
    [sym_let_operator] = ACTIONS(1693),
    [sym_let_and_operator] = ACTIONS(1693),
    [sym__uppercase_identifier] = ACTIONS(1693),
    [aux_sym_directive_token1] = ACTIONS(1691),
    [aux_sym_tag_token1] = ACTIONS(1693),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(964)] = {
    [sym_attribute] = STATE(964),
    [sym__lowercase_identifier] = ACTIONS(1591),
    [anon_sym_SEMI_SEMI] = ACTIONS(1593),
    [anon_sym_let] = ACTIONS(1591),
    [anon_sym_and] = ACTIONS(1591),
    [anon_sym_TILDE] = ACTIONS(1591),
    [anon_sym_QMARK] = ACTIONS(1591),
    [anon_sym_LPAREN] = ACTIONS(1593),
    [anon_sym_external] = ACTIONS(1591),
    [anon_sym_type] = ACTIONS(1591),
    [anon_sym_COMMA] = ACTIONS(1593),
    [anon_sym_PIPE] = ACTIONS(1591),
    [anon_sym_DOT] = ACTIONS(1593),
    [anon_sym_LBRACK] = ACTIONS(1591),
    [anon_sym_RBRACK] = ACTIONS(1593),
    [anon_sym_true] = ACTIONS(1591),
    [anon_sym_false] = ACTIONS(1591),
    [anon_sym_LBRACE] = ACTIONS(1591),
    [anon_sym_SEMI] = ACTIONS(1591),
    [anon_sym_exception] = ACTIONS(1591),
    [anon_sym_module] = ACTIONS(1591),
    [anon_sym_open] = ACTIONS(1591),
    [anon_sym_include] = ACTIONS(1591),
    [anon_sym_class] = ACTIONS(1591),
    [anon_sym_val] = ACTIONS(1591),
    [anon_sym_end] = ACTIONS(1591),
    [anon_sym_object] = ACTIONS(1591),
    [anon_sym_in] = ACTIONS(1591),
    [anon_sym_AMP] = ACTIONS(1591),
    [anon_sym_POUND] = ACTIONS(1591),
    [anon_sym_COLON_COLON] = ACTIONS(1593),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1593),
    [anon_sym_else] = ACTIONS(1591),
    [anon_sym_new] = ACTIONS(1591),
    [anon_sym_LBRACE_LT] = ACTIONS(1593),
    [anon_sym_begin] = ACTIONS(1591),
    [sym_ocamlyacc_value] = ACTIONS(1593),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1591),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1593),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1591),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1591),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1593),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1593),
    [aux_sym_number_token1] = ACTIONS(1593),
    [aux_sym_signed_number_token1] = ACTIONS(1591),
    [anon_sym_SQUOTE] = ACTIONS(1593),
    [anon_sym_DQUOTE] = ACTIONS(1593),
    [sym_prefix_operator] = ACTIONS(1593),
    [aux_sym_sign_operator_token1] = ACTIONS(1591),
    [sym_hash_operator] = ACTIONS(1593),
    [anon_sym_lsl] = ACTIONS(1591),
    [anon_sym_lsr] = ACTIONS(1591),
    [anon_sym_asr] = ACTIONS(1591),
    [aux_sym_pow_operator_token1] = ACTIONS(1593),
    [anon_sym_mod] = ACTIONS(1591),
    [anon_sym_land] = ACTIONS(1591),
    [anon_sym_lor] = ACTIONS(1591),
    [anon_sym_lxor] = ACTIONS(1591),
    [aux_sym_mult_operator_token1] = ACTIONS(1591),
    [aux_sym_add_operator_token1] = ACTIONS(1591),
    [sym_concat_operator] = ACTIONS(1593),
    [sym_rel_operator] = ACTIONS(1591),
    [anon_sym_AMP_AMP] = ACTIONS(1591),
    [anon_sym_or] = ACTIONS(1591),
    [anon_sym_PIPE_PIPE] = ACTIONS(1591),
    [sym_assign_operator] = ACTIONS(1593),
    [sym_let_operator] = ACTIONS(1593),
    [sym_let_and_operator] = ACTIONS(1593),
    [sym__uppercase_identifier] = ACTIONS(1593),
    [aux_sym_directive_token1] = ACTIONS(1591),
    [aux_sym_tag_token1] = ACTIONS(1593),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(965)] = {
    [sym_attribute] = STATE(965),
    [sym__lowercase_identifier] = ACTIONS(1777),
    [anon_sym_SEMI_SEMI] = ACTIONS(1779),
    [anon_sym_let] = ACTIONS(1777),
    [anon_sym_and] = ACTIONS(1777),
    [anon_sym_TILDE] = ACTIONS(1777),
    [anon_sym_QMARK] = ACTIONS(1777),
    [anon_sym_LPAREN] = ACTIONS(1779),
    [anon_sym_external] = ACTIONS(1777),
    [anon_sym_type] = ACTIONS(1777),
    [anon_sym_COMMA] = ACTIONS(1779),
    [anon_sym_PIPE] = ACTIONS(1777),
    [anon_sym_DOT] = ACTIONS(1779),
    [anon_sym_LBRACK] = ACTIONS(1777),
    [anon_sym_RBRACK] = ACTIONS(1779),
    [anon_sym_true] = ACTIONS(1777),
    [anon_sym_false] = ACTIONS(1777),
    [anon_sym_LBRACE] = ACTIONS(1777),
    [anon_sym_SEMI] = ACTIONS(1777),
    [anon_sym_exception] = ACTIONS(1777),
    [anon_sym_module] = ACTIONS(1777),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_include] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1777),
    [anon_sym_val] = ACTIONS(1777),
    [anon_sym_end] = ACTIONS(1777),
    [anon_sym_object] = ACTIONS(1777),
    [anon_sym_in] = ACTIONS(1777),
    [anon_sym_AMP] = ACTIONS(1777),
    [anon_sym_POUND] = ACTIONS(1777),
    [anon_sym_COLON_COLON] = ACTIONS(1779),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1779),
    [anon_sym_else] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1777),
    [anon_sym_LBRACE_LT] = ACTIONS(1779),
    [anon_sym_begin] = ACTIONS(1777),
    [sym_ocamlyacc_value] = ACTIONS(1779),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1777),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1779),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1777),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1777),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1779),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1779),
    [aux_sym_number_token1] = ACTIONS(1779),
    [aux_sym_signed_number_token1] = ACTIONS(1777),
    [anon_sym_SQUOTE] = ACTIONS(1779),
    [anon_sym_DQUOTE] = ACTIONS(1779),
    [sym_prefix_operator] = ACTIONS(1779),
    [aux_sym_sign_operator_token1] = ACTIONS(1777),
    [sym_hash_operator] = ACTIONS(1779),
    [anon_sym_lsl] = ACTIONS(1777),
    [anon_sym_lsr] = ACTIONS(1777),
    [anon_sym_asr] = ACTIONS(1777),
    [aux_sym_pow_operator_token1] = ACTIONS(1779),
    [anon_sym_mod] = ACTIONS(1777),
    [anon_sym_land] = ACTIONS(1777),
    [anon_sym_lor] = ACTIONS(1777),
    [anon_sym_lxor] = ACTIONS(1777),
    [aux_sym_mult_operator_token1] = ACTIONS(1777),
    [aux_sym_add_operator_token1] = ACTIONS(1777),
    [sym_concat_operator] = ACTIONS(1779),
    [sym_rel_operator] = ACTIONS(1777),
    [anon_sym_AMP_AMP] = ACTIONS(1777),
    [anon_sym_or] = ACTIONS(1777),
    [anon_sym_PIPE_PIPE] = ACTIONS(1777),
    [sym_assign_operator] = ACTIONS(1779),
    [sym_let_operator] = ACTIONS(1779),
    [sym_let_and_operator] = ACTIONS(1779),
    [sym__uppercase_identifier] = ACTIONS(1779),
    [aux_sym_directive_token1] = ACTIONS(1777),
    [aux_sym_tag_token1] = ACTIONS(1779),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(966)] = {
    [sym_attribute] = STATE(966),
    [sym__lowercase_identifier] = ACTIONS(1781),
    [anon_sym_SEMI_SEMI] = ACTIONS(1783),
    [anon_sym_let] = ACTIONS(1781),
    [anon_sym_and] = ACTIONS(1781),
    [anon_sym_TILDE] = ACTIONS(1781),
    [anon_sym_QMARK] = ACTIONS(1781),
    [anon_sym_LPAREN] = ACTIONS(1783),
    [anon_sym_external] = ACTIONS(1781),
    [anon_sym_type] = ACTIONS(1781),
    [anon_sym_COMMA] = ACTIONS(1783),
    [anon_sym_PIPE] = ACTIONS(1781),
    [anon_sym_DOT] = ACTIONS(1783),
    [anon_sym_LBRACK] = ACTIONS(1781),
    [anon_sym_RBRACK] = ACTIONS(1783),
    [anon_sym_true] = ACTIONS(1781),
    [anon_sym_false] = ACTIONS(1781),
    [anon_sym_LBRACE] = ACTIONS(1781),
    [anon_sym_SEMI] = ACTIONS(1781),
    [anon_sym_exception] = ACTIONS(1781),
    [anon_sym_module] = ACTIONS(1781),
    [anon_sym_open] = ACTIONS(1781),
    [anon_sym_include] = ACTIONS(1781),
    [anon_sym_class] = ACTIONS(1781),
    [anon_sym_val] = ACTIONS(1781),
    [anon_sym_end] = ACTIONS(1781),
    [anon_sym_object] = ACTIONS(1781),
    [anon_sym_in] = ACTIONS(1781),
    [anon_sym_AMP] = ACTIONS(1781),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_COLON_COLON] = ACTIONS(1783),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_else] = ACTIONS(1781),
    [anon_sym_new] = ACTIONS(1781),
    [anon_sym_LBRACE_LT] = ACTIONS(1783),
    [anon_sym_begin] = ACTIONS(1781),
    [sym_ocamlyacc_value] = ACTIONS(1783),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1781),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1783),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1781),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1781),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1783),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1783),
    [aux_sym_number_token1] = ACTIONS(1783),
    [aux_sym_signed_number_token1] = ACTIONS(1781),
    [anon_sym_SQUOTE] = ACTIONS(1783),
    [anon_sym_DQUOTE] = ACTIONS(1783),
    [sym_prefix_operator] = ACTIONS(1783),
    [aux_sym_sign_operator_token1] = ACTIONS(1781),
    [sym_hash_operator] = ACTIONS(1783),
    [anon_sym_lsl] = ACTIONS(1781),
    [anon_sym_lsr] = ACTIONS(1781),
    [anon_sym_asr] = ACTIONS(1781),
    [aux_sym_pow_operator_token1] = ACTIONS(1783),
    [anon_sym_mod] = ACTIONS(1781),
    [anon_sym_land] = ACTIONS(1781),
    [anon_sym_lor] = ACTIONS(1781),
    [anon_sym_lxor] = ACTIONS(1781),
    [aux_sym_mult_operator_token1] = ACTIONS(1781),
    [aux_sym_add_operator_token1] = ACTIONS(1781),
    [sym_concat_operator] = ACTIONS(1783),
    [sym_rel_operator] = ACTIONS(1781),
    [anon_sym_AMP_AMP] = ACTIONS(1781),
    [anon_sym_or] = ACTIONS(1781),
    [anon_sym_PIPE_PIPE] = ACTIONS(1781),
    [sym_assign_operator] = ACTIONS(1783),
    [sym_let_operator] = ACTIONS(1783),
    [sym_let_and_operator] = ACTIONS(1783),
    [sym__uppercase_identifier] = ACTIONS(1783),
    [aux_sym_directive_token1] = ACTIONS(1781),
    [aux_sym_tag_token1] = ACTIONS(1783),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(967)] = {
    [sym_parameter] = STATE(2207),
    [sym__class_typed] = STATE(3874),
    [sym__parenthesized_abstract_type] = STATE(2207),
    [sym__delimited_pattern] = STATE(2286),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2286),
    [sym_polymorphic_variant_pattern] = STATE(2286),
    [sym_record_pattern] = STATE(2287),
    [sym_list_pattern] = STATE(2287),
    [sym_array_pattern] = STATE(2287),
    [sym_local_open_pattern] = STATE(2286),
    [sym_package_pattern] = STATE(2286),
    [sym_parenthesized_pattern] = STATE(2287),
    [sym_attribute] = STATE(967),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2286),
    [sym_quoted_extension] = STATE(2286),
    [sym__constant] = STATE(2212),
    [sym__signed_constant] = STATE(2286),
    [sym_number] = STATE(2280),
    [sym_signed_number] = STATE(2212),
    [sym_character] = STATE(2280),
    [sym_string] = STATE(2280),
    [sym_quoted_string] = STATE(2280),
    [sym_boolean] = STATE(2280),
    [sym_unit] = STATE(2287),
    [sym__simple_value_pattern] = STATE(2286),
    [sym_parenthesized_operator] = STATE(2286),
    [sym_module_path] = STATE(6556),
    [sym_constructor_path] = STATE(2286),
    [sym__constructor_name] = STATE(2597),
    [sym_tag] = STATE(2286),
    [aux_sym_expression_item_repeat1] = STATE(3765),
    [aux_sym_let_binding_repeat1] = STATE(970),
    [sym__lowercase_identifier] = ACTIONS(1467),
    [anon_sym_SEMI_SEMI] = ACTIONS(2021),
    [anon_sym_and] = ACTIONS(2023),
    [anon_sym_EQ] = ACTIONS(2101),
    [anon_sym_TILDE] = ACTIONS(2091),
    [anon_sym_QMARK] = ACTIONS(2091),
    [anon_sym_LPAREN] = ACTIONS(2093),
    [anon_sym_external] = ACTIONS(2023),
    [anon_sym_type] = ACTIONS(2023),
    [anon_sym_COLON2] = ACTIONS(2095),
    [anon_sym_LBRACK] = ACTIONS(1485),
    [anon_sym_RBRACK] = ACTIONS(2021),
    [anon_sym_true] = ACTIONS(1487),
    [anon_sym_false] = ACTIONS(1487),
    [anon_sym_LBRACE] = ACTIONS(1489),
    [anon_sym_exception] = ACTIONS(2023),
    [anon_sym_module] = ACTIONS(2023),
    [anon_sym_open] = ACTIONS(2023),
    [anon_sym_include] = ACTIONS(2023),
    [anon_sym_class] = ACTIONS(2023),
    [anon_sym_val] = ACTIONS(2023),
    [anon_sym_end] = ACTIONS(2023),
    [anon_sym_POUND] = ACTIONS(1495),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1499),
    [anon_sym_begin] = ACTIONS(1501),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2021),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2021),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2021),
    [aux_sym_number_token1] = ACTIONS(1509),
    [aux_sym_signed_number_token1] = ACTIONS(1511),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE] = ACTIONS(1515),
    [sym__uppercase_identifier] = ACTIONS(2097),
    [aux_sym_tag_token1] = ACTIONS(2099),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(968)] = {
    [sym_parameter] = STATE(2207),
    [sym__class_typed] = STATE(3875),
    [sym__parenthesized_abstract_type] = STATE(2207),
    [sym__delimited_pattern] = STATE(2286),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2286),
    [sym_polymorphic_variant_pattern] = STATE(2286),
    [sym_record_pattern] = STATE(2287),
    [sym_list_pattern] = STATE(2287),
    [sym_array_pattern] = STATE(2287),
    [sym_local_open_pattern] = STATE(2286),
    [sym_package_pattern] = STATE(2286),
    [sym_parenthesized_pattern] = STATE(2287),
    [sym_attribute] = STATE(968),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2286),
    [sym_quoted_extension] = STATE(2286),
    [sym__constant] = STATE(2212),
    [sym__signed_constant] = STATE(2286),
    [sym_number] = STATE(2280),
    [sym_signed_number] = STATE(2212),
    [sym_character] = STATE(2280),
    [sym_string] = STATE(2280),
    [sym_quoted_string] = STATE(2280),
    [sym_boolean] = STATE(2280),
    [sym_unit] = STATE(2287),
    [sym__simple_value_pattern] = STATE(2286),
    [sym_parenthesized_operator] = STATE(2286),
    [sym_module_path] = STATE(6556),
    [sym_constructor_path] = STATE(2286),
    [sym__constructor_name] = STATE(2597),
    [sym_tag] = STATE(2286),
    [aux_sym_expression_item_repeat1] = STATE(3584),
    [aux_sym_let_binding_repeat1] = STATE(1130),
    [sym__lowercase_identifier] = ACTIONS(1467),
    [anon_sym_SEMI_SEMI] = ACTIONS(2027),
    [anon_sym_and] = ACTIONS(2029),
    [anon_sym_EQ] = ACTIONS(2103),
    [anon_sym_TILDE] = ACTIONS(2091),
    [anon_sym_QMARK] = ACTIONS(2091),
    [anon_sym_LPAREN] = ACTIONS(2093),
    [anon_sym_external] = ACTIONS(2029),
    [anon_sym_type] = ACTIONS(2029),
    [anon_sym_COLON2] = ACTIONS(2095),
    [anon_sym_LBRACK] = ACTIONS(1485),
    [anon_sym_RBRACK] = ACTIONS(2027),
    [anon_sym_true] = ACTIONS(1487),
    [anon_sym_false] = ACTIONS(1487),
    [anon_sym_LBRACE] = ACTIONS(1489),
    [anon_sym_exception] = ACTIONS(2029),
    [anon_sym_module] = ACTIONS(2029),
    [anon_sym_open] = ACTIONS(2029),
    [anon_sym_include] = ACTIONS(2029),
    [anon_sym_class] = ACTIONS(2029),
    [anon_sym_val] = ACTIONS(2029),
    [anon_sym_end] = ACTIONS(2029),
    [anon_sym_POUND] = ACTIONS(1495),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1499),
    [anon_sym_begin] = ACTIONS(1501),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2027),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2027),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2027),
    [aux_sym_number_token1] = ACTIONS(1509),
    [aux_sym_signed_number_token1] = ACTIONS(1511),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE] = ACTIONS(1515),
    [sym__uppercase_identifier] = ACTIONS(2097),
    [aux_sym_tag_token1] = ACTIONS(2099),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(969)] = {
    [sym_attribute] = STATE(969),
    [sym__lowercase_identifier] = ACTIONS(1627),
    [anon_sym_SEMI_SEMI] = ACTIONS(1629),
    [anon_sym_let] = ACTIONS(1627),
    [anon_sym_and] = ACTIONS(1627),
    [anon_sym_TILDE] = ACTIONS(1627),
    [anon_sym_QMARK] = ACTIONS(1627),
    [anon_sym_LPAREN] = ACTIONS(1629),
    [anon_sym_external] = ACTIONS(1627),
    [anon_sym_type] = ACTIONS(1627),
    [anon_sym_COMMA] = ACTIONS(1629),
    [anon_sym_PIPE] = ACTIONS(1627),
    [anon_sym_DOT] = ACTIONS(1629),
    [anon_sym_LBRACK] = ACTIONS(1627),
    [anon_sym_RBRACK] = ACTIONS(1629),
    [anon_sym_true] = ACTIONS(1627),
    [anon_sym_false] = ACTIONS(1627),
    [anon_sym_LBRACE] = ACTIONS(1627),
    [anon_sym_SEMI] = ACTIONS(1627),
    [anon_sym_exception] = ACTIONS(1627),
    [anon_sym_module] = ACTIONS(1627),
    [anon_sym_open] = ACTIONS(1627),
    [anon_sym_include] = ACTIONS(1627),
    [anon_sym_class] = ACTIONS(1627),
    [anon_sym_val] = ACTIONS(1627),
    [anon_sym_end] = ACTIONS(1627),
    [anon_sym_object] = ACTIONS(1627),
    [anon_sym_in] = ACTIONS(1627),
    [anon_sym_AMP] = ACTIONS(1627),
    [anon_sym_POUND] = ACTIONS(1627),
    [anon_sym_COLON_COLON] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1629),
    [anon_sym_else] = ACTIONS(1627),
    [anon_sym_new] = ACTIONS(1627),
    [anon_sym_LBRACE_LT] = ACTIONS(1629),
    [anon_sym_begin] = ACTIONS(1627),
    [sym_ocamlyacc_value] = ACTIONS(1629),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1627),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1629),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1627),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1627),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1629),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1629),
    [aux_sym_number_token1] = ACTIONS(1629),
    [aux_sym_signed_number_token1] = ACTIONS(1627),
    [anon_sym_SQUOTE] = ACTIONS(1629),
    [anon_sym_DQUOTE] = ACTIONS(1629),
    [sym_prefix_operator] = ACTIONS(1629),
    [aux_sym_sign_operator_token1] = ACTIONS(1627),
    [sym_hash_operator] = ACTIONS(1629),
    [anon_sym_lsl] = ACTIONS(1627),
    [anon_sym_lsr] = ACTIONS(1627),
    [anon_sym_asr] = ACTIONS(1627),
    [aux_sym_pow_operator_token1] = ACTIONS(1629),
    [anon_sym_mod] = ACTIONS(1627),
    [anon_sym_land] = ACTIONS(1627),
    [anon_sym_lor] = ACTIONS(1627),
    [anon_sym_lxor] = ACTIONS(1627),
    [aux_sym_mult_operator_token1] = ACTIONS(1627),
    [aux_sym_add_operator_token1] = ACTIONS(1627),
    [sym_concat_operator] = ACTIONS(1629),
    [sym_rel_operator] = ACTIONS(1627),
    [anon_sym_AMP_AMP] = ACTIONS(1627),
    [anon_sym_or] = ACTIONS(1627),
    [anon_sym_PIPE_PIPE] = ACTIONS(1627),
    [sym_assign_operator] = ACTIONS(1629),
    [sym_let_operator] = ACTIONS(1629),
    [sym_let_and_operator] = ACTIONS(1629),
    [sym__uppercase_identifier] = ACTIONS(1629),
    [aux_sym_directive_token1] = ACTIONS(1627),
    [aux_sym_tag_token1] = ACTIONS(1629),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(970)] = {
    [sym_parameter] = STATE(2207),
    [sym__class_typed] = STATE(3885),
    [sym__parenthesized_abstract_type] = STATE(2207),
    [sym__delimited_pattern] = STATE(2286),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2286),
    [sym_polymorphic_variant_pattern] = STATE(2286),
    [sym_record_pattern] = STATE(2287),
    [sym_list_pattern] = STATE(2287),
    [sym_array_pattern] = STATE(2287),
    [sym_local_open_pattern] = STATE(2286),
    [sym_package_pattern] = STATE(2286),
    [sym_parenthesized_pattern] = STATE(2287),
    [sym_attribute] = STATE(970),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2286),
    [sym_quoted_extension] = STATE(2286),
    [sym__constant] = STATE(2212),
    [sym__signed_constant] = STATE(2286),
    [sym_number] = STATE(2280),
    [sym_signed_number] = STATE(2212),
    [sym_character] = STATE(2280),
    [sym_string] = STATE(2280),
    [sym_quoted_string] = STATE(2280),
    [sym_boolean] = STATE(2280),
    [sym_unit] = STATE(2287),
    [sym__simple_value_pattern] = STATE(2286),
    [sym_parenthesized_operator] = STATE(2286),
    [sym_module_path] = STATE(6556),
    [sym_constructor_path] = STATE(2286),
    [sym__constructor_name] = STATE(2597),
    [sym_tag] = STATE(2286),
    [aux_sym_expression_item_repeat1] = STATE(3626),
    [aux_sym_let_binding_repeat1] = STATE(1130),
    [sym__lowercase_identifier] = ACTIONS(1467),
    [anon_sym_SEMI_SEMI] = ACTIONS(2057),
    [anon_sym_and] = ACTIONS(2059),
    [anon_sym_EQ] = ACTIONS(2105),
    [anon_sym_TILDE] = ACTIONS(2091),
    [anon_sym_QMARK] = ACTIONS(2091),
    [anon_sym_LPAREN] = ACTIONS(2093),
    [anon_sym_external] = ACTIONS(2059),
    [anon_sym_type] = ACTIONS(2059),
    [anon_sym_COLON2] = ACTIONS(2095),
    [anon_sym_LBRACK] = ACTIONS(1485),
    [anon_sym_RBRACK] = ACTIONS(2057),
    [anon_sym_true] = ACTIONS(1487),
    [anon_sym_false] = ACTIONS(1487),
    [anon_sym_LBRACE] = ACTIONS(1489),
    [anon_sym_exception] = ACTIONS(2059),
    [anon_sym_module] = ACTIONS(2059),
    [anon_sym_open] = ACTIONS(2059),
    [anon_sym_include] = ACTIONS(2059),
    [anon_sym_class] = ACTIONS(2059),
    [anon_sym_val] = ACTIONS(2059),
    [anon_sym_end] = ACTIONS(2059),
    [anon_sym_POUND] = ACTIONS(1495),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1499),
    [anon_sym_begin] = ACTIONS(1501),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2057),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2057),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2057),
    [aux_sym_number_token1] = ACTIONS(1509),
    [aux_sym_signed_number_token1] = ACTIONS(1511),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE] = ACTIONS(1515),
    [sym__uppercase_identifier] = ACTIONS(2097),
    [aux_sym_tag_token1] = ACTIONS(2099),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(971)] = {
    [sym_attribute] = STATE(971),
    [sym__lowercase_identifier] = ACTIONS(1713),
    [anon_sym_SEMI_SEMI] = ACTIONS(1715),
    [anon_sym_let] = ACTIONS(1713),
    [anon_sym_and] = ACTIONS(1713),
    [anon_sym_TILDE] = ACTIONS(1713),
    [anon_sym_QMARK] = ACTIONS(1713),
    [anon_sym_LPAREN] = ACTIONS(1715),
    [anon_sym_external] = ACTIONS(1713),
    [anon_sym_type] = ACTIONS(1713),
    [anon_sym_COMMA] = ACTIONS(1715),
    [anon_sym_PIPE] = ACTIONS(1713),
    [anon_sym_DOT] = ACTIONS(1715),
    [anon_sym_LBRACK] = ACTIONS(1713),
    [anon_sym_RBRACK] = ACTIONS(1715),
    [anon_sym_true] = ACTIONS(1713),
    [anon_sym_false] = ACTIONS(1713),
    [anon_sym_LBRACE] = ACTIONS(1713),
    [anon_sym_SEMI] = ACTIONS(1713),
    [anon_sym_exception] = ACTIONS(1713),
    [anon_sym_module] = ACTIONS(1713),
    [anon_sym_open] = ACTIONS(1713),
    [anon_sym_include] = ACTIONS(1713),
    [anon_sym_class] = ACTIONS(1713),
    [anon_sym_val] = ACTIONS(1713),
    [anon_sym_end] = ACTIONS(1713),
    [anon_sym_object] = ACTIONS(1713),
    [anon_sym_in] = ACTIONS(1713),
    [anon_sym_AMP] = ACTIONS(1713),
    [anon_sym_POUND] = ACTIONS(1713),
    [anon_sym_COLON_COLON] = ACTIONS(1715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1715),
    [anon_sym_else] = ACTIONS(1713),
    [anon_sym_new] = ACTIONS(1713),
    [anon_sym_LBRACE_LT] = ACTIONS(1715),
    [anon_sym_begin] = ACTIONS(1713),
    [sym_ocamlyacc_value] = ACTIONS(1715),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1713),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1715),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1713),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1713),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1715),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1715),
    [aux_sym_number_token1] = ACTIONS(1715),
    [aux_sym_signed_number_token1] = ACTIONS(1713),
    [anon_sym_SQUOTE] = ACTIONS(1715),
    [anon_sym_DQUOTE] = ACTIONS(1715),
    [sym_prefix_operator] = ACTIONS(1715),
    [aux_sym_sign_operator_token1] = ACTIONS(1713),
    [sym_hash_operator] = ACTIONS(1715),
    [anon_sym_lsl] = ACTIONS(1713),
    [anon_sym_lsr] = ACTIONS(1713),
    [anon_sym_asr] = ACTIONS(1713),
    [aux_sym_pow_operator_token1] = ACTIONS(1715),
    [anon_sym_mod] = ACTIONS(1713),
    [anon_sym_land] = ACTIONS(1713),
    [anon_sym_lor] = ACTIONS(1713),
    [anon_sym_lxor] = ACTIONS(1713),
    [aux_sym_mult_operator_token1] = ACTIONS(1713),
    [aux_sym_add_operator_token1] = ACTIONS(1713),
    [sym_concat_operator] = ACTIONS(1715),
    [sym_rel_operator] = ACTIONS(1713),
    [anon_sym_AMP_AMP] = ACTIONS(1713),
    [anon_sym_or] = ACTIONS(1713),
    [anon_sym_PIPE_PIPE] = ACTIONS(1713),
    [sym_assign_operator] = ACTIONS(1715),
    [sym_let_operator] = ACTIONS(1715),
    [sym_let_and_operator] = ACTIONS(1715),
    [sym__uppercase_identifier] = ACTIONS(1715),
    [aux_sym_directive_token1] = ACTIONS(1713),
    [aux_sym_tag_token1] = ACTIONS(1715),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(972)] = {
    [sym_parameter] = STATE(2207),
    [sym__class_typed] = STATE(3893),
    [sym__parenthesized_abstract_type] = STATE(2207),
    [sym__delimited_pattern] = STATE(2286),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2286),
    [sym_polymorphic_variant_pattern] = STATE(2286),
    [sym_record_pattern] = STATE(2287),
    [sym_list_pattern] = STATE(2287),
    [sym_array_pattern] = STATE(2287),
    [sym_local_open_pattern] = STATE(2286),
    [sym_package_pattern] = STATE(2286),
    [sym_parenthesized_pattern] = STATE(2287),
    [sym_attribute] = STATE(972),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2286),
    [sym_quoted_extension] = STATE(2286),
    [sym__constant] = STATE(2212),
    [sym__signed_constant] = STATE(2286),
    [sym_number] = STATE(2280),
    [sym_signed_number] = STATE(2212),
    [sym_character] = STATE(2280),
    [sym_string] = STATE(2280),
    [sym_quoted_string] = STATE(2280),
    [sym_boolean] = STATE(2280),
    [sym_unit] = STATE(2287),
    [sym__simple_value_pattern] = STATE(2286),
    [sym_parenthesized_operator] = STATE(2286),
    [sym_module_path] = STATE(6556),
    [sym_constructor_path] = STATE(2286),
    [sym__constructor_name] = STATE(2597),
    [sym_tag] = STATE(2286),
    [aux_sym_expression_item_repeat1] = STATE(3625),
    [aux_sym_let_binding_repeat1] = STATE(973),
    [sym__lowercase_identifier] = ACTIONS(1467),
    [anon_sym_SEMI_SEMI] = ACTIONS(1975),
    [anon_sym_and] = ACTIONS(1977),
    [anon_sym_EQ] = ACTIONS(2107),
    [anon_sym_TILDE] = ACTIONS(2091),
    [anon_sym_QMARK] = ACTIONS(2091),
    [anon_sym_LPAREN] = ACTIONS(2093),
    [anon_sym_external] = ACTIONS(1977),
    [anon_sym_type] = ACTIONS(1977),
    [anon_sym_COLON2] = ACTIONS(2095),
    [anon_sym_LBRACK] = ACTIONS(1485),
    [anon_sym_RBRACK] = ACTIONS(1975),
    [anon_sym_true] = ACTIONS(1487),
    [anon_sym_false] = ACTIONS(1487),
    [anon_sym_LBRACE] = ACTIONS(1489),
    [anon_sym_exception] = ACTIONS(1977),
    [anon_sym_module] = ACTIONS(1977),
    [anon_sym_open] = ACTIONS(1977),
    [anon_sym_include] = ACTIONS(1977),
    [anon_sym_class] = ACTIONS(1977),
    [anon_sym_val] = ACTIONS(1977),
    [anon_sym_end] = ACTIONS(1977),
    [anon_sym_POUND] = ACTIONS(1495),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1499),
    [anon_sym_begin] = ACTIONS(1501),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1975),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1975),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1975),
    [aux_sym_number_token1] = ACTIONS(1509),
    [aux_sym_signed_number_token1] = ACTIONS(1511),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE] = ACTIONS(1515),
    [sym__uppercase_identifier] = ACTIONS(2097),
    [aux_sym_tag_token1] = ACTIONS(2099),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(973)] = {
    [sym_parameter] = STATE(2207),
    [sym__class_typed] = STATE(3898),
    [sym__parenthesized_abstract_type] = STATE(2207),
    [sym__delimited_pattern] = STATE(2286),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2286),
    [sym_polymorphic_variant_pattern] = STATE(2286),
    [sym_record_pattern] = STATE(2287),
    [sym_list_pattern] = STATE(2287),
    [sym_array_pattern] = STATE(2287),
    [sym_local_open_pattern] = STATE(2286),
    [sym_package_pattern] = STATE(2286),
    [sym_parenthesized_pattern] = STATE(2287),
    [sym_attribute] = STATE(973),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2286),
    [sym_quoted_extension] = STATE(2286),
    [sym__constant] = STATE(2212),
    [sym__signed_constant] = STATE(2286),
    [sym_number] = STATE(2280),
    [sym_signed_number] = STATE(2212),
    [sym_character] = STATE(2280),
    [sym_string] = STATE(2280),
    [sym_quoted_string] = STATE(2280),
    [sym_boolean] = STATE(2280),
    [sym_unit] = STATE(2287),
    [sym__simple_value_pattern] = STATE(2286),
    [sym_parenthesized_operator] = STATE(2286),
    [sym_module_path] = STATE(6556),
    [sym_constructor_path] = STATE(2286),
    [sym__constructor_name] = STATE(2597),
    [sym_tag] = STATE(2286),
    [aux_sym_expression_item_repeat1] = STATE(3721),
    [aux_sym_let_binding_repeat1] = STATE(1130),
    [sym__lowercase_identifier] = ACTIONS(1467),
    [anon_sym_SEMI_SEMI] = ACTIONS(2063),
    [anon_sym_and] = ACTIONS(2065),
    [anon_sym_EQ] = ACTIONS(2109),
    [anon_sym_TILDE] = ACTIONS(2091),
    [anon_sym_QMARK] = ACTIONS(2091),
    [anon_sym_LPAREN] = ACTIONS(2093),
    [anon_sym_external] = ACTIONS(2065),
    [anon_sym_type] = ACTIONS(2065),
    [anon_sym_COLON2] = ACTIONS(2095),
    [anon_sym_LBRACK] = ACTIONS(1485),
    [anon_sym_RBRACK] = ACTIONS(2063),
    [anon_sym_true] = ACTIONS(1487),
    [anon_sym_false] = ACTIONS(1487),
    [anon_sym_LBRACE] = ACTIONS(1489),
    [anon_sym_exception] = ACTIONS(2065),
    [anon_sym_module] = ACTIONS(2065),
    [anon_sym_open] = ACTIONS(2065),
    [anon_sym_include] = ACTIONS(2065),
    [anon_sym_class] = ACTIONS(2065),
    [anon_sym_val] = ACTIONS(2065),
    [anon_sym_end] = ACTIONS(2065),
    [anon_sym_POUND] = ACTIONS(1495),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1499),
    [anon_sym_begin] = ACTIONS(1501),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2063),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2063),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2063),
    [aux_sym_number_token1] = ACTIONS(1509),
    [aux_sym_signed_number_token1] = ACTIONS(1511),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE] = ACTIONS(1515),
    [sym__uppercase_identifier] = ACTIONS(2097),
    [aux_sym_tag_token1] = ACTIONS(2099),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(974)] = {
    [sym_parameter] = STATE(2207),
    [sym__class_typed] = STATE(3899),
    [sym__parenthesized_abstract_type] = STATE(2207),
    [sym__delimited_pattern] = STATE(2286),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2286),
    [sym_polymorphic_variant_pattern] = STATE(2286),
    [sym_record_pattern] = STATE(2287),
    [sym_list_pattern] = STATE(2287),
    [sym_array_pattern] = STATE(2287),
    [sym_local_open_pattern] = STATE(2286),
    [sym_package_pattern] = STATE(2286),
    [sym_parenthesized_pattern] = STATE(2287),
    [sym_attribute] = STATE(974),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2286),
    [sym_quoted_extension] = STATE(2286),
    [sym__constant] = STATE(2212),
    [sym__signed_constant] = STATE(2286),
    [sym_number] = STATE(2280),
    [sym_signed_number] = STATE(2212),
    [sym_character] = STATE(2280),
    [sym_string] = STATE(2280),
    [sym_quoted_string] = STATE(2280),
    [sym_boolean] = STATE(2280),
    [sym_unit] = STATE(2287),
    [sym__simple_value_pattern] = STATE(2286),
    [sym_parenthesized_operator] = STATE(2286),
    [sym_module_path] = STATE(6556),
    [sym_constructor_path] = STATE(2286),
    [sym__constructor_name] = STATE(2597),
    [sym_tag] = STATE(2286),
    [aux_sym_expression_item_repeat1] = STATE(3723),
    [aux_sym_let_binding_repeat1] = STATE(975),
    [sym__lowercase_identifier] = ACTIONS(1467),
    [anon_sym_SEMI_SEMI] = ACTIONS(2069),
    [anon_sym_and] = ACTIONS(2071),
    [anon_sym_EQ] = ACTIONS(2111),
    [anon_sym_TILDE] = ACTIONS(2091),
    [anon_sym_QMARK] = ACTIONS(2091),
    [anon_sym_LPAREN] = ACTIONS(2093),
    [anon_sym_external] = ACTIONS(2071),
    [anon_sym_type] = ACTIONS(2071),
    [anon_sym_COLON2] = ACTIONS(2095),
    [anon_sym_LBRACK] = ACTIONS(1485),
    [anon_sym_RBRACK] = ACTIONS(2069),
    [anon_sym_true] = ACTIONS(1487),
    [anon_sym_false] = ACTIONS(1487),
    [anon_sym_LBRACE] = ACTIONS(1489),
    [anon_sym_exception] = ACTIONS(2071),
    [anon_sym_module] = ACTIONS(2071),
    [anon_sym_open] = ACTIONS(2071),
    [anon_sym_include] = ACTIONS(2071),
    [anon_sym_class] = ACTIONS(2071),
    [anon_sym_val] = ACTIONS(2071),
    [anon_sym_end] = ACTIONS(2071),
    [anon_sym_POUND] = ACTIONS(1495),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1499),
    [anon_sym_begin] = ACTIONS(1501),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2069),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2069),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2069),
    [aux_sym_number_token1] = ACTIONS(1509),
    [aux_sym_signed_number_token1] = ACTIONS(1511),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE] = ACTIONS(1515),
    [sym__uppercase_identifier] = ACTIONS(2097),
    [aux_sym_tag_token1] = ACTIONS(2099),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(975)] = {
    [sym_parameter] = STATE(2207),
    [sym__class_typed] = STATE(3903),
    [sym__parenthesized_abstract_type] = STATE(2207),
    [sym__delimited_pattern] = STATE(2286),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2286),
    [sym_polymorphic_variant_pattern] = STATE(2286),
    [sym_record_pattern] = STATE(2287),
    [sym_list_pattern] = STATE(2287),
    [sym_array_pattern] = STATE(2287),
    [sym_local_open_pattern] = STATE(2286),
    [sym_package_pattern] = STATE(2286),
    [sym_parenthesized_pattern] = STATE(2287),
    [sym_attribute] = STATE(975),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2286),
    [sym_quoted_extension] = STATE(2286),
    [sym__constant] = STATE(2212),
    [sym__signed_constant] = STATE(2286),
    [sym_number] = STATE(2280),
    [sym_signed_number] = STATE(2212),
    [sym_character] = STATE(2280),
    [sym_string] = STATE(2280),
    [sym_quoted_string] = STATE(2280),
    [sym_boolean] = STATE(2280),
    [sym_unit] = STATE(2287),
    [sym__simple_value_pattern] = STATE(2286),
    [sym_parenthesized_operator] = STATE(2286),
    [sym_module_path] = STATE(6556),
    [sym_constructor_path] = STATE(2286),
    [sym__constructor_name] = STATE(2597),
    [sym_tag] = STATE(2286),
    [aux_sym_expression_item_repeat1] = STATE(3596),
    [aux_sym_let_binding_repeat1] = STATE(1130),
    [sym__lowercase_identifier] = ACTIONS(1467),
    [anon_sym_SEMI_SEMI] = ACTIONS(2033),
    [anon_sym_and] = ACTIONS(2035),
    [anon_sym_EQ] = ACTIONS(2113),
    [anon_sym_TILDE] = ACTIONS(2091),
    [anon_sym_QMARK] = ACTIONS(2091),
    [anon_sym_LPAREN] = ACTIONS(2093),
    [anon_sym_external] = ACTIONS(2035),
    [anon_sym_type] = ACTIONS(2035),
    [anon_sym_COLON2] = ACTIONS(2095),
    [anon_sym_LBRACK] = ACTIONS(1485),
    [anon_sym_RBRACK] = ACTIONS(2033),
    [anon_sym_true] = ACTIONS(1487),
    [anon_sym_false] = ACTIONS(1487),
    [anon_sym_LBRACE] = ACTIONS(1489),
    [anon_sym_exception] = ACTIONS(2035),
    [anon_sym_module] = ACTIONS(2035),
    [anon_sym_open] = ACTIONS(2035),
    [anon_sym_include] = ACTIONS(2035),
    [anon_sym_class] = ACTIONS(2035),
    [anon_sym_val] = ACTIONS(2035),
    [anon_sym_end] = ACTIONS(2035),
    [anon_sym_POUND] = ACTIONS(1495),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1499),
    [anon_sym_begin] = ACTIONS(1501),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2033),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2033),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2033),
    [aux_sym_number_token1] = ACTIONS(1509),
    [aux_sym_signed_number_token1] = ACTIONS(1511),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE] = ACTIONS(1515),
    [sym__uppercase_identifier] = ACTIONS(2097),
    [aux_sym_tag_token1] = ACTIONS(2099),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(976)] = {
    [sym_parameter] = STATE(2207),
    [sym__class_typed] = STATE(3904),
    [sym__parenthesized_abstract_type] = STATE(2207),
    [sym__delimited_pattern] = STATE(2286),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2286),
    [sym_polymorphic_variant_pattern] = STATE(2286),
    [sym_record_pattern] = STATE(2287),
    [sym_list_pattern] = STATE(2287),
    [sym_array_pattern] = STATE(2287),
    [sym_local_open_pattern] = STATE(2286),
    [sym_package_pattern] = STATE(2286),
    [sym_parenthesized_pattern] = STATE(2287),
    [sym_attribute] = STATE(976),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2286),
    [sym_quoted_extension] = STATE(2286),
    [sym__constant] = STATE(2212),
    [sym__signed_constant] = STATE(2286),
    [sym_number] = STATE(2280),
    [sym_signed_number] = STATE(2212),
    [sym_character] = STATE(2280),
    [sym_string] = STATE(2280),
    [sym_quoted_string] = STATE(2280),
    [sym_boolean] = STATE(2280),
    [sym_unit] = STATE(2287),
    [sym__simple_value_pattern] = STATE(2286),
    [sym_parenthesized_operator] = STATE(2286),
    [sym_module_path] = STATE(6556),
    [sym_constructor_path] = STATE(2286),
    [sym__constructor_name] = STATE(2597),
    [sym_tag] = STATE(2286),
    [aux_sym_expression_item_repeat1] = STATE(3602),
    [aux_sym_let_binding_repeat1] = STATE(977),
    [sym__lowercase_identifier] = ACTIONS(1467),
    [anon_sym_SEMI_SEMI] = ACTIONS(2039),
    [anon_sym_and] = ACTIONS(2041),
    [anon_sym_EQ] = ACTIONS(2115),
    [anon_sym_TILDE] = ACTIONS(2091),
    [anon_sym_QMARK] = ACTIONS(2091),
    [anon_sym_LPAREN] = ACTIONS(2093),
    [anon_sym_external] = ACTIONS(2041),
    [anon_sym_type] = ACTIONS(2041),
    [anon_sym_COLON2] = ACTIONS(2095),
    [anon_sym_LBRACK] = ACTIONS(1485),
    [anon_sym_RBRACK] = ACTIONS(2039),
    [anon_sym_true] = ACTIONS(1487),
    [anon_sym_false] = ACTIONS(1487),
    [anon_sym_LBRACE] = ACTIONS(1489),
    [anon_sym_exception] = ACTIONS(2041),
    [anon_sym_module] = ACTIONS(2041),
    [anon_sym_open] = ACTIONS(2041),
    [anon_sym_include] = ACTIONS(2041),
    [anon_sym_class] = ACTIONS(2041),
    [anon_sym_val] = ACTIONS(2041),
    [anon_sym_end] = ACTIONS(2041),
    [anon_sym_POUND] = ACTIONS(1495),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1499),
    [anon_sym_begin] = ACTIONS(1501),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2039),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2039),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2039),
    [aux_sym_number_token1] = ACTIONS(1509),
    [aux_sym_signed_number_token1] = ACTIONS(1511),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE] = ACTIONS(1515),
    [sym__uppercase_identifier] = ACTIONS(2097),
    [aux_sym_tag_token1] = ACTIONS(2099),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(977)] = {
    [sym_parameter] = STATE(2207),
    [sym__class_typed] = STATE(3906),
    [sym__parenthesized_abstract_type] = STATE(2207),
    [sym__delimited_pattern] = STATE(2286),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2286),
    [sym_polymorphic_variant_pattern] = STATE(2286),
    [sym_record_pattern] = STATE(2287),
    [sym_list_pattern] = STATE(2287),
    [sym_array_pattern] = STATE(2287),
    [sym_local_open_pattern] = STATE(2286),
    [sym_package_pattern] = STATE(2286),
    [sym_parenthesized_pattern] = STATE(2287),
    [sym_attribute] = STATE(977),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2286),
    [sym_quoted_extension] = STATE(2286),
    [sym__constant] = STATE(2212),
    [sym__signed_constant] = STATE(2286),
    [sym_number] = STATE(2280),
    [sym_signed_number] = STATE(2212),
    [sym_character] = STATE(2280),
    [sym_string] = STATE(2280),
    [sym_quoted_string] = STATE(2280),
    [sym_boolean] = STATE(2280),
    [sym_unit] = STATE(2287),
    [sym__simple_value_pattern] = STATE(2286),
    [sym_parenthesized_operator] = STATE(2286),
    [sym_module_path] = STATE(6556),
    [sym_constructor_path] = STATE(2286),
    [sym__constructor_name] = STATE(2597),
    [sym_tag] = STATE(2286),
    [aux_sym_expression_item_repeat1] = STATE(3736),
    [aux_sym_let_binding_repeat1] = STATE(1130),
    [sym__lowercase_identifier] = ACTIONS(1467),
    [anon_sym_SEMI_SEMI] = ACTIONS(2051),
    [anon_sym_and] = ACTIONS(2053),
    [anon_sym_EQ] = ACTIONS(2117),
    [anon_sym_TILDE] = ACTIONS(2091),
    [anon_sym_QMARK] = ACTIONS(2091),
    [anon_sym_LPAREN] = ACTIONS(2093),
    [anon_sym_external] = ACTIONS(2053),
    [anon_sym_type] = ACTIONS(2053),
    [anon_sym_COLON2] = ACTIONS(2095),
    [anon_sym_LBRACK] = ACTIONS(1485),
    [anon_sym_RBRACK] = ACTIONS(2051),
    [anon_sym_true] = ACTIONS(1487),
    [anon_sym_false] = ACTIONS(1487),
    [anon_sym_LBRACE] = ACTIONS(1489),
    [anon_sym_exception] = ACTIONS(2053),
    [anon_sym_module] = ACTIONS(2053),
    [anon_sym_open] = ACTIONS(2053),
    [anon_sym_include] = ACTIONS(2053),
    [anon_sym_class] = ACTIONS(2053),
    [anon_sym_val] = ACTIONS(2053),
    [anon_sym_end] = ACTIONS(2053),
    [anon_sym_POUND] = ACTIONS(1495),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1499),
    [anon_sym_begin] = ACTIONS(1501),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2051),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2051),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2051),
    [aux_sym_number_token1] = ACTIONS(1509),
    [aux_sym_signed_number_token1] = ACTIONS(1511),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE] = ACTIONS(1515),
    [sym__uppercase_identifier] = ACTIONS(2097),
    [aux_sym_tag_token1] = ACTIONS(2099),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(978)] = {
    [sym_attribute] = STATE(978),
    [sym__lowercase_identifier] = ACTIONS(1847),
    [anon_sym_SEMI_SEMI] = ACTIONS(1849),
    [anon_sym_let] = ACTIONS(1847),
    [anon_sym_and] = ACTIONS(1847),
    [anon_sym_TILDE] = ACTIONS(1847),
    [anon_sym_QMARK] = ACTIONS(1847),
    [anon_sym_LPAREN] = ACTIONS(1849),
    [anon_sym_external] = ACTIONS(1847),
    [anon_sym_type] = ACTIONS(1847),
    [anon_sym_COMMA] = ACTIONS(1849),
    [anon_sym_PIPE] = ACTIONS(1847),
    [anon_sym_DOT] = ACTIONS(1849),
    [anon_sym_LBRACK] = ACTIONS(1847),
    [anon_sym_RBRACK] = ACTIONS(1849),
    [anon_sym_true] = ACTIONS(1847),
    [anon_sym_false] = ACTIONS(1847),
    [anon_sym_LBRACE] = ACTIONS(1847),
    [anon_sym_SEMI] = ACTIONS(1847),
    [anon_sym_exception] = ACTIONS(1847),
    [anon_sym_module] = ACTIONS(1847),
    [anon_sym_open] = ACTIONS(1847),
    [anon_sym_include] = ACTIONS(1847),
    [anon_sym_class] = ACTIONS(1847),
    [anon_sym_val] = ACTIONS(1847),
    [anon_sym_end] = ACTIONS(1847),
    [anon_sym_object] = ACTIONS(1847),
    [anon_sym_in] = ACTIONS(1847),
    [anon_sym_AMP] = ACTIONS(1847),
    [anon_sym_POUND] = ACTIONS(1847),
    [anon_sym_COLON_COLON] = ACTIONS(1849),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_else] = ACTIONS(1847),
    [anon_sym_new] = ACTIONS(1847),
    [anon_sym_LBRACE_LT] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1847),
    [sym_ocamlyacc_value] = ACTIONS(1849),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1847),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1849),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1847),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1847),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1849),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1849),
    [aux_sym_number_token1] = ACTIONS(1849),
    [aux_sym_signed_number_token1] = ACTIONS(1847),
    [anon_sym_SQUOTE] = ACTIONS(1849),
    [anon_sym_DQUOTE] = ACTIONS(1849),
    [sym_prefix_operator] = ACTIONS(1849),
    [aux_sym_sign_operator_token1] = ACTIONS(1847),
    [sym_hash_operator] = ACTIONS(1849),
    [anon_sym_lsl] = ACTIONS(1847),
    [anon_sym_lsr] = ACTIONS(1847),
    [anon_sym_asr] = ACTIONS(1847),
    [aux_sym_pow_operator_token1] = ACTIONS(1849),
    [anon_sym_mod] = ACTIONS(1847),
    [anon_sym_land] = ACTIONS(1847),
    [anon_sym_lor] = ACTIONS(1847),
    [anon_sym_lxor] = ACTIONS(1847),
    [aux_sym_mult_operator_token1] = ACTIONS(1847),
    [aux_sym_add_operator_token1] = ACTIONS(1847),
    [sym_concat_operator] = ACTIONS(1849),
    [sym_rel_operator] = ACTIONS(1847),
    [anon_sym_AMP_AMP] = ACTIONS(1847),
    [anon_sym_or] = ACTIONS(1847),
    [anon_sym_PIPE_PIPE] = ACTIONS(1847),
    [sym_assign_operator] = ACTIONS(1849),
    [sym_let_operator] = ACTIONS(1849),
    [sym_let_and_operator] = ACTIONS(1849),
    [sym__uppercase_identifier] = ACTIONS(1849),
    [aux_sym_directive_token1] = ACTIONS(1847),
    [aux_sym_tag_token1] = ACTIONS(1849),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(979)] = {
    [sym_attribute] = STATE(979),
    [sym__lowercase_identifier] = ACTIONS(1745),
    [anon_sym_SEMI_SEMI] = ACTIONS(1747),
    [anon_sym_let] = ACTIONS(1745),
    [anon_sym_and] = ACTIONS(1745),
    [anon_sym_TILDE] = ACTIONS(1745),
    [anon_sym_QMARK] = ACTIONS(1745),
    [anon_sym_LPAREN] = ACTIONS(1747),
    [anon_sym_external] = ACTIONS(1745),
    [anon_sym_type] = ACTIONS(1745),
    [anon_sym_COMMA] = ACTIONS(1747),
    [anon_sym_PIPE] = ACTIONS(1745),
    [anon_sym_DOT] = ACTIONS(1747),
    [anon_sym_LBRACK] = ACTIONS(1745),
    [anon_sym_RBRACK] = ACTIONS(1747),
    [anon_sym_true] = ACTIONS(1745),
    [anon_sym_false] = ACTIONS(1745),
    [anon_sym_LBRACE] = ACTIONS(1745),
    [anon_sym_SEMI] = ACTIONS(1745),
    [anon_sym_exception] = ACTIONS(1745),
    [anon_sym_module] = ACTIONS(1745),
    [anon_sym_open] = ACTIONS(1745),
    [anon_sym_include] = ACTIONS(1745),
    [anon_sym_class] = ACTIONS(1745),
    [anon_sym_val] = ACTIONS(1745),
    [anon_sym_end] = ACTIONS(1745),
    [anon_sym_object] = ACTIONS(1745),
    [anon_sym_in] = ACTIONS(1745),
    [anon_sym_AMP] = ACTIONS(1745),
    [anon_sym_POUND] = ACTIONS(1745),
    [anon_sym_COLON_COLON] = ACTIONS(1747),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1747),
    [anon_sym_else] = ACTIONS(1745),
    [anon_sym_new] = ACTIONS(1745),
    [anon_sym_LBRACE_LT] = ACTIONS(1747),
    [anon_sym_begin] = ACTIONS(1745),
    [sym_ocamlyacc_value] = ACTIONS(1747),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1745),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1747),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1745),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1745),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1747),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1747),
    [aux_sym_number_token1] = ACTIONS(1747),
    [aux_sym_signed_number_token1] = ACTIONS(1745),
    [anon_sym_SQUOTE] = ACTIONS(1747),
    [anon_sym_DQUOTE] = ACTIONS(1747),
    [sym_prefix_operator] = ACTIONS(1747),
    [aux_sym_sign_operator_token1] = ACTIONS(1745),
    [sym_hash_operator] = ACTIONS(1747),
    [anon_sym_lsl] = ACTIONS(1745),
    [anon_sym_lsr] = ACTIONS(1745),
    [anon_sym_asr] = ACTIONS(1745),
    [aux_sym_pow_operator_token1] = ACTIONS(1747),
    [anon_sym_mod] = ACTIONS(1745),
    [anon_sym_land] = ACTIONS(1745),
    [anon_sym_lor] = ACTIONS(1745),
    [anon_sym_lxor] = ACTIONS(1745),
    [aux_sym_mult_operator_token1] = ACTIONS(1745),
    [aux_sym_add_operator_token1] = ACTIONS(1745),
    [sym_concat_operator] = ACTIONS(1747),
    [sym_rel_operator] = ACTIONS(1745),
    [anon_sym_AMP_AMP] = ACTIONS(1745),
    [anon_sym_or] = ACTIONS(1745),
    [anon_sym_PIPE_PIPE] = ACTIONS(1745),
    [sym_assign_operator] = ACTIONS(1747),
    [sym_let_operator] = ACTIONS(1747),
    [sym_let_and_operator] = ACTIONS(1747),
    [sym__uppercase_identifier] = ACTIONS(1747),
    [aux_sym_directive_token1] = ACTIONS(1745),
    [aux_sym_tag_token1] = ACTIONS(1747),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(980)] = {
    [sym_attribute] = STATE(980),
    [sym__lowercase_identifier] = ACTIONS(1967),
    [anon_sym_TILDE] = ACTIONS(1967),
    [anon_sym_QMARK] = ACTIONS(1967),
    [anon_sym_COLON] = ACTIONS(2119),
    [anon_sym_LPAREN] = ACTIONS(1971),
    [anon_sym_RPAREN] = ACTIONS(1971),
    [anon_sym_COMMA] = ACTIONS(1971),
    [anon_sym_PIPE] = ACTIONS(1967),
    [anon_sym_COLON2] = ACTIONS(1967),
    [anon_sym_DASH_GT] = ACTIONS(1967),
    [anon_sym_LBRACK] = ACTIONS(1967),
    [anon_sym_RBRACK] = ACTIONS(1971),
    [anon_sym_true] = ACTIONS(1967),
    [anon_sym_false] = ACTIONS(1967),
    [anon_sym_LBRACE] = ACTIONS(1967),
    [anon_sym_SEMI] = ACTIONS(1971),
    [anon_sym_RBRACE] = ACTIONS(1971),
    [anon_sym_constraint] = ACTIONS(1967),
    [anon_sym_val] = ACTIONS(1967),
    [anon_sym_end] = ACTIONS(1967),
    [anon_sym_with] = ACTIONS(1967),
    [anon_sym_COLON_GT] = ACTIONS(1971),
    [anon_sym_object] = ACTIONS(1967),
    [anon_sym_inherit] = ACTIONS(1967),
    [anon_sym_method] = ACTIONS(1967),
    [anon_sym_initializer] = ACTIONS(1967),
    [anon_sym_AMP] = ACTIONS(1967),
    [anon_sym_COLON_COLON] = ACTIONS(1971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1971),
    [anon_sym_then] = ACTIONS(1967),
    [anon_sym_else] = ACTIONS(1967),
    [anon_sym_do] = ACTIONS(1967),
    [anon_sym_done] = ACTIONS(1967),
    [anon_sym_to] = ACTIONS(1967),
    [anon_sym_downto] = ACTIONS(1967),
    [anon_sym_new] = ACTIONS(1967),
    [anon_sym_LBRACE_LT] = ACTIONS(1971),
    [anon_sym_GT_RBRACE] = ACTIONS(1971),
    [anon_sym_begin] = ACTIONS(1967),
    [sym_ocamlyacc_value] = ACTIONS(1971),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1967),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1971),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1967),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1967),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1971),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1971),
    [aux_sym_number_token1] = ACTIONS(1971),
    [aux_sym_signed_number_token1] = ACTIONS(1967),
    [anon_sym_SQUOTE] = ACTIONS(1971),
    [anon_sym_DQUOTE] = ACTIONS(1971),
    [sym_prefix_operator] = ACTIONS(1971),
    [aux_sym_sign_operator_token1] = ACTIONS(1967),
    [anon_sym_lsl] = ACTIONS(1967),
    [anon_sym_lsr] = ACTIONS(1967),
    [anon_sym_asr] = ACTIONS(1967),
    [aux_sym_pow_operator_token1] = ACTIONS(1971),
    [anon_sym_mod] = ACTIONS(1967),
    [anon_sym_land] = ACTIONS(1967),
    [anon_sym_lor] = ACTIONS(1967),
    [anon_sym_lxor] = ACTIONS(1967),
    [aux_sym_mult_operator_token1] = ACTIONS(1967),
    [aux_sym_add_operator_token1] = ACTIONS(1967),
    [sym_concat_operator] = ACTIONS(1971),
    [sym_rel_operator] = ACTIONS(1967),
    [anon_sym_AMP_AMP] = ACTIONS(1967),
    [anon_sym_or] = ACTIONS(1967),
    [anon_sym_PIPE_PIPE] = ACTIONS(1967),
    [sym_assign_operator] = ACTIONS(1971),
    [sym__uppercase_identifier] = ACTIONS(1971),
    [aux_sym_tag_token1] = ACTIONS(1971),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(981)] = {
    [sym_attribute] = STATE(981),
    [sym__lowercase_identifier] = ACTIONS(1567),
    [anon_sym_SEMI_SEMI] = ACTIONS(1569),
    [anon_sym_let] = ACTIONS(1567),
    [anon_sym_and] = ACTIONS(1567),
    [anon_sym_TILDE] = ACTIONS(1567),
    [anon_sym_QMARK] = ACTIONS(1567),
    [anon_sym_LPAREN] = ACTIONS(1569),
    [anon_sym_external] = ACTIONS(1567),
    [anon_sym_type] = ACTIONS(1567),
    [anon_sym_COMMA] = ACTIONS(1569),
    [anon_sym_PIPE] = ACTIONS(1567),
    [anon_sym_DOT] = ACTIONS(1569),
    [anon_sym_LBRACK] = ACTIONS(1567),
    [anon_sym_RBRACK] = ACTIONS(1569),
    [anon_sym_true] = ACTIONS(1567),
    [anon_sym_false] = ACTIONS(1567),
    [anon_sym_LBRACE] = ACTIONS(1567),
    [anon_sym_SEMI] = ACTIONS(1567),
    [anon_sym_exception] = ACTIONS(1567),
    [anon_sym_module] = ACTIONS(1567),
    [anon_sym_open] = ACTIONS(1567),
    [anon_sym_include] = ACTIONS(1567),
    [anon_sym_class] = ACTIONS(1567),
    [anon_sym_val] = ACTIONS(1567),
    [anon_sym_end] = ACTIONS(1567),
    [anon_sym_object] = ACTIONS(1567),
    [anon_sym_in] = ACTIONS(1567),
    [anon_sym_AMP] = ACTIONS(1567),
    [anon_sym_POUND] = ACTIONS(1567),
    [anon_sym_COLON_COLON] = ACTIONS(1569),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1569),
    [anon_sym_else] = ACTIONS(1567),
    [anon_sym_new] = ACTIONS(1567),
    [anon_sym_LBRACE_LT] = ACTIONS(1569),
    [anon_sym_begin] = ACTIONS(1567),
    [sym_ocamlyacc_value] = ACTIONS(1569),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1567),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1569),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1567),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1567),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1569),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1569),
    [aux_sym_number_token1] = ACTIONS(1569),
    [aux_sym_signed_number_token1] = ACTIONS(1567),
    [anon_sym_SQUOTE] = ACTIONS(1569),
    [anon_sym_DQUOTE] = ACTIONS(1569),
    [sym_prefix_operator] = ACTIONS(1569),
    [aux_sym_sign_operator_token1] = ACTIONS(1567),
    [sym_hash_operator] = ACTIONS(1569),
    [anon_sym_lsl] = ACTIONS(1567),
    [anon_sym_lsr] = ACTIONS(1567),
    [anon_sym_asr] = ACTIONS(1567),
    [aux_sym_pow_operator_token1] = ACTIONS(1569),
    [anon_sym_mod] = ACTIONS(1567),
    [anon_sym_land] = ACTIONS(1567),
    [anon_sym_lor] = ACTIONS(1567),
    [anon_sym_lxor] = ACTIONS(1567),
    [aux_sym_mult_operator_token1] = ACTIONS(1567),
    [aux_sym_add_operator_token1] = ACTIONS(1567),
    [sym_concat_operator] = ACTIONS(1569),
    [sym_rel_operator] = ACTIONS(1567),
    [anon_sym_AMP_AMP] = ACTIONS(1567),
    [anon_sym_or] = ACTIONS(1567),
    [anon_sym_PIPE_PIPE] = ACTIONS(1567),
    [sym_assign_operator] = ACTIONS(1569),
    [sym_let_operator] = ACTIONS(1569),
    [sym_let_and_operator] = ACTIONS(1569),
    [sym__uppercase_identifier] = ACTIONS(1569),
    [aux_sym_directive_token1] = ACTIONS(1567),
    [aux_sym_tag_token1] = ACTIONS(1569),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(982)] = {
    [sym_parameter] = STATE(2473),
    [sym__parenthesized_abstract_type] = STATE(2473),
    [sym__delimited_pattern] = STATE(2474),
    [sym__simple_pattern] = STATE(2446),
    [sym_typed_pattern] = STATE(2474),
    [sym_polymorphic_variant_pattern] = STATE(2474),
    [sym_record_pattern] = STATE(2428),
    [sym_list_pattern] = STATE(2428),
    [sym_array_pattern] = STATE(2428),
    [sym_local_open_pattern] = STATE(2474),
    [sym_package_pattern] = STATE(2474),
    [sym_parenthesized_pattern] = STATE(2428),
    [sym_attribute] = STATE(982),
    [sym_extension] = STATE(2474),
    [sym_quoted_extension] = STATE(2474),
    [sym__constant] = STATE(2426),
    [sym__signed_constant] = STATE(2474),
    [sym_number] = STATE(2434),
    [sym_signed_number] = STATE(2426),
    [sym_character] = STATE(2434),
    [sym_string] = STATE(2434),
    [sym_quoted_string] = STATE(2434),
    [sym_boolean] = STATE(2434),
    [sym_unit] = STATE(2428),
    [sym__simple_value_pattern] = STATE(2474),
    [sym_parenthesized_operator] = STATE(2474),
    [sym_module_path] = STATE(6496),
    [sym_constructor_path] = STATE(2474),
    [sym__constructor_name] = STATE(2476),
    [sym_tag] = STATE(2474),
    [aux_sym_let_binding_repeat1] = STATE(982),
    [sym__lowercase_identifier] = ACTIONS(2121),
    [anon_sym_SEMI_SEMI] = ACTIONS(2124),
    [anon_sym_let] = ACTIONS(2126),
    [anon_sym_and] = ACTIONS(2126),
    [anon_sym_EQ] = ACTIONS(2124),
    [anon_sym_TILDE] = ACTIONS(2128),
    [anon_sym_QMARK] = ACTIONS(2128),
    [anon_sym_LPAREN] = ACTIONS(2131),
    [anon_sym_external] = ACTIONS(2126),
    [anon_sym_type] = ACTIONS(2126),
    [anon_sym_COLON2] = ACTIONS(2124),
    [anon_sym_LBRACK] = ACTIONS(2134),
    [anon_sym_RBRACK] = ACTIONS(2124),
    [anon_sym_true] = ACTIONS(2137),
    [anon_sym_false] = ACTIONS(2137),
    [anon_sym_LBRACE] = ACTIONS(2140),
    [anon_sym_exception] = ACTIONS(2126),
    [anon_sym_module] = ACTIONS(2126),
    [anon_sym_open] = ACTIONS(2126),
    [anon_sym_include] = ACTIONS(2126),
    [anon_sym_class] = ACTIONS(2126),
    [anon_sym_val] = ACTIONS(2126),
    [anon_sym_end] = ACTIONS(2126),
    [anon_sym_POUND] = ACTIONS(2143),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2146),
    [anon_sym_begin] = ACTIONS(2149),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2126),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2124),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2152),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2155),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2124),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2124),
    [aux_sym_number_token1] = ACTIONS(2158),
    [aux_sym_signed_number_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2164),
    [anon_sym_DQUOTE] = ACTIONS(2167),
    [sym_let_operator] = ACTIONS(2124),
    [sym__uppercase_identifier] = ACTIONS(2170),
    [aux_sym_directive_token1] = ACTIONS(2126),
    [aux_sym_tag_token1] = ACTIONS(2173),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(983)] = {
    [sym_attribute] = STATE(983),
    [sym__lowercase_identifier] = ACTIONS(1655),
    [anon_sym_SEMI_SEMI] = ACTIONS(1657),
    [anon_sym_let] = ACTIONS(1655),
    [anon_sym_and] = ACTIONS(1655),
    [anon_sym_TILDE] = ACTIONS(1655),
    [anon_sym_QMARK] = ACTIONS(1655),
    [anon_sym_LPAREN] = ACTIONS(1657),
    [anon_sym_external] = ACTIONS(1655),
    [anon_sym_type] = ACTIONS(1655),
    [anon_sym_COMMA] = ACTIONS(1657),
    [anon_sym_PIPE] = ACTIONS(1655),
    [anon_sym_DOT] = ACTIONS(1657),
    [anon_sym_LBRACK] = ACTIONS(1655),
    [anon_sym_RBRACK] = ACTIONS(1657),
    [anon_sym_true] = ACTIONS(1655),
    [anon_sym_false] = ACTIONS(1655),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_SEMI] = ACTIONS(1655),
    [anon_sym_exception] = ACTIONS(1655),
    [anon_sym_module] = ACTIONS(1655),
    [anon_sym_open] = ACTIONS(1655),
    [anon_sym_include] = ACTIONS(1655),
    [anon_sym_class] = ACTIONS(1655),
    [anon_sym_val] = ACTIONS(1655),
    [anon_sym_end] = ACTIONS(1655),
    [anon_sym_object] = ACTIONS(1655),
    [anon_sym_in] = ACTIONS(1655),
    [anon_sym_AMP] = ACTIONS(1655),
    [anon_sym_POUND] = ACTIONS(1655),
    [anon_sym_COLON_COLON] = ACTIONS(1657),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1657),
    [anon_sym_else] = ACTIONS(1655),
    [anon_sym_new] = ACTIONS(1655),
    [anon_sym_LBRACE_LT] = ACTIONS(1657),
    [anon_sym_begin] = ACTIONS(1655),
    [sym_ocamlyacc_value] = ACTIONS(1657),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1655),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1657),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1655),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1655),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1657),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1657),
    [aux_sym_number_token1] = ACTIONS(1657),
    [aux_sym_signed_number_token1] = ACTIONS(1655),
    [anon_sym_SQUOTE] = ACTIONS(1657),
    [anon_sym_DQUOTE] = ACTIONS(1657),
    [sym_prefix_operator] = ACTIONS(1657),
    [aux_sym_sign_operator_token1] = ACTIONS(1655),
    [sym_hash_operator] = ACTIONS(1657),
    [anon_sym_lsl] = ACTIONS(1655),
    [anon_sym_lsr] = ACTIONS(1655),
    [anon_sym_asr] = ACTIONS(1655),
    [aux_sym_pow_operator_token1] = ACTIONS(1657),
    [anon_sym_mod] = ACTIONS(1655),
    [anon_sym_land] = ACTIONS(1655),
    [anon_sym_lor] = ACTIONS(1655),
    [anon_sym_lxor] = ACTIONS(1655),
    [aux_sym_mult_operator_token1] = ACTIONS(1655),
    [aux_sym_add_operator_token1] = ACTIONS(1655),
    [sym_concat_operator] = ACTIONS(1657),
    [sym_rel_operator] = ACTIONS(1655),
    [anon_sym_AMP_AMP] = ACTIONS(1655),
    [anon_sym_or] = ACTIONS(1655),
    [anon_sym_PIPE_PIPE] = ACTIONS(1655),
    [sym_assign_operator] = ACTIONS(1657),
    [sym_let_operator] = ACTIONS(1657),
    [sym_let_and_operator] = ACTIONS(1657),
    [sym__uppercase_identifier] = ACTIONS(1657),
    [aux_sym_directive_token1] = ACTIONS(1655),
    [aux_sym_tag_token1] = ACTIONS(1657),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(984)] = {
    [sym_attribute] = STATE(984),
    [sym__lowercase_identifier] = ACTIONS(1699),
    [anon_sym_SEMI_SEMI] = ACTIONS(1701),
    [anon_sym_let] = ACTIONS(1699),
    [anon_sym_and] = ACTIONS(1699),
    [anon_sym_TILDE] = ACTIONS(1699),
    [anon_sym_QMARK] = ACTIONS(1699),
    [anon_sym_LPAREN] = ACTIONS(1701),
    [anon_sym_external] = ACTIONS(1699),
    [anon_sym_type] = ACTIONS(1699),
    [anon_sym_COMMA] = ACTIONS(1701),
    [anon_sym_PIPE] = ACTIONS(1699),
    [anon_sym_DOT] = ACTIONS(1841),
    [anon_sym_LBRACK] = ACTIONS(1699),
    [anon_sym_RBRACK] = ACTIONS(1701),
    [anon_sym_true] = ACTIONS(1699),
    [anon_sym_false] = ACTIONS(1699),
    [anon_sym_LBRACE] = ACTIONS(1699),
    [anon_sym_SEMI] = ACTIONS(1699),
    [anon_sym_exception] = ACTIONS(1699),
    [anon_sym_module] = ACTIONS(1699),
    [anon_sym_open] = ACTIONS(1699),
    [anon_sym_include] = ACTIONS(1699),
    [anon_sym_class] = ACTIONS(1699),
    [anon_sym_val] = ACTIONS(1699),
    [anon_sym_end] = ACTIONS(1699),
    [anon_sym_object] = ACTIONS(1699),
    [anon_sym_in] = ACTIONS(1699),
    [anon_sym_AMP] = ACTIONS(1699),
    [anon_sym_POUND] = ACTIONS(1699),
    [anon_sym_COLON_COLON] = ACTIONS(1701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1701),
    [anon_sym_else] = ACTIONS(1699),
    [anon_sym_new] = ACTIONS(1699),
    [anon_sym_LBRACE_LT] = ACTIONS(1701),
    [anon_sym_begin] = ACTIONS(1699),
    [sym_ocamlyacc_value] = ACTIONS(1701),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1699),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1701),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1699),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1699),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1701),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1701),
    [aux_sym_number_token1] = ACTIONS(1701),
    [aux_sym_signed_number_token1] = ACTIONS(1699),
    [anon_sym_SQUOTE] = ACTIONS(1701),
    [anon_sym_DQUOTE] = ACTIONS(1701),
    [sym_prefix_operator] = ACTIONS(1701),
    [aux_sym_sign_operator_token1] = ACTIONS(1699),
    [sym_hash_operator] = ACTIONS(1701),
    [anon_sym_lsl] = ACTIONS(1699),
    [anon_sym_lsr] = ACTIONS(1699),
    [anon_sym_asr] = ACTIONS(1699),
    [aux_sym_pow_operator_token1] = ACTIONS(1701),
    [anon_sym_mod] = ACTIONS(1699),
    [anon_sym_land] = ACTIONS(1699),
    [anon_sym_lor] = ACTIONS(1699),
    [anon_sym_lxor] = ACTIONS(1699),
    [aux_sym_mult_operator_token1] = ACTIONS(1699),
    [aux_sym_add_operator_token1] = ACTIONS(1699),
    [sym_concat_operator] = ACTIONS(1701),
    [sym_rel_operator] = ACTIONS(1699),
    [anon_sym_AMP_AMP] = ACTIONS(1699),
    [anon_sym_or] = ACTIONS(1699),
    [anon_sym_PIPE_PIPE] = ACTIONS(1699),
    [sym_assign_operator] = ACTIONS(1701),
    [sym_let_operator] = ACTIONS(1701),
    [sym_let_and_operator] = ACTIONS(1701),
    [sym__uppercase_identifier] = ACTIONS(1701),
    [aux_sym_directive_token1] = ACTIONS(1699),
    [aux_sym_tag_token1] = ACTIONS(1701),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(985)] = {
    [sym_attribute] = STATE(985),
    [sym__lowercase_identifier] = ACTIONS(1843),
    [anon_sym_SEMI_SEMI] = ACTIONS(1845),
    [anon_sym_let] = ACTIONS(1843),
    [anon_sym_and] = ACTIONS(1843),
    [anon_sym_TILDE] = ACTIONS(1843),
    [anon_sym_QMARK] = ACTIONS(1843),
    [anon_sym_LPAREN] = ACTIONS(1845),
    [anon_sym_external] = ACTIONS(1843),
    [anon_sym_type] = ACTIONS(1843),
    [anon_sym_COMMA] = ACTIONS(1845),
    [anon_sym_PIPE] = ACTIONS(1843),
    [anon_sym_DOT] = ACTIONS(1845),
    [anon_sym_LBRACK] = ACTIONS(1843),
    [anon_sym_RBRACK] = ACTIONS(1845),
    [anon_sym_true] = ACTIONS(1843),
    [anon_sym_false] = ACTIONS(1843),
    [anon_sym_LBRACE] = ACTIONS(1843),
    [anon_sym_SEMI] = ACTIONS(1843),
    [anon_sym_exception] = ACTIONS(1843),
    [anon_sym_module] = ACTIONS(1843),
    [anon_sym_open] = ACTIONS(1843),
    [anon_sym_include] = ACTIONS(1843),
    [anon_sym_class] = ACTIONS(1843),
    [anon_sym_val] = ACTIONS(1843),
    [anon_sym_end] = ACTIONS(1843),
    [anon_sym_object] = ACTIONS(1843),
    [anon_sym_in] = ACTIONS(1843),
    [anon_sym_AMP] = ACTIONS(1843),
    [anon_sym_POUND] = ACTIONS(1843),
    [anon_sym_COLON_COLON] = ACTIONS(1845),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1845),
    [anon_sym_else] = ACTIONS(1843),
    [anon_sym_new] = ACTIONS(1843),
    [anon_sym_LBRACE_LT] = ACTIONS(1845),
    [anon_sym_begin] = ACTIONS(1843),
    [sym_ocamlyacc_value] = ACTIONS(1845),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1843),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1845),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1843),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1843),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1845),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1845),
    [aux_sym_number_token1] = ACTIONS(1845),
    [aux_sym_signed_number_token1] = ACTIONS(1843),
    [anon_sym_SQUOTE] = ACTIONS(1845),
    [anon_sym_DQUOTE] = ACTIONS(1845),
    [sym_prefix_operator] = ACTIONS(1845),
    [aux_sym_sign_operator_token1] = ACTIONS(1843),
    [sym_hash_operator] = ACTIONS(1845),
    [anon_sym_lsl] = ACTIONS(1843),
    [anon_sym_lsr] = ACTIONS(1843),
    [anon_sym_asr] = ACTIONS(1843),
    [aux_sym_pow_operator_token1] = ACTIONS(1845),
    [anon_sym_mod] = ACTIONS(1843),
    [anon_sym_land] = ACTIONS(1843),
    [anon_sym_lor] = ACTIONS(1843),
    [anon_sym_lxor] = ACTIONS(1843),
    [aux_sym_mult_operator_token1] = ACTIONS(1843),
    [aux_sym_add_operator_token1] = ACTIONS(1843),
    [sym_concat_operator] = ACTIONS(1845),
    [sym_rel_operator] = ACTIONS(1843),
    [anon_sym_AMP_AMP] = ACTIONS(1843),
    [anon_sym_or] = ACTIONS(1843),
    [anon_sym_PIPE_PIPE] = ACTIONS(1843),
    [sym_assign_operator] = ACTIONS(1845),
    [sym_let_operator] = ACTIONS(1845),
    [sym_let_and_operator] = ACTIONS(1845),
    [sym__uppercase_identifier] = ACTIONS(1845),
    [aux_sym_directive_token1] = ACTIONS(1843),
    [aux_sym_tag_token1] = ACTIONS(1845),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(986)] = {
    [sym_attribute] = STATE(986),
    [sym__lowercase_identifier] = ACTIONS(1663),
    [anon_sym_SEMI_SEMI] = ACTIONS(1665),
    [anon_sym_let] = ACTIONS(1663),
    [anon_sym_and] = ACTIONS(1663),
    [anon_sym_TILDE] = ACTIONS(1663),
    [anon_sym_QMARK] = ACTIONS(1663),
    [anon_sym_LPAREN] = ACTIONS(1665),
    [anon_sym_external] = ACTIONS(1663),
    [anon_sym_type] = ACTIONS(1663),
    [anon_sym_COMMA] = ACTIONS(1665),
    [anon_sym_PIPE] = ACTIONS(1663),
    [anon_sym_DOT] = ACTIONS(835),
    [anon_sym_LBRACK] = ACTIONS(1663),
    [anon_sym_RBRACK] = ACTIONS(1665),
    [anon_sym_true] = ACTIONS(1663),
    [anon_sym_false] = ACTIONS(1663),
    [anon_sym_LBRACE] = ACTIONS(1663),
    [anon_sym_SEMI] = ACTIONS(1663),
    [anon_sym_exception] = ACTIONS(1663),
    [anon_sym_module] = ACTIONS(1663),
    [anon_sym_open] = ACTIONS(1663),
    [anon_sym_include] = ACTIONS(1663),
    [anon_sym_class] = ACTIONS(1663),
    [anon_sym_val] = ACTIONS(1663),
    [anon_sym_end] = ACTIONS(1663),
    [anon_sym_object] = ACTIONS(1663),
    [anon_sym_in] = ACTIONS(1663),
    [anon_sym_AMP] = ACTIONS(1663),
    [anon_sym_POUND] = ACTIONS(837),
    [anon_sym_COLON_COLON] = ACTIONS(1665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1665),
    [anon_sym_else] = ACTIONS(1663),
    [anon_sym_new] = ACTIONS(1663),
    [anon_sym_LBRACE_LT] = ACTIONS(1665),
    [anon_sym_begin] = ACTIONS(1663),
    [sym_ocamlyacc_value] = ACTIONS(1665),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1663),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1665),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1663),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1663),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1665),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1665),
    [aux_sym_number_token1] = ACTIONS(1665),
    [aux_sym_signed_number_token1] = ACTIONS(1663),
    [anon_sym_SQUOTE] = ACTIONS(1665),
    [anon_sym_DQUOTE] = ACTIONS(1665),
    [sym_prefix_operator] = ACTIONS(1665),
    [aux_sym_sign_operator_token1] = ACTIONS(1663),
    [sym_hash_operator] = ACTIONS(839),
    [anon_sym_lsl] = ACTIONS(1663),
    [anon_sym_lsr] = ACTIONS(1663),
    [anon_sym_asr] = ACTIONS(1663),
    [aux_sym_pow_operator_token1] = ACTIONS(1665),
    [anon_sym_mod] = ACTIONS(1663),
    [anon_sym_land] = ACTIONS(1663),
    [anon_sym_lor] = ACTIONS(1663),
    [anon_sym_lxor] = ACTIONS(1663),
    [aux_sym_mult_operator_token1] = ACTIONS(1663),
    [aux_sym_add_operator_token1] = ACTIONS(1663),
    [sym_concat_operator] = ACTIONS(1665),
    [sym_rel_operator] = ACTIONS(1663),
    [anon_sym_AMP_AMP] = ACTIONS(1663),
    [anon_sym_or] = ACTIONS(1663),
    [anon_sym_PIPE_PIPE] = ACTIONS(1663),
    [sym_assign_operator] = ACTIONS(1665),
    [sym_let_operator] = ACTIONS(1665),
    [sym_let_and_operator] = ACTIONS(1665),
    [sym__uppercase_identifier] = ACTIONS(1665),
    [aux_sym_directive_token1] = ACTIONS(1663),
    [aux_sym_tag_token1] = ACTIONS(1665),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(987)] = {
    [sym_attribute] = STATE(987),
    [sym__lowercase_identifier] = ACTIONS(1705),
    [anon_sym_SEMI_SEMI] = ACTIONS(1707),
    [anon_sym_let] = ACTIONS(1705),
    [anon_sym_and] = ACTIONS(1705),
    [anon_sym_TILDE] = ACTIONS(1705),
    [anon_sym_QMARK] = ACTIONS(1705),
    [anon_sym_LPAREN] = ACTIONS(1707),
    [anon_sym_external] = ACTIONS(1705),
    [anon_sym_type] = ACTIONS(1705),
    [anon_sym_COMMA] = ACTIONS(1707),
    [anon_sym_PIPE] = ACTIONS(1705),
    [anon_sym_DOT] = ACTIONS(1707),
    [anon_sym_LBRACK] = ACTIONS(1705),
    [anon_sym_RBRACK] = ACTIONS(1707),
    [anon_sym_true] = ACTIONS(1705),
    [anon_sym_false] = ACTIONS(1705),
    [anon_sym_LBRACE] = ACTIONS(1705),
    [anon_sym_SEMI] = ACTIONS(1705),
    [anon_sym_exception] = ACTIONS(1705),
    [anon_sym_module] = ACTIONS(1705),
    [anon_sym_open] = ACTIONS(1705),
    [anon_sym_include] = ACTIONS(1705),
    [anon_sym_class] = ACTIONS(1705),
    [anon_sym_val] = ACTIONS(1705),
    [anon_sym_end] = ACTIONS(1705),
    [anon_sym_object] = ACTIONS(1705),
    [anon_sym_in] = ACTIONS(1705),
    [anon_sym_AMP] = ACTIONS(1705),
    [anon_sym_POUND] = ACTIONS(1705),
    [anon_sym_COLON_COLON] = ACTIONS(1707),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1707),
    [anon_sym_else] = ACTIONS(1705),
    [anon_sym_new] = ACTIONS(1705),
    [anon_sym_LBRACE_LT] = ACTIONS(1707),
    [anon_sym_begin] = ACTIONS(1705),
    [sym_ocamlyacc_value] = ACTIONS(1707),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1705),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1707),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1705),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1705),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1707),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1707),
    [aux_sym_number_token1] = ACTIONS(1707),
    [aux_sym_signed_number_token1] = ACTIONS(1705),
    [anon_sym_SQUOTE] = ACTIONS(1707),
    [anon_sym_DQUOTE] = ACTIONS(1707),
    [sym_prefix_operator] = ACTIONS(1707),
    [aux_sym_sign_operator_token1] = ACTIONS(1705),
    [sym_hash_operator] = ACTIONS(1707),
    [anon_sym_lsl] = ACTIONS(1705),
    [anon_sym_lsr] = ACTIONS(1705),
    [anon_sym_asr] = ACTIONS(1705),
    [aux_sym_pow_operator_token1] = ACTIONS(1707),
    [anon_sym_mod] = ACTIONS(1705),
    [anon_sym_land] = ACTIONS(1705),
    [anon_sym_lor] = ACTIONS(1705),
    [anon_sym_lxor] = ACTIONS(1705),
    [aux_sym_mult_operator_token1] = ACTIONS(1705),
    [aux_sym_add_operator_token1] = ACTIONS(1705),
    [sym_concat_operator] = ACTIONS(1707),
    [sym_rel_operator] = ACTIONS(1705),
    [anon_sym_AMP_AMP] = ACTIONS(1705),
    [anon_sym_or] = ACTIONS(1705),
    [anon_sym_PIPE_PIPE] = ACTIONS(1705),
    [sym_assign_operator] = ACTIONS(1707),
    [sym_let_operator] = ACTIONS(1707),
    [sym_let_and_operator] = ACTIONS(1707),
    [sym__uppercase_identifier] = ACTIONS(1707),
    [aux_sym_directive_token1] = ACTIONS(1705),
    [aux_sym_tag_token1] = ACTIONS(1707),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(988)] = {
    [sym_attribute] = STATE(988),
    [sym__lowercase_identifier] = ACTIONS(1667),
    [anon_sym_SEMI_SEMI] = ACTIONS(1669),
    [anon_sym_let] = ACTIONS(1667),
    [anon_sym_and] = ACTIONS(1667),
    [anon_sym_TILDE] = ACTIONS(1667),
    [anon_sym_QMARK] = ACTIONS(1667),
    [anon_sym_LPAREN] = ACTIONS(1669),
    [anon_sym_external] = ACTIONS(1667),
    [anon_sym_type] = ACTIONS(1667),
    [anon_sym_COMMA] = ACTIONS(1669),
    [anon_sym_PIPE] = ACTIONS(1667),
    [anon_sym_DOT] = ACTIONS(1669),
    [anon_sym_LBRACK] = ACTIONS(1667),
    [anon_sym_RBRACK] = ACTIONS(1669),
    [anon_sym_true] = ACTIONS(1667),
    [anon_sym_false] = ACTIONS(1667),
    [anon_sym_LBRACE] = ACTIONS(1667),
    [anon_sym_SEMI] = ACTIONS(1667),
    [anon_sym_exception] = ACTIONS(1667),
    [anon_sym_module] = ACTIONS(1667),
    [anon_sym_open] = ACTIONS(1667),
    [anon_sym_include] = ACTIONS(1667),
    [anon_sym_class] = ACTIONS(1667),
    [anon_sym_val] = ACTIONS(1667),
    [anon_sym_end] = ACTIONS(1667),
    [anon_sym_object] = ACTIONS(1667),
    [anon_sym_in] = ACTIONS(1667),
    [anon_sym_AMP] = ACTIONS(1667),
    [anon_sym_POUND] = ACTIONS(1667),
    [anon_sym_COLON_COLON] = ACTIONS(1669),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1669),
    [anon_sym_else] = ACTIONS(1667),
    [anon_sym_new] = ACTIONS(1667),
    [anon_sym_LBRACE_LT] = ACTIONS(1669),
    [anon_sym_begin] = ACTIONS(1667),
    [sym_ocamlyacc_value] = ACTIONS(1669),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1667),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1669),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1667),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1667),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1669),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1669),
    [aux_sym_number_token1] = ACTIONS(1669),
    [aux_sym_signed_number_token1] = ACTIONS(1667),
    [anon_sym_SQUOTE] = ACTIONS(1669),
    [anon_sym_DQUOTE] = ACTIONS(1669),
    [sym_prefix_operator] = ACTIONS(1669),
    [aux_sym_sign_operator_token1] = ACTIONS(1667),
    [sym_hash_operator] = ACTIONS(1669),
    [anon_sym_lsl] = ACTIONS(1667),
    [anon_sym_lsr] = ACTIONS(1667),
    [anon_sym_asr] = ACTIONS(1667),
    [aux_sym_pow_operator_token1] = ACTIONS(1669),
    [anon_sym_mod] = ACTIONS(1667),
    [anon_sym_land] = ACTIONS(1667),
    [anon_sym_lor] = ACTIONS(1667),
    [anon_sym_lxor] = ACTIONS(1667),
    [aux_sym_mult_operator_token1] = ACTIONS(1667),
    [aux_sym_add_operator_token1] = ACTIONS(1667),
    [sym_concat_operator] = ACTIONS(1669),
    [sym_rel_operator] = ACTIONS(1667),
    [anon_sym_AMP_AMP] = ACTIONS(1667),
    [anon_sym_or] = ACTIONS(1667),
    [anon_sym_PIPE_PIPE] = ACTIONS(1667),
    [sym_assign_operator] = ACTIONS(1669),
    [sym_let_operator] = ACTIONS(1669),
    [sym_let_and_operator] = ACTIONS(1669),
    [sym__uppercase_identifier] = ACTIONS(1669),
    [aux_sym_directive_token1] = ACTIONS(1667),
    [aux_sym_tag_token1] = ACTIONS(1669),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(989)] = {
    [sym_attribute] = STATE(989),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_SEMI_SEMI] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(1757),
    [anon_sym_and] = ACTIONS(1757),
    [anon_sym_TILDE] = ACTIONS(1757),
    [anon_sym_QMARK] = ACTIONS(1757),
    [anon_sym_LPAREN] = ACTIONS(1759),
    [anon_sym_external] = ACTIONS(1757),
    [anon_sym_type] = ACTIONS(1757),
    [anon_sym_COMMA] = ACTIONS(1759),
    [anon_sym_PIPE] = ACTIONS(1757),
    [anon_sym_DOT] = ACTIONS(1759),
    [anon_sym_LBRACK] = ACTIONS(1757),
    [anon_sym_RBRACK] = ACTIONS(1759),
    [anon_sym_true] = ACTIONS(1757),
    [anon_sym_false] = ACTIONS(1757),
    [anon_sym_LBRACE] = ACTIONS(1757),
    [anon_sym_SEMI] = ACTIONS(1757),
    [anon_sym_exception] = ACTIONS(1757),
    [anon_sym_module] = ACTIONS(1757),
    [anon_sym_open] = ACTIONS(1757),
    [anon_sym_include] = ACTIONS(1757),
    [anon_sym_class] = ACTIONS(1757),
    [anon_sym_val] = ACTIONS(1757),
    [anon_sym_end] = ACTIONS(1757),
    [anon_sym_object] = ACTIONS(1757),
    [anon_sym_in] = ACTIONS(1757),
    [anon_sym_AMP] = ACTIONS(1757),
    [anon_sym_POUND] = ACTIONS(1757),
    [anon_sym_COLON_COLON] = ACTIONS(1759),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1759),
    [anon_sym_else] = ACTIONS(1757),
    [anon_sym_new] = ACTIONS(1757),
    [anon_sym_LBRACE_LT] = ACTIONS(1759),
    [anon_sym_begin] = ACTIONS(1757),
    [sym_ocamlyacc_value] = ACTIONS(1759),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1757),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1759),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1757),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1757),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1759),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1759),
    [aux_sym_number_token1] = ACTIONS(1759),
    [aux_sym_signed_number_token1] = ACTIONS(1757),
    [anon_sym_SQUOTE] = ACTIONS(1759),
    [anon_sym_DQUOTE] = ACTIONS(1759),
    [sym_prefix_operator] = ACTIONS(1759),
    [aux_sym_sign_operator_token1] = ACTIONS(1757),
    [sym_hash_operator] = ACTIONS(1759),
    [anon_sym_lsl] = ACTIONS(1757),
    [anon_sym_lsr] = ACTIONS(1757),
    [anon_sym_asr] = ACTIONS(1757),
    [aux_sym_pow_operator_token1] = ACTIONS(1759),
    [anon_sym_mod] = ACTIONS(1757),
    [anon_sym_land] = ACTIONS(1757),
    [anon_sym_lor] = ACTIONS(1757),
    [anon_sym_lxor] = ACTIONS(1757),
    [aux_sym_mult_operator_token1] = ACTIONS(1757),
    [aux_sym_add_operator_token1] = ACTIONS(1757),
    [sym_concat_operator] = ACTIONS(1759),
    [sym_rel_operator] = ACTIONS(1757),
    [anon_sym_AMP_AMP] = ACTIONS(1757),
    [anon_sym_or] = ACTIONS(1757),
    [anon_sym_PIPE_PIPE] = ACTIONS(1757),
    [sym_assign_operator] = ACTIONS(1759),
    [sym_let_operator] = ACTIONS(1759),
    [sym_let_and_operator] = ACTIONS(1759),
    [sym__uppercase_identifier] = ACTIONS(1759),
    [aux_sym_directive_token1] = ACTIONS(1757),
    [aux_sym_tag_token1] = ACTIONS(1759),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(990)] = {
    [sym_attribute] = STATE(990),
    [sym__lowercase_identifier] = ACTIONS(1729),
    [anon_sym_SEMI_SEMI] = ACTIONS(1731),
    [anon_sym_let] = ACTIONS(1729),
    [anon_sym_and] = ACTIONS(1729),
    [anon_sym_TILDE] = ACTIONS(1729),
    [anon_sym_QMARK] = ACTIONS(1729),
    [anon_sym_LPAREN] = ACTIONS(1731),
    [anon_sym_external] = ACTIONS(1729),
    [anon_sym_type] = ACTIONS(1729),
    [anon_sym_COMMA] = ACTIONS(1731),
    [anon_sym_PIPE] = ACTIONS(1729),
    [anon_sym_DOT] = ACTIONS(1731),
    [anon_sym_LBRACK] = ACTIONS(1729),
    [anon_sym_RBRACK] = ACTIONS(1731),
    [anon_sym_true] = ACTIONS(1729),
    [anon_sym_false] = ACTIONS(1729),
    [anon_sym_LBRACE] = ACTIONS(1729),
    [anon_sym_SEMI] = ACTIONS(1729),
    [anon_sym_exception] = ACTIONS(1729),
    [anon_sym_module] = ACTIONS(1729),
    [anon_sym_open] = ACTIONS(1729),
    [anon_sym_include] = ACTIONS(1729),
    [anon_sym_class] = ACTIONS(1729),
    [anon_sym_val] = ACTIONS(1729),
    [anon_sym_end] = ACTIONS(1729),
    [anon_sym_object] = ACTIONS(1729),
    [anon_sym_in] = ACTIONS(1729),
    [anon_sym_AMP] = ACTIONS(1729),
    [anon_sym_POUND] = ACTIONS(1729),
    [anon_sym_COLON_COLON] = ACTIONS(1731),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1731),
    [anon_sym_else] = ACTIONS(1729),
    [anon_sym_new] = ACTIONS(1729),
    [anon_sym_LBRACE_LT] = ACTIONS(1731),
    [anon_sym_begin] = ACTIONS(1729),
    [sym_ocamlyacc_value] = ACTIONS(1731),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1729),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1731),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1729),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1729),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1731),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1731),
    [aux_sym_number_token1] = ACTIONS(1731),
    [aux_sym_signed_number_token1] = ACTIONS(1729),
    [anon_sym_SQUOTE] = ACTIONS(1731),
    [anon_sym_DQUOTE] = ACTIONS(1731),
    [sym_prefix_operator] = ACTIONS(1731),
    [aux_sym_sign_operator_token1] = ACTIONS(1729),
    [sym_hash_operator] = ACTIONS(1731),
    [anon_sym_lsl] = ACTIONS(1729),
    [anon_sym_lsr] = ACTIONS(1729),
    [anon_sym_asr] = ACTIONS(1729),
    [aux_sym_pow_operator_token1] = ACTIONS(1731),
    [anon_sym_mod] = ACTIONS(1729),
    [anon_sym_land] = ACTIONS(1729),
    [anon_sym_lor] = ACTIONS(1729),
    [anon_sym_lxor] = ACTIONS(1729),
    [aux_sym_mult_operator_token1] = ACTIONS(1729),
    [aux_sym_add_operator_token1] = ACTIONS(1729),
    [sym_concat_operator] = ACTIONS(1731),
    [sym_rel_operator] = ACTIONS(1729),
    [anon_sym_AMP_AMP] = ACTIONS(1729),
    [anon_sym_or] = ACTIONS(1729),
    [anon_sym_PIPE_PIPE] = ACTIONS(1729),
    [sym_assign_operator] = ACTIONS(1731),
    [sym_let_operator] = ACTIONS(1731),
    [sym_let_and_operator] = ACTIONS(1731),
    [sym__uppercase_identifier] = ACTIONS(1731),
    [aux_sym_directive_token1] = ACTIONS(1729),
    [aux_sym_tag_token1] = ACTIONS(1731),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(991)] = {
    [sym_attribute] = STATE(991),
    [sym__lowercase_identifier] = ACTIONS(1737),
    [anon_sym_SEMI_SEMI] = ACTIONS(1739),
    [anon_sym_let] = ACTIONS(1737),
    [anon_sym_and] = ACTIONS(1737),
    [anon_sym_TILDE] = ACTIONS(1737),
    [anon_sym_QMARK] = ACTIONS(1737),
    [anon_sym_LPAREN] = ACTIONS(1739),
    [anon_sym_external] = ACTIONS(1737),
    [anon_sym_type] = ACTIONS(1737),
    [anon_sym_COMMA] = ACTIONS(1739),
    [anon_sym_PIPE] = ACTIONS(1737),
    [anon_sym_DOT] = ACTIONS(1739),
    [anon_sym_LBRACK] = ACTIONS(1737),
    [anon_sym_RBRACK] = ACTIONS(1739),
    [anon_sym_true] = ACTIONS(1737),
    [anon_sym_false] = ACTIONS(1737),
    [anon_sym_LBRACE] = ACTIONS(1737),
    [anon_sym_SEMI] = ACTIONS(1737),
    [anon_sym_exception] = ACTIONS(1737),
    [anon_sym_module] = ACTIONS(1737),
    [anon_sym_open] = ACTIONS(1737),
    [anon_sym_include] = ACTIONS(1737),
    [anon_sym_class] = ACTIONS(1737),
    [anon_sym_val] = ACTIONS(1737),
    [anon_sym_end] = ACTIONS(1737),
    [anon_sym_object] = ACTIONS(1737),
    [anon_sym_in] = ACTIONS(1737),
    [anon_sym_AMP] = ACTIONS(1737),
    [anon_sym_POUND] = ACTIONS(1737),
    [anon_sym_COLON_COLON] = ACTIONS(1739),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1739),
    [anon_sym_else] = ACTIONS(1737),
    [anon_sym_new] = ACTIONS(1737),
    [anon_sym_LBRACE_LT] = ACTIONS(1739),
    [anon_sym_begin] = ACTIONS(1737),
    [sym_ocamlyacc_value] = ACTIONS(1739),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1737),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1739),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1737),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1737),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1739),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1739),
    [aux_sym_number_token1] = ACTIONS(1739),
    [aux_sym_signed_number_token1] = ACTIONS(1737),
    [anon_sym_SQUOTE] = ACTIONS(1739),
    [anon_sym_DQUOTE] = ACTIONS(1739),
    [sym_prefix_operator] = ACTIONS(1739),
    [aux_sym_sign_operator_token1] = ACTIONS(1737),
    [sym_hash_operator] = ACTIONS(1739),
    [anon_sym_lsl] = ACTIONS(1737),
    [anon_sym_lsr] = ACTIONS(1737),
    [anon_sym_asr] = ACTIONS(1737),
    [aux_sym_pow_operator_token1] = ACTIONS(1739),
    [anon_sym_mod] = ACTIONS(1737),
    [anon_sym_land] = ACTIONS(1737),
    [anon_sym_lor] = ACTIONS(1737),
    [anon_sym_lxor] = ACTIONS(1737),
    [aux_sym_mult_operator_token1] = ACTIONS(1737),
    [aux_sym_add_operator_token1] = ACTIONS(1737),
    [sym_concat_operator] = ACTIONS(1739),
    [sym_rel_operator] = ACTIONS(1737),
    [anon_sym_AMP_AMP] = ACTIONS(1737),
    [anon_sym_or] = ACTIONS(1737),
    [anon_sym_PIPE_PIPE] = ACTIONS(1737),
    [sym_assign_operator] = ACTIONS(1739),
    [sym_let_operator] = ACTIONS(1739),
    [sym_let_and_operator] = ACTIONS(1739),
    [sym__uppercase_identifier] = ACTIONS(1739),
    [aux_sym_directive_token1] = ACTIONS(1737),
    [aux_sym_tag_token1] = ACTIONS(1739),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(992)] = {
    [sym_attribute] = STATE(992),
    [sym__lowercase_identifier] = ACTIONS(1821),
    [anon_sym_SEMI_SEMI] = ACTIONS(1823),
    [anon_sym_let] = ACTIONS(1821),
    [anon_sym_and] = ACTIONS(1821),
    [anon_sym_TILDE] = ACTIONS(1821),
    [anon_sym_QMARK] = ACTIONS(1821),
    [anon_sym_LPAREN] = ACTIONS(1823),
    [anon_sym_external] = ACTIONS(1821),
    [anon_sym_type] = ACTIONS(1821),
    [anon_sym_COMMA] = ACTIONS(1823),
    [anon_sym_PIPE] = ACTIONS(1821),
    [anon_sym_DOT] = ACTIONS(1823),
    [anon_sym_LBRACK] = ACTIONS(1821),
    [anon_sym_RBRACK] = ACTIONS(1823),
    [anon_sym_true] = ACTIONS(1821),
    [anon_sym_false] = ACTIONS(1821),
    [anon_sym_LBRACE] = ACTIONS(1821),
    [anon_sym_SEMI] = ACTIONS(1821),
    [anon_sym_exception] = ACTIONS(1821),
    [anon_sym_module] = ACTIONS(1821),
    [anon_sym_open] = ACTIONS(1821),
    [anon_sym_include] = ACTIONS(1821),
    [anon_sym_class] = ACTIONS(1821),
    [anon_sym_val] = ACTIONS(1821),
    [anon_sym_end] = ACTIONS(1821),
    [anon_sym_object] = ACTIONS(1821),
    [anon_sym_in] = ACTIONS(1821),
    [anon_sym_AMP] = ACTIONS(1821),
    [anon_sym_POUND] = ACTIONS(1821),
    [anon_sym_COLON_COLON] = ACTIONS(1823),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1823),
    [anon_sym_else] = ACTIONS(1821),
    [anon_sym_new] = ACTIONS(1821),
    [anon_sym_LBRACE_LT] = ACTIONS(1823),
    [anon_sym_begin] = ACTIONS(1821),
    [sym_ocamlyacc_value] = ACTIONS(1823),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1821),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1823),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1821),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1821),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1823),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1823),
    [aux_sym_number_token1] = ACTIONS(1823),
    [aux_sym_signed_number_token1] = ACTIONS(1821),
    [anon_sym_SQUOTE] = ACTIONS(1823),
    [anon_sym_DQUOTE] = ACTIONS(1823),
    [sym_prefix_operator] = ACTIONS(1823),
    [aux_sym_sign_operator_token1] = ACTIONS(1821),
    [sym_hash_operator] = ACTIONS(1823),
    [anon_sym_lsl] = ACTIONS(1821),
    [anon_sym_lsr] = ACTIONS(1821),
    [anon_sym_asr] = ACTIONS(1821),
    [aux_sym_pow_operator_token1] = ACTIONS(1823),
    [anon_sym_mod] = ACTIONS(1821),
    [anon_sym_land] = ACTIONS(1821),
    [anon_sym_lor] = ACTIONS(1821),
    [anon_sym_lxor] = ACTIONS(1821),
    [aux_sym_mult_operator_token1] = ACTIONS(1821),
    [aux_sym_add_operator_token1] = ACTIONS(1821),
    [sym_concat_operator] = ACTIONS(1823),
    [sym_rel_operator] = ACTIONS(1821),
    [anon_sym_AMP_AMP] = ACTIONS(1821),
    [anon_sym_or] = ACTIONS(1821),
    [anon_sym_PIPE_PIPE] = ACTIONS(1821),
    [sym_assign_operator] = ACTIONS(1823),
    [sym_let_operator] = ACTIONS(1823),
    [sym_let_and_operator] = ACTIONS(1823),
    [sym__uppercase_identifier] = ACTIONS(1823),
    [aux_sym_directive_token1] = ACTIONS(1821),
    [aux_sym_tag_token1] = ACTIONS(1823),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(993)] = {
    [sym_attribute] = STATE(993),
    [sym__lowercase_identifier] = ACTIONS(1575),
    [anon_sym_SEMI_SEMI] = ACTIONS(1577),
    [anon_sym_let] = ACTIONS(1575),
    [anon_sym_and] = ACTIONS(1575),
    [anon_sym_TILDE] = ACTIONS(1575),
    [anon_sym_QMARK] = ACTIONS(1575),
    [anon_sym_LPAREN] = ACTIONS(1577),
    [anon_sym_external] = ACTIONS(1575),
    [anon_sym_type] = ACTIONS(1575),
    [anon_sym_COMMA] = ACTIONS(1577),
    [anon_sym_PIPE] = ACTIONS(1575),
    [anon_sym_DOT] = ACTIONS(1577),
    [anon_sym_LBRACK] = ACTIONS(1575),
    [anon_sym_RBRACK] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(1575),
    [anon_sym_false] = ACTIONS(1575),
    [anon_sym_LBRACE] = ACTIONS(1575),
    [anon_sym_SEMI] = ACTIONS(1575),
    [anon_sym_exception] = ACTIONS(1575),
    [anon_sym_module] = ACTIONS(1575),
    [anon_sym_open] = ACTIONS(1575),
    [anon_sym_include] = ACTIONS(1575),
    [anon_sym_class] = ACTIONS(1575),
    [anon_sym_val] = ACTIONS(1575),
    [anon_sym_end] = ACTIONS(1575),
    [anon_sym_object] = ACTIONS(1575),
    [anon_sym_in] = ACTIONS(1575),
    [anon_sym_AMP] = ACTIONS(1575),
    [anon_sym_POUND] = ACTIONS(1575),
    [anon_sym_COLON_COLON] = ACTIONS(1577),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1577),
    [anon_sym_else] = ACTIONS(1575),
    [anon_sym_new] = ACTIONS(1575),
    [anon_sym_LBRACE_LT] = ACTIONS(1577),
    [anon_sym_begin] = ACTIONS(1575),
    [sym_ocamlyacc_value] = ACTIONS(1577),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1575),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1577),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1575),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1575),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1577),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1577),
    [aux_sym_number_token1] = ACTIONS(1577),
    [aux_sym_signed_number_token1] = ACTIONS(1575),
    [anon_sym_SQUOTE] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(1577),
    [sym_prefix_operator] = ACTIONS(1577),
    [aux_sym_sign_operator_token1] = ACTIONS(1575),
    [sym_hash_operator] = ACTIONS(1577),
    [anon_sym_lsl] = ACTIONS(1575),
    [anon_sym_lsr] = ACTIONS(1575),
    [anon_sym_asr] = ACTIONS(1575),
    [aux_sym_pow_operator_token1] = ACTIONS(1577),
    [anon_sym_mod] = ACTIONS(1575),
    [anon_sym_land] = ACTIONS(1575),
    [anon_sym_lor] = ACTIONS(1575),
    [anon_sym_lxor] = ACTIONS(1575),
    [aux_sym_mult_operator_token1] = ACTIONS(1575),
    [aux_sym_add_operator_token1] = ACTIONS(1575),
    [sym_concat_operator] = ACTIONS(1577),
    [sym_rel_operator] = ACTIONS(1575),
    [anon_sym_AMP_AMP] = ACTIONS(1575),
    [anon_sym_or] = ACTIONS(1575),
    [anon_sym_PIPE_PIPE] = ACTIONS(1575),
    [sym_assign_operator] = ACTIONS(1577),
    [sym_let_operator] = ACTIONS(1577),
    [sym_let_and_operator] = ACTIONS(1577),
    [sym__uppercase_identifier] = ACTIONS(1577),
    [aux_sym_directive_token1] = ACTIONS(1575),
    [aux_sym_tag_token1] = ACTIONS(1577),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(994)] = {
    [sym_attribute] = STATE(994),
    [sym__lowercase_identifier] = ACTIONS(1595),
    [anon_sym_SEMI_SEMI] = ACTIONS(1597),
    [anon_sym_let] = ACTIONS(1595),
    [anon_sym_and] = ACTIONS(1595),
    [anon_sym_TILDE] = ACTIONS(1595),
    [anon_sym_QMARK] = ACTIONS(1595),
    [anon_sym_LPAREN] = ACTIONS(1597),
    [anon_sym_external] = ACTIONS(1595),
    [anon_sym_type] = ACTIONS(1595),
    [anon_sym_COMMA] = ACTIONS(1597),
    [anon_sym_PIPE] = ACTIONS(1595),
    [anon_sym_DOT] = ACTIONS(1597),
    [anon_sym_LBRACK] = ACTIONS(1595),
    [anon_sym_RBRACK] = ACTIONS(1597),
    [anon_sym_true] = ACTIONS(1595),
    [anon_sym_false] = ACTIONS(1595),
    [anon_sym_LBRACE] = ACTIONS(1595),
    [anon_sym_SEMI] = ACTIONS(1595),
    [anon_sym_exception] = ACTIONS(1595),
    [anon_sym_module] = ACTIONS(1595),
    [anon_sym_open] = ACTIONS(1595),
    [anon_sym_include] = ACTIONS(1595),
    [anon_sym_class] = ACTIONS(1595),
    [anon_sym_val] = ACTIONS(1595),
    [anon_sym_end] = ACTIONS(1595),
    [anon_sym_object] = ACTIONS(1595),
    [anon_sym_in] = ACTIONS(1595),
    [anon_sym_AMP] = ACTIONS(1595),
    [anon_sym_POUND] = ACTIONS(1595),
    [anon_sym_COLON_COLON] = ACTIONS(1597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1597),
    [anon_sym_else] = ACTIONS(1595),
    [anon_sym_new] = ACTIONS(1595),
    [anon_sym_LBRACE_LT] = ACTIONS(1597),
    [anon_sym_begin] = ACTIONS(1595),
    [sym_ocamlyacc_value] = ACTIONS(1597),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1595),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1597),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1595),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1595),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1597),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1597),
    [aux_sym_number_token1] = ACTIONS(1597),
    [aux_sym_signed_number_token1] = ACTIONS(1595),
    [anon_sym_SQUOTE] = ACTIONS(1597),
    [anon_sym_DQUOTE] = ACTIONS(1597),
    [sym_prefix_operator] = ACTIONS(1597),
    [aux_sym_sign_operator_token1] = ACTIONS(1595),
    [sym_hash_operator] = ACTIONS(1597),
    [anon_sym_lsl] = ACTIONS(1595),
    [anon_sym_lsr] = ACTIONS(1595),
    [anon_sym_asr] = ACTIONS(1595),
    [aux_sym_pow_operator_token1] = ACTIONS(1597),
    [anon_sym_mod] = ACTIONS(1595),
    [anon_sym_land] = ACTIONS(1595),
    [anon_sym_lor] = ACTIONS(1595),
    [anon_sym_lxor] = ACTIONS(1595),
    [aux_sym_mult_operator_token1] = ACTIONS(1595),
    [aux_sym_add_operator_token1] = ACTIONS(1595),
    [sym_concat_operator] = ACTIONS(1597),
    [sym_rel_operator] = ACTIONS(1595),
    [anon_sym_AMP_AMP] = ACTIONS(1595),
    [anon_sym_or] = ACTIONS(1595),
    [anon_sym_PIPE_PIPE] = ACTIONS(1595),
    [sym_assign_operator] = ACTIONS(1597),
    [sym_let_operator] = ACTIONS(1597),
    [sym_let_and_operator] = ACTIONS(1597),
    [sym__uppercase_identifier] = ACTIONS(1597),
    [aux_sym_directive_token1] = ACTIONS(1595),
    [aux_sym_tag_token1] = ACTIONS(1597),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(995)] = {
    [sym_attribute] = STATE(995),
    [sym__lowercase_identifier] = ACTIONS(1683),
    [anon_sym_SEMI_SEMI] = ACTIONS(1685),
    [anon_sym_let] = ACTIONS(1683),
    [anon_sym_and] = ACTIONS(1683),
    [anon_sym_TILDE] = ACTIONS(1683),
    [anon_sym_QMARK] = ACTIONS(1683),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_external] = ACTIONS(1683),
    [anon_sym_type] = ACTIONS(1683),
    [anon_sym_COMMA] = ACTIONS(1685),
    [anon_sym_PIPE] = ACTIONS(1683),
    [anon_sym_DOT] = ACTIONS(1685),
    [anon_sym_LBRACK] = ACTIONS(1683),
    [anon_sym_RBRACK] = ACTIONS(1685),
    [anon_sym_true] = ACTIONS(1683),
    [anon_sym_false] = ACTIONS(1683),
    [anon_sym_LBRACE] = ACTIONS(1683),
    [anon_sym_SEMI] = ACTIONS(1683),
    [anon_sym_exception] = ACTIONS(1683),
    [anon_sym_module] = ACTIONS(1683),
    [anon_sym_open] = ACTIONS(1683),
    [anon_sym_include] = ACTIONS(1683),
    [anon_sym_class] = ACTIONS(1683),
    [anon_sym_val] = ACTIONS(1683),
    [anon_sym_end] = ACTIONS(1683),
    [anon_sym_object] = ACTIONS(1683),
    [anon_sym_in] = ACTIONS(1683),
    [anon_sym_AMP] = ACTIONS(1683),
    [anon_sym_POUND] = ACTIONS(1683),
    [anon_sym_COLON_COLON] = ACTIONS(1685),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1685),
    [anon_sym_else] = ACTIONS(1683),
    [anon_sym_new] = ACTIONS(1683),
    [anon_sym_LBRACE_LT] = ACTIONS(1685),
    [anon_sym_begin] = ACTIONS(1683),
    [sym_ocamlyacc_value] = ACTIONS(1685),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1683),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1685),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1683),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1683),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1685),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1685),
    [aux_sym_number_token1] = ACTIONS(1685),
    [aux_sym_signed_number_token1] = ACTIONS(1683),
    [anon_sym_SQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1685),
    [sym_prefix_operator] = ACTIONS(1685),
    [aux_sym_sign_operator_token1] = ACTIONS(1683),
    [sym_hash_operator] = ACTIONS(1685),
    [anon_sym_lsl] = ACTIONS(1683),
    [anon_sym_lsr] = ACTIONS(1683),
    [anon_sym_asr] = ACTIONS(1683),
    [aux_sym_pow_operator_token1] = ACTIONS(1685),
    [anon_sym_mod] = ACTIONS(1683),
    [anon_sym_land] = ACTIONS(1683),
    [anon_sym_lor] = ACTIONS(1683),
    [anon_sym_lxor] = ACTIONS(1683),
    [aux_sym_mult_operator_token1] = ACTIONS(1683),
    [aux_sym_add_operator_token1] = ACTIONS(1683),
    [sym_concat_operator] = ACTIONS(1685),
    [sym_rel_operator] = ACTIONS(1683),
    [anon_sym_AMP_AMP] = ACTIONS(1683),
    [anon_sym_or] = ACTIONS(1683),
    [anon_sym_PIPE_PIPE] = ACTIONS(1683),
    [sym_assign_operator] = ACTIONS(1685),
    [sym_let_operator] = ACTIONS(1685),
    [sym_let_and_operator] = ACTIONS(1685),
    [sym__uppercase_identifier] = ACTIONS(1685),
    [aux_sym_directive_token1] = ACTIONS(1683),
    [aux_sym_tag_token1] = ACTIONS(1685),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(996)] = {
    [sym_attribute] = STATE(996),
    [sym__lowercase_identifier] = ACTIONS(1687),
    [anon_sym_SEMI_SEMI] = ACTIONS(1689),
    [anon_sym_let] = ACTIONS(1687),
    [anon_sym_and] = ACTIONS(1687),
    [anon_sym_TILDE] = ACTIONS(1687),
    [anon_sym_QMARK] = ACTIONS(1687),
    [anon_sym_LPAREN] = ACTIONS(1689),
    [anon_sym_external] = ACTIONS(1687),
    [anon_sym_type] = ACTIONS(1687),
    [anon_sym_COMMA] = ACTIONS(1689),
    [anon_sym_PIPE] = ACTIONS(1687),
    [anon_sym_DOT] = ACTIONS(1689),
    [anon_sym_LBRACK] = ACTIONS(1687),
    [anon_sym_RBRACK] = ACTIONS(1689),
    [anon_sym_true] = ACTIONS(1687),
    [anon_sym_false] = ACTIONS(1687),
    [anon_sym_LBRACE] = ACTIONS(1687),
    [anon_sym_SEMI] = ACTIONS(1687),
    [anon_sym_exception] = ACTIONS(1687),
    [anon_sym_module] = ACTIONS(1687),
    [anon_sym_open] = ACTIONS(1687),
    [anon_sym_include] = ACTIONS(1687),
    [anon_sym_class] = ACTIONS(1687),
    [anon_sym_val] = ACTIONS(1687),
    [anon_sym_end] = ACTIONS(1687),
    [anon_sym_object] = ACTIONS(1687),
    [anon_sym_in] = ACTIONS(1687),
    [anon_sym_AMP] = ACTIONS(1687),
    [anon_sym_POUND] = ACTIONS(1687),
    [anon_sym_COLON_COLON] = ACTIONS(1689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1689),
    [anon_sym_else] = ACTIONS(1687),
    [anon_sym_new] = ACTIONS(1687),
    [anon_sym_LBRACE_LT] = ACTIONS(1689),
    [anon_sym_begin] = ACTIONS(1687),
    [sym_ocamlyacc_value] = ACTIONS(1689),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1687),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1689),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1687),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1687),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1689),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1689),
    [aux_sym_number_token1] = ACTIONS(1689),
    [aux_sym_signed_number_token1] = ACTIONS(1687),
    [anon_sym_SQUOTE] = ACTIONS(1689),
    [anon_sym_DQUOTE] = ACTIONS(1689),
    [sym_prefix_operator] = ACTIONS(1689),
    [aux_sym_sign_operator_token1] = ACTIONS(1687),
    [sym_hash_operator] = ACTIONS(1689),
    [anon_sym_lsl] = ACTIONS(1687),
    [anon_sym_lsr] = ACTIONS(1687),
    [anon_sym_asr] = ACTIONS(1687),
    [aux_sym_pow_operator_token1] = ACTIONS(1689),
    [anon_sym_mod] = ACTIONS(1687),
    [anon_sym_land] = ACTIONS(1687),
    [anon_sym_lor] = ACTIONS(1687),
    [anon_sym_lxor] = ACTIONS(1687),
    [aux_sym_mult_operator_token1] = ACTIONS(1687),
    [aux_sym_add_operator_token1] = ACTIONS(1687),
    [sym_concat_operator] = ACTIONS(1689),
    [sym_rel_operator] = ACTIONS(1687),
    [anon_sym_AMP_AMP] = ACTIONS(1687),
    [anon_sym_or] = ACTIONS(1687),
    [anon_sym_PIPE_PIPE] = ACTIONS(1687),
    [sym_assign_operator] = ACTIONS(1689),
    [sym_let_operator] = ACTIONS(1689),
    [sym_let_and_operator] = ACTIONS(1689),
    [sym__uppercase_identifier] = ACTIONS(1689),
    [aux_sym_directive_token1] = ACTIONS(1687),
    [aux_sym_tag_token1] = ACTIONS(1689),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(997)] = {
    [sym_attribute] = STATE(997),
    [sym__lowercase_identifier] = ACTIONS(1709),
    [anon_sym_SEMI_SEMI] = ACTIONS(1711),
    [anon_sym_let] = ACTIONS(1709),
    [anon_sym_and] = ACTIONS(1709),
    [anon_sym_TILDE] = ACTIONS(1709),
    [anon_sym_QMARK] = ACTIONS(1709),
    [anon_sym_LPAREN] = ACTIONS(1711),
    [anon_sym_external] = ACTIONS(1709),
    [anon_sym_type] = ACTIONS(1709),
    [anon_sym_COMMA] = ACTIONS(1711),
    [anon_sym_PIPE] = ACTIONS(1709),
    [anon_sym_DOT] = ACTIONS(1711),
    [anon_sym_LBRACK] = ACTIONS(1709),
    [anon_sym_RBRACK] = ACTIONS(1711),
    [anon_sym_true] = ACTIONS(1709),
    [anon_sym_false] = ACTIONS(1709),
    [anon_sym_LBRACE] = ACTIONS(1709),
    [anon_sym_SEMI] = ACTIONS(1709),
    [anon_sym_exception] = ACTIONS(1709),
    [anon_sym_module] = ACTIONS(1709),
    [anon_sym_open] = ACTIONS(1709),
    [anon_sym_include] = ACTIONS(1709),
    [anon_sym_class] = ACTIONS(1709),
    [anon_sym_val] = ACTIONS(1709),
    [anon_sym_end] = ACTIONS(1709),
    [anon_sym_object] = ACTIONS(1709),
    [anon_sym_in] = ACTIONS(1709),
    [anon_sym_AMP] = ACTIONS(1709),
    [anon_sym_POUND] = ACTIONS(1709),
    [anon_sym_COLON_COLON] = ACTIONS(1711),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1711),
    [anon_sym_else] = ACTIONS(1709),
    [anon_sym_new] = ACTIONS(1709),
    [anon_sym_LBRACE_LT] = ACTIONS(1711),
    [anon_sym_begin] = ACTIONS(1709),
    [sym_ocamlyacc_value] = ACTIONS(1711),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1709),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1711),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1709),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1709),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1711),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1711),
    [aux_sym_number_token1] = ACTIONS(1711),
    [aux_sym_signed_number_token1] = ACTIONS(1709),
    [anon_sym_SQUOTE] = ACTIONS(1711),
    [anon_sym_DQUOTE] = ACTIONS(1711),
    [sym_prefix_operator] = ACTIONS(1711),
    [aux_sym_sign_operator_token1] = ACTIONS(1709),
    [sym_hash_operator] = ACTIONS(1711),
    [anon_sym_lsl] = ACTIONS(1709),
    [anon_sym_lsr] = ACTIONS(1709),
    [anon_sym_asr] = ACTIONS(1709),
    [aux_sym_pow_operator_token1] = ACTIONS(1711),
    [anon_sym_mod] = ACTIONS(1709),
    [anon_sym_land] = ACTIONS(1709),
    [anon_sym_lor] = ACTIONS(1709),
    [anon_sym_lxor] = ACTIONS(1709),
    [aux_sym_mult_operator_token1] = ACTIONS(1709),
    [aux_sym_add_operator_token1] = ACTIONS(1709),
    [sym_concat_operator] = ACTIONS(1711),
    [sym_rel_operator] = ACTIONS(1709),
    [anon_sym_AMP_AMP] = ACTIONS(1709),
    [anon_sym_or] = ACTIONS(1709),
    [anon_sym_PIPE_PIPE] = ACTIONS(1709),
    [sym_assign_operator] = ACTIONS(1711),
    [sym_let_operator] = ACTIONS(1711),
    [sym_let_and_operator] = ACTIONS(1711),
    [sym__uppercase_identifier] = ACTIONS(1711),
    [aux_sym_directive_token1] = ACTIONS(1709),
    [aux_sym_tag_token1] = ACTIONS(1711),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(998)] = {
    [sym_attribute] = STATE(998),
    [sym__lowercase_identifier] = ACTIONS(1611),
    [anon_sym_SEMI_SEMI] = ACTIONS(1613),
    [anon_sym_let] = ACTIONS(1611),
    [anon_sym_and] = ACTIONS(1611),
    [anon_sym_TILDE] = ACTIONS(1611),
    [anon_sym_QMARK] = ACTIONS(1611),
    [anon_sym_LPAREN] = ACTIONS(1613),
    [anon_sym_external] = ACTIONS(1611),
    [anon_sym_type] = ACTIONS(1611),
    [anon_sym_COMMA] = ACTIONS(1613),
    [anon_sym_PIPE] = ACTIONS(1611),
    [anon_sym_DOT] = ACTIONS(1613),
    [anon_sym_LBRACK] = ACTIONS(1611),
    [anon_sym_RBRACK] = ACTIONS(1613),
    [anon_sym_true] = ACTIONS(1611),
    [anon_sym_false] = ACTIONS(1611),
    [anon_sym_LBRACE] = ACTIONS(1611),
    [anon_sym_SEMI] = ACTIONS(1611),
    [anon_sym_exception] = ACTIONS(1611),
    [anon_sym_module] = ACTIONS(1611),
    [anon_sym_open] = ACTIONS(1611),
    [anon_sym_include] = ACTIONS(1611),
    [anon_sym_class] = ACTIONS(1611),
    [anon_sym_val] = ACTIONS(1611),
    [anon_sym_end] = ACTIONS(1611),
    [anon_sym_object] = ACTIONS(1611),
    [anon_sym_in] = ACTIONS(1611),
    [anon_sym_AMP] = ACTIONS(1611),
    [anon_sym_POUND] = ACTIONS(1611),
    [anon_sym_COLON_COLON] = ACTIONS(1613),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1613),
    [anon_sym_else] = ACTIONS(1611),
    [anon_sym_new] = ACTIONS(1611),
    [anon_sym_LBRACE_LT] = ACTIONS(1613),
    [anon_sym_begin] = ACTIONS(1611),
    [sym_ocamlyacc_value] = ACTIONS(1613),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1611),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1613),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1611),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1611),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1613),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1613),
    [aux_sym_number_token1] = ACTIONS(1613),
    [aux_sym_signed_number_token1] = ACTIONS(1611),
    [anon_sym_SQUOTE] = ACTIONS(1613),
    [anon_sym_DQUOTE] = ACTIONS(1613),
    [sym_prefix_operator] = ACTIONS(1613),
    [aux_sym_sign_operator_token1] = ACTIONS(1611),
    [sym_hash_operator] = ACTIONS(1613),
    [anon_sym_lsl] = ACTIONS(1611),
    [anon_sym_lsr] = ACTIONS(1611),
    [anon_sym_asr] = ACTIONS(1611),
    [aux_sym_pow_operator_token1] = ACTIONS(1613),
    [anon_sym_mod] = ACTIONS(1611),
    [anon_sym_land] = ACTIONS(1611),
    [anon_sym_lor] = ACTIONS(1611),
    [anon_sym_lxor] = ACTIONS(1611),
    [aux_sym_mult_operator_token1] = ACTIONS(1611),
    [aux_sym_add_operator_token1] = ACTIONS(1611),
    [sym_concat_operator] = ACTIONS(1613),
    [sym_rel_operator] = ACTIONS(1611),
    [anon_sym_AMP_AMP] = ACTIONS(1611),
    [anon_sym_or] = ACTIONS(1611),
    [anon_sym_PIPE_PIPE] = ACTIONS(1611),
    [sym_assign_operator] = ACTIONS(1613),
    [sym_let_operator] = ACTIONS(1613),
    [sym_let_and_operator] = ACTIONS(1613),
    [sym__uppercase_identifier] = ACTIONS(1613),
    [aux_sym_directive_token1] = ACTIONS(1611),
    [aux_sym_tag_token1] = ACTIONS(1613),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(999)] = {
    [sym_attribute] = STATE(999),
    [sym__lowercase_identifier] = ACTIONS(1623),
    [anon_sym_SEMI_SEMI] = ACTIONS(1625),
    [anon_sym_let] = ACTIONS(1623),
    [anon_sym_and] = ACTIONS(1623),
    [anon_sym_TILDE] = ACTIONS(1623),
    [anon_sym_QMARK] = ACTIONS(1623),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_external] = ACTIONS(1623),
    [anon_sym_type] = ACTIONS(1623),
    [anon_sym_COMMA] = ACTIONS(1625),
    [anon_sym_PIPE] = ACTIONS(1623),
    [anon_sym_DOT] = ACTIONS(1625),
    [anon_sym_LBRACK] = ACTIONS(1623),
    [anon_sym_RBRACK] = ACTIONS(1625),
    [anon_sym_true] = ACTIONS(1623),
    [anon_sym_false] = ACTIONS(1623),
    [anon_sym_LBRACE] = ACTIONS(1623),
    [anon_sym_SEMI] = ACTIONS(1623),
    [anon_sym_exception] = ACTIONS(1623),
    [anon_sym_module] = ACTIONS(1623),
    [anon_sym_open] = ACTIONS(1623),
    [anon_sym_include] = ACTIONS(1623),
    [anon_sym_class] = ACTIONS(1623),
    [anon_sym_val] = ACTIONS(1623),
    [anon_sym_end] = ACTIONS(1623),
    [anon_sym_object] = ACTIONS(1623),
    [anon_sym_in] = ACTIONS(1623),
    [anon_sym_AMP] = ACTIONS(1623),
    [anon_sym_POUND] = ACTIONS(1623),
    [anon_sym_COLON_COLON] = ACTIONS(1625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1625),
    [anon_sym_else] = ACTIONS(1623),
    [anon_sym_new] = ACTIONS(1623),
    [anon_sym_LBRACE_LT] = ACTIONS(1625),
    [anon_sym_begin] = ACTIONS(1623),
    [sym_ocamlyacc_value] = ACTIONS(1625),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1623),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1625),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1623),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1623),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1625),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1625),
    [aux_sym_number_token1] = ACTIONS(1625),
    [aux_sym_signed_number_token1] = ACTIONS(1623),
    [anon_sym_SQUOTE] = ACTIONS(1625),
    [anon_sym_DQUOTE] = ACTIONS(1625),
    [sym_prefix_operator] = ACTIONS(1625),
    [aux_sym_sign_operator_token1] = ACTIONS(1623),
    [sym_hash_operator] = ACTIONS(1625),
    [anon_sym_lsl] = ACTIONS(1623),
    [anon_sym_lsr] = ACTIONS(1623),
    [anon_sym_asr] = ACTIONS(1623),
    [aux_sym_pow_operator_token1] = ACTIONS(1625),
    [anon_sym_mod] = ACTIONS(1623),
    [anon_sym_land] = ACTIONS(1623),
    [anon_sym_lor] = ACTIONS(1623),
    [anon_sym_lxor] = ACTIONS(1623),
    [aux_sym_mult_operator_token1] = ACTIONS(1623),
    [aux_sym_add_operator_token1] = ACTIONS(1623),
    [sym_concat_operator] = ACTIONS(1625),
    [sym_rel_operator] = ACTIONS(1623),
    [anon_sym_AMP_AMP] = ACTIONS(1623),
    [anon_sym_or] = ACTIONS(1623),
    [anon_sym_PIPE_PIPE] = ACTIONS(1623),
    [sym_assign_operator] = ACTIONS(1625),
    [sym_let_operator] = ACTIONS(1625),
    [sym_let_and_operator] = ACTIONS(1625),
    [sym__uppercase_identifier] = ACTIONS(1625),
    [aux_sym_directive_token1] = ACTIONS(1623),
    [aux_sym_tag_token1] = ACTIONS(1625),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1000)] = {
    [sym_attribute] = STATE(1000),
    [sym__lowercase_identifier] = ACTIONS(1721),
    [anon_sym_SEMI_SEMI] = ACTIONS(1723),
    [anon_sym_let] = ACTIONS(1721),
    [anon_sym_and] = ACTIONS(1721),
    [anon_sym_TILDE] = ACTIONS(1721),
    [anon_sym_QMARK] = ACTIONS(1721),
    [anon_sym_LPAREN] = ACTIONS(1723),
    [anon_sym_external] = ACTIONS(1721),
    [anon_sym_type] = ACTIONS(1721),
    [anon_sym_COMMA] = ACTIONS(1723),
    [anon_sym_PIPE] = ACTIONS(1721),
    [anon_sym_DOT] = ACTIONS(1723),
    [anon_sym_LBRACK] = ACTIONS(1721),
    [anon_sym_RBRACK] = ACTIONS(1723),
    [anon_sym_true] = ACTIONS(1721),
    [anon_sym_false] = ACTIONS(1721),
    [anon_sym_LBRACE] = ACTIONS(1721),
    [anon_sym_SEMI] = ACTIONS(1721),
    [anon_sym_exception] = ACTIONS(1721),
    [anon_sym_module] = ACTIONS(1721),
    [anon_sym_open] = ACTIONS(1721),
    [anon_sym_include] = ACTIONS(1721),
    [anon_sym_class] = ACTIONS(1721),
    [anon_sym_val] = ACTIONS(1721),
    [anon_sym_end] = ACTIONS(1721),
    [anon_sym_object] = ACTIONS(1721),
    [anon_sym_in] = ACTIONS(1721),
    [anon_sym_AMP] = ACTIONS(1721),
    [anon_sym_POUND] = ACTIONS(1721),
    [anon_sym_COLON_COLON] = ACTIONS(1723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1723),
    [anon_sym_else] = ACTIONS(1721),
    [anon_sym_new] = ACTIONS(1721),
    [anon_sym_LBRACE_LT] = ACTIONS(1723),
    [anon_sym_begin] = ACTIONS(1721),
    [sym_ocamlyacc_value] = ACTIONS(1723),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1721),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1723),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1721),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1721),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1723),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1723),
    [aux_sym_number_token1] = ACTIONS(1723),
    [aux_sym_signed_number_token1] = ACTIONS(1721),
    [anon_sym_SQUOTE] = ACTIONS(1723),
    [anon_sym_DQUOTE] = ACTIONS(1723),
    [sym_prefix_operator] = ACTIONS(1723),
    [aux_sym_sign_operator_token1] = ACTIONS(1721),
    [sym_hash_operator] = ACTIONS(1723),
    [anon_sym_lsl] = ACTIONS(1721),
    [anon_sym_lsr] = ACTIONS(1721),
    [anon_sym_asr] = ACTIONS(1721),
    [aux_sym_pow_operator_token1] = ACTIONS(1723),
    [anon_sym_mod] = ACTIONS(1721),
    [anon_sym_land] = ACTIONS(1721),
    [anon_sym_lor] = ACTIONS(1721),
    [anon_sym_lxor] = ACTIONS(1721),
    [aux_sym_mult_operator_token1] = ACTIONS(1721),
    [aux_sym_add_operator_token1] = ACTIONS(1721),
    [sym_concat_operator] = ACTIONS(1723),
    [sym_rel_operator] = ACTIONS(1721),
    [anon_sym_AMP_AMP] = ACTIONS(1721),
    [anon_sym_or] = ACTIONS(1721),
    [anon_sym_PIPE_PIPE] = ACTIONS(1721),
    [sym_assign_operator] = ACTIONS(1723),
    [sym_let_operator] = ACTIONS(1723),
    [sym_let_and_operator] = ACTIONS(1723),
    [sym__uppercase_identifier] = ACTIONS(1723),
    [aux_sym_directive_token1] = ACTIONS(1721),
    [aux_sym_tag_token1] = ACTIONS(1723),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1001)] = {
    [sym_attribute] = STATE(1001),
    [sym__lowercase_identifier] = ACTIONS(1733),
    [anon_sym_SEMI_SEMI] = ACTIONS(1735),
    [anon_sym_let] = ACTIONS(1733),
    [anon_sym_and] = ACTIONS(1733),
    [anon_sym_TILDE] = ACTIONS(1733),
    [anon_sym_QMARK] = ACTIONS(1733),
    [anon_sym_LPAREN] = ACTIONS(1735),
    [anon_sym_external] = ACTIONS(1733),
    [anon_sym_type] = ACTIONS(1733),
    [anon_sym_COMMA] = ACTIONS(1735),
    [anon_sym_PIPE] = ACTIONS(1733),
    [anon_sym_DOT] = ACTIONS(1735),
    [anon_sym_LBRACK] = ACTIONS(1733),
    [anon_sym_RBRACK] = ACTIONS(1735),
    [anon_sym_true] = ACTIONS(1733),
    [anon_sym_false] = ACTIONS(1733),
    [anon_sym_LBRACE] = ACTIONS(1733),
    [anon_sym_SEMI] = ACTIONS(1733),
    [anon_sym_exception] = ACTIONS(1733),
    [anon_sym_module] = ACTIONS(1733),
    [anon_sym_open] = ACTIONS(1733),
    [anon_sym_include] = ACTIONS(1733),
    [anon_sym_class] = ACTIONS(1733),
    [anon_sym_val] = ACTIONS(1733),
    [anon_sym_end] = ACTIONS(1733),
    [anon_sym_object] = ACTIONS(1733),
    [anon_sym_in] = ACTIONS(1733),
    [anon_sym_AMP] = ACTIONS(1733),
    [anon_sym_POUND] = ACTIONS(1733),
    [anon_sym_COLON_COLON] = ACTIONS(1735),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1735),
    [anon_sym_else] = ACTIONS(1733),
    [anon_sym_new] = ACTIONS(1733),
    [anon_sym_LBRACE_LT] = ACTIONS(1735),
    [anon_sym_begin] = ACTIONS(1733),
    [sym_ocamlyacc_value] = ACTIONS(1735),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1733),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1735),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1733),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1733),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1735),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1735),
    [aux_sym_number_token1] = ACTIONS(1735),
    [aux_sym_signed_number_token1] = ACTIONS(1733),
    [anon_sym_SQUOTE] = ACTIONS(1735),
    [anon_sym_DQUOTE] = ACTIONS(1735),
    [sym_prefix_operator] = ACTIONS(1735),
    [aux_sym_sign_operator_token1] = ACTIONS(1733),
    [sym_hash_operator] = ACTIONS(1735),
    [anon_sym_lsl] = ACTIONS(1733),
    [anon_sym_lsr] = ACTIONS(1733),
    [anon_sym_asr] = ACTIONS(1733),
    [aux_sym_pow_operator_token1] = ACTIONS(1735),
    [anon_sym_mod] = ACTIONS(1733),
    [anon_sym_land] = ACTIONS(1733),
    [anon_sym_lor] = ACTIONS(1733),
    [anon_sym_lxor] = ACTIONS(1733),
    [aux_sym_mult_operator_token1] = ACTIONS(1733),
    [aux_sym_add_operator_token1] = ACTIONS(1733),
    [sym_concat_operator] = ACTIONS(1735),
    [sym_rel_operator] = ACTIONS(1733),
    [anon_sym_AMP_AMP] = ACTIONS(1733),
    [anon_sym_or] = ACTIONS(1733),
    [anon_sym_PIPE_PIPE] = ACTIONS(1733),
    [sym_assign_operator] = ACTIONS(1735),
    [sym_let_operator] = ACTIONS(1735),
    [sym_let_and_operator] = ACTIONS(1735),
    [sym__uppercase_identifier] = ACTIONS(1735),
    [aux_sym_directive_token1] = ACTIONS(1733),
    [aux_sym_tag_token1] = ACTIONS(1735),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1002)] = {
    [sym_attribute] = STATE(1002),
    [sym__lowercase_identifier] = ACTIONS(1813),
    [anon_sym_SEMI_SEMI] = ACTIONS(1815),
    [anon_sym_let] = ACTIONS(1813),
    [anon_sym_and] = ACTIONS(1813),
    [anon_sym_TILDE] = ACTIONS(1813),
    [anon_sym_QMARK] = ACTIONS(1813),
    [anon_sym_LPAREN] = ACTIONS(1815),
    [anon_sym_external] = ACTIONS(1813),
    [anon_sym_type] = ACTIONS(1813),
    [anon_sym_COMMA] = ACTIONS(1815),
    [anon_sym_PIPE] = ACTIONS(1813),
    [anon_sym_DOT] = ACTIONS(1815),
    [anon_sym_LBRACK] = ACTIONS(1813),
    [anon_sym_RBRACK] = ACTIONS(1815),
    [anon_sym_true] = ACTIONS(1813),
    [anon_sym_false] = ACTIONS(1813),
    [anon_sym_LBRACE] = ACTIONS(1813),
    [anon_sym_SEMI] = ACTIONS(1813),
    [anon_sym_exception] = ACTIONS(1813),
    [anon_sym_module] = ACTIONS(1813),
    [anon_sym_open] = ACTIONS(1813),
    [anon_sym_include] = ACTIONS(1813),
    [anon_sym_class] = ACTIONS(1813),
    [anon_sym_val] = ACTIONS(1813),
    [anon_sym_end] = ACTIONS(1813),
    [anon_sym_object] = ACTIONS(1813),
    [anon_sym_in] = ACTIONS(1813),
    [anon_sym_AMP] = ACTIONS(1813),
    [anon_sym_POUND] = ACTIONS(1813),
    [anon_sym_COLON_COLON] = ACTIONS(1815),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1815),
    [anon_sym_else] = ACTIONS(1813),
    [anon_sym_new] = ACTIONS(1813),
    [anon_sym_LBRACE_LT] = ACTIONS(1815),
    [anon_sym_begin] = ACTIONS(1813),
    [sym_ocamlyacc_value] = ACTIONS(1815),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1813),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1815),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1813),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1813),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1815),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1815),
    [aux_sym_number_token1] = ACTIONS(1815),
    [aux_sym_signed_number_token1] = ACTIONS(1813),
    [anon_sym_SQUOTE] = ACTIONS(1815),
    [anon_sym_DQUOTE] = ACTIONS(1815),
    [sym_prefix_operator] = ACTIONS(1815),
    [aux_sym_sign_operator_token1] = ACTIONS(1813),
    [sym_hash_operator] = ACTIONS(1815),
    [anon_sym_lsl] = ACTIONS(1813),
    [anon_sym_lsr] = ACTIONS(1813),
    [anon_sym_asr] = ACTIONS(1813),
    [aux_sym_pow_operator_token1] = ACTIONS(1815),
    [anon_sym_mod] = ACTIONS(1813),
    [anon_sym_land] = ACTIONS(1813),
    [anon_sym_lor] = ACTIONS(1813),
    [anon_sym_lxor] = ACTIONS(1813),
    [aux_sym_mult_operator_token1] = ACTIONS(1813),
    [aux_sym_add_operator_token1] = ACTIONS(1813),
    [sym_concat_operator] = ACTIONS(1815),
    [sym_rel_operator] = ACTIONS(1813),
    [anon_sym_AMP_AMP] = ACTIONS(1813),
    [anon_sym_or] = ACTIONS(1813),
    [anon_sym_PIPE_PIPE] = ACTIONS(1813),
    [sym_assign_operator] = ACTIONS(1815),
    [sym_let_operator] = ACTIONS(1815),
    [sym_let_and_operator] = ACTIONS(1815),
    [sym__uppercase_identifier] = ACTIONS(1815),
    [aux_sym_directive_token1] = ACTIONS(1813),
    [aux_sym_tag_token1] = ACTIONS(1815),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1003)] = {
    [sym_attribute] = STATE(1003),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_SEMI_SEMI] = ACTIONS(1819),
    [anon_sym_let] = ACTIONS(1817),
    [anon_sym_and] = ACTIONS(1817),
    [anon_sym_TILDE] = ACTIONS(1817),
    [anon_sym_QMARK] = ACTIONS(1817),
    [anon_sym_LPAREN] = ACTIONS(1819),
    [anon_sym_external] = ACTIONS(1817),
    [anon_sym_type] = ACTIONS(1817),
    [anon_sym_COMMA] = ACTIONS(1819),
    [anon_sym_PIPE] = ACTIONS(1817),
    [anon_sym_DOT] = ACTIONS(1819),
    [anon_sym_LBRACK] = ACTIONS(1817),
    [anon_sym_RBRACK] = ACTIONS(1819),
    [anon_sym_true] = ACTIONS(1817),
    [anon_sym_false] = ACTIONS(1817),
    [anon_sym_LBRACE] = ACTIONS(1817),
    [anon_sym_SEMI] = ACTIONS(1817),
    [anon_sym_exception] = ACTIONS(1817),
    [anon_sym_module] = ACTIONS(1817),
    [anon_sym_open] = ACTIONS(1817),
    [anon_sym_include] = ACTIONS(1817),
    [anon_sym_class] = ACTIONS(1817),
    [anon_sym_val] = ACTIONS(1817),
    [anon_sym_end] = ACTIONS(1817),
    [anon_sym_object] = ACTIONS(1817),
    [anon_sym_in] = ACTIONS(1817),
    [anon_sym_AMP] = ACTIONS(1817),
    [anon_sym_POUND] = ACTIONS(1817),
    [anon_sym_COLON_COLON] = ACTIONS(1819),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1819),
    [anon_sym_else] = ACTIONS(1817),
    [anon_sym_new] = ACTIONS(1817),
    [anon_sym_LBRACE_LT] = ACTIONS(1819),
    [anon_sym_begin] = ACTIONS(1817),
    [sym_ocamlyacc_value] = ACTIONS(1819),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1817),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1819),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1817),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1817),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1819),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1819),
    [aux_sym_number_token1] = ACTIONS(1819),
    [aux_sym_signed_number_token1] = ACTIONS(1817),
    [anon_sym_SQUOTE] = ACTIONS(1819),
    [anon_sym_DQUOTE] = ACTIONS(1819),
    [sym_prefix_operator] = ACTIONS(1819),
    [aux_sym_sign_operator_token1] = ACTIONS(1817),
    [sym_hash_operator] = ACTIONS(1819),
    [anon_sym_lsl] = ACTIONS(1817),
    [anon_sym_lsr] = ACTIONS(1817),
    [anon_sym_asr] = ACTIONS(1817),
    [aux_sym_pow_operator_token1] = ACTIONS(1819),
    [anon_sym_mod] = ACTIONS(1817),
    [anon_sym_land] = ACTIONS(1817),
    [anon_sym_lor] = ACTIONS(1817),
    [anon_sym_lxor] = ACTIONS(1817),
    [aux_sym_mult_operator_token1] = ACTIONS(1817),
    [aux_sym_add_operator_token1] = ACTIONS(1817),
    [sym_concat_operator] = ACTIONS(1819),
    [sym_rel_operator] = ACTIONS(1817),
    [anon_sym_AMP_AMP] = ACTIONS(1817),
    [anon_sym_or] = ACTIONS(1817),
    [anon_sym_PIPE_PIPE] = ACTIONS(1817),
    [sym_assign_operator] = ACTIONS(1819),
    [sym_let_operator] = ACTIONS(1819),
    [sym_let_and_operator] = ACTIONS(1819),
    [sym__uppercase_identifier] = ACTIONS(1819),
    [aux_sym_directive_token1] = ACTIONS(1817),
    [aux_sym_tag_token1] = ACTIONS(1819),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1004)] = {
    [sym_attribute] = STATE(1004),
    [sym__lowercase_identifier] = ACTIONS(1825),
    [anon_sym_SEMI_SEMI] = ACTIONS(1827),
    [anon_sym_let] = ACTIONS(1825),
    [anon_sym_and] = ACTIONS(1825),
    [anon_sym_TILDE] = ACTIONS(1825),
    [anon_sym_QMARK] = ACTIONS(1825),
    [anon_sym_LPAREN] = ACTIONS(1827),
    [anon_sym_external] = ACTIONS(1825),
    [anon_sym_type] = ACTIONS(1825),
    [anon_sym_COMMA] = ACTIONS(1827),
    [anon_sym_PIPE] = ACTIONS(1825),
    [anon_sym_DOT] = ACTIONS(1827),
    [anon_sym_LBRACK] = ACTIONS(1825),
    [anon_sym_RBRACK] = ACTIONS(1827),
    [anon_sym_true] = ACTIONS(1825),
    [anon_sym_false] = ACTIONS(1825),
    [anon_sym_LBRACE] = ACTIONS(1825),
    [anon_sym_SEMI] = ACTIONS(1825),
    [anon_sym_exception] = ACTIONS(1825),
    [anon_sym_module] = ACTIONS(1825),
    [anon_sym_open] = ACTIONS(1825),
    [anon_sym_include] = ACTIONS(1825),
    [anon_sym_class] = ACTIONS(1825),
    [anon_sym_val] = ACTIONS(1825),
    [anon_sym_end] = ACTIONS(1825),
    [anon_sym_object] = ACTIONS(1825),
    [anon_sym_in] = ACTIONS(1825),
    [anon_sym_AMP] = ACTIONS(1825),
    [anon_sym_POUND] = ACTIONS(1825),
    [anon_sym_COLON_COLON] = ACTIONS(1827),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1827),
    [anon_sym_else] = ACTIONS(1825),
    [anon_sym_new] = ACTIONS(1825),
    [anon_sym_LBRACE_LT] = ACTIONS(1827),
    [anon_sym_begin] = ACTIONS(1825),
    [sym_ocamlyacc_value] = ACTIONS(1827),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1825),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1827),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1825),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1825),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1827),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1827),
    [aux_sym_number_token1] = ACTIONS(1827),
    [aux_sym_signed_number_token1] = ACTIONS(1825),
    [anon_sym_SQUOTE] = ACTIONS(1827),
    [anon_sym_DQUOTE] = ACTIONS(1827),
    [sym_prefix_operator] = ACTIONS(1827),
    [aux_sym_sign_operator_token1] = ACTIONS(1825),
    [sym_hash_operator] = ACTIONS(1827),
    [anon_sym_lsl] = ACTIONS(1825),
    [anon_sym_lsr] = ACTIONS(1825),
    [anon_sym_asr] = ACTIONS(1825),
    [aux_sym_pow_operator_token1] = ACTIONS(1827),
    [anon_sym_mod] = ACTIONS(1825),
    [anon_sym_land] = ACTIONS(1825),
    [anon_sym_lor] = ACTIONS(1825),
    [anon_sym_lxor] = ACTIONS(1825),
    [aux_sym_mult_operator_token1] = ACTIONS(1825),
    [aux_sym_add_operator_token1] = ACTIONS(1825),
    [sym_concat_operator] = ACTIONS(1827),
    [sym_rel_operator] = ACTIONS(1825),
    [anon_sym_AMP_AMP] = ACTIONS(1825),
    [anon_sym_or] = ACTIONS(1825),
    [anon_sym_PIPE_PIPE] = ACTIONS(1825),
    [sym_assign_operator] = ACTIONS(1827),
    [sym_let_operator] = ACTIONS(1827),
    [sym_let_and_operator] = ACTIONS(1827),
    [sym__uppercase_identifier] = ACTIONS(1827),
    [aux_sym_directive_token1] = ACTIONS(1825),
    [aux_sym_tag_token1] = ACTIONS(1827),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1005)] = {
    [sym_attribute] = STATE(1005),
    [sym__lowercase_identifier] = ACTIONS(1705),
    [anon_sym_SEMI_SEMI] = ACTIONS(1707),
    [anon_sym_let] = ACTIONS(1705),
    [anon_sym_and] = ACTIONS(1705),
    [anon_sym_TILDE] = ACTIONS(1705),
    [anon_sym_QMARK] = ACTIONS(1705),
    [anon_sym_LPAREN] = ACTIONS(1707),
    [anon_sym_external] = ACTIONS(1705),
    [anon_sym_type] = ACTIONS(1705),
    [anon_sym_COMMA] = ACTIONS(1707),
    [anon_sym_PIPE] = ACTIONS(1705),
    [anon_sym_DOT] = ACTIONS(1707),
    [anon_sym_LBRACK] = ACTIONS(1705),
    [anon_sym_RBRACK] = ACTIONS(1707),
    [anon_sym_true] = ACTIONS(1705),
    [anon_sym_false] = ACTIONS(1705),
    [anon_sym_LBRACE] = ACTIONS(1705),
    [anon_sym_SEMI] = ACTIONS(1705),
    [anon_sym_exception] = ACTIONS(1705),
    [anon_sym_module] = ACTIONS(1705),
    [anon_sym_open] = ACTIONS(1705),
    [anon_sym_include] = ACTIONS(1705),
    [anon_sym_class] = ACTIONS(1705),
    [anon_sym_val] = ACTIONS(1705),
    [anon_sym_end] = ACTIONS(1705),
    [anon_sym_object] = ACTIONS(1705),
    [anon_sym_AMP] = ACTIONS(1705),
    [anon_sym_POUND] = ACTIONS(1705),
    [anon_sym_COLON_COLON] = ACTIONS(1707),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1707),
    [anon_sym_LT_DASH] = ACTIONS(2176),
    [anon_sym_else] = ACTIONS(1705),
    [anon_sym_new] = ACTIONS(1705),
    [anon_sym_LBRACE_LT] = ACTIONS(1707),
    [anon_sym_begin] = ACTIONS(1705),
    [sym_ocamlyacc_value] = ACTIONS(1707),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1705),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1707),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1705),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1705),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1707),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1707),
    [aux_sym_number_token1] = ACTIONS(1707),
    [aux_sym_signed_number_token1] = ACTIONS(1705),
    [anon_sym_SQUOTE] = ACTIONS(1707),
    [anon_sym_DQUOTE] = ACTIONS(1707),
    [sym_prefix_operator] = ACTIONS(1707),
    [aux_sym_sign_operator_token1] = ACTIONS(1705),
    [sym_hash_operator] = ACTIONS(1707),
    [anon_sym_lsl] = ACTIONS(1705),
    [anon_sym_lsr] = ACTIONS(1705),
    [anon_sym_asr] = ACTIONS(1705),
    [aux_sym_pow_operator_token1] = ACTIONS(1707),
    [anon_sym_mod] = ACTIONS(1705),
    [anon_sym_land] = ACTIONS(1705),
    [anon_sym_lor] = ACTIONS(1705),
    [anon_sym_lxor] = ACTIONS(1705),
    [aux_sym_mult_operator_token1] = ACTIONS(1705),
    [aux_sym_add_operator_token1] = ACTIONS(1705),
    [sym_concat_operator] = ACTIONS(1707),
    [sym_rel_operator] = ACTIONS(1705),
    [anon_sym_AMP_AMP] = ACTIONS(1705),
    [anon_sym_or] = ACTIONS(1705),
    [anon_sym_PIPE_PIPE] = ACTIONS(1705),
    [sym_assign_operator] = ACTIONS(1707),
    [sym_let_operator] = ACTIONS(1707),
    [sym_let_and_operator] = ACTIONS(1707),
    [sym__uppercase_identifier] = ACTIONS(1707),
    [aux_sym_directive_token1] = ACTIONS(1705),
    [aux_sym_tag_token1] = ACTIONS(1707),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1006)] = {
    [sym_attribute] = STATE(1006),
    [sym__lowercase_identifier] = ACTIONS(1829),
    [anon_sym_SEMI_SEMI] = ACTIONS(1831),
    [anon_sym_let] = ACTIONS(1829),
    [anon_sym_and] = ACTIONS(1829),
    [anon_sym_TILDE] = ACTIONS(1829),
    [anon_sym_QMARK] = ACTIONS(1829),
    [anon_sym_LPAREN] = ACTIONS(1831),
    [anon_sym_external] = ACTIONS(1829),
    [anon_sym_type] = ACTIONS(1829),
    [anon_sym_COMMA] = ACTIONS(1831),
    [anon_sym_PIPE] = ACTIONS(1829),
    [anon_sym_DOT] = ACTIONS(1831),
    [anon_sym_LBRACK] = ACTIONS(1829),
    [anon_sym_RBRACK] = ACTIONS(1831),
    [anon_sym_true] = ACTIONS(1829),
    [anon_sym_false] = ACTIONS(1829),
    [anon_sym_LBRACE] = ACTIONS(1829),
    [anon_sym_SEMI] = ACTIONS(1829),
    [anon_sym_exception] = ACTIONS(1829),
    [anon_sym_module] = ACTIONS(1829),
    [anon_sym_open] = ACTIONS(1829),
    [anon_sym_include] = ACTIONS(1829),
    [anon_sym_class] = ACTIONS(1829),
    [anon_sym_val] = ACTIONS(1829),
    [anon_sym_end] = ACTIONS(1829),
    [anon_sym_object] = ACTIONS(1829),
    [anon_sym_in] = ACTIONS(1829),
    [anon_sym_AMP] = ACTIONS(1829),
    [anon_sym_POUND] = ACTIONS(1829),
    [anon_sym_COLON_COLON] = ACTIONS(1831),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1831),
    [anon_sym_else] = ACTIONS(1829),
    [anon_sym_new] = ACTIONS(1829),
    [anon_sym_LBRACE_LT] = ACTIONS(1831),
    [anon_sym_begin] = ACTIONS(1829),
    [sym_ocamlyacc_value] = ACTIONS(1831),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1829),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1831),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1829),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1829),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1831),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1831),
    [aux_sym_number_token1] = ACTIONS(1831),
    [aux_sym_signed_number_token1] = ACTIONS(1829),
    [anon_sym_SQUOTE] = ACTIONS(1831),
    [anon_sym_DQUOTE] = ACTIONS(1831),
    [sym_prefix_operator] = ACTIONS(1831),
    [aux_sym_sign_operator_token1] = ACTIONS(1829),
    [sym_hash_operator] = ACTIONS(1831),
    [anon_sym_lsl] = ACTIONS(1829),
    [anon_sym_lsr] = ACTIONS(1829),
    [anon_sym_asr] = ACTIONS(1829),
    [aux_sym_pow_operator_token1] = ACTIONS(1831),
    [anon_sym_mod] = ACTIONS(1829),
    [anon_sym_land] = ACTIONS(1829),
    [anon_sym_lor] = ACTIONS(1829),
    [anon_sym_lxor] = ACTIONS(1829),
    [aux_sym_mult_operator_token1] = ACTIONS(1829),
    [aux_sym_add_operator_token1] = ACTIONS(1829),
    [sym_concat_operator] = ACTIONS(1831),
    [sym_rel_operator] = ACTIONS(1829),
    [anon_sym_AMP_AMP] = ACTIONS(1829),
    [anon_sym_or] = ACTIONS(1829),
    [anon_sym_PIPE_PIPE] = ACTIONS(1829),
    [sym_assign_operator] = ACTIONS(1831),
    [sym_let_operator] = ACTIONS(1831),
    [sym_let_and_operator] = ACTIONS(1831),
    [sym__uppercase_identifier] = ACTIONS(1831),
    [aux_sym_directive_token1] = ACTIONS(1829),
    [aux_sym_tag_token1] = ACTIONS(1831),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1007)] = {
    [sym_attribute] = STATE(1007),
    [sym__lowercase_identifier] = ACTIONS(1833),
    [anon_sym_SEMI_SEMI] = ACTIONS(1835),
    [anon_sym_let] = ACTIONS(1833),
    [anon_sym_and] = ACTIONS(1833),
    [anon_sym_TILDE] = ACTIONS(1833),
    [anon_sym_QMARK] = ACTIONS(1833),
    [anon_sym_LPAREN] = ACTIONS(1835),
    [anon_sym_external] = ACTIONS(1833),
    [anon_sym_type] = ACTIONS(1833),
    [anon_sym_COMMA] = ACTIONS(1835),
    [anon_sym_PIPE] = ACTIONS(1833),
    [anon_sym_DOT] = ACTIONS(1835),
    [anon_sym_LBRACK] = ACTIONS(1833),
    [anon_sym_RBRACK] = ACTIONS(1835),
    [anon_sym_true] = ACTIONS(1833),
    [anon_sym_false] = ACTIONS(1833),
    [anon_sym_LBRACE] = ACTIONS(1833),
    [anon_sym_SEMI] = ACTIONS(1833),
    [anon_sym_exception] = ACTIONS(1833),
    [anon_sym_module] = ACTIONS(1833),
    [anon_sym_open] = ACTIONS(1833),
    [anon_sym_include] = ACTIONS(1833),
    [anon_sym_class] = ACTIONS(1833),
    [anon_sym_val] = ACTIONS(1833),
    [anon_sym_end] = ACTIONS(1833),
    [anon_sym_object] = ACTIONS(1833),
    [anon_sym_in] = ACTIONS(1833),
    [anon_sym_AMP] = ACTIONS(1833),
    [anon_sym_POUND] = ACTIONS(1833),
    [anon_sym_COLON_COLON] = ACTIONS(1835),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1835),
    [anon_sym_else] = ACTIONS(1833),
    [anon_sym_new] = ACTIONS(1833),
    [anon_sym_LBRACE_LT] = ACTIONS(1835),
    [anon_sym_begin] = ACTIONS(1833),
    [sym_ocamlyacc_value] = ACTIONS(1835),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1833),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1835),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1833),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1833),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1835),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1835),
    [aux_sym_number_token1] = ACTIONS(1835),
    [aux_sym_signed_number_token1] = ACTIONS(1833),
    [anon_sym_SQUOTE] = ACTIONS(1835),
    [anon_sym_DQUOTE] = ACTIONS(1835),
    [sym_prefix_operator] = ACTIONS(1835),
    [aux_sym_sign_operator_token1] = ACTIONS(1833),
    [sym_hash_operator] = ACTIONS(1835),
    [anon_sym_lsl] = ACTIONS(1833),
    [anon_sym_lsr] = ACTIONS(1833),
    [anon_sym_asr] = ACTIONS(1833),
    [aux_sym_pow_operator_token1] = ACTIONS(1835),
    [anon_sym_mod] = ACTIONS(1833),
    [anon_sym_land] = ACTIONS(1833),
    [anon_sym_lor] = ACTIONS(1833),
    [anon_sym_lxor] = ACTIONS(1833),
    [aux_sym_mult_operator_token1] = ACTIONS(1833),
    [aux_sym_add_operator_token1] = ACTIONS(1833),
    [sym_concat_operator] = ACTIONS(1835),
    [sym_rel_operator] = ACTIONS(1833),
    [anon_sym_AMP_AMP] = ACTIONS(1833),
    [anon_sym_or] = ACTIONS(1833),
    [anon_sym_PIPE_PIPE] = ACTIONS(1833),
    [sym_assign_operator] = ACTIONS(1835),
    [sym_let_operator] = ACTIONS(1835),
    [sym_let_and_operator] = ACTIONS(1835),
    [sym__uppercase_identifier] = ACTIONS(1835),
    [aux_sym_directive_token1] = ACTIONS(1833),
    [aux_sym_tag_token1] = ACTIONS(1835),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1008)] = {
    [sym_attribute] = STATE(1008),
    [sym__lowercase_identifier] = ACTIONS(1785),
    [anon_sym_SEMI_SEMI] = ACTIONS(1787),
    [anon_sym_let] = ACTIONS(1785),
    [anon_sym_and] = ACTIONS(1785),
    [anon_sym_TILDE] = ACTIONS(1785),
    [anon_sym_QMARK] = ACTIONS(1785),
    [anon_sym_LPAREN] = ACTIONS(1787),
    [anon_sym_external] = ACTIONS(1785),
    [anon_sym_type] = ACTIONS(1785),
    [anon_sym_COMMA] = ACTIONS(1787),
    [anon_sym_PIPE] = ACTIONS(1785),
    [anon_sym_DOT] = ACTIONS(1787),
    [anon_sym_LBRACK] = ACTIONS(1785),
    [anon_sym_RBRACK] = ACTIONS(1787),
    [anon_sym_true] = ACTIONS(1785),
    [anon_sym_false] = ACTIONS(1785),
    [anon_sym_LBRACE] = ACTIONS(1785),
    [anon_sym_SEMI] = ACTIONS(1785),
    [anon_sym_exception] = ACTIONS(1785),
    [anon_sym_module] = ACTIONS(1785),
    [anon_sym_open] = ACTIONS(1785),
    [anon_sym_include] = ACTIONS(1785),
    [anon_sym_class] = ACTIONS(1785),
    [anon_sym_val] = ACTIONS(1785),
    [anon_sym_end] = ACTIONS(1785),
    [anon_sym_object] = ACTIONS(1785),
    [anon_sym_in] = ACTIONS(1785),
    [anon_sym_AMP] = ACTIONS(1785),
    [anon_sym_POUND] = ACTIONS(1785),
    [anon_sym_COLON_COLON] = ACTIONS(1787),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1787),
    [anon_sym_else] = ACTIONS(1785),
    [anon_sym_new] = ACTIONS(1785),
    [anon_sym_LBRACE_LT] = ACTIONS(1787),
    [anon_sym_begin] = ACTIONS(1785),
    [sym_ocamlyacc_value] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1785),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1787),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1785),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1785),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1787),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1787),
    [aux_sym_number_token1] = ACTIONS(1787),
    [aux_sym_signed_number_token1] = ACTIONS(1785),
    [anon_sym_SQUOTE] = ACTIONS(1787),
    [anon_sym_DQUOTE] = ACTIONS(1787),
    [sym_prefix_operator] = ACTIONS(1787),
    [aux_sym_sign_operator_token1] = ACTIONS(1785),
    [sym_hash_operator] = ACTIONS(1787),
    [anon_sym_lsl] = ACTIONS(1785),
    [anon_sym_lsr] = ACTIONS(1785),
    [anon_sym_asr] = ACTIONS(1785),
    [aux_sym_pow_operator_token1] = ACTIONS(1787),
    [anon_sym_mod] = ACTIONS(1785),
    [anon_sym_land] = ACTIONS(1785),
    [anon_sym_lor] = ACTIONS(1785),
    [anon_sym_lxor] = ACTIONS(1785),
    [aux_sym_mult_operator_token1] = ACTIONS(1785),
    [aux_sym_add_operator_token1] = ACTIONS(1785),
    [sym_concat_operator] = ACTIONS(1787),
    [sym_rel_operator] = ACTIONS(1785),
    [anon_sym_AMP_AMP] = ACTIONS(1785),
    [anon_sym_or] = ACTIONS(1785),
    [anon_sym_PIPE_PIPE] = ACTIONS(1785),
    [sym_assign_operator] = ACTIONS(1787),
    [sym_let_operator] = ACTIONS(1787),
    [sym_let_and_operator] = ACTIONS(1787),
    [sym__uppercase_identifier] = ACTIONS(1787),
    [aux_sym_directive_token1] = ACTIONS(1785),
    [aux_sym_tag_token1] = ACTIONS(1787),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1009)] = {
    [sym_attribute] = STATE(1009),
    [sym__lowercase_identifier] = ACTIONS(1671),
    [anon_sym_SEMI_SEMI] = ACTIONS(1673),
    [anon_sym_let] = ACTIONS(1671),
    [anon_sym_and] = ACTIONS(1671),
    [anon_sym_TILDE] = ACTIONS(1671),
    [anon_sym_QMARK] = ACTIONS(1671),
    [anon_sym_LPAREN] = ACTIONS(1673),
    [anon_sym_external] = ACTIONS(1671),
    [anon_sym_type] = ACTIONS(1671),
    [anon_sym_COMMA] = ACTIONS(1673),
    [anon_sym_PIPE] = ACTIONS(1671),
    [anon_sym_DOT] = ACTIONS(1673),
    [anon_sym_LBRACK] = ACTIONS(1671),
    [anon_sym_RBRACK] = ACTIONS(1673),
    [anon_sym_true] = ACTIONS(1671),
    [anon_sym_false] = ACTIONS(1671),
    [anon_sym_LBRACE] = ACTIONS(1671),
    [anon_sym_SEMI] = ACTIONS(1671),
    [anon_sym_exception] = ACTIONS(1671),
    [anon_sym_module] = ACTIONS(1671),
    [anon_sym_open] = ACTIONS(1671),
    [anon_sym_include] = ACTIONS(1671),
    [anon_sym_class] = ACTIONS(1671),
    [anon_sym_val] = ACTIONS(1671),
    [anon_sym_end] = ACTIONS(1671),
    [anon_sym_object] = ACTIONS(1671),
    [anon_sym_in] = ACTIONS(1671),
    [anon_sym_AMP] = ACTIONS(1671),
    [anon_sym_POUND] = ACTIONS(1671),
    [anon_sym_COLON_COLON] = ACTIONS(1673),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1673),
    [anon_sym_else] = ACTIONS(1671),
    [anon_sym_new] = ACTIONS(1671),
    [anon_sym_LBRACE_LT] = ACTIONS(1673),
    [anon_sym_begin] = ACTIONS(1671),
    [sym_ocamlyacc_value] = ACTIONS(1673),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1671),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1673),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1671),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1671),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1673),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1673),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1671),
    [anon_sym_SQUOTE] = ACTIONS(1673),
    [anon_sym_DQUOTE] = ACTIONS(1673),
    [sym_prefix_operator] = ACTIONS(1673),
    [aux_sym_sign_operator_token1] = ACTIONS(1671),
    [sym_hash_operator] = ACTIONS(1673),
    [anon_sym_lsl] = ACTIONS(1671),
    [anon_sym_lsr] = ACTIONS(1671),
    [anon_sym_asr] = ACTIONS(1671),
    [aux_sym_pow_operator_token1] = ACTIONS(1673),
    [anon_sym_mod] = ACTIONS(1671),
    [anon_sym_land] = ACTIONS(1671),
    [anon_sym_lor] = ACTIONS(1671),
    [anon_sym_lxor] = ACTIONS(1671),
    [aux_sym_mult_operator_token1] = ACTIONS(1671),
    [aux_sym_add_operator_token1] = ACTIONS(1671),
    [sym_concat_operator] = ACTIONS(1673),
    [sym_rel_operator] = ACTIONS(1671),
    [anon_sym_AMP_AMP] = ACTIONS(1671),
    [anon_sym_or] = ACTIONS(1671),
    [anon_sym_PIPE_PIPE] = ACTIONS(1671),
    [sym_assign_operator] = ACTIONS(1673),
    [sym_let_operator] = ACTIONS(1673),
    [sym_let_and_operator] = ACTIONS(1673),
    [sym__uppercase_identifier] = ACTIONS(1673),
    [aux_sym_directive_token1] = ACTIONS(1671),
    [aux_sym_tag_token1] = ACTIONS(1673),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1010)] = {
    [sym_attribute] = STATE(1010),
    [sym__lowercase_identifier] = ACTIONS(1675),
    [anon_sym_SEMI_SEMI] = ACTIONS(1677),
    [anon_sym_let] = ACTIONS(1675),
    [anon_sym_and] = ACTIONS(1675),
    [anon_sym_TILDE] = ACTIONS(1675),
    [anon_sym_QMARK] = ACTIONS(1675),
    [anon_sym_LPAREN] = ACTIONS(1677),
    [anon_sym_external] = ACTIONS(1675),
    [anon_sym_type] = ACTIONS(1675),
    [anon_sym_COMMA] = ACTIONS(1677),
    [anon_sym_PIPE] = ACTIONS(1675),
    [anon_sym_DOT] = ACTIONS(1677),
    [anon_sym_LBRACK] = ACTIONS(1675),
    [anon_sym_RBRACK] = ACTIONS(1677),
    [anon_sym_true] = ACTIONS(1675),
    [anon_sym_false] = ACTIONS(1675),
    [anon_sym_LBRACE] = ACTIONS(1675),
    [anon_sym_SEMI] = ACTIONS(1675),
    [anon_sym_exception] = ACTIONS(1675),
    [anon_sym_module] = ACTIONS(1675),
    [anon_sym_open] = ACTIONS(1675),
    [anon_sym_include] = ACTIONS(1675),
    [anon_sym_class] = ACTIONS(1675),
    [anon_sym_val] = ACTIONS(1675),
    [anon_sym_end] = ACTIONS(1675),
    [anon_sym_object] = ACTIONS(1675),
    [anon_sym_in] = ACTIONS(1675),
    [anon_sym_AMP] = ACTIONS(1675),
    [anon_sym_POUND] = ACTIONS(1675),
    [anon_sym_COLON_COLON] = ACTIONS(1677),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1677),
    [anon_sym_else] = ACTIONS(1675),
    [anon_sym_new] = ACTIONS(1675),
    [anon_sym_LBRACE_LT] = ACTIONS(1677),
    [anon_sym_begin] = ACTIONS(1675),
    [sym_ocamlyacc_value] = ACTIONS(1677),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1675),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1677),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1675),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1675),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1677),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1677),
    [aux_sym_number_token1] = ACTIONS(1677),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1677),
    [sym_prefix_operator] = ACTIONS(1677),
    [aux_sym_sign_operator_token1] = ACTIONS(1675),
    [sym_hash_operator] = ACTIONS(1677),
    [anon_sym_lsl] = ACTIONS(1675),
    [anon_sym_lsr] = ACTIONS(1675),
    [anon_sym_asr] = ACTIONS(1675),
    [aux_sym_pow_operator_token1] = ACTIONS(1677),
    [anon_sym_mod] = ACTIONS(1675),
    [anon_sym_land] = ACTIONS(1675),
    [anon_sym_lor] = ACTIONS(1675),
    [anon_sym_lxor] = ACTIONS(1675),
    [aux_sym_mult_operator_token1] = ACTIONS(1675),
    [aux_sym_add_operator_token1] = ACTIONS(1675),
    [sym_concat_operator] = ACTIONS(1677),
    [sym_rel_operator] = ACTIONS(1675),
    [anon_sym_AMP_AMP] = ACTIONS(1675),
    [anon_sym_or] = ACTIONS(1675),
    [anon_sym_PIPE_PIPE] = ACTIONS(1675),
    [sym_assign_operator] = ACTIONS(1677),
    [sym_let_operator] = ACTIONS(1677),
    [sym_let_and_operator] = ACTIONS(1677),
    [sym__uppercase_identifier] = ACTIONS(1677),
    [aux_sym_directive_token1] = ACTIONS(1675),
    [aux_sym_tag_token1] = ACTIONS(1677),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1011)] = {
    [sym_attribute] = STATE(1011),
    [sym__lowercase_identifier] = ACTIONS(1679),
    [anon_sym_SEMI_SEMI] = ACTIONS(1681),
    [anon_sym_let] = ACTIONS(1679),
    [anon_sym_and] = ACTIONS(1679),
    [anon_sym_TILDE] = ACTIONS(1679),
    [anon_sym_QMARK] = ACTIONS(1679),
    [anon_sym_LPAREN] = ACTIONS(1681),
    [anon_sym_external] = ACTIONS(1679),
    [anon_sym_type] = ACTIONS(1679),
    [anon_sym_COMMA] = ACTIONS(1681),
    [anon_sym_PIPE] = ACTIONS(1679),
    [anon_sym_DOT] = ACTIONS(1681),
    [anon_sym_LBRACK] = ACTIONS(1679),
    [anon_sym_RBRACK] = ACTIONS(1681),
    [anon_sym_true] = ACTIONS(1679),
    [anon_sym_false] = ACTIONS(1679),
    [anon_sym_LBRACE] = ACTIONS(1679),
    [anon_sym_SEMI] = ACTIONS(1679),
    [anon_sym_exception] = ACTIONS(1679),
    [anon_sym_module] = ACTIONS(1679),
    [anon_sym_open] = ACTIONS(1679),
    [anon_sym_include] = ACTIONS(1679),
    [anon_sym_class] = ACTIONS(1679),
    [anon_sym_val] = ACTIONS(1679),
    [anon_sym_end] = ACTIONS(1679),
    [anon_sym_object] = ACTIONS(1679),
    [anon_sym_in] = ACTIONS(1679),
    [anon_sym_AMP] = ACTIONS(1679),
    [anon_sym_POUND] = ACTIONS(1679),
    [anon_sym_COLON_COLON] = ACTIONS(1681),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1681),
    [anon_sym_else] = ACTIONS(1679),
    [anon_sym_new] = ACTIONS(1679),
    [anon_sym_LBRACE_LT] = ACTIONS(1681),
    [anon_sym_begin] = ACTIONS(1679),
    [sym_ocamlyacc_value] = ACTIONS(1681),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1679),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1681),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1679),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1679),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1681),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1681),
    [aux_sym_number_token1] = ACTIONS(1681),
    [aux_sym_signed_number_token1] = ACTIONS(1679),
    [anon_sym_SQUOTE] = ACTIONS(1681),
    [anon_sym_DQUOTE] = ACTIONS(1681),
    [sym_prefix_operator] = ACTIONS(1681),
    [aux_sym_sign_operator_token1] = ACTIONS(1679),
    [sym_hash_operator] = ACTIONS(1681),
    [anon_sym_lsl] = ACTIONS(1679),
    [anon_sym_lsr] = ACTIONS(1679),
    [anon_sym_asr] = ACTIONS(1679),
    [aux_sym_pow_operator_token1] = ACTIONS(1681),
    [anon_sym_mod] = ACTIONS(1679),
    [anon_sym_land] = ACTIONS(1679),
    [anon_sym_lor] = ACTIONS(1679),
    [anon_sym_lxor] = ACTIONS(1679),
    [aux_sym_mult_operator_token1] = ACTIONS(1679),
    [aux_sym_add_operator_token1] = ACTIONS(1679),
    [sym_concat_operator] = ACTIONS(1681),
    [sym_rel_operator] = ACTIONS(1679),
    [anon_sym_AMP_AMP] = ACTIONS(1679),
    [anon_sym_or] = ACTIONS(1679),
    [anon_sym_PIPE_PIPE] = ACTIONS(1679),
    [sym_assign_operator] = ACTIONS(1681),
    [sym_let_operator] = ACTIONS(1681),
    [sym_let_and_operator] = ACTIONS(1681),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_directive_token1] = ACTIONS(1679),
    [aux_sym_tag_token1] = ACTIONS(1681),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1012)] = {
    [sym_attribute] = STATE(1012),
    [sym__lowercase_identifier] = ACTIONS(1695),
    [anon_sym_SEMI_SEMI] = ACTIONS(1697),
    [anon_sym_let] = ACTIONS(1695),
    [anon_sym_and] = ACTIONS(1695),
    [anon_sym_TILDE] = ACTIONS(1695),
    [anon_sym_QMARK] = ACTIONS(1695),
    [anon_sym_LPAREN] = ACTIONS(1697),
    [anon_sym_external] = ACTIONS(1695),
    [anon_sym_type] = ACTIONS(1695),
    [anon_sym_COMMA] = ACTIONS(1697),
    [anon_sym_PIPE] = ACTIONS(1695),
    [anon_sym_DOT] = ACTIONS(1697),
    [anon_sym_LBRACK] = ACTIONS(1695),
    [anon_sym_RBRACK] = ACTIONS(1697),
    [anon_sym_true] = ACTIONS(1695),
    [anon_sym_false] = ACTIONS(1695),
    [anon_sym_LBRACE] = ACTIONS(1695),
    [anon_sym_SEMI] = ACTIONS(1695),
    [anon_sym_exception] = ACTIONS(1695),
    [anon_sym_module] = ACTIONS(1695),
    [anon_sym_open] = ACTIONS(1695),
    [anon_sym_include] = ACTIONS(1695),
    [anon_sym_class] = ACTIONS(1695),
    [anon_sym_val] = ACTIONS(1695),
    [anon_sym_end] = ACTIONS(1695),
    [anon_sym_object] = ACTIONS(1695),
    [anon_sym_in] = ACTIONS(1695),
    [anon_sym_AMP] = ACTIONS(1695),
    [anon_sym_POUND] = ACTIONS(1695),
    [anon_sym_COLON_COLON] = ACTIONS(1697),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1697),
    [anon_sym_else] = ACTIONS(1695),
    [anon_sym_new] = ACTIONS(1695),
    [anon_sym_LBRACE_LT] = ACTIONS(1697),
    [anon_sym_begin] = ACTIONS(1695),
    [sym_ocamlyacc_value] = ACTIONS(1697),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1695),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1697),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1695),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1695),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1697),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1697),
    [aux_sym_number_token1] = ACTIONS(1697),
    [aux_sym_signed_number_token1] = ACTIONS(1695),
    [anon_sym_SQUOTE] = ACTIONS(1697),
    [anon_sym_DQUOTE] = ACTIONS(1697),
    [sym_prefix_operator] = ACTIONS(1697),
    [aux_sym_sign_operator_token1] = ACTIONS(1695),
    [sym_hash_operator] = ACTIONS(1697),
    [anon_sym_lsl] = ACTIONS(1695),
    [anon_sym_lsr] = ACTIONS(1695),
    [anon_sym_asr] = ACTIONS(1695),
    [aux_sym_pow_operator_token1] = ACTIONS(1697),
    [anon_sym_mod] = ACTIONS(1695),
    [anon_sym_land] = ACTIONS(1695),
    [anon_sym_lor] = ACTIONS(1695),
    [anon_sym_lxor] = ACTIONS(1695),
    [aux_sym_mult_operator_token1] = ACTIONS(1695),
    [aux_sym_add_operator_token1] = ACTIONS(1695),
    [sym_concat_operator] = ACTIONS(1697),
    [sym_rel_operator] = ACTIONS(1695),
    [anon_sym_AMP_AMP] = ACTIONS(1695),
    [anon_sym_or] = ACTIONS(1695),
    [anon_sym_PIPE_PIPE] = ACTIONS(1695),
    [sym_assign_operator] = ACTIONS(1697),
    [sym_let_operator] = ACTIONS(1697),
    [sym_let_and_operator] = ACTIONS(1697),
    [sym__uppercase_identifier] = ACTIONS(1697),
    [aux_sym_directive_token1] = ACTIONS(1695),
    [aux_sym_tag_token1] = ACTIONS(1697),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1013)] = {
    [sym_attribute] = STATE(1013),
    [sym__lowercase_identifier] = ACTIONS(1797),
    [anon_sym_SEMI_SEMI] = ACTIONS(1799),
    [anon_sym_let] = ACTIONS(1797),
    [anon_sym_and] = ACTIONS(1797),
    [anon_sym_TILDE] = ACTIONS(1797),
    [anon_sym_QMARK] = ACTIONS(1797),
    [anon_sym_LPAREN] = ACTIONS(1799),
    [anon_sym_external] = ACTIONS(1797),
    [anon_sym_type] = ACTIONS(1797),
    [anon_sym_COMMA] = ACTIONS(1799),
    [anon_sym_PIPE] = ACTIONS(1797),
    [anon_sym_DOT] = ACTIONS(1799),
    [anon_sym_LBRACK] = ACTIONS(1797),
    [anon_sym_RBRACK] = ACTIONS(1799),
    [anon_sym_true] = ACTIONS(1797),
    [anon_sym_false] = ACTIONS(1797),
    [anon_sym_LBRACE] = ACTIONS(1797),
    [anon_sym_SEMI] = ACTIONS(1797),
    [anon_sym_exception] = ACTIONS(1797),
    [anon_sym_module] = ACTIONS(1797),
    [anon_sym_open] = ACTIONS(1797),
    [anon_sym_include] = ACTIONS(1797),
    [anon_sym_class] = ACTIONS(1797),
    [anon_sym_val] = ACTIONS(1797),
    [anon_sym_end] = ACTIONS(1797),
    [anon_sym_object] = ACTIONS(1797),
    [anon_sym_in] = ACTIONS(1797),
    [anon_sym_AMP] = ACTIONS(1797),
    [anon_sym_POUND] = ACTIONS(1797),
    [anon_sym_COLON_COLON] = ACTIONS(1799),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1799),
    [anon_sym_else] = ACTIONS(1797),
    [anon_sym_new] = ACTIONS(1797),
    [anon_sym_LBRACE_LT] = ACTIONS(1799),
    [anon_sym_begin] = ACTIONS(1797),
    [sym_ocamlyacc_value] = ACTIONS(1799),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1797),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1799),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1797),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1797),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1799),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1799),
    [aux_sym_signed_number_token1] = ACTIONS(1797),
    [anon_sym_SQUOTE] = ACTIONS(1799),
    [anon_sym_DQUOTE] = ACTIONS(1799),
    [sym_prefix_operator] = ACTIONS(1799),
    [aux_sym_sign_operator_token1] = ACTIONS(1797),
    [sym_hash_operator] = ACTIONS(1799),
    [anon_sym_lsl] = ACTIONS(1797),
    [anon_sym_lsr] = ACTIONS(1797),
    [anon_sym_asr] = ACTIONS(1797),
    [aux_sym_pow_operator_token1] = ACTIONS(1799),
    [anon_sym_mod] = ACTIONS(1797),
    [anon_sym_land] = ACTIONS(1797),
    [anon_sym_lor] = ACTIONS(1797),
    [anon_sym_lxor] = ACTIONS(1797),
    [aux_sym_mult_operator_token1] = ACTIONS(1797),
    [aux_sym_add_operator_token1] = ACTIONS(1797),
    [sym_concat_operator] = ACTIONS(1799),
    [sym_rel_operator] = ACTIONS(1797),
    [anon_sym_AMP_AMP] = ACTIONS(1797),
    [anon_sym_or] = ACTIONS(1797),
    [anon_sym_PIPE_PIPE] = ACTIONS(1797),
    [sym_assign_operator] = ACTIONS(1799),
    [sym_let_operator] = ACTIONS(1799),
    [sym_let_and_operator] = ACTIONS(1799),
    [sym__uppercase_identifier] = ACTIONS(1799),
    [aux_sym_directive_token1] = ACTIONS(1797),
    [aux_sym_tag_token1] = ACTIONS(1799),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1014)] = {
    [sym_attribute] = STATE(1014),
    [sym__lowercase_identifier] = ACTIONS(1851),
    [anon_sym_SEMI_SEMI] = ACTIONS(1853),
    [anon_sym_let] = ACTIONS(1851),
    [anon_sym_and] = ACTIONS(1851),
    [anon_sym_TILDE] = ACTIONS(1851),
    [anon_sym_QMARK] = ACTIONS(1851),
    [anon_sym_LPAREN] = ACTIONS(1853),
    [anon_sym_external] = ACTIONS(1851),
    [anon_sym_type] = ACTIONS(1851),
    [anon_sym_COMMA] = ACTIONS(1853),
    [anon_sym_PIPE] = ACTIONS(1851),
    [anon_sym_DOT] = ACTIONS(1853),
    [anon_sym_LBRACK] = ACTIONS(1851),
    [anon_sym_RBRACK] = ACTIONS(1853),
    [anon_sym_true] = ACTIONS(1851),
    [anon_sym_false] = ACTIONS(1851),
    [anon_sym_LBRACE] = ACTIONS(1851),
    [anon_sym_SEMI] = ACTIONS(1851),
    [anon_sym_exception] = ACTIONS(1851),
    [anon_sym_module] = ACTIONS(1851),
    [anon_sym_open] = ACTIONS(1851),
    [anon_sym_include] = ACTIONS(1851),
    [anon_sym_class] = ACTIONS(1851),
    [anon_sym_val] = ACTIONS(1851),
    [anon_sym_end] = ACTIONS(1851),
    [anon_sym_object] = ACTIONS(1851),
    [anon_sym_in] = ACTIONS(1851),
    [anon_sym_AMP] = ACTIONS(1851),
    [anon_sym_POUND] = ACTIONS(1851),
    [anon_sym_COLON_COLON] = ACTIONS(1853),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1853),
    [anon_sym_else] = ACTIONS(1851),
    [anon_sym_new] = ACTIONS(1851),
    [anon_sym_LBRACE_LT] = ACTIONS(1853),
    [anon_sym_begin] = ACTIONS(1851),
    [sym_ocamlyacc_value] = ACTIONS(1853),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1851),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1853),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1851),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1851),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1853),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1853),
    [aux_sym_number_token1] = ACTIONS(1853),
    [aux_sym_signed_number_token1] = ACTIONS(1851),
    [anon_sym_SQUOTE] = ACTIONS(1853),
    [anon_sym_DQUOTE] = ACTIONS(1853),
    [sym_prefix_operator] = ACTIONS(1853),
    [aux_sym_sign_operator_token1] = ACTIONS(1851),
    [sym_hash_operator] = ACTIONS(1853),
    [anon_sym_lsl] = ACTIONS(1851),
    [anon_sym_lsr] = ACTIONS(1851),
    [anon_sym_asr] = ACTIONS(1851),
    [aux_sym_pow_operator_token1] = ACTIONS(1853),
    [anon_sym_mod] = ACTIONS(1851),
    [anon_sym_land] = ACTIONS(1851),
    [anon_sym_lor] = ACTIONS(1851),
    [anon_sym_lxor] = ACTIONS(1851),
    [aux_sym_mult_operator_token1] = ACTIONS(1851),
    [aux_sym_add_operator_token1] = ACTIONS(1851),
    [sym_concat_operator] = ACTIONS(1853),
    [sym_rel_operator] = ACTIONS(1851),
    [anon_sym_AMP_AMP] = ACTIONS(1851),
    [anon_sym_or] = ACTIONS(1851),
    [anon_sym_PIPE_PIPE] = ACTIONS(1851),
    [sym_assign_operator] = ACTIONS(1853),
    [sym_let_operator] = ACTIONS(1853),
    [sym_let_and_operator] = ACTIONS(1853),
    [sym__uppercase_identifier] = ACTIONS(1853),
    [aux_sym_directive_token1] = ACTIONS(1851),
    [aux_sym_tag_token1] = ACTIONS(1853),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1015)] = {
    [sym_attribute] = STATE(1015),
    [sym__lowercase_identifier] = ACTIONS(1717),
    [anon_sym_SEMI_SEMI] = ACTIONS(1719),
    [anon_sym_let] = ACTIONS(1717),
    [anon_sym_and] = ACTIONS(1717),
    [anon_sym_TILDE] = ACTIONS(1717),
    [anon_sym_QMARK] = ACTIONS(1717),
    [anon_sym_LPAREN] = ACTIONS(1719),
    [anon_sym_external] = ACTIONS(1717),
    [anon_sym_type] = ACTIONS(1717),
    [anon_sym_COMMA] = ACTIONS(1719),
    [anon_sym_PIPE] = ACTIONS(1717),
    [anon_sym_DOT] = ACTIONS(1719),
    [anon_sym_LBRACK] = ACTIONS(1717),
    [anon_sym_RBRACK] = ACTIONS(1719),
    [anon_sym_true] = ACTIONS(1717),
    [anon_sym_false] = ACTIONS(1717),
    [anon_sym_LBRACE] = ACTIONS(1717),
    [anon_sym_SEMI] = ACTIONS(1717),
    [anon_sym_exception] = ACTIONS(1717),
    [anon_sym_module] = ACTIONS(1717),
    [anon_sym_open] = ACTIONS(1717),
    [anon_sym_include] = ACTIONS(1717),
    [anon_sym_class] = ACTIONS(1717),
    [anon_sym_val] = ACTIONS(1717),
    [anon_sym_end] = ACTIONS(1717),
    [anon_sym_object] = ACTIONS(1717),
    [anon_sym_in] = ACTIONS(1717),
    [anon_sym_AMP] = ACTIONS(1717),
    [anon_sym_POUND] = ACTIONS(1717),
    [anon_sym_COLON_COLON] = ACTIONS(1719),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1719),
    [anon_sym_else] = ACTIONS(1717),
    [anon_sym_new] = ACTIONS(1717),
    [anon_sym_LBRACE_LT] = ACTIONS(1719),
    [anon_sym_begin] = ACTIONS(1717),
    [sym_ocamlyacc_value] = ACTIONS(1719),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1717),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1719),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1717),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1717),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1719),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1719),
    [aux_sym_number_token1] = ACTIONS(1719),
    [aux_sym_signed_number_token1] = ACTIONS(1717),
    [anon_sym_SQUOTE] = ACTIONS(1719),
    [anon_sym_DQUOTE] = ACTIONS(1719),
    [sym_prefix_operator] = ACTIONS(1719),
    [aux_sym_sign_operator_token1] = ACTIONS(1717),
    [sym_hash_operator] = ACTIONS(1719),
    [anon_sym_lsl] = ACTIONS(1717),
    [anon_sym_lsr] = ACTIONS(1717),
    [anon_sym_asr] = ACTIONS(1717),
    [aux_sym_pow_operator_token1] = ACTIONS(1719),
    [anon_sym_mod] = ACTIONS(1717),
    [anon_sym_land] = ACTIONS(1717),
    [anon_sym_lor] = ACTIONS(1717),
    [anon_sym_lxor] = ACTIONS(1717),
    [aux_sym_mult_operator_token1] = ACTIONS(1717),
    [aux_sym_add_operator_token1] = ACTIONS(1717),
    [sym_concat_operator] = ACTIONS(1719),
    [sym_rel_operator] = ACTIONS(1717),
    [anon_sym_AMP_AMP] = ACTIONS(1717),
    [anon_sym_or] = ACTIONS(1717),
    [anon_sym_PIPE_PIPE] = ACTIONS(1717),
    [sym_assign_operator] = ACTIONS(1719),
    [sym_let_operator] = ACTIONS(1719),
    [sym_let_and_operator] = ACTIONS(1719),
    [sym__uppercase_identifier] = ACTIONS(1719),
    [aux_sym_directive_token1] = ACTIONS(1717),
    [aux_sym_tag_token1] = ACTIONS(1719),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1016)] = {
    [sym_attribute] = STATE(1016),
    [sym__lowercase_identifier] = ACTIONS(1837),
    [anon_sym_SEMI_SEMI] = ACTIONS(1839),
    [anon_sym_let] = ACTIONS(1837),
    [anon_sym_and] = ACTIONS(1837),
    [anon_sym_TILDE] = ACTIONS(1837),
    [anon_sym_QMARK] = ACTIONS(1837),
    [anon_sym_LPAREN] = ACTIONS(1839),
    [anon_sym_external] = ACTIONS(1837),
    [anon_sym_type] = ACTIONS(1837),
    [anon_sym_COMMA] = ACTIONS(1839),
    [anon_sym_PIPE] = ACTIONS(1837),
    [anon_sym_DOT] = ACTIONS(1839),
    [anon_sym_LBRACK] = ACTIONS(1837),
    [anon_sym_RBRACK] = ACTIONS(1839),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1837),
    [anon_sym_SEMI] = ACTIONS(1837),
    [anon_sym_exception] = ACTIONS(1837),
    [anon_sym_module] = ACTIONS(1837),
    [anon_sym_open] = ACTIONS(1837),
    [anon_sym_include] = ACTIONS(1837),
    [anon_sym_class] = ACTIONS(1837),
    [anon_sym_val] = ACTIONS(1837),
    [anon_sym_end] = ACTIONS(1837),
    [anon_sym_object] = ACTIONS(1837),
    [anon_sym_in] = ACTIONS(1837),
    [anon_sym_AMP] = ACTIONS(1837),
    [anon_sym_POUND] = ACTIONS(1837),
    [anon_sym_COLON_COLON] = ACTIONS(1839),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1839),
    [anon_sym_else] = ACTIONS(1837),
    [anon_sym_new] = ACTIONS(1837),
    [anon_sym_LBRACE_LT] = ACTIONS(1839),
    [anon_sym_begin] = ACTIONS(1837),
    [sym_ocamlyacc_value] = ACTIONS(1839),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1837),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1839),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1837),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1837),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1839),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1839),
    [aux_sym_number_token1] = ACTIONS(1839),
    [aux_sym_signed_number_token1] = ACTIONS(1837),
    [anon_sym_SQUOTE] = ACTIONS(1839),
    [anon_sym_DQUOTE] = ACTIONS(1839),
    [sym_prefix_operator] = ACTIONS(1839),
    [aux_sym_sign_operator_token1] = ACTIONS(1837),
    [sym_hash_operator] = ACTIONS(1839),
    [anon_sym_lsl] = ACTIONS(1837),
    [anon_sym_lsr] = ACTIONS(1837),
    [anon_sym_asr] = ACTIONS(1837),
    [aux_sym_pow_operator_token1] = ACTIONS(1839),
    [anon_sym_mod] = ACTIONS(1837),
    [anon_sym_land] = ACTIONS(1837),
    [anon_sym_lor] = ACTIONS(1837),
    [anon_sym_lxor] = ACTIONS(1837),
    [aux_sym_mult_operator_token1] = ACTIONS(1837),
    [aux_sym_add_operator_token1] = ACTIONS(1837),
    [sym_concat_operator] = ACTIONS(1839),
    [sym_rel_operator] = ACTIONS(1837),
    [anon_sym_AMP_AMP] = ACTIONS(1837),
    [anon_sym_or] = ACTIONS(1837),
    [anon_sym_PIPE_PIPE] = ACTIONS(1837),
    [sym_assign_operator] = ACTIONS(1839),
    [sym_let_operator] = ACTIONS(1839),
    [sym_let_and_operator] = ACTIONS(1839),
    [sym__uppercase_identifier] = ACTIONS(1839),
    [aux_sym_directive_token1] = ACTIONS(1837),
    [aux_sym_tag_token1] = ACTIONS(1839),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1017)] = {
    [sym_attribute] = STATE(1017),
    [sym__lowercase_identifier] = ACTIONS(1571),
    [anon_sym_SEMI_SEMI] = ACTIONS(1573),
    [anon_sym_let] = ACTIONS(1571),
    [anon_sym_and] = ACTIONS(1571),
    [anon_sym_TILDE] = ACTIONS(1571),
    [anon_sym_QMARK] = ACTIONS(1571),
    [anon_sym_LPAREN] = ACTIONS(1573),
    [anon_sym_external] = ACTIONS(1571),
    [anon_sym_type] = ACTIONS(1571),
    [anon_sym_COMMA] = ACTIONS(1573),
    [anon_sym_PIPE] = ACTIONS(1571),
    [anon_sym_DOT] = ACTIONS(1573),
    [anon_sym_LBRACK] = ACTIONS(1571),
    [anon_sym_RBRACK] = ACTIONS(1573),
    [anon_sym_true] = ACTIONS(1571),
    [anon_sym_false] = ACTIONS(1571),
    [anon_sym_LBRACE] = ACTIONS(1571),
    [anon_sym_SEMI] = ACTIONS(1571),
    [anon_sym_exception] = ACTIONS(1571),
    [anon_sym_module] = ACTIONS(1571),
    [anon_sym_open] = ACTIONS(1571),
    [anon_sym_include] = ACTIONS(1571),
    [anon_sym_class] = ACTIONS(1571),
    [anon_sym_val] = ACTIONS(1571),
    [anon_sym_end] = ACTIONS(1571),
    [anon_sym_object] = ACTIONS(1571),
    [anon_sym_in] = ACTIONS(1571),
    [anon_sym_AMP] = ACTIONS(1571),
    [anon_sym_POUND] = ACTIONS(1571),
    [anon_sym_COLON_COLON] = ACTIONS(1573),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1573),
    [anon_sym_else] = ACTIONS(1571),
    [anon_sym_new] = ACTIONS(1571),
    [anon_sym_LBRACE_LT] = ACTIONS(1573),
    [anon_sym_begin] = ACTIONS(1571),
    [sym_ocamlyacc_value] = ACTIONS(1573),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1571),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1573),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1571),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1571),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1573),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1573),
    [aux_sym_number_token1] = ACTIONS(1573),
    [aux_sym_signed_number_token1] = ACTIONS(1571),
    [anon_sym_SQUOTE] = ACTIONS(1573),
    [anon_sym_DQUOTE] = ACTIONS(1573),
    [sym_prefix_operator] = ACTIONS(1573),
    [aux_sym_sign_operator_token1] = ACTIONS(1571),
    [sym_hash_operator] = ACTIONS(1573),
    [anon_sym_lsl] = ACTIONS(1571),
    [anon_sym_lsr] = ACTIONS(1571),
    [anon_sym_asr] = ACTIONS(1571),
    [aux_sym_pow_operator_token1] = ACTIONS(1573),
    [anon_sym_mod] = ACTIONS(1571),
    [anon_sym_land] = ACTIONS(1571),
    [anon_sym_lor] = ACTIONS(1571),
    [anon_sym_lxor] = ACTIONS(1571),
    [aux_sym_mult_operator_token1] = ACTIONS(1571),
    [aux_sym_add_operator_token1] = ACTIONS(1571),
    [sym_concat_operator] = ACTIONS(1573),
    [sym_rel_operator] = ACTIONS(1571),
    [anon_sym_AMP_AMP] = ACTIONS(1571),
    [anon_sym_or] = ACTIONS(1571),
    [anon_sym_PIPE_PIPE] = ACTIONS(1571),
    [sym_assign_operator] = ACTIONS(1573),
    [sym_let_operator] = ACTIONS(1573),
    [sym_let_and_operator] = ACTIONS(1573),
    [sym__uppercase_identifier] = ACTIONS(1573),
    [aux_sym_directive_token1] = ACTIONS(1571),
    [aux_sym_tag_token1] = ACTIONS(1573),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1018)] = {
    [sym_attribute] = STATE(1018),
    [sym__lowercase_identifier] = ACTIONS(1805),
    [anon_sym_SEMI_SEMI] = ACTIONS(1807),
    [anon_sym_let] = ACTIONS(1805),
    [anon_sym_and] = ACTIONS(1805),
    [anon_sym_TILDE] = ACTIONS(1805),
    [anon_sym_QMARK] = ACTIONS(1805),
    [anon_sym_LPAREN] = ACTIONS(1807),
    [anon_sym_external] = ACTIONS(1805),
    [anon_sym_type] = ACTIONS(1805),
    [anon_sym_COMMA] = ACTIONS(1807),
    [anon_sym_PIPE] = ACTIONS(1805),
    [anon_sym_DOT] = ACTIONS(1807),
    [anon_sym_LBRACK] = ACTIONS(1805),
    [anon_sym_RBRACK] = ACTIONS(1807),
    [anon_sym_true] = ACTIONS(1805),
    [anon_sym_false] = ACTIONS(1805),
    [anon_sym_LBRACE] = ACTIONS(1805),
    [anon_sym_SEMI] = ACTIONS(1805),
    [anon_sym_exception] = ACTIONS(1805),
    [anon_sym_module] = ACTIONS(1805),
    [anon_sym_open] = ACTIONS(1805),
    [anon_sym_include] = ACTIONS(1805),
    [anon_sym_class] = ACTIONS(1805),
    [anon_sym_val] = ACTIONS(1805),
    [anon_sym_end] = ACTIONS(1805),
    [anon_sym_object] = ACTIONS(1805),
    [anon_sym_in] = ACTIONS(1805),
    [anon_sym_AMP] = ACTIONS(1805),
    [anon_sym_POUND] = ACTIONS(1805),
    [anon_sym_COLON_COLON] = ACTIONS(1807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1807),
    [anon_sym_else] = ACTIONS(1805),
    [anon_sym_new] = ACTIONS(1805),
    [anon_sym_LBRACE_LT] = ACTIONS(1807),
    [anon_sym_begin] = ACTIONS(1805),
    [sym_ocamlyacc_value] = ACTIONS(1807),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1805),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1807),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1805),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1805),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1807),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1807),
    [aux_sym_number_token1] = ACTIONS(1807),
    [aux_sym_signed_number_token1] = ACTIONS(1805),
    [anon_sym_SQUOTE] = ACTIONS(1807),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym_prefix_operator] = ACTIONS(1807),
    [aux_sym_sign_operator_token1] = ACTIONS(1805),
    [sym_hash_operator] = ACTIONS(1807),
    [anon_sym_lsl] = ACTIONS(1805),
    [anon_sym_lsr] = ACTIONS(1805),
    [anon_sym_asr] = ACTIONS(1805),
    [aux_sym_pow_operator_token1] = ACTIONS(1807),
    [anon_sym_mod] = ACTIONS(1805),
    [anon_sym_land] = ACTIONS(1805),
    [anon_sym_lor] = ACTIONS(1805),
    [anon_sym_lxor] = ACTIONS(1805),
    [aux_sym_mult_operator_token1] = ACTIONS(1805),
    [aux_sym_add_operator_token1] = ACTIONS(1805),
    [sym_concat_operator] = ACTIONS(1807),
    [sym_rel_operator] = ACTIONS(1805),
    [anon_sym_AMP_AMP] = ACTIONS(1805),
    [anon_sym_or] = ACTIONS(1805),
    [anon_sym_PIPE_PIPE] = ACTIONS(1805),
    [sym_assign_operator] = ACTIONS(1807),
    [sym_let_operator] = ACTIONS(1807),
    [sym_let_and_operator] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1807),
    [aux_sym_directive_token1] = ACTIONS(1805),
    [aux_sym_tag_token1] = ACTIONS(1807),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1019)] = {
    [sym_attribute] = STATE(1019),
    [sym__lowercase_identifier] = ACTIONS(1579),
    [anon_sym_SEMI_SEMI] = ACTIONS(1581),
    [anon_sym_let] = ACTIONS(1579),
    [anon_sym_and] = ACTIONS(1579),
    [anon_sym_TILDE] = ACTIONS(1579),
    [anon_sym_QMARK] = ACTIONS(1579),
    [anon_sym_LPAREN] = ACTIONS(1581),
    [anon_sym_external] = ACTIONS(1579),
    [anon_sym_type] = ACTIONS(1579),
    [anon_sym_COMMA] = ACTIONS(1581),
    [anon_sym_PIPE] = ACTIONS(1579),
    [anon_sym_DOT] = ACTIONS(1581),
    [anon_sym_LBRACK] = ACTIONS(1579),
    [anon_sym_RBRACK] = ACTIONS(1581),
    [anon_sym_true] = ACTIONS(1579),
    [anon_sym_false] = ACTIONS(1579),
    [anon_sym_LBRACE] = ACTIONS(1579),
    [anon_sym_SEMI] = ACTIONS(1579),
    [anon_sym_exception] = ACTIONS(1579),
    [anon_sym_module] = ACTIONS(1579),
    [anon_sym_open] = ACTIONS(1579),
    [anon_sym_include] = ACTIONS(1579),
    [anon_sym_class] = ACTIONS(1579),
    [anon_sym_val] = ACTIONS(1579),
    [anon_sym_end] = ACTIONS(1579),
    [anon_sym_object] = ACTIONS(1579),
    [anon_sym_in] = ACTIONS(1579),
    [anon_sym_AMP] = ACTIONS(1579),
    [anon_sym_POUND] = ACTIONS(1579),
    [anon_sym_COLON_COLON] = ACTIONS(1581),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1581),
    [anon_sym_else] = ACTIONS(1579),
    [anon_sym_new] = ACTIONS(1579),
    [anon_sym_LBRACE_LT] = ACTIONS(1581),
    [anon_sym_begin] = ACTIONS(1579),
    [sym_ocamlyacc_value] = ACTIONS(1581),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1579),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1581),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1579),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1579),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1581),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1581),
    [aux_sym_number_token1] = ACTIONS(1581),
    [aux_sym_signed_number_token1] = ACTIONS(1579),
    [anon_sym_SQUOTE] = ACTIONS(1581),
    [anon_sym_DQUOTE] = ACTIONS(1581),
    [sym_prefix_operator] = ACTIONS(1581),
    [aux_sym_sign_operator_token1] = ACTIONS(1579),
    [sym_hash_operator] = ACTIONS(1581),
    [anon_sym_lsl] = ACTIONS(1579),
    [anon_sym_lsr] = ACTIONS(1579),
    [anon_sym_asr] = ACTIONS(1579),
    [aux_sym_pow_operator_token1] = ACTIONS(1581),
    [anon_sym_mod] = ACTIONS(1579),
    [anon_sym_land] = ACTIONS(1579),
    [anon_sym_lor] = ACTIONS(1579),
    [anon_sym_lxor] = ACTIONS(1579),
    [aux_sym_mult_operator_token1] = ACTIONS(1579),
    [aux_sym_add_operator_token1] = ACTIONS(1579),
    [sym_concat_operator] = ACTIONS(1581),
    [sym_rel_operator] = ACTIONS(1579),
    [anon_sym_AMP_AMP] = ACTIONS(1579),
    [anon_sym_or] = ACTIONS(1579),
    [anon_sym_PIPE_PIPE] = ACTIONS(1579),
    [sym_assign_operator] = ACTIONS(1581),
    [sym_let_operator] = ACTIONS(1581),
    [sym_let_and_operator] = ACTIONS(1581),
    [sym__uppercase_identifier] = ACTIONS(1581),
    [aux_sym_directive_token1] = ACTIONS(1579),
    [aux_sym_tag_token1] = ACTIONS(1581),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1020)] = {
    [sym_attribute] = STATE(1020),
    [sym__lowercase_identifier] = ACTIONS(1583),
    [anon_sym_SEMI_SEMI] = ACTIONS(1585),
    [anon_sym_let] = ACTIONS(1583),
    [anon_sym_and] = ACTIONS(1583),
    [anon_sym_TILDE] = ACTIONS(1583),
    [anon_sym_QMARK] = ACTIONS(1583),
    [anon_sym_LPAREN] = ACTIONS(1585),
    [anon_sym_external] = ACTIONS(1583),
    [anon_sym_type] = ACTIONS(1583),
    [anon_sym_COMMA] = ACTIONS(1585),
    [anon_sym_PIPE] = ACTIONS(1583),
    [anon_sym_DOT] = ACTIONS(835),
    [anon_sym_LBRACK] = ACTIONS(1583),
    [anon_sym_RBRACK] = ACTIONS(1585),
    [anon_sym_true] = ACTIONS(1583),
    [anon_sym_false] = ACTIONS(1583),
    [anon_sym_LBRACE] = ACTIONS(1583),
    [anon_sym_SEMI] = ACTIONS(1583),
    [anon_sym_exception] = ACTIONS(1583),
    [anon_sym_module] = ACTIONS(1583),
    [anon_sym_open] = ACTIONS(1583),
    [anon_sym_include] = ACTIONS(1583),
    [anon_sym_class] = ACTIONS(1583),
    [anon_sym_val] = ACTIONS(1583),
    [anon_sym_end] = ACTIONS(1583),
    [anon_sym_object] = ACTIONS(1583),
    [anon_sym_in] = ACTIONS(1583),
    [anon_sym_AMP] = ACTIONS(1583),
    [anon_sym_POUND] = ACTIONS(1583),
    [anon_sym_COLON_COLON] = ACTIONS(1585),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1585),
    [anon_sym_else] = ACTIONS(1583),
    [anon_sym_new] = ACTIONS(1583),
    [anon_sym_LBRACE_LT] = ACTIONS(1585),
    [anon_sym_begin] = ACTIONS(1583),
    [sym_ocamlyacc_value] = ACTIONS(1585),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1583),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1585),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1583),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1583),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1585),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1585),
    [aux_sym_number_token1] = ACTIONS(1585),
    [aux_sym_signed_number_token1] = ACTIONS(1583),
    [anon_sym_SQUOTE] = ACTIONS(1585),
    [anon_sym_DQUOTE] = ACTIONS(1585),
    [sym_prefix_operator] = ACTIONS(1585),
    [aux_sym_sign_operator_token1] = ACTIONS(1583),
    [sym_hash_operator] = ACTIONS(1585),
    [anon_sym_lsl] = ACTIONS(1583),
    [anon_sym_lsr] = ACTIONS(1583),
    [anon_sym_asr] = ACTIONS(1583),
    [aux_sym_pow_operator_token1] = ACTIONS(1585),
    [anon_sym_mod] = ACTIONS(1583),
    [anon_sym_land] = ACTIONS(1583),
    [anon_sym_lor] = ACTIONS(1583),
    [anon_sym_lxor] = ACTIONS(1583),
    [aux_sym_mult_operator_token1] = ACTIONS(1583),
    [aux_sym_add_operator_token1] = ACTIONS(1583),
    [sym_concat_operator] = ACTIONS(1585),
    [sym_rel_operator] = ACTIONS(1583),
    [anon_sym_AMP_AMP] = ACTIONS(1583),
    [anon_sym_or] = ACTIONS(1583),
    [anon_sym_PIPE_PIPE] = ACTIONS(1583),
    [sym_assign_operator] = ACTIONS(1585),
    [sym_let_operator] = ACTIONS(1585),
    [sym_let_and_operator] = ACTIONS(1585),
    [sym__uppercase_identifier] = ACTIONS(1585),
    [aux_sym_directive_token1] = ACTIONS(1583),
    [aux_sym_tag_token1] = ACTIONS(1585),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1021)] = {
    [sym_attribute] = STATE(1021),
    [sym__lowercase_identifier] = ACTIONS(1587),
    [anon_sym_SEMI_SEMI] = ACTIONS(1589),
    [anon_sym_let] = ACTIONS(1587),
    [anon_sym_and] = ACTIONS(1587),
    [anon_sym_TILDE] = ACTIONS(1587),
    [anon_sym_QMARK] = ACTIONS(1587),
    [anon_sym_LPAREN] = ACTIONS(1589),
    [anon_sym_external] = ACTIONS(1587),
    [anon_sym_type] = ACTIONS(1587),
    [anon_sym_COMMA] = ACTIONS(1589),
    [anon_sym_PIPE] = ACTIONS(1587),
    [anon_sym_DOT] = ACTIONS(1589),
    [anon_sym_LBRACK] = ACTIONS(1587),
    [anon_sym_RBRACK] = ACTIONS(1589),
    [anon_sym_true] = ACTIONS(1587),
    [anon_sym_false] = ACTIONS(1587),
    [anon_sym_LBRACE] = ACTIONS(1587),
    [anon_sym_SEMI] = ACTIONS(1587),
    [anon_sym_exception] = ACTIONS(1587),
    [anon_sym_module] = ACTIONS(1587),
    [anon_sym_open] = ACTIONS(1587),
    [anon_sym_include] = ACTIONS(1587),
    [anon_sym_class] = ACTIONS(1587),
    [anon_sym_val] = ACTIONS(1587),
    [anon_sym_end] = ACTIONS(1587),
    [anon_sym_object] = ACTIONS(1587),
    [anon_sym_in] = ACTIONS(1587),
    [anon_sym_AMP] = ACTIONS(1587),
    [anon_sym_POUND] = ACTIONS(1587),
    [anon_sym_COLON_COLON] = ACTIONS(1589),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1589),
    [anon_sym_else] = ACTIONS(1587),
    [anon_sym_new] = ACTIONS(1587),
    [anon_sym_LBRACE_LT] = ACTIONS(1589),
    [anon_sym_begin] = ACTIONS(1587),
    [sym_ocamlyacc_value] = ACTIONS(1589),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1587),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1589),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1587),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1587),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1589),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1589),
    [aux_sym_number_token1] = ACTIONS(1589),
    [aux_sym_signed_number_token1] = ACTIONS(1587),
    [anon_sym_SQUOTE] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(1589),
    [sym_prefix_operator] = ACTIONS(1589),
    [aux_sym_sign_operator_token1] = ACTIONS(1587),
    [sym_hash_operator] = ACTIONS(1589),
    [anon_sym_lsl] = ACTIONS(1587),
    [anon_sym_lsr] = ACTIONS(1587),
    [anon_sym_asr] = ACTIONS(1587),
    [aux_sym_pow_operator_token1] = ACTIONS(1589),
    [anon_sym_mod] = ACTIONS(1587),
    [anon_sym_land] = ACTIONS(1587),
    [anon_sym_lor] = ACTIONS(1587),
    [anon_sym_lxor] = ACTIONS(1587),
    [aux_sym_mult_operator_token1] = ACTIONS(1587),
    [aux_sym_add_operator_token1] = ACTIONS(1587),
    [sym_concat_operator] = ACTIONS(1589),
    [sym_rel_operator] = ACTIONS(1587),
    [anon_sym_AMP_AMP] = ACTIONS(1587),
    [anon_sym_or] = ACTIONS(1587),
    [anon_sym_PIPE_PIPE] = ACTIONS(1587),
    [sym_assign_operator] = ACTIONS(1589),
    [sym_let_operator] = ACTIONS(1589),
    [sym_let_and_operator] = ACTIONS(1589),
    [sym__uppercase_identifier] = ACTIONS(1589),
    [aux_sym_directive_token1] = ACTIONS(1587),
    [aux_sym_tag_token1] = ACTIONS(1589),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1022)] = {
    [sym_attribute] = STATE(1022),
    [sym__lowercase_identifier] = ACTIONS(1599),
    [anon_sym_SEMI_SEMI] = ACTIONS(1601),
    [anon_sym_let] = ACTIONS(1599),
    [anon_sym_and] = ACTIONS(1599),
    [anon_sym_TILDE] = ACTIONS(1599),
    [anon_sym_QMARK] = ACTIONS(1599),
    [anon_sym_LPAREN] = ACTIONS(1601),
    [anon_sym_external] = ACTIONS(1599),
    [anon_sym_type] = ACTIONS(1599),
    [anon_sym_COMMA] = ACTIONS(1601),
    [anon_sym_PIPE] = ACTIONS(1599),
    [anon_sym_DOT] = ACTIONS(1601),
    [anon_sym_LBRACK] = ACTIONS(1599),
    [anon_sym_RBRACK] = ACTIONS(1601),
    [anon_sym_true] = ACTIONS(1599),
    [anon_sym_false] = ACTIONS(1599),
    [anon_sym_LBRACE] = ACTIONS(1599),
    [anon_sym_SEMI] = ACTIONS(1599),
    [anon_sym_exception] = ACTIONS(1599),
    [anon_sym_module] = ACTIONS(1599),
    [anon_sym_open] = ACTIONS(1599),
    [anon_sym_include] = ACTIONS(1599),
    [anon_sym_class] = ACTIONS(1599),
    [anon_sym_val] = ACTIONS(1599),
    [anon_sym_end] = ACTIONS(1599),
    [anon_sym_object] = ACTIONS(1599),
    [anon_sym_in] = ACTIONS(1599),
    [anon_sym_AMP] = ACTIONS(1599),
    [anon_sym_POUND] = ACTIONS(1599),
    [anon_sym_COLON_COLON] = ACTIONS(1601),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1601),
    [anon_sym_else] = ACTIONS(1599),
    [anon_sym_new] = ACTIONS(1599),
    [anon_sym_LBRACE_LT] = ACTIONS(1601),
    [anon_sym_begin] = ACTIONS(1599),
    [sym_ocamlyacc_value] = ACTIONS(1601),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1599),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1601),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1599),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1599),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1601),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1601),
    [aux_sym_number_token1] = ACTIONS(1601),
    [aux_sym_signed_number_token1] = ACTIONS(1599),
    [anon_sym_SQUOTE] = ACTIONS(1601),
    [anon_sym_DQUOTE] = ACTIONS(1601),
    [sym_prefix_operator] = ACTIONS(1601),
    [aux_sym_sign_operator_token1] = ACTIONS(1599),
    [sym_hash_operator] = ACTIONS(1601),
    [anon_sym_lsl] = ACTIONS(1599),
    [anon_sym_lsr] = ACTIONS(1599),
    [anon_sym_asr] = ACTIONS(1599),
    [aux_sym_pow_operator_token1] = ACTIONS(1601),
    [anon_sym_mod] = ACTIONS(1599),
    [anon_sym_land] = ACTIONS(1599),
    [anon_sym_lor] = ACTIONS(1599),
    [anon_sym_lxor] = ACTIONS(1599),
    [aux_sym_mult_operator_token1] = ACTIONS(1599),
    [aux_sym_add_operator_token1] = ACTIONS(1599),
    [sym_concat_operator] = ACTIONS(1601),
    [sym_rel_operator] = ACTIONS(1599),
    [anon_sym_AMP_AMP] = ACTIONS(1599),
    [anon_sym_or] = ACTIONS(1599),
    [anon_sym_PIPE_PIPE] = ACTIONS(1599),
    [sym_assign_operator] = ACTIONS(1601),
    [sym_let_operator] = ACTIONS(1601),
    [sym_let_and_operator] = ACTIONS(1601),
    [sym__uppercase_identifier] = ACTIONS(1601),
    [aux_sym_directive_token1] = ACTIONS(1599),
    [aux_sym_tag_token1] = ACTIONS(1601),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1023)] = {
    [sym_attribute] = STATE(1023),
    [sym__lowercase_identifier] = ACTIONS(1603),
    [anon_sym_SEMI_SEMI] = ACTIONS(1605),
    [anon_sym_let] = ACTIONS(1603),
    [anon_sym_and] = ACTIONS(1603),
    [anon_sym_TILDE] = ACTIONS(1603),
    [anon_sym_QMARK] = ACTIONS(1603),
    [anon_sym_LPAREN] = ACTIONS(1605),
    [anon_sym_external] = ACTIONS(1603),
    [anon_sym_type] = ACTIONS(1603),
    [anon_sym_COMMA] = ACTIONS(1605),
    [anon_sym_PIPE] = ACTIONS(1603),
    [anon_sym_DOT] = ACTIONS(1605),
    [anon_sym_LBRACK] = ACTIONS(1603),
    [anon_sym_RBRACK] = ACTIONS(1605),
    [anon_sym_true] = ACTIONS(1603),
    [anon_sym_false] = ACTIONS(1603),
    [anon_sym_LBRACE] = ACTIONS(1603),
    [anon_sym_SEMI] = ACTIONS(1603),
    [anon_sym_exception] = ACTIONS(1603),
    [anon_sym_module] = ACTIONS(1603),
    [anon_sym_open] = ACTIONS(1603),
    [anon_sym_include] = ACTIONS(1603),
    [anon_sym_class] = ACTIONS(1603),
    [anon_sym_val] = ACTIONS(1603),
    [anon_sym_end] = ACTIONS(1603),
    [anon_sym_object] = ACTIONS(1603),
    [anon_sym_in] = ACTIONS(1603),
    [anon_sym_AMP] = ACTIONS(1603),
    [anon_sym_POUND] = ACTIONS(1603),
    [anon_sym_COLON_COLON] = ACTIONS(1605),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1605),
    [anon_sym_else] = ACTIONS(1603),
    [anon_sym_new] = ACTIONS(1603),
    [anon_sym_LBRACE_LT] = ACTIONS(1605),
    [anon_sym_begin] = ACTIONS(1603),
    [sym_ocamlyacc_value] = ACTIONS(1605),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1603),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1605),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1603),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1603),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1605),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1605),
    [aux_sym_number_token1] = ACTIONS(1605),
    [aux_sym_signed_number_token1] = ACTIONS(1603),
    [anon_sym_SQUOTE] = ACTIONS(1605),
    [anon_sym_DQUOTE] = ACTIONS(1605),
    [sym_prefix_operator] = ACTIONS(1605),
    [aux_sym_sign_operator_token1] = ACTIONS(1603),
    [sym_hash_operator] = ACTIONS(1605),
    [anon_sym_lsl] = ACTIONS(1603),
    [anon_sym_lsr] = ACTIONS(1603),
    [anon_sym_asr] = ACTIONS(1603),
    [aux_sym_pow_operator_token1] = ACTIONS(1605),
    [anon_sym_mod] = ACTIONS(1603),
    [anon_sym_land] = ACTIONS(1603),
    [anon_sym_lor] = ACTIONS(1603),
    [anon_sym_lxor] = ACTIONS(1603),
    [aux_sym_mult_operator_token1] = ACTIONS(1603),
    [aux_sym_add_operator_token1] = ACTIONS(1603),
    [sym_concat_operator] = ACTIONS(1605),
    [sym_rel_operator] = ACTIONS(1603),
    [anon_sym_AMP_AMP] = ACTIONS(1603),
    [anon_sym_or] = ACTIONS(1603),
    [anon_sym_PIPE_PIPE] = ACTIONS(1603),
    [sym_assign_operator] = ACTIONS(1605),
    [sym_let_operator] = ACTIONS(1605),
    [sym_let_and_operator] = ACTIONS(1605),
    [sym__uppercase_identifier] = ACTIONS(1605),
    [aux_sym_directive_token1] = ACTIONS(1603),
    [aux_sym_tag_token1] = ACTIONS(1605),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1024)] = {
    [sym_attribute] = STATE(1024),
    [sym__lowercase_identifier] = ACTIONS(1607),
    [anon_sym_SEMI_SEMI] = ACTIONS(1609),
    [anon_sym_let] = ACTIONS(1607),
    [anon_sym_and] = ACTIONS(1607),
    [anon_sym_TILDE] = ACTIONS(1607),
    [anon_sym_QMARK] = ACTIONS(1607),
    [anon_sym_LPAREN] = ACTIONS(1609),
    [anon_sym_external] = ACTIONS(1607),
    [anon_sym_type] = ACTIONS(1607),
    [anon_sym_COMMA] = ACTIONS(1609),
    [anon_sym_PIPE] = ACTIONS(1607),
    [anon_sym_DOT] = ACTIONS(1609),
    [anon_sym_LBRACK] = ACTIONS(1607),
    [anon_sym_RBRACK] = ACTIONS(1609),
    [anon_sym_true] = ACTIONS(1607),
    [anon_sym_false] = ACTIONS(1607),
    [anon_sym_LBRACE] = ACTIONS(1607),
    [anon_sym_SEMI] = ACTIONS(1607),
    [anon_sym_exception] = ACTIONS(1607),
    [anon_sym_module] = ACTIONS(1607),
    [anon_sym_open] = ACTIONS(1607),
    [anon_sym_include] = ACTIONS(1607),
    [anon_sym_class] = ACTIONS(1607),
    [anon_sym_val] = ACTIONS(1607),
    [anon_sym_end] = ACTIONS(1607),
    [anon_sym_object] = ACTIONS(1607),
    [anon_sym_in] = ACTIONS(1607),
    [anon_sym_AMP] = ACTIONS(1607),
    [anon_sym_POUND] = ACTIONS(1607),
    [anon_sym_COLON_COLON] = ACTIONS(1609),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1609),
    [anon_sym_else] = ACTIONS(1607),
    [anon_sym_new] = ACTIONS(1607),
    [anon_sym_LBRACE_LT] = ACTIONS(1609),
    [anon_sym_begin] = ACTIONS(1607),
    [sym_ocamlyacc_value] = ACTIONS(1609),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1607),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1609),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1607),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1607),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1609),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1609),
    [aux_sym_number_token1] = ACTIONS(1609),
    [aux_sym_signed_number_token1] = ACTIONS(1607),
    [anon_sym_SQUOTE] = ACTIONS(1609),
    [anon_sym_DQUOTE] = ACTIONS(1609),
    [sym_prefix_operator] = ACTIONS(1609),
    [aux_sym_sign_operator_token1] = ACTIONS(1607),
    [sym_hash_operator] = ACTIONS(1609),
    [anon_sym_lsl] = ACTIONS(1607),
    [anon_sym_lsr] = ACTIONS(1607),
    [anon_sym_asr] = ACTIONS(1607),
    [aux_sym_pow_operator_token1] = ACTIONS(1609),
    [anon_sym_mod] = ACTIONS(1607),
    [anon_sym_land] = ACTIONS(1607),
    [anon_sym_lor] = ACTIONS(1607),
    [anon_sym_lxor] = ACTIONS(1607),
    [aux_sym_mult_operator_token1] = ACTIONS(1607),
    [aux_sym_add_operator_token1] = ACTIONS(1607),
    [sym_concat_operator] = ACTIONS(1609),
    [sym_rel_operator] = ACTIONS(1607),
    [anon_sym_AMP_AMP] = ACTIONS(1607),
    [anon_sym_or] = ACTIONS(1607),
    [anon_sym_PIPE_PIPE] = ACTIONS(1607),
    [sym_assign_operator] = ACTIONS(1609),
    [sym_let_operator] = ACTIONS(1609),
    [sym_let_and_operator] = ACTIONS(1609),
    [sym__uppercase_identifier] = ACTIONS(1609),
    [aux_sym_directive_token1] = ACTIONS(1607),
    [aux_sym_tag_token1] = ACTIONS(1609),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1025)] = {
    [sym_attribute] = STATE(1025),
    [sym__lowercase_identifier] = ACTIONS(1619),
    [anon_sym_SEMI_SEMI] = ACTIONS(1621),
    [anon_sym_let] = ACTIONS(1619),
    [anon_sym_and] = ACTIONS(1619),
    [anon_sym_TILDE] = ACTIONS(1619),
    [anon_sym_QMARK] = ACTIONS(1619),
    [anon_sym_LPAREN] = ACTIONS(1621),
    [anon_sym_external] = ACTIONS(1619),
    [anon_sym_type] = ACTIONS(1619),
    [anon_sym_COMMA] = ACTIONS(1621),
    [anon_sym_PIPE] = ACTIONS(1619),
    [anon_sym_DOT] = ACTIONS(1621),
    [anon_sym_LBRACK] = ACTIONS(1619),
    [anon_sym_RBRACK] = ACTIONS(1621),
    [anon_sym_true] = ACTIONS(1619),
    [anon_sym_false] = ACTIONS(1619),
    [anon_sym_LBRACE] = ACTIONS(1619),
    [anon_sym_SEMI] = ACTIONS(1619),
    [anon_sym_exception] = ACTIONS(1619),
    [anon_sym_module] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(1619),
    [anon_sym_include] = ACTIONS(1619),
    [anon_sym_class] = ACTIONS(1619),
    [anon_sym_val] = ACTIONS(1619),
    [anon_sym_end] = ACTIONS(1619),
    [anon_sym_object] = ACTIONS(1619),
    [anon_sym_in] = ACTIONS(1619),
    [anon_sym_AMP] = ACTIONS(1619),
    [anon_sym_POUND] = ACTIONS(1619),
    [anon_sym_COLON_COLON] = ACTIONS(1621),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1621),
    [anon_sym_else] = ACTIONS(1619),
    [anon_sym_new] = ACTIONS(1619),
    [anon_sym_LBRACE_LT] = ACTIONS(1621),
    [anon_sym_begin] = ACTIONS(1619),
    [sym_ocamlyacc_value] = ACTIONS(1621),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1619),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1621),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1619),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1619),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1621),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1621),
    [aux_sym_number_token1] = ACTIONS(1621),
    [aux_sym_signed_number_token1] = ACTIONS(1619),
    [anon_sym_SQUOTE] = ACTIONS(1621),
    [anon_sym_DQUOTE] = ACTIONS(1621),
    [sym_prefix_operator] = ACTIONS(1621),
    [aux_sym_sign_operator_token1] = ACTIONS(1619),
    [sym_hash_operator] = ACTIONS(1621),
    [anon_sym_lsl] = ACTIONS(1619),
    [anon_sym_lsr] = ACTIONS(1619),
    [anon_sym_asr] = ACTIONS(1619),
    [aux_sym_pow_operator_token1] = ACTIONS(1621),
    [anon_sym_mod] = ACTIONS(1619),
    [anon_sym_land] = ACTIONS(1619),
    [anon_sym_lor] = ACTIONS(1619),
    [anon_sym_lxor] = ACTIONS(1619),
    [aux_sym_mult_operator_token1] = ACTIONS(1619),
    [aux_sym_add_operator_token1] = ACTIONS(1619),
    [sym_concat_operator] = ACTIONS(1621),
    [sym_rel_operator] = ACTIONS(1619),
    [anon_sym_AMP_AMP] = ACTIONS(1619),
    [anon_sym_or] = ACTIONS(1619),
    [anon_sym_PIPE_PIPE] = ACTIONS(1619),
    [sym_assign_operator] = ACTIONS(1621),
    [sym_let_operator] = ACTIONS(1621),
    [sym_let_and_operator] = ACTIONS(1621),
    [sym__uppercase_identifier] = ACTIONS(1621),
    [aux_sym_directive_token1] = ACTIONS(1619),
    [aux_sym_tag_token1] = ACTIONS(1621),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1026)] = {
    [sym_attribute] = STATE(1026),
    [sym__lowercase_identifier] = ACTIONS(1651),
    [anon_sym_SEMI_SEMI] = ACTIONS(1653),
    [anon_sym_let] = ACTIONS(1651),
    [anon_sym_and] = ACTIONS(1651),
    [anon_sym_TILDE] = ACTIONS(1651),
    [anon_sym_QMARK] = ACTIONS(1651),
    [anon_sym_LPAREN] = ACTIONS(1653),
    [anon_sym_external] = ACTIONS(1651),
    [anon_sym_type] = ACTIONS(1651),
    [anon_sym_COMMA] = ACTIONS(1653),
    [anon_sym_PIPE] = ACTIONS(1651),
    [anon_sym_DOT] = ACTIONS(1653),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_RBRACK] = ACTIONS(1653),
    [anon_sym_true] = ACTIONS(1651),
    [anon_sym_false] = ACTIONS(1651),
    [anon_sym_LBRACE] = ACTIONS(1651),
    [anon_sym_SEMI] = ACTIONS(1651),
    [anon_sym_exception] = ACTIONS(1651),
    [anon_sym_module] = ACTIONS(1651),
    [anon_sym_open] = ACTIONS(1651),
    [anon_sym_include] = ACTIONS(1651),
    [anon_sym_class] = ACTIONS(1651),
    [anon_sym_val] = ACTIONS(1651),
    [anon_sym_end] = ACTIONS(1651),
    [anon_sym_object] = ACTIONS(1651),
    [anon_sym_in] = ACTIONS(1651),
    [anon_sym_AMP] = ACTIONS(1651),
    [anon_sym_POUND] = ACTIONS(1651),
    [anon_sym_COLON_COLON] = ACTIONS(1653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1653),
    [anon_sym_else] = ACTIONS(1651),
    [anon_sym_new] = ACTIONS(1651),
    [anon_sym_LBRACE_LT] = ACTIONS(1653),
    [anon_sym_begin] = ACTIONS(1651),
    [sym_ocamlyacc_value] = ACTIONS(1653),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1651),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1653),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1651),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1653),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1653),
    [aux_sym_number_token1] = ACTIONS(1653),
    [aux_sym_signed_number_token1] = ACTIONS(1651),
    [anon_sym_SQUOTE] = ACTIONS(1653),
    [anon_sym_DQUOTE] = ACTIONS(1653),
    [sym_prefix_operator] = ACTIONS(1653),
    [aux_sym_sign_operator_token1] = ACTIONS(1651),
    [sym_hash_operator] = ACTIONS(1653),
    [anon_sym_lsl] = ACTIONS(1651),
    [anon_sym_lsr] = ACTIONS(1651),
    [anon_sym_asr] = ACTIONS(1651),
    [aux_sym_pow_operator_token1] = ACTIONS(1653),
    [anon_sym_mod] = ACTIONS(1651),
    [anon_sym_land] = ACTIONS(1651),
    [anon_sym_lor] = ACTIONS(1651),
    [anon_sym_lxor] = ACTIONS(1651),
    [aux_sym_mult_operator_token1] = ACTIONS(1651),
    [aux_sym_add_operator_token1] = ACTIONS(1651),
    [sym_concat_operator] = ACTIONS(1653),
    [sym_rel_operator] = ACTIONS(1651),
    [anon_sym_AMP_AMP] = ACTIONS(1651),
    [anon_sym_or] = ACTIONS(1651),
    [anon_sym_PIPE_PIPE] = ACTIONS(1651),
    [sym_assign_operator] = ACTIONS(1653),
    [sym_let_operator] = ACTIONS(1653),
    [sym_let_and_operator] = ACTIONS(1653),
    [sym__uppercase_identifier] = ACTIONS(1653),
    [aux_sym_directive_token1] = ACTIONS(1651),
    [aux_sym_tag_token1] = ACTIONS(1653),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1027)] = {
    [sym_attribute] = STATE(1027),
    [sym__lowercase_identifier] = ACTIONS(1631),
    [anon_sym_SEMI_SEMI] = ACTIONS(1633),
    [anon_sym_let] = ACTIONS(1631),
    [anon_sym_and] = ACTIONS(1631),
    [anon_sym_TILDE] = ACTIONS(1631),
    [anon_sym_QMARK] = ACTIONS(1631),
    [anon_sym_LPAREN] = ACTIONS(1633),
    [anon_sym_external] = ACTIONS(1631),
    [anon_sym_type] = ACTIONS(1631),
    [anon_sym_COMMA] = ACTIONS(1633),
    [anon_sym_PIPE] = ACTIONS(1631),
    [anon_sym_DOT] = ACTIONS(1633),
    [anon_sym_LBRACK] = ACTIONS(1631),
    [anon_sym_RBRACK] = ACTIONS(1633),
    [anon_sym_true] = ACTIONS(1631),
    [anon_sym_false] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1631),
    [anon_sym_SEMI] = ACTIONS(1631),
    [anon_sym_exception] = ACTIONS(1631),
    [anon_sym_module] = ACTIONS(1631),
    [anon_sym_open] = ACTIONS(1631),
    [anon_sym_include] = ACTIONS(1631),
    [anon_sym_class] = ACTIONS(1631),
    [anon_sym_val] = ACTIONS(1631),
    [anon_sym_end] = ACTIONS(1631),
    [anon_sym_object] = ACTIONS(1631),
    [anon_sym_in] = ACTIONS(1631),
    [anon_sym_AMP] = ACTIONS(1631),
    [anon_sym_POUND] = ACTIONS(1631),
    [anon_sym_COLON_COLON] = ACTIONS(1633),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1633),
    [anon_sym_else] = ACTIONS(1631),
    [anon_sym_new] = ACTIONS(1631),
    [anon_sym_LBRACE_LT] = ACTIONS(1633),
    [anon_sym_begin] = ACTIONS(1631),
    [sym_ocamlyacc_value] = ACTIONS(1633),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1631),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1633),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1631),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1631),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1633),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1633),
    [aux_sym_number_token1] = ACTIONS(1633),
    [aux_sym_signed_number_token1] = ACTIONS(1631),
    [anon_sym_SQUOTE] = ACTIONS(1633),
    [anon_sym_DQUOTE] = ACTIONS(1633),
    [sym_prefix_operator] = ACTIONS(1633),
    [aux_sym_sign_operator_token1] = ACTIONS(1631),
    [sym_hash_operator] = ACTIONS(1633),
    [anon_sym_lsl] = ACTIONS(1631),
    [anon_sym_lsr] = ACTIONS(1631),
    [anon_sym_asr] = ACTIONS(1631),
    [aux_sym_pow_operator_token1] = ACTIONS(1633),
    [anon_sym_mod] = ACTIONS(1631),
    [anon_sym_land] = ACTIONS(1631),
    [anon_sym_lor] = ACTIONS(1631),
    [anon_sym_lxor] = ACTIONS(1631),
    [aux_sym_mult_operator_token1] = ACTIONS(1631),
    [aux_sym_add_operator_token1] = ACTIONS(1631),
    [sym_concat_operator] = ACTIONS(1633),
    [sym_rel_operator] = ACTIONS(1631),
    [anon_sym_AMP_AMP] = ACTIONS(1631),
    [anon_sym_or] = ACTIONS(1631),
    [anon_sym_PIPE_PIPE] = ACTIONS(1631),
    [sym_assign_operator] = ACTIONS(1633),
    [sym_let_operator] = ACTIONS(1633),
    [sym_let_and_operator] = ACTIONS(1633),
    [sym__uppercase_identifier] = ACTIONS(1633),
    [aux_sym_directive_token1] = ACTIONS(1631),
    [aux_sym_tag_token1] = ACTIONS(1633),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1028)] = {
    [sym_attribute] = STATE(1028),
    [sym__lowercase_identifier] = ACTIONS(1635),
    [anon_sym_SEMI_SEMI] = ACTIONS(1637),
    [anon_sym_let] = ACTIONS(1635),
    [anon_sym_and] = ACTIONS(1635),
    [anon_sym_TILDE] = ACTIONS(1635),
    [anon_sym_QMARK] = ACTIONS(1635),
    [anon_sym_LPAREN] = ACTIONS(1637),
    [anon_sym_external] = ACTIONS(1635),
    [anon_sym_type] = ACTIONS(1635),
    [anon_sym_COMMA] = ACTIONS(1637),
    [anon_sym_PIPE] = ACTIONS(1635),
    [anon_sym_DOT] = ACTIONS(1637),
    [anon_sym_LBRACK] = ACTIONS(1635),
    [anon_sym_RBRACK] = ACTIONS(1637),
    [anon_sym_true] = ACTIONS(1635),
    [anon_sym_false] = ACTIONS(1635),
    [anon_sym_LBRACE] = ACTIONS(1635),
    [anon_sym_SEMI] = ACTIONS(1635),
    [anon_sym_exception] = ACTIONS(1635),
    [anon_sym_module] = ACTIONS(1635),
    [anon_sym_open] = ACTIONS(1635),
    [anon_sym_include] = ACTIONS(1635),
    [anon_sym_class] = ACTIONS(1635),
    [anon_sym_val] = ACTIONS(1635),
    [anon_sym_end] = ACTIONS(1635),
    [anon_sym_object] = ACTIONS(1635),
    [anon_sym_in] = ACTIONS(1635),
    [anon_sym_AMP] = ACTIONS(1635),
    [anon_sym_POUND] = ACTIONS(1635),
    [anon_sym_COLON_COLON] = ACTIONS(1637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1637),
    [anon_sym_else] = ACTIONS(1635),
    [anon_sym_new] = ACTIONS(1635),
    [anon_sym_LBRACE_LT] = ACTIONS(1637),
    [anon_sym_begin] = ACTIONS(1635),
    [sym_ocamlyacc_value] = ACTIONS(1637),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1635),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1637),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1635),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1635),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1637),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1637),
    [aux_sym_number_token1] = ACTIONS(1637),
    [aux_sym_signed_number_token1] = ACTIONS(1635),
    [anon_sym_SQUOTE] = ACTIONS(1637),
    [anon_sym_DQUOTE] = ACTIONS(1637),
    [sym_prefix_operator] = ACTIONS(1637),
    [aux_sym_sign_operator_token1] = ACTIONS(1635),
    [sym_hash_operator] = ACTIONS(1637),
    [anon_sym_lsl] = ACTIONS(1635),
    [anon_sym_lsr] = ACTIONS(1635),
    [anon_sym_asr] = ACTIONS(1635),
    [aux_sym_pow_operator_token1] = ACTIONS(1637),
    [anon_sym_mod] = ACTIONS(1635),
    [anon_sym_land] = ACTIONS(1635),
    [anon_sym_lor] = ACTIONS(1635),
    [anon_sym_lxor] = ACTIONS(1635),
    [aux_sym_mult_operator_token1] = ACTIONS(1635),
    [aux_sym_add_operator_token1] = ACTIONS(1635),
    [sym_concat_operator] = ACTIONS(1637),
    [sym_rel_operator] = ACTIONS(1635),
    [anon_sym_AMP_AMP] = ACTIONS(1635),
    [anon_sym_or] = ACTIONS(1635),
    [anon_sym_PIPE_PIPE] = ACTIONS(1635),
    [sym_assign_operator] = ACTIONS(1637),
    [sym_let_operator] = ACTIONS(1637),
    [sym_let_and_operator] = ACTIONS(1637),
    [sym__uppercase_identifier] = ACTIONS(1637),
    [aux_sym_directive_token1] = ACTIONS(1635),
    [aux_sym_tag_token1] = ACTIONS(1637),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1029)] = {
    [sym_attribute] = STATE(1029),
    [sym__lowercase_identifier] = ACTIONS(1639),
    [anon_sym_SEMI_SEMI] = ACTIONS(1641),
    [anon_sym_let] = ACTIONS(1639),
    [anon_sym_and] = ACTIONS(1639),
    [anon_sym_TILDE] = ACTIONS(1639),
    [anon_sym_QMARK] = ACTIONS(1639),
    [anon_sym_LPAREN] = ACTIONS(1641),
    [anon_sym_external] = ACTIONS(1639),
    [anon_sym_type] = ACTIONS(1639),
    [anon_sym_COMMA] = ACTIONS(1641),
    [anon_sym_PIPE] = ACTIONS(1639),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACK] = ACTIONS(1639),
    [anon_sym_RBRACK] = ACTIONS(1641),
    [anon_sym_true] = ACTIONS(1639),
    [anon_sym_false] = ACTIONS(1639),
    [anon_sym_LBRACE] = ACTIONS(1639),
    [anon_sym_SEMI] = ACTIONS(1639),
    [anon_sym_exception] = ACTIONS(1639),
    [anon_sym_module] = ACTIONS(1639),
    [anon_sym_open] = ACTIONS(1639),
    [anon_sym_include] = ACTIONS(1639),
    [anon_sym_class] = ACTIONS(1639),
    [anon_sym_val] = ACTIONS(1639),
    [anon_sym_end] = ACTIONS(1639),
    [anon_sym_object] = ACTIONS(1639),
    [anon_sym_in] = ACTIONS(1639),
    [anon_sym_AMP] = ACTIONS(1639),
    [anon_sym_POUND] = ACTIONS(1639),
    [anon_sym_COLON_COLON] = ACTIONS(1641),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1641),
    [anon_sym_else] = ACTIONS(1639),
    [anon_sym_new] = ACTIONS(1639),
    [anon_sym_LBRACE_LT] = ACTIONS(1641),
    [anon_sym_begin] = ACTIONS(1639),
    [sym_ocamlyacc_value] = ACTIONS(1641),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1639),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1641),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1639),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1639),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1641),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1641),
    [aux_sym_number_token1] = ACTIONS(1641),
    [aux_sym_signed_number_token1] = ACTIONS(1639),
    [anon_sym_SQUOTE] = ACTIONS(1641),
    [anon_sym_DQUOTE] = ACTIONS(1641),
    [sym_prefix_operator] = ACTIONS(1641),
    [aux_sym_sign_operator_token1] = ACTIONS(1639),
    [sym_hash_operator] = ACTIONS(1641),
    [anon_sym_lsl] = ACTIONS(1639),
    [anon_sym_lsr] = ACTIONS(1639),
    [anon_sym_asr] = ACTIONS(1639),
    [aux_sym_pow_operator_token1] = ACTIONS(1641),
    [anon_sym_mod] = ACTIONS(1639),
    [anon_sym_land] = ACTIONS(1639),
    [anon_sym_lor] = ACTIONS(1639),
    [anon_sym_lxor] = ACTIONS(1639),
    [aux_sym_mult_operator_token1] = ACTIONS(1639),
    [aux_sym_add_operator_token1] = ACTIONS(1639),
    [sym_concat_operator] = ACTIONS(1641),
    [sym_rel_operator] = ACTIONS(1639),
    [anon_sym_AMP_AMP] = ACTIONS(1639),
    [anon_sym_or] = ACTIONS(1639),
    [anon_sym_PIPE_PIPE] = ACTIONS(1639),
    [sym_assign_operator] = ACTIONS(1641),
    [sym_let_operator] = ACTIONS(1641),
    [sym_let_and_operator] = ACTIONS(1641),
    [sym__uppercase_identifier] = ACTIONS(1641),
    [aux_sym_directive_token1] = ACTIONS(1639),
    [aux_sym_tag_token1] = ACTIONS(1641),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1030)] = {
    [sym_attribute] = STATE(1030),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_SEMI_SEMI] = ACTIONS(1645),
    [anon_sym_let] = ACTIONS(1643),
    [anon_sym_and] = ACTIONS(1643),
    [anon_sym_TILDE] = ACTIONS(1643),
    [anon_sym_QMARK] = ACTIONS(1643),
    [anon_sym_LPAREN] = ACTIONS(1645),
    [anon_sym_external] = ACTIONS(1643),
    [anon_sym_type] = ACTIONS(1643),
    [anon_sym_COMMA] = ACTIONS(1645),
    [anon_sym_PIPE] = ACTIONS(1643),
    [anon_sym_DOT] = ACTIONS(1645),
    [anon_sym_LBRACK] = ACTIONS(1643),
    [anon_sym_RBRACK] = ACTIONS(1645),
    [anon_sym_true] = ACTIONS(1643),
    [anon_sym_false] = ACTIONS(1643),
    [anon_sym_LBRACE] = ACTIONS(1643),
    [anon_sym_SEMI] = ACTIONS(1643),
    [anon_sym_exception] = ACTIONS(1643),
    [anon_sym_module] = ACTIONS(1643),
    [anon_sym_open] = ACTIONS(1643),
    [anon_sym_include] = ACTIONS(1643),
    [anon_sym_class] = ACTIONS(1643),
    [anon_sym_val] = ACTIONS(1643),
    [anon_sym_end] = ACTIONS(1643),
    [anon_sym_object] = ACTIONS(1643),
    [anon_sym_in] = ACTIONS(1643),
    [anon_sym_AMP] = ACTIONS(1643),
    [anon_sym_POUND] = ACTIONS(1643),
    [anon_sym_COLON_COLON] = ACTIONS(1645),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1645),
    [anon_sym_else] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1643),
    [anon_sym_LBRACE_LT] = ACTIONS(1645),
    [anon_sym_begin] = ACTIONS(1643),
    [sym_ocamlyacc_value] = ACTIONS(1645),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1643),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1645),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1643),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1643),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1645),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1645),
    [aux_sym_number_token1] = ACTIONS(1645),
    [aux_sym_signed_number_token1] = ACTIONS(1643),
    [anon_sym_SQUOTE] = ACTIONS(1645),
    [anon_sym_DQUOTE] = ACTIONS(1645),
    [sym_prefix_operator] = ACTIONS(1645),
    [aux_sym_sign_operator_token1] = ACTIONS(1643),
    [sym_hash_operator] = ACTIONS(1645),
    [anon_sym_lsl] = ACTIONS(1643),
    [anon_sym_lsr] = ACTIONS(1643),
    [anon_sym_asr] = ACTIONS(1643),
    [aux_sym_pow_operator_token1] = ACTIONS(1645),
    [anon_sym_mod] = ACTIONS(1643),
    [anon_sym_land] = ACTIONS(1643),
    [anon_sym_lor] = ACTIONS(1643),
    [anon_sym_lxor] = ACTIONS(1643),
    [aux_sym_mult_operator_token1] = ACTIONS(1643),
    [aux_sym_add_operator_token1] = ACTIONS(1643),
    [sym_concat_operator] = ACTIONS(1645),
    [sym_rel_operator] = ACTIONS(1643),
    [anon_sym_AMP_AMP] = ACTIONS(1643),
    [anon_sym_or] = ACTIONS(1643),
    [anon_sym_PIPE_PIPE] = ACTIONS(1643),
    [sym_assign_operator] = ACTIONS(1645),
    [sym_let_operator] = ACTIONS(1645),
    [sym_let_and_operator] = ACTIONS(1645),
    [sym__uppercase_identifier] = ACTIONS(1645),
    [aux_sym_directive_token1] = ACTIONS(1643),
    [aux_sym_tag_token1] = ACTIONS(1645),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1031)] = {
    [sym_attribute] = STATE(1031),
    [sym__lowercase_identifier] = ACTIONS(1647),
    [anon_sym_SEMI_SEMI] = ACTIONS(1649),
    [anon_sym_let] = ACTIONS(1647),
    [anon_sym_and] = ACTIONS(1647),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_QMARK] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1649),
    [anon_sym_external] = ACTIONS(1647),
    [anon_sym_type] = ACTIONS(1647),
    [anon_sym_COMMA] = ACTIONS(1649),
    [anon_sym_PIPE] = ACTIONS(1647),
    [anon_sym_DOT] = ACTIONS(1649),
    [anon_sym_LBRACK] = ACTIONS(1647),
    [anon_sym_RBRACK] = ACTIONS(1649),
    [anon_sym_true] = ACTIONS(1647),
    [anon_sym_false] = ACTIONS(1647),
    [anon_sym_LBRACE] = ACTIONS(1647),
    [anon_sym_SEMI] = ACTIONS(1647),
    [anon_sym_exception] = ACTIONS(1647),
    [anon_sym_module] = ACTIONS(1647),
    [anon_sym_open] = ACTIONS(1647),
    [anon_sym_include] = ACTIONS(1647),
    [anon_sym_class] = ACTIONS(1647),
    [anon_sym_val] = ACTIONS(1647),
    [anon_sym_end] = ACTIONS(1647),
    [anon_sym_object] = ACTIONS(1647),
    [anon_sym_in] = ACTIONS(1647),
    [anon_sym_AMP] = ACTIONS(1647),
    [anon_sym_POUND] = ACTIONS(1647),
    [anon_sym_COLON_COLON] = ACTIONS(1649),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1649),
    [anon_sym_else] = ACTIONS(1647),
    [anon_sym_new] = ACTIONS(1647),
    [anon_sym_LBRACE_LT] = ACTIONS(1649),
    [anon_sym_begin] = ACTIONS(1647),
    [sym_ocamlyacc_value] = ACTIONS(1649),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1647),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1649),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1647),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1647),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1649),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1649),
    [aux_sym_number_token1] = ACTIONS(1649),
    [aux_sym_signed_number_token1] = ACTIONS(1647),
    [anon_sym_SQUOTE] = ACTIONS(1649),
    [anon_sym_DQUOTE] = ACTIONS(1649),
    [sym_prefix_operator] = ACTIONS(1649),
    [aux_sym_sign_operator_token1] = ACTIONS(1647),
    [sym_hash_operator] = ACTIONS(1649),
    [anon_sym_lsl] = ACTIONS(1647),
    [anon_sym_lsr] = ACTIONS(1647),
    [anon_sym_asr] = ACTIONS(1647),
    [aux_sym_pow_operator_token1] = ACTIONS(1649),
    [anon_sym_mod] = ACTIONS(1647),
    [anon_sym_land] = ACTIONS(1647),
    [anon_sym_lor] = ACTIONS(1647),
    [anon_sym_lxor] = ACTIONS(1647),
    [aux_sym_mult_operator_token1] = ACTIONS(1647),
    [aux_sym_add_operator_token1] = ACTIONS(1647),
    [sym_concat_operator] = ACTIONS(1649),
    [sym_rel_operator] = ACTIONS(1647),
    [anon_sym_AMP_AMP] = ACTIONS(1647),
    [anon_sym_or] = ACTIONS(1647),
    [anon_sym_PIPE_PIPE] = ACTIONS(1647),
    [sym_assign_operator] = ACTIONS(1649),
    [sym_let_operator] = ACTIONS(1649),
    [sym_let_and_operator] = ACTIONS(1649),
    [sym__uppercase_identifier] = ACTIONS(1649),
    [aux_sym_directive_token1] = ACTIONS(1647),
    [aux_sym_tag_token1] = ACTIONS(1649),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1032)] = {
    [sym_attribute] = STATE(1032),
    [sym__lowercase_identifier] = ACTIONS(1809),
    [anon_sym_SEMI_SEMI] = ACTIONS(1811),
    [anon_sym_let] = ACTIONS(1809),
    [anon_sym_and] = ACTIONS(1809),
    [anon_sym_TILDE] = ACTIONS(1809),
    [anon_sym_QMARK] = ACTIONS(1809),
    [anon_sym_LPAREN] = ACTIONS(1811),
    [anon_sym_external] = ACTIONS(1809),
    [anon_sym_type] = ACTIONS(1809),
    [anon_sym_COMMA] = ACTIONS(1811),
    [anon_sym_PIPE] = ACTIONS(1809),
    [anon_sym_DOT] = ACTIONS(1811),
    [anon_sym_LBRACK] = ACTIONS(1809),
    [anon_sym_RBRACK] = ACTIONS(1811),
    [anon_sym_true] = ACTIONS(1809),
    [anon_sym_false] = ACTIONS(1809),
    [anon_sym_LBRACE] = ACTIONS(1809),
    [anon_sym_SEMI] = ACTIONS(1809),
    [anon_sym_exception] = ACTIONS(1809),
    [anon_sym_module] = ACTIONS(1809),
    [anon_sym_open] = ACTIONS(1809),
    [anon_sym_include] = ACTIONS(1809),
    [anon_sym_class] = ACTIONS(1809),
    [anon_sym_val] = ACTIONS(1809),
    [anon_sym_end] = ACTIONS(1809),
    [anon_sym_object] = ACTIONS(1809),
    [anon_sym_in] = ACTIONS(1809),
    [anon_sym_AMP] = ACTIONS(1809),
    [anon_sym_POUND] = ACTIONS(1809),
    [anon_sym_COLON_COLON] = ACTIONS(1811),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1811),
    [anon_sym_else] = ACTIONS(1809),
    [anon_sym_new] = ACTIONS(1809),
    [anon_sym_LBRACE_LT] = ACTIONS(1811),
    [anon_sym_begin] = ACTIONS(1809),
    [sym_ocamlyacc_value] = ACTIONS(1811),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1809),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1811),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1809),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1809),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1811),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1811),
    [aux_sym_number_token1] = ACTIONS(1811),
    [aux_sym_signed_number_token1] = ACTIONS(1809),
    [anon_sym_SQUOTE] = ACTIONS(1811),
    [anon_sym_DQUOTE] = ACTIONS(1811),
    [sym_prefix_operator] = ACTIONS(1811),
    [aux_sym_sign_operator_token1] = ACTIONS(1809),
    [sym_hash_operator] = ACTIONS(1811),
    [anon_sym_lsl] = ACTIONS(1809),
    [anon_sym_lsr] = ACTIONS(1809),
    [anon_sym_asr] = ACTIONS(1809),
    [aux_sym_pow_operator_token1] = ACTIONS(1811),
    [anon_sym_mod] = ACTIONS(1809),
    [anon_sym_land] = ACTIONS(1809),
    [anon_sym_lor] = ACTIONS(1809),
    [anon_sym_lxor] = ACTIONS(1809),
    [aux_sym_mult_operator_token1] = ACTIONS(1809),
    [aux_sym_add_operator_token1] = ACTIONS(1809),
    [sym_concat_operator] = ACTIONS(1811),
    [sym_rel_operator] = ACTIONS(1809),
    [anon_sym_AMP_AMP] = ACTIONS(1809),
    [anon_sym_or] = ACTIONS(1809),
    [anon_sym_PIPE_PIPE] = ACTIONS(1809),
    [sym_assign_operator] = ACTIONS(1811),
    [sym_let_operator] = ACTIONS(1811),
    [sym_let_and_operator] = ACTIONS(1811),
    [sym__uppercase_identifier] = ACTIONS(1811),
    [aux_sym_directive_token1] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1033)] = {
    [sym_attribute] = STATE(1033),
    [sym__lowercase_identifier] = ACTIONS(1705),
    [anon_sym_TILDE] = ACTIONS(1705),
    [anon_sym_QMARK] = ACTIONS(1705),
    [anon_sym_LPAREN] = ACTIONS(1707),
    [anon_sym_RPAREN] = ACTIONS(1707),
    [anon_sym_COMMA] = ACTIONS(1707),
    [anon_sym_PIPE] = ACTIONS(1705),
    [anon_sym_COLON2] = ACTIONS(1705),
    [anon_sym_DOT] = ACTIONS(1707),
    [anon_sym_DASH_GT] = ACTIONS(1705),
    [anon_sym_LBRACK] = ACTIONS(1705),
    [anon_sym_RBRACK] = ACTIONS(1707),
    [anon_sym_true] = ACTIONS(1705),
    [anon_sym_false] = ACTIONS(1705),
    [anon_sym_LBRACE] = ACTIONS(1705),
    [anon_sym_SEMI] = ACTIONS(1707),
    [anon_sym_RBRACE] = ACTIONS(1707),
    [anon_sym_constraint] = ACTIONS(1705),
    [anon_sym_val] = ACTIONS(1705),
    [anon_sym_end] = ACTIONS(1705),
    [anon_sym_with] = ACTIONS(1705),
    [anon_sym_COLON_GT] = ACTIONS(1707),
    [anon_sym_object] = ACTIONS(1705),
    [anon_sym_inherit] = ACTIONS(1705),
    [anon_sym_method] = ACTIONS(1705),
    [anon_sym_initializer] = ACTIONS(1705),
    [anon_sym_AMP] = ACTIONS(1705),
    [anon_sym_POUND] = ACTIONS(1705),
    [anon_sym_COLON_COLON] = ACTIONS(1707),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1707),
    [anon_sym_LT_DASH] = ACTIONS(2178),
    [anon_sym_then] = ACTIONS(1705),
    [anon_sym_else] = ACTIONS(1705),
    [anon_sym_new] = ACTIONS(1705),
    [anon_sym_LBRACE_LT] = ACTIONS(1707),
    [anon_sym_GT_RBRACE] = ACTIONS(1707),
    [anon_sym_begin] = ACTIONS(1705),
    [sym_ocamlyacc_value] = ACTIONS(1707),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1705),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1707),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1705),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1705),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1707),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1707),
    [aux_sym_number_token1] = ACTIONS(1707),
    [aux_sym_signed_number_token1] = ACTIONS(1705),
    [anon_sym_SQUOTE] = ACTIONS(1707),
    [anon_sym_DQUOTE] = ACTIONS(1707),
    [sym_prefix_operator] = ACTIONS(1707),
    [aux_sym_sign_operator_token1] = ACTIONS(1705),
    [sym_hash_operator] = ACTIONS(1707),
    [anon_sym_lsl] = ACTIONS(1705),
    [anon_sym_lsr] = ACTIONS(1705),
    [anon_sym_asr] = ACTIONS(1705),
    [aux_sym_pow_operator_token1] = ACTIONS(1707),
    [anon_sym_mod] = ACTIONS(1705),
    [anon_sym_land] = ACTIONS(1705),
    [anon_sym_lor] = ACTIONS(1705),
    [anon_sym_lxor] = ACTIONS(1705),
    [aux_sym_mult_operator_token1] = ACTIONS(1705),
    [aux_sym_add_operator_token1] = ACTIONS(1705),
    [sym_concat_operator] = ACTIONS(1707),
    [sym_rel_operator] = ACTIONS(1705),
    [anon_sym_AMP_AMP] = ACTIONS(1705),
    [anon_sym_or] = ACTIONS(1705),
    [anon_sym_PIPE_PIPE] = ACTIONS(1705),
    [sym_assign_operator] = ACTIONS(1707),
    [sym__uppercase_identifier] = ACTIONS(1707),
    [aux_sym_tag_token1] = ACTIONS(1707),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1034)] = {
    [sym_attribute] = STATE(1034),
    [sym__lowercase_identifier] = ACTIONS(1559),
    [anon_sym_SEMI_SEMI] = ACTIONS(1561),
    [anon_sym_let] = ACTIONS(1559),
    [anon_sym_and] = ACTIONS(1559),
    [anon_sym_TILDE] = ACTIONS(1559),
    [anon_sym_QMARK] = ACTIONS(1559),
    [anon_sym_LPAREN] = ACTIONS(1561),
    [anon_sym_external] = ACTIONS(1559),
    [anon_sym_type] = ACTIONS(1559),
    [anon_sym_COMMA] = ACTIONS(1561),
    [anon_sym_PIPE] = ACTIONS(1559),
    [anon_sym_DOT] = ACTIONS(1561),
    [anon_sym_LBRACK] = ACTIONS(1559),
    [anon_sym_RBRACK] = ACTIONS(1561),
    [anon_sym_true] = ACTIONS(1559),
    [anon_sym_false] = ACTIONS(1559),
    [anon_sym_LBRACE] = ACTIONS(1559),
    [anon_sym_SEMI] = ACTIONS(1559),
    [anon_sym_exception] = ACTIONS(1559),
    [anon_sym_module] = ACTIONS(1559),
    [anon_sym_open] = ACTIONS(1559),
    [anon_sym_include] = ACTIONS(1559),
    [anon_sym_class] = ACTIONS(1559),
    [anon_sym_val] = ACTIONS(1559),
    [anon_sym_end] = ACTIONS(1559),
    [anon_sym_object] = ACTIONS(1559),
    [anon_sym_AMP] = ACTIONS(1559),
    [anon_sym_POUND] = ACTIONS(1559),
    [anon_sym_COLON_COLON] = ACTIONS(1561),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1561),
    [anon_sym_LT_DASH] = ACTIONS(1559),
    [anon_sym_else] = ACTIONS(1559),
    [anon_sym_new] = ACTIONS(1559),
    [anon_sym_LBRACE_LT] = ACTIONS(1561),
    [anon_sym_begin] = ACTIONS(1559),
    [sym_ocamlyacc_value] = ACTIONS(1561),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1559),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1561),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1559),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1559),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1561),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1561),
    [aux_sym_number_token1] = ACTIONS(1561),
    [aux_sym_signed_number_token1] = ACTIONS(1559),
    [anon_sym_SQUOTE] = ACTIONS(1561),
    [anon_sym_DQUOTE] = ACTIONS(1561),
    [sym_prefix_operator] = ACTIONS(1561),
    [aux_sym_sign_operator_token1] = ACTIONS(1559),
    [sym_hash_operator] = ACTIONS(1561),
    [anon_sym_lsl] = ACTIONS(1559),
    [anon_sym_lsr] = ACTIONS(1559),
    [anon_sym_asr] = ACTIONS(1559),
    [aux_sym_pow_operator_token1] = ACTIONS(1561),
    [anon_sym_mod] = ACTIONS(1559),
    [anon_sym_land] = ACTIONS(1559),
    [anon_sym_lor] = ACTIONS(1559),
    [anon_sym_lxor] = ACTIONS(1559),
    [aux_sym_mult_operator_token1] = ACTIONS(1559),
    [aux_sym_add_operator_token1] = ACTIONS(1559),
    [sym_concat_operator] = ACTIONS(1561),
    [sym_rel_operator] = ACTIONS(1559),
    [anon_sym_AMP_AMP] = ACTIONS(1559),
    [anon_sym_or] = ACTIONS(1559),
    [anon_sym_PIPE_PIPE] = ACTIONS(1559),
    [sym_assign_operator] = ACTIONS(1561),
    [sym_let_operator] = ACTIONS(1561),
    [sym__uppercase_identifier] = ACTIONS(1561),
    [aux_sym_directive_token1] = ACTIONS(1559),
    [aux_sym_tag_token1] = ACTIONS(1561),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1035)] = {
    [sym_attribute] = STATE(1035),
    [sym__lowercase_identifier] = ACTIONS(1563),
    [anon_sym_SEMI_SEMI] = ACTIONS(1565),
    [anon_sym_let] = ACTIONS(1563),
    [anon_sym_and] = ACTIONS(1563),
    [anon_sym_TILDE] = ACTIONS(1563),
    [anon_sym_QMARK] = ACTIONS(1563),
    [anon_sym_LPAREN] = ACTIONS(1565),
    [anon_sym_external] = ACTIONS(1563),
    [anon_sym_type] = ACTIONS(1563),
    [anon_sym_COMMA] = ACTIONS(1565),
    [anon_sym_PIPE] = ACTIONS(1563),
    [anon_sym_DOT] = ACTIONS(1565),
    [anon_sym_LBRACK] = ACTIONS(1563),
    [anon_sym_RBRACK] = ACTIONS(1565),
    [anon_sym_true] = ACTIONS(1563),
    [anon_sym_false] = ACTIONS(1563),
    [anon_sym_LBRACE] = ACTIONS(1563),
    [anon_sym_SEMI] = ACTIONS(1563),
    [anon_sym_exception] = ACTIONS(1563),
    [anon_sym_module] = ACTIONS(1563),
    [anon_sym_open] = ACTIONS(1563),
    [anon_sym_include] = ACTIONS(1563),
    [anon_sym_class] = ACTIONS(1563),
    [anon_sym_val] = ACTIONS(1563),
    [anon_sym_end] = ACTIONS(1563),
    [anon_sym_object] = ACTIONS(1563),
    [anon_sym_AMP] = ACTIONS(1563),
    [anon_sym_POUND] = ACTIONS(1563),
    [anon_sym_COLON_COLON] = ACTIONS(1565),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1565),
    [anon_sym_LT_DASH] = ACTIONS(1563),
    [anon_sym_else] = ACTIONS(1563),
    [anon_sym_new] = ACTIONS(1563),
    [anon_sym_LBRACE_LT] = ACTIONS(1565),
    [anon_sym_begin] = ACTIONS(1563),
    [sym_ocamlyacc_value] = ACTIONS(1565),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1563),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1565),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1563),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1563),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1565),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1565),
    [aux_sym_number_token1] = ACTIONS(1565),
    [aux_sym_signed_number_token1] = ACTIONS(1563),
    [anon_sym_SQUOTE] = ACTIONS(1565),
    [anon_sym_DQUOTE] = ACTIONS(1565),
    [sym_prefix_operator] = ACTIONS(1565),
    [aux_sym_sign_operator_token1] = ACTIONS(1563),
    [sym_hash_operator] = ACTIONS(1565),
    [anon_sym_lsl] = ACTIONS(1563),
    [anon_sym_lsr] = ACTIONS(1563),
    [anon_sym_asr] = ACTIONS(1563),
    [aux_sym_pow_operator_token1] = ACTIONS(1565),
    [anon_sym_mod] = ACTIONS(1563),
    [anon_sym_land] = ACTIONS(1563),
    [anon_sym_lor] = ACTIONS(1563),
    [anon_sym_lxor] = ACTIONS(1563),
    [aux_sym_mult_operator_token1] = ACTIONS(1563),
    [aux_sym_add_operator_token1] = ACTIONS(1563),
    [sym_concat_operator] = ACTIONS(1565),
    [sym_rel_operator] = ACTIONS(1563),
    [anon_sym_AMP_AMP] = ACTIONS(1563),
    [anon_sym_or] = ACTIONS(1563),
    [anon_sym_PIPE_PIPE] = ACTIONS(1563),
    [sym_assign_operator] = ACTIONS(1565),
    [sym_let_operator] = ACTIONS(1565),
    [sym__uppercase_identifier] = ACTIONS(1565),
    [aux_sym_directive_token1] = ACTIONS(1563),
    [aux_sym_tag_token1] = ACTIONS(1565),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1036)] = {
    [sym_attribute] = STATE(1036),
    [sym__lowercase_identifier] = ACTIONS(1539),
    [anon_sym_SEMI_SEMI] = ACTIONS(1541),
    [anon_sym_let] = ACTIONS(1539),
    [anon_sym_and] = ACTIONS(1539),
    [anon_sym_TILDE] = ACTIONS(1539),
    [anon_sym_QMARK] = ACTIONS(1539),
    [anon_sym_LPAREN] = ACTIONS(1541),
    [anon_sym_external] = ACTIONS(1539),
    [anon_sym_type] = ACTIONS(1539),
    [anon_sym_COMMA] = ACTIONS(1541),
    [anon_sym_PIPE] = ACTIONS(1539),
    [anon_sym_DOT] = ACTIONS(1541),
    [anon_sym_LBRACK] = ACTIONS(1539),
    [anon_sym_RBRACK] = ACTIONS(1541),
    [anon_sym_true] = ACTIONS(1539),
    [anon_sym_false] = ACTIONS(1539),
    [anon_sym_LBRACE] = ACTIONS(1539),
    [anon_sym_SEMI] = ACTIONS(1539),
    [anon_sym_exception] = ACTIONS(1539),
    [anon_sym_module] = ACTIONS(1539),
    [anon_sym_open] = ACTIONS(1539),
    [anon_sym_include] = ACTIONS(1539),
    [anon_sym_class] = ACTIONS(1539),
    [anon_sym_val] = ACTIONS(1539),
    [anon_sym_end] = ACTIONS(1539),
    [anon_sym_object] = ACTIONS(1539),
    [anon_sym_AMP] = ACTIONS(1539),
    [anon_sym_POUND] = ACTIONS(1539),
    [anon_sym_COLON_COLON] = ACTIONS(1541),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1541),
    [anon_sym_LT_DASH] = ACTIONS(1539),
    [anon_sym_else] = ACTIONS(1539),
    [anon_sym_new] = ACTIONS(1539),
    [anon_sym_LBRACE_LT] = ACTIONS(1541),
    [anon_sym_begin] = ACTIONS(1539),
    [sym_ocamlyacc_value] = ACTIONS(1541),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1539),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1541),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1539),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1539),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1541),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1541),
    [aux_sym_number_token1] = ACTIONS(1541),
    [aux_sym_signed_number_token1] = ACTIONS(1539),
    [anon_sym_SQUOTE] = ACTIONS(1541),
    [anon_sym_DQUOTE] = ACTIONS(1541),
    [sym_prefix_operator] = ACTIONS(1541),
    [aux_sym_sign_operator_token1] = ACTIONS(1539),
    [sym_hash_operator] = ACTIONS(1541),
    [anon_sym_lsl] = ACTIONS(1539),
    [anon_sym_lsr] = ACTIONS(1539),
    [anon_sym_asr] = ACTIONS(1539),
    [aux_sym_pow_operator_token1] = ACTIONS(1541),
    [anon_sym_mod] = ACTIONS(1539),
    [anon_sym_land] = ACTIONS(1539),
    [anon_sym_lor] = ACTIONS(1539),
    [anon_sym_lxor] = ACTIONS(1539),
    [aux_sym_mult_operator_token1] = ACTIONS(1539),
    [aux_sym_add_operator_token1] = ACTIONS(1539),
    [sym_concat_operator] = ACTIONS(1541),
    [sym_rel_operator] = ACTIONS(1539),
    [anon_sym_AMP_AMP] = ACTIONS(1539),
    [anon_sym_or] = ACTIONS(1539),
    [anon_sym_PIPE_PIPE] = ACTIONS(1539),
    [sym_assign_operator] = ACTIONS(1541),
    [sym_let_operator] = ACTIONS(1541),
    [sym__uppercase_identifier] = ACTIONS(1541),
    [aux_sym_directive_token1] = ACTIONS(1539),
    [aux_sym_tag_token1] = ACTIONS(1541),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1037)] = {
    [sym_attribute] = STATE(1037),
    [sym__lowercase_identifier] = ACTIONS(1615),
    [anon_sym_TILDE] = ACTIONS(1615),
    [anon_sym_QMARK] = ACTIONS(1615),
    [anon_sym_LPAREN] = ACTIONS(1617),
    [anon_sym_RPAREN] = ACTIONS(1617),
    [anon_sym_COMMA] = ACTIONS(1617),
    [anon_sym_PIPE] = ACTIONS(1615),
    [anon_sym_COLON2] = ACTIONS(1615),
    [anon_sym_DASH_GT] = ACTIONS(1615),
    [anon_sym_LBRACK] = ACTIONS(1615),
    [anon_sym_RBRACK] = ACTIONS(1617),
    [anon_sym_true] = ACTIONS(1615),
    [anon_sym_false] = ACTIONS(1615),
    [anon_sym_LBRACE] = ACTIONS(1615),
    [anon_sym_SEMI] = ACTIONS(1617),
    [anon_sym_RBRACE] = ACTIONS(1617),
    [anon_sym_constraint] = ACTIONS(1615),
    [anon_sym_val] = ACTIONS(1615),
    [anon_sym_end] = ACTIONS(1615),
    [anon_sym_with] = ACTIONS(1615),
    [anon_sym_COLON_GT] = ACTIONS(1617),
    [anon_sym_object] = ACTIONS(1615),
    [anon_sym_inherit] = ACTIONS(1615),
    [anon_sym_method] = ACTIONS(1615),
    [anon_sym_initializer] = ACTIONS(1615),
    [anon_sym_AMP] = ACTIONS(1615),
    [anon_sym_COLON_COLON] = ACTIONS(1617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1617),
    [anon_sym_then] = ACTIONS(1615),
    [anon_sym_else] = ACTIONS(1615),
    [anon_sym_do] = ACTIONS(1615),
    [anon_sym_done] = ACTIONS(1615),
    [anon_sym_to] = ACTIONS(1615),
    [anon_sym_downto] = ACTIONS(1615),
    [anon_sym_new] = ACTIONS(1615),
    [anon_sym_LBRACE_LT] = ACTIONS(1617),
    [anon_sym_GT_RBRACE] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1615),
    [sym_ocamlyacc_value] = ACTIONS(1617),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1615),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1617),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1615),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1615),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1617),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1617),
    [aux_sym_number_token1] = ACTIONS(1617),
    [aux_sym_signed_number_token1] = ACTIONS(1615),
    [anon_sym_SQUOTE] = ACTIONS(1617),
    [anon_sym_DQUOTE] = ACTIONS(1617),
    [sym_prefix_operator] = ACTIONS(1617),
    [aux_sym_sign_operator_token1] = ACTIONS(1615),
    [anon_sym_lsl] = ACTIONS(1615),
    [anon_sym_lsr] = ACTIONS(1615),
    [anon_sym_asr] = ACTIONS(1615),
    [aux_sym_pow_operator_token1] = ACTIONS(1617),
    [anon_sym_mod] = ACTIONS(1615),
    [anon_sym_land] = ACTIONS(1615),
    [anon_sym_lor] = ACTIONS(1615),
    [anon_sym_lxor] = ACTIONS(1615),
    [aux_sym_mult_operator_token1] = ACTIONS(1615),
    [aux_sym_add_operator_token1] = ACTIONS(1615),
    [sym_concat_operator] = ACTIONS(1617),
    [sym_rel_operator] = ACTIONS(1615),
    [anon_sym_AMP_AMP] = ACTIONS(1615),
    [anon_sym_or] = ACTIONS(1615),
    [anon_sym_PIPE_PIPE] = ACTIONS(1615),
    [sym_assign_operator] = ACTIONS(1617),
    [sym__uppercase_identifier] = ACTIONS(1617),
    [aux_sym_tag_token1] = ACTIONS(1617),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1038)] = {
    [sym__parenthesized_abstract_type] = STATE(1333),
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4560),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6673),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1038),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(4808),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1038),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1120),
    [sym__lowercase_identifier] = ACTIONS(2180),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(2182),
    [anon_sym_COMMA] = ACTIONS(1295),
    [anon_sym_PIPE] = ACTIONS(1295),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_RBRACK] = ACTIONS(1295),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_SEMI] = ACTIONS(1295),
    [anon_sym_RBRACE] = ACTIONS(1295),
    [anon_sym_exception] = ACTIONS(2184),
    [anon_sym_as] = ACTIONS(1297),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_COLON_COLON] = ACTIONS(1295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(2188),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2190),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1039)] = {
    [sym_attribute] = STATE(1039),
    [sym__lowercase_identifier] = ACTIONS(1531),
    [anon_sym_SEMI_SEMI] = ACTIONS(1533),
    [anon_sym_let] = ACTIONS(1531),
    [anon_sym_and] = ACTIONS(1531),
    [anon_sym_TILDE] = ACTIONS(1531),
    [anon_sym_QMARK] = ACTIONS(1531),
    [anon_sym_LPAREN] = ACTIONS(1533),
    [anon_sym_external] = ACTIONS(1531),
    [anon_sym_type] = ACTIONS(1531),
    [anon_sym_COMMA] = ACTIONS(1533),
    [anon_sym_PIPE] = ACTIONS(1531),
    [anon_sym_DOT] = ACTIONS(1533),
    [anon_sym_LBRACK] = ACTIONS(1531),
    [anon_sym_RBRACK] = ACTIONS(1533),
    [anon_sym_true] = ACTIONS(1531),
    [anon_sym_false] = ACTIONS(1531),
    [anon_sym_LBRACE] = ACTIONS(1531),
    [anon_sym_SEMI] = ACTIONS(1531),
    [anon_sym_exception] = ACTIONS(1531),
    [anon_sym_module] = ACTIONS(1531),
    [anon_sym_open] = ACTIONS(1531),
    [anon_sym_include] = ACTIONS(1531),
    [anon_sym_class] = ACTIONS(1531),
    [anon_sym_val] = ACTIONS(1531),
    [anon_sym_end] = ACTIONS(1531),
    [anon_sym_object] = ACTIONS(1531),
    [anon_sym_AMP] = ACTIONS(1531),
    [anon_sym_POUND] = ACTIONS(1531),
    [anon_sym_COLON_COLON] = ACTIONS(1533),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1533),
    [anon_sym_LT_DASH] = ACTIONS(1531),
    [anon_sym_else] = ACTIONS(1531),
    [anon_sym_new] = ACTIONS(1531),
    [anon_sym_LBRACE_LT] = ACTIONS(1533),
    [anon_sym_begin] = ACTIONS(1531),
    [sym_ocamlyacc_value] = ACTIONS(1533),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1531),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1533),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1531),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1531),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1533),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1533),
    [aux_sym_number_token1] = ACTIONS(1533),
    [aux_sym_signed_number_token1] = ACTIONS(1531),
    [anon_sym_SQUOTE] = ACTIONS(1533),
    [anon_sym_DQUOTE] = ACTIONS(1533),
    [sym_prefix_operator] = ACTIONS(1533),
    [aux_sym_sign_operator_token1] = ACTIONS(1531),
    [sym_hash_operator] = ACTIONS(1533),
    [anon_sym_lsl] = ACTIONS(1531),
    [anon_sym_lsr] = ACTIONS(1531),
    [anon_sym_asr] = ACTIONS(1531),
    [aux_sym_pow_operator_token1] = ACTIONS(1533),
    [anon_sym_mod] = ACTIONS(1531),
    [anon_sym_land] = ACTIONS(1531),
    [anon_sym_lor] = ACTIONS(1531),
    [anon_sym_lxor] = ACTIONS(1531),
    [aux_sym_mult_operator_token1] = ACTIONS(1531),
    [aux_sym_add_operator_token1] = ACTIONS(1531),
    [sym_concat_operator] = ACTIONS(1533),
    [sym_rel_operator] = ACTIONS(1531),
    [anon_sym_AMP_AMP] = ACTIONS(1531),
    [anon_sym_or] = ACTIONS(1531),
    [anon_sym_PIPE_PIPE] = ACTIONS(1531),
    [sym_assign_operator] = ACTIONS(1533),
    [sym_let_operator] = ACTIONS(1533),
    [sym__uppercase_identifier] = ACTIONS(1533),
    [aux_sym_directive_token1] = ACTIONS(1531),
    [aux_sym_tag_token1] = ACTIONS(1533),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1040)] = {
    [sym_attribute] = STATE(1040),
    [sym__lowercase_identifier] = ACTIONS(1535),
    [anon_sym_SEMI_SEMI] = ACTIONS(1537),
    [anon_sym_let] = ACTIONS(1535),
    [anon_sym_and] = ACTIONS(1535),
    [anon_sym_TILDE] = ACTIONS(1535),
    [anon_sym_QMARK] = ACTIONS(1535),
    [anon_sym_LPAREN] = ACTIONS(1537),
    [anon_sym_external] = ACTIONS(1535),
    [anon_sym_type] = ACTIONS(1535),
    [anon_sym_COMMA] = ACTIONS(1537),
    [anon_sym_PIPE] = ACTIONS(1535),
    [anon_sym_DOT] = ACTIONS(1537),
    [anon_sym_LBRACK] = ACTIONS(1535),
    [anon_sym_RBRACK] = ACTIONS(1537),
    [anon_sym_true] = ACTIONS(1535),
    [anon_sym_false] = ACTIONS(1535),
    [anon_sym_LBRACE] = ACTIONS(1535),
    [anon_sym_SEMI] = ACTIONS(1535),
    [anon_sym_exception] = ACTIONS(1535),
    [anon_sym_module] = ACTIONS(1535),
    [anon_sym_open] = ACTIONS(1535),
    [anon_sym_include] = ACTIONS(1535),
    [anon_sym_class] = ACTIONS(1535),
    [anon_sym_val] = ACTIONS(1535),
    [anon_sym_end] = ACTIONS(1535),
    [anon_sym_object] = ACTIONS(1535),
    [anon_sym_AMP] = ACTIONS(1535),
    [anon_sym_POUND] = ACTIONS(1535),
    [anon_sym_COLON_COLON] = ACTIONS(1537),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1537),
    [anon_sym_LT_DASH] = ACTIONS(1535),
    [anon_sym_else] = ACTIONS(1535),
    [anon_sym_new] = ACTIONS(1535),
    [anon_sym_LBRACE_LT] = ACTIONS(1537),
    [anon_sym_begin] = ACTIONS(1535),
    [sym_ocamlyacc_value] = ACTIONS(1537),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1535),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1537),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1535),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1535),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1537),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1537),
    [aux_sym_number_token1] = ACTIONS(1537),
    [aux_sym_signed_number_token1] = ACTIONS(1535),
    [anon_sym_SQUOTE] = ACTIONS(1537),
    [anon_sym_DQUOTE] = ACTIONS(1537),
    [sym_prefix_operator] = ACTIONS(1537),
    [aux_sym_sign_operator_token1] = ACTIONS(1535),
    [sym_hash_operator] = ACTIONS(1537),
    [anon_sym_lsl] = ACTIONS(1535),
    [anon_sym_lsr] = ACTIONS(1535),
    [anon_sym_asr] = ACTIONS(1535),
    [aux_sym_pow_operator_token1] = ACTIONS(1537),
    [anon_sym_mod] = ACTIONS(1535),
    [anon_sym_land] = ACTIONS(1535),
    [anon_sym_lor] = ACTIONS(1535),
    [anon_sym_lxor] = ACTIONS(1535),
    [aux_sym_mult_operator_token1] = ACTIONS(1535),
    [aux_sym_add_operator_token1] = ACTIONS(1535),
    [sym_concat_operator] = ACTIONS(1537),
    [sym_rel_operator] = ACTIONS(1535),
    [anon_sym_AMP_AMP] = ACTIONS(1535),
    [anon_sym_or] = ACTIONS(1535),
    [anon_sym_PIPE_PIPE] = ACTIONS(1535),
    [sym_assign_operator] = ACTIONS(1537),
    [sym_let_operator] = ACTIONS(1537),
    [sym__uppercase_identifier] = ACTIONS(1537),
    [aux_sym_directive_token1] = ACTIONS(1535),
    [aux_sym_tag_token1] = ACTIONS(1537),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1041)] = {
    [sym_attribute] = STATE(1041),
    [sym__lowercase_identifier] = ACTIONS(2017),
    [anon_sym_TILDE] = ACTIONS(2017),
    [anon_sym_QMARK] = ACTIONS(2017),
    [anon_sym_LPAREN] = ACTIONS(2019),
    [anon_sym_RPAREN] = ACTIONS(2019),
    [anon_sym_COMMA] = ACTIONS(2019),
    [anon_sym_PIPE] = ACTIONS(2017),
    [anon_sym_COLON2] = ACTIONS(2017),
    [anon_sym_DASH_GT] = ACTIONS(2017),
    [anon_sym_LBRACK] = ACTIONS(2017),
    [anon_sym_RBRACK] = ACTIONS(2019),
    [anon_sym_true] = ACTIONS(2017),
    [anon_sym_false] = ACTIONS(2017),
    [anon_sym_LBRACE] = ACTIONS(2017),
    [anon_sym_SEMI] = ACTIONS(2019),
    [anon_sym_RBRACE] = ACTIONS(2019),
    [anon_sym_constraint] = ACTIONS(2017),
    [anon_sym_val] = ACTIONS(2017),
    [anon_sym_end] = ACTIONS(2017),
    [anon_sym_with] = ACTIONS(2017),
    [anon_sym_COLON_GT] = ACTIONS(2019),
    [anon_sym_object] = ACTIONS(2017),
    [anon_sym_inherit] = ACTIONS(2017),
    [anon_sym_method] = ACTIONS(2017),
    [anon_sym_initializer] = ACTIONS(2017),
    [anon_sym_AMP] = ACTIONS(2017),
    [anon_sym_COLON_COLON] = ACTIONS(2019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2019),
    [anon_sym_then] = ACTIONS(2017),
    [anon_sym_else] = ACTIONS(2017),
    [anon_sym_do] = ACTIONS(2017),
    [anon_sym_done] = ACTIONS(2017),
    [anon_sym_to] = ACTIONS(2017),
    [anon_sym_downto] = ACTIONS(2017),
    [anon_sym_new] = ACTIONS(2017),
    [anon_sym_LBRACE_LT] = ACTIONS(2019),
    [anon_sym_GT_RBRACE] = ACTIONS(2019),
    [anon_sym_begin] = ACTIONS(2017),
    [sym_ocamlyacc_value] = ACTIONS(2019),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2017),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2019),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2019),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2019),
    [aux_sym_number_token1] = ACTIONS(2019),
    [aux_sym_signed_number_token1] = ACTIONS(2017),
    [anon_sym_SQUOTE] = ACTIONS(2019),
    [anon_sym_DQUOTE] = ACTIONS(2019),
    [sym_prefix_operator] = ACTIONS(2019),
    [aux_sym_sign_operator_token1] = ACTIONS(2017),
    [anon_sym_lsl] = ACTIONS(2017),
    [anon_sym_lsr] = ACTIONS(2017),
    [anon_sym_asr] = ACTIONS(2017),
    [aux_sym_pow_operator_token1] = ACTIONS(2019),
    [anon_sym_mod] = ACTIONS(2017),
    [anon_sym_land] = ACTIONS(2017),
    [anon_sym_lor] = ACTIONS(2017),
    [anon_sym_lxor] = ACTIONS(2017),
    [aux_sym_mult_operator_token1] = ACTIONS(2017),
    [aux_sym_add_operator_token1] = ACTIONS(2017),
    [sym_concat_operator] = ACTIONS(2019),
    [sym_rel_operator] = ACTIONS(2017),
    [anon_sym_AMP_AMP] = ACTIONS(2017),
    [anon_sym_or] = ACTIONS(2017),
    [anon_sym_PIPE_PIPE] = ACTIONS(2017),
    [sym_assign_operator] = ACTIONS(2019),
    [sym__uppercase_identifier] = ACTIONS(2019),
    [aux_sym_tag_token1] = ACTIONS(2019),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1042)] = {
    [sym__parenthesized_abstract_type] = STATE(1330),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4351),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(6815),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1042),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(1042),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1097),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(2079),
    [anon_sym_COMMA] = ACTIONS(2075),
    [anon_sym_PIPE] = ACTIONS(2075),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_RBRACK] = ACTIONS(2075),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_SEMI] = ACTIONS(2075),
    [anon_sym_RBRACE] = ACTIONS(2075),
    [anon_sym_exception] = ACTIONS(2198),
    [anon_sym_as] = ACTIONS(2081),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_COLON_COLON] = ACTIONS(2075),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(2200),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2202),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1043)] = {
    [sym_attribute] = STATE(1043),
    [sym__lowercase_identifier] = ACTIONS(1555),
    [anon_sym_SEMI_SEMI] = ACTIONS(1557),
    [anon_sym_let] = ACTIONS(1555),
    [anon_sym_and] = ACTIONS(1555),
    [anon_sym_TILDE] = ACTIONS(1555),
    [anon_sym_QMARK] = ACTIONS(1555),
    [anon_sym_LPAREN] = ACTIONS(1557),
    [anon_sym_external] = ACTIONS(1555),
    [anon_sym_type] = ACTIONS(1555),
    [anon_sym_COMMA] = ACTIONS(1557),
    [anon_sym_PIPE] = ACTIONS(1555),
    [anon_sym_DOT] = ACTIONS(1557),
    [anon_sym_LBRACK] = ACTIONS(1555),
    [anon_sym_RBRACK] = ACTIONS(1557),
    [anon_sym_true] = ACTIONS(1555),
    [anon_sym_false] = ACTIONS(1555),
    [anon_sym_LBRACE] = ACTIONS(1555),
    [anon_sym_SEMI] = ACTIONS(1555),
    [anon_sym_exception] = ACTIONS(1555),
    [anon_sym_module] = ACTIONS(1555),
    [anon_sym_open] = ACTIONS(1555),
    [anon_sym_include] = ACTIONS(1555),
    [anon_sym_class] = ACTIONS(1555),
    [anon_sym_val] = ACTIONS(1555),
    [anon_sym_end] = ACTIONS(1555),
    [anon_sym_object] = ACTIONS(1555),
    [anon_sym_AMP] = ACTIONS(1555),
    [anon_sym_POUND] = ACTIONS(1555),
    [anon_sym_COLON_COLON] = ACTIONS(1557),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1557),
    [anon_sym_LT_DASH] = ACTIONS(1555),
    [anon_sym_else] = ACTIONS(1555),
    [anon_sym_new] = ACTIONS(1555),
    [anon_sym_LBRACE_LT] = ACTIONS(1557),
    [anon_sym_begin] = ACTIONS(1555),
    [sym_ocamlyacc_value] = ACTIONS(1557),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1555),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1557),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1555),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1555),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1557),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1557),
    [aux_sym_number_token1] = ACTIONS(1557),
    [aux_sym_signed_number_token1] = ACTIONS(1555),
    [anon_sym_SQUOTE] = ACTIONS(1557),
    [anon_sym_DQUOTE] = ACTIONS(1557),
    [sym_prefix_operator] = ACTIONS(1557),
    [aux_sym_sign_operator_token1] = ACTIONS(1555),
    [sym_hash_operator] = ACTIONS(1557),
    [anon_sym_lsl] = ACTIONS(1555),
    [anon_sym_lsr] = ACTIONS(1555),
    [anon_sym_asr] = ACTIONS(1555),
    [aux_sym_pow_operator_token1] = ACTIONS(1557),
    [anon_sym_mod] = ACTIONS(1555),
    [anon_sym_land] = ACTIONS(1555),
    [anon_sym_lor] = ACTIONS(1555),
    [anon_sym_lxor] = ACTIONS(1555),
    [aux_sym_mult_operator_token1] = ACTIONS(1555),
    [aux_sym_add_operator_token1] = ACTIONS(1555),
    [sym_concat_operator] = ACTIONS(1557),
    [sym_rel_operator] = ACTIONS(1555),
    [anon_sym_AMP_AMP] = ACTIONS(1555),
    [anon_sym_or] = ACTIONS(1555),
    [anon_sym_PIPE_PIPE] = ACTIONS(1555),
    [sym_assign_operator] = ACTIONS(1557),
    [sym_let_operator] = ACTIONS(1557),
    [sym__uppercase_identifier] = ACTIONS(1557),
    [aux_sym_directive_token1] = ACTIONS(1555),
    [aux_sym_tag_token1] = ACTIONS(1557),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1044)] = {
    [sym_attribute] = STATE(1044),
    [sym__lowercase_identifier] = ACTIONS(1551),
    [anon_sym_SEMI_SEMI] = ACTIONS(1553),
    [anon_sym_let] = ACTIONS(1551),
    [anon_sym_and] = ACTIONS(1551),
    [anon_sym_TILDE] = ACTIONS(1551),
    [anon_sym_QMARK] = ACTIONS(1551),
    [anon_sym_LPAREN] = ACTIONS(1553),
    [anon_sym_external] = ACTIONS(1551),
    [anon_sym_type] = ACTIONS(1551),
    [anon_sym_COMMA] = ACTIONS(1553),
    [anon_sym_PIPE] = ACTIONS(1551),
    [anon_sym_DOT] = ACTIONS(1553),
    [anon_sym_LBRACK] = ACTIONS(1551),
    [anon_sym_RBRACK] = ACTIONS(1553),
    [anon_sym_true] = ACTIONS(1551),
    [anon_sym_false] = ACTIONS(1551),
    [anon_sym_LBRACE] = ACTIONS(1551),
    [anon_sym_SEMI] = ACTIONS(1551),
    [anon_sym_exception] = ACTIONS(1551),
    [anon_sym_module] = ACTIONS(1551),
    [anon_sym_open] = ACTIONS(1551),
    [anon_sym_include] = ACTIONS(1551),
    [anon_sym_class] = ACTIONS(1551),
    [anon_sym_val] = ACTIONS(1551),
    [anon_sym_end] = ACTIONS(1551),
    [anon_sym_object] = ACTIONS(1551),
    [anon_sym_AMP] = ACTIONS(1551),
    [anon_sym_POUND] = ACTIONS(1551),
    [anon_sym_COLON_COLON] = ACTIONS(1553),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1553),
    [anon_sym_LT_DASH] = ACTIONS(1551),
    [anon_sym_else] = ACTIONS(1551),
    [anon_sym_new] = ACTIONS(1551),
    [anon_sym_LBRACE_LT] = ACTIONS(1553),
    [anon_sym_begin] = ACTIONS(1551),
    [sym_ocamlyacc_value] = ACTIONS(1553),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1551),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1553),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1551),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1551),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1553),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1553),
    [aux_sym_number_token1] = ACTIONS(1553),
    [aux_sym_signed_number_token1] = ACTIONS(1551),
    [anon_sym_SQUOTE] = ACTIONS(1553),
    [anon_sym_DQUOTE] = ACTIONS(1553),
    [sym_prefix_operator] = ACTIONS(1553),
    [aux_sym_sign_operator_token1] = ACTIONS(1551),
    [sym_hash_operator] = ACTIONS(1553),
    [anon_sym_lsl] = ACTIONS(1551),
    [anon_sym_lsr] = ACTIONS(1551),
    [anon_sym_asr] = ACTIONS(1551),
    [aux_sym_pow_operator_token1] = ACTIONS(1553),
    [anon_sym_mod] = ACTIONS(1551),
    [anon_sym_land] = ACTIONS(1551),
    [anon_sym_lor] = ACTIONS(1551),
    [anon_sym_lxor] = ACTIONS(1551),
    [aux_sym_mult_operator_token1] = ACTIONS(1551),
    [aux_sym_add_operator_token1] = ACTIONS(1551),
    [sym_concat_operator] = ACTIONS(1553),
    [sym_rel_operator] = ACTIONS(1551),
    [anon_sym_AMP_AMP] = ACTIONS(1551),
    [anon_sym_or] = ACTIONS(1551),
    [anon_sym_PIPE_PIPE] = ACTIONS(1551),
    [sym_assign_operator] = ACTIONS(1553),
    [sym_let_operator] = ACTIONS(1553),
    [sym__uppercase_identifier] = ACTIONS(1553),
    [aux_sym_directive_token1] = ACTIONS(1551),
    [aux_sym_tag_token1] = ACTIONS(1553),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1045)] = {
    [sym_attribute] = STATE(1045),
    [sym__lowercase_identifier] = ACTIONS(1543),
    [anon_sym_SEMI_SEMI] = ACTIONS(1545),
    [anon_sym_let] = ACTIONS(1543),
    [anon_sym_and] = ACTIONS(1543),
    [anon_sym_TILDE] = ACTIONS(1543),
    [anon_sym_QMARK] = ACTIONS(1543),
    [anon_sym_LPAREN] = ACTIONS(1545),
    [anon_sym_external] = ACTIONS(1543),
    [anon_sym_type] = ACTIONS(1543),
    [anon_sym_COMMA] = ACTIONS(1545),
    [anon_sym_PIPE] = ACTIONS(1543),
    [anon_sym_DOT] = ACTIONS(1545),
    [anon_sym_LBRACK] = ACTIONS(1543),
    [anon_sym_RBRACK] = ACTIONS(1545),
    [anon_sym_true] = ACTIONS(1543),
    [anon_sym_false] = ACTIONS(1543),
    [anon_sym_LBRACE] = ACTIONS(1543),
    [anon_sym_SEMI] = ACTIONS(1543),
    [anon_sym_exception] = ACTIONS(1543),
    [anon_sym_module] = ACTIONS(1543),
    [anon_sym_open] = ACTIONS(1543),
    [anon_sym_include] = ACTIONS(1543),
    [anon_sym_class] = ACTIONS(1543),
    [anon_sym_val] = ACTIONS(1543),
    [anon_sym_end] = ACTIONS(1543),
    [anon_sym_object] = ACTIONS(1543),
    [anon_sym_AMP] = ACTIONS(1543),
    [anon_sym_POUND] = ACTIONS(1543),
    [anon_sym_COLON_COLON] = ACTIONS(1545),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1545),
    [anon_sym_LT_DASH] = ACTIONS(1543),
    [anon_sym_else] = ACTIONS(1543),
    [anon_sym_new] = ACTIONS(1543),
    [anon_sym_LBRACE_LT] = ACTIONS(1545),
    [anon_sym_begin] = ACTIONS(1543),
    [sym_ocamlyacc_value] = ACTIONS(1545),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1543),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1545),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1543),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1543),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1545),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1545),
    [aux_sym_number_token1] = ACTIONS(1545),
    [aux_sym_signed_number_token1] = ACTIONS(1543),
    [anon_sym_SQUOTE] = ACTIONS(1545),
    [anon_sym_DQUOTE] = ACTIONS(1545),
    [sym_prefix_operator] = ACTIONS(1545),
    [aux_sym_sign_operator_token1] = ACTIONS(1543),
    [sym_hash_operator] = ACTIONS(1545),
    [anon_sym_lsl] = ACTIONS(1543),
    [anon_sym_lsr] = ACTIONS(1543),
    [anon_sym_asr] = ACTIONS(1543),
    [aux_sym_pow_operator_token1] = ACTIONS(1545),
    [anon_sym_mod] = ACTIONS(1543),
    [anon_sym_land] = ACTIONS(1543),
    [anon_sym_lor] = ACTIONS(1543),
    [anon_sym_lxor] = ACTIONS(1543),
    [aux_sym_mult_operator_token1] = ACTIONS(1543),
    [aux_sym_add_operator_token1] = ACTIONS(1543),
    [sym_concat_operator] = ACTIONS(1545),
    [sym_rel_operator] = ACTIONS(1543),
    [anon_sym_AMP_AMP] = ACTIONS(1543),
    [anon_sym_or] = ACTIONS(1543),
    [anon_sym_PIPE_PIPE] = ACTIONS(1543),
    [sym_assign_operator] = ACTIONS(1545),
    [sym_let_operator] = ACTIONS(1545),
    [sym__uppercase_identifier] = ACTIONS(1545),
    [aux_sym_directive_token1] = ACTIONS(1543),
    [aux_sym_tag_token1] = ACTIONS(1545),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1046)] = {
    [sym_attribute] = STATE(1046),
    [sym__lowercase_identifier] = ACTIONS(1547),
    [anon_sym_SEMI_SEMI] = ACTIONS(1549),
    [anon_sym_let] = ACTIONS(1547),
    [anon_sym_and] = ACTIONS(1547),
    [anon_sym_TILDE] = ACTIONS(1547),
    [anon_sym_QMARK] = ACTIONS(1547),
    [anon_sym_LPAREN] = ACTIONS(1549),
    [anon_sym_external] = ACTIONS(1547),
    [anon_sym_type] = ACTIONS(1547),
    [anon_sym_COMMA] = ACTIONS(1549),
    [anon_sym_PIPE] = ACTIONS(1547),
    [anon_sym_DOT] = ACTIONS(1549),
    [anon_sym_LBRACK] = ACTIONS(1547),
    [anon_sym_RBRACK] = ACTIONS(1549),
    [anon_sym_true] = ACTIONS(1547),
    [anon_sym_false] = ACTIONS(1547),
    [anon_sym_LBRACE] = ACTIONS(1547),
    [anon_sym_SEMI] = ACTIONS(1547),
    [anon_sym_exception] = ACTIONS(1547),
    [anon_sym_module] = ACTIONS(1547),
    [anon_sym_open] = ACTIONS(1547),
    [anon_sym_include] = ACTIONS(1547),
    [anon_sym_class] = ACTIONS(1547),
    [anon_sym_val] = ACTIONS(1547),
    [anon_sym_end] = ACTIONS(1547),
    [anon_sym_object] = ACTIONS(1547),
    [anon_sym_AMP] = ACTIONS(1547),
    [anon_sym_POUND] = ACTIONS(1547),
    [anon_sym_COLON_COLON] = ACTIONS(1549),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1549),
    [anon_sym_LT_DASH] = ACTIONS(1547),
    [anon_sym_else] = ACTIONS(1547),
    [anon_sym_new] = ACTIONS(1547),
    [anon_sym_LBRACE_LT] = ACTIONS(1549),
    [anon_sym_begin] = ACTIONS(1547),
    [sym_ocamlyacc_value] = ACTIONS(1549),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1547),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1549),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1547),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1547),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1549),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1549),
    [aux_sym_number_token1] = ACTIONS(1549),
    [aux_sym_signed_number_token1] = ACTIONS(1547),
    [anon_sym_SQUOTE] = ACTIONS(1549),
    [anon_sym_DQUOTE] = ACTIONS(1549),
    [sym_prefix_operator] = ACTIONS(1549),
    [aux_sym_sign_operator_token1] = ACTIONS(1547),
    [sym_hash_operator] = ACTIONS(1549),
    [anon_sym_lsl] = ACTIONS(1547),
    [anon_sym_lsr] = ACTIONS(1547),
    [anon_sym_asr] = ACTIONS(1547),
    [aux_sym_pow_operator_token1] = ACTIONS(1549),
    [anon_sym_mod] = ACTIONS(1547),
    [anon_sym_land] = ACTIONS(1547),
    [anon_sym_lor] = ACTIONS(1547),
    [anon_sym_lxor] = ACTIONS(1547),
    [aux_sym_mult_operator_token1] = ACTIONS(1547),
    [aux_sym_add_operator_token1] = ACTIONS(1547),
    [sym_concat_operator] = ACTIONS(1549),
    [sym_rel_operator] = ACTIONS(1547),
    [anon_sym_AMP_AMP] = ACTIONS(1547),
    [anon_sym_or] = ACTIONS(1547),
    [anon_sym_PIPE_PIPE] = ACTIONS(1547),
    [sym_assign_operator] = ACTIONS(1549),
    [sym_let_operator] = ACTIONS(1549),
    [sym__uppercase_identifier] = ACTIONS(1549),
    [aux_sym_directive_token1] = ACTIONS(1547),
    [aux_sym_tag_token1] = ACTIONS(1549),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1047)] = {
    [sym_attribute] = STATE(1047),
    [sym__lowercase_identifier] = ACTIONS(1611),
    [anon_sym_SEMI_SEMI] = ACTIONS(1613),
    [anon_sym_let] = ACTIONS(1611),
    [anon_sym_and] = ACTIONS(1611),
    [anon_sym_TILDE] = ACTIONS(1611),
    [anon_sym_QMARK] = ACTIONS(1611),
    [anon_sym_LPAREN] = ACTIONS(1613),
    [anon_sym_external] = ACTIONS(1611),
    [anon_sym_type] = ACTIONS(1611),
    [anon_sym_COMMA] = ACTIONS(1613),
    [anon_sym_PIPE] = ACTIONS(1611),
    [anon_sym_DOT] = ACTIONS(1613),
    [anon_sym_LBRACK] = ACTIONS(1611),
    [anon_sym_RBRACK] = ACTIONS(1613),
    [anon_sym_true] = ACTIONS(1611),
    [anon_sym_false] = ACTIONS(1611),
    [anon_sym_LBRACE] = ACTIONS(1611),
    [anon_sym_SEMI] = ACTIONS(1611),
    [anon_sym_exception] = ACTIONS(1611),
    [anon_sym_module] = ACTIONS(1611),
    [anon_sym_open] = ACTIONS(1611),
    [anon_sym_include] = ACTIONS(1611),
    [anon_sym_class] = ACTIONS(1611),
    [anon_sym_val] = ACTIONS(1611),
    [anon_sym_end] = ACTIONS(1611),
    [anon_sym_object] = ACTIONS(1611),
    [anon_sym_AMP] = ACTIONS(1611),
    [anon_sym_POUND] = ACTIONS(1611),
    [anon_sym_COLON_COLON] = ACTIONS(1613),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1613),
    [anon_sym_else] = ACTIONS(1611),
    [anon_sym_new] = ACTIONS(1611),
    [anon_sym_LBRACE_LT] = ACTIONS(1613),
    [anon_sym_begin] = ACTIONS(1611),
    [sym_ocamlyacc_value] = ACTIONS(1613),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1611),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1613),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1611),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1611),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1613),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1613),
    [aux_sym_number_token1] = ACTIONS(1613),
    [aux_sym_signed_number_token1] = ACTIONS(1611),
    [anon_sym_SQUOTE] = ACTIONS(1613),
    [anon_sym_DQUOTE] = ACTIONS(1613),
    [sym_prefix_operator] = ACTIONS(1613),
    [aux_sym_sign_operator_token1] = ACTIONS(1611),
    [sym_hash_operator] = ACTIONS(1613),
    [anon_sym_lsl] = ACTIONS(1611),
    [anon_sym_lsr] = ACTIONS(1611),
    [anon_sym_asr] = ACTIONS(1611),
    [aux_sym_pow_operator_token1] = ACTIONS(1613),
    [anon_sym_mod] = ACTIONS(1611),
    [anon_sym_land] = ACTIONS(1611),
    [anon_sym_lor] = ACTIONS(1611),
    [anon_sym_lxor] = ACTIONS(1611),
    [aux_sym_mult_operator_token1] = ACTIONS(1611),
    [aux_sym_add_operator_token1] = ACTIONS(1611),
    [sym_concat_operator] = ACTIONS(1613),
    [sym_rel_operator] = ACTIONS(1611),
    [anon_sym_AMP_AMP] = ACTIONS(1611),
    [anon_sym_or] = ACTIONS(1611),
    [anon_sym_PIPE_PIPE] = ACTIONS(1611),
    [sym_assign_operator] = ACTIONS(1613),
    [sym_let_operator] = ACTIONS(1613),
    [sym__uppercase_identifier] = ACTIONS(1613),
    [aux_sym_directive_token1] = ACTIONS(1611),
    [aux_sym_tag_token1] = ACTIONS(1613),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1048)] = {
    [sym_attribute] = STATE(1048),
    [sym__lowercase_identifier] = ACTIONS(1591),
    [anon_sym_SEMI_SEMI] = ACTIONS(1593),
    [anon_sym_let] = ACTIONS(1591),
    [anon_sym_and] = ACTIONS(1591),
    [anon_sym_TILDE] = ACTIONS(1591),
    [anon_sym_QMARK] = ACTIONS(1591),
    [anon_sym_LPAREN] = ACTIONS(1593),
    [anon_sym_external] = ACTIONS(1591),
    [anon_sym_type] = ACTIONS(1591),
    [anon_sym_COMMA] = ACTIONS(1593),
    [anon_sym_PIPE] = ACTIONS(1591),
    [anon_sym_DOT] = ACTIONS(1593),
    [anon_sym_LBRACK] = ACTIONS(1591),
    [anon_sym_RBRACK] = ACTIONS(1593),
    [anon_sym_true] = ACTIONS(1591),
    [anon_sym_false] = ACTIONS(1591),
    [anon_sym_LBRACE] = ACTIONS(1591),
    [anon_sym_SEMI] = ACTIONS(1591),
    [anon_sym_exception] = ACTIONS(1591),
    [anon_sym_module] = ACTIONS(1591),
    [anon_sym_open] = ACTIONS(1591),
    [anon_sym_include] = ACTIONS(1591),
    [anon_sym_class] = ACTIONS(1591),
    [anon_sym_val] = ACTIONS(1591),
    [anon_sym_end] = ACTIONS(1591),
    [anon_sym_object] = ACTIONS(1591),
    [anon_sym_AMP] = ACTIONS(1591),
    [anon_sym_POUND] = ACTIONS(1591),
    [anon_sym_COLON_COLON] = ACTIONS(1593),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1593),
    [anon_sym_else] = ACTIONS(1591),
    [anon_sym_new] = ACTIONS(1591),
    [anon_sym_LBRACE_LT] = ACTIONS(1593),
    [anon_sym_begin] = ACTIONS(1591),
    [sym_ocamlyacc_value] = ACTIONS(1593),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1591),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1593),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1591),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1591),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1593),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1593),
    [aux_sym_number_token1] = ACTIONS(1593),
    [aux_sym_signed_number_token1] = ACTIONS(1591),
    [anon_sym_SQUOTE] = ACTIONS(1593),
    [anon_sym_DQUOTE] = ACTIONS(1593),
    [sym_prefix_operator] = ACTIONS(1593),
    [aux_sym_sign_operator_token1] = ACTIONS(1591),
    [sym_hash_operator] = ACTIONS(1593),
    [anon_sym_lsl] = ACTIONS(1591),
    [anon_sym_lsr] = ACTIONS(1591),
    [anon_sym_asr] = ACTIONS(1591),
    [aux_sym_pow_operator_token1] = ACTIONS(1593),
    [anon_sym_mod] = ACTIONS(1591),
    [anon_sym_land] = ACTIONS(1591),
    [anon_sym_lor] = ACTIONS(1591),
    [anon_sym_lxor] = ACTIONS(1591),
    [aux_sym_mult_operator_token1] = ACTIONS(1591),
    [aux_sym_add_operator_token1] = ACTIONS(1591),
    [sym_concat_operator] = ACTIONS(1593),
    [sym_rel_operator] = ACTIONS(1591),
    [anon_sym_AMP_AMP] = ACTIONS(1591),
    [anon_sym_or] = ACTIONS(1591),
    [anon_sym_PIPE_PIPE] = ACTIONS(1591),
    [sym_assign_operator] = ACTIONS(1593),
    [sym_let_operator] = ACTIONS(1593),
    [sym__uppercase_identifier] = ACTIONS(1593),
    [aux_sym_directive_token1] = ACTIONS(1591),
    [aux_sym_tag_token1] = ACTIONS(1593),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1049)] = {
    [sym_attribute] = STATE(1049),
    [sym__lowercase_identifier] = ACTIONS(1675),
    [anon_sym_SEMI_SEMI] = ACTIONS(1677),
    [anon_sym_let] = ACTIONS(1675),
    [anon_sym_and] = ACTIONS(1675),
    [anon_sym_TILDE] = ACTIONS(1675),
    [anon_sym_QMARK] = ACTIONS(1675),
    [anon_sym_LPAREN] = ACTIONS(1677),
    [anon_sym_external] = ACTIONS(1675),
    [anon_sym_type] = ACTIONS(1675),
    [anon_sym_COMMA] = ACTIONS(1677),
    [anon_sym_PIPE] = ACTIONS(1675),
    [anon_sym_DOT] = ACTIONS(1677),
    [anon_sym_LBRACK] = ACTIONS(1675),
    [anon_sym_RBRACK] = ACTIONS(1677),
    [anon_sym_true] = ACTIONS(1675),
    [anon_sym_false] = ACTIONS(1675),
    [anon_sym_LBRACE] = ACTIONS(1675),
    [anon_sym_SEMI] = ACTIONS(1675),
    [anon_sym_exception] = ACTIONS(1675),
    [anon_sym_module] = ACTIONS(1675),
    [anon_sym_open] = ACTIONS(1675),
    [anon_sym_include] = ACTIONS(1675),
    [anon_sym_class] = ACTIONS(1675),
    [anon_sym_val] = ACTIONS(1675),
    [anon_sym_end] = ACTIONS(1675),
    [anon_sym_object] = ACTIONS(1675),
    [anon_sym_AMP] = ACTIONS(1675),
    [anon_sym_POUND] = ACTIONS(1675),
    [anon_sym_COLON_COLON] = ACTIONS(1677),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1677),
    [anon_sym_else] = ACTIONS(1675),
    [anon_sym_new] = ACTIONS(1675),
    [anon_sym_LBRACE_LT] = ACTIONS(1677),
    [anon_sym_begin] = ACTIONS(1675),
    [sym_ocamlyacc_value] = ACTIONS(1677),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1675),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1677),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1675),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1675),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1677),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1677),
    [aux_sym_number_token1] = ACTIONS(1677),
    [aux_sym_signed_number_token1] = ACTIONS(1675),
    [anon_sym_SQUOTE] = ACTIONS(1677),
    [anon_sym_DQUOTE] = ACTIONS(1677),
    [sym_prefix_operator] = ACTIONS(1677),
    [aux_sym_sign_operator_token1] = ACTIONS(1675),
    [sym_hash_operator] = ACTIONS(1677),
    [anon_sym_lsl] = ACTIONS(1675),
    [anon_sym_lsr] = ACTIONS(1675),
    [anon_sym_asr] = ACTIONS(1675),
    [aux_sym_pow_operator_token1] = ACTIONS(1677),
    [anon_sym_mod] = ACTIONS(1675),
    [anon_sym_land] = ACTIONS(1675),
    [anon_sym_lor] = ACTIONS(1675),
    [anon_sym_lxor] = ACTIONS(1675),
    [aux_sym_mult_operator_token1] = ACTIONS(1675),
    [aux_sym_add_operator_token1] = ACTIONS(1675),
    [sym_concat_operator] = ACTIONS(1677),
    [sym_rel_operator] = ACTIONS(1675),
    [anon_sym_AMP_AMP] = ACTIONS(1675),
    [anon_sym_or] = ACTIONS(1675),
    [anon_sym_PIPE_PIPE] = ACTIONS(1675),
    [sym_assign_operator] = ACTIONS(1677),
    [sym_let_operator] = ACTIONS(1677),
    [sym__uppercase_identifier] = ACTIONS(1677),
    [aux_sym_directive_token1] = ACTIONS(1675),
    [aux_sym_tag_token1] = ACTIONS(1677),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1050)] = {
    [sym_attribute] = STATE(1050),
    [sym__lowercase_identifier] = ACTIONS(1837),
    [anon_sym_SEMI_SEMI] = ACTIONS(1839),
    [anon_sym_let] = ACTIONS(1837),
    [anon_sym_and] = ACTIONS(1837),
    [anon_sym_TILDE] = ACTIONS(1837),
    [anon_sym_QMARK] = ACTIONS(1837),
    [anon_sym_LPAREN] = ACTIONS(1839),
    [anon_sym_external] = ACTIONS(1837),
    [anon_sym_type] = ACTIONS(1837),
    [anon_sym_COMMA] = ACTIONS(1839),
    [anon_sym_PIPE] = ACTIONS(1837),
    [anon_sym_DOT] = ACTIONS(1839),
    [anon_sym_LBRACK] = ACTIONS(1837),
    [anon_sym_RBRACK] = ACTIONS(1839),
    [anon_sym_true] = ACTIONS(1837),
    [anon_sym_false] = ACTIONS(1837),
    [anon_sym_LBRACE] = ACTIONS(1837),
    [anon_sym_SEMI] = ACTIONS(1837),
    [anon_sym_exception] = ACTIONS(1837),
    [anon_sym_module] = ACTIONS(1837),
    [anon_sym_open] = ACTIONS(1837),
    [anon_sym_include] = ACTIONS(1837),
    [anon_sym_class] = ACTIONS(1837),
    [anon_sym_val] = ACTIONS(1837),
    [anon_sym_end] = ACTIONS(1837),
    [anon_sym_object] = ACTIONS(1837),
    [anon_sym_AMP] = ACTIONS(1837),
    [anon_sym_POUND] = ACTIONS(1837),
    [anon_sym_COLON_COLON] = ACTIONS(1839),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1839),
    [anon_sym_else] = ACTIONS(1837),
    [anon_sym_new] = ACTIONS(1837),
    [anon_sym_LBRACE_LT] = ACTIONS(1839),
    [anon_sym_begin] = ACTIONS(1837),
    [sym_ocamlyacc_value] = ACTIONS(1839),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1837),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1839),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1837),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1837),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1839),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1839),
    [aux_sym_number_token1] = ACTIONS(1839),
    [aux_sym_signed_number_token1] = ACTIONS(1837),
    [anon_sym_SQUOTE] = ACTIONS(1839),
    [anon_sym_DQUOTE] = ACTIONS(1839),
    [sym_prefix_operator] = ACTIONS(1839),
    [aux_sym_sign_operator_token1] = ACTIONS(1837),
    [sym_hash_operator] = ACTIONS(1839),
    [anon_sym_lsl] = ACTIONS(1837),
    [anon_sym_lsr] = ACTIONS(1837),
    [anon_sym_asr] = ACTIONS(1837),
    [aux_sym_pow_operator_token1] = ACTIONS(1839),
    [anon_sym_mod] = ACTIONS(1837),
    [anon_sym_land] = ACTIONS(1837),
    [anon_sym_lor] = ACTIONS(1837),
    [anon_sym_lxor] = ACTIONS(1837),
    [aux_sym_mult_operator_token1] = ACTIONS(1837),
    [aux_sym_add_operator_token1] = ACTIONS(1837),
    [sym_concat_operator] = ACTIONS(1839),
    [sym_rel_operator] = ACTIONS(1837),
    [anon_sym_AMP_AMP] = ACTIONS(1837),
    [anon_sym_or] = ACTIONS(1837),
    [anon_sym_PIPE_PIPE] = ACTIONS(1837),
    [sym_assign_operator] = ACTIONS(1839),
    [sym_let_operator] = ACTIONS(1839),
    [sym__uppercase_identifier] = ACTIONS(1839),
    [aux_sym_directive_token1] = ACTIONS(1837),
    [aux_sym_tag_token1] = ACTIONS(1839),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1051)] = {
    [sym_attribute] = STATE(1051),
    [sym__lowercase_identifier] = ACTIONS(1843),
    [anon_sym_SEMI_SEMI] = ACTIONS(1845),
    [anon_sym_let] = ACTIONS(1843),
    [anon_sym_and] = ACTIONS(1843),
    [anon_sym_TILDE] = ACTIONS(1843),
    [anon_sym_QMARK] = ACTIONS(1843),
    [anon_sym_LPAREN] = ACTIONS(1845),
    [anon_sym_external] = ACTIONS(1843),
    [anon_sym_type] = ACTIONS(1843),
    [anon_sym_COMMA] = ACTIONS(1845),
    [anon_sym_PIPE] = ACTIONS(1843),
    [anon_sym_DOT] = ACTIONS(1845),
    [anon_sym_LBRACK] = ACTIONS(1843),
    [anon_sym_RBRACK] = ACTIONS(1845),
    [anon_sym_true] = ACTIONS(1843),
    [anon_sym_false] = ACTIONS(1843),
    [anon_sym_LBRACE] = ACTIONS(1843),
    [anon_sym_SEMI] = ACTIONS(1843),
    [anon_sym_exception] = ACTIONS(1843),
    [anon_sym_module] = ACTIONS(1843),
    [anon_sym_open] = ACTIONS(1843),
    [anon_sym_include] = ACTIONS(1843),
    [anon_sym_class] = ACTIONS(1843),
    [anon_sym_val] = ACTIONS(1843),
    [anon_sym_end] = ACTIONS(1843),
    [anon_sym_object] = ACTIONS(1843),
    [anon_sym_AMP] = ACTIONS(1843),
    [anon_sym_POUND] = ACTIONS(1843),
    [anon_sym_COLON_COLON] = ACTIONS(1845),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1845),
    [anon_sym_else] = ACTIONS(1843),
    [anon_sym_new] = ACTIONS(1843),
    [anon_sym_LBRACE_LT] = ACTIONS(1845),
    [anon_sym_begin] = ACTIONS(1843),
    [sym_ocamlyacc_value] = ACTIONS(1845),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1843),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1845),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1843),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1843),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1845),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1845),
    [aux_sym_number_token1] = ACTIONS(1845),
    [aux_sym_signed_number_token1] = ACTIONS(1843),
    [anon_sym_SQUOTE] = ACTIONS(1845),
    [anon_sym_DQUOTE] = ACTIONS(1845),
    [sym_prefix_operator] = ACTIONS(1845),
    [aux_sym_sign_operator_token1] = ACTIONS(1843),
    [sym_hash_operator] = ACTIONS(1845),
    [anon_sym_lsl] = ACTIONS(1843),
    [anon_sym_lsr] = ACTIONS(1843),
    [anon_sym_asr] = ACTIONS(1843),
    [aux_sym_pow_operator_token1] = ACTIONS(1845),
    [anon_sym_mod] = ACTIONS(1843),
    [anon_sym_land] = ACTIONS(1843),
    [anon_sym_lor] = ACTIONS(1843),
    [anon_sym_lxor] = ACTIONS(1843),
    [aux_sym_mult_operator_token1] = ACTIONS(1843),
    [aux_sym_add_operator_token1] = ACTIONS(1843),
    [sym_concat_operator] = ACTIONS(1845),
    [sym_rel_operator] = ACTIONS(1843),
    [anon_sym_AMP_AMP] = ACTIONS(1843),
    [anon_sym_or] = ACTIONS(1843),
    [anon_sym_PIPE_PIPE] = ACTIONS(1843),
    [sym_assign_operator] = ACTIONS(1845),
    [sym_let_operator] = ACTIONS(1845),
    [sym__uppercase_identifier] = ACTIONS(1845),
    [aux_sym_directive_token1] = ACTIONS(1843),
    [aux_sym_tag_token1] = ACTIONS(1845),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1052)] = {
    [sym_attribute] = STATE(1052),
    [sym__lowercase_identifier] = ACTIONS(1705),
    [anon_sym_SEMI_SEMI] = ACTIONS(1707),
    [anon_sym_let] = ACTIONS(1705),
    [anon_sym_and] = ACTIONS(1705),
    [anon_sym_TILDE] = ACTIONS(1705),
    [anon_sym_QMARK] = ACTIONS(1705),
    [anon_sym_LPAREN] = ACTIONS(1707),
    [anon_sym_external] = ACTIONS(1705),
    [anon_sym_type] = ACTIONS(1705),
    [anon_sym_COMMA] = ACTIONS(1707),
    [anon_sym_PIPE] = ACTIONS(1705),
    [anon_sym_DOT] = ACTIONS(1707),
    [anon_sym_LBRACK] = ACTIONS(1705),
    [anon_sym_RBRACK] = ACTIONS(1707),
    [anon_sym_true] = ACTIONS(1705),
    [anon_sym_false] = ACTIONS(1705),
    [anon_sym_LBRACE] = ACTIONS(1705),
    [anon_sym_SEMI] = ACTIONS(1705),
    [anon_sym_exception] = ACTIONS(1705),
    [anon_sym_module] = ACTIONS(1705),
    [anon_sym_open] = ACTIONS(1705),
    [anon_sym_include] = ACTIONS(1705),
    [anon_sym_class] = ACTIONS(1705),
    [anon_sym_val] = ACTIONS(1705),
    [anon_sym_end] = ACTIONS(1705),
    [anon_sym_object] = ACTIONS(1705),
    [anon_sym_AMP] = ACTIONS(1705),
    [anon_sym_POUND] = ACTIONS(1705),
    [anon_sym_COLON_COLON] = ACTIONS(1707),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1707),
    [anon_sym_else] = ACTIONS(1705),
    [anon_sym_new] = ACTIONS(1705),
    [anon_sym_LBRACE_LT] = ACTIONS(1707),
    [anon_sym_begin] = ACTIONS(1705),
    [sym_ocamlyacc_value] = ACTIONS(1707),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1705),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1707),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1705),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1705),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1707),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1707),
    [aux_sym_number_token1] = ACTIONS(1707),
    [aux_sym_signed_number_token1] = ACTIONS(1705),
    [anon_sym_SQUOTE] = ACTIONS(1707),
    [anon_sym_DQUOTE] = ACTIONS(1707),
    [sym_prefix_operator] = ACTIONS(1707),
    [aux_sym_sign_operator_token1] = ACTIONS(1705),
    [sym_hash_operator] = ACTIONS(1707),
    [anon_sym_lsl] = ACTIONS(1705),
    [anon_sym_lsr] = ACTIONS(1705),
    [anon_sym_asr] = ACTIONS(1705),
    [aux_sym_pow_operator_token1] = ACTIONS(1707),
    [anon_sym_mod] = ACTIONS(1705),
    [anon_sym_land] = ACTIONS(1705),
    [anon_sym_lor] = ACTIONS(1705),
    [anon_sym_lxor] = ACTIONS(1705),
    [aux_sym_mult_operator_token1] = ACTIONS(1705),
    [aux_sym_add_operator_token1] = ACTIONS(1705),
    [sym_concat_operator] = ACTIONS(1707),
    [sym_rel_operator] = ACTIONS(1705),
    [anon_sym_AMP_AMP] = ACTIONS(1705),
    [anon_sym_or] = ACTIONS(1705),
    [anon_sym_PIPE_PIPE] = ACTIONS(1705),
    [sym_assign_operator] = ACTIONS(1707),
    [sym_let_operator] = ACTIONS(1707),
    [sym__uppercase_identifier] = ACTIONS(1707),
    [aux_sym_directive_token1] = ACTIONS(1705),
    [aux_sym_tag_token1] = ACTIONS(1707),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1053)] = {
    [sym_attribute] = STATE(1053),
    [sym__lowercase_identifier] = ACTIONS(1571),
    [anon_sym_SEMI_SEMI] = ACTIONS(1573),
    [anon_sym_let] = ACTIONS(1571),
    [anon_sym_and] = ACTIONS(1571),
    [anon_sym_TILDE] = ACTIONS(1571),
    [anon_sym_QMARK] = ACTIONS(1571),
    [anon_sym_LPAREN] = ACTIONS(1573),
    [anon_sym_external] = ACTIONS(1571),
    [anon_sym_type] = ACTIONS(1571),
    [anon_sym_COMMA] = ACTIONS(1573),
    [anon_sym_PIPE] = ACTIONS(1571),
    [anon_sym_DOT] = ACTIONS(1573),
    [anon_sym_LBRACK] = ACTIONS(1571),
    [anon_sym_RBRACK] = ACTIONS(1573),
    [anon_sym_true] = ACTIONS(1571),
    [anon_sym_false] = ACTIONS(1571),
    [anon_sym_LBRACE] = ACTIONS(1571),
    [anon_sym_SEMI] = ACTIONS(1571),
    [anon_sym_exception] = ACTIONS(1571),
    [anon_sym_module] = ACTIONS(1571),
    [anon_sym_open] = ACTIONS(1571),
    [anon_sym_include] = ACTIONS(1571),
    [anon_sym_class] = ACTIONS(1571),
    [anon_sym_val] = ACTIONS(1571),
    [anon_sym_end] = ACTIONS(1571),
    [anon_sym_object] = ACTIONS(1571),
    [anon_sym_AMP] = ACTIONS(1571),
    [anon_sym_POUND] = ACTIONS(1571),
    [anon_sym_COLON_COLON] = ACTIONS(1573),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1573),
    [anon_sym_else] = ACTIONS(1571),
    [anon_sym_new] = ACTIONS(1571),
    [anon_sym_LBRACE_LT] = ACTIONS(1573),
    [anon_sym_begin] = ACTIONS(1571),
    [sym_ocamlyacc_value] = ACTIONS(1573),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1571),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1573),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1571),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1571),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1573),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1573),
    [aux_sym_number_token1] = ACTIONS(1573),
    [aux_sym_signed_number_token1] = ACTIONS(1571),
    [anon_sym_SQUOTE] = ACTIONS(1573),
    [anon_sym_DQUOTE] = ACTIONS(1573),
    [sym_prefix_operator] = ACTIONS(1573),
    [aux_sym_sign_operator_token1] = ACTIONS(1571),
    [sym_hash_operator] = ACTIONS(1573),
    [anon_sym_lsl] = ACTIONS(1571),
    [anon_sym_lsr] = ACTIONS(1571),
    [anon_sym_asr] = ACTIONS(1571),
    [aux_sym_pow_operator_token1] = ACTIONS(1573),
    [anon_sym_mod] = ACTIONS(1571),
    [anon_sym_land] = ACTIONS(1571),
    [anon_sym_lor] = ACTIONS(1571),
    [anon_sym_lxor] = ACTIONS(1571),
    [aux_sym_mult_operator_token1] = ACTIONS(1571),
    [aux_sym_add_operator_token1] = ACTIONS(1571),
    [sym_concat_operator] = ACTIONS(1573),
    [sym_rel_operator] = ACTIONS(1571),
    [anon_sym_AMP_AMP] = ACTIONS(1571),
    [anon_sym_or] = ACTIONS(1571),
    [anon_sym_PIPE_PIPE] = ACTIONS(1571),
    [sym_assign_operator] = ACTIONS(1573),
    [sym_let_operator] = ACTIONS(1573),
    [sym__uppercase_identifier] = ACTIONS(1573),
    [aux_sym_directive_token1] = ACTIONS(1571),
    [aux_sym_tag_token1] = ACTIONS(1573),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1054)] = {
    [sym_attribute] = STATE(1054),
    [sym__lowercase_identifier] = ACTIONS(1575),
    [anon_sym_SEMI_SEMI] = ACTIONS(1577),
    [anon_sym_let] = ACTIONS(1575),
    [anon_sym_and] = ACTIONS(1575),
    [anon_sym_TILDE] = ACTIONS(1575),
    [anon_sym_QMARK] = ACTIONS(1575),
    [anon_sym_LPAREN] = ACTIONS(1577),
    [anon_sym_external] = ACTIONS(1575),
    [anon_sym_type] = ACTIONS(1575),
    [anon_sym_COMMA] = ACTIONS(1577),
    [anon_sym_PIPE] = ACTIONS(1575),
    [anon_sym_DOT] = ACTIONS(1577),
    [anon_sym_LBRACK] = ACTIONS(1575),
    [anon_sym_RBRACK] = ACTIONS(1577),
    [anon_sym_true] = ACTIONS(1575),
    [anon_sym_false] = ACTIONS(1575),
    [anon_sym_LBRACE] = ACTIONS(1575),
    [anon_sym_SEMI] = ACTIONS(1575),
    [anon_sym_exception] = ACTIONS(1575),
    [anon_sym_module] = ACTIONS(1575),
    [anon_sym_open] = ACTIONS(1575),
    [anon_sym_include] = ACTIONS(1575),
    [anon_sym_class] = ACTIONS(1575),
    [anon_sym_val] = ACTIONS(1575),
    [anon_sym_end] = ACTIONS(1575),
    [anon_sym_object] = ACTIONS(1575),
    [anon_sym_AMP] = ACTIONS(1575),
    [anon_sym_POUND] = ACTIONS(1575),
    [anon_sym_COLON_COLON] = ACTIONS(1577),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1577),
    [anon_sym_else] = ACTIONS(1575),
    [anon_sym_new] = ACTIONS(1575),
    [anon_sym_LBRACE_LT] = ACTIONS(1577),
    [anon_sym_begin] = ACTIONS(1575),
    [sym_ocamlyacc_value] = ACTIONS(1577),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1575),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1577),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1575),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1575),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1577),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1577),
    [aux_sym_number_token1] = ACTIONS(1577),
    [aux_sym_signed_number_token1] = ACTIONS(1575),
    [anon_sym_SQUOTE] = ACTIONS(1577),
    [anon_sym_DQUOTE] = ACTIONS(1577),
    [sym_prefix_operator] = ACTIONS(1577),
    [aux_sym_sign_operator_token1] = ACTIONS(1575),
    [sym_hash_operator] = ACTIONS(1577),
    [anon_sym_lsl] = ACTIONS(1575),
    [anon_sym_lsr] = ACTIONS(1575),
    [anon_sym_asr] = ACTIONS(1575),
    [aux_sym_pow_operator_token1] = ACTIONS(1577),
    [anon_sym_mod] = ACTIONS(1575),
    [anon_sym_land] = ACTIONS(1575),
    [anon_sym_lor] = ACTIONS(1575),
    [anon_sym_lxor] = ACTIONS(1575),
    [aux_sym_mult_operator_token1] = ACTIONS(1575),
    [aux_sym_add_operator_token1] = ACTIONS(1575),
    [sym_concat_operator] = ACTIONS(1577),
    [sym_rel_operator] = ACTIONS(1575),
    [anon_sym_AMP_AMP] = ACTIONS(1575),
    [anon_sym_or] = ACTIONS(1575),
    [anon_sym_PIPE_PIPE] = ACTIONS(1575),
    [sym_assign_operator] = ACTIONS(1577),
    [sym_let_operator] = ACTIONS(1577),
    [sym__uppercase_identifier] = ACTIONS(1577),
    [aux_sym_directive_token1] = ACTIONS(1575),
    [aux_sym_tag_token1] = ACTIONS(1577),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1055)] = {
    [sym_attribute] = STATE(1055),
    [sym__lowercase_identifier] = ACTIONS(1595),
    [anon_sym_SEMI_SEMI] = ACTIONS(1597),
    [anon_sym_let] = ACTIONS(1595),
    [anon_sym_and] = ACTIONS(1595),
    [anon_sym_TILDE] = ACTIONS(1595),
    [anon_sym_QMARK] = ACTIONS(1595),
    [anon_sym_LPAREN] = ACTIONS(1597),
    [anon_sym_external] = ACTIONS(1595),
    [anon_sym_type] = ACTIONS(1595),
    [anon_sym_COMMA] = ACTIONS(1597),
    [anon_sym_PIPE] = ACTIONS(1595),
    [anon_sym_DOT] = ACTIONS(1597),
    [anon_sym_LBRACK] = ACTIONS(1595),
    [anon_sym_RBRACK] = ACTIONS(1597),
    [anon_sym_true] = ACTIONS(1595),
    [anon_sym_false] = ACTIONS(1595),
    [anon_sym_LBRACE] = ACTIONS(1595),
    [anon_sym_SEMI] = ACTIONS(1595),
    [anon_sym_exception] = ACTIONS(1595),
    [anon_sym_module] = ACTIONS(1595),
    [anon_sym_open] = ACTIONS(1595),
    [anon_sym_include] = ACTIONS(1595),
    [anon_sym_class] = ACTIONS(1595),
    [anon_sym_val] = ACTIONS(1595),
    [anon_sym_end] = ACTIONS(1595),
    [anon_sym_object] = ACTIONS(1595),
    [anon_sym_AMP] = ACTIONS(1595),
    [anon_sym_POUND] = ACTIONS(1595),
    [anon_sym_COLON_COLON] = ACTIONS(1597),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1597),
    [anon_sym_else] = ACTIONS(1595),
    [anon_sym_new] = ACTIONS(1595),
    [anon_sym_LBRACE_LT] = ACTIONS(1597),
    [anon_sym_begin] = ACTIONS(1595),
    [sym_ocamlyacc_value] = ACTIONS(1597),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1595),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1597),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1595),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1595),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1597),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1597),
    [aux_sym_number_token1] = ACTIONS(1597),
    [aux_sym_signed_number_token1] = ACTIONS(1595),
    [anon_sym_SQUOTE] = ACTIONS(1597),
    [anon_sym_DQUOTE] = ACTIONS(1597),
    [sym_prefix_operator] = ACTIONS(1597),
    [aux_sym_sign_operator_token1] = ACTIONS(1595),
    [sym_hash_operator] = ACTIONS(1597),
    [anon_sym_lsl] = ACTIONS(1595),
    [anon_sym_lsr] = ACTIONS(1595),
    [anon_sym_asr] = ACTIONS(1595),
    [aux_sym_pow_operator_token1] = ACTIONS(1597),
    [anon_sym_mod] = ACTIONS(1595),
    [anon_sym_land] = ACTIONS(1595),
    [anon_sym_lor] = ACTIONS(1595),
    [anon_sym_lxor] = ACTIONS(1595),
    [aux_sym_mult_operator_token1] = ACTIONS(1595),
    [aux_sym_add_operator_token1] = ACTIONS(1595),
    [sym_concat_operator] = ACTIONS(1597),
    [sym_rel_operator] = ACTIONS(1595),
    [anon_sym_AMP_AMP] = ACTIONS(1595),
    [anon_sym_or] = ACTIONS(1595),
    [anon_sym_PIPE_PIPE] = ACTIONS(1595),
    [sym_assign_operator] = ACTIONS(1597),
    [sym_let_operator] = ACTIONS(1597),
    [sym__uppercase_identifier] = ACTIONS(1597),
    [aux_sym_directive_token1] = ACTIONS(1595),
    [aux_sym_tag_token1] = ACTIONS(1597),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1056)] = {
    [sym_attribute] = STATE(1056),
    [sym__lowercase_identifier] = ACTIONS(1705),
    [anon_sym_SEMI_SEMI] = ACTIONS(1707),
    [anon_sym_let] = ACTIONS(1705),
    [anon_sym_TILDE] = ACTIONS(1705),
    [anon_sym_QMARK] = ACTIONS(1705),
    [anon_sym_LPAREN] = ACTIONS(1707),
    [anon_sym_external] = ACTIONS(1705),
    [anon_sym_type] = ACTIONS(1705),
    [anon_sym_COMMA] = ACTIONS(1707),
    [anon_sym_PIPE] = ACTIONS(1705),
    [anon_sym_DOT] = ACTIONS(1707),
    [anon_sym_LBRACK] = ACTIONS(1705),
    [anon_sym_RBRACK] = ACTIONS(1707),
    [anon_sym_true] = ACTIONS(1705),
    [anon_sym_false] = ACTIONS(1705),
    [anon_sym_LBRACE] = ACTIONS(1705),
    [anon_sym_SEMI] = ACTIONS(1705),
    [anon_sym_exception] = ACTIONS(1705),
    [anon_sym_module] = ACTIONS(1705),
    [anon_sym_open] = ACTIONS(1705),
    [anon_sym_include] = ACTIONS(1705),
    [anon_sym_class] = ACTIONS(1705),
    [anon_sym_val] = ACTIONS(1705),
    [anon_sym_end] = ACTIONS(1705),
    [anon_sym_object] = ACTIONS(1705),
    [anon_sym_AMP] = ACTIONS(1705),
    [anon_sym_POUND] = ACTIONS(1705),
    [anon_sym_COLON_COLON] = ACTIONS(1707),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1707),
    [anon_sym_LT_DASH] = ACTIONS(2204),
    [anon_sym_else] = ACTIONS(1705),
    [anon_sym_new] = ACTIONS(1705),
    [anon_sym_LBRACE_LT] = ACTIONS(1707),
    [anon_sym_begin] = ACTIONS(1705),
    [sym_ocamlyacc_value] = ACTIONS(1707),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1705),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1707),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1705),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1705),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1707),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1707),
    [aux_sym_number_token1] = ACTIONS(1707),
    [aux_sym_signed_number_token1] = ACTIONS(1705),
    [anon_sym_SQUOTE] = ACTIONS(1707),
    [anon_sym_DQUOTE] = ACTIONS(1707),
    [sym_prefix_operator] = ACTIONS(1707),
    [aux_sym_sign_operator_token1] = ACTIONS(1705),
    [sym_hash_operator] = ACTIONS(1707),
    [anon_sym_lsl] = ACTIONS(1705),
    [anon_sym_lsr] = ACTIONS(1705),
    [anon_sym_asr] = ACTIONS(1705),
    [aux_sym_pow_operator_token1] = ACTIONS(1707),
    [anon_sym_mod] = ACTIONS(1705),
    [anon_sym_land] = ACTIONS(1705),
    [anon_sym_lor] = ACTIONS(1705),
    [anon_sym_lxor] = ACTIONS(1705),
    [aux_sym_mult_operator_token1] = ACTIONS(1705),
    [aux_sym_add_operator_token1] = ACTIONS(1705),
    [sym_concat_operator] = ACTIONS(1707),
    [sym_rel_operator] = ACTIONS(1705),
    [anon_sym_AMP_AMP] = ACTIONS(1705),
    [anon_sym_or] = ACTIONS(1705),
    [anon_sym_PIPE_PIPE] = ACTIONS(1705),
    [sym_assign_operator] = ACTIONS(1707),
    [sym_let_operator] = ACTIONS(1707),
    [sym__uppercase_identifier] = ACTIONS(1707),
    [aux_sym_directive_token1] = ACTIONS(1705),
    [aux_sym_tag_token1] = ACTIONS(1707),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1057)] = {
    [sym_attribute] = STATE(1057),
    [sym__lowercase_identifier] = ACTIONS(1709),
    [anon_sym_SEMI_SEMI] = ACTIONS(1711),
    [anon_sym_let] = ACTIONS(1709),
    [anon_sym_and] = ACTIONS(1709),
    [anon_sym_TILDE] = ACTIONS(1709),
    [anon_sym_QMARK] = ACTIONS(1709),
    [anon_sym_LPAREN] = ACTIONS(1711),
    [anon_sym_external] = ACTIONS(1709),
    [anon_sym_type] = ACTIONS(1709),
    [anon_sym_COMMA] = ACTIONS(1711),
    [anon_sym_PIPE] = ACTIONS(1709),
    [anon_sym_DOT] = ACTIONS(1711),
    [anon_sym_LBRACK] = ACTIONS(1709),
    [anon_sym_RBRACK] = ACTIONS(1711),
    [anon_sym_true] = ACTIONS(1709),
    [anon_sym_false] = ACTIONS(1709),
    [anon_sym_LBRACE] = ACTIONS(1709),
    [anon_sym_SEMI] = ACTIONS(1709),
    [anon_sym_exception] = ACTIONS(1709),
    [anon_sym_module] = ACTIONS(1709),
    [anon_sym_open] = ACTIONS(1709),
    [anon_sym_include] = ACTIONS(1709),
    [anon_sym_class] = ACTIONS(1709),
    [anon_sym_val] = ACTIONS(1709),
    [anon_sym_end] = ACTIONS(1709),
    [anon_sym_object] = ACTIONS(1709),
    [anon_sym_AMP] = ACTIONS(1709),
    [anon_sym_POUND] = ACTIONS(1709),
    [anon_sym_COLON_COLON] = ACTIONS(1711),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1711),
    [anon_sym_else] = ACTIONS(1709),
    [anon_sym_new] = ACTIONS(1709),
    [anon_sym_LBRACE_LT] = ACTIONS(1711),
    [anon_sym_begin] = ACTIONS(1709),
    [sym_ocamlyacc_value] = ACTIONS(1711),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1709),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1711),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1709),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1709),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1711),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1711),
    [aux_sym_number_token1] = ACTIONS(1711),
    [aux_sym_signed_number_token1] = ACTIONS(1709),
    [anon_sym_SQUOTE] = ACTIONS(1711),
    [anon_sym_DQUOTE] = ACTIONS(1711),
    [sym_prefix_operator] = ACTIONS(1711),
    [aux_sym_sign_operator_token1] = ACTIONS(1709),
    [sym_hash_operator] = ACTIONS(1711),
    [anon_sym_lsl] = ACTIONS(1709),
    [anon_sym_lsr] = ACTIONS(1709),
    [anon_sym_asr] = ACTIONS(1709),
    [aux_sym_pow_operator_token1] = ACTIONS(1711),
    [anon_sym_mod] = ACTIONS(1709),
    [anon_sym_land] = ACTIONS(1709),
    [anon_sym_lor] = ACTIONS(1709),
    [anon_sym_lxor] = ACTIONS(1709),
    [aux_sym_mult_operator_token1] = ACTIONS(1709),
    [aux_sym_add_operator_token1] = ACTIONS(1709),
    [sym_concat_operator] = ACTIONS(1711),
    [sym_rel_operator] = ACTIONS(1709),
    [anon_sym_AMP_AMP] = ACTIONS(1709),
    [anon_sym_or] = ACTIONS(1709),
    [anon_sym_PIPE_PIPE] = ACTIONS(1709),
    [sym_assign_operator] = ACTIONS(1711),
    [sym_let_operator] = ACTIONS(1711),
    [sym__uppercase_identifier] = ACTIONS(1711),
    [aux_sym_directive_token1] = ACTIONS(1709),
    [aux_sym_tag_token1] = ACTIONS(1711),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1058)] = {
    [sym_attribute] = STATE(1058),
    [sym__lowercase_identifier] = ACTIONS(1639),
    [anon_sym_SEMI_SEMI] = ACTIONS(1641),
    [anon_sym_let] = ACTIONS(1639),
    [anon_sym_and] = ACTIONS(1639),
    [anon_sym_TILDE] = ACTIONS(1639),
    [anon_sym_QMARK] = ACTIONS(1639),
    [anon_sym_LPAREN] = ACTIONS(1641),
    [anon_sym_external] = ACTIONS(1639),
    [anon_sym_type] = ACTIONS(1639),
    [anon_sym_COMMA] = ACTIONS(1641),
    [anon_sym_PIPE] = ACTIONS(1639),
    [anon_sym_DOT] = ACTIONS(1641),
    [anon_sym_LBRACK] = ACTIONS(1639),
    [anon_sym_RBRACK] = ACTIONS(1641),
    [anon_sym_true] = ACTIONS(1639),
    [anon_sym_false] = ACTIONS(1639),
    [anon_sym_LBRACE] = ACTIONS(1639),
    [anon_sym_SEMI] = ACTIONS(1639),
    [anon_sym_exception] = ACTIONS(1639),
    [anon_sym_module] = ACTIONS(1639),
    [anon_sym_open] = ACTIONS(1639),
    [anon_sym_include] = ACTIONS(1639),
    [anon_sym_class] = ACTIONS(1639),
    [anon_sym_val] = ACTIONS(1639),
    [anon_sym_end] = ACTIONS(1639),
    [anon_sym_object] = ACTIONS(1639),
    [anon_sym_AMP] = ACTIONS(1639),
    [anon_sym_POUND] = ACTIONS(1639),
    [anon_sym_COLON_COLON] = ACTIONS(1641),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1641),
    [anon_sym_else] = ACTIONS(1639),
    [anon_sym_new] = ACTIONS(1639),
    [anon_sym_LBRACE_LT] = ACTIONS(1641),
    [anon_sym_begin] = ACTIONS(1639),
    [sym_ocamlyacc_value] = ACTIONS(1641),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1639),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1641),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1639),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1639),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1641),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1641),
    [aux_sym_number_token1] = ACTIONS(1641),
    [aux_sym_signed_number_token1] = ACTIONS(1639),
    [anon_sym_SQUOTE] = ACTIONS(1641),
    [anon_sym_DQUOTE] = ACTIONS(1641),
    [sym_prefix_operator] = ACTIONS(1641),
    [aux_sym_sign_operator_token1] = ACTIONS(1639),
    [sym_hash_operator] = ACTIONS(1641),
    [anon_sym_lsl] = ACTIONS(1639),
    [anon_sym_lsr] = ACTIONS(1639),
    [anon_sym_asr] = ACTIONS(1639),
    [aux_sym_pow_operator_token1] = ACTIONS(1641),
    [anon_sym_mod] = ACTIONS(1639),
    [anon_sym_land] = ACTIONS(1639),
    [anon_sym_lor] = ACTIONS(1639),
    [anon_sym_lxor] = ACTIONS(1639),
    [aux_sym_mult_operator_token1] = ACTIONS(1639),
    [aux_sym_add_operator_token1] = ACTIONS(1639),
    [sym_concat_operator] = ACTIONS(1641),
    [sym_rel_operator] = ACTIONS(1639),
    [anon_sym_AMP_AMP] = ACTIONS(1639),
    [anon_sym_or] = ACTIONS(1639),
    [anon_sym_PIPE_PIPE] = ACTIONS(1639),
    [sym_assign_operator] = ACTIONS(1641),
    [sym_let_operator] = ACTIONS(1641),
    [sym__uppercase_identifier] = ACTIONS(1641),
    [aux_sym_directive_token1] = ACTIONS(1639),
    [aux_sym_tag_token1] = ACTIONS(1641),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1059)] = {
    [sym_attribute] = STATE(1059),
    [sym__lowercase_identifier] = ACTIONS(1643),
    [anon_sym_SEMI_SEMI] = ACTIONS(1645),
    [anon_sym_let] = ACTIONS(1643),
    [anon_sym_and] = ACTIONS(1643),
    [anon_sym_TILDE] = ACTIONS(1643),
    [anon_sym_QMARK] = ACTIONS(1643),
    [anon_sym_LPAREN] = ACTIONS(1645),
    [anon_sym_external] = ACTIONS(1643),
    [anon_sym_type] = ACTIONS(1643),
    [anon_sym_COMMA] = ACTIONS(1645),
    [anon_sym_PIPE] = ACTIONS(1643),
    [anon_sym_DOT] = ACTIONS(1645),
    [anon_sym_LBRACK] = ACTIONS(1643),
    [anon_sym_RBRACK] = ACTIONS(1645),
    [anon_sym_true] = ACTIONS(1643),
    [anon_sym_false] = ACTIONS(1643),
    [anon_sym_LBRACE] = ACTIONS(1643),
    [anon_sym_SEMI] = ACTIONS(1643),
    [anon_sym_exception] = ACTIONS(1643),
    [anon_sym_module] = ACTIONS(1643),
    [anon_sym_open] = ACTIONS(1643),
    [anon_sym_include] = ACTIONS(1643),
    [anon_sym_class] = ACTIONS(1643),
    [anon_sym_val] = ACTIONS(1643),
    [anon_sym_end] = ACTIONS(1643),
    [anon_sym_object] = ACTIONS(1643),
    [anon_sym_AMP] = ACTIONS(1643),
    [anon_sym_POUND] = ACTIONS(1643),
    [anon_sym_COLON_COLON] = ACTIONS(1645),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1645),
    [anon_sym_else] = ACTIONS(1643),
    [anon_sym_new] = ACTIONS(1643),
    [anon_sym_LBRACE_LT] = ACTIONS(1645),
    [anon_sym_begin] = ACTIONS(1643),
    [sym_ocamlyacc_value] = ACTIONS(1645),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1643),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1645),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1643),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1643),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1645),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1645),
    [aux_sym_number_token1] = ACTIONS(1645),
    [aux_sym_signed_number_token1] = ACTIONS(1643),
    [anon_sym_SQUOTE] = ACTIONS(1645),
    [anon_sym_DQUOTE] = ACTIONS(1645),
    [sym_prefix_operator] = ACTIONS(1645),
    [aux_sym_sign_operator_token1] = ACTIONS(1643),
    [sym_hash_operator] = ACTIONS(1645),
    [anon_sym_lsl] = ACTIONS(1643),
    [anon_sym_lsr] = ACTIONS(1643),
    [anon_sym_asr] = ACTIONS(1643),
    [aux_sym_pow_operator_token1] = ACTIONS(1645),
    [anon_sym_mod] = ACTIONS(1643),
    [anon_sym_land] = ACTIONS(1643),
    [anon_sym_lor] = ACTIONS(1643),
    [anon_sym_lxor] = ACTIONS(1643),
    [aux_sym_mult_operator_token1] = ACTIONS(1643),
    [aux_sym_add_operator_token1] = ACTIONS(1643),
    [sym_concat_operator] = ACTIONS(1645),
    [sym_rel_operator] = ACTIONS(1643),
    [anon_sym_AMP_AMP] = ACTIONS(1643),
    [anon_sym_or] = ACTIONS(1643),
    [anon_sym_PIPE_PIPE] = ACTIONS(1643),
    [sym_assign_operator] = ACTIONS(1645),
    [sym_let_operator] = ACTIONS(1645),
    [sym__uppercase_identifier] = ACTIONS(1645),
    [aux_sym_directive_token1] = ACTIONS(1643),
    [aux_sym_tag_token1] = ACTIONS(1645),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1060)] = {
    [sym_attribute] = STATE(1060),
    [sym__lowercase_identifier] = ACTIONS(1647),
    [anon_sym_SEMI_SEMI] = ACTIONS(1649),
    [anon_sym_let] = ACTIONS(1647),
    [anon_sym_and] = ACTIONS(1647),
    [anon_sym_TILDE] = ACTIONS(1647),
    [anon_sym_QMARK] = ACTIONS(1647),
    [anon_sym_LPAREN] = ACTIONS(1649),
    [anon_sym_external] = ACTIONS(1647),
    [anon_sym_type] = ACTIONS(1647),
    [anon_sym_COMMA] = ACTIONS(1649),
    [anon_sym_PIPE] = ACTIONS(1647),
    [anon_sym_DOT] = ACTIONS(1649),
    [anon_sym_LBRACK] = ACTIONS(1647),
    [anon_sym_RBRACK] = ACTIONS(1649),
    [anon_sym_true] = ACTIONS(1647),
    [anon_sym_false] = ACTIONS(1647),
    [anon_sym_LBRACE] = ACTIONS(1647),
    [anon_sym_SEMI] = ACTIONS(1647),
    [anon_sym_exception] = ACTIONS(1647),
    [anon_sym_module] = ACTIONS(1647),
    [anon_sym_open] = ACTIONS(1647),
    [anon_sym_include] = ACTIONS(1647),
    [anon_sym_class] = ACTIONS(1647),
    [anon_sym_val] = ACTIONS(1647),
    [anon_sym_end] = ACTIONS(1647),
    [anon_sym_object] = ACTIONS(1647),
    [anon_sym_AMP] = ACTIONS(1647),
    [anon_sym_POUND] = ACTIONS(1647),
    [anon_sym_COLON_COLON] = ACTIONS(1649),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1649),
    [anon_sym_else] = ACTIONS(1647),
    [anon_sym_new] = ACTIONS(1647),
    [anon_sym_LBRACE_LT] = ACTIONS(1649),
    [anon_sym_begin] = ACTIONS(1647),
    [sym_ocamlyacc_value] = ACTIONS(1649),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1647),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1649),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1647),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1647),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1649),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1649),
    [aux_sym_number_token1] = ACTIONS(1649),
    [aux_sym_signed_number_token1] = ACTIONS(1647),
    [anon_sym_SQUOTE] = ACTIONS(1649),
    [anon_sym_DQUOTE] = ACTIONS(1649),
    [sym_prefix_operator] = ACTIONS(1649),
    [aux_sym_sign_operator_token1] = ACTIONS(1647),
    [sym_hash_operator] = ACTIONS(1649),
    [anon_sym_lsl] = ACTIONS(1647),
    [anon_sym_lsr] = ACTIONS(1647),
    [anon_sym_asr] = ACTIONS(1647),
    [aux_sym_pow_operator_token1] = ACTIONS(1649),
    [anon_sym_mod] = ACTIONS(1647),
    [anon_sym_land] = ACTIONS(1647),
    [anon_sym_lor] = ACTIONS(1647),
    [anon_sym_lxor] = ACTIONS(1647),
    [aux_sym_mult_operator_token1] = ACTIONS(1647),
    [aux_sym_add_operator_token1] = ACTIONS(1647),
    [sym_concat_operator] = ACTIONS(1649),
    [sym_rel_operator] = ACTIONS(1647),
    [anon_sym_AMP_AMP] = ACTIONS(1647),
    [anon_sym_or] = ACTIONS(1647),
    [anon_sym_PIPE_PIPE] = ACTIONS(1647),
    [sym_assign_operator] = ACTIONS(1649),
    [sym_let_operator] = ACTIONS(1649),
    [sym__uppercase_identifier] = ACTIONS(1649),
    [aux_sym_directive_token1] = ACTIONS(1647),
    [aux_sym_tag_token1] = ACTIONS(1649),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1061)] = {
    [sym_attribute] = STATE(1061),
    [sym__lowercase_identifier] = ACTIONS(1651),
    [anon_sym_SEMI_SEMI] = ACTIONS(1653),
    [anon_sym_let] = ACTIONS(1651),
    [anon_sym_and] = ACTIONS(1651),
    [anon_sym_TILDE] = ACTIONS(1651),
    [anon_sym_QMARK] = ACTIONS(1651),
    [anon_sym_LPAREN] = ACTIONS(1653),
    [anon_sym_external] = ACTIONS(1651),
    [anon_sym_type] = ACTIONS(1651),
    [anon_sym_COMMA] = ACTIONS(1653),
    [anon_sym_PIPE] = ACTIONS(1651),
    [anon_sym_DOT] = ACTIONS(1653),
    [anon_sym_LBRACK] = ACTIONS(1651),
    [anon_sym_RBRACK] = ACTIONS(1653),
    [anon_sym_true] = ACTIONS(1651),
    [anon_sym_false] = ACTIONS(1651),
    [anon_sym_LBRACE] = ACTIONS(1651),
    [anon_sym_SEMI] = ACTIONS(1651),
    [anon_sym_exception] = ACTIONS(1651),
    [anon_sym_module] = ACTIONS(1651),
    [anon_sym_open] = ACTIONS(1651),
    [anon_sym_include] = ACTIONS(1651),
    [anon_sym_class] = ACTIONS(1651),
    [anon_sym_val] = ACTIONS(1651),
    [anon_sym_end] = ACTIONS(1651),
    [anon_sym_object] = ACTIONS(1651),
    [anon_sym_AMP] = ACTIONS(1651),
    [anon_sym_POUND] = ACTIONS(1651),
    [anon_sym_COLON_COLON] = ACTIONS(1653),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1653),
    [anon_sym_else] = ACTIONS(1651),
    [anon_sym_new] = ACTIONS(1651),
    [anon_sym_LBRACE_LT] = ACTIONS(1653),
    [anon_sym_begin] = ACTIONS(1651),
    [sym_ocamlyacc_value] = ACTIONS(1653),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1651),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1653),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1651),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1651),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1653),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1653),
    [aux_sym_number_token1] = ACTIONS(1653),
    [aux_sym_signed_number_token1] = ACTIONS(1651),
    [anon_sym_SQUOTE] = ACTIONS(1653),
    [anon_sym_DQUOTE] = ACTIONS(1653),
    [sym_prefix_operator] = ACTIONS(1653),
    [aux_sym_sign_operator_token1] = ACTIONS(1651),
    [sym_hash_operator] = ACTIONS(1653),
    [anon_sym_lsl] = ACTIONS(1651),
    [anon_sym_lsr] = ACTIONS(1651),
    [anon_sym_asr] = ACTIONS(1651),
    [aux_sym_pow_operator_token1] = ACTIONS(1653),
    [anon_sym_mod] = ACTIONS(1651),
    [anon_sym_land] = ACTIONS(1651),
    [anon_sym_lor] = ACTIONS(1651),
    [anon_sym_lxor] = ACTIONS(1651),
    [aux_sym_mult_operator_token1] = ACTIONS(1651),
    [aux_sym_add_operator_token1] = ACTIONS(1651),
    [sym_concat_operator] = ACTIONS(1653),
    [sym_rel_operator] = ACTIONS(1651),
    [anon_sym_AMP_AMP] = ACTIONS(1651),
    [anon_sym_or] = ACTIONS(1651),
    [anon_sym_PIPE_PIPE] = ACTIONS(1651),
    [sym_assign_operator] = ACTIONS(1653),
    [sym_let_operator] = ACTIONS(1653),
    [sym__uppercase_identifier] = ACTIONS(1653),
    [aux_sym_directive_token1] = ACTIONS(1651),
    [aux_sym_tag_token1] = ACTIONS(1653),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1062)] = {
    [sym_attribute] = STATE(1062),
    [sym__lowercase_identifier] = ACTIONS(1671),
    [anon_sym_SEMI_SEMI] = ACTIONS(1673),
    [anon_sym_let] = ACTIONS(1671),
    [anon_sym_and] = ACTIONS(1671),
    [anon_sym_TILDE] = ACTIONS(1671),
    [anon_sym_QMARK] = ACTIONS(1671),
    [anon_sym_LPAREN] = ACTIONS(1673),
    [anon_sym_external] = ACTIONS(1671),
    [anon_sym_type] = ACTIONS(1671),
    [anon_sym_COMMA] = ACTIONS(1673),
    [anon_sym_PIPE] = ACTIONS(1671),
    [anon_sym_DOT] = ACTIONS(1673),
    [anon_sym_LBRACK] = ACTIONS(1671),
    [anon_sym_RBRACK] = ACTIONS(1673),
    [anon_sym_true] = ACTIONS(1671),
    [anon_sym_false] = ACTIONS(1671),
    [anon_sym_LBRACE] = ACTIONS(1671),
    [anon_sym_SEMI] = ACTIONS(1671),
    [anon_sym_exception] = ACTIONS(1671),
    [anon_sym_module] = ACTIONS(1671),
    [anon_sym_open] = ACTIONS(1671),
    [anon_sym_include] = ACTIONS(1671),
    [anon_sym_class] = ACTIONS(1671),
    [anon_sym_val] = ACTIONS(1671),
    [anon_sym_end] = ACTIONS(1671),
    [anon_sym_object] = ACTIONS(1671),
    [anon_sym_AMP] = ACTIONS(1671),
    [anon_sym_POUND] = ACTIONS(1671),
    [anon_sym_COLON_COLON] = ACTIONS(1673),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1673),
    [anon_sym_else] = ACTIONS(1671),
    [anon_sym_new] = ACTIONS(1671),
    [anon_sym_LBRACE_LT] = ACTIONS(1673),
    [anon_sym_begin] = ACTIONS(1671),
    [sym_ocamlyacc_value] = ACTIONS(1673),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1671),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1673),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1671),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1671),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1673),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1673),
    [aux_sym_number_token1] = ACTIONS(1673),
    [aux_sym_signed_number_token1] = ACTIONS(1671),
    [anon_sym_SQUOTE] = ACTIONS(1673),
    [anon_sym_DQUOTE] = ACTIONS(1673),
    [sym_prefix_operator] = ACTIONS(1673),
    [aux_sym_sign_operator_token1] = ACTIONS(1671),
    [sym_hash_operator] = ACTIONS(1673),
    [anon_sym_lsl] = ACTIONS(1671),
    [anon_sym_lsr] = ACTIONS(1671),
    [anon_sym_asr] = ACTIONS(1671),
    [aux_sym_pow_operator_token1] = ACTIONS(1673),
    [anon_sym_mod] = ACTIONS(1671),
    [anon_sym_land] = ACTIONS(1671),
    [anon_sym_lor] = ACTIONS(1671),
    [anon_sym_lxor] = ACTIONS(1671),
    [aux_sym_mult_operator_token1] = ACTIONS(1671),
    [aux_sym_add_operator_token1] = ACTIONS(1671),
    [sym_concat_operator] = ACTIONS(1673),
    [sym_rel_operator] = ACTIONS(1671),
    [anon_sym_AMP_AMP] = ACTIONS(1671),
    [anon_sym_or] = ACTIONS(1671),
    [anon_sym_PIPE_PIPE] = ACTIONS(1671),
    [sym_assign_operator] = ACTIONS(1673),
    [sym_let_operator] = ACTIONS(1673),
    [sym__uppercase_identifier] = ACTIONS(1673),
    [aux_sym_directive_token1] = ACTIONS(1671),
    [aux_sym_tag_token1] = ACTIONS(1673),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1063)] = {
    [sym_attribute] = STATE(1063),
    [sym__lowercase_identifier] = ACTIONS(1655),
    [anon_sym_SEMI_SEMI] = ACTIONS(1657),
    [anon_sym_let] = ACTIONS(1655),
    [anon_sym_and] = ACTIONS(1655),
    [anon_sym_TILDE] = ACTIONS(1655),
    [anon_sym_QMARK] = ACTIONS(1655),
    [anon_sym_LPAREN] = ACTIONS(1657),
    [anon_sym_external] = ACTIONS(1655),
    [anon_sym_type] = ACTIONS(1655),
    [anon_sym_COMMA] = ACTIONS(1657),
    [anon_sym_PIPE] = ACTIONS(1655),
    [anon_sym_DOT] = ACTIONS(1657),
    [anon_sym_LBRACK] = ACTIONS(1655),
    [anon_sym_RBRACK] = ACTIONS(1657),
    [anon_sym_true] = ACTIONS(1655),
    [anon_sym_false] = ACTIONS(1655),
    [anon_sym_LBRACE] = ACTIONS(1655),
    [anon_sym_SEMI] = ACTIONS(1655),
    [anon_sym_exception] = ACTIONS(1655),
    [anon_sym_module] = ACTIONS(1655),
    [anon_sym_open] = ACTIONS(1655),
    [anon_sym_include] = ACTIONS(1655),
    [anon_sym_class] = ACTIONS(1655),
    [anon_sym_val] = ACTIONS(1655),
    [anon_sym_end] = ACTIONS(1655),
    [anon_sym_object] = ACTIONS(1655),
    [anon_sym_AMP] = ACTIONS(1655),
    [anon_sym_POUND] = ACTIONS(1655),
    [anon_sym_COLON_COLON] = ACTIONS(1657),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1657),
    [anon_sym_else] = ACTIONS(1655),
    [anon_sym_new] = ACTIONS(1655),
    [anon_sym_LBRACE_LT] = ACTIONS(1657),
    [anon_sym_begin] = ACTIONS(1655),
    [sym_ocamlyacc_value] = ACTIONS(1657),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1655),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1657),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1655),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1655),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1657),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1657),
    [aux_sym_number_token1] = ACTIONS(1657),
    [aux_sym_signed_number_token1] = ACTIONS(1655),
    [anon_sym_SQUOTE] = ACTIONS(1657),
    [anon_sym_DQUOTE] = ACTIONS(1657),
    [sym_prefix_operator] = ACTIONS(1657),
    [aux_sym_sign_operator_token1] = ACTIONS(1655),
    [sym_hash_operator] = ACTIONS(1657),
    [anon_sym_lsl] = ACTIONS(1655),
    [anon_sym_lsr] = ACTIONS(1655),
    [anon_sym_asr] = ACTIONS(1655),
    [aux_sym_pow_operator_token1] = ACTIONS(1657),
    [anon_sym_mod] = ACTIONS(1655),
    [anon_sym_land] = ACTIONS(1655),
    [anon_sym_lor] = ACTIONS(1655),
    [anon_sym_lxor] = ACTIONS(1655),
    [aux_sym_mult_operator_token1] = ACTIONS(1655),
    [aux_sym_add_operator_token1] = ACTIONS(1655),
    [sym_concat_operator] = ACTIONS(1657),
    [sym_rel_operator] = ACTIONS(1655),
    [anon_sym_AMP_AMP] = ACTIONS(1655),
    [anon_sym_or] = ACTIONS(1655),
    [anon_sym_PIPE_PIPE] = ACTIONS(1655),
    [sym_assign_operator] = ACTIONS(1657),
    [sym_let_operator] = ACTIONS(1657),
    [sym__uppercase_identifier] = ACTIONS(1657),
    [aux_sym_directive_token1] = ACTIONS(1655),
    [aux_sym_tag_token1] = ACTIONS(1657),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1064)] = {
    [sym_attribute] = STATE(1064),
    [sym__lowercase_identifier] = ACTIONS(1679),
    [anon_sym_SEMI_SEMI] = ACTIONS(1681),
    [anon_sym_let] = ACTIONS(1679),
    [anon_sym_and] = ACTIONS(1679),
    [anon_sym_TILDE] = ACTIONS(1679),
    [anon_sym_QMARK] = ACTIONS(1679),
    [anon_sym_LPAREN] = ACTIONS(1681),
    [anon_sym_external] = ACTIONS(1679),
    [anon_sym_type] = ACTIONS(1679),
    [anon_sym_COMMA] = ACTIONS(1681),
    [anon_sym_PIPE] = ACTIONS(1679),
    [anon_sym_DOT] = ACTIONS(1681),
    [anon_sym_LBRACK] = ACTIONS(1679),
    [anon_sym_RBRACK] = ACTIONS(1681),
    [anon_sym_true] = ACTIONS(1679),
    [anon_sym_false] = ACTIONS(1679),
    [anon_sym_LBRACE] = ACTIONS(1679),
    [anon_sym_SEMI] = ACTIONS(1679),
    [anon_sym_exception] = ACTIONS(1679),
    [anon_sym_module] = ACTIONS(1679),
    [anon_sym_open] = ACTIONS(1679),
    [anon_sym_include] = ACTIONS(1679),
    [anon_sym_class] = ACTIONS(1679),
    [anon_sym_val] = ACTIONS(1679),
    [anon_sym_end] = ACTIONS(1679),
    [anon_sym_object] = ACTIONS(1679),
    [anon_sym_AMP] = ACTIONS(1679),
    [anon_sym_POUND] = ACTIONS(1679),
    [anon_sym_COLON_COLON] = ACTIONS(1681),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1681),
    [anon_sym_else] = ACTIONS(1679),
    [anon_sym_new] = ACTIONS(1679),
    [anon_sym_LBRACE_LT] = ACTIONS(1681),
    [anon_sym_begin] = ACTIONS(1679),
    [sym_ocamlyacc_value] = ACTIONS(1681),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1679),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1681),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1679),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1679),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1681),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1681),
    [aux_sym_number_token1] = ACTIONS(1681),
    [aux_sym_signed_number_token1] = ACTIONS(1679),
    [anon_sym_SQUOTE] = ACTIONS(1681),
    [anon_sym_DQUOTE] = ACTIONS(1681),
    [sym_prefix_operator] = ACTIONS(1681),
    [aux_sym_sign_operator_token1] = ACTIONS(1679),
    [sym_hash_operator] = ACTIONS(1681),
    [anon_sym_lsl] = ACTIONS(1679),
    [anon_sym_lsr] = ACTIONS(1679),
    [anon_sym_asr] = ACTIONS(1679),
    [aux_sym_pow_operator_token1] = ACTIONS(1681),
    [anon_sym_mod] = ACTIONS(1679),
    [anon_sym_land] = ACTIONS(1679),
    [anon_sym_lor] = ACTIONS(1679),
    [anon_sym_lxor] = ACTIONS(1679),
    [aux_sym_mult_operator_token1] = ACTIONS(1679),
    [aux_sym_add_operator_token1] = ACTIONS(1679),
    [sym_concat_operator] = ACTIONS(1681),
    [sym_rel_operator] = ACTIONS(1679),
    [anon_sym_AMP_AMP] = ACTIONS(1679),
    [anon_sym_or] = ACTIONS(1679),
    [anon_sym_PIPE_PIPE] = ACTIONS(1679),
    [sym_assign_operator] = ACTIONS(1681),
    [sym_let_operator] = ACTIONS(1681),
    [sym__uppercase_identifier] = ACTIONS(1681),
    [aux_sym_directive_token1] = ACTIONS(1679),
    [aux_sym_tag_token1] = ACTIONS(1681),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1065)] = {
    [sym_attribute] = STATE(1065),
    [sym__lowercase_identifier] = ACTIONS(1733),
    [anon_sym_SEMI_SEMI] = ACTIONS(1735),
    [anon_sym_let] = ACTIONS(1733),
    [anon_sym_and] = ACTIONS(1733),
    [anon_sym_TILDE] = ACTIONS(1733),
    [anon_sym_QMARK] = ACTIONS(1733),
    [anon_sym_LPAREN] = ACTIONS(1735),
    [anon_sym_external] = ACTIONS(1733),
    [anon_sym_type] = ACTIONS(1733),
    [anon_sym_COMMA] = ACTIONS(1735),
    [anon_sym_PIPE] = ACTIONS(1733),
    [anon_sym_DOT] = ACTIONS(1735),
    [anon_sym_LBRACK] = ACTIONS(1733),
    [anon_sym_RBRACK] = ACTIONS(1735),
    [anon_sym_true] = ACTIONS(1733),
    [anon_sym_false] = ACTIONS(1733),
    [anon_sym_LBRACE] = ACTIONS(1733),
    [anon_sym_SEMI] = ACTIONS(1733),
    [anon_sym_exception] = ACTIONS(1733),
    [anon_sym_module] = ACTIONS(1733),
    [anon_sym_open] = ACTIONS(1733),
    [anon_sym_include] = ACTIONS(1733),
    [anon_sym_class] = ACTIONS(1733),
    [anon_sym_val] = ACTIONS(1733),
    [anon_sym_end] = ACTIONS(1733),
    [anon_sym_object] = ACTIONS(1733),
    [anon_sym_AMP] = ACTIONS(1733),
    [anon_sym_POUND] = ACTIONS(1733),
    [anon_sym_COLON_COLON] = ACTIONS(1735),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1735),
    [anon_sym_else] = ACTIONS(1733),
    [anon_sym_new] = ACTIONS(1733),
    [anon_sym_LBRACE_LT] = ACTIONS(1735),
    [anon_sym_begin] = ACTIONS(1733),
    [sym_ocamlyacc_value] = ACTIONS(1735),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1733),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1735),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1733),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1733),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1735),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1735),
    [aux_sym_number_token1] = ACTIONS(1735),
    [aux_sym_signed_number_token1] = ACTIONS(1733),
    [anon_sym_SQUOTE] = ACTIONS(1735),
    [anon_sym_DQUOTE] = ACTIONS(1735),
    [sym_prefix_operator] = ACTIONS(1735),
    [aux_sym_sign_operator_token1] = ACTIONS(1733),
    [sym_hash_operator] = ACTIONS(1735),
    [anon_sym_lsl] = ACTIONS(1733),
    [anon_sym_lsr] = ACTIONS(1733),
    [anon_sym_asr] = ACTIONS(1733),
    [aux_sym_pow_operator_token1] = ACTIONS(1735),
    [anon_sym_mod] = ACTIONS(1733),
    [anon_sym_land] = ACTIONS(1733),
    [anon_sym_lor] = ACTIONS(1733),
    [anon_sym_lxor] = ACTIONS(1733),
    [aux_sym_mult_operator_token1] = ACTIONS(1733),
    [aux_sym_add_operator_token1] = ACTIONS(1733),
    [sym_concat_operator] = ACTIONS(1735),
    [sym_rel_operator] = ACTIONS(1733),
    [anon_sym_AMP_AMP] = ACTIONS(1733),
    [anon_sym_or] = ACTIONS(1733),
    [anon_sym_PIPE_PIPE] = ACTIONS(1733),
    [sym_assign_operator] = ACTIONS(1735),
    [sym_let_operator] = ACTIONS(1735),
    [sym__uppercase_identifier] = ACTIONS(1735),
    [aux_sym_directive_token1] = ACTIONS(1733),
    [aux_sym_tag_token1] = ACTIONS(1735),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1066)] = {
    [sym_attribute] = STATE(1066),
    [sym__lowercase_identifier] = ACTIONS(1817),
    [anon_sym_SEMI_SEMI] = ACTIONS(1819),
    [anon_sym_let] = ACTIONS(1817),
    [anon_sym_and] = ACTIONS(1817),
    [anon_sym_TILDE] = ACTIONS(1817),
    [anon_sym_QMARK] = ACTIONS(1817),
    [anon_sym_LPAREN] = ACTIONS(1819),
    [anon_sym_external] = ACTIONS(1817),
    [anon_sym_type] = ACTIONS(1817),
    [anon_sym_COMMA] = ACTIONS(1819),
    [anon_sym_PIPE] = ACTIONS(1817),
    [anon_sym_DOT] = ACTIONS(1819),
    [anon_sym_LBRACK] = ACTIONS(1817),
    [anon_sym_RBRACK] = ACTIONS(1819),
    [anon_sym_true] = ACTIONS(1817),
    [anon_sym_false] = ACTIONS(1817),
    [anon_sym_LBRACE] = ACTIONS(1817),
    [anon_sym_SEMI] = ACTIONS(1817),
    [anon_sym_exception] = ACTIONS(1817),
    [anon_sym_module] = ACTIONS(1817),
    [anon_sym_open] = ACTIONS(1817),
    [anon_sym_include] = ACTIONS(1817),
    [anon_sym_class] = ACTIONS(1817),
    [anon_sym_val] = ACTIONS(1817),
    [anon_sym_end] = ACTIONS(1817),
    [anon_sym_object] = ACTIONS(1817),
    [anon_sym_AMP] = ACTIONS(1817),
    [anon_sym_POUND] = ACTIONS(1817),
    [anon_sym_COLON_COLON] = ACTIONS(1819),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1819),
    [anon_sym_else] = ACTIONS(1817),
    [anon_sym_new] = ACTIONS(1817),
    [anon_sym_LBRACE_LT] = ACTIONS(1819),
    [anon_sym_begin] = ACTIONS(1817),
    [sym_ocamlyacc_value] = ACTIONS(1819),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1817),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1819),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1817),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1817),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1819),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1819),
    [aux_sym_number_token1] = ACTIONS(1819),
    [aux_sym_signed_number_token1] = ACTIONS(1817),
    [anon_sym_SQUOTE] = ACTIONS(1819),
    [anon_sym_DQUOTE] = ACTIONS(1819),
    [sym_prefix_operator] = ACTIONS(1819),
    [aux_sym_sign_operator_token1] = ACTIONS(1817),
    [sym_hash_operator] = ACTIONS(1819),
    [anon_sym_lsl] = ACTIONS(1817),
    [anon_sym_lsr] = ACTIONS(1817),
    [anon_sym_asr] = ACTIONS(1817),
    [aux_sym_pow_operator_token1] = ACTIONS(1819),
    [anon_sym_mod] = ACTIONS(1817),
    [anon_sym_land] = ACTIONS(1817),
    [anon_sym_lor] = ACTIONS(1817),
    [anon_sym_lxor] = ACTIONS(1817),
    [aux_sym_mult_operator_token1] = ACTIONS(1817),
    [aux_sym_add_operator_token1] = ACTIONS(1817),
    [sym_concat_operator] = ACTIONS(1819),
    [sym_rel_operator] = ACTIONS(1817),
    [anon_sym_AMP_AMP] = ACTIONS(1817),
    [anon_sym_or] = ACTIONS(1817),
    [anon_sym_PIPE_PIPE] = ACTIONS(1817),
    [sym_assign_operator] = ACTIONS(1819),
    [sym_let_operator] = ACTIONS(1819),
    [sym__uppercase_identifier] = ACTIONS(1819),
    [aux_sym_directive_token1] = ACTIONS(1817),
    [aux_sym_tag_token1] = ACTIONS(1819),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1067)] = {
    [sym_attribute] = STATE(1067),
    [sym__lowercase_identifier] = ACTIONS(1721),
    [anon_sym_SEMI_SEMI] = ACTIONS(1723),
    [anon_sym_let] = ACTIONS(1721),
    [anon_sym_and] = ACTIONS(1721),
    [anon_sym_TILDE] = ACTIONS(1721),
    [anon_sym_QMARK] = ACTIONS(1721),
    [anon_sym_LPAREN] = ACTIONS(1723),
    [anon_sym_external] = ACTIONS(1721),
    [anon_sym_type] = ACTIONS(1721),
    [anon_sym_COMMA] = ACTIONS(1723),
    [anon_sym_PIPE] = ACTIONS(1721),
    [anon_sym_DOT] = ACTIONS(1723),
    [anon_sym_LBRACK] = ACTIONS(1721),
    [anon_sym_RBRACK] = ACTIONS(1723),
    [anon_sym_true] = ACTIONS(1721),
    [anon_sym_false] = ACTIONS(1721),
    [anon_sym_LBRACE] = ACTIONS(1721),
    [anon_sym_SEMI] = ACTIONS(1721),
    [anon_sym_exception] = ACTIONS(1721),
    [anon_sym_module] = ACTIONS(1721),
    [anon_sym_open] = ACTIONS(1721),
    [anon_sym_include] = ACTIONS(1721),
    [anon_sym_class] = ACTIONS(1721),
    [anon_sym_val] = ACTIONS(1721),
    [anon_sym_end] = ACTIONS(1721),
    [anon_sym_object] = ACTIONS(1721),
    [anon_sym_AMP] = ACTIONS(1721),
    [anon_sym_POUND] = ACTIONS(1721),
    [anon_sym_COLON_COLON] = ACTIONS(1723),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1723),
    [anon_sym_else] = ACTIONS(1721),
    [anon_sym_new] = ACTIONS(1721),
    [anon_sym_LBRACE_LT] = ACTIONS(1723),
    [anon_sym_begin] = ACTIONS(1721),
    [sym_ocamlyacc_value] = ACTIONS(1723),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1721),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1723),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1721),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1721),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1723),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1723),
    [aux_sym_number_token1] = ACTIONS(1723),
    [aux_sym_signed_number_token1] = ACTIONS(1721),
    [anon_sym_SQUOTE] = ACTIONS(1723),
    [anon_sym_DQUOTE] = ACTIONS(1723),
    [sym_prefix_operator] = ACTIONS(1723),
    [aux_sym_sign_operator_token1] = ACTIONS(1721),
    [sym_hash_operator] = ACTIONS(1723),
    [anon_sym_lsl] = ACTIONS(1721),
    [anon_sym_lsr] = ACTIONS(1721),
    [anon_sym_asr] = ACTIONS(1721),
    [aux_sym_pow_operator_token1] = ACTIONS(1723),
    [anon_sym_mod] = ACTIONS(1721),
    [anon_sym_land] = ACTIONS(1721),
    [anon_sym_lor] = ACTIONS(1721),
    [anon_sym_lxor] = ACTIONS(1721),
    [aux_sym_mult_operator_token1] = ACTIONS(1721),
    [aux_sym_add_operator_token1] = ACTIONS(1721),
    [sym_concat_operator] = ACTIONS(1723),
    [sym_rel_operator] = ACTIONS(1721),
    [anon_sym_AMP_AMP] = ACTIONS(1721),
    [anon_sym_or] = ACTIONS(1721),
    [anon_sym_PIPE_PIPE] = ACTIONS(1721),
    [sym_assign_operator] = ACTIONS(1723),
    [sym_let_operator] = ACTIONS(1723),
    [sym__uppercase_identifier] = ACTIONS(1723),
    [aux_sym_directive_token1] = ACTIONS(1721),
    [aux_sym_tag_token1] = ACTIONS(1723),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1068)] = {
    [sym_attribute] = STATE(1068),
    [sym__lowercase_identifier] = ACTIONS(1737),
    [anon_sym_SEMI_SEMI] = ACTIONS(1739),
    [anon_sym_let] = ACTIONS(1737),
    [anon_sym_and] = ACTIONS(1737),
    [anon_sym_TILDE] = ACTIONS(1737),
    [anon_sym_QMARK] = ACTIONS(1737),
    [anon_sym_LPAREN] = ACTIONS(1739),
    [anon_sym_external] = ACTIONS(1737),
    [anon_sym_type] = ACTIONS(1737),
    [anon_sym_COMMA] = ACTIONS(1739),
    [anon_sym_PIPE] = ACTIONS(1737),
    [anon_sym_DOT] = ACTIONS(1739),
    [anon_sym_LBRACK] = ACTIONS(1737),
    [anon_sym_RBRACK] = ACTIONS(1739),
    [anon_sym_true] = ACTIONS(1737),
    [anon_sym_false] = ACTIONS(1737),
    [anon_sym_LBRACE] = ACTIONS(1737),
    [anon_sym_SEMI] = ACTIONS(1737),
    [anon_sym_exception] = ACTIONS(1737),
    [anon_sym_module] = ACTIONS(1737),
    [anon_sym_open] = ACTIONS(1737),
    [anon_sym_include] = ACTIONS(1737),
    [anon_sym_class] = ACTIONS(1737),
    [anon_sym_val] = ACTIONS(1737),
    [anon_sym_end] = ACTIONS(1737),
    [anon_sym_object] = ACTIONS(1737),
    [anon_sym_AMP] = ACTIONS(1737),
    [anon_sym_POUND] = ACTIONS(1737),
    [anon_sym_COLON_COLON] = ACTIONS(1739),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1739),
    [anon_sym_else] = ACTIONS(1737),
    [anon_sym_new] = ACTIONS(1737),
    [anon_sym_LBRACE_LT] = ACTIONS(1739),
    [anon_sym_begin] = ACTIONS(1737),
    [sym_ocamlyacc_value] = ACTIONS(1739),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1737),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1739),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1737),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1737),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1739),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1739),
    [aux_sym_number_token1] = ACTIONS(1739),
    [aux_sym_signed_number_token1] = ACTIONS(1737),
    [anon_sym_SQUOTE] = ACTIONS(1739),
    [anon_sym_DQUOTE] = ACTIONS(1739),
    [sym_prefix_operator] = ACTIONS(1739),
    [aux_sym_sign_operator_token1] = ACTIONS(1737),
    [sym_hash_operator] = ACTIONS(1739),
    [anon_sym_lsl] = ACTIONS(1737),
    [anon_sym_lsr] = ACTIONS(1737),
    [anon_sym_asr] = ACTIONS(1737),
    [aux_sym_pow_operator_token1] = ACTIONS(1739),
    [anon_sym_mod] = ACTIONS(1737),
    [anon_sym_land] = ACTIONS(1737),
    [anon_sym_lor] = ACTIONS(1737),
    [anon_sym_lxor] = ACTIONS(1737),
    [aux_sym_mult_operator_token1] = ACTIONS(1737),
    [aux_sym_add_operator_token1] = ACTIONS(1737),
    [sym_concat_operator] = ACTIONS(1739),
    [sym_rel_operator] = ACTIONS(1737),
    [anon_sym_AMP_AMP] = ACTIONS(1737),
    [anon_sym_or] = ACTIONS(1737),
    [anon_sym_PIPE_PIPE] = ACTIONS(1737),
    [sym_assign_operator] = ACTIONS(1739),
    [sym_let_operator] = ACTIONS(1739),
    [sym__uppercase_identifier] = ACTIONS(1739),
    [aux_sym_directive_token1] = ACTIONS(1737),
    [aux_sym_tag_token1] = ACTIONS(1739),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1069)] = {
    [sym_attribute] = STATE(1069),
    [sym__lowercase_identifier] = ACTIONS(1687),
    [anon_sym_SEMI_SEMI] = ACTIONS(1689),
    [anon_sym_let] = ACTIONS(1687),
    [anon_sym_and] = ACTIONS(1687),
    [anon_sym_TILDE] = ACTIONS(1687),
    [anon_sym_QMARK] = ACTIONS(1687),
    [anon_sym_LPAREN] = ACTIONS(1689),
    [anon_sym_external] = ACTIONS(1687),
    [anon_sym_type] = ACTIONS(1687),
    [anon_sym_COMMA] = ACTIONS(1689),
    [anon_sym_PIPE] = ACTIONS(1687),
    [anon_sym_DOT] = ACTIONS(1689),
    [anon_sym_LBRACK] = ACTIONS(1687),
    [anon_sym_RBRACK] = ACTIONS(1689),
    [anon_sym_true] = ACTIONS(1687),
    [anon_sym_false] = ACTIONS(1687),
    [anon_sym_LBRACE] = ACTIONS(1687),
    [anon_sym_SEMI] = ACTIONS(1687),
    [anon_sym_exception] = ACTIONS(1687),
    [anon_sym_module] = ACTIONS(1687),
    [anon_sym_open] = ACTIONS(1687),
    [anon_sym_include] = ACTIONS(1687),
    [anon_sym_class] = ACTIONS(1687),
    [anon_sym_val] = ACTIONS(1687),
    [anon_sym_end] = ACTIONS(1687),
    [anon_sym_object] = ACTIONS(1687),
    [anon_sym_AMP] = ACTIONS(1687),
    [anon_sym_POUND] = ACTIONS(1687),
    [anon_sym_COLON_COLON] = ACTIONS(1689),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1689),
    [anon_sym_else] = ACTIONS(1687),
    [anon_sym_new] = ACTIONS(1687),
    [anon_sym_LBRACE_LT] = ACTIONS(1689),
    [anon_sym_begin] = ACTIONS(1687),
    [sym_ocamlyacc_value] = ACTIONS(1689),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1687),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1689),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1687),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1687),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1689),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1689),
    [aux_sym_number_token1] = ACTIONS(1689),
    [aux_sym_signed_number_token1] = ACTIONS(1687),
    [anon_sym_SQUOTE] = ACTIONS(1689),
    [anon_sym_DQUOTE] = ACTIONS(1689),
    [sym_prefix_operator] = ACTIONS(1689),
    [aux_sym_sign_operator_token1] = ACTIONS(1687),
    [sym_hash_operator] = ACTIONS(1689),
    [anon_sym_lsl] = ACTIONS(1687),
    [anon_sym_lsr] = ACTIONS(1687),
    [anon_sym_asr] = ACTIONS(1687),
    [aux_sym_pow_operator_token1] = ACTIONS(1689),
    [anon_sym_mod] = ACTIONS(1687),
    [anon_sym_land] = ACTIONS(1687),
    [anon_sym_lor] = ACTIONS(1687),
    [anon_sym_lxor] = ACTIONS(1687),
    [aux_sym_mult_operator_token1] = ACTIONS(1687),
    [aux_sym_add_operator_token1] = ACTIONS(1687),
    [sym_concat_operator] = ACTIONS(1689),
    [sym_rel_operator] = ACTIONS(1687),
    [anon_sym_AMP_AMP] = ACTIONS(1687),
    [anon_sym_or] = ACTIONS(1687),
    [anon_sym_PIPE_PIPE] = ACTIONS(1687),
    [sym_assign_operator] = ACTIONS(1689),
    [sym_let_operator] = ACTIONS(1689),
    [sym__uppercase_identifier] = ACTIONS(1689),
    [aux_sym_directive_token1] = ACTIONS(1687),
    [aux_sym_tag_token1] = ACTIONS(1689),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1070)] = {
    [sym_attribute] = STATE(1070),
    [sym__lowercase_identifier] = ACTIONS(1599),
    [anon_sym_SEMI_SEMI] = ACTIONS(1601),
    [anon_sym_let] = ACTIONS(1599),
    [anon_sym_and] = ACTIONS(1599),
    [anon_sym_TILDE] = ACTIONS(1599),
    [anon_sym_QMARK] = ACTIONS(1599),
    [anon_sym_LPAREN] = ACTIONS(1601),
    [anon_sym_external] = ACTIONS(1599),
    [anon_sym_type] = ACTIONS(1599),
    [anon_sym_COMMA] = ACTIONS(1601),
    [anon_sym_PIPE] = ACTIONS(1599),
    [anon_sym_DOT] = ACTIONS(1601),
    [anon_sym_LBRACK] = ACTIONS(1599),
    [anon_sym_RBRACK] = ACTIONS(1601),
    [anon_sym_true] = ACTIONS(1599),
    [anon_sym_false] = ACTIONS(1599),
    [anon_sym_LBRACE] = ACTIONS(1599),
    [anon_sym_SEMI] = ACTIONS(1599),
    [anon_sym_exception] = ACTIONS(1599),
    [anon_sym_module] = ACTIONS(1599),
    [anon_sym_open] = ACTIONS(1599),
    [anon_sym_include] = ACTIONS(1599),
    [anon_sym_class] = ACTIONS(1599),
    [anon_sym_val] = ACTIONS(1599),
    [anon_sym_end] = ACTIONS(1599),
    [anon_sym_object] = ACTIONS(1599),
    [anon_sym_AMP] = ACTIONS(1599),
    [anon_sym_POUND] = ACTIONS(1599),
    [anon_sym_COLON_COLON] = ACTIONS(1601),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1601),
    [anon_sym_else] = ACTIONS(1599),
    [anon_sym_new] = ACTIONS(1599),
    [anon_sym_LBRACE_LT] = ACTIONS(1601),
    [anon_sym_begin] = ACTIONS(1599),
    [sym_ocamlyacc_value] = ACTIONS(1601),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1599),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1601),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1599),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1599),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1601),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1601),
    [aux_sym_number_token1] = ACTIONS(1601),
    [aux_sym_signed_number_token1] = ACTIONS(1599),
    [anon_sym_SQUOTE] = ACTIONS(1601),
    [anon_sym_DQUOTE] = ACTIONS(1601),
    [sym_prefix_operator] = ACTIONS(1601),
    [aux_sym_sign_operator_token1] = ACTIONS(1599),
    [sym_hash_operator] = ACTIONS(1601),
    [anon_sym_lsl] = ACTIONS(1599),
    [anon_sym_lsr] = ACTIONS(1599),
    [anon_sym_asr] = ACTIONS(1599),
    [aux_sym_pow_operator_token1] = ACTIONS(1601),
    [anon_sym_mod] = ACTIONS(1599),
    [anon_sym_land] = ACTIONS(1599),
    [anon_sym_lor] = ACTIONS(1599),
    [anon_sym_lxor] = ACTIONS(1599),
    [aux_sym_mult_operator_token1] = ACTIONS(1599),
    [aux_sym_add_operator_token1] = ACTIONS(1599),
    [sym_concat_operator] = ACTIONS(1601),
    [sym_rel_operator] = ACTIONS(1599),
    [anon_sym_AMP_AMP] = ACTIONS(1599),
    [anon_sym_or] = ACTIONS(1599),
    [anon_sym_PIPE_PIPE] = ACTIONS(1599),
    [sym_assign_operator] = ACTIONS(1601),
    [sym_let_operator] = ACTIONS(1601),
    [sym__uppercase_identifier] = ACTIONS(1601),
    [aux_sym_directive_token1] = ACTIONS(1599),
    [aux_sym_tag_token1] = ACTIONS(1601),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1071)] = {
    [sym_attribute] = STATE(1071),
    [sym__lowercase_identifier] = ACTIONS(1587),
    [anon_sym_SEMI_SEMI] = ACTIONS(1589),
    [anon_sym_let] = ACTIONS(1587),
    [anon_sym_and] = ACTIONS(1587),
    [anon_sym_TILDE] = ACTIONS(1587),
    [anon_sym_QMARK] = ACTIONS(1587),
    [anon_sym_LPAREN] = ACTIONS(1589),
    [anon_sym_external] = ACTIONS(1587),
    [anon_sym_type] = ACTIONS(1587),
    [anon_sym_COMMA] = ACTIONS(1589),
    [anon_sym_PIPE] = ACTIONS(1587),
    [anon_sym_DOT] = ACTIONS(1589),
    [anon_sym_LBRACK] = ACTIONS(1587),
    [anon_sym_RBRACK] = ACTIONS(1589),
    [anon_sym_true] = ACTIONS(1587),
    [anon_sym_false] = ACTIONS(1587),
    [anon_sym_LBRACE] = ACTIONS(1587),
    [anon_sym_SEMI] = ACTIONS(1587),
    [anon_sym_exception] = ACTIONS(1587),
    [anon_sym_module] = ACTIONS(1587),
    [anon_sym_open] = ACTIONS(1587),
    [anon_sym_include] = ACTIONS(1587),
    [anon_sym_class] = ACTIONS(1587),
    [anon_sym_val] = ACTIONS(1587),
    [anon_sym_end] = ACTIONS(1587),
    [anon_sym_object] = ACTIONS(1587),
    [anon_sym_AMP] = ACTIONS(1587),
    [anon_sym_POUND] = ACTIONS(1587),
    [anon_sym_COLON_COLON] = ACTIONS(1589),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1589),
    [anon_sym_else] = ACTIONS(1587),
    [anon_sym_new] = ACTIONS(1587),
    [anon_sym_LBRACE_LT] = ACTIONS(1589),
    [anon_sym_begin] = ACTIONS(1587),
    [sym_ocamlyacc_value] = ACTIONS(1589),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1587),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1589),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1587),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1587),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1589),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1589),
    [aux_sym_number_token1] = ACTIONS(1589),
    [aux_sym_signed_number_token1] = ACTIONS(1587),
    [anon_sym_SQUOTE] = ACTIONS(1589),
    [anon_sym_DQUOTE] = ACTIONS(1589),
    [sym_prefix_operator] = ACTIONS(1589),
    [aux_sym_sign_operator_token1] = ACTIONS(1587),
    [sym_hash_operator] = ACTIONS(1589),
    [anon_sym_lsl] = ACTIONS(1587),
    [anon_sym_lsr] = ACTIONS(1587),
    [anon_sym_asr] = ACTIONS(1587),
    [aux_sym_pow_operator_token1] = ACTIONS(1589),
    [anon_sym_mod] = ACTIONS(1587),
    [anon_sym_land] = ACTIONS(1587),
    [anon_sym_lor] = ACTIONS(1587),
    [anon_sym_lxor] = ACTIONS(1587),
    [aux_sym_mult_operator_token1] = ACTIONS(1587),
    [aux_sym_add_operator_token1] = ACTIONS(1587),
    [sym_concat_operator] = ACTIONS(1589),
    [sym_rel_operator] = ACTIONS(1587),
    [anon_sym_AMP_AMP] = ACTIONS(1587),
    [anon_sym_or] = ACTIONS(1587),
    [anon_sym_PIPE_PIPE] = ACTIONS(1587),
    [sym_assign_operator] = ACTIONS(1589),
    [sym_let_operator] = ACTIONS(1589),
    [sym__uppercase_identifier] = ACTIONS(1589),
    [aux_sym_directive_token1] = ACTIONS(1587),
    [aux_sym_tag_token1] = ACTIONS(1589),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1072)] = {
    [sym_attribute] = STATE(1072),
    [sym__lowercase_identifier] = ACTIONS(1833),
    [anon_sym_SEMI_SEMI] = ACTIONS(1835),
    [anon_sym_let] = ACTIONS(1833),
    [anon_sym_and] = ACTIONS(1833),
    [anon_sym_TILDE] = ACTIONS(1833),
    [anon_sym_QMARK] = ACTIONS(1833),
    [anon_sym_LPAREN] = ACTIONS(1835),
    [anon_sym_external] = ACTIONS(1833),
    [anon_sym_type] = ACTIONS(1833),
    [anon_sym_COMMA] = ACTIONS(1835),
    [anon_sym_PIPE] = ACTIONS(1833),
    [anon_sym_DOT] = ACTIONS(1835),
    [anon_sym_LBRACK] = ACTIONS(1833),
    [anon_sym_RBRACK] = ACTIONS(1835),
    [anon_sym_true] = ACTIONS(1833),
    [anon_sym_false] = ACTIONS(1833),
    [anon_sym_LBRACE] = ACTIONS(1833),
    [anon_sym_SEMI] = ACTIONS(1833),
    [anon_sym_exception] = ACTIONS(1833),
    [anon_sym_module] = ACTIONS(1833),
    [anon_sym_open] = ACTIONS(1833),
    [anon_sym_include] = ACTIONS(1833),
    [anon_sym_class] = ACTIONS(1833),
    [anon_sym_val] = ACTIONS(1833),
    [anon_sym_end] = ACTIONS(1833),
    [anon_sym_object] = ACTIONS(1833),
    [anon_sym_AMP] = ACTIONS(1833),
    [anon_sym_POUND] = ACTIONS(1833),
    [anon_sym_COLON_COLON] = ACTIONS(1835),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1835),
    [anon_sym_else] = ACTIONS(1833),
    [anon_sym_new] = ACTIONS(1833),
    [anon_sym_LBRACE_LT] = ACTIONS(1835),
    [anon_sym_begin] = ACTIONS(1833),
    [sym_ocamlyacc_value] = ACTIONS(1835),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1833),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1835),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1833),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1833),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1835),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1835),
    [aux_sym_number_token1] = ACTIONS(1835),
    [aux_sym_signed_number_token1] = ACTIONS(1833),
    [anon_sym_SQUOTE] = ACTIONS(1835),
    [anon_sym_DQUOTE] = ACTIONS(1835),
    [sym_prefix_operator] = ACTIONS(1835),
    [aux_sym_sign_operator_token1] = ACTIONS(1833),
    [sym_hash_operator] = ACTIONS(1835),
    [anon_sym_lsl] = ACTIONS(1833),
    [anon_sym_lsr] = ACTIONS(1833),
    [anon_sym_asr] = ACTIONS(1833),
    [aux_sym_pow_operator_token1] = ACTIONS(1835),
    [anon_sym_mod] = ACTIONS(1833),
    [anon_sym_land] = ACTIONS(1833),
    [anon_sym_lor] = ACTIONS(1833),
    [anon_sym_lxor] = ACTIONS(1833),
    [aux_sym_mult_operator_token1] = ACTIONS(1833),
    [aux_sym_add_operator_token1] = ACTIONS(1833),
    [sym_concat_operator] = ACTIONS(1835),
    [sym_rel_operator] = ACTIONS(1833),
    [anon_sym_AMP_AMP] = ACTIONS(1833),
    [anon_sym_or] = ACTIONS(1833),
    [anon_sym_PIPE_PIPE] = ACTIONS(1833),
    [sym_assign_operator] = ACTIONS(1835),
    [sym_let_operator] = ACTIONS(1835),
    [sym__uppercase_identifier] = ACTIONS(1835),
    [aux_sym_directive_token1] = ACTIONS(1833),
    [aux_sym_tag_token1] = ACTIONS(1835),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1073)] = {
    [sym_attribute] = STATE(1073),
    [sym__lowercase_identifier] = ACTIONS(1769),
    [anon_sym_SEMI_SEMI] = ACTIONS(1771),
    [anon_sym_let] = ACTIONS(1769),
    [anon_sym_and] = ACTIONS(1769),
    [anon_sym_TILDE] = ACTIONS(1769),
    [anon_sym_QMARK] = ACTIONS(1769),
    [anon_sym_LPAREN] = ACTIONS(1771),
    [anon_sym_external] = ACTIONS(1769),
    [anon_sym_type] = ACTIONS(1769),
    [anon_sym_COMMA] = ACTIONS(1771),
    [anon_sym_PIPE] = ACTIONS(1769),
    [anon_sym_DOT] = ACTIONS(1771),
    [anon_sym_LBRACK] = ACTIONS(1769),
    [anon_sym_RBRACK] = ACTIONS(1771),
    [anon_sym_true] = ACTIONS(1769),
    [anon_sym_false] = ACTIONS(1769),
    [anon_sym_LBRACE] = ACTIONS(1769),
    [anon_sym_SEMI] = ACTIONS(1769),
    [anon_sym_exception] = ACTIONS(1769),
    [anon_sym_module] = ACTIONS(1769),
    [anon_sym_open] = ACTIONS(1769),
    [anon_sym_include] = ACTIONS(1769),
    [anon_sym_class] = ACTIONS(1769),
    [anon_sym_val] = ACTIONS(1769),
    [anon_sym_end] = ACTIONS(1769),
    [anon_sym_object] = ACTIONS(1769),
    [anon_sym_AMP] = ACTIONS(1769),
    [anon_sym_POUND] = ACTIONS(1769),
    [anon_sym_COLON_COLON] = ACTIONS(1771),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1771),
    [anon_sym_else] = ACTIONS(1769),
    [anon_sym_new] = ACTIONS(1769),
    [anon_sym_LBRACE_LT] = ACTIONS(1771),
    [anon_sym_begin] = ACTIONS(1769),
    [sym_ocamlyacc_value] = ACTIONS(1771),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1769),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1771),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1769),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1769),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1771),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1771),
    [aux_sym_number_token1] = ACTIONS(1771),
    [aux_sym_signed_number_token1] = ACTIONS(1769),
    [anon_sym_SQUOTE] = ACTIONS(1771),
    [anon_sym_DQUOTE] = ACTIONS(1771),
    [sym_prefix_operator] = ACTIONS(1771),
    [aux_sym_sign_operator_token1] = ACTIONS(1769),
    [sym_hash_operator] = ACTIONS(1771),
    [anon_sym_lsl] = ACTIONS(1769),
    [anon_sym_lsr] = ACTIONS(1769),
    [anon_sym_asr] = ACTIONS(1769),
    [aux_sym_pow_operator_token1] = ACTIONS(1771),
    [anon_sym_mod] = ACTIONS(1769),
    [anon_sym_land] = ACTIONS(1769),
    [anon_sym_lor] = ACTIONS(1769),
    [anon_sym_lxor] = ACTIONS(1769),
    [aux_sym_mult_operator_token1] = ACTIONS(1769),
    [aux_sym_add_operator_token1] = ACTIONS(1769),
    [sym_concat_operator] = ACTIONS(1771),
    [sym_rel_operator] = ACTIONS(1769),
    [anon_sym_AMP_AMP] = ACTIONS(1769),
    [anon_sym_or] = ACTIONS(1769),
    [anon_sym_PIPE_PIPE] = ACTIONS(1769),
    [sym_assign_operator] = ACTIONS(1771),
    [sym_let_operator] = ACTIONS(1771),
    [sym__uppercase_identifier] = ACTIONS(1771),
    [aux_sym_directive_token1] = ACTIONS(1769),
    [aux_sym_tag_token1] = ACTIONS(1771),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1074)] = {
    [sym_attribute] = STATE(1074),
    [sym__lowercase_identifier] = ACTIONS(1789),
    [anon_sym_SEMI_SEMI] = ACTIONS(1791),
    [anon_sym_let] = ACTIONS(1789),
    [anon_sym_and] = ACTIONS(1789),
    [anon_sym_TILDE] = ACTIONS(1789),
    [anon_sym_QMARK] = ACTIONS(1789),
    [anon_sym_LPAREN] = ACTIONS(1791),
    [anon_sym_external] = ACTIONS(1789),
    [anon_sym_type] = ACTIONS(1789),
    [anon_sym_COMMA] = ACTIONS(1791),
    [anon_sym_PIPE] = ACTIONS(1789),
    [anon_sym_DOT] = ACTIONS(1791),
    [anon_sym_LBRACK] = ACTIONS(1789),
    [anon_sym_RBRACK] = ACTIONS(1791),
    [anon_sym_true] = ACTIONS(1789),
    [anon_sym_false] = ACTIONS(1789),
    [anon_sym_LBRACE] = ACTIONS(1789),
    [anon_sym_SEMI] = ACTIONS(1789),
    [anon_sym_exception] = ACTIONS(1789),
    [anon_sym_module] = ACTIONS(1789),
    [anon_sym_open] = ACTIONS(1789),
    [anon_sym_include] = ACTIONS(1789),
    [anon_sym_class] = ACTIONS(1789),
    [anon_sym_val] = ACTIONS(1789),
    [anon_sym_end] = ACTIONS(1789),
    [anon_sym_object] = ACTIONS(1789),
    [anon_sym_AMP] = ACTIONS(1789),
    [anon_sym_POUND] = ACTIONS(1789),
    [anon_sym_COLON_COLON] = ACTIONS(1791),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1791),
    [anon_sym_else] = ACTIONS(1789),
    [anon_sym_new] = ACTIONS(1789),
    [anon_sym_LBRACE_LT] = ACTIONS(1791),
    [anon_sym_begin] = ACTIONS(1789),
    [sym_ocamlyacc_value] = ACTIONS(1791),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1789),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1791),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1789),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1789),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1791),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1791),
    [aux_sym_number_token1] = ACTIONS(1791),
    [aux_sym_signed_number_token1] = ACTIONS(1789),
    [anon_sym_SQUOTE] = ACTIONS(1791),
    [anon_sym_DQUOTE] = ACTIONS(1791),
    [sym_prefix_operator] = ACTIONS(1791),
    [aux_sym_sign_operator_token1] = ACTIONS(1789),
    [sym_hash_operator] = ACTIONS(1791),
    [anon_sym_lsl] = ACTIONS(1789),
    [anon_sym_lsr] = ACTIONS(1789),
    [anon_sym_asr] = ACTIONS(1789),
    [aux_sym_pow_operator_token1] = ACTIONS(1791),
    [anon_sym_mod] = ACTIONS(1789),
    [anon_sym_land] = ACTIONS(1789),
    [anon_sym_lor] = ACTIONS(1789),
    [anon_sym_lxor] = ACTIONS(1789),
    [aux_sym_mult_operator_token1] = ACTIONS(1789),
    [aux_sym_add_operator_token1] = ACTIONS(1789),
    [sym_concat_operator] = ACTIONS(1791),
    [sym_rel_operator] = ACTIONS(1789),
    [anon_sym_AMP_AMP] = ACTIONS(1789),
    [anon_sym_or] = ACTIONS(1789),
    [anon_sym_PIPE_PIPE] = ACTIONS(1789),
    [sym_assign_operator] = ACTIONS(1791),
    [sym_let_operator] = ACTIONS(1791),
    [sym__uppercase_identifier] = ACTIONS(1791),
    [aux_sym_directive_token1] = ACTIONS(1789),
    [aux_sym_tag_token1] = ACTIONS(1791),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1075)] = {
    [sym_attribute] = STATE(1075),
    [sym__lowercase_identifier] = ACTIONS(1603),
    [anon_sym_SEMI_SEMI] = ACTIONS(1605),
    [anon_sym_let] = ACTIONS(1603),
    [anon_sym_and] = ACTIONS(1603),
    [anon_sym_TILDE] = ACTIONS(1603),
    [anon_sym_QMARK] = ACTIONS(1603),
    [anon_sym_LPAREN] = ACTIONS(1605),
    [anon_sym_external] = ACTIONS(1603),
    [anon_sym_type] = ACTIONS(1603),
    [anon_sym_COMMA] = ACTIONS(1605),
    [anon_sym_PIPE] = ACTIONS(1603),
    [anon_sym_DOT] = ACTIONS(1605),
    [anon_sym_LBRACK] = ACTIONS(1603),
    [anon_sym_RBRACK] = ACTIONS(1605),
    [anon_sym_true] = ACTIONS(1603),
    [anon_sym_false] = ACTIONS(1603),
    [anon_sym_LBRACE] = ACTIONS(1603),
    [anon_sym_SEMI] = ACTIONS(1603),
    [anon_sym_exception] = ACTIONS(1603),
    [anon_sym_module] = ACTIONS(1603),
    [anon_sym_open] = ACTIONS(1603),
    [anon_sym_include] = ACTIONS(1603),
    [anon_sym_class] = ACTIONS(1603),
    [anon_sym_val] = ACTIONS(1603),
    [anon_sym_end] = ACTIONS(1603),
    [anon_sym_object] = ACTIONS(1603),
    [anon_sym_AMP] = ACTIONS(1603),
    [anon_sym_POUND] = ACTIONS(1603),
    [anon_sym_COLON_COLON] = ACTIONS(1605),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1605),
    [anon_sym_else] = ACTIONS(1603),
    [anon_sym_new] = ACTIONS(1603),
    [anon_sym_LBRACE_LT] = ACTIONS(1605),
    [anon_sym_begin] = ACTIONS(1603),
    [sym_ocamlyacc_value] = ACTIONS(1605),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1603),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1605),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1603),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1603),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1605),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1605),
    [aux_sym_number_token1] = ACTIONS(1605),
    [aux_sym_signed_number_token1] = ACTIONS(1603),
    [anon_sym_SQUOTE] = ACTIONS(1605),
    [anon_sym_DQUOTE] = ACTIONS(1605),
    [sym_prefix_operator] = ACTIONS(1605),
    [aux_sym_sign_operator_token1] = ACTIONS(1603),
    [sym_hash_operator] = ACTIONS(1605),
    [anon_sym_lsl] = ACTIONS(1603),
    [anon_sym_lsr] = ACTIONS(1603),
    [anon_sym_asr] = ACTIONS(1603),
    [aux_sym_pow_operator_token1] = ACTIONS(1605),
    [anon_sym_mod] = ACTIONS(1603),
    [anon_sym_land] = ACTIONS(1603),
    [anon_sym_lor] = ACTIONS(1603),
    [anon_sym_lxor] = ACTIONS(1603),
    [aux_sym_mult_operator_token1] = ACTIONS(1603),
    [aux_sym_add_operator_token1] = ACTIONS(1603),
    [sym_concat_operator] = ACTIONS(1605),
    [sym_rel_operator] = ACTIONS(1603),
    [anon_sym_AMP_AMP] = ACTIONS(1603),
    [anon_sym_or] = ACTIONS(1603),
    [anon_sym_PIPE_PIPE] = ACTIONS(1603),
    [sym_assign_operator] = ACTIONS(1605),
    [sym_let_operator] = ACTIONS(1605),
    [sym__uppercase_identifier] = ACTIONS(1605),
    [aux_sym_directive_token1] = ACTIONS(1603),
    [aux_sym_tag_token1] = ACTIONS(1605),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1076)] = {
    [sym_attribute] = STATE(1076),
    [sym__lowercase_identifier] = ACTIONS(1691),
    [anon_sym_SEMI_SEMI] = ACTIONS(1693),
    [anon_sym_let] = ACTIONS(1691),
    [anon_sym_and] = ACTIONS(1691),
    [anon_sym_TILDE] = ACTIONS(1691),
    [anon_sym_QMARK] = ACTIONS(1691),
    [anon_sym_LPAREN] = ACTIONS(1693),
    [anon_sym_external] = ACTIONS(1691),
    [anon_sym_type] = ACTIONS(1691),
    [anon_sym_COMMA] = ACTIONS(1693),
    [anon_sym_PIPE] = ACTIONS(1691),
    [anon_sym_DOT] = ACTIONS(1693),
    [anon_sym_LBRACK] = ACTIONS(1691),
    [anon_sym_RBRACK] = ACTIONS(1693),
    [anon_sym_true] = ACTIONS(1691),
    [anon_sym_false] = ACTIONS(1691),
    [anon_sym_LBRACE] = ACTIONS(1691),
    [anon_sym_SEMI] = ACTIONS(1691),
    [anon_sym_exception] = ACTIONS(1691),
    [anon_sym_module] = ACTIONS(1691),
    [anon_sym_open] = ACTIONS(1691),
    [anon_sym_include] = ACTIONS(1691),
    [anon_sym_class] = ACTIONS(1691),
    [anon_sym_val] = ACTIONS(1691),
    [anon_sym_end] = ACTIONS(1691),
    [anon_sym_object] = ACTIONS(1691),
    [anon_sym_AMP] = ACTIONS(1691),
    [anon_sym_POUND] = ACTIONS(1691),
    [anon_sym_COLON_COLON] = ACTIONS(1693),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1693),
    [anon_sym_else] = ACTIONS(1691),
    [anon_sym_new] = ACTIONS(1691),
    [anon_sym_LBRACE_LT] = ACTIONS(1693),
    [anon_sym_begin] = ACTIONS(1691),
    [sym_ocamlyacc_value] = ACTIONS(1693),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1691),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1693),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1691),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1691),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1693),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1693),
    [aux_sym_number_token1] = ACTIONS(1693),
    [aux_sym_signed_number_token1] = ACTIONS(1691),
    [anon_sym_SQUOTE] = ACTIONS(1693),
    [anon_sym_DQUOTE] = ACTIONS(1693),
    [sym_prefix_operator] = ACTIONS(1693),
    [aux_sym_sign_operator_token1] = ACTIONS(1691),
    [sym_hash_operator] = ACTIONS(1693),
    [anon_sym_lsl] = ACTIONS(1691),
    [anon_sym_lsr] = ACTIONS(1691),
    [anon_sym_asr] = ACTIONS(1691),
    [aux_sym_pow_operator_token1] = ACTIONS(1693),
    [anon_sym_mod] = ACTIONS(1691),
    [anon_sym_land] = ACTIONS(1691),
    [anon_sym_lor] = ACTIONS(1691),
    [anon_sym_lxor] = ACTIONS(1691),
    [aux_sym_mult_operator_token1] = ACTIONS(1691),
    [aux_sym_add_operator_token1] = ACTIONS(1691),
    [sym_concat_operator] = ACTIONS(1693),
    [sym_rel_operator] = ACTIONS(1691),
    [anon_sym_AMP_AMP] = ACTIONS(1691),
    [anon_sym_or] = ACTIONS(1691),
    [anon_sym_PIPE_PIPE] = ACTIONS(1691),
    [sym_assign_operator] = ACTIONS(1693),
    [sym_let_operator] = ACTIONS(1693),
    [sym__uppercase_identifier] = ACTIONS(1693),
    [aux_sym_directive_token1] = ACTIONS(1691),
    [aux_sym_tag_token1] = ACTIONS(1693),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1077)] = {
    [sym_attribute] = STATE(1077),
    [sym__lowercase_identifier] = ACTIONS(1847),
    [anon_sym_SEMI_SEMI] = ACTIONS(1849),
    [anon_sym_let] = ACTIONS(1847),
    [anon_sym_and] = ACTIONS(1847),
    [anon_sym_TILDE] = ACTIONS(1847),
    [anon_sym_QMARK] = ACTIONS(1847),
    [anon_sym_LPAREN] = ACTIONS(1849),
    [anon_sym_external] = ACTIONS(1847),
    [anon_sym_type] = ACTIONS(1847),
    [anon_sym_COMMA] = ACTIONS(1849),
    [anon_sym_PIPE] = ACTIONS(1847),
    [anon_sym_DOT] = ACTIONS(1849),
    [anon_sym_LBRACK] = ACTIONS(1847),
    [anon_sym_RBRACK] = ACTIONS(1849),
    [anon_sym_true] = ACTIONS(1847),
    [anon_sym_false] = ACTIONS(1847),
    [anon_sym_LBRACE] = ACTIONS(1847),
    [anon_sym_SEMI] = ACTIONS(1847),
    [anon_sym_exception] = ACTIONS(1847),
    [anon_sym_module] = ACTIONS(1847),
    [anon_sym_open] = ACTIONS(1847),
    [anon_sym_include] = ACTIONS(1847),
    [anon_sym_class] = ACTIONS(1847),
    [anon_sym_val] = ACTIONS(1847),
    [anon_sym_end] = ACTIONS(1847),
    [anon_sym_object] = ACTIONS(1847),
    [anon_sym_AMP] = ACTIONS(1847),
    [anon_sym_POUND] = ACTIONS(1847),
    [anon_sym_COLON_COLON] = ACTIONS(1849),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1849),
    [anon_sym_else] = ACTIONS(1847),
    [anon_sym_new] = ACTIONS(1847),
    [anon_sym_LBRACE_LT] = ACTIONS(1849),
    [anon_sym_begin] = ACTIONS(1847),
    [sym_ocamlyacc_value] = ACTIONS(1849),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1847),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1849),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1847),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1847),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1849),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1849),
    [aux_sym_number_token1] = ACTIONS(1849),
    [aux_sym_signed_number_token1] = ACTIONS(1847),
    [anon_sym_SQUOTE] = ACTIONS(1849),
    [anon_sym_DQUOTE] = ACTIONS(1849),
    [sym_prefix_operator] = ACTIONS(1849),
    [aux_sym_sign_operator_token1] = ACTIONS(1847),
    [sym_hash_operator] = ACTIONS(1849),
    [anon_sym_lsl] = ACTIONS(1847),
    [anon_sym_lsr] = ACTIONS(1847),
    [anon_sym_asr] = ACTIONS(1847),
    [aux_sym_pow_operator_token1] = ACTIONS(1849),
    [anon_sym_mod] = ACTIONS(1847),
    [anon_sym_land] = ACTIONS(1847),
    [anon_sym_lor] = ACTIONS(1847),
    [anon_sym_lxor] = ACTIONS(1847),
    [aux_sym_mult_operator_token1] = ACTIONS(1847),
    [aux_sym_add_operator_token1] = ACTIONS(1847),
    [sym_concat_operator] = ACTIONS(1849),
    [sym_rel_operator] = ACTIONS(1847),
    [anon_sym_AMP_AMP] = ACTIONS(1847),
    [anon_sym_or] = ACTIONS(1847),
    [anon_sym_PIPE_PIPE] = ACTIONS(1847),
    [sym_assign_operator] = ACTIONS(1849),
    [sym_let_operator] = ACTIONS(1849),
    [sym__uppercase_identifier] = ACTIONS(1849),
    [aux_sym_directive_token1] = ACTIONS(1847),
    [aux_sym_tag_token1] = ACTIONS(1849),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1078)] = {
    [sym_attribute] = STATE(1078),
    [sym__lowercase_identifier] = ACTIONS(1615),
    [anon_sym_SEMI_SEMI] = ACTIONS(1617),
    [anon_sym_let] = ACTIONS(1615),
    [anon_sym_and] = ACTIONS(1615),
    [anon_sym_TILDE] = ACTIONS(1615),
    [anon_sym_QMARK] = ACTIONS(1615),
    [anon_sym_LPAREN] = ACTIONS(1617),
    [anon_sym_external] = ACTIONS(1615),
    [anon_sym_type] = ACTIONS(1615),
    [anon_sym_COMMA] = ACTIONS(1617),
    [anon_sym_PIPE] = ACTIONS(1615),
    [anon_sym_DOT] = ACTIONS(969),
    [anon_sym_LBRACK] = ACTIONS(1615),
    [anon_sym_RBRACK] = ACTIONS(1617),
    [anon_sym_true] = ACTIONS(1615),
    [anon_sym_false] = ACTIONS(1615),
    [anon_sym_LBRACE] = ACTIONS(1615),
    [anon_sym_SEMI] = ACTIONS(1615),
    [anon_sym_exception] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(1615),
    [anon_sym_open] = ACTIONS(1615),
    [anon_sym_include] = ACTIONS(1615),
    [anon_sym_class] = ACTIONS(1615),
    [anon_sym_val] = ACTIONS(1615),
    [anon_sym_end] = ACTIONS(1615),
    [anon_sym_object] = ACTIONS(1615),
    [anon_sym_AMP] = ACTIONS(1615),
    [anon_sym_POUND] = ACTIONS(971),
    [anon_sym_COLON_COLON] = ACTIONS(1617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1617),
    [anon_sym_else] = ACTIONS(1615),
    [anon_sym_new] = ACTIONS(1615),
    [anon_sym_LBRACE_LT] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1615),
    [sym_ocamlyacc_value] = ACTIONS(1617),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1615),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1617),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1615),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1615),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1617),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1617),
    [aux_sym_number_token1] = ACTIONS(1617),
    [aux_sym_signed_number_token1] = ACTIONS(1615),
    [anon_sym_SQUOTE] = ACTIONS(1617),
    [anon_sym_DQUOTE] = ACTIONS(1617),
    [sym_prefix_operator] = ACTIONS(1617),
    [aux_sym_sign_operator_token1] = ACTIONS(1615),
    [sym_hash_operator] = ACTIONS(973),
    [anon_sym_lsl] = ACTIONS(1615),
    [anon_sym_lsr] = ACTIONS(1615),
    [anon_sym_asr] = ACTIONS(1615),
    [aux_sym_pow_operator_token1] = ACTIONS(1617),
    [anon_sym_mod] = ACTIONS(1615),
    [anon_sym_land] = ACTIONS(1615),
    [anon_sym_lor] = ACTIONS(1615),
    [anon_sym_lxor] = ACTIONS(1615),
    [aux_sym_mult_operator_token1] = ACTIONS(1615),
    [aux_sym_add_operator_token1] = ACTIONS(1615),
    [sym_concat_operator] = ACTIONS(1617),
    [sym_rel_operator] = ACTIONS(1615),
    [anon_sym_AMP_AMP] = ACTIONS(1615),
    [anon_sym_or] = ACTIONS(1615),
    [anon_sym_PIPE_PIPE] = ACTIONS(1615),
    [sym_assign_operator] = ACTIONS(1617),
    [sym_let_operator] = ACTIONS(1617),
    [sym__uppercase_identifier] = ACTIONS(1617),
    [aux_sym_directive_token1] = ACTIONS(1615),
    [aux_sym_tag_token1] = ACTIONS(1617),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1079)] = {
    [sym_attribute] = STATE(1079),
    [sym__lowercase_identifier] = ACTIONS(1713),
    [anon_sym_SEMI_SEMI] = ACTIONS(1715),
    [anon_sym_let] = ACTIONS(1713),
    [anon_sym_and] = ACTIONS(1713),
    [anon_sym_TILDE] = ACTIONS(1713),
    [anon_sym_QMARK] = ACTIONS(1713),
    [anon_sym_LPAREN] = ACTIONS(1715),
    [anon_sym_external] = ACTIONS(1713),
    [anon_sym_type] = ACTIONS(1713),
    [anon_sym_COMMA] = ACTIONS(1715),
    [anon_sym_PIPE] = ACTIONS(1713),
    [anon_sym_DOT] = ACTIONS(1715),
    [anon_sym_LBRACK] = ACTIONS(1713),
    [anon_sym_RBRACK] = ACTIONS(1715),
    [anon_sym_true] = ACTIONS(1713),
    [anon_sym_false] = ACTIONS(1713),
    [anon_sym_LBRACE] = ACTIONS(1713),
    [anon_sym_SEMI] = ACTIONS(1713),
    [anon_sym_exception] = ACTIONS(1713),
    [anon_sym_module] = ACTIONS(1713),
    [anon_sym_open] = ACTIONS(1713),
    [anon_sym_include] = ACTIONS(1713),
    [anon_sym_class] = ACTIONS(1713),
    [anon_sym_val] = ACTIONS(1713),
    [anon_sym_end] = ACTIONS(1713),
    [anon_sym_object] = ACTIONS(1713),
    [anon_sym_AMP] = ACTIONS(1713),
    [anon_sym_POUND] = ACTIONS(1713),
    [anon_sym_COLON_COLON] = ACTIONS(1715),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1715),
    [anon_sym_else] = ACTIONS(1713),
    [anon_sym_new] = ACTIONS(1713),
    [anon_sym_LBRACE_LT] = ACTIONS(1715),
    [anon_sym_begin] = ACTIONS(1713),
    [sym_ocamlyacc_value] = ACTIONS(1715),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1713),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1715),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1713),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1713),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1715),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1715),
    [aux_sym_number_token1] = ACTIONS(1715),
    [aux_sym_signed_number_token1] = ACTIONS(1713),
    [anon_sym_SQUOTE] = ACTIONS(1715),
    [anon_sym_DQUOTE] = ACTIONS(1715),
    [sym_prefix_operator] = ACTIONS(1715),
    [aux_sym_sign_operator_token1] = ACTIONS(1713),
    [sym_hash_operator] = ACTIONS(1715),
    [anon_sym_lsl] = ACTIONS(1713),
    [anon_sym_lsr] = ACTIONS(1713),
    [anon_sym_asr] = ACTIONS(1713),
    [aux_sym_pow_operator_token1] = ACTIONS(1715),
    [anon_sym_mod] = ACTIONS(1713),
    [anon_sym_land] = ACTIONS(1713),
    [anon_sym_lor] = ACTIONS(1713),
    [anon_sym_lxor] = ACTIONS(1713),
    [aux_sym_mult_operator_token1] = ACTIONS(1713),
    [aux_sym_add_operator_token1] = ACTIONS(1713),
    [sym_concat_operator] = ACTIONS(1715),
    [sym_rel_operator] = ACTIONS(1713),
    [anon_sym_AMP_AMP] = ACTIONS(1713),
    [anon_sym_or] = ACTIONS(1713),
    [anon_sym_PIPE_PIPE] = ACTIONS(1713),
    [sym_assign_operator] = ACTIONS(1715),
    [sym_let_operator] = ACTIONS(1715),
    [sym__uppercase_identifier] = ACTIONS(1715),
    [aux_sym_directive_token1] = ACTIONS(1713),
    [aux_sym_tag_token1] = ACTIONS(1715),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1080)] = {
    [sym_attribute] = STATE(1080),
    [sym__lowercase_identifier] = ACTIONS(1967),
    [anon_sym_SEMI_SEMI] = ACTIONS(1971),
    [anon_sym_let] = ACTIONS(1967),
    [anon_sym_and] = ACTIONS(1967),
    [anon_sym_TILDE] = ACTIONS(1967),
    [anon_sym_QMARK] = ACTIONS(1967),
    [anon_sym_COLON] = ACTIONS(2206),
    [anon_sym_LPAREN] = ACTIONS(1971),
    [anon_sym_external] = ACTIONS(1967),
    [anon_sym_type] = ACTIONS(1967),
    [anon_sym_COMMA] = ACTIONS(1971),
    [anon_sym_PIPE] = ACTIONS(1967),
    [anon_sym_LBRACK] = ACTIONS(1967),
    [anon_sym_RBRACK] = ACTIONS(1971),
    [anon_sym_true] = ACTIONS(1967),
    [anon_sym_false] = ACTIONS(1967),
    [anon_sym_LBRACE] = ACTIONS(1967),
    [anon_sym_SEMI] = ACTIONS(1967),
    [anon_sym_exception] = ACTIONS(1967),
    [anon_sym_module] = ACTIONS(1967),
    [anon_sym_open] = ACTIONS(1967),
    [anon_sym_include] = ACTIONS(1967),
    [anon_sym_class] = ACTIONS(1967),
    [anon_sym_val] = ACTIONS(1967),
    [anon_sym_end] = ACTIONS(1967),
    [anon_sym_object] = ACTIONS(1967),
    [anon_sym_in] = ACTIONS(1967),
    [anon_sym_AMP] = ACTIONS(1967),
    [anon_sym_COLON_COLON] = ACTIONS(1971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1971),
    [anon_sym_else] = ACTIONS(1967),
    [anon_sym_new] = ACTIONS(1967),
    [anon_sym_LBRACE_LT] = ACTIONS(1971),
    [anon_sym_begin] = ACTIONS(1967),
    [sym_ocamlyacc_value] = ACTIONS(1971),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1967),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1971),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1967),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1967),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1971),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1971),
    [aux_sym_number_token1] = ACTIONS(1971),
    [aux_sym_signed_number_token1] = ACTIONS(1967),
    [anon_sym_SQUOTE] = ACTIONS(1971),
    [anon_sym_DQUOTE] = ACTIONS(1971),
    [sym_prefix_operator] = ACTIONS(1971),
    [aux_sym_sign_operator_token1] = ACTIONS(1967),
    [anon_sym_lsl] = ACTIONS(1967),
    [anon_sym_lsr] = ACTIONS(1967),
    [anon_sym_asr] = ACTIONS(1967),
    [aux_sym_pow_operator_token1] = ACTIONS(1971),
    [anon_sym_mod] = ACTIONS(1967),
    [anon_sym_land] = ACTIONS(1967),
    [anon_sym_lor] = ACTIONS(1967),
    [anon_sym_lxor] = ACTIONS(1967),
    [aux_sym_mult_operator_token1] = ACTIONS(1967),
    [aux_sym_add_operator_token1] = ACTIONS(1967),
    [sym_concat_operator] = ACTIONS(1971),
    [sym_rel_operator] = ACTIONS(1967),
    [anon_sym_AMP_AMP] = ACTIONS(1967),
    [anon_sym_or] = ACTIONS(1967),
    [anon_sym_PIPE_PIPE] = ACTIONS(1967),
    [sym_assign_operator] = ACTIONS(1971),
    [sym_let_operator] = ACTIONS(1971),
    [sym_let_and_operator] = ACTIONS(1971),
    [sym__uppercase_identifier] = ACTIONS(1971),
    [aux_sym_directive_token1] = ACTIONS(1971),
    [aux_sym_tag_token1] = ACTIONS(1971),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1081)] = {
    [sym_attribute] = STATE(1081),
    [sym__lowercase_identifier] = ACTIONS(1801),
    [anon_sym_SEMI_SEMI] = ACTIONS(1803),
    [anon_sym_let] = ACTIONS(1801),
    [anon_sym_and] = ACTIONS(1801),
    [anon_sym_TILDE] = ACTIONS(1801),
    [anon_sym_QMARK] = ACTIONS(1801),
    [anon_sym_LPAREN] = ACTIONS(1803),
    [anon_sym_external] = ACTIONS(1801),
    [anon_sym_type] = ACTIONS(1801),
    [anon_sym_COMMA] = ACTIONS(1803),
    [anon_sym_PIPE] = ACTIONS(1801),
    [anon_sym_DOT] = ACTIONS(1803),
    [anon_sym_LBRACK] = ACTIONS(1801),
    [anon_sym_RBRACK] = ACTIONS(1803),
    [anon_sym_true] = ACTIONS(1801),
    [anon_sym_false] = ACTIONS(1801),
    [anon_sym_LBRACE] = ACTIONS(1801),
    [anon_sym_SEMI] = ACTIONS(1801),
    [anon_sym_exception] = ACTIONS(1801),
    [anon_sym_module] = ACTIONS(1801),
    [anon_sym_open] = ACTIONS(1801),
    [anon_sym_include] = ACTIONS(1801),
    [anon_sym_class] = ACTIONS(1801),
    [anon_sym_val] = ACTIONS(1801),
    [anon_sym_end] = ACTIONS(1801),
    [anon_sym_object] = ACTIONS(1801),
    [anon_sym_AMP] = ACTIONS(1801),
    [anon_sym_POUND] = ACTIONS(1801),
    [anon_sym_COLON_COLON] = ACTIONS(1803),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1803),
    [anon_sym_else] = ACTIONS(1801),
    [anon_sym_new] = ACTIONS(1801),
    [anon_sym_LBRACE_LT] = ACTIONS(1803),
    [anon_sym_begin] = ACTIONS(1801),
    [sym_ocamlyacc_value] = ACTIONS(1803),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1801),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1803),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1801),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1801),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1803),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1803),
    [aux_sym_number_token1] = ACTIONS(1803),
    [aux_sym_signed_number_token1] = ACTIONS(1801),
    [anon_sym_SQUOTE] = ACTIONS(1803),
    [anon_sym_DQUOTE] = ACTIONS(1803),
    [sym_prefix_operator] = ACTIONS(1803),
    [aux_sym_sign_operator_token1] = ACTIONS(1801),
    [sym_hash_operator] = ACTIONS(1803),
    [anon_sym_lsl] = ACTIONS(1801),
    [anon_sym_lsr] = ACTIONS(1801),
    [anon_sym_asr] = ACTIONS(1801),
    [aux_sym_pow_operator_token1] = ACTIONS(1803),
    [anon_sym_mod] = ACTIONS(1801),
    [anon_sym_land] = ACTIONS(1801),
    [anon_sym_lor] = ACTIONS(1801),
    [anon_sym_lxor] = ACTIONS(1801),
    [aux_sym_mult_operator_token1] = ACTIONS(1801),
    [aux_sym_add_operator_token1] = ACTIONS(1801),
    [sym_concat_operator] = ACTIONS(1803),
    [sym_rel_operator] = ACTIONS(1801),
    [anon_sym_AMP_AMP] = ACTIONS(1801),
    [anon_sym_or] = ACTIONS(1801),
    [anon_sym_PIPE_PIPE] = ACTIONS(1801),
    [sym_assign_operator] = ACTIONS(1803),
    [sym_let_operator] = ACTIONS(1803),
    [sym__uppercase_identifier] = ACTIONS(1803),
    [aux_sym_directive_token1] = ACTIONS(1801),
    [aux_sym_tag_token1] = ACTIONS(1803),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1082)] = {
    [sym_attribute] = STATE(1082),
    [sym__lowercase_identifier] = ACTIONS(1729),
    [anon_sym_SEMI_SEMI] = ACTIONS(1731),
    [anon_sym_let] = ACTIONS(1729),
    [anon_sym_and] = ACTIONS(1729),
    [anon_sym_TILDE] = ACTIONS(1729),
    [anon_sym_QMARK] = ACTIONS(1729),
    [anon_sym_LPAREN] = ACTIONS(1731),
    [anon_sym_external] = ACTIONS(1729),
    [anon_sym_type] = ACTIONS(1729),
    [anon_sym_COMMA] = ACTIONS(1731),
    [anon_sym_PIPE] = ACTIONS(1729),
    [anon_sym_DOT] = ACTIONS(1731),
    [anon_sym_LBRACK] = ACTIONS(1729),
    [anon_sym_RBRACK] = ACTIONS(1731),
    [anon_sym_true] = ACTIONS(1729),
    [anon_sym_false] = ACTIONS(1729),
    [anon_sym_LBRACE] = ACTIONS(1729),
    [anon_sym_SEMI] = ACTIONS(1729),
    [anon_sym_exception] = ACTIONS(1729),
    [anon_sym_module] = ACTIONS(1729),
    [anon_sym_open] = ACTIONS(1729),
    [anon_sym_include] = ACTIONS(1729),
    [anon_sym_class] = ACTIONS(1729),
    [anon_sym_val] = ACTIONS(1729),
    [anon_sym_end] = ACTIONS(1729),
    [anon_sym_object] = ACTIONS(1729),
    [anon_sym_AMP] = ACTIONS(1729),
    [anon_sym_POUND] = ACTIONS(1729),
    [anon_sym_COLON_COLON] = ACTIONS(1731),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1731),
    [anon_sym_else] = ACTIONS(1729),
    [anon_sym_new] = ACTIONS(1729),
    [anon_sym_LBRACE_LT] = ACTIONS(1731),
    [anon_sym_begin] = ACTIONS(1729),
    [sym_ocamlyacc_value] = ACTIONS(1731),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1729),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1731),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1729),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1729),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1731),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1731),
    [aux_sym_number_token1] = ACTIONS(1731),
    [aux_sym_signed_number_token1] = ACTIONS(1729),
    [anon_sym_SQUOTE] = ACTIONS(1731),
    [anon_sym_DQUOTE] = ACTIONS(1731),
    [sym_prefix_operator] = ACTIONS(1731),
    [aux_sym_sign_operator_token1] = ACTIONS(1729),
    [sym_hash_operator] = ACTIONS(1731),
    [anon_sym_lsl] = ACTIONS(1729),
    [anon_sym_lsr] = ACTIONS(1729),
    [anon_sym_asr] = ACTIONS(1729),
    [aux_sym_pow_operator_token1] = ACTIONS(1731),
    [anon_sym_mod] = ACTIONS(1729),
    [anon_sym_land] = ACTIONS(1729),
    [anon_sym_lor] = ACTIONS(1729),
    [anon_sym_lxor] = ACTIONS(1729),
    [aux_sym_mult_operator_token1] = ACTIONS(1729),
    [aux_sym_add_operator_token1] = ACTIONS(1729),
    [sym_concat_operator] = ACTIONS(1731),
    [sym_rel_operator] = ACTIONS(1729),
    [anon_sym_AMP_AMP] = ACTIONS(1729),
    [anon_sym_or] = ACTIONS(1729),
    [anon_sym_PIPE_PIPE] = ACTIONS(1729),
    [sym_assign_operator] = ACTIONS(1731),
    [sym_let_operator] = ACTIONS(1731),
    [sym__uppercase_identifier] = ACTIONS(1731),
    [aux_sym_directive_token1] = ACTIONS(1729),
    [aux_sym_tag_token1] = ACTIONS(1731),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1083)] = {
    [sym_attribute] = STATE(1083),
    [sym__lowercase_identifier] = ACTIONS(1659),
    [anon_sym_SEMI_SEMI] = ACTIONS(1661),
    [anon_sym_let] = ACTIONS(1659),
    [anon_sym_and] = ACTIONS(1659),
    [anon_sym_TILDE] = ACTIONS(1659),
    [anon_sym_QMARK] = ACTIONS(1659),
    [anon_sym_LPAREN] = ACTIONS(1661),
    [anon_sym_external] = ACTIONS(1659),
    [anon_sym_type] = ACTIONS(1659),
    [anon_sym_COMMA] = ACTIONS(1661),
    [anon_sym_PIPE] = ACTIONS(1659),
    [anon_sym_DOT] = ACTIONS(1661),
    [anon_sym_LBRACK] = ACTIONS(1659),
    [anon_sym_RBRACK] = ACTIONS(1661),
    [anon_sym_true] = ACTIONS(1659),
    [anon_sym_false] = ACTIONS(1659),
    [anon_sym_LBRACE] = ACTIONS(1659),
    [anon_sym_SEMI] = ACTIONS(1659),
    [anon_sym_exception] = ACTIONS(1659),
    [anon_sym_module] = ACTIONS(1659),
    [anon_sym_open] = ACTIONS(1659),
    [anon_sym_include] = ACTIONS(1659),
    [anon_sym_class] = ACTIONS(1659),
    [anon_sym_val] = ACTIONS(1659),
    [anon_sym_end] = ACTIONS(1659),
    [anon_sym_object] = ACTIONS(1659),
    [anon_sym_AMP] = ACTIONS(1659),
    [anon_sym_POUND] = ACTIONS(1659),
    [anon_sym_COLON_COLON] = ACTIONS(1661),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1661),
    [anon_sym_else] = ACTIONS(1659),
    [anon_sym_new] = ACTIONS(1659),
    [anon_sym_LBRACE_LT] = ACTIONS(1661),
    [anon_sym_begin] = ACTIONS(1659),
    [sym_ocamlyacc_value] = ACTIONS(1661),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1659),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1661),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1659),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1659),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1661),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1661),
    [aux_sym_number_token1] = ACTIONS(1661),
    [aux_sym_signed_number_token1] = ACTIONS(1659),
    [anon_sym_SQUOTE] = ACTIONS(1661),
    [anon_sym_DQUOTE] = ACTIONS(1661),
    [sym_prefix_operator] = ACTIONS(1661),
    [aux_sym_sign_operator_token1] = ACTIONS(1659),
    [sym_hash_operator] = ACTIONS(1661),
    [anon_sym_lsl] = ACTIONS(1659),
    [anon_sym_lsr] = ACTIONS(1659),
    [anon_sym_asr] = ACTIONS(1659),
    [aux_sym_pow_operator_token1] = ACTIONS(1661),
    [anon_sym_mod] = ACTIONS(1659),
    [anon_sym_land] = ACTIONS(1659),
    [anon_sym_lor] = ACTIONS(1659),
    [anon_sym_lxor] = ACTIONS(1659),
    [aux_sym_mult_operator_token1] = ACTIONS(1659),
    [aux_sym_add_operator_token1] = ACTIONS(1659),
    [sym_concat_operator] = ACTIONS(1661),
    [sym_rel_operator] = ACTIONS(1659),
    [anon_sym_AMP_AMP] = ACTIONS(1659),
    [anon_sym_or] = ACTIONS(1659),
    [anon_sym_PIPE_PIPE] = ACTIONS(1659),
    [sym_assign_operator] = ACTIONS(1661),
    [sym_let_operator] = ACTIONS(1661),
    [sym__uppercase_identifier] = ACTIONS(1661),
    [aux_sym_directive_token1] = ACTIONS(1659),
    [aux_sym_tag_token1] = ACTIONS(1661),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1084)] = {
    [sym_attribute] = STATE(1084),
    [sym__lowercase_identifier] = ACTIONS(1663),
    [anon_sym_SEMI_SEMI] = ACTIONS(1665),
    [anon_sym_let] = ACTIONS(1663),
    [anon_sym_and] = ACTIONS(1663),
    [anon_sym_TILDE] = ACTIONS(1663),
    [anon_sym_QMARK] = ACTIONS(1663),
    [anon_sym_LPAREN] = ACTIONS(1665),
    [anon_sym_external] = ACTIONS(1663),
    [anon_sym_type] = ACTIONS(1663),
    [anon_sym_COMMA] = ACTIONS(1665),
    [anon_sym_PIPE] = ACTIONS(1663),
    [anon_sym_DOT] = ACTIONS(969),
    [anon_sym_LBRACK] = ACTIONS(1663),
    [anon_sym_RBRACK] = ACTIONS(1665),
    [anon_sym_true] = ACTIONS(1663),
    [anon_sym_false] = ACTIONS(1663),
    [anon_sym_LBRACE] = ACTIONS(1663),
    [anon_sym_SEMI] = ACTIONS(1663),
    [anon_sym_exception] = ACTIONS(1663),
    [anon_sym_module] = ACTIONS(1663),
    [anon_sym_open] = ACTIONS(1663),
    [anon_sym_include] = ACTIONS(1663),
    [anon_sym_class] = ACTIONS(1663),
    [anon_sym_val] = ACTIONS(1663),
    [anon_sym_end] = ACTIONS(1663),
    [anon_sym_object] = ACTIONS(1663),
    [anon_sym_AMP] = ACTIONS(1663),
    [anon_sym_POUND] = ACTIONS(971),
    [anon_sym_COLON_COLON] = ACTIONS(1665),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1665),
    [anon_sym_else] = ACTIONS(1663),
    [anon_sym_new] = ACTIONS(1663),
    [anon_sym_LBRACE_LT] = ACTIONS(1665),
    [anon_sym_begin] = ACTIONS(1663),
    [sym_ocamlyacc_value] = ACTIONS(1665),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1663),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1665),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1663),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1663),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1665),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1665),
    [aux_sym_number_token1] = ACTIONS(1665),
    [aux_sym_signed_number_token1] = ACTIONS(1663),
    [anon_sym_SQUOTE] = ACTIONS(1665),
    [anon_sym_DQUOTE] = ACTIONS(1665),
    [sym_prefix_operator] = ACTIONS(1665),
    [aux_sym_sign_operator_token1] = ACTIONS(1663),
    [sym_hash_operator] = ACTIONS(973),
    [anon_sym_lsl] = ACTIONS(1663),
    [anon_sym_lsr] = ACTIONS(1663),
    [anon_sym_asr] = ACTIONS(1663),
    [aux_sym_pow_operator_token1] = ACTIONS(1665),
    [anon_sym_mod] = ACTIONS(1663),
    [anon_sym_land] = ACTIONS(1663),
    [anon_sym_lor] = ACTIONS(1663),
    [anon_sym_lxor] = ACTIONS(1663),
    [aux_sym_mult_operator_token1] = ACTIONS(1663),
    [aux_sym_add_operator_token1] = ACTIONS(1663),
    [sym_concat_operator] = ACTIONS(1665),
    [sym_rel_operator] = ACTIONS(1663),
    [anon_sym_AMP_AMP] = ACTIONS(1663),
    [anon_sym_or] = ACTIONS(1663),
    [anon_sym_PIPE_PIPE] = ACTIONS(1663),
    [sym_assign_operator] = ACTIONS(1665),
    [sym_let_operator] = ACTIONS(1665),
    [sym__uppercase_identifier] = ACTIONS(1665),
    [aux_sym_directive_token1] = ACTIONS(1663),
    [aux_sym_tag_token1] = ACTIONS(1665),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1085)] = {
    [sym_attribute] = STATE(1085),
    [sym__lowercase_identifier] = ACTIONS(1695),
    [anon_sym_SEMI_SEMI] = ACTIONS(1697),
    [anon_sym_let] = ACTIONS(1695),
    [anon_sym_and] = ACTIONS(1695),
    [anon_sym_TILDE] = ACTIONS(1695),
    [anon_sym_QMARK] = ACTIONS(1695),
    [anon_sym_LPAREN] = ACTIONS(1697),
    [anon_sym_external] = ACTIONS(1695),
    [anon_sym_type] = ACTIONS(1695),
    [anon_sym_COMMA] = ACTIONS(1697),
    [anon_sym_PIPE] = ACTIONS(1695),
    [anon_sym_DOT] = ACTIONS(1697),
    [anon_sym_LBRACK] = ACTIONS(1695),
    [anon_sym_RBRACK] = ACTIONS(1697),
    [anon_sym_true] = ACTIONS(1695),
    [anon_sym_false] = ACTIONS(1695),
    [anon_sym_LBRACE] = ACTIONS(1695),
    [anon_sym_SEMI] = ACTIONS(1695),
    [anon_sym_exception] = ACTIONS(1695),
    [anon_sym_module] = ACTIONS(1695),
    [anon_sym_open] = ACTIONS(1695),
    [anon_sym_include] = ACTIONS(1695),
    [anon_sym_class] = ACTIONS(1695),
    [anon_sym_val] = ACTIONS(1695),
    [anon_sym_end] = ACTIONS(1695),
    [anon_sym_object] = ACTIONS(1695),
    [anon_sym_AMP] = ACTIONS(1695),
    [anon_sym_POUND] = ACTIONS(1695),
    [anon_sym_COLON_COLON] = ACTIONS(1697),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1697),
    [anon_sym_else] = ACTIONS(1695),
    [anon_sym_new] = ACTIONS(1695),
    [anon_sym_LBRACE_LT] = ACTIONS(1697),
    [anon_sym_begin] = ACTIONS(1695),
    [sym_ocamlyacc_value] = ACTIONS(1697),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1695),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1697),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1695),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1695),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1697),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1697),
    [aux_sym_number_token1] = ACTIONS(1697),
    [aux_sym_signed_number_token1] = ACTIONS(1695),
    [anon_sym_SQUOTE] = ACTIONS(1697),
    [anon_sym_DQUOTE] = ACTIONS(1697),
    [sym_prefix_operator] = ACTIONS(1697),
    [aux_sym_sign_operator_token1] = ACTIONS(1695),
    [sym_hash_operator] = ACTIONS(1697),
    [anon_sym_lsl] = ACTIONS(1695),
    [anon_sym_lsr] = ACTIONS(1695),
    [anon_sym_asr] = ACTIONS(1695),
    [aux_sym_pow_operator_token1] = ACTIONS(1697),
    [anon_sym_mod] = ACTIONS(1695),
    [anon_sym_land] = ACTIONS(1695),
    [anon_sym_lor] = ACTIONS(1695),
    [anon_sym_lxor] = ACTIONS(1695),
    [aux_sym_mult_operator_token1] = ACTIONS(1695),
    [aux_sym_add_operator_token1] = ACTIONS(1695),
    [sym_concat_operator] = ACTIONS(1697),
    [sym_rel_operator] = ACTIONS(1695),
    [anon_sym_AMP_AMP] = ACTIONS(1695),
    [anon_sym_or] = ACTIONS(1695),
    [anon_sym_PIPE_PIPE] = ACTIONS(1695),
    [sym_assign_operator] = ACTIONS(1697),
    [sym_let_operator] = ACTIONS(1697),
    [sym__uppercase_identifier] = ACTIONS(1697),
    [aux_sym_directive_token1] = ACTIONS(1695),
    [aux_sym_tag_token1] = ACTIONS(1697),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1086)] = {
    [sym_attribute] = STATE(1086),
    [sym__lowercase_identifier] = ACTIONS(1683),
    [anon_sym_SEMI_SEMI] = ACTIONS(1685),
    [anon_sym_let] = ACTIONS(1683),
    [anon_sym_and] = ACTIONS(1683),
    [anon_sym_TILDE] = ACTIONS(1683),
    [anon_sym_QMARK] = ACTIONS(1683),
    [anon_sym_LPAREN] = ACTIONS(1685),
    [anon_sym_external] = ACTIONS(1683),
    [anon_sym_type] = ACTIONS(1683),
    [anon_sym_COMMA] = ACTIONS(1685),
    [anon_sym_PIPE] = ACTIONS(1683),
    [anon_sym_DOT] = ACTIONS(1685),
    [anon_sym_LBRACK] = ACTIONS(1683),
    [anon_sym_RBRACK] = ACTIONS(1685),
    [anon_sym_true] = ACTIONS(1683),
    [anon_sym_false] = ACTIONS(1683),
    [anon_sym_LBRACE] = ACTIONS(1683),
    [anon_sym_SEMI] = ACTIONS(1683),
    [anon_sym_exception] = ACTIONS(1683),
    [anon_sym_module] = ACTIONS(1683),
    [anon_sym_open] = ACTIONS(1683),
    [anon_sym_include] = ACTIONS(1683),
    [anon_sym_class] = ACTIONS(1683),
    [anon_sym_val] = ACTIONS(1683),
    [anon_sym_end] = ACTIONS(1683),
    [anon_sym_object] = ACTIONS(1683),
    [anon_sym_AMP] = ACTIONS(1683),
    [anon_sym_POUND] = ACTIONS(1683),
    [anon_sym_COLON_COLON] = ACTIONS(1685),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1685),
    [anon_sym_else] = ACTIONS(1683),
    [anon_sym_new] = ACTIONS(1683),
    [anon_sym_LBRACE_LT] = ACTIONS(1685),
    [anon_sym_begin] = ACTIONS(1683),
    [sym_ocamlyacc_value] = ACTIONS(1685),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1683),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1685),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1683),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1683),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1685),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1685),
    [aux_sym_number_token1] = ACTIONS(1685),
    [aux_sym_signed_number_token1] = ACTIONS(1683),
    [anon_sym_SQUOTE] = ACTIONS(1685),
    [anon_sym_DQUOTE] = ACTIONS(1685),
    [sym_prefix_operator] = ACTIONS(1685),
    [aux_sym_sign_operator_token1] = ACTIONS(1683),
    [sym_hash_operator] = ACTIONS(1685),
    [anon_sym_lsl] = ACTIONS(1683),
    [anon_sym_lsr] = ACTIONS(1683),
    [anon_sym_asr] = ACTIONS(1683),
    [aux_sym_pow_operator_token1] = ACTIONS(1685),
    [anon_sym_mod] = ACTIONS(1683),
    [anon_sym_land] = ACTIONS(1683),
    [anon_sym_lor] = ACTIONS(1683),
    [anon_sym_lxor] = ACTIONS(1683),
    [aux_sym_mult_operator_token1] = ACTIONS(1683),
    [aux_sym_add_operator_token1] = ACTIONS(1683),
    [sym_concat_operator] = ACTIONS(1685),
    [sym_rel_operator] = ACTIONS(1683),
    [anon_sym_AMP_AMP] = ACTIONS(1683),
    [anon_sym_or] = ACTIONS(1683),
    [anon_sym_PIPE_PIPE] = ACTIONS(1683),
    [sym_assign_operator] = ACTIONS(1685),
    [sym_let_operator] = ACTIONS(1685),
    [sym__uppercase_identifier] = ACTIONS(1685),
    [aux_sym_directive_token1] = ACTIONS(1683),
    [aux_sym_tag_token1] = ACTIONS(1685),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1087)] = {
    [sym_attribute] = STATE(1087),
    [sym__lowercase_identifier] = ACTIONS(1607),
    [anon_sym_SEMI_SEMI] = ACTIONS(1609),
    [anon_sym_let] = ACTIONS(1607),
    [anon_sym_and] = ACTIONS(1607),
    [anon_sym_TILDE] = ACTIONS(1607),
    [anon_sym_QMARK] = ACTIONS(1607),
    [anon_sym_LPAREN] = ACTIONS(1609),
    [anon_sym_external] = ACTIONS(1607),
    [anon_sym_type] = ACTIONS(1607),
    [anon_sym_COMMA] = ACTIONS(1609),
    [anon_sym_PIPE] = ACTIONS(1607),
    [anon_sym_DOT] = ACTIONS(1609),
    [anon_sym_LBRACK] = ACTIONS(1607),
    [anon_sym_RBRACK] = ACTIONS(1609),
    [anon_sym_true] = ACTIONS(1607),
    [anon_sym_false] = ACTIONS(1607),
    [anon_sym_LBRACE] = ACTIONS(1607),
    [anon_sym_SEMI] = ACTIONS(1607),
    [anon_sym_exception] = ACTIONS(1607),
    [anon_sym_module] = ACTIONS(1607),
    [anon_sym_open] = ACTIONS(1607),
    [anon_sym_include] = ACTIONS(1607),
    [anon_sym_class] = ACTIONS(1607),
    [anon_sym_val] = ACTIONS(1607),
    [anon_sym_end] = ACTIONS(1607),
    [anon_sym_object] = ACTIONS(1607),
    [anon_sym_AMP] = ACTIONS(1607),
    [anon_sym_POUND] = ACTIONS(1607),
    [anon_sym_COLON_COLON] = ACTIONS(1609),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1609),
    [anon_sym_else] = ACTIONS(1607),
    [anon_sym_new] = ACTIONS(1607),
    [anon_sym_LBRACE_LT] = ACTIONS(1609),
    [anon_sym_begin] = ACTIONS(1607),
    [sym_ocamlyacc_value] = ACTIONS(1609),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1607),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1609),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1607),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1607),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1609),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1609),
    [aux_sym_number_token1] = ACTIONS(1609),
    [aux_sym_signed_number_token1] = ACTIONS(1607),
    [anon_sym_SQUOTE] = ACTIONS(1609),
    [anon_sym_DQUOTE] = ACTIONS(1609),
    [sym_prefix_operator] = ACTIONS(1609),
    [aux_sym_sign_operator_token1] = ACTIONS(1607),
    [sym_hash_operator] = ACTIONS(1609),
    [anon_sym_lsl] = ACTIONS(1607),
    [anon_sym_lsr] = ACTIONS(1607),
    [anon_sym_asr] = ACTIONS(1607),
    [aux_sym_pow_operator_token1] = ACTIONS(1609),
    [anon_sym_mod] = ACTIONS(1607),
    [anon_sym_land] = ACTIONS(1607),
    [anon_sym_lor] = ACTIONS(1607),
    [anon_sym_lxor] = ACTIONS(1607),
    [aux_sym_mult_operator_token1] = ACTIONS(1607),
    [aux_sym_add_operator_token1] = ACTIONS(1607),
    [sym_concat_operator] = ACTIONS(1609),
    [sym_rel_operator] = ACTIONS(1607),
    [anon_sym_AMP_AMP] = ACTIONS(1607),
    [anon_sym_or] = ACTIONS(1607),
    [anon_sym_PIPE_PIPE] = ACTIONS(1607),
    [sym_assign_operator] = ACTIONS(1609),
    [sym_let_operator] = ACTIONS(1609),
    [sym__uppercase_identifier] = ACTIONS(1609),
    [aux_sym_directive_token1] = ACTIONS(1607),
    [aux_sym_tag_token1] = ACTIONS(1609),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1088)] = {
    [sym__parenthesized_abstract_type] = STATE(1308),
    [sym__delimited_pattern] = STATE(4970),
    [sym__simple_pattern] = STATE(4826),
    [sym__effect_pattern] = STATE(4830),
    [sym__pattern] = STATE(4878),
    [sym_alias_pattern] = STATE(4830),
    [sym_typed_pattern] = STATE(4970),
    [sym__or_pattern_anonymous] = STATE(4832),
    [sym_constructor_pattern] = STATE(4826),
    [sym_tag_pattern] = STATE(4826),
    [sym_polymorphic_variant_pattern] = STATE(4970),
    [sym_labeled_tuple_element_pattern] = STATE(7167),
    [sym__tuple_pattern] = STATE(4833),
    [sym_record_pattern] = STATE(4978),
    [sym_list_pattern] = STATE(4978),
    [sym_cons_pattern] = STATE(4830),
    [sym_array_pattern] = STATE(4978),
    [sym_range_pattern] = STATE(4830),
    [sym_lazy_pattern] = STATE(4826),
    [sym_local_open_pattern] = STATE(4970),
    [sym_package_pattern] = STATE(4970),
    [sym_parenthesized_pattern] = STATE(4978),
    [sym_exception_pattern] = STATE(4830),
    [sym_effect_pattern] = STATE(4830),
    [sym_attribute] = STATE(1088),
    [sym_extension] = STATE(4970),
    [sym_quoted_extension] = STATE(4970),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4725),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4978),
    [sym__simple_value_pattern] = STATE(4970),
    [sym_parenthesized_operator] = STATE(4970),
    [sym_module_path] = STATE(6366),
    [sym_constructor_path] = STATE(1088),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1127),
    [sym__lowercase_identifier] = ACTIONS(2208),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(2210),
    [anon_sym_COMMA] = ACTIONS(2075),
    [anon_sym_PIPE] = ACTIONS(2081),
    [anon_sym_LBRACK] = ACTIONS(2212),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2216),
    [anon_sym_SEMI] = ACTIONS(2075),
    [anon_sym_exception] = ACTIONS(2218),
    [anon_sym_as] = ACTIONS(2081),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_COLON_COLON] = ACTIONS(2075),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2222),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2075),
    [anon_sym_lazy] = ACTIONS(2224),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2228),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1089)] = {
    [sym_attribute] = STATE(1089),
    [sym__lowercase_identifier] = ACTIONS(1765),
    [anon_sym_SEMI_SEMI] = ACTIONS(1767),
    [anon_sym_let] = ACTIONS(1765),
    [anon_sym_and] = ACTIONS(1765),
    [anon_sym_TILDE] = ACTIONS(1765),
    [anon_sym_QMARK] = ACTIONS(1765),
    [anon_sym_LPAREN] = ACTIONS(1767),
    [anon_sym_external] = ACTIONS(1765),
    [anon_sym_type] = ACTIONS(1765),
    [anon_sym_COMMA] = ACTIONS(1767),
    [anon_sym_PIPE] = ACTIONS(1765),
    [anon_sym_DOT] = ACTIONS(1767),
    [anon_sym_LBRACK] = ACTIONS(1765),
    [anon_sym_RBRACK] = ACTIONS(1767),
    [anon_sym_true] = ACTIONS(1765),
    [anon_sym_false] = ACTIONS(1765),
    [anon_sym_LBRACE] = ACTIONS(1765),
    [anon_sym_SEMI] = ACTIONS(1765),
    [anon_sym_exception] = ACTIONS(1765),
    [anon_sym_module] = ACTIONS(1765),
    [anon_sym_open] = ACTIONS(1765),
    [anon_sym_include] = ACTIONS(1765),
    [anon_sym_class] = ACTIONS(1765),
    [anon_sym_val] = ACTIONS(1765),
    [anon_sym_end] = ACTIONS(1765),
    [anon_sym_object] = ACTIONS(1765),
    [anon_sym_AMP] = ACTIONS(1765),
    [anon_sym_POUND] = ACTIONS(1765),
    [anon_sym_COLON_COLON] = ACTIONS(1767),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1767),
    [anon_sym_else] = ACTIONS(1765),
    [anon_sym_new] = ACTIONS(1765),
    [anon_sym_LBRACE_LT] = ACTIONS(1767),
    [anon_sym_begin] = ACTIONS(1765),
    [sym_ocamlyacc_value] = ACTIONS(1767),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1765),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1767),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1765),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1765),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1767),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1767),
    [aux_sym_number_token1] = ACTIONS(1767),
    [aux_sym_signed_number_token1] = ACTIONS(1765),
    [anon_sym_SQUOTE] = ACTIONS(1767),
    [anon_sym_DQUOTE] = ACTIONS(1767),
    [sym_prefix_operator] = ACTIONS(1767),
    [aux_sym_sign_operator_token1] = ACTIONS(1765),
    [sym_hash_operator] = ACTIONS(1767),
    [anon_sym_lsl] = ACTIONS(1765),
    [anon_sym_lsr] = ACTIONS(1765),
    [anon_sym_asr] = ACTIONS(1765),
    [aux_sym_pow_operator_token1] = ACTIONS(1767),
    [anon_sym_mod] = ACTIONS(1765),
    [anon_sym_land] = ACTIONS(1765),
    [anon_sym_lor] = ACTIONS(1765),
    [anon_sym_lxor] = ACTIONS(1765),
    [aux_sym_mult_operator_token1] = ACTIONS(1765),
    [aux_sym_add_operator_token1] = ACTIONS(1765),
    [sym_concat_operator] = ACTIONS(1767),
    [sym_rel_operator] = ACTIONS(1765),
    [anon_sym_AMP_AMP] = ACTIONS(1765),
    [anon_sym_or] = ACTIONS(1765),
    [anon_sym_PIPE_PIPE] = ACTIONS(1765),
    [sym_assign_operator] = ACTIONS(1767),
    [sym_let_operator] = ACTIONS(1767),
    [sym__uppercase_identifier] = ACTIONS(1767),
    [aux_sym_directive_token1] = ACTIONS(1765),
    [aux_sym_tag_token1] = ACTIONS(1767),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1090)] = {
    [sym_attribute] = STATE(1090),
    [sym__lowercase_identifier] = ACTIONS(1773),
    [anon_sym_SEMI_SEMI] = ACTIONS(1775),
    [anon_sym_let] = ACTIONS(1773),
    [anon_sym_and] = ACTIONS(1773),
    [anon_sym_TILDE] = ACTIONS(1773),
    [anon_sym_QMARK] = ACTIONS(1773),
    [anon_sym_LPAREN] = ACTIONS(1775),
    [anon_sym_external] = ACTIONS(1773),
    [anon_sym_type] = ACTIONS(1773),
    [anon_sym_COMMA] = ACTIONS(1775),
    [anon_sym_PIPE] = ACTIONS(1773),
    [anon_sym_DOT] = ACTIONS(1775),
    [anon_sym_LBRACK] = ACTIONS(1773),
    [anon_sym_RBRACK] = ACTIONS(1775),
    [anon_sym_true] = ACTIONS(1773),
    [anon_sym_false] = ACTIONS(1773),
    [anon_sym_LBRACE] = ACTIONS(1773),
    [anon_sym_SEMI] = ACTIONS(1773),
    [anon_sym_exception] = ACTIONS(1773),
    [anon_sym_module] = ACTIONS(1773),
    [anon_sym_open] = ACTIONS(1773),
    [anon_sym_include] = ACTIONS(1773),
    [anon_sym_class] = ACTIONS(1773),
    [anon_sym_val] = ACTIONS(1773),
    [anon_sym_end] = ACTIONS(1773),
    [anon_sym_object] = ACTIONS(1773),
    [anon_sym_AMP] = ACTIONS(1773),
    [anon_sym_POUND] = ACTIONS(1773),
    [anon_sym_COLON_COLON] = ACTIONS(1775),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1775),
    [anon_sym_else] = ACTIONS(1773),
    [anon_sym_new] = ACTIONS(1773),
    [anon_sym_LBRACE_LT] = ACTIONS(1775),
    [anon_sym_begin] = ACTIONS(1773),
    [sym_ocamlyacc_value] = ACTIONS(1775),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1773),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1775),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1773),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1773),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1775),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1775),
    [aux_sym_number_token1] = ACTIONS(1775),
    [aux_sym_signed_number_token1] = ACTIONS(1773),
    [anon_sym_SQUOTE] = ACTIONS(1775),
    [anon_sym_DQUOTE] = ACTIONS(1775),
    [sym_prefix_operator] = ACTIONS(1775),
    [aux_sym_sign_operator_token1] = ACTIONS(1773),
    [sym_hash_operator] = ACTIONS(1775),
    [anon_sym_lsl] = ACTIONS(1773),
    [anon_sym_lsr] = ACTIONS(1773),
    [anon_sym_asr] = ACTIONS(1773),
    [aux_sym_pow_operator_token1] = ACTIONS(1775),
    [anon_sym_mod] = ACTIONS(1773),
    [anon_sym_land] = ACTIONS(1773),
    [anon_sym_lor] = ACTIONS(1773),
    [anon_sym_lxor] = ACTIONS(1773),
    [aux_sym_mult_operator_token1] = ACTIONS(1773),
    [aux_sym_add_operator_token1] = ACTIONS(1773),
    [sym_concat_operator] = ACTIONS(1775),
    [sym_rel_operator] = ACTIONS(1773),
    [anon_sym_AMP_AMP] = ACTIONS(1773),
    [anon_sym_or] = ACTIONS(1773),
    [anon_sym_PIPE_PIPE] = ACTIONS(1773),
    [sym_assign_operator] = ACTIONS(1775),
    [sym_let_operator] = ACTIONS(1775),
    [sym__uppercase_identifier] = ACTIONS(1775),
    [aux_sym_directive_token1] = ACTIONS(1773),
    [aux_sym_tag_token1] = ACTIONS(1775),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1091)] = {
    [sym_attribute] = STATE(1091),
    [sym__lowercase_identifier] = ACTIONS(1821),
    [anon_sym_SEMI_SEMI] = ACTIONS(1823),
    [anon_sym_let] = ACTIONS(1821),
    [anon_sym_and] = ACTIONS(1821),
    [anon_sym_TILDE] = ACTIONS(1821),
    [anon_sym_QMARK] = ACTIONS(1821),
    [anon_sym_LPAREN] = ACTIONS(1823),
    [anon_sym_external] = ACTIONS(1821),
    [anon_sym_type] = ACTIONS(1821),
    [anon_sym_COMMA] = ACTIONS(1823),
    [anon_sym_PIPE] = ACTIONS(1821),
    [anon_sym_DOT] = ACTIONS(1823),
    [anon_sym_LBRACK] = ACTIONS(1821),
    [anon_sym_RBRACK] = ACTIONS(1823),
    [anon_sym_true] = ACTIONS(1821),
    [anon_sym_false] = ACTIONS(1821),
    [anon_sym_LBRACE] = ACTIONS(1821),
    [anon_sym_SEMI] = ACTIONS(1821),
    [anon_sym_exception] = ACTIONS(1821),
    [anon_sym_module] = ACTIONS(1821),
    [anon_sym_open] = ACTIONS(1821),
    [anon_sym_include] = ACTIONS(1821),
    [anon_sym_class] = ACTIONS(1821),
    [anon_sym_val] = ACTIONS(1821),
    [anon_sym_end] = ACTIONS(1821),
    [anon_sym_object] = ACTIONS(1821),
    [anon_sym_AMP] = ACTIONS(1821),
    [anon_sym_POUND] = ACTIONS(1821),
    [anon_sym_COLON_COLON] = ACTIONS(1823),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1823),
    [anon_sym_else] = ACTIONS(1821),
    [anon_sym_new] = ACTIONS(1821),
    [anon_sym_LBRACE_LT] = ACTIONS(1823),
    [anon_sym_begin] = ACTIONS(1821),
    [sym_ocamlyacc_value] = ACTIONS(1823),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1821),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1823),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1821),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1821),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1823),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1823),
    [aux_sym_number_token1] = ACTIONS(1823),
    [aux_sym_signed_number_token1] = ACTIONS(1821),
    [anon_sym_SQUOTE] = ACTIONS(1823),
    [anon_sym_DQUOTE] = ACTIONS(1823),
    [sym_prefix_operator] = ACTIONS(1823),
    [aux_sym_sign_operator_token1] = ACTIONS(1821),
    [sym_hash_operator] = ACTIONS(1823),
    [anon_sym_lsl] = ACTIONS(1821),
    [anon_sym_lsr] = ACTIONS(1821),
    [anon_sym_asr] = ACTIONS(1821),
    [aux_sym_pow_operator_token1] = ACTIONS(1823),
    [anon_sym_mod] = ACTIONS(1821),
    [anon_sym_land] = ACTIONS(1821),
    [anon_sym_lor] = ACTIONS(1821),
    [anon_sym_lxor] = ACTIONS(1821),
    [aux_sym_mult_operator_token1] = ACTIONS(1821),
    [aux_sym_add_operator_token1] = ACTIONS(1821),
    [sym_concat_operator] = ACTIONS(1823),
    [sym_rel_operator] = ACTIONS(1821),
    [anon_sym_AMP_AMP] = ACTIONS(1821),
    [anon_sym_or] = ACTIONS(1821),
    [anon_sym_PIPE_PIPE] = ACTIONS(1821),
    [sym_assign_operator] = ACTIONS(1823),
    [sym_let_operator] = ACTIONS(1823),
    [sym__uppercase_identifier] = ACTIONS(1823),
    [aux_sym_directive_token1] = ACTIONS(1821),
    [aux_sym_tag_token1] = ACTIONS(1823),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1092)] = {
    [sym_attribute] = STATE(1092),
    [sym__lowercase_identifier] = ACTIONS(1777),
    [anon_sym_SEMI_SEMI] = ACTIONS(1779),
    [anon_sym_let] = ACTIONS(1777),
    [anon_sym_and] = ACTIONS(1777),
    [anon_sym_TILDE] = ACTIONS(1777),
    [anon_sym_QMARK] = ACTIONS(1777),
    [anon_sym_LPAREN] = ACTIONS(1779),
    [anon_sym_external] = ACTIONS(1777),
    [anon_sym_type] = ACTIONS(1777),
    [anon_sym_COMMA] = ACTIONS(1779),
    [anon_sym_PIPE] = ACTIONS(1777),
    [anon_sym_DOT] = ACTIONS(1779),
    [anon_sym_LBRACK] = ACTIONS(1777),
    [anon_sym_RBRACK] = ACTIONS(1779),
    [anon_sym_true] = ACTIONS(1777),
    [anon_sym_false] = ACTIONS(1777),
    [anon_sym_LBRACE] = ACTIONS(1777),
    [anon_sym_SEMI] = ACTIONS(1777),
    [anon_sym_exception] = ACTIONS(1777),
    [anon_sym_module] = ACTIONS(1777),
    [anon_sym_open] = ACTIONS(1777),
    [anon_sym_include] = ACTIONS(1777),
    [anon_sym_class] = ACTIONS(1777),
    [anon_sym_val] = ACTIONS(1777),
    [anon_sym_end] = ACTIONS(1777),
    [anon_sym_object] = ACTIONS(1777),
    [anon_sym_AMP] = ACTIONS(1777),
    [anon_sym_POUND] = ACTIONS(1777),
    [anon_sym_COLON_COLON] = ACTIONS(1779),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1779),
    [anon_sym_else] = ACTIONS(1777),
    [anon_sym_new] = ACTIONS(1777),
    [anon_sym_LBRACE_LT] = ACTIONS(1779),
    [anon_sym_begin] = ACTIONS(1777),
    [sym_ocamlyacc_value] = ACTIONS(1779),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1777),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1779),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1777),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1777),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1779),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1779),
    [aux_sym_number_token1] = ACTIONS(1779),
    [aux_sym_signed_number_token1] = ACTIONS(1777),
    [anon_sym_SQUOTE] = ACTIONS(1779),
    [anon_sym_DQUOTE] = ACTIONS(1779),
    [sym_prefix_operator] = ACTIONS(1779),
    [aux_sym_sign_operator_token1] = ACTIONS(1777),
    [sym_hash_operator] = ACTIONS(1779),
    [anon_sym_lsl] = ACTIONS(1777),
    [anon_sym_lsr] = ACTIONS(1777),
    [anon_sym_asr] = ACTIONS(1777),
    [aux_sym_pow_operator_token1] = ACTIONS(1779),
    [anon_sym_mod] = ACTIONS(1777),
    [anon_sym_land] = ACTIONS(1777),
    [anon_sym_lor] = ACTIONS(1777),
    [anon_sym_lxor] = ACTIONS(1777),
    [aux_sym_mult_operator_token1] = ACTIONS(1777),
    [aux_sym_add_operator_token1] = ACTIONS(1777),
    [sym_concat_operator] = ACTIONS(1779),
    [sym_rel_operator] = ACTIONS(1777),
    [anon_sym_AMP_AMP] = ACTIONS(1777),
    [anon_sym_or] = ACTIONS(1777),
    [anon_sym_PIPE_PIPE] = ACTIONS(1777),
    [sym_assign_operator] = ACTIONS(1779),
    [sym_let_operator] = ACTIONS(1779),
    [sym__uppercase_identifier] = ACTIONS(1779),
    [aux_sym_directive_token1] = ACTIONS(1777),
    [aux_sym_tag_token1] = ACTIONS(1779),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1093)] = {
    [sym_attribute] = STATE(1093),
    [sym__lowercase_identifier] = ACTIONS(1781),
    [anon_sym_SEMI_SEMI] = ACTIONS(1783),
    [anon_sym_let] = ACTIONS(1781),
    [anon_sym_and] = ACTIONS(1781),
    [anon_sym_TILDE] = ACTIONS(1781),
    [anon_sym_QMARK] = ACTIONS(1781),
    [anon_sym_LPAREN] = ACTIONS(1783),
    [anon_sym_external] = ACTIONS(1781),
    [anon_sym_type] = ACTIONS(1781),
    [anon_sym_COMMA] = ACTIONS(1783),
    [anon_sym_PIPE] = ACTIONS(1781),
    [anon_sym_DOT] = ACTIONS(1783),
    [anon_sym_LBRACK] = ACTIONS(1781),
    [anon_sym_RBRACK] = ACTIONS(1783),
    [anon_sym_true] = ACTIONS(1781),
    [anon_sym_false] = ACTIONS(1781),
    [anon_sym_LBRACE] = ACTIONS(1781),
    [anon_sym_SEMI] = ACTIONS(1781),
    [anon_sym_exception] = ACTIONS(1781),
    [anon_sym_module] = ACTIONS(1781),
    [anon_sym_open] = ACTIONS(1781),
    [anon_sym_include] = ACTIONS(1781),
    [anon_sym_class] = ACTIONS(1781),
    [anon_sym_val] = ACTIONS(1781),
    [anon_sym_end] = ACTIONS(1781),
    [anon_sym_object] = ACTIONS(1781),
    [anon_sym_AMP] = ACTIONS(1781),
    [anon_sym_POUND] = ACTIONS(1781),
    [anon_sym_COLON_COLON] = ACTIONS(1783),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1783),
    [anon_sym_else] = ACTIONS(1781),
    [anon_sym_new] = ACTIONS(1781),
    [anon_sym_LBRACE_LT] = ACTIONS(1783),
    [anon_sym_begin] = ACTIONS(1781),
    [sym_ocamlyacc_value] = ACTIONS(1783),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1781),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1783),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1781),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1781),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1783),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1783),
    [aux_sym_number_token1] = ACTIONS(1783),
    [aux_sym_signed_number_token1] = ACTIONS(1781),
    [anon_sym_SQUOTE] = ACTIONS(1783),
    [anon_sym_DQUOTE] = ACTIONS(1783),
    [sym_prefix_operator] = ACTIONS(1783),
    [aux_sym_sign_operator_token1] = ACTIONS(1781),
    [sym_hash_operator] = ACTIONS(1783),
    [anon_sym_lsl] = ACTIONS(1781),
    [anon_sym_lsr] = ACTIONS(1781),
    [anon_sym_asr] = ACTIONS(1781),
    [aux_sym_pow_operator_token1] = ACTIONS(1783),
    [anon_sym_mod] = ACTIONS(1781),
    [anon_sym_land] = ACTIONS(1781),
    [anon_sym_lor] = ACTIONS(1781),
    [anon_sym_lxor] = ACTIONS(1781),
    [aux_sym_mult_operator_token1] = ACTIONS(1781),
    [aux_sym_add_operator_token1] = ACTIONS(1781),
    [sym_concat_operator] = ACTIONS(1783),
    [sym_rel_operator] = ACTIONS(1781),
    [anon_sym_AMP_AMP] = ACTIONS(1781),
    [anon_sym_or] = ACTIONS(1781),
    [anon_sym_PIPE_PIPE] = ACTIONS(1781),
    [sym_assign_operator] = ACTIONS(1783),
    [sym_let_operator] = ACTIONS(1783),
    [sym__uppercase_identifier] = ACTIONS(1783),
    [aux_sym_directive_token1] = ACTIONS(1781),
    [aux_sym_tag_token1] = ACTIONS(1783),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1094)] = {
    [sym__parenthesized_abstract_type] = STATE(1321),
    [sym__delimited_binding_pattern] = STATE(4914),
    [sym__simple_binding_pattern] = STATE(4921),
    [sym__effect_binding_pattern] = STATE(4846),
    [sym__binding_pattern] = STATE(4940),
    [sym_alias_binding_pattern] = STATE(4846),
    [sym_typed_binding_pattern] = STATE(4914),
    [sym__or_binding_pattern_anonymous] = STATE(4848),
    [sym_constructor_binding_pattern] = STATE(4921),
    [sym_tag_binding_pattern] = STATE(4921),
    [sym_polymorphic_variant_pattern] = STATE(4914),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6450),
    [sym__tuple_binding_pattern] = STATE(4850),
    [sym_record_binding_pattern] = STATE(4922),
    [sym_list_binding_pattern] = STATE(4922),
    [sym_cons_binding_pattern] = STATE(4846),
    [sym_array_binding_pattern] = STATE(4922),
    [sym_range_pattern] = STATE(4846),
    [sym_lazy_binding_pattern] = STATE(4921),
    [sym_local_open_binding_pattern] = STATE(4914),
    [sym_package_pattern] = STATE(4914),
    [sym_parenthesized_binding_pattern] = STATE(4922),
    [sym_exception_binding_pattern] = STATE(4846),
    [sym_effect_binding_pattern] = STATE(4846),
    [sym_attribute] = STATE(1094),
    [sym_extension] = STATE(4914),
    [sym_quoted_extension] = STATE(4914),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4739),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4922),
    [sym__value_name] = STATE(4914),
    [sym_parenthesized_operator] = STATE(4863),
    [sym_module_path] = STATE(6362),
    [sym_constructor_path] = STATE(1094),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1128),
    [sym__lowercase_identifier] = ACTIONS(2246),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(2248),
    [anon_sym_COMMA] = ACTIONS(1295),
    [anon_sym_PIPE] = ACTIONS(1297),
    [anon_sym_LBRACK] = ACTIONS(2250),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2252),
    [anon_sym_SEMI] = ACTIONS(1295),
    [anon_sym_exception] = ACTIONS(2254),
    [anon_sym_as] = ACTIONS(1297),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_COLON_COLON] = ACTIONS(1295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2256),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1295),
    [anon_sym_lazy] = ACTIONS(2258),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2260),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1095)] = {
    [sym_attribute] = STATE(1095),
    [sym__lowercase_identifier] = ACTIONS(1745),
    [anon_sym_SEMI_SEMI] = ACTIONS(1747),
    [anon_sym_let] = ACTIONS(1745),
    [anon_sym_TILDE] = ACTIONS(1745),
    [anon_sym_QMARK] = ACTIONS(1745),
    [anon_sym_LPAREN] = ACTIONS(1747),
    [anon_sym_external] = ACTIONS(1745),
    [anon_sym_type] = ACTIONS(1745),
    [anon_sym_COMMA] = ACTIONS(1747),
    [anon_sym_PIPE] = ACTIONS(1745),
    [anon_sym_DOT] = ACTIONS(1747),
    [anon_sym_LBRACK] = ACTIONS(1745),
    [anon_sym_RBRACK] = ACTIONS(1747),
    [anon_sym_true] = ACTIONS(1745),
    [anon_sym_false] = ACTIONS(1745),
    [anon_sym_LBRACE] = ACTIONS(1745),
    [anon_sym_SEMI] = ACTIONS(1745),
    [anon_sym_exception] = ACTIONS(1745),
    [anon_sym_module] = ACTIONS(1745),
    [anon_sym_open] = ACTIONS(1745),
    [anon_sym_include] = ACTIONS(1745),
    [anon_sym_class] = ACTIONS(1745),
    [anon_sym_val] = ACTIONS(1745),
    [anon_sym_end] = ACTIONS(1745),
    [anon_sym_object] = ACTIONS(1745),
    [anon_sym_AMP] = ACTIONS(1745),
    [anon_sym_POUND] = ACTIONS(1745),
    [anon_sym_COLON_COLON] = ACTIONS(1747),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1747),
    [anon_sym_LT_DASH] = ACTIONS(1955),
    [anon_sym_else] = ACTIONS(1745),
    [anon_sym_new] = ACTIONS(1745),
    [anon_sym_LBRACE_LT] = ACTIONS(1747),
    [anon_sym_begin] = ACTIONS(1745),
    [sym_ocamlyacc_value] = ACTIONS(1747),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1745),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1747),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1745),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1745),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1747),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1747),
    [aux_sym_number_token1] = ACTIONS(1747),
    [aux_sym_signed_number_token1] = ACTIONS(1745),
    [anon_sym_SQUOTE] = ACTIONS(1747),
    [anon_sym_DQUOTE] = ACTIONS(1747),
    [sym_prefix_operator] = ACTIONS(1747),
    [aux_sym_sign_operator_token1] = ACTIONS(1745),
    [sym_hash_operator] = ACTIONS(1747),
    [anon_sym_lsl] = ACTIONS(1745),
    [anon_sym_lsr] = ACTIONS(1745),
    [anon_sym_asr] = ACTIONS(1745),
    [aux_sym_pow_operator_token1] = ACTIONS(1747),
    [anon_sym_mod] = ACTIONS(1745),
    [anon_sym_land] = ACTIONS(1745),
    [anon_sym_lor] = ACTIONS(1745),
    [anon_sym_lxor] = ACTIONS(1745),
    [aux_sym_mult_operator_token1] = ACTIONS(1745),
    [aux_sym_add_operator_token1] = ACTIONS(1745),
    [sym_concat_operator] = ACTIONS(1747),
    [sym_rel_operator] = ACTIONS(1745),
    [anon_sym_AMP_AMP] = ACTIONS(1745),
    [anon_sym_or] = ACTIONS(1745),
    [anon_sym_PIPE_PIPE] = ACTIONS(1745),
    [sym_assign_operator] = ACTIONS(1747),
    [sym_let_operator] = ACTIONS(1747),
    [sym__uppercase_identifier] = ACTIONS(1747),
    [aux_sym_directive_token1] = ACTIONS(1745),
    [aux_sym_tag_token1] = ACTIONS(1747),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1096)] = {
    [sym_attribute] = STATE(1096),
    [sym__lowercase_identifier] = ACTIONS(1619),
    [anon_sym_SEMI_SEMI] = ACTIONS(1621),
    [anon_sym_let] = ACTIONS(1619),
    [anon_sym_and] = ACTIONS(1619),
    [anon_sym_TILDE] = ACTIONS(1619),
    [anon_sym_QMARK] = ACTIONS(1619),
    [anon_sym_LPAREN] = ACTIONS(1621),
    [anon_sym_external] = ACTIONS(1619),
    [anon_sym_type] = ACTIONS(1619),
    [anon_sym_COMMA] = ACTIONS(1621),
    [anon_sym_PIPE] = ACTIONS(1619),
    [anon_sym_DOT] = ACTIONS(1621),
    [anon_sym_LBRACK] = ACTIONS(1619),
    [anon_sym_RBRACK] = ACTIONS(1621),
    [anon_sym_true] = ACTIONS(1619),
    [anon_sym_false] = ACTIONS(1619),
    [anon_sym_LBRACE] = ACTIONS(1619),
    [anon_sym_SEMI] = ACTIONS(1619),
    [anon_sym_exception] = ACTIONS(1619),
    [anon_sym_module] = ACTIONS(1619),
    [anon_sym_open] = ACTIONS(1619),
    [anon_sym_include] = ACTIONS(1619),
    [anon_sym_class] = ACTIONS(1619),
    [anon_sym_val] = ACTIONS(1619),
    [anon_sym_end] = ACTIONS(1619),
    [anon_sym_object] = ACTIONS(1619),
    [anon_sym_AMP] = ACTIONS(1619),
    [anon_sym_POUND] = ACTIONS(1619),
    [anon_sym_COLON_COLON] = ACTIONS(1621),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1621),
    [anon_sym_else] = ACTIONS(1619),
    [anon_sym_new] = ACTIONS(1619),
    [anon_sym_LBRACE_LT] = ACTIONS(1621),
    [anon_sym_begin] = ACTIONS(1619),
    [sym_ocamlyacc_value] = ACTIONS(1621),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1619),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1621),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1619),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1619),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1621),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1621),
    [aux_sym_number_token1] = ACTIONS(1621),
    [aux_sym_signed_number_token1] = ACTIONS(1619),
    [anon_sym_SQUOTE] = ACTIONS(1621),
    [anon_sym_DQUOTE] = ACTIONS(1621),
    [sym_prefix_operator] = ACTIONS(1621),
    [aux_sym_sign_operator_token1] = ACTIONS(1619),
    [sym_hash_operator] = ACTIONS(1621),
    [anon_sym_lsl] = ACTIONS(1619),
    [anon_sym_lsr] = ACTIONS(1619),
    [anon_sym_asr] = ACTIONS(1619),
    [aux_sym_pow_operator_token1] = ACTIONS(1621),
    [anon_sym_mod] = ACTIONS(1619),
    [anon_sym_land] = ACTIONS(1619),
    [anon_sym_lor] = ACTIONS(1619),
    [anon_sym_lxor] = ACTIONS(1619),
    [aux_sym_mult_operator_token1] = ACTIONS(1619),
    [aux_sym_add_operator_token1] = ACTIONS(1619),
    [sym_concat_operator] = ACTIONS(1621),
    [sym_rel_operator] = ACTIONS(1619),
    [anon_sym_AMP_AMP] = ACTIONS(1619),
    [anon_sym_or] = ACTIONS(1619),
    [anon_sym_PIPE_PIPE] = ACTIONS(1619),
    [sym_assign_operator] = ACTIONS(1621),
    [sym_let_operator] = ACTIONS(1621),
    [sym__uppercase_identifier] = ACTIONS(1621),
    [aux_sym_directive_token1] = ACTIONS(1619),
    [aux_sym_tag_token1] = ACTIONS(1621),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1097)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4355),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(6815),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1097),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(1042),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1097),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_COMMA] = ACTIONS(2075),
    [anon_sym_PIPE] = ACTIONS(2075),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_RBRACK] = ACTIONS(2075),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_SEMI] = ACTIONS(2075),
    [anon_sym_RBRACE] = ACTIONS(2075),
    [anon_sym_exception] = ACTIONS(2198),
    [anon_sym_as] = ACTIONS(2081),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_COLON_COLON] = ACTIONS(2075),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(2200),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2202),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1098)] = {
    [sym_attribute] = STATE(1098),
    [sym__lowercase_identifier] = ACTIONS(1829),
    [anon_sym_SEMI_SEMI] = ACTIONS(1831),
    [anon_sym_let] = ACTIONS(1829),
    [anon_sym_and] = ACTIONS(1829),
    [anon_sym_TILDE] = ACTIONS(1829),
    [anon_sym_QMARK] = ACTIONS(1829),
    [anon_sym_LPAREN] = ACTIONS(1831),
    [anon_sym_external] = ACTIONS(1829),
    [anon_sym_type] = ACTIONS(1829),
    [anon_sym_COMMA] = ACTIONS(1831),
    [anon_sym_PIPE] = ACTIONS(1829),
    [anon_sym_DOT] = ACTIONS(1831),
    [anon_sym_LBRACK] = ACTIONS(1829),
    [anon_sym_RBRACK] = ACTIONS(1831),
    [anon_sym_true] = ACTIONS(1829),
    [anon_sym_false] = ACTIONS(1829),
    [anon_sym_LBRACE] = ACTIONS(1829),
    [anon_sym_SEMI] = ACTIONS(1829),
    [anon_sym_exception] = ACTIONS(1829),
    [anon_sym_module] = ACTIONS(1829),
    [anon_sym_open] = ACTIONS(1829),
    [anon_sym_include] = ACTIONS(1829),
    [anon_sym_class] = ACTIONS(1829),
    [anon_sym_val] = ACTIONS(1829),
    [anon_sym_end] = ACTIONS(1829),
    [anon_sym_object] = ACTIONS(1829),
    [anon_sym_AMP] = ACTIONS(1829),
    [anon_sym_POUND] = ACTIONS(1829),
    [anon_sym_COLON_COLON] = ACTIONS(1831),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1831),
    [anon_sym_else] = ACTIONS(1829),
    [anon_sym_new] = ACTIONS(1829),
    [anon_sym_LBRACE_LT] = ACTIONS(1831),
    [anon_sym_begin] = ACTIONS(1829),
    [sym_ocamlyacc_value] = ACTIONS(1831),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1829),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1831),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1829),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1829),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1831),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1831),
    [aux_sym_number_token1] = ACTIONS(1831),
    [aux_sym_signed_number_token1] = ACTIONS(1829),
    [anon_sym_SQUOTE] = ACTIONS(1831),
    [anon_sym_DQUOTE] = ACTIONS(1831),
    [sym_prefix_operator] = ACTIONS(1831),
    [aux_sym_sign_operator_token1] = ACTIONS(1829),
    [sym_hash_operator] = ACTIONS(1831),
    [anon_sym_lsl] = ACTIONS(1829),
    [anon_sym_lsr] = ACTIONS(1829),
    [anon_sym_asr] = ACTIONS(1829),
    [aux_sym_pow_operator_token1] = ACTIONS(1831),
    [anon_sym_mod] = ACTIONS(1829),
    [anon_sym_land] = ACTIONS(1829),
    [anon_sym_lor] = ACTIONS(1829),
    [anon_sym_lxor] = ACTIONS(1829),
    [aux_sym_mult_operator_token1] = ACTIONS(1829),
    [aux_sym_add_operator_token1] = ACTIONS(1829),
    [sym_concat_operator] = ACTIONS(1831),
    [sym_rel_operator] = ACTIONS(1829),
    [anon_sym_AMP_AMP] = ACTIONS(1829),
    [anon_sym_or] = ACTIONS(1829),
    [anon_sym_PIPE_PIPE] = ACTIONS(1829),
    [sym_assign_operator] = ACTIONS(1831),
    [sym_let_operator] = ACTIONS(1831),
    [sym__uppercase_identifier] = ACTIONS(1831),
    [aux_sym_directive_token1] = ACTIONS(1829),
    [aux_sym_tag_token1] = ACTIONS(1831),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1099)] = {
    [sym_attribute] = STATE(1099),
    [sym__lowercase_identifier] = ACTIONS(1699),
    [anon_sym_SEMI_SEMI] = ACTIONS(1701),
    [anon_sym_let] = ACTIONS(1699),
    [anon_sym_and] = ACTIONS(1699),
    [anon_sym_TILDE] = ACTIONS(1699),
    [anon_sym_QMARK] = ACTIONS(1699),
    [anon_sym_LPAREN] = ACTIONS(1701),
    [anon_sym_external] = ACTIONS(1699),
    [anon_sym_type] = ACTIONS(1699),
    [anon_sym_COMMA] = ACTIONS(1701),
    [anon_sym_PIPE] = ACTIONS(1699),
    [anon_sym_DOT] = ACTIONS(1703),
    [anon_sym_LBRACK] = ACTIONS(1699),
    [anon_sym_RBRACK] = ACTIONS(1701),
    [anon_sym_true] = ACTIONS(1699),
    [anon_sym_false] = ACTIONS(1699),
    [anon_sym_LBRACE] = ACTIONS(1699),
    [anon_sym_SEMI] = ACTIONS(1699),
    [anon_sym_exception] = ACTIONS(1699),
    [anon_sym_module] = ACTIONS(1699),
    [anon_sym_open] = ACTIONS(1699),
    [anon_sym_include] = ACTIONS(1699),
    [anon_sym_class] = ACTIONS(1699),
    [anon_sym_val] = ACTIONS(1699),
    [anon_sym_end] = ACTIONS(1699),
    [anon_sym_object] = ACTIONS(1699),
    [anon_sym_AMP] = ACTIONS(1699),
    [anon_sym_POUND] = ACTIONS(1699),
    [anon_sym_COLON_COLON] = ACTIONS(1701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1701),
    [anon_sym_else] = ACTIONS(1699),
    [anon_sym_new] = ACTIONS(1699),
    [anon_sym_LBRACE_LT] = ACTIONS(1701),
    [anon_sym_begin] = ACTIONS(1699),
    [sym_ocamlyacc_value] = ACTIONS(1701),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1699),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1701),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1699),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1699),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1701),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1701),
    [aux_sym_number_token1] = ACTIONS(1701),
    [aux_sym_signed_number_token1] = ACTIONS(1699),
    [anon_sym_SQUOTE] = ACTIONS(1701),
    [anon_sym_DQUOTE] = ACTIONS(1701),
    [sym_prefix_operator] = ACTIONS(1701),
    [aux_sym_sign_operator_token1] = ACTIONS(1699),
    [sym_hash_operator] = ACTIONS(1701),
    [anon_sym_lsl] = ACTIONS(1699),
    [anon_sym_lsr] = ACTIONS(1699),
    [anon_sym_asr] = ACTIONS(1699),
    [aux_sym_pow_operator_token1] = ACTIONS(1701),
    [anon_sym_mod] = ACTIONS(1699),
    [anon_sym_land] = ACTIONS(1699),
    [anon_sym_lor] = ACTIONS(1699),
    [anon_sym_lxor] = ACTIONS(1699),
    [aux_sym_mult_operator_token1] = ACTIONS(1699),
    [aux_sym_add_operator_token1] = ACTIONS(1699),
    [sym_concat_operator] = ACTIONS(1701),
    [sym_rel_operator] = ACTIONS(1699),
    [anon_sym_AMP_AMP] = ACTIONS(1699),
    [anon_sym_or] = ACTIONS(1699),
    [anon_sym_PIPE_PIPE] = ACTIONS(1699),
    [sym_assign_operator] = ACTIONS(1701),
    [sym_let_operator] = ACTIONS(1701),
    [sym__uppercase_identifier] = ACTIONS(1701),
    [aux_sym_directive_token1] = ACTIONS(1699),
    [aux_sym_tag_token1] = ACTIONS(1701),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1100)] = {
    [sym_attribute] = STATE(1100),
    [sym__lowercase_identifier] = ACTIONS(1567),
    [anon_sym_SEMI_SEMI] = ACTIONS(1569),
    [anon_sym_let] = ACTIONS(1567),
    [anon_sym_and] = ACTIONS(1567),
    [anon_sym_TILDE] = ACTIONS(1567),
    [anon_sym_QMARK] = ACTIONS(1567),
    [anon_sym_LPAREN] = ACTIONS(1569),
    [anon_sym_external] = ACTIONS(1567),
    [anon_sym_type] = ACTIONS(1567),
    [anon_sym_COMMA] = ACTIONS(1569),
    [anon_sym_PIPE] = ACTIONS(1567),
    [anon_sym_DOT] = ACTIONS(1569),
    [anon_sym_LBRACK] = ACTIONS(1567),
    [anon_sym_RBRACK] = ACTIONS(1569),
    [anon_sym_true] = ACTIONS(1567),
    [anon_sym_false] = ACTIONS(1567),
    [anon_sym_LBRACE] = ACTIONS(1567),
    [anon_sym_SEMI] = ACTIONS(1567),
    [anon_sym_exception] = ACTIONS(1567),
    [anon_sym_module] = ACTIONS(1567),
    [anon_sym_open] = ACTIONS(1567),
    [anon_sym_include] = ACTIONS(1567),
    [anon_sym_class] = ACTIONS(1567),
    [anon_sym_val] = ACTIONS(1567),
    [anon_sym_end] = ACTIONS(1567),
    [anon_sym_object] = ACTIONS(1567),
    [anon_sym_AMP] = ACTIONS(1567),
    [anon_sym_POUND] = ACTIONS(1567),
    [anon_sym_COLON_COLON] = ACTIONS(1569),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1569),
    [anon_sym_else] = ACTIONS(1567),
    [anon_sym_new] = ACTIONS(1567),
    [anon_sym_LBRACE_LT] = ACTIONS(1569),
    [anon_sym_begin] = ACTIONS(1567),
    [sym_ocamlyacc_value] = ACTIONS(1569),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1567),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1569),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1567),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1567),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1569),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1569),
    [aux_sym_number_token1] = ACTIONS(1569),
    [aux_sym_signed_number_token1] = ACTIONS(1567),
    [anon_sym_SQUOTE] = ACTIONS(1569),
    [anon_sym_DQUOTE] = ACTIONS(1569),
    [sym_prefix_operator] = ACTIONS(1569),
    [aux_sym_sign_operator_token1] = ACTIONS(1567),
    [sym_hash_operator] = ACTIONS(1569),
    [anon_sym_lsl] = ACTIONS(1567),
    [anon_sym_lsr] = ACTIONS(1567),
    [anon_sym_asr] = ACTIONS(1567),
    [aux_sym_pow_operator_token1] = ACTIONS(1569),
    [anon_sym_mod] = ACTIONS(1567),
    [anon_sym_land] = ACTIONS(1567),
    [anon_sym_lor] = ACTIONS(1567),
    [anon_sym_lxor] = ACTIONS(1567),
    [aux_sym_mult_operator_token1] = ACTIONS(1567),
    [aux_sym_add_operator_token1] = ACTIONS(1567),
    [sym_concat_operator] = ACTIONS(1569),
    [sym_rel_operator] = ACTIONS(1567),
    [anon_sym_AMP_AMP] = ACTIONS(1567),
    [anon_sym_or] = ACTIONS(1567),
    [anon_sym_PIPE_PIPE] = ACTIONS(1567),
    [sym_assign_operator] = ACTIONS(1569),
    [sym_let_operator] = ACTIONS(1569),
    [sym__uppercase_identifier] = ACTIONS(1569),
    [aux_sym_directive_token1] = ACTIONS(1567),
    [aux_sym_tag_token1] = ACTIONS(1569),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1101)] = {
    [sym_attribute] = STATE(1101),
    [sym__lowercase_identifier] = ACTIONS(1797),
    [anon_sym_SEMI_SEMI] = ACTIONS(1799),
    [anon_sym_let] = ACTIONS(1797),
    [anon_sym_and] = ACTIONS(1797),
    [anon_sym_TILDE] = ACTIONS(1797),
    [anon_sym_QMARK] = ACTIONS(1797),
    [anon_sym_LPAREN] = ACTIONS(1799),
    [anon_sym_external] = ACTIONS(1797),
    [anon_sym_type] = ACTIONS(1797),
    [anon_sym_COMMA] = ACTIONS(1799),
    [anon_sym_PIPE] = ACTIONS(1797),
    [anon_sym_DOT] = ACTIONS(1799),
    [anon_sym_LBRACK] = ACTIONS(1797),
    [anon_sym_RBRACK] = ACTIONS(1799),
    [anon_sym_true] = ACTIONS(1797),
    [anon_sym_false] = ACTIONS(1797),
    [anon_sym_LBRACE] = ACTIONS(1797),
    [anon_sym_SEMI] = ACTIONS(1797),
    [anon_sym_exception] = ACTIONS(1797),
    [anon_sym_module] = ACTIONS(1797),
    [anon_sym_open] = ACTIONS(1797),
    [anon_sym_include] = ACTIONS(1797),
    [anon_sym_class] = ACTIONS(1797),
    [anon_sym_val] = ACTIONS(1797),
    [anon_sym_end] = ACTIONS(1797),
    [anon_sym_object] = ACTIONS(1797),
    [anon_sym_AMP] = ACTIONS(1797),
    [anon_sym_POUND] = ACTIONS(1797),
    [anon_sym_COLON_COLON] = ACTIONS(1799),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1799),
    [anon_sym_else] = ACTIONS(1797),
    [anon_sym_new] = ACTIONS(1797),
    [anon_sym_LBRACE_LT] = ACTIONS(1799),
    [anon_sym_begin] = ACTIONS(1797),
    [sym_ocamlyacc_value] = ACTIONS(1799),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1797),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1799),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1797),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1797),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1799),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1799),
    [aux_sym_number_token1] = ACTIONS(1799),
    [aux_sym_signed_number_token1] = ACTIONS(1797),
    [anon_sym_SQUOTE] = ACTIONS(1799),
    [anon_sym_DQUOTE] = ACTIONS(1799),
    [sym_prefix_operator] = ACTIONS(1799),
    [aux_sym_sign_operator_token1] = ACTIONS(1797),
    [sym_hash_operator] = ACTIONS(1799),
    [anon_sym_lsl] = ACTIONS(1797),
    [anon_sym_lsr] = ACTIONS(1797),
    [anon_sym_asr] = ACTIONS(1797),
    [aux_sym_pow_operator_token1] = ACTIONS(1799),
    [anon_sym_mod] = ACTIONS(1797),
    [anon_sym_land] = ACTIONS(1797),
    [anon_sym_lor] = ACTIONS(1797),
    [anon_sym_lxor] = ACTIONS(1797),
    [aux_sym_mult_operator_token1] = ACTIONS(1797),
    [aux_sym_add_operator_token1] = ACTIONS(1797),
    [sym_concat_operator] = ACTIONS(1799),
    [sym_rel_operator] = ACTIONS(1797),
    [anon_sym_AMP_AMP] = ACTIONS(1797),
    [anon_sym_or] = ACTIONS(1797),
    [anon_sym_PIPE_PIPE] = ACTIONS(1797),
    [sym_assign_operator] = ACTIONS(1799),
    [sym_let_operator] = ACTIONS(1799),
    [sym__uppercase_identifier] = ACTIONS(1799),
    [aux_sym_directive_token1] = ACTIONS(1797),
    [aux_sym_tag_token1] = ACTIONS(1799),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1102)] = {
    [sym_attribute] = STATE(1102),
    [sym__lowercase_identifier] = ACTIONS(1741),
    [anon_sym_SEMI_SEMI] = ACTIONS(1743),
    [anon_sym_let] = ACTIONS(1741),
    [anon_sym_and] = ACTIONS(1741),
    [anon_sym_TILDE] = ACTIONS(1741),
    [anon_sym_QMARK] = ACTIONS(1741),
    [anon_sym_LPAREN] = ACTIONS(1743),
    [anon_sym_external] = ACTIONS(1741),
    [anon_sym_type] = ACTIONS(1741),
    [anon_sym_COMMA] = ACTIONS(1743),
    [anon_sym_PIPE] = ACTIONS(1741),
    [anon_sym_DOT] = ACTIONS(1743),
    [anon_sym_LBRACK] = ACTIONS(1741),
    [anon_sym_RBRACK] = ACTIONS(1743),
    [anon_sym_true] = ACTIONS(1741),
    [anon_sym_false] = ACTIONS(1741),
    [anon_sym_LBRACE] = ACTIONS(1741),
    [anon_sym_SEMI] = ACTIONS(1741),
    [anon_sym_exception] = ACTIONS(1741),
    [anon_sym_module] = ACTIONS(1741),
    [anon_sym_open] = ACTIONS(1741),
    [anon_sym_include] = ACTIONS(1741),
    [anon_sym_class] = ACTIONS(1741),
    [anon_sym_val] = ACTIONS(1741),
    [anon_sym_end] = ACTIONS(1741),
    [anon_sym_object] = ACTIONS(1741),
    [anon_sym_AMP] = ACTIONS(1741),
    [anon_sym_POUND] = ACTIONS(1741),
    [anon_sym_COLON_COLON] = ACTIONS(1743),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1743),
    [anon_sym_else] = ACTIONS(1741),
    [anon_sym_new] = ACTIONS(1741),
    [anon_sym_LBRACE_LT] = ACTIONS(1743),
    [anon_sym_begin] = ACTIONS(1741),
    [sym_ocamlyacc_value] = ACTIONS(1743),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1741),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1743),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1741),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1741),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1743),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1743),
    [aux_sym_number_token1] = ACTIONS(1743),
    [aux_sym_signed_number_token1] = ACTIONS(1741),
    [anon_sym_SQUOTE] = ACTIONS(1743),
    [anon_sym_DQUOTE] = ACTIONS(1743),
    [sym_prefix_operator] = ACTIONS(1743),
    [aux_sym_sign_operator_token1] = ACTIONS(1741),
    [sym_hash_operator] = ACTIONS(1743),
    [anon_sym_lsl] = ACTIONS(1741),
    [anon_sym_lsr] = ACTIONS(1741),
    [anon_sym_asr] = ACTIONS(1741),
    [aux_sym_pow_operator_token1] = ACTIONS(1743),
    [anon_sym_mod] = ACTIONS(1741),
    [anon_sym_land] = ACTIONS(1741),
    [anon_sym_lor] = ACTIONS(1741),
    [anon_sym_lxor] = ACTIONS(1741),
    [aux_sym_mult_operator_token1] = ACTIONS(1741),
    [aux_sym_add_operator_token1] = ACTIONS(1741),
    [sym_concat_operator] = ACTIONS(1743),
    [sym_rel_operator] = ACTIONS(1741),
    [anon_sym_AMP_AMP] = ACTIONS(1741),
    [anon_sym_or] = ACTIONS(1741),
    [anon_sym_PIPE_PIPE] = ACTIONS(1741),
    [sym_assign_operator] = ACTIONS(1743),
    [sym_let_operator] = ACTIONS(1743),
    [sym__uppercase_identifier] = ACTIONS(1743),
    [aux_sym_directive_token1] = ACTIONS(1741),
    [aux_sym_tag_token1] = ACTIONS(1743),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1103)] = {
    [sym_attribute] = STATE(1103),
    [sym__lowercase_identifier] = ACTIONS(1851),
    [anon_sym_SEMI_SEMI] = ACTIONS(1853),
    [anon_sym_let] = ACTIONS(1851),
    [anon_sym_and] = ACTIONS(1851),
    [anon_sym_TILDE] = ACTIONS(1851),
    [anon_sym_QMARK] = ACTIONS(1851),
    [anon_sym_LPAREN] = ACTIONS(1853),
    [anon_sym_external] = ACTIONS(1851),
    [anon_sym_type] = ACTIONS(1851),
    [anon_sym_COMMA] = ACTIONS(1853),
    [anon_sym_PIPE] = ACTIONS(1851),
    [anon_sym_DOT] = ACTIONS(1853),
    [anon_sym_LBRACK] = ACTIONS(1851),
    [anon_sym_RBRACK] = ACTIONS(1853),
    [anon_sym_true] = ACTIONS(1851),
    [anon_sym_false] = ACTIONS(1851),
    [anon_sym_LBRACE] = ACTIONS(1851),
    [anon_sym_SEMI] = ACTIONS(1851),
    [anon_sym_exception] = ACTIONS(1851),
    [anon_sym_module] = ACTIONS(1851),
    [anon_sym_open] = ACTIONS(1851),
    [anon_sym_include] = ACTIONS(1851),
    [anon_sym_class] = ACTIONS(1851),
    [anon_sym_val] = ACTIONS(1851),
    [anon_sym_end] = ACTIONS(1851),
    [anon_sym_object] = ACTIONS(1851),
    [anon_sym_AMP] = ACTIONS(1851),
    [anon_sym_POUND] = ACTIONS(1851),
    [anon_sym_COLON_COLON] = ACTIONS(1853),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1853),
    [anon_sym_else] = ACTIONS(1851),
    [anon_sym_new] = ACTIONS(1851),
    [anon_sym_LBRACE_LT] = ACTIONS(1853),
    [anon_sym_begin] = ACTIONS(1851),
    [sym_ocamlyacc_value] = ACTIONS(1853),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1851),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1853),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1851),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1851),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1853),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1853),
    [aux_sym_number_token1] = ACTIONS(1853),
    [aux_sym_signed_number_token1] = ACTIONS(1851),
    [anon_sym_SQUOTE] = ACTIONS(1853),
    [anon_sym_DQUOTE] = ACTIONS(1853),
    [sym_prefix_operator] = ACTIONS(1853),
    [aux_sym_sign_operator_token1] = ACTIONS(1851),
    [sym_hash_operator] = ACTIONS(1853),
    [anon_sym_lsl] = ACTIONS(1851),
    [anon_sym_lsr] = ACTIONS(1851),
    [anon_sym_asr] = ACTIONS(1851),
    [aux_sym_pow_operator_token1] = ACTIONS(1853),
    [anon_sym_mod] = ACTIONS(1851),
    [anon_sym_land] = ACTIONS(1851),
    [anon_sym_lor] = ACTIONS(1851),
    [anon_sym_lxor] = ACTIONS(1851),
    [aux_sym_mult_operator_token1] = ACTIONS(1851),
    [aux_sym_add_operator_token1] = ACTIONS(1851),
    [sym_concat_operator] = ACTIONS(1853),
    [sym_rel_operator] = ACTIONS(1851),
    [anon_sym_AMP_AMP] = ACTIONS(1851),
    [anon_sym_or] = ACTIONS(1851),
    [anon_sym_PIPE_PIPE] = ACTIONS(1851),
    [sym_assign_operator] = ACTIONS(1853),
    [sym_let_operator] = ACTIONS(1853),
    [sym__uppercase_identifier] = ACTIONS(1853),
    [aux_sym_directive_token1] = ACTIONS(1851),
    [aux_sym_tag_token1] = ACTIONS(1853),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1104)] = {
    [sym_attribute] = STATE(1104),
    [sym__lowercase_identifier] = ACTIONS(1749),
    [anon_sym_SEMI_SEMI] = ACTIONS(1751),
    [anon_sym_let] = ACTIONS(1749),
    [anon_sym_and] = ACTIONS(1749),
    [anon_sym_TILDE] = ACTIONS(1749),
    [anon_sym_QMARK] = ACTIONS(1749),
    [anon_sym_LPAREN] = ACTIONS(1751),
    [anon_sym_external] = ACTIONS(1749),
    [anon_sym_type] = ACTIONS(1749),
    [anon_sym_COMMA] = ACTIONS(1751),
    [anon_sym_PIPE] = ACTIONS(1749),
    [anon_sym_DOT] = ACTIONS(1751),
    [anon_sym_LBRACK] = ACTIONS(1749),
    [anon_sym_RBRACK] = ACTIONS(1751),
    [anon_sym_true] = ACTIONS(1749),
    [anon_sym_false] = ACTIONS(1749),
    [anon_sym_LBRACE] = ACTIONS(1749),
    [anon_sym_SEMI] = ACTIONS(1749),
    [anon_sym_exception] = ACTIONS(1749),
    [anon_sym_module] = ACTIONS(1749),
    [anon_sym_open] = ACTIONS(1749),
    [anon_sym_include] = ACTIONS(1749),
    [anon_sym_class] = ACTIONS(1749),
    [anon_sym_val] = ACTIONS(1749),
    [anon_sym_end] = ACTIONS(1749),
    [anon_sym_object] = ACTIONS(1749),
    [anon_sym_AMP] = ACTIONS(1749),
    [anon_sym_POUND] = ACTIONS(1749),
    [anon_sym_COLON_COLON] = ACTIONS(1751),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1751),
    [anon_sym_else] = ACTIONS(1749),
    [anon_sym_new] = ACTIONS(1749),
    [anon_sym_LBRACE_LT] = ACTIONS(1751),
    [anon_sym_begin] = ACTIONS(1749),
    [sym_ocamlyacc_value] = ACTIONS(1751),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1749),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1751),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1749),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1749),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1751),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1751),
    [aux_sym_number_token1] = ACTIONS(1751),
    [aux_sym_signed_number_token1] = ACTIONS(1749),
    [anon_sym_SQUOTE] = ACTIONS(1751),
    [anon_sym_DQUOTE] = ACTIONS(1751),
    [sym_prefix_operator] = ACTIONS(1751),
    [aux_sym_sign_operator_token1] = ACTIONS(1749),
    [sym_hash_operator] = ACTIONS(1751),
    [anon_sym_lsl] = ACTIONS(1749),
    [anon_sym_lsr] = ACTIONS(1749),
    [anon_sym_asr] = ACTIONS(1749),
    [aux_sym_pow_operator_token1] = ACTIONS(1751),
    [anon_sym_mod] = ACTIONS(1749),
    [anon_sym_land] = ACTIONS(1749),
    [anon_sym_lor] = ACTIONS(1749),
    [anon_sym_lxor] = ACTIONS(1749),
    [aux_sym_mult_operator_token1] = ACTIONS(1749),
    [aux_sym_add_operator_token1] = ACTIONS(1749),
    [sym_concat_operator] = ACTIONS(1751),
    [sym_rel_operator] = ACTIONS(1749),
    [anon_sym_AMP_AMP] = ACTIONS(1749),
    [anon_sym_or] = ACTIONS(1749),
    [anon_sym_PIPE_PIPE] = ACTIONS(1749),
    [sym_assign_operator] = ACTIONS(1751),
    [sym_let_operator] = ACTIONS(1751),
    [sym__uppercase_identifier] = ACTIONS(1751),
    [aux_sym_directive_token1] = ACTIONS(1749),
    [aux_sym_tag_token1] = ACTIONS(1751),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1105)] = {
    [sym_attribute] = STATE(1105),
    [sym__lowercase_identifier] = ACTIONS(1579),
    [anon_sym_SEMI_SEMI] = ACTIONS(1581),
    [anon_sym_let] = ACTIONS(1579),
    [anon_sym_and] = ACTIONS(1579),
    [anon_sym_TILDE] = ACTIONS(1579),
    [anon_sym_QMARK] = ACTIONS(1579),
    [anon_sym_LPAREN] = ACTIONS(1581),
    [anon_sym_external] = ACTIONS(1579),
    [anon_sym_type] = ACTIONS(1579),
    [anon_sym_COMMA] = ACTIONS(1581),
    [anon_sym_PIPE] = ACTIONS(1579),
    [anon_sym_DOT] = ACTIONS(1581),
    [anon_sym_LBRACK] = ACTIONS(1579),
    [anon_sym_RBRACK] = ACTIONS(1581),
    [anon_sym_true] = ACTIONS(1579),
    [anon_sym_false] = ACTIONS(1579),
    [anon_sym_LBRACE] = ACTIONS(1579),
    [anon_sym_SEMI] = ACTIONS(1579),
    [anon_sym_exception] = ACTIONS(1579),
    [anon_sym_module] = ACTIONS(1579),
    [anon_sym_open] = ACTIONS(1579),
    [anon_sym_include] = ACTIONS(1579),
    [anon_sym_class] = ACTIONS(1579),
    [anon_sym_val] = ACTIONS(1579),
    [anon_sym_end] = ACTIONS(1579),
    [anon_sym_object] = ACTIONS(1579),
    [anon_sym_AMP] = ACTIONS(1579),
    [anon_sym_POUND] = ACTIONS(1579),
    [anon_sym_COLON_COLON] = ACTIONS(1581),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1581),
    [anon_sym_else] = ACTIONS(1579),
    [anon_sym_new] = ACTIONS(1579),
    [anon_sym_LBRACE_LT] = ACTIONS(1581),
    [anon_sym_begin] = ACTIONS(1579),
    [sym_ocamlyacc_value] = ACTIONS(1581),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1579),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1581),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1579),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1579),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1581),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1581),
    [aux_sym_number_token1] = ACTIONS(1581),
    [aux_sym_signed_number_token1] = ACTIONS(1579),
    [anon_sym_SQUOTE] = ACTIONS(1581),
    [anon_sym_DQUOTE] = ACTIONS(1581),
    [sym_prefix_operator] = ACTIONS(1581),
    [aux_sym_sign_operator_token1] = ACTIONS(1579),
    [sym_hash_operator] = ACTIONS(1581),
    [anon_sym_lsl] = ACTIONS(1579),
    [anon_sym_lsr] = ACTIONS(1579),
    [anon_sym_asr] = ACTIONS(1579),
    [aux_sym_pow_operator_token1] = ACTIONS(1581),
    [anon_sym_mod] = ACTIONS(1579),
    [anon_sym_land] = ACTIONS(1579),
    [anon_sym_lor] = ACTIONS(1579),
    [anon_sym_lxor] = ACTIONS(1579),
    [aux_sym_mult_operator_token1] = ACTIONS(1579),
    [aux_sym_add_operator_token1] = ACTIONS(1579),
    [sym_concat_operator] = ACTIONS(1581),
    [sym_rel_operator] = ACTIONS(1579),
    [anon_sym_AMP_AMP] = ACTIONS(1579),
    [anon_sym_or] = ACTIONS(1579),
    [anon_sym_PIPE_PIPE] = ACTIONS(1579),
    [sym_assign_operator] = ACTIONS(1581),
    [sym_let_operator] = ACTIONS(1581),
    [sym__uppercase_identifier] = ACTIONS(1581),
    [aux_sym_directive_token1] = ACTIONS(1579),
    [aux_sym_tag_token1] = ACTIONS(1581),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1106)] = {
    [sym_attribute] = STATE(1106),
    [sym__lowercase_identifier] = ACTIONS(1753),
    [anon_sym_SEMI_SEMI] = ACTIONS(1755),
    [anon_sym_let] = ACTIONS(1753),
    [anon_sym_and] = ACTIONS(1753),
    [anon_sym_TILDE] = ACTIONS(1753),
    [anon_sym_QMARK] = ACTIONS(1753),
    [anon_sym_LPAREN] = ACTIONS(1755),
    [anon_sym_external] = ACTIONS(1753),
    [anon_sym_type] = ACTIONS(1753),
    [anon_sym_COMMA] = ACTIONS(1755),
    [anon_sym_PIPE] = ACTIONS(1753),
    [anon_sym_DOT] = ACTIONS(1755),
    [anon_sym_LBRACK] = ACTIONS(1753),
    [anon_sym_RBRACK] = ACTIONS(1755),
    [anon_sym_true] = ACTIONS(1753),
    [anon_sym_false] = ACTIONS(1753),
    [anon_sym_LBRACE] = ACTIONS(1753),
    [anon_sym_SEMI] = ACTIONS(1753),
    [anon_sym_exception] = ACTIONS(1753),
    [anon_sym_module] = ACTIONS(1753),
    [anon_sym_open] = ACTIONS(1753),
    [anon_sym_include] = ACTIONS(1753),
    [anon_sym_class] = ACTIONS(1753),
    [anon_sym_val] = ACTIONS(1753),
    [anon_sym_end] = ACTIONS(1753),
    [anon_sym_object] = ACTIONS(1753),
    [anon_sym_AMP] = ACTIONS(1753),
    [anon_sym_POUND] = ACTIONS(1753),
    [anon_sym_COLON_COLON] = ACTIONS(1755),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1755),
    [anon_sym_else] = ACTIONS(1753),
    [anon_sym_new] = ACTIONS(1753),
    [anon_sym_LBRACE_LT] = ACTIONS(1755),
    [anon_sym_begin] = ACTIONS(1753),
    [sym_ocamlyacc_value] = ACTIONS(1755),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1753),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1755),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1753),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1753),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1755),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1755),
    [aux_sym_number_token1] = ACTIONS(1755),
    [aux_sym_signed_number_token1] = ACTIONS(1753),
    [anon_sym_SQUOTE] = ACTIONS(1755),
    [anon_sym_DQUOTE] = ACTIONS(1755),
    [sym_prefix_operator] = ACTIONS(1755),
    [aux_sym_sign_operator_token1] = ACTIONS(1753),
    [sym_hash_operator] = ACTIONS(1755),
    [anon_sym_lsl] = ACTIONS(1753),
    [anon_sym_lsr] = ACTIONS(1753),
    [anon_sym_asr] = ACTIONS(1753),
    [aux_sym_pow_operator_token1] = ACTIONS(1755),
    [anon_sym_mod] = ACTIONS(1753),
    [anon_sym_land] = ACTIONS(1753),
    [anon_sym_lor] = ACTIONS(1753),
    [anon_sym_lxor] = ACTIONS(1753),
    [aux_sym_mult_operator_token1] = ACTIONS(1753),
    [aux_sym_add_operator_token1] = ACTIONS(1753),
    [sym_concat_operator] = ACTIONS(1755),
    [sym_rel_operator] = ACTIONS(1753),
    [anon_sym_AMP_AMP] = ACTIONS(1753),
    [anon_sym_or] = ACTIONS(1753),
    [anon_sym_PIPE_PIPE] = ACTIONS(1753),
    [sym_assign_operator] = ACTIONS(1755),
    [sym_let_operator] = ACTIONS(1755),
    [sym__uppercase_identifier] = ACTIONS(1755),
    [aux_sym_directive_token1] = ACTIONS(1753),
    [aux_sym_tag_token1] = ACTIONS(1755),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1107)] = {
    [sym_attribute] = STATE(1107),
    [sym__lowercase_identifier] = ACTIONS(1761),
    [anon_sym_SEMI_SEMI] = ACTIONS(1763),
    [anon_sym_let] = ACTIONS(1761),
    [anon_sym_and] = ACTIONS(1761),
    [anon_sym_TILDE] = ACTIONS(1761),
    [anon_sym_QMARK] = ACTIONS(1761),
    [anon_sym_LPAREN] = ACTIONS(1763),
    [anon_sym_external] = ACTIONS(1761),
    [anon_sym_type] = ACTIONS(1761),
    [anon_sym_COMMA] = ACTIONS(1763),
    [anon_sym_PIPE] = ACTIONS(1761),
    [anon_sym_DOT] = ACTIONS(1763),
    [anon_sym_LBRACK] = ACTIONS(1761),
    [anon_sym_RBRACK] = ACTIONS(1763),
    [anon_sym_true] = ACTIONS(1761),
    [anon_sym_false] = ACTIONS(1761),
    [anon_sym_LBRACE] = ACTIONS(1761),
    [anon_sym_SEMI] = ACTIONS(1761),
    [anon_sym_exception] = ACTIONS(1761),
    [anon_sym_module] = ACTIONS(1761),
    [anon_sym_open] = ACTIONS(1761),
    [anon_sym_include] = ACTIONS(1761),
    [anon_sym_class] = ACTIONS(1761),
    [anon_sym_val] = ACTIONS(1761),
    [anon_sym_end] = ACTIONS(1761),
    [anon_sym_object] = ACTIONS(1761),
    [anon_sym_AMP] = ACTIONS(1761),
    [anon_sym_POUND] = ACTIONS(1761),
    [anon_sym_COLON_COLON] = ACTIONS(1763),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1763),
    [anon_sym_else] = ACTIONS(1761),
    [anon_sym_new] = ACTIONS(1761),
    [anon_sym_LBRACE_LT] = ACTIONS(1763),
    [anon_sym_begin] = ACTIONS(1761),
    [sym_ocamlyacc_value] = ACTIONS(1763),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1761),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1763),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1761),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1761),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1763),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1763),
    [aux_sym_number_token1] = ACTIONS(1763),
    [aux_sym_signed_number_token1] = ACTIONS(1761),
    [anon_sym_SQUOTE] = ACTIONS(1763),
    [anon_sym_DQUOTE] = ACTIONS(1763),
    [sym_prefix_operator] = ACTIONS(1763),
    [aux_sym_sign_operator_token1] = ACTIONS(1761),
    [sym_hash_operator] = ACTIONS(1763),
    [anon_sym_lsl] = ACTIONS(1761),
    [anon_sym_lsr] = ACTIONS(1761),
    [anon_sym_asr] = ACTIONS(1761),
    [aux_sym_pow_operator_token1] = ACTIONS(1763),
    [anon_sym_mod] = ACTIONS(1761),
    [anon_sym_land] = ACTIONS(1761),
    [anon_sym_lor] = ACTIONS(1761),
    [anon_sym_lxor] = ACTIONS(1761),
    [aux_sym_mult_operator_token1] = ACTIONS(1761),
    [aux_sym_add_operator_token1] = ACTIONS(1761),
    [sym_concat_operator] = ACTIONS(1763),
    [sym_rel_operator] = ACTIONS(1761),
    [anon_sym_AMP_AMP] = ACTIONS(1761),
    [anon_sym_or] = ACTIONS(1761),
    [anon_sym_PIPE_PIPE] = ACTIONS(1761),
    [sym_assign_operator] = ACTIONS(1763),
    [sym_let_operator] = ACTIONS(1763),
    [sym__uppercase_identifier] = ACTIONS(1763),
    [aux_sym_directive_token1] = ACTIONS(1761),
    [aux_sym_tag_token1] = ACTIONS(1763),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1108)] = {
    [sym_attribute] = STATE(1108),
    [sym__lowercase_identifier] = ACTIONS(1805),
    [anon_sym_SEMI_SEMI] = ACTIONS(1807),
    [anon_sym_let] = ACTIONS(1805),
    [anon_sym_and] = ACTIONS(1805),
    [anon_sym_TILDE] = ACTIONS(1805),
    [anon_sym_QMARK] = ACTIONS(1805),
    [anon_sym_LPAREN] = ACTIONS(1807),
    [anon_sym_external] = ACTIONS(1805),
    [anon_sym_type] = ACTIONS(1805),
    [anon_sym_COMMA] = ACTIONS(1807),
    [anon_sym_PIPE] = ACTIONS(1805),
    [anon_sym_DOT] = ACTIONS(1807),
    [anon_sym_LBRACK] = ACTIONS(1805),
    [anon_sym_RBRACK] = ACTIONS(1807),
    [anon_sym_true] = ACTIONS(1805),
    [anon_sym_false] = ACTIONS(1805),
    [anon_sym_LBRACE] = ACTIONS(1805),
    [anon_sym_SEMI] = ACTIONS(1805),
    [anon_sym_exception] = ACTIONS(1805),
    [anon_sym_module] = ACTIONS(1805),
    [anon_sym_open] = ACTIONS(1805),
    [anon_sym_include] = ACTIONS(1805),
    [anon_sym_class] = ACTIONS(1805),
    [anon_sym_val] = ACTIONS(1805),
    [anon_sym_end] = ACTIONS(1805),
    [anon_sym_object] = ACTIONS(1805),
    [anon_sym_AMP] = ACTIONS(1805),
    [anon_sym_POUND] = ACTIONS(1805),
    [anon_sym_COLON_COLON] = ACTIONS(1807),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1807),
    [anon_sym_else] = ACTIONS(1805),
    [anon_sym_new] = ACTIONS(1805),
    [anon_sym_LBRACE_LT] = ACTIONS(1807),
    [anon_sym_begin] = ACTIONS(1805),
    [sym_ocamlyacc_value] = ACTIONS(1807),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1805),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1807),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1805),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1805),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1807),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1807),
    [aux_sym_number_token1] = ACTIONS(1807),
    [aux_sym_signed_number_token1] = ACTIONS(1805),
    [anon_sym_SQUOTE] = ACTIONS(1807),
    [anon_sym_DQUOTE] = ACTIONS(1807),
    [sym_prefix_operator] = ACTIONS(1807),
    [aux_sym_sign_operator_token1] = ACTIONS(1805),
    [sym_hash_operator] = ACTIONS(1807),
    [anon_sym_lsl] = ACTIONS(1805),
    [anon_sym_lsr] = ACTIONS(1805),
    [anon_sym_asr] = ACTIONS(1805),
    [aux_sym_pow_operator_token1] = ACTIONS(1807),
    [anon_sym_mod] = ACTIONS(1805),
    [anon_sym_land] = ACTIONS(1805),
    [anon_sym_lor] = ACTIONS(1805),
    [anon_sym_lxor] = ACTIONS(1805),
    [aux_sym_mult_operator_token1] = ACTIONS(1805),
    [aux_sym_add_operator_token1] = ACTIONS(1805),
    [sym_concat_operator] = ACTIONS(1807),
    [sym_rel_operator] = ACTIONS(1805),
    [anon_sym_AMP_AMP] = ACTIONS(1805),
    [anon_sym_or] = ACTIONS(1805),
    [anon_sym_PIPE_PIPE] = ACTIONS(1805),
    [sym_assign_operator] = ACTIONS(1807),
    [sym_let_operator] = ACTIONS(1807),
    [sym__uppercase_identifier] = ACTIONS(1807),
    [aux_sym_directive_token1] = ACTIONS(1805),
    [aux_sym_tag_token1] = ACTIONS(1807),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1109)] = {
    [sym_attribute] = STATE(1109),
    [sym__lowercase_identifier] = ACTIONS(1793),
    [anon_sym_SEMI_SEMI] = ACTIONS(1795),
    [anon_sym_let] = ACTIONS(1793),
    [anon_sym_and] = ACTIONS(1793),
    [anon_sym_TILDE] = ACTIONS(1793),
    [anon_sym_QMARK] = ACTIONS(1793),
    [anon_sym_LPAREN] = ACTIONS(1795),
    [anon_sym_external] = ACTIONS(1793),
    [anon_sym_type] = ACTIONS(1793),
    [anon_sym_COMMA] = ACTIONS(1795),
    [anon_sym_PIPE] = ACTIONS(1793),
    [anon_sym_DOT] = ACTIONS(1795),
    [anon_sym_LBRACK] = ACTIONS(1793),
    [anon_sym_RBRACK] = ACTIONS(1795),
    [anon_sym_true] = ACTIONS(1793),
    [anon_sym_false] = ACTIONS(1793),
    [anon_sym_LBRACE] = ACTIONS(1793),
    [anon_sym_SEMI] = ACTIONS(1793),
    [anon_sym_exception] = ACTIONS(1793),
    [anon_sym_module] = ACTIONS(1793),
    [anon_sym_open] = ACTIONS(1793),
    [anon_sym_include] = ACTIONS(1793),
    [anon_sym_class] = ACTIONS(1793),
    [anon_sym_val] = ACTIONS(1793),
    [anon_sym_end] = ACTIONS(1793),
    [anon_sym_object] = ACTIONS(1793),
    [anon_sym_AMP] = ACTIONS(1793),
    [anon_sym_POUND] = ACTIONS(1793),
    [anon_sym_COLON_COLON] = ACTIONS(1795),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1795),
    [anon_sym_else] = ACTIONS(1793),
    [anon_sym_new] = ACTIONS(1793),
    [anon_sym_LBRACE_LT] = ACTIONS(1795),
    [anon_sym_begin] = ACTIONS(1793),
    [sym_ocamlyacc_value] = ACTIONS(1795),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1793),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1795),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1793),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1795),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1795),
    [aux_sym_number_token1] = ACTIONS(1795),
    [aux_sym_signed_number_token1] = ACTIONS(1793),
    [anon_sym_SQUOTE] = ACTIONS(1795),
    [anon_sym_DQUOTE] = ACTIONS(1795),
    [sym_prefix_operator] = ACTIONS(1795),
    [aux_sym_sign_operator_token1] = ACTIONS(1793),
    [sym_hash_operator] = ACTIONS(1795),
    [anon_sym_lsl] = ACTIONS(1793),
    [anon_sym_lsr] = ACTIONS(1793),
    [anon_sym_asr] = ACTIONS(1793),
    [aux_sym_pow_operator_token1] = ACTIONS(1795),
    [anon_sym_mod] = ACTIONS(1793),
    [anon_sym_land] = ACTIONS(1793),
    [anon_sym_lor] = ACTIONS(1793),
    [anon_sym_lxor] = ACTIONS(1793),
    [aux_sym_mult_operator_token1] = ACTIONS(1793),
    [aux_sym_add_operator_token1] = ACTIONS(1793),
    [sym_concat_operator] = ACTIONS(1795),
    [sym_rel_operator] = ACTIONS(1793),
    [anon_sym_AMP_AMP] = ACTIONS(1793),
    [anon_sym_or] = ACTIONS(1793),
    [anon_sym_PIPE_PIPE] = ACTIONS(1793),
    [sym_assign_operator] = ACTIONS(1795),
    [sym_let_operator] = ACTIONS(1795),
    [sym__uppercase_identifier] = ACTIONS(1795),
    [aux_sym_directive_token1] = ACTIONS(1793),
    [aux_sym_tag_token1] = ACTIONS(1795),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1110)] = {
    [sym_attribute] = STATE(1110),
    [sym__lowercase_identifier] = ACTIONS(1757),
    [anon_sym_SEMI_SEMI] = ACTIONS(1759),
    [anon_sym_let] = ACTIONS(1757),
    [anon_sym_and] = ACTIONS(1757),
    [anon_sym_TILDE] = ACTIONS(1757),
    [anon_sym_QMARK] = ACTIONS(1757),
    [anon_sym_LPAREN] = ACTIONS(1759),
    [anon_sym_external] = ACTIONS(1757),
    [anon_sym_type] = ACTIONS(1757),
    [anon_sym_COMMA] = ACTIONS(1759),
    [anon_sym_PIPE] = ACTIONS(1757),
    [anon_sym_DOT] = ACTIONS(1759),
    [anon_sym_LBRACK] = ACTIONS(1757),
    [anon_sym_RBRACK] = ACTIONS(1759),
    [anon_sym_true] = ACTIONS(1757),
    [anon_sym_false] = ACTIONS(1757),
    [anon_sym_LBRACE] = ACTIONS(1757),
    [anon_sym_SEMI] = ACTIONS(1757),
    [anon_sym_exception] = ACTIONS(1757),
    [anon_sym_module] = ACTIONS(1757),
    [anon_sym_open] = ACTIONS(1757),
    [anon_sym_include] = ACTIONS(1757),
    [anon_sym_class] = ACTIONS(1757),
    [anon_sym_val] = ACTIONS(1757),
    [anon_sym_end] = ACTIONS(1757),
    [anon_sym_object] = ACTIONS(1757),
    [anon_sym_AMP] = ACTIONS(1757),
    [anon_sym_POUND] = ACTIONS(1757),
    [anon_sym_COLON_COLON] = ACTIONS(1759),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1759),
    [anon_sym_else] = ACTIONS(1757),
    [anon_sym_new] = ACTIONS(1757),
    [anon_sym_LBRACE_LT] = ACTIONS(1759),
    [anon_sym_begin] = ACTIONS(1757),
    [sym_ocamlyacc_value] = ACTIONS(1759),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1757),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1759),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1757),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1757),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1759),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1759),
    [aux_sym_number_token1] = ACTIONS(1759),
    [aux_sym_signed_number_token1] = ACTIONS(1757),
    [anon_sym_SQUOTE] = ACTIONS(1759),
    [anon_sym_DQUOTE] = ACTIONS(1759),
    [sym_prefix_operator] = ACTIONS(1759),
    [aux_sym_sign_operator_token1] = ACTIONS(1757),
    [sym_hash_operator] = ACTIONS(1759),
    [anon_sym_lsl] = ACTIONS(1757),
    [anon_sym_lsr] = ACTIONS(1757),
    [anon_sym_asr] = ACTIONS(1757),
    [aux_sym_pow_operator_token1] = ACTIONS(1759),
    [anon_sym_mod] = ACTIONS(1757),
    [anon_sym_land] = ACTIONS(1757),
    [anon_sym_lor] = ACTIONS(1757),
    [anon_sym_lxor] = ACTIONS(1757),
    [aux_sym_mult_operator_token1] = ACTIONS(1757),
    [aux_sym_add_operator_token1] = ACTIONS(1757),
    [sym_concat_operator] = ACTIONS(1759),
    [sym_rel_operator] = ACTIONS(1757),
    [anon_sym_AMP_AMP] = ACTIONS(1757),
    [anon_sym_or] = ACTIONS(1757),
    [anon_sym_PIPE_PIPE] = ACTIONS(1757),
    [sym_assign_operator] = ACTIONS(1759),
    [sym_let_operator] = ACTIONS(1759),
    [sym__uppercase_identifier] = ACTIONS(1759),
    [aux_sym_directive_token1] = ACTIONS(1757),
    [aux_sym_tag_token1] = ACTIONS(1759),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1111)] = {
    [sym_attribute] = STATE(1111),
    [sym__lowercase_identifier] = ACTIONS(1809),
    [anon_sym_SEMI_SEMI] = ACTIONS(1811),
    [anon_sym_let] = ACTIONS(1809),
    [anon_sym_and] = ACTIONS(1809),
    [anon_sym_TILDE] = ACTIONS(1809),
    [anon_sym_QMARK] = ACTIONS(1809),
    [anon_sym_LPAREN] = ACTIONS(1811),
    [anon_sym_external] = ACTIONS(1809),
    [anon_sym_type] = ACTIONS(1809),
    [anon_sym_COMMA] = ACTIONS(1811),
    [anon_sym_PIPE] = ACTIONS(1809),
    [anon_sym_DOT] = ACTIONS(1811),
    [anon_sym_LBRACK] = ACTIONS(1809),
    [anon_sym_RBRACK] = ACTIONS(1811),
    [anon_sym_true] = ACTIONS(1809),
    [anon_sym_false] = ACTIONS(1809),
    [anon_sym_LBRACE] = ACTIONS(1809),
    [anon_sym_SEMI] = ACTIONS(1809),
    [anon_sym_exception] = ACTIONS(1809),
    [anon_sym_module] = ACTIONS(1809),
    [anon_sym_open] = ACTIONS(1809),
    [anon_sym_include] = ACTIONS(1809),
    [anon_sym_class] = ACTIONS(1809),
    [anon_sym_val] = ACTIONS(1809),
    [anon_sym_end] = ACTIONS(1809),
    [anon_sym_object] = ACTIONS(1809),
    [anon_sym_AMP] = ACTIONS(1809),
    [anon_sym_POUND] = ACTIONS(1809),
    [anon_sym_COLON_COLON] = ACTIONS(1811),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1811),
    [anon_sym_else] = ACTIONS(1809),
    [anon_sym_new] = ACTIONS(1809),
    [anon_sym_LBRACE_LT] = ACTIONS(1811),
    [anon_sym_begin] = ACTIONS(1809),
    [sym_ocamlyacc_value] = ACTIONS(1811),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1809),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1811),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1809),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1809),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1811),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1811),
    [aux_sym_number_token1] = ACTIONS(1811),
    [aux_sym_signed_number_token1] = ACTIONS(1809),
    [anon_sym_SQUOTE] = ACTIONS(1811),
    [anon_sym_DQUOTE] = ACTIONS(1811),
    [sym_prefix_operator] = ACTIONS(1811),
    [aux_sym_sign_operator_token1] = ACTIONS(1809),
    [sym_hash_operator] = ACTIONS(1811),
    [anon_sym_lsl] = ACTIONS(1809),
    [anon_sym_lsr] = ACTIONS(1809),
    [anon_sym_asr] = ACTIONS(1809),
    [aux_sym_pow_operator_token1] = ACTIONS(1811),
    [anon_sym_mod] = ACTIONS(1809),
    [anon_sym_land] = ACTIONS(1809),
    [anon_sym_lor] = ACTIONS(1809),
    [anon_sym_lxor] = ACTIONS(1809),
    [aux_sym_mult_operator_token1] = ACTIONS(1809),
    [aux_sym_add_operator_token1] = ACTIONS(1809),
    [sym_concat_operator] = ACTIONS(1811),
    [sym_rel_operator] = ACTIONS(1809),
    [anon_sym_AMP_AMP] = ACTIONS(1809),
    [anon_sym_or] = ACTIONS(1809),
    [anon_sym_PIPE_PIPE] = ACTIONS(1809),
    [sym_assign_operator] = ACTIONS(1811),
    [sym_let_operator] = ACTIONS(1811),
    [sym__uppercase_identifier] = ACTIONS(1811),
    [aux_sym_directive_token1] = ACTIONS(1809),
    [aux_sym_tag_token1] = ACTIONS(1811),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1112)] = {
    [sym_attribute] = STATE(1112),
    [sym__lowercase_identifier] = ACTIONS(1583),
    [anon_sym_SEMI_SEMI] = ACTIONS(1585),
    [anon_sym_let] = ACTIONS(1583),
    [anon_sym_and] = ACTIONS(1583),
    [anon_sym_TILDE] = ACTIONS(1583),
    [anon_sym_QMARK] = ACTIONS(1583),
    [anon_sym_LPAREN] = ACTIONS(1585),
    [anon_sym_external] = ACTIONS(1583),
    [anon_sym_type] = ACTIONS(1583),
    [anon_sym_COMMA] = ACTIONS(1585),
    [anon_sym_PIPE] = ACTIONS(1583),
    [anon_sym_DOT] = ACTIONS(969),
    [anon_sym_LBRACK] = ACTIONS(1583),
    [anon_sym_RBRACK] = ACTIONS(1585),
    [anon_sym_true] = ACTIONS(1583),
    [anon_sym_false] = ACTIONS(1583),
    [anon_sym_LBRACE] = ACTIONS(1583),
    [anon_sym_SEMI] = ACTIONS(1583),
    [anon_sym_exception] = ACTIONS(1583),
    [anon_sym_module] = ACTIONS(1583),
    [anon_sym_open] = ACTIONS(1583),
    [anon_sym_include] = ACTIONS(1583),
    [anon_sym_class] = ACTIONS(1583),
    [anon_sym_val] = ACTIONS(1583),
    [anon_sym_end] = ACTIONS(1583),
    [anon_sym_object] = ACTIONS(1583),
    [anon_sym_AMP] = ACTIONS(1583),
    [anon_sym_POUND] = ACTIONS(1583),
    [anon_sym_COLON_COLON] = ACTIONS(1585),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1585),
    [anon_sym_else] = ACTIONS(1583),
    [anon_sym_new] = ACTIONS(1583),
    [anon_sym_LBRACE_LT] = ACTIONS(1585),
    [anon_sym_begin] = ACTIONS(1583),
    [sym_ocamlyacc_value] = ACTIONS(1585),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1583),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1585),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1583),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1583),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1585),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1585),
    [aux_sym_number_token1] = ACTIONS(1585),
    [aux_sym_signed_number_token1] = ACTIONS(1583),
    [anon_sym_SQUOTE] = ACTIONS(1585),
    [anon_sym_DQUOTE] = ACTIONS(1585),
    [sym_prefix_operator] = ACTIONS(1585),
    [aux_sym_sign_operator_token1] = ACTIONS(1583),
    [sym_hash_operator] = ACTIONS(1585),
    [anon_sym_lsl] = ACTIONS(1583),
    [anon_sym_lsr] = ACTIONS(1583),
    [anon_sym_asr] = ACTIONS(1583),
    [aux_sym_pow_operator_token1] = ACTIONS(1585),
    [anon_sym_mod] = ACTIONS(1583),
    [anon_sym_land] = ACTIONS(1583),
    [anon_sym_lor] = ACTIONS(1583),
    [anon_sym_lxor] = ACTIONS(1583),
    [aux_sym_mult_operator_token1] = ACTIONS(1583),
    [aux_sym_add_operator_token1] = ACTIONS(1583),
    [sym_concat_operator] = ACTIONS(1585),
    [sym_rel_operator] = ACTIONS(1583),
    [anon_sym_AMP_AMP] = ACTIONS(1583),
    [anon_sym_or] = ACTIONS(1583),
    [anon_sym_PIPE_PIPE] = ACTIONS(1583),
    [sym_assign_operator] = ACTIONS(1585),
    [sym_let_operator] = ACTIONS(1585),
    [sym__uppercase_identifier] = ACTIONS(1585),
    [aux_sym_directive_token1] = ACTIONS(1583),
    [aux_sym_tag_token1] = ACTIONS(1585),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1113)] = {
    [sym_attribute] = STATE(1113),
    [sym__lowercase_identifier] = ACTIONS(1813),
    [anon_sym_SEMI_SEMI] = ACTIONS(1815),
    [anon_sym_let] = ACTIONS(1813),
    [anon_sym_and] = ACTIONS(1813),
    [anon_sym_TILDE] = ACTIONS(1813),
    [anon_sym_QMARK] = ACTIONS(1813),
    [anon_sym_LPAREN] = ACTIONS(1815),
    [anon_sym_external] = ACTIONS(1813),
    [anon_sym_type] = ACTIONS(1813),
    [anon_sym_COMMA] = ACTIONS(1815),
    [anon_sym_PIPE] = ACTIONS(1813),
    [anon_sym_DOT] = ACTIONS(1815),
    [anon_sym_LBRACK] = ACTIONS(1813),
    [anon_sym_RBRACK] = ACTIONS(1815),
    [anon_sym_true] = ACTIONS(1813),
    [anon_sym_false] = ACTIONS(1813),
    [anon_sym_LBRACE] = ACTIONS(1813),
    [anon_sym_SEMI] = ACTIONS(1813),
    [anon_sym_exception] = ACTIONS(1813),
    [anon_sym_module] = ACTIONS(1813),
    [anon_sym_open] = ACTIONS(1813),
    [anon_sym_include] = ACTIONS(1813),
    [anon_sym_class] = ACTIONS(1813),
    [anon_sym_val] = ACTIONS(1813),
    [anon_sym_end] = ACTIONS(1813),
    [anon_sym_object] = ACTIONS(1813),
    [anon_sym_AMP] = ACTIONS(1813),
    [anon_sym_POUND] = ACTIONS(1813),
    [anon_sym_COLON_COLON] = ACTIONS(1815),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1815),
    [anon_sym_else] = ACTIONS(1813),
    [anon_sym_new] = ACTIONS(1813),
    [anon_sym_LBRACE_LT] = ACTIONS(1815),
    [anon_sym_begin] = ACTIONS(1813),
    [sym_ocamlyacc_value] = ACTIONS(1815),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1813),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1815),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1813),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1813),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1815),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1815),
    [aux_sym_number_token1] = ACTIONS(1815),
    [aux_sym_signed_number_token1] = ACTIONS(1813),
    [anon_sym_SQUOTE] = ACTIONS(1815),
    [anon_sym_DQUOTE] = ACTIONS(1815),
    [sym_prefix_operator] = ACTIONS(1815),
    [aux_sym_sign_operator_token1] = ACTIONS(1813),
    [sym_hash_operator] = ACTIONS(1815),
    [anon_sym_lsl] = ACTIONS(1813),
    [anon_sym_lsr] = ACTIONS(1813),
    [anon_sym_asr] = ACTIONS(1813),
    [aux_sym_pow_operator_token1] = ACTIONS(1815),
    [anon_sym_mod] = ACTIONS(1813),
    [anon_sym_land] = ACTIONS(1813),
    [anon_sym_lor] = ACTIONS(1813),
    [anon_sym_lxor] = ACTIONS(1813),
    [aux_sym_mult_operator_token1] = ACTIONS(1813),
    [aux_sym_add_operator_token1] = ACTIONS(1813),
    [sym_concat_operator] = ACTIONS(1815),
    [sym_rel_operator] = ACTIONS(1813),
    [anon_sym_AMP_AMP] = ACTIONS(1813),
    [anon_sym_or] = ACTIONS(1813),
    [anon_sym_PIPE_PIPE] = ACTIONS(1813),
    [sym_assign_operator] = ACTIONS(1815),
    [sym_let_operator] = ACTIONS(1815),
    [sym__uppercase_identifier] = ACTIONS(1815),
    [aux_sym_directive_token1] = ACTIONS(1813),
    [aux_sym_tag_token1] = ACTIONS(1815),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1114)] = {
    [sym_attribute] = STATE(1114),
    [sym__lowercase_identifier] = ACTIONS(1627),
    [anon_sym_SEMI_SEMI] = ACTIONS(1629),
    [anon_sym_let] = ACTIONS(1627),
    [anon_sym_and] = ACTIONS(1627),
    [anon_sym_TILDE] = ACTIONS(1627),
    [anon_sym_QMARK] = ACTIONS(1627),
    [anon_sym_LPAREN] = ACTIONS(1629),
    [anon_sym_external] = ACTIONS(1627),
    [anon_sym_type] = ACTIONS(1627),
    [anon_sym_COMMA] = ACTIONS(1629),
    [anon_sym_PIPE] = ACTIONS(1627),
    [anon_sym_DOT] = ACTIONS(1629),
    [anon_sym_LBRACK] = ACTIONS(1627),
    [anon_sym_RBRACK] = ACTIONS(1629),
    [anon_sym_true] = ACTIONS(1627),
    [anon_sym_false] = ACTIONS(1627),
    [anon_sym_LBRACE] = ACTIONS(1627),
    [anon_sym_SEMI] = ACTIONS(1627),
    [anon_sym_exception] = ACTIONS(1627),
    [anon_sym_module] = ACTIONS(1627),
    [anon_sym_open] = ACTIONS(1627),
    [anon_sym_include] = ACTIONS(1627),
    [anon_sym_class] = ACTIONS(1627),
    [anon_sym_val] = ACTIONS(1627),
    [anon_sym_end] = ACTIONS(1627),
    [anon_sym_object] = ACTIONS(1627),
    [anon_sym_AMP] = ACTIONS(1627),
    [anon_sym_POUND] = ACTIONS(1627),
    [anon_sym_COLON_COLON] = ACTIONS(1629),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1629),
    [anon_sym_else] = ACTIONS(1627),
    [anon_sym_new] = ACTIONS(1627),
    [anon_sym_LBRACE_LT] = ACTIONS(1629),
    [anon_sym_begin] = ACTIONS(1627),
    [sym_ocamlyacc_value] = ACTIONS(1629),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1627),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1629),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1627),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1627),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1629),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1629),
    [aux_sym_number_token1] = ACTIONS(1629),
    [aux_sym_signed_number_token1] = ACTIONS(1627),
    [anon_sym_SQUOTE] = ACTIONS(1629),
    [anon_sym_DQUOTE] = ACTIONS(1629),
    [sym_prefix_operator] = ACTIONS(1629),
    [aux_sym_sign_operator_token1] = ACTIONS(1627),
    [sym_hash_operator] = ACTIONS(1629),
    [anon_sym_lsl] = ACTIONS(1627),
    [anon_sym_lsr] = ACTIONS(1627),
    [anon_sym_asr] = ACTIONS(1627),
    [aux_sym_pow_operator_token1] = ACTIONS(1629),
    [anon_sym_mod] = ACTIONS(1627),
    [anon_sym_land] = ACTIONS(1627),
    [anon_sym_lor] = ACTIONS(1627),
    [anon_sym_lxor] = ACTIONS(1627),
    [aux_sym_mult_operator_token1] = ACTIONS(1627),
    [aux_sym_add_operator_token1] = ACTIONS(1627),
    [sym_concat_operator] = ACTIONS(1629),
    [sym_rel_operator] = ACTIONS(1627),
    [anon_sym_AMP_AMP] = ACTIONS(1627),
    [anon_sym_or] = ACTIONS(1627),
    [anon_sym_PIPE_PIPE] = ACTIONS(1627),
    [sym_assign_operator] = ACTIONS(1629),
    [sym_let_operator] = ACTIONS(1629),
    [sym__uppercase_identifier] = ACTIONS(1629),
    [aux_sym_directive_token1] = ACTIONS(1627),
    [aux_sym_tag_token1] = ACTIONS(1629),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1115)] = {
    [sym_attribute] = STATE(1115),
    [sym__lowercase_identifier] = ACTIONS(1631),
    [anon_sym_SEMI_SEMI] = ACTIONS(1633),
    [anon_sym_let] = ACTIONS(1631),
    [anon_sym_and] = ACTIONS(1631),
    [anon_sym_TILDE] = ACTIONS(1631),
    [anon_sym_QMARK] = ACTIONS(1631),
    [anon_sym_LPAREN] = ACTIONS(1633),
    [anon_sym_external] = ACTIONS(1631),
    [anon_sym_type] = ACTIONS(1631),
    [anon_sym_COMMA] = ACTIONS(1633),
    [anon_sym_PIPE] = ACTIONS(1631),
    [anon_sym_DOT] = ACTIONS(1633),
    [anon_sym_LBRACK] = ACTIONS(1631),
    [anon_sym_RBRACK] = ACTIONS(1633),
    [anon_sym_true] = ACTIONS(1631),
    [anon_sym_false] = ACTIONS(1631),
    [anon_sym_LBRACE] = ACTIONS(1631),
    [anon_sym_SEMI] = ACTIONS(1631),
    [anon_sym_exception] = ACTIONS(1631),
    [anon_sym_module] = ACTIONS(1631),
    [anon_sym_open] = ACTIONS(1631),
    [anon_sym_include] = ACTIONS(1631),
    [anon_sym_class] = ACTIONS(1631),
    [anon_sym_val] = ACTIONS(1631),
    [anon_sym_end] = ACTIONS(1631),
    [anon_sym_object] = ACTIONS(1631),
    [anon_sym_AMP] = ACTIONS(1631),
    [anon_sym_POUND] = ACTIONS(1631),
    [anon_sym_COLON_COLON] = ACTIONS(1633),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1633),
    [anon_sym_else] = ACTIONS(1631),
    [anon_sym_new] = ACTIONS(1631),
    [anon_sym_LBRACE_LT] = ACTIONS(1633),
    [anon_sym_begin] = ACTIONS(1631),
    [sym_ocamlyacc_value] = ACTIONS(1633),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1631),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1633),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1631),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1631),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1633),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1633),
    [aux_sym_number_token1] = ACTIONS(1633),
    [aux_sym_signed_number_token1] = ACTIONS(1631),
    [anon_sym_SQUOTE] = ACTIONS(1633),
    [anon_sym_DQUOTE] = ACTIONS(1633),
    [sym_prefix_operator] = ACTIONS(1633),
    [aux_sym_sign_operator_token1] = ACTIONS(1631),
    [sym_hash_operator] = ACTIONS(1633),
    [anon_sym_lsl] = ACTIONS(1631),
    [anon_sym_lsr] = ACTIONS(1631),
    [anon_sym_asr] = ACTIONS(1631),
    [aux_sym_pow_operator_token1] = ACTIONS(1633),
    [anon_sym_mod] = ACTIONS(1631),
    [anon_sym_land] = ACTIONS(1631),
    [anon_sym_lor] = ACTIONS(1631),
    [anon_sym_lxor] = ACTIONS(1631),
    [aux_sym_mult_operator_token1] = ACTIONS(1631),
    [aux_sym_add_operator_token1] = ACTIONS(1631),
    [sym_concat_operator] = ACTIONS(1633),
    [sym_rel_operator] = ACTIONS(1631),
    [anon_sym_AMP_AMP] = ACTIONS(1631),
    [anon_sym_or] = ACTIONS(1631),
    [anon_sym_PIPE_PIPE] = ACTIONS(1631),
    [sym_assign_operator] = ACTIONS(1633),
    [sym_let_operator] = ACTIONS(1633),
    [sym__uppercase_identifier] = ACTIONS(1633),
    [aux_sym_directive_token1] = ACTIONS(1631),
    [aux_sym_tag_token1] = ACTIONS(1633),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1116)] = {
    [sym_attribute] = STATE(1116),
    [sym__lowercase_identifier] = ACTIONS(1623),
    [anon_sym_SEMI_SEMI] = ACTIONS(1625),
    [anon_sym_let] = ACTIONS(1623),
    [anon_sym_and] = ACTIONS(1623),
    [anon_sym_TILDE] = ACTIONS(1623),
    [anon_sym_QMARK] = ACTIONS(1623),
    [anon_sym_LPAREN] = ACTIONS(1625),
    [anon_sym_external] = ACTIONS(1623),
    [anon_sym_type] = ACTIONS(1623),
    [anon_sym_COMMA] = ACTIONS(1625),
    [anon_sym_PIPE] = ACTIONS(1623),
    [anon_sym_DOT] = ACTIONS(1625),
    [anon_sym_LBRACK] = ACTIONS(1623),
    [anon_sym_RBRACK] = ACTIONS(1625),
    [anon_sym_true] = ACTIONS(1623),
    [anon_sym_false] = ACTIONS(1623),
    [anon_sym_LBRACE] = ACTIONS(1623),
    [anon_sym_SEMI] = ACTIONS(1623),
    [anon_sym_exception] = ACTIONS(1623),
    [anon_sym_module] = ACTIONS(1623),
    [anon_sym_open] = ACTIONS(1623),
    [anon_sym_include] = ACTIONS(1623),
    [anon_sym_class] = ACTIONS(1623),
    [anon_sym_val] = ACTIONS(1623),
    [anon_sym_end] = ACTIONS(1623),
    [anon_sym_object] = ACTIONS(1623),
    [anon_sym_AMP] = ACTIONS(1623),
    [anon_sym_POUND] = ACTIONS(1623),
    [anon_sym_COLON_COLON] = ACTIONS(1625),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1625),
    [anon_sym_else] = ACTIONS(1623),
    [anon_sym_new] = ACTIONS(1623),
    [anon_sym_LBRACE_LT] = ACTIONS(1625),
    [anon_sym_begin] = ACTIONS(1623),
    [sym_ocamlyacc_value] = ACTIONS(1625),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1623),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1625),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1623),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1623),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1625),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1625),
    [aux_sym_number_token1] = ACTIONS(1625),
    [aux_sym_signed_number_token1] = ACTIONS(1623),
    [anon_sym_SQUOTE] = ACTIONS(1625),
    [anon_sym_DQUOTE] = ACTIONS(1625),
    [sym_prefix_operator] = ACTIONS(1625),
    [aux_sym_sign_operator_token1] = ACTIONS(1623),
    [sym_hash_operator] = ACTIONS(1625),
    [anon_sym_lsl] = ACTIONS(1623),
    [anon_sym_lsr] = ACTIONS(1623),
    [anon_sym_asr] = ACTIONS(1623),
    [aux_sym_pow_operator_token1] = ACTIONS(1625),
    [anon_sym_mod] = ACTIONS(1623),
    [anon_sym_land] = ACTIONS(1623),
    [anon_sym_lor] = ACTIONS(1623),
    [anon_sym_lxor] = ACTIONS(1623),
    [aux_sym_mult_operator_token1] = ACTIONS(1623),
    [aux_sym_add_operator_token1] = ACTIONS(1623),
    [sym_concat_operator] = ACTIONS(1625),
    [sym_rel_operator] = ACTIONS(1623),
    [anon_sym_AMP_AMP] = ACTIONS(1623),
    [anon_sym_or] = ACTIONS(1623),
    [anon_sym_PIPE_PIPE] = ACTIONS(1623),
    [sym_assign_operator] = ACTIONS(1625),
    [sym_let_operator] = ACTIONS(1625),
    [sym__uppercase_identifier] = ACTIONS(1625),
    [aux_sym_directive_token1] = ACTIONS(1623),
    [aux_sym_tag_token1] = ACTIONS(1625),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1117)] = {
    [sym_attribute] = STATE(1117),
    [sym__lowercase_identifier] = ACTIONS(1635),
    [anon_sym_SEMI_SEMI] = ACTIONS(1637),
    [anon_sym_let] = ACTIONS(1635),
    [anon_sym_and] = ACTIONS(1635),
    [anon_sym_TILDE] = ACTIONS(1635),
    [anon_sym_QMARK] = ACTIONS(1635),
    [anon_sym_LPAREN] = ACTIONS(1637),
    [anon_sym_external] = ACTIONS(1635),
    [anon_sym_type] = ACTIONS(1635),
    [anon_sym_COMMA] = ACTIONS(1637),
    [anon_sym_PIPE] = ACTIONS(1635),
    [anon_sym_DOT] = ACTIONS(1637),
    [anon_sym_LBRACK] = ACTIONS(1635),
    [anon_sym_RBRACK] = ACTIONS(1637),
    [anon_sym_true] = ACTIONS(1635),
    [anon_sym_false] = ACTIONS(1635),
    [anon_sym_LBRACE] = ACTIONS(1635),
    [anon_sym_SEMI] = ACTIONS(1635),
    [anon_sym_exception] = ACTIONS(1635),
    [anon_sym_module] = ACTIONS(1635),
    [anon_sym_open] = ACTIONS(1635),
    [anon_sym_include] = ACTIONS(1635),
    [anon_sym_class] = ACTIONS(1635),
    [anon_sym_val] = ACTIONS(1635),
    [anon_sym_end] = ACTIONS(1635),
    [anon_sym_object] = ACTIONS(1635),
    [anon_sym_AMP] = ACTIONS(1635),
    [anon_sym_POUND] = ACTIONS(1635),
    [anon_sym_COLON_COLON] = ACTIONS(1637),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1637),
    [anon_sym_else] = ACTIONS(1635),
    [anon_sym_new] = ACTIONS(1635),
    [anon_sym_LBRACE_LT] = ACTIONS(1637),
    [anon_sym_begin] = ACTIONS(1635),
    [sym_ocamlyacc_value] = ACTIONS(1637),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1635),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1637),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1635),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1635),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1637),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1637),
    [aux_sym_number_token1] = ACTIONS(1637),
    [aux_sym_signed_number_token1] = ACTIONS(1635),
    [anon_sym_SQUOTE] = ACTIONS(1637),
    [anon_sym_DQUOTE] = ACTIONS(1637),
    [sym_prefix_operator] = ACTIONS(1637),
    [aux_sym_sign_operator_token1] = ACTIONS(1635),
    [sym_hash_operator] = ACTIONS(1637),
    [anon_sym_lsl] = ACTIONS(1635),
    [anon_sym_lsr] = ACTIONS(1635),
    [anon_sym_asr] = ACTIONS(1635),
    [aux_sym_pow_operator_token1] = ACTIONS(1637),
    [anon_sym_mod] = ACTIONS(1635),
    [anon_sym_land] = ACTIONS(1635),
    [anon_sym_lor] = ACTIONS(1635),
    [anon_sym_lxor] = ACTIONS(1635),
    [aux_sym_mult_operator_token1] = ACTIONS(1635),
    [aux_sym_add_operator_token1] = ACTIONS(1635),
    [sym_concat_operator] = ACTIONS(1637),
    [sym_rel_operator] = ACTIONS(1635),
    [anon_sym_AMP_AMP] = ACTIONS(1635),
    [anon_sym_or] = ACTIONS(1635),
    [anon_sym_PIPE_PIPE] = ACTIONS(1635),
    [sym_assign_operator] = ACTIONS(1637),
    [sym_let_operator] = ACTIONS(1637),
    [sym__uppercase_identifier] = ACTIONS(1637),
    [aux_sym_directive_token1] = ACTIONS(1635),
    [aux_sym_tag_token1] = ACTIONS(1637),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1118)] = {
    [sym_attribute] = STATE(1118),
    [sym__lowercase_identifier] = ACTIONS(1667),
    [anon_sym_SEMI_SEMI] = ACTIONS(1669),
    [anon_sym_let] = ACTIONS(1667),
    [anon_sym_and] = ACTIONS(1667),
    [anon_sym_TILDE] = ACTIONS(1667),
    [anon_sym_QMARK] = ACTIONS(1667),
    [anon_sym_LPAREN] = ACTIONS(1669),
    [anon_sym_external] = ACTIONS(1667),
    [anon_sym_type] = ACTIONS(1667),
    [anon_sym_COMMA] = ACTIONS(1669),
    [anon_sym_PIPE] = ACTIONS(1667),
    [anon_sym_DOT] = ACTIONS(1669),
    [anon_sym_LBRACK] = ACTIONS(1667),
    [anon_sym_RBRACK] = ACTIONS(1669),
    [anon_sym_true] = ACTIONS(1667),
    [anon_sym_false] = ACTIONS(1667),
    [anon_sym_LBRACE] = ACTIONS(1667),
    [anon_sym_SEMI] = ACTIONS(1667),
    [anon_sym_exception] = ACTIONS(1667),
    [anon_sym_module] = ACTIONS(1667),
    [anon_sym_open] = ACTIONS(1667),
    [anon_sym_include] = ACTIONS(1667),
    [anon_sym_class] = ACTIONS(1667),
    [anon_sym_val] = ACTIONS(1667),
    [anon_sym_end] = ACTIONS(1667),
    [anon_sym_object] = ACTIONS(1667),
    [anon_sym_AMP] = ACTIONS(1667),
    [anon_sym_POUND] = ACTIONS(1667),
    [anon_sym_COLON_COLON] = ACTIONS(1669),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1669),
    [anon_sym_else] = ACTIONS(1667),
    [anon_sym_new] = ACTIONS(1667),
    [anon_sym_LBRACE_LT] = ACTIONS(1669),
    [anon_sym_begin] = ACTIONS(1667),
    [sym_ocamlyacc_value] = ACTIONS(1669),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1667),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1669),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1667),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1667),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1669),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1669),
    [aux_sym_number_token1] = ACTIONS(1669),
    [aux_sym_signed_number_token1] = ACTIONS(1667),
    [anon_sym_SQUOTE] = ACTIONS(1669),
    [anon_sym_DQUOTE] = ACTIONS(1669),
    [sym_prefix_operator] = ACTIONS(1669),
    [aux_sym_sign_operator_token1] = ACTIONS(1667),
    [sym_hash_operator] = ACTIONS(1669),
    [anon_sym_lsl] = ACTIONS(1667),
    [anon_sym_lsr] = ACTIONS(1667),
    [anon_sym_asr] = ACTIONS(1667),
    [aux_sym_pow_operator_token1] = ACTIONS(1669),
    [anon_sym_mod] = ACTIONS(1667),
    [anon_sym_land] = ACTIONS(1667),
    [anon_sym_lor] = ACTIONS(1667),
    [anon_sym_lxor] = ACTIONS(1667),
    [aux_sym_mult_operator_token1] = ACTIONS(1667),
    [aux_sym_add_operator_token1] = ACTIONS(1667),
    [sym_concat_operator] = ACTIONS(1669),
    [sym_rel_operator] = ACTIONS(1667),
    [anon_sym_AMP_AMP] = ACTIONS(1667),
    [anon_sym_or] = ACTIONS(1667),
    [anon_sym_PIPE_PIPE] = ACTIONS(1667),
    [sym_assign_operator] = ACTIONS(1669),
    [sym_let_operator] = ACTIONS(1669),
    [sym__uppercase_identifier] = ACTIONS(1669),
    [aux_sym_directive_token1] = ACTIONS(1667),
    [aux_sym_tag_token1] = ACTIONS(1669),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1119)] = {
    [sym_attribute] = STATE(1119),
    [sym__lowercase_identifier] = ACTIONS(1699),
    [anon_sym_SEMI_SEMI] = ACTIONS(1701),
    [anon_sym_let] = ACTIONS(1699),
    [anon_sym_and] = ACTIONS(1699),
    [anon_sym_TILDE] = ACTIONS(1699),
    [anon_sym_QMARK] = ACTIONS(1699),
    [anon_sym_LPAREN] = ACTIONS(1701),
    [anon_sym_external] = ACTIONS(1699),
    [anon_sym_type] = ACTIONS(1699),
    [anon_sym_COMMA] = ACTIONS(1701),
    [anon_sym_PIPE] = ACTIONS(1699),
    [anon_sym_DOT] = ACTIONS(1841),
    [anon_sym_LBRACK] = ACTIONS(1699),
    [anon_sym_RBRACK] = ACTIONS(1701),
    [anon_sym_true] = ACTIONS(1699),
    [anon_sym_false] = ACTIONS(1699),
    [anon_sym_LBRACE] = ACTIONS(1699),
    [anon_sym_SEMI] = ACTIONS(1699),
    [anon_sym_exception] = ACTIONS(1699),
    [anon_sym_module] = ACTIONS(1699),
    [anon_sym_open] = ACTIONS(1699),
    [anon_sym_include] = ACTIONS(1699),
    [anon_sym_class] = ACTIONS(1699),
    [anon_sym_val] = ACTIONS(1699),
    [anon_sym_end] = ACTIONS(1699),
    [anon_sym_object] = ACTIONS(1699),
    [anon_sym_AMP] = ACTIONS(1699),
    [anon_sym_POUND] = ACTIONS(1699),
    [anon_sym_COLON_COLON] = ACTIONS(1701),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1701),
    [anon_sym_else] = ACTIONS(1699),
    [anon_sym_new] = ACTIONS(1699),
    [anon_sym_LBRACE_LT] = ACTIONS(1701),
    [anon_sym_begin] = ACTIONS(1699),
    [sym_ocamlyacc_value] = ACTIONS(1701),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1699),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1701),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1699),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1699),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1701),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1701),
    [aux_sym_number_token1] = ACTIONS(1701),
    [aux_sym_signed_number_token1] = ACTIONS(1699),
    [anon_sym_SQUOTE] = ACTIONS(1701),
    [anon_sym_DQUOTE] = ACTIONS(1701),
    [sym_prefix_operator] = ACTIONS(1701),
    [aux_sym_sign_operator_token1] = ACTIONS(1699),
    [sym_hash_operator] = ACTIONS(1701),
    [anon_sym_lsl] = ACTIONS(1699),
    [anon_sym_lsr] = ACTIONS(1699),
    [anon_sym_asr] = ACTIONS(1699),
    [aux_sym_pow_operator_token1] = ACTIONS(1701),
    [anon_sym_mod] = ACTIONS(1699),
    [anon_sym_land] = ACTIONS(1699),
    [anon_sym_lor] = ACTIONS(1699),
    [anon_sym_lxor] = ACTIONS(1699),
    [aux_sym_mult_operator_token1] = ACTIONS(1699),
    [aux_sym_add_operator_token1] = ACTIONS(1699),
    [sym_concat_operator] = ACTIONS(1701),
    [sym_rel_operator] = ACTIONS(1699),
    [anon_sym_AMP_AMP] = ACTIONS(1699),
    [anon_sym_or] = ACTIONS(1699),
    [anon_sym_PIPE_PIPE] = ACTIONS(1699),
    [sym_assign_operator] = ACTIONS(1701),
    [sym_let_operator] = ACTIONS(1701),
    [sym__uppercase_identifier] = ACTIONS(1701),
    [aux_sym_directive_token1] = ACTIONS(1699),
    [aux_sym_tag_token1] = ACTIONS(1701),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1120)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4565),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6673),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1120),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(4808),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1038),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1120),
    [sym__lowercase_identifier] = ACTIONS(2180),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_COMMA] = ACTIONS(1295),
    [anon_sym_PIPE] = ACTIONS(1295),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_RBRACK] = ACTIONS(1295),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_SEMI] = ACTIONS(1295),
    [anon_sym_RBRACE] = ACTIONS(1295),
    [anon_sym_exception] = ACTIONS(2184),
    [anon_sym_as] = ACTIONS(1297),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_COLON_COLON] = ACTIONS(1295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(2188),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2190),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1121)] = {
    [sym_attribute] = STATE(1121),
    [sym__lowercase_identifier] = ACTIONS(1717),
    [anon_sym_SEMI_SEMI] = ACTIONS(1719),
    [anon_sym_let] = ACTIONS(1717),
    [anon_sym_and] = ACTIONS(1717),
    [anon_sym_TILDE] = ACTIONS(1717),
    [anon_sym_QMARK] = ACTIONS(1717),
    [anon_sym_LPAREN] = ACTIONS(1719),
    [anon_sym_external] = ACTIONS(1717),
    [anon_sym_type] = ACTIONS(1717),
    [anon_sym_COMMA] = ACTIONS(1719),
    [anon_sym_PIPE] = ACTIONS(1717),
    [anon_sym_DOT] = ACTIONS(1719),
    [anon_sym_LBRACK] = ACTIONS(1717),
    [anon_sym_RBRACK] = ACTIONS(1719),
    [anon_sym_true] = ACTIONS(1717),
    [anon_sym_false] = ACTIONS(1717),
    [anon_sym_LBRACE] = ACTIONS(1717),
    [anon_sym_SEMI] = ACTIONS(1717),
    [anon_sym_exception] = ACTIONS(1717),
    [anon_sym_module] = ACTIONS(1717),
    [anon_sym_open] = ACTIONS(1717),
    [anon_sym_include] = ACTIONS(1717),
    [anon_sym_class] = ACTIONS(1717),
    [anon_sym_val] = ACTIONS(1717),
    [anon_sym_end] = ACTIONS(1717),
    [anon_sym_object] = ACTIONS(1717),
    [anon_sym_AMP] = ACTIONS(1717),
    [anon_sym_POUND] = ACTIONS(1717),
    [anon_sym_COLON_COLON] = ACTIONS(1719),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1719),
    [anon_sym_else] = ACTIONS(1717),
    [anon_sym_new] = ACTIONS(1717),
    [anon_sym_LBRACE_LT] = ACTIONS(1719),
    [anon_sym_begin] = ACTIONS(1717),
    [sym_ocamlyacc_value] = ACTIONS(1719),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1717),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1719),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1717),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1717),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1719),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1719),
    [aux_sym_number_token1] = ACTIONS(1719),
    [aux_sym_signed_number_token1] = ACTIONS(1717),
    [anon_sym_SQUOTE] = ACTIONS(1719),
    [anon_sym_DQUOTE] = ACTIONS(1719),
    [sym_prefix_operator] = ACTIONS(1719),
    [aux_sym_sign_operator_token1] = ACTIONS(1717),
    [sym_hash_operator] = ACTIONS(1719),
    [anon_sym_lsl] = ACTIONS(1717),
    [anon_sym_lsr] = ACTIONS(1717),
    [anon_sym_asr] = ACTIONS(1717),
    [aux_sym_pow_operator_token1] = ACTIONS(1719),
    [anon_sym_mod] = ACTIONS(1717),
    [anon_sym_land] = ACTIONS(1717),
    [anon_sym_lor] = ACTIONS(1717),
    [anon_sym_lxor] = ACTIONS(1717),
    [aux_sym_mult_operator_token1] = ACTIONS(1717),
    [aux_sym_add_operator_token1] = ACTIONS(1717),
    [sym_concat_operator] = ACTIONS(1719),
    [sym_rel_operator] = ACTIONS(1717),
    [anon_sym_AMP_AMP] = ACTIONS(1717),
    [anon_sym_or] = ACTIONS(1717),
    [anon_sym_PIPE_PIPE] = ACTIONS(1717),
    [sym_assign_operator] = ACTIONS(1719),
    [sym_let_operator] = ACTIONS(1719),
    [sym__uppercase_identifier] = ACTIONS(1719),
    [aux_sym_directive_token1] = ACTIONS(1717),
    [aux_sym_tag_token1] = ACTIONS(1719),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1122)] = {
    [sym_attribute] = STATE(1122),
    [sym__lowercase_identifier] = ACTIONS(1825),
    [anon_sym_SEMI_SEMI] = ACTIONS(1827),
    [anon_sym_let] = ACTIONS(1825),
    [anon_sym_and] = ACTIONS(1825),
    [anon_sym_TILDE] = ACTIONS(1825),
    [anon_sym_QMARK] = ACTIONS(1825),
    [anon_sym_LPAREN] = ACTIONS(1827),
    [anon_sym_external] = ACTIONS(1825),
    [anon_sym_type] = ACTIONS(1825),
    [anon_sym_COMMA] = ACTIONS(1827),
    [anon_sym_PIPE] = ACTIONS(1825),
    [anon_sym_DOT] = ACTIONS(1827),
    [anon_sym_LBRACK] = ACTIONS(1825),
    [anon_sym_RBRACK] = ACTIONS(1827),
    [anon_sym_true] = ACTIONS(1825),
    [anon_sym_false] = ACTIONS(1825),
    [anon_sym_LBRACE] = ACTIONS(1825),
    [anon_sym_SEMI] = ACTIONS(1825),
    [anon_sym_exception] = ACTIONS(1825),
    [anon_sym_module] = ACTIONS(1825),
    [anon_sym_open] = ACTIONS(1825),
    [anon_sym_include] = ACTIONS(1825),
    [anon_sym_class] = ACTIONS(1825),
    [anon_sym_val] = ACTIONS(1825),
    [anon_sym_end] = ACTIONS(1825),
    [anon_sym_object] = ACTIONS(1825),
    [anon_sym_AMP] = ACTIONS(1825),
    [anon_sym_POUND] = ACTIONS(1825),
    [anon_sym_COLON_COLON] = ACTIONS(1827),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1827),
    [anon_sym_else] = ACTIONS(1825),
    [anon_sym_new] = ACTIONS(1825),
    [anon_sym_LBRACE_LT] = ACTIONS(1827),
    [anon_sym_begin] = ACTIONS(1825),
    [sym_ocamlyacc_value] = ACTIONS(1827),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1825),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1827),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1825),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1825),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1827),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1827),
    [aux_sym_number_token1] = ACTIONS(1827),
    [aux_sym_signed_number_token1] = ACTIONS(1825),
    [anon_sym_SQUOTE] = ACTIONS(1827),
    [anon_sym_DQUOTE] = ACTIONS(1827),
    [sym_prefix_operator] = ACTIONS(1827),
    [aux_sym_sign_operator_token1] = ACTIONS(1825),
    [sym_hash_operator] = ACTIONS(1827),
    [anon_sym_lsl] = ACTIONS(1825),
    [anon_sym_lsr] = ACTIONS(1825),
    [anon_sym_asr] = ACTIONS(1825),
    [aux_sym_pow_operator_token1] = ACTIONS(1827),
    [anon_sym_mod] = ACTIONS(1825),
    [anon_sym_land] = ACTIONS(1825),
    [anon_sym_lor] = ACTIONS(1825),
    [anon_sym_lxor] = ACTIONS(1825),
    [aux_sym_mult_operator_token1] = ACTIONS(1825),
    [aux_sym_add_operator_token1] = ACTIONS(1825),
    [sym_concat_operator] = ACTIONS(1827),
    [sym_rel_operator] = ACTIONS(1825),
    [anon_sym_AMP_AMP] = ACTIONS(1825),
    [anon_sym_or] = ACTIONS(1825),
    [anon_sym_PIPE_PIPE] = ACTIONS(1825),
    [sym_assign_operator] = ACTIONS(1827),
    [sym_let_operator] = ACTIONS(1827),
    [sym__uppercase_identifier] = ACTIONS(1827),
    [aux_sym_directive_token1] = ACTIONS(1825),
    [aux_sym_tag_token1] = ACTIONS(1827),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1123)] = {
    [sym__parenthesized_abstract_type] = STATE(1337),
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4560),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1123),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(2182),
    [anon_sym_RPAREN] = ACTIONS(1295),
    [anon_sym_COMMA] = ACTIONS(1295),
    [anon_sym_PIPE] = ACTIONS(1295),
    [anon_sym_COLON2] = ACTIONS(1297),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_as] = ACTIONS(1297),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(1295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1124)] = {
    [sym_attribute] = STATE(1124),
    [sym__lowercase_identifier] = ACTIONS(1725),
    [anon_sym_SEMI_SEMI] = ACTIONS(1727),
    [anon_sym_let] = ACTIONS(1725),
    [anon_sym_and] = ACTIONS(1725),
    [anon_sym_TILDE] = ACTIONS(1725),
    [anon_sym_QMARK] = ACTIONS(1725),
    [anon_sym_LPAREN] = ACTIONS(1727),
    [anon_sym_external] = ACTIONS(1725),
    [anon_sym_type] = ACTIONS(1725),
    [anon_sym_COMMA] = ACTIONS(1727),
    [anon_sym_PIPE] = ACTIONS(1725),
    [anon_sym_DOT] = ACTIONS(1727),
    [anon_sym_LBRACK] = ACTIONS(1725),
    [anon_sym_RBRACK] = ACTIONS(1727),
    [anon_sym_true] = ACTIONS(1725),
    [anon_sym_false] = ACTIONS(1725),
    [anon_sym_LBRACE] = ACTIONS(1725),
    [anon_sym_SEMI] = ACTIONS(1725),
    [anon_sym_exception] = ACTIONS(1725),
    [anon_sym_module] = ACTIONS(1725),
    [anon_sym_open] = ACTIONS(1725),
    [anon_sym_include] = ACTIONS(1725),
    [anon_sym_class] = ACTIONS(1725),
    [anon_sym_val] = ACTIONS(1725),
    [anon_sym_end] = ACTIONS(1725),
    [anon_sym_object] = ACTIONS(1725),
    [anon_sym_AMP] = ACTIONS(1725),
    [anon_sym_POUND] = ACTIONS(1725),
    [anon_sym_COLON_COLON] = ACTIONS(1727),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1727),
    [anon_sym_else] = ACTIONS(1725),
    [anon_sym_new] = ACTIONS(1725),
    [anon_sym_LBRACE_LT] = ACTIONS(1727),
    [anon_sym_begin] = ACTIONS(1725),
    [sym_ocamlyacc_value] = ACTIONS(1727),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1725),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1727),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1725),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1725),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1727),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1727),
    [aux_sym_number_token1] = ACTIONS(1727),
    [aux_sym_signed_number_token1] = ACTIONS(1725),
    [anon_sym_SQUOTE] = ACTIONS(1727),
    [anon_sym_DQUOTE] = ACTIONS(1727),
    [sym_prefix_operator] = ACTIONS(1727),
    [aux_sym_sign_operator_token1] = ACTIONS(1725),
    [sym_hash_operator] = ACTIONS(1727),
    [anon_sym_lsl] = ACTIONS(1725),
    [anon_sym_lsr] = ACTIONS(1725),
    [anon_sym_asr] = ACTIONS(1725),
    [aux_sym_pow_operator_token1] = ACTIONS(1727),
    [anon_sym_mod] = ACTIONS(1725),
    [anon_sym_land] = ACTIONS(1725),
    [anon_sym_lor] = ACTIONS(1725),
    [anon_sym_lxor] = ACTIONS(1725),
    [aux_sym_mult_operator_token1] = ACTIONS(1725),
    [aux_sym_add_operator_token1] = ACTIONS(1725),
    [sym_concat_operator] = ACTIONS(1727),
    [sym_rel_operator] = ACTIONS(1725),
    [anon_sym_AMP_AMP] = ACTIONS(1725),
    [anon_sym_or] = ACTIONS(1725),
    [anon_sym_PIPE_PIPE] = ACTIONS(1725),
    [sym_assign_operator] = ACTIONS(1727),
    [sym_let_operator] = ACTIONS(1727),
    [sym__uppercase_identifier] = ACTIONS(1727),
    [aux_sym_directive_token1] = ACTIONS(1725),
    [aux_sym_tag_token1] = ACTIONS(1727),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1125)] = {
    [sym_attribute] = STATE(1125),
    [sym__lowercase_identifier] = ACTIONS(1745),
    [anon_sym_SEMI_SEMI] = ACTIONS(1747),
    [anon_sym_let] = ACTIONS(1745),
    [anon_sym_and] = ACTIONS(1745),
    [anon_sym_TILDE] = ACTIONS(1745),
    [anon_sym_QMARK] = ACTIONS(1745),
    [anon_sym_LPAREN] = ACTIONS(1747),
    [anon_sym_external] = ACTIONS(1745),
    [anon_sym_type] = ACTIONS(1745),
    [anon_sym_COMMA] = ACTIONS(1747),
    [anon_sym_PIPE] = ACTIONS(1745),
    [anon_sym_DOT] = ACTIONS(1747),
    [anon_sym_LBRACK] = ACTIONS(1745),
    [anon_sym_RBRACK] = ACTIONS(1747),
    [anon_sym_true] = ACTIONS(1745),
    [anon_sym_false] = ACTIONS(1745),
    [anon_sym_LBRACE] = ACTIONS(1745),
    [anon_sym_SEMI] = ACTIONS(1745),
    [anon_sym_exception] = ACTIONS(1745),
    [anon_sym_module] = ACTIONS(1745),
    [anon_sym_open] = ACTIONS(1745),
    [anon_sym_include] = ACTIONS(1745),
    [anon_sym_class] = ACTIONS(1745),
    [anon_sym_val] = ACTIONS(1745),
    [anon_sym_end] = ACTIONS(1745),
    [anon_sym_object] = ACTIONS(1745),
    [anon_sym_AMP] = ACTIONS(1745),
    [anon_sym_POUND] = ACTIONS(1745),
    [anon_sym_COLON_COLON] = ACTIONS(1747),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1747),
    [anon_sym_else] = ACTIONS(1745),
    [anon_sym_new] = ACTIONS(1745),
    [anon_sym_LBRACE_LT] = ACTIONS(1747),
    [anon_sym_begin] = ACTIONS(1745),
    [sym_ocamlyacc_value] = ACTIONS(1747),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1745),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1747),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1745),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1745),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1747),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1747),
    [aux_sym_number_token1] = ACTIONS(1747),
    [aux_sym_signed_number_token1] = ACTIONS(1745),
    [anon_sym_SQUOTE] = ACTIONS(1747),
    [anon_sym_DQUOTE] = ACTIONS(1747),
    [sym_prefix_operator] = ACTIONS(1747),
    [aux_sym_sign_operator_token1] = ACTIONS(1745),
    [sym_hash_operator] = ACTIONS(1747),
    [anon_sym_lsl] = ACTIONS(1745),
    [anon_sym_lsr] = ACTIONS(1745),
    [anon_sym_asr] = ACTIONS(1745),
    [aux_sym_pow_operator_token1] = ACTIONS(1747),
    [anon_sym_mod] = ACTIONS(1745),
    [anon_sym_land] = ACTIONS(1745),
    [anon_sym_lor] = ACTIONS(1745),
    [anon_sym_lxor] = ACTIONS(1745),
    [aux_sym_mult_operator_token1] = ACTIONS(1745),
    [aux_sym_add_operator_token1] = ACTIONS(1745),
    [sym_concat_operator] = ACTIONS(1747),
    [sym_rel_operator] = ACTIONS(1745),
    [anon_sym_AMP_AMP] = ACTIONS(1745),
    [anon_sym_or] = ACTIONS(1745),
    [anon_sym_PIPE_PIPE] = ACTIONS(1745),
    [sym_assign_operator] = ACTIONS(1747),
    [sym_let_operator] = ACTIONS(1747),
    [sym__uppercase_identifier] = ACTIONS(1747),
    [aux_sym_directive_token1] = ACTIONS(1745),
    [aux_sym_tag_token1] = ACTIONS(1747),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1126)] = {
    [sym_attribute] = STATE(1126),
    [sym__lowercase_identifier] = ACTIONS(1785),
    [anon_sym_SEMI_SEMI] = ACTIONS(1787),
    [anon_sym_let] = ACTIONS(1785),
    [anon_sym_and] = ACTIONS(1785),
    [anon_sym_TILDE] = ACTIONS(1785),
    [anon_sym_QMARK] = ACTIONS(1785),
    [anon_sym_LPAREN] = ACTIONS(1787),
    [anon_sym_external] = ACTIONS(1785),
    [anon_sym_type] = ACTIONS(1785),
    [anon_sym_COMMA] = ACTIONS(1787),
    [anon_sym_PIPE] = ACTIONS(1785),
    [anon_sym_DOT] = ACTIONS(1787),
    [anon_sym_LBRACK] = ACTIONS(1785),
    [anon_sym_RBRACK] = ACTIONS(1787),
    [anon_sym_true] = ACTIONS(1785),
    [anon_sym_false] = ACTIONS(1785),
    [anon_sym_LBRACE] = ACTIONS(1785),
    [anon_sym_SEMI] = ACTIONS(1785),
    [anon_sym_exception] = ACTIONS(1785),
    [anon_sym_module] = ACTIONS(1785),
    [anon_sym_open] = ACTIONS(1785),
    [anon_sym_include] = ACTIONS(1785),
    [anon_sym_class] = ACTIONS(1785),
    [anon_sym_val] = ACTIONS(1785),
    [anon_sym_end] = ACTIONS(1785),
    [anon_sym_object] = ACTIONS(1785),
    [anon_sym_AMP] = ACTIONS(1785),
    [anon_sym_POUND] = ACTIONS(1785),
    [anon_sym_COLON_COLON] = ACTIONS(1787),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1787),
    [anon_sym_else] = ACTIONS(1785),
    [anon_sym_new] = ACTIONS(1785),
    [anon_sym_LBRACE_LT] = ACTIONS(1787),
    [anon_sym_begin] = ACTIONS(1785),
    [sym_ocamlyacc_value] = ACTIONS(1787),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1785),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1787),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1785),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1785),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1787),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1787),
    [aux_sym_number_token1] = ACTIONS(1787),
    [aux_sym_signed_number_token1] = ACTIONS(1785),
    [anon_sym_SQUOTE] = ACTIONS(1787),
    [anon_sym_DQUOTE] = ACTIONS(1787),
    [sym_prefix_operator] = ACTIONS(1787),
    [aux_sym_sign_operator_token1] = ACTIONS(1785),
    [sym_hash_operator] = ACTIONS(1787),
    [anon_sym_lsl] = ACTIONS(1785),
    [anon_sym_lsr] = ACTIONS(1785),
    [anon_sym_asr] = ACTIONS(1785),
    [aux_sym_pow_operator_token1] = ACTIONS(1787),
    [anon_sym_mod] = ACTIONS(1785),
    [anon_sym_land] = ACTIONS(1785),
    [anon_sym_lor] = ACTIONS(1785),
    [anon_sym_lxor] = ACTIONS(1785),
    [aux_sym_mult_operator_token1] = ACTIONS(1785),
    [aux_sym_add_operator_token1] = ACTIONS(1785),
    [sym_concat_operator] = ACTIONS(1787),
    [sym_rel_operator] = ACTIONS(1785),
    [anon_sym_AMP_AMP] = ACTIONS(1785),
    [anon_sym_or] = ACTIONS(1785),
    [anon_sym_PIPE_PIPE] = ACTIONS(1785),
    [sym_assign_operator] = ACTIONS(1787),
    [sym_let_operator] = ACTIONS(1787),
    [sym__uppercase_identifier] = ACTIONS(1787),
    [aux_sym_directive_token1] = ACTIONS(1785),
    [aux_sym_tag_token1] = ACTIONS(1787),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1127)] = {
    [sym__delimited_pattern] = STATE(4970),
    [sym__simple_pattern] = STATE(4826),
    [sym__effect_pattern] = STATE(4830),
    [sym__pattern] = STATE(4879),
    [sym_alias_pattern] = STATE(4830),
    [sym_typed_pattern] = STATE(4970),
    [sym__or_pattern_anonymous] = STATE(4832),
    [sym_constructor_pattern] = STATE(4826),
    [sym_tag_pattern] = STATE(4826),
    [sym_polymorphic_variant_pattern] = STATE(4970),
    [sym_labeled_tuple_element_pattern] = STATE(7167),
    [sym__tuple_pattern] = STATE(4833),
    [sym_record_pattern] = STATE(4978),
    [sym_list_pattern] = STATE(4978),
    [sym_cons_pattern] = STATE(4830),
    [sym_array_pattern] = STATE(4978),
    [sym_range_pattern] = STATE(4830),
    [sym_lazy_pattern] = STATE(4826),
    [sym_local_open_pattern] = STATE(4970),
    [sym_package_pattern] = STATE(4970),
    [sym_parenthesized_pattern] = STATE(4978),
    [sym_exception_pattern] = STATE(4830),
    [sym_effect_pattern] = STATE(4830),
    [sym_attribute] = STATE(1127),
    [sym_extension] = STATE(4970),
    [sym_quoted_extension] = STATE(4970),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4725),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4978),
    [sym__simple_value_pattern] = STATE(4970),
    [sym_parenthesized_operator] = STATE(4970),
    [sym_module_path] = STATE(6366),
    [sym_constructor_path] = STATE(1088),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1127),
    [sym__lowercase_identifier] = ACTIONS(2208),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(2262),
    [anon_sym_COMMA] = ACTIONS(2075),
    [anon_sym_PIPE] = ACTIONS(2081),
    [anon_sym_LBRACK] = ACTIONS(2212),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2216),
    [anon_sym_SEMI] = ACTIONS(2075),
    [anon_sym_exception] = ACTIONS(2218),
    [anon_sym_as] = ACTIONS(2081),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_COLON_COLON] = ACTIONS(2075),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2222),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2075),
    [anon_sym_lazy] = ACTIONS(2224),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2228),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1128)] = {
    [sym__delimited_binding_pattern] = STATE(4914),
    [sym__simple_binding_pattern] = STATE(4921),
    [sym__effect_binding_pattern] = STATE(4846),
    [sym__binding_pattern] = STATE(4942),
    [sym_alias_binding_pattern] = STATE(4846),
    [sym_typed_binding_pattern] = STATE(4914),
    [sym__or_binding_pattern_anonymous] = STATE(4848),
    [sym_constructor_binding_pattern] = STATE(4921),
    [sym_tag_binding_pattern] = STATE(4921),
    [sym_polymorphic_variant_pattern] = STATE(4914),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6450),
    [sym__tuple_binding_pattern] = STATE(4850),
    [sym_record_binding_pattern] = STATE(4922),
    [sym_list_binding_pattern] = STATE(4922),
    [sym_cons_binding_pattern] = STATE(4846),
    [sym_array_binding_pattern] = STATE(4922),
    [sym_range_pattern] = STATE(4846),
    [sym_lazy_binding_pattern] = STATE(4921),
    [sym_local_open_binding_pattern] = STATE(4914),
    [sym_package_pattern] = STATE(4914),
    [sym_parenthesized_binding_pattern] = STATE(4922),
    [sym_exception_binding_pattern] = STATE(4846),
    [sym_effect_binding_pattern] = STATE(4846),
    [sym_attribute] = STATE(1128),
    [sym_extension] = STATE(4914),
    [sym_quoted_extension] = STATE(4914),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4739),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4922),
    [sym__value_name] = STATE(4914),
    [sym_parenthesized_operator] = STATE(4863),
    [sym_module_path] = STATE(6362),
    [sym_constructor_path] = STATE(1094),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1128),
    [sym__lowercase_identifier] = ACTIONS(2246),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(2264),
    [anon_sym_COMMA] = ACTIONS(1295),
    [anon_sym_PIPE] = ACTIONS(1297),
    [anon_sym_LBRACK] = ACTIONS(2250),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2252),
    [anon_sym_SEMI] = ACTIONS(1295),
    [anon_sym_exception] = ACTIONS(2254),
    [anon_sym_as] = ACTIONS(1297),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_COLON_COLON] = ACTIONS(1295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2256),
    [anon_sym_PIPE_RBRACK] = ACTIONS(1295),
    [anon_sym_lazy] = ACTIONS(2258),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2260),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1129)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4565),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1129),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_RPAREN] = ACTIONS(1295),
    [anon_sym_COMMA] = ACTIONS(1295),
    [anon_sym_PIPE] = ACTIONS(1295),
    [anon_sym_COLON2] = ACTIONS(1297),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_as] = ACTIONS(1297),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(1295),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1130)] = {
    [sym_parameter] = STATE(2207),
    [sym__parenthesized_abstract_type] = STATE(2207),
    [sym__delimited_pattern] = STATE(2286),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2286),
    [sym_polymorphic_variant_pattern] = STATE(2286),
    [sym_record_pattern] = STATE(2287),
    [sym_list_pattern] = STATE(2287),
    [sym_array_pattern] = STATE(2287),
    [sym_local_open_pattern] = STATE(2286),
    [sym_package_pattern] = STATE(2286),
    [sym_parenthesized_pattern] = STATE(2287),
    [sym_attribute] = STATE(1130),
    [sym_extension] = STATE(2286),
    [sym_quoted_extension] = STATE(2286),
    [sym__constant] = STATE(2212),
    [sym__signed_constant] = STATE(2286),
    [sym_number] = STATE(2280),
    [sym_signed_number] = STATE(2212),
    [sym_character] = STATE(2280),
    [sym_string] = STATE(2280),
    [sym_quoted_string] = STATE(2280),
    [sym_boolean] = STATE(2280),
    [sym_unit] = STATE(2287),
    [sym__simple_value_pattern] = STATE(2286),
    [sym_parenthesized_operator] = STATE(2286),
    [sym_module_path] = STATE(6556),
    [sym_constructor_path] = STATE(2286),
    [sym__constructor_name] = STATE(2597),
    [sym_tag] = STATE(2286),
    [aux_sym_let_binding_repeat1] = STATE(1130),
    [sym__lowercase_identifier] = ACTIONS(2266),
    [anon_sym_SEMI_SEMI] = ACTIONS(2124),
    [anon_sym_and] = ACTIONS(2126),
    [anon_sym_EQ] = ACTIONS(2124),
    [anon_sym_TILDE] = ACTIONS(2269),
    [anon_sym_QMARK] = ACTIONS(2269),
    [anon_sym_LPAREN] = ACTIONS(2272),
    [anon_sym_external] = ACTIONS(2126),
    [anon_sym_type] = ACTIONS(2126),
    [anon_sym_COLON2] = ACTIONS(2124),
    [anon_sym_LBRACK] = ACTIONS(2275),
    [anon_sym_RBRACK] = ACTIONS(2124),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2281),
    [anon_sym_exception] = ACTIONS(2126),
    [anon_sym_module] = ACTIONS(2126),
    [anon_sym_open] = ACTIONS(2126),
    [anon_sym_include] = ACTIONS(2126),
    [anon_sym_class] = ACTIONS(2126),
    [anon_sym_val] = ACTIONS(2126),
    [anon_sym_end] = ACTIONS(2126),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2287),
    [anon_sym_begin] = ACTIONS(2290),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2126),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2124),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2293),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2124),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2124),
    [aux_sym_number_token1] = ACTIONS(2299),
    [aux_sym_signed_number_token1] = ACTIONS(2302),
    [anon_sym_SQUOTE] = ACTIONS(2305),
    [anon_sym_DQUOTE] = ACTIONS(2308),
    [sym__uppercase_identifier] = ACTIONS(2311),
    [aux_sym_tag_token1] = ACTIONS(2314),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1131)] = {
    [sym_attribute] = STATE(1131),
    [sym__lowercase_identifier] = ACTIONS(1615),
    [anon_sym_SEMI_SEMI] = ACTIONS(1617),
    [anon_sym_let] = ACTIONS(1615),
    [anon_sym_and] = ACTIONS(1615),
    [anon_sym_TILDE] = ACTIONS(1615),
    [anon_sym_QMARK] = ACTIONS(1615),
    [anon_sym_LPAREN] = ACTIONS(1617),
    [anon_sym_external] = ACTIONS(1615),
    [anon_sym_type] = ACTIONS(1615),
    [anon_sym_COMMA] = ACTIONS(1617),
    [anon_sym_PIPE] = ACTIONS(1615),
    [anon_sym_LBRACK] = ACTIONS(1615),
    [anon_sym_RBRACK] = ACTIONS(1617),
    [anon_sym_true] = ACTIONS(1615),
    [anon_sym_false] = ACTIONS(1615),
    [anon_sym_LBRACE] = ACTIONS(1615),
    [anon_sym_SEMI] = ACTIONS(1615),
    [anon_sym_exception] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(1615),
    [anon_sym_open] = ACTIONS(1615),
    [anon_sym_include] = ACTIONS(1615),
    [anon_sym_class] = ACTIONS(1615),
    [anon_sym_val] = ACTIONS(1615),
    [anon_sym_end] = ACTIONS(1615),
    [anon_sym_object] = ACTIONS(1615),
    [anon_sym_in] = ACTIONS(1615),
    [anon_sym_AMP] = ACTIONS(1615),
    [anon_sym_COLON_COLON] = ACTIONS(1617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1617),
    [anon_sym_else] = ACTIONS(1615),
    [anon_sym_new] = ACTIONS(1615),
    [anon_sym_LBRACE_LT] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1615),
    [sym_ocamlyacc_value] = ACTIONS(1617),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1615),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1617),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1615),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1615),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1617),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1617),
    [aux_sym_number_token1] = ACTIONS(1617),
    [aux_sym_signed_number_token1] = ACTIONS(1615),
    [anon_sym_SQUOTE] = ACTIONS(1617),
    [anon_sym_DQUOTE] = ACTIONS(1617),
    [sym_prefix_operator] = ACTIONS(1617),
    [aux_sym_sign_operator_token1] = ACTIONS(1615),
    [anon_sym_lsl] = ACTIONS(1615),
    [anon_sym_lsr] = ACTIONS(1615),
    [anon_sym_asr] = ACTIONS(1615),
    [aux_sym_pow_operator_token1] = ACTIONS(1617),
    [anon_sym_mod] = ACTIONS(1615),
    [anon_sym_land] = ACTIONS(1615),
    [anon_sym_lor] = ACTIONS(1615),
    [anon_sym_lxor] = ACTIONS(1615),
    [aux_sym_mult_operator_token1] = ACTIONS(1615),
    [aux_sym_add_operator_token1] = ACTIONS(1615),
    [sym_concat_operator] = ACTIONS(1617),
    [sym_rel_operator] = ACTIONS(1615),
    [anon_sym_AMP_AMP] = ACTIONS(1615),
    [anon_sym_or] = ACTIONS(1615),
    [anon_sym_PIPE_PIPE] = ACTIONS(1615),
    [sym_assign_operator] = ACTIONS(1617),
    [sym_let_operator] = ACTIONS(1617),
    [sym_let_and_operator] = ACTIONS(1617),
    [sym__uppercase_identifier] = ACTIONS(1617),
    [aux_sym_directive_token1] = ACTIONS(1617),
    [aux_sym_tag_token1] = ACTIONS(1617),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1132)] = {
    [sym_attribute] = STATE(1132),
    [sym__lowercase_identifier] = ACTIONS(2017),
    [anon_sym_SEMI_SEMI] = ACTIONS(2019),
    [anon_sym_let] = ACTIONS(2017),
    [anon_sym_and] = ACTIONS(2017),
    [anon_sym_TILDE] = ACTIONS(2017),
    [anon_sym_QMARK] = ACTIONS(2017),
    [anon_sym_LPAREN] = ACTIONS(2019),
    [anon_sym_external] = ACTIONS(2017),
    [anon_sym_type] = ACTIONS(2017),
    [anon_sym_COMMA] = ACTIONS(2019),
    [anon_sym_PIPE] = ACTIONS(2017),
    [anon_sym_LBRACK] = ACTIONS(2017),
    [anon_sym_RBRACK] = ACTIONS(2019),
    [anon_sym_true] = ACTIONS(2017),
    [anon_sym_false] = ACTIONS(2017),
    [anon_sym_LBRACE] = ACTIONS(2017),
    [anon_sym_SEMI] = ACTIONS(2017),
    [anon_sym_exception] = ACTIONS(2017),
    [anon_sym_module] = ACTIONS(2017),
    [anon_sym_open] = ACTIONS(2017),
    [anon_sym_include] = ACTIONS(2017),
    [anon_sym_class] = ACTIONS(2017),
    [anon_sym_val] = ACTIONS(2017),
    [anon_sym_end] = ACTIONS(2017),
    [anon_sym_object] = ACTIONS(2017),
    [anon_sym_in] = ACTIONS(2017),
    [anon_sym_AMP] = ACTIONS(2017),
    [anon_sym_COLON_COLON] = ACTIONS(2019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2019),
    [anon_sym_else] = ACTIONS(2017),
    [anon_sym_new] = ACTIONS(2017),
    [anon_sym_LBRACE_LT] = ACTIONS(2019),
    [anon_sym_begin] = ACTIONS(2017),
    [sym_ocamlyacc_value] = ACTIONS(2019),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2017),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2019),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2019),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2019),
    [aux_sym_number_token1] = ACTIONS(2019),
    [aux_sym_signed_number_token1] = ACTIONS(2017),
    [anon_sym_SQUOTE] = ACTIONS(2019),
    [anon_sym_DQUOTE] = ACTIONS(2019),
    [sym_prefix_operator] = ACTIONS(2019),
    [aux_sym_sign_operator_token1] = ACTIONS(2017),
    [anon_sym_lsl] = ACTIONS(2017),
    [anon_sym_lsr] = ACTIONS(2017),
    [anon_sym_asr] = ACTIONS(2017),
    [aux_sym_pow_operator_token1] = ACTIONS(2019),
    [anon_sym_mod] = ACTIONS(2017),
    [anon_sym_land] = ACTIONS(2017),
    [anon_sym_lor] = ACTIONS(2017),
    [anon_sym_lxor] = ACTIONS(2017),
    [aux_sym_mult_operator_token1] = ACTIONS(2017),
    [aux_sym_add_operator_token1] = ACTIONS(2017),
    [sym_concat_operator] = ACTIONS(2019),
    [sym_rel_operator] = ACTIONS(2017),
    [anon_sym_AMP_AMP] = ACTIONS(2017),
    [anon_sym_or] = ACTIONS(2017),
    [anon_sym_PIPE_PIPE] = ACTIONS(2017),
    [sym_assign_operator] = ACTIONS(2019),
    [sym_let_operator] = ACTIONS(2019),
    [sym_let_and_operator] = ACTIONS(2019),
    [sym__uppercase_identifier] = ACTIONS(2019),
    [aux_sym_directive_token1] = ACTIONS(2019),
    [aux_sym_tag_token1] = ACTIONS(2019),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1133)] = {
    [sym__match_cases] = STATE(2148),
    [sym_match_case] = STATE(2095),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4726),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1133),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__attribute] = STATE(1197),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2317),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1134)] = {
    [sym__match_cases] = STATE(2157),
    [sym_match_case] = STATE(3297),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4762),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1134),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__attribute] = STATE(1149),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2319),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1135)] = {
    [sym__match_cases] = STATE(2157),
    [sym_match_case] = STATE(3277),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4695),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1135),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__attribute] = STATE(1174),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2321),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1136)] = {
    [sym__match_cases] = STATE(2935),
    [sym_match_case] = STATE(3130),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4775),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1136),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__attribute] = STATE(1201),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2323),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1137)] = {
    [sym__match_cases] = STATE(2157),
    [sym_match_case] = STATE(2185),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4806),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1137),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__attribute] = STATE(1178),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2325),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1138)] = {
    [sym__match_cases] = STATE(2227),
    [sym_match_case] = STATE(2185),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4806),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1138),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__attribute] = STATE(1168),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2325),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1139)] = {
    [sym__match_cases] = STATE(2148),
    [sym_match_case] = STATE(2310),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4688),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1139),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__attribute] = STATE(1182),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2327),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1140)] = {
    [sym__match_cases] = STATE(2227),
    [sym_match_case] = STATE(3277),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4695),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1140),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__attribute] = STATE(1152),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2321),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1141)] = {
    [sym__match_cases] = STATE(2148),
    [sym_match_case] = STATE(2166),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4784),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1141),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__attribute] = STATE(1156),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2329),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1142)] = {
    [sym_attribute] = STATE(1142),
    [sym__lowercase_identifier] = ACTIONS(1967),
    [anon_sym_SEMI_SEMI] = ACTIONS(1971),
    [anon_sym_let] = ACTIONS(1967),
    [anon_sym_and] = ACTIONS(1967),
    [anon_sym_TILDE] = ACTIONS(1967),
    [anon_sym_QMARK] = ACTIONS(1967),
    [anon_sym_COLON] = ACTIONS(2331),
    [anon_sym_LPAREN] = ACTIONS(1971),
    [anon_sym_external] = ACTIONS(1967),
    [anon_sym_type] = ACTIONS(1967),
    [anon_sym_COMMA] = ACTIONS(1971),
    [anon_sym_PIPE] = ACTIONS(1967),
    [anon_sym_LBRACK] = ACTIONS(1967),
    [anon_sym_RBRACK] = ACTIONS(1971),
    [anon_sym_true] = ACTIONS(1967),
    [anon_sym_false] = ACTIONS(1967),
    [anon_sym_LBRACE] = ACTIONS(1967),
    [anon_sym_SEMI] = ACTIONS(1967),
    [anon_sym_exception] = ACTIONS(1967),
    [anon_sym_module] = ACTIONS(1967),
    [anon_sym_open] = ACTIONS(1967),
    [anon_sym_include] = ACTIONS(1967),
    [anon_sym_class] = ACTIONS(1967),
    [anon_sym_val] = ACTIONS(1967),
    [anon_sym_end] = ACTIONS(1967),
    [anon_sym_object] = ACTIONS(1967),
    [anon_sym_AMP] = ACTIONS(1967),
    [anon_sym_COLON_COLON] = ACTIONS(1971),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1971),
    [anon_sym_else] = ACTIONS(1967),
    [anon_sym_new] = ACTIONS(1967),
    [anon_sym_LBRACE_LT] = ACTIONS(1971),
    [anon_sym_begin] = ACTIONS(1967),
    [sym_ocamlyacc_value] = ACTIONS(1971),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1967),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1971),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1967),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1967),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1971),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1971),
    [aux_sym_number_token1] = ACTIONS(1971),
    [aux_sym_signed_number_token1] = ACTIONS(1967),
    [anon_sym_SQUOTE] = ACTIONS(1971),
    [anon_sym_DQUOTE] = ACTIONS(1971),
    [sym_prefix_operator] = ACTIONS(1971),
    [aux_sym_sign_operator_token1] = ACTIONS(1967),
    [anon_sym_lsl] = ACTIONS(1967),
    [anon_sym_lsr] = ACTIONS(1967),
    [anon_sym_asr] = ACTIONS(1967),
    [aux_sym_pow_operator_token1] = ACTIONS(1971),
    [anon_sym_mod] = ACTIONS(1967),
    [anon_sym_land] = ACTIONS(1967),
    [anon_sym_lor] = ACTIONS(1967),
    [anon_sym_lxor] = ACTIONS(1967),
    [aux_sym_mult_operator_token1] = ACTIONS(1967),
    [aux_sym_add_operator_token1] = ACTIONS(1967),
    [sym_concat_operator] = ACTIONS(1971),
    [sym_rel_operator] = ACTIONS(1967),
    [anon_sym_AMP_AMP] = ACTIONS(1967),
    [anon_sym_or] = ACTIONS(1967),
    [anon_sym_PIPE_PIPE] = ACTIONS(1967),
    [sym_assign_operator] = ACTIONS(1971),
    [sym_let_operator] = ACTIONS(1971),
    [sym__uppercase_identifier] = ACTIONS(1971),
    [aux_sym_directive_token1] = ACTIONS(1971),
    [aux_sym_tag_token1] = ACTIONS(1971),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1143)] = {
    [sym__match_cases] = STATE(2935),
    [sym_match_case] = STATE(3218),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4752),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1143),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__attribute] = STATE(1166),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2333),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1144)] = {
    [sym_attribute] = STATE(1144),
    [sym__lowercase_identifier] = ACTIONS(1615),
    [anon_sym_SEMI_SEMI] = ACTIONS(1617),
    [anon_sym_let] = ACTIONS(1615),
    [anon_sym_and] = ACTIONS(1615),
    [anon_sym_TILDE] = ACTIONS(1615),
    [anon_sym_QMARK] = ACTIONS(1615),
    [anon_sym_LPAREN] = ACTIONS(1617),
    [anon_sym_external] = ACTIONS(1615),
    [anon_sym_type] = ACTIONS(1615),
    [anon_sym_COMMA] = ACTIONS(1617),
    [anon_sym_PIPE] = ACTIONS(1615),
    [anon_sym_LBRACK] = ACTIONS(1615),
    [anon_sym_RBRACK] = ACTIONS(1617),
    [anon_sym_true] = ACTIONS(1615),
    [anon_sym_false] = ACTIONS(1615),
    [anon_sym_LBRACE] = ACTIONS(1615),
    [anon_sym_SEMI] = ACTIONS(1615),
    [anon_sym_exception] = ACTIONS(1615),
    [anon_sym_module] = ACTIONS(1615),
    [anon_sym_open] = ACTIONS(1615),
    [anon_sym_include] = ACTIONS(1615),
    [anon_sym_class] = ACTIONS(1615),
    [anon_sym_val] = ACTIONS(1615),
    [anon_sym_end] = ACTIONS(1615),
    [anon_sym_object] = ACTIONS(1615),
    [anon_sym_AMP] = ACTIONS(1615),
    [anon_sym_COLON_COLON] = ACTIONS(1617),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1617),
    [anon_sym_else] = ACTIONS(1615),
    [anon_sym_new] = ACTIONS(1615),
    [anon_sym_LBRACE_LT] = ACTIONS(1617),
    [anon_sym_begin] = ACTIONS(1615),
    [sym_ocamlyacc_value] = ACTIONS(1617),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1615),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(1617),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1615),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1615),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(1617),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(1617),
    [aux_sym_number_token1] = ACTIONS(1617),
    [aux_sym_signed_number_token1] = ACTIONS(1615),
    [anon_sym_SQUOTE] = ACTIONS(1617),
    [anon_sym_DQUOTE] = ACTIONS(1617),
    [sym_prefix_operator] = ACTIONS(1617),
    [aux_sym_sign_operator_token1] = ACTIONS(1615),
    [anon_sym_lsl] = ACTIONS(1615),
    [anon_sym_lsr] = ACTIONS(1615),
    [anon_sym_asr] = ACTIONS(1615),
    [aux_sym_pow_operator_token1] = ACTIONS(1617),
    [anon_sym_mod] = ACTIONS(1615),
    [anon_sym_land] = ACTIONS(1615),
    [anon_sym_lor] = ACTIONS(1615),
    [anon_sym_lxor] = ACTIONS(1615),
    [aux_sym_mult_operator_token1] = ACTIONS(1615),
    [aux_sym_add_operator_token1] = ACTIONS(1615),
    [sym_concat_operator] = ACTIONS(1617),
    [sym_rel_operator] = ACTIONS(1615),
    [anon_sym_AMP_AMP] = ACTIONS(1615),
    [anon_sym_or] = ACTIONS(1615),
    [anon_sym_PIPE_PIPE] = ACTIONS(1615),
    [sym_assign_operator] = ACTIONS(1617),
    [sym_let_operator] = ACTIONS(1617),
    [sym__uppercase_identifier] = ACTIONS(1617),
    [aux_sym_directive_token1] = ACTIONS(1617),
    [aux_sym_tag_token1] = ACTIONS(1617),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1145)] = {
    [sym_parameter] = STATE(2207),
    [sym__parenthesized_abstract_type] = STATE(2207),
    [sym__delimited_pattern] = STATE(2286),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2286),
    [sym_polymorphic_variant_pattern] = STATE(2286),
    [sym_record_pattern] = STATE(2287),
    [sym_list_pattern] = STATE(2287),
    [sym_array_pattern] = STATE(2287),
    [sym_local_open_pattern] = STATE(2286),
    [sym_package_pattern] = STATE(2286),
    [sym_parenthesized_pattern] = STATE(2287),
    [sym_attribute] = STATE(1145),
    [sym_extension] = STATE(2286),
    [sym_quoted_extension] = STATE(2286),
    [sym__constant] = STATE(2212),
    [sym__signed_constant] = STATE(2286),
    [sym_number] = STATE(2280),
    [sym_signed_number] = STATE(2212),
    [sym_character] = STATE(2280),
    [sym_string] = STATE(2280),
    [sym_quoted_string] = STATE(2280),
    [sym_boolean] = STATE(2280),
    [sym_unit] = STATE(2287),
    [sym__simple_value_pattern] = STATE(2286),
    [sym_parenthesized_operator] = STATE(2286),
    [sym_module_path] = STATE(6390),
    [sym_constructor_path] = STATE(2286),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(2286),
    [aux_sym_let_binding_repeat1] = STATE(1145),
    [sym__lowercase_identifier] = ACTIONS(2266),
    [anon_sym_and] = ACTIONS(2126),
    [anon_sym_EQ] = ACTIONS(2124),
    [anon_sym_TILDE] = ACTIONS(2335),
    [anon_sym_QMARK] = ACTIONS(2335),
    [anon_sym_LPAREN] = ACTIONS(2338),
    [anon_sym_COLON2] = ACTIONS(2126),
    [anon_sym_DASH_GT] = ACTIONS(2124),
    [anon_sym_LBRACK] = ACTIONS(2275),
    [anon_sym_true] = ACTIONS(2278),
    [anon_sym_false] = ACTIONS(2278),
    [anon_sym_LBRACE] = ACTIONS(2281),
    [anon_sym_constraint] = ACTIONS(2126),
    [anon_sym_val] = ACTIONS(2126),
    [anon_sym_end] = ACTIONS(2126),
    [anon_sym_COLON_GT] = ACTIONS(2124),
    [anon_sym_inherit] = ACTIONS(2126),
    [anon_sym_method] = ACTIONS(2126),
    [anon_sym_in] = ACTIONS(2126),
    [anon_sym_initializer] = ACTIONS(2126),
    [anon_sym_POUND] = ACTIONS(2284),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2287),
    [anon_sym_begin] = ACTIONS(2290),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2126),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2124),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2293),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2296),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2124),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2124),
    [aux_sym_number_token1] = ACTIONS(2299),
    [aux_sym_signed_number_token1] = ACTIONS(2341),
    [anon_sym_SQUOTE] = ACTIONS(2305),
    [anon_sym_DQUOTE] = ACTIONS(2308),
    [sym__uppercase_identifier] = ACTIONS(2344),
    [aux_sym_tag_token1] = ACTIONS(2347),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1146)] = {
    [sym_attribute] = STATE(1146),
    [sym__lowercase_identifier] = ACTIONS(2017),
    [anon_sym_SEMI_SEMI] = ACTIONS(2019),
    [anon_sym_let] = ACTIONS(2017),
    [anon_sym_and] = ACTIONS(2017),
    [anon_sym_TILDE] = ACTIONS(2017),
    [anon_sym_QMARK] = ACTIONS(2017),
    [anon_sym_LPAREN] = ACTIONS(2019),
    [anon_sym_external] = ACTIONS(2017),
    [anon_sym_type] = ACTIONS(2017),
    [anon_sym_COMMA] = ACTIONS(2019),
    [anon_sym_PIPE] = ACTIONS(2017),
    [anon_sym_LBRACK] = ACTIONS(2017),
    [anon_sym_RBRACK] = ACTIONS(2019),
    [anon_sym_true] = ACTIONS(2017),
    [anon_sym_false] = ACTIONS(2017),
    [anon_sym_LBRACE] = ACTIONS(2017),
    [anon_sym_SEMI] = ACTIONS(2017),
    [anon_sym_exception] = ACTIONS(2017),
    [anon_sym_module] = ACTIONS(2017),
    [anon_sym_open] = ACTIONS(2017),
    [anon_sym_include] = ACTIONS(2017),
    [anon_sym_class] = ACTIONS(2017),
    [anon_sym_val] = ACTIONS(2017),
    [anon_sym_end] = ACTIONS(2017),
    [anon_sym_object] = ACTIONS(2017),
    [anon_sym_AMP] = ACTIONS(2017),
    [anon_sym_COLON_COLON] = ACTIONS(2019),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2019),
    [anon_sym_else] = ACTIONS(2017),
    [anon_sym_new] = ACTIONS(2017),
    [anon_sym_LBRACE_LT] = ACTIONS(2019),
    [anon_sym_begin] = ACTIONS(2017),
    [sym_ocamlyacc_value] = ACTIONS(2019),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2017),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2019),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2017),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2019),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2019),
    [aux_sym_number_token1] = ACTIONS(2019),
    [aux_sym_signed_number_token1] = ACTIONS(2017),
    [anon_sym_SQUOTE] = ACTIONS(2019),
    [anon_sym_DQUOTE] = ACTIONS(2019),
    [sym_prefix_operator] = ACTIONS(2019),
    [aux_sym_sign_operator_token1] = ACTIONS(2017),
    [anon_sym_lsl] = ACTIONS(2017),
    [anon_sym_lsr] = ACTIONS(2017),
    [anon_sym_asr] = ACTIONS(2017),
    [aux_sym_pow_operator_token1] = ACTIONS(2019),
    [anon_sym_mod] = ACTIONS(2017),
    [anon_sym_land] = ACTIONS(2017),
    [anon_sym_lor] = ACTIONS(2017),
    [anon_sym_lxor] = ACTIONS(2017),
    [aux_sym_mult_operator_token1] = ACTIONS(2017),
    [aux_sym_add_operator_token1] = ACTIONS(2017),
    [sym_concat_operator] = ACTIONS(2019),
    [sym_rel_operator] = ACTIONS(2017),
    [anon_sym_AMP_AMP] = ACTIONS(2017),
    [anon_sym_or] = ACTIONS(2017),
    [anon_sym_PIPE_PIPE] = ACTIONS(2017),
    [sym_assign_operator] = ACTIONS(2019),
    [sym_let_operator] = ACTIONS(2019),
    [sym__uppercase_identifier] = ACTIONS(2019),
    [aux_sym_directive_token1] = ACTIONS(2019),
    [aux_sym_tag_token1] = ACTIONS(2019),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1147)] = {
    [sym__match_cases] = STATE(2142),
    [sym_match_case] = STATE(3277),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4695),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1147),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2321),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1148)] = {
    [sym_parameter] = STATE(2207),
    [sym__maybe_polymorphic_typed] = STATE(4267),
    [sym__parenthesized_abstract_type] = STATE(2207),
    [sym__delimited_pattern] = STATE(2286),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2286),
    [sym_polymorphic_variant_pattern] = STATE(2286),
    [sym_record_pattern] = STATE(2287),
    [sym_list_pattern] = STATE(2287),
    [sym_array_pattern] = STATE(2287),
    [sym_local_open_pattern] = STATE(2286),
    [sym_package_pattern] = STATE(2286),
    [sym_parenthesized_pattern] = STATE(2287),
    [sym_attribute] = STATE(1148),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2286),
    [sym_quoted_extension] = STATE(2286),
    [sym__constant] = STATE(2212),
    [sym__signed_constant] = STATE(2286),
    [sym_number] = STATE(2280),
    [sym_signed_number] = STATE(2212),
    [sym_character] = STATE(2280),
    [sym_string] = STATE(2280),
    [sym_quoted_string] = STATE(2280),
    [sym_boolean] = STATE(2280),
    [sym_unit] = STATE(2287),
    [sym__simple_value_pattern] = STATE(2286),
    [sym_parenthesized_operator] = STATE(2286),
    [sym_module_path] = STATE(6390),
    [sym_constructor_path] = STATE(2286),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(2286),
    [aux_sym_expression_item_repeat1] = STATE(4378),
    [aux_sym_let_binding_repeat1] = STATE(1194),
    [sym__lowercase_identifier] = ACTIONS(1467),
    [anon_sym_EQ] = ACTIONS(2350),
    [anon_sym_TILDE] = ACTIONS(1475),
    [anon_sym_QMARK] = ACTIONS(1475),
    [anon_sym_LPAREN] = ACTIONS(1477),
    [anon_sym_COLON2] = ACTIONS(2352),
    [anon_sym_LBRACK] = ACTIONS(1485),
    [anon_sym_true] = ACTIONS(1487),
    [anon_sym_false] = ACTIONS(1487),
    [anon_sym_LBRACE] = ACTIONS(1489),
    [anon_sym_constraint] = ACTIONS(2354),
    [anon_sym_val] = ACTIONS(2354),
    [anon_sym_end] = ACTIONS(2354),
    [anon_sym_inherit] = ACTIONS(2354),
    [anon_sym_method] = ACTIONS(2354),
    [anon_sym_initializer] = ACTIONS(2354),
    [anon_sym_POUND] = ACTIONS(1495),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1499),
    [anon_sym_begin] = ACTIONS(1501),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2356),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2356),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2356),
    [aux_sym_number_token1] = ACTIONS(1509),
    [aux_sym_signed_number_token1] = ACTIONS(1511),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE] = ACTIONS(1515),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1149)] = {
    [sym__match_cases] = STATE(2153),
    [sym_match_case] = STATE(3297),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4762),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1149),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2319),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1150)] = {
    [sym__match_cases] = STATE(2168),
    [sym_match_case] = STATE(2166),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4784),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1150),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2329),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1151)] = {
    [sym__match_cases] = STATE(2889),
    [sym_match_case] = STATE(3130),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4775),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1151),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2323),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1152)] = {
    [sym__match_cases] = STATE(2222),
    [sym_match_case] = STATE(3277),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4695),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1152),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2321),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1153)] = {
    [sym_parameter] = STATE(2207),
    [sym__maybe_polymorphic_typed] = STATE(4296),
    [sym__parenthesized_abstract_type] = STATE(2207),
    [sym__delimited_pattern] = STATE(2286),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2286),
    [sym_polymorphic_variant_pattern] = STATE(2286),
    [sym_record_pattern] = STATE(2287),
    [sym_list_pattern] = STATE(2287),
    [sym_array_pattern] = STATE(2287),
    [sym_local_open_pattern] = STATE(2286),
    [sym_package_pattern] = STATE(2286),
    [sym_parenthesized_pattern] = STATE(2287),
    [sym_attribute] = STATE(1153),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2286),
    [sym_quoted_extension] = STATE(2286),
    [sym__constant] = STATE(2212),
    [sym__signed_constant] = STATE(2286),
    [sym_number] = STATE(2280),
    [sym_signed_number] = STATE(2212),
    [sym_character] = STATE(2280),
    [sym_string] = STATE(2280),
    [sym_quoted_string] = STATE(2280),
    [sym_boolean] = STATE(2280),
    [sym_unit] = STATE(2287),
    [sym__simple_value_pattern] = STATE(2286),
    [sym_parenthesized_operator] = STATE(2286),
    [sym_module_path] = STATE(6390),
    [sym_constructor_path] = STATE(2286),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(2286),
    [aux_sym_expression_item_repeat1] = STATE(4432),
    [aux_sym_let_binding_repeat1] = STATE(1145),
    [sym__lowercase_identifier] = ACTIONS(1467),
    [anon_sym_EQ] = ACTIONS(2358),
    [anon_sym_TILDE] = ACTIONS(1475),
    [anon_sym_QMARK] = ACTIONS(1475),
    [anon_sym_LPAREN] = ACTIONS(1477),
    [anon_sym_COLON2] = ACTIONS(2352),
    [anon_sym_LBRACK] = ACTIONS(1485),
    [anon_sym_true] = ACTIONS(1487),
    [anon_sym_false] = ACTIONS(1487),
    [anon_sym_LBRACE] = ACTIONS(1489),
    [anon_sym_constraint] = ACTIONS(2360),
    [anon_sym_val] = ACTIONS(2360),
    [anon_sym_end] = ACTIONS(2360),
    [anon_sym_inherit] = ACTIONS(2360),
    [anon_sym_method] = ACTIONS(2360),
    [anon_sym_initializer] = ACTIONS(2360),
    [anon_sym_POUND] = ACTIONS(1495),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1499),
    [anon_sym_begin] = ACTIONS(1501),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2362),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2362),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2362),
    [aux_sym_number_token1] = ACTIONS(1509),
    [aux_sym_signed_number_token1] = ACTIONS(1511),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE] = ACTIONS(1515),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1154)] = {
    [sym__match_cases] = STATE(2170),
    [sym_match_case] = STATE(2166),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4784),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1154),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2329),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1155)] = {
    [sym__match_cases] = STATE(2895),
    [sym_match_case] = STATE(3130),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4775),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1155),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2323),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1156)] = {
    [sym__match_cases] = STATE(2121),
    [sym_match_case] = STATE(2166),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4784),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1156),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2329),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1157)] = {
    [sym_parameter] = STATE(2207),
    [sym__maybe_polymorphic_typed] = STATE(4242),
    [sym__parenthesized_abstract_type] = STATE(2207),
    [sym__delimited_pattern] = STATE(2286),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2286),
    [sym_polymorphic_variant_pattern] = STATE(2286),
    [sym_record_pattern] = STATE(2287),
    [sym_list_pattern] = STATE(2287),
    [sym_array_pattern] = STATE(2287),
    [sym_local_open_pattern] = STATE(2286),
    [sym_package_pattern] = STATE(2286),
    [sym_parenthesized_pattern] = STATE(2287),
    [sym_attribute] = STATE(1157),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2286),
    [sym_quoted_extension] = STATE(2286),
    [sym__constant] = STATE(2212),
    [sym__signed_constant] = STATE(2286),
    [sym_number] = STATE(2280),
    [sym_signed_number] = STATE(2212),
    [sym_character] = STATE(2280),
    [sym_string] = STATE(2280),
    [sym_quoted_string] = STATE(2280),
    [sym_boolean] = STATE(2280),
    [sym_unit] = STATE(2287),
    [sym__simple_value_pattern] = STATE(2286),
    [sym_parenthesized_operator] = STATE(2286),
    [sym_module_path] = STATE(6390),
    [sym_constructor_path] = STATE(2286),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(2286),
    [aux_sym_expression_item_repeat1] = STATE(4410),
    [aux_sym_let_binding_repeat1] = STATE(1176),
    [sym__lowercase_identifier] = ACTIONS(1467),
    [anon_sym_EQ] = ACTIONS(2364),
    [anon_sym_TILDE] = ACTIONS(1475),
    [anon_sym_QMARK] = ACTIONS(1475),
    [anon_sym_LPAREN] = ACTIONS(1477),
    [anon_sym_COLON2] = ACTIONS(2352),
    [anon_sym_LBRACK] = ACTIONS(1485),
    [anon_sym_true] = ACTIONS(1487),
    [anon_sym_false] = ACTIONS(1487),
    [anon_sym_LBRACE] = ACTIONS(1489),
    [anon_sym_constraint] = ACTIONS(2366),
    [anon_sym_val] = ACTIONS(2366),
    [anon_sym_end] = ACTIONS(2366),
    [anon_sym_inherit] = ACTIONS(2366),
    [anon_sym_method] = ACTIONS(2366),
    [anon_sym_initializer] = ACTIONS(2366),
    [anon_sym_POUND] = ACTIONS(1495),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1499),
    [anon_sym_begin] = ACTIONS(1501),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2368),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2368),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2368),
    [aux_sym_number_token1] = ACTIONS(1509),
    [aux_sym_signed_number_token1] = ACTIONS(1511),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE] = ACTIONS(1515),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1158)] = {
    [sym__match_cases] = STATE(2145),
    [sym_match_case] = STATE(2185),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4806),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1158),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2325),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1159)] = {
    [sym__match_cases] = STATE(2139),
    [sym_match_case] = STATE(3297),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4762),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1159),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2319),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1160)] = {
    [sym__match_cases] = STATE(2145),
    [sym_match_case] = STATE(3297),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4762),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1160),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2319),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1161)] = {
    [sym__match_cases] = STATE(2956),
    [sym_match_case] = STATE(3130),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4775),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1161),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2323),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1162)] = {
    [sym__match_cases] = STATE(2275),
    [sym_match_case] = STATE(3277),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4695),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1162),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2321),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1163)] = {
    [sym__match_cases] = STATE(2170),
    [sym_match_case] = STATE(2095),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4726),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1163),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2317),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1164)] = {
    [sym__match_cases] = STATE(2144),
    [sym_match_case] = STATE(2185),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4806),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1164),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2325),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1165)] = {
    [sym__match_cases] = STATE(2142),
    [sym_match_case] = STATE(2185),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4806),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1165),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2325),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1166)] = {
    [sym__match_cases] = STATE(2908),
    [sym_match_case] = STATE(3218),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4752),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1166),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2333),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1167)] = {
    [sym__match_cases] = STATE(2923),
    [sym_match_case] = STATE(3130),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4775),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1167),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2323),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1168)] = {
    [sym__match_cases] = STATE(2222),
    [sym_match_case] = STATE(2185),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4806),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1168),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2325),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1169)] = {
    [sym__match_cases] = STATE(2295),
    [sym_match_case] = STATE(2185),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4806),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1169),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2325),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1170)] = {
    [sym_parameter] = STATE(2207),
    [sym__maybe_polymorphic_typed] = STATE(4264),
    [sym__parenthesized_abstract_type] = STATE(2207),
    [sym__delimited_pattern] = STATE(2286),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2286),
    [sym_polymorphic_variant_pattern] = STATE(2286),
    [sym_record_pattern] = STATE(2287),
    [sym_list_pattern] = STATE(2287),
    [sym_array_pattern] = STATE(2287),
    [sym_local_open_pattern] = STATE(2286),
    [sym_package_pattern] = STATE(2286),
    [sym_parenthesized_pattern] = STATE(2287),
    [sym_attribute] = STATE(1170),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2286),
    [sym_quoted_extension] = STATE(2286),
    [sym__constant] = STATE(2212),
    [sym__signed_constant] = STATE(2286),
    [sym_number] = STATE(2280),
    [sym_signed_number] = STATE(2212),
    [sym_character] = STATE(2280),
    [sym_string] = STATE(2280),
    [sym_quoted_string] = STATE(2280),
    [sym_boolean] = STATE(2280),
    [sym_unit] = STATE(2287),
    [sym__simple_value_pattern] = STATE(2286),
    [sym_parenthesized_operator] = STATE(2286),
    [sym_module_path] = STATE(6390),
    [sym_constructor_path] = STATE(2286),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(2286),
    [aux_sym_expression_item_repeat1] = STATE(4417),
    [aux_sym_let_binding_repeat1] = STATE(1153),
    [sym__lowercase_identifier] = ACTIONS(1467),
    [anon_sym_EQ] = ACTIONS(2370),
    [anon_sym_TILDE] = ACTIONS(1475),
    [anon_sym_QMARK] = ACTIONS(1475),
    [anon_sym_LPAREN] = ACTIONS(1477),
    [anon_sym_COLON2] = ACTIONS(2352),
    [anon_sym_LBRACK] = ACTIONS(1485),
    [anon_sym_true] = ACTIONS(1487),
    [anon_sym_false] = ACTIONS(1487),
    [anon_sym_LBRACE] = ACTIONS(1489),
    [anon_sym_constraint] = ACTIONS(2372),
    [anon_sym_val] = ACTIONS(2372),
    [anon_sym_end] = ACTIONS(2372),
    [anon_sym_inherit] = ACTIONS(2372),
    [anon_sym_method] = ACTIONS(2372),
    [anon_sym_initializer] = ACTIONS(2372),
    [anon_sym_POUND] = ACTIONS(1495),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1499),
    [anon_sym_begin] = ACTIONS(1501),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2374),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2374),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2374),
    [aux_sym_number_token1] = ACTIONS(1509),
    [aux_sym_signed_number_token1] = ACTIONS(1511),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE] = ACTIONS(1515),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1171)] = {
    [sym__match_cases] = STATE(2168),
    [sym_match_case] = STATE(2310),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4688),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1171),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2327),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1172)] = {
    [sym__match_cases] = STATE(2170),
    [sym_match_case] = STATE(2310),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4688),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1172),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2327),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1173)] = {
    [sym__match_cases] = STATE(2274),
    [sym_match_case] = STATE(3277),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4695),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1173),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2321),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1174)] = {
    [sym__match_cases] = STATE(2153),
    [sym_match_case] = STATE(3277),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4695),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1174),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2321),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1175)] = {
    [sym__match_cases] = STATE(2895),
    [sym_match_case] = STATE(3218),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4752),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1175),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2333),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1176)] = {
    [sym_parameter] = STATE(2207),
    [sym__maybe_polymorphic_typed] = STATE(4301),
    [sym__parenthesized_abstract_type] = STATE(2207),
    [sym__delimited_pattern] = STATE(2286),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2286),
    [sym_polymorphic_variant_pattern] = STATE(2286),
    [sym_record_pattern] = STATE(2287),
    [sym_list_pattern] = STATE(2287),
    [sym_array_pattern] = STATE(2287),
    [sym_local_open_pattern] = STATE(2286),
    [sym_package_pattern] = STATE(2286),
    [sym_parenthesized_pattern] = STATE(2287),
    [sym_attribute] = STATE(1176),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2286),
    [sym_quoted_extension] = STATE(2286),
    [sym__constant] = STATE(2212),
    [sym__signed_constant] = STATE(2286),
    [sym_number] = STATE(2280),
    [sym_signed_number] = STATE(2212),
    [sym_character] = STATE(2280),
    [sym_string] = STATE(2280),
    [sym_quoted_string] = STATE(2280),
    [sym_boolean] = STATE(2280),
    [sym_unit] = STATE(2287),
    [sym__simple_value_pattern] = STATE(2286),
    [sym_parenthesized_operator] = STATE(2286),
    [sym_module_path] = STATE(6390),
    [sym_constructor_path] = STATE(2286),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(2286),
    [aux_sym_expression_item_repeat1] = STATE(4349),
    [aux_sym_let_binding_repeat1] = STATE(1145),
    [sym__lowercase_identifier] = ACTIONS(1467),
    [anon_sym_EQ] = ACTIONS(2376),
    [anon_sym_TILDE] = ACTIONS(1475),
    [anon_sym_QMARK] = ACTIONS(1475),
    [anon_sym_LPAREN] = ACTIONS(1477),
    [anon_sym_COLON2] = ACTIONS(2352),
    [anon_sym_LBRACK] = ACTIONS(1485),
    [anon_sym_true] = ACTIONS(1487),
    [anon_sym_false] = ACTIONS(1487),
    [anon_sym_LBRACE] = ACTIONS(1489),
    [anon_sym_constraint] = ACTIONS(2378),
    [anon_sym_val] = ACTIONS(2378),
    [anon_sym_end] = ACTIONS(2378),
    [anon_sym_inherit] = ACTIONS(2378),
    [anon_sym_method] = ACTIONS(2378),
    [anon_sym_initializer] = ACTIONS(2378),
    [anon_sym_POUND] = ACTIONS(1495),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1499),
    [anon_sym_begin] = ACTIONS(1501),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2380),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2380),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2380),
    [aux_sym_number_token1] = ACTIONS(1509),
    [aux_sym_signed_number_token1] = ACTIONS(1511),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE] = ACTIONS(1515),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1177)] = {
    [sym__match_cases] = STATE(2159),
    [sym_match_case] = STATE(2166),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4784),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1177),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2329),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1178)] = {
    [sym__match_cases] = STATE(2153),
    [sym_match_case] = STATE(2185),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4806),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1178),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2325),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1179)] = {
    [sym__match_cases] = STATE(2139),
    [sym_match_case] = STATE(3277),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4695),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1179),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2321),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1180)] = {
    [sym__match_cases] = STATE(2144),
    [sym_match_case] = STATE(3297),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4762),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1180),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2319),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1181)] = {
    [sym__match_cases] = STATE(2274),
    [sym_match_case] = STATE(2185),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4806),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1181),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2325),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1182)] = {
    [sym__match_cases] = STATE(2121),
    [sym_match_case] = STATE(2310),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4688),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1182),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2327),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1183)] = {
    [sym__match_cases] = STATE(2145),
    [sym_match_case] = STATE(3277),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4695),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1183),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2321),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1184)] = {
    [sym__match_cases] = STATE(2159),
    [sym_match_case] = STATE(2310),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4688),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1184),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2327),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1185)] = {
    [sym_parameter] = STATE(2207),
    [sym__coerced] = STATE(6467),
    [sym__maybe_polymorphic_typed] = STATE(5611),
    [sym__parenthesized_abstract_type] = STATE(2207),
    [sym__delimited_pattern] = STATE(2286),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2286),
    [sym_polymorphic_variant_pattern] = STATE(2286),
    [sym_record_pattern] = STATE(2287),
    [sym_list_pattern] = STATE(2287),
    [sym_array_pattern] = STATE(2287),
    [sym_local_open_pattern] = STATE(2286),
    [sym_package_pattern] = STATE(2286),
    [sym_parenthesized_pattern] = STATE(2287),
    [sym_attribute] = STATE(1185),
    [sym_item_attribute] = STATE(5502),
    [sym_extension] = STATE(2286),
    [sym_quoted_extension] = STATE(2286),
    [sym__constant] = STATE(2212),
    [sym__signed_constant] = STATE(2286),
    [sym_number] = STATE(2280),
    [sym_signed_number] = STATE(2212),
    [sym_character] = STATE(2280),
    [sym_string] = STATE(2280),
    [sym_quoted_string] = STATE(2280),
    [sym_boolean] = STATE(2280),
    [sym_unit] = STATE(2287),
    [sym__simple_value_pattern] = STATE(2286),
    [sym_parenthesized_operator] = STATE(2286),
    [sym_module_path] = STATE(6390),
    [sym_constructor_path] = STATE(2286),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(2286),
    [aux_sym_expression_item_repeat1] = STATE(4864),
    [aux_sym_let_binding_repeat1] = STATE(1519),
    [sym__lowercase_identifier] = ACTIONS(1467),
    [anon_sym_and] = ACTIONS(1471),
    [anon_sym_EQ] = ACTIONS(2382),
    [anon_sym_TILDE] = ACTIONS(1475),
    [anon_sym_QMARK] = ACTIONS(1475),
    [anon_sym_LPAREN] = ACTIONS(1477),
    [anon_sym_COMMA] = ACTIONS(2384),
    [anon_sym_PIPE] = ACTIONS(2386),
    [anon_sym_COLON2] = ACTIONS(1483),
    [anon_sym_LBRACK] = ACTIONS(1485),
    [anon_sym_true] = ACTIONS(1487),
    [anon_sym_false] = ACTIONS(1487),
    [anon_sym_LBRACE] = ACTIONS(1489),
    [anon_sym_COLON_GT] = ACTIONS(1491),
    [anon_sym_in] = ACTIONS(1471),
    [anon_sym_as] = ACTIONS(2388),
    [anon_sym_POUND] = ACTIONS(1495),
    [anon_sym_COLON_COLON] = ACTIONS(2390),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1499),
    [anon_sym_begin] = ACTIONS(1501),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(2392),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2394),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2396),
    [aux_sym_number_token1] = ACTIONS(1509),
    [aux_sym_signed_number_token1] = ACTIONS(1511),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE] = ACTIONS(1515),
    [sym_let_and_operator] = ACTIONS(1469),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1186)] = {
    [sym__match_cases] = STATE(2275),
    [sym_match_case] = STATE(2185),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4806),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1186),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2325),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1187)] = {
    [sym__match_cases] = STATE(2144),
    [sym_match_case] = STATE(3277),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4695),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1187),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2321),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1188)] = {
    [sym__match_cases] = STATE(2295),
    [sym_match_case] = STATE(3277),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4695),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1188),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2321),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1189)] = {
    [sym__match_cases] = STATE(2889),
    [sym_match_case] = STATE(3218),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4752),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1189),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2333),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1190)] = {
    [sym__match_cases] = STATE(2142),
    [sym_match_case] = STATE(3297),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4762),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1190),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2319),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1191)] = {
    [sym__match_cases] = STATE(2184),
    [sym_match_case] = STATE(3277),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4695),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1191),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2321),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1192)] = {
    [sym__match_cases] = STATE(2122),
    [sym_match_case] = STATE(2166),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4784),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1192),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2329),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1193)] = {
    [sym__match_cases] = STATE(2122),
    [sym_match_case] = STATE(2310),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4688),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1193),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2327),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1194)] = {
    [sym_parameter] = STATE(2207),
    [sym__maybe_polymorphic_typed] = STATE(4259),
    [sym__parenthesized_abstract_type] = STATE(2207),
    [sym__delimited_pattern] = STATE(2286),
    [sym__simple_pattern] = STATE(2243),
    [sym_typed_pattern] = STATE(2286),
    [sym_polymorphic_variant_pattern] = STATE(2286),
    [sym_record_pattern] = STATE(2287),
    [sym_list_pattern] = STATE(2287),
    [sym_array_pattern] = STATE(2287),
    [sym_local_open_pattern] = STATE(2286),
    [sym_package_pattern] = STATE(2286),
    [sym_parenthesized_pattern] = STATE(2287),
    [sym_attribute] = STATE(1194),
    [sym_item_attribute] = STATE(3494),
    [sym_extension] = STATE(2286),
    [sym_quoted_extension] = STATE(2286),
    [sym__constant] = STATE(2212),
    [sym__signed_constant] = STATE(2286),
    [sym_number] = STATE(2280),
    [sym_signed_number] = STATE(2212),
    [sym_character] = STATE(2280),
    [sym_string] = STATE(2280),
    [sym_quoted_string] = STATE(2280),
    [sym_boolean] = STATE(2280),
    [sym_unit] = STATE(2287),
    [sym__simple_value_pattern] = STATE(2286),
    [sym_parenthesized_operator] = STATE(2286),
    [sym_module_path] = STATE(6390),
    [sym_constructor_path] = STATE(2286),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(2286),
    [aux_sym_expression_item_repeat1] = STATE(4400),
    [aux_sym_let_binding_repeat1] = STATE(1145),
    [sym__lowercase_identifier] = ACTIONS(1467),
    [anon_sym_EQ] = ACTIONS(2398),
    [anon_sym_TILDE] = ACTIONS(1475),
    [anon_sym_QMARK] = ACTIONS(1475),
    [anon_sym_LPAREN] = ACTIONS(1477),
    [anon_sym_COLON2] = ACTIONS(2352),
    [anon_sym_LBRACK] = ACTIONS(1485),
    [anon_sym_true] = ACTIONS(1487),
    [anon_sym_false] = ACTIONS(1487),
    [anon_sym_LBRACE] = ACTIONS(1489),
    [anon_sym_constraint] = ACTIONS(2400),
    [anon_sym_val] = ACTIONS(2400),
    [anon_sym_end] = ACTIONS(2400),
    [anon_sym_inherit] = ACTIONS(2400),
    [anon_sym_method] = ACTIONS(2400),
    [anon_sym_initializer] = ACTIONS(2400),
    [anon_sym_POUND] = ACTIONS(1495),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1499),
    [anon_sym_begin] = ACTIONS(1501),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT] = ACTIONS(1999),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(2402),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1505),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1507),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(2402),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(2402),
    [aux_sym_number_token1] = ACTIONS(1509),
    [aux_sym_signed_number_token1] = ACTIONS(1511),
    [anon_sym_SQUOTE] = ACTIONS(1513),
    [anon_sym_DQUOTE] = ACTIONS(1515),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1195)] = {
    [sym__match_cases] = STATE(2122),
    [sym_match_case] = STATE(2095),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4726),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1195),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2317),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1196)] = {
    [sym__match_cases] = STATE(2956),
    [sym_match_case] = STATE(3218),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4752),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1196),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2333),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1197)] = {
    [sym__match_cases] = STATE(2121),
    [sym_match_case] = STATE(2095),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4726),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1197),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2317),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1198)] = {
    [sym__match_cases] = STATE(2923),
    [sym_match_case] = STATE(3218),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4752),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1198),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2333),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1199)] = {
    [sym__match_cases] = STATE(2168),
    [sym_match_case] = STATE(2095),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4726),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1199),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2317),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1200)] = {
    [sym__match_cases] = STATE(2159),
    [sym_match_case] = STATE(2095),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4726),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1200),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2317),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1201)] = {
    [sym__match_cases] = STATE(2908),
    [sym_match_case] = STATE(3130),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4775),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1201),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2323),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1202)] = {
    [sym__match_cases] = STATE(2139),
    [sym_match_case] = STATE(2185),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4806),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1202),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2325),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1203)] = {
    [sym__match_cases] = STATE(2184),
    [sym_match_case] = STATE(2185),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4806),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1203),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_PIPE] = ACTIONS(2325),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1204)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4771),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(6815),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym__sequence_pattern_content] = STATE(6998),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1204),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(1042),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1097),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_RBRACK] = ACTIONS(2404),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(2198),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(2200),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2202),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1205)] = {
    [sym__delimited_pattern] = STATE(4970),
    [sym__simple_pattern] = STATE(4826),
    [sym__effect_pattern] = STATE(4830),
    [sym__pattern] = STATE(4698),
    [sym_alias_pattern] = STATE(4830),
    [sym_typed_pattern] = STATE(4970),
    [sym__or_pattern_anonymous] = STATE(4832),
    [sym_constructor_pattern] = STATE(4826),
    [sym_tag_pattern] = STATE(4826),
    [sym_polymorphic_variant_pattern] = STATE(4970),
    [sym_labeled_tuple_element_pattern] = STATE(7167),
    [sym__tuple_pattern] = STATE(4833),
    [sym_record_pattern] = STATE(4978),
    [sym_list_pattern] = STATE(4978),
    [sym_cons_pattern] = STATE(4830),
    [sym_array_pattern] = STATE(4978),
    [sym__sequence_pattern_content] = STATE(6477),
    [sym_range_pattern] = STATE(4830),
    [sym_lazy_pattern] = STATE(4826),
    [sym_local_open_pattern] = STATE(4970),
    [sym_package_pattern] = STATE(4970),
    [sym_parenthesized_pattern] = STATE(4978),
    [sym_exception_pattern] = STATE(4830),
    [sym_effect_pattern] = STATE(4830),
    [sym_attribute] = STATE(1205),
    [sym_extension] = STATE(4970),
    [sym_quoted_extension] = STATE(4970),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4725),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4978),
    [sym__simple_value_pattern] = STATE(4970),
    [sym_parenthesized_operator] = STATE(4970),
    [sym_module_path] = STATE(6366),
    [sym_constructor_path] = STATE(1088),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1127),
    [sym__lowercase_identifier] = ACTIONS(2208),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(2262),
    [anon_sym_LBRACK] = ACTIONS(2212),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2216),
    [anon_sym_exception] = ACTIONS(2218),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2222),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2406),
    [anon_sym_lazy] = ACTIONS(2224),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2228),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1206)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5096),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1206),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(469),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(1195),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1207)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(5473),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1207),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__attribute] = STATE(1311),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1208)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(5197),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1208),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_RPAREN] = ACTIONS(469),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(1195),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1209)] = {
    [sym__delimited_binding_pattern] = STATE(2051),
    [sym__simple_binding_pattern] = STATE(2085),
    [sym__effect_binding_pattern] = STATE(2060),
    [sym__binding_pattern] = STATE(2024),
    [sym_alias_binding_pattern] = STATE(2060),
    [sym_typed_binding_pattern] = STATE(2051),
    [sym__or_binding_pattern_anonymous] = STATE(2061),
    [sym_constructor_binding_pattern] = STATE(2085),
    [sym_tag_binding_pattern] = STATE(2085),
    [sym_polymorphic_variant_pattern] = STATE(2051),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7213),
    [sym__tuple_binding_pattern] = STATE(2062),
    [sym_record_binding_pattern] = STATE(2087),
    [sym_list_binding_pattern] = STATE(2087),
    [sym_cons_binding_pattern] = STATE(2060),
    [sym_array_binding_pattern] = STATE(2087),
    [sym_range_pattern] = STATE(2060),
    [sym_lazy_binding_pattern] = STATE(2085),
    [sym_local_open_binding_pattern] = STATE(2051),
    [sym_package_pattern] = STATE(2051),
    [sym_parenthesized_binding_pattern] = STATE(2087),
    [sym_exception_binding_pattern] = STATE(2060),
    [sym_effect_binding_pattern] = STATE(2060),
    [sym_attribute] = STATE(1209),
    [sym_extension] = STATE(2051),
    [sym_quoted_extension] = STATE(2051),
    [sym__attribute] = STATE(1302),
    [sym__constant] = STATE(1823),
    [sym__signed_constant] = STATE(1825),
    [sym_number] = STATE(1803),
    [sym_signed_number] = STATE(1823),
    [sym_character] = STATE(1803),
    [sym_string] = STATE(1803),
    [sym_quoted_string] = STATE(1803),
    [sym_boolean] = STATE(1803),
    [sym_unit] = STATE(2087),
    [sym__value_name] = STATE(2051),
    [sym_parenthesized_operator] = STATE(2035),
    [sym_module_path] = STATE(6855),
    [sym_constructor_path] = STATE(714),
    [sym__constructor_name] = STATE(1698),
    [sym_tag] = STATE(716),
    [sym__lowercase_identifier] = ACTIONS(1293),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1347),
    [anon_sym_LBRACK] = ACTIONS(1303),
    [anon_sym_true] = ACTIONS(1305),
    [anon_sym_false] = ACTIONS(1305),
    [anon_sym_LBRACE] = ACTIONS(1307),
    [anon_sym_exception] = ACTIONS(1309),
    [anon_sym_POUND] = ACTIONS(1311),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1313),
    [anon_sym_lazy] = ACTIONS(1315),
    [anon_sym_begin] = ACTIONS(1317),
    [anon_sym_effect] = ACTIONS(1319),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2408),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2410),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1325),
    [aux_sym_signed_number_token1] = ACTIONS(1327),
    [anon_sym_SQUOTE] = ACTIONS(1329),
    [anon_sym_DQUOTE] = ACTIONS(1331),
    [sym__uppercase_identifier] = ACTIONS(1333),
    [aux_sym_tag_token1] = ACTIONS(1335),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1210)] = {
    [sym__parenthesized_abstract_type] = STATE(1320),
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(5439),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1210),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(2182),
    [anon_sym_COMMA] = ACTIONS(1295),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1211)] = {
    [sym__delimited_binding_pattern] = STATE(4914),
    [sym__simple_binding_pattern] = STATE(4921),
    [sym__effect_binding_pattern] = STATE(4846),
    [sym__binding_pattern] = STATE(4692),
    [sym_alias_binding_pattern] = STATE(4846),
    [sym_typed_binding_pattern] = STATE(4914),
    [sym__or_binding_pattern_anonymous] = STATE(4848),
    [sym_constructor_binding_pattern] = STATE(4921),
    [sym_tag_binding_pattern] = STATE(4921),
    [sym_polymorphic_variant_pattern] = STATE(4914),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6450),
    [sym__tuple_binding_pattern] = STATE(4850),
    [sym_record_binding_pattern] = STATE(4922),
    [sym_list_binding_pattern] = STATE(4922),
    [sym_cons_binding_pattern] = STATE(4846),
    [sym_array_binding_pattern] = STATE(4922),
    [sym__sequence_binding_pattern_content] = STATE(7110),
    [sym_range_pattern] = STATE(4846),
    [sym_lazy_binding_pattern] = STATE(4921),
    [sym_local_open_binding_pattern] = STATE(4914),
    [sym_package_pattern] = STATE(4914),
    [sym_parenthesized_binding_pattern] = STATE(4922),
    [sym_exception_binding_pattern] = STATE(4846),
    [sym_effect_binding_pattern] = STATE(4846),
    [sym_attribute] = STATE(1211),
    [sym_extension] = STATE(4914),
    [sym_quoted_extension] = STATE(4914),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4739),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4922),
    [sym__value_name] = STATE(4914),
    [sym_parenthesized_operator] = STATE(4863),
    [sym_module_path] = STATE(6362),
    [sym_constructor_path] = STATE(1094),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1128),
    [sym__lowercase_identifier] = ACTIONS(2246),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(2264),
    [anon_sym_LBRACK] = ACTIONS(2250),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2252),
    [anon_sym_exception] = ACTIONS(2254),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2256),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2412),
    [anon_sym_lazy] = ACTIONS(2258),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2260),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1212)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(5159),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1212),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_RPAREN] = ACTIONS(1139),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(1145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1213)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5169),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1213),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(1139),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(1145),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1214)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4501),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1214),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__attribute] = STATE(1365),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1215)] = {
    [sym_let_binding] = STATE(5451),
    [sym_open_module] = STATE(7065),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(1215),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym__attribute] = STATE(1400),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1411),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1965),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1216)] = {
    [sym__delimited_pattern] = STATE(4970),
    [sym__simple_pattern] = STATE(4826),
    [sym__effect_pattern] = STATE(4830),
    [sym__pattern] = STATE(4698),
    [sym_alias_pattern] = STATE(4830),
    [sym_typed_pattern] = STATE(4970),
    [sym__or_pattern_anonymous] = STATE(4832),
    [sym_constructor_pattern] = STATE(4826),
    [sym_tag_pattern] = STATE(4826),
    [sym_polymorphic_variant_pattern] = STATE(4970),
    [sym_labeled_tuple_element_pattern] = STATE(7167),
    [sym__tuple_pattern] = STATE(4833),
    [sym_record_pattern] = STATE(4978),
    [sym_list_pattern] = STATE(4978),
    [sym_cons_pattern] = STATE(4830),
    [sym_array_pattern] = STATE(4978),
    [sym__sequence_pattern_content] = STATE(7024),
    [sym_range_pattern] = STATE(4830),
    [sym_lazy_pattern] = STATE(4826),
    [sym_local_open_pattern] = STATE(4970),
    [sym_package_pattern] = STATE(4970),
    [sym_parenthesized_pattern] = STATE(4978),
    [sym_exception_pattern] = STATE(4830),
    [sym_effect_pattern] = STATE(4830),
    [sym_attribute] = STATE(1216),
    [sym_extension] = STATE(4970),
    [sym_quoted_extension] = STATE(4970),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4725),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4978),
    [sym__simple_value_pattern] = STATE(4970),
    [sym_parenthesized_operator] = STATE(4970),
    [sym_module_path] = STATE(6366),
    [sym_constructor_path] = STATE(1088),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1127),
    [sym__lowercase_identifier] = ACTIONS(2208),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(2262),
    [anon_sym_LBRACK] = ACTIONS(2212),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2216),
    [anon_sym_exception] = ACTIONS(2218),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2222),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2414),
    [anon_sym_lazy] = ACTIONS(2224),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2228),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1217)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(5063),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1217),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_RPAREN] = ACTIONS(1041),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(1059),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1218)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5096),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1218),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(469),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(1113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1219)] = {
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2961),
    [sym__binding_pattern] = STATE(3026),
    [sym_alias_binding_pattern] = STATE(2961),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_anonymous] = STATE(2962),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6582),
    [sym__tuple_binding_pattern] = STATE(2964),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern] = STATE(2961),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2961),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_exception_binding_pattern] = STATE(2961),
    [sym_effect_binding_pattern] = STATE(2961),
    [sym_attribute] = STATE(1219),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym__attribute] = STATE(1379),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1959),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [anon_sym_effect] = ACTIONS(1961),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1965),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1220)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4341),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1220),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__attribute] = STATE(1364),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1221)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4703),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6673),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1221),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__attribute] = STATE(1335),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(4808),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1038),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1120),
    [sym__lowercase_identifier] = ACTIONS(2180),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(2184),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(2188),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2190),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1222)] = {
    [sym__delimited_binding_pattern] = STATE(4914),
    [sym__simple_binding_pattern] = STATE(4921),
    [sym__effect_binding_pattern] = STATE(4846),
    [sym__binding_pattern] = STATE(4822),
    [sym_alias_binding_pattern] = STATE(4846),
    [sym_typed_binding_pattern] = STATE(4914),
    [sym__or_binding_pattern_anonymous] = STATE(4848),
    [sym_constructor_binding_pattern] = STATE(4921),
    [sym_tag_binding_pattern] = STATE(4921),
    [sym_polymorphic_variant_pattern] = STATE(4914),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6450),
    [sym__tuple_binding_pattern] = STATE(4850),
    [sym_record_binding_pattern] = STATE(4922),
    [sym_list_binding_pattern] = STATE(4922),
    [sym_cons_binding_pattern] = STATE(4846),
    [sym_array_binding_pattern] = STATE(4922),
    [sym_range_pattern] = STATE(4846),
    [sym_lazy_binding_pattern] = STATE(4921),
    [sym_local_open_binding_pattern] = STATE(4914),
    [sym_package_pattern] = STATE(4914),
    [sym_parenthesized_binding_pattern] = STATE(4922),
    [sym_exception_binding_pattern] = STATE(4846),
    [sym_effect_binding_pattern] = STATE(4846),
    [sym_attribute] = STATE(1222),
    [sym_extension] = STATE(4914),
    [sym_quoted_extension] = STATE(4914),
    [sym__attribute] = STATE(1368),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4739),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4922),
    [sym__value_name] = STATE(4914),
    [sym_parenthesized_operator] = STATE(4863),
    [sym_module_path] = STATE(6362),
    [sym_constructor_path] = STATE(1094),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1128),
    [sym__lowercase_identifier] = ACTIONS(2246),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(2264),
    [anon_sym_LBRACK] = ACTIONS(2250),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2252),
    [anon_sym_exception] = ACTIONS(2254),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2256),
    [anon_sym_lazy] = ACTIONS(2258),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2260),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1223)] = {
    [sym_let_binding] = STATE(5451),
    [sym_open_module] = STATE(6651),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(1223),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym__attribute] = STATE(1400),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1411),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1965),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1224)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4926),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1224),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__attribute] = STATE(1329),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1225)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5050),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1225),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__attribute] = STATE(1342),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1226)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4793),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6673),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym__sequence_binding_pattern_content] = STATE(6674),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1226),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(4808),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1038),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1120),
    [sym__lowercase_identifier] = ACTIONS(2180),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_RBRACK] = ACTIONS(2416),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(2184),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(2188),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2190),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1227)] = {
    [sym__delimited_binding_pattern] = STATE(4914),
    [sym__simple_binding_pattern] = STATE(4921),
    [sym__effect_binding_pattern] = STATE(4846),
    [sym__binding_pattern] = STATE(4692),
    [sym_alias_binding_pattern] = STATE(4846),
    [sym_typed_binding_pattern] = STATE(4914),
    [sym__or_binding_pattern_anonymous] = STATE(4848),
    [sym_constructor_binding_pattern] = STATE(4921),
    [sym_tag_binding_pattern] = STATE(4921),
    [sym_polymorphic_variant_pattern] = STATE(4914),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6450),
    [sym__tuple_binding_pattern] = STATE(4850),
    [sym_record_binding_pattern] = STATE(4922),
    [sym_list_binding_pattern] = STATE(4922),
    [sym_cons_binding_pattern] = STATE(4846),
    [sym_array_binding_pattern] = STATE(4922),
    [sym__sequence_binding_pattern_content] = STATE(6681),
    [sym_range_pattern] = STATE(4846),
    [sym_lazy_binding_pattern] = STATE(4921),
    [sym_local_open_binding_pattern] = STATE(4914),
    [sym_package_pattern] = STATE(4914),
    [sym_parenthesized_binding_pattern] = STATE(4922),
    [sym_exception_binding_pattern] = STATE(4846),
    [sym_effect_binding_pattern] = STATE(4846),
    [sym_attribute] = STATE(1227),
    [sym_extension] = STATE(4914),
    [sym_quoted_extension] = STATE(4914),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4739),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4922),
    [sym__value_name] = STATE(4914),
    [sym_parenthesized_operator] = STATE(4863),
    [sym_module_path] = STATE(6362),
    [sym_constructor_path] = STATE(1094),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1128),
    [sym__lowercase_identifier] = ACTIONS(2246),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(2264),
    [anon_sym_LBRACK] = ACTIONS(2250),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2252),
    [anon_sym_exception] = ACTIONS(2254),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2256),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2418),
    [anon_sym_lazy] = ACTIONS(2258),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2260),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1228)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4793),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6673),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym__sequence_binding_pattern_content] = STATE(6754),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1228),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(4808),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1038),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1120),
    [sym__lowercase_identifier] = ACTIONS(2180),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_RBRACK] = ACTIONS(2420),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(2184),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(2188),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2190),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1229)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4793),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6673),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym__sequence_binding_pattern_content] = STATE(7217),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1229),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(4808),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1038),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1120),
    [sym__lowercase_identifier] = ACTIONS(2180),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_RBRACK] = ACTIONS(2422),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(2184),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(2188),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2190),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1230)] = {
    [sym__delimited_binding_pattern] = STATE(4914),
    [sym__simple_binding_pattern] = STATE(4921),
    [sym__effect_binding_pattern] = STATE(4846),
    [sym__binding_pattern] = STATE(4692),
    [sym_alias_binding_pattern] = STATE(4846),
    [sym_typed_binding_pattern] = STATE(4914),
    [sym__or_binding_pattern_anonymous] = STATE(4848),
    [sym_constructor_binding_pattern] = STATE(4921),
    [sym_tag_binding_pattern] = STATE(4921),
    [sym_polymorphic_variant_pattern] = STATE(4914),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6450),
    [sym__tuple_binding_pattern] = STATE(4850),
    [sym_record_binding_pattern] = STATE(4922),
    [sym_list_binding_pattern] = STATE(4922),
    [sym_cons_binding_pattern] = STATE(4846),
    [sym_array_binding_pattern] = STATE(4922),
    [sym__sequence_binding_pattern_content] = STATE(7223),
    [sym_range_pattern] = STATE(4846),
    [sym_lazy_binding_pattern] = STATE(4921),
    [sym_local_open_binding_pattern] = STATE(4914),
    [sym_package_pattern] = STATE(4914),
    [sym_parenthesized_binding_pattern] = STATE(4922),
    [sym_exception_binding_pattern] = STATE(4846),
    [sym_effect_binding_pattern] = STATE(4846),
    [sym_attribute] = STATE(1230),
    [sym_extension] = STATE(4914),
    [sym_quoted_extension] = STATE(4914),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4739),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4922),
    [sym__value_name] = STATE(4914),
    [sym_parenthesized_operator] = STATE(4863),
    [sym_module_path] = STATE(6362),
    [sym_constructor_path] = STATE(1094),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1128),
    [sym__lowercase_identifier] = ACTIONS(2246),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(2264),
    [anon_sym_LBRACK] = ACTIONS(2250),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2252),
    [anon_sym_exception] = ACTIONS(2254),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2256),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2424),
    [anon_sym_lazy] = ACTIONS(2258),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2260),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1231)] = {
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2961),
    [sym__binding_pattern] = STATE(2972),
    [sym_alias_binding_pattern] = STATE(2961),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_anonymous] = STATE(2962),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6582),
    [sym__tuple_binding_pattern] = STATE(2964),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern] = STATE(2961),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2961),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_exception_binding_pattern] = STATE(2961),
    [sym_effect_binding_pattern] = STATE(2961),
    [sym_attribute] = STATE(1231),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym__attribute] = STATE(1332),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1959),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [anon_sym_effect] = ACTIONS(1961),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1965),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1232)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(5197),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1232),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_RPAREN] = ACTIONS(469),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(475),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1233)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5256),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1233),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(1155),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(1163),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1234)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4771),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(6815),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym__sequence_pattern_content] = STATE(7249),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1234),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(1042),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1097),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_RBRACK] = ACTIONS(2426),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(2198),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(2200),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2202),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1235)] = {
    [sym__delimited_pattern] = STATE(4970),
    [sym__simple_pattern] = STATE(4826),
    [sym__effect_pattern] = STATE(4830),
    [sym__pattern] = STATE(4698),
    [sym_alias_pattern] = STATE(4830),
    [sym_typed_pattern] = STATE(4970),
    [sym__or_pattern_anonymous] = STATE(4832),
    [sym_constructor_pattern] = STATE(4826),
    [sym_tag_pattern] = STATE(4826),
    [sym_polymorphic_variant_pattern] = STATE(4970),
    [sym_labeled_tuple_element_pattern] = STATE(7167),
    [sym__tuple_pattern] = STATE(4833),
    [sym_record_pattern] = STATE(4978),
    [sym_list_pattern] = STATE(4978),
    [sym_cons_pattern] = STATE(4830),
    [sym_array_pattern] = STATE(4978),
    [sym__sequence_pattern_content] = STATE(7253),
    [sym_range_pattern] = STATE(4830),
    [sym_lazy_pattern] = STATE(4826),
    [sym_local_open_pattern] = STATE(4970),
    [sym_package_pattern] = STATE(4970),
    [sym_parenthesized_pattern] = STATE(4978),
    [sym_exception_pattern] = STATE(4830),
    [sym_effect_pattern] = STATE(4830),
    [sym_attribute] = STATE(1235),
    [sym_extension] = STATE(4970),
    [sym_quoted_extension] = STATE(4970),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4725),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4978),
    [sym__simple_value_pattern] = STATE(4970),
    [sym_parenthesized_operator] = STATE(4970),
    [sym_module_path] = STATE(6366),
    [sym_constructor_path] = STATE(1088),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1127),
    [sym__lowercase_identifier] = ACTIONS(2208),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(2262),
    [anon_sym_LBRACK] = ACTIONS(2212),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2216),
    [anon_sym_exception] = ACTIONS(2218),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2222),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2428),
    [anon_sym_lazy] = ACTIONS(2224),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2228),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1236)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4609),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6673),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1236),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__attribute] = STATE(1378),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(4808),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1038),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1120),
    [sym__lowercase_identifier] = ACTIONS(2180),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(2184),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(2188),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2190),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1237)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(5159),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1237),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_RPAREN] = ACTIONS(1139),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(1185),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1238)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5169),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1238),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(1139),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(1185),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1239)] = {
    [sym_let_binding] = STATE(5451),
    [sym_open_module] = STATE(6423),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(1239),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym__attribute] = STATE(1400),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1411),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1965),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1240)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4771),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(6815),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym__sequence_pattern_content] = STATE(6785),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1240),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(1042),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1097),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_RBRACK] = ACTIONS(2430),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(2198),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(2200),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2202),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1241)] = {
    [sym_let_binding] = STATE(5451),
    [sym_open_module] = STATE(6527),
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2976),
    [sym__binding_pattern_no_exn] = STATE(1185),
    [sym_alias_binding_pattern_no_exn] = STATE(2976),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2976),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6580),
    [sym__tuple_binding_pattern_no_exn] = STATE(2976),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern_no_exn] = STATE(2976),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2976),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_attribute] = STATE(1241),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym__attribute] = STATE(1400),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_rec] = ACTIONS(1411),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_open] = ACTIONS(1429),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1965),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1242)] = {
    [sym__delimited_binding_pattern] = STATE(2051),
    [sym__simple_binding_pattern] = STATE(2085),
    [sym__effect_binding_pattern] = STATE(2060),
    [sym__binding_pattern] = STATE(1922),
    [sym_alias_binding_pattern] = STATE(2060),
    [sym_typed_binding_pattern] = STATE(2051),
    [sym__or_binding_pattern_anonymous] = STATE(2061),
    [sym_constructor_binding_pattern] = STATE(2085),
    [sym_tag_binding_pattern] = STATE(2085),
    [sym_polymorphic_variant_pattern] = STATE(2051),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7213),
    [sym__tuple_binding_pattern] = STATE(2062),
    [sym_record_binding_pattern] = STATE(2087),
    [sym_list_binding_pattern] = STATE(2087),
    [sym_cons_binding_pattern] = STATE(2060),
    [sym_array_binding_pattern] = STATE(2087),
    [sym_range_pattern] = STATE(2060),
    [sym_lazy_binding_pattern] = STATE(2085),
    [sym_local_open_binding_pattern] = STATE(2051),
    [sym_package_pattern] = STATE(2051),
    [sym_parenthesized_binding_pattern] = STATE(2087),
    [sym_exception_binding_pattern] = STATE(2060),
    [sym_effect_binding_pattern] = STATE(2060),
    [sym_attribute] = STATE(1242),
    [sym_extension] = STATE(2051),
    [sym_quoted_extension] = STATE(2051),
    [sym__attribute] = STATE(1336),
    [sym__constant] = STATE(1823),
    [sym__signed_constant] = STATE(1825),
    [sym_number] = STATE(1803),
    [sym_signed_number] = STATE(1823),
    [sym_character] = STATE(1803),
    [sym_string] = STATE(1803),
    [sym_quoted_string] = STATE(1803),
    [sym_boolean] = STATE(1803),
    [sym_unit] = STATE(2087),
    [sym__value_name] = STATE(2051),
    [sym_parenthesized_operator] = STATE(2035),
    [sym_module_path] = STATE(6855),
    [sym_constructor_path] = STATE(714),
    [sym__constructor_name] = STATE(1698),
    [sym_tag] = STATE(716),
    [sym__lowercase_identifier] = ACTIONS(1293),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1347),
    [anon_sym_LBRACK] = ACTIONS(1303),
    [anon_sym_true] = ACTIONS(1305),
    [anon_sym_false] = ACTIONS(1305),
    [anon_sym_LBRACE] = ACTIONS(1307),
    [anon_sym_exception] = ACTIONS(1309),
    [anon_sym_POUND] = ACTIONS(1311),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1313),
    [anon_sym_lazy] = ACTIONS(1315),
    [anon_sym_begin] = ACTIONS(1317),
    [anon_sym_effect] = ACTIONS(1319),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2408),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2410),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1325),
    [aux_sym_signed_number_token1] = ACTIONS(1327),
    [anon_sym_SQUOTE] = ACTIONS(1329),
    [anon_sym_DQUOTE] = ACTIONS(1331),
    [sym__uppercase_identifier] = ACTIONS(1333),
    [aux_sym_tag_token1] = ACTIONS(1335),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1243)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(5197),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1243),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_RPAREN] = ACTIONS(469),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(1113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1244)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5256),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1244),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(1155),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(1113),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1245)] = {
    [sym__signature] = STATE(7055),
    [sym_external] = STATE(4240),
    [sym_type_definition] = STATE(4240),
    [sym_exception_definition] = STATE(4240),
    [sym_module_definition] = STATE(4240),
    [sym_module_type_definition] = STATE(4240),
    [sym_open_module] = STATE(4240),
    [sym_class_definition] = STATE(4240),
    [sym_class_type_definition] = STATE(4240),
    [sym__signature_item] = STATE(4212),
    [sym_value_specification] = STATE(4240),
    [sym_include_module_type] = STATE(4240),
    [sym__delimited_type] = STATE(2574),
    [sym__simple_type] = STATE(3803),
    [sym__type] = STATE(6127),
    [sym_function_type] = STATE(2650),
    [sym_labeled_argument_type] = STATE(7246),
    [sym__proper_tuple_type] = STATE(4229),
    [sym__labeled_tuple_type] = STATE(2650),
    [sym_labeled_tuple_element_type] = STATE(6978),
    [sym_constructed_type] = STATE(2574),
    [sym_aliased_type] = STATE(2650),
    [sym_local_open_type] = STATE(2574),
    [sym_polymorphic_variant_type] = STATE(2586),
    [sym_package_type] = STATE(2586),
    [sym_object_type] = STATE(2574),
    [sym_hash_type] = STATE(2574),
    [sym_parenthesized_type] = STATE(2586),
    [sym_attribute] = STATE(1245),
    [sym_floating_attribute] = STATE(4240),
    [sym_extension] = STATE(2574),
    [sym_quoted_extension] = STATE(2574),
    [sym_item_extension] = STATE(4240),
    [sym_quoted_item_extension] = STATE(4240),
    [sym_extended_module_path] = STATE(6301),
    [sym_type_constructor_path] = STATE(2574),
    [sym_type_variable] = STATE(2574),
    [aux_sym__structure_repeat1] = STATE(3139),
    [aux_sym__signature_repeat1] = STATE(2959),
    [sym__lowercase_identifier] = ACTIONS(2432),
    [anon_sym_SEMI_SEMI] = ACTIONS(2434),
    [anon_sym_QMARK] = ACTIONS(9),
    [anon_sym_LPAREN] = ACTIONS(11),
    [anon_sym_external] = ACTIONS(2436),
    [anon_sym_type] = ACTIONS(2438),
    [anon_sym_LBRACK] = ACTIONS(13),
    [anon_sym_RBRACK] = ACTIONS(2440),
    [anon_sym_exception] = ACTIONS(2442),
    [anon_sym_module] = ACTIONS(2444),
    [anon_sym_open] = ACTIONS(249),
    [anon_sym_include] = ACTIONS(2446),
    [anon_sym_class] = ACTIONS(2448),
    [anon_sym_val] = ACTIONS(2450),
    [anon_sym_LBRACK_GT] = ACTIONS(15),
    [anon_sym_LBRACK_LT] = ACTIONS(17),
    [anon_sym_LT] = ACTIONS(19),
    [anon_sym_POUND] = ACTIONS(21),
    [aux_sym_attribute_token1] = ACTIONS(3),
    [anon_sym_LBRACK_AT_AT_AT] = ACTIONS(257),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2452),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2454),
    [anon_sym_LBRACK_PERCENT_PERCENT] = ACTIONS(259),
    [anon_sym_LBRACE_PERCENT_PERCENT] = ACTIONS(261),
    [sym__uppercase_identifier] = ACTIONS(29),
    [aux_sym_type_variable_token1] = ACTIONS(31),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1246)] = {
    [sym__delimited_pattern] = STATE(4970),
    [sym__simple_pattern] = STATE(4826),
    [sym__effect_pattern] = STATE(4830),
    [sym__pattern] = STATE(4868),
    [sym_alias_pattern] = STATE(4830),
    [sym_typed_pattern] = STATE(4970),
    [sym__or_pattern_anonymous] = STATE(4832),
    [sym_constructor_pattern] = STATE(4826),
    [sym_tag_pattern] = STATE(4826),
    [sym_polymorphic_variant_pattern] = STATE(4970),
    [sym_labeled_tuple_element_pattern] = STATE(7167),
    [sym__tuple_pattern] = STATE(4833),
    [sym_record_pattern] = STATE(4978),
    [sym_list_pattern] = STATE(4978),
    [sym_cons_pattern] = STATE(4830),
    [sym_array_pattern] = STATE(4978),
    [sym_range_pattern] = STATE(4830),
    [sym_lazy_pattern] = STATE(4826),
    [sym_local_open_pattern] = STATE(4970),
    [sym_package_pattern] = STATE(4970),
    [sym_parenthesized_pattern] = STATE(4978),
    [sym_exception_pattern] = STATE(4830),
    [sym_effect_pattern] = STATE(4830),
    [sym_attribute] = STATE(1246),
    [sym_extension] = STATE(4970),
    [sym_quoted_extension] = STATE(4970),
    [sym__attribute] = STATE(1347),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4725),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4978),
    [sym__simple_value_pattern] = STATE(4970),
    [sym_parenthesized_operator] = STATE(4970),
    [sym_module_path] = STATE(6366),
    [sym_constructor_path] = STATE(1088),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1127),
    [sym__lowercase_identifier] = ACTIONS(2208),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(2262),
    [anon_sym_LBRACK] = ACTIONS(2212),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2216),
    [anon_sym_exception] = ACTIONS(2218),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2224),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2228),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1247)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5096),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1247),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(469),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(1191),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1248)] = {
    [sym__delimited_pattern] = STATE(4970),
    [sym__simple_pattern] = STATE(4826),
    [sym__effect_pattern] = STATE(4830),
    [sym__pattern] = STATE(4874),
    [sym_alias_pattern] = STATE(4830),
    [sym_typed_pattern] = STATE(4970),
    [sym__or_pattern_anonymous] = STATE(4832),
    [sym_constructor_pattern] = STATE(4826),
    [sym_tag_pattern] = STATE(4826),
    [sym_polymorphic_variant_pattern] = STATE(4970),
    [sym_labeled_tuple_element_pattern] = STATE(7167),
    [sym__tuple_pattern] = STATE(4833),
    [sym_record_pattern] = STATE(4978),
    [sym_list_pattern] = STATE(4978),
    [sym_cons_pattern] = STATE(4830),
    [sym_array_pattern] = STATE(4978),
    [sym_range_pattern] = STATE(4830),
    [sym_lazy_pattern] = STATE(4826),
    [sym_local_open_pattern] = STATE(4970),
    [sym_package_pattern] = STATE(4970),
    [sym_parenthesized_pattern] = STATE(4978),
    [sym_exception_pattern] = STATE(4830),
    [sym_effect_pattern] = STATE(4830),
    [sym_attribute] = STATE(1248),
    [sym_extension] = STATE(4970),
    [sym_quoted_extension] = STATE(4970),
    [sym__attribute] = STATE(1357),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4725),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4978),
    [sym__simple_value_pattern] = STATE(4970),
    [sym_parenthesized_operator] = STATE(4970),
    [sym_module_path] = STATE(6366),
    [sym_constructor_path] = STATE(1088),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1127),
    [sym__lowercase_identifier] = ACTIONS(2208),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(2262),
    [anon_sym_LBRACK] = ACTIONS(2212),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2216),
    [anon_sym_exception] = ACTIONS(2218),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2224),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2228),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1249)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4609),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1249),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__attribute] = STATE(1346),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1250)] = {
    [sym__delimited_pattern] = STATE(4970),
    [sym__simple_pattern] = STATE(4826),
    [sym__effect_pattern] = STATE(4830),
    [sym__pattern] = STATE(4698),
    [sym_alias_pattern] = STATE(4830),
    [sym_typed_pattern] = STATE(4970),
    [sym__or_pattern_anonymous] = STATE(4832),
    [sym_constructor_pattern] = STATE(4826),
    [sym_tag_pattern] = STATE(4826),
    [sym_polymorphic_variant_pattern] = STATE(4970),
    [sym_labeled_tuple_element_pattern] = STATE(7167),
    [sym__tuple_pattern] = STATE(4833),
    [sym_record_pattern] = STATE(4978),
    [sym_list_pattern] = STATE(4978),
    [sym_cons_pattern] = STATE(4830),
    [sym_array_pattern] = STATE(4978),
    [sym__sequence_pattern_content] = STATE(6807),
    [sym_range_pattern] = STATE(4830),
    [sym_lazy_pattern] = STATE(4826),
    [sym_local_open_pattern] = STATE(4970),
    [sym_package_pattern] = STATE(4970),
    [sym_parenthesized_pattern] = STATE(4978),
    [sym_exception_pattern] = STATE(4830),
    [sym_effect_pattern] = STATE(4830),
    [sym_attribute] = STATE(1250),
    [sym_extension] = STATE(4970),
    [sym_quoted_extension] = STATE(4970),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4725),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4978),
    [sym__simple_value_pattern] = STATE(4970),
    [sym_parenthesized_operator] = STATE(4970),
    [sym_module_path] = STATE(6366),
    [sym_constructor_path] = STATE(1088),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1127),
    [sym__lowercase_identifier] = ACTIONS(2208),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(2262),
    [anon_sym_LBRACK] = ACTIONS(2212),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2216),
    [anon_sym_exception] = ACTIONS(2218),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2222),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2456),
    [anon_sym_lazy] = ACTIONS(2224),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2228),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1251)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5361),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1251),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__attribute] = STATE(1353),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1252)] = {
    [sym__parenthesized_abstract_type] = STATE(1374),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5386),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1252),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(2079),
    [anon_sym_COMMA] = ACTIONS(2075),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1253)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(5145),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1253),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_RPAREN] = ACTIONS(1087),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(1093),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1254)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5152),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1254),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(1127),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(1133),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1255)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4691),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(6815),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1255),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__attribute] = STATE(1354),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(1042),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1097),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(2198),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(2200),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2202),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1256)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4341),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(6815),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1256),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__attribute] = STATE(1358),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(1042),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1097),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(2198),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(2200),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2202),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1257)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4793),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6673),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym__sequence_binding_pattern_content] = STATE(6452),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1257),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(4808),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1038),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1120),
    [sym__lowercase_identifier] = ACTIONS(2180),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_RBRACK] = ACTIONS(2458),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(2184),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(2188),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2190),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1258)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5096),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1258),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_RPAREN] = ACTIONS(469),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(475),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1259)] = {
    [sym__delimited_binding_pattern] = STATE(4914),
    [sym__simple_binding_pattern] = STATE(4921),
    [sym__effect_binding_pattern] = STATE(4846),
    [sym__binding_pattern] = STATE(4692),
    [sym_alias_binding_pattern] = STATE(4846),
    [sym_typed_binding_pattern] = STATE(4914),
    [sym__or_binding_pattern_anonymous] = STATE(4848),
    [sym_constructor_binding_pattern] = STATE(4921),
    [sym_tag_binding_pattern] = STATE(4921),
    [sym_polymorphic_variant_pattern] = STATE(4914),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6450),
    [sym__tuple_binding_pattern] = STATE(4850),
    [sym_record_binding_pattern] = STATE(4922),
    [sym_list_binding_pattern] = STATE(4922),
    [sym_cons_binding_pattern] = STATE(4846),
    [sym_array_binding_pattern] = STATE(4922),
    [sym__sequence_binding_pattern_content] = STATE(6458),
    [sym_range_pattern] = STATE(4846),
    [sym_lazy_binding_pattern] = STATE(4921),
    [sym_local_open_binding_pattern] = STATE(4914),
    [sym_package_pattern] = STATE(4914),
    [sym_parenthesized_binding_pattern] = STATE(4922),
    [sym_exception_binding_pattern] = STATE(4846),
    [sym_effect_binding_pattern] = STATE(4846),
    [sym_attribute] = STATE(1259),
    [sym_extension] = STATE(4914),
    [sym_quoted_extension] = STATE(4914),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4739),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4922),
    [sym__value_name] = STATE(4914),
    [sym_parenthesized_operator] = STATE(4863),
    [sym_module_path] = STATE(6362),
    [sym_constructor_path] = STATE(1094),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1128),
    [sym__lowercase_identifier] = ACTIONS(2246),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(2264),
    [anon_sym_LBRACK] = ACTIONS(2250),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2252),
    [anon_sym_exception] = ACTIONS(2254),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2256),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2460),
    [anon_sym_lazy] = ACTIONS(2258),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2260),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1260)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(5197),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1260),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_RPAREN] = ACTIONS(469),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_COLON_COLON] = ACTIONS(1191),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1261)] = {
    [sym__delimited_binding_pattern] = STATE(4914),
    [sym__simple_binding_pattern] = STATE(4921),
    [sym__effect_binding_pattern] = STATE(4846),
    [sym__binding_pattern] = STATE(4918),
    [sym_alias_binding_pattern] = STATE(4846),
    [sym_typed_binding_pattern] = STATE(4914),
    [sym__or_binding_pattern_anonymous] = STATE(4848),
    [sym_constructor_binding_pattern] = STATE(4921),
    [sym_tag_binding_pattern] = STATE(4921),
    [sym_polymorphic_variant_pattern] = STATE(4914),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6450),
    [sym__tuple_binding_pattern] = STATE(4850),
    [sym_record_binding_pattern] = STATE(4922),
    [sym_list_binding_pattern] = STATE(4922),
    [sym_cons_binding_pattern] = STATE(4846),
    [sym_array_binding_pattern] = STATE(4922),
    [sym_range_pattern] = STATE(4846),
    [sym_lazy_binding_pattern] = STATE(4921),
    [sym_local_open_binding_pattern] = STATE(4914),
    [sym_package_pattern] = STATE(4914),
    [sym_parenthesized_binding_pattern] = STATE(4922),
    [sym_exception_binding_pattern] = STATE(4846),
    [sym_effect_binding_pattern] = STATE(4846),
    [sym_attribute] = STATE(1261),
    [sym_extension] = STATE(4914),
    [sym_quoted_extension] = STATE(4914),
    [sym__attribute] = STATE(1317),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4739),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4922),
    [sym__value_name] = STATE(4914),
    [sym_parenthesized_operator] = STATE(4863),
    [sym_module_path] = STATE(6362),
    [sym_constructor_path] = STATE(1094),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1128),
    [sym__lowercase_identifier] = ACTIONS(2246),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(2264),
    [anon_sym_LBRACK] = ACTIONS(2250),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2252),
    [anon_sym_exception] = ACTIONS(2254),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2256),
    [anon_sym_lazy] = ACTIONS(2258),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2260),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1262)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5240),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1262),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__attribute] = STATE(1381),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1263)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5250),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1263),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__attribute] = STATE(1383),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1264)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5252),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1264),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__attribute] = STATE(1377),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1265)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4771),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(6815),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym__sequence_pattern_content] = STATE(6470),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1265),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(1042),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1097),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_RBRACK] = ACTIONS(2462),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(2198),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(2200),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2202),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1266)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4866),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6673),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1266),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(4808),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1038),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1120),
    [sym__lowercase_identifier] = ACTIONS(2180),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_RBRACK] = ACTIONS(2464),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(2184),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(2188),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2190),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1267)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4728),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(4730),
    [sym__tuple_pattern] = STATE(4316),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1267),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(1042),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1097),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2466),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_DOT_DOT] = ACTIONS(2468),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(2198),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(2200),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2202),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1268)] = {
    [sym_match_case] = STATE(2150),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4784),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1268),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1269)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4844),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(4852),
    [sym__tuple_binding_pattern] = STATE(4581),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1269),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(2470),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_DOT_DOT] = ACTIONS(2472),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1270)] = {
    [sym_match_case] = STATE(2273),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4806),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1270),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1271)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4932),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(6815),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1271),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(1042),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1097),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_RBRACK] = ACTIONS(2474),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(2198),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(2200),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2202),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1272)] = {
    [sym_match_case] = STATE(2306),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4688),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1272),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1273)] = {
    [sym__delimited_binding_pattern] = STATE(2051),
    [sym__simple_binding_pattern] = STATE(2085),
    [sym__effect_binding_pattern] = STATE(2060),
    [sym__binding_pattern] = STATE(2009),
    [sym_alias_binding_pattern] = STATE(2060),
    [sym_typed_binding_pattern] = STATE(2051),
    [sym__or_binding_pattern_anonymous] = STATE(2061),
    [sym_constructor_binding_pattern] = STATE(2085),
    [sym_tag_binding_pattern] = STATE(2085),
    [sym_polymorphic_variant_pattern] = STATE(2051),
    [sym_labeled_tuple_element_binding_pattern] = STATE(2027),
    [sym__tuple_binding_pattern] = STATE(2033),
    [sym_record_binding_pattern] = STATE(2087),
    [sym_list_binding_pattern] = STATE(2087),
    [sym_cons_binding_pattern] = STATE(2060),
    [sym_array_binding_pattern] = STATE(2087),
    [sym_range_pattern] = STATE(2060),
    [sym_lazy_binding_pattern] = STATE(2085),
    [sym_local_open_binding_pattern] = STATE(2051),
    [sym_package_pattern] = STATE(2051),
    [sym_parenthesized_binding_pattern] = STATE(2087),
    [sym_exception_binding_pattern] = STATE(2060),
    [sym_effect_binding_pattern] = STATE(2060),
    [sym_attribute] = STATE(1273),
    [sym_extension] = STATE(2051),
    [sym_quoted_extension] = STATE(2051),
    [sym__constant] = STATE(1823),
    [sym__signed_constant] = STATE(1825),
    [sym_number] = STATE(1803),
    [sym_signed_number] = STATE(1823),
    [sym_character] = STATE(1803),
    [sym_string] = STATE(1803),
    [sym_quoted_string] = STATE(1803),
    [sym_boolean] = STATE(1803),
    [sym_unit] = STATE(2087),
    [sym__value_name] = STATE(2051),
    [sym_parenthesized_operator] = STATE(2035),
    [sym_module_path] = STATE(6855),
    [sym_constructor_path] = STATE(714),
    [sym__constructor_name] = STATE(1698),
    [sym_tag] = STATE(716),
    [sym__lowercase_identifier] = ACTIONS(1293),
    [anon_sym_TILDE] = ACTIONS(2476),
    [anon_sym_LPAREN] = ACTIONS(1347),
    [anon_sym_DOT_DOT] = ACTIONS(2478),
    [anon_sym_LBRACK] = ACTIONS(1303),
    [anon_sym_true] = ACTIONS(1305),
    [anon_sym_false] = ACTIONS(1305),
    [anon_sym_LBRACE] = ACTIONS(1307),
    [anon_sym_exception] = ACTIONS(1309),
    [anon_sym_POUND] = ACTIONS(1311),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1313),
    [anon_sym_lazy] = ACTIONS(1315),
    [anon_sym_begin] = ACTIONS(1317),
    [anon_sym_effect] = ACTIONS(1319),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2408),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2410),
    [aux_sym_number_token1] = ACTIONS(1325),
    [aux_sym_signed_number_token1] = ACTIONS(1327),
    [anon_sym_SQUOTE] = ACTIONS(1329),
    [anon_sym_DQUOTE] = ACTIONS(1331),
    [sym__uppercase_identifier] = ACTIONS(1333),
    [aux_sym_tag_token1] = ACTIONS(1335),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1274)] = {
    [sym__delimited_binding_pattern] = STATE(4914),
    [sym__simple_binding_pattern] = STATE(4921),
    [sym__effect_binding_pattern] = STATE(4846),
    [sym__binding_pattern] = STATE(4888),
    [sym_alias_binding_pattern] = STATE(4846),
    [sym_typed_binding_pattern] = STATE(4914),
    [sym__or_binding_pattern_anonymous] = STATE(4848),
    [sym_constructor_binding_pattern] = STATE(4921),
    [sym_tag_binding_pattern] = STATE(4921),
    [sym_polymorphic_variant_pattern] = STATE(4914),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6450),
    [sym__tuple_binding_pattern] = STATE(4850),
    [sym_record_binding_pattern] = STATE(4922),
    [sym_list_binding_pattern] = STATE(4922),
    [sym_cons_binding_pattern] = STATE(4846),
    [sym_array_binding_pattern] = STATE(4922),
    [sym_range_pattern] = STATE(4846),
    [sym_lazy_binding_pattern] = STATE(4921),
    [sym_local_open_binding_pattern] = STATE(4914),
    [sym_package_pattern] = STATE(4914),
    [sym_parenthesized_binding_pattern] = STATE(4922),
    [sym_exception_binding_pattern] = STATE(4846),
    [sym_effect_binding_pattern] = STATE(4846),
    [sym_attribute] = STATE(1274),
    [sym_extension] = STATE(4914),
    [sym_quoted_extension] = STATE(4914),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4739),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4922),
    [sym__value_name] = STATE(4914),
    [sym_parenthesized_operator] = STATE(4863),
    [sym_module_path] = STATE(6362),
    [sym_constructor_path] = STATE(1094),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1128),
    [sym__lowercase_identifier] = ACTIONS(2246),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(2264),
    [anon_sym_LBRACK] = ACTIONS(2250),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2252),
    [anon_sym_exception] = ACTIONS(2254),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2256),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2480),
    [anon_sym_lazy] = ACTIONS(2258),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2260),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1275)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4536),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(4514),
    [sym__tuple_pattern] = STATE(4316),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1275),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2482),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_DOT_DOT] = ACTIONS(2468),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1276)] = {
    [sym__delimited_pattern] = STATE(4970),
    [sym__simple_pattern] = STATE(4826),
    [sym__effect_pattern] = STATE(4830),
    [sym__pattern] = STATE(4860),
    [sym_alias_pattern] = STATE(4830),
    [sym_typed_pattern] = STATE(4970),
    [sym__or_pattern_anonymous] = STATE(4832),
    [sym_constructor_pattern] = STATE(4826),
    [sym_tag_pattern] = STATE(4826),
    [sym_polymorphic_variant_pattern] = STATE(4970),
    [sym_labeled_tuple_element_pattern] = STATE(4867),
    [sym__tuple_pattern] = STATE(4920),
    [sym_record_pattern] = STATE(4978),
    [sym_list_pattern] = STATE(4978),
    [sym_cons_pattern] = STATE(4830),
    [sym_array_pattern] = STATE(4978),
    [sym_range_pattern] = STATE(4830),
    [sym_lazy_pattern] = STATE(4826),
    [sym_local_open_pattern] = STATE(4970),
    [sym_package_pattern] = STATE(4970),
    [sym_parenthesized_pattern] = STATE(4978),
    [sym_exception_pattern] = STATE(4830),
    [sym_effect_pattern] = STATE(4830),
    [sym_attribute] = STATE(1276),
    [sym_extension] = STATE(4970),
    [sym_quoted_extension] = STATE(4970),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4725),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4978),
    [sym__simple_value_pattern] = STATE(4970),
    [sym_parenthesized_operator] = STATE(4970),
    [sym_module_path] = STATE(6366),
    [sym_constructor_path] = STATE(1088),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1127),
    [sym__lowercase_identifier] = ACTIONS(2208),
    [anon_sym_TILDE] = ACTIONS(2484),
    [anon_sym_LPAREN] = ACTIONS(2262),
    [anon_sym_DOT_DOT] = ACTIONS(2486),
    [anon_sym_LBRACK] = ACTIONS(2212),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2216),
    [anon_sym_exception] = ACTIONS(2218),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2224),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2228),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1277)] = {
    [sym_match_case] = STATE(3278),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4695),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1277),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1278)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(5322),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1278),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_COMMA] = ACTIONS(1295),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1279)] = {
    [sym_match_case] = STATE(2150),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4726),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1279),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1280)] = {
    [sym_match_case] = STATE(2281),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4806),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1280),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1281)] = {
    [sym_match_case] = STATE(2890),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4775),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1281),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1282)] = {
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2961),
    [sym__binding_pattern] = STATE(3006),
    [sym_alias_binding_pattern] = STATE(2961),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_anonymous] = STATE(2962),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(3007),
    [sym__tuple_binding_pattern] = STATE(3008),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern] = STATE(2961),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2961),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_exception_binding_pattern] = STATE(2961),
    [sym_effect_binding_pattern] = STATE(2961),
    [sym_attribute] = STATE(1282),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_TILDE] = ACTIONS(2488),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_DOT_DOT] = ACTIONS(2490),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1959),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [anon_sym_effect] = ACTIONS(1961),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1965),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1283)] = {
    [sym_match_case] = STATE(3189),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4752),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1283),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1284)] = {
    [sym_match_case] = STATE(2890),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4752),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1284),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1285)] = {
    [sym__delimited_binding_pattern] = STATE(4914),
    [sym__simple_binding_pattern] = STATE(4921),
    [sym__effect_binding_pattern] = STATE(4846),
    [sym__binding_pattern] = STATE(4975),
    [sym_alias_binding_pattern] = STATE(4846),
    [sym_typed_binding_pattern] = STATE(4914),
    [sym__or_binding_pattern_anonymous] = STATE(4848),
    [sym_constructor_binding_pattern] = STATE(4921),
    [sym_tag_binding_pattern] = STATE(4921),
    [sym_polymorphic_variant_pattern] = STATE(4914),
    [sym_labeled_tuple_element_binding_pattern] = STATE(4976),
    [sym__tuple_binding_pattern] = STATE(4884),
    [sym_record_binding_pattern] = STATE(4922),
    [sym_list_binding_pattern] = STATE(4922),
    [sym_cons_binding_pattern] = STATE(4846),
    [sym_array_binding_pattern] = STATE(4922),
    [sym_range_pattern] = STATE(4846),
    [sym_lazy_binding_pattern] = STATE(4921),
    [sym_local_open_binding_pattern] = STATE(4914),
    [sym_package_pattern] = STATE(4914),
    [sym_parenthesized_binding_pattern] = STATE(4922),
    [sym_exception_binding_pattern] = STATE(4846),
    [sym_effect_binding_pattern] = STATE(4846),
    [sym_attribute] = STATE(1285),
    [sym_extension] = STATE(4914),
    [sym_quoted_extension] = STATE(4914),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4739),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4922),
    [sym__value_name] = STATE(4914),
    [sym_parenthesized_operator] = STATE(4863),
    [sym_module_path] = STATE(6362),
    [sym_constructor_path] = STATE(1094),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1128),
    [sym__lowercase_identifier] = ACTIONS(2246),
    [anon_sym_TILDE] = ACTIONS(2492),
    [anon_sym_LPAREN] = ACTIONS(2264),
    [anon_sym_DOT_DOT] = ACTIONS(2494),
    [anon_sym_LBRACK] = ACTIONS(2250),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2252),
    [anon_sym_exception] = ACTIONS(2254),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2256),
    [anon_sym_lazy] = ACTIONS(2258),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2260),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1286)] = {
    [sym_match_case] = STATE(2118),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4784),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1286),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1287)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4711),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(4712),
    [sym__tuple_binding_pattern] = STATE(4581),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1287),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(4808),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1038),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1120),
    [sym__lowercase_identifier] = ACTIONS(2180),
    [anon_sym_TILDE] = ACTIONS(2496),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_DOT_DOT] = ACTIONS(2472),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(2184),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(2188),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2190),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1288)] = {
    [sym_match_case] = STATE(3411),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4762),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1288),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1289)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4932),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(6815),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1289),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(1042),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1097),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_RBRACK] = ACTIONS(2498),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(2198),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(2200),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2202),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1290)] = {
    [sym__delimited_pattern] = STATE(4970),
    [sym__simple_pattern] = STATE(4826),
    [sym__effect_pattern] = STATE(4830),
    [sym__pattern] = STATE(4898),
    [sym_alias_pattern] = STATE(4830),
    [sym_typed_pattern] = STATE(4970),
    [sym__or_pattern_anonymous] = STATE(4832),
    [sym_constructor_pattern] = STATE(4826),
    [sym_tag_pattern] = STATE(4826),
    [sym_polymorphic_variant_pattern] = STATE(4970),
    [sym_labeled_tuple_element_pattern] = STATE(7167),
    [sym__tuple_pattern] = STATE(4833),
    [sym_record_pattern] = STATE(4978),
    [sym_list_pattern] = STATE(4978),
    [sym_cons_pattern] = STATE(4830),
    [sym_array_pattern] = STATE(4978),
    [sym_range_pattern] = STATE(4830),
    [sym_lazy_pattern] = STATE(4826),
    [sym_local_open_pattern] = STATE(4970),
    [sym_package_pattern] = STATE(4970),
    [sym_parenthesized_pattern] = STATE(4978),
    [sym_exception_pattern] = STATE(4830),
    [sym_effect_pattern] = STATE(4830),
    [sym_attribute] = STATE(1290),
    [sym_extension] = STATE(4970),
    [sym_quoted_extension] = STATE(4970),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4725),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4978),
    [sym__simple_value_pattern] = STATE(4970),
    [sym_parenthesized_operator] = STATE(4970),
    [sym_module_path] = STATE(6366),
    [sym_constructor_path] = STATE(1088),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1127),
    [sym__lowercase_identifier] = ACTIONS(2208),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(2262),
    [anon_sym_LBRACK] = ACTIONS(2212),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2216),
    [anon_sym_exception] = ACTIONS(2218),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2222),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2498),
    [anon_sym_lazy] = ACTIONS(2224),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2228),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1291)] = {
    [sym_match_case] = STATE(2273),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4695),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1291),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1292)] = {
    [sym_match_case] = STATE(3273),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4762),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1292),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1293)] = {
    [sym__delimited_binding_pattern] = STATE(2051),
    [sym__simple_binding_pattern] = STATE(2085),
    [sym__effect_binding_pattern] = STATE(2060),
    [sym__binding_pattern] = STATE(1929),
    [sym_alias_binding_pattern] = STATE(2060),
    [sym_typed_binding_pattern] = STATE(2051),
    [sym__or_binding_pattern_anonymous] = STATE(2061),
    [sym_constructor_binding_pattern] = STATE(2085),
    [sym_tag_binding_pattern] = STATE(2085),
    [sym_polymorphic_variant_pattern] = STATE(2051),
    [sym_labeled_tuple_element_binding_pattern] = STATE(1930),
    [sym__tuple_binding_pattern] = STATE(1926),
    [sym_record_binding_pattern] = STATE(2087),
    [sym_list_binding_pattern] = STATE(2087),
    [sym_cons_binding_pattern] = STATE(2060),
    [sym_array_binding_pattern] = STATE(2087),
    [sym_range_pattern] = STATE(2060),
    [sym_lazy_binding_pattern] = STATE(2085),
    [sym_local_open_binding_pattern] = STATE(2051),
    [sym_package_pattern] = STATE(2051),
    [sym_parenthesized_binding_pattern] = STATE(2087),
    [sym_exception_binding_pattern] = STATE(2060),
    [sym_effect_binding_pattern] = STATE(2060),
    [sym_attribute] = STATE(1293),
    [sym_extension] = STATE(2051),
    [sym_quoted_extension] = STATE(2051),
    [sym__constant] = STATE(1823),
    [sym__signed_constant] = STATE(1825),
    [sym_number] = STATE(1803),
    [sym_signed_number] = STATE(1823),
    [sym_character] = STATE(1803),
    [sym_string] = STATE(1803),
    [sym_quoted_string] = STATE(1803),
    [sym_boolean] = STATE(1803),
    [sym_unit] = STATE(2087),
    [sym__value_name] = STATE(2051),
    [sym_parenthesized_operator] = STATE(2035),
    [sym_module_path] = STATE(6855),
    [sym_constructor_path] = STATE(714),
    [sym__constructor_name] = STATE(1698),
    [sym_tag] = STATE(716),
    [sym__lowercase_identifier] = ACTIONS(1293),
    [anon_sym_TILDE] = ACTIONS(2476),
    [anon_sym_LPAREN] = ACTIONS(1347),
    [anon_sym_DOT_DOT] = ACTIONS(2500),
    [anon_sym_LBRACK] = ACTIONS(1303),
    [anon_sym_true] = ACTIONS(1305),
    [anon_sym_false] = ACTIONS(1305),
    [anon_sym_LBRACE] = ACTIONS(1307),
    [anon_sym_exception] = ACTIONS(1309),
    [anon_sym_POUND] = ACTIONS(1311),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1313),
    [anon_sym_lazy] = ACTIONS(1315),
    [anon_sym_begin] = ACTIONS(1317),
    [anon_sym_effect] = ACTIONS(1319),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2408),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2410),
    [aux_sym_number_token1] = ACTIONS(1325),
    [aux_sym_signed_number_token1] = ACTIONS(1327),
    [anon_sym_SQUOTE] = ACTIONS(1329),
    [anon_sym_DQUOTE] = ACTIONS(1331),
    [sym__uppercase_identifier] = ACTIONS(1333),
    [aux_sym_tag_token1] = ACTIONS(1335),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1294)] = {
    [sym_match_case] = STATE(2150),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4688),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1294),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1295)] = {
    [sym__delimited_binding_pattern] = STATE(4914),
    [sym__simple_binding_pattern] = STATE(4921),
    [sym__effect_binding_pattern] = STATE(4846),
    [sym__binding_pattern] = STATE(4888),
    [sym_alias_binding_pattern] = STATE(4846),
    [sym_typed_binding_pattern] = STATE(4914),
    [sym__or_binding_pattern_anonymous] = STATE(4848),
    [sym_constructor_binding_pattern] = STATE(4921),
    [sym_tag_binding_pattern] = STATE(4921),
    [sym_polymorphic_variant_pattern] = STATE(4914),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6450),
    [sym__tuple_binding_pattern] = STATE(4850),
    [sym_record_binding_pattern] = STATE(4922),
    [sym_list_binding_pattern] = STATE(4922),
    [sym_cons_binding_pattern] = STATE(4846),
    [sym_array_binding_pattern] = STATE(4922),
    [sym_range_pattern] = STATE(4846),
    [sym_lazy_binding_pattern] = STATE(4921),
    [sym_local_open_binding_pattern] = STATE(4914),
    [sym_package_pattern] = STATE(4914),
    [sym_parenthesized_binding_pattern] = STATE(4922),
    [sym_exception_binding_pattern] = STATE(4846),
    [sym_effect_binding_pattern] = STATE(4846),
    [sym_attribute] = STATE(1295),
    [sym_extension] = STATE(4914),
    [sym_quoted_extension] = STATE(4914),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4739),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4922),
    [sym__value_name] = STATE(4914),
    [sym_parenthesized_operator] = STATE(4863),
    [sym_module_path] = STATE(6362),
    [sym_constructor_path] = STATE(1094),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1128),
    [sym__lowercase_identifier] = ACTIONS(2246),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(2264),
    [anon_sym_LBRACK] = ACTIONS(2250),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2252),
    [anon_sym_exception] = ACTIONS(2254),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2256),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2464),
    [anon_sym_lazy] = ACTIONS(2258),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2260),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1296)] = {
    [sym_match_case] = STATE(2105),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4726),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1296),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1297)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5263),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1297),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_COMMA] = ACTIONS(2075),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1298)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4866),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6673),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1298),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(4808),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1038),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1120),
    [sym__lowercase_identifier] = ACTIONS(2180),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_RBRACK] = ACTIONS(2480),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(2184),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(2188),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2190),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1299)] = {
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2961),
    [sym__binding_pattern] = STATE(3106),
    [sym_alias_binding_pattern] = STATE(2961),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_anonymous] = STATE(2962),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(3107),
    [sym__tuple_binding_pattern] = STATE(2969),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern] = STATE(2961),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2961),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_exception_binding_pattern] = STATE(2961),
    [sym_effect_binding_pattern] = STATE(2961),
    [sym_attribute] = STATE(1299),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_TILDE] = ACTIONS(2488),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_DOT_DOT] = ACTIONS(2502),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1959),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [anon_sym_effect] = ACTIONS(1961),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1965),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1300)] = {
    [sym__delimited_pattern] = STATE(4970),
    [sym__simple_pattern] = STATE(4826),
    [sym__effect_pattern] = STATE(4830),
    [sym__pattern] = STATE(4898),
    [sym_alias_pattern] = STATE(4830),
    [sym_typed_pattern] = STATE(4970),
    [sym__or_pattern_anonymous] = STATE(4832),
    [sym_constructor_pattern] = STATE(4826),
    [sym_tag_pattern] = STATE(4826),
    [sym_polymorphic_variant_pattern] = STATE(4970),
    [sym_labeled_tuple_element_pattern] = STATE(7167),
    [sym__tuple_pattern] = STATE(4833),
    [sym_record_pattern] = STATE(4978),
    [sym_list_pattern] = STATE(4978),
    [sym_cons_pattern] = STATE(4830),
    [sym_array_pattern] = STATE(4978),
    [sym_range_pattern] = STATE(4830),
    [sym_lazy_pattern] = STATE(4826),
    [sym_local_open_pattern] = STATE(4970),
    [sym_package_pattern] = STATE(4970),
    [sym_parenthesized_pattern] = STATE(4978),
    [sym_exception_pattern] = STATE(4830),
    [sym_effect_pattern] = STATE(4830),
    [sym_attribute] = STATE(1300),
    [sym_extension] = STATE(4970),
    [sym_quoted_extension] = STATE(4970),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4725),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4978),
    [sym__simple_value_pattern] = STATE(4970),
    [sym_parenthesized_operator] = STATE(4970),
    [sym_module_path] = STATE(6366),
    [sym_constructor_path] = STATE(1088),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1127),
    [sym__lowercase_identifier] = ACTIONS(2208),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(2262),
    [anon_sym_LBRACK] = ACTIONS(2212),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2216),
    [anon_sym_exception] = ACTIONS(2218),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2222),
    [anon_sym_PIPE_RBRACK] = ACTIONS(2474),
    [anon_sym_lazy] = ACTIONS(2224),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2228),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1301)] = {
    [sym_match_case] = STATE(3133),
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4775),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1301),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1302)] = {
    [sym__delimited_binding_pattern] = STATE(2051),
    [sym__simple_binding_pattern] = STATE(2085),
    [sym__effect_binding_pattern] = STATE(2060),
    [sym__binding_pattern] = STATE(1927),
    [sym_alias_binding_pattern] = STATE(2060),
    [sym_typed_binding_pattern] = STATE(2051),
    [sym__or_binding_pattern_anonymous] = STATE(2061),
    [sym_constructor_binding_pattern] = STATE(2085),
    [sym_tag_binding_pattern] = STATE(2085),
    [sym_polymorphic_variant_pattern] = STATE(2051),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7213),
    [sym__tuple_binding_pattern] = STATE(2062),
    [sym_record_binding_pattern] = STATE(2087),
    [sym_list_binding_pattern] = STATE(2087),
    [sym_cons_binding_pattern] = STATE(2060),
    [sym_array_binding_pattern] = STATE(2087),
    [sym_range_pattern] = STATE(2060),
    [sym_lazy_binding_pattern] = STATE(2085),
    [sym_local_open_binding_pattern] = STATE(2051),
    [sym_package_pattern] = STATE(2051),
    [sym_parenthesized_binding_pattern] = STATE(2087),
    [sym_exception_binding_pattern] = STATE(2060),
    [sym_effect_binding_pattern] = STATE(2060),
    [sym_attribute] = STATE(1302),
    [sym_extension] = STATE(2051),
    [sym_quoted_extension] = STATE(2051),
    [sym__constant] = STATE(1823),
    [sym__signed_constant] = STATE(1825),
    [sym_number] = STATE(1803),
    [sym_signed_number] = STATE(1823),
    [sym_character] = STATE(1803),
    [sym_string] = STATE(1803),
    [sym_quoted_string] = STATE(1803),
    [sym_boolean] = STATE(1803),
    [sym_unit] = STATE(2087),
    [sym__value_name] = STATE(2051),
    [sym_parenthesized_operator] = STATE(2035),
    [sym_module_path] = STATE(6855),
    [sym_constructor_path] = STATE(714),
    [sym__constructor_name] = STATE(1698),
    [sym_tag] = STATE(716),
    [sym__lowercase_identifier] = ACTIONS(1293),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1347),
    [anon_sym_LBRACK] = ACTIONS(1303),
    [anon_sym_true] = ACTIONS(1305),
    [anon_sym_false] = ACTIONS(1305),
    [anon_sym_LBRACE] = ACTIONS(1307),
    [anon_sym_exception] = ACTIONS(1309),
    [anon_sym_POUND] = ACTIONS(1311),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1313),
    [anon_sym_lazy] = ACTIONS(1315),
    [anon_sym_begin] = ACTIONS(1317),
    [anon_sym_effect] = ACTIONS(1319),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2408),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2410),
    [aux_sym_number_token1] = ACTIONS(1325),
    [aux_sym_signed_number_token1] = ACTIONS(1327),
    [anon_sym_SQUOTE] = ACTIONS(1329),
    [anon_sym_DQUOTE] = ACTIONS(1331),
    [sym__uppercase_identifier] = ACTIONS(1333),
    [aux_sym_tag_token1] = ACTIONS(1335),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1303)] = {
    [sym__delimited_binding_pattern] = STATE(2051),
    [sym__simple_binding_pattern] = STATE(2085),
    [sym__effect_binding_pattern] = STATE(2060),
    [sym__binding_pattern] = STATE(2042),
    [sym_alias_binding_pattern] = STATE(2060),
    [sym_typed_binding_pattern] = STATE(2051),
    [sym__or_binding_pattern_anonymous] = STATE(2079),
    [sym_constructor_binding_pattern] = STATE(2085),
    [sym_tag_binding_pattern] = STATE(2085),
    [sym_polymorphic_variant_pattern] = STATE(2051),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7213),
    [sym__tuple_binding_pattern] = STATE(2062),
    [sym_record_binding_pattern] = STATE(2087),
    [sym_list_binding_pattern] = STATE(2087),
    [sym_cons_binding_pattern] = STATE(2060),
    [sym_array_binding_pattern] = STATE(2087),
    [sym_range_pattern] = STATE(2060),
    [sym_lazy_binding_pattern] = STATE(2085),
    [sym_local_open_binding_pattern] = STATE(2051),
    [sym_package_pattern] = STATE(2051),
    [sym_parenthesized_binding_pattern] = STATE(2087),
    [sym_exception_binding_pattern] = STATE(2060),
    [sym_effect_binding_pattern] = STATE(2060),
    [sym_attribute] = STATE(1303),
    [sym_extension] = STATE(2051),
    [sym_quoted_extension] = STATE(2051),
    [sym__constant] = STATE(1823),
    [sym__signed_constant] = STATE(1825),
    [sym_number] = STATE(1803),
    [sym_signed_number] = STATE(1823),
    [sym_character] = STATE(1803),
    [sym_string] = STATE(1803),
    [sym_quoted_string] = STATE(1803),
    [sym_boolean] = STATE(1803),
    [sym_unit] = STATE(2087),
    [sym__value_name] = STATE(2051),
    [sym_parenthesized_operator] = STATE(2035),
    [sym_module_path] = STATE(6855),
    [sym_constructor_path] = STATE(714),
    [sym__constructor_name] = STATE(1698),
    [sym_tag] = STATE(716),
    [sym__lowercase_identifier] = ACTIONS(1293),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1347),
    [anon_sym_LBRACK] = ACTIONS(1303),
    [anon_sym_true] = ACTIONS(1305),
    [anon_sym_false] = ACTIONS(1305),
    [anon_sym_LBRACE] = ACTIONS(1307),
    [anon_sym_exception] = ACTIONS(1309),
    [anon_sym_POUND] = ACTIONS(1311),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1313),
    [anon_sym_lazy] = ACTIONS(1315),
    [anon_sym_begin] = ACTIONS(1317),
    [anon_sym_effect] = ACTIONS(1319),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2408),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2410),
    [aux_sym_number_token1] = ACTIONS(1325),
    [aux_sym_signed_number_token1] = ACTIONS(1327),
    [anon_sym_SQUOTE] = ACTIONS(1329),
    [anon_sym_DQUOTE] = ACTIONS(1331),
    [sym__uppercase_identifier] = ACTIONS(1333),
    [aux_sym_tag_token1] = ACTIONS(1335),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1304)] = {
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2961),
    [sym__binding_pattern] = STATE(3009),
    [sym_alias_binding_pattern] = STATE(2961),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_anonymous] = STATE(3010),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6582),
    [sym__tuple_binding_pattern] = STATE(2964),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern] = STATE(2961),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2961),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_exception_binding_pattern] = STATE(2961),
    [sym_effect_binding_pattern] = STATE(2961),
    [sym_attribute] = STATE(1304),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1959),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [anon_sym_effect] = ACTIONS(1961),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1965),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1305)] = {
    [sym__delimited_pattern] = STATE(4970),
    [sym__simple_pattern] = STATE(4826),
    [sym__effect_pattern] = STATE(4830),
    [sym__pattern] = STATE(4875),
    [sym_alias_pattern] = STATE(4830),
    [sym_typed_pattern] = STATE(4970),
    [sym__or_pattern_anonymous] = STATE(4930),
    [sym_constructor_pattern] = STATE(4826),
    [sym_tag_pattern] = STATE(4826),
    [sym_polymorphic_variant_pattern] = STATE(4970),
    [sym_labeled_tuple_element_pattern] = STATE(7167),
    [sym__tuple_pattern] = STATE(4833),
    [sym_record_pattern] = STATE(4978),
    [sym_list_pattern] = STATE(4978),
    [sym_cons_pattern] = STATE(4830),
    [sym_array_pattern] = STATE(4978),
    [sym_range_pattern] = STATE(4830),
    [sym_lazy_pattern] = STATE(4826),
    [sym_local_open_pattern] = STATE(4970),
    [sym_package_pattern] = STATE(4970),
    [sym_parenthesized_pattern] = STATE(4978),
    [sym_exception_pattern] = STATE(4830),
    [sym_effect_pattern] = STATE(4830),
    [sym_attribute] = STATE(1305),
    [sym_extension] = STATE(4970),
    [sym_quoted_extension] = STATE(4970),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4725),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4978),
    [sym__simple_value_pattern] = STATE(4970),
    [sym_parenthesized_operator] = STATE(4970),
    [sym_module_path] = STATE(6366),
    [sym_constructor_path] = STATE(1088),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1127),
    [sym__lowercase_identifier] = ACTIONS(2208),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(2262),
    [anon_sym_LBRACK] = ACTIONS(2212),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2216),
    [anon_sym_exception] = ACTIONS(2218),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2224),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2228),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1306)] = {
    [sym__delimited_pattern] = STATE(4970),
    [sym__simple_pattern] = STATE(4826),
    [sym__effect_pattern] = STATE(4830),
    [sym__pattern] = STATE(4889),
    [sym_alias_pattern] = STATE(4830),
    [sym_typed_pattern] = STATE(4970),
    [sym__or_pattern_anonymous] = STATE(4832),
    [sym_constructor_pattern] = STATE(4826),
    [sym_tag_pattern] = STATE(4826),
    [sym_polymorphic_variant_pattern] = STATE(4970),
    [sym_labeled_tuple_element_pattern] = STATE(7167),
    [sym__tuple_pattern] = STATE(4833),
    [sym_record_pattern] = STATE(4978),
    [sym_list_pattern] = STATE(4978),
    [sym_cons_pattern] = STATE(4830),
    [sym_array_pattern] = STATE(4978),
    [sym_range_pattern] = STATE(4830),
    [sym_lazy_pattern] = STATE(4826),
    [sym_local_open_pattern] = STATE(4970),
    [sym_package_pattern] = STATE(4970),
    [sym_parenthesized_pattern] = STATE(4978),
    [sym_exception_pattern] = STATE(4830),
    [sym_effect_pattern] = STATE(4830),
    [sym_attribute] = STATE(1306),
    [sym_extension] = STATE(4970),
    [sym_quoted_extension] = STATE(4970),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4725),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4978),
    [sym__simple_value_pattern] = STATE(4970),
    [sym_parenthesized_operator] = STATE(4970),
    [sym_module_path] = STATE(6366),
    [sym_constructor_path] = STATE(1088),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1127),
    [sym__lowercase_identifier] = ACTIONS(2208),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(2262),
    [anon_sym_LBRACK] = ACTIONS(2212),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2216),
    [anon_sym_exception] = ACTIONS(2218),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2224),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2228),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1307)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4759),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1307),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1308)] = {
    [sym__delimited_pattern] = STATE(4970),
    [sym__simple_pattern] = STATE(4826),
    [sym__effect_pattern] = STATE(4830),
    [sym__pattern] = STATE(4947),
    [sym_alias_pattern] = STATE(4830),
    [sym_typed_pattern] = STATE(4970),
    [sym__or_pattern_anonymous] = STATE(4832),
    [sym_constructor_pattern] = STATE(4826),
    [sym_tag_pattern] = STATE(4826),
    [sym_polymorphic_variant_pattern] = STATE(4970),
    [sym_labeled_tuple_element_pattern] = STATE(7167),
    [sym__tuple_pattern] = STATE(4833),
    [sym_record_pattern] = STATE(4978),
    [sym_list_pattern] = STATE(4978),
    [sym_cons_pattern] = STATE(4830),
    [sym_array_pattern] = STATE(4978),
    [sym_range_pattern] = STATE(4830),
    [sym_lazy_pattern] = STATE(4826),
    [sym_local_open_pattern] = STATE(4970),
    [sym_package_pattern] = STATE(4970),
    [sym_parenthesized_pattern] = STATE(4978),
    [sym_exception_pattern] = STATE(4830),
    [sym_effect_pattern] = STATE(4830),
    [sym_attribute] = STATE(1308),
    [sym_extension] = STATE(4970),
    [sym_quoted_extension] = STATE(4970),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4725),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4978),
    [sym__simple_value_pattern] = STATE(4970),
    [sym_parenthesized_operator] = STATE(4970),
    [sym_module_path] = STATE(6366),
    [sym_constructor_path] = STATE(1088),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1127),
    [sym__lowercase_identifier] = ACTIONS(2208),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(2262),
    [anon_sym_LBRACK] = ACTIONS(2212),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2216),
    [anon_sym_exception] = ACTIONS(2218),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2224),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2228),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1309)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4929),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(6815),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1309),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(1042),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1097),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(2198),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(2200),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2202),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1310)] = {
    [sym__delimited_binding_pattern] = STATE(2051),
    [sym__simple_binding_pattern] = STATE(2085),
    [sym__effect_binding_pattern] = STATE(2060),
    [sym__binding_pattern] = STATE(2048),
    [sym_alias_binding_pattern] = STATE(2060),
    [sym_typed_binding_pattern] = STATE(2051),
    [sym__or_binding_pattern_anonymous] = STATE(2061),
    [sym_constructor_binding_pattern] = STATE(2085),
    [sym_tag_binding_pattern] = STATE(2085),
    [sym_polymorphic_variant_pattern] = STATE(2051),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7213),
    [sym__tuple_binding_pattern] = STATE(2062),
    [sym_record_binding_pattern] = STATE(2087),
    [sym_list_binding_pattern] = STATE(2087),
    [sym_cons_binding_pattern] = STATE(2060),
    [sym_array_binding_pattern] = STATE(2087),
    [sym_range_pattern] = STATE(2060),
    [sym_lazy_binding_pattern] = STATE(2085),
    [sym_local_open_binding_pattern] = STATE(2051),
    [sym_package_pattern] = STATE(2051),
    [sym_parenthesized_binding_pattern] = STATE(2087),
    [sym_exception_binding_pattern] = STATE(2060),
    [sym_effect_binding_pattern] = STATE(2060),
    [sym_attribute] = STATE(1310),
    [sym_extension] = STATE(2051),
    [sym_quoted_extension] = STATE(2051),
    [sym__constant] = STATE(1823),
    [sym__signed_constant] = STATE(1825),
    [sym_number] = STATE(1803),
    [sym_signed_number] = STATE(1823),
    [sym_character] = STATE(1803),
    [sym_string] = STATE(1803),
    [sym_quoted_string] = STATE(1803),
    [sym_boolean] = STATE(1803),
    [sym_unit] = STATE(2087),
    [sym__value_name] = STATE(2051),
    [sym_parenthesized_operator] = STATE(2035),
    [sym_module_path] = STATE(6855),
    [sym_constructor_path] = STATE(714),
    [sym__constructor_name] = STATE(1698),
    [sym_tag] = STATE(716),
    [sym__lowercase_identifier] = ACTIONS(1293),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1347),
    [anon_sym_LBRACK] = ACTIONS(1303),
    [anon_sym_true] = ACTIONS(1305),
    [anon_sym_false] = ACTIONS(1305),
    [anon_sym_LBRACE] = ACTIONS(1307),
    [anon_sym_exception] = ACTIONS(1309),
    [anon_sym_POUND] = ACTIONS(1311),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1313),
    [anon_sym_lazy] = ACTIONS(1315),
    [anon_sym_begin] = ACTIONS(1317),
    [anon_sym_effect] = ACTIONS(1319),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2408),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2410),
    [aux_sym_number_token1] = ACTIONS(1325),
    [aux_sym_signed_number_token1] = ACTIONS(1327),
    [anon_sym_SQUOTE] = ACTIONS(1329),
    [anon_sym_DQUOTE] = ACTIONS(1331),
    [sym__uppercase_identifier] = ACTIONS(1333),
    [aux_sym_tag_token1] = ACTIONS(1335),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1311)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(5390),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1311),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1312)] = {
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2961),
    [sym__binding_pattern] = STATE(3012),
    [sym_alias_binding_pattern] = STATE(2961),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_anonymous] = STATE(2962),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6582),
    [sym__tuple_binding_pattern] = STATE(2964),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern] = STATE(2961),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2961),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_exception_binding_pattern] = STATE(2961),
    [sym_effect_binding_pattern] = STATE(2961),
    [sym_attribute] = STATE(1312),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1959),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [anon_sym_effect] = ACTIONS(1961),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1965),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1313)] = {
    [sym__delimited_binding_pattern] = STATE(4914),
    [sym__simple_binding_pattern] = STATE(4921),
    [sym__effect_binding_pattern] = STATE(4846),
    [sym__binding_pattern] = STATE(4888),
    [sym_alias_binding_pattern] = STATE(4846),
    [sym_typed_binding_pattern] = STATE(4914),
    [sym__or_binding_pattern_anonymous] = STATE(4848),
    [sym_constructor_binding_pattern] = STATE(4921),
    [sym_tag_binding_pattern] = STATE(4921),
    [sym_polymorphic_variant_pattern] = STATE(4914),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6450),
    [sym__tuple_binding_pattern] = STATE(4850),
    [sym_record_binding_pattern] = STATE(4922),
    [sym_list_binding_pattern] = STATE(4922),
    [sym_cons_binding_pattern] = STATE(4846),
    [sym_array_binding_pattern] = STATE(4922),
    [sym_range_pattern] = STATE(4846),
    [sym_lazy_binding_pattern] = STATE(4921),
    [sym_local_open_binding_pattern] = STATE(4914),
    [sym_package_pattern] = STATE(4914),
    [sym_parenthesized_binding_pattern] = STATE(4922),
    [sym_exception_binding_pattern] = STATE(4846),
    [sym_effect_binding_pattern] = STATE(4846),
    [sym_attribute] = STATE(1313),
    [sym_extension] = STATE(4914),
    [sym_quoted_extension] = STATE(4914),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4739),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4922),
    [sym__value_name] = STATE(4914),
    [sym_parenthesized_operator] = STATE(4863),
    [sym_module_path] = STATE(6362),
    [sym_constructor_path] = STATE(1094),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1128),
    [sym__lowercase_identifier] = ACTIONS(2246),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(2264),
    [anon_sym_LBRACK] = ACTIONS(2250),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2252),
    [anon_sym_exception] = ACTIONS(2254),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2256),
    [anon_sym_lazy] = ACTIONS(2258),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2260),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1314)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4776),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1314),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1315)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4729),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1315),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1316)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4862),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1316),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1317)] = {
    [sym__delimited_binding_pattern] = STATE(4914),
    [sym__simple_binding_pattern] = STATE(4921),
    [sym__effect_binding_pattern] = STATE(4846),
    [sym__binding_pattern] = STATE(4823),
    [sym_alias_binding_pattern] = STATE(4846),
    [sym_typed_binding_pattern] = STATE(4914),
    [sym__or_binding_pattern_anonymous] = STATE(4848),
    [sym_constructor_binding_pattern] = STATE(4921),
    [sym_tag_binding_pattern] = STATE(4921),
    [sym_polymorphic_variant_pattern] = STATE(4914),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6450),
    [sym__tuple_binding_pattern] = STATE(4850),
    [sym_record_binding_pattern] = STATE(4922),
    [sym_list_binding_pattern] = STATE(4922),
    [sym_cons_binding_pattern] = STATE(4846),
    [sym_array_binding_pattern] = STATE(4922),
    [sym_range_pattern] = STATE(4846),
    [sym_lazy_binding_pattern] = STATE(4921),
    [sym_local_open_binding_pattern] = STATE(4914),
    [sym_package_pattern] = STATE(4914),
    [sym_parenthesized_binding_pattern] = STATE(4922),
    [sym_exception_binding_pattern] = STATE(4846),
    [sym_effect_binding_pattern] = STATE(4846),
    [sym_attribute] = STATE(1317),
    [sym_extension] = STATE(4914),
    [sym_quoted_extension] = STATE(4914),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4739),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4922),
    [sym__value_name] = STATE(4914),
    [sym_parenthesized_operator] = STATE(4863),
    [sym_module_path] = STATE(6362),
    [sym_constructor_path] = STATE(1094),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1128),
    [sym__lowercase_identifier] = ACTIONS(2246),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(2264),
    [anon_sym_LBRACK] = ACTIONS(2250),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2252),
    [anon_sym_exception] = ACTIONS(2254),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2256),
    [anon_sym_lazy] = ACTIONS(2258),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2260),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1318)] = {
    [sym__delimited_pattern] = STATE(4970),
    [sym__simple_pattern] = STATE(4826),
    [sym__effect_pattern] = STATE(4830),
    [sym__pattern] = STATE(4898),
    [sym_alias_pattern] = STATE(4830),
    [sym_typed_pattern] = STATE(4970),
    [sym__or_pattern_anonymous] = STATE(4832),
    [sym_constructor_pattern] = STATE(4826),
    [sym_tag_pattern] = STATE(4826),
    [sym_polymorphic_variant_pattern] = STATE(4970),
    [sym_labeled_tuple_element_pattern] = STATE(7167),
    [sym__tuple_pattern] = STATE(4833),
    [sym_record_pattern] = STATE(4978),
    [sym_list_pattern] = STATE(4978),
    [sym_cons_pattern] = STATE(4830),
    [sym_array_pattern] = STATE(4978),
    [sym_range_pattern] = STATE(4830),
    [sym_lazy_pattern] = STATE(4826),
    [sym_local_open_pattern] = STATE(4970),
    [sym_package_pattern] = STATE(4970),
    [sym_parenthesized_pattern] = STATE(4978),
    [sym_exception_pattern] = STATE(4830),
    [sym_effect_pattern] = STATE(4830),
    [sym_attribute] = STATE(1318),
    [sym_extension] = STATE(4970),
    [sym_quoted_extension] = STATE(4970),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4725),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4978),
    [sym__simple_value_pattern] = STATE(4970),
    [sym_parenthesized_operator] = STATE(4970),
    [sym_module_path] = STATE(6366),
    [sym_constructor_path] = STATE(1088),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1127),
    [sym__lowercase_identifier] = ACTIONS(2208),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(2262),
    [anon_sym_LBRACK] = ACTIONS(2212),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2216),
    [anon_sym_exception] = ACTIONS(2218),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2224),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2228),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1319)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4732),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1319),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1320)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(5425),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1320),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1321)] = {
    [sym__delimited_binding_pattern] = STATE(4914),
    [sym__simple_binding_pattern] = STATE(4921),
    [sym__effect_binding_pattern] = STATE(4846),
    [sym__binding_pattern] = STATE(4835),
    [sym_alias_binding_pattern] = STATE(4846),
    [sym_typed_binding_pattern] = STATE(4914),
    [sym__or_binding_pattern_anonymous] = STATE(4848),
    [sym_constructor_binding_pattern] = STATE(4921),
    [sym_tag_binding_pattern] = STATE(4921),
    [sym_polymorphic_variant_pattern] = STATE(4914),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6450),
    [sym__tuple_binding_pattern] = STATE(4850),
    [sym_record_binding_pattern] = STATE(4922),
    [sym_list_binding_pattern] = STATE(4922),
    [sym_cons_binding_pattern] = STATE(4846),
    [sym_array_binding_pattern] = STATE(4922),
    [sym_range_pattern] = STATE(4846),
    [sym_lazy_binding_pattern] = STATE(4921),
    [sym_local_open_binding_pattern] = STATE(4914),
    [sym_package_pattern] = STATE(4914),
    [sym_parenthesized_binding_pattern] = STATE(4922),
    [sym_exception_binding_pattern] = STATE(4846),
    [sym_effect_binding_pattern] = STATE(4846),
    [sym_attribute] = STATE(1321),
    [sym_extension] = STATE(4914),
    [sym_quoted_extension] = STATE(4914),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4739),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4922),
    [sym__value_name] = STATE(4914),
    [sym_parenthesized_operator] = STATE(4863),
    [sym_module_path] = STATE(6362),
    [sym_constructor_path] = STATE(1094),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1128),
    [sym__lowercase_identifier] = ACTIONS(2246),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(2264),
    [anon_sym_LBRACK] = ACTIONS(2250),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2252),
    [anon_sym_exception] = ACTIONS(2254),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2256),
    [anon_sym_lazy] = ACTIONS(2258),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2260),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1322)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4734),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4323),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(6815),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1322),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(1042),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1097),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(2198),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(2200),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2202),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1323)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4740),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1323),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1324)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4908),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6673),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1324),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(4808),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1038),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1120),
    [sym__lowercase_identifier] = ACTIONS(2180),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(2184),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(2188),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2190),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1325)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4506),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4323),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1325),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1326)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4744),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(6815),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1326),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(1042),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1097),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(2198),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(2200),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2202),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1327)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4741),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1327),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1328)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4503),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1328),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1329)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4841),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1329),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1330)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4330),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(6815),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1330),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(1042),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1097),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(2198),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(2200),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2202),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1331)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4746),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1331),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1332)] = {
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2961),
    [sym__binding_pattern] = STATE(3049),
    [sym_alias_binding_pattern] = STATE(2961),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_anonymous] = STATE(2962),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6582),
    [sym__tuple_binding_pattern] = STATE(2964),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern] = STATE(2961),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2961),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_exception_binding_pattern] = STATE(2961),
    [sym_effect_binding_pattern] = STATE(2961),
    [sym_attribute] = STATE(1332),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1959),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [anon_sym_effect] = ACTIONS(1961),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1965),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1333)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4546),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6673),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1333),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(4808),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1038),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1120),
    [sym__lowercase_identifier] = ACTIONS(2180),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(2184),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(2188),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2190),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1334)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4748),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1334),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1335)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4708),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6673),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1335),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(4808),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1038),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1120),
    [sym__lowercase_identifier] = ACTIONS(2180),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(2184),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(2188),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2190),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1336)] = {
    [sym__delimited_binding_pattern] = STATE(2051),
    [sym__simple_binding_pattern] = STATE(2085),
    [sym__effect_binding_pattern] = STATE(2060),
    [sym__binding_pattern] = STATE(2072),
    [sym_alias_binding_pattern] = STATE(2060),
    [sym_typed_binding_pattern] = STATE(2051),
    [sym__or_binding_pattern_anonymous] = STATE(2061),
    [sym_constructor_binding_pattern] = STATE(2085),
    [sym_tag_binding_pattern] = STATE(2085),
    [sym_polymorphic_variant_pattern] = STATE(2051),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7213),
    [sym__tuple_binding_pattern] = STATE(2062),
    [sym_record_binding_pattern] = STATE(2087),
    [sym_list_binding_pattern] = STATE(2087),
    [sym_cons_binding_pattern] = STATE(2060),
    [sym_array_binding_pattern] = STATE(2087),
    [sym_range_pattern] = STATE(2060),
    [sym_lazy_binding_pattern] = STATE(2085),
    [sym_local_open_binding_pattern] = STATE(2051),
    [sym_package_pattern] = STATE(2051),
    [sym_parenthesized_binding_pattern] = STATE(2087),
    [sym_exception_binding_pattern] = STATE(2060),
    [sym_effect_binding_pattern] = STATE(2060),
    [sym_attribute] = STATE(1336),
    [sym_extension] = STATE(2051),
    [sym_quoted_extension] = STATE(2051),
    [sym__constant] = STATE(1823),
    [sym__signed_constant] = STATE(1825),
    [sym_number] = STATE(1803),
    [sym_signed_number] = STATE(1823),
    [sym_character] = STATE(1803),
    [sym_string] = STATE(1803),
    [sym_quoted_string] = STATE(1803),
    [sym_boolean] = STATE(1803),
    [sym_unit] = STATE(2087),
    [sym__value_name] = STATE(2051),
    [sym_parenthesized_operator] = STATE(2035),
    [sym_module_path] = STATE(6855),
    [sym_constructor_path] = STATE(714),
    [sym__constructor_name] = STATE(1698),
    [sym_tag] = STATE(716),
    [sym__lowercase_identifier] = ACTIONS(1293),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1347),
    [anon_sym_LBRACK] = ACTIONS(1303),
    [anon_sym_true] = ACTIONS(1305),
    [anon_sym_false] = ACTIONS(1305),
    [anon_sym_LBRACE] = ACTIONS(1307),
    [anon_sym_exception] = ACTIONS(1309),
    [anon_sym_POUND] = ACTIONS(1311),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1313),
    [anon_sym_lazy] = ACTIONS(1315),
    [anon_sym_begin] = ACTIONS(1317),
    [anon_sym_effect] = ACTIONS(1319),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2408),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2410),
    [aux_sym_number_token1] = ACTIONS(1325),
    [aux_sym_signed_number_token1] = ACTIONS(1327),
    [anon_sym_SQUOTE] = ACTIONS(1329),
    [anon_sym_DQUOTE] = ACTIONS(1331),
    [sym__uppercase_identifier] = ACTIONS(1333),
    [aux_sym_tag_token1] = ACTIONS(1335),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1337)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4546),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1337),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1338)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4750),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1338),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1339)] = {
    [sym__delimited_binding_pattern] = STATE(2051),
    [sym__simple_binding_pattern] = STATE(2085),
    [sym__effect_binding_pattern] = STATE(2060),
    [sym__binding_pattern] = STATE(1932),
    [sym_alias_binding_pattern] = STATE(2060),
    [sym_typed_binding_pattern] = STATE(2051),
    [sym__or_binding_pattern_anonymous] = STATE(1931),
    [sym_constructor_binding_pattern] = STATE(2085),
    [sym_tag_binding_pattern] = STATE(2085),
    [sym_polymorphic_variant_pattern] = STATE(2051),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7213),
    [sym__tuple_binding_pattern] = STATE(2062),
    [sym_record_binding_pattern] = STATE(2087),
    [sym_list_binding_pattern] = STATE(2087),
    [sym_cons_binding_pattern] = STATE(2060),
    [sym_array_binding_pattern] = STATE(2087),
    [sym_range_pattern] = STATE(2060),
    [sym_lazy_binding_pattern] = STATE(2085),
    [sym_local_open_binding_pattern] = STATE(2051),
    [sym_package_pattern] = STATE(2051),
    [sym_parenthesized_binding_pattern] = STATE(2087),
    [sym_exception_binding_pattern] = STATE(2060),
    [sym_effect_binding_pattern] = STATE(2060),
    [sym_attribute] = STATE(1339),
    [sym_extension] = STATE(2051),
    [sym_quoted_extension] = STATE(2051),
    [sym__constant] = STATE(1823),
    [sym__signed_constant] = STATE(1825),
    [sym_number] = STATE(1803),
    [sym_signed_number] = STATE(1823),
    [sym_character] = STATE(1803),
    [sym_string] = STATE(1803),
    [sym_quoted_string] = STATE(1803),
    [sym_boolean] = STATE(1803),
    [sym_unit] = STATE(2087),
    [sym__value_name] = STATE(2051),
    [sym_parenthesized_operator] = STATE(2035),
    [sym_module_path] = STATE(6855),
    [sym_constructor_path] = STATE(714),
    [sym__constructor_name] = STATE(1698),
    [sym_tag] = STATE(716),
    [sym__lowercase_identifier] = ACTIONS(1293),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1347),
    [anon_sym_LBRACK] = ACTIONS(1303),
    [anon_sym_true] = ACTIONS(1305),
    [anon_sym_false] = ACTIONS(1305),
    [anon_sym_LBRACE] = ACTIONS(1307),
    [anon_sym_exception] = ACTIONS(1309),
    [anon_sym_POUND] = ACTIONS(1311),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1313),
    [anon_sym_lazy] = ACTIONS(1315),
    [anon_sym_begin] = ACTIONS(1317),
    [anon_sym_effect] = ACTIONS(1319),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2408),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2410),
    [aux_sym_number_token1] = ACTIONS(1325),
    [aux_sym_signed_number_token1] = ACTIONS(1327),
    [anon_sym_SQUOTE] = ACTIONS(1329),
    [anon_sym_DQUOTE] = ACTIONS(1331),
    [sym__uppercase_identifier] = ACTIONS(1333),
    [aux_sym_tag_token1] = ACTIONS(1335),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1340)] = {
    [sym__delimited_binding_pattern] = STATE(2051),
    [sym__simple_binding_pattern] = STATE(2085),
    [sym__effect_binding_pattern] = STATE(2060),
    [sym__binding_pattern] = STATE(1933),
    [sym_alias_binding_pattern] = STATE(2060),
    [sym_typed_binding_pattern] = STATE(2051),
    [sym__or_binding_pattern_anonymous] = STATE(2061),
    [sym_constructor_binding_pattern] = STATE(2085),
    [sym_tag_binding_pattern] = STATE(2085),
    [sym_polymorphic_variant_pattern] = STATE(2051),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7213),
    [sym__tuple_binding_pattern] = STATE(2062),
    [sym_record_binding_pattern] = STATE(2087),
    [sym_list_binding_pattern] = STATE(2087),
    [sym_cons_binding_pattern] = STATE(2060),
    [sym_array_binding_pattern] = STATE(2087),
    [sym_range_pattern] = STATE(2060),
    [sym_lazy_binding_pattern] = STATE(2085),
    [sym_local_open_binding_pattern] = STATE(2051),
    [sym_package_pattern] = STATE(2051),
    [sym_parenthesized_binding_pattern] = STATE(2087),
    [sym_exception_binding_pattern] = STATE(2060),
    [sym_effect_binding_pattern] = STATE(2060),
    [sym_attribute] = STATE(1340),
    [sym_extension] = STATE(2051),
    [sym_quoted_extension] = STATE(2051),
    [sym__constant] = STATE(1823),
    [sym__signed_constant] = STATE(1825),
    [sym_number] = STATE(1803),
    [sym_signed_number] = STATE(1823),
    [sym_character] = STATE(1803),
    [sym_string] = STATE(1803),
    [sym_quoted_string] = STATE(1803),
    [sym_boolean] = STATE(1803),
    [sym_unit] = STATE(2087),
    [sym__value_name] = STATE(2051),
    [sym_parenthesized_operator] = STATE(2035),
    [sym_module_path] = STATE(6855),
    [sym_constructor_path] = STATE(714),
    [sym__constructor_name] = STATE(1698),
    [sym_tag] = STATE(716),
    [sym__lowercase_identifier] = ACTIONS(1293),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1347),
    [anon_sym_LBRACK] = ACTIONS(1303),
    [anon_sym_true] = ACTIONS(1305),
    [anon_sym_false] = ACTIONS(1305),
    [anon_sym_LBRACE] = ACTIONS(1307),
    [anon_sym_exception] = ACTIONS(1309),
    [anon_sym_POUND] = ACTIONS(1311),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1313),
    [anon_sym_lazy] = ACTIONS(1315),
    [anon_sym_begin] = ACTIONS(1317),
    [anon_sym_effect] = ACTIONS(1319),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2408),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2410),
    [aux_sym_number_token1] = ACTIONS(1325),
    [aux_sym_signed_number_token1] = ACTIONS(1327),
    [anon_sym_SQUOTE] = ACTIONS(1329),
    [anon_sym_DQUOTE] = ACTIONS(1331),
    [sym__uppercase_identifier] = ACTIONS(1333),
    [aux_sym_tag_token1] = ACTIONS(1335),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1341)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4751),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1341),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1342)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5216),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1342),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1343)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4713),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6673),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1343),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(4808),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1038),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1120),
    [sym__lowercase_identifier] = ACTIONS(2180),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(2184),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(2188),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2190),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1344)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4714),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6673),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1344),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(4808),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1038),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1120),
    [sym__lowercase_identifier] = ACTIONS(2180),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(2184),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(2188),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2190),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1345)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4755),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1345),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1346)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4594),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1346),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1347)] = {
    [sym__delimited_pattern] = STATE(4970),
    [sym__simple_pattern] = STATE(4826),
    [sym__effect_pattern] = STATE(4830),
    [sym__pattern] = STATE(4842),
    [sym_alias_pattern] = STATE(4830),
    [sym_typed_pattern] = STATE(4970),
    [sym__or_pattern_anonymous] = STATE(4832),
    [sym_constructor_pattern] = STATE(4826),
    [sym_tag_pattern] = STATE(4826),
    [sym_polymorphic_variant_pattern] = STATE(4970),
    [sym_labeled_tuple_element_pattern] = STATE(7167),
    [sym__tuple_pattern] = STATE(4833),
    [sym_record_pattern] = STATE(4978),
    [sym_list_pattern] = STATE(4978),
    [sym_cons_pattern] = STATE(4830),
    [sym_array_pattern] = STATE(4978),
    [sym_range_pattern] = STATE(4830),
    [sym_lazy_pattern] = STATE(4826),
    [sym_local_open_pattern] = STATE(4970),
    [sym_package_pattern] = STATE(4970),
    [sym_parenthesized_pattern] = STATE(4978),
    [sym_exception_pattern] = STATE(4830),
    [sym_effect_pattern] = STATE(4830),
    [sym_attribute] = STATE(1347),
    [sym_extension] = STATE(4970),
    [sym_quoted_extension] = STATE(4970),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4725),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4978),
    [sym__simple_value_pattern] = STATE(4970),
    [sym_parenthesized_operator] = STATE(4970),
    [sym_module_path] = STATE(6366),
    [sym_constructor_path] = STATE(1088),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1127),
    [sym__lowercase_identifier] = ACTIONS(2208),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(2262),
    [anon_sym_LBRACK] = ACTIONS(2212),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2216),
    [anon_sym_exception] = ACTIONS(2218),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2224),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2228),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1348)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4758),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1348),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1349)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4811),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1349),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1350)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4743),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1350),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1351)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4941),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(6815),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1351),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(1042),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1097),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(2198),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(2200),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2202),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1352)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4761),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1352),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1353)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5457),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1353),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1354)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4720),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(6815),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1354),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(1042),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1097),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(2198),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(2200),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2202),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1355)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4763),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1355),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1356)] = {
    [sym__delimited_binding_pattern] = STATE(2051),
    [sym__simple_binding_pattern] = STATE(2085),
    [sym__effect_binding_pattern] = STATE(2060),
    [sym__binding_pattern] = STATE(2000),
    [sym_alias_binding_pattern] = STATE(2060),
    [sym_typed_binding_pattern] = STATE(2051),
    [sym__or_binding_pattern_anonymous] = STATE(2061),
    [sym_constructor_binding_pattern] = STATE(2085),
    [sym_tag_binding_pattern] = STATE(2085),
    [sym_polymorphic_variant_pattern] = STATE(2051),
    [sym_labeled_tuple_element_binding_pattern] = STATE(7213),
    [sym__tuple_binding_pattern] = STATE(2062),
    [sym_record_binding_pattern] = STATE(2087),
    [sym_list_binding_pattern] = STATE(2087),
    [sym_cons_binding_pattern] = STATE(2060),
    [sym_array_binding_pattern] = STATE(2087),
    [sym_range_pattern] = STATE(2060),
    [sym_lazy_binding_pattern] = STATE(2085),
    [sym_local_open_binding_pattern] = STATE(2051),
    [sym_package_pattern] = STATE(2051),
    [sym_parenthesized_binding_pattern] = STATE(2087),
    [sym_exception_binding_pattern] = STATE(2060),
    [sym_effect_binding_pattern] = STATE(2060),
    [sym_attribute] = STATE(1356),
    [sym_extension] = STATE(2051),
    [sym_quoted_extension] = STATE(2051),
    [sym__constant] = STATE(1823),
    [sym__signed_constant] = STATE(1825),
    [sym_number] = STATE(1803),
    [sym_signed_number] = STATE(1823),
    [sym_character] = STATE(1803),
    [sym_string] = STATE(1803),
    [sym_quoted_string] = STATE(1803),
    [sym_boolean] = STATE(1803),
    [sym_unit] = STATE(2087),
    [sym__value_name] = STATE(2051),
    [sym_parenthesized_operator] = STATE(2035),
    [sym_module_path] = STATE(6855),
    [sym_constructor_path] = STATE(714),
    [sym__constructor_name] = STATE(1698),
    [sym_tag] = STATE(716),
    [sym__lowercase_identifier] = ACTIONS(1293),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1347),
    [anon_sym_LBRACK] = ACTIONS(1303),
    [anon_sym_true] = ACTIONS(1305),
    [anon_sym_false] = ACTIONS(1305),
    [anon_sym_LBRACE] = ACTIONS(1307),
    [anon_sym_exception] = ACTIONS(1309),
    [anon_sym_POUND] = ACTIONS(1311),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1313),
    [anon_sym_lazy] = ACTIONS(1315),
    [anon_sym_begin] = ACTIONS(1317),
    [anon_sym_effect] = ACTIONS(1319),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2408),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2410),
    [aux_sym_number_token1] = ACTIONS(1325),
    [aux_sym_signed_number_token1] = ACTIONS(1327),
    [anon_sym_SQUOTE] = ACTIONS(1329),
    [anon_sym_DQUOTE] = ACTIONS(1331),
    [sym__uppercase_identifier] = ACTIONS(1333),
    [aux_sym_tag_token1] = ACTIONS(1335),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1357)] = {
    [sym__delimited_pattern] = STATE(4970),
    [sym__simple_pattern] = STATE(4826),
    [sym__effect_pattern] = STATE(4830),
    [sym__pattern] = STATE(4905),
    [sym_alias_pattern] = STATE(4830),
    [sym_typed_pattern] = STATE(4970),
    [sym__or_pattern_anonymous] = STATE(4832),
    [sym_constructor_pattern] = STATE(4826),
    [sym_tag_pattern] = STATE(4826),
    [sym_polymorphic_variant_pattern] = STATE(4970),
    [sym_labeled_tuple_element_pattern] = STATE(7167),
    [sym__tuple_pattern] = STATE(4833),
    [sym_record_pattern] = STATE(4978),
    [sym_list_pattern] = STATE(4978),
    [sym_cons_pattern] = STATE(4830),
    [sym_array_pattern] = STATE(4978),
    [sym_range_pattern] = STATE(4830),
    [sym_lazy_pattern] = STATE(4826),
    [sym_local_open_pattern] = STATE(4970),
    [sym_package_pattern] = STATE(4970),
    [sym_parenthesized_pattern] = STATE(4978),
    [sym_exception_pattern] = STATE(4830),
    [sym_effect_pattern] = STATE(4830),
    [sym_attribute] = STATE(1357),
    [sym_extension] = STATE(4970),
    [sym_quoted_extension] = STATE(4970),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4725),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4978),
    [sym__simple_value_pattern] = STATE(4970),
    [sym_parenthesized_operator] = STATE(4970),
    [sym_module_path] = STATE(6366),
    [sym_constructor_path] = STATE(1088),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1127),
    [sym__lowercase_identifier] = ACTIONS(2208),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(2262),
    [anon_sym_LBRACK] = ACTIONS(2212),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2216),
    [anon_sym_exception] = ACTIONS(2218),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2222),
    [anon_sym_lazy] = ACTIONS(2224),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2228),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1358)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4424),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(6815),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1358),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(1042),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1097),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(2198),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(2200),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2202),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1359)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4765),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1359),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1360)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4866),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6673),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1360),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(4808),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1038),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1120),
    [sym__lowercase_identifier] = ACTIONS(2180),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(2184),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(2188),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2190),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1361)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4330),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1361),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1362)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4766),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1362),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1363)] = {
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2961),
    [sym__binding_pattern] = STATE(3077),
    [sym_alias_binding_pattern] = STATE(2961),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_anonymous] = STATE(2962),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6582),
    [sym__tuple_binding_pattern] = STATE(2964),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern] = STATE(2961),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2961),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_exception_binding_pattern] = STATE(2961),
    [sym_effect_binding_pattern] = STATE(2961),
    [sym_attribute] = STATE(1363),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1959),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [anon_sym_effect] = ACTIONS(1961),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1965),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1364)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4424),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1364),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1365)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4516),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1365),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1366)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4769),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1366),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1367)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4932),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(6815),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1367),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(1042),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1097),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(2198),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(2200),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2202),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1368)] = {
    [sym__delimited_binding_pattern] = STATE(4914),
    [sym__simple_binding_pattern] = STATE(4921),
    [sym__effect_binding_pattern] = STATE(4846),
    [sym__binding_pattern] = STATE(4972),
    [sym_alias_binding_pattern] = STATE(4846),
    [sym_typed_binding_pattern] = STATE(4914),
    [sym__or_binding_pattern_anonymous] = STATE(4848),
    [sym_constructor_binding_pattern] = STATE(4921),
    [sym_tag_binding_pattern] = STATE(4921),
    [sym_polymorphic_variant_pattern] = STATE(4914),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6450),
    [sym__tuple_binding_pattern] = STATE(4850),
    [sym_record_binding_pattern] = STATE(4922),
    [sym_list_binding_pattern] = STATE(4922),
    [sym_cons_binding_pattern] = STATE(4846),
    [sym_array_binding_pattern] = STATE(4922),
    [sym_range_pattern] = STATE(4846),
    [sym_lazy_binding_pattern] = STATE(4921),
    [sym_local_open_binding_pattern] = STATE(4914),
    [sym_package_pattern] = STATE(4914),
    [sym_parenthesized_binding_pattern] = STATE(4922),
    [sym_exception_binding_pattern] = STATE(4846),
    [sym_effect_binding_pattern] = STATE(4846),
    [sym_attribute] = STATE(1368),
    [sym_extension] = STATE(4914),
    [sym_quoted_extension] = STATE(4914),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4739),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4922),
    [sym__value_name] = STATE(4914),
    [sym_parenthesized_operator] = STATE(4863),
    [sym_module_path] = STATE(6362),
    [sym_constructor_path] = STATE(1094),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1128),
    [sym__lowercase_identifier] = ACTIONS(2246),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(2264),
    [anon_sym_LBRACK] = ACTIONS(2250),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2252),
    [anon_sym_exception] = ACTIONS(2254),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2256),
    [anon_sym_lazy] = ACTIONS(2258),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2260),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1369)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4770),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1369),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1370)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4772),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1370),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1371)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4773),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1371),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1372)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4892),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6673),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1372),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(4808),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1038),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1120),
    [sym__lowercase_identifier] = ACTIONS(2180),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(2184),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(2188),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2190),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1373)] = {
    [sym_let_binding] = STATE(3832),
    [sym__delimited_binding_pattern] = STATE(2051),
    [sym__simple_binding_pattern] = STATE(2085),
    [sym__effect_binding_pattern] = STATE(2086),
    [sym__binding_pattern_no_exn] = STATE(742),
    [sym_alias_binding_pattern_no_exn] = STATE(2086),
    [sym_typed_binding_pattern] = STATE(2051),
    [sym__or_binding_pattern_no_exn_anonymous] = STATE(2086),
    [sym_constructor_binding_pattern] = STATE(2085),
    [sym_tag_binding_pattern] = STATE(2085),
    [sym_polymorphic_variant_pattern] = STATE(2051),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6652),
    [sym__tuple_binding_pattern_no_exn] = STATE(2086),
    [sym_record_binding_pattern] = STATE(2087),
    [sym_list_binding_pattern] = STATE(2087),
    [sym_cons_binding_pattern_no_exn] = STATE(2086),
    [sym_array_binding_pattern] = STATE(2087),
    [sym_range_pattern] = STATE(2086),
    [sym_lazy_binding_pattern] = STATE(2085),
    [sym_local_open_binding_pattern] = STATE(2051),
    [sym_package_pattern] = STATE(2051),
    [sym_parenthesized_binding_pattern] = STATE(2087),
    [sym_attribute] = STATE(1373),
    [sym_extension] = STATE(2051),
    [sym_quoted_extension] = STATE(2051),
    [sym__attribute] = STATE(1411),
    [sym__constant] = STATE(1823),
    [sym__signed_constant] = STATE(1825),
    [sym_number] = STATE(1803),
    [sym_signed_number] = STATE(1823),
    [sym_character] = STATE(1803),
    [sym_string] = STATE(1803),
    [sym_quoted_string] = STATE(1803),
    [sym_boolean] = STATE(1803),
    [sym_unit] = STATE(2087),
    [sym__value_name] = STATE(2051),
    [sym_parenthesized_operator] = STATE(2035),
    [sym_module_path] = STATE(6855),
    [sym_constructor_path] = STATE(714),
    [sym__constructor_name] = STATE(1698),
    [sym_tag] = STATE(716),
    [sym__lowercase_identifier] = ACTIONS(1293),
    [anon_sym_rec] = ACTIONS(2504),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1347),
    [anon_sym_LBRACK] = ACTIONS(1303),
    [anon_sym_true] = ACTIONS(1305),
    [anon_sym_false] = ACTIONS(1305),
    [anon_sym_LBRACE] = ACTIONS(1307),
    [anon_sym_POUND] = ACTIONS(1311),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1313),
    [anon_sym_lazy] = ACTIONS(1315),
    [anon_sym_begin] = ACTIONS(1317),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2408),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2410),
    [anon_sym_PERCENT] = ACTIONS(1451),
    [aux_sym_number_token1] = ACTIONS(1325),
    [aux_sym_signed_number_token1] = ACTIONS(1327),
    [anon_sym_SQUOTE] = ACTIONS(1329),
    [anon_sym_DQUOTE] = ACTIONS(1331),
    [sym__uppercase_identifier] = ACTIONS(1333),
    [aux_sym_tag_token1] = ACTIONS(1335),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1374)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5397),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1374),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1375)] = {
    [sym__delimited_binding_pattern] = STATE(4914),
    [sym__simple_binding_pattern] = STATE(4921),
    [sym__effect_binding_pattern] = STATE(4846),
    [sym__binding_pattern] = STATE(4980),
    [sym_alias_binding_pattern] = STATE(4846),
    [sym_typed_binding_pattern] = STATE(4914),
    [sym__or_binding_pattern_anonymous] = STATE(4885),
    [sym_constructor_binding_pattern] = STATE(4921),
    [sym_tag_binding_pattern] = STATE(4921),
    [sym_polymorphic_variant_pattern] = STATE(4914),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6450),
    [sym__tuple_binding_pattern] = STATE(4850),
    [sym_record_binding_pattern] = STATE(4922),
    [sym_list_binding_pattern] = STATE(4922),
    [sym_cons_binding_pattern] = STATE(4846),
    [sym_array_binding_pattern] = STATE(4922),
    [sym_range_pattern] = STATE(4846),
    [sym_lazy_binding_pattern] = STATE(4921),
    [sym_local_open_binding_pattern] = STATE(4914),
    [sym_package_pattern] = STATE(4914),
    [sym_parenthesized_binding_pattern] = STATE(4922),
    [sym_exception_binding_pattern] = STATE(4846),
    [sym_effect_binding_pattern] = STATE(4846),
    [sym_attribute] = STATE(1375),
    [sym_extension] = STATE(4914),
    [sym_quoted_extension] = STATE(4914),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4739),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4922),
    [sym__value_name] = STATE(4914),
    [sym_parenthesized_operator] = STATE(4863),
    [sym_module_path] = STATE(6362),
    [sym_constructor_path] = STATE(1094),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1128),
    [sym__lowercase_identifier] = ACTIONS(2246),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(2264),
    [anon_sym_LBRACK] = ACTIONS(2250),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2252),
    [anon_sym_exception] = ACTIONS(2254),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2256),
    [anon_sym_lazy] = ACTIONS(2258),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2260),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1376)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(4689),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1376),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1377)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5253),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1377),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1378)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4594),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4600),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6673),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1378),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(4808),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1038),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1120),
    [sym__lowercase_identifier] = ACTIONS(2180),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(2184),
    [anon_sym_POUND] = ACTIONS(2186),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(2188),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(2190),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2192),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2194),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1379)] = {
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2961),
    [sym__binding_pattern] = STATE(3099),
    [sym_alias_binding_pattern] = STATE(2961),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_anonymous] = STATE(2962),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6582),
    [sym__tuple_binding_pattern] = STATE(2964),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern] = STATE(2961),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2961),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_exception_binding_pattern] = STATE(2961),
    [sym_effect_binding_pattern] = STATE(2961),
    [sym_attribute] = STATE(1379),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1959),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [anon_sym_effect] = ACTIONS(1961),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1965),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1380)] = {
    [sym__delimited_binding_pattern] = STATE(4914),
    [sym__simple_binding_pattern] = STATE(4921),
    [sym__effect_binding_pattern] = STATE(4846),
    [sym__binding_pattern] = STATE(4985),
    [sym_alias_binding_pattern] = STATE(4846),
    [sym_typed_binding_pattern] = STATE(4914),
    [sym__or_binding_pattern_anonymous] = STATE(4848),
    [sym_constructor_binding_pattern] = STATE(4921),
    [sym_tag_binding_pattern] = STATE(4921),
    [sym_polymorphic_variant_pattern] = STATE(4914),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6450),
    [sym__tuple_binding_pattern] = STATE(4850),
    [sym_record_binding_pattern] = STATE(4922),
    [sym_list_binding_pattern] = STATE(4922),
    [sym_cons_binding_pattern] = STATE(4846),
    [sym_array_binding_pattern] = STATE(4922),
    [sym_range_pattern] = STATE(4846),
    [sym_lazy_binding_pattern] = STATE(4921),
    [sym_local_open_binding_pattern] = STATE(4914),
    [sym_package_pattern] = STATE(4914),
    [sym_parenthesized_binding_pattern] = STATE(4922),
    [sym_exception_binding_pattern] = STATE(4846),
    [sym_effect_binding_pattern] = STATE(4846),
    [sym_attribute] = STATE(1380),
    [sym_extension] = STATE(4914),
    [sym_quoted_extension] = STATE(4914),
    [sym__constant] = STATE(4717),
    [sym__signed_constant] = STATE(4739),
    [sym_number] = STATE(4693),
    [sym_signed_number] = STATE(4717),
    [sym_character] = STATE(4693),
    [sym_string] = STATE(4693),
    [sym_quoted_string] = STATE(4693),
    [sym_boolean] = STATE(4693),
    [sym_unit] = STATE(4922),
    [sym__value_name] = STATE(4914),
    [sym_parenthesized_operator] = STATE(4863),
    [sym_module_path] = STATE(6362),
    [sym_constructor_path] = STATE(1094),
    [sym__constructor_name] = STATE(3229),
    [sym_tag] = STATE(1128),
    [sym__lowercase_identifier] = ACTIONS(2246),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(2264),
    [anon_sym_LBRACK] = ACTIONS(2250),
    [anon_sym_true] = ACTIONS(2214),
    [anon_sym_false] = ACTIONS(2214),
    [anon_sym_LBRACE] = ACTIONS(2252),
    [anon_sym_exception] = ACTIONS(2254),
    [anon_sym_POUND] = ACTIONS(2220),
    [anon_sym_LBRACK_PIPE] = ACTIONS(2256),
    [anon_sym_lazy] = ACTIONS(2258),
    [anon_sym_begin] = ACTIONS(2226),
    [anon_sym_effect] = ACTIONS(2260),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(2230),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(2232),
    [aux_sym_number_token1] = ACTIONS(2234),
    [aux_sym_signed_number_token1] = ACTIONS(2236),
    [anon_sym_SQUOTE] = ACTIONS(2238),
    [anon_sym_DQUOTE] = ACTIONS(2240),
    [sym__uppercase_identifier] = ACTIONS(2242),
    [aux_sym_tag_token1] = ACTIONS(2244),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1381)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5243),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1381),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1382)] = {
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2961),
    [sym__binding_pattern] = STATE(3108),
    [sym_alias_binding_pattern] = STATE(2961),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_anonymous] = STATE(2970),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6582),
    [sym__tuple_binding_pattern] = STATE(2964),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern] = STATE(2961),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2961),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_exception_binding_pattern] = STATE(2961),
    [sym_effect_binding_pattern] = STATE(2961),
    [sym_attribute] = STATE(1382),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1959),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [anon_sym_effect] = ACTIONS(1961),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1965),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1383)] = {
    [sym__delimited_pattern] = STATE(4428),
    [sym__simple_pattern] = STATE(4430),
    [sym__effect_pattern] = STATE(4405),
    [sym__pattern] = STATE(5251),
    [sym_alias_pattern] = STATE(4405),
    [sym_typed_pattern] = STATE(4428),
    [sym__or_pattern_anonymous] = STATE(4322),
    [sym_constructor_pattern] = STATE(4430),
    [sym_tag_pattern] = STATE(4430),
    [sym_polymorphic_variant_pattern] = STATE(4428),
    [sym_labeled_tuple_element_pattern] = STATE(7064),
    [sym__tuple_pattern] = STATE(4418),
    [sym_record_pattern] = STATE(4414),
    [sym_list_pattern] = STATE(4414),
    [sym_cons_pattern] = STATE(4405),
    [sym_array_pattern] = STATE(4414),
    [sym_range_pattern] = STATE(4405),
    [sym_lazy_pattern] = STATE(4430),
    [sym_local_open_pattern] = STATE(4428),
    [sym_package_pattern] = STATE(4428),
    [sym_parenthesized_pattern] = STATE(4414),
    [sym_exception_pattern] = STATE(4405),
    [sym_effect_pattern] = STATE(4405),
    [sym_attribute] = STATE(1383),
    [sym_extension] = STATE(4428),
    [sym_quoted_extension] = STATE(4428),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4285),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4414),
    [sym__simple_value_pattern] = STATE(4428),
    [sym_parenthesized_operator] = STATE(4428),
    [sym_module_path] = STATE(6987),
    [sym_constructor_path] = STATE(908),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(936),
    [sym__lowercase_identifier] = ACTIONS(1099),
    [anon_sym_TILDE] = ACTIONS(2077),
    [anon_sym_LPAREN] = ACTIONS(1103),
    [anon_sym_LBRACK] = ACTIONS(1105),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1107),
    [anon_sym_exception] = ACTIONS(1109),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1115),
    [anon_sym_lazy] = ACTIONS(1117),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1119),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1384)] = {
    [sym__delimited_binding_pattern] = STATE(3023),
    [sym__simple_binding_pattern] = STATE(3024),
    [sym__effect_binding_pattern] = STATE(2961),
    [sym__binding_pattern] = STATE(3110),
    [sym_alias_binding_pattern] = STATE(2961),
    [sym_typed_binding_pattern] = STATE(3023),
    [sym__or_binding_pattern_anonymous] = STATE(2962),
    [sym_constructor_binding_pattern] = STATE(3024),
    [sym_tag_binding_pattern] = STATE(3024),
    [sym_polymorphic_variant_pattern] = STATE(3023),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6582),
    [sym__tuple_binding_pattern] = STATE(2964),
    [sym_record_binding_pattern] = STATE(3051),
    [sym_list_binding_pattern] = STATE(3051),
    [sym_cons_binding_pattern] = STATE(2961),
    [sym_array_binding_pattern] = STATE(3051),
    [sym_range_pattern] = STATE(2961),
    [sym_lazy_binding_pattern] = STATE(3024),
    [sym_local_open_binding_pattern] = STATE(3023),
    [sym_package_pattern] = STATE(3023),
    [sym_parenthesized_binding_pattern] = STATE(3051),
    [sym_exception_binding_pattern] = STATE(2961),
    [sym_effect_binding_pattern] = STATE(2961),
    [sym_attribute] = STATE(1384),
    [sym_extension] = STATE(3023),
    [sym_quoted_extension] = STATE(3023),
    [sym__constant] = STATE(2940),
    [sym__signed_constant] = STATE(2924),
    [sym_number] = STATE(2929),
    [sym_signed_number] = STATE(2940),
    [sym_character] = STATE(2929),
    [sym_string] = STATE(2929),
    [sym_quoted_string] = STATE(2929),
    [sym_boolean] = STATE(2929),
    [sym_unit] = STATE(3051),
    [sym__value_name] = STATE(3023),
    [sym_parenthesized_operator] = STATE(3035),
    [sym_module_path] = STATE(6493),
    [sym_constructor_path] = STATE(838),
    [sym__constructor_name] = STATE(2768),
    [sym_tag] = STATE(862),
    [sym__lowercase_identifier] = ACTIONS(1409),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1413),
    [anon_sym_LBRACK] = ACTIONS(1419),
    [anon_sym_true] = ACTIONS(1421),
    [anon_sym_false] = ACTIONS(1421),
    [anon_sym_LBRACE] = ACTIONS(1423),
    [anon_sym_exception] = ACTIONS(1959),
    [anon_sym_POUND] = ACTIONS(1433),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1435),
    [anon_sym_lazy] = ACTIONS(1437),
    [anon_sym_begin] = ACTIONS(1439),
    [anon_sym_effect] = ACTIONS(1961),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(1963),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(1965),
    [aux_sym_number_token1] = ACTIONS(1453),
    [aux_sym_signed_number_token1] = ACTIONS(1455),
    [anon_sym_SQUOTE] = ACTIONS(1457),
    [anon_sym_DQUOTE] = ACTIONS(1459),
    [sym__uppercase_identifier] = ACTIONS(1461),
    [aux_sym_tag_token1] = ACTIONS(1463),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
  [STATE(1385)] = {
    [sym__delimited_binding_pattern] = STATE(4548),
    [sym__simple_binding_pattern] = STATE(4553),
    [sym__effect_binding_pattern] = STATE(4599),
    [sym__binding_pattern] = STATE(4855),
    [sym_alias_binding_pattern] = STATE(4599),
    [sym_typed_binding_pattern] = STATE(4548),
    [sym__or_binding_pattern_anonymous] = STATE(4579),
    [sym_constructor_binding_pattern] = STATE(4553),
    [sym_tag_binding_pattern] = STATE(4553),
    [sym_polymorphic_variant_pattern] = STATE(4548),
    [sym_labeled_tuple_element_binding_pattern] = STATE(6639),
    [sym__tuple_binding_pattern] = STATE(4602),
    [sym_record_binding_pattern] = STATE(4564),
    [sym_list_binding_pattern] = STATE(4564),
    [sym_cons_binding_pattern] = STATE(4599),
    [sym_array_binding_pattern] = STATE(4564),
    [sym_range_pattern] = STATE(4599),
    [sym_lazy_binding_pattern] = STATE(4553),
    [sym_local_open_binding_pattern] = STATE(4548),
    [sym_package_pattern] = STATE(4548),
    [sym_parenthesized_binding_pattern] = STATE(4564),
    [sym_exception_binding_pattern] = STATE(4599),
    [sym_effect_binding_pattern] = STATE(4599),
    [sym_attribute] = STATE(1385),
    [sym_extension] = STATE(4548),
    [sym_quoted_extension] = STATE(4548),
    [sym__constant] = STATE(4254),
    [sym__signed_constant] = STATE(4524),
    [sym_number] = STATE(4100),
    [sym_signed_number] = STATE(4254),
    [sym_character] = STATE(4100),
    [sym_string] = STATE(4100),
    [sym_quoted_string] = STATE(4100),
    [sym_boolean] = STATE(4100),
    [sym_unit] = STATE(4564),
    [sym__value_name] = STATE(4548),
    [sym_parenthesized_operator] = STATE(3501),
    [sym_module_path] = STATE(6379),
    [sym_constructor_path] = STATE(1123),
    [sym__constructor_name] = STATE(2008),
    [sym_tag] = STATE(1129),
    [sym__lowercase_identifier] = ACTIONS(1035),
    [anon_sym_TILDE] = ACTIONS(1299),
    [anon_sym_LPAREN] = ACTIONS(1039),
    [anon_sym_LBRACK] = ACTIONS(1047),
    [anon_sym_true] = ACTIONS(1049),
    [anon_sym_false] = ACTIONS(1049),
    [anon_sym_LBRACE] = ACTIONS(1051),
    [anon_sym_exception] = ACTIONS(1053),
    [anon_sym_POUND] = ACTIONS(2083),
    [anon_sym_LBRACK_PIPE] = ACTIONS(1061),
    [anon_sym_lazy] = ACTIONS(1063),
    [anon_sym_begin] = ACTIONS(1065),
    [anon_sym_effect] = ACTIONS(1067),
    [aux_sym_attribute_token1] = ACTIONS(23),
    [anon_sym_LBRACK_PERCENT] = ACTIONS(25),
    [anon_sym_LBRACE_PERCENT] = ACTIONS(27),
    [aux_sym_number_token1] = ACTIONS(1069),
    [aux_sym_signed_number_token1] = ACTIONS(2196),
    [anon_sym_SQUOTE] = ACTIONS(1073),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [sym__uppercase_identifier] = ACTIONS(1079),
    [aux_sym_tag_token1] = ACTIONS(1081),
    [sym_comment] = ACTIONS(5),
    [sym_line_number_directive] = ACTIONS(5),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 34,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2506), 1,
      sym__lowercase_identifier,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(1386), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(4678), 1,
      sym_field_path,
    STATE(5317), 1,
      sym__simple_expression,
    STATE(5886), 1,
      sym_field_expression,
    STATE(7186), 1,
      sym__quoted_string,
    STATE(7248), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [132] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2534), 1,
      anon_sym_LT_DASH,
    STATE(1387), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1707), 21,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1705), 37,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [208] = 37,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1197), 1,
      anon_sym_RPAREN,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1409), 1,
      sym__lowercase_identifier,
    ACTIONS(1413), 1,
      anon_sym_LPAREN,
    ACTIONS(1419), 1,
      anon_sym_LBRACK,
    ACTIONS(1423), 1,
      anon_sym_LBRACE,
    ACTIONS(1433), 1,
      anon_sym_POUND,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1437), 1,
      anon_sym_lazy,
    ACTIONS(1439), 1,
      anon_sym_begin,
    ACTIONS(1453), 1,
      aux_sym_number_token1,
    ACTIONS(1455), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1457), 1,
      anon_sym_SQUOTE,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(1461), 1,
      sym__uppercase_identifier,
    ACTIONS(1463), 1,
      aux_sym_tag_token1,
    ACTIONS(1963), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1965), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(838), 1,
      sym_constructor_path,
    STATE(862), 1,
      sym_tag,
    STATE(1185), 1,
      sym__binding_pattern_no_exn,
    STATE(1388), 1,
      sym_attribute,
    STATE(2768), 1,
      sym__constructor_name,
    STATE(2924), 1,
      sym__signed_constant,
    STATE(3035), 1,
      sym_parenthesized_operator,
    STATE(5451), 1,
      sym_let_binding,
    STATE(6493), 1,
      sym_module_path,
    STATE(6580), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1421), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2940), 2,
      sym__constant,
      sym_signed_number,
    STATE(3024), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2929), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3051), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2976), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3023), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [346] = 37,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1237), 1,
      anon_sym_RPAREN,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1409), 1,
      sym__lowercase_identifier,
    ACTIONS(1413), 1,
      anon_sym_LPAREN,
    ACTIONS(1419), 1,
      anon_sym_LBRACK,
    ACTIONS(1423), 1,
      anon_sym_LBRACE,
    ACTIONS(1433), 1,
      anon_sym_POUND,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1437), 1,
      anon_sym_lazy,
    ACTIONS(1439), 1,
      anon_sym_begin,
    ACTIONS(1453), 1,
      aux_sym_number_token1,
    ACTIONS(1455), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1457), 1,
      anon_sym_SQUOTE,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(1461), 1,
      sym__uppercase_identifier,
    ACTIONS(1463), 1,
      aux_sym_tag_token1,
    ACTIONS(1963), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1965), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(838), 1,
      sym_constructor_path,
    STATE(862), 1,
      sym_tag,
    STATE(1185), 1,
      sym__binding_pattern_no_exn,
    STATE(1389), 1,
      sym_attribute,
    STATE(2768), 1,
      sym__constructor_name,
    STATE(2924), 1,
      sym__signed_constant,
    STATE(3035), 1,
      sym_parenthesized_operator,
    STATE(5451), 1,
      sym_let_binding,
    STATE(6493), 1,
      sym_module_path,
    STATE(6580), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1421), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2940), 2,
      sym__constant,
      sym_signed_number,
    STATE(3024), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2929), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3051), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2976), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3023), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [484] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2536), 1,
      anon_sym_LT_DASH,
    STATE(1390), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1707), 25,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      anon_sym_GT_RBRACE,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1705), 33,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_object,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_then,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [560] = 37,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1213), 1,
      anon_sym_RPAREN,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1409), 1,
      sym__lowercase_identifier,
    ACTIONS(1413), 1,
      anon_sym_LPAREN,
    ACTIONS(1419), 1,
      anon_sym_LBRACK,
    ACTIONS(1423), 1,
      anon_sym_LBRACE,
    ACTIONS(1433), 1,
      anon_sym_POUND,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1437), 1,
      anon_sym_lazy,
    ACTIONS(1439), 1,
      anon_sym_begin,
    ACTIONS(1453), 1,
      aux_sym_number_token1,
    ACTIONS(1455), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1457), 1,
      anon_sym_SQUOTE,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(1461), 1,
      sym__uppercase_identifier,
    ACTIONS(1463), 1,
      aux_sym_tag_token1,
    ACTIONS(1963), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1965), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(838), 1,
      sym_constructor_path,
    STATE(862), 1,
      sym_tag,
    STATE(1185), 1,
      sym__binding_pattern_no_exn,
    STATE(1391), 1,
      sym_attribute,
    STATE(2768), 1,
      sym__constructor_name,
    STATE(2924), 1,
      sym__signed_constant,
    STATE(3035), 1,
      sym_parenthesized_operator,
    STATE(5451), 1,
      sym_let_binding,
    STATE(6493), 1,
      sym_module_path,
    STATE(6580), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1421), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2940), 2,
      sym__constant,
      sym_signed_number,
    STATE(3024), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2929), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3051), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2976), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3023), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [698] = 37,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1211), 1,
      anon_sym_RPAREN,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1409), 1,
      sym__lowercase_identifier,
    ACTIONS(1413), 1,
      anon_sym_LPAREN,
    ACTIONS(1419), 1,
      anon_sym_LBRACK,
    ACTIONS(1423), 1,
      anon_sym_LBRACE,
    ACTIONS(1433), 1,
      anon_sym_POUND,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1437), 1,
      anon_sym_lazy,
    ACTIONS(1439), 1,
      anon_sym_begin,
    ACTIONS(1453), 1,
      aux_sym_number_token1,
    ACTIONS(1455), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1457), 1,
      anon_sym_SQUOTE,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(1461), 1,
      sym__uppercase_identifier,
    ACTIONS(1463), 1,
      aux_sym_tag_token1,
    ACTIONS(1963), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1965), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(838), 1,
      sym_constructor_path,
    STATE(862), 1,
      sym_tag,
    STATE(1185), 1,
      sym__binding_pattern_no_exn,
    STATE(1392), 1,
      sym_attribute,
    STATE(2768), 1,
      sym__constructor_name,
    STATE(2924), 1,
      sym__signed_constant,
    STATE(3035), 1,
      sym_parenthesized_operator,
    STATE(5451), 1,
      sym_let_binding,
    STATE(6493), 1,
      sym_module_path,
    STATE(6580), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1421), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2940), 2,
      sym__constant,
      sym_signed_number,
    STATE(3024), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2929), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3051), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2976), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3023), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [836] = 37,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1229), 1,
      anon_sym_RPAREN,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1409), 1,
      sym__lowercase_identifier,
    ACTIONS(1413), 1,
      anon_sym_LPAREN,
    ACTIONS(1419), 1,
      anon_sym_LBRACK,
    ACTIONS(1423), 1,
      anon_sym_LBRACE,
    ACTIONS(1433), 1,
      anon_sym_POUND,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1437), 1,
      anon_sym_lazy,
    ACTIONS(1439), 1,
      anon_sym_begin,
    ACTIONS(1453), 1,
      aux_sym_number_token1,
    ACTIONS(1455), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1457), 1,
      anon_sym_SQUOTE,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(1461), 1,
      sym__uppercase_identifier,
    ACTIONS(1463), 1,
      aux_sym_tag_token1,
    ACTIONS(1963), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1965), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(838), 1,
      sym_constructor_path,
    STATE(862), 1,
      sym_tag,
    STATE(1185), 1,
      sym__binding_pattern_no_exn,
    STATE(1393), 1,
      sym_attribute,
    STATE(2768), 1,
      sym__constructor_name,
    STATE(2924), 1,
      sym__signed_constant,
    STATE(3035), 1,
      sym_parenthesized_operator,
    STATE(5451), 1,
      sym_let_binding,
    STATE(6493), 1,
      sym_module_path,
    STATE(6580), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1421), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2940), 2,
      sym__constant,
      sym_signed_number,
    STATE(3024), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2929), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3051), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2976), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3023), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [974] = 37,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1205), 1,
      anon_sym_RPAREN,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1409), 1,
      sym__lowercase_identifier,
    ACTIONS(1413), 1,
      anon_sym_LPAREN,
    ACTIONS(1419), 1,
      anon_sym_LBRACK,
    ACTIONS(1423), 1,
      anon_sym_LBRACE,
    ACTIONS(1433), 1,
      anon_sym_POUND,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1437), 1,
      anon_sym_lazy,
    ACTIONS(1439), 1,
      anon_sym_begin,
    ACTIONS(1453), 1,
      aux_sym_number_token1,
    ACTIONS(1455), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1457), 1,
      anon_sym_SQUOTE,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(1461), 1,
      sym__uppercase_identifier,
    ACTIONS(1463), 1,
      aux_sym_tag_token1,
    ACTIONS(1963), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1965), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(838), 1,
      sym_constructor_path,
    STATE(862), 1,
      sym_tag,
    STATE(1185), 1,
      sym__binding_pattern_no_exn,
    STATE(1394), 1,
      sym_attribute,
    STATE(2768), 1,
      sym__constructor_name,
    STATE(2924), 1,
      sym__signed_constant,
    STATE(3035), 1,
      sym_parenthesized_operator,
    STATE(5451), 1,
      sym_let_binding,
    STATE(6493), 1,
      sym_module_path,
    STATE(6580), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1421), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2940), 2,
      sym__constant,
      sym_signed_number,
    STATE(3024), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2929), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3051), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2976), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3023), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [1112] = 37,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1221), 1,
      anon_sym_RPAREN,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1409), 1,
      sym__lowercase_identifier,
    ACTIONS(1413), 1,
      anon_sym_LPAREN,
    ACTIONS(1419), 1,
      anon_sym_LBRACK,
    ACTIONS(1423), 1,
      anon_sym_LBRACE,
    ACTIONS(1433), 1,
      anon_sym_POUND,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1437), 1,
      anon_sym_lazy,
    ACTIONS(1439), 1,
      anon_sym_begin,
    ACTIONS(1453), 1,
      aux_sym_number_token1,
    ACTIONS(1455), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1457), 1,
      anon_sym_SQUOTE,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(1461), 1,
      sym__uppercase_identifier,
    ACTIONS(1463), 1,
      aux_sym_tag_token1,
    ACTIONS(1963), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1965), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(838), 1,
      sym_constructor_path,
    STATE(862), 1,
      sym_tag,
    STATE(1185), 1,
      sym__binding_pattern_no_exn,
    STATE(1395), 1,
      sym_attribute,
    STATE(2768), 1,
      sym__constructor_name,
    STATE(2924), 1,
      sym__signed_constant,
    STATE(3035), 1,
      sym_parenthesized_operator,
    STATE(5451), 1,
      sym_let_binding,
    STATE(6493), 1,
      sym_module_path,
    STATE(6580), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1421), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2940), 2,
      sym__constant,
      sym_signed_number,
    STATE(3024), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2929), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3051), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2976), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3023), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [1250] = 37,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1235), 1,
      anon_sym_RPAREN,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1409), 1,
      sym__lowercase_identifier,
    ACTIONS(1413), 1,
      anon_sym_LPAREN,
    ACTIONS(1419), 1,
      anon_sym_LBRACK,
    ACTIONS(1423), 1,
      anon_sym_LBRACE,
    ACTIONS(1433), 1,
      anon_sym_POUND,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1437), 1,
      anon_sym_lazy,
    ACTIONS(1439), 1,
      anon_sym_begin,
    ACTIONS(1453), 1,
      aux_sym_number_token1,
    ACTIONS(1455), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1457), 1,
      anon_sym_SQUOTE,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(1461), 1,
      sym__uppercase_identifier,
    ACTIONS(1463), 1,
      aux_sym_tag_token1,
    ACTIONS(1963), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1965), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(838), 1,
      sym_constructor_path,
    STATE(862), 1,
      sym_tag,
    STATE(1185), 1,
      sym__binding_pattern_no_exn,
    STATE(1396), 1,
      sym_attribute,
    STATE(2768), 1,
      sym__constructor_name,
    STATE(2924), 1,
      sym__signed_constant,
    STATE(3035), 1,
      sym_parenthesized_operator,
    STATE(5451), 1,
      sym_let_binding,
    STATE(6493), 1,
      sym_module_path,
    STATE(6580), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1421), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2940), 2,
      sym__constant,
      sym_signed_number,
    STATE(3024), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2929), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3051), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2976), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3023), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [1388] = 37,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1199), 1,
      anon_sym_RPAREN,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1409), 1,
      sym__lowercase_identifier,
    ACTIONS(1413), 1,
      anon_sym_LPAREN,
    ACTIONS(1419), 1,
      anon_sym_LBRACK,
    ACTIONS(1423), 1,
      anon_sym_LBRACE,
    ACTIONS(1433), 1,
      anon_sym_POUND,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1437), 1,
      anon_sym_lazy,
    ACTIONS(1439), 1,
      anon_sym_begin,
    ACTIONS(1453), 1,
      aux_sym_number_token1,
    ACTIONS(1455), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1457), 1,
      anon_sym_SQUOTE,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(1461), 1,
      sym__uppercase_identifier,
    ACTIONS(1463), 1,
      aux_sym_tag_token1,
    ACTIONS(1963), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1965), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(838), 1,
      sym_constructor_path,
    STATE(862), 1,
      sym_tag,
    STATE(1185), 1,
      sym__binding_pattern_no_exn,
    STATE(1397), 1,
      sym_attribute,
    STATE(2768), 1,
      sym__constructor_name,
    STATE(2924), 1,
      sym__signed_constant,
    STATE(3035), 1,
      sym_parenthesized_operator,
    STATE(5451), 1,
      sym_let_binding,
    STATE(6493), 1,
      sym_module_path,
    STATE(6580), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1421), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2940), 2,
      sym__constant,
      sym_signed_number,
    STATE(3024), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2929), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3051), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2976), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3023), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [1526] = 37,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1183), 1,
      anon_sym_RPAREN,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1409), 1,
      sym__lowercase_identifier,
    ACTIONS(1413), 1,
      anon_sym_LPAREN,
    ACTIONS(1419), 1,
      anon_sym_LBRACK,
    ACTIONS(1423), 1,
      anon_sym_LBRACE,
    ACTIONS(1433), 1,
      anon_sym_POUND,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1437), 1,
      anon_sym_lazy,
    ACTIONS(1439), 1,
      anon_sym_begin,
    ACTIONS(1453), 1,
      aux_sym_number_token1,
    ACTIONS(1455), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1457), 1,
      anon_sym_SQUOTE,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(1461), 1,
      sym__uppercase_identifier,
    ACTIONS(1463), 1,
      aux_sym_tag_token1,
    ACTIONS(1963), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1965), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(838), 1,
      sym_constructor_path,
    STATE(862), 1,
      sym_tag,
    STATE(1185), 1,
      sym__binding_pattern_no_exn,
    STATE(1398), 1,
      sym_attribute,
    STATE(2768), 1,
      sym__constructor_name,
    STATE(2924), 1,
      sym__signed_constant,
    STATE(3035), 1,
      sym_parenthesized_operator,
    STATE(5451), 1,
      sym_let_binding,
    STATE(6493), 1,
      sym_module_path,
    STATE(6580), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1421), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2940), 2,
      sym__constant,
      sym_signed_number,
    STATE(3024), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2929), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3051), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2976), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3023), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [1664] = 34,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2506), 1,
      sym__lowercase_identifier,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(1399), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(4678), 1,
      sym_field_path,
    STATE(5283), 1,
      sym__simple_expression,
    STATE(5575), 1,
      sym_field_expression,
    STATE(7153), 1,
      sym__quoted_string,
    STATE(7248), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [1796] = 37,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1409), 1,
      sym__lowercase_identifier,
    ACTIONS(1413), 1,
      anon_sym_LPAREN,
    ACTIONS(1419), 1,
      anon_sym_LBRACK,
    ACTIONS(1423), 1,
      anon_sym_LBRACE,
    ACTIONS(1433), 1,
      anon_sym_POUND,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1437), 1,
      anon_sym_lazy,
    ACTIONS(1439), 1,
      anon_sym_begin,
    ACTIONS(1453), 1,
      aux_sym_number_token1,
    ACTIONS(1455), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1457), 1,
      anon_sym_SQUOTE,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(1461), 1,
      sym__uppercase_identifier,
    ACTIONS(1463), 1,
      aux_sym_tag_token1,
    ACTIONS(1517), 1,
      anon_sym_rec,
    ACTIONS(1963), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1965), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(838), 1,
      sym_constructor_path,
    STATE(862), 1,
      sym_tag,
    STATE(1185), 1,
      sym__binding_pattern_no_exn,
    STATE(1400), 1,
      sym_attribute,
    STATE(2768), 1,
      sym__constructor_name,
    STATE(2924), 1,
      sym__signed_constant,
    STATE(3035), 1,
      sym_parenthesized_operator,
    STATE(5505), 1,
      sym_let_binding,
    STATE(6493), 1,
      sym_module_path,
    STATE(6580), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1421), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2940), 2,
      sym__constant,
      sym_signed_number,
    STATE(3024), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2929), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3051), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2976), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3023), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [1934] = 37,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1193), 1,
      anon_sym_RPAREN,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1409), 1,
      sym__lowercase_identifier,
    ACTIONS(1413), 1,
      anon_sym_LPAREN,
    ACTIONS(1419), 1,
      anon_sym_LBRACK,
    ACTIONS(1423), 1,
      anon_sym_LBRACE,
    ACTIONS(1433), 1,
      anon_sym_POUND,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1437), 1,
      anon_sym_lazy,
    ACTIONS(1439), 1,
      anon_sym_begin,
    ACTIONS(1453), 1,
      aux_sym_number_token1,
    ACTIONS(1455), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1457), 1,
      anon_sym_SQUOTE,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(1461), 1,
      sym__uppercase_identifier,
    ACTIONS(1463), 1,
      aux_sym_tag_token1,
    ACTIONS(1963), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1965), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(838), 1,
      sym_constructor_path,
    STATE(862), 1,
      sym_tag,
    STATE(1185), 1,
      sym__binding_pattern_no_exn,
    STATE(1401), 1,
      sym_attribute,
    STATE(2768), 1,
      sym__constructor_name,
    STATE(2924), 1,
      sym__signed_constant,
    STATE(3035), 1,
      sym_parenthesized_operator,
    STATE(5451), 1,
      sym_let_binding,
    STATE(6493), 1,
      sym_module_path,
    STATE(6580), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1421), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2940), 2,
      sym__constant,
      sym_signed_number,
    STATE(3024), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2929), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3051), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2976), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3023), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [2072] = 34,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2506), 1,
      sym__lowercase_identifier,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(1402), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(4678), 1,
      sym_field_path,
    STATE(5391), 1,
      sym__simple_expression,
    STATE(5595), 1,
      sym_field_expression,
    STATE(7248), 1,
      sym_module_path,
    STATE(7307), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [2204] = 34,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2506), 1,
      sym__lowercase_identifier,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(1403), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(4678), 1,
      sym_field_path,
    STATE(5391), 1,
      sym__simple_expression,
    STATE(5595), 1,
      sym_field_expression,
    STATE(6787), 1,
      sym__quoted_string,
    STATE(7248), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [2336] = 34,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2506), 1,
      sym__lowercase_identifier,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(1404), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(4678), 1,
      sym_field_path,
    STATE(5409), 1,
      sym__simple_expression,
    STATE(5693), 1,
      sym_field_expression,
    STATE(6801), 1,
      sym__quoted_string,
    STATE(7248), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [2468] = 34,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2506), 1,
      sym__lowercase_identifier,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(1405), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(4678), 1,
      sym_field_path,
    STATE(5419), 1,
      sym__simple_expression,
    STATE(5751), 1,
      sym_field_expression,
    STATE(6677), 1,
      sym__quoted_string,
    STATE(7248), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [2600] = 34,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2506), 1,
      sym__lowercase_identifier,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(1406), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(4678), 1,
      sym_field_path,
    STATE(5427), 1,
      sym__simple_expression,
    STATE(5789), 1,
      sym_field_expression,
    STATE(6773), 1,
      sym__quoted_string,
    STATE(7248), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [2732] = 34,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2506), 1,
      sym__lowercase_identifier,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(1407), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(4678), 1,
      sym_field_path,
    STATE(5435), 1,
      sym__simple_expression,
    STATE(5845), 1,
      sym_field_expression,
    STATE(6839), 1,
      sym__quoted_string,
    STATE(7248), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [2864] = 34,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2506), 1,
      sym__lowercase_identifier,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(1408), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(4678), 1,
      sym_field_path,
    STATE(5443), 1,
      sym__simple_expression,
    STATE(5892), 1,
      sym_field_expression,
    STATE(6956), 1,
      sym__quoted_string,
    STATE(7248), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [2996] = 34,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2506), 1,
      sym__lowercase_identifier,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(1409), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(4678), 1,
      sym_field_path,
    STATE(5450), 1,
      sym__simple_expression,
    STATE(5765), 1,
      sym_field_expression,
    STATE(7032), 1,
      sym__quoted_string,
    STATE(7248), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [3128] = 34,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2506), 1,
      sym__lowercase_identifier,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(1410), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(4678), 1,
      sym_field_path,
    STATE(5456), 1,
      sym__simple_expression,
    STATE(5562), 1,
      sym_field_expression,
    STATE(7076), 1,
      sym__quoted_string,
    STATE(7248), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [3260] = 37,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1293), 1,
      sym__lowercase_identifier,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1303), 1,
      anon_sym_LBRACK,
    ACTIONS(1307), 1,
      anon_sym_LBRACE,
    ACTIONS(1311), 1,
      anon_sym_POUND,
    ACTIONS(1313), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1315), 1,
      anon_sym_lazy,
    ACTIONS(1317), 1,
      anon_sym_begin,
    ACTIONS(1325), 1,
      aux_sym_number_token1,
    ACTIONS(1327), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1329), 1,
      anon_sym_SQUOTE,
    ACTIONS(1331), 1,
      anon_sym_DQUOTE,
    ACTIONS(1333), 1,
      sym__uppercase_identifier,
    ACTIONS(1335), 1,
      aux_sym_tag_token1,
    ACTIONS(1347), 1,
      anon_sym_LPAREN,
    ACTIONS(2408), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2410), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2538), 1,
      anon_sym_rec,
    STATE(714), 1,
      sym_constructor_path,
    STATE(716), 1,
      sym_tag,
    STATE(742), 1,
      sym__binding_pattern_no_exn,
    STATE(1411), 1,
      sym_attribute,
    STATE(1698), 1,
      sym__constructor_name,
    STATE(1825), 1,
      sym__signed_constant,
    STATE(2035), 1,
      sym_parenthesized_operator,
    STATE(3834), 1,
      sym_let_binding,
    STATE(6652), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6855), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1305), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1823), 2,
      sym__constant,
      sym_signed_number,
    STATE(2085), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(1803), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2087), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2086), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2051), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [3398] = 36,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1293), 1,
      sym__lowercase_identifier,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1303), 1,
      anon_sym_LBRACK,
    ACTIONS(1307), 1,
      anon_sym_LBRACE,
    ACTIONS(1311), 1,
      anon_sym_POUND,
    ACTIONS(1313), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1315), 1,
      anon_sym_lazy,
    ACTIONS(1317), 1,
      anon_sym_begin,
    ACTIONS(1325), 1,
      aux_sym_number_token1,
    ACTIONS(1327), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1329), 1,
      anon_sym_SQUOTE,
    ACTIONS(1331), 1,
      anon_sym_DQUOTE,
    ACTIONS(1333), 1,
      sym__uppercase_identifier,
    ACTIONS(1335), 1,
      aux_sym_tag_token1,
    ACTIONS(1347), 1,
      anon_sym_LPAREN,
    ACTIONS(2408), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2410), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(714), 1,
      sym_constructor_path,
    STATE(716), 1,
      sym_tag,
    STATE(742), 1,
      sym__binding_pattern_no_exn,
    STATE(1412), 1,
      sym_attribute,
    STATE(1698), 1,
      sym__constructor_name,
    STATE(1825), 1,
      sym__signed_constant,
    STATE(2035), 1,
      sym_parenthesized_operator,
    STATE(3834), 1,
      sym_let_binding,
    STATE(6652), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6855), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1305), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1823), 2,
      sym__constant,
      sym_signed_number,
    STATE(2085), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(1803), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2087), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2086), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2051), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [3533] = 36,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1293), 1,
      sym__lowercase_identifier,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1303), 1,
      anon_sym_LBRACK,
    ACTIONS(1307), 1,
      anon_sym_LBRACE,
    ACTIONS(1311), 1,
      anon_sym_POUND,
    ACTIONS(1313), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1315), 1,
      anon_sym_lazy,
    ACTIONS(1317), 1,
      anon_sym_begin,
    ACTIONS(1325), 1,
      aux_sym_number_token1,
    ACTIONS(1327), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1329), 1,
      anon_sym_SQUOTE,
    ACTIONS(1331), 1,
      anon_sym_DQUOTE,
    ACTIONS(1333), 1,
      sym__uppercase_identifier,
    ACTIONS(1335), 1,
      aux_sym_tag_token1,
    ACTIONS(1347), 1,
      anon_sym_LPAREN,
    ACTIONS(2408), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2410), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(714), 1,
      sym_constructor_path,
    STATE(716), 1,
      sym_tag,
    STATE(733), 1,
      sym__binding_pattern_no_exn,
    STATE(1413), 1,
      sym_attribute,
    STATE(1698), 1,
      sym__constructor_name,
    STATE(1825), 1,
      sym__signed_constant,
    STATE(2035), 1,
      sym_parenthesized_operator,
    STATE(3605), 1,
      sym_let_binding,
    STATE(6652), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6855), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1305), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1823), 2,
      sym__constant,
      sym_signed_number,
    STATE(2085), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(1803), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2087), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2086), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2051), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [3668] = 36,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1293), 1,
      sym__lowercase_identifier,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1303), 1,
      anon_sym_LBRACK,
    ACTIONS(1307), 1,
      anon_sym_LBRACE,
    ACTIONS(1311), 1,
      anon_sym_POUND,
    ACTIONS(1313), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1315), 1,
      anon_sym_lazy,
    ACTIONS(1317), 1,
      anon_sym_begin,
    ACTIONS(1325), 1,
      aux_sym_number_token1,
    ACTIONS(1327), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1329), 1,
      anon_sym_SQUOTE,
    ACTIONS(1331), 1,
      anon_sym_DQUOTE,
    ACTIONS(1333), 1,
      sym__uppercase_identifier,
    ACTIONS(1335), 1,
      aux_sym_tag_token1,
    ACTIONS(1347), 1,
      anon_sym_LPAREN,
    ACTIONS(2408), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2410), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(714), 1,
      sym_constructor_path,
    STATE(716), 1,
      sym_tag,
    STATE(733), 1,
      sym__binding_pattern_no_exn,
    STATE(1414), 1,
      sym_attribute,
    STATE(1698), 1,
      sym__constructor_name,
    STATE(1825), 1,
      sym__signed_constant,
    STATE(2035), 1,
      sym_parenthesized_operator,
    STATE(3646), 1,
      sym_let_binding,
    STATE(6652), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6855), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1305), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1823), 2,
      sym__constant,
      sym_signed_number,
    STATE(2085), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(1803), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2087), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2086), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2051), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [3803] = 36,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1409), 1,
      sym__lowercase_identifier,
    ACTIONS(1413), 1,
      anon_sym_LPAREN,
    ACTIONS(1419), 1,
      anon_sym_LBRACK,
    ACTIONS(1423), 1,
      anon_sym_LBRACE,
    ACTIONS(1433), 1,
      anon_sym_POUND,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1437), 1,
      anon_sym_lazy,
    ACTIONS(1439), 1,
      anon_sym_begin,
    ACTIONS(1453), 1,
      aux_sym_number_token1,
    ACTIONS(1455), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1457), 1,
      anon_sym_SQUOTE,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(1461), 1,
      sym__uppercase_identifier,
    ACTIONS(1463), 1,
      aux_sym_tag_token1,
    ACTIONS(1963), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1965), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(838), 1,
      sym_constructor_path,
    STATE(862), 1,
      sym_tag,
    STATE(1185), 1,
      sym__binding_pattern_no_exn,
    STATE(1415), 1,
      sym_attribute,
    STATE(2768), 1,
      sym__constructor_name,
    STATE(2924), 1,
      sym__signed_constant,
    STATE(3035), 1,
      sym_parenthesized_operator,
    STATE(5579), 1,
      sym_let_binding,
    STATE(6493), 1,
      sym_module_path,
    STATE(6580), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1421), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2940), 2,
      sym__constant,
      sym_signed_number,
    STATE(3024), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2929), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3051), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2976), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3023), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [3938] = 36,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1409), 1,
      sym__lowercase_identifier,
    ACTIONS(1413), 1,
      anon_sym_LPAREN,
    ACTIONS(1419), 1,
      anon_sym_LBRACK,
    ACTIONS(1423), 1,
      anon_sym_LBRACE,
    ACTIONS(1433), 1,
      anon_sym_POUND,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1437), 1,
      anon_sym_lazy,
    ACTIONS(1439), 1,
      anon_sym_begin,
    ACTIONS(1453), 1,
      aux_sym_number_token1,
    ACTIONS(1455), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1457), 1,
      anon_sym_SQUOTE,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(1461), 1,
      sym__uppercase_identifier,
    ACTIONS(1463), 1,
      aux_sym_tag_token1,
    ACTIONS(1963), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1965), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(838), 1,
      sym_constructor_path,
    STATE(862), 1,
      sym_tag,
    STATE(1185), 1,
      sym__binding_pattern_no_exn,
    STATE(1416), 1,
      sym_attribute,
    STATE(2768), 1,
      sym__constructor_name,
    STATE(2924), 1,
      sym__signed_constant,
    STATE(3035), 1,
      sym_parenthesized_operator,
    STATE(5451), 1,
      sym_let_binding,
    STATE(6493), 1,
      sym_module_path,
    STATE(6580), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1421), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2940), 2,
      sym__constant,
      sym_signed_number,
    STATE(3024), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2929), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3051), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2976), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3023), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [4073] = 36,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1409), 1,
      sym__lowercase_identifier,
    ACTIONS(1413), 1,
      anon_sym_LPAREN,
    ACTIONS(1419), 1,
      anon_sym_LBRACK,
    ACTIONS(1423), 1,
      anon_sym_LBRACE,
    ACTIONS(1433), 1,
      anon_sym_POUND,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1437), 1,
      anon_sym_lazy,
    ACTIONS(1439), 1,
      anon_sym_begin,
    ACTIONS(1453), 1,
      aux_sym_number_token1,
    ACTIONS(1455), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1457), 1,
      anon_sym_SQUOTE,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(1461), 1,
      sym__uppercase_identifier,
    ACTIONS(1463), 1,
      aux_sym_tag_token1,
    ACTIONS(1963), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1965), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(838), 1,
      sym_constructor_path,
    STATE(862), 1,
      sym_tag,
    STATE(1185), 1,
      sym__binding_pattern_no_exn,
    STATE(1417), 1,
      sym_attribute,
    STATE(2768), 1,
      sym__constructor_name,
    STATE(2924), 1,
      sym__signed_constant,
    STATE(3035), 1,
      sym_parenthesized_operator,
    STATE(5366), 1,
      sym_let_binding,
    STATE(6493), 1,
      sym_module_path,
    STATE(6580), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1421), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2940), 2,
      sym__constant,
      sym_signed_number,
    STATE(3024), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2929), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3051), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2976), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3023), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [4208] = 36,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1409), 1,
      sym__lowercase_identifier,
    ACTIONS(1413), 1,
      anon_sym_LPAREN,
    ACTIONS(1419), 1,
      anon_sym_LBRACK,
    ACTIONS(1423), 1,
      anon_sym_LBRACE,
    ACTIONS(1433), 1,
      anon_sym_POUND,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1437), 1,
      anon_sym_lazy,
    ACTIONS(1439), 1,
      anon_sym_begin,
    ACTIONS(1453), 1,
      aux_sym_number_token1,
    ACTIONS(1455), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1457), 1,
      anon_sym_SQUOTE,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(1461), 1,
      sym__uppercase_identifier,
    ACTIONS(1463), 1,
      aux_sym_tag_token1,
    ACTIONS(1963), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1965), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(838), 1,
      sym_constructor_path,
    STATE(862), 1,
      sym_tag,
    STATE(1185), 1,
      sym__binding_pattern_no_exn,
    STATE(1418), 1,
      sym_attribute,
    STATE(2768), 1,
      sym__constructor_name,
    STATE(2924), 1,
      sym__signed_constant,
    STATE(3035), 1,
      sym_parenthesized_operator,
    STATE(5505), 1,
      sym_let_binding,
    STATE(6493), 1,
      sym_module_path,
    STATE(6580), 1,
      sym_labeled_tuple_element_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1421), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2940), 2,
      sym__constant,
      sym_signed_number,
    STATE(3024), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(2929), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3051), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2976), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(3023), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [4343] = 36,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1293), 1,
      sym__lowercase_identifier,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1303), 1,
      anon_sym_LBRACK,
    ACTIONS(1307), 1,
      anon_sym_LBRACE,
    ACTIONS(1311), 1,
      anon_sym_POUND,
    ACTIONS(1313), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1315), 1,
      anon_sym_lazy,
    ACTIONS(1317), 1,
      anon_sym_begin,
    ACTIONS(1325), 1,
      aux_sym_number_token1,
    ACTIONS(1327), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1329), 1,
      anon_sym_SQUOTE,
    ACTIONS(1331), 1,
      anon_sym_DQUOTE,
    ACTIONS(1333), 1,
      sym__uppercase_identifier,
    ACTIONS(1335), 1,
      aux_sym_tag_token1,
    ACTIONS(1347), 1,
      anon_sym_LPAREN,
    ACTIONS(2408), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2410), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(714), 1,
      sym_constructor_path,
    STATE(716), 1,
      sym_tag,
    STATE(742), 1,
      sym__binding_pattern_no_exn,
    STATE(1419), 1,
      sym_attribute,
    STATE(1698), 1,
      sym__constructor_name,
    STATE(1825), 1,
      sym__signed_constant,
    STATE(2035), 1,
      sym_parenthesized_operator,
    STATE(3778), 1,
      sym_let_binding,
    STATE(6652), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6855), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1305), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1823), 2,
      sym__constant,
      sym_signed_number,
    STATE(2085), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(1803), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2087), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2086), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2051), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [4478] = 36,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1293), 1,
      sym__lowercase_identifier,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1303), 1,
      anon_sym_LBRACK,
    ACTIONS(1307), 1,
      anon_sym_LBRACE,
    ACTIONS(1311), 1,
      anon_sym_POUND,
    ACTIONS(1313), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1315), 1,
      anon_sym_lazy,
    ACTIONS(1317), 1,
      anon_sym_begin,
    ACTIONS(1325), 1,
      aux_sym_number_token1,
    ACTIONS(1327), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1329), 1,
      anon_sym_SQUOTE,
    ACTIONS(1331), 1,
      anon_sym_DQUOTE,
    ACTIONS(1333), 1,
      sym__uppercase_identifier,
    ACTIONS(1335), 1,
      aux_sym_tag_token1,
    ACTIONS(1347), 1,
      anon_sym_LPAREN,
    ACTIONS(2408), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2410), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(714), 1,
      sym_constructor_path,
    STATE(716), 1,
      sym_tag,
    STATE(742), 1,
      sym__binding_pattern_no_exn,
    STATE(1420), 1,
      sym_attribute,
    STATE(1698), 1,
      sym__constructor_name,
    STATE(1825), 1,
      sym__signed_constant,
    STATE(2035), 1,
      sym_parenthesized_operator,
    STATE(3796), 1,
      sym_let_binding,
    STATE(6652), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6855), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1305), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1823), 2,
      sym__constant,
      sym_signed_number,
    STATE(2085), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(1803), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2087), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2086), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2051), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [4613] = 36,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1293), 1,
      sym__lowercase_identifier,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1303), 1,
      anon_sym_LBRACK,
    ACTIONS(1307), 1,
      anon_sym_LBRACE,
    ACTIONS(1311), 1,
      anon_sym_POUND,
    ACTIONS(1313), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1315), 1,
      anon_sym_lazy,
    ACTIONS(1317), 1,
      anon_sym_begin,
    ACTIONS(1325), 1,
      aux_sym_number_token1,
    ACTIONS(1327), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1329), 1,
      anon_sym_SQUOTE,
    ACTIONS(1331), 1,
      anon_sym_DQUOTE,
    ACTIONS(1333), 1,
      sym__uppercase_identifier,
    ACTIONS(1335), 1,
      aux_sym_tag_token1,
    ACTIONS(1347), 1,
      anon_sym_LPAREN,
    ACTIONS(2408), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2410), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(714), 1,
      sym_constructor_path,
    STATE(716), 1,
      sym_tag,
    STATE(733), 1,
      sym__binding_pattern_no_exn,
    STATE(1421), 1,
      sym_attribute,
    STATE(1698), 1,
      sym__constructor_name,
    STATE(1825), 1,
      sym__signed_constant,
    STATE(2035), 1,
      sym_parenthesized_operator,
    STATE(3778), 1,
      sym_let_binding,
    STATE(6652), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6855), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1305), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1823), 2,
      sym__constant,
      sym_signed_number,
    STATE(2085), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(1803), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2087), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2086), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2051), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [4748] = 36,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1293), 1,
      sym__lowercase_identifier,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1303), 1,
      anon_sym_LBRACK,
    ACTIONS(1307), 1,
      anon_sym_LBRACE,
    ACTIONS(1311), 1,
      anon_sym_POUND,
    ACTIONS(1313), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1315), 1,
      anon_sym_lazy,
    ACTIONS(1317), 1,
      anon_sym_begin,
    ACTIONS(1325), 1,
      aux_sym_number_token1,
    ACTIONS(1327), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1329), 1,
      anon_sym_SQUOTE,
    ACTIONS(1331), 1,
      anon_sym_DQUOTE,
    ACTIONS(1333), 1,
      sym__uppercase_identifier,
    ACTIONS(1335), 1,
      aux_sym_tag_token1,
    ACTIONS(1347), 1,
      anon_sym_LPAREN,
    ACTIONS(2408), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2410), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(714), 1,
      sym_constructor_path,
    STATE(716), 1,
      sym_tag,
    STATE(733), 1,
      sym__binding_pattern_no_exn,
    STATE(1422), 1,
      sym_attribute,
    STATE(1698), 1,
      sym__constructor_name,
    STATE(1825), 1,
      sym__signed_constant,
    STATE(2035), 1,
      sym_parenthesized_operator,
    STATE(3623), 1,
      sym_let_binding,
    STATE(6652), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6855), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1305), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1823), 2,
      sym__constant,
      sym_signed_number,
    STATE(2085), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(1803), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2087), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2086), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2051), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [4883] = 36,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1293), 1,
      sym__lowercase_identifier,
    ACTIONS(1299), 1,
      anon_sym_TILDE,
    ACTIONS(1303), 1,
      anon_sym_LBRACK,
    ACTIONS(1307), 1,
      anon_sym_LBRACE,
    ACTIONS(1311), 1,
      anon_sym_POUND,
    ACTIONS(1313), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1315), 1,
      anon_sym_lazy,
    ACTIONS(1317), 1,
      anon_sym_begin,
    ACTIONS(1325), 1,
      aux_sym_number_token1,
    ACTIONS(1327), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1329), 1,
      anon_sym_SQUOTE,
    ACTIONS(1331), 1,
      anon_sym_DQUOTE,
    ACTIONS(1333), 1,
      sym__uppercase_identifier,
    ACTIONS(1335), 1,
      aux_sym_tag_token1,
    ACTIONS(1347), 1,
      anon_sym_LPAREN,
    ACTIONS(2408), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2410), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(714), 1,
      sym_constructor_path,
    STATE(716), 1,
      sym_tag,
    STATE(742), 1,
      sym__binding_pattern_no_exn,
    STATE(1423), 1,
      sym_attribute,
    STATE(1698), 1,
      sym__constructor_name,
    STATE(1825), 1,
      sym__signed_constant,
    STATE(2035), 1,
      sym_parenthesized_operator,
    STATE(3832), 1,
      sym_let_binding,
    STATE(6652), 1,
      sym_labeled_tuple_element_binding_pattern,
    STATE(6855), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1305), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1823), 2,
      sym__constant,
      sym_signed_number,
    STATE(2085), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(1803), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2087), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2086), 6,
      sym__effect_binding_pattern,
      sym_alias_binding_pattern_no_exn,
      sym__or_binding_pattern_no_exn_anonymous,
      sym__tuple_binding_pattern_no_exn,
      sym_cons_binding_pattern_no_exn,
      sym_range_pattern,
    STATE(2051), 8,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__value_name,
  [5018] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2540), 1,
      anon_sym_EQ,
    ACTIONS(2542), 1,
      anon_sym_COLON2,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(1424), 1,
      sym_attribute,
    STATE(1444), 1,
      aux_sym_let_binding_repeat1,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(4967), 1,
      sym__class_typed,
    STATE(5129), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(1977), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [5152] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2546), 1,
      sym__lowercase_identifier,
    ACTIONS(2548), 1,
      anon_sym_LPAREN,
    ACTIONS(2550), 1,
      anon_sym_LBRACK,
    ACTIONS(2554), 1,
      anon_sym_LBRACE,
    ACTIONS(2556), 1,
      anon_sym_object,
    ACTIONS(2558), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2560), 1,
      anon_sym_new,
    ACTIONS(2562), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2564), 1,
      anon_sym_begin,
    ACTIONS(2566), 1,
      sym_ocamlyacc_value,
    ACTIONS(2568), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2570), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2572), 1,
      anon_sym_PERCENT,
    ACTIONS(2574), 1,
      aux_sym_number_token1,
    ACTIONS(2576), 1,
      anon_sym_SQUOTE,
    ACTIONS(2578), 1,
      anon_sym_DQUOTE,
    ACTIONS(2580), 1,
      sym_prefix_operator,
    ACTIONS(2582), 1,
      sym__uppercase_identifier,
    ACTIONS(2584), 1,
      aux_sym_tag_token1,
    STATE(1425), 1,
      sym_attribute,
    STATE(1496), 1,
      sym__attribute,
    STATE(1924), 1,
      sym_parenthesized_operator,
    STATE(1977), 1,
      sym__constructor_name,
    STATE(2029), 1,
      sym__simple_expression,
    STATE(2083), 1,
      sym__value_name,
    STATE(6755), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2552), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1951), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1953), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1937), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [5278] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2546), 1,
      sym__lowercase_identifier,
    ACTIONS(2548), 1,
      anon_sym_LPAREN,
    ACTIONS(2550), 1,
      anon_sym_LBRACK,
    ACTIONS(2554), 1,
      anon_sym_LBRACE,
    ACTIONS(2556), 1,
      anon_sym_object,
    ACTIONS(2558), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2560), 1,
      anon_sym_new,
    ACTIONS(2562), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2564), 1,
      anon_sym_begin,
    ACTIONS(2566), 1,
      sym_ocamlyacc_value,
    ACTIONS(2568), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2570), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2572), 1,
      anon_sym_PERCENT,
    ACTIONS(2574), 1,
      aux_sym_number_token1,
    ACTIONS(2576), 1,
      anon_sym_SQUOTE,
    ACTIONS(2578), 1,
      anon_sym_DQUOTE,
    ACTIONS(2580), 1,
      sym_prefix_operator,
    ACTIONS(2582), 1,
      sym__uppercase_identifier,
    ACTIONS(2584), 1,
      aux_sym_tag_token1,
    STATE(1426), 1,
      sym_attribute,
    STATE(1497), 1,
      sym__attribute,
    STATE(1924), 1,
      sym_parenthesized_operator,
    STATE(1977), 1,
      sym__constructor_name,
    STATE(2030), 1,
      sym__simple_expression,
    STATE(2083), 1,
      sym__value_name,
    STATE(6755), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2552), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1951), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1953), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1937), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [5404] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2542), 1,
      anon_sym_COLON2,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2586), 1,
      anon_sym_EQ,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1427), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(5026), 1,
      sym__class_typed,
    STATE(5211), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2035), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [5538] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2542), 1,
      anon_sym_COLON2,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2588), 1,
      anon_sym_EQ,
    STATE(1428), 1,
      sym_attribute,
    STATE(1430), 1,
      aux_sym_let_binding_repeat1,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(5027), 1,
      sym__class_typed,
    STATE(5042), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2041), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [5672] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2506), 1,
      sym__lowercase_identifier,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(1429), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(4678), 1,
      sym_field_path,
    STATE(5409), 1,
      sym__simple_expression,
    STATE(5693), 1,
      sym_field_expression,
    STATE(7248), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [5798] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2542), 1,
      anon_sym_COLON2,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2590), 1,
      anon_sym_EQ,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1430), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(5035), 1,
      sym__class_typed,
    STATE(5239), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2053), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [5932] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2542), 1,
      anon_sym_COLON2,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2592), 1,
      anon_sym_EQ,
    STATE(1431), 1,
      sym_attribute,
    STATE(1442), 1,
      aux_sym_let_binding_repeat1,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(4943), 1,
      sym__class_typed,
    STATE(5094), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2047), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [6066] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2506), 1,
      sym__lowercase_identifier,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(1432), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(4678), 1,
      sym_field_path,
    STATE(5419), 1,
      sym__simple_expression,
    STATE(5751), 1,
      sym_field_expression,
    STATE(7248), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [6192] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2506), 1,
      sym__lowercase_identifier,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(1433), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(4678), 1,
      sym_field_path,
    STATE(5427), 1,
      sym__simple_expression,
    STATE(5789), 1,
      sym_field_expression,
    STATE(7248), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [6318] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2572), 1,
      anon_sym_PERCENT,
    ACTIONS(2594), 1,
      sym__lowercase_identifier,
    ACTIONS(2596), 1,
      anon_sym_LPAREN,
    ACTIONS(2598), 1,
      anon_sym_LBRACK,
    ACTIONS(2602), 1,
      anon_sym_LBRACE,
    ACTIONS(2604), 1,
      anon_sym_object,
    ACTIONS(2606), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2608), 1,
      anon_sym_new,
    ACTIONS(2610), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2612), 1,
      anon_sym_begin,
    ACTIONS(2614), 1,
      sym_ocamlyacc_value,
    ACTIONS(2616), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2618), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2620), 1,
      aux_sym_number_token1,
    ACTIONS(2622), 1,
      anon_sym_SQUOTE,
    ACTIONS(2624), 1,
      anon_sym_DQUOTE,
    ACTIONS(2626), 1,
      sym_prefix_operator,
    ACTIONS(2628), 1,
      sym__uppercase_identifier,
    ACTIONS(2630), 1,
      aux_sym_tag_token1,
    STATE(1434), 1,
      sym_attribute,
    STATE(1489), 1,
      sym__attribute,
    STATE(1819), 1,
      sym__simple_expression,
    STATE(1822), 1,
      sym_parenthesized_operator,
    STATE(1833), 1,
      sym__constructor_name,
    STATE(1838), 1,
      sym__value_name,
    STATE(6490), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2600), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1837), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1897), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1896), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [6444] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2506), 1,
      sym__lowercase_identifier,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(1435), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(4678), 1,
      sym_field_path,
    STATE(5435), 1,
      sym__simple_expression,
    STATE(5845), 1,
      sym_field_expression,
    STATE(7248), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [6570] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2572), 1,
      anon_sym_PERCENT,
    ACTIONS(2594), 1,
      sym__lowercase_identifier,
    ACTIONS(2596), 1,
      anon_sym_LPAREN,
    ACTIONS(2598), 1,
      anon_sym_LBRACK,
    ACTIONS(2602), 1,
      anon_sym_LBRACE,
    ACTIONS(2604), 1,
      anon_sym_object,
    ACTIONS(2606), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2608), 1,
      anon_sym_new,
    ACTIONS(2610), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2612), 1,
      anon_sym_begin,
    ACTIONS(2614), 1,
      sym_ocamlyacc_value,
    ACTIONS(2616), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2618), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2620), 1,
      aux_sym_number_token1,
    ACTIONS(2622), 1,
      anon_sym_SQUOTE,
    ACTIONS(2624), 1,
      anon_sym_DQUOTE,
    ACTIONS(2626), 1,
      sym_prefix_operator,
    ACTIONS(2628), 1,
      sym__uppercase_identifier,
    ACTIONS(2630), 1,
      aux_sym_tag_token1,
    STATE(1436), 1,
      sym_attribute,
    STATE(1490), 1,
      sym__attribute,
    STATE(1820), 1,
      sym__simple_expression,
    STATE(1822), 1,
      sym_parenthesized_operator,
    STATE(1833), 1,
      sym__constructor_name,
    STATE(1838), 1,
      sym__value_name,
    STATE(6490), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2600), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1837), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1897), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1896), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [6696] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2506), 1,
      sym__lowercase_identifier,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(1437), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(4678), 1,
      sym_field_path,
    STATE(5443), 1,
      sym__simple_expression,
    STATE(5892), 1,
      sym_field_expression,
    STATE(7248), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [6822] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2506), 1,
      sym__lowercase_identifier,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(1438), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(4678), 1,
      sym_field_path,
    STATE(5450), 1,
      sym__simple_expression,
    STATE(5765), 1,
      sym_field_expression,
    STATE(7248), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [6948] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2572), 1,
      anon_sym_PERCENT,
    ACTIONS(2632), 1,
      sym__lowercase_identifier,
    ACTIONS(2634), 1,
      anon_sym_LPAREN,
    ACTIONS(2636), 1,
      anon_sym_LBRACK,
    ACTIONS(2640), 1,
      anon_sym_LBRACE,
    ACTIONS(2642), 1,
      anon_sym_object,
    ACTIONS(2644), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2646), 1,
      anon_sym_new,
    ACTIONS(2648), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2650), 1,
      anon_sym_begin,
    ACTIONS(2652), 1,
      sym_ocamlyacc_value,
    ACTIONS(2654), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2656), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2658), 1,
      aux_sym_number_token1,
    ACTIONS(2660), 1,
      anon_sym_SQUOTE,
    ACTIONS(2662), 1,
      anon_sym_DQUOTE,
    ACTIONS(2664), 1,
      sym_prefix_operator,
    ACTIONS(2666), 1,
      sym__uppercase_identifier,
    ACTIONS(2668), 1,
      aux_sym_tag_token1,
    STATE(1439), 1,
      sym_attribute,
    STATE(1498), 1,
      sym__attribute,
    STATE(2673), 1,
      sym__simple_expression,
    STATE(2675), 1,
      sym_parenthesized_operator,
    STATE(2695), 1,
      sym__constructor_name,
    STATE(2761), 1,
      sym__value_name,
    STATE(6718), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2638), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2683), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2719), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [7074] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2506), 1,
      sym__lowercase_identifier,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(1440), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(4678), 1,
      sym_field_path,
    STATE(5456), 1,
      sym__simple_expression,
    STATE(5562), 1,
      sym_field_expression,
    STATE(7248), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [7200] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2542), 1,
      anon_sym_COLON2,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2670), 1,
      anon_sym_EQ,
    STATE(1441), 1,
      sym_attribute,
    STATE(1450), 1,
      aux_sym_let_binding_repeat1,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(4849), 1,
      sym__class_typed,
    STATE(5052), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2023), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [7334] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2542), 1,
      anon_sym_COLON2,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2672), 1,
      anon_sym_EQ,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1442), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(4851), 1,
      sym__class_typed,
    STATE(5062), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2029), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [7468] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2572), 1,
      anon_sym_PERCENT,
    ACTIONS(2632), 1,
      sym__lowercase_identifier,
    ACTIONS(2634), 1,
      anon_sym_LPAREN,
    ACTIONS(2636), 1,
      anon_sym_LBRACK,
    ACTIONS(2640), 1,
      anon_sym_LBRACE,
    ACTIONS(2642), 1,
      anon_sym_object,
    ACTIONS(2644), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2646), 1,
      anon_sym_new,
    ACTIONS(2648), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2650), 1,
      anon_sym_begin,
    ACTIONS(2652), 1,
      sym_ocamlyacc_value,
    ACTIONS(2654), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2656), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2658), 1,
      aux_sym_number_token1,
    ACTIONS(2660), 1,
      anon_sym_SQUOTE,
    ACTIONS(2662), 1,
      anon_sym_DQUOTE,
    ACTIONS(2664), 1,
      sym_prefix_operator,
    ACTIONS(2666), 1,
      sym__uppercase_identifier,
    ACTIONS(2668), 1,
      aux_sym_tag_token1,
    STATE(1443), 1,
      sym_attribute,
    STATE(1499), 1,
      sym__attribute,
    STATE(2675), 1,
      sym_parenthesized_operator,
    STATE(2695), 1,
      sym__constructor_name,
    STATE(2704), 1,
      sym__simple_expression,
    STATE(2761), 1,
      sym__value_name,
    STATE(6718), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2638), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2683), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2719), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [7594] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2542), 1,
      anon_sym_COLON2,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2674), 1,
      anon_sym_EQ,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1444), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(5012), 1,
      sym__class_typed,
    STATE(5182), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2065), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [7728] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2572), 1,
      anon_sym_PERCENT,
    ACTIONS(2676), 1,
      sym__lowercase_identifier,
    ACTIONS(2678), 1,
      anon_sym_LPAREN,
    ACTIONS(2682), 1,
      anon_sym_LBRACE,
    ACTIONS(2684), 1,
      anon_sym_begin,
    ACTIONS(2686), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2688), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2690), 1,
      aux_sym_number_token1,
    ACTIONS(2692), 1,
      anon_sym_SQUOTE,
    ACTIONS(2694), 1,
      anon_sym_DQUOTE,
    ACTIONS(2696), 1,
      sym_prefix_operator,
    ACTIONS(2698), 1,
      aux_sym_tag_token1,
    STATE(1445), 1,
      sym_attribute,
    STATE(1472), 1,
      sym__attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(1812), 1,
      sym__simple_expression,
    STATE(6354), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2680), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1868), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [7854] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2506), 1,
      sym__lowercase_identifier,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(1446), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(4678), 1,
      sym_field_path,
    STATE(5283), 1,
      sym__simple_expression,
    STATE(5575), 1,
      sym_field_expression,
    STATE(7248), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [7980] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2542), 1,
      anon_sym_COLON2,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2700), 1,
      anon_sym_EQ,
    STATE(1427), 1,
      aux_sym_let_binding_repeat1,
    STATE(1447), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(5015), 1,
      sym__class_typed,
    STATE(5185), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2071), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [8114] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2506), 1,
      sym__lowercase_identifier,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(1448), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(4678), 1,
      sym_field_path,
    STATE(5317), 1,
      sym__simple_expression,
    STATE(5886), 1,
      sym_field_expression,
    STATE(7248), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [8240] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2572), 1,
      anon_sym_PERCENT,
    ACTIONS(2676), 1,
      sym__lowercase_identifier,
    ACTIONS(2678), 1,
      anon_sym_LPAREN,
    ACTIONS(2682), 1,
      anon_sym_LBRACE,
    ACTIONS(2684), 1,
      anon_sym_begin,
    ACTIONS(2686), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2688), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2690), 1,
      aux_sym_number_token1,
    ACTIONS(2692), 1,
      anon_sym_SQUOTE,
    ACTIONS(2694), 1,
      anon_sym_DQUOTE,
    ACTIONS(2696), 1,
      sym_prefix_operator,
    ACTIONS(2698), 1,
      aux_sym_tag_token1,
    STATE(1449), 1,
      sym_attribute,
    STATE(1473), 1,
      sym__attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(1813), 1,
      sym__simple_expression,
    STATE(6354), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2680), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1868), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [8366] = 36,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2542), 1,
      anon_sym_COLON2,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2702), 1,
      anon_sym_EQ,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1450), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(4928), 1,
      sym__class_typed,
    STATE(5070), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(2059), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [8500] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2506), 1,
      sym__lowercase_identifier,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(1451), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(4678), 1,
      sym_field_path,
    STATE(5391), 1,
      sym__simple_expression,
    STATE(5595), 1,
      sym_field_expression,
    STATE(7248), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [8626] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(353), 1,
      anon_sym_LPAREN,
    ACTIONS(355), 1,
      anon_sym_LBRACK,
    ACTIONS(359), 1,
      anon_sym_LBRACE,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(365), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(383), 1,
      anon_sym_new,
    ACTIONS(385), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(387), 1,
      anon_sym_begin,
    ACTIONS(389), 1,
      sym_ocamlyacc_value,
    ACTIONS(395), 1,
      aux_sym_number_token1,
    ACTIONS(397), 1,
      anon_sym_SQUOTE,
    ACTIONS(399), 1,
      anon_sym_DQUOTE,
    ACTIONS(401), 1,
      sym_prefix_operator,
    ACTIONS(405), 1,
      sym__uppercase_identifier,
    ACTIONS(407), 1,
      aux_sym_tag_token1,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(743), 1,
      sym__lowercase_identifier,
    ACTIONS(1197), 1,
      anon_sym_RPAREN,
    STATE(771), 1,
      sym__simple_expression,
    STATE(791), 1,
      sym__value_name,
    STATE(796), 1,
      sym__constructor_name,
    STATE(830), 1,
      sym_parenthesized_operator,
    STATE(1452), 1,
      sym_attribute,
    STATE(7343), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(357), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(795), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(785), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(794), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [8749] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(353), 1,
      anon_sym_LPAREN,
    ACTIONS(355), 1,
      anon_sym_LBRACK,
    ACTIONS(359), 1,
      anon_sym_LBRACE,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(365), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(383), 1,
      anon_sym_new,
    ACTIONS(385), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(387), 1,
      anon_sym_begin,
    ACTIONS(389), 1,
      sym_ocamlyacc_value,
    ACTIONS(395), 1,
      aux_sym_number_token1,
    ACTIONS(397), 1,
      anon_sym_SQUOTE,
    ACTIONS(399), 1,
      anon_sym_DQUOTE,
    ACTIONS(401), 1,
      sym_prefix_operator,
    ACTIONS(405), 1,
      sym__uppercase_identifier,
    ACTIONS(407), 1,
      aux_sym_tag_token1,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(743), 1,
      sym__lowercase_identifier,
    ACTIONS(1235), 1,
      anon_sym_RPAREN,
    STATE(771), 1,
      sym__simple_expression,
    STATE(791), 1,
      sym__value_name,
    STATE(796), 1,
      sym__constructor_name,
    STATE(830), 1,
      sym_parenthesized_operator,
    STATE(1453), 1,
      sym_attribute,
    STATE(7343), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(357), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(795), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(785), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(794), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [8872] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(353), 1,
      anon_sym_LPAREN,
    ACTIONS(355), 1,
      anon_sym_LBRACK,
    ACTIONS(359), 1,
      anon_sym_LBRACE,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(365), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(383), 1,
      anon_sym_new,
    ACTIONS(385), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(387), 1,
      anon_sym_begin,
    ACTIONS(389), 1,
      sym_ocamlyacc_value,
    ACTIONS(395), 1,
      aux_sym_number_token1,
    ACTIONS(397), 1,
      anon_sym_SQUOTE,
    ACTIONS(399), 1,
      anon_sym_DQUOTE,
    ACTIONS(401), 1,
      sym_prefix_operator,
    ACTIONS(405), 1,
      sym__uppercase_identifier,
    ACTIONS(407), 1,
      aux_sym_tag_token1,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(743), 1,
      sym__lowercase_identifier,
    ACTIONS(1237), 1,
      anon_sym_RPAREN,
    STATE(771), 1,
      sym__simple_expression,
    STATE(791), 1,
      sym__value_name,
    STATE(796), 1,
      sym__constructor_name,
    STATE(830), 1,
      sym_parenthesized_operator,
    STATE(1454), 1,
      sym_attribute,
    STATE(7343), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(357), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(795), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(785), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(794), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [8995] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(353), 1,
      anon_sym_LPAREN,
    ACTIONS(355), 1,
      anon_sym_LBRACK,
    ACTIONS(359), 1,
      anon_sym_LBRACE,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(365), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(383), 1,
      anon_sym_new,
    ACTIONS(385), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(387), 1,
      anon_sym_begin,
    ACTIONS(389), 1,
      sym_ocamlyacc_value,
    ACTIONS(395), 1,
      aux_sym_number_token1,
    ACTIONS(397), 1,
      anon_sym_SQUOTE,
    ACTIONS(399), 1,
      anon_sym_DQUOTE,
    ACTIONS(401), 1,
      sym_prefix_operator,
    ACTIONS(405), 1,
      sym__uppercase_identifier,
    ACTIONS(407), 1,
      aux_sym_tag_token1,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(743), 1,
      sym__lowercase_identifier,
    ACTIONS(1213), 1,
      anon_sym_RPAREN,
    STATE(771), 1,
      sym__simple_expression,
    STATE(791), 1,
      sym__value_name,
    STATE(796), 1,
      sym__constructor_name,
    STATE(830), 1,
      sym_parenthesized_operator,
    STATE(1455), 1,
      sym_attribute,
    STATE(7343), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(357), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(795), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(785), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(794), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [9118] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(353), 1,
      anon_sym_LPAREN,
    ACTIONS(355), 1,
      anon_sym_LBRACK,
    ACTIONS(359), 1,
      anon_sym_LBRACE,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(365), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(383), 1,
      anon_sym_new,
    ACTIONS(385), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(387), 1,
      anon_sym_begin,
    ACTIONS(389), 1,
      sym_ocamlyacc_value,
    ACTIONS(395), 1,
      aux_sym_number_token1,
    ACTIONS(397), 1,
      anon_sym_SQUOTE,
    ACTIONS(399), 1,
      anon_sym_DQUOTE,
    ACTIONS(401), 1,
      sym_prefix_operator,
    ACTIONS(405), 1,
      sym__uppercase_identifier,
    ACTIONS(407), 1,
      aux_sym_tag_token1,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(743), 1,
      sym__lowercase_identifier,
    ACTIONS(1199), 1,
      anon_sym_RPAREN,
    STATE(771), 1,
      sym__simple_expression,
    STATE(791), 1,
      sym__value_name,
    STATE(796), 1,
      sym__constructor_name,
    STATE(830), 1,
      sym_parenthesized_operator,
    STATE(1456), 1,
      sym_attribute,
    STATE(7343), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(357), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(795), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(785), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(794), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [9241] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(353), 1,
      anon_sym_LPAREN,
    ACTIONS(355), 1,
      anon_sym_LBRACK,
    ACTIONS(359), 1,
      anon_sym_LBRACE,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(365), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(383), 1,
      anon_sym_new,
    ACTIONS(385), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(387), 1,
      anon_sym_begin,
    ACTIONS(389), 1,
      sym_ocamlyacc_value,
    ACTIONS(395), 1,
      aux_sym_number_token1,
    ACTIONS(397), 1,
      anon_sym_SQUOTE,
    ACTIONS(399), 1,
      anon_sym_DQUOTE,
    ACTIONS(401), 1,
      sym_prefix_operator,
    ACTIONS(405), 1,
      sym__uppercase_identifier,
    ACTIONS(407), 1,
      aux_sym_tag_token1,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(743), 1,
      sym__lowercase_identifier,
    ACTIONS(1193), 1,
      anon_sym_RPAREN,
    STATE(771), 1,
      sym__simple_expression,
    STATE(791), 1,
      sym__value_name,
    STATE(796), 1,
      sym__constructor_name,
    STATE(830), 1,
      sym_parenthesized_operator,
    STATE(1457), 1,
      sym_attribute,
    STATE(7343), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(357), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(795), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(785), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(794), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [9364] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(353), 1,
      anon_sym_LPAREN,
    ACTIONS(355), 1,
      anon_sym_LBRACK,
    ACTIONS(359), 1,
      anon_sym_LBRACE,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(365), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(383), 1,
      anon_sym_new,
    ACTIONS(385), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(387), 1,
      anon_sym_begin,
    ACTIONS(389), 1,
      sym_ocamlyacc_value,
    ACTIONS(395), 1,
      aux_sym_number_token1,
    ACTIONS(397), 1,
      anon_sym_SQUOTE,
    ACTIONS(399), 1,
      anon_sym_DQUOTE,
    ACTIONS(401), 1,
      sym_prefix_operator,
    ACTIONS(405), 1,
      sym__uppercase_identifier,
    ACTIONS(407), 1,
      aux_sym_tag_token1,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(743), 1,
      sym__lowercase_identifier,
    ACTIONS(1229), 1,
      anon_sym_RPAREN,
    STATE(771), 1,
      sym__simple_expression,
    STATE(791), 1,
      sym__value_name,
    STATE(796), 1,
      sym__constructor_name,
    STATE(830), 1,
      sym_parenthesized_operator,
    STATE(1458), 1,
      sym_attribute,
    STATE(7343), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(357), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(795), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(785), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(794), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [9487] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(353), 1,
      anon_sym_LPAREN,
    ACTIONS(355), 1,
      anon_sym_LBRACK,
    ACTIONS(359), 1,
      anon_sym_LBRACE,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(365), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(383), 1,
      anon_sym_new,
    ACTIONS(385), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(387), 1,
      anon_sym_begin,
    ACTIONS(389), 1,
      sym_ocamlyacc_value,
    ACTIONS(395), 1,
      aux_sym_number_token1,
    ACTIONS(397), 1,
      anon_sym_SQUOTE,
    ACTIONS(399), 1,
      anon_sym_DQUOTE,
    ACTIONS(401), 1,
      sym_prefix_operator,
    ACTIONS(405), 1,
      sym__uppercase_identifier,
    ACTIONS(407), 1,
      aux_sym_tag_token1,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(743), 1,
      sym__lowercase_identifier,
    ACTIONS(1221), 1,
      anon_sym_RPAREN,
    STATE(771), 1,
      sym__simple_expression,
    STATE(791), 1,
      sym__value_name,
    STATE(796), 1,
      sym__constructor_name,
    STATE(830), 1,
      sym_parenthesized_operator,
    STATE(1459), 1,
      sym_attribute,
    STATE(7343), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(357), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(795), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(785), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(794), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [9610] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(353), 1,
      anon_sym_LPAREN,
    ACTIONS(355), 1,
      anon_sym_LBRACK,
    ACTIONS(359), 1,
      anon_sym_LBRACE,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(365), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(383), 1,
      anon_sym_new,
    ACTIONS(385), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(387), 1,
      anon_sym_begin,
    ACTIONS(389), 1,
      sym_ocamlyacc_value,
    ACTIONS(395), 1,
      aux_sym_number_token1,
    ACTIONS(397), 1,
      anon_sym_SQUOTE,
    ACTIONS(399), 1,
      anon_sym_DQUOTE,
    ACTIONS(401), 1,
      sym_prefix_operator,
    ACTIONS(405), 1,
      sym__uppercase_identifier,
    ACTIONS(407), 1,
      aux_sym_tag_token1,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(743), 1,
      sym__lowercase_identifier,
    ACTIONS(1183), 1,
      anon_sym_RPAREN,
    STATE(771), 1,
      sym__simple_expression,
    STATE(791), 1,
      sym__value_name,
    STATE(796), 1,
      sym__constructor_name,
    STATE(830), 1,
      sym_parenthesized_operator,
    STATE(1460), 1,
      sym_attribute,
    STATE(7343), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(357), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(795), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(785), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(794), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [9733] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(353), 1,
      anon_sym_LPAREN,
    ACTIONS(355), 1,
      anon_sym_LBRACK,
    ACTIONS(359), 1,
      anon_sym_LBRACE,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(365), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(383), 1,
      anon_sym_new,
    ACTIONS(385), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(387), 1,
      anon_sym_begin,
    ACTIONS(389), 1,
      sym_ocamlyacc_value,
    ACTIONS(395), 1,
      aux_sym_number_token1,
    ACTIONS(397), 1,
      anon_sym_SQUOTE,
    ACTIONS(399), 1,
      anon_sym_DQUOTE,
    ACTIONS(401), 1,
      sym_prefix_operator,
    ACTIONS(405), 1,
      sym__uppercase_identifier,
    ACTIONS(407), 1,
      aux_sym_tag_token1,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(743), 1,
      sym__lowercase_identifier,
    ACTIONS(1211), 1,
      anon_sym_RPAREN,
    STATE(771), 1,
      sym__simple_expression,
    STATE(791), 1,
      sym__value_name,
    STATE(796), 1,
      sym__constructor_name,
    STATE(830), 1,
      sym_parenthesized_operator,
    STATE(1461), 1,
      sym_attribute,
    STATE(7343), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(357), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(795), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(785), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(794), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [9856] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(353), 1,
      anon_sym_LPAREN,
    ACTIONS(355), 1,
      anon_sym_LBRACK,
    ACTIONS(359), 1,
      anon_sym_LBRACE,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(365), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(383), 1,
      anon_sym_new,
    ACTIONS(385), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(387), 1,
      anon_sym_begin,
    ACTIONS(389), 1,
      sym_ocamlyacc_value,
    ACTIONS(395), 1,
      aux_sym_number_token1,
    ACTIONS(397), 1,
      anon_sym_SQUOTE,
    ACTIONS(399), 1,
      anon_sym_DQUOTE,
    ACTIONS(401), 1,
      sym_prefix_operator,
    ACTIONS(405), 1,
      sym__uppercase_identifier,
    ACTIONS(407), 1,
      aux_sym_tag_token1,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(743), 1,
      sym__lowercase_identifier,
    ACTIONS(1205), 1,
      anon_sym_RPAREN,
    STATE(771), 1,
      sym__simple_expression,
    STATE(791), 1,
      sym__value_name,
    STATE(796), 1,
      sym__constructor_name,
    STATE(830), 1,
      sym_parenthesized_operator,
    STATE(1462), 1,
      sym_attribute,
    STATE(7343), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(357), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(795), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(785), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(794), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [9979] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(613), 1,
      anon_sym_LPAREN,
    ACTIONS(615), 1,
      anon_sym_LBRACK,
    ACTIONS(619), 1,
      anon_sym_LBRACE,
    ACTIONS(621), 1,
      anon_sym_object,
    ACTIONS(625), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(643), 1,
      anon_sym_new,
    ACTIONS(645), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(647), 1,
      anon_sym_begin,
    ACTIONS(649), 1,
      sym_ocamlyacc_value,
    ACTIONS(651), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(653), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(655), 1,
      aux_sym_number_token1,
    ACTIONS(657), 1,
      anon_sym_SQUOTE,
    ACTIONS(659), 1,
      anon_sym_DQUOTE,
    ACTIONS(661), 1,
      sym_prefix_operator,
    ACTIONS(665), 1,
      sym__uppercase_identifier,
    ACTIONS(667), 1,
      aux_sym_tag_token1,
    ACTIONS(1337), 1,
      sym__lowercase_identifier,
    STATE(1463), 1,
      sym_attribute,
    STATE(1539), 1,
      sym__constructor_name,
    STATE(1550), 1,
      sym__simple_expression,
    STATE(1580), 1,
      sym__value_name,
    STATE(1593), 1,
      sym_parenthesized_operator,
    STATE(6641), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(617), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1596), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1545), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1543), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10099] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2676), 1,
      sym__lowercase_identifier,
    ACTIONS(2678), 1,
      anon_sym_LPAREN,
    ACTIONS(2682), 1,
      anon_sym_LBRACE,
    ACTIONS(2684), 1,
      anon_sym_begin,
    ACTIONS(2686), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2688), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2690), 1,
      aux_sym_number_token1,
    ACTIONS(2692), 1,
      anon_sym_SQUOTE,
    ACTIONS(2694), 1,
      anon_sym_DQUOTE,
    ACTIONS(2696), 1,
      sym_prefix_operator,
    ACTIONS(2698), 1,
      aux_sym_tag_token1,
    STATE(1464), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(1845), 1,
      sym__simple_expression,
    STATE(6354), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2680), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1868), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10219] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1855), 1,
      sym__lowercase_identifier,
    ACTIONS(1859), 1,
      anon_sym_LPAREN,
    ACTIONS(1861), 1,
      anon_sym_LBRACK,
    ACTIONS(1865), 1,
      anon_sym_LBRACE,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(1869), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1871), 1,
      anon_sym_new,
    ACTIONS(1873), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(1875), 1,
      anon_sym_begin,
    ACTIONS(1877), 1,
      sym_ocamlyacc_value,
    ACTIONS(1883), 1,
      aux_sym_number_token1,
    ACTIONS(1885), 1,
      anon_sym_SQUOTE,
    ACTIONS(1887), 1,
      anon_sym_DQUOTE,
    ACTIONS(1889), 1,
      sym_prefix_operator,
    ACTIONS(1891), 1,
      sym__uppercase_identifier,
    ACTIONS(1893), 1,
      aux_sym_tag_token1,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1465), 1,
      sym_attribute,
    STATE(2317), 1,
      sym__simple_expression,
    STATE(2345), 1,
      sym__value_name,
    STATE(2369), 1,
      sym__constructor_name,
    STATE(2400), 1,
      sym_parenthesized_operator,
    STATE(6791), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1863), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2361), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2330), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2321), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10339] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(39), 1,
      anon_sym_LPAREN,
    ACTIONS(45), 1,
      anon_sym_LBRACK,
    ACTIONS(49), 1,
      anon_sym_LBRACE,
    ACTIONS(51), 1,
      anon_sym_object,
    ACTIONS(55), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(73), 1,
      anon_sym_new,
    ACTIONS(75), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(77), 1,
      anon_sym_begin,
    ACTIONS(79), 1,
      sym_ocamlyacc_value,
    ACTIONS(87), 1,
      aux_sym_number_token1,
    ACTIONS(91), 1,
      anon_sym_SQUOTE,
    ACTIONS(93), 1,
      anon_sym_DQUOTE,
    ACTIONS(95), 1,
      sym_prefix_operator,
    ACTIONS(101), 1,
      sym__uppercase_identifier,
    ACTIONS(103), 1,
      aux_sym_tag_token1,
    ACTIONS(681), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(683), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(821), 1,
      sym__lowercase_identifier,
    STATE(868), 1,
      sym__value_name,
    STATE(883), 1,
      sym_parenthesized_operator,
    STATE(886), 1,
      sym__constructor_name,
    STATE(915), 1,
      sym__simple_expression,
    STATE(1466), 1,
      sym_attribute,
    STATE(6680), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(47), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(894), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(911), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(910), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10459] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2546), 1,
      sym__lowercase_identifier,
    ACTIONS(2548), 1,
      anon_sym_LPAREN,
    ACTIONS(2550), 1,
      anon_sym_LBRACK,
    ACTIONS(2554), 1,
      anon_sym_LBRACE,
    ACTIONS(2556), 1,
      anon_sym_object,
    ACTIONS(2558), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2560), 1,
      anon_sym_new,
    ACTIONS(2562), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2564), 1,
      anon_sym_begin,
    ACTIONS(2566), 1,
      sym_ocamlyacc_value,
    ACTIONS(2568), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2570), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2574), 1,
      aux_sym_number_token1,
    ACTIONS(2576), 1,
      anon_sym_SQUOTE,
    ACTIONS(2578), 1,
      anon_sym_DQUOTE,
    ACTIONS(2580), 1,
      sym_prefix_operator,
    ACTIONS(2582), 1,
      sym__uppercase_identifier,
    ACTIONS(2584), 1,
      aux_sym_tag_token1,
    STATE(1467), 1,
      sym_attribute,
    STATE(1924), 1,
      sym_parenthesized_operator,
    STATE(1977), 1,
      sym__constructor_name,
    STATE(2011), 1,
      sym__simple_expression,
    STATE(2083), 1,
      sym__value_name,
    STATE(6755), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2552), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1951), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1953), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1937), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10579] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(39), 1,
      anon_sym_LPAREN,
    ACTIONS(45), 1,
      anon_sym_LBRACK,
    ACTIONS(49), 1,
      anon_sym_LBRACE,
    ACTIONS(51), 1,
      anon_sym_object,
    ACTIONS(55), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(73), 1,
      anon_sym_new,
    ACTIONS(75), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(77), 1,
      anon_sym_begin,
    ACTIONS(79), 1,
      sym_ocamlyacc_value,
    ACTIONS(87), 1,
      aux_sym_number_token1,
    ACTIONS(91), 1,
      anon_sym_SQUOTE,
    ACTIONS(93), 1,
      anon_sym_DQUOTE,
    ACTIONS(95), 1,
      sym_prefix_operator,
    ACTIONS(101), 1,
      sym__uppercase_identifier,
    ACTIONS(103), 1,
      aux_sym_tag_token1,
    ACTIONS(681), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(683), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(821), 1,
      sym__lowercase_identifier,
    STATE(868), 1,
      sym__value_name,
    STATE(883), 1,
      sym_parenthesized_operator,
    STATE(886), 1,
      sym__constructor_name,
    STATE(923), 1,
      sym__simple_expression,
    STATE(1468), 1,
      sym_attribute,
    STATE(6680), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(47), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(894), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(911), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(910), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10699] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2546), 1,
      sym__lowercase_identifier,
    ACTIONS(2548), 1,
      anon_sym_LPAREN,
    ACTIONS(2550), 1,
      anon_sym_LBRACK,
    ACTIONS(2554), 1,
      anon_sym_LBRACE,
    ACTIONS(2556), 1,
      anon_sym_object,
    ACTIONS(2558), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2560), 1,
      anon_sym_new,
    ACTIONS(2562), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2564), 1,
      anon_sym_begin,
    ACTIONS(2566), 1,
      sym_ocamlyacc_value,
    ACTIONS(2568), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2570), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2574), 1,
      aux_sym_number_token1,
    ACTIONS(2576), 1,
      anon_sym_SQUOTE,
    ACTIONS(2578), 1,
      anon_sym_DQUOTE,
    ACTIONS(2580), 1,
      sym_prefix_operator,
    ACTIONS(2582), 1,
      sym__uppercase_identifier,
    ACTIONS(2584), 1,
      aux_sym_tag_token1,
    STATE(1469), 1,
      sym_attribute,
    STATE(1924), 1,
      sym_parenthesized_operator,
    STATE(1977), 1,
      sym__constructor_name,
    STATE(2044), 1,
      sym__simple_expression,
    STATE(2083), 1,
      sym__value_name,
    STATE(6755), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2552), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1951), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1953), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1937), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10819] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(109), 1,
      anon_sym_LPAREN,
    ACTIONS(111), 1,
      anon_sym_LBRACK,
    ACTIONS(115), 1,
      anon_sym_LBRACE,
    ACTIONS(117), 1,
      anon_sym_object,
    ACTIONS(121), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(139), 1,
      anon_sym_new,
    ACTIONS(141), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(143), 1,
      anon_sym_begin,
    ACTIONS(145), 1,
      sym_ocamlyacc_value,
    ACTIONS(151), 1,
      aux_sym_number_token1,
    ACTIONS(153), 1,
      anon_sym_SQUOTE,
    ACTIONS(155), 1,
      anon_sym_DQUOTE,
    ACTIONS(157), 1,
      sym_prefix_operator,
    ACTIONS(161), 1,
      sym__uppercase_identifier,
    ACTIONS(163), 1,
      aux_sym_tag_token1,
    ACTIONS(831), 1,
      sym__lowercase_identifier,
    ACTIONS(1179), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1181), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(963), 1,
      sym__value_name,
    STATE(971), 1,
      sym__constructor_name,
    STATE(978), 1,
      sym_parenthesized_operator,
    STATE(986), 1,
      sym__simple_expression,
    STATE(1470), 1,
      sym_attribute,
    STATE(6598), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(113), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(997), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(988), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(987), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [10939] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(353), 1,
      anon_sym_LPAREN,
    ACTIONS(355), 1,
      anon_sym_LBRACK,
    ACTIONS(359), 1,
      anon_sym_LBRACE,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(365), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(383), 1,
      anon_sym_new,
    ACTIONS(385), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(387), 1,
      anon_sym_begin,
    ACTIONS(389), 1,
      sym_ocamlyacc_value,
    ACTIONS(395), 1,
      aux_sym_number_token1,
    ACTIONS(397), 1,
      anon_sym_SQUOTE,
    ACTIONS(399), 1,
      anon_sym_DQUOTE,
    ACTIONS(401), 1,
      sym_prefix_operator,
    ACTIONS(405), 1,
      sym__uppercase_identifier,
    ACTIONS(407), 1,
      aux_sym_tag_token1,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(743), 1,
      sym__lowercase_identifier,
    STATE(771), 1,
      sym__simple_expression,
    STATE(791), 1,
      sym__value_name,
    STATE(796), 1,
      sym__constructor_name,
    STATE(830), 1,
      sym_parenthesized_operator,
    STATE(1471), 1,
      sym_attribute,
    STATE(7343), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(357), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(795), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(785), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(794), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11059] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2676), 1,
      sym__lowercase_identifier,
    ACTIONS(2678), 1,
      anon_sym_LPAREN,
    ACTIONS(2682), 1,
      anon_sym_LBRACE,
    ACTIONS(2684), 1,
      anon_sym_begin,
    ACTIONS(2686), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2688), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2690), 1,
      aux_sym_number_token1,
    ACTIONS(2692), 1,
      anon_sym_SQUOTE,
    ACTIONS(2694), 1,
      anon_sym_DQUOTE,
    ACTIONS(2696), 1,
      sym_prefix_operator,
    ACTIONS(2698), 1,
      aux_sym_tag_token1,
    STATE(1472), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(1892), 1,
      sym__simple_expression,
    STATE(6354), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2680), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1868), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11179] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2676), 1,
      sym__lowercase_identifier,
    ACTIONS(2678), 1,
      anon_sym_LPAREN,
    ACTIONS(2682), 1,
      anon_sym_LBRACE,
    ACTIONS(2684), 1,
      anon_sym_begin,
    ACTIONS(2686), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2688), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2690), 1,
      aux_sym_number_token1,
    ACTIONS(2692), 1,
      anon_sym_SQUOTE,
    ACTIONS(2694), 1,
      anon_sym_DQUOTE,
    ACTIONS(2696), 1,
      sym_prefix_operator,
    ACTIONS(2698), 1,
      aux_sym_tag_token1,
    STATE(1473), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(1893), 1,
      sym__simple_expression,
    STATE(6354), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2680), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1868), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11299] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1474), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2710), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2708), 41,
      anon_sym_SEMI_SEMI,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [11369] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(183), 1,
      anon_sym_LPAREN,
    ACTIONS(185), 1,
      anon_sym_LBRACK,
    ACTIONS(189), 1,
      anon_sym_LBRACE,
    ACTIONS(191), 1,
      anon_sym_object,
    ACTIONS(195), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(205), 1,
      anon_sym_new,
    ACTIONS(207), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(209), 1,
      anon_sym_begin,
    ACTIONS(211), 1,
      sym_ocamlyacc_value,
    ACTIONS(217), 1,
      aux_sym_number_token1,
    ACTIONS(219), 1,
      anon_sym_SQUOTE,
    ACTIONS(221), 1,
      anon_sym_DQUOTE,
    ACTIONS(223), 1,
      sym_prefix_operator,
    ACTIONS(227), 1,
      sym__uppercase_identifier,
    ACTIONS(229), 1,
      aux_sym_tag_token1,
    ACTIONS(965), 1,
      sym__lowercase_identifier,
    ACTIONS(1173), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1175), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1076), 1,
      sym__value_name,
    STATE(1077), 1,
      sym_parenthesized_operator,
    STATE(1079), 1,
      sym__constructor_name,
    STATE(1112), 1,
      sym__simple_expression,
    STATE(1475), 1,
      sym_attribute,
    STATE(7309), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(187), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1057), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1118), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1052), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11489] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2546), 1,
      sym__lowercase_identifier,
    ACTIONS(2548), 1,
      anon_sym_LPAREN,
    ACTIONS(2550), 1,
      anon_sym_LBRACK,
    ACTIONS(2554), 1,
      anon_sym_LBRACE,
    ACTIONS(2556), 1,
      anon_sym_object,
    ACTIONS(2558), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2560), 1,
      anon_sym_new,
    ACTIONS(2562), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2564), 1,
      anon_sym_begin,
    ACTIONS(2566), 1,
      sym_ocamlyacc_value,
    ACTIONS(2568), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2570), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2574), 1,
      aux_sym_number_token1,
    ACTIONS(2576), 1,
      anon_sym_SQUOTE,
    ACTIONS(2578), 1,
      anon_sym_DQUOTE,
    ACTIONS(2580), 1,
      sym_prefix_operator,
    ACTIONS(2582), 1,
      sym__uppercase_identifier,
    ACTIONS(2584), 1,
      aux_sym_tag_token1,
    STATE(1476), 1,
      sym_attribute,
    STATE(1924), 1,
      sym_parenthesized_operator,
    STATE(1977), 1,
      sym__constructor_name,
    STATE(2052), 1,
      sym__simple_expression,
    STATE(2083), 1,
      sym__value_name,
    STATE(6755), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2552), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1951), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1953), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1937), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11609] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(39), 1,
      anon_sym_LPAREN,
    ACTIONS(45), 1,
      anon_sym_LBRACK,
    ACTIONS(49), 1,
      anon_sym_LBRACE,
    ACTIONS(51), 1,
      anon_sym_object,
    ACTIONS(55), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(73), 1,
      anon_sym_new,
    ACTIONS(75), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(77), 1,
      anon_sym_begin,
    ACTIONS(79), 1,
      sym_ocamlyacc_value,
    ACTIONS(87), 1,
      aux_sym_number_token1,
    ACTIONS(91), 1,
      anon_sym_SQUOTE,
    ACTIONS(93), 1,
      anon_sym_DQUOTE,
    ACTIONS(95), 1,
      sym_prefix_operator,
    ACTIONS(101), 1,
      sym__uppercase_identifier,
    ACTIONS(103), 1,
      aux_sym_tag_token1,
    ACTIONS(681), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(683), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(821), 1,
      sym__lowercase_identifier,
    STATE(868), 1,
      sym__value_name,
    STATE(883), 1,
      sym_parenthesized_operator,
    STATE(886), 1,
      sym__constructor_name,
    STATE(909), 1,
      sym__simple_expression,
    STATE(1477), 1,
      sym_attribute,
    STATE(6680), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(47), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(894), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(911), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(910), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11729] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2632), 1,
      sym__lowercase_identifier,
    ACTIONS(2634), 1,
      anon_sym_LPAREN,
    ACTIONS(2636), 1,
      anon_sym_LBRACK,
    ACTIONS(2640), 1,
      anon_sym_LBRACE,
    ACTIONS(2642), 1,
      anon_sym_object,
    ACTIONS(2644), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2646), 1,
      anon_sym_new,
    ACTIONS(2648), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2650), 1,
      anon_sym_begin,
    ACTIONS(2652), 1,
      sym_ocamlyacc_value,
    ACTIONS(2654), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2656), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2658), 1,
      aux_sym_number_token1,
    ACTIONS(2660), 1,
      anon_sym_SQUOTE,
    ACTIONS(2662), 1,
      anon_sym_DQUOTE,
    ACTIONS(2664), 1,
      sym_prefix_operator,
    ACTIONS(2666), 1,
      sym__uppercase_identifier,
    ACTIONS(2668), 1,
      aux_sym_tag_token1,
    STATE(1478), 1,
      sym_attribute,
    STATE(2675), 1,
      sym_parenthesized_operator,
    STATE(2695), 1,
      sym__constructor_name,
    STATE(2727), 1,
      sym__simple_expression,
    STATE(2761), 1,
      sym__value_name,
    STATE(6718), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2638), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2683), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2719), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11849] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2594), 1,
      sym__lowercase_identifier,
    ACTIONS(2596), 1,
      anon_sym_LPAREN,
    ACTIONS(2598), 1,
      anon_sym_LBRACK,
    ACTIONS(2602), 1,
      anon_sym_LBRACE,
    ACTIONS(2604), 1,
      anon_sym_object,
    ACTIONS(2606), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2608), 1,
      anon_sym_new,
    ACTIONS(2610), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2612), 1,
      anon_sym_begin,
    ACTIONS(2614), 1,
      sym_ocamlyacc_value,
    ACTIONS(2616), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2618), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2620), 1,
      aux_sym_number_token1,
    ACTIONS(2622), 1,
      anon_sym_SQUOTE,
    ACTIONS(2624), 1,
      anon_sym_DQUOTE,
    ACTIONS(2626), 1,
      sym_prefix_operator,
    ACTIONS(2628), 1,
      sym__uppercase_identifier,
    ACTIONS(2630), 1,
      aux_sym_tag_token1,
    STATE(1479), 1,
      sym_attribute,
    STATE(1822), 1,
      sym_parenthesized_operator,
    STATE(1833), 1,
      sym__constructor_name,
    STATE(1838), 1,
      sym__value_name,
    STATE(1903), 1,
      sym__simple_expression,
    STATE(6490), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2600), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1837), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1897), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1896), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [11969] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1480), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2714), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2712), 41,
      anon_sym_SEMI_SEMI,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [12039] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1855), 1,
      sym__lowercase_identifier,
    ACTIONS(1859), 1,
      anon_sym_LPAREN,
    ACTIONS(1861), 1,
      anon_sym_LBRACK,
    ACTIONS(1865), 1,
      anon_sym_LBRACE,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(1869), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1871), 1,
      anon_sym_new,
    ACTIONS(1873), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(1875), 1,
      anon_sym_begin,
    ACTIONS(1877), 1,
      sym_ocamlyacc_value,
    ACTIONS(1883), 1,
      aux_sym_number_token1,
    ACTIONS(1885), 1,
      anon_sym_SQUOTE,
    ACTIONS(1887), 1,
      anon_sym_DQUOTE,
    ACTIONS(1889), 1,
      sym_prefix_operator,
    ACTIONS(1891), 1,
      sym__uppercase_identifier,
    ACTIONS(1893), 1,
      aux_sym_tag_token1,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1481), 1,
      sym_attribute,
    STATE(2345), 1,
      sym__value_name,
    STATE(2357), 1,
      sym__simple_expression,
    STATE(2369), 1,
      sym__constructor_name,
    STATE(2400), 1,
      sym_parenthesized_operator,
    STATE(6791), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1863), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2361), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2330), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2321), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12159] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1855), 1,
      sym__lowercase_identifier,
    ACTIONS(1859), 1,
      anon_sym_LPAREN,
    ACTIONS(1861), 1,
      anon_sym_LBRACK,
    ACTIONS(1865), 1,
      anon_sym_LBRACE,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(1869), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1871), 1,
      anon_sym_new,
    ACTIONS(1873), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(1875), 1,
      anon_sym_begin,
    ACTIONS(1877), 1,
      sym_ocamlyacc_value,
    ACTIONS(1883), 1,
      aux_sym_number_token1,
    ACTIONS(1885), 1,
      anon_sym_SQUOTE,
    ACTIONS(1887), 1,
      anon_sym_DQUOTE,
    ACTIONS(1889), 1,
      sym_prefix_operator,
    ACTIONS(1891), 1,
      sym__uppercase_identifier,
    ACTIONS(1893), 1,
      aux_sym_tag_token1,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1482), 1,
      sym_attribute,
    STATE(2345), 1,
      sym__value_name,
    STATE(2369), 1,
      sym__constructor_name,
    STATE(2396), 1,
      sym__simple_expression,
    STATE(2400), 1,
      sym_parenthesized_operator,
    STATE(6791), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1863), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2361), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2330), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2321), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12279] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(183), 1,
      anon_sym_LPAREN,
    ACTIONS(185), 1,
      anon_sym_LBRACK,
    ACTIONS(189), 1,
      anon_sym_LBRACE,
    ACTIONS(191), 1,
      anon_sym_object,
    ACTIONS(195), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(205), 1,
      anon_sym_new,
    ACTIONS(207), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(209), 1,
      anon_sym_begin,
    ACTIONS(211), 1,
      sym_ocamlyacc_value,
    ACTIONS(217), 1,
      aux_sym_number_token1,
    ACTIONS(219), 1,
      anon_sym_SQUOTE,
    ACTIONS(221), 1,
      anon_sym_DQUOTE,
    ACTIONS(223), 1,
      sym_prefix_operator,
    ACTIONS(227), 1,
      sym__uppercase_identifier,
    ACTIONS(229), 1,
      aux_sym_tag_token1,
    ACTIONS(965), 1,
      sym__lowercase_identifier,
    ACTIONS(1173), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1175), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1076), 1,
      sym__value_name,
    STATE(1077), 1,
      sym_parenthesized_operator,
    STATE(1079), 1,
      sym__constructor_name,
    STATE(1084), 1,
      sym__simple_expression,
    STATE(1483), 1,
      sym_attribute,
    STATE(7309), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(187), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1057), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1118), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1052), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12399] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2594), 1,
      sym__lowercase_identifier,
    ACTIONS(2596), 1,
      anon_sym_LPAREN,
    ACTIONS(2598), 1,
      anon_sym_LBRACK,
    ACTIONS(2602), 1,
      anon_sym_LBRACE,
    ACTIONS(2604), 1,
      anon_sym_object,
    ACTIONS(2606), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2608), 1,
      anon_sym_new,
    ACTIONS(2610), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2612), 1,
      anon_sym_begin,
    ACTIONS(2614), 1,
      sym_ocamlyacc_value,
    ACTIONS(2616), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2618), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2620), 1,
      aux_sym_number_token1,
    ACTIONS(2622), 1,
      anon_sym_SQUOTE,
    ACTIONS(2624), 1,
      anon_sym_DQUOTE,
    ACTIONS(2626), 1,
      sym_prefix_operator,
    ACTIONS(2628), 1,
      sym__uppercase_identifier,
    ACTIONS(2630), 1,
      aux_sym_tag_token1,
    STATE(1484), 1,
      sym_attribute,
    STATE(1822), 1,
      sym_parenthesized_operator,
    STATE(1833), 1,
      sym__constructor_name,
    STATE(1838), 1,
      sym__value_name,
    STATE(1914), 1,
      sym__simple_expression,
    STATE(6490), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2600), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1837), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1897), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1896), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12519] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2632), 1,
      sym__lowercase_identifier,
    ACTIONS(2634), 1,
      anon_sym_LPAREN,
    ACTIONS(2636), 1,
      anon_sym_LBRACK,
    ACTIONS(2640), 1,
      anon_sym_LBRACE,
    ACTIONS(2642), 1,
      anon_sym_object,
    ACTIONS(2644), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2646), 1,
      anon_sym_new,
    ACTIONS(2648), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2650), 1,
      anon_sym_begin,
    ACTIONS(2652), 1,
      sym_ocamlyacc_value,
    ACTIONS(2654), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2656), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2658), 1,
      aux_sym_number_token1,
    ACTIONS(2660), 1,
      anon_sym_SQUOTE,
    ACTIONS(2662), 1,
      anon_sym_DQUOTE,
    ACTIONS(2664), 1,
      sym_prefix_operator,
    ACTIONS(2666), 1,
      sym__uppercase_identifier,
    ACTIONS(2668), 1,
      aux_sym_tag_token1,
    STATE(1485), 1,
      sym_attribute,
    STATE(2675), 1,
      sym_parenthesized_operator,
    STATE(2695), 1,
      sym__constructor_name,
    STATE(2744), 1,
      sym__simple_expression,
    STATE(2761), 1,
      sym__value_name,
    STATE(6718), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2638), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2683), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2719), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12639] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1486), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2718), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2716), 41,
      anon_sym_SEMI_SEMI,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [12709] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2676), 1,
      sym__lowercase_identifier,
    ACTIONS(2678), 1,
      anon_sym_LPAREN,
    ACTIONS(2682), 1,
      anon_sym_LBRACE,
    ACTIONS(2684), 1,
      anon_sym_begin,
    ACTIONS(2686), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2688), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2690), 1,
      aux_sym_number_token1,
    ACTIONS(2692), 1,
      anon_sym_SQUOTE,
    ACTIONS(2694), 1,
      anon_sym_DQUOTE,
    ACTIONS(2696), 1,
      sym_prefix_operator,
    ACTIONS(2698), 1,
      aux_sym_tag_token1,
    STATE(1487), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(1746), 1,
      sym__simple_expression,
    STATE(6354), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2680), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1868), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12829] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(613), 1,
      anon_sym_LPAREN,
    ACTIONS(615), 1,
      anon_sym_LBRACK,
    ACTIONS(619), 1,
      anon_sym_LBRACE,
    ACTIONS(621), 1,
      anon_sym_object,
    ACTIONS(625), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(643), 1,
      anon_sym_new,
    ACTIONS(645), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(647), 1,
      anon_sym_begin,
    ACTIONS(649), 1,
      sym_ocamlyacc_value,
    ACTIONS(651), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(653), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(655), 1,
      aux_sym_number_token1,
    ACTIONS(657), 1,
      anon_sym_SQUOTE,
    ACTIONS(659), 1,
      anon_sym_DQUOTE,
    ACTIONS(661), 1,
      sym_prefix_operator,
    ACTIONS(665), 1,
      sym__uppercase_identifier,
    ACTIONS(667), 1,
      aux_sym_tag_token1,
    ACTIONS(1337), 1,
      sym__lowercase_identifier,
    STATE(1488), 1,
      sym_attribute,
    STATE(1539), 1,
      sym__constructor_name,
    STATE(1541), 1,
      sym__simple_expression,
    STATE(1580), 1,
      sym__value_name,
    STATE(1593), 1,
      sym_parenthesized_operator,
    STATE(6641), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(617), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1596), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1545), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1543), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [12949] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2594), 1,
      sym__lowercase_identifier,
    ACTIONS(2596), 1,
      anon_sym_LPAREN,
    ACTIONS(2598), 1,
      anon_sym_LBRACK,
    ACTIONS(2602), 1,
      anon_sym_LBRACE,
    ACTIONS(2604), 1,
      anon_sym_object,
    ACTIONS(2606), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2608), 1,
      anon_sym_new,
    ACTIONS(2610), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2612), 1,
      anon_sym_begin,
    ACTIONS(2614), 1,
      sym_ocamlyacc_value,
    ACTIONS(2616), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2618), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2620), 1,
      aux_sym_number_token1,
    ACTIONS(2622), 1,
      anon_sym_SQUOTE,
    ACTIONS(2624), 1,
      anon_sym_DQUOTE,
    ACTIONS(2626), 1,
      sym_prefix_operator,
    ACTIONS(2628), 1,
      sym__uppercase_identifier,
    ACTIONS(2630), 1,
      aux_sym_tag_token1,
    STATE(1489), 1,
      sym_attribute,
    STATE(1822), 1,
      sym_parenthesized_operator,
    STATE(1833), 1,
      sym__constructor_name,
    STATE(1838), 1,
      sym__value_name,
    STATE(1840), 1,
      sym__simple_expression,
    STATE(6490), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2600), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1837), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1897), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1896), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13069] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2594), 1,
      sym__lowercase_identifier,
    ACTIONS(2596), 1,
      anon_sym_LPAREN,
    ACTIONS(2598), 1,
      anon_sym_LBRACK,
    ACTIONS(2602), 1,
      anon_sym_LBRACE,
    ACTIONS(2604), 1,
      anon_sym_object,
    ACTIONS(2606), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2608), 1,
      anon_sym_new,
    ACTIONS(2610), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2612), 1,
      anon_sym_begin,
    ACTIONS(2614), 1,
      sym_ocamlyacc_value,
    ACTIONS(2616), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2618), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2620), 1,
      aux_sym_number_token1,
    ACTIONS(2622), 1,
      anon_sym_SQUOTE,
    ACTIONS(2624), 1,
      anon_sym_DQUOTE,
    ACTIONS(2626), 1,
      sym_prefix_operator,
    ACTIONS(2628), 1,
      sym__uppercase_identifier,
    ACTIONS(2630), 1,
      aux_sym_tag_token1,
    STATE(1490), 1,
      sym_attribute,
    STATE(1822), 1,
      sym_parenthesized_operator,
    STATE(1833), 1,
      sym__constructor_name,
    STATE(1838), 1,
      sym__value_name,
    STATE(1841), 1,
      sym__simple_expression,
    STATE(6490), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2600), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1837), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1897), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1896), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13189] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(109), 1,
      anon_sym_LPAREN,
    ACTIONS(111), 1,
      anon_sym_LBRACK,
    ACTIONS(115), 1,
      anon_sym_LBRACE,
    ACTIONS(117), 1,
      anon_sym_object,
    ACTIONS(121), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(139), 1,
      anon_sym_new,
    ACTIONS(141), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(143), 1,
      anon_sym_begin,
    ACTIONS(145), 1,
      sym_ocamlyacc_value,
    ACTIONS(151), 1,
      aux_sym_number_token1,
    ACTIONS(153), 1,
      anon_sym_SQUOTE,
    ACTIONS(155), 1,
      anon_sym_DQUOTE,
    ACTIONS(157), 1,
      sym_prefix_operator,
    ACTIONS(161), 1,
      sym__uppercase_identifier,
    ACTIONS(163), 1,
      aux_sym_tag_token1,
    ACTIONS(831), 1,
      sym__lowercase_identifier,
    ACTIONS(1179), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1181), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(963), 1,
      sym__value_name,
    STATE(971), 1,
      sym__constructor_name,
    STATE(978), 1,
      sym_parenthesized_operator,
    STATE(998), 1,
      sym__simple_expression,
    STATE(1491), 1,
      sym_attribute,
    STATE(6598), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(113), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(997), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(988), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(987), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13309] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2676), 1,
      sym__lowercase_identifier,
    ACTIONS(2678), 1,
      anon_sym_LPAREN,
    ACTIONS(2682), 1,
      anon_sym_LBRACE,
    ACTIONS(2684), 1,
      anon_sym_begin,
    ACTIONS(2686), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2688), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2690), 1,
      aux_sym_number_token1,
    ACTIONS(2692), 1,
      anon_sym_SQUOTE,
    ACTIONS(2694), 1,
      anon_sym_DQUOTE,
    ACTIONS(2696), 1,
      sym_prefix_operator,
    ACTIONS(2698), 1,
      aux_sym_tag_token1,
    STATE(1492), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(1884), 1,
      sym__simple_expression,
    STATE(6354), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2680), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1868), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13429] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(109), 1,
      anon_sym_LPAREN,
    ACTIONS(111), 1,
      anon_sym_LBRACK,
    ACTIONS(115), 1,
      anon_sym_LBRACE,
    ACTIONS(117), 1,
      anon_sym_object,
    ACTIONS(121), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(139), 1,
      anon_sym_new,
    ACTIONS(141), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(143), 1,
      anon_sym_begin,
    ACTIONS(145), 1,
      sym_ocamlyacc_value,
    ACTIONS(151), 1,
      aux_sym_number_token1,
    ACTIONS(153), 1,
      anon_sym_SQUOTE,
    ACTIONS(155), 1,
      anon_sym_DQUOTE,
    ACTIONS(157), 1,
      sym_prefix_operator,
    ACTIONS(161), 1,
      sym__uppercase_identifier,
    ACTIONS(163), 1,
      aux_sym_tag_token1,
    ACTIONS(831), 1,
      sym__lowercase_identifier,
    ACTIONS(1179), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1181), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(963), 1,
      sym__value_name,
    STATE(971), 1,
      sym__constructor_name,
    STATE(978), 1,
      sym_parenthesized_operator,
    STATE(1020), 1,
      sym__simple_expression,
    STATE(1493), 1,
      sym_attribute,
    STATE(6598), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(113), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(997), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(988), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(987), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13549] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(2676), 1,
      sym__lowercase_identifier,
    STATE(1494), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(5215), 1,
      sym__simple_expression,
    STATE(6553), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13669] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(353), 1,
      anon_sym_LPAREN,
    ACTIONS(355), 1,
      anon_sym_LBRACK,
    ACTIONS(359), 1,
      anon_sym_LBRACE,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(365), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(383), 1,
      anon_sym_new,
    ACTIONS(385), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(387), 1,
      anon_sym_begin,
    ACTIONS(389), 1,
      sym_ocamlyacc_value,
    ACTIONS(395), 1,
      aux_sym_number_token1,
    ACTIONS(397), 1,
      anon_sym_SQUOTE,
    ACTIONS(399), 1,
      anon_sym_DQUOTE,
    ACTIONS(401), 1,
      sym_prefix_operator,
    ACTIONS(405), 1,
      sym__uppercase_identifier,
    ACTIONS(407), 1,
      aux_sym_tag_token1,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(743), 1,
      sym__lowercase_identifier,
    STATE(764), 1,
      sym__simple_expression,
    STATE(791), 1,
      sym__value_name,
    STATE(796), 1,
      sym__constructor_name,
    STATE(830), 1,
      sym_parenthesized_operator,
    STATE(1495), 1,
      sym_attribute,
    STATE(7343), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(357), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(795), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(785), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(794), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13789] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2546), 1,
      sym__lowercase_identifier,
    ACTIONS(2548), 1,
      anon_sym_LPAREN,
    ACTIONS(2550), 1,
      anon_sym_LBRACK,
    ACTIONS(2554), 1,
      anon_sym_LBRACE,
    ACTIONS(2556), 1,
      anon_sym_object,
    ACTIONS(2558), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2560), 1,
      anon_sym_new,
    ACTIONS(2562), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2564), 1,
      anon_sym_begin,
    ACTIONS(2566), 1,
      sym_ocamlyacc_value,
    ACTIONS(2568), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2570), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2574), 1,
      aux_sym_number_token1,
    ACTIONS(2576), 1,
      anon_sym_SQUOTE,
    ACTIONS(2578), 1,
      anon_sym_DQUOTE,
    ACTIONS(2580), 1,
      sym_prefix_operator,
    ACTIONS(2582), 1,
      sym__uppercase_identifier,
    ACTIONS(2584), 1,
      aux_sym_tag_token1,
    STATE(1496), 1,
      sym_attribute,
    STATE(1924), 1,
      sym_parenthesized_operator,
    STATE(1977), 1,
      sym__constructor_name,
    STATE(2083), 1,
      sym__value_name,
    STATE(2093), 1,
      sym__simple_expression,
    STATE(6755), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2552), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1951), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1953), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1937), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [13909] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2546), 1,
      sym__lowercase_identifier,
    ACTIONS(2548), 1,
      anon_sym_LPAREN,
    ACTIONS(2550), 1,
      anon_sym_LBRACK,
    ACTIONS(2554), 1,
      anon_sym_LBRACE,
    ACTIONS(2556), 1,
      anon_sym_object,
    ACTIONS(2558), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2560), 1,
      anon_sym_new,
    ACTIONS(2562), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2564), 1,
      anon_sym_begin,
    ACTIONS(2566), 1,
      sym_ocamlyacc_value,
    ACTIONS(2568), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2570), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2574), 1,
      aux_sym_number_token1,
    ACTIONS(2576), 1,
      anon_sym_SQUOTE,
    ACTIONS(2578), 1,
      anon_sym_DQUOTE,
    ACTIONS(2580), 1,
      sym_prefix_operator,
    ACTIONS(2582), 1,
      sym__uppercase_identifier,
    ACTIONS(2584), 1,
      aux_sym_tag_token1,
    STATE(1497), 1,
      sym_attribute,
    STATE(1924), 1,
      sym_parenthesized_operator,
    STATE(1959), 1,
      sym__simple_expression,
    STATE(1977), 1,
      sym__constructor_name,
    STATE(2083), 1,
      sym__value_name,
    STATE(6755), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2552), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1951), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1953), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1937), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14029] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2632), 1,
      sym__lowercase_identifier,
    ACTIONS(2634), 1,
      anon_sym_LPAREN,
    ACTIONS(2636), 1,
      anon_sym_LBRACK,
    ACTIONS(2640), 1,
      anon_sym_LBRACE,
    ACTIONS(2642), 1,
      anon_sym_object,
    ACTIONS(2644), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2646), 1,
      anon_sym_new,
    ACTIONS(2648), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2650), 1,
      anon_sym_begin,
    ACTIONS(2652), 1,
      sym_ocamlyacc_value,
    ACTIONS(2654), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2656), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2658), 1,
      aux_sym_number_token1,
    ACTIONS(2660), 1,
      anon_sym_SQUOTE,
    ACTIONS(2662), 1,
      anon_sym_DQUOTE,
    ACTIONS(2664), 1,
      sym_prefix_operator,
    ACTIONS(2666), 1,
      sym__uppercase_identifier,
    ACTIONS(2668), 1,
      aux_sym_tag_token1,
    STATE(1498), 1,
      sym_attribute,
    STATE(2675), 1,
      sym_parenthesized_operator,
    STATE(2695), 1,
      sym__constructor_name,
    STATE(2761), 1,
      sym__value_name,
    STATE(2762), 1,
      sym__simple_expression,
    STATE(6718), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2638), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2683), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2719), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14149] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2632), 1,
      sym__lowercase_identifier,
    ACTIONS(2634), 1,
      anon_sym_LPAREN,
    ACTIONS(2636), 1,
      anon_sym_LBRACK,
    ACTIONS(2640), 1,
      anon_sym_LBRACE,
    ACTIONS(2642), 1,
      anon_sym_object,
    ACTIONS(2644), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2646), 1,
      anon_sym_new,
    ACTIONS(2648), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2650), 1,
      anon_sym_begin,
    ACTIONS(2652), 1,
      sym_ocamlyacc_value,
    ACTIONS(2654), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2656), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2658), 1,
      aux_sym_number_token1,
    ACTIONS(2660), 1,
      anon_sym_SQUOTE,
    ACTIONS(2662), 1,
      anon_sym_DQUOTE,
    ACTIONS(2664), 1,
      sym_prefix_operator,
    ACTIONS(2666), 1,
      sym__uppercase_identifier,
    ACTIONS(2668), 1,
      aux_sym_tag_token1,
    STATE(1499), 1,
      sym_attribute,
    STATE(2675), 1,
      sym_parenthesized_operator,
    STATE(2695), 1,
      sym__constructor_name,
    STATE(2710), 1,
      sym__simple_expression,
    STATE(2761), 1,
      sym__value_name,
    STATE(6718), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2638), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2683), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2719), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14269] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2594), 1,
      sym__lowercase_identifier,
    ACTIONS(2596), 1,
      anon_sym_LPAREN,
    ACTIONS(2598), 1,
      anon_sym_LBRACK,
    ACTIONS(2602), 1,
      anon_sym_LBRACE,
    ACTIONS(2604), 1,
      anon_sym_object,
    ACTIONS(2606), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2608), 1,
      anon_sym_new,
    ACTIONS(2610), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2612), 1,
      anon_sym_begin,
    ACTIONS(2614), 1,
      sym_ocamlyacc_value,
    ACTIONS(2616), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2618), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2620), 1,
      aux_sym_number_token1,
    ACTIONS(2622), 1,
      anon_sym_SQUOTE,
    ACTIONS(2624), 1,
      anon_sym_DQUOTE,
    ACTIONS(2626), 1,
      sym_prefix_operator,
    ACTIONS(2628), 1,
      sym__uppercase_identifier,
    ACTIONS(2630), 1,
      aux_sym_tag_token1,
    STATE(1500), 1,
      sym_attribute,
    STATE(1822), 1,
      sym_parenthesized_operator,
    STATE(1833), 1,
      sym__constructor_name,
    STATE(1838), 1,
      sym__value_name,
    STATE(1872), 1,
      sym__simple_expression,
    STATE(6490), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2600), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1837), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1897), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1896), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14389] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(183), 1,
      anon_sym_LPAREN,
    ACTIONS(185), 1,
      anon_sym_LBRACK,
    ACTIONS(189), 1,
      anon_sym_LBRACE,
    ACTIONS(191), 1,
      anon_sym_object,
    ACTIONS(195), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(205), 1,
      anon_sym_new,
    ACTIONS(207), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(209), 1,
      anon_sym_begin,
    ACTIONS(211), 1,
      sym_ocamlyacc_value,
    ACTIONS(217), 1,
      aux_sym_number_token1,
    ACTIONS(219), 1,
      anon_sym_SQUOTE,
    ACTIONS(221), 1,
      anon_sym_DQUOTE,
    ACTIONS(223), 1,
      sym_prefix_operator,
    ACTIONS(227), 1,
      sym__uppercase_identifier,
    ACTIONS(229), 1,
      aux_sym_tag_token1,
    ACTIONS(965), 1,
      sym__lowercase_identifier,
    ACTIONS(1173), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1175), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1047), 1,
      sym__simple_expression,
    STATE(1076), 1,
      sym__value_name,
    STATE(1077), 1,
      sym_parenthesized_operator,
    STATE(1079), 1,
      sym__constructor_name,
    STATE(1501), 1,
      sym_attribute,
    STATE(7309), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(187), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1057), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1118), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1052), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14509] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(613), 1,
      anon_sym_LPAREN,
    ACTIONS(615), 1,
      anon_sym_LBRACK,
    ACTIONS(619), 1,
      anon_sym_LBRACE,
    ACTIONS(621), 1,
      anon_sym_object,
    ACTIONS(625), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(643), 1,
      anon_sym_new,
    ACTIONS(645), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(647), 1,
      anon_sym_begin,
    ACTIONS(649), 1,
      sym_ocamlyacc_value,
    ACTIONS(651), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(653), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(655), 1,
      aux_sym_number_token1,
    ACTIONS(657), 1,
      anon_sym_SQUOTE,
    ACTIONS(659), 1,
      anon_sym_DQUOTE,
    ACTIONS(661), 1,
      sym_prefix_operator,
    ACTIONS(665), 1,
      sym__uppercase_identifier,
    ACTIONS(667), 1,
      aux_sym_tag_token1,
    ACTIONS(1337), 1,
      sym__lowercase_identifier,
    STATE(1502), 1,
      sym_attribute,
    STATE(1539), 1,
      sym__constructor_name,
    STATE(1558), 1,
      sym__simple_expression,
    STATE(1580), 1,
      sym__value_name,
    STATE(1593), 1,
      sym_parenthesized_operator,
    STATE(6641), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(617), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1596), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1545), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1543), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14629] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2632), 1,
      sym__lowercase_identifier,
    ACTIONS(2634), 1,
      anon_sym_LPAREN,
    ACTIONS(2636), 1,
      anon_sym_LBRACK,
    ACTIONS(2640), 1,
      anon_sym_LBRACE,
    ACTIONS(2642), 1,
      anon_sym_object,
    ACTIONS(2644), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2646), 1,
      anon_sym_new,
    ACTIONS(2648), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2650), 1,
      anon_sym_begin,
    ACTIONS(2652), 1,
      sym_ocamlyacc_value,
    ACTIONS(2654), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2656), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2658), 1,
      aux_sym_number_token1,
    ACTIONS(2660), 1,
      anon_sym_SQUOTE,
    ACTIONS(2662), 1,
      anon_sym_DQUOTE,
    ACTIONS(2664), 1,
      sym_prefix_operator,
    ACTIONS(2666), 1,
      sym__uppercase_identifier,
    ACTIONS(2668), 1,
      aux_sym_tag_token1,
    STATE(1503), 1,
      sym_attribute,
    STATE(2675), 1,
      sym_parenthesized_operator,
    STATE(2695), 1,
      sym__constructor_name,
    STATE(2711), 1,
      sym__simple_expression,
    STATE(2761), 1,
      sym__value_name,
    STATE(6718), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2638), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2683), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2719), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(2713), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14749] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(2676), 1,
      sym__lowercase_identifier,
    STATE(1504), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(5354), 1,
      sym__simple_expression,
    STATE(6553), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14869] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(353), 1,
      anon_sym_LPAREN,
    ACTIONS(355), 1,
      anon_sym_LBRACK,
    ACTIONS(359), 1,
      anon_sym_LBRACE,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(365), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(383), 1,
      anon_sym_new,
    ACTIONS(385), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(387), 1,
      anon_sym_begin,
    ACTIONS(389), 1,
      sym_ocamlyacc_value,
    ACTIONS(395), 1,
      aux_sym_number_token1,
    ACTIONS(397), 1,
      anon_sym_SQUOTE,
    ACTIONS(399), 1,
      anon_sym_DQUOTE,
    ACTIONS(401), 1,
      sym_prefix_operator,
    ACTIONS(405), 1,
      sym__uppercase_identifier,
    ACTIONS(407), 1,
      aux_sym_tag_token1,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(743), 1,
      sym__lowercase_identifier,
    STATE(784), 1,
      sym__simple_expression,
    STATE(791), 1,
      sym__value_name,
    STATE(796), 1,
      sym__constructor_name,
    STATE(830), 1,
      sym_parenthesized_operator,
    STATE(1505), 1,
      sym_attribute,
    STATE(7343), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(357), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(795), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(785), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(794), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [14989] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2508), 1,
      anon_sym_LPAREN,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2512), 1,
      anon_sym_LBRACE,
    ACTIONS(2514), 1,
      anon_sym_object,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2518), 1,
      anon_sym_new,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2524), 1,
      sym_ocamlyacc_value,
    ACTIONS(2526), 1,
      sym_prefix_operator,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(2676), 1,
      sym__lowercase_identifier,
    STATE(1506), 1,
      sym_attribute,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(1697), 1,
      sym__value_name,
    STATE(1746), 1,
      sym__simple_expression,
    STATE(6553), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
    STATE(1677), 18,
      sym__delimited_expression,
      sym_typed_expression,
      sym_prefix_expression,
      sym_hash_expression,
      sym_field_get_expression,
      sym_array_get_expression,
      sym_string_get_expression,
      sym_bigarray_get_expression,
      sym_local_open_expression,
      sym_new_expression,
      sym_method_invocation,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym__constant,
      sym_value_path,
      sym_constructor_path,
      sym_tag,
  [15109] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1507), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1561), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1559), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [15178] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1508), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1565), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1563), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [15247] = 34,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1483), 1,
      anon_sym_COLON2,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2720), 1,
      anon_sym_EQ,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1509), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(5725), 1,
      sym__maybe_polymorphic_typed,
    STATE(6390), 1,
      sym_module_path,
    STATE(6410), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [15374] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2726), 1,
      aux_sym_attribute_id_token1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1510), 2,
      sym_attribute,
      aux_sym_attribute_id_repeat1,
    ACTIONS(2724), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2722), 32,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [15445] = 34,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1483), 1,
      anon_sym_COLON2,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2729), 1,
      anon_sym_EQ,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1511), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(5866), 1,
      sym__maybe_polymorphic_typed,
    STATE(6390), 1,
      sym_module_path,
    STATE(6999), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [15572] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1512), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1541), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1539), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [15641] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1513), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1557), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1555), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [15710] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1514), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1553), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1551), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [15779] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2735), 1,
      aux_sym_attribute_id_token1,
    STATE(1510), 1,
      aux_sym_attribute_id_repeat1,
    STATE(1515), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2733), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2731), 32,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [15852] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1516), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1545), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1543), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [15921] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1955), 1,
      anon_sym_LT_DASH,
    STATE(1517), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1747), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1745), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [15992] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1518), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1549), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1547), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [16061] = 34,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1483), 1,
      anon_sym_COLON2,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2737), 1,
      anon_sym_EQ,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1519), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(5628), 1,
      sym__maybe_polymorphic_typed,
    STATE(6390), 1,
      sym_module_path,
    STATE(6575), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [16188] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2739), 1,
      anon_sym_LT_DASH,
    STATE(1520), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1707), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1705), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [16259] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2735), 1,
      aux_sym_attribute_id_token1,
    STATE(1515), 1,
      aux_sym_attribute_id_repeat1,
    STATE(1521), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2743), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2741), 32,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [16332] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1522), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1537), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1535), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [16401] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1523), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1533), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1531), 32,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LT_DASH,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [16470] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1524), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1795), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1793), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [16538] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1525), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1723), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1721), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [16606] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1526), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1735), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1733), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [16674] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2745), 1,
      anon_sym_LT_DASH,
    STATE(1527), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1707), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1705), 30,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [16744] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1528), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1815), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1813), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [16812] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1529), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1819), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1817), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [16880] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1341), 1,
      anon_sym_DOT,
    ACTIONS(1343), 1,
      anon_sym_POUND,
    ACTIONS(1345), 1,
      sym_hash_operator,
    STATE(1530), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1617), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1615), 30,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [16954] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1531), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1827), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1825), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17022] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1532), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1831), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1829), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17090] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1533), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1835), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1833), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17158] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1534), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1787), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1785), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17226] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1535), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1799), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1797), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17294] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1536), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1807), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1805), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17362] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1537), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1759), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1757), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17430] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1538), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2724), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
      aux_sym_attribute_id_token1,
    ACTIONS(2722), 32,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [17498] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1539), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1715), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1713), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17566] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1540), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1811), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1809), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17634] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1341), 1,
      anon_sym_DOT,
    ACTIONS(1343), 1,
      anon_sym_POUND,
    ACTIONS(1345), 1,
      sym_hash_operator,
    STATE(1541), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1665), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1663), 30,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17708] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1542), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1727), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1725), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17776] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1543), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1707), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1705), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17844] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1544), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1771), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1769), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17912] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1545), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1669), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1667), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [17980] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1546), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1731), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1729), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18048] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1547), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1823), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1821), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18116] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1548), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1577), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1575), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18184] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1549), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1597), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1595), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18252] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1550), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1613), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1611), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18320] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1551), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1673), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1671), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18388] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1552), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1677), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1675), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18456] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1553), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1681), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1679), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18524] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1554), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1853), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1851), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18592] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1555), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1719), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1717), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18660] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1556), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1839), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1837), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18728] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1557), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1581), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1579), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18796] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1341), 1,
      anon_sym_DOT,
    STATE(1558), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1585), 21,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1583), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18866] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2747), 1,
      anon_sym_LT_DASH,
    STATE(1559), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1707), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1705), 32,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [18936] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1560), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1589), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1587), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19004] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1561), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1791), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1789), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19072] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1562), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1601), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1599), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19140] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1563), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1605), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1603), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19208] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1564), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1625), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1623), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19276] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1565), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1803), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1801), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19344] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1566), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1633), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1631), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19412] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1567), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1637), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1635), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19480] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1568), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1641), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1639), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19548] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1569), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1645), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1643), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19616] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1570), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1649), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1647), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19684] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1571), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1651), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19752] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_DOT,
    STATE(1572), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1701), 21,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1699), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19822] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1573), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1661), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1659), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19890] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1574), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1845), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1843), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [19958] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1575), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1767), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1765), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20026] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1576), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1775), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1773), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20094] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1703), 1,
      anon_sym_DOT,
    STATE(1577), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1701), 21,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1699), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20164] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1578), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1685), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1683), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20232] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1579), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1743), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1741), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20300] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1580), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1693), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1691), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20368] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1581), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1689), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1687), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20436] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1582), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1739), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1737), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20504] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1583), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1697), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1695), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20572] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1584), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1593), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1591), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20640] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1585), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1621), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1619), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20708] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1586), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1657), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1655), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20776] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1587), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1779), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1777), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20844] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1588), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1783), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1781), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20912] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1589), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1569), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1567), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [20980] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1590), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1751), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1749), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21048] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1591), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1755), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1753), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21116] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1592), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1763), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1761), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21184] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1593), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1849), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1847), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21252] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1594), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1747), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1745), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21320] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1595), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1573), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1571), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21388] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1596), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1711), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1709), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21456] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1597), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1629), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1627), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21524] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1598), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1609), 22,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1607), 31,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [21592] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2751), 1,
      anon_sym_DASH_GT,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1599), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6361), 1,
      sym__simple_typed,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [21713] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2755), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1600), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6349), 1,
      sym__simple_typed,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [21834] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2757), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1601), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(6914), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [21955] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2759), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1602), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(6941), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [22076] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2761), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1603), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(6911), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [22197] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2763), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1604), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(7057), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [22318] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2765), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1605), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(7079), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [22439] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2767), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1606), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(7236), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [22560] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2769), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1607), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(7263), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [22681] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2771), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1608), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(7259), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [22802] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2773), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1609), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(7294), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [22923] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2775), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1610), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(6471), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [23044] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2777), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1611), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(7093), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [23165] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2779), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1612), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(7172), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [23286] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2781), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1613), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(6507), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [23407] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2783), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1614), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(7345), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [23528] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2785), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1615), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(6568), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [23649] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2787), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1616), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(6577), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [23770] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2789), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1617), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(6737), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [23891] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2791), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1618), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(6665), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [24012] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2793), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1619), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(6610), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [24133] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2795), 1,
      anon_sym_LT_DASH,
    STATE(1620), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1707), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1705), 31,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [24202] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2797), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1621), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(6646), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [24323] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2799), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1622), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(7036), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [24444] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2801), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1623), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(7037), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [24565] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2803), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1624), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(7094), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [24686] = 32,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2749), 1,
      anon_sym_COLON2,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2805), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1625), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    STATE(7274), 1,
      sym__simple_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [24807] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2809), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1626), 2,
      sym_attribute,
      aux_sym__structure_repeat1,
    ACTIONS(2812), 18,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2807), 31,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [24875] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2814), 1,
      anon_sym_COLON,
    STATE(1627), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1971), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1967), 30,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [24943] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1451), 1,
      anon_sym_PERCENT,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1613), 1,
      aux_sym_let_binding_repeat1,
    STATE(1628), 1,
      sym_attribute,
    STATE(1705), 1,
      sym__attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [25061] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1451), 1,
      anon_sym_PERCENT,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1625), 1,
      aux_sym_let_binding_repeat1,
    STATE(1629), 1,
      sym_attribute,
    STATE(1712), 1,
      sym__attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [25179] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1451), 1,
      anon_sym_PERCENT,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1619), 1,
      aux_sym_let_binding_repeat1,
    STATE(1630), 1,
      sym_attribute,
    STATE(1717), 1,
      sym__attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [25297] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1451), 1,
      anon_sym_PERCENT,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1601), 1,
      aux_sym_let_binding_repeat1,
    STATE(1631), 1,
      sym_attribute,
    STATE(1720), 1,
      sym__attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [25415] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1451), 1,
      anon_sym_PERCENT,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1603), 1,
      aux_sym_let_binding_repeat1,
    STATE(1632), 1,
      sym_attribute,
    STATE(1678), 1,
      sym__attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [25533] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1451), 1,
      anon_sym_PERCENT,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1604), 1,
      aux_sym_let_binding_repeat1,
    STATE(1633), 1,
      sym_attribute,
    STATE(1725), 1,
      sym__attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [25651] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1451), 1,
      anon_sym_PERCENT,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1607), 1,
      aux_sym_let_binding_repeat1,
    STATE(1634), 1,
      sym_attribute,
    STATE(1729), 1,
      sym__attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [25769] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1451), 1,
      anon_sym_PERCENT,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1610), 1,
      aux_sym_let_binding_repeat1,
    STATE(1635), 1,
      sym_attribute,
    STATE(1733), 1,
      sym__attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [25887] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1451), 1,
      anon_sym_PERCENT,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1612), 1,
      aux_sym_let_binding_repeat1,
    STATE(1636), 1,
      sym_attribute,
    STATE(1738), 1,
      sym__attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [26005] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1451), 1,
      anon_sym_PERCENT,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1615), 1,
      aux_sym_let_binding_repeat1,
    STATE(1637), 1,
      sym_attribute,
    STATE(1741), 1,
      sym__attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [26123] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2816), 1,
      anon_sym_LT_DASH,
    STATE(1638), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1707), 21,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1705), 29,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [26191] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1451), 1,
      anon_sym_PERCENT,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1624), 1,
      aux_sym_let_binding_repeat1,
    STATE(1639), 1,
      sym_attribute,
    STATE(1743), 1,
      sym__attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [26309] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2818), 1,
      anon_sym_LT_DASH,
    STATE(1640), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1707), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1705), 30,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_do,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [26377] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1451), 1,
      anon_sym_PERCENT,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1606), 1,
      aux_sym_let_binding_repeat1,
    STATE(1641), 1,
      sym_attribute,
    STATE(1688), 1,
      sym__attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [26495] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1451), 1,
      anon_sym_PERCENT,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1622), 1,
      aux_sym_let_binding_repeat1,
    STATE(1642), 1,
      sym_attribute,
    STATE(1672), 1,
      sym__attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [26613] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1643), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1745), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1747), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [26678] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1644), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1617), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1615), 30,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [26743] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_DOT,
    STATE(1645), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1701), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1699), 26,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [26810] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1646), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2822), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(2820), 31,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [26875] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      anon_sym_COMMA,
    ACTIONS(2832), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2842), 1,
      sym_concat_operator,
    ACTIONS(2844), 1,
      sym_rel_operator,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2850), 1,
      sym_assign_operator,
    STATE(675), 1,
      sym_pow_operator,
    STATE(676), 1,
      sym_mult_operator,
    STATE(677), 1,
      sym_add_operator,
    STATE(678), 1,
      sym_and_operator,
    STATE(679), 1,
      sym_or_operator,
    STATE(1647), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2826), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2824), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [26974] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      anon_sym_COMMA,
    ACTIONS(2832), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2842), 1,
      sym_concat_operator,
    ACTIONS(2844), 1,
      sym_rel_operator,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2850), 1,
      sym_assign_operator,
    ACTIONS(2856), 1,
      anon_sym_SEMI,
    STATE(675), 1,
      sym_pow_operator,
    STATE(676), 1,
      sym_mult_operator,
    STATE(677), 1,
      sym_add_operator,
    STATE(678), 1,
      sym_and_operator,
    STATE(679), 1,
      sym_or_operator,
    STATE(1648), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2854), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2852), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [27075] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2858), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1649), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [27190] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1703), 1,
      anon_sym_DOT,
    STATE(1650), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1701), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1699), 26,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [27257] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(675), 1,
      sym_pow_operator,
    STATE(676), 1,
      sym_mult_operator,
    STATE(677), 1,
      sym_add_operator,
    STATE(678), 1,
      sym_and_operator,
    STATE(679), 1,
      sym_or_operator,
    STATE(1651), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 27,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [27334] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      anon_sym_COMMA,
    ACTIONS(2832), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2842), 1,
      sym_concat_operator,
    ACTIONS(2844), 1,
      sym_rel_operator,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2850), 1,
      sym_assign_operator,
    STATE(675), 1,
      sym_pow_operator,
    STATE(676), 1,
      sym_mult_operator,
    STATE(677), 1,
      sym_add_operator,
    STATE(678), 1,
      sym_and_operator,
    STATE(679), 1,
      sym_or_operator,
    STATE(1652), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2866), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2864), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [27433] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(675), 1,
      sym_pow_operator,
    STATE(676), 1,
      sym_mult_operator,
    STATE(677), 1,
      sym_add_operator,
    STATE(678), 1,
      sym_and_operator,
    STATE(679), 1,
      sym_or_operator,
    STATE(1653), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [27514] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1654), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2019), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2017), 30,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_else,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [27579] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2868), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1655), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [27694] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2870), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1656), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [27809] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(675), 1,
      sym_pow_operator,
    STATE(676), 1,
      sym_mult_operator,
    STATE(677), 1,
      sym_add_operator,
    STATE(678), 1,
      sym_and_operator,
    STATE(679), 1,
      sym_or_operator,
    STATE(1657), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2874), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2872), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [27884] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2876), 1,
      anon_sym_LT_DASH,
    STATE(1658), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1707), 20,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      aux_sym_pow_operator_token1,
      sym_concat_operator,
      sym_assign_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1705), 29,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_else,
      anon_sym_do,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_or,
      anon_sym_PIPE_PIPE,
      sym__lowercase_identifier,
  [27951] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      anon_sym_COMMA,
    ACTIONS(2832), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2842), 1,
      sym_concat_operator,
    ACTIONS(2844), 1,
      sym_rel_operator,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2850), 1,
      sym_assign_operator,
    ACTIONS(2856), 1,
      anon_sym_SEMI,
    STATE(675), 1,
      sym_pow_operator,
    STATE(676), 1,
      sym_mult_operator,
    STATE(677), 1,
      sym_add_operator,
    STATE(678), 1,
      sym_and_operator,
    STATE(679), 1,
      sym_or_operator,
    STATE(1659), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2710), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2708), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [28052] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      anon_sym_COMMA,
    ACTIONS(2832), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2842), 1,
      sym_concat_operator,
    ACTIONS(2844), 1,
      sym_rel_operator,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    STATE(675), 1,
      sym_pow_operator,
    STATE(676), 1,
      sym_mult_operator,
    STATE(677), 1,
      sym_add_operator,
    STATE(678), 1,
      sym_and_operator,
    STATE(679), 1,
      sym_or_operator,
    STATE(1660), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2880), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2878), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [28149] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2832), 1,
      anon_sym_COLON_COLON,
    STATE(675), 1,
      sym_pow_operator,
    STATE(676), 1,
      sym_mult_operator,
    STATE(677), 1,
      sym_add_operator,
    STATE(678), 1,
      sym_and_operator,
    STATE(679), 1,
      sym_or_operator,
    STATE(1661), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2884), 9,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [28234] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1662), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1847), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1849), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [28299] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2753), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2886), 1,
      anon_sym_DASH_GT,
    STATE(1145), 1,
      aux_sym_let_binding_repeat1,
    STATE(1663), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [28414] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2832), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2842), 1,
      sym_concat_operator,
    STATE(675), 1,
      sym_pow_operator,
    STATE(676), 1,
      sym_mult_operator,
    STATE(677), 1,
      sym_add_operator,
    STATE(678), 1,
      sym_and_operator,
    STATE(679), 1,
      sym_or_operator,
    STATE(1664), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 9,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [28501] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(675), 1,
      sym_pow_operator,
    STATE(676), 1,
      sym_mult_operator,
    STATE(677), 1,
      sym_add_operator,
    STATE(678), 1,
      sym_and_operator,
    STATE(679), 1,
      sym_or_operator,
    STATE(1665), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 27,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [28578] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2832), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2842), 1,
      sym_concat_operator,
    STATE(675), 1,
      sym_pow_operator,
    STATE(676), 1,
      sym_mult_operator,
    STATE(677), 1,
      sym_add_operator,
    STATE(678), 1,
      sym_and_operator,
    STATE(679), 1,
      sym_or_operator,
    STATE(1666), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 9,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [28665] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      anon_sym_COMMA,
    ACTIONS(2832), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2842), 1,
      sym_concat_operator,
    ACTIONS(2844), 1,
      sym_rel_operator,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2850), 1,
      sym_assign_operator,
    STATE(675), 1,
      sym_pow_operator,
    STATE(676), 1,
      sym_mult_operator,
    STATE(677), 1,
      sym_add_operator,
    STATE(678), 1,
      sym_and_operator,
    STATE(679), 1,
      sym_or_operator,
    STATE(1667), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2890), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2888), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [28764] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2832), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2842), 1,
      sym_concat_operator,
    ACTIONS(2844), 1,
      sym_rel_operator,
    STATE(675), 1,
      sym_pow_operator,
    STATE(676), 1,
      sym_mult_operator,
    STATE(677), 1,
      sym_add_operator,
    STATE(678), 1,
      sym_and_operator,
    STATE(679), 1,
      sym_or_operator,
    STATE(1668), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 7,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [28855] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      anon_sym_COMMA,
    ACTIONS(2832), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2842), 1,
      sym_concat_operator,
    ACTIONS(2844), 1,
      sym_rel_operator,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2850), 1,
      sym_assign_operator,
    ACTIONS(2856), 1,
      anon_sym_SEMI,
    STATE(675), 1,
      sym_pow_operator,
    STATE(676), 1,
      sym_mult_operator,
    STATE(677), 1,
      sym_add_operator,
    STATE(678), 1,
      sym_and_operator,
    STATE(679), 1,
      sym_or_operator,
    STATE(1669), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2718), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2716), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [28956] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      anon_sym_COMMA,
    ACTIONS(2832), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2842), 1,
      sym_concat_operator,
    ACTIONS(2844), 1,
      sym_rel_operator,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2850), 1,
      sym_assign_operator,
    STATE(675), 1,
      sym_pow_operator,
    STATE(676), 1,
      sym_mult_operator,
    STATE(677), 1,
      sym_add_operator,
    STATE(678), 1,
      sym_and_operator,
    STATE(679), 1,
      sym_or_operator,
    STATE(1670), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2860), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [29055] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2832), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2842), 1,
      sym_concat_operator,
    ACTIONS(2844), 1,
      sym_rel_operator,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    STATE(675), 1,
      sym_pow_operator,
    STATE(676), 1,
      sym_mult_operator,
    STATE(677), 1,
      sym_add_operator,
    STATE(678), 1,
      sym_and_operator,
    STATE(679), 1,
      sym_or_operator,
    STATE(1671), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2860), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [29150] = 29,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1623), 1,
      aux_sym_let_binding_repeat1,
    STATE(1672), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [29262] = 29,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1649), 1,
      aux_sym_let_binding_repeat1,
    STATE(1673), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [29374] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(620), 1,
      sym_pow_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(713), 1,
      sym_mult_operator,
    STATE(1674), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 27,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [29450] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(620), 1,
      sym_pow_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(713), 1,
      sym_mult_operator,
    STATE(1675), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 27,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [29526] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(620), 1,
      sym_pow_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(713), 1,
      sym_mult_operator,
    STATE(1676), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 11,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [29606] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1677), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1705), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1707), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [29670] = 29,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1600), 1,
      aux_sym_let_binding_repeat1,
    STATE(1678), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [29782] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1679), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1837), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1839), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [29846] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1680), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1599), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1601), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [29910] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1039), 1,
      anon_sym_LPAREN,
    ACTIONS(1047), 1,
      anon_sym_LBRACK,
    ACTIONS(1051), 1,
      anon_sym_LBRACE,
    ACTIONS(1061), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1065), 1,
      anon_sym_begin,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(2083), 1,
      anon_sym_POUND,
    ACTIONS(2196), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2676), 1,
      sym__lowercase_identifier,
    ACTIONS(2892), 1,
      anon_sym_lazy,
    STATE(1210), 1,
      sym_constructor_path,
    STATE(1278), 1,
      sym_tag,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1681), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(6379), 1,
      sym_module_path,
    STATE(7004), 1,
      sym__effect_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4254), 2,
      sym__constant,
      sym_signed_number,
    STATE(4553), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4564), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4548), 9,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
  [30026] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1682), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1603), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1605), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [30090] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1683), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1607), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1609), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [30154] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1684), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1619), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1621), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [30218] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1703), 1,
      anon_sym_DOT,
    STATE(1685), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1699), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1701), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [30284] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1686), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1531), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1533), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [30348] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1065), 1,
      anon_sym_begin,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1099), 1,
      sym__lowercase_identifier,
    ACTIONS(1103), 1,
      anon_sym_LPAREN,
    ACTIONS(1105), 1,
      anon_sym_LBRACK,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1115), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2083), 1,
      anon_sym_POUND,
    ACTIONS(2196), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2894), 1,
      anon_sym_lazy,
    STATE(1252), 1,
      sym_constructor_path,
    STATE(1297), 1,
      sym_tag,
    STATE(1687), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(6479), 1,
      sym__effect_pattern,
    STATE(6987), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4254), 2,
      sym__constant,
      sym_signed_number,
    STATE(4430), 4,
      sym__simple_pattern,
      sym_constructor_pattern,
      sym_tag_pattern,
      sym_lazy_pattern,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4414), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(4428), 10,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
  [30462] = 29,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1608), 1,
      aux_sym_let_binding_repeat1,
    STATE(1688), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [30574] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1689), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1579), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1581), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [30638] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1690), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1571), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1573), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [30702] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1039), 1,
      anon_sym_LPAREN,
    ACTIONS(1047), 1,
      anon_sym_LBRACK,
    ACTIONS(1051), 1,
      anon_sym_LBRACE,
    ACTIONS(1061), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1065), 1,
      anon_sym_begin,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(2083), 1,
      anon_sym_POUND,
    ACTIONS(2196), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2676), 1,
      sym__lowercase_identifier,
    ACTIONS(2892), 1,
      anon_sym_lazy,
    STATE(1210), 1,
      sym_constructor_path,
    STATE(1278), 1,
      sym_tag,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1691), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(6379), 1,
      sym_module_path,
    STATE(6609), 1,
      sym__effect_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4254), 2,
      sym__constant,
      sym_signed_number,
    STATE(4553), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4564), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4548), 9,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
  [30818] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1692), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1639), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1641), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [30882] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1693), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1643), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1645), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [30946] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1694), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1713), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1715), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [31010] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1065), 1,
      anon_sym_begin,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1099), 1,
      sym__lowercase_identifier,
    ACTIONS(1103), 1,
      anon_sym_LPAREN,
    ACTIONS(1105), 1,
      anon_sym_LBRACK,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1115), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2083), 1,
      anon_sym_POUND,
    ACTIONS(2196), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2894), 1,
      anon_sym_lazy,
    STATE(1252), 1,
      sym_constructor_path,
    STATE(1297), 1,
      sym_tag,
    STATE(1695), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(6987), 1,
      sym_module_path,
    STATE(7058), 1,
      sym__effect_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4254), 2,
      sym__constant,
      sym_signed_number,
    STATE(4430), 4,
      sym__simple_pattern,
      sym_constructor_pattern,
      sym_tag_pattern,
      sym_lazy_pattern,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4414), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(4428), 10,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
  [31124] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2896), 1,
      anon_sym_COMMA,
    ACTIONS(2898), 1,
      anon_sym_SEMI,
    ACTIONS(2900), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2902), 1,
      sym_concat_operator,
    ACTIONS(2904), 1,
      sym_rel_operator,
    ACTIONS(2906), 1,
      sym_assign_operator,
    STATE(620), 1,
      sym_pow_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(713), 1,
      sym_mult_operator,
    STATE(1696), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2710), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2708), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [31224] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1697), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1691), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1693), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [31288] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1698), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1715), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1713), 26,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [31352] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1699), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1647), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1649), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [31416] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1700), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1575), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1577), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [31480] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1701), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1651), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1653), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [31544] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1702), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1625), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1623), 26,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [31608] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1703), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1595), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1597), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [31672] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1704), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1737), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1739), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [31736] = 29,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1617), 1,
      aux_sym_let_binding_repeat1,
    STATE(1705), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [31848] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1039), 1,
      anon_sym_LPAREN,
    ACTIONS(1047), 1,
      anon_sym_LBRACK,
    ACTIONS(1051), 1,
      anon_sym_LBRACE,
    ACTIONS(1061), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1065), 1,
      anon_sym_begin,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(2083), 1,
      anon_sym_POUND,
    ACTIONS(2196), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2676), 1,
      sym__lowercase_identifier,
    ACTIONS(2892), 1,
      anon_sym_lazy,
    STATE(1210), 1,
      sym_constructor_path,
    STATE(1278), 1,
      sym_tag,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1706), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(6379), 1,
      sym_module_path,
    STATE(7328), 1,
      sym__effect_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4254), 2,
      sym__constant,
      sym_signed_number,
    STATE(4553), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4564), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4548), 9,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
  [31964] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2896), 1,
      anon_sym_COMMA,
    ACTIONS(2898), 1,
      anon_sym_SEMI,
    ACTIONS(2900), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2902), 1,
      sym_concat_operator,
    ACTIONS(2904), 1,
      sym_rel_operator,
    ACTIONS(2906), 1,
      sym_assign_operator,
    STATE(620), 1,
      sym_pow_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(713), 1,
      sym_mult_operator,
    STATE(1707), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2718), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2716), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [32064] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1708), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1627), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1629), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [32128] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1709), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1631), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1633), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [32192] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1710), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1655), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1657), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [32256] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1711), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1809), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1811), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [32320] = 29,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1599), 1,
      aux_sym_let_binding_repeat1,
    STATE(1712), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [32432] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1039), 1,
      anon_sym_LPAREN,
    ACTIONS(1047), 1,
      anon_sym_LBRACK,
    ACTIONS(1051), 1,
      anon_sym_LBRACE,
    ACTIONS(1061), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1065), 1,
      anon_sym_begin,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(2083), 1,
      anon_sym_POUND,
    ACTIONS(2196), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2676), 1,
      sym__lowercase_identifier,
    ACTIONS(2892), 1,
      anon_sym_lazy,
    STATE(1210), 1,
      sym_constructor_path,
    STATE(1278), 1,
      sym_tag,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1713), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(6379), 1,
      sym_module_path,
    STATE(6633), 1,
      sym__effect_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4254), 2,
      sym__constant,
      sym_signed_number,
    STATE(4553), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4564), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4548), 9,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
  [32548] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1714), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1793), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1795), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [32612] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1715), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1671), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1673), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [32676] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1716), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1675), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1677), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [32740] = 29,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1621), 1,
      aux_sym_let_binding_repeat1,
    STATE(1717), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [32852] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1718), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1667), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1669), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [32916] = 31,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1039), 1,
      anon_sym_LPAREN,
    ACTIONS(1047), 1,
      anon_sym_LBRACK,
    ACTIONS(1051), 1,
      anon_sym_LBRACE,
    ACTIONS(1061), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1065), 1,
      anon_sym_begin,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(2083), 1,
      anon_sym_POUND,
    ACTIONS(2196), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2676), 1,
      sym__lowercase_identifier,
    ACTIONS(2892), 1,
      anon_sym_lazy,
    STATE(1210), 1,
      sym_constructor_path,
    STATE(1278), 1,
      sym_tag,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1719), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(6379), 1,
      sym_module_path,
    STATE(7171), 1,
      sym__effect_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4254), 2,
      sym__constant,
      sym_signed_number,
    STATE(4553), 4,
      sym__simple_binding_pattern,
      sym_constructor_binding_pattern,
      sym_tag_binding_pattern,
      sym_lazy_binding_pattern,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4564), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4548), 9,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
  [33032] = 29,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1602), 1,
      aux_sym_let_binding_repeat1,
    STATE(1720), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [33144] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2896), 1,
      anon_sym_COMMA,
    ACTIONS(2900), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2902), 1,
      sym_concat_operator,
    ACTIONS(2904), 1,
      sym_rel_operator,
    STATE(620), 1,
      sym_pow_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(713), 1,
      sym_mult_operator,
    STATE(1721), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2880), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2878), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [33240] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2900), 1,
      anon_sym_COLON_COLON,
    STATE(620), 1,
      sym_pow_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(713), 1,
      sym_mult_operator,
    STATE(1722), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2884), 8,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [33324] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1723), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1659), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1661), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [33388] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1724), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1761), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1763), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [33452] = 29,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1605), 1,
      aux_sym_let_binding_repeat1,
    STATE(1725), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [33564] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2900), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2902), 1,
      sym_concat_operator,
    STATE(620), 1,
      sym_pow_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(713), 1,
      sym_mult_operator,
    STATE(1726), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 8,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [33650] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1727), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1559), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1561), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [33714] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2896), 1,
      anon_sym_COMMA,
    ACTIONS(2900), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2902), 1,
      sym_concat_operator,
    ACTIONS(2904), 1,
      sym_rel_operator,
    ACTIONS(2906), 1,
      sym_assign_operator,
    STATE(620), 1,
      sym_pow_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(713), 1,
      sym_mult_operator,
    STATE(1728), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2860), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [33812] = 29,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1609), 1,
      aux_sym_let_binding_repeat1,
    STATE(1729), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [33924] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1730), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1563), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1565), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [33988] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2900), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2902), 1,
      sym_concat_operator,
    ACTIONS(2904), 1,
      sym_rel_operator,
    STATE(620), 1,
      sym_pow_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(713), 1,
      sym_mult_operator,
    STATE(1731), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 6,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [34078] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2900), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2902), 1,
      sym_concat_operator,
    ACTIONS(2904), 1,
      sym_rel_operator,
    STATE(620), 1,
      sym_pow_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(713), 1,
      sym_mult_operator,
    STATE(1732), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2860), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [34172] = 29,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1611), 1,
      aux_sym_let_binding_repeat1,
    STATE(1733), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [34284] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2896), 1,
      anon_sym_COMMA,
    ACTIONS(2900), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2902), 1,
      sym_concat_operator,
    ACTIONS(2904), 1,
      sym_rel_operator,
    ACTIONS(2906), 1,
      sym_assign_operator,
    STATE(620), 1,
      sym_pow_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(713), 1,
      sym_mult_operator,
    STATE(1734), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2826), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2824), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [34382] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2896), 1,
      anon_sym_COMMA,
    ACTIONS(2900), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2902), 1,
      sym_concat_operator,
    ACTIONS(2904), 1,
      sym_rel_operator,
    ACTIONS(2906), 1,
      sym_assign_operator,
    STATE(620), 1,
      sym_pow_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(713), 1,
      sym_mult_operator,
    STATE(1735), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2866), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2864), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [34480] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1736), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1539), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1541), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [34544] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1737), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1811), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1809), 26,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [34608] = 29,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1614), 1,
      aux_sym_let_binding_repeat1,
    STATE(1738), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [34720] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2896), 1,
      anon_sym_COMMA,
    ACTIONS(2900), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2902), 1,
      sym_concat_operator,
    ACTIONS(2904), 1,
      sym_rel_operator,
    ACTIONS(2906), 1,
      sym_assign_operator,
    STATE(620), 1,
      sym_pow_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(713), 1,
      sym_mult_operator,
    STATE(1739), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2890), 5,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2888), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [34818] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1740), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1679), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1681), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [34882] = 29,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1616), 1,
      aux_sym_let_binding_repeat1,
    STATE(1741), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [34994] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1742), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1635), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1637), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [35058] = 29,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1618), 1,
      aux_sym_let_binding_repeat1,
    STATE(1743), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [35170] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1744), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1765), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1767), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [35234] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1745), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1773), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1775), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [35298] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1746), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1611), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1613), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [35362] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1747), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1777), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1779), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [35426] = 29,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1656), 1,
      aux_sym_let_binding_repeat1,
    STATE(1748), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [35538] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1749), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1781), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1783), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [35602] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1750), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1821), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1823), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [35666] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_DOT,
    STATE(1751), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1699), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1701), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [35732] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1752), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1695), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1697), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [35796] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1753), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1843), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1845), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [35860] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1754), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1567), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1569), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [35924] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1755), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1851), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1853), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [35988] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1756), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1555), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1557), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [36052] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1757), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1551), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1553), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [36116] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1758), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1587), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1589), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [36180] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1759), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1591), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1593), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [36244] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1760), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1717), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1719), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [36308] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1761), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1729), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1731), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [36372] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1762), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1845), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1843), 26,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [36436] = 29,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1663), 1,
      aux_sym_let_binding_repeat1,
    STATE(1763), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [36548] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2896), 1,
      anon_sym_COMMA,
    ACTIONS(2898), 1,
      anon_sym_SEMI,
    ACTIONS(2900), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2902), 1,
      sym_concat_operator,
    ACTIONS(2904), 1,
      sym_rel_operator,
    ACTIONS(2906), 1,
      sym_assign_operator,
    STATE(620), 1,
      sym_pow_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(713), 1,
      sym_mult_operator,
    STATE(1764), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2854), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2852), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [36648] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1765), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1543), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1545), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [36712] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1766), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1741), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1743), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [36776] = 29,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1477), 1,
      anon_sym_LPAREN,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1655), 1,
      aux_sym_let_binding_repeat1,
    STATE(1767), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2243), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1475), 2,
      anon_sym_TILDE,
      anon_sym_QMARK,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2207), 2,
      sym_parameter,
      sym__parenthesized_abstract_type,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [36888] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1768), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1749), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1751), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [36952] = 30,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1065), 1,
      anon_sym_begin,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1099), 1,
      sym__lowercase_identifier,
    ACTIONS(1103), 1,
      anon_sym_LPAREN,
    ACTIONS(1105), 1,
      anon_sym_LBRACK,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1115), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2083), 1,
      anon_sym_POUND,
    ACTIONS(2196), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2894), 1,
      anon_sym_lazy,
    STATE(1252), 1,
      sym_constructor_path,
    STATE(1297), 1,
      sym_tag,
    STATE(1769), 1,
      sym_attribute,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(6813), 1,
      sym__effect_pattern,
    STATE(6987), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4254), 2,
      sym__constant,
      sym_signed_number,
    STATE(4430), 4,
      sym__simple_pattern,
      sym_constructor_pattern,
      sym_tag_pattern,
      sym_lazy_pattern,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4414), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(4428), 10,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
  [37066] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1770), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1753), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1755), 33,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [37130] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(620), 1,
      sym_pow_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(713), 1,
      sym_mult_operator,
    STATE(1771), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2874), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2872), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [37204] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2900), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2902), 1,
      sym_concat_operator,
    STATE(620), 1,
      sym_pow_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(713), 1,
      sym_mult_operator,
    STATE(1772), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 8,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [37290] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1773), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1651), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1653), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [37353] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1774), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1655), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1657), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [37416] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1775), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2710), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2708), 34,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [37479] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1776), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1785), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1787), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [37542] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1777), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1777), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1779), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [37605] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(534), 1,
      sym_pow_operator,
    STATE(535), 1,
      sym_mult_operator,
    STATE(536), 1,
      sym_add_operator,
    STATE(537), 1,
      sym_and_operator,
    STATE(538), 1,
      sym_or_operator,
    STATE(1778), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 27,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [37680] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(534), 1,
      sym_pow_operator,
    STATE(535), 1,
      sym_mult_operator,
    STATE(536), 1,
      sym_add_operator,
    STATE(537), 1,
      sym_and_operator,
    STATE(538), 1,
      sym_or_operator,
    STATE(1779), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 27,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [37755] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    STATE(534), 1,
      sym_pow_operator,
    STATE(535), 1,
      sym_mult_operator,
    STATE(536), 1,
      sym_add_operator,
    STATE(537), 1,
      sym_and_operator,
    STATE(538), 1,
      sym_or_operator,
    STATE(1780), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 11,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 23,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [37834] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1781), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1781), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1783), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [37897] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1782), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1567), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1569), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [37960] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_DOT,
    STATE(1783), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1701), 23,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1699), 24,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [38025] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1784), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1797), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1799), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [38088] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1785), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1749), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1751), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [38151] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1786), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1753), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1755), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [38214] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1787), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1761), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1763), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [38277] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1788), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1805), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1807), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [38340] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1789), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1793), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1795), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [38403] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1790), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1757), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1759), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [38466] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1791), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1725), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1727), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [38529] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1792), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2718), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2716), 34,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [38592] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      anon_sym_COMMA,
    ACTIONS(2832), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2842), 1,
      sym_concat_operator,
    ACTIONS(2844), 1,
      sym_rel_operator,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2850), 1,
      sym_assign_operator,
    ACTIONS(2908), 1,
      anon_sym_SEMI,
    STATE(675), 1,
      sym_pow_operator,
    STATE(676), 1,
      sym_mult_operator,
    STATE(677), 1,
      sym_add_operator,
    STATE(678), 1,
      sym_and_operator,
    STATE(679), 1,
      sym_or_operator,
    STATE(1793), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2854), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2852), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [38691] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1794), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1631), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1633), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [38754] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1795), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1683), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(1685), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [38817] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1796), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1771), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [38880] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1797), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1789), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1791), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [38943] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1798), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1801), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1803), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39006] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1799), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1687), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(1689), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [39069] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1800), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1639), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1641), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39132] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1801), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1643), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1645), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39195] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1802), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1647), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1649), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39258] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1803), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1709), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(1711), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [39321] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1804), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1635), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1637), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39384] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1805), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1659), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1661), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39447] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1806), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1559), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1561), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39510] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1807), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1563), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1565), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39573] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1808), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1539), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1541), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39636] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1809), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1765), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1767), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39699] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1810), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1773), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1775), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39762] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1811), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1745), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1747), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [39825] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2914), 1,
      anon_sym_DOT,
    ACTIONS(2916), 1,
      anon_sym_POUND,
    ACTIONS(2918), 1,
      sym_hash_operator,
    STATE(1812), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2912), 15,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2910), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [39894] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2914), 1,
      anon_sym_DOT,
    ACTIONS(2916), 1,
      anon_sym_POUND,
    ACTIONS(2918), 1,
      sym_hash_operator,
    STATE(1813), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2922), 15,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2920), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [39963] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1814), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1555), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1557), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [40026] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1815), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1551), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1553), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [40089] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1816), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1543), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1545), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [40152] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1817), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1741), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1743), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [40215] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1818), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1547), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1549), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [40278] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2924), 1,
      anon_sym_DOT,
    ACTIONS(2926), 1,
      anon_sym_POUND,
    ACTIONS(2928), 1,
      sym_hash_operator,
    STATE(1819), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2912), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2910), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [40347] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2924), 1,
      anon_sym_DOT,
    ACTIONS(2926), 1,
      anon_sym_POUND,
    ACTIONS(2928), 1,
      sym_hash_operator,
    STATE(1820), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2922), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2920), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [40416] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1821), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1733), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(1735), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [40479] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1822), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1847), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1849), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [40542] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1823), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2930), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2932), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [40605] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      anon_sym_COMMA,
    ACTIONS(2832), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2842), 1,
      sym_concat_operator,
    ACTIONS(2844), 1,
      sym_rel_operator,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2850), 1,
      sym_assign_operator,
    ACTIONS(2908), 1,
      anon_sym_SEMI,
    STATE(675), 1,
      sym_pow_operator,
    STATE(676), 1,
      sym_mult_operator,
    STATE(677), 1,
      sym_add_operator,
    STATE(678), 1,
      sym_and_operator,
    STATE(679), 1,
      sym_or_operator,
    STATE(1824), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2710), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2708), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [40704] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2934), 1,
      anon_sym_DOT_DOT,
    STATE(1825), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1295), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1297), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [40769] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2940), 1,
      anon_sym_COLON,
    STATE(1826), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2938), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2936), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [40834] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1827), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1683), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1685), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [40897] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      anon_sym_COMMA,
    ACTIONS(2832), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2842), 1,
      sym_concat_operator,
    ACTIONS(2844), 1,
      sym_rel_operator,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2850), 1,
      sym_assign_operator,
    ACTIONS(2908), 1,
      anon_sym_SEMI,
    STATE(675), 1,
      sym_pow_operator,
    STATE(676), 1,
      sym_mult_operator,
    STATE(677), 1,
      sym_add_operator,
    STATE(678), 1,
      sym_and_operator,
    STATE(679), 1,
      sym_or_operator,
    STATE(1828), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2718), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2716), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [40996] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1829), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2942), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(2944), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [41059] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1830), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1535), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1537), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [41122] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1831), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1687), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1689), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [41185] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1703), 1,
      anon_sym_DOT,
    STATE(1832), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1699), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1701), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [41250] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1833), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1713), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1715), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [41313] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1834), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1817), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(1819), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [41376] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_DOT,
    STATE(1835), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1699), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1701), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [41441] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1836), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1843), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1845), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [41504] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1837), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1709), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1711), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [41567] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1838), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1691), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1693), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [41630] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1839), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1809), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1811), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [41693] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2924), 1,
      anon_sym_DOT,
    ACTIONS(2926), 1,
      anon_sym_POUND,
    ACTIONS(2928), 1,
      sym_hash_operator,
    STATE(1840), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2948), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2946), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [41762] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2924), 1,
      anon_sym_DOT,
    ACTIONS(2926), 1,
      anon_sym_POUND,
    ACTIONS(2928), 1,
      sym_hash_operator,
    STATE(1841), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2952), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2950), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [41831] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1842), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1829), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(1831), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [41894] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1843), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1833), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(1835), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [41957] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1844), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1547), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1549), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [42020] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2914), 1,
      anon_sym_DOT,
    STATE(1845), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1583), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1585), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [42085] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2954), 1,
      anon_sym_COMMA,
    ACTIONS(2956), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2958), 1,
      sym_concat_operator,
    ACTIONS(2960), 1,
      sym_rel_operator,
    STATE(534), 1,
      sym_pow_operator,
    STATE(535), 1,
      sym_mult_operator,
    STATE(536), 1,
      sym_add_operator,
    STATE(537), 1,
      sym_and_operator,
    STATE(538), 1,
      sym_or_operator,
    STATE(1846), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2880), 4,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2878), 19,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
  [42180] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2956), 1,
      anon_sym_COLON_COLON,
    STATE(534), 1,
      sym_pow_operator,
    STATE(535), 1,
      sym_mult_operator,
    STATE(536), 1,
      sym_add_operator,
    STATE(537), 1,
      sym_and_operator,
    STATE(538), 1,
      sym_or_operator,
    STATE(1847), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2884), 8,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 22,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [42263] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2954), 1,
      anon_sym_COMMA,
    ACTIONS(2956), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2958), 1,
      sym_concat_operator,
    ACTIONS(2960), 1,
      sym_rel_operator,
    ACTIONS(2962), 1,
      anon_sym_SEMI,
    ACTIONS(2964), 1,
      sym_assign_operator,
    STATE(534), 1,
      sym_pow_operator,
    STATE(535), 1,
      sym_mult_operator,
    STATE(536), 1,
      sym_add_operator,
    STATE(537), 1,
      sym_and_operator,
    STATE(538), 1,
      sym_or_operator,
    STATE(1848), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2854), 4,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2852), 17,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [42362] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2956), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2958), 1,
      sym_concat_operator,
    STATE(534), 1,
      sym_pow_operator,
    STATE(535), 1,
      sym_mult_operator,
    STATE(536), 1,
      sym_add_operator,
    STATE(537), 1,
      sym_and_operator,
    STATE(538), 1,
      sym_or_operator,
    STATE(1849), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 8,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 21,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
  [42447] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2956), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2958), 1,
      sym_concat_operator,
    STATE(534), 1,
      sym_pow_operator,
    STATE(535), 1,
      sym_mult_operator,
    STATE(536), 1,
      sym_add_operator,
    STATE(537), 1,
      sym_and_operator,
    STATE(538), 1,
      sym_or_operator,
    STATE(1850), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 8,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 21,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
  [42532] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2954), 1,
      anon_sym_COMMA,
    ACTIONS(2956), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2958), 1,
      sym_concat_operator,
    ACTIONS(2960), 1,
      sym_rel_operator,
    ACTIONS(2964), 1,
      sym_assign_operator,
    STATE(534), 1,
      sym_pow_operator,
    STATE(535), 1,
      sym_mult_operator,
    STATE(536), 1,
      sym_add_operator,
    STATE(537), 1,
      sym_and_operator,
    STATE(538), 1,
      sym_or_operator,
    STATE(1851), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 4,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2860), 18,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [42629] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2956), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2958), 1,
      sym_concat_operator,
    ACTIONS(2960), 1,
      sym_rel_operator,
    STATE(534), 1,
      sym_pow_operator,
    STATE(535), 1,
      sym_mult_operator,
    STATE(536), 1,
      sym_add_operator,
    STATE(537), 1,
      sym_and_operator,
    STATE(538), 1,
      sym_or_operator,
    STATE(1852), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 5,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 21,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
  [42718] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1853), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1607), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1609), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [42781] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2954), 1,
      anon_sym_COMMA,
    ACTIONS(2956), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2958), 1,
      sym_concat_operator,
    ACTIONS(2960), 1,
      sym_rel_operator,
    ACTIONS(2964), 1,
      sym_assign_operator,
    STATE(534), 1,
      sym_pow_operator,
    STATE(535), 1,
      sym_mult_operator,
    STATE(536), 1,
      sym_add_operator,
    STATE(537), 1,
      sym_and_operator,
    STATE(538), 1,
      sym_or_operator,
    STATE(1854), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2826), 4,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2824), 18,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [42878] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1855), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1623), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1625), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [42941] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1703), 1,
      anon_sym_DOT,
    STATE(1856), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1701), 23,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1699), 24,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [43006] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1857), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1721), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1723), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43069] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2954), 1,
      anon_sym_COMMA,
    ACTIONS(2956), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2958), 1,
      sym_concat_operator,
    ACTIONS(2960), 1,
      sym_rel_operator,
    ACTIONS(2964), 1,
      sym_assign_operator,
    STATE(534), 1,
      sym_pow_operator,
    STATE(535), 1,
      sym_mult_operator,
    STATE(536), 1,
      sym_add_operator,
    STATE(537), 1,
      sym_and_operator,
    STATE(538), 1,
      sym_or_operator,
    STATE(1858), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2866), 4,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2864), 18,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [43166] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1859), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1737), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1739), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43229] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1860), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1571), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1573), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43292] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1861), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2714), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2712), 34,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [43355] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1862), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1535), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1537), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43418] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1863), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1733), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1735), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43481] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2954), 1,
      anon_sym_COMMA,
    ACTIONS(2956), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2958), 1,
      sym_concat_operator,
    ACTIONS(2960), 1,
      sym_rel_operator,
    ACTIONS(2964), 1,
      sym_assign_operator,
    STATE(534), 1,
      sym_pow_operator,
    STATE(535), 1,
      sym_mult_operator,
    STATE(536), 1,
      sym_add_operator,
    STATE(537), 1,
      sym_and_operator,
    STATE(538), 1,
      sym_or_operator,
    STATE(1864), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2890), 4,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2888), 18,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [43578] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1865), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1813), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1815), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43641] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1866), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1683), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1685), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [43704] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1867), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1687), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1689), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [43767] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1868), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1709), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1711), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [43830] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1869), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1817), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1819), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [43893] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1870), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1623), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1625), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [43956] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1871), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1695), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1697), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [44019] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2924), 1,
      anon_sym_DOT,
    ACTIONS(2926), 1,
      anon_sym_POUND,
    ACTIONS(2928), 1,
      sym_hash_operator,
    STATE(1872), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2968), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2966), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [44088] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1873), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1721), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1723), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [44151] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1874), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1733), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1735), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [44214] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1875), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1827), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [44277] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1876), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1829), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1831), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [44340] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1877), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1813), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1815), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [44403] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1878), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1817), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1819), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [44466] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1879), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1833), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1835), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [44529] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1880), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1827), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [44592] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1881), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1829), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1831), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [44655] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1882), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1833), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1835), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [44718] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1883), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1591), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1593), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [44781] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2914), 1,
      anon_sym_DOT,
    ACTIONS(2916), 1,
      anon_sym_POUND,
    ACTIONS(2918), 1,
      sym_hash_operator,
    STATE(1884), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2968), 15,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2966), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [44850] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1885), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1785), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1787), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [44913] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1886), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1797), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1799), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [44976] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1887), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1619), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1621), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [45039] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(534), 1,
      sym_pow_operator,
    STATE(535), 1,
      sym_mult_operator,
    STATE(536), 1,
      sym_add_operator,
    STATE(537), 1,
      sym_and_operator,
    STATE(538), 1,
      sym_or_operator,
    STATE(1888), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2874), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2872), 31,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [45112] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1889), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1805), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1807), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [45175] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1890), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1627), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1629), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [45238] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1891), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1757), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1759), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [45301] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2914), 1,
      anon_sym_DOT,
    ACTIONS(2916), 1,
      anon_sym_POUND,
    ACTIONS(2918), 1,
      sym_hash_operator,
    STATE(1892), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2948), 15,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2946), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [45370] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2914), 1,
      anon_sym_DOT,
    ACTIONS(2916), 1,
      anon_sym_POUND,
    ACTIONS(2918), 1,
      sym_hash_operator,
    STATE(1893), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2952), 15,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(2950), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [45439] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2954), 1,
      anon_sym_COMMA,
    ACTIONS(2956), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2958), 1,
      sym_concat_operator,
    ACTIONS(2960), 1,
      sym_rel_operator,
    ACTIONS(2962), 1,
      anon_sym_SEMI,
    ACTIONS(2964), 1,
      sym_assign_operator,
    STATE(534), 1,
      sym_pow_operator,
    STATE(535), 1,
      sym_mult_operator,
    STATE(536), 1,
      sym_add_operator,
    STATE(537), 1,
      sym_and_operator,
    STATE(538), 1,
      sym_or_operator,
    STATE(1894), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2710), 4,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2708), 17,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [45538] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2954), 1,
      anon_sym_COMMA,
    ACTIONS(2956), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2958), 1,
      sym_concat_operator,
    ACTIONS(2960), 1,
      sym_rel_operator,
    ACTIONS(2962), 1,
      anon_sym_SEMI,
    ACTIONS(2964), 1,
      sym_assign_operator,
    STATE(534), 1,
      sym_pow_operator,
    STATE(535), 1,
      sym_mult_operator,
    STATE(536), 1,
      sym_add_operator,
    STATE(537), 1,
      sym_and_operator,
    STATE(538), 1,
      sym_or_operator,
    STATE(1895), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2718), 4,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2716), 17,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [45637] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1896), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1705), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1707), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [45700] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1897), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1667), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1669), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [45763] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1898), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1725), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1727), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [45826] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1899), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1729), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1731), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [45889] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1900), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1821), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1823), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [45952] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1901), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1575), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1577), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [46015] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1902), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1595), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1597), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [46078] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1903), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1611), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1613), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [46141] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1904), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1771), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [46204] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1905), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1789), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1791), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [46267] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1906), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1671), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1673), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [46330] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1907), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1675), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1677), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [46393] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1908), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1679), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1681), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [46456] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1909), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1851), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1853), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [46519] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1910), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1717), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1719), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [46582] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1911), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1837), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1839), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [46645] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1912), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1531), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1533), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [46708] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1913), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1579), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1581), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [46771] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2924), 1,
      anon_sym_DOT,
    STATE(1914), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1583), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1585), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [46836] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1915), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1587), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1589), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [46899] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1916), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1801), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      aux_sym_directive_token1,
    ACTIONS(1803), 32,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
  [46962] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1917), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1599), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1601), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [47025] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1918), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1603), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1605), 35,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [47088] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2956), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2958), 1,
      sym_concat_operator,
    ACTIONS(2960), 1,
      sym_rel_operator,
    STATE(534), 1,
      sym_pow_operator,
    STATE(535), 1,
      sym_mult_operator,
    STATE(536), 1,
      sym_add_operator,
    STATE(537), 1,
      sym_and_operator,
    STATE(538), 1,
      sym_or_operator,
    STATE(1919), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 4,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2860), 20,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
  [47181] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1920), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1551), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1553), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [47243] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1921), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1803), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1801), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [47305] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1922), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2972), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2970), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [47367] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2974), 1,
      anon_sym_COMMA,
    ACTIONS(2976), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2978), 1,
      sym_concat_operator,
    ACTIONS(2980), 1,
      sym_rel_operator,
    ACTIONS(2982), 1,
      sym_assign_operator,
    STATE(578), 1,
      sym_pow_operator,
    STATE(579), 1,
      sym_mult_operator,
    STATE(580), 1,
      sym_add_operator,
    STATE(581), 1,
      sym_and_operator,
    STATE(582), 1,
      sym_or_operator,
    STATE(1923), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2866), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2864), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [47463] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1924), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1847), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1849), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [47525] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1925), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2986), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2984), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [47587] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1926), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2986), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2984), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [47649] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2992), 1,
      anon_sym_COMMA,
    ACTIONS(2994), 1,
      anon_sym_PIPE,
    ACTIONS(2996), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_COLON_COLON,
    STATE(1927), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2990), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2988), 23,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [47719] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1928), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1745), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1747), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [47781] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2992), 1,
      anon_sym_COMMA,
    ACTIONS(2998), 1,
      anon_sym_COLON_COLON,
    STATE(1929), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2986), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2984), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [47847] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2992), 1,
      anon_sym_COMMA,
    STATE(1930), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2986), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2984), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [47911] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1931), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3002), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3000), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [47973] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2992), 1,
      anon_sym_COMMA,
    ACTIONS(2994), 1,
      anon_sym_PIPE,
    ACTIONS(2998), 1,
      anon_sym_COLON_COLON,
    STATE(1932), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3002), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3000), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [48041] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2998), 1,
      anon_sym_COLON_COLON,
    STATE(1933), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3006), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3004), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [48105] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1934), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3010), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3008), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [48167] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1935), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1799), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1797), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [48229] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2974), 1,
      anon_sym_COMMA,
    ACTIONS(2976), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2978), 1,
      sym_concat_operator,
    ACTIONS(2980), 1,
      sym_rel_operator,
    STATE(578), 1,
      sym_pow_operator,
    STATE(579), 1,
      sym_mult_operator,
    STATE(580), 1,
      sym_add_operator,
    STATE(581), 1,
      sym_and_operator,
    STATE(582), 1,
      sym_or_operator,
    STATE(1936), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2880), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2878), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [48323] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1937), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1705), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1707), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48385] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1938), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1815), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1813), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [48447] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1939), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1555), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1557), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48509] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1940), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1571), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1573), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48571] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2974), 1,
      anon_sym_COMMA,
    ACTIONS(2976), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2978), 1,
      sym_concat_operator,
    ACTIONS(2980), 1,
      sym_rel_operator,
    ACTIONS(2982), 1,
      sym_assign_operator,
    ACTIONS(3012), 1,
      anon_sym_SEMI,
    STATE(578), 1,
      sym_pow_operator,
    STATE(579), 1,
      sym_mult_operator,
    STATE(580), 1,
      sym_add_operator,
    STATE(581), 1,
      sym_and_operator,
    STATE(582), 1,
      sym_or_operator,
    STATE(1941), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2710), 3,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2708), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [48669] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1942), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1721), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1723), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48731] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1943), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3016), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3014), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [48793] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1944), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1827), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48855] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1945), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3020), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3018), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [48917] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1946), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1547), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1549), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [48979] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1947), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1695), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1697), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49041] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1948), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1807), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1805), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [49103] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1949), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1683), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1685), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49165] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1950), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1687), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1689), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49227] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1951), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1709), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1711), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49289] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1952), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3024), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3022), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [49351] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1953), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1667), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1669), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49413] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1954), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3028), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3026), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [49475] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1955), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1591), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1593), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49537] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1956), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1543), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1545), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49599] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1957), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1813), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1815), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49661] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1958), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1619), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1621), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49723] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3030), 1,
      anon_sym_DOT,
    ACTIONS(3032), 1,
      anon_sym_POUND,
    ACTIONS(3034), 1,
      sym_hash_operator,
    STATE(1959), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2952), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2950), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49791] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1960), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1829), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1831), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49853] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1961), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1785), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1787), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49915] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1962), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1797), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1799), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [49977] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1963), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1805), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1807), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50039] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1964), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1827), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1825), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [50101] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1965), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1757), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1759), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50163] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1966), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1627), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1629), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50225] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1967), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1535), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1537), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50287] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1968), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1623), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1625), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50349] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1969), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3038), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3036), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [50411] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1970), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3042), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3040), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [50473] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1971), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1655), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1657), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50535] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1972), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1733), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1735), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50597] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1973), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1817), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1819), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50659] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1974), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1833), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1835), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50721] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1975), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3046), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3044), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [50783] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1703), 1,
      anon_sym_DOT,
    STATE(1976), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1699), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1701), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50847] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1977), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1713), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1715), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50909] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1978), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1809), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1811), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [50971] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_DOT,
    STATE(1979), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1699), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1701), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51035] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1980), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1843), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1845), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51097] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1981), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1759), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1757), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [51159] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1982), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1725), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1727), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51221] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1983), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1771), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51283] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1984), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1789), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1791), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51345] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1985), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1801), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1803), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51407] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1986), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1777), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1779), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51469] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1987), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1781), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1783), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51531] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1988), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1567), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1569), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51593] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1989), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1749), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1751), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51655] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1990), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1753), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1755), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51717] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1991), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1761), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1763), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51779] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1992), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1793), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1795), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [51841] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2896), 1,
      anon_sym_COMMA,
    ACTIONS(2900), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2902), 1,
      sym_concat_operator,
    ACTIONS(2904), 1,
      sym_rel_operator,
    ACTIONS(2906), 1,
      sym_assign_operator,
    ACTIONS(3048), 1,
      anon_sym_SEMI,
    STATE(620), 1,
      sym_pow_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(713), 1,
      sym_mult_operator,
    STATE(1993), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2710), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2708), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [51939] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1994), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3052), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3050), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [52001] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1995), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3056), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3054), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [52063] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1996), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3060), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3058), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [52125] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1997), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3064), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3062), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [52187] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1998), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1845), 23,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1843), 24,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [52249] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(1999), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1729), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1731), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52311] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2998), 1,
      anon_sym_COLON_COLON,
    STATE(2000), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3068), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3066), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [52375] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2001), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3072), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3070), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [52437] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2974), 1,
      anon_sym_COMMA,
    ACTIONS(2976), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2978), 1,
      sym_concat_operator,
    ACTIONS(2980), 1,
      sym_rel_operator,
    ACTIONS(2982), 1,
      sym_assign_operator,
    ACTIONS(3012), 1,
      anon_sym_SEMI,
    STATE(578), 1,
      sym_pow_operator,
    STATE(579), 1,
      sym_mult_operator,
    STATE(580), 1,
      sym_add_operator,
    STATE(581), 1,
      sym_and_operator,
    STATE(582), 1,
      sym_or_operator,
    STATE(2002), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2854), 3,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2852), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [52535] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2003), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1821), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1823), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52597] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2974), 1,
      anon_sym_COMMA,
    ACTIONS(2976), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2978), 1,
      sym_concat_operator,
    ACTIONS(2980), 1,
      sym_rel_operator,
    ACTIONS(2982), 1,
      sym_assign_operator,
    STATE(578), 1,
      sym_pow_operator,
    STATE(579), 1,
      sym_mult_operator,
    STATE(580), 1,
      sym_add_operator,
    STATE(581), 1,
      sym_and_operator,
    STATE(582), 1,
      sym_or_operator,
    STATE(2004), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2890), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2888), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [52693] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2005), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1575), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1577), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52755] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2006), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1595), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1597), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [52817] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2976), 1,
      anon_sym_COLON_COLON,
    STATE(578), 1,
      sym_pow_operator,
    STATE(579), 1,
      sym_mult_operator,
    STATE(580), 1,
      sym_add_operator,
    STATE(581), 1,
      sym_and_operator,
    STATE(582), 1,
      sym_or_operator,
    STATE(2007), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2884), 7,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [52899] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2008), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1715), 23,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1713), 24,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [52961] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2992), 1,
      anon_sym_COMMA,
    ACTIONS(2998), 1,
      anon_sym_COLON_COLON,
    STATE(2009), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3076), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3074), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [53027] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2976), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2978), 1,
      sym_concat_operator,
    STATE(578), 1,
      sym_pow_operator,
    STATE(579), 1,
      sym_mult_operator,
    STATE(580), 1,
      sym_add_operator,
    STATE(581), 1,
      sym_and_operator,
    STATE(582), 1,
      sym_or_operator,
    STATE(2010), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 7,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [53111] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2011), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1611), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1613), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [53173] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2012), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1787), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1785), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [53235] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2013), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3080), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3078), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [53297] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2976), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2978), 1,
      sym_concat_operator,
    STATE(578), 1,
      sym_pow_operator,
    STATE(579), 1,
      sym_mult_operator,
    STATE(580), 1,
      sym_add_operator,
    STATE(581), 1,
      sym_and_operator,
    STATE(582), 1,
      sym_or_operator,
    STATE(2014), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 7,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [53381] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2974), 1,
      anon_sym_COMMA,
    ACTIONS(2976), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2978), 1,
      sym_concat_operator,
    ACTIONS(2980), 1,
      sym_rel_operator,
    ACTIONS(2982), 1,
      sym_assign_operator,
    STATE(578), 1,
      sym_pow_operator,
    STATE(579), 1,
      sym_mult_operator,
    STATE(580), 1,
      sym_add_operator,
    STATE(581), 1,
      sym_and_operator,
    STATE(582), 1,
      sym_or_operator,
    STATE(2015), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2860), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [53477] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(578), 1,
      sym_pow_operator,
    STATE(579), 1,
      sym_mult_operator,
    STATE(580), 1,
      sym_add_operator,
    STATE(581), 1,
      sym_and_operator,
    STATE(582), 1,
      sym_or_operator,
    STATE(2016), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 26,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [53551] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(578), 1,
      sym_pow_operator,
    STATE(579), 1,
      sym_mult_operator,
    STATE(580), 1,
      sym_add_operator,
    STATE(581), 1,
      sym_and_operator,
    STATE(582), 1,
      sym_or_operator,
    STATE(2017), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 26,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [53625] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(578), 1,
      sym_pow_operator,
    STATE(579), 1,
      sym_mult_operator,
    STATE(580), 1,
      sym_add_operator,
    STATE(581), 1,
      sym_and_operator,
    STATE(582), 1,
      sym_or_operator,
    STATE(2018), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 10,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [53703] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2976), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2978), 1,
      sym_concat_operator,
    ACTIONS(2980), 1,
      sym_rel_operator,
    STATE(578), 1,
      sym_pow_operator,
    STATE(579), 1,
      sym_mult_operator,
    STATE(580), 1,
      sym_add_operator,
    STATE(581), 1,
      sym_and_operator,
    STATE(582), 1,
      sym_or_operator,
    STATE(2019), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 5,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [53791] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2020), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1671), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1673), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [53853] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2021), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1675), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1677), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [53915] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2896), 1,
      anon_sym_COMMA,
    ACTIONS(2900), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2902), 1,
      sym_concat_operator,
    ACTIONS(2904), 1,
      sym_rel_operator,
    ACTIONS(2906), 1,
      sym_assign_operator,
    ACTIONS(3048), 1,
      anon_sym_SEMI,
    STATE(620), 1,
      sym_pow_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(713), 1,
      sym_mult_operator,
    STATE(2022), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2854), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2852), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [54013] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2023), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1679), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1681), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [54075] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2992), 1,
      anon_sym_COMMA,
    ACTIONS(2994), 1,
      anon_sym_PIPE,
    ACTIONS(2996), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_COLON_COLON,
    STATE(2024), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3084), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3082), 23,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [54145] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2025), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1625), 23,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1623), 24,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [54207] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2026), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1727), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1725), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [54269] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2992), 1,
      anon_sym_COMMA,
    STATE(2027), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3076), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3074), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [54333] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2896), 1,
      anon_sym_COMMA,
    ACTIONS(2900), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2902), 1,
      sym_concat_operator,
    ACTIONS(2904), 1,
      sym_rel_operator,
    ACTIONS(2906), 1,
      sym_assign_operator,
    ACTIONS(3048), 1,
      anon_sym_SEMI,
    STATE(620), 1,
      sym_pow_operator,
    STATE(621), 1,
      sym_add_operator,
    STATE(622), 1,
      sym_and_operator,
    STATE(623), 1,
      sym_or_operator,
    STATE(713), 1,
      sym_mult_operator,
    STATE(2028), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2718), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2716), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [54431] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3030), 1,
      anon_sym_DOT,
    ACTIONS(3032), 1,
      anon_sym_POUND,
    ACTIONS(3034), 1,
      sym_hash_operator,
    STATE(2029), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2912), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2910), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [54499] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3030), 1,
      anon_sym_DOT,
    ACTIONS(3032), 1,
      anon_sym_POUND,
    ACTIONS(3034), 1,
      sym_hash_operator,
    STATE(2030), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2922), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2920), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [54567] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2031), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1851), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1853), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [54629] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2032), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1717), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1719), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [54691] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2033), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3076), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3074), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [54753] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2034), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1837), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1839), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [54815] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2035), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1849), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1847), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [54877] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2036), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3088), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3086), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [54939] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2037), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3092), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3090), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55001] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2038), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1747), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1745), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55063] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2039), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1531), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1533), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [55125] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2040), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1579), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1581), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [55187] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2041), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1741), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1743), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [55249] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2992), 1,
      anon_sym_COMMA,
    ACTIONS(2994), 1,
      anon_sym_PIPE,
    ACTIONS(2998), 1,
      anon_sym_COLON_COLON,
    STATE(2042), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3096), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3094), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55317] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2043), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3100), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3098), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55379] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3030), 1,
      anon_sym_DOT,
    STATE(2044), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1583), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1585), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [55443] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2976), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2978), 1,
      sym_concat_operator,
    ACTIONS(2980), 1,
      sym_rel_operator,
    STATE(578), 1,
      sym_pow_operator,
    STATE(579), 1,
      sym_mult_operator,
    STATE(580), 1,
      sym_add_operator,
    STATE(581), 1,
      sym_and_operator,
    STATE(582), 1,
      sym_or_operator,
    STATE(2045), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2860), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [55535] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2046), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1587), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1589), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [55597] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2974), 1,
      anon_sym_COMMA,
    ACTIONS(2976), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2978), 1,
      sym_concat_operator,
    ACTIONS(2980), 1,
      sym_rel_operator,
    ACTIONS(2982), 1,
      sym_assign_operator,
    STATE(578), 1,
      sym_pow_operator,
    STATE(579), 1,
      sym_mult_operator,
    STATE(580), 1,
      sym_add_operator,
    STATE(581), 1,
      sym_and_operator,
    STATE(582), 1,
      sym_or_operator,
    STATE(2047), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2826), 4,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2824), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [55693] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2048), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3104), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3102), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55755] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2049), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3076), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3074), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55817] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2050), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1599), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1601), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [55879] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2051), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1295), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1297), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [55941] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3030), 1,
      anon_sym_DOT,
    ACTIONS(3032), 1,
      anon_sym_POUND,
    ACTIONS(3034), 1,
      sym_hash_operator,
    STATE(2052), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2968), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2966), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [56009] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2053), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1603), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1605), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [56071] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2054), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1607), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1609), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [56133] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2055), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1659), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1661), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [56195] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2056), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3108), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3106), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [56257] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2057), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3112), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3110), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [56319] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2058), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1765), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1767), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [56381] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2059), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3116), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3114), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [56443] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2060), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3120), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3118), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [56505] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2061), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3124), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3122), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [56567] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2062), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3128), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3126), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [56629] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2063), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3132), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3130), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [56691] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2064), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3136), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3134), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [56753] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2065), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1723), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1721), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [56815] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2066), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1773), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1775), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [56877] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2067), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1631), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1633), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [56939] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2068), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3140), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3138), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57001] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2069), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3144), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3142), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57063] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2070), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1635), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1637), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [57125] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2071), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3148), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3146), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57187] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2072), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3152), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3150), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57249] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2073), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1811), 23,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1809), 24,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [57311] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(578), 1,
      sym_pow_operator,
    STATE(579), 1,
      sym_mult_operator,
    STATE(580), 1,
      sym_add_operator,
    STATE(581), 1,
      sym_and_operator,
    STATE(582), 1,
      sym_or_operator,
    STATE(2074), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2874), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2872), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [57383] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2075), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1639), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1641), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [57445] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2076), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3156), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3154), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57507] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2077), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1643), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1645), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [57569] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2078), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1771), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1769), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57631] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2079), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3096), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3094), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57693] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2080), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1791), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1789), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [57755] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2081), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1559), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1561), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [57817] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2082), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1647), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1649), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [57879] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2083), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1691), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1693), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [57941] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2084), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1563), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1565), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [58003] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2085), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3160), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3158), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [58065] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2086), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3164), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3162), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [58127] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2087), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3168), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3166), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [58189] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2974), 1,
      anon_sym_COMMA,
    ACTIONS(2976), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2978), 1,
      sym_concat_operator,
    ACTIONS(2980), 1,
      sym_rel_operator,
    ACTIONS(2982), 1,
      sym_assign_operator,
    ACTIONS(3012), 1,
      anon_sym_SEMI,
    STATE(578), 1,
      sym_pow_operator,
    STATE(579), 1,
      sym_mult_operator,
    STATE(580), 1,
      sym_add_operator,
    STATE(581), 1,
      sym_and_operator,
    STATE(582), 1,
      sym_or_operator,
    STATE(2088), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2718), 3,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2716), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [58287] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2089), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3172), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3170), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [58349] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2090), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1737), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1739), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [58411] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2091), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1651), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1653), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [58473] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2092), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1539), 13,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1541), 34,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [58535] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3030), 1,
      anon_sym_DOT,
    ACTIONS(3032), 1,
      anon_sym_POUND,
    ACTIONS(3034), 1,
      sym_hash_operator,
    STATE(2093), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2948), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2946), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [58603] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2094), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3176), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3174), 28,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [58664] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3182), 1,
      anon_sym_PIPE,
    STATE(2095), 1,
      sym_attribute,
    STATE(2113), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3180), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3178), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [58729] = 33,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3184), 1,
      anon_sym_LPAREN,
    ACTIONS(3186), 1,
      anon_sym_external,
    ACTIONS(3188), 1,
      anon_sym_private,
    ACTIONS(3190), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3192), 1,
      anon_sym_PIPE,
    ACTIONS(3194), 1,
      anon_sym_LBRACK,
    ACTIONS(3198), 1,
      anon_sym_LBRACE,
    ACTIONS(3200), 1,
      sym__uppercase_identifier,
    STATE(2096), 1,
      sym_attribute,
    STATE(4185), 1,
      sym__simple_type,
    STATE(4535), 1,
      sym__type,
    STATE(4610), 1,
      sym__proper_tuple_type,
    STATE(5004), 1,
      sym_constructor_declaration,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6481), 1,
      sym_labeled_tuple_element_type,
    STATE(7168), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3196), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4628), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4300), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(4633), 3,
      sym_variant_declaration,
      sym_record_declaration,
      sym_external_declaration,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [58846] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2097), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3204), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3202), 29,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [58907] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3182), 1,
      anon_sym_PIPE,
    STATE(2098), 1,
      sym_attribute,
    STATE(2099), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3208), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3206), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [58972] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3214), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2099), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3212), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3210), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [59035] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3217), 1,
      anon_sym_COLON,
    STATE(2100), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3221), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3219), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [59098] = 33,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3223), 1,
      anon_sym_LPAREN,
    ACTIONS(3225), 1,
      anon_sym_external,
    ACTIONS(3227), 1,
      anon_sym_private,
    ACTIONS(3229), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3231), 1,
      anon_sym_PIPE,
    ACTIONS(3233), 1,
      anon_sym_LBRACK,
    ACTIONS(3237), 1,
      anon_sym_LBRACE,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3243), 1,
      sym__uppercase_identifier,
    STATE(2101), 1,
      sym_attribute,
    STATE(3209), 1,
      sym__simple_type,
    STATE(3505), 1,
      sym__type,
    STATE(3557), 1,
      sym__proper_tuple_type,
    STATE(3870), 1,
      sym_constructor_declaration,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6883), 1,
      sym_labeled_argument_type,
    STATE(6885), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3235), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3600), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3667), 3,
      sym_variant_declaration,
      sym_record_declaration,
      sym_external_declaration,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [59215] = 33,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3184), 1,
      anon_sym_LPAREN,
    ACTIONS(3186), 1,
      anon_sym_external,
    ACTIONS(3192), 1,
      anon_sym_PIPE,
    ACTIONS(3194), 1,
      anon_sym_LBRACK,
    ACTIONS(3198), 1,
      anon_sym_LBRACE,
    ACTIONS(3200), 1,
      sym__uppercase_identifier,
    ACTIONS(3245), 1,
      anon_sym_private,
    ACTIONS(3247), 1,
      anon_sym_DOT_DOT,
    STATE(2102), 1,
      sym_attribute,
    STATE(4185), 1,
      sym__simple_type,
    STATE(4517), 1,
      sym__type,
    STATE(4610), 1,
      sym__proper_tuple_type,
    STATE(5004), 1,
      sym_constructor_declaration,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6481), 1,
      sym_labeled_tuple_element_type,
    STATE(7168), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3196), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4628), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4300), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(4650), 3,
      sym_variant_declaration,
      sym_record_declaration,
      sym_external_declaration,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [59332] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2103), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3251), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3249), 29,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [59393] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2104), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3255), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3253), 29,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [59454] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3182), 1,
      anon_sym_PIPE,
    STATE(2098), 1,
      aux_sym__match_cases_repeat1,
    STATE(2105), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3259), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3257), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [59519] = 33,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3223), 1,
      anon_sym_LPAREN,
    ACTIONS(3225), 1,
      anon_sym_external,
    ACTIONS(3231), 1,
      anon_sym_PIPE,
    ACTIONS(3233), 1,
      anon_sym_LBRACK,
    ACTIONS(3237), 1,
      anon_sym_LBRACE,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3243), 1,
      sym__uppercase_identifier,
    ACTIONS(3261), 1,
      anon_sym_private,
    ACTIONS(3263), 1,
      anon_sym_DOT_DOT,
    STATE(2106), 1,
      sym_attribute,
    STATE(3209), 1,
      sym__simple_type,
    STATE(3528), 1,
      sym__type,
    STATE(3557), 1,
      sym__proper_tuple_type,
    STATE(3870), 1,
      sym_constructor_declaration,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6883), 1,
      sym_labeled_argument_type,
    STATE(6885), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3235), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3600), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3662), 3,
      sym_variant_declaration,
      sym_record_declaration,
      sym_external_declaration,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [59636] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2107), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3267), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3265), 29,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [59697] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3273), 1,
      anon_sym_else,
    STATE(2108), 1,
      sym_attribute,
    STATE(2126), 1,
      sym_else_clause,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3271), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3269), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [59762] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3273), 1,
      anon_sym_else,
    STATE(2109), 1,
      sym_attribute,
    STATE(2119), 1,
      sym_else_clause,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3277), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3275), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [59827] = 33,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3225), 1,
      anon_sym_external,
    ACTIONS(3237), 1,
      anon_sym_LBRACE,
    ACTIONS(3243), 1,
      sym__uppercase_identifier,
    ACTIONS(3279), 1,
      sym__lowercase_identifier,
    ACTIONS(3281), 1,
      anon_sym_LPAREN,
    ACTIONS(3283), 1,
      anon_sym_private,
    ACTIONS(3285), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3287), 1,
      anon_sym_PIPE,
    ACTIONS(3289), 1,
      anon_sym_LBRACK,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    STATE(2110), 1,
      sym_attribute,
    STATE(2917), 1,
      sym__simple_type,
    STATE(3290), 1,
      sym__type,
    STATE(3350), 1,
      sym__proper_tuple_type,
    STATE(3549), 1,
      sym_constructor_declaration,
    STATE(6244), 1,
      sym_extended_module_path,
    STATE(7252), 1,
      sym_labeled_argument_type,
    STATE(7344), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3235), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3440), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3474), 3,
      sym_variant_declaration,
      sym_record_declaration,
      sym_external_declaration,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [59944] = 33,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3225), 1,
      anon_sym_external,
    ACTIONS(3237), 1,
      anon_sym_LBRACE,
    ACTIONS(3243), 1,
      sym__uppercase_identifier,
    ACTIONS(3279), 1,
      sym__lowercase_identifier,
    ACTIONS(3281), 1,
      anon_sym_LPAREN,
    ACTIONS(3287), 1,
      anon_sym_PIPE,
    ACTIONS(3289), 1,
      anon_sym_LBRACK,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3305), 1,
      anon_sym_private,
    ACTIONS(3307), 1,
      anon_sym_DOT_DOT,
    STATE(2111), 1,
      sym_attribute,
    STATE(2917), 1,
      sym__simple_type,
    STATE(3289), 1,
      sym__type,
    STATE(3350), 1,
      sym__proper_tuple_type,
    STATE(3549), 1,
      sym_constructor_declaration,
    STATE(6244), 1,
      sym_extended_module_path,
    STATE(7252), 1,
      sym_labeled_argument_type,
    STATE(7344), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3235), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3440), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3460), 3,
      sym_variant_declaration,
      sym_record_declaration,
      sym_external_declaration,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [60061] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3309), 1,
      anon_sym_COLON,
    STATE(2112), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3221), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3219), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [60124] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3182), 1,
      anon_sym_PIPE,
    STATE(2099), 1,
      aux_sym__match_cases_repeat1,
    STATE(2113), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3259), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3257), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [60189] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2114), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3313), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3311), 29,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [60250] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2115), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3317), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3315), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [60310] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2116), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3321), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3319), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [60370] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2117), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3325), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3323), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [60430] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3327), 1,
      anon_sym_PIPE,
    STATE(2118), 1,
      sym_attribute,
    STATE(2173), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3259), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3257), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [60494] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2119), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3325), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3323), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [60554] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2974), 1,
      anon_sym_COMMA,
    ACTIONS(2976), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2978), 1,
      sym_concat_operator,
    ACTIONS(2980), 1,
      sym_rel_operator,
    ACTIONS(2982), 1,
      sym_assign_operator,
    ACTIONS(3329), 1,
      anon_sym_SEMI,
    STATE(578), 1,
      sym_pow_operator,
    STATE(579), 1,
      sym_mult_operator,
    STATE(580), 1,
      sym_add_operator,
    STATE(581), 1,
      sym_and_operator,
    STATE(582), 1,
      sym_or_operator,
    STATE(2120), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2718), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2716), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [60650] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2121), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3333), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3331), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [60710] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2122), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3337), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3335), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [60770] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2123), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3341), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3339), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [60830] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2124), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3345), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3343), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [60890] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2125), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2880), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2878), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [60950] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2126), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3349), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3347), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [61010] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2127), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2880), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2878), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [61070] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2128), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3353), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3351), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61130] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2129), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3357), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3355), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [61190] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2130), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3361), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3359), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61250] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2131), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3365), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3363), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [61310] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2132), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3369), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3367), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [61370] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2133), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(749), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(747), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [61430] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2134), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3361), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3359), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [61490] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2135), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3357), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3355), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61550] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2136), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3349), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3347), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61610] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      anon_sym_COMMA,
    STATE(2137), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2880), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2878), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [61672] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2138), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3373), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3371), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61732] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2139), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3377), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3375), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61792] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2140), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3381), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3379), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [61852] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2141), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3385), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3383), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [61912] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2142), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3389), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3387), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [61972] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2143), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3393), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3391), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [62032] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2144), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3397), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3395), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62092] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2145), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3337), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3335), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62152] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2974), 1,
      anon_sym_COMMA,
    ACTIONS(2976), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2978), 1,
      sym_concat_operator,
    ACTIONS(2980), 1,
      sym_rel_operator,
    ACTIONS(2982), 1,
      sym_assign_operator,
    ACTIONS(3329), 1,
      anon_sym_SEMI,
    STATE(578), 1,
      sym_pow_operator,
    STATE(579), 1,
      sym_mult_operator,
    STATE(580), 1,
      sym_add_operator,
    STATE(581), 1,
      sym_and_operator,
    STATE(582), 1,
      sym_or_operator,
    STATE(2146), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2710), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2708), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [62248] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2147), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3401), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3399), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62308] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2148), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3405), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3403), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [62368] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2149), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3409), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3407), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [62428] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2150), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3212), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3210), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [62488] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2151), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(749), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(747), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62548] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2152), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3317), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3315), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [62608] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2153), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3333), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3331), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62668] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2154), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3381), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3379), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62728] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2155), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3413), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3411), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [62788] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2156), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3373), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3371), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [62848] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2157), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3405), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3403), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [62908] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3415), 1,
      anon_sym_else,
    STATE(2119), 1,
      sym_else_clause,
    STATE(2158), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3277), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3275), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [62972] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2159), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3377), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3375), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [63032] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2160), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3369), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3367), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63092] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3415), 1,
      anon_sym_else,
    STATE(2126), 1,
      sym_else_clause,
    STATE(2161), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3271), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3269), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [63156] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2162), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3365), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3363), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63216] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2163), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3345), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3343), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63276] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3327), 1,
      anon_sym_PIPE,
    STATE(2164), 1,
      sym_attribute,
    STATE(2174), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3259), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3257), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [63340] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2165), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3401), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3399), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [63400] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3327), 1,
      anon_sym_PIPE,
    STATE(2164), 1,
      aux_sym__match_cases_repeat1,
    STATE(2166), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3180), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3178), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [63464] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2167), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3419), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3417), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [63524] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2168), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3397), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3395), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [63584] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2169), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3393), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3391), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63644] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2170), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3389), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3387), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [63704] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2171), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3423), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3421), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63764] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2172), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3385), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3383), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [63824] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3327), 1,
      anon_sym_PIPE,
    STATE(2173), 1,
      sym_attribute,
    STATE(2174), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3208), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3206), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [63888] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3425), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2174), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3212), 12,
      anon_sym_let,
      anon_sym_and,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3210), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [63950] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2175), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3423), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3421), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [64010] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3428), 1,
      anon_sym_COLON,
    STATE(2176), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3221), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3219), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64072] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2177), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3409), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3407), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64132] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2974), 1,
      anon_sym_COMMA,
    ACTIONS(2976), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2978), 1,
      sym_concat_operator,
    ACTIONS(2980), 1,
      sym_rel_operator,
    ACTIONS(2982), 1,
      sym_assign_operator,
    ACTIONS(3329), 1,
      anon_sym_SEMI,
    STATE(578), 1,
      sym_pow_operator,
    STATE(579), 1,
      sym_mult_operator,
    STATE(580), 1,
      sym_add_operator,
    STATE(581), 1,
      sym_and_operator,
    STATE(582), 1,
      sym_or_operator,
    STATE(2178), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2838), 2,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
    ACTIONS(2854), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2840), 3,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2852), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [64228] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2179), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3353), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3351), 31,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [64288] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2180), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3341), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3339), 33,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64348] = 27,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1293), 1,
      sym__lowercase_identifier,
    ACTIONS(1303), 1,
      anon_sym_LBRACK,
    ACTIONS(1307), 1,
      anon_sym_LBRACE,
    ACTIONS(1311), 1,
      anon_sym_POUND,
    ACTIONS(1313), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1317), 1,
      anon_sym_begin,
    ACTIONS(1325), 1,
      aux_sym_number_token1,
    ACTIONS(1327), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1329), 1,
      anon_sym_SQUOTE,
    ACTIONS(1331), 1,
      anon_sym_DQUOTE,
    ACTIONS(1333), 1,
      sym__uppercase_identifier,
    ACTIONS(1335), 1,
      aux_sym_tag_token1,
    ACTIONS(1347), 1,
      anon_sym_LPAREN,
    ACTIONS(2408), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2410), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1698), 1,
      sym__constructor_name,
    STATE(1975), 1,
      sym__simple_binding_pattern,
    STATE(2035), 1,
      sym_parenthesized_operator,
    STATE(2181), 1,
      sym_attribute,
    STATE(6855), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1305), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1823), 2,
      sym__constant,
      sym_signed_number,
    STATE(1803), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2087), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2051), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [64451] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2182), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1819), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1817), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [64510] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2896), 1,
      anon_sym_COMMA,
    STATE(2183), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2880), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2878), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [64571] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2184), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3337), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3335), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64630] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3430), 1,
      anon_sym_PIPE,
    STATE(2185), 1,
      sym_attribute,
    STATE(2282), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3180), 11,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3178), 31,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64693] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2186), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1827), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1825), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [64752] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2187), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1831), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1829), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [64811] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2188), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1835), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1833), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [64870] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2189), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3369), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3367), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [64929] = 32,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3237), 1,
      anon_sym_LBRACE,
    ACTIONS(3243), 1,
      sym__uppercase_identifier,
    ACTIONS(3279), 1,
      sym__lowercase_identifier,
    ACTIONS(3281), 1,
      anon_sym_LPAREN,
    ACTIONS(3283), 1,
      anon_sym_private,
    ACTIONS(3285), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3287), 1,
      anon_sym_PIPE,
    ACTIONS(3289), 1,
      anon_sym_LBRACK,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    STATE(2190), 1,
      sym_attribute,
    STATE(2917), 1,
      sym__simple_type,
    STATE(3290), 1,
      sym__type,
    STATE(3350), 1,
      sym__proper_tuple_type,
    STATE(3549), 1,
      sym_constructor_declaration,
    STATE(6244), 1,
      sym_extended_module_path,
    STATE(7252), 1,
      sym_labeled_argument_type,
    STATE(7344), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3235), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3440), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3474), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [65042] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2191), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3434), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3432), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [65101] = 32,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3237), 1,
      anon_sym_LBRACE,
    ACTIONS(3243), 1,
      sym__uppercase_identifier,
    ACTIONS(3279), 1,
      sym__lowercase_identifier,
    ACTIONS(3281), 1,
      anon_sym_LPAREN,
    ACTIONS(3287), 1,
      anon_sym_PIPE,
    ACTIONS(3289), 1,
      anon_sym_LBRACK,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3305), 1,
      anon_sym_private,
    ACTIONS(3307), 1,
      anon_sym_DOT_DOT,
    STATE(2192), 1,
      sym_attribute,
    STATE(2917), 1,
      sym__simple_type,
    STATE(3289), 1,
      sym__type,
    STATE(3350), 1,
      sym__proper_tuple_type,
    STATE(3549), 1,
      sym_constructor_declaration,
    STATE(6244), 1,
      sym_extended_module_path,
    STATE(7252), 1,
      sym_labeled_argument_type,
    STATE(7344), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3235), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3440), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3460), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [65214] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2193), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3108), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3106), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [65273] = 27,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1409), 1,
      sym__lowercase_identifier,
    ACTIONS(1413), 1,
      anon_sym_LPAREN,
    ACTIONS(1419), 1,
      anon_sym_LBRACK,
    ACTIONS(1423), 1,
      anon_sym_LBRACE,
    ACTIONS(1433), 1,
      anon_sym_POUND,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1439), 1,
      anon_sym_begin,
    ACTIONS(1453), 1,
      aux_sym_number_token1,
    ACTIONS(1455), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1457), 1,
      anon_sym_SQUOTE,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(1461), 1,
      sym__uppercase_identifier,
    ACTIONS(1463), 1,
      aux_sym_tag_token1,
    ACTIONS(1963), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1965), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2194), 1,
      sym_attribute,
    STATE(2768), 1,
      sym__constructor_name,
    STATE(2987), 1,
      sym__simple_binding_pattern,
    STATE(3035), 1,
      sym_parenthesized_operator,
    STATE(6493), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1421), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2940), 2,
      sym__constant,
      sym_signed_number,
    STATE(2929), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3051), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3023), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [65376] = 27,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1047), 1,
      anon_sym_LBRACK,
    ACTIONS(1051), 1,
      anon_sym_LBRACE,
    ACTIONS(1061), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1065), 1,
      anon_sym_begin,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2180), 1,
      sym__lowercase_identifier,
    ACTIONS(2186), 1,
      anon_sym_POUND,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2196), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3436), 1,
      anon_sym_LPAREN,
    ACTIONS(3438), 1,
      sym__uppercase_identifier,
    STATE(2195), 1,
      sym_attribute,
    STATE(4454), 1,
      sym__constructor_name,
    STATE(4580), 1,
      sym__simple_binding_pattern,
    STATE(4808), 1,
      sym_parenthesized_operator,
    STATE(6643), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4254), 2,
      sym__constant,
      sym_signed_number,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4564), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4548), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [65479] = 27,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1409), 1,
      sym__lowercase_identifier,
    ACTIONS(1413), 1,
      anon_sym_LPAREN,
    ACTIONS(1419), 1,
      anon_sym_LBRACK,
    ACTIONS(1423), 1,
      anon_sym_LBRACE,
    ACTIONS(1433), 1,
      anon_sym_POUND,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1439), 1,
      anon_sym_begin,
    ACTIONS(1453), 1,
      aux_sym_number_token1,
    ACTIONS(1455), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1457), 1,
      anon_sym_SQUOTE,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(1461), 1,
      sym__uppercase_identifier,
    ACTIONS(1463), 1,
      aux_sym_tag_token1,
    ACTIONS(1963), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(1965), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2196), 1,
      sym_attribute,
    STATE(2768), 1,
      sym__constructor_name,
    STATE(3002), 1,
      sym__simple_binding_pattern,
    STATE(3035), 1,
      sym_parenthesized_operator,
    STATE(6493), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1421), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2940), 2,
      sym__constant,
      sym_signed_number,
    STATE(2929), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(3051), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(3023), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [65582] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2197), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3385), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3383), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [65641] = 27,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1293), 1,
      sym__lowercase_identifier,
    ACTIONS(1303), 1,
      anon_sym_LBRACK,
    ACTIONS(1307), 1,
      anon_sym_LBRACE,
    ACTIONS(1311), 1,
      anon_sym_POUND,
    ACTIONS(1313), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1317), 1,
      anon_sym_begin,
    ACTIONS(1325), 1,
      aux_sym_number_token1,
    ACTIONS(1327), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1329), 1,
      anon_sym_SQUOTE,
    ACTIONS(1331), 1,
      anon_sym_DQUOTE,
    ACTIONS(1333), 1,
      sym__uppercase_identifier,
    ACTIONS(1335), 1,
      aux_sym_tag_token1,
    ACTIONS(1347), 1,
      anon_sym_LPAREN,
    ACTIONS(2408), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2410), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(1698), 1,
      sym__constructor_name,
    STATE(2035), 1,
      sym_parenthesized_operator,
    STATE(2076), 1,
      sym__simple_binding_pattern,
    STATE(2198), 1,
      sym_attribute,
    STATE(6855), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1305), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1823), 2,
      sym__constant,
      sym_signed_number,
    STATE(1803), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2087), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(2051), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [65744] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2199), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3442), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3440), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [65803] = 26,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2208), 1,
      sym__lowercase_identifier,
    ACTIONS(2212), 1,
      anon_sym_LBRACK,
    ACTIONS(2216), 1,
      anon_sym_LBRACE,
    ACTIONS(2220), 1,
      anon_sym_POUND,
    ACTIONS(2222), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2226), 1,
      anon_sym_begin,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2234), 1,
      aux_sym_number_token1,
    ACTIONS(2236), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2238), 1,
      anon_sym_SQUOTE,
    ACTIONS(2240), 1,
      anon_sym_DQUOTE,
    ACTIONS(3444), 1,
      anon_sym_LPAREN,
    ACTIONS(3446), 1,
      sym__uppercase_identifier,
    ACTIONS(3448), 1,
      aux_sym_tag_token1,
    STATE(2200), 1,
      sym_attribute,
    STATE(4897), 1,
      sym__simple_pattern,
    STATE(5025), 1,
      sym__constructor_name,
    STATE(6720), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2214), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4717), 2,
      sym__constant,
      sym_signed_number,
    STATE(4693), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4978), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(4970), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [65904] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3450), 1,
      anon_sym_else,
    STATE(2201), 1,
      sym_attribute,
    STATE(2270), 1,
      sym_else_clause,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3277), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3275), 30,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [65967] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2202), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2710), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2708), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [66026] = 26,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2208), 1,
      sym__lowercase_identifier,
    ACTIONS(2212), 1,
      anon_sym_LBRACK,
    ACTIONS(2216), 1,
      anon_sym_LBRACE,
    ACTIONS(2220), 1,
      anon_sym_POUND,
    ACTIONS(2222), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2226), 1,
      anon_sym_begin,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2234), 1,
      aux_sym_number_token1,
    ACTIONS(2236), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2238), 1,
      anon_sym_SQUOTE,
    ACTIONS(2240), 1,
      anon_sym_DQUOTE,
    ACTIONS(3444), 1,
      anon_sym_LPAREN,
    ACTIONS(3446), 1,
      sym__uppercase_identifier,
    ACTIONS(3448), 1,
      aux_sym_tag_token1,
    STATE(2203), 1,
      sym_attribute,
    STATE(4899), 1,
      sym__simple_pattern,
    STATE(5025), 1,
      sym__constructor_name,
    STATE(6720), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2214), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4717), 2,
      sym__constant,
      sym_signed_number,
    STATE(4693), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4978), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(4970), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [66127] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2204), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(749), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(747), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [66186] = 26,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1079), 1,
      sym__uppercase_identifier,
    ACTIONS(1081), 1,
      aux_sym_tag_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3452), 1,
      anon_sym_LPAREN,
    STATE(2008), 1,
      sym__constructor_name,
    STATE(2205), 1,
      sym_attribute,
    STATE(2238), 1,
      sym__simple_pattern,
    STATE(6390), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [66287] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2206), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3361), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3359), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [66346] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2207), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3456), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3454), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [66405] = 27,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2220), 1,
      anon_sym_POUND,
    ACTIONS(2226), 1,
      anon_sym_begin,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2234), 1,
      aux_sym_number_token1,
    ACTIONS(2236), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2238), 1,
      anon_sym_SQUOTE,
    ACTIONS(2240), 1,
      anon_sym_DQUOTE,
    ACTIONS(2246), 1,
      sym__lowercase_identifier,
    ACTIONS(2250), 1,
      anon_sym_LBRACK,
    ACTIONS(2252), 1,
      anon_sym_LBRACE,
    ACTIONS(2256), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3446), 1,
      sym__uppercase_identifier,
    ACTIONS(3448), 1,
      aux_sym_tag_token1,
    ACTIONS(3458), 1,
      anon_sym_LPAREN,
    STATE(2208), 1,
      sym_attribute,
    STATE(4863), 1,
      sym_parenthesized_operator,
    STATE(4960), 1,
      sym__simple_binding_pattern,
    STATE(5025), 1,
      sym__constructor_name,
    STATE(6682), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2214), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4717), 2,
      sym__constant,
      sym_signed_number,
    STATE(4693), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4922), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4914), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [66508] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2209), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2880), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2878), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [66567] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2210), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3038), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3036), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [66626] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2211), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3393), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3391), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [66685] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2212), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2932), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2930), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [66744] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2213), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3052), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3050), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [66803] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2214), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3357), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3355), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [66862] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2215), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1787), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1785), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [66921] = 32,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3223), 1,
      anon_sym_LPAREN,
    ACTIONS(3227), 1,
      anon_sym_private,
    ACTIONS(3229), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3231), 1,
      anon_sym_PIPE,
    ACTIONS(3233), 1,
      anon_sym_LBRACK,
    ACTIONS(3237), 1,
      anon_sym_LBRACE,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3243), 1,
      sym__uppercase_identifier,
    STATE(2216), 1,
      sym_attribute,
    STATE(3209), 1,
      sym__simple_type,
    STATE(3505), 1,
      sym__type,
    STATE(3557), 1,
      sym__proper_tuple_type,
    STATE(3870), 1,
      sym_constructor_declaration,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6883), 1,
      sym_labeled_argument_type,
    STATE(6885), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3235), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3600), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3667), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [67034] = 27,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2220), 1,
      anon_sym_POUND,
    ACTIONS(2226), 1,
      anon_sym_begin,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2234), 1,
      aux_sym_number_token1,
    ACTIONS(2236), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2238), 1,
      anon_sym_SQUOTE,
    ACTIONS(2240), 1,
      anon_sym_DQUOTE,
    ACTIONS(2246), 1,
      sym__lowercase_identifier,
    ACTIONS(2250), 1,
      anon_sym_LBRACK,
    ACTIONS(2252), 1,
      anon_sym_LBRACE,
    ACTIONS(2256), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(3446), 1,
      sym__uppercase_identifier,
    ACTIONS(3448), 1,
      aux_sym_tag_token1,
    ACTIONS(3458), 1,
      anon_sym_LPAREN,
    STATE(2217), 1,
      sym_attribute,
    STATE(4863), 1,
      sym_parenthesized_operator,
    STATE(4944), 1,
      sym__simple_binding_pattern,
    STATE(5025), 1,
      sym__constructor_name,
    STATE(6682), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2214), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4717), 2,
      sym__constant,
      sym_signed_number,
    STATE(4693), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4922), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4914), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [67137] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2218), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1799), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1797), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [67196] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2219), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2718), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2716), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [67255] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2220), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1807), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1805), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [67314] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2221), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1759), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1757), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [67373] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2222), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3333), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3331), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [67432] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2223), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2714), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2712), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [67491] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2224), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3462), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3460), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [67550] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2225), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3466), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3464), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [67609] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2226), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3353), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3351), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [67668] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2227), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3405), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3403), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [67727] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2228), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3341), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3339), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [67786] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2229), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3423), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3421), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [67845] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2230), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3470), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3468), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [67904] = 26,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1065), 1,
      anon_sym_begin,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1099), 1,
      sym__lowercase_identifier,
    ACTIONS(1105), 1,
      anon_sym_LBRACK,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1115), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2186), 1,
      anon_sym_POUND,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2196), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3438), 1,
      sym__uppercase_identifier,
    ACTIONS(3472), 1,
      anon_sym_LPAREN,
    STATE(2231), 1,
      sym_attribute,
    STATE(4435), 1,
      sym__simple_pattern,
    STATE(4454), 1,
      sym__constructor_name,
    STATE(6683), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4254), 2,
      sym__constant,
      sym_signed_number,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4414), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(4428), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [68005] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2232), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3476), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3474), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [68064] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2233), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3381), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3379), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [68123] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2234), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3480), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3478), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [68182] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2235), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3484), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3482), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [68241] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2236), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3080), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3078), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [68300] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3450), 1,
      anon_sym_else,
    STATE(2237), 1,
      sym_attribute,
    STATE(2262), 1,
      sym_else_clause,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3271), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3269), 30,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [68363] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2238), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3488), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3486), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [68422] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2239), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3492), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3490), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [68481] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2240), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3496), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3494), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [68540] = 27,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1035), 1,
      sym__lowercase_identifier,
    ACTIONS(1047), 1,
      anon_sym_LBRACK,
    ACTIONS(1051), 1,
      anon_sym_LBRACE,
    ACTIONS(1061), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1065), 1,
      anon_sym_begin,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2083), 1,
      anon_sym_POUND,
    ACTIONS(2196), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3498), 1,
      anon_sym_LPAREN,
    ACTIONS(3500), 1,
      sym__uppercase_identifier,
    STATE(2241), 1,
      sym_attribute,
    STATE(3151), 1,
      sym__constructor_name,
    STATE(3501), 1,
      sym_parenthesized_operator,
    STATE(4580), 1,
      sym__simple_binding_pattern,
    STATE(6600), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4254), 2,
      sym__constant,
      sym_signed_number,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4564), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4548), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [68643] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2242), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3504), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3502), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [68702] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2243), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3508), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3506), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [68761] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2244), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1815), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1813), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [68820] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2245), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3321), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3319), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [68879] = 26,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1065), 1,
      anon_sym_begin,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1099), 1,
      sym__lowercase_identifier,
    ACTIONS(1105), 1,
      anon_sym_LBRACK,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1115), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2083), 1,
      anon_sym_POUND,
    ACTIONS(2196), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3500), 1,
      sym__uppercase_identifier,
    ACTIONS(3510), 1,
      anon_sym_LPAREN,
    STATE(2246), 1,
      sym_attribute,
    STATE(3151), 1,
      sym__constructor_name,
    STATE(4435), 1,
      sym__simple_pattern,
    STATE(6644), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4254), 2,
      sym__constant,
      sym_signed_number,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4414), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(4428), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [68980] = 32,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3184), 1,
      anon_sym_LPAREN,
    ACTIONS(3188), 1,
      anon_sym_private,
    ACTIONS(3190), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3192), 1,
      anon_sym_PIPE,
    ACTIONS(3194), 1,
      anon_sym_LBRACK,
    ACTIONS(3198), 1,
      anon_sym_LBRACE,
    ACTIONS(3200), 1,
      sym__uppercase_identifier,
    STATE(2247), 1,
      sym_attribute,
    STATE(4185), 1,
      sym__simple_type,
    STATE(4535), 1,
      sym__type,
    STATE(4610), 1,
      sym__proper_tuple_type,
    STATE(5004), 1,
      sym_constructor_declaration,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6481), 1,
      sym_labeled_tuple_element_type,
    STATE(7168), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3196), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4628), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4633), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4300), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [69093] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2248), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3514), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3512), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [69152] = 26,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1973), 1,
      sym__lowercase_identifier,
    ACTIONS(1987), 1,
      anon_sym_LBRACK,
    ACTIONS(1991), 1,
      anon_sym_LBRACE,
    ACTIONS(1993), 1,
      anon_sym_POUND,
    ACTIONS(1995), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1997), 1,
      anon_sym_begin,
    ACTIONS(2005), 1,
      aux_sym_number_token1,
    ACTIONS(2007), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2009), 1,
      anon_sym_SQUOTE,
    ACTIONS(2011), 1,
      anon_sym_DQUOTE,
    ACTIONS(2013), 1,
      sym__uppercase_identifier,
    ACTIONS(2015), 1,
      aux_sym_tag_token1,
    ACTIONS(3516), 1,
      anon_sym_LPAREN,
    ACTIONS(3518), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3520), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2249), 1,
      sym_attribute,
    STATE(2469), 1,
      sym__simple_pattern,
    STATE(2476), 1,
      sym__constructor_name,
    STATE(6496), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1989), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2426), 2,
      sym__constant,
      sym_signed_number,
    STATE(2428), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2434), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2474), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [69253] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2250), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3524), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3522), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [69312] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2251), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3528), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3526), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [69371] = 27,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1047), 1,
      anon_sym_LBRACK,
    ACTIONS(1051), 1,
      anon_sym_LBRACE,
    ACTIONS(1061), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1065), 1,
      anon_sym_begin,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2083), 1,
      anon_sym_POUND,
    ACTIONS(2196), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(2676), 1,
      sym__lowercase_identifier,
    ACTIONS(3530), 1,
      anon_sym_LPAREN,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(2252), 1,
      sym_attribute,
    STATE(4580), 1,
      sym__simple_binding_pattern,
    STATE(6555), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4254), 2,
      sym__constant,
      sym_signed_number,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4564), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4548), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [69474] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2253), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3534), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3532), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [69533] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2254), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1727), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1725), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [69592] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2255), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3056), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3054), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [69651] = 32,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3184), 1,
      anon_sym_LPAREN,
    ACTIONS(3192), 1,
      anon_sym_PIPE,
    ACTIONS(3194), 1,
      anon_sym_LBRACK,
    ACTIONS(3198), 1,
      anon_sym_LBRACE,
    ACTIONS(3200), 1,
      sym__uppercase_identifier,
    ACTIONS(3245), 1,
      anon_sym_private,
    ACTIONS(3247), 1,
      anon_sym_DOT_DOT,
    STATE(2256), 1,
      sym_attribute,
    STATE(4185), 1,
      sym__simple_type,
    STATE(4517), 1,
      sym__type,
    STATE(4610), 1,
      sym__proper_tuple_type,
    STATE(5004), 1,
      sym_constructor_declaration,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6481), 1,
      sym_labeled_tuple_element_type,
    STATE(7168), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3196), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4628), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4650), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4300), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [69764] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2257), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1771), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1769), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [69823] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2258), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1791), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1789), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [69882] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2259), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2944), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2942), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [69941] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2260), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1685), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1683), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [70000] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2261), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3345), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3343), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [70059] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2262), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3349), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3347), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [70118] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2263), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1689), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1687), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [70177] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2264), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3538), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3536), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [70236] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2265), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3409), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3407), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [70295] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2266), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3401), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3399), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [70354] = 26,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1065), 1,
      anon_sym_begin,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1099), 1,
      sym__lowercase_identifier,
    ACTIONS(1105), 1,
      anon_sym_LBRACK,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1115), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2083), 1,
      anon_sym_POUND,
    ACTIONS(2196), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3540), 1,
      anon_sym_LPAREN,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(2267), 1,
      sym_attribute,
    STATE(4435), 1,
      sym__simple_pattern,
    STATE(6601), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4254), 2,
      sym__constant,
      sym_signed_number,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4414), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(4428), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [70455] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2268), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1803), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1801), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [70514] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2269), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3317), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3315), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [70573] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2270), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3325), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3323), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [70632] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2271), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3365), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3363), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [70691] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3430), 1,
      anon_sym_PIPE,
    STATE(2272), 1,
      sym_attribute,
    STATE(2276), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3208), 11,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3206), 31,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [70754] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2273), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3212), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3210), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [70813] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2274), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3397), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3395), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [70872] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2275), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3389), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3387), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [70931] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3542), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2276), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3212), 11,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3210), 31,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [70992] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2277), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3413), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3411), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [71051] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2278), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3419), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3417), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [71110] = 32,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3223), 1,
      anon_sym_LPAREN,
    ACTIONS(3231), 1,
      anon_sym_PIPE,
    ACTIONS(3233), 1,
      anon_sym_LBRACK,
    ACTIONS(3237), 1,
      anon_sym_LBRACE,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3243), 1,
      sym__uppercase_identifier,
    ACTIONS(3261), 1,
      anon_sym_private,
    ACTIONS(3263), 1,
      anon_sym_DOT_DOT,
    STATE(2279), 1,
      sym_attribute,
    STATE(3209), 1,
      sym__simple_type,
    STATE(3528), 1,
      sym__type,
    STATE(3557), 1,
      sym__proper_tuple_type,
    STATE(3870), 1,
      sym_constructor_declaration,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6883), 1,
      sym_labeled_argument_type,
    STATE(6885), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3235), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3600), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3662), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [71223] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2280), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1711), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1709), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [71282] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3430), 1,
      anon_sym_PIPE,
    STATE(2272), 1,
      aux_sym__match_cases_repeat1,
    STATE(2281), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3259), 11,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3257), 31,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [71345] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3430), 1,
      anon_sym_PIPE,
    STATE(2276), 1,
      aux_sym__match_cases_repeat1,
    STATE(2282), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3259), 11,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3257), 31,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [71408] = 27,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1047), 1,
      anon_sym_LBRACK,
    ACTIONS(1051), 1,
      anon_sym_LBRACE,
    ACTIONS(1061), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1065), 1,
      anon_sym_begin,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2180), 1,
      sym__lowercase_identifier,
    ACTIONS(2186), 1,
      anon_sym_POUND,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2196), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3436), 1,
      anon_sym_LPAREN,
    ACTIONS(3438), 1,
      sym__uppercase_identifier,
    STATE(2283), 1,
      sym_attribute,
    STATE(4454), 1,
      sym__constructor_name,
    STATE(4612), 1,
      sym__simple_binding_pattern,
    STATE(4808), 1,
      sym_parenthesized_operator,
    STATE(6643), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4254), 2,
      sym__constant,
      sym_signed_number,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4564), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4548), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [71511] = 26,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1065), 1,
      anon_sym_begin,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1099), 1,
      sym__lowercase_identifier,
    ACTIONS(1105), 1,
      anon_sym_LBRACK,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1115), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2083), 1,
      anon_sym_POUND,
    ACTIONS(2196), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3500), 1,
      sym__uppercase_identifier,
    ACTIONS(3510), 1,
      anon_sym_LPAREN,
    STATE(2284), 1,
      sym_attribute,
    STATE(3151), 1,
      sym__constructor_name,
    STATE(4309), 1,
      sym__simple_pattern,
    STATE(6644), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4254), 2,
      sym__constant,
      sym_signed_number,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4414), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(4428), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [71612] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2285), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3547), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3545), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [71671] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2286), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2075), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2081), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [71730] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2287), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3551), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3549), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [71789] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2288), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1723), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1721), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [71848] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2289), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3555), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3553), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [71907] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2290), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3559), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3557), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [71966] = 26,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1467), 1,
      sym__lowercase_identifier,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1489), 1,
      anon_sym_LBRACE,
    ACTIONS(1495), 1,
      anon_sym_POUND,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1501), 1,
      anon_sym_begin,
    ACTIONS(1509), 1,
      aux_sym_number_token1,
    ACTIONS(1511), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1513), 1,
      anon_sym_SQUOTE,
    ACTIONS(1515), 1,
      anon_sym_DQUOTE,
    ACTIONS(2097), 1,
      sym__uppercase_identifier,
    ACTIONS(2099), 1,
      aux_sym_tag_token1,
    ACTIONS(2394), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2396), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3561), 1,
      anon_sym_LPAREN,
    STATE(2238), 1,
      sym__simple_pattern,
    STATE(2291), 1,
      sym_attribute,
    STATE(2597), 1,
      sym__constructor_name,
    STATE(6556), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1487), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2212), 2,
      sym__constant,
      sym_signed_number,
    STATE(2280), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(2286), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [72067] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2292), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3565), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3563), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [72126] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2293), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1735), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1733), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [72185] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2294), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3373), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3371), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [72244] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2295), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3377), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3375), 32,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [72303] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2296), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3569), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3567), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [72362] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2297), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3573), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3571), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [72421] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2298), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3016), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3014), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [72480] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2299), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3020), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3018), 26,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [72539] = 27,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1035), 1,
      sym__lowercase_identifier,
    ACTIONS(1047), 1,
      anon_sym_LBRACK,
    ACTIONS(1051), 1,
      anon_sym_LBRACE,
    ACTIONS(1061), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1065), 1,
      anon_sym_begin,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2083), 1,
      anon_sym_POUND,
    ACTIONS(2196), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3498), 1,
      anon_sym_LPAREN,
    ACTIONS(3500), 1,
      sym__uppercase_identifier,
    STATE(2300), 1,
      sym_attribute,
    STATE(3151), 1,
      sym__constructor_name,
    STATE(3501), 1,
      sym_parenthesized_operator,
    STATE(4612), 1,
      sym__simple_binding_pattern,
    STATE(6600), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4254), 2,
      sym__constant,
      sym_signed_number,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4564), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
    STATE(4548), 11,
      sym__delimited_binding_pattern,
      sym_typed_binding_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_binding_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__value_name,
      sym_constructor_path,
      sym_tag,
  [72642] = 26,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1065), 1,
      anon_sym_begin,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1099), 1,
      sym__lowercase_identifier,
    ACTIONS(1105), 1,
      anon_sym_LBRACK,
    ACTIONS(1107), 1,
      anon_sym_LBRACE,
    ACTIONS(1115), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2186), 1,
      anon_sym_POUND,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2196), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3438), 1,
      sym__uppercase_identifier,
    ACTIONS(3472), 1,
      anon_sym_LPAREN,
    STATE(2301), 1,
      sym_attribute,
    STATE(4309), 1,
      sym__simple_pattern,
    STATE(4454), 1,
      sym__constructor_name,
    STATE(6683), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1049), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4254), 2,
      sym__constant,
      sym_signed_number,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    STATE(4414), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
    STATE(4428), 12,
      sym__delimited_pattern,
      sym_typed_pattern,
      sym_polymorphic_variant_pattern,
      sym_local_open_pattern,
      sym_package_pattern,
      sym_extension,
      sym_quoted_extension,
      sym__signed_constant,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
      sym_constructor_path,
      sym_tag,
  [72743] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3575), 1,
      anon_sym_PIPE,
    STATE(2302), 1,
      sym_attribute,
    STATE(2304), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3208), 11,
      anon_sym_let,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3206), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [72805] = 31,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3237), 1,
      anon_sym_LBRACE,
    ACTIONS(3243), 1,
      sym__uppercase_identifier,
    ACTIONS(3279), 1,
      sym__lowercase_identifier,
    ACTIONS(3281), 1,
      anon_sym_LPAREN,
    ACTIONS(3287), 1,
      anon_sym_PIPE,
    ACTIONS(3289), 1,
      anon_sym_LBRACK,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3577), 1,
      anon_sym_DOT_DOT,
    STATE(2303), 1,
      sym_attribute,
    STATE(2917), 1,
      sym__simple_type,
    STATE(3335), 1,
      sym__type,
    STATE(3350), 1,
      sym__proper_tuple_type,
    STATE(3549), 1,
      sym_constructor_declaration,
    STATE(6244), 1,
      sym_extended_module_path,
    STATE(7252), 1,
      sym_labeled_argument_type,
    STATE(7344), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3235), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3440), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3469), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [72915] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3579), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2304), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3212), 11,
      anon_sym_let,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3210), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [72975] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3582), 1,
      anon_sym_else,
    STATE(2126), 1,
      sym_else_clause,
    STATE(2305), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3271), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3269), 29,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [73037] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3575), 1,
      anon_sym_PIPE,
    STATE(2302), 1,
      aux_sym__match_cases_repeat1,
    STATE(2306), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3259), 11,
      anon_sym_let,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3257), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [73099] = 31,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3184), 1,
      anon_sym_LPAREN,
    ACTIONS(3192), 1,
      anon_sym_PIPE,
    ACTIONS(3194), 1,
      anon_sym_LBRACK,
    ACTIONS(3198), 1,
      anon_sym_LBRACE,
    ACTIONS(3200), 1,
      sym__uppercase_identifier,
    ACTIONS(3584), 1,
      anon_sym_DOT_DOT,
    STATE(2307), 1,
      sym_attribute,
    STATE(4185), 1,
      sym__simple_type,
    STATE(4544), 1,
      sym__type,
    STATE(4610), 1,
      sym__proper_tuple_type,
    STATE(5004), 1,
      sym_constructor_declaration,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6481), 1,
      sym_labeled_tuple_element_type,
    STATE(7168), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3196), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4628), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4646), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4300), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [73209] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3575), 1,
      anon_sym_PIPE,
    STATE(2304), 1,
      aux_sym__match_cases_repeat1,
    STATE(2308), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3259), 11,
      anon_sym_let,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3257), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [73271] = 31,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3223), 1,
      anon_sym_LPAREN,
    ACTIONS(3231), 1,
      anon_sym_PIPE,
    ACTIONS(3233), 1,
      anon_sym_LBRACK,
    ACTIONS(3237), 1,
      anon_sym_LBRACE,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3243), 1,
      sym__uppercase_identifier,
    ACTIONS(3586), 1,
      anon_sym_DOT_DOT,
    STATE(2309), 1,
      sym_attribute,
    STATE(3209), 1,
      sym__simple_type,
    STATE(3529), 1,
      sym__type,
    STATE(3557), 1,
      sym__proper_tuple_type,
    STATE(3870), 1,
      sym_constructor_declaration,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6883), 1,
      sym_labeled_argument_type,
    STATE(6885), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3235), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3600), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3664), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [73381] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3575), 1,
      anon_sym_PIPE,
    STATE(2308), 1,
      aux_sym__match_cases_repeat1,
    STATE(2310), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3180), 11,
      anon_sym_let,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3178), 30,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [73443] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2954), 1,
      anon_sym_COMMA,
    STATE(2311), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2880), 12,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2878), 30,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [73503] = 31,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3223), 1,
      anon_sym_LPAREN,
    ACTIONS(3231), 1,
      anon_sym_PIPE,
    ACTIONS(3233), 1,
      anon_sym_LBRACK,
    ACTIONS(3237), 1,
      anon_sym_LBRACE,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3243), 1,
      sym__uppercase_identifier,
    ACTIONS(3588), 1,
      anon_sym_DOT_DOT,
    STATE(2312), 1,
      sym_attribute,
    STATE(3209), 1,
      sym__simple_type,
    STATE(3537), 1,
      sym__type,
    STATE(3557), 1,
      sym__proper_tuple_type,
    STATE(3870), 1,
      sym_constructor_declaration,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6883), 1,
      sym_labeled_argument_type,
    STATE(6885), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3235), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3600), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3674), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [73613] = 31,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3237), 1,
      anon_sym_LBRACE,
    ACTIONS(3243), 1,
      sym__uppercase_identifier,
    ACTIONS(3279), 1,
      sym__lowercase_identifier,
    ACTIONS(3281), 1,
      anon_sym_LPAREN,
    ACTIONS(3287), 1,
      anon_sym_PIPE,
    ACTIONS(3289), 1,
      anon_sym_LBRACK,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3590), 1,
      anon_sym_DOT_DOT,
    STATE(2313), 1,
      sym_attribute,
    STATE(2917), 1,
      sym__simple_type,
    STATE(3318), 1,
      sym__type,
    STATE(3350), 1,
      sym__proper_tuple_type,
    STATE(3549), 1,
      sym_constructor_declaration,
    STATE(6244), 1,
      sym_extended_module_path,
    STATE(7252), 1,
      sym_labeled_argument_type,
    STATE(7344), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3235), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3440), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3456), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [73723] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3582), 1,
      anon_sym_else,
    STATE(2119), 1,
      sym_else_clause,
    STATE(2314), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3277), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3275), 29,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [73785] = 31,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3184), 1,
      anon_sym_LPAREN,
    ACTIONS(3192), 1,
      anon_sym_PIPE,
    ACTIONS(3194), 1,
      anon_sym_LBRACK,
    ACTIONS(3198), 1,
      anon_sym_LBRACE,
    ACTIONS(3200), 1,
      sym__uppercase_identifier,
    ACTIONS(3592), 1,
      anon_sym_DOT_DOT,
    STATE(2315), 1,
      sym_attribute,
    STATE(4185), 1,
      sym__simple_type,
    STATE(4563), 1,
      sym__type,
    STATE(4610), 1,
      sym__proper_tuple_type,
    STATE(5004), 1,
      sym_constructor_declaration,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6481), 1,
      sym_labeled_tuple_element_type,
    STATE(7168), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3196), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4628), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4658), 2,
      sym_variant_declaration,
      sym_record_declaration,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4300), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [73895] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2974), 1,
      anon_sym_COMMA,
    STATE(2316), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2880), 12,
      anon_sym_let,
      anon_sym_PIPE,
      anon_sym_SEMI,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      anon_sym_mod,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2878), 29,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_operator,
      aux_sym_directive_token1,
  [73954] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3594), 1,
      anon_sym_DOT,
    ACTIONS(3596), 1,
      anon_sym_POUND,
    ACTIONS(3598), 1,
      sym_hash_operator,
    STATE(2317), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1665), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1663), 23,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74016] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2318), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1653), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1651), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74072] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2319), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1625), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1623), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74128] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2320), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1661), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1659), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74184] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2321), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1707), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1705), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74240] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2322), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3080), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3078), 21,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [74296] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2323), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1727), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1725), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74352] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2324), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1561), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1559), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74408] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2325), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1565), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1563), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74464] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2326), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1541), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1539), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74520] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2327), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1685), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1683), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74576] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2328), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1767), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1765), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74632] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2329), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1775), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1773), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74688] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2330), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1669), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1667), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74744] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2331), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1657), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1655), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74800] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2332), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1787), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1785), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74856] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2333), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3255), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(3253), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74912] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2334), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1799), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1797), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [74968] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2335), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1557), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1555), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75024] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2336), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1553), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1551), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75080] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2337), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1545), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1543), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75136] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2338), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1807), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1805), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75192] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2339), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1743), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1741), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75248] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2340), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1689), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1687), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75304] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2341), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3251), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(3249), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75360] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2342), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3204), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(3202), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75416] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2343), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1735), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1733), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75472] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3600), 1,
      anon_sym_COLON,
    STATE(2344), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3547), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3545), 22,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [75530] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2345), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1693), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1691), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75586] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2346), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1803), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1801), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75642] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2347), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1739), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1737), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75698] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2348), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1731), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1729), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75754] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2349), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1823), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1821), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75810] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2350), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1573), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1571), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75866] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2351), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1573), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1571), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75922] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2352), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1577), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1575), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [75978] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2353), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1771), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1769), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76034] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2354), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1597), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1595), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76090] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2355), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1759), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1757), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76146] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2356), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1697), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1695), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76202] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2357), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1613), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1611), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76258] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_DOT,
    STATE(2358), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1701), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1699), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76316] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3594), 1,
      anon_sym_DOT,
    ACTIONS(3596), 1,
      anon_sym_POUND,
    ACTIONS(3598), 1,
      sym_hash_operator,
    STATE(2359), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1617), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1615), 23,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76378] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2360), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1827), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1825), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76434] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2361), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1711), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1709), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76490] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2362), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1771), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1769), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76546] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2363), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1831), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1829), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76602] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2364), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1629), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1627), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76658] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1703), 1,
      anon_sym_DOT,
    STATE(2365), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1701), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1699), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76716] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2366), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1593), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1591), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76772] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2367), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1791), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1789), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76828] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2368), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1779), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1777), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76884] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2369), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1715), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1713), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76940] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2370), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1673), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1671), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [76996] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2371), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1621), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1619), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77052] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2372), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1677), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1675), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77108] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2373), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1649), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1647), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77164] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2374), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1681), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1679), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77220] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2375), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1819), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1817), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77276] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2376), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1783), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1781), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77332] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2377), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1853), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1851), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77388] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2378), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1719), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1717), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77444] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2379), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1569), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1567), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77500] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2380), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1835), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1833), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77556] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2381), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1839), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1837), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77612] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2382), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1657), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1655), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77668] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2383), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1697), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1695), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77724] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2384), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1723), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1721), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77780] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2385), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1621), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1619), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77836] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2386), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1779), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1777), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77892] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2387), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1783), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1781), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [77948] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2388), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1569), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1567), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78004] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2389), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1845), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1843), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78060] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2390), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1533), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1531), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78116] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2391), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1581), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1579), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78172] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2392), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1751), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1749), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78228] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2393), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1755), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1753), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78284] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2394), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1763), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1761), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78340] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2395), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1795), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1793), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78396] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3594), 1,
      anon_sym_DOT,
    STATE(2396), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1585), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1583), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78454] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1703), 1,
      anon_sym_DOT,
    STATE(2397), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1701), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1699), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [78512] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2398), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1751), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1749), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78568] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2399), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1537), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1535), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78624] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2400), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1849), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1847), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78680] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2401), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1747), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1745), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78736] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2402), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1589), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1587), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78792] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2403), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1755), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1753), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78848] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2404), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1763), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1761), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78904] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2405), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1739), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1737), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [78960] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2406), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1601), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1599), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79016] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2407), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1605), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1603), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79072] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2408), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1609), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1607), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79128] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2409), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1549), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1547), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79184] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2410), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1811), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1809), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79240] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2411), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1791), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1789), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79296] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2412), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1629), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1627), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79352] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2413), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1633), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1631), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79408] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2414), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3267), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(3265), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79464] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2415), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1795), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1793), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79520] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2416), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1637), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1635), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79576] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2417), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1803), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(1801), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79632] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_DOT,
    STATE(2418), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1701), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1699), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [79690] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2419), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1815), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1813), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79746] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2420), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1727), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1725), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79802] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2421), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1641), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1639), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79858] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2422), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1645), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_DOT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_hash_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1643), 24,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_POUND,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79914] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2423), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3313), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
      aux_sym_tag_token1,
    ACTIONS(3311), 24,
      anon_sym_let,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [79970] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2424), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3514), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3512), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80025] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2425), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2714), 12,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2712), 28,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [80080] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2426), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2932), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2930), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80135] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2427), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3108), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3106), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80190] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2428), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3551), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3549), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80245] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2429), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1819), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1817), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80300] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2430), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1827), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1825), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80355] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2431), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3484), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3482), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80410] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2432), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3555), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3553), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80465] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2433), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3565), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3563), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80520] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2434), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1711), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1709), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80575] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2435), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3052), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3050), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80630] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2436), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1831), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1829), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80685] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2437), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3476), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3474), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80740] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2438), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3538), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3536), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80795] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2439), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3492), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3490), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80850] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2440), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3080), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3078), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [80905] = 32,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(1195), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3602), 1,
      anon_sym_RPAREN,
    ACTIONS(3604), 1,
      anon_sym_module,
    STATE(2441), 1,
      sym_attribute,
    STATE(3181), 1,
      sym_type_variable,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5476), 1,
      sym__type,
    STATE(5489), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6640), 1,
      sym__abstract_type,
    STATE(6886), 1,
      sym_polymorphic_type,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [81014] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2442), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3496), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3494), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81069] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2443), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3534), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3532), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81124] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2444), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3504), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3502), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81179] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2445), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1625), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1623), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81234] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2446), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3508), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3506), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81289] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2447), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3056), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3054), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81344] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2448), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1835), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1833), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81399] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2449), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3524), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3522), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81454] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2450), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3176), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3174), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81509] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2451), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1787), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1785), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81564] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2452), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2944), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2942), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81619] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2453), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3573), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3571), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81674] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2454), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3470), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3468), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81729] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2455), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1807), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1805), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81784] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2456), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3480), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3478), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81839] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2457), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1727), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1725), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81894] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2458), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1811), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1809), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [81949] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2459), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1723), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1721), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82004] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2460), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1771), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1769), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82059] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2461), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1759), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1757), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82114] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2462), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3559), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3557), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82169] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2463), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3462), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3460), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82224] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2464), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1815), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1813), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82279] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2465), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1791), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1789), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82334] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2466), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3528), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3526), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82389] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2467), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3569), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3567), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82444] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2468), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2718), 12,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2716), 28,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [82499] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2469), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3488), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3486), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82554] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2470), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1799), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1797), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82609] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2471), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3442), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3440), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82664] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2472), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3547), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3545), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82719] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2473), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3456), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3454), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82774] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2474), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2075), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2081), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82829] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2475), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3038), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3036), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82884] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2476), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1715), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1713), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82939] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2477), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3466), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3464), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [82994] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2478), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1803), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1801), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83049] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2479), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3434), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3432), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83104] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2480), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1689), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1687), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83159] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2481), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3016), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3014), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83214] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2482), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3020), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3018), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83269] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2483), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1685), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1683), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83324] = 32,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(1191), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3604), 1,
      anon_sym_module,
    ACTIONS(3606), 1,
      anon_sym_RPAREN,
    STATE(2484), 1,
      sym_attribute,
    STATE(3181), 1,
      sym_type_variable,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5476), 1,
      sym__type,
    STATE(5489), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6640), 1,
      sym__abstract_type,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7237), 1,
      sym_polymorphic_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [83433] = 32,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(1191), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3606), 1,
      anon_sym_RPAREN,
    ACTIONS(3608), 1,
      anon_sym_module,
    STATE(2485), 1,
      sym_attribute,
    STATE(3181), 1,
      sym_type_variable,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5345), 1,
      sym__type,
    STATE(5489), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6339), 1,
      sym_polymorphic_type,
    STATE(6640), 1,
      sym__abstract_type,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [83542] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2486), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1845), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1843), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83597] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2487), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2710), 12,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2708), 28,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_then,
      anon_sym_else,
      anon_sym_do,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [83652] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2488), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1735), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1733), 21,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [83707] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2489), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3052), 19,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3050), 21,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [83762] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3610), 1,
      anon_sym_COLON,
    STATE(2490), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1971), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1967), 23,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [83818] = 30,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3606), 1,
      anon_sym_RBRACK,
    ACTIONS(3612), 1,
      anon_sym_PIPE,
    STATE(2491), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(6271), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7051), 1,
      sym__tag_spec,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [83922] = 30,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3614), 1,
      anon_sym_PIPE,
    ACTIONS(3616), 1,
      anon_sym_RBRACK,
    STATE(2492), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(5887), 1,
      sym__tag_spec,
    STATE(5930), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [84026] = 30,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3602), 1,
      anon_sym_RBRACK,
    ACTIONS(3618), 1,
      anon_sym_PIPE,
    STATE(2493), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(6099), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7203), 1,
      sym__tag_spec,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [84130] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3622), 1,
      sym__lowercase_identifier,
    ACTIONS(3626), 1,
      anon_sym_POUND,
    STATE(2494), 1,
      sym_attribute,
    STATE(2604), 1,
      sym_type_constructor_path,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3620), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(3624), 18,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
  [84194] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3622), 1,
      sym__lowercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    STATE(2495), 1,
      sym_attribute,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3628), 16,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(3630), 18,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
  [84258] = 30,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3606), 1,
      anon_sym_RBRACK,
    ACTIONS(3618), 1,
      anon_sym_PIPE,
    STATE(2496), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(6099), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7203), 1,
      sym__tag_spec,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [84362] = 30,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3634), 1,
      anon_sym_PIPE,
    ACTIONS(3636), 1,
      anon_sym_RBRACK,
    STATE(2497), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(5571), 1,
      sym__tag_spec,
    STATE(5930), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [84466] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2498), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3640), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(3638), 20,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
  [84520] = 30,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3642), 1,
      anon_sym_PIPE,
    ACTIONS(3644), 1,
      anon_sym_RBRACK,
    STATE(2499), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(5889), 1,
      sym__tag_spec,
    STATE(5930), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [84624] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2854), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3646), 1,
      anon_sym_COMMA,
    ACTIONS(3648), 1,
      anon_sym_SEMI,
    ACTIONS(3650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3652), 1,
      sym_concat_operator,
    ACTIONS(3654), 1,
      sym_rel_operator,
    ACTIONS(3656), 1,
      sym_assign_operator,
    STATE(662), 1,
      sym_pow_operator,
    STATE(663), 1,
      sym_mult_operator,
    STATE(664), 1,
      sym_add_operator,
    STATE(665), 1,
      sym_and_operator,
    STATE(666), 1,
      sym_or_operator,
    STATE(2500), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2852), 10,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [84713] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2501), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2019), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2017), 23,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [84766] = 29,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3658), 1,
      anon_sym_RBRACK,
    STATE(2502), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(5587), 1,
      sym__tag_spec,
    STATE(5930), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [84867] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3660), 1,
      anon_sym_COLON2,
    STATE(2503), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3052), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3050), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [84922] = 30,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    STATE(2504), 1,
      sym_attribute,
    STATE(3181), 1,
      sym_type_variable,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4150), 1,
      sym__polymorphic_type,
    STATE(4152), 1,
      sym_polymorphic_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(4239), 1,
      sym__type,
    STATE(5489), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6640), 1,
      sym__abstract_type,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [85025] = 30,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3662), 1,
      sym__lowercase_identifier,
    ACTIONS(3664), 1,
      anon_sym_LPAREN,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3672), 1,
      anon_sym_LT,
    ACTIONS(3674), 1,
      anon_sym_POUND,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    STATE(2505), 1,
      sym_attribute,
    STATE(3181), 1,
      sym_type_variable,
    STATE(4150), 1,
      sym__polymorphic_type,
    STATE(4152), 1,
      sym_polymorphic_type,
    STATE(4297), 1,
      sym__simple_type,
    STATE(4792), 1,
      sym__proper_tuple_type,
    STATE(5466), 1,
      sym__type,
    STATE(5494), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(5999), 1,
      sym_extended_module_path,
    STATE(7131), 1,
      sym_labeled_argument_type,
    STATE(7142), 1,
      sym__abstract_type,
    STATE(7170), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4300), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4452), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [85128] = 30,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3604), 1,
      anon_sym_module,
    STATE(2506), 1,
      sym_attribute,
    STATE(3181), 1,
      sym_type_variable,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5476), 1,
      sym__type,
    STATE(5489), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6534), 1,
      sym_polymorphic_type,
    STATE(6640), 1,
      sym__abstract_type,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [85231] = 23,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    ACTIONS(3680), 1,
      anon_sym_SEMI,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3686), 1,
      sym_rel_operator,
    ACTIONS(3688), 1,
      sym_assign_operator,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(2507), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2710), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2708), 9,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [85320] = 23,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    ACTIONS(3680), 1,
      anon_sym_SEMI,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3686), 1,
      sym_rel_operator,
    ACTIONS(3688), 1,
      sym_assign_operator,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(2508), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2718), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2716), 9,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [85409] = 30,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3604), 1,
      anon_sym_module,
    STATE(2509), 1,
      sym_attribute,
    STATE(3181), 1,
      sym_type_variable,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5476), 1,
      sym__type,
    STATE(5489), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6640), 1,
      sym__abstract_type,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    STATE(7323), 1,
      sym_polymorphic_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [85512] = 23,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    ACTIONS(3680), 1,
      anon_sym_SEMI,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3686), 1,
      sym_rel_operator,
    ACTIONS(3688), 1,
      sym_assign_operator,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(2510), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2852), 9,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [85601] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(2511), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 19,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [85666] = 29,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3618), 1,
      anon_sym_PIPE,
    STATE(2512), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5121), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(6099), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7203), 1,
      sym__tag_spec,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [85767] = 30,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3690), 1,
      anon_sym_module,
    STATE(2513), 1,
      sym_attribute,
    STATE(3181), 1,
      sym_type_variable,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5303), 1,
      sym__type,
    STATE(5489), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6640), 1,
      sym__abstract_type,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7175), 1,
      sym_polymorphic_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [85870] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3686), 1,
      sym_rel_operator,
    ACTIONS(3688), 1,
      sym_assign_operator,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(2514), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2826), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2824), 10,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [85957] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(2515), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2874), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2872), 23,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [86020] = 30,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3692), 1,
      anon_sym_LPAREN,
    STATE(2516), 1,
      sym_attribute,
    STATE(3209), 1,
      sym__simple_type,
    STATE(3296), 1,
      sym_type_variable,
    STATE(3557), 1,
      sym__proper_tuple_type,
    STATE(4053), 1,
      sym__polymorphic_type,
    STATE(4055), 1,
      sym_polymorphic_type,
    STATE(4141), 1,
      sym__type,
    STATE(5374), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6883), 1,
      sym_labeled_argument_type,
    STATE(6885), 1,
      sym_labeled_tuple_element_type,
    STATE(7176), 1,
      sym__abstract_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [86123] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2718), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3646), 1,
      anon_sym_COMMA,
    ACTIONS(3648), 1,
      anon_sym_SEMI,
    ACTIONS(3650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3652), 1,
      sym_concat_operator,
    ACTIONS(3654), 1,
      sym_rel_operator,
    ACTIONS(3656), 1,
      sym_assign_operator,
    STATE(662), 1,
      sym_pow_operator,
    STATE(663), 1,
      sym_mult_operator,
    STATE(664), 1,
      sym_add_operator,
    STATE(665), 1,
      sym_and_operator,
    STATE(666), 1,
      sym_or_operator,
    STATE(2517), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2716), 10,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [86212] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3686), 1,
      sym_rel_operator,
    ACTIONS(3688), 1,
      sym_assign_operator,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(2518), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2866), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2864), 10,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [86299] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(2519), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 6,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 13,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_or,
      sym_assign_operator,
  [86374] = 29,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3694), 1,
      anon_sym_RBRACK,
    STATE(2520), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(5548), 1,
      sym__tag_spec,
    STATE(5930), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [86475] = 30,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(3279), 1,
      sym__lowercase_identifier,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3696), 1,
      anon_sym_LPAREN,
    ACTIONS(3698), 1,
      anon_sym_LBRACK,
    STATE(2521), 1,
      sym_attribute,
    STATE(2917), 1,
      sym__simple_type,
    STATE(3125), 1,
      sym_type_variable,
    STATE(3350), 1,
      sym__proper_tuple_type,
    STATE(3902), 1,
      sym__type,
    STATE(4053), 1,
      sym__polymorphic_type,
    STATE(4055), 1,
      sym_polymorphic_type,
    STATE(5356), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6244), 1,
      sym_extended_module_path,
    STATE(7252), 1,
      sym_labeled_argument_type,
    STATE(7301), 1,
      sym__abstract_type,
    STATE(7344), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [86578] = 30,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3608), 1,
      anon_sym_module,
    STATE(2522), 1,
      sym_attribute,
    STATE(3181), 1,
      sym_type_variable,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5345), 1,
      sym__type,
    STATE(5489), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6339), 1,
      sym_polymorphic_type,
    STATE(6640), 1,
      sym__abstract_type,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [86681] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3686), 1,
      sym_rel_operator,
    ACTIONS(3688), 1,
      sym_assign_operator,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(2523), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2862), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 10,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [86768] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2710), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3646), 1,
      anon_sym_COMMA,
    ACTIONS(3648), 1,
      anon_sym_SEMI,
    ACTIONS(3650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3652), 1,
      sym_concat_operator,
    ACTIONS(3654), 1,
      sym_rel_operator,
    ACTIONS(3656), 1,
      sym_assign_operator,
    STATE(662), 1,
      sym_pow_operator,
    STATE(663), 1,
      sym_mult_operator,
    STATE(664), 1,
      sym_add_operator,
    STATE(665), 1,
      sym_and_operator,
    STATE(666), 1,
      sym_or_operator,
    STATE(2524), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2708), 10,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [86857] = 29,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3700), 1,
      anon_sym_PIPE,
    STATE(2525), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(6242), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6889), 1,
      sym__tag_spec,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [86958] = 29,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3702), 1,
      anon_sym_PIPE,
    STATE(2526), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5313), 1,
      sym__tag_spec,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(5930), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [87059] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(662), 1,
      sym_pow_operator,
    STATE(663), 1,
      sym_mult_operator,
    STATE(664), 1,
      sym_add_operator,
    STATE(665), 1,
      sym_and_operator,
    STATE(666), 1,
      sym_or_operator,
    STATE(2527), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2874), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2872), 25,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [87122] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(2528), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 9,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 15,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [87191] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3646), 1,
      anon_sym_COMMA,
    ACTIONS(3650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3652), 1,
      sym_concat_operator,
    ACTIONS(3654), 1,
      sym_rel_operator,
    STATE(662), 1,
      sym_pow_operator,
    STATE(663), 1,
      sym_mult_operator,
    STATE(664), 1,
      sym_add_operator,
    STATE(665), 1,
      sym_and_operator,
    STATE(666), 1,
      sym_or_operator,
    STATE(2529), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2878), 12,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
  [87276] = 30,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3604), 1,
      anon_sym_module,
    STATE(2530), 1,
      sym_attribute,
    STATE(3181), 1,
      sym_type_variable,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5476), 1,
      sym__type,
    STATE(5489), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6640), 1,
      sym__abstract_type,
    STATE(6808), 1,
      sym_polymorphic_type,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [87379] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3686), 1,
      sym_rel_operator,
    ACTIONS(3688), 1,
      sym_assign_operator,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(2531), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2890), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2888), 10,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
  [87466] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3650), 1,
      anon_sym_COLON_COLON,
    STATE(662), 1,
      sym_pow_operator,
    STATE(663), 1,
      sym_mult_operator,
    STATE(664), 1,
      sym_add_operator,
    STATE(665), 1,
      sym_and_operator,
    STATE(666), 1,
      sym_or_operator,
    STATE(2532), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2884), 4,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [87539] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3652), 1,
      sym_concat_operator,
    STATE(662), 1,
      sym_pow_operator,
    STATE(663), 1,
      sym_mult_operator,
    STATE(664), 1,
      sym_add_operator,
    STATE(665), 1,
      sym_and_operator,
    STATE(666), 1,
      sym_or_operator,
    STATE(2533), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 4,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [87614] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3652), 1,
      sym_concat_operator,
    STATE(662), 1,
      sym_pow_operator,
    STATE(663), 1,
      sym_mult_operator,
    STATE(664), 1,
      sym_add_operator,
    STATE(665), 1,
      sym_and_operator,
    STATE(666), 1,
      sym_or_operator,
    STATE(2534), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 4,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [87689] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2862), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3646), 1,
      anon_sym_COMMA,
    ACTIONS(3650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3652), 1,
      sym_concat_operator,
    ACTIONS(3654), 1,
      sym_rel_operator,
    ACTIONS(3656), 1,
      sym_assign_operator,
    STATE(662), 1,
      sym_pow_operator,
    STATE(663), 1,
      sym_mult_operator,
    STATE(664), 1,
      sym_add_operator,
    STATE(665), 1,
      sym_and_operator,
    STATE(666), 1,
      sym_or_operator,
    STATE(2535), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 11,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [87776] = 29,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3704), 1,
      anon_sym_PIPE,
    STATE(2536), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5376), 1,
      sym__tag_spec,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(5930), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [87877] = 29,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3618), 1,
      anon_sym_PIPE,
    STATE(2537), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5219), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(6099), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7203), 1,
      sym__tag_spec,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [87978] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(662), 1,
      sym_pow_operator,
    STATE(663), 1,
      sym_mult_operator,
    STATE(664), 1,
      sym_add_operator,
    STATE(665), 1,
      sym_and_operator,
    STATE(666), 1,
      sym_or_operator,
    STATE(2538), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 21,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [88043] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(662), 1,
      sym_pow_operator,
    STATE(663), 1,
      sym_mult_operator,
    STATE(664), 1,
      sym_add_operator,
    STATE(665), 1,
      sym_and_operator,
    STATE(666), 1,
      sym_or_operator,
    STATE(2539), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 21,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [88108] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    STATE(662), 1,
      sym_pow_operator,
    STATE(663), 1,
      sym_mult_operator,
    STATE(664), 1,
      sym_add_operator,
    STATE(665), 1,
      sym_and_operator,
    STATE(666), 1,
      sym_or_operator,
    STATE(2540), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 7,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [88177] = 30,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    STATE(2541), 1,
      sym_attribute,
    STATE(3181), 1,
      sym_type_variable,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4150), 1,
      sym__polymorphic_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5489), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(5563), 1,
      sym__type,
    STATE(6149), 1,
      sym_polymorphic_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6640), 1,
      sym__abstract_type,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [88280] = 30,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3706), 1,
      anon_sym_LPAREN,
    STATE(2542), 1,
      sym_attribute,
    STATE(3181), 1,
      sym_type_variable,
    STATE(3843), 1,
      sym__simple_type,
    STATE(4150), 1,
      sym__polymorphic_type,
    STATE(4152), 1,
      sym_polymorphic_type,
    STATE(4236), 1,
      sym__proper_tuple_type,
    STATE(4239), 1,
      sym__type,
    STATE(5272), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6887), 1,
      sym__abstract_type,
    STATE(7048), 1,
      sym_labeled_argument_type,
    STATE(7050), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [88383] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3652), 1,
      sym_concat_operator,
    ACTIONS(3654), 1,
      sym_rel_operator,
    STATE(662), 1,
      sym_pow_operator,
    STATE(663), 1,
      sym_mult_operator,
    STATE(664), 1,
      sym_add_operator,
    STATE(665), 1,
      sym_and_operator,
    STATE(666), 1,
      sym_or_operator,
    STATE(2543), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2862), 2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_PIPE_PIPE,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 14,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_or,
      sym_assign_operator,
  [88462] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2862), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3652), 1,
      sym_concat_operator,
    ACTIONS(3654), 1,
      sym_rel_operator,
    STATE(662), 1,
      sym_pow_operator,
    STATE(663), 1,
      sym_mult_operator,
    STATE(664), 1,
      sym_add_operator,
    STATE(665), 1,
      sym_and_operator,
    STATE(666), 1,
      sym_or_operator,
    STATE(2544), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 13,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_assign_operator,
  [88545] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3646), 1,
      anon_sym_COMMA,
    ACTIONS(3650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3652), 1,
      sym_concat_operator,
    ACTIONS(3654), 1,
      sym_rel_operator,
    ACTIONS(3656), 1,
      sym_assign_operator,
    STATE(662), 1,
      sym_pow_operator,
    STATE(663), 1,
      sym_mult_operator,
    STATE(664), 1,
      sym_add_operator,
    STATE(665), 1,
      sym_and_operator,
    STATE(666), 1,
      sym_or_operator,
    STATE(2545), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2824), 11,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [88632] = 29,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3708), 1,
      anon_sym_PIPE,
    STATE(2546), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5364), 1,
      sym__tag_spec,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(5930), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [88733] = 30,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3604), 1,
      anon_sym_module,
    STATE(2547), 1,
      sym_attribute,
    STATE(3181), 1,
      sym_type_variable,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5476), 1,
      sym__type,
    STATE(5489), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6356), 1,
      sym_polymorphic_type,
    STATE(6640), 1,
      sym__abstract_type,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [88836] = 29,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3710), 1,
      anon_sym_RBRACK,
    STATE(2548), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(5873), 1,
      sym__tag_spec,
    STATE(5930), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [88937] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2866), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3646), 1,
      anon_sym_COMMA,
    ACTIONS(3650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3652), 1,
      sym_concat_operator,
    ACTIONS(3654), 1,
      sym_rel_operator,
    ACTIONS(3656), 1,
      sym_assign_operator,
    STATE(662), 1,
      sym_pow_operator,
    STATE(663), 1,
      sym_mult_operator,
    STATE(664), 1,
      sym_add_operator,
    STATE(665), 1,
      sym_and_operator,
    STATE(666), 1,
      sym_or_operator,
    STATE(2549), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2864), 11,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [89024] = 30,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3604), 1,
      anon_sym_module,
    STATE(2550), 1,
      sym_attribute,
    STATE(3181), 1,
      sym_type_variable,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5476), 1,
      sym__type,
    STATE(5489), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6640), 1,
      sym__abstract_type,
    STATE(6736), 1,
      sym_polymorphic_type,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [89127] = 30,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3604), 1,
      anon_sym_module,
    STATE(2551), 1,
      sym_attribute,
    STATE(3181), 1,
      sym_type_variable,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5476), 1,
      sym__type,
    STATE(5489), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6640), 1,
      sym__abstract_type,
    STATE(6886), 1,
      sym_polymorphic_type,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [89230] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2552), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1617), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1615), 23,
      anon_sym_and,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [89283] = 30,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3604), 1,
      anon_sym_module,
    STATE(2553), 1,
      sym_attribute,
    STATE(3181), 1,
      sym_type_variable,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5476), 1,
      sym__type,
    STATE(5489), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6640), 1,
      sym__abstract_type,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7018), 1,
      sym_polymorphic_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [89386] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2890), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3646), 1,
      anon_sym_COMMA,
    ACTIONS(3650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3652), 1,
      sym_concat_operator,
    ACTIONS(3654), 1,
      sym_rel_operator,
    ACTIONS(3656), 1,
      sym_assign_operator,
    STATE(662), 1,
      sym_pow_operator,
    STATE(663), 1,
      sym_mult_operator,
    STATE(664), 1,
      sym_add_operator,
    STATE(665), 1,
      sym_and_operator,
    STATE(666), 1,
      sym_or_operator,
    STATE(2554), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2888), 11,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [89473] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3712), 1,
      anon_sym_COLON,
    STATE(2555), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3547), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3545), 20,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [89528] = 30,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3604), 1,
      anon_sym_module,
    STATE(2556), 1,
      sym_attribute,
    STATE(3181), 1,
      sym_type_variable,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5476), 1,
      sym__type,
    STATE(5489), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6433), 1,
      sym_polymorphic_type,
    STATE(6640), 1,
      sym__abstract_type,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [89631] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1703), 1,
      anon_sym_DOT,
    STATE(2557), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1701), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1699), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [89686] = 20,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3686), 1,
      sym_rel_operator,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(2558), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2862), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 12,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      sym_assign_operator,
  [89769] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(2559), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 19,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [89834] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(2560), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2884), 6,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 14,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [89907] = 30,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3604), 1,
      anon_sym_module,
    STATE(2561), 1,
      sym_attribute,
    STATE(3181), 1,
      sym_type_variable,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5476), 1,
      sym__type,
    STATE(5489), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6640), 1,
      sym__abstract_type,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7237), 1,
      sym_polymorphic_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [90010] = 30,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3604), 1,
      anon_sym_module,
    STATE(2562), 1,
      sym_attribute,
    STATE(3181), 1,
      sym_type_variable,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5476), 1,
      sym__type,
    STATE(5489), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6640), 1,
      sym__abstract_type,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    STATE(7271), 1,
      sym_polymorphic_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [90113] = 29,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3618), 1,
      anon_sym_PIPE,
    STATE(2563), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(6099), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7203), 1,
      sym__tag_spec,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [90214] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(2564), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 6,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 13,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_or,
      sym_assign_operator,
  [90289] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_DOT,
    STATE(2565), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1701), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1699), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [90344] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3686), 1,
      sym_rel_operator,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(2566), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2880), 2,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2878), 11,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      sym_assign_operator,
  [90429] = 18,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3686), 1,
      sym_rel_operator,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(2567), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2862), 3,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_PIPE_PIPE,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 13,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_or,
      sym_assign_operator,
  [90508] = 29,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3618), 1,
      anon_sym_PIPE,
    STATE(2568), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5254), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(6099), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7203), 1,
      sym__tag_spec,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [90609] = 29,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(3612), 1,
      anon_sym_PIPE,
    STATE(2569), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(6271), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7051), 1,
      sym__tag_spec,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [90710] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2570), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3714), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3716), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [90762] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2571), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3718), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3720), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [90814] = 24,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3722), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(3725), 1,
      anon_sym_let,
    ACTIONS(3728), 1,
      anon_sym_external,
    ACTIONS(3731), 1,
      anon_sym_type,
    ACTIONS(3736), 1,
      anon_sym_exception,
    ACTIONS(3739), 1,
      anon_sym_module,
    ACTIONS(3742), 1,
      anon_sym_open,
    ACTIONS(3745), 1,
      anon_sym_include,
    ACTIONS(3748), 1,
      anon_sym_class,
    ACTIONS(3751), 1,
      anon_sym_val,
    ACTIONS(3754), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(3757), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(3760), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(3763), 1,
      sym_let_operator,
    ACTIONS(3766), 1,
      aux_sym_directive_token1,
    STATE(44), 1,
      aux_sym__structure_repeat1,
    STATE(2634), 1,
      sym_directive,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3734), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(2572), 2,
      sym_attribute,
      aux_sym__structure_repeat2,
    STATE(4059), 2,
      sym_toplevel_directive,
      sym__structure_item,
    STATE(4110), 4,
      sym__local_structure_item,
      sym_value_definition,
      sym_include_module,
      sym_value_specification,
    STATE(4107), 11,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [90904] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2573), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3769), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3771), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [90956] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2574), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3773), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3775), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [91008] = 25,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(231), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(259), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(261), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(267), 1,
      aux_sym_directive_token1,
    ACTIONS(3777), 1,
      anon_sym_let,
    ACTIONS(3779), 1,
      anon_sym_external,
    ACTIONS(3781), 1,
      anon_sym_type,
    ACTIONS(3783), 1,
      anon_sym_exception,
    ACTIONS(3785), 1,
      anon_sym_module,
    ACTIONS(3787), 1,
      anon_sym_open,
    ACTIONS(3789), 1,
      anon_sym_include,
    ACTIONS(3791), 1,
      anon_sym_class,
    ACTIONS(3793), 1,
      anon_sym_val,
    ACTIONS(3795), 1,
      sym_let_operator,
    STATE(40), 1,
      aux_sym__structure_repeat1,
    STATE(2572), 1,
      aux_sym__structure_repeat2,
    STATE(2575), 1,
      sym_attribute,
    STATE(2634), 1,
      sym_directive,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(331), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(4059), 2,
      sym_toplevel_directive,
      sym__structure_item,
    STATE(4110), 4,
      sym__local_structure_item,
      sym_value_definition,
      sym_include_module,
      sym_value_specification,
    STATE(4107), 11,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [91100] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2576), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3797), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3799), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [91152] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2577), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3801), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3803), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [91204] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2578), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3805), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3807), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [91256] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2579), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3809), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3811), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [91308] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2580), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3620), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3624), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [91360] = 29,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(1043), 1,
      anon_sym_type,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    STATE(2581), 1,
      sym_attribute,
    STATE(3181), 1,
      sym_type_variable,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4214), 1,
      sym__type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5489), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6640), 1,
      sym__abstract_type,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7218), 1,
      sym_polymorphic_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [91460] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2582), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3813), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3815), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [91512] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2583), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3817), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3819), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [91564] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2584), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3821), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3823), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [91616] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2585), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3825), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3827), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [91668] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2586), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3829), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3831), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [91720] = 26,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3833), 1,
      sym__lowercase_identifier,
    ACTIONS(3835), 1,
      anon_sym_let,
    ACTIONS(3837), 1,
      anon_sym_LPAREN,
    ACTIONS(3839), 1,
      anon_sym_LBRACK,
    ACTIONS(3841), 1,
      anon_sym_object,
    STATE(2587), 1,
      sym_attribute,
    STATE(3638), 1,
      sym__class_type,
    STATE(4724), 1,
      sym__simple_type,
    STATE(6254), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3326), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(3757), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(7317), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3462), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2574), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [91814] = 28,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(2588), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(5873), 1,
      sym__tag_spec,
    STATE(5930), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [91912] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2589), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3843), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3845), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [91964] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2590), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3847), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3849), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [92016] = 25,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(231), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(259), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(261), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(267), 1,
      aux_sym_directive_token1,
    ACTIONS(3777), 1,
      anon_sym_let,
    ACTIONS(3779), 1,
      anon_sym_external,
    ACTIONS(3781), 1,
      anon_sym_type,
    ACTIONS(3783), 1,
      anon_sym_exception,
    ACTIONS(3785), 1,
      anon_sym_module,
    ACTIONS(3787), 1,
      anon_sym_open,
    ACTIONS(3789), 1,
      anon_sym_include,
    ACTIONS(3791), 1,
      anon_sym_class,
    ACTIONS(3793), 1,
      anon_sym_val,
    ACTIONS(3795), 1,
      sym_let_operator,
    STATE(42), 1,
      aux_sym__structure_repeat1,
    STATE(2591), 1,
      sym_attribute,
    STATE(2592), 1,
      aux_sym__structure_repeat2,
    STATE(2634), 1,
      sym_directive,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(335), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(4059), 2,
      sym_toplevel_directive,
      sym__structure_item,
    STATE(4110), 4,
      sym__local_structure_item,
      sym_value_definition,
      sym_include_module,
      sym_value_specification,
    STATE(4107), 11,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [92108] = 25,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(231), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(259), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(261), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(267), 1,
      aux_sym_directive_token1,
    ACTIONS(3777), 1,
      anon_sym_let,
    ACTIONS(3779), 1,
      anon_sym_external,
    ACTIONS(3781), 1,
      anon_sym_type,
    ACTIONS(3783), 1,
      anon_sym_exception,
    ACTIONS(3785), 1,
      anon_sym_module,
    ACTIONS(3787), 1,
      anon_sym_open,
    ACTIONS(3789), 1,
      anon_sym_include,
    ACTIONS(3791), 1,
      anon_sym_class,
    ACTIONS(3793), 1,
      anon_sym_val,
    ACTIONS(3795), 1,
      sym_let_operator,
    STATE(38), 1,
      aux_sym__structure_repeat1,
    STATE(2572), 1,
      aux_sym__structure_repeat2,
    STATE(2592), 1,
      sym_attribute,
    STATE(2634), 1,
      sym_directive,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(345), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(4059), 2,
      sym_toplevel_directive,
      sym__structure_item,
    STATE(4110), 4,
      sym__local_structure_item,
      sym_value_definition,
      sym_include_module,
      sym_value_specification,
    STATE(4107), 11,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [92200] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2593), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3851), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3853), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [92252] = 28,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(2594), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5407), 1,
      sym__tag_spec,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(5930), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [92350] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2595), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3855), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3857), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [92402] = 26,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3833), 1,
      sym__lowercase_identifier,
    ACTIONS(3835), 1,
      anon_sym_let,
    ACTIONS(3837), 1,
      anon_sym_LPAREN,
    ACTIONS(3839), 1,
      anon_sym_LBRACK,
    ACTIONS(3841), 1,
      anon_sym_object,
    STATE(2596), 1,
      sym_attribute,
    STATE(3614), 1,
      sym__class_type,
    STATE(4724), 1,
      sym__simple_type,
    STATE(6254), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3326), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(3757), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(7317), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3462), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2574), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [92496] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2597), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1715), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1713), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [92548] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2598), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1811), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1809), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [92600] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2599), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1845), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1843), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [92652] = 25,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(231), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(259), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(261), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(267), 1,
      aux_sym_directive_token1,
    ACTIONS(3777), 1,
      anon_sym_let,
    ACTIONS(3779), 1,
      anon_sym_external,
    ACTIONS(3781), 1,
      anon_sym_type,
    ACTIONS(3783), 1,
      anon_sym_exception,
    ACTIONS(3785), 1,
      anon_sym_module,
    ACTIONS(3787), 1,
      anon_sym_open,
    ACTIONS(3789), 1,
      anon_sym_include,
    ACTIONS(3791), 1,
      anon_sym_class,
    ACTIONS(3793), 1,
      anon_sym_val,
    ACTIONS(3795), 1,
      sym_let_operator,
    STATE(38), 1,
      aux_sym__structure_repeat1,
    STATE(2575), 1,
      aux_sym__structure_repeat2,
    STATE(2600), 1,
      sym_attribute,
    STATE(2634), 1,
      sym_directive,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(345), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(4059), 2,
      sym_toplevel_directive,
      sym__structure_item,
    STATE(4110), 4,
      sym__local_structure_item,
      sym_value_definition,
      sym_include_module,
      sym_value_specification,
    STATE(4107), 11,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [92744] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2601), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1625), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(1623), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [92796] = 26,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3859), 1,
      sym__lowercase_identifier,
    ACTIONS(3861), 1,
      anon_sym_let,
    ACTIONS(3863), 1,
      anon_sym_LPAREN,
    ACTIONS(3865), 1,
      anon_sym_LBRACK,
    ACTIONS(3867), 1,
      anon_sym_object,
    STATE(2602), 1,
      sym_attribute,
    STATE(4733), 1,
      sym__simple_type,
    STATE(5513), 1,
      sym__class_type,
    STATE(6308), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3473), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(5509), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(7303), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(5534), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2574), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [92890] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2603), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3869), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3871), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [92942] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2604), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3873), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3875), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [92994] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2605), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3877), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3879), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [93046] = 26,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3835), 1,
      anon_sym_let,
    ACTIONS(3839), 1,
      anon_sym_LBRACK,
    ACTIONS(3841), 1,
      anon_sym_object,
    ACTIONS(3859), 1,
      sym__lowercase_identifier,
    ACTIONS(3881), 1,
      anon_sym_LPAREN,
    STATE(2606), 1,
      sym_attribute,
    STATE(3638), 1,
      sym__class_type,
    STATE(4785), 1,
      sym__simple_type,
    STATE(6308), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3473), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(3757), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(6714), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3462), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2574), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [93140] = 26,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3835), 1,
      anon_sym_let,
    ACTIONS(3839), 1,
      anon_sym_LBRACK,
    ACTIONS(3841), 1,
      anon_sym_object,
    ACTIONS(3859), 1,
      sym__lowercase_identifier,
    ACTIONS(3881), 1,
      anon_sym_LPAREN,
    STATE(2607), 1,
      sym_attribute,
    STATE(3614), 1,
      sym__class_type,
    STATE(4785), 1,
      sym__simple_type,
    STATE(6308), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3473), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(3757), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(6714), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3462), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2574), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [93234] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2608), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3883), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3885), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [93286] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2609), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3887), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3889), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [93338] = 26,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3835), 1,
      anon_sym_let,
    ACTIONS(3839), 1,
      anon_sym_LBRACK,
    ACTIONS(3841), 1,
      anon_sym_object,
    ACTIONS(3859), 1,
      sym__lowercase_identifier,
    ACTIONS(3881), 1,
      anon_sym_LPAREN,
    STATE(2610), 1,
      sym_attribute,
    STATE(3585), 1,
      sym__class_type,
    STATE(4785), 1,
      sym__simple_type,
    STATE(6308), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3473), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(3757), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(6714), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3462), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2574), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [93432] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2611), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3891), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3893), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [93484] = 28,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(2612), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(5930), 1,
      sym_tag_specification,
    STATE(5937), 1,
      sym__tag_spec,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [93582] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2613), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3895), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3897), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [93634] = 28,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(2614), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(5552), 1,
      sym__tag_spec,
    STATE(5930), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [93732] = 28,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(2615), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(5548), 1,
      sym__tag_spec,
    STATE(5930), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [93830] = 28,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(2616), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5454), 1,
      sym__type,
    STATE(5514), 1,
      sym__tag_spec,
    STATE(5540), 1,
      sym_tag,
    STATE(5930), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [93928] = 28,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(2617), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(5620), 1,
      sym__tag_spec,
    STATE(5930), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [94026] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2618), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3899), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3901), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [94078] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2710), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3646), 1,
      anon_sym_COMMA,
    ACTIONS(3650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3652), 1,
      sym_concat_operator,
    ACTIONS(3654), 1,
      sym_rel_operator,
    ACTIONS(3656), 1,
      sym_assign_operator,
    ACTIONS(3903), 1,
      anon_sym_SEMI,
    STATE(662), 1,
      sym_pow_operator,
    STATE(663), 1,
      sym_mult_operator,
    STATE(664), 1,
      sym_add_operator,
    STATE(665), 1,
      sym_and_operator,
    STATE(666), 1,
      sym_or_operator,
    STATE(2619), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2708), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [94166] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2718), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3646), 1,
      anon_sym_COMMA,
    ACTIONS(3650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3652), 1,
      sym_concat_operator,
    ACTIONS(3654), 1,
      sym_rel_operator,
    ACTIONS(3656), 1,
      sym_assign_operator,
    ACTIONS(3903), 1,
      anon_sym_SEMI,
    STATE(662), 1,
      sym_pow_operator,
    STATE(663), 1,
      sym_mult_operator,
    STATE(664), 1,
      sym_add_operator,
    STATE(665), 1,
      sym_and_operator,
    STATE(666), 1,
      sym_or_operator,
    STATE(2620), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2716), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [94254] = 26,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3841), 1,
      anon_sym_object,
    ACTIONS(3859), 1,
      sym__lowercase_identifier,
    ACTIONS(3905), 1,
      anon_sym_let,
    ACTIONS(3907), 1,
      anon_sym_LPAREN,
    ACTIONS(3909), 1,
      anon_sym_LBRACK,
    STATE(2621), 1,
      sym_attribute,
    STATE(3638), 1,
      sym__class_type,
    STATE(4716), 1,
      sym__simple_type,
    STATE(6308), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3473), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(3757), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(7315), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3462), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2574), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [94348] = 26,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3841), 1,
      anon_sym_object,
    ACTIONS(3859), 1,
      sym__lowercase_identifier,
    ACTIONS(3905), 1,
      anon_sym_let,
    ACTIONS(3907), 1,
      anon_sym_LPAREN,
    ACTIONS(3909), 1,
      anon_sym_LBRACK,
    STATE(2622), 1,
      sym_attribute,
    STATE(3614), 1,
      sym__class_type,
    STATE(4716), 1,
      sym__simple_type,
    STATE(6308), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3473), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(3757), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(7315), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3462), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2574), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [94442] = 26,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3841), 1,
      anon_sym_object,
    ACTIONS(3859), 1,
      sym__lowercase_identifier,
    ACTIONS(3905), 1,
      anon_sym_let,
    ACTIONS(3907), 1,
      anon_sym_LPAREN,
    ACTIONS(3909), 1,
      anon_sym_LBRACK,
    STATE(2623), 1,
      sym_attribute,
    STATE(3585), 1,
      sym__class_type,
    STATE(4716), 1,
      sym__simple_type,
    STATE(6308), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3473), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(3757), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(7315), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3462), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2574), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [94536] = 28,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(2624), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(5587), 1,
      sym__tag_spec,
    STATE(5930), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [94634] = 28,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(2625), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5315), 1,
      sym__tag_spec,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(5930), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [94732] = 28,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(2626), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5454), 1,
      sym__type,
    STATE(5540), 1,
      sym_tag,
    STATE(5604), 1,
      sym__tag_spec,
    STATE(5930), 1,
      sym_tag_specification,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [94830] = 26,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3859), 1,
      sym__lowercase_identifier,
    ACTIONS(3861), 1,
      anon_sym_let,
    ACTIONS(3863), 1,
      anon_sym_LPAREN,
    ACTIONS(3865), 1,
      anon_sym_LBRACK,
    ACTIONS(3867), 1,
      anon_sym_object,
    STATE(2627), 1,
      sym_attribute,
    STATE(4733), 1,
      sym__simple_type,
    STATE(5526), 1,
      sym__class_type,
    STATE(6308), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3473), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(5509), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(7303), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(5534), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2574), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [94924] = 26,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3833), 1,
      sym__lowercase_identifier,
    ACTIONS(3835), 1,
      anon_sym_let,
    ACTIONS(3837), 1,
      anon_sym_LPAREN,
    ACTIONS(3839), 1,
      anon_sym_LBRACK,
    ACTIONS(3841), 1,
      anon_sym_object,
    STATE(2628), 1,
      sym_attribute,
    STATE(3585), 1,
      sym__class_type,
    STATE(4724), 1,
      sym__simple_type,
    STATE(6254), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3326), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(3757), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(7317), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3462), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2574), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [95018] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2629), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3911), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3913), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [95070] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2854), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3646), 1,
      anon_sym_COMMA,
    ACTIONS(3650), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3652), 1,
      sym_concat_operator,
    ACTIONS(3654), 1,
      sym_rel_operator,
    ACTIONS(3656), 1,
      sym_assign_operator,
    ACTIONS(3903), 1,
      anon_sym_SEMI,
    STATE(662), 1,
      sym_pow_operator,
    STATE(663), 1,
      sym_mult_operator,
    STATE(664), 1,
      sym_add_operator,
    STATE(665), 1,
      sym_and_operator,
    STATE(666), 1,
      sym_or_operator,
    STATE(2630), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2852), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [95158] = 26,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3859), 1,
      sym__lowercase_identifier,
    ACTIONS(3861), 1,
      anon_sym_let,
    ACTIONS(3863), 1,
      anon_sym_LPAREN,
    ACTIONS(3865), 1,
      anon_sym_LBRACK,
    ACTIONS(3867), 1,
      anon_sym_object,
    STATE(2631), 1,
      sym_attribute,
    STATE(4733), 1,
      sym__simple_type,
    STATE(5447), 1,
      sym__class_type,
    STATE(6308), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3473), 2,
      sym_extension,
      sym_quoted_extension,
    STATE(5509), 2,
      sym__simple_class_type,
      sym_class_function_type,
    STATE(7303), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(5534), 4,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_class_type_path,
    STATE(2574), 7,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_type_constructor_path,
      sym_type_variable,
  [95252] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2632), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3915), 18,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3917), 19,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [95304] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3923), 1,
      anon_sym_RPAREN,
    STATE(2633), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3921), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3919), 20,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [95357] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1035), 1,
      sym__lowercase_identifier,
    ACTIONS(3929), 1,
      anon_sym_LPAREN,
    ACTIONS(3933), 1,
      anon_sym_LBRACE,
    ACTIONS(3935), 1,
      aux_sym_number_token1,
    ACTIONS(3937), 1,
      anon_sym_SQUOTE,
    ACTIONS(3939), 1,
      anon_sym_DQUOTE,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    STATE(2634), 1,
      sym_attribute,
    STATE(3501), 1,
      sym_parenthesized_operator,
    STATE(4041), 1,
      sym_module_path,
    STATE(4096), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3931), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4078), 2,
      sym__constant,
      sym_value_path,
    STATE(4117), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
    ACTIONS(3925), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
    ACTIONS(3927), 10,
      anon_sym_let,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
  [95434] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3943), 1,
      anon_sym_COLON,
    STATE(2635), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3547), 16,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3545), 19,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_initializer,
      anon_sym_begin,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [95487] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(2636), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3947), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3945), 21,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_end,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [95538] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2637), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3951), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3949), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [95588] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2638), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3955), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3953), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [95638] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(2639), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3921), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3919), 20,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [95688] = 23,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2854), 1,
      anon_sym_COLON2,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3688), 1,
      sym_assign_operator,
    ACTIONS(3957), 1,
      anon_sym_SEMI,
    ACTIONS(3959), 1,
      sym_rel_operator,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(2640), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2852), 7,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
  [95774] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2641), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3963), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3961), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [95824] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(2642), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3967), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3965), 20,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [95874] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(2643), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3971), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3969), 20,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [95924] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(2644), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3975), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3973), 20,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [95974] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(2645), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3923), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3977), 20,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [96024] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(2646), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3981), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      aux_sym_sign_operator_token1,
      sym_let_operator,
      sym_match_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3979), 20,
      anon_sym_let,
      anon_sym_TILDE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_fun,
      anon_sym_if,
      anon_sym_while,
      anon_sym_for,
      anon_sym_match,
      anon_sym_function,
      anon_sym_try,
      anon_sym_assert,
      anon_sym_lazy,
      anon_sym_new,
      anon_sym_begin,
      aux_sym_signed_number_token1,
      sym__lowercase_identifier,
  [96074] = 23,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2710), 1,
      anon_sym_COLON2,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3688), 1,
      sym_assign_operator,
    ACTIONS(3957), 1,
      anon_sym_SEMI,
    ACTIONS(3959), 1,
      sym_rel_operator,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(2647), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2708), 7,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
  [96160] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2648), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3985), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3983), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [96210] = 23,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2718), 1,
      anon_sym_COLON2,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3688), 1,
      sym_assign_operator,
    ACTIONS(3957), 1,
      anon_sym_SEMI,
    ACTIONS(3959), 1,
      sym_rel_operator,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(2649), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2716), 7,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_then,
  [96296] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2650), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3989), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3987), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [96346] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2651), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3993), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3991), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [96396] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2652), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3997), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3995), 33,
      ts_builtin_sym_end,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [96446] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2653), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1789), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1791), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [96495] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2654), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1737), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1739), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [96544] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2655), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1571), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1573), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [96593] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2656), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1547), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1549), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [96642] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2657), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1695), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1697), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [96691] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2658), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1591), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1593), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [96740] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2659), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1619), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1621), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [96789] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2660), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1627), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1629), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [96838] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2661), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1535), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1537), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [96887] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2662), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1655), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1657), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [96936] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2663), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1777), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1779), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [96985] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2664), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1781), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1783), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97034] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2665), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1567), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1569), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97083] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2666), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1749), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1751), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97132] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2667), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1753), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1755), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97181] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2668), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1761), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1763), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97230] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2669), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1793), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1795), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97279] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3999), 1,
      aux_sym_attribute_id_token1,
    STATE(2670), 1,
      sym_attribute,
    STATE(2671), 1,
      aux_sym_attribute_id_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2733), 15,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2731), 17,
      anon_sym_rec,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_class,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [97332] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4001), 1,
      aux_sym_attribute_id_token1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2671), 2,
      sym_attribute,
      aux_sym_attribute_id_repeat1,
    ACTIONS(2724), 15,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2722), 17,
      anon_sym_rec,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_class,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [97383] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(4010), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4012), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2672), 1,
      sym_attribute,
    STATE(3167), 1,
      sym__simple_module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4006), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4004), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [97442] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4014), 1,
      anon_sym_DOT,
    ACTIONS(4016), 1,
      anon_sym_POUND,
    ACTIONS(4018), 1,
      sym_hash_operator,
    STATE(2673), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2912), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2910), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97497] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1703), 1,
      anon_sym_DOT,
    STATE(2674), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1699), 13,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1701), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [97548] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2675), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1847), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1849), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97597] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_DOT,
    STATE(2676), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1699), 13,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1701), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [97648] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2677), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1651), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1653), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97697] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2678), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1745), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1747), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97746] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2679), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1721), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1723), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97795] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2680), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1827), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97844] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2681), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1683), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1685), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97893] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2682), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1687), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1689), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97942] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2683), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1709), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1711), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [97991] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2684), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1813), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1815), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98040] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2685), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1829), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1831), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98089] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2686), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1785), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1787), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98138] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2687), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1797), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1799), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98187] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2688), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1805), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1807), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98236] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2689), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1757), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1759), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98285] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2690), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1623), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1625), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98334] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2691), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1733), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1735), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98383] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2692), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1817), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1819), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98432] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2693), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1833), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1835), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98481] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1703), 1,
      anon_sym_DOT,
    STATE(2694), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1699), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1701), 23,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98532] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2695), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1713), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1715), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98581] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2696), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1809), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1811), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98630] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_DOT,
    STATE(2697), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1699), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1701), 23,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98681] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2698), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1843), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1845), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98730] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2699), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1725), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1727), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98779] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2700), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1771), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98828] = 25,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3604), 1,
      anon_sym_module,
    STATE(2701), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(6198), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [98917] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2702), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1801), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1803), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [98966] = 25,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3604), 1,
      anon_sym_module,
    STATE(2703), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5476), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [99055] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4014), 1,
      anon_sym_DOT,
    ACTIONS(4016), 1,
      anon_sym_POUND,
    ACTIONS(4018), 1,
      sym_hash_operator,
    STATE(2704), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2922), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2920), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99110] = 25,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(4020), 1,
      anon_sym_LPAREN,
    ACTIONS(4022), 1,
      anon_sym_private,
    STATE(2705), 1,
      sym_attribute,
    STATE(4185), 1,
      sym__simple_type,
    STATE(4463), 1,
      sym__type,
    STATE(4610), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6481), 1,
      sym_labeled_tuple_element_type,
    STATE(7168), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4300), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [99199] = 25,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(4020), 1,
      anon_sym_LPAREN,
    ACTIONS(4024), 1,
      anon_sym_private,
    STATE(2706), 1,
      sym_attribute,
    STATE(4185), 1,
      sym__simple_type,
    STATE(4477), 1,
      sym__type,
    STATE(4610), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6481), 1,
      sym_labeled_tuple_element_type,
    STATE(7168), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4300), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [99288] = 25,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(4026), 1,
      anon_sym_private,
    STATE(2707), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(4608), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [99377] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2708), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1659), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1661), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99426] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2709), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1647), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1649), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99475] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4014), 1,
      anon_sym_DOT,
    ACTIONS(4016), 1,
      anon_sym_POUND,
    ACTIONS(4018), 1,
      sym_hash_operator,
    STATE(2710), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2952), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2950), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99530] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4014), 1,
      anon_sym_DOT,
    ACTIONS(4016), 1,
      anon_sym_POUND,
    ACTIONS(4018), 1,
      sym_hash_operator,
    STATE(2711), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2968), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2966), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99585] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2712), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1559), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1561), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99634] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2713), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1705), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1707), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99683] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2714), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1563), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1565), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99732] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2715), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1539), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1541), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99781] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2716), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1765), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1767), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99830] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2717), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1773), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1775), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [99879] = 25,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3690), 1,
      anon_sym_module,
    STATE(2718), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5303), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [99968] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2719), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1667), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1669), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100017] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2720), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1729), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1731), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100066] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2721), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1821), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1823), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100115] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4028), 1,
      anon_sym_else,
    STATE(2136), 1,
      sym_else_clause,
    STATE(2722), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3271), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3269), 22,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_GT_RBRACE,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [100168] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2723), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1575), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1577), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100217] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2724), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1595), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1597), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100266] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3999), 1,
      aux_sym_attribute_id_token1,
    STATE(2670), 1,
      aux_sym_attribute_id_repeat1,
    STATE(2725), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2743), 15,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(2741), 17,
      anon_sym_rec,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_class,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [100319] = 25,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4030), 1,
      anon_sym_AMP,
    STATE(2726), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5024), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [100408] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2727), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1611), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1613), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100457] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2728), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1555), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1557), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100506] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2729), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1671), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1673), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100555] = 25,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3608), 1,
      anon_sym_module,
    STATE(2730), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5345), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [100644] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2731), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1675), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1677), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100693] = 25,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3690), 1,
      anon_sym_module,
    STATE(2732), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6312), 1,
      sym__type,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [100782] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2733), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1679), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1681), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [100831] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(4010), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4012), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(2734), 1,
      sym_attribute,
    STATE(3167), 1,
      sym__simple_module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4034), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4032), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [100890] = 25,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3692), 1,
      anon_sym_LPAREN,
    ACTIONS(4036), 1,
      anon_sym_private,
    STATE(2735), 1,
      sym_attribute,
    STATE(3209), 1,
      sym__simple_type,
    STATE(3476), 1,
      sym__type,
    STATE(3557), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6883), 1,
      sym_labeled_argument_type,
    STATE(6885), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [100979] = 25,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3692), 1,
      anon_sym_LPAREN,
    ACTIONS(4038), 1,
      anon_sym_private,
    STATE(2736), 1,
      sym_attribute,
    STATE(3209), 1,
      sym__simple_type,
    STATE(3478), 1,
      sym__type,
    STATE(3557), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6883), 1,
      sym_labeled_argument_type,
    STATE(6885), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [101068] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2737), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1851), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1853), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101117] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2738), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1717), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1719), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101166] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4028), 1,
      anon_sym_else,
    STATE(2117), 1,
      sym_else_clause,
    STATE(2739), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3277), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3275), 22,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_GT_RBRACE,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [101219] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2740), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1551), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1553), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101268] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2741), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1837), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1839), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101317] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2742), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1531), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1533), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101366] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2743), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1579), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1581), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101415] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4014), 1,
      anon_sym_DOT,
    STATE(2744), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1583), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1585), 23,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101466] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2745), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1587), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1589), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101515] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2746), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1543), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1545), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101564] = 25,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(3608), 1,
      anon_sym_module,
    STATE(2747), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(6178), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [101653] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4040), 1,
      anon_sym_else,
    STATE(2117), 1,
      sym_else_clause,
    STATE(2748), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3277), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3275), 24,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [101706] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2749), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1599), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1601), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101755] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2750), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1603), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1605), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101804] = 25,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3279), 1,
      sym__lowercase_identifier,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3696), 1,
      anon_sym_LPAREN,
    ACTIONS(3698), 1,
      anon_sym_LBRACK,
    ACTIONS(4042), 1,
      anon_sym_private,
    STATE(2751), 1,
      sym_attribute,
    STATE(2917), 1,
      sym__simple_type,
    STATE(3254), 1,
      sym__type,
    STATE(3350), 1,
      sym__proper_tuple_type,
    STATE(6244), 1,
      sym_extended_module_path,
    STATE(7252), 1,
      sym_labeled_argument_type,
    STATE(7344), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [101893] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2752), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1607), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1609), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101942] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2753), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1631), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1633), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [101991] = 25,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3279), 1,
      sym__lowercase_identifier,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3696), 1,
      anon_sym_LPAREN,
    ACTIONS(3698), 1,
      anon_sym_LBRACK,
    ACTIONS(4044), 1,
      anon_sym_private,
    STATE(2754), 1,
      sym_attribute,
    STATE(2917), 1,
      sym__simple_type,
    STATE(3256), 1,
      sym__type,
    STATE(3350), 1,
      sym__proper_tuple_type,
    STATE(6244), 1,
      sym_extended_module_path,
    STATE(7252), 1,
      sym_labeled_argument_type,
    STATE(7344), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [102080] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2755), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1635), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1637), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [102129] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2756), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1639), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1641), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [102178] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4040), 1,
      anon_sym_else,
    STATE(2136), 1,
      sym_else_clause,
    STATE(2757), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3271), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3269), 24,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [102231] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2758), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1741), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1743), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [102280] = 25,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2432), 1,
      sym__lowercase_identifier,
    ACTIONS(4046), 1,
      anon_sym_private,
    STATE(2759), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(4545), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [102369] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2760), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1643), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1645), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [102418] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2761), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1691), 10,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_POUND,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(1693), 24,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [102467] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4014), 1,
      anon_sym_DOT,
    ACTIONS(4016), 1,
      anon_sym_POUND,
    ACTIONS(4018), 1,
      sym_hash_operator,
    STATE(2762), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2948), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2946), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [102522] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2763), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5340), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [102608] = 24,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3696), 1,
      anon_sym_LPAREN,
    ACTIONS(3698), 1,
      anon_sym_LBRACK,
    ACTIONS(4048), 1,
      sym__lowercase_identifier,
    STATE(2764), 1,
      sym_attribute,
    STATE(2917), 1,
      sym__simple_type,
    STATE(3262), 1,
      sym__type,
    STATE(3350), 1,
      sym__proper_tuple_type,
    STATE(6244), 1,
      sym_extended_module_path,
    STATE(7252), 1,
      sym_labeled_argument_type,
    STATE(7344), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [102694] = 24,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3696), 1,
      anon_sym_LPAREN,
    ACTIONS(3698), 1,
      anon_sym_LBRACK,
    ACTIONS(4048), 1,
      sym__lowercase_identifier,
    STATE(2648), 1,
      sym__type,
    STATE(2765), 1,
      sym_attribute,
    STATE(2917), 1,
      sym__simple_type,
    STATE(3350), 1,
      sym__proper_tuple_type,
    STATE(6244), 1,
      sym_extended_module_path,
    STATE(7252), 1,
      sym_labeled_argument_type,
    STATE(7344), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [102780] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2766), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5344), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [102866] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2767), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(6054), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [102952] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2768), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1713), 13,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1715), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [103000] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2769), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1843), 13,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1845), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [103048] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2770), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1623), 13,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1625), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [103096] = 24,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3696), 1,
      anon_sym_LPAREN,
    ACTIONS(3698), 1,
      anon_sym_LBRACK,
    ACTIONS(4048), 1,
      sym__lowercase_identifier,
    STATE(2651), 1,
      sym__type,
    STATE(2771), 1,
      sym_attribute,
    STATE(2917), 1,
      sym__simple_type,
    STATE(3350), 1,
      sym__proper_tuple_type,
    STATE(6244), 1,
      sym_extended_module_path,
    STATE(7252), 1,
      sym_labeled_argument_type,
    STATE(7344), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103182] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2772), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1809), 13,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_in,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1811), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [103230] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2773), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5184), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103316] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2651), 1,
      sym__type,
    STATE(2774), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103402] = 24,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3696), 1,
      anon_sym_LPAREN,
    ACTIONS(3698), 1,
      anon_sym_LBRACK,
    ACTIONS(4048), 1,
      sym__lowercase_identifier,
    STATE(2775), 1,
      sym_attribute,
    STATE(2917), 1,
      sym__simple_type,
    STATE(3350), 1,
      sym__proper_tuple_type,
    STATE(3896), 1,
      sym__type,
    STATE(6244), 1,
      sym_extended_module_path,
    STATE(7252), 1,
      sym_labeled_argument_type,
    STATE(7344), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103488] = 24,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3696), 1,
      anon_sym_LPAREN,
    ACTIONS(3698), 1,
      anon_sym_LBRACK,
    ACTIONS(4048), 1,
      sym__lowercase_identifier,
    STATE(2776), 1,
      sym_attribute,
    STATE(2917), 1,
      sym__simple_type,
    STATE(3350), 1,
      sym__proper_tuple_type,
    STATE(3897), 1,
      sym__type,
    STATE(6244), 1,
      sym_extended_module_path,
    STATE(7252), 1,
      sym_labeled_argument_type,
    STATE(7344), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103574] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2777), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2724), 16,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
      aux_sym_attribute_id_token1,
    ACTIONS(2722), 17,
      anon_sym_rec,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_class,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [103622] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2778), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(4545), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103708] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2779), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(4593), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103794] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4020), 1,
      anon_sym_LPAREN,
    STATE(2780), 1,
      sym_attribute,
    STATE(4185), 1,
      sym__simple_type,
    STATE(4245), 1,
      sym__type,
    STATE(4610), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6481), 1,
      sym_labeled_tuple_element_type,
    STATE(7168), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4300), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103880] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3692), 1,
      anon_sym_LPAREN,
    STATE(2781), 1,
      sym_attribute,
    STATE(3209), 1,
      sym__simple_type,
    STATE(3557), 1,
      sym__proper_tuple_type,
    STATE(4139), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6883), 1,
      sym_labeled_argument_type,
    STATE(6885), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [103966] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3692), 1,
      anon_sym_LPAREN,
    STATE(2782), 1,
      sym_attribute,
    STATE(3209), 1,
      sym__simple_type,
    STATE(3557), 1,
      sym__proper_tuple_type,
    STATE(4140), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6883), 1,
      sym_labeled_argument_type,
    STATE(6885), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [104052] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4020), 1,
      anon_sym_LPAREN,
    STATE(2783), 1,
      sym_attribute,
    STATE(4185), 1,
      sym__simple_type,
    STATE(4265), 1,
      sym__type,
    STATE(4610), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6481), 1,
      sym_labeled_tuple_element_type,
    STATE(7168), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4300), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [104138] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4020), 1,
      anon_sym_LPAREN,
    STATE(2784), 1,
      sym_attribute,
    STATE(4185), 1,
      sym__simple_type,
    STATE(4477), 1,
      sym__type,
    STATE(4610), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6481), 1,
      sym_labeled_tuple_element_type,
    STATE(7168), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4300), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [104224] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4020), 1,
      anon_sym_LPAREN,
    STATE(2785), 1,
      sym_attribute,
    STATE(4185), 1,
      sym__simple_type,
    STATE(4482), 1,
      sym__type,
    STATE(4610), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6481), 1,
      sym_labeled_tuple_element_type,
    STATE(7168), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4300), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [104310] = 24,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3664), 1,
      anon_sym_LPAREN,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3672), 1,
      anon_sym_LT,
    ACTIONS(3674), 1,
      anon_sym_POUND,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4050), 1,
      sym__lowercase_identifier,
    STATE(2786), 1,
      sym_attribute,
    STATE(4245), 1,
      sym__type,
    STATE(4297), 1,
      sym__simple_type,
    STATE(4792), 1,
      sym__proper_tuple_type,
    STATE(5999), 1,
      sym_extended_module_path,
    STATE(7131), 1,
      sym_labeled_argument_type,
    STATE(7170), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4300), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4452), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [104396] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2648), 1,
      sym__type,
    STATE(2787), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [104482] = 24,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3664), 1,
      anon_sym_LPAREN,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3672), 1,
      anon_sym_LT,
    ACTIONS(3674), 1,
      anon_sym_POUND,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4050), 1,
      sym__lowercase_identifier,
    STATE(2788), 1,
      sym_attribute,
    STATE(4297), 1,
      sym__simple_type,
    STATE(4792), 1,
      sym__proper_tuple_type,
    STATE(5464), 1,
      sym__type,
    STATE(5999), 1,
      sym_extended_module_path,
    STATE(7131), 1,
      sym_labeled_argument_type,
    STATE(7170), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4300), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4452), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [104568] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3692), 1,
      anon_sym_LPAREN,
    STATE(2789), 1,
      sym_attribute,
    STATE(3209), 1,
      sym__simple_type,
    STATE(3262), 1,
      sym__type,
    STATE(3557), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6883), 1,
      sym_labeled_argument_type,
    STATE(6885), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [104654] = 24,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3664), 1,
      anon_sym_LPAREN,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3672), 1,
      anon_sym_LT,
    ACTIONS(3674), 1,
      anon_sym_POUND,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4050), 1,
      sym__lowercase_identifier,
    STATE(2790), 1,
      sym_attribute,
    STATE(4297), 1,
      sym__simple_type,
    STATE(4792), 1,
      sym__proper_tuple_type,
    STATE(5479), 1,
      sym__type,
    STATE(5999), 1,
      sym_extended_module_path,
    STATE(7131), 1,
      sym_labeled_argument_type,
    STATE(7170), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4300), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4452), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [104740] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2791), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5017), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [104826] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3706), 1,
      anon_sym_LPAREN,
    STATE(2792), 1,
      sym_attribute,
    STATE(3843), 1,
      sym__simple_type,
    STATE(4236), 1,
      sym__proper_tuple_type,
    STATE(4294), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(7048), 1,
      sym_labeled_argument_type,
    STATE(7050), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [104912] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4052), 1,
      anon_sym_COMMA,
    ACTIONS(4054), 1,
      anon_sym_SEMI,
    ACTIONS(4056), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4058), 1,
      sym_concat_operator,
    ACTIONS(4060), 1,
      sym_rel_operator,
    ACTIONS(4062), 1,
      sym_assign_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2793), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2710), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2708), 4,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [104996] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4052), 1,
      anon_sym_COMMA,
    ACTIONS(4054), 1,
      anon_sym_SEMI,
    ACTIONS(4056), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4058), 1,
      sym_concat_operator,
    ACTIONS(4060), 1,
      sym_rel_operator,
    ACTIONS(4062), 1,
      sym_assign_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2794), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2718), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2716), 4,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [105080] = 24,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3664), 1,
      anon_sym_LPAREN,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3672), 1,
      anon_sym_LT,
    ACTIONS(3674), 1,
      anon_sym_POUND,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4050), 1,
      sym__lowercase_identifier,
    STATE(2795), 1,
      sym_attribute,
    STATE(4265), 1,
      sym__type,
    STATE(4297), 1,
      sym__simple_type,
    STATE(4792), 1,
      sym__proper_tuple_type,
    STATE(5999), 1,
      sym_extended_module_path,
    STATE(7131), 1,
      sym_labeled_argument_type,
    STATE(7170), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4300), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4452), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [105166] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2796), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2874), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2872), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [105224] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2797), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [105284] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2798), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [105344] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2799), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 8,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [105408] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4052), 1,
      anon_sym_COMMA,
    ACTIONS(4054), 1,
      anon_sym_SEMI,
    ACTIONS(4056), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4058), 1,
      sym_concat_operator,
    ACTIONS(4060), 1,
      sym_rel_operator,
    ACTIONS(4062), 1,
      sym_assign_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2800), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2854), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2852), 4,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [105492] = 24,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3664), 1,
      anon_sym_LPAREN,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3672), 1,
      anon_sym_LT,
    ACTIONS(3674), 1,
      anon_sym_POUND,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4050), 1,
      sym__lowercase_identifier,
    STATE(2801), 1,
      sym_attribute,
    STATE(4297), 1,
      sym__simple_type,
    STATE(4792), 1,
      sym__proper_tuple_type,
    STATE(5299), 1,
      sym__type,
    STATE(5999), 1,
      sym_extended_module_path,
    STATE(7131), 1,
      sym_labeled_argument_type,
    STATE(7170), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4300), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4452), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [105578] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2802), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(4249), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [105664] = 24,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3664), 1,
      anon_sym_LPAREN,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3672), 1,
      anon_sym_LT,
    ACTIONS(3674), 1,
      anon_sym_POUND,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4050), 1,
      sym__lowercase_identifier,
    STATE(2803), 1,
      sym_attribute,
    STATE(4297), 1,
      sym__simple_type,
    STATE(4792), 1,
      sym__proper_tuple_type,
    STATE(5214), 1,
      sym__type,
    STATE(5999), 1,
      sym_extended_module_path,
    STATE(7131), 1,
      sym_labeled_argument_type,
    STATE(7170), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4300), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4452), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [105750] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2804), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(4230), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [105836] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2805), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5529), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [105922] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2806), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(4234), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106008] = 24,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3696), 1,
      anon_sym_LPAREN,
    ACTIONS(3698), 1,
      anon_sym_LBRACK,
    ACTIONS(4048), 1,
      sym__lowercase_identifier,
    STATE(2807), 1,
      sym_attribute,
    STATE(2917), 1,
      sym__simple_type,
    STATE(3258), 1,
      sym__type,
    STATE(3350), 1,
      sym__proper_tuple_type,
    STATE(6244), 1,
      sym_extended_module_path,
    STATE(7252), 1,
      sym_labeled_argument_type,
    STATE(7344), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106094] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2808), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4214), 1,
      sym__type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106180] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2809), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6306), 1,
      sym__type,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106266] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3692), 1,
      anon_sym_LPAREN,
    STATE(2651), 1,
      sym__type,
    STATE(2810), 1,
      sym_attribute,
    STATE(3209), 1,
      sym__simple_type,
    STATE(3557), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6883), 1,
      sym_labeled_argument_type,
    STATE(6885), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106352] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3706), 1,
      anon_sym_LPAREN,
    STATE(2811), 1,
      sym_attribute,
    STATE(3843), 1,
      sym__simple_type,
    STATE(4230), 1,
      sym__type,
    STATE(4236), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(7048), 1,
      sym_labeled_argument_type,
    STATE(7050), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106438] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3706), 1,
      anon_sym_LPAREN,
    STATE(2812), 1,
      sym_attribute,
    STATE(3843), 1,
      sym__simple_type,
    STATE(4234), 1,
      sym__type,
    STATE(4236), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(7048), 1,
      sym_labeled_argument_type,
    STATE(7050), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106524] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3692), 1,
      anon_sym_LPAREN,
    STATE(2648), 1,
      sym__type,
    STATE(2813), 1,
      sym_attribute,
    STATE(3209), 1,
      sym__simple_type,
    STATE(3557), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6883), 1,
      sym_labeled_argument_type,
    STATE(6885), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106610] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3706), 1,
      anon_sym_LPAREN,
    STATE(2814), 1,
      sym_attribute,
    STATE(3843), 1,
      sym__simple_type,
    STATE(4236), 1,
      sym__proper_tuple_type,
    STATE(4249), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(7048), 1,
      sym_labeled_argument_type,
    STATE(7050), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106696] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3706), 1,
      anon_sym_LPAREN,
    STATE(2815), 1,
      sym_attribute,
    STATE(3843), 1,
      sym__simple_type,
    STATE(4214), 1,
      sym__type,
    STATE(4236), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(7048), 1,
      sym_labeled_argument_type,
    STATE(7050), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106782] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2816), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(6208), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106868] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2817), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5311), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [106954] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3692), 1,
      anon_sym_LPAREN,
    STATE(2818), 1,
      sym_attribute,
    STATE(3209), 1,
      sym__simple_type,
    STATE(3478), 1,
      sym__type,
    STATE(3557), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6883), 1,
      sym_labeled_argument_type,
    STATE(6885), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [107040] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3692), 1,
      anon_sym_LPAREN,
    STATE(2819), 1,
      sym_attribute,
    STATE(3209), 1,
      sym__simple_type,
    STATE(3479), 1,
      sym__type,
    STATE(3557), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6883), 1,
      sym_labeled_argument_type,
    STATE(6885), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [107126] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4052), 1,
      anon_sym_COMMA,
    ACTIONS(4056), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4058), 1,
      sym_concat_operator,
    ACTIONS(4060), 1,
      sym_rel_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2820), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2880), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2878), 6,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_assign_operator,
      sym_let_and_operator,
  [107206] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4056), 1,
      anon_sym_COLON_COLON,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2821), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2884), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 10,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [107274] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4056), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4058), 1,
      sym_concat_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2822), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [107344] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4056), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4058), 1,
      sym_concat_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2823), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [107414] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4052), 1,
      anon_sym_COMMA,
    ACTIONS(4056), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4058), 1,
      sym_concat_operator,
    ACTIONS(4060), 1,
      sym_rel_operator,
    ACTIONS(4062), 1,
      sym_assign_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2824), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2862), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [107496] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4056), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4058), 1,
      sym_concat_operator,
    ACTIONS(4060), 1,
      sym_rel_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2825), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2862), 3,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_PIPE_PIPE,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 8,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [107570] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4056), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4058), 1,
      sym_concat_operator,
    ACTIONS(4060), 1,
      sym_rel_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2826), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2862), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_assign_operator,
      sym_let_and_operator,
  [107648] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4052), 1,
      anon_sym_COMMA,
    ACTIONS(4056), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4058), 1,
      sym_concat_operator,
    ACTIONS(4060), 1,
      sym_rel_operator,
    ACTIONS(4062), 1,
      sym_assign_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2827), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2826), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2824), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [107730] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4052), 1,
      anon_sym_COMMA,
    ACTIONS(4056), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4058), 1,
      sym_concat_operator,
    ACTIONS(4060), 1,
      sym_rel_operator,
    ACTIONS(4062), 1,
      sym_assign_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2828), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2866), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2864), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [107812] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4052), 1,
      anon_sym_COMMA,
    ACTIONS(4056), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4058), 1,
      sym_concat_operator,
    ACTIONS(4060), 1,
      sym_rel_operator,
    ACTIONS(4062), 1,
      sym_assign_operator,
    STATE(633), 1,
      sym_pow_operator,
    STATE(634), 1,
      sym_mult_operator,
    STATE(635), 1,
      sym_add_operator,
    STATE(636), 1,
      sym_and_operator,
    STATE(637), 1,
      sym_or_operator,
    STATE(2829), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2890), 2,
      anon_sym_and,
      anon_sym_PIPE,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2888), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [107894] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2830), 1,
      sym_attribute,
    STATE(3262), 1,
      sym__type,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [107980] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3706), 1,
      anon_sym_LPAREN,
    STATE(2651), 1,
      sym__type,
    STATE(2831), 1,
      sym_attribute,
    STATE(3843), 1,
      sym__simple_type,
    STATE(4236), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(7048), 1,
      sym_labeled_argument_type,
    STATE(7050), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [108066] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2832), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5972), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [108152] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3706), 1,
      anon_sym_LPAREN,
    STATE(2648), 1,
      sym__type,
    STATE(2833), 1,
      sym_attribute,
    STATE(3843), 1,
      sym__simple_type,
    STATE(4236), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(7048), 1,
      sym_labeled_argument_type,
    STATE(7050), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [108238] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2834), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(6164), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [108324] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4020), 1,
      anon_sym_LPAREN,
    STATE(2835), 1,
      sym_attribute,
    STATE(4185), 1,
      sym__simple_type,
    STATE(4556), 1,
      sym__type,
    STATE(4610), 1,
      sym__proper_tuple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6481), 1,
      sym_labeled_tuple_element_type,
    STATE(7168), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4300), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [108410] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3646), 1,
      anon_sym_COMMA,
    STATE(2836), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2880), 8,
      anon_sym_AMP,
      anon_sym_LBRACK_AT_AT,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2878), 24,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [108460] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2837), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(6230), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [108546] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2838), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5388), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [108632] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2839), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5362), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [108718] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2840), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5985), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [108804] = 24,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3696), 1,
      anon_sym_LPAREN,
    ACTIONS(3698), 1,
      anon_sym_LBRACK,
    ACTIONS(4048), 1,
      sym__lowercase_identifier,
    STATE(2841), 1,
      sym_attribute,
    STATE(2917), 1,
      sym__simple_type,
    STATE(3256), 1,
      sym__type,
    STATE(3350), 1,
      sym__proper_tuple_type,
    STATE(6244), 1,
      sym_extended_module_path,
    STATE(7252), 1,
      sym_labeled_argument_type,
    STATE(7344), 1,
      sym_labeled_tuple_element_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [108890] = 24,
    ACTIONS(7), 1,
      sym__lowercase_identifier,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(11), 1,
      anon_sym_LPAREN,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2842), 1,
      sym_attribute,
    STATE(3803), 1,
      sym__simple_type,
    STATE(4229), 1,
      sym__proper_tuple_type,
    STATE(5455), 1,
      sym__type,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6978), 1,
      sym_labeled_tuple_element_type,
    STATE(7246), 1,
      sym_labeled_argument_type,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2650), 3,
      sym_function_type,
      sym__labeled_tuple_type,
      sym_aliased_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [108976] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    STATE(2843), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2880), 10,
      anon_sym_COLON2,
      anon_sym_DASH_GT,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      sym_rel_operator,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2878), 22,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_COLON_COLON,
      anon_sym_then,
      anon_sym_else,
      anon_sym_GT_RBRACE,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      anon_sym_or,
      sym_assign_operator,
  [109026] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2890), 1,
      anon_sym_and,
    ACTIONS(4064), 1,
      anon_sym_COMMA,
    ACTIONS(4066), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4068), 1,
      sym_concat_operator,
    ACTIONS(4070), 1,
      sym_rel_operator,
    ACTIONS(4072), 1,
      sym_assign_operator,
    STATE(705), 1,
      sym_pow_operator,
    STATE(706), 1,
      sym_mult_operator,
    STATE(707), 1,
      sym_add_operator,
    STATE(708), 1,
      sym_and_operator,
    STATE(709), 1,
      sym_or_operator,
    STATE(2844), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2888), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [109107] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4074), 1,
      sym__lowercase_identifier,
    ACTIONS(4076), 1,
      anon_sym_POUND,
    STATE(2845), 1,
      sym_attribute,
    STATE(2981), 1,
      sym_type_constructor_path,
    STATE(6029), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3620), 11,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(3624), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [109164] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4078), 1,
      anon_sym_COLON,
    STATE(2846), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3221), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3219), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [109213] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(565), 1,
      sym_pow_operator,
    STATE(566), 1,
      sym_mult_operator,
    STATE(567), 1,
      sym_add_operator,
    STATE(568), 1,
      sym_and_operator,
    STATE(569), 1,
      sym_or_operator,
    STATE(2847), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2874), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2872), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [109270] = 23,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2710), 1,
      anon_sym_PIPE,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4080), 1,
      anon_sym_COMMA,
    ACTIONS(4082), 1,
      anon_sym_SEMI,
    ACTIONS(4084), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4086), 1,
      sym_concat_operator,
    ACTIONS(4088), 1,
      sym_rel_operator,
    ACTIONS(4090), 1,
      sym_assign_operator,
    STATE(565), 1,
      sym_pow_operator,
    STATE(566), 1,
      sym_mult_operator,
    STATE(567), 1,
      sym_add_operator,
    STATE(568), 1,
      sym_and_operator,
    STATE(569), 1,
      sym_or_operator,
    STATE(2848), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2708), 4,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [109353] = 23,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2718), 1,
      anon_sym_PIPE,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4080), 1,
      anon_sym_COMMA,
    ACTIONS(4082), 1,
      anon_sym_SEMI,
    ACTIONS(4084), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4086), 1,
      sym_concat_operator,
    ACTIONS(4088), 1,
      sym_rel_operator,
    ACTIONS(4090), 1,
      sym_assign_operator,
    STATE(565), 1,
      sym_pow_operator,
    STATE(566), 1,
      sym_mult_operator,
    STATE(567), 1,
      sym_add_operator,
    STATE(568), 1,
      sym_and_operator,
    STATE(569), 1,
      sym_or_operator,
    STATE(2849), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2716), 4,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [109436] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 1,
      anon_sym_and,
    ACTIONS(4064), 1,
      anon_sym_COMMA,
    ACTIONS(4066), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4068), 1,
      sym_concat_operator,
    ACTIONS(4070), 1,
      sym_rel_operator,
    STATE(705), 1,
      sym_pow_operator,
    STATE(706), 1,
      sym_mult_operator,
    STATE(707), 1,
      sym_add_operator,
    STATE(708), 1,
      sym_and_operator,
    STATE(709), 1,
      sym_or_operator,
    STATE(2850), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2878), 6,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_assign_operator,
      sym_let_and_operator,
  [109515] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4066), 1,
      anon_sym_COLON_COLON,
    STATE(705), 1,
      sym_pow_operator,
    STATE(706), 1,
      sym_mult_operator,
    STATE(707), 1,
      sym_add_operator,
    STATE(708), 1,
      sym_and_operator,
    STATE(709), 1,
      sym_or_operator,
    STATE(2851), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2884), 4,
      anon_sym_and,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 10,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [109582] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4066), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4068), 1,
      sym_concat_operator,
    STATE(705), 1,
      sym_pow_operator,
    STATE(706), 1,
      sym_mult_operator,
    STATE(707), 1,
      sym_add_operator,
    STATE(708), 1,
      sym_and_operator,
    STATE(709), 1,
      sym_or_operator,
    STATE(2852), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 4,
      anon_sym_and,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [109651] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4066), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4068), 1,
      sym_concat_operator,
    STATE(705), 1,
      sym_pow_operator,
    STATE(706), 1,
      sym_mult_operator,
    STATE(707), 1,
      sym_add_operator,
    STATE(708), 1,
      sym_and_operator,
    STATE(709), 1,
      sym_or_operator,
    STATE(2853), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 4,
      anon_sym_and,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [109720] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2862), 1,
      anon_sym_and,
    ACTIONS(4064), 1,
      anon_sym_COMMA,
    ACTIONS(4066), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4068), 1,
      sym_concat_operator,
    ACTIONS(4070), 1,
      sym_rel_operator,
    ACTIONS(4072), 1,
      sym_assign_operator,
    STATE(705), 1,
      sym_pow_operator,
    STATE(706), 1,
      sym_mult_operator,
    STATE(707), 1,
      sym_add_operator,
    STATE(708), 1,
      sym_and_operator,
    STATE(709), 1,
      sym_or_operator,
    STATE(2854), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [109801] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(565), 1,
      sym_pow_operator,
    STATE(566), 1,
      sym_mult_operator,
    STATE(567), 1,
      sym_add_operator,
    STATE(568), 1,
      sym_and_operator,
    STATE(569), 1,
      sym_or_operator,
    STATE(2855), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [109860] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(565), 1,
      sym_pow_operator,
    STATE(566), 1,
      sym_mult_operator,
    STATE(567), 1,
      sym_add_operator,
    STATE(568), 1,
      sym_and_operator,
    STATE(569), 1,
      sym_or_operator,
    STATE(2856), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [109919] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    STATE(565), 1,
      sym_pow_operator,
    STATE(566), 1,
      sym_mult_operator,
    STATE(567), 1,
      sym_add_operator,
    STATE(568), 1,
      sym_and_operator,
    STATE(569), 1,
      sym_or_operator,
    STATE(2857), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 7,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [109982] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4066), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4068), 1,
      sym_concat_operator,
    ACTIONS(4070), 1,
      sym_rel_operator,
    STATE(705), 1,
      sym_pow_operator,
    STATE(706), 1,
      sym_mult_operator,
    STATE(707), 1,
      sym_add_operator,
    STATE(708), 1,
      sym_and_operator,
    STATE(709), 1,
      sym_or_operator,
    STATE(2858), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2862), 2,
      anon_sym_and,
      anon_sym_PIPE_PIPE,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 8,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [110055] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2862), 1,
      anon_sym_and,
    ACTIONS(4066), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4068), 1,
      sym_concat_operator,
    ACTIONS(4070), 1,
      sym_rel_operator,
    STATE(705), 1,
      sym_pow_operator,
    STATE(706), 1,
      sym_mult_operator,
    STATE(707), 1,
      sym_add_operator,
    STATE(708), 1,
      sym_and_operator,
    STATE(709), 1,
      sym_or_operator,
    STATE(2859), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_assign_operator,
      sym_let_and_operator,
  [110132] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      anon_sym_and,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4064), 1,
      anon_sym_COMMA,
    ACTIONS(4066), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4068), 1,
      sym_concat_operator,
    ACTIONS(4070), 1,
      sym_rel_operator,
    ACTIONS(4072), 1,
      sym_assign_operator,
    STATE(705), 1,
      sym_pow_operator,
    STATE(706), 1,
      sym_mult_operator,
    STATE(707), 1,
      sym_add_operator,
    STATE(708), 1,
      sym_and_operator,
    STATE(709), 1,
      sym_or_operator,
    STATE(2860), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2824), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [110213] = 22,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2866), 1,
      anon_sym_and,
    ACTIONS(4064), 1,
      anon_sym_COMMA,
    ACTIONS(4066), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4068), 1,
      sym_concat_operator,
    ACTIONS(4070), 1,
      sym_rel_operator,
    ACTIONS(4072), 1,
      sym_assign_operator,
    STATE(705), 1,
      sym_pow_operator,
    STATE(706), 1,
      sym_mult_operator,
    STATE(707), 1,
      sym_add_operator,
    STATE(708), 1,
      sym_and_operator,
    STATE(709), 1,
      sym_or_operator,
    STATE(2861), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2864), 5,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [110294] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2854), 1,
      anon_sym_and,
    ACTIONS(4064), 1,
      anon_sym_COMMA,
    ACTIONS(4066), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4068), 1,
      sym_concat_operator,
    ACTIONS(4070), 1,
      sym_rel_operator,
    ACTIONS(4072), 1,
      sym_assign_operator,
    ACTIONS(4092), 1,
      anon_sym_SEMI,
    STATE(705), 1,
      sym_pow_operator,
    STATE(706), 1,
      sym_mult_operator,
    STATE(707), 1,
      sym_add_operator,
    STATE(708), 1,
      sym_and_operator,
    STATE(709), 1,
      sym_or_operator,
    STATE(2862), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2852), 4,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [110377] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2863), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3947), 15,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
    ACTIONS(3945), 17,
      anon_sym_rec,
      anon_sym_external,
      anon_sym_type,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_class,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [110424] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4094), 1,
      aux_sym_attribute_id_token1,
    STATE(2864), 1,
      sym_attribute,
    STATE(2883), 1,
      aux_sym_attribute_id_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2743), 9,
      anon_sym_LPAREN,
      anon_sym_PLUS,
      anon_sym_BANG,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
    ACTIONS(2741), 21,
      anon_sym_rec,
      anon_sym_nonrec,
      anon_sym_DASH,
      anon_sym__,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_constraint,
      anon_sym_module,
      anon_sym_virtual,
      anon_sym_val,
      anon_sym_sig,
      anon_sym_end,
      anon_sym_functor,
      anon_sym_struct,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [110475] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4084), 1,
      anon_sym_COLON_COLON,
    STATE(565), 1,
      sym_pow_operator,
    STATE(566), 1,
      sym_mult_operator,
    STATE(567), 1,
      sym_add_operator,
    STATE(568), 1,
      sym_and_operator,
    STATE(569), 1,
      sym_or_operator,
    STATE(2865), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2884), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 10,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [110542] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4084), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4086), 1,
      sym_concat_operator,
    STATE(565), 1,
      sym_pow_operator,
    STATE(566), 1,
      sym_mult_operator,
    STATE(567), 1,
      sym_add_operator,
    STATE(568), 1,
      sym_and_operator,
    STATE(569), 1,
      sym_or_operator,
    STATE(2866), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [110611] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4084), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4086), 1,
      sym_concat_operator,
    STATE(565), 1,
      sym_pow_operator,
    STATE(566), 1,
      sym_mult_operator,
    STATE(567), 1,
      sym_add_operator,
    STATE(568), 1,
      sym_and_operator,
    STATE(569), 1,
      sym_or_operator,
    STATE(2867), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [110680] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2862), 1,
      anon_sym_PIPE,
    ACTIONS(4080), 1,
      anon_sym_COMMA,
    ACTIONS(4084), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4086), 1,
      sym_concat_operator,
    ACTIONS(4088), 1,
      sym_rel_operator,
    ACTIONS(4090), 1,
      sym_assign_operator,
    STATE(565), 1,
      sym_pow_operator,
    STATE(566), 1,
      sym_mult_operator,
    STATE(567), 1,
      sym_add_operator,
    STATE(568), 1,
      sym_and_operator,
    STATE(569), 1,
      sym_or_operator,
    STATE(2868), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [110761] = 18,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4084), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4086), 1,
      sym_concat_operator,
    ACTIONS(4088), 1,
      sym_rel_operator,
    STATE(565), 1,
      sym_pow_operator,
    STATE(566), 1,
      sym_mult_operator,
    STATE(567), 1,
      sym_add_operator,
    STATE(568), 1,
      sym_and_operator,
    STATE(569), 1,
      sym_or_operator,
    STATE(2869), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2862), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_PIPE,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 8,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_or,
      sym_assign_operator,
  [110834] = 20,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2862), 1,
      anon_sym_PIPE,
    ACTIONS(4084), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4086), 1,
      sym_concat_operator,
    ACTIONS(4088), 1,
      sym_rel_operator,
    STATE(565), 1,
      sym_pow_operator,
    STATE(566), 1,
      sym_mult_operator,
    STATE(567), 1,
      sym_add_operator,
    STATE(568), 1,
      sym_and_operator,
    STATE(569), 1,
      sym_or_operator,
    STATE(2870), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_assign_operator,
  [110911] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      anon_sym_PIPE,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4080), 1,
      anon_sym_COMMA,
    ACTIONS(4084), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4086), 1,
      sym_concat_operator,
    ACTIONS(4088), 1,
      sym_rel_operator,
    ACTIONS(4090), 1,
      sym_assign_operator,
    STATE(565), 1,
      sym_pow_operator,
    STATE(566), 1,
      sym_mult_operator,
    STATE(567), 1,
      sym_add_operator,
    STATE(568), 1,
      sym_and_operator,
    STATE(569), 1,
      sym_or_operator,
    STATE(2871), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2824), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [110992] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2866), 1,
      anon_sym_PIPE,
    ACTIONS(4080), 1,
      anon_sym_COMMA,
    ACTIONS(4084), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4086), 1,
      sym_concat_operator,
    ACTIONS(4088), 1,
      sym_rel_operator,
    ACTIONS(4090), 1,
      sym_assign_operator,
    STATE(565), 1,
      sym_pow_operator,
    STATE(566), 1,
      sym_mult_operator,
    STATE(567), 1,
      sym_add_operator,
    STATE(568), 1,
      sym_and_operator,
    STATE(569), 1,
      sym_or_operator,
    STATE(2872), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2864), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [111073] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2890), 1,
      anon_sym_PIPE,
    ACTIONS(4080), 1,
      anon_sym_COMMA,
    ACTIONS(4084), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4086), 1,
      sym_concat_operator,
    ACTIONS(4088), 1,
      sym_rel_operator,
    ACTIONS(4090), 1,
      sym_assign_operator,
    STATE(565), 1,
      sym_pow_operator,
    STATE(566), 1,
      sym_mult_operator,
    STATE(567), 1,
      sym_add_operator,
    STATE(568), 1,
      sym_and_operator,
    STATE(569), 1,
      sym_or_operator,
    STATE(2873), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2888), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [111154] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2874), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3638), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
    ACTIONS(3640), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [111201] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4096), 1,
      aux_sym_attribute_id_token1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2875), 2,
      sym_attribute,
      aux_sym_attribute_id_repeat1,
    ACTIONS(2724), 9,
      anon_sym_LPAREN,
      anon_sym_PLUS,
      anon_sym_BANG,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
    ACTIONS(2722), 21,
      anon_sym_rec,
      anon_sym_nonrec,
      anon_sym_DASH,
      anon_sym__,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_constraint,
      anon_sym_module,
      anon_sym_virtual,
      anon_sym_val,
      anon_sym_sig,
      anon_sym_end,
      anon_sym_functor,
      anon_sym_struct,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [111250] = 23,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2854), 1,
      anon_sym_PIPE,
    ACTIONS(4080), 1,
      anon_sym_COMMA,
    ACTIONS(4082), 1,
      anon_sym_SEMI,
    ACTIONS(4084), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4086), 1,
      sym_concat_operator,
    ACTIONS(4088), 1,
      sym_rel_operator,
    ACTIONS(4090), 1,
      sym_assign_operator,
    STATE(565), 1,
      sym_pow_operator,
    STATE(566), 1,
      sym_mult_operator,
    STATE(567), 1,
      sym_add_operator,
    STATE(568), 1,
      sym_and_operator,
    STATE(569), 1,
      sym_or_operator,
    STATE(2876), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2852), 4,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [111333] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(705), 1,
      sym_pow_operator,
    STATE(706), 1,
      sym_mult_operator,
    STATE(707), 1,
      sym_add_operator,
    STATE(708), 1,
      sym_and_operator,
    STATE(709), 1,
      sym_or_operator,
    STATE(2877), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2874), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2872), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [111390] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(705), 1,
      sym_pow_operator,
    STATE(706), 1,
      sym_mult_operator,
    STATE(707), 1,
      sym_add_operator,
    STATE(708), 1,
      sym_and_operator,
    STATE(709), 1,
      sym_or_operator,
    STATE(2878), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [111449] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(705), 1,
      sym_pow_operator,
    STATE(706), 1,
      sym_mult_operator,
    STATE(707), 1,
      sym_add_operator,
    STATE(708), 1,
      sym_and_operator,
    STATE(709), 1,
      sym_or_operator,
    STATE(2879), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [111508] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    STATE(705), 1,
      sym_pow_operator,
    STATE(706), 1,
      sym_mult_operator,
    STATE(707), 1,
      sym_add_operator,
    STATE(708), 1,
      sym_and_operator,
    STATE(709), 1,
      sym_or_operator,
    STATE(2880), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 7,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [111571] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2710), 1,
      anon_sym_and,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4064), 1,
      anon_sym_COMMA,
    ACTIONS(4066), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4068), 1,
      sym_concat_operator,
    ACTIONS(4070), 1,
      sym_rel_operator,
    ACTIONS(4072), 1,
      sym_assign_operator,
    ACTIONS(4092), 1,
      anon_sym_SEMI,
    STATE(705), 1,
      sym_pow_operator,
    STATE(706), 1,
      sym_mult_operator,
    STATE(707), 1,
      sym_add_operator,
    STATE(708), 1,
      sym_and_operator,
    STATE(709), 1,
      sym_or_operator,
    STATE(2881), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2708), 4,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [111654] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2718), 1,
      anon_sym_and,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4064), 1,
      anon_sym_COMMA,
    ACTIONS(4066), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4068), 1,
      sym_concat_operator,
    ACTIONS(4070), 1,
      sym_rel_operator,
    ACTIONS(4072), 1,
      sym_assign_operator,
    ACTIONS(4092), 1,
      anon_sym_SEMI,
    STATE(705), 1,
      sym_pow_operator,
    STATE(706), 1,
      sym_mult_operator,
    STATE(707), 1,
      sym_add_operator,
    STATE(708), 1,
      sym_and_operator,
    STATE(709), 1,
      sym_or_operator,
    STATE(2882), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2716), 4,
      anon_sym_in,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [111737] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4094), 1,
      aux_sym_attribute_id_token1,
    STATE(2875), 1,
      aux_sym_attribute_id_repeat1,
    STATE(2883), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2733), 9,
      anon_sym_LPAREN,
      anon_sym_PLUS,
      anon_sym_BANG,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
    ACTIONS(2731), 21,
      anon_sym_rec,
      anon_sym_nonrec,
      anon_sym_DASH,
      anon_sym__,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_constraint,
      anon_sym_module,
      anon_sym_virtual,
      anon_sym_val,
      anon_sym_sig,
      anon_sym_end,
      anon_sym_functor,
      anon_sym_struct,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [111788] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 1,
      anon_sym_PIPE,
    ACTIONS(4080), 1,
      anon_sym_COMMA,
    ACTIONS(4084), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4086), 1,
      sym_concat_operator,
    ACTIONS(4088), 1,
      sym_rel_operator,
    STATE(565), 1,
      sym_pow_operator,
    STATE(566), 1,
      sym_mult_operator,
    STATE(567), 1,
      sym_add_operator,
    STATE(568), 1,
      sym_and_operator,
    STATE(569), 1,
      sym_or_operator,
    STATE(2884), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2878), 6,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_assign_operator,
  [111867] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2885), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3381), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3379), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [111913] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2886), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3317), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3315), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [111959] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4099), 1,
      anon_sym_COMMA,
    ACTIONS(4101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4103), 1,
      sym_concat_operator,
    ACTIONS(4105), 1,
      sym_rel_operator,
    ACTIONS(4107), 1,
      sym_assign_operator,
    STATE(525), 1,
      sym_and_operator,
    STATE(549), 1,
      sym_pow_operator,
    STATE(550), 1,
      sym_mult_operator,
    STATE(551), 1,
      sym_add_operator,
    STATE(552), 1,
      sym_or_operator,
    STATE(2887), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2864), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [112037] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2888), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3373), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3371), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [112083] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2889), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3377), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3375), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [112129] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2890), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3212), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3210), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [112175] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2718), 1,
      anon_sym_and,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4064), 1,
      anon_sym_COMMA,
    ACTIONS(4066), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4068), 1,
      sym_concat_operator,
    ACTIONS(4070), 1,
      sym_rel_operator,
    ACTIONS(4072), 1,
      sym_assign_operator,
    ACTIONS(4109), 1,
      anon_sym_SEMI,
    STATE(705), 1,
      sym_pow_operator,
    STATE(706), 1,
      sym_mult_operator,
    STATE(707), 1,
      sym_add_operator,
    STATE(708), 1,
      sym_and_operator,
    STATE(709), 1,
      sym_or_operator,
    STATE(2891), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2716), 3,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [112257] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2892), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2718), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2716), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [112303] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2893), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3413), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3411), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [112349] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2894), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3419), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3417), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [112395] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2895), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3337), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3335), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [112441] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2896), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2714), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2712), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [112487] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4099), 1,
      anon_sym_COMMA,
    ACTIONS(4101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4103), 1,
      sym_concat_operator,
    ACTIONS(4105), 1,
      sym_rel_operator,
    ACTIONS(4107), 1,
      sym_assign_operator,
    STATE(525), 1,
      sym_and_operator,
    STATE(549), 1,
      sym_pow_operator,
    STATE(550), 1,
      sym_mult_operator,
    STATE(551), 1,
      sym_add_operator,
    STATE(552), 1,
      sym_or_operator,
    STATE(2897), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2888), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [112565] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(525), 1,
      sym_and_operator,
    STATE(549), 1,
      sym_pow_operator,
    STATE(550), 1,
      sym_mult_operator,
    STATE(551), 1,
      sym_add_operator,
    STATE(552), 1,
      sym_or_operator,
    STATE(2898), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2874), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2872), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [112621] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2854), 1,
      anon_sym_and,
    ACTIONS(4064), 1,
      anon_sym_COMMA,
    ACTIONS(4066), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4068), 1,
      sym_concat_operator,
    ACTIONS(4070), 1,
      sym_rel_operator,
    ACTIONS(4072), 1,
      sym_assign_operator,
    ACTIONS(4109), 1,
      anon_sym_SEMI,
    STATE(705), 1,
      sym_pow_operator,
    STATE(706), 1,
      sym_mult_operator,
    STATE(707), 1,
      sym_add_operator,
    STATE(708), 1,
      sym_and_operator,
    STATE(709), 1,
      sym_or_operator,
    STATE(2899), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2852), 3,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [112703] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(525), 1,
      sym_and_operator,
    STATE(549), 1,
      sym_pow_operator,
    STATE(550), 1,
      sym_mult_operator,
    STATE(551), 1,
      sym_add_operator,
    STATE(552), 1,
      sym_or_operator,
    STATE(2900), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [112761] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(525), 1,
      sym_and_operator,
    STATE(549), 1,
      sym_pow_operator,
    STATE(550), 1,
      sym_mult_operator,
    STATE(551), 1,
      sym_add_operator,
    STATE(552), 1,
      sym_or_operator,
    STATE(2901), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 15,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [112819] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    STATE(525), 1,
      sym_and_operator,
    STATE(549), 1,
      sym_pow_operator,
    STATE(550), 1,
      sym_mult_operator,
    STATE(551), 1,
      sym_add_operator,
    STATE(552), 1,
      sym_or_operator,
    STATE(2902), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 6,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 11,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [112881] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2903), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3341), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3339), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [112927] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4101), 1,
      anon_sym_COLON_COLON,
    STATE(525), 1,
      sym_and_operator,
    STATE(549), 1,
      sym_pow_operator,
    STATE(550), 1,
      sym_mult_operator,
    STATE(551), 1,
      sym_add_operator,
    STATE(552), 1,
      sym_or_operator,
    STATE(2904), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2884), 3,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 10,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [112993] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2905), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3357), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3355), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [113039] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2906), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3345), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3343), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [113085] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2907), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3409), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3407), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [113131] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2908), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3333), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3331), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [113177] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(439), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(445), 1,
      sym_rel_operator,
    ACTIONS(447), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(471), 1,
      anon_sym_DOT,
    ACTIONS(4113), 1,
      anon_sym_or,
    STATE(2909), 1,
      sym_attribute,
    STATE(7187), 1,
      sym__infix_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(425), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(443), 2,
      sym_concat_operator,
      sym_assign_operator,
    ACTIONS(441), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(437), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(4111), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(479), 5,
      sym_prefix_operator,
      sym_hash_operator,
      sym_let_operator,
      sym_let_and_operator,
      sym_match_operator,
    STATE(6595), 5,
      sym_pow_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_and_operator,
      sym_or_operator,
  [113245] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4099), 1,
      anon_sym_COMMA,
    ACTIONS(4101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4103), 1,
      sym_concat_operator,
    ACTIONS(4105), 1,
      sym_rel_operator,
    ACTIONS(4107), 1,
      sym_assign_operator,
    STATE(525), 1,
      sym_and_operator,
    STATE(549), 1,
      sym_pow_operator,
    STATE(550), 1,
      sym_mult_operator,
    STATE(551), 1,
      sym_add_operator,
    STATE(552), 1,
      sym_or_operator,
    STATE(2910), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [113323] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2911), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3423), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3421), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [113369] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4099), 1,
      anon_sym_COMMA,
    ACTIONS(4101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4103), 1,
      sym_concat_operator,
    ACTIONS(4105), 1,
      sym_rel_operator,
    ACTIONS(4107), 1,
      sym_assign_operator,
    ACTIONS(4115), 1,
      anon_sym_SEMI,
    STATE(525), 1,
      sym_and_operator,
    STATE(549), 1,
      sym_pow_operator,
    STATE(550), 1,
      sym_mult_operator,
    STATE(551), 1,
      sym_add_operator,
    STATE(552), 1,
      sym_or_operator,
    STATE(2912), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2852), 4,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [113449] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2913), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2880), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2878), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [113495] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2914), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3385), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3383), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [113541] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4099), 1,
      anon_sym_COMMA,
    ACTIONS(4101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4103), 1,
      sym_concat_operator,
    ACTIONS(4105), 1,
      sym_rel_operator,
    ACTIONS(4107), 1,
      sym_assign_operator,
    STATE(525), 1,
      sym_and_operator,
    STATE(549), 1,
      sym_pow_operator,
    STATE(550), 1,
      sym_mult_operator,
    STATE(551), 1,
      sym_add_operator,
    STATE(552), 1,
      sym_or_operator,
    STATE(2915), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2824), 5,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
  [113619] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2916), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(749), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(747), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [113665] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4074), 1,
      sym__lowercase_identifier,
    ACTIONS(4117), 1,
      anon_sym_DASH_GT,
    ACTIONS(4119), 1,
      anon_sym_STAR,
    ACTIONS(4121), 1,
      anon_sym_POUND,
    STATE(2917), 1,
      sym_attribute,
    STATE(3055), 1,
      sym_type_constructor_path,
    STATE(6029), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3987), 8,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(3989), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [113725] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2918), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3325), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3323), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [113771] = 18,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2862), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4103), 1,
      sym_concat_operator,
    ACTIONS(4105), 1,
      sym_rel_operator,
    STATE(525), 1,
      sym_and_operator,
    STATE(549), 1,
      sym_pow_operator,
    STATE(550), 1,
      sym_mult_operator,
    STATE(551), 1,
      sym_add_operator,
    STATE(552), 1,
      sym_or_operator,
    STATE(2919), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 8,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_or,
      sym_assign_operator,
  [113843] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2920), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3365), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3363), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [113889] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4103), 1,
      sym_concat_operator,
    STATE(525), 1,
      sym_and_operator,
    STATE(549), 1,
      sym_pow_operator,
    STATE(550), 1,
      sym_mult_operator,
    STATE(551), 1,
      sym_add_operator,
    STATE(552), 1,
      sym_or_operator,
    STATE(2921), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2862), 3,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [113957] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2922), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3321), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3319), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [114003] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2923), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3389), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3387), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [114049] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4123), 1,
      anon_sym_DOT_DOT,
    STATE(2924), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1297), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1295), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [114097] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4125), 1,
      anon_sym_COLON,
    STATE(2925), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2936), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2938), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [114145] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2926), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3369), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3367), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [114191] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2927), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1683), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1685), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [114237] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2928), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1687), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1689), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [114283] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2929), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1709), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1711), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [114329] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2930), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3361), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3359), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [114375] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(439), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(445), 1,
      sym_rel_operator,
    ACTIONS(447), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4113), 1,
      anon_sym_or,
    ACTIONS(4127), 1,
      anon_sym_DOT,
    STATE(2931), 1,
      sym_attribute,
    STATE(7008), 1,
      sym__infix_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(425), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(443), 2,
      sym_concat_operator,
      sym_assign_operator,
    ACTIONS(441), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(437), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(4111), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(4129), 5,
      sym_prefix_operator,
      sym_hash_operator,
      sym_let_operator,
      sym_let_and_operator,
      sym_match_operator,
    STATE(6595), 5,
      sym_pow_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_and_operator,
      sym_or_operator,
  [114443] = 20,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4099), 1,
      anon_sym_COMMA,
    ACTIONS(4101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4103), 1,
      sym_concat_operator,
    ACTIONS(4105), 1,
      sym_rel_operator,
    STATE(525), 1,
      sym_and_operator,
    STATE(549), 1,
      sym_pow_operator,
    STATE(550), 1,
      sym_mult_operator,
    STATE(551), 1,
      sym_add_operator,
    STATE(552), 1,
      sym_or_operator,
    STATE(2932), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2878), 6,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_assign_operator,
  [114519] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4074), 1,
      sym__lowercase_identifier,
    ACTIONS(4121), 1,
      anon_sym_POUND,
    STATE(2933), 1,
      sym_attribute,
    STATE(3055), 1,
      sym_type_constructor_path,
    STATE(6029), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3628), 10,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(3630), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [114575] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2934), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1733), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1735), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [114621] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2935), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3405), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3403), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [114667] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2936), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3393), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3391), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [114713] = 23,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2710), 1,
      anon_sym_and,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4064), 1,
      anon_sym_COMMA,
    ACTIONS(4066), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4068), 1,
      sym_concat_operator,
    ACTIONS(4070), 1,
      sym_rel_operator,
    ACTIONS(4072), 1,
      sym_assign_operator,
    ACTIONS(4109), 1,
      anon_sym_SEMI,
    STATE(705), 1,
      sym_pow_operator,
    STATE(706), 1,
      sym_mult_operator,
    STATE(707), 1,
      sym_add_operator,
    STATE(708), 1,
      sym_and_operator,
    STATE(709), 1,
      sym_or_operator,
    STATE(2937), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2708), 3,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [114795] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4074), 1,
      sym__lowercase_identifier,
    ACTIONS(4121), 1,
      anon_sym_POUND,
    ACTIONS(4135), 1,
      anon_sym_STAR,
    STATE(2938), 1,
      sym_attribute,
    STATE(3055), 1,
      sym_type_constructor_path,
    STATE(6029), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4131), 9,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4133), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [114853] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2939), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2710), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2708), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [114899] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2940), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2930), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2932), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [114945] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2941), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1817), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1819), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [114991] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2942), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1829), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1831), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [115037] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2943), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1833), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1835), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [115083] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4103), 1,
      sym_concat_operator,
    ACTIONS(4105), 1,
      sym_rel_operator,
    STATE(525), 1,
      sym_and_operator,
    STATE(549), 1,
      sym_pow_operator,
    STATE(550), 1,
      sym_mult_operator,
    STATE(551), 1,
      sym_add_operator,
    STATE(552), 1,
      sym_or_operator,
    STATE(2944), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_assign_operator,
  [115157] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4103), 1,
      sym_concat_operator,
    STATE(525), 1,
      sym_and_operator,
    STATE(549), 1,
      sym_pow_operator,
    STATE(550), 1,
      sym_mult_operator,
    STATE(551), 1,
      sym_add_operator,
    STATE(552), 1,
      sym_or_operator,
    STATE(2945), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2862), 3,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [115225] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4099), 1,
      anon_sym_COMMA,
    ACTIONS(4101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4103), 1,
      sym_concat_operator,
    ACTIONS(4105), 1,
      sym_rel_operator,
    ACTIONS(4107), 1,
      sym_assign_operator,
    ACTIONS(4115), 1,
      anon_sym_SEMI,
    STATE(525), 1,
      sym_and_operator,
    STATE(549), 1,
      sym_pow_operator,
    STATE(550), 1,
      sym_mult_operator,
    STATE(551), 1,
      sym_add_operator,
    STATE(552), 1,
      sym_or_operator,
    STATE(2946), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2708), 4,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [115305] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2947), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3401), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3399), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [115351] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(439), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(445), 1,
      sym_rel_operator,
    ACTIONS(447), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1045), 1,
      anon_sym_DOT,
    ACTIONS(4113), 1,
      anon_sym_or,
    STATE(2948), 1,
      sym_attribute,
    STATE(6980), 1,
      sym__infix_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(425), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(443), 2,
      sym_concat_operator,
      sym_assign_operator,
    ACTIONS(441), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(437), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(4111), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(1077), 5,
      sym_prefix_operator,
      sym_hash_operator,
      sym_let_operator,
      sym_let_and_operator,
      sym_match_operator,
    STATE(6595), 5,
      sym_pow_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_and_operator,
      sym_or_operator,
  [115419] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2949), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2942), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2944), 21,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [115465] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(439), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(445), 1,
      sym_rel_operator,
    ACTIONS(447), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1141), 1,
      anon_sym_DOT,
    ACTIONS(4113), 1,
      anon_sym_or,
    STATE(2950), 1,
      sym_attribute,
    STATE(7169), 1,
      sym__infix_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(425), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(443), 2,
      sym_concat_operator,
      sym_assign_operator,
    ACTIONS(441), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(437), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(4111), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(1147), 5,
      sym_prefix_operator,
      sym_hash_operator,
      sym_let_operator,
      sym_let_and_operator,
      sym_match_operator,
    STATE(6595), 5,
      sym_pow_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_and_operator,
      sym_or_operator,
  [115533] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2951), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2724), 10,
      anon_sym_LPAREN,
      anon_sym_PLUS,
      anon_sym_BANG,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
      aux_sym_attribute_id_token1,
    ACTIONS(2722), 21,
      anon_sym_rec,
      anon_sym_nonrec,
      anon_sym_DASH,
      anon_sym__,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_constraint,
      anon_sym_module,
      anon_sym_virtual,
      anon_sym_val,
      anon_sym_sig,
      anon_sym_end,
      anon_sym_functor,
      anon_sym_struct,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [115579] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(439), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(445), 1,
      sym_rel_operator,
    ACTIONS(447), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1089), 1,
      anon_sym_DOT,
    ACTIONS(4113), 1,
      anon_sym_or,
    STATE(2952), 1,
      sym_attribute,
    STATE(7342), 1,
      sym__infix_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(425), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(443), 2,
      sym_concat_operator,
      sym_assign_operator,
    ACTIONS(441), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(437), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(4111), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(1095), 5,
      sym_prefix_operator,
      sym_hash_operator,
      sym_let_operator,
      sym_let_and_operator,
      sym_match_operator,
    STATE(6595), 5,
      sym_pow_operator,
      sym_mult_operator,
      sym_add_operator,
      sym_and_operator,
      sym_or_operator,
  [115647] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2953), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3353), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3351), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [115693] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4099), 1,
      anon_sym_COMMA,
    ACTIONS(4101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4103), 1,
      sym_concat_operator,
    ACTIONS(4105), 1,
      sym_rel_operator,
    ACTIONS(4107), 1,
      sym_assign_operator,
    ACTIONS(4115), 1,
      anon_sym_SEMI,
    STATE(525), 1,
      sym_and_operator,
    STATE(549), 1,
      sym_pow_operator,
    STATE(550), 1,
      sym_mult_operator,
    STATE(551), 1,
      sym_add_operator,
    STATE(552), 1,
      sym_or_operator,
    STATE(2954), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2716), 4,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [115773] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2955), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3349), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3347), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [115819] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2956), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3397), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3395), 22,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [115865] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2957), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3052), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3050), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [115910] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2958), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3947), 9,
      anon_sym_LPAREN,
      anon_sym_PLUS,
      anon_sym_BANG,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
    ACTIONS(3945), 21,
      anon_sym_rec,
      anon_sym_nonrec,
      anon_sym_DASH,
      anon_sym__,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_constraint,
      anon_sym_module,
      anon_sym_virtual,
      anon_sym_val,
      anon_sym_sig,
      anon_sym_end,
      anon_sym_functor,
      anon_sym_struct,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym__lowercase_identifier,
  [115955] = 20,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(259), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(261), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(2434), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(3787), 1,
      anon_sym_open,
    ACTIONS(4137), 1,
      anon_sym_external,
    ACTIONS(4139), 1,
      anon_sym_type,
    ACTIONS(4143), 1,
      anon_sym_exception,
    ACTIONS(4145), 1,
      anon_sym_module,
    ACTIONS(4147), 1,
      anon_sym_include,
    ACTIONS(4149), 1,
      anon_sym_class,
    ACTIONS(4151), 1,
      anon_sym_val,
    STATE(2959), 1,
      sym_attribute,
    STATE(3025), 1,
      aux_sym__signature_repeat1,
    STATE(3163), 1,
      aux_sym__structure_repeat1,
    STATE(4212), 1,
      sym__signature_item,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4141), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(4240), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [116030] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4099), 1,
      anon_sym_COMMA,
    ACTIONS(4101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4103), 1,
      sym_concat_operator,
    ACTIONS(4105), 1,
      sym_rel_operator,
    ACTIONS(4107), 1,
      sym_assign_operator,
    ACTIONS(4153), 1,
      anon_sym_SEMI,
    STATE(525), 1,
      sym_and_operator,
    STATE(549), 1,
      sym_pow_operator,
    STATE(550), 1,
      sym_mult_operator,
    STATE(551), 1,
      sym_add_operator,
    STATE(552), 1,
      sym_or_operator,
    STATE(2960), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2708), 3,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [116109] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2961), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3118), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3120), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116154] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2962), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3122), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3124), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116199] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_DOT,
    STATE(2963), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1699), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1701), 26,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [116246] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2964), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3126), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3128), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116291] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2965), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2984), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2986), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116336] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2966), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3022), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3024), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116381] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2967), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3086), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3088), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116426] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2968), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3080), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3078), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [116471] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2969), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2984), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2986), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116516] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2970), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3000), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3002), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116561] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2971), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3821), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3823), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [116606] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2972), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2970), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2972), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116651] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2973), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3130), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3132), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116696] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2974), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3134), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3136), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116741] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(4010), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4012), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4157), 1,
      anon_sym_let,
    STATE(2975), 1,
      sym_attribute,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3743), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4155), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [116802] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2976), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3162), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3164), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116847] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2977), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3008), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3010), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116892] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2978), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3040), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3042), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116937] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2979), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3062), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3064), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [116982] = 24,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3688), 1,
      sym_assign_operator,
    ACTIONS(3959), 1,
      sym_rel_operator,
    ACTIONS(4159), 1,
      anon_sym_RPAREN,
    ACTIONS(4161), 1,
      anon_sym_COLON2,
    ACTIONS(4163), 1,
      anon_sym_COLON_GT,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(2980), 1,
      sym_attribute,
    STATE(6261), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [117065] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2981), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3873), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3875), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [117110] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2982), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3877), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3879), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [117155] = 23,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2854), 1,
      anon_sym_PIPE,
    ACTIONS(4165), 1,
      anon_sym_COMMA,
    ACTIONS(4167), 1,
      anon_sym_SEMI,
    ACTIONS(4169), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4171), 1,
      sym_concat_operator,
    ACTIONS(4173), 1,
      sym_rel_operator,
    ACTIONS(4175), 1,
      sym_assign_operator,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(2983), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2852), 2,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [117236] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2984), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3825), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3827), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [117281] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2985), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3869), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3871), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [117326] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2986), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1809), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1811), 27,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [117371] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2987), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3154), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3156), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [117416] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2988), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3843), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3845), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [117461] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2989), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3718), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3720), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [117506] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2990), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1727), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(1725), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [117551] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(259), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(261), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(2434), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(3787), 1,
      anon_sym_open,
    ACTIONS(4137), 1,
      anon_sym_external,
    ACTIONS(4139), 1,
      anon_sym_type,
    ACTIONS(4143), 1,
      anon_sym_exception,
    ACTIONS(4145), 1,
      anon_sym_module,
    ACTIONS(4147), 1,
      anon_sym_include,
    ACTIONS(4149), 1,
      anon_sym_class,
    ACTIONS(4151), 1,
      anon_sym_val,
    ACTIONS(4177), 1,
      anon_sym_end,
    STATE(2959), 1,
      aux_sym__signature_repeat1,
    STATE(2991), 1,
      sym_attribute,
    STATE(3139), 1,
      aux_sym__structure_repeat1,
    STATE(4212), 1,
      sym__signature_item,
    STATE(6995), 1,
      sym__signature,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4240), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [117628] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2992), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1771), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(1769), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [117673] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2993), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1791), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(1789), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [117718] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2994), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3883), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3885), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [117763] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2995), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3847), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3849), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [117808] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2996), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1803), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(1801), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [117853] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2997), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4179), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1841), 25,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [117898] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2998), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3070), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3072), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [117943] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(2999), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3887), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3889), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [117988] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3000), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3714), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3716), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [118033] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4099), 1,
      anon_sym_COMMA,
    ACTIONS(4101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4103), 1,
      sym_concat_operator,
    ACTIONS(4105), 1,
      sym_rel_operator,
    ACTIONS(4107), 1,
      sym_assign_operator,
    ACTIONS(4153), 1,
      anon_sym_SEMI,
    STATE(525), 1,
      sym_and_operator,
    STATE(549), 1,
      sym_pow_operator,
    STATE(550), 1,
      sym_mult_operator,
    STATE(551), 1,
      sym_add_operator,
    STATE(552), 1,
      sym_or_operator,
    STATE(3001), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2716), 3,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [118112] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3002), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3044), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3046), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [118157] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3003), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3895), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3897), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [118202] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3004), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3801), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3803), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [118247] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3005), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3074), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3076), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [118292] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4181), 1,
      anon_sym_COMMA,
    ACTIONS(4183), 1,
      anon_sym_COLON_COLON,
    STATE(3006), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3074), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3076), 18,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [118341] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4181), 1,
      anon_sym_COMMA,
    STATE(3007), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3074), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3076), 19,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [118388] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3008), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3074), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3076), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [118433] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4181), 1,
      anon_sym_COMMA,
    ACTIONS(4183), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4185), 1,
      anon_sym_PIPE,
    STATE(3009), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3094), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3096), 17,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [118484] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3010), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3094), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3096), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [118529] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3011), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3026), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3028), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [118574] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4183), 1,
      anon_sym_COLON_COLON,
    STATE(3012), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3066), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3068), 19,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [118621] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3013), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3058), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3060), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [118666] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3014), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3805), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3807), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [118711] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3015), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3110), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3112), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [118756] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3016), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3809), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3811), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [118801] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3660), 1,
      anon_sym_COLON2,
    STATE(3017), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3052), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3050), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [118848] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4099), 1,
      anon_sym_COMMA,
    ACTIONS(4101), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4103), 1,
      sym_concat_operator,
    ACTIONS(4105), 1,
      sym_rel_operator,
    ACTIONS(4107), 1,
      sym_assign_operator,
    ACTIONS(4153), 1,
      anon_sym_SEMI,
    STATE(525), 1,
      sym_and_operator,
    STATE(549), 1,
      sym_pow_operator,
    STATE(550), 1,
      sym_mult_operator,
    STATE(551), 1,
      sym_add_operator,
    STATE(552), 1,
      sym_or_operator,
    STATE(3018), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2852), 3,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [118927] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3019), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3170), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3172), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [118972] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3020), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3899), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3901), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [119017] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3021), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1745), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1747), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [119062] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3022), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1721), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1723), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [119107] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3023), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1297), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1295), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [119152] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3024), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3158), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3160), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [119197] = 19,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4187), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(4190), 1,
      anon_sym_external,
    ACTIONS(4193), 1,
      anon_sym_type,
    ACTIONS(4198), 1,
      anon_sym_exception,
    ACTIONS(4201), 1,
      anon_sym_module,
    ACTIONS(4204), 1,
      anon_sym_open,
    ACTIONS(4207), 1,
      anon_sym_include,
    ACTIONS(4210), 1,
      anon_sym_class,
    ACTIONS(4213), 1,
      anon_sym_val,
    ACTIONS(4216), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(4219), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(4222), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    STATE(3268), 1,
      aux_sym__structure_repeat1,
    STATE(4212), 1,
      sym__signature_item,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4196), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(3025), 2,
      sym_attribute,
      aux_sym__signature_repeat1,
    STATE(4240), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [119270] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4181), 1,
      anon_sym_COMMA,
    ACTIONS(4183), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4185), 1,
      anon_sym_PIPE,
    ACTIONS(4225), 1,
      anon_sym_as,
    STATE(3026), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3082), 9,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3084), 17,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [119323] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3027), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3114), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3116), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [119368] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3028), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3138), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3140), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [119413] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3029), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3050), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3052), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [119458] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3030), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3142), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3144), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [119503] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3031), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3829), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3831), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [119548] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3032), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3146), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3148), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [119593] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(606), 1,
      sym_pow_operator,
    STATE(607), 1,
      sym_mult_operator,
    STATE(608), 1,
      sym_add_operator,
    STATE(609), 1,
      sym_and_operator,
    STATE(610), 1,
      sym_or_operator,
    STATE(3033), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2874), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2872), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [119648] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 1,
      anon_sym_PIPE,
    ACTIONS(4227), 1,
      anon_sym_COMMA,
    ACTIONS(4229), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4231), 1,
      sym_concat_operator,
    ACTIONS(4233), 1,
      sym_rel_operator,
    STATE(606), 1,
      sym_pow_operator,
    STATE(607), 1,
      sym_mult_operator,
    STATE(608), 1,
      sym_add_operator,
    STATE(609), 1,
      sym_and_operator,
    STATE(610), 1,
      sym_or_operator,
    STATE(3034), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2878), 4,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      sym_assign_operator,
  [119725] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3035), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1847), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1849), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [119770] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4229), 1,
      anon_sym_COLON_COLON,
    STATE(606), 1,
      sym_pow_operator,
    STATE(607), 1,
      sym_mult_operator,
    STATE(608), 1,
      sym_add_operator,
    STATE(609), 1,
      sym_and_operator,
    STATE(610), 1,
      sym_or_operator,
    STATE(3036), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2884), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 8,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [119835] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4229), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4231), 1,
      sym_concat_operator,
    STATE(606), 1,
      sym_pow_operator,
    STATE(607), 1,
      sym_mult_operator,
    STATE(608), 1,
      sym_add_operator,
    STATE(609), 1,
      sym_and_operator,
    STATE(610), 1,
      sym_or_operator,
    STATE(3037), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [119902] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4229), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4231), 1,
      sym_concat_operator,
    STATE(606), 1,
      sym_pow_operator,
    STATE(607), 1,
      sym_mult_operator,
    STATE(608), 1,
      sym_add_operator,
    STATE(609), 1,
      sym_and_operator,
    STATE(610), 1,
      sym_or_operator,
    STATE(3038), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [119969] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2862), 1,
      anon_sym_PIPE,
    ACTIONS(4227), 1,
      anon_sym_COMMA,
    ACTIONS(4229), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4231), 1,
      sym_concat_operator,
    ACTIONS(4233), 1,
      sym_rel_operator,
    ACTIONS(4235), 1,
      sym_assign_operator,
    STATE(606), 1,
      sym_pow_operator,
    STATE(607), 1,
      sym_mult_operator,
    STATE(608), 1,
      sym_add_operator,
    STATE(609), 1,
      sym_and_operator,
    STATE(610), 1,
      sym_or_operator,
    STATE(3039), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [120048] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(606), 1,
      sym_pow_operator,
    STATE(607), 1,
      sym_mult_operator,
    STATE(608), 1,
      sym_add_operator,
    STATE(609), 1,
      sym_and_operator,
    STATE(610), 1,
      sym_or_operator,
    STATE(3040), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 13,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [120105] = 24,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3688), 1,
      sym_assign_operator,
    ACTIONS(3959), 1,
      sym_rel_operator,
    ACTIONS(4161), 1,
      anon_sym_COLON2,
    ACTIONS(4237), 1,
      anon_sym_RPAREN,
    ACTIONS(4239), 1,
      anon_sym_COLON_GT,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(3041), 1,
      sym_attribute,
    STATE(5987), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [120188] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    STATE(606), 1,
      sym_pow_operator,
    STATE(607), 1,
      sym_mult_operator,
    STATE(608), 1,
      sym_add_operator,
    STATE(609), 1,
      sym_and_operator,
    STATE(610), 1,
      sym_or_operator,
    STATE(3042), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 7,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [120249] = 18,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4229), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4231), 1,
      sym_concat_operator,
    ACTIONS(4233), 1,
      sym_rel_operator,
    STATE(606), 1,
      sym_pow_operator,
    STATE(607), 1,
      sym_mult_operator,
    STATE(608), 1,
      sym_add_operator,
    STATE(609), 1,
      sym_and_operator,
    STATE(610), 1,
      sym_or_operator,
    STATE(3043), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2862), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_PIPE,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 6,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_or,
      sym_assign_operator,
  [120320] = 20,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2862), 1,
      anon_sym_PIPE,
    ACTIONS(4229), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4231), 1,
      sym_concat_operator,
    ACTIONS(4233), 1,
      sym_rel_operator,
    STATE(606), 1,
      sym_pow_operator,
    STATE(607), 1,
      sym_mult_operator,
    STATE(608), 1,
      sym_add_operator,
    STATE(609), 1,
      sym_and_operator,
    STATE(610), 1,
      sym_or_operator,
    STATE(3044), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      sym_assign_operator,
  [120395] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      anon_sym_PIPE,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4227), 1,
      anon_sym_COMMA,
    ACTIONS(4229), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4231), 1,
      sym_concat_operator,
    ACTIONS(4233), 1,
      sym_rel_operator,
    ACTIONS(4235), 1,
      sym_assign_operator,
    STATE(606), 1,
      sym_pow_operator,
    STATE(607), 1,
      sym_mult_operator,
    STATE(608), 1,
      sym_add_operator,
    STATE(609), 1,
      sym_and_operator,
    STATE(610), 1,
      sym_or_operator,
    STATE(3045), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2824), 3,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [120474] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2866), 1,
      anon_sym_PIPE,
    ACTIONS(4227), 1,
      anon_sym_COMMA,
    ACTIONS(4229), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4231), 1,
      sym_concat_operator,
    ACTIONS(4233), 1,
      sym_rel_operator,
    ACTIONS(4235), 1,
      sym_assign_operator,
    STATE(606), 1,
      sym_pow_operator,
    STATE(607), 1,
      sym_mult_operator,
    STATE(608), 1,
      sym_add_operator,
    STATE(609), 1,
      sym_and_operator,
    STATE(610), 1,
      sym_or_operator,
    STATE(3046), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2864), 3,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [120553] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3047), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3078), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3080), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [120598] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2890), 1,
      anon_sym_PIPE,
    ACTIONS(4227), 1,
      anon_sym_COMMA,
    ACTIONS(4229), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4231), 1,
      sym_concat_operator,
    ACTIONS(4233), 1,
      sym_rel_operator,
    ACTIONS(4235), 1,
      sym_assign_operator,
    STATE(606), 1,
      sym_pow_operator,
    STATE(607), 1,
      sym_mult_operator,
    STATE(608), 1,
      sym_add_operator,
    STATE(609), 1,
      sym_and_operator,
    STATE(610), 1,
      sym_or_operator,
    STATE(3048), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2888), 3,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [120677] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3049), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3150), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3152), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [120722] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3050), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1813), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1815), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [120767] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3051), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3166), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3168), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [120812] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3052), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1827), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [120857] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3053), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3090), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3092), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [120902] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3054), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3098), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3100), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [120947] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3055), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3769), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3771), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [120992] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3056), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1785), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1787), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [121037] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3057), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1797), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1799), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [121082] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3058), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1805), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1807), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [121127] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3059), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1757), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1759), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [121172] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3060), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4241), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1703), 25,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [121217] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2880), 1,
      anon_sym_PIPE,
    ACTIONS(4165), 1,
      anon_sym_COMMA,
    ACTIONS(4169), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4171), 1,
      sym_concat_operator,
    ACTIONS(4173), 1,
      sym_rel_operator,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3061), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2878), 4,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_assign_operator,
  [121294] = 23,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2710), 1,
      anon_sym_PIPE,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4227), 1,
      anon_sym_COMMA,
    ACTIONS(4229), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4231), 1,
      sym_concat_operator,
    ACTIONS(4233), 1,
      sym_rel_operator,
    ACTIONS(4235), 1,
      sym_assign_operator,
    ACTIONS(4243), 1,
      anon_sym_SEMI,
    STATE(606), 1,
      sym_pow_operator,
    STATE(607), 1,
      sym_mult_operator,
    STATE(608), 1,
      sym_add_operator,
    STATE(609), 1,
      sym_and_operator,
    STATE(610), 1,
      sym_or_operator,
    STATE(3062), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2708), 2,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [121375] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4169), 1,
      anon_sym_COLON_COLON,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3063), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2884), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2882), 8,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [121440] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4169), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4171), 1,
      sym_concat_operator,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3064), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [121507] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4169), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4171), 1,
      sym_concat_operator,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3065), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 4,
      anon_sym_PIPE,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [121574] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2862), 1,
      anon_sym_PIPE,
    ACTIONS(4165), 1,
      anon_sym_COMMA,
    ACTIONS(4169), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4171), 1,
      sym_concat_operator,
    ACTIONS(4173), 1,
      sym_rel_operator,
    ACTIONS(4175), 1,
      sym_assign_operator,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3066), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [121653] = 18,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4169), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4171), 1,
      sym_concat_operator,
    ACTIONS(4173), 1,
      sym_rel_operator,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3067), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2862), 2,
      anon_sym_PIPE,
      anon_sym_PIPE_PIPE,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 6,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      anon_sym_or,
      sym_assign_operator,
  [121724] = 20,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2862), 1,
      anon_sym_PIPE,
    ACTIONS(4169), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4171), 1,
      sym_concat_operator,
    ACTIONS(4173), 1,
      sym_rel_operator,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3068), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_assign_operator,
  [121799] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2826), 1,
      anon_sym_PIPE,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4165), 1,
      anon_sym_COMMA,
    ACTIONS(4169), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4171), 1,
      sym_concat_operator,
    ACTIONS(4173), 1,
      sym_rel_operator,
    ACTIONS(4175), 1,
      sym_assign_operator,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3069), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2824), 3,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [121878] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2866), 1,
      anon_sym_PIPE,
    ACTIONS(4165), 1,
      anon_sym_COMMA,
    ACTIONS(4169), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4171), 1,
      sym_concat_operator,
    ACTIONS(4173), 1,
      sym_rel_operator,
    ACTIONS(4175), 1,
      sym_assign_operator,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3070), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2864), 3,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [121957] = 23,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2718), 1,
      anon_sym_PIPE,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4227), 1,
      anon_sym_COMMA,
    ACTIONS(4229), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4231), 1,
      sym_concat_operator,
    ACTIONS(4233), 1,
      sym_rel_operator,
    ACTIONS(4235), 1,
      sym_assign_operator,
    ACTIONS(4243), 1,
      anon_sym_SEMI,
    STATE(606), 1,
      sym_pow_operator,
    STATE(607), 1,
      sym_mult_operator,
    STATE(608), 1,
      sym_add_operator,
    STATE(609), 1,
      sym_and_operator,
    STATE(610), 1,
      sym_or_operator,
    STATE(3071), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2716), 2,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [122038] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2890), 1,
      anon_sym_PIPE,
    ACTIONS(4165), 1,
      anon_sym_COMMA,
    ACTIONS(4169), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4171), 1,
      sym_concat_operator,
    ACTIONS(4173), 1,
      sym_rel_operator,
    ACTIONS(4175), 1,
      sym_assign_operator,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3072), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2888), 3,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [122117] = 23,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2710), 1,
      anon_sym_PIPE,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4165), 1,
      anon_sym_COMMA,
    ACTIONS(4167), 1,
      anon_sym_SEMI,
    ACTIONS(4169), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4171), 1,
      sym_concat_operator,
    ACTIONS(4173), 1,
      sym_rel_operator,
    ACTIONS(4175), 1,
      sym_assign_operator,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3073), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2708), 2,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [122198] = 23,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2718), 1,
      anon_sym_PIPE,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4165), 1,
      anon_sym_COMMA,
    ACTIONS(4167), 1,
      anon_sym_SEMI,
    ACTIONS(4169), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4171), 1,
      sym_concat_operator,
    ACTIONS(4173), 1,
      sym_rel_operator,
    ACTIONS(4175), 1,
      sym_assign_operator,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3074), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2716), 2,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [122279] = 23,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2854), 1,
      anon_sym_PIPE,
    ACTIONS(4227), 1,
      anon_sym_COMMA,
    ACTIONS(4229), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4231), 1,
      sym_concat_operator,
    ACTIONS(4233), 1,
      sym_rel_operator,
    ACTIONS(4235), 1,
      sym_assign_operator,
    ACTIONS(4243), 1,
      anon_sym_SEMI,
    STATE(606), 1,
      sym_pow_operator,
    STATE(607), 1,
      sym_mult_operator,
    STATE(608), 1,
      sym_add_operator,
    STATE(609), 1,
      sym_and_operator,
    STATE(610), 1,
      sym_or_operator,
    STATE(3075), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2852), 2,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [122360] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3076), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1725), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1727), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [122405] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3077), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3102), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3104), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [122450] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3078), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3620), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3624), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [122495] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4249), 1,
      anon_sym_DOT,
    STATE(3079), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4247), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4245), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [122542] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3080), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3054), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3056), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [122587] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(4010), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4012), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4253), 1,
      anon_sym_let,
    STATE(3081), 1,
      sym_attribute,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3712), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4251), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [122648] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3082), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3797), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3799), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [122693] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(4010), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4012), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4253), 1,
      anon_sym_let,
    STATE(3083), 1,
      sym_attribute,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3713), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4251), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [122754] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3084), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1771), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [122799] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3085), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1789), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1791), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [122844] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4255), 1,
      anon_sym_COLON2,
    STATE(3086), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3052), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3050), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [122891] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3087), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3851), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3853), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [122936] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3088), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3813), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3815), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [122981] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(4010), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4012), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4259), 1,
      anon_sym_let,
    STATE(3089), 1,
      sym_attribute,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3615), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4257), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [123042] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3090), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3855), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3857), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [123087] = 21,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(259), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(261), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(2434), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(3787), 1,
      anon_sym_open,
    ACTIONS(4137), 1,
      anon_sym_external,
    ACTIONS(4139), 1,
      anon_sym_type,
    ACTIONS(4143), 1,
      anon_sym_exception,
    ACTIONS(4145), 1,
      anon_sym_module,
    ACTIONS(4147), 1,
      anon_sym_include,
    ACTIONS(4149), 1,
      anon_sym_class,
    ACTIONS(4151), 1,
      anon_sym_val,
    ACTIONS(4261), 1,
      anon_sym_end,
    STATE(2959), 1,
      aux_sym__signature_repeat1,
    STATE(3091), 1,
      sym_attribute,
    STATE(3139), 1,
      aux_sym__structure_repeat1,
    STATE(4212), 1,
      sym__signature_item,
    STATE(6469), 1,
      sym__signature,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4240), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [123164] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3092), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3891), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3893), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [123209] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3093), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1801), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(1803), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [123254] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3094), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3911), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3913), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [123299] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(4010), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4012), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4259), 1,
      anon_sym_let,
    STATE(3095), 1,
      sym_attribute,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3616), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4257), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [123360] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3096), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3773), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3775), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [123405] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(4010), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4012), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4265), 1,
      anon_sym_let,
    STATE(3097), 1,
      sym_attribute,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3618), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4263), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [123466] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3098), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3915), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3917), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [123511] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4181), 1,
      anon_sym_COMMA,
    ACTIONS(4183), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4185), 1,
      anon_sym_PIPE,
    ACTIONS(4225), 1,
      anon_sym_as,
    STATE(3099), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2988), 9,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2990), 17,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [123564] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3100), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3036), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3038), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [123609] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3101), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2874), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2872), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [123664] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3102), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 13,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [123721] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3103), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 13,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [123778] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    STATE(647), 1,
      sym_pow_operator,
    STATE(648), 1,
      sym_mult_operator,
    STATE(649), 1,
      sym_add_operator,
    STATE(650), 1,
      sym_and_operator,
    STATE(651), 1,
      sym_or_operator,
    STATE(3104), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 7,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [123839] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1703), 1,
      anon_sym_DOT,
    STATE(3105), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1699), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1701), 26,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [123886] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4181), 1,
      anon_sym_COMMA,
    ACTIONS(4183), 1,
      anon_sym_COLON_COLON,
    STATE(3106), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2984), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2986), 18,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [123935] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4181), 1,
      anon_sym_COMMA,
    STATE(3107), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2984), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2986), 19,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [123982] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4181), 1,
      anon_sym_COMMA,
    ACTIONS(4183), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4185), 1,
      anon_sym_PIPE,
    STATE(3108), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3000), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3002), 17,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [124033] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3109), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3817), 13,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3819), 17,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [124078] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4183), 1,
      anon_sym_COLON_COLON,
    STATE(3110), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3004), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3006), 19,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [124125] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3111), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3014), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3016), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [124170] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3112), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3018), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3020), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [124215] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3113), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3106), 10,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_as,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3108), 20,
      anon_sym_EQ,
      anon_sym_TILDE,
      anon_sym_QMARK,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_and_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [124260] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(606), 1,
      sym_pow_operator,
    STATE(607), 1,
      sym_mult_operator,
    STATE(608), 1,
      sym_add_operator,
    STATE(609), 1,
      sym_and_operator,
    STATE(610), 1,
      sym_or_operator,
    STATE(3114), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 13,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [124317] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4267), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4269), 1,
      sym_concat_operator,
    ACTIONS(4271), 1,
      sym_rel_operator,
    STATE(692), 1,
      sym_pow_operator,
    STATE(693), 1,
      sym_mult_operator,
    STATE(694), 1,
      sym_add_operator,
    STATE(695), 1,
      sym_and_operator,
    STATE(696), 1,
      sym_or_operator,
    STATE(3115), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_assign_operator,
  [124389] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4273), 1,
      anon_sym_else,
    STATE(2918), 1,
      sym_else_clause,
    STATE(3116), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3277), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3275), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [124437] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4275), 1,
      anon_sym_PIPE,
    STATE(3117), 1,
      sym_attribute,
    STATE(3118), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3208), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3206), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [124485] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4277), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3118), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3212), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3210), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [124531] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3119), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1843), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1845), 26,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [124575] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(4010), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4012), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4282), 1,
      anon_sym_let,
    STATE(3120), 1,
      sym_attribute,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3810), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4280), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [124635] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3121), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4286), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4284), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [124679] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3122), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4290), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4288), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [124723] = 23,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3237), 1,
      anon_sym_LBRACE,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    STATE(3123), 1,
      sym_attribute,
    STATE(3261), 1,
      sym__simple_type,
    STATE(3296), 1,
      sym_type_variable,
    STATE(3540), 1,
      sym_record_declaration,
    STATE(5372), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(7256), 1,
      sym__constructor_argument,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [124803] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3124), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4298), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4296), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [124847] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3125), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4300), 2,
      anon_sym_DOT,
      aux_sym_type_variable_token1,
    ACTIONS(3773), 11,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym__uppercase_identifier,
    ACTIONS(3775), 16,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
      aux_sym_directive_token1,
  [124893] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(4010), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4012), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4304), 1,
      anon_sym_let,
    STATE(3126), 1,
      sym_attribute,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3815), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4302), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [124953] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3127), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4308), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4306), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [124997] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3128), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1725), 9,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_POUND,
      anon_sym_when,
      sym__lowercase_identifier,
    ACTIONS(1727), 20,
      ts_builtin_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym_hash_operator,
      sym__uppercase_identifier,
  [125041] = 23,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3237), 1,
      anon_sym_LBRACE,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3698), 1,
      anon_sym_LBRACK,
    ACTIONS(4310), 1,
      sym__lowercase_identifier,
    ACTIONS(4312), 1,
      anon_sym_LPAREN,
    STATE(3125), 1,
      sym_type_variable,
    STATE(3129), 1,
      sym_attribute,
    STATE(3156), 1,
      sym__simple_type,
    STATE(3540), 1,
      sym_record_declaration,
    STATE(5302), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6244), 1,
      sym_extended_module_path,
    STATE(7208), 1,
      sym__constructor_argument,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [125121] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4275), 1,
      anon_sym_PIPE,
    STATE(3130), 1,
      sym_attribute,
    STATE(3134), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3180), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3178), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [125169] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3131), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4316), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4314), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [125213] = 23,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4227), 1,
      anon_sym_COMMA,
    ACTIONS(4229), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4231), 1,
      sym_concat_operator,
    ACTIONS(4233), 1,
      sym_rel_operator,
    ACTIONS(4235), 1,
      sym_assign_operator,
    ACTIONS(4318), 1,
      anon_sym_SEMI,
    ACTIONS(4320), 1,
      anon_sym_PIPE_RBRACK,
    STATE(606), 1,
      sym_pow_operator,
    STATE(607), 1,
      sym_mult_operator,
    STATE(608), 1,
      sym_add_operator,
    STATE(609), 1,
      sym_and_operator,
    STATE(610), 1,
      sym_or_operator,
    STATE(3132), 1,
      sym_attribute,
    STATE(5832), 1,
      aux_sym__sequence_expression_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [125293] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4275), 1,
      anon_sym_PIPE,
    STATE(3117), 1,
      aux_sym__match_cases_repeat1,
    STATE(3133), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3259), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3257), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [125341] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4275), 1,
      anon_sym_PIPE,
    STATE(3118), 1,
      aux_sym__match_cases_repeat1,
    STATE(3134), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3259), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3257), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [125389] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3135), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 9,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_POUND,
      anon_sym_when,
      sym__lowercase_identifier,
    ACTIONS(1771), 20,
      ts_builtin_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym_hash_operator,
      sym__uppercase_identifier,
  [125433] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3136), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4324), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4322), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [125477] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3137), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1789), 9,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_POUND,
      anon_sym_when,
      sym__lowercase_identifier,
    ACTIONS(1791), 20,
      ts_builtin_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym_hash_operator,
      sym__uppercase_identifier,
  [125521] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3138), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4328), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4326), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [125565] = 19,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(259), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(261), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(2434), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(3787), 1,
      anon_sym_open,
    ACTIONS(4137), 1,
      anon_sym_external,
    ACTIONS(4139), 1,
      anon_sym_type,
    ACTIONS(4143), 1,
      anon_sym_exception,
    ACTIONS(4145), 1,
      anon_sym_module,
    ACTIONS(4147), 1,
      anon_sym_include,
    ACTIONS(4149), 1,
      anon_sym_class,
    ACTIONS(4151), 1,
      anon_sym_val,
    STATE(3139), 1,
      sym_attribute,
    STATE(4207), 1,
      aux_sym__structure_repeat1,
    STATE(4222), 1,
      sym__signature_item,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4141), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(4240), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [125637] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3140), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4332), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4330), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [125681] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(692), 1,
      sym_pow_operator,
    STATE(693), 1,
      sym_mult_operator,
    STATE(694), 1,
      sym_add_operator,
    STATE(695), 1,
      sym_and_operator,
    STATE(696), 1,
      sym_or_operator,
    STATE(3141), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2874), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2872), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [125735] = 20,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4267), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4269), 1,
      sym_concat_operator,
    ACTIONS(4271), 1,
      sym_rel_operator,
    ACTIONS(4334), 1,
      anon_sym_COMMA,
    STATE(692), 1,
      sym_pow_operator,
    STATE(693), 1,
      sym_mult_operator,
    STATE(694), 1,
      sym_add_operator,
    STATE(695), 1,
      sym_and_operator,
    STATE(696), 1,
      sym_or_operator,
    STATE(3142), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2878), 4,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_assign_operator,
  [125809] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(4010), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4012), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4338), 1,
      anon_sym_let,
    STATE(3143), 1,
      sym_attribute,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3819), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4336), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [125869] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4267), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4269), 1,
      sym_concat_operator,
    STATE(692), 1,
      sym_pow_operator,
    STATE(693), 1,
      sym_mult_operator,
    STATE(694), 1,
      sym_add_operator,
    STATE(695), 1,
      sym_and_operator,
    STATE(696), 1,
      sym_or_operator,
    STATE(3144), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2862), 3,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [125935] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4267), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4269), 1,
      sym_concat_operator,
    STATE(692), 1,
      sym_pow_operator,
    STATE(693), 1,
      sym_mult_operator,
    STATE(694), 1,
      sym_add_operator,
    STATE(695), 1,
      sym_and_operator,
    STATE(696), 1,
      sym_or_operator,
    STATE(3145), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2862), 3,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 7,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [126001] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4267), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4269), 1,
      sym_concat_operator,
    ACTIONS(4271), 1,
      sym_rel_operator,
    ACTIONS(4334), 1,
      anon_sym_COMMA,
    ACTIONS(4340), 1,
      sym_assign_operator,
    STATE(692), 1,
      sym_pow_operator,
    STATE(693), 1,
      sym_mult_operator,
    STATE(694), 1,
      sym_add_operator,
    STATE(695), 1,
      sym_and_operator,
    STATE(696), 1,
      sym_or_operator,
    STATE(3146), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2860), 3,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [126077] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(692), 1,
      sym_pow_operator,
    STATE(693), 1,
      sym_mult_operator,
    STATE(694), 1,
      sym_add_operator,
    STATE(695), 1,
      sym_and_operator,
    STATE(696), 1,
      sym_or_operator,
    STATE(3147), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 13,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [126133] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(692), 1,
      sym_pow_operator,
    STATE(693), 1,
      sym_mult_operator,
    STATE(694), 1,
      sym_add_operator,
    STATE(695), 1,
      sym_and_operator,
    STATE(696), 1,
      sym_or_operator,
    STATE(3148), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2862), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 13,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [126189] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    STATE(692), 1,
      sym_pow_operator,
    STATE(693), 1,
      sym_mult_operator,
    STATE(694), 1,
      sym_add_operator,
    STATE(695), 1,
      sym_and_operator,
    STATE(696), 1,
      sym_or_operator,
    STATE(3149), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2862), 6,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2860), 9,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [126249] = 18,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2862), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4267), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4269), 1,
      sym_concat_operator,
    ACTIONS(4271), 1,
      sym_rel_operator,
    STATE(692), 1,
      sym_pow_operator,
    STATE(693), 1,
      sym_mult_operator,
    STATE(694), 1,
      sym_add_operator,
    STATE(695), 1,
      sym_and_operator,
    STATE(696), 1,
      sym_or_operator,
    STATE(3150), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2860), 6,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      anon_sym_or,
      sym_assign_operator,
  [126319] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3151), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1713), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1715), 26,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [126363] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4267), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4269), 1,
      sym_concat_operator,
    ACTIONS(4271), 1,
      sym_rel_operator,
    ACTIONS(4334), 1,
      anon_sym_COMMA,
    ACTIONS(4340), 1,
      sym_assign_operator,
    STATE(692), 1,
      sym_pow_operator,
    STATE(693), 1,
      sym_mult_operator,
    STATE(694), 1,
      sym_add_operator,
    STATE(695), 1,
      sym_and_operator,
    STATE(696), 1,
      sym_or_operator,
    STATE(3152), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2824), 3,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [126439] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4267), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4269), 1,
      sym_concat_operator,
    ACTIONS(4271), 1,
      sym_rel_operator,
    ACTIONS(4334), 1,
      anon_sym_COMMA,
    ACTIONS(4340), 1,
      sym_assign_operator,
    STATE(692), 1,
      sym_pow_operator,
    STATE(693), 1,
      sym_mult_operator,
    STATE(694), 1,
      sym_add_operator,
    STATE(695), 1,
      sym_and_operator,
    STATE(696), 1,
      sym_or_operator,
    STATE(3153), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2864), 3,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [126515] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4267), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4269), 1,
      sym_concat_operator,
    ACTIONS(4271), 1,
      sym_rel_operator,
    ACTIONS(4334), 1,
      anon_sym_COMMA,
    ACTIONS(4340), 1,
      sym_assign_operator,
    STATE(692), 1,
      sym_pow_operator,
    STATE(693), 1,
      sym_mult_operator,
    STATE(694), 1,
      sym_add_operator,
    STATE(695), 1,
      sym_and_operator,
    STATE(696), 1,
      sym_or_operator,
    STATE(3154), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2888), 3,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [126591] = 23,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3237), 1,
      anon_sym_LBRACE,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    STATE(3155), 1,
      sym_attribute,
    STATE(3181), 1,
      sym_type_variable,
    STATE(3261), 1,
      sym__simple_type,
    STATE(3540), 1,
      sym_record_declaration,
    STATE(5331), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(7234), 1,
      sym__constructor_argument,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 8,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
  [126671] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4074), 1,
      sym__lowercase_identifier,
    ACTIONS(4121), 1,
      anon_sym_POUND,
    ACTIONS(4346), 1,
      anon_sym_DASH_GT,
    ACTIONS(4348), 1,
      anon_sym_STAR,
    STATE(3055), 1,
      sym_type_constructor_path,
    STATE(3156), 1,
      sym_attribute,
    STATE(4651), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(6029), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4342), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4344), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [126731] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3157), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4247), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4245), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [126775] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(4010), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4012), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4352), 1,
      anon_sym_let,
    STATE(3158), 1,
      sym_attribute,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3826), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4350), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [126835] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4267), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4269), 1,
      sym_concat_operator,
    ACTIONS(4271), 1,
      sym_rel_operator,
    ACTIONS(4334), 1,
      anon_sym_COMMA,
    ACTIONS(4340), 1,
      sym_assign_operator,
    ACTIONS(4354), 1,
      anon_sym_SEMI,
    STATE(692), 1,
      sym_pow_operator,
    STATE(693), 1,
      sym_mult_operator,
    STATE(694), 1,
      sym_add_operator,
    STATE(695), 1,
      sym_and_operator,
    STATE(696), 1,
      sym_or_operator,
    STATE(3159), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2708), 2,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [126913] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4267), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4269), 1,
      sym_concat_operator,
    ACTIONS(4271), 1,
      sym_rel_operator,
    ACTIONS(4334), 1,
      anon_sym_COMMA,
    ACTIONS(4340), 1,
      sym_assign_operator,
    ACTIONS(4354), 1,
      anon_sym_SEMI,
    STATE(692), 1,
      sym_pow_operator,
    STATE(693), 1,
      sym_mult_operator,
    STATE(694), 1,
      sym_add_operator,
    STATE(695), 1,
      sym_and_operator,
    STATE(696), 1,
      sym_or_operator,
    STATE(3160), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2716), 2,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [126991] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4267), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4269), 1,
      sym_concat_operator,
    ACTIONS(4271), 1,
      sym_rel_operator,
    ACTIONS(4334), 1,
      anon_sym_COMMA,
    ACTIONS(4340), 1,
      sym_assign_operator,
    ACTIONS(4354), 1,
      anon_sym_SEMI,
    STATE(692), 1,
      sym_pow_operator,
    STATE(693), 1,
      sym_mult_operator,
    STATE(694), 1,
      sym_add_operator,
    STATE(695), 1,
      sym_and_operator,
    STATE(696), 1,
      sym_or_operator,
    STATE(3161), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2852), 2,
      anon_sym_else,
      anon_sym_do,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [127069] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3162), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4358), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4356), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [127113] = 19,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(259), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(261), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(2434), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(3787), 1,
      anon_sym_open,
    ACTIONS(4137), 1,
      anon_sym_external,
    ACTIONS(4139), 1,
      anon_sym_type,
    ACTIONS(4143), 1,
      anon_sym_exception,
    ACTIONS(4145), 1,
      anon_sym_module,
    ACTIONS(4147), 1,
      anon_sym_include,
    ACTIONS(4149), 1,
      anon_sym_class,
    ACTIONS(4151), 1,
      anon_sym_val,
    STATE(3163), 1,
      sym_attribute,
    STATE(4207), 1,
      aux_sym__structure_repeat1,
    STATE(4222), 1,
      sym__signature_item,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4360), 2,
      anon_sym_RBRACK,
      anon_sym_end,
    STATE(4240), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [127185] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3164), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4364), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4362), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [127229] = 23,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3688), 1,
      sym_assign_operator,
    ACTIONS(3959), 1,
      sym_rel_operator,
    ACTIONS(4320), 1,
      anon_sym_RBRACK,
    ACTIONS(4366), 1,
      anon_sym_SEMI,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(3165), 1,
      sym_attribute,
    STATE(5618), 1,
      aux_sym__sequence_expression_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [127309] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3166), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1801), 9,
      anon_sym_and,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_POUND,
      anon_sym_when,
      sym__lowercase_identifier,
    ACTIONS(1803), 20,
      ts_builtin_sym_end,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      sym_hash_operator,
      sym__uppercase_identifier,
  [127353] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3167), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4370), 5,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4368), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [127397] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4273), 1,
      anon_sym_else,
    STATE(2955), 1,
      sym_else_clause,
    STATE(3168), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3271), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3269), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [127445] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(4010), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4012), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4374), 1,
      anon_sym_let,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(3169), 1,
      sym_attribute,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3827), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4372), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [127505] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4074), 1,
      sym__lowercase_identifier,
    ACTIONS(4121), 1,
      anon_sym_POUND,
    ACTIONS(4346), 1,
      anon_sym_DASH_GT,
    ACTIONS(4348), 1,
      anon_sym_STAR,
    STATE(3055), 1,
      sym_type_constructor_path,
    STATE(3170), 1,
      sym_attribute,
    STATE(4651), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(6029), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4376), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4378), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [127565] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(4267), 1,
      anon_sym_COLON_COLON,
    STATE(692), 1,
      sym_pow_operator,
    STATE(693), 1,
      sym_mult_operator,
    STATE(694), 1,
      sym_add_operator,
    STATE(695), 1,
      sym_and_operator,
    STATE(696), 1,
      sym_or_operator,
    STATE(3171), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2884), 3,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
    ACTIONS(2882), 8,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_else,
      anon_sym_do,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [127629] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    STATE(3172), 1,
      sym_attribute,
    STATE(3550), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4382), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3440), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    ACTIONS(4380), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [127684] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3672), 1,
      anon_sym_LT,
    ACTIONS(3674), 1,
      anon_sym_POUND,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4392), 1,
      sym__lowercase_identifier,
    ACTIONS(4394), 1,
      anon_sym_LPAREN,
    ACTIONS(4396), 1,
      anon_sym_DOT_DOT,
    ACTIONS(4398), 1,
      anon_sym_GT,
    STATE(3173), 1,
      sym_attribute,
    STATE(4630), 1,
      sym__simple_type,
    STATE(5574), 1,
      sym_method_type,
    STATE(5999), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4452), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [127759] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3672), 1,
      anon_sym_LT,
    ACTIONS(3674), 1,
      anon_sym_POUND,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4392), 1,
      sym__lowercase_identifier,
    ACTIONS(4394), 1,
      anon_sym_LPAREN,
    ACTIONS(4400), 1,
      anon_sym_DOT_DOT,
    ACTIONS(4402), 1,
      anon_sym_GT,
    STATE(3174), 1,
      sym_attribute,
    STATE(4757), 1,
      sym__simple_type,
    STATE(5994), 1,
      sym_method_type,
    STATE(5999), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4452), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [127834] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3688), 1,
      sym_assign_operator,
    ACTIONS(3959), 1,
      sym_rel_operator,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(3175), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(4404), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [127909] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4408), 1,
      anon_sym_let,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(4414), 1,
      anon_sym_COLON2,
    STATE(3176), 1,
      sym_attribute,
    STATE(3197), 1,
      aux_sym_module_binding_repeat1,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3515), 1,
      sym__module_typed,
    STATE(3700), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4410), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(4406), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [127970] = 20,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4416), 1,
      sym__lowercase_identifier,
    ACTIONS(4418), 1,
      anon_sym_LPAREN,
    STATE(3177), 1,
      sym_attribute,
    STATE(4754), 1,
      sym__simple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(7016), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [128043] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3672), 1,
      anon_sym_LT,
    ACTIONS(3674), 1,
      anon_sym_POUND,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4392), 1,
      sym__lowercase_identifier,
    ACTIONS(4394), 1,
      anon_sym_LPAREN,
    ACTIONS(4420), 1,
      anon_sym_DOT_DOT,
    ACTIONS(4422), 1,
      anon_sym_GT,
    STATE(3178), 1,
      sym_attribute,
    STATE(4757), 1,
      sym__simple_type,
    STATE(5994), 1,
      sym_method_type,
    STATE(5999), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4452), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [128118] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4074), 1,
      sym__lowercase_identifier,
    ACTIONS(4121), 1,
      anon_sym_POUND,
    ACTIONS(4426), 1,
      anon_sym_STAR,
    STATE(3055), 1,
      sym_type_constructor_path,
    STATE(3179), 1,
      sym_attribute,
    STATE(3518), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(6029), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4346), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4424), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [128175] = 20,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4416), 1,
      sym__lowercase_identifier,
    ACTIONS(4428), 1,
      anon_sym_LPAREN,
    STATE(3180), 1,
      sym_attribute,
    STATE(4783), 1,
      sym__simple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(6529), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [128248] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3181), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4300), 2,
      anon_sym_DOT,
      aux_sym_type_variable_token1,
    ACTIONS(3775), 11,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_in,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(3773), 15,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [128293] = 21,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3198), 1,
      anon_sym_LBRACE,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    STATE(3182), 1,
      sym_attribute,
    STATE(4394), 1,
      sym__simple_type,
    STATE(5077), 1,
      sym__constructor_argument,
    STATE(5079), 1,
      sym_record_declaration,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [128368] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3686), 1,
      sym_rel_operator,
    ACTIONS(3688), 1,
      sym_assign_operator,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(3183), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(4430), 2,
      anon_sym_SEMI,
      anon_sym_GT_RBRACE,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [128443] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4432), 1,
      anon_sym_else,
    STATE(2918), 1,
      sym_else_clause,
    STATE(3184), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3277), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3275), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [128490] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4432), 1,
      anon_sym_else,
    STATE(2955), 1,
      sym_else_clause,
    STATE(3185), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3271), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3269), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [128537] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4434), 1,
      anon_sym_else,
    STATE(2918), 1,
      sym_else_clause,
    STATE(3186), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3277), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3275), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [128584] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4436), 1,
      anon_sym_PIPE,
    STATE(3187), 1,
      sym_attribute,
    STATE(3188), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3208), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3206), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [128631] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4438), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3188), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3212), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3210), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [128676] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4436), 1,
      anon_sym_PIPE,
    STATE(3187), 1,
      aux_sym__match_cases_repeat1,
    STATE(3189), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3259), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3257), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [128723] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4436), 1,
      anon_sym_PIPE,
    STATE(3188), 1,
      aux_sym__match_cases_repeat1,
    STATE(3190), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3259), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3257), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [128770] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3672), 1,
      anon_sym_LT,
    ACTIONS(3674), 1,
      anon_sym_POUND,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4392), 1,
      sym__lowercase_identifier,
    ACTIONS(4394), 1,
      anon_sym_LPAREN,
    ACTIONS(4441), 1,
      anon_sym_DOT_DOT,
    ACTIONS(4443), 1,
      anon_sym_GT,
    STATE(3191), 1,
      sym_attribute,
    STATE(4653), 1,
      sym__simple_type,
    STATE(5732), 1,
      sym_method_type,
    STATE(5999), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4452), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [128845] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2854), 1,
      anon_sym_DASH_GT,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3688), 1,
      sym_assign_operator,
    ACTIONS(3959), 1,
      sym_rel_operator,
    ACTIONS(4445), 1,
      anon_sym_SEMI,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(3192), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [128922] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1703), 1,
      anon_sym_DOT,
    STATE(3193), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1699), 11,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1701), 16,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [128967] = 21,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3237), 1,
      anon_sym_LBRACE,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    STATE(3194), 1,
      sym_attribute,
    STATE(3249), 1,
      sym__simple_type,
    STATE(3540), 1,
      sym_record_declaration,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(7265), 1,
      sym__constructor_argument,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [129042] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3672), 1,
      anon_sym_LT,
    ACTIONS(3674), 1,
      anon_sym_POUND,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4392), 1,
      sym__lowercase_identifier,
    ACTIONS(4394), 1,
      anon_sym_LPAREN,
    ACTIONS(4447), 1,
      anon_sym_DOT_DOT,
    ACTIONS(4449), 1,
      anon_sym_GT,
    STATE(3195), 1,
      sym_attribute,
    STATE(4757), 1,
      sym__simple_type,
    STATE(5994), 1,
      sym_method_type,
    STATE(5999), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4452), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [129117] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3672), 1,
      anon_sym_LT,
    ACTIONS(3674), 1,
      anon_sym_POUND,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4392), 1,
      sym__lowercase_identifier,
    ACTIONS(4394), 1,
      anon_sym_LPAREN,
    ACTIONS(4451), 1,
      anon_sym_DOT_DOT,
    ACTIONS(4453), 1,
      anon_sym_GT,
    STATE(3196), 1,
      sym_attribute,
    STATE(4757), 1,
      sym__simple_type,
    STATE(5994), 1,
      sym_method_type,
    STATE(5999), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4452), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [129192] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(4414), 1,
      anon_sym_COLON2,
    ACTIONS(4457), 1,
      anon_sym_let,
    STATE(3197), 1,
      sym_attribute,
    STATE(3310), 1,
      aux_sym_module_binding_repeat1,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3506), 1,
      sym__module_typed,
    STATE(3677), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4459), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(4455), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [129253] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4052), 1,
      anon_sym_COMMA,
    STATE(3198), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2880), 9,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2878), 18,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [129298] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3199), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1725), 4,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1727), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [129341] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3672), 1,
      anon_sym_LT,
    ACTIONS(3674), 1,
      anon_sym_POUND,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4392), 1,
      sym__lowercase_identifier,
    ACTIONS(4394), 1,
      anon_sym_LPAREN,
    ACTIONS(4461), 1,
      anon_sym_DOT_DOT,
    ACTIONS(4463), 1,
      anon_sym_GT,
    STATE(3200), 1,
      sym_attribute,
    STATE(4625), 1,
      sym__simple_type,
    STATE(5895), 1,
      sym_method_type,
    STATE(5999), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4452), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [129416] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3201), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 4,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1771), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [129459] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3202), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1789), 4,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1791), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [129502] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4434), 1,
      anon_sym_else,
    STATE(2955), 1,
      sym_else_clause,
    STATE(3203), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3271), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3269), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [129549] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3672), 1,
      anon_sym_LT,
    ACTIONS(3674), 1,
      anon_sym_POUND,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4392), 1,
      sym__lowercase_identifier,
    ACTIONS(4394), 1,
      anon_sym_LPAREN,
    ACTIONS(4465), 1,
      anon_sym_DOT_DOT,
    ACTIONS(4467), 1,
      anon_sym_GT,
    STATE(3204), 1,
      sym_attribute,
    STATE(4757), 1,
      sym__simple_type,
    STATE(5994), 1,
      sym_method_type,
    STATE(5999), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4452), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [129624] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3672), 1,
      anon_sym_LT,
    ACTIONS(3674), 1,
      anon_sym_POUND,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4392), 1,
      sym__lowercase_identifier,
    ACTIONS(4394), 1,
      anon_sym_LPAREN,
    ACTIONS(4469), 1,
      anon_sym_DOT_DOT,
    ACTIONS(4471), 1,
      anon_sym_GT,
    STATE(3205), 1,
      sym_attribute,
    STATE(4757), 1,
      sym__simple_type,
    STATE(5994), 1,
      sym_method_type,
    STATE(5999), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4452), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [129699] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3688), 1,
      sym_assign_operator,
    ACTIONS(3959), 1,
      sym_rel_operator,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(3206), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(4473), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [129774] = 21,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3237), 1,
      anon_sym_LBRACE,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    STATE(3207), 1,
      sym_attribute,
    STATE(3249), 1,
      sym__simple_type,
    STATE(3540), 1,
      sym_record_declaration,
    STATE(6301), 1,
      sym_extended_module_path,
    STATE(6789), 1,
      sym__constructor_argument,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [129849] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3237), 1,
      anon_sym_LBRACE,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3698), 1,
      anon_sym_LBRACK,
    ACTIONS(4310), 1,
      sym__lowercase_identifier,
    ACTIONS(4312), 1,
      anon_sym_LPAREN,
    STATE(3170), 1,
      sym__simple_type,
    STATE(3208), 1,
      sym_attribute,
    STATE(3540), 1,
      sym_record_declaration,
    STATE(6244), 1,
      sym_extended_module_path,
    STATE(6341), 1,
      sym__constructor_argument,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [129924] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3622), 1,
      sym__lowercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    ACTIONS(4475), 1,
      anon_sym_DASH_GT,
    ACTIONS(4477), 1,
      anon_sym_STAR,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(3209), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3987), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(3989), 14,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
  [129981] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3622), 1,
      sym__lowercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    ACTIONS(4479), 1,
      anon_sym_STAR,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(3210), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4131), 8,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4133), 14,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
  [130036] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4227), 1,
      anon_sym_COMMA,
    ACTIONS(4229), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4231), 1,
      sym_concat_operator,
    ACTIONS(4233), 1,
      sym_rel_operator,
    ACTIONS(4235), 1,
      sym_assign_operator,
    STATE(606), 1,
      sym_pow_operator,
    STATE(607), 1,
      sym_mult_operator,
    STATE(608), 1,
      sym_add_operator,
    STATE(609), 1,
      sym_and_operator,
    STATE(610), 1,
      sym_or_operator,
    STATE(3211), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(4481), 2,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [130111] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2708), 1,
      anon_sym_do,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4267), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4269), 1,
      sym_concat_operator,
    ACTIONS(4271), 1,
      sym_rel_operator,
    ACTIONS(4334), 1,
      anon_sym_COMMA,
    ACTIONS(4340), 1,
      sym_assign_operator,
    ACTIONS(4483), 1,
      anon_sym_SEMI,
    STATE(692), 1,
      sym_pow_operator,
    STATE(693), 1,
      sym_mult_operator,
    STATE(694), 1,
      sym_add_operator,
    STATE(695), 1,
      sym_and_operator,
    STATE(696), 1,
      sym_or_operator,
    STATE(3212), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [130188] = 21,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3237), 1,
      anon_sym_LBRACE,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    STATE(3213), 1,
      sym_attribute,
    STATE(3419), 1,
      sym__simple_type,
    STATE(3540), 1,
      sym_record_declaration,
    STATE(3593), 1,
      sym__constructor_argument,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [130263] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2710), 1,
      anon_sym_DASH_GT,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3688), 1,
      sym_assign_operator,
    ACTIONS(3959), 1,
      sym_rel_operator,
    ACTIONS(4445), 1,
      anon_sym_SEMI,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(3214), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [130340] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2716), 1,
      anon_sym_do,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(4267), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4269), 1,
      sym_concat_operator,
    ACTIONS(4271), 1,
      sym_rel_operator,
    ACTIONS(4334), 1,
      anon_sym_COMMA,
    ACTIONS(4340), 1,
      sym_assign_operator,
    ACTIONS(4483), 1,
      anon_sym_SEMI,
    STATE(692), 1,
      sym_pow_operator,
    STATE(693), 1,
      sym_mult_operator,
    STATE(694), 1,
      sym_add_operator,
    STATE(695), 1,
      sym_and_operator,
    STATE(696), 1,
      sym_or_operator,
    STATE(3215), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [130417] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_DOT,
    STATE(3216), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1699), 11,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1701), 16,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [130462] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2718), 1,
      anon_sym_DASH_GT,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3688), 1,
      sym_assign_operator,
    ACTIONS(3959), 1,
      sym_rel_operator,
    ACTIONS(4445), 1,
      anon_sym_SEMI,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(3217), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [130539] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4436), 1,
      anon_sym_PIPE,
    STATE(3190), 1,
      aux_sym__match_cases_repeat1,
    STATE(3218), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3180), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3178), 19,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [130586] = 20,
    ACTIONS(9), 1,
      anon_sym_QMARK,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4416), 1,
      sym__lowercase_identifier,
    ACTIONS(4418), 1,
      anon_sym_LPAREN,
    STATE(3219), 1,
      sym_attribute,
    STATE(4779), 1,
      sym__simple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(7016), 2,
      sym_labeled_argument_type,
      sym__proper_tuple_type,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [130659] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3220), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1801), 4,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(1803), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [130702] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    ACTIONS(3682), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3684), 1,
      sym_concat_operator,
    ACTIONS(3688), 1,
      sym_assign_operator,
    ACTIONS(3959), 1,
      sym_rel_operator,
    STATE(590), 1,
      sym_pow_operator,
    STATE(591), 1,
      sym_mult_operator,
    STATE(592), 1,
      sym_add_operator,
    STATE(593), 1,
      sym_and_operator,
    STATE(594), 1,
      sym_or_operator,
    STATE(3221), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(4481), 2,
      anon_sym_RBRACK,
      anon_sym_SEMI,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [130777] = 22,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2838), 1,
      aux_sym_mult_operator_token1,
    ACTIONS(2846), 1,
      anon_sym_or,
    ACTIONS(2848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2852), 1,
      anon_sym_do,
    ACTIONS(4267), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4269), 1,
      sym_concat_operator,
    ACTIONS(4271), 1,
      sym_rel_operator,
    ACTIONS(4334), 1,
      anon_sym_COMMA,
    ACTIONS(4340), 1,
      sym_assign_operator,
    ACTIONS(4483), 1,
      anon_sym_SEMI,
    STATE(692), 1,
      sym_pow_operator,
    STATE(693), 1,
      sym_mult_operator,
    STATE(694), 1,
      sym_add_operator,
    STATE(695), 1,
      sym_and_operator,
    STATE(696), 1,
      sym_or_operator,
    STATE(3222), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2830), 2,
      anon_sym_AMP,
      anon_sym_AMP_AMP,
    ACTIONS(2834), 3,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_add_operator_token1,
    ACTIONS(2836), 4,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
    ACTIONS(2840), 4,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
  [130854] = 21,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3237), 1,
      anon_sym_LBRACE,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3698), 1,
      anon_sym_LBRACK,
    ACTIONS(4310), 1,
      sym__lowercase_identifier,
    ACTIONS(4312), 1,
      anon_sym_LPAREN,
    STATE(3179), 1,
      sym__simple_type,
    STATE(3223), 1,
      sym_attribute,
    STATE(3540), 1,
      sym_record_declaration,
    STATE(3593), 1,
      sym__constructor_argument,
    STATE(6244), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [130929] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3224), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1727), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(1725), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [130971] = 20,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    ACTIONS(4485), 1,
      sym__lowercase_identifier,
    STATE(2641), 1,
      sym__tuple_type_rhs,
    STATE(3225), 1,
      sym_attribute,
    STATE(3846), 1,
      sym__simple_type,
    STATE(4205), 1,
      sym_labeled_tuple_element_type,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [131043] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3226), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1771), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(1769), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [131085] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3227), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1791), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(1789), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [131127] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3052), 1,
      anon_sym_PLUS_EQ,
    ACTIONS(4489), 1,
      anon_sym_let,
    ACTIONS(4491), 1,
      anon_sym_EQ,
    ACTIONS(4493), 1,
      anon_sym_COLON_EQ,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    STATE(3228), 1,
      sym_attribute,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3482), 1,
      aux_sym_type_binding_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3756), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4487), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [131187] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(3229), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1713), 11,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1715), 16,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [131229] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(3230), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1623), 11,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1625), 16,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [131271] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3231), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1803), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(1801), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [131313] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(3232), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1809), 11,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1811), 16,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [131355] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3233), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3640), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3638), 25,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [131397] = 20,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3672), 1,
      anon_sym_LT,
    ACTIONS(3674), 1,
      anon_sym_POUND,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4394), 1,
      anon_sym_LPAREN,
    ACTIONS(4497), 1,
      sym__lowercase_identifier,
    STATE(3234), 1,
      sym_attribute,
    STATE(4250), 1,
      sym__simple_type,
    STATE(4251), 1,
      sym__tuple_type_rhs,
    STATE(4621), 1,
      sym_labeled_tuple_element_type,
    STATE(5999), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4452), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [131469] = 20,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3672), 1,
      anon_sym_LT,
    ACTIONS(3674), 1,
      anon_sym_POUND,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4394), 1,
      anon_sym_LPAREN,
    ACTIONS(4497), 1,
      sym__lowercase_identifier,
    STATE(3235), 1,
      sym_attribute,
    STATE(4250), 1,
      sym__simple_type,
    STATE(4260), 1,
      sym__tuple_type_rhs,
    STATE(4621), 1,
      sym_labeled_tuple_element_type,
    STATE(5999), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4452), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [131541] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4499), 1,
      anon_sym_else,
    STATE(2955), 1,
      sym_else_clause,
    STATE(3236), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3271), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3269), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [131587] = 20,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    ACTIONS(4501), 1,
      sym__lowercase_identifier,
    STATE(2652), 1,
      sym__tuple_type_rhs,
    STATE(3210), 1,
      sym__simple_type,
    STATE(3237), 1,
      sym_attribute,
    STATE(3519), 1,
      sym_labeled_tuple_element_type,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [131659] = 20,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    ACTIONS(4501), 1,
      sym__lowercase_identifier,
    STATE(2637), 1,
      sym__tuple_type_rhs,
    STATE(3210), 1,
      sym__simple_type,
    STATE(3238), 1,
      sym_attribute,
    STATE(3519), 1,
      sym_labeled_tuple_element_type,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [131731] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3239), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4505), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4507), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4503), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [131775] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3052), 1,
      anon_sym_PLUS_EQ,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4511), 1,
      anon_sym_let,
    ACTIONS(4513), 1,
      anon_sym_EQ,
    ACTIONS(4515), 1,
      anon_sym_COLON_EQ,
    STATE(3240), 1,
      sym_attribute,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3453), 1,
      aux_sym_type_binding_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3645), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4509), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [131835] = 20,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    ACTIONS(4485), 1,
      sym__lowercase_identifier,
    STATE(3241), 1,
      sym_attribute,
    STATE(4179), 1,
      sym__simple_type,
    STATE(4274), 1,
      sym__tuple_type_rhs,
    STATE(4531), 1,
      sym_labeled_tuple_element_type,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [131907] = 20,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3672), 1,
      anon_sym_LT,
    ACTIONS(3674), 1,
      anon_sym_POUND,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4394), 1,
      anon_sym_LPAREN,
    ACTIONS(4497), 1,
      sym__lowercase_identifier,
    STATE(3242), 1,
      sym_attribute,
    STATE(4250), 1,
      sym__simple_type,
    STATE(4274), 1,
      sym__tuple_type_rhs,
    STATE(4621), 1,
      sym_labeled_tuple_element_type,
    STATE(5999), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4452), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [131979] = 20,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    ACTIONS(4517), 1,
      sym__lowercase_identifier,
    STATE(2652), 1,
      sym__tuple_type_rhs,
    STATE(3243), 1,
      sym_attribute,
    STATE(3844), 1,
      sym__simple_type,
    STATE(4201), 1,
      sym_labeled_tuple_element_type,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [132051] = 20,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3698), 1,
      anon_sym_LBRACK,
    ACTIONS(4312), 1,
      anon_sym_LPAREN,
    ACTIONS(4519), 1,
      sym__lowercase_identifier,
    STATE(2641), 1,
      sym__tuple_type_rhs,
    STATE(2938), 1,
      sym__simple_type,
    STATE(3244), 1,
      sym_attribute,
    STATE(3295), 1,
      sym_labeled_tuple_element_type,
    STATE(6244), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [132123] = 20,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    ACTIONS(4517), 1,
      sym__lowercase_identifier,
    STATE(2641), 1,
      sym__tuple_type_rhs,
    STATE(3245), 1,
      sym_attribute,
    STATE(3844), 1,
      sym__simple_type,
    STATE(4201), 1,
      sym_labeled_tuple_element_type,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [132195] = 20,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    ACTIONS(4501), 1,
      sym__lowercase_identifier,
    STATE(2641), 1,
      sym__tuple_type_rhs,
    STATE(3210), 1,
      sym__simple_type,
    STATE(3246), 1,
      sym_attribute,
    STATE(3519), 1,
      sym_labeled_tuple_element_type,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [132267] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(4527), 1,
      anon_sym_DOT,
    STATE(3247), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4523), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4521), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [132313] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3248), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4531), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4533), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4529), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [132357] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3622), 1,
      sym__lowercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    ACTIONS(4346), 1,
      anon_sym_DASH_GT,
    ACTIONS(4348), 1,
      anon_sym_STAR,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(3249), 1,
      sym_attribute,
    STATE(4651), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4376), 6,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4378), 13,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [132415] = 20,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    ACTIONS(4485), 1,
      sym__lowercase_identifier,
    STATE(3250), 1,
      sym_attribute,
    STATE(4179), 1,
      sym__simple_type,
    STATE(4251), 1,
      sym__tuple_type_rhs,
    STATE(4531), 1,
      sym_labeled_tuple_element_type,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [132487] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4535), 1,
      anon_sym_COLON2,
    STATE(3251), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3052), 11,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3050), 15,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [132531] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3252), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4537), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4533), 25,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [132573] = 20,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    ACTIONS(4485), 1,
      sym__lowercase_identifier,
    STATE(3253), 1,
      sym_attribute,
    STATE(4179), 1,
      sym__simple_type,
    STATE(4260), 1,
      sym__tuple_type_rhs,
    STATE(4531), 1,
      sym_labeled_tuple_element_type,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [132645] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4543), 1,
      anon_sym_as,
    STATE(3254), 1,
      sym_attribute,
    STATE(3274), 1,
      aux_sym_type_binding_repeat1,
    STATE(3309), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4541), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4539), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [132695] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4064), 1,
      anon_sym_COMMA,
    STATE(3255), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2880), 8,
      anon_sym_and,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2878), 18,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_LBRACK_AT_AT,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
      sym_let_and_operator,
  [132739] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4543), 1,
      anon_sym_as,
    STATE(3256), 1,
      sym_attribute,
    STATE(3275), 1,
      aux_sym_type_binding_repeat1,
    STATE(3309), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4547), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4545), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [132789] = 20,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    ACTIONS(4485), 1,
      sym__lowercase_identifier,
    STATE(2637), 1,
      sym__tuple_type_rhs,
    STATE(3257), 1,
      sym_attribute,
    STATE(3846), 1,
      sym__simple_type,
    STATE(4205), 1,
      sym_labeled_tuple_element_type,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [132861] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4543), 1,
      anon_sym_as,
    STATE(3258), 1,
      sym_attribute,
    STATE(3276), 1,
      aux_sym_type_binding_repeat1,
    STATE(3309), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4551), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4549), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [132911] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4080), 1,
      anon_sym_COMMA,
    STATE(3259), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2880), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2878), 18,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [132955] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3260), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4553), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4507), 25,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [132997] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3622), 1,
      sym__lowercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    ACTIONS(4346), 1,
      anon_sym_DASH_GT,
    ACTIONS(4348), 1,
      anon_sym_STAR,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(3261), 1,
      sym_attribute,
    STATE(4651), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4342), 6,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4344), 13,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [133055] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4543), 1,
      anon_sym_as,
    STATE(3262), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4557), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4555), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [133099] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3263), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4561), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4559), 25,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [133141] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4074), 1,
      sym__lowercase_identifier,
    ACTIONS(4121), 1,
      anon_sym_POUND,
    STATE(3055), 1,
      sym_type_constructor_path,
    STATE(3264), 1,
      sym_attribute,
    STATE(6029), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4563), 8,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4565), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [133193] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(3265), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1843), 11,
      anon_sym_PIPE,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_as,
      anon_sym_lazy,
      anon_sym_begin,
      anon_sym_effect,
      sym__lowercase_identifier,
    ACTIONS(1845), 16,
      anon_sym_TILDE,
      anon_sym_LPAREN,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_PIPE,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      aux_sym_signed_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [133235] = 20,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3698), 1,
      anon_sym_LBRACK,
    ACTIONS(4312), 1,
      anon_sym_LPAREN,
    ACTIONS(4519), 1,
      sym__lowercase_identifier,
    STATE(2652), 1,
      sym__tuple_type_rhs,
    STATE(2938), 1,
      sym__simple_type,
    STATE(3266), 1,
      sym_attribute,
    STATE(3295), 1,
      sym_labeled_tuple_element_type,
    STATE(6244), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [133307] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4499), 1,
      anon_sym_else,
    STATE(2918), 1,
      sym_else_clause,
    STATE(3267), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3277), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3275), 18,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [133353] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(259), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(261), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(2434), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(3787), 1,
      anon_sym_open,
    ACTIONS(4137), 1,
      anon_sym_external,
    ACTIONS(4139), 1,
      anon_sym_type,
    ACTIONS(4143), 1,
      anon_sym_exception,
    ACTIONS(4145), 1,
      anon_sym_module,
    ACTIONS(4147), 1,
      anon_sym_include,
    ACTIONS(4149), 1,
      anon_sym_class,
    ACTIONS(4151), 1,
      anon_sym_val,
    STATE(3268), 1,
      sym_attribute,
    STATE(4207), 1,
      aux_sym__structure_repeat1,
    STATE(4222), 1,
      sym__signature_item,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4240), 13,
      sym_external,
      sym_type_definition,
      sym_exception_definition,
      sym_module_definition,
      sym_module_type_definition,
      sym_open_module,
      sym_class_definition,
      sym_class_type_definition,
      sym_value_specification,
      sym_include_module_type,
      sym_floating_attribute,
      sym_item_extension,
      sym_quoted_item_extension,
  [133421] = 20,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    ACTIONS(4485), 1,
      sym__lowercase_identifier,
    STATE(2652), 1,
      sym__tuple_type_rhs,
    STATE(3269), 1,
      sym_attribute,
    STATE(3846), 1,
      sym__simple_type,
    STATE(4205), 1,
      sym_labeled_tuple_element_type,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [133493] = 20,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3698), 1,
      anon_sym_LBRACK,
    ACTIONS(4312), 1,
      anon_sym_LPAREN,
    ACTIONS(4519), 1,
      sym__lowercase_identifier,
    STATE(2637), 1,
      sym__tuple_type_rhs,
    STATE(2938), 1,
      sym__simple_type,
    STATE(3270), 1,
      sym_attribute,
    STATE(3295), 1,
      sym_labeled_tuple_element_type,
    STATE(6244), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [133565] = 20,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    ACTIONS(4517), 1,
      sym__lowercase_identifier,
    STATE(2637), 1,
      sym__tuple_type_rhs,
    STATE(3271), 1,
      sym_attribute,
    STATE(3844), 1,
      sym__simple_type,
    STATE(4201), 1,
      sym_labeled_tuple_element_type,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [133637] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4567), 1,
      anon_sym_PIPE,
    STATE(3272), 1,
      sym_attribute,
    STATE(3283), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3208), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3206), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [133682] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4569), 1,
      anon_sym_PIPE,
    STATE(3273), 1,
      sym_attribute,
    STATE(3306), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3259), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3257), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [133727] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    STATE(3274), 1,
      sym_attribute,
    STATE(3294), 1,
      aux_sym_type_binding_repeat1,
    STATE(3309), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4573), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4571), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [133774] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    STATE(3275), 1,
      sym_attribute,
    STATE(3294), 1,
      aux_sym_type_binding_repeat1,
    STATE(3309), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4577), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4575), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [133821] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    STATE(3276), 1,
      sym_attribute,
    STATE(3294), 1,
      aux_sym_type_binding_repeat1,
    STATE(3309), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4581), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4579), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [133868] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4567), 1,
      anon_sym_PIPE,
    STATE(3277), 1,
      sym_attribute,
    STATE(3279), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3180), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3178), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [133913] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4567), 1,
      anon_sym_PIPE,
    STATE(3272), 1,
      aux_sym__match_cases_repeat1,
    STATE(3278), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3259), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3257), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [133958] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4567), 1,
      anon_sym_PIPE,
    STATE(3279), 1,
      sym_attribute,
    STATE(3283), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3259), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3257), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [134003] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4074), 1,
      sym__lowercase_identifier,
    ACTIONS(4121), 1,
      anon_sym_POUND,
    STATE(3055), 1,
      sym_type_constructor_path,
    STATE(3280), 1,
      sym_attribute,
    STATE(6029), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4583), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4585), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [134054] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3281), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4589), 12,
      anon_sym_SEMI_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_let_operator,
      sym__uppercase_identifier,
      aux_sym_directive_token1,
    ACTIONS(4587), 14,
      anon_sym_let,
      anon_sym_external,
      anon_sym_type,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      sym__lowercase_identifier,
  [134095] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4591), 1,
      anon_sym_else,
    STATE(2117), 1,
      sym_else_clause,
    STATE(3282), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3277), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3275), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [134140] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4593), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3283), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3212), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3210), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [134183] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4569), 1,
      anon_sym_PIPE,
    STATE(3284), 1,
      sym_attribute,
    STATE(3308), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3259), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3257), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [134228] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3285), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4598), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4596), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [134269] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4074), 1,
      sym__lowercase_identifier,
    ACTIONS(4121), 1,
      anon_sym_POUND,
    STATE(3055), 1,
      sym_type_constructor_path,
    STATE(3286), 1,
      sym_attribute,
    STATE(6029), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4376), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4378), 14,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [134320] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3672), 1,
      anon_sym_LT,
    ACTIONS(3674), 1,
      anon_sym_POUND,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4392), 1,
      sym__lowercase_identifier,
    ACTIONS(4394), 1,
      anon_sym_LPAREN,
    STATE(3287), 1,
      sym_attribute,
    STATE(4757), 1,
      sym__simple_type,
    STATE(5994), 1,
      sym_method_type,
    STATE(5999), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4452), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [134389] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3288), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1699), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4507), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(1701), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [134432] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4543), 1,
      anon_sym_as,
    ACTIONS(4602), 1,
      anon_sym_let,
    ACTIONS(4604), 1,
      anon_sym_EQ,
    STATE(3289), 1,
      sym_attribute,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3444), 1,
      aux_sym_type_binding_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3708), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4600), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [134489] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4543), 1,
      anon_sym_as,
    ACTIONS(4608), 1,
      anon_sym_let,
    ACTIONS(4610), 1,
      anon_sym_EQ,
    STATE(3290), 1,
      sym_attribute,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3437), 1,
      aux_sym_type_binding_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3607), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4606), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [134546] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3291), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4614), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4612), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [134587] = 20,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4618), 1,
      anon_sym_let,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4622), 1,
      anon_sym_BANG,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4626), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3292), 1,
      sym_attribute,
    STATE(4253), 1,
      sym__class_expression,
    STATE(6430), 1,
      sym_module_path,
    STATE(6711), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [134658] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4628), 1,
      anon_sym_else,
    STATE(2270), 1,
      sym_else_clause,
    STATE(3293), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3277), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3275), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [134703] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4634), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4632), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    STATE(3294), 2,
      sym_attribute,
      aux_sym_type_binding_repeat1,
    ACTIONS(4630), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [134748] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4135), 1,
      anon_sym_STAR,
    STATE(3295), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4133), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4131), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [134791] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3296), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4300), 2,
      anon_sym_DOT,
      aux_sym_type_variable_token1,
    ACTIONS(3773), 10,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
    ACTIONS(3775), 14,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
  [134834] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4569), 1,
      anon_sym_PIPE,
    STATE(3284), 1,
      aux_sym__match_cases_repeat1,
    STATE(3297), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3180), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3178), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [134879] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3298), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1785), 3,
      anon_sym_COLON2,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1787), 23,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
  [134920] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3299), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1797), 3,
      anon_sym_COLON2,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1799), 23,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
  [134961] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3050), 1,
      sym__lowercase_identifier,
    ACTIONS(4637), 1,
      anon_sym_COLON2,
    STATE(3300), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3052), 4,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_POUND,
      sym__uppercase_identifier,
    ACTIONS(3891), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(3893), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [135008] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3622), 1,
      sym__lowercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(3301), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4563), 8,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4565), 13,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [135059] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3302), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1805), 3,
      anon_sym_COLON2,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1807), 23,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
  [135100] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3303), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1525), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1523), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [135141] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3304), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1757), 3,
      anon_sym_COLON2,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1759), 23,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
  [135182] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4099), 1,
      anon_sym_COMMA,
    STATE(3305), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2880), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2878), 18,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_done,
      anon_sym_to,
      anon_sym_downto,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [135225] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4569), 1,
      anon_sym_PIPE,
    STATE(3306), 1,
      sym_attribute,
    STATE(3308), 1,
      aux_sym__match_cases_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3208), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3206), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [135270] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4591), 1,
      anon_sym_else,
    STATE(2136), 1,
      sym_else_clause,
    STATE(3307), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3271), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3269), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [135315] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4639), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3308), 2,
      sym_attribute,
      aux_sym__match_cases_repeat1,
    ACTIONS(3212), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3210), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [135358] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3309), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4644), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4642), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [135399] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4650), 1,
      anon_sym_LPAREN,
    STATE(3449), 1,
      sym_module_parameter,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3310), 2,
      sym_attribute,
      aux_sym_module_binding_repeat1,
    ACTIONS(4648), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4646), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [135444] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4628), 1,
      anon_sym_else,
    STATE(2262), 1,
      sym_else_clause,
    STATE(3311), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3271), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3269), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [135489] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3312), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1813), 3,
      anon_sym_COLON2,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1815), 23,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_hash_operator,
  [135530] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3313), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4655), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4653), 24,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [135571] = 18,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    STATE(2495), 1,
      sym__simple_type,
    STATE(3314), 1,
      sym_attribute,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [135637] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(4657), 1,
      anon_sym_COLON2,
    STATE(3315), 1,
      sym_attribute,
    STATE(3433), 1,
      aux_sym_module_binding_repeat1,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3515), 1,
      sym__module_typed,
    STATE(3700), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4410), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(4406), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [135693] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3316), 1,
      sym_attribute,
    STATE(3630), 1,
      sym__class_expression,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [135761] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3317), 1,
      sym_attribute,
    STATE(3583), 1,
      sym__class_expression,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [135829] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4543), 1,
      anon_sym_as,
    ACTIONS(4685), 1,
      anon_sym_let,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3318), 1,
      sym_attribute,
    STATE(3443), 1,
      aux_sym_type_binding_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3707), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4683), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [135883] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3319), 1,
      sym_attribute,
    STATE(3697), 1,
      sym__class_expression,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [135951] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4618), 1,
      anon_sym_let,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4626), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3313), 1,
      sym__class_expression,
    STATE(3320), 1,
      sym_attribute,
    STATE(6430), 1,
      sym_module_path,
    STATE(6711), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [136019] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3321), 1,
      sym_attribute,
    STATE(3703), 1,
      sym__class_expression,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [136087] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3322), 1,
      sym_attribute,
    STATE(3704), 1,
      sym__class_expression,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [136155] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3323), 1,
      sym_attribute,
    STATE(3719), 1,
      sym__class_expression,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [136223] = 18,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    STATE(3301), 1,
      sym__simple_type,
    STATE(3324), 1,
      sym_attribute,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [136289] = 18,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    STATE(3325), 1,
      sym_attribute,
    STATE(3458), 1,
      sym__simple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [136355] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3775), 1,
      sym__lowercase_identifier,
    STATE(3326), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3773), 4,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_POUND,
      sym__uppercase_identifier,
    ACTIONS(4687), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(4689), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [136399] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4618), 1,
      anon_sym_let,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4626), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3327), 1,
      sym_attribute,
    STATE(5706), 1,
      sym__class_expression,
    STATE(6430), 1,
      sym_module_path,
    STATE(6711), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [136467] = 18,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    STATE(3328), 1,
      sym_attribute,
    STATE(3475), 1,
      sym__simple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [136533] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3050), 1,
      sym__lowercase_identifier,
    ACTIONS(4637), 1,
      anon_sym_COLON2,
    STATE(3329), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3052), 4,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_POUND,
      sym__uppercase_identifier,
    ACTIONS(3891), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(3893), 12,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [136579] = 18,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    STATE(2495), 1,
      sym__simple_type,
    STATE(3330), 1,
      sym_attribute,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [136645] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3331), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1769), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(1771), 16,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [136685] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3332), 1,
      sym_attribute,
    STATE(3697), 1,
      sym__class_expression,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [136753] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3333), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1789), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(1791), 16,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [136793] = 18,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3672), 1,
      anon_sym_LT,
    ACTIONS(3674), 1,
      anon_sym_POUND,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4394), 1,
      anon_sym_LPAREN,
    ACTIONS(4691), 1,
      sym__lowercase_identifier,
    STATE(3334), 1,
      sym_attribute,
    STATE(4272), 1,
      sym__simple_type,
    STATE(5999), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(4452), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [136859] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4543), 1,
      anon_sym_as,
    ACTIONS(4695), 1,
      anon_sym_let,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3335), 1,
      sym_attribute,
    STATE(3451), 1,
      aux_sym_type_binding_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3598), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4693), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [136913] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3336), 1,
      sym_attribute,
    STATE(3632), 1,
      sym__class_expression,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [136981] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3337), 1,
      sym_attribute,
    STATE(3611), 1,
      sym__class_expression,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [137049] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3338), 1,
      sym_attribute,
    STATE(3573), 1,
      sym__class_expression,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [137117] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3339), 1,
      sym_attribute,
    STATE(3577), 1,
      sym__class_expression,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [137185] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3340), 1,
      sym_attribute,
    STATE(3703), 1,
      sym__class_expression,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [137253] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3341), 1,
      sym_attribute,
    STATE(3648), 1,
      sym__class_expression,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [137321] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3342), 1,
      sym_attribute,
    STATE(3704), 1,
      sym__class_expression,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [137389] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3343), 1,
      sym_attribute,
    STATE(3716), 1,
      sym__class_expression,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [137457] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3344), 1,
      sym_attribute,
    STATE(3726), 1,
      sym__class_expression,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [137525] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3345), 1,
      sym_attribute,
    STATE(3719), 1,
      sym__class_expression,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [137593] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4699), 1,
      anon_sym_let,
    ACTIONS(4701), 1,
      anon_sym_LBRACE,
    ACTIONS(4703), 1,
      anon_sym_DQUOTE,
    STATE(3346), 1,
      sym_attribute,
    STATE(3493), 1,
      aux_sym_external_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3849), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3818), 2,
      sym_string,
      sym_quoted_string,
    ACTIONS(4697), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [137647] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3078), 1,
      sym__lowercase_identifier,
    STATE(3347), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3080), 4,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_POUND,
      sym__uppercase_identifier,
    ACTIONS(3887), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
    ACTIONS(3889), 13,
      anon_sym_let,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
      aux_sym_directive_token1,
  [137691] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4165), 1,
      anon_sym_COMMA,
    STATE(3348), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2880), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2878), 16,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [137733] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3349), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4505), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4503), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [137773] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4117), 1,
      anon_sym_DASH_GT,
    STATE(3350), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3989), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3987), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [137815] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(3351), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3321), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3319), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [137855] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3352), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4707), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4705), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [137895] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3285), 1,
      sym__class_expression,
    STATE(3353), 1,
      sym_attribute,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [137963] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3291), 1,
      sym__class_expression,
    STATE(3354), 1,
      sym_attribute,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [138031] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3313), 1,
      sym__class_expression,
    STATE(3355), 1,
      sym_attribute,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [138099] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3356), 1,
      sym_attribute,
    STATE(3748), 1,
      sym__class_expression,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [138167] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3357), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4711), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4709), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [138207] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3358), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4715), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4713), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [138247] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3359), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4531), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4529), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [138287] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3360), 1,
      sym_attribute,
    STATE(3611), 1,
      sym__class_expression,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [138355] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3361), 1,
      sym_attribute,
    STATE(3573), 1,
      sym__class_expression,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [138423] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3362), 1,
      sym_attribute,
    STATE(3748), 1,
      sym__class_expression,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [138491] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3363), 1,
      sym_attribute,
    STATE(5053), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [138559] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3364), 1,
      sym_attribute,
    STATE(3750), 1,
      sym__class_expression,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [138627] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3365), 1,
      sym_attribute,
    STATE(3750), 1,
      sym__class_expression,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [138695] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3366), 1,
      sym_attribute,
    STATE(3751), 1,
      sym__class_expression,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [138763] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3367), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4723), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4721), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [138803] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4382), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    STATE(3368), 1,
      sym_attribute,
    STATE(3881), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3600), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    ACTIONS(4380), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [138855] = 18,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3698), 1,
      anon_sym_LBRACK,
    ACTIONS(4310), 1,
      sym__lowercase_identifier,
    ACTIONS(4312), 1,
      anon_sym_LPAREN,
    STATE(3264), 1,
      sym__simple_type,
    STATE(3369), 1,
      sym_attribute,
    STATE(6244), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [138921] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3370), 1,
      sym_attribute,
    STATE(5067), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [138989] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3371), 1,
      sym_attribute,
    STATE(5073), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139057] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3372), 1,
      sym_attribute,
    STATE(5074), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139125] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3373), 1,
      sym_attribute,
    STATE(3716), 1,
      sym__class_expression,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139193] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3285), 1,
      sym__class_expression,
    STATE(3374), 1,
      sym_attribute,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139261] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3375), 1,
      sym_attribute,
    STATE(3580), 1,
      sym__class_expression,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139329] = 18,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    STATE(3301), 1,
      sym__simple_type,
    STATE(3376), 1,
      sym_attribute,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [139395] = 18,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    STATE(3377), 1,
      sym_attribute,
    STATE(3458), 1,
      sym__simple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [139461] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3378), 1,
      sym_attribute,
    STATE(3751), 1,
      sym__class_expression,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139529] = 18,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3698), 1,
      anon_sym_LBRACK,
    ACTIONS(4310), 1,
      sym__lowercase_identifier,
    ACTIONS(4312), 1,
      anon_sym_LPAREN,
    STATE(3286), 1,
      sym__simple_type,
    STATE(3379), 1,
      sym_attribute,
    STATE(6244), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [139595] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3380), 1,
      sym_attribute,
    STATE(5133), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139663] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3381), 1,
      sym_attribute,
    STATE(3726), 1,
      sym__class_expression,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139731] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3382), 1,
      sym_attribute,
    STATE(5139), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139799] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3383), 1,
      sym_attribute,
    STATE(3577), 1,
      sym__class_expression,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139867] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3291), 1,
      sym__class_expression,
    STATE(3384), 1,
      sym_attribute,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [139935] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3313), 1,
      sym__class_expression,
    STATE(3385), 1,
      sym_attribute,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140003] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3386), 1,
      sym_attribute,
    STATE(5179), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140071] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3387), 1,
      sym_attribute,
    STATE(5192), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140139] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4618), 1,
      anon_sym_let,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4626), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3388), 1,
      sym_attribute,
    STATE(5857), 1,
      sym__class_expression,
    STATE(6430), 1,
      sym_module_path,
    STATE(6711), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140207] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4618), 1,
      anon_sym_let,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4626), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3389), 1,
      sym_attribute,
    STATE(4287), 1,
      sym__class_expression,
    STATE(6430), 1,
      sym_module_path,
    STATE(6711), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140275] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3390), 1,
      sym_attribute,
    STATE(3648), 1,
      sym__class_expression,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140343] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3391), 1,
      sym_attribute,
    STATE(3647), 1,
      sym__class_expression,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140411] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3392), 1,
      sym_attribute,
    STATE(3580), 1,
      sym__class_expression,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140479] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3393), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4727), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4725), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [140519] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3394), 1,
      sym_attribute,
    STATE(5746), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140587] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3395), 1,
      sym_attribute,
    STATE(3641), 1,
      sym__class_expression,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140655] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4701), 1,
      anon_sym_LBRACE,
    ACTIONS(4703), 1,
      anon_sym_DQUOTE,
    ACTIONS(4731), 1,
      anon_sym_let,
    STATE(3396), 1,
      sym_attribute,
    STATE(3493), 1,
      aux_sym_external_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3824), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3818), 2,
      sym_string,
      sym_quoted_string,
    ACTIONS(4729), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [140709] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3397), 1,
      sym_attribute,
    STATE(5205), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140777] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3398), 1,
      sym_attribute,
    STATE(5206), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140845] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3399), 1,
      sym_attribute,
    STATE(5208), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140913] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3400), 1,
      sym_attribute,
    STATE(3644), 1,
      sym__class_expression,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [140981] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3401), 1,
      sym_attribute,
    STATE(5228), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [141049] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3402), 1,
      sym_attribute,
    STATE(5233), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [141117] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3403), 1,
      sym_attribute,
    STATE(5235), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [141185] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3404), 1,
      sym_attribute,
    STATE(5237), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [141253] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3405), 1,
      sym_attribute,
    STATE(5101), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [141321] = 18,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3239), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3241), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    STATE(2495), 1,
      sym__simple_type,
    STATE(3406), 1,
      sym_attribute,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [141387] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3407), 1,
      sym_attribute,
    STATE(5055), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [141455] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3408), 1,
      sym_attribute,
    STATE(5143), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [141523] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3409), 1,
      sym_attribute,
    STATE(5180), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [141591] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3410), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1801), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(1803), 16,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [141631] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(3411), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3212), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3210), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [141671] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4227), 1,
      anon_sym_COMMA,
    STATE(3412), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2880), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2878), 16,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [141713] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(3413), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3413), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3411), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [141753] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(3414), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3419), 8,
      anon_sym_PIPE,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3417), 17,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_else,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [141793] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3415), 1,
      sym_attribute,
    STATE(3630), 1,
      sym__class_expression,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [141861] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4733), 1,
      anon_sym_else,
    STATE(2117), 1,
      sym_else_clause,
    STATE(3416), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3277), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3275), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [141905] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3417), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4737), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4735), 23,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [141945] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3418), 1,
      sym_attribute,
    STATE(3632), 1,
      sym__class_expression,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [142013] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3622), 1,
      sym__lowercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    ACTIONS(4739), 1,
      anon_sym_STAR,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(3419), 1,
      sym_attribute,
    STATE(3807), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4346), 6,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4424), 12,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT,
  [142067] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(1867), 1,
      anon_sym_object,
    ACTIONS(2704), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2706), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4675), 1,
      sym__lowercase_identifier,
    ACTIONS(4677), 1,
      anon_sym_let,
    ACTIONS(4679), 1,
      anon_sym_LBRACK,
    ACTIONS(4681), 1,
      anon_sym_fun,
    STATE(832), 1,
      sym__simple_class_expression,
    STATE(3420), 1,
      sym_attribute,
    STATE(3641), 1,
      sym__class_expression,
    STATE(6513), 1,
      sym_value_definition,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [142135] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3421), 1,
      sym_attribute,
    STATE(5818), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [142203] = 18,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    STATE(3422), 1,
      sym_attribute,
    STATE(3475), 1,
      sym__simple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [142269] = 18,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3698), 1,
      anon_sym_LBRACK,
    ACTIONS(4310), 1,
      sym__lowercase_identifier,
    ACTIONS(4312), 1,
      anon_sym_LPAREN,
    STATE(3280), 1,
      sym__simple_type,
    STATE(3423), 1,
      sym_attribute,
    STATE(6244), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [142335] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3424), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1725), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(1727), 16,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [142375] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3425), 1,
      sym_attribute,
    STATE(3647), 1,
      sym__class_expression,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [142443] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4618), 1,
      anon_sym_let,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4626), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3285), 1,
      sym__class_expression,
    STATE(3426), 1,
      sym_attribute,
    STATE(6430), 1,
      sym_module_path,
    STATE(6711), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [142511] = 18,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3295), 1,
      anon_sym_LT,
    ACTIONS(3297), 1,
      anon_sym_POUND,
    ACTIONS(3299), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(3301), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3303), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3698), 1,
      anon_sym_LBRACK,
    ACTIONS(4310), 1,
      sym__lowercase_identifier,
    ACTIONS(4312), 1,
      anon_sym_LPAREN,
    STATE(2933), 1,
      sym__simple_type,
    STATE(3427), 1,
      sym_attribute,
    STATE(6244), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(3096), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [142577] = 18,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(19), 1,
      anon_sym_LT,
    ACTIONS(21), 1,
      anon_sym_POUND,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(4294), 1,
      anon_sym_LPAREN,
    STATE(3428), 1,
      sym_attribute,
    STATE(4871), 1,
      sym__simple_type,
    STATE(6301), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
    STATE(2574), 9,
      sym__delimited_type,
      sym_constructed_type,
      sym_local_open_type,
      sym_object_type,
      sym_hash_type,
      sym_extension,
      sym_quoted_extension,
      sym_type_constructor_path,
      sym_type_variable,
  [142643] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3429), 1,
      sym_attribute,
    STATE(3644), 1,
      sym__class_expression,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [142711] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4743), 1,
      anon_sym_let,
    ACTIONS(4745), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3430), 2,
      sym_attribute,
      aux_sym_expression_item_repeat1,
    ACTIONS(4741), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [142755] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4618), 1,
      anon_sym_let,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4626), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3291), 1,
      sym__class_expression,
    STATE(3431), 1,
      sym_attribute,
    STATE(6430), 1,
      sym_module_path,
    STATE(6711), 1,
      sym_value_definition,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [142823] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4659), 1,
      sym__lowercase_identifier,
    ACTIONS(4661), 1,
      anon_sym_let,
    ACTIONS(4663), 1,
      anon_sym_LPAREN,
    ACTIONS(4665), 1,
      anon_sym_LBRACK,
    ACTIONS(4667), 1,
      anon_sym_object,
    ACTIONS(4669), 1,
      anon_sym_fun,
    ACTIONS(4671), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4673), 1,
      anon_sym_LBRACE_PERCENT,
    STATE(749), 1,
      sym__simple_class_expression,
    STATE(3432), 1,
      sym_attribute,
    STATE(3583), 1,
      sym__class_expression,
    STATE(6516), 1,
      sym_value_definition,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3303), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2342), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [142891] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(4657), 1,
      anon_sym_COLON2,
    STATE(3310), 1,
      aux_sym_module_binding_repeat1,
    STATE(3433), 1,
      sym_attribute,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3506), 1,
      sym__module_typed,
    STATE(3677), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4459), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(4455), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [142947] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3434), 1,
      sym_attribute,
    STATE(5824), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [143015] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4733), 1,
      anon_sym_else,
    STATE(2136), 1,
      sym_else_clause,
    STATE(3435), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3271), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(3269), 16,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [143059] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(97), 1,
      sym_let_operator,
    ACTIONS(361), 1,
      anon_sym_object,
    ACTIONS(415), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(417), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4616), 1,
      sym__lowercase_identifier,
    ACTIONS(4620), 1,
      anon_sym_LPAREN,
    ACTIONS(4624), 1,
      anon_sym_LBRACK,
    ACTIONS(4717), 1,
      anon_sym_let,
    ACTIONS(4719), 1,
      anon_sym_fun,
    STATE(837), 1,
      sym__simple_class_expression,
    STATE(3436), 1,
      sym_attribute,
    STATE(5134), 1,
      sym__class_expression,
    STATE(6397), 1,
      sym_value_definition,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5829), 4,
      sym_class_function,
      sym_class_application,
      sym_let_class_expression,
      sym_let_open_class_expression,
    STATE(2097), 7,
      sym_instantiated_class,
      sym_typed_class_expression,
      sym_parenthesized_class_expression,
      sym_object_expression,
      sym_extension,
      sym_quoted_extension,
      sym_class_path,
  [143127] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4750), 1,
      anon_sym_let,
    STATE(3294), 1,
      aux_sym_type_binding_repeat1,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3437), 1,
      sym_attribute,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3710), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4748), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143178] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4754), 1,
      anon_sym_let,
    STATE(3294), 1,
      aux_sym_type_binding_repeat1,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3438), 1,
      sym_attribute,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3737), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4752), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143229] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3439), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1699), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1701), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143268] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4760), 1,
      anon_sym_EQ,
    ACTIONS(4762), 1,
      anon_sym_of,
    ACTIONS(4764), 1,
      anon_sym_COLON2,
    STATE(3440), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4758), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4756), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143313] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3441), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4768), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4766), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143352] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4334), 1,
      anon_sym_COMMA,
    STATE(3442), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2880), 7,
      anon_sym_AMP,
      aux_sym_signed_number_token1,
      aux_sym_sign_operator_token1,
      aux_sym_mult_operator_token1,
      aux_sym_add_operator_token1,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
    ACTIONS(2878), 16,
      anon_sym_SEMI,
      anon_sym_COLON_COLON,
      anon_sym_else,
      anon_sym_do,
      anon_sym_lsl,
      anon_sym_lsr,
      anon_sym_asr,
      aux_sym_pow_operator_token1,
      anon_sym_mod,
      anon_sym_land,
      anon_sym_lor,
      anon_sym_lxor,
      sym_concat_operator,
      sym_rel_operator,
      anon_sym_or,
      sym_assign_operator,
  [143393] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4772), 1,
      anon_sym_let,
    STATE(3294), 1,
      aux_sym_type_binding_repeat1,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3443), 1,
      sym_attribute,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3753), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4770), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143444] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4776), 1,
      anon_sym_let,
    STATE(3294), 1,
      aux_sym_type_binding_repeat1,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3444), 1,
      sym_attribute,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3599), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4774), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143495] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3052), 1,
      anon_sym_PLUS_EQ,
    ACTIONS(4778), 1,
      anon_sym_EQ,
    ACTIONS(4780), 1,
      anon_sym_COLON_EQ,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3445), 1,
      sym_attribute,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3663), 1,
      aux_sym_type_binding_repeat1,
    STATE(3756), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4487), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [143550] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3446), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4786), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4784), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143589] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3447), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4790), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4788), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143628] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4794), 1,
      anon_sym_let,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3448), 1,
      sym_attribute,
    STATE(3470), 1,
      aux_sym_type_binding_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3705), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4792), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143679] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3449), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4798), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4796), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143718] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3450), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4802), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4800), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143757] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4806), 1,
      anon_sym_let,
    STATE(3294), 1,
      aux_sym_type_binding_repeat1,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3451), 1,
      sym_attribute,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3702), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4804), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143808] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4810), 1,
      anon_sym_let,
    STATE(3294), 1,
      aux_sym_type_binding_repeat1,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3452), 1,
      sym_attribute,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3668), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4808), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143859] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4814), 1,
      anon_sym_let,
    STATE(3294), 1,
      aux_sym_type_binding_repeat1,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3453), 1,
      sym_attribute,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3595), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4812), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143910] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3454), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4818), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4816), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [143949] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4822), 1,
      anon_sym_let,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3452), 1,
      aux_sym_type_binding_repeat1,
    STATE(3455), 1,
      sym_attribute,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3747), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4820), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144000] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4685), 1,
      anon_sym_let,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3443), 1,
      aux_sym_type_binding_repeat1,
    STATE(3456), 1,
      sym_attribute,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3707), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4683), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144051] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3052), 1,
      anon_sym_PLUS_EQ,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    ACTIONS(4824), 1,
      anon_sym_EQ,
    ACTIONS(4826), 1,
      anon_sym_COLON_EQ,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3457), 1,
      sym_attribute,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3645), 1,
      aux_sym_expression_item_repeat1,
    STATE(3657), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4509), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [144106] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3622), 1,
      sym__lowercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(3458), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4376), 6,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4378), 13,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [144155] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4830), 1,
      anon_sym_let,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3438), 1,
      aux_sym_type_binding_repeat1,
    STATE(3459), 1,
      sym_attribute,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3671), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4828), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144206] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4602), 1,
      anon_sym_let,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3444), 1,
      aux_sym_type_binding_repeat1,
    STATE(3460), 1,
      sym_attribute,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3708), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4600), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144257] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3078), 1,
      sym__lowercase_identifier,
    STATE(3461), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3080), 4,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_POUND,
      sym__uppercase_identifier,
    ACTIONS(3887), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(3889), 12,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [144300] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3462), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4689), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4687), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144339] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4832), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4723), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    STATE(3463), 2,
      sym_attribute,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(4721), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144380] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4837), 1,
      anon_sym_let,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3464), 1,
      sym_attribute,
    STATE(3466), 1,
      aux_sym_type_binding_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3755), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4835), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144431] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3465), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4841), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4839), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144470] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4845), 1,
      anon_sym_let,
    STATE(3294), 1,
      aux_sym_type_binding_repeat1,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3466), 1,
      sym_attribute,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3689), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4843), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144521] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4851), 1,
      anon_sym_and,
    STATE(3463), 1,
      aux_sym_module_type_constraint_repeat1,
    STATE(3467), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4849), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4847), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144564] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3468), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4855), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4853), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144603] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4695), 1,
      anon_sym_let,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3451), 1,
      aux_sym_type_binding_repeat1,
    STATE(3469), 1,
      sym_attribute,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3598), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4693), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144654] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4859), 1,
      anon_sym_let,
    STATE(3294), 1,
      aux_sym_type_binding_repeat1,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3470), 1,
      sym_attribute,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3752), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4857), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144705] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4851), 1,
      anon_sym_and,
    STATE(3467), 1,
      aux_sym_module_type_constraint_repeat1,
    STATE(3471), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4863), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4861), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144748] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3472), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4867), 3,
      anon_sym_let,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4865), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144787] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3775), 1,
      sym__lowercase_identifier,
    STATE(3473), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3773), 4,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_POUND,
      sym__uppercase_identifier,
    ACTIONS(4687), 7,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4689), 12,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [144830] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4608), 1,
      anon_sym_let,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3437), 1,
      aux_sym_type_binding_repeat1,
    STATE(3474), 1,
      sym_attribute,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3607), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4606), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [144881] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3622), 1,
      sym__lowercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(3475), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4583), 6,
      anon_sym_SEMI_SEMI,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4585), 13,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [144930] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4541), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4543), 1,
      anon_sym_as,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3476), 1,
      sym_attribute,
    STATE(3520), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4539), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [144977] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3477), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4871), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4869), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145016] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4543), 1,
      anon_sym_as,
    ACTIONS(4547), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3478), 1,
      sym_attribute,
    STATE(3483), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4545), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [145063] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4543), 1,
      anon_sym_as,
    ACTIONS(4551), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3479), 1,
      sym_attribute,
    STATE(3521), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4549), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [145110] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3480), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4875), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4873), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145149] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3481), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4879), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4877), 22,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145188] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4495), 1,
      anon_sym_constraint,
    ACTIONS(4883), 1,
      anon_sym_let,
    STATE(3294), 1,
      aux_sym_type_binding_repeat1,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3482), 1,
      sym_attribute,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3717), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4881), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145239] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4577), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3483), 1,
      sym_attribute,
    STATE(3504), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4575), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [145283] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3484), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4887), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4885), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145321] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3485), 1,
      sym_attribute,
    STATE(3507), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4891), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(4889), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [145365] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3486), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4895), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4893), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145403] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3487), 1,
      sym_attribute,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4899), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(4897), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [145447] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3488), 1,
      sym_attribute,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4903), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(4901), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [145491] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3489), 1,
      sym_attribute,
    STATE(3511), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4907), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(4905), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [145535] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3490), 1,
      sym_attribute,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4911), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(4909), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [145579] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3491), 1,
      sym_attribute,
    STATE(3512), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4915), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(4913), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [145623] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3492), 1,
      sym_attribute,
    STATE(3513), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4919), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(4917), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [145667] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4925), 1,
      anon_sym_LBRACE,
    ACTIONS(4928), 1,
      anon_sym_DQUOTE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4923), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    STATE(3493), 2,
      sym_attribute,
      aux_sym_external_repeat1,
    STATE(3818), 2,
      sym_string,
      sym_quoted_string,
    ACTIONS(4921), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145711] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3494), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4933), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4931), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145749] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4935), 1,
      anon_sym_STAR,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4565), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    STATE(3495), 2,
      sym_attribute,
      aux_sym__constructor_argument_repeat1,
    ACTIONS(4563), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145789] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3496), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1745), 2,
      anon_sym_let,
      anon_sym_COLON2,
    ACTIONS(1747), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [145827] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4942), 1,
      anon_sym_RPAREN,
    ACTIONS(4944), 1,
      anon_sym__,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(4954), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3497), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5860), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [145893] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3498), 1,
      sym_attribute,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4958), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(4956), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [145937] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(4962), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3499), 1,
      sym_attribute,
    STATE(3651), 1,
      sym__attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5544), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [146003] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4968), 1,
      anon_sym_DASH_GT,
    ACTIONS(4970), 1,
      anon_sym_with,
    STATE(3500), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4966), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4964), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [146045] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3501), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1847), 2,
      anon_sym_let,
      anon_sym_COLON2,
    ACTIONS(1849), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [146083] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4942), 1,
      anon_sym_RPAREN,
    ACTIONS(4944), 1,
      anon_sym__,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(4954), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3502), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5610), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [146149] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(4972), 1,
      anon_sym_DASH_GT,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4976), 1,
      anon_sym_functor,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3503), 1,
      sym_attribute,
    STATE(3653), 1,
      sym__attribute,
    STATE(4005), 1,
      sym__module_type,
    STATE(5277), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [146215] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4632), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4984), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3504), 2,
      sym_attribute,
      aux_sym_type_binding_repeat1,
    ACTIONS(4630), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [146257] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4543), 1,
      anon_sym_as,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    ACTIONS(4987), 1,
      anon_sym_EQ,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3505), 1,
      sym_attribute,
    STATE(3607), 1,
      aux_sym_expression_item_repeat1,
    STATE(3675), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4606), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [146309] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4991), 1,
      anon_sym_let,
    ACTIONS(4993), 1,
      anon_sym_EQ,
    ACTIONS(4995), 1,
      anon_sym_COLON_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3506), 1,
      sym_attribute,
    STATE(3601), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4989), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [146357] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3507), 1,
      sym_attribute,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4999), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(4997), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [146401] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3487), 1,
      aux_sym_expression_item_repeat1,
    STATE(3508), 1,
      sym_attribute,
    STATE(3701), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5003), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5001), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [146445] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3488), 1,
      aux_sym_expression_item_repeat1,
    STATE(3509), 1,
      sym_attribute,
    STATE(3701), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5007), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5005), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [146489] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3490), 1,
      aux_sym_expression_item_repeat1,
    STATE(3510), 1,
      sym_attribute,
    STATE(3701), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5011), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5009), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [146533] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3511), 1,
      sym_attribute,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5015), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5013), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [146577] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3512), 1,
      sym_attribute,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5019), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5017), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [146621] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3513), 1,
      sym_attribute,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5023), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5021), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [146665] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3514), 1,
      sym_attribute,
    STATE(3523), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5027), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5025), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [146709] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5031), 1,
      anon_sym_let,
    ACTIONS(5033), 1,
      anon_sym_EQ,
    ACTIONS(5035), 1,
      anon_sym_COLON_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3515), 1,
      sym_attribute,
    STATE(3659), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5029), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [146757] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5037), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3516), 1,
      sym_attribute,
    STATE(3635), 1,
      sym__attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5704), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [146823] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5039), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3517), 1,
      sym_attribute,
    STATE(3728), 1,
      sym__attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5585), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [146889] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4426), 1,
      anon_sym_STAR,
    STATE(3495), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(3518), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5043), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5041), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [146931] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4133), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4479), 1,
      anon_sym_STAR,
    STATE(3519), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4131), 21,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [146971] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4573), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3504), 1,
      aux_sym_type_binding_repeat1,
    STATE(3520), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4571), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [147015] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4581), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3504), 1,
      aux_sym_type_binding_repeat1,
    STATE(3521), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4579), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [147059] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5045), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3701), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3522), 2,
      sym_attribute,
      aux_sym_expression_item_repeat1,
    ACTIONS(4743), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(4741), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [147101] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3523), 1,
      sym_attribute,
    STATE(3701), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5050), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5048), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [147145] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4968), 1,
      anon_sym_DASH_GT,
    ACTIONS(4970), 1,
      anon_sym_with,
    STATE(3524), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5054), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5052), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147187] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4968), 1,
      anon_sym_DASH_GT,
    ACTIONS(4970), 1,
      anon_sym_with,
    STATE(3525), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5058), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5056), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147229] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4968), 1,
      anon_sym_DASH_GT,
    ACTIONS(4970), 1,
      anon_sym_with,
    STATE(3526), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5062), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5060), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147271] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4968), 1,
      anon_sym_DASH_GT,
    ACTIONS(4970), 1,
      anon_sym_with,
    STATE(3527), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5066), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5064), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147313] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4543), 1,
      anon_sym_as,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    ACTIONS(5068), 1,
      anon_sym_EQ,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3528), 1,
      sym_attribute,
    STATE(3666), 1,
      aux_sym_type_binding_repeat1,
    STATE(3708), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4600), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [147365] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4543), 1,
      anon_sym_as,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3529), 1,
      sym_attribute,
    STATE(3598), 1,
      aux_sym_expression_item_repeat1,
    STATE(3672), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4693), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [147414] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5070), 1,
      anon_sym_LPAREN,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5076), 1,
      anon_sym_end,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    STATE(3530), 1,
      sym_attribute,
    STATE(3848), 1,
      sym__attribute,
    STATE(4034), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [147477] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5088), 1,
      anon_sym_LPAREN,
    ACTIONS(5090), 1,
      anon_sym_end,
    STATE(3531), 1,
      sym_attribute,
    STATE(3808), 1,
      sym__attribute,
    STATE(3914), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [147540] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5094), 1,
      anon_sym_let,
    ACTIONS(5096), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3532), 1,
      sym_attribute,
    STATE(3734), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5092), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147585] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5102), 1,
      anon_sym_PIPE,
    STATE(3533), 1,
      sym_attribute,
    STATE(3536), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5100), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5098), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147626] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5106), 1,
      anon_sym_let,
    ACTIONS(5108), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3534), 1,
      sym_attribute,
    STATE(3613), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5104), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147671] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3535), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5112), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5110), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147708] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5118), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5116), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    STATE(3536), 2,
      sym_attribute,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5114), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147747] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4543), 1,
      anon_sym_as,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3537), 1,
      sym_attribute,
    STATE(3680), 1,
      aux_sym_type_binding_repeat1,
    STATE(3707), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4683), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [147796] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5123), 1,
      anon_sym_let,
    ACTIONS(5125), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3538), 1,
      sym_attribute,
    STATE(3633), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5121), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147841] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5127), 1,
      aux_sym_attribute_id_token1,
    STATE(3539), 1,
      sym_attribute,
    STATE(3542), 1,
      aux_sym_attribute_id_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2741), 8,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2743), 12,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [147882] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3540), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4424), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4346), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [147919] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5129), 1,
      anon_sym_LPAREN,
    ACTIONS(5131), 1,
      anon_sym_end,
    STATE(3541), 1,
      sym_attribute,
    STATE(3798), 1,
      sym__attribute,
    STATE(3872), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [147982] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5127), 1,
      aux_sym_attribute_id_token1,
    STATE(3542), 1,
      sym_attribute,
    STATE(3555), 1,
      aux_sym_attribute_id_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2731), 8,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2733), 12,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [148023] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4968), 1,
      anon_sym_DASH_GT,
    ACTIONS(4970), 1,
      anon_sym_with,
    ACTIONS(5135), 1,
      anon_sym_let,
    STATE(3543), 1,
      sym_attribute,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3823), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5133), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148070] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5139), 1,
      anon_sym_let,
    ACTIONS(5141), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3544), 1,
      sym_attribute,
    STATE(3718), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5137), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148115] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5143), 1,
      anon_sym_LPAREN,
    ACTIONS(5145), 1,
      anon_sym_end,
    STATE(3545), 1,
      sym_attribute,
    STATE(3777), 1,
      sym__attribute,
    STATE(3930), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [148178] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5149), 1,
      anon_sym_let,
    STATE(3546), 1,
      sym_attribute,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3842), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5151), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(5147), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148223] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5155), 1,
      anon_sym_let,
    ACTIONS(5157), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3547), 1,
      sym_attribute,
    STATE(3575), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5153), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148268] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3548), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5161), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5159), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148305] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5102), 1,
      anon_sym_PIPE,
    STATE(3549), 1,
      sym_attribute,
    STATE(3551), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4382), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4380), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148346] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5102), 1,
      anon_sym_PIPE,
    STATE(3533), 1,
      aux_sym_variant_declaration_repeat1,
    STATE(3550), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5165), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5163), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148387] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5102), 1,
      anon_sym_PIPE,
    STATE(3536), 1,
      aux_sym_variant_declaration_repeat1,
    STATE(3551), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5165), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5163), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148428] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5169), 1,
      anon_sym_let,
    ACTIONS(5171), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3552), 1,
      sym_attribute,
    STATE(3624), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5167), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148473] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5175), 1,
      anon_sym_let,
    ACTIONS(5177), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3553), 1,
      sym_attribute,
    STATE(3594), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5173), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148518] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5179), 1,
      anon_sym_LPAREN,
    ACTIONS(5181), 1,
      anon_sym_end,
    STATE(3554), 1,
      sym_attribute,
    STATE(3825), 1,
      sym__attribute,
    STATE(3957), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [148581] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5183), 1,
      aux_sym_attribute_id_token1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3555), 2,
      sym_attribute,
      aux_sym_attribute_id_repeat1,
    ACTIONS(2722), 8,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2724), 12,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [148620] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5188), 1,
      anon_sym_let,
    ACTIONS(5190), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3556), 1,
      sym_attribute,
    STATE(3754), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5186), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148665] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3989), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4475), 1,
      anon_sym_DASH_GT,
    STATE(3557), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3987), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [148704] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5192), 1,
      anon_sym_LPAREN,
    ACTIONS(5194), 1,
      anon_sym_end,
    STATE(3558), 1,
      sym_attribute,
    STATE(3836), 1,
      sym__attribute,
    STATE(3997), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [148767] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5196), 1,
      anon_sym_LPAREN,
    ACTIONS(5198), 1,
      anon_sym_end,
    STATE(3559), 1,
      sym_attribute,
    STATE(3847), 1,
      sym__attribute,
    STATE(4018), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [148830] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5200), 1,
      anon_sym_LPAREN,
    ACTIONS(5202), 1,
      anon_sym_end,
    STATE(3560), 1,
      sym_attribute,
    STATE(3816), 1,
      sym__attribute,
    STATE(4037), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [148893] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4968), 1,
      anon_sym_DASH_GT,
    ACTIONS(4970), 1,
      anon_sym_with,
    ACTIONS(5206), 1,
      anon_sym_let,
    STATE(3561), 1,
      sym_attribute,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3802), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5204), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148940] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5210), 1,
      anon_sym_let,
    STATE(3562), 1,
      sym_attribute,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3837), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5212), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(5208), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [148985] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5214), 1,
      anon_sym_LPAREN,
    ACTIONS(5216), 1,
      anon_sym_end,
    STATE(3563), 1,
      sym_attribute,
    STATE(3817), 1,
      sym__attribute,
    STATE(3863), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [149048] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3564), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5220), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5218), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149085] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5222), 1,
      anon_sym_LPAREN,
    ACTIONS(5224), 1,
      anon_sym_end,
    STATE(3565), 1,
      sym_attribute,
    STATE(3779), 1,
      sym__attribute,
    STATE(3905), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [149148] = 18,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5226), 1,
      anon_sym_LPAREN,
    ACTIONS(5228), 1,
      anon_sym_end,
    STATE(3566), 1,
      sym_attribute,
    STATE(3782), 1,
      sym__attribute,
    STATE(3943), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [149211] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(221), 1,
      anon_sym_DQUOTE,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5230), 1,
      anon_sym_LBRACE,
    STATE(3567), 1,
      sym_attribute,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3811), 1,
      aux_sym_external_repeat1,
    STATE(3824), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3818), 2,
      sym_string,
      sym_quoted_string,
    ACTIONS(4729), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [149260] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5234), 1,
      anon_sym_let,
    ACTIONS(5236), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3568), 1,
      sym_attribute,
    STATE(3649), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5232), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149305] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(221), 1,
      anon_sym_DQUOTE,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5230), 1,
      anon_sym_LBRACE,
    STATE(3569), 1,
      sym_attribute,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3811), 1,
      aux_sym_external_repeat1,
    STATE(3849), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3818), 2,
      sym_string,
      sym_quoted_string,
    ACTIONS(4697), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [149354] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3570), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3545), 2,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3547), 20,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [149391] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5238), 1,
      anon_sym_COLON2,
    STATE(3571), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3050), 6,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3052), 15,
      ts_builtin_sym_end,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT,
      sym__uppercase_identifier,
  [149430] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5242), 1,
      anon_sym_let,
    ACTIONS(5244), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3572), 1,
      sym_attribute,
    STATE(3709), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5240), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149475] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5248), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3573), 1,
      sym_attribute,
    STATE(3639), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5246), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149517] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    ACTIONS(5250), 1,
      anon_sym_DASH_GT,
    ACTIONS(5252), 1,
      anon_sym_functor,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3574), 1,
      sym_attribute,
    STATE(5248), 1,
      sym__module_type,
    STATE(5531), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [149577] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5256), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3575), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5254), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149619] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5260), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3576), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5258), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149661] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5264), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3577), 1,
      sym_attribute,
    STATE(3627), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5262), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149703] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5268), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3578), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5266), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149745] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5272), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3579), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5270), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149787] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5276), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3580), 1,
      sym_attribute,
    STATE(3581), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5274), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149829] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5280), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3581), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5278), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149871] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3582), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5738), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [149931] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5284), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3583), 1,
      sym_attribute,
    STATE(3699), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5282), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [149973] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5288), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3584), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5286), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150015] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3585), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5292), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5290), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150051] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5298), 1,
      anon_sym_and,
    ACTIONS(5301), 1,
      sym_let_and_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5296), 2,
      anon_sym_let,
      anon_sym_in,
    STATE(3586), 2,
      sym_attribute,
      aux_sym_value_definition_repeat1,
    ACTIONS(5294), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150091] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5306), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3587), 1,
      sym_attribute,
    STATE(3729), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5304), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150133] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3588), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5849), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [150193] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3589), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5355), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [150253] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3590), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4895), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4893), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [150289] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3591), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5515), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [150349] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3592), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1817), 3,
      anon_sym_COLON2,
      anon_sym_DOT,
      anon_sym_POUND,
    ACTIONS(1819), 18,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
  [150385] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3593), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4344), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4342), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150421] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5310), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3594), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5308), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150463] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5314), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3595), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5312), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150505] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5318), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3596), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5316), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150547] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5322), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3597), 1,
      sym_attribute,
    STATE(3622), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5320), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150589] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4806), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3598), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4804), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150631] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5326), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3599), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5324), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150673] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4758), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4760), 1,
      anon_sym_EQ,
    ACTIONS(5328), 1,
      anon_sym_of,
    ACTIONS(5330), 1,
      anon_sym_COLON2,
    STATE(3600), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4756), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [150715] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5334), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3601), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5332), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150757] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2053), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3602), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2051), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150799] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3603), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5116), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5114), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150835] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5338), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3604), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5336), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150877] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5344), 1,
      anon_sym_and,
    ACTIONS(5346), 1,
      sym_let_and_operator,
    STATE(3605), 1,
      sym_attribute,
    STATE(3758), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5342), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(5340), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [150919] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3606), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5817), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [150979] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4750), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3607), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4748), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151021] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5350), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3608), 1,
      sym_attribute,
    STATE(3711), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5348), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151063] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5354), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3609), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5352), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151105] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5358), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3610), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5356), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151147] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5362), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3611), 1,
      sym_attribute,
    STATE(3629), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5360), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151189] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5366), 1,
      anon_sym_let,
    STATE(3612), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5364), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151225] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5370), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3613), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5368), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151267] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3614), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5374), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5372), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151303] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5378), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3615), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5376), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151345] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5378), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3616), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5376), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151387] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5380), 1,
      sym__lowercase_identifier,
    ACTIONS(5382), 1,
      anon_sym_LPAREN,
    ACTIONS(5384), 1,
      anon_sym_DASH_GT,
    ACTIONS(5386), 1,
      anon_sym_module,
    ACTIONS(5388), 1,
      anon_sym_sig,
    ACTIONS(5390), 1,
      anon_sym_functor,
    ACTIONS(5392), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3617), 1,
      sym_attribute,
    STATE(4801), 1,
      sym__module_type,
    STATE(5446), 1,
      aux_sym_module_binding_repeat1,
    STATE(6087), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4686), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [151447] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5396), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3618), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5394), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151489] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5380), 1,
      sym__lowercase_identifier,
    ACTIONS(5382), 1,
      anon_sym_LPAREN,
    ACTIONS(5384), 1,
      anon_sym_DASH_GT,
    ACTIONS(5386), 1,
      anon_sym_module,
    ACTIONS(5388), 1,
      anon_sym_sig,
    ACTIONS(5390), 1,
      anon_sym_functor,
    ACTIONS(5392), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3619), 1,
      sym_attribute,
    STATE(4807), 1,
      sym__module_type,
    STATE(5446), 1,
      aux_sym_module_binding_repeat1,
    STATE(6087), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4686), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [151549] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3620), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5657), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [151609] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5400), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3621), 1,
      sym_attribute,
    STATE(3715), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5398), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151651] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5404), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3622), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5402), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151693] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5344), 1,
      anon_sym_and,
    ACTIONS(5346), 1,
      sym_let_and_operator,
    STATE(3623), 1,
      sym_attribute,
    STATE(3694), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5408), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(5406), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151735] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5412), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3624), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5410), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151777] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2065), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3625), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2063), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151819] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5416), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3626), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5414), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151861] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5420), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3627), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5418), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [151903] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3628), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5381), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [151963] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5424), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3629), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5422), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152005] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5428), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3630), 1,
      sym_attribute,
    STATE(3724), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5426), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152047] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5344), 1,
      anon_sym_and,
    ACTIONS(5346), 1,
      sym_let_and_operator,
    STATE(3586), 1,
      aux_sym_value_definition_repeat1,
    STATE(3631), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5342), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(5340), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152089] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5432), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3632), 1,
      sym_attribute,
    STATE(3725), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5430), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152131] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5436), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3633), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5434), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152173] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(3634), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2722), 8,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(2724), 13,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
      aux_sym_attribute_id_token1,
  [152209] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(5438), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3635), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5880), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [152269] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    ACTIONS(5440), 1,
      anon_sym_DASH_GT,
    ACTIONS(5442), 1,
      anon_sym_functor,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3561), 1,
      sym__module_type,
    STATE(3636), 1,
      sym_attribute,
    STATE(5487), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [152329] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5380), 1,
      sym__lowercase_identifier,
    ACTIONS(5382), 1,
      anon_sym_LPAREN,
    ACTIONS(5384), 1,
      anon_sym_DASH_GT,
    ACTIONS(5386), 1,
      anon_sym_module,
    ACTIONS(5388), 1,
      anon_sym_sig,
    ACTIONS(5390), 1,
      anon_sym_functor,
    ACTIONS(5392), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3637), 1,
      sym_attribute,
    STATE(4810), 1,
      sym__module_type,
    STATE(5446), 1,
      aux_sym_module_binding_repeat1,
    STATE(6087), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4686), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [152389] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3638), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5446), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5444), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152425] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5450), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3639), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5448), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152467] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5454), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3640), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5452), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152509] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5458), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3641), 1,
      sym_attribute,
    STATE(3733), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5456), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152551] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5380), 1,
      sym__lowercase_identifier,
    ACTIONS(5382), 1,
      anon_sym_LPAREN,
    ACTIONS(5384), 1,
      anon_sym_DASH_GT,
    ACTIONS(5386), 1,
      anon_sym_module,
    ACTIONS(5388), 1,
      anon_sym_sig,
    ACTIONS(5390), 1,
      anon_sym_functor,
    ACTIONS(5392), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3642), 1,
      sym_attribute,
    STATE(4684), 1,
      sym__module_type,
    STATE(5446), 1,
      aux_sym_module_binding_repeat1,
    STATE(6087), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4686), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [152611] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5380), 1,
      sym__lowercase_identifier,
    ACTIONS(5382), 1,
      anon_sym_LPAREN,
    ACTIONS(5384), 1,
      anon_sym_DASH_GT,
    ACTIONS(5386), 1,
      anon_sym_module,
    ACTIONS(5388), 1,
      anon_sym_sig,
    ACTIONS(5390), 1,
      anon_sym_functor,
    ACTIONS(5392), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3643), 1,
      sym_attribute,
    STATE(4700), 1,
      sym__module_type,
    STATE(5446), 1,
      aux_sym_module_binding_repeat1,
    STATE(6087), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4686), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [152671] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5462), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3644), 1,
      sym_attribute,
    STATE(3696), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5460), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152713] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4814), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3645), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4812), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152755] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5344), 1,
      anon_sym_and,
    ACTIONS(5346), 1,
      sym_let_and_operator,
    STATE(3631), 1,
      aux_sym_value_definition_repeat1,
    STATE(3646), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5466), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(5464), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152797] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5470), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3647), 1,
      sym_attribute,
    STATE(3693), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5468), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152839] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5474), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3640), 1,
      aux_sym_expression_item_repeat1,
    STATE(3648), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5472), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152881] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5478), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3649), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5476), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [152923] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4972), 1,
      anon_sym_DASH_GT,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4976), 1,
      anon_sym_functor,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3650), 1,
      sym_attribute,
    STATE(3833), 1,
      sym__module_type,
    STATE(5277), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [152983] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(5480), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3651), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5614), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [153043] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5484), 1,
      anon_sym_let,
    STATE(3652), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5482), 20,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153079] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4972), 1,
      anon_sym_DASH_GT,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4976), 1,
      anon_sym_functor,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3653), 1,
      sym_attribute,
    STATE(3909), 1,
      sym__module_type,
    STATE(5277), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [153139] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4972), 1,
      anon_sym_DASH_GT,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4976), 1,
      anon_sym_functor,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3654), 1,
      sym_attribute,
    STATE(3792), 1,
      sym__module_type,
    STATE(5277), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [153199] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3655), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5670), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [153259] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4972), 1,
      anon_sym_DASH_GT,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4976), 1,
      anon_sym_functor,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3656), 1,
      sym_attribute,
    STATE(3793), 1,
      sym__module_type,
    STATE(5277), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [153319] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3504), 1,
      aux_sym_type_binding_repeat1,
    STATE(3595), 1,
      aux_sym_expression_item_repeat1,
    STATE(3657), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4812), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [153365] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3658), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5351), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [153425] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5488), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3659), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5486), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153467] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4972), 1,
      anon_sym_DASH_GT,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4976), 1,
      anon_sym_functor,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3660), 1,
      sym_attribute,
    STATE(3799), 1,
      sym__module_type,
    STATE(5277), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [153527] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3661), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5733), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [153587] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3662), 1,
      sym_attribute,
    STATE(3666), 1,
      aux_sym_type_binding_repeat1,
    STATE(3708), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4600), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [153633] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3504), 1,
      aux_sym_type_binding_repeat1,
    STATE(3663), 1,
      sym_attribute,
    STATE(3717), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4881), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [153679] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3598), 1,
      aux_sym_expression_item_repeat1,
    STATE(3664), 1,
      sym_attribute,
    STATE(3672), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4693), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [153725] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3665), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5406), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [153785] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3504), 1,
      aux_sym_type_binding_repeat1,
    STATE(3599), 1,
      aux_sym_expression_item_repeat1,
    STATE(3666), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4774), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [153831] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3607), 1,
      aux_sym_expression_item_repeat1,
    STATE(3667), 1,
      sym_attribute,
    STATE(3675), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4606), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [153877] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5492), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3668), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5490), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [153919] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3669), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5792), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [153979] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3670), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4887), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4885), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [154015] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4754), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3671), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4752), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154057] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3504), 1,
      aux_sym_type_binding_repeat1,
    STATE(3672), 1,
      sym_attribute,
    STATE(3702), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4804), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [154103] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3673), 1,
      sym_attribute,
    STATE(3679), 1,
      aux_sym_type_binding_repeat1,
    STATE(3705), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4792), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [154149] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3674), 1,
      sym_attribute,
    STATE(3680), 1,
      aux_sym_type_binding_repeat1,
    STATE(3707), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4683), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [154195] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3504), 1,
      aux_sym_type_binding_repeat1,
    STATE(3675), 1,
      sym_attribute,
    STATE(3710), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4748), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [154241] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3676), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5841), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [154301] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5496), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3677), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5494), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154343] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3678), 1,
      sym_attribute,
    STATE(3682), 1,
      aux_sym_type_binding_repeat1,
    STATE(3747), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4820), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [154389] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3504), 1,
      aux_sym_type_binding_repeat1,
    STATE(3679), 1,
      sym_attribute,
    STATE(3752), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4857), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [154435] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3504), 1,
      aux_sym_type_binding_repeat1,
    STATE(3680), 1,
      sym_attribute,
    STATE(3753), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4770), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [154481] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3681), 1,
      sym_attribute,
    STATE(3684), 1,
      aux_sym_type_binding_repeat1,
    STATE(3755), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4835), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [154527] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3504), 1,
      aux_sym_type_binding_repeat1,
    STATE(3668), 1,
      aux_sym_expression_item_repeat1,
    STATE(3682), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4808), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [154573] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3671), 1,
      aux_sym_expression_item_repeat1,
    STATE(3683), 1,
      sym_attribute,
    STATE(3685), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4828), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [154619] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3504), 1,
      aux_sym_type_binding_repeat1,
    STATE(3684), 1,
      sym_attribute,
    STATE(3689), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4843), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [154665] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4782), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3504), 1,
      aux_sym_type_binding_repeat1,
    STATE(3685), 1,
      sym_attribute,
    STATE(3737), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4752), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [154711] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5380), 1,
      sym__lowercase_identifier,
    ACTIONS(5382), 1,
      anon_sym_LPAREN,
    ACTIONS(5388), 1,
      anon_sym_sig,
    ACTIONS(5392), 1,
      sym__uppercase_identifier,
    ACTIONS(5498), 1,
      anon_sym_DASH_GT,
    ACTIONS(5500), 1,
      anon_sym_module,
    ACTIONS(5502), 1,
      anon_sym_functor,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3686), 1,
      sym_attribute,
    STATE(4923), 1,
      sym__module_type,
    STATE(5327), 1,
      aux_sym_module_binding_repeat1,
    STATE(6087), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4686), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [154771] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5380), 1,
      sym__lowercase_identifier,
    ACTIONS(5382), 1,
      anon_sym_LPAREN,
    ACTIONS(5388), 1,
      anon_sym_sig,
    ACTIONS(5392), 1,
      sym__uppercase_identifier,
    ACTIONS(5498), 1,
      anon_sym_DASH_GT,
    ACTIONS(5500), 1,
      anon_sym_module,
    ACTIONS(5502), 1,
      anon_sym_functor,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3687), 1,
      sym_attribute,
    STATE(5327), 1,
      aux_sym_module_binding_repeat1,
    STATE(5369), 1,
      sym__module_type,
    STATE(6087), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4686), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [154831] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5380), 1,
      sym__lowercase_identifier,
    ACTIONS(5382), 1,
      anon_sym_LPAREN,
    ACTIONS(5388), 1,
      anon_sym_sig,
    ACTIONS(5392), 1,
      sym__uppercase_identifier,
    ACTIONS(5498), 1,
      anon_sym_DASH_GT,
    ACTIONS(5500), 1,
      anon_sym_module,
    ACTIONS(5502), 1,
      anon_sym_functor,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3688), 1,
      sym_attribute,
    STATE(5327), 1,
      aux_sym_module_binding_repeat1,
    STATE(5382), 1,
      sym__module_type,
    STATE(6087), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4686), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [154891] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5506), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3689), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5504), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154933] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5510), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3690), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5508), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [154975] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5514), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3691), 1,
      sym_attribute,
    STATE(3739), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5512), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155017] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5380), 1,
      sym__lowercase_identifier,
    ACTIONS(5382), 1,
      anon_sym_LPAREN,
    ACTIONS(5388), 1,
      anon_sym_sig,
    ACTIONS(5392), 1,
      sym__uppercase_identifier,
    ACTIONS(5498), 1,
      anon_sym_DASH_GT,
    ACTIONS(5500), 1,
      anon_sym_module,
    ACTIONS(5502), 1,
      anon_sym_functor,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3692), 1,
      sym_attribute,
    STATE(5327), 1,
      aux_sym_module_binding_repeat1,
    STATE(5414), 1,
      sym__module_type,
    STATE(6087), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4686), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [155077] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5518), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3693), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5516), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155119] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5344), 1,
      anon_sym_and,
    ACTIONS(5346), 1,
      sym_let_and_operator,
    STATE(3586), 1,
      aux_sym_value_definition_repeat1,
    STATE(3694), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5522), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(5520), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155161] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    ACTIONS(5250), 1,
      anon_sym_DASH_GT,
    ACTIONS(5252), 1,
      anon_sym_functor,
    STATE(3357), 1,
      sym__module_type,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3695), 1,
      sym_attribute,
    STATE(5531), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [155221] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5526), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3696), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5524), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155263] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5530), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3697), 1,
      sym_attribute,
    STATE(3744), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5528), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155305] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    ACTIONS(5440), 1,
      anon_sym_DASH_GT,
    ACTIONS(5442), 1,
      anon_sym_functor,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3500), 1,
      sym__module_type,
    STATE(3698), 1,
      sym_attribute,
    STATE(5487), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [155365] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5534), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3699), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5532), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155407] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4457), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3700), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4455), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155449] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3701), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4933), 4,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4931), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [155485] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5538), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3702), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5536), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155527] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5542), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3703), 1,
      sym_attribute,
    STATE(3745), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5540), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155569] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5546), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3704), 1,
      sym_attribute,
    STATE(3746), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5544), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155611] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4859), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3705), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4857), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155653] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3706), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1833), 3,
      anon_sym_COLON2,
      anon_sym_DOT,
      anon_sym_POUND,
    ACTIONS(1835), 18,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
  [155689] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4772), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3707), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4770), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155731] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4776), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3708), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4774), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155773] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5550), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3709), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5548), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155815] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5554), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3710), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5552), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155857] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5558), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3711), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5556), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155899] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5562), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3712), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5560), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155941] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5562), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3713), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5560), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [155983] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5380), 1,
      sym__lowercase_identifier,
    ACTIONS(5382), 1,
      anon_sym_LPAREN,
    ACTIONS(5388), 1,
      anon_sym_sig,
    ACTIONS(5392), 1,
      sym__uppercase_identifier,
    ACTIONS(5498), 1,
      anon_sym_DASH_GT,
    ACTIONS(5500), 1,
      anon_sym_module,
    ACTIONS(5502), 1,
      anon_sym_functor,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3714), 1,
      sym_attribute,
    STATE(5327), 1,
      aux_sym_module_binding_repeat1,
    STATE(5333), 1,
      sym__module_type,
    STATE(6087), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4686), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [156043] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5566), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3715), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5564), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156085] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5570), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3716), 1,
      sym_attribute,
    STATE(3766), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5568), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156127] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5574), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3717), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5572), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156169] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5578), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3718), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5576), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156211] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5582), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3719), 1,
      sym_attribute,
    STATE(3749), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5580), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156253] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5586), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3609), 1,
      aux_sym_expression_item_repeat1,
    STATE(3720), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5584), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156295] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5590), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3721), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5588), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156337] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3722), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1733), 3,
      anon_sym_COLON2,
      anon_sym_DOT,
      anon_sym_POUND,
    ACTIONS(1735), 18,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      anon_sym_LBRACK_AT_AT,
      sym_hash_operator,
  [156373] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2035), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3723), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2033), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156415] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5594), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3724), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5592), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156457] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5598), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3725), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5596), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156499] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5602), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3604), 1,
      aux_sym_expression_item_repeat1,
    STATE(3726), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5600), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156541] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5606), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3610), 1,
      aux_sym_expression_item_repeat1,
    STATE(3727), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5604), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156583] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(5608), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3728), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5601), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [156643] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5612), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3729), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5610), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156685] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3730), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5629), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [156745] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4946), 1,
      anon_sym_DASH_GT,
    ACTIONS(4948), 1,
      anon_sym_module,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4952), 1,
      anon_sym_functor,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3731), 1,
      sym_attribute,
    STATE(5328), 1,
      aux_sym_module_binding_repeat1,
    STATE(5649), 1,
      sym__module_type,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [156805] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5380), 1,
      sym__lowercase_identifier,
    ACTIONS(5382), 1,
      anon_sym_LPAREN,
    ACTIONS(5388), 1,
      anon_sym_sig,
    ACTIONS(5392), 1,
      sym__uppercase_identifier,
    ACTIONS(5498), 1,
      anon_sym_DASH_GT,
    ACTIONS(5500), 1,
      anon_sym_module,
    ACTIONS(5502), 1,
      anon_sym_functor,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3732), 1,
      sym_attribute,
    STATE(4964), 1,
      sym__module_type,
    STATE(5327), 1,
      aux_sym_module_binding_repeat1,
    STATE(6087), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4686), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [156865] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5616), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3733), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5614), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156907] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5620), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3734), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5618), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156949] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2029), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3735), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2027), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [156991] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5624), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3736), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5622), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157033] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5628), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3737), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5626), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157075] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4972), 1,
      anon_sym_DASH_GT,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4976), 1,
      anon_sym_functor,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3357), 1,
      sym__module_type,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3738), 1,
      sym_attribute,
    STATE(5277), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [157135] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5632), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3739), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5630), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157177] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5380), 1,
      sym__lowercase_identifier,
    ACTIONS(5382), 1,
      anon_sym_LPAREN,
    ACTIONS(5384), 1,
      anon_sym_DASH_GT,
    ACTIONS(5386), 1,
      anon_sym_module,
    ACTIONS(5388), 1,
      anon_sym_sig,
    ACTIONS(5390), 1,
      anon_sym_functor,
    ACTIONS(5392), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3740), 1,
      sym_attribute,
    STATE(4682), 1,
      sym__module_type,
    STATE(5446), 1,
      aux_sym_module_binding_repeat1,
    STATE(6087), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4686), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [157237] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5636), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3576), 1,
      aux_sym_expression_item_repeat1,
    STATE(3741), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5634), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157279] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    ACTIONS(5440), 1,
      anon_sym_DASH_GT,
    ACTIONS(5442), 1,
      anon_sym_functor,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3543), 1,
      sym__module_type,
    STATE(3742), 1,
      sym_attribute,
    STATE(5487), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [157339] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5640), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3743), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5638), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157381] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5644), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3744), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5642), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157423] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5648), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3745), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5646), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157465] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5652), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3746), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5650), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157507] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4810), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3747), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4808), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157549] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5656), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3748), 1,
      sym_attribute,
    STATE(3776), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5654), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157591] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5660), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3749), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5658), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157633] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5664), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3578), 1,
      aux_sym_expression_item_repeat1,
    STATE(3750), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5662), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157675] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5668), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3579), 1,
      aux_sym_expression_item_repeat1,
    STATE(3751), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5666), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157717] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5672), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3752), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5670), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157759] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5676), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3753), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5674), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157801] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5680), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3754), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5678), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157843] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4845), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3755), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4843), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157885] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4883), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3756), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4881), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157927] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3757), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5684), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5682), 19,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [157963] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5344), 1,
      anon_sym_and,
    ACTIONS(5346), 1,
      sym_let_and_operator,
    STATE(3586), 1,
      aux_sym_value_definition_repeat1,
    STATE(3758), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5408), 2,
      anon_sym_let,
      anon_sym_in,
    ACTIONS(5406), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [158005] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5688), 1,
      anon_sym_let,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3690), 1,
      aux_sym_expression_item_repeat1,
    STATE(3759), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5686), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [158047] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4972), 1,
      anon_sym_DASH_GT,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4976), 1,
      anon_sym_functor,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3760), 1,
      sym_attribute,
    STATE(3835), 1,
      sym__module_type,
    STATE(5277), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158107] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4972), 1,
      anon_sym_DASH_GT,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4976), 1,
      anon_sym_functor,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3761), 1,
      sym_attribute,
    STATE(3994), 1,
      sym__module_type,
    STATE(5277), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158167] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4849), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5690), 1,
      anon_sym_and,
    STATE(3762), 1,
      sym_attribute,
    STATE(3764), 1,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4847), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [158207] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4972), 1,
      anon_sym_DASH_GT,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4976), 1,
      anon_sym_functor,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3763), 1,
      sym_attribute,
    STATE(3995), 1,
      sym__module_type,
    STATE(5277), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158267] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4723), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5692), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3764), 2,
      sym_attribute,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(4721), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [158305] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(2059), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3765), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2057), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [158347] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5697), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3766), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5695), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [158389] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    ACTIONS(5440), 1,
      anon_sym_DASH_GT,
    ACTIONS(5442), 1,
      anon_sym_functor,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3527), 1,
      sym__module_type,
    STATE(3767), 1,
      sym_attribute,
    STATE(5487), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158449] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    ACTIONS(5440), 1,
      anon_sym_DASH_GT,
    ACTIONS(5442), 1,
      anon_sym_functor,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3524), 1,
      sym__module_type,
    STATE(3768), 1,
      sym_attribute,
    STATE(5487), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158509] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    ACTIONS(5440), 1,
      anon_sym_DASH_GT,
    ACTIONS(5442), 1,
      anon_sym_functor,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3525), 1,
      sym__module_type,
    STATE(3769), 1,
      sym_attribute,
    STATE(5487), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158569] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    ACTIONS(5440), 1,
      anon_sym_DASH_GT,
    ACTIONS(5442), 1,
      anon_sym_functor,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3526), 1,
      sym__module_type,
    STATE(3770), 1,
      sym_attribute,
    STATE(5487), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158629] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4863), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5690), 1,
      anon_sym_and,
    STATE(3762), 1,
      aux_sym_module_type_constraint_repeat1,
    STATE(3771), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4861), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_with,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [158669] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    ACTIONS(5440), 1,
      anon_sym_DASH_GT,
    ACTIONS(5442), 1,
      anon_sym_functor,
    STATE(3357), 1,
      sym__module_type,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3772), 1,
      sym_attribute,
    STATE(5487), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158729] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    ACTIONS(5250), 1,
      anon_sym_DASH_GT,
    ACTIONS(5252), 1,
      anon_sym_functor,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3773), 1,
      sym_attribute,
    STATE(5249), 1,
      sym__module_type,
    STATE(5531), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158789] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    ACTIONS(5250), 1,
      anon_sym_DASH_GT,
    ACTIONS(5252), 1,
      anon_sym_functor,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3774), 1,
      sym_attribute,
    STATE(5246), 1,
      sym__module_type,
    STATE(5531), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158849] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4938), 1,
      sym__lowercase_identifier,
    ACTIONS(4940), 1,
      anon_sym_LPAREN,
    ACTIONS(4950), 1,
      anon_sym_sig,
    ACTIONS(4974), 1,
      anon_sym_module,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    ACTIONS(5250), 1,
      anon_sym_DASH_GT,
    ACTIONS(5252), 1,
      anon_sym_functor,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(3775), 1,
      sym_attribute,
    STATE(5247), 1,
      sym__module_type,
    STATE(5531), 1,
      aux_sym_module_binding_repeat1,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3352), 8,
      sym_signature,
      sym_module_type_constraint,
      sym_module_type_of,
      sym_functor_type,
      sym_parenthesized_module_type,
      sym_extension,
      sym_quoted_extension,
      sym_module_type_path,
  [158909] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5701), 1,
      anon_sym_let,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3776), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5699), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [158951] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5703), 1,
      anon_sym_LPAREN,
    ACTIONS(5705), 1,
      anon_sym_end,
    STATE(3777), 1,
      sym_attribute,
    STATE(3933), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [159008] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3778), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5296), 3,
      anon_sym_let,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(5294), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      sym_let_and_operator,
      aux_sym_directive_token1,
  [159043] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5707), 1,
      anon_sym_LPAREN,
    ACTIONS(5709), 1,
      anon_sym_end,
    STATE(3779), 1,
      sym_attribute,
    STATE(3913), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [159100] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5713), 1,
      anon_sym_let,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3780), 1,
      sym_attribute,
    STATE(3812), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5711), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [159141] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5717), 1,
      anon_sym_let,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3781), 1,
      sym_attribute,
    STATE(3785), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5715), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [159182] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5719), 1,
      anon_sym_LPAREN,
    ACTIONS(5721), 1,
      anon_sym_end,
    STATE(3782), 1,
      sym_attribute,
    STATE(3946), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [159239] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5725), 1,
      anon_sym_let,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3783), 1,
      sym_attribute,
    STATE(3822), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5723), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [159280] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(3784), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3945), 8,
      anon_sym_LBRACK,
      anon_sym_true,
      anon_sym_false,
      anon_sym_LBRACE,
      anon_sym_object,
      anon_sym_new,
      anon_sym_begin,
      sym__lowercase_identifier,
    ACTIONS(3947), 12,
      anon_sym_LPAREN,
      anon_sym_LBRACK_PIPE,
      anon_sym_LBRACE_LT,
      sym_ocamlyacc_value,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
      aux_sym_number_token1,
      anon_sym_SQUOTE,
      anon_sym_DQUOTE,
      sym_prefix_operator,
      sym__uppercase_identifier,
      aux_sym_tag_token1,
  [159315] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5729), 1,
      anon_sym_let,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3785), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5727), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [159356] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3786), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3889), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3887), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [159391] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5733), 1,
      anon_sym_let,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3787), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5731), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [159432] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5737), 1,
      anon_sym_BANG,
    ACTIONS(5739), 1,
      anon_sym_functor,
    ACTIONS(5741), 1,
      anon_sym_struct,
    STATE(3079), 1,
      sym_module_path,
    STATE(3158), 1,
      sym__module_expression,
    STATE(3788), 1,
      sym_attribute,
    STATE(4068), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [159489] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5729), 1,
      anon_sym_let,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3789), 1,
      sym_attribute,
    STATE(3804), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5727), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [159530] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5745), 1,
      anon_sym_let,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3790), 1,
      sym_attribute,
    STATE(3806), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5743), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [159571] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5342), 1,
      anon_sym_let,
    ACTIONS(5747), 1,
      anon_sym_and,
    ACTIONS(5749), 1,
      sym_let_and_operator,
    STATE(3791), 1,
      sym_attribute,
    STATE(3795), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5340), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [159612] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5054), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5751), 1,
      anon_sym_DASH_GT,
    ACTIONS(5753), 1,
      anon_sym_with,
    STATE(3792), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5052), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [159651] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5058), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5751), 1,
      anon_sym_DASH_GT,
    ACTIONS(5753), 1,
      anon_sym_with,
    STATE(3793), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5056), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [159690] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5408), 1,
      anon_sym_let,
    ACTIONS(5747), 1,
      anon_sym_and,
    ACTIONS(5749), 1,
      sym_let_and_operator,
    STATE(3794), 1,
      sym_attribute,
    STATE(3795), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5406), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [159731] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5296), 1,
      anon_sym_let,
    ACTIONS(5755), 1,
      anon_sym_and,
    ACTIONS(5758), 1,
      sym_let_and_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3795), 2,
      sym_attribute,
      aux_sym_value_definition_repeat1,
    ACTIONS(5294), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [159770] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5408), 1,
      anon_sym_let,
    ACTIONS(5747), 1,
      anon_sym_and,
    ACTIONS(5749), 1,
      sym_let_and_operator,
    STATE(3796), 1,
      sym_attribute,
    STATE(3800), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5406), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [159811] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5763), 1,
      anon_sym_let,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3797), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5761), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [159852] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5765), 1,
      anon_sym_LPAREN,
    ACTIONS(5767), 1,
      anon_sym_end,
    STATE(3798), 1,
      sym_attribute,
    STATE(3971), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [159909] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5062), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5751), 1,
      anon_sym_DASH_GT,
    ACTIONS(5753), 1,
      anon_sym_with,
    STATE(3799), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5060), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [159948] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5522), 1,
      anon_sym_let,
    ACTIONS(5747), 1,
      anon_sym_and,
    ACTIONS(5749), 1,
      sym_let_and_operator,
    STATE(3795), 1,
      aux_sym_value_definition_repeat1,
    STATE(3800), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5520), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [159989] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5771), 1,
      anon_sym_let,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3801), 1,
      sym_attribute,
    STATE(3813), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5769), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [160030] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5775), 1,
      anon_sym_let,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3802), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5773), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [160071] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3622), 1,
      sym__lowercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    ACTIONS(5777), 1,
      anon_sym_DASH_GT,
    ACTIONS(5779), 1,
      anon_sym_STAR,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(3803), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3989), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_as,
    ACTIONS(3987), 9,
      ts_builtin_sym_end,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
  [160120] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5783), 1,
      anon_sym_let,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3804), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5781), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [160161] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5787), 1,
      anon_sym_let,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3805), 1,
      sym_attribute,
    STATE(3809), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5785), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [160202] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5787), 1,
      anon_sym_let,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3806), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5785), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [160243] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4739), 1,
      anon_sym_STAR,
    ACTIONS(5043), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3807), 1,
      sym_attribute,
    STATE(3821), 1,
      aux_sym__constructor_argument_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5041), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [160282] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5789), 1,
      anon_sym_LPAREN,
    ACTIONS(5791), 1,
      anon_sym_end,
    STATE(3808), 1,
      sym_attribute,
    STATE(3992), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [160339] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5795), 1,
      anon_sym_let,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3809), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5793), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [160380] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5799), 1,
      anon_sym_let,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3810), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5797), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [160421] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4923), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5801), 1,
      anon_sym_LBRACE,
    ACTIONS(5804), 1,
      anon_sym_DQUOTE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3811), 2,
      sym_attribute,
      aux_sym_external_repeat1,
    STATE(3818), 2,
      sym_string,
      sym_quoted_string,
    ACTIONS(4921), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [160462] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5809), 1,
      anon_sym_let,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3812), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5807), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [160503] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5813), 1,
      anon_sym_let,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3813), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5811), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [160544] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5813), 1,
      anon_sym_let,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3787), 1,
      aux_sym_expression_item_repeat1,
    STATE(3814), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5811), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [160585] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5817), 1,
      anon_sym_let,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3815), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5815), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [160626] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5819), 1,
      anon_sym_LPAREN,
    ACTIONS(5821), 1,
      anon_sym_end,
    STATE(3816), 1,
      sym_attribute,
    STATE(3895), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [160683] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5823), 1,
      anon_sym_LPAREN,
    ACTIONS(5825), 1,
      anon_sym_end,
    STATE(3817), 1,
      sym_attribute,
    STATE(3867), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [160740] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3818), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5829), 3,
      anon_sym_let,
      anon_sym_LBRACE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5827), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_DQUOTE,
      sym_let_operator,
      aux_sym_directive_token1,
  [160775] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5833), 1,
      anon_sym_let,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3819), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5831), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [160816] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3820), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5837), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5835), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [160851] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4565), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5839), 1,
      anon_sym_STAR,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3821), 2,
      sym_attribute,
      aux_sym__constructor_argument_repeat1,
    ACTIONS(4563), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [160888] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5844), 1,
      anon_sym_let,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3822), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5842), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [160929] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5848), 1,
      anon_sym_let,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3823), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5846), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [160970] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5852), 1,
      anon_sym_let,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3824), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5850), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161011] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5854), 1,
      anon_sym_LPAREN,
    ACTIONS(5856), 1,
      anon_sym_end,
    STATE(3825), 1,
      sym_attribute,
    STATE(3959), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [161068] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5860), 1,
      anon_sym_let,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3826), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5858), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161109] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5864), 1,
      anon_sym_let,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3827), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5862), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161150] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3828), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1733), 3,
      anon_sym_let,
      anon_sym_LBRACE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1735), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_DQUOTE,
      sym_let_operator,
      aux_sym_directive_token1,
  [161185] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3829), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1817), 3,
      anon_sym_let,
      anon_sym_LBRACE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1819), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_DQUOTE,
      sym_let_operator,
      aux_sym_directive_token1,
  [161220] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3830), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1833), 3,
      anon_sym_let,
      anon_sym_LBRACE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1835), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      anon_sym_DQUOTE,
      sym_let_operator,
      aux_sym_directive_token1,
  [161255] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5868), 1,
      anon_sym_let,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3797), 1,
      aux_sym_expression_item_repeat1,
    STATE(3831), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5866), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161296] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5466), 1,
      anon_sym_let,
    ACTIONS(5747), 1,
      anon_sym_and,
    ACTIONS(5749), 1,
      sym_let_and_operator,
    STATE(3791), 1,
      aux_sym_value_definition_repeat1,
    STATE(3832), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5464), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161337] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5066), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5751), 1,
      anon_sym_DASH_GT,
    ACTIONS(5753), 1,
      anon_sym_with,
    STATE(3833), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5064), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [161376] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5342), 1,
      anon_sym_let,
    ACTIONS(5747), 1,
      anon_sym_and,
    ACTIONS(5749), 1,
      sym_let_and_operator,
    STATE(3794), 1,
      aux_sym_value_definition_repeat1,
    STATE(3834), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5340), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161417] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4966), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5751), 1,
      anon_sym_DASH_GT,
    ACTIONS(5753), 1,
      anon_sym_with,
    STATE(3835), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4964), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_COLON_EQ,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [161456] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5870), 1,
      anon_sym_LPAREN,
    ACTIONS(5872), 1,
      anon_sym_end,
    STATE(3836), 1,
      sym_attribute,
    STATE(4000), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [161513] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5876), 1,
      anon_sym_let,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3837), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5874), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161554] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3838), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1733), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1735), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161589] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3839), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1817), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1819), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161624] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3840), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1833), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(1835), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161659] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(3841), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3893), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3891), 18,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161694] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5880), 1,
      anon_sym_let,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3842), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5878), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [161735] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3622), 1,
      sym__lowercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    ACTIONS(5882), 1,
      anon_sym_DASH_GT,
    ACTIONS(5884), 1,
      anon_sym_STAR,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(3843), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3987), 5,
      anon_sym_EQ,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(3989), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
  [161784] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3622), 1,
      sym__lowercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    ACTIONS(5886), 1,
      anon_sym_STAR,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(3844), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4131), 6,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_COLON_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(4133), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
  [161831] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5888), 1,
      anon_sym_LPAREN,
    ACTIONS(5890), 1,
      anon_sym_BANG,
    ACTIONS(5892), 1,
      anon_sym_functor,
    ACTIONS(5894), 1,
      anon_sym_struct,
    ACTIONS(5896), 1,
      sym__uppercase_identifier,
    STATE(3845), 1,
      sym_attribute,
    STATE(4074), 1,
      sym__attribute,
    STATE(4256), 1,
      sym__module_expression,
    STATE(4479), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4495), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [161888] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3622), 1,
      sym__lowercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    ACTIONS(5898), 1,
      anon_sym_STAR,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(3846), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4133), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_as,
    ACTIONS(4131), 10,
      ts_builtin_sym_end,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_AMP,
  [161935] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5900), 1,
      anon_sym_LPAREN,
    ACTIONS(5902), 1,
      anon_sym_end,
    STATE(3847), 1,
      sym_attribute,
    STATE(4020), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [161992] = 16,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5904), 1,
      anon_sym_LPAREN,
    ACTIONS(5906), 1,
      anon_sym_end,
    STATE(3848), 1,
      sym_attribute,
    STATE(4011), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [162049] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5910), 1,
      anon_sym_let,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3849), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5908), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [162090] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3842), 1,
      aux_sym_expression_item_repeat1,
    STATE(3850), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5912), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(5147), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [162130] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(3851), 1,
      sym_attribute,
    STATE(4061), 1,
      sym__attribute,
    STATE(4383), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [162184] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5916), 1,
      anon_sym_end,
    STATE(3852), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [162238] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5918), 1,
      anon_sym_end,
    STATE(3853), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [162292] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5920), 1,
      anon_sym_end,
    STATE(3854), 1,
      sym_attribute,
    STATE(3855), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [162346] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5922), 1,
      anon_sym_end,
    STATE(3855), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [162400] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(3856), 1,
      sym_attribute,
    STATE(4090), 1,
      sym__attribute,
    STATE(4358), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [162454] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5926), 1,
      anon_sym_let,
    ACTIONS(5928), 1,
      anon_sym_and,
    STATE(3857), 1,
      sym_attribute,
    STATE(4006), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5924), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [162492] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5932), 1,
      anon_sym_let,
    ACTIONS(5934), 1,
      anon_sym_and,
    STATE(3858), 1,
      sym_attribute,
    STATE(3977), 1,
      aux_sym_class_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5930), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [162530] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5936), 1,
      anon_sym_end,
    STATE(3859), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [162584] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5888), 1,
      anon_sym_LPAREN,
    ACTIONS(5892), 1,
      anon_sym_functor,
    ACTIONS(5894), 1,
      anon_sym_struct,
    ACTIONS(5896), 1,
      sym__uppercase_identifier,
    STATE(3860), 1,
      sym_attribute,
    STATE(4066), 1,
      sym__attribute,
    STATE(4262), 1,
      sym__module_expression,
    STATE(4479), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4495), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [162638] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5938), 1,
      anon_sym_end,
    STATE(3861), 1,
      sym_attribute,
    STATE(3873), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [162692] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5940), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3754), 1,
      aux_sym_expression_item_repeat1,
    STATE(3862), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5186), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [162732] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5825), 1,
      anon_sym_end,
    STATE(3863), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [162786] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5942), 1,
      anon_sym_end,
    STATE(3864), 1,
      sym_attribute,
    STATE(3878), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [162840] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5944), 1,
      anon_sym_end,
    STATE(3865), 1,
      sym_attribute,
    STATE(3886), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [162894] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2632), 1,
      sym__lowercase_identifier,
    ACTIONS(2636), 1,
      anon_sym_LBRACK,
    ACTIONS(2644), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2648), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2650), 1,
      anon_sym_begin,
    ACTIONS(5946), 1,
      anon_sym_LPAREN,
    ACTIONS(5948), 1,
      anon_sym_LBRACE,
    ACTIONS(5950), 1,
      sym__uppercase_identifier,
    STATE(2658), 1,
      sym__value_name,
    STATE(2675), 1,
      sym_parenthesized_operator,
    STATE(2698), 1,
      sym__constructor_name,
    STATE(2745), 1,
      sym__delimited_expression,
    STATE(3866), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2719), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [162950] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5952), 1,
      anon_sym_end,
    STATE(3867), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163004] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(45), 1,
      anon_sym_LBRACK,
    ACTIONS(55), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(75), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(77), 1,
      anon_sym_begin,
    ACTIONS(821), 1,
      sym__lowercase_identifier,
    ACTIONS(5954), 1,
      anon_sym_LPAREN,
    ACTIONS(5956), 1,
      anon_sym_LBRACE,
    ACTIONS(5958), 1,
      sym__uppercase_identifier,
    STATE(872), 1,
      sym__value_name,
    STATE(883), 1,
      sym_parenthesized_operator,
    STATE(891), 1,
      sym__constructor_name,
    STATE(924), 1,
      sym__delimited_expression,
    STATE(3868), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(911), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [163060] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5960), 1,
      anon_sym_end,
    STATE(3869), 1,
      sym_attribute,
    STATE(3871), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163114] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4382), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5962), 1,
      anon_sym_PIPE,
    STATE(3870), 1,
      sym_attribute,
    STATE(3892), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4380), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [163152] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5964), 1,
      anon_sym_end,
    STATE(3871), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163206] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5767), 1,
      anon_sym_end,
    STATE(3872), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163260] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5966), 1,
      anon_sym_end,
    STATE(3873), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163314] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5968), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3613), 1,
      aux_sym_expression_item_repeat1,
    STATE(3874), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5104), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [163354] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5970), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3649), 1,
      aux_sym_expression_item_repeat1,
    STATE(3875), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5232), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [163394] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5972), 1,
      anon_sym_end,
    STATE(3876), 1,
      sym_attribute,
    STATE(3887), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163448] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5974), 1,
      anon_sym_end,
    STATE(3877), 1,
      sym_attribute,
    STATE(3888), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163502] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5976), 1,
      anon_sym_end,
    STATE(3878), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163556] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5978), 1,
      anon_sym_end,
    STATE(3879), 1,
      sym_attribute,
    STATE(3883), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163610] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5980), 1,
      anon_sym_end,
    STATE(3880), 1,
      sym_attribute,
    STATE(3921), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163664] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5165), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5962), 1,
      anon_sym_PIPE,
    STATE(3881), 1,
      sym_attribute,
    STATE(3924), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5163), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [163702] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5982), 1,
      anon_sym_end,
    STATE(3882), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163756] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5984), 1,
      anon_sym_end,
    STATE(3883), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163810] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5986), 1,
      anon_sym_end,
    STATE(3884), 1,
      sym_attribute,
    STATE(3941), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163864] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5988), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3633), 1,
      aux_sym_expression_item_repeat1,
    STATE(3885), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5121), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [163904] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5990), 1,
      anon_sym_end,
    STATE(3886), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [163958] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5992), 1,
      anon_sym_end,
    STATE(3887), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [164012] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5994), 1,
      anon_sym_end,
    STATE(3888), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [164066] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5996), 1,
      anon_sym_end,
    STATE(3889), 1,
      sym_attribute,
    STATE(3891), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [164120] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6000), 1,
      anon_sym_let,
    ACTIONS(6002), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3890), 2,
      sym_attribute,
      aux_sym_type_definition_repeat1,
    ACTIONS(5998), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [164156] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6005), 1,
      anon_sym_end,
    STATE(3891), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [164210] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5165), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5962), 1,
      anon_sym_PIPE,
    STATE(3892), 1,
      sym_attribute,
    STATE(3926), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5163), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [164248] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6007), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3718), 1,
      aux_sym_expression_item_repeat1,
    STATE(3893), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5137), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [164288] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6009), 1,
      anon_sym_end,
    STATE(3894), 1,
      sym_attribute,
    STATE(4001), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [164342] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6011), 1,
      anon_sym_end,
    STATE(3895), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [164396] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4543), 1,
      anon_sym_as,
    STATE(3896), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6015), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6013), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [164432] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4543), 1,
      anon_sym_as,
    STATE(3897), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6019), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6017), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [164468] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6021), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3624), 1,
      aux_sym_expression_item_repeat1,
    STATE(3898), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5167), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [164508] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6023), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3594), 1,
      aux_sym_expression_item_repeat1,
    STATE(3899), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5173), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [164548] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6027), 1,
      anon_sym_let,
    ACTIONS(6029), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3900), 2,
      sym_attribute,
      aux_sym_module_definition_repeat1,
    ACTIONS(6025), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [164584] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6034), 1,
      anon_sym_let,
    ACTIONS(6036), 1,
      anon_sym_and,
    STATE(3901), 1,
      sym_attribute,
    STATE(3987), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6032), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [164622] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4543), 1,
      anon_sym_as,
    STATE(3902), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6040), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6038), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [164658] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6042), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3709), 1,
      aux_sym_expression_item_repeat1,
    STATE(3903), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5240), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [164698] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6044), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3734), 1,
      aux_sym_expression_item_repeat1,
    STATE(3904), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5092), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [164738] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5709), 1,
      anon_sym_end,
    STATE(3905), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [164792] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6046), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(3575), 1,
      aux_sym_expression_item_repeat1,
    STATE(3906), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5153), 15,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [164832] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6034), 1,
      anon_sym_let,
    ACTIONS(6036), 1,
      anon_sym_and,
    STATE(3900), 1,
      aux_sym_module_definition_repeat1,
    STATE(3907), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6032), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [164870] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6036), 1,
      anon_sym_and,
    ACTIONS(6050), 1,
      anon_sym_let,
    STATE(3908), 1,
      sym_attribute,
    STATE(3928), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6048), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [164908] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5751), 1,
      anon_sym_DASH_GT,
    ACTIONS(5753), 1,
      anon_sym_with,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3909), 1,
      sym_attribute,
    STATE(4082), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6052), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [164950] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6054), 1,
      anon_sym_end,
    STATE(3910), 1,
      sym_attribute,
    STATE(4003), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [165004] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6058), 1,
      anon_sym_let,
    ACTIONS(6060), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3911), 2,
      sym_attribute,
      aux_sym_class_definition_repeat1,
    ACTIONS(6056), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [165040] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2546), 1,
      sym__lowercase_identifier,
    ACTIONS(2550), 1,
      anon_sym_LBRACK,
    ACTIONS(2558), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2562), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2564), 1,
      anon_sym_begin,
    ACTIONS(6063), 1,
      anon_sym_LPAREN,
    ACTIONS(6065), 1,
      anon_sym_LBRACE,
    ACTIONS(6067), 1,
      sym__uppercase_identifier,
    STATE(1924), 1,
      sym_parenthesized_operator,
    STATE(1955), 1,
      sym__value_name,
    STATE(1980), 1,
      sym__constructor_name,
    STATE(2046), 1,
      sym__delimited_expression,
    STATE(3912), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1953), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [165096] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6069), 1,
      anon_sym_end,
    STATE(3913), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [165150] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5791), 1,
      anon_sym_end,
    STATE(3914), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [165204] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5739), 1,
      anon_sym_functor,
    ACTIONS(5741), 1,
      anon_sym_struct,
    STATE(3079), 1,
      sym_module_path,
    STATE(3169), 1,
      sym__module_expression,
    STATE(3915), 1,
      sym_attribute,
    STATE(4076), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [165258] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6071), 1,
      anon_sym_end,
    STATE(3916), 1,
      sym_attribute,
    STATE(3919), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [165312] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(355), 1,
      anon_sym_LBRACK,
    ACTIONS(365), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(385), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(387), 1,
      anon_sym_begin,
    ACTIONS(743), 1,
      sym__lowercase_identifier,
    ACTIONS(6073), 1,
      anon_sym_LPAREN,
    ACTIONS(6075), 1,
      anon_sym_LBRACE,
    ACTIONS(6077), 1,
      sym__uppercase_identifier,
    STATE(765), 1,
      sym__delimited_expression,
    STATE(766), 1,
      sym__value_name,
    STATE(829), 1,
      sym__constructor_name,
    STATE(830), 1,
      sym_parenthesized_operator,
    STATE(3917), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(785), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [165368] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(6079), 1,
      sym__lowercase_identifier,
    ACTIONS(6081), 1,
      anon_sym_LPAREN,
    ACTIONS(6083), 1,
      anon_sym_LBRACE,
    ACTIONS(6085), 1,
      sym__uppercase_identifier,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1753), 1,
      sym__constructor_name,
    STATE(1758), 1,
      sym__delimited_expression,
    STATE(1759), 1,
      sym__value_name,
    STATE(3918), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [165424] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6087), 1,
      anon_sym_end,
    STATE(3919), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [165478] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6089), 1,
      anon_sym_end,
    STATE(3920), 1,
      sym_attribute,
    STATE(3925), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [165532] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6091), 1,
      anon_sym_end,
    STATE(3921), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [165586] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6093), 1,
      anon_sym_end,
    STATE(3922), 1,
      sym_attribute,
    STATE(3927), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [165640] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6036), 1,
      anon_sym_and,
    ACTIONS(6097), 1,
      anon_sym_let,
    STATE(3907), 1,
      aux_sym_module_definition_repeat1,
    STATE(3923), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6095), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [165678] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5100), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5962), 1,
      anon_sym_PIPE,
    STATE(3924), 1,
      sym_attribute,
    STATE(3926), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5098), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [165716] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6099), 1,
      anon_sym_end,
    STATE(3925), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [165770] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5116), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6101), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3926), 2,
      sym_attribute,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5114), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [165806] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6104), 1,
      anon_sym_end,
    STATE(3927), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [165860] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6036), 1,
      anon_sym_and,
    ACTIONS(6097), 1,
      anon_sym_let,
    STATE(3900), 1,
      aux_sym_module_definition_repeat1,
    STATE(3928), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6095), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [165898] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6106), 1,
      anon_sym_end,
    STATE(3929), 1,
      sym_attribute,
    STATE(3932), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [165952] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5705), 1,
      anon_sym_end,
    STATE(3930), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166006] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2594), 1,
      sym__lowercase_identifier,
    ACTIONS(2598), 1,
      anon_sym_LBRACK,
    ACTIONS(2606), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2610), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2612), 1,
      anon_sym_begin,
    ACTIONS(6108), 1,
      anon_sym_LPAREN,
    ACTIONS(6110), 1,
      anon_sym_LBRACE,
    ACTIONS(6112), 1,
      sym__uppercase_identifier,
    STATE(1822), 1,
      sym_parenthesized_operator,
    STATE(1836), 1,
      sym__constructor_name,
    STATE(1883), 1,
      sym__value_name,
    STATE(1915), 1,
      sym__delimited_expression,
    STATE(3931), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1897), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [166062] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6114), 1,
      anon_sym_end,
    STATE(3932), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166116] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6116), 1,
      anon_sym_end,
    STATE(3933), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166170] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5926), 1,
      anon_sym_let,
    ACTIONS(5928), 1,
      anon_sym_and,
    STATE(3890), 1,
      aux_sym_type_definition_repeat1,
    STATE(3934), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5924), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [166208] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6118), 1,
      anon_sym_end,
    STATE(3935), 1,
      sym_attribute,
    STATE(3938), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166262] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5928), 1,
      anon_sym_and,
    ACTIONS(6122), 1,
      anon_sym_let,
    STATE(3936), 1,
      sym_attribute,
    STATE(3960), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6120), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [166300] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5739), 1,
      anon_sym_functor,
    ACTIONS(5741), 1,
      anon_sym_struct,
    STATE(3079), 1,
      sym_module_path,
    STATE(3120), 1,
      sym__module_expression,
    STATE(3937), 1,
      sym_attribute,
    STATE(4101), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [166354] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6124), 1,
      anon_sym_end,
    STATE(3938), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166408] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6126), 1,
      anon_sym_end,
    STATE(3939), 1,
      sym_attribute,
    STATE(4035), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166462] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6128), 1,
      anon_sym_end,
    STATE(3940), 1,
      sym_attribute,
    STATE(3942), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166516] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6130), 1,
      anon_sym_end,
    STATE(3941), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166570] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6132), 1,
      anon_sym_end,
    STATE(3942), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166624] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5721), 1,
      anon_sym_end,
    STATE(3943), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166678] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1855), 1,
      sym__lowercase_identifier,
    ACTIONS(1861), 1,
      anon_sym_LBRACK,
    ACTIONS(1869), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(1873), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(1875), 1,
      anon_sym_begin,
    ACTIONS(6134), 1,
      anon_sym_LPAREN,
    ACTIONS(6136), 1,
      anon_sym_LBRACE,
    ACTIONS(6138), 1,
      sym__uppercase_identifier,
    STATE(2366), 1,
      sym__value_name,
    STATE(2389), 1,
      sym__constructor_name,
    STATE(2400), 1,
      sym_parenthesized_operator,
    STATE(2402), 1,
      sym__delimited_expression,
    STATE(3944), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2330), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [166734] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6140), 1,
      anon_sym_end,
    STATE(3945), 1,
      sym_attribute,
    STATE(3948), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166788] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6142), 1,
      anon_sym_end,
    STATE(3946), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166842] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6144), 1,
      anon_sym_end,
    STATE(3947), 1,
      sym_attribute,
    STATE(3949), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166896] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6146), 1,
      anon_sym_end,
    STATE(3948), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [166950] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6148), 1,
      anon_sym_end,
    STATE(3949), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167004] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6150), 1,
      anon_sym_end,
    STATE(3950), 1,
      sym_attribute,
    STATE(3951), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167058] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6152), 1,
      anon_sym_end,
    STATE(3951), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167112] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    ACTIONS(6154), 1,
      anon_sym_RPAREN,
    ACTIONS(6156), 1,
      anon_sym_val,
    STATE(3079), 1,
      sym_module_path,
    STATE(3952), 1,
      sym_attribute,
    STATE(4318), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [167166] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2676), 1,
      sym__lowercase_identifier,
    ACTIONS(2684), 1,
      anon_sym_begin,
    ACTIONS(6083), 1,
      anon_sym_LBRACE,
    ACTIONS(6085), 1,
      sym__uppercase_identifier,
    ACTIONS(6158), 1,
      anon_sym_LPAREN,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1753), 1,
      sym__constructor_name,
    STATE(1758), 1,
      sym__delimited_expression,
    STATE(1759), 1,
      sym__value_name,
    STATE(3953), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [167222] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    ACTIONS(6160), 1,
      anon_sym_RPAREN,
    ACTIONS(6162), 1,
      anon_sym_val,
    STATE(3079), 1,
      sym_module_path,
    STATE(3954), 1,
      sym_attribute,
    STATE(4313), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [167276] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6164), 1,
      anon_sym_end,
    STATE(3955), 1,
      sym_attribute,
    STATE(3956), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167330] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6166), 1,
      anon_sym_end,
    STATE(3956), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167384] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5856), 1,
      anon_sym_end,
    STATE(3957), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167438] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2510), 1,
      anon_sym_LBRACK,
    ACTIONS(2516), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(2520), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(2522), 1,
      anon_sym_begin,
    ACTIONS(2676), 1,
      sym__lowercase_identifier,
    ACTIONS(6081), 1,
      anon_sym_LPAREN,
    ACTIONS(6083), 1,
      anon_sym_LBRACE,
    ACTIONS(6085), 1,
      sym__uppercase_identifier,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(1753), 1,
      sym__constructor_name,
    STATE(1758), 1,
      sym__delimited_expression,
    STATE(1759), 1,
      sym__value_name,
    STATE(3958), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1718), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [167494] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6168), 1,
      anon_sym_end,
    STATE(3959), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167548] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5928), 1,
      anon_sym_and,
    ACTIONS(6172), 1,
      anon_sym_let,
    STATE(3890), 1,
      aux_sym_type_definition_repeat1,
    STATE(3960), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6170), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [167586] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6174), 1,
      anon_sym_end,
    STATE(3961), 1,
      sym_attribute,
    STATE(3966), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167640] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(3962), 1,
      sym_attribute,
    STATE(4086), 1,
      sym__attribute,
    STATE(4427), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [167694] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6178), 1,
      anon_sym_let,
    ACTIONS(6180), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3963), 2,
      sym_attribute,
      aux_sym_class_type_definition_repeat1,
    ACTIONS(6176), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [167730] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6183), 1,
      anon_sym_end,
    STATE(3964), 1,
      sym_attribute,
    STATE(3970), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167784] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6187), 1,
      anon_sym_let,
    ACTIONS(6189), 1,
      anon_sym_and,
    STATE(3963), 1,
      aux_sym_class_type_definition_repeat1,
    STATE(3965), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6185), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [167822] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6191), 1,
      anon_sym_end,
    STATE(3966), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167876] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6189), 1,
      anon_sym_and,
    ACTIONS(6195), 1,
      anon_sym_let,
    STATE(3965), 1,
      aux_sym_class_type_definition_repeat1,
    STATE(3967), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6193), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [167914] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6197), 1,
      anon_sym_end,
    STATE(3968), 1,
      sym_attribute,
    STATE(3979), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [167968] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6199), 1,
      anon_sym_end,
    STATE(3969), 1,
      sym_attribute,
    STATE(3975), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [168022] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6201), 1,
      anon_sym_end,
    STATE(3970), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [168076] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6203), 1,
      anon_sym_end,
    STATE(3971), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [168130] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(3972), 1,
      sym_attribute,
    STATE(4104), 1,
      sym__attribute,
    STATE(4356), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [168184] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6187), 1,
      anon_sym_let,
    ACTIONS(6189), 1,
      anon_sym_and,
    STATE(3973), 1,
      sym_attribute,
    STATE(4002), 1,
      aux_sym_class_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6185), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [168222] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6205), 1,
      anon_sym_end,
    STATE(3974), 1,
      sym_attribute,
    STATE(3976), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [168276] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6207), 1,
      anon_sym_end,
    STATE(3975), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [168330] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6209), 1,
      anon_sym_end,
    STATE(3976), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [168384] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5934), 1,
      anon_sym_and,
    ACTIONS(6213), 1,
      anon_sym_let,
    STATE(3911), 1,
      aux_sym_class_definition_repeat1,
    STATE(3977), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6211), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [168422] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5934), 1,
      anon_sym_and,
    ACTIONS(6213), 1,
      anon_sym_let,
    STATE(3978), 1,
      sym_attribute,
    STATE(4021), 1,
      aux_sym_class_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6211), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [168460] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6215), 1,
      anon_sym_end,
    STATE(3979), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [168514] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(3980), 1,
      sym_attribute,
    STATE(4114), 1,
      sym__attribute,
    STATE(4304), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [168568] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(3981), 1,
      sym_attribute,
    STATE(4119), 1,
      sym__attribute,
    STATE(4321), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [168622] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(185), 1,
      anon_sym_LBRACK,
    ACTIONS(195), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(207), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(209), 1,
      anon_sym_begin,
    ACTIONS(965), 1,
      sym__lowercase_identifier,
    ACTIONS(6217), 1,
      anon_sym_LPAREN,
    ACTIONS(6219), 1,
      anon_sym_LBRACE,
    ACTIONS(6221), 1,
      sym__uppercase_identifier,
    STATE(1048), 1,
      sym__value_name,
    STATE(1051), 1,
      sym__constructor_name,
    STATE(1071), 1,
      sym__delimited_expression,
    STATE(1077), 1,
      sym_parenthesized_operator,
    STATE(3982), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1118), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [168678] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(3983), 1,
      sym_attribute,
    STATE(4087), 1,
      sym__attribute,
    STATE(4342), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [168732] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(3984), 1,
      sym_attribute,
    STATE(4112), 1,
      sym__attribute,
    STATE(4362), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [168786] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(3985), 1,
      sym_attribute,
    STATE(4067), 1,
      sym__attribute,
    STATE(4393), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [168840] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(3986), 1,
      sym_attribute,
    STATE(4073), 1,
      sym__attribute,
    STATE(4416), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [168894] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6036), 1,
      anon_sym_and,
    ACTIONS(6225), 1,
      anon_sym_let,
    STATE(3900), 1,
      aux_sym_module_definition_repeat1,
    STATE(3987), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6223), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [168932] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6227), 1,
      anon_sym_end,
    STATE(3988), 1,
      sym_attribute,
    STATE(3996), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [168986] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6229), 1,
      anon_sym_end,
    STATE(3989), 1,
      sym_attribute,
    STATE(3990), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169040] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6231), 1,
      anon_sym_end,
    STATE(3990), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169094] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5928), 1,
      anon_sym_and,
    ACTIONS(6235), 1,
      anon_sym_let,
    STATE(3934), 1,
      aux_sym_type_definition_repeat1,
    STATE(3991), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6233), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [169132] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6237), 1,
      anon_sym_end,
    STATE(3992), 1,
      sym_attribute,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169186] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6239), 1,
      anon_sym_constraint,
    ACTIONS(6242), 1,
      anon_sym_val,
    ACTIONS(6245), 1,
      anon_sym_end,
    ACTIONS(6247), 1,
      anon_sym_inherit,
    ACTIONS(6250), 1,
      anon_sym_method,
    ACTIONS(6253), 1,
      anon_sym_initializer,
    ACTIONS(6256), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(6259), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(6262), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3993), 2,
      sym_attribute,
      aux_sym_object_expression_repeat1,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169238] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5751), 1,
      anon_sym_DASH_GT,
    ACTIONS(5753), 1,
      anon_sym_with,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3802), 1,
      aux_sym_expression_item_repeat1,
    STATE(3994), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5204), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [169280] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5751), 1,
      anon_sym_DASH_GT,
    ACTIONS(5753), 1,
      anon_sym_with,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3823), 1,
      aux_sym_expression_item_repeat1,
    STATE(3995), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5133), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [169322] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6265), 1,
      anon_sym_end,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    STATE(3996), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169376] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5872), 1,
      anon_sym_end,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    STATE(3997), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169430] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(111), 1,
      anon_sym_LBRACK,
    ACTIONS(121), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(141), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(143), 1,
      anon_sym_begin,
    ACTIONS(831), 1,
      sym__lowercase_identifier,
    ACTIONS(6267), 1,
      anon_sym_LPAREN,
    ACTIONS(6269), 1,
      anon_sym_LBRACE,
    ACTIONS(6271), 1,
      sym__uppercase_identifier,
    STATE(964), 1,
      sym__value_name,
    STATE(978), 1,
      sym_parenthesized_operator,
    STATE(985), 1,
      sym__constructor_name,
    STATE(1021), 1,
      sym__delimited_expression,
    STATE(3998), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(988), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [169486] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6273), 1,
      anon_sym_end,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    STATE(3999), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169540] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6275), 1,
      anon_sym_end,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    STATE(4000), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169594] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6277), 1,
      anon_sym_end,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    STATE(4001), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169648] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6189), 1,
      anon_sym_and,
    ACTIONS(6281), 1,
      anon_sym_let,
    STATE(3963), 1,
      aux_sym_class_type_definition_repeat1,
    STATE(4002), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6279), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [169686] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6283), 1,
      anon_sym_end,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    STATE(4003), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169740] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6285), 1,
      anon_sym_end,
    STATE(4004), 1,
      sym_attribute,
    STATE(4007), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169794] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5751), 1,
      anon_sym_DASH_GT,
    ACTIONS(5753), 1,
      anon_sym_with,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(4005), 1,
      sym_attribute,
    STATE(4071), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6287), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [169836] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5928), 1,
      anon_sym_and,
    ACTIONS(6122), 1,
      anon_sym_let,
    STATE(3890), 1,
      aux_sym_type_definition_repeat1,
    STATE(4006), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6120), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [169874] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6289), 1,
      anon_sym_end,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    STATE(4007), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169928] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6291), 1,
      anon_sym_end,
    STATE(4008), 1,
      sym_attribute,
    STATE(4012), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [169982] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6293), 1,
      anon_sym_end,
    STATE(4009), 1,
      sym_attribute,
    STATE(4013), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170036] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6295), 1,
      anon_sym_COLON2,
    STATE(4010), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3050), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT,
      sym__lowercase_identifier,
    ACTIONS(3052), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_COLON_GT,
      anon_sym_POUND,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym__uppercase_identifier,
  [170072] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6297), 1,
      anon_sym_end,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    STATE(4011), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170126] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6299), 1,
      anon_sym_end,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    STATE(4012), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170180] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6301), 1,
      anon_sym_end,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    STATE(4013), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170234] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6303), 1,
      anon_sym_end,
    STATE(4014), 1,
      sym_attribute,
    STATE(4015), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170288] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6305), 1,
      anon_sym_end,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    STATE(4015), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170342] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6307), 1,
      anon_sym_end,
    STATE(3999), 1,
      aux_sym_object_expression_repeat1,
    STATE(4016), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170396] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6309), 1,
      anon_sym_end,
    STATE(3852), 1,
      aux_sym_object_expression_repeat1,
    STATE(4017), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170450] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5902), 1,
      anon_sym_end,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    STATE(4018), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170504] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(615), 1,
      anon_sym_LBRACK,
    ACTIONS(625), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(645), 1,
      anon_sym_LBRACE_LT,
    ACTIONS(647), 1,
      anon_sym_begin,
    ACTIONS(1337), 1,
      sym__lowercase_identifier,
    ACTIONS(6311), 1,
      anon_sym_LPAREN,
    ACTIONS(6313), 1,
      anon_sym_LBRACE,
    ACTIONS(6315), 1,
      sym__uppercase_identifier,
    STATE(1560), 1,
      sym__delimited_expression,
    STATE(1574), 1,
      sym__constructor_name,
    STATE(1584), 1,
      sym__value_name,
    STATE(1593), 1,
      sym_parenthesized_operator,
    STATE(4019), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(1545), 7,
      sym_list_expression,
      sym_array_expression,
      sym_record_expression,
      sym_package_expression,
      sym_object_copy_expression,
      sym_parenthesized_expression,
      sym_unit,
  [170560] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6317), 1,
      anon_sym_end,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    STATE(4020), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170614] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(5934), 1,
      anon_sym_and,
    ACTIONS(6321), 1,
      anon_sym_let,
    STATE(3911), 1,
      aux_sym_class_definition_repeat1,
    STATE(4021), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6319), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [170652] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6323), 1,
      anon_sym_end,
    STATE(4022), 1,
      sym_attribute,
    STATE(4023), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170706] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6325), 1,
      anon_sym_end,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    STATE(4023), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170760] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6327), 1,
      anon_sym_end,
    STATE(4024), 1,
      sym_attribute,
    STATE(4026), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170814] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6329), 1,
      anon_sym_end,
    STATE(4025), 1,
      sym_attribute,
    STATE(4027), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170868] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6331), 1,
      anon_sym_end,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    STATE(4026), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170922] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6333), 1,
      anon_sym_end,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    STATE(4027), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [170976] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6335), 1,
      anon_sym_end,
    STATE(4028), 1,
      sym_attribute,
    STATE(4029), 1,
      aux_sym_object_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [171030] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6337), 1,
      anon_sym_end,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    STATE(4029), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [171084] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4030), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1623), 2,
      anon_sym_COLON2,
      anon_sym_POUND,
    ACTIONS(1625), 17,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_COLON_COLON,
      anon_sym_when,
      sym_hash_operator,
      aux_sym_tag_token1,
  [171118] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6339), 1,
      anon_sym_end,
    STATE(3882), 1,
      aux_sym_object_expression_repeat1,
    STATE(4031), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [171172] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6341), 1,
      anon_sym_end,
    STATE(3853), 1,
      aux_sym_object_expression_repeat1,
    STATE(4032), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [171226] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(3837), 1,
      aux_sym_expression_item_repeat1,
    STATE(4033), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6343), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
    ACTIONS(5208), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [171266] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5906), 1,
      anon_sym_end,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    STATE(4034), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [171320] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6345), 1,
      anon_sym_end,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    STATE(4035), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [171374] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6347), 1,
      anon_sym_end,
    STATE(3859), 1,
      aux_sym_object_expression_repeat1,
    STATE(4036), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [171428] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5074), 1,
      anon_sym_val,
    ACTIONS(5078), 1,
      anon_sym_inherit,
    ACTIONS(5080), 1,
      anon_sym_method,
    ACTIONS(5082), 1,
      anon_sym_initializer,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5821), 1,
      anon_sym_end,
    STATE(3993), 1,
      aux_sym_object_expression_repeat1,
    STATE(4037), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4540), 2,
      sym__class_field,
      sym_floating_attribute,
    STATE(4617), 7,
      sym_type_parameter_constraint,
      sym_inheritance_definition,
      sym_instance_variable_definition,
      sym_method_definition,
      sym_class_initializer,
      sym_item_extension,
      sym_quoted_item_extension,
  [171482] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6351), 1,
      anon_sym_let,
    ACTIONS(6353), 1,
      anon_sym_in,
    STATE(4038), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6349), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171517] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6178), 1,
      anon_sym_let,
    STATE(4039), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6176), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171550] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6355), 1,
      anon_sym_EQ,
    ACTIONS(6357), 1,
      anon_sym_COLON2,
    ACTIONS(6361), 1,
      anon_sym_COLON_GT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4040), 1,
      sym_attribute,
    STATE(4224), 1,
      sym__coerced,
    STATE(4268), 1,
      sym__typed,
    STATE(4290), 1,
      sym__type_constrained,
    STATE(4320), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6359), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [171599] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6365), 1,
      anon_sym_let,
    ACTIONS(6367), 1,
      anon_sym_DOT,
    STATE(4041), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6363), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171634] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6357), 1,
      anon_sym_COLON2,
    ACTIONS(6361), 1,
      anon_sym_COLON_GT,
    ACTIONS(6369), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4042), 1,
      sym_attribute,
    STATE(4224), 1,
      sym__coerced,
    STATE(4268), 1,
      sym__typed,
    STATE(4271), 1,
      sym__type_constrained,
    STATE(4413), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6371), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [171683] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(6373), 1,
      sym__lowercase_identifier,
    ACTIONS(6375), 1,
      anon_sym_LPAREN,
    ACTIONS(6377), 1,
      anon_sym_nonrec,
    ACTIONS(6381), 1,
      anon_sym__,
    STATE(4043), 1,
      sym_attribute,
    STATE(4156), 1,
      sym__attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5467), 1,
      sym__type_params,
    STATE(5561), 1,
      sym_type_binding,
    STATE(6063), 1,
      sym__type_param,
    STATE(6196), 1,
      sym_extended_module_path,
    STATE(6383), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [171744] = 14,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    ACTIONS(6156), 1,
      anon_sym_val,
    STATE(3079), 1,
      sym_module_path,
    STATE(4044), 1,
      sym_attribute,
    STATE(4318), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [171795] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6000), 1,
      anon_sym_let,
    STATE(4045), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5998), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171828] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(6375), 1,
      anon_sym_LPAREN,
    ACTIONS(6381), 1,
      anon_sym__,
    ACTIONS(6383), 1,
      sym__lowercase_identifier,
    ACTIONS(6385), 1,
      anon_sym_nonrec,
    STATE(3991), 1,
      sym_type_binding,
    STATE(4046), 1,
      sym_attribute,
    STATE(4137), 1,
      sym__attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5379), 1,
      sym__type_params,
    STATE(6063), 1,
      sym__type_param,
    STATE(6196), 1,
      sym_extended_module_path,
    STATE(6446), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [171889] = 19,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(6375), 1,
      anon_sym_LPAREN,
    ACTIONS(6381), 1,
      anon_sym__,
    ACTIONS(6387), 1,
      sym__lowercase_identifier,
    ACTIONS(6389), 1,
      anon_sym_nonrec,
    STATE(4047), 1,
      sym_attribute,
    STATE(4130), 1,
      sym_type_binding,
    STATE(4134), 1,
      sym__attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5436), 1,
      sym__type_params,
    STATE(6063), 1,
      sym__type_param,
    STATE(6196), 1,
      sym_extended_module_path,
    STATE(6407), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [171950] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6058), 1,
      anon_sym_let,
    STATE(4048), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6056), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [171983] = 14,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    ACTIONS(6162), 1,
      anon_sym_val,
    STATE(3079), 1,
      sym_module_path,
    STATE(4049), 1,
      sym_attribute,
    STATE(4313), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [172034] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4758), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5328), 1,
      anon_sym_of,
    ACTIONS(5330), 1,
      anon_sym_COLON2,
    ACTIONS(6391), 1,
      anon_sym_EQ,
    STATE(4050), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4756), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [172073] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6027), 1,
      anon_sym_let,
    STATE(4051), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6025), 17,
      anon_sym_SEMI_SEMI,
      anon_sym_and,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172106] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6393), 1,
      anon_sym_LPAREN,
    ACTIONS(6395), 1,
      anon_sym_val,
    ACTIONS(6397), 1,
      anon_sym_end,
    ACTIONS(6399), 1,
      anon_sym_inherit,
    ACTIONS(6401), 1,
      anon_sym_method,
    STATE(4052), 1,
      sym_attribute,
    STATE(4093), 1,
      aux_sym_class_body_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4623), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(4624), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [172159] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4053), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6405), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6403), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172192] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6357), 1,
      anon_sym_COLON2,
    ACTIONS(6361), 1,
      anon_sym_COLON_GT,
    ACTIONS(6407), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4054), 1,
      sym_attribute,
    STATE(4224), 1,
      sym__coerced,
    STATE(4261), 1,
      sym__type_constrained,
    STATE(4268), 1,
      sym__typed,
    STATE(4332), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6409), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [172241] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4055), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6040), 2,
      anon_sym_let,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6038), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172274] = 15,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6395), 1,
      anon_sym_val,
    ACTIONS(6399), 1,
      anon_sym_inherit,
    ACTIONS(6401), 1,
      anon_sym_method,
    ACTIONS(6411), 1,
      anon_sym_LPAREN,
    ACTIONS(6413), 1,
      anon_sym_end,
    STATE(4056), 1,
      sym_attribute,
    STATE(4095), 1,
      aux_sym_class_body_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4623), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(4624), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [172327] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6395), 1,
      anon_sym_val,
    ACTIONS(6399), 1,
      anon_sym_inherit,
    ACTIONS(6401), 1,
      anon_sym_method,
    ACTIONS(6415), 1,
      anon_sym_end,
    STATE(4057), 1,
      sym_attribute,
    STATE(4070), 1,
      aux_sym_class_body_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4623), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(4624), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [172377] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5888), 1,
      anon_sym_LPAREN,
    ACTIONS(5894), 1,
      anon_sym_struct,
    ACTIONS(5896), 1,
      sym__uppercase_identifier,
    ACTIONS(6417), 1,
      anon_sym_functor,
    STATE(4058), 1,
      sym_attribute,
    STATE(4218), 1,
      sym__module_expression,
    STATE(4479), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4495), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [172425] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6421), 1,
      anon_sym_let,
    STATE(4059), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6419), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172457] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1785), 1,
      anon_sym_let,
    STATE(4060), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1787), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172489] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(4061), 1,
      sym_attribute,
    STATE(4346), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [172537] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1797), 1,
      anon_sym_let,
    STATE(4062), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1799), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172569] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1805), 1,
      anon_sym_let,
    STATE(4063), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1807), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172601] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5888), 1,
      anon_sym_LPAREN,
    ACTIONS(5894), 1,
      anon_sym_struct,
    ACTIONS(5896), 1,
      sym__uppercase_identifier,
    ACTIONS(6417), 1,
      anon_sym_functor,
    STATE(4064), 1,
      sym_attribute,
    STATE(4219), 1,
      sym__module_expression,
    STATE(4479), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4495), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [172649] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1757), 1,
      anon_sym_let,
    STATE(4065), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1759), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [172681] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5888), 1,
      anon_sym_LPAREN,
    ACTIONS(5892), 1,
      anon_sym_functor,
    ACTIONS(5894), 1,
      anon_sym_struct,
    ACTIONS(5896), 1,
      sym__uppercase_identifier,
    STATE(4066), 1,
      sym_attribute,
    STATE(4284), 1,
      sym__module_expression,
    STATE(4479), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4495), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [172729] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(4067), 1,
      sym_attribute,
    STATE(4397), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [172777] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5739), 1,
      anon_sym_functor,
    ACTIONS(5741), 1,
      anon_sym_struct,
    STATE(3079), 1,
      sym_module_path,
    STATE(3120), 1,
      sym__module_expression,
    STATE(4068), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [172825] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5888), 1,
      anon_sym_LPAREN,
    ACTIONS(5894), 1,
      anon_sym_struct,
    ACTIONS(5896), 1,
      sym__uppercase_identifier,
    ACTIONS(6417), 1,
      anon_sym_functor,
    STATE(4069), 1,
      sym_attribute,
    STATE(4220), 1,
      sym__module_expression,
    STATE(4479), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4495), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [172873] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6395), 1,
      anon_sym_val,
    ACTIONS(6399), 1,
      anon_sym_inherit,
    ACTIONS(6401), 1,
      anon_sym_method,
    ACTIONS(6423), 1,
      anon_sym_end,
    STATE(4070), 1,
      sym_attribute,
    STATE(4079), 1,
      aux_sym_class_body_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4623), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(4624), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [172923] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(4071), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6425), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [172959] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4072), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1687), 3,
      anon_sym_COLON2,
      anon_sym_DOT,
      anon_sym_POUND,
    ACTIONS(1689), 14,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      sym_hash_operator,
  [172991] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(4073), 1,
      sym_attribute,
    STATE(4420), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173039] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5888), 1,
      anon_sym_LPAREN,
    ACTIONS(5892), 1,
      anon_sym_functor,
    ACTIONS(5894), 1,
      anon_sym_struct,
    ACTIONS(5896), 1,
      sym__uppercase_identifier,
    STATE(4074), 1,
      sym_attribute,
    STATE(4262), 1,
      sym__module_expression,
    STATE(4479), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4495), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173087] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5888), 1,
      anon_sym_LPAREN,
    ACTIONS(5894), 1,
      anon_sym_struct,
    ACTIONS(5896), 1,
      sym__uppercase_identifier,
    ACTIONS(6417), 1,
      anon_sym_functor,
    STATE(4075), 1,
      sym_attribute,
    STATE(4223), 1,
      sym__module_expression,
    STATE(4479), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4495), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173135] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5739), 1,
      anon_sym_functor,
    ACTIONS(5741), 1,
      anon_sym_struct,
    STATE(3079), 1,
      sym_module_path,
    STATE(3143), 1,
      sym__module_expression,
    STATE(4076), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173183] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(4077), 1,
      sym_attribute,
    STATE(4273), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173231] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6365), 1,
      anon_sym_let,
    STATE(4078), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6363), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173263] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6427), 1,
      anon_sym_constraint,
    ACTIONS(6430), 1,
      anon_sym_val,
    ACTIONS(6433), 1,
      anon_sym_end,
    ACTIONS(6435), 1,
      anon_sym_inherit,
    ACTIONS(6438), 1,
      anon_sym_method,
    ACTIONS(6441), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(6444), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(6447), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4079), 2,
      sym_attribute,
      aux_sym_class_body_type_repeat1,
    STATE(4623), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(4624), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [173311] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1813), 1,
      anon_sym_let,
    STATE(4080), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1815), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173343] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5739), 1,
      anon_sym_functor,
    ACTIONS(5741), 1,
      anon_sym_struct,
    STATE(3079), 1,
      sym_module_path,
    STATE(3083), 1,
      sym__module_expression,
    STATE(4081), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173391] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1503), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3522), 1,
      aux_sym_expression_item_repeat1,
    STATE(3701), 1,
      sym_item_attribute,
    STATE(4082), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6450), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [173427] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5739), 1,
      anon_sym_functor,
    ACTIONS(5741), 1,
      anon_sym_struct,
    STATE(3079), 1,
      sym_module_path,
    STATE(3097), 1,
      sym__module_expression,
    STATE(4083), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173475] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5739), 1,
      anon_sym_functor,
    ACTIONS(5741), 1,
      anon_sym_struct,
    STATE(2672), 1,
      sym__module_expression,
    STATE(3079), 1,
      sym_module_path,
    STATE(4084), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173523] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5739), 1,
      anon_sym_functor,
    ACTIONS(5741), 1,
      anon_sym_struct,
    STATE(2734), 1,
      sym__module_expression,
    STATE(3079), 1,
      sym_module_path,
    STATE(4085), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173571] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(4086), 1,
      sym_attribute,
    STATE(4302), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173619] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(4087), 1,
      sym_attribute,
    STATE(4343), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173667] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4088), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1683), 3,
      anon_sym_COLON2,
      anon_sym_DOT,
      anon_sym_POUND,
    ACTIONS(1685), 14,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      sym_hash_operator,
  [173699] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4089), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1829), 3,
      anon_sym_COLON2,
      anon_sym_DOT,
      anon_sym_POUND,
    ACTIONS(1831), 14,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      sym_hash_operator,
  [173731] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(4090), 1,
      sym_attribute,
    STATE(4339), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173779] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1829), 1,
      anon_sym_let,
    STATE(4091), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1831), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [173811] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5888), 1,
      anon_sym_LPAREN,
    ACTIONS(5892), 1,
      anon_sym_functor,
    ACTIONS(5894), 1,
      anon_sym_struct,
    ACTIONS(5896), 1,
      sym__uppercase_identifier,
    STATE(4092), 1,
      sym_attribute,
    STATE(4244), 1,
      sym__module_expression,
    STATE(4479), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4495), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173859] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6395), 1,
      anon_sym_val,
    ACTIONS(6399), 1,
      anon_sym_inherit,
    ACTIONS(6401), 1,
      anon_sym_method,
    ACTIONS(6452), 1,
      anon_sym_end,
    STATE(4079), 1,
      aux_sym_class_body_type_repeat1,
    STATE(4093), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4623), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(4624), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [173909] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5888), 1,
      anon_sym_LPAREN,
    ACTIONS(5894), 1,
      anon_sym_struct,
    ACTIONS(5896), 1,
      sym__uppercase_identifier,
    ACTIONS(6417), 1,
      anon_sym_functor,
    STATE(4094), 1,
      sym_attribute,
    STATE(4227), 1,
      sym__module_expression,
    STATE(4479), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4495), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [173957] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6395), 1,
      anon_sym_val,
    ACTIONS(6399), 1,
      anon_sym_inherit,
    ACTIONS(6401), 1,
      anon_sym_method,
    ACTIONS(6454), 1,
      anon_sym_end,
    STATE(4079), 1,
      aux_sym_class_body_type_repeat1,
    STATE(4095), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4623), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(4624), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [174007] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1691), 1,
      anon_sym_let,
    STATE(4096), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1693), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174039] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5739), 1,
      anon_sym_functor,
    ACTIONS(5741), 1,
      anon_sym_struct,
    STATE(3079), 1,
      sym_module_path,
    STATE(3089), 1,
      sym__module_expression,
    STATE(4097), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174087] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5888), 1,
      anon_sym_LPAREN,
    ACTIONS(5894), 1,
      anon_sym_struct,
    ACTIONS(5896), 1,
      sym__uppercase_identifier,
    ACTIONS(6417), 1,
      anon_sym_functor,
    STATE(4098), 1,
      sym_attribute,
    STATE(4228), 1,
      sym__module_expression,
    STATE(4479), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4495), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174135] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5888), 1,
      anon_sym_LPAREN,
    ACTIONS(5894), 1,
      anon_sym_struct,
    ACTIONS(5896), 1,
      sym__uppercase_identifier,
    ACTIONS(6417), 1,
      anon_sym_functor,
    STATE(4099), 1,
      sym_attribute,
    STATE(4138), 1,
      sym__module_expression,
    STATE(4479), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4495), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174183] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4100), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1709), 3,
      anon_sym_COLON2,
      anon_sym_DOT,
      anon_sym_POUND,
    ACTIONS(1711), 14,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      sym_hash_operator,
  [174215] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5739), 1,
      anon_sym_functor,
    ACTIONS(5741), 1,
      anon_sym_struct,
    STATE(3079), 1,
      sym_module_path,
    STATE(3126), 1,
      sym__module_expression,
    STATE(4101), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174263] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5739), 1,
      anon_sym_functor,
    ACTIONS(5741), 1,
      anon_sym_struct,
    STATE(3079), 1,
      sym_module_path,
    STATE(3081), 1,
      sym__module_expression,
    STATE(4102), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174311] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5739), 1,
      anon_sym_functor,
    ACTIONS(5741), 1,
      anon_sym_struct,
    STATE(3079), 1,
      sym_module_path,
    STATE(3095), 1,
      sym__module_expression,
    STATE(4103), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174359] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(4104), 1,
      sym_attribute,
    STATE(4369), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174407] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1683), 1,
      anon_sym_let,
    STATE(4105), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1685), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174439] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(4106), 1,
      sym_attribute,
    STATE(4233), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174487] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6458), 1,
      anon_sym_let,
    STATE(4107), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6456), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174519] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1687), 1,
      anon_sym_let,
    STATE(4108), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1689), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174551] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5888), 1,
      anon_sym_LPAREN,
    ACTIONS(5892), 1,
      anon_sym_functor,
    ACTIONS(5894), 1,
      anon_sym_struct,
    ACTIONS(5896), 1,
      sym__uppercase_identifier,
    STATE(4109), 1,
      sym_attribute,
    STATE(4278), 1,
      sym__module_expression,
    STATE(4479), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4495), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174599] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6351), 1,
      anon_sym_let,
    STATE(4110), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6349), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174631] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6460), 1,
      anon_sym_let,
    STATE(4111), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3734), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174663] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(4112), 1,
      sym_attribute,
    STATE(4366), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174711] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1591), 1,
      anon_sym_let,
    STATE(4113), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1593), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174743] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(4114), 1,
      sym_attribute,
    STATE(4306), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174791] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6395), 1,
      anon_sym_val,
    ACTIONS(6399), 1,
      anon_sym_inherit,
    ACTIONS(6401), 1,
      anon_sym_method,
    ACTIONS(6462), 1,
      anon_sym_end,
    STATE(4115), 1,
      sym_attribute,
    STATE(4118), 1,
      aux_sym_class_body_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4623), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(4624), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [174841] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5739), 1,
      anon_sym_functor,
    ACTIONS(5741), 1,
      anon_sym_struct,
    STATE(2975), 1,
      sym__module_expression,
    STATE(3079), 1,
      sym_module_path,
    STATE(4116), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [174889] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1709), 1,
      anon_sym_let,
    STATE(4117), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1711), 16,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
      sym_let_operator,
      aux_sym_directive_token1,
  [174921] = 14,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(257), 1,
      anon_sym_LBRACK_AT_AT_AT,
    ACTIONS(5072), 1,
      anon_sym_constraint,
    ACTIONS(5084), 1,
      anon_sym_LBRACK_PERCENT_PERCENT,
    ACTIONS(5086), 1,
      anon_sym_LBRACE_PERCENT_PERCENT,
    ACTIONS(6395), 1,
      anon_sym_val,
    ACTIONS(6399), 1,
      anon_sym_inherit,
    ACTIONS(6401), 1,
      anon_sym_method,
    ACTIONS(6464), 1,
      anon_sym_end,
    STATE(4079), 1,
      aux_sym_class_body_type_repeat1,
    STATE(4118), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4623), 2,
      sym__class_field_specification,
      sym_floating_attribute,
    STATE(4624), 6,
      sym_inheritance_specification,
      sym_instance_variable_specification,
      sym_method_specification,
      sym_type_parameter_constraint,
      sym_item_extension,
      sym_quoted_item_extension,
  [174971] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(5735), 1,
      anon_sym_LPAREN,
    ACTIONS(5741), 1,
      anon_sym_struct,
    ACTIONS(5914), 1,
      anon_sym_functor,
    STATE(3079), 1,
      sym_module_path,
    STATE(4119), 1,
      sym_attribute,
    STATE(4326), 1,
      sym__module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3157), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [175019] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(5888), 1,
      anon_sym_LPAREN,
    ACTIONS(5894), 1,
      anon_sym_struct,
    ACTIONS(5896), 1,
      sym__uppercase_identifier,
    ACTIONS(6417), 1,
      anon_sym_functor,
    STATE(4120), 1,
      sym_attribute,
    STATE(4173), 1,
      sym__module_expression,
    STATE(4479), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4495), 4,
      sym__simple_module_expression,
      sym_structure,
      sym_functor,
      sym_module_application,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [175067] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6466), 1,
      anon_sym_DQUOTE,
    STATE(4121), 1,
      sym_attribute,
    STATE(4194), 1,
      aux_sym_string_content_repeat1,
    STATE(4247), 1,
      sym_escape_sequence,
    STATE(6901), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6470), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6472), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6468), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [175108] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6474), 1,
      anon_sym_DQUOTE,
    STATE(4122), 1,
      sym_attribute,
    STATE(4194), 1,
      aux_sym_string_content_repeat1,
    STATE(4247), 1,
      sym_escape_sequence,
    STATE(6792), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6470), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6472), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6468), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [175149] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6476), 1,
      anon_sym_DQUOTE,
    STATE(4123), 1,
      sym_attribute,
    STATE(4194), 1,
      aux_sym_string_content_repeat1,
    STATE(4247), 1,
      sym_escape_sequence,
    STATE(6845), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6470), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6472), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6468), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [175190] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4124), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1721), 2,
      anon_sym_COLON2,
      anon_sym_POUND,
    ACTIONS(1723), 14,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      sym_hash_operator,
  [175221] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6478), 1,
      anon_sym_DQUOTE,
    STATE(4125), 1,
      sym_attribute,
    STATE(4194), 1,
      aux_sym_string_content_repeat1,
    STATE(4247), 1,
      sym_escape_sequence,
    STATE(6967), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6470), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6472), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6468), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [175262] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6480), 1,
      anon_sym_DQUOTE,
    STATE(4126), 1,
      sym_attribute,
    STATE(4194), 1,
      aux_sym_string_content_repeat1,
    STATE(4247), 1,
      sym_escape_sequence,
    STATE(7040), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6470), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6472), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6468), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [175303] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6482), 1,
      anon_sym_DQUOTE,
    STATE(4127), 1,
      sym_attribute,
    STATE(4194), 1,
      aux_sym_string_content_repeat1,
    STATE(4247), 1,
      sym_escape_sequence,
    STATE(7078), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6470), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6472), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6468), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [175344] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6484), 1,
      anon_sym_and,
    STATE(4128), 1,
      sym_attribute,
    STATE(4154), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6170), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [175377] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6486), 1,
      anon_sym_and,
    STATE(4129), 1,
      sym_attribute,
    STATE(4164), 1,
      aux_sym_class_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6319), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [175410] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6484), 1,
      anon_sym_and,
    STATE(4130), 1,
      sym_attribute,
    STATE(4132), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6233), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [175443] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6484), 1,
      anon_sym_and,
    STATE(4131), 1,
      sym_attribute,
    STATE(4149), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5924), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [175476] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6484), 1,
      anon_sym_and,
    STATE(4132), 1,
      sym_attribute,
    STATE(4154), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5924), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [175509] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6488), 1,
      anon_sym_DQUOTE,
    STATE(4133), 1,
      sym_attribute,
    STATE(4194), 1,
      aux_sym_string_content_repeat1,
    STATE(4247), 1,
      sym_escape_sequence,
    STATE(7190), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6470), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6472), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6468), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [175550] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6375), 1,
      anon_sym_LPAREN,
    ACTIONS(6381), 1,
      anon_sym__,
    ACTIONS(6387), 1,
      sym__lowercase_identifier,
    ACTIONS(6490), 1,
      anon_sym_nonrec,
    STATE(4131), 1,
      sym_type_binding,
    STATE(4134), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5436), 1,
      sym__type_params,
    STATE(6063), 1,
      sym__type_param,
    STATE(6196), 1,
      sym_extended_module_path,
    STATE(6407), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [175605] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6492), 1,
      anon_sym_DQUOTE,
    STATE(4135), 1,
      sym_attribute,
    STATE(4194), 1,
      aux_sym_string_content_repeat1,
    STATE(4247), 1,
      sym_escape_sequence,
    STATE(6928), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6470), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6472), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6468), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [175646] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6494), 1,
      anon_sym_DQUOTE,
    STATE(4136), 1,
      sym_attribute,
    STATE(4194), 1,
      aux_sym_string_content_repeat1,
    STATE(4247), 1,
      sym_escape_sequence,
    STATE(6567), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6470), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6472), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6468), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [175687] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6375), 1,
      anon_sym_LPAREN,
    ACTIONS(6381), 1,
      anon_sym__,
    ACTIONS(6383), 1,
      sym__lowercase_identifier,
    ACTIONS(6496), 1,
      anon_sym_nonrec,
    STATE(3857), 1,
      sym_type_binding,
    STATE(4137), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5379), 1,
      sym__type_params,
    STATE(6063), 1,
      sym__type_param,
    STATE(6196), 1,
      sym_extended_module_path,
    STATE(6446), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [175742] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6498), 1,
      anon_sym_LPAREN,
    STATE(4138), 1,
      sym_attribute,
    STATE(4538), 1,
      sym__simple_module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4004), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [175781] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6015), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6500), 1,
      anon_sym_as,
    STATE(4139), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6013), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [175814] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6019), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6500), 1,
      anon_sym_as,
    STATE(4140), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6017), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [175847] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6040), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6500), 1,
      anon_sym_as,
    STATE(4141), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6038), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [175880] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6486), 1,
      anon_sym_and,
    STATE(4142), 1,
      sym_attribute,
    STATE(4146), 1,
      aux_sym_class_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5930), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [175913] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4143), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1825), 2,
      anon_sym_COLON2,
      anon_sym_POUND,
    ACTIONS(1827), 14,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
      sym_hash_operator,
  [175944] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6502), 1,
      anon_sym_DQUOTE,
    STATE(4144), 1,
      sym_attribute,
    STATE(4194), 1,
      aux_sym_string_content_repeat1,
    STATE(4247), 1,
      sym_escape_sequence,
    STATE(6861), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6470), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6472), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6468), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [175985] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6504), 1,
      anon_sym_DQUOTE,
    STATE(4145), 1,
      sym_attribute,
    STATE(4194), 1,
      aux_sym_string_content_repeat1,
    STATE(4247), 1,
      sym_escape_sequence,
    STATE(6699), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6470), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6472), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6468), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176026] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6486), 1,
      anon_sym_and,
    STATE(4146), 1,
      sym_attribute,
    STATE(4164), 1,
      aux_sym_class_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6211), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [176059] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6486), 1,
      anon_sym_and,
    STATE(4129), 1,
      aux_sym_class_definition_repeat1,
    STATE(4147), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6211), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [176092] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6506), 1,
      anon_sym_DQUOTE,
    STATE(4148), 1,
      sym_attribute,
    STATE(4194), 1,
      aux_sym_string_content_repeat1,
    STATE(4247), 1,
      sym_escape_sequence,
    STATE(7035), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6470), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6472), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6468), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176133] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6484), 1,
      anon_sym_and,
    STATE(4149), 1,
      sym_attribute,
    STATE(4154), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6120), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [176166] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6405), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4150), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6403), 15,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [176197] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6508), 1,
      anon_sym_DQUOTE,
    STATE(4151), 1,
      sym_attribute,
    STATE(4194), 1,
      aux_sym_string_content_repeat1,
    STATE(4247), 1,
      sym_escape_sequence,
    STATE(7245), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6470), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6472), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6468), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176238] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6040), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4152), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6038), 15,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [176269] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6510), 1,
      anon_sym_DQUOTE,
    STATE(4153), 1,
      sym_attribute,
    STATE(4194), 1,
      aux_sym_string_content_repeat1,
    STATE(4247), 1,
      sym_escape_sequence,
    STATE(6384), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6470), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6472), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6468), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176310] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6512), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4154), 2,
      sym_attribute,
      aux_sym_type_definition_repeat1,
    ACTIONS(5998), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [176341] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6484), 1,
      anon_sym_and,
    STATE(4128), 1,
      aux_sym_type_definition_repeat1,
    STATE(4155), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6120), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [176374] = 17,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6373), 1,
      sym__lowercase_identifier,
    ACTIONS(6375), 1,
      anon_sym_LPAREN,
    ACTIONS(6381), 1,
      anon_sym__,
    ACTIONS(6515), 1,
      anon_sym_nonrec,
    STATE(4156), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5467), 1,
      sym__type_params,
    STATE(5737), 1,
      sym_type_binding,
    STATE(6063), 1,
      sym__type_param,
    STATE(6196), 1,
      sym_extended_module_path,
    STATE(6383), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [176429] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6517), 1,
      anon_sym_DQUOTE,
    STATE(4157), 1,
      sym_attribute,
    STATE(4194), 1,
      aux_sym_string_content_repeat1,
    STATE(4247), 1,
      sym_escape_sequence,
    STATE(6592), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6470), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6472), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6468), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176470] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6519), 1,
      anon_sym_DQUOTE,
    STATE(4158), 1,
      sym_attribute,
    STATE(4194), 1,
      aux_sym_string_content_repeat1,
    STATE(4247), 1,
      sym_escape_sequence,
    STATE(6388), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6470), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6472), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6468), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176511] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1725), 1,
      anon_sym_PIPE,
    STATE(4159), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1727), 15,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [176542] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1769), 1,
      anon_sym_PIPE,
    STATE(4160), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1771), 15,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [176573] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1789), 1,
      anon_sym_PIPE,
    STATE(4161), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1791), 15,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [176604] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6521), 1,
      anon_sym_DQUOTE,
    STATE(4162), 1,
      sym_attribute,
    STATE(4194), 1,
      aux_sym_string_content_repeat1,
    STATE(4247), 1,
      sym_escape_sequence,
    STATE(6930), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6470), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6472), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6468), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176645] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1801), 1,
      anon_sym_PIPE,
    STATE(4163), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1803), 15,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [176676] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6523), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4164), 2,
      sym_attribute,
      aux_sym_class_definition_repeat1,
    ACTIONS(6056), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [176707] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6526), 1,
      anon_sym_DQUOTE,
    STATE(4165), 1,
      sym_attribute,
    STATE(4194), 1,
      aux_sym_string_content_repeat1,
    STATE(4247), 1,
      sym_escape_sequence,
    STATE(6550), 1,
      sym_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6470), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6472), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6468), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [176748] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6528), 1,
      anon_sym_and,
    STATE(4166), 1,
      sym_attribute,
    STATE(4168), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6095), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [176781] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6528), 1,
      anon_sym_and,
    STATE(4167), 1,
      sym_attribute,
    STATE(4168), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6032), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [176814] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6530), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4168), 2,
      sym_attribute,
      aux_sym_module_definition_repeat1,
    ACTIONS(6025), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [176845] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6528), 1,
      anon_sym_and,
    STATE(4169), 1,
      sym_attribute,
    STATE(4170), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6032), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [176878] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6528), 1,
      anon_sym_and,
    STATE(4168), 1,
      aux_sym_module_definition_repeat1,
    STATE(4170), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6223), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [176911] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6528), 1,
      anon_sym_and,
    STATE(4166), 1,
      aux_sym_module_definition_repeat1,
    STATE(4171), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6048), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [176944] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6528), 1,
      anon_sym_and,
    STATE(4167), 1,
      aux_sym_module_definition_repeat1,
    STATE(4172), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6095), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [176977] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6498), 1,
      anon_sym_LPAREN,
    STATE(4173), 1,
      sym_attribute,
    STATE(4538), 1,
      sym__simple_module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
    ACTIONS(4032), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [177016] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3861), 1,
      anon_sym_let,
    ACTIONS(3867), 1,
      anon_sym_object,
    ACTIONS(6533), 1,
      sym__lowercase_identifier,
    ACTIONS(6535), 1,
      anon_sym_LBRACK,
    STATE(4174), 1,
      sym_attribute,
    STATE(5127), 1,
      sym__simple_class_type,
    STATE(5975), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5534), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [177062] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6373), 1,
      sym__lowercase_identifier,
    ACTIONS(6375), 1,
      anon_sym_LPAREN,
    ACTIONS(6381), 1,
      anon_sym__,
    STATE(4175), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5467), 1,
      sym__type_params,
    STATE(6063), 1,
      sym__type_param,
    STATE(6196), 1,
      sym_extended_module_path,
    STATE(6320), 1,
      sym_type_binding,
    STATE(6383), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [177114] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6373), 1,
      sym__lowercase_identifier,
    ACTIONS(6375), 1,
      anon_sym_LPAREN,
    ACTIONS(6381), 1,
      anon_sym__,
    STATE(4176), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5467), 1,
      sym__type_params,
    STATE(5653), 1,
      sym_type_binding,
    STATE(6063), 1,
      sym__type_param,
    STATE(6196), 1,
      sym_extended_module_path,
    STATE(6383), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [177166] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4177), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3050), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3052), 13,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      sym__uppercase_identifier,
  [177196] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1325), 1,
      aux_sym_number_token1,
    ACTIONS(1327), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1329), 1,
      anon_sym_SQUOTE,
    ACTIONS(1331), 1,
      anon_sym_DQUOTE,
    ACTIONS(6539), 1,
      anon_sym_LBRACE,
    STATE(2037), 1,
      sym__signed_constant,
    STATE(4178), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6537), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(1823), 2,
      sym__constant,
      sym_signed_number,
    STATE(1803), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
  [177240] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3622), 1,
      sym__lowercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    ACTIONS(6541), 1,
      anon_sym_STAR,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(4179), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4131), 4,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4133), 5,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
  [177282] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3841), 1,
      anon_sym_object,
    ACTIONS(3905), 1,
      anon_sym_let,
    ACTIONS(6543), 1,
      sym__lowercase_identifier,
    ACTIONS(6545), 1,
      anon_sym_LBRACK,
    STATE(3468), 1,
      sym__simple_class_type,
    STATE(4180), 1,
      sym_attribute,
    STATE(5974), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3462), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [177328] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4181), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1955), 2,
      anon_sym_COLON2,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6547), 13,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_GT_RBRACE,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [177358] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6375), 1,
      anon_sym_LPAREN,
    ACTIONS(6381), 1,
      anon_sym__,
    ACTIONS(6387), 1,
      sym__lowercase_identifier,
    STATE(4131), 1,
      sym_type_binding,
    STATE(4182), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5436), 1,
      sym__type_params,
    STATE(6063), 1,
      sym__type_param,
    STATE(6196), 1,
      sym_extended_module_path,
    STATE(6407), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [177410] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3835), 1,
      anon_sym_let,
    ACTIONS(3841), 1,
      anon_sym_object,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6549), 1,
      sym__lowercase_identifier,
    ACTIONS(6551), 1,
      anon_sym_LBRACK,
    STATE(3621), 1,
      sym__simple_class_type,
    STATE(4183), 1,
      sym_attribute,
    STATE(6336), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3462), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [177456] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6375), 1,
      anon_sym_LPAREN,
    ACTIONS(6381), 1,
      anon_sym__,
    ACTIONS(6383), 1,
      sym__lowercase_identifier,
    STATE(3936), 1,
      sym_type_binding,
    STATE(4184), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5379), 1,
      sym__type_params,
    STATE(6063), 1,
      sym__type_param,
    STATE(6196), 1,
      sym_extended_module_path,
    STATE(6446), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [177508] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3622), 1,
      sym__lowercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    ACTIONS(6553), 1,
      anon_sym_DASH_GT,
    ACTIONS(6555), 1,
      anon_sym_STAR,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(4185), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3987), 3,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3989), 5,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
  [177552] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3835), 1,
      anon_sym_let,
    ACTIONS(3841), 1,
      anon_sym_object,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6549), 1,
      sym__lowercase_identifier,
    ACTIONS(6551), 1,
      anon_sym_LBRACK,
    STATE(3597), 1,
      sym__simple_class_type,
    STATE(4186), 1,
      sym_attribute,
    STATE(6336), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3462), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [177598] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6375), 1,
      anon_sym_LPAREN,
    ACTIONS(6381), 1,
      anon_sym__,
    ACTIONS(6387), 1,
      sym__lowercase_identifier,
    STATE(4045), 1,
      sym_type_binding,
    STATE(4187), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5436), 1,
      sym__type_params,
    STATE(6063), 1,
      sym__type_param,
    STATE(6196), 1,
      sym_extended_module_path,
    STATE(6407), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [177650] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3835), 1,
      anon_sym_let,
    ACTIONS(3841), 1,
      anon_sym_object,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6549), 1,
      sym__lowercase_identifier,
    ACTIONS(6551), 1,
      anon_sym_LBRACK,
    STATE(3759), 1,
      sym__simple_class_type,
    STATE(4188), 1,
      sym_attribute,
    STATE(6336), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3462), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [177696] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6375), 1,
      anon_sym_LPAREN,
    ACTIONS(6381), 1,
      anon_sym__,
    ACTIONS(6387), 1,
      sym__lowercase_identifier,
    STATE(4155), 1,
      sym_type_binding,
    STATE(4189), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5436), 1,
      sym__type_params,
    STATE(6063), 1,
      sym__type_param,
    STATE(6196), 1,
      sym_extended_module_path,
    STATE(6407), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [177748] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3835), 1,
      anon_sym_let,
    ACTIONS(3841), 1,
      anon_sym_object,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6549), 1,
      sym__lowercase_identifier,
    ACTIONS(6551), 1,
      anon_sym_LBRACK,
    STATE(3468), 1,
      sym__simple_class_type,
    STATE(4190), 1,
      sym_attribute,
    STATE(6336), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3462), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [177794] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3861), 1,
      anon_sym_let,
    ACTIONS(3867), 1,
      anon_sym_object,
    ACTIONS(6533), 1,
      sym__lowercase_identifier,
    ACTIONS(6535), 1,
      anon_sym_LBRACK,
    STATE(4191), 1,
      sym_attribute,
    STATE(5401), 1,
      sym__simple_class_type,
    STATE(5975), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5534), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [177840] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3861), 1,
      anon_sym_let,
    ACTIONS(3867), 1,
      anon_sym_object,
    ACTIONS(6533), 1,
      sym__lowercase_identifier,
    ACTIONS(6535), 1,
      anon_sym_LBRACK,
    STATE(4192), 1,
      sym_attribute,
    STATE(5203), 1,
      sym__simple_class_type,
    STATE(5975), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5534), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [177886] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6373), 1,
      sym__lowercase_identifier,
    ACTIONS(6375), 1,
      anon_sym_LPAREN,
    ACTIONS(6381), 1,
      anon_sym__,
    STATE(4193), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5467), 1,
      sym__type_params,
    STATE(5737), 1,
      sym_type_binding,
    STATE(6063), 1,
      sym__type_param,
    STATE(6196), 1,
      sym_extended_module_path,
    STATE(6383), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [177938] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6557), 1,
      anon_sym_DQUOTE,
    STATE(4194), 1,
      sym_attribute,
    STATE(4210), 1,
      aux_sym_string_content_repeat1,
    STATE(4247), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6470), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6472), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6468), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [177976] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3861), 1,
      anon_sym_let,
    ACTIONS(3867), 1,
      anon_sym_object,
    ACTIONS(6533), 1,
      sym__lowercase_identifier,
    ACTIONS(6535), 1,
      anon_sym_LBRACK,
    STATE(4195), 1,
      sym_attribute,
    STATE(5066), 1,
      sym__simple_class_type,
    STATE(5975), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5534), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [178022] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3841), 1,
      anon_sym_object,
    ACTIONS(6543), 1,
      sym__lowercase_identifier,
    ACTIONS(6545), 1,
      anon_sym_LBRACK,
    ACTIONS(6559), 1,
      anon_sym_let,
    STATE(3468), 1,
      sym__simple_class_type,
    STATE(4196), 1,
      sym_attribute,
    STATE(5974), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3462), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [178068] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3835), 1,
      anon_sym_let,
    ACTIONS(3841), 1,
      anon_sym_object,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6549), 1,
      sym__lowercase_identifier,
    ACTIONS(6551), 1,
      anon_sym_LBRACK,
    STATE(3741), 1,
      sym__simple_class_type,
    STATE(4197), 1,
      sym_attribute,
    STATE(6336), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3462), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [178114] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2192), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2194), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3841), 1,
      anon_sym_object,
    ACTIONS(6543), 1,
      sym__lowercase_identifier,
    ACTIONS(6545), 1,
      anon_sym_LBRACK,
    ACTIONS(6559), 1,
      anon_sym_let,
    STATE(4198), 1,
      sym_attribute,
    STATE(4445), 1,
      sym__simple_class_type,
    STATE(5974), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3462), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [178160] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1069), 1,
      aux_sym_number_token1,
    ACTIONS(1073), 1,
      anon_sym_SQUOTE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(2196), 1,
      aux_sym_signed_number_token1,
    ACTIONS(6563), 1,
      anon_sym_LBRACE,
    STATE(4199), 1,
      sym_attribute,
    STATE(4425), 1,
      sym__signed_constant,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6561), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4254), 2,
      sym__constant,
      sym_signed_number,
    STATE(4100), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
  [178204] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3861), 1,
      anon_sym_let,
    ACTIONS(3867), 1,
      anon_sym_object,
    ACTIONS(6533), 1,
      sym__lowercase_identifier,
    ACTIONS(6535), 1,
      anon_sym_LBRACK,
    STATE(4200), 1,
      sym_attribute,
    STATE(5222), 1,
      sym__simple_class_type,
    STATE(5975), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5534), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [178250] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4133), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5886), 1,
      anon_sym_STAR,
    STATE(4201), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4131), 13,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [178282] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6375), 1,
      anon_sym_LPAREN,
    ACTIONS(6381), 1,
      anon_sym__,
    ACTIONS(6383), 1,
      sym__lowercase_identifier,
    STATE(4045), 1,
      sym_type_binding,
    STATE(4202), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5379), 1,
      sym__type_params,
    STATE(6063), 1,
      sym__type_param,
    STATE(6196), 1,
      sym_extended_module_path,
    STATE(6446), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [178334] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1453), 1,
      aux_sym_number_token1,
    ACTIONS(1455), 1,
      aux_sym_signed_number_token1,
    ACTIONS(1457), 1,
      anon_sym_SQUOTE,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(6567), 1,
      anon_sym_LBRACE,
    STATE(3053), 1,
      sym__signed_constant,
    STATE(4203), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6565), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(2940), 2,
      sym__constant,
      sym_signed_number,
    STATE(2929), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
  [178378] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2234), 1,
      aux_sym_number_token1,
    ACTIONS(2236), 1,
      aux_sym_signed_number_token1,
    ACTIONS(2238), 1,
      anon_sym_SQUOTE,
    ACTIONS(2240), 1,
      anon_sym_DQUOTE,
    ACTIONS(6571), 1,
      anon_sym_LBRACE,
    STATE(4204), 1,
      sym_attribute,
    STATE(4828), 1,
      sym__signed_constant,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6569), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4717), 2,
      sym__constant,
      sym_signed_number,
    STATE(4693), 5,
      sym_number,
      sym_character,
      sym_string,
      sym_quoted_string,
      sym_boolean,
  [178422] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5898), 1,
      anon_sym_STAR,
    STATE(4205), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4131), 14,
      ts_builtin_sym_end,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_AMP,
  [178452] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3835), 1,
      anon_sym_let,
    ACTIONS(3841), 1,
      anon_sym_object,
    ACTIONS(4978), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(4980), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6549), 1,
      sym__lowercase_identifier,
    ACTIONS(6551), 1,
      anon_sym_LBRACK,
    STATE(3691), 1,
      sym__simple_class_type,
    STATE(4206), 1,
      sym_attribute,
    STATE(6336), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3462), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [178498] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6573), 1,
      anon_sym_SEMI_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4207), 2,
      sym_attribute,
      aux_sym__structure_repeat1,
    ACTIONS(2812), 13,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [178528] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(3861), 1,
      anon_sym_let,
    ACTIONS(3867), 1,
      anon_sym_object,
    ACTIONS(6533), 1,
      sym__lowercase_identifier,
    ACTIONS(6535), 1,
      anon_sym_LBRACK,
    STATE(4208), 1,
      sym_attribute,
    STATE(5259), 1,
      sym__simple_class_type,
    STATE(5975), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5534), 6,
      sym_instantiated_class_type,
      sym_class_body_type,
      sym_let_open_class_type,
      sym_extension,
      sym_quoted_extension,
      sym_class_type_path,
  [178574] = 16,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(6375), 1,
      anon_sym_LPAREN,
    ACTIONS(6381), 1,
      anon_sym__,
    ACTIONS(6383), 1,
      sym__lowercase_identifier,
    STATE(3857), 1,
      sym_type_binding,
    STATE(4209), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5379), 1,
      sym__type_params,
    STATE(6063), 1,
      sym__type_param,
    STATE(6196), 1,
      sym_extended_module_path,
    STATE(6446), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [178626] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6576), 1,
      anon_sym_DQUOTE,
    STATE(4247), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6581), 2,
      sym__null,
      aux_sym_string_content_token2,
    STATE(4210), 2,
      sym_attribute,
      aux_sym_string_content_repeat1,
    ACTIONS(6584), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
    ACTIONS(6578), 6,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [178662] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4211), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3078), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3080), 13,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      anon_sym_COLON_COLON,
      sym__uppercase_identifier,
  [178692] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4212), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6587), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [178719] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3622), 1,
      sym__lowercase_identifier,
    ACTIONS(6375), 1,
      anon_sym_LPAREN,
    ACTIONS(6381), 1,
      anon_sym__,
    STATE(4213), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5335), 1,
      sym__type_params,
    STATE(6031), 1,
      sym_type_constructor_path,
    STATE(6063), 1,
      sym__type_param,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [178768] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(6591), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4214), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6589), 12,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [178799] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6595), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4215), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6593), 13,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [178828] = 14,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3192), 1,
      anon_sym_PIPE,
    ACTIONS(6597), 1,
      anon_sym_LPAREN,
    ACTIONS(6599), 1,
      anon_sym_private,
    ACTIONS(6601), 1,
      anon_sym_DOT_DOT,
    ACTIONS(6603), 1,
      anon_sym_LBRACK,
    ACTIONS(6607), 1,
      anon_sym_LBRACE,
    ACTIONS(6609), 1,
      sym__uppercase_identifier,
    STATE(4216), 1,
      sym_attribute,
    STATE(5004), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6605), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4628), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4657), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [178875] = 14,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3231), 1,
      anon_sym_PIPE,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    ACTIONS(6611), 1,
      anon_sym_private,
    ACTIONS(6613), 1,
      anon_sym_DOT_DOT,
    ACTIONS(6615), 1,
      anon_sym_LBRACE,
    STATE(3870), 1,
      sym_constructor_declaration,
    STATE(4217), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3600), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3681), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [178922] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6498), 1,
      anon_sym_LPAREN,
    STATE(4218), 1,
      sym_attribute,
    STATE(4538), 1,
      sym__simple_module_expression,
    STATE(5123), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4257), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [178965] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6498), 1,
      anon_sym_LPAREN,
    STATE(4219), 1,
      sym_attribute,
    STATE(4538), 1,
      sym__simple_module_expression,
    STATE(5124), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4257), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [179008] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6498), 1,
      anon_sym_LPAREN,
    STATE(4220), 1,
      sym_attribute,
    STATE(4538), 1,
      sym__simple_module_expression,
    STATE(5125), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4263), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [179051] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3622), 1,
      sym__lowercase_identifier,
    ACTIONS(6375), 1,
      anon_sym_LPAREN,
    ACTIONS(6381), 1,
      anon_sym__,
    STATE(4221), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5415), 1,
      sym__type_params,
    STATE(6063), 1,
      sym__type_param,
    STATE(6196), 1,
      sym_extended_module_path,
    STATE(6277), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [179100] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4222), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4196), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179127] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6498), 1,
      anon_sym_LPAREN,
    STATE(4223), 1,
      sym_attribute,
    STATE(4538), 1,
      sym__simple_module_expression,
    STATE(5061), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4155), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [179170] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6619), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4224), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6617), 13,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179199] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4225), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2822), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179226] = 14,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3192), 1,
      anon_sym_PIPE,
    ACTIONS(6597), 1,
      anon_sym_LPAREN,
    ACTIONS(6603), 1,
      anon_sym_LBRACK,
    ACTIONS(6607), 1,
      anon_sym_LBRACE,
    ACTIONS(6609), 1,
      sym__uppercase_identifier,
    ACTIONS(6621), 1,
      anon_sym_private,
    ACTIONS(6623), 1,
      anon_sym_DOT_DOT,
    STATE(4226), 1,
      sym_attribute,
    STATE(5004), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6605), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4628), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4669), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [179273] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6498), 1,
      anon_sym_LPAREN,
    STATE(4227), 1,
      sym_attribute,
    STATE(4538), 1,
      sym__simple_module_expression,
    STATE(5176), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4251), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [179316] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2230), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(2232), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6498), 1,
      anon_sym_LPAREN,
    STATE(4228), 1,
      sym_attribute,
    STATE(4538), 1,
      sym__simple_module_expression,
    STATE(5177), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4251), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [179359] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5777), 1,
      anon_sym_DASH_GT,
    STATE(4229), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3987), 13,
      ts_builtin_sym_end,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_AMP,
  [179388] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6015), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6500), 1,
      anon_sym_as,
    STATE(4230), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6013), 12,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179419] = 14,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3287), 1,
      anon_sym_PIPE,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    ACTIONS(6615), 1,
      anon_sym_LBRACE,
    ACTIONS(6625), 1,
      anon_sym_private,
    ACTIONS(6627), 1,
      anon_sym_DOT_DOT,
    STATE(3549), 1,
      sym_constructor_declaration,
    STATE(4231), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3440), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3464), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [179466] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3622), 1,
      sym__lowercase_identifier,
    ACTIONS(6375), 1,
      anon_sym_LPAREN,
    ACTIONS(6381), 1,
      anon_sym__,
    STATE(4232), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5305), 1,
      sym__type_params,
    STATE(6063), 1,
      sym__type_param,
    STATE(6196), 1,
      sym_extended_module_path,
    STATE(6333), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [179515] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4006), 1,
      anon_sym_COLON2,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4233), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4004), 4,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [179554] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6019), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6500), 1,
      anon_sym_as,
    STATE(4234), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6017), 12,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179585] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1809), 1,
      anon_sym_COLON2,
    STATE(4235), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1811), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
  [179614] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(3989), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5882), 1,
      anon_sym_DASH_GT,
    STATE(4236), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3987), 12,
      anon_sym_EQ,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_as,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179645] = 14,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3231), 1,
      anon_sym_PIPE,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    ACTIONS(6615), 1,
      anon_sym_LBRACE,
    ACTIONS(6629), 1,
      anon_sym_private,
    ACTIONS(6631), 1,
      anon_sym_DOT_DOT,
    STATE(3870), 1,
      sym_constructor_declaration,
    STATE(4237), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3600), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3673), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [179692] = 15,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(3622), 1,
      sym__lowercase_identifier,
    ACTIONS(6375), 1,
      anon_sym_LPAREN,
    ACTIONS(6381), 1,
      anon_sym__,
    STATE(4238), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5472), 1,
      sym__type_params,
    STATE(6063), 1,
      sym__type_param,
    STATE(6165), 1,
      sym_type_constructor_path,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [179741] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6040), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6500), 1,
      anon_sym_as,
    STATE(4239), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6038), 12,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_COLON_GT,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179772] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4240), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6633), 14,
      anon_sym_SEMI_SEMI,
      anon_sym_external,
      anon_sym_type,
      anon_sym_RBRACK,
      anon_sym_exception,
      anon_sym_module,
      anon_sym_open,
      anon_sym_include,
      anon_sym_class,
      anon_sym_val,
      anon_sym_end,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179799] = 14,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3287), 1,
      anon_sym_PIPE,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    ACTIONS(6615), 1,
      anon_sym_LBRACE,
    ACTIONS(6635), 1,
      anon_sym_private,
    ACTIONS(6637), 1,
      anon_sym_DOT_DOT,
    STATE(3549), 1,
      sym_constructor_declaration,
    STATE(4241), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3440), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3448), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [179846] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6639), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4242), 1,
      sym_attribute,
    STATE(4347), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6641), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [179880] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1105), 1,
      anon_sym_LBRACK,
    ACTIONS(1115), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6643), 1,
      anon_sym_LPAREN,
    ACTIONS(6645), 1,
      anon_sym_LBRACE,
    ACTIONS(6647), 1,
      anon_sym_begin,
    ACTIONS(6649), 1,
      sym__uppercase_identifier,
    STATE(4243), 1,
      sym_attribute,
    STATE(4329), 1,
      sym__delimited_pattern,
    STATE(4493), 1,
      sym__constructor_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4414), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [179922] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6498), 1,
      anon_sym_LPAREN,
    STATE(4244), 1,
      sym_attribute,
    STATE(4538), 1,
      sym__simple_module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4032), 4,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [179958] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4245), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3991), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [179984] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3192), 1,
      anon_sym_PIPE,
    ACTIONS(6597), 1,
      anon_sym_LPAREN,
    ACTIONS(6603), 1,
      anon_sym_LBRACK,
    ACTIONS(6607), 1,
      anon_sym_LBRACE,
    ACTIONS(6609), 1,
      sym__uppercase_identifier,
    ACTIONS(6651), 1,
      anon_sym_DOT_DOT,
    STATE(4246), 1,
      sym_attribute,
    STATE(5004), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6605), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4628), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4665), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [180028] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4247), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6655), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6653), 11,
      anon_sym_DQUOTE,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [180056] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3231), 1,
      anon_sym_PIPE,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    ACTIONS(6615), 1,
      anon_sym_LBRACE,
    ACTIONS(6657), 1,
      anon_sym_DOT_DOT,
    STATE(3870), 1,
      sym_constructor_declaration,
    STATE(4248), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3600), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3678), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [180100] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(6661), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4249), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6659), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [180130] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4133), 1,
      anon_sym_as,
    ACTIONS(6663), 1,
      sym__lowercase_identifier,
    ACTIONS(6665), 1,
      anon_sym_STAR,
    ACTIONS(6667), 1,
      anon_sym_POUND,
    STATE(4250), 1,
      sym_attribute,
    STATE(4480), 1,
      sym_type_constructor_path,
    STATE(6026), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4131), 6,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
  [180170] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4251), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3995), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [180196] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1105), 1,
      anon_sym_LBRACK,
    ACTIONS(1115), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6645), 1,
      anon_sym_LBRACE,
    ACTIONS(6647), 1,
      anon_sym_begin,
    ACTIONS(6669), 1,
      anon_sym_LPAREN,
    ACTIONS(6671), 1,
      sym__uppercase_identifier,
    STATE(1998), 1,
      sym__constructor_name,
    STATE(4252), 1,
      sym_attribute,
    STATE(4329), 1,
      sym__delimited_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4414), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [180238] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6675), 1,
      anon_sym_as,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4253), 1,
      sym_attribute,
    STATE(4381), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6673), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [180272] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2930), 1,
      anon_sym_COLON2,
    STATE(4254), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2932), 12,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [180300] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6671), 1,
      sym__uppercase_identifier,
    ACTIONS(6677), 1,
      anon_sym_LPAREN,
    ACTIONS(6679), 1,
      anon_sym_LBRACE,
    ACTIONS(6681), 1,
      anon_sym_begin,
    STATE(1998), 1,
      sym__constructor_name,
    STATE(2250), 1,
      sym__delimited_pattern,
    STATE(4255), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [180342] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4350), 1,
      anon_sym_in,
    ACTIONS(6498), 1,
      anon_sym_LPAREN,
    STATE(4256), 1,
      sym_attribute,
    STATE(4538), 1,
      sym__simple_module_expression,
    STATE(5304), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [180384] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2212), 1,
      anon_sym_LBRACK,
    ACTIONS(2222), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6683), 1,
      anon_sym_LPAREN,
    ACTIONS(6685), 1,
      anon_sym_LBRACE,
    ACTIONS(6687), 1,
      anon_sym_begin,
    ACTIONS(6689), 1,
      sym__uppercase_identifier,
    STATE(4257), 1,
      sym_attribute,
    STATE(4894), 1,
      sym__delimited_pattern,
    STATE(5039), 1,
      sym__constructor_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4978), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [180426] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4258), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3953), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [180452] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6691), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4259), 1,
      sym_attribute,
    STATE(4353), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6693), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [180486] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4260), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3949), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [180512] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6695), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4261), 1,
      sym_attribute,
    STATE(4357), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6697), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [180546] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4280), 1,
      anon_sym_in,
    ACTIONS(6498), 1,
      anon_sym_LPAREN,
    STATE(4262), 1,
      sym_attribute,
    STATE(4538), 1,
      sym__simple_module_expression,
    STATE(5396), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [180588] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4263), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3640), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3638), 11,
      anon_sym_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
      aux_sym_type_variable_token1,
  [180616] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6699), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4264), 1,
      sym_attribute,
    STATE(4350), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6701), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [180650] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4265), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3983), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [180676] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      anon_sym_COLON2,
    STATE(4266), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2944), 12,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [180704] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6703), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4267), 1,
      sym_attribute,
    STATE(4395), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6705), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [180738] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6361), 1,
      anon_sym_COLON_GT,
    ACTIONS(6709), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(4215), 1,
      sym__coerced,
    STATE(4268), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6707), 10,
      anon_sym_EQ,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [180770] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1047), 1,
      anon_sym_LBRACK,
    ACTIONS(1061), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6647), 1,
      anon_sym_begin,
    ACTIONS(6649), 1,
      sym__uppercase_identifier,
    ACTIONS(6711), 1,
      anon_sym_LPAREN,
    ACTIONS(6713), 1,
      anon_sym_LBRACE,
    STATE(4269), 1,
      sym_attribute,
    STATE(4493), 1,
      sym__constructor_name,
    STATE(4575), 1,
      sym__delimited_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4564), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [180812] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1485), 1,
      anon_sym_LBRACK,
    ACTIONS(1499), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6679), 1,
      anon_sym_LBRACE,
    ACTIONS(6681), 1,
      anon_sym_begin,
    ACTIONS(6715), 1,
      anon_sym_LPAREN,
    ACTIONS(6717), 1,
      sym__uppercase_identifier,
    STATE(2250), 1,
      sym__delimited_pattern,
    STATE(2599), 1,
      sym__constructor_name,
    STATE(4270), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2287), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [180854] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6719), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4271), 1,
      sym_attribute,
    STATE(4373), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6721), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [180888] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3630), 1,
      anon_sym_as,
    ACTIONS(6663), 1,
      sym__lowercase_identifier,
    ACTIONS(6667), 1,
      anon_sym_POUND,
    STATE(4272), 1,
      sym_attribute,
    STATE(4480), 1,
      sym_type_constructor_path,
    STATE(6026), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3628), 7,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
  [180926] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4273), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4032), 4,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [180962] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4274), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3961), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [180988] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3192), 1,
      anon_sym_PIPE,
    ACTIONS(6597), 1,
      anon_sym_LPAREN,
    ACTIONS(6603), 1,
      anon_sym_LBRACK,
    ACTIONS(6607), 1,
      anon_sym_LBRACE,
    ACTIONS(6609), 1,
      sym__uppercase_identifier,
    ACTIONS(6723), 1,
      anon_sym_DOT_DOT,
    STATE(4275), 1,
      sym_attribute,
    STATE(5004), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6605), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4628), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(4676), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [181032] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4276), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6727), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(6725), 11,
      anon_sym_DQUOTE,
      aux_sym_string_content_token1,
      aux_sym_string_content_token3,
      aux_sym_string_content_token4,
      aux_sym_string_content_token5,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [181060] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1105), 1,
      anon_sym_LBRACK,
    ACTIONS(1115), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6645), 1,
      anon_sym_LBRACE,
    ACTIONS(6647), 1,
      anon_sym_begin,
    ACTIONS(6729), 1,
      anon_sym_LPAREN,
    ACTIONS(6731), 1,
      sym__uppercase_identifier,
    STATE(3119), 1,
      sym__constructor_name,
    STATE(4277), 1,
      sym_attribute,
    STATE(4329), 1,
      sym__delimited_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4414), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [181102] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(6498), 1,
      anon_sym_LPAREN,
    STATE(4278), 1,
      sym_attribute,
    STATE(4538), 1,
      sym__simple_module_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4004), 4,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [181138] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1047), 1,
      anon_sym_LBRACK,
    ACTIONS(1061), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6647), 1,
      anon_sym_begin,
    ACTIONS(6713), 1,
      anon_sym_LBRACE,
    ACTIONS(6731), 1,
      sym__uppercase_identifier,
    ACTIONS(6733), 1,
      anon_sym_LPAREN,
    STATE(3119), 1,
      sym__constructor_name,
    STATE(4279), 1,
      sym_attribute,
    STATE(4575), 1,
      sym__delimited_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4564), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [181180] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1047), 1,
      anon_sym_LBRACK,
    ACTIONS(1061), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6085), 1,
      sym__uppercase_identifier,
    ACTIONS(6647), 1,
      anon_sym_begin,
    ACTIONS(6713), 1,
      anon_sym_LBRACE,
    ACTIONS(6735), 1,
      anon_sym_LPAREN,
    STATE(1753), 1,
      sym__constructor_name,
    STATE(4280), 1,
      sym_attribute,
    STATE(4575), 1,
      sym__delimited_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4564), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [181222] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1105), 1,
      anon_sym_LBRACK,
    ACTIONS(1115), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6085), 1,
      sym__uppercase_identifier,
    ACTIONS(6645), 1,
      anon_sym_LBRACE,
    ACTIONS(6647), 1,
      anon_sym_begin,
    ACTIONS(6737), 1,
      anon_sym_LPAREN,
    STATE(1753), 1,
      sym__constructor_name,
    STATE(4281), 1,
      sym_attribute,
    STATE(4329), 1,
      sym__delimited_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4414), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [181264] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2250), 1,
      anon_sym_LBRACK,
    ACTIONS(2256), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6687), 1,
      anon_sym_begin,
    ACTIONS(6739), 1,
      anon_sym_LPAREN,
    ACTIONS(6741), 1,
      anon_sym_LBRACE,
    ACTIONS(6743), 1,
      sym__uppercase_identifier,
    STATE(3265), 1,
      sym__constructor_name,
    STATE(4282), 1,
      sym_attribute,
    STATE(4831), 1,
      sym__delimited_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4922), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [181306] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2212), 1,
      anon_sym_LBRACK,
    ACTIONS(2222), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6685), 1,
      anon_sym_LBRACE,
    ACTIONS(6687), 1,
      anon_sym_begin,
    ACTIONS(6743), 1,
      sym__uppercase_identifier,
    ACTIONS(6745), 1,
      anon_sym_LPAREN,
    STATE(3265), 1,
      sym__constructor_name,
    STATE(4283), 1,
      sym_attribute,
    STATE(4894), 1,
      sym__delimited_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4978), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [181348] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4302), 1,
      anon_sym_in,
    ACTIONS(6498), 1,
      anon_sym_LPAREN,
    STATE(4284), 1,
      sym_attribute,
    STATE(4538), 1,
      sym__simple_module_expression,
    STATE(5394), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4500), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [181390] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2081), 1,
      anon_sym_COLON2,
    ACTIONS(6747), 1,
      anon_sym_DOT_DOT,
    STATE(4285), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2075), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [181420] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1047), 1,
      anon_sym_LBRACK,
    ACTIONS(1061), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6647), 1,
      anon_sym_begin,
    ACTIONS(6671), 1,
      sym__uppercase_identifier,
    ACTIONS(6713), 1,
      anon_sym_LBRACE,
    ACTIONS(6749), 1,
      anon_sym_LPAREN,
    STATE(1998), 1,
      sym__constructor_name,
    STATE(4286), 1,
      sym_attribute,
    STATE(4575), 1,
      sym__delimited_binding_pattern,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4564), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [181462] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6753), 1,
      anon_sym_as,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4287), 1,
      sym_attribute,
    STATE(4392), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6751), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [181496] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3287), 1,
      anon_sym_PIPE,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    ACTIONS(6615), 1,
      anon_sym_LBRACE,
    ACTIONS(6755), 1,
      anon_sym_DOT_DOT,
    STATE(3549), 1,
      sym_constructor_declaration,
    STATE(4288), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3440), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3455), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [181540] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1419), 1,
      anon_sym_LBRACK,
    ACTIONS(1435), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6757), 1,
      anon_sym_LPAREN,
    ACTIONS(6759), 1,
      anon_sym_LBRACE,
    ACTIONS(6761), 1,
      anon_sym_begin,
    ACTIONS(6763), 1,
      sym__uppercase_identifier,
    STATE(2769), 1,
      sym__constructor_name,
    STATE(3054), 1,
      sym__delimited_binding_pattern,
    STATE(4289), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3051), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [181582] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6765), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4290), 1,
      sym_attribute,
    STATE(4388), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6767), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [181616] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3231), 1,
      anon_sym_PIPE,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    ACTIONS(6615), 1,
      anon_sym_LBRACE,
    ACTIONS(6769), 1,
      anon_sym_DOT_DOT,
    STATE(3870), 1,
      sym_constructor_declaration,
    STATE(4291), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3600), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3683), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [181660] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1987), 1,
      anon_sym_LBRACK,
    ACTIONS(1995), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6771), 1,
      anon_sym_LPAREN,
    ACTIONS(6773), 1,
      anon_sym_LBRACE,
    ACTIONS(6775), 1,
      anon_sym_begin,
    ACTIONS(6777), 1,
      sym__uppercase_identifier,
    STATE(2449), 1,
      sym__delimited_pattern,
    STATE(2486), 1,
      sym__constructor_name,
    STATE(4292), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2428), 5,
      sym_record_pattern,
      sym_list_pattern,
      sym_array_pattern,
      sym_parenthesized_pattern,
      sym_unit,
  [181702] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3624), 1,
      anon_sym_as,
    ACTIONS(6663), 1,
      sym__lowercase_identifier,
    ACTIONS(6779), 1,
      anon_sym_POUND,
    STATE(4293), 1,
      sym_attribute,
    STATE(4439), 1,
      sym_type_constructor_path,
    STATE(6026), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3620), 7,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
  [181740] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6500), 1,
      anon_sym_as,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4294), 1,
      sym_attribute,
    STATE(4363), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6781), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [181774] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1303), 1,
      anon_sym_LBRACK,
    ACTIONS(1313), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6783), 1,
      anon_sym_LPAREN,
    ACTIONS(6785), 1,
      anon_sym_LBRACE,
    ACTIONS(6787), 1,
      anon_sym_begin,
    ACTIONS(6789), 1,
      sym__uppercase_identifier,
    STATE(1762), 1,
      sym__constructor_name,
    STATE(2043), 1,
      sym__delimited_binding_pattern,
    STATE(4295), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2087), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [181816] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6791), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4296), 1,
      sym_attribute,
    STATE(4389), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6793), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [181850] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3989), 1,
      anon_sym_as,
    ACTIONS(6663), 1,
      sym__lowercase_identifier,
    ACTIONS(6667), 1,
      anon_sym_POUND,
    ACTIONS(6795), 1,
      anon_sym_DASH_GT,
    ACTIONS(6797), 1,
      anon_sym_STAR,
    STATE(4297), 1,
      sym_attribute,
    STATE(4480), 1,
      sym_type_constructor_path,
    STATE(6026), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3987), 5,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
  [181892] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3287), 1,
      anon_sym_PIPE,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    ACTIONS(6615), 1,
      anon_sym_LBRACE,
    ACTIONS(6799), 1,
      anon_sym_DOT_DOT,
    STATE(3549), 1,
      sym_constructor_declaration,
    STATE(4298), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3440), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    STATE(3459), 2,
      sym_variant_declaration,
      sym_record_declaration,
  [181936] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2250), 1,
      anon_sym_LBRACK,
    ACTIONS(2256), 1,
      anon_sym_LBRACK_PIPE,
    ACTIONS(6687), 1,
      anon_sym_begin,
    ACTIONS(6689), 1,
      sym__uppercase_identifier,
    ACTIONS(6741), 1,
      anon_sym_LBRACE,
    ACTIONS(6801), 1,
      anon_sym_LPAREN,
    STATE(4299), 1,
      sym_attribute,
    STATE(4831), 1,
      sym__delimited_binding_pattern,
    STATE(5039), 1,
      sym__constructor_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4922), 5,
      sym_record_binding_pattern,
      sym_list_binding_pattern,
      sym_array_binding_pattern,
      sym_parenthesized_binding_pattern,
      sym_unit,
  [181978] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4300), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3987), 13,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_COLON_GT,
      anon_sym_in,
      anon_sym_as,
      anon_sym_GT,
      anon_sym_LBRACK_AT_AT,
  [182004] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6803), 1,
      anon_sym_EQ,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4301), 1,
      sym_attribute,
    STATE(4325), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6805), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182038] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6807), 1,
      anon_sym_RPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4302), 1,
      sym_attribute,
    STATE(6819), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [182077] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6597), 1,
      anon_sym_LPAREN,
    ACTIONS(6603), 1,
      anon_sym_LBRACK,
    ACTIONS(6609), 1,
      sym__uppercase_identifier,
    STATE(4303), 1,
      sym_attribute,
    STATE(4909), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6605), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4628), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
    ACTIONS(4380), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [182114] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(6811), 1,
      anon_sym_RPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4304), 1,
      sym_attribute,
    STATE(6637), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [182153] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4305), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6813), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182184] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(6815), 1,
      anon_sym_RPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4306), 1,
      sym_attribute,
    STATE(6671), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [182223] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3482), 1,
      anon_sym_COLON2,
    STATE(4307), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3484), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182250] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3563), 1,
      anon_sym_COLON2,
    STATE(4308), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3565), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182277] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6819), 1,
      anon_sym_COLON2,
    STATE(4309), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6817), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182304] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4310), 1,
      sym_attribute,
    STATE(4411), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6821), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182335] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4311), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6823), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182366] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6827), 1,
      anon_sym_LPAREN,
    ACTIONS(6829), 1,
      anon_sym_COLON2,
    STATE(4312), 1,
      sym_attribute,
    STATE(4607), 1,
      aux_sym_module_binding_repeat1,
    STATE(4683), 1,
      sym__module_typed,
    STATE(4760), 1,
      sym_module_parameter,
    STATE(5136), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4455), 2,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(6825), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [182409] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(6831), 1,
      anon_sym_RPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4313), 1,
      sym_attribute,
    STATE(7243), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [182448] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4314), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6833), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182479] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4315), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6835), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182510] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6839), 1,
      anon_sym_COLON2,
    STATE(4316), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6837), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182537] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4317), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6841), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182568] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(6843), 1,
      anon_sym_RPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4318), 1,
      sym_attribute,
    STATE(7044), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [182607] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4319), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5793), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182638] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4320), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6371), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182669] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(6845), 1,
      anon_sym_RPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4321), 1,
      sym_attribute,
    STATE(6933), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [182708] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6849), 1,
      anon_sym_COLON2,
    STATE(4322), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6847), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182735] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6853), 1,
      anon_sym_COLON2,
    STATE(4323), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6851), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182762] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6857), 1,
      anon_sym_COLON2,
    STATE(4324), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6855), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182789] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4325), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6859), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182820] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(6861), 1,
      anon_sym_RPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4326), 1,
      sym_attribute,
    STATE(6965), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [182859] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4327), 1,
      sym_attribute,
    STATE(4437), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6863), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182890] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4305), 1,
      aux_sym_expression_item_repeat1,
    STATE(4328), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6865), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [182921] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3522), 1,
      anon_sym_COLON2,
    STATE(4329), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3524), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182948] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6869), 1,
      anon_sym_COLON2,
    STATE(4330), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6867), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [182975] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4331), 1,
      sym_attribute,
    STATE(4382), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6871), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183006] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4332), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6359), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183037] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3036), 1,
      anon_sym_COLON2,
    STATE(4333), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3038), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [183064] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6873), 1,
      anon_sym_COLON2,
    STATE(4334), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3050), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3052), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [183093] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4335), 1,
      sym_attribute,
    STATE(4398), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6875), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183124] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3468), 1,
      anon_sym_COLON2,
    STATE(4336), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3470), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [183151] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3054), 1,
      anon_sym_COLON2,
    STATE(4337), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3056), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [183178] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4338), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6877), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183209] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(6879), 1,
      anon_sym_RPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4339), 1,
      sym_attribute,
    STATE(6547), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [183248] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3478), 1,
      anon_sym_COLON2,
    STATE(4340), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3480), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [183275] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6883), 1,
      anon_sym_COLON2,
    STATE(4341), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6881), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [183302] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(6885), 1,
      anon_sym_RPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4342), 1,
      sym_attribute,
    STATE(7072), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [183341] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(6887), 1,
      anon_sym_RPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4343), 1,
      sym_attribute,
    STATE(7098), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [183380] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4344), 1,
      sym_attribute,
    STATE(4419), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6889), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183411] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3474), 1,
      anon_sym_COLON2,
    STATE(4345), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3476), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [183438] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(6891), 1,
      anon_sym_RPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4346), 1,
      sym_attribute,
    STATE(6492), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [183477] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4347), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6893), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183508] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4348), 1,
      sym_attribute,
    STATE(4370), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6895), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183539] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4349), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6897), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183570] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4350), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6899), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183601] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6903), 1,
      anon_sym_COLON2,
    STATE(4351), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6901), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [183628] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4352), 1,
      sym_attribute,
    STATE(4380), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6905), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183659] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4353), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6907), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183690] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3660), 1,
      anon_sym_COLON2,
    STATE(4354), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3050), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3052), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [183719] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6911), 1,
      anon_sym_COLON2,
    STATE(4355), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6909), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [183746] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(6913), 1,
      anon_sym_RPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4356), 1,
      sym_attribute,
    STATE(6351), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [183785] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4357), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6915), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183816] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(6917), 1,
      anon_sym_RPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4358), 1,
      sym_attribute,
    STATE(6776), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [183855] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4359), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6919), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183886] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4360), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5781), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183917] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4319), 1,
      aux_sym_expression_item_repeat1,
    STATE(4361), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5785), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [183948] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(6921), 1,
      anon_sym_RPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4362), 1,
      sym_attribute,
    STATE(7275), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [183987] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4363), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6923), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184018] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4359), 1,
      aux_sym_expression_item_repeat1,
    STATE(4364), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6925), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184049] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4315), 1,
      aux_sym_expression_item_repeat1,
    STATE(4365), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6927), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184080] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(6929), 1,
      anon_sym_RPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4366), 1,
      sym_attribute,
    STATE(7305), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [184119] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3014), 1,
      anon_sym_COLON2,
    STATE(4367), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3016), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [184146] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3018), 1,
      anon_sym_COLON2,
    STATE(4368), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3020), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [184173] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(6931), 1,
      anon_sym_RPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4369), 1,
      sym_attribute,
    STATE(6391), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [184212] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4370), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6933), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184243] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4317), 1,
      aux_sym_expression_item_repeat1,
    STATE(4371), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6935), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184274] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4372), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6937), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184305] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4373), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6939), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184336] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6943), 1,
      anon_sym_COLON2,
    STATE(4374), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6941), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [184363] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3460), 1,
      anon_sym_COLON2,
    STATE(4375), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3462), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [184390] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4376), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5785), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184421] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(6827), 1,
      anon_sym_LPAREN,
    ACTIONS(6829), 1,
      anon_sym_COLON2,
    STATE(4312), 1,
      aux_sym_module_binding_repeat1,
    STATE(4377), 1,
      sym_attribute,
    STATE(4735), 1,
      sym__module_typed,
    STATE(4760), 1,
      sym_module_parameter,
    STATE(5086), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4406), 2,
      anon_sym_and,
      anon_sym_in,
    ACTIONS(6945), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [184464] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4378), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2402), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184495] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4379), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6947), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184526] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4380), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6949), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184557] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4381), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6951), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184588] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4382), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6953), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184619] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(6955), 1,
      anon_sym_RPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4383), 1,
      sym_attribute,
    STATE(7128), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [184658] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4384), 1,
      sym_attribute,
    STATE(4423), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5715), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184689] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4372), 1,
      aux_sym_expression_item_repeat1,
    STATE(4385), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6957), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184720] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4386), 1,
      sym_attribute,
    STATE(4396), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6959), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184751] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3490), 1,
      anon_sym_COLON2,
    STATE(4387), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3492), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [184778] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4388), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6961), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184809] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4389), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6963), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184840] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4311), 1,
      aux_sym_expression_item_repeat1,
    STATE(4390), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6965), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184871] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4391), 1,
      sym_attribute,
    STATE(4434), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6967), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184902] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4392), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6969), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [184933] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(6971), 1,
      anon_sym_RPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4393), 1,
      sym_attribute,
    STATE(6798), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [184972] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3622), 1,
      sym__lowercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    ACTIONS(4348), 1,
      anon_sym_STAR,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(4394), 1,
      sym_attribute,
    STATE(4651), 1,
      aux_sym__constructor_argument_repeat1,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4346), 2,
      anon_sym_PIPE,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4424), 3,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
  [185013] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4395), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6973), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185044] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4396), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6975), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185075] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(6977), 1,
      anon_sym_RPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4397), 1,
      sym_attribute,
    STATE(7346), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [185114] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4398), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6979), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185145] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4338), 1,
      aux_sym_expression_item_repeat1,
    STATE(4399), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6981), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185176] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4400), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6983), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185207] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4401), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6985), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185238] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4402), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6987), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185269] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3106), 1,
      anon_sym_COLON2,
    STATE(4403), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3108), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [185296] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4314), 1,
      aux_sym_expression_item_repeat1,
    STATE(4404), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6989), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185327] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6993), 1,
      anon_sym_COLON2,
    STATE(4405), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6991), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [185354] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4401), 1,
      aux_sym_expression_item_repeat1,
    STATE(4406), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6995), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185385] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3494), 1,
      anon_sym_COLON2,
    STATE(4407), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3496), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [185412] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4408), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6997), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185443] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1703), 1,
      anon_sym_DOT,
    STATE(4409), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1701), 11,
      anon_sym_and,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
  [185470] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4410), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2380), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185501] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4411), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6999), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185532] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3536), 1,
      anon_sym_COLON2,
    STATE(4412), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3538), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [185559] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4413), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7001), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185590] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3549), 1,
      anon_sym_COLON2,
    STATE(4414), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3551), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [185617] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1841), 1,
      anon_sym_DOT,
    STATE(4415), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1701), 11,
      anon_sym_and,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
  [185644] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(7003), 1,
      anon_sym_RPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4416), 1,
      sym_attribute,
    STATE(7308), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [185683] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4417), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2362), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185714] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7007), 1,
      anon_sym_COLON2,
    STATE(4418), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7005), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [185741] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4419), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7009), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185772] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(7011), 1,
      anon_sym_RPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4420), 1,
      sym_attribute,
    STATE(6378), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [185811] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4402), 1,
      aux_sym_expression_item_repeat1,
    STATE(4421), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7013), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185842] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3502), 1,
      anon_sym_COLON2,
    STATE(4422), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3504), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [185869] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4423), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5727), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185900] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7017), 1,
      anon_sym_COLON2,
    STATE(4424), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7015), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [185927] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3090), 1,
      anon_sym_COLON2,
    STATE(4425), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3092), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [185954] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4360), 1,
      aux_sym_expression_item_repeat1,
    STATE(4426), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5727), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [185985] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(25), 1,
      anon_sym_LBRACK_PERCENT,
    ACTIONS(27), 1,
      anon_sym_LBRACE_PERCENT,
    ACTIONS(4008), 1,
      anon_sym_LPAREN,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(7019), 1,
      anon_sym_RPAREN,
    STATE(3167), 1,
      sym__simple_module_expression,
    STATE(4427), 1,
      sym_attribute,
    STATE(6675), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3162), 5,
      sym_typed_module_expression,
      sym_packed_module,
      sym_parenthesized_module_expression,
      sym_extension,
      sym_quoted_extension,
  [186024] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2081), 1,
      anon_sym_COLON2,
    STATE(4428), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2075), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [186051] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4376), 1,
      aux_sym_expression_item_repeat1,
    STATE(4429), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5743), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [186082] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7023), 1,
      anon_sym_COLON2,
    STATE(4430), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7021), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [186109] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6839), 1,
      anon_sym_COLON2,
    STATE(4431), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6837), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [186136] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4432), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7025), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [186167] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4379), 1,
      aux_sym_expression_item_repeat1,
    STATE(4433), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7027), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [186198] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4434), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7029), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [186229] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7033), 1,
      anon_sym_COLON2,
    STATE(4435), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7031), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [186256] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3553), 1,
      anon_sym_COLON2,
    STATE(4436), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3555), 11,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [186283] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4437), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7035), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [186314] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4408), 1,
      aux_sym_expression_item_repeat1,
    STATE(4438), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7037), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [186345] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4439), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3875), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3873), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186371] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4440), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3897), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3895), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186397] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4441), 1,
      sym_attribute,
    STATE(4470), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7039), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [186427] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4442), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3716), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3714), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186453] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4443), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3803), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3801), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186479] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4444), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3901), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3899), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186505] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4445), 1,
      sym_attribute,
    STATE(4478), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7041), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [186535] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4446), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3807), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3805), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186561] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4447), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3811), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3809), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186587] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4448), 1,
      sym_attribute,
    STATE(4450), 1,
      aux_sym_expression_item_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7043), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [186617] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4449), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3827), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3825), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186643] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4450), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7045), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [186673] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7049), 1,
      anon_sym_COLON,
    ACTIONS(7051), 1,
      anon_sym_COLON2,
    STATE(4451), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7047), 9,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [186701] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4452), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3775), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3773), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186727] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4453), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3720), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3718), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186753] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4454), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1715), 11,
      anon_sym_and,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
  [186777] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4455), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3815), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3813), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186803] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3231), 1,
      anon_sym_PIPE,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    ACTIONS(7053), 1,
      anon_sym_private,
    STATE(3720), 1,
      sym_variant_declaration,
    STATE(3870), 1,
      sym_constructor_declaration,
    STATE(4456), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3600), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [186843] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4457), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3624), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3620), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186869] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4458), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3831), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3829), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186895] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4459), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3889), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3887), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [186921] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4460), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7055), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [186951] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3052), 1,
      anon_sym_PLUS_EQ,
    ACTIONS(7057), 1,
      anon_sym_EQ,
    ACTIONS(7059), 1,
      anon_sym_COLON_EQ,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4461), 1,
      sym_attribute,
    STATE(4635), 1,
      aux_sym_type_binding_repeat1,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(5166), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4487), 2,
      anon_sym_and,
      anon_sym_in,
  [186993] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4462), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1841), 11,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [187017] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    ACTIONS(7063), 1,
      anon_sym_as,
    STATE(4463), 1,
      sym_attribute,
    STATE(4522), 1,
      aux_sym_type_binding_repeat1,
    STATE(4643), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4539), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [187049] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4464), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3819), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3817), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187075] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3231), 1,
      anon_sym_PIPE,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    ACTIONS(7065), 1,
      anon_sym_private,
    STATE(3587), 1,
      sym_variant_declaration,
    STATE(3870), 1,
      sym_constructor_declaration,
    STATE(4465), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3600), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [187115] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4466), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1703), 11,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [187139] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4467), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7067), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [187169] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3287), 1,
      anon_sym_PIPE,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    ACTIONS(7069), 1,
      anon_sym_private,
    STATE(3549), 1,
      sym_constructor_declaration,
    STATE(3720), 1,
      sym_variant_declaration,
    STATE(4468), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3440), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [187209] = 13,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(3052), 1,
      anon_sym_PLUS_EQ,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    ACTIONS(7071), 1,
      anon_sym_EQ,
    ACTIONS(7073), 1,
      anon_sym_COLON_EQ,
    STATE(4469), 1,
      sym_attribute,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4644), 1,
      aux_sym_type_binding_repeat1,
    STATE(5242), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4509), 2,
      anon_sym_and,
      anon_sym_in,
  [187251] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4470), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7075), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [187281] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4471), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3845), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3843), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187307] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4472), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3849), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3847), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187333] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4473), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3885), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3883), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187359] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4467), 1,
      aux_sym_expression_item_repeat1,
    STATE(4474), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7077), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [187389] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3192), 1,
      anon_sym_PIPE,
    ACTIONS(6597), 1,
      anon_sym_LPAREN,
    ACTIONS(6603), 1,
      anon_sym_LBRACK,
    ACTIONS(6609), 1,
      sym__uppercase_identifier,
    ACTIONS(7079), 1,
      anon_sym_private,
    STATE(4475), 1,
      sym_attribute,
    STATE(5004), 1,
      sym_constructor_declaration,
    STATE(5065), 1,
      sym_variant_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6605), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4628), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [187429] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4476), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3799), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3797), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187455] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    ACTIONS(7063), 1,
      anon_sym_as,
    STATE(4477), 1,
      sym_attribute,
    STATE(4526), 1,
      aux_sym_type_binding_repeat1,
    STATE(4643), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4545), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [187487] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3430), 1,
      aux_sym_expression_item_repeat1,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4478), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7081), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [187517] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7083), 1,
      anon_sym_DOT,
    STATE(4479), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4245), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [187543] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4480), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3771), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3769), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187569] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4481), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3853), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3851), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187595] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    ACTIONS(7063), 1,
      anon_sym_as,
    STATE(4482), 1,
      sym_attribute,
    STATE(4534), 1,
      aux_sym_type_binding_repeat1,
    STATE(4643), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4549), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [187627] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3192), 1,
      anon_sym_PIPE,
    ACTIONS(6597), 1,
      anon_sym_LPAREN,
    ACTIONS(6603), 1,
      anon_sym_LBRACK,
    ACTIONS(6609), 1,
      sym__uppercase_identifier,
    ACTIONS(7085), 1,
      anon_sym_private,
    STATE(4483), 1,
      sym_attribute,
    STATE(5004), 1,
      sym_constructor_declaration,
    STATE(5057), 1,
      sym_variant_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6605), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4628), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [187667] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4484), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3857), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3855), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187693] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4485), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3893), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3891), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187719] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4486), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3917), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3915), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187745] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1999), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(3494), 1,
      sym_item_attribute,
    STATE(4460), 1,
      aux_sym_expression_item_repeat1,
    STATE(4487), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7087), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [187775] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4488), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3913), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3911), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187801] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3287), 1,
      anon_sym_PIPE,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    ACTIONS(7089), 1,
      anon_sym_private,
    STATE(3549), 1,
      sym_constructor_declaration,
    STATE(3587), 1,
      sym_variant_declaration,
    STATE(4489), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3440), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [187841] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4490), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3871), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3869), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187867] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4491), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3823), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3821), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187893] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4492), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3879), 2,
      anon_sym_as,
      sym__lowercase_identifier,
    ACTIONS(3877), 9,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_GT,
      anon_sym_POUND,
      sym__uppercase_identifier,
  [187919] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4493), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1845), 11,
      anon_sym_and,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK_AT_AT,
  [187943] = 11,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7091), 1,
      sym__lowercase_identifier,
    ACTIONS(7093), 1,
      anon_sym_LPAREN,
    ACTIONS(7095), 1,
      sym__uppercase_identifier,
    STATE(2571), 1,
      sym__delimited_type,
    STATE(4494), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
  [187980] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4495), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4245), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [188003] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3287), 1,
      anon_sym_PIPE,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    STATE(3549), 1,
      sym_constructor_declaration,
    STATE(3608), 1,
      sym_variant_declaration,
    STATE(4496), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3440), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [188040] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4497), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4314), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [188063] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4498), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4306), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [188086] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7097), 1,
      anon_sym_constraint,
    STATE(4643), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4499), 2,
      sym_attribute,
      aux_sym_type_binding_repeat1,
    ACTIONS(4630), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [188113] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4500), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4356), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [188136] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7106), 1,
      anon_sym_COLON2,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    STATE(4501), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7100), 5,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_when,
  [188169] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4502), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4284), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [188192] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7114), 1,
      anon_sym_COLON2,
    STATE(4503), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7112), 8,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_when,
  [188219] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3231), 1,
      anon_sym_PIPE,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    STATE(3608), 1,
      sym_variant_declaration,
    STATE(3870), 1,
      sym_constructor_declaration,
    STATE(4504), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3600), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [188256] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3231), 1,
      anon_sym_PIPE,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    STATE(3727), 1,
      sym_variant_declaration,
    STATE(3870), 1,
      sym_constructor_declaration,
    STATE(4505), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3600), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [188293] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6853), 1,
      anon_sym_COLON2,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    STATE(4506), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6851), 6,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_when,
  [188324] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7116), 1,
      anon_sym_EQ,
    ACTIONS(7118), 1,
      anon_sym_RPAREN,
    ACTIONS(7120), 1,
      anon_sym_COLON2,
    STATE(4507), 1,
      sym_attribute,
    STATE(6542), 1,
      sym__maybe_polymorphic_typed,
    STATE(6549), 1,
      sym__polymorphic_typed,
    STATE(7080), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [188365] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4508), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4507), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(1701), 8,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [188390] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3287), 1,
      anon_sym_PIPE,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    STATE(3549), 1,
      sym_constructor_declaration,
    STATE(3727), 1,
      sym_variant_declaration,
    STATE(4509), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3440), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [188427] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4510), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1535), 2,
      anon_sym_COLON2,
      anon_sym_POUND,
    ACTIONS(1537), 8,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_COLON_GT,
      sym_hash_operator,
  [188452] = 11,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7093), 1,
      anon_sym_LPAREN,
    ACTIONS(7095), 1,
      sym__uppercase_identifier,
    ACTIONS(7122), 1,
      sym__lowercase_identifier,
    STATE(2571), 1,
      sym__delimited_type,
    STATE(4511), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
  [188489] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4512), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4330), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [188512] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4513), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4362), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [188535] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6839), 1,
      anon_sym_COLON2,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    STATE(4514), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6837), 8,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_when,
  [188562] = 11,
    ACTIONS(13), 1,
      anon_sym_LBRACK,
    ACTIONS(15), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(17), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7093), 1,
      anon_sym_LPAREN,
    ACTIONS(7095), 1,
      sym__uppercase_identifier,
    ACTIONS(7124), 1,
      sym__lowercase_identifier,
    STATE(2571), 1,
      sym__delimited_type,
    STATE(4515), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(2586), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
  [188599] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7128), 1,
      anon_sym_COLON2,
    STATE(4516), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7126), 5,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_when,
  [188632] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    ACTIONS(7063), 1,
      anon_sym_as,
    ACTIONS(7130), 1,
      anon_sym_EQ,
    STATE(4517), 1,
      sym_attribute,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4659), 1,
      aux_sym_type_binding_repeat1,
    STATE(5117), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4606), 2,
      anon_sym_and,
      anon_sym_in,
  [188671] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3192), 1,
      anon_sym_PIPE,
    ACTIONS(6597), 1,
      anon_sym_LPAREN,
    ACTIONS(6603), 1,
      anon_sym_LBRACK,
    ACTIONS(6609), 1,
      sym__uppercase_identifier,
    STATE(4518), 1,
      sym_attribute,
    STATE(5004), 1,
      sym_constructor_declaration,
    STATE(5118), 1,
      sym_variant_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6605), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4628), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [188708] = 13,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7120), 1,
      anon_sym_COLON2,
    ACTIONS(7132), 1,
      anon_sym_EQ,
    ACTIONS(7134), 1,
      anon_sym_RPAREN,
    STATE(4519), 1,
      sym_attribute,
    STATE(6345), 1,
      sym__typed,
    STATE(6963), 1,
      sym__maybe_polymorphic_typed,
    STATE(6996), 1,
      sym__polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [188749] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3192), 1,
      anon_sym_PIPE,
    ACTIONS(6597), 1,
      anon_sym_LPAREN,
    ACTIONS(6603), 1,
      anon_sym_LBRACK,
    ACTIONS(6609), 1,
      sym__uppercase_identifier,
    STATE(4520), 1,
      sym_attribute,
    STATE(5004), 1,
      sym_constructor_declaration,
    STATE(5058), 1,
      sym_variant_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6605), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4628), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [188786] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4521), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4288), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [188809] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4499), 1,
      aux_sym_type_binding_repeat1,
    STATE(4522), 1,
      sym_attribute,
    STATE(4643), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4571), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [188838] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3291), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3293), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(3698), 1,
      anon_sym_LBRACK,
    ACTIONS(7095), 1,
      sym__uppercase_identifier,
    ACTIONS(7136), 1,
      sym__lowercase_identifier,
    ACTIONS(7138), 1,
      anon_sym_LPAREN,
    STATE(2989), 1,
      sym__delimited_type,
    STATE(4523), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3031), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
  [188875] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1297), 1,
      anon_sym_COLON2,
    ACTIONS(6747), 1,
      anon_sym_DOT_DOT,
    STATE(4524), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1295), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [188902] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4525), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4322), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [188925] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4499), 1,
      aux_sym_type_binding_repeat1,
    STATE(4526), 1,
      sym_attribute,
    STATE(4643), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4575), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [188954] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4527), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4296), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [188977] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    STATE(3801), 1,
      sym_constructor_declaration,
    STATE(4528), 1,
      sym_attribute,
    STATE(4620), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3440), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [189014] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4529), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4326), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [189037] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3666), 1,
      anon_sym_LBRACK,
    ACTIONS(3668), 1,
      anon_sym_LBRACK_GT,
    ACTIONS(3670), 1,
      anon_sym_LBRACK_LT,
    ACTIONS(7095), 1,
      sym__uppercase_identifier,
    ACTIONS(7140), 1,
      sym__lowercase_identifier,
    ACTIONS(7142), 1,
      anon_sym_LPAREN,
    STATE(4453), 1,
      sym__delimited_type,
    STATE(4530), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4458), 3,
      sym_polymorphic_variant_type,
      sym_package_type,
      sym_parenthesized_type,
  [189074] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6541), 1,
      anon_sym_STAR,
    STATE(4531), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4131), 9,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
  [189099] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    STATE(3801), 1,
      sym_constructor_declaration,
    STATE(4532), 1,
      sym_attribute,
    STATE(4634), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4050), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [189136] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(6597), 1,
      anon_sym_LPAREN,
    ACTIONS(6603), 1,
      anon_sym_LBRACK,
    ACTIONS(6609), 1,
      sym__uppercase_identifier,
    STATE(4533), 1,
      sym_attribute,
    STATE(4629), 1,
      sym__attribute,
    STATE(5499), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6605), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4628), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [189173] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4499), 1,
      aux_sym_type_binding_repeat1,
    STATE(4534), 1,
      sym_attribute,
    STATE(4643), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4579), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [189202] = 12,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    ACTIONS(7063), 1,
      anon_sym_as,
    ACTIONS(7144), 1,
      anon_sym_EQ,
    STATE(4535), 1,
      sym_attribute,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4681), 1,
      aux_sym_type_binding_repeat1,
    STATE(5046), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4600), 2,
      anon_sym_and,
      anon_sym_in,
  [189241] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6839), 1,
      anon_sym_COLON2,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    STATE(4536), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6837), 7,
      anon_sym_EQ,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_when,
  [189270] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4537), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1547), 2,
      anon_sym_COLON2,
      anon_sym_POUND,
    ACTIONS(1549), 8,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_DOT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_with,
      anon_sym_COLON_GT,
      sym_hash_operator,
  [189295] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4538), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4368), 10,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_LBRACK_PERCENT,
      anon_sym_LBRACE_PERCENT,
  [189318] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7148), 1,
      aux_sym_character_content_token2,
    STATE(4539), 1,
      sym_attribute,
    STATE(6809), 1,
      sym_escape_sequence,
    STATE(6964), 1,
      sym_character_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7146), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7150), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [189348] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4540), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7152), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [189370] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7148), 1,
      aux_sym_character_content_token2,
    STATE(4541), 1,
      sym_attribute,
    STATE(6809), 1,
      sym_escape_sequence,
    STATE(6851), 1,
      sym_character_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7146), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7150), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [189400] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3070), 1,
      anon_sym_COLON2,
    STATE(4542), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3072), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [189424] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4543), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4533), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4529), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [189448] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    ACTIONS(7063), 1,
      anon_sym_as,
    STATE(4544), 1,
      sym_attribute,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4656), 1,
      aux_sym_type_binding_repeat1,
    STATE(5105), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4693), 2,
      anon_sym_and,
      anon_sym_in,
  [189484] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4543), 1,
      anon_sym_as,
    ACTIONS(7154), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(4545), 1,
      sym_attribute,
    STATE(4636), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4545), 5,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [189514] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3102), 1,
      anon_sym_COLON2,
    STATE(4546), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3104), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [189538] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4547), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4559), 9,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [189560] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1297), 1,
      anon_sym_COLON2,
    STATE(4548), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1295), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [189584] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7160), 1,
      sym__right_quoted_string_delimiter,
    STATE(4549), 1,
      sym_attribute,
    STATE(4632), 1,
      aux_sym_quoted_string_content_repeat1,
    STATE(7158), 1,
      sym_quoted_string_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7158), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7156), 4,
      aux_sym_string_content_token1,
      aux_sym_quoted_string_content_token1,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [189614] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4697), 1,
      anon_sym_in,
    ACTIONS(6567), 1,
      anon_sym_LBRACE,
    STATE(4550), 1,
      sym_attribute,
    STATE(4756), 1,
      aux_sym_external_repeat1,
    STATE(5329), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5492), 2,
      sym_string,
      sym_quoted_string,
  [189650] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7148), 1,
      aux_sym_character_content_token2,
    STATE(4551), 1,
      sym_attribute,
    STATE(6809), 1,
      sym_escape_sequence,
    STATE(7033), 1,
      sym_character_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7146), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7150), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [189680] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7162), 1,
      sym__lowercase_identifier,
    ACTIONS(7164), 1,
      anon_sym_LPAREN,
    ACTIONS(7166), 1,
      anon_sym_LBRACK,
    ACTIONS(7168), 1,
      anon_sym_LBRACE,
    ACTIONS(7170), 1,
      sym_indexing_operator,
    STATE(1039), 1,
      sym_field_path,
    STATE(4552), 1,
      sym_attribute,
    STATE(5545), 1,
      sym_indexing_operator_path,
    STATE(6468), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [189718] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3158), 1,
      anon_sym_COLON2,
    STATE(4553), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3160), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [189742] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7170), 1,
      sym_indexing_operator,
    ACTIONS(7172), 1,
      sym__lowercase_identifier,
    ACTIONS(7174), 1,
      anon_sym_LPAREN,
    ACTIONS(7176), 1,
      anon_sym_LBRACK,
    ACTIONS(7178), 1,
      anon_sym_LBRACE,
    STATE(2039), 1,
      sym_field_path,
    STATE(4554), 1,
      sym_attribute,
    STATE(5890), 1,
      sym_indexing_operator_path,
    STATE(6770), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [189780] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7148), 1,
      aux_sym_character_content_token2,
    STATE(4555), 1,
      sym_attribute,
    STATE(6809), 1,
      sym_escape_sequence,
    STATE(6835), 1,
      sym_character_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7146), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7150), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [189810] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7063), 1,
      anon_sym_as,
    STATE(4556), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4555), 8,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [189834] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7170), 1,
      sym_indexing_operator,
    ACTIONS(7180), 1,
      sym__lowercase_identifier,
    ACTIONS(7182), 1,
      anon_sym_LPAREN,
    ACTIONS(7184), 1,
      anon_sym_LBRACK,
    ACTIONS(7186), 1,
      anon_sym_LBRACE,
    STATE(942), 1,
      sym_field_path,
    STATE(4557), 1,
      sym_attribute,
    STATE(5870), 1,
      sym_indexing_operator_path,
    STATE(6617), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [189872] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7148), 1,
      aux_sym_character_content_token2,
    STATE(4558), 1,
      sym_attribute,
    STATE(6809), 1,
      sym_escape_sequence,
    STATE(7244), 1,
      sym_character_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7146), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7150), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [189902] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4559), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4507), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4503), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [189926] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3130), 1,
      anon_sym_COLON2,
    STATE(4560), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3132), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [189950] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7170), 1,
      sym_indexing_operator,
    ACTIONS(7188), 1,
      sym__lowercase_identifier,
    ACTIONS(7190), 1,
      anon_sym_LPAREN,
    ACTIONS(7192), 1,
      anon_sym_LBRACK,
    ACTIONS(7194), 1,
      anon_sym_LBRACE,
    STATE(854), 1,
      sym_field_path,
    STATE(4561), 1,
      sym_attribute,
    STATE(5881), 1,
      sym_indexing_operator_path,
    STATE(6697), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [189988] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3114), 1,
      anon_sym_COLON2,
    STATE(4562), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3116), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190012] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    ACTIONS(7063), 1,
      anon_sym_as,
    STATE(4563), 1,
      sym_attribute,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4668), 1,
      aux_sym_type_binding_repeat1,
    STATE(5168), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4683), 2,
      anon_sym_and,
      anon_sym_in,
  [190048] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3166), 1,
      anon_sym_COLON2,
    STATE(4564), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3168), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190072] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3134), 1,
      anon_sym_COLON2,
    STATE(4565), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3136), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190096] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7170), 1,
      sym_indexing_operator,
    ACTIONS(7196), 1,
      sym__lowercase_identifier,
    ACTIONS(7198), 1,
      anon_sym_LPAREN,
    ACTIONS(7200), 1,
      anon_sym_LBRACK,
    ACTIONS(7202), 1,
      anon_sym_LBRACE,
    STATE(751), 1,
      sym_field_path,
    STATE(4566), 1,
      sym_attribute,
    STATE(5835), 1,
      sym_indexing_operator_path,
    STATE(6545), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [190134] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7170), 1,
      sym_indexing_operator,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7206), 1,
      anon_sym_LPAREN,
    ACTIONS(7208), 1,
      anon_sym_LBRACK,
    ACTIONS(7210), 1,
      anon_sym_LBRACE,
    STATE(1686), 1,
      sym_field_path,
    STATE(4567), 1,
      sym_attribute,
    STATE(5858), 1,
      sym_indexing_operator_path,
    STATE(6404), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [190172] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7170), 1,
      sym_indexing_operator,
    ACTIONS(7212), 1,
      sym__lowercase_identifier,
    ACTIONS(7214), 1,
      anon_sym_LPAREN,
    ACTIONS(7216), 1,
      anon_sym_LBRACK,
    ACTIONS(7218), 1,
      anon_sym_LBRACE,
    STATE(1523), 1,
      sym_field_path,
    STATE(4568), 1,
      sym_attribute,
    STATE(5875), 1,
      sym_indexing_operator_path,
    STATE(6658), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [190210] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7148), 1,
      aux_sym_character_content_token2,
    STATE(4569), 1,
      sym_attribute,
    STATE(6578), 1,
      sym_character_content,
    STATE(6809), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7146), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7150), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [190240] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3086), 1,
      anon_sym_COLON2,
    STATE(4570), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3088), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190264] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1535), 1,
      anon_sym_COLON2,
    ACTIONS(1745), 1,
      anon_sym_POUND,
    STATE(4571), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1747), 3,
      anon_sym_DOT,
      anon_sym_with,
      sym_hash_operator,
    ACTIONS(1537), 4,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
  [190292] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3022), 1,
      anon_sym_COLON2,
    STATE(4572), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3024), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190316] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4573), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3638), 9,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
  [190338] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4574), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4507), 9,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [190360] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3098), 1,
      anon_sym_COLON2,
    STATE(4575), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3100), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190384] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7170), 1,
      sym_indexing_operator,
    ACTIONS(7220), 1,
      sym__lowercase_identifier,
    ACTIONS(7222), 1,
      anon_sym_LPAREN,
    ACTIONS(7224), 1,
      anon_sym_LBRACK,
    ACTIONS(7226), 1,
      anon_sym_LBRACE,
    STATE(2742), 1,
      sym_field_path,
    STATE(4576), 1,
      sym_attribute,
    STATE(5884), 1,
      sym_indexing_operator_path,
    STATE(6734), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [190422] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7148), 1,
      aux_sym_character_content_token2,
    STATE(4577), 1,
      sym_attribute,
    STATE(6809), 1,
      sym_escape_sequence,
    STATE(6921), 1,
      sym_character_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7146), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7150), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [190452] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7148), 1,
      aux_sym_character_content_token2,
    STATE(4578), 1,
      sym_attribute,
    STATE(6809), 1,
      sym_escape_sequence,
    STATE(7189), 1,
      sym_character_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7146), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7150), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [190482] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3000), 1,
      anon_sym_COLON2,
    STATE(4579), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3002), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190506] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3154), 1,
      anon_sym_COLON2,
    STATE(4580), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3156), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190530] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2984), 1,
      anon_sym_COLON2,
    STATE(4581), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2986), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190554] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7228), 1,
      anon_sym_LPAREN,
    ACTIONS(7230), 1,
      anon_sym_DOT,
    STATE(4582), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4521), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [190580] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7148), 1,
      aux_sym_character_content_token2,
    STATE(4583), 1,
      sym_attribute,
    STATE(6566), 1,
      sym_character_content,
    STATE(6809), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7146), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7150), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [190610] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3040), 1,
      anon_sym_COLON2,
    STATE(4584), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3042), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190634] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3138), 1,
      anon_sym_COLON2,
    STATE(4585), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3140), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190658] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7170), 1,
      sym_indexing_operator,
    ACTIONS(7206), 1,
      anon_sym_LPAREN,
    ACTIONS(7208), 1,
      anon_sym_LBRACK,
    ACTIONS(7210), 1,
      anon_sym_LBRACE,
    ACTIONS(7232), 1,
      sym__lowercase_identifier,
    STATE(1686), 1,
      sym_field_path,
    STATE(4586), 1,
      sym_attribute,
    STATE(5858), 1,
      sym_indexing_operator_path,
    STATE(6519), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [190696] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3110), 1,
      anon_sym_COLON2,
    STATE(4587), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3112), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190720] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3142), 1,
      anon_sym_COLON2,
    STATE(4588), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3144), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190744] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4589), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4533), 9,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_DOT,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [190766] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7148), 1,
      aux_sym_character_content_token2,
    STATE(4590), 1,
      sym_attribute,
    STATE(6548), 1,
      sym_character_content,
    STATE(6809), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7146), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7150), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [190796] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7148), 1,
      aux_sym_character_content_token2,
    STATE(4591), 1,
      sym_attribute,
    STATE(6698), 1,
      sym_character_content,
    STATE(6809), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7146), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7150), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [190826] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3146), 1,
      anon_sym_COLON2,
    STATE(4592), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3148), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190850] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4543), 1,
      anon_sym_as,
    ACTIONS(7154), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(4593), 1,
      sym_attribute,
    STATE(4641), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4549), 5,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [190880] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3150), 1,
      anon_sym_COLON2,
    STATE(4594), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3152), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [190904] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7148), 1,
      aux_sym_character_content_token2,
    STATE(4595), 1,
      sym_attribute,
    STATE(6778), 1,
      sym_character_content,
    STATE(6809), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7146), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7150), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [190934] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7148), 1,
      aux_sym_character_content_token2,
    STATE(4596), 1,
      sym_attribute,
    STATE(6809), 1,
      sym_escape_sequence,
    STATE(6842), 1,
      sym_character_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7146), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7150), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [190964] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7148), 1,
      aux_sym_character_content_token2,
    STATE(4597), 1,
      sym_attribute,
    STATE(6809), 1,
      sym_escape_sequence,
    STATE(6899), 1,
      sym_character_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7146), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7150), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [190994] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7170), 1,
      sym_indexing_operator,
    ACTIONS(7234), 1,
      sym__lowercase_identifier,
    ACTIONS(7236), 1,
      anon_sym_LPAREN,
    ACTIONS(7238), 1,
      anon_sym_LBRACK,
    ACTIONS(7240), 1,
      anon_sym_LBRACE,
    STATE(2390), 1,
      sym_field_path,
    STATE(4598), 1,
      sym_attribute,
    STATE(5896), 1,
      sym_indexing_operator_path,
    STATE(6796), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [191032] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3118), 1,
      anon_sym_COLON2,
    STATE(4599), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3120), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [191056] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3122), 1,
      anon_sym_COLON2,
    STATE(4600), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3124), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [191080] = 12,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7170), 1,
      sym_indexing_operator,
    ACTIONS(7242), 1,
      sym__lowercase_identifier,
    ACTIONS(7244), 1,
      anon_sym_LPAREN,
    ACTIONS(7246), 1,
      anon_sym_LBRACK,
    ACTIONS(7248), 1,
      anon_sym_LBRACE,
    STATE(1912), 1,
      sym_field_path,
    STATE(4601), 1,
      sym_attribute,
    STATE(5864), 1,
      sym_indexing_operator_path,
    STATE(6574), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [191118] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3126), 1,
      anon_sym_COLON2,
    STATE(4602), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3128), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [191142] = 11,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4729), 1,
      anon_sym_in,
    ACTIONS(6567), 1,
      anon_sym_LBRACE,
    STATE(4603), 1,
      sym_attribute,
    STATE(4756), 1,
      aux_sym_external_repeat1,
    STATE(5392), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5492), 2,
      sym_string,
      sym_quoted_string,
  [191178] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7148), 1,
      aux_sym_character_content_token2,
    STATE(4604), 1,
      sym_attribute,
    STATE(6809), 1,
      sym_escape_sequence,
    STATE(7038), 1,
      sym_character_content,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7146), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7150), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [191208] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3062), 1,
      anon_sym_COLON2,
    STATE(4605), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3064), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [191232] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7148), 1,
      aux_sym_character_content_token2,
    STATE(4606), 1,
      sym_attribute,
    STATE(6386), 1,
      sym_character_content,
    STATE(6809), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7146), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7150), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [191262] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4648), 1,
      anon_sym_COLON2,
    ACTIONS(7250), 1,
      anon_sym_LPAREN,
    STATE(4760), 1,
      sym_module_parameter,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4607), 2,
      sym_attribute,
      aux_sym_module_binding_repeat1,
    ACTIONS(4646), 5,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [191290] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4543), 1,
      anon_sym_as,
    ACTIONS(7154), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(4608), 1,
      sym_attribute,
    STATE(4667), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4539), 5,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [191320] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2970), 1,
      anon_sym_COLON2,
    STATE(4609), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2972), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [191344] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(6553), 1,
      anon_sym_DASH_GT,
    STATE(4610), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3987), 8,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_as,
      anon_sym_LBRACK_AT_AT,
  [191368] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2984), 1,
      anon_sym_COLON2,
    STATE(4611), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2986), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [191392] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3044), 1,
      anon_sym_COLON2,
    STATE(4612), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3046), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [191416] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7148), 1,
      aux_sym_character_content_token2,
    STATE(4613), 1,
      sym_attribute,
    STATE(6367), 1,
      sym_character_content,
    STATE(6809), 1,
      sym_escape_sequence,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7146), 2,
      sym__null,
      aux_sym_character_content_token1,
    ACTIONS(7150), 4,
      aux_sym_escape_sequence_token1,
      aux_sym_escape_sequence_token2,
      aux_sym_escape_sequence_token3,
      aux_sym_escape_sequence_token4,
  [191446] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3008), 1,
      anon_sym_COLON2,
    STATE(4614), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3010), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [191470] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1547), 1,
      anon_sym_COLON2,
    ACTIONS(1745), 1,
      anon_sym_POUND,
    STATE(4615), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1747), 3,
      anon_sym_DOT,
      anon_sym_with,
      sym_hash_operator,
    ACTIONS(1549), 4,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
  [191498] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3058), 1,
      anon_sym_COLON2,
    STATE(4616), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3060), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [191522] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4617), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7253), 9,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_initializer,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [191544] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3170), 1,
      anon_sym_COLON2,
    STATE(4618), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3172), 8,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [191568] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7255), 1,
      anon_sym__,
    STATE(4619), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5922), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [191599] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    STATE(3814), 1,
      sym_constructor_declaration,
    STATE(4620), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3440), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [191630] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6665), 1,
      anon_sym_STAR,
    STATE(4621), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4131), 7,
      anon_sym_EQ,
      anon_sym_DASH_GT,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_as,
      anon_sym_GT,
  [191653] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7255), 1,
      anon_sym__,
    STATE(4622), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5756), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [191684] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4623), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7257), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [191705] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4624), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7259), 8,
      anon_sym_constraint,
      anon_sym_val,
      anon_sym_end,
      anon_sym_inherit,
      anon_sym_method,
      anon_sym_LBRACK_AT_AT_AT,
      anon_sym_LBRACK_PERCENT_PERCENT,
      anon_sym_LBRACE_PERCENT_PERCENT,
  [191726] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4691), 1,
      sym__lowercase_identifier,
    ACTIONS(6667), 1,
      anon_sym_POUND,
    ACTIONS(7261), 1,
      anon_sym_SEMI,
    ACTIONS(7263), 1,
      anon_sym_GT,
    STATE(4480), 1,
      sym_type_constructor_path,
    STATE(4625), 1,
      sym_attribute,
    STATE(5554), 1,
      aux_sym_object_type_repeat1,
    STATE(6026), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [191761] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4626), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4869), 8,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [191782] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4627), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1701), 8,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [191803] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7265), 1,
      anon_sym_EQ,
    ACTIONS(7267), 1,
      anon_sym_of,
    ACTIONS(7269), 1,
      anon_sym_COLON2,
    STATE(4628), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4756), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [191830] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6597), 1,
      anon_sym_LPAREN,
    ACTIONS(6603), 1,
      anon_sym_LBRACK,
    ACTIONS(6609), 1,
      sym__uppercase_identifier,
    STATE(4629), 1,
      sym_attribute,
    STATE(5297), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6605), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4628), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [191861] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4691), 1,
      sym__lowercase_identifier,
    ACTIONS(6667), 1,
      anon_sym_POUND,
    ACTIONS(7271), 1,
      anon_sym_SEMI,
    ACTIONS(7273), 1,
      anon_sym_GT,
    STATE(4480), 1,
      sym_type_constructor_path,
    STATE(4630), 1,
      sym_attribute,
    STATE(5590), 1,
      aux_sym_object_type_repeat1,
    STATE(6026), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [191896] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7255), 1,
      anon_sym__,
    STATE(4631), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5745), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [191927] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7275), 1,
      sym__right_quoted_string_delimiter,
    STATE(4632), 1,
      sym_attribute,
    STATE(4674), 1,
      aux_sym_quoted_string_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7158), 2,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7156), 4,
      aux_sym_string_content_token1,
      aux_sym_quoted_string_content_token1,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [191954] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4633), 1,
      sym_attribute,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4681), 1,
      aux_sym_type_binding_repeat1,
    STATE(5046), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4600), 2,
      anon_sym_and,
      anon_sym_in,
  [191987] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    STATE(3814), 1,
      sym_constructor_declaration,
    STATE(4634), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4050), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [192018] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4499), 1,
      aux_sym_type_binding_repeat1,
    STATE(4635), 1,
      sym_attribute,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(5054), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4881), 2,
      anon_sym_and,
      anon_sym_in,
  [192051] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7154), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(4636), 1,
      sym_attribute,
    STATE(4670), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4575), 5,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [192078] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7255), 1,
      anon_sym__,
    STATE(4637), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5594), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [192109] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7277), 1,
      anon_sym_COLON,
    STATE(4638), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2938), 7,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [192132] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7279), 1,
      anon_sym_COLON,
    STATE(4639), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7047), 7,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [192155] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7134), 1,
      anon_sym_RPAREN,
    ACTIONS(7281), 1,
      anon_sym_COLON2,
    STATE(4640), 1,
      sym_attribute,
    STATE(6345), 1,
      sym__typed,
    STATE(6347), 1,
      sym__polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [192190] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7154), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(4641), 1,
      sym_attribute,
    STATE(4670), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4579), 5,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [192217] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7283), 1,
      anon_sym_and,
    STATE(4642), 1,
      sym_attribute,
    STATE(4654), 1,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4847), 6,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [192242] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4643), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4642), 8,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_constraint,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [192263] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4499), 1,
      aux_sym_type_binding_repeat1,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4644), 1,
      sym_attribute,
    STATE(5189), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4812), 2,
      anon_sym_and,
      anon_sym_in,
  [192296] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7255), 1,
      anon_sym__,
    STATE(4645), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5878), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [192327] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4646), 1,
      sym_attribute,
    STATE(4656), 1,
      aux_sym_type_binding_repeat1,
    STATE(5105), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4693), 2,
      anon_sym_and,
      anon_sym_in,
  [192360] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7255), 1,
      anon_sym__,
    STATE(4647), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5918), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [192391] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6597), 1,
      anon_sym_LPAREN,
    ACTIONS(6603), 1,
      anon_sym_LBRACK,
    ACTIONS(6609), 1,
      sym__uppercase_identifier,
    STATE(4648), 1,
      sym_attribute,
    STATE(5110), 1,
      sym_constructor_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6605), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(4628), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [192422] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4649), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4766), 8,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_PIPE,
      anon_sym_of,
      anon_sym_COLON2,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [192443] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4650), 1,
      sym_attribute,
    STATE(4659), 1,
      aux_sym_type_binding_repeat1,
    STATE(5117), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4606), 2,
      anon_sym_and,
      anon_sym_in,
  [192476] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4348), 1,
      anon_sym_STAR,
    STATE(4651), 1,
      sym_attribute,
    STATE(4660), 1,
      aux_sym__constructor_argument_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5041), 6,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [192501] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7255), 1,
      anon_sym__,
    STATE(4652), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5631), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [192532] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4691), 1,
      sym__lowercase_identifier,
    ACTIONS(6667), 1,
      anon_sym_POUND,
    ACTIONS(7285), 1,
      anon_sym_SEMI,
    ACTIONS(7287), 1,
      anon_sym_GT,
    STATE(4480), 1,
      sym_type_constructor_path,
    STATE(4653), 1,
      sym_attribute,
    STATE(5690), 1,
      aux_sym_object_type_repeat1,
    STATE(6026), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [192567] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7289), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4654), 2,
      sym_attribute,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(4721), 6,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [192590] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    STATE(3603), 1,
      sym_constructor_declaration,
    STATE(4655), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3600), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [192621] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4499), 1,
      aux_sym_type_binding_repeat1,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4656), 1,
      sym_attribute,
    STATE(5261), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4804), 2,
      anon_sym_and,
      anon_sym_in,
  [192654] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4657), 1,
      sym_attribute,
    STATE(4666), 1,
      aux_sym_type_binding_repeat1,
    STATE(5165), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4792), 2,
      anon_sym_and,
      anon_sym_in,
  [192687] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4658), 1,
      sym_attribute,
    STATE(4668), 1,
      aux_sym_type_binding_repeat1,
    STATE(5168), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4683), 2,
      anon_sym_and,
      anon_sym_in,
  [192720] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4499), 1,
      aux_sym_type_binding_repeat1,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4659), 1,
      sym_attribute,
    STATE(5170), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4748), 2,
      anon_sym_and,
      anon_sym_in,
  [192753] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7292), 1,
      anon_sym_STAR,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4660), 2,
      sym_attribute,
      aux_sym__constructor_argument_repeat1,
    ACTIONS(4563), 6,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_DASH_GT,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [192776] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7255), 1,
      anon_sym__,
    STATE(4661), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5942), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [192807] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4384), 1,
      anon_sym_LPAREN,
    ACTIONS(4386), 1,
      anon_sym_LBRACK,
    ACTIONS(4390), 1,
      sym__uppercase_identifier,
    STATE(3603), 1,
      sym_constructor_declaration,
    STATE(4662), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4388), 2,
      anon_sym_true,
      anon_sym_false,
    STATE(3440), 2,
      sym__constructor_declaration_name,
      sym__constructor_name,
  [192838] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7255), 1,
      anon_sym__,
    STATE(4663), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5616), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [192869] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7255), 1,
      anon_sym__,
    STATE(4664), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5637), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [192900] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4665), 1,
      sym_attribute,
    STATE(4675), 1,
      aux_sym_type_binding_repeat1,
    STATE(5199), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4820), 2,
      anon_sym_and,
      anon_sym_in,
  [192933] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4499), 1,
      aux_sym_type_binding_repeat1,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4666), 1,
      sym_attribute,
    STATE(5200), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4857), 2,
      anon_sym_and,
      anon_sym_in,
  [192966] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7154), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    STATE(4667), 1,
      sym_attribute,
    STATE(4670), 1,
      aux_sym_type_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4571), 5,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [192993] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4499), 1,
      aux_sym_type_binding_repeat1,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4668), 1,
      sym_attribute,
    STATE(5201), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4770), 2,
      anon_sym_and,
      anon_sym_in,
  [193026] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4669), 1,
      sym_attribute,
    STATE(4677), 1,
      aux_sym_type_binding_repeat1,
    STATE(5202), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4835), 2,
      anon_sym_and,
      anon_sym_in,
  [193059] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7295), 1,
      anon_sym_constraint,
    STATE(3309), 1,
      sym_type_constraint,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4670), 2,
      sym_attribute,
      aux_sym_type_binding_repeat1,
    ACTIONS(4630), 5,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [193084] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7255), 1,
      anon_sym__,
    STATE(4671), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5722), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [193115] = 11,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7118), 1,
      anon_sym_RPAREN,
    ACTIONS(7281), 1,
      anon_sym_COLON2,
    STATE(4672), 1,
      sym_attribute,
    STATE(6352), 1,
      sym__polymorphic_typed,
    STATE(7080), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193150] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7255), 1,
      anon_sym__,
    STATE(4673), 1,
      sym_attribute,
    STATE(4749), 1,
      aux_sym__type_param_repeat1,
    STATE(5082), 1,
      sym_type_variable,
    STATE(5727), 1,
      sym__type_param,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [193181] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7304), 1,
      sym__right_quoted_string_delimiter,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7301), 2,
      sym__null,
      aux_sym_string_content_token2,
    STATE(4674), 2,
      sym_attribute,
      aux_sym_quoted_string_content_repeat1,
    ACTIONS(7298), 4,
      aux_sym_string_content_token1,
      aux_sym_quoted_string_content_token1,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [193206] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4499), 1,
      aux_sym_type_binding_repeat1,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4675), 1,
      sym_attribute,
    STATE(5217), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4808), 2,
      anon_sym_and,
      anon_sym_in,
  [193239] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4676), 1,
      sym_attribute,
    STATE(4679), 1,
      aux_sym_type_binding_repeat1,
    STATE(5218), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4828), 2,
      anon_sym_and,
      anon_sym_in,
  [193272] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4499), 1,
      aux_sym_type_binding_repeat1,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4677), 1,
      sym_attribute,
    STATE(5220), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4843), 2,
      anon_sym_and,
      anon_sym_in,
  [193305] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7306), 1,
      anon_sym_EQ,
    ACTIONS(7308), 1,
      anon_sym_COLON2,
    ACTIONS(7312), 1,
      anon_sym_COLON_GT,
    STATE(4224), 1,
      sym__coerced,
    STATE(4678), 1,
      sym_attribute,
    STATE(5230), 1,
      sym__typed,
    STATE(5678), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7310), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [193338] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4499), 1,
      aux_sym_type_binding_repeat1,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4679), 1,
      sym_attribute,
    STATE(5257), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4752), 2,
      anon_sym_and,
      anon_sym_in,
  [193371] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7283), 1,
      anon_sym_and,
    STATE(4642), 1,
      aux_sym_module_type_constraint_repeat1,
    STATE(4680), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4861), 6,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [193396] = 10,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7061), 1,
      anon_sym_constraint,
    STATE(4499), 1,
      aux_sym_type_binding_repeat1,
    STATE(4643), 1,
      sym_type_constraint,
    STATE(4681), 1,
      sym_attribute,
    STATE(5108), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4774), 2,
      anon_sym_and,
      anon_sym_in,
  [193429] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4682), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4709), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [193449] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7314), 1,
      anon_sym_EQ,
    ACTIONS(7316), 1,
      anon_sym_COLON_EQ,
    STATE(4683), 1,
      sym_attribute,
    STATE(5064), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4989), 2,
      anon_sym_and,
      anon_sym_in,
  [193479] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7318), 1,
      anon_sym_DASH_GT,
    ACTIONS(7320), 1,
      anon_sym_with,
    STATE(4684), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5064), 5,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [193503] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4685), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4713), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [193523] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4686), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4705), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [193543] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7322), 1,
      anon_sym_RPAREN,
    ACTIONS(7324), 1,
      anon_sym_COMMA,
    ACTIONS(7326), 1,
      anon_sym_PIPE,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7330), 1,
      anon_sym_as,
    ACTIONS(7332), 1,
      anon_sym_COLON_COLON,
    STATE(4687), 1,
      sym_attribute,
    STATE(7026), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193575] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7334), 1,
      anon_sym_DASH_GT,
    ACTIONS(7336), 1,
      anon_sym_when,
    STATE(4688), 1,
      sym_attribute,
    STATE(7195), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193607] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7338), 1,
      anon_sym_RBRACK,
    ACTIONS(7340), 1,
      anon_sym_when,
    STATE(4689), 1,
      sym_attribute,
    STATE(7141), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193639] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7342), 1,
      anon_sym_RPAREN,
    STATE(4690), 1,
      sym_attribute,
    STATE(6663), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193671] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7344), 1,
      anon_sym_COMMA,
    ACTIONS(7346), 1,
      anon_sym_PIPE,
    ACTIONS(7348), 1,
      anon_sym_COLON_COLON,
    STATE(4691), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7100), 3,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [193699] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7350), 1,
      anon_sym_COMMA,
    ACTIONS(7352), 1,
      anon_sym_PIPE,
    ACTIONS(7354), 1,
      anon_sym_SEMI,
    ACTIONS(7356), 1,
      anon_sym_as,
    ACTIONS(7358), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7360), 1,
      anon_sym_PIPE_RBRACK,
    STATE(4692), 1,
      sym_attribute,
    STATE(5626), 1,
      aux_sym__sequence_binding_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193731] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1709), 1,
      anon_sym_PIPE,
    STATE(4693), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1711), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [193753] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7362), 1,
      sym__lowercase_identifier,
    ACTIONS(7364), 1,
      anon_sym__,
    ACTIONS(7366), 1,
      anon_sym_RBRACE,
    STATE(4694), 1,
      sym_attribute,
    STATE(5069), 1,
      sym_field_path,
    STATE(6202), 1,
      sym_field_binding_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193785] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7336), 1,
      anon_sym_when,
    ACTIONS(7368), 1,
      anon_sym_DASH_GT,
    STATE(4695), 1,
      sym_attribute,
    STATE(7306), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193817] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7051), 1,
      anon_sym_PIPE,
    ACTIONS(7370), 1,
      anon_sym_COLON,
    STATE(4696), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7047), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [193841] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7324), 1,
      anon_sym_COMMA,
    ACTIONS(7326), 1,
      anon_sym_PIPE,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7330), 1,
      anon_sym_as,
    ACTIONS(7332), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7372), 1,
      anon_sym_RPAREN,
    STATE(4697), 1,
      sym_attribute,
    STATE(7334), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193873] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7374), 1,
      anon_sym_COMMA,
    ACTIONS(7376), 1,
      anon_sym_PIPE,
    ACTIONS(7378), 1,
      anon_sym_SEMI,
    ACTIONS(7380), 1,
      anon_sym_as,
    ACTIONS(7382), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7384), 1,
      anon_sym_PIPE_RBRACK,
    STATE(4698), 1,
      sym_attribute,
    STATE(5685), 1,
      aux_sym__sequence_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [193905] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4867), 1,
      anon_sym_COLON2,
    STATE(4699), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4865), 6,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [193927] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7318), 1,
      anon_sym_DASH_GT,
    ACTIONS(7320), 1,
      anon_sym_with,
    STATE(4700), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5060), 5,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [193951] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7386), 1,
      anon_sym_rec,
    ACTIONS(7388), 1,
      anon_sym_type,
    STATE(4701), 1,
      sym_attribute,
    STATE(5424), 1,
      sym__attribute,
    STATE(5570), 1,
      sym_module_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7390), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [193981] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4702), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4503), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [194001] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7392), 1,
      anon_sym_COMMA,
    ACTIONS(7394), 1,
      anon_sym_PIPE,
    ACTIONS(7396), 1,
      anon_sym_as,
    ACTIONS(7398), 1,
      anon_sym_COLON_COLON,
    STATE(4703), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3084), 3,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [194029] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7362), 1,
      sym__lowercase_identifier,
    ACTIONS(7400), 1,
      anon_sym__,
    ACTIONS(7402), 1,
      anon_sym_RBRACE,
    STATE(4704), 1,
      sym_attribute,
    STATE(5069), 1,
      sym_field_path,
    STATE(6202), 1,
      sym_field_binding_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194061] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7362), 1,
      sym__lowercase_identifier,
    ACTIONS(7404), 1,
      anon_sym__,
    ACTIONS(7406), 1,
      anon_sym_RBRACE,
    STATE(4705), 1,
      sym_attribute,
    STATE(5141), 1,
      sym_field_path,
    STATE(6332), 1,
      sym_field_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194093] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7362), 1,
      sym__lowercase_identifier,
    ACTIONS(7408), 1,
      anon_sym__,
    ACTIONS(7410), 1,
      anon_sym_RBRACE,
    STATE(4706), 1,
      sym_attribute,
    STATE(5069), 1,
      sym_field_path,
    STATE(6202), 1,
      sym_field_binding_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194125] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7362), 1,
      sym__lowercase_identifier,
    ACTIONS(7412), 1,
      anon_sym__,
    ACTIONS(7414), 1,
      anon_sym_RBRACE,
    STATE(4707), 1,
      sym_attribute,
    STATE(5069), 1,
      sym_field_path,
    STATE(6202), 1,
      sym_field_binding_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194157] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7392), 1,
      anon_sym_COMMA,
    ACTIONS(7394), 1,
      anon_sym_PIPE,
    ACTIONS(7396), 1,
      anon_sym_as,
    ACTIONS(7398), 1,
      anon_sym_COLON_COLON,
    STATE(4708), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2990), 3,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [194185] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7362), 1,
      sym__lowercase_identifier,
    ACTIONS(7416), 1,
      anon_sym__,
    ACTIONS(7418), 1,
      anon_sym_RBRACE,
    STATE(4709), 1,
      sym_attribute,
    STATE(5069), 1,
      sym_field_path,
    STATE(6202), 1,
      sym_field_binding_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194217] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7362), 1,
      sym__lowercase_identifier,
    ACTIONS(7420), 1,
      anon_sym__,
    ACTIONS(7422), 1,
      anon_sym_RBRACE,
    STATE(4710), 1,
      sym_attribute,
    STATE(5141), 1,
      sym_field_path,
    STATE(6332), 1,
      sym_field_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194249] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7392), 1,
      anon_sym_COMMA,
    ACTIONS(7398), 1,
      anon_sym_COLON_COLON,
    STATE(4711), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2986), 5,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
  [194273] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7392), 1,
      anon_sym_COMMA,
    STATE(4712), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2986), 6,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [194295] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7392), 1,
      anon_sym_COMMA,
    ACTIONS(7394), 1,
      anon_sym_PIPE,
    ACTIONS(7398), 1,
      anon_sym_COLON_COLON,
    STATE(4713), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3002), 4,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
  [194321] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7398), 1,
      anon_sym_COLON_COLON,
    STATE(4714), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3006), 6,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
  [194343] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7362), 1,
      sym__lowercase_identifier,
    ACTIONS(7424), 1,
      anon_sym__,
    ACTIONS(7426), 1,
      anon_sym_RBRACE,
    STATE(4715), 1,
      sym_attribute,
    STATE(5141), 1,
      sym_field_path,
    STATE(6332), 1,
      sym_field_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194375] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(5779), 1,
      anon_sym_STAR,
    ACTIONS(7428), 1,
      anon_sym_DASH_GT,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(4716), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194407] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2930), 1,
      anon_sym_PIPE,
    STATE(4717), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2932), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [194429] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7430), 1,
      sym__lowercase_identifier,
    ACTIONS(7432), 1,
      anon_sym_type,
    ACTIONS(7434), 1,
      anon_sym_LBRACK,
    ACTIONS(7436), 1,
      anon_sym_virtual,
    STATE(4718), 1,
      sym_attribute,
    STATE(5538), 1,
      sym__attribute,
    STATE(5582), 1,
      sym_class_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194461] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(4719), 1,
      sym_attribute,
    STATE(5069), 1,
      sym_field_path,
    STATE(5639), 1,
      sym_field_binding_pattern,
    STATE(6997), 1,
      sym_module_path,
    STATE(7307), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194493] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7344), 1,
      anon_sym_COMMA,
    ACTIONS(7346), 1,
      anon_sym_PIPE,
    ACTIONS(7348), 1,
      anon_sym_COLON_COLON,
    STATE(4720), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7126), 3,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [194521] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(4721), 1,
      sym_attribute,
    STATE(5069), 1,
      sym_field_path,
    STATE(5609), 1,
      sym_field_binding_pattern,
    STATE(6997), 1,
      sym_module_path,
    STATE(7184), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194553] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(4722), 1,
      sym_attribute,
    STATE(5141), 1,
      sym_field_path,
    STATE(5612), 1,
      sym_field_pattern,
    STATE(6997), 1,
      sym_module_path,
    STATE(7184), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194585] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(4723), 1,
      sym_attribute,
    STATE(5141), 1,
      sym_field_path,
    STATE(5652), 1,
      sym_field_pattern,
    STATE(6532), 1,
      sym__quoted_string,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194617] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(5779), 1,
      anon_sym_STAR,
    ACTIONS(7438), 1,
      anon_sym_DASH_GT,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(4724), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194649] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2081), 1,
      anon_sym_PIPE,
    ACTIONS(7440), 1,
      anon_sym_DOT_DOT,
    STATE(4725), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2075), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [194673] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7336), 1,
      anon_sym_when,
    ACTIONS(7442), 1,
      anon_sym_DASH_GT,
    STATE(4726), 1,
      sym_attribute,
    STATE(6615), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194705] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7444), 1,
      sym__lowercase_identifier,
    ACTIONS(7446), 1,
      anon_sym_type,
    ACTIONS(7448), 1,
      anon_sym_LBRACK,
    ACTIONS(7450), 1,
      anon_sym_virtual,
    STATE(4142), 1,
      sym_class_binding,
    STATE(4727), 1,
      sym_attribute,
    STATE(5410), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194737] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7344), 1,
      anon_sym_COMMA,
    ACTIONS(7348), 1,
      anon_sym_COLON_COLON,
    STATE(4728), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6837), 5,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
  [194761] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7452), 1,
      anon_sym_RPAREN,
    STATE(4729), 1,
      sym_attribute,
    STATE(6422), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194793] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7344), 1,
      anon_sym_COMMA,
    STATE(4730), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6837), 6,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [194815] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7454), 1,
      anon_sym_rec,
    ACTIONS(7456), 1,
      anon_sym_type,
    STATE(3908), 1,
      sym_module_binding,
    STATE(4731), 1,
      sym_attribute,
    STATE(5268), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7458), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [194845] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7460), 1,
      anon_sym_RPAREN,
    STATE(4732), 1,
      sym_attribute,
    STATE(6435), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194877] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(5779), 1,
      anon_sym_STAR,
    ACTIONS(7462), 1,
      anon_sym_DASH_GT,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(4733), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194909] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7344), 1,
      anon_sym_COMMA,
    ACTIONS(7346), 1,
      anon_sym_PIPE,
    ACTIONS(7348), 1,
      anon_sym_COLON_COLON,
    STATE(4734), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6851), 4,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
  [194935] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7464), 1,
      anon_sym_EQ,
    ACTIONS(7466), 1,
      anon_sym_COLON_EQ,
    STATE(4735), 1,
      sym_attribute,
    STATE(5107), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5029), 2,
      anon_sym_and,
      anon_sym_in,
  [194965] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(4736), 1,
      sym_attribute,
    STATE(5069), 1,
      sym_field_path,
    STATE(5699), 1,
      sym_field_binding_pattern,
    STATE(6908), 1,
      sym__quoted_string,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [194997] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2936), 1,
      anon_sym_COLON2,
    ACTIONS(7468), 1,
      anon_sym_COLON,
    STATE(4737), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2938), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [195021] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(4738), 1,
      sym_attribute,
    STATE(5141), 1,
      sym_field_path,
    STATE(5701), 1,
      sym_field_pattern,
    STATE(6531), 1,
      sym__quoted_string,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195053] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1297), 1,
      anon_sym_PIPE,
    ACTIONS(7440), 1,
      anon_sym_DOT_DOT,
    STATE(4739), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1295), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [195077] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7470), 1,
      anon_sym_RPAREN,
    STATE(4740), 1,
      sym_attribute,
    STATE(6526), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195109] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7472), 1,
      anon_sym_RPAREN,
    STATE(4741), 1,
      sym_attribute,
    STATE(6535), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195141] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7362), 1,
      sym__lowercase_identifier,
    ACTIONS(7474), 1,
      anon_sym__,
    ACTIONS(7476), 1,
      anon_sym_RBRACE,
    STATE(4742), 1,
      sym_attribute,
    STATE(5141), 1,
      sym_field_path,
    STATE(6332), 1,
      sym_field_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195173] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7478), 1,
      anon_sym_RPAREN,
    STATE(4743), 1,
      sym_attribute,
    STATE(6977), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195205] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7348), 1,
      anon_sym_COLON_COLON,
    STATE(4744), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7112), 6,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
  [195227] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7362), 1,
      sym__lowercase_identifier,
    ACTIONS(7480), 1,
      anon_sym__,
    ACTIONS(7482), 1,
      anon_sym_RBRACE,
    STATE(4745), 1,
      sym_attribute,
    STATE(5069), 1,
      sym_field_path,
    STATE(6202), 1,
      sym_field_binding_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195259] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7484), 1,
      anon_sym_RPAREN,
    STATE(4746), 1,
      sym_attribute,
    STATE(6579), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195291] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4747), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4721), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [195311] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7486), 1,
      anon_sym_RPAREN,
    STATE(4748), 1,
      sym_attribute,
    STATE(6585), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195343] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7488), 1,
      anon_sym__,
    STATE(4749), 1,
      sym_attribute,
    STATE(4880), 1,
      aux_sym__type_param_repeat1,
    STATE(5154), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6379), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [195371] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7490), 1,
      anon_sym_RPAREN,
    STATE(4750), 1,
      sym_attribute,
    STATE(6622), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195403] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7492), 1,
      anon_sym_RPAREN,
    STATE(4751), 1,
      sym_attribute,
    STATE(6628), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195435] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7336), 1,
      anon_sym_when,
    ACTIONS(7494), 1,
      anon_sym_DASH_GT,
    STATE(4752), 1,
      sym_attribute,
    STATE(6919), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195467] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7362), 1,
      sym__lowercase_identifier,
    ACTIONS(7496), 1,
      anon_sym__,
    ACTIONS(7498), 1,
      anon_sym_RBRACE,
    STATE(4753), 1,
      sym_attribute,
    STATE(5141), 1,
      sym_field_path,
    STATE(6332), 1,
      sym_field_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195499] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(5779), 1,
      anon_sym_STAR,
    ACTIONS(7500), 1,
      anon_sym_DASH_GT,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(4754), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195531] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7502), 1,
      anon_sym_RPAREN,
    STATE(4755), 1,
      sym_attribute,
    STATE(6662), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195563] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7504), 1,
      anon_sym_LBRACE,
    ACTIONS(7507), 1,
      anon_sym_DQUOTE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4921), 2,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
    STATE(4756), 2,
      sym_attribute,
      aux_sym_external_repeat1,
    STATE(5492), 2,
      sym_string,
      sym_quoted_string,
  [195589] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4691), 1,
      sym__lowercase_identifier,
    ACTIONS(6667), 1,
      anon_sym_POUND,
    STATE(4480), 1,
      sym_type_constructor_path,
    STATE(4757), 1,
      sym_attribute,
    STATE(6026), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7510), 2,
      anon_sym_SEMI,
      anon_sym_GT,
  [195619] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7512), 1,
      anon_sym_RPAREN,
    STATE(4758), 1,
      sym_attribute,
    STATE(6667), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195651] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7514), 1,
      anon_sym_RPAREN,
    STATE(4759), 1,
      sym_attribute,
    STATE(6642), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195683] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4798), 1,
      anon_sym_COLON2,
    STATE(4760), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4796), 6,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [195705] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7516), 1,
      anon_sym_RPAREN,
    STATE(4761), 1,
      sym_attribute,
    STATE(6701), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195737] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7336), 1,
      anon_sym_when,
    ACTIONS(7518), 1,
      anon_sym_DASH_GT,
    STATE(4762), 1,
      sym_attribute,
    STATE(6546), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195769] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7520), 1,
      anon_sym_RPAREN,
    STATE(4763), 1,
      sym_attribute,
    STATE(6706), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195801] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1733), 1,
      anon_sym_PIPE,
    STATE(4764), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1735), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [195823] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7522), 1,
      anon_sym_RPAREN,
    STATE(4765), 1,
      sym_attribute,
    STATE(6738), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195855] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7524), 1,
      anon_sym_RPAREN,
    STATE(4766), 1,
      sym_attribute,
    STATE(6743), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195887] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(4767), 1,
      sym_attribute,
    STATE(5141), 1,
      sym_field_path,
    STATE(5844), 1,
      sym_field_pattern,
    STATE(6997), 1,
      sym_module_path,
    STATE(7307), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195919] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7526), 1,
      anon_sym_RPAREN,
    STATE(4768), 1,
      sym_attribute,
    STATE(6722), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195951] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7528), 1,
      anon_sym_RPAREN,
    STATE(4769), 1,
      sym_attribute,
    STATE(6774), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [195983] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7530), 1,
      anon_sym_RPAREN,
    STATE(4770), 1,
      sym_attribute,
    STATE(6779), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196015] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7344), 1,
      anon_sym_COMMA,
    ACTIONS(7346), 1,
      anon_sym_PIPE,
    ACTIONS(7348), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7384), 1,
      anon_sym_RBRACK,
    ACTIONS(7532), 1,
      anon_sym_SEMI,
    STATE(4771), 1,
      sym_attribute,
    STATE(5647), 1,
      aux_sym__sequence_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196047] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7534), 1,
      anon_sym_RPAREN,
    STATE(4772), 1,
      sym_attribute,
    STATE(6800), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196079] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7536), 1,
      anon_sym_RPAREN,
    STATE(4773), 1,
      sym_attribute,
    STATE(6802), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196111] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7324), 1,
      anon_sym_COMMA,
    ACTIONS(7326), 1,
      anon_sym_PIPE,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7330), 1,
      anon_sym_as,
    ACTIONS(7332), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7538), 1,
      anon_sym_RPAREN,
    STATE(4774), 1,
      sym_attribute,
    STATE(6616), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196143] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7336), 1,
      anon_sym_when,
    ACTIONS(7540), 1,
      anon_sym_DASH_GT,
    STATE(4775), 1,
      sym_attribute,
    STATE(7062), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196175] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7542), 1,
      anon_sym_RPAREN,
    STATE(4776), 1,
      sym_attribute,
    STATE(6741), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196207] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1817), 1,
      anon_sym_PIPE,
    STATE(4777), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1819), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [196229] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7362), 1,
      sym__lowercase_identifier,
    ACTIONS(7544), 1,
      anon_sym__,
    ACTIONS(7546), 1,
      anon_sym_RBRACE,
    STATE(4778), 1,
      sym_attribute,
    STATE(5069), 1,
      sym_field_path,
    STATE(6202), 1,
      sym_field_binding_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196261] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(7500), 1,
      anon_sym_DASH_GT,
    ACTIONS(7548), 1,
      anon_sym_STAR,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(4779), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196293] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2936), 1,
      anon_sym_PIPE,
    ACTIONS(7551), 1,
      anon_sym_COLON,
    STATE(4780), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2938), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [196317] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7324), 1,
      anon_sym_COMMA,
    ACTIONS(7326), 1,
      anon_sym_PIPE,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7330), 1,
      anon_sym_as,
    ACTIONS(7332), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7553), 1,
      anon_sym_RPAREN,
    STATE(4781), 1,
      sym_attribute,
    STATE(6455), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196349] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7362), 1,
      sym__lowercase_identifier,
    ACTIONS(7555), 1,
      anon_sym__,
    ACTIONS(7557), 1,
      anon_sym_RBRACE,
    STATE(4782), 1,
      sym_attribute,
    STATE(5141), 1,
      sym_field_path,
    STATE(6332), 1,
      sym_field_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196381] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(5779), 1,
      anon_sym_STAR,
    ACTIONS(7559), 1,
      anon_sym_DASH_GT,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(4783), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196413] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7336), 1,
      anon_sym_when,
    ACTIONS(7561), 1,
      anon_sym_DASH_GT,
    STATE(4784), 1,
      sym_attribute,
    STATE(7266), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196445] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(5779), 1,
      anon_sym_STAR,
    ACTIONS(7563), 1,
      anon_sym_DASH_GT,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(4785), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196477] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1683), 1,
      anon_sym_PIPE,
    STATE(4786), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1685), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [196499] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2942), 1,
      anon_sym_PIPE,
    STATE(4787), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2944), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [196521] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7565), 1,
      anon_sym_COLON2,
    STATE(4788), 1,
      sym_attribute,
    STATE(6112), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3052), 5,
      anon_sym_SEMI,
      anon_sym_GT,
      anon_sym_POUND,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [196545] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(4789), 1,
      sym_attribute,
    STATE(5069), 1,
      sym_field_path,
    STATE(5675), 1,
      sym_field_binding_pattern,
    STATE(6434), 1,
      sym__quoted_string,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196577] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    STATE(4790), 1,
      sym_attribute,
    STATE(6678), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4507), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4503), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
  [196603] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1687), 1,
      anon_sym_PIPE,
    STATE(4791), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1689), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [196625] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6795), 1,
      anon_sym_DASH_GT,
    STATE(4792), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3987), 6,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
      anon_sym_as,
      anon_sym_GT,
  [196647] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7360), 1,
      anon_sym_RBRACK,
    ACTIONS(7392), 1,
      anon_sym_COMMA,
    ACTIONS(7394), 1,
      anon_sym_PIPE,
    ACTIONS(7396), 1,
      anon_sym_as,
    ACTIONS(7398), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7567), 1,
      anon_sym_SEMI,
    STATE(4793), 1,
      sym_attribute,
    STATE(5634), 1,
      aux_sym__sequence_binding_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196679] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7362), 1,
      sym__lowercase_identifier,
    ACTIONS(7569), 1,
      anon_sym__,
    ACTIONS(7571), 1,
      anon_sym_RBRACE,
    STATE(4794), 1,
      sym_attribute,
    STATE(5069), 1,
      sym_field_path,
    STATE(6202), 1,
      sym_field_binding_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196711] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1829), 1,
      anon_sym_PIPE,
    STATE(4795), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1831), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [196733] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4796), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4735), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [196753] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7362), 1,
      sym__lowercase_identifier,
    ACTIONS(7573), 1,
      anon_sym__,
    ACTIONS(7575), 1,
      anon_sym_RBRACE,
    STATE(4797), 1,
      sym_attribute,
    STATE(5141), 1,
      sym_field_path,
    STATE(6332), 1,
      sym_field_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196785] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1699), 1,
      anon_sym_PIPE,
    ACTIONS(1703), 1,
      anon_sym_DOT,
    STATE(4798), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1701), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [196809] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4841), 1,
      anon_sym_COLON2,
    STATE(4799), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4839), 6,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_LPAREN,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [196831] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4800), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4725), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [196851] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7318), 1,
      anon_sym_DASH_GT,
    ACTIONS(7320), 1,
      anon_sym_with,
    STATE(4801), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5052), 5,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [196875] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7362), 1,
      sym__lowercase_identifier,
    ACTIONS(7577), 1,
      anon_sym__,
    ACTIONS(7579), 1,
      anon_sym_RBRACE,
    STATE(4802), 1,
      sym_attribute,
    STATE(5141), 1,
      sym_field_path,
    STATE(6332), 1,
      sym_field_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196907] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1833), 1,
      anon_sym_PIPE,
    STATE(4803), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1835), 6,
      anon_sym_DOT_DOT,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [196929] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4804), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1747), 7,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [196949] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7446), 1,
      anon_sym_type,
    ACTIONS(7581), 1,
      sym__lowercase_identifier,
    ACTIONS(7583), 1,
      anon_sym_LBRACK,
    ACTIONS(7585), 1,
      anon_sym_virtual,
    STATE(3858), 1,
      sym_class_binding,
    STATE(4805), 1,
      sym_attribute,
    STATE(5269), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [196981] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7336), 1,
      anon_sym_when,
    ACTIONS(7587), 1,
      anon_sym_DASH_GT,
    STATE(4806), 1,
      sym_attribute,
    STATE(6927), 1,
      sym_guard,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197013] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7318), 1,
      anon_sym_DASH_GT,
    ACTIONS(7320), 1,
      anon_sym_with,
    STATE(4807), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5056), 5,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [197037] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4808), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1849), 7,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [197057] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4809), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4529), 7,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [197077] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7318), 1,
      anon_sym_DASH_GT,
    ACTIONS(7320), 1,
      anon_sym_with,
    STATE(4810), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4964), 5,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [197101] = 10,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7589), 1,
      anon_sym_RPAREN,
    STATE(4811), 1,
      sym_attribute,
    STATE(6721), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197133] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1699), 1,
      anon_sym_PIPE,
    ACTIONS(1841), 1,
      anon_sym_DOT,
    STATE(4812), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1701), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197157] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7591), 1,
      anon_sym_rec,
    ACTIONS(7593), 1,
      anon_sym_type,
    STATE(4171), 1,
      sym_module_binding,
    STATE(4813), 1,
      sym_attribute,
    STATE(5497), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7595), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [197187] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(4814), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7599), 3,
      sym__right_quoted_string_delimiter,
      sym__null,
      aux_sym_string_content_token2,
    ACTIONS(7597), 4,
      aux_sym_string_content_token1,
      aux_sym_quoted_string_content_token1,
      sym_conversion_specification,
      sym_pretty_printing_indication,
  [197209] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6943), 1,
      anon_sym_PIPE,
    STATE(4815), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6941), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197230] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7601), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4816), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197259] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3146), 1,
      anon_sym_PIPE,
    STATE(4817), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3148), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197280] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7603), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4818), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197309] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1825), 1,
      anon_sym_PIPE,
    STATE(4819), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1827), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197330] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7605), 1,
      sym__lowercase_identifier,
    ACTIONS(7607), 1,
      anon_sym_LBRACK,
    ACTIONS(7609), 1,
      anon_sym_virtual,
    STATE(4820), 1,
      sym_attribute,
    STATE(5365), 1,
      sym__attribute,
    STATE(5784), 1,
      sym_class_type_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197359] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7611), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4821), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197388] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7350), 1,
      anon_sym_COMMA,
    ACTIONS(7352), 1,
      anon_sym_PIPE,
    ACTIONS(7356), 1,
      anon_sym_as,
    ACTIONS(7358), 1,
      anon_sym_COLON_COLON,
    STATE(4822), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3084), 2,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
  [197415] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3150), 1,
      anon_sym_PIPE,
    STATE(4823), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3152), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197436] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7613), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4824), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197465] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7615), 1,
      sym__lowercase_identifier,
    ACTIONS(7617), 1,
      anon_sym_BANG,
    STATE(4054), 1,
      sym__instance_variable_name,
    STATE(4825), 1,
      sym_attribute,
    STATE(5091), 1,
      aux_sym_instance_variable_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7619), 2,
      anon_sym_mutable,
      anon_sym_virtual,
  [197492] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7023), 1,
      anon_sym_PIPE,
    STATE(4826), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7021), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197513] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1745), 1,
      anon_sym_PIPE,
    STATE(4827), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1747), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197534] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3090), 1,
      anon_sym_PIPE,
    STATE(4828), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3092), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197555] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5208), 1,
      anon_sym_in,
    STATE(4829), 1,
      sym_attribute,
    STATE(5309), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7621), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [197582] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6993), 1,
      anon_sym_PIPE,
    STATE(4830), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6991), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197603] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3098), 1,
      anon_sym_PIPE,
    STATE(4831), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3100), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197624] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6849), 1,
      anon_sym_PIPE,
    STATE(4832), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6847), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197645] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7007), 1,
      anon_sym_PIPE,
    STATE(4833), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7005), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197666] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7623), 1,
      anon_sym_and,
    STATE(4834), 1,
      sym_attribute,
    STATE(4877), 1,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4847), 4,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [197689] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3102), 1,
      anon_sym_PIPE,
    STATE(4835), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3104), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197710] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3482), 1,
      anon_sym_PIPE,
    STATE(4836), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3484), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197731] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3474), 1,
      anon_sym_PIPE,
    STATE(4837), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3476), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197752] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3478), 1,
      anon_sym_PIPE,
    STATE(4838), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3480), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197773] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3490), 1,
      anon_sym_PIPE,
    STATE(4839), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3492), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197794] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3494), 1,
      anon_sym_PIPE,
    STATE(4840), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3496), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197815] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2988), 1,
      anon_sym_COLON2,
    ACTIONS(2990), 1,
      anon_sym_RPAREN,
    ACTIONS(7324), 1,
      anon_sym_COMMA,
    ACTIONS(7326), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_as,
    ACTIONS(7332), 1,
      anon_sym_COLON_COLON,
    STATE(4841), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197844] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7374), 1,
      anon_sym_COMMA,
    ACTIONS(7376), 1,
      anon_sym_PIPE,
    ACTIONS(7380), 1,
      anon_sym_as,
    ACTIONS(7382), 1,
      anon_sym_COLON_COLON,
    STATE(4842), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7126), 2,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
  [197871] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3502), 1,
      anon_sym_PIPE,
    STATE(4843), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3504), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197892] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2984), 1,
      anon_sym_COLON2,
    ACTIONS(7324), 1,
      anon_sym_COMMA,
    ACTIONS(7332), 1,
      anon_sym_COLON_COLON,
    STATE(4844), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2986), 3,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_as,
  [197917] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7625), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4845), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [197946] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3118), 1,
      anon_sym_PIPE,
    STATE(4846), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3120), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197967] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1785), 1,
      anon_sym_PIPE,
    STATE(4847), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1787), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [197988] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3122), 1,
      anon_sym_PIPE,
    STATE(4848), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3124), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198009] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7627), 1,
      anon_sym_EQ,
    STATE(4849), 1,
      sym_attribute,
    STATE(5068), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5104), 2,
      anon_sym_and,
      anon_sym_in,
  [198036] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3126), 1,
      anon_sym_PIPE,
    STATE(4850), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3128), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198057] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7629), 1,
      anon_sym_EQ,
    STATE(4851), 1,
      sym_attribute,
    STATE(5075), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5232), 2,
      anon_sym_and,
      anon_sym_in,
  [198084] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2984), 1,
      anon_sym_COLON2,
    ACTIONS(7324), 1,
      anon_sym_COMMA,
    STATE(4852), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2986), 4,
      anon_sym_RPAREN,
      anon_sym_PIPE,
      anon_sym_as,
      anon_sym_COLON_COLON,
  [198107] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1797), 1,
      anon_sym_PIPE,
    STATE(4853), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1799), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198128] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7631), 1,
      sym__lowercase_identifier,
    STATE(1901), 1,
      sym_class_path,
    STATE(4854), 1,
      sym_attribute,
    STATE(5385), 1,
      sym__attribute,
    STATE(6503), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [198157] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3000), 1,
      anon_sym_COLON2,
    ACTIONS(7324), 1,
      anon_sym_COMMA,
    ACTIONS(7326), 1,
      anon_sym_PIPE,
    ACTIONS(7332), 1,
      anon_sym_COLON_COLON,
    STATE(4855), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3002), 2,
      anon_sym_RPAREN,
      anon_sym_as,
  [198184] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3058), 1,
      anon_sym_PIPE,
    STATE(4856), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3060), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198205] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1805), 1,
      anon_sym_PIPE,
    STATE(4857), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1807), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198226] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1757), 1,
      anon_sym_PIPE,
    STATE(4858), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1759), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198247] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3050), 1,
      anon_sym_PIPE,
    STATE(4859), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3052), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198268] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6839), 1,
      anon_sym_PIPE,
    ACTIONS(7374), 1,
      anon_sym_COMMA,
    ACTIONS(7382), 1,
      anon_sym_COLON_COLON,
    STATE(4860), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6837), 3,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_PIPE_RBRACK,
  [198293] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7633), 1,
      anon_sym_RPAREN,
    STATE(4224), 1,
      sym__coerced,
    STATE(4861), 1,
      sym_attribute,
    STATE(5928), 1,
      sym__typed,
    STATE(7260), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [198322] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3004), 1,
      anon_sym_COLON2,
    ACTIONS(7332), 1,
      anon_sym_COLON_COLON,
    STATE(4862), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3006), 4,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
      anon_sym_as,
  [198345] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1847), 1,
      anon_sym_PIPE,
    STATE(4863), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1849), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198366] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2392), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4958), 1,
      anon_sym_and,
    STATE(4864), 1,
      sym_attribute,
    STATE(4906), 1,
      aux_sym_expression_item_repeat1,
    STATE(5502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4956), 2,
      anon_sym_in,
      sym_let_and_operator,
  [198393] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3078), 1,
      anon_sym_PIPE,
    STATE(4865), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3080), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198414] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7392), 1,
      anon_sym_COMMA,
    ACTIONS(7394), 1,
      anon_sym_PIPE,
    ACTIONS(7396), 1,
      anon_sym_as,
    ACTIONS(7398), 1,
      anon_sym_COLON_COLON,
    STATE(4866), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7635), 2,
      anon_sym_RBRACK,
      anon_sym_SEMI,
  [198441] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6839), 1,
      anon_sym_PIPE,
    ACTIONS(7374), 1,
      anon_sym_COMMA,
    STATE(4867), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6837), 4,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198464] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7374), 1,
      anon_sym_COMMA,
    ACTIONS(7376), 1,
      anon_sym_PIPE,
    ACTIONS(7380), 1,
      anon_sym_as,
    ACTIONS(7382), 1,
      anon_sym_COLON_COLON,
    STATE(4868), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7100), 2,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
  [198491] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7637), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4869), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [198520] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7639), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4870), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [198549] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(3632), 1,
      anon_sym_POUND,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(7641), 1,
      anon_sym_DASH_GT,
    STATE(2573), 1,
      sym_type_constructor_path,
    STATE(4871), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [198578] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3460), 1,
      anon_sym_PIPE,
    STATE(4872), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3462), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198599] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7643), 1,
      anon_sym_COLON2,
    STATE(4873), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4507), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4503), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
  [198622] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6883), 1,
      anon_sym_PIPE,
    STATE(4874), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6881), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198643] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7374), 1,
      anon_sym_COMMA,
    ACTIONS(7376), 1,
      anon_sym_PIPE,
    ACTIONS(7382), 1,
      anon_sym_COLON_COLON,
    STATE(4875), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6851), 3,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_PIPE_RBRACK,
  [198668] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7645), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4876), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [198697] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7647), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4877), 2,
      sym_attribute,
      aux_sym_module_type_constraint_repeat1,
    ACTIONS(4721), 4,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [198718] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6903), 1,
      anon_sym_PIPE,
    STATE(4878), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6901), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198739] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6911), 1,
      anon_sym_PIPE,
    STATE(4879), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6909), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198760] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7653), 2,
      anon_sym__,
      aux_sym_type_variable_token1,
    STATE(4880), 2,
      sym_attribute,
      aux_sym__type_param_repeat1,
    ACTIONS(7650), 3,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
  [198781] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3170), 1,
      anon_sym_PIPE,
    STATE(4881), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3172), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198802] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7655), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4882), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [198831] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2984), 1,
      anon_sym_PIPE,
    STATE(4883), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2986), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198852] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2984), 1,
      anon_sym_PIPE,
    STATE(4884), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2986), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198873] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3000), 1,
      anon_sym_PIPE,
    STATE(4885), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3002), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198894] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3008), 1,
      anon_sym_PIPE,
    STATE(4886), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3010), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [198915] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7657), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4887), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [198944] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7350), 1,
      anon_sym_COMMA,
    ACTIONS(7352), 1,
      anon_sym_PIPE,
    ACTIONS(7356), 1,
      anon_sym_as,
    ACTIONS(7358), 1,
      anon_sym_COLON_COLON,
    STATE(4888), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7635), 2,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
  [198971] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7114), 1,
      anon_sym_PIPE,
    ACTIONS(7382), 1,
      anon_sym_COLON_COLON,
    STATE(4889), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7112), 4,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_PIPE_RBRACK,
  [198994] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2392), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4999), 1,
      anon_sym_and,
    STATE(4890), 1,
      sym_attribute,
    STATE(4906), 1,
      aux_sym_expression_item_repeat1,
    STATE(5502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4997), 2,
      anon_sym_in,
      sym_let_and_operator,
  [199021] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2392), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5003), 1,
      anon_sym_and,
    STATE(4891), 1,
      sym_attribute,
    STATE(4946), 1,
      aux_sym_expression_item_repeat1,
    STATE(5502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5001), 2,
      anon_sym_in,
      sym_let_and_operator,
  [199048] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7392), 1,
      anon_sym_COMMA,
    ACTIONS(7394), 1,
      anon_sym_PIPE,
    ACTIONS(7396), 1,
      anon_sym_as,
    ACTIONS(7398), 1,
      anon_sym_COLON_COLON,
    STATE(4892), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7659), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [199075] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2392), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5007), 1,
      anon_sym_and,
    STATE(4893), 1,
      sym_attribute,
    STATE(4948), 1,
      aux_sym_expression_item_repeat1,
    STATE(5502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5005), 2,
      anon_sym_in,
      sym_let_and_operator,
  [199102] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3522), 1,
      anon_sym_PIPE,
    STATE(4894), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3524), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199123] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2392), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5011), 1,
      anon_sym_and,
    STATE(4895), 1,
      sym_attribute,
    STATE(4950), 1,
      aux_sym_expression_item_repeat1,
    STATE(5502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5009), 2,
      anon_sym_in,
      sym_let_and_operator,
  [199150] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7661), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4896), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199179] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7033), 1,
      anon_sym_PIPE,
    STATE(4897), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7031), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199200] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7374), 1,
      anon_sym_COMMA,
    ACTIONS(7376), 1,
      anon_sym_PIPE,
    ACTIONS(7380), 1,
      anon_sym_as,
    ACTIONS(7382), 1,
      anon_sym_COLON_COLON,
    STATE(4898), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7663), 2,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
  [199227] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6819), 1,
      anon_sym_PIPE,
    STATE(4899), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6817), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199248] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7665), 1,
      sym__lowercase_identifier,
    ACTIONS(7667), 1,
      anon_sym_LBRACK,
    ACTIONS(7669), 1,
      anon_sym_virtual,
    STATE(3967), 1,
      sym_class_type_binding,
    STATE(4900), 1,
      sym_attribute,
    STATE(5405), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199277] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3086), 1,
      anon_sym_PIPE,
    STATE(4901), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3088), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199298] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7671), 1,
      sym__lowercase_identifier,
    STATE(993), 1,
      sym_class_path,
    STATE(4902), 1,
      sym_attribute,
    STATE(5395), 1,
      sym__attribute,
    STATE(6606), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199327] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7673), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4903), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199356] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7675), 1,
      sym__lowercase_identifier,
    ACTIONS(7677), 1,
      anon_sym_LPAREN,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(4904), 1,
      sym_attribute,
    STATE(5347), 1,
      sym__attribute,
    STATE(5951), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199385] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7017), 1,
      anon_sym_PIPE,
    STATE(4905), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7015), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199406] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4743), 1,
      anon_sym_and,
    ACTIONS(7679), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4741), 2,
      anon_sym_in,
      sym_let_and_operator,
    STATE(4906), 2,
      sym_attribute,
      aux_sym_expression_item_repeat1,
  [199431] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7682), 1,
      anon_sym_RPAREN,
    STATE(4224), 1,
      sym__coerced,
    STATE(4907), 1,
      sym_attribute,
    STATE(5928), 1,
      sym__typed,
    STATE(6597), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199460] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7392), 1,
      anon_sym_COMMA,
    ACTIONS(7394), 1,
      anon_sym_PIPE,
    ACTIONS(7396), 1,
      anon_sym_as,
    ACTIONS(7398), 1,
      anon_sym_COLON_COLON,
    STATE(4908), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7684), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [199487] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7686), 1,
      anon_sym_PIPE,
    STATE(4909), 1,
      sym_attribute,
    STATE(4956), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5163), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [199510] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7688), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4910), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199539] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3536), 1,
      anon_sym_PIPE,
    STATE(4911), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3538), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199560] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7690), 1,
      anon_sym_RPAREN,
    STATE(4224), 1,
      sym__coerced,
    STATE(4912), 1,
      sym_attribute,
    STATE(5928), 1,
      sym__typed,
    STATE(7063), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199589] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7686), 1,
      anon_sym_PIPE,
    STATE(4913), 1,
      sym_attribute,
    STATE(4957), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5163), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [199612] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1297), 1,
      anon_sym_PIPE,
    STATE(4914), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1295), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199633] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1623), 1,
      anon_sym_PIPE,
    STATE(4915), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1625), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199654] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3114), 1,
      anon_sym_PIPE,
    STATE(4916), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3116), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199675] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6839), 1,
      anon_sym_PIPE,
    STATE(4917), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6837), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199696] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2970), 1,
      anon_sym_PIPE,
    STATE(4918), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2972), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199717] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7692), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4919), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199746] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6839), 1,
      anon_sym_PIPE,
    STATE(4920), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6837), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199767] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3158), 1,
      anon_sym_PIPE,
    STATE(4921), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3160), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199788] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3166), 1,
      anon_sym_PIPE,
    STATE(4922), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3168), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [199809] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5204), 1,
      anon_sym_in,
    ACTIONS(7320), 1,
      anon_sym_with,
    ACTIONS(7694), 1,
      anon_sym_DASH_GT,
    STATE(4923), 1,
      sym_attribute,
    STATE(5470), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199838] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5147), 1,
      anon_sym_in,
    STATE(4924), 1,
      sym_attribute,
    STATE(5323), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7696), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [199865] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7698), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4925), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199894] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3082), 1,
      anon_sym_COLON2,
    ACTIONS(3084), 1,
      anon_sym_RPAREN,
    ACTIONS(7324), 1,
      anon_sym_COMMA,
    ACTIONS(7326), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_as,
    ACTIONS(7332), 1,
      anon_sym_COLON_COLON,
    STATE(4926), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199923] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7700), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4927), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [199952] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7702), 1,
      anon_sym_EQ,
    STATE(4928), 1,
      sym_attribute,
    STATE(5135), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5121), 2,
      anon_sym_and,
      anon_sym_in,
  [199979] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7344), 1,
      anon_sym_COMMA,
    ACTIONS(7346), 1,
      anon_sym_PIPE,
    ACTIONS(7348), 1,
      anon_sym_COLON_COLON,
    STATE(4929), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7704), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [200006] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6853), 1,
      anon_sym_PIPE,
    STATE(4930), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6851), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200027] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7706), 1,
      sym__lowercase_identifier,
    STATE(762), 1,
      sym_class_path,
    STATE(4931), 1,
      sym_attribute,
    STATE(5468), 1,
      sym__attribute,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200056] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7344), 1,
      anon_sym_COMMA,
    ACTIONS(7346), 1,
      anon_sym_PIPE,
    ACTIONS(7348), 1,
      anon_sym_COLON_COLON,
    STATE(4932), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7663), 2,
      anon_sym_RBRACK,
      anon_sym_SEMI,
  [200083] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7708), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4933), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200112] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7623), 1,
      anon_sym_and,
    STATE(4834), 1,
      aux_sym_module_type_constraint_repeat1,
    STATE(4934), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4861), 4,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
      anon_sym_COLON_GT,
  [200135] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7710), 1,
      sym__lowercase_identifier,
    STATE(1700), 1,
      sym_class_path,
    STATE(4935), 1,
      sym_attribute,
    STATE(5523), 1,
      sym__attribute,
    STATE(6377), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200164] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7712), 1,
      anon_sym_RPAREN,
    STATE(4224), 1,
      sym__coerced,
    STATE(4936), 1,
      sym_attribute,
    STATE(5928), 1,
      sym__typed,
    STATE(7296), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200193] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1721), 1,
      anon_sym_PIPE,
    STATE(4937), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1723), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200214] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6857), 1,
      anon_sym_PIPE,
    STATE(4938), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6855), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200235] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7714), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4939), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200264] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3130), 1,
      anon_sym_PIPE,
    STATE(4940), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3132), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200285] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7344), 1,
      anon_sym_COMMA,
    ACTIONS(7346), 1,
      anon_sym_PIPE,
    ACTIONS(7348), 1,
      anon_sym_COLON_COLON,
    STATE(4941), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7716), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [200312] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3134), 1,
      anon_sym_PIPE,
    STATE(4942), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3136), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200333] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7718), 1,
      anon_sym_EQ,
    STATE(4943), 1,
      sym_attribute,
    STATE(5060), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5186), 2,
      anon_sym_and,
      anon_sym_in,
  [200360] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3044), 1,
      anon_sym_PIPE,
    STATE(4944), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3046), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200381] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7720), 1,
      anon_sym_RPAREN,
    STATE(4224), 1,
      sym__coerced,
    STATE(4945), 1,
      sym_attribute,
    STATE(5928), 1,
      sym__typed,
    STATE(6903), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200410] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2392), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4899), 1,
      anon_sym_and,
    STATE(4906), 1,
      aux_sym_expression_item_repeat1,
    STATE(4946), 1,
      sym_attribute,
    STATE(5502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4897), 2,
      anon_sym_in,
      sym_let_and_operator,
  [200437] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6869), 1,
      anon_sym_PIPE,
    STATE(4947), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6867), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200458] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2392), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4903), 1,
      anon_sym_and,
    STATE(4906), 1,
      aux_sym_expression_item_repeat1,
    STATE(4948), 1,
      sym_attribute,
    STATE(5502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4901), 2,
      anon_sym_in,
      sym_let_and_operator,
  [200485] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2392), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4907), 1,
      anon_sym_and,
    STATE(4949), 1,
      sym_attribute,
    STATE(4992), 1,
      aux_sym_expression_item_repeat1,
    STATE(5502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4905), 2,
      anon_sym_in,
      sym_let_and_operator,
  [200512] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2392), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4911), 1,
      anon_sym_and,
    STATE(4906), 1,
      aux_sym_expression_item_repeat1,
    STATE(4950), 1,
      sym_attribute,
    STATE(5502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4909), 2,
      anon_sym_in,
      sym_let_and_operator,
  [200539] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2392), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4915), 1,
      anon_sym_and,
    STATE(4951), 1,
      sym_attribute,
    STATE(4993), 1,
      aux_sym_expression_item_repeat1,
    STATE(5502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4913), 2,
      anon_sym_in,
      sym_let_and_operator,
  [200566] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2392), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4919), 1,
      anon_sym_and,
    STATE(4952), 1,
      sym_attribute,
    STATE(4994), 1,
      aux_sym_expression_item_repeat1,
    STATE(5502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4917), 2,
      anon_sym_in,
      sym_let_and_operator,
  [200593] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7722), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4953), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200622] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7724), 1,
      sym__lowercase_identifier,
    STATE(1054), 1,
      sym_class_path,
    STATE(4954), 1,
      sym_attribute,
    STATE(5441), 1,
      sym__attribute,
    STATE(6593), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200651] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7726), 1,
      anon_sym_RPAREN,
    STATE(4224), 1,
      sym__coerced,
    STATE(4955), 1,
      sym_attribute,
    STATE(5928), 1,
      sym__typed,
    STATE(6484), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200680] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7686), 1,
      anon_sym_PIPE,
    STATE(4956), 1,
      sym_attribute,
    STATE(4957), 1,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5098), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [200703] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7728), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4957), 2,
      sym_attribute,
      aux_sym_variant_declaration_repeat1,
    ACTIONS(5114), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [200724] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7731), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4958), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200753] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7733), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4959), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200782] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3154), 1,
      anon_sym_PIPE,
    STATE(4960), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3156), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200803] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7735), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4961), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200832] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3545), 1,
      anon_sym_PIPE,
    STATE(4962), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3547), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200853] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3036), 1,
      anon_sym_PIPE,
    STATE(4963), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3038), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [200874] = 9,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5133), 1,
      anon_sym_in,
    ACTIONS(7320), 1,
      anon_sym_with,
    ACTIONS(7694), 1,
      anon_sym_DASH_GT,
    STATE(4964), 1,
      sym_attribute,
    STATE(5367), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200903] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7675), 1,
      sym__lowercase_identifier,
    ACTIONS(7677), 1,
      anon_sym_LPAREN,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(4965), 1,
      sym_attribute,
    STATE(5284), 1,
      sym__attribute,
    STATE(6240), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200932] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7737), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4966), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [200961] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7739), 1,
      anon_sym_EQ,
    STATE(4967), 1,
      sym_attribute,
    STATE(5181), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5137), 2,
      anon_sym_and,
      anon_sym_in,
  [200988] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7675), 1,
      sym__lowercase_identifier,
    ACTIONS(7677), 1,
      anon_sym_LPAREN,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(4968), 1,
      sym_attribute,
    STATE(5393), 1,
      sym__attribute,
    STATE(6182), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201017] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7741), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4969), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201046] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2081), 1,
      anon_sym_PIPE,
    STATE(4970), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2075), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201067] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3138), 1,
      anon_sym_PIPE,
    STATE(4971), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3140), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201088] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7350), 1,
      anon_sym_COMMA,
    ACTIONS(7352), 1,
      anon_sym_PIPE,
    ACTIONS(7356), 1,
      anon_sym_as,
    ACTIONS(7358), 1,
      anon_sym_COLON_COLON,
    STATE(4972), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2990), 2,
      anon_sym_SEMI,
      anon_sym_PIPE_RBRACK,
  [201115] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7743), 1,
      sym__lowercase_identifier,
    STATE(913), 1,
      sym_class_path,
    STATE(4973), 1,
      sym_attribute,
    STATE(5483), 1,
      sym__attribute,
    STATE(6725), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201144] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7745), 1,
      anon_sym_RPAREN,
    STATE(4224), 1,
      sym__coerced,
    STATE(4974), 1,
      sym_attribute,
    STATE(5928), 1,
      sym__typed,
    STATE(7047), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201173] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2984), 1,
      anon_sym_PIPE,
    ACTIONS(7350), 1,
      anon_sym_COMMA,
    ACTIONS(7358), 1,
      anon_sym_COLON_COLON,
    STATE(4975), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2986), 3,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_PIPE_RBRACK,
  [201198] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2984), 1,
      anon_sym_PIPE,
    ACTIONS(7350), 1,
      anon_sym_COMMA,
    STATE(4976), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2986), 4,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201221] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7747), 1,
      anon_sym_COLON2,
    STATE(4977), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4507), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4503), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
  [201244] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3549), 1,
      anon_sym_PIPE,
    STATE(4978), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3551), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201265] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7749), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4979), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201294] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7350), 1,
      anon_sym_COMMA,
    ACTIONS(7352), 1,
      anon_sym_PIPE,
    ACTIONS(7358), 1,
      anon_sym_COLON_COLON,
    STATE(4980), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3002), 3,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_PIPE_RBRACK,
  [201319] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7751), 1,
      anon_sym_COLON2,
    STATE(4981), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4507), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4503), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
  [201342] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7753), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4982), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201371] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4637), 1,
      anon_sym_COLON2,
    STATE(4983), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3052), 5,
      anon_sym_DASH_GT,
      anon_sym_STAR,
      anon_sym_POUND,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [201392] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2392), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(4891), 1,
      anon_sym_and,
    STATE(4890), 1,
      aux_sym_expression_item_repeat1,
    STATE(4984), 1,
      sym_attribute,
    STATE(5502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4889), 2,
      anon_sym_in,
      sym_let_and_operator,
  [201419] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3004), 1,
      anon_sym_PIPE,
    ACTIONS(7358), 1,
      anon_sym_COLON_COLON,
    STATE(4985), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3006), 4,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_PIPE_RBRACK,
  [201442] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7755), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4986), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201471] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7757), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4987), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201500] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7759), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4988), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201529] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3040), 1,
      anon_sym_PIPE,
    STATE(4989), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3042), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201550] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7761), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4990), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201579] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3062), 1,
      anon_sym_PIPE,
    STATE(4991), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3064), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201600] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2392), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5015), 1,
      anon_sym_and,
    STATE(4906), 1,
      aux_sym_expression_item_repeat1,
    STATE(4992), 1,
      sym_attribute,
    STATE(5502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5013), 2,
      anon_sym_in,
      sym_let_and_operator,
  [201627] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2392), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5019), 1,
      anon_sym_and,
    STATE(4906), 1,
      aux_sym_expression_item_repeat1,
    STATE(4993), 1,
      sym_attribute,
    STATE(5502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5017), 2,
      anon_sym_in,
      sym_let_and_operator,
  [201654] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2392), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5023), 1,
      anon_sym_and,
    STATE(4906), 1,
      aux_sym_expression_item_repeat1,
    STATE(4994), 1,
      sym_attribute,
    STATE(5502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5021), 2,
      anon_sym_in,
      sym_let_and_operator,
  [201681] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2392), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5027), 1,
      anon_sym_and,
    STATE(4995), 1,
      sym_attribute,
    STATE(5023), 1,
      aux_sym_expression_item_repeat1,
    STATE(5502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5025), 2,
      anon_sym_in,
      sym_let_and_operator,
  [201708] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7763), 1,
      sym__lowercase_identifier,
    STATE(2723), 1,
      sym_class_path,
    STATE(4996), 1,
      sym_attribute,
    STATE(5517), 1,
      sym__attribute,
    STATE(6761), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201737] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7765), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(4997), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201766] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1813), 1,
      anon_sym_PIPE,
    STATE(4998), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1815), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201787] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7767), 1,
      anon_sym_RPAREN,
    STATE(4224), 1,
      sym__coerced,
    STATE(4999), 1,
      sym_attribute,
    STATE(5928), 1,
      sym__typed,
    STATE(7257), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201816] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3553), 1,
      anon_sym_PIPE,
    STATE(5000), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3555), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201837] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3054), 1,
      anon_sym_PIPE,
    STATE(5001), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3056), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201858] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3563), 1,
      anon_sym_PIPE,
    STATE(5002), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3565), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201879] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7769), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(5003), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201908] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7686), 1,
      anon_sym_PIPE,
    STATE(4913), 1,
      aux_sym_variant_declaration_repeat1,
    STATE(5004), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4380), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [201931] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7771), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(5005), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [201960] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3142), 1,
      anon_sym_PIPE,
    STATE(5006), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3144), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [201981] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3070), 1,
      anon_sym_PIPE,
    STATE(5007), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3072), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [202002] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7773), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(5008), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202031] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3014), 1,
      anon_sym_PIPE,
    STATE(5009), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3016), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [202052] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3018), 1,
      anon_sym_PIPE,
    STATE(5010), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3020), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [202073] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7775), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(5011), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202102] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7777), 1,
      anon_sym_EQ,
    STATE(5012), 1,
      sym_attribute,
    STATE(5207), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5167), 2,
      anon_sym_and,
      anon_sym_in,
  [202129] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7779), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(5013), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202158] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3468), 1,
      anon_sym_PIPE,
    STATE(5014), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3470), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [202179] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7781), 1,
      anon_sym_EQ,
    STATE(5015), 1,
      sym_attribute,
    STATE(5209), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5173), 2,
      anon_sym_and,
      anon_sym_in,
  [202206] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7783), 1,
      anon_sym_COLON2,
    STATE(5016), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4507), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4503), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
  [202229] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(7787), 1,
      anon_sym_AMP,
    STATE(5017), 1,
      sym_attribute,
    STATE(5173), 1,
      aux_sym_tag_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7785), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [202254] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7789), 1,
      anon_sym_COLON2,
    STATE(5018), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4507), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4503), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
  [202277] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7791), 1,
      sym__lowercase_identifier,
    STATE(2005), 1,
      sym_class_path,
    STATE(5019), 1,
      sym_attribute,
    STATE(5512), 1,
      sym__attribute,
    STATE(6794), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202306] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7793), 1,
      anon_sym_RPAREN,
    STATE(4224), 1,
      sym__coerced,
    STATE(5020), 1,
      sym_attribute,
    STATE(5928), 1,
      sym__typed,
    STATE(6834), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202335] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    ACTIONS(7795), 1,
      anon_sym_RPAREN,
    STATE(4224), 1,
      sym__coerced,
    STATE(5021), 1,
      sym_attribute,
    STATE(5928), 1,
      sym__typed,
    STATE(6343), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202364] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3022), 1,
      anon_sym_PIPE,
    STATE(5022), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3024), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [202385] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2392), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5050), 1,
      anon_sym_and,
    STATE(4906), 1,
      aux_sym_expression_item_repeat1,
    STATE(5023), 1,
      sym_attribute,
    STATE(5502), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5048), 2,
      anon_sym_in,
      sym_let_and_operator,
  [202412] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(7787), 1,
      anon_sym_AMP,
    STATE(5024), 1,
      sym_attribute,
    STATE(5227), 1,
      aux_sym_tag_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7797), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [202437] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1713), 1,
      anon_sym_PIPE,
    STATE(5025), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1715), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [202458] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7799), 1,
      anon_sym_EQ,
    STATE(5026), 1,
      sym_attribute,
    STATE(5236), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5240), 2,
      anon_sym_and,
      anon_sym_in,
  [202485] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7801), 1,
      anon_sym_EQ,
    STATE(5027), 1,
      sym_attribute,
    STATE(5238), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5092), 2,
      anon_sym_and,
      anon_sym_in,
  [202512] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7803), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(5028), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202541] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7675), 1,
      sym__lowercase_identifier,
    ACTIONS(7677), 1,
      anon_sym_LPAREN,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(5029), 1,
      sym_attribute,
    STATE(5465), 1,
      sym__attribute,
    STATE(6142), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202570] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3106), 1,
      anon_sym_PIPE,
    STATE(5030), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3108), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [202591] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7805), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(5031), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202620] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1809), 1,
      anon_sym_PIPE,
    STATE(5032), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1811), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [202641] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7807), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(5033), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202670] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7675), 1,
      sym__lowercase_identifier,
    ACTIONS(7677), 1,
      anon_sym_LPAREN,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(5034), 1,
      sym_attribute,
    STATE(5490), 1,
      sym__attribute,
    STATE(6201), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202699] = 8,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(7809), 1,
      anon_sym_EQ,
    STATE(5035), 1,
      sym_attribute,
    STATE(5212), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5153), 2,
      anon_sym_and,
      anon_sym_in,
  [202726] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    ACTIONS(7811), 1,
      anon_sym_RBRACE,
    STATE(4678), 1,
      sym_field_path,
    STATE(5036), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202755] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7813), 1,
      anon_sym_COLON2,
    STATE(5037), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4507), 2,
      anon_sym_LPAREN,
      anon_sym_DOT,
    ACTIONS(4503), 3,
      anon_sym_RPAREN,
      anon_sym_DASH_GT,
      anon_sym_with,
  [202778] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3110), 1,
      anon_sym_PIPE,
    STATE(5038), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3112), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [202799] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1843), 1,
      anon_sym_PIPE,
    STATE(5039), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1845), 5,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_PIPE_RBRACK,
  [202820] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    STATE(2352), 1,
      sym_class_path,
    STATE(5040), 1,
      sym_attribute,
    STATE(5319), 1,
      sym__attribute,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202849] = 9,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    ACTIONS(7817), 1,
      sym__lowercase_identifier,
    STATE(1548), 1,
      sym_class_path,
    STATE(5041), 1,
      sym_attribute,
    STATE(5474), 1,
      sym__attribute,
    STATE(6688), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202878] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5042), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2051), 2,
      anon_sym_and,
      anon_sym_in,
  [202902] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(7821), 1,
      anon_sym_SEMI,
    ACTIONS(7823), 1,
      anon_sym_GT_RBRACE,
    STATE(5043), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(5758), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202928] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(5044), 1,
      sym_attribute,
    STATE(5069), 1,
      sym_field_path,
    STATE(6202), 1,
      sym_field_binding_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202954] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3438), 1,
      sym__uppercase_identifier,
    ACTIONS(7825), 1,
      anon_sym_LPAREN,
    STATE(4454), 1,
      sym__constructor_name,
    STATE(5045), 1,
      sym_attribute,
    STATE(5077), 1,
      sym_constructor_path,
    STATE(6619), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [202980] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5046), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4774), 2,
      anon_sym_and,
      anon_sym_in,
  [203004] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5047), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5258), 2,
      anon_sym_and,
      anon_sym_in,
  [203028] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(221), 1,
      anon_sym_DQUOTE,
    ACTIONS(7827), 1,
      anon_sym_LBRACE,
    STATE(3567), 1,
      aux_sym_external_repeat1,
    STATE(5048), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3818), 2,
      sym_string,
      sym_quoted_string,
  [203052] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    STATE(4224), 1,
      sym__coerced,
    STATE(5049), 1,
      sym_attribute,
    STATE(5928), 1,
      sym__typed,
    STATE(6660), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203078] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7829), 1,
      anon_sym_EQ,
    STATE(5050), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203104] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(4678), 1,
      sym_field_path,
    STATE(5051), 1,
      sym_attribute,
    STATE(5641), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203130] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5052), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2057), 2,
      anon_sym_and,
      anon_sym_in,
  [203154] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5053), 1,
      sym_attribute,
    STATE(5071), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5262), 2,
      anon_sym_and,
      anon_sym_in,
  [203178] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5054), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5572), 2,
      anon_sym_and,
      anon_sym_in,
  [203202] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5055), 1,
      sym_attribute,
    STATE(5158), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5662), 2,
      anon_sym_and,
      anon_sym_in,
  [203226] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    STATE(4224), 1,
      sym__coerced,
    STATE(5056), 1,
      sym_attribute,
    STATE(5928), 1,
      sym__typed,
    STATE(6843), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203252] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5057), 1,
      sym_attribute,
    STATE(5119), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5584), 2,
      anon_sym_and,
      anon_sym_in,
  [203276] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5058), 1,
      sym_attribute,
    STATE(5120), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5604), 2,
      anon_sym_and,
      anon_sym_in,
  [203300] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5059), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5610), 2,
      anon_sym_and,
      anon_sym_in,
  [203324] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5060), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5678), 2,
      anon_sym_and,
      anon_sym_in,
  [203348] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5061), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5638), 2,
      anon_sym_and,
      anon_sym_in,
  [203372] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5062), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5286), 2,
      anon_sym_and,
      anon_sym_in,
  [203396] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7324), 1,
      anon_sym_COMMA,
    ACTIONS(7326), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_as,
    ACTIONS(7332), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7553), 1,
      anon_sym_RPAREN,
    STATE(5063), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203422] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5064), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5332), 2,
      anon_sym_and,
      anon_sym_in,
  [203446] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5059), 1,
      aux_sym_expression_item_repeat1,
    STATE(5065), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5304), 2,
      anon_sym_and,
      anon_sym_in,
  [203470] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5066), 1,
      sym_attribute,
    STATE(5128), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5320), 2,
      anon_sym_and,
      anon_sym_in,
  [203494] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5067), 1,
      sym_attribute,
    STATE(5132), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5360), 2,
      anon_sym_and,
      anon_sym_in,
  [203518] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5068), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5368), 2,
      anon_sym_and,
      anon_sym_in,
  [203542] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7831), 1,
      anon_sym_EQ,
    ACTIONS(7833), 1,
      anon_sym_COLON2,
    STATE(5069), 1,
      sym_attribute,
    STATE(5700), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7835), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [203566] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5070), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5414), 2,
      anon_sym_and,
      anon_sym_in,
  [203590] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5071), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5418), 2,
      anon_sym_and,
      anon_sym_in,
  [203614] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(5072), 1,
      sym_attribute,
    STATE(5141), 1,
      sym_field_path,
    STATE(6332), 1,
      sym_field_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203640] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5073), 1,
      sym_attribute,
    STATE(5137), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5246), 2,
      anon_sym_and,
      anon_sym_in,
  [203664] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5074), 1,
      sym_attribute,
    STATE(5138), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5472), 2,
      anon_sym_and,
      anon_sym_in,
  [203688] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5075), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5476), 2,
      anon_sym_and,
      anon_sym_in,
  [203712] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3500), 1,
      sym__uppercase_identifier,
    ACTIONS(7837), 1,
      anon_sym_LPAREN,
    STATE(3151), 1,
      sym__constructor_name,
    STATE(3593), 1,
      sym_constructor_path,
    STATE(5076), 1,
      sym_attribute,
    STATE(7160), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203738] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5077), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4342), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [203756] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(7839), 1,
      anon_sym_SEMI,
    ACTIONS(7841), 1,
      anon_sym_GT_RBRACE,
    STATE(5078), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(5598), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203782] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5079), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4346), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [203800] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(7843), 1,
      anon_sym_SEMI,
    ACTIONS(7845), 1,
      anon_sym_GT_RBRACE,
    STATE(5080), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(5697), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203826] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    STATE(4224), 1,
      sym__coerced,
    STATE(5081), 1,
      sym_attribute,
    STATE(5928), 1,
      sym__typed,
    STATE(6344), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203852] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5082), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7847), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [203870] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(4678), 1,
      sym_field_path,
    STATE(5083), 1,
      sym_attribute,
    STATE(5713), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203896] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(7849), 1,
      anon_sym_SEMI,
    ACTIONS(7851), 1,
      anon_sym_GT_RBRACE,
    STATE(5084), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(5600), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203922] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4310), 1,
      sym__lowercase_identifier,
    ACTIONS(7853), 1,
      anon_sym_POUND,
    STATE(3004), 1,
      sym_type_constructor_path,
    STATE(5085), 1,
      sym_attribute,
    STATE(6029), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203948] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5086), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4455), 2,
      anon_sym_and,
      anon_sym_in,
  [203972] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    STATE(4224), 1,
      sym__coerced,
    STATE(5087), 1,
      sym_attribute,
    STATE(5928), 1,
      sym__typed,
    STATE(6412), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [203998] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7855), 1,
      anon_sym_type,
    ACTIONS(7857), 1,
      anon_sym_module,
    STATE(5088), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4747), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [204020] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7615), 1,
      sym__lowercase_identifier,
    STATE(4040), 1,
      sym__instance_variable_name,
    STATE(5089), 1,
      sym_attribute,
    STATE(5156), 1,
      aux_sym_instance_variable_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7619), 2,
      anon_sym_mutable,
      anon_sym_virtual,
  [204044] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(7859), 1,
      anon_sym_SEMI,
    ACTIONS(7861), 1,
      anon_sym_GT_RBRACE,
    STATE(5090), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(5798), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204070] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7615), 1,
      sym__lowercase_identifier,
    STATE(4040), 1,
      sym__instance_variable_name,
    STATE(5091), 1,
      sym_attribute,
    STATE(5358), 1,
      aux_sym_instance_variable_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7619), 2,
      anon_sym_mutable,
      anon_sym_virtual,
  [204094] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(4678), 1,
      sym_field_path,
    STATE(5092), 1,
      sym_attribute,
    STATE(5769), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204120] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    STATE(4224), 1,
      sym__coerced,
    STATE(5093), 1,
      sym_attribute,
    STATE(5928), 1,
      sym__typed,
    STATE(6684), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204146] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5094), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2027), 2,
      anon_sym_and,
      anon_sym_in,
  [204170] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5095), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5699), 2,
      anon_sym_and,
      anon_sym_in,
  [204194] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7526), 1,
      anon_sym_RPAREN,
    STATE(5096), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204220] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(4678), 1,
      sym_field_path,
    STATE(5097), 1,
      sym_attribute,
    STATE(5834), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204246] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7863), 1,
      sym__lowercase_identifier,
    ACTIONS(7867), 1,
      anon_sym_BANG,
    STATE(5098), 1,
      sym_attribute,
    STATE(5286), 1,
      aux_sym_method_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7865), 2,
      anon_sym_private,
      anon_sym_virtual,
  [204270] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    STATE(4224), 1,
      sym__coerced,
    STATE(5099), 1,
      sym_attribute,
    STATE(5928), 1,
      sym__typed,
    STATE(7071), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204296] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    STATE(4224), 1,
      sym__coerced,
    STATE(5100), 1,
      sym_attribute,
    STATE(5928), 1,
      sym__typed,
    STATE(6971), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204322] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5095), 1,
      aux_sym_expression_item_repeat1,
    STATE(5101), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5654), 2,
      anon_sym_and,
      anon_sym_in,
  [204346] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(6567), 1,
      anon_sym_LBRACE,
    STATE(4550), 1,
      aux_sym_external_repeat1,
    STATE(5102), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5492), 2,
      sym_string,
      sym_quoted_string,
  [204370] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(4678), 1,
      sym_field_path,
    STATE(5103), 1,
      sym_attribute,
    STATE(5855), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204396] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(5104), 1,
      sym_attribute,
    STATE(5141), 1,
      sym_field_path,
    STATE(5844), 1,
      sym_field_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204422] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5105), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4804), 2,
      anon_sym_and,
      anon_sym_in,
  [204446] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7855), 1,
      anon_sym_type,
    ACTIONS(7857), 1,
      anon_sym_module,
    STATE(5106), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4680), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [204468] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5107), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5486), 2,
      anon_sym_and,
      anon_sym_in,
  [204492] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5108), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5324), 2,
      anon_sym_and,
      anon_sym_in,
  [204516] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5109), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5110), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [204534] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5110), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5114), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [204552] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4703), 1,
      anon_sym_DQUOTE,
    ACTIONS(7869), 1,
      anon_sym_LBRACE,
    STATE(3346), 1,
      aux_sym_external_repeat1,
    STATE(5111), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3818), 2,
      sym_string,
      sym_quoted_string,
  [204576] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    STATE(4224), 1,
      sym__coerced,
    STATE(5112), 1,
      sym_attribute,
    STATE(5928), 1,
      sym__typed,
    STATE(6456), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204602] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(4678), 1,
      sym_field_path,
    STATE(5113), 1,
      sym_attribute,
    STATE(5698), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204628] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7615), 1,
      sym__lowercase_identifier,
    STATE(5114), 1,
      sym_attribute,
    STATE(5183), 1,
      aux_sym_instance_variable_specification_repeat1,
    STATE(6055), 1,
      sym__instance_variable_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7619), 2,
      anon_sym_mutable,
      anon_sym_virtual,
  [204652] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(4678), 1,
      sym_field_path,
    STATE(5115), 1,
      sym_attribute,
    STATE(5929), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204678] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3929), 1,
      anon_sym_LPAREN,
    ACTIONS(7871), 1,
      sym__lowercase_identifier,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    STATE(3501), 1,
      sym_parenthesized_operator,
    STATE(4113), 1,
      sym__value_name,
    STATE(5116), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204704] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5117), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4748), 2,
      anon_sym_and,
      anon_sym_in,
  [204728] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5118), 1,
      sym_attribute,
    STATE(5172), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5348), 2,
      anon_sym_and,
      anon_sym_in,
  [204752] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5119), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5352), 2,
      anon_sym_and,
      anon_sym_in,
  [204776] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5120), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5356), 2,
      anon_sym_and,
      anon_sym_in,
  [204800] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(7875), 1,
      anon_sym_COMMA,
    ACTIONS(7877), 1,
      anon_sym_PIPE,
    ACTIONS(7879), 1,
      anon_sym_RBRACK,
    STATE(5121), 1,
      sym_attribute,
    STATE(5826), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204826] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(4678), 1,
      sym_field_path,
    STATE(5122), 1,
      sym_attribute,
    STATE(5555), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [204852] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5123), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5376), 2,
      anon_sym_and,
      anon_sym_in,
  [204876] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5124), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5376), 2,
      anon_sym_and,
      anon_sym_in,
  [204900] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5125), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5394), 2,
      anon_sym_and,
      anon_sym_in,
  [204924] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5126), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5646), 2,
      anon_sym_and,
      anon_sym_in,
  [204948] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5127), 1,
      sym_attribute,
    STATE(5178), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5398), 2,
      anon_sym_and,
      anon_sym_in,
  [204972] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5128), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5402), 2,
      anon_sym_and,
      anon_sym_in,
  [204996] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5129), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2063), 2,
      anon_sym_and,
      anon_sym_in,
  [205020] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(4678), 1,
      sym_field_path,
    STATE(5130), 1,
      sym_attribute,
    STATE(5568), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205046] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    STATE(4224), 1,
      sym__coerced,
    STATE(5131), 1,
      sym_attribute,
    STATE(5928), 1,
      sym__typed,
    STATE(6732), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205072] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5132), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5422), 2,
      anon_sym_and,
      anon_sym_in,
  [205096] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5133), 1,
      sym_attribute,
    STATE(5190), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5426), 2,
      anon_sym_and,
      anon_sym_in,
  [205120] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5134), 1,
      sym_attribute,
    STATE(5191), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5430), 2,
      anon_sym_and,
      anon_sym_in,
  [205144] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5135), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5434), 2,
      anon_sym_and,
      anon_sym_in,
  [205168] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5136), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5494), 2,
      anon_sym_and,
      anon_sym_in,
  [205192] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5137), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5448), 2,
      anon_sym_and,
      anon_sym_in,
  [205216] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5138), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5452), 2,
      anon_sym_and,
      anon_sym_in,
  [205240] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5139), 1,
      sym_attribute,
    STATE(5195), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5456), 2,
      anon_sym_and,
      anon_sym_in,
  [205264] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(7881), 1,
      anon_sym_SEMI,
    ACTIONS(7883), 1,
      anon_sym_GT_RBRACE,
    STATE(5140), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(5848), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205290] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7833), 1,
      anon_sym_COLON2,
    ACTIONS(7885), 1,
      anon_sym_EQ,
    STATE(5141), 1,
      sym_attribute,
    STATE(5663), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7887), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [205314] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7889), 1,
      anon_sym_type,
    ACTIONS(7891), 1,
      anon_sym_module,
    STATE(5142), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3367), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [205336] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5143), 1,
      sym_attribute,
    STATE(5174), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5666), 2,
      anon_sym_and,
      anon_sym_in,
  [205360] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5144), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7893), 5,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_BANG,
      anon_sym__,
      aux_sym_type_variable_token1,
  [205378] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7324), 1,
      anon_sym_COMMA,
    ACTIONS(7326), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_as,
    ACTIONS(7332), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7372), 1,
      anon_sym_RPAREN,
    STATE(5145), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205404] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(5069), 1,
      sym_field_path,
    STATE(5146), 1,
      sym_attribute,
    STATE(5675), 1,
      sym_field_binding_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205430] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7895), 1,
      anon_sym_type,
    ACTIONS(7897), 1,
      anon_sym_module,
    STATE(5147), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3471), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [205452] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(5069), 1,
      sym_field_path,
    STATE(5148), 1,
      sym_attribute,
    STATE(5699), 1,
      sym_field_binding_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205478] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(4678), 1,
      sym_field_path,
    STATE(5149), 1,
      sym_attribute,
    STATE(5681), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205504] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5150), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7899), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [205522] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(5141), 1,
      sym_field_path,
    STATE(5151), 1,
      sym_attribute,
    STATE(5701), 1,
      sym_field_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205548] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7134), 1,
      anon_sym_RPAREN,
    STATE(5152), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205574] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5153), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5278), 2,
      anon_sym_and,
      anon_sym_in,
  [205598] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5154), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7901), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [205616] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(7903), 1,
      anon_sym_SEMI,
    ACTIONS(7905), 1,
      anon_sym_GT_RBRACE,
    STATE(5155), 1,
      sym_attribute,
    STATE(5577), 1,
      sym_instance_variable_expression,
    STATE(5578), 1,
      sym__instance_variable_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205642] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7615), 1,
      sym__lowercase_identifier,
    STATE(4042), 1,
      sym__instance_variable_name,
    STATE(5156), 1,
      sym_attribute,
    STATE(5358), 1,
      aux_sym_instance_variable_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7619), 2,
      anon_sym_mutable,
      anon_sym_virtual,
  [205666] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5157), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7907), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [205684] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5158), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5266), 2,
      anon_sym_and,
      anon_sym_in,
  [205708] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7324), 1,
      anon_sym_COMMA,
    ACTIONS(7326), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_as,
    ACTIONS(7332), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7538), 1,
      anon_sym_RPAREN,
    STATE(5159), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205734] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5160), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5658), 2,
      anon_sym_and,
      anon_sym_in,
  [205758] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(221), 1,
      anon_sym_DQUOTE,
    ACTIONS(7827), 1,
      anon_sym_LBRACE,
    STATE(3569), 1,
      aux_sym_external_repeat1,
    STATE(5161), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3818), 2,
      sym_string,
      sym_quoted_string,
  [205782] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    ACTIONS(7909), 1,
      anon_sym_POUND,
    STATE(2577), 1,
      sym_type_constructor_path,
    STATE(5162), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205808] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5163), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5159), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [205826] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(7911), 1,
      anon_sym_SEMI,
    ACTIONS(7913), 1,
      anon_sym_GT_RBRACE,
    STATE(5164), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(5915), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205852] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5165), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4857), 2,
      anon_sym_and,
      anon_sym_in,
  [205876] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5166), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4881), 2,
      anon_sym_and,
      anon_sym_in,
  [205900] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(4678), 1,
      sym_field_path,
    STATE(5167), 1,
      sym_attribute,
    STATE(6144), 1,
      sym_field_expression,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205926] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5168), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4770), 2,
      anon_sym_and,
      anon_sym_in,
  [205950] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7342), 1,
      anon_sym_RPAREN,
    STATE(5169), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [205976] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5170), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5552), 2,
      anon_sym_and,
      anon_sym_in,
  [206000] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2528), 1,
      sym__uppercase_identifier,
    ACTIONS(7915), 1,
      anon_sym_LPAREN,
    STATE(1694), 1,
      sym__constructor_name,
    STATE(3593), 1,
      sym_constructor_path,
    STATE(5171), 1,
      sym_attribute,
    STATE(6418), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [206026] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5172), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5556), 2,
      anon_sym_and,
      anon_sym_in,
  [206050] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7787), 1,
      anon_sym_AMP,
    STATE(5173), 1,
      sym_attribute,
    STATE(5193), 1,
      aux_sym_tag_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7917), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [206072] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5174), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5270), 2,
      anon_sym_and,
      anon_sym_in,
  [206096] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4691), 1,
      sym__lowercase_identifier,
    ACTIONS(7919), 1,
      anon_sym_POUND,
    STATE(4443), 1,
      sym_type_constructor_path,
    STATE(5175), 1,
      sym_attribute,
    STATE(6026), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [206122] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5176), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5560), 2,
      anon_sym_and,
      anon_sym_in,
  [206146] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5177), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5560), 2,
      anon_sym_and,
      anon_sym_in,
  [206170] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5178), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5564), 2,
      anon_sym_and,
      anon_sym_in,
  [206194] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5179), 1,
      sym_attribute,
    STATE(5204), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5568), 2,
      anon_sym_and,
      anon_sym_in,
  [206218] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5153), 1,
      aux_sym_expression_item_repeat1,
    STATE(5180), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5274), 2,
      anon_sym_and,
      anon_sym_in,
  [206242] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5181), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5576), 2,
      anon_sym_and,
      anon_sym_in,
  [206266] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5182), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5588), 2,
      anon_sym_and,
      anon_sym_in,
  [206290] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7615), 1,
      sym__lowercase_identifier,
    STATE(5183), 1,
      sym_attribute,
    STATE(5358), 1,
      aux_sym_instance_variable_specification_repeat1,
    STATE(6139), 1,
      sym__instance_variable_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7619), 2,
      anon_sym_mutable,
      anon_sym_virtual,
  [206314] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    STATE(5184), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7921), 4,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
      anon_sym_AMP,
  [206334] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5185), 1,
      sym_attribute,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2033), 2,
      anon_sym_and,
      anon_sym_in,
  [206358] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(5069), 1,
      sym_field_path,
    STATE(5186), 1,
      sym_attribute,
    STATE(5639), 1,
      sym_field_binding_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [206384] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7923), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4741), 2,
      anon_sym_and,
      anon_sym_in,
    STATE(5187), 2,
      sym_attribute,
      aux_sym_expression_item_repeat1,
  [206406] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(7328), 1,
      anon_sym_COLON2,
    STATE(4224), 1,
      sym__coerced,
    STATE(5188), 1,
      sym_attribute,
    STATE(5928), 1,
      sym__typed,
    STATE(6679), 1,
      sym__type_constrained,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [206432] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5189), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5312), 2,
      anon_sym_and,
      anon_sym_in,
  [206456] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5190), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5592), 2,
      anon_sym_and,
      anon_sym_in,
  [206480] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5191), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5596), 2,
      anon_sym_and,
      anon_sym_in,
  [206504] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5192), 1,
      sym_attribute,
    STATE(5213), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5600), 2,
      anon_sym_and,
      anon_sym_in,
  [206528] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7926), 1,
      anon_sym_AMP,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5193), 2,
      sym_attribute,
      aux_sym_tag_specification_repeat1,
    ACTIONS(7921), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [206548] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(7929), 1,
      anon_sym_SEMI,
    ACTIONS(7931), 1,
      anon_sym_GT_RBRACE,
    STATE(5194), 1,
      sym_attribute,
    STATE(5546), 1,
      sym_instance_variable_expression,
    STATE(5578), 1,
      sym__instance_variable_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [206574] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5195), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5614), 2,
      anon_sym_and,
      anon_sym_in,
  [206598] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7895), 1,
      anon_sym_type,
    ACTIONS(7897), 1,
      anon_sym_module,
    STATE(5196), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3367), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [206620] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7322), 1,
      anon_sym_RPAREN,
    ACTIONS(7324), 1,
      anon_sym_COMMA,
    ACTIONS(7326), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_as,
    ACTIONS(7332), 1,
      anon_sym_COLON_COLON,
    STATE(5197), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [206646] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5198), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5218), 5,
      anon_sym_and,
      anon_sym_PIPE,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [206664] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5199), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4808), 2,
      anon_sym_and,
      anon_sym_in,
  [206688] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5200), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5670), 2,
      anon_sym_and,
      anon_sym_in,
  [206712] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5201), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5674), 2,
      anon_sym_and,
      anon_sym_in,
  [206736] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5202), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4843), 2,
      anon_sym_and,
      anon_sym_in,
  [206760] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5203), 1,
      sym_attribute,
    STATE(5221), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5686), 2,
      anon_sym_and,
      anon_sym_in,
  [206784] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5204), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5695), 2,
      anon_sym_and,
      anon_sym_in,
  [206808] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5205), 1,
      sym_attribute,
    STATE(5223), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5468), 2,
      anon_sym_and,
      anon_sym_in,
  [206832] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5206), 1,
      sym_attribute,
    STATE(5226), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5460), 2,
      anon_sym_and,
      anon_sym_in,
  [206856] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5207), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5410), 2,
      anon_sym_and,
      anon_sym_in,
  [206880] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5208), 1,
      sym_attribute,
    STATE(5232), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5282), 2,
      anon_sym_and,
      anon_sym_in,
  [206904] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5209), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5308), 2,
      anon_sym_and,
      anon_sym_in,
  [206928] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(5141), 1,
      sym_field_path,
    STATE(5210), 1,
      sym_attribute,
    STATE(5652), 1,
      sym_field_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [206954] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5211), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5316), 2,
      anon_sym_and,
      anon_sym_in,
  [206978] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5212), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5254), 2,
      anon_sym_and,
      anon_sym_in,
  [207002] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5213), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5336), 2,
      anon_sym_and,
      anon_sym_in,
  [207026] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7933), 1,
      anon_sym_as,
    STATE(5214), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6589), 4,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COLON_GT,
  [207046] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1583), 1,
      anon_sym_POUND,
    ACTIONS(7935), 1,
      anon_sym_DOT,
    STATE(5215), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1585), 3,
      anon_sym_COMMA,
      anon_sym_with,
      sym_hash_operator,
  [207068] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7937), 1,
      anon_sym_EQ,
    STATE(5216), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207094] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5217), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5490), 2,
      anon_sym_and,
      anon_sym_in,
  [207118] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5218), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4752), 2,
      anon_sym_and,
      anon_sym_in,
  [207142] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(7875), 1,
      anon_sym_COMMA,
    ACTIONS(7877), 1,
      anon_sym_PIPE,
    ACTIONS(7939), 1,
      anon_sym_RBRACK,
    STATE(5219), 1,
      sym_attribute,
    STATE(5783), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207168] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5220), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5504), 2,
      anon_sym_and,
      anon_sym_in,
  [207192] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5221), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5508), 2,
      anon_sym_and,
      anon_sym_in,
  [207216] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5222), 1,
      sym_attribute,
    STATE(5258), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5512), 2,
      anon_sym_and,
      anon_sym_in,
  [207240] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5223), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5516), 2,
      anon_sym_and,
      anon_sym_in,
  [207264] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1459), 1,
      anon_sym_DQUOTE,
    ACTIONS(6567), 1,
      anon_sym_LBRACE,
    STATE(4603), 1,
      aux_sym_external_repeat1,
    STATE(5224), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5492), 2,
      sym_string,
      sym_quoted_string,
  [207288] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4703), 1,
      anon_sym_DQUOTE,
    ACTIONS(7869), 1,
      anon_sym_LBRACE,
    STATE(3396), 1,
      aux_sym_external_repeat1,
    STATE(5225), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3818), 2,
      sym_string,
      sym_quoted_string,
  [207312] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5226), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5524), 2,
      anon_sym_and,
      anon_sym_in,
  [207336] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7787), 1,
      anon_sym_AMP,
    STATE(5193), 1,
      aux_sym_tag_specification_repeat1,
    STATE(5227), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7785), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [207358] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5228), 1,
      sym_attribute,
    STATE(5260), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5528), 2,
      anon_sym_and,
      anon_sym_in,
  [207382] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7941), 1,
      anon_sym_type,
    ACTIONS(7943), 1,
      anon_sym_module,
    STATE(5229), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3367), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [207404] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7312), 1,
      anon_sym_COLON_GT,
    STATE(4215), 1,
      sym__coerced,
    STATE(5230), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6707), 3,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [207426] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(5069), 1,
      sym_field_path,
    STATE(5231), 1,
      sym_attribute,
    STATE(5609), 1,
      sym_field_binding_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207452] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5232), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5532), 2,
      anon_sym_and,
      anon_sym_in,
  [207476] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5126), 1,
      aux_sym_expression_item_repeat1,
    STATE(5233), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5540), 2,
      anon_sym_and,
      anon_sym_in,
  [207500] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7889), 1,
      anon_sym_type,
    ACTIONS(7891), 1,
      anon_sym_module,
    STATE(5234), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4934), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [207522] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5235), 1,
      sym_attribute,
    STATE(5245), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5544), 2,
      anon_sym_and,
      anon_sym_in,
  [207546] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5236), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5548), 2,
      anon_sym_and,
      anon_sym_in,
  [207570] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5160), 1,
      aux_sym_expression_item_repeat1,
    STATE(5237), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5580), 2,
      anon_sym_and,
      anon_sym_in,
  [207594] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5238), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5618), 2,
      anon_sym_and,
      anon_sym_in,
  [207618] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5239), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5622), 2,
      anon_sym_and,
      anon_sym_in,
  [207642] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7945), 1,
      anon_sym_EQ,
    STATE(5240), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207668] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7941), 1,
      anon_sym_type,
    ACTIONS(7943), 1,
      anon_sym_module,
    STATE(5241), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3771), 3,
      sym_constrain_type,
      sym_constrain_module,
      sym_constrain_module_type,
  [207690] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5242), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4812), 2,
      anon_sym_and,
      anon_sym_in,
  [207714] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7947), 1,
      anon_sym_EQ,
    STATE(5243), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207740] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7204), 1,
      sym__lowercase_identifier,
    STATE(5141), 1,
      sym_field_path,
    STATE(5244), 1,
      sym_attribute,
    STATE(5612), 1,
      sym_field_pattern,
    STATE(6997), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207766] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5245), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5650), 2,
      anon_sym_and,
      anon_sym_in,
  [207790] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7949), 1,
      anon_sym_DASH_GT,
    ACTIONS(7951), 1,
      anon_sym_with,
    STATE(5246), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5052), 3,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [207812] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7949), 1,
      anon_sym_DASH_GT,
    ACTIONS(7951), 1,
      anon_sym_with,
    STATE(5247), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5056), 3,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [207834] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7949), 1,
      anon_sym_DASH_GT,
    ACTIONS(7951), 1,
      anon_sym_with,
    STATE(5248), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5060), 3,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [207856] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7949), 1,
      anon_sym_DASH_GT,
    ACTIONS(7951), 1,
      anon_sym_with,
    STATE(5249), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5064), 3,
      anon_sym_and,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [207878] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7953), 1,
      anon_sym_EQ,
    STATE(5250), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207904] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7955), 1,
      anon_sym_EQ,
    STATE(5251), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207930] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7957), 1,
      anon_sym_EQ,
    STATE(5252), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207956] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7959), 1,
      anon_sym_EQ,
    STATE(5253), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [207982] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(7875), 1,
      anon_sym_COMMA,
    ACTIONS(7877), 1,
      anon_sym_PIPE,
    ACTIONS(7961), 1,
      anon_sym_RBRACK,
    STATE(5254), 1,
      sym_attribute,
    STATE(5650), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208008] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(7963), 1,
      anon_sym_SEMI,
    ACTIONS(7965), 1,
      anon_sym_GT_RBRACE,
    STATE(5255), 1,
      sym_attribute,
    STATE(5564), 1,
      sym_instance_variable_expression,
    STATE(5578), 1,
      sym__instance_variable_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208034] = 8,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    ACTIONS(7118), 1,
      anon_sym_RPAREN,
    STATE(5256), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208060] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5257), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5626), 2,
      anon_sym_and,
      anon_sym_in,
  [208084] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5258), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5630), 2,
      anon_sym_and,
      anon_sym_in,
  [208108] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5047), 1,
      aux_sym_expression_item_repeat1,
    STATE(5259), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5634), 2,
      anon_sym_and,
      anon_sym_in,
  [208132] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5260), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5642), 2,
      anon_sym_and,
      anon_sym_in,
  [208156] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5261), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5536), 2,
      anon_sym_and,
      anon_sym_in,
  [208180] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7967), 1,
      sym__lowercase_identifier,
    ACTIONS(7969), 1,
      anon_sym_RBRACE,
    ACTIONS(7971), 1,
      anon_sym_mutable,
    STATE(5262), 1,
      sym_attribute,
    STATE(6106), 1,
      sym_field_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208203] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6909), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    STATE(5263), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208226] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(7973), 1,
      anon_sym_RBRACK,
    STATE(5264), 1,
      sym_attribute,
    STATE(5496), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6146), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208249] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    ACTIONS(7975), 1,
      sym__lowercase_identifier,
    STATE(5265), 1,
      sym_attribute,
    STATE(5956), 1,
      sym_module_type_path,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208272] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(7977), 1,
      anon_sym_RBRACK,
    STATE(5266), 1,
      sym_attribute,
    STATE(5496), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6146), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208295] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(7979), 1,
      sym__lowercase_identifier,
    STATE(2583), 1,
      sym_class_type_path,
    STATE(5267), 1,
      sym_attribute,
    STATE(5974), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208318] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7981), 1,
      anon_sym_rec,
    STATE(3923), 1,
      sym_module_binding,
    STATE(5268), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7458), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [208339] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7581), 1,
      sym__lowercase_identifier,
    ACTIONS(7583), 1,
      anon_sym_LBRACK,
    ACTIONS(7585), 1,
      anon_sym_virtual,
    STATE(3978), 1,
      sym_class_binding,
    STATE(5269), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208362] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7677), 1,
      anon_sym_LPAREN,
    ACTIONS(7983), 1,
      sym__lowercase_identifier,
    STATE(4614), 1,
      sym__value_name,
    STATE(4808), 1,
      sym_parenthesized_operator,
    STATE(5270), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208385] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(7985), 1,
      anon_sym_GT_RBRACE,
    STATE(5271), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(6195), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208408] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(7987), 1,
      anon_sym_DOT,
    STATE(5272), 1,
      sym_attribute,
    STATE(5306), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6205), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208431] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5781), 1,
      anon_sym_in,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5273), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208454] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7989), 1,
      sym__lowercase_identifier,
    STATE(5274), 1,
      sym_attribute,
    STATE(5421), 1,
      aux_sym_method_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7865), 2,
      anon_sym_private,
      anon_sym_virtual,
  [208475] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(7991), 1,
      anon_sym_DASH_GT,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5275), 1,
      sym_attribute,
    STATE(5289), 1,
      aux_sym_module_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208498] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7993), 1,
      sym__lowercase_identifier,
    ACTIONS(7995), 1,
      anon_sym_LPAREN,
    STATE(5276), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4938), 2,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
  [208519] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(7991), 1,
      anon_sym_DASH_GT,
    STATE(3310), 1,
      aux_sym_module_binding_repeat1,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5277), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208542] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7997), 1,
      sym__lowercase_identifier,
    ACTIONS(7999), 1,
      anon_sym_LPAREN,
    STATE(2977), 1,
      sym__value_name,
    STATE(3035), 1,
      sym_parenthesized_operator,
    STATE(5278), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208565] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8003), 1,
      anon_sym_RBRACK,
    ACTIONS(8005), 1,
      anon_sym_GT,
    STATE(5279), 1,
      sym_attribute,
    STATE(5341), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208588] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5785), 1,
      anon_sym_in,
    STATE(5280), 1,
      sym_attribute,
    STATE(5295), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208611] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(8003), 1,
      anon_sym_RBRACK,
    STATE(5281), 1,
      sym_attribute,
    STATE(5496), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6146), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208634] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8007), 1,
      sym__lowercase_identifier,
    ACTIONS(8009), 1,
      anon_sym_LPAREN,
    STATE(1954), 1,
      sym__value_name,
    STATE(2035), 1,
      sym_parenthesized_operator,
    STATE(5282), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208657] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2916), 1,
      anon_sym_POUND,
    ACTIONS(7935), 1,
      anon_sym_DOT,
    ACTIONS(8011), 1,
      anon_sym_with,
    ACTIONS(8013), 1,
      sym_hash_operator,
    STATE(5283), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208680] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7675), 1,
      sym__lowercase_identifier,
    ACTIONS(7677), 1,
      anon_sym_LPAREN,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(5284), 1,
      sym_attribute,
    STATE(5970), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208703] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5785), 1,
      anon_sym_in,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5285), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208726] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7989), 1,
      sym__lowercase_identifier,
    STATE(5286), 1,
      sym_attribute,
    STATE(5485), 1,
      aux_sym_method_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7865), 2,
      anon_sym_private,
      anon_sym_virtual,
  [208747] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(8015), 1,
      anon_sym_GT_RBRACE,
    STATE(5287), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(6195), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208770] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    STATE(5288), 1,
      sym_attribute,
    STATE(6266), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8017), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [208791] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(8019), 1,
      anon_sym_DASH_GT,
    STATE(3310), 1,
      aux_sym_module_binding_repeat1,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5289), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208814] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5731), 1,
      anon_sym_in,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5290), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208837] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(8021), 1,
      sym__lowercase_identifier,
    STATE(5001), 1,
      sym_type_constructor_path,
    STATE(5291), 1,
      sym_attribute,
    STATE(5977), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208860] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(8023), 1,
      anon_sym_GT_RBRACE,
    STATE(5292), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(6195), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208883] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(8025), 1,
      anon_sym_DASH_GT,
    STATE(3310), 1,
      aux_sym_module_binding_repeat1,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5293), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208906] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    STATE(2104), 1,
      sym_class_path,
    STATE(5294), 1,
      sym_attribute,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208929] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5793), 1,
      anon_sym_in,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5295), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208952] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7967), 1,
      sym__lowercase_identifier,
    ACTIONS(7971), 1,
      anon_sym_mutable,
    ACTIONS(8027), 1,
      anon_sym_RBRACE,
    STATE(5296), 1,
      sym_attribute,
    STATE(6106), 1,
      sym_field_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208975] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5811), 1,
      anon_sym_in,
    STATE(5290), 1,
      aux_sym_expression_item_repeat1,
    STATE(5297), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [208998] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    STATE(5298), 1,
      sym_attribute,
    STATE(5989), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8029), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [209019] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7933), 1,
      anon_sym_as,
    STATE(5299), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6659), 3,
      anon_sym_EQ,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [209038] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5715), 1,
      anon_sym_in,
    STATE(5300), 1,
      sym_attribute,
    STATE(5458), 1,
      aux_sym_expression_item_repeat1,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209061] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(8031), 1,
      anon_sym_GT_RBRACE,
    STATE(5301), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(6195), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209084] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(8033), 1,
      anon_sym_DOT,
    STATE(5302), 1,
      sym_attribute,
    STATE(5306), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(6205), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209107] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(7875), 1,
      anon_sym_COMMA,
    ACTIONS(8035), 1,
      anon_sym_RPAREN,
    STATE(5303), 1,
      sym_attribute,
    STATE(5547), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209130] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5858), 1,
      anon_sym_in,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5304), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209153] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    STATE(5305), 1,
      sym_attribute,
    STATE(5978), 1,
      sym_type_constructor_path,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209176] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8037), 1,
      anon_sym_DOT,
    ACTIONS(8039), 1,
      aux_sym_type_variable_token1,
    STATE(6205), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5306), 2,
      sym_attribute,
      aux_sym_constructor_declaration_repeat1,
  [209197] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7677), 1,
      anon_sym_LPAREN,
    ACTIONS(8042), 1,
      sym__lowercase_identifier,
    STATE(5307), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4324), 2,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
  [209218] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(8044), 1,
      anon_sym_GT_RBRACE,
    STATE(5308), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(6195), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209241] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5874), 1,
      anon_sym_in,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5309), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209264] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5340), 1,
      anon_sym_in,
    ACTIONS(8046), 1,
      anon_sym_and,
    ACTIONS(8048), 1,
      sym_let_and_operator,
    STATE(5310), 1,
      sym_attribute,
    STATE(5535), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209287] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(7875), 1,
      anon_sym_COMMA,
    ACTIONS(7879), 1,
      anon_sym_RBRACK,
    STATE(5311), 1,
      sym_attribute,
    STATE(5826), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209310] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8050), 1,
      anon_sym_RBRACK,
    ACTIONS(8052), 1,
      anon_sym_GT,
    STATE(5312), 1,
      sym_attribute,
    STATE(5341), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209333] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3694), 1,
      anon_sym_RBRACK,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8054), 1,
      anon_sym_GT,
    STATE(5313), 1,
      sym_attribute,
    STATE(5527), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209356] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(8056), 1,
      anon_sym_GT_RBRACE,
    STATE(5314), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(6195), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209379] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8058), 1,
      anon_sym_RBRACK,
    ACTIONS(8060), 1,
      anon_sym_GT,
    STATE(5279), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5315), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209402] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(8062), 1,
      sym__lowercase_identifier,
    STATE(5316), 1,
      sym_attribute,
    STATE(5403), 1,
      sym_class_type_path,
    STATE(5975), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209425] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2916), 1,
      anon_sym_POUND,
    ACTIONS(7935), 1,
      anon_sym_DOT,
    ACTIONS(8013), 1,
      sym_hash_operator,
    ACTIONS(8064), 1,
      anon_sym_with,
    STATE(5317), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209448] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(8066), 1,
      sym__lowercase_identifier,
    STATE(2333), 1,
      sym_class_path,
    STATE(5318), 1,
      sym_attribute,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209471] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    STATE(2378), 1,
      sym_class_path,
    STATE(5319), 1,
      sym_attribute,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209494] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(8068), 1,
      anon_sym_DASH_GT,
    STATE(3310), 1,
      aux_sym_module_binding_repeat1,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5320), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209517] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8058), 1,
      anon_sym_RBRACK,
    ACTIONS(8060), 1,
      anon_sym_GT,
    STATE(5321), 1,
      sym_attribute,
    STATE(5341), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209540] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3136), 1,
      anon_sym_COMMA,
    ACTIONS(7326), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_as,
    ACTIONS(7332), 1,
      anon_sym_COLON_COLON,
    STATE(5322), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209563] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5878), 1,
      anon_sym_in,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5323), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209586] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(8070), 1,
      sym__lowercase_identifier,
    STATE(2447), 1,
      sym_type_constructor_path,
    STATE(5324), 1,
      sym_attribute,
    STATE(6015), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209609] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(8072), 1,
      anon_sym_GT_RBRACE,
    STATE(5325), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(6195), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209632] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(8074), 1,
      anon_sym_DASH_GT,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5326), 1,
      sym_attribute,
    STATE(5349), 1,
      aux_sym_module_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209655] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(8074), 1,
      anon_sym_DASH_GT,
    STATE(3310), 1,
      aux_sym_module_binding_repeat1,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5327), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209678] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(8076), 1,
      anon_sym_DASH_GT,
    STATE(3310), 1,
      aux_sym_module_binding_repeat1,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5328), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209701] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5908), 1,
      anon_sym_in,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5329), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209724] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7997), 1,
      sym__lowercase_identifier,
    ACTIONS(7999), 1,
      anon_sym_LPAREN,
    STATE(3011), 1,
      sym__value_name,
    STATE(3035), 1,
      sym_parenthesized_operator,
    STATE(5330), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209747] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(8078), 1,
      anon_sym_DOT,
    STATE(5306), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(5331), 1,
      sym_attribute,
    STATE(6205), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209770] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(8080), 1,
      anon_sym_RBRACK,
    STATE(5332), 1,
      sym_attribute,
    STATE(5496), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6146), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209793] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7320), 1,
      anon_sym_with,
    ACTIONS(7694), 1,
      anon_sym_DASH_GT,
    STATE(5333), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5064), 2,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [209814] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5334), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3887), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [209831] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    STATE(5335), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    STATE(6314), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209854] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(8082), 1,
      sym__lowercase_identifier,
    STATE(2255), 1,
      sym_type_constructor_path,
    STATE(5336), 1,
      sym_attribute,
    STATE(5949), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209877] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5337), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(3891), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [209894] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7967), 1,
      sym__lowercase_identifier,
    ACTIONS(7971), 1,
      anon_sym_mutable,
    ACTIONS(8084), 1,
      anon_sym_RBRACE,
    STATE(5338), 1,
      sym_attribute,
    STATE(6106), 1,
      sym_field_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209917] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(8086), 1,
      anon_sym_GT_RBRACE,
    STATE(5339), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(6195), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209940] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(7875), 1,
      anon_sym_COMMA,
    ACTIONS(7961), 1,
      anon_sym_RBRACK,
    STATE(5340), 1,
      sym_attribute,
    STATE(5650), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [209963] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8088), 1,
      anon_sym_PIPE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8091), 2,
      anon_sym_RBRACK,
      anon_sym_GT,
    STATE(5341), 2,
      sym_attribute,
      aux_sym_polymorphic_variant_type_repeat1,
  [209982] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8093), 1,
      sym__lowercase_identifier,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8096), 2,
      anon_sym_RPAREN,
      anon_sym_DOT,
    STATE(5342), 2,
      sym_attribute,
      aux_sym__abstract_type_repeat1,
  [210001] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(8098), 1,
      sym__lowercase_identifier,
    STATE(3094), 1,
      sym_class_type_path,
    STATE(5343), 1,
      sym_attribute,
    STATE(6250), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210024] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(7875), 1,
      anon_sym_COMMA,
    ACTIONS(7939), 1,
      anon_sym_RBRACK,
    STATE(5344), 1,
      sym_attribute,
    STATE(5783), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210047] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(7875), 1,
      anon_sym_COMMA,
    ACTIONS(8100), 1,
      anon_sym_RPAREN,
    STATE(5345), 1,
      sym_attribute,
    STATE(5586), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210070] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7677), 1,
      anon_sym_LPAREN,
    ACTIONS(8042), 1,
      sym__lowercase_identifier,
    STATE(5346), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4352), 2,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
  [210091] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7675), 1,
      sym__lowercase_identifier,
    ACTIONS(7677), 1,
      anon_sym_LPAREN,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(5347), 1,
      sym_attribute,
    STATE(6145), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210114] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(8062), 1,
      sym__lowercase_identifier,
    STATE(5348), 1,
      sym_attribute,
    STATE(5448), 1,
      sym_class_type_path,
    STATE(5975), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210137] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(8102), 1,
      anon_sym_DASH_GT,
    STATE(3310), 1,
      aux_sym_module_binding_repeat1,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5349), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210160] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    STATE(5350), 1,
      sym_attribute,
    STATE(6005), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8104), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [210181] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    STATE(5351), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4964), 2,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [210202] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(8108), 1,
      anon_sym_DASH_GT,
    STATE(3310), 1,
      aux_sym_module_binding_repeat1,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5352), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210225] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7430), 1,
      sym__lowercase_identifier,
    ACTIONS(7434), 1,
      anon_sym_LBRACK,
    ACTIONS(7436), 1,
      anon_sym_virtual,
    STATE(5353), 1,
      sym_attribute,
    STATE(6121), 1,
      sym_class_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210248] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2916), 1,
      anon_sym_POUND,
    ACTIONS(2966), 1,
      anon_sym_COMMA,
    ACTIONS(7935), 1,
      anon_sym_DOT,
    ACTIONS(8013), 1,
      sym_hash_operator,
    STATE(5354), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210271] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    STATE(5355), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5056), 2,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [210292] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(8110), 1,
      anon_sym_DOT,
    STATE(5306), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(5356), 1,
      sym_attribute,
    STATE(6205), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210315] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(8112), 1,
      sym__lowercase_identifier,
    STATE(4442), 1,
      sym_class_type_path,
    STATE(5357), 1,
      sym_attribute,
    STATE(6014), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210338] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8114), 1,
      sym__lowercase_identifier,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8116), 2,
      anon_sym_mutable,
      anon_sym_virtual,
    STATE(5358), 2,
      sym_attribute,
      aux_sym_instance_variable_specification_repeat1,
  [210357] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(8119), 1,
      anon_sym_DASH_GT,
    STATE(3310), 1,
      aux_sym_module_binding_repeat1,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5359), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210380] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(8076), 1,
      anon_sym_DASH_GT,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5360), 1,
      sym_attribute,
    STATE(5471), 1,
      aux_sym_module_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210403] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6881), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    STATE(5361), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210426] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(7875), 1,
      anon_sym_COMMA,
    ACTIONS(8121), 1,
      anon_sym_RBRACK,
    STATE(5362), 1,
      sym_attribute,
    STATE(5645), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210449] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(8123), 1,
      anon_sym_GT_RBRACE,
    STATE(5363), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(6195), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210472] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3658), 1,
      anon_sym_RBRACK,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8125), 1,
      anon_sym_GT,
    STATE(5321), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5364), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210495] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7605), 1,
      sym__lowercase_identifier,
    ACTIONS(7607), 1,
      anon_sym_LBRACK,
    ACTIONS(7609), 1,
      anon_sym_virtual,
    STATE(5365), 1,
      sym_attribute,
    STATE(5672), 1,
      sym_class_type_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210518] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5406), 1,
      anon_sym_in,
    ACTIONS(8046), 1,
      anon_sym_and,
    ACTIONS(8048), 1,
      sym_let_and_operator,
    STATE(5366), 1,
      sym_attribute,
    STATE(5411), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210541] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5846), 1,
      anon_sym_in,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5367), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210564] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    STATE(5368), 1,
      sym_attribute,
    STATE(6280), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8127), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [210585] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7320), 1,
      anon_sym_with,
    ACTIONS(7694), 1,
      anon_sym_DASH_GT,
    STATE(5369), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5052), 2,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [210606] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(8129), 1,
      anon_sym_GT_RBRACE,
    STATE(5370), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(6195), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210629] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(8098), 1,
      sym__lowercase_identifier,
    STATE(3000), 1,
      sym_class_type_path,
    STATE(5371), 1,
      sym_attribute,
    STATE(6250), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210652] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(8131), 1,
      anon_sym_DOT,
    STATE(5306), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(5372), 1,
      sym_attribute,
    STATE(6205), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210675] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(8133), 1,
      anon_sym_GT_RBRACE,
    STATE(5373), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(6195), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210698] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(8135), 1,
      anon_sym_DOT,
    STATE(5306), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(5374), 1,
      sym_attribute,
    STATE(6205), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210721] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8137), 1,
      anon_sym_EQ,
    ACTIONS(8139), 1,
      anon_sym_RPAREN,
    ACTIONS(8141), 1,
      anon_sym_COLON2,
    STATE(5375), 1,
      sym_attribute,
    STATE(6018), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210744] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3710), 1,
      anon_sym_RBRACK,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8143), 1,
      anon_sym_GT,
    STATE(5376), 1,
      sym_attribute,
    STATE(5444), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210767] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(8145), 1,
      anon_sym_GT_RBRACE,
    STATE(5377), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(6195), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210790] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4691), 1,
      sym__lowercase_identifier,
    STATE(4337), 1,
      sym_type_constructor_path,
    STATE(5378), 1,
      sym_attribute,
    STATE(6026), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210813] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(8147), 1,
      sym__lowercase_identifier,
    STATE(5379), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    STATE(7108), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210836] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(8149), 1,
      sym__lowercase_identifier,
    STATE(3447), 1,
      sym_class_type_path,
    STATE(5380), 1,
      sym_attribute,
    STATE(6336), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210859] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    STATE(5381), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5064), 2,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [210880] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7320), 1,
      anon_sym_with,
    ACTIONS(7694), 1,
      anon_sym_DASH_GT,
    STATE(5382), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5056), 2,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [210901] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(8151), 1,
      sym__lowercase_identifier,
    STATE(3080), 1,
      sym_type_constructor_path,
    STATE(5383), 1,
      sym_attribute,
    STATE(5996), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210924] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(8050), 1,
      anon_sym_RBRACK,
    STATE(5384), 1,
      sym_attribute,
    STATE(5496), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6146), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210947] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7631), 1,
      sym__lowercase_identifier,
    STATE(1910), 1,
      sym_class_path,
    STATE(5385), 1,
      sym_attribute,
    STATE(6503), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210970] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6901), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    STATE(5386), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [210993] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7967), 1,
      sym__lowercase_identifier,
    ACTIONS(7971), 1,
      anon_sym_mutable,
    ACTIONS(8153), 1,
      anon_sym_RBRACE,
    STATE(5387), 1,
      sym_attribute,
    STATE(6106), 1,
      sym_field_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211016] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(7875), 1,
      anon_sym_COMMA,
    ACTIONS(8155), 1,
      anon_sym_RBRACK,
    STATE(5388), 1,
      sym_attribute,
    STATE(5872), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211039] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8157), 1,
      aux_sym_attribute_id_token1,
    STATE(5389), 1,
      sym_attribute,
    STATE(5453), 1,
      aux_sym_attribute_id_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2743), 2,
      sym__left_quoted_string_delimiter,
      aux_sym_quoted_extension_token1,
  [211060] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3152), 1,
      anon_sym_COMMA,
    ACTIONS(7326), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_as,
    ACTIONS(7332), 1,
      anon_sym_COLON_COLON,
    STATE(5390), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211083] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2916), 1,
      anon_sym_POUND,
    ACTIONS(7935), 1,
      anon_sym_DOT,
    ACTIONS(8013), 1,
      sym_hash_operator,
    ACTIONS(8159), 1,
      anon_sym_with,
    STATE(5391), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211106] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5850), 1,
      anon_sym_in,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5392), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211129] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7675), 1,
      sym__lowercase_identifier,
    ACTIONS(7677), 1,
      anon_sym_LPAREN,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(5393), 1,
      sym_attribute,
    STATE(6323), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211152] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5815), 1,
      anon_sym_in,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5394), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211175] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7671), 1,
      sym__lowercase_identifier,
    STATE(1015), 1,
      sym_class_path,
    STATE(5395), 1,
      sym_attribute,
    STATE(6606), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211198] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5797), 1,
      anon_sym_in,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5396), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211221] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6867), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    STATE(5397), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211244] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7706), 1,
      sym__lowercase_identifier,
    STATE(2114), 1,
      sym_class_path,
    STATE(5398), 1,
      sym_attribute,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211267] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(8161), 1,
      anon_sym_GT_RBRACE,
    STATE(5399), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(6195), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211290] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8163), 1,
      sym__lowercase_identifier,
    STATE(5400), 1,
      sym_attribute,
    STATE(5485), 1,
      aux_sym_method_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7865), 2,
      anon_sym_private,
      anon_sym_virtual,
  [211311] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5401), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4853), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [211328] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(8165), 1,
      anon_sym_DASH_GT,
    STATE(3310), 1,
      aux_sym_module_binding_repeat1,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5402), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211351] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5403), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4788), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [211368] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(7979), 1,
      sym__lowercase_identifier,
    STATE(3446), 1,
      sym_class_type_path,
    STATE(5404), 1,
      sym_attribute,
    STATE(5974), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211391] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7665), 1,
      sym__lowercase_identifier,
    ACTIONS(7667), 1,
      anon_sym_LBRACK,
    ACTIONS(7669), 1,
      anon_sym_virtual,
    STATE(3973), 1,
      sym_class_type_binding,
    STATE(5405), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211414] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    STATE(5406), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5052), 2,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [211435] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8167), 1,
      anon_sym_RBRACK,
    ACTIONS(8169), 1,
      anon_sym_GT,
    STATE(5312), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5407), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211458] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7677), 1,
      anon_sym_LPAREN,
    ACTIONS(8042), 1,
      sym__lowercase_identifier,
    STATE(5408), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(4399), 2,
      sym__simple_value_pattern,
      sym_parenthesized_operator,
  [211479] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2916), 1,
      anon_sym_POUND,
    ACTIONS(7935), 1,
      anon_sym_DOT,
    ACTIONS(8013), 1,
      sym_hash_operator,
    ACTIONS(8171), 1,
      anon_sym_with,
    STATE(5409), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211502] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7444), 1,
      sym__lowercase_identifier,
    ACTIONS(7448), 1,
      anon_sym_LBRACK,
    ACTIONS(7450), 1,
      anon_sym_virtual,
    STATE(4147), 1,
      sym_class_binding,
    STATE(5410), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211525] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5520), 1,
      anon_sym_in,
    ACTIONS(8046), 1,
      anon_sym_and,
    ACTIONS(8048), 1,
      sym_let_and_operator,
    STATE(5411), 1,
      sym_attribute,
    STATE(5535), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211548] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(8173), 1,
      anon_sym_RBRACK,
    STATE(5412), 1,
      sym_attribute,
    STATE(5496), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6146), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211571] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    STATE(5413), 1,
      sym_attribute,
    STATE(6221), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8175), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [211592] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(7320), 1,
      anon_sym_with,
    ACTIONS(7694), 1,
      anon_sym_DASH_GT,
    STATE(5414), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5060), 2,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [211613] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    STATE(5415), 1,
      sym_attribute,
    STATE(6036), 1,
      sym_type_constructor_path,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211636] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    ACTIONS(7975), 1,
      sym__lowercase_identifier,
    STATE(5416), 1,
      sym_attribute,
    STATE(6041), 1,
      sym_module_type_path,
    STATE(6325), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211659] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7815), 1,
      sym__lowercase_identifier,
    STATE(2114), 1,
      sym_class_path,
    STATE(5417), 1,
      sym_attribute,
    STATE(6649), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211682] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7665), 1,
      sym__lowercase_identifier,
    ACTIONS(7667), 1,
      anon_sym_LBRACK,
    ACTIONS(7669), 1,
      anon_sym_virtual,
    STATE(4039), 1,
      sym_class_type_binding,
    STATE(5418), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211705] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2916), 1,
      anon_sym_POUND,
    ACTIONS(7935), 1,
      anon_sym_DOT,
    ACTIONS(8013), 1,
      sym_hash_operator,
    ACTIONS(8177), 1,
      anon_sym_with,
    STATE(5419), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211728] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(8179), 1,
      anon_sym_GT_RBRACE,
    STATE(5420), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(6195), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211751] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8181), 1,
      sym__lowercase_identifier,
    STATE(5421), 1,
      sym_attribute,
    STATE(5485), 1,
      aux_sym_method_specification_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7865), 2,
      anon_sym_private,
      anon_sym_virtual,
  [211772] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5406), 1,
      anon_sym_in,
    ACTIONS(8046), 1,
      anon_sym_and,
    ACTIONS(8048), 1,
      sym_let_and_operator,
    STATE(5422), 1,
      sym_attribute,
    STATE(5535), 1,
      aux_sym_value_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211795] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5811), 1,
      anon_sym_in,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5423), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211818] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8183), 1,
      anon_sym_rec,
    STATE(5424), 1,
      sym_attribute,
    STATE(5764), 1,
      sym_module_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7390), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [211839] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3104), 1,
      anon_sym_COMMA,
    ACTIONS(7326), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_as,
    ACTIONS(7332), 1,
      anon_sym_COLON_COLON,
    STATE(5425), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211862] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5426), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5835), 4,
      anon_sym_and,
      anon_sym_constraint,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [211879] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2916), 1,
      anon_sym_POUND,
    ACTIONS(7935), 1,
      anon_sym_DOT,
    ACTIONS(8013), 1,
      sym_hash_operator,
    ACTIONS(8185), 1,
      anon_sym_with,
    STATE(5427), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211902] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(8187), 1,
      anon_sym_RBRACK,
    STATE(5428), 1,
      sym_attribute,
    STATE(5496), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6146), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211925] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7677), 1,
      anon_sym_LPAREN,
    ACTIONS(7871), 1,
      sym__lowercase_identifier,
    STATE(3501), 1,
      sym_parenthesized_operator,
    STATE(4614), 1,
      sym__value_name,
    STATE(5429), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211948] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7995), 1,
      anon_sym_LPAREN,
    ACTIONS(8189), 1,
      sym__lowercase_identifier,
    STATE(4863), 1,
      sym_parenthesized_operator,
    STATE(4886), 1,
      sym__value_name,
    STATE(5430), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211971] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(8112), 1,
      sym__lowercase_identifier,
    STATE(4490), 1,
      sym_class_type_path,
    STATE(5431), 1,
      sym_attribute,
    STATE(6014), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [211994] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(8098), 1,
      sym__lowercase_identifier,
    STATE(2985), 1,
      sym_class_type_path,
    STATE(5432), 1,
      sym_attribute,
    STATE(6250), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212017] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4887), 1,
      anon_sym_and,
    STATE(5433), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4885), 3,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [212036] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(8191), 1,
      anon_sym_GT_RBRACE,
    STATE(5434), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(6195), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212059] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2916), 1,
      anon_sym_POUND,
    ACTIONS(7935), 1,
      anon_sym_DOT,
    ACTIONS(8013), 1,
      sym_hash_operator,
    ACTIONS(8193), 1,
      anon_sym_with,
    STATE(5435), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212082] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(8195), 1,
      sym__lowercase_identifier,
    STATE(5436), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    STATE(6488), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212105] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5437), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4873), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [212122] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8197), 1,
      sym__lowercase_identifier,
    STATE(5400), 1,
      aux_sym_method_specification_repeat1,
    STATE(5438), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7865), 2,
      anon_sym_private,
      anon_sym_virtual,
  [212143] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3132), 1,
      anon_sym_COMMA,
    ACTIONS(7326), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_as,
    ACTIONS(7332), 1,
      anon_sym_COLON_COLON,
    STATE(5439), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212166] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(8199), 1,
      anon_sym_DASH_GT,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5359), 1,
      aux_sym_module_binding_repeat1,
    STATE(5440), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212189] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7724), 1,
      sym__lowercase_identifier,
    STATE(1121), 1,
      sym_class_path,
    STATE(5441), 1,
      sym_attribute,
    STATE(6593), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212212] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7605), 1,
      sym__lowercase_identifier,
    ACTIONS(7607), 1,
      anon_sym_LBRACK,
    ACTIONS(7609), 1,
      anon_sym_virtual,
    STATE(5442), 1,
      sym_attribute,
    STATE(6315), 1,
      sym_class_type_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212235] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2916), 1,
      anon_sym_POUND,
    ACTIONS(7935), 1,
      anon_sym_DOT,
    ACTIONS(8013), 1,
      sym_hash_operator,
    ACTIONS(8201), 1,
      anon_sym_with,
    STATE(5443), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212258] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8167), 1,
      anon_sym_RBRACK,
    ACTIONS(8169), 1,
      anon_sym_GT,
    STATE(5341), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5444), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212281] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8205), 1,
      anon_sym_COMMA,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8203), 2,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    STATE(5445), 2,
      sym_attribute,
      aux_sym_instantiated_class_type_repeat1,
  [212300] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(8199), 1,
      anon_sym_DASH_GT,
    STATE(3310), 1,
      aux_sym_module_binding_repeat1,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5446), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212323] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5447), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5372), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [212340] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5448), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4784), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [212357] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5449), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4800), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [212374] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2916), 1,
      anon_sym_POUND,
    ACTIONS(7935), 1,
      anon_sym_DOT,
    ACTIONS(8013), 1,
      sym_hash_operator,
    ACTIONS(8208), 1,
      anon_sym_with,
    STATE(5450), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212397] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5464), 1,
      anon_sym_in,
    ACTIONS(8046), 1,
      anon_sym_and,
    ACTIONS(8048), 1,
      sym_let_and_operator,
    STATE(5310), 1,
      aux_sym_value_definition_repeat1,
    STATE(5451), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212420] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(7979), 1,
      sym__lowercase_identifier,
    STATE(2603), 1,
      sym_class_type_path,
    STATE(5452), 1,
      sym_attribute,
    STATE(5974), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212443] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8157), 1,
      aux_sym_attribute_id_token1,
    STATE(5453), 1,
      sym_attribute,
    STATE(5481), 1,
      aux_sym_attribute_id_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2733), 2,
      sym__left_quoted_string_delimiter,
      aux_sym_quoted_extension_token1,
  [212464] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    STATE(5454), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7877), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [212483] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(7875), 1,
      anon_sym_COMMA,
    ACTIONS(8210), 1,
      anon_sym_RBRACK,
    STATE(5455), 1,
      sym_attribute,
    STATE(5730), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212506] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2916), 1,
      anon_sym_POUND,
    ACTIONS(7935), 1,
      anon_sym_DOT,
    ACTIONS(8013), 1,
      sym_hash_operator,
    ACTIONS(8212), 1,
      anon_sym_with,
    STATE(5456), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212529] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7015), 1,
      anon_sym_COMMA,
    ACTIONS(7104), 1,
      anon_sym_PIPE,
    ACTIONS(7108), 1,
      anon_sym_as,
    ACTIONS(7110), 1,
      anon_sym_COLON_COLON,
    STATE(5457), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212552] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5727), 1,
      anon_sym_in,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5458), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212575] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(8098), 1,
      sym__lowercase_identifier,
    STATE(3109), 1,
      sym_class_type_path,
    STATE(5459), 1,
      sym_attribute,
    STATE(6250), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212598] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(8214), 1,
      anon_sym_GT_RBRACE,
    STATE(5460), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(6195), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212621] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5727), 1,
      anon_sym_in,
    STATE(5273), 1,
      aux_sym_expression_item_repeat1,
    STATE(5461), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212644] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5743), 1,
      anon_sym_in,
    STATE(5285), 1,
      aux_sym_expression_item_repeat1,
    STATE(5462), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212667] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(7979), 1,
      sym__lowercase_identifier,
    STATE(2629), 1,
      sym_class_type_path,
    STATE(5463), 1,
      sym_attribute,
    STATE(5974), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212690] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7933), 1,
      anon_sym_as,
    STATE(5464), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6013), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_GT,
  [212709] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7675), 1,
      sym__lowercase_identifier,
    ACTIONS(7677), 1,
      anon_sym_LPAREN,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(5465), 1,
      sym_attribute,
    STATE(6153), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212732] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7933), 1,
      anon_sym_as,
    STATE(5466), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6038), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_GT,
  [212751] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(8216), 1,
      sym__lowercase_identifier,
    STATE(5467), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    STATE(6841), 1,
      sym_type_constructor_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212774] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7706), 1,
      sym__lowercase_identifier,
    STATE(797), 1,
      sym_class_path,
    STATE(5468), 1,
      sym_attribute,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212797] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(8218), 1,
      anon_sym_GT_RBRACE,
    STATE(5469), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(6195), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212820] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5773), 1,
      anon_sym_in,
    STATE(5187), 1,
      aux_sym_expression_item_repeat1,
    STATE(5470), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212843] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(8220), 1,
      anon_sym_DASH_GT,
    STATE(3310), 1,
      aux_sym_module_binding_repeat1,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5471), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212866] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    STATE(5472), 1,
      sym_attribute,
    STATE(6168), 1,
      sym_type_constructor_path,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212889] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2972), 1,
      anon_sym_COMMA,
    ACTIONS(7326), 1,
      anon_sym_PIPE,
    ACTIONS(7330), 1,
      anon_sym_as,
    ACTIONS(7332), 1,
      anon_sym_COLON_COLON,
    STATE(5473), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212912] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7817), 1,
      sym__lowercase_identifier,
    STATE(1555), 1,
      sym_class_path,
    STATE(5474), 1,
      sym_attribute,
    STATE(6688), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212935] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(8222), 1,
      anon_sym_RBRACK,
    STATE(5475), 1,
      sym_attribute,
    STATE(5496), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6146), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212958] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(7875), 1,
      anon_sym_COMMA,
    ACTIONS(8224), 1,
      anon_sym_RPAREN,
    STATE(5476), 1,
      sym_attribute,
    STATE(5846), 1,
      aux_sym_instantiated_class_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [212981] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    STATE(5477), 1,
      sym_attribute,
    STATE(5965), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8226), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [213002] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3939), 1,
      anon_sym_DQUOTE,
    ACTIONS(8228), 1,
      anon_sym_LBRACE,
    STATE(5478), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(3820), 2,
      sym_string,
      sym_quoted_string,
  [213023] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7933), 1,
      anon_sym_as,
    STATE(5479), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6017), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_GT,
  [213042] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8232), 1,
      anon_sym_COMMA,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8230), 2,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    STATE(5480), 2,
      sym_attribute,
      aux_sym__type_params_repeat1,
  [213061] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(8235), 1,
      aux_sym_attribute_id_token1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2724), 2,
      sym__left_quoted_string_delimiter,
      aux_sym_quoted_extension_token1,
    STATE(5481), 2,
      sym_attribute,
      aux_sym_attribute_id_repeat1,
  [213080] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(8238), 1,
      anon_sym_GT_RBRACE,
    STATE(5482), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(6195), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213103] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7743), 1,
      sym__lowercase_identifier,
    STATE(920), 1,
      sym_class_path,
    STATE(5483), 1,
      sym_attribute,
    STATE(6725), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213126] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    ACTIONS(8240), 1,
      anon_sym_RBRACK,
    STATE(5484), 1,
      sym_attribute,
    STATE(5496), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(6146), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213149] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8242), 1,
      sym__lowercase_identifier,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8244), 2,
      anon_sym_private,
      anon_sym_virtual,
    STATE(5485), 2,
      sym_attribute,
      aux_sym_method_specification_repeat1,
  [213168] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(8247), 1,
      anon_sym_DASH_GT,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5486), 1,
      sym_attribute,
    STATE(5488), 1,
      aux_sym_module_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213191] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(8247), 1,
      anon_sym_DASH_GT,
    STATE(3310), 1,
      aux_sym_module_binding_repeat1,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5487), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213214] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(8249), 1,
      anon_sym_DASH_GT,
    STATE(3310), 1,
      aux_sym_module_binding_repeat1,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5488), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213237] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(8251), 1,
      anon_sym_DOT,
    STATE(5306), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(5489), 1,
      sym_attribute,
    STATE(6205), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213260] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7675), 1,
      sym__lowercase_identifier,
    ACTIONS(7677), 1,
      anon_sym_LPAREN,
    STATE(1662), 1,
      sym_parenthesized_operator,
    STATE(5490), 1,
      sym_attribute,
    STATE(6207), 1,
      sym__value_name,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213283] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(8149), 1,
      sym__lowercase_identifier,
    STATE(3446), 1,
      sym_class_type_path,
    STATE(5491), 1,
      sym_attribute,
    STATE(6336), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213306] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5492), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5827), 4,
      anon_sym_LBRACE,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      anon_sym_DQUOTE,
  [213323] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(8253), 1,
      anon_sym_GT_RBRACE,
    STATE(5493), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(6195), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213346] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    ACTIONS(8255), 1,
      anon_sym_DOT,
    STATE(5306), 1,
      aux_sym_constructor_declaration_repeat1,
    STATE(5494), 1,
      sym_attribute,
    STATE(6205), 1,
      sym_type_variable,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213369] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7973), 1,
      anon_sym_RBRACK,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8257), 1,
      anon_sym_GT,
    STATE(5341), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5495), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213392] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8259), 1,
      anon_sym_RBRACK,
    ACTIONS(8261), 1,
      aux_sym_tag_token1,
    STATE(6146), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5496), 2,
      sym_attribute,
      aux_sym_polymorphic_variant_type_repeat2,
  [213413] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8264), 1,
      anon_sym_rec,
    STATE(4172), 1,
      sym_module_binding,
    STATE(5497), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7595), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [213434] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7706), 1,
      sym__lowercase_identifier,
    STATE(2104), 1,
      sym_class_path,
    STATE(5498), 1,
      sym_attribute,
    STATE(6430), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213457] = 7,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2544), 1,
      anon_sym_LBRACK_AT_AT,
    ACTIONS(5769), 1,
      anon_sym_in,
    STATE(5423), 1,
      aux_sym_expression_item_repeat1,
    STATE(5499), 1,
      sym_attribute,
    STATE(5646), 1,
      sym_item_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213480] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    STATE(5500), 1,
      sym_attribute,
    STATE(6043), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8266), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [213501] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    ACTIONS(8268), 1,
      anon_sym_GT_RBRACE,
    STATE(5501), 1,
      sym_attribute,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(6195), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213524] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4933), 1,
      anon_sym_and,
    STATE(5502), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4931), 3,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [213543] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(4292), 1,
      sym__lowercase_identifier,
    STATE(4337), 1,
      sym_type_constructor_path,
    STATE(5503), 1,
      sym_attribute,
    STATE(6196), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213566] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8270), 1,
      sym__lowercase_identifier,
    STATE(5342), 1,
      aux_sym__abstract_type_repeat1,
    STATE(5504), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8272), 2,
      anon_sym_RPAREN,
      anon_sym_DOT,
  [213587] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5340), 1,
      anon_sym_in,
    ACTIONS(8046), 1,
      anon_sym_and,
    ACTIONS(8048), 1,
      sym_let_and_operator,
    STATE(5422), 1,
      aux_sym_value_definition_repeat1,
    STATE(5505), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213610] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    STATE(5506), 1,
      sym_attribute,
    STATE(6143), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8274), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [213631] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8141), 1,
      anon_sym_COLON2,
    ACTIONS(8276), 1,
      anon_sym_EQ,
    ACTIONS(8278), 1,
      anon_sym_RPAREN,
    STATE(5507), 1,
      sym_attribute,
    STATE(6188), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213654] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7444), 1,
      sym__lowercase_identifier,
    ACTIONS(7448), 1,
      anon_sym_LBRACK,
    ACTIONS(7450), 1,
      anon_sym_virtual,
    STATE(4048), 1,
      sym_class_binding,
    STATE(5508), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213677] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5509), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5682), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [213694] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(8112), 1,
      sym__lowercase_identifier,
    STATE(4464), 1,
      sym_class_type_path,
    STATE(5510), 1,
      sym_attribute,
    STATE(6014), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213717] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(8280), 1,
      sym__lowercase_identifier,
    STATE(1995), 1,
      sym_type_constructor_path,
    STATE(5511), 1,
      sym_attribute,
    STATE(6252), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213740] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7791), 1,
      sym__lowercase_identifier,
    STATE(2032), 1,
      sym_class_path,
    STATE(5512), 1,
      sym_attribute,
    STATE(6794), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213763] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5513), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5290), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [213780] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8282), 1,
      anon_sym_RBRACK,
    ACTIONS(8284), 1,
      anon_sym_GT,
    STATE(5495), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5514), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213803] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    STATE(5515), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5060), 2,
      anon_sym_RPAREN,
      anon_sym_COLON_GT,
  [213824] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5516), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4816), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [213841] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7763), 1,
      sym__lowercase_identifier,
    STATE(2738), 1,
      sym_class_path,
    STATE(5517), 1,
      sym_attribute,
    STATE(6761), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213864] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7581), 1,
      sym__lowercase_identifier,
    ACTIONS(7583), 1,
      anon_sym_LBRACK,
    ACTIONS(7585), 1,
      anon_sym_virtual,
    STATE(4048), 1,
      sym_class_binding,
    STATE(5518), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213887] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(8112), 1,
      sym__lowercase_identifier,
    STATE(4488), 1,
      sym_class_type_path,
    STATE(5519), 1,
      sym_attribute,
    STATE(6014), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213910] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(7979), 1,
      sym__lowercase_identifier,
    STATE(3447), 1,
      sym_class_type_path,
    STATE(5520), 1,
      sym_attribute,
    STATE(5974), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213933] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5521), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4877), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [213950] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    STATE(5522), 1,
      sym_attribute,
    STATE(6326), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8286), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [213971] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(7710), 1,
      sym__lowercase_identifier,
    STATE(1760), 1,
      sym_class_path,
    STATE(5523), 1,
      sym_attribute,
    STATE(6377), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [213994] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    ACTIONS(7979), 1,
      sym__lowercase_identifier,
    STATE(2570), 1,
      sym_class_type_path,
    STATE(5524), 1,
      sym_attribute,
    STATE(5974), 1,
      sym_extended_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214017] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(6563), 1,
      anon_sym_LBRACE,
    STATE(5525), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5426), 2,
      sym_string,
      sym_quoted_string,
  [214038] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5526), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5444), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [214055] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8282), 1,
      anon_sym_RBRACK,
    ACTIONS(8284), 1,
      anon_sym_GT,
    STATE(5341), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5527), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214078] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(4895), 1,
      anon_sym_and,
    STATE(5528), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4893), 3,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
      sym_let_and_operator,
  [214097] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    STATE(5529), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8203), 3,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [214116] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(8288), 1,
      anon_sym_DASH_GT,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5530), 1,
      sym_attribute,
    STATE(5532), 1,
      aux_sym_module_binding_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214139] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(8288), 1,
      anon_sym_DASH_GT,
    STATE(3310), 1,
      aux_sym_module_binding_repeat1,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5531), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214162] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    ACTIONS(8290), 1,
      anon_sym_DASH_GT,
    STATE(3310), 1,
      aux_sym_module_binding_repeat1,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5532), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214185] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(8066), 1,
      sym__lowercase_identifier,
    STATE(2423), 1,
      sym_class_path,
    STATE(5533), 1,
      sym_attribute,
    STATE(6562), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214208] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5534), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4687), 4,
      anon_sym_and,
      anon_sym_EQ,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [214225] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5294), 1,
      anon_sym_in,
    ACTIONS(8292), 1,
      anon_sym_and,
    ACTIONS(8295), 1,
      sym_let_and_operator,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5535), 2,
      sym_attribute,
      aux_sym_value_definition_repeat1,
  [214246] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8007), 1,
      sym__lowercase_identifier,
    ACTIONS(8009), 1,
      anon_sym_LPAREN,
    STATE(1934), 1,
      sym__value_name,
    STATE(2035), 1,
      sym_parenthesized_operator,
    STATE(5536), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214269] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    ACTIONS(7975), 1,
      sym__lowercase_identifier,
    STATE(5537), 1,
      sym_attribute,
    STATE(6325), 1,
      sym_extended_module_path,
    STATE(6327), 1,
      sym_module_type_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214292] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7430), 1,
      sym__lowercase_identifier,
    ACTIONS(7434), 1,
      anon_sym_LBRACK,
    ACTIONS(7436), 1,
      anon_sym_virtual,
    STATE(5538), 1,
      sym_attribute,
    STATE(5803), 1,
      sym_class_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214315] = 7,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4982), 1,
      sym__uppercase_identifier,
    ACTIONS(7975), 1,
      sym__lowercase_identifier,
    STATE(5539), 1,
      sym_attribute,
    STATE(6325), 1,
      sym_extended_module_path,
    STATE(6335), 1,
      sym_module_type_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214338] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8300), 1,
      anon_sym_of,
    STATE(5540), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8298), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [214357] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(23), 1,
      sym_attribute_id,
    STATE(5541), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [214375] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(2863), 1,
      sym_attribute_id,
    STATE(5542), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8304), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [214393] = 5,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2722), 1,
      aux_sym_attribute_id_token1,
    STATE(5543), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(2724), 2,
      sym__left_quoted_string_delimiter,
      aux_sym_quoted_extension_token1,
  [214411] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    ACTIONS(8306), 1,
      anon_sym_RPAREN,
    STATE(5544), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214431] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8308), 1,
      anon_sym_LPAREN,
    ACTIONS(8310), 1,
      anon_sym_LBRACK,
    ACTIONS(8312), 1,
      anon_sym_LBRACE,
    STATE(5545), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214451] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8314), 1,
      anon_sym_SEMI,
    ACTIONS(8316), 1,
      anon_sym_GT_RBRACE,
    STATE(5546), 1,
      sym_attribute,
    STATE(5553), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214471] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7875), 1,
      anon_sym_COMMA,
    ACTIONS(8318), 1,
      anon_sym_RPAREN,
    STATE(5445), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(5547), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214491] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8282), 1,
      anon_sym_RBRACK,
    STATE(5548), 1,
      sym_attribute,
    STATE(5619), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214511] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8282), 1,
      anon_sym_RBRACK,
    STATE(5341), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5549), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214531] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7803), 1,
      anon_sym_RBRACE,
    ACTIONS(8320), 1,
      anon_sym_SEMI,
    STATE(5550), 1,
      sym_attribute,
    STATE(5684), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214551] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5551), 1,
      sym_attribute,
    STATE(5613), 1,
      sym_attribute_id,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [214569] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5552), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8091), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [214585] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8218), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(8324), 1,
      anon_sym_SEMI,
    STATE(5553), 1,
      sym_attribute,
    STATE(5694), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214605] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4467), 1,
      anon_sym_GT,
    ACTIONS(8326), 1,
      anon_sym_SEMI,
    STATE(5554), 1,
      sym_attribute,
    STATE(5581), 1,
      aux_sym_object_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214625] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8328), 1,
      anon_sym_SEMI,
    ACTIONS(8330), 1,
      anon_sym_RBRACE,
    STATE(5555), 1,
      sym_attribute,
    STATE(5558), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214645] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(5484), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(5556), 1,
      sym_attribute,
    STATE(6146), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214665] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(8332), 1,
      sym__lowercase_identifier,
    ACTIONS(8334), 1,
      sym_indexing_operator,
    STATE(5557), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214685] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7807), 1,
      anon_sym_RBRACE,
    ACTIONS(8336), 1,
      anon_sym_SEMI,
    STATE(5558), 1,
      sym_attribute,
    STATE(5684), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214705] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(4527), 1,
      anon_sym_DOT,
    ACTIONS(8338), 1,
      anon_sym_RPAREN,
    STATE(5559), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214725] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(469), 1,
      anon_sym_end,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    STATE(5560), 1,
      sym_attribute,
    STATE(6704), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214745] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6233), 1,
      anon_sym_in,
    ACTIONS(8340), 1,
      anon_sym_and,
    STATE(5561), 1,
      sym_attribute,
    STATE(5744), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214765] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8342), 1,
      anon_sym_SEMI,
    ACTIONS(8344), 1,
      anon_sym_RBRACE,
    STATE(5562), 1,
      sym_attribute,
    STATE(5566), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214785] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6038), 1,
      anon_sym_EQ,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(6589), 1,
      anon_sym_RPAREN,
    STATE(5563), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214805] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8346), 1,
      anon_sym_SEMI,
    ACTIONS(8348), 1,
      anon_sym_GT_RBRACE,
    STATE(5564), 1,
      sym_attribute,
    STATE(5567), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214825] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4942), 1,
      anon_sym_RPAREN,
    STATE(5565), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8350), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [214843] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7603), 1,
      anon_sym_RBRACE,
    ACTIONS(8352), 1,
      anon_sym_SEMI,
    STATE(5566), 1,
      sym_attribute,
    STATE(5684), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214863] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8145), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(8354), 1,
      anon_sym_SEMI,
    STATE(5567), 1,
      sym_attribute,
    STATE(5694), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214883] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8356), 1,
      anon_sym_SEMI,
    ACTIONS(8358), 1,
      anon_sym_RBRACE,
    STATE(5568), 1,
      sym_attribute,
    STATE(5573), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214903] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(28), 1,
      sym_attribute_id,
    STATE(5569), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [214921] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6048), 1,
      anon_sym_in,
    ACTIONS(8360), 1,
      anon_sym_and,
    STATE(5570), 1,
      sym_attribute,
    STATE(5780), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214941] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3658), 1,
      anon_sym_RBRACK,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    STATE(5571), 1,
      sym_attribute,
    STATE(5588), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214961] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(8334), 1,
      sym_indexing_operator,
    ACTIONS(8362), 1,
      sym__lowercase_identifier,
    STATE(5572), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [214981] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7655), 1,
      anon_sym_RBRACE,
    ACTIONS(8364), 1,
      anon_sym_SEMI,
    STATE(5573), 1,
      sym_attribute,
    STATE(5684), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215001] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7271), 1,
      anon_sym_SEMI,
    ACTIONS(7273), 1,
      anon_sym_GT,
    STATE(5574), 1,
      sym_attribute,
    STATE(5590), 1,
      aux_sym_object_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215021] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8366), 1,
      anon_sym_SEMI,
    ACTIONS(8368), 1,
      anon_sym_RBRACE,
    STATE(5575), 1,
      sym_attribute,
    STATE(5660), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215041] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5576), 1,
      sym_attribute,
    STATE(5764), 1,
      sym_module_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7390), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [215059] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8370), 1,
      anon_sym_SEMI,
    ACTIONS(8372), 1,
      anon_sym_GT_RBRACE,
    STATE(5577), 1,
      sym_attribute,
    STATE(5836), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215079] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8374), 1,
      anon_sym_EQ,
    STATE(5578), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8376), 2,
      anon_sym_SEMI,
      anon_sym_GT_RBRACE,
  [215097] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5296), 1,
      anon_sym_and,
    STATE(5579), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5294), 2,
      anon_sym_in,
      sym_let_and_operator,
  [215115] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7819), 1,
      sym__lowercase_identifier,
    STATE(5578), 1,
      sym__instance_variable_name,
    STATE(5580), 1,
      sym_attribute,
    STATE(6195), 1,
      sym_instance_variable_expression,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215135] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7510), 1,
      anon_sym_GT,
    ACTIONS(8378), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5581), 2,
      sym_attribute,
      aux_sym_object_type_repeat1,
  [215153] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5930), 1,
      anon_sym_in,
    ACTIONS(8381), 1,
      anon_sym_and,
    STATE(5582), 1,
      sym_attribute,
    STATE(5795), 1,
      aux_sym_class_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215173] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(8334), 1,
      sym_indexing_operator,
    ACTIONS(8383), 1,
      sym__lowercase_identifier,
    STATE(5583), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215193] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(3784), 1,
      sym_attribute_id,
    STATE(5584), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8385), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [215211] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    ACTIONS(8387), 1,
      anon_sym_RPAREN,
    STATE(5585), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215231] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7875), 1,
      anon_sym_COMMA,
    ACTIONS(8389), 1,
      anon_sym_RPAREN,
    STATE(5445), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(5586), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215251] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8058), 1,
      anon_sym_RBRACK,
    STATE(5587), 1,
      sym_attribute,
    STATE(5602), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215271] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8058), 1,
      anon_sym_RBRACK,
    STATE(5341), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5588), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215291] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8391), 1,
      sym__lowercase_identifier,
    ACTIONS(8393), 1,
      anon_sym_LPAREN,
    STATE(2443), 1,
      sym__simple_value_pattern,
    STATE(5589), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215311] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4402), 1,
      anon_sym_GT,
    ACTIONS(8395), 1,
      anon_sym_SEMI,
    STATE(5581), 1,
      aux_sym_object_type_repeat1,
    STATE(5590), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215331] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1155), 1,
      anon_sym_end,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    STATE(5591), 1,
      sym_attribute,
    STATE(6459), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215351] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(9), 1,
      sym_attribute_id,
    STATE(5592), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [215369] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8397), 1,
      aux_sym_quoted_extension_token1,
    STATE(5593), 1,
      sym_attribute,
    STATE(7163), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215389] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5594), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8230), 3,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [215405] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8399), 1,
      anon_sym_SEMI,
    ACTIONS(8401), 1,
      anon_sym_RBRACE,
    STATE(5595), 1,
      sym_attribute,
    STATE(5615), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215425] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(10), 1,
      sym_attribute_id,
    STATE(5596), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [215443] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(11), 1,
      sym_attribute_id,
    STATE(5597), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [215461] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8403), 1,
      anon_sym_SEMI,
    ACTIONS(8405), 1,
      anon_sym_GT_RBRACE,
    STATE(5598), 1,
      sym_attribute,
    STATE(5668), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215481] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(5264), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(5599), 1,
      sym_attribute,
    STATE(6146), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215501] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8407), 1,
      anon_sym_SEMI,
    ACTIONS(8409), 1,
      anon_sym_GT_RBRACE,
    STATE(5600), 1,
      sym_attribute,
    STATE(5623), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215521] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    ACTIONS(8411), 1,
      anon_sym_RPAREN,
    STATE(5601), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215541] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8003), 1,
      anon_sym_RBRACK,
    STATE(5341), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5602), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215561] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5402), 1,
      aux_sym_module_binding_repeat1,
    STATE(5603), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215581] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8003), 1,
      anon_sym_RBRACK,
    STATE(5604), 1,
      sym_attribute,
    STATE(5630), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215601] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8413), 1,
      anon_sym_SEMI,
    ACTIONS(8416), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5605), 2,
      sym_attribute,
      aux_sym_record_declaration_repeat1,
  [215619] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8418), 1,
      aux_sym_quoted_extension_token1,
    STATE(5606), 1,
      sym_attribute,
    STATE(6942), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215639] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(12), 1,
      sym_attribute_id,
    STATE(5607), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [215657] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(5475), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(5608), 1,
      sym_attribute,
    STATE(6146), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215677] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8420), 1,
      anon_sym_SEMI,
    ACTIONS(8422), 1,
      anon_sym_RBRACE,
    STATE(5609), 1,
      sym_attribute,
    STATE(5635), 1,
      aux_sym_record_binding_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215697] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    ACTIONS(8424), 1,
      anon_sym_RPAREN,
    STATE(5610), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215717] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(8426), 1,
      anon_sym_EQ,
    STATE(5611), 1,
      sym_attribute,
    STATE(6572), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215737] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8428), 1,
      anon_sym_SEMI,
    ACTIONS(8430), 1,
      anon_sym_RBRACE,
    STATE(5612), 1,
      sym_attribute,
    STATE(5638), 1,
      aux_sym_record_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215757] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8432), 1,
      aux_sym_quoted_extension_token1,
    STATE(5613), 1,
      sym_attribute,
    STATE(6551), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215777] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    ACTIONS(8434), 1,
      anon_sym_RPAREN,
    STATE(5614), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215797] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7769), 1,
      anon_sym_RBRACE,
    ACTIONS(8436), 1,
      anon_sym_SEMI,
    STATE(5615), 1,
      sym_attribute,
    STATE(5684), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215817] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(8440), 1,
      anon_sym_RBRACK,
    STATE(5616), 1,
      sym_attribute,
    STATE(5636), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215837] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6120), 1,
      anon_sym_in,
    ACTIONS(8340), 1,
      anon_sym_and,
    STATE(5617), 1,
      sym_attribute,
    STATE(5642), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215857] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1281), 1,
      anon_sym_RBRACK,
    ACTIONS(8442), 1,
      anon_sym_SEMI,
    STATE(5618), 1,
      sym_attribute,
    STATE(5669), 1,
      aux_sym__sequence_expression_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215877] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7973), 1,
      anon_sym_RBRACK,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    STATE(5341), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5619), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215897] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7973), 1,
      anon_sym_RBRACK,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    STATE(5620), 1,
      sym_attribute,
    STATE(5673), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215917] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5621), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8444), 3,
      anon_sym_RPAREN,
      anon_sym_DOT,
      sym__lowercase_identifier,
  [215933] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(8446), 1,
      anon_sym_RBRACK,
    STATE(5480), 1,
      aux_sym__type_params_repeat1,
    STATE(5622), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215953] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8179), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(8448), 1,
      anon_sym_SEMI,
    STATE(5623), 1,
      sym_attribute,
    STATE(5694), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215973] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(8450), 1,
      anon_sym_type,
    STATE(5624), 1,
      sym_attribute,
    STATE(5731), 1,
      sym_module_path,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [215993] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(5266), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(5625), 1,
      sym_attribute,
    STATE(6146), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216013] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2480), 1,
      anon_sym_PIPE_RBRACK,
    ACTIONS(8452), 1,
      anon_sym_SEMI,
    STATE(5626), 1,
      sym_attribute,
    STATE(5640), 1,
      aux_sym__sequence_binding_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216033] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8456), 1,
      anon_sym_LBRACK,
    STATE(5627), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8454), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
  [216051] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(8458), 1,
      anon_sym_EQ,
    STATE(5628), 1,
      sym_attribute,
    STATE(6700), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216071] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    ACTIONS(8460), 1,
      anon_sym_RPAREN,
    STATE(5629), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216091] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8187), 1,
      anon_sym_RBRACK,
    STATE(5341), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5630), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216111] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(8446), 1,
      anon_sym_RBRACK,
    STATE(5631), 1,
      sym_attribute,
    STATE(5770), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216131] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7967), 1,
      sym__lowercase_identifier,
    ACTIONS(7971), 1,
      anon_sym_mutable,
    STATE(5632), 1,
      sym_attribute,
    STATE(5750), 1,
      sym_field_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216151] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(8462), 1,
      anon_sym_RPAREN,
    STATE(5633), 1,
      sym_attribute,
    STATE(6797), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216171] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2480), 1,
      anon_sym_RBRACK,
    ACTIONS(8464), 1,
      anon_sym_SEMI,
    STATE(5634), 1,
      sym_attribute,
    STATE(5851), 1,
      aux_sym__sequence_binding_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216191] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7546), 1,
      anon_sym_RBRACE,
    ACTIONS(8466), 1,
      anon_sym_SEMI,
    STATE(5635), 1,
      sym_attribute,
    STATE(5856), 1,
      aux_sym_record_binding_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216211] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(8468), 1,
      anon_sym_RBRACK,
    STATE(5480), 1,
      aux_sym__type_params_repeat1,
    STATE(5636), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216231] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(8468), 1,
      anon_sym_RBRACK,
    STATE(5637), 1,
      sym_attribute,
    STATE(5643), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216251] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7557), 1,
      anon_sym_RBRACE,
    ACTIONS(8470), 1,
      anon_sym_SEMI,
    STATE(5638), 1,
      sym_attribute,
    STATE(5920), 1,
      aux_sym_record_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216271] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8472), 1,
      anon_sym_SEMI,
    ACTIONS(8474), 1,
      anon_sym_RBRACE,
    STATE(5639), 1,
      sym_attribute,
    STATE(5680), 1,
      aux_sym_record_binding_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216291] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7635), 1,
      anon_sym_PIPE_RBRACK,
    ACTIONS(8476), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5640), 2,
      sym_attribute,
      aux_sym__sequence_binding_pattern_content_repeat1,
  [216309] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8479), 1,
      anon_sym_SEMI,
    ACTIONS(8481), 1,
      anon_sym_RBRACE,
    STATE(5641), 1,
      sym_attribute,
    STATE(5654), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216329] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5998), 1,
      anon_sym_in,
    ACTIONS(8483), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5642), 2,
      sym_attribute,
      aux_sym_type_definition_repeat1,
  [216347] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(8486), 1,
      anon_sym_RBRACK,
    STATE(5480), 1,
      aux_sym__type_params_repeat1,
    STATE(5643), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216367] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(3923), 1,
      sym_module_binding,
    STATE(5644), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7458), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [216385] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7875), 1,
      anon_sym_COMMA,
    ACTIONS(8488), 1,
      anon_sym_RBRACK,
    STATE(5445), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(5645), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216405] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5646), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4931), 3,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [216421] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2498), 1,
      anon_sym_RBRACK,
    ACTIONS(8490), 1,
      anon_sym_SEMI,
    STATE(5647), 1,
      sym_attribute,
    STATE(5938), 1,
      aux_sym__sequence_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216441] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(8334), 1,
      sym_indexing_operator,
    ACTIONS(8492), 1,
      sym__lowercase_identifier,
    STATE(5648), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216461] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    ACTIONS(8494), 1,
      anon_sym_RPAREN,
    STATE(5649), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216481] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7875), 1,
      anon_sym_COMMA,
    ACTIONS(8496), 1,
      anon_sym_RBRACK,
    STATE(5445), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(5650), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216501] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(8498), 1,
      anon_sym_RPAREN,
    STATE(5651), 1,
      sym_attribute,
    STATE(6852), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216521] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8500), 1,
      anon_sym_SEMI,
    ACTIONS(8502), 1,
      anon_sym_RBRACE,
    STATE(5652), 1,
      sym_attribute,
    STATE(5686), 1,
      aux_sym_record_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216541] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6120), 1,
      anon_sym_in,
    ACTIONS(8340), 1,
      anon_sym_and,
    STATE(5653), 1,
      sym_attribute,
    STATE(5728), 1,
      aux_sym_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216561] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7611), 1,
      anon_sym_RBRACE,
    ACTIONS(8504), 1,
      anon_sym_SEMI,
    STATE(5654), 1,
      sym_attribute,
    STATE(5684), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216581] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6085), 1,
      sym__uppercase_identifier,
    ACTIONS(7915), 1,
      anon_sym_LPAREN,
    STATE(1753), 1,
      sym__constructor_name,
    STATE(5655), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216601] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8506), 1,
      anon_sym_SEMI,
    ACTIONS(8508), 1,
      anon_sym_RBRACE,
    STATE(5656), 1,
      sym_attribute,
    STATE(5692), 1,
      aux_sym_record_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216621] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    ACTIONS(8510), 1,
      anon_sym_RPAREN,
    STATE(5657), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216641] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7714), 1,
      anon_sym_RBRACE,
    ACTIONS(8512), 1,
      anon_sym_SEMI,
    STATE(5658), 1,
      sym_attribute,
    STATE(5684), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216661] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6032), 1,
      anon_sym_in,
    ACTIONS(8360), 1,
      anon_sym_and,
    STATE(5659), 1,
      sym_attribute,
    STATE(5662), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216681] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7625), 1,
      anon_sym_RBRACE,
    ACTIONS(8514), 1,
      anon_sym_SEMI,
    STATE(5660), 1,
      sym_attribute,
    STATE(5684), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216701] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7575), 1,
      anon_sym_RBRACE,
    ACTIONS(8516), 1,
      anon_sym_SEMI,
    STATE(5661), 1,
      sym_attribute,
    STATE(5920), 1,
      aux_sym_record_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216721] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6025), 1,
      anon_sym_in,
    ACTIONS(8518), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5662), 2,
      sym_attribute,
      aux_sym_module_definition_repeat1,
  [216739] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8521), 1,
      anon_sym_EQ,
    STATE(5663), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8523), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [216757] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8240), 1,
      anon_sym_RBRACK,
    STATE(5341), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5664), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216777] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5320), 1,
      aux_sym_module_binding_repeat1,
    STATE(5665), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216797] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8525), 1,
      anon_sym_RPAREN,
    STATE(5666), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8527), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [216815] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6032), 1,
      anon_sym_in,
    ACTIONS(8360), 1,
      anon_sym_and,
    STATE(5667), 1,
      sym_attribute,
    STATE(5736), 1,
      aux_sym_module_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216835] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8133), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(8529), 1,
      anon_sym_SEMI,
    STATE(5668), 1,
      sym_attribute,
    STATE(5694), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216855] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4481), 1,
      anon_sym_RBRACK,
    ACTIONS(8531), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5669), 2,
      sym_attribute,
      aux_sym__sequence_expression_content_repeat1,
  [216873] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    ACTIONS(8534), 1,
      anon_sym_RPAREN,
    STATE(5670), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216893] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6185), 1,
      anon_sym_in,
    ACTIONS(8536), 1,
      anon_sym_and,
    STATE(5671), 1,
      sym_attribute,
    STATE(5739), 1,
      aux_sym_class_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216913] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6185), 1,
      anon_sym_in,
    ACTIONS(8536), 1,
      anon_sym_and,
    STATE(5672), 1,
      sym_attribute,
    STATE(5740), 1,
      aux_sym_class_type_definition_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216933] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8222), 1,
      anon_sym_RBRACK,
    STATE(5341), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5673), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216953] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(8538), 1,
      anon_sym_RPAREN,
    STATE(5674), 1,
      sym_attribute,
    STATE(6457), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216973] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8540), 1,
      anon_sym_SEMI,
    ACTIONS(8542), 1,
      anon_sym_RBRACE,
    STATE(5675), 1,
      sym_attribute,
    STATE(5696), 1,
      aux_sym_record_binding_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [216993] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6056), 1,
      anon_sym_in,
    ACTIONS(8544), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5676), 2,
      sym_attribute,
      aux_sym_class_definition_repeat1,
  [217011] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6319), 1,
      anon_sym_in,
    ACTIONS(8381), 1,
      anon_sym_and,
    STATE(5676), 1,
      aux_sym_class_definition_repeat1,
    STATE(5677), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217031] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8547), 1,
      anon_sym_EQ,
    STATE(5678), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8549), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [217049] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7765), 1,
      anon_sym_RBRACE,
    ACTIONS(8551), 1,
      anon_sym_SEMI,
    STATE(5679), 1,
      sym_attribute,
    STATE(5684), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217069] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7410), 1,
      anon_sym_RBRACE,
    ACTIONS(8553), 1,
      anon_sym_SEMI,
    STATE(5680), 1,
      sym_attribute,
    STATE(5856), 1,
      aux_sym_record_binding_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217089] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8555), 1,
      anon_sym_SEMI,
    ACTIONS(8557), 1,
      anon_sym_RBRACE,
    STATE(5679), 1,
      aux_sym_record_expression_repeat1,
    STATE(5681), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217109] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6731), 1,
      sym__uppercase_identifier,
    ACTIONS(7837), 1,
      anon_sym_LPAREN,
    STATE(3119), 1,
      sym__constructor_name,
    STATE(5682), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217129] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4051), 1,
      sym_module_binding,
    STATE(5683), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7458), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [217147] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8559), 1,
      anon_sym_SEMI,
    ACTIONS(8562), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5684), 2,
      sym_attribute,
      aux_sym_record_expression_repeat1,
  [217165] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2498), 1,
      anon_sym_PIPE_RBRACK,
    ACTIONS(8564), 1,
      anon_sym_SEMI,
    STATE(5685), 1,
      sym_attribute,
    STATE(5741), 1,
      aux_sym__sequence_pattern_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217185] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7426), 1,
      anon_sym_RBRACE,
    ACTIONS(8566), 1,
      anon_sym_SEMI,
    STATE(5686), 1,
      sym_attribute,
    STATE(5920), 1,
      aux_sym_record_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217205] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(3901), 1,
      sym_module_binding,
    STATE(5687), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7458), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [217223] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(5412), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(5688), 1,
      sym_attribute,
    STATE(6146), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217243] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1139), 1,
      anon_sym_end,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    STATE(5689), 1,
      sym_attribute,
    STATE(7227), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217263] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4449), 1,
      anon_sym_GT,
    ACTIONS(8568), 1,
      anon_sym_SEMI,
    STATE(5581), 1,
      aux_sym_object_type_repeat1,
    STATE(5690), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217283] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5691), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4885), 3,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [217299] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8153), 1,
      anon_sym_RBRACE,
    ACTIONS(8570), 1,
      anon_sym_SEMI,
    STATE(5605), 1,
      aux_sym_record_declaration_repeat1,
    STATE(5692), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217319] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8572), 1,
      anon_sym_SEMI,
    ACTIONS(8574), 1,
      anon_sym_RBRACE,
    STATE(5693), 1,
      sym_attribute,
    STATE(5703), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217339] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8576), 1,
      anon_sym_SEMI,
    ACTIONS(8579), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5694), 2,
      sym_attribute,
      aux_sym_object_copy_expression_repeat1,
  [217357] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(2958), 1,
      sym_attribute_id,
    STATE(5695), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8581), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [217375] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7366), 1,
      anon_sym_RBRACE,
    ACTIONS(8583), 1,
      anon_sym_SEMI,
    STATE(5696), 1,
      sym_attribute,
    STATE(5856), 1,
      aux_sym_record_binding_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217395] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8585), 1,
      anon_sym_SEMI,
    ACTIONS(8587), 1,
      anon_sym_GT_RBRACE,
    STATE(5697), 1,
      sym_attribute,
    STATE(5705), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217415] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8589), 1,
      anon_sym_SEMI,
    ACTIONS(8591), 1,
      anon_sym_RBRACE,
    STATE(5698), 1,
      sym_attribute,
    STATE(5757), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217435] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8593), 1,
      anon_sym_SEMI,
    ACTIONS(8595), 1,
      anon_sym_RBRACE,
    STATE(5699), 1,
      sym_attribute,
    STATE(5709), 1,
      aux_sym_record_binding_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217455] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8597), 1,
      anon_sym_EQ,
    STATE(5700), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8599), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [217473] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8601), 1,
      anon_sym_SEMI,
    ACTIONS(8603), 1,
      anon_sym_RBRACE,
    STATE(5701), 1,
      sym_attribute,
    STATE(5712), 1,
      aux_sym_record_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217493] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8605), 1,
      aux_sym_quoted_extension_token1,
    STATE(5702), 1,
      sym_attribute,
    STATE(6812), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217513] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7639), 1,
      anon_sym_RBRACE,
    ACTIONS(8607), 1,
      anon_sym_SEMI,
    STATE(5684), 1,
      aux_sym_record_expression_repeat1,
    STATE(5703), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217533] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    ACTIONS(8609), 1,
      anon_sym_RPAREN,
    STATE(5704), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217553] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7985), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(8611), 1,
      anon_sym_SEMI,
    STATE(5694), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(5705), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217573] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8613), 1,
      anon_sym_RPAREN,
    ACTIONS(8615), 1,
      anon_sym_COLON2,
    STATE(5706), 1,
      sym_attribute,
    STATE(6638), 1,
      sym__class_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217593] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(8617), 1,
      anon_sym_RPAREN,
    STATE(5707), 1,
      sym_attribute,
    STATE(6381), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217613] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(14), 1,
      sym_attribute_id,
    STATE(5708), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [217631] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7402), 1,
      anon_sym_RBRACE,
    ACTIONS(8619), 1,
      anon_sym_SEMI,
    STATE(5709), 1,
      sym_attribute,
    STATE(5856), 1,
      aux_sym_record_binding_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217651] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(8334), 1,
      sym_indexing_operator,
    ACTIONS(8621), 1,
      sym__lowercase_identifier,
    STATE(5710), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217671] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8623), 1,
      aux_sym_quoted_extension_token1,
    STATE(5711), 1,
      sym_attribute,
    STATE(6876), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217691] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7406), 1,
      anon_sym_RBRACE,
    ACTIONS(8625), 1,
      anon_sym_SEMI,
    STATE(5712), 1,
      sym_attribute,
    STATE(5920), 1,
      aux_sym_record_pattern_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217711] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8627), 1,
      anon_sym_SEMI,
    ACTIONS(8629), 1,
      anon_sym_RBRACE,
    STATE(5713), 1,
      sym_attribute,
    STATE(5721), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217731] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(5281), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(5714), 1,
      sym_attribute,
    STATE(6146), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217751] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5352), 1,
      aux_sym_module_binding_repeat1,
    STATE(5715), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217771] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(8334), 1,
      sym_indexing_operator,
    ACTIONS(8631), 1,
      sym__lowercase_identifier,
    STATE(5716), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217791] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(15), 1,
      sym_attribute_id,
    STATE(5717), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [217809] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(5428), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(5718), 1,
      sym_attribute,
    STATE(6146), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217829] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(8633), 1,
      anon_sym_RPAREN,
    STATE(5719), 1,
      sym_attribute,
    STATE(6427), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217849] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(8635), 1,
      anon_sym_EQ,
    STATE(5720), 1,
      sym_attribute,
    STATE(6409), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217869] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7657), 1,
      anon_sym_RBRACE,
    ACTIONS(8637), 1,
      anon_sym_SEMI,
    STATE(5684), 1,
      aux_sym_record_expression_repeat1,
    STATE(5721), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217889] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(8639), 1,
      anon_sym_RBRACK,
    STATE(5722), 1,
      sym_attribute,
    STATE(5726), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217909] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(8334), 1,
      sym_indexing_operator,
    ACTIONS(8641), 1,
      sym__lowercase_identifier,
    STATE(5723), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217929] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(5332), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(5724), 1,
      sym_attribute,
    STATE(6146), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217949] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(8643), 1,
      anon_sym_EQ,
    STATE(5725), 1,
      sym_attribute,
    STATE(6431), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217969] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(8645), 1,
      anon_sym_RBRACK,
    STATE(5480), 1,
      aux_sym__type_params_repeat1,
    STATE(5726), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [217989] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(8645), 1,
      anon_sym_RBRACK,
    STATE(5727), 1,
      sym_attribute,
    STATE(5729), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218009] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6170), 1,
      anon_sym_in,
    ACTIONS(8340), 1,
      anon_sym_and,
    STATE(5642), 1,
      aux_sym_type_definition_repeat1,
    STATE(5728), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218029] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(8647), 1,
      anon_sym_RBRACK,
    STATE(5480), 1,
      aux_sym__type_params_repeat1,
    STATE(5729), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218049] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7875), 1,
      anon_sym_COMMA,
    ACTIONS(8649), 1,
      anon_sym_RBRACK,
    STATE(5445), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(5730), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218069] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4249), 1,
      anon_sym_DOT,
    STATE(5731), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8651), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [218087] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7285), 1,
      anon_sym_SEMI,
    ACTIONS(7287), 1,
      anon_sym_GT,
    STATE(5690), 1,
      aux_sym_object_type_repeat1,
    STATE(5732), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218107] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    ACTIONS(8653), 1,
      anon_sym_RPAREN,
    STATE(5733), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218127] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(8655), 1,
      anon_sym_EQ,
    STATE(5734), 1,
      sym_attribute,
    STATE(6991), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218147] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(8657), 1,
      anon_sym_RPAREN,
    STATE(5735), 1,
      sym_attribute,
    STATE(6715), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218167] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6223), 1,
      anon_sym_in,
    ACTIONS(8360), 1,
      anon_sym_and,
    STATE(5662), 1,
      aux_sym_module_definition_repeat1,
    STATE(5736), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218187] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5924), 1,
      anon_sym_in,
    ACTIONS(8340), 1,
      anon_sym_and,
    STATE(5617), 1,
      aux_sym_type_definition_repeat1,
    STATE(5737), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218207] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    ACTIONS(8659), 1,
      anon_sym_RPAREN,
    STATE(5738), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218227] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6176), 1,
      anon_sym_in,
    ACTIONS(8661), 1,
      anon_sym_and,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5739), 2,
      sym_attribute,
      aux_sym_class_type_definition_repeat1,
  [218245] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6279), 1,
      anon_sym_in,
    ACTIONS(8536), 1,
      anon_sym_and,
    STATE(5739), 1,
      aux_sym_class_type_definition_repeat1,
    STATE(5740), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218265] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7663), 1,
      anon_sym_PIPE_RBRACK,
    ACTIONS(8664), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5741), 2,
      sym_attribute,
      aux_sym__sequence_pattern_content_repeat1,
  [218283] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(16), 1,
      sym_attribute_id,
    STATE(5742), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [218301] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8667), 1,
      aux_sym_quoted_extension_token1,
    STATE(5743), 1,
      sym_attribute,
    STATE(7097), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218321] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5924), 1,
      anon_sym_in,
    ACTIONS(8340), 1,
      anon_sym_and,
    STATE(5642), 1,
      aux_sym_type_definition_repeat1,
    STATE(5744), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218341] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(8669), 1,
      anon_sym_RBRACK,
    STATE(5622), 1,
      aux_sym__type_params_repeat1,
    STATE(5745), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218361] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5746), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4596), 3,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [218377] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(17), 1,
      sym_attribute_id,
    STATE(5747), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [218395] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1087), 1,
      anon_sym_end,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    STATE(5748), 1,
      sym_attribute,
    STATE(6583), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218415] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(8671), 1,
      anon_sym_RBRACK,
    STATE(5480), 1,
      aux_sym__type_params_repeat1,
    STATE(5749), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218435] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8673), 1,
      anon_sym_SEMI,
    ACTIONS(8675), 1,
      anon_sym_RBRACE,
    STATE(5750), 1,
      sym_attribute,
    STATE(5788), 1,
      aux_sym_record_declaration_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218455] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8677), 1,
      anon_sym_SEMI,
    ACTIONS(8679), 1,
      anon_sym_RBRACE,
    STATE(5751), 1,
      sym_attribute,
    STATE(5763), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218475] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(8681), 1,
      anon_sym_RBRACK,
    STATE(5480), 1,
      aux_sym__type_params_repeat1,
    STATE(5752), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218495] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(18), 1,
      sym_attribute_id,
    STATE(5753), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [218513] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8683), 1,
      aux_sym_quoted_extension_token1,
    STATE(5754), 1,
      sym_attribute,
    STATE(6506), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218533] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5755), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8685), 3,
      anon_sym_mutable,
      anon_sym_virtual,
      sym__lowercase_identifier,
  [218549] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(8681), 1,
      anon_sym_RBRACK,
    STATE(5749), 1,
      aux_sym__type_params_repeat1,
    STATE(5756), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218569] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7731), 1,
      anon_sym_RBRACE,
    ACTIONS(8687), 1,
      anon_sym_SEMI,
    STATE(5684), 1,
      aux_sym_record_expression_repeat1,
    STATE(5757), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218589] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8689), 1,
      anon_sym_SEMI,
    ACTIONS(8691), 1,
      anon_sym_GT_RBRACE,
    STATE(5758), 1,
      sym_attribute,
    STATE(5947), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218609] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6649), 1,
      sym__uppercase_identifier,
    ACTIONS(7825), 1,
      anon_sym_LPAREN,
    STATE(4493), 1,
      sym__constructor_name,
    STATE(5759), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218629] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5760), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4893), 3,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [218645] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(19), 1,
      sym_attribute_id,
    STATE(5761), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [218663] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8693), 1,
      aux_sym_quoted_extension_token1,
    STATE(5762), 1,
      sym_attribute,
    STATE(6753), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218683] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7692), 1,
      anon_sym_RBRACE,
    ACTIONS(8695), 1,
      anon_sym_SEMI,
    STATE(5684), 1,
      aux_sym_record_expression_repeat1,
    STATE(5763), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218703] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6095), 1,
      anon_sym_in,
    ACTIONS(8360), 1,
      anon_sym_and,
    STATE(5659), 1,
      aux_sym_module_definition_repeat1,
    STATE(5764), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218723] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8697), 1,
      anon_sym_SEMI,
    ACTIONS(8699), 1,
      anon_sym_RBRACE,
    STATE(5550), 1,
      aux_sym_record_expression_repeat1,
    STATE(5765), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218743] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(8701), 1,
      anon_sym_RPAREN,
    STATE(5766), 1,
      sym_attribute,
    STATE(6664), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218763] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(20), 1,
      sym_attribute_id,
    STATE(5767), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [218781] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8703), 1,
      aux_sym_quoted_extension_token1,
    STATE(5768), 1,
      sym_attribute,
    STATE(7003), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218801] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8705), 1,
      anon_sym_SEMI,
    ACTIONS(8707), 1,
      anon_sym_RBRACE,
    STATE(5769), 1,
      sym_attribute,
    STATE(5776), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218821] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(8709), 1,
      anon_sym_RBRACK,
    STATE(5480), 1,
      aux_sym__type_params_repeat1,
    STATE(5770), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218841] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(8334), 1,
      sym_indexing_operator,
    ACTIONS(8711), 1,
      sym__lowercase_identifier,
    STATE(5771), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218861] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(8713), 1,
      anon_sym_RPAREN,
    STATE(5772), 1,
      sym_attribute,
    STATE(6691), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218881] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7967), 1,
      sym__lowercase_identifier,
    ACTIONS(7971), 1,
      anon_sym_mutable,
    STATE(5773), 1,
      sym_attribute,
    STATE(6106), 1,
      sym_field_declaration,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218901] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(21), 1,
      sym_attribute_id,
    STATE(5774), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [218919] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8715), 1,
      aux_sym_quoted_extension_token1,
    STATE(5775), 1,
      sym_attribute,
    STATE(7157), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218939] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7700), 1,
      anon_sym_RBRACE,
    ACTIONS(8717), 1,
      anon_sym_SEMI,
    STATE(5684), 1,
      aux_sym_record_expression_repeat1,
    STATE(5776), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218959] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(22), 1,
      sym_attribute_id,
    STATE(5777), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [218977] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8719), 1,
      aux_sym_quoted_extension_token1,
    STATE(5778), 1,
      sym_attribute,
    STATE(7338), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [218997] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5779), 1,
      sym_attribute,
    STATE(6048), 1,
      sym_module_binding,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7390), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [219015] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6095), 1,
      anon_sym_in,
    ACTIONS(8360), 1,
      anon_sym_and,
    STATE(5662), 1,
      aux_sym_module_definition_repeat1,
    STATE(5780), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219035] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5667), 1,
      sym_module_binding,
    STATE(5781), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7390), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [219053] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(36), 1,
      sym_attribute_id,
    STATE(5782), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219071] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7875), 1,
      anon_sym_COMMA,
    ACTIONS(8721), 1,
      anon_sym_RBRACK,
    STATE(5445), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(5783), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219091] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6193), 1,
      anon_sym_in,
    ACTIONS(8536), 1,
      anon_sym_and,
    STATE(5671), 1,
      aux_sym_class_type_definition_repeat1,
    STATE(5784), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219111] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8723), 1,
      aux_sym_quoted_extension_token1,
    STATE(5785), 1,
      sym_attribute,
    STATE(6620), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219131] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4249), 1,
      anon_sym_DOT,
    STATE(5786), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8725), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [219149] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1127), 1,
      anon_sym_end,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    STATE(5787), 1,
      sym_attribute,
    STATE(6897), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219169] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8027), 1,
      anon_sym_RBRACE,
    ACTIONS(8727), 1,
      anon_sym_SEMI,
    STATE(5605), 1,
      aux_sym_record_declaration_repeat1,
    STATE(5788), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219189] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8729), 1,
      anon_sym_SEMI,
    ACTIONS(8731), 1,
      anon_sym_RBRACE,
    STATE(5789), 1,
      sym_attribute,
    STATE(5825), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219209] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(24), 1,
      sym_attribute_id,
    STATE(5790), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219227] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8733), 1,
      aux_sym_quoted_extension_token1,
    STATE(5791), 1,
      sym_attribute,
    STATE(6487), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219247] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    ACTIONS(8735), 1,
      anon_sym_RPAREN,
    STATE(5792), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219267] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(25), 1,
      sym_attribute_id,
    STATE(5793), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219285] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8737), 1,
      aux_sym_quoted_extension_token1,
    STATE(5794), 1,
      sym_attribute,
    STATE(6666), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219305] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6211), 1,
      anon_sym_in,
    ACTIONS(8381), 1,
      anon_sym_and,
    STATE(5676), 1,
      aux_sym_class_definition_repeat1,
    STATE(5795), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219325] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(26), 1,
      sym_attribute_id,
    STATE(5796), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219343] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8739), 1,
      aux_sym_quoted_extension_token1,
    STATE(5797), 1,
      sym_attribute,
    STATE(6748), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219363] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8741), 1,
      anon_sym_SEMI,
    ACTIONS(8743), 1,
      anon_sym_GT_RBRACE,
    STATE(5798), 1,
      sym_attribute,
    STATE(5830), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219383] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(27), 1,
      sym_attribute_id,
    STATE(5799), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219401] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8745), 1,
      aux_sym_quoted_extension_token1,
    STATE(5800), 1,
      sym_attribute,
    STATE(6824), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219421] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6), 1,
      sym_attribute_id,
    STATE(5801), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219439] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8747), 1,
      aux_sym_quoted_extension_token1,
    STATE(5802), 1,
      sym_attribute,
    STATE(6881), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219459] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6211), 1,
      anon_sym_in,
    ACTIONS(8381), 1,
      anon_sym_and,
    STATE(5677), 1,
      aux_sym_class_definition_repeat1,
    STATE(5803), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219479] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(29), 1,
      sym_attribute_id,
    STATE(5804), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219497] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8749), 1,
      aux_sym_quoted_extension_token1,
    STATE(5805), 1,
      sym_attribute,
    STATE(6939), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219517] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(30), 1,
      sym_attribute_id,
    STATE(5806), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219535] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8751), 1,
      aux_sym_quoted_extension_token1,
    STATE(5807), 1,
      sym_attribute,
    STATE(7019), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219555] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(31), 1,
      sym_attribute_id,
    STATE(5808), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219573] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8753), 1,
      aux_sym_quoted_extension_token1,
    STATE(5809), 1,
      sym_attribute,
    STATE(7074), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219593] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(32), 1,
      sym_attribute_id,
    STATE(5810), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219611] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8755), 1,
      aux_sym_quoted_extension_token1,
    STATE(5811), 1,
      sym_attribute,
    STATE(7102), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219631] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(33), 1,
      sym_attribute_id,
    STATE(5812), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219649] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8757), 1,
      aux_sym_quoted_extension_token1,
    STATE(5813), 1,
      sym_attribute,
    STATE(7111), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219669] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(34), 1,
      sym_attribute_id,
    STATE(5814), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219687] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8759), 1,
      aux_sym_quoted_extension_token1,
    STATE(5815), 1,
      sym_attribute,
    STATE(7139), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219707] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5593), 1,
      sym_attribute_id,
    STATE(5816), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219725] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    ACTIONS(8761), 1,
      anon_sym_RPAREN,
    STATE(5817), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219745] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5818), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4612), 3,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [219761] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(35), 1,
      sym_attribute_id,
    STATE(5819), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219779] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5606), 1,
      sym_attribute_id,
    STATE(5820), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219797] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5821), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8763), 3,
      anon_sym_private,
      anon_sym_virtual,
      sym__lowercase_identifier,
  [219813] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(13), 1,
      sym_attribute_id,
    STATE(5822), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [219831] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1041), 1,
      anon_sym_end,
    ACTIONS(4960), 1,
      anon_sym_PERCENT,
    STATE(5823), 1,
      sym_attribute,
    STATE(7144), 1,
      sym__attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219851] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5824), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4653), 3,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [219867] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7722), 1,
      anon_sym_RBRACE,
    ACTIONS(8765), 1,
      anon_sym_SEMI,
    STATE(5684), 1,
      aux_sym_record_expression_repeat1,
    STATE(5825), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219887] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7875), 1,
      anon_sym_COMMA,
    ACTIONS(8767), 1,
      anon_sym_RBRACK,
    STATE(5445), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(5826), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219907] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8769), 1,
      sym__lowercase_identifier,
    ACTIONS(8771), 1,
      anon_sym_LPAREN,
    STATE(2253), 1,
      sym__simple_value_pattern,
    STATE(5827), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219927] = 6,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    ACTIONS(8773), 1,
      aux_sym_quoted_extension_token1,
    STATE(5828), 1,
      sym_attribute,
    STATE(6533), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219947] = 4,
    ACTIONS(3), 1,
      aux_sym_attribute_token1,
    STATE(5829), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(1523), 3,
      anon_sym_and,
      anon_sym_in,
      anon_sym_LBRACK_AT_AT,
  [219963] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8191), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(8775), 1,
      anon_sym_SEMI,
    STATE(5694), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(5830), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [219983] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8777), 1,
      anon_sym_LPAREN,
    ACTIONS(8779), 1,
      anon_sym_LBRACK,
    ACTIONS(8781), 1,
      anon_sym_LBRACE,
    STATE(5831), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220003] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1281), 1,
      anon_sym_PIPE_RBRACK,
    ACTIONS(8783), 1,
      anon_sym_SEMI,
    STATE(5832), 1,
      sym_attribute,
    STATE(5837), 1,
      aux_sym__sequence_expression_content_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220023] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(8785), 1,
      anon_sym_RPAREN,
    STATE(5833), 1,
      sym_attribute,
    STATE(6959), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220043] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8787), 1,
      anon_sym_SEMI,
    ACTIONS(8789), 1,
      anon_sym_RBRACE,
    STATE(5834), 1,
      sym_attribute,
    STATE(5840), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220063] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8791), 1,
      anon_sym_LPAREN,
    ACTIONS(8793), 1,
      anon_sym_LBRACK,
    ACTIONS(8795), 1,
      anon_sym_LBRACE,
    STATE(5835), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220083] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8015), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(8797), 1,
      anon_sym_SEMI,
    STATE(5694), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(5836), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220103] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4481), 1,
      anon_sym_PIPE_RBRACK,
    ACTIONS(8799), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5837), 2,
      sym_attribute,
      aux_sym__sequence_expression_content_repeat1,
  [220121] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(8334), 1,
      sym_indexing_operator,
    ACTIONS(8802), 1,
      sym__lowercase_identifier,
    STATE(5838), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220141] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(8804), 1,
      anon_sym_RPAREN,
    STATE(5839), 1,
      sym_attribute,
    STATE(6975), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220161] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7735), 1,
      anon_sym_RBRACE,
    ACTIONS(8806), 1,
      anon_sym_SEMI,
    STATE(5684), 1,
      aux_sym_record_expression_repeat1,
    STATE(5840), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220181] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    ACTIONS(8808), 1,
      anon_sym_RPAREN,
    STATE(5841), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220201] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(8810), 1,
      anon_sym_RPAREN,
    STATE(5842), 1,
      sym_attribute,
    STATE(6832), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220221] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    ACTIONS(8812), 1,
      anon_sym_RPAREN,
    STATE(5843), 1,
      sym_attribute,
    STATE(6983), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220241] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8814), 1,
      anon_sym_SEMI,
    ACTIONS(8816), 1,
      anon_sym_RBRACE,
    STATE(5661), 1,
      aux_sym_record_pattern_repeat1,
    STATE(5844), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220261] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8818), 1,
      anon_sym_SEMI,
    ACTIONS(8820), 1,
      anon_sym_RBRACE,
    STATE(5845), 1,
      sym_attribute,
    STATE(5852), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220281] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7875), 1,
      anon_sym_COMMA,
    ACTIONS(8822), 1,
      anon_sym_RPAREN,
    STATE(5445), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(5846), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220301] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(8824), 1,
      anon_sym_RPAREN,
    STATE(5480), 1,
      aux_sym__type_params_repeat1,
    STATE(5847), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220321] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8826), 1,
      anon_sym_SEMI,
    ACTIONS(8828), 1,
      anon_sym_GT_RBRACE,
    STATE(5848), 1,
      sym_attribute,
    STATE(5853), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220341] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    ACTIONS(8830), 1,
      anon_sym_RPAREN,
    STATE(5849), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220361] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5711), 1,
      sym_attribute_id,
    STATE(5850), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220379] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7635), 1,
      anon_sym_RBRACK,
    ACTIONS(8832), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5851), 2,
      sym_attribute,
      aux_sym__sequence_binding_pattern_content_repeat1,
  [220397] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7753), 1,
      anon_sym_RBRACE,
    ACTIONS(8835), 1,
      anon_sym_SEMI,
    STATE(5684), 1,
      aux_sym_record_expression_repeat1,
    STATE(5852), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220417] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8031), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(8837), 1,
      anon_sym_SEMI,
    STATE(5694), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(5853), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220437] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8839), 1,
      anon_sym_LPAREN,
    ACTIONS(8841), 1,
      anon_sym_LBRACK,
    ACTIONS(8843), 1,
      anon_sym_LBRACE,
    STATE(5854), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220457] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8845), 1,
      anon_sym_SEMI,
    ACTIONS(8847), 1,
      anon_sym_RBRACE,
    STATE(5855), 1,
      sym_attribute,
    STATE(5863), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220477] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8849), 1,
      anon_sym_SEMI,
    ACTIONS(8852), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5856), 2,
      sym_attribute,
      aux_sym_record_binding_pattern_repeat1,
  [220495] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8615), 1,
      anon_sym_COLON2,
    ACTIONS(8854), 1,
      anon_sym_RPAREN,
    STATE(5857), 1,
      sym_attribute,
    STATE(7086), 1,
      sym__class_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220515] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8856), 1,
      anon_sym_LPAREN,
    ACTIONS(8858), 1,
      anon_sym_LBRACK,
    ACTIONS(8860), 1,
      anon_sym_LBRACE,
    STATE(5858), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220535] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5743), 1,
      sym_attribute_id,
    STATE(5859), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220553] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    ACTIONS(8862), 1,
      anon_sym_RPAREN,
    STATE(5860), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220573] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(8334), 1,
      sym_indexing_operator,
    ACTIONS(8864), 1,
      sym__lowercase_identifier,
    STATE(5861), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220593] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8866), 1,
      anon_sym_LPAREN,
    ACTIONS(8868), 1,
      anon_sym_LBRACK,
    ACTIONS(8870), 1,
      anon_sym_LBRACE,
    STATE(5862), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220613] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7757), 1,
      anon_sym_RBRACE,
    ACTIONS(8872), 1,
      anon_sym_SEMI,
    STATE(5684), 1,
      aux_sym_record_expression_repeat1,
    STATE(5863), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220633] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8874), 1,
      anon_sym_LPAREN,
    ACTIONS(8876), 1,
      anon_sym_LBRACK,
    ACTIONS(8878), 1,
      anon_sym_LBRACE,
    STATE(5864), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220653] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5754), 1,
      sym_attribute_id,
    STATE(5865), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220671] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(8880), 1,
      anon_sym_EQ,
    STATE(5866), 1,
      sym_attribute,
    STATE(7261), 1,
      sym__coerced,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220691] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5702), 1,
      sym_attribute_id,
    STATE(5867), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220709] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8882), 1,
      anon_sym_LPAREN,
    ACTIONS(8884), 1,
      anon_sym_LBRACK,
    ACTIONS(8886), 1,
      anon_sym_LBRACE,
    STATE(5868), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220729] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(8888), 1,
      anon_sym_type,
    STATE(5786), 1,
      sym_module_path,
    STATE(5869), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220749] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8890), 1,
      anon_sym_LPAREN,
    ACTIONS(8892), 1,
      anon_sym_LBRACK,
    ACTIONS(8894), 1,
      anon_sym_LBRACE,
    STATE(5870), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220769] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5762), 1,
      sym_attribute_id,
    STATE(5871), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220787] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7875), 1,
      anon_sym_COMMA,
    ACTIONS(8896), 1,
      anon_sym_RBRACK,
    STATE(5445), 1,
      aux_sym_instantiated_class_type_repeat1,
    STATE(5872), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220807] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8167), 1,
      anon_sym_RBRACK,
    STATE(5873), 1,
      sym_attribute,
    STATE(5935), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220827] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8898), 1,
      anon_sym_LPAREN,
    ACTIONS(8900), 1,
      anon_sym_LBRACK,
    ACTIONS(8902), 1,
      anon_sym_LBRACE,
    STATE(5874), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220847] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8904), 1,
      anon_sym_LPAREN,
    ACTIONS(8906), 1,
      anon_sym_LBRACK,
    ACTIONS(8908), 1,
      anon_sym_LBRACE,
    STATE(5875), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220867] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5768), 1,
      sym_attribute_id,
    STATE(5876), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [220885] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7967), 1,
      sym__lowercase_identifier,
    ACTIONS(7971), 1,
      anon_sym_mutable,
    STATE(5656), 1,
      sym_field_declaration,
    STATE(5877), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220905] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(8910), 1,
      anon_sym_RBRACK,
    STATE(5752), 1,
      aux_sym__type_params_repeat1,
    STATE(5878), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220925] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8912), 1,
      anon_sym_LPAREN,
    ACTIONS(8914), 1,
      anon_sym_LBRACK,
    ACTIONS(8916), 1,
      anon_sym_LBRACE,
    STATE(5879), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220945] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7951), 1,
      anon_sym_with,
    ACTIONS(8106), 1,
      anon_sym_DASH_GT,
    ACTIONS(8918), 1,
      anon_sym_RPAREN,
    STATE(5880), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220965] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8920), 1,
      anon_sym_LPAREN,
    ACTIONS(8922), 1,
      anon_sym_LBRACK,
    ACTIONS(8924), 1,
      anon_sym_LBRACE,
    STATE(5881), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [220985] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5775), 1,
      sym_attribute_id,
    STATE(5882), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [221003] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8926), 1,
      anon_sym_LPAREN,
    ACTIONS(8928), 1,
      anon_sym_LBRACK,
    ACTIONS(8930), 1,
      anon_sym_LBRACE,
    STATE(5883), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221023] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8932), 1,
      anon_sym_LPAREN,
    ACTIONS(8934), 1,
      anon_sym_LBRACK,
    ACTIONS(8936), 1,
      anon_sym_LBRACE,
    STATE(5884), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221043] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5778), 1,
      sym_attribute_id,
    STATE(5885), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [221061] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8938), 1,
      anon_sym_SEMI,
    ACTIONS(8940), 1,
      anon_sym_RBRACE,
    STATE(5658), 1,
      aux_sym_record_expression_repeat1,
    STATE(5886), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221081] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3694), 1,
      anon_sym_RBRACK,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    STATE(5549), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5887), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221101] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8942), 1,
      anon_sym_LPAREN,
    ACTIONS(8944), 1,
      anon_sym_LBRACK,
    ACTIONS(8946), 1,
      anon_sym_LBRACE,
    STATE(5888), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221121] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3710), 1,
      anon_sym_RBRACK,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    STATE(5889), 1,
      sym_attribute,
    STATE(5926), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221141] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8948), 1,
      anon_sym_LPAREN,
    ACTIONS(8950), 1,
      anon_sym_LBRACK,
    ACTIONS(8952), 1,
      anon_sym_LBRACE,
    STATE(5890), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221161] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5785), 1,
      sym_attribute_id,
    STATE(5891), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [221179] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8954), 1,
      anon_sym_SEMI,
    ACTIONS(8956), 1,
      anon_sym_RBRACE,
    STATE(5892), 1,
      sym_attribute,
    STATE(5924), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221199] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8958), 1,
      anon_sym_LPAREN,
    ACTIONS(8960), 1,
      anon_sym_LBRACK,
    ACTIONS(8962), 1,
      anon_sym_LBRACE,
    STATE(5893), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221219] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8964), 1,
      anon_sym_LPAREN,
    ACTIONS(8966), 1,
      anon_sym_LBRACK,
    ACTIONS(8968), 1,
      anon_sym_LBRACE,
    STATE(5894), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221239] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7261), 1,
      anon_sym_SEMI,
    ACTIONS(7263), 1,
      anon_sym_GT,
    STATE(5554), 1,
      aux_sym_object_type_repeat1,
    STATE(5895), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221259] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8970), 1,
      anon_sym_LPAREN,
    ACTIONS(8972), 1,
      anon_sym_LBRACK,
    ACTIONS(8974), 1,
      anon_sym_LBRACE,
    STATE(5896), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221279] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5791), 1,
      sym_attribute_id,
    STATE(5897), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [221297] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8976), 1,
      anon_sym_LPAREN,
    ACTIONS(8978), 1,
      anon_sym_LBRACK,
    ACTIONS(8980), 1,
      anon_sym_LBRACE,
    STATE(5898), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221317] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5794), 1,
      sym_attribute_id,
    STATE(5899), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [221335] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8982), 1,
      anon_sym_LPAREN,
    ACTIONS(8984), 1,
      anon_sym_LBRACK,
    ACTIONS(8986), 1,
      anon_sym_LBRACE,
    STATE(5900), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221355] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5797), 1,
      sym_attribute_id,
    STATE(5901), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [221373] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8988), 1,
      anon_sym_LPAREN,
    ACTIONS(8990), 1,
      anon_sym_LBRACK,
    ACTIONS(8992), 1,
      anon_sym_LBRACE,
    STATE(5902), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221393] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5800), 1,
      sym_attribute_id,
    STATE(5903), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [221411] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8994), 1,
      anon_sym_LPAREN,
    ACTIONS(8996), 1,
      anon_sym_LBRACK,
    ACTIONS(8998), 1,
      anon_sym_LBRACE,
    STATE(5904), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221431] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5802), 1,
      sym_attribute_id,
    STATE(5905), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [221449] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9000), 1,
      anon_sym_LPAREN,
    ACTIONS(9002), 1,
      anon_sym_LBRACK,
    ACTIONS(9004), 1,
      anon_sym_LBRACE,
    STATE(5906), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221469] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5805), 1,
      sym_attribute_id,
    STATE(5907), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [221487] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9006), 1,
      anon_sym_LPAREN,
    ACTIONS(9008), 1,
      anon_sym_LBRACK,
    ACTIONS(9010), 1,
      anon_sym_LBRACE,
    STATE(5908), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221507] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5807), 1,
      sym_attribute_id,
    STATE(5909), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [221525] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9012), 1,
      anon_sym_LPAREN,
    ACTIONS(9014), 1,
      anon_sym_LBRACK,
    ACTIONS(9016), 1,
      anon_sym_LBRACE,
    STATE(5910), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221545] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5809), 1,
      sym_attribute_id,
    STATE(5911), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [221563] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5811), 1,
      sym_attribute_id,
    STATE(5912), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [221581] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5813), 1,
      sym_attribute_id,
    STATE(5913), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [221599] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5815), 1,
      sym_attribute_id,
    STATE(5914), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [221617] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9018), 1,
      anon_sym_SEMI,
    ACTIONS(9020), 1,
      anon_sym_GT_RBRACE,
    STATE(5915), 1,
      sym_attribute,
    STATE(5927), 1,
      aux_sym_object_copy_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221637] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5828), 1,
      sym_attribute_id,
    STATE(5916), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8322), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [221655] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4412), 1,
      anon_sym_LPAREN,
    STATE(3449), 1,
      sym_module_parameter,
    STATE(5293), 1,
      aux_sym_module_binding_repeat1,
    STATE(5917), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221675] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(9022), 1,
      anon_sym_RBRACK,
    STATE(5918), 1,
      sym_attribute,
    STATE(5921), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221695] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8771), 1,
      anon_sym_LPAREN,
    ACTIONS(9024), 1,
      sym__lowercase_identifier,
    STATE(2253), 1,
      sym__simple_value_pattern,
    STATE(5919), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221715] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9026), 1,
      anon_sym_SEMI,
    ACTIONS(9029), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5920), 2,
      sym_attribute,
      aux_sym_record_pattern_repeat1,
  [221733] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(9031), 1,
      anon_sym_RBRACK,
    STATE(5480), 1,
      aux_sym__type_params_repeat1,
    STATE(5921), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221753] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(9031), 1,
      anon_sym_RBRACK,
    STATE(5922), 1,
      sym_attribute,
    STATE(5923), 1,
      aux_sym__type_params_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221773] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(9033), 1,
      anon_sym_RBRACK,
    STATE(5480), 1,
      aux_sym__type_params_repeat1,
    STATE(5923), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221793] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7771), 1,
      anon_sym_RBRACE,
    ACTIONS(9035), 1,
      anon_sym_SEMI,
    STATE(5684), 1,
      aux_sym_record_expression_repeat1,
    STATE(5924), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221813] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(2636), 1,
      sym_attribute_id,
    STATE(5925), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [221831] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8167), 1,
      anon_sym_RBRACK,
    STATE(5341), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5926), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221851] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8123), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(9037), 1,
      anon_sym_SEMI,
    STATE(5694), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(5927), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221871] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1491), 1,
      anon_sym_COLON_GT,
    ACTIONS(6707), 1,
      anon_sym_RPAREN,
    STATE(4215), 1,
      sym__coerced,
    STATE(5928), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221891] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9039), 1,
      anon_sym_SEMI,
    ACTIONS(9041), 1,
      anon_sym_RBRACE,
    STATE(5929), 1,
      sym_attribute,
    STATE(5934), 1,
      aux_sym_record_expression_repeat1,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221911] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5930), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7877), 3,
      anon_sym_PIPE,
      anon_sym_RBRACK,
      anon_sym_GT,
  [221927] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4051), 1,
      sym_module_binding,
    STATE(5931), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7595), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [221945] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4169), 1,
      sym_module_binding,
    STATE(5932), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7595), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [221963] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(8334), 1,
      sym_indexing_operator,
    ACTIONS(9043), 1,
      sym__lowercase_identifier,
    STATE(5933), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [221983] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7775), 1,
      anon_sym_RBRACE,
    ACTIONS(9045), 1,
      anon_sym_SEMI,
    STATE(5684), 1,
      aux_sym_record_expression_repeat1,
    STATE(5934), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222003] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8050), 1,
      anon_sym_RBRACK,
    STATE(5341), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5935), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222023] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(7), 1,
      sym_attribute_id,
    STATE(5936), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [222041] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8001), 1,
      anon_sym_PIPE,
    ACTIONS(8050), 1,
      anon_sym_RBRACK,
    STATE(5664), 1,
      aux_sym_polymorphic_variant_type_repeat1,
    STATE(5937), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222061] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7663), 1,
      anon_sym_RBRACK,
    ACTIONS(9047), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    STATE(5938), 2,
      sym_attribute,
      aux_sym__sequence_pattern_content_repeat1,
  [222079] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(4527), 1,
      anon_sym_DOT,
    ACTIONS(9050), 1,
      anon_sym_RPAREN,
    STATE(5939), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222099] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(4172), 1,
      sym_module_binding,
    STATE(5940), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7595), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [222117] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(9052), 1,
      anon_sym_type,
    STATE(5786), 1,
      sym_module_path,
    STATE(5941), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222137] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8438), 1,
      anon_sym_COMMA,
    ACTIONS(9054), 1,
      anon_sym_RPAREN,
    STATE(5847), 1,
      aux_sym__type_params_repeat1,
    STATE(5942), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222157] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(8), 1,
      sym_attribute_id,
    STATE(5943), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8302), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [222175] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2530), 1,
      aux_sym_tag_token1,
    STATE(5384), 1,
      aux_sym_polymorphic_variant_type_repeat2,
    STATE(5944), 1,
      sym_attribute,
    STATE(6146), 1,
      sym_tag,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222195] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3941), 1,
      sym__uppercase_identifier,
    ACTIONS(9056), 1,
      anon_sym_type,
    STATE(5786), 1,
      sym_module_path,
    STATE(5945), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222215] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9060), 1,
      anon_sym_LBRACK,
    STATE(5946), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9058), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
  [222233] = 6,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8086), 1,
      anon_sym_GT_RBRACE,
    ACTIONS(9062), 1,
      anon_sym_SEMI,
    STATE(5694), 1,
      aux_sym_object_copy_expression_repeat1,
    STATE(5947), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222253] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7476), 1,
      anon_sym_RBRACE,
    ACTIONS(9064), 1,
      anon_sym_SEMI,
    STATE(5948), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222270] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(9066), 1,
      anon_sym_DOT,
    STATE(5949), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222287] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7498), 1,
      anon_sym_RBRACE,
    ACTIONS(9068), 1,
      anon_sym_SEMI,
    STATE(5950), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222304] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9070), 1,
      anon_sym_COLON2,
    STATE(3831), 1,
      sym__maybe_polymorphic_typed,
    STATE(5951), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222321] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9072), 1,
      anon_sym_RPAREN,
    ACTIONS(9074), 1,
      anon_sym_LT_DASH,
    STATE(5952), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222338] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9076), 1,
      anon_sym_RPAREN,
    ACTIONS(9078), 1,
      anon_sym_LT_DASH,
    STATE(5953), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222355] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(5954), 1,
      sym_attribute,
    STATE(7207), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222372] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7095), 1,
      sym__uppercase_identifier,
    ACTIONS(9080), 1,
      sym__lowercase_identifier,
    STATE(5955), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222389] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5956), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9082), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [222404] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7095), 1,
      sym__uppercase_identifier,
    ACTIONS(9084), 1,
      sym__lowercase_identifier,
    STATE(5957), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222421] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    STATE(3247), 1,
      sym_extended_module_path,
    STATE(5958), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222438] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(5959), 1,
      sym_attribute,
    STATE(7184), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222455] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7095), 1,
      sym__uppercase_identifier,
    ACTIONS(9086), 1,
      sym__lowercase_identifier,
    STATE(5960), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222472] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9088), 1,
      anon_sym_RPAREN,
    ACTIONS(9090), 1,
      anon_sym_LT_DASH,
    STATE(5961), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222489] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5962), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9092), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [222504] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5963), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9094), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [222519] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9096), 1,
      anon_sym_COLON2,
    STATE(5964), 1,
      sym_attribute,
    STATE(6380), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222536] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5965), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9098), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [222551] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9100), 1,
      anon_sym_RPAREN,
    ACTIONS(9102), 1,
      anon_sym_SEMI,
    STATE(5966), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222568] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9100), 1,
      anon_sym_RBRACK,
    ACTIONS(9104), 1,
      anon_sym_SEMI,
    STATE(5967), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222585] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9100), 1,
      anon_sym_RBRACE,
    ACTIONS(9106), 1,
      anon_sym_SEMI,
    STATE(5968), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222602] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9108), 1,
      anon_sym_then,
    STATE(2722), 1,
      sym_then_clause,
    STATE(5969), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222619] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9110), 1,
      anon_sym_COLON2,
    STATE(3780), 1,
      sym__maybe_polymorphic_typed,
    STATE(5970), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222636] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9112), 1,
      anon_sym_RPAREN,
    ACTIONS(9114), 1,
      anon_sym_LT_DASH,
    STATE(5971), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222653] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(9116), 1,
      anon_sym_EQ,
    STATE(5972), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222670] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9118), 1,
      anon_sym_then,
    STATE(3203), 1,
      sym_then_clause,
    STATE(5973), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222687] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(9120), 1,
      anon_sym_DOT,
    STATE(5974), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222704] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(9122), 1,
      anon_sym_DOT,
    STATE(5975), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222721] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9124), 1,
      anon_sym_RPAREN,
    ACTIONS(9126), 1,
      anon_sym_LT_DASH,
    STATE(5976), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222738] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(9128), 1,
      anon_sym_DOT,
    STATE(5977), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222755] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5978), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9130), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [222770] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5979), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9132), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [222785] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9096), 1,
      anon_sym_COLON2,
    STATE(5980), 1,
      sym_attribute,
    STATE(7293), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222802] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(5981), 1,
      sym_attribute,
    STATE(6581), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222819] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(9134), 1,
      sym__lowercase_identifier,
    STATE(5982), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222836] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(5983), 1,
      sym_attribute,
    STATE(6908), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222853] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9136), 1,
      anon_sym_do,
    STATE(2135), 1,
      sym_do_clause,
    STATE(5984), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222870] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(9138), 1,
      anon_sym_RPAREN,
    STATE(5985), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222887] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9140), 1,
      sym__lowercase_identifier,
    ACTIONS(9142), 1,
      anon_sym_LPAREN,
    STATE(5986), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222904] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9144), 1,
      anon_sym_RPAREN,
    ACTIONS(9146), 1,
      anon_sym_COLON_GT,
    STATE(5987), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222921] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9148), 1,
      sym__uppercase_identifier,
    STATE(4582), 1,
      sym_extended_module_path,
    STATE(5988), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222938] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5989), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9150), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [222953] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9152), 1,
      anon_sym_RPAREN,
    ACTIONS(9154), 1,
      anon_sym_SEMI,
    STATE(5990), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222970] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9152), 1,
      anon_sym_RBRACK,
    ACTIONS(9156), 1,
      anon_sym_SEMI,
    STATE(5991), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [222987] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9152), 1,
      anon_sym_RBRACE,
    ACTIONS(9158), 1,
      anon_sym_SEMI,
    STATE(5992), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223004] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9160), 1,
      anon_sym_do,
    STATE(2143), 1,
      sym_do_clause,
    STATE(5993), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223021] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5994), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(7510), 2,
      anon_sym_SEMI,
      anon_sym_GT,
  [223036] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7571), 1,
      anon_sym_RBRACE,
    ACTIONS(9162), 1,
      anon_sym_SEMI,
    STATE(5995), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223053] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(9164), 1,
      anon_sym_DOT,
    STATE(5996), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223070] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(5997), 1,
      sym_attribute,
    STATE(6896), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223087] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(5998), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4865), 2,
      anon_sym_LPAREN,
      anon_sym_DASH_GT,
  [223102] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(9166), 1,
      anon_sym_DOT,
    STATE(5999), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223119] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6000), 1,
      sym_attribute,
    STATE(6894), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223136] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9168), 1,
      anon_sym_RPAREN,
    ACTIONS(9170), 1,
      anon_sym_SEMI,
    STATE(6001), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223153] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9168), 1,
      anon_sym_RBRACK,
    ACTIONS(9172), 1,
      anon_sym_SEMI,
    STATE(6002), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223170] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7579), 1,
      anon_sym_RBRACE,
    ACTIONS(9174), 1,
      anon_sym_SEMI,
    STATE(6003), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223187] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    STATE(6004), 1,
      sym_attribute,
    STATE(6357), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223204] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6005), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9176), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [223219] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9178), 1,
      anon_sym_RPAREN,
    ACTIONS(9180), 1,
      anon_sym_SEMI,
    STATE(6006), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223236] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9178), 1,
      anon_sym_RBRACK,
    ACTIONS(9182), 1,
      anon_sym_SEMI,
    STATE(6007), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223253] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9178), 1,
      anon_sym_RBRACE,
    ACTIONS(9184), 1,
      anon_sym_SEMI,
    STATE(6008), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223270] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(9186), 1,
      sym__lowercase_identifier,
    STATE(6009), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223287] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9168), 1,
      anon_sym_RBRACE,
    ACTIONS(9188), 1,
      anon_sym_SEMI,
    STATE(6010), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223304] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9190), 1,
      anon_sym_RPAREN,
    ACTIONS(9192), 1,
      anon_sym_LT_DASH,
    STATE(6011), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223321] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9194), 1,
      anon_sym_RPAREN,
    ACTIONS(9196), 1,
      anon_sym_LT_DASH,
    STATE(6012), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223338] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7565), 1,
      anon_sym_COLON2,
    STATE(6013), 1,
      sym_attribute,
    STATE(6171), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223355] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(9198), 1,
      anon_sym_DOT,
    STATE(6014), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223372] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(9200), 1,
      anon_sym_DOT,
    STATE(6015), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223389] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9202), 1,
      anon_sym_then,
    STATE(3184), 1,
      sym_then_clause,
    STATE(6016), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223406] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6017), 1,
      sym_attribute,
    STATE(7041), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223423] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9204), 1,
      anon_sym_EQ,
    ACTIONS(9206), 1,
      anon_sym_RPAREN,
    STATE(6018), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223440] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9208), 1,
      anon_sym_RPAREN,
    ACTIONS(9210), 1,
      anon_sym_LT_DASH,
    STATE(6019), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223457] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9212), 1,
      anon_sym_RPAREN,
    ACTIONS(9214), 1,
      anon_sym_SEMI,
    STATE(6020), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223474] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9212), 1,
      anon_sym_RBRACK,
    ACTIONS(9216), 1,
      anon_sym_SEMI,
    STATE(6021), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223491] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9212), 1,
      anon_sym_RBRACE,
    ACTIONS(9218), 1,
      anon_sym_SEMI,
    STATE(6022), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223508] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9220), 1,
      sym__lowercase_identifier,
    ACTIONS(9222), 1,
      anon_sym_LPAREN,
    STATE(6023), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223525] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6024), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9224), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [223540] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9226), 1,
      anon_sym_RPAREN,
    ACTIONS(9228), 1,
      anon_sym_LT_DASH,
    STATE(6025), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223557] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(9230), 1,
      anon_sym_DOT,
    STATE(6026), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223574] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    STATE(5559), 1,
      sym_extended_module_path,
    STATE(6027), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223591] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9232), 1,
      anon_sym_SEMI,
    ACTIONS(9234), 1,
      anon_sym_RBRACE,
    STATE(6028), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223608] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(9236), 1,
      anon_sym_DOT,
    STATE(6029), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223625] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9238), 1,
      sym__lowercase_identifier,
    ACTIONS(9240), 1,
      anon_sym_LPAREN,
    STATE(6030), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223642] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6031), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9242), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [223657] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6032), 1,
      sym_attribute,
    STATE(7250), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223674] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9244), 1,
      anon_sym_RPAREN,
    ACTIONS(9246), 1,
      anon_sym_SEMI,
    STATE(6033), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223691] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9244), 1,
      anon_sym_RBRACK,
    ACTIONS(9248), 1,
      anon_sym_SEMI,
    STATE(6034), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223708] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9244), 1,
      anon_sym_RBRACE,
    ACTIONS(9250), 1,
      anon_sym_SEMI,
    STATE(6035), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223725] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6036), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9252), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [223740] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9254), 1,
      anon_sym_RPAREN,
    ACTIONS(9256), 1,
      anon_sym_LT_DASH,
    STATE(6037), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223757] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9258), 1,
      anon_sym_then,
    STATE(2201), 1,
      sym_then_clause,
    STATE(6038), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223774] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9260), 1,
      anon_sym_SEMI,
    ACTIONS(9262), 1,
      anon_sym_RBRACE,
    STATE(6039), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223791] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6040), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9264), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [223806] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6041), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9266), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [223821] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6042), 1,
      sym_attribute,
    STATE(7181), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223838] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6043), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9268), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [223853] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9270), 1,
      anon_sym_SEMI,
    ACTIONS(9272), 1,
      anon_sym_RBRACE,
    STATE(6044), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223870] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6045), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9274), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [223885] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9276), 1,
      anon_sym_do,
    STATE(2930), 1,
      sym_do_clause,
    STATE(6046), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223902] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9278), 1,
      anon_sym_SEMI,
    ACTIONS(9280), 1,
      anon_sym_RBRACE,
    STATE(6047), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223919] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6048), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6025), 2,
      anon_sym_and,
      anon_sym_in,
  [223934] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9282), 1,
      anon_sym_RPAREN,
    ACTIONS(9284), 1,
      anon_sym_SEMI,
    STATE(6049), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223951] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9282), 1,
      anon_sym_RBRACK,
    ACTIONS(9286), 1,
      anon_sym_SEMI,
    STATE(6050), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223968] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9282), 1,
      anon_sym_RBRACE,
    ACTIONS(9288), 1,
      anon_sym_SEMI,
    STATE(6051), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [223985] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9290), 1,
      anon_sym_SEMI,
    ACTIONS(9292), 1,
      anon_sym_RBRACE,
    STATE(6052), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224002] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9294), 1,
      anon_sym_then,
    STATE(3282), 1,
      sym_then_clause,
    STATE(6053), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224019] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(9296), 1,
      anon_sym_RPAREN,
    STATE(6054), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224036] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9298), 1,
      anon_sym_COLON2,
    STATE(4474), 1,
      sym__typed,
    STATE(6055), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224053] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9300), 1,
      anon_sym_RPAREN,
    ACTIONS(9302), 1,
      anon_sym_LT_DASH,
    STATE(6056), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224070] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9304), 1,
      sym__lowercase_identifier,
    ACTIONS(9306), 1,
      anon_sym_LBRACK,
    STATE(6057), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224087] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7095), 1,
      sym__uppercase_identifier,
    ACTIONS(9308), 1,
      sym__lowercase_identifier,
    STATE(6058), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224104] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6059), 1,
      sym_attribute,
    STATE(7012), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224121] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6060), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9310), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [224136] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9276), 1,
      anon_sym_do,
    STATE(2936), 1,
      sym_do_clause,
    STATE(6061), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224153] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6062), 1,
      sym_attribute,
    STATE(7191), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224170] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6063), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9312), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [224185] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9314), 1,
      anon_sym_RPAREN,
    ACTIONS(9316), 1,
      anon_sym_SEMI,
    STATE(6064), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224202] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9314), 1,
      anon_sym_RBRACK,
    ACTIONS(9318), 1,
      anon_sym_SEMI,
    STATE(6065), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224219] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9314), 1,
      anon_sym_RBRACE,
    ACTIONS(9320), 1,
      anon_sym_SEMI,
    STATE(6066), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224236] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(9322), 1,
      sym__lowercase_identifier,
    STATE(6067), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224253] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2352), 1,
      anon_sym_COLON2,
    STATE(4441), 1,
      sym__maybe_polymorphic_typed,
    STATE(6068), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224270] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6069), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9324), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [224285] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6070), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9326), 2,
      anon_sym_to,
      anon_sym_downto,
  [224300] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9328), 1,
      anon_sym_RPAREN,
    ACTIONS(9330), 1,
      anon_sym_LT_DASH,
    STATE(6071), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224317] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6072), 1,
      sym_attribute,
    STATE(6558), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224334] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9258), 1,
      anon_sym_then,
    STATE(2237), 1,
      sym_then_clause,
    STATE(6073), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224351] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9332), 1,
      anon_sym_RPAREN,
    ACTIONS(9334), 1,
      anon_sym_SEMI,
    STATE(6074), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224368] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9332), 1,
      anon_sym_RBRACK,
    ACTIONS(9336), 1,
      anon_sym_SEMI,
    STATE(6075), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224385] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9332), 1,
      anon_sym_RBRACE,
    ACTIONS(9338), 1,
      anon_sym_SEMI,
    STATE(6076), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224402] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9340), 1,
      anon_sym_then,
    STATE(2108), 1,
      sym_then_clause,
    STATE(6077), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224419] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9342), 1,
      sym__lowercase_identifier,
    ACTIONS(9344), 1,
      anon_sym_LPAREN,
    STATE(6078), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224436] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9346), 1,
      aux_sym_type_variable_token1,
    STATE(2638), 1,
      sym_type_variable,
    STATE(6079), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224453] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9276), 1,
      anon_sym_do,
    STATE(2905), 1,
      sym_do_clause,
    STATE(6080), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224470] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6081), 1,
      sym_attribute,
    STATE(6703), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224487] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9348), 1,
      anon_sym_RPAREN,
    ACTIONS(9350), 1,
      anon_sym_SEMI,
    STATE(6082), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224504] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9348), 1,
      anon_sym_RBRACK,
    ACTIONS(9352), 1,
      anon_sym_SEMI,
    STATE(6083), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224521] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9348), 1,
      anon_sym_RBRACE,
    ACTIONS(9354), 1,
      anon_sym_SEMI,
    STATE(6084), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224538] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9136), 1,
      anon_sym_do,
    STATE(2138), 1,
      sym_do_clause,
    STATE(6085), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224555] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(9356), 1,
      sym__lowercase_identifier,
    STATE(6086), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224572] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(9358), 1,
      anon_sym_DOT,
    STATE(6087), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224589] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6088), 1,
      sym_attribute,
    STATE(6795), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224606] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9360), 1,
      anon_sym_RPAREN,
    ACTIONS(9362), 1,
      anon_sym_SEMI,
    STATE(6089), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224623] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9360), 1,
      anon_sym_RBRACK,
    ACTIONS(9364), 1,
      anon_sym_SEMI,
    STATE(6090), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224640] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9360), 1,
      anon_sym_RBRACE,
    ACTIONS(9366), 1,
      anon_sym_SEMI,
    STATE(6091), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224657] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9368), 1,
      anon_sym_open,
    STATE(6092), 1,
      sym_attribute,
    STATE(6788), 1,
      sym_open_module,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224674] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9370), 1,
      anon_sym_SEMI,
    ACTIONS(9372), 1,
      anon_sym_RBRACE,
    STATE(6093), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224691] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6094), 1,
      sym_attribute,
    STATE(6854), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224708] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9374), 1,
      anon_sym_RPAREN,
    ACTIONS(9376), 1,
      anon_sym_SEMI,
    STATE(6095), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224725] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9374), 1,
      anon_sym_RBRACK,
    ACTIONS(9378), 1,
      anon_sym_SEMI,
    STATE(6096), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224742] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9374), 1,
      anon_sym_RBRACE,
    ACTIONS(9380), 1,
      anon_sym_SEMI,
    STATE(6097), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224759] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9382), 1,
      anon_sym_then,
    STATE(2305), 1,
      sym_then_clause,
    STATE(6098), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224776] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3710), 1,
      anon_sym_RBRACK,
    ACTIONS(7877), 1,
      anon_sym_PIPE,
    STATE(6099), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224793] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1195), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3602), 1,
      anon_sym_RPAREN,
    STATE(6100), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224810] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6101), 1,
      sym_attribute,
    STATE(6904), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224827] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9384), 1,
      anon_sym_RPAREN,
    ACTIONS(9386), 1,
      anon_sym_SEMI,
    STATE(6102), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224844] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9384), 1,
      anon_sym_RBRACK,
    ACTIONS(9388), 1,
      anon_sym_SEMI,
    STATE(6103), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224861] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9384), 1,
      anon_sym_RBRACE,
    ACTIONS(9390), 1,
      anon_sym_SEMI,
    STATE(6104), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224878] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6105), 1,
      sym_attribute,
    STATE(6632), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224895] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6106), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8416), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [224910] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6107), 1,
      sym_attribute,
    STATE(6970), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224927] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9392), 1,
      anon_sym_RPAREN,
    ACTIONS(9394), 1,
      anon_sym_SEMI,
    STATE(6108), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224944] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9392), 1,
      anon_sym_RBRACK,
    ACTIONS(9396), 1,
      anon_sym_SEMI,
    STATE(6109), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224961] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9392), 1,
      anon_sym_RBRACE,
    ACTIONS(9398), 1,
      anon_sym_SEMI,
    STATE(6110), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224978] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9222), 1,
      anon_sym_LPAREN,
    ACTIONS(9400), 1,
      sym__lowercase_identifier,
    STATE(6111), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [224995] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6112), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9402), 2,
      anon_sym_SEMI,
      anon_sym_GT,
  [225010] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6113), 1,
      sym_attribute,
    STATE(7046), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225027] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9404), 1,
      anon_sym_RPAREN,
    ACTIONS(9406), 1,
      anon_sym_SEMI,
    STATE(6114), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225044] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9404), 1,
      anon_sym_RBRACK,
    ACTIONS(9408), 1,
      anon_sym_SEMI,
    STATE(6115), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225061] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9404), 1,
      anon_sym_RBRACE,
    ACTIONS(9410), 1,
      anon_sym_SEMI,
    STATE(6116), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225078] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9412), 1,
      anon_sym_then,
    STATE(2158), 1,
      sym_then_clause,
    STATE(6117), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225095] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9414), 1,
      anon_sym_RPAREN,
    ACTIONS(9416), 1,
      anon_sym_LT_DASH,
    STATE(6118), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225112] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6119), 1,
      sym_attribute,
    STATE(7084), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225129] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9418), 1,
      sym__lowercase_identifier,
    ACTIONS(9420), 1,
      anon_sym_LPAREN,
    STATE(6120), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225146] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6121), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6056), 2,
      anon_sym_and,
      anon_sym_in,
  [225161] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6122), 1,
      sym_attribute,
    STATE(7104), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225178] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9422), 1,
      sym__lowercase_identifier,
    ACTIONS(9424), 1,
      anon_sym_LPAREN,
    STATE(6123), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225195] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9426), 1,
      sym__lowercase_identifier,
    ACTIONS(9428), 1,
      anon_sym_LPAREN,
    STATE(6124), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225212] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6125), 1,
      sym_attribute,
    STATE(7117), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225229] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6126), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9430), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [225244] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(7338), 1,
      anon_sym_RBRACK,
    STATE(6127), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225261] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6128), 1,
      sym_attribute,
    STATE(7143), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225278] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6129), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9432), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [225293] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9434), 1,
      sym__lowercase_identifier,
    ACTIONS(9436), 1,
      anon_sym_LPAREN,
    STATE(6130), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225310] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6131), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9438), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [225325] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9344), 1,
      anon_sym_LPAREN,
    ACTIONS(9440), 1,
      sym__lowercase_identifier,
    STATE(6132), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225342] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9442), 1,
      aux_sym_type_variable_token1,
    STATE(4258), 1,
      sym_type_variable,
    STATE(6133), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225359] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9160), 1,
      anon_sym_do,
    STATE(2129), 1,
      sym_do_clause,
    STATE(6134), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225376] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6135), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9444), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [225391] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9446), 1,
      sym__lowercase_identifier,
    ACTIONS(9448), 1,
      anon_sym_LPAREN,
    STATE(6136), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225408] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6137), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9450), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [225423] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9452), 1,
      sym__lowercase_identifier,
    ACTIONS(9454), 1,
      anon_sym_LPAREN,
    STATE(6138), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225440] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9298), 1,
      anon_sym_COLON2,
    STATE(4448), 1,
      sym__typed,
    STATE(6139), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225457] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9222), 1,
      anon_sym_LPAREN,
    ACTIONS(9456), 1,
      sym__lowercase_identifier,
    STATE(6140), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225474] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9420), 1,
      anon_sym_LPAREN,
    ACTIONS(9458), 1,
      sym__lowercase_identifier,
    STATE(6141), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225491] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8141), 1,
      anon_sym_COLON2,
    STATE(6142), 1,
      sym_attribute,
    STATE(6483), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225508] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6143), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9460), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [225523] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6144), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8562), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [225538] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9070), 1,
      anon_sym_COLON2,
    STATE(3780), 1,
      sym__maybe_polymorphic_typed,
    STATE(6145), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225555] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6146), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9462), 2,
      anon_sym_RBRACK,
      aux_sym_tag_token1,
  [225570] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9096), 1,
      anon_sym_COLON2,
    STATE(6147), 1,
      sym_attribute,
    STATE(6943), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225587] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3219), 1,
      anon_sym_COMMA,
    ACTIONS(9464), 1,
      anon_sym_COLON,
    STATE(6148), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225604] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6038), 1,
      anon_sym_EQ,
    ACTIONS(9466), 1,
      anon_sym_RPAREN,
    STATE(6149), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225621] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(9468), 1,
      sym__lowercase_identifier,
    STATE(6150), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225638] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6151), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9470), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [225653] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9472), 1,
      anon_sym_then,
    STATE(3116), 1,
      sym_then_clause,
    STATE(6152), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225670] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8141), 1,
      anon_sym_COLON2,
    STATE(6153), 1,
      sym_attribute,
    STATE(6599), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225687] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2352), 1,
      anon_sym_COLON2,
    STATE(4487), 1,
      sym__maybe_polymorphic_typed,
    STATE(6154), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225704] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9474), 1,
      sym__lowercase_identifier,
    ACTIONS(9476), 1,
      anon_sym_LBRACK,
    STATE(6155), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225721] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7095), 1,
      sym__uppercase_identifier,
    ACTIONS(9478), 1,
      sym__lowercase_identifier,
    STATE(6156), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225738] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7095), 1,
      sym__uppercase_identifier,
    ACTIONS(9480), 1,
      sym__lowercase_identifier,
    STATE(6157), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225755] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6158), 1,
      sym_attribute,
    STATE(6661), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225772] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7095), 1,
      sym__uppercase_identifier,
    ACTIONS(9482), 1,
      sym__lowercase_identifier,
    STATE(6159), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225789] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9484), 1,
      sym__lowercase_identifier,
    ACTIONS(9486), 1,
      anon_sym_LPAREN,
    STATE(6160), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225806] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9136), 1,
      anon_sym_do,
    STATE(2130), 1,
      sym_do_clause,
    STATE(6161), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225823] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7565), 1,
      anon_sym_COLON2,
    STATE(6060), 1,
      sym__maybe_polymorphic_typed,
    STATE(6162), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225840] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9368), 1,
      anon_sym_open,
    STATE(6163), 1,
      sym_attribute,
    STATE(6735), 1,
      sym_open_module,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225857] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(9488), 1,
      anon_sym_EQ,
    STATE(6164), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225874] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6165), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9490), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [225889] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7095), 1,
      sym__uppercase_identifier,
    ACTIONS(7140), 1,
      sym__lowercase_identifier,
    STATE(6166), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225906] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6167), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9492), 2,
      anon_sym_to,
      anon_sym_downto,
  [225921] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6168), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9494), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [225936] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6169), 1,
      sym_attribute,
    STATE(6434), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225953] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9496), 1,
      anon_sym_RPAREN,
    ACTIONS(9498), 1,
      anon_sym_LT_DASH,
    STATE(6170), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [225970] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6171), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9500), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [225985] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9502), 1,
      anon_sym_RPAREN,
    ACTIONS(9504), 1,
      anon_sym_LT_DASH,
    STATE(6172), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226002] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6809), 1,
      anon_sym_COLON2,
    STATE(6173), 1,
      sym_attribute,
    STATE(6961), 1,
      sym__module_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226019] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9506), 1,
      sym__lowercase_identifier,
    ACTIONS(9508), 1,
      anon_sym_LPAREN,
    STATE(6174), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226036] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8042), 1,
      sym__lowercase_identifier,
    STATE(5507), 1,
      sym__simple_value_pattern,
    STATE(6175), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226053] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9510), 1,
      anon_sym_RPAREN,
    ACTIONS(9512), 1,
      anon_sym_LT_DASH,
    STATE(6176), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226070] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7047), 1,
      anon_sym_COMMA,
    ACTIONS(9514), 1,
      anon_sym_COLON,
    STATE(6177), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226087] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(9516), 1,
      anon_sym_RPAREN,
    STATE(6178), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226104] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6179), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9518), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [226119] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6180), 1,
      sym_attribute,
    STATE(7186), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226136] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9520), 1,
      anon_sym_then,
    STATE(3416), 1,
      sym_then_clause,
    STATE(6181), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226153] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8141), 1,
      anon_sym_COLON2,
    STATE(6182), 1,
      sym_attribute,
    STATE(7327), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226170] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9108), 1,
      anon_sym_then,
    STATE(2739), 1,
      sym_then_clause,
    STATE(6183), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226187] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9136), 1,
      anon_sym_do,
    STATE(2169), 1,
      sym_do_clause,
    STATE(6184), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226204] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7095), 1,
      sym__uppercase_identifier,
    ACTIONS(9522), 1,
      sym__lowercase_identifier,
    STATE(6185), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226221] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9420), 1,
      anon_sym_LPAREN,
    ACTIONS(9524), 1,
      sym__lowercase_identifier,
    STATE(6186), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226238] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9118), 1,
      anon_sym_then,
    STATE(3186), 1,
      sym_then_clause,
    STATE(6187), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226255] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9526), 1,
      anon_sym_EQ,
    ACTIONS(9528), 1,
      anon_sym_RPAREN,
    STATE(6188), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226272] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6189), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9530), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [226287] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9532), 1,
      sym__lowercase_identifier,
    ACTIONS(9534), 1,
      anon_sym_LBRACK,
    STATE(6190), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226304] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9202), 1,
      anon_sym_then,
    STATE(3185), 1,
      sym_then_clause,
    STATE(6191), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226321] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9536), 1,
      anon_sym_then,
    STATE(3293), 1,
      sym_then_clause,
    STATE(6192), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226338] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9538), 1,
      sym__lowercase_identifier,
    ACTIONS(9540), 1,
      anon_sym_LPAREN,
    STATE(6193), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226355] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9520), 1,
      anon_sym_then,
    STATE(3435), 1,
      sym_then_clause,
    STATE(6194), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226372] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6195), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8579), 2,
      anon_sym_SEMI,
      anon_sym_GT_RBRACE,
  [226387] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(9542), 1,
      anon_sym_DOT,
    STATE(6196), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226404] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9544), 1,
      anon_sym_then,
    STATE(3236), 1,
      sym_then_clause,
    STATE(6197), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226421] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(9546), 1,
      anon_sym_RPAREN,
    STATE(6198), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226438] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9548), 1,
      anon_sym_do,
    STATE(2214), 1,
      sym_do_clause,
    STATE(6199), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226455] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7414), 1,
      anon_sym_RBRACE,
    ACTIONS(9550), 1,
      anon_sym_SEMI,
    STATE(6200), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226472] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8141), 1,
      anon_sym_COLON2,
    STATE(6201), 1,
      sym_attribute,
    STATE(6392), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226489] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6202), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(8852), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [226504] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(9552), 1,
      sym__lowercase_identifier,
    STATE(6203), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226521] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8270), 1,
      sym__lowercase_identifier,
    STATE(5504), 1,
      aux_sym__abstract_type_repeat1,
    STATE(6204), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226538] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6205), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(4300), 2,
      anon_sym_DOT,
      aux_sym_type_variable_token1,
  [226553] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(29), 1,
      sym__uppercase_identifier,
    STATE(5939), 1,
      sym_extended_module_path,
    STATE(6206), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226570] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8141), 1,
      anon_sym_COLON2,
    STATE(6207), 1,
      sym_attribute,
    STATE(6416), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226587] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(9554), 1,
      anon_sym_EQ,
    STATE(6208), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226604] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9556), 1,
      sym__lowercase_identifier,
    ACTIONS(9558), 1,
      anon_sym_LBRACK,
    STATE(6209), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226621] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6210), 1,
      sym_attribute,
    STATE(7015), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226638] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(9560), 1,
      sym__lowercase_identifier,
    STATE(6211), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226655] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6212), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9562), 2,
      anon_sym_to,
      anon_sym_downto,
  [226670] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6213), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9564), 2,
      anon_sym_to,
      anon_sym_downto,
  [226685] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(8492), 1,
      sym__lowercase_identifier,
    STATE(6214), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226702] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9566), 1,
      sym__lowercase_identifier,
    ACTIONS(9568), 1,
      anon_sym_LPAREN,
    STATE(6215), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226719] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6216), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9570), 2,
      anon_sym_to,
      anon_sym_downto,
  [226734] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9572), 1,
      anon_sym_RPAREN,
    ACTIONS(9574), 1,
      anon_sym_LT_DASH,
    STATE(6217), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226751] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9294), 1,
      anon_sym_then,
    STATE(3307), 1,
      sym_then_clause,
    STATE(6218), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226768] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9576), 1,
      anon_sym_RPAREN,
    ACTIONS(9578), 1,
      anon_sym_LT_DASH,
    STATE(6219), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226785] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9580), 1,
      sym__lowercase_identifier,
    ACTIONS(9582), 1,
      sym__uppercase_identifier,
    STATE(6220), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226802] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6221), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9584), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [226817] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6222), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9586), 2,
      anon_sym_to,
      anon_sym_downto,
  [226832] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6223), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9588), 2,
      anon_sym_to,
      anon_sym_downto,
  [226847] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6224), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9590), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [226862] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2938), 1,
      anon_sym_COMMA,
    ACTIONS(9592), 1,
      anon_sym_COLON,
    STATE(6225), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226879] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8042), 1,
      sym__lowercase_identifier,
    STATE(5375), 1,
      sym__simple_value_pattern,
    STATE(6226), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226896] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9548), 1,
      anon_sym_do,
    STATE(2294), 1,
      sym_do_clause,
    STATE(6227), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226913] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9412), 1,
      anon_sym_then,
    STATE(2161), 1,
      sym_then_clause,
    STATE(6228), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226930] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(9594), 1,
      sym__lowercase_identifier,
    STATE(6229), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226947] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(9596), 1,
      anon_sym_EQ,
    STATE(6230), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226964] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(9598), 1,
      sym__lowercase_identifier,
    STATE(6231), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226981] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9600), 1,
      anon_sym_RPAREN,
    ACTIONS(9602), 1,
      anon_sym_LT_DASH,
    STATE(6232), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [226998] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6233), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9604), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [227013] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9606), 1,
      anon_sym_RPAREN,
    ACTIONS(9608), 1,
      anon_sym_LT_DASH,
    STATE(6234), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227030] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6235), 1,
      sym_attribute,
    STATE(6385), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227047] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9276), 1,
      anon_sym_do,
    STATE(2888), 1,
      sym_do_clause,
    STATE(6236), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227064] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9472), 1,
      anon_sym_then,
    STATE(3168), 1,
      sym_then_clause,
    STATE(6237), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227081] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6238), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9610), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [227096] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9536), 1,
      anon_sym_then,
    STATE(3311), 1,
      sym_then_clause,
    STATE(6239), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227113] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9110), 1,
      anon_sym_COLON2,
    STATE(3831), 1,
      sym__maybe_polymorphic_typed,
    STATE(6240), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227130] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6241), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9612), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [227145] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3694), 1,
      anon_sym_RBRACK,
    ACTIONS(7877), 1,
      anon_sym_PIPE,
    STATE(6242), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227162] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9614), 1,
      anon_sym_RPAREN,
    ACTIONS(9616), 1,
      anon_sym_LT_DASH,
    STATE(6243), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227179] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(9618), 1,
      anon_sym_DOT,
    STATE(6244), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227196] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9620), 1,
      anon_sym_RPAREN,
    ACTIONS(9622), 1,
      anon_sym_LT_DASH,
    STATE(6245), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227213] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6246), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9624), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [227228] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9626), 1,
      sym__lowercase_identifier,
    ACTIONS(9628), 1,
      sym__uppercase_identifier,
    STATE(6247), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227245] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7095), 1,
      sym__uppercase_identifier,
    ACTIONS(7136), 1,
      sym__lowercase_identifier,
    STATE(6248), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227262] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9382), 1,
      anon_sym_then,
    STATE(2314), 1,
      sym_then_clause,
    STATE(6249), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227279] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(9630), 1,
      anon_sym_DOT,
    STATE(6250), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227296] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(9632), 1,
      sym__lowercase_identifier,
    STATE(6251), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227313] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(9634), 1,
      anon_sym_DOT,
    STATE(6252), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227330] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6253), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9636), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [227345] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(9638), 1,
      anon_sym_DOT,
    STATE(6254), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227362] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7095), 1,
      sym__uppercase_identifier,
    ACTIONS(7122), 1,
      sym__lowercase_identifier,
    STATE(6255), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227379] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9160), 1,
      anon_sym_do,
    STATE(2156), 1,
      sym_do_clause,
    STATE(6256), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227396] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6257), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9640), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [227411] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9642), 1,
      anon_sym_RPAREN,
    ACTIONS(9644), 1,
      anon_sym_LT_DASH,
    STATE(6258), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227428] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6259), 1,
      sym_attribute,
    STATE(6432), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227445] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6260), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9646), 2,
      anon_sym_DASH_GT,
      anon_sym_RBRACK,
  [227460] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9648), 1,
      anon_sym_RPAREN,
    ACTIONS(9650), 1,
      anon_sym_COLON_GT,
    STATE(6261), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227477] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9652), 1,
      anon_sym_RPAREN,
    ACTIONS(9654), 1,
      anon_sym_LT_DASH,
    STATE(6262), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227494] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3923), 1,
      anon_sym_RPAREN,
    ACTIONS(9656), 1,
      aux_sym_number_token1,
    STATE(6263), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227511] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9096), 1,
      anon_sym_COLON2,
    STATE(6264), 1,
      sym_attribute,
    STATE(6451), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227528] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6265), 1,
      sym_attribute,
    STATE(7307), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227545] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6266), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9658), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [227560] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9544), 1,
      anon_sym_then,
    STATE(3267), 1,
      sym_then_clause,
    STATE(6267), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227577] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6268), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9660), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [227592] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7418), 1,
      anon_sym_RBRACE,
    ACTIONS(9662), 1,
      anon_sym_SEMI,
    STATE(6269), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227609] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7422), 1,
      anon_sym_RBRACE,
    ACTIONS(9664), 1,
      anon_sym_SEMI,
    STATE(6270), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227626] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3658), 1,
      anon_sym_RBRACK,
    ACTIONS(7877), 1,
      anon_sym_PIPE,
    STATE(6271), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227643] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9666), 1,
      anon_sym_RPAREN,
    ACTIONS(9668), 1,
      anon_sym_LT_DASH,
    STATE(6272), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227660] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9670), 1,
      anon_sym_RPAREN,
    ACTIONS(9672), 1,
      anon_sym_LT_DASH,
    STATE(6273), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227677] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9674), 1,
      anon_sym_RPAREN,
    ACTIONS(9676), 1,
      anon_sym_LT_DASH,
    STATE(6274), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227694] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9096), 1,
      anon_sym_COLON2,
    STATE(6275), 1,
      sym_attribute,
    STATE(6478), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227711] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9678), 1,
      sym__lowercase_identifier,
    ACTIONS(9680), 1,
      anon_sym_LBRACK,
    STATE(6276), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227728] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6277), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9682), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [227743] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6278), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9684), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [227758] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9096), 1,
      anon_sym_COLON2,
    STATE(6279), 1,
      sym_attribute,
    STATE(6786), 1,
      sym__typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227775] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6280), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9686), 2,
      anon_sym__,
      sym__uppercase_identifier,
  [227790] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9688), 1,
      anon_sym_RPAREN,
    ACTIONS(9690), 1,
      anon_sym_SEMI,
    STATE(6281), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227807] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9692), 1,
      anon_sym_RPAREN,
    ACTIONS(9694), 1,
      anon_sym_SEMI,
    STATE(6282), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227824] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9692), 1,
      anon_sym_RBRACK,
    ACTIONS(9696), 1,
      anon_sym_SEMI,
    STATE(6283), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227841] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9692), 1,
      anon_sym_RBRACE,
    ACTIONS(9698), 1,
      anon_sym_SEMI,
    STATE(6284), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227858] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9688), 1,
      anon_sym_RBRACK,
    ACTIONS(9700), 1,
      anon_sym_SEMI,
    STATE(6285), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227875] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9688), 1,
      anon_sym_RBRACE,
    ACTIONS(9702), 1,
      anon_sym_SEMI,
    STATE(6286), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227892] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9368), 1,
      anon_sym_open,
    STATE(6287), 1,
      sym_attribute,
    STATE(7264), 1,
      sym_open_module,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227909] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9704), 1,
      anon_sym_SEMI,
    ACTIONS(9706), 1,
      anon_sym_RBRACE,
    STATE(6288), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227926] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3676), 1,
      aux_sym_type_variable_token1,
    STATE(4258), 1,
      sym_type_variable,
    STATE(6289), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227943] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7095), 1,
      sym__uppercase_identifier,
    ACTIONS(9708), 1,
      sym__lowercase_identifier,
    STATE(6290), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227960] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9710), 1,
      anon_sym_RPAREN,
    ACTIONS(9712), 1,
      anon_sym_LT_DASH,
    STATE(6291), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227977] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9714), 1,
      anon_sym_RPAREN,
    ACTIONS(9716), 1,
      anon_sym_LT_DASH,
    STATE(6292), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [227994] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9718), 1,
      anon_sym_SEMI,
    ACTIONS(9720), 1,
      anon_sym_RBRACE,
    STATE(6293), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228011] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6294), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9722), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [228026] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6295), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9724), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [228041] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7095), 1,
      sym__uppercase_identifier,
    ACTIONS(9726), 1,
      sym__lowercase_identifier,
    STATE(6296), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228058] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9728), 1,
      sym__lowercase_identifier,
    ACTIONS(9730), 1,
      anon_sym_LPAREN,
    STATE(6297), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228075] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9732), 1,
      anon_sym_RPAREN,
    ACTIONS(9734), 1,
      anon_sym_LT_DASH,
    STATE(6298), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228092] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9548), 1,
      anon_sym_do,
    STATE(2206), 1,
      sym_do_clause,
    STATE(6299), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228109] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    ACTIONS(9736), 1,
      sym__lowercase_identifier,
    STATE(6300), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228126] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(9738), 1,
      anon_sym_DOT,
    STATE(6301), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228143] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9740), 1,
      anon_sym_RPAREN,
    ACTIONS(9742), 1,
      anon_sym_LT_DASH,
    STATE(6302), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228160] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9548), 1,
      anon_sym_do,
    STATE(2211), 1,
      sym_do_clause,
    STATE(6303), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228177] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9744), 1,
      anon_sym_RPAREN,
    ACTIONS(9746), 1,
      anon_sym_LT_DASH,
    STATE(6304), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228194] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9748), 1,
      anon_sym_then,
    STATE(2748), 1,
      sym_then_clause,
    STATE(6305), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228211] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(9750), 1,
      anon_sym_EQ,
    STATE(6306), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228228] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9368), 1,
      anon_sym_open,
    STATE(6307), 1,
      sym_attribute,
    STATE(6596), 1,
      sym_open_module,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228245] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(9752), 1,
      anon_sym_DOT,
    STATE(6308), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228262] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7482), 1,
      anon_sym_RBRACE,
    ACTIONS(9754), 1,
      anon_sym_SEMI,
    STATE(6309), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228279] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9756), 1,
      anon_sym_RPAREN,
    ACTIONS(9758), 1,
      anon_sym_LT_DASH,
    STATE(6310), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228296] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2532), 1,
      sym__left_quoted_string_delimiter,
    STATE(6311), 1,
      sym_attribute,
    STATE(7235), 1,
      sym__quoted_string,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228313] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(9760), 1,
      anon_sym_RPAREN,
    STATE(6312), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228330] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9340), 1,
      anon_sym_then,
    STATE(2109), 1,
      sym_then_clause,
    STATE(6313), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228347] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6314), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9762), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [228362] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6315), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(6176), 2,
      anon_sym_and,
      anon_sym_in,
  [228377] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6316), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9764), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [228392] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(31), 1,
      aux_sym_type_variable_token1,
    STATE(2638), 1,
      sym_type_variable,
    STATE(6317), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228409] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1191), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3606), 1,
      anon_sym_RPAREN,
    STATE(6318), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228426] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9766), 1,
      anon_sym_RPAREN,
    ACTIONS(9768), 1,
      anon_sym_LT_DASH,
    STATE(6319), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228443] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6320), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(5998), 2,
      anon_sym_and,
      anon_sym_in,
  [228458] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6321), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9770), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [228473] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6322), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9772), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [228488] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8141), 1,
      anon_sym_COLON2,
    STATE(6323), 1,
      sym_attribute,
    STATE(6966), 1,
      sym__maybe_polymorphic_typed,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228505] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6324), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9774), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [228520] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(9776), 1,
      anon_sym_DOT,
    STATE(6325), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228537] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6326), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9778), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [228552] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6327), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9780), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [228567] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6328), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9782), 2,
      sym__lowercase_identifier,
      sym__uppercase_identifier,
  [228582] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9748), 1,
      anon_sym_then,
    STATE(2757), 1,
      sym_then_clause,
    STATE(6329), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228599] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9160), 1,
      anon_sym_do,
    STATE(2134), 1,
      sym_do_clause,
    STATE(6330), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228616] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6500), 1,
      anon_sym_as,
    ACTIONS(9784), 1,
      ts_builtin_sym_end,
    STATE(6331), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228633] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6332), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9029), 2,
      anon_sym_SEMI,
      anon_sym_RBRACE,
  [228648] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6333), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9786), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [228663] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9788), 1,
      sym__lowercase_identifier,
    ACTIONS(9790), 1,
      anon_sym_LPAREN,
    STATE(6334), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228680] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6335), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9792), 2,
      anon_sym_EQ,
      anon_sym_COLON_EQ,
  [228695] = 5,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4525), 1,
      anon_sym_LPAREN,
    ACTIONS(9794), 1,
      anon_sym_DOT,
    STATE(6336), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228712] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    STATE(6337), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
    ACTIONS(9796), 2,
      anon_sym_to,
      anon_sym_downto,
  [228727] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9798), 1,
      anon_sym_RPAREN,
    STATE(6338), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228741] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9800), 1,
      anon_sym_RPAREN,
    STATE(6339), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228755] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9802), 1,
      anon_sym_type,
    STATE(6340), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228769] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9804), 1,
      anon_sym_DASH_GT,
    STATE(6341), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228783] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9806), 1,
      anon_sym_PIPE_RBRACK,
    STATE(6342), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228797] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9808), 1,
      anon_sym_RPAREN,
    STATE(6343), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228811] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9810), 1,
      anon_sym_RPAREN,
    STATE(6344), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228825] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9812), 1,
      anon_sym_RPAREN,
    STATE(6345), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228839] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9814), 1,
      anon_sym_DASH_GT,
    STATE(6346), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228853] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9816), 1,
      anon_sym_RPAREN,
    STATE(6347), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228867] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7793), 1,
      anon_sym_end,
    STATE(6348), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228881] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9818), 1,
      anon_sym_DASH_GT,
    STATE(6349), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228895] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9820), 1,
      anon_sym_done,
    STATE(6350), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228909] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9822), 1,
      anon_sym_RPAREN,
    STATE(6351), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228923] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9824), 1,
      anon_sym_RPAREN,
    STATE(6352), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228937] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9826), 1,
      anon_sym_RPAREN,
    STATE(6353), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228951] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9828), 1,
      anon_sym_DOT,
    STATE(6354), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228965] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9830), 1,
      anon_sym_type,
    STATE(6355), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228979] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9832), 1,
      anon_sym_RPAREN,
    STATE(6356), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [228993] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9834), 1,
      anon_sym_RPAREN,
    STATE(6357), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229007] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9836), 1,
      anon_sym_RBRACK,
    STATE(6358), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229021] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9838), 1,
      anon_sym_RBRACE,
    STATE(6359), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229035] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9840), 1,
      anon_sym_DASH_GT,
    STATE(6360), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229049] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9842), 1,
      anon_sym_DASH_GT,
    STATE(6361), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229063] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9844), 1,
      anon_sym_DOT,
    STATE(6362), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229077] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9846), 1,
      anon_sym_done,
    STATE(6363), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229091] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9848), 1,
      anon_sym_RBRACK,
    STATE(6364), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229105] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7633), 1,
      anon_sym_RPAREN,
    STATE(6365), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229119] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9850), 1,
      anon_sym_DOT,
    STATE(6366), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229133] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9852), 1,
      anon_sym_SQUOTE2,
    STATE(6367), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229147] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9854), 1,
      anon_sym_in,
    STATE(6368), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229161] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9856), 1,
      anon_sym_RPAREN,
    STATE(6369), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229175] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8316), 1,
      anon_sym_GT_RBRACE,
    STATE(6370), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229189] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9858), 1,
      anon_sym_RBRACK,
    STATE(6371), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229203] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9860), 1,
      anon_sym_RBRACE,
    STATE(6372), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229217] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9862), 1,
      anon_sym_with,
    STATE(6373), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229231] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8409), 1,
      anon_sym_GT_RBRACE,
    STATE(6374), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229245] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9864), 1,
      anon_sym_RBRACK,
    STATE(6375), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229259] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9866), 1,
      anon_sym_with,
    STATE(6376), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229273] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9868), 1,
      anon_sym_DOT,
    STATE(6377), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229287] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9870), 1,
      anon_sym_RPAREN,
    STATE(6378), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229301] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9872), 1,
      anon_sym_DOT,
    STATE(6379), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229315] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9874), 1,
      anon_sym_RPAREN,
    STATE(6380), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229329] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9876), 1,
      anon_sym_RPAREN,
    STATE(6381), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229343] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7726), 1,
      anon_sym_end,
    STATE(6382), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229357] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9878), 1,
      anon_sym_PLUS_EQ,
    STATE(6383), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229371] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9880), 1,
      anon_sym_DQUOTE,
    STATE(6384), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229385] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9882), 1,
      anon_sym_RBRACE,
    STATE(6385), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229399] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9884), 1,
      anon_sym_SQUOTE2,
    STATE(6386), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229413] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9886), 1,
      anon_sym_EQ,
    STATE(6387), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229427] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9888), 1,
      anon_sym_DQUOTE,
    STATE(6388), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229441] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9890), 1,
      anon_sym_RBRACK,
    STATE(6389), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229455] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9892), 1,
      anon_sym_DOT,
    STATE(6390), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229469] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9894), 1,
      anon_sym_RPAREN,
    STATE(6391), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229483] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9896), 1,
      anon_sym_EQ,
    STATE(6392), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229497] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9898), 1,
      anon_sym_RPAREN,
    STATE(6393), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229511] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7287), 1,
      anon_sym_GT,
    STATE(6394), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229525] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9900), 1,
      anon_sym_in,
    STATE(6395), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229539] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9902), 1,
      anon_sym_RBRACK,
    STATE(6396), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229553] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9904), 1,
      anon_sym_in,
    STATE(6397), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229567] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9906), 1,
      anon_sym_with,
    STATE(6398), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229581] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9908), 1,
      anon_sym_RBRACE,
    STATE(6399), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229595] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9910), 1,
      anon_sym_with,
    STATE(6400), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229609] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7273), 1,
      anon_sym_GT,
    STATE(6401), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229623] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9912), 1,
      sym__lowercase_identifier,
    STATE(6402), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229637] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9914), 1,
      anon_sym_RBRACK,
    STATE(6403), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229651] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9916), 1,
      anon_sym_DOT,
    STATE(6404), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229665] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9918), 1,
      anon_sym_RBRACK,
    STATE(6405), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229679] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9920), 1,
      anon_sym_RPAREN,
    STATE(6406), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229693] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9922), 1,
      anon_sym_PLUS_EQ,
    STATE(6407), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229707] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9924), 1,
      anon_sym_RPAREN,
    STATE(6408), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229721] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9926), 1,
      anon_sym_EQ,
    STATE(6409), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229735] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9928), 1,
      anon_sym_EQ,
    STATE(6410), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229749] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9930), 1,
      anon_sym_RBRACK,
    STATE(6411), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229763] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9932), 1,
      anon_sym_RPAREN,
    STATE(6412), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229777] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9934), 1,
      anon_sym_RBRACE,
    STATE(6413), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229791] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9936), 1,
      sym__lowercase_identifier,
    STATE(6414), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229805] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9938), 1,
      anon_sym_RPAREN,
    STATE(6415), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229819] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9940), 1,
      anon_sym_EQ,
    STATE(6416), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229833] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(475), 1,
      anon_sym_COLON_COLON,
    STATE(6417), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229847] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9942), 1,
      anon_sym_DOT,
    STATE(6418), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229861] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9944), 1,
      anon_sym_with,
    STATE(6419), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229875] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4422), 1,
      anon_sym_GT,
    STATE(6420), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229889] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9946), 1,
      anon_sym_RBRACK,
    STATE(6421), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229903] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9948), 1,
      anon_sym_RPAREN,
    STATE(6422), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229917] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9950), 1,
      anon_sym_in,
    STATE(6423), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229931] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9952), 1,
      anon_sym_RPAREN,
    STATE(6424), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229945] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9954), 1,
      anon_sym_RBRACK,
    STATE(6425), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229959] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9956), 1,
      anon_sym_RBRACE,
    STATE(6426), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229973] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9958), 1,
      anon_sym_RPAREN,
    STATE(6427), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [229987] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9706), 1,
      anon_sym_RBRACE,
    STATE(6428), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230001] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9960), 1,
      anon_sym_with,
    STATE(6429), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230015] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9962), 1,
      anon_sym_DOT,
    STATE(6430), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230029] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9964), 1,
      anon_sym_EQ,
    STATE(6431), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230043] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9966), 1,
      anon_sym_RBRACE,
    STATE(6432), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230057] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9968), 1,
      anon_sym_RPAREN,
    STATE(6433), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230071] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9970), 1,
      anon_sym_RBRACE,
    STATE(6434), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230085] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9972), 1,
      anon_sym_RPAREN,
    STATE(6435), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230099] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9720), 1,
      anon_sym_RBRACE,
    STATE(6436), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230113] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9974), 1,
      anon_sym_RPAREN,
    STATE(6437), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230127] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9974), 1,
      anon_sym_RBRACK,
    STATE(6438), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230141] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9974), 1,
      anon_sym_RBRACE,
    STATE(6439), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230155] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9976), 1,
      anon_sym_RBRACE,
    STATE(6440), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230169] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9978), 1,
      anon_sym_EQ,
    STATE(6441), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230183] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9980), 1,
      anon_sym_RPAREN,
    STATE(6442), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230197] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9982), 1,
      anon_sym_RPAREN,
    STATE(6443), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230211] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9984), 1,
      anon_sym_RBRACK,
    STATE(6444), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230225] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3602), 1,
      anon_sym_RBRACK,
    STATE(6445), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230239] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9986), 1,
      anon_sym_PLUS_EQ,
    STATE(6446), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230253] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9988), 1,
      anon_sym_RPAREN,
    STATE(6447), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230267] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9990), 1,
      anon_sym_RBRACE,
    STATE(6448), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230281] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9992), 1,
      anon_sym_RBRACE,
    STATE(6449), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230295] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7350), 1,
      anon_sym_COMMA,
    STATE(6450), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230309] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9994), 1,
      anon_sym_RPAREN,
    STATE(6451), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230323] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9996), 1,
      anon_sym_RBRACK,
    STATE(6452), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230337] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9998), 1,
      anon_sym_RPAREN,
    STATE(6453), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230351] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10000), 1,
      anon_sym_EQ,
    STATE(6454), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230365] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10002), 1,
      anon_sym_RPAREN,
    STATE(6455), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230379] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10004), 1,
      anon_sym_RPAREN,
    STATE(6456), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230393] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10006), 1,
      anon_sym_RPAREN,
    STATE(6457), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230407] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10008), 1,
      anon_sym_PIPE_RBRACK,
    STATE(6458), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230421] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10010), 1,
      anon_sym_end,
    STATE(6459), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230435] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3975), 1,
      anon_sym_RPAREN,
    STATE(6460), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230449] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7095), 1,
      sym__uppercase_identifier,
    STATE(6461), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230463] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7767), 1,
      anon_sym_RPAREN,
    STATE(6462), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230477] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10012), 1,
      anon_sym_in,
    STATE(6463), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230491] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10014), 1,
      anon_sym_RBRACE,
    STATE(6464), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230505] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10016), 1,
      anon_sym_of,
    STATE(6465), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230519] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7795), 1,
      anon_sym_end,
    STATE(6466), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230533] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10018), 1,
      anon_sym_EQ,
    STATE(6467), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230547] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10020), 1,
      anon_sym_DOT,
    STATE(6468), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230561] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10022), 1,
      anon_sym_end,
    STATE(6469), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230575] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10024), 1,
      anon_sym_RBRACK,
    STATE(6470), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230589] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10026), 1,
      anon_sym_DASH_GT,
    STATE(6471), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230603] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7712), 1,
      anon_sym_RPAREN,
    STATE(6472), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230617] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3923), 1,
      anon_sym_RPAREN,
    STATE(6473), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230631] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3646), 1,
      anon_sym_COMMA,
    STATE(6474), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230645] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10028), 1,
      anon_sym_LT_DASH,
    STATE(6475), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230659] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10030), 1,
      anon_sym_EQ,
    STATE(6476), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230673] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10032), 1,
      anon_sym_PIPE_RBRACK,
    STATE(6477), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230687] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10034), 1,
      anon_sym_RPAREN,
    STATE(6478), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230701] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10036), 1,
      anon_sym_COMMA,
    STATE(6479), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230715] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10038), 1,
      anon_sym_DOT_DOT,
    STATE(6480), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230729] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10040), 1,
      anon_sym_STAR,
    STATE(6481), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230743] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10042), 1,
      sym__lowercase_identifier,
    STATE(6482), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230757] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10044), 1,
      anon_sym_EQ,
    STATE(6483), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230771] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10046), 1,
      anon_sym_RPAREN,
    STATE(6484), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230785] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10048), 1,
      anon_sym_RBRACK,
    STATE(6485), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230799] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10050), 1,
      sym__lowercase_identifier,
    STATE(6486), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230813] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10052), 1,
      anon_sym_RBRACE,
    STATE(6487), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230827] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10054), 1,
      anon_sym_PLUS_EQ,
    STATE(6488), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230841] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10056), 1,
      anon_sym_in,
    STATE(6489), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230855] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10058), 1,
      anon_sym_DOT,
    STATE(6490), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230869] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10060), 1,
      anon_sym_RPAREN,
    STATE(6491), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230883] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10062), 1,
      anon_sym_RPAREN,
    STATE(6492), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230897] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10064), 1,
      anon_sym_DOT,
    STATE(6493), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230911] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4064), 1,
      anon_sym_COMMA,
    STATE(6494), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230925] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10066), 1,
      anon_sym_LT_DASH,
    STATE(6495), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230939] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10068), 1,
      anon_sym_DOT,
    STATE(6496), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230953] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10070), 1,
      anon_sym_RBRACK,
    STATE(6497), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230967] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10072), 1,
      anon_sym_RBRACK,
    STATE(6498), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230981] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10074), 1,
      anon_sym_RBRACE,
    STATE(6499), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [230995] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10076), 1,
      anon_sym_EQ,
    STATE(6500), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231009] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10078), 1,
      anon_sym_with,
    STATE(6501), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231023] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10080), 1,
      anon_sym_with,
    STATE(6502), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231037] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10082), 1,
      anon_sym_DOT,
    STATE(6503), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231051] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10084), 1,
      anon_sym_RPAREN,
    STATE(6504), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231065] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10086), 1,
      sym__lowercase_identifier,
    STATE(6505), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231079] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10088), 1,
      anon_sym_RBRACE,
    STATE(6506), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231093] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10090), 1,
      anon_sym_DASH_GT,
    STATE(6507), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231107] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10092), 1,
      anon_sym_RPAREN,
    STATE(6508), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231121] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10094), 1,
      anon_sym_RPAREN,
    STATE(6509), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231135] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10096), 1,
      aux_sym_number_token1,
    STATE(6510), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231149] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10098), 1,
      anon_sym_in,
    STATE(6511), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231163] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10100), 1,
      anon_sym_RPAREN,
    STATE(6512), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231177] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10102), 1,
      anon_sym_in,
    STATE(6513), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231191] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10104), 1,
      anon_sym_with,
    STATE(6514), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231205] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10106), 1,
      anon_sym_with,
    STATE(6515), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231219] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10108), 1,
      anon_sym_in,
    STATE(6516), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231233] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10110), 1,
      anon_sym_RPAREN,
    STATE(6517), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231247] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10112), 1,
      anon_sym_RPAREN,
    STATE(6518), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231261] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10114), 1,
      anon_sym_DOT,
    STATE(6519), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231275] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7712), 1,
      anon_sym_end,
    STATE(6520), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231289] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10116), 1,
      anon_sym_with,
    STATE(6521), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231303] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10118), 1,
      anon_sym_in,
    STATE(6522), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231317] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4453), 1,
      anon_sym_GT,
    STATE(6523), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231331] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10120), 1,
      anon_sym_with,
    STATE(6524), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231345] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10122), 1,
      anon_sym_RBRACK,
    STATE(6525), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231359] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10124), 1,
      anon_sym_RPAREN,
    STATE(6526), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231373] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10126), 1,
      anon_sym_in,
    STATE(6527), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231387] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10128), 1,
      anon_sym_in,
    STATE(6528), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231401] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7559), 1,
      anon_sym_DASH_GT,
    STATE(6529), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231415] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10130), 1,
      anon_sym_RBRACK,
    STATE(6530), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231429] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10132), 1,
      anon_sym_RBRACE,
    STATE(6531), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231443] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10134), 1,
      anon_sym_RBRACE,
    STATE(6532), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231457] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10136), 1,
      anon_sym_RBRACE,
    STATE(6533), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231471] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10138), 1,
      anon_sym_RPAREN,
    STATE(6534), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231485] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10140), 1,
      anon_sym_RPAREN,
    STATE(6535), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231499] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10142), 1,
      sym__lowercase_identifier,
    STATE(6536), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231513] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10144), 1,
      anon_sym_RPAREN,
    STATE(6537), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231527] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10144), 1,
      anon_sym_RBRACK,
    STATE(6538), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231541] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10144), 1,
      anon_sym_RBRACE,
    STATE(6539), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231555] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10146), 1,
      anon_sym_COLON2,
    STATE(6540), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231569] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10148), 1,
      anon_sym_DOT_DOT,
    STATE(6541), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231583] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10150), 1,
      anon_sym_EQ,
    STATE(6542), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231597] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10152), 1,
      anon_sym_PIPE_RBRACK,
    STATE(6543), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231611] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3606), 1,
      anon_sym_RBRACK,
    STATE(6544), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231625] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10154), 1,
      anon_sym_DOT,
    STATE(6545), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231639] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10156), 1,
      anon_sym_DASH_GT,
    STATE(6546), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231653] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10158), 1,
      anon_sym_RPAREN,
    STATE(6547), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231667] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10160), 1,
      anon_sym_SQUOTE2,
    STATE(6548), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231681] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10162), 1,
      anon_sym_RPAREN,
    STATE(6549), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231695] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10164), 1,
      anon_sym_DQUOTE,
    STATE(6550), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231709] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10166), 1,
      anon_sym_RBRACE,
    STATE(6551), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231723] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9656), 1,
      aux_sym_number_token1,
    STATE(6552), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231737] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10168), 1,
      anon_sym_DOT,
    STATE(6553), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231751] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10170), 1,
      anon_sym_end,
    STATE(6554), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231765] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10172), 1,
      anon_sym_DOT,
    STATE(6555), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231779] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10174), 1,
      anon_sym_DOT,
    STATE(6556), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231793] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8691), 1,
      anon_sym_GT_RBRACE,
    STATE(6557), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231807] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10176), 1,
      anon_sym_RBRACE,
    STATE(6558), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231821] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7682), 1,
      anon_sym_end,
    STATE(6559), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231835] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10178), 1,
      anon_sym_with,
    STATE(6560), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231849] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10180), 1,
      anon_sym_with,
    STATE(6561), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231863] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10182), 1,
      anon_sym_DOT,
    STATE(6562), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231877] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10184), 1,
      anon_sym_in,
    STATE(6563), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231891] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10186), 1,
      anon_sym_DASH_GT,
    STATE(6564), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231905] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10188), 1,
      anon_sym_RBRACE,
    STATE(6565), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231919] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10190), 1,
      anon_sym_SQUOTE2,
    STATE(6566), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231933] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10192), 1,
      anon_sym_DQUOTE,
    STATE(6567), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231947] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10194), 1,
      anon_sym_DASH_GT,
    STATE(6568), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231961] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10196), 1,
      anon_sym_with,
    STATE(6569), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231975] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10198), 1,
      anon_sym_with,
    STATE(6570), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [231989] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10200), 1,
      anon_sym_in,
    STATE(6571), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232003] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10202), 1,
      anon_sym_EQ,
    STATE(6572), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232017] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10204), 1,
      sym__lowercase_identifier,
    STATE(6573), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232031] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10206), 1,
      anon_sym_DOT,
    STATE(6574), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232045] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10208), 1,
      anon_sym_EQ,
    STATE(6575), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232059] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2828), 1,
      anon_sym_COMMA,
    STATE(6576), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232073] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10210), 1,
      anon_sym_DASH_GT,
    STATE(6577), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232087] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10212), 1,
      anon_sym_SQUOTE2,
    STATE(6578), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232101] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10214), 1,
      anon_sym_RPAREN,
    STATE(6579), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232115] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2384), 1,
      anon_sym_COMMA,
    STATE(6580), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232129] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10216), 1,
      anon_sym_RBRACE,
    STATE(6581), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232143] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4181), 1,
      anon_sym_COMMA,
    STATE(6582), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232157] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10218), 1,
      anon_sym_end,
    STATE(6583), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232171] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10220), 1,
      anon_sym_in,
    STATE(6584), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232185] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10222), 1,
      anon_sym_RPAREN,
    STATE(6585), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232199] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10224), 1,
      anon_sym_RPAREN,
    STATE(6586), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232213] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10224), 1,
      anon_sym_RBRACK,
    STATE(6587), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232227] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10224), 1,
      anon_sym_RBRACE,
    STATE(6588), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232241] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10226), 1,
      anon_sym_RBRACK,
    STATE(6589), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232255] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10228), 1,
      sym__lowercase_identifier,
    STATE(6590), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232269] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10230), 1,
      anon_sym_LT_DASH,
    STATE(6591), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232283] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10232), 1,
      anon_sym_DQUOTE,
    STATE(6592), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232297] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10234), 1,
      anon_sym_DOT,
    STATE(6593), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232311] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10236), 1,
      anon_sym_DASH_GT,
    STATE(6594), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232325] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10238), 1,
      anon_sym_RPAREN,
    STATE(6595), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232339] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10240), 1,
      anon_sym_in,
    STATE(6596), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232353] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10242), 1,
      anon_sym_RPAREN,
    STATE(6597), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232367] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10244), 1,
      anon_sym_DOT,
    STATE(6598), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232381] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10246), 1,
      anon_sym_EQ,
    STATE(6599), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232395] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10248), 1,
      anon_sym_DOT,
    STATE(6600), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232409] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10250), 1,
      anon_sym_DOT,
    STATE(6601), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232423] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10252), 1,
      anon_sym_RPAREN,
    STATE(6602), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232437] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10254), 1,
      anon_sym_DASH_GT,
    STATE(6603), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232451] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10256), 1,
      anon_sym_with,
    STATE(6604), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232465] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10258), 1,
      anon_sym_with,
    STATE(6605), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232479] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10260), 1,
      anon_sym_DOT,
    STATE(6606), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232493] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10262), 1,
      anon_sym_GT,
    STATE(6607), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232507] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10264), 1,
      anon_sym_RBRACK,
    STATE(6608), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232521] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10266), 1,
      anon_sym_COMMA,
    STATE(6609), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232535] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10268), 1,
      anon_sym_DASH_GT,
    STATE(6610), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232549] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10270), 1,
      sym__lowercase_identifier,
    STATE(6611), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232563] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10272), 1,
      anon_sym_with,
    STATE(6612), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232577] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10274), 1,
      anon_sym_with,
    STATE(6613), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232591] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4471), 1,
      anon_sym_GT,
    STATE(6614), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232605] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10276), 1,
      anon_sym_DASH_GT,
    STATE(6615), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232619] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10278), 1,
      anon_sym_RPAREN,
    STATE(6616), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232633] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10280), 1,
      anon_sym_DOT,
    STATE(6617), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232647] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1195), 1,
      anon_sym_COLON_COLON,
    STATE(6618), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232661] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10282), 1,
      anon_sym_DOT,
    STATE(6619), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232675] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10284), 1,
      anon_sym_RBRACE,
    STATE(6620), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232689] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10286), 1,
      anon_sym_in,
    STATE(6621), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232703] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10288), 1,
      anon_sym_RPAREN,
    STATE(6622), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232717] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10290), 1,
      anon_sym_EQ,
    STATE(6623), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232731] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10292), 1,
      anon_sym_RPAREN,
    STATE(6624), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232745] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10294), 1,
      anon_sym_RPAREN,
    STATE(6625), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232759] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10296), 1,
      anon_sym_end,
    STATE(6626), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232773] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3981), 1,
      anon_sym_RPAREN,
    STATE(6627), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232787] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10298), 1,
      anon_sym_RPAREN,
    STATE(6628), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232801] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10300), 1,
      anon_sym_RPAREN,
    STATE(6629), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232815] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10300), 1,
      anon_sym_RBRACK,
    STATE(6630), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232829] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10300), 1,
      anon_sym_RBRACE,
    STATE(6631), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232843] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10302), 1,
      anon_sym_RBRACE,
    STATE(6632), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232857] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10304), 1,
      anon_sym_COMMA,
    STATE(6633), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232871] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10306), 1,
      anon_sym_end,
    STATE(6634), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232885] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3967), 1,
      anon_sym_RPAREN,
    STATE(6635), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232899] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1211), 1,
      anon_sym_RPAREN,
    STATE(6636), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232913] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10308), 1,
      anon_sym_RPAREN,
    STATE(6637), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232927] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10310), 1,
      anon_sym_RPAREN,
    STATE(6638), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232941] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7324), 1,
      anon_sym_COMMA,
    STATE(6639), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232955] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10312), 1,
      anon_sym_DOT,
    STATE(6640), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232969] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10314), 1,
      anon_sym_DOT,
    STATE(6641), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232983] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10316), 1,
      anon_sym_RPAREN,
    STATE(6642), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [232997] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10318), 1,
      anon_sym_DOT,
    STATE(6643), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233011] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10320), 1,
      anon_sym_DOT,
    STATE(6644), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233025] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10322), 1,
      sym__lowercase_identifier,
    STATE(6645), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233039] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10324), 1,
      anon_sym_DASH_GT,
    STATE(6646), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233053] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10326), 1,
      anon_sym_with,
    STATE(6647), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233067] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10328), 1,
      anon_sym_with,
    STATE(6648), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233081] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10330), 1,
      anon_sym_DOT,
    STATE(6649), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233095] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10332), 1,
      anon_sym_RPAREN,
    STATE(6650), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233109] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10334), 1,
      anon_sym_in,
    STATE(6651), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233123] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1479), 1,
      anon_sym_COMMA,
    STATE(6652), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233137] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10336), 1,
      anon_sym_with,
    STATE(6653), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233151] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10338), 1,
      anon_sym_with,
    STATE(6654), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233165] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10340), 1,
      anon_sym_RPAREN,
    STATE(6655), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233179] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10342), 1,
      anon_sym_RBRACK,
    STATE(6656), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233193] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10344), 1,
      anon_sym_RBRACE,
    STATE(6657), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233207] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10346), 1,
      anon_sym_DOT,
    STATE(6658), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233221] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10348), 1,
      anon_sym_RBRACK,
    STATE(6659), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233235] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10350), 1,
      anon_sym_RPAREN,
    STATE(6660), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233249] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10352), 1,
      anon_sym_RBRACE,
    STATE(6661), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233263] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10354), 1,
      anon_sym_RPAREN,
    STATE(6662), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233277] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10356), 1,
      anon_sym_RPAREN,
    STATE(6663), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233291] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10358), 1,
      anon_sym_RPAREN,
    STATE(6664), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233305] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10360), 1,
      anon_sym_DASH_GT,
    STATE(6665), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233319] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10362), 1,
      anon_sym_RBRACE,
    STATE(6666), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233333] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10364), 1,
      anon_sym_RPAREN,
    STATE(6667), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233347] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10366), 1,
      anon_sym_RPAREN,
    STATE(6668), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233361] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10366), 1,
      anon_sym_RBRACK,
    STATE(6669), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233375] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10366), 1,
      anon_sym_RBRACE,
    STATE(6670), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233389] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10368), 1,
      anon_sym_RPAREN,
    STATE(6671), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233403] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10370), 1,
      anon_sym_RPAREN,
    STATE(6672), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233417] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7392), 1,
      anon_sym_COMMA,
    STATE(6673), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233431] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10372), 1,
      anon_sym_RBRACK,
    STATE(6674), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233445] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10374), 1,
      anon_sym_RPAREN,
    STATE(6675), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233459] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1229), 1,
      anon_sym_RPAREN,
    STATE(6676), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233473] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10376), 1,
      anon_sym_RBRACE,
    STATE(6677), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233487] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10378), 1,
      anon_sym_RPAREN,
    STATE(6678), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233501] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10380), 1,
      anon_sym_RPAREN,
    STATE(6679), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233515] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10382), 1,
      anon_sym_DOT,
    STATE(6680), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233529] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10384), 1,
      anon_sym_PIPE_RBRACK,
    STATE(6681), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233543] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10386), 1,
      anon_sym_DOT,
    STATE(6682), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233557] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10388), 1,
      anon_sym_DOT,
    STATE(6683), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233571] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10390), 1,
      anon_sym_RPAREN,
    STATE(6684), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233585] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10392), 1,
      anon_sym_RPAREN,
    STATE(6685), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233599] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10394), 1,
      anon_sym_with,
    STATE(6686), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233613] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10396), 1,
      anon_sym_with,
    STATE(6687), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233627] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10398), 1,
      anon_sym_DOT,
    STATE(6688), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233641] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10400), 1,
      anon_sym_RBRACK,
    STATE(6689), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233655] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10402), 1,
      anon_sym_RBRACE,
    STATE(6690), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233669] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10404), 1,
      anon_sym_RPAREN,
    STATE(6691), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233683] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10406), 1,
      anon_sym_with,
    STATE(6692), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233697] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10408), 1,
      anon_sym_with,
    STATE(6693), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233711] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10410), 1,
      anon_sym_RPAREN,
    STATE(6694), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233725] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10412), 1,
      anon_sym_RPAREN,
    STATE(6695), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233739] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10414), 1,
      anon_sym_RBRACK,
    STATE(6696), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233753] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10416), 1,
      anon_sym_DOT,
    STATE(6697), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233767] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10418), 1,
      anon_sym_SQUOTE2,
    STATE(6698), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233781] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10420), 1,
      anon_sym_DQUOTE,
    STATE(6699), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233795] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10422), 1,
      anon_sym_EQ,
    STATE(6700), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233809] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10424), 1,
      anon_sym_RPAREN,
    STATE(6701), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233823] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10426), 1,
      anon_sym_RBRACE,
    STATE(6702), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233837] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10428), 1,
      anon_sym_RBRACE,
    STATE(6703), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233851] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10430), 1,
      anon_sym_end,
    STATE(6704), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233865] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10432), 1,
      sym__lowercase_identifier,
    STATE(6705), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233879] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10434), 1,
      anon_sym_RPAREN,
    STATE(6706), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233893] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10436), 1,
      anon_sym_RPAREN,
    STATE(6707), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233907] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10436), 1,
      anon_sym_RBRACK,
    STATE(6708), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233921] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10436), 1,
      anon_sym_RBRACE,
    STATE(6709), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233935] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10438), 1,
      anon_sym_RPAREN,
    STATE(6710), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233949] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10440), 1,
      anon_sym_in,
    STATE(6711), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233963] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10442), 1,
      anon_sym_of,
    STATE(6712), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233977] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10444), 1,
      anon_sym_in,
    STATE(6713), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [233991] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7563), 1,
      anon_sym_DASH_GT,
    STATE(6714), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234005] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10446), 1,
      anon_sym_RPAREN,
    STATE(6715), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234019] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10448), 1,
      anon_sym_DASH_GT,
    STATE(6716), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234033] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7795), 1,
      anon_sym_RPAREN,
    STATE(6717), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234047] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10450), 1,
      anon_sym_DOT,
    STATE(6718), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234061] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10452), 1,
      anon_sym_RPAREN,
    STATE(6719), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234075] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10454), 1,
      anon_sym_DOT,
    STATE(6720), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234089] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10456), 1,
      anon_sym_RPAREN,
    STATE(6721), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234103] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10458), 1,
      anon_sym_RPAREN,
    STATE(6722), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234117] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10460), 1,
      anon_sym_with,
    STATE(6723), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234131] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10462), 1,
      anon_sym_with,
    STATE(6724), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234145] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10464), 1,
      anon_sym_DOT,
    STATE(6725), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234159] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10466), 1,
      anon_sym_end,
    STATE(6726), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234173] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10468), 1,
      anon_sym_RPAREN,
    STATE(6727), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234187] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9280), 1,
      anon_sym_RBRACE,
    STATE(6728), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234201] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10470), 1,
      anon_sym_with,
    STATE(6729), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234215] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10472), 1,
      anon_sym_with,
    STATE(6730), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234229] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10474), 1,
      anon_sym_RPAREN,
    STATE(6731), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234243] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10476), 1,
      anon_sym_RPAREN,
    STATE(6732), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234257] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9292), 1,
      anon_sym_RBRACE,
    STATE(6733), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234271] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10478), 1,
      anon_sym_DOT,
    STATE(6734), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234285] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10480), 1,
      anon_sym_in,
    STATE(6735), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234299] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10482), 1,
      anon_sym_RPAREN,
    STATE(6736), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234313] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10484), 1,
      anon_sym_DASH_GT,
    STATE(6737), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234327] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10486), 1,
      anon_sym_RPAREN,
    STATE(6738), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234341] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10488), 1,
      anon_sym_RBRACK,
    STATE(6739), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234355] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10490), 1,
      anon_sym_RPAREN,
    STATE(6740), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234369] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10492), 1,
      anon_sym_RPAREN,
    STATE(6741), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234383] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10494), 1,
      anon_sym_done,
    STATE(6742), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234397] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10496), 1,
      anon_sym_RPAREN,
    STATE(6743), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234411] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10498), 1,
      anon_sym_RPAREN,
    STATE(6744), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234425] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10498), 1,
      anon_sym_RBRACK,
    STATE(6745), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234439] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10498), 1,
      anon_sym_RBRACE,
    STATE(6746), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234453] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10500), 1,
      anon_sym_RBRACK,
    STATE(6747), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234467] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10502), 1,
      anon_sym_RBRACE,
    STATE(6748), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234481] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2954), 1,
      anon_sym_COMMA,
    STATE(6749), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234495] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10504), 1,
      anon_sym_RPAREN,
    STATE(6750), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234509] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10506), 1,
      anon_sym_LT_DASH,
    STATE(6751), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234523] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10508), 1,
      anon_sym_RBRACK,
    STATE(6752), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234537] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10510), 1,
      anon_sym_RBRACE,
    STATE(6753), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234551] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10512), 1,
      anon_sym_RBRACK,
    STATE(6754), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234565] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10514), 1,
      anon_sym_DOT,
    STATE(6755), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234579] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1205), 1,
      anon_sym_RPAREN,
    STATE(6756), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234593] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10516), 1,
      anon_sym_RPAREN,
    STATE(6757), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234607] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10516), 1,
      anon_sym_RBRACK,
    STATE(6758), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234621] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10518), 1,
      anon_sym_with,
    STATE(6759), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234635] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10520), 1,
      anon_sym_with,
    STATE(6760), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234649] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10522), 1,
      anon_sym_DOT,
    STATE(6761), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234663] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10524), 1,
      aux_sym_number_token1,
    STATE(6762), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234677] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10526), 1,
      anon_sym_in,
    STATE(6763), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234691] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10528), 1,
      anon_sym_RPAREN,
    STATE(6764), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234705] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10530), 1,
      anon_sym_with,
    STATE(6765), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234719] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10532), 1,
      anon_sym_with,
    STATE(6766), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234733] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10516), 1,
      anon_sym_RBRACE,
    STATE(6767), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234747] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10534), 1,
      anon_sym_RPAREN,
    STATE(6768), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234761] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1213), 1,
      anon_sym_RPAREN,
    STATE(6769), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234775] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10536), 1,
      anon_sym_DOT,
    STATE(6770), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234789] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10538), 1,
      anon_sym_RBRACK,
    STATE(6771), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234803] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10540), 1,
      anon_sym_RBRACE,
    STATE(6772), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234817] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10542), 1,
      anon_sym_RBRACE,
    STATE(6773), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234831] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10544), 1,
      anon_sym_RPAREN,
    STATE(6774), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234845] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10546), 1,
      anon_sym_RBRACK,
    STATE(6775), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234859] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10548), 1,
      anon_sym_RPAREN,
    STATE(6776), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234873] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10550), 1,
      anon_sym_in,
    STATE(6777), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234887] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10552), 1,
      anon_sym_SQUOTE2,
    STATE(6778), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234901] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10554), 1,
      anon_sym_RPAREN,
    STATE(6779), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234915] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10556), 1,
      anon_sym_RPAREN,
    STATE(6780), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234929] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10556), 1,
      anon_sym_RBRACK,
    STATE(6781), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234943] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10556), 1,
      anon_sym_RBRACE,
    STATE(6782), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234957] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10558), 1,
      anon_sym_RBRACK,
    STATE(6783), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234971] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10560), 1,
      anon_sym_EQ,
    STATE(6784), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234985] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10562), 1,
      anon_sym_RBRACK,
    STATE(6785), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [234999] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10564), 1,
      anon_sym_RPAREN,
    STATE(6786), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235013] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10566), 1,
      anon_sym_RBRACE,
    STATE(6787), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235027] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10568), 1,
      anon_sym_in,
    STATE(6788), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235041] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10570), 1,
      anon_sym_DASH_GT,
    STATE(6789), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235055] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10572), 1,
      anon_sym_DASH_GT,
    STATE(6790), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235069] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10574), 1,
      anon_sym_DOT,
    STATE(6791), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235083] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10576), 1,
      anon_sym_DQUOTE,
    STATE(6792), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235097] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10578), 1,
      anon_sym_PIPE_RBRACK,
    STATE(6793), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235111] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10580), 1,
      anon_sym_DOT,
    STATE(6794), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235125] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10582), 1,
      anon_sym_RBRACE,
    STATE(6795), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235139] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10584), 1,
      anon_sym_DOT,
    STATE(6796), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235153] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10586), 1,
      anon_sym_RPAREN,
    STATE(6797), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235167] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10588), 1,
      anon_sym_RPAREN,
    STATE(6798), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235181] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10590), 1,
      anon_sym_RPAREN,
    STATE(6799), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235195] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10592), 1,
      anon_sym_RPAREN,
    STATE(6800), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235209] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10594), 1,
      anon_sym_RBRACE,
    STATE(6801), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235223] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10596), 1,
      anon_sym_RPAREN,
    STATE(6802), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235237] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10598), 1,
      anon_sym_RPAREN,
    STATE(6803), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235251] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10598), 1,
      anon_sym_RBRACK,
    STATE(6804), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235265] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10598), 1,
      anon_sym_RBRACE,
    STATE(6805), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235279] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10600), 1,
      sym__lowercase_identifier,
    STATE(6806), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235293] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10602), 1,
      anon_sym_PIPE_RBRACK,
    STATE(6807), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235307] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10604), 1,
      anon_sym_RPAREN,
    STATE(6808), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235321] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10606), 1,
      anon_sym_SQUOTE2,
    STATE(6809), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235335] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8743), 1,
      anon_sym_GT_RBRACE,
    STATE(6810), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235349] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10608), 1,
      anon_sym_RPAREN,
    STATE(6811), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235363] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10610), 1,
      anon_sym_RBRACE,
    STATE(6812), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235377] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10612), 1,
      anon_sym_COMMA,
    STATE(6813), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235391] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10614), 1,
      anon_sym_EQ,
    STATE(6814), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235405] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7344), 1,
      anon_sym_COMMA,
    STATE(6815), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235419] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10616), 1,
      anon_sym_RPAREN,
    STATE(6816), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235433] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10616), 1,
      anon_sym_RBRACK,
    STATE(6817), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235447] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10616), 1,
      anon_sym_RBRACE,
    STATE(6818), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235461] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10618), 1,
      anon_sym_RPAREN,
    STATE(6819), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235475] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7720), 1,
      anon_sym_RPAREN,
    STATE(6820), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235489] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6727), 1,
      anon_sym_SQUOTE2,
    STATE(6821), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235503] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10620), 1,
      anon_sym_RBRACK,
    STATE(6822), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235517] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10622), 1,
      anon_sym_RBRACK,
    STATE(6823), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235531] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10624), 1,
      anon_sym_RBRACE,
    STATE(6824), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235545] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10626), 1,
      sym__lowercase_identifier,
    STATE(6825), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235559] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10628), 1,
      anon_sym_RPAREN,
    STATE(6826), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235573] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10628), 1,
      anon_sym_RBRACK,
    STATE(6827), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235587] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10628), 1,
      anon_sym_RBRACE,
    STATE(6828), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235601] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10630), 1,
      anon_sym_RPAREN,
    STATE(6829), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235615] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10632), 1,
      anon_sym_EQ,
    STATE(6830), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235629] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7720), 1,
      anon_sym_end,
    STATE(6831), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235643] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10634), 1,
      anon_sym_RPAREN,
    STATE(6832), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235657] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1221), 1,
      anon_sym_RPAREN,
    STATE(6833), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235671] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10636), 1,
      anon_sym_RPAREN,
    STATE(6834), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235685] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10638), 1,
      anon_sym_SQUOTE2,
    STATE(6835), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235699] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10640), 1,
      anon_sym_RPAREN,
    STATE(6836), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235713] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10640), 1,
      anon_sym_RBRACK,
    STATE(6837), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235727] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10640), 1,
      anon_sym_RBRACE,
    STATE(6838), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235741] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10642), 1,
      anon_sym_RBRACE,
    STATE(6839), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235755] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9272), 1,
      anon_sym_RBRACE,
    STATE(6840), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235769] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10644), 1,
      anon_sym_PLUS_EQ,
    STATE(6841), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235783] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10646), 1,
      anon_sym_SQUOTE2,
    STATE(6842), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235797] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10648), 1,
      anon_sym_RPAREN,
    STATE(6843), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235811] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10650), 1,
      anon_sym_RPAREN,
    STATE(6844), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235825] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10652), 1,
      anon_sym_DQUOTE,
    STATE(6845), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235839] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10654), 1,
      anon_sym_RPAREN,
    STATE(6846), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235853] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10654), 1,
      anon_sym_RBRACK,
    STATE(6847), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235867] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10654), 1,
      anon_sym_RBRACE,
    STATE(6848), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235881] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10656), 1,
      anon_sym_RBRACK,
    STATE(6849), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235895] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10658), 1,
      anon_sym_RBRACE,
    STATE(6850), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235909] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10660), 1,
      anon_sym_SQUOTE2,
    STATE(6851), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235923] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10662), 1,
      anon_sym_RPAREN,
    STATE(6852), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235937] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9234), 1,
      anon_sym_RBRACE,
    STATE(6853), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235951] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10664), 1,
      anon_sym_RBRACE,
    STATE(6854), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235965] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10666), 1,
      anon_sym_DOT,
    STATE(6855), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235979] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10668), 1,
      anon_sym_RPAREN,
    STATE(6856), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [235993] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10668), 1,
      anon_sym_RBRACK,
    STATE(6857), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236007] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10668), 1,
      anon_sym_RBRACE,
    STATE(6858), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236021] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10670), 1,
      sym__lowercase_identifier,
    STATE(6859), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236035] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9262), 1,
      anon_sym_RBRACE,
    STATE(6860), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236049] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10672), 1,
      anon_sym_DQUOTE,
    STATE(6861), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236063] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10674), 1,
      anon_sym_RPAREN,
    STATE(6862), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236077] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10676), 1,
      anon_sym_RPAREN,
    STATE(6863), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236091] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10678), 1,
      anon_sym_EQ,
    STATE(6864), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236105] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10680), 1,
      anon_sym_PIPE_RBRACK,
    STATE(6865), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236119] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10682), 1,
      anon_sym_RPAREN,
    STATE(6866), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236133] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10682), 1,
      anon_sym_RBRACK,
    STATE(6867), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236147] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10682), 1,
      anon_sym_RBRACE,
    STATE(6868), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236161] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10684), 1,
      sym__lowercase_identifier,
    STATE(6869), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236175] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10686), 1,
      sym__lowercase_identifier,
    STATE(6870), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236189] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10688), 1,
      anon_sym_DASH_GT,
    STATE(6871), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236203] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10690), 1,
      anon_sym_RPAREN,
    STATE(6872), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236217] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10692), 1,
      anon_sym_RPAREN,
    STATE(6873), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236231] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2712), 1,
      anon_sym_DASH_GT,
    STATE(6874), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236245] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10694), 1,
      anon_sym_end,
    STATE(6875), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236259] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10696), 1,
      anon_sym_RBRACE,
    STATE(6876), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236273] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10698), 1,
      anon_sym_RBRACE,
    STATE(6877), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236287] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8348), 1,
      anon_sym_GT_RBRACE,
    STATE(6878), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236301] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10700), 1,
      anon_sym_RBRACK,
    STATE(6879), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236315] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10702), 1,
      anon_sym_RBRACE,
    STATE(6880), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236329] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10704), 1,
      anon_sym_RBRACE,
    STATE(6881), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236343] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10706), 1,
      anon_sym_PIPE_RBRACK,
    STATE(6882), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236357] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4475), 1,
      anon_sym_DASH_GT,
    STATE(6883), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236371] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10708), 1,
      sym__uppercase_identifier,
    STATE(6884), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236385] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10710), 1,
      anon_sym_STAR,
    STATE(6885), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236399] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10712), 1,
      anon_sym_RPAREN,
    STATE(6886), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236413] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10714), 1,
      anon_sym_DOT,
    STATE(6887), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236427] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10716), 1,
      sym__lowercase_identifier,
    STATE(6888), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236441] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10718), 1,
      anon_sym_PIPE,
    STATE(6889), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236455] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10720), 1,
      anon_sym_RPAREN,
    STATE(6890), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236469] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10722), 1,
      sym__lowercase_identifier,
    STATE(6891), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236483] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1235), 1,
      anon_sym_RPAREN,
    STATE(6892), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236497] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10724), 1,
      anon_sym_done,
    STATE(6893), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236511] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10726), 1,
      anon_sym_RBRACE,
    STATE(6894), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236525] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10728), 1,
      anon_sym_RPAREN,
    STATE(6895), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236539] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10730), 1,
      anon_sym_RBRACE,
    STATE(6896), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236553] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10732), 1,
      anon_sym_end,
    STATE(6897), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236567] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10734), 1,
      anon_sym_in,
    STATE(6898), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236581] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10736), 1,
      anon_sym_SQUOTE2,
    STATE(6899), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236595] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10738), 1,
      anon_sym_DASH_GT,
    STATE(6900), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236609] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10740), 1,
      anon_sym_DQUOTE,
    STATE(6901), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236623] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10742), 1,
      anon_sym_DASH_GT,
    STATE(6902), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236637] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10744), 1,
      anon_sym_RPAREN,
    STATE(6903), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236651] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10746), 1,
      anon_sym_RBRACE,
    STATE(6904), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236665] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10748), 1,
      anon_sym_RPAREN,
    STATE(6905), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236679] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10750), 1,
      anon_sym_in,
    STATE(6906), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236693] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10752), 1,
      sym__lowercase_identifier,
    STATE(6907), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236707] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10754), 1,
      anon_sym_RBRACE,
    STATE(6908), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236721] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10756), 1,
      anon_sym_in,
    STATE(6909), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236735] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10758), 1,
      anon_sym_RPAREN,
    STATE(6910), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236749] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10760), 1,
      anon_sym_DASH_GT,
    STATE(6911), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236763] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10762), 1,
      ts_builtin_sym_end,
    STATE(6912), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236777] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7690), 1,
      anon_sym_end,
    STATE(6913), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236791] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10764), 1,
      anon_sym_DASH_GT,
    STATE(6914), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236805] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10766), 1,
      anon_sym_LT_DASH,
    STATE(6915), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236819] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10768), 1,
      anon_sym_RPAREN,
    STATE(6916), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236833] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10770), 1,
      anon_sym_EQ,
    STATE(6917), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236847] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10772), 1,
      anon_sym_DOT_DOT,
    STATE(6918), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236861] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10774), 1,
      anon_sym_DASH_GT,
    STATE(6919), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236875] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10776), 1,
      anon_sym_of,
    STATE(6920), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236889] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10778), 1,
      anon_sym_SQUOTE2,
    STATE(6921), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236903] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10780), 1,
      aux_sym_number_token1,
    STATE(6922), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236917] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2440), 1,
      anon_sym_RBRACK,
    STATE(6923), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236931] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10782), 1,
      sym__lowercase_identifier,
    STATE(6924), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236945] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10784), 1,
      sym__lowercase_identifier,
    STATE(6925), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236959] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10786), 1,
      sym__lowercase_identifier,
    STATE(6926), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236973] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10788), 1,
      anon_sym_DASH_GT,
    STATE(6927), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [236987] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10790), 1,
      anon_sym_DQUOTE,
    STATE(6928), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237001] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10792), 1,
      anon_sym_end,
    STATE(6929), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237015] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10794), 1,
      anon_sym_DQUOTE,
    STATE(6930), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237029] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10796), 1,
      anon_sym_in,
    STATE(6931), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237043] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9372), 1,
      anon_sym_RBRACE,
    STATE(6932), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237057] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10798), 1,
      anon_sym_RPAREN,
    STATE(6933), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237071] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10800), 1,
      anon_sym_RPAREN,
    STATE(6934), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237085] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10802), 1,
      anon_sym_RBRACK,
    STATE(6935), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237099] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10804), 1,
      anon_sym_end,
    STATE(6936), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237113] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10806), 1,
      sym__lowercase_identifier,
    STATE(6937), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237127] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10808), 1,
      anon_sym_RBRACK,
    STATE(6938), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237141] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10810), 1,
      anon_sym_RBRACE,
    STATE(6939), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237155] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10812), 1,
      anon_sym_RBRACK,
    STATE(6940), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237169] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10814), 1,
      anon_sym_DASH_GT,
    STATE(6941), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237183] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10816), 1,
      anon_sym_RBRACE,
    STATE(6942), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237197] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10818), 1,
      anon_sym_RPAREN,
    STATE(6943), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237211] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10820), 1,
      anon_sym_RPAREN,
    STATE(6944), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237225] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10822), 1,
      sym__lowercase_identifier,
    STATE(6945), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237239] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2708), 1,
      anon_sym_DASH_GT,
    STATE(6946), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237253] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10824), 1,
      anon_sym_RBRACK,
    STATE(6947), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237267] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10826), 1,
      anon_sym_RPAREN,
    STATE(6948), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237281] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1199), 1,
      anon_sym_RPAREN,
    STATE(6949), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237295] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10828), 1,
      anon_sym_DOT_DOT,
    STATE(6950), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237309] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10830), 1,
      anon_sym_DOT_DOT,
    STATE(6951), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237323] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10832), 1,
      anon_sym_DOT_DOT,
    STATE(6952), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237337] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10834), 1,
      anon_sym_RBRACK,
    STATE(6953), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237351] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10836), 1,
      sym__lowercase_identifier,
    STATE(6954), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237365] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10838), 1,
      anon_sym_RBRACE,
    STATE(6955), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237379] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10840), 1,
      anon_sym_RBRACE,
    STATE(6956), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237393] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10842), 1,
      anon_sym_RPAREN,
    STATE(6957), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237407] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10844), 1,
      anon_sym_RPAREN,
    STATE(6958), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237421] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10846), 1,
      anon_sym_RPAREN,
    STATE(6959), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237435] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10848), 1,
      sym__lowercase_identifier,
    STATE(6960), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237449] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10850), 1,
      anon_sym_RPAREN,
    STATE(6961), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237463] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10852), 1,
      sym__lowercase_identifier,
    STATE(6962), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237477] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10854), 1,
      anon_sym_EQ,
    STATE(6963), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237491] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10856), 1,
      anon_sym_SQUOTE2,
    STATE(6964), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237505] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10858), 1,
      anon_sym_RPAREN,
    STATE(6965), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237519] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10860), 1,
      anon_sym_EQ,
    STATE(6966), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237533] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10862), 1,
      anon_sym_DQUOTE,
    STATE(6967), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237547] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10864), 1,
      anon_sym_RBRACE,
    STATE(6968), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237561] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10866), 1,
      anon_sym_RBRACE,
    STATE(6969), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237575] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10868), 1,
      anon_sym_RBRACE,
    STATE(6970), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237589] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10870), 1,
      anon_sym_RPAREN,
    STATE(6971), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237603] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10872), 1,
      anon_sym_RPAREN,
    STATE(6972), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237617] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10874), 1,
      anon_sym_RBRACK,
    STATE(6973), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237631] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10876), 1,
      anon_sym_RBRACE,
    STATE(6974), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237645] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10878), 1,
      anon_sym_RPAREN,
    STATE(6975), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237659] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10880), 1,
      anon_sym_RPAREN,
    STATE(6976), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237673] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10882), 1,
      anon_sym_RPAREN,
    STATE(6977), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237687] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10884), 1,
      anon_sym_STAR,
    STATE(6978), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237701] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7745), 1,
      anon_sym_RPAREN,
    STATE(6979), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237715] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10886), 1,
      anon_sym_RPAREN,
    STATE(6980), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237729] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10888), 1,
      anon_sym_RPAREN,
    STATE(6981), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237743] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10890), 1,
      anon_sym_RPAREN,
    STATE(6982), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237757] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10892), 1,
      anon_sym_RPAREN,
    STATE(6983), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237771] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10894), 1,
      anon_sym_RPAREN,
    STATE(6984), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237785] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10896), 1,
      anon_sym_GT,
    STATE(6985), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237799] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10890), 1,
      anon_sym_RBRACK,
    STATE(6986), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237813] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10898), 1,
      anon_sym_DOT,
    STATE(6987), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237827] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7690), 1,
      anon_sym_RPAREN,
    STATE(6988), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237841] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10890), 1,
      anon_sym_RBRACE,
    STATE(6989), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237855] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10900), 1,
      sym__uppercase_identifier,
    STATE(6990), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237869] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10902), 1,
      anon_sym_EQ,
    STATE(6991), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237883] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10904), 1,
      sym__lowercase_identifier,
    STATE(6992), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237897] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10906), 1,
      anon_sym_RPAREN,
    STATE(6993), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237911] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10908), 1,
      anon_sym_RPAREN,
    STATE(6994), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237925] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10910), 1,
      anon_sym_end,
    STATE(6995), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237939] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10912), 1,
      anon_sym_RPAREN,
    STATE(6996), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237953] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10914), 1,
      anon_sym_DOT,
    STATE(6997), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237967] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10916), 1,
      anon_sym_RBRACK,
    STATE(6998), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237981] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10918), 1,
      anon_sym_EQ,
    STATE(6999), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [237995] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4080), 1,
      anon_sym_COMMA,
    STATE(7000), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238009] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10920), 1,
      anon_sym_LT_DASH,
    STATE(7001), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238023] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10922), 1,
      anon_sym_RBRACK,
    STATE(7002), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238037] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10924), 1,
      anon_sym_RBRACE,
    STATE(7003), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238051] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10926), 1,
      anon_sym_COMMA,
    STATE(7004), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238065] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10928), 1,
      anon_sym_in,
    STATE(7005), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238079] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10930), 1,
      anon_sym_RPAREN,
    STATE(7006), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238093] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10932), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7007), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238107] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10934), 1,
      anon_sym_RPAREN,
    STATE(7008), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238121] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10936), 1,
      anon_sym_RPAREN,
    STATE(7009), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238135] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10938), 1,
      sym__lowercase_identifier,
    STATE(7010), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238149] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10940), 1,
      anon_sym_in,
    STATE(7011), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238163] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10942), 1,
      anon_sym_RBRACE,
    STATE(7012), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238177] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10944), 1,
      anon_sym_RBRACK,
    STATE(7013), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238191] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10946), 1,
      anon_sym_RBRACK,
    STATE(7014), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238205] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10948), 1,
      anon_sym_RBRACE,
    STATE(7015), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238219] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7500), 1,
      anon_sym_DASH_GT,
    STATE(7016), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238233] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10950), 1,
      anon_sym_RBRACK,
    STATE(7017), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238247] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10952), 1,
      anon_sym_RPAREN,
    STATE(7018), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238261] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10954), 1,
      anon_sym_RBRACE,
    STATE(7019), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238275] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10956), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7020), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238289] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10958), 1,
      anon_sym_RPAREN,
    STATE(7021), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238303] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7726), 1,
      anon_sym_RPAREN,
    STATE(7022), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238317] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1183), 1,
      anon_sym_RPAREN,
    STATE(7023), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238331] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10960), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7024), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238345] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8828), 1,
      anon_sym_GT_RBRACE,
    STATE(7025), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238359] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10962), 1,
      anon_sym_RPAREN,
    STATE(7026), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238373] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7745), 1,
      anon_sym_end,
    STATE(7027), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238387] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4334), 1,
      anon_sym_COMMA,
    STATE(7028), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238401] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10964), 1,
      anon_sym_LT_DASH,
    STATE(7029), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238415] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10966), 1,
      anon_sym_in,
    STATE(7030), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238429] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10968), 1,
      anon_sym_RPAREN,
    STATE(7031), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238443] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10970), 1,
      anon_sym_RBRACE,
    STATE(7032), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238457] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10972), 1,
      anon_sym_SQUOTE2,
    STATE(7033), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238471] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10974), 1,
      anon_sym_in,
    STATE(7034), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238485] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10976), 1,
      anon_sym_DQUOTE,
    STATE(7035), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238499] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10978), 1,
      anon_sym_DASH_GT,
    STATE(7036), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238513] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10980), 1,
      anon_sym_DASH_GT,
    STATE(7037), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238527] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10982), 1,
      anon_sym_SQUOTE2,
    STATE(7038), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238541] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10984), 1,
      sym__lowercase_identifier,
    STATE(7039), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238555] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10986), 1,
      anon_sym_DQUOTE,
    STATE(7040), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238569] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10988), 1,
      anon_sym_RBRACE,
    STATE(7041), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238583] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10990), 1,
      anon_sym_in,
    STATE(7042), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238597] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10992), 1,
      sym__lowercase_identifier,
    STATE(7043), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238611] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10994), 1,
      anon_sym_RPAREN,
    STATE(7044), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238625] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10996), 1,
      anon_sym_RPAREN,
    STATE(7045), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238639] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(10998), 1,
      anon_sym_RBRACE,
    STATE(7046), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238653] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11000), 1,
      anon_sym_RPAREN,
    STATE(7047), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238667] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5882), 1,
      anon_sym_DASH_GT,
    STATE(7048), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238681] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11002), 1,
      anon_sym_in,
    STATE(7049), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238695] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11004), 1,
      anon_sym_STAR,
    STATE(7050), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238709] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11006), 1,
      anon_sym_PIPE,
    STATE(7051), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238723] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11008), 1,
      anon_sym_RPAREN,
    STATE(7052), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238737] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11010), 1,
      anon_sym_RBRACK,
    STATE(7053), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238751] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11012), 1,
      anon_sym_RPAREN,
    STATE(7054), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238765] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7338), 1,
      anon_sym_RBRACK,
    STATE(7055), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238779] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11014), 1,
      anon_sym_DASH_GT,
    STATE(7056), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238793] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11016), 1,
      anon_sym_DASH_GT,
    STATE(7057), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238807] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11018), 1,
      anon_sym_COMMA,
    STATE(7058), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238821] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5364), 1,
      anon_sym_in,
    STATE(7059), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238835] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11020), 1,
      sym__lowercase_identifier,
    STATE(7060), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238849] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11022), 1,
      anon_sym_RPAREN,
    STATE(7061), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238863] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11024), 1,
      anon_sym_DASH_GT,
    STATE(7062), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238877] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11026), 1,
      anon_sym_RPAREN,
    STATE(7063), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238891] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7102), 1,
      anon_sym_COMMA,
    STATE(7064), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238905] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11028), 1,
      anon_sym_in,
    STATE(7065), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238919] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11030), 1,
      anon_sym_in,
    STATE(7066), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238933] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11032), 1,
      anon_sym_end,
    STATE(7067), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238947] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11034), 1,
      anon_sym_RBRACE,
    STATE(7068), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238961] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11036), 1,
      sym__lowercase_identifier,
    STATE(7069), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238975] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11038), 1,
      anon_sym_RBRACK,
    STATE(7070), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [238989] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11040), 1,
      anon_sym_RPAREN,
    STATE(7071), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239003] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11042), 1,
      anon_sym_RPAREN,
    STATE(7072), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239017] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11044), 1,
      anon_sym_with,
    STATE(7073), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239031] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11046), 1,
      anon_sym_RBRACE,
    STATE(7074), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239045] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11048), 1,
      anon_sym_RPAREN,
    STATE(7075), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239059] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11050), 1,
      anon_sym_RBRACE,
    STATE(7076), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239073] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11052), 1,
      sym__lowercase_identifier,
    STATE(7077), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239087] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11054), 1,
      anon_sym_DQUOTE,
    STATE(7078), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239101] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11056), 1,
      anon_sym_DASH_GT,
    STATE(7079), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239115] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11058), 1,
      anon_sym_RPAREN,
    STATE(7080), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239129] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11060), 1,
      anon_sym_DOT_DOT,
    STATE(7081), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239143] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11062), 1,
      anon_sym_DOT_DOT,
    STATE(7082), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239157] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11064), 1,
      anon_sym_DOT_DOT,
    STATE(7083), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239171] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11066), 1,
      anon_sym_RBRACE,
    STATE(7084), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239185] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11068), 1,
      anon_sym_in,
    STATE(7085), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239199] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11070), 1,
      anon_sym_RPAREN,
    STATE(7086), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239213] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4099), 1,
      anon_sym_COMMA,
    STATE(7087), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239227] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11072), 1,
      anon_sym_LT_DASH,
    STATE(7088), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239241] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11074), 1,
      anon_sym_RPAREN,
    STATE(7089), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239255] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11076), 1,
      anon_sym_RBRACK,
    STATE(7090), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239269] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11078), 1,
      sym__lowercase_identifier,
    STATE(7091), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239283] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11080), 1,
      anon_sym_RBRACE,
    STATE(7092), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239297] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11082), 1,
      anon_sym_DASH_GT,
    STATE(7093), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239311] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11084), 1,
      anon_sym_DASH_GT,
    STATE(7094), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239325] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11086), 1,
      anon_sym_RBRACK,
    STATE(7095), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239339] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11088), 1,
      sym__lowercase_identifier,
    STATE(7096), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239353] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11090), 1,
      anon_sym_RBRACE,
    STATE(7097), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239367] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11092), 1,
      anon_sym_RPAREN,
    STATE(7098), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239381] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11094), 1,
      anon_sym_with,
    STATE(7099), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239395] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11096), 1,
      anon_sym_RBRACK,
    STATE(7100), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239409] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11098), 1,
      sym__lowercase_identifier,
    STATE(7101), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239423] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11100), 1,
      anon_sym_RBRACE,
    STATE(7102), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239437] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11102), 1,
      anon_sym_RPAREN,
    STATE(7103), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239451] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11104), 1,
      anon_sym_RBRACE,
    STATE(7104), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239465] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11106), 1,
      anon_sym_DOT_DOT,
    STATE(7105), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239479] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11108), 1,
      anon_sym_DOT_DOT,
    STATE(7106), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239493] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11110), 1,
      anon_sym_DOT_DOT,
    STATE(7107), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239507] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11112), 1,
      anon_sym_PLUS_EQ,
    STATE(7108), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239521] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11114), 1,
      anon_sym_RBRACK,
    STATE(7109), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239535] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11116), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7110), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239549] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11118), 1,
      anon_sym_RBRACE,
    STATE(7111), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239563] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11120), 1,
      anon_sym_RPAREN,
    STATE(7112), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239577] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11122), 1,
      anon_sym_RBRACK,
    STATE(7113), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239591] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11124), 1,
      anon_sym_RBRACE,
    STATE(7114), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239605] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11126), 1,
      anon_sym_RPAREN,
    STATE(7115), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239619] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2974), 1,
      anon_sym_COMMA,
    STATE(7116), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239633] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11128), 1,
      anon_sym_RBRACE,
    STATE(7117), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239647] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11130), 1,
      aux_sym_number_token1,
    STATE(7118), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239661] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11132), 1,
      sym__lowercase_identifier,
    STATE(7119), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239675] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11134), 1,
      aux_sym_number_token1,
    STATE(7120), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239689] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11136), 1,
      anon_sym_DOT_DOT,
    STATE(7121), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239703] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11138), 1,
      anon_sym_DOT_DOT,
    STATE(7122), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239717] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11140), 1,
      anon_sym_DOT_DOT,
    STATE(7123), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239731] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5482), 1,
      anon_sym_in,
    STATE(7124), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239745] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2716), 1,
      anon_sym_DASH_GT,
    STATE(7125), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239759] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11142), 1,
      anon_sym_in,
    STATE(7126), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239773] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11144), 1,
      anon_sym_type,
    STATE(7127), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239787] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11146), 1,
      anon_sym_RPAREN,
    STATE(7128), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239801] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11148), 1,
      anon_sym_RPAREN,
    STATE(7129), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239815] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11150), 1,
      anon_sym_RPAREN,
    STATE(7130), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239829] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6795), 1,
      anon_sym_DASH_GT,
    STATE(7131), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239843] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11152), 1,
      anon_sym_RBRACK,
    STATE(7132), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239857] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7682), 1,
      anon_sym_RPAREN,
    STATE(7133), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239871] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11154), 1,
      sym__lowercase_identifier,
    STATE(7134), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239885] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1197), 1,
      anon_sym_RPAREN,
    STATE(7135), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239899] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11156), 1,
      anon_sym_DOT_DOT,
    STATE(7136), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239913] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11158), 1,
      anon_sym_DOT_DOT,
    STATE(7137), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239927] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11160), 1,
      anon_sym_DOT_DOT,
    STATE(7138), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239941] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11162), 1,
      anon_sym_RBRACE,
    STATE(7139), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239955] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7793), 1,
      anon_sym_RPAREN,
    STATE(7140), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239969] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11164), 1,
      anon_sym_RBRACK,
    STATE(7141), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239983] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11166), 1,
      anon_sym_DOT,
    STATE(7142), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [239997] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11168), 1,
      anon_sym_RBRACE,
    STATE(7143), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240011] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11170), 1,
      anon_sym_end,
    STATE(7144), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240025] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11172), 1,
      anon_sym_in,
    STATE(7145), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240039] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11174), 1,
      anon_sym_RPAREN,
    STATE(7146), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240053] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11176), 1,
      anon_sym_RBRACK,
    STATE(7147), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240067] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11178), 1,
      anon_sym_in,
    STATE(7148), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240081] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8372), 1,
      anon_sym_GT_RBRACE,
    STATE(7149), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240095] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11180), 1,
      anon_sym_DOT_DOT,
    STATE(7150), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240109] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11182), 1,
      anon_sym_DOT_DOT,
    STATE(7151), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240123] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11184), 1,
      anon_sym_DOT_DOT,
    STATE(7152), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240137] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11186), 1,
      anon_sym_RBRACE,
    STATE(7153), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240151] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6456), 1,
      anon_sym_in,
    STATE(7154), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240165] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1191), 1,
      anon_sym_COLON_COLON,
    STATE(7155), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240179] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11188), 1,
      anon_sym_RBRACK,
    STATE(7156), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240193] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11190), 1,
      anon_sym_RBRACE,
    STATE(7157), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240207] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11192), 1,
      sym__right_quoted_string_delimiter,
    STATE(7158), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240221] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11194), 1,
      anon_sym_in,
    STATE(7159), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240235] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11196), 1,
      anon_sym_DOT,
    STATE(7160), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240249] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11198), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7161), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240263] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11200), 1,
      anon_sym_RPAREN,
    STATE(7162), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240277] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11202), 1,
      anon_sym_RBRACE,
    STATE(7163), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240291] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11204), 1,
      anon_sym_DOT_DOT,
    STATE(7164), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240305] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11206), 1,
      anon_sym_DOT_DOT,
    STATE(7165), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240319] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11208), 1,
      anon_sym_DOT_DOT,
    STATE(7166), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240333] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7374), 1,
      anon_sym_COMMA,
    STATE(7167), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240347] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(6553), 1,
      anon_sym_DASH_GT,
    STATE(7168), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240361] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11210), 1,
      anon_sym_RPAREN,
    STATE(7169), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240375] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11212), 1,
      anon_sym_STAR,
    STATE(7170), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240389] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11214), 1,
      anon_sym_COMMA,
    STATE(7171), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240403] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11216), 1,
      anon_sym_DASH_GT,
    STATE(7172), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240417] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11218), 1,
      anon_sym_in,
    STATE(7173), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240431] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11220), 1,
      anon_sym_RPAREN,
    STATE(7174), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240445] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11222), 1,
      anon_sym_RPAREN,
    STATE(7175), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240459] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11224), 1,
      anon_sym_DOT,
    STATE(7176), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240473] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8587), 1,
      anon_sym_GT_RBRACE,
    STATE(7177), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240487] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11226), 1,
      anon_sym_DOT_DOT,
    STATE(7178), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240501] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11228), 1,
      anon_sym_DOT_DOT,
    STATE(7179), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240515] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11230), 1,
      anon_sym_DOT_DOT,
    STATE(7180), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240529] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11232), 1,
      anon_sym_RBRACE,
    STATE(7181), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240543] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11234), 1,
      anon_sym_RBRACK,
    STATE(7182), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240557] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7633), 1,
      anon_sym_end,
    STATE(7183), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240571] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11236), 1,
      anon_sym_RBRACE,
    STATE(7184), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240585] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7263), 1,
      anon_sym_GT,
    STATE(7185), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240599] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11238), 1,
      anon_sym_RBRACE,
    STATE(7186), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240613] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1237), 1,
      anon_sym_RPAREN,
    STATE(7187), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240627] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11240), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7188), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240641] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11242), 1,
      anon_sym_SQUOTE2,
    STATE(7189), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240655] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11244), 1,
      anon_sym_DQUOTE,
    STATE(7190), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240669] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11246), 1,
      anon_sym_RBRACE,
    STATE(7191), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240683] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11248), 1,
      anon_sym_DOT_DOT,
    STATE(7192), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240697] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11250), 1,
      anon_sym_DOT_DOT,
    STATE(7193), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240711] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11252), 1,
      anon_sym_DOT_DOT,
    STATE(7194), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240725] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11254), 1,
      anon_sym_DASH_GT,
    STATE(7195), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240739] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4227), 1,
      anon_sym_COMMA,
    STATE(7196), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240753] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11256), 1,
      anon_sym_LT_DASH,
    STATE(7197), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240767] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9020), 1,
      anon_sym_GT_RBRACE,
    STATE(7198), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240781] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11258), 1,
      anon_sym_DOT_DOT,
    STATE(7199), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240795] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11260), 1,
      anon_sym_DOT_DOT,
    STATE(7200), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240809] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11262), 1,
      anon_sym_DOT_DOT,
    STATE(7201), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240823] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11264), 1,
      sym__lowercase_identifier,
    STATE(7202), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240837] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11266), 1,
      anon_sym_PIPE,
    STATE(7203), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240851] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11268), 1,
      anon_sym_DOT_DOT,
    STATE(7204), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240865] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11270), 1,
      anon_sym_DOT_DOT,
    STATE(7205), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240879] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11272), 1,
      anon_sym_DOT_DOT,
    STATE(7206), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240893] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11274), 1,
      anon_sym_RBRACE,
    STATE(7207), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240907] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11276), 1,
      anon_sym_DASH_GT,
    STATE(7208), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240921] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11278), 1,
      anon_sym_DOT_DOT,
    STATE(7209), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240935] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11280), 1,
      anon_sym_DOT_DOT,
    STATE(7210), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240949] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11282), 1,
      anon_sym_DOT_DOT,
    STATE(7211), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240963] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11284), 1,
      anon_sym_in,
    STATE(7212), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240977] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2992), 1,
      anon_sym_COMMA,
    STATE(7213), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [240991] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11286), 1,
      anon_sym_DOT_DOT,
    STATE(7214), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241005] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11288), 1,
      anon_sym_DOT_DOT,
    STATE(7215), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241019] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11290), 1,
      anon_sym_DOT_DOT,
    STATE(7216), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241033] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11292), 1,
      anon_sym_RBRACK,
    STATE(7217), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241047] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(9466), 1,
      anon_sym_RPAREN,
    STATE(7218), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241061] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11294), 1,
      anon_sym_DOT_DOT,
    STATE(7219), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241075] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11296), 1,
      anon_sym_DOT_DOT,
    STATE(7220), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241089] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11298), 1,
      anon_sym_DOT_DOT,
    STATE(7221), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241103] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11300), 1,
      anon_sym_RPAREN,
    STATE(7222), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241117] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11302), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7223), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241131] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11304), 1,
      anon_sym_DOT_DOT,
    STATE(7224), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241145] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11306), 1,
      anon_sym_DOT_DOT,
    STATE(7225), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241159] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11308), 1,
      anon_sym_DOT_DOT,
    STATE(7226), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241173] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11310), 1,
      anon_sym_end,
    STATE(7227), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241187] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11312), 1,
      anon_sym_of,
    STATE(7228), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241201] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11314), 1,
      anon_sym_DOT_DOT,
    STATE(7229), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241215] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11316), 1,
      anon_sym_DOT_DOT,
    STATE(7230), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241229] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11318), 1,
      anon_sym_DOT_DOT,
    STATE(7231), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241243] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11320), 1,
      anon_sym_in,
    STATE(7232), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241257] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11322), 1,
      anon_sym_in,
    STATE(7233), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241271] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11324), 1,
      anon_sym_DASH_GT,
    STATE(7234), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241285] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11326), 1,
      anon_sym_RBRACE,
    STATE(7235), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241299] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11328), 1,
      anon_sym_DASH_GT,
    STATE(7236), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241313] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11330), 1,
      anon_sym_RPAREN,
    STATE(7237), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241327] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7767), 1,
      anon_sym_end,
    STATE(7238), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241341] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11332), 1,
      anon_sym_in,
    STATE(7239), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241355] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11334), 1,
      anon_sym_type,
    STATE(7240), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241369] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11336), 1,
      sym_indexing_operator,
    STATE(7241), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241383] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11338), 1,
      anon_sym_in,
    STATE(7242), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241397] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11340), 1,
      anon_sym_RPAREN,
    STATE(7243), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241411] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11342), 1,
      anon_sym_SQUOTE2,
    STATE(7244), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241425] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11344), 1,
      anon_sym_DQUOTE,
    STATE(7245), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241439] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(5777), 1,
      anon_sym_DASH_GT,
    STATE(7246), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241453] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11346), 1,
      sym__lowercase_identifier,
    STATE(7247), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241467] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11348), 1,
      anon_sym_DOT,
    STATE(7248), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241481] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11350), 1,
      anon_sym_RBRACK,
    STATE(7249), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241495] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11352), 1,
      anon_sym_RBRACE,
    STATE(7250), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241509] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11354), 1,
      anon_sym_in,
    STATE(7251), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241523] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4117), 1,
      anon_sym_DASH_GT,
    STATE(7252), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241537] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11356), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7253), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241551] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11358), 1,
      anon_sym_GT,
    STATE(7254), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241565] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11360), 1,
      anon_sym_end,
    STATE(7255), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241579] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11362), 1,
      anon_sym_DASH_GT,
    STATE(7256), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241593] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11364), 1,
      anon_sym_RPAREN,
    STATE(7257), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241607] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11366), 1,
      anon_sym_RBRACK,
    STATE(7258), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241621] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11368), 1,
      anon_sym_DASH_GT,
    STATE(7259), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241635] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11370), 1,
      anon_sym_RPAREN,
    STATE(7260), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241649] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11372), 1,
      anon_sym_EQ,
    STATE(7261), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241663] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11374), 1,
      sym__lowercase_identifier,
    STATE(7262), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241677] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11376), 1,
      anon_sym_DASH_GT,
    STATE(7263), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241691] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11378), 1,
      anon_sym_in,
    STATE(7264), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241705] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11380), 1,
      anon_sym_DASH_GT,
    STATE(7265), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241719] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11382), 1,
      anon_sym_DASH_GT,
    STATE(7266), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241733] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11384), 1,
      anon_sym_RBRACK,
    STATE(7267), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241747] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11386), 1,
      anon_sym_RBRACK,
    STATE(7268), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241761] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11388), 1,
      anon_sym_RBRACE,
    STATE(7269), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241775] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11390), 1,
      anon_sym_RPAREN,
    STATE(7270), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241789] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11392), 1,
      anon_sym_RPAREN,
    STATE(7271), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241803] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11394), 1,
      anon_sym_end,
    STATE(7272), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241817] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11396), 1,
      sym_indexing_operator,
    STATE(7273), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241831] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11398), 1,
      anon_sym_DASH_GT,
    STATE(7274), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241845] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11400), 1,
      anon_sym_RPAREN,
    STATE(7275), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241859] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11402), 1,
      anon_sym_EQ,
    STATE(7276), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241873] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11404), 1,
      sym_indexing_operator,
    STATE(7277), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241887] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11406), 1,
      anon_sym_RBRACK,
    STATE(7278), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241901] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11408), 1,
      sym__lowercase_identifier,
    STATE(7279), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241915] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11410), 1,
      sym_indexing_operator,
    STATE(7280), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241929] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11412), 1,
      sym_indexing_operator,
    STATE(7281), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241943] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11414), 1,
      sym_indexing_operator,
    STATE(7282), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241957] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11416), 1,
      sym_indexing_operator,
    STATE(7283), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241971] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11418), 1,
      sym_indexing_operator,
    STATE(7284), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241985] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11420), 1,
      sym_indexing_operator,
    STATE(7285), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [241999] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11422), 1,
      sym_indexing_operator,
    STATE(7286), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242013] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11424), 1,
      sym_indexing_operator,
    STATE(7287), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242027] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11426), 1,
      sym_indexing_operator,
    STATE(7288), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242041] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11428), 1,
      sym_indexing_operator,
    STATE(7289), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242055] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11430), 1,
      sym_indexing_operator,
    STATE(7290), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242069] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11432), 1,
      sym_indexing_operator,
    STATE(7291), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242083] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11434), 1,
      sym_indexing_operator,
    STATE(7292), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242097] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11436), 1,
      anon_sym_RPAREN,
    STATE(7293), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242111] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11438), 1,
      anon_sym_DASH_GT,
    STATE(7294), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242125] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(2896), 1,
      anon_sym_COMMA,
    STATE(7295), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242139] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11440), 1,
      anon_sym_RPAREN,
    STATE(7296), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242153] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11442), 1,
      anon_sym_LT_DASH,
    STATE(7297), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242167] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11444), 1,
      anon_sym_RPAREN,
    STATE(7298), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242181] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11446), 1,
      anon_sym_end,
    STATE(7299), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242195] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11448), 1,
      anon_sym_RBRACK,
    STATE(7300), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242209] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11450), 1,
      anon_sym_DOT,
    STATE(7301), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242223] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11452), 1,
      anon_sym_RBRACE,
    STATE(7302), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242237] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7462), 1,
      anon_sym_DASH_GT,
    STATE(7303), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242251] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11454), 1,
      anon_sym_RBRACE,
    STATE(7304), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242265] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11456), 1,
      anon_sym_RPAREN,
    STATE(7305), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242279] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11458), 1,
      anon_sym_DASH_GT,
    STATE(7306), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242293] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11460), 1,
      anon_sym_RBRACE,
    STATE(7307), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242307] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11462), 1,
      anon_sym_RPAREN,
    STATE(7308), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242321] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11464), 1,
      anon_sym_DOT,
    STATE(7309), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242335] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11466), 1,
      sym_indexing_operator,
    STATE(7310), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242349] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11468), 1,
      anon_sym_RPAREN,
    STATE(7311), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242363] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11470), 1,
      anon_sym_RBRACK,
    STATE(7312), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242377] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11472), 1,
      anon_sym_RBRACE,
    STATE(7313), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242391] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11474), 1,
      anon_sym_RPAREN,
    STATE(7314), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242405] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7428), 1,
      anon_sym_DASH_GT,
    STATE(7315), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242419] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11476), 1,
      anon_sym_DASH_GT,
    STATE(7316), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242433] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7438), 1,
      anon_sym_DASH_GT,
    STATE(7317), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242447] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11478), 1,
      anon_sym_RBRACE,
    STATE(7318), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242461] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11480), 1,
      anon_sym_end,
    STATE(7319), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242475] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(1193), 1,
      anon_sym_RPAREN,
    STATE(7320), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242489] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11482), 1,
      anon_sym_RPAREN,
    STATE(7321), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242503] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(8405), 1,
      anon_sym_GT_RBRACE,
    STATE(7322), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242517] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11484), 1,
      anon_sym_RPAREN,
    STATE(7323), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242531] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4052), 1,
      anon_sym_COMMA,
    STATE(7324), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242545] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11486), 1,
      anon_sym_LT_DASH,
    STATE(7325), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242559] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11488), 1,
      anon_sym_in,
    STATE(7326), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242573] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11490), 1,
      anon_sym_EQ,
    STATE(7327), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242587] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11492), 1,
      anon_sym_COMMA,
    STATE(7328), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242601] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(4165), 1,
      anon_sym_COMMA,
    STATE(7329), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242615] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11494), 1,
      anon_sym_LT_DASH,
    STATE(7330), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242629] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11496), 1,
      anon_sym_with,
    STATE(7331), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242643] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3971), 1,
      anon_sym_RPAREN,
    STATE(7332), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242657] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(7873), 1,
      sym__uppercase_identifier,
    STATE(7333), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242671] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11498), 1,
      anon_sym_RPAREN,
    STATE(7334), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242685] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(3678), 1,
      anon_sym_COMMA,
    STATE(7335), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242699] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11500), 1,
      anon_sym_LT_DASH,
    STATE(7336), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242713] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11502), 1,
      anon_sym_RBRACK,
    STATE(7337), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242727] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11504), 1,
      anon_sym_RBRACE,
    STATE(7338), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242741] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11506), 1,
      anon_sym_in,
    STATE(7339), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242755] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11508), 1,
      anon_sym_RPAREN,
    STATE(7340), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242769] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11510), 1,
      anon_sym_PIPE_RBRACK,
    STATE(7341), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242783] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11512), 1,
      anon_sym_RPAREN,
    STATE(7342), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242797] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11514), 1,
      anon_sym_DOT,
    STATE(7343), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242811] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11516), 1,
      anon_sym_STAR,
    STATE(7344), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242825] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11518), 1,
      anon_sym_DASH_GT,
    STATE(7345), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242839] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11520), 1,
      anon_sym_RPAREN,
    STATE(7346), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242853] = 4,
    ACTIONS(23), 1,
      aux_sym_attribute_token1,
    ACTIONS(11522), 1,
      anon_sym_with,
    STATE(7347), 1,
      sym_attribute,
    ACTIONS(5), 2,
      sym_comment,
      sym_line_number_directive,
  [242867] = 1,
    ACTIONS(11524), 1,
      ts_builtin_sym_end,
  [242871] = 1,
    ACTIONS(11526), 1,
      ts_builtin_sym_end,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(1386)] = 0,
  [SMALL_STATE(1387)] = 132,
  [SMALL_STATE(1388)] = 208,
  [SMALL_STATE(1389)] = 346,
  [SMALL_STATE(1390)] = 484,
  [SMALL_STATE(1391)] = 560,
  [SMALL_STATE(1392)] = 698,
  [SMALL_STATE(1393)] = 836,
  [SMALL_STATE(1394)] = 974,
  [SMALL_STATE(1395)] = 1112,
  [SMALL_STATE(1396)] = 1250,
  [SMALL_STATE(1397)] = 1388,
  [SMALL_STATE(1398)] = 1526,
  [SMALL_STATE(1399)] = 1664,
  [SMALL_STATE(1400)] = 1796,
  [SMALL_STATE(1401)] = 1934,
  [SMALL_STATE(1402)] = 2072,
  [SMALL_STATE(1403)] = 2204,
  [SMALL_STATE(1404)] = 2336,
  [SMALL_STATE(1405)] = 2468,
  [SMALL_STATE(1406)] = 2600,
  [SMALL_STATE(1407)] = 2732,
  [SMALL_STATE(1408)] = 2864,
  [SMALL_STATE(1409)] = 2996,
  [SMALL_STATE(1410)] = 3128,
  [SMALL_STATE(1411)] = 3260,
  [SMALL_STATE(1412)] = 3398,
  [SMALL_STATE(1413)] = 3533,
  [SMALL_STATE(1414)] = 3668,
  [SMALL_STATE(1415)] = 3803,
  [SMALL_STATE(1416)] = 3938,
  [SMALL_STATE(1417)] = 4073,
  [SMALL_STATE(1418)] = 4208,
  [SMALL_STATE(1419)] = 4343,
  [SMALL_STATE(1420)] = 4478,
  [SMALL_STATE(1421)] = 4613,
  [SMALL_STATE(1422)] = 4748,
  [SMALL_STATE(1423)] = 4883,
  [SMALL_STATE(1424)] = 5018,
  [SMALL_STATE(1425)] = 5152,
  [SMALL_STATE(1426)] = 5278,
  [SMALL_STATE(1427)] = 5404,
  [SMALL_STATE(1428)] = 5538,
  [SMALL_STATE(1429)] = 5672,
  [SMALL_STATE(1430)] = 5798,
  [SMALL_STATE(1431)] = 5932,
  [SMALL_STATE(1432)] = 6066,
  [SMALL_STATE(1433)] = 6192,
  [SMALL_STATE(1434)] = 6318,
  [SMALL_STATE(1435)] = 6444,
  [SMALL_STATE(1436)] = 6570,
  [SMALL_STATE(1437)] = 6696,
  [SMALL_STATE(1438)] = 6822,
  [SMALL_STATE(1439)] = 6948,
  [SMALL_STATE(1440)] = 7074,
  [SMALL_STATE(1441)] = 7200,
  [SMALL_STATE(1442)] = 7334,
  [SMALL_STATE(1443)] = 7468,
  [SMALL_STATE(1444)] = 7594,
  [SMALL_STATE(1445)] = 7728,
  [SMALL_STATE(1446)] = 7854,
  [SMALL_STATE(1447)] = 7980,
  [SMALL_STATE(1448)] = 8114,
  [SMALL_STATE(1449)] = 8240,
  [SMALL_STATE(1450)] = 8366,
  [SMALL_STATE(1451)] = 8500,
  [SMALL_STATE(1452)] = 8626,
  [SMALL_STATE(1453)] = 8749,
  [SMALL_STATE(1454)] = 8872,
  [SMALL_STATE(1455)] = 8995,
  [SMALL_STATE(1456)] = 9118,
  [SMALL_STATE(1457)] = 9241,
  [SMALL_STATE(1458)] = 9364,
  [SMALL_STATE(1459)] = 9487,
  [SMALL_STATE(1460)] = 9610,
  [SMALL_STATE(1461)] = 9733,
  [SMALL_STATE(1462)] = 9856,
  [SMALL_STATE(1463)] = 9979,
  [SMALL_STATE(1464)] = 10099,
  [SMALL_STATE(1465)] = 10219,
  [SMALL_STATE(1466)] = 10339,
  [SMALL_STATE(1467)] = 10459,
  [SMALL_STATE(1468)] = 10579,
  [SMALL_STATE(1469)] = 10699,
  [SMALL_STATE(1470)] = 10819,
  [SMALL_STATE(1471)] = 10939,
  [SMALL_STATE(1472)] = 11059,
  [SMALL_STATE(1473)] = 11179,
  [SMALL_STATE(1474)] = 11299,
  [SMALL_STATE(1475)] = 11369,
  [SMALL_STATE(1476)] = 11489,
  [SMALL_STATE(1477)] = 11609,
  [SMALL_STATE(1478)] = 11729,
  [SMALL_STATE(1479)] = 11849,
  [SMALL_STATE(1480)] = 11969,
  [SMALL_STATE(1481)] = 12039,
  [SMALL_STATE(1482)] = 12159,
  [SMALL_STATE(1483)] = 12279,
  [SMALL_STATE(1484)] = 12399,
  [SMALL_STATE(1485)] = 12519,
  [SMALL_STATE(1486)] = 12639,
  [SMALL_STATE(1487)] = 12709,
  [SMALL_STATE(1488)] = 12829,
  [SMALL_STATE(1489)] = 12949,
  [SMALL_STATE(1490)] = 13069,
  [SMALL_STATE(1491)] = 13189,
  [SMALL_STATE(1492)] = 13309,
  [SMALL_STATE(1493)] = 13429,
  [SMALL_STATE(1494)] = 13549,
  [SMALL_STATE(1495)] = 13669,
  [SMALL_STATE(1496)] = 13789,
  [SMALL_STATE(1497)] = 13909,
  [SMALL_STATE(1498)] = 14029,
  [SMALL_STATE(1499)] = 14149,
  [SMALL_STATE(1500)] = 14269,
  [SMALL_STATE(1501)] = 14389,
  [SMALL_STATE(1502)] = 14509,
  [SMALL_STATE(1503)] = 14629,
  [SMALL_STATE(1504)] = 14749,
  [SMALL_STATE(1505)] = 14869,
  [SMALL_STATE(1506)] = 14989,
  [SMALL_STATE(1507)] = 15109,
  [SMALL_STATE(1508)] = 15178,
  [SMALL_STATE(1509)] = 15247,
  [SMALL_STATE(1510)] = 15374,
  [SMALL_STATE(1511)] = 15445,
  [SMALL_STATE(1512)] = 15572,
  [SMALL_STATE(1513)] = 15641,
  [SMALL_STATE(1514)] = 15710,
  [SMALL_STATE(1515)] = 15779,
  [SMALL_STATE(1516)] = 15852,
  [SMALL_STATE(1517)] = 15921,
  [SMALL_STATE(1518)] = 15992,
  [SMALL_STATE(1519)] = 16061,
  [SMALL_STATE(1520)] = 16188,
  [SMALL_STATE(1521)] = 16259,
  [SMALL_STATE(1522)] = 16332,
  [SMALL_STATE(1523)] = 16401,
  [SMALL_STATE(1524)] = 16470,
  [SMALL_STATE(1525)] = 16538,
  [SMALL_STATE(1526)] = 16606,
  [SMALL_STATE(1527)] = 16674,
  [SMALL_STATE(1528)] = 16744,
  [SMALL_STATE(1529)] = 16812,
  [SMALL_STATE(1530)] = 16880,
  [SMALL_STATE(1531)] = 16954,
  [SMALL_STATE(1532)] = 17022,
  [SMALL_STATE(1533)] = 17090,
  [SMALL_STATE(1534)] = 17158,
  [SMALL_STATE(1535)] = 17226,
  [SMALL_STATE(1536)] = 17294,
  [SMALL_STATE(1537)] = 17362,
  [SMALL_STATE(1538)] = 17430,
  [SMALL_STATE(1539)] = 17498,
  [SMALL_STATE(1540)] = 17566,
  [SMALL_STATE(1541)] = 17634,
  [SMALL_STATE(1542)] = 17708,
  [SMALL_STATE(1543)] = 17776,
  [SMALL_STATE(1544)] = 17844,
  [SMALL_STATE(1545)] = 17912,
  [SMALL_STATE(1546)] = 17980,
  [SMALL_STATE(1547)] = 18048,
  [SMALL_STATE(1548)] = 18116,
  [SMALL_STATE(1549)] = 18184,
  [SMALL_STATE(1550)] = 18252,
  [SMALL_STATE(1551)] = 18320,
  [SMALL_STATE(1552)] = 18388,
  [SMALL_STATE(1553)] = 18456,
  [SMALL_STATE(1554)] = 18524,
  [SMALL_STATE(1555)] = 18592,
  [SMALL_STATE(1556)] = 18660,
  [SMALL_STATE(1557)] = 18728,
  [SMALL_STATE(1558)] = 18796,
  [SMALL_STATE(1559)] = 18866,
  [SMALL_STATE(1560)] = 18936,
  [SMALL_STATE(1561)] = 19004,
  [SMALL_STATE(1562)] = 19072,
  [SMALL_STATE(1563)] = 19140,
  [SMALL_STATE(1564)] = 19208,
  [SMALL_STATE(1565)] = 19276,
  [SMALL_STATE(1566)] = 19344,
  [SMALL_STATE(1567)] = 19412,
  [SMALL_STATE(1568)] = 19480,
  [SMALL_STATE(1569)] = 19548,
  [SMALL_STATE(1570)] = 19616,
  [SMALL_STATE(1571)] = 19684,
  [SMALL_STATE(1572)] = 19752,
  [SMALL_STATE(1573)] = 19822,
  [SMALL_STATE(1574)] = 19890,
  [SMALL_STATE(1575)] = 19958,
  [SMALL_STATE(1576)] = 20026,
  [SMALL_STATE(1577)] = 20094,
  [SMALL_STATE(1578)] = 20164,
  [SMALL_STATE(1579)] = 20232,
  [SMALL_STATE(1580)] = 20300,
  [SMALL_STATE(1581)] = 20368,
  [SMALL_STATE(1582)] = 20436,
  [SMALL_STATE(1583)] = 20504,
  [SMALL_STATE(1584)] = 20572,
  [SMALL_STATE(1585)] = 20640,
  [SMALL_STATE(1586)] = 20708,
  [SMALL_STATE(1587)] = 20776,
  [SMALL_STATE(1588)] = 20844,
  [SMALL_STATE(1589)] = 20912,
  [SMALL_STATE(1590)] = 20980,
  [SMALL_STATE(1591)] = 21048,
  [SMALL_STATE(1592)] = 21116,
  [SMALL_STATE(1593)] = 21184,
  [SMALL_STATE(1594)] = 21252,
  [SMALL_STATE(1595)] = 21320,
  [SMALL_STATE(1596)] = 21388,
  [SMALL_STATE(1597)] = 21456,
  [SMALL_STATE(1598)] = 21524,
  [SMALL_STATE(1599)] = 21592,
  [SMALL_STATE(1600)] = 21713,
  [SMALL_STATE(1601)] = 21834,
  [SMALL_STATE(1602)] = 21955,
  [SMALL_STATE(1603)] = 22076,
  [SMALL_STATE(1604)] = 22197,
  [SMALL_STATE(1605)] = 22318,
  [SMALL_STATE(1606)] = 22439,
  [SMALL_STATE(1607)] = 22560,
  [SMALL_STATE(1608)] = 22681,
  [SMALL_STATE(1609)] = 22802,
  [SMALL_STATE(1610)] = 22923,
  [SMALL_STATE(1611)] = 23044,
  [SMALL_STATE(1612)] = 23165,
  [SMALL_STATE(1613)] = 23286,
  [SMALL_STATE(1614)] = 23407,
  [SMALL_STATE(1615)] = 23528,
  [SMALL_STATE(1616)] = 23649,
  [SMALL_STATE(1617)] = 23770,
  [SMALL_STATE(1618)] = 23891,
  [SMALL_STATE(1619)] = 24012,
  [SMALL_STATE(1620)] = 24133,
  [SMALL_STATE(1621)] = 24202,
  [SMALL_STATE(1622)] = 24323,
  [SMALL_STATE(1623)] = 24444,
  [SMALL_STATE(1624)] = 24565,
  [SMALL_STATE(1625)] = 24686,
  [SMALL_STATE(1626)] = 24807,
  [SMALL_STATE(1627)] = 24875,
  [SMALL_STATE(1628)] = 24943,
  [SMALL_STATE(1629)] = 25061,
  [SMALL_STATE(1630)] = 25179,
  [SMALL_STATE(1631)] = 25297,
  [SMALL_STATE(1632)] = 25415,
  [SMALL_STATE(1633)] = 25533,
  [SMALL_STATE(1634)] = 25651,
  [SMALL_STATE(1635)] = 25769,
  [SMALL_STATE(1636)] = 25887,
  [SMALL_STATE(1637)] = 26005,
  [SMALL_STATE(1638)] = 26123,
  [SMALL_STATE(1639)] = 26191,
  [SMALL_STATE(1640)] = 26309,
  [SMALL_STATE(1641)] = 26377,
  [SMALL_STATE(1642)] = 26495,
  [SMALL_STATE(1643)] = 26613,
  [SMALL_STATE(1644)] = 26678,
  [SMALL_STATE(1645)] = 26743,
  [SMALL_STATE(1646)] = 26810,
  [SMALL_STATE(1647)] = 26875,
  [SMALL_STATE(1648)] = 26974,
  [SMALL_STATE(1649)] = 27075,
  [SMALL_STATE(1650)] = 27190,
  [SMALL_STATE(1651)] = 27257,
  [SMALL_STATE(1652)] = 27334,
  [SMALL_STATE(1653)] = 27433,
  [SMALL_STATE(1654)] = 27514,
  [SMALL_STATE(1655)] = 27579,
  [SMALL_STATE(1656)] = 27694,
  [SMALL_STATE(1657)] = 27809,
  [SMALL_STATE(1658)] = 27884,
  [SMALL_STATE(1659)] = 27951,
  [SMALL_STATE(1660)] = 28052,
  [SMALL_STATE(1661)] = 28149,
  [SMALL_STATE(1662)] = 28234,
  [SMALL_STATE(1663)] = 28299,
  [SMALL_STATE(1664)] = 28414,
  [SMALL_STATE(1665)] = 28501,
  [SMALL_STATE(1666)] = 28578,
  [SMALL_STATE(1667)] = 28665,
  [SMALL_STATE(1668)] = 28764,
  [SMALL_STATE(1669)] = 28855,
  [SMALL_STATE(1670)] = 28956,
  [SMALL_STATE(1671)] = 29055,
  [SMALL_STATE(1672)] = 29150,
  [SMALL_STATE(1673)] = 29262,
  [SMALL_STATE(1674)] = 29374,
  [SMALL_STATE(1675)] = 29450,
  [SMALL_STATE(1676)] = 29526,
  [SMALL_STATE(1677)] = 29606,
  [SMALL_STATE(1678)] = 29670,
  [SMALL_STATE(1679)] = 29782,
  [SMALL_STATE(1680)] = 29846,
  [SMALL_STATE(1681)] = 29910,
  [SMALL_STATE(1682)] = 30026,
  [SMALL_STATE(1683)] = 30090,
  [SMALL_STATE(1684)] = 30154,
  [SMALL_STATE(1685)] = 30218,
  [SMALL_STATE(1686)] = 30284,
  [SMALL_STATE(1687)] = 30348,
  [SMALL_STATE(1688)] = 30462,
  [SMALL_STATE(1689)] = 30574,
  [SMALL_STATE(1690)] = 30638,
  [SMALL_STATE(1691)] = 30702,
  [SMALL_STATE(1692)] = 30818,
  [SMALL_STATE(1693)] = 30882,
  [SMALL_STATE(1694)] = 30946,
  [SMALL_STATE(1695)] = 31010,
  [SMALL_STATE(1696)] = 31124,
  [SMALL_STATE(1697)] = 31224,
  [SMALL_STATE(1698)] = 31288,
  [SMALL_STATE(1699)] = 31352,
  [SMALL_STATE(1700)] = 31416,
  [SMALL_STATE(1701)] = 31480,
  [SMALL_STATE(1702)] = 31544,
  [SMALL_STATE(1703)] = 31608,
  [SMALL_STATE(1704)] = 31672,
  [SMALL_STATE(1705)] = 31736,
  [SMALL_STATE(1706)] = 31848,
  [SMALL_STATE(1707)] = 31964,
  [SMALL_STATE(1708)] = 32064,
  [SMALL_STATE(1709)] = 32128,
  [SMALL_STATE(1710)] = 32192,
  [SMALL_STATE(1711)] = 32256,
  [SMALL_STATE(1712)] = 32320,
  [SMALL_STATE(1713)] = 32432,
  [SMALL_STATE(1714)] = 32548,
  [SMALL_STATE(1715)] = 32612,
  [SMALL_STATE(1716)] = 32676,
  [SMALL_STATE(1717)] = 32740,
  [SMALL_STATE(1718)] = 32852,
  [SMALL_STATE(1719)] = 32916,
  [SMALL_STATE(1720)] = 33032,
  [SMALL_STATE(1721)] = 33144,
  [SMALL_STATE(1722)] = 33240,
  [SMALL_STATE(1723)] = 33324,
  [SMALL_STATE(1724)] = 33388,
  [SMALL_STATE(1725)] = 33452,
  [SMALL_STATE(1726)] = 33564,
  [SMALL_STATE(1727)] = 33650,
  [SMALL_STATE(1728)] = 33714,
  [SMALL_STATE(1729)] = 33812,
  [SMALL_STATE(1730)] = 33924,
  [SMALL_STATE(1731)] = 33988,
  [SMALL_STATE(1732)] = 34078,
  [SMALL_STATE(1733)] = 34172,
  [SMALL_STATE(1734)] = 34284,
  [SMALL_STATE(1735)] = 34382,
  [SMALL_STATE(1736)] = 34480,
  [SMALL_STATE(1737)] = 34544,
  [SMALL_STATE(1738)] = 34608,
  [SMALL_STATE(1739)] = 34720,
  [SMALL_STATE(1740)] = 34818,
  [SMALL_STATE(1741)] = 34882,
  [SMALL_STATE(1742)] = 34994,
  [SMALL_STATE(1743)] = 35058,
  [SMALL_STATE(1744)] = 35170,
  [SMALL_STATE(1745)] = 35234,
  [SMALL_STATE(1746)] = 35298,
  [SMALL_STATE(1747)] = 35362,
  [SMALL_STATE(1748)] = 35426,
  [SMALL_STATE(1749)] = 35538,
  [SMALL_STATE(1750)] = 35602,
  [SMALL_STATE(1751)] = 35666,
  [SMALL_STATE(1752)] = 35732,
  [SMALL_STATE(1753)] = 35796,
  [SMALL_STATE(1754)] = 35860,
  [SMALL_STATE(1755)] = 35924,
  [SMALL_STATE(1756)] = 35988,
  [SMALL_STATE(1757)] = 36052,
  [SMALL_STATE(1758)] = 36116,
  [SMALL_STATE(1759)] = 36180,
  [SMALL_STATE(1760)] = 36244,
  [SMALL_STATE(1761)] = 36308,
  [SMALL_STATE(1762)] = 36372,
  [SMALL_STATE(1763)] = 36436,
  [SMALL_STATE(1764)] = 36548,
  [SMALL_STATE(1765)] = 36648,
  [SMALL_STATE(1766)] = 36712,
  [SMALL_STATE(1767)] = 36776,
  [SMALL_STATE(1768)] = 36888,
  [SMALL_STATE(1769)] = 36952,
  [SMALL_STATE(1770)] = 37066,
  [SMALL_STATE(1771)] = 37130,
  [SMALL_STATE(1772)] = 37204,
  [SMALL_STATE(1773)] = 37290,
  [SMALL_STATE(1774)] = 37353,
  [SMALL_STATE(1775)] = 37416,
  [SMALL_STATE(1776)] = 37479,
  [SMALL_STATE(1777)] = 37542,
  [SMALL_STATE(1778)] = 37605,
  [SMALL_STATE(1779)] = 37680,
  [SMALL_STATE(1780)] = 37755,
  [SMALL_STATE(1781)] = 37834,
  [SMALL_STATE(1782)] = 37897,
  [SMALL_STATE(1783)] = 37960,
  [SMALL_STATE(1784)] = 38025,
  [SMALL_STATE(1785)] = 38088,
  [SMALL_STATE(1786)] = 38151,
  [SMALL_STATE(1787)] = 38214,
  [SMALL_STATE(1788)] = 38277,
  [SMALL_STATE(1789)] = 38340,
  [SMALL_STATE(1790)] = 38403,
  [SMALL_STATE(1791)] = 38466,
  [SMALL_STATE(1792)] = 38529,
  [SMALL_STATE(1793)] = 38592,
  [SMALL_STATE(1794)] = 38691,
  [SMALL_STATE(1795)] = 38754,
  [SMALL_STATE(1796)] = 38817,
  [SMALL_STATE(1797)] = 38880,
  [SMALL_STATE(1798)] = 38943,
  [SMALL_STATE(1799)] = 39006,
  [SMALL_STATE(1800)] = 39069,
  [SMALL_STATE(1801)] = 39132,
  [SMALL_STATE(1802)] = 39195,
  [SMALL_STATE(1803)] = 39258,
  [SMALL_STATE(1804)] = 39321,
  [SMALL_STATE(1805)] = 39384,
  [SMALL_STATE(1806)] = 39447,
  [SMALL_STATE(1807)] = 39510,
  [SMALL_STATE(1808)] = 39573,
  [SMALL_STATE(1809)] = 39636,
  [SMALL_STATE(1810)] = 39699,
  [SMALL_STATE(1811)] = 39762,
  [SMALL_STATE(1812)] = 39825,
  [SMALL_STATE(1813)] = 39894,
  [SMALL_STATE(1814)] = 39963,
  [SMALL_STATE(1815)] = 40026,
  [SMALL_STATE(1816)] = 40089,
  [SMALL_STATE(1817)] = 40152,
  [SMALL_STATE(1818)] = 40215,
  [SMALL_STATE(1819)] = 40278,
  [SMALL_STATE(1820)] = 40347,
  [SMALL_STATE(1821)] = 40416,
  [SMALL_STATE(1822)] = 40479,
  [SMALL_STATE(1823)] = 40542,
  [SMALL_STATE(1824)] = 40605,
  [SMALL_STATE(1825)] = 40704,
  [SMALL_STATE(1826)] = 40769,
  [SMALL_STATE(1827)] = 40834,
  [SMALL_STATE(1828)] = 40897,
  [SMALL_STATE(1829)] = 40996,
  [SMALL_STATE(1830)] = 41059,
  [SMALL_STATE(1831)] = 41122,
  [SMALL_STATE(1832)] = 41185,
  [SMALL_STATE(1833)] = 41250,
  [SMALL_STATE(1834)] = 41313,
  [SMALL_STATE(1835)] = 41376,
  [SMALL_STATE(1836)] = 41441,
  [SMALL_STATE(1837)] = 41504,
  [SMALL_STATE(1838)] = 41567,
  [SMALL_STATE(1839)] = 41630,
  [SMALL_STATE(1840)] = 41693,
  [SMALL_STATE(1841)] = 41762,
  [SMALL_STATE(1842)] = 41831,
  [SMALL_STATE(1843)] = 41894,
  [SMALL_STATE(1844)] = 41957,
  [SMALL_STATE(1845)] = 42020,
  [SMALL_STATE(1846)] = 42085,
  [SMALL_STATE(1847)] = 42180,
  [SMALL_STATE(1848)] = 42263,
  [SMALL_STATE(1849)] = 42362,
  [SMALL_STATE(1850)] = 42447,
  [SMALL_STATE(1851)] = 42532,
  [SMALL_STATE(1852)] = 42629,
  [SMALL_STATE(1853)] = 42718,
  [SMALL_STATE(1854)] = 42781,
  [SMALL_STATE(1855)] = 42878,
  [SMALL_STATE(1856)] = 42941,
  [SMALL_STATE(1857)] = 43006,
  [SMALL_STATE(1858)] = 43069,
  [SMALL_STATE(1859)] = 43166,
  [SMALL_STATE(1860)] = 43229,
  [SMALL_STATE(1861)] = 43292,
  [SMALL_STATE(1862)] = 43355,
  [SMALL_STATE(1863)] = 43418,
  [SMALL_STATE(1864)] = 43481,
  [SMALL_STATE(1865)] = 43578,
  [SMALL_STATE(1866)] = 43641,
  [SMALL_STATE(1867)] = 43704,
  [SMALL_STATE(1868)] = 43767,
  [SMALL_STATE(1869)] = 43830,
  [SMALL_STATE(1870)] = 43893,
  [SMALL_STATE(1871)] = 43956,
  [SMALL_STATE(1872)] = 44019,
  [SMALL_STATE(1873)] = 44088,
  [SMALL_STATE(1874)] = 44151,
  [SMALL_STATE(1875)] = 44214,
  [SMALL_STATE(1876)] = 44277,
  [SMALL_STATE(1877)] = 44340,
  [SMALL_STATE(1878)] = 44403,
  [SMALL_STATE(1879)] = 44466,
  [SMALL_STATE(1880)] = 44529,
  [SMALL_STATE(1881)] = 44592,
  [SMALL_STATE(1882)] = 44655,
  [SMALL_STATE(1883)] = 44718,
  [SMALL_STATE(1884)] = 44781,
  [SMALL_STATE(1885)] = 44850,
  [SMALL_STATE(1886)] = 44913,
  [SMALL_STATE(1887)] = 44976,
  [SMALL_STATE(1888)] = 45039,
  [SMALL_STATE(1889)] = 45112,
  [SMALL_STATE(1890)] = 45175,
  [SMALL_STATE(1891)] = 45238,
  [SMALL_STATE(1892)] = 45301,
  [SMALL_STATE(1893)] = 45370,
  [SMALL_STATE(1894)] = 45439,
  [SMALL_STATE(1895)] = 45538,
  [SMALL_STATE(1896)] = 45637,
  [SMALL_STATE(1897)] = 45700,
  [SMALL_STATE(1898)] = 45763,
  [SMALL_STATE(1899)] = 45826,
  [SMALL_STATE(1900)] = 45889,
  [SMALL_STATE(1901)] = 45952,
  [SMALL_STATE(1902)] = 46015,
  [SMALL_STATE(1903)] = 46078,
  [SMALL_STATE(1904)] = 46141,
  [SMALL_STATE(1905)] = 46204,
  [SMALL_STATE(1906)] = 46267,
  [SMALL_STATE(1907)] = 46330,
  [SMALL_STATE(1908)] = 46393,
  [SMALL_STATE(1909)] = 46456,
  [SMALL_STATE(1910)] = 46519,
  [SMALL_STATE(1911)] = 46582,
  [SMALL_STATE(1912)] = 46645,
  [SMALL_STATE(1913)] = 46708,
  [SMALL_STATE(1914)] = 46771,
  [SMALL_STATE(1915)] = 46836,
  [SMALL_STATE(1916)] = 46899,
  [SMALL_STATE(1917)] = 46962,
  [SMALL_STATE(1918)] = 47025,
  [SMALL_STATE(1919)] = 47088,
  [SMALL_STATE(1920)] = 47181,
  [SMALL_STATE(1921)] = 47243,
  [SMALL_STATE(1922)] = 47305,
  [SMALL_STATE(1923)] = 47367,
  [SMALL_STATE(1924)] = 47463,
  [SMALL_STATE(1925)] = 47525,
  [SMALL_STATE(1926)] = 47587,
  [SMALL_STATE(1927)] = 47649,
  [SMALL_STATE(1928)] = 47719,
  [SMALL_STATE(1929)] = 47781,
  [SMALL_STATE(1930)] = 47847,
  [SMALL_STATE(1931)] = 47911,
  [SMALL_STATE(1932)] = 47973,
  [SMALL_STATE(1933)] = 48041,
  [SMALL_STATE(1934)] = 48105,
  [SMALL_STATE(1935)] = 48167,
  [SMALL_STATE(1936)] = 48229,
  [SMALL_STATE(1937)] = 48323,
  [SMALL_STATE(1938)] = 48385,
  [SMALL_STATE(1939)] = 48447,
  [SMALL_STATE(1940)] = 48509,
  [SMALL_STATE(1941)] = 48571,
  [SMALL_STATE(1942)] = 48669,
  [SMALL_STATE(1943)] = 48731,
  [SMALL_STATE(1944)] = 48793,
  [SMALL_STATE(1945)] = 48855,
  [SMALL_STATE(1946)] = 48917,
  [SMALL_STATE(1947)] = 48979,
  [SMALL_STATE(1948)] = 49041,
  [SMALL_STATE(1949)] = 49103,
  [SMALL_STATE(1950)] = 49165,
  [SMALL_STATE(1951)] = 49227,
  [SMALL_STATE(1952)] = 49289,
  [SMALL_STATE(1953)] = 49351,
  [SMALL_STATE(1954)] = 49413,
  [SMALL_STATE(1955)] = 49475,
  [SMALL_STATE(1956)] = 49537,
  [SMALL_STATE(1957)] = 49599,
  [SMALL_STATE(1958)] = 49661,
  [SMALL_STATE(1959)] = 49723,
  [SMALL_STATE(1960)] = 49791,
  [SMALL_STATE(1961)] = 49853,
  [SMALL_STATE(1962)] = 49915,
  [SMALL_STATE(1963)] = 49977,
  [SMALL_STATE(1964)] = 50039,
  [SMALL_STATE(1965)] = 50101,
  [SMALL_STATE(1966)] = 50163,
  [SMALL_STATE(1967)] = 50225,
  [SMALL_STATE(1968)] = 50287,
  [SMALL_STATE(1969)] = 50349,
  [SMALL_STATE(1970)] = 50411,
  [SMALL_STATE(1971)] = 50473,
  [SMALL_STATE(1972)] = 50535,
  [SMALL_STATE(1973)] = 50597,
  [SMALL_STATE(1974)] = 50659,
  [SMALL_STATE(1975)] = 50721,
  [SMALL_STATE(1976)] = 50783,
  [SMALL_STATE(1977)] = 50847,
  [SMALL_STATE(1978)] = 50909,
  [SMALL_STATE(1979)] = 50971,
  [SMALL_STATE(1980)] = 51035,
  [SMALL_STATE(1981)] = 51097,
  [SMALL_STATE(1982)] = 51159,
  [SMALL_STATE(1983)] = 51221,
  [SMALL_STATE(1984)] = 51283,
  [SMALL_STATE(1985)] = 51345,
  [SMALL_STATE(1986)] = 51407,
  [SMALL_STATE(1987)] = 51469,
  [SMALL_STATE(1988)] = 51531,
  [SMALL_STATE(1989)] = 51593,
  [SMALL_STATE(1990)] = 51655,
  [SMALL_STATE(1991)] = 51717,
  [SMALL_STATE(1992)] = 51779,
  [SMALL_STATE(1993)] = 51841,
  [SMALL_STATE(1994)] = 51939,
  [SMALL_STATE(1995)] = 52001,
  [SMALL_STATE(1996)] = 52063,
  [SMALL_STATE(1997)] = 52125,
  [SMALL_STATE(1998)] = 52187,
  [SMALL_STATE(1999)] = 52249,
  [SMALL_STATE(2000)] = 52311,
  [SMALL_STATE(2001)] = 52375,
  [SMALL_STATE(2002)] = 52437,
  [SMALL_STATE(2003)] = 52535,
  [SMALL_STATE(2004)] = 52597,
  [SMALL_STATE(2005)] = 52693,
  [SMALL_STATE(2006)] = 52755,
  [SMALL_STATE(2007)] = 52817,
  [SMALL_STATE(2008)] = 52899,
  [SMALL_STATE(2009)] = 52961,
  [SMALL_STATE(2010)] = 53027,
  [SMALL_STATE(2011)] = 53111,
  [SMALL_STATE(2012)] = 53173,
  [SMALL_STATE(2013)] = 53235,
  [SMALL_STATE(2014)] = 53297,
  [SMALL_STATE(2015)] = 53381,
  [SMALL_STATE(2016)] = 53477,
  [SMALL_STATE(2017)] = 53551,
  [SMALL_STATE(2018)] = 53625,
  [SMALL_STATE(2019)] = 53703,
  [SMALL_STATE(2020)] = 53791,
  [SMALL_STATE(2021)] = 53853,
  [SMALL_STATE(2022)] = 53915,
  [SMALL_STATE(2023)] = 54013,
  [SMALL_STATE(2024)] = 54075,
  [SMALL_STATE(2025)] = 54145,
  [SMALL_STATE(2026)] = 54207,
  [SMALL_STATE(2027)] = 54269,
  [SMALL_STATE(2028)] = 54333,
  [SMALL_STATE(2029)] = 54431,
  [SMALL_STATE(2030)] = 54499,
  [SMALL_STATE(2031)] = 54567,
  [SMALL_STATE(2032)] = 54629,
  [SMALL_STATE(2033)] = 54691,
  [SMALL_STATE(2034)] = 54753,
  [SMALL_STATE(2035)] = 54815,
  [SMALL_STATE(2036)] = 54877,
  [SMALL_STATE(2037)] = 54939,
  [SMALL_STATE(2038)] = 55001,
  [SMALL_STATE(2039)] = 55063,
  [SMALL_STATE(2040)] = 55125,
  [SMALL_STATE(2041)] = 55187,
  [SMALL_STATE(2042)] = 55249,
  [SMALL_STATE(2043)] = 55317,
  [SMALL_STATE(2044)] = 55379,
  [SMALL_STATE(2045)] = 55443,
  [SMALL_STATE(2046)] = 55535,
  [SMALL_STATE(2047)] = 55597,
  [SMALL_STATE(2048)] = 55693,
  [SMALL_STATE(2049)] = 55755,
  [SMALL_STATE(2050)] = 55817,
  [SMALL_STATE(2051)] = 55879,
  [SMALL_STATE(2052)] = 55941,
  [SMALL_STATE(2053)] = 56009,
  [SMALL_STATE(2054)] = 56071,
  [SMALL_STATE(2055)] = 56133,
  [SMALL_STATE(2056)] = 56195,
  [SMALL_STATE(2057)] = 56257,
  [SMALL_STATE(2058)] = 56319,
  [SMALL_STATE(2059)] = 56381,
  [SMALL_STATE(2060)] = 56443,
  [SMALL_STATE(2061)] = 56505,
  [SMALL_STATE(2062)] = 56567,
  [SMALL_STATE(2063)] = 56629,
  [SMALL_STATE(2064)] = 56691,
  [SMALL_STATE(2065)] = 56753,
  [SMALL_STATE(2066)] = 56815,
  [SMALL_STATE(2067)] = 56877,
  [SMALL_STATE(2068)] = 56939,
  [SMALL_STATE(2069)] = 57001,
  [SMALL_STATE(2070)] = 57063,
  [SMALL_STATE(2071)] = 57125,
  [SMALL_STATE(2072)] = 57187,
  [SMALL_STATE(2073)] = 57249,
  [SMALL_STATE(2074)] = 57311,
  [SMALL_STATE(2075)] = 57383,
  [SMALL_STATE(2076)] = 57445,
  [SMALL_STATE(2077)] = 57507,
  [SMALL_STATE(2078)] = 57569,
  [SMALL_STATE(2079)] = 57631,
  [SMALL_STATE(2080)] = 57693,
  [SMALL_STATE(2081)] = 57755,
  [SMALL_STATE(2082)] = 57817,
  [SMALL_STATE(2083)] = 57879,
  [SMALL_STATE(2084)] = 57941,
  [SMALL_STATE(2085)] = 58003,
  [SMALL_STATE(2086)] = 58065,
  [SMALL_STATE(2087)] = 58127,
  [SMALL_STATE(2088)] = 58189,
  [SMALL_STATE(2089)] = 58287,
  [SMALL_STATE(2090)] = 58349,
  [SMALL_STATE(2091)] = 58411,
  [SMALL_STATE(2092)] = 58473,
  [SMALL_STATE(2093)] = 58535,
  [SMALL_STATE(2094)] = 58603,
  [SMALL_STATE(2095)] = 58664,
  [SMALL_STATE(2096)] = 58729,
  [SMALL_STATE(2097)] = 58846,
  [SMALL_STATE(2098)] = 58907,
  [SMALL_STATE(2099)] = 58972,
  [SMALL_STATE(2100)] = 59035,
  [SMALL_STATE(2101)] = 59098,
  [SMALL_STATE(2102)] = 59215,
  [SMALL_STATE(2103)] = 59332,
  [SMALL_STATE(2104)] = 59393,
  [SMALL_STATE(2105)] = 59454,
  [SMALL_STATE(2106)] = 59519,
  [SMALL_STATE(2107)] = 59636,
  [SMALL_STATE(2108)] = 59697,
  [SMALL_STATE(2109)] = 59762,
  [SMALL_STATE(2110)] = 59827,
  [SMALL_STATE(2111)] = 59944,
  [SMALL_STATE(2112)] = 60061,
  [SMALL_STATE(2113)] = 60124,
  [SMALL_STATE(2114)] = 60189,
  [SMALL_STATE(2115)] = 60250,
  [SMALL_STATE(2116)] = 60310,
  [SMALL_STATE(2117)] = 60370,
  [SMALL_STATE(2118)] = 60430,
  [SMALL_STATE(2119)] = 60494,
  [SMALL_STATE(2120)] = 60554,
  [SMALL_STATE(2121)] = 60650,
  [SMALL_STATE(2122)] = 60710,
  [SMALL_STATE(2123)] = 60770,
  [SMALL_STATE(2124)] = 60830,
  [SMALL_STATE(2125)] = 60890,
  [SMALL_STATE(2126)] = 60950,
  [SMALL_STATE(2127)] = 61010,
  [SMALL_STATE(2128)] = 61070,
  [SMALL_STATE(2129)] = 61130,
  [SMALL_STATE(2130)] = 61190,
  [SMALL_STATE(2131)] = 61250,
  [SMALL_STATE(2132)] = 61310,
  [SMALL_STATE(2133)] = 61370,
  [SMALL_STATE(2134)] = 61430,
  [SMALL_STATE(2135)] = 61490,
  [SMALL_STATE(2136)] = 61550,
  [SMALL_STATE(2137)] = 61610,
  [SMALL_STATE(2138)] = 61672,
  [SMALL_STATE(2139)] = 61732,
  [SMALL_STATE(2140)] = 61792,
  [SMALL_STATE(2141)] = 61852,
  [SMALL_STATE(2142)] = 61912,
  [SMALL_STATE(2143)] = 61972,
  [SMALL_STATE(2144)] = 62032,
  [SMALL_STATE(2145)] = 62092,
  [SMALL_STATE(2146)] = 62152,
  [SMALL_STATE(2147)] = 62248,
  [SMALL_STATE(2148)] = 62308,
  [SMALL_STATE(2149)] = 62368,
  [SMALL_STATE(2150)] = 62428,
  [SMALL_STATE(2151)] = 62488,
  [SMALL_STATE(2152)] = 62548,
  [SMALL_STATE(2153)] = 62608,
  [SMALL_STATE(2154)] = 62668,
  [SMALL_STATE(2155)] = 62728,
  [SMALL_STATE(2156)] = 62788,
  [SMALL_STATE(2157)] = 62848,
  [SMALL_STATE(2158)] = 62908,
  [SMALL_STATE(2159)] = 62972,
  [SMALL_STATE(2160)] = 63032,
  [SMALL_STATE(2161)] = 63092,
  [SMALL_STATE(2162)] = 63156,
  [SMALL_STATE(2163)] = 63216,
  [SMALL_STATE(2164)] = 63276,
  [SMALL_STATE(2165)] = 63340,
  [SMALL_STATE(2166)] = 63400,
  [SMALL_STATE(2167)] = 63464,
  [SMALL_STATE(2168)] = 63524,
  [SMALL_STATE(2169)] = 63584,
  [SMALL_STATE(2170)] = 63644,
  [SMALL_STATE(2171)] = 63704,
  [SMALL_STATE(2172)] = 63764,
  [SMALL_STATE(2173)] = 63824,
  [SMALL_STATE(2174)] = 63888,
  [SMALL_STATE(2175)] = 63950,
  [SMALL_STATE(2176)] = 64010,
  [SMALL_STATE(2177)] = 64072,
  [SMALL_STATE(2178)] = 64132,
  [SMALL_STATE(2179)] = 64228,
  [SMALL_STATE(2180)] = 64288,
  [SMALL_STATE(2181)] = 64348,
  [SMALL_STATE(2182)] = 64451,
  [SMALL_STATE(2183)] = 64510,
  [SMALL_STATE(2184)] = 64571,
  [SMALL_STATE(2185)] = 64630,
  [SMALL_STATE(2186)] = 64693,
  [SMALL_STATE(2187)] = 64752,
  [SMALL_STATE(2188)] = 64811,
  [SMALL_STATE(2189)] = 64870,
  [SMALL_STATE(2190)] = 64929,
  [SMALL_STATE(2191)] = 65042,
  [SMALL_STATE(2192)] = 65101,
  [SMALL_STATE(2193)] = 65214,
  [SMALL_STATE(2194)] = 65273,
  [SMALL_STATE(2195)] = 65376,
  [SMALL_STATE(2196)] = 65479,
  [SMALL_STATE(2197)] = 65582,
  [SMALL_STATE(2198)] = 65641,
  [SMALL_STATE(2199)] = 65744,
  [SMALL_STATE(2200)] = 65803,
  [SMALL_STATE(2201)] = 65904,
  [SMALL_STATE(2202)] = 65967,
  [SMALL_STATE(2203)] = 66026,
  [SMALL_STATE(2204)] = 66127,
  [SMALL_STATE(2205)] = 66186,
  [SMALL_STATE(2206)] = 66287,
  [SMALL_STATE(2207)] = 66346,
  [SMALL_STATE(2208)] = 66405,
  [SMALL_STATE(2209)] = 66508,
  [SMALL_STATE(2210)] = 66567,
  [SMALL_STATE(2211)] = 66626,
  [SMALL_STATE(2212)] = 66685,
  [SMALL_STATE(2213)] = 66744,
  [SMALL_STATE(2214)] = 66803,
  [SMALL_STATE(2215)] = 66862,
  [SMALL_STATE(2216)] = 66921,
  [SMALL_STATE(2217)] = 67034,
  [SMALL_STATE(2218)] = 67137,
  [SMALL_STATE(2219)] = 67196,
  [SMALL_STATE(2220)] = 67255,
  [SMALL_STATE(2221)] = 67314,
  [SMALL_STATE(2222)] = 67373,
  [SMALL_STATE(2223)] = 67432,
  [SMALL_STATE(2224)] = 67491,
  [SMALL_STATE(2225)] = 67550,
  [SMALL_STATE(2226)] = 67609,
  [SMALL_STATE(2227)] = 67668,
  [SMALL_STATE(2228)] = 67727,
  [SMALL_STATE(2229)] = 67786,
  [SMALL_STATE(2230)] = 67845,
  [SMALL_STATE(2231)] = 67904,
  [SMALL_STATE(2232)] = 68005,
  [SMALL_STATE(2233)] = 68064,
  [SMALL_STATE(2234)] = 68123,
  [SMALL_STATE(2235)] = 68182,
  [SMALL_STATE(2236)] = 68241,
  [SMALL_STATE(2237)] = 68300,
  [SMALL_STATE(2238)] = 68363,
  [SMALL_STATE(2239)] = 68422,
  [SMALL_STATE(2240)] = 68481,
  [SMALL_STATE(2241)] = 68540,
  [SMALL_STATE(2242)] = 68643,
  [SMALL_STATE(2243)] = 68702,
  [SMALL_STATE(2244)] = 68761,
  [SMALL_STATE(2245)] = 68820,
  [SMALL_STATE(2246)] = 68879,
  [SMALL_STATE(2247)] = 68980,
  [SMALL_STATE(2248)] = 69093,
  [SMALL_STATE(2249)] = 69152,
  [SMALL_STATE(2250)] = 69253,
  [SMALL_STATE(2251)] = 69312,
  [SMALL_STATE(2252)] = 69371,
  [SMALL_STATE(2253)] = 69474,
  [SMALL_STATE(2254)] = 69533,
  [SMALL_STATE(2255)] = 69592,
  [SMALL_STATE(2256)] = 69651,
  [SMALL_STATE(2257)] = 69764,
  [SMALL_STATE(2258)] = 69823,
  [SMALL_STATE(2259)] = 69882,
  [SMALL_STATE(2260)] = 69941,
  [SMALL_STATE(2261)] = 70000,
  [SMALL_STATE(2262)] = 70059,
  [SMALL_STATE(2263)] = 70118,
  [SMALL_STATE(2264)] = 70177,
  [SMALL_STATE(2265)] = 70236,
  [SMALL_STATE(2266)] = 70295,
  [SMALL_STATE(2267)] = 70354,
  [SMALL_STATE(2268)] = 70455,
  [SMALL_STATE(2269)] = 70514,
  [SMALL_STATE(2270)] = 70573,
  [SMALL_STATE(2271)] = 70632,
  [SMALL_STATE(2272)] = 70691,
  [SMALL_STATE(2273)] = 70754,
  [SMALL_STATE(2274)] = 70813,
  [SMALL_STATE(2275)] = 70872,
  [SMALL_STATE(2276)] = 70931,
  [SMALL_STATE(2277)] = 70992,
  [SMALL_STATE(2278)] = 71051,
  [SMALL_STATE(2279)] = 71110,
  [SMALL_STATE(2280)] = 71223,
  [SMALL_STATE(2281)] = 71282,
  [SMALL_STATE(2282)] = 71345,
  [SMALL_STATE(2283)] = 71408,
  [SMALL_STATE(2284)] = 71511,
  [SMALL_STATE(2285)] = 71612,
  [SMALL_STATE(2286)] = 71671,
  [SMALL_STATE(2287)] = 71730,
  [SMALL_STATE(2288)] = 71789,
  [SMALL_STATE(2289)] = 71848,
  [SMALL_STATE(2290)] = 71907,
  [SMALL_STATE(2291)] = 71966,
  [SMALL_STATE(2292)] = 72067,
  [SMALL_STATE(2293)] = 72126,
  [SMALL_STATE(2294)] = 72185,
  [SMALL_STATE(2295)] = 72244,
  [SMALL_STATE(2296)] = 72303,
  [SMALL_STATE(2297)] = 72362,
  [SMALL_STATE(2298)] = 72421,
  [SMALL_STATE(2299)] = 72480,
  [SMALL_STATE(2300)] = 72539,
  [SMALL_STATE(2301)] = 72642,
  [SMALL_STATE(2302)] = 72743,
  [SMALL_STATE(2303)] = 72805,
  [SMALL_STATE(2304)] = 72915,
  [SMALL_STATE(2305)] = 72975,
  [SMALL_STATE(2306)] = 73037,
  [SMALL_STATE(2307)] = 73099,
  [SMALL_STATE(2308)] = 73209,
  [SMALL_STATE(2309)] = 73271,
  [SMALL_STATE(2310)] = 73381,
  [SMALL_STATE(2311)] = 73443,
  [SMALL_STATE(2312)] = 73503,
  [SMALL_STATE(2313)] = 73613,
  [SMALL_STATE(2314)] = 73723,
  [SMALL_STATE(2315)] = 73785,
  [SMALL_STATE(2316)] = 73895,
  [SMALL_STATE(2317)] = 73954,
  [SMALL_STATE(2318)] = 74016,
  [SMALL_STATE(2319)] = 74072,
  [SMALL_STATE(2320)] = 74128,
  [SMALL_STATE(2321)] = 74184,
  [SMALL_STATE(2322)] = 74240,
  [SMALL_STATE(2323)] = 74296,
  [SMALL_STATE(2324)] = 74352,
  [SMALL_STATE(2325)] = 74408,
  [SMALL_STATE(2326)] = 74464,
  [SMALL_STATE(2327)] = 74520,
  [SMALL_STATE(2328)] = 74576,
  [SMALL_STATE(2329)] = 74632,
  [SMALL_STATE(2330)] = 74688,
  [SMALL_STATE(2331)] = 74744,
  [SMALL_STATE(2332)] = 74800,
  [SMALL_STATE(2333)] = 74856,
  [SMALL_STATE(2334)] = 74912,
  [SMALL_STATE(2335)] = 74968,
  [SMALL_STATE(2336)] = 75024,
  [SMALL_STATE(2337)] = 75080,
  [SMALL_STATE(2338)] = 75136,
  [SMALL_STATE(2339)] = 75192,
  [SMALL_STATE(2340)] = 75248,
  [SMALL_STATE(2341)] = 75304,
  [SMALL_STATE(2342)] = 75360,
  [SMALL_STATE(2343)] = 75416,
  [SMALL_STATE(2344)] = 75472,
  [SMALL_STATE(2345)] = 75530,
  [SMALL_STATE(2346)] = 75586,
  [SMALL_STATE(2347)] = 75642,
  [SMALL_STATE(2348)] = 75698,
  [SMALL_STATE(2349)] = 75754,
  [SMALL_STATE(2350)] = 75810,
  [SMALL_STATE(2351)] = 75866,
  [SMALL_STATE(2352)] = 75922,
  [SMALL_STATE(2353)] = 75978,
  [SMALL_STATE(2354)] = 76034,
  [SMALL_STATE(2355)] = 76090,
  [SMALL_STATE(2356)] = 76146,
  [SMALL_STATE(2357)] = 76202,
  [SMALL_STATE(2358)] = 76258,
  [SMALL_STATE(2359)] = 76316,
  [SMALL_STATE(2360)] = 76378,
  [SMALL_STATE(2361)] = 76434,
  [SMALL_STATE(2362)] = 76490,
  [SMALL_STATE(2363)] = 76546,
  [SMALL_STATE(2364)] = 76602,
  [SMALL_STATE(2365)] = 76658,
  [SMALL_STATE(2366)] = 76716,
  [SMALL_STATE(2367)] = 76772,
  [SMALL_STATE(2368)] = 76828,
  [SMALL_STATE(2369)] = 76884,
  [SMALL_STATE(2370)] = 76940,
  [SMALL_STATE(2371)] = 76996,
  [SMALL_STATE(2372)] = 77052,
  [SMALL_STATE(2373)] = 77108,
  [SMALL_STATE(2374)] = 77164,
  [SMALL_STATE(2375)] = 77220,
  [SMALL_STATE(2376)] = 77276,
  [SMALL_STATE(2377)] = 77332,
  [SMALL_STATE(2378)] = 77388,
  [SMALL_STATE(2379)] = 77444,
  [SMALL_STATE(2380)] = 77500,
  [SMALL_STATE(2381)] = 77556,
  [SMALL_STATE(2382)] = 77612,
  [SMALL_STATE(2383)] = 77668,
  [SMALL_STATE(2384)] = 77724,
  [SMALL_STATE(2385)] = 77780,
  [SMALL_STATE(2386)] = 77836,
  [SMALL_STATE(2387)] = 77892,
  [SMALL_STATE(2388)] = 77948,
  [SMALL_STATE(2389)] = 78004,
  [SMALL_STATE(2390)] = 78060,
  [SMALL_STATE(2391)] = 78116,
  [SMALL_STATE(2392)] = 78172,
  [SMALL_STATE(2393)] = 78228,
  [SMALL_STATE(2394)] = 78284,
  [SMALL_STATE(2395)] = 78340,
  [SMALL_STATE(2396)] = 78396,
  [SMALL_STATE(2397)] = 78454,
  [SMALL_STATE(2398)] = 78512,
  [SMALL_STATE(2399)] = 78568,
  [SMALL_STATE(2400)] = 78624,
  [SMALL_STATE(2401)] = 78680,
  [SMALL_STATE(2402)] = 78736,
  [SMALL_STATE(2403)] = 78792,
  [SMALL_STATE(2404)] = 78848,
  [SMALL_STATE(2405)] = 78904,
  [SMALL_STATE(2406)] = 78960,
  [SMALL_STATE(2407)] = 79016,
  [SMALL_STATE(2408)] = 79072,
  [SMALL_STATE(2409)] = 79128,
  [SMALL_STATE(2410)] = 79184,
  [SMALL_STATE(2411)] = 79240,
  [SMALL_STATE(2412)] = 79296,
  [SMALL_STATE(2413)] = 79352,
  [SMALL_STATE(2414)] = 79408,
  [SMALL_STATE(2415)] = 79464,
  [SMALL_STATE(2416)] = 79520,
  [SMALL_STATE(2417)] = 79576,
  [SMALL_STATE(2418)] = 79632,
  [SMALL_STATE(2419)] = 79690,
  [SMALL_STATE(2420)] = 79746,
  [SMALL_STATE(2421)] = 79802,
  [SMALL_STATE(2422)] = 79858,
  [SMALL_STATE(2423)] = 79914,
  [SMALL_STATE(2424)] = 79970,
  [SMALL_STATE(2425)] = 80025,
  [SMALL_STATE(2426)] = 80080,
  [SMALL_STATE(2427)] = 80135,
  [SMALL_STATE(2428)] = 80190,
  [SMALL_STATE(2429)] = 80245,
  [SMALL_STATE(2430)] = 80300,
  [SMALL_STATE(2431)] = 80355,
  [SMALL_STATE(2432)] = 80410,
  [SMALL_STATE(2433)] = 80465,
  [SMALL_STATE(2434)] = 80520,
  [SMALL_STATE(2435)] = 80575,
  [SMALL_STATE(2436)] = 80630,
  [SMALL_STATE(2437)] = 80685,
  [SMALL_STATE(2438)] = 80740,
  [SMALL_STATE(2439)] = 80795,
  [SMALL_STATE(2440)] = 80850,
  [SMALL_STATE(2441)] = 80905,
  [SMALL_STATE(2442)] = 81014,
  [SMALL_STATE(2443)] = 81069,
  [SMALL_STATE(2444)] = 81124,
  [SMALL_STATE(2445)] = 81179,
  [SMALL_STATE(2446)] = 81234,
  [SMALL_STATE(2447)] = 81289,
  [SMALL_STATE(2448)] = 81344,
  [SMALL_STATE(2449)] = 81399,
  [SMALL_STATE(2450)] = 81454,
  [SMALL_STATE(2451)] = 81509,
  [SMALL_STATE(2452)] = 81564,
  [SMALL_STATE(2453)] = 81619,
  [SMALL_STATE(2454)] = 81674,
  [SMALL_STATE(2455)] = 81729,
  [SMALL_STATE(2456)] = 81784,
  [SMALL_STATE(2457)] = 81839,
  [SMALL_STATE(2458)] = 81894,
  [SMALL_STATE(2459)] = 81949,
  [SMALL_STATE(2460)] = 82004,
  [SMALL_STATE(2461)] = 82059,
  [SMALL_STATE(2462)] = 82114,
  [SMALL_STATE(2463)] = 82169,
  [SMALL_STATE(2464)] = 82224,
  [SMALL_STATE(2465)] = 82279,
  [SMALL_STATE(2466)] = 82334,
  [SMALL_STATE(2467)] = 82389,
  [SMALL_STATE(2468)] = 82444,
  [SMALL_STATE(2469)] = 82499,
  [SMALL_STATE(2470)] = 82554,
  [SMALL_STATE(2471)] = 82609,
  [SMALL_STATE(2472)] = 82664,
  [SMALL_STATE(2473)] = 82719,
  [SMALL_STATE(2474)] = 82774,
  [SMALL_STATE(2475)] = 82829,
  [SMALL_STATE(2476)] = 82884,
  [SMALL_STATE(2477)] = 82939,
  [SMALL_STATE(2478)] = 82994,
  [SMALL_STATE(2479)] = 83049,
  [SMALL_STATE(2480)] = 83104,
  [SMALL_STATE(2481)] = 83159,
  [SMALL_STATE(2482)] = 83214,
  [SMALL_STATE(2483)] = 83269,
  [SMALL_STATE(2484)] = 83324,
  [SMALL_STATE(2485)] = 83433,
  [SMALL_STATE(2486)] = 83542,
  [SMALL_STATE(2487)] = 83597,
  [SMALL_STATE(2488)] = 83652,
  [SMALL_STATE(2489)] = 83707,
  [SMALL_STATE(2490)] = 83762,
  [SMALL_STATE(2491)] = 83818,
  [SMALL_STATE(2492)] = 83922,
  [SMALL_STATE(2493)] = 84026,
  [SMALL_STATE(2494)] = 84130,
  [SMALL_STATE(2495)] = 84194,
  [SMALL_STATE(2496)] = 84258,
  [SMALL_STATE(2497)] = 84362,
  [SMALL_STATE(2498)] = 84466,
  [SMALL_STATE(2499)] = 84520,
  [SMALL_STATE(2500)] = 84624,
  [SMALL_STATE(2501)] = 84713,
  [SMALL_STATE(2502)] = 84766,
  [SMALL_STATE(2503)] = 84867,
  [SMALL_STATE(2504)] = 84922,
  [SMALL_STATE(2505)] = 85025,
  [SMALL_STATE(2506)] = 85128,
  [SMALL_STATE(2507)] = 85231,
  [SMALL_STATE(2508)] = 85320,
  [SMALL_STATE(2509)] = 85409,
  [SMALL_STATE(2510)] = 85512,
  [SMALL_STATE(2511)] = 85601,
  [SMALL_STATE(2512)] = 85666,
  [SMALL_STATE(2513)] = 85767,
  [SMALL_STATE(2514)] = 85870,
  [SMALL_STATE(2515)] = 85957,
  [SMALL_STATE(2516)] = 86020,
  [SMALL_STATE(2517)] = 86123,
  [SMALL_STATE(2518)] = 86212,
  [SMALL_STATE(2519)] = 86299,
  [SMALL_STATE(2520)] = 86374,
  [SMALL_STATE(2521)] = 86475,
  [SMALL_STATE(2522)] = 86578,
  [SMALL_STATE(2523)] = 86681,
  [SMALL_STATE(2524)] = 86768,
  [SMALL_STATE(2525)] = 86857,
  [SMALL_STATE(2526)] = 86958,
  [SMALL_STATE(2527)] = 87059,
  [SMALL_STATE(2528)] = 87122,
  [SMALL_STATE(2529)] = 87191,
  [SMALL_STATE(2530)] = 87276,
  [SMALL_STATE(2531)] = 87379,
  [SMALL_STATE(2532)] = 87466,
  [SMALL_STATE(2533)] = 87539,
  [SMALL_STATE(2534)] = 87614,
  [SMALL_STATE(2535)] = 87689,
  [SMALL_STATE(2536)] = 87776,
  [SMALL_STATE(2537)] = 87877,
  [SMALL_STATE(2538)] = 87978,
  [SMALL_STATE(2539)] = 88043,
  [SMALL_STATE(2540)] = 88108,
  [SMALL_STATE(2541)] = 88177,
  [SMALL_STATE(2542)] = 88280,
  [SMALL_STATE(2543)] = 88383,
  [SMALL_STATE(2544)] = 88462,
  [SMALL_STATE(2545)] = 88545,
  [SMALL_STATE(2546)] = 88632,
  [SMALL_STATE(2547)] = 88733,
  [SMALL_STATE(2548)] = 88836,
  [SMALL_STATE(2549)] = 88937,
  [SMALL_STATE(2550)] = 89024,
  [SMALL_STATE(2551)] = 89127,
  [SMALL_STATE(2552)] = 89230,
  [SMALL_STATE(2553)] = 89283,
  [SMALL_STATE(2554)] = 89386,
  [SMALL_STATE(2555)] = 89473,
  [SMALL_STATE(2556)] = 89528,
  [SMALL_STATE(2557)] = 89631,
  [SMALL_STATE(2558)] = 89686,
  [SMALL_STATE(2559)] = 89769,
  [SMALL_STATE(2560)] = 89834,
  [SMALL_STATE(2561)] = 89907,
  [SMALL_STATE(2562)] = 90010,
  [SMALL_STATE(2563)] = 90113,
  [SMALL_STATE(2564)] = 90214,
  [SMALL_STATE(2565)] = 90289,
  [SMALL_STATE(2566)] = 90344,
  [SMALL_STATE(2567)] = 90429,
  [SMALL_STATE(2568)] = 90508,
  [SMALL_STATE(2569)] = 90609,
  [SMALL_STATE(2570)] = 90710,
  [SMALL_STATE(2571)] = 90762,
  [SMALL_STATE(2572)] = 90814,
  [SMALL_STATE(2573)] = 90904,
  [SMALL_STATE(2574)] = 90956,
  [SMALL_STATE(2575)] = 91008,
  [SMALL_STATE(2576)] = 91100,
  [SMALL_STATE(2577)] = 91152,
  [SMALL_STATE(2578)] = 91204,
  [SMALL_STATE(2579)] = 91256,
  [SMALL_STATE(2580)] = 91308,
  [SMALL_STATE(2581)] = 91360,
  [SMALL_STATE(2582)] = 91460,
  [SMALL_STATE(2583)] = 91512,
  [SMALL_STATE(2584)] = 91564,
  [SMALL_STATE(2585)] = 91616,
  [SMALL_STATE(2586)] = 91668,
  [SMALL_STATE(2587)] = 91720,
  [SMALL_STATE(2588)] = 91814,
  [SMALL_STATE(2589)] = 91912,
  [SMALL_STATE(2590)] = 91964,
  [SMALL_STATE(2591)] = 92016,
  [SMALL_STATE(2592)] = 92108,
  [SMALL_STATE(2593)] = 92200,
  [SMALL_STATE(2594)] = 92252,
  [SMALL_STATE(2595)] = 92350,
  [SMALL_STATE(2596)] = 92402,
  [SMALL_STATE(2597)] = 92496,
  [SMALL_STATE(2598)] = 92548,
  [SMALL_STATE(2599)] = 92600,
  [SMALL_STATE(2600)] = 92652,
  [SMALL_STATE(2601)] = 92744,
  [SMALL_STATE(2602)] = 92796,
  [SMALL_STATE(2603)] = 92890,
  [SMALL_STATE(2604)] = 92942,
  [SMALL_STATE(2605)] = 92994,
  [SMALL_STATE(2606)] = 93046,
  [SMALL_STATE(2607)] = 93140,
  [SMALL_STATE(2608)] = 93234,
  [SMALL_STATE(2609)] = 93286,
  [SMALL_STATE(2610)] = 93338,
  [SMALL_STATE(2611)] = 93432,
  [SMALL_STATE(2612)] = 93484,
  [SMALL_STATE(2613)] = 93582,
  [SMALL_STATE(2614)] = 93634,
  [SMALL_STATE(2615)] = 93732,
  [SMALL_STATE(2616)] = 93830,
  [SMALL_STATE(2617)] = 93928,
  [SMALL_STATE(2618)] = 94026,
  [SMALL_STATE(2619)] = 94078,
  [SMALL_STATE(2620)] = 94166,
  [SMALL_STATE(2621)] = 94254,
  [SMALL_STATE(2622)] = 94348,
  [SMALL_STATE(2623)] = 94442,
  [SMALL_STATE(2624)] = 94536,
  [SMALL_STATE(2625)] = 94634,
  [SMALL_STATE(2626)] = 94732,
  [SMALL_STATE(2627)] = 94830,
  [SMALL_STATE(2628)] = 94924,
  [SMALL_STATE(2629)] = 95018,
  [SMALL_STATE(2630)] = 95070,
  [SMALL_STATE(2631)] = 95158,
  [SMALL_STATE(2632)] = 95252,
  [SMALL_STATE(2633)] = 95304,
  [SMALL_STATE(2634)] = 95357,
  [SMALL_STATE(2635)] = 95434,
  [SMALL_STATE(2636)] = 95487,
  [SMALL_STATE(2637)] = 95538,
  [SMALL_STATE(2638)] = 95588,
  [SMALL_STATE(2639)] = 95638,
  [SMALL_STATE(2640)] = 95688,
  [SMALL_STATE(2641)] = 95774,
  [SMALL_STATE(2642)] = 95824,
  [SMALL_STATE(2643)] = 95874,
  [SMALL_STATE(2644)] = 95924,
  [SMALL_STATE(2645)] = 95974,
  [SMALL_STATE(2646)] = 96024,
  [SMALL_STATE(2647)] = 96074,
  [SMALL_STATE(2648)] = 96160,
  [SMALL_STATE(2649)] = 96210,
  [SMALL_STATE(2650)] = 96296,
  [SMALL_STATE(2651)] = 96346,
  [SMALL_STATE(2652)] = 96396,
  [SMALL_STATE(2653)] = 96446,
  [SMALL_STATE(2654)] = 96495,
  [SMALL_STATE(2655)] = 96544,
  [SMALL_STATE(2656)] = 96593,
  [SMALL_STATE(2657)] = 96642,
  [SMALL_STATE(2658)] = 96691,
  [SMALL_STATE(2659)] = 96740,
  [SMALL_STATE(2660)] = 96789,
  [SMALL_STATE(2661)] = 96838,
  [SMALL_STATE(2662)] = 96887,
  [SMALL_STATE(2663)] = 96936,
  [SMALL_STATE(2664)] = 96985,
  [SMALL_STATE(2665)] = 97034,
  [SMALL_STATE(2666)] = 97083,
  [SMALL_STATE(2667)] = 97132,
  [SMALL_STATE(2668)] = 97181,
  [SMALL_STATE(2669)] = 97230,
  [SMALL_STATE(2670)] = 97279,
  [SMALL_STATE(2671)] = 97332,
  [SMALL_STATE(2672)] = 97383,
  [SMALL_STATE(2673)] = 97442,
  [SMALL_STATE(2674)] = 97497,
  [SMALL_STATE(2675)] = 97548,
  [SMALL_STATE(2676)] = 97597,
  [SMALL_STATE(2677)] = 97648,
  [SMALL_STATE(2678)] = 97697,
  [SMALL_STATE(2679)] = 97746,
  [SMALL_STATE(2680)] = 97795,
  [SMALL_STATE(2681)] = 97844,
  [SMALL_STATE(2682)] = 97893,
  [SMALL_STATE(2683)] = 97942,
  [SMALL_STATE(2684)] = 97991,
  [SMALL_STATE(2685)] = 98040,
  [SMALL_STATE(2686)] = 98089,
  [SMALL_STATE(2687)] = 98138,
  [SMALL_STATE(2688)] = 98187,
  [SMALL_STATE(2689)] = 98236,
  [SMALL_STATE(2690)] = 98285,
  [SMALL_STATE(2691)] = 98334,
  [SMALL_STATE(2692)] = 98383,
  [SMALL_STATE(2693)] = 98432,
  [SMALL_STATE(2694)] = 98481,
  [SMALL_STATE(2695)] = 98532,
  [SMALL_STATE(2696)] = 98581,
  [SMALL_STATE(2697)] = 98630,
  [SMALL_STATE(2698)] = 98681,
  [SMALL_STATE(2699)] = 98730,
  [SMALL_STATE(2700)] = 98779,
  [SMALL_STATE(2701)] = 98828,
  [SMALL_STATE(2702)] = 98917,
  [SMALL_STATE(2703)] = 98966,
  [SMALL_STATE(2704)] = 99055,
  [SMALL_STATE(2705)] = 99110,
  [SMALL_STATE(2706)] = 99199,
  [SMALL_STATE(2707)] = 99288,
  [SMALL_STATE(2708)] = 99377,
  [SMALL_STATE(2709)] = 99426,
  [SMALL_STATE(2710)] = 99475,
  [SMALL_STATE(2711)] = 99530,
  [SMALL_STATE(2712)] = 99585,
  [SMALL_STATE(2713)] = 99634,
  [SMALL_STATE(2714)] = 99683,
  [SMALL_STATE(2715)] = 99732,
  [SMALL_STATE(2716)] = 99781,
  [SMALL_STATE(2717)] = 99830,
  [SMALL_STATE(2718)] = 99879,
  [SMALL_STATE(2719)] = 99968,
  [SMALL_STATE(2720)] = 100017,
  [SMALL_STATE(2721)] = 100066,
  [SMALL_STATE(2722)] = 100115,
  [SMALL_STATE(2723)] = 100168,
  [SMALL_STATE(2724)] = 100217,
  [SMALL_STATE(2725)] = 100266,
  [SMALL_STATE(2726)] = 100319,
  [SMALL_STATE(2727)] = 100408,
  [SMALL_STATE(2728)] = 100457,
  [SMALL_STATE(2729)] = 100506,
  [SMALL_STATE(2730)] = 100555,
  [SMALL_STATE(2731)] = 100644,
  [SMALL_STATE(2732)] = 100693,
  [SMALL_STATE(2733)] = 100782,
  [SMALL_STATE(2734)] = 100831,
  [SMALL_STATE(2735)] = 100890,
  [SMALL_STATE(2736)] = 100979,
  [SMALL_STATE(2737)] = 101068,
  [SMALL_STATE(2738)] = 101117,
  [SMALL_STATE(2739)] = 101166,
  [SMALL_STATE(2740)] = 101219,
  [SMALL_STATE(2741)] = 101268,
  [SMALL_STATE(2742)] = 101317,
  [SMALL_STATE(2743)] = 101366,
  [SMALL_STATE(2744)] = 101415,
  [SMALL_STATE(2745)] = 101466,
  [SMALL_STATE(2746)] = 101515,
  [SMALL_STATE(2747)] = 101564,
  [SMALL_STATE(2748)] = 101653,
  [SMALL_STATE(2749)] = 101706,
  [SMALL_STATE(2750)] = 101755,
  [SMALL_STATE(2751)] = 101804,
  [SMALL_STATE(2752)] = 101893,
  [SMALL_STATE(2753)] = 101942,
  [SMALL_STATE(2754)] = 101991,
  [SMALL_STATE(2755)] = 102080,
  [SMALL_STATE(2756)] = 102129,
  [SMALL_STATE(2757)] = 102178,
  [SMALL_STATE(2758)] = 102231,
  [SMALL_STATE(2759)] = 102280,
  [SMALL_STATE(2760)] = 102369,
  [SMALL_STATE(2761)] = 102418,
  [SMALL_STATE(2762)] = 102467,
  [SMALL_STATE(2763)] = 102522,
  [SMALL_STATE(2764)] = 102608,
  [SMALL_STATE(2765)] = 102694,
  [SMALL_STATE(2766)] = 102780,
  [SMALL_STATE(2767)] = 102866,
  [SMALL_STATE(2768)] = 102952,
  [SMALL_STATE(2769)] = 103000,
  [SMALL_STATE(2770)] = 103048,
  [SMALL_STATE(2771)] = 103096,
  [SMALL_STATE(2772)] = 103182,
  [SMALL_STATE(2773)] = 103230,
  [SMALL_STATE(2774)] = 103316,
  [SMALL_STATE(2775)] = 103402,
  [SMALL_STATE(2776)] = 103488,
  [SMALL_STATE(2777)] = 103574,
  [SMALL_STATE(2778)] = 103622,
  [SMALL_STATE(2779)] = 103708,
  [SMALL_STATE(2780)] = 103794,
  [SMALL_STATE(2781)] = 103880,
  [SMALL_STATE(2782)] = 103966,
  [SMALL_STATE(2783)] = 104052,
  [SMALL_STATE(2784)] = 104138,
  [SMALL_STATE(2785)] = 104224,
  [SMALL_STATE(2786)] = 104310,
  [SMALL_STATE(2787)] = 104396,
  [SMALL_STATE(2788)] = 104482,
  [SMALL_STATE(2789)] = 104568,
  [SMALL_STATE(2790)] = 104654,
  [SMALL_STATE(2791)] = 104740,
  [SMALL_STATE(2792)] = 104826,
  [SMALL_STATE(2793)] = 104912,
  [SMALL_STATE(2794)] = 104996,
  [SMALL_STATE(2795)] = 105080,
  [SMALL_STATE(2796)] = 105166,
  [SMALL_STATE(2797)] = 105224,
  [SMALL_STATE(2798)] = 105284,
  [SMALL_STATE(2799)] = 105344,
  [SMALL_STATE(2800)] = 105408,
  [SMALL_STATE(2801)] = 105492,
  [SMALL_STATE(2802)] = 105578,
  [SMALL_STATE(2803)] = 105664,
  [SMALL_STATE(2804)] = 105750,
  [SMALL_STATE(2805)] = 105836,
  [SMALL_STATE(2806)] = 105922,
  [SMALL_STATE(2807)] = 106008,
  [SMALL_STATE(2808)] = 106094,
  [SMALL_STATE(2809)] = 106180,
  [SMALL_STATE(2810)] = 106266,
  [SMALL_STATE(2811)] = 106352,
  [SMALL_STATE(2812)] = 106438,
  [SMALL_STATE(2813)] = 106524,
  [SMALL_STATE(2814)] = 106610,
  [SMALL_STATE(2815)] = 106696,
  [SMALL_STATE(2816)] = 106782,
  [SMALL_STATE(2817)] = 106868,
  [SMALL_STATE(2818)] = 106954,
  [SMALL_STATE(2819)] = 107040,
  [SMALL_STATE(2820)] = 107126,
  [SMALL_STATE(2821)] = 107206,
  [SMALL_STATE(2822)] = 107274,
  [SMALL_STATE(2823)] = 107344,
  [SMALL_STATE(2824)] = 107414,
  [SMALL_STATE(2825)] = 107496,
  [SMALL_STATE(2826)] = 107570,
  [SMALL_STATE(2827)] = 107648,
  [SMALL_STATE(2828)] = 107730,
  [SMALL_STATE(2829)] = 107812,
  [SMALL_STATE(2830)] = 107894,
  [SMALL_STATE(2831)] = 107980,
  [SMALL_STATE(2832)] = 108066,
  [SMALL_STATE(2833)] = 108152,
  [SMALL_STATE(2834)] = 108238,
  [SMALL_STATE(2835)] = 108324,
  [SMALL_STATE(2836)] = 108410,
  [SMALL_STATE(2837)] = 108460,
  [SMALL_STATE(2838)] = 108546,
  [SMALL_STATE(2839)] = 108632,
  [SMALL_STATE(2840)] = 108718,
  [SMALL_STATE(2841)] = 108804,
  [SMALL_STATE(2842)] = 108890,
  [SMALL_STATE(2843)] = 108976,
  [SMALL_STATE(2844)] = 109026,
  [SMALL_STATE(2845)] = 109107,
  [SMALL_STATE(2846)] = 109164,
  [SMALL_STATE(2847)] = 109213,
  [SMALL_STATE(2848)] = 109270,
  [SMALL_STATE(2849)] = 109353,
  [SMALL_STATE(2850)] = 109436,
  [SMALL_STATE(2851)] = 109515,
  [SMALL_STATE(2852)] = 109582,
  [SMALL_STATE(2853)] = 109651,
  [SMALL_STATE(2854)] = 109720,
  [SMALL_STATE(2855)] = 109801,
  [SMALL_STATE(2856)] = 109860,
  [SMALL_STATE(2857)] = 109919,
  [SMALL_STATE(2858)] = 109982,
  [SMALL_STATE(2859)] = 110055,
  [SMALL_STATE(2860)] = 110132,
  [SMALL_STATE(2861)] = 110213,
  [SMALL_STATE(2862)] = 110294,
  [SMALL_STATE(2863)] = 110377,
  [SMALL_STATE(2864)] = 110424,
  [SMALL_STATE(2865)] = 110475,
  [SMALL_STATE(2866)] = 110542,
  [SMALL_STATE(2867)] = 110611,
  [SMALL_STATE(2868)] = 110680,
  [SMALL_STATE(2869)] = 110761,
  [SMALL_STATE(2870)] = 110834,
  [SMALL_STATE(2871)] = 110911,
  [SMALL_STATE(2872)] = 110992,
  [SMALL_STATE(2873)] = 111073,
  [SMALL_STATE(2874)] = 111154,
  [SMALL_STATE(2875)] = 111201,
  [SMALL_STATE(2876)] = 111250,
  [SMALL_STATE(2877)] = 111333,
  [SMALL_STATE(2878)] = 111390,
  [SMALL_STATE(2879)] = 111449,
  [SMALL_STATE(2880)] = 111508,
  [SMALL_STATE(2881)] = 111571,
  [SMALL_STATE(2882)] = 111654,
  [SMALL_STATE(2883)] = 111737,
  [SMALL_STATE(2884)] = 111788,
  [SMALL_STATE(2885)] = 111867,
  [SMALL_STATE(2886)] = 111913,
  [SMALL_STATE(2887)] = 111959,
  [SMALL_STATE(2888)] = 112037,
  [SMALL_STATE(2889)] = 112083,
  [SMALL_STATE(2890)] = 112129,
  [SMALL_STATE(2891)] = 112175,
  [SMALL_STATE(2892)] = 112257,
  [SMALL_STATE(2893)] = 112303,
  [SMALL_STATE(2894)] = 112349,
  [SMALL_STATE(2895)] = 112395,
  [SMALL_STATE(2896)] = 112441,
  [SMALL_STATE(2897)] = 112487,
  [SMALL_STATE(2898)] = 112565,
  [SMALL_STATE(2899)] = 112621,
  [SMALL_STATE(2900)] = 112703,
  [SMALL_STATE(2901)] = 112761,
  [SMALL_STATE(2902)] = 112819,
  [SMALL_STATE(2903)] = 112881,
  [SMALL_STATE(2904)] = 112927,
  [SMALL_STATE(2905)] = 112993,
  [SMALL_STATE(2906)] = 113039,
  [SMALL_STATE(2907)] = 113085,
  [SMALL_STATE(2908)] = 113131,
  [SMALL_STATE(2909)] = 113177,
  [SMALL_STATE(2910)] = 113245,
  [SMALL_STATE(2911)] = 113323,
  [SMALL_STATE(2912)] = 113369,
  [SMALL_STATE(2913)] = 113449,
  [SMALL_STATE(2914)] = 113495,
  [SMALL_STATE(2915)] = 113541,
  [SMALL_STATE(2916)] = 113619,
  [SMALL_STATE(2917)] = 113665,
  [SMALL_STATE(2918)] = 113725,
  [SMALL_STATE(2919)] = 113771,
  [SMALL_STATE(2920)] = 113843,
  [SMALL_STATE(2921)] = 113889,
  [SMALL_STATE(2922)] = 113957,
  [SMALL_STATE(2923)] = 114003,
  [SMALL_STATE(2924)] = 114049,
  [SMALL_STATE(2925)] = 114097,
  [SMALL_STATE(2926)] = 114145,
  [SMALL_STATE(2927)] = 114191,
  [SMALL_STATE(2928)] = 114237,
  [SMALL_STATE(2929)] = 114283,
  [SMALL_STATE(2930)] = 114329,
  [SMALL_STATE(2931)] = 114375,
  [SMALL_STATE(2932)] = 114443,
  [SMALL_STATE(2933)] = 114519,
  [SMALL_STATE(2934)] = 114575,
  [SMALL_STATE(2935)] = 114621,
  [SMALL_STATE(2936)] = 114667,
  [SMALL_STATE(2937)] = 114713,
  [SMALL_STATE(2938)] = 114795,
  [SMALL_STATE(2939)] = 114853,
  [SMALL_STATE(2940)] = 114899,
  [SMALL_STATE(2941)] = 114945,
  [SMALL_STATE(2942)] = 114991,
  [SMALL_STATE(2943)] = 115037,
  [SMALL_STATE(2944)] = 115083,
  [SMALL_STATE(2945)] = 115157,
  [SMALL_STATE(2946)] = 115225,
  [SMALL_STATE(2947)] = 115305,
  [SMALL_STATE(2948)] = 115351,
  [SMALL_STATE(2949)] = 115419,
  [SMALL_STATE(2950)] = 115465,
  [SMALL_STATE(2951)] = 115533,
  [SMALL_STATE(2952)] = 115579,
  [SMALL_STATE(2953)] = 115647,
  [SMALL_STATE(2954)] = 115693,
  [SMALL_STATE(2955)] = 115773,
  [SMALL_STATE(2956)] = 115819,
  [SMALL_STATE(2957)] = 115865,
  [SMALL_STATE(2958)] = 115910,
  [SMALL_STATE(2959)] = 115955,
  [SMALL_STATE(2960)] = 116030,
  [SMALL_STATE(2961)] = 116109,
  [SMALL_STATE(2962)] = 116154,
  [SMALL_STATE(2963)] = 116199,
  [SMALL_STATE(2964)] = 116246,
  [SMALL_STATE(2965)] = 116291,
  [SMALL_STATE(2966)] = 116336,
  [SMALL_STATE(2967)] = 116381,
  [SMALL_STATE(2968)] = 116426,
  [SMALL_STATE(2969)] = 116471,
  [SMALL_STATE(2970)] = 116516,
  [SMALL_STATE(2971)] = 116561,
  [SMALL_STATE(2972)] = 116606,
  [SMALL_STATE(2973)] = 116651,
  [SMALL_STATE(2974)] = 116696,
  [SMALL_STATE(2975)] = 116741,
  [SMALL_STATE(2976)] = 116802,
  [SMALL_STATE(2977)] = 116847,
  [SMALL_STATE(2978)] = 116892,
  [SMALL_STATE(2979)] = 116937,
  [SMALL_STATE(2980)] = 116982,
  [SMALL_STATE(2981)] = 117065,
  [SMALL_STATE(2982)] = 117110,
  [SMALL_STATE(2983)] = 117155,
  [SMALL_STATE(2984)] = 117236,
  [SMALL_STATE(2985)] = 117281,
  [SMALL_STATE(2986)] = 117326,
  [SMALL_STATE(2987)] = 117371,
  [SMALL_STATE(2988)] = 117416,
  [SMALL_STATE(2989)] = 117461,
  [SMALL_STATE(2990)] = 117506,
  [SMALL_STATE(2991)] = 117551,
  [SMALL_STATE(2992)] = 117628,
  [SMALL_STATE(2993)] = 117673,
  [SMALL_STATE(2994)] = 117718,
  [SMALL_STATE(2995)] = 117763,
  [SMALL_STATE(2996)] = 117808,
  [SMALL_STATE(2997)] = 117853,
  [SMALL_STATE(2998)] = 117898,
  [SMALL_STATE(2999)] = 117943,
  [SMALL_STATE(3000)] = 117988,
  [SMALL_STATE(3001)] = 118033,
  [SMALL_STATE(3002)] = 118112,
  [SMALL_STATE(3003)] = 118157,
  [SMALL_STATE(3004)] = 118202,
  [SMALL_STATE(3005)] = 118247,
  [SMALL_STATE(3006)] = 118292,
  [SMALL_STATE(3007)] = 118341,
  [SMALL_STATE(3008)] = 118388,
  [SMALL_STATE(3009)] = 118433,
  [SMALL_STATE(3010)] = 118484,
  [SMALL_STATE(3011)] = 118529,
  [SMALL_STATE(3012)] = 118574,
  [SMALL_STATE(3013)] = 118621,
  [SMALL_STATE(3014)] = 118666,
  [SMALL_STATE(3015)] = 118711,
  [SMALL_STATE(3016)] = 118756,
  [SMALL_STATE(3017)] = 118801,
  [SMALL_STATE(3018)] = 118848,
  [SMALL_STATE(3019)] = 118927,
  [SMALL_STATE(3020)] = 118972,
  [SMALL_STATE(3021)] = 119017,
  [SMALL_STATE(3022)] = 119062,
  [SMALL_STATE(3023)] = 119107,
  [SMALL_STATE(3024)] = 119152,
  [SMALL_STATE(3025)] = 119197,
  [SMALL_STATE(3026)] = 119270,
  [SMALL_STATE(3027)] = 119323,
  [SMALL_STATE(3028)] = 119368,
  [SMALL_STATE(3029)] = 119413,
  [SMALL_STATE(3030)] = 119458,
  [SMALL_STATE(3031)] = 119503,
  [SMALL_STATE(3032)] = 119548,
  [SMALL_STATE(3033)] = 119593,
  [SMALL_STATE(3034)] = 119648,
  [SMALL_STATE(3035)] = 119725,
  [SMALL_STATE(3036)] = 119770,
  [SMALL_STATE(3037)] = 119835,
  [SMALL_STATE(3038)] = 119902,
  [SMALL_STATE(3039)] = 119969,
  [SMALL_STATE(3040)] = 120048,
  [SMALL_STATE(3041)] = 120105,
  [SMALL_STATE(3042)] = 120188,
  [SMALL_STATE(3043)] = 120249,
  [SMALL_STATE(3044)] = 120320,
  [SMALL_STATE(3045)] = 120395,
  [SMALL_STATE(3046)] = 120474,
  [SMALL_STATE(3047)] = 120553,
  [SMALL_STATE(3048)] = 120598,
  [SMALL_STATE(3049)] = 120677,
  [SMALL_STATE(3050)] = 120722,
  [SMALL_STATE(3051)] = 120767,
  [SMALL_STATE(3052)] = 120812,
  [SMALL_STATE(3053)] = 120857,
  [SMALL_STATE(3054)] = 120902,
  [SMALL_STATE(3055)] = 120947,
  [SMALL_STATE(3056)] = 120992,
  [SMALL_STATE(3057)] = 121037,
  [SMALL_STATE(3058)] = 121082,
  [SMALL_STATE(3059)] = 121127,
  [SMALL_STATE(3060)] = 121172,
  [SMALL_STATE(3061)] = 121217,
  [SMALL_STATE(3062)] = 121294,
  [SMALL_STATE(3063)] = 121375,
  [SMALL_STATE(3064)] = 121440,
  [SMALL_STATE(3065)] = 121507,
  [SMALL_STATE(3066)] = 121574,
  [SMALL_STATE(3067)] = 121653,
  [SMALL_STATE(3068)] = 121724,
  [SMALL_STATE(3069)] = 121799,
  [SMALL_STATE(3070)] = 121878,
  [SMALL_STATE(3071)] = 121957,
  [SMALL_STATE(3072)] = 122038,
  [SMALL_STATE(3073)] = 122117,
  [SMALL_STATE(3074)] = 122198,
  [SMALL_STATE(3075)] = 122279,
  [SMALL_STATE(3076)] = 122360,
  [SMALL_STATE(3077)] = 122405,
  [SMALL_STATE(3078)] = 122450,
  [SMALL_STATE(3079)] = 122495,
  [SMALL_STATE(3080)] = 122542,
  [SMALL_STATE(3081)] = 122587,
  [SMALL_STATE(3082)] = 122648,
  [SMALL_STATE(3083)] = 122693,
  [SMALL_STATE(3084)] = 122754,
  [SMALL_STATE(3085)] = 122799,
  [SMALL_STATE(3086)] = 122844,
  [SMALL_STATE(3087)] = 122891,
  [SMALL_STATE(3088)] = 122936,
  [SMALL_STATE(3089)] = 122981,
  [SMALL_STATE(3090)] = 123042,
  [SMALL_STATE(3091)] = 123087,
  [SMALL_STATE(3092)] = 123164,
  [SMALL_STATE(3093)] = 123209,
  [SMALL_STATE(3094)] = 123254,
  [SMALL_STATE(3095)] = 123299,
  [SMALL_STATE(3096)] = 123360,
  [SMALL_STATE(3097)] = 123405,
  [SMALL_STATE(3098)] = 123466,
  [SMALL_STATE(3099)] = 123511,
  [SMALL_STATE(3100)] = 123564,
  [SMALL_STATE(3101)] = 123609,
  [SMALL_STATE(3102)] = 123664,
  [SMALL_STATE(3103)] = 123721,
  [SMALL_STATE(3104)] = 123778,
  [SMALL_STATE(3105)] = 123839,
  [SMALL_STATE(3106)] = 123886,
  [SMALL_STATE(3107)] = 123935,
  [SMALL_STATE(3108)] = 123982,
  [SMALL_STATE(3109)] = 124033,
  [SMALL_STATE(3110)] = 124078,
  [SMALL_STATE(3111)] = 124125,
  [SMALL_STATE(3112)] = 124170,
  [SMALL_STATE(3113)] = 124215,
  [SMALL_STATE(3114)] = 124260,
  [SMALL_STATE(3115)] = 124317,
  [SMALL_STATE(3116)] = 124389,
  [SMALL_STATE(3117)] = 124437,
  [SMALL_STATE(3118)] = 124485,
  [SMALL_STATE(3119)] = 124531,
  [SMALL_STATE(3120)] = 124575,
  [SMALL_STATE(3121)] = 124635,
  [SMALL_STATE(3122)] = 124679,
  [SMALL_STATE(3123)] = 124723,
  [SMALL_STATE(3124)] = 124803,
  [SMALL_STATE(3125)] = 124847,
  [SMALL_STATE(3126)] = 124893,
  [SMALL_STATE(3127)] = 124953,
  [SMALL_STATE(3128)] = 124997,
  [SMALL_STATE(3129)] = 125041,
  [SMALL_STATE(3130)] = 125121,
  [SMALL_STATE(3131)] = 125169,
  [SMALL_STATE(3132)] = 125213,
  [SMALL_STATE(3133)] = 125293,
  [SMALL_STATE(3134)] = 125341,
  [SMALL_STATE(3135)] = 125389,
  [SMALL_STATE(3136)] = 125433,
  [SMALL_STATE(3137)] = 125477,
  [SMALL_STATE(3138)] = 125521,
  [SMALL_STATE(3139)] = 125565,
  [SMALL_STATE(3140)] = 125637,
  [SMALL_STATE(3141)] = 125681,
  [SMALL_STATE(3142)] = 125735,
  [SMALL_STATE(3143)] = 125809,
  [SMALL_STATE(3144)] = 125869,
  [SMALL_STATE(3145)] = 125935,
  [SMALL_STATE(3146)] = 126001,
  [SMALL_STATE(3147)] = 126077,
  [SMALL_STATE(3148)] = 126133,
  [SMALL_STATE(3149)] = 126189,
  [SMALL_STATE(3150)] = 126249,
  [SMALL_STATE(3151)] = 126319,
  [SMALL_STATE(3152)] = 126363,
  [SMALL_STATE(3153)] = 126439,
  [SMALL_STATE(3154)] = 126515,
  [SMALL_STATE(3155)] = 126591,
  [SMALL_STATE(3156)] = 126671,
  [SMALL_STATE(3157)] = 126731,
  [SMALL_STATE(3158)] = 126775,
  [SMALL_STATE(3159)] = 126835,
  [SMALL_STATE(3160)] = 126913,
  [SMALL_STATE(3161)] = 126991,
  [SMALL_STATE(3162)] = 127069,
  [SMALL_STATE(3163)] = 127113,
  [SMALL_STATE(3164)] = 127185,
  [SMALL_STATE(3165)] = 127229,
  [SMALL_STATE(3166)] = 127309,
  [SMALL_STATE(3167)] = 127353,
  [SMALL_STATE(3168)] = 127397,
  [SMALL_STATE(3169)] = 127445,
  [SMALL_STATE(3170)] = 127505,
  [SMALL_STATE(3171)] = 127565,
  [SMALL_STATE(3172)] = 127629,
  [SMALL_STATE(3173)] = 127684,
  [SMALL_STATE(3174)] = 127759,
  [SMALL_STATE(3175)] = 127834,
  [SMALL_STATE(3176)] = 127909,
  [SMALL_STATE(3177)] = 127970,
  [SMALL_STATE(3178)] = 128043,
  [SMALL_STATE(3179)] = 128118,
  [SMALL_STATE(3180)] = 128175,
  [SMALL_STATE(3181)] = 128248,
  [SMALL_STATE(3182)] = 128293,
  [SMALL_STATE(3183)] = 128368,
  [SMALL_STATE(3184)] = 128443,
  [SMALL_STATE(3185)] = 128490,
  [SMALL_STATE(3186)] = 128537,
  [SMALL_STATE(3187)] = 128584,
  [SMALL_STATE(3188)] = 128631,
  [SMALL_STATE(3189)] = 128676,
  [SMALL_STATE(3190)] = 128723,
  [SMALL_STATE(3191)] = 128770,
  [SMALL_STATE(3192)] = 128845,
  [SMALL_STATE(3193)] = 128922,
  [SMALL_STATE(3194)] = 128967,
  [SMALL_STATE(3195)] = 129042,
  [SMALL_STATE(3196)] = 129117,
  [SMALL_STATE(3197)] = 129192,
  [SMALL_STATE(3198)] = 129253,
  [SMALL_STATE(3199)] = 129298,
  [SMALL_STATE(3200)] = 129341,
  [SMALL_STATE(3201)] = 129416,
  [SMALL_STATE(3202)] = 129459,
  [SMALL_STATE(3203)] = 129502,
  [SMALL_STATE(3204)] = 129549,
  [SMALL_STATE(3205)] = 129624,
  [SMALL_STATE(3206)] = 129699,
  [SMALL_STATE(3207)] = 129774,
  [SMALL_STATE(3208)] = 129849,
  [SMALL_STATE(3209)] = 129924,
  [SMALL_STATE(3210)] = 129981,
  [SMALL_STATE(3211)] = 130036,
  [SMALL_STATE(3212)] = 130111,
  [SMALL_STATE(3213)] = 130188,
  [SMALL_STATE(3214)] = 130263,
  [SMALL_STATE(3215)] = 130340,
  [SMALL_STATE(3216)] = 130417,
  [SMALL_STATE(3217)] = 130462,
  [SMALL_STATE(3218)] = 130539,
  [SMALL_STATE(3219)] = 130586,
  [SMALL_STATE(3220)] = 130659,
  [SMALL_STATE(3221)] = 130702,
  [SMALL_STATE(3222)] = 130777,
  [SMALL_STATE(3223)] = 130854,
  [SMALL_STATE(3224)] = 130929,
  [SMALL_STATE(3225)] = 130971,
  [SMALL_STATE(3226)] = 131043,
  [SMALL_STATE(3227)] = 131085,
  [SMALL_STATE(3228)] = 131127,
  [SMALL_STATE(3229)] = 131187,
  [SMALL_STATE(3230)] = 131229,
  [SMALL_STATE(3231)] = 131271,
  [SMALL_STATE(3232)] = 131313,
  [SMALL_STATE(3233)] = 131355,
  [SMALL_STATE(3234)] = 131397,
  [SMALL_STATE(3235)] = 131469,
  [SMALL_STATE(3236)] = 131541,
  [SMALL_STATE(3237)] = 131587,
  [SMALL_STATE(3238)] = 131659,
  [SMALL_STATE(3239)] = 131731,
  [SMALL_STATE(3240)] = 131775,
  [SMALL_STATE(3241)] = 131835,
  [SMALL_STATE(3242)] = 131907,
  [SMALL_STATE(3243)] = 131979,
  [SMALL_STATE(3244)] = 132051,
  [SMALL_STATE(3245)] = 132123,
  [SMALL_STATE(3246)] = 132195,
  [SMALL_STATE(3247)] = 132267,
  [SMALL_STATE(3248)] = 132313,
  [SMALL_STATE(3249)] = 132357,
  [SMALL_STATE(3250)] = 132415,
  [SMALL_STATE(3251)] = 132487,
  [SMALL_STATE(3252)] = 132531,
  [SMALL_STATE(3253)] = 132573,
  [SMALL_STATE(3254)] = 132645,
  [SMALL_STATE(3255)] = 132695,
  [SMALL_STATE(3256)] = 132739,
  [SMALL_STATE(3257)] = 132789,
  [SMALL_STATE(3258)] = 132861,
  [SMALL_STATE(3259)] = 132911,
  [SMALL_STATE(3260)] = 132955,
  [SMALL_STATE(3261)] = 132997,
  [SMALL_STATE(3262)] = 133055,
  [SMALL_STATE(3263)] = 133099,
  [SMALL_STATE(3264)] = 133141,
  [SMALL_STATE(3265)] = 133193,
  [SMALL_STATE(3266)] = 133235,
  [SMALL_STATE(3267)] = 133307,
  [SMALL_STATE(3268)] = 133353,
  [SMALL_STATE(3269)] = 133421,
  [SMALL_STATE(3270)] = 133493,
  [SMALL_STATE(3271)] = 133565,
  [SMALL_STATE(3272)] = 133637,
  [SMALL_STATE(3273)] = 133682,
  [SMALL_STATE(3274)] = 133727,
  [SMALL_STATE(3275)] = 133774,
  [SMALL_STATE(3276)] = 133821,
  [SMALL_STATE(3277)] = 133868,
  [SMALL_STATE(3278)] = 133913,
  [SMALL_STATE(3279)] = 133958,
  [SMALL_STATE(3280)] = 134003,
  [SMALL_STATE(3281)] = 134054,
  [SMALL_STATE(3282)] = 134095,
  [SMALL_STATE(3283)] = 134140,
  [SMALL_STATE(3284)] = 134183,
  [SMALL_STATE(3285)] = 134228,
  [SMALL_STATE(3286)] = 134269,
  [SMALL_STATE(3287)] = 134320,
  [SMALL_STATE(3288)] = 134389,
  [SMALL_STATE(3289)] = 134432,
  [SMALL_STATE(3290)] = 134489,
  [SMALL_STATE(3291)] = 134546,
  [SMALL_STATE(3292)] = 134587,
  [SMALL_STATE(3293)] = 134658,
  [SMALL_STATE(3294)] = 134703,
  [SMALL_STATE(3295)] = 134748,
  [SMALL_STATE(3296)] = 134791,
  [SMALL_STATE(3297)] = 134834,
  [SMALL_STATE(3298)] = 134879,
  [SMALL_STATE(3299)] = 134920,
  [SMALL_STATE(3300)] = 134961,
  [SMALL_STATE(3301)] = 135008,
  [SMALL_STATE(3302)] = 135059,
  [SMALL_STATE(3303)] = 135100,
  [SMALL_STATE(3304)] = 135141,
  [SMALL_STATE(3305)] = 135182,
  [SMALL_STATE(3306)] = 135225,
  [SMALL_STATE(3307)] = 135270,
  [SMALL_STATE(3308)] = 135315,
  [SMALL_STATE(3309)] = 135358,
  [SMALL_STATE(3310)] = 135399,
  [SMALL_STATE(3311)] = 135444,
  [SMALL_STATE(3312)] = 135489,
  [SMALL_STATE(3313)] = 135530,
  [SMALL_STATE(3314)] = 135571,
  [SMALL_STATE(3315)] = 135637,
  [SMALL_STATE(3316)] = 135693,
  [SMALL_STATE(3317)] = 135761,
  [SMALL_STATE(3318)] = 135829,
  [SMALL_STATE(3319)] = 135883,
  [SMALL_STATE(3320)] = 135951,
  [SMALL_STATE(3321)] = 136019,
  [SMALL_STATE(3322)] = 136087,
  [SMALL_STATE(3323)] = 136155,
  [SMALL_STATE(3324)] = 136223,
  [SMALL_STATE(3325)] = 136289,
  [SMALL_STATE(3326)] = 136355,
  [SMALL_STATE(3327)] = 136399,
  [SMALL_STATE(3328)] = 136467,
  [SMALL_STATE(3329)] = 136533,
  [SMALL_STATE(3330)] = 136579,
  [SMALL_STATE(3331)] = 136645,
  [SMALL_STATE(3332)] = 136685,
  [SMALL_STATE(3333)] = 136753,
  [SMALL_STATE(3334)] = 136793,
  [SMALL_STATE(3335)] = 136859,
  [SMALL_STATE(3336)] = 136913,
  [SMALL_STATE(3337)] = 136981,
  [SMALL_STATE(3338)] = 137049,
  [SMALL_STATE(3339)] = 137117,
  [SMALL_STATE(3340)] = 137185,
  [SMALL_STATE(3341)] = 137253,
  [SMALL_STATE(3342)] = 137321,
  [SMALL_STATE(3343)] = 137389,
  [SMALL_STATE(3344)] = 137457,
  [SMALL_STATE(3345)] = 137525,
  [SMALL_STATE(3346)] = 137593,
  [SMALL_STATE(3347)] = 137647,
  [SMALL_STATE(3348)] = 137691,
  [SMALL_STATE(3349)] = 137733,
  [SMALL_STATE(3350)] = 137773,
  [SMALL_STATE(3351)] = 137815,
  [SMALL_STATE(3352)] = 137855,
  [SMALL_STATE(3353)] = 137895,
  [SMALL_STATE(3354)] = 137963,
  [SMALL_STATE(3355)] = 138031,
  [SMALL_STATE(3356)] = 138099,
  [SMALL_STATE(3357)] = 138167,
  [SMALL_STATE(3358)] = 138207,
  [SMALL_STATE(3359)] = 138247,
  [SMALL_STATE(3360)] = 138287,
  [SMALL_STATE(3361)] = 138355,
  [SMALL_STATE(3362)] = 138423,
  [SMALL_STATE(3363)] = 138491,
  [SMALL_STATE(3364)] = 138559,
  [SMALL_STATE(3365)] = 138627,
  [SMALL_STATE(3366)] = 138695,
  [SMALL_STATE(3367)] = 138763,
  [SMALL_STATE(3368)] = 138803,
  [SMALL_STATE(3369)] = 138855,
  [SMALL_STATE(3370)] = 138921,
  [SMALL_STATE(3371)] = 138989,
  [SMALL_STATE(3372)] = 139057,
  [SMALL_STATE(3373)] = 139125,
  [SMALL_STATE(3374)] = 139193,
  [SMALL_STATE(3375)] = 139261,
  [SMALL_STATE(3376)] = 139329,
  [SMALL_STATE(3377)] = 139395,
  [SMALL_STATE(3378)] = 139461,
  [SMALL_STATE(3379)] = 139529,
  [SMALL_STATE(3380)] = 139595,
  [SMALL_STATE(3381)] = 139663,
  [SMALL_STATE(3382)] = 139731,
  [SMALL_STATE(3383)] = 139799,
  [SMALL_STATE(3384)] = 139867,
  [SMALL_STATE(3385)] = 139935,
  [SMALL_STATE(3386)] = 140003,
  [SMALL_STATE(3387)] = 140071,
  [SMALL_STATE(3388)] = 140139,
  [SMALL_STATE(3389)] = 140207,
  [SMALL_STATE(3390)] = 140275,
  [SMALL_STATE(3391)] = 140343,
  [SMALL_STATE(3392)] = 140411,
  [SMALL_STATE(3393)] = 140479,
  [SMALL_STATE(3394)] = 140519,
  [SMALL_STATE(3395)] = 140587,
  [SMALL_STATE(3396)] = 140655,
  [SMALL_STATE(3397)] = 140709,
  [SMALL_STATE(3398)] = 140777,
  [SMALL_STATE(3399)] = 140845,
  [SMALL_STATE(3400)] = 140913,
  [SMALL_STATE(3401)] = 140981,
  [SMALL_STATE(3402)] = 141049,
  [SMALL_STATE(3403)] = 141117,
  [SMALL_STATE(3404)] = 141185,
  [SMALL_STATE(3405)] = 141253,
  [SMALL_STATE(3406)] = 141321,
  [SMALL_STATE(3407)] = 141387,
  [SMALL_STATE(3408)] = 141455,
  [SMALL_STATE(3409)] = 141523,
  [SMALL_STATE(3410)] = 141591,
  [SMALL_STATE(3411)] = 141631,
  [SMALL_STATE(3412)] = 141671,
  [SMALL_STATE(3413)] = 141713,
  [SMALL_STATE(3414)] = 141753,
  [SMALL_STATE(3415)] = 141793,
  [SMALL_STATE(3416)] = 141861,
  [SMALL_STATE(3417)] = 141905,
  [SMALL_STATE(3418)] = 141945,
  [SMALL_STATE(3419)] = 142013,
  [SMALL_STATE(3420)] = 142067,
  [SMALL_STATE(3421)] = 142135,
  [SMALL_STATE(3422)] = 142203,
  [SMALL_STATE(3423)] = 142269,
  [SMALL_STATE(3424)] = 142335,
  [SMALL_STATE(3425)] = 142375,
  [SMALL_STATE(3426)] = 142443,
  [SMALL_STATE(3427)] = 142511,
  [SMALL_STATE(3428)] = 142577,
  [SMALL_STATE(3429)] = 142643,
  [SMALL_STATE(3430)] = 142711,
  [SMALL_STATE(3431)] = 142755,
  [SMALL_STATE(3432)] = 142823,
  [SMALL_STATE(3433)] = 142891,
  [SMALL_STATE(3434)] = 142947,
  [SMALL_STATE(3435)] = 143015,
  [SMALL_STATE(3436)] = 143059,
  [SMALL_STATE(3437)] = 143127,
  [SMALL_STATE(3438)] = 143178,
  [SMALL_STATE(3439)] = 143229,
  [SMALL_STATE(3440)] = 143268,
  [SMALL_STATE(3441)] = 143313,
  [SMALL_STATE(3442)] = 143352,
  [SMALL_STATE(3443)] = 143393,
  [SMALL_STATE(3444)] = 143444,
  [SMALL_STATE(3445)] = 143495,
  [SMALL_STATE(3446)] = 143550,
  [SMALL_STATE(3447)] = 143589,
  [SMALL_STATE(3448)] = 143628,
  [SMALL_STATE(3449)] = 143679,
  [SMALL_STATE(3450)] = 143718,
  [SMALL_STATE(3451)] = 143757,
  [SMALL_STATE(3452)] = 143808,
  [SMALL_STATE(3453)] = 143859,
  [SMALL_STATE(3454)] = 143910,
  [SMALL_STATE(3455)] = 143949,
  [SMALL_STATE(3456)] = 144000,
  [SMALL_STATE(3457)] = 144051,
  [SMALL_STATE(3458)] = 144106,
  [SMALL_STATE(3459)] = 144155,
  [SMALL_STATE(3460)] = 144206,
  [SMALL_STATE(3461)] = 144257,
  [SMALL_STATE(3462)] = 144300,
  [SMALL_STATE(3463)] = 144339,
  [SMALL_STATE(3464)] = 144380,
  [SMALL_STATE(3465)] = 144431,
  [SMALL_STATE(3466)] = 144470,
  [SMALL_STATE(3467)] = 144521,
  [SMALL_STATE(3468)] = 144564,
  [SMALL_STATE(3469)] = 144603,
  [SMALL_STATE(3470)] = 144654,
  [SMALL_STATE(3471)] = 144705,
  [SMALL_STATE(3472)] = 144748,
  [SMALL_STATE(3473)] = 144787,
  [SMALL_STATE(3474)] = 144830,
  [SMALL_STATE(3475)] = 144881,
  [SMALL_STATE(3476)] = 144930,
  [SMALL_STATE(3477)] = 144977,
  [SMALL_STATE(3478)] = 145016,
  [SMALL_STATE(3479)] = 145063,
  [SMALL_STATE(3480)] = 145110,
  [SMALL_STATE(3481)] = 145149,
  [SMALL_STATE(3482)] = 145188,
  [SMALL_STATE(3483)] = 145239,
  [SMALL_STATE(3484)] = 145283,
  [SMALL_STATE(3485)] = 145321,
  [SMALL_STATE(3486)] = 145365,
  [SMALL_STATE(3487)] = 145403,
  [SMALL_STATE(3488)] = 145447,
  [SMALL_STATE(3489)] = 145491,
  [SMALL_STATE(3490)] = 145535,
  [SMALL_STATE(3491)] = 145579,
  [SMALL_STATE(3492)] = 145623,
  [SMALL_STATE(3493)] = 145667,
  [SMALL_STATE(3494)] = 145711,
  [SMALL_STATE(3495)] = 145749,
  [SMALL_STATE(3496)] = 145789,
  [SMALL_STATE(3497)] = 145827,
  [SMALL_STATE(3498)] = 145893,
  [SMALL_STATE(3499)] = 145937,
  [SMALL_STATE(3500)] = 146003,
  [SMALL_STATE(3501)] = 146045,
  [SMALL_STATE(3502)] = 146083,
  [SMALL_STATE(3503)] = 146149,
  [SMALL_STATE(3504)] = 146215,
  [SMALL_STATE(3505)] = 146257,
  [SMALL_STATE(3506)] = 146309,
  [SMALL_STATE(3507)] = 146357,
  [SMALL_STATE(3508)] = 146401,
  [SMALL_STATE(3509)] = 146445,
  [SMALL_STATE(3510)] = 146489,
  [SMALL_STATE(3511)] = 146533,
  [SMALL_STATE(3512)] = 146577,
  [SMALL_STATE(3513)] = 146621,
  [SMALL_STATE(3514)] = 146665,
  [SMALL_STATE(3515)] = 146709,
  [SMALL_STATE(3516)] = 146757,
  [SMALL_STATE(3517)] = 146823,
  [SMALL_STATE(3518)] = 146889,
  [SMALL_STATE(3519)] = 146931,
  [SMALL_STATE(3520)] = 146971,
  [SMALL_STATE(3521)] = 147015,
  [SMALL_STATE(3522)] = 147059,
  [SMALL_STATE(3523)] = 147101,
  [SMALL_STATE(3524)] = 147145,
  [SMALL_STATE(3525)] = 147187,
  [SMALL_STATE(3526)] = 147229,
  [SMALL_STATE(3527)] = 147271,
  [SMALL_STATE(3528)] = 147313,
  [SMALL_STATE(3529)] = 147365,
  [SMALL_STATE(3530)] = 147414,
  [SMALL_STATE(3531)] = 147477,
  [SMALL_STATE(3532)] = 147540,
  [SMALL_STATE(3533)] = 147585,
  [SMALL_STATE(3534)] = 147626,
  [SMALL_STATE(3535)] = 147671,
  [SMALL_STATE(3536)] = 147708,
  [SMALL_STATE(3537)] = 147747,
  [SMALL_STATE(3538)] = 147796,
  [SMALL_STATE(3539)] = 147841,
  [SMALL_STATE(3540)] = 147882,
  [SMALL_STATE(3541)] = 147919,
  [SMALL_STATE(3542)] = 147982,
  [SMALL_STATE(3543)] = 148023,
  [SMALL_STATE(3544)] = 148070,
  [SMALL_STATE(3545)] = 148115,
  [SMALL_STATE(3546)] = 148178,
  [SMALL_STATE(3547)] = 148223,
  [SMALL_STATE(3548)] = 148268,
  [SMALL_STATE(3549)] = 148305,
  [SMALL_STATE(3550)] = 148346,
  [SMALL_STATE(3551)] = 148387,
  [SMALL_STATE(3552)] = 148428,
  [SMALL_STATE(3553)] = 148473,
  [SMALL_STATE(3554)] = 148518,
  [SMALL_STATE(3555)] = 148581,
  [SMALL_STATE(3556)] = 148620,
  [SMALL_STATE(3557)] = 148665,
  [SMALL_STATE(3558)] = 148704,
  [SMALL_STATE(3559)] = 148767,
  [SMALL_STATE(3560)] = 148830,
  [SMALL_STATE(3561)] = 148893,
  [SMALL_STATE(3562)] = 148940,
  [SMALL_STATE(3563)] = 148985,
  [SMALL_STATE(3564)] = 149048,
  [SMALL_STATE(3565)] = 149085,
  [SMALL_STATE(3566)] = 149148,
  [SMALL_STATE(3567)] = 149211,
  [SMALL_STATE(3568)] = 149260,
  [SMALL_STATE(3569)] = 149305,
  [SMALL_STATE(3570)] = 149354,
  [SMALL_STATE(3571)] = 149391,
  [SMALL_STATE(3572)] = 149430,
  [SMALL_STATE(3573)] = 149475,
  [SMALL_STATE(3574)] = 149517,
  [SMALL_STATE(3575)] = 149577,
  [SMALL_STATE(3576)] = 149619,
  [SMALL_STATE(3577)] = 149661,
  [SMALL_STATE(3578)] = 149703,
  [SMALL_STATE(3579)] = 149745,
  [SMALL_STATE(3580)] = 149787,
  [SMALL_STATE(3581)] = 149829,
  [SMALL_STATE(3582)] = 149871,
  [SMALL_STATE(3583)] = 149931,
  [SMALL_STATE(3584)] = 149973,
  [SMALL_STATE(3585)] = 150015,
  [SMALL_STATE(3586)] = 150051,
  [SMALL_STATE(3587)] = 150091,
  [SMALL_STATE(3588)] = 150133,
  [SMALL_STATE(3589)] = 150193,
  [SMALL_STATE(3590)] = 150253,
  [SMALL_STATE(3591)] = 150289,
  [SMALL_STATE(3592)] = 150349,
  [SMALL_STATE(3593)] = 150385,
  [SMALL_STATE(3594)] = 150421,
  [SMALL_STATE(3595)] = 150463,
  [SMALL_STATE(3596)] = 150505,
  [SMALL_STATE(3597)] = 150547,
  [SMALL_STATE(3598)] = 150589,
  [SMALL_STATE(3599)] = 150631,
  [SMALL_STATE(3600)] = 150673,
  [SMALL_STATE(3601)] = 150715,
  [SMALL_STATE(3602)] = 150757,
  [SMALL_STATE(3603)] = 150799,
  [SMALL_STATE(3604)] = 150835,
  [SMALL_STATE(3605)] = 150877,
  [SMALL_STATE(3606)] = 150919,
  [SMALL_STATE(3607)] = 150979,
  [SMALL_STATE(3608)] = 151021,
  [SMALL_STATE(3609)] = 151063,
  [SMALL_STATE(3610)] = 151105,
  [SMALL_STATE(3611)] = 151147,
  [SMALL_STATE(3612)] = 151189,
  [SMALL_STATE(3613)] = 151225,
  [SMALL_STATE(3614)] = 151267,
  [SMALL_STATE(3615)] = 151303,
  [SMALL_STATE(3616)] = 151345,
  [SMALL_STATE(3617)] = 151387,
  [SMALL_STATE(3618)] = 151447,
  [SMALL_STATE(3619)] = 151489,
  [SMALL_STATE(3620)] = 151549,
  [SMALL_STATE(3621)] = 151609,
  [SMALL_STATE(3622)] = 151651,
  [SMALL_STATE(3623)] = 151693,
  [SMALL_STATE(3624)] = 151735,
  [SMALL_STATE(3625)] = 151777,
  [SMALL_STATE(3626)] = 151819,
  [SMALL_STATE(3627)] = 151861,
  [SMALL_STATE(3628)] = 151903,
  [SMALL_STATE(3629)] = 151963,
  [SMALL_STATE(3630)] = 152005,
  [SMALL_STATE(3631)] = 152047,
  [SMALL_STATE(3632)] = 152089,
  [SMALL_STATE(3633)] = 152131,
  [SMALL_STATE(3634)] = 152173,
  [SMALL_STATE(3635)] = 152209,
  [SMALL_STATE(3636)] = 152269,
  [SMALL_STATE(3637)] = 152329,
  [SMALL_STATE(3638)] = 152389,
  [SMALL_STATE(3639)] = 152425,
  [SMALL_STATE(3640)] = 152467,
  [SMALL_STATE(3641)] = 152509,
  [SMALL_STATE(3642)] = 152551,
  [SMALL_STATE(3643)] = 152611,
  [SMALL_STATE(3644)] = 152671,
  [SMALL_STATE(3645)] = 152713,
  [SMALL_STATE(3646)] = 152755,
  [SMALL_STATE(3647)] = 152797,
  [SMALL_STATE(3648)] = 152839,
  [SMALL_STATE(3649)] = 152881,
  [SMALL_STATE(3650)] = 152923,
  [SMALL_STATE(3651)] = 152983,
  [SMALL_STATE(3652)] = 153043,
  [SMALL_STATE(3653)] = 153079,
  [SMALL_STATE(3654)] = 153139,
  [SMALL_STATE(3655)] = 153199,
  [SMALL_STATE(3656)] = 153259,
  [SMALL_STATE(3657)] = 153319,
  [SMALL_STATE(3658)] = 153365,
  [SMALL_STATE(3659)] = 153425,
  [SMALL_STATE(3660)] = 153467,
  [SMALL_STATE(3661)] = 153527,
  [SMALL_STATE(3662)] = 153587,
  [SMALL_STATE(3663)] = 153633,
  [SMALL_STATE(3664)] = 153679,
  [SMALL_STATE(3665)] = 153725,
  [SMALL_STATE(3666)] = 153785,
  [SMALL_STATE(3667)] = 153831,
  [SMALL_STATE(3668)] = 153877,
  [SMALL_STATE(3669)] = 153919,
  [SMALL_STATE(3670)] = 153979,
  [SMALL_STATE(3671)] = 154015,
  [SMALL_STATE(3672)] = 154057,
  [SMALL_STATE(3673)] = 154103,
  [SMALL_STATE(3674)] = 154149,
  [SMALL_STATE(3675)] = 154195,
  [SMALL_STATE(3676)] = 154241,
  [SMALL_STATE(3677)] = 154301,
  [SMALL_STATE(3678)] = 154343,
  [SMALL_STATE(3679)] = 154389,
  [SMALL_STATE(3680)] = 154435,
  [SMALL_STATE(3681)] = 154481,
  [SMALL_STATE(3682)] = 154527,
  [SMALL_STATE(3683)] = 154573,
  [SMALL_STATE(3684)] = 154619,
  [SMALL_STATE(3685)] = 154665,
  [SMALL_STATE(3686)] = 154711,
  [SMALL_STATE(3687)] = 154771,
  [SMALL_STATE(3688)] = 154831,
  [SMALL_STATE(3689)] = 154891,
  [SMALL_STATE(3690)] = 154933,
  [SMALL_STATE(3691)] = 154975,
  [SMALL_STATE(3692)] = 155017,
  [SMALL_STATE(3693)] = 155077,
  [SMALL_STATE(3694)] = 155119,
  [SMALL_STATE(3695)] = 155161,
  [SMALL_STATE(3696)] = 155221,
  [SMALL_STATE(3697)] = 155263,
  [SMALL_STATE(3698)] = 155305,
  [SMALL_STATE(3699)] = 155365,
  [SMALL_STATE(3700)] = 155407,
  [SMALL_STATE(3701)] = 155449,
  [SMALL_STATE(3702)] = 155485,
  [SMALL_STATE(3703)] = 155527,
  [SMALL_STATE(3704)] = 155569,
  [SMALL_STATE(3705)] = 155611,
  [SMALL_STATE(3706)] = 155653,
  [SMALL_STATE(3707)] = 155689,
  [SMALL_STATE(3708)] = 155731,
  [SMALL_STATE(3709)] = 155773,
  [SMALL_STATE(3710)] = 155815,
  [SMALL_STATE(3711)] = 155857,
  [SMALL_STATE(3712)] = 155899,
  [SMALL_STATE(3713)] = 155941,
  [SMALL_STATE(3714)] = 155983,
  [SMALL_STATE(3715)] = 156043,
  [SMALL_STATE(3716)] = 156085,
  [SMALL_STATE(3717)] = 156127,
  [SMALL_STATE(3718)] = 156169,
  [SMALL_STATE(3719)] = 156211,
  [SMALL_STATE(3720)] = 156253,
  [SMALL_STATE(3721)] = 156295,
  [SMALL_STATE(3722)] = 156337,
  [SMALL_STATE(3723)] = 156373,
  [SMALL_STATE(3724)] = 156415,
  [SMALL_STATE(3725)] = 156457,
  [SMALL_STATE(3726)] = 156499,
  [SMALL_STATE(3727)] = 156541,
  [SMALL_STATE(3728)] = 156583,
  [SMALL_STATE(3729)] = 156643,
  [SMALL_STATE(3730)] = 156685,
  [SMALL_STATE(3731)] = 156745,
  [SMALL_STATE(3732)] = 156805,
  [SMALL_STATE(3733)] = 156865,
  [SMALL_STATE(3734)] = 156907,
  [SMALL_STATE(3735)] = 156949,
  [SMALL_STATE(3736)] = 156991,
  [SMALL_STATE(3737)] = 157033,
  [SMALL_STATE(3738)] = 157075,
  [SMALL_STATE(3739)] = 157135,
  [SMALL_STATE(3740)] = 157177,
  [SMALL_STATE(3741)] = 157237,
  [SMALL_STATE(3742)] = 157279,
  [SMALL_STATE(3743)] = 157339,
  [SMALL_STATE(3744)] = 157381,
  [SMALL_STATE(3745)] = 157423,
  [SMALL_STATE(3746)] = 157465,
  [SMALL_STATE(3747)] = 157507,
  [SMALL_STATE(3748)] = 157549,
  [SMALL_STATE(3749)] = 157591,
  [SMALL_STATE(3750)] = 157633,
  [SMALL_STATE(3751)] = 157675,
  [SMALL_STATE(3752)] = 157717,
  [SMALL_STATE(3753)] = 157759,
  [SMALL_STATE(3754)] = 157801,
  [SMALL_STATE(3755)] = 157843,
  [SMALL_STATE(3756)] = 157885,
  [SMALL_STATE(3757)] = 157927,
  [SMALL_STATE(3758)] = 157963,
  [SMALL_STATE(3759)] = 158005,
  [SMALL_STATE(3760)] = 158047,
  [SMALL_STATE(3761)] = 158107,
  [SMALL_STATE(3762)] = 158167,
  [SMALL_STATE(3763)] = 158207,
  [SMALL_STATE(3764)] = 158267,
  [SMALL_STATE(3765)] = 158305,
  [SMALL_STATE(3766)] = 158347,
  [SMALL_STATE(3767)] = 158389,
  [SMALL_STATE(3768)] = 158449,
  [SMALL_STATE(3769)] = 158509,
  [SMALL_STATE(3770)] = 158569,
  [SMALL_STATE(3771)] = 158629,
  [SMALL_STATE(3772)] = 158669,
  [SMALL_STATE(3773)] = 158729,
  [SMALL_STATE(3774)] = 158789,
  [SMALL_STATE(3775)] = 158849,
  [SMALL_STATE(3776)] = 158909,
  [SMALL_STATE(3777)] = 158951,
  [SMALL_STATE(3778)] = 159008,
  [SMALL_STATE(3779)] = 159043,
  [SMALL_STATE(3780)] = 159100,
  [SMALL_STATE(3781)] = 159141,
  [SMALL_STATE(3782)] = 159182,
  [SMALL_STATE(3783)] = 159239,
  [SMALL_STATE(3784)] = 159280,
  [SMALL_STATE(3785)] = 159315,
  [SMALL_STATE(3786)] = 159356,
  [SMALL_STATE(3787)] = 159391,
  [SMALL_STATE(3788)] = 159432,
  [SMALL_STATE(3789)] = 159489,
  [SMALL_STATE(3790)] = 159530,
  [SMALL_STATE(3791)] = 159571,
  [SMALL_STATE(3792)] = 159612,
  [SMALL_STATE(3793)] = 159651,
  [SMALL_STATE(3794)] = 159690,
  [SMALL_STATE(3795)] = 159731,
  [SMALL_STATE(3796)] = 159770,
  [SMALL_STATE(3797)] = 159811,
  [SMALL_STATE(3798)] = 159852,
  [SMALL_STATE(3799)] = 159909,
  [SMALL_STATE(3800)] = 159948,
  [SMALL_STATE(3801)] = 159989,
  [SMALL_STATE(3802)] = 160030,
  [SMALL_STATE(3803)] = 160071,
  [SMALL_STATE(3804)] = 160120,
  [SMALL_STATE(3805)] = 160161,
  [SMALL_STATE(3806)] = 160202,
  [SMALL_STATE(3807)] = 160243,
  [SMALL_STATE(3808)] = 160282,
  [SMALL_STATE(3809)] = 160339,
  [SMALL_STATE(3810)] = 160380,
  [SMALL_STATE(3811)] = 160421,
  [SMALL_STATE(3812)] = 160462,
  [SMALL_STATE(3813)] = 160503,
  [SMALL_STATE(3814)] = 160544,
  [SMALL_STATE(3815)] = 160585,
  [SMALL_STATE(3816)] = 160626,
  [SMALL_STATE(3817)] = 160683,
  [SMALL_STATE(3818)] = 160740,
  [SMALL_STATE(3819)] = 160775,
  [SMALL_STATE(3820)] = 160816,
  [SMALL_STATE(3821)] = 160851,
  [SMALL_STATE(3822)] = 160888,
  [SMALL_STATE(3823)] = 160929,
  [SMALL_STATE(3824)] = 160970,
  [SMALL_STATE(3825)] = 161011,
  [SMALL_STATE(3826)] = 161068,
  [SMALL_STATE(3827)] = 161109,
  [SMALL_STATE(3828)] = 161150,
  [SMALL_STATE(3829)] = 161185,
  [SMALL_STATE(3830)] = 161220,
  [SMALL_STATE(3831)] = 161255,
  [SMALL_STATE(3832)] = 161296,
  [SMALL_STATE(3833)] = 161337,
  [SMALL_STATE(3834)] = 161376,
  [SMALL_STATE(3835)] = 161417,
  [SMALL_STATE(3836)] = 161456,
  [SMALL_STATE(3837)] = 161513,
  [SMALL_STATE(3838)] = 161554,
  [SMALL_STATE(3839)] = 161589,
  [SMALL_STATE(3840)] = 161624,
  [SMALL_STATE(3841)] = 161659,
  [SMALL_STATE(3842)] = 161694,
  [SMALL_STATE(3843)] = 161735,
  [SMALL_STATE(3844)] = 161784,
  [SMALL_STATE(3845)] = 161831,
  [SMALL_STATE(3846)] = 161888,
  [SMALL_STATE(3847)] = 161935,
  [SMALL_STATE(3848)] = 161992,
  [SMALL_STATE(3849)] = 162049,
  [SMALL_STATE(3850)] = 162090,
  [SMALL_STATE(3851)] = 162130,
  [SMALL_STATE(3852)] = 162184,
  [SMALL_STATE(3853)] = 162238,
  [SMALL_STATE(3854)] = 162292,
  [SMALL_STATE(3855)] = 162346,
  [SMALL_STATE(3856)] = 162400,
  [SMALL_STATE(3857)] = 162454,
  [SMALL_STATE(3858)] = 162492,
  [SMALL_STATE(3859)] = 162530,
  [SMALL_STATE(3860)] = 162584,
  [SMALL_STATE(3861)] = 162638,
  [SMALL_STATE(3862)] = 162692,
  [SMALL_STATE(3863)] = 162732,
  [SMALL_STATE(3864)] = 162786,
  [SMALL_STATE(3865)] = 162840,
  [SMALL_STATE(3866)] = 162894,
  [SMALL_STATE(3867)] = 162950,
  [SMALL_STATE(3868)] = 163004,
  [SMALL_STATE(3869)] = 163060,
  [SMALL_STATE(3870)] = 163114,
  [SMALL_STATE(3871)] = 163152,
  [SMALL_STATE(3872)] = 163206,
  [SMALL_STATE(3873)] = 163260,
  [SMALL_STATE(3874)] = 163314,
  [SMALL_STATE(3875)] = 163354,
  [SMALL_STATE(3876)] = 163394,
  [SMALL_STATE(3877)] = 163448,
  [SMALL_STATE(3878)] = 163502,
  [SMALL_STATE(3879)] = 163556,
  [SMALL_STATE(3880)] = 163610,
  [SMALL_STATE(3881)] = 163664,
  [SMALL_STATE(3882)] = 163702,
  [SMALL_STATE(3883)] = 163756,
  [SMALL_STATE(3884)] = 163810,
  [SMALL_STATE(3885)] = 163864,
  [SMALL_STATE(3886)] = 163904,
  [SMALL_STATE(3887)] = 163958,
  [SMALL_STATE(3888)] = 164012,
  [SMALL_STATE(3889)] = 164066,
  [SMALL_STATE(3890)] = 164120,
  [SMALL_STATE(3891)] = 164156,
  [SMALL_STATE(3892)] = 164210,
  [SMALL_STATE(3893)] = 164248,
  [SMALL_STATE(3894)] = 164288,
  [SMALL_STATE(3895)] = 164342,
  [SMALL_STATE(3896)] = 164396,
  [SMALL_STATE(3897)] = 164432,
  [SMALL_STATE(3898)] = 164468,
  [SMALL_STATE(3899)] = 164508,
  [SMALL_STATE(3900)] = 164548,
  [SMALL_STATE(3901)] = 164584,
  [SMALL_STATE(3902)] = 164622,
  [SMALL_STATE(3903)] = 164658,
  [SMALL_STATE(3904)] = 164698,
  [SMALL_STATE(3905)] = 164738,
  [SMALL_STATE(3906)] = 164792,
  [SMALL_STATE(3907)] = 164832,
  [SMALL_STATE(3908)] = 164870,
  [SMALL_STATE(3909)] = 164908,
  [SMALL_STATE(3910)] = 164950,
  [SMALL_STATE(3911)] = 165004,
  [SMALL_STATE(3912)] = 165040,
  [SMALL_STATE(3913)] = 165096,
  [SMALL_STATE(3914)] = 165150,
  [SMALL_STATE(3915)] = 165204,
  [SMALL_STATE(3916)] = 165258,
  [SMALL_STATE(3917)] = 165312,
  [SMALL_STATE(3918)] = 165368,
  [SMALL_STATE(3919)] = 165424,
  [SMALL_STATE(3920)] = 165478,
  [SMALL_STATE(3921)] = 165532,
  [SMALL_STATE(3922)] = 165586,
  [SMALL_STATE(3923)] = 165640,
  [SMALL_STATE(3924)] = 165678,
  [SMALL_STATE(3925)] = 165716,
  [SMALL_STATE(3926)] = 165770,
  [SMALL_STATE(3927)] = 165806,
  [SMALL_STATE(3928)] = 165860,
  [SMALL_STATE(3929)] = 165898,
  [SMALL_STATE(3930)] = 165952,
  [SMALL_STATE(3931)] = 166006,
  [SMALL_STATE(3932)] = 166062,
  [SMALL_STATE(3933)] = 166116,
  [SMALL_STATE(3934)] = 166170,
  [SMALL_STATE(3935)] = 166208,
  [SMALL_STATE(3936)] = 166262,
  [SMALL_STATE(3937)] = 166300,
  [SMALL_STATE(3938)] = 166354,
  [SMALL_STATE(3939)] = 166408,
  [SMALL_STATE(3940)] = 166462,
  [SMALL_STATE(3941)] = 166516,
  [SMALL_STATE(3942)] = 166570,
  [SMALL_STATE(3943)] = 166624,
  [SMALL_STATE(3944)] = 166678,
  [SMALL_STATE(3945)] = 166734,
  [SMALL_STATE(3946)] = 166788,
  [SMALL_STATE(3947)] = 166842,
  [SMALL_STATE(3948)] = 166896,
  [SMALL_STATE(3949)] = 166950,
  [SMALL_STATE(3950)] = 167004,
  [SMALL_STATE(3951)] = 167058,
  [SMALL_STATE(3952)] = 167112,
  [SMALL_STATE(3953)] = 167166,
  [SMALL_STATE(3954)] = 167222,
  [SMALL_STATE(3955)] = 167276,
  [SMALL_STATE(3956)] = 167330,
  [SMALL_STATE(3957)] = 167384,
  [SMALL_STATE(3958)] = 167438,
  [SMALL_STATE(3959)] = 167494,
  [SMALL_STATE(3960)] = 167548,
  [SMALL_STATE(3961)] = 167586,
  [SMALL_STATE(3962)] = 167640,
  [SMALL_STATE(3963)] = 167694,
  [SMALL_STATE(3964)] = 167730,
  [SMALL_STATE(3965)] = 167784,
  [SMALL_STATE(3966)] = 167822,
  [SMALL_STATE(3967)] = 167876,
  [SMALL_STATE(3968)] = 167914,
  [SMALL_STATE(3969)] = 167968,
  [SMALL_STATE(3970)] = 168022,
  [SMALL_STATE(3971)] = 168076,
  [SMALL_STATE(3972)] = 168130,
  [SMALL_STATE(3973)] = 168184,
  [SMALL_STATE(3974)] = 168222,
  [SMALL_STATE(3975)] = 168276,
  [SMALL_STATE(3976)] = 168330,
  [SMALL_STATE(3977)] = 168384,
  [SMALL_STATE(3978)] = 168422,
  [SMALL_STATE(3979)] = 168460,
  [SMALL_STATE(3980)] = 168514,
  [SMALL_STATE(3981)] = 168568,
  [SMALL_STATE(3982)] = 168622,
  [SMALL_STATE(3983)] = 168678,
  [SMALL_STATE(3984)] = 168732,
  [SMALL_STATE(3985)] = 168786,
  [SMALL_STATE(3986)] = 168840,
  [SMALL_STATE(3987)] = 168894,
  [SMALL_STATE(3988)] = 168932,
  [SMALL_STATE(3989)] = 168986,
  [SMALL_STATE(3990)] = 169040,
  [SMALL_STATE(3991)] = 169094,
  [SMALL_STATE(3992)] = 169132,
  [SMALL_STATE(3993)] = 169186,
  [SMALL_STATE(3994)] = 169238,
  [SMALL_STATE(3995)] = 169280,
  [SMALL_STATE(3996)] = 169322,
  [SMALL_STATE(3997)] = 169376,
  [SMALL_STATE(3998)] = 169430,
  [SMALL_STATE(3999)] = 169486,
  [SMALL_STATE(4000)] = 169540,
  [SMALL_STATE(4001)] = 169594,
  [SMALL_STATE(4002)] = 169648,
  [SMALL_STATE(4003)] = 169686,
  [SMALL_STATE(4004)] = 169740,
  [SMALL_STATE(4005)] = 169794,
  [SMALL_STATE(4006)] = 169836,
  [SMALL_STATE(4007)] = 169874,
  [SMALL_STATE(4008)] = 169928,
  [SMALL_STATE(4009)] = 169982,
  [SMALL_STATE(4010)] = 170036,
  [SMALL_STATE(4011)] = 170072,
  [SMALL_STATE(4012)] = 170126,
  [SMALL_STATE(4013)] = 170180,
  [SMALL_STATE(4014)] = 170234,
  [SMALL_STATE(4015)] = 170288,
  [SMALL_STATE(4016)] = 170342,
  [SMALL_STATE(4017)] = 170396,
  [SMALL_STATE(4018)] = 170450,
  [SMALL_STATE(4019)] = 170504,
  [SMALL_STATE(4020)] = 170560,
  [SMALL_STATE(4021)] = 170614,
  [SMALL_STATE(4022)] = 170652,
  [SMALL_STATE(4023)] = 170706,
  [SMALL_STATE(4024)] = 170760,
  [SMALL_STATE(4025)] = 170814,
  [SMALL_STATE(4026)] = 170868,
  [SMALL_STATE(4027)] = 170922,
  [SMALL_STATE(4028)] = 170976,
  [SMALL_STATE(4029)] = 171030,
  [SMALL_STATE(4030)] = 171084,
  [SMALL_STATE(4031)] = 171118,
  [SMALL_STATE(4032)] = 171172,
  [SMALL_STATE(4033)] = 171226,
  [SMALL_STATE(4034)] = 171266,
  [SMALL_STATE(4035)] = 171320,
  [SMALL_STATE(4036)] = 171374,
  [SMALL_STATE(4037)] = 171428,
  [SMALL_STATE(4038)] = 171482,
  [SMALL_STATE(4039)] = 171517,
  [SMALL_STATE(4040)] = 171550,
  [SMALL_STATE(4041)] = 171599,
  [SMALL_STATE(4042)] = 171634,
  [SMALL_STATE(4043)] = 171683,
  [SMALL_STATE(4044)] = 171744,
  [SMALL_STATE(4045)] = 171795,
  [SMALL_STATE(4046)] = 171828,
  [SMALL_STATE(4047)] = 171889,
  [SMALL_STATE(4048)] = 171950,
  [SMALL_STATE(4049)] = 171983,
  [SMALL_STATE(4050)] = 172034,
  [SMALL_STATE(4051)] = 172073,
  [SMALL_STATE(4052)] = 172106,
  [SMALL_STATE(4053)] = 172159,
  [SMALL_STATE(4054)] = 172192,
  [SMALL_STATE(4055)] = 172241,
  [SMALL_STATE(4056)] = 172274,
  [SMALL_STATE(4057)] = 172327,
  [SMALL_STATE(4058)] = 172377,
  [SMALL_STATE(4059)] = 172425,
  [SMALL_STATE(4060)] = 172457,
  [SMALL_STATE(4061)] = 172489,
  [SMALL_STATE(4062)] = 172537,
  [SMALL_STATE(4063)] = 172569,
  [SMALL_STATE(4064)] = 172601,
  [SMALL_STATE(4065)] = 172649,
  [SMALL_STATE(4066)] = 172681,
  [SMALL_STATE(4067)] = 172729,
  [SMALL_STATE(4068)] = 172777,
  [SMALL_STATE(4069)] = 172825,
  [SMALL_STATE(4070)] = 172873,
  [SMALL_STATE(4071)] = 172923,
  [SMALL_STATE(4072)] = 172959,
  [SMALL_STATE(4073)] = 172991,
  [SMALL_STATE(4074)] = 173039,
  [SMALL_STATE(4075)] = 173087,
  [SMALL_STATE(4076)] = 173135,
  [SMALL_STATE(4077)] = 173183,
  [SMALL_STATE(4078)] = 173231,
  [SMALL_STATE(4079)] = 173263,
  [SMALL_STATE(4080)] = 173311,
  [SMALL_STATE(4081)] = 173343,
  [SMALL_STATE(4082)] = 173391,
  [SMALL_STATE(4083)] = 173427,
  [SMALL_STATE(4084)] = 173475,
  [SMALL_STATE(4085)] = 173523,
  [SMALL_STATE(4086)] = 173571,
  [SMALL_STATE(4087)] = 173619,
  [SMALL_STATE(4088)] = 173667,
  [SMALL_STATE(4089)] = 173699,
  [SMALL_STATE(4090)] = 173731,
  [SMALL_STATE(4091)] = 173779,
  [SMALL_STATE(4092)] = 173811,
  [SMALL_STATE(4093)] = 173859,
  [SMALL_STATE(4094)] = 173909,
  [SMALL_STATE(4095)] = 173957,
  [SMALL_STATE(4096)] = 174007,
  [SMALL_STATE(4097)] = 174039,
  [SMALL_STATE(4098)] = 174087,
  [SMALL_STATE(4099)] = 174135,
  [SMALL_STATE(4100)] = 174183,
  [SMALL_STATE(4101)] = 174215,
  [SMALL_STATE(4102)] = 174263,
  [SMALL_STATE(4103)] = 174311,
  [SMALL_STATE(4104)] = 174359,
  [SMALL_STATE(4105)] = 174407,
  [SMALL_STATE(4106)] = 174439,
  [SMALL_STATE(4107)] = 174487,
  [SMALL_STATE(4108)] = 174519,
  [SMALL_STATE(4109)] = 174551,
  [SMALL_STATE(4110)] = 174599,
  [SMALL_STATE(4111)] = 174631,
  [SMALL_STATE(4112)] = 174663,
  [SMALL_STATE(4113)] = 174711,
  [SMALL_STATE(4114)] = 174743,
  [SMALL_STATE(4115)] = 174791,
  [SMALL_STATE(4116)] = 174841,
  [SMALL_STATE(4117)] = 174889,
  [SMALL_STATE(4118)] = 174921,
  [SMALL_STATE(4119)] = 174971,
  [SMALL_STATE(4120)] = 175019,
  [SMALL_STATE(4121)] = 175067,
  [SMALL_STATE(4122)] = 175108,
  [SMALL_STATE(4123)] = 175149,
  [SMALL_STATE(4124)] = 175190,
  [SMALL_STATE(4125)] = 175221,
  [SMALL_STATE(4126)] = 175262,
  [SMALL_STATE(4127)] = 175303,
  [SMALL_STATE(4128)] = 175344,
  [SMALL_STATE(4129)] = 175377,
  [SMALL_STATE(4130)] = 175410,
  [SMALL_STATE(4131)] = 175443,
  [SMALL_STATE(4132)] = 175476,
  [SMALL_STATE(4133)] = 175509,
  [SMALL_STATE(4134)] = 175550,
  [SMALL_STATE(4135)] = 175605,
  [SMALL_STATE(4136)] = 175646,
  [SMALL_STATE(4137)] = 175687,
  [SMALL_STATE(4138)] = 175742,
  [SMALL_STATE(4139)] = 175781,
  [SMALL_STATE(4140)] = 175814,
  [SMALL_STATE(4141)] = 175847,
  [SMALL_STATE(4142)] = 175880,
  [SMALL_STATE(4143)] = 175913,
  [SMALL_STATE(4144)] = 175944,
  [SMALL_STATE(4145)] = 175985,
  [SMALL_STATE(4146)] = 176026,
  [SMALL_STATE(4147)] = 176059,
  [SMALL_STATE(4148)] = 176092,
  [SMALL_STATE(4149)] = 176133,
  [SMALL_STATE(4150)] = 176166,
  [SMALL_STATE(4151)] = 176197,
  [SMALL_STATE(4152)] = 176238,
  [SMALL_STATE(4153)] = 176269,
  [SMALL_STATE(4154)] = 176310,
  [SMALL_STATE(4155)] = 176341,
  [SMALL_STATE(4156)] = 176374,
  [SMALL_STATE(4157)] = 176429,
  [SMALL_STATE(4158)] = 176470,
  [SMALL_STATE(4159)] = 176511,
  [SMALL_STATE(4160)] = 176542,
  [SMALL_STATE(4161)] = 176573,
  [SMALL_STATE(4162)] = 176604,
  [SMALL_STATE(4163)] = 176645,
  [SMALL_STATE(4164)] = 176676,
  [SMALL_STATE(4165)] = 176707,
  [SMALL_STATE(4166)] = 176748,
  [SMALL_STATE(4167)] = 176781,
  [SMALL_STATE(4168)] = 176814,
  [SMALL_STATE(4169)] = 176845,
  [SMALL_STATE(4170)] = 176878,
  [SMALL_STATE(4171)] = 176911,
  [SMALL_STATE(4172)] = 176944,
  [SMALL_STATE(4173)] = 176977,
  [SMALL_STATE(4174)] = 177016,
  [SMALL_STATE(4175)] = 177062,
  [SMALL_STATE(4176)] = 177114,
  [SMALL_STATE(4177)] = 177166,
  [SMALL_STATE(4178)] = 177196,
  [SMALL_STATE(4179)] = 177240,
  [SMALL_STATE(4180)] = 177282,
  [SMALL_STATE(4181)] = 177328,
  [SMALL_STATE(4182)] = 177358,
  [SMALL_STATE(4183)] = 177410,
  [SMALL_STATE(4184)] = 177456,
  [SMALL_STATE(4185)] = 177508,
  [SMALL_STATE(4186)] = 177552,
  [SMALL_STATE(4187)] = 177598,
  [SMALL_STATE(4188)] = 177650,
  [SMALL_STATE(4189)] = 177696,
  [SMALL_STATE(4190)] = 177748,
  [SMALL_STATE(4191)] = 177794,
  [SMALL_STATE(4192)] = 177840,
  [SMALL_STATE(4193)] = 177886,
  [SMALL_STATE(4194)] = 177938,
  [SMALL_STATE(4195)] = 177976,
  [SMALL_STATE(4196)] = 178022,
  [SMALL_STATE(4197)] = 178068,
  [SMALL_STATE(4198)] = 178114,
  [SMALL_STATE(4199)] = 178160,
  [SMALL_STATE(4200)] = 178204,
  [SMALL_STATE(4201)] = 178250,
  [SMALL_STATE(4202)] = 178282,
  [SMALL_STATE(4203)] = 178334,
  [SMALL_STATE(4204)] = 178378,
  [SMALL_STATE(4205)] = 178422,
  [SMALL_STATE(4206)] = 178452,
  [SMALL_STATE(4207)] = 178498,
  [SMALL_STATE(4208)] = 178528,
  [SMALL_STATE(4209)] = 178574,
  [SMALL_STATE(4210)] = 178626,
  [SMALL_STATE(4211)] = 178662,
  [SMALL_STATE(4212)] = 178692,
  [SMALL_STATE(4213)] = 178719,
  [SMALL_STATE(4214)] = 178768,
  [SMALL_STATE(4215)] = 178799,
  [SMALL_STATE(4216)] = 178828,
  [SMALL_STATE(4217)] = 178875,
  [SMALL_STATE(4218)] = 178922,
  [SMALL_STATE(4219)] = 178965,
  [SMALL_STATE(4220)] = 179008,
  [SMALL_STATE(4221)] = 179051,
  [SMALL_STATE(4222)] = 179100,
  [SMALL_STATE(4223)] = 179127,
  [SMALL_STATE(4224)] = 179170,
  [SMALL_STATE(4225)] = 179199,
  [SMALL_STATE(4226)] = 179226,
  [SMALL_STATE(4227)] = 179273,
  [SMALL_STATE(4228)] = 179316,
  [SMALL_STATE(4229)] = 179359,
  [SMALL_STATE(4230)] = 179388,
  [SMALL_STATE(4231)] = 179419,
  [SMALL_STATE(4232)] = 179466,
  [SMALL_STATE(4233)] = 179515,
  [SMALL_STATE(4234)] = 179554,
  [SMALL_STATE(4235)] = 179585,
  [SMALL_STATE(4236)] = 179614,
  [SMALL_STATE(4237)] = 179645,
  [SMALL_STATE(4238)] = 179692,
  [SMALL_STATE(4239)] = 179741,
  [SMALL_STATE(4240)] = 179772,
  [SMALL_STATE(4241)] = 179799,
  [SMALL_STATE(4242)] = 179846,
  [SMALL_STATE(4243)] = 179880,
  [SMALL_STATE(4244)] = 179922,
  [SMALL_STATE(4245)] = 179958,
  [SMALL_STATE(4246)] = 179984,
  [SMALL_STATE(4247)] = 180028,
  [SMALL_STATE(4248)] = 180056,
  [SMALL_STATE(4249)] = 180100,
  [SMALL_STATE(4250)] = 180130,
  [SMALL_STATE(4251)] = 180170,
  [SMALL_STATE(4252)] = 180196,
  [SMALL_STATE(4253)] = 180238,
  [SMALL_STATE(4254)] = 180272,
  [SMALL_STATE(4255)] = 180300,
  [SMALL_STATE(4256)] = 180342,
  [SMALL_STATE(4257)] = 180384,
  [SMALL_STATE(4258)] = 180426,
  [SMALL_STATE(4259)] = 180452,
  [SMALL_STATE(4260)] = 180486,
  [SMALL_STATE(4261)] = 180512,
  [SMALL_STATE(4262)] = 180546,
  [SMALL_STATE(4263)] = 180588,
  [SMALL_STATE(4264)] = 180616,
  [SMALL_STATE(4265)] = 180650,
  [SMALL_STATE(4266)] = 180676,
  [SMALL_STATE(4267)] = 180704,
  [SMALL_STATE(4268)] = 180738,
  [SMALL_STATE(4269)] = 180770,
  [SMALL_STATE(4270)] = 180812,
  [SMALL_STATE(4271)] = 180854,
  [SMALL_STATE(4272)] = 180888,
  [SMALL_STATE(4273)] = 180926,
  [SMALL_STATE(4274)] = 180962,
  [SMALL_STATE(4275)] = 180988,
  [SMALL_STATE(4276)] = 181032,
  [SMALL_STATE(4277)] = 181060,
  [SMALL_STATE(4278)] = 181102,
  [SMALL_STATE(4279)] = 181138,
  [SMALL_STATE(4280)] = 181180,
  [SMALL_STATE(4281)] = 181222,
  [SMALL_STATE(4282)] = 181264,
  [SMALL_STATE(4283)] = 181306,
  [SMALL_STATE(4284)] = 181348,
  [SMALL_STATE(4285)] = 181390,
  [SMALL_STATE(4286)] = 181420,
  [SMALL_STATE(4287)] = 181462,
  [SMALL_STATE(4288)] = 181496,
  [SMALL_STATE(4289)] = 181540,
  [SMALL_STATE(4290)] = 181582,
  [SMALL_STATE(4291)] = 181616,
  [SMALL_STATE(4292)] = 181660,
  [SMALL_STATE(4293)] = 181702,
  [SMALL_STATE(4294)] = 181740,
  [SMALL_STATE(4295)] = 181774,
  [SMALL_STATE(4296)] = 181816,
  [SMALL_STATE(4297)] = 181850,
  [SMALL_STATE(4298)] = 181892,
  [SMALL_STATE(4299)] = 181936,
  [SMALL_STATE(4300)] = 181978,
  [SMALL_STATE(4301)] = 182004,
  [SMALL_STATE(4302)] = 182038,
  [SMALL_STATE(4303)] = 182077,
  [SMALL_STATE(4304)] = 182114,
  [SMALL_STATE(4305)] = 182153,
  [SMALL_STATE(4306)] = 182184,
  [SMALL_STATE(4307)] = 182223,
  [SMALL_STATE(4308)] = 182250,
  [SMALL_STATE(4309)] = 182277,
  [SMALL_STATE(4310)] = 182304,
  [SMALL_STATE(4311)] = 182335,
  [SMALL_STATE(4312)] = 182366,
  [SMALL_STATE(4313)] = 182409,
  [SMALL_STATE(4314)] = 182448,
  [SMALL_STATE(4315)] = 182479,
  [SMALL_STATE(4316)] = 182510,
  [SMALL_STATE(4317)] = 182537,
  [SMALL_STATE(4318)] = 182568,
  [SMALL_STATE(4319)] = 182607,
  [SMALL_STATE(4320)] = 182638,
  [SMALL_STATE(4321)] = 182669,
  [SMALL_STATE(4322)] = 182708,
  [SMALL_STATE(4323)] = 182735,
  [SMALL_STATE(4324)] = 182762,
  [SMALL_STATE(4325)] = 182789,
  [SMALL_STATE(4326)] = 182820,
  [SMALL_STATE(4327)] = 182859,
  [SMALL_STATE(4328)] = 182890,
  [SMALL_STATE(4329)] = 182921,
  [SMALL_STATE(4330)] = 182948,
  [SMALL_STATE(4331)] = 182975,
  [SMALL_STATE(4332)] = 183006,
  [SMALL_STATE(4333)] = 183037,
  [SMALL_STATE(4334)] = 183064,
  [SMALL_STATE(4335)] = 183093,
  [SMALL_STATE(4336)] = 183124,
  [SMALL_STATE(4337)] = 183151,
  [SMALL_STATE(4338)] = 183178,
  [SMALL_STATE(4339)] = 183209,
  [SMALL_STATE(4340)] = 183248,
  [SMALL_STATE(4341)] = 183275,
  [SMALL_STATE(4342)] = 183302,
  [SMALL_STATE(4343)] = 183341,
  [SMALL_STATE(4344)] = 183380,
  [SMALL_STATE(4345)] = 183411,
  [SMALL_STATE(4346)] = 183438,
  [SMALL_STATE(4347)] = 183477,
  [SMALL_STATE(4348)] = 183508,
  [SMALL_STATE(4349)] = 183539,
  [SMALL_STATE(4350)] = 183570,
  [SMALL_STATE(4351)] = 183601,
  [SMALL_STATE(4352)] = 183628,
  [SMALL_STATE(4353)] = 183659,
  [SMALL_STATE(4354)] = 183690,
  [SMALL_STATE(4355)] = 183719,
  [SMALL_STATE(4356)] = 183746,
  [SMALL_STATE(4357)] = 183785,
  [SMALL_STATE(4358)] = 183816,
  [SMALL_STATE(4359)] = 183855,
  [SMALL_STATE(4360)] = 183886,
  [SMALL_STATE(4361)] = 183917,
  [SMALL_STATE(4362)] = 183948,
  [SMALL_STATE(4363)] = 183987,
  [SMALL_STATE(4364)] = 184018,
  [SMALL_STATE(4365)] = 184049,
  [SMALL_STATE(4366)] = 184080,
  [SMALL_STATE(4367)] = 184119,
  [SMALL_STATE(4368)] = 184146,
  [SMALL_STATE(4369)] = 184173,
  [SMALL_STATE(4370)] = 184212,
  [SMALL_STATE(4371)] = 184243,
  [SMALL_STATE(4372)] = 184274,
  [SMALL_STATE(4373)] = 184305,
  [SMALL_STATE(4374)] = 184336,
  [SMALL_STATE(4375)] = 184363,
  [SMALL_STATE(4376)] = 184390,
  [SMALL_STATE(4377)] = 184421,
  [SMALL_STATE(4378)] = 184464,
  [SMALL_STATE(4379)] = 184495,
  [SMALL_STATE(4380)] = 184526,
  [SMALL_STATE(4381)] = 184557,
  [SMALL_STATE(4382)] = 184588,
  [SMALL_STATE(4383)] = 184619,
  [SMALL_STATE(4384)] = 184658,
  [SMALL_STATE(4385)] = 184689,
  [SMALL_STATE(4386)] = 184720,
  [SMALL_STATE(4387)] = 184751,
  [SMALL_STATE(4388)] = 184778,
  [SMALL_STATE(4389)] = 184809,
  [SMALL_STATE(4390)] = 184840,
  [SMALL_STATE(4391)] = 184871,
  [SMALL_STATE(4392)] = 184902,
  [SMALL_STATE(4393)] = 184933,
  [SMALL_STATE(4394)] = 184972,
  [SMALL_STATE(4395)] = 185013,
  [SMALL_STATE(4396)] = 185044,
  [SMALL_STATE(4397)] = 185075,
  [SMALL_STATE(4398)] = 185114,
  [SMALL_STATE(4399)] = 185145,
  [SMALL_STATE(4400)] = 185176,
  [SMALL_STATE(4401)] = 185207,
  [SMALL_STATE(4402)] = 185238,
  [SMALL_STATE(4403)] = 185269,
  [SMALL_STATE(4404)] = 185296,
  [SMALL_STATE(4405)] = 185327,
  [SMALL_STATE(4406)] = 185354,
  [SMALL_STATE(4407)] = 185385,
  [SMALL_STATE(4408)] = 185412,
  [SMALL_STATE(4409)] = 185443,
  [SMALL_STATE(4410)] = 185470,
  [SMALL_STATE(4411)] = 185501,
  [SMALL_STATE(4412)] = 185532,
  [SMALL_STATE(4413)] = 185559,
  [SMALL_STATE(4414)] = 185590,
  [SMALL_STATE(4415)] = 185617,
  [SMALL_STATE(4416)] = 185644,
  [SMALL_STATE(4417)] = 185683,
  [SMALL_STATE(4418)] = 185714,
  [SMALL_STATE(4419)] = 185741,
  [SMALL_STATE(4420)] = 185772,
  [SMALL_STATE(4421)] = 185811,
  [SMALL_STATE(4422)] = 185842,
  [SMALL_STATE(4423)] = 185869,
  [SMALL_STATE(4424)] = 185900,
  [SMALL_STATE(4425)] = 185927,
  [SMALL_STATE(4426)] = 185954,
  [SMALL_STATE(4427)] = 185985,
  [SMALL_STATE(4428)] = 186024,
  [SMALL_STATE(4429)] = 186051,
  [SMALL_STATE(4430)] = 186082,
  [SMALL_STATE(4431)] = 186109,
  [SMALL_STATE(4432)] = 186136,
  [SMALL_STATE(4433)] = 186167,
  [SMALL_STATE(4434)] = 186198,
  [SMALL_STATE(4435)] = 186229,
  [SMALL_STATE(4436)] = 186256,
  [SMALL_STATE(4437)] = 186283,
  [SMALL_STATE(4438)] = 186314,
  [SMALL_STATE(4439)] = 186345,
  [SMALL_STATE(4440)] = 186371,
  [SMALL_STATE(4441)] = 186397,
  [SMALL_STATE(4442)] = 186427,
  [SMALL_STATE(4443)] = 186453,
  [SMALL_STATE(4444)] = 186479,
  [SMALL_STATE(4445)] = 186505,
  [SMALL_STATE(4446)] = 186535,
  [SMALL_STATE(4447)] = 186561,
  [SMALL_STATE(4448)] = 186587,
  [SMALL_STATE(4449)] = 186617,
  [SMALL_STATE(4450)] = 186643,
  [SMALL_STATE(4451)] = 186673,
  [SMALL_STATE(4452)] = 186701,
  [SMALL_STATE(4453)] = 186727,
  [SMALL_STATE(4454)] = 186753,
  [SMALL_STATE(4455)] = 186777,
  [SMALL_STATE(4456)] = 186803,
  [SMALL_STATE(4457)] = 186843,
  [SMALL_STATE(4458)] = 186869,
  [SMALL_STATE(4459)] = 186895,
  [SMALL_STATE(4460)] = 186921,
  [SMALL_STATE(4461)] = 186951,
  [SMALL_STATE(4462)] = 186993,
  [SMALL_STATE(4463)] = 187017,
  [SMALL_STATE(4464)] = 187049,
  [SMALL_STATE(4465)] = 187075,
  [SMALL_STATE(4466)] = 187115,
  [SMALL_STATE(4467)] = 187139,
  [SMALL_STATE(4468)] = 187169,
  [SMALL_STATE(4469)] = 187209,
  [SMALL_STATE(4470)] = 187251,
  [SMALL_STATE(4471)] = 187281,
  [SMALL_STATE(4472)] = 187307,
  [SMALL_STATE(4473)] = 187333,
  [SMALL_STATE(4474)] = 187359,
  [SMALL_STATE(4475)] = 187389,
  [SMALL_STATE(4476)] = 187429,
  [SMALL_STATE(4477)] = 187455,
  [SMALL_STATE(4478)] = 187487,
  [SMALL_STATE(4479)] = 187517,
  [SMALL_STATE(4480)] = 187543,
  [SMALL_STATE(4481)] = 187569,
  [SMALL_STATE(4482)] = 187595,
  [SMALL_STATE(4483)] = 187627,
  [SMALL_STATE(4484)] = 187667,
  [SMALL_STATE(4485)] = 187693,
  [SMALL_STATE(4486)] = 187719,
  [SMALL_STATE(4487)] = 187745,
  [SMALL_STATE(4488)] = 187775,
  [SMALL_STATE(4489)] = 187801,
  [SMALL_STATE(4490)] = 187841,
  [SMALL_STATE(4491)] = 187867,
  [SMALL_STATE(4492)] = 187893,
  [SMALL_STATE(4493)] = 187919,
  [SMALL_STATE(4494)] = 187943,
  [SMALL_STATE(4495)] = 187980,
  [SMALL_STATE(4496)] = 188003,
  [SMALL_STATE(4497)] = 188040,
  [SMALL_STATE(4498)] = 188063,
  [SMALL_STATE(4499)] = 188086,
  [SMALL_STATE(4500)] = 188113,
  [SMALL_STATE(4501)] = 188136,
  [SMALL_STATE(4502)] = 188169,
  [SMALL_STATE(4503)] = 188192,
  [SMALL_STATE(4504)] = 188219,
  [SMALL_STATE(4505)] = 188256,
  [SMALL_STATE(4506)] = 188293,
  [SMALL_STATE(4507)] = 188324,
  [SMALL_STATE(4508)] = 188365,
  [SMALL_STATE(4509)] = 188390,
  [SMALL_STATE(4510)] = 188427,
  [SMALL_STATE(4511)] = 188452,
  [SMALL_STATE(4512)] = 188489,
  [SMALL_STATE(4513)] = 188512,
  [SMALL_STATE(4514)] = 188535,
  [SMALL_STATE(4515)] = 188562,
  [SMALL_STATE(4516)] = 188599,
  [SMALL_STATE(4517)] = 188632,
  [SMALL_STATE(4518)] = 188671,
  [SMALL_STATE(4519)] = 188708,
  [SMALL_STATE(4520)] = 188749,
  [SMALL_STATE(4521)] = 188786,
  [SMALL_STATE(4522)] = 188809,
  [SMALL_STATE(4523)] = 188838,
  [SMALL_STATE(4524)] = 188875,
  [SMALL_STATE(4525)] = 188902,
  [SMALL_STATE(4526)] = 188925,
  [SMALL_STATE(4527)] = 188954,
  [SMALL_STATE(4528)] = 188977,
  [SMALL_STATE(4529)] = 189014,
  [SMALL_STATE(4530)] = 189037,
  [SMALL_STATE(4531)] = 189074,
  [SMALL_STATE(4532)] = 189099,
  [SMALL_STATE(4533)] = 189136,
  [SMALL_STATE(4534)] = 189173,
  [SMALL_STATE(4535)] = 189202,
  [SMALL_STATE(4536)] = 189241,
  [SMALL_STATE(4537)] = 189270,
  [SMALL_STATE(4538)] = 189295,
  [SMALL_STATE(4539)] = 189318,
  [SMALL_STATE(4540)] = 189348,
  [SMALL_STATE(4541)] = 189370,
  [SMALL_STATE(4542)] = 189400,
  [SMALL_STATE(4543)] = 189424,
  [SMALL_STATE(4544)] = 189448,
  [SMALL_STATE(4545)] = 189484,
  [SMALL_STATE(4546)] = 189514,
  [SMALL_STATE(4547)] = 189538,
  [SMALL_STATE(4548)] = 189560,
  [SMALL_STATE(4549)] = 189584,
  [SMALL_STATE(4550)] = 189614,
  [SMALL_STATE(4551)] = 189650,
  [SMALL_STATE(4552)] = 189680,
  [SMALL_STATE(4553)] = 189718,
  [SMALL_STATE(4554)] = 189742,
  [SMALL_STATE(4555)] = 189780,
  [SMALL_STATE(4556)] = 189810,
  [SMALL_STATE(4557)] = 189834,
  [SMALL_STATE(4558)] = 189872,
  [SMALL_STATE(4559)] = 189902,
  [SMALL_STATE(4560)] = 189926,
  [SMALL_STATE(4561)] = 189950,
  [SMALL_STATE(4562)] = 189988,
  [SMALL_STATE(4563)] = 190012,
  [SMALL_STATE(4564)] = 190048,
  [SMALL_STATE(4565)] = 190072,
  [SMALL_STATE(4566)] = 190096,
  [SMALL_STATE(4567)] = 190134,
  [SMALL_STATE(4568)] = 190172,
  [SMALL_STATE(4569)] = 190210,
  [SMALL_STATE(4570)] = 190240,
  [SMALL_STATE(4571)] = 190264,
  [SMALL_STATE(4572)] = 190292,
  [SMALL_STATE(4573)] = 190316,
  [SMALL_STATE(4574)] = 190338,
  [SMALL_STATE(4575)] = 190360,
  [SMALL_STATE(4576)] = 190384,
  [SMALL_STATE(4577)] = 190422,
  [SMALL_STATE(4578)] = 190452,
  [SMALL_STATE(4579)] = 190482,
  [SMALL_STATE(4580)] = 190506,
  [SMALL_STATE(4581)] = 190530,
  [SMALL_STATE(4582)] = 190554,
  [SMALL_STATE(4583)] = 190580,
  [SMALL_STATE(4584)] = 190610,
  [SMALL_STATE(4585)] = 190634,
  [SMALL_STATE(4586)] = 190658,
  [SMALL_STATE(4587)] = 190696,
  [SMALL_STATE(4588)] = 190720,
  [SMALL_STATE(4589)] = 190744,
  [SMALL_STATE(4590)] = 190766,
  [SMALL_STATE(4591)] = 190796,
  [SMALL_STATE(4592)] = 190826,
  [SMALL_STATE(4593)] = 190850,
  [SMALL_STATE(4594)] = 190880,
  [SMALL_STATE(4595)] = 190904,
  [SMALL_STATE(4596)] = 190934,
  [SMALL_STATE(4597)] = 190964,
  [SMALL_STATE(4598)] = 190994,
  [SMALL_STATE(4599)] = 191032,
  [SMALL_STATE(4600)] = 191056,
  [SMALL_STATE(4601)] = 191080,
  [SMALL_STATE(4602)] = 191118,
  [SMALL_STATE(4603)] = 191142,
  [SMALL_STATE(4604)] = 191178,
  [SMALL_STATE(4605)] = 191208,
  [SMALL_STATE(4606)] = 191232,
  [SMALL_STATE(4607)] = 191262,
  [SMALL_STATE(4608)] = 191290,
  [SMALL_STATE(4609)] = 191320,
  [SMALL_STATE(4610)] = 191344,
  [SMALL_STATE(4611)] = 191368,
  [SMALL_STATE(4612)] = 191392,
  [SMALL_STATE(4613)] = 191416,
  [SMALL_STATE(4614)] = 191446,
  [SMALL_STATE(4615)] = 191470,
  [SMALL_STATE(4616)] = 191498,
  [SMALL_STATE(4617)] = 191522,
  [SMALL_STATE(4618)] = 191544,
  [SMALL_STATE(4619)] = 191568,
  [SMALL_STATE(4620)] = 191599,
  [SMALL_STATE(4621)] = 191630,
  [SMALL_STATE(4622)] = 191653,
  [SMALL_STATE(4623)] = 191684,
  [SMALL_STATE(4624)] = 191705,
  [SMALL_STATE(4625)] = 191726,
  [SMALL_STATE(4626)] = 191761,
  [SMALL_STATE(4627)] = 191782,
  [SMALL_STATE(4628)] = 191803,
  [SMALL_STATE(4629)] = 191830,
  [SMALL_STATE(4630)] = 191861,
  [SMALL_STATE(4631)] = 191896,
  [SMALL_STATE(4632)] = 191927,
  [SMALL_STATE(4633)] = 191954,
  [SMALL_STATE(4634)] = 191987,
  [SMALL_STATE(4635)] = 192018,
  [SMALL_STATE(4636)] = 192051,
  [SMALL_STATE(4637)] = 192078,
  [SMALL_STATE(4638)] = 192109,
  [SMALL_STATE(4639)] = 192132,
  [SMALL_STATE(4640)] = 192155,
  [SMALL_STATE(4641)] = 192190,
  [SMALL_STATE(4642)] = 192217,
  [SMALL_STATE(4643)] = 192242,
  [SMALL_STATE(4644)] = 192263,
  [SMALL_STATE(4645)] = 192296,
  [SMALL_STATE(4646)] = 192327,
  [SMALL_STATE(4647)] = 192360,
  [SMALL_STATE(4648)] = 192391,
  [SMALL_STATE(4649)] = 192422,
  [SMALL_STATE(4650)] = 192443,
  [SMALL_STATE(4651)] = 192476,
  [SMALL_STATE(4652)] = 192501,
  [SMALL_STATE(4653)] = 192532,
  [SMALL_STATE(4654)] = 192567,
  [SMALL_STATE(4655)] = 192590,
  [SMALL_STATE(4656)] = 192621,
  [SMALL_STATE(4657)] = 192654,
  [SMALL_STATE(4658)] = 192687,
  [SMALL_STATE(4659)] = 192720,
  [SMALL_STATE(4660)] = 192753,
  [SMALL_STATE(4661)] = 192776,
  [SMALL_STATE(4662)] = 192807,
  [SMALL_STATE(4663)] = 192838,
  [SMALL_STATE(4664)] = 192869,
  [SMALL_STATE(4665)] = 192900,
  [SMALL_STATE(4666)] = 192933,
  [SMALL_STATE(4667)] = 192966,
  [SMALL_STATE(4668)] = 192993,
  [SMALL_STATE(4669)] = 193026,
  [SMALL_STATE(4670)] = 193059,
  [SMALL_STATE(4671)] = 193084,
  [SMALL_STATE(4672)] = 193115,
  [SMALL_STATE(4673)] = 193150,
  [SMALL_STATE(4674)] = 193181,
  [SMALL_STATE(4675)] = 193206,
  [SMALL_STATE(4676)] = 193239,
  [SMALL_STATE(4677)] = 193272,
  [SMALL_STATE(4678)] = 193305,
  [SMALL_STATE(4679)] = 193338,
  [SMALL_STATE(4680)] = 193371,
  [SMALL_STATE(4681)] = 193396,
  [SMALL_STATE(4682)] = 193429,
  [SMALL_STATE(4683)] = 193449,
  [SMALL_STATE(4684)] = 193479,
  [SMALL_STATE(4685)] = 193503,
  [SMALL_STATE(4686)] = 193523,
  [SMALL_STATE(4687)] = 193543,
  [SMALL_STATE(4688)] = 193575,
  [SMALL_STATE(4689)] = 193607,
  [SMALL_STATE(4690)] = 193639,
  [SMALL_STATE(4691)] = 193671,
  [SMALL_STATE(4692)] = 193699,
  [SMALL_STATE(4693)] = 193731,
  [SMALL_STATE(4694)] = 193753,
  [SMALL_STATE(4695)] = 193785,
  [SMALL_STATE(4696)] = 193817,
  [SMALL_STATE(4697)] = 193841,
  [SMALL_STATE(4698)] = 193873,
  [SMALL_STATE(4699)] = 193905,
  [SMALL_STATE(4700)] = 193927,
  [SMALL_STATE(4701)] = 193951,
  [SMALL_STATE(4702)] = 193981,
  [SMALL_STATE(4703)] = 194001,
  [SMALL_STATE(4704)] = 194029,
  [SMALL_STATE(4705)] = 194061,
  [SMALL_STATE(4706)] = 194093,
  [SMALL_STATE(4707)] = 194125,
  [SMALL_STATE(4708)] = 194157,
  [SMALL_STATE(4709)] = 194185,
  [SMALL_STATE(4710)] = 194217,
  [SMALL_STATE(4711)] = 194249,
  [SMALL_STATE(4712)] = 194273,
  [SMALL_STATE(4713)] = 194295,
  [SMALL_STATE(4714)] = 194321,
  [SMALL_STATE(4715)] = 194343,
  [SMALL_STATE(4716)] = 194375,
  [SMALL_STATE(4717)] = 194407,
  [SMALL_STATE(4718)] = 194429,
  [SMALL_STATE(4719)] = 194461,
  [SMALL_STATE(4720)] = 194493,
  [SMALL_STATE(4721)] = 194521,
  [SMALL_STATE(4722)] = 194553,
  [SMALL_STATE(4723)] = 194585,
  [SMALL_STATE(4724)] = 194617,
  [SMALL_STATE(4725)] = 194649,
  [SMALL_STATE(4726)] = 194673,
  [SMALL_STATE(4727)] = 194705,
  [SMALL_STATE(4728)] = 194737,
  [SMALL_STATE(4729)] = 194761,
  [SMALL_STATE(4730)] = 194793,
  [SMALL_STATE(4731)] = 194815,
  [SMALL_STATE(4732)] = 194845,
  [SMALL_STATE(4733)] = 194877,
  [SMALL_STATE(4734)] = 194909,
  [SMALL_STATE(4735)] = 194935,
  [SMALL_STATE(4736)] = 194965,
  [SMALL_STATE(4737)] = 194997,
  [SMALL_STATE(4738)] = 195021,
  [SMALL_STATE(4739)] = 195053,
  [SMALL_STATE(4740)] = 195077,
  [SMALL_STATE(4741)] = 195109,
  [SMALL_STATE(4742)] = 195141,
  [SMALL_STATE(4743)] = 195173,
  [SMALL_STATE(4744)] = 195205,
  [SMALL_STATE(4745)] = 195227,
  [SMALL_STATE(4746)] = 195259,
  [SMALL_STATE(4747)] = 195291,
  [SMALL_STATE(4748)] = 195311,
  [SMALL_STATE(4749)] = 195343,
  [SMALL_STATE(4750)] = 195371,
  [SMALL_STATE(4751)] = 195403,
  [SMALL_STATE(4752)] = 195435,
  [SMALL_STATE(4753)] = 195467,
  [SMALL_STATE(4754)] = 195499,
  [SMALL_STATE(4755)] = 195531,
  [SMALL_STATE(4756)] = 195563,
  [SMALL_STATE(4757)] = 195589,
  [SMALL_STATE(4758)] = 195619,
  [SMALL_STATE(4759)] = 195651,
  [SMALL_STATE(4760)] = 195683,
  [SMALL_STATE(4761)] = 195705,
  [SMALL_STATE(4762)] = 195737,
  [SMALL_STATE(4763)] = 195769,
  [SMALL_STATE(4764)] = 195801,
  [SMALL_STATE(4765)] = 195823,
  [SMALL_STATE(4766)] = 195855,
  [SMALL_STATE(4767)] = 195887,
  [SMALL_STATE(4768)] = 195919,
  [SMALL_STATE(4769)] = 195951,
  [SMALL_STATE(4770)] = 195983,
  [SMALL_STATE(4771)] = 196015,
  [SMALL_STATE(4772)] = 196047,
  [SMALL_STATE(4773)] = 196079,
  [SMALL_STATE(4774)] = 196111,
  [SMALL_STATE(4775)] = 196143,
  [SMALL_STATE(4776)] = 196175,
  [SMALL_STATE(4777)] = 196207,
  [SMALL_STATE(4778)] = 196229,
  [SMALL_STATE(4779)] = 196261,
  [SMALL_STATE(4780)] = 196293,
  [SMALL_STATE(4781)] = 196317,
  [SMALL_STATE(4782)] = 196349,
  [SMALL_STATE(4783)] = 196381,
  [SMALL_STATE(4784)] = 196413,
  [SMALL_STATE(4785)] = 196445,
  [SMALL_STATE(4786)] = 196477,
  [SMALL_STATE(4787)] = 196499,
  [SMALL_STATE(4788)] = 196521,
  [SMALL_STATE(4789)] = 196545,
  [SMALL_STATE(4790)] = 196577,
  [SMALL_STATE(4791)] = 196603,
  [SMALL_STATE(4792)] = 196625,
  [SMALL_STATE(4793)] = 196647,
  [SMALL_STATE(4794)] = 196679,
  [SMALL_STATE(4795)] = 196711,
  [SMALL_STATE(4796)] = 196733,
  [SMALL_STATE(4797)] = 196753,
  [SMALL_STATE(4798)] = 196785,
  [SMALL_STATE(4799)] = 196809,
  [SMALL_STATE(4800)] = 196831,
  [SMALL_STATE(4801)] = 196851,
  [SMALL_STATE(4802)] = 196875,
  [SMALL_STATE(4803)] = 196907,
  [SMALL_STATE(4804)] = 196929,
  [SMALL_STATE(4805)] = 196949,
  [SMALL_STATE(4806)] = 196981,
  [SMALL_STATE(4807)] = 197013,
  [SMALL_STATE(4808)] = 197037,
  [SMALL_STATE(4809)] = 197057,
  [SMALL_STATE(4810)] = 197077,
  [SMALL_STATE(4811)] = 197101,
  [SMALL_STATE(4812)] = 197133,
  [SMALL_STATE(4813)] = 197157,
  [SMALL_STATE(4814)] = 197187,
  [SMALL_STATE(4815)] = 197209,
  [SMALL_STATE(4816)] = 197230,
  [SMALL_STATE(4817)] = 197259,
  [SMALL_STATE(4818)] = 197280,
  [SMALL_STATE(4819)] = 197309,
  [SMALL_STATE(4820)] = 197330,
  [SMALL_STATE(4821)] = 197359,
  [SMALL_STATE(4822)] = 197388,
  [SMALL_STATE(4823)] = 197415,
  [SMALL_STATE(4824)] = 197436,
  [SMALL_STATE(4825)] = 197465,
  [SMALL_STATE(4826)] = 197492,
  [SMALL_STATE(4827)] = 197513,
  [SMALL_STATE(4828)] = 197534,
  [SMALL_STATE(4829)] = 197555,
  [SMALL_STATE(4830)] = 197582,
  [SMALL_STATE(4831)] = 197603,
  [SMALL_STATE(4832)] = 197624,
  [SMALL_STATE(4833)] = 197645,
  [SMALL_STATE(4834)] = 197666,
  [SMALL_STATE(4835)] = 197689,
  [SMALL_STATE(4836)] = 197710,
  [SMALL_STATE(4837)] = 197731,
  [SMALL_STATE(4838)] = 197752,
  [SMALL_STATE(4839)] = 197773,
  [SMALL_STATE(4840)] = 197794,
  [SMALL_STATE(4841)] = 197815,
  [SMALL_STATE(4842)] = 197844,
  [SMALL_STATE(4843)] = 197871,
  [SMALL_STATE(4844)] = 197892,
  [SMALL_STATE(4845)] = 197917,
  [SMALL_STATE(4846)] = 197946,
  [SMALL_STATE(4847)] = 197967,
  [SMALL_STATE(4848)] = 197988,
  [SMALL_STATE(4849)] = 198009,
  [SMALL_STATE(4850)] = 198036,
  [SMALL_STATE(4851)] = 198057,
  [SMALL_STATE(4852)] = 198084,
  [SMALL_STATE(4853)] = 198107,
  [SMALL_STATE(4854)] = 198128,
  [SMALL_STATE(4855)] = 198157,
  [SMALL_STATE(4856)] = 198184,
  [SMALL_STATE(4857)] = 198205,
  [SMALL_STATE(4858)] = 198226,
  [SMALL_STATE(4859)] = 198247,
  [SMALL_STATE(4860)] = 198268,
  [SMALL_STATE(4861)] = 198293,
  [SMALL_STATE(4862)] = 198322,
  [SMALL_STATE(4863)] = 198345,
  [SMALL_STATE(4864)] = 198366,
  [SMALL_STATE(4865)] = 198393,
  [SMALL_STATE(4866)] = 198414,
  [SMALL_STATE(4867)] = 198441,
  [SMALL_STATE(4868)] = 198464,
  [SMALL_STATE(4869)] = 198491,
  [SMALL_STATE(4870)] = 198520,
  [SMALL_STATE(4871)] = 198549,
  [SMALL_STATE(4872)] = 198578,
  [SMALL_STATE(4873)] = 198599,
  [SMALL_STATE(4874)] = 198622,
  [SMALL_STATE(4875)] = 198643,
  [SMALL_STATE(4876)] = 198668,
  [SMALL_STATE(4877)] = 198697,
  [SMALL_STATE(4878)] = 198718,
  [SMALL_STATE(4879)] = 198739,
  [SMALL_STATE(4880)] = 198760,
  [SMALL_STATE(4881)] = 198781,
  [SMALL_STATE(4882)] = 198802,
  [SMALL_STATE(4883)] = 198831,
  [SMALL_STATE(4884)] = 198852,
  [SMALL_STATE(4885)] = 198873,
  [SMALL_STATE(4886)] = 198894,
  [SMALL_STATE(4887)] = 198915,
  [SMALL_STATE(4888)] = 198944,
  [SMALL_STATE(4889)] = 198971,
  [SMALL_STATE(4890)] = 198994,
  [SMALL_STATE(4891)] = 199021,
  [SMALL_STATE(4892)] = 199048,
  [SMALL_STATE(4893)] = 199075,
  [SMALL_STATE(4894)] = 199102,
  [SMALL_STATE(4895)] = 199123,
  [SMALL_STATE(4896)] = 199150,
  [SMALL_STATE(4897)] = 199179,
  [SMALL_STATE(4898)] = 199200,
  [SMALL_STATE(4899)] = 199227,
  [SMALL_STATE(4900)] = 199248,
  [SMALL_STATE(4901)] = 199277,
  [SMALL_STATE(4902)] = 199298,
  [SMALL_STATE(4903)] = 199327,
  [SMALL_STATE(4904)] = 199356,
  [SMALL_STATE(4905)] = 199385,
  [SMALL_STATE(4906)] = 199406,
  [SMALL_STATE(4907)] = 199431,
  [SMALL_STATE(4908)] = 199460,
  [SMALL_STATE(4909)] = 199487,
  [SMALL_STATE(4910)] = 199510,
  [SMALL_STATE(4911)] = 199539,
  [SMALL_STATE(4912)] = 199560,
  [SMALL_STATE(4913)] = 199589,
  [SMALL_STATE(4914)] = 199612,
  [SMALL_STATE(4915)] = 199633,
  [SMALL_STATE(4916)] = 199654,
  [SMALL_STATE(4917)] = 199675,
  [SMALL_STATE(4918)] = 199696,
  [SMALL_STATE(4919)] = 199717,
  [SMALL_STATE(4920)] = 199746,
  [SMALL_STATE(4921)] = 199767,
  [SMALL_STATE(4922)] = 199788,
  [SMALL_STATE(4923)] = 199809,
  [SMALL_STATE(4924)] = 199838,
  [SMALL_STATE(4925)] = 199865,
  [SMALL_STATE(4926)] = 199894,
  [SMALL_STATE(4927)] = 199923,
  [SMALL_STATE(4928)] = 199952,
  [SMALL_STATE(4929)] = 199979,
  [SMALL_STATE(4930)] = 200006,
  [SMALL_STATE(4931)] = 200027,
  [SMALL_STATE(4932)] = 200056,
  [SMALL_STATE(4933)] = 200083,
  [SMALL_STATE(4934)] = 200112,
  [SMALL_STATE(4935)] = 200135,
  [SMALL_STATE(4936)] = 200164,
  [SMALL_STATE(4937)] = 200193,
  [SMALL_STATE(4938)] = 200214,
  [SMALL_STATE(4939)] = 200235,
  [SMALL_STATE(4940)] = 200264,
  [SMALL_STATE(4941)] = 200285,
  [SMALL_STATE(4942)] = 200312,
  [SMALL_STATE(4943)] = 200333,
  [SMALL_STATE(4944)] = 200360,
  [SMALL_STATE(4945)] = 200381,
  [SMALL_STATE(4946)] = 200410,
  [SMALL_STATE(4947)] = 200437,
  [SMALL_STATE(4948)] = 200458,
  [SMALL_STATE(4949)] = 200485,
  [SMALL_STATE(4950)] = 200512,
  [SMALL_STATE(4951)] = 200539,
  [SMALL_STATE(4952)] = 200566,
  [SMALL_STATE(4953)] = 200593,
  [SMALL_STATE(4954)] = 200622,
  [SMALL_STATE(4955)] = 200651,
  [SMALL_STATE(4956)] = 200680,
  [SMALL_STATE(4957)] = 200703,
  [SMALL_STATE(4958)] = 200724,
  [SMALL_STATE(4959)] = 200753,
  [SMALL_STATE(4960)] = 200782,
  [SMALL_STATE(4961)] = 200803,
  [SMALL_STATE(4962)] = 200832,
  [SMALL_STATE(4963)] = 200853,
  [SMALL_STATE(4964)] = 200874,
  [SMALL_STATE(4965)] = 200903,
  [SMALL_STATE(4966)] = 200932,
  [SMALL_STATE(4967)] = 200961,
  [SMALL_STATE(4968)] = 200988,
  [SMALL_STATE(4969)] = 201017,
  [SMALL_STATE(4970)] = 201046,
  [SMALL_STATE(4971)] = 201067,
  [SMALL_STATE(4972)] = 201088,
  [SMALL_STATE(4973)] = 201115,
  [SMALL_STATE(4974)] = 201144,
  [SMALL_STATE(4975)] = 201173,
  [SMALL_STATE(4976)] = 201198,
  [SMALL_STATE(4977)] = 201221,
  [SMALL_STATE(4978)] = 201244,
  [SMALL_STATE(4979)] = 201265,
  [SMALL_STATE(4980)] = 201294,
  [SMALL_STATE(4981)] = 201319,
  [SMALL_STATE(4982)] = 201342,
  [SMALL_STATE(4983)] = 201371,
  [SMALL_STATE(4984)] = 201392,
  [SMALL_STATE(4985)] = 201419,
  [SMALL_STATE(4986)] = 201442,
  [SMALL_STATE(4987)] = 201471,
  [SMALL_STATE(4988)] = 201500,
  [SMALL_STATE(4989)] = 201529,
  [SMALL_STATE(4990)] = 201550,
  [SMALL_STATE(4991)] = 201579,
  [SMALL_STATE(4992)] = 201600,
  [SMALL_STATE(4993)] = 201627,
  [SMALL_STATE(4994)] = 201654,
  [SMALL_STATE(4995)] = 201681,
  [SMALL_STATE(4996)] = 201708,
  [SMALL_STATE(4997)] = 201737,
  [SMALL_STATE(4998)] = 201766,
  [SMALL_STATE(4999)] = 201787,
  [SMALL_STATE(5000)] = 201816,
  [SMALL_STATE(5001)] = 201837,
  [SMALL_STATE(5002)] = 201858,
  [SMALL_STATE(5003)] = 201879,
  [SMALL_STATE(5004)] = 201908,
  [SMALL_STATE(5005)] = 201931,
  [SMALL_STATE(5006)] = 201960,
  [SMALL_STATE(5007)] = 201981,
  [SMALL_STATE(5008)] = 202002,
  [SMALL_STATE(5009)] = 202031,
  [SMALL_STATE(5010)] = 202052,
  [SMALL_STATE(5011)] = 202073,
  [SMALL_STATE(5012)] = 202102,
  [SMALL_STATE(5013)] = 202129,
  [SMALL_STATE(5014)] = 202158,
  [SMALL_STATE(5015)] = 202179,
  [SMALL_STATE(5016)] = 202206,
  [SMALL_STATE(5017)] = 202229,
  [SMALL_STATE(5018)] = 202254,
  [SMALL_STATE(5019)] = 202277,
  [SMALL_STATE(5020)] = 202306,
  [SMALL_STATE(5021)] = 202335,
  [SMALL_STATE(5022)] = 202364,
  [SMALL_STATE(5023)] = 202385,
  [SMALL_STATE(5024)] = 202412,
  [SMALL_STATE(5025)] = 202437,
  [SMALL_STATE(5026)] = 202458,
  [SMALL_STATE(5027)] = 202485,
  [SMALL_STATE(5028)] = 202512,
  [SMALL_STATE(5029)] = 202541,
  [SMALL_STATE(5030)] = 202570,
  [SMALL_STATE(5031)] = 202591,
  [SMALL_STATE(5032)] = 202620,
  [SMALL_STATE(5033)] = 202641,
  [SMALL_STATE(5034)] = 202670,
  [SMALL_STATE(5035)] = 202699,
  [SMALL_STATE(5036)] = 202726,
  [SMALL_STATE(5037)] = 202755,
  [SMALL_STATE(5038)] = 202778,
  [SMALL_STATE(5039)] = 202799,
  [SMALL_STATE(5040)] = 202820,
  [SMALL_STATE(5041)] = 202849,
  [SMALL_STATE(5042)] = 202878,
  [SMALL_STATE(5043)] = 202902,
  [SMALL_STATE(5044)] = 202928,
  [SMALL_STATE(5045)] = 202954,
  [SMALL_STATE(5046)] = 202980,
  [SMALL_STATE(5047)] = 203004,
  [SMALL_STATE(5048)] = 203028,
  [SMALL_STATE(5049)] = 203052,
  [SMALL_STATE(5050)] = 203078,
  [SMALL_STATE(5051)] = 203104,
  [SMALL_STATE(5052)] = 203130,
  [SMALL_STATE(5053)] = 203154,
  [SMALL_STATE(5054)] = 203178,
  [SMALL_STATE(5055)] = 203202,
  [SMALL_STATE(5056)] = 203226,
  [SMALL_STATE(5057)] = 203252,
  [SMALL_STATE(5058)] = 203276,
  [SMALL_STATE(5059)] = 203300,
  [SMALL_STATE(5060)] = 203324,
  [SMALL_STATE(5061)] = 203348,
  [SMALL_STATE(5062)] = 203372,
  [SMALL_STATE(5063)] = 203396,
  [SMALL_STATE(5064)] = 203422,
  [SMALL_STATE(5065)] = 203446,
  [SMALL_STATE(5066)] = 203470,
  [SMALL_STATE(5067)] = 203494,
  [SMALL_STATE(5068)] = 203518,
  [SMALL_STATE(5069)] = 203542,
  [SMALL_STATE(5070)] = 203566,
  [SMALL_STATE(5071)] = 203590,
  [SMALL_STATE(5072)] = 203614,
  [SMALL_STATE(5073)] = 203640,
  [SMALL_STATE(5074)] = 203664,
  [SMALL_STATE(5075)] = 203688,
  [SMALL_STATE(5076)] = 203712,
  [SMALL_STATE(5077)] = 203738,
  [SMALL_STATE(5078)] = 203756,
  [SMALL_STATE(5079)] = 203782,
  [SMALL_STATE(5080)] = 203800,
  [SMALL_STATE(5081)] = 203826,
  [SMALL_STATE(5082)] = 203852,
  [SMALL_STATE(5083)] = 203870,
  [SMALL_STATE(5084)] = 203896,
  [SMALL_STATE(5085)] = 203922,
  [SMALL_STATE(5086)] = 203948,
  [SMALL_STATE(5087)] = 203972,
  [SMALL_STATE(5088)] = 203998,
  [SMALL_STATE(5089)] = 204020,
  [SMALL_STATE(5090)] = 204044,
  [SMALL_STATE(5091)] = 204070,
  [SMALL_STATE(5092)] = 204094,
  [SMALL_STATE(5093)] = 204120,
  [SMALL_STATE(5094)] = 204146,
  [SMALL_STATE(5095)] = 204170,
  [SMALL_STATE(5096)] = 204194,
  [SMALL_STATE(5097)] = 204220,
  [SMALL_STATE(5098)] = 204246,
  [SMALL_STATE(5099)] = 204270,
  [SMALL_STATE(5100)] = 204296,
  [SMALL_STATE(5101)] = 204322,
  [SMALL_STATE(5102)] = 204346,
  [SMALL_STATE(5103)] = 204370,
  [SMALL_STATE(5104)] = 204396,
  [SMALL_STATE(5105)] = 204422,
  [SMALL_STATE(5106)] = 204446,
  [SMALL_STATE(5107)] = 204468,
  [SMALL_STATE(5108)] = 204492,
  [SMALL_STATE(5109)] = 204516,
  [SMALL_STATE(5110)] = 204534,
  [SMALL_STATE(5111)] = 204552,
  [SMALL_STATE(5112)] = 204576,
  [SMALL_STATE(5113)] = 204602,
  [SMALL_STATE(5114)] = 204628,
  [SMALL_STATE(5115)] = 204652,
  [SMALL_STATE(5116)] = 204678,
  [SMALL_STATE(5117)] = 204704,
  [SMALL_STATE(5118)] = 204728,
  [SMALL_STATE(5119)] = 204752,
  [SMALL_STATE(5120)] = 204776,
  [SMALL_STATE(5121)] = 204800,
  [SMALL_STATE(5122)] = 204826,
  [SMALL_STATE(5123)] = 204852,
  [SMALL_STATE(5124)] = 204876,
  [SMALL_STATE(5125)] = 204900,
  [SMALL_STATE(5126)] = 204924,
  [SMALL_STATE(5127)] = 204948,
  [SMALL_STATE(5128)] = 204972,
  [SMALL_STATE(5129)] = 204996,
  [SMALL_STATE(5130)] = 205020,
  [SMALL_STATE(5131)] = 205046,
  [SMALL_STATE(5132)] = 205072,
  [SMALL_STATE(5133)] = 205096,
  [SMALL_STATE(5134)] = 205120,
  [SMALL_STATE(5135)] = 205144,
  [SMALL_STATE(5136)] = 205168,
  [SMALL_STATE(5137)] = 205192,
  [SMALL_STATE(5138)] = 205216,
  [SMALL_STATE(5139)] = 205240,
  [SMALL_STATE(5140)] = 205264,
  [SMALL_STATE(5141)] = 205290,
  [SMALL_STATE(5142)] = 205314,
  [SMALL_STATE(5143)] = 205336,
  [SMALL_STATE(5144)] = 205360,
  [SMALL_STATE(5145)] = 205378,
  [SMALL_STATE(5146)] = 205404,
  [SMALL_STATE(5147)] = 205430,
  [SMALL_STATE(5148)] = 205452,
  [SMALL_STATE(5149)] = 205478,
  [SMALL_STATE(5150)] = 205504,
  [SMALL_STATE(5151)] = 205522,
  [SMALL_STATE(5152)] = 205548,
  [SMALL_STATE(5153)] = 205574,
  [SMALL_STATE(5154)] = 205598,
  [SMALL_STATE(5155)] = 205616,
  [SMALL_STATE(5156)] = 205642,
  [SMALL_STATE(5157)] = 205666,
  [SMALL_STATE(5158)] = 205684,
  [SMALL_STATE(5159)] = 205708,
  [SMALL_STATE(5160)] = 205734,
  [SMALL_STATE(5161)] = 205758,
  [SMALL_STATE(5162)] = 205782,
  [SMALL_STATE(5163)] = 205808,
  [SMALL_STATE(5164)] = 205826,
  [SMALL_STATE(5165)] = 205852,
  [SMALL_STATE(5166)] = 205876,
  [SMALL_STATE(5167)] = 205900,
  [SMALL_STATE(5168)] = 205926,
  [SMALL_STATE(5169)] = 205950,
  [SMALL_STATE(5170)] = 205976,
  [SMALL_STATE(5171)] = 206000,
  [SMALL_STATE(5172)] = 206026,
  [SMALL_STATE(5173)] = 206050,
  [SMALL_STATE(5174)] = 206072,
  [SMALL_STATE(5175)] = 206096,
  [SMALL_STATE(5176)] = 206122,
  [SMALL_STATE(5177)] = 206146,
  [SMALL_STATE(5178)] = 206170,
  [SMALL_STATE(5179)] = 206194,
  [SMALL_STATE(5180)] = 206218,
  [SMALL_STATE(5181)] = 206242,
  [SMALL_STATE(5182)] = 206266,
  [SMALL_STATE(5183)] = 206290,
  [SMALL_STATE(5184)] = 206314,
  [SMALL_STATE(5185)] = 206334,
  [SMALL_STATE(5186)] = 206358,
  [SMALL_STATE(5187)] = 206384,
  [SMALL_STATE(5188)] = 206406,
  [SMALL_STATE(5189)] = 206432,
  [SMALL_STATE(5190)] = 206456,
  [SMALL_STATE(5191)] = 206480,
  [SMALL_STATE(5192)] = 206504,
  [SMALL_STATE(5193)] = 206528,
  [SMALL_STATE(5194)] = 206548,
  [SMALL_STATE(5195)] = 206574,
  [SMALL_STATE(5196)] = 206598,
  [SMALL_STATE(5197)] = 206620,
  [SMALL_STATE(5198)] = 206646,
  [SMALL_STATE(5199)] = 206664,
  [SMALL_STATE(5200)] = 206688,
  [SMALL_STATE(5201)] = 206712,
  [SMALL_STATE(5202)] = 206736,
  [SMALL_STATE(5203)] = 206760,
  [SMALL_STATE(5204)] = 206784,
  [SMALL_STATE(5205)] = 206808,
  [SMALL_STATE(5206)] = 206832,
  [SMALL_STATE(5207)] = 206856,
  [SMALL_STATE(5208)] = 206880,
  [SMALL_STATE(5209)] = 206904,
  [SMALL_STATE(5210)] = 206928,
  [SMALL_STATE(5211)] = 206954,
  [SMALL_STATE(5212)] = 206978,
  [SMALL_STATE(5213)] = 207002,
  [SMALL_STATE(5214)] = 207026,
  [SMALL_STATE(5215)] = 207046,
  [SMALL_STATE(5216)] = 207068,
  [SMALL_STATE(5217)] = 207094,
  [SMALL_STATE(5218)] = 207118,
  [SMALL_STATE(5219)] = 207142,
  [SMALL_STATE(5220)] = 207168,
  [SMALL_STATE(5221)] = 207192,
  [SMALL_STATE(5222)] = 207216,
  [SMALL_STATE(5223)] = 207240,
  [SMALL_STATE(5224)] = 207264,
  [SMALL_STATE(5225)] = 207288,
  [SMALL_STATE(5226)] = 207312,
  [SMALL_STATE(5227)] = 207336,
  [SMALL_STATE(5228)] = 207358,
  [SMALL_STATE(5229)] = 207382,
  [SMALL_STATE(5230)] = 207404,
  [SMALL_STATE(5231)] = 207426,
  [SMALL_STATE(5232)] = 207452,
  [SMALL_STATE(5233)] = 207476,
  [SMALL_STATE(5234)] = 207500,
  [SMALL_STATE(5235)] = 207522,
  [SMALL_STATE(5236)] = 207546,
  [SMALL_STATE(5237)] = 207570,
  [SMALL_STATE(5238)] = 207594,
  [SMALL_STATE(5239)] = 207618,
  [SMALL_STATE(5240)] = 207642,
  [SMALL_STATE(5241)] = 207668,
  [SMALL_STATE(5242)] = 207690,
  [SMALL_STATE(5243)] = 207714,
  [SMALL_STATE(5244)] = 207740,
  [SMALL_STATE(5245)] = 207766,
  [SMALL_STATE(5246)] = 207790,
  [SMALL_STATE(5247)] = 207812,
  [SMALL_STATE(5248)] = 207834,
  [SMALL_STATE(5249)] = 207856,
  [SMALL_STATE(5250)] = 207878,
  [SMALL_STATE(5251)] = 207904,
  [SMALL_STATE(5252)] = 207930,
  [SMALL_STATE(5253)] = 207956,
  [SMALL_STATE(5254)] = 207982,
  [SMALL_STATE(5255)] = 208008,
  [SMALL_STATE(5256)] = 208034,
  [SMALL_STATE(5257)] = 208060,
  [SMALL_STATE(5258)] = 208084,
  [SMALL_STATE(5259)] = 208108,
  [SMALL_STATE(5260)] = 208132,
  [SMALL_STATE(5261)] = 208156,
  [SMALL_STATE(5262)] = 208180,
  [SMALL_STATE(5263)] = 208203,
  [SMALL_STATE(5264)] = 208226,
  [SMALL_STATE(5265)] = 208249,
  [SMALL_STATE(5266)] = 208272,
  [SMALL_STATE(5267)] = 208295,
  [SMALL_STATE(5268)] = 208318,
  [SMALL_STATE(5269)] = 208339,
  [SMALL_STATE(5270)] = 208362,
  [SMALL_STATE(5271)] = 208385,
  [SMALL_STATE(5272)] = 208408,
  [SMALL_STATE(5273)] = 208431,
  [SMALL_STATE(5274)] = 208454,
  [SMALL_STATE(5275)] = 208475,
  [SMALL_STATE(5276)] = 208498,
  [SMALL_STATE(5277)] = 208519,
  [SMALL_STATE(5278)] = 208542,
  [SMALL_STATE(5279)] = 208565,
  [SMALL_STATE(5280)] = 208588,
  [SMALL_STATE(5281)] = 208611,
  [SMALL_STATE(5282)] = 208634,
  [SMALL_STATE(5283)] = 208657,
  [SMALL_STATE(5284)] = 208680,
  [SMALL_STATE(5285)] = 208703,
  [SMALL_STATE(5286)] = 208726,
  [SMALL_STATE(5287)] = 208747,
  [SMALL_STATE(5288)] = 208770,
  [SMALL_STATE(5289)] = 208791,
  [SMALL_STATE(5290)] = 208814,
  [SMALL_STATE(5291)] = 208837,
  [SMALL_STATE(5292)] = 208860,
  [SMALL_STATE(5293)] = 208883,
  [SMALL_STATE(5294)] = 208906,
  [SMALL_STATE(5295)] = 208929,
  [SMALL_STATE(5296)] = 208952,
  [SMALL_STATE(5297)] = 208975,
  [SMALL_STATE(5298)] = 208998,
  [SMALL_STATE(5299)] = 209019,
  [SMALL_STATE(5300)] = 209038,
  [SMALL_STATE(5301)] = 209061,
  [SMALL_STATE(5302)] = 209084,
  [SMALL_STATE(5303)] = 209107,
  [SMALL_STATE(5304)] = 209130,
  [SMALL_STATE(5305)] = 209153,
  [SMALL_STATE(5306)] = 209176,
  [SMALL_STATE(5307)] = 209197,
  [SMALL_STATE(5308)] = 209218,
  [SMALL_STATE(5309)] = 209241,
  [SMALL_STATE(5310)] = 209264,
  [SMALL_STATE(5311)] = 209287,
  [SMALL_STATE(5312)] = 209310,
  [SMALL_STATE(5313)] = 209333,
  [SMALL_STATE(5314)] = 209356,
  [SMALL_STATE(5315)] = 209379,
  [SMALL_STATE(5316)] = 209402,
  [SMALL_STATE(5317)] = 209425,
  [SMALL_STATE(5318)] = 209448,
  [SMALL_STATE(5319)] = 209471,
  [SMALL_STATE(5320)] = 209494,
  [SMALL_STATE(5321)] = 209517,
  [SMALL_STATE(5322)] = 209540,
  [SMALL_STATE(5323)] = 209563,
  [SMALL_STATE(5324)] = 209586,
  [SMALL_STATE(5325)] = 209609,
  [SMALL_STATE(5326)] = 209632,
  [SMALL_STATE(5327)] = 209655,
  [SMALL_STATE(5328)] = 209678,
  [SMALL_STATE(5329)] = 209701,
  [SMALL_STATE(5330)] = 209724,
  [SMALL_STATE(5331)] = 209747,
  [SMALL_STATE(5332)] = 209770,
  [SMALL_STATE(5333)] = 209793,
  [SMALL_STATE(5334)] = 209814,
  [SMALL_STATE(5335)] = 209831,
  [SMALL_STATE(5336)] = 209854,
  [SMALL_STATE(5337)] = 209877,
  [SMALL_STATE(5338)] = 209894,
  [SMALL_STATE(5339)] = 209917,
  [SMALL_STATE(5340)] = 209940,
  [SMALL_STATE(5341)] = 209963,
  [SMALL_STATE(5342)] = 209982,
  [SMALL_STATE(5343)] = 210001,
  [SMALL_STATE(5344)] = 210024,
  [SMALL_STATE(5345)] = 210047,
  [SMALL_STATE(5346)] = 210070,
  [SMALL_STATE(5347)] = 210091,
  [SMALL_STATE(5348)] = 210114,
  [SMALL_STATE(5349)] = 210137,
  [SMALL_STATE(5350)] = 210160,
  [SMALL_STATE(5351)] = 210181,
  [SMALL_STATE(5352)] = 210202,
  [SMALL_STATE(5353)] = 210225,
  [SMALL_STATE(5354)] = 210248,
  [SMALL_STATE(5355)] = 210271,
  [SMALL_STATE(5356)] = 210292,
  [SMALL_STATE(5357)] = 210315,
  [SMALL_STATE(5358)] = 210338,
  [SMALL_STATE(5359)] = 210357,
  [SMALL_STATE(5360)] = 210380,
  [SMALL_STATE(5361)] = 210403,
  [SMALL_STATE(5362)] = 210426,
  [SMALL_STATE(5363)] = 210449,
  [SMALL_STATE(5364)] = 210472,
  [SMALL_STATE(5365)] = 210495,
  [SMALL_STATE(5366)] = 210518,
  [SMALL_STATE(5367)] = 210541,
  [SMALL_STATE(5368)] = 210564,
  [SMALL_STATE(5369)] = 210585,
  [SMALL_STATE(5370)] = 210606,
  [SMALL_STATE(5371)] = 210629,
  [SMALL_STATE(5372)] = 210652,
  [SMALL_STATE(5373)] = 210675,
  [SMALL_STATE(5374)] = 210698,
  [SMALL_STATE(5375)] = 210721,
  [SMALL_STATE(5376)] = 210744,
  [SMALL_STATE(5377)] = 210767,
  [SMALL_STATE(5378)] = 210790,
  [SMALL_STATE(5379)] = 210813,
  [SMALL_STATE(5380)] = 210836,
  [SMALL_STATE(5381)] = 210859,
  [SMALL_STATE(5382)] = 210880,
  [SMALL_STATE(5383)] = 210901,
  [SMALL_STATE(5384)] = 210924,
  [SMALL_STATE(5385)] = 210947,
  [SMALL_STATE(5386)] = 210970,
  [SMALL_STATE(5387)] = 210993,
  [SMALL_STATE(5388)] = 211016,
  [SMALL_STATE(5389)] = 211039,
  [SMALL_STATE(5390)] = 211060,
  [SMALL_STATE(5391)] = 211083,
  [SMALL_STATE(5392)] = 211106,
  [SMALL_STATE(5393)] = 211129,
  [SMALL_STATE(5394)] = 211152,
  [SMALL_STATE(5395)] = 211175,
  [SMALL_STATE(5396)] = 211198,
  [SMALL_STATE(5397)] = 211221,
  [SMALL_STATE(5398)] = 211244,
  [SMALL_STATE(5399)] = 211267,
  [SMALL_STATE(5400)] = 211290,
  [SMALL_STATE(5401)] = 211311,
  [SMALL_STATE(5402)] = 211328,
  [SMALL_STATE(5403)] = 211351,
  [SMALL_STATE(5404)] = 211368,
  [SMALL_STATE(5405)] = 211391,
  [SMALL_STATE(5406)] = 211414,
  [SMALL_STATE(5407)] = 211435,
  [SMALL_STATE(5408)] = 211458,
  [SMALL_STATE(5409)] = 211479,
  [SMALL_STATE(5410)] = 211502,
  [SMALL_STATE(5411)] = 211525,
  [SMALL_STATE(5412)] = 211548,
  [SMALL_STATE(5413)] = 211571,
  [SMALL_STATE(5414)] = 211592,
  [SMALL_STATE(5415)] = 211613,
  [SMALL_STATE(5416)] = 211636,
  [SMALL_STATE(5417)] = 211659,
  [SMALL_STATE(5418)] = 211682,
  [SMALL_STATE(5419)] = 211705,
  [SMALL_STATE(5420)] = 211728,
  [SMALL_STATE(5421)] = 211751,
  [SMALL_STATE(5422)] = 211772,
  [SMALL_STATE(5423)] = 211795,
  [SMALL_STATE(5424)] = 211818,
  [SMALL_STATE(5425)] = 211839,
  [SMALL_STATE(5426)] = 211862,
  [SMALL_STATE(5427)] = 211879,
  [SMALL_STATE(5428)] = 211902,
  [SMALL_STATE(5429)] = 211925,
  [SMALL_STATE(5430)] = 211948,
  [SMALL_STATE(5431)] = 211971,
  [SMALL_STATE(5432)] = 211994,
  [SMALL_STATE(5433)] = 212017,
  [SMALL_STATE(5434)] = 212036,
  [SMALL_STATE(5435)] = 212059,
  [SMALL_STATE(5436)] = 212082,
  [SMALL_STATE(5437)] = 212105,
  [SMALL_STATE(5438)] = 212122,
  [SMALL_STATE(5439)] = 212143,
  [SMALL_STATE(5440)] = 212166,
  [SMALL_STATE(5441)] = 212189,
  [SMALL_STATE(5442)] = 212212,
  [SMALL_STATE(5443)] = 212235,
  [SMALL_STATE(5444)] = 212258,
  [SMALL_STATE(5445)] = 212281,
  [SMALL_STATE(5446)] = 212300,
  [SMALL_STATE(5447)] = 212323,
  [SMALL_STATE(5448)] = 212340,
  [SMALL_STATE(5449)] = 212357,
  [SMALL_STATE(5450)] = 212374,
  [SMALL_STATE(5451)] = 212397,
  [SMALL_STATE(5452)] = 212420,
  [SMALL_STATE(5453)] = 212443,
  [SMALL_STATE(5454)] = 212464,
  [SMALL_STATE(5455)] = 212483,
  [SMALL_STATE(5456)] = 212506,
  [SMALL_STATE(5457)] = 212529,
  [SMALL_STATE(5458)] = 212552,
  [SMALL_STATE(5459)] = 212575,
  [SMALL_STATE(5460)] = 212598,
  [SMALL_STATE(5461)] = 212621,
  [SMALL_STATE(5462)] = 212644,
  [SMALL_STATE(5463)] = 212667,
  [SMALL_STATE(5464)] = 212690,
  [SMALL_STATE(5465)] = 212709,
  [SMALL_STATE(5466)] = 212732,
  [SMALL_STATE(5467)] = 212751,
  [SMALL_STATE(5468)] = 212774,
  [SMALL_STATE(5469)] = 212797,
  [SMALL_STATE(5470)] = 212820,
  [SMALL_STATE(5471)] = 212843,
  [SMALL_STATE(5472)] = 212866,
  [SMALL_STATE(5473)] = 212889,
  [SMALL_STATE(5474)] = 212912,
  [SMALL_STATE(5475)] = 212935,
  [SMALL_STATE(5476)] = 212958,
  [SMALL_STATE(5477)] = 212981,
  [SMALL_STATE(5478)] = 213002,
  [SMALL_STATE(5479)] = 213023,
  [SMALL_STATE(5480)] = 213042,
  [SMALL_STATE(5481)] = 213061,
  [SMALL_STATE(5482)] = 213080,
  [SMALL_STATE(5483)] = 213103,
  [SMALL_STATE(5484)] = 213126,
  [SMALL_STATE(5485)] = 213149,
  [SMALL_STATE(5486)] = 213168,
  [SMALL_STATE(5487)] = 213191,
  [SMALL_STATE(5488)] = 213214,
  [SMALL_STATE(5489)] = 213237,
  [SMALL_STATE(5490)] = 213260,
  [SMALL_STATE(5491)] = 213283,
  [SMALL_STATE(5492)] = 213306,
  [SMALL_STATE(5493)] = 213323,
  [SMALL_STATE(5494)] = 213346,
  [SMALL_STATE(5495)] = 213369,
  [SMALL_STATE(5496)] = 213392,
  [SMALL_STATE(5497)] = 213413,
  [SMALL_STATE(5498)] = 213434,
  [SMALL_STATE(5499)] = 213457,
  [SMALL_STATE(5500)] = 213480,
  [SMALL_STATE(5501)] = 213501,
  [SMALL_STATE(5502)] = 213524,
  [SMALL_STATE(5503)] = 213543,
  [SMALL_STATE(5504)] = 213566,
  [SMALL_STATE(5505)] = 213587,
  [SMALL_STATE(5506)] = 213610,
  [SMALL_STATE(5507)] = 213631,
  [SMALL_STATE(5508)] = 213654,
  [SMALL_STATE(5509)] = 213677,
  [SMALL_STATE(5510)] = 213694,
  [SMALL_STATE(5511)] = 213717,
  [SMALL_STATE(5512)] = 213740,
  [SMALL_STATE(5513)] = 213763,
  [SMALL_STATE(5514)] = 213780,
  [SMALL_STATE(5515)] = 213803,
  [SMALL_STATE(5516)] = 213824,
  [SMALL_STATE(5517)] = 213841,
  [SMALL_STATE(5518)] = 213864,
  [SMALL_STATE(5519)] = 213887,
  [SMALL_STATE(5520)] = 213910,
  [SMALL_STATE(5521)] = 213933,
  [SMALL_STATE(5522)] = 213950,
  [SMALL_STATE(5523)] = 213971,
  [SMALL_STATE(5524)] = 213994,
  [SMALL_STATE(5525)] = 214017,
  [SMALL_STATE(5526)] = 214038,
  [SMALL_STATE(5527)] = 214055,
  [SMALL_STATE(5528)] = 214078,
  [SMALL_STATE(5529)] = 214097,
  [SMALL_STATE(5530)] = 214116,
  [SMALL_STATE(5531)] = 214139,
  [SMALL_STATE(5532)] = 214162,
  [SMALL_STATE(5533)] = 214185,
  [SMALL_STATE(5534)] = 214208,
  [SMALL_STATE(5535)] = 214225,
  [SMALL_STATE(5536)] = 214246,
  [SMALL_STATE(5537)] = 214269,
  [SMALL_STATE(5538)] = 214292,
  [SMALL_STATE(5539)] = 214315,
  [SMALL_STATE(5540)] = 214338,
  [SMALL_STATE(5541)] = 214357,
  [SMALL_STATE(5542)] = 214375,
  [SMALL_STATE(5543)] = 214393,
  [SMALL_STATE(5544)] = 214411,
  [SMALL_STATE(5545)] = 214431,
  [SMALL_STATE(5546)] = 214451,
  [SMALL_STATE(5547)] = 214471,
  [SMALL_STATE(5548)] = 214491,
  [SMALL_STATE(5549)] = 214511,
  [SMALL_STATE(5550)] = 214531,
  [SMALL_STATE(5551)] = 214551,
  [SMALL_STATE(5552)] = 214569,
  [SMALL_STATE(5553)] = 214585,
  [SMALL_STATE(5554)] = 214605,
  [SMALL_STATE(5555)] = 214625,
  [SMALL_STATE(5556)] = 214645,
  [SMALL_STATE(5557)] = 214665,
  [SMALL_STATE(5558)] = 214685,
  [SMALL_STATE(5559)] = 214705,
  [SMALL_STATE(5560)] = 214725,
  [SMALL_STATE(5561)] = 214745,
  [SMALL_STATE(5562)] = 214765,
  [SMALL_STATE(5563)] = 214785,
  [SMALL_STATE(5564)] = 214805,
  [SMALL_STATE(5565)] = 214825,
  [SMALL_STATE(5566)] = 214843,
  [SMALL_STATE(5567)] = 214863,
  [SMALL_STATE(5568)] = 214883,
  [SMALL_STATE(5569)] = 214903,
  [SMALL_STATE(5570)] = 214921,
  [SMALL_STATE(5571)] = 214941,
  [SMALL_STATE(5572)] = 214961,
  [SMALL_STATE(5573)] = 214981,
  [SMALL_STATE(5574)] = 215001,
  [SMALL_STATE(5575)] = 215021,
  [SMALL_STATE(5576)] = 215041,
  [SMALL_STATE(5577)] = 215059,
  [SMALL_STATE(5578)] = 215079,
  [SMALL_STATE(5579)] = 215097,
  [SMALL_STATE(5580)] = 215115,
  [SMALL_STATE(5581)] = 215135,
  [SMALL_STATE(5582)] = 215153,
  [SMALL_STATE(5583)] = 215173,
  [SMALL_STATE(5584)] = 215193,
  [SMALL_STATE(5585)] = 215211,
  [SMALL_STATE(5586)] = 215231,
  [SMALL_STATE(5587)] = 215251,
  [SMALL_STATE(5588)] = 215271,
  [SMALL_STATE(5589)] = 215291,
  [SMALL_STATE(5590)] = 215311,
  [SMALL_STATE(5591)] = 215331,
  [SMALL_STATE(5592)] = 215351,
  [SMALL_STATE(5593)] = 215369,
  [SMALL_STATE(5594)] = 215389,
  [SMALL_STATE(5595)] = 215405,
  [SMALL_STATE(5596)] = 215425,
  [SMALL_STATE(5597)] = 215443,
  [SMALL_STATE(5598)] = 215461,
  [SMALL_STATE(5599)] = 215481,
  [SMALL_STATE(5600)] = 215501,
  [SMALL_STATE(5601)] = 215521,
  [SMALL_STATE(5602)] = 215541,
  [SMALL_STATE(5603)] = 215561,
  [SMALL_STATE(5604)] = 215581,
  [SMALL_STATE(5605)] = 215601,
  [SMALL_STATE(5606)] = 215619,
  [SMALL_STATE(5607)] = 215639,
  [SMALL_STATE(5608)] = 215657,
  [SMALL_STATE(5609)] = 215677,
  [SMALL_STATE(5610)] = 215697,
  [SMALL_STATE(5611)] = 215717,
  [SMALL_STATE(5612)] = 215737,
  [SMALL_STATE(5613)] = 215757,
  [SMALL_STATE(5614)] = 215777,
  [SMALL_STATE(5615)] = 215797,
  [SMALL_STATE(5616)] = 215817,
  [SMALL_STATE(5617)] = 215837,
  [SMALL_STATE(5618)] = 215857,
  [SMALL_STATE(5619)] = 215877,
  [SMALL_STATE(5620)] = 215897,
  [SMALL_STATE(5621)] = 215917,
  [SMALL_STATE(5622)] = 215933,
  [SMALL_STATE(5623)] = 215953,
  [SMALL_STATE(5624)] = 215973,
  [SMALL_STATE(5625)] = 215993,
  [SMALL_STATE(5626)] = 216013,
  [SMALL_STATE(5627)] = 216033,
  [SMALL_STATE(5628)] = 216051,
  [SMALL_STATE(5629)] = 216071,
  [SMALL_STATE(5630)] = 216091,
  [SMALL_STATE(5631)] = 216111,
  [SMALL_STATE(5632)] = 216131,
  [SMALL_STATE(5633)] = 216151,
  [SMALL_STATE(5634)] = 216171,
  [SMALL_STATE(5635)] = 216191,
  [SMALL_STATE(5636)] = 216211,
  [SMALL_STATE(5637)] = 216231,
  [SMALL_STATE(5638)] = 216251,
  [SMALL_STATE(5639)] = 216271,
  [SMALL_STATE(5640)] = 216291,
  [SMALL_STATE(5641)] = 216309,
  [SMALL_STATE(5642)] = 216329,
  [SMALL_STATE(5643)] = 216347,
  [SMALL_STATE(5644)] = 216367,
  [SMALL_STATE(5645)] = 216385,
  [SMALL_STATE(5646)] = 216405,
  [SMALL_STATE(5647)] = 216421,
  [SMALL_STATE(5648)] = 216441,
  [SMALL_STATE(5649)] = 216461,
  [SMALL_STATE(5650)] = 216481,
  [SMALL_STATE(5651)] = 216501,
  [SMALL_STATE(5652)] = 216521,
  [SMALL_STATE(5653)] = 216541,
  [SMALL_STATE(5654)] = 216561,
  [SMALL_STATE(5655)] = 216581,
  [SMALL_STATE(5656)] = 216601,
  [SMALL_STATE(5657)] = 216621,
  [SMALL_STATE(5658)] = 216641,
  [SMALL_STATE(5659)] = 216661,
  [SMALL_STATE(5660)] = 216681,
  [SMALL_STATE(5661)] = 216701,
  [SMALL_STATE(5662)] = 216721,
  [SMALL_STATE(5663)] = 216739,
  [SMALL_STATE(5664)] = 216757,
  [SMALL_STATE(5665)] = 216777,
  [SMALL_STATE(5666)] = 216797,
  [SMALL_STATE(5667)] = 216815,
  [SMALL_STATE(5668)] = 216835,
  [SMALL_STATE(5669)] = 216855,
  [SMALL_STATE(5670)] = 216873,
  [SMALL_STATE(5671)] = 216893,
  [SMALL_STATE(5672)] = 216913,
  [SMALL_STATE(5673)] = 216933,
  [SMALL_STATE(5674)] = 216953,
  [SMALL_STATE(5675)] = 216973,
  [SMALL_STATE(5676)] = 216993,
  [SMALL_STATE(5677)] = 217011,
  [SMALL_STATE(5678)] = 217031,
  [SMALL_STATE(5679)] = 217049,
  [SMALL_STATE(5680)] = 217069,
  [SMALL_STATE(5681)] = 217089,
  [SMALL_STATE(5682)] = 217109,
  [SMALL_STATE(5683)] = 217129,
  [SMALL_STATE(5684)] = 217147,
  [SMALL_STATE(5685)] = 217165,
  [SMALL_STATE(5686)] = 217185,
  [SMALL_STATE(5687)] = 217205,
  [SMALL_STATE(5688)] = 217223,
  [SMALL_STATE(5689)] = 217243,
  [SMALL_STATE(5690)] = 217263,
  [SMALL_STATE(5691)] = 217283,
  [SMALL_STATE(5692)] = 217299,
  [SMALL_STATE(5693)] = 217319,
  [SMALL_STATE(5694)] = 217339,
  [SMALL_STATE(5695)] = 217357,
  [SMALL_STATE(5696)] = 217375,
  [SMALL_STATE(5697)] = 217395,
  [SMALL_STATE(5698)] = 217415,
  [SMALL_STATE(5699)] = 217435,
  [SMALL_STATE(5700)] = 217455,
  [SMALL_STATE(5701)] = 217473,
  [SMALL_STATE(5702)] = 217493,
  [SMALL_STATE(5703)] = 217513,
  [SMALL_STATE(5704)] = 217533,
  [SMALL_STATE(5705)] = 217553,
  [SMALL_STATE(5706)] = 217573,
  [SMALL_STATE(5707)] = 217593,
  [SMALL_STATE(5708)] = 217613,
  [SMALL_STATE(5709)] = 217631,
  [SMALL_STATE(5710)] = 217651,
  [SMALL_STATE(5711)] = 217671,
  [SMALL_STATE(5712)] = 217691,
  [SMALL_STATE(5713)] = 217711,
  [SMALL_STATE(5714)] = 217731,
  [SMALL_STATE(5715)] = 217751,
  [SMALL_STATE(5716)] = 217771,
  [SMALL_STATE(5717)] = 217791,
  [SMALL_STATE(5718)] = 217809,
  [SMALL_STATE(5719)] = 217829,
  [SMALL_STATE(5720)] = 217849,
  [SMALL_STATE(5721)] = 217869,
  [SMALL_STATE(5722)] = 217889,
  [SMALL_STATE(5723)] = 217909,
  [SMALL_STATE(5724)] = 217929,
  [SMALL_STATE(5725)] = 217949,
  [SMALL_STATE(5726)] = 217969,
  [SMALL_STATE(5727)] = 217989,
  [SMALL_STATE(5728)] = 218009,
  [SMALL_STATE(5729)] = 218029,
  [SMALL_STATE(5730)] = 218049,
  [SMALL_STATE(5731)] = 218069,
  [SMALL_STATE(5732)] = 218087,
  [SMALL_STATE(5733)] = 218107,
  [SMALL_STATE(5734)] = 218127,
  [SMALL_STATE(5735)] = 218147,
  [SMALL_STATE(5736)] = 218167,
  [SMALL_STATE(5737)] = 218187,
  [SMALL_STATE(5738)] = 218207,
  [SMALL_STATE(5739)] = 218227,
  [SMALL_STATE(5740)] = 218245,
  [SMALL_STATE(5741)] = 218265,
  [SMALL_STATE(5742)] = 218283,
  [SMALL_STATE(5743)] = 218301,
  [SMALL_STATE(5744)] = 218321,
  [SMALL_STATE(5745)] = 218341,
  [SMALL_STATE(5746)] = 218361,
  [SMALL_STATE(5747)] = 218377,
  [SMALL_STATE(5748)] = 218395,
  [SMALL_STATE(5749)] = 218415,
  [SMALL_STATE(5750)] = 218435,
  [SMALL_STATE(5751)] = 218455,
  [SMALL_STATE(5752)] = 218475,
  [SMALL_STATE(5753)] = 218495,
  [SMALL_STATE(5754)] = 218513,
  [SMALL_STATE(5755)] = 218533,
  [SMALL_STATE(5756)] = 218549,
  [SMALL_STATE(5757)] = 218569,
  [SMALL_STATE(5758)] = 218589,
  [SMALL_STATE(5759)] = 218609,
  [SMALL_STATE(5760)] = 218629,
  [SMALL_STATE(5761)] = 218645,
  [SMALL_STATE(5762)] = 218663,
  [SMALL_STATE(5763)] = 218683,
  [SMALL_STATE(5764)] = 218703,
  [SMALL_STATE(5765)] = 218723,
  [SMALL_STATE(5766)] = 218743,
  [SMALL_STATE(5767)] = 218763,
  [SMALL_STATE(5768)] = 218781,
  [SMALL_STATE(5769)] = 218801,
  [SMALL_STATE(5770)] = 218821,
  [SMALL_STATE(5771)] = 218841,
  [SMALL_STATE(5772)] = 218861,
  [SMALL_STATE(5773)] = 218881,
  [SMALL_STATE(5774)] = 218901,
  [SMALL_STATE(5775)] = 218919,
  [SMALL_STATE(5776)] = 218939,
  [SMALL_STATE(5777)] = 218959,
  [SMALL_STATE(5778)] = 218977,
  [SMALL_STATE(5779)] = 218997,
  [SMALL_STATE(5780)] = 219015,
  [SMALL_STATE(5781)] = 219035,
  [SMALL_STATE(5782)] = 219053,
  [SMALL_STATE(5783)] = 219071,
  [SMALL_STATE(5784)] = 219091,
  [SMALL_STATE(5785)] = 219111,
  [SMALL_STATE(5786)] = 219131,
  [SMALL_STATE(5787)] = 219149,
  [SMALL_STATE(5788)] = 219169,
  [SMALL_STATE(5789)] = 219189,
  [SMALL_STATE(5790)] = 219209,
  [SMALL_STATE(5791)] = 219227,
  [SMALL_STATE(5792)] = 219247,
  [SMALL_STATE(5793)] = 219267,
  [SMALL_STATE(5794)] = 219285,
  [SMALL_STATE(5795)] = 219305,
  [SMALL_STATE(5796)] = 219325,
  [SMALL_STATE(5797)] = 219343,
  [SMALL_STATE(5798)] = 219363,
  [SMALL_STATE(5799)] = 219383,
  [SMALL_STATE(5800)] = 219401,
  [SMALL_STATE(5801)] = 219421,
  [SMALL_STATE(5802)] = 219439,
  [SMALL_STATE(5803)] = 219459,
  [SMALL_STATE(5804)] = 219479,
  [SMALL_STATE(5805)] = 219497,
  [SMALL_STATE(5806)] = 219517,
  [SMALL_STATE(5807)] = 219535,
  [SMALL_STATE(5808)] = 219555,
  [SMALL_STATE(5809)] = 219573,
  [SMALL_STATE(5810)] = 219593,
  [SMALL_STATE(5811)] = 219611,
  [SMALL_STATE(5812)] = 219631,
  [SMALL_STATE(5813)] = 219649,
  [SMALL_STATE(5814)] = 219669,
  [SMALL_STATE(5815)] = 219687,
  [SMALL_STATE(5816)] = 219707,
  [SMALL_STATE(5817)] = 219725,
  [SMALL_STATE(5818)] = 219745,
  [SMALL_STATE(5819)] = 219761,
  [SMALL_STATE(5820)] = 219779,
  [SMALL_STATE(5821)] = 219797,
  [SMALL_STATE(5822)] = 219813,
  [SMALL_STATE(5823)] = 219831,
  [SMALL_STATE(5824)] = 219851,
  [SMALL_STATE(5825)] = 219867,
  [SMALL_STATE(5826)] = 219887,
  [SMALL_STATE(5827)] = 219907,
  [SMALL_STATE(5828)] = 219927,
  [SMALL_STATE(5829)] = 219947,
  [SMALL_STATE(5830)] = 219963,
  [SMALL_STATE(5831)] = 219983,
  [SMALL_STATE(5832)] = 220003,
  [SMALL_STATE(5833)] = 220023,
  [SMALL_STATE(5834)] = 220043,
  [SMALL_STATE(5835)] = 220063,
  [SMALL_STATE(5836)] = 220083,
  [SMALL_STATE(5837)] = 220103,
  [SMALL_STATE(5838)] = 220121,
  [SMALL_STATE(5839)] = 220141,
  [SMALL_STATE(5840)] = 220161,
  [SMALL_STATE(5841)] = 220181,
  [SMALL_STATE(5842)] = 220201,
  [SMALL_STATE(5843)] = 220221,
  [SMALL_STATE(5844)] = 220241,
  [SMALL_STATE(5845)] = 220261,
  [SMALL_STATE(5846)] = 220281,
  [SMALL_STATE(5847)] = 220301,
  [SMALL_STATE(5848)] = 220321,
  [SMALL_STATE(5849)] = 220341,
  [SMALL_STATE(5850)] = 220361,
  [SMALL_STATE(5851)] = 220379,
  [SMALL_STATE(5852)] = 220397,
  [SMALL_STATE(5853)] = 220417,
  [SMALL_STATE(5854)] = 220437,
  [SMALL_STATE(5855)] = 220457,
  [SMALL_STATE(5856)] = 220477,
  [SMALL_STATE(5857)] = 220495,
  [SMALL_STATE(5858)] = 220515,
  [SMALL_STATE(5859)] = 220535,
  [SMALL_STATE(5860)] = 220553,
  [SMALL_STATE(5861)] = 220573,
  [SMALL_STATE(5862)] = 220593,
  [SMALL_STATE(5863)] = 220613,
  [SMALL_STATE(5864)] = 220633,
  [SMALL_STATE(5865)] = 220653,
  [SMALL_STATE(5866)] = 220671,
  [SMALL_STATE(5867)] = 220691,
  [SMALL_STATE(5868)] = 220709,
  [SMALL_STATE(5869)] = 220729,
  [SMALL_STATE(5870)] = 220749,
  [SMALL_STATE(5871)] = 220769,
  [SMALL_STATE(5872)] = 220787,
  [SMALL_STATE(5873)] = 220807,
  [SMALL_STATE(5874)] = 220827,
  [SMALL_STATE(5875)] = 220847,
  [SMALL_STATE(5876)] = 220867,
  [SMALL_STATE(5877)] = 220885,
  [SMALL_STATE(5878)] = 220905,
  [SMALL_STATE(5879)] = 220925,
  [SMALL_STATE(5880)] = 220945,
  [SMALL_STATE(5881)] = 220965,
  [SMALL_STATE(5882)] = 220985,
  [SMALL_STATE(5883)] = 221003,
  [SMALL_STATE(5884)] = 221023,
  [SMALL_STATE(5885)] = 221043,
  [SMALL_STATE(5886)] = 221061,
  [SMALL_STATE(5887)] = 221081,
  [SMALL_STATE(5888)] = 221101,
  [SMALL_STATE(5889)] = 221121,
  [SMALL_STATE(5890)] = 221141,
  [SMALL_STATE(5891)] = 221161,
  [SMALL_STATE(5892)] = 221179,
  [SMALL_STATE(5893)] = 221199,
  [SMALL_STATE(5894)] = 221219,
  [SMALL_STATE(5895)] = 221239,
  [SMALL_STATE(5896)] = 221259,
  [SMALL_STATE(5897)] = 221279,
  [SMALL_STATE(5898)] = 221297,
  [SMALL_STATE(5899)] = 221317,
  [SMALL_STATE(5900)] = 221335,
  [SMALL_STATE(5901)] = 221355,
  [SMALL_STATE(5902)] = 221373,
  [SMALL_STATE(5903)] = 221393,
  [SMALL_STATE(5904)] = 221411,
  [SMALL_STATE(5905)] = 221431,
  [SMALL_STATE(5906)] = 221449,
  [SMALL_STATE(5907)] = 221469,
  [SMALL_STATE(5908)] = 221487,
  [SMALL_STATE(5909)] = 221507,
  [SMALL_STATE(5910)] = 221525,
  [SMALL_STATE(5911)] = 221545,
  [SMALL_STATE(5912)] = 221563,
  [SMALL_STATE(5913)] = 221581,
  [SMALL_STATE(5914)] = 221599,
  [SMALL_STATE(5915)] = 221617,
  [SMALL_STATE(5916)] = 221637,
  [SMALL_STATE(5917)] = 221655,
  [SMALL_STATE(5918)] = 221675,
  [SMALL_STATE(5919)] = 221695,
  [SMALL_STATE(5920)] = 221715,
  [SMALL_STATE(5921)] = 221733,
  [SMALL_STATE(5922)] = 221753,
  [SMALL_STATE(5923)] = 221773,
  [SMALL_STATE(5924)] = 221793,
  [SMALL_STATE(5925)] = 221813,
  [SMALL_STATE(5926)] = 221831,
  [SMALL_STATE(5927)] = 221851,
  [SMALL_STATE(5928)] = 221871,
  [SMALL_STATE(5929)] = 221891,
  [SMALL_STATE(5930)] = 221911,
  [SMALL_STATE(5931)] = 221927,
  [SMALL_STATE(5932)] = 221945,
  [SMALL_STATE(5933)] = 221963,
  [SMALL_STATE(5934)] = 221983,
  [SMALL_STATE(5935)] = 222003,
  [SMALL_STATE(5936)] = 222023,
  [SMALL_STATE(5937)] = 222041,
  [SMALL_STATE(5938)] = 222061,
  [SMALL_STATE(5939)] = 222079,
  [SMALL_STATE(5940)] = 222099,
  [SMALL_STATE(5941)] = 222117,
  [SMALL_STATE(5942)] = 222137,
  [SMALL_STATE(5943)] = 222157,
  [SMALL_STATE(5944)] = 222175,
  [SMALL_STATE(5945)] = 222195,
  [SMALL_STATE(5946)] = 222215,
  [SMALL_STATE(5947)] = 222233,
  [SMALL_STATE(5948)] = 222253,
  [SMALL_STATE(5949)] = 222270,
  [SMALL_STATE(5950)] = 222287,
  [SMALL_STATE(5951)] = 222304,
  [SMALL_STATE(5952)] = 222321,
  [SMALL_STATE(5953)] = 222338,
  [SMALL_STATE(5954)] = 222355,
  [SMALL_STATE(5955)] = 222372,
  [SMALL_STATE(5956)] = 222389,
  [SMALL_STATE(5957)] = 222404,
  [SMALL_STATE(5958)] = 222421,
  [SMALL_STATE(5959)] = 222438,
  [SMALL_STATE(5960)] = 222455,
  [SMALL_STATE(5961)] = 222472,
  [SMALL_STATE(5962)] = 222489,
  [SMALL_STATE(5963)] = 222504,
  [SMALL_STATE(5964)] = 222519,
  [SMALL_STATE(5965)] = 222536,
  [SMALL_STATE(5966)] = 222551,
  [SMALL_STATE(5967)] = 222568,
  [SMALL_STATE(5968)] = 222585,
  [SMALL_STATE(5969)] = 222602,
  [SMALL_STATE(5970)] = 222619,
  [SMALL_STATE(5971)] = 222636,
  [SMALL_STATE(5972)] = 222653,
  [SMALL_STATE(5973)] = 222670,
  [SMALL_STATE(5974)] = 222687,
  [SMALL_STATE(5975)] = 222704,
  [SMALL_STATE(5976)] = 222721,
  [SMALL_STATE(5977)] = 222738,
  [SMALL_STATE(5978)] = 222755,
  [SMALL_STATE(5979)] = 222770,
  [SMALL_STATE(5980)] = 222785,
  [SMALL_STATE(5981)] = 222802,
  [SMALL_STATE(5982)] = 222819,
  [SMALL_STATE(5983)] = 222836,
  [SMALL_STATE(5984)] = 222853,
  [SMALL_STATE(5985)] = 222870,
  [SMALL_STATE(5986)] = 222887,
  [SMALL_STATE(5987)] = 222904,
  [SMALL_STATE(5988)] = 222921,
  [SMALL_STATE(5989)] = 222938,
  [SMALL_STATE(5990)] = 222953,
  [SMALL_STATE(5991)] = 222970,
  [SMALL_STATE(5992)] = 222987,
  [SMALL_STATE(5993)] = 223004,
  [SMALL_STATE(5994)] = 223021,
  [SMALL_STATE(5995)] = 223036,
  [SMALL_STATE(5996)] = 223053,
  [SMALL_STATE(5997)] = 223070,
  [SMALL_STATE(5998)] = 223087,
  [SMALL_STATE(5999)] = 223102,
  [SMALL_STATE(6000)] = 223119,
  [SMALL_STATE(6001)] = 223136,
  [SMALL_STATE(6002)] = 223153,
  [SMALL_STATE(6003)] = 223170,
  [SMALL_STATE(6004)] = 223187,
  [SMALL_STATE(6005)] = 223204,
  [SMALL_STATE(6006)] = 223219,
  [SMALL_STATE(6007)] = 223236,
  [SMALL_STATE(6008)] = 223253,
  [SMALL_STATE(6009)] = 223270,
  [SMALL_STATE(6010)] = 223287,
  [SMALL_STATE(6011)] = 223304,
  [SMALL_STATE(6012)] = 223321,
  [SMALL_STATE(6013)] = 223338,
  [SMALL_STATE(6014)] = 223355,
  [SMALL_STATE(6015)] = 223372,
  [SMALL_STATE(6016)] = 223389,
  [SMALL_STATE(6017)] = 223406,
  [SMALL_STATE(6018)] = 223423,
  [SMALL_STATE(6019)] = 223440,
  [SMALL_STATE(6020)] = 223457,
  [SMALL_STATE(6021)] = 223474,
  [SMALL_STATE(6022)] = 223491,
  [SMALL_STATE(6023)] = 223508,
  [SMALL_STATE(6024)] = 223525,
  [SMALL_STATE(6025)] = 223540,
  [SMALL_STATE(6026)] = 223557,
  [SMALL_STATE(6027)] = 223574,
  [SMALL_STATE(6028)] = 223591,
  [SMALL_STATE(6029)] = 223608,
  [SMALL_STATE(6030)] = 223625,
  [SMALL_STATE(6031)] = 223642,
  [SMALL_STATE(6032)] = 223657,
  [SMALL_STATE(6033)] = 223674,
  [SMALL_STATE(6034)] = 223691,
  [SMALL_STATE(6035)] = 223708,
  [SMALL_STATE(6036)] = 223725,
  [SMALL_STATE(6037)] = 223740,
  [SMALL_STATE(6038)] = 223757,
  [SMALL_STATE(6039)] = 223774,
  [SMALL_STATE(6040)] = 223791,
  [SMALL_STATE(6041)] = 223806,
  [SMALL_STATE(6042)] = 223821,
  [SMALL_STATE(6043)] = 223838,
  [SMALL_STATE(6044)] = 223853,
  [SMALL_STATE(6045)] = 223870,
  [SMALL_STATE(6046)] = 223885,
  [SMALL_STATE(6047)] = 223902,
  [SMALL_STATE(6048)] = 223919,
  [SMALL_STATE(6049)] = 223934,
  [SMALL_STATE(6050)] = 223951,
  [SMALL_STATE(6051)] = 223968,
  [SMALL_STATE(6052)] = 223985,
  [SMALL_STATE(6053)] = 224002,
  [SMALL_STATE(6054)] = 224019,
  [SMALL_STATE(6055)] = 224036,
  [SMALL_STATE(6056)] = 224053,
  [SMALL_STATE(6057)] = 224070,
  [SMALL_STATE(6058)] = 224087,
  [SMALL_STATE(6059)] = 224104,
  [SMALL_STATE(6060)] = 224121,
  [SMALL_STATE(6061)] = 224136,
  [SMALL_STATE(6062)] = 224153,
  [SMALL_STATE(6063)] = 224170,
  [SMALL_STATE(6064)] = 224185,
  [SMALL_STATE(6065)] = 224202,
  [SMALL_STATE(6066)] = 224219,
  [SMALL_STATE(6067)] = 224236,
  [SMALL_STATE(6068)] = 224253,
  [SMALL_STATE(6069)] = 224270,
  [SMALL_STATE(6070)] = 224285,
  [SMALL_STATE(6071)] = 224300,
  [SMALL_STATE(6072)] = 224317,
  [SMALL_STATE(6073)] = 224334,
  [SMALL_STATE(6074)] = 224351,
  [SMALL_STATE(6075)] = 224368,
  [SMALL_STATE(6076)] = 224385,
  [SMALL_STATE(6077)] = 224402,
  [SMALL_STATE(6078)] = 224419,
  [SMALL_STATE(6079)] = 224436,
  [SMALL_STATE(6080)] = 224453,
  [SMALL_STATE(6081)] = 224470,
  [SMALL_STATE(6082)] = 224487,
  [SMALL_STATE(6083)] = 224504,
  [SMALL_STATE(6084)] = 224521,
  [SMALL_STATE(6085)] = 224538,
  [SMALL_STATE(6086)] = 224555,
  [SMALL_STATE(6087)] = 224572,
  [SMALL_STATE(6088)] = 224589,
  [SMALL_STATE(6089)] = 224606,
  [SMALL_STATE(6090)] = 224623,
  [SMALL_STATE(6091)] = 224640,
  [SMALL_STATE(6092)] = 224657,
  [SMALL_STATE(6093)] = 224674,
  [SMALL_STATE(6094)] = 224691,
  [SMALL_STATE(6095)] = 224708,
  [SMALL_STATE(6096)] = 224725,
  [SMALL_STATE(6097)] = 224742,
  [SMALL_STATE(6098)] = 224759,
  [SMALL_STATE(6099)] = 224776,
  [SMALL_STATE(6100)] = 224793,
  [SMALL_STATE(6101)] = 224810,
  [SMALL_STATE(6102)] = 224827,
  [SMALL_STATE(6103)] = 224844,
  [SMALL_STATE(6104)] = 224861,
  [SMALL_STATE(6105)] = 224878,
  [SMALL_STATE(6106)] = 224895,
  [SMALL_STATE(6107)] = 224910,
  [SMALL_STATE(6108)] = 224927,
  [SMALL_STATE(6109)] = 224944,
  [SMALL_STATE(6110)] = 224961,
  [SMALL_STATE(6111)] = 224978,
  [SMALL_STATE(6112)] = 224995,
  [SMALL_STATE(6113)] = 225010,
  [SMALL_STATE(6114)] = 225027,
  [SMALL_STATE(6115)] = 225044,
  [SMALL_STATE(6116)] = 225061,
  [SMALL_STATE(6117)] = 225078,
  [SMALL_STATE(6118)] = 225095,
  [SMALL_STATE(6119)] = 225112,
  [SMALL_STATE(6120)] = 225129,
  [SMALL_STATE(6121)] = 225146,
  [SMALL_STATE(6122)] = 225161,
  [SMALL_STATE(6123)] = 225178,
  [SMALL_STATE(6124)] = 225195,
  [SMALL_STATE(6125)] = 225212,
  [SMALL_STATE(6126)] = 225229,
  [SMALL_STATE(6127)] = 225244,
  [SMALL_STATE(6128)] = 225261,
  [SMALL_STATE(6129)] = 225278,
  [SMALL_STATE(6130)] = 225293,
  [SMALL_STATE(6131)] = 225310,
  [SMALL_STATE(6132)] = 225325,
  [SMALL_STATE(6133)] = 225342,
  [SMALL_STATE(6134)] = 225359,
  [SMALL_STATE(6135)] = 225376,
  [SMALL_STATE(6136)] = 225391,
  [SMALL_STATE(6137)] = 225408,
  [SMALL_STATE(6138)] = 225423,
  [SMALL_STATE(6139)] = 225440,
  [SMALL_STATE(6140)] = 225457,
  [SMALL_STATE(6141)] = 225474,
  [SMALL_STATE(6142)] = 225491,
  [SMALL_STATE(6143)] = 225508,
  [SMALL_STATE(6144)] = 225523,
  [SMALL_STATE(6145)] = 225538,
  [SMALL_STATE(6146)] = 225555,
  [SMALL_STATE(6147)] = 225570,
  [SMALL_STATE(6148)] = 225587,
  [SMALL_STATE(6149)] = 225604,
  [SMALL_STATE(6150)] = 225621,
  [SMALL_STATE(6151)] = 225638,
  [SMALL_STATE(6152)] = 225653,
  [SMALL_STATE(6153)] = 225670,
  [SMALL_STATE(6154)] = 225687,
  [SMALL_STATE(6155)] = 225704,
  [SMALL_STATE(6156)] = 225721,
  [SMALL_STATE(6157)] = 225738,
  [SMALL_STATE(6158)] = 225755,
  [SMALL_STATE(6159)] = 225772,
  [SMALL_STATE(6160)] = 225789,
  [SMALL_STATE(6161)] = 225806,
  [SMALL_STATE(6162)] = 225823,
  [SMALL_STATE(6163)] = 225840,
  [SMALL_STATE(6164)] = 225857,
  [SMALL_STATE(6165)] = 225874,
  [SMALL_STATE(6166)] = 225889,
  [SMALL_STATE(6167)] = 225906,
  [SMALL_STATE(6168)] = 225921,
  [SMALL_STATE(6169)] = 225936,
  [SMALL_STATE(6170)] = 225953,
  [SMALL_STATE(6171)] = 225970,
  [SMALL_STATE(6172)] = 225985,
  [SMALL_STATE(6173)] = 226002,
  [SMALL_STATE(6174)] = 226019,
  [SMALL_STATE(6175)] = 226036,
  [SMALL_STATE(6176)] = 226053,
  [SMALL_STATE(6177)] = 226070,
  [SMALL_STATE(6178)] = 226087,
  [SMALL_STATE(6179)] = 226104,
  [SMALL_STATE(6180)] = 226119,
  [SMALL_STATE(6181)] = 226136,
  [SMALL_STATE(6182)] = 226153,
  [SMALL_STATE(6183)] = 226170,
  [SMALL_STATE(6184)] = 226187,
  [SMALL_STATE(6185)] = 226204,
  [SMALL_STATE(6186)] = 226221,
  [SMALL_STATE(6187)] = 226238,
  [SMALL_STATE(6188)] = 226255,
  [SMALL_STATE(6189)] = 226272,
  [SMALL_STATE(6190)] = 226287,
  [SMALL_STATE(6191)] = 226304,
  [SMALL_STATE(6192)] = 226321,
  [SMALL_STATE(6193)] = 226338,
  [SMALL_STATE(6194)] = 226355,
  [SMALL_STATE(6195)] = 226372,
  [SMALL_STATE(6196)] = 226387,
  [SMALL_STATE(6197)] = 226404,
  [SMALL_STATE(6198)] = 226421,
  [SMALL_STATE(6199)] = 226438,
  [SMALL_STATE(6200)] = 226455,
  [SMALL_STATE(6201)] = 226472,
  [SMALL_STATE(6202)] = 226489,
  [SMALL_STATE(6203)] = 226504,
  [SMALL_STATE(6204)] = 226521,
  [SMALL_STATE(6205)] = 226538,
  [SMALL_STATE(6206)] = 226553,
  [SMALL_STATE(6207)] = 226570,
  [SMALL_STATE(6208)] = 226587,
  [SMALL_STATE(6209)] = 226604,
  [SMALL_STATE(6210)] = 226621,
  [SMALL_STATE(6211)] = 226638,
  [SMALL_STATE(6212)] = 226655,
  [SMALL_STATE(6213)] = 226670,
  [SMALL_STATE(6214)] = 226685,
  [SMALL_STATE(6215)] = 226702,
  [SMALL_STATE(6216)] = 226719,
  [SMALL_STATE(6217)] = 226734,
  [SMALL_STATE(6218)] = 226751,
  [SMALL_STATE(6219)] = 226768,
  [SMALL_STATE(6220)] = 226785,
  [SMALL_STATE(6221)] = 226802,
  [SMALL_STATE(6222)] = 226817,
  [SMALL_STATE(6223)] = 226832,
  [SMALL_STATE(6224)] = 226847,
  [SMALL_STATE(6225)] = 226862,
  [SMALL_STATE(6226)] = 226879,
  [SMALL_STATE(6227)] = 226896,
  [SMALL_STATE(6228)] = 226913,
  [SMALL_STATE(6229)] = 226930,
  [SMALL_STATE(6230)] = 226947,
  [SMALL_STATE(6231)] = 226964,
  [SMALL_STATE(6232)] = 226981,
  [SMALL_STATE(6233)] = 226998,
  [SMALL_STATE(6234)] = 227013,
  [SMALL_STATE(6235)] = 227030,
  [SMALL_STATE(6236)] = 227047,
  [SMALL_STATE(6237)] = 227064,
  [SMALL_STATE(6238)] = 227081,
  [SMALL_STATE(6239)] = 227096,
  [SMALL_STATE(6240)] = 227113,
  [SMALL_STATE(6241)] = 227130,
  [SMALL_STATE(6242)] = 227145,
  [SMALL_STATE(6243)] = 227162,
  [SMALL_STATE(6244)] = 227179,
  [SMALL_STATE(6245)] = 227196,
  [SMALL_STATE(6246)] = 227213,
  [SMALL_STATE(6247)] = 227228,
  [SMALL_STATE(6248)] = 227245,
  [SMALL_STATE(6249)] = 227262,
  [SMALL_STATE(6250)] = 227279,
  [SMALL_STATE(6251)] = 227296,
  [SMALL_STATE(6252)] = 227313,
  [SMALL_STATE(6253)] = 227330,
  [SMALL_STATE(6254)] = 227345,
  [SMALL_STATE(6255)] = 227362,
  [SMALL_STATE(6256)] = 227379,
  [SMALL_STATE(6257)] = 227396,
  [SMALL_STATE(6258)] = 227411,
  [SMALL_STATE(6259)] = 227428,
  [SMALL_STATE(6260)] = 227445,
  [SMALL_STATE(6261)] = 227460,
  [SMALL_STATE(6262)] = 227477,
  [SMALL_STATE(6263)] = 227494,
  [SMALL_STATE(6264)] = 227511,
  [SMALL_STATE(6265)] = 227528,
  [SMALL_STATE(6266)] = 227545,
  [SMALL_STATE(6267)] = 227560,
  [SMALL_STATE(6268)] = 227577,
  [SMALL_STATE(6269)] = 227592,
  [SMALL_STATE(6270)] = 227609,
  [SMALL_STATE(6271)] = 227626,
  [SMALL_STATE(6272)] = 227643,
  [SMALL_STATE(6273)] = 227660,
  [SMALL_STATE(6274)] = 227677,
  [SMALL_STATE(6275)] = 227694,
  [SMALL_STATE(6276)] = 227711,
  [SMALL_STATE(6277)] = 227728,
  [SMALL_STATE(6278)] = 227743,
  [SMALL_STATE(6279)] = 227758,
  [SMALL_STATE(6280)] = 227775,
  [SMALL_STATE(6281)] = 227790,
  [SMALL_STATE(6282)] = 227807,
  [SMALL_STATE(6283)] = 227824,
  [SMALL_STATE(6284)] = 227841,
  [SMALL_STATE(6285)] = 227858,
  [SMALL_STATE(6286)] = 227875,
  [SMALL_STATE(6287)] = 227892,
  [SMALL_STATE(6288)] = 227909,
  [SMALL_STATE(6289)] = 227926,
  [SMALL_STATE(6290)] = 227943,
  [SMALL_STATE(6291)] = 227960,
  [SMALL_STATE(6292)] = 227977,
  [SMALL_STATE(6293)] = 227994,
  [SMALL_STATE(6294)] = 228011,
  [SMALL_STATE(6295)] = 228026,
  [SMALL_STATE(6296)] = 228041,
  [SMALL_STATE(6297)] = 228058,
  [SMALL_STATE(6298)] = 228075,
  [SMALL_STATE(6299)] = 228092,
  [SMALL_STATE(6300)] = 228109,
  [SMALL_STATE(6301)] = 228126,
  [SMALL_STATE(6302)] = 228143,
  [SMALL_STATE(6303)] = 228160,
  [SMALL_STATE(6304)] = 228177,
  [SMALL_STATE(6305)] = 228194,
  [SMALL_STATE(6306)] = 228211,
  [SMALL_STATE(6307)] = 228228,
  [SMALL_STATE(6308)] = 228245,
  [SMALL_STATE(6309)] = 228262,
  [SMALL_STATE(6310)] = 228279,
  [SMALL_STATE(6311)] = 228296,
  [SMALL_STATE(6312)] = 228313,
  [SMALL_STATE(6313)] = 228330,
  [SMALL_STATE(6314)] = 228347,
  [SMALL_STATE(6315)] = 228362,
  [SMALL_STATE(6316)] = 228377,
  [SMALL_STATE(6317)] = 228392,
  [SMALL_STATE(6318)] = 228409,
  [SMALL_STATE(6319)] = 228426,
  [SMALL_STATE(6320)] = 228443,
  [SMALL_STATE(6321)] = 228458,
  [SMALL_STATE(6322)] = 228473,
  [SMALL_STATE(6323)] = 228488,
  [SMALL_STATE(6324)] = 228505,
  [SMALL_STATE(6325)] = 228520,
  [SMALL_STATE(6326)] = 228537,
  [SMALL_STATE(6327)] = 228552,
  [SMALL_STATE(6328)] = 228567,
  [SMALL_STATE(6329)] = 228582,
  [SMALL_STATE(6330)] = 228599,
  [SMALL_STATE(6331)] = 228616,
  [SMALL_STATE(6332)] = 228633,
  [SMALL_STATE(6333)] = 228648,
  [SMALL_STATE(6334)] = 228663,
  [SMALL_STATE(6335)] = 228680,
  [SMALL_STATE(6336)] = 228695,
  [SMALL_STATE(6337)] = 228712,
  [SMALL_STATE(6338)] = 228727,
  [SMALL_STATE(6339)] = 228741,
  [SMALL_STATE(6340)] = 228755,
  [SMALL_STATE(6341)] = 228769,
  [SMALL_STATE(6342)] = 228783,
  [SMALL_STATE(6343)] = 228797,
  [SMALL_STATE(6344)] = 228811,
  [SMALL_STATE(6345)] = 228825,
  [SMALL_STATE(6346)] = 228839,
  [SMALL_STATE(6347)] = 228853,
  [SMALL_STATE(6348)] = 228867,
  [SMALL_STATE(6349)] = 228881,
  [SMALL_STATE(6350)] = 228895,
  [SMALL_STATE(6351)] = 228909,
  [SMALL_STATE(6352)] = 228923,
  [SMALL_STATE(6353)] = 228937,
  [SMALL_STATE(6354)] = 228951,
  [SMALL_STATE(6355)] = 228965,
  [SMALL_STATE(6356)] = 228979,
  [SMALL_STATE(6357)] = 228993,
  [SMALL_STATE(6358)] = 229007,
  [SMALL_STATE(6359)] = 229021,
  [SMALL_STATE(6360)] = 229035,
  [SMALL_STATE(6361)] = 229049,
  [SMALL_STATE(6362)] = 229063,
  [SMALL_STATE(6363)] = 229077,
  [SMALL_STATE(6364)] = 229091,
  [SMALL_STATE(6365)] = 229105,
  [SMALL_STATE(6366)] = 229119,
  [SMALL_STATE(6367)] = 229133,
  [SMALL_STATE(6368)] = 229147,
  [SMALL_STATE(6369)] = 229161,
  [SMALL_STATE(6370)] = 229175,
  [SMALL_STATE(6371)] = 229189,
  [SMALL_STATE(6372)] = 229203,
  [SMALL_STATE(6373)] = 229217,
  [SMALL_STATE(6374)] = 229231,
  [SMALL_STATE(6375)] = 229245,
  [SMALL_STATE(6376)] = 229259,
  [SMALL_STATE(6377)] = 229273,
  [SMALL_STATE(6378)] = 229287,
  [SMALL_STATE(6379)] = 229301,
  [SMALL_STATE(6380)] = 229315,
  [SMALL_STATE(6381)] = 229329,
  [SMALL_STATE(6382)] = 229343,
  [SMALL_STATE(6383)] = 229357,
  [SMALL_STATE(6384)] = 229371,
  [SMALL_STATE(6385)] = 229385,
  [SMALL_STATE(6386)] = 229399,
  [SMALL_STATE(6387)] = 229413,
  [SMALL_STATE(6388)] = 229427,
  [SMALL_STATE(6389)] = 229441,
  [SMALL_STATE(6390)] = 229455,
  [SMALL_STATE(6391)] = 229469,
  [SMALL_STATE(6392)] = 229483,
  [SMALL_STATE(6393)] = 229497,
  [SMALL_STATE(6394)] = 229511,
  [SMALL_STATE(6395)] = 229525,
  [SMALL_STATE(6396)] = 229539,
  [SMALL_STATE(6397)] = 229553,
  [SMALL_STATE(6398)] = 229567,
  [SMALL_STATE(6399)] = 229581,
  [SMALL_STATE(6400)] = 229595,
  [SMALL_STATE(6401)] = 229609,
  [SMALL_STATE(6402)] = 229623,
  [SMALL_STATE(6403)] = 229637,
  [SMALL_STATE(6404)] = 229651,
  [SMALL_STATE(6405)] = 229665,
  [SMALL_STATE(6406)] = 229679,
  [SMALL_STATE(6407)] = 229693,
  [SMALL_STATE(6408)] = 229707,
  [SMALL_STATE(6409)] = 229721,
  [SMALL_STATE(6410)] = 229735,
  [SMALL_STATE(6411)] = 229749,
  [SMALL_STATE(6412)] = 229763,
  [SMALL_STATE(6413)] = 229777,
  [SMALL_STATE(6414)] = 229791,
  [SMALL_STATE(6415)] = 229805,
  [SMALL_STATE(6416)] = 229819,
  [SMALL_STATE(6417)] = 229833,
  [SMALL_STATE(6418)] = 229847,
  [SMALL_STATE(6419)] = 229861,
  [SMALL_STATE(6420)] = 229875,
  [SMALL_STATE(6421)] = 229889,
  [SMALL_STATE(6422)] = 229903,
  [SMALL_STATE(6423)] = 229917,
  [SMALL_STATE(6424)] = 229931,
  [SMALL_STATE(6425)] = 229945,
  [SMALL_STATE(6426)] = 229959,
  [SMALL_STATE(6427)] = 229973,
  [SMALL_STATE(6428)] = 229987,
  [SMALL_STATE(6429)] = 230001,
  [SMALL_STATE(6430)] = 230015,
  [SMALL_STATE(6431)] = 230029,
  [SMALL_STATE(6432)] = 230043,
  [SMALL_STATE(6433)] = 230057,
  [SMALL_STATE(6434)] = 230071,
  [SMALL_STATE(6435)] = 230085,
  [SMALL_STATE(6436)] = 230099,
  [SMALL_STATE(6437)] = 230113,
  [SMALL_STATE(6438)] = 230127,
  [SMALL_STATE(6439)] = 230141,
  [SMALL_STATE(6440)] = 230155,
  [SMALL_STATE(6441)] = 230169,
  [SMALL_STATE(6442)] = 230183,
  [SMALL_STATE(6443)] = 230197,
  [SMALL_STATE(6444)] = 230211,
  [SMALL_STATE(6445)] = 230225,
  [SMALL_STATE(6446)] = 230239,
  [SMALL_STATE(6447)] = 230253,
  [SMALL_STATE(6448)] = 230267,
  [SMALL_STATE(6449)] = 230281,
  [SMALL_STATE(6450)] = 230295,
  [SMALL_STATE(6451)] = 230309,
  [SMALL_STATE(6452)] = 230323,
  [SMALL_STATE(6453)] = 230337,
  [SMALL_STATE(6454)] = 230351,
  [SMALL_STATE(6455)] = 230365,
  [SMALL_STATE(6456)] = 230379,
  [SMALL_STATE(6457)] = 230393,
  [SMALL_STATE(6458)] = 230407,
  [SMALL_STATE(6459)] = 230421,
  [SMALL_STATE(6460)] = 230435,
  [SMALL_STATE(6461)] = 230449,
  [SMALL_STATE(6462)] = 230463,
  [SMALL_STATE(6463)] = 230477,
  [SMALL_STATE(6464)] = 230491,
  [SMALL_STATE(6465)] = 230505,
  [SMALL_STATE(6466)] = 230519,
  [SMALL_STATE(6467)] = 230533,
  [SMALL_STATE(6468)] = 230547,
  [SMALL_STATE(6469)] = 230561,
  [SMALL_STATE(6470)] = 230575,
  [SMALL_STATE(6471)] = 230589,
  [SMALL_STATE(6472)] = 230603,
  [SMALL_STATE(6473)] = 230617,
  [SMALL_STATE(6474)] = 230631,
  [SMALL_STATE(6475)] = 230645,
  [SMALL_STATE(6476)] = 230659,
  [SMALL_STATE(6477)] = 230673,
  [SMALL_STATE(6478)] = 230687,
  [SMALL_STATE(6479)] = 230701,
  [SMALL_STATE(6480)] = 230715,
  [SMALL_STATE(6481)] = 230729,
  [SMALL_STATE(6482)] = 230743,
  [SMALL_STATE(6483)] = 230757,
  [SMALL_STATE(6484)] = 230771,
  [SMALL_STATE(6485)] = 230785,
  [SMALL_STATE(6486)] = 230799,
  [SMALL_STATE(6487)] = 230813,
  [SMALL_STATE(6488)] = 230827,
  [SMALL_STATE(6489)] = 230841,
  [SMALL_STATE(6490)] = 230855,
  [SMALL_STATE(6491)] = 230869,
  [SMALL_STATE(6492)] = 230883,
  [SMALL_STATE(6493)] = 230897,
  [SMALL_STATE(6494)] = 230911,
  [SMALL_STATE(6495)] = 230925,
  [SMALL_STATE(6496)] = 230939,
  [SMALL_STATE(6497)] = 230953,
  [SMALL_STATE(6498)] = 230967,
  [SMALL_STATE(6499)] = 230981,
  [SMALL_STATE(6500)] = 230995,
  [SMALL_STATE(6501)] = 231009,
  [SMALL_STATE(6502)] = 231023,
  [SMALL_STATE(6503)] = 231037,
  [SMALL_STATE(6504)] = 231051,
  [SMALL_STATE(6505)] = 231065,
  [SMALL_STATE(6506)] = 231079,
  [SMALL_STATE(6507)] = 231093,
  [SMALL_STATE(6508)] = 231107,
  [SMALL_STATE(6509)] = 231121,
  [SMALL_STATE(6510)] = 231135,
  [SMALL_STATE(6511)] = 231149,
  [SMALL_STATE(6512)] = 231163,
  [SMALL_STATE(6513)] = 231177,
  [SMALL_STATE(6514)] = 231191,
  [SMALL_STATE(6515)] = 231205,
  [SMALL_STATE(6516)] = 231219,
  [SMALL_STATE(6517)] = 231233,
  [SMALL_STATE(6518)] = 231247,
  [SMALL_STATE(6519)] = 231261,
  [SMALL_STATE(6520)] = 231275,
  [SMALL_STATE(6521)] = 231289,
  [SMALL_STATE(6522)] = 231303,
  [SMALL_STATE(6523)] = 231317,
  [SMALL_STATE(6524)] = 231331,
  [SMALL_STATE(6525)] = 231345,
  [SMALL_STATE(6526)] = 231359,
  [SMALL_STATE(6527)] = 231373,
  [SMALL_STATE(6528)] = 231387,
  [SMALL_STATE(6529)] = 231401,
  [SMALL_STATE(6530)] = 231415,
  [SMALL_STATE(6531)] = 231429,
  [SMALL_STATE(6532)] = 231443,
  [SMALL_STATE(6533)] = 231457,
  [SMALL_STATE(6534)] = 231471,
  [SMALL_STATE(6535)] = 231485,
  [SMALL_STATE(6536)] = 231499,
  [SMALL_STATE(6537)] = 231513,
  [SMALL_STATE(6538)] = 231527,
  [SMALL_STATE(6539)] = 231541,
  [SMALL_STATE(6540)] = 231555,
  [SMALL_STATE(6541)] = 231569,
  [SMALL_STATE(6542)] = 231583,
  [SMALL_STATE(6543)] = 231597,
  [SMALL_STATE(6544)] = 231611,
  [SMALL_STATE(6545)] = 231625,
  [SMALL_STATE(6546)] = 231639,
  [SMALL_STATE(6547)] = 231653,
  [SMALL_STATE(6548)] = 231667,
  [SMALL_STATE(6549)] = 231681,
  [SMALL_STATE(6550)] = 231695,
  [SMALL_STATE(6551)] = 231709,
  [SMALL_STATE(6552)] = 231723,
  [SMALL_STATE(6553)] = 231737,
  [SMALL_STATE(6554)] = 231751,
  [SMALL_STATE(6555)] = 231765,
  [SMALL_STATE(6556)] = 231779,
  [SMALL_STATE(6557)] = 231793,
  [SMALL_STATE(6558)] = 231807,
  [SMALL_STATE(6559)] = 231821,
  [SMALL_STATE(6560)] = 231835,
  [SMALL_STATE(6561)] = 231849,
  [SMALL_STATE(6562)] = 231863,
  [SMALL_STATE(6563)] = 231877,
  [SMALL_STATE(6564)] = 231891,
  [SMALL_STATE(6565)] = 231905,
  [SMALL_STATE(6566)] = 231919,
  [SMALL_STATE(6567)] = 231933,
  [SMALL_STATE(6568)] = 231947,
  [SMALL_STATE(6569)] = 231961,
  [SMALL_STATE(6570)] = 231975,
  [SMALL_STATE(6571)] = 231989,
  [SMALL_STATE(6572)] = 232003,
  [SMALL_STATE(6573)] = 232017,
  [SMALL_STATE(6574)] = 232031,
  [SMALL_STATE(6575)] = 232045,
  [SMALL_STATE(6576)] = 232059,
  [SMALL_STATE(6577)] = 232073,
  [SMALL_STATE(6578)] = 232087,
  [SMALL_STATE(6579)] = 232101,
  [SMALL_STATE(6580)] = 232115,
  [SMALL_STATE(6581)] = 232129,
  [SMALL_STATE(6582)] = 232143,
  [SMALL_STATE(6583)] = 232157,
  [SMALL_STATE(6584)] = 232171,
  [SMALL_STATE(6585)] = 232185,
  [SMALL_STATE(6586)] = 232199,
  [SMALL_STATE(6587)] = 232213,
  [SMALL_STATE(6588)] = 232227,
  [SMALL_STATE(6589)] = 232241,
  [SMALL_STATE(6590)] = 232255,
  [SMALL_STATE(6591)] = 232269,
  [SMALL_STATE(6592)] = 232283,
  [SMALL_STATE(6593)] = 232297,
  [SMALL_STATE(6594)] = 232311,
  [SMALL_STATE(6595)] = 232325,
  [SMALL_STATE(6596)] = 232339,
  [SMALL_STATE(6597)] = 232353,
  [SMALL_STATE(6598)] = 232367,
  [SMALL_STATE(6599)] = 232381,
  [SMALL_STATE(6600)] = 232395,
  [SMALL_STATE(6601)] = 232409,
  [SMALL_STATE(6602)] = 232423,
  [SMALL_STATE(6603)] = 232437,
  [SMALL_STATE(6604)] = 232451,
  [SMALL_STATE(6605)] = 232465,
  [SMALL_STATE(6606)] = 232479,
  [SMALL_STATE(6607)] = 232493,
  [SMALL_STATE(6608)] = 232507,
  [SMALL_STATE(6609)] = 232521,
  [SMALL_STATE(6610)] = 232535,
  [SMALL_STATE(6611)] = 232549,
  [SMALL_STATE(6612)] = 232563,
  [SMALL_STATE(6613)] = 232577,
  [SMALL_STATE(6614)] = 232591,
  [SMALL_STATE(6615)] = 232605,
  [SMALL_STATE(6616)] = 232619,
  [SMALL_STATE(6617)] = 232633,
  [SMALL_STATE(6618)] = 232647,
  [SMALL_STATE(6619)] = 232661,
  [SMALL_STATE(6620)] = 232675,
  [SMALL_STATE(6621)] = 232689,
  [SMALL_STATE(6622)] = 232703,
  [SMALL_STATE(6623)] = 232717,
  [SMALL_STATE(6624)] = 232731,
  [SMALL_STATE(6625)] = 232745,
  [SMALL_STATE(6626)] = 232759,
  [SMALL_STATE(6627)] = 232773,
  [SMALL_STATE(6628)] = 232787,
  [SMALL_STATE(6629)] = 232801,
  [SMALL_STATE(6630)] = 232815,
  [SMALL_STATE(6631)] = 232829,
  [SMALL_STATE(6632)] = 232843,
  [SMALL_STATE(6633)] = 232857,
  [SMALL_STATE(6634)] = 232871,
  [SMALL_STATE(6635)] = 232885,
  [SMALL_STATE(6636)] = 232899,
  [SMALL_STATE(6637)] = 232913,
  [SMALL_STATE(6638)] = 232927,
  [SMALL_STATE(6639)] = 232941,
  [SMALL_STATE(6640)] = 232955,
  [SMALL_STATE(6641)] = 232969,
  [SMALL_STATE(6642)] = 232983,
  [SMALL_STATE(6643)] = 232997,
  [SMALL_STATE(6644)] = 233011,
  [SMALL_STATE(6645)] = 233025,
  [SMALL_STATE(6646)] = 233039,
  [SMALL_STATE(6647)] = 233053,
  [SMALL_STATE(6648)] = 233067,
  [SMALL_STATE(6649)] = 233081,
  [SMALL_STATE(6650)] = 233095,
  [SMALL_STATE(6651)] = 233109,
  [SMALL_STATE(6652)] = 233123,
  [SMALL_STATE(6653)] = 233137,
  [SMALL_STATE(6654)] = 233151,
  [SMALL_STATE(6655)] = 233165,
  [SMALL_STATE(6656)] = 233179,
  [SMALL_STATE(6657)] = 233193,
  [SMALL_STATE(6658)] = 233207,
  [SMALL_STATE(6659)] = 233221,
  [SMALL_STATE(6660)] = 233235,
  [SMALL_STATE(6661)] = 233249,
  [SMALL_STATE(6662)] = 233263,
  [SMALL_STATE(6663)] = 233277,
  [SMALL_STATE(6664)] = 233291,
  [SMALL_STATE(6665)] = 233305,
  [SMALL_STATE(6666)] = 233319,
  [SMALL_STATE(6667)] = 233333,
  [SMALL_STATE(6668)] = 233347,
  [SMALL_STATE(6669)] = 233361,
  [SMALL_STATE(6670)] = 233375,
  [SMALL_STATE(6671)] = 233389,
  [SMALL_STATE(6672)] = 233403,
  [SMALL_STATE(6673)] = 233417,
  [SMALL_STATE(6674)] = 233431,
  [SMALL_STATE(6675)] = 233445,
  [SMALL_STATE(6676)] = 233459,
  [SMALL_STATE(6677)] = 233473,
  [SMALL_STATE(6678)] = 233487,
  [SMALL_STATE(6679)] = 233501,
  [SMALL_STATE(6680)] = 233515,
  [SMALL_STATE(6681)] = 233529,
  [SMALL_STATE(6682)] = 233543,
  [SMALL_STATE(6683)] = 233557,
  [SMALL_STATE(6684)] = 233571,
  [SMALL_STATE(6685)] = 233585,
  [SMALL_STATE(6686)] = 233599,
  [SMALL_STATE(6687)] = 233613,
  [SMALL_STATE(6688)] = 233627,
  [SMALL_STATE(6689)] = 233641,
  [SMALL_STATE(6690)] = 233655,
  [SMALL_STATE(6691)] = 233669,
  [SMALL_STATE(6692)] = 233683,
  [SMALL_STATE(6693)] = 233697,
  [SMALL_STATE(6694)] = 233711,
  [SMALL_STATE(6695)] = 233725,
  [SMALL_STATE(6696)] = 233739,
  [SMALL_STATE(6697)] = 233753,
  [SMALL_STATE(6698)] = 233767,
  [SMALL_STATE(6699)] = 233781,
  [SMALL_STATE(6700)] = 233795,
  [SMALL_STATE(6701)] = 233809,
  [SMALL_STATE(6702)] = 233823,
  [SMALL_STATE(6703)] = 233837,
  [SMALL_STATE(6704)] = 233851,
  [SMALL_STATE(6705)] = 233865,
  [SMALL_STATE(6706)] = 233879,
  [SMALL_STATE(6707)] = 233893,
  [SMALL_STATE(6708)] = 233907,
  [SMALL_STATE(6709)] = 233921,
  [SMALL_STATE(6710)] = 233935,
  [SMALL_STATE(6711)] = 233949,
  [SMALL_STATE(6712)] = 233963,
  [SMALL_STATE(6713)] = 233977,
  [SMALL_STATE(6714)] = 233991,
  [SMALL_STATE(6715)] = 234005,
  [SMALL_STATE(6716)] = 234019,
  [SMALL_STATE(6717)] = 234033,
  [SMALL_STATE(6718)] = 234047,
  [SMALL_STATE(6719)] = 234061,
  [SMALL_STATE(6720)] = 234075,
  [SMALL_STATE(6721)] = 234089,
  [SMALL_STATE(6722)] = 234103,
  [SMALL_STATE(6723)] = 234117,
  [SMALL_STATE(6724)] = 234131,
  [SMALL_STATE(6725)] = 234145,
  [SMALL_STATE(6726)] = 234159,
  [SMALL_STATE(6727)] = 234173,
  [SMALL_STATE(6728)] = 234187,
  [SMALL_STATE(6729)] = 234201,
  [SMALL_STATE(6730)] = 234215,
  [SMALL_STATE(6731)] = 234229,
  [SMALL_STATE(6732)] = 234243,
  [SMALL_STATE(6733)] = 234257,
  [SMALL_STATE(6734)] = 234271,
  [SMALL_STATE(6735)] = 234285,
  [SMALL_STATE(6736)] = 234299,
  [SMALL_STATE(6737)] = 234313,
  [SMALL_STATE(6738)] = 234327,
  [SMALL_STATE(6739)] = 234341,
  [SMALL_STATE(6740)] = 234355,
  [SMALL_STATE(6741)] = 234369,
  [SMALL_STATE(6742)] = 234383,
  [SMALL_STATE(6743)] = 234397,
  [SMALL_STATE(6744)] = 234411,
  [SMALL_STATE(6745)] = 234425,
  [SMALL_STATE(6746)] = 234439,
  [SMALL_STATE(6747)] = 234453,
  [SMALL_STATE(6748)] = 234467,
  [SMALL_STATE(6749)] = 234481,
  [SMALL_STATE(6750)] = 234495,
  [SMALL_STATE(6751)] = 234509,
  [SMALL_STATE(6752)] = 234523,
  [SMALL_STATE(6753)] = 234537,
  [SMALL_STATE(6754)] = 234551,
  [SMALL_STATE(6755)] = 234565,
  [SMALL_STATE(6756)] = 234579,
  [SMALL_STATE(6757)] = 234593,
  [SMALL_STATE(6758)] = 234607,
  [SMALL_STATE(6759)] = 234621,
  [SMALL_STATE(6760)] = 234635,
  [SMALL_STATE(6761)] = 234649,
  [SMALL_STATE(6762)] = 234663,
  [SMALL_STATE(6763)] = 234677,
  [SMALL_STATE(6764)] = 234691,
  [SMALL_STATE(6765)] = 234705,
  [SMALL_STATE(6766)] = 234719,
  [SMALL_STATE(6767)] = 234733,
  [SMALL_STATE(6768)] = 234747,
  [SMALL_STATE(6769)] = 234761,
  [SMALL_STATE(6770)] = 234775,
  [SMALL_STATE(6771)] = 234789,
  [SMALL_STATE(6772)] = 234803,
  [SMALL_STATE(6773)] = 234817,
  [SMALL_STATE(6774)] = 234831,
  [SMALL_STATE(6775)] = 234845,
  [SMALL_STATE(6776)] = 234859,
  [SMALL_STATE(6777)] = 234873,
  [SMALL_STATE(6778)] = 234887,
  [SMALL_STATE(6779)] = 234901,
  [SMALL_STATE(6780)] = 234915,
  [SMALL_STATE(6781)] = 234929,
  [SMALL_STATE(6782)] = 234943,
  [SMALL_STATE(6783)] = 234957,
  [SMALL_STATE(6784)] = 234971,
  [SMALL_STATE(6785)] = 234985,
  [SMALL_STATE(6786)] = 234999,
  [SMALL_STATE(6787)] = 235013,
  [SMALL_STATE(6788)] = 235027,
  [SMALL_STATE(6789)] = 235041,
  [SMALL_STATE(6790)] = 235055,
  [SMALL_STATE(6791)] = 235069,
  [SMALL_STATE(6792)] = 235083,
  [SMALL_STATE(6793)] = 235097,
  [SMALL_STATE(6794)] = 235111,
  [SMALL_STATE(6795)] = 235125,
  [SMALL_STATE(6796)] = 235139,
  [SMALL_STATE(6797)] = 235153,
  [SMALL_STATE(6798)] = 235167,
  [SMALL_STATE(6799)] = 235181,
  [SMALL_STATE(6800)] = 235195,
  [SMALL_STATE(6801)] = 235209,
  [SMALL_STATE(6802)] = 235223,
  [SMALL_STATE(6803)] = 235237,
  [SMALL_STATE(6804)] = 235251,
  [SMALL_STATE(6805)] = 235265,
  [SMALL_STATE(6806)] = 235279,
  [SMALL_STATE(6807)] = 235293,
  [SMALL_STATE(6808)] = 235307,
  [SMALL_STATE(6809)] = 235321,
  [SMALL_STATE(6810)] = 235335,
  [SMALL_STATE(6811)] = 235349,
  [SMALL_STATE(6812)] = 235363,
  [SMALL_STATE(6813)] = 235377,
  [SMALL_STATE(6814)] = 235391,
  [SMALL_STATE(6815)] = 235405,
  [SMALL_STATE(6816)] = 235419,
  [SMALL_STATE(6817)] = 235433,
  [SMALL_STATE(6818)] = 235447,
  [SMALL_STATE(6819)] = 235461,
  [SMALL_STATE(6820)] = 235475,
  [SMALL_STATE(6821)] = 235489,
  [SMALL_STATE(6822)] = 235503,
  [SMALL_STATE(6823)] = 235517,
  [SMALL_STATE(6824)] = 235531,
  [SMALL_STATE(6825)] = 235545,
  [SMALL_STATE(6826)] = 235559,
  [SMALL_STATE(6827)] = 235573,
  [SMALL_STATE(6828)] = 235587,
  [SMALL_STATE(6829)] = 235601,
  [SMALL_STATE(6830)] = 235615,
  [SMALL_STATE(6831)] = 235629,
  [SMALL_STATE(6832)] = 235643,
  [SMALL_STATE(6833)] = 235657,
  [SMALL_STATE(6834)] = 235671,
  [SMALL_STATE(6835)] = 235685,
  [SMALL_STATE(6836)] = 235699,
  [SMALL_STATE(6837)] = 235713,
  [SMALL_STATE(6838)] = 235727,
  [SMALL_STATE(6839)] = 235741,
  [SMALL_STATE(6840)] = 235755,
  [SMALL_STATE(6841)] = 235769,
  [SMALL_STATE(6842)] = 235783,
  [SMALL_STATE(6843)] = 235797,
  [SMALL_STATE(6844)] = 235811,
  [SMALL_STATE(6845)] = 235825,
  [SMALL_STATE(6846)] = 235839,
  [SMALL_STATE(6847)] = 235853,
  [SMALL_STATE(6848)] = 235867,
  [SMALL_STATE(6849)] = 235881,
  [SMALL_STATE(6850)] = 235895,
  [SMALL_STATE(6851)] = 235909,
  [SMALL_STATE(6852)] = 235923,
  [SMALL_STATE(6853)] = 235937,
  [SMALL_STATE(6854)] = 235951,
  [SMALL_STATE(6855)] = 235965,
  [SMALL_STATE(6856)] = 235979,
  [SMALL_STATE(6857)] = 235993,
  [SMALL_STATE(6858)] = 236007,
  [SMALL_STATE(6859)] = 236021,
  [SMALL_STATE(6860)] = 236035,
  [SMALL_STATE(6861)] = 236049,
  [SMALL_STATE(6862)] = 236063,
  [SMALL_STATE(6863)] = 236077,
  [SMALL_STATE(6864)] = 236091,
  [SMALL_STATE(6865)] = 236105,
  [SMALL_STATE(6866)] = 236119,
  [SMALL_STATE(6867)] = 236133,
  [SMALL_STATE(6868)] = 236147,
  [SMALL_STATE(6869)] = 236161,
  [SMALL_STATE(6870)] = 236175,
  [SMALL_STATE(6871)] = 236189,
  [SMALL_STATE(6872)] = 236203,
  [SMALL_STATE(6873)] = 236217,
  [SMALL_STATE(6874)] = 236231,
  [SMALL_STATE(6875)] = 236245,
  [SMALL_STATE(6876)] = 236259,
  [SMALL_STATE(6877)] = 236273,
  [SMALL_STATE(6878)] = 236287,
  [SMALL_STATE(6879)] = 236301,
  [SMALL_STATE(6880)] = 236315,
  [SMALL_STATE(6881)] = 236329,
  [SMALL_STATE(6882)] = 236343,
  [SMALL_STATE(6883)] = 236357,
  [SMALL_STATE(6884)] = 236371,
  [SMALL_STATE(6885)] = 236385,
  [SMALL_STATE(6886)] = 236399,
  [SMALL_STATE(6887)] = 236413,
  [SMALL_STATE(6888)] = 236427,
  [SMALL_STATE(6889)] = 236441,
  [SMALL_STATE(6890)] = 236455,
  [SMALL_STATE(6891)] = 236469,
  [SMALL_STATE(6892)] = 236483,
  [SMALL_STATE(6893)] = 236497,
  [SMALL_STATE(6894)] = 236511,
  [SMALL_STATE(6895)] = 236525,
  [SMALL_STATE(6896)] = 236539,
  [SMALL_STATE(6897)] = 236553,
  [SMALL_STATE(6898)] = 236567,
  [SMALL_STATE(6899)] = 236581,
  [SMALL_STATE(6900)] = 236595,
  [SMALL_STATE(6901)] = 236609,
  [SMALL_STATE(6902)] = 236623,
  [SMALL_STATE(6903)] = 236637,
  [SMALL_STATE(6904)] = 236651,
  [SMALL_STATE(6905)] = 236665,
  [SMALL_STATE(6906)] = 236679,
  [SMALL_STATE(6907)] = 236693,
  [SMALL_STATE(6908)] = 236707,
  [SMALL_STATE(6909)] = 236721,
  [SMALL_STATE(6910)] = 236735,
  [SMALL_STATE(6911)] = 236749,
  [SMALL_STATE(6912)] = 236763,
  [SMALL_STATE(6913)] = 236777,
  [SMALL_STATE(6914)] = 236791,
  [SMALL_STATE(6915)] = 236805,
  [SMALL_STATE(6916)] = 236819,
  [SMALL_STATE(6917)] = 236833,
  [SMALL_STATE(6918)] = 236847,
  [SMALL_STATE(6919)] = 236861,
  [SMALL_STATE(6920)] = 236875,
  [SMALL_STATE(6921)] = 236889,
  [SMALL_STATE(6922)] = 236903,
  [SMALL_STATE(6923)] = 236917,
  [SMALL_STATE(6924)] = 236931,
  [SMALL_STATE(6925)] = 236945,
  [SMALL_STATE(6926)] = 236959,
  [SMALL_STATE(6927)] = 236973,
  [SMALL_STATE(6928)] = 236987,
  [SMALL_STATE(6929)] = 237001,
  [SMALL_STATE(6930)] = 237015,
  [SMALL_STATE(6931)] = 237029,
  [SMALL_STATE(6932)] = 237043,
  [SMALL_STATE(6933)] = 237057,
  [SMALL_STATE(6934)] = 237071,
  [SMALL_STATE(6935)] = 237085,
  [SMALL_STATE(6936)] = 237099,
  [SMALL_STATE(6937)] = 237113,
  [SMALL_STATE(6938)] = 237127,
  [SMALL_STATE(6939)] = 237141,
  [SMALL_STATE(6940)] = 237155,
  [SMALL_STATE(6941)] = 237169,
  [SMALL_STATE(6942)] = 237183,
  [SMALL_STATE(6943)] = 237197,
  [SMALL_STATE(6944)] = 237211,
  [SMALL_STATE(6945)] = 237225,
  [SMALL_STATE(6946)] = 237239,
  [SMALL_STATE(6947)] = 237253,
  [SMALL_STATE(6948)] = 237267,
  [SMALL_STATE(6949)] = 237281,
  [SMALL_STATE(6950)] = 237295,
  [SMALL_STATE(6951)] = 237309,
  [SMALL_STATE(6952)] = 237323,
  [SMALL_STATE(6953)] = 237337,
  [SMALL_STATE(6954)] = 237351,
  [SMALL_STATE(6955)] = 237365,
  [SMALL_STATE(6956)] = 237379,
  [SMALL_STATE(6957)] = 237393,
  [SMALL_STATE(6958)] = 237407,
  [SMALL_STATE(6959)] = 237421,
  [SMALL_STATE(6960)] = 237435,
  [SMALL_STATE(6961)] = 237449,
  [SMALL_STATE(6962)] = 237463,
  [SMALL_STATE(6963)] = 237477,
  [SMALL_STATE(6964)] = 237491,
  [SMALL_STATE(6965)] = 237505,
  [SMALL_STATE(6966)] = 237519,
  [SMALL_STATE(6967)] = 237533,
  [SMALL_STATE(6968)] = 237547,
  [SMALL_STATE(6969)] = 237561,
  [SMALL_STATE(6970)] = 237575,
  [SMALL_STATE(6971)] = 237589,
  [SMALL_STATE(6972)] = 237603,
  [SMALL_STATE(6973)] = 237617,
  [SMALL_STATE(6974)] = 237631,
  [SMALL_STATE(6975)] = 237645,
  [SMALL_STATE(6976)] = 237659,
  [SMALL_STATE(6977)] = 237673,
  [SMALL_STATE(6978)] = 237687,
  [SMALL_STATE(6979)] = 237701,
  [SMALL_STATE(6980)] = 237715,
  [SMALL_STATE(6981)] = 237729,
  [SMALL_STATE(6982)] = 237743,
  [SMALL_STATE(6983)] = 237757,
  [SMALL_STATE(6984)] = 237771,
  [SMALL_STATE(6985)] = 237785,
  [SMALL_STATE(6986)] = 237799,
  [SMALL_STATE(6987)] = 237813,
  [SMALL_STATE(6988)] = 237827,
  [SMALL_STATE(6989)] = 237841,
  [SMALL_STATE(6990)] = 237855,
  [SMALL_STATE(6991)] = 237869,
  [SMALL_STATE(6992)] = 237883,
  [SMALL_STATE(6993)] = 237897,
  [SMALL_STATE(6994)] = 237911,
  [SMALL_STATE(6995)] = 237925,
  [SMALL_STATE(6996)] = 237939,
  [SMALL_STATE(6997)] = 237953,
  [SMALL_STATE(6998)] = 237967,
  [SMALL_STATE(6999)] = 237981,
  [SMALL_STATE(7000)] = 237995,
  [SMALL_STATE(7001)] = 238009,
  [SMALL_STATE(7002)] = 238023,
  [SMALL_STATE(7003)] = 238037,
  [SMALL_STATE(7004)] = 238051,
  [SMALL_STATE(7005)] = 238065,
  [SMALL_STATE(7006)] = 238079,
  [SMALL_STATE(7007)] = 238093,
  [SMALL_STATE(7008)] = 238107,
  [SMALL_STATE(7009)] = 238121,
  [SMALL_STATE(7010)] = 238135,
  [SMALL_STATE(7011)] = 238149,
  [SMALL_STATE(7012)] = 238163,
  [SMALL_STATE(7013)] = 238177,
  [SMALL_STATE(7014)] = 238191,
  [SMALL_STATE(7015)] = 238205,
  [SMALL_STATE(7016)] = 238219,
  [SMALL_STATE(7017)] = 238233,
  [SMALL_STATE(7018)] = 238247,
  [SMALL_STATE(7019)] = 238261,
  [SMALL_STATE(7020)] = 238275,
  [SMALL_STATE(7021)] = 238289,
  [SMALL_STATE(7022)] = 238303,
  [SMALL_STATE(7023)] = 238317,
  [SMALL_STATE(7024)] = 238331,
  [SMALL_STATE(7025)] = 238345,
  [SMALL_STATE(7026)] = 238359,
  [SMALL_STATE(7027)] = 238373,
  [SMALL_STATE(7028)] = 238387,
  [SMALL_STATE(7029)] = 238401,
  [SMALL_STATE(7030)] = 238415,
  [SMALL_STATE(7031)] = 238429,
  [SMALL_STATE(7032)] = 238443,
  [SMALL_STATE(7033)] = 238457,
  [SMALL_STATE(7034)] = 238471,
  [SMALL_STATE(7035)] = 238485,
  [SMALL_STATE(7036)] = 238499,
  [SMALL_STATE(7037)] = 238513,
  [SMALL_STATE(7038)] = 238527,
  [SMALL_STATE(7039)] = 238541,
  [SMALL_STATE(7040)] = 238555,
  [SMALL_STATE(7041)] = 238569,
  [SMALL_STATE(7042)] = 238583,
  [SMALL_STATE(7043)] = 238597,
  [SMALL_STATE(7044)] = 238611,
  [SMALL_STATE(7045)] = 238625,
  [SMALL_STATE(7046)] = 238639,
  [SMALL_STATE(7047)] = 238653,
  [SMALL_STATE(7048)] = 238667,
  [SMALL_STATE(7049)] = 238681,
  [SMALL_STATE(7050)] = 238695,
  [SMALL_STATE(7051)] = 238709,
  [SMALL_STATE(7052)] = 238723,
  [SMALL_STATE(7053)] = 238737,
  [SMALL_STATE(7054)] = 238751,
  [SMALL_STATE(7055)] = 238765,
  [SMALL_STATE(7056)] = 238779,
  [SMALL_STATE(7057)] = 238793,
  [SMALL_STATE(7058)] = 238807,
  [SMALL_STATE(7059)] = 238821,
  [SMALL_STATE(7060)] = 238835,
  [SMALL_STATE(7061)] = 238849,
  [SMALL_STATE(7062)] = 238863,
  [SMALL_STATE(7063)] = 238877,
  [SMALL_STATE(7064)] = 238891,
  [SMALL_STATE(7065)] = 238905,
  [SMALL_STATE(7066)] = 238919,
  [SMALL_STATE(7067)] = 238933,
  [SMALL_STATE(7068)] = 238947,
  [SMALL_STATE(7069)] = 238961,
  [SMALL_STATE(7070)] = 238975,
  [SMALL_STATE(7071)] = 238989,
  [SMALL_STATE(7072)] = 239003,
  [SMALL_STATE(7073)] = 239017,
  [SMALL_STATE(7074)] = 239031,
  [SMALL_STATE(7075)] = 239045,
  [SMALL_STATE(7076)] = 239059,
  [SMALL_STATE(7077)] = 239073,
  [SMALL_STATE(7078)] = 239087,
  [SMALL_STATE(7079)] = 239101,
  [SMALL_STATE(7080)] = 239115,
  [SMALL_STATE(7081)] = 239129,
  [SMALL_STATE(7082)] = 239143,
  [SMALL_STATE(7083)] = 239157,
  [SMALL_STATE(7084)] = 239171,
  [SMALL_STATE(7085)] = 239185,
  [SMALL_STATE(7086)] = 239199,
  [SMALL_STATE(7087)] = 239213,
  [SMALL_STATE(7088)] = 239227,
  [SMALL_STATE(7089)] = 239241,
  [SMALL_STATE(7090)] = 239255,
  [SMALL_STATE(7091)] = 239269,
  [SMALL_STATE(7092)] = 239283,
  [SMALL_STATE(7093)] = 239297,
  [SMALL_STATE(7094)] = 239311,
  [SMALL_STATE(7095)] = 239325,
  [SMALL_STATE(7096)] = 239339,
  [SMALL_STATE(7097)] = 239353,
  [SMALL_STATE(7098)] = 239367,
  [SMALL_STATE(7099)] = 239381,
  [SMALL_STATE(7100)] = 239395,
  [SMALL_STATE(7101)] = 239409,
  [SMALL_STATE(7102)] = 239423,
  [SMALL_STATE(7103)] = 239437,
  [SMALL_STATE(7104)] = 239451,
  [SMALL_STATE(7105)] = 239465,
  [SMALL_STATE(7106)] = 239479,
  [SMALL_STATE(7107)] = 239493,
  [SMALL_STATE(7108)] = 239507,
  [SMALL_STATE(7109)] = 239521,
  [SMALL_STATE(7110)] = 239535,
  [SMALL_STATE(7111)] = 239549,
  [SMALL_STATE(7112)] = 239563,
  [SMALL_STATE(7113)] = 239577,
  [SMALL_STATE(7114)] = 239591,
  [SMALL_STATE(7115)] = 239605,
  [SMALL_STATE(7116)] = 239619,
  [SMALL_STATE(7117)] = 239633,
  [SMALL_STATE(7118)] = 239647,
  [SMALL_STATE(7119)] = 239661,
  [SMALL_STATE(7120)] = 239675,
  [SMALL_STATE(7121)] = 239689,
  [SMALL_STATE(7122)] = 239703,
  [SMALL_STATE(7123)] = 239717,
  [SMALL_STATE(7124)] = 239731,
  [SMALL_STATE(7125)] = 239745,
  [SMALL_STATE(7126)] = 239759,
  [SMALL_STATE(7127)] = 239773,
  [SMALL_STATE(7128)] = 239787,
  [SMALL_STATE(7129)] = 239801,
  [SMALL_STATE(7130)] = 239815,
  [SMALL_STATE(7131)] = 239829,
  [SMALL_STATE(7132)] = 239843,
  [SMALL_STATE(7133)] = 239857,
  [SMALL_STATE(7134)] = 239871,
  [SMALL_STATE(7135)] = 239885,
  [SMALL_STATE(7136)] = 239899,
  [SMALL_STATE(7137)] = 239913,
  [SMALL_STATE(7138)] = 239927,
  [SMALL_STATE(7139)] = 239941,
  [SMALL_STATE(7140)] = 239955,
  [SMALL_STATE(7141)] = 239969,
  [SMALL_STATE(7142)] = 239983,
  [SMALL_STATE(7143)] = 239997,
  [SMALL_STATE(7144)] = 240011,
  [SMALL_STATE(7145)] = 240025,
  [SMALL_STATE(7146)] = 240039,
  [SMALL_STATE(7147)] = 240053,
  [SMALL_STATE(7148)] = 240067,
  [SMALL_STATE(7149)] = 240081,
  [SMALL_STATE(7150)] = 240095,
  [SMALL_STATE(7151)] = 240109,
  [SMALL_STATE(7152)] = 240123,
  [SMALL_STATE(7153)] = 240137,
  [SMALL_STATE(7154)] = 240151,
  [SMALL_STATE(7155)] = 240165,
  [SMALL_STATE(7156)] = 240179,
  [SMALL_STATE(7157)] = 240193,
  [SMALL_STATE(7158)] = 240207,
  [SMALL_STATE(7159)] = 240221,
  [SMALL_STATE(7160)] = 240235,
  [SMALL_STATE(7161)] = 240249,
  [SMALL_STATE(7162)] = 240263,
  [SMALL_STATE(7163)] = 240277,
  [SMALL_STATE(7164)] = 240291,
  [SMALL_STATE(7165)] = 240305,
  [SMALL_STATE(7166)] = 240319,
  [SMALL_STATE(7167)] = 240333,
  [SMALL_STATE(7168)] = 240347,
  [SMALL_STATE(7169)] = 240361,
  [SMALL_STATE(7170)] = 240375,
  [SMALL_STATE(7171)] = 240389,
  [SMALL_STATE(7172)] = 240403,
  [SMALL_STATE(7173)] = 240417,
  [SMALL_STATE(7174)] = 240431,
  [SMALL_STATE(7175)] = 240445,
  [SMALL_STATE(7176)] = 240459,
  [SMALL_STATE(7177)] = 240473,
  [SMALL_STATE(7178)] = 240487,
  [SMALL_STATE(7179)] = 240501,
  [SMALL_STATE(7180)] = 240515,
  [SMALL_STATE(7181)] = 240529,
  [SMALL_STATE(7182)] = 240543,
  [SMALL_STATE(7183)] = 240557,
  [SMALL_STATE(7184)] = 240571,
  [SMALL_STATE(7185)] = 240585,
  [SMALL_STATE(7186)] = 240599,
  [SMALL_STATE(7187)] = 240613,
  [SMALL_STATE(7188)] = 240627,
  [SMALL_STATE(7189)] = 240641,
  [SMALL_STATE(7190)] = 240655,
  [SMALL_STATE(7191)] = 240669,
  [SMALL_STATE(7192)] = 240683,
  [SMALL_STATE(7193)] = 240697,
  [SMALL_STATE(7194)] = 240711,
  [SMALL_STATE(7195)] = 240725,
  [SMALL_STATE(7196)] = 240739,
  [SMALL_STATE(7197)] = 240753,
  [SMALL_STATE(7198)] = 240767,
  [SMALL_STATE(7199)] = 240781,
  [SMALL_STATE(7200)] = 240795,
  [SMALL_STATE(7201)] = 240809,
  [SMALL_STATE(7202)] = 240823,
  [SMALL_STATE(7203)] = 240837,
  [SMALL_STATE(7204)] = 240851,
  [SMALL_STATE(7205)] = 240865,
  [SMALL_STATE(7206)] = 240879,
  [SMALL_STATE(7207)] = 240893,
  [SMALL_STATE(7208)] = 240907,
  [SMALL_STATE(7209)] = 240921,
  [SMALL_STATE(7210)] = 240935,
  [SMALL_STATE(7211)] = 240949,
  [SMALL_STATE(7212)] = 240963,
  [SMALL_STATE(7213)] = 240977,
  [SMALL_STATE(7214)] = 240991,
  [SMALL_STATE(7215)] = 241005,
  [SMALL_STATE(7216)] = 241019,
  [SMALL_STATE(7217)] = 241033,
  [SMALL_STATE(7218)] = 241047,
  [SMALL_STATE(7219)] = 241061,
  [SMALL_STATE(7220)] = 241075,
  [SMALL_STATE(7221)] = 241089,
  [SMALL_STATE(7222)] = 241103,
  [SMALL_STATE(7223)] = 241117,
  [SMALL_STATE(7224)] = 241131,
  [SMALL_STATE(7225)] = 241145,
  [SMALL_STATE(7226)] = 241159,
  [SMALL_STATE(7227)] = 241173,
  [SMALL_STATE(7228)] = 241187,
  [SMALL_STATE(7229)] = 241201,
  [SMALL_STATE(7230)] = 241215,
  [SMALL_STATE(7231)] = 241229,
  [SMALL_STATE(7232)] = 241243,
  [SMALL_STATE(7233)] = 241257,
  [SMALL_STATE(7234)] = 241271,
  [SMALL_STATE(7235)] = 241285,
  [SMALL_STATE(7236)] = 241299,
  [SMALL_STATE(7237)] = 241313,
  [SMALL_STATE(7238)] = 241327,
  [SMALL_STATE(7239)] = 241341,
  [SMALL_STATE(7240)] = 241355,
  [SMALL_STATE(7241)] = 241369,
  [SMALL_STATE(7242)] = 241383,
  [SMALL_STATE(7243)] = 241397,
  [SMALL_STATE(7244)] = 241411,
  [SMALL_STATE(7245)] = 241425,
  [SMALL_STATE(7246)] = 241439,
  [SMALL_STATE(7247)] = 241453,
  [SMALL_STATE(7248)] = 241467,
  [SMALL_STATE(7249)] = 241481,
  [SMALL_STATE(7250)] = 241495,
  [SMALL_STATE(7251)] = 241509,
  [SMALL_STATE(7252)] = 241523,
  [SMALL_STATE(7253)] = 241537,
  [SMALL_STATE(7254)] = 241551,
  [SMALL_STATE(7255)] = 241565,
  [SMALL_STATE(7256)] = 241579,
  [SMALL_STATE(7257)] = 241593,
  [SMALL_STATE(7258)] = 241607,
  [SMALL_STATE(7259)] = 241621,
  [SMALL_STATE(7260)] = 241635,
  [SMALL_STATE(7261)] = 241649,
  [SMALL_STATE(7262)] = 241663,
  [SMALL_STATE(7263)] = 241677,
  [SMALL_STATE(7264)] = 241691,
  [SMALL_STATE(7265)] = 241705,
  [SMALL_STATE(7266)] = 241719,
  [SMALL_STATE(7267)] = 241733,
  [SMALL_STATE(7268)] = 241747,
  [SMALL_STATE(7269)] = 241761,
  [SMALL_STATE(7270)] = 241775,
  [SMALL_STATE(7271)] = 241789,
  [SMALL_STATE(7272)] = 241803,
  [SMALL_STATE(7273)] = 241817,
  [SMALL_STATE(7274)] = 241831,
  [SMALL_STATE(7275)] = 241845,
  [SMALL_STATE(7276)] = 241859,
  [SMALL_STATE(7277)] = 241873,
  [SMALL_STATE(7278)] = 241887,
  [SMALL_STATE(7279)] = 241901,
  [SMALL_STATE(7280)] = 241915,
  [SMALL_STATE(7281)] = 241929,
  [SMALL_STATE(7282)] = 241943,
  [SMALL_STATE(7283)] = 241957,
  [SMALL_STATE(7284)] = 241971,
  [SMALL_STATE(7285)] = 241985,
  [SMALL_STATE(7286)] = 241999,
  [SMALL_STATE(7287)] = 242013,
  [SMALL_STATE(7288)] = 242027,
  [SMALL_STATE(7289)] = 242041,
  [SMALL_STATE(7290)] = 242055,
  [SMALL_STATE(7291)] = 242069,
  [SMALL_STATE(7292)] = 242083,
  [SMALL_STATE(7293)] = 242097,
  [SMALL_STATE(7294)] = 242111,
  [SMALL_STATE(7295)] = 242125,
  [SMALL_STATE(7296)] = 242139,
  [SMALL_STATE(7297)] = 242153,
  [SMALL_STATE(7298)] = 242167,
  [SMALL_STATE(7299)] = 242181,
  [SMALL_STATE(7300)] = 242195,
  [SMALL_STATE(7301)] = 242209,
  [SMALL_STATE(7302)] = 242223,
  [SMALL_STATE(7303)] = 242237,
  [SMALL_STATE(7304)] = 242251,
  [SMALL_STATE(7305)] = 242265,
  [SMALL_STATE(7306)] = 242279,
  [SMALL_STATE(7307)] = 242293,
  [SMALL_STATE(7308)] = 242307,
  [SMALL_STATE(7309)] = 242321,
  [SMALL_STATE(7310)] = 242335,
  [SMALL_STATE(7311)] = 242349,
  [SMALL_STATE(7312)] = 242363,
  [SMALL_STATE(7313)] = 242377,
  [SMALL_STATE(7314)] = 242391,
  [SMALL_STATE(7315)] = 242405,
  [SMALL_STATE(7316)] = 242419,
  [SMALL_STATE(7317)] = 242433,
  [SMALL_STATE(7318)] = 242447,
  [SMALL_STATE(7319)] = 242461,
  [SMALL_STATE(7320)] = 242475,
  [SMALL_STATE(7321)] = 242489,
  [SMALL_STATE(7322)] = 242503,
  [SMALL_STATE(7323)] = 242517,
  [SMALL_STATE(7324)] = 242531,
  [SMALL_STATE(7325)] = 242545,
  [SMALL_STATE(7326)] = 242559,
  [SMALL_STATE(7327)] = 242573,
  [SMALL_STATE(7328)] = 242587,
  [SMALL_STATE(7329)] = 242601,
  [SMALL_STATE(7330)] = 242615,
  [SMALL_STATE(7331)] = 242629,
  [SMALL_STATE(7332)] = 242643,
  [SMALL_STATE(7333)] = 242657,
  [SMALL_STATE(7334)] = 242671,
  [SMALL_STATE(7335)] = 242685,
  [SMALL_STATE(7336)] = 242699,
  [SMALL_STATE(7337)] = 242713,
  [SMALL_STATE(7338)] = 242727,
  [SMALL_STATE(7339)] = 242741,
  [SMALL_STATE(7340)] = 242755,
  [SMALL_STATE(7341)] = 242769,
  [SMALL_STATE(7342)] = 242783,
  [SMALL_STATE(7343)] = 242797,
  [SMALL_STATE(7344)] = 242811,
  [SMALL_STATE(7345)] = 242825,
  [SMALL_STATE(7346)] = 242839,
  [SMALL_STATE(7347)] = 242853,
  [SMALL_STATE(7348)] = 242867,
  [SMALL_STATE(7349)] = 242871,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5782),
  [5] = {.entry = {.count = 1, .reusable = true}}, SHIFT_EXTRA(),
  [7] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2503),
  [9] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7279),
  [11] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2553),
  [13] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2563),
  [15] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2499),
  [17] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2536),
  [19] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3191),
  [21] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5463),
  [23] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5782),
  [25] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5822),
  [27] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5867),
  [29] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3260),
  [31] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6024),
  [33] = {.entry = {.count = 1, .reusable = false}}, SHIFT(846),
  [35] = {.entry = {.count = 1, .reusable = false}}, SHIFT(726),
  [37] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6078),
  [39] = {.entry = {.count = 1, .reusable = true}}, SHIFT(59),
  [41] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression_anonymous, 2, 0, 0),
  [43] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__sequence_expression_anonymous, 2, 0, 0),
  [45] = {.entry = {.count = 1, .reusable = false}}, SHIFT(290),
  [47] = {.entry = {.count = 1, .reusable = false}}, SHIFT(892),
  [49] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1407),
  [51] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3560),
  [53] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1633),
  [55] = {.entry = {.count = 1, .reusable = true}}, SHIFT(291),
  [57] = {.entry = {.count = 1, .reusable = false}}, SHIFT(145),
  [59] = {.entry = {.count = 1, .reusable = false}}, SHIFT(141),
  [61] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1264),
  [63] = {.entry = {.count = 1, .reusable = false}}, SHIFT(163),
  [65] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1138),
  [67] = {.entry = {.count = 1, .reusable = false}}, SHIFT(164),
  [69] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1425),
  [71] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1426),
  [73] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4973),
  [75] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5140),
  [77] = {.entry = {.count = 1, .reusable = false}}, SHIFT(104),
  [79] = {.entry = {.count = 1, .reusable = true}}, SHIFT(910),
  [81] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5806),
  [83] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5909),
  [85] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5925),
  [87] = {.entry = {.count = 1, .reusable = true}}, SHIFT(893),
  [89] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2639),
  [91] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4596),
  [93] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4123),
  [95] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1466),
  [97] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1416),
  [99] = {.entry = {.count = 1, .reusable = true}}, SHIFT(479),
  [101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(885),
  [103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6278),
  [105] = {.entry = {.count = 1, .reusable = false}}, SHIFT(937),
  [107] = {.entry = {.count = 1, .reusable = false}}, SHIFT(723),
  [109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(55),
  [111] = {.entry = {.count = 1, .reusable = false}}, SHIFT(274),
  [113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(995),
  [115] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1405),
  [117] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3554),
  [119] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1629),
  [121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(275),
  [123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(142),
  [125] = {.entry = {.count = 1, .reusable = false}}, SHIFT(121),
  [127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1225),
  [129] = {.entry = {.count = 1, .reusable = false}}, SHIFT(158),
  [131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1133),
  [133] = {.entry = {.count = 1, .reusable = false}}, SHIFT(114),
  [135] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1445),
  [137] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1449),
  [139] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4902),
  [141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5043),
  [143] = {.entry = {.count = 1, .reusable = false}}, SHIFT(101),
  [145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(987),
  [147] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5796),
  [149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5901),
  [151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(996),
  [153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4591),
  [155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4145),
  [157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1491),
  [159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(212),
  [161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(957),
  [163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6137),
  [165] = {.entry = {.count = 1, .reusable = false}}, SHIFT(727),
  [167] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1635),
  [169] = {.entry = {.count = 1, .reusable = false}}, SHIFT(147),
  [171] = {.entry = {.count = 1, .reusable = false}}, SHIFT(167),
  [173] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1141),
  [175] = {.entry = {.count = 1, .reusable = false}}, SHIFT(115),
  [177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(489),
  [179] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1095),
  [181] = {.entry = {.count = 1, .reusable = false}}, SHIFT(732),
  [183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(66),
  [185] = {.entry = {.count = 1, .reusable = false}}, SHIFT(252),
  [187] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1086),
  [189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1386),
  [191] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3531),
  [193] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1639),
  [195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(277),
  [197] = {.entry = {.count = 1, .reusable = false}}, SHIFT(120),
  [199] = {.entry = {.count = 1, .reusable = false}}, SHIFT(123),
  [201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1139),
  [203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(132),
  [205] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4954),
  [207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5155),
  [209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(95),
  [211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1052),
  [213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5708),
  [215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5850),
  [217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1069),
  [219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4555),
  [221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4135),
  [223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1501),
  [225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(262),
  [227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1099),
  [229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6241),
  [231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1646),
  [233] = {.entry = {.count = 1, .reusable = false}}, SHIFT(730),
  [235] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1376),
  [237] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4968),
  [239] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4046),
  [241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1245),
  [243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2323),
  [245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4528),
  [247] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4731),
  [249] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3788),
  [251] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3915),
  [253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4805),
  [255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4965),
  [257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5936),
  [259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5943),
  [261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5551),
  [263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2639),
  [265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1414),
  [267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6151),
  [269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3612),
  [271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3781),
  [273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3424),
  [275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7059),
  [277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5300),
  [279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3484),
  [281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3128),
  [283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1124),
  [285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5691),
  [287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2026),
  [289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5433),
  [291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(799),
  [293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3199),
  [295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2254),
  [297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1898),
  [299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4159),
  [301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3076),
  [303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1791),
  [305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2457),
  [307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(949),
  [309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2990),
  [311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3670),
  [313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1542),
  [315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(935),
  [317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3224),
  [319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2420),
  [321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2699),
  [323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1982),
  [325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4384),
  [327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7349),
  [329] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3140),
  [331] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__structure, 3, 0, 0),
  [333] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__structure, 3, 0, 0),
  [335] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__structure, 1, 0, 0),
  [337] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__structure, 1, 0, 0),
  [339] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__structure, 4, 0, 0),
  [341] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__structure, 4, 0, 0),
  [343] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4512),
  [345] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__structure, 2, 0, 0),
  [347] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__structure, 2, 0, 0),
  [349] = {.entry = {.count = 1, .reusable = false}}, SHIFT(835),
  [351] = {.entry = {.count = 1, .reusable = false}}, SHIFT(721),
  [353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(47),
  [355] = {.entry = {.count = 1, .reusable = false}}, SHIFT(324),
  [357] = {.entry = {.count = 1, .reusable = false}}, SHIFT(789),
  [359] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1399),
  [361] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3541),
  [363] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1628),
  [365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(325),
  [367] = {.entry = {.count = 1, .reusable = false}}, SHIFT(140),
  [369] = {.entry = {.count = 1, .reusable = false}}, SHIFT(135),
  [371] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1262),
  [373] = {.entry = {.count = 1, .reusable = false}}, SHIFT(150),
  [375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1137),
  [377] = {.entry = {.count = 1, .reusable = false}}, SHIFT(151),
  [379] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1434),
  [381] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1436),
  [383] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4931),
  [385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5078),
  [387] = {.entry = {.count = 1, .reusable = false}}, SHIFT(100),
  [389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(794),
  [391] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5753),
  [393] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5865),
  [395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(790),
  [397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4578),
  [399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4133),
  [401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1471),
  [403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(440),
  [405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(793),
  [407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6322),
  [409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(720),
  [411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1632),
  [413] = {.entry = {.count = 1, .reusable = false}}, SHIFT(134),
  [415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5753),
  [417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5865),
  [419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(887),
  [421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7289),
  [423] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3983),
  [425] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6635),
  [427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6890),
  [429] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2633),
  [431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1459),
  [433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6833),
  [435] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7332),
  [437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7332),
  [439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6460),
  [441] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6473),
  [443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6595),
  [445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6595),
  [447] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6627),
  [449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1395),
  [451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(198),
  [453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(798),
  [455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7277),
  [457] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3851),
  [459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7130),
  [461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1452),
  [463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7135),
  [465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1388),
  [467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(181),
  [469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4124),
  [471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7273),
  [473] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3962),
  [475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6512),
  [477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1454),
  [479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7187),
  [481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1389),
  [483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(210),
  [485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1873),
  [487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7281),
  [489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1455),
  [491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6769),
  [493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1391),
  [495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(193),
  [497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1857),
  [499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7285),
  [501] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3972),
  [503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6624),
  [505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1461),
  [507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6636),
  [509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1392),
  [511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(192),
  [513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1000),
  [515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7287),
  [517] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3980),
  [519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6672),
  [521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1458),
  [523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6676),
  [525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1393),
  [527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(202),
  [529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1525),
  [531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7288),
  [533] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3981),
  [535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6829),
  [537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1462),
  [539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6756),
  [541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1394),
  [543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(186),
  [545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2679),
  [547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7290),
  [549] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3984),
  [551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6944),
  [553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1453),
  [555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6892),
  [557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1396),
  [559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(207),
  [561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1942),
  [563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7291),
  [565] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3985),
  [567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7075),
  [569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1456),
  [571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6949),
  [573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1397),
  [575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(182),
  [577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2384),
  [579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7292),
  [581] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3986),
  [583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7103),
  [585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1460),
  [587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7023),
  [589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1398),
  [591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(169),
  [593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1067),
  [595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7310),
  [597] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3856),
  [599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7146),
  [601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1457),
  [603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7320),
  [605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1401),
  [607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(177),
  [609] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1517),
  [611] = {.entry = {.count = 1, .reusable = false}}, SHIFT(729),
  [613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(57),
  [615] = {.entry = {.count = 1, .reusable = false}}, SHIFT(281),
  [617] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1578),
  [619] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1406),
  [621] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3559),
  [623] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1637),
  [625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(282),
  [627] = {.entry = {.count = 1, .reusable = false}}, SHIFT(149),
  [629] = {.entry = {.count = 1, .reusable = false}}, SHIFT(139),
  [631] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1263),
  [633] = {.entry = {.count = 1, .reusable = false}}, SHIFT(161),
  [635] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1136),
  [637] = {.entry = {.count = 1, .reusable = false}}, SHIFT(162),
  [639] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1439),
  [641] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1443),
  [643] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5041),
  [645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5090),
  [647] = {.entry = {.count = 1, .reusable = false}}, SHIFT(102),
  [649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1543),
  [651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5804),
  [653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5907),
  [655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1581),
  [657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4595),
  [659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4122),
  [661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1463),
  [663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(474),
  [665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1577),
  [667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6268),
  [669] = {.entry = {.count = 1, .reusable = false}}, SHIFT(728),
  [671] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1636),
  [673] = {.entry = {.count = 1, .reusable = false}}, SHIFT(148),
  [675] = {.entry = {.count = 1, .reusable = false}}, SHIFT(159),
  [677] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1143),
  [679] = {.entry = {.count = 1, .reusable = false}}, SHIFT(160),
  [681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5806),
  [683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5909),
  [685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(470),
  [687] = {.entry = {.count = 1, .reusable = false}}, SHIFT(725),
  [689] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1631),
  [691] = {.entry = {.count = 1, .reusable = false}}, SHIFT(144),
  [693] = {.entry = {.count = 1, .reusable = false}}, SHIFT(724),
  [695] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1630),
  [697] = {.entry = {.count = 1, .reusable = false}}, SHIFT(143),
  [699] = {.entry = {.count = 1, .reusable = false}}, SHIFT(731),
  [701] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1641),
  [703] = {.entry = {.count = 1, .reusable = false}}, SHIFT(138),
  [705] = {.entry = {.count = 1, .reusable = false}}, SHIFT(154),
  [707] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1134),
  [709] = {.entry = {.count = 1, .reusable = false}}, SHIFT(155),
  [711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(455),
  [713] = {.entry = {.count = 1, .reusable = false}}, SHIFT(722),
  [715] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1634),
  [717] = {.entry = {.count = 1, .reusable = false}}, SHIFT(146),
  [719] = {.entry = {.count = 1, .reusable = false}}, SHIFT(165),
  [721] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1140),
  [723] = {.entry = {.count = 1, .reusable = false}}, SHIFT(166),
  [725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(484),
  [727] = {.entry = {.count = 1, .reusable = false}}, SHIFT(719),
  [729] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1642),
  [731] = {.entry = {.count = 1, .reusable = false}}, SHIFT(153),
  [733] = {.entry = {.count = 1, .reusable = false}}, SHIFT(156),
  [735] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1135),
  [737] = {.entry = {.count = 1, .reusable = false}}, SHIFT(157),
  [739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(460),
  [741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5925),
  [743] = {.entry = {.count = 1, .reusable = false}}, SHIFT(804),
  [745] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6124),
  [747] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression, 1, 0, 0),
  [749] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression, 1, 0, 0),
  [751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4566),
  [753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6414),
  [755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1495),
  [757] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(804),
  [760] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61),
  [762] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(6124),
  [765] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(47),
  [768] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61),
  [770] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(324),
  [773] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(789),
  [776] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1399),
  [779] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(3541),
  [782] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(325),
  [785] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(4931),
  [788] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5078),
  [791] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(100),
  [794] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(794),
  [797] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5753),
  [800] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5865),
  [803] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(790),
  [806] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(4578),
  [809] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(4133),
  [812] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1471),
  [815] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(793),
  [818] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(6322),
  [821] = {.entry = {.count = 1, .reusable = false}}, SHIFT(884),
  [823] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5986),
  [825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4561),
  [827] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7039),
  [829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1468),
  [831] = {.entry = {.count = 1, .reusable = false}}, SHIFT(979),
  [833] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6123),
  [835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4557),
  [837] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6573),
  [839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1493),
  [841] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_application_expression, 2, 0, 30),
  [843] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_application_expression, 2, 0, 30),
  [845] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(884),
  [848] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5986),
  [851] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(59),
  [854] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(290),
  [857] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(892),
  [860] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1407),
  [863] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(3560),
  [866] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(291),
  [869] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(4973),
  [872] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5140),
  [875] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(104),
  [878] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(910),
  [881] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5806),
  [884] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5909),
  [887] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(893),
  [890] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(4596),
  [893] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(4123),
  [896] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1466),
  [899] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(885),
  [902] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(6278),
  [905] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(979),
  [908] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(6123),
  [911] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(55),
  [914] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(274),
  [917] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(995),
  [920] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1405),
  [923] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(3554),
  [926] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(275),
  [929] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(4902),
  [932] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5043),
  [935] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(101),
  [938] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(987),
  [941] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5796),
  [944] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5901),
  [947] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(996),
  [950] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(4591),
  [953] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(4145),
  [956] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1491),
  [959] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(957),
  [962] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(6137),
  [965] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1125),
  [967] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6174),
  [969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4552),
  [971] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7010),
  [973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1475),
  [975] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1125),
  [978] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(6174),
  [981] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(66),
  [984] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(252),
  [987] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1086),
  [990] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1386),
  [993] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(3531),
  [996] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(277),
  [999] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(4954),
  [1002] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5155),
  [1005] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(95),
  [1008] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1052),
  [1011] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5708),
  [1014] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5850),
  [1017] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1069),
  [1020] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(4555),
  [1023] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(4135),
  [1026] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1501),
  [1029] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1099),
  [1032] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(6241),
  [1035] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3496),
  [1037] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6023),
  [1039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(195),
  [1041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2065),
  [1043] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6204),
  [1045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7241),
  [1047] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1226),
  [1049] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4088),
  [1051] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4719),
  [1053] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1224),
  [1055] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5413),
  [1057] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5503),
  [1059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6873),
  [1061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1227),
  [1063] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1249),
  [1065] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5560),
  [1067] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1719),
  [1069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4072),
  [1071] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6263),
  [1073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4606),
  [1075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4158),
  [1077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6980),
  [1079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1856),
  [1081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6129),
  [1083] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1067),
  [1085] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1873),
  [1087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3022),
  [1089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7284),
  [1091] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5298),
  [1093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7340),
  [1095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7342),
  [1097] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1857),
  [1099] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3570),
  [1101] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6186),
  [1103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(191),
  [1105] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1204),
  [1107] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4767),
  [1109] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1214),
  [1111] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5288),
  [1113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7174),
  [1115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1216),
  [1117] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1220),
  [1119] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1695),
  [1121] = {.entry = {.count = 1, .reusable = false}}, SHIFT(798),
  [1123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1000),
  [1125] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1525),
  [1127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2459),
  [1129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7286),
  [1131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5350),
  [1133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6504),
  [1135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6518),
  [1137] = {.entry = {.count = 1, .reusable = false}}, SHIFT(887),
  [1139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4937),
  [1141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7283),
  [1143] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5477),
  [1145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7162),
  [1147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7169),
  [1149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2679),
  [1151] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1942),
  [1153] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2384),
  [1155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2288),
  [1157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7280),
  [1159] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5368),
  [1161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6517),
  [1163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6750),
  [1165] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4124),
  [1167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2893),
  [1169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2277),
  [1171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2155),
  [1173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5708),
  [1175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5850),
  [1177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3413),
  [1179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5796),
  [1181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5901),
  [1183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2419),
  [1185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7021),
  [1187] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2680),
  [1189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2115),
  [1191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6764),
  [1193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1113),
  [1195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7006),
  [1197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(821),
  [1199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1957),
  [1201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1944),
  [1203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2152),
  [1205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1528),
  [1207] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4143),
  [1209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1531),
  [1211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1865),
  [1213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1877),
  [1215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(824),
  [1217] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1122),
  [1219] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1880),
  [1221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(895),
  [1223] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1875),
  [1225] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2886),
  [1227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(889),
  [1229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1002),
  [1231] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1004),
  [1233] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2269),
  [1235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2684),
  [1237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3312),
  [1239] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2360),
  [1241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1761),
  [1243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1750),
  [1245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1082),
  [1247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1899),
  [1249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1900),
  [1251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(990),
  [1253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(992),
  [1255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1091),
  [1257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1546),
  [1259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1547),
  [1261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(863),
  [1263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(912),
  [1265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2720),
  [1267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2721),
  [1269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1999),
  [1271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2003),
  [1273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2348),
  [1275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2349),
  [1277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(800),
  [1279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(823),
  [1281] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression_content, 2, 0, 0),
  [1283] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression_content, 3, 0, 0),
  [1285] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6136),
  [1287] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6138),
  [1289] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6132),
  [1291] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6297),
  [1293] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2038),
  [1295] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_binding_pattern, 1, 0, 0),
  [1297] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_binding_pattern, 1, 0, 0),
  [1299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6023),
  [1301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(94),
  [1303] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1228),
  [1305] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1795),
  [1307] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4789),
  [1309] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1209),
  [1311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5511),
  [1313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1211),
  [1315] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1242),
  [1317] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5823),
  [1319] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1706),
  [1321] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5742),
  [1323] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5859),
  [1325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1799),
  [1327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6922),
  [1329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4613),
  [1331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4153),
  [1333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1650),
  [1335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6257),
  [1337] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1594),
  [1339] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6215),
  [1341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4568),
  [1343] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6888),
  [1345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1502),
  [1347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(187),
  [1349] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1594),
  [1352] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(6215),
  [1355] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(57),
  [1358] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(281),
  [1361] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1578),
  [1364] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1406),
  [1367] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(3559),
  [1370] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(282),
  [1373] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5041),
  [1376] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5090),
  [1379] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(102),
  [1382] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1543),
  [1385] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5804),
  [1388] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5907),
  [1391] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1581),
  [1394] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(4595),
  [1397] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(4122),
  [1400] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1463),
  [1403] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1577),
  [1406] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(6268),
  [1409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3021),
  [1411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1418),
  [1413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(185),
  [1415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5029),
  [1417] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4043),
  [1419] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1229),
  [1421] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2927),
  [1423] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4736),
  [1425] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4533),
  [1427] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4701),
  [1429] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3845),
  [1431] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4718),
  [1433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5383),
  [1435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1230),
  [1437] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1231),
  [1439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5748),
  [1441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5596),
  [1443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5541),
  [1445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5891),
  [1447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5597),
  [1449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5820),
  [1451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5542),
  [1453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2928),
  [1455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6510),
  [1457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4569),
  [1459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4157),
  [1461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2674),
  [1463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6238),
  [1465] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1413),
  [1467] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2285),
  [1469] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 1, 0, 17),
  [1471] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 1, 0, 17),
  [1473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(322),
  [1475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5919),
  [1477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(110),
  [1479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1273),
  [1481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1303),
  [1483] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2504),
  [1485] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1240),
  [1487] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2260),
  [1489] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4723),
  [1491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2802),
  [1493] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5282),
  [1495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5336),
  [1497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1356),
  [1499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1250),
  [1501] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5591),
  [1503] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5569),
  [1505] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5767),
  [1507] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5876),
  [1509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2263),
  [1511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7118),
  [1513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4583),
  [1515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4136),
  [1517] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1417),
  [1519] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1422),
  [1521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(231),
  [1523] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_expression, 1, 0, 0),
  [1525] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_expression, 1, 0, 0),
  [1527] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_application, 2, 0, 92),
  [1529] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_application, 2, 0, 92),
  [1531] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_get_expression, 3, 0, 58),
  [1533] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_get_expression, 3, 0, 58),
  [1535] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_path, 3, 0, 79),
  [1537] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_path, 3, 0, 79),
  [1539] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_bigarray_get_expression, 5, 0, 136),
  [1541] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_bigarray_get_expression, 5, 0, 136),
  [1543] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_bigarray_get_expression, 6, 0, 173),
  [1545] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_bigarray_get_expression, 6, 0, 173),
  [1547] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_path, 1, 0, 24),
  [1549] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_path, 1, 0, 24),
  [1551] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_get_expression, 6, 0, 174),
  [1553] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string_get_expression, 6, 0, 174),
  [1555] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_get_expression, 6, 0, 173),
  [1557] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_get_expression, 6, 0, 173),
  [1559] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_get_expression, 5, 0, 136),
  [1561] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_get_expression, 5, 0, 136),
  [1563] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_get_expression, 5, 0, 137),
  [1565] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string_get_expression, 5, 0, 137),
  [1567] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 6, 0, 170),
  [1569] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 6, 0, 170),
  [1571] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_path, 1, 0, 26),
  [1573] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_path, 1, 0, 26),
  [1575] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_new_expression, 2, 0, 0),
  [1577] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_new_expression, 2, 0, 0),
  [1579] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_invocation, 3, 0, 59),
  [1581] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_invocation, 3, 0, 59),
  [1583] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_expression, 3, 0, 60),
  [1585] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_expression, 3, 0, 60),
  [1587] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_local_open_expression, 3, 0, 56),
  [1589] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_local_open_expression, 3, 0, 56),
  [1591] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_path, 3, 0, 0),
  [1593] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_path, 3, 0, 0),
  [1595] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_copy_expression, 2, 0, 0),
  [1597] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_copy_expression, 2, 0, 0),
  [1599] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_expression, 4, 0, 47),
  [1601] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_expression, 4, 0, 47),
  [1603] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_typed_expression, 4, 0, 76),
  [1605] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_typed_expression, 4, 0, 76),
  [1607] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_expression, 4, 0, 0),
  [1609] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_expression, 4, 0, 0),
  [1611] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_prefix_expression, 2, 0, 28),
  [1613] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_prefix_expression, 2, 0, 28),
  [1615] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 1, 0, 29),
  [1617] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 1, 0, 29),
  [1619] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 4, 0, 0),
  [1621] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 4, 0, 0),
  [1623] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tag, 2, 0, 0),
  [1625] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tag, 2, 0, 0),
  [1627] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_path, 3, 0, 96),
  [1629] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_path, 3, 0, 96),
  [1631] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_copy_expression, 4, 0, 0),
  [1633] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_copy_expression, 4, 0, 0),
  [1635] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_expression, 4, 0, 56),
  [1637] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_expression, 4, 0, 56),
  [1639] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_expression, 5, 0, 109),
  [1641] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_expression, 5, 0, 109),
  [1643] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_expression, 5, 0, 83),
  [1645] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_expression, 5, 0, 83),
  [1647] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_expression, 5, 0, 114),
  [1649] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_expression, 5, 0, 114),
  [1651] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_expression, 5, 0, 0),
  [1653] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_expression, 5, 0, 0),
  [1655] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 5, 0, 127),
  [1657] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 5, 0, 127),
  [1659] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_copy_expression, 5, 0, 0),
  [1661] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_copy_expression, 5, 0, 0),
  [1663] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_argument, 4, 0, 74),
  [1665] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_argument, 4, 0, 74),
  [1667] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__delimited_expression, 1, 0, 0),
  [1669] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__delimited_expression, 1, 0, 0),
  [1671] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_expression, 3, 0, 27),
  [1673] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_expression, 3, 0, 27),
  [1675] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_expression, 3, 0, 0),
  [1677] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_expression, 3, 0, 0),
  [1679] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_expression, 3, 0, 0),
  [1681] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_expression, 3, 0, 0),
  [1683] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_boolean, 1, 0, 0),
  [1685] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_boolean, 1, 0, 0),
  [1687] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_number, 1, 0, 0),
  [1689] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_number, 1, 0, 0),
  [1691] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_path, 1, 0, 0),
  [1693] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_path, 1, 0, 0),
  [1695] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 3, 0, 0),
  [1697] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 3, 0, 0),
  [1699] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constructor_name, 1, 0, 6),
  [1701] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constructor_name, 1, 0, 6),
  [1703] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_path, 1, 0, 2),
  [1705] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_expression, 1, 0, 0),
  [1707] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_expression, 1, 0, 0),
  [1709] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constant, 1, 0, 0),
  [1711] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constant, 1, 0, 0),
  [1713] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_path, 1, 0, 0),
  [1715] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_path, 1, 0, 0),
  [1717] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_new_expression, 3, 0, 0),
  [1719] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_new_expression, 3, 0, 0),
  [1721] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unit, 2, 0, 0),
  [1723] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unit, 2, 0, 0),
  [1725] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_extension, 3, 0, 0),
  [1727] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extension, 3, 0, 0),
  [1729] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_expression, 2, 0, 0),
  [1731] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_expression, 2, 0, 0),
  [1733] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 2, 0, 0),
  [1735] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 2, 0, 0),
  [1737] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 2, 0, 0),
  [1739] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 2, 0, 0),
  [1741] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_expression, 7, 0, 114),
  [1743] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_expression, 7, 0, 114),
  [1745] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__value_name, 1, 0, 4),
  [1747] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__value_name, 1, 0, 4),
  [1749] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 7, 0, 160),
  [1751] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 7, 0, 160),
  [1753] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 7, 0, 170),
  [1755] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 7, 0, 170),
  [1757] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_operator, 9, 0, 0),
  [1759] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_operator, 9, 0, 0),
  [1761] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 7, 0, 195),
  [1763] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 7, 0, 195),
  [1765] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_expression, 6, 0, 146),
  [1767] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_expression, 6, 0, 146),
  [1769] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_extension, 4, 0, 0),
  [1771] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extension, 4, 0, 0),
  [1773] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_expression, 6, 0, 114),
  [1775] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_expression, 6, 0, 114),
  [1777] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 6, 0, 127),
  [1779] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 6, 0, 127),
  [1781] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 6, 0, 160),
  [1783] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 6, 0, 160),
  [1785] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_operator, 6, 0, 0),
  [1787] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_operator, 6, 0, 0),
  [1789] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_extension, 4, 0, 0),
  [1791] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_extension, 4, 0, 0),
  [1793] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_expression, 8, 0, 195),
  [1795] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_expression, 8, 0, 195),
  [1797] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_operator, 7, 0, 0),
  [1799] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_operator, 7, 0, 0),
  [1801] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_extension, 5, 0, 0),
  [1803] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_extension, 5, 0, 0),
  [1805] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_operator, 8, 0, 0),
  [1807] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_operator, 8, 0, 0),
  [1809] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constructor_name, 3, 0, 38),
  [1811] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constructor_name, 3, 0, 38),
  [1813] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_operator, 3, 0, 0),
  [1815] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_operator, 3, 0, 0),
  [1817] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_string, 3, 0, 0),
  [1819] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_string, 3, 0, 0),
  [1821] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_expression, 2, 0, 0),
  [1823] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_expression, 2, 0, 0),
  [1825] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unit, 3, 0, 0),
  [1827] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unit, 3, 0, 0),
  [1829] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_character, 3, 0, 0),
  [1831] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_character, 3, 0, 0),
  [1833] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 3, 0, 0),
  [1835] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 3, 0, 0),
  [1837] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_copy_expression, 3, 0, 0),
  [1839] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_copy_expression, 3, 0, 0),
  [1841] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_path, 3, 0, 15),
  [1843] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_path, 3, 0, 0),
  [1845] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_path, 3, 0, 0),
  [1847] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__value_name, 1, 0, 0),
  [1849] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__value_name, 1, 0, 0),
  [1851] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_expression, 3, 0, 0),
  [1853] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_expression, 3, 0, 0),
  [1855] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2401),
  [1857] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6130),
  [1859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(65),
  [1861] = {.entry = {.count = 1, .reusable = false}}, SHIFT(312),
  [1863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2327),
  [1865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1410),
  [1867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3563),
  [1869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(313),
  [1871] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5040),
  [1873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5255),
  [1875] = {.entry = {.count = 1, .reusable = false}}, SHIFT(109),
  [1877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2321),
  [1879] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5810),
  [1881] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5912),
  [1883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2340),
  [1885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4604),
  [1887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4127),
  [1889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1481),
  [1891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2365),
  [1893] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5979),
  [1895] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(2401),
  [1898] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(6130),
  [1901] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(65),
  [1904] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(312),
  [1907] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(2327),
  [1910] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1410),
  [1913] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(3563),
  [1916] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(313),
  [1919] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5040),
  [1922] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5255),
  [1925] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(109),
  [1928] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(2321),
  [1931] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5810),
  [1934] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5912),
  [1937] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(2340),
  [1940] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(4604),
  [1943] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(4127),
  [1946] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(1481),
  [1949] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(2365),
  [1952] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_application_repeat1, 2, 0, 61), SHIFT_REPEAT(5979),
  [1955] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__instance_variable_name, 1, 0, 5),
  [1957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(97),
  [1959] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1219),
  [1961] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1713),
  [1963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5541),
  [1965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5891),
  [1967] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_argument, 2, 0, 18),
  [1969] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1505),
  [1971] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_argument, 2, 0, 18),
  [1973] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2472),
  [1975] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 122),
  [1977] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 122),
  [1979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3343),
  [1981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5589),
  [1983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(103),
  [1985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2628),
  [1987] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1234),
  [1989] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2483),
  [1991] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4738),
  [1993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5324),
  [1995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1235),
  [1997] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5787),
  [1999] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5607),
  [2001] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5793),
  [2003] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5899),
  [2005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2480),
  [2007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6762),
  [2009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4590),
  [2011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4165),
  [2013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2397),
  [2015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6246),
  [2017] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_argument, 5, 0, 107),
  [2019] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_argument, 5, 0, 107),
  [2021] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 2, 0, 49),
  [2023] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 2, 0, 49),
  [2025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3337),
  [2027] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 2, 0, 26),
  [2029] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 2, 0, 26),
  [2031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3341),
  [2033] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 156),
  [2035] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 156),
  [2037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3342),
  [2039] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 186),
  [2041] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 186),
  [2043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3345),
  [2045] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 1, 0, 26),
  [2047] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 1, 0, 26),
  [2049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3383),
  [2051] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 186),
  [2053] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 186),
  [2055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3366),
  [2057] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 3, 0, 49),
  [2059] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 3, 0, 49),
  [2061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3336),
  [2063] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 122),
  [2065] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 122),
  [2067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3429),
  [2069] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 156),
  [2071] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 156),
  [2073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3432),
  [2075] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_pattern, 1, 0, 0),
  [2077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6186),
  [2079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(99),
  [2081] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_pattern, 1, 0, 0),
  [2083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5503),
  [2085] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6552),
  [2087] = {.entry = {.count = 1, .reusable = false}}, SHIFT(680),
  [2089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3339),
  [2091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5827),
  [2093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(111),
  [2095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2610),
  [2097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2557),
  [2099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6253),
  [2101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3360),
  [2103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3390),
  [2105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3418),
  [2107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3373),
  [2109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3400),
  [2111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3317),
  [2113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3322),
  [2115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3323),
  [2117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3378),
  [2119] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1477),
  [2121] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2472),
  [2124] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0),
  [2126] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0),
  [2128] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5589),
  [2131] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(103),
  [2134] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(1234),
  [2137] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2483),
  [2140] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(4738),
  [2143] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5324),
  [2146] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(1235),
  [2149] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5787),
  [2152] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5793),
  [2155] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5899),
  [2158] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2480),
  [2161] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6762),
  [2164] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(4590),
  [2167] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(4165),
  [2170] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2397),
  [2173] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6246),
  [2176] = {.entry = {.count = 1, .reusable = false}}, SHIFT(624),
  [2178] = {.entry = {.count = 1, .reusable = false}}, SHIFT(539),
  [2180] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4804),
  [2182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(105),
  [2184] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1221),
  [2186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5378),
  [2188] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1236),
  [2190] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1681),
  [2192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5592),
  [2194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5816),
  [2196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6552),
  [2198] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1255),
  [2200] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1256),
  [2202] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1769),
  [2204] = {.entry = {.count = 1, .reusable = false}}, SHIFT(583),
  [2206] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1470),
  [2208] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4962),
  [2210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(106),
  [2212] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1265),
  [2214] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4786),
  [2216] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4722),
  [2218] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1246),
  [2220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5291),
  [2222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1205),
  [2224] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1248),
  [2226] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5689),
  [2228] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1687),
  [2230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5777),
  [2232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5885),
  [2234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4791),
  [2236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7120),
  [2238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4558),
  [2240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4151),
  [2242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3193),
  [2244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6179),
  [2246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4827),
  [2248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(113),
  [2250] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1257),
  [2252] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4721),
  [2254] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1222),
  [2256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1259),
  [2258] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1261),
  [2260] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1691),
  [2262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(206),
  [2264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(209),
  [2266] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2285),
  [2269] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5827),
  [2272] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(111),
  [2275] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(1240),
  [2278] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2260),
  [2281] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(4723),
  [2284] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5336),
  [2287] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(1250),
  [2290] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5591),
  [2293] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5767),
  [2296] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5876),
  [2299] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2263),
  [2302] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(7118),
  [2305] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(4583),
  [2308] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(4136),
  [2311] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2557),
  [2314] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6253),
  [2317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1296),
  [2319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1292),
  [2321] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1277),
  [2323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1301),
  [2325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1280),
  [2327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1272),
  [2329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1286),
  [2331] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1483),
  [2333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1283),
  [2335] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5919),
  [2338] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(110),
  [2341] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(7118),
  [2344] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(1856),
  [2347] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(6129),
  [2350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(483),
  [2352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2542),
  [2354] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 4, 0, 132),
  [2356] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 132),
  [2358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(473),
  [2360] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 4, 0, 93),
  [2362] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 93),
  [2364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(317),
  [2366] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 2, 0, 53),
  [2368] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 2, 0, 53),
  [2370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(335),
  [2372] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 3, 0, 93),
  [2374] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 3, 0, 93),
  [2376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(353),
  [2378] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 3, 0, 53),
  [2380] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 3, 0, 53),
  [2382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(336),
  [2384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1282),
  [2386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1304),
  [2388] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5330),
  [2390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1312),
  [2392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5747),
  [2394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5767),
  [2396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5876),
  [2398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(226),
  [2400] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 5, 0, 132),
  [2402] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 132),
  [2404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4336),
  [2406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4838),
  [2408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5742),
  [2410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5859),
  [2412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2036),
  [2414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4340),
  [2416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4572),
  [2418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4570),
  [2420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1952),
  [2422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2966),
  [2424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2967),
  [2426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2454),
  [2428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2456),
  [2430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2230),
  [2432] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2503),
  [2434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4225),
  [2436] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5034),
  [2438] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4047),
  [2440] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_payload, 1, 0, 0),
  [2442] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4532),
  [2444] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4813),
  [2446] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3503),
  [2448] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4727),
  [2450] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4904),
  [2452] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5822),
  [2454] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5867),
  [2456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2234),
  [2458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5022),
  [2460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4901),
  [2462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5014),
  [2464] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_binding_pattern_content, 3, 0, 0),
  [2466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6120),
  [2468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4431),
  [2470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6140),
  [2472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4611),
  [2474] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_pattern_content, 3, 0, 0),
  [2476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6193),
  [2478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2049),
  [2480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_binding_pattern_content, 2, 0, 0),
  [2482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6141),
  [2484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6030),
  [2486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4917),
  [2488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6160),
  [2490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3005),
  [2492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6334),
  [2494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4883),
  [2496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6111),
  [2498] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_pattern_content, 2, 0, 0),
  [2500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1925),
  [2502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2965),
  [2504] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1412),
  [2506] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4615),
  [2508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(48),
  [2510] = {.entry = {.count = 1, .reusable = false}}, SHIFT(224),
  [2512] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1402),
  [2514] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3530),
  [2516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(225),
  [2518] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4935),
  [2520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5084),
  [2522] = {.entry = {.count = 1, .reusable = false}}, SHIFT(112),
  [2524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1677),
  [2526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1506),
  [2528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1685),
  [2530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6069),
  [2532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4549),
  [2534] = {.entry = {.count = 1, .reusable = false}}, SHIFT(667),
  [2536] = {.entry = {.count = 1, .reusable = false}}, SHIFT(595),
  [2538] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1420),
  [2540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3386),
  [2542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2602),
  [2544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5717),
  [2546] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1928),
  [2548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(63),
  [2550] = {.entry = {.count = 1, .reusable = false}}, SHIFT(306),
  [2552] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1949),
  [2554] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1409),
  [2556] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3566),
  [2558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(307),
  [2560] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5019),
  [2562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5194),
  [2564] = {.entry = {.count = 1, .reusable = false}}, SHIFT(108),
  [2566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1937),
  [2568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5814),
  [2570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5914),
  [2572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5584),
  [2574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1950),
  [2576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4539),
  [2578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4126),
  [2580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1467),
  [2582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1976),
  [2584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5963),
  [2586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3403),
  [2588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3404),
  [2590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3408),
  [2592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3363),
  [2594] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1811),
  [2596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(53),
  [2598] = {.entry = {.count = 1, .reusable = false}}, SHIFT(260),
  [2600] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1827),
  [2602] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1404),
  [2604] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3545),
  [2606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(261),
  [2608] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4854),
  [2610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5080),
  [2612] = {.entry = {.count = 1, .reusable = false}}, SHIFT(98),
  [2614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1896),
  [2616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5790),
  [2618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5897),
  [2620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1831),
  [2622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4577),
  [2624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4162),
  [2626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1479),
  [2628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1832),
  [2630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6324),
  [2632] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2678),
  [2634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(61),
  [2636] = {.entry = {.count = 1, .reusable = false}}, SHIFT(298),
  [2638] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2681),
  [2640] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1408),
  [2642] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3565),
  [2644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(299),
  [2646] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4996),
  [2648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5164),
  [2650] = {.entry = {.count = 1, .reusable = false}}, SHIFT(107),
  [2652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2713),
  [2654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5812),
  [2656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5913),
  [2658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2682),
  [2660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4597),
  [2662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4125),
  [2664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1478),
  [2666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2694),
  [2668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6294),
  [2670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3370),
  [2672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3372),
  [2674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3398),
  [2676] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1643),
  [2678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(51),
  [2680] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1866),
  [2682] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1403),
  [2684] = {.entry = {.count = 1, .reusable = false}}, SHIFT(96),
  [2686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5774),
  [2688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5882),
  [2690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1867),
  [2692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4541),
  [2694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4144),
  [2696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1487),
  [2698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6045),
  [2700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3399),
  [2702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3436),
  [2704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5810),
  [2706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5912),
  [2708] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression_anonymous, 3, 0, 0),
  [2710] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__sequence_expression_anonymous, 3, 0, 0),
  [2712] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression, 1, 0, 8),
  [2714] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__sequence_expression, 1, 0, 8),
  [2716] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression_anonymous, 4, 0, 0),
  [2718] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__sequence_expression_anonymous, 4, 0, 0),
  [2720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(236),
  [2722] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_attribute_id_repeat1, 2, 0, 0),
  [2724] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attribute_id_repeat1, 2, 0, 0),
  [2726] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_id_repeat1, 2, 0, 0), SHIFT_REPEAT(5962),
  [2729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(304),
  [2731] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_id, 2, 0, 0),
  [2733] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_id, 2, 0, 0),
  [2735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5962),
  [2737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(368),
  [2739] = {.entry = {.count = 1, .reusable = false}}, SHIFT(638),
  [2741] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_id, 1, 0, 0),
  [2743] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_id, 1, 0, 0),
  [2745] = {.entry = {.count = 1, .reusable = false}}, SHIFT(710),
  [2747] = {.entry = {.count = 1, .reusable = false}}, SHIFT(570),
  [2749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3428),
  [2751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(280),
  [2753] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7118),
  [2755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(264),
  [2757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(292),
  [2759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(296),
  [2761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(246),
  [2763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(309),
  [2765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(311),
  [2767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(355),
  [2769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(214),
  [2771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(357),
  [2773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(216),
  [2775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(453),
  [2777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(458),
  [2779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(254),
  [2781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(402),
  [2783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(256),
  [2785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(469),
  [2787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(478),
  [2789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(436),
  [2791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(318),
  [2793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(244),
  [2795] = {.entry = {.count = 1, .reusable = false}}, SHIFT(553),
  [2797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(248),
  [2799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(446),
  [2801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(448),
  [2803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(242),
  [2805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(276),
  [2807] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__structure_repeat1, 2, 0, 0),
  [2809] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat1, 2, 0, 0), SHIFT_REPEAT(1646),
  [2812] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__structure_repeat1, 2, 0, 0),
  [2814] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1488),
  [2816] = {.entry = {.count = 1, .reusable = false}}, SHIFT(611),
  [2818] = {.entry = {.count = 1, .reusable = false}}, SHIFT(652),
  [2820] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__structure_repeat1, 1, 0, 0),
  [2822] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__structure_repeat1, 1, 0, 0),
  [2824] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_set_expression, 3, 0, 57),
  [2826] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_set_expression, 3, 0, 57),
  [2828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(670),
  [2830] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2642),
  [2832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(671),
  [2834] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2645),
  [2836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2643),
  [2838] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2644),
  [2840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2644),
  [2842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(672),
  [2844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(673),
  [2846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2646),
  [2848] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2646),
  [2850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(674),
  [2852] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression, 1, 0, 0),
  [2854] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__sequence_expression, 1, 0, 0),
  [2856] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3),
  [2858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3320),
  [2860] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_infix_expression, 3, 0, 60),
  [2862] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_infix_expression, 3, 0, 60),
  [2864] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_then_clause, 2, 0, 27),
  [2866] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_then_clause, 2, 0, 27),
  [2868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3355),
  [2870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3385),
  [2872] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sign_expression, 2, 0, 28),
  [2874] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sign_expression, 2, 0, 28),
  [2876] = {.entry = {.count = 1, .reusable = false}}, SHIFT(697),
  [2878] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tuple_expression, 3, 0, 0),
  [2880] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__tuple_expression, 3, 0, 0),
  [2882] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cons_expression, 3, 0, 62),
  [2884] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cons_expression, 3, 0, 62),
  [2886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3434),
  [2888] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_clause, 2, 0, 27),
  [2890] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_clause, 2, 0, 27),
  [2892] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1207),
  [2894] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1251),
  [2896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(615),
  [2898] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4),
  [2900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(616),
  [2902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(617),
  [2904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(618),
  [2906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(619),
  [2908] = {.entry = {.count = 1, .reusable = false}}, SHIFT(75),
  [2910] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assert_expression, 2, 0, 27),
  [2912] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assert_expression, 2, 0, 27),
  [2914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4586),
  [2916] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6402),
  [2918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1464),
  [2920] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lazy_expression, 2, 0, 27),
  [2922] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lazy_expression, 2, 0, 27),
  [2924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4601),
  [2926] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7202),
  [2928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1484),
  [2930] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__signed_constant, 1, 0, 0),
  [2932] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__signed_constant, 1, 0, 0),
  [2934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4178),
  [2936] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element_binding_pattern, 2, 0, 18),
  [2938] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element_binding_pattern, 2, 0, 18),
  [2940] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2198),
  [2942] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_signed_number, 2, 0, 0),
  [2944] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signed_number, 2, 0, 0),
  [2946] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assert_expression, 3, 0, 56),
  [2948] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assert_expression, 3, 0, 56),
  [2950] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lazy_expression, 3, 0, 56),
  [2952] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lazy_expression, 3, 0, 56),
  [2954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(529),
  [2956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(530),
  [2958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(531),
  [2960] = {.entry = {.count = 1, .reusable = false}}, SHIFT(532),
  [2962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2),
  [2964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(533),
  [2966] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element, 4, 0, 74),
  [2968] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element, 4, 0, 74),
  [2970] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lazy_binding_pattern, 2, 0, 37),
  [2972] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lazy_binding_pattern, 2, 0, 37),
  [2974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(573),
  [2976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(574),
  [2978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(575),
  [2980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(576),
  [2982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(577),
  [2984] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__tuple_binding_pattern, 3, 0, 0),
  [2986] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tuple_binding_pattern, 3, 0, 0),
  [2988] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exception_binding_pattern, 3, 0, 69),
  [2990] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exception_binding_pattern, 3, 0, 69),
  [2992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1293),
  [2994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1339),
  [2996] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5536),
  [2998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1340),
  [3000] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__or_binding_pattern_anonymous, 3, 0, 0),
  [3002] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__or_binding_pattern_anonymous, 3, 0, 0),
  [3004] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cons_binding_pattern, 3, 0, 62),
  [3006] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cons_binding_pattern, 3, 0, 62),
  [3008] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_alias_binding_pattern, 3, 0, 73),
  [3010] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_alias_binding_pattern, 3, 0, 73),
  [3012] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5),
  [3014] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_pattern, 5, 0, 140),
  [3016] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_pattern, 5, 0, 140),
  [3018] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_pattern, 5, 0, 141),
  [3020] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_pattern, 5, 0, 141),
  [3022] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_binding_pattern, 2, 0, 0),
  [3024] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_binding_pattern, 2, 0, 0),
  [3026] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_alias_binding_pattern_no_exn, 3, 0, 73),
  [3028] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_alias_binding_pattern_no_exn, 3, 0, 73),
  [3030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4554),
  [3032] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6859),
  [3034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1469),
  [3036] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_pattern, 4, 0, 15),
  [3038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_pattern, 4, 0, 15),
  [3040] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_typed_binding_pattern, 4, 0, 102),
  [3042] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_typed_binding_pattern, 4, 0, 102),
  [3044] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_effect_binding_pattern, 4, 0, 108),
  [3046] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_effect_binding_pattern, 4, 0, 108),
  [3048] = {.entry = {.count = 1, .reusable = false}}, SHIFT(76),
  [3050] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_constructor_path, 1, 0, 1),
  [3052] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_constructor_path, 1, 0, 1),
  [3054] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_variant_pattern, 2, 0, 0),
  [3056] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_variant_pattern, 2, 0, 0),
  [3058] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_binding_pattern, 5, 0, 0),
  [3060] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_binding_pattern, 5, 0, 0),
  [3062] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_binding_pattern, 4, 0, 0),
  [3064] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_binding_pattern, 4, 0, 0),
  [3066] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cons_binding_pattern_no_exn, 3, 0, 62),
  [3068] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cons_binding_pattern_no_exn, 3, 0, 62),
  [3070] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element_binding_pattern, 5, 0, 139),
  [3072] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element_binding_pattern, 5, 0, 139),
  [3074] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__tuple_binding_pattern_no_exn, 3, 0, 0),
  [3076] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tuple_binding_pattern_no_exn, 3, 0, 0),
  [3078] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_constructor_path, 3, 0, 14),
  [3080] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_constructor_path, 3, 0, 14),
  [3082] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exception_binding_pattern, 2, 0, 37),
  [3084] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exception_binding_pattern, 2, 0, 37),
  [3086] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_binding_pattern, 2, 0, 0),
  [3088] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_binding_pattern, 2, 0, 0),
  [3090] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_range_pattern, 3, 0, 62),
  [3092] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_range_pattern, 3, 0, 62),
  [3094] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__or_binding_pattern_no_exn_anonymous, 3, 0, 0),
  [3096] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__or_binding_pattern_no_exn_anonymous, 3, 0, 0),
  [3098] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_local_open_binding_pattern, 3, 0, 69),
  [3100] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_local_open_binding_pattern, 3, 0, 69),
  [3102] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_binding_pattern, 3, 0, 69),
  [3104] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_binding_pattern, 3, 0, 69),
  [3106] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_pattern, 6, 0, 177),
  [3108] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_pattern, 6, 0, 177),
  [3110] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_binding_pattern, 6, 0, 0),
  [3112] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_binding_pattern, 6, 0, 0),
  [3114] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_binding_pattern, 3, 0, 0),
  [3116] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_binding_pattern, 3, 0, 0),
  [3118] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__binding_pattern, 1, 0, 0),
  [3120] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__binding_pattern, 1, 0, 0),
  [3122] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__binding_pattern, 1, 0, 20),
  [3124] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__binding_pattern, 1, 0, 20),
  [3126] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__binding_pattern, 1, 0, 21),
  [3128] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__binding_pattern, 1, 0, 21),
  [3130] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_binding_pattern, 2, 0, 37),
  [3132] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_binding_pattern, 2, 0, 37),
  [3134] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tag_binding_pattern, 2, 0, 37),
  [3136] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tag_binding_pattern, 2, 0, 37),
  [3138] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_binding_pattern, 3, 0, 0),
  [3140] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_binding_pattern, 3, 0, 0),
  [3142] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_binding_pattern, 3, 0, 0),
  [3144] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_binding_pattern, 3, 0, 0),
  [3146] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_binding_pattern, 3, 0, 0),
  [3148] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_binding_pattern, 3, 0, 0),
  [3150] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lazy_binding_pattern, 3, 0, 69),
  [3152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lazy_binding_pattern, 3, 0, 69),
  [3154] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element_binding_pattern, 4, 0, 101),
  [3156] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element_binding_pattern, 4, 0, 101),
  [3158] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__effect_binding_pattern, 1, 0, 0),
  [3160] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__effect_binding_pattern, 1, 0, 0),
  [3162] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__binding_pattern_no_exn, 1, 0, 0),
  [3164] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__binding_pattern_no_exn, 1, 0, 0),
  [3166] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__delimited_binding_pattern, 1, 0, 0),
  [3168] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__delimited_binding_pattern, 1, 0, 0),
  [3170] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_binding_pattern, 7, 0, 0),
  [3172] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_binding_pattern, 7, 0, 0),
  [3174] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__parenthesized_abstract_type, 3, 0, 0),
  [3176] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__parenthesized_abstract_type, 3, 0, 0),
  [3178] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__match_cases, 1, 0, 0),
  [3180] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__match_cases, 1, 0, 0),
  [3182] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1279),
  [3184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2441),
  [3186] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5525),
  [3188] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2307),
  [3190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4633),
  [3192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4303),
  [3194] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2493),
  [3196] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4626),
  [3198] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5632),
  [3200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4508),
  [3202] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_class_expression, 1, 0, 0),
  [3204] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_class_expression, 1, 0, 0),
  [3206] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__match_cases, 3, 0, 0),
  [3208] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__match_cases, 3, 0, 0),
  [3210] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0),
  [3212] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0),
  [3214] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1279),
  [3217] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1500),
  [3219] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element, 2, 0, 18),
  [3221] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element, 2, 0, 18),
  [3223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2484),
  [3225] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5478),
  [3227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2312),
  [3229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3667),
  [3231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3368),
  [3233] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2496),
  [3235] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3477),
  [3237] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5877),
  [3239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5808),
  [3241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5911),
  [3243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3288),
  [3245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2315),
  [3247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4650),
  [3249] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_typed_class_expression, 4, 0, 163),
  [3251] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_typed_class_expression, 4, 0, 163),
  [3253] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_instantiated_class, 5, 0, 0),
  [3255] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instantiated_class, 5, 0, 0),
  [3257] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__match_cases, 2, 0, 0),
  [3259] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__match_cases, 2, 0, 0),
  [3261] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2309),
  [3263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3662),
  [3265] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_class_expression, 3, 0, 0),
  [3267] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_class_expression, 3, 0, 0),
  [3269] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_expression, 3, 0, 55),
  [3271] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_expression, 3, 0, 55),
  [3273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(684),
  [3275] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_expression, 4, 0, 95),
  [3277] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_expression, 4, 0, 95),
  [3279] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3017),
  [3281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2485),
  [3283] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2313),
  [3285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3474),
  [3287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3172),
  [3289] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2491),
  [3291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2497),
  [3293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2546),
  [3295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3173),
  [3297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5343),
  [3299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5799),
  [3301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5903),
  [3303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6224),
  [3305] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2303),
  [3307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3460),
  [3309] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1492),
  [3311] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_instantiated_class, 4, 0, 0),
  [3313] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instantiated_class, 4, 0, 0),
  [3315] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_clause, 2, 0, 0),
  [3317] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_clause, 2, 0, 0),
  [3319] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_case, 4, 0, 105),
  [3321] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_case, 4, 0, 105),
  [3323] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_expression, 5, 0, 95),
  [3325] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_expression, 5, 0, 95),
  [3327] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1268),
  [3329] = {.entry = {.count = 1, .reusable = false}}, SHIFT(79),
  [3331] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_expression, 3, 0, 0),
  [3333] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_expression, 3, 0, 0),
  [3335] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_expression, 4, 0, 27),
  [3337] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_expression, 4, 0, 27),
  [3339] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_expression, 5, 0, 106),
  [3341] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_expression, 5, 0, 106),
  [3343] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fun_expression, 5, 0, 106),
  [3345] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fun_expression, 5, 0, 106),
  [3347] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_expression, 4, 0, 55),
  [3349] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_expression, 4, 0, 55),
  [3351] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fun_expression, 4, 0, 70),
  [3353] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fun_expression, 4, 0, 70),
  [3355] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_expression, 3, 0, 55),
  [3357] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_expression, 3, 0, 55),
  [3359] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_expression, 7, 0, 197),
  [3361] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_expression, 7, 0, 197),
  [3363] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_clause, 3, 0, 0),
  [3365] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_clause, 3, 0, 0),
  [3367] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fun_expression, 6, 0, 172),
  [3369] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fun_expression, 6, 0, 172),
  [3371] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_expression, 4, 0, 95),
  [3373] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_expression, 4, 0, 95),
  [3375] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_expression, 4, 0, 27),
  [3377] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_expression, 4, 0, 27),
  [3379] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element, 5, 0, 107),
  [3381] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element, 5, 0, 107),
  [3383] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression, 1, 0, 7),
  [3385] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression, 1, 0, 7),
  [3387] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_expression, 5, 0, 56),
  [3389] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_expression, 5, 0, 56),
  [3391] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_expression, 8, 0, 215),
  [3393] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_expression, 8, 0, 215),
  [3395] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_expression, 5, 0, 56),
  [3397] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_expression, 5, 0, 56),
  [3399] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_expression, 4, 0, 70),
  [3401] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_expression, 4, 0, 70),
  [3403] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_expression, 2, 0, 0),
  [3405] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_expression, 2, 0, 0),
  [3407] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fun_expression, 5, 0, 135),
  [3409] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fun_expression, 5, 0, 135),
  [3411] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_refutation_case, 1, 0, 0),
  [3413] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_refutation_case, 1, 0, 0),
  [3415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(626),
  [3417] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_case, 3, 0, 71),
  [3419] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_case, 3, 0, 71),
  [3421] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_expression, 3, 0, 57),
  [3423] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_expression, 3, 0, 57),
  [3425] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1268),
  [3428] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1476),
  [3430] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1270),
  [3432] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 7, 0, 214),
  [3434] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 7, 0, 214),
  [3436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(180),
  [3438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4409),
  [3440] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 4, 0, 102),
  [3442] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 4, 0, 102),
  [3444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(171),
  [3446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4798),
  [3448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6316),
  [3450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(541),
  [3452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(188),
  [3454] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_let_binding_repeat1, 1, 0, 0),
  [3456] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_let_binding_repeat1, 1, 0, 0),
  [3458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(172),
  [3460] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_pattern, 5, 0, 0),
  [3462] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern, 5, 0, 0),
  [3464] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 7, 0, 213),
  [3466] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 7, 0, 213),
  [3468] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_pattern, 2, 0, 0),
  [3470] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_pattern, 2, 0, 0),
  [3472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(179),
  [3474] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_pattern, 3, 0, 0),
  [3476] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_pattern, 3, 0, 0),
  [3478] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_pattern, 2, 0, 0),
  [3480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_pattern, 2, 0, 0),
  [3482] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_pattern, 6, 0, 0),
  [3484] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern, 6, 0, 0),
  [3486] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 4, 0, 101),
  [3488] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 4, 0, 101),
  [3490] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_pattern, 3, 0, 0),
  [3492] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_pattern, 3, 0, 0),
  [3494] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_pattern, 3, 0, 0),
  [3496] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern, 3, 0, 0),
  [3498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(176),
  [3500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3105),
  [3502] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_pattern, 3, 0, 0),
  [3504] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_pattern, 3, 0, 0),
  [3506] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 1, 0, 17),
  [3508] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 1, 0, 17),
  [3510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(175),
  [3512] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 5, 0, 171),
  [3514] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 5, 0, 171),
  [3516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(178),
  [3518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5793),
  [3520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5899),
  [3522] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_local_open_pattern, 3, 0, 69),
  [3524] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_local_open_pattern, 3, 0, 69),
  [3526] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 8, 0, 224),
  [3528] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 8, 0, 224),
  [3530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(170),
  [3532] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 2, 0, 37),
  [3534] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 2, 0, 37),
  [3536] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_pattern, 7, 0, 0),
  [3538] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern, 7, 0, 0),
  [3540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(205),
  [3542] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1270),
  [3545] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_value_pattern, 1, 0, 19),
  [3547] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_value_pattern, 1, 0, 19),
  [3549] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__delimited_pattern, 1, 0, 0),
  [3551] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__delimited_pattern, 1, 0, 0),
  [3553] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_typed_pattern, 4, 0, 102),
  [3555] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_typed_pattern, 4, 0, 102),
  [3557] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 4, 0, 69),
  [3559] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 4, 0, 69),
  [3561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(208),
  [3563] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_pattern, 4, 0, 0),
  [3565] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_pattern, 4, 0, 0),
  [3567] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 9, 0, 229),
  [3569] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 9, 0, 229),
  [3571] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameter, 6, 0, 196),
  [3573] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 6, 0, 196),
  [3575] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1294),
  [3577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3469),
  [3579] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1294),
  [3582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(554),
  [3584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4646),
  [3586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3664),
  [3588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3674),
  [3590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3456),
  [3592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4658),
  [3594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4598),
  [3596] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6992),
  [3598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1482),
  [3600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2249),
  [3602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4649),
  [3604] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3516),
  [3606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3441),
  [3608] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3517),
  [3610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1465),
  [3612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2624),
  [3614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2520),
  [3616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4481),
  [3618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2588),
  [3620] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_type, 3, 0, 0),
  [3622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2489),
  [3624] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_type, 3, 0, 0),
  [3626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5524),
  [3628] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element_type, 3, 0, 11),
  [3630] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element_type, 3, 0, 11),
  [3632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5452),
  [3634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2502),
  [3636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3087),
  [3638] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_variable, 2, 0, 0),
  [3640] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_variable, 2, 0, 0),
  [3642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2548),
  [3644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2593),
  [3646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(657),
  [3648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(43),
  [3650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(658),
  [3652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(659),
  [3654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(660),
  [3656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(661),
  [3658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3003),
  [3660] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3219),
  [3662] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4354),
  [3664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2513),
  [3666] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2525),
  [3668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2492),
  [3670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2526),
  [3672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3200),
  [3674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5519),
  [3676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6233),
  [3678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(585),
  [3680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(45),
  [3682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(586),
  [3684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(587),
  [3686] = {.entry = {.count = 1, .reusable = false}}, SHIFT(588),
  [3688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(589),
  [3690] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3499),
  [3692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2561),
  [3694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4440),
  [3696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2522),
  [3698] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2569),
  [3700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2615),
  [3702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2616),
  [3704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2594),
  [3706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2562),
  [3708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2625),
  [3710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2613),
  [3712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2291),
  [3714] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_type, 5, 0, 0),
  [3716] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_type, 5, 0, 0),
  [3718] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_local_open_type, 3, 0, 16),
  [3720] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_local_open_type, 3, 0, 16),
  [3722] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(1646),
  [3725] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(1373),
  [3728] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(4968),
  [3731] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(4046),
  [3734] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0),
  [3736] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(4528),
  [3739] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(4731),
  [3742] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(3788),
  [3745] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(3915),
  [3748] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(4805),
  [3751] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(4965),
  [3754] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(5936),
  [3757] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(5943),
  [3760] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(5551),
  [3763] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(1423),
  [3766] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0), SHIFT_REPEAT(6151),
  [3769] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructed_type, 2, 0, 0),
  [3771] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructed_type, 2, 0, 0),
  [3773] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_type, 1, 0, 0),
  [3775] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_type, 1, 0, 0),
  [3777] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1373),
  [3779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4968),
  [3781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4046),
  [3783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4528),
  [3785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4731),
  [3787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3788),
  [3789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3915),
  [3791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4805),
  [3793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4965),
  [3795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1423),
  [3797] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_type, 5, 0, 65),
  [3799] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_type, 5, 0, 65),
  [3801] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructed_type, 5, 0, 0),
  [3803] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructed_type, 5, 0, 0),
  [3805] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_variant_type, 5, 0, 0),
  [3807] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_variant_type, 5, 0, 0),
  [3809] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_type, 5, 0, 0),
  [3811] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_type, 5, 0, 0),
  [3813] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_type, 6, 0, 100),
  [3815] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_type, 6, 0, 100),
  [3817] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_type, 6, 0, 0),
  [3819] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_type, 6, 0, 0),
  [3821] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_variant_type, 6, 0, 0),
  [3823] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_variant_type, 6, 0, 0),
  [3825] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_type, 6, 0, 0),
  [3827] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_type, 6, 0, 0),
  [3829] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__delimited_type, 1, 0, 0),
  [3831] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__delimited_type, 1, 0, 0),
  [3833] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3300),
  [3835] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6092),
  [3837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2530),
  [3839] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2537),
  [3841] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4056),
  [3843] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_type, 7, 0, 138),
  [3845] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_type, 7, 0, 138),
  [3847] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_variant_type, 7, 0, 0),
  [3849] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_variant_type, 7, 0, 0),
  [3851] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_variant_type, 2, 0, 0),
  [3853] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_variant_type, 2, 0, 0),
  [3855] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_type, 2, 0, 0),
  [3857] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_type, 2, 0, 0),
  [3859] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3329),
  [3861] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6163),
  [3863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2550),
  [3865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2568),
  [3867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4052),
  [3869] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_type, 3, 0, 0),
  [3871] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_type, 3, 0, 0),
  [3873] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructed_type, 4, 0, 0),
  [3875] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructed_type, 4, 0, 0),
  [3877] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_variant_type, 4, 0, 0),
  [3879] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_variant_type, 4, 0, 0),
  [3881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2556),
  [3883] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_type, 4, 0, 0),
  [3885] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_type, 4, 0, 0),
  [3887] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_path, 3, 0, 36),
  [3889] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_path, 3, 0, 36),
  [3891] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_path, 1, 0, 3),
  [3893] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_path, 1, 0, 3),
  [3895] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_variant_type, 3, 0, 0),
  [3897] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_variant_type, 3, 0, 0),
  [3899] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_type, 3, 0, 0),
  [3901] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_type, 3, 0, 0),
  [3903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(88),
  [3905] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6287),
  [3907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2506),
  [3909] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2512),
  [3911] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_hash_type, 2, 0, 0),
  [3913] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_hash_type, 2, 0, 0),
  [3915] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_package_type, 4, 0, 33),
  [3917] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_package_type, 4, 0, 33),
  [3919] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_sign_operator, 1, 0, 0),
  [3921] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sign_operator, 1, 0, 0),
  [3923] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_add_operator, 1, 0, 0),
  [3925] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_toplevel_directive, 1, 0, 0),
  [3927] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_toplevel_directive, 1, 0, 0),
  [3929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2931),
  [3931] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4105),
  [3933] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6210),
  [3935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4108),
  [3937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4551),
  [3939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4148),
  [3941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3060),
  [3943] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2205),
  [3945] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__attribute, 2, 0, 0),
  [3947] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__attribute, 2, 0, 0),
  [3949] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__labeled_tuple_type, 3, 0, 0),
  [3951] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__labeled_tuple_type, 3, 0, 0),
  [3953] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_aliased_type, 3, 0, 13),
  [3955] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_aliased_type, 3, 0, 13),
  [3957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(90),
  [3959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(588),
  [3961] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tuple_type_rhs, 3, 0, 0),
  [3963] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__tuple_type_rhs, 3, 0, 0),
  [3965] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_and_operator, 1, 0, 0),
  [3967] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_and_operator, 1, 0, 0),
  [3969] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pow_operator, 1, 0, 0),
  [3971] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pow_operator, 1, 0, 0),
  [3973] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_mult_operator, 1, 0, 0),
  [3975] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mult_operator, 1, 0, 0),
  [3977] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_add_operator, 1, 0, 0),
  [3979] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_or_operator, 1, 0, 0),
  [3981] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_or_operator, 1, 0, 0),
  [3983] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 5, 0, 66),
  [3985] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 5, 0, 66),
  [3987] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type, 1, 0, 0),
  [3989] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type, 1, 0, 0),
  [3991] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 3, 0, 12),
  [3993] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 3, 0, 12),
  [3995] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__proper_tuple_type, 3, 0, 0),
  [3997] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__proper_tuple_type, 3, 0, 0),
  [3999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6135),
  [4001] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_id_repeat1, 2, 0, 0), SHIFT_REPEAT(6135),
  [4004] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_functor, 4, 0, 70),
  [4006] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_functor, 4, 0, 70),
  [4008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3952),
  [4010] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5761),
  [4012] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5871),
  [4014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4576),
  [4016] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7247),
  [4018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1485),
  [4020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2551),
  [4022] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2784),
  [4024] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2785),
  [4026] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2778),
  [4028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(597),
  [4030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2791),
  [4032] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_of, 4, 0, 83),
  [4034] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_of, 4, 0, 83),
  [4036] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2818),
  [4038] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2819),
  [4040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(669),
  [4042] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2841),
  [4044] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2807),
  [4046] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2779),
  [4048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3017),
  [4050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4354),
  [4052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(628),
  [4054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(68),
  [4056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(629),
  [4058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(630),
  [4060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(631),
  [4062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(632),
  [4064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(699),
  [4066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(700),
  [4068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(701),
  [4070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(703),
  [4072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(704),
  [4074] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2957),
  [4076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5371),
  [4078] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1503),
  [4080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(560),
  [4082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(69),
  [4084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(561),
  [4086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(562),
  [4088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(563),
  [4090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(564),
  [4092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(70),
  [4094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6131),
  [4096] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_id_repeat1, 2, 0, 0), SHIFT_REPEAT(6131),
  [4099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(544),
  [4101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(545),
  [4103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(546),
  [4105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(547),
  [4107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(548),
  [4109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(92),
  [4111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6460),
  [4113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6627),
  [4115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(71),
  [4117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2771),
  [4119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3266),
  [4121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5432),
  [4123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4203),
  [4125] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2194),
  [4127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7282),
  [4129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7008),
  [4131] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tuple_type_rhs, 1, 0, 0),
  [4133] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__tuple_type_rhs, 1, 0, 0),
  [4135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3244),
  [4137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5034),
  [4139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4047),
  [4141] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__signature, 1, 0, 0),
  [4143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4532),
  [4145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4813),
  [4147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3503),
  [4149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4727),
  [4151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4904),
  [4153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(93),
  [4155] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 3, 0, 81),
  [4157] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 3, 0, 81),
  [4159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3131),
  [4161] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3658),
  [4163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3582),
  [4165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(642),
  [4167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(73),
  [4169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(643),
  [4171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(644),
  [4173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(645),
  [4175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(646),
  [4177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3358),
  [4179] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_path, 3, 0, 15),
  [4181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1299),
  [4183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1384),
  [4185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1382),
  [4187] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(4225),
  [4190] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(5034),
  [4193] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(4047),
  [4196] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0),
  [4198] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(4532),
  [4201] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(4813),
  [4204] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(3788),
  [4207] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(3503),
  [4210] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(4727),
  [4213] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(4904),
  [4216] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(5936),
  [4219] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(5943),
  [4222] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 2, 0, 0), SHIFT_REPEAT(5551),
  [4225] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5278),
  [4227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(601),
  [4229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(602),
  [4231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(603),
  [4233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(604),
  [4235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(605),
  [4237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4497),
  [4239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3669),
  [4241] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_path, 1, 0, 2),
  [4243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(72),
  [4245] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__module_expression, 1, 0, 0),
  [4247] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__module_expression, 1, 0, 0),
  [4249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7333),
  [4251] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 5, 0, 152),
  [4253] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 5, 0, 152),
  [4255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3427),
  [4257] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 4, 0, 117),
  [4259] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 4, 0, 117),
  [4261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4685),
  [4263] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 4, 0, 118),
  [4265] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 4, 0, 118),
  [4267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(688),
  [4269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(689),
  [4271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(690),
  [4273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(640),
  [4275] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1281),
  [4277] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1281),
  [4280] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_open_module, 3, 0, 47),
  [4282] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_open_module, 3, 0, 47),
  [4284] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_application, 3, 0, 84),
  [4286] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_application, 3, 0, 84),
  [4288] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_packed_module, 5, 0, 153),
  [4290] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_packed_module, 5, 0, 153),
  [4292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2489),
  [4294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2703),
  [4296] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_module_expression, 3, 0, 0),
  [4298] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_module_expression, 3, 0, 0),
  [4300] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_constructor_declaration_repeat1, 1, 0, 0),
  [4302] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_open_module, 4, 0, 83),
  [4304] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_open_module, 4, 0, 83),
  [4306] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_typed_module_expression, 4, 0, 120),
  [4308] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_typed_module_expression, 4, 0, 120),
  [4310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2957),
  [4312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2730),
  [4314] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_packed_module, 4, 0, 119),
  [4316] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_packed_module, 4, 0, 119),
  [4318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(515),
  [4320] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_expression_content, 1, 0, 0),
  [4322] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_packed_module, 7, 0, 201),
  [4324] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_packed_module, 7, 0, 201),
  [4326] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_structure, 3, 0, 0),
  [4328] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_structure, 3, 0, 0),
  [4330] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_structure, 2, 0, 0),
  [4332] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_structure, 2, 0, 0),
  [4334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(687),
  [4336] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module, 3, 0, 47),
  [4338] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_include_module, 3, 0, 47),
  [4340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(691),
  [4342] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_declaration, 3, 0, 0),
  [4344] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_declaration, 3, 0, 0),
  [4346] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constructor_argument, 1, 0, 0),
  [4348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3376),
  [4350] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_open_module, 2, 0, 25),
  [4352] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_open_module, 2, 0, 25),
  [4354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(74),
  [4356] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_module_expression, 1, 0, 0),
  [4358] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_module_expression, 1, 0, 0),
  [4360] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__signature, 2, 0, 0),
  [4362] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_packed_module, 6, 0, 182),
  [4364] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_packed_module, 6, 0, 182),
  [4366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(511),
  [4368] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_application, 2, 0, 48),
  [4370] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_application, 2, 0, 48),
  [4372] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module, 2, 0, 25),
  [4374] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_include_module, 2, 0, 25),
  [4376] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_declaration, 5, 0, 0),
  [4378] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_declaration, 5, 0, 0),
  [4380] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variant_declaration, 1, 0, 0),
  [4382] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variant_declaration, 1, 0, 0),
  [4384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6318),
  [4386] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6544),
  [4388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3477),
  [4390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3439),
  [4392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4788),
  [4394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2718),
  [4396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6401),
  [4398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3090),
  [4400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6420),
  [4402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2994),
  [4404] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_expression, 3, 0, 57),
  [4406] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 1, 0, 2),
  [4408] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 1, 0, 2),
  [4410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4116),
  [4412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5565),
  [4414] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3698),
  [4416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4983),
  [4418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2509),
  [4420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6607),
  [4422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3016),
  [4424] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constructor_argument, 1, 0, 0),
  [4426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3369),
  [4428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2547),
  [4430] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_expression, 3, 0, 56),
  [4432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(572),
  [4434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(526),
  [4436] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1284),
  [4438] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1284),
  [4441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6394),
  [4443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2595),
  [4445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(168),
  [4447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6523),
  [4449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2608),
  [4451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7254),
  [4453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2579),
  [4455] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 2, 0, 2),
  [4457] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 2, 0, 2),
  [4459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4083),
  [4461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7185),
  [4463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4484),
  [4465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6614),
  [4467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4473),
  [4469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6985),
  [4471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4447),
  [4473] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_expression, 4, 0, 115),
  [4475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2810),
  [4477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3237),
  [4479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3246),
  [4481] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__sequence_expression_content_repeat1, 2, 0, 0),
  [4483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(152),
  [4485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3571),
  [4487] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 2, 0, 41),
  [4489] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 2, 0, 41),
  [4491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2110),
  [4493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2190),
  [4495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2809),
  [4497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4334),
  [4499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(556),
  [4501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3251),
  [4503] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_path, 1, 0, 9),
  [4505] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_path, 1, 0, 9),
  [4507] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extended_module_path, 1, 0, 2),
  [4509] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 1, 0, 23),
  [4511] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 1, 0, 23),
  [4513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2111),
  [4515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2192),
  [4517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4010),
  [4519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3086),
  [4521] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_module, 4, 0, 176),
  [4523] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_module, 4, 0, 176),
  [4525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6206),
  [4527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6461),
  [4529] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_path, 3, 0, 45),
  [4531] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_path, 3, 0, 45),
  [4533] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extended_module_path, 3, 0, 15),
  [4535] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3406),
  [4537] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_extended_module_path, 3, 0, 15),
  [4539] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_type, 4, 0, 175),
  [4541] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_type, 4, 0, 175),
  [4543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6079),
  [4545] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_type, 5, 0, 198),
  [4547] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_type, 5, 0, 198),
  [4549] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_type, 6, 0, 216),
  [4551] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_type, 6, 0, 216),
  [4553] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_extended_module_path, 1, 0, 2),
  [4555] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_constraint, 4, 0, 128),
  [4557] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_constraint, 4, 0, 128),
  [4559] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extended_module_path, 4, 0, 0),
  [4561] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_extended_module_path, 4, 0, 0),
  [4563] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__constructor_argument_repeat1, 2, 0, 0),
  [4565] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__constructor_argument_repeat1, 2, 0, 0),
  [4567] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1291),
  [4569] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1288),
  [4571] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_type, 5, 0, 175),
  [4573] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_type, 5, 0, 175),
  [4575] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_type, 6, 0, 198),
  [4577] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_type, 6, 0, 198),
  [4579] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_type, 7, 0, 216),
  [4581] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_type, 7, 0, 216),
  [4583] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_declaration, 7, 0, 0),
  [4585] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_declaration, 7, 0, 0),
  [4587] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_directive, 2, 0, 0),
  [4589] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_directive, 2, 0, 0),
  [4591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(613),
  [4593] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1291),
  [4596] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_class_expression, 3, 0, 57),
  [4598] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_class_expression, 3, 0, 57),
  [4600] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 3, 0, 77),
  [4602] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 3, 0, 77),
  [4604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4241),
  [4606] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, 0, 111),
  [4608] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 4, 0, 111),
  [4610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4231),
  [4612] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_open_class_expression, 4, 0, 70),
  [4614] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_open_class_expression, 4, 0, 70),
  [4616] = {.entry = {.count = 1, .reusable = false}}, SHIFT(761),
  [4618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1215),
  [4620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3388),
  [4622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3389),
  [4624] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2838),
  [4626] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1673),
  [4628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(654),
  [4630] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_binding_repeat1, 2, 0, 0),
  [4632] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_type_binding_repeat1, 2, 0, 0),
  [4634] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2809),
  [4637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3177),
  [4639] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__match_cases_repeat1, 2, 0, 0), SHIFT_REPEAT(1288),
  [4642] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_binding_repeat1, 1, 0, 0),
  [4644] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_type_binding_repeat1, 1, 0, 0),
  [4646] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_binding_repeat1, 2, 0, 0),
  [4648] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_module_binding_repeat1, 2, 0, 0),
  [4650] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5565),
  [4653] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_function, 4, 0, 70),
  [4655] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_function, 4, 0, 70),
  [4657] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3760),
  [4659] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2351),
  [4661] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1223),
  [4663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3327),
  [4665] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2839),
  [4667] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3558),
  [4669] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1748),
  [4671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5801),
  [4673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5905),
  [4675] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2350),
  [4677] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1241),
  [4679] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2842),
  [4681] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1767),
  [4683] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 149),
  [4685] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 149),
  [4687] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_class_type, 1, 0, 0),
  [4689] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__simple_class_type, 1, 0, 0),
  [4691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4177),
  [4693] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, 0, 110),
  [4695] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 4, 0, 110),
  [4697] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external, 6, 0, 89),
  [4699] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external, 6, 0, 89),
  [4701] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6000),
  [4703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4121),
  [4705] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__module_type, 1, 0, 0),
  [4707] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__module_type, 1, 0, 0),
  [4709] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrain_module_type, 5, 0, 199),
  [4711] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constrain_module_type, 5, 0, 199),
  [4713] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signature, 2, 0, 0),
  [4715] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_signature, 2, 0, 0),
  [4717] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1239),
  [4719] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1763),
  [4721] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_type_constraint_repeat1, 2, 0, 0),
  [4723] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_module_type_constraint_repeat1, 2, 0, 0),
  [4725] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_signature, 3, 0, 0),
  [4727] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_signature, 3, 0, 0),
  [4729] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external, 5, 0, 51),
  [4731] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external, 5, 0, 51),
  [4733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(702),
  [4735] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_module_type, 3, 0, 0),
  [4737] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_module_type, 3, 0, 0),
  [4739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3324),
  [4741] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_expression_item_repeat1, 2, 0, 0),
  [4743] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_expression_item_repeat1, 2, 0, 0),
  [4745] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_expression_item_repeat1, 2, 0, 0), SHIFT_REPEAT(5607),
  [4748] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 111),
  [4750] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 111),
  [4752] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 8, 0, 200),
  [4754] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 8, 0, 200),
  [4756] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_declaration, 1, 0, 0),
  [4758] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_declaration, 1, 0, 0),
  [4760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5076),
  [4762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3223),
  [4764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3129),
  [4766] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constructor_declaration_name, 2, 0, 0),
  [4768] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constructor_declaration_name, 2, 0, 0),
  [4770] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 6, 0, 149),
  [4772] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 6, 0, 149),
  [4774] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, 0, 77),
  [4776] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 4, 0, 77),
  [4778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2101),
  [4780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2216),
  [4782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2834),
  [4784] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instantiated_class_type, 5, 0, 0),
  [4786] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_instantiated_class_type, 5, 0, 0),
  [4788] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instantiated_class_type, 4, 0, 0),
  [4790] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_instantiated_class_type, 4, 0, 0),
  [4792] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 148),
  [4794] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 148),
  [4796] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_binding_repeat1, 1, 0, 0),
  [4798] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_module_binding_repeat1, 1, 0, 0),
  [4800] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_body_type, 5, 0, 208),
  [4802] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_body_type, 5, 0, 208),
  [4804] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 110),
  [4806] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 110),
  [4808] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 7, 0, 180),
  [4810] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 7, 0, 180),
  [4812] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 2, 0, 23),
  [4814] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 2, 0, 23),
  [4816] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_body_type, 6, 0, 208),
  [4818] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_body_type, 6, 0, 208),
  [4820] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 6, 0, 180),
  [4822] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 6, 0, 180),
  [4824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2106),
  [4826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2279),
  [4828] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 7, 0, 200),
  [4830] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 7, 0, 200),
  [4832] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_type_constraint_repeat1, 2, 0, 0), SHIFT_REPEAT(5196),
  [4835] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 6, 0, 181),
  [4837] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 6, 0, 181),
  [4839] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_parameter, 2, 0, 0),
  [4841] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_parameter, 2, 0, 0),
  [4843] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 7, 0, 181),
  [4845] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 7, 0, 181),
  [4847] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_constraint, 4, 0, 64),
  [4849] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_constraint, 4, 0, 64),
  [4851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5196),
  [4853] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_open_class_type, 4, 0, 70),
  [4855] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_open_class_type, 4, 0, 70),
  [4857] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 6, 0, 148),
  [4859] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 6, 0, 148),
  [4861] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_constraint, 3, 0, 64),
  [4863] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_constraint, 3, 0, 64),
  [4865] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_parameter, 4, 0, 98),
  [4867] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_parameter, 4, 0, 98),
  [4869] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constructor_declaration_name, 1, 0, 0),
  [4871] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constructor_declaration_name, 1, 0, 0),
  [4873] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_body_type, 2, 0, 0),
  [4875] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_body_type, 2, 0, 0),
  [4877] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_body_type, 3, 0, 0),
  [4879] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_body_type, 3, 0, 0),
  [4881] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 3, 0, 41),
  [4883] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 3, 0, 41),
  [4885] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_item_attribute, 3, 0, 0),
  [4887] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_item_attribute, 3, 0, 0),
  [4889] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 3, 0, 71),
  [4891] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 3, 0, 71),
  [4893] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_item_attribute, 4, 0, 0),
  [4895] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_item_attribute, 4, 0, 0),
  [4897] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 5, 0, 103),
  [4899] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 5, 0, 103),
  [4901] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 5, 0, 104),
  [4903] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 5, 0, 104),
  [4905] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 5, 0, 143),
  [4907] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 5, 0, 143),
  [4909] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 5, 0, 105),
  [4911] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 5, 0, 105),
  [4913] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 5, 0, 144),
  [4915] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 5, 0, 144),
  [4917] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 5, 0, 145),
  [4919] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 5, 0, 145),
  [4921] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0),
  [4923] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0),
  [4925] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0), SHIFT_REPEAT(6000),
  [4928] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0), SHIFT_REPEAT(4121),
  [4931] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_expression_item_repeat1, 1, 0, 0),
  [4933] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_expression_item_repeat1, 1, 0, 0),
  [4935] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__constructor_argument_repeat1, 2, 0, 0), SHIFT_REPEAT(3369),
  [4938] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3349),
  [4940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3497),
  [4942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3465),
  [4944] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6004),
  [4946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3628),
  [4948] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7127),
  [4950] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2991),
  [4952] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5360),
  [4954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4790),
  [4956] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 2, 0, 17),
  [4958] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 2, 0, 17),
  [4960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5695),
  [4962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4977),
  [4964] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__module_typed, 2, 0, 43),
  [4966] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__module_typed, 2, 0, 43),
  [4968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3769),
  [4970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5147),
  [4972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3650),
  [4974] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7240),
  [4976] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5275),
  [4978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5761),
  [4980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5871),
  [4982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3239),
  [4984] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2834),
  [4987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4217),
  [4989] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 3, 0, 82),
  [4991] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 3, 0, 82),
  [4993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4102),
  [4995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4081),
  [4997] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 4, 0, 71),
  [4999] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 4, 0, 71),
  [5001] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 4, 0, 103),
  [5003] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 4, 0, 103),
  [5005] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 4, 0, 104),
  [5007] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 4, 0, 104),
  [5009] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 4, 0, 105),
  [5011] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 4, 0, 105),
  [5013] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 6, 0, 143),
  [5015] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 6, 0, 143),
  [5017] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 6, 0, 144),
  [5019] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 6, 0, 144),
  [5021] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 6, 0, 145),
  [5023] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 6, 0, 145),
  [5025] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 6, 0, 178),
  [5027] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 6, 0, 178),
  [5029] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 2, 0, 46),
  [5031] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 2, 0, 46),
  [5033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4097),
  [5035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4103),
  [5037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4873),
  [5039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5016),
  [5041] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__constructor_argument, 2, 0, 0),
  [5043] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__constructor_argument, 2, 0, 0),
  [5045] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_expression_item_repeat1, 2, 0, 0), SHIFT_REPEAT(5569),
  [5048] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_binding, 7, 0, 178),
  [5050] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_binding, 7, 0, 178),
  [5052] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_functor_type, 3, 0, 63),
  [5054] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_functor_type, 3, 0, 63),
  [5056] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_functor_type, 3, 0, 12),
  [5058] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_functor_type, 3, 0, 12),
  [5060] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_functor_type, 4, 0, 99),
  [5062] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_functor_type, 4, 0, 99),
  [5064] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_functor_type, 2, 0, 32),
  [5066] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_functor_type, 2, 0, 32),
  [5068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4237),
  [5070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1315),
  [5072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2832),
  [5074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4825),
  [5076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1704),
  [5078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3292),
  [5080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5098),
  [5082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(267),
  [5084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5819),
  [5086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5916),
  [5088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1350),
  [5090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1068),
  [5092] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 207),
  [5094] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 207),
  [5096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3365),
  [5098] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variant_declaration, 3, 0, 0),
  [5100] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variant_declaration, 3, 0, 0),
  [5102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4662),
  [5104] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 3, 0, 85),
  [5106] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 3, 0, 85),
  [5108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3316),
  [5110] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_declaration, 3, 0, 0),
  [5112] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_declaration, 3, 0, 0),
  [5114] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_variant_declaration_repeat1, 2, 0, 0),
  [5116] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_variant_declaration_repeat1, 2, 0, 0),
  [5118] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_variant_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(4662),
  [5121] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 124),
  [5123] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 124),
  [5125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3344),
  [5127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6328),
  [5129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1307),
  [5131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(802),
  [5133] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 6, 0, 151),
  [5135] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 6, 0, 151),
  [5137] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 155),
  [5139] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 155),
  [5141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3425),
  [5143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1323),
  [5145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1859),
  [5147] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 3, 0, 45),
  [5149] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 3, 0, 45),
  [5151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3636),
  [5153] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 222),
  [5155] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 222),
  [5157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3375),
  [5159] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_declaration, 4, 0, 0),
  [5161] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_declaration, 4, 0, 0),
  [5163] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variant_declaration, 2, 0, 0),
  [5165] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variant_declaration, 2, 0, 0),
  [5167] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 184),
  [5169] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 184),
  [5171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3332),
  [5173] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 185),
  [5175] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 185),
  [5177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3340),
  [5179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1331),
  [5181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(991),
  [5183] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_id_repeat1, 2, 0, 0), SHIFT_REPEAT(6328),
  [5186] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 2, 0, 50),
  [5188] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 2, 0, 50),
  [5190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3338),
  [5192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1338),
  [5194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2405),
  [5196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1345),
  [5198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1582),
  [5200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1352),
  [5202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(869),
  [5204] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 5, 0, 116),
  [5206] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 5, 0, 116),
  [5208] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 4, 0, 80),
  [5210] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 4, 0, 80),
  [5212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3742),
  [5214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1359),
  [5216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2347),
  [5218] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_record_declaration, 5, 0, 0),
  [5220] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_record_declaration, 5, 0, 0),
  [5222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1366),
  [5224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2654),
  [5226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1370),
  [5228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2090),
  [5230] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6180),
  [5232] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 3, 0, 88),
  [5234] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 3, 0, 88),
  [5236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3395),
  [5238] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3330),
  [5240] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 206),
  [5242] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 206),
  [5244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3362),
  [5246] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 125),
  [5248] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 125),
  [5250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3773),
  [5252] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5530),
  [5254] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 222),
  [5256] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 222),
  [5258] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 9, 0, 225),
  [5260] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 9, 0, 225),
  [5262] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 3, 0, 86),
  [5264] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 3, 0, 86),
  [5266] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 10, 0, 227),
  [5268] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 10, 0, 227),
  [5270] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 10, 0, 228),
  [5272] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 10, 0, 228),
  [5274] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 10, 0, 230),
  [5276] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 10, 0, 230),
  [5278] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 11, 0, 230),
  [5280] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 11, 0, 230),
  [5282] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 205),
  [5284] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 205),
  [5286] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 3, 0, 26),
  [5288] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 3, 0, 26),
  [5290] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_typed, 2, 0, 87),
  [5292] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_typed, 2, 0, 87),
  [5294] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0),
  [5296] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0),
  [5298] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1421),
  [5301] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1421),
  [5304] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 3, 0, 78),
  [5306] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 3, 0, 78),
  [5308] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 185),
  [5310] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 185),
  [5312] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 3, 0, 23),
  [5314] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 3, 0, 23),
  [5316] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 156),
  [5318] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 156),
  [5320] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 3, 0, 121),
  [5322] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 3, 0, 121),
  [5324] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 77),
  [5326] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 77),
  [5328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3213),
  [5330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3123),
  [5332] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 4, 0, 82),
  [5334] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 4, 0, 82),
  [5336] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 187),
  [5338] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 187),
  [5340] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_definition, 3, 0, 0),
  [5342] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_definition, 3, 0, 0),
  [5344] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1421),
  [5346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1421),
  [5348] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 150),
  [5350] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 150),
  [5352] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 112),
  [5354] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 112),
  [5356] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 5, 0, 113),
  [5358] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 5, 0, 113),
  [5360] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 123),
  [5362] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 123),
  [5364] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_floating_attribute, 3, 0, 0),
  [5366] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_floating_attribute, 3, 0, 0),
  [5368] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 85),
  [5370] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 85),
  [5372] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_function_type, 5, 0, 66),
  [5374] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_function_type, 5, 0, 66),
  [5376] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 5, 0, 117),
  [5378] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 5, 0, 117),
  [5380] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4702),
  [5382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3502),
  [5384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3642),
  [5386] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6340),
  [5388] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3091),
  [5390] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5440),
  [5392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4559),
  [5394] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 5, 0, 118),
  [5396] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 5, 0, 118),
  [5398] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 4, 0, 154),
  [5400] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 4, 0, 154),
  [5402] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 4, 0, 121),
  [5404] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 4, 0, 121),
  [5406] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_definition, 4, 0, 0),
  [5408] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_definition, 4, 0, 0),
  [5410] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 184),
  [5412] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 184),
  [5414] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 49),
  [5416] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 49),
  [5418] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 86),
  [5420] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 86),
  [5422] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 123),
  [5424] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 123),
  [5426] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 157),
  [5428] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 157),
  [5430] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 158),
  [5432] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 158),
  [5434] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 124),
  [5436] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 124),
  [5438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5037),
  [5440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3767),
  [5442] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5486),
  [5444] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_function_type, 3, 0, 12),
  [5446] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_function_type, 3, 0, 12),
  [5448] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 125),
  [5450] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 125),
  [5452] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 126),
  [5454] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 126),
  [5456] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 5, 0, 159),
  [5458] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 5, 0, 159),
  [5460] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 204),
  [5462] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 204),
  [5464] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_definition, 2, 0, 0),
  [5466] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_definition, 2, 0, 0),
  [5468] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 203),
  [5470] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 203),
  [5472] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 126),
  [5474] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 126),
  [5476] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 4, 0, 88),
  [5478] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 4, 0, 88),
  [5480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4981),
  [5482] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_floating_attribute, 4, 0, 0),
  [5484] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_floating_attribute, 4, 0, 0),
  [5486] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 3, 0, 46),
  [5488] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 3, 0, 46),
  [5490] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 8, 0, 180),
  [5492] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 8, 0, 180),
  [5494] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 3, 0, 2),
  [5496] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 3, 0, 2),
  [5498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3714),
  [5500] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6355),
  [5502] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5326),
  [5504] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 8, 0, 181),
  [5506] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 8, 0, 181),
  [5508] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 7, 0, 202),
  [5510] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 7, 0, 202),
  [5512] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 7, 0, 217),
  [5514] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 7, 0, 217),
  [5516] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 203),
  [5518] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 203),
  [5520] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_definition, 5, 0, 0),
  [5522] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_definition, 5, 0, 0),
  [5524] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 204),
  [5526] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 204),
  [5528] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 218),
  [5530] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 218),
  [5532] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 205),
  [5534] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 205),
  [5536] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 6, 0, 110),
  [5538] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 6, 0, 110),
  [5540] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 219),
  [5542] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 219),
  [5544] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 220),
  [5546] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 220),
  [5548] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 206),
  [5550] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 206),
  [5552] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 6, 0, 111),
  [5554] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 6, 0, 111),
  [5556] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 6, 0, 150),
  [5558] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 6, 0, 150),
  [5560] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 6, 0, 152),
  [5562] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 6, 0, 152),
  [5564] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 5, 0, 154),
  [5566] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 5, 0, 154),
  [5568] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 183),
  [5570] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 183),
  [5572] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, 0, 41),
  [5574] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 4, 0, 41),
  [5576] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 155),
  [5578] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 155),
  [5580] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 221),
  [5582] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 221),
  [5584] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, 0, 112),
  [5586] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 4, 0, 112),
  [5588] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 122),
  [5590] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 122),
  [5592] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 157),
  [5594] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 157),
  [5596] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 158),
  [5598] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 158),
  [5600] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 187),
  [5602] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 187),
  [5604] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, 0, 113),
  [5606] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 4, 0, 113),
  [5608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5018),
  [5610] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, 0, 78),
  [5612] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 4, 0, 78),
  [5614] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 6, 0, 159),
  [5616] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 6, 0, 159),
  [5618] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 207),
  [5620] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 207),
  [5622] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 8, 0, 186),
  [5624] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 8, 0, 186),
  [5626] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 9, 0, 200),
  [5628] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 9, 0, 200),
  [5630] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 8, 0, 217),
  [5632] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 8, 0, 217),
  [5634] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 8, 0, 225),
  [5636] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 8, 0, 225),
  [5638] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_binding, 4, 0, 81),
  [5640] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_binding, 4, 0, 81),
  [5642] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 218),
  [5644] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 218),
  [5646] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 219),
  [5648] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 219),
  [5650] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 220),
  [5652] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 220),
  [5654] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 226),
  [5656] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 226),
  [5658] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 221),
  [5660] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 221),
  [5662] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 227),
  [5664] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 227),
  [5666] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 9, 0, 228),
  [5668] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 9, 0, 228),
  [5670] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 7, 0, 148),
  [5672] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 7, 0, 148),
  [5674] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 7, 0, 149),
  [5676] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_binding, 7, 0, 149),
  [5678] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 3, 0, 50),
  [5680] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 3, 0, 50),
  [5682] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_type, 1, 0, 0),
  [5684] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__class_type, 1, 0, 0),
  [5686] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_binding, 6, 0, 202),
  [5688] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_binding, 6, 0, 202),
  [5690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5229),
  [5692] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_type_constraint_repeat1, 2, 0, 0), SHIFT_REPEAT(5229),
  [5695] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 7, 0, 183),
  [5697] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 7, 0, 183),
  [5699] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_binding, 10, 0, 226),
  [5701] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_binding, 10, 0, 226),
  [5703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1327),
  [5705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1871),
  [5707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1369),
  [5709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2657),
  [5711] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_specification, 4, 0, 89),
  [5713] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_specification, 4, 0, 89),
  [5715] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_item_extension, 3, 0, 0),
  [5717] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_item_extension, 3, 0, 0),
  [5719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1371),
  [5721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1947),
  [5723] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_item, 1, 0, 0),
  [5725] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_item, 1, 0, 0),
  [5727] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_item_extension, 4, 0, 0),
  [5729] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_item_extension, 4, 0, 0),
  [5731] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exception_definition, 4, 0, 0),
  [5733] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exception_definition, 4, 0, 0),
  [5735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4044),
  [5737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3937),
  [5739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5665),
  [5741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(37),
  [5743] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_item_extension, 4, 0, 0),
  [5745] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_item_extension, 4, 0, 0),
  [5747] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1419),
  [5749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1419),
  [5751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3656),
  [5753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5241),
  [5755] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1419),
  [5758] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1419),
  [5761] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_specification, 4, 0, 51),
  [5763] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_specification, 4, 0, 51),
  [5765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1314),
  [5767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(792),
  [5769] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exception_definition, 2, 0, 0),
  [5771] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exception_definition, 2, 0, 0),
  [5773] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 6, 0, 116),
  [5775] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 6, 0, 116),
  [5777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2774),
  [5779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3269),
  [5781] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_item_extension, 5, 0, 0),
  [5783] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_item_extension, 5, 0, 0),
  [5785] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_item_extension, 5, 0, 0),
  [5787] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_item_extension, 5, 0, 0),
  [5789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1349),
  [5791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1085),
  [5793] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_item_extension, 6, 0, 0),
  [5795] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_quoted_item_extension, 6, 0, 0),
  [5797] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_open_module, 4, 0, 47),
  [5799] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_open_module, 4, 0, 47),
  [5801] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0), SHIFT_REPEAT(6180),
  [5804] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0), SHIFT_REPEAT(4135),
  [5807] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_specification, 5, 0, 89),
  [5809] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_specification, 5, 0, 89),
  [5811] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exception_definition, 3, 0, 0),
  [5813] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exception_definition, 3, 0, 0),
  [5815] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_open_module, 5, 0, 83),
  [5817] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_open_module, 5, 0, 83),
  [5819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1355),
  [5821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(871),
  [5823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1362),
  [5825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2356),
  [5827] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_external_repeat1, 1, 0, 0),
  [5829] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_external_repeat1, 1, 0, 0),
  [5831] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module, 4, 0, 47),
  [5833] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_include_module, 4, 0, 47),
  [5835] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external_declaration, 2, 0, 0),
  [5837] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external_declaration, 2, 0, 0),
  [5839] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__constructor_argument_repeat1, 2, 0, 0), SHIFT_REPEAT(3324),
  [5842] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_item, 2, 0, 0),
  [5844] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_item, 2, 0, 0),
  [5846] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 7, 0, 151),
  [5848] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 7, 0, 151),
  [5850] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external, 6, 0, 51),
  [5852] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external, 6, 0, 51),
  [5854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1334),
  [5856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1012),
  [5858] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_open_module, 3, 0, 25),
  [5860] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_open_module, 3, 0, 25),
  [5862] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module, 3, 0, 25),
  [5864] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_include_module, 3, 0, 25),
  [5866] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value_specification, 3, 0, 51),
  [5868] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value_specification, 3, 0, 51),
  [5870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1341),
  [5872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2383),
  [5874] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 5, 0, 80),
  [5876] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 5, 0, 80),
  [5878] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_type_definition, 4, 0, 45),
  [5880] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_type_definition, 4, 0, 45),
  [5882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2831),
  [5884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3243),
  [5886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3245),
  [5888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4049),
  [5890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3860),
  [5892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5603),
  [5894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(41),
  [5896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4466),
  [5898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3225),
  [5900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1348),
  [5902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1583),
  [5904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1319),
  [5906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1752),
  [5908] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external, 7, 0, 89),
  [5910] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external, 7, 0, 89),
  [5912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3761),
  [5914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5917),
  [5916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(879),
  [5918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(880),
  [5920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(881),
  [5922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(882),
  [5924] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 3, 0, 0),
  [5926] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 3, 0, 0),
  [5928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4202),
  [5930] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 2, 0, 0),
  [5932] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 2, 0, 0),
  [5934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5518),
  [5936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1747),
  [5938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1749),
  [5940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3361),
  [5942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1754),
  [5944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1100),
  [5946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(60),
  [5948] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1437),
  [5950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2697),
  [5952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2371),
  [5954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(46),
  [5956] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1435),
  [5958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(890),
  [5960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2382),
  [5962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4655),
  [5964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2386),
  [5966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1768),
  [5968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3415),
  [5970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3420),
  [5972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2387),
  [5974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2388),
  [5976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1770),
  [5978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1724),
  [5980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1063),
  [5982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(812),
  [5984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1714),
  [5986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(813),
  [5988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3381),
  [5990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1106),
  [5992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2392),
  [5994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2393),
  [5996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2394),
  [5998] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_definition_repeat1, 2, 0, 0),
  [6000] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_type_definition_repeat1, 2, 0, 0),
  [6002] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4202),
  [6005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2395),
  [6007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3391),
  [6009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1107),
  [6011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(873),
  [6013] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_type, 3, 0, 34),
  [6015] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_type, 3, 0, 34),
  [6017] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_polymorphic_type, 3, 0, 16),
  [6019] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_polymorphic_type, 3, 0, 16),
  [6021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3319),
  [6023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3321),
  [6025] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_definition_repeat1, 2, 0, 0),
  [6027] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_module_definition_repeat1, 2, 0, 0),
  [6029] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5683),
  [6032] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_definition, 4, 0, 0),
  [6034] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_definition, 4, 0, 0),
  [6036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5683),
  [6038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__polymorphic_type, 1, 0, 0),
  [6040] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__polymorphic_type, 1, 0, 0),
  [6042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3356),
  [6044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3364),
  [6046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3392),
  [6048] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_definition, 2, 0, 0),
  [6050] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_definition, 2, 0, 0),
  [6052] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module_type, 3, 0, 33),
  [6054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(760),
  [6056] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_definition_repeat1, 2, 0, 0),
  [6058] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_definition_repeat1, 2, 0, 0),
  [6060] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5518),
  [6063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(62),
  [6065] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1438),
  [6067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1979),
  [6069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2659),
  [6071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2662),
  [6073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(58),
  [6075] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1446),
  [6077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(828),
  [6079] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4571),
  [6081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(50),
  [6083] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1451),
  [6085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1751),
  [6087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2663),
  [6089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2664),
  [6091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1092),
  [6093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2665),
  [6095] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_definition, 3, 0, 0),
  [6097] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_definition, 3, 0, 0),
  [6099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2666),
  [6101] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_variant_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(4655),
  [6104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2667),
  [6106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2668),
  [6108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(52),
  [6110] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1429),
  [6112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1835),
  [6114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2669),
  [6116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1887),
  [6118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(875),
  [6120] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 4, 0, 0),
  [6122] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 4, 0, 0),
  [6124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(876),
  [6126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1093),
  [6128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1774),
  [6130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(805),
  [6132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1777),
  [6134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(64),
  [6136] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1440),
  [6138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2358),
  [6140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1781),
  [6142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1958),
  [6144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1782),
  [6146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1785),
  [6148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1786),
  [6150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1787),
  [6152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1789),
  [6154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3121),
  [6156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(682),
  [6158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(49),
  [6160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4502),
  [6162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(598),
  [6164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1971),
  [6166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1986),
  [6168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1025),
  [6170] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 5, 0, 0),
  [6172] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 5, 0, 0),
  [6174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(983),
  [6176] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_type_definition_repeat1, 2, 0, 0),
  [6178] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_type_definition_repeat1, 2, 0, 0),
  [6180] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_type_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5418),
  [6183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1987),
  [6185] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_definition, 4, 0, 0),
  [6187] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_definition, 4, 0, 0),
  [6189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5418),
  [6191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(965),
  [6193] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_definition, 3, 0, 0),
  [6195] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_definition, 3, 0, 0),
  [6197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1988),
  [6199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(966),
  [6201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1989),
  [6203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(773),
  [6205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(981),
  [6207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(954),
  [6209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(955),
  [6211] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 3, 0, 0),
  [6213] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 3, 0, 0),
  [6215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1990),
  [6217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(67),
  [6219] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1448),
  [6221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1119),
  [6223] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_definition, 5, 0, 0),
  [6225] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_definition, 5, 0, 0),
  [6227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(956),
  [6229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1991),
  [6231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1992),
  [6233] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_definition, 2, 0, 0),
  [6235] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_definition, 2, 0, 0),
  [6237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1096),
  [6239] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(2832),
  [6242] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(4825),
  [6245] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0),
  [6247] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(3292),
  [6250] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(5098),
  [6253] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(267),
  [6256] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(5936),
  [6259] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(5819),
  [6262] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(5916),
  [6265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(958),
  [6267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(54),
  [6269] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1432),
  [6271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(984),
  [6273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(816),
  [6275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2385),
  [6277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1109),
  [6279] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_type_definition, 5, 0, 0),
  [6281] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_type_definition, 5, 0, 0),
  [6283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(806),
  [6285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2331),
  [6287] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module_type, 2, 0, 43),
  [6289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2368),
  [6291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2376),
  [6293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2379),
  [6295] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3314),
  [6297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1684),
  [6299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2398),
  [6301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2403),
  [6303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2404),
  [6305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2415),
  [6307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(808),
  [6309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(877),
  [6311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(56),
  [6313] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1433),
  [6315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1572),
  [6317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1585),
  [6319] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_definition, 4, 0, 0),
  [6321] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_definition, 4, 0, 0),
  [6323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1586),
  [6325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1587),
  [6327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1588),
  [6329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1589),
  [6331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1590),
  [6333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1591),
  [6335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1592),
  [6337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1524),
  [6339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(782),
  [6341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(878),
  [6343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3763),
  [6345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1104),
  [6347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1710),
  [6349] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__structure_item, 1, 0, 0),
  [6351] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__structure_item, 1, 0, 0),
  [6353] = {.entry = {.count = 1, .reusable = false}}, SHIFT(284),
  [6355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(333),
  [6357] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2815),
  [6359] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 3, 0, 0),
  [6361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2814),
  [6363] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_toplevel_directive, 2, 0, 0),
  [6365] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_toplevel_directive, 2, 0, 0),
  [6367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5116),
  [6369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(438),
  [6371] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 4, 0, 0),
  [6373] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4469),
  [6375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4661),
  [6377] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4193),
  [6379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5144),
  [6381] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5157),
  [6383] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3240),
  [6385] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4209),
  [6387] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3457),
  [6389] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4182),
  [6391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5171),
  [6393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2840),
  [6395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5114),
  [6397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5437),
  [6399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4198),
  [6401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5438),
  [6403] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__maybe_polymorphic_typed, 2, 0, 35),
  [6405] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__maybe_polymorphic_typed, 2, 0, 35),
  [6407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(315),
  [6409] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 2, 0, 0),
  [6411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2767),
  [6413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3480),
  [6415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3450),
  [6417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5715),
  [6419] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__structure_repeat2, 1, 0, 0),
  [6421] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__structure_repeat2, 1, 0, 0),
  [6423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3454),
  [6425] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module_type, 3, 0, 43),
  [6427] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0), SHIFT_REPEAT(2832),
  [6430] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0), SHIFT_REPEAT(5114),
  [6433] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0),
  [6435] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0), SHIFT_REPEAT(4198),
  [6438] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0), SHIFT_REPEAT(5438),
  [6441] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0), SHIFT_REPEAT(5936),
  [6444] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0), SHIFT_REPEAT(5819),
  [6447] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 2, 0, 0), SHIFT_REPEAT(5916),
  [6450] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_include_module_type, 4, 0, 33),
  [6452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5521),
  [6454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3481),
  [6456] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__local_structure_item, 1, 0, 0),
  [6458] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__local_structure_item, 1, 0, 0),
  [6460] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__structure_repeat2, 2, 0, 0),
  [6462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5449),
  [6464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5516),
  [6466] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3828),
  [6468] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4247),
  [6470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4247),
  [6472] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4276),
  [6474] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1526),
  [6476] = {.entry = {.count = 1, .reusable = false}}, SHIFT(902),
  [6478] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2691),
  [6480] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1972),
  [6482] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2343),
  [6484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4187),
  [6486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5508),
  [6488] = {.entry = {.count = 1, .reusable = false}}, SHIFT(801),
  [6490] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4189),
  [6492] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1065),
  [6494] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2293),
  [6496] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4184),
  [6498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3954),
  [6500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6317),
  [6502] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1874),
  [6504] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1001),
  [6506] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3838),
  [6508] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4764),
  [6510] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1821),
  [6512] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4187),
  [6515] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4176),
  [6517] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2934),
  [6519] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3722),
  [6521] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1863),
  [6523] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5508),
  [6526] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2488),
  [6528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5931),
  [6530] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5931),
  [6533] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5337),
  [6535] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2763),
  [6537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1795),
  [6539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6169),
  [6541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3241),
  [6543] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2611),
  [6545] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2817),
  [6547] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__instance_variable_name, 1, 0, 5),
  [6549] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3841),
  [6551] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2766),
  [6553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2780),
  [6555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3250),
  [6557] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_content, 1, 0, 0),
  [6559] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6307),
  [6561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4088),
  [6563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6265),
  [6565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2927),
  [6567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5983),
  [6569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4786),
  [6571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5959),
  [6573] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__structure_repeat1, 2, 0, 0), SHIFT_REPEAT(4225),
  [6576] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_content_repeat1, 2, 0, 0),
  [6578] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_string_content_repeat1, 2, 0, 0), SHIFT_REPEAT(4247),
  [6581] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_string_content_repeat1, 2, 0, 0), SHIFT_REPEAT(4247),
  [6584] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_string_content_repeat1, 2, 0, 0), SHIFT_REPEAT(4276),
  [6587] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__signature_repeat1, 1, 0, 0),
  [6589] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__typed, 2, 0, 35),
  [6591] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__typed, 2, 0, 35),
  [6593] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_constrained, 2, 0, 75),
  [6595] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_constrained, 2, 0, 75),
  [6597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6100),
  [6599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4246),
  [6601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4657),
  [6603] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6445),
  [6605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4626),
  [6607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5632),
  [6609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4627),
  [6611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4291),
  [6613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3681),
  [6615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5877),
  [6617] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_constrained, 1, 0, 40),
  [6619] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_constrained, 1, 0, 40),
  [6621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4275),
  [6623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4669),
  [6625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4298),
  [6627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3464),
  [6629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4248),
  [6631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3673),
  [6633] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__signature_item, 1, 0, 0),
  [6635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4288),
  [6637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3448),
  [6639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(338),
  [6641] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 3, 0, 94),
  [6643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1206),
  [6645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5104),
  [6647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5560),
  [6649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4415),
  [6651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4665),
  [6653] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_content_repeat1, 1, 0, 0),
  [6655] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_string_content_repeat1, 1, 0, 0),
  [6657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3678),
  [6659] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__coerced, 2, 0, 72),
  [6661] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__coerced, 2, 0, 72),
  [6663] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4177),
  [6665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3242),
  [6667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5431),
  [6669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1218),
  [6671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1783),
  [6673] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 2, 0, 52),
  [6675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5408),
  [6677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1244),
  [6679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5210),
  [6681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5591),
  [6683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1238),
  [6685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5244),
  [6687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5689),
  [6689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4812),
  [6691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(228),
  [6693] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 193),
  [6695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(334),
  [6697] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 3, 0, 90),
  [6699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(465),
  [6701] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 131),
  [6703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(223),
  [6705] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 167),
  [6707] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_constrained, 1, 0, 39),
  [6709] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type_constrained, 1, 0, 39),
  [6711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1208),
  [6713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5186),
  [6715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1233),
  [6717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2565),
  [6719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(221),
  [6721] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 5, 0, 161),
  [6723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4676),
  [6725] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_escape_sequence, 1, 0, 0),
  [6727] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_escape_sequence, 1, 0, 0),
  [6729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1247),
  [6731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2963),
  [6733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1260),
  [6735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1232),
  [6737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1258),
  [6739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1212),
  [6741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5231),
  [6743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3216),
  [6745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1213),
  [6747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4199),
  [6749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1243),
  [6751] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 3, 0, 91),
  [6753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5346),
  [6755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3455),
  [6757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1253),
  [6759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5148),
  [6761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5748),
  [6763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2676),
  [6765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(422),
  [6767] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 4, 0, 129),
  [6769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3683),
  [6771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1254),
  [6773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5151),
  [6775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5787),
  [6777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2418),
  [6779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5357),
  [6781] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameter_constraint, 4, 0, 128),
  [6783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1217),
  [6785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5146),
  [6787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5823),
  [6789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1645),
  [6791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(222),
  [6793] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 166),
  [6795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2786),
  [6797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3234),
  [6799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3459),
  [6801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1237),
  [6803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(363),
  [6805] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 134),
  [6807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1693),
  [6809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3658),
  [6811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1022),
  [6813] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 8, 0, 212),
  [6815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1030),
  [6817] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_effect_pattern, 4, 0, 108),
  [6819] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_effect_pattern, 4, 0, 108),
  [6821] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 8, 0, 223),
  [6823] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 192),
  [6825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4069),
  [6827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5666),
  [6829] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3637),
  [6831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4527),
  [6833] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 194),
  [6835] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 7, 0, 188),
  [6837] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tuple_pattern, 3, 0, 0),
  [6839] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__tuple_pattern, 3, 0, 0),
  [6841] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 7, 0, 189),
  [6843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3124),
  [6845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1562),
  [6847] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern, 1, 0, 20),
  [6849] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__pattern, 1, 0, 20),
  [6851] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__or_pattern_anonymous, 3, 0, 0),
  [6853] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__or_pattern_anonymous, 3, 0, 0),
  [6855] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_alias_pattern, 3, 0, 73),
  [6857] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_alias_pattern, 3, 0, 73),
  [6859] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 134),
  [6861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1569),
  [6863] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 211),
  [6865] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 212),
  [6867] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_pattern, 3, 0, 69),
  [6869] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_pattern, 3, 0, 69),
  [6871] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 165),
  [6873] = {.entry = {.count = 1, .reusable = false}}, SHIFT(3334),
  [6875] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 133),
  [6877] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 5, 0, 130),
  [6879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1059),
  [6881] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lazy_pattern, 2, 0, 37),
  [6883] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lazy_pattern, 2, 0, 37),
  [6885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(925),
  [6887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(931),
  [6889] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 4, 0, 70),
  [6891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(779),
  [6893] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 94),
  [6895] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 7, 0, 209),
  [6897] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 53),
  [6899] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 131),
  [6901] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constructor_pattern, 2, 0, 37),
  [6903] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_constructor_pattern, 2, 0, 37),
  [6905] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 5, 0, 164),
  [6907] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 193),
  [6909] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tag_pattern, 2, 0, 37),
  [6911] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tag_pattern, 2, 0, 37),
  [6913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1917),
  [6915] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 4, 0, 90),
  [6917] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1070),
  [6919] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 6, 0, 106),
  [6921] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2749),
  [6923] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameter_constraint, 5, 0, 128),
  [6925] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 5, 0, 106),
  [6927] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 6, 0, 188),
  [6929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2760),
  [6931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1801),
  [6933] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 8, 0, 209),
  [6935] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 6, 0, 189),
  [6937] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 190),
  [6939] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 6, 0, 161),
  [6941] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element_pattern, 5, 0, 139),
  [6943] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element_pattern, 5, 0, 139),
  [6945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4075),
  [6947] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 6, 0, 162),
  [6949] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 6, 0, 164),
  [6951] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 3, 0, 52),
  [6953] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 165),
  [6955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(768),
  [6957] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 190),
  [6959] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 191),
  [6961] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 5, 0, 129),
  [6963] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 166),
  [6965] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 192),
  [6967] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_initializer, 2, 0, 54),
  [6969] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 4, 0, 91),
  [6971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2050),
  [6973] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 167),
  [6975] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 191),
  [6977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2077),
  [6979] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 133),
  [6981] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_definition, 4, 0, 130),
  [6983] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 132),
  [6985] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 168),
  [6987] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 169),
  [6989] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 194),
  [6991] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern, 1, 0, 0),
  [6993] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__pattern, 1, 0, 0),
  [6995] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 168),
  [6997] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 8, 0, 210),
  [6999] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 9, 0, 223),
  [7001] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 5, 0, 0),
  [7003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2406),
  [7005] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern, 1, 0, 21),
  [7007] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__pattern, 1, 0, 21),
  [7009] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 5, 0, 70),
  [7011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2422),
  [7013] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 169),
  [7015] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lazy_pattern, 3, 0, 69),
  [7017] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lazy_pattern, 3, 0, 69),
  [7019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1680),
  [7021] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__effect_pattern, 1, 0, 0),
  [7023] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__effect_pattern, 1, 0, 0),
  [7025] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 93),
  [7027] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_definition, 5, 0, 162),
  [7029] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_initializer, 3, 0, 54),
  [7031] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element_pattern, 4, 0, 101),
  [7033] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element_pattern, 4, 0, 101),
  [7035] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 8, 0, 211),
  [7037] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 210),
  [7039] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_specification, 3, 0, 94),
  [7041] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_specification, 2, 0, 87),
  [7043] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_specification, 4, 0, 89),
  [7045] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_specification, 5, 0, 89),
  [7047] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_tuple_element_pattern, 2, 0, 18),
  [7049] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2246),
  [7051] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_tuple_element_pattern, 2, 0, 18),
  [7053] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4504),
  [7055] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_specification, 5, 0, 131),
  [7057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2102),
  [7059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2256),
  [7061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2816),
  [7063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6133),
  [7065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4505),
  [7067] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_specification, 4, 0, 51),
  [7069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4496),
  [7071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2096),
  [7073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2247),
  [7075] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_specification, 4, 0, 94),
  [7077] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_specification, 3, 0, 51),
  [7079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4520),
  [7081] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inheritance_specification, 3, 0, 87),
  [7083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6884),
  [7085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4518),
  [7087] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_specification, 4, 0, 131),
  [7089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4509),
  [7091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3461),
  [7093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2701),
  [7095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3252),
  [7097] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2816),
  [7100] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exception_pattern, 2, 0, 37),
  [7102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1275),
  [7104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1325),
  [7106] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exception_pattern, 2, 0, 37),
  [7108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5307),
  [7110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1328),
  [7112] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_cons_pattern, 3, 0, 62),
  [7114] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_cons_pattern, 3, 0, 62),
  [7116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(227),
  [7118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2232),
  [7120] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2541),
  [7122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2322),
  [7124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3347),
  [7126] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_exception_pattern, 3, 0, 69),
  [7128] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_exception_pattern, 3, 0, 69),
  [7130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4226),
  [7132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(350),
  [7134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2437),
  [7136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2968),
  [7138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2747),
  [7140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4211),
  [7142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2732),
  [7144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4216),
  [7146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6809),
  [7148] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6809),
  [7150] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6821),
  [7152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_object_expression_repeat1, 1, 0, 0),
  [7154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2837),
  [7156] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4814),
  [7158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4814),
  [7160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7068),
  [7162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1046),
  [7164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(269),
  [7166] = {.entry = {.count = 1, .reusable = false}}, SHIFT(271),
  [7168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(272),
  [7170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5946),
  [7172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1946),
  [7174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(423),
  [7176] = {.entry = {.count = 1, .reusable = false}}, SHIFT(424),
  [7178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(425),
  [7180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(940),
  [7182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(396),
  [7184] = {.entry = {.count = 1, .reusable = false}}, SHIFT(397),
  [7186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(398),
  [7188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(851),
  [7190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(410),
  [7192] = {.entry = {.count = 1, .reusable = false}}, SHIFT(411),
  [7194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(412),
  [7196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(755),
  [7198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(339),
  [7200] = {.entry = {.count = 1, .reusable = false}}, SHIFT(340),
  [7202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(341),
  [7204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4537),
  [7206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(372),
  [7208] = {.entry = {.count = 1, .reusable = false}}, SHIFT(373),
  [7210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(374),
  [7212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1518),
  [7214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(403),
  [7216] = {.entry = {.count = 1, .reusable = false}}, SHIFT(404),
  [7218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(405),
  [7220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2656),
  [7222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(416),
  [7224] = {.entry = {.count = 1, .reusable = false}}, SHIFT(417),
  [7226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(418),
  [7228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6027),
  [7230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6990),
  [7232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1818),
  [7234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2409),
  [7236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(430),
  [7238] = {.entry = {.count = 1, .reusable = false}}, SHIFT(431),
  [7240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(432),
  [7242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1844),
  [7244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(386),
  [7246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(387),
  [7248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(388),
  [7250] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(5666),
  [7253] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_field, 1, 0, 0),
  [7255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5157),
  [7257] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_body_type_repeat1, 1, 0, 0),
  [7259] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__class_field_specification, 1, 0, 0),
  [7261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3204),
  [7263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4444),
  [7265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5045),
  [7267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3182),
  [7269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3155),
  [7271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3174),
  [7273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3020),
  [7275] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_quoted_string_content, 1, 0, 0),
  [7277] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2195),
  [7279] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2231),
  [7281] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2581),
  [7283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5088),
  [7285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3195),
  [7287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2618),
  [7289] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_type_constraint_repeat1, 2, 0, 0), SHIFT_REPEAT(5088),
  [7292] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__constructor_argument_repeat1, 2, 0, 0), SHIFT_REPEAT(3376),
  [7295] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_binding_repeat1, 2, 0, 0), SHIFT_REPEAT(2837),
  [7298] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_quoted_string_content_repeat1, 2, 0, 0), SHIFT_REPEAT(4814),
  [7301] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_quoted_string_content_repeat1, 2, 0, 0), SHIFT_REPEAT(4814),
  [7304] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_quoted_string_content_repeat1, 2, 0, 0),
  [7306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(681),
  [7308] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2803),
  [7310] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_expression, 1, 0, 0),
  [7312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2801),
  [7314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4094),
  [7316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4098),
  [7318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3619),
  [7320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5106),
  [7322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4562),
  [7324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1269),
  [7326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1385),
  [7328] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2808),
  [7330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5429),
  [7332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1316),
  [7334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(119),
  [7336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(346),
  [7338] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_payload, 2, 0, 0),
  [7340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(361),
  [7342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4837),
  [7344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1267),
  [7346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1322),
  [7348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1326),
  [7350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1285),
  [7352] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1375),
  [7354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1274),
  [7356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5430),
  [7358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1380),
  [7360] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_binding_pattern_content, 1, 0, 0),
  [7362] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4537),
  [7364] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6200),
  [7366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1997),
  [7368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(124),
  [7370] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2200),
  [7372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3027),
  [7374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1276),
  [7376] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1305),
  [7378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1290),
  [7380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5276),
  [7382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1306),
  [7384] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__sequence_pattern_content, 1, 0, 0),
  [7386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5576),
  [7388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5500),
  [7390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4377),
  [7392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1287),
  [7394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1343),
  [7396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5270),
  [7398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1344),
  [7400] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6269),
  [7402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2979),
  [7404] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6270),
  [7406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2433),
  [7408] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6309),
  [7410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4605),
  [7412] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6044),
  [7414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1996),
  [7416] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6288),
  [7418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3013),
  [7420] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6293),
  [7422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2463),
  [7424] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5950),
  [7426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2292),
  [7428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2621),
  [7430] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1431),
  [7432] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4820),
  [7434] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4663),
  [7436] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6190),
  [7438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2587),
  [7440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4204),
  [7442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(130),
  [7444] = {.entry = {.count = 1, .reusable = false}}, SHIFT(961),
  [7446] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4900),
  [7448] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4671),
  [7450] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6057),
  [7452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4036),
  [7454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5644),
  [7456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5506),
  [7458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3176),
  [7460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3864),
  [7462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2627),
  [7464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4058),
  [7466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4064),
  [7468] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2241),
  [7470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3940),
  [7472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3947),
  [7474] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6093),
  [7476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4375),
  [7478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3880),
  [7480] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6047),
  [7482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4616),
  [7484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3961),
  [7486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3974),
  [7488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5150),
  [7490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4004),
  [7492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4009),
  [7494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(117),
  [7496] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6052),
  [7498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2224),
  [7500] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_argument_type, 3, 0, 11),
  [7502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4022),
  [7504] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0), SHIFT_REPEAT(5983),
  [7507] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_external_repeat1, 2, 0, 0), SHIFT_REPEAT(4157),
  [7510] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_object_type_repeat1, 2, 0, 0),
  [7512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4025),
  [7514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4031),
  [7516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3935),
  [7518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(125),
  [7520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4032),
  [7522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3869),
  [7524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3877),
  [7526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4345),
  [7528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3916),
  [7530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3922),
  [7532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1289),
  [7534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3955),
  [7536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3968),
  [7538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4916),
  [7540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(122),
  [7542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3910),
  [7544] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5995),
  [7546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4991),
  [7548] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_labeled_tuple_element_type, 3, 0, 11), SHIFT(3269),
  [7551] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2208),
  [7553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2059),
  [7555] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6003),
  [7557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5002),
  [7559] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_argument_type, 4, 0, 31),
  [7561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(127),
  [7563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2606),
  [7565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2505),
  [7567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1298),
  [7569] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6028),
  [7571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4856),
  [7573] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5948),
  [7575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4308),
  [7577] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6039),
  [7579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4872),
  [7581] = {.entry = {.count = 1, .reusable = false}}, SHIFT(855),
  [7583] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4631),
  [7585] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6209),
  [7587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(136),
  [7589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3865),
  [7591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5940),
  [7593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5522),
  [7595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3315),
  [7597] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_quoted_string_content_repeat1, 1, 0, 0),
  [7599] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_quoted_string_content_repeat1, 1, 0, 0),
  [7601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1701),
  [7603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2408),
  [7605] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6500),
  [7607] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4647),
  [7609] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6155),
  [7611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1745),
  [7613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1766),
  [7615] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4181),
  [7617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5089),
  [7619] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5755),
  [7621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3732),
  [7623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5142),
  [7625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(770),
  [7627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3380),
  [7629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3382),
  [7631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1860),
  [7633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1906),
  [7635] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__sequence_binding_pattern_content_repeat1, 2, 0, 0),
  [7637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2318),
  [7639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1853),
  [7641] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__simple_typed, 2, 0, 35),
  [7643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3588),
  [7645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1773),
  [7647] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_type_constraint_repeat1, 2, 0, 0), SHIFT_REPEAT(5142),
  [7650] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__type_param_repeat1, 2, 0, 0), SHIFT_REPEAT(5144),
  [7653] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__type_param_repeat1, 2, 0, 0),
  [7655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2329),
  [7657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1810),
  [7659] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_binding_pattern, 3, 0, 69),
  [7661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1817),
  [7663] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__sequence_pattern_content_repeat1, 2, 0, 0),
  [7665] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6441),
  [7667] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4645),
  [7669] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6276),
  [7671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1017),
  [7673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2339),
  [7675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1643),
  [7677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2909),
  [7679] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_item_repeat1, 2, 0, 0), SHIFT_REPEAT(5747),
  [7682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1009),
  [7684] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_binding_pattern, 4, 0, 142),
  [7686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4648),
  [7688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(781),
  [7690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2370),
  [7692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1024),
  [7694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3688),
  [7696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3686),
  [7698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1026),
  [7700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(962),
  [7702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3387),
  [7704] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_pattern, 4, 0, 142),
  [7706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(761),
  [7708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(953),
  [7710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1690),
  [7712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1062),
  [7714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1087),
  [7716] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_pattern, 3, 0, 69),
  [7718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3371),
  [7720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1551),
  [7722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1598),
  [7724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1053),
  [7726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1715),
  [7728] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_variant_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(4648),
  [7731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(811),
  [7733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1571),
  [7735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1576),
  [7737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1579),
  [7739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3397),
  [7741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1061),
  [7743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(870),
  [7745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(916),
  [7747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3655),
  [7749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1102),
  [7751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3661),
  [7753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(927),
  [7755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(933),
  [7757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(865),
  [7759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(803),
  [7761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(867),
  [7763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2655),
  [7765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1090),
  [7767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2729),
  [7769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1683),
  [7771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2752),
  [7773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2677),
  [7775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2717),
  [7777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3401),
  [7779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2758),
  [7781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3402),
  [7783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3730),
  [7785] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tag_specification, 4, 0, 0),
  [7787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2773),
  [7789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3731),
  [7791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1940),
  [7793] = {.entry = {.count = 1, .reusable = true}}, SHIFT(786),
  [7795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2020),
  [7797] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tag_specification, 3, 0, 0),
  [7799] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3405),
  [7801] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3407),
  [7803] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2054),
  [7805] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2091),
  [7807] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2066),
  [7809] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3409),
  [7811] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2041),
  [7813] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3606),
  [7815] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2350),
  [7817] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1595),
  [7819] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4181),
  [7821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6557),
  [7823] = {.entry = {.count = 1, .reusable = true}}, SHIFT(994),
  [7825] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6618),
  [7827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6180),
  [7829] = {.entry = {.count = 1, .reusable = true}}, SHIFT(259),
  [7831] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1372),
  [7833] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2803),
  [7835] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_binding_pattern, 1, 0, 0),
  [7837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7155),
  [7839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7322),
  [7841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(767),
  [7843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7177),
  [7845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1902),
  [7847] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_param, 1, 0, 0),
  [7849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6374),
  [7851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1703),
  [7853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5459),
  [7855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4232),
  [7857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5624),
  [7859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6810),
  [7861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1549),
  [7863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1157),
  [7865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5821),
  [7867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5274),
  [7869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6000),
  [7871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3496),
  [7873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2997),
  [7875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2805),
  [7877] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__tag_spec, 1, 0, 0),
  [7879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5520),
  [7881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7025),
  [7883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(914),
  [7885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1351),
  [7887] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_pattern, 1, 0, 0),
  [7889] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4213),
  [7891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5869),
  [7893] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__type_param_repeat1, 1, 0, 0),
  [7895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4221),
  [7897] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5945),
  [7899] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_param, 2, 0, 42),
  [7901] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_param, 2, 0, 0),
  [7903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7149),
  [7905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1055),
  [7907] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_param, 1, 0, 22),
  [7909] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5267),
  [7911] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7198),
  [7913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2724),
  [7915] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6417),
  [7917] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tag_specification, 5, 0, 0),
  [7919] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5510),
  [7921] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_tag_specification_repeat1, 2, 0, 0),
  [7923] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_expression_item_repeat1, 2, 0, 0), SHIFT_REPEAT(5717),
  [7926] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_tag_specification_repeat1, 2, 0, 0), SHIFT_REPEAT(2773),
  [7929] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6370),
  [7931] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2006),
  [7933] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6289),
  [7935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4567),
  [7937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(320),
  [7939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5380),
  [7941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4238),
  [7943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5941),
  [7945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(492),
  [7947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(493),
  [7949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3775),
  [7951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5234),
  [7953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(494),
  [7955] = {.entry = {.count = 1, .reusable = true}}, SHIFT(495),
  [7957] = {.entry = {.count = 1, .reusable = true}}, SHIFT(496),
  [7959] = {.entry = {.count = 1, .reusable = true}}, SHIFT(497),
  [7961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5316),
  [7963] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6878),
  [7965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2354),
  [7967] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6013),
  [7969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5198),
  [7971] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6806),
  [7973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4446),
  [7975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3349),
  [7977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4472),
  [7979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2611),
  [7981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5687),
  [7983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4804),
  [7985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1794),
  [7987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2812),
  [7989] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1170),
  [7991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3654),
  [7993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4962),
  [7995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2950),
  [7997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3021),
  [7999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2952),
  [8001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2614),
  [8003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3014),
  [8005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5724),
  [8007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2038),
  [8009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2948),
  [8011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5113),
  [8013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1494),
  [8015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1115),
  [8017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5674),
  [8019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3660),
  [8021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4859),
  [8023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2320),
  [8025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4106),
  [8027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5163),
  [8029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5766),
  [8031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(928),
  [8033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3208),
  [8035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4293),
  [8037] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_constructor_declaration_repeat1, 2, 0, 0),
  [8039] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_constructor_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(6024),
  [8042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3570),
  [8044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(934),
  [8046] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1415),
  [8048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1415),
  [8050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2578),
  [8052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5688),
  [8054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5599),
  [8056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1723),
  [8058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2982),
  [8060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5718),
  [8062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5337),
  [8064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5149),
  [8066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2351),
  [8068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4084),
  [8070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2435),
  [8072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1805),
  [8074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3687),
  [8076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3665),
  [8078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3207),
  [8080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2995),
  [8082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2213),
  [8084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3564),
  [8086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1027),
  [8088] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_polymorphic_variant_type_repeat1, 2, 0, 0), SHIFT_REPEAT(2614),
  [8091] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_polymorphic_variant_type_repeat1, 2, 0, 0),
  [8093] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__abstract_type_repeat1, 2, 0, 0), SHIFT_REPEAT(5621),
  [8096] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__abstract_type_repeat1, 2, 0, 0),
  [8098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3092),
  [8100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2845),
  [8102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3692),
  [8104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5833),
  [8106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3589),
  [8108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4099),
  [8110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2776),
  [8112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4485),
  [8114] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_instance_variable_specification_repeat1, 2, 0, 0),
  [8116] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_instance_variable_specification_repeat1, 2, 0, 0), SHIFT_REPEAT(5755),
  [8119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3643),
  [8121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5533),
  [8123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2753),
  [8125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5714),
  [8127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5633),
  [8129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2708),
  [8131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3194),
  [8133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(776),
  [8135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2782),
  [8137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(345),
  [8139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2462),
  [8141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2504),
  [8143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5944),
  [8145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2413),
  [8147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3228),
  [8149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3841),
  [8151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3029),
  [8153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3548),
  [8155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5398),
  [8157] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6295),
  [8159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5051),
  [8161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(959),
  [8163] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6154),
  [8165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4109),
  [8167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2605),
  [8169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5556),
  [8171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5083),
  [8173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2590),
  [8175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5843),
  [8177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5092),
  [8179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1709),
  [8181] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1148),
  [8183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5781),
  [8185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5097),
  [8187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2971),
  [8189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4827),
  [8191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1566),
  [8193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5103),
  [8195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3445),
  [8197] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6068),
  [8199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3617),
  [8201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5115),
  [8203] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_instantiated_class_type_repeat1, 2, 0, 0),
  [8205] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_instantiated_class_type_repeat1, 2, 0, 0), SHIFT_REPEAT(2805),
  [8208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5122),
  [8210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5417),
  [8212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5130),
  [8214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1573),
  [8216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4461),
  [8218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2067),
  [8220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3591),
  [8222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4491),
  [8224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2494),
  [8226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5707),
  [8228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6210),
  [8230] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__type_params_repeat1, 2, 0, 0),
  [8232] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__type_params_repeat1, 2, 0, 0), SHIFT_REPEAT(4637),
  [8235] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_attribute_id_repeat1, 2, 0, 0), SHIFT_REPEAT(6295),
  [8238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2055),
  [8240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2584),
  [8242] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_method_specification_repeat1, 2, 0, 0),
  [8244] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_method_specification_repeat1, 2, 0, 0), SHIFT_REPEAT(5821),
  [8247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3768),
  [8249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3770),
  [8251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2806),
  [8253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(783),
  [8255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2790),
  [8257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5625),
  [8259] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_polymorphic_variant_type_repeat2, 2, 0, 0),
  [8261] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_polymorphic_variant_type_repeat2, 2, 0, 0), SHIFT_REPEAT(6069),
  [8264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5932),
  [8266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4924),
  [8268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1083),
  [8270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5621),
  [8272] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__abstract_type, 2, 0, 0),
  [8274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3546),
  [8276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(360),
  [8278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2290),
  [8280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1994),
  [8282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4492),
  [8284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5608),
  [8286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3850),
  [8288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3774),
  [8290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3574),
  [8292] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1415),
  [8295] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_value_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(1415),
  [8298] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tag_specification, 1, 0, 0),
  [8300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2726),
  [8302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1521),
  [8304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2725),
  [8306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4486),
  [8308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(323),
  [8310] = {.entry = {.count = 1, .reusable = false}}, SHIFT(326),
  [8312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(327),
  [8314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5469),
  [8316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2034),
  [8318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5175),
  [8320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5031),
  [8322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5389),
  [8324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5482),
  [8326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3205),
  [8328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5033),
  [8330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2082),
  [8332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2661),
  [8334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5627),
  [8336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5036),
  [8338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4547),
  [8340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4175),
  [8342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4818),
  [8344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2374),
  [8346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5377),
  [8348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2381),
  [8350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6004),
  [8352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4869),
  [8354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5292),
  [8356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4882),
  [8358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2373),
  [8360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5779),
  [8362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1967),
  [8364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4903),
  [8366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4845),
  [8368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(788),
  [8370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5287),
  [8372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1050),
  [8374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(527),
  [8376] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_instance_variable_expression, 1, 0, 0),
  [8378] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_type_repeat1, 2, 0, 0), SHIFT_REPEAT(3287),
  [8381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5353),
  [8383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2399),
  [8385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3539),
  [8387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3098),
  [8389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5085),
  [8391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2344),
  [8393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6226),
  [8395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3178),
  [8397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6105),
  [8399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5003),
  [8401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1740),
  [8403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5373),
  [8405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(827),
  [8407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5420),
  [8409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1679),
  [8411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3082),
  [8413] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_record_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(5773),
  [8416] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_record_declaration_repeat1, 2, 0, 0),
  [8418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6235),
  [8420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4778),
  [8422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5006),
  [8424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4796),
  [8426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(364),
  [8428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4782),
  [8430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4840),
  [8432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6311),
  [8434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4476),
  [8436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4816),
  [8438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4637),
  [8440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7262),
  [8442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(518),
  [8444] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__abstract_type_repeat1, 1, 0, 1),
  [8446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6645),
  [8448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5314),
  [8450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5416),
  [8452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1295),
  [8454] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_indexing_operator_path, 3, 0, 0),
  [8456] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_indexing_operator_path, 3, 0, 0),
  [8458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(380),
  [8460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3088),
  [8462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2210),
  [8464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1266),
  [8466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4794),
  [8468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6590),
  [8470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4802),
  [8472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4706),
  [8474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4588),
  [8476] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__sequence_binding_pattern_content_repeat1, 2, 0, 0), SHIFT_REPEAT(1313),
  [8479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4821),
  [8481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1699),
  [8483] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(4175),
  [8486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6891),
  [8488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5318),
  [8490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1271),
  [8492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4510),
  [8494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2988),
  [8496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5348),
  [8498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2299),
  [8500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4715),
  [8502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2240),
  [8504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4824),
  [8506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5387),
  [8508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3535),
  [8510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3136),
  [8512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4969),
  [8514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4910),
  [8516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4742),
  [8518] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5779),
  [8521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1309),
  [8523] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_pattern, 2, 0, 68),
  [8525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4799),
  [8527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6173),
  [8529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5493),
  [8531] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__sequence_expression_content_repeat1, 2, 0, 0), SHIFT_REPEAT(558),
  [8534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4455),
  [8536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5442),
  [8538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4333),
  [8540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4694),
  [8542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2069),
  [8544] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5353),
  [8547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(542),
  [8549] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_expression, 2, 0, 44),
  [8551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4979),
  [8553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4745),
  [8555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4997),
  [8557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1060),
  [8559] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_record_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(5167),
  [8562] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_record_expression_repeat1, 2, 0, 0),
  [8564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1300),
  [8566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4753),
  [8568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3196),
  [8570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5338),
  [8572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4870),
  [8574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1908),
  [8576] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_copy_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(5580),
  [8579] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_object_copy_expression_repeat1, 2, 0, 0),
  [8581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2864),
  [8583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4707),
  [8585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5271),
  [8587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1911),
  [8589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4958),
  [8591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(780),
  [8593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4704),
  [8595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3030),
  [8597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1324),
  [8599] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_binding_pattern, 2, 0, 68),
  [8601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4705),
  [8603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2442),
  [8605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6062),
  [8607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4876),
  [8609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2632),
  [8611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5325),
  [8613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2414),
  [8615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2623),
  [8617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4963),
  [8619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4709),
  [8621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1040),
  [8623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6259),
  [8625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4710),
  [8627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4887),
  [8629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1802),
  [8631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1830),
  [8633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5010),
  [8635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(235),
  [8637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4896),
  [8639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6705),
  [8641] = {.entry = {.count = 1, .reusable = true}}, SHIFT(752),
  [8643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(239),
  [8645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6825),
  [8647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6869),
  [8649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5294),
  [8651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5988),
  [8653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4471),
  [8655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [8657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4368),
  [8659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3164),
  [8661] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_type_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(5442),
  [8664] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__sequence_pattern_content_repeat1, 2, 0, 0), SHIFT_REPEAT(1318),
  [8667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5954),
  [8669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6482),
  [8671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6870),
  [8673] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5296),
  [8675] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5109),
  [8677] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4919),
  [8679] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1011),
  [8681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6486),
  [8683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5981),
  [8685] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_instance_variable_specification_repeat1, 1, 0, 0),
  [8687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4988),
  [8689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5339),
  [8691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1016),
  [8693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5997),
  [8695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4925),
  [8697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5028),
  [8699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2023),
  [8701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3100),
  [8703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6017),
  [8705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4927),
  [8707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1031),
  [8709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6611),
  [8711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1862),
  [8713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3112),
  [8715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6032),
  [8717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4933),
  [8719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6042),
  [8721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5491),
  [8723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6059),
  [8725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5958),
  [8727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5262),
  [8729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4953),
  [8731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1553),
  [8733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6072),
  [8735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4513),
  [8737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6081),
  [8739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6088),
  [8741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5434),
  [8743] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1556),
  [8745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6094),
  [8747] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6101),
  [8749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6107),
  [8751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6113),
  [8753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6119),
  [8755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6122),
  [8757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6125),
  [8759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6128),
  [8761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2589),
  [8763] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_method_specification_repeat1, 1, 0, 0),
  [8765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4959),
  [8767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5404),
  [8769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2555),
  [8771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6175),
  [8773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6158),
  [8775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5460),
  [8777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6281),
  [8779] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6285),
  [8781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6286),
  [8783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(520),
  [8785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2475),
  [8787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4961),
  [8789] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1570),
  [8791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(342),
  [8793] = {.entry = {.count = 1, .reusable = false}}, SHIFT(343),
  [8795] = {.entry = {.count = 1, .reusable = true}}, SHIFT(344),
  [8797] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5501),
  [8799] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__sequence_expression_content_repeat1, 2, 0, 0), SHIFT_REPEAT(599),
  [8802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(941),
  [8804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2482),
  [8806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4966),
  [8808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4525),
  [8810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1945),
  [8812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1969),
  [8814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4797),
  [8816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4407),
  [8818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4982),
  [8820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(918),
  [8822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5162),
  [8824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6040),
  [8826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5301),
  [8828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(921),
  [8830] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2582),
  [8832] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__sequence_binding_pattern_content_repeat1, 2, 0, 0), SHIFT_REPEAT(1360),
  [8835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4986),
  [8837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5308),
  [8839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6282),
  [8841] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6283),
  [8843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6284),
  [8845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4987),
  [8847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(932),
  [8849] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_record_binding_pattern_repeat1, 2, 0, 0), SHIFT_REPEAT(5044),
  [8852] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_record_binding_pattern_repeat1, 2, 0, 0),
  [8854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2107),
  [8856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(375),
  [8858] = {.entry = {.count = 1, .reusable = false}}, SHIFT(376),
  [8860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(377),
  [8862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3417),
  [8864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1522),
  [8866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5966),
  [8868] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5967),
  [8870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5968),
  [8872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4990),
  [8874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(389),
  [8876] = {.entry = {.count = 1, .reusable = false}}, SHIFT(390),
  [8878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(391),
  [8880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(331),
  [8882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5990),
  [8884] = {.entry = {.count = 1, .reusable = false}}, SHIFT(5991),
  [8886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5992),
  [8888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5265),
  [8890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(399),
  [8892] = {.entry = {.count = 1, .reusable = false}}, SHIFT(400),
  [8894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(401),
  [8896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5498),
  [8898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6006),
  [8900] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6007),
  [8902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6008),
  [8904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(406),
  [8906] = {.entry = {.count = 1, .reusable = false}}, SHIFT(407),
  [8908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(408),
  [8910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6536),
  [8912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6020),
  [8914] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6021),
  [8916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6022),
  [8918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2576),
  [8920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(413),
  [8922] = {.entry = {.count = 1, .reusable = false}}, SHIFT(414),
  [8924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(415),
  [8926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6033),
  [8928] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6034),
  [8930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6035),
  [8932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(419),
  [8934] = {.entry = {.count = 1, .reusable = false}}, SHIFT(420),
  [8936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(421),
  [8938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4939),
  [8940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1064),
  [8942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6049),
  [8944] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6050),
  [8946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6051),
  [8948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(426),
  [8950] = {.entry = {.count = 1, .reusable = false}}, SHIFT(427),
  [8952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(428),
  [8954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5005),
  [8956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2733),
  [8958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6001),
  [8960] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6002),
  [8962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6010),
  [8964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6064),
  [8966] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6065),
  [8968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6066),
  [8970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(433),
  [8972] = {.entry = {.count = 1, .reusable = false}}, SHIFT(434),
  [8974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(435),
  [8976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6074),
  [8978] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6075),
  [8980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6076),
  [8982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6082),
  [8984] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6083),
  [8986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6084),
  [8988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6089),
  [8990] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6090),
  [8992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6091),
  [8994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6095),
  [8996] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6096),
  [8998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6097),
  [9000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6102),
  [9002] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6103),
  [9004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6104),
  [9006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6108),
  [9008] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6109),
  [9010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6110),
  [9012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6114),
  [9014] = {.entry = {.count = 1, .reusable = false}}, SHIFT(6115),
  [9016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6116),
  [9018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5363),
  [9020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2741),
  [9022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6954),
  [9024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2635),
  [9026] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_record_pattern_repeat1, 2, 0, 0), SHIFT_REPEAT(5072),
  [9029] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_record_pattern_repeat1, 2, 0, 0),
  [9031] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6960),
  [9033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6962),
  [9035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5008),
  [9037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5370),
  [9039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5011),
  [9041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2709),
  [9043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(852),
  [9045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5013),
  [9047] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__sequence_pattern_content_repeat1, 2, 0, 0), SHIFT_REPEAT(1367),
  [9050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3263),
  [9052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5539),
  [9054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6189),
  [9056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5537),
  [9058] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_indexing_operator_path, 1, 0, 0),
  [9060] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_indexing_operator_path, 1, 0, 0),
  [9062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5399),
  [9064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6932),
  [9066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6159),
  [9068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6733),
  [9070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2516),
  [9072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1961),
  [9074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6981),
  [9076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2012),
  [9078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6740),
  [9080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5334),
  [9082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3695),
  [9084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2609),
  [9086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4865),
  [9088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1963),
  [9090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6993),
  [9092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1538),
  [9094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1968),
  [9096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2808),
  [9098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5719),
  [9100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6272),
  [9102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7105),
  [9104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7106),
  [9106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7107),
  [9108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(596),
  [9110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2521),
  [9112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2332),
  [9114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7052),
  [9116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2792),
  [9118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(712),
  [9120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5957),
  [9122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5955),
  [9124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2338),
  [9126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7061),
  [9128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5960),
  [9130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2706),
  [9132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2319),
  [9134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(874),
  [9136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(174),
  [9138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4115),
  [9140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(980),
  [9142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7119),
  [9144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4521),
  [9146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3676),
  [9148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4574),
  [9150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5772),
  [9152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6025),
  [9154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7121),
  [9156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7122),
  [9158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7123),
  [9160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(184),
  [9162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6853),
  [9164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6058),
  [9166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4530),
  [9168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6234),
  [9170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6480),
  [9172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6541),
  [9174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6860),
  [9176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5839),
  [9178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6170),
  [9180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7136),
  [9182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7137),
  [9184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7138),
  [9186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2364),
  [9188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6918),
  [9190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3298),
  [9192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6863),
  [9194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1108),
  [9196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6650),
  [9198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6156),
  [9200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6185),
  [9202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(571),
  [9204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(348),
  [9206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2424),
  [9208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1776),
  [9210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6415),
  [9212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6217),
  [9214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7150),
  [9216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7151),
  [9218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7152),
  [9220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6225),
  [9222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6505),
  [9224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2498),
  [9226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2215),
  [9228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6694),
  [9230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6166),
  [9232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6877),
  [9234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5038),
  [9236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6248),
  [9238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4696),
  [9240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6925),
  [9242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2707),
  [9244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6310),
  [9246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7164),
  [9248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7165),
  [9250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7166),
  [9252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2754),
  [9254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2220),
  [9256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6710),
  [9258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(540),
  [9260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6880),
  [9262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4836),
  [9264] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_params, 4, 0, 0),
  [9266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3740),
  [9268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4829),
  [9270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6464),
  [9272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2057),
  [9274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1870),
  [9276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(200),
  [9278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6968),
  [9280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4587),
  [9282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6071),
  [9284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7178),
  [9286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7179),
  [9288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7180),
  [9290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6969),
  [9292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2235),
  [9294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(612),
  [9296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4057),
  [9298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2815),
  [9300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3302),
  [9302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6895),
  [9304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(967),
  [9306] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4673),
  [9308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3047),
  [9310] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 3, 0, 179),
  [9312] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_params, 1, 0, 0),
  [9314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6019),
  [9316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7192),
  [9318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7193),
  [9320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7194),
  [9322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1114),
  [9324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4030),
  [9326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(394),
  [9328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3056),
  [9330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6443),
  [9332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6219),
  [9334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7199),
  [9336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7200),
  [9338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7201),
  [9340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(683),
  [9342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6148),
  [9344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6926),
  [9346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6126),
  [9348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6258),
  [9350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7204),
  [9352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7205),
  [9354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7206),
  [9356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(775),
  [9358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6220),
  [9360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6273),
  [9362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7209),
  [9364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7210),
  [9366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7211),
  [9368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3845),
  [9370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6565),
  [9372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4307),
  [9374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6291),
  [9376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7214),
  [9378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7215),
  [9380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7216),
  [9382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(711),
  [9384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6298),
  [9386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7219),
  [9388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7220),
  [9390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7221),
  [9392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5952),
  [9394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7224),
  [9396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7225),
  [9398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7226),
  [9400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4638),
  [9402] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_type, 2, 0, 10),
  [9404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5971),
  [9406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7229),
  [9408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7230),
  [9410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7231),
  [9412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(625),
  [9414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1788),
  [9416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6447),
  [9418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4639),
  [9420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6945),
  [9422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1080),
  [9424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7077),
  [9426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(839),
  [9428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6937),
  [9430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3233),
  [9432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2025),
  [9434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2490),
  [9436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7134),
  [9438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2951),
  [9440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2112),
  [9442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6321),
  [9444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2777),
  [9446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2176),
  [9448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7091),
  [9450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(999),
  [9452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2846),
  [9454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7060),
  [9456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4737),
  [9458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4451),
  [9460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3562),
  [9462] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_polymorphic_variant_type_repeat2, 1, 0, 0),
  [9464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1504),
  [9466] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__polymorphic_typed, 2, 0, 35),
  [9468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(969),
  [9470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3281),
  [9472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(639),
  [9474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6623),
  [9476] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4619),
  [9478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4459),
  [9480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3786),
  [9482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2236),
  [9484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2925),
  [9486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7096),
  [9488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2789),
  [9490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2735),
  [9492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(349),
  [9494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2736),
  [9496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1885),
  [9498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6976),
  [9500] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 2, 0, 147),
  [9502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1948),
  [9504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7009),
  [9506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1142),
  [9508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7043),
  [9510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1889),
  [9512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6984),
  [9514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2267),
  [9516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3078),
  [9518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3230),
  [9520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(698),
  [9522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2440),
  [9524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6177),
  [9526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(219),
  [9528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2248),
  [9530] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type_params, 3, 0, 0),
  [9532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1441),
  [9534] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4664),
  [9536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(653),
  [9538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1826),
  [9540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6924),
  [9542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6255),
  [9544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(555),
  [9546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2580),
  [9548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(204),
  [9550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6840),
  [9552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2412),
  [9554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2835),
  [9556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(848),
  [9558] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4652),
  [9560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1966),
  [9562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(378),
  [9564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(379),
  [9566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1627),
  [9568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7101),
  [9570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(220),
  [9572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4060),
  [9574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7115),
  [9576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2451),
  [9578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6602),
  [9580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4809),
  [9582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4543),
  [9584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5842),
  [9586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(392),
  [9588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(393),
  [9590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2874),
  [9592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2252),
  [9594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2660),
  [9596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2830),
  [9598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1890),
  [9600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4063),
  [9602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7129),
  [9604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4263),
  [9606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1126),
  [9608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6957),
  [9610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2770),
  [9612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1116),
  [9614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3058),
  [9616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6509),
  [9618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4523),
  [9620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2455),
  [9622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6625),
  [9624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2445),
  [9626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3359),
  [9628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3248),
  [9630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6290),
  [9632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1708),
  [9634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6296),
  [9636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2601),
  [9638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4515),
  [9640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1702),
  [9642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1008),
  [9644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6719),
  [9646] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_guard, 2, 0, 27),
  [9648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3122),
  [9650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3620),
  [9652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1018),
  [9654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6727),
  [9656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4266),
  [9658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5735),
  [9660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1564),
  [9662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6428),
  [9664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6436),
  [9666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(814),
  [9668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6442),
  [9670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1534),
  [9672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6799),
  [9674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1536),
  [9676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6811),
  [9678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7276),
  [9680] = {.entry = {.count = 1, .reusable = false}}, SHIFT(4622),
  [9682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2751),
  [9684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(901),
  [9686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5651),
  [9688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5953),
  [9690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6950),
  [9692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6011),
  [9694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7081),
  [9696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7082),
  [9698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7083),
  [9700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6951),
  [9702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6952),
  [9704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6448),
  [9706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3015),
  [9708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2999),
  [9710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(897),
  [9712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6862),
  [9714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(899),
  [9716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6872),
  [9718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6449),
  [9720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2431),
  [9722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2690),
  [9724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5543),
  [9726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2013),
  [9728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2100),
  [9730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6907),
  [9732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2686),
  [9734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6910),
  [9736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1597),
  [9738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4511),
  [9740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(819),
  [9742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6453),
  [9744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2688),
  [9746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6916),
  [9748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(668),
  [9750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2764),
  [9752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4494),
  [9754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6728),
  [9756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4847),
  [9758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7314),
  [9760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4457),
  [9762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2759),
  [9764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4915),
  [9766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4857),
  [9768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7321),
  [9770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4573),
  [9772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(774),
  [9774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1855),
  [9776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6247),
  [9778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4033),
  [9780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3772),
  [9782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3634),
  [9784] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type, 1, 0, 0),
  [9786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2705),
  [9788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4780),
  [9790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7069),
  [9792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3738),
  [9794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6157),
  [9796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(347),
  [9798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2297),
  [9800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6603),
  [9802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6465),
  [9804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3423),
  [9806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2031),
  [9808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2053),
  [9810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2233),
  [9812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2432),
  [9814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2783),
  [9816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2471),
  [9818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(268),
  [9820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2162),
  [9822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1800),
  [9824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2199),
  [9826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2324),
  [9828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3953),
  [9830] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7228),
  [9832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6594),
  [9834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3472),
  [9836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2325),
  [9838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2326),
  [9840] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_argument_type, 5, 0, 67),
  [9842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(286),
  [9844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4282),
  [9846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2271),
  [9848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2021),
  [9850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4283),
  [9852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1842),
  [9854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(241),
  [9856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1806),
  [9858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1807),
  [9860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1808),
  [9862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1159),
  [9864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5528),
  [9866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1160),
  [9868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6251),
  [9870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2328),
  [9872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4286),
  [9874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2998),
  [9876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5009),
  [9878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4475),
  [9880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1843),
  [9882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5280),
  [9884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4089),
  [9886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(234),
  [9888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3706),
  [9890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3590),
  [9892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4255),
  [9894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1809),
  [9896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5048),
  [9898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(758),
  [9900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(384),
  [9902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(759),
  [9904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3394),
  [9906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1180),
  [9908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(753),
  [9910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1190),
  [9912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1689),
  [9914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3486),
  [9916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5648),
  [9918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(787),
  [9920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1939),
  [9922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4465),
  [9924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2335),
  [9926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(237),
  [9928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(238),
  [9930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2336),
  [9932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1654),
  [9934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2337),
  [9936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(763),
  [9938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1784),
  [9940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5161),
  [9942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5655),
  [9944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1202),
  [9946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1920),
  [9948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3861),
  [9950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3421),
  [9952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1814),
  [9954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1815),
  [9956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1816),
  [9958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5030),
  [9960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1158),
  [9962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6086),
  [9964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(240),
  [9966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1081),
  [9968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6871),
  [9970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1834),
  [9972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3879),
  [9974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6056),
  [9976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1956),
  [9978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4186),
  [9980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(817),
  [9982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3057),
  [9984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7348),
  [9986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4489),
  [9988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1790),
  [9990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3019),
  [9992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2438),
  [9994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2001),
  [9996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4971),
  [9998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(807),
  [10000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(300),
  [10002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1970),
  [10004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2140),
  [10006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4367),
  [10008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4817),
  [10010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2186),
  [10012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(429),
  [10014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2089),
  [10016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4120),
  [10018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(362),
  [10020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5710),
  [10022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4800),
  [10024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4839),
  [10026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(459),
  [10028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(667),
  [10030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4188),
  [10032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4843),
  [10034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4815),
  [10036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2203),
  [10038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6982),
  [10040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3253),
  [10042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(842),
  [10044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5224),
  [10046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1682),
  [10048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1796),
  [10050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6864),
  [10052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1797),
  [10054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4456),
  [10056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(468),
  [10058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3931),
  [10060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1043),
  [10062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(809),
  [10064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4289),
  [10066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(710),
  [10068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4292),
  [10070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1044),
  [10072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(810),
  [10074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1045),
  [10076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4195),
  [10078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1179),
  [10080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1183),
  [10082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6231),
  [10084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2458),
  [10086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6147),
  [10088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(815),
  [10090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(437),
  [10092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2225),
  [10094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3059),
  [10096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2949),
  [10098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(243),
  [10100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1711),
  [10102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3353),
  [10104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1187),
  [10106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1147),
  [10108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3374),
  [10110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2244),
  [10112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2464),
  [10114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5716),
  [10116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1164),
  [10118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(305),
  [10120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1165),
  [10122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4426),
  [10124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3945),
  [10126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3354),
  [10128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(303),
  [10130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1010),
  [10132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2429),
  [10134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2182),
  [10136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4429),
  [10138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7316),
  [10140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3950),
  [10142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6476),
  [10144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6302),
  [10146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3180),
  [10148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6986),
  [10150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(230),
  [10152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1014),
  [10154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5723),
  [10156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(129),
  [10158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1089),
  [10160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2436),
  [10162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2191),
  [10164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2448),
  [10166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3790),
  [10168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3958),
  [10170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1742),
  [10172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4280),
  [10174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4270),
  [10176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1798),
  [10178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1200),
  [10180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1195),
  [10182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6009),
  [10184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(477),
  [10186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2795),
  [10188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4412),
  [10190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2187),
  [10192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2188),
  [10194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(482),
  [10196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1199),
  [10198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1163),
  [10200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(245),
  [10202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(369),
  [10204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1019),
  [10206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5771),
  [10208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(370),
  [10210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(487),
  [10212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2942),
  [10214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3969),
  [10216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(818),
  [10218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3052),
  [10220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(247),
  [10222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3988),
  [10224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6037),
  [10226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3135),
  [10228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1447),
  [10230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(680),
  [10232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2943),
  [10234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6067),
  [10236] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_argument_type, 6, 0, 97),
  [10238] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__infix_operator, 1, 0, 0),
  [10240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4196),
  [10242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1023),
  [10244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3998),
  [10246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5102),
  [10248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4279),
  [10250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4281),
  [10252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2470),
  [10254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2765),
  [10256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1189),
  [10258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1175),
  [10260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6150),
  [10262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2984),
  [10264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3084),
  [10266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2217),
  [10268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(249),
  [10270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(853),
  [10272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1196),
  [10274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1198),
  [10276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(131),
  [10278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4989),
  [10280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5838),
  [10282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5759),
  [10284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3085),
  [10286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(253),
  [10288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4008),
  [10290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4174),
  [10292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1839),
  [10294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2461),
  [10296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1028),
  [10298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4014),
  [10300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6176),
  [10302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3410),
  [10304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2196),
  [10306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2070),
  [10308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1029),
  [10310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2341),
  [10312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2804),
  [10314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4019),
  [10316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3884),
  [10318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4269),
  [10320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4277),
  [10322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(861),
  [10324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(250),
  [10326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1151),
  [10328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1155),
  [10330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6203),
  [10332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1110),
  [10334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3384),
  [10336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1161),
  [10338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1167),
  [10340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(946),
  [10342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(943),
  [10344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(945),
  [10346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5861),
  [10348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2460),
  [10350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2885),
  [10352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4361),
  [10354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4024),
  [10356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5000),
  [10358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3111),
  [10360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(382),
  [10362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2465),
  [10364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4028),
  [10366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6232),
  [10368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(960),
  [10370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1032),
  [10372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4585),
  [10374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1692),
  [10376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1003),
  [10378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5998),
  [10380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(847),
  [10382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3868),
  [10384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4592),
  [10386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4299),
  [10388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4243),
  [10390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1041),
  [10392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(944),
  [10394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1169),
  [10396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1203),
  [10398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6300),
  [10400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(947),
  [10402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(939),
  [10404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3113),
  [10406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1181),
  [10408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1186),
  [10410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2218),
  [10412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(757),
  [10414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(756),
  [10416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5933),
  [10418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1006),
  [10420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1007),
  [10422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(383),
  [10424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4017),
  [10426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(754),
  [10428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2478),
  [10430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4143),
  [10432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(972),
  [10434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3854),
  [10436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6319),
  [10438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2221),
  [10440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3426),
  [10442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4077),
  [10444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(263),
  [10446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4403),
  [10448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2596),
  [10450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3866),
  [10452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1013),
  [10454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4257),
  [10456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3894),
  [10458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4436),
  [10460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1188),
  [10462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1191),
  [10464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5982),
  [10466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1117),
  [10468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(989),
  [10470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1173),
  [10472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1162),
  [10474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2251),
  [10476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1146),
  [10478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5557),
  [10480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4191),
  [10482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6902),
  [10484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(439),
  [10486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3876),
  [10488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2372),
  [10490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1935),
  [10492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4016),
  [10494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2920),
  [10496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3889),
  [10498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6243),
  [10500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(950),
  [10502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(951),
  [10504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2598),
  [10506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(539),
  [10508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3201),
  [10510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3202),
  [10512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2068),
  [10514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3912),
  [10516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6172),
  [10518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1177),
  [10520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1192),
  [10522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6229),
  [10524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2452),
  [10526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(289),
  [10528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2986),
  [10530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1150),
  [10532] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1154),
  [10534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1727),
  [10536] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5572),
  [10538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1730),
  [10540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1736),
  [10542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1529),
  [10544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3920),
  [10546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5760),
  [10548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1058),
  [10550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(488),
  [10552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1532),
  [10554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3929),
  [10556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6118),
  [10558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1552),
  [10560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4192),
  [10562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2239),
  [10564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5007),
  [10566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1878),
  [10568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4190),
  [10570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3422),
  [10572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2787),
  [10574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3944),
  [10576] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1533),
  [10578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1554),
  [10580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6211),
  [10582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(952),
  [10584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5583),
  [10586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2298),
  [10588] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2075),
  [10590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1535),
  [10592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3964),
  [10594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1869),
  [10596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3989),
  [10598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6245),
  [10600] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6162),
  [10602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2242),
  [10604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6716),
  [10606] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_character_content, 1, 0, 0),
  [10608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1537),
  [10610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3137),
  [10612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2301),
  [10614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4200),
  [10616] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6262),
  [10618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1744),
  [10620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1073),
  [10622] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2992),
  [10624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2993),
  [10626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(974),
  [10628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6274),
  [10630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1540),
  [10632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4208),
  [10634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2056),
  [10636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(769),
  [10638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1098),
  [10640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6292),
  [10642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(903),
  [10644] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4483),
  [10646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(896),
  [10648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1132),
  [10650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1756),
  [10652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(904),
  [10654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6304),
  [10656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1757),
  [10658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1765),
  [10660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1881),
  [10662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2193),
  [10664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2996),
  [10666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4295),
  [10668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5961),
  [10670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2040),
  [10672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1882),
  [10674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(898),
  [10676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3299),
  [10678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4206),
  [10680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2377),
  [10682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5976),
  [10684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(976),
  [10686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6917),
  [10688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2607),
  [10690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(900),
  [10692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1737),
  [10694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4529),
  [10696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1074),
  [10698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4881),
  [10700] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2353),
  [10702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4911),
  [10704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2367),
  [10706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1103),
  [10708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4462),
  [10710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3238),
  [10712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6346),
  [10714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2811),
  [10716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1557),
  [10718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2617),
  [10720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(888),
  [10722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1428),
  [10724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2131),
  [10726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3829),
  [10728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3304),
  [10730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3220),
  [10732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2430),
  [10734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(295),
  [10736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2685),
  [10738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2813),
  [10740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3830),
  [10742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2631),
  [10744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1563),
  [10746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2417),
  [10748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2450),
  [10750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(464),
  [10752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5099),
  [10754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2941),
  [10756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(352),
  [10758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2687),
  [10760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(266),
  [10762] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [10764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(297),
  [10766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(583),
  [10768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2689),
  [10770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4197),
  [10772] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6989),
  [10774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(137),
  [10776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4085),
  [10778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1876),
  [10780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1829),
  [10782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6264),
  [10784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6275),
  [10786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5112),
  [10788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(118),
  [10790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1072),
  [10792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1567),
  [10794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1879),
  [10796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(233),
  [10798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1568),
  [10800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2453),
  [10802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1544),
  [10804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(777),
  [10806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5188),
  [10808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7124),
  [10810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1561),
  [10812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5461),
  [10814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(301),
  [10816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5462),
  [10818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4542),
  [10820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2696),
  [10822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5980),
  [10824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1049),
  [10826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1507),
  [10828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6757),
  [10830] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6758),
  [10832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6767),
  [10834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1508),
  [10836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6784),
  [10838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1512),
  [10840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2692),
  [10842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1101),
  [10844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2296),
  [10846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2481),
  [10848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6814),
  [10850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4699),
  [10852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6830),
  [10854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(351),
  [10856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1960),
  [10858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1575),
  [10860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5111),
  [10862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2693),
  [10864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4618),
  [10866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2264),
  [10868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1565),
  [10870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2501),
  [10872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1513),
  [10874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1514),
  [10876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1516),
  [10878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2427),
  [10880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1886),
  [10882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3939),
  [10884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3257),
  [10886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1938),
  [10888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1962),
  [10890] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6012),
  [10892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1943),
  [10894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1891),
  [10896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4449),
  [10898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4252),
  [10900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4589),
  [10902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(329),
  [10904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2391),
  [10906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1965),
  [10908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2477),
  [10910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3393),
  [10912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2479),
  [10914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6214),
  [10916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4387),
  [10918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(330),
  [10920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(570),
  [10922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2257),
  [10924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2258),
  [10926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2283),
  [10928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(308),
  [10930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4235),
  [10932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(831),
  [10934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4080),
  [10936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1981),
  [10938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1105),
  [10940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(444),
  [10942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3093),
  [10944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(917),
  [10946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(905),
  [10948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3839),
  [10950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3652),
  [10952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6790),
  [10954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(906),
  [10956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(919),
  [10958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5032),
  [10960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4422),
  [10962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4584),
  [10964] = {.entry = {.count = 1, .reusable = true}}, SHIFT(697),
  [10966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(445),
  [10968] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2466),
  [10970] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1973),
  [10972] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4091),
  [10974] = {.entry = {.count = 1, .reusable = true}}, SHIFT(447),
  [10976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3840),
  [10978] = {.entry = {.count = 1, .reusable = true}}, SHIFT(449),
  [10980] = {.entry = {.count = 1, .reusable = true}}, SHIFT(450),
  [10982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2363),
  [10984] = {.entry = {.count = 1, .reusable = true}}, SHIFT(922),
  [10986] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1974),
  [10988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2268),
  [10990] = {.entry = {.count = 1, .reusable = true}}, SHIFT(310),
  [10992] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5131),
  [10994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3127),
  [10996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2467),
  [10998] = {.entry = {.count = 1, .reusable = true}}, SHIFT(907),
  [11000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(926),
  [11002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(255),
  [11004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3271),
  [11006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2626),
  [11008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2334),
  [11010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3331),
  [11012] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2094),
  [11014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2833),
  [11016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(314),
  [11018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2284),
  [11020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5049),
  [11022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2355),
  [11024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(116),
  [11026] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2407),
  [11028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3431),
  [11030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(284),
  [11032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(929),
  [11034] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_string, 2, 0, 0),
  [11036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6279),
  [11038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3226),
  [11040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2154),
  [11042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(930),
  [11044] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1184),
  [11046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3227),
  [11048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1978),
  [11050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2375),
  [11052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5056),
  [11054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2380),
  [11056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(316),
  [11058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2289),
  [11060] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6437),
  [11062] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6438),
  [11064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6439),
  [11066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3231),
  [11068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(321),
  [11070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2103),
  [11072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(553),
  [11074] = {.entry = {.count = 1, .reusable = true}}, SHIFT(858),
  [11076] = {.entry = {.count = 1, .reusable = true}}, SHIFT(859),
  [11078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5081),
  [11080] = {.entry = {.count = 1, .reusable = true}}, SHIFT(840),
  [11082] = {.entry = {.count = 1, .reusable = true}}, SHIFT(463),
  [11084] = {.entry = {.count = 1, .reusable = true}}, SHIFT(319),
  [11086] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2078),
  [11088] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5964),
  [11090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2080),
  [11092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(864),
  [11094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1172),
  [11096] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2362),
  [11098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5087),
  [11100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2411),
  [11102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2410),
  [11104] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2346),
  [11106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6537),
  [11108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6538),
  [11110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6539),
  [11112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4468),
  [11114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2700),
  [11116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2071),
  [11118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2653),
  [11120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(843),
  [11122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(841),
  [11124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(844),
  [11126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4062),
  [11128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2702),
  [11130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2259),
  [11132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5093),
  [11134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4787),
  [11136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6586),
  [11138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6587),
  [11140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6588),
  [11142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(273),
  [11144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6712),
  [11146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(778),
  [11148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4065),
  [11150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(820),
  [11152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1983),
  [11154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5100),
  [11156] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6629),
  [11158] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6630),
  [11160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6631),
  [11162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1984),
  [11164] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_payload, 3, 0, 0),
  [11166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2788),
  [11168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1985),
  [11170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1964),
  [11172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(287),
  [11174] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1111),
  [11176] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1907),
  [11178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(395),
  [11180] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6668),
  [11182] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6669),
  [11184] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6670),
  [11186] = {.entry = {.count = 1, .reusable = true}}, SHIFT(822),
  [11188] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1904),
  [11190] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1905),
  [11192] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7269),
  [11194] = {.entry = {.count = 1, .reusable = true}}, SHIFT(213),
  [11196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5682),
  [11198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1909),
  [11200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3232),
  [11202] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3333),
  [11204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6707),
  [11206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6708),
  [11208] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6709),
  [11210] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4998),
  [11212] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3235),
  [11214] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2300),
  [11216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(257),
  [11218] = {.entry = {.count = 1, .reusable = true}}, SHIFT(251),
  [11220] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2073),
  [11222] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6564),
  [11224] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2781),
  [11226] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6744),
  [11228] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6745),
  [11230] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6746),
  [11232] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4163),
  [11234] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2731),
  [11236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4777),
  [11238] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1066),
  [11240] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2737),
  [11242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(825),
  [11244] = {.entry = {.count = 1, .reusable = true}}, SHIFT(826),
  [11246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3166),
  [11248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6780),
  [11250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6781),
  [11252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6782),
  [11254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(133),
  [11256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(611),
  [11258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6803),
  [11260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6804),
  [11262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6805),
  [11264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1913),
  [11266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2612),
  [11268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6816),
  [11270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6817),
  [11272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6818),
  [11274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1921),
  [11276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3379),
  [11278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6826),
  [11280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6827),
  [11282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6828),
  [11284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(354),
  [11286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6836),
  [11288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6837),
  [11290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6838),
  [11292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3028),
  [11294] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6846),
  [11296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6847),
  [11298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6848),
  [11300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2081),
  [11302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3032),
  [11304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6856),
  [11306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6857),
  [11308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6858),
  [11310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4819),
  [11312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4092),
  [11314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6866),
  [11316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6867),
  [11318] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6868),
  [11320] = {.entry = {.count = 1, .reusable = true}}, SHIFT(279),
  [11322] = {.entry = {.count = 1, .reusable = true}}, SHIFT(356),
  [11324] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3377),
  [11326] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3805),
  [11328] = {.entry = {.count = 1, .reusable = true}}, SHIFT(358),
  [11330] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6900),
  [11332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(270),
  [11334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6920),
  [11336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5831),
  [11338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(443),
  [11340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4498),
  [11342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4795),
  [11344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4803),
  [11346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2743),
  [11348] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3918),
  [11350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2439),
  [11352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1916),
  [11354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(215),
  [11356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2444),
  [11358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2585),
  [11360] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3138),
  [11362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3325),
  [11364] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2750),
  [11366] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2084),
  [11368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(359),
  [11370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1918),
  [11372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(409),
  [11374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1424),
  [11376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(217),
  [11378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4180),
  [11380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3328),
  [11382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(128),
  [11384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3789),
  [11386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1716),
  [11388] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__quoted_string, 3, 0, 0),
  [11390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1034),
  [11392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7056),
  [11394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2755),
  [11396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5854),
  [11398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(283),
  [11400] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2756),
  [11402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4183),
  [11404] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5862),
  [11406] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1035),
  [11408] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6540),
  [11410] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5868),
  [11412] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5874),
  [11414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5879),
  [11416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5883),
  [11418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5888),
  [11420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5894),
  [11422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5898),
  [11424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5900),
  [11426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5902),
  [11428] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5904),
  [11430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5906),
  [11432] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5908),
  [11434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5910),
  [11436] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4374),
  [11438] = {.entry = {.count = 1, .reusable = true}}, SHIFT(218),
  [11440] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1075),
  [11442] = {.entry = {.count = 1, .reusable = true}}, SHIFT(624),
  [11444] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2712),
  [11446] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2416),
  [11448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2714),
  [11450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2775),
  [11452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2715),
  [11454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1036),
  [11456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2716),
  [11458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(126),
  [11460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3592),
  [11462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2421),
  [11464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3982),
  [11466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5893),
  [11468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2728),
  [11470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2740),
  [11472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2746),
  [11474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4853),
  [11476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2622),
  [11478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2092),
  [11480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1804),
  [11482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4858),
  [11484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6360),
  [11486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(638),
  [11488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(454),
  [11490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5225),
  [11492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2181),
  [11494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(652),
  [11496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1193),
  [11498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2978),
  [11500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(595),
  [11502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4160),
  [11504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4161),
  [11506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(451),
  [11508] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2772),
  [11510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1755),
  [11512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3050),
  [11514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3917),
  [11516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3270),
  [11518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(258),
  [11520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2058),
  [11522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1171),
  [11524] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 4, 0, 0),
  [11526] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 3, 0, 0),
};

enum ts_external_scanner_symbol_identifiers {
  ts_external_token_comment = 0,
  ts_external_token__left_quoted_string_delimiter = 1,
  ts_external_token__right_quoted_string_delimiter = 2,
  ts_external_token_DQUOTE = 3,
  ts_external_token_line_number_directive = 4,
  ts_external_token__null = 5,
  ts_external_token__error_sentinel = 6,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token_comment] = sym_comment,
  [ts_external_token__left_quoted_string_delimiter] = sym__left_quoted_string_delimiter,
  [ts_external_token__right_quoted_string_delimiter] = sym__right_quoted_string_delimiter,
  [ts_external_token_DQUOTE] = anon_sym_DQUOTE,
  [ts_external_token_line_number_directive] = sym_line_number_directive,
  [ts_external_token__null] = sym__null,
  [ts_external_token__error_sentinel] = sym__error_sentinel,
};

static const bool ts_external_scanner_states[10][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token_comment] = true,
    [ts_external_token__left_quoted_string_delimiter] = true,
    [ts_external_token__right_quoted_string_delimiter] = true,
    [ts_external_token_DQUOTE] = true,
    [ts_external_token_line_number_directive] = true,
    [ts_external_token__null] = true,
    [ts_external_token__error_sentinel] = true,
  },
  [2] = {
    [ts_external_token_comment] = true,
    [ts_external_token_line_number_directive] = true,
  },
  [3] = {
    [ts_external_token_comment] = true,
    [ts_external_token_DQUOTE] = true,
    [ts_external_token_line_number_directive] = true,
  },
  [4] = {
    [ts_external_token_comment] = true,
    [ts_external_token__left_quoted_string_delimiter] = true,
    [ts_external_token_DQUOTE] = true,
    [ts_external_token_line_number_directive] = true,
  },
  [5] = {
    [ts_external_token_comment] = true,
    [ts_external_token_DQUOTE] = true,
    [ts_external_token_line_number_directive] = true,
    [ts_external_token__null] = true,
  },
  [6] = {
    [ts_external_token_comment] = true,
    [ts_external_token_line_number_directive] = true,
    [ts_external_token__null] = true,
  },
  [7] = {
    [ts_external_token_comment] = true,
    [ts_external_token__right_quoted_string_delimiter] = true,
    [ts_external_token_line_number_directive] = true,
    [ts_external_token__null] = true,
  },
  [8] = {
    [ts_external_token_comment] = true,
    [ts_external_token__left_quoted_string_delimiter] = true,
    [ts_external_token_line_number_directive] = true,
  },
  [9] = {
    [ts_external_token_comment] = true,
    [ts_external_token__right_quoted_string_delimiter] = true,
    [ts_external_token_line_number_directive] = true,
  },
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_ocaml_type_external_scanner_create(void);
void tree_sitter_ocaml_type_external_scanner_destroy(void *);
bool tree_sitter_ocaml_type_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_ocaml_type_external_scanner_serialize(void *, char *);
void tree_sitter_ocaml_type_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_ocaml_type(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .supertype_map_slices = ts_supertype_map_slices,
    .supertype_map_entries = ts_supertype_map_entries,
    .supertype_symbols = ts_supertype_symbols,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .keyword_lex_fn = ts_lex_keywords,
    .keyword_capture_token = sym__lowercase_identifier,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_ocaml_type_external_scanner_create,
      tree_sitter_ocaml_type_external_scanner_destroy,
      tree_sitter_ocaml_type_external_scanner_scan,
      tree_sitter_ocaml_type_external_scanner_serialize,
      tree_sitter_ocaml_type_external_scanner_deserialize,
    },
    .primary_state_ids = ts_primary_state_ids,
    .name = "ocaml_type",
    .reserved_words = &ts_reserved_words[0][0],
    .max_reserved_word_set_size = 55,
    .metadata = {
      .major_version = 0,
      .minor_version = 25,
      .patch_version = 0,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
